function mpc = case1354pegase
%CASE1354PEGASE  Power flow data for medium part of European system.
%   Please see CASEFORMAT for details on the case file format.
%
%   This case accurately represents the size and complexity of part of the
%   European high voltage transmission network. The network contains 1,354
%   buses, 260 generators, and 1,991 branches and it operates at 380 and
%   220 kV. Please note that the data are fictitious and do not correspond
%   to real world data. They can thus be used to validate methods and tools
%   but should not be used for operation and planning of the European grid.
%
%   The data stems from the Pan European Grid Advanced Simulation and State
%   Estimation (PEGASE) project, part of the 7th Framework Program of the
%   European Union (https://www.fp7-pegase.com/).
%
%   When publishing results based on this data, please cite:
%
%     C. Josz, S. Fliscounakis, J. Maeght, and P. Panciatici, "AC Power Flow
%     Data in MATPOWER and QCQP Format: iTesla, RTE Snapshots, and PEGASE"
%     https://arxiv.org/abs/1603.01533
%
%     S. Fliscounakis, P. Panciatici, F. Capitanescu, and L. Wehenkel,
%     "Contingency ranking with respect to overloads in very large power
%     systems taking into account uncertainty, preventive and corrective
%     actions", Power Systems, IEEE Trans. on, (28)4:4909-4917, 2013.
%     https://doi.org/10.1109/TPWRS.2013.2251015
%
%   Remarks:
%
%   1. Line flow limits are 100 MVA (at 1 p.u. voltage) lower than the
%   current flow limits found in PEGASE data.
%
%   2. PEGASE data contains asymmetric shunt conductance and susceptance in
%   the PI transmission line model of branches. Thus total line charging
%   susceptance of branches is set to 0 p.u. and the nodal representation
%   of shunt condutance and susceptance is used. As a result, power flow
%   equations are left unchanged compared with original PEGASE data.
%   However, line flow constraints in the optimal flow problem are
%   modified.
%
%   3. Identical linear costs are used for all generators to form a loss
%   minimizing OPF objective function.
%
%   4. Since some parts of the network are aggregated, some generators
%   (e.g. with negative PMIN) represent aggregations of multiple loads
%   and generators.
%
%   Contacts:
%     Cédric Josz, Stéphane Fliscounakis, Jean Maeght, Patrick Panciatici
%     firstname.lastname@rte-france.com
%     Réseau de Transport d'Electricité (French Transmission System Operator)
%     Département Expertise Système, Immeuble "Le Colbert"
%     9 rue de la Porte de Buc, 78000 Versailles Cedex, France
%
%   March 18th, 2015

%   MATPOWER
%   Copyright (c) 2015, 2016 by Cédric Josz, Stéphane Fliscounakis, Jean Maeght,
%   and Patrick Panciatici
%   Licensed under the Creative Commons Attribution 4.0 International license,
%   https://creativecommons.org/licenses/by/4.0/

%% MATPOWER Case Format : Version 2
mpc.version = '2';

%%-----  Power Flow Data  -----%%
%% system MVA base
mpc.baseMVA = 100;

%% bus data
%	bus_i	type	Pd	Qd	Gs	Bs	area	Vm	Va	baseKV	zone	Vmax	Vmin
mpc.bus = [
	3	1	151	48.8	0	4.69	0	1.016674	-21.761632	220	5	1.1	0.9;
	4	1	171.41	23.4	0	2.1	0	1.026472	-7.007354	220	5	1.1	0.9;
	10	1	134	24.7	0	12.44	0	1.038534	-24.281644	220	5	1.1	0.9;
	21	1	161.2	39.3	0	30.71	0	1.036195	-21.958828	220	5	1.1	0.9;
	22	1	0	-0	0	0.85	0	1.035058	-18.483339	220	5	1.1	0.9;
	26	1	167.3	54.8	0	110.54	0	1.032267	-3.318448	380	5	1.1	0.9;
	29	1	152.9	49.4	0	27.65	0	0.994409	-13.816418	380	5	1.1	0.9;
	44	1	-0	0	0	7.27	0	1.045829	-22.181066	220	5	1.1	0.9;
	53	1	48.65	16.1	0	7.14	0	1.063422	-9.070013	220	5	1.1	0.9;
	58	1	0	0	0	2.31	0	0.983702	-21.159412	380	5	1.1	0.9;
	59	1	119.2	20.7	0	3.31	0	1.036808	-23.631999	220	5	1.1	0.9;
	90	1	363.25	115.39	0	3.95	0	1.042187	-27.546292	220	5	1.1	0.9;
	96	1	-8.4	28.66	0	3.49	0	1.030441	-15.689479	220	5	1.1	0.9;
	113	1	-0	-0	0	6.532711	0	1.030718	-14.242909	380	5	1.1	0.9;
	115	1	0	-0	0	4.21	0	1.066797	-19.783678	220	5	1.1	0.9;
	118	1	704.53	-35.33	0	13.102834	0	1.020256	-11.430028	380	5	1.1	0.9;
	124	2	0	0	0	0	0	1.081537	8.330657	220	5	1.1	0.9;
	128	1	23.73	12.7	0	0.52	0	1.020482	-16.063366	220	5	1.1	0.9;
	145	1	38.34	5.6	0	7.64	0	1.039096	-8.903623	220	5	1.1	0.9;
	148	1	32.34	12	0	6.1	0	1.041568	-7.991267	220	5	1.1	0.9;
	150	2	0	0	0	0	0	1.062805	2.531831	220	5	1.1	0.9;
	163	1	147.76	32.73	0	8.34	0	1.047047	-10.672351	220	5	1.1	0.9;
	171	1	100.8	13.2	0	5.24	0	1.069649	-25.153324	220	5	1.1	0.9;
	174	1	370.5	-17	0	9.91	0	1.074846	-27.942789	220	5	1.1	0.9;
	184	1	-0	-0	0	2.42	0	1.042771	-11.94851	220	5	1.1	0.9;
	188	1	36.84	13.9	0	3.3	0	1.018752	-19.550413	220	5	1.1	0.9;
	195	1	73.14	21.55	0	8.52	0	1.056337	-8.791116	220	5	1.1	0.9;
	196	1	0	-0	0	12.48	0	1.074825	-26.319857	220	5	1.1	0.9;
	198	1	39.74	11.8	0	1.44	0	1.01673	-16.306681	220	5	1.1	0.9;
	207	1	321.98	84.4	0	1.34	0	1.026604	-13.457588	220	5	1.1	0.9;
	216	1	155.6	54.7	0	2.21	0	1.039362	-24.590496	220	5	1.1	0.9;
	217	1	105.2	0.1	0	7.36	0	1.042762	-2.603638	220	5	1.1	0.9;
	218	1	69.83	15.12	0	2.23	0	1.045923	-13.114892	220	5	1.1	0.9;
	221	2	0	0	0	0	0	0.984682	-21.807603	380	5	1.1	0.9;
	225	1	0	0	0	1.15	0	1.063761	-20.174043	220	5	1.1	0.9;
	280	1	119.6	29.6	0	5.34	0	1.035725	-11.193679	220	5	1.1	0.9;
	283	1	27.33	11.9	0	1.17	0	1.035204	-12.094637	220	5	1.1	0.9;
	292	1	13.82	1.4	0	0.7	0	1.047544	-14.765137	220	5	1.1	0.9;
	305	1	103	12	0	0.63	0	1.067188	-32.434719	220	5	1.1	0.9;
	306	1	-0	0	0	4.76	0	1.030032	-18.823594	220	5	1.1	0.9;
	314	1	163.1	46.3	0	10.6	0	1.005548	-16.020947	220	5	1.1	0.9;
	333	1	524.5	43	0	6.85	0	1.014579	-42.137692	220	5	1.1	0.9;
	338	2	0	0	0	0	0	1.048249	-6.162533	220	5	1.1	0.9;
	346	1	197.5	48.3	0	0.71	0	1.040694	-21.298393	220	5	1.1	0.9;
	350	1	110.4	26.2	0	5.62	0	1.043185	-24.295117	220	5	1.1	0.9;
	352	2	0	0	0	0	0	1.055378	0.262953	380	5	1.1	0.9;
	367	1	65.17	-51.5	0	4.85	0	1.07164	-23.599251	220	5	1.1	0.9;
	401	1	56.5	36.29	0	3.08	0	1.026548	-14.262083	220	5	1.1	0.9;
	408	1	129.1	7.4	0	0.04	0	1.039791	-11.549917	220	5	1.1	0.9;
	410	1	170.89	50.38	0	6.44	0	1.041564	-26.439994	220	5	1.1	0.9;
	413	2	0	0	0	0	0	1.05013	-6.606455	220	5	1.1	0.9;
	416	1	70.68	13.5	0	6.96	0	1.0458	-6.728632	220	5	1.1	0.9;
	426	1	123.1	6.5	0	3.25	0	1.040489	-18.395441	220	5	1.1	0.9;
	432	1	-0	-0	0	68.517663	0	1.047774	2.403578	380	5	1.1	0.9;
	444	1	0	-0	0	2.45	0	1.036267	-12.201434	220	5	1.1	0.9;
	449	1	0	0	0	29.42	0	1.021739	-8.489958	380	5	1.1	0.9;
	453	2	0	0	0	0	0	1.018914	-7.857046	380	5	1.1	0.9;
	455	1	274.32	5.66	0	13.24	0	1.045882	-30.596498	220	5	1.1	0.9;
	490	1	20.52	4	0	2.77	0	1.036451	-23.014287	220	5	1.1	0.9;
	500	1	144.4	35.5	0	7.66	0	1.040423	-15.042452	220	5	1.1	0.9;
	502	1	-0	0	0	1.96	0	1.043564	-6.675538	220	5	1.1	0.9;
	506	1	-0	0	0	0.4	0	1.063297	-9.078728	380	5	1.1	0.9;
	513	1	57.46	17.1	0	1.61	0	1.038461	-7.443414	220	5	1.1	0.9;
	516	2	0	0	0	0	0	1.075592	-33.453515	220	5	1.1	0.9;
	520	1	118.5	20.4	0	2.57	0	1.048239	-27.108761	220	5	1.1	0.9;
	549	1	0	-0	0	5.15	0	1.074517	-10.959606	220	5	1.1	0.9;
	554	1	0	-0	0	1.97	0	1.024071	-20.793083	220	5	1.1	0.9;
	556	1	16.52	9	0	1.17	0	1.03694	-22.762979	220	5	1.1	0.9;
	561	1	0	-0	0	0.32	0	1.055451	-11.021763	220	5	1.1	0.9;
	564	2	0	0	0	0	0	1.069817	-47.078276	220	5	1.1	0.9;
	575	1	0	-0	0	0.67	0	1.04436	-11.163788	220	5	1.1	0.9;
	583	2	0	0	0	0	0	1.045397	-15.568574	220	5	1.1	0.9;
	594	1	67.57	17	0	5.75	0	1.036683	-8.869667	220	5	1.1	0.9;
	601	1	140.6	42.6	0	5.71	0	1.040685	-14.418335	220	5	1.1	0.9;
	604	1	33.24	4.4	0	24.54	0	0.996939	-12.93567	380	5	1.1	0.9;
	608	1	150.9	10.4	0	4.77	0	1.071753	-32.66829	220	5	1.1	0.9;
	609	1	-0	-0	0	5.67	0	1.025344	-14.779536	220	5	1.1	0.9;
	615	2	0	0	0	0	0	1.063876	-10.607341	220	5	1.1	0.9;
	616	2	0	0	0	0	0	1.048713	-12.997468	220	5	1.1	0.9;
	619	1	72.88	14.4	0	2.96	0	1.046742	-25.252227	220	5	1.1	0.9;
	639	2	0	0	0	0	0	1.051376	-29.999643	220	5	1.1	0.9;
	641	1	45.32	18.5	0	0.74	0	1.03515	-13.956622	220	5	1.1	0.9;
	658	1	258.4	66.7	0	5.87	0	1.033849	-17.442414	220	5	1.1	0.9;
	661	1	-0	-0	0	0.6	0	1.040021	2.414523	380	5	1.1	0.9;
	666	1	-5.61	-2.79	0	150.93	0	1.03076	-5.413919	380	5	1.1	0.9;
	678	1	232.1	89.8	0	0.09	0	0.986375	-18.62447	220	5	1.1	0.9;
	682	2	0	0	0	0	0	1.0691	-10.045044	220	5	1.1	0.9;
	687	1	-0	0	0	5.6	0	1.05742	-4.068613	220	5	1.1	0.9;
	707	1	83.99	21.2	0	5.36	0	1.054495	-25.620212	220	5	1.1	0.9;
	718	1	135.11	41	0	5.31	0	1.031148	-24.884639	220	5	1.1	0.9;
	720	1	-0	-0	0	23.554121	0	0.981824	-26.763103	380	5	1.1	0.9;
	726	1	400.3	189.9	0	0.53	0	1.048069	-4.933448	220	5	1.1	0.9;
	742	1	0	-0	0	2.83	0	1.05759	-4.070122	220	5	1.1	0.9;
	747	1	69.58	16	0	3.03	0	1.073922	-26.380554	220	5	1.1	0.9;
	749	2	0	0	0	0	0	1.057884	-28.950791	220	5	1.1	0.9;
	750	1	-0	-0	0	18.63	0	1.041141	-24.444747	220	5	1.1	0.9;
	757	2	0	0	0	0	0	1.017567	-13.765564	380	5	1.1	0.9;
	766	1	-0	0	0	7.85	0	1.047952	-19.8489	220	5	1.1	0.9;
	769	1	152.57	40	0	6.12	0	1.024104	-10.886119	220	5	1.1	0.9;
	772	1	343.9	141.9	0	5.55	0	1.034361	-14.167643	220	5	1.1	0.9;
	776	2	0	0	0	0	0	1.026368	-13.331099	220	5	1.1	0.9;
	778	2	0	0	0	0	0	1.049411	-6.680953	220	5	1.1	0.9;
	789	1	-0	-0	0	6.63	0	1.051689	-5.954643	220	5	1.1	0.9;
	795	2	0	0	0	0	0	1.035522	-15.484388	220	5	1.1	0.9;
	800	1	50.56	18	0	0.11	0	1.034778	-12.128589	220	5	1.1	0.9;
	803	2	0	0	0	0	0	1.041369	-24.768342	220	5	1.1	0.9;
	804	1	-0	-0	0	7.26	0	1.05585	-7.563603	220	5	1.1	0.9;
	805	1	124.5	29.6	0	15.4	0	1.05673	-11.12828	220	5	1.1	0.9;
	809	1	0	-0	0	58.088687	0	1.027525	-0.064839	380	5	1.1	0.9;
	819	1	0	-0	0	0.22	0	1.052169	-0.720463	220	5	1.1	0.9;
	823	2	0	0	0	0	0	1.044097	-2.229251	380	5	1.1	0.9;
	839	1	-0	-0	0	8.17514	0	1.024385	-1.386321	380	5	1.1	0.9;
	851	2	0	0	0	0	0	1.00025	-17.401395	220	5	1.1	0.9;
	858	2	0	0	0	0	0	1.031745	1.110134	380	5	1.1	0.9;
	870	1	110	3.4	0	11.95	0	1.040762	-24.602002	220	5	1.1	0.9;
	871	1	108.89	-21.76	0	3.72	0	1.056945	-11.955754	220	5	1.1	0.9;
	883	1	192.5	78.3	0	6.65	0	1.017176	-10.133417	220	5	1.1	0.9;
	891	2	0	0	0	0	0	1.064525	-18.97841	380	5	1.1	0.9;
	892	1	-0	-0	0	9.1	0	1.069234	-34.99461	220	5	1.1	0.9;
	900	1	0	0	0	0.15	0	1.050278	-15.385577	220	5	1.1	0.9;
	903	1	82.69	20.1	0	1.36	0	1.073885	-26.405705	220	5	1.1	0.9;
	905	1	631.71	-54.16	0	7.534129	0	1.035643	-15.075813	380	5	1.1	0.9;
	907	1	-0	-0	0	0.05	0	1.038555	-9.767436	220	5	1.1	0.9;
	908	1	125.9	-0.5	0	0.21	0	1.065214	-7.929625	220	5	1.1	0.9;
	920	1	34.44	8.5	0	1.11	0	1.045537	-20.769903	220	5	1.1	0.9;
	923	1	62.07	14.7	0	4.84	0	1.035596	-23.090183	220	5	1.1	0.9;
	933	1	-0	0	0	22.75	0	1.019946	-21.525393	220	5	1.1	0.9;
	935	1	-0	0	0	9.46	0	1.049182	-24.470238	220	5	1.1	0.9;
	953	1	108.7	-12.15	0	37.11	0	1.065973	-13.014498	380	5	1.1	0.9;
	954	1	66.27	13.6	0	0.67	0	1.044324	-21.525875	220	5	1.1	0.9;
	960	1	-0	-0	0	108.728602	0	1.063295	-9.078724	380	5	1.1	0.9;
	964	1	88.5	16.5	0	146.44	0	1.015553	-11.659761	380	5	1.1	0.9;
	972	2	0	0	0	0	0	1.072621	5.35893	380	5	1.3	0.7;
	980	1	14.71	4.6	0	7.88	0	1.047164	-25.138433	220	5	1.1	0.9;
	1001	2	0	0	0	0	0	1.074143	-33.62199	220	5	1.1	0.9;
	1002	2	0	0	0	0	0	1.041506	-20.086455	220	5	1.1	0.9;
	1005	1	0	0	0	0.81	0	1.053955	-12.379476	220	5	1.1	0.9;
	1015	1	44.55	0	0	4.78	0	1.056277	-5.398682	220	5	1.1	0.9;
	1026	1	201.4	53	0	9.67	0	1.028182	-29.352232	220	5	1.1	0.9;
	1027	1	-5.91	-2.33	0	30.393595	0	0.990542	-18.181223	380	5	1.1	0.9;
	1033	1	26.33	5.3	0	4.31	0	1.040934	-24.521854	220	5	1.1	0.9;
	1035	1	382.5	57.9	0	8.16	0	1.067815	-32.322562	220	5	1.1	0.9;
	1039	1	34.64	-0	0	0.91	0	1.034441	-19.543294	220	5	1.1	0.9;
	1040	1	31.94	8.8	0	3.79	0	1.038913	-9.027471	220	5	1.1	0.9;
	1043	2	0	0	0	0	0	1.028946	-3.693426	380	5	1.1	0.9;
	1051	1	-0	0	0	3.79	0	1.041627	-12.829065	220	5	1.1	0.9;
	1077	1	-0	0	0	0.1	0	1.035206	-12.094658	220	5	1.1	0.9;
	1078	1	-0	-0	0	4.74	0	1.061509	-3.472987	220	5	1.1	0.9;
	1081	1	277.86	58.95	0	182.880644	0	1.038644	-10.509999	380	5	1.1	0.9;
	1083	2	0	0	0	0	0	1.049472	-8.613737	220	5	1.1	0.9;
	1090	1	56.06	18.7	0	4.24	0	1.0101	-21.692579	220	5	1.1	0.9;
	1093	2	0	0	0	0	0	1.027056	-8.507308	380	5	1.1	0.9;
	1096	1	0	-0	0	0.16	0	1.050189	-1.116839	220	5	1.1	0.9;
	1100	2	0	0	0	0	0	1.042801	-4.348181	220	5	1.1	0.9;
	1101	1	-0	-0	0	2.05	0	1.033508	-14.209132	220	5	1.1	0.9;
	1102	2	0	0	0	0	0	1.049879	-6.64195	220	5	1.1	0.9;
	1105	1	29.83	7.3	0	6.49	0	1.074759	-26.325932	220	5	1.1	0.9;
	1111	1	0	0	0	1.1	0	1.036823	-8.871578	220	5	1.1	0.9;
	1129	1	190.47	55.03	0	6.7	0	1.003325	-22.532961	220	5	1.1	0.9;
	1136	1	16.22	2	0	0.71	0	1.034786	-18.539009	220	5	1.1	0.9;
	1137	1	-0	-0	0	0.05	0	1.048666	-24.42816	220	5	1.1	0.9;
	1146	1	32.84	-0	0	1.41	0	1.032638	-5.627486	220	5	1.1	0.9;
	1151	1	-0	-0	0	18.82	0	1.043354	-20.606997	220	5	1.1	0.9;
	1153	1	12.41	3.9	0	2.23	0	1.034654	-6.058841	220	5	1.1	0.9;
	1156	1	65.67	24	0	9.95	0	1.038055	-22.501868	220	5	1.1	0.9;
	1159	1	112.4	13	0	3.99	0	1.052615	-36.244174	220	5	1.1	0.9;
	1172	1	-0	-0	0	2.3	0	1.074468	-25.55724	220	5	1.1	0.9;
	1179	1	86.6	23.4	0	0.39	0	1.040896	-2.802576	220	5	1.1	0.9;
	1183	1	199.4	-19.9	0	6.68	0	1.066643	-33.081849	220	5	1.1	0.9;
	1187	1	67.57	21.2	0	5.71	0	1.039857	-4.955282	220	5	1.1	0.9;
	1188	1	30.93	6.6	0	9.09	0	1.054495	-14.222981	220	5	1.1	0.9;
	1194	1	129.6	13.3	0	13.98	0	1.03896	-15.127161	380	5	1.1	0.9;
	1198	1	0	0	0	30.637696	0	0.987964	-14.637206	380	5	1.1	0.9;
	1201	1	114.1	23.5	0	1.08	0	1.050304	-19.627666	220	5	1.1	0.9;
	1216	1	0	0	0	4.93	0	1.040626	-10.966354	220	5	1.1	0.9;
	1233	1	0	-0	0	1.72	0	1.037981	-22.526704	220	5	1.1	0.9;
	1234	1	71.78	17.9	0	1.96	0	1.022628	-16.524518	220	5	1.1	0.9;
	1237	2	0	0	0	0	0	1.108028	-6.285992	380	5	1.3	0.7;
	1248	1	0	0	0	1.17	0	1.047205	-10.673893	220	5	1.1	0.9;
	1249	1	348.93	26.47	0	9.36	0	1.019402	-11.26449	380	5	1.1	0.9;
	1251	2	0	0	0	0	0	1.027102	-8.516785	380	5	1.1	0.9;
	1262	1	-0	0	0	10.73	0	1.038607	-22.034538	220	5	1.1	0.9;
	1265	1	173	-21.3	0	5.93	0	1.066705	-50.057306	220	5	1.1	0.9;
	1275	1	52.46	2.9	0	0.35	0	1.032647	-18.856368	220	5	1.1	0.9;
	1295	2	0	0	0	0	0	1.03682	-8.643487	220	5	1.1	0.9;
	1301	1	479	99.2	0	10.68	0	1.05167	-28.720012	220	5	1.1	0.9;
	1305	1	-0	-0	0	4.34	0	1.038253	-22.269637	220	5	1.1	0.9;
	1311	1	37.94	15.2	0	35.31	0	1.033986	-21.654604	380	5	1.1	0.9;
	1326	1	0	0	0	0.98	0	1.069384	-34.995605	220	5	1.1	0.9;
	1334	1	0	0	0	0.59	0	1.025188	-16.314102	220	5	1.1	0.9;
	1341	2	0	0	0	0	0	1.032899	-14.204982	220	5	1.1	0.9;
	1343	1	-0	-0	0	58.3	0	0.996377	-14.200565	380	5	1.1	0.9;
	1354	2	0	0	0	0	0	1.041045	-7.375163	220	5	1.1	0.9;
	1355	1	-3.29	-1.32	0	190.704588	0	1.0328	-8.554156	380	5	1.1	0.9;
	1364	1	40.85	14.5	0	0.59	0	1.039851	-12.092524	220	5	1.1	0.9;
	1380	1	145.2	21.4	0	1.09	0	1.033661	-22.574327	220	5	1.1	0.9;
	1394	1	30.93	7.6	0	4.69	0	1.074659	-26.314245	220	5	1.1	0.9;
	1397	1	0	-0	0	2.14	0	1.037924	-12.820941	220	5	1.1	0.9;
	1398	1	196.7	44.2	0	0.1	0	1.017322	-19.003234	220	5	1.1	0.9;
	1401	1	11.91	2.9	0	5.27	0	1.038647	-14.758146	220	5	1.1	0.9;
	1414	1	144.7	20.3	0	6.35	0	1.022683	-27.895964	220	5	1.1	0.9;
	1415	1	0	-0	0	0.3	0	1.032141	-11.506028	220	5	1.1	0.9;
	1422	2	0	0	0	0	0	1.04353	-20.966559	220	5	1.1	0.9;
	1435	1	-0	0	0	3.85	0	1.044674	-24.345986	220	5	1.1	0.9;
	1436	2	0	0	0	0	0	1.047684	-12.917585	220	5	1.1	0.9;
	1448	1	0	0	0	9.49	0	1.075069	-15.319825	220	5	1.1	0.9;
	1459	1	-0	-0	0	3.38	0	1.042934	-21.817757	220	5	1.1	0.9;
	1462	1	-0	0	0	0.42	0	1.043552	-20.967069	220	5	1.1	0.9;
	1465	1	-11.03	-5.11	0	29.706803	0	0.988454	-15.339654	380	5	1.1	0.9;
	1478	2	0	0	0	0	0	1.048068	-15.82908	220	5	1.1	0.9;
	1483	1	0	0	0	4.86	0	1.064156	-26.867437	220	5	1.1	0.9;
	1486	1	79.59	18.8	0	6.78	0	1.034956	-23.34269	220	5	1.1	0.9;
	1494	1	-1.85	-1.63	0	31.424588	0	1.034255	-11.521816	380	5	1.1	0.9;
	1502	1	-0	0	0	37.116157	0	1.027764	-26.693479	380	5	1.1	0.9;
	1504	1	0	-0	0	1.84	0	1.025322	-14.28951	220	5	1.1	0.9;
	1526	1	-10.86	-5.93	0	12.54	0	1.042753	-11.9479	220	5	1.1	0.9;
	1538	1	36.34	15.2	0	0.3	0	1.033256	-8.75699	220	5	1.1	0.9;
	1539	1	73.58	3.24	0	52.35	0	1.059384	-15.296455	220	5	1.1	0.9;
	1541	1	125.4	45.8	0	0	0	1.039282	-23.538177	220	5	1.1	0.9;
	1545	1	42.55	8.2	0	2.06	0	1.042222	-23.472695	220	5	1.1	0.9;
	1547	1	-0	-0	0	1.02	0	1.040481	-15.042995	220	5	1.1	0.9;
	1551	1	-0	-0	0	7.99	0	1.043158	-20.41875	220	5	1.1	0.9;
	1552	1	20.42	8.1	0	35.15	0	1.042604	-16.620759	380	5	1.1	0.9;
	1554	1	0	-0	0	4.96	0	1.020729	-16.043428	220	5	1.1	0.9;
	1556	1	-0	-0	0	0.95	0	1.061796	-11.391738	220	5	1.1	0.9;
	1562	1	77.39	24.1	0	2.67	0	1.042236	-25.854486	220	5	1.1	0.9;
	1566	1	156.9	-4	0	2.48	0	1.036219	-24.756749	220	5	1.1	0.9;
	1568	1	69.08	21.5	0	1.65	0	1.046858	-25.154702	220	5	1.1	0.9;
	1578	1	0	-0	0	1.38	0	1.040417	-12.029853	220	5	1.1	0.9;
	1584	1	0	-0	0	3.24	0	1.026871	-14.239921	220	5	1.1	0.9;
	1592	1	28.33	44.8	0	0.51	0	1.037512	-2.335282	220	5	1.1	0.9;
	1595	1	71.77	13.57	0	3.84	0	1.01584	-29.283839	220	5	1.1	0.9;
	1604	2	0	0	0	0	0	1.043862	-6.640235	220	5	1.1	0.9;
	1605	1	-0	0	0	0.05	0	1.076968	-24.357645	220	5	1.1	0.9;
	1607	1	73.78	-0	0	2.69	0	1.030671	-5.514136	380	5	1.1	0.9;
	1609	1	0	0	0	6.55	0	1.05016	-20.676376	220	5	1.1	0.9;
	1625	1	73.18	22.3	0	8.77	0	1.047827	-19.847235	220	5	1.1	0.9;
	1629	1	41.95	8.5	0	1.05	0	1.043667	-11.062381	220	5	1.1	0.9;
	1642	2	0	0	0	0	0	1.028554	0.427747	380	5	1.1	0.9;
	1643	1	95.71	25.9	0	11.38	0	1.042277	-25.887944	220	5	1.1	0.9;
	1644	1	449.56	245.72	0	94.55	0	1.054472	3.4279	380	5	1.1	0.9;
	1662	1	3.67	-1.86	0	3.89	0	1.046228	-20.80193	220	5	1.1	0.9;
	1672	1	-28.13	-0.83	0	0.1	0	1.061821	-11.405066	220	5	1.1	0.9;
	1680	2	0	0	0	0	0	1.04395	-2.225166	380	5	1.1	0.9;
	1692	1	83.19	25.8	0	0.33	0	1.047275	-10.47462	220	5	1.1	0.9;
	1704	1	62.16	8	0	2.99	0	1.038923	-13.052366	220	5	1.1	0.9;
	1708	2	0	0	0	0	0	1.037938	-13.186223	380	5	1.1	0.9;
	1709	1	-15.38	-4.94	0	33.241519	0	1.007462	-14.734072	380	5	1.1	0.9;
	1711	1	25.13	-61.6	0	3.68	0	1.052686	-29.767092	220	5	1.1	0.9;
	1721	2	0	0	0	0	0	1.062238	2.466713	220	5	1.1	0.9;
	1730	1	-17.18	-4.35	0	44.419365	0	1.025799	-14.51076	380	5	1.1	0.9;
	1742	1	0	0	0	6.68	0	1.030326	-32.491405	220	5	1.1	0.9;
	1746	1	-0	0	0	1.86	0	1.049354	-24.089302	220	5	1.1	0.9;
	1750	1	58.46	6.6	0	10.13	0	1.036578	-19.191311	220	5	1.1	0.9;
	1754	2	0	0	0	0	0	1.04286	-17.149587	380	5	1.1	0.9;
	1758	1	0	-0	0	121.29	0	1.03465	-4.424306	380	5	1.1	0.9;
	1763	1	-0	-0	0	0.28	0	1.034875	-30.808254	380	5	1.1	0.9;
	1767	1	333.5	-1.4	0	8.37	0	1.049612	-45.569227	220	5	1.1	0.9;
	1768	1	73.98	22.5	0	1.15	0	1.042927	-11.13756	220	5	1.1	0.9;
	1775	1	-0	0	0	1.64	0	1.05106	-25.97492	220	5	1.1	0.9;
	1794	2	0	0	0	0	0	1.05295	-0.236871	380	5	1.1	0.9;
	1798	1	-0.63	-0.83	0	22.810288	0	0.987325	-7.424983	380	5	1.1	0.9;
	1803	1	-0	-0	0	0.05	0	1.038883	-22.036572	220	5	1.1	0.9;
	1808	2	0	0	0	0	0	1.046224	-0.834506	220	5	1.1	0.9;
	1813	1	88.92	19.79	0	2.81	0	1.029057	-7.519944	220	5	1.1	0.9;
	1817	1	108.91	33.7	0	56.24	0	1.069539	-12.567622	380	5	1.1	0.9;
	1833	1	0	0	0	3.33	0	1.036371	-22.480232	220	5	1.1	0.9;
	1838	1	0	-0	0	3.02	0	1.037419	-22.162507	220	5	1.1	0.9;
	1844	1	-28.14	-0.76	0	0.1	0	1.061839	-11.404513	220	5	1.1	0.9;
	1851	2	0	0	0	0	0	1.04762	-12.927099	220	5	1.1	0.9;
	1852	2	0	0	0	0	0	1.054325	-25.744509	220	5	1.1	0.9;
	1857	1	0	0	0	1.53	0	1.028608	-11.055937	220	5	1.1	0.9;
	1860	1	88.8	-3.2	0	13.63	0	1.060858	-25.532569	220	5	1.1	0.9;
	1866	1	209.2	23.5	0	6.44	0	1.04474	-28.285319	220	5	1.1	0.9;
	1868	1	11.71	3.8	0	2.96	0	1.035258	-15.877966	220	5	1.1	0.9;
	1876	1	-0	-0	0	6.273568	0	1.030797	-14.242223	380	5	1.1	0.9;
	1883	1	27.33	5	0	6.77	0	1.029787	-16.425673	220	5	1.1	0.9;
	1888	2	0	0	0	0	0	1.039377	-2.376673	220	5	1.1	0.9;
	1895	1	20.82	-0	0	0.14	0	1.070353	7.588403	220	5	1.1	0.9;
	1896	1	-0	-0	0	1.18	0	1.021131	-6.153674	380	5	1.1	0.9;
	1910	1	-0	0	0	2.68	0	1.049243	-6.695385	220	5	1.1	0.9;
	1914	2	0	0	0	0	0	1.035636	-20.619184	220	5	1.1	0.9;
	1917	1	197.5	-31	0	17.34	0	1.050486	-25.132946	220	5	1.1	0.9;
	1923	1	192.11	45.5	0	68.04	0	1.002407	-11.646061	380	5	1.1	0.9;
	1940	1	0	0	0	1.19	0	1.03614	-18.037347	220	5	1.1	0.9;
	1959	2	0	0	0	0	0	1.064692	-10.511132	220	5	1.1	0.9;
	1965	1	248.7	8.1	0	12.46	0	1.066548	-49.702693	220	5	1.1	0.9;
	1973	1	210.1	66.2	0	59.36	0	1.074741	-26.300211	220	5	1.1	0.9;
	1979	1	0	0	0	0.14	0	1.074144	-33.621995	220	5	1.1	0.9;
	1980	1	177.3	36.4	0	0.7	0	1.040983	-21.42173	220	5	1.1	0.9;
	1998	1	25.73	7.3	0	2.82	0	1.033294	-20.141362	220	5	1.1	0.9;
	2012	1	41.45	4.9	0	18.43	0	1.038506	-18.675791	220	5	1.1	0.9;
	2019	1	72.08	21.3	0	7.19	0	1.044528	-21.078894	220	5	1.1	0.9;
	2020	1	0	0	0	9.98	0	1.051074	-24.096673	220	5	1.1	0.9;
	2021	1	3.9	1	0	3.81	0	1.050096	-24.96566	220	5	1.1	0.9;
	2035	2	0	0	0	0	0	1.03088	0.693557	380	5	1.1	0.9;
	2042	1	14.02	6.5	0	0.09	0	1.037536	-5.903671	220	5	1.1	0.9;
	2043	1	3.2	1.7	0	0.05	0	1.043517	-20.967534	220	5	1.1	0.9;
	2044	1	0	0	0	0.14	0	1.032138	-12.426968	220	5	1.1	0.9;
	2050	2	0	0	0	0	0	1.009914	-1.144693	380	5	1.1	0.9;
	2056	1	-0	0	0	7.338808	0	1.038215	-9.076082	380	5	1.1	0.9;
	2057	1	120.2	20.5	0	19.94	0	1.04971	-30.773674	220	5	1.1	0.9;
	2072	1	101.2	2.1	0	38.43	0	1.065746	-17.405584	380	5	1.1	0.9;
	2078	1	77.19	18.6	0	2.22	0	1.020675	-14.721188	220	5	1.1	0.9;
	2079	1	0	-0	0	13.94	0	1.050272	-12.128304	220	5	1.1	0.9;
	2083	1	121.4	20.2	0	16.61	0	1.035	-18.482915	220	5	1.1	0.9;
	2085	2	0	0	0	0	0	1.040351	2.467063	380	5	1.1	0.9;
	2088	1	-0	0	0	6.02	0	1.052176	-12.052739	220	5	1.1	0.9;
	2089	1	199.86	56.93	0	46.83	0	1.04312	-25.770522	220	5	1.1	0.9;
	2093	1	-5.64	-2.84	0	34.582195	0	1.016671	-8.086067	380	5	1.1	0.9;
	2101	1	-0	-0	0	38.32	0	1.02998	-5.549174	380	5	1.1	0.9;
	2128	1	79.09	18.7	0	10.8	0	1.035961	-23.062582	220	5	1.1	0.9;
	2129	1	0	0	0	56.686515	0	1.016495	-8.242431	380	5	1.1	0.9;
	2132	1	107.3	14.2	0	20.23	0	1.039669	-15.783021	220	5	1.1	0.9;
	2142	1	0	-0	0	7.63	0	1.050391	-7.902365	220	5	1.1	0.9;
	2155	1	63.67	20.2	0	6.13	0	1.037645	-5.893797	220	5	1.1	0.9;
	2161	1	35.66	-9.39	0	7.89	0	1.048364	-28.662661	220	5	1.1	0.9;
	2166	1	20.92	7.7	0	1.21	0	1.052887	-14.673236	220	5	1.1	0.9;
	2174	1	0	0	0	61.795782	0	1.011944	-13.202451	380	5	1.1	0.9;
	2177	2	0	0	0	0	0	1.07333	-28.416356	220	5	1.1	0.9;
	2183	1	0	0	0	32.849821	0	0.995876	-16.181883	380	5	1.1	0.9;
	2189	1	-0	0	0	0.18	0	1.047761	-6.28234	220	5	1.1	0.9;
	2197	2	0	0	0	0	0	1.035528	-6.012095	380	5	1.1	0.9;
	2208	1	74.98	26.9	0	2.96	0	1.051126	-13.420372	220	5	1.1	0.9;
	2229	1	0	-0	0	2.91	0	1.062959	-13.080678	220	5	1.1	0.9;
	2230	1	0	-0	0	1.71	0	1.036513	-22.481463	220	5	1.1	0.9;
	2231	1	0	0	0	0.45	0	1.049391	1.166976	220	5	1.1	0.9;
	2242	1	-0	-0	0	2.76	0	1.02489	-19.020985	220	5	1.1	0.9;
	2252	1	100.7	-4.2	0	1.57	0	1.068023	-47.917821	220	5	1.1	0.9;
	2273	1	141.3	32.4	0	7.29	0	1.074359	-26.339387	220	5	1.1	0.9;
	2276	2	0	0	0	0	0	1.05324	-23.758819	220	5	1.1	0.9;
	2286	1	208.88	60.8	0	6.68	0	1.023269	-8.717089	220	5	1.1	0.9;
	2288	1	170.9	39	0	11.25	0	1.034468	-8.57662	220	5	1.1	0.9;
	2291	2	0	0	0	0	0	1.072528	-10.610728	220	5	1.1	0.9;
	2303	1	0	-0	0	7.94	0	1.03744	-15.856049	220	5	1.1	0.9;
	2304	1	88.7	-0.1	0	2.83	0	1.058495	-25.832763	220	5	1.1	0.9;
	2308	1	32.04	6.6	0	0.44	0	1.038612	-24.078628	220	5	1.1	0.9;
	2313	1	0	-0	0	0.41	0	1.046654	-21.265608	220	5	1.1	0.9;
	2319	1	39.64	20.9	0	0.65	0	1.061069	-8.962289	220	5	1.1	0.9;
	2327	1	114.5	25.2	0	17.57	0	1.047194	-25.133812	220	5	1.1	0.9;
	2328	1	-0	0	0	0.4	0	1.017489	-14.250876	220	5	1.1	0.9;
	2337	1	103.4	27	0	1.41	0	1.045493	-29.7818	220	5	1.1	0.9;
	2340	1	96.91	13.9	0	3.98	0	1.03958	-12.236094	220	5	1.1	0.9;
	2341	1	-0	0	0	3.28	0	1.065118	-26.070582	220	5	1.1	0.9;
	2359	2	0	0	0	0	0	1.024884	-9.921886	380	5	1.1	0.9;
	2360	1	215.5	31.3	0	14.23	0	1.044563	-21.085033	220	5	1.1	0.9;
	2361	1	16.9	3.85	0	25.09	0	1.052288	-29.004707	220	5	1.1	0.9;
	2365	1	37.94	15.4	0	117.386167	0	1.033702	-22.17424	380	5	1.1	0.9;
	2372	1	0	-0	0	14.490904	0	0.991274	-17.482034	380	5	1.1	0.9;
	2377	1	101.7	17.6	0	13.14	0	1.036655	-14.897845	220	5	1.1	0.9;
	2393	1	20.82	0	0	0.13	0	1.070495	7.700354	220	5	1.1	0.9;
	2406	1	0	0	0	1.69	0	1.048666	-24.428158	220	5	1.1	0.9;
	2421	2	0	0	0	0	0	1.047766	-23.702161	220	5	1.1	0.9;
	2424	1	44.55	16.8	0	2.31	0	1.034979	-22.416002	220	5	1.1	0.9;
	2425	2	0	0	0	0	0	1.041186	-9.882232	220	5	1.1	0.9;
	2426	2	0	0	0	0	0	1.037687	-13.762974	220	5	1.1	0.9;
	2430	1	-0	0	0	6.21	0	1.041369	-11.921713	220	5	1.1	0.9;
	2432	1	291.2	35.4	0	21.87	0	1.030479	-32.300033	220	5	1.1	0.9;
	2438	1	-0	0	0	0.64	0	1.046626	-21.265307	220	5	1.1	0.9;
	2446	2	0	0	0	0	0	1.06157	7.557722	380	5	1.1	0.9;
	2457	1	30.73	9.6	0	2.75	0	1.040059	-24.619174	220	5	1.1	0.9;
	2458	1	132	26	0	27.18	0	1.006537	-13.460583	380	5	1.1	0.9;
	2467	1	0	-0	0	11.65	0	1.018095	-14.100158	220	5	1.1	0.9;
	2468	2	0	0	0	0	0	1.048132	-6.215141	220	5	1.1	0.9;
	2475	1	0	0	0	0.21	0	1.065573	-7.808732	220	5	1.1	0.9;
	2479	1	-2.65	-1.17	0	65.673108	0	1.008158	-2.627166	380	5	1.1	0.9;
	2481	2	0	0	0	0	0	1.042777	-18.185589	220	5	1.1	0.9;
	2489	2	0	0	0	0	0	1.049366	1.167288	220	5	1.1	0.9;
	2503	1	-0	0	0	2.09	0	1.012856	-21.48175	220	5	1.1	0.9;
	2510	1	115.87	29.41	0	7.92	0	1.028486	-12.08393	220	5	1.1	0.9;
	2518	1	-5.29	-1.91	0	42.622979	0	0.999088	-11.536983	380	5	1.1	0.9;
	2526	1	238.45	46.86	0	125.534496	0	1.044376	-19.855823	380	5	1.1	0.9;
	2527	1	59.77	16.7	0	0.64	0	1.043988	-4.63491	220	5	1.1	0.9;
	2535	1	51.56	2.1	0	0.18	0	1.037544	-2.896669	220	5	1.1	0.9;
	2550	2	0	0	0	0	0	1.049384	-6.666685	220	5	1.1	0.9;
	2558	1	159.27	38.31	0	2.18	0	1.023305	-16.640241	220	5	1.1	0.9;
	2563	1	52.86	13	0	6.04	0	1.046977	-4.707678	220	5	1.1	0.9;
	2575	1	62.97	17.4	0	0.3	0	1.04159	-2.724929	220	5	1.1	0.9;
	2591	1	253.77	84.6	0	20.71	0	1.030923	-15.233022	220	5	1.1	0.9;
	2597	1	14.72	5.4	0	0.72	0	1.033106	-10.863404	220	5	1.1	0.9;
	2598	1	25.63	12.7	0	8.42	0	1.040506	-15.286014	220	5	1.1	0.9;
	2600	2	0	0	0	0	0	1.038821	-13.64341	220	5	1.1	0.9;
	2627	2	0	0	0	0	0	1.061488	7.521439	380	5	1.1	0.9;
	2629	1	203.1	65.2	0	0.64	0	1.045498	-24.492103	220	5	1.1	0.9;
	2641	1	0	0	0	1.72	0	1.042124	-11.251293	220	5	1.1	0.9;
	2644	1	0	0	0	0.19	0	1.07369	-24.808555	220	5	1.1	0.9;
	2653	2	0	0	0	0	0	1.036517	-23.007774	220	5	1.1	0.9;
	2654	1	207.1	50.4	0	4.93	0	1.019228	-18.770253	220	5	1.1	0.9;
	2656	1	-0.72	0	0	69.582294	0	1.034308	-22.749906	380	5	1.1	0.9;
	2676	1	123.24	23.65	0	6.44	0	1.035697	-14.93988	220	5	1.1	0.9;
	2689	1	0	0	0	0.45	0	1.034001	-3.213772	220	5	1.1	0.9;
	2695	1	16.22	8.9	0	2.2	0	1.038344	-21.004788	220	5	1.1	0.9;
	2702	1	151.7	27.6	0	6.49	0	1.043222	-25.609448	220	5	1.1	0.9;
	2719	2	0	0	0	0	0	1.035384	-15.495326	220	5	1.1	0.9;
	2721	1	-0	0	0	10.040547	0	0.983623	-22.371652	380	5	1.1	0.9;
	2732	1	85.59	20.72	0	142.79	0	1.013054	-10.064991	380	5	1.1	0.9;
	2749	1	55.56	26.5	0	1.09	0	1.049925	-30.124895	220	5	1.1	0.9;
	2751	1	34.54	8.6	0	5.05	0	1.042162	-21.896741	220	5	1.1	0.9;
	2754	1	-0	0	0	3.72	0	1.051994	-13.365671	220	5	1.1	0.9;
	2770	1	85.39	14.3	0	0.35	0	1.047946	-9.667144	220	5	1.1	0.9;
	2782	1	-28.13	-0.81	0	0.1	0	1.06184	-11.404239	220	5	1.1	0.9;
	2786	2	0	0	0	0	0	1.073795	6.427789	380	5	1.3	0.7;
	2794	1	0	-0	0	0.39	0	1.035004	-18.482955	220	5	1.1	0.9;
	2795	1	-0	-0	0	1.94	0	1.070164	-32.44656	220	5	1.1	0.9;
	2797	2	0	0	0	0	0	1.042226	-4.483231	220	5	1.1	0.9;
	2799	2	0	0	0	0	0	1.05396	-12.355638	220	5	1.1	0.9;
	2801	1	0	-0	0	12.17	0	1.041023	-24.502823	220	5	1.1	0.9;
	2806	1	-0	-0	0	0.39	0	1.050477	0.170291	220	5	1.1	0.9;
	2815	1	25.33	14.2	0	4.05	0	1.03655	-22.984441	220	5	1.1	0.9;
	2816	2	0	0	0	0	0	1.053944	-0.390904	380	5	1.1	0.9;
	2841	2	0	0	0	0	0	1.08248	-26.223645	220	5	1.1	0.9;
	2842	2	0	0	0	0	0	1.043712	-6.650222	220	5	1.1	0.9;
	2847	1	40.14	8.8	0	0.82	0	1.050663	-3.956106	220	5	1.1	0.9;
	2848	1	453.23	139.16	0	25.583218	0	0.998047	-7.830482	380	5	1.1	0.9;
	2850	1	-0	-0	0	4.76	0	1.03307	-18.741426	220	5	1.1	0.9;
	2854	1	-12.33	-2.65	0	69.323028	0	1.044838	-10.677353	380	5	1.1	0.9;
	2863	1	37.73	13.84	0	0.09	0	1.028402	-11.074681	220	5	1.1	0.9;
	2866	1	-0	0	0	4.21	0	1.03784	-7.272198	220	5	1.1	0.9;
	2872	2	0	0	0	0	0	1.063841	-10.631082	220	5	1.1	0.9;
	2877	1	29.93	-0	0	0.46	0	1.041807	-14.920901	220	5	1.1	0.9;
	2878	2	0	0	0	0	0	1.057626	-10.952158	220	5	1.1	0.9;
	2886	2	0	0	0	0	0	1.055151	-14.128145	220	5	1.1	0.9;
	2888	1	-0	0	0	0.22	0	1.013722	-21.258074	220	5	1.1	0.9;
	2889	1	-0	0	0	3.2	0	1.050185	-1.116811	220	5	1.1	0.9;
	2898	1	56.76	13.4	0	1.18	0	1.035844	-23.070363	220	5	1.1	0.9;
	2902	2	0	0	0	0	0	1.059072	-23.273273	220	5	1.1	0.9;
	2910	1	-0	0	0	2.87	0	1.040554	-2.262281	220	5	1.1	0.9;
	2918	1	0	0	0	53.25003	0	1.02385	-28.205516	380	5	1.1	0.9;
	2919	1	-0	0	0	32.049116	0	0.991588	-18.308905	380	5	1.1	0.9;
	2924	1	19.92	7.4	0	17.28	0	1.057573	-33.080856	220	5	1.1	0.9;
	2928	1	0	-0	0	92.38	0	1.074568	-26.335879	220	5	1.1	0.9;
	2930	2	0	0	0	0	0	1.047695	-12.912307	220	5	1.1	0.9;
	2931	1	230.7	19.3	0	4.56	0	1.043622	-32.26389	220	5	1.1	0.9;
	2934	2	0	0	0	0	0	1.051149	-6.52803	220	5	1.1	0.9;
	2938	1	0	-0	0	137.29	0	1.005521	-7.42176	380	5	1.1	0.9;
	2940	1	103	28.7	0	2.88	0	1.01671	-14.352513	220	5	1.1	0.9;
	2949	1	-0	-0	0	11.03	0	1.050433	-4.837602	220	5	1.1	0.9;
	2961	1	159.9	53.2	0	7.99	0	1.041215	-24.387727	220	5	1.1	0.9;
	2967	1	0	-0	0	0.19	0	1.046973	-17.07434	220	5	1.1	0.9;
	2968	1	112.51	16.51	0	1.63	0	1.044527	-8.952764	220	5	1.1	0.9;
	2972	1	97.82	20.59	0	6.67	0	1.038342	-15.064035	220	5	1.1	0.9;
	2980	1	-0	0	0	0.42	0	1.063297	-9.078731	380	5	1.1	0.9;
	2981	1	-0	0	0	1.68	0	1.075422	-15.322665	220	5	1.1	0.9;
	2985	2	0	0	0	0	0	1.053607	-8.47854	220	5	1.1	0.9;
	2995	1	0	-0	0	3.9	0	1.051891	-3.763961	220	5	1.1	0.9;
	3013	1	2.2	1.1	0	0.67	0	1.049575	-5.241034	220	5	1.1	0.9;
	3018	2	0	0	0	0	0	1.041705	-12.798969	220	5	1.1	0.9;
	3019	1	0	-0	0	1.37	0	1.036837	-22.763335	220	5	1.1	0.9;
	3021	1	74.38	8.8	0	6.1	0	1.042535	-11.458053	220	5	1.1	0.9;
	3022	1	149	19.8	0	23.89	0	1.038441	-14.928463	380	5	1.1	0.9;
	3028	2	0	0	0	0	0	1.030654	0.605264	380	5	1.1	0.9;
	3036	1	116.93	33.6	0	3.35	0	1.046808	-25.175878	220	5	1.1	0.9;
	3037	1	221	146.8	0	8.35	0	1.04857	-6.585978	220	5	1.1	0.9;
	3044	1	29.33	5.6	0	6.56	0	1.044991	-21.014316	220	5	1.1	0.9;
	3051	1	0	0	0	0	0	1.031148	-24.884639	220	5	1.1	0.9;
	3069	1	89.6	22.7	0	4.06	0	1.033466	-8.740353	220	5	1.1	0.9;
	3070	1	0	-0	0	4.27	0	1.013716	-21.258029	220	5	1.1	0.9;
	3071	1	0	-0	0	2.83	0	1.044715	-7.284056	220	5	1.1	0.9;
	3072	1	92.9	27.7	0	5.22	0	1.050676	-30.097241	220	5	1.1	0.9;
	3075	1	30.33	10.5	0	0.33	0	1.020662	-16.898848	220	5	1.1	0.9;
	3082	1	0	0	0	75.076038	0	0.998938	-15.122127	380	5	1.1	0.9;
	3083	1	567.97	130.73	0	1.09	0	1.068389	-27.641331	220	5	1.1	0.9;
	3085	1	64.67	24.5	0	1.46	0	1.067218	-27.697899	220	5	1.1	0.9;
	3112	1	-0	0	0	0	0	1.057299	2.403578	220	5	1.1	0.9;
	3114	2	0	0	0	0	0	1.071964	-11.598423	220	5	1.1	0.9;
	3121	1	242.16	73.64	0	12.79	0	1.031627	-7.214885	220	5	1.1	0.9;
	3126	1	104.78	20.92	0	6.17	0	1.021547	-17.925447	220	5	1.1	0.9;
	3133	2	0	0	0	0	0	1.065372	-9.169038	220	5	1.1	0.9;
	3134	2	0	0	0	0	0	1.023117	-15.704702	220	5	1.1	0.9;
	3137	1	0	-0	0	0.03	0	1.065118	-26.070583	220	5	1.1	0.9;
	3145	1	860.95	-182.67	0	18.309116	0	0.993017	-41.441031	380	5	1.1	0.9;
	3166	1	184.91	32.8	0	8.35	0	1.039048	-17.527734	220	5	1.1	0.9;
	3183	2	0	0	0	0	0	1.055515	-11.832574	220	5	1.1	0.9;
	3184	1	14.12	7.7	0	1.42	0	1.040299	-21.338667	220	5	1.1	0.9;
	3187	1	107	34.7	0	1.54	0	1.042404	-12.953299	220	5	1.1	0.9;
	3191	1	127.8	15.2	0	9.34	0	1.040776	-12.071062	220	5	1.1	0.9;
	3200	1	86.2	19.9	0	5.43	0	1.036168	-23.053307	220	5	1.1	0.9;
	3204	1	8.51	3.2	0	0.96	0	1.017023	-10.267075	220	5	1.1	0.9;
	3205	2	0	0	0	0	0	1.041982	-1.292668	380	5	1.1	0.9;
	3218	2	0	0	0	0	0	1.028507	-11.253521	220	5	1.1	0.9;
	3221	1	76.38	24.8	0	3.24	0	1.047766	-12.146786	220	5	1.1	0.9;
	3231	1	251.3	42.6	0	5.47	0	1.025161	-14.287368	220	5	1.1	0.9;
	3239	1	-5.06	-1.84	0	187.327025	0	1.056229	-11.611515	380	5	1.1	0.9;
	3240	2	0	0	0	0	0	1.050305	-1.129073	220	5	1.1	0.9;
	3241	1	132	27	0	4.66	0	1.030192	-17.957576	220	5	1.1	0.9;
	3243	1	-0	0	0	1.18	0	0.990554	-18.181283	380	5	1.1	0.9;
	3246	1	20.9	-171.5	0	0.21	0	1.080513	7.863586	220	5	1.1	0.9;
	3248	1	84.19	17.3	0	6.2	0	1.05498	-11.325947	220	5	1.1	0.9;
	3255	1	0	0	0	1.44	0	1.038882	-22.03657	220	5	1.1	0.9;
	3276	1	-0	-0	0	0.13	0	1.050076	-7.893785	220	5	1.1	0.9;
	3293	1	33.14	11.6	0	4.51	0	1.038075	-11.39812	220	5	1.1	0.9;
	3306	2	0	0	0	0	0	1.071978	8.08029	220	5	1.1	0.9;
	3324	2	0	0	0	0	0	1.014126	-16.129669	220	5	1.1	0.9;
	3325	1	-0	0	0	0.91	0	1.033998	-20.055869	220	5	1.1	0.9;
	3327	1	-0	-0	0	2.27	0	1.038066	-12.156606	220	5	1.1	0.9;
	3331	1	-0	-0	0	4.56	0	1.0175	-10.215862	220	5	1.1	0.9;
	3344	1	83.29	24.1	0	1.23	0	1.039524	-16.595479	220	5	1.1	0.9;
	3346	2	0	0	0	0	0	1.01594	-12.063974	380	5	1.1	0.9;
	3353	2	0	0	0	0	0	1.045739	-3.042188	220	5	1.1	0.9;
	3364	2	0	0	0	0	0	1.0538	-12.359552	220	5	1.1	0.9;
	3377	1	77.49	21.4	0	0.05	0	1.018192	-10.51029	220	5	1.1	0.9;
	3390	2	0	0	0	0	0	1.039984	-11.495624	220	5	1.1	0.9;
	3391	1	302.5	65.1	0	8.87	0	1.059691	-20.937372	220	5	1.1	0.9;
	3400	1	0	0	0	17.31	0	1.047571	-25.085016	220	5	1.1	0.9;
	3401	1	23.22	6.3	0	0.66	0	1.034711	-14.123031	220	5	1.1	0.9;
	3412	1	-0	0	0	0	0	1.057299	2.403578	220	5	1.1	0.9;
	3422	2	0	0	0	0	0	1.063657	-10.680951	220	5	1.1	0.9;
	3430	1	125.1	44	0	6.04	0	1.015255	-16.040907	220	5	1.1	0.9;
	3435	1	-0	0	0	10.13	0	1.043606	-25.55539	220	5	1.1	0.9;
	3436	2	0	0	0	0	0	1.060777	-13.286433	220	5	1.1	0.9;
	3445	1	45.14	-6.64	0	88.82001	0	0.999706	-5.879329	380	5	1.1	0.9;
	3450	1	78.99	16.2	0	1.97	0	1.046643	-21.48337	220	5	1.1	0.9;
	3481	1	142.5	23.6	0	4.24	0	1.023602	-20.789174	220	5	1.1	0.9;
	3483	1	157.78	276.03	0	91.192543	0	1.051877	2.264367	380	5	1.1	0.9;
	3485	1	-0	0	0	6.501668	0	0.991148	-16.268999	380	5	1.1	0.9;
	3486	1	101.5	-18.3	0	8.47	0	1.037237	-20.871205	220	5	1.1	0.9;
	3488	1	55.66	4.5	0	1.33	0	1.038891	-12.374272	220	5	1.1	0.9;
	3492	2	0	0	0	0	0	1.04852	-8.570568	220	5	1.1	0.9;
	3496	1	99.3	23.3	0	1.88	0	1.03602	-21.337931	220	5	1.1	0.9;
	3498	1	138.5	19.3	0	5.26	0	1.04022	-10.181185	220	5	1.1	0.9;
	3499	1	180.4	71.6	0	81.042938	0	1.035853	-3.63424	380	5	1.1	0.9;
	3502	1	98.2	19.1	0	10.92	0	1.020854	-10.552336	220	5	1.1	0.9;
	3503	1	-0	0	0	1.93	0	1.032128	-11.505891	220	5	1.1	0.9;
	3513	2	0	0	0	0	0	1.028771	-3.782955	380	5	1.1	0.9;
	3519	1	-0	0	0	68.053762	0	1.033967	-22.784898	380	5	1.1	0.9;
	3520	1	-0	0	0	2.47	0	1.037612	-5.896123	220	5	1.1	0.9;
	3526	1	33.34	13.2	0	0.76	0	1.056871	-4.128916	220	5	1.1	0.9;
	3535	1	-0	0	0	0.49	0	1.035787	-24.451001	220	5	1.1	0.9;
	3541	1	0	-0	0	3.34	0	1.039193	-23.983552	220	5	1.1	0.9;
	3543	1	-0	-0	0	14.079026	0	0.991625	-17.553669	380	5	1.1	0.9;
	3545	1	20.12	7.6	0	0.89	0	1.049206	-8.009952	220	5	1.1	0.9;
	3557	1	87.2	31.8	0	5.96	0	1.048975	-30.444694	220	5	1.1	0.9;
	3558	1	-0	0	0	2.51	0	1.062152	2.457422	220	5	1.1	0.9;
	3565	2	0	0	0	0	0	1.081614	-26.430835	220	5	1.1	0.9;
	3577	1	-38.7	-12.99	0	13.47383	0	1.031951	-5.185856	380	5	1.1	0.9;
	3579	1	277.72	7.35	0	3.58	0	1.072032	-39.86741	220	5	1.1	0.9;
	3580	2	0	0	0	0	0	1.07602	-33.404876	220	5	1.1	0.9;
	3589	1	37.54	18.3	0	0.22	0	1.042938	-6.728326	220	5	1.1	0.9;
	3594	1	31.43	6.6	0	3.32	0	1.034276	-19.803995	220	5	1.1	0.9;
	3601	1	0	0	0	0.23	0	1.040409	-21.360272	220	5	1.1	0.9;
	3602	1	86.6	14.9	0	1.34	0	1.032076	-18.933744	220	5	1.1	0.9;
	3608	1	0	-0	0	58.146813	0	1.00943	-41.64982	380	5	1.1	0.9;
	3609	1	-28.13	-0.82	0	0.1	0	1.061834	-11.405533	220	5	1.1	0.9;
	3610	1	185.8	31.8	0	3.73	0	1.03662	-22.744804	220	5	1.1	0.9;
	3611	2	0	0	0	0	0	1.03905	-18.955066	220	5	1.1	0.9;
	3613	1	152.5	14.4	0	24.61	0	1.03039	-28.045674	380	5	1.1	0.9;
	3643	1	24.23	6.4	0	3.9	0	1.033238	-19.364356	220	5	1.1	0.9;
	3645	1	42.55	17.6	0	17.28	0	1.057259	-33.103595	220	5	1.1	0.9;
	3649	1	-0	-0	0	4.86	0	1.049277	-36.568687	220	5	1.1	0.9;
	3654	1	20.72	5	0	6.64	0	1.074155	-26.393413	220	5	1.1	0.9;
	3656	2	0	0	0	0	0	1.070983	-11.746409	220	5	1.1	0.9;
	3657	1	131	1.3	0	120.45	0	1.075593	-26.08278	220	5	1.1	0.9;
	3661	2	0	0	0	0	0	1.059676	4.513427	380	5	1.1	0.9;
	3670	1	28.03	6.5	0	6.32	0	1.036121	-21.979516	220	5	1.1	0.9;
	3672	1	82.99	25.9	0	0.23	0	1.032626	-12.873557	220	5	1.1	0.9;
	3674	1	155.05	39.56	0	6.27	0	1.042158	-14.835571	220	5	1.1	0.9;
	3680	1	-0	-0	0	12.4	0	1.035367	-15.507057	220	5	1.1	0.9;
	3697	1	-32.71	-13.76	0	135.136611	0	1.04888	-20.581455	380	5	1.1	0.9;
	3698	2	0	0	0	0	0	1.041813	-24.312887	220	5	1.1	0.9;
	3701	1	-0	0	0	0.43	0	1.048029	-25.022699	220	5	1.1	0.9;
	3707	1	77.49	17.3	0	0.1	0	1.050811	-15.306372	220	5	1.1	0.9;
	3718	1	179.4	58.3	0	0.64	0	1.04576	-25.247909	220	5	1.1	0.9;
	3737	1	42.48	10.36	0	0.37	0	1.036909	-2.934488	220	5	1.1	0.9;
	3740	1	0	-0	0	13.03	0	1.03798	-21.218929	220	5	1.1	0.9;
	3741	2	0	0	0	0	0	1.031629	0.717314	380	5	1.1	0.9;
	3749	1	115.1	16.4	0	7.67	0	1.041217	-17.23173	220	5	1.1	0.9;
	3758	1	83.49	17.4	0	19.32	0	1.023928	-19.012865	220	5	1.1	0.9;
	3760	1	163.5	44.9	0	1.17	0	1.058074	-20.932224	220	5	1.1	0.9;
	3768	1	19.82	10.9	0	0.05	0	1.042031	-21.905161	220	5	1.1	0.9;
	3775	1	55.54	15.06	0	6.28	0	1.001347	-23.504647	220	5	1.1	0.9;
	3786	1	-0	-0	0	26.98	0	1.052535	-19.650048	220	5	1.1	0.9;
	3793	1	197.7	67.65	0	13.58	0	1.034254	-20.635259	220	5	1.1	0.9;
	3794	1	165.3	7.4	0	28.32	0	1.030869	-28.137864	380	5	1.1	0.9;
	3809	2	0	0	0	0	0	1.084785	-33.980103	220	5	1.1	0.9;
	3814	1	130.5	24.7	0	0.4	0	1.041172	-21.968498	220	5	1.1	0.9;
	3817	1	0	-0	0	1.02	0	1.089268	-15.202915	220	5	1.1	0.9;
	3818	1	44.45	10.9	0	2.23	0	1.042892	-25.785119	220	5	1.1	0.9;
	3825	2	0	0	0	0	0	1.026994	-8.529973	380	5	1.1	0.9;
	3830	1	-0	0	0	97.326028	0	0.99448	-13.765558	380	5	1.1	0.9;
	3834	1	160	69.2	0	0.19	0	1.036267	-22.248651	220	5	1.1	0.9;
	3855	1	-38.73	-15.32	0	97.23	0	1.07551	-26.170776	220	5	1.1	0.9;
	3857	1	-0	-0	0	11.68	0	1.081278	-23.127573	220	5	1.1	0.9;
	3865	1	-0	-0	0	2.26	0	1.027017	-14.866102	220	5	1.1	0.9;
	3866	1	201.06	19.46	0	2.87	0	1.048033	-27.549636	220	5	1.1	0.9;
	3869	2	0	0	0	0	0	1.031907	1.193974	380	5	1.1	0.9;
	3876	2	0	0	0	0	0	1.044501	-9.587109	220	5	1.1	0.9;
	3880	1	-28.13	-0.82	0	0.09	0	1.061821	-11.404764	220	5	1.1	0.9;
	3894	1	0	0	0	4.77	0	1.02402	-15.319245	220	5	1.1	0.9;
	3903	1	0	0	0	4.36	0	1.051358	-24.515174	220	5	1.1	0.9;
	3906	1	0	0	0	98.2	0	1.037271	-13.381903	380	5	1.1	0.9;
	3912	1	0	-0	0	19.73	0	1.038244	-21.078792	220	5	1.1	0.9;
	3916	2	0	0	0	0	0	1.043758	-14.004059	220	5	1.1	0.9;
	3918	1	-11.8	-5.47	0	88.050735	0	1.02393	-1.867594	380	5	1.1	0.9;
	3919	1	-0	0	0	39.95	0	1.024896	-6.075842	380	5	1.1	0.9;
	3925	1	89.3	25.1	0	2.98	0	1.026482	-13.116396	220	5	1.1	0.9;
	3928	1	105.2	24.2	0	0.25	0	1.046535	-4.757211	220	5	1.1	0.9;
	3929	1	48.95	11.6	0	2	0	1.03267	-20.217595	220	5	1.1	0.9;
	3951	2	0	0	0	0	0	1.027439	-12.908691	220	5	1.1	0.9;
	3956	1	-0	-0	0	1.18	0	1.049338	-24.08911	220	5	1.1	0.9;
	3962	1	19.42	11.8	0	2.6	0	1.041036	-23.700249	220	5	1.1	0.9;
	3969	1	-0	-0	0	0.84	0	1.018397	-10.494183	220	5	1.1	0.9;
	3971	2	0	0	0	0	0	1.063145	-8.854993	220	5	1.1	0.9;
	3975	1	148	26.1	0	9.44	0	0.986357	-24.771259	220	5	1.1	0.9;
	3985	1	77.19	26.5	0	6.2	0	1.047148	-0.061532	220	5	1.1	0.9;
	3994	1	36.34	10.4	0	2.67	0	1.035393	-2.725702	220	5	1.1	0.9;
	3997	1	26.13	9.7	0	2.37	0	1.042584	-11.956467	220	5	1.1	0.9;
	3999	1	110.8	36.9	0	23.2	0	1.051666	-6.639084	220	5	1.1	0.9;
	4000	1	-0	0	0	24.103274	0	0.991622	-15.520854	380	5	1.1	0.9;
	4005	1	171.18	28.49	0	0	0	1.067869	-31.620043	220	5	1.1	0.9;
	4024	2	0	0	0	0	0	1.040328	2.440561	380	5	1.1	0.9;
	4025	1	0	0	0	2.67	0	1.035538	-2.7146	220	5	1.1	0.9;
	4031	1	0	-0	0	6.42	0	1.063769	-10.640029	220	5	1.1	0.9;
	4032	1	265.7	113.5	0	1.85	0	0.999578	-11.699686	220	5	1.1	0.9;
	4039	1	-0	-0	0	1.18	0	1.037985	-22.526852	220	5	1.1	0.9;
	4049	1	-0	-0	0	0.89	0	1.017178	-10.133503	220	5	1.1	0.9;
	4054	1	0	-0	0	2.71	0	1.035023	-18.483102	220	5	1.1	0.9;
	4056	2	0	0	0	0	0	1.038526	-9.78553	220	5	1.1	0.9;
	4060	1	-0	0	0	2.08	0	1.066903	-14.674014	220	5	1.1	0.9;
	4084	2	0	0	0	0	0	1.063234	-8.796721	220	5	1.1	0.9;
	4100	1	-0	-0	0	1.23	0	1.063607	-10.700214	220	5	1.1	0.9;
	4103	1	81.09	12.5	0	2.4	0	1.037109	-21.700054	220	5	1.1	0.9;
	4110	1	13.01	3.8	0	3.59	0	1.048623	0.630118	220	5	1.1	0.9;
	4114	1	9.91	5.4	0	4.6	0	1.048622	-24.482065	220	5	1.1	0.9;
	4118	2	0	0	0	0	0	1.027195	-14.802318	220	5	1.1	0.9;
	4125	2	0	0	0	0	0	1.003502	-8.417156	380	5	1.1	0.9;
	4127	1	39.14	7.7	0	11.68	0	1.033235	-20.156948	220	5	1.1	0.9;
	4128	2	0	0	0	0	0	1.019761	-15.71382	220	5	1.1	0.9;
	4134	1	-0	0	0	0.09	0	1.074144	-33.622001	220	5	1.1	0.9;
	4141	1	-0.62	0	0	75.020288	0	1.032772	-21.096452	380	5	1.1	0.9;
	4143	1	11.11	4.1	0	0.8	0	1.05009	-1.152786	220	5	1.1	0.9;
	4144	1	310.2	4.5	0	8.51	0	1.048354	-19.735876	220	5	1.1	0.9;
	4157	1	0	0	0	1.2	0	1.074571	-26.335983	220	5	1.1	0.9;
	4185	1	87.5	21.3	0	7.52	0	1.042708	-25.806182	220	5	1.1	0.9;
	4186	1	-0	0	0	5.06	0	1.036176	-23.132591	220	5	1.1	0.9;
	4189	1	200.65	30.72	0	9.72	0	1.037191	-26.003681	220	5	1.1	0.9;
	4196	1	-3.2	-1.05	0	48.471563	0	1.029571	-7.552855	380	5	1.1	0.9;
	4197	1	9.01	4.9	0	0.64	0	1.06874	0.692131	220	5	1.1	0.9;
	4205	1	5.81	1.7	0	0.42	0	1.052019	-12.684377	220	5	1.1	0.9;
	4215	1	-0	-0	0	17.29	0	1.046292	-23.032269	220	5	1.1	0.9;
	4231	3	0	0	0	0	0	1.049182	0	380	5	1.1	0.9;
	4234	1	0	0	0	8.95	0	1.046192	-17.159976	220	5	1.1	0.9;
	4235	1	-0	-0	0	12.68	0	1.053821	-19.119448	220	5	1.1	0.9;
	4239	1	190.7	52	0	20.57	0	1.052709	-13.186066	220	5	1.1	0.9;
	4245	1	24.93	9.6	0	0.09	0	1.041034	-9.897904	220	5	1.1	0.9;
	4251	1	47.15	15.7	0	5.41	0	1.026731	-17.726744	220	5	1.1	0.9;
	4267	1	70.78	19.6	0	7.9	0	1.038616	-8.918342	220	5	1.1	0.9;
	4281	1	-0	-0	0	11.4	0	1.05107	-11.692848	220	5	1.1	0.9;
	4300	1	81.49	25.1	0	0.36	0	1.00381	-16.189999	220	5	1.1	0.9;
	4301	1	-0	-0	0	2.27	0	1.038253	-21.079157	220	5	1.1	0.9;
	4306	1	0	-0	0	0.3	0	1.034875	-30.808256	380	5	1.1	0.9;
	4313	1	133.4	29.4	0	10.87	0	1.043554	-21.132272	220	5	1.1	0.9;
	4314	1	26.13	2.7	0	1.7	0	1.044059	-14.213235	220	5	1.1	0.9;
	4324	1	307.42	-38.3	0	0	0	1.06402	-29.724601	220	5	1.1	0.9;
	4331	2	0	0	0	0	0	1.028646	-11.048944	220	5	1.1	0.9;
	4339	1	-0	-0	0	42.345005	0	0.994972	-15.805227	380	5	1.1	0.9;
	4353	1	165.8	44	0	6.15	0	1.055169	-26.615369	220	5	1.1	0.9;
	4355	1	65.07	15	0	1.37	0	1.07349	-26.435642	220	5	1.1	0.9;
	4357	1	235	58.9	0	14.47	0	1.046909	-15.944189	220	5	1.1	0.9;
	4368	1	90.6	15.8	0	7.99	0	1.060945	-25.082183	220	5	1.1	0.9;
	4395	2	0	0	0	0	0	1.039834	-20.6569	220	5	1.1	0.9;
	4402	1	-0	0	0	1.02	0	1.089271	-15.202973	220	5	1.1	0.9;
	4410	1	45.05	13.2	0	44.93	0	1.046754	-36.877378	220	5	1.1	0.9;
	4418	1	-0	0	0	0.05	0	1.076968	-24.357646	220	5	1.1	0.9;
	4419	2	0	0	0	0	0	1.059565	4.496453	380	5	1.1	0.9;
	4426	1	485.5	-63.38	0	34.92	0	1.068079	-46.755413	220	5	1.1	0.9;
	4432	1	-0	0	0	0.8	0	1.07457	-26.335948	220	5	1.1	0.9;
	4435	1	-0	-0	0	102.217726	0	1.006778	-8.596561	380	5	1.1	0.9;
	4454	1	-0	-0	0	1	0	0.987785	-18.481884	220	5	1.1	0.9;
	4480	2	0	0	0	0	0	1.071852	7.973164	220	5	1.1	0.9;
	4482	2	0	0	0	0	0	1.049914	-6.455681	220	5	1.1	0.9;
	4484	1	69.78	25	0	14.24	0	1.051896	-13.325368	220	5	1.1	0.9;
	4491	1	0	-0	0	1.75	0	1.032064	-28.901791	220	5	1.1	0.9;
	4494	1	-0	0	0	55.514618	0	1.024773	-0.869931	380	5	1.1	0.9;
	4504	1	170.1	-9.3	0	5.97	0	1.072324	-42.518571	220	5	1.1	0.9;
	4505	1	95.51	17.1	0	8.5	0	1.047176	-15.911012	220	5	1.1	0.9;
	4506	2	0	0	0	0	0	1.036047	-5.941934	220	5	1.1	0.9;
	4511	1	81.89	19.9	0	5.1	0	1.041941	-25.929686	220	5	1.1	0.9;
	4513	1	92.1	29.5	0	3.14	0	1.030552	-9.829625	220	5	1.1	0.9;
	4520	1	-0	-0	0	0.94	0	1.017179	-10.133508	220	5	1.1	0.9;
	4525	1	-0	0	0	6.12	0	1.060611	-8.826883	220	5	1.1	0.9;
	4529	1	48.48	8.7	0	10.14	0	1.055612	-15.125162	220	5	1.1	0.9;
	4541	1	0	-0	0	150.797812	0	1.023455	-9.910829	380	5	1.1	0.9;
	4544	1	-0	-0	0	14.64	0	1.047769	-14.697837	220	5	1.1	0.9;
	4549	1	0	-0	0	0.3	0	1.03214	-11.506046	220	5	1.1	0.9;
	4550	1	0	-0	0	124.2	0	1.019271	-10.38876	380	5	1.1	0.9;
	4554	1	119	21.9	0	0.95	0	1.041048	-9.799798	220	5	1.1	0.9;
	4562	1	130.5	25.7	0	2.27	0	1.02147	-16.819485	220	5	1.1	0.9;
	4566	2	0	0	0	0	0	1.073282	-27.676255	220	5	1.1	0.9;
	4580	1	11.01	3.9	0	7.98	0	1.045713	-12.00096	220	5	1.1	0.9;
	4594	1	-1.73	-1.71	0	22.062314	0	0.998779	-5.591046	380	5	1.1	0.9;
	4598	1	-0	0	0	52.958443	0	1.024757	-7.997137	380	5	1.1	0.9;
	4615	1	42.55	17.6	0	1.2	0	1.066957	-25.393864	220	5	1.1	0.9;
	4623	1	0	-0	0	80.388213	0	0.9943	-16.866228	380	5	1.1	0.9;
	4624	2	0	0	0	0	0	1.042135	-15.685962	220	5	1.1	0.9;
	4656	1	93.9	-0.4	0	3.66	0	1.052781	-19.605607	220	5	1.1	0.9;
	4661	2	0	0	0	0	0	1.049813	-26.16746	220	5	1.1	0.9;
	4674	1	124	29.2	0	0.53	0	1.032396	-12.4247	220	5	1.1	0.9;
	4679	1	102.7	17.8	0	48.05	0	1.042344	-20.616305	220	5	1.1	0.9;
	4683	1	38.84	8.5	0	7.31	0	1.052059	-3.729961	220	5	1.1	0.9;
	4685	1	46.15	11.2	0	10.16	0	1.042999	-25.790941	220	5	1.1	0.9;
	4689	1	109.7	15	0	30.97	0	1.048918	-30.792772	220	5	1.1	0.9;
	4701	2	0	0	0	0	0	1.043063	-9.879302	380	5	1.1	0.9;
	4710	1	195.4	76.84	0	8.59	0	1.026719	-9.176978	220	5	1.1	0.9;
	4711	1	-0	0	0	16.11	0	1.042937	-11.415152	220	5	1.1	0.9;
	4725	1	4.91	1.6	0	4.15	0	1.014202	-10.601951	220	5	1.1	0.9;
	4728	1	3.1	0.1	0	1.34	0	1.040401	-21.360189	220	5	1.1	0.9;
	4729	1	0	0	0	2.38	0	1.032006	-8.433928	220	5	1.1	0.9;
	4734	1	0	0	0	25.339116	0	1.014516	-10.190708	380	5	1.1	0.9;
	4738	1	101.89	25.72	0	5.11	0	1.06866	-10.715961	220	5	1.1	0.9;
	4747	1	93.3	27.8	0	6.1	0	1.035644	-12.325346	220	5	1.1	0.9;
	4748	1	-0	-0	0	1.49	0	1.032136	-12.426948	220	5	1.1	0.9;
	4765	1	33.04	10.1	0	0.43	0	1.03734	-12.246628	220	5	1.1	0.9;
	4783	2	0	0	0	0	0	1.071981	8.277038	220	5	1.1	0.9;
	4787	1	71.98	27.1	0	7.2	0	0.99123	-14.709219	380	5	1.1	0.9;
	4816	2	0	0	0	0	0	1.067305	-23.812093	220	5	1.1	0.9;
	4819	2	0	0	0	0	0	1.031907	1.213157	380	5	1.1	0.9;
	4823	2	0	0	0	0	0	1.067114	-14.648331	220	5	1.1	0.9;
	4826	1	103.6	16	0	11.22	0	1.076968	-24.357644	220	5	1.1	0.9;
	4829	1	171.9	39.1	0	4.23	0	1.020877	-5.66958	220	5	1.1	0.9;
	4831	1	130.4	44.33	0	5.22	0	1.023656	-15.867691	220	5	1.1	0.9;
	4850	2	0	0	0	0	0	1.054975	-13.889155	220	5	1.1	0.9;
	4852	1	45.55	-59.9	0	30.04	0	1.040672	-20.328499	220	5	1.1	0.9;
	4864	1	-28.21	-6	0	0.05	0	1.037977	-12.814047	220	5	1.1	0.9;
	4867	1	65.77	27.3	0	0.49	0	1.031537	-3.407292	220	5	1.1	0.9;
	4874	1	-0.58	-53.03	0	6.85	0	1.07611	-39.379591	220	5	1.1	0.9;
	4880	2	0	0	0	0	0	1.034991	-15.897969	220	5	1.1	0.9;
	4885	1	166.4	50.4	0	1.23	0	1.074189	-26.349591	220	5	1.1	0.9;
	4889	1	74.98	19.8	0	6.47	0	1.034333	-12.232863	220	5	1.1	0.9;
	4907	1	125.16	29.2	0	2.69	0	1.041821	-13.302552	220	5	1.1	0.9;
	4908	1	50.06	9.9	0	11.56	0	1.037684	-21.273347	220	5	1.1	0.9;
	4914	1	14.91	-8.73	0	34.66	0	1.036628	-22.980806	220	5	1.1	0.9;
	4918	2	0	0	0	0	0	1.057325	-11.046567	220	5	1.1	0.9;
	4925	1	-0	0	0	0.3	0	1.036525	-22.48156	220	5	1.1	0.9;
	4936	1	312.2	28	0	48.6	0	1.049019	-30.699604	220	5	1.1	0.9;
	4939	1	56.76	31.1	0	0.37	0	1.043989	-4.549265	220	5	1.1	0.9;
	4942	1	109.9	33.6	0	3.72	0	1.028311	-12.898354	220	5	1.1	0.9;
	4950	1	218.7	-9.1	0	49	0	1.036717	-40.104666	220	5	1.1	0.9;
	4951	1	183.95	-6.47	0	5.74	0	1.053531	-1.663312	220	5	1.1	0.9;
	4952	2	0	0	0	0	0	1.045386	-17.502534	220	5	1.1	0.9;
	4970	1	-0	-0	0	171.931977	0	1.064315	-17.672609	380	5	1.1	0.9;
	4974	1	89.08	26.9	0	4.68	0	1.023886	-13.4778	220	5	1.1	0.9;
	5002	1	-0	0	0	4.34	0	1.071581	-8.674239	220	5	1.1	0.9;
	5003	1	17.87	-8.59	0	0.21	0	1.042135	-9.676575	220	5	1.1	0.9;
	5004	2	0	0	0	0	0	1.0663	-15.374553	220	5	1.1	0.9;
	5007	1	0	0	0	40.291745	0	1.021829	-15.303707	380	5	1.1	0.9;
	5016	1	-0	0	0	8.347383	0	1.02139	-5.019103	380	5	1.1	0.9;
	5019	2	0	0	0	0	0	1.047693	-12.9179	220	5	1.1	0.9;
	5049	1	136.6	29	0	2.05	0	1.01997	-20.429759	220	5	1.1	0.9;
	5051	2	0	0	0	0	0	1.037373	-16.452704	220	5	1.1	0.9;
	5067	2	0	0	0	0	0	1.021924	-7.972254	380	5	1.1	0.9;
	5077	1	30.93	5.8	0	4.45	0	1.055018	-26.635396	220	5	1.1	0.9;
	5083	1	80.39	10.8	0	0.35	0	1.039314	-12.19306	220	5	1.1	0.9;
	5093	1	70.78	28	0	1.49	0	1.010775	-14.861295	220	5	1.1	0.9;
	5099	1	0	0	0	9.34	0	1.058325	1.539585	220	5	1.1	0.9;
	5106	1	39.94	13	0	2.47	0	1.071273	-31.590024	220	5	1.1	0.9;
	5110	2	0	0	0	0	0	1.032061	-10.28337	220	5	1.1	0.9;
	5120	2	0	0	0	0	0	1.055539	-23.597086	220	5	1.1	0.9;
	5131	1	-0	0	0	7.84	0	1.074641	-26.336733	220	5	1.1	0.9;
	5137	1	16.92	0	0	0.57	0	1.025017	-16.369795	220	5	1.1	0.9;
	5144	2	0	0	0	0	0	1.035201	-3.127844	380	5	1.1	0.9;
	5146	1	229.9	48.9	0	8.36	0	1.01309	-34.0847	220	5	1.1	0.9;
	5174	1	0	-0	0	8.19	0	1.045055	-21.030104	220	5	1.1	0.9;
	5179	1	179.8	5.67	0	12.83	0	1.049521	-30.437104	220	5	1.1	0.9;
	5182	1	0	0	0	0.18	0	1.04247	-7.991956	220	5	1.1	0.9;
	5212	1	-0	0	0	0.2	0	1.043532	-11.000335	220	5	1.1	0.9;
	5213	1	246.8	36.3	0	16.01	0	1.046557	-26.37211	220	5	1.1	0.9;
	5215	1	-0	-0	0	0.89	0	1.029858	-5.840941	220	5	1.1	0.9;
	5233	1	-0	-0	0	14.04	0	1.040607	-10.965139	220	5	1.1	0.9;
	5237	2	0	0	0	0	0	1.052719	-14.75067	220	5	1.1	0.9;
	5241	1	0	-0	0	70.905048	0	1.030298	-3.61649	380	5	1.1	0.9;
	5256	1	91	36	0	5.47	0	1.00434	-16.654648	220	5	1.1	0.9;
	5257	1	-0	0	0	0.18	0	1.04776	-6.282341	220	5	1.1	0.9;
	5270	1	-0	-0	0	7.4	0	1.040679	-21.118583	220	5	1.1	0.9;
	5278	2	0	0	0	0	0	1.042037	-24.259299	220	5	1.1	0.9;
	5286	1	17.22	5.1	0	0.99	0	1.037376	-22.164674	220	5	1.1	0.9;
	5288	1	-0	-0	0	3.06	0	0.998946	-15.122249	380	5	1.1	0.9;
	5297	1	55.26	11	0	7.68	0	1.041403	-24.362719	220	5	1.1	0.9;
	5300	1	-0	-0	0	1.3	0	1.045246	-10.990672	220	5	1.1	0.9;
	5308	1	0	-0	0	42.19206	0	0.984486	-18.190432	380	5	1.1	0.9;
	5317	1	293.2	87.4	0	26.36	0	1.028707	-44.346536	220	5	1.1	0.9;
	5334	1	17.52	6.6	0	3.79	0	1.033829	-12.746707	220	5	1.1	0.9;
	5340	2	0	0	0	0	0	1.054941	-11.154638	220	5	1.1	0.9;
	5341	1	64.27	17.5	0	0.43	0	1.011142	-21.65338	220	5	1.1	0.9;
	5350	1	0	0	0	28.208233	0	0.981529	-24.398889	380	5	1.1	0.9;
	5351	1	0	0	0	3.28	0	1.034562	-14.13742	220	5	1.1	0.9;
	5354	1	70.18	19.5	0	3.59	0	1.045497	-8.643995	220	5	1.1	0.9;
	5362	1	152.73	-17.2	0	45.85	0	1.043688	-24.235922	220	5	1.1	0.9;
	5365	2	0	0	0	0	0	1.061754	-11.4194	220	5	1.1	0.9;
	5379	2	0	0	0	0	0	1.107976	-6.815844	380	5	1.3	0.7;
	5383	1	0	-0	0	24.7	0	1.048549	-24.463345	220	5	1.1	0.9;
	5388	1	159.5	42	0	2.3	0	1.013391	-14.763695	220	5	1.1	0.9;
	5393	1	-0	0	0	9.43	0	1.065867	-15.42107	220	5	1.1	0.9;
	5395	2	0	0	0	0	0	1.048137	-27.33134	220	5	1.1	0.9;
	5400	1	241.62	52.78	0	0.82	0	1.025608	-14.012865	220	5	1.1	0.9;
	5410	1	69.78	26	0	5.18	0	1.045719	-21.375392	220	5	1.1	0.9;
	5413	1	-0	-0	0	5.57	0	1.036008	-18.03631	220	5	1.1	0.9;
	5417	1	32.74	5.3	0	3.07	0	1.057776	-12.050099	220	5	1.1	0.9;
	5418	1	-0	0	0	3.96	0	1.037717	-21.273778	220	5	1.1	0.9;
	5419	1	349.3	144.3	0	0.54	0	1.021201	-15.134475	220	5	1.1	0.9;
	5420	1	56.96	1.94	0	26.48	0	1.024419	-22.859247	380	5	1.1	0.9;
	5421	1	-0	-0	0	0.92	0	0.998945	-15.122174	380	5	1.1	0.9;
	5441	1	138.01	34.55	0	8.44	0	1.037781	-2.820227	220	5	1.1	0.9;
	5455	1	-0	-0	0	5.27	0	1.042465	-7.991918	220	5	1.1	0.9;
	5458	1	0	0	0	9.09	0	1.038913	-21.424694	220	5	1.1	0.9;
	5460	1	-15.89	-6.42	0	38.657071	0	0.992989	-17.9255	380	5	1.1	0.9;
	5461	2	0	0	0	0	0	1.06213	-18.41436	380	5	1.1	0.9;
	5469	1	106.6	-18.8	0	4.1	0	1.067374	-38.727555	220	5	1.1	0.9;
	5477	1	-0	-0	0	18.02	0	1.075392	-26.166931	220	5	1.1	0.9;
	5481	2	0	0	0	0	0	1.063711	-8.933875	380	5	1.1	0.9;
	5482	2	0	0	0	0	0	1.044486	-7.317519	220	5	1.1	0.9;
	5486	2	0	0	0	0	0	1.059636	4.495854	380	5	1.1	0.9;
	5488	2	0	0	0	0	0	1.041881	-29.8855	220	5	1.1	0.9;
	5490	2	0	0	0	0	0	1.068845	-15.202054	380	5	1.1	0.9;
	5502	1	-0	-0	0	0.58	0	1.038483	-20.512891	220	5	1.1	0.9;
	5519	1	0	-0	0	1.98	0	1.048798	-24.5238	220	5	1.1	0.9;
	5522	1	-0	-0	0	1.18	0	1.068862	0.701223	220	5	1.1	0.9;
	5525	1	236.33	59.35	0	18.71	0	1.03913	-14.017648	220	5	1.1	0.9;
	5529	1	0	0	0	0.36	0	1.037801	-2.820407	220	5	1.1	0.9;
	5533	2	0	0	0	0	0	1.038212	-13.663692	220	5	1.1	0.9;
	5546	2	0	0	0	0	0	1.05378	-23.67431	220	5	1.1	0.9;
	5550	1	0	-0	0	10.83	0	1.051152	-25.756092	220	5	1.1	0.9;
	5564	2	0	0	0	0	0	1.04123	-12.598948	220	5	1.1	0.9;
	5567	1	0	0	0	3.59	0	1.017813	-18.930548	220	5	1.1	0.9;
	5571	1	-0	0	0	1.01	0	1.036516	-22.481578	220	5	1.1	0.9;
	5573	1	0	-0	0	2.94	0	1.05903	-23.279891	220	5	1.1	0.9;
	5574	1	99.7	6	0	2.06	0	1.03922	-11.857544	220	5	1.1	0.9;
	5586	1	-0	0	0	0	0	1.057299	2.403578	220	5	1.1	0.9;
	5589	1	0	0	0	173.904854	0	1.034874	-30.808253	380	5	1.1	0.9;
	5610	1	99.5	9.2	0	27.55	0	1.045824	-24.167858	220	5	1.1	0.9;
	5616	1	225.2	35.2	0	11.99	0	1.076287	-35.225281	220	5	1.1	0.9;
	5627	1	-0	-0	0	0.53	0	1.030337	-32.491502	220	5	1.1	0.9;
	5630	1	0	0	0	11.52	0	1.043599	-4.085534	220	5	1.1	0.9;
	5641	1	-0	-0	0	5.98	0	1.049836	-26.047052	220	5	1.1	0.9;
	5648	1	-0	-0	0	115.528858	0	0.995233	-13.451687	380	5	1.1	0.9;
	5653	1	96.59	7.28	0	2.52	0	1.045504	-7.661043	220	5	1.1	0.9;
	5658	2	0	0	0	0	0	1.017689	-23.66504	380	5	1.1	0.9;
	5664	2	0	0	0	0	0	1.070507	0.956534	220	5	1.1	0.9;
	5666	1	124.8	34.7	0	3.66	0	1.046065	-10.74401	220	5	1.1	0.9;
	5686	1	52.26	19.6	0	2.64	0	1.030503	-10.223727	220	5	1.1	0.9;
	5688	1	14.12	5.2	0	1.44	0	1.033877	-8.079704	220	5	1.1	0.9;
	5691	1	195.1	-14.2	0	12.89	0	1.061392	-45.226837	220	5	1.1	0.9;
	5695	1	142.8	20.3	0	6.77	0	1.047646	-27.137982	220	5	1.1	0.9;
	5699	1	-0	0	0	0.59	0	1.049081	-6.522248	220	5	1.1	0.9;
	5709	2	0	0	0	0	0	1.027078	-8.525839	380	5	1.1	0.9;
	5712	1	183.07	44.65	0	2.77	0	1.0445	-4.578143	220	5	1.1	0.9;
	5720	1	337.6	-15.4	0	6.27	0	1.045667	-24.989207	220	5	1.1	0.9;
	5723	1	141.2	45.4	0	7.55	0	1.045129	-9.640756	220	5	1.1	0.9;
	5735	1	161.3	10.04	0	4.38	0	1.060348	-26.902429	220	5	1.1	0.9;
	5738	1	-0	-0	0	8.13	0	1.013795	-10.544291	220	5	1.1	0.9;
	5743	1	-0	0	0	4.28	0	1.052165	-24.123493	220	5	1.1	0.9;
	5753	1	-0	-0	0	1.81	0	1.033984	-3.213627	220	5	1.1	0.9;
	5764	1	99.4	4.3	0	3.69	0	1.034962	-25.335623	220	5	1.1	0.9;
	5781	2	0	0	0	0	0	1.0396	-11.754534	380	5	1.1	0.9;
	5789	1	0	0	0	0.35	0	1.005562	-16.021128	220	5	1.1	0.9;
	5799	1	-0	-0	0	1.38	0	1.043907	-32.265811	220	5	1.1	0.9;
	5803	1	0	-0	0	0.35	0	1.043774	-14.004171	220	5	1.1	0.9;
	5814	2	0	0	0	0	0	1.044051	-11.822108	220	5	1.1	0.9;
	5836	1	244.09	48.92	0	5.96	0	1.040266	-12.0279	220	5	1.1	0.9;
	5837	1	-0	-0	0	132.04744	0	1.019818	-10.148769	380	5	1.1	0.9;
	5853	1	122.5	38.9	0	0.84	0	1.042789	-7.464608	220	5	1.1	0.9;
	5856	2	0	0	0	0	0	1.038417	-11.615407	220	5	1.1	0.9;
	5857	1	78.39	21.2	0	0.97	0	1.057665	-21.071135	220	5	1.1	0.9;
	5881	2	0	0	0	0	0	1.065605	-10.395605	220	5	1.1	0.9;
	5891	1	14.02	5.5	0	3.3	0	1.036617	-12.108533	220	5	1.1	0.9;
	5907	1	186.1	14.8	0	17.28	0	1.046917	-8.47729	220	5	1.1	0.9;
	5918	1	73.83	16.41	0	15.77	0	1.053364	-25.787722	220	5	1.1	0.9;
	5926	1	0	-0	0	3.98	0	1.072067	-47.920607	220	5	1.1	0.9;
	5935	1	-0	0	0	5.75	0	1.065025	-22.144776	220	5	1.1	0.9;
	5940	2	0	0	0	0	0	1.030075	-14.582177	220	5	1.1	0.9;
	5944	1	71.88	21.7	0	9.63	0	1.059622	-25.872698	220	5	1.1	0.9;
	5957	1	208.06	36.6	0	3.55	0	1.025527	-16.209872	220	5	1.1	0.9;
	5971	2	0	0	0	0	0	1.070056	4.938459	380	5	1.3	0.7;
	5983	2	0	0	0	0	0	1.066029	-14.886912	380	5	1.1	0.9;
	5987	1	24.04	1.68	0	4.02	0	1.036825	-13.937479	220	5	1.1	0.9;
	5990	1	0	-0	0	4.53	0	1.023107	-10.218557	220	5	1.1	0.9;
	5993	1	0	-0	0	2.75	0	1.043528	-11.000279	220	5	1.1	0.9;
	5994	2	0	0	0	0	0	1.029184	-15.420668	220	5	1.1	0.9;
	6010	1	0	0	0	1.35	0	1.08135	-23.128077	220	5	1.1	0.9;
	6031	1	0	0	0	2.22	0	1.041572	-15.751421	220	5	1.1	0.9;
	6036	2	0	0	0	0	0	1.043738	-0.899698	380	5	1.1	0.9;
	6041	1	0	0	0	0.53	0	1.072109	-47.920883	220	5	1.1	0.9;
	6053	1	412.1	79.3	0	7.73	0	1.058134	-26.592239	220	5	1.1	0.9;
	6071	1	163.5	35.6	0	4.59	0	1.035726	-16.340255	220	5	1.1	0.9;
	6101	1	103	40.2	0	1.13	0	1.02679	-10.858615	220	5	1.1	0.9;
	6104	1	93.12	4.99	0	0.18	0	1.056405	-12.076426	220	5	1.1	0.9;
	6110	1	78.59	16	0	1.87	0	1.056871	-35.945281	220	5	1.1	0.9;
	6112	1	45.28	15.27	0	7.88	0	1.024189	-9.306431	220	5	1.1	0.9;
	6114	1	0	0	0	8.78	0	1.048769	-10.298802	220	5	1.1	0.9;
	6115	1	-0	-0	0	1.18	0	1.02848	-11.273668	220	5	1.1	0.9;
	6119	1	221.25	72.56	0	8.23	0	1.039102	-15.995867	220	5	1.1	0.9;
	6146	1	0	-0	0	3.15	0	1.028373	-12.079992	220	5	1.1	0.9;
	6151	1	60.67	33.2	0	0.31	0	1.032517	-24.013495	220	5	1.1	0.9;
	6153	2	0	0	0	0	0	0.988112	-14.90068	380	5	1.1	0.9;
	6163	1	0	-0	0	4.96	0	1.036223	-21.959198	220	5	1.1	0.9;
	6168	2	0	0	0	0	0	1.040343	-13.438292	220	5	1.1	0.9;
	6178	1	91	28.4	0	10.57	0	1.040244	-24.602196	220	5	1.1	0.9;
	6194	1	0	-0	0	3.14	0	1.026211	-16.329018	220	5	1.1	0.9;
	6199	1	-4.82	-0.88	0	224.85	0	1.072632	-12.599054	380	5	1.1	0.9;
	6203	1	14.02	5.6	0	2.34	0	1.070264	-27.353716	220	5	1.1	0.9;
	6206	1	-7.62	-2.52	0	33.197994	0	1.037591	0.03956	380	5	1.1	0.9;
	6219	1	-0	-0	0	0.22	0	1.033837	-12.746776	220	5	1.1	0.9;
	6220	1	231.9	27.6	0	0.36	0	1.055297	-22.931727	220	5	1.1	0.9;
	6224	1	0	0	0	58.990785	0	1.000545	-12.907246	380	5	1.1	0.9;
	6231	1	-0	0	0	11.5	0	1.034263	-8.032463	220	5	1.1	0.9;
	6232	1	49.25	12	0	7.28	0	1.042941	-25.794518	220	5	1.1	0.9;
	6240	1	0	-0	0	6.31	0	1.075696	-34.242457	220	5	1.1	0.9;
	6246	1	1769.94	-353.95	0	62.59	0	1.045739	-36.412972	380	5	1.1	0.9;
	6252	1	164.7	44.5	0	3.67	0	1.020382	-29.885564	220	5	1.1	0.9;
	6253	1	128.8	29.3	0	26.74	0	1.04318	-25.753603	220	5	1.1	0.9;
	6267	1	91.3	18.1	0	3.05	0	1.041192	-24.394509	220	5	1.1	0.9;
	6271	1	3.7	2.6	0	4.45	0	1.054472	-20.10439	220	5	1.1	0.9;
	6290	1	47.65	20.4	0	3.63	0	1.062557	-12.82926	220	5	1.1	0.9;
	6291	2	0	0	0	0	0	1.040598	-23.840293	220	5	1.1	0.9;
	6306	1	0	-0	0	2.13	0	1.077228	-24.297873	220	5	1.1	0.9;
	6308	1	0	-0	0	0.05	0	1.044666	-25.545467	220	5	1.1	0.9;
	6313	1	48.95	14.9	0	4.04	0	1.033893	-23.915745	220	5	1.1	0.9;
	6331	2	0	0	0	0	0	1.052181	-0.845895	220	5	1.1	0.9;
	6332	2	0	0	0	0	0	1.062214	-11.259336	220	5	1.1	0.9;
	6337	1	0	0	0	8.26	0	1.090736	-9.663194	380	5	1.1	0.9;
	6351	2	0	0	0	0	0	1.064446	-16.194964	220	5	1.1	0.9;
	6357	1	430.8	107.9	0	11.64	0	0.986449	-24.787083	220	5	1.1	0.9;
	6368	2	0	0	0	0	0	1.043835	-8.297319	220	5	1.1	0.9;
	6371	1	-0	-0	0	0.35	0	1.04624	-0.834636	220	5	1.1	0.9;
	6376	2	0	0	0	0	0	1.046714	-0.304027	220	5	1.1	0.9;
	6382	1	75.43	10.49	0	7.91	0	1.042008	-9.653397	220	5	1.1	0.9;
	6384	1	159.9	47.4	0	3.65	0	1.042953	-1.650086	220	5	1.1	0.9;
	6405	1	76.48	19.8	0	5.41	0	1.016603	-16.051975	220	5	1.1	0.9;
	6416	1	67.67	16.4	0	3.7	0	1.044399	-9.544276	220	5	1.1	0.9;
	6426	1	10.21	3.8	0	4.6	0	1.054991	1.525091	220	5	1.1	0.9;
	6427	1	83.19	22.9	0	28.5	0	1.00477	-13.123224	380	5	1.1	0.9;
	6429	2	0	0	0	0	0	1.039574	-18.343476	220	5	1.1	0.9;
	6430	1	0	-0	0	6.03	0	1.052192	-28.928753	220	5	1.1	0.9;
	6450	1	5.51	2.1	0	1.92	0	1.010198	-10.933937	220	5	1.1	0.9;
	6455	1	20.52	13.3	0	0.17	0	1.038267	-22.056487	220	5	1.1	0.9;
	6472	1	0	-0	0	9.93	0	1.042806	-24.004462	220	5	1.1	0.9;
	6474	2	0	0	0	0	0	1.06557	-7.808697	220	5	1.1	0.9;
	6475	1	-0	0	0	149.047157	0	1.026549	-8.701988	380	5	1.1	0.9;
	6478	1	141.2	34.8	0	2.37	0	1.012128	-14.390515	220	5	1.1	0.9;
	6486	1	31.74	17.4	0	0.89	0	1.048545	-3.989815	220	5	1.1	0.9;
	6495	1	49.66	12.3	0	2.2	0	1.043059	-21.768062	220	5	1.1	0.9;
	6510	1	88.2	26.7	0	8.91	0	1.050118	-26.338765	220	5	1.1	0.9;
	6516	2	0	0	0	0	0	1.052452	-3.60194	220	5	1.1	0.9;
	6521	1	0	0	0	2.3	0	1.046651	-4.738353	220	5	1.1	0.9;
	6532	1	120.69	18.1	0	15.46	0	1.075728	-31.435419	220	5	1.1	0.9;
	6552	2	0	0	0	0	0	1.048796	-21.140924	220	5	1.1	0.9;
	6555	1	-37.81	-21.64	0	83.362314	0	1.008884	-43.693382	380	5	1.1	0.9;
	6556	1	-0	-0	0	4.61	0	1.052054	-12.679594	220	5	1.1	0.9;
	6563	1	0	-0	0	8.94	0	1.040617	-24.64966	220	5	1.1	0.9;
	6565	1	0	-0	0	4.05	0	1.033616	-7.247992	220	5	1.1	0.9;
	6570	1	-0	0	0	32.43	0	1.045563	-20.766442	220	5	1.1	0.9;
	6581	1	153.8	23.4	0	87.73	0	1.042943	-2.447442	380	5	1.1	0.9;
	6612	1	41.14	15.25	0	3.96	0	1.047964	-31.701384	220	5	1.1	0.9;
	6616	1	-0	-0	0	3.06	0	1.044855	-15.904052	220	5	1.1	0.9;
	6619	1	-0	-0	0	10.67	0	1.050275	-15.385558	220	5	1.1	0.9;
	6624	1	-37.33	-11.88	0	145.18	0	1.058707	4.26077	380	5	1.1	0.9;
	6629	1	62.63	16.3	0	2.03	0	1.054419	-11.400101	220	5	1.1	0.9;
	6630	1	134.1	24.8	0	13.73	0	1.039511	-24.152951	220	5	1.1	0.9;
	6636	1	-0	0	0	0.31	0	1.036596	0.577218	220	5	1.1	0.9;
	6638	1	61.65	10.24	0	3.48	0	1.020424	-14.95076	220	5	1.1	0.9;
	6639	1	388.04	52.61	0	2.87	0	1.050408	-2.453064	220	5	1.1	0.9;
	6648	1	-26.03	-0.84	0	0.1	0	1.06183	-11.406171	220	5	1.1	0.9;
	6664	1	-0	0	0	10.68	0	1.06971	-46.294482	220	5	1.1	0.9;
	6675	1	37.94	7.5	0	11.89	0	1.035344	-23.290216	220	5	1.1	0.9;
	6684	1	-0	0	0	9.84	0	1.044665	-25.545459	220	5	1.1	0.9;
	6691	1	42.55	12.5	0	1.26	0	1.083988	-34.063221	220	5	1.1	0.9;
	6692	1	10.33	3.3	0	4.65	0	1.02441	-15.580999	220	5	1.1	0.9;
	6697	1	309	74.5	0	3.69	0	1.0463	-37.100468	220	5	1.1	0.9;
	6714	1	-0	-0	0	0.3	0	1.046238	-0.834605	220	5	1.1	0.9;
	6723	1	0	0	0	1.58	0	1.032514	-9.464819	220	5	1.1	0.9;
	6730	1	-0	0	0	2.18	0	1.038253	-21.079078	220	5	1.1	0.9;
	6734	2	0	0	0	0	0	1.061492	-3.472634	220	5	1.1	0.9;
	6735	1	-0	-0	0	0.4	0	1.041984	-1.292675	380	5	1.1	0.9;
	6737	1	0	-0	0	9.74	0	1.021139	-10.532887	220	5	1.1	0.9;
	6738	1	-4.91	-2.53	0	120.576812	0	1.029435	-24.781868	380	5	1.1	0.9;
	6742	1	137.4	23.9	0	7.26	0	1.037902	-21.502542	220	5	1.1	0.9;
	6744	1	244.12	95.29	0	7.55	0	1.033513	-10.221422	220	5	1.1	0.9;
	6757	1	0	0	0	6.57	0	1.040789	-1.178761	380	5	1.1	0.9;
	6763	1	0	-0	0	5.46	0	1.043751	-24.236704	220	5	1.1	0.9;
	6772	1	61.37	23.2	0	0.22	0	1.045385	-24.09701	220	5	1.1	0.9;
	6773	1	0	-0	0	2.18	0	1.041981	-24.273746	220	5	1.1	0.9;
	6785	1	0	0	0	8.9	0	1.054082	-0.784591	220	5	1.1	0.9;
	6791	1	46.35	9.1	0	18.2	0	1.04692	-25.227196	220	5	1.1	0.9;
	6802	1	-0	0	0	7.43	0	1.034691	-6.054841	220	5	1.1	0.9;
	6806	1	135.84	15.97	0	4.57	0	1.046591	-21.264934	220	5	1.1	0.9;
	6807	2	0	0	0	0	0	1.006037	-16.618507	220	5	1.1	0.9;
	6816	2	0	0	0	0	0	1.053803	-24.150678	220	5	1.1	0.9;
	6820	2	0	0	0	0	0	1.049557	-6.650937	220	5	1.1	0.9;
	6828	1	58.16	16.5	0	6.24	0	1.037156	-13.39716	220	5	1.1	0.9;
	6831	2	0	0	0	0	0	1.031347	-9.785828	220	5	1.1	0.9;
	6837	1	47.75	11	0	0.05	0	1.050311	-7.914424	220	5	1.1	0.9;
	6842	1	98	21.7	0	6	0	1.037762	-8.446373	220	5	1.1	0.9;
	6844	1	-0	-0	0	1.04	0	1.015966	-29.284947	220	5	1.1	0.9;
	6845	2	0	0	0	0	0	1.055216	-11.135397	220	5	1.1	0.9;
	6846	1	44.35	19.3	0	3.25	0	1.053078	-4.617698	220	5	1.1	0.9;
	6852	2	0	0	0	0	0	1.031296	-19.801967	220	5	1.1	0.9;
	6854	1	0	0	0	0	0	1.054902	-11.343079	220	5	1.1	0.9;
	6857	2	0	0	0	0	0	1.021411	-8.599284	380	5	1.1	0.9;
	6880	1	0	0	0	1.36	0	1.045265	-4.460667	220	5	1.1	0.9;
	6887	1	-0	-0	0	3.13	0	1.045229	-3.097099	220	5	1.1	0.9;
	6888	2	0	0	0	0	0	1.039234	-13.670722	220	5	1.1	0.9;
	6889	1	-12.49	-5.83	0	6.235224	0	1.040018	2.414551	380	5	1.1	0.9;
	6891	1	135.61	26.86	0	1.93	0	1.042154	-14.327292	220	5	1.1	0.9;
	6897	1	176.8	39.4	0	28.24	0	1.05642	-16.592196	380	5	1.1	0.9;
	6901	1	-37.84	-19.23	0	28.140258	0	1.010989	-32.14614	380	5	1.1	0.9;
	6908	1	0	0	0	3.01	0	1.037832	-40.113954	220	5	1.1	0.9;
	6909	1	122.5	8.7	0	15.85	0	1.036123	-23.128738	220	5	1.1	0.9;
	6921	1	-37.34	-13.33	0	251.529665	0	1.036075	-1.608828	380	5	1.1	0.9;
	6922	1	113.6	44.7	0	1.91	0	1.036504	-8.622098	220	5	1.1	0.9;
	6926	1	124.5	101.4	0	0.53	0	1.033188	-12.416059	220	5	1.1	0.9;
	6940	1	0	-0	0	10.46	0	1.057214	-27.586432	220	5	1.1	0.9;
	6947	2	0	0	0	0	0	1.051367	-3.749893	220	5	1.1	0.9;
	6952	1	180.9	33.2	0	29.17	0	1.074941	-25.158601	220	5	1.1	0.9;
	6954	1	18.12	10	0	1.03	0	1.033937	-20.057453	220	5	1.1	0.9;
	6961	1	56.56	18.4	0	3.29	0	1.046702	-0.041292	220	5	1.1	0.9;
	6969	2	0	0	0	0	0	1.059486	-23.158005	220	5	1.1	0.9;
	6982	2	0	0	0	0	0	1.036531	-20.798066	220	5	1.1	0.9;
	6989	1	0	0	0	1.91	0	1.046262	-24.006134	220	5	1.1	0.9;
	6990	1	-0	0	0	4.54	0	1.049424	-24.091607	220	5	1.1	0.9;
	7014	1	0	0	0	0.3	0	1.042775	-2.603788	220	5	1.1	0.9;
	7019	1	52.56	12.8	0	2.87	0	1.042599	-25.81851	220	5	1.1	0.9;
	7021	1	226.8	53.3	0	3.26	0	1.00593	-17.187001	220	5	1.1	0.9;
	7030	1	0	0	0	0.19	0	1.065574	-7.808728	220	5	1.1	0.9;
	7036	2	0	0	0	0	0	1.041241	-12.591424	220	5	1.1	0.9;
	7042	1	220.98	91	0	1.16	0	1.05498	-30.089269	220	5	1.1	0.9;
	7047	1	339.9	70.5	0	28.91	0	1.036489	-21.901119	220	5	1.1	0.9;
	7049	2	0	0	0	0	0	1.049669	-14.344403	220	5	1.1	0.9;
	7050	1	120	26	0	22.01	0	1.037777	-21.239311	220	5	1.1	0.9;
	7052	1	111	36.6	0	2.21	0	1.047794	-12.393218	220	5	1.1	0.9;
	7056	2	0	0	0	0	0	1.006597	-4.279873	380	5	1.1	0.9;
	7069	1	258.79	42.82	0	9.28	0	1.041721	-7.971411	220	5	1.1	0.9;
	7070	1	142.9	25.8	0	24.18	0	1.051529	-13.392684	220	5	1.1	0.9;
	7076	1	0	-0	0	0.36	0	1.032543	-12.400826	220	5	1.1	0.9;
	7092	1	44.65	8.8	0	20.17	0	1.048118	-25.098572	220	5	1.1	0.9;
	7098	1	0	-0	0	7.17	0	1.040402	-2.260554	220	5	1.1	0.9;
	7115	2	0	0	0	0	0	1.008722	-16.289324	220	5	1.1	0.9;
	7119	1	149.53	34.9	0	8.9	0	1.011547	-16.3375	220	5	1.1	0.9;
	7124	1	-0	0	0	16.73	0	1.071843	-27.185513	220	5	1.1	0.9;
	7129	1	33.34	5	0	0.38	0	1.03765	-2.840915	220	5	1.1	0.9;
	7132	1	222	90.5	0	0.16	0	1.017449	-15.909765	220	5	1.1	0.9;
	7133	1	102.3	19	0	1.84	0	1.034901	-14.267058	220	5	1.1	0.9;
	7144	1	8.48	-1.82	0	0.37	0	1.03015	-13.008541	220	5	1.1	0.9;
	7148	1	-0	-0	0	10.89	0	1.050811	-23.312468	220	5	1.1	0.9;
	7159	2	0	0	0	0	0	1.059575	4.500856	380	5	1.1	0.9;
	7162	1	-0	0	0	2.48	0	1.051693	-13.395189	220	5	1.1	0.9;
	7163	1	-0	0	0	2.9	0	1.04999	-24.866445	220	5	1.1	0.9;
	7164	1	-18.6	-6.72	0	152.683271	0	1.027379	-22.063848	380	5	1.1	0.9;
	7165	1	-0	-0	0	9.26	0	1.071048	-34.202221	220	5	1.1	0.9;
	7178	1	-0	-0	0	2.45	0	1.017469	-14.250696	220	5	1.1	0.9;
	7183	2	0	0	0	0	0	1.107733	-6.868685	380	5	1.3	0.7;
	7202	1	18.62	6.1	0	2	0	1.050711	-3.937783	220	5	1.1	0.9;
	7209	2	0	0	0	0	0	1.051113	-6.533675	220	5	1.1	0.9;
	7222	1	0	-0	0	5	0	1.068964	-23.827934	220	5	1.1	0.9;
	7226	1	144	31.3	0	3.18	0	1.022196	-15.530828	220	5	1.1	0.9;
	7231	1	0	0	0	0.05	0	1.039985	-11.495627	220	5	1.1	0.9;
	7253	1	7.51	3.1	0	0	0	1.048698	-2.446191	220	5	1.1	0.9;
	7256	1	392.25	92.02	0	9.45	0	1.022568	-35.394958	220	5	1.1	0.9;
	7259	1	82.39	0	0	0.15	0	1.06537	-7.876917	220	5	1.1	0.9;
	7264	1	0	-0	0	3.42	0	1.042846	-10.973594	220	5	1.1	0.9;
	7266	1	-0	-0	0	2.12	0	1.032954	-5.485507	220	5	1.1	0.9;
	7267	2	0	0	0	0	0	1.010817	-7.692119	380	5	1.1	0.9;
	7273	1	219.6	36.4	0	17.89	0	1.042886	-24.322652	220	5	1.1	0.9;
	7274	1	54.93	16.5	0	6.75	0	1.040156	-3.899721	220	5	1.1	0.9;
	7282	2	0	0	0	0	0	1.034538	-4.68119	380	5	1.1	0.9;
	7284	1	0	0	0	0	0	1.094644	-0.869931	220	5	1.1	0.9;
	7289	1	0	-0	0	26.5	0	1.054049	-6.174395	220	5	1.1	0.9;
	7309	1	0	-0	0	1.32	0	1.054902	-11.343079	220	5	1.1	0.9;
	7316	1	-0	0	0	0.09	0	1.044717	-7.284069	220	5	1.1	0.9;
	7325	1	0	0	0	0.41	0	1.052061	-12.679657	220	5	1.1	0.9;
	7327	2	0	0	0	0	0	1.043172	-18.073929	220	5	1.1	0.9;
	7328	2	0	0	0	0	0	0.999675	-14.476061	380	5	1.1	0.9;
	7338	1	0	-0	0	2.98	0	1.045891	-14.743514	220	5	1.1	0.9;
	7341	1	37.44	0	0	0.67	0	1.040136	-2.402429	220	5	1.1	0.9;
	7342	1	169	70.8	0	5.23	0	1.018398	-16.108009	220	5	1.1	0.9;
	7351	1	61.67	16.3	0	0.05	0	1.037551	-2.8443	220	5	1.1	0.9;
	7353	1	127.3	26.7	0	7.02	0	0.985936	-22.42059	220	5	1.1	0.9;
	7361	1	-0	-0	0	14.29	0	1.053815	-15.271705	220	5	1.1	0.9;
	7367	1	1.5	0.4	0	0.15	0	1.028631	-18.905659	220	5	1.1	0.9;
	7373	1	0	-0	0	1.81	0	1.071954	-23.60091	220	5	1.1	0.9;
	7377	1	0	0	0	0.63	0	1.014168	-16.130077	220	5	1.1	0.9;
	7380	1	105.67	0.56	0	18.05	0	0.992016	-14.57117	380	5	1.1	0.9;
	7396	1	-0	0	0	5.68	0	1.019741	-19.557023	220	5	1.1	0.9;
	7422	1	65.67	25.6	0	2.34	0	1.033674	-10.805957	220	5	1.1	0.9;
	7437	1	18.32	5.3	0	7.32	0	1.04383	-25.64578	220	5	1.1	0.9;
	7438	1	121.3	20.7	0	2.39	0	1.035924	-12.727575	220	5	1.1	0.9;
	7464	1	197.65	28.23	0	5.87	0	1.041056	-12.617278	220	5	1.1	0.9;
	7466	2	0	0	0	0	0	1.057257	-34.871885	220	5	1.1	0.9;
	7471	1	-18.95	-6.93	0	33.11439	0	1.020554	-6.279485	380	5	1.1	0.9;
	7473	1	113.44	71.42	0	11.08	0	1.035336	-16.025005	220	5	1.1	0.9;
	7474	2	0	0	0	0	0	1.027108	-14.200631	220	5	1.1	0.9;
	7485	1	44.35	10.8	0	8.72	0	1.042606	-25.826706	220	5	1.1	0.9;
	7491	1	0	-0	0	4.29	0	1.012242	-21.474958	220	5	1.1	0.9;
	7495	2	0	0	0	0	0	1.035256	-15.196734	220	5	1.1	0.9;
	7507	1	4	0.8	0	17.65	0	1.04914	-24.08765	220	5	1.1	0.9;
	7513	1	0	-0	0	38.549494	0	0.993009	-15.030612	380	5	1.1	0.9;
	7519	1	97.97	31.8	0	5.02	0	1.036786	-10.798224	220	5	1.1	0.9;
	7520	2	0	0	0	0	0	1.037447	-9.123264	380	5	1.1	0.9;
	7522	2	0	0	0	0	0	0.99994	-10.953994	380	5	1.1	0.9;
	7523	1	-0	0	0	2.11	0	1.052161	-0.72038	220	5	1.1	0.9;
	7530	1	0	0	0	31.242294	0	1.046869	-19.820201	380	5	1.1	0.9;
	7537	1	0	0	0	17.184121	0	0.989983	-15.6248	380	5	1.1	0.9;
	7539	1	40.85	15.7	0	2.08	0	1.053724	-12.373239	220	5	1.1	0.9;
	7541	1	346.42	89.04	0	43.55	0	1.054084	-16.465783	380	5	1.1	0.9;
	7571	1	-0	-0	0	13.9	0	1.044943	-13.985525	220	5	1.1	0.9;
	7576	1	0	0	0	0.5	0	1.04028	-12.028046	220	5	1.1	0.9;
	7577	1	0	0	0	6.69	0	1.0481	-10.467484	220	5	1.1	0.9;
	7579	1	136.9	11.4	0	3.05	0	1.048022	-26.50227	220	5	1.1	0.9;
	7580	1	0	0	0	0.17	0	1.05695	-11.9558	220	5	1.1	0.9;
	7582	1	0	-0	0	3.44	0	1.035758	-24.450751	220	5	1.1	0.9;
	7583	1	5.01	1.8	0	3.68	0	1.04744	-0.936009	220	5	1.1	0.9;
	7599	1	0	-0	0	2.13	0	1.034704	-6.055108	220	5	1.1	0.9;
	7624	1	32.74	10.5	0	1.78	0	1.01592	-20.903943	220	5	1.1	0.9;
	7626	1	78.89	45.6	0	0.04	0	1.032312	-12.417124	220	5	1.1	0.9;
	7635	1	0	0	0	1.86	0	1.050003	-7.891572	220	5	1.1	0.9;
	7640	1	68.78	35	0	34.52	0	1.033083	-6.005598	380	5	1.1	0.9;
	7641	2	0	0	0	0	0	1.051427	-6.243524	220	5	1.1	0.9;
	7647	1	-0	-0	0	0.1	0	1.063763	-20.174055	220	5	1.1	0.9;
	7653	1	33.94	11.4	0	0	0	1.0368	-22.767059	220	5	1.1	0.9;
	7663	1	37.14	2.7	0	2.33	0	1.042204	-11.956017	220	5	1.1	0.9;
	7691	1	0	-0	0	96.48	0	1.026843	-3.944271	380	5	1.1	0.9;
	7694	1	-14.96	-4.7	0	0.04	0	1.027078	-14.861465	220	5	1.1	0.9;
	7697	2	0	0	0	0	0	1.052424	-3.583012	220	5	1.1	0.9;
	7700	1	66.17	27.4	0	14.12	0	1.018432	-10.528155	220	5	1.1	0.9;
	7702	1	-0	-0	0	11.68	0	1.025153	-16.313747	220	5	1.1	0.9;
	7726	1	0	0	0	2.03	0	1.018323	-10.493485	220	5	1.1	0.9;
	7738	1	31.84	9.2	0	5.23	0	1.027755	-21.847912	220	5	1.1	0.9;
	7752	1	524.5	126	0	7.45	0	1.03043	-18.412915	220	5	1.1	0.9;
	7766	1	78.89	25.6	0	1.6	0	1.033279	-12.689648	220	5	1.1	0.9;
	7770	1	207.9	32.36	0	8.31	0	1.031073	-44.333742	220	5	1.1	0.9;
	7772	1	223.6	51.8	0	18.45	0	1.053846	-25.615931	220	5	1.1	0.9;
	7775	1	151.5	7.9	0	5	0	1.054984	-28.678026	220	5	1.1	0.9;
	7776	2	0	0	0	0	0	1.072261	6.181044	380	5	1.3	0.7;
	7778	1	153	21.5	0	28.85	0	1.057591	-16.510132	380	5	1.1	0.9;
	7791	1	-0	0	0	1.33	0	1.03015	-12.988362	220	5	1.1	0.9;
	7797	1	-0	-0	0	2.09	0	1.051084	-24.096828	220	5	1.1	0.9;
	7808	2	0	0	0	0	0	1.054562	-0.188042	380	5	1.1	0.9;
	7809	1	169.5	27.7	0	3.04	0	1.024952	-11.538551	220	5	1.1	0.9;
	7824	1	9.21	3.4	0	2.63	0	1.047756	-6.282298	220	5	1.1	0.9;
	7831	1	-17.17	-7.1	0	57.019803	0	1.011199	-11.417611	380	5	1.1	0.9;
	7840	1	-19.28	29.9	0	1.6	0	1.049049	-2.414341	220	5	1.1	0.9;
	7842	2	0	0	0	0	0	1.001912	-12.534976	380	5	1.1	0.9;
	7847	1	0	-0	0	1.55	0	1.040499	-8.878125	220	5	1.1	0.9;
	7857	1	0	0	0	2.36	0	1.050458	0.170436	220	5	1.1	0.9;
	7862	1	0	0	0	9.74	0	1.040593	-21.115869	220	5	1.1	0.9;
	7865	1	84.99	19.6	0	4.32	0	1.074298	-26.36902	220	5	1.1	0.9;
	7873	1	-20.94	-6.16	0	11.99	0	1.021235	-15.229234	220	5	1.1	0.9;
	7881	1	-0	-0	0	7.99	0	1.055617	-14.956341	220	5	1.1	0.9;
	7883	1	47.55	17.8	0	0.32	0	1.050929	-28.78941	220	5	1.1	0.9;
	7885	1	142.4	20	0	2.42	0	1.037839	-21.110646	220	5	1.1	0.9;
	7886	1	162.5	42.7	0	12.79	0	1.018149	-19.303853	220	5	1.1	0.9;
	7892	1	36.94	11	0	0.52	0	1.053815	-11.294264	220	5	1.1	0.9;
	7895	1	338.7	52.5	0	11.36	0	1.059845	-19.561088	220	5	1.1	0.9;
	7903	1	0	-0	0	7.78	0	1.072768	-41.998744	220	5	1.1	0.9;
	7905	1	347.1	87.2	0	88.35	0	1.046655	-17.893306	380	5	1.1	0.9;
	7913	2	0	0	0	0	0	1.063764	-8.906151	380	5	1.1	0.9;
	7923	1	0	-0	0	0.97	0	1.038632	-22.138858	220	5	1.1	0.9;
	7937	1	56.06	17	0	2.4	0	1.040168	-7.402075	220	5	1.1	0.9;
	7943	1	38.44	2	0	0.56	0	1.047179	-16.175311	220	5	1.1	0.9;
	7945	1	17.92	5.8	0	0.84	0	1.031087	-5.74962	220	5	1.1	0.9;
	7955	1	111.6	8.9	0	16.19	0	1.068312	-23.349161	220	5	1.1	0.9;
	7961	1	50.56	8.1	0	8.02	0	1.034289	-8.616758	220	5	1.1	0.9;
	7967	1	0	-0	0	5.97	0	1.043007	-11.416934	220	5	1.1	0.9;
	7972	1	237.9	68.5	0	5.12	0	1.035175	-19.298443	220	5	1.1	0.9;
	7974	1	234	13.1	0	51.03	0	1.033995	-20.055775	220	5	1.1	0.9;
	7982	1	112	19.7	0	38.04	0	1.036111	-23.13213	220	5	1.1	0.9;
	7988	1	-3.55	-1.43	0	11.834121	0	1.0339	-2.136803	380	5	1.1	0.9;
	7989	1	0	0	0	0.27	0	1.052062	-3.729978	220	5	1.1	0.9;
	7994	1	11.91	3.3	0	6.84	0	1.015764	-16.263273	220	5	1.1	0.9;
	7998	2	0	0	0	0	0	1.048854	4.223749	380	5	1.1	0.9;
	8005	1	-0	-0	0	16.77	0	1.027697	-13.719399	220	5	1.1	0.9;
	8030	1	-0	0	0	122.13	0	1.034376	-4.451962	380	5	1.1	0.9;
	8035	1	-0	0	0	4.22	0	1.023645	-16.359288	220	5	1.1	0.9;
	8043	2	0	0	0	0	0	1.066913	-11.239517	220	5	1.1	0.9;
	8057	1	69.18	9.5	0	3.67	0	1.045952	-8.643584	220	5	1.1	0.9;
	8060	1	0	-0	0	9.54	0	1.05168	-13.395033	220	5	1.1	0.9;
	8104	1	127.2	40.8	0	0.2	0	1.070946	-25.178857	220	5	1.1	0.9;
	8107	1	68.18	16.1	0	3.35	0	1.034732	-23.363006	220	5	1.1	0.9;
	8109	2	0	0	0	0	0	1.067387	-21.117921	380	5	1.1	0.9;
	8112	1	87.9	29.3	0	1.21	0	1.031544	-3.541639	220	5	1.1	0.9;
	8128	1	-0	0	0	8.1	0	1.090736	-9.663123	380	5	1.1	0.9;
	8151	1	0	0	0	14.23	0	1.044491	-25.577335	220	5	1.1	0.9;
	8158	2	0	0	0	0	0	1.052602	-0.559615	220	5	1.1	0.9;
	8165	1	0	-0	0	2.42	0	1.061504	-3.473077	220	5	1.1	0.9;
	8180	1	291.6	47	0	2.38	0	1.075394	-30.774767	220	5	1.1	0.9;
	8189	1	0	-0	0	9.54	0	1.042203	-18.144075	220	5	1.1	0.9;
	8190	1	0	-0	0	0.32	0	1.048784	-10.298904	220	5	1.1	0.9;
	8191	1	0	0	0	11.38	0	1.051859	-15.264149	220	5	1.1	0.9;
	8195	1	-101.01	-129.29	0	37.176703	0	1.02377	-17.045599	380	5	1.1	0.9;
	8200	1	0	-0	0	39.58	0	1.021119	-6.153606	380	5	1.1	0.9;
	8209	1	83.39	16.5	0	16.43	0	1.033486	-20.136338	220	5	1.1	0.9;
	8214	1	58.97	11.1	0	1.05	0	1.04259	-11.470366	220	5	1.1	0.9;
	8222	2	0	0	0	0	0	1.02335	-10.438616	380	5	1.1	0.9;
	8225	2	0	0	0	0	0	1.071504	-11.705551	220	5	1.1	0.9;
	8250	1	196.1	32.1	0	3.86	0	1.04037	-25.998622	220	5	1.1	0.9;
	8255	1	39.04	13.1	0	3.22	0	1.021879	-13.796655	220	5	1.1	0.9;
	8265	1	40.44	20.5	0	0.85	0	1.066641	-23.859918	220	5	1.1	0.9;
	8267	2	0	0	0	0	0	1.043326	-9.93055	380	5	1.1	0.9;
	8291	1	0	-0	0	6.52	0	1.043396	-7.697439	220	5	1.1	0.9;
	8293	1	112.7	14.5	0	1.24	0	1.048233	-24.806368	220	5	1.1	0.9;
	8294	1	14.22	6.3	0	0.38	0	1.043617	-11.857424	220	5	1.1	0.9;
	8307	1	-0	0	0	0.14	0	1.051092	-3.882875	220	5	1.1	0.9;
	8310	1	-0	0	0	2.46	0	1.049352	-8.638203	220	5	1.1	0.9;
	8311	2	0	0	0	0	0	1.035451	-15.519296	220	5	1.1	0.9;
	8312	2	0	0	0	0	0	1.07814	-12.220054	380	5	1.1	0.9;
	8316	1	0	0	0	0.72	0	1.084023	-26.486866	220	5	1.1	0.9;
	8328	1	82.39	14.4	0	13.37	0	1.041343	-24.415623	220	5	1.1	0.9;
	8331	1	-0	-0	0	0.05	0	1.047611	-12.932386	220	5	1.1	0.9;
	8334	1	263.96	-98.64	0	27.848535	0	1.034142	-14.749149	380	5	1.1	0.9;
	8347	1	-2.09	-1.43	0	48.358243	0	0.988493	-18.64268	380	5	1.1	0.9;
	8361	1	126.54	29.1	0	12.55	0	1.073653	-26.422933	220	5	1.1	0.9;
	8367	1	77.99	19.3	0	3.04	0	1.034784	-6.699666	220	5	1.1	0.9;
	8369	1	-0	-0	0	1.23	0	1.02695	-9.179276	220	5	1.1	0.9;
	8373	1	216.3	-6.9	0	2.44	0	1.071546	-48.132392	220	5	1.1	0.9;
	8397	1	-0	0	0	4.05	0	1.042817	-25.347834	220	5	1.1	0.9;
	8405	1	40.75	12.5	0	2.61	0	1.036479	-12.357404	220	5	1.1	0.9;
	8406	1	0	0	0	14.38	0	1.050892	-25.13855	220	5	1.1	0.9;
	8411	1	39.64	9.6	0	1.8	0	1.042819	-25.804727	220	5	1.1	0.9;
	8439	1	380.8	62.8	0	7.11	0	0.998187	-23.461096	220	5	1.1	0.9;
	8448	1	286.1	89.3	0	8.89	0	1.018069	-20.595036	220	5	1.1	0.9;
	8458	2	0	0	0	0	0	1.042925	-9.400848	380	5	1.1	0.9;
	8466	1	0	-0	0	1.97	0	1.045302	-4.473537	220	5	1.1	0.9;
	8467	1	216.2	75.3	0	2.74	0	0.98622	-22.981008	220	5	1.1	0.9;
	8468	1	21.42	11.8	0	28.85	0	1.005238	-14.948026	380	5	1.1	0.9;
	8473	2	0	0	0	0	0	1.039202	-8.879096	220	5	1.1	0.9;
	8475	1	-0	0	0	8.35	0	1.072229	-47.841521	220	5	1.1	0.9;
	8477	1	0	0	0	2.46	0	1.068335	-10.11476	220	5	1.1	0.9;
	8478	1	-0	0	0	3.01	0	1.04799	-8.490097	220	5	1.1	0.9;
	8485	1	149	99.8	0	9.27	0	1.039726	-7.369105	220	5	1.1	0.9;
	8486	2	0	0	0	0	0	1.04752	-19.530978	380	5	1.1	0.9;
	8487	1	44.45	7.68	0	17.83	0	1.003338	-3.846694	380	5	1.1	0.9;
	8494	1	235.97	34.52	0	7.15	0	1.08964	-9.846898	380	5	1.1	0.9;
	8497	1	38.84	14.2	0	3.5	0	1.046629	-14.553599	220	5	1.1	0.9;
	8507	1	37.44	9.1	0	7.87	0	1.074417	-26.347125	220	5	1.1	0.9;
	8511	1	93.6	-43.5	0	33.51	0	1.04106	-24.505876	220	5	1.1	0.9;
	8515	2	0	0	0	0	0	1.053872	-12.378436	220	5	1.1	0.9;
	8522	2	0	0	0	0	0	1.019671	-15.723378	220	5	1.1	0.9;
	8535	1	72.68	24.3	0	3.12	0	1.036308	-21.925576	220	5	1.1	0.9;
	8542	1	87.7	26	0	47.58	0	1.044937	-21.028222	220	5	1.1	0.9;
	8546	1	5.01	1.8	0	0.77	0	1.038205	-11.643532	220	5	1.1	0.9;
	8560	1	-0	0	0	1.11	0	0.984498	-18.190495	380	5	1.1	0.9;
	8564	2	0	0	0	0	0	1.019727	-19.556828	220	5	1.1	0.9;
	8568	1	175.7	56.2	0	0.13	0	1.027945	-6.041782	220	5	1.1	0.9;
	8578	1	0	-0	0	4.55	0	1.049043	-21.931929	220	5	1.1	0.9;
	8592	1	-0	0	0	0.43	0	1.051916	-3.764146	220	5	1.1	0.9;
	8626	1	136	36.1	0	6.07	0	0.99463	-17.562106	220	5	1.1	0.9;
	8627	1	-0	0	0	4.26	0	1.048517	-26.395781	220	5	1.1	0.9;
	8628	1	277.1	37	0	21.03	0	1.042001	-23.598682	220	5	1.1	0.9;
	8636	1	31.43	10.6	0	1.61	0	1.041292	-21.579777	220	5	1.1	0.9;
	8651	1	162.49	45.11	0	5.38	0	1.024808	-17.313776	220	5	1.1	0.9;
	8653	1	44.55	7.5	0	0.6	0	1.036488	-16.021006	220	5	1.1	0.9;
	8656	1	5.51	1.3	0	0.09	0	1.031987	-8.436898	220	5	1.1	0.9;
	8669	1	74.98	10.5	0	12.96	0	1.050943	-15.2714	220	5	1.1	0.9;
	8670	2	0	0	0	0	0	1.049066	-6.522112	220	5	1.1	0.9;
	8672	1	-0	-0	0	28.515005	0	1.023418	-20.934804	380	5	1.1	0.9;
	8676	2	0	0	0	0	0	1.018762	-21.841814	220	5	1.1	0.9;
	8677	1	0	0	0	2.29	0	1.018085	-10.22697	220	5	1.1	0.9;
	8683	2	0	0	0	0	0	1.041975	-24.273671	220	5	1.1	0.9;
	8689	1	42.65	18.2	0	0	0	1.036224	-20.697926	220	5	1.1	0.9;
	8691	1	305.04	77.16	0	17.43	0	1.047013	-33.585967	220	5	1.1	0.9;
	8704	1	180.7	71.8	0	2.46	0	1.039389	-7.999499	220	5	1.1	0.9;
	8707	1	101.9	24.3	0	8.53	0	1.032359	-11.707737	220	5	1.1	0.9;
	8711	1	0	0	0	3.9	0	1.040912	-12.07314	220	5	1.1	0.9;
	8721	2	0	0	0	0	0	1.054052	-15.864679	380	5	1.1	0.9;
	8722	1	-0	0	0	2.13	0	1.072192	-25.045427	220	5	1.1	0.9;
	8732	1	66.27	19.3	0	12.94	0	1.038041	-21.15003	220	5	1.1	0.9;
	8743	1	11.61	7.6	0	6.21	0	1.035613	-34.477305	220	5	1.1	0.9;
	8748	1	279.7	4.6	0	3.77	0	1.067084	-49.622846	220	5	1.1	0.9;
	8763	1	-233.23	29.6	0	207.55	0	1.04005	2.599263	380	5	1.1	0.9;
	8765	1	-0	-0	0	11.69	0	1.051157	-15.257422	220	5	1.1	0.9;
	8787	1	22.83	8	0	3.71	0	1.066984	-16.98525	220	5	1.1	0.9;
	8788	1	143.8	46.45	0	3.26	0	1.046176	-5.977711	220	5	1.1	0.9;
	8791	1	-0	-0	0	0.36	0	1.057978	-22.356166	220	5	1.1	0.9;
	8795	2	0	0	0	0	0	1.054157	-0.136311	220	5	1.1	0.9;
	8804	1	22.02	4.4	0	4.83	0	1.049065	-9.481469	220	5	1.1	0.9;
	8807	2	0	0	0	0	0	1.05109	-3.882868	220	5	1.1	0.9;
	8808	1	203.6	19.81	0	15.46	0	1.060353	-25.817931	220	5	1.1	0.9;
	8809	1	101.2	-2.9	0	0.51	0	1.029974	-32.515321	220	5	1.1	0.9;
	8818	2	0	0	0	0	0	1.050791	0.636035	220	5	1.1	0.9;
	8825	1	406.04	95.5	0	35.34	0	1.014963	-2.776003	380	5	1.1	0.9;
	8829	1	91.6	27.1	0	10.95	0	1.044583	-21.076938	220	5	1.1	0.9;
	8834	1	-0	0	0	1.52	0	1.003582	-22.536497	220	5	1.1	0.9;
	8835	1	113.7	22.3	0	15.55	0	1.036698	-22.906459	220	5	1.1	0.9;
	8843	1	208.7	63	0	3.45	0	1.054844	-11.164607	220	5	1.1	0.9;
	8846	1	20.82	0	0	0.17	0	1.070172	7.811263	220	5	1.1	0.9;
	8853	1	410.3	91.6	0	13.25	0	1.052314	-36.519464	220	5	1.1	0.9;
	8854	1	69.18	35.1	0	12.59	0	1.027794	-44.429467	220	5	1.1	0.9;
	8860	1	57.16	19.7	0	22.17	0	1.020431	-10.190501	220	5	1.1	0.9;
	8864	1	63.31	-9.8	0	5.36	0	1.048228	-31.670524	220	5	1.1	0.9;
	8872	2	0	0	0	0	0	1.053844	-15.88923	380	5	1.1	0.9;
	8873	1	0	-0	0	1.87	0	1.071142	-27.284425	220	5	1.1	0.9;
	8874	1	285.2	61.4	0	26.04	0	1.046146	-20.00976	220	5	1.1	0.9;
	8877	1	-0	0	0	1.03	0	1.039567	-20.646198	220	5	1.1	0.9;
	8879	1	0	0	0	1.47	0	1.036915	-22.747121	220	5	1.1	0.9;
	8886	1	0	-0	0	-1.392036	0	0.983701	-3.785162	380	5	1.1	0.9;
	8887	1	0	0	0	102.70867	0	1.045962	1.729833	380	5	1.1	0.9;
	8893	1	278.3	33.6	0	25.9	0	1.05128	-15.163258	220	5	1.1	0.9;
	8900	1	3.1	1.6	0	0.09	0	1.047741	-23.703947	220	5	1.1	0.9;
	8903	2	0	0	0	0	0	1.020268	-11.421507	380	5	1.1	0.9;
	8906	1	91	21	0	11.23	0	1.036	-23.062052	220	5	1.1	0.9;
	8913	1	-0	0	0	1.13	0	1.040075	-20.555337	220	5	1.1	0.9;
	8930	1	139.4	21.2	0	6.21	0	0.991658	-18.577239	220	5	1.1	0.9;
	8931	1	-0.56	-0	0	108.21	0	1.090352	-9.661963	380	5	1.1	0.9;
	8947	1	3.7	0.2	0	0.05	0	1.03998	-11.496674	220	5	1.1	0.9;
	8950	2	0	0	0	0	0	1.047268	-14.839334	220	5	1.1	0.9;
	8961	2	0	0	0	0	0	1.035459	-14.187287	220	5	1.1	0.9;
	8975	1	0	-0	0	11.04	0	1.04225	-24.276259	220	5	1.1	0.9;
	8976	2	0	0	0	0	0	1.046972	-17.074322	220	5	1.1	0.9;
	8980	1	63.97	12.8	0	0.17	0	1.042994	-11.072487	220	5	1.1	0.9;
	8989	1	88.5	24.5	0	2.23	0	1.043398	-25.575347	220	5	1.1	0.9;
	8992	1	-0	-0	0	14.64	0	1.051114	-3.785395	220	5	1.1	0.9;
	8997	2	0	0	0	0	0	0.997131	-13.27089	380	5	1.1	0.9;
	8999	1	41.15	10.3	0	2.41	0	1.044391	-21.558302	220	5	1.1	0.9;
	9002	1	119.02	12.54	0	23.45	0	1.021009	-10.528192	220	5	1.1	0.9;
	9011	1	88.4	25.2	0	12.9	0	1.036275	-23.026084	220	5	1.1	0.9;
	9012	1	25.63	6.5	0	0.52	0	1.048758	1.093286	220	5	1.1	0.9;
	9014	1	70.78	19.6	0	8.1	0	1.051072	-6.681904	220	5	1.1	0.9;
	9018	1	55.46	16.9	0	1.92	0	1.022715	-6.602667	220	5	1.1	0.9;
	9019	1	91.4	18.1	0	4.89	0	1.040421	-24.677298	220	5	1.1	0.9;
	9021	1	152.18	21.63	0	1.91	0	1.035437	-6.308005	220	5	1.1	0.9;
	9033	1	76.88	16.4	0	6.61	0	1.049786	-26.382004	220	5	1.1	0.9;
	9045	1	60.57	18.9	0	1.46	0	1.039905	-24.629587	220	5	1.1	0.9;
	9051	1	0	-0	0	2.15	0	1.036594	0.577238	220	5	1.1	0.9;
	9059	1	0	-0	0	2	0	1.072487	-27.192131	220	5	1.1	0.9;
	9065	1	-0	-0	0	0.59	0	1.040021	2.414524	380	5	1.1	0.9;
	9066	1	23.26	3.01	0	9.29	0	1.019989	-20.806005	220	5	1.1	0.9;
	9067	2	0	0	0	0	0	1.040924	1.549232	220	5	1.1	0.9;
	9091	1	0	0	0	3.65	0	1.038554	-9.767436	220	5	1.1	0.9;
	9101	2	0	0	0	0	0	1.02885	-21.628013	380	5	1.1	0.9;
	9108	2	0	0	0	0	0	1.046626	-16.441009	220	5	1.1	0.9;
	9109	1	0	-0	0	0.31	0	1.036596	0.577218	220	5	1.1	0.9;
	9112	1	0	-0	0	19.71	0	1.047611	-12.93238	220	5	1.1	0.9;
	9119	1	-0	0	0	1.24	0	1.04289	-20.527494	220	5	1.1	0.9;
	9128	1	50.56	8.7	0	0.19	0	1.042244	-2.676445	220	5	1.1	0.9;
	9130	1	245.51	9.17	0	24.25	0	1.049965	-30.721639	220	5	1.1	0.9;
	9131	1	-0	-0	0	6.62	0	1.051124	-24.097314	220	5	1.1	0.9;
	9137	2	0	0	0	0	0	1.049768	-26.378636	220	5	1.1	0.9;
	9140	2	0	0	0	0	0	1.049624	-5.232971	220	5	1.1	0.9;
	9150	2	0	0	0	0	0	1.037969	-13.200551	380	5	1.1	0.9;
	9155	1	-0	-0	0	5.01	0	1.02382	-20.860292	220	5	1.1	0.9;
	9158	1	54.96	9.42	0	5.92	0	1.025798	-14.56483	220	5	1.1	0.9;
	9164	1	7.71	2.4	0	6.93	0	1.001887	-19.403088	220	5	1.1	0.9;
	9173	1	66.07	15.6	0	5.57	0	1.032906	-20.202148	220	5	1.1	0.9;
	9174	2	0	0	0	0	0	1.03623	-32.377948	380	5	1.1	0.9;
	9176	1	-0	0	0	1.28	0	1.047842	-12.93497	220	5	1.1	0.9;
	9180	2	0	0	0	0	0	1.061189	-8.603491	220	5	1.1	0.9;
	9181	1	2.43	1.35	0	0.1	0	1.032498	-9.466095	220	5	1.1	0.9;
	9185	1	66.97	18.6	0	18.5	0	1.051815	-6.62338	220	5	1.1	0.9;
	9189	1	62.57	-61	0	55.14	0	1.048019	-25.022568	220	5	1.1	0.9;
	9191	1	-0	-0	0	2.34	0	1.073685	-24.80852	220	5	1.1	0.9;
	9203	1	627.36	-73.13	0	27.185221	0	1.001613	-12.605605	380	5	1.1	0.9;
	9213	1	-0	-0	0	10.210952	0	1.023742	-4.653017	380	5	1.1	0.9;
	9217	1	-0	0	0	6.33	0	1.040155	-8.875709	220	5	1.1	0.9;
	9222	1	-5.15	-2.04	0	154.446402	0	1.053345	-16.073005	380	5	1.1	0.9;
	9231	1	242.08	59.98	0	5.6	0	1.029833	-19.150137	220	5	1.1	0.9;
	9241	1	0	-0	0	0.36	0	1.049081	-9.481609	220	5	1.1	0.9;
];

%% generator data
%	bus	Pg	Qg	Qmax	Qmin	Vg	mBase	status	Pmax	Pmin	Pc1	Pc2	Qc1min	Qc1max	Qc2min	Qc2max	ramp_agc	ramp_10	ramp_30	ramp_q	apf
mpc.gen = [
	124	861.3	56.53	440.32	-172.59	1.081537	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	150	67.29	-7.99	77.09	-22.54	1.062805	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	221	-155.39	97.68	183.92	-65.3	0.984682	100	1	100	-207.18	0	0	0	0	0	0	0	0	0	0	0;
	338	83.7	2.5	55.54	-19.16	1.048249	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	352	1250.4	278.94	926.35	-297.68	1.055378	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	413	231	19.11	124.12	-49.84	1.05013	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	453	260.1	57.98	257.23	-74.4	1.018914	100	1	540	0	0	0	0	0	0	0	0	0	0	0	0;
	516	539.4	190.43	261.79	-106.67	1.075592	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	564	-51.67	-3.31	18.47	-19.35	1.069817	100	1	100	-68.89	0	0	0	0	0	0	0	0	0	0	0;
	583	-270	3.65	17.88	-7.03	1.045397	100	1	100	-360	0	0	0	0	0	0	0	0	0	0	0;
	615	38.2	-0.39	17.09	-7.79	1.063876	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	616	379.2	37.76	217.9	-81.16	1.048713	100	1	480	160	0	0	0	0	0	0	0	0	0	0	0;
	639	-46.2	112.55	135.52	-38.9	1.051376	100	1	100	-61.6	0	0	0	0	0	0	0	0	0	0	0;
	682	37.4	8.53	38.77	-11.06	1.0691	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	749	198.88	-11.17	105.69	-43.84	1.057884	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	757	116.7	13.66	51.3	-23.41	1.017567	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	776	-111.4	3.36	18.16	-6.73	1.026368	100	1	100	-148.53	0	0	0	0	0	0	0	0	0	0	0;
	778	16.5	-0.67	12.37	-9.36	1.049411	100	1	33.58	0	0	0	0	0	0	0	0	0	0	0	0;
	795	58.4	7.75	36.75	-13.04	1.035522	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	803	-89.2	-1.52	74.03	-25.54	1.041369	100	1	100	-118.93	0	0	0	0	0	0	0	0	0	0	0;
	823	1274.2	289.74	926.68	-298.31	1.044097	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	851	59.56	50.82	96.64	-27.81	1.00025	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	858	855.6	147.86	440.67	-172.15	1.031745	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	891	1702.8	651.53	1097.41	-373.91	1.064525	100	1	2400	800	0	0	0	0	0	0	0	0	0	0	0;
	972	1221.2	393.14	934.73	-291.52	1.072621	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	1001	92.1	25.97	55.02	-19.81	1.074143	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1002	171	10.89	89.55	-35.15	1.041506	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	1043	829.5	49	387.88	-163.4	1.028946	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	1083	91.8	9.85	54.91	-19.88	1.049472	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1093	837	83.03	550.69	-185.15	1.027056	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	1100	58.2	5.6	36.8	-13.01	1.042801	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1102	-124.15	7.32	38.05	-11.59	1.049879	100	1	100	-165.54	0	0	0	0	0	0	0	0	0	0	0;
	1237	1504.8	364.78	906.61	-319.44	1.108028	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	1251	837	77.94	550.69	-185.15	1.027102	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	1295	-124.1	4.52	36.92	-12.88	1.03682	100	1	100	-165.47	0	0	0	0	0	0	0	0	0	0	0;
	1341	67.2	29.74	57.03	-17.62	1.032899	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1354	-246.5	13.72	38.08	-11.73	1.041045	100	1	100	-328.67	0	0	0	0	0	0	0	0	0	0	0;
	1422	-101	2.33	18.09	-6.79	1.04353	100	1	100	-134.67	0	0	0	0	0	0	0	0	0	0	0;
	1436	49.4	2.95	37.72	-12.09	1.047684	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1478	62.6	7.49	36.4	-13.44	1.048068	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1604	50.6	13.86	37.66	-12.17	1.043862	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1642	50.4	11.76	37.74	-12.11	1.028554	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1680	1250.4	274.36	928.89	-296.1	1.04395	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	1708	860.8	121.94	439.37	-173.3	1.037938	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	1721	28.1	1.73	18.52	-6.39	1.062238	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1754	40	3.62	16.99	-7.92	1.04286	100	1	41.06	0	0	0	0	0	0	0	0	0	0	0	0;
	1794	1250.4	253.94	926.35	-297.68	1.05295	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	1808	-0.6	-13.17	111.23	-38.16	1.046224	100	1	100	-0.8	0	0	0	0	0	0	0	0	0	0	0;
	1851	15.6	-1.26	11.76	-8.89	1.04762	100	1	31.92	0	0	0	0	0	0	0	0	0	0	0	0;
	1852	-241.8	10.63	58.16	-16.52	1.054325	100	1	100	-322.4	0	0	0	0	0	0	0	0	0	0	0;
	1888	573.68	42.43	396.02	-155.83	1.039377	100	1	912.78	312.78	0	0	0	0	0	0	0	0	0	0	0;
	1914	115.5	-45.22	160.36	-150.95	1.035636	100	1	472.16	39.35	0	0	0	0	0	0	0	0	0	0	0;
	1959	31	0.56	18.14	-6.75	1.064692	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2035	855.6	149.85	441.85	-171.27	1.03088	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	2050	836.2	-16.43	962.2	-263.65	1.009914	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	2085	176	31.77	88.39	-36.19	1.040351	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2177	-121.3	13.33	58.35	-16.26	1.07333	100	1	100	-161.73	0	0	0	0	0	0	0	0	0	0	0;
	2197	-53.29	-1.42	19.51	-14.74	1.035528	100	1	100	-71.05	0	0	0	0	0	0	0	0	0	0	0;
	2276	38.9	-22.64	41.67	-36.5	1.05324	100	1	119.3	0	0	0	0	0	0	0	0	0	0	0	0;
	2291	110.5	13.69	95.69	-28.93	1.072528	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2359	78.2	7.2	34.35	-15.5	1.024884	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2421	-320.4	4.66	37.39	-12.45	1.047766	100	1	100	-427.2	0	0	0	0	0	0	0	0	0	0	0;
	2425	-100.5	5.31	54.3	-20.49	1.041186	100	1	100	-134	0	0	0	0	0	0	0	0	0	0	0;
	2426	-43.2	-30.65	105.34	-99.02	1.037687	100	1	100	-57.6	0	0	0	0	0	0	0	0	0	0	0;
	2446	1367	286.48	921.62	-304.73	1.06157	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	2468	46.8	5.15	38.02	-11.81	1.048132	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2481	31.8	11.5	39.12	-10.71	1.042777	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2489	286.5	-15.8	162.4	-61.79	1.049366	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	2550	54.3	4.31	58.23	-16.49	1.049384	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2600	76.48	16.9	76.08	-23.59	1.038821	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	2627	1367	302.84	921.52	-304.79	1.061488	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	2653	41.92	3.94	51.94	-22.8	1.036517	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2719	77.6	-15.35	70.15	-58.1	1.035384	100	1	196.72	0	0	0	0	0	0	0	0	0	0	0	0;
	2786	1221.2	372.42	934.81	-291.47	1.073795	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	2797	-95.1	-11.62	88.33	-36.33	1.042226	100	1	100	-126.8	0	0	0	0	0	0	0	0	0	0	0;
	2799	66.8	-1.21	77.91	-21.66	1.05396	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	2816	1250.4	270.9	926.26	-297.74	1.053944	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	2841	80.7	-3.29	55.98	-18.76	1.08248	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2842	31.6	5.75	18.19	-6.74	1.043712	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2872	4.5	0.33	3.31	-2.54	1.063841	100	1	9.05	0	0	0	0	0	0	0	0	0	0	0	0;
	2878	8.95	-11.72	72.95	-68.71	1.057626	100	1	214.86	0	0	0	0	0	0	0	0	0	0	0	0;
	2886	60.9	10.71	57.69	-17.02	1.055151	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2902	20.9	1.11	19.19	-5.72	1.059072	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2930	53.2	2.49	37.24	-12.54	1.047695	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2934	3.4	-0.16	2.51	-1.91	1.051149	100	1	6.83	0	0	0	0	0	0	0	0	0	0	0	0;
	2985	46.6	-69.84	184.46	-132.67	1.053607	100	1	493.35	0	0	0	0	0	0	0	0	0	0	0	0;
	3018	162.5	-4.77	89.61	-34.82	1.041705	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	3028	855.7	146.13	441.85	-171.27	1.030654	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	3114	38.6	4.12	17.33	-7.61	1.071964	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	3133	85.54	0.39	113.97	-35.56	1.065372	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	3134	-136.3	-11.71	70.25	-29.33	1.023117	100	1	100	-181.73	0	0	0	0	0	0	0	0	0	0	0;
	3183	98.7	-3.57	53.7	-20.98	1.055515	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3205	416.7	119.72	243.61	-88.46	1.041982	100	1	540	0	0	0	0	0	0	0	0	0	0	0	0;
	3218	-107.4	4.44	36.79	-13	1.028507	100	1	100	-143.2	0	0	0	0	0	0	0	0	0	0	0;
	3240	-42.3	-25.61	62.23	-51.37	1.050305	100	1	100	-56.4	0	0	0	0	0	0	0	0	0	0	0;
	3306	861.3	143.67	439.2	-173.43	1.071978	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	3324	-26.48	12.72	38.13	-11.64	1.014126	100	1	100	-35.3	0	0	0	0	0	0	0	0	0	0	0;
	3346	56.8	11.49	37.1	-12.76	1.01594	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3353	19.07	5.42	38.2	-11.61	1.045739	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3364	44.5	-21.96	47.19	-41.44	1.0538	100	1	135.27	0	0	0	0	0	0	0	0	0	0	0	0;
	3390	2.4	12.68	54.59	-20.19	1.039984	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3422	53.7	-3.24	58.19	-16.5	1.063657	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3436	521	217.43	550.97	-184.89	1.060777	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	3492	-370.8	2.02	53.75	-21.08	1.04852	100	1	100	-494.4	0	0	0	0	0	0	0	0	0	0	0;
	3513	829.8	51	387.88	-163.4	1.028771	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	3565	15.88	2.27	38.83	-10.97	1.081614	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3580	539	189.12	261.7	-107.36	1.07602	100	1	600	0	0	0	0	0	0	0	0	0	0	0	0;
	3611	-125.4	3.51	54.59	-20.12	1.03905	100	1	100	-167.2	0	0	0	0	0	0	0	0	0	0	0;
	3656	-26.2	12.26	116.07	-33.32	1.070983	100	1	100	-34.93	0	0	0	0	0	0	0	0	0	0	0;
	3661	861.2	118.75	441.04	-172.05	1.059676	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	3698	-12.58	3.09	36.34	-13.47	1.041813	100	1	100	-16.78	0	0	0	0	0	0	0	0	0	0	0;
	3741	854.7	190.06	548.57	-187.16	1.031629	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	3809	257	122.15	313.01	-116.73	1.084785	100	1	700	233.33	0	0	0	0	0	0	0	0	0	0	0;
	3825	837.3	54.8	388.17	-163.4	1.026994	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	3869	855.6	148.23	440.69	-172.14	1.031907	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	3876	-25.1	6.45	34.27	-15.52	1.044501	100	1	100	-33.47	0	0	0	0	0	0	0	0	0	0	0;
	3916	-21.27	11.22	38.02	-11.77	1.043758	100	1	100	-28.37	0	0	0	0	0	0	0	0	0	0	0;
	3951	-27.1	-74.05	126.16	-127.56	1.027439	100	1	100	-36.13	0	0	0	0	0	0	0	0	0	0	0;
	3971	-134.5	9.64	55.67	-19.09	1.063145	100	1	100	-179.33	0	0	0	0	0	0	0	0	0	0	0;
	4024	89.5	40.38	97.07	-27.44	1.040328	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	4056	-76.9	3.73	38.05	-11.77	1.038526	100	1	100	-102.53	0	0	0	0	0	0	0	0	0	0	0;
	4084	366.1	46.44	264.94	-103.74	1.063234	100	1	600	0	0	0	0	0	0	0	0	0	0	0	0;
	4118	41	5.09	38.37	-11.4	1.027195	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4125	59.6	16.3	36.89	-13	1.003502	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4128	67.5	24.27	57.03	-17.63	1.019761	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4231	2641.24	771.6	Inf	-Inf	1.049182	100	1	4188.95	1333.33	0	0	0	0	0	0	0	0	0	0	0;
	4331	60.2	6.92	36.73	-13.12	1.028646	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4395	-129.1	6.73	38.48	-11.33	1.039834	100	1	100	-172.13	0	0	0	0	0	0	0	0	0	0	0;
	4419	861.3	121.55	441.04	-172.05	1.059565	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	4480	861.3	143.79	439.2	-173.43	1.071852	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	4482	35.2	13.53	38.85	-10.95	1.049914	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4506	-171.8	12.84	37.86	-11.95	1.036047	100	1	100	-229.07	0	0	0	0	0	0	0	0	0	0	0;
	4566	-30.39	6.81	37.94	-11.87	1.073282	100	1	100	-40.51	0	0	0	0	0	0	0	0	0	0	0;
	4624	68	8.25	35.62	-14.19	1.042135	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4661	-78.9	2.19	19.14	-5.77	1.049813	100	1	100	-105.2	0	0	0	0	0	0	0	0	0	0	0;
	4701	93.2	-41.31	98.4	-86.29	1.043063	100	1	281.85	0	0	0	0	0	0	0	0	0	0	0	0;
	4783	861.3	141.62	439.2	-173.43	1.071981	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	4816	-26.08	10.07	38.37	-11.44	1.067305	100	1	100	-34.77	0	0	0	0	0	0	0	0	0	0	0;
	4819	855.6	150.1	440.69	-172.14	1.031907	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	4823	10.3	1.59	4.44	-2.07	1.067114	100	1	10.46	0	0	0	0	0	0	0	0	0	0	0	0;
	4850	88.8	11.86	76.52	-23.19	1.054975	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	4880	-40.3	5.06	75.62	-23.96	1.034991	100	1	100	-53.73	0	0	0	0	0	0	0	0	0	0	0;
	4918	41.69	8.64	37.27	-12.57	1.057325	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4952	-12.57	4.47	37.23	-12.61	1.045386	100	1	100	-16.77	0	0	0	0	0	0	0	0	0	0	0;
	5004	54.3	4.85	74.73	-70.41	1.0663	100	1	220.12	0	0	0	0	0	0	0	0	0	0	0	0;
	5019	49.4	2.95	37.72	-12.09	1.047693	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5051	83.48	-1.38	74.55	-24.97	1.037373	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5067	542	142.39	258.91	-109.58	1.021924	100	1	600	0	0	0	0	0	0	0	0	0	0	0	0;
	5110	-52.51	8.27	38.28	-11.5	1.032061	100	1	100	-70.01	0	0	0	0	0	0	0	0	0	0	0;
	5120	75.6	-5.04	56.35	-18.42	1.055539	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5144	1713.6	139.22	882.86	-343.24	1.035201	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	5237	106.8	10.9	75.02	-24.74	1.052719	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5278	119.4	-0.06	51.3	-23.49	1.042037	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5340	-159	11.05	36.37	-13.43	1.054941	100	1	100	-212	0	0	0	0	0	0	0	0	0	0	0;
	5365	19.59	1.13	18.14	-6.75	1.061754	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	5379	1250.6	358.22	926.47	-297.63	1.107976	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	5395	-395.8	3.01	58.06	-16.66	1.048137	100	1	100	-527.73	0	0	0	0	0	0	0	0	0	0	0;
	5461	43.4	10.92	38.26	-11.55	1.06213	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5481	1368.6	146.68	920.31	-305.64	1.063711	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	5482	56.8	-22.21	101.32	-48.12	1.044486	100	1	249.06	0	0	0	0	0	0	0	0	0	0	0	0;
	5486	861.3	119.59	441.04	-172.05	1.059636	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	5488	-96.4	14.63	37.49	-12.33	1.041881	100	1	100	-128.53	0	0	0	0	0	0	0	0	0	0	0;
	5490	3424.8	295.7	1765.36	-686.67	1.068845	100	1	4000	1333.33	0	0	0	0	0	0	0	0	0	0	0;
	5533	173	14.6	88.87	-35.71	1.038212	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	5546	21.5	1.97	19.16	-5.75	1.05378	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	5564	57.6	11.81	36.8	-12.99	1.04123	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5658	229.8	77.84	103.18	-46.25	1.017689	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	5664	128	14.87	72.71	-27.05	1.070507	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5709	837	75.6	550.69	-185.15	1.027078	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	5781	1102.83	396.26	779.29	-324.79	1.0396	100	1	1800	600	0	0	0	0	0	0	0	0	0	0	0;
	5814	181.2	41.72	196.98	-77.17	1.044051	100	1	440	146.67	0	0	0	0	0	0	0	0	0	0	0;
	5856	305.7	19.99	214.13	-85	1.038417	100	1	480	160	0	0	0	0	0	0	0	0	0	0	0;
	5881	60	7.13	36.72	-13.13	1.065605	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5940	-74.9	4.51	17.21	-7.73	1.030075	100	1	100	-99.87	0	0	0	0	0	0	0	0	0	0	0;
	5971	1221.2	370.23	934.54	-291.64	1.070056	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	5983	-102.5	3.55	37.27	-12.59	1.066029	100	1	100	-136.67	0	0	0	0	0	0	0	0	0	0	0;
	5994	-87.3	-16.58	60.51	-61.08	1.029184	100	1	100	-116.4	0	0	0	0	0	0	0	0	0	0	0;
	6036	820.2	300.86	390.4	-161.16	1.043738	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6153	94.3	-78.07	249.65	-261.82	0.988112	100	1	769.94	32.08	0	0	0	0	0	0	0	0	0	0	0;
	6168	26.3	0.4	19.86	-14.93	1.040343	100	1	53.76	0	0	0	0	0	0	0	0	0	0	0	0;
	6291	55.5	24.55	58.12	-16.6	1.040598	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6331	48.8	5.01	37.76	-12.05	1.052181	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6332	96	-9.58	96.56	-27.95	1.062214	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	6351	22.08	14.99	38.29	-11.56	1.064446	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6368	-145.3	3	51.11	-23.46	1.043835	100	1	100	-193.73	0	0	0	0	0	0	0	0	0	0	0;
	6376	-27.8	-23.17	148.65	-50.73	1.046714	100	1	100	-37.07	0	0	0	0	0	0	0	0	0	0	0;
	6429	171	18.2	88.83	-35.69	1.039574	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	6474	718.5	-12.86	438.72	-174.15	1.06557	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	6516	228.2	12.73	126.73	-47.83	1.052452	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	6552	-204.2	5.93	37.58	-12.16	1.048796	100	1	100	-272.27	0	0	0	0	0	0	0	0	0	0	0;
	6734	382.7	-77.78	311.52	-132.09	1.061492	100	1	720	0	0	0	0	0	0	0	0	0	0	0	0;
	6807	44.07	31.99	77.14	-22.46	1.006037	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	6816	16.53	-8.62	77.35	-22.26	1.053803	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	6820	15.9	-0.62	11.99	-9.06	1.049557	100	1	32.53	0	0	0	0	0	0	0	0	0	0	0	0;
	6831	-57.5	4.01	35.62	-14.25	1.031347	100	1	100	-76.67	0	0	0	0	0	0	0	0	0	0	0;
	6845	102.9	17.09	53.24	-21.47	1.055216	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6852	-344.9	1.84	18.56	-6.34	1.031296	100	1	100	-459.87	0	0	0	0	0	0	0	0	0	0	0;
	6857	2463.6	391.48	1865.83	-585.95	1.021411	100	1	4000	1333.33	0	0	0	0	0	0	0	0	0	0	0;
	6888	133	6.97	189.31	-59.87	1.039234	100	1	300	88.67	0	0	0	0	0	0	0	0	0	0	0;
	6947	227.5	5.93	126.09	-48.29	1.051367	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	6969	39.8	-1.81	38.53	-11.27	1.059486	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6982	-12.6	9.53	56.22	-18.44	1.036531	100	1	100	-16.8	0	0	0	0	0	0	0	0	0	0	0;
	7036	65.8	11.07	36.16	-13.72	1.041241	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7049	55.5	16.47	58.18	-16.56	1.049669	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7056	448.85	137.55	267.36	-100.98	1.006597	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7115	65.87	13.66	95.06	-83.34	1.008722	100	1	272.28	0	0	0	0	0	0	0	0	0	0	0	0;
	7159	861.3	121.21	441.04	-172.05	1.059575	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	7183	1250.6	358.08	926.47	-297.63	1.107733	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	7209	7.3	0.36	3.12	-1.46	1.051113	100	1	7.35	0	0	0	0	0	0	0	0	0	0	0	0;
	7267	-39.7	22.41	89.84	-34.84	1.010817	100	1	100	-52.93	0	0	0	0	0	0	0	0	0	0	0;
	7282	2340.4	94.47	1862.29	-588.16	1.034538	100	1	4000	1333.33	0	0	0	0	0	0	0	0	0	0	0;
	7327	59.06	19.56	76.11	-23.62	1.043172	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	7328	53	24.18	37.45	-12.39	0.999675	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7466	-202.53	33.59	75.85	-77.06	1.057257	100	1	100	-270.04	0	0	0	0	0	0	0	0	0	0	0;
	7474	76.5	8.44	56.27	-18.41	1.027108	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7495	75.8	67.52	214.23	-59.77	1.035256	100	1	200	50.53	0	0	0	0	0	0	0	0	0	0	0;
	7520	702.4	277.3	350.36	-140.73	1.037447	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	7522	508	235.59	371.11	-119.67	0.99994	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	7641	769.2	22.2	378.21	-162.35	1.051427	100	1	880	200	0	0	0	0	0	0	0	0	0	0	0;
	7697	228.2	12.72	126.73	-47.83	1.052424	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	7776	1221.2	354.58	934.73	-291.52	1.072261	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	7808	1250.4	279.8	926.35	-297.68	1.054562	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	7842	165	23.34	90.08	-34.56	1.001912	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	7913	1368.6	146.28	920.41	-305.58	1.063764	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	7998	811.8	175.43	391.57	-160.95	1.048854	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	8043	-27	4.11	75.67	-23.96	1.066913	100	1	100	-36	0	0	0	0	0	0	0	0	0	0	0;
	8109	992	864.28	Inf	-Inf	1.067387	100	1	1041.6	0	0	0	0	0	0	0	0	0	0	0	0;
	8158	153	-2.71	91.67	-33.01	1.052602	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	8222	66.8	17.58	35.76	-14.05	1.02335	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8225	17	6.35	19.49	-5.42	1.071504	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	8267	203.4	-3.04	106.84	-42.56	1.043326	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	8311	-8.75	3.43	17.8	-7.13	1.035451	100	1	100	-11.66	0	0	0	0	0	0	0	0	0	0	0;
	8312	1702.2	433.08	1096.33	-374.64	1.07814	100	1	2400	800	0	0	0	0	0	0	0	0	0	0	0;
	8458	517.1	-58.18	474.42	-138.49	1.042925	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	8473	-135	-7.13	90.21	-34.45	1.039202	100	1	100	-180	0	0	0	0	0	0	0	0	0	0	0;
	8486	78.8	9.58	34.16	-15.67	1.04752	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8515	32.3	-18.17	93.03	-31.43	1.053872	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	8522	-75.69	1.02	3.61	-1.69	1.019671	100	1	100	-100.92	0	0	0	0	0	0	0	0	0	0	0;
	8564	-65	20.7	51.91	-22.86	1.019727	100	1	100	-86.67	0	0	0	0	0	0	0	0	0	0	0;
	8670	-138.3	10.82	37.4	-12.46	1.049066	100	1	100	-184.4	0	0	0	0	0	0	0	0	0	0	0;
	8676	-259.1	12.69	34.92	-14.96	1.018762	100	1	100	-345.47	0	0	0	0	0	0	0	0	0	0	0;
	8683	-37.28	0.32	54.32	-20.39	1.041975	100	1	100	-49.7	0	0	0	0	0	0	0	0	0	0	0;
	8721	1232.6	152.55	932.68	-293.14	1.054052	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	8795	166.2	-6.14	111.7	-37.85	1.054157	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	8807	-304	3.53	35.84	-14.07	1.05109	100	1	100	-405.33	0	0	0	0	0	0	0	0	0	0	0;
	8818	44.66	-2.57	76.87	-22.75	1.050791	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	8872	1232.6	140.05	932.68	-293.14	1.053844	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	8903	17	-1.08	28.99	-28.41	1.020268	100	1	86.73	0	0	0	0	0	0	0	0	0	0	0	0;
	8950	-255.7	7.67	38.23	-11.61	1.047268	100	1	100	-340.94	0	0	0	0	0	0	0	0	0	0	0;
	8961	-61.8	-11.18	53.52	-50.38	1.035459	100	1	100	-82.4	0	0	0	0	0	0	0	0	0	0	0;
	8976	110	15.22	95.65	-28.95	1.046972	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	8997	78.2	-5.23	107.59	-101.48	0.997131	100	1	317.09	0	0	0	0	0	0	0	0	0	0	0	0;
	9067	369.6	48.34	198.07	-76.34	1.040924	100	1	440	146.67	0	0	0	0	0	0	0	0	0	0	0;
	9101	698.8	0.54	442.99	-169.92	1.02885	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	9108	-239.23	-1.6	18.65	-14.16	1.046626	100	1	100	-318.97	0	0	0	0	0	0	0	0	0	0	0;
	9137	42.7	18.81	335.96	-93.23	1.049768	100	1	100	28.47	0	0	0	0	0	0	0	0	0	0	0;
	9140	-37.2	28.17	73.19	-26.48	1.049624	100	1	100	-49.6	0	0	0	0	0	0	0	0	0	0	0;
	9150	860.8	117.32	439.38	-173.28	1.037969	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	9174	-331.4	142.7	175	-49.16	1.03623	100	1	100	-441.87	0	0	0	0	0	0	0	0	0	0	0;
	9180	72	7.71	77.69	-21.97	1.061189	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
];

%% branch data
%	fbus	tbus	r	x	b	rateA	rateB	rateC	ratio	angle	status	angmin	angmax
mpc.branch = [
	7351	5441	0.00018	0.000781	0	0	0	0	0	0	1	-360	360;
	4314	7571	0.002961	0.01669	0	319	0	0	0	0	1	-360	360;
	5803	3916	0.00056	0.00432	0	0	0	0	0	0	1	-360	360;
	6757	6036	0.0002	0.00246	0	657	0	0	0	0	1	-360	360;
	6757	6921	0.0003	0.00377	0	1315	0	0	0	0	1	-360	360;
	5019	9112	0.00014	0.00057	0	0	0	0	0	0	1	-360	360;
	2930	9112	0.00013	0.000729	0	0	0	0	0	0	1	-360	360;
	1436	9112	0.00012	0.000581	0	0	0	0	0	0	1	-360	360;
	1851	9112	0.00011	0.00064	0	0	0	0	0	0	1	-360	360;
	8331	9112	0.00019	0.00056	0	0	0	0	0	0	1	-360	360;
	6220	8791	0.00062	0.004909	0	529	0	0	0	0	1	-360	360;
	9203	7842	7e-05	0.00076	0	887	0	0	0	0	1	-360	360;
	9203	8997	0.00066	0.00622	0	1611	0	0	0	0	1	-360	360;
	9203	2129	0.00054	0.00752	0	0	0	0	0	0	1	-360	360;
	9203	29	0.00105	0.00877	0	1315	0	0	0	0	1	-360	360;
	118	8903	8e-05	0.00091	0	657	0	0	0	0	1	-360	360;
	118	4598	0.00056	0.00719	0	953	0	0	0	0	1	-360	360;
	2426	5533	0.000531	0.002039	0	319	0	0	0	0	1	-360	360;
	2426	4831	0.008659	0.03769	0	300	0	0	0	0	1	-360	360;
	2426	8961	0.003849	0.021099	0	0	0	0	0	0	1	-360	360;
	2426	8522	0.00688	0.035969	0	300	0	0	0	0	1	-360	360;
	2426	6888	0.006729	0.03713	0	0	0	0	0	0	1	-360	360;
	2426	3401	0.00111	0.006651	0	414	0	0	0	0	1	-360	360;
	2426	641	0.0023	0.008909	0	0	0	0	0	0	1	-360	360;
	2426	2600	0.000841	0.00313	0	0	0	0	0	0	1	-360	360;
	2426	5351	0.003521	0.017479	0	300	0	0	0	0	1	-360	360;
	3051	718	3.9e-05	0.00025	0	510	0	0	0	0	1	-360	360;
	3036	2327	0.000159	0.000729	0	357	0	0	0	0	1	-360	360;
	8651	7473	0.005281	0.032099	0	395	0	0	0	0	1	-360	360;
	8651	7473	0.005651	0.029419	0	376	0	0	0	0	1	-360	360;
	903	3654	0.000271	0.00036	0	0	0	0	0	0	1	-360	360;
	8732	7050	0.00044	0.00143	0	472	0	0	0	0	1	-360	360;
	8732	3912	0.000159	0.00074	0	453	0	0	0	0	1	-360	360;
	2273	1973	0.00019	0.000591	0	338	0	0	0	0	1	-360	360;
	5131	2928	0.00019	0.00087	0	0	0	0	0	0	1	-360	360;
	6629	4918	0.002591	0.013659	0	319	0	0	0	0	1	-360	360;
	6629	7309	0.00305	0.01168	0	0	0	0	0	0	1	-360	360;
	413	1102	0.000159	0.000599	0	433	0	0	0	0	1	-360	360;
	413	1102	0.000159	0.00062	0	395	0	0	0	0	1	-360	360;
	3248	4918	0.00101	0.004229	0	472	0	0	0	0	1	-360	360;
	3248	4281	0.00319	0.018651	0	453	0	0	0	0	1	-360	360;
	3248	7309	0.000729	0.00337	0	0	0	0	0	0	1	-360	360;
	1146	1187	0.002961	0.012831	0	281	0	0	0	0	1	-360	360;
	1146	7945	0.000591	0.00339	0	0	0	0	0	0	1	-360	360;
	8900	2421	0.00024	0.001219	0	0	0	0	0	0	1	-360	360;
	6071	2972	0.006219	0.036341	0	376	0	0	0	0	1	-360	360;
	6071	6031	0.001841	0.01193	0	472	0	0	0	0	1	-360	360;
	3293	7264	0.001409	0.01113	0	414	0	0	0	0	1	-360	360;
	3293	283	0.00151	0.00768	0	376	0	0	0	0	1	-360	360;
	3293	3069	0.00599	0.03313	0	395	0	0	0	0	1	-360	360;
	3293	3390	0.001771	0.010151	0	357	0	0	0	0	1	-360	360;
	4689	4936	0.00074	0.00311	0	0	0	0	0	0	1	-360	360;
	4689	4936	0.000729	0.003159	0	0	0	0	0	0	1	-360	360;
	3218	7809	0.005271	0.01906	0	281	0	0	0	0	1	-360	360;
	3218	5455	0.00906	0.053151	0	319	0	0	0	0	1	-360	360;
	3218	1857	0.00405	0.0153	0	0	0	0	0	0	1	-360	360;
	6806	5410	0.00055	0.003961	0	395	0	0	0	0	1	-360	360;
	6806	5410	0.00055	0.003729	0	338	0	0	0	0	1	-360	360;
	6806	1609	0.00107	0.007789	0	414	0	0	0	0	1	-360	360;
	6806	1609	0.00112	0.00775	0	472	0	0	0	0	1	-360	360;
	6806	1609	0.001031	0.00775	0	0	0	0	0	0	1	-360	360;
	6806	2438	0.00062	0.003219	0	0	0	0	0	0	1	-360	360;
	6806	954	0.00068	0.002659	0	414	0	0	0	0	1	-360	360;
	9140	502	0.00193	0.012479	0	491	0	0	0	0	1	-360	360;
	9140	3013	0.001159	0.007229	0	0	0	0	0	0	1	-360	360;
	9140	2949	0.00137	0.008729	0	395	0	0	0	0	1	-360	360;
	9140	2949	0.001401	0.00855	0	433	0	0	0	0	1	-360	360;
	9140	2949	0.001331	0.008521	0	433	0	0	0	0	1	-360	360;
	7070	2208	0.00031	0.00081	0	281	0	0	0	0	1	-360	360;
	7070	8060	0.000341	0.00119	0	281	0	0	0	0	1	-360	360;
	7070	4484	0.00031	0.001281	0	376	0	0	0	0	1	-360	360;
	7070	4239	0.00094	0.0035	0	433	0	0	0	0	1	-360	360;
	8535	7047	7e-05	0.000651	0	0	0	0	0	0	1	-360	360;
	5477	3657	0.00019	0.00124	0	491	0	0	0	0	1	-360	360;
	5477	2928	0.000331	0.002539	0	433	0	0	0	0	1	-360	360;
	3145	2918	0.0016	0.02113	0	0	0	0	0	0	1	-360	360;
	7770	5317	0.000909	0.004711	0	357	0	0	0	0	1	-360	360;
	7770	5317	0.00088	0.004969	0	319	0	0	0	0	1	-360	360;
	3	8676	0.001841	0.0088	0	0	0	0	0	0	1	-360	360;
	3	8189	0.01013	0.04238	0	395	0	0	0	0	1	-360	360;
	7267	2938	0.00059	0.00802	0	723	0	0	0	0	1	-360	360;
	7267	2938	0.00054	0.0079	0	657	0	0	0	0	1	-360	360;
	7267	3918	0.00171	0.02128	0	986	0	0	0	0	1	-360	360;
	7267	3918	0.00174	0.02084	0	854	0	0	0	0	1	-360	360;
	7267	6581	0.00086	0.01433	0	1151	0	0	0	0	1	-360	360;
	7267	6581	0.00087	0.01404	0	723	0	0	0	0	1	-360	360;
	7267	3830	0.00126	0.01447	0	1118	0	0	0	0	1	-360	360;
	7267	5648	0.00122	0.01509	0	1052	0	0	0	0	1	-360	360;
	7267	5648	0.00154	0.019	0	1282	0	0	0	0	1	-360	360;
	4918	4907	0.00474	0.03564	0	472	0	0	0	0	1	-360	360;
	4918	5957	0.006	0.047849	0	433	0	0	0	0	1	-360	360;
	4918	5233	0.008039	0.03375	0	281	0	0	0	0	1	-360	360;
	6854	7309	2.1e-05	0.000211	0	510	0	0	0	0	1	-360	360;
	8311	3680	0.00045	0.002341	0	0	0	0	0	0	1	-360	360;
	2770	8804	0.000669	0.004281	0	433	0	0	0	0	1	-360	360;
	9241	8804	0.00068	0.00419	0	0	0	0	0	0	1	-360	360;
	3083	7124	0.000729	0.00431	0	510	0	0	0	0	1	-360	360;
	3083	7124	0.000591	0.004651	0	453	0	0	0	0	1	-360	360;
	3083	8873	0.00099	0.005081	0	338	0	0	0	0	1	-360	360;
	1566	7582	0.00212	0.013031	0	0	0	0	0	0	1	-360	360;
	1566	3541	0.002159	0.01275	0	0	0	0	0	0	1	-360	360;
	8683	3698	0.00025	0.00095	0	0	0	0	0	0	1	-360	360;
	8683	3698	0.000271	0.00089	0	0	0	0	0	0	1	-360	360;
	8683	2801	0.001961	0.012409	0	281	0	0	0	0	1	-360	360;
	8683	8328	0.00044	0.00207	0	472	0	0	0	0	1	-360	360;
	8683	8511	0.000469	0.00143	0	529	0	0	0	0	1	-360	360;
	8683	6570	0.001271	0.01406	0	529	0	0	0	0	1	-360	360;
	8683	6570	0.00124	0.013021	0	529	0	0	0	0	1	-360	360;
	8683	8975	0.000409	0.00239	0	0	0	0	0	0	1	-360	360;
	8683	216	0.00094	0.005	0	338	0	0	0	0	1	-360	360;
	8683	2961	0.000219	0.0007	0	529	0	0	0	0	1	-360	360;
	8683	6773	6e-05	0.00025	0	0	0	0	0	0	1	-360	360;
	8683	5278	5e-05	0.000229	0	0	0	0	0	0	1	-360	360;
	3768	2751	0.00019	0.000909	0	0	0	0	0	0	1	-360	360;
	2276	5546	0.001711	0.00782	0	0	0	0	0	0	1	-360	360;
	2276	6816	0.002651	0.01176	0	319	0	0	0	0	1	-360	360;
	7766	444	0.001409	0.010599	0	0	0	0	0	0	1	-360	360;
	7766	5334	0.001419	0.010219	0	0	0	0	0	0	1	-360	360;
	839	3918	0.00013	0.00212	0	756	0	0	0	0	1	-360	360;
	839	809	0.00026	0.00313	0	1282	0	0	0	0	1	-360	360;
	4506	2286	0.0047	0.0318	0	376	0	0	0	0	1	-360	360;
	4506	6802	0.00136	0.00768	0	338	0	0	0	0	1	-360	360;
	4506	3520	0.005669	0.027789	0	300	0	0	0	0	1	-360	360;
	7273	6271	0.004409	0.033091	0	281	0	0	0	0	1	-360	360;
	7273	1435	0.01056	0.044469	0	0	0	0	0	0	1	-360	360;
	7273	7148	0.01963	0.07736	0	338	0	0	0	0	1	-360	360;
	7273	7437	0.01157	0.04694	0	357	0	0	0	0	1	-360	360;
	6104	871	0.000479	0.00255	0	357	0	0	0	0	1	-360	360;
	7580	871	0.000469	0.00236	0	0	0	0	0	0	1	-360	360;
	1188	1539	0.01237	0.067039	0	281	0	0	0	0	1	-360	360;
	1188	8515	0.0132	0.056229	0	281	0	0	0	0	1	-360	360;
	3971	4084	0.00026	0.001711	0	0	0	0	0	0	1	-360	360;
	3971	4084	0.000271	0.001669	0	0	0	0	0	0	1	-360	360;
	6845	5340	0.000211	0.000401	0	0	0	0	0	0	1	-360	360;
	8294	5814	0.00105	0.005159	0	0	0	0	0	0	1	-360	360;
	7115	6807	0.00212	0.00931	0	319	0	0	0	0	1	-360	360;
	619	6791	0.000151	0.00068	0	0	0	0	0	0	1	-360	360;
	5418	4908	0.00019	0.00082	0	0	0	0	0	0	1	-360	360;
	5120	6816	0.003219	0.01405	0	300	0	0	0	0	1	-360	360;
	6692	9158	0.007271	0.033159	0	281	0	0	0	0	1	-360	360;
	6692	8035	0.006659	0.03132	0	300	0	0	0	0	1	-360	360;
	3445	2848	0.00057	0.00668	0	1480	0	0	0	0	1	-360	360;
	3445	7056	0.00057	0.00758	0	1414	0	0	0	0	1	-360	360;
	3445	7056	0.00039	0.00607	0	821	0	0	0	0	1	-360	360;
	3445	1642	0.00155	0.0179	0	0	0	0	0	0	1	-360	360;
	3445	4594	0.00036	0.00485	0	1545	0	0	0	0	1	-360	360;
	6114	1692	0.000581	0.004229	0	338	0	0	0	0	1	-360	360;
	6114	8190	0.00056	0.00431	0	0	0	0	0	0	1	-360	360;
	6114	5836	0.003099	0.020531	0	472	0	0	0	0	1	-360	360;
	6114	3191	0.00118	0.016919	0	510	0	0	0	0	1	-360	360;
	6114	5300	0.00188	0.01407	0	0	0	0	0	0	1	-360	360;
	6114	4889	0.00545	0.04082	0	357	0	0	0	0	1	-360	360;
	7264	444	0.00214	0.01639	0	453	0	0	0	0	1	-360	360;
	7264	4747	0.002479	0.01549	0	0	0	0	0	0	1	-360	360;
	7982	9011	0.000159	0.001229	0	472	0	0	0	0	1	-360	360;
	7982	6675	0.00026	0.00162	0	453	0	0	0	0	1	-360	360;
	7982	6909	0.000159	0.00113	0	376	0	0	0	0	1	-360	360;
	7982	8835	0.000461	0.003219	0	472	0	0	0	0	1	-360	360;
	7982	4186	0.000159	0.001229	0	0	0	0	0	0	1	-360	360;
	1604	502	0.000229	0.00139	0	376	0	0	0	0	1	-360	360;
	2842	502	0.0002	0.00156	0	338	0	0	0	0	1	-360	360;
	1354	513	0.00145	0.00824	0	281	0	0	0	0	1	-360	360;
	1354	502	0.00069	0.005	0	491	0	0	0	0	1	-360	360;
	891	3697	0.00036	0.00455	0	1809	0	0	0	0	1	-360	360;
	891	3697	0.00036	0.00453	0	1776	0	0	0	0	1	-360	360;
	8312	7541	0.00066	0.00997	0	1348	0	0	0	0	1	-360	360;
	8312	7541	0.00064	0.01016	0	986	0	0	0	0	1	-360	360;
	6306	4826	9.9e-05	0.00044	0	529	0	0	0	0	1	-360	360;
	4826	4418	6e-05	0.000469	0	0	0	0	0	0	1	-360	360;
	4826	1605	3.9e-05	0.000341	0	0	0	0	0	0	1	-360	360;
	4826	6952	0.002531	0.01943	0	433	0	0	0	0	1	-360	360;
	4826	7124	0.00945	0.06418	0	414	0	0	0	0	1	-360	360;
	9128	217	0.000591	0.002831	0	357	0	0	0	0	1	-360	360;
	7014	217	0.00087	0.00387	0	0	0	0	0	0	1	-360	360;
	720	5350	0.0015	0.01785	0	1183	0	0	0	0	1	-360	360;
	720	5350	0.00147	0.0182	0	1151	0	0	0	0	1	-360	360;
	1711	639	0.00151	0.0113	0	376	0	0	0	0	1	-360	360;
	1711	639	0.001479	0.011521	0	395	0	0	0	0	1	-360	360;
	1711	9130	0.00174	0.010341	0	414	0	0	0	0	1	-360	360;
	1711	9130	0.001659	0.01206	0	453	0	0	0	0	1	-360	360;
	8209	1998	0.00069	0.000479	0	0	0	0	0	0	1	-360	360;
	8209	7974	0.00044	0.00139	0	0	0	0	0	0	1	-360	360;
	1035	8180	0.002039	0.01601	0	0	0	0	0	0	1	-360	360;
	1035	7466	0.01349	0.05662	0	357	0	0	0	0	1	-360	360;
	1035	305	0.00082	0.00438	0	300	0	0	0	0	1	-360	360;
	1035	305	0.0008	0.004461	0	281	0	0	0	0	1	-360	360;
	1035	5106	0.001091	0.00888	0	453	0	0	0	0	1	-360	360;
	1035	5106	0.00111	0.007039	0	491	0	0	0	0	1	-360	360;
	2527	8466	0.000849	0.00537	0	433	0	0	0	0	1	-360	360;
	8669	8893	0.00061	0.00275	0	357	0	0	0	0	1	-360	360;
	6206	7691	0.00141	0.0188	0	1315	0	0	0	0	1	-360	360;
	6206	6624	0.00121	0.01427	0	1545	0	0	0	0	1	-360	360;
	3240	8158	0.001711	0.007159	0	338	0	0	0	0	1	-360	360;
	3240	8795	0.00545	0.02381	0	300	0	0	0	0	1	-360	360;
	3240	4143	0.000771	0.004331	0	0	0	0	0	0	1	-360	360;
	3240	2797	0.00414	0.03282	0	453	0	0	0	0	1	-360	360;
	3240	6331	0.00264	0.011521	0	338	0	0	0	0	1	-360	360;
	3240	6384	0.00261	0.011	0	376	0	0	0	0	1	-360	360;
	3240	7523	0.00239	0.00919	0	319	0	0	0	0	1	-360	360;
	3240	2889	0.000539	0.00226	0	0	0	0	0	0	1	-360	360;
	3240	6887	0.00325	0.017979	0	472	0	0	0	0	1	-360	360;
	1249	4541	0.00076	0.0071	0	1216	0	0	0	0	1	-360	360;
	819	7523	0.000651	0.00343	0	0	0	0	0	0	1	-360	360;
	2304	4353	0.001219	0.00812	0	338	0	0	0	0	1	-360	360;
	2304	4816	0.00243	0.01439	0	491	0	0	0	0	1	-360	360;
	9174	6246	0.0006	0.00839	0	1151	0	0	0	0	1	-360	360;
	9174	6246	0.00059	0.00856	0	1085	0	0	0	0	1	-360	360;
	9174	5658	0.00087	0.01328	0	1677	0	0	0	0	1	-360	360;
	9174	5658	0.00073	0.01595	0	1249	0	0	0	0	1	-360	360;
	2924	3645	0.00163	0.002039	0	0	0	0	0	0	1	-360	360;
	2924	3645	0.001659	0.002	0	0	0	0	0	0	1	-360	360;
	59	5764	0.00339	0.019669	0	472	0	0	0	0	1	-360	360;
	59	2360	0.00239	0.017281	0	529	0	0	0	0	1	-360	360;
	6510	9033	0.00032	0.00112	0	395	0	0	0	0	1	-360	360;
	6510	5918	0.0018	0.01401	0	395	0	0	0	0	1	-360	360;
	6510	5918	0.00212	0.012469	0	433	0	0	0	0	1	-360	360;
	1275	2850	0.00063	0.00411	0	0	0	0	0	0	1	-360	360;
	6426	4110	0.009729	0.03505	0	0	0	0	0	0	1	-360	360;
	6426	3558	0.01007	0.03056	0	0	0	0	0	0	1	-360	360;
	1758	2816	0.00042	0.00623	0	0	0	0	0	0	1	-360	360;
	1758	352	0.00036	0.00722	0	1743	0	0	0	0	1	-360	360;
	1758	6581	0.00053	0.0078	0	854	0	0	0	0	1	-360	360;
	1758	1923	0.00109	0.0173	0	723	0	0	0	0	1	-360	360;
	1758	5837	0.00071	0.01144	0	0	0	0	0	0	1	-360	360;
	1758	3830	0.00157	0.01884	0	1151	0	0	0	0	1	-360	360;
	1758	3577	0.00032	0.00384	0	756	0	0	0	0	1	-360	360;
	8030	1794	0.00041	0.00649	0	1447	0	0	0	0	1	-360	360;
	8030	7808	0.0004	0.00658	0	0	0	0	0	0	1	-360	360;
	8030	6581	0.00054	0.00763	0	854	0	0	0	0	1	-360	360;
	8030	1923	0.00107	0.01602	0	821	0	0	0	0	1	-360	360;
	8030	5837	0.0007	0.01165	0	0	0	0	0	0	1	-360	360;
	8030	5648	0.0016	0.01844	0	1249	0	0	0	0	1	-360	360;
	8030	3577	0.00032	0.00376	0	854	0	0	0	0	1	-360	360;
	4615	171	0.00212	0.01207	0	395	0	0	0	0	1	-360	360;
	3137	2341	5e-05	0.00037	0	0	0	0	0	0	1	-360	360;
	6271	8787	0.00355	0.024651	0	319	0	0	0	0	1	-360	360;
	6252	9137	0.00675	0.04145	0	433	0	0	0	0	1	-360	360;
	5533	7133	0.00195	0.014341	0	338	0	0	0	0	1	-360	360;
	8361	4355	0.00018	0.00043	0	281	0	0	0	0	1	-360	360;
	8361	2928	0.00037	0.000969	0	357	0	0	0	0	1	-360	360;
	609	1341	0.00257	0.017169	0	376	0	0	0	0	1	-360	360;
	609	5419	0.00076	0.00399	0	395	0	0	0	0	1	-360	360;
	609	5419	0.00075	0.00406	0	395	0	0	0	0	1	-360	360;
	609	2079	0.003901	0.023961	0	472	0	0	0	0	1	-360	360;
	609	772	0.00249	0.019211	0	357	0	0	0	0	1	-360	360;
	3654	2928	0.00025	0.001159	0	376	0	0	0	0	1	-360	360;
	1355	7282	0.00195	0.02924	0	0	0	0	0	0	1	-360	360;
	1355	2197	0.00152	0.01527	0	1381	0	0	0	0	1	-360	360;
	1355	7640	0.00149	0.01558	0	1513	0	0	0	0	1	-360	360;
	1355	1081	0.00253	0.0276	0	1282	0	0	0	0	1	-360	360;
	1355	1081	0.00263	0.02648	0	1480	0	0	0	0	1	-360	360;
	1355	2359	0.00077	0.0092	0	1315	0	0	0	0	1	-360	360;
	1355	4196	0.00112	0.01705	0	657	0	0	0	0	1	-360	360;
	5856	8546	0.00226	0.01101	0	0	0	0	0	0	1	-360	360;
	5856	5987	0.004531	0.029039	0	395	0	0	0	0	1	-360	360;
	5856	8005	0.002081	0.030409	0	0	0	0	0	0	1	-360	360;
	5856	3925	0.0053	0.03257	0	338	0	0	0	0	1	-360	360;
	5856	3324	0.00513	0.041901	0	453	0	0	0	0	1	-360	360;
	5856	4748	0.001461	0.01056	0	357	0	0	0	0	1	-360	360;
	8265	4816	0.0005	0.002591	0	0	0	0	0	0	1	-360	360;
	6151	6313	0.00055	0.0033	0	357	0	0	0	0	1	-360	360;
	4005	6532	0.002289	0.01763	0	376	0	0	0	0	1	-360	360;
	4005	2177	0.0075	0.043531	0	357	0	0	0	0	1	-360	360;
	5723	2079	0.00138	0.019961	0	0	0	0	0	0	1	-360	360;
	5723	789	0.00126	0.018729	0	491	0	0	0	0	1	-360	360;
	2327	9189	0.00038	0.001331	0	395	0	0	0	0	1	-360	360;
	2327	3400	0.00019	0.001461	0	0	0	0	0	0	1	-360	360;
	207	4747	0.002151	0.013401	0	376	0	0	0	0	1	-360	360;
	207	7791	0.000789	0.005669	0	472	0	0	0	0	1	-360	360;
	3205	6735	3e-05	0.00042	0	657	0	0	0	0	1	-360	360;
	3205	666	0.00079	0.00899	0	1282	0	0	0	0	1	-360	360;
	3205	666	0.0008	0.00879	0	1447	0	0	0	0	1	-360	360;
	3205	4435	0.00334	0.03393	0	1282	0	0	0	0	1	-360	360;
	3205	1644	0.00063	0.00789	0	0	0	0	0	0	1	-360	360;
	3205	3483	0.0006	0.00842	0	1216	0	0	0	0	1	-360	360;
	4683	7989	0.00011	0.00125	0	0	0	0	0	0	1	-360	360;
	4683	6516	0.00012	0.001091	0	0	0	0	0	0	1	-360	360;
	4683	7697	9.9e-05	0.00125	0	0	0	0	0	0	1	-360	360;
	4683	416	0.00294	0.02443	0	529	0	0	0	0	1	-360	360;
	4683	2847	0.00144	0.011151	0	395	0	0	0	0	1	-360	360;
	4683	2563	0.001831	0.010849	0	491	0	0	0	0	1	-360	360;
	4683	2563	0.001599	0.01055	0	472	0	0	0	0	1	-360	360;
	4683	2995	0.00144	0.019591	0	281	0	0	0	0	1	-360	360;
	5691	1001	0.0133	0.09551	0	510	0	0	0	0	1	-360	360;
	5691	4426	0.009591	0.073961	0	414	0	0	0	0	1	-360	360;
	5350	2721	0.0007	0.00671	0	1447	0	0	0	0	1	-360	360;
	5350	2721	0.00072	0.00658	0	1480	0	0	0	0	1	-360	360;
	2361	4936	0.001901	0.00964	0	529	0	0	0	0	1	-360	360;
	2361	5179	0.001711	0.01189	0	510	0	0	0	0	1	-360	360;
	2044	4748	0.00025	0.00187	0	0	0	0	0	0	1	-360	360;
	4197	5522	0.00043	0.002211	0	0	0	0	0	0	1	-360	360;
	5699	8670	0.000401	0.00243	0	0	0	0	0	0	1	-360	360;
	7809	5110	0.00336	0.017	0	357	0	0	0	0	1	-360	360;
	7222	4816	0.005521	0.031031	0	0	0	0	0	0	1	-360	360;
	4402	3817	9.9e-05	0.00026	0	0	0	0	0	0	1	-360	360;
	6857	2732	0.00045	0.00718	0	986	0	0	0	0	1	-360	360;
	6857	4550	0.00044	0.00665	0	821	0	0	0	0	1	-360	360;
	6857	4339	0.00103	0.01516	0	1183	0	0	0	0	1	-360	360;
	6857	7513	0.00101	0.01546	0	854	0	0	0	0	1	-360	360;
	5016	26	0.00045	0.00608	0	1414	0	0	0	0	1	-360	360;
	5016	9213	0.0004	0.00501	0	1249	0	0	0	0	1	-360	360;
	6112	9002	0.001729	0.01337	0	491	0	0	0	0	1	-360	360;
	6112	6231	0.00264	0.01489	0	414	0	0	0	0	1	-360	360;
	6112	883	0.002341	0.018961	0	414	0	0	0	0	1	-360	360;
	6112	5738	0.001841	0.01363	0	376	0	0	0	0	1	-360	360;
	8860	7700	0.001219	0.00949	0	395	0	0	0	0	1	-360	360;
	8860	9002	0.002271	0.01151	0	395	0	0	0	0	1	-360	360;
	8860	5990	0.010781	0.05775	0	0	0	0	0	0	1	-360	360;
	8860	6450	0.00251	0.013919	0	300	0	0	0	0	1	-360	360;
	8860	7726	0.00101	0.007331	0	414	0	0	0	0	1	-360	360;
	2489	4110	0.00061	0.003909	0	529	0	0	0	0	1	-360	360;
	3985	6376	0.004289	0.019331	0	0	0	0	0	0	1	-360	360;
	3985	4110	0.0008	0.004401	0	529	0	0	0	0	1	-360	360;
	3985	4829	0.008099	0.03519	0	510	0	0	0	0	1	-360	360;
	3985	7857	0.00063	0.004591	0	414	0	0	0	0	1	-360	360;
	2806	7857	0.000651	0.00456	0	0	0	0	0	0	1	-360	360;
	2489	9012	0.00114	0.00581	0	0	0	0	0	0	1	-360	360;
	2489	2231	0.001211	0.00538	0	0	0	0	0	0	1	-360	360;
	658	3793	0.00418	0.030159	0	472	0	0	0	0	1	-360	360;
	658	8497	0.00324	0.02413	0	510	0	0	0	0	1	-360	360;
	658	8497	0.003169	0.024591	0	529	0	0	0	0	1	-360	360;
	3502	9002	0.000219	0.00088	0	395	0	0	0	0	1	-360	360;
	3502	9002	0.000289	0.00095	0	414	0	0	0	0	1	-360	360;
	6313	44	0.0035	0.022211	0	453	0	0	0	0	1	-360	360;
	6313	7582	0.00313	0.02256	0	0	0	0	0	0	1	-360	360;
	5944	8808	0.000831	0.00162	0	0	0	0	0	0	1	-360	360;
	3183	8515	0.002099	0.01068	0	300	0	0	0	0	1	-360	360;
	1005	8515	0.00224	0.00976	0	0	0	0	0	0	1	-360	360;
	8787	1448	0.00169	0.01231	0	529	0	0	0	0	1	-360	360;
	6636	9051	0.00011	0.00076	0	0	0	0	0	0	1	-360	360;
	9109	9051	0.00011	0.00068	0	0	0	0	0	0	1	-360	360;
	6961	6376	0.005479	0.021909	0	0	0	0	0	0	1	-360	360;
	6961	8818	0.003331	0.017039	0	338	0	0	0	0	1	-360	360;
	5077	4353	0.000479	0.001271	0	0	0	0	0	0	1	-360	360;
	7471	5241	0.00109	0.01629	0	723	0	0	0	0	1	-360	360;
	7471	8200	4e-05	0.00063	0	1710	0	0	0	0	1	-360	360;
	549	6290	0.0053	0.024531	0	472	0	0	0	0	1	-360	360;
	549	2291	0.004659	0.021919	0	319	0	0	0	0	1	-360	360;
	549	3656	0.001659	0.01043	0	472	0	0	0	0	1	-360	360;
	5002	4144	0.006961	0.053539	0	395	0	0	0	0	1	-360	360;
	1888	5441	0.00036	0.00288	0	529	0	0	0	0	1	-360	360;
	1888	5441	0.00038	0.003	0	529	0	0	0	0	1	-360	360;
	5490	4970	0.00048	0.00643	0	0	0	0	0	0	1	-360	360;
	5490	4970	0.00051	0.0066	0	0	0	0	0	0	1	-360	360;
	5490	4970	0.00046	0.00729	0	1776	0	0	0	0	1	-360	360;
	5490	5461	0.0006	0.00868	0	1710	0	0	0	0	1	-360	360;
	5490	7778	0.00068	0.01076	0	1842	0	0	0	0	1	-360	360;
	5490	6897	0.00066	0.01096	0	1809	0	0	0	0	1	-360	360;
	3075	4562	0.00101	0.005099	0	300	0	0	0	0	1	-360	360;
	564	8475	0.003901	0.02437	0	0	0	0	0	0	1	-360	360;
	564	2252	0.00276	0.016461	0	453	0	0	0	0	1	-360	360;
	564	7903	0.00588	0.047081	0	0	0	0	0	0	1	-360	360;
	4331	1857	3.9e-05	0.000219	0	0	0	0	0	0	1	-360	360;
	2877	3674	0.00131	0.00538	0	0	0	0	0	0	1	-360	360;
	2854	1081	0.00679	0.04679	0	920	0	0	0	0	1	-360	360;
	2854	8267	0.00138	0.01466	0	986	0	0	0	0	1	-360	360;
	9112	9176	0.003531	0.01724	0	0	0	0	0	0	1	-360	360;
	9112	1422	0.01324	0.06936	0	376	0	0	0	0	1	-360	360;
	9112	8515	0.00363	0.01743	0	338	0	0	0	0	1	-360	360;
	9112	8515	0.00257	0.01662	0	433	0	0	0	0	1	-360	360;
	9112	5051	0.01399	0.056289	0	319	0	0	0	0	1	-360	360;
	9112	1051	0.00343	0.01751	0	357	0	0	0	0	1	-360	360;
	9112	1397	0.00301	0.015409	0	0	0	0	0	0	1	-360	360;
	9112	1584	0.00857	0.033091	0	376	0	0	0	0	1	-360	360;
	6168	6888	0.003341	0.01689	0	0	0	0	0	0	1	-360	360;
	7653	3019	3.9e-05	0.000219	0	0	0	0	0	0	1	-360	360;
	7341	7098	0.00087	0.00714	0	357	0	0	0	0	1	-360	360;
	4701	8267	0.00011	0.001	0	657	0	0	0	0	1	-360	360;
	1672	5365	0.000219	0.00101	0	0	0	0	0	0	1	-360	360;
	3880	5365	0.000219	0.001031	0	0	0	0	0	0	1	-360	360;
	6648	5365	0.000271	0.00101	0	0	0	0	0	0	1	-360	360;
	3609	5365	0.000271	0.000979	0	0	0	0	0	0	1	-360	360;
	2782	5365	0.000289	0.00107	0	0	0	0	0	0	1	-360	360;
	1844	5365	0.000289	0.00105	0	0	0	0	0	0	1	-360	360;
	4353	6053	0.001591	0.012289	0	281	0	0	0	0	1	-360	360;
	2166	5237	0.000271	0.001901	0	395	0	0	0	0	1	-360	360;
	2166	2886	0.00186	0.010711	0	414	0	0	0	0	1	-360	360;
	1043	3513	1e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	4049	883	0.000169	0.000271	0	0	0	0	0	0	1	-360	360;
	4520	883	0.000169	0.00031	0	0	0	0	0	0	1	-360	360;
	1704	5814	0.00526	0.03787	0	453	0	0	0	0	1	-360	360;
	8107	1486	0.000229	0.000599	0	0	0	0	0	0	1	-360	360;
	5664	5522	0.00089	0.004091	0	338	0	0	0	0	1	-360	360;
	4482	8670	0.00093	0.004	0	300	0	0	0	0	1	-360	360;
	8795	7523	0.00289	0.013099	0	300	0	0	0	0	1	-360	360;
	2337	6430	0.002599	0.016479	0	433	0	0	0	0	1	-360	360;
	455	7775	0.00369	0.02712	0	453	0	0	0	0	1	-360	360;
	455	3857	0.0062	0.04864	0	529	0	0	0	0	1	-360	360;
	455	333	0.008021	0.059031	0	529	0	0	0	0	1	-360	360;
	455	6940	0.004979	0.0295	0	472	0	0	0	0	1	-360	360;
	8225	3656	0.001159	0.005281	0	0	0	0	0	0	1	-360	360;
	5413	7752	0.00361	0.01824	0	0	0	0	0	0	1	-360	360;
	5413	1940	0.001521	0.01075	0	0	0	0	0	0	1	-360	360;
	5413	4234	0.00868	0.041711	0	0	0	0	0	0	1	-360	360;
	2591	6828	0.01118	0.05986	0	0	0	0	0	0	1	-360	360;
	2591	7571	0.003229	0.02543	0	491	0	0	0	0	1	-360	360;
	2591	7571	0.00218	0.01869	0	0	0	0	0	0	1	-360	360;
	2591	7571	0.001901	0.021909	0	472	0	0	0	0	1	-360	360;
	2591	7994	0.00974	0.043969	0	300	0	0	0	0	1	-360	360;
	2591	7226	0.003461	0.01814	0	357	0	0	0	0	1	-360	360;
	2591	1554	0.00487	0.027669	0	319	0	0	0	0	1	-360	360;
	9011	1156	0.000651	0.003961	0	510	0	0	0	0	1	-360	360;
	8689	5502	0.00169	0.00887	0	300	0	0	0	0	1	-360	360;
	3085	8873	0.001711	0.01336	0	414	0	0	0	0	1	-360	360;
	1568	980	0.00031	0.000539	0	300	0	0	0	0	1	-360	360;
	3204	3331	0.0027	0.01157	0	0	0	0	0	0	1	-360	360;
	3557	639	0.000909	0.006901	0	414	0	0	0	0	1	-360	360;
	3557	4936	0.001729	0.01214	0	414	0	0	0	0	1	-360	360;
	416	3492	0.00614	0.03106	0	338	0	0	0	0	1	-360	360;
	416	7937	0.00418	0.02388	0	414	0	0	0	0	1	-360	360;
	3255	1262	0.00238	0.017771	0	0	0	0	0	0	1	-360	360;
	3255	1803	8.1e-05	0.00063	0	0	0	0	0	0	1	-360	360;
	7373	367	0.001599	0.01618	0	0	0	0	0	0	1	-360	360;
	4251	2132	0.00532	0.026961	0	319	0	0	0	0	1	-360	360;
	4251	2132	0.00361	0.02814	0	453	0	0	0	0	1	-360	360;
	4251	5567	0.002469	0.01025	0	433	0	0	0	0	1	-360	360;
	6376	1808	0.005229	0.02675	0	300	0	0	0	0	1	-360	360;
	6376	1808	0.00512	0.02726	0	319	0	0	0	0	1	-360	360;
	6376	7857	0.00355	0.01694	0	300	0	0	0	0	1	-360	360;
	9185	7289	0.00139	0.007659	0	357	0	0	0	0	1	-360	360;
	9185	3999	0.000169	0.000669	0	414	0	0	0	0	1	-360	360;
	8653	2303	0.00111	0.00712	0	376	0	0	0	0	1	-360	360;
	908	6474	0.00031	0.001901	0	433	0	0	0	0	1	-360	360;
	7030	6474	0.000281	0.002091	0	0	0	0	0	0	1	-360	360;
	4205	6556	0.000281	0.001651	0	0	0	0	0	0	1	-360	360;
	7325	6556	0.000271	0.00169	0	0	0	0	0	0	1	-360	360;
	5857	3391	0.001521	0.00843	0	281	0	0	0	0	1	-360	360;
	5857	3760	0.000841	0.0062	0	433	0	0	0	0	1	-360	360;
	3707	8765	0.00019	0.00126	0	414	0	0	0	0	1	-360	360;
	900	6619	0.000219	0.00143	0	0	0	0	0	0	1	-360	360;
	2918	8109	0.00121	0.01796	0	1776	0	0	0	0	1	-360	360;
	2918	8109	0.0014	0.01512	0	0	0	0	0	0	1	-360	360;
	6532	2177	0.008409	0.063651	0	0	0	0	0	0	1	-360	360;
	6532	5317	0.00918	0.07344	0	510	0	0	0	0	1	-360	360;
	6532	608	0.004219	0.0345	0	414	0	0	0	0	1	-360	360;
	6532	608	0.0042	0.031651	0	357	0	0	0	0	1	-360	360;
	4494	8763	0.00121	0.01381	0	1447	0	0	0	0	1	-360	360;
	4494	7056	0.00107	0.01326	0	1480	0	0	0	0	1	-360	360;
	6828	5814	0.003401	0.02607	0	0	0	0	0	0	1	-360	360;
	3670	21	0.000289	0.00138	0	319	0	0	0	0	1	-360	360;
	5137	7702	0.001031	0.006039	0	0	0	0	0	0	1	-360	360;
	1334	7702	0.00105	0.005909	0	0	0	0	0	0	1	-360	360;
	8448	7624	0.001091	0.007901	0	319	0	0	0	0	1	-360	360;
	8448	1129	0.00332	0.024021	0	453	0	0	0	0	1	-360	360;
	8448	3070	0.00344	0.021419	0	300	0	0	0	0	1	-360	360;
	8448	306	0.00095	0.010419	0	414	0	0	0	0	1	-360	360;
	8448	306	0.00093	0.01062	0	376	0	0	0	0	1	-360	360;
	8448	7491	0.00474	0.025409	0	0	0	0	0	0	1	-360	360;
	4685	2089	0.000271	0.000841	0	281	0	0	0	0	1	-360	360;
	7485	6253	0.000271	0.00119	0	281	0	0	0	0	1	-360	360;
	7485	1562	0.000289	0.00076	0	0	0	0	0	0	1	-360	360;
	7776	1644	0.00031	0.00454	0	1842	0	0	0	0	1	-360	360;
	5971	3483	0.00032	0.0044	0	0	0	0	0	0	1	-360	360;
	2786	1644	0.00028	0.00494	0	0	0	0	0	0	1	-360	360;
	972	3483	0.00028	0.00508	0	0	0	0	0	0	1	-360	360;
	8906	923	0.000531	0.00093	0	0	0	0	0	0	1	-360	360;
	8906	4914	0.0003	0.001031	0	395	0	0	0	0	1	-360	360;
	2656	9101	0.00139	0.02163	0	657	0	0	0	0	1	-360	360;
	2656	7164	0.00164	0.02321	0	1776	0	0	0	0	1	-360	360;
	3519	9101	0.00136	0.02205	0	657	0	0	0	0	1	-360	360;
	3519	7164	0.0016	0.02367	0	657	0	0	0	0	1	-360	360;
	1775	9137	0.002091	0.01174	0	281	0	0	0	0	1	-360	360;
	1775	9137	0.001719	0.00961	0	453	0	0	0	0	1	-360	360;
	5641	1026	0.012341	0.076781	0	319	0	0	0	0	1	-360	360;
	5641	9137	0.00176	0.009419	0	491	0	0	0	0	1	-360	360;
	1096	2889	0.00031	0.00182	0	0	0	0	0	0	1	-360	360;
	7380	2732	0.00386	0.02348	0	1183	0	0	0	0	1	-360	360;
	7380	4787	0.00017	0.00104	0	1052	0	0	0	0	1	-360	360;
	3543	5308	0.00061	0.00808	0	1513	0	0	0	0	1	-360	360;
	3543	3830	0.00059	0.0067	0	0	0	0	0	0	1	-360	360;
	2372	1027	0.00059	0.00748	0	1381	0	0	0	0	1	-360	360;
	2372	5648	0.00061	0.0063	0	0	0	0	0	0	1	-360	360;
	367	1172	0.001031	0.008789	0	529	0	0	0	0	1	-360	360;
	367	9191	0.00051	0.006169	0	529	0	0	0	0	1	-360	360;
	367	8722	0.0005	0.00619	0	529	0	0	0	0	1	-360	360;
	1833	3610	0.00049	0.00376	0	472	0	0	0	0	1	-360	360;
	1833	3610	0.000539	0.003229	0	433	0	0	0	0	1	-360	360;
	1833	3962	0.001031	0.00882	0	0	0	0	0	0	1	-360	360;
	1833	1838	0.000711	0.00463	0	357	0	0	0	0	1	-360	360;
	1833	2230	0.000711	0.004539	0	0	0	0	0	0	1	-360	360;
	5388	2467	0.00206	0.015341	0	414	0	0	0	0	1	-360	360;
	5388	2467	0.002021	0.01564	0	414	0	0	0	0	1	-360	360;
	6110	8853	0.00094	0.005781	0	510	0	0	0	0	1	-360	360;
	6110	1001	0.002219	0.016969	0	529	0	0	0	0	1	-360	360;
	7583	1808	0.00238	0.01187	0	281	0	0	0	0	1	-360	360;
	7583	2889	0.00707	0.03095	0	281	0	0	0	0	1	-360	360;
	26	6624	0.00256	0.03401	0	1545	0	0	0	0	1	-360	360;
	26	2479	0.00203	0.02416	0	1052	0	0	0	0	1	-360	360;
	26	4598	0.001	0.01459	0	0	0	0	0	0	1	-360	360;
	26	4231	0.00103	0.00987	0	1513	0	0	0	0	1	-360	360;
	26	5144	0.00038	0.00483	0	1348	0	0	0	0	1	-360	360;
	26	5144	0.00041	0.00398	0	1216	0	0	0	0	1	-360	360;
	7691	2479	0.00167	0.01721	0	1282	0	0	0	0	1	-360	360;
	7691	2129	0.00098	0.01487	0	887	0	0	0	0	1	-360	360;
	7691	9213	0.00042	0.00342	0	1249	0	0	0	0	1	-360	360;
	7691	9213	0.00019	0.00248	0	1480	0	0	0	0	1	-360	360;
	7691	4231	0.00105	0.00966	0	1809	0	0	0	0	1	-360	360;
	7691	3513	0.00043	0.00473	0	1282	0	0	0	0	1	-360	360;
	7691	3513	0.00041	0.00397	0	1216	0	0	0	0	1	-360	360;
	1876	3906	0.00051	0.00514	0	1118	0	0	0	0	1	-360	360;
	113	3906	0.00053	0.00503	0	1315	0	0	0	0	1	-360	360;
	7752	3758	0.00455	0.022771	0	300	0	0	0	0	1	-360	360;
	7752	3758	0.00557	0.02055	0	281	0	0	0	0	1	-360	360;
	7752	3758	0.00339	0.02026	0	0	0	0	0	0	1	-360	360;
	4952	8976	0.00237	0.01068	0	338	0	0	0	0	1	-360	360;
	4952	8189	0.0037	0.019479	0	281	0	0	0	0	1	-360	360;
	6119	7886	0.01407	0.056849	0	0	0	0	0	0	1	-360	360;
	6119	4544	0.001669	0.015409	0	510	0	0	0	0	1	-360	360;
	6119	4544	0.001711	0.015091	0	0	0	0	0	0	1	-360	360;
	7905	3697	0.00172	0.01619	0	1578	0	0	0	0	1	-360	360;
	7905	7541	0.00029	0.00447	0	887	0	0	0	0	1	-360	360;
	7905	7541	0.00033	0.00387	0	1019	0	0	0	0	1	-360	360;
	7905	4141	0.00079	0.01154	0	887	0	0	0	0	1	-360	360;
	7905	4141	0.00077	0.01177	0	887	0	0	0	0	1	-360	360;
	7905	3022	0.00127	0.01498	0	1447	0	0	0	0	1	-360	360;
	1768	163	0.00214	0.012599	0	395	0	0	0	0	1	-360	360;
	4245	2425	0.000159	0.00125	0	357	0	0	0	0	1	-360	360;
	3698	5297	0.000211	0.000669	0	357	0	0	0	0	1	-360	360;
	750	8328	0.000341	0.001169	0	414	0	0	0	0	1	-360	360;
	750	870	0.00024	0.001469	0	472	0	0	0	0	1	-360	360;
	750	2961	0.00018	0.000669	0	433	0	0	0	0	1	-360	360;
	3499	7988	0.00033	0.00683	0	821	0	0	0	0	1	-360	360;
	3499	2197	0.00121	0.01211	0	1315	0	0	0	0	1	-360	360;
	3499	7640	0.00123	0.01185	0	1348	0	0	0	0	1	-360	360;
	3499	4231	0.00058	0.0097	0	1085	0	0	0	0	1	-360	360;
	3499	5144	0.00103	0.01292	0	1151	0	0	0	0	1	-360	360;
	3499	3513	0.00116	0.01266	0	821	0	0	0	0	1	-360	360;
	7289	3999	0.0013	0.006841	0	357	0	0	0	0	1	-360	360;
	3018	1051	6e-05	0.000349	0	338	0	0	0	0	1	-360	360;
	2702	3435	0.000169	0.0007	0	300	0	0	0	0	1	-360	360;
	9131	2020	0.000169	0.000719	0	0	0	0	0	0	1	-360	360;
	6308	6684	0.000271	0.001031	0	0	0	0	0	0	1	-360	360;
	4831	3680	0.008271	0.036841	0	0	0	0	0	0	1	-360	360;
	5410	2751	0.002969	0.02299	0	319	0	0	0	0	1	-360	360;
	3082	4623	0.00124	0.01273	0	1118	0	0	0	0	1	-360	360;
	3082	4339	0.00106	0.00989	0	1381	0	0	0	0	1	-360	360;
	3082	5288	7e-05	0.00029	0	657	0	0	0	0	1	-360	360;
	3082	5421	9e-05	0.00079	0	657	0	0	0	0	1	-360	360;
	3082	6475	0.00196	0.02049	0	1513	0	0	0	0	1	-360	360;
	3082	8468	0.00158	0.01519	0	1183	0	0	0	0	1	-360	360;
	6224	2732	0.00172	0.01811	0	1513	0	0	0	0	1	-360	360;
	6224	7513	0.00096	0.01121	0	1644	0	0	0	0	1	-360	360;
	6224	6475	0.00147	0.01707	0	0	0	0	0	0	1	-360	360;
	2174	4623	0.00129	0.0129	0	1216	0	0	0	0	1	-360	360;
	2174	4550	0.00179	0.01737	0	1545	0	0	0	0	1	-360	360;
	2174	6475	0.00131	0.01935	0	1677	0	0	0	0	1	-360	360;
	2083	9231	0.003901	0.01614	0	319	0	0	0	0	1	-360	360;
	2083	9231	0.002591	0.014771	0	453	0	0	0	0	1	-360	360;
	2083	9231	0.00224	0.0173	0	376	0	0	0	0	1	-360	360;
	2083	4054	0.00012	0.000849	0	0	0	0	0	0	1	-360	360;
	2083	3481	0.00394	0.03039	0	395	0	0	0	0	1	-360	360;
	2083	3602	0.00189	0.01	0	0	0	0	0	0	1	-360	360;
	2083	2794	0.00032	0.00257	0	0	0	0	0	0	1	-360	360;
	2083	2794	0.00038	0.00224	0	0	0	0	0	0	1	-360	360;
	2083	8467	0.00751	0.03955	0	395	0	0	0	0	1	-360	360;
	2083	1136	0.00086	0.006531	0	0	0	0	0	0	1	-360	360;
	2083	22	0.00087	0.006591	0	0	0	0	0	0	1	-360	360;
	8853	1001	0.003031	0.02276	0	529	0	0	0	0	1	-360	360;
	8853	4950	0.006331	0.03911	0	0	0	0	0	0	1	-360	360;
	8853	4410	0.00605	0.02986	0	281	0	0	0	0	1	-360	360;
	8853	6697	0.003919	0.01612	0	338	0	0	0	0	1	-360	360;
	8853	1159	0.002039	0.0108	0	338	0	0	0	0	1	-360	360;
	8853	7165	0.006289	0.026271	0	414	0	0	0	0	1	-360	360;
	2938	2848	0.00042	0.00547	0	756	0	0	0	0	1	-360	360;
	2938	2848	0.00043	0.00536	0	657	0	0	0	0	1	-360	360;
	2938	7522	0.00147	0.01373	0	1677	0	0	0	0	1	-360	360;
	2938	7522	0.00131	0.01555	0	1447	0	0	0	0	1	-360	360;
	2938	1642	0.00221	0.02217	0	1710	0	0	0	0	1	-360	360;
	2938	809	0.00224	0.0217	0	1710	0	0	0	0	1	-360	360;
	9189	6791	0.00095	0.003289	0	300	0	0	0	0	1	-360	360;
	9189	3400	0.000229	0.00082	0	0	0	0	0	0	1	-360	360;
	9189	7092	0.00068	0.00306	0	0	0	0	0	0	1	-360	360;
	9189	3718	0.00049	0.00256	0	319	0	0	0	0	1	-360	360;
	9189	3701	0.000531	0.0022	0	0	0	0	0	0	1	-360	360;
	9189	7507	0.00136	0.00956	0	414	0	0	0	0	1	-360	360;
	9189	7507	0.0012	0.010729	0	376	0	0	0	0	1	-360	360;
	9189	7507	0.00105	0.00861	0	0	0	0	0	0	1	-360	360;
	9189	2406	0.000591	0.006969	0	0	0	0	0	0	1	-360	360;
	2801	8511	0.00025	0.0008	0	0	0	0	0	0	1	-360	360;
	2801	216	0.00113	0.00632	0	300	0	0	0	0	1	-360	360;
	453	2129	0.0003	0.00275	0	854	0	0	0	0	1	-360	360;
	8691	6952	0.00836	0.051659	0	529	0	0	0	0	1	-360	360;
	8691	6952	0.01338	0.073469	0	491	0	0	0	0	1	-360	360;
	8691	1183	0.0027	0.02089	0	0	0	0	0	0	1	-360	360;
	8691	4410	0.003419	0.02636	0	510	0	0	0	0	1	-360	360;
	8691	8743	0.009531	0.033711	0	357	0	0	0	0	1	-360	360;
	2446	8763	0.0004	0.00707	0	0	0	0	0	0	1	-360	360;
	2627	8763	0.00035	0.00701	0	0	0	0	0	0	1	-360	360;
	4951	6785	0.00194	0.01751	0	433	0	0	0	0	1	-360	360;
	4951	6785	0.00156	0.01936	0	414	0	0	0	0	1	-360	360;
	4623	964	0.00078	0.00941	0	1249	0	0	0	0	1	-360	360;
	4623	964	0.00074	0.00969	0	1480	0	0	0	0	1	-360	360;
	4623	8347	0.00042	0.00532	0	1578	0	0	0	0	1	-360	360;
	4623	8347	0.0004	0.00473	0	1743	0	0	0	0	1	-360	360;
	4623	8347	0.0006	0.00499	0	1447	0	0	0	0	1	-360	360;
	1609	6552	0.00086	0.00638	0	510	0	0	0	0	1	-360	360;
	1609	6552	0.000969	0.00625	0	510	0	0	0	0	1	-360	360;
	1609	8999	0.002781	0.01725	0	357	0	0	0	0	1	-360	360;
	4368	4816	0.005031	0.028081	0	453	0	0	0	0	1	-360	360;
	5481	960	1e-05	0.00021	0	0	0	0	0	0	1	-360	360;
	7913	960	1e-05	0.00025	0	0	0	0	0	0	1	-360	360;
	506	960	2e-05	0.00031	0	657	0	0	0	0	1	-360	360;
	2980	960	3e-05	0.0004	0	657	0	0	0	0	1	-360	360;
	4554	6382	0.000841	0.00474	0	433	0	0	0	0	1	-360	360;
	4554	6382	0.00082	0.00487	0	433	0	0	0	0	1	-360	360;
	6844	1595	0.00186	0.011979	0	0	0	0	0	0	1	-360	360;
	3672	5334	0.00056	0.003021	0	281	0	0	0	0	1	-360	360;
	6219	5334	0.00055	0.003081	0	0	0	0	0	0	1	-360	360;
	8997	6153	0.00114	0.01063	0	0	0	0	0	0	1	-360	360;
	1852	5935	0.003719	0.02295	0	529	0	0	0	0	1	-360	360;
	1852	5695	0.006669	0.04288	0	414	0	0	0	0	1	-360	360;
	2931	5799	0.00243	0.0198	0	0	0	0	0	0	1	-360	360;
	2931	7124	0.00845	0.04339	0	433	0	0	0	0	1	-360	360;
	2598	3674	0.00219	0.015169	0	281	0	0	0	0	1	-360	360;
	2598	2132	0.00357	0.0302	0	319	0	0	0	0	1	-360	360;
	2558	7571	0.005849	0.032419	0	319	0	0	0	0	1	-360	360;
	8834	1129	0.00406	0.01688	0	0	0	0	0	0	1	-360	360;
	7377	3324	0.00113	0.0065	0	0	0	0	0	0	1	-360	360;
	1265	4426	0.0068	0.04613	0	414	0	0	0	0	1	-360	360;
	1265	1965	0.00337	0.02011	0	319	0	0	0	0	1	-360	360;
	8060	7162	0.00011	0.0005	0	0	0	0	0	0	1	-360	360;
	6922	5482	0.003711	0.018521	0	338	0	0	0	0	1	-360	360;
	6922	1295	0.000271	0.00145	0	357	0	0	0	0	1	-360	360;
	8369	4710	0.00326	0.01824	0	0	0	0	0	0	1	-360	360;
	8961	7133	0.00118	0.00712	0	319	0	0	0	0	1	-360	360;
	8961	6888	0.00482	0.02232	0	281	0	0	0	0	1	-360	360;
	4185	7019	0.00013	0.000469	0	281	0	0	0	0	1	-360	360;
	4185	6253	0.000229	0.00075	0	319	0	0	0	0	1	-360	360;
	7895	3391	0.004591	0.035021	0	433	0	0	0	0	1	-360	360;
	7895	1448	0.005469	0.03924	0	510	0	0	0	0	1	-360	360;
	7895	1448	0.00536	0.0403	0	0	0	0	0	0	1	-360	360;
	7895	115	0.00424	0.033031	0	0	0	0	0	0	1	-360	360;
	5589	4306	2e-05	0.00024	0	657	0	0	0	0	1	-360	360;
	5589	1763	1e-05	0.00021	0	657	0	0	0	0	1	-360	360;
	5589	3697	0.00286	0.03653	0	1545	0	0	0	0	1	-360	360;
	5589	3794	0.00117	0.01183	0	1644	0	0	0	0	1	-360	360;
	5589	3613	0.00103	0.01336	0	1381	0	0	0	0	1	-360	360;
	5589	1502	0.00156	0.01969	0	1447	0	0	0	0	1	-360	360;
	5589	6555	0.00449	0.05772	0	1677	0	0	0	0	1	-360	360;
	5589	3608	0.00259	0.03469	0	1513	0	0	0	0	1	-360	360;
	1001	4134	0.000219	0.00118	0	0	0	0	0	0	1	-360	360;
	1001	1979	6e-05	0.00095	0	0	0	0	0	0	1	-360	360;
	1001	516	6e-05	0.000651	0	529	0	0	0	0	1	-360	360;
	1001	3580	8.1e-05	0.000841	0	529	0	0	0	0	1	-360	360;
	1001	3809	0.00274	0.020831	0	414	0	0	0	0	1	-360	360;
	1001	5616	0.004781	0.027729	0	453	0	0	0	0	1	-360	360;
	1001	5616	0.002849	0.02093	0	472	0	0	0	0	1	-360	360;
	1001	6697	0.00513	0.03014	0	414	0	0	0	0	1	-360	360;
	1001	7165	0.00113	0.00651	0	0	0	0	0	0	1	-360	360;
	1001	892	0.00125	0.00707	0	472	0	0	0	0	1	-360	360;
	7438	5334	0.003349	0.01463	0	0	0	0	0	0	1	-360	360;
	7438	1526	0.00245	0.0122	0	395	0	0	0	0	1	-360	360;
	1301	6952	0.00587	0.036521	0	300	0	0	0	0	1	-360	360;
	1301	6952	0.00569	0.03774	0	0	0	0	0	0	1	-360	360;
	1301	6952	0.006031	0.03545	0	529	0	0	0	0	1	-360	360;
	1301	7883	0.000531	0.00301	0	0	0	0	0	0	1	-360	360;
	1301	6430	0.00076	0.006099	0	414	0	0	0	0	1	-360	360;
	778	7209	0.008841	0.028919	0	0	0	0	0	0	1	-360	360;
	778	1910	0.00169	0.007091	0	0	0	0	0	0	1	-360	360;
	778	6820	0.001031	0.00361	0	0	0	0	0	0	1	-360	360;
	778	3071	0.00813	0.030961	0	0	0	0	0	0	1	-360	360;
	5297	6267	0.00013	0.00068	0	433	0	0	0	0	1	-360	360;
	3657	3855	0.000211	0.000729	0	529	0	0	0	0	1	-360	360;
	3657	3855	0.000169	0.000669	0	491	0	0	0	0	1	-360	360;
	3657	2928	0.000669	0.00251	0	395	0	0	0	0	1	-360	360;
	3657	2928	0.000711	0.00238	0	376	0	0	0	0	1	-360	360;
	3657	1172	0.000229	0.00238	0	491	0	0	0	0	1	-360	360;
	3657	9191	0.0007	0.0065	0	453	0	0	0	0	1	-360	360;
	3657	8722	0.00061	0.006031	0	414	0	0	0	0	1	-360	360;
	6472	5362	0.000281	0.0018	0	529	0	0	0	0	1	-360	360;
	6472	3962	0.000211	0.002409	0	0	0	0	0	0	1	-360	360;
	1198	1343	0.00056	0.00773	0	657	0	0	0	0	1	-360	360;
	1198	29	0.0023	0.01581	0	1118	0	0	0	0	1	-360	360;
	7473	3680	0.005039	0.032151	0	0	0	0	0	0	1	-360	360;
	7473	96	0.00364	0.01713	0	357	0	0	0	0	1	-360	360;
	7473	1868	0.002169	0.012409	0	0	0	0	0	0	1	-360	360;
	346	3184	0.00113	0.004719	0	0	0	0	0	0	1	-360	360;
	346	6570	0.00113	0.004979	0	376	0	0	0	0	1	-360	360;
	3184	4728	0.00264	0.01174	0	0	0	0	0	0	1	-360	360;
	2654	7886	0.00494	0.03181	0	376	0	0	0	0	1	-360	360;
	2654	2132	0.003711	0.023961	0	510	0	0	0	0	1	-360	360;
	4144	8874	0.001409	0.01101	0	357	0	0	0	0	1	-360	360;
	4144	1625	0.001169	0.00593	0	281	0	0	0	0	1	-360	360;
	4144	1625	0.00118	0.00587	0	0	0	0	0	0	1	-360	360;
	4144	3786	0.00264	0.012469	0	433	0	0	0	0	1	-360	360;
	4144	3786	0.002349	0.01414	0	395	0	0	0	0	1	-360	360;
	3422	4100	0.00014	0.0007	0	0	0	0	0	0	1	-360	360;
	6624	4419	3e-05	0.00054	0	1776	0	0	0	0	1	-360	360;
	6624	7159	3e-05	0.00055	0	1710	0	0	0	0	1	-360	360;
	6624	3661	4e-05	0.00058	0	1677	0	0	0	0	1	-360	360;
	6624	5486	4e-05	0.00054	0	1776	0	0	0	0	1	-360	360;
	6624	4231	0.00171	0.02197	0	1282	0	0	0	0	1	-360	360;
	6624	4231	0.00164	0.02064	0	1545	0	0	0	0	1	-360	360;
	6624	6921	0.00136	0.01756	0	1611	0	0	0	0	1	-360	360;
	6624	432	0.00097	0.00942	0	1381	0	0	0	0	1	-360	360;
	6624	432	0.00086	0.01109	0	1381	0	0	0	0	1	-360	360;
	6624	432	0.00085	0.0113	0	1348	0	0	0	0	1	-360	360;
	3353	6887	0.000979	0.0058	0	300	0	0	0	0	1	-360	360;
	2479	2050	0.00052	0.00593	0	1809	0	0	0	0	1	-360	360;
	2479	2050	0.00046	0.00672	0	1809	0	0	0	0	1	-360	360;
	5482	8704	0.001151	0.00636	0	472	0	0	0	0	1	-360	360;
	5482	8704	0.001031	0.007211	0	491	0	0	0	0	1	-360	360;
	5482	4710	0.002341	0.02286	0	433	0	0	0	0	1	-360	360;
	5482	1295	0.002281	0.02518	0	300	0	0	0	0	1	-360	360;
	5482	1910	0.005479	0.02114	0	281	0	0	0	0	1	-360	360;
	5482	2985	0.011159	0.08856	0	376	0	0	0	0	1	-360	360;
	5482	5853	0.000669	0.00249	0	300	0	0	0	0	1	-360	360;
	5482	7824	0.00193	0.01643	0	472	0	0	0	0	1	-360	360;
	5482	3071	0.00024	0.0017	0	338	0	0	0	0	1	-360	360;
	4525	195	0.010159	0.065849	0	0	0	0	0	0	1	-360	360;
	9066	7367	0.00294	0.02131	0	414	0	0	0	0	1	-360	360;
	9066	9155	0.018841	0.07469	0	0	0	0	0	0	1	-360	360;
	9066	6357	0.015599	0.052169	0	300	0	0	0	0	1	-360	360;
	7886	3126	0.005831	0.03969	0	395	0	0	0	0	1	-360	360;
	7886	8564	0.00087	0.00661	0	357	0	0	0	0	1	-360	360;
	7886	5567	0.00637	0.024659	0	281	0	0	0	0	1	-360	360;
	6416	3069	0.0027	0.020419	0	453	0	0	0	0	1	-360	360;
	6416	3876	0.000409	0.00314	0	338	0	0	0	0	1	-360	360;
	6416	8804	0.001669	0.01301	0	376	0	0	0	0	1	-360	360;
	6416	8804	0.00164	0.01325	0	395	0	0	0	0	1	-360	360;
	7831	4550	0.00089	0.00556	0	1183	0	0	0	0	1	-360	360;
	7831	4125	0.00186	0.01809	0	1513	0	0	0	0	1	-360	360;
	7831	3346	0.00381	0.02853	0	920	0	0	0	0	1	-360	360;
	5525	2377	0.012419	0.072151	0	300	0	0	0	0	1	-360	360;
	5525	1401	0.0042	0.02556	0	300	0	0	0	0	1	-360	360;
	5525	2850	0.010771	0.0493	0	376	0	0	0	0	1	-360	360;
	5525	2430	0.010159	0.04995	0	300	0	0	0	0	1	-360	360;
	5525	9164	0.013169	0.05113	0	395	0	0	0	0	1	-360	360;
	8511	1033	0.000469	0.001159	0	300	0	0	0	0	1	-360	360;
	8511	6178	0.000229	0.00105	0	357	0	0	0	0	1	-360	360;
	3400	980	0.000271	0.001271	0	433	0	0	0	0	1	-360	360;
	2101	4598	0.00099	0.01194	0	1545	0	0	0	0	1	-360	360;
	2101	4541	0.00077	0.01184	0	0	0	0	0	0	1	-360	360;
	2101	5144	0.00034	0.00426	0	0	0	0	0	0	1	-360	360;
	3919	2129	0.00101	0.01168	0	1644	0	0	0	0	1	-360	360;
	3919	4541	0.00086	0.01046	0	1710	0	0	0	0	1	-360	360;
	3919	3513	0.00035	0.00417	0	0	0	0	0	0	1	-360	360;
	6675	1486	0.000159	0.00069	0	433	0	0	0	0	1	-360	360;
	4711	5836	0.000781	0.00488	0	529	0	0	0	0	1	-360	360;
	4711	575	0.000781	0.00511	0	433	0	0	0	0	1	-360	360;
	4711	8214	0.00064	0.003539	0	357	0	0	0	0	1	-360	360;
	4711	8214	0.00057	0.003781	0	338	0	0	0	0	1	-360	360;
	4711	3021	0.000531	0.00111	0	0	0	0	0	0	1	-360	360;
	4711	7967	0.000521	0.00113	0	0	0	0	0	0	1	-360	360;
	4711	5993	0.00176	0.00843	0	338	0	0	0	0	1	-360	360;
	4711	8804	0.002461	0.01975	0	472	0	0	0	0	1	-360	360;
	4513	6831	0.000591	0.00105	0	0	0	0	0	0	1	-360	360;
	2043	1422	0.00012	0.00064	0	0	0	0	0	0	1	-360	360;
	6714	1808	0.000581	0.004419	0	0	0	0	0	0	1	-360	360;
	6371	1808	0.000651	0.004331	0	0	0	0	0	0	1	-360	360;
	513	2866	0.001289	0.008289	0	395	0	0	0	0	1	-360	360;
	4970	9222	0.00233	0.03502	0	657	0	0	0	0	1	-360	360;
	4970	9222	0.00238	0.03781	0	657	0	0	0	0	1	-360	360;
	4970	2072	0.00052	0.00605	0	1282	0	0	0	0	1	-360	360;
	4970	5461	0.00023	0.00275	0	1414	0	0	0	0	1	-360	360;
	4970	8109	0.00134	0.01343	0	1611	0	0	0	0	1	-360	360;
	4970	8109	0.00116	0.01552	0	1381	0	0	0	0	1	-360	360;
	7955	707	0.004169	0.027919	0	414	0	0	0	0	1	-360	360;
	7955	4816	0.00063	0.008	0	0	0	0	0	0	1	-360	360;
	7955	1866	0.004841	0.028581	0	529	0	0	0	0	1	-360	360;
	7955	5488	0.007419	0.05787	0	491	0	0	0	0	1	-360	360;
	5093	2467	0.002719	0.020331	0	395	0	0	0	0	1	-360	360;
	2328	7178	0.000789	0.005	0	0	0	0	0	0	1	-360	360;
	9180	195	0.00238	0.013219	0	357	0	0	0	0	1	-360	360;
	9180	3133	0.003771	0.02701	0	395	0	0	0	0	1	-360	360;
	292	4544	0.001271	0.00939	0	0	0	0	0	0	1	-360	360;
	9002	6737	0.000841	0.00131	0	0	0	0	0	0	1	-360	360;
	8439	3758	0.005711	0.043039	0	491	0	0	0	0	1	-360	360;
	8439	3758	0.00562	0.041901	0	395	0	0	0	0	1	-360	360;
	4974	2940	0.00356	0.019781	0	433	0	0	0	0	1	-360	360;
	4974	163	0.004841	0.03149	0	433	0	0	0	0	1	-360	360;
	747	8507	0.00057	0.00107	0	0	0	0	0	0	1	-360	360;
	6909	4914	0.000539	0.00213	0	414	0	0	0	0	1	-360	360;
	444	4747	0.0007	0.00425	0	0	0	0	0	0	1	-360	360;
	1459	6495	0.0012	0.009531	0	0	0	0	0	0	1	-360	360;
	1459	2751	0.00206	0.015901	0	0	0	0	0	0	1	-360	360;
	3601	4728	0.00063	0.00312	0	0	0	0	0	0	1	-360	360;
	1251	6475	3e-05	0.00041	0	1381	0	0	0	0	1	-360	360;
	1093	6475	2e-05	0.00043	0	0	0	0	0	0	1	-360	360;
	5709	6475	3e-05	0.00039	0	1183	0	0	0	0	1	-360	360;
	3825	6475	3e-05	0.00038	0	0	0	0	0	0	1	-360	360;
	1415	3503	0.0008	0.004151	0	0	0	0	0	0	1	-360	360;
	4549	3503	0.000901	0.00381	0	0	0	0	0	0	1	-360	360;
	3072	639	0.000591	0.004341	0	376	0	0	0	0	1	-360	360;
	3072	639	0.000591	0.00405	0	395	0	0	0	0	1	-360	360;
	5007	4541	0.00218	0.0275	0	1052	0	0	0	0	1	-360	360;
	5007	8195	0.00184	0.01677	0	986	0	0	0	0	1	-360	360;
	757	964	0.00058	0.00807	0	986	0	0	0	0	1	-360	360;
	757	964	0.00056	0.00823	0	821	0	0	0	0	1	-360	360;
	757	3906	0.00123	0.01146	0	1315	0	0	0	0	1	-360	360;
	757	3906	0.0011	0.01298	0	1381	0	0	0	0	1	-360	360;
	757	8347	0.00156	0.01585	0	1677	0	0	0	0	1	-360	360;
	757	2919	0.00226	0.02781	0	1381	0	0	0	0	1	-360	360;
	3758	188	0.00614	0.02831	0	300	0	0	0	0	1	-360	360;
	3758	2242	0.00513	0.03401	0	0	0	0	0	0	1	-360	360;
	3903	4114	0.013219	0.0597	0	0	0	0	0	0	1	-360	360;
	5743	6990	0.012969	0.06087	0	0	0	0	0	0	1	-360	360;
	8626	8930	0.00437	0.032831	0	453	0	0	0	0	1	-360	360;
	8626	2467	0.005651	0.033229	0	491	0	0	0	0	1	-360	360;
	2161	5488	0.008969	0.03625	0	338	0	0	0	0	1	-360	360;
	2161	7772	0.012409	0.05645	0	319	0	0	0	0	1	-360	360;
	5003	6382	0.00038	0.002271	0	300	0	0	0	0	1	-360	360;
	6474	4084	0.00138	0.01089	0	472	0	0	0	0	1	-360	360;
	6474	4084	0.001539	0.012521	0	491	0	0	0	0	1	-360	360;
	6474	53	0.00205	0.014229	0	433	0	0	0	0	1	-360	360;
	6474	7259	0.00026	0.00164	0	395	0	0	0	0	1	-360	360;
	6474	2475	0.000289	0.00169	0	0	0	0	0	0	1	-360	360;
	5334	7577	0.003349	0.051211	0	357	0	0	0	0	1	-360	360;
	7961	2288	0.00038	0.001479	0	0	0	0	0	0	1	-360	360;
	1592	7098	0.00082	0.00625	0	453	0	0	0	0	1	-360	360;
	3187	2079	0.00275	0.015581	0	472	0	0	0	0	1	-360	360;
	6041	5926	0.000909	0.00739	0	0	0	0	0	0	1	-360	360;
	3697	8486	0.00082	0.01038	0	1447	0	0	0	0	1	-360	360;
	3697	7530	0.00093	0.01017	0	1282	0	0	0	0	1	-360	360;
	3697	1502	0.00137	0.01431	0	1776	0	0	0	0	1	-360	360;
	3697	1552	0.00215	0.02021	0	1578	0	0	0	0	1	-360	360;
	6952	3793	0.01189	0.058841	0	357	0	0	0	0	1	-360	360;
	6738	2526	0.00104	0.01241	0	1151	0	0	0	0	1	-360	360;
	6738	2526	0.00092	0.014	0	1151	0	0	0	0	1	-360	360;
	6738	2365	0.00055	0.00845	0	1052	0	0	0	0	1	-360	360;
	6738	2365	0.00054	0.00861	0	887	0	0	0	0	1	-360	360;
	6738	3794	0.00105	0.01057	0	1710	0	0	0	0	1	-360	360;
	6738	3613	0.00089	0.01122	0	1513	0	0	0	0	1	-360	360;
	6738	6901	0.00223	0.02584	0	1677	0	0	0	0	1	-360	360;
	8180	5106	0.00101	0.007841	0	453	0	0	0	0	1	-360	360;
	8180	5106	0.001081	0.008099	0	472	0	0	0	0	1	-360	360;
	6240	7165	0.011081	0.068479	0	0	0	0	0	0	1	-360	360;
	9222	8721	2e-05	0.00033	0	0	0	0	0	0	1	-360	360;
	9222	8872	1e-05	0.00029	0	1677	0	0	0	0	1	-360	360;
	9222	8672	0.00071	0.00892	0	1480	0	0	0	0	1	-360	360;
	9222	7164	0.00136	0.01526	0	1348	0	0	0	0	1	-360	360;
	5935	7772	0.005401	0.036021	0	491	0	0	0	0	1	-360	360;
	3134	3430	0.00437	0.033719	0	357	0	0	0	0	1	-360	360;
	3134	1051	0.005151	0.034919	0	414	0	0	0	0	1	-360	360;
	283	800	0.00038	0.00139	0	0	0	0	0	0	1	-360	360;
	283	1077	0.00037	0.0015	0	0	0	0	0	0	1	-360	360;
	283	4747	0.00069	0.004461	0	433	0	0	0	0	1	-360	360;
	3391	5213	0.00688	0.040591	0	281	0	0	0	0	1	-360	360;
	3391	115	0.001341	0.008349	0	376	0	0	0	0	1	-360	360;
	3391	115	0.001229	0.009669	0	529	0	0	0	0	1	-360	360;
	4867	5753	0.00137	0.006039	0	319	0	0	0	0	1	-360	360;
	7539	6556	0.001919	0.01261	0	376	0	0	0	0	1	-360	360;
	7539	871	0.00163	0.0095	0	433	0	0	0	0	1	-360	360;
	8893	8874	0.00488	0.0497	0	472	0	0	0	0	1	-360	360;
	8893	4505	0.001289	0.0087	0	453	0	0	0	0	1	-360	360;
	8893	3436	0.00211	0.01524	0	472	0	0	0	0	1	-360	360;
	8893	3436	0.00206	0.015539	0	491	0	0	0	0	1	-360	360;
	8893	3436	0.00276	0.01631	0	414	0	0	0	0	1	-360	360;
	8893	6556	0.00305	0.019521	0	510	0	0	0	0	1	-360	360;
	8893	8765	0.000409	0.0023	0	395	0	0	0	0	1	-360	360;
	8893	6619	0.00036	0.00268	0	414	0	0	0	0	1	-360	360;
	8165	6734	0.00032	0.00049	0	0	0	0	0	0	1	-360	360;
	1078	6734	0.00013	0.000349	0	0	0	0	0	0	1	-360	360;
	7988	4231	0.0002	0.00327	0	0	0	0	0	0	1	-360	360;
	804	2142	0.000919	0.00561	0	433	0	0	0	0	1	-360	360;
	804	2142	0.00094	0.00549	0	433	0	0	0	0	1	-360	360;
	804	7519	0.00175	0.01976	0	433	0	0	0	0	1	-360	360;
	804	6744	0.001729	0.019331	0	376	0	0	0	0	1	-360	360;
	2286	6802	0.00406	0.02581	0	472	0	0	0	0	1	-360	360;
	2286	769	0.00526	0.025961	0	414	0	0	0	0	1	-360	360;
	3239	1081	0.00179	0.02098	0	1249	0	0	0	0	1	-360	360;
	3239	5983	0.00162	0.02148	0	1282	0	0	0	0	1	-360	360;
	3239	8267	0.00492	0.04991	0	953	0	0	0	0	1	-360	360;
	3239	960	0.00109	0.01361	0	986	0	0	0	0	1	-360	360;
	3239	960	0.00097	0.0153	0	789	0	0	0	0	1	-360	360;
	3239	3906	0.00322	0.03022	0	1381	0	0	0	0	1	-360	360;
	1539	7361	0.032831	0.074151	0	0	0	0	0	0	1	-360	360;
	1539	6852	0.01411	0.058711	0	414	0	0	0	0	1	-360	360;
	1539	8515	0.02725	0.102979	0	338	0	0	0	0	1	-360	360;
	1539	3166	0.012229	0.04049	0	300	0	0	0	0	1	-360	360;
	1539	2676	0.01994	0.066719	0	0	0	0	0	0	1	-360	360;
	1539	3793	0.018591	0.07462	0	338	0	0	0	0	1	-360	360;
	1539	4529	0.008979	0.03251	0	0	0	0	0	0	1	-360	360;
	1539	4234	0.016531	0.086031	0	281	0	0	0	0	1	-360	360;
	8628	44	0.003591	0.02137	0	491	0	0	0	0	1	-360	360;
	8628	6630	0.00124	0.008	0	300	0	0	0	0	1	-360	360;
	8628	6630	0.001271	0.007831	0	338	0	0	0	0	1	-360	360;
	8628	1545	0.000469	0.00294	0	395	0	0	0	0	1	-360	360;
	8628	4215	0.000841	0.003289	0	395	0	0	0	0	1	-360	360;
	7163	1917	0.00219	0.016539	0	300	0	0	0	0	1	-360	360;
	7163	5519	0.002901	0.020901	0	300	0	0	0	0	1	-360	360;
	7647	225	0.0002	0.001211	0	0	0	0	0	0	1	-360	360;
	4864	1397	9.1e-05	0.000479	0	0	0	0	0	0	1	-360	360;
	7021	7119	0.005461	0.029469	0	319	0	0	0	0	1	-360	360;
	7021	3324	0.001979	0.011289	0	491	0	0	0	0	1	-360	360;
	1248	163	0.0023	0.01289	0	0	0	0	0	0	1	-360	360;
	9045	2457	0.000159	0.00037	0	0	0	0	0	0	1	-360	360;
	8874	3786	0.002789	0.01562	0	433	0	0	0	0	1	-360	360;
	5891	5574	0.002531	0.012331	0	281	0	0	0	0	1	-360	360;
	5891	7663	0.00357	0.023289	0	300	0	0	0	0	1	-360	360;
	5891	4889	0.00076	0.006039	0	414	0	0	0	0	1	-360	360;
	3997	1526	0.000469	0.00075	0	0	0	0	0	0	1	-360	360;
	184	1526	0.00044	0.00069	0	0	0	0	0	0	1	-360	360;
	2972	2088	0.00561	0.035521	0	491	0	0	0	0	1	-360	360;
	8676	933	0.004539	0.03468	0	0	0	0	0	0	1	-360	360;
	8676	933	0.00175	0.0067	0	319	0	0	0	0	1	-360	360;
	8676	3643	0.007461	0.031419	0	300	0	0	0	0	1	-360	360;
	8704	6723	0.00224	0.012909	0	510	0	0	0	0	1	-360	360;
	7047	21	0.0003	0.000961	0	319	0	0	0	0	1	-360	360;
	7047	21	0.00032	0.001409	0	0	0	0	0	0	1	-360	360;
	7047	1551	0.002031	0.0157	0	453	0	0	0	0	1	-360	360;
	7047	4103	0.00038	0.00306	0	414	0	0	0	0	1	-360	360;
	7047	7862	0.00118	0.0083	0	453	0	0	0	0	1	-360	360;
	7047	5458	0.001229	0.00787	0	472	0	0	0	0	1	-360	360;
	5574	5666	0.00301	0.01538	0	414	0	0	0	0	1	-360	360;
	6742	5458	0.00063	0.00114	0	281	0	0	0	0	1	-360	360;
	5270	7862	0.00064	0.00111	0	0	0	0	0	0	1	-360	360;
	7775	3857	0.004919	0.03837	0	529	0	0	0	0	1	-360	360;
	6290	3436	0.001831	0.008349	0	338	0	0	0	0	1	-360	360;
	3430	401	0.00536	0.03188	0	433	0	0	0	0	1	-360	360;
	2967	8976	0.00038	0.001461	0	0	0	0	0	0	1	-360	360;
	2967	8976	0.000289	0.001521	0	0	0	0	0	0	1	-360	360;
	8976	2481	0.00305	0.016841	0	376	0	0	0	0	1	-360	360;
	8976	9108	0.001781	0.012729	0	357	0	0	0	0	1	-360	360;
	2056	7520	0.00041	0.00461	0	854	0	0	0	0	1	-360	360;
	2056	7520	0.0004	0.0047	0	920	0	0	0	0	1	-360	360;
	4084	5340	0.00213	0.01401	0	529	0	0	0	0	1	-360	360;
	4084	2319	0.001271	0.00887	0	338	0	0	0	0	1	-360	360;
	4084	2878	0.00462	0.02743	0	433	0	0	0	0	1	-360	360;
	2597	7422	0.001539	0.00776	0	0	0	0	0	0	1	-360	360;
	7183	8931	0.00025	0.00478	0	0	0	0	0	0	1	-360	360;
	5379	8931	0.00025	0.00487	0	0	0	0	0	0	1	-360	360;
	8128	8931	0.00025	0.00435	0	723	0	0	0	0	1	-360	360;
	6337	8931	0.00026	0.00426	0	657	0	0	0	0	1	-360	360;
	1237	8931	0.00025	0.00479	0	0	0	0	0	0	1	-360	360;
	4127	9173	0.000169	0.00076	0	338	0	0	0	0	1	-360	360;
	4127	7974	0.00039	0.00126	0	395	0	0	0	0	1	-360	360;
	2432	4816	0.011229	0.074659	0	529	0	0	0	0	1	-360	360;
	2432	1866	0.014521	0.058081	0	300	0	0	0	0	1	-360	360;
	2432	8809	0.00064	0.003919	0	433	0	0	0	0	1	-360	360;
	2432	5146	0.014789	0.07768	0	338	0	0	0	0	1	-360	360;
	2432	6940	0.00938	0.05355	0	472	0	0	0	0	1	-360	360;
	2432	1742	0.00124	0.004289	0	319	0	0	0	0	1	-360	360;
	5213	8627	0.00968	0.04388	0	0	0	0	0	0	1	-360	360;
	5213	174	0.008531	0.067151	0	357	0	0	0	0	1	-360	360;
	5213	7437	0.00731	0.04106	0	338	0	0	0	0	1	-360	360;
	7342	5940	0.004349	0.02555	0	433	0	0	0	0	1	-360	360;
	7342	7495	0.00599	0.03462	0	376	0	0	0	0	1	-360	360;
	8808	4816	0.002961	0.021591	0	453	0	0	0	0	1	-360	360;
	8808	4816	0.002901	0.02206	0	472	0	0	0	0	1	-360	360;
	8808	6053	0.002979	0.016969	0	376	0	0	0	0	1	-360	360;
	5764	4189	0.00312	0.020021	0	376	0	0	0	0	1	-360	360;
	8507	1973	0.000219	0.000909	0	0	0	0	0	0	1	-360	360;
	1394	1973	0.000219	0.00093	0	357	0	0	0	0	1	-360	360;
	2898	2128	0.000151	0.000289	0	0	0	0	0	0	1	-360	360;
	4942	8255	0.00563	0.043599	0	0	0	0	0	0	1	-360	360;
	4942	4748	0.000969	0.00612	0	414	0	0	0	0	1	-360	360;
	3486	3594	0.005021	0.020719	0	338	0	0	0	0	1	-360	360;
	3486	4679	0.00988	0.041039	0	0	0	0	0	0	1	-360	360;
	4656	4235	0.003159	0.02011	0	376	0	0	0	0	1	-360	360;
	4656	4235	0.003219	0.01969	0	376	0	0	0	0	1	-360	360;
	7129	5441	0.00049	0.00238	0	0	0	0	0	0	1	-360	360;
	7129	5441	0.000479	0.002419	0	0	0	0	0	0	1	-360	360;
	8250	5720	0.00264	0.019729	0	433	0	0	0	0	1	-360	360;
	8250	5720	0.002581	0.02011	0	395	0	0	0	0	1	-360	360;
	6831	6723	0.00037	0.002831	0	453	0	0	0	0	1	-360	360;
	6831	5686	0.00213	0.012789	0	414	0	0	0	0	1	-360	360;
	2377	805	0.018669	0.092281	0	338	0	0	0	0	1	-360	360;
	907	9091	3.9e-05	0.000401	0	0	0	0	0	0	1	-360	360;
	4056	9091	6e-05	0.00044	0	281	0	0	0	0	1	-360	360;
	6837	2142	6e-05	0.0005	0	414	0	0	0	0	1	-360	360;
	5666	2641	0.001271	0.007031	0	376	0	0	0	0	1	-360	360;
	5666	7577	0.00055	0.003159	0	510	0	0	0	0	1	-360	360;
	5666	7577	0.00055	0.00338	0	491	0	0	0	0	1	-360	360;
	5666	7577	0.00105	0.004771	0	395	0	0	0	0	1	-360	360;
	5666	1629	0.003289	0.015039	0	281	0	0	0	0	1	-360	360;
	726	687	0.00036	0.004349	0	510	0	0	0	0	1	-360	360;
	3526	687	0.00038	0.003669	0	0	0	0	0	0	1	-360	360;
	6382	2425	0.00118	0.00701	0	0	0	0	0	0	1	-360	360;
	6382	2425	0.001229	0.00686	0	433	0	0	0	0	1	-360	360;
	6382	5907	0.00301	0.0217	0	0	0	0	0	0	1	-360	360;
	6382	5907	0.003409	0.01924	0	0	0	0	0	0	1	-360	360;
	6382	5907	0.003409	0.01938	0	472	0	0	0	0	1	-360	360;
	7847	9217	0.002719	0.02131	0	0	0	0	0	0	1	-360	360;
	2981	1448	0.00295	0.01956	0	0	0	0	0	0	1	-360	360;
	5110	5686	0.001091	0.00831	0	376	0	0	0	0	1	-360	360;
	2968	5907	0.00111	0.008211	0	453	0	0	0	0	1	-360	360;
	5441	2155	0.002469	0.015271	0	529	0	0	0	0	1	-360	360;
	5441	8112	0.00212	0.01601	0	0	0	0	0	0	1	-360	360;
	5441	2535	0.00037	0.0028	0	433	0	0	0	0	1	-360	360;
	5441	7098	0.00162	0.01238	0	0	0	0	0	0	1	-360	360;
	5441	7098	0.001909	0.01162	0	0	0	0	0	0	1	-360	360;
	5441	7098	0.001909	0.010729	0	453	0	0	0	0	1	-360	360;
	5441	3737	0.00089	0.00526	0	319	0	0	0	0	1	-360	360;
	5441	5529	0.00087	0.00536	0	0	0	0	0	0	1	-360	360;
	5441	5753	0.001521	0.01176	0	433	0	0	0	0	1	-360	360;
	9019	6563	0.00014	0.000591	0	376	0	0	0	0	1	-360	360;
	1341	1101	0.003531	0.028781	0	0	0	0	0	0	1	-360	360;
	2797	8291	0.004599	0.03657	0	453	0	0	0	0	1	-360	360;
	2797	8367	0.00307	0.019159	0	491	0	0	0	0	1	-360	360;
	2797	1100	0.000581	0.004461	0	414	0	0	0	0	1	-360	360;
	2797	6887	0.00164	0.01161	0	491	0	0	0	0	1	-360	360;
	4939	6880	0.00057	0.003281	0	433	0	0	0	0	1	-360	360;
	7738	6982	0.020419	0.063669	0	0	0	0	0	0	1	-360	360;
	6565	3121	0.01424	0.04751	0	0	0	0	0	0	1	-360	360;
	8707	280	0.00157	0.008849	0	453	0	0	0	0	1	-360	360;
	8707	2510	0.01218	0.07524	0	0	0	0	0	0	1	-360	360;
	1625	766	0.00037	0.00151	0	0	0	0	0	0	1	-360	360;
	8411	6232	0.000219	0.000531	0	0	0	0	0	0	1	-360	360;
	3951	1397	0.002849	0.0137	0	0	0	0	0	0	1	-360	360;
	964	7282	0.00116	0.01692	0	1183	0	0	0	0	1	-360	360;
	964	7282	0.00133	0.01462	0	1315	0	0	0	0	1	-360	360;
	964	6475	0.00063	0.00821	0	887	0	0	0	0	1	-360	360;
	964	6475	0.00064	0.00804	0	1183	0	0	0	0	1	-360	360;
	5658	5420	0.00126	0.01272	0	1282	0	0	0	0	1	-360	360;
	5658	8267	0.00337	0.0494	0	1644	0	0	0	0	1	-360	360;
	5658	7328	0.0014	0.0214	0	0	0	0	0	0	1	-360	360;
	5658	7328	0.00137	0.02182	0	0	0	0	0	0	1	-360	360;
	5658	7164	0.00212	0.02063	0	1118	0	0	0	0	1	-360	360;
	5550	410	0.00186	0.01045	0	510	0	0	0	0	1	-360	360;
	5550	7579	0.00426	0.021771	0	338	0	0	0	0	1	-360	360;
	5550	7579	0.002659	0.020539	0	395	0	0	0	0	1	-360	360;
	5550	4189	0.00606	0.049919	0	357	0	0	0	0	1	-360	360;
	7282	4196	0.00109	0.01192	0	789	0	0	0	0	1	-360	360;
	1526	3916	0.013021	0.049219	0	338	0	0	0	0	1	-360	360;
	1526	4747	0.00236	0.01239	0	338	0	0	0	0	1	-360	360;
	1526	4747	0.0027	0.015169	0	319	0	0	0	0	1	-360	360;
	4541	3346	0.00285	0.03283	0	1414	0	0	0	0	1	-360	360;
	4541	3346	0.00529	0.03969	0	920	0	0	0	0	1	-360	360;
	1343	3346	0.00254	0.03721	0	1447	0	0	0	0	1	-360	360;
	3680	2719	0.000159	0.00069	0	0	0	0	0	0	1	-360	360;
	3680	2719	0.000229	0.000979	0	0	0	0	0	0	1	-360	360;
	3680	2719	0.00024	0.00082	0	0	0	0	0	0	1	-360	360;
	3680	795	0.000169	0.00075	0	0	0	0	0	0	1	-360	360;
	3680	4880	0.00368	0.01756	0	0	0	0	0	0	1	-360	360;
	3680	1868	0.002599	0.020659	0	0	0	0	0	0	1	-360	360;
	7464	5564	0.00018	0.00064	0	0	0	0	0	0	1	-360	360;
	7464	7036	0.000159	0.000771	0	281	0	0	0	0	1	-360	360;
	7464	5940	0.00357	0.01981	0	376	0	0	0	0	1	-360	360;
	7464	7495	0.006021	0.03205	0	433	0	0	0	0	1	-360	360;
	4850	2886	0.001781	0.01164	0	433	0	0	0	0	1	-360	360;
	4850	7049	0.002781	0.018831	0	414	0	0	0	0	1	-360	360;
	2510	6146	0.000349	0.00281	0	0	0	0	0	0	1	-360	360;
	90	1860	0.00313	0.02444	0	510	0	0	0	0	1	-360	360;
	90	1860	0.00368	0.02074	0	491	0	0	0	0	1	-360	360;
	2142	3276	9.1e-05	0.001539	0	0	0	0	0	0	1	-360	360;
	2142	3545	0.00251	0.01111	0	0	0	0	0	0	1	-360	360;
	2142	6842	0.003081	0.017961	0	319	0	0	0	0	1	-360	360;
	2142	6744	0.005341	0.030419	0	338	0	0	0	0	1	-360	360;
	2142	7635	9.9e-05	0.00126	0	414	0	0	0	0	1	-360	360;
	9091	1040	0.002169	0.017901	0	319	0	0	0	0	1	-360	360;
	2457	6178	0.00011	0.00038	0	0	0	0	0	0	1	-360	360;
	1422	1462	0.00212	0.00494	0	0	0	0	0	0	1	-360	360;
	1422	7772	0.018901	0.077349	0	376	0	0	0	0	1	-360	360;
	7881	5365	0.00957	0.0395	0	357	0	0	0	0	1	-360	360;
	7881	6982	0.02324	0.063469	0	300	0	0	0	0	1	-360	360;
	6926	3327	0.000531	0.004331	0	510	0	0	0	0	1	-360	360;
	1959	615	0.002581	0.006229	0	0	0	0	0	0	1	-360	360;
	6455	1262	0.00031	0.002211	0	319	0	0	0	0	1	-360	360;
	803	6982	0.01087	0.06114	0	338	0	0	0	0	1	-360	360;
	803	5918	0.011581	0.061099	0	319	0	0	0	0	1	-360	360;
	7624	3070	0.002099	0.016271	0	300	0	0	0	0	1	-360	360;
	500	1547	0.00093	0.005461	0	0	0	0	0	0	1	-360	360;
	500	5237	0.009591	0.037039	0	319	0	0	0	0	1	-360	360;
	500	772	0.002591	0.017211	0	453	0	0	0	0	1	-360	360;
	500	8950	0.00164	0.013651	0	395	0	0	0	0	1	-360	360;
	2754	2079	0.003219	0.014719	0	357	0	0	0	0	1	-360	360;
	2754	2229	0.00401	0.022229	0	357	0	0	0	0	1	-360	360;
	2754	7049	0.00126	0.010211	0	491	0	0	0	0	1	-360	360;
	4118	7873	0.009169	0.03993	0	0	0	0	0	0	1	-360	360;
	4118	3865	0.001349	0.006081	0	0	0	0	0	0	1	-360	360;
	1754	8486	0.00156	0.01931	0	1480	0	0	0	0	1	-360	360;
	1754	7530	0.00152	0.01968	0	1381	0	0	0	0	1	-360	360;
	1754	7778	0.00094	0.01433	0	1578	0	0	0	0	1	-360	360;
	1754	6897	0.00092	0.01461	0	1348	0	0	0	0	1	-360	360;
	1754	960	0.00137	0.02012	0	1019	0	0	0	0	1	-360	360;
	1754	960	0.00134	0.02051	0	821	0	0	0	0	1	-360	360;
	3857	4491	0.00314	0.042151	0	472	0	0	0	0	1	-360	360;
	3857	6010	0.000651	0.00494	0	0	0	0	0	0	1	-360	360;
	3857	6940	0.003961	0.02426	0	491	0	0	0	0	1	-360	360;
	858	1642	0.00013	0.0015	0	0	0	0	0	0	1	-360	360;
	3028	809	0.00013	0.00147	0	1776	0	0	0	0	1	-360	360;
	4819	1642	0.00011	0.00172	0	1776	0	0	0	0	1	-360	360;
	2035	809	0.00012	0.00166	0	0	0	0	0	0	1	-360	360;
	3869	1642	0.00012	0.00168	0	1809	0	0	0	0	1	-360	360;
	3741	809	0.00012	0.00172	0	1809	0	0	0	0	1	-360	360;
	742	687	0.00093	0.005669	0	0	0	0	0	0	1	-360	360;
	2910	7098	0.00105	0.00506	0	0	0	0	0	0	1	-360	360;
	5688	6231	0.00113	0.006539	0	0	0	0	0	0	1	-360	360;
	5182	5455	0.00037	0.002919	0	0	0	0	0	0	1	-360	360;
	707	5695	0.005169	0.03738	0	376	0	0	0	0	1	-360	360;
	3814	954	0.00125	0.006669	0	357	0	0	0	0	1	-360	360;
	2313	2438	0.001281	0.00661	0	0	0	0	0	0	1	-360	360;
	195	2985	0.00226	0.0138	0	357	0	0	0	0	1	-360	360;
	2518	3513	0.00213	0.02677	0	1513	0	0	0	0	1	-360	360;
	2518	3346	0.00154	0.01919	0	1249	0	0	0	0	1	-360	360;
	4544	601	0.00687	0.04413	0	0	0	0	0	0	1	-360	360;
	4544	3241	0.0045	0.04725	0	357	0	0	0	0	1	-360	360;
	4544	7338	0.00138	0.00837	0	0	0	0	0	0	1	-360	360;
	3069	1538	0.000419	0.001901	0	0	0	0	0	0	1	-360	360;
	3069	1538	0.00043	0.00186	0	0	0	0	0	0	1	-360	360;
	6115	4729	0.003331	0.015099	0	491	0	0	0	0	1	-360	360;
	3344	6616	0.002099	0.01631	0	433	0	0	0	0	1	-360	360;
	3809	5616	0.001331	0.01082	0	491	0	0	0	0	1	-360	360;
	3809	6691	0.00199	0.00818	0	281	0	0	0	0	1	-360	360;
	3809	6691	0.00169	0.008831	0	0	0	0	0	0	1	-360	360;
	1326	892	0.001771	0.014281	0	0	0	0	0	0	1	-360	360;
	601	616	0.003099	0.01844	0	395	0	0	0	0	1	-360	360;
	3496	1551	0.003409	0.01819	0	338	0	0	0	0	1	-360	360;
	4925	2230	0.00057	0.00369	0	0	0	0	0	0	1	-360	360;
	8057	5907	0.001081	0.00468	0	0	0	0	0	0	1	-360	360;
	5354	5907	0.00105	0.004771	0	0	0	0	0	0	1	-360	360;
	3674	616	0.001979	0.01549	0	491	0	0	0	0	1	-360	360;
	3674	7338	0.001469	0.015151	0	0	0	0	0	0	1	-360	360;
	5836	5083	0.000711	0.003969	0	319	0	0	0	0	1	-360	360;
	5836	7576	0.00051	0.00274	0	0	0	0	0	0	1	-360	360;
	5836	3191	0.000521	0.00306	0	395	0	0	0	0	1	-360	360;
	5836	3488	0.00188	0.011841	0	0	0	0	0	0	1	-360	360;
	5836	1578	0.002469	0.010539	0	0	0	0	0	0	1	-360	360;
	575	5300	0.00045	0.00351	0	414	0	0	0	0	1	-360	360;
	3929	9173	0.00037	0.000659	0	0	0	0	0	0	1	-360	360;
	2089	3818	0.000401	0.0007	0	0	0	0	0	0	1	-360	360;
	2089	6232	0.00014	0.000531	0	0	0	0	0	0	1	-360	360;
	2089	1643	0.000289	0.0013	0	395	0	0	0	0	1	-360	360;
	2089	6253	7e-05	0.000289	0	0	0	0	0	0	1	-360	360;
	2089	6684	0.00051	0.002151	0	281	0	0	0	0	1	-360	360;
	2089	8151	0.00037	0.00157	0	433	0	0	0	0	1	-360	360;
	5067	449	0.00022	0.00277	0	1348	0	0	0	0	1	-360	360;
	5067	449	0.00025	0.00271	0	1447	0	0	0	0	1	-360	360;
	7641	3037	0.000271	0.002081	0	472	0	0	0	0	1	-360	360;
	7641	3037	0.00025	0.00231	0	433	0	0	0	0	1	-360	360;
	3191	2340	0.00095	0.00331	0	0	0	0	0	0	1	-360	360;
	3191	8711	0.00093	0.00336	0	0	0	0	0	0	1	-360	360;
	8104	8722	0.000341	0.002211	0	491	0	0	0	0	1	-360	360;
	2644	9191	0.00032	0.00232	0	0	0	0	0	0	1	-360	360;
	4128	8522	3.9e-05	0.000271	0	414	0	0	0	0	1	-360	360;
	604	7522	0.00075	0.0101	0	1513	0	0	0	0	1	-360	360;
	604	4000	0.00118	0.01455	0	1447	0	0	0	0	1	-360	360;
	2155	9021	0.0017	0.0098	0	414	0	0	0	0	1	-360	360;
	2155	9021	0.001521	0.01111	0	0	0	0	0	0	1	-360	360;
	2155	2042	0.000169	0.001401	0	0	0	0	0	0	1	-360	360;
	2155	687	0.002591	0.01725	0	510	0	0	0	0	1	-360	360;
	2155	3520	0.000169	0.001159	0	338	0	0	0	0	1	-360	360;
	2155	4729	0.00307	0.01339	0	433	0	0	0	0	1	-360	360;
	2641	4889	0.002591	0.013521	0	414	0	0	0	0	1	-360	360;
	9067	9051	0.000581	0.005031	0	453	0	0	0	0	1	-360	360;
	6969	5573	0.001419	0.00595	0	281	0	0	0	0	1	-360	360;
	666	1494	0.00205	0.02664	0	1348	0	0	0	0	1	-360	360;
	666	2732	0.00295	0.03721	0	1414	0	0	0	0	1	-360	360;
	666	4550	0.00335	0.03648	0	1381	0	0	0	0	1	-360	360;
	666	1607	0.00022	0.00251	0	1151	0	0	0	0	1	-360	360;
	666	2093	0.00059	0.00651	0	1545	0	0	0	0	1	-360	360;
	3565	2177	0.008539	0.041979	0	281	0	0	0	0	1	-360	360;
	3565	2841	0.001349	0.0052	0	300	0	0	0	0	1	-360	360;
	3611	7972	0.00245	0.013909	0	414	0	0	0	0	1	-360	360;
	3611	2481	0.001469	0.008419	0	453	0	0	0	0	1	-360	360;
	5420	7164	0.00084	0.00836	0	1249	0	0	0	0	1	-360	360;
	7119	2078	0.002591	0.017789	0	414	0	0	0	0	1	-360	360;
	7119	7873	0.009651	0.05968	0	319	0	0	0	0	1	-360	360;
	7422	7519	0.00143	0.00799	0	281	0	0	0	0	1	-360	360;
	7422	6744	0.001969	0.014909	0	472	0	0	0	0	1	-360	360;
	1026	1860	0.01039	0.06406	0	395	0	0	0	0	1	-360	360;
	4765	3327	0.000719	0.005331	0	338	0	0	0	0	1	-360	360;
	6036	6921	0.00052	0.00565	0	1315	0	0	0	0	1	-360	360;
	8670	7069	0.002409	0.019479	0	395	0	0	0	0	1	-360	360;
	8670	7069	0.00256	0.019151	0	319	0	0	0	0	1	-360	360;
	8670	2866	0.00656	0.02288	0	281	0	0	0	0	1	-360	360;
	4032	6450	0.0027	0.014901	0	300	0	0	0	0	1	-360	360;
	4032	5738	0.003591	0.01362	0	300	0	0	0	0	1	-360	360;
	7327	2481	0.002211	0.01188	0	0	0	0	0	0	1	-360	360;
	7327	8189	0.00068	0.00357	0	300	0	0	0	0	1	-360	360;
	7474	1584	0.000211	0.000969	0	300	0	0	0	0	1	-360	360;
	8947	3390	9.9e-05	0.000539	0	0	0	0	0	0	1	-360	360;
	1808	6384	0.00645	0.0263	0	319	0	0	0	0	1	-360	360;
	4624	6031	0.00063	0.001901	0	0	0	0	0	0	1	-360	360;
	1721	3558	0.000281	0.000669	0	0	0	0	0	0	1	-360	360;
	7694	3865	0.000219	0.00064	0	0	0	0	0	0	1	-360	360;
	9181	6723	0.00014	0.00105	0	0	0	0	0	0	1	-360	360;
	8373	8748	0.003219	0.020729	0	433	0	0	0	0	1	-360	360;
	8373	8475	0.00051	0.003151	0	472	0	0	0	0	1	-360	360;
	8373	8475	0.000409	0.00331	0	414	0	0	0	0	1	-360	360;
	5630	6639	0.002289	0.017479	0	491	0	0	0	0	1	-360	360;
	5630	1187	0.002831	0.01895	0	433	0	0	0	0	1	-360	360;
	5630	7274	0.006289	0.036521	0	0	0	0	0	0	1	-360	360;
	5630	8992	0.005651	0.032169	0	338	0	0	0	0	1	-360	360;
	5630	7266	0.00414	0.02262	0	376	0	0	0	0	1	-360	360;
	3492	6734	0.01095	0.05814	0	357	0	0	0	0	1	-360	360;
	3492	1015	0.009	0.0368	0	395	0	0	0	0	1	-360	360;
	3492	8310	0.000229	0.00181	0	453	0	0	0	0	1	-360	360;
	1980	6570	0.00119	0.00726	0	510	0	0	0	0	1	-360	360;
	9137	749	0.011771	0.051669	0	529	0	0	0	0	1	-360	360;
	7522	2183	0.00209	0.02279	0	1611	0	0	0	0	1	-360	360;
	7571	3916	0.002031	0.01443	0	0	0	0	0	0	1	-360	360;
	7571	6891	0.00145	0.008479	0	453	0	0	0	0	1	-360	360;
	2526	6199	0.00124	0.01433	0	1414	0	0	0	0	1	-360	360;
	2526	6199	0.00127	0.01402	0	1776	0	0	0	0	1	-360	360;
	1483	3649	0.013039	0.052841	0	529	0	0	0	0	1	-360	360;
	2365	6199	0.00235	0.02739	0	854	0	0	0	0	1	-360	360;
	2365	6199	0.0024	0.02681	0	1118	0	0	0	0	1	-360	360;
	2365	4141	0.00135	0.02056	0	657	0	0	0	0	1	-360	360;
	2365	1311	0.00087	0.01124	0	854	0	0	0	0	1	-360	360;
	8316	174	0.0012	0.00818	0	529	0	0	0	0	1	-360	360;
	174	6430	0.006961	0.05561	0	395	0	0	0	0	1	-360	360;
	5340	7892	0.00105	0.00763	0	376	0	0	0	0	1	-360	360;
	5340	561	0.000219	0.00182	0	0	0	0	0	0	1	-360	360;
	5340	8843	3.9e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	5340	8843	4.5e-05	0.000195	0	0	0	0	0	0	1	-360	360;
	5340	4239	0.0035	0.022651	0	433	0	0	0	0	1	-360	360;
	3866	1860	0.004281	0.02319	0	357	0	0	0	0	1	-360	360;
	3866	5395	0.003729	0.014021	0	319	0	0	0	0	1	-360	360;
	7042	2177	0.00212	0.01582	0	472	0	0	0	0	1	-360	360;
	8672	7164	0.00056	0.00784	0	657	0	0	0	0	1	-360	360;
	1860	7772	0.004401	0.01813	0	0	0	0	0	0	1	-360	360;
	1860	7772	0.005021	0.02144	0	0	0	0	0	0	1	-360	360;
	4885	2928	0.00018	0.000219	0	281	0	0	0	0	1	-360	360;
	4157	2928	0.000151	0.000211	0	0	0	0	0	0	1	-360	360;
	4432	2928	0.000151	0.000211	0	0	0	0	0	0	1	-360	360;
	2072	5983	0.00282	0.02591	0	1381	0	0	0	0	1	-360	360;
	3450	6552	0.00131	0.008539	0	433	0	0	0	0	1	-360	360;
	3956	7507	0.002159	0.015969	0	0	0	0	0	0	1	-360	360;
	5286	1838	0.00019	0.000281	0	0	0	0	0	0	1	-360	360;
	5571	2230	0.0002	0.000281	0	0	0	0	0	0	1	-360	360;
	2078	8005	0.00139	0.00774	0	472	0	0	0	0	1	-360	360;
	639	6351	0.01101	0.06739	0	529	0	0	0	0	1	-360	360;
	639	2749	0.00063	0.00463	0	338	0	0	0	0	1	-360	360;
	639	9130	0.002341	0.009599	0	376	0	0	0	0	1	-360	360;
	410	4189	0.00375	0.021659	0	376	0	0	0	0	1	-360	360;
	1494	5781	0.00056	0.00749	0	1183	0	0	0	0	1	-360	360;
	583	6616	0.00118	0.00926	0	433	0	0	0	0	1	-360	360;
	583	6031	0.002039	0.013081	0	414	0	0	0	0	1	-360	360;
	1746	7507	0.00155	0.010961	0	0	0	0	0	0	1	-360	360;
	3276	7635	3.9e-05	0.00036	0	319	0	0	0	0	1	-360	360;
	1105	196	0.00024	0.000409	0	281	0	0	0	0	1	-360	360;
	8522	7132	0.000341	0.001659	0	453	0	0	0	0	1	-360	360;
	8522	5351	0.005461	0.024919	0	357	0	0	0	0	1	-360	360;
	8763	7056	0.00158	0.02413	0	854	0	0	0	0	1	-360	360;
	8763	8825	0.00155	0.02458	0	756	0	0	0	0	1	-360	360;
	8763	6889	0.00021	0.00256	0	1315	0	0	0	0	1	-360	360;
	8763	6889	0.00024	0.00243	0	1414	0	0	0	0	1	-360	360;
	8763	8887	0.00144	0.02266	0	1414	0	0	0	0	1	-360	360;
	8763	8887	0.0014	0.02098	0	1447	0	0	0	0	1	-360	360;
	8763	8487	0.00086	0.01074	0	0	0	0	0	0	1	-360	360;
	8748	8475	0.003341	0.02611	0	453	0	0	0	0	1	-360	360;
	217	6785	0.00387	0.02063	0	414	0	0	0	0	1	-360	360;
	217	6785	0.00405	0.018271	0	414	0	0	0	0	1	-360	360;
	217	1179	0.00101	0.00462	0	300	0	0	0	0	1	-360	360;
	217	2575	0.00088	0.00389	0	319	0	0	0	0	1	-360	360;
	217	7274	0.005081	0.039	0	414	0	0	0	0	1	-360	360;
	7972	8189	0.00257	0.0208	0	453	0	0	0	0	1	-360	360;
	7972	8189	0.00251	0.01926	0	414	0	0	0	0	1	-360	360;
	7972	1039	0.002531	0.01314	0	0	0	0	0	0	1	-360	360;
	7253	7840	0.00132	0.00869	0	0	0	0	0	0	1	-360	360;
	4395	8913	0.00082	0.004021	0	319	0	0	0	0	1	-360	360;
	4395	4852	0.00143	0.007479	0	338	0	0	0	0	1	-360	360;
	2795	1183	0.000659	0.005771	0	0	0	0	0	0	1	-360	360;
	2795	1183	0.000591	0.006539	0	0	0	0	0	0	1	-360	360;
	1305	6291	0.00589	0.025599	0	376	0	0	0	0	1	-360	360;
	1305	6982	0.00614	0.02343	0	338	0	0	0	0	1	-360	360;
	2732	1798	0.00155	0.01994	0	1381	0	0	0	0	1	-360	360;
	2732	3346	0.00205	0.03218	0	1480	0	0	0	0	1	-360	360;
	4550	3346	0.00233	0.03154	0	1480	0	0	0	0	1	-360	360;
	4725	5738	0.002341	0.01982	0	0	0	0	0	0	1	-360	360;
	1083	8310	8.1e-05	0.000521	0	433	0	0	0	0	1	-360	360;
	1398	5567	9.9e-05	0.00069	0	491	0	0	0	0	1	-360	360;
	4281	4580	0.00276	0.016419	0	453	0	0	0	0	1	-360	360;
	7050	4908	0.00036	0.001229	0	357	0	0	0	0	1	-360	360;
	7050	3740	0.000229	0.000711	0	414	0	0	0	0	1	-360	360;
	6563	870	0.00013	0.00099	0	414	0	0	0	0	1	-360	360;
	8999	6495	0.001159	0.00689	0	319	0	0	0	0	1	-360	360;
	6888	5994	0.008531	0.03045	0	281	0	0	0	0	1	-360	360;
	5400	8005	0.000781	0.004531	0	414	0	0	0	0	1	-360	360;
	5400	8005	0.00075	0.00474	0	433	0	0	0	0	1	-360	360;
	5881	4031	0.002159	0.008349	0	0	0	0	0	0	1	-360	360;
	8195	8334	0.00162	0.01651	0	1447	0	0	0	0	1	-360	360;
	933	6357	0.01236	0.044211	0	300	0	0	0	0	1	-360	360;
	933	6357	0.00825	0.03989	0	319	0	0	0	0	1	-360	360;
	933	3775	0.02187	0.066961	0	0	0	0	0	0	1	-360	360;
	933	3975	0.01212	0.06661	0	0	0	0	0	0	1	-360	360;
	3918	1642	0.00036	0.00534	0	789	0	0	0	0	1	-360	360;
	6802	1153	0.000229	0.00063	0	0	0	0	0	0	1	-360	360;
	6802	7599	0.000219	0.000581	0	0	0	0	0	0	1	-360	360;
	4025	3994	0.000281	0.00063	0	0	0	0	0	0	1	-360	360;
	1595	1414	0.005021	0.035659	0	319	0	0	0	0	1	-360	360;
	3114	3656	0.00174	0.007901	0	0	0	0	0	0	1	-360	360;
	5237	2229	0.003919	0.020581	0	395	0	0	0	0	1	-360	360;
	5237	9108	0.003341	0.016349	0	395	0	0	0	0	1	-360	360;
	5237	1478	0.003271	0.016669	0	357	0	0	0	0	1	-360	360;
	2799	8515	0.000151	0.000659	0	319	0	0	0	0	1	-360	360;
	6405	7873	0.00736	0.035659	0	0	0	0	0	0	1	-360	360;
	6405	3894	0.00944	0.042169	0	0	0	0	0	0	1	-360	360;
	4435	4734	0.00068	0.00652	0	756	0	0	0	0	1	-360	360;
	4435	3483	0.00188	0.02281	0	1282	0	0	0	0	1	-360	360;
	2079	7052	0.00168	0.00956	0	433	0	0	0	0	1	-360	360;
	2079	7052	0.001669	0.00981	0	414	0	0	0	0	1	-360	360;
	2079	772	0.002229	0.01599	0	491	0	0	0	0	1	-360	360;
	2079	772	0.00218	0.01631	0	510	0	0	0	0	1	-360	360;
	1081	6475	0.00238	0.02124	0	953	0	0	0	0	1	-360	360;
	1081	6475	0.00194	0.02023	0	1151	0	0	0	0	1	-360	360;
	7361	3166	0.013831	0.05955	0	0	0	0	0	0	1	-360	360;
	7361	426	0.00876	0.04875	0	319	0	0	0	0	1	-360	360;
	2863	1857	0.0002	0.00099	0	300	0	0	0	0	1	-360	360;
	1364	2430	0.00105	0.00826	0	376	0	0	0	0	1	-360	360;
	2177	749	0.0172	0.06637	0	376	0	0	0	0	1	-360	360;
	3594	2850	0.00305	0.015781	0	376	0	0	0	0	1	-360	360;
	6555	3608	0.00208	0.02105	0	1381	0	0	0	0	1	-360	360;
	8475	1965	0.00614	0.043729	0	376	0	0	0	0	1	-360	360;
	8475	5926	0.000229	0.001849	0	453	0	0	0	0	1	-360	360;
	5735	6053	0.00386	0.025789	0	338	0	0	0	0	1	-360	360;
	5735	2341	0.002031	0.012	0	491	0	0	0	0	1	-360	360;
	2629	7507	0.000591	0.004	0	395	0	0	0	0	1	-360	360;
	1137	2406	9.1e-05	0.00051	0	0	0	0	0	0	1	-360	360;
	7577	5814	0.00732	0.039969	0	319	0	0	0	0	1	-360	360;
	7577	3231	0.004581	0.02031	0	529	0	0	0	0	1	-360	360;
	7663	5814	0.00132	0.00537	0	357	0	0	0	0	1	-360	360;
	1183	1159	0.007521	0.0387	0	376	0	0	0	0	1	-360	360;
	6852	5051	0.005159	0.031281	0	433	0	0	0	0	1	-360	360;
	8307	8807	9.9e-05	0.001159	0	0	0	0	0	0	1	-360	360;
	6947	8807	9.9e-05	0.00113	0	0	0	0	0	0	1	-360	360;
	4505	4357	0.00019	0.000909	0	281	0	0	0	0	1	-360	360;
	7144	7791	0.00107	0.00413	0	0	0	0	0	0	1	-360	360;
	6889	2085	9e-05	0.00058	0	1151	0	0	0	0	1	-360	360;
	6889	4024	9e-05	0.00059	0	1052	0	0	0	0	1	-360	360;
	6889	9065	8e-05	0.00053	0	657	0	0	0	0	1	-360	360;
	6889	661	8e-05	0.00052	0	657	0	0	0	0	1	-360	360;
	6612	8864	0.000271	0.00151	0	395	0	0	0	0	1	-360	360;
	8931	6199	0.00053	0.00585	0	1809	0	0	0	0	1	-360	360;
	8931	6199	0.00047	0.00663	0	1348	0	0	0	0	1	-360	360;
	8931	6199	0.00046	0.00676	0	1151	0	0	0	0	1	-360	360;
	8931	6199	0.00045	0.00689	0	1151	0	0	0	0	1	-360	360;
	8931	8494	0.00028	0.00338	0	1480	0	0	0	0	1	-360	360;
	8931	8494	0.0003	0.00319	0	1249	0	0	0	0	1	-360	360;
	1643	4511	0.00025	0.00101	0	376	0	0	0	0	1	-360	360;
	196	1973	0.000281	0.00119	0	0	0	0	0	0	1	-360	360;
	2902	5573	0.00018	0.00063	0	0	0	0	0	0	1	-360	360;
	4950	4410	0.00382	0.02868	0	0	0	0	0	0	1	-360	360;
	4950	333	0.00361	0.0235	0	472	0	0	0	0	1	-360	360;
	4950	6908	0.00538	0.03569	0	0	0	0	0	0	1	-360	360;
	3377	7726	6e-05	0.000409	0	414	0	0	0	0	1	-360	360;
	1027	3243	9e-05	0.00102	0	657	0	0	0	0	1	-360	360;
	1027	2458	0.0014	0.01649	0	1611	0	0	0	0	1	-360	360;
	1027	2919	0.00034	0.00427	0	1019	0	0	0	0	1	-360	360;
	5308	8560	0.0001	0.00109	0	657	0	0	0	0	1	-360	360;
	5308	6427	0.00137	0.01682	0	1677	0	0	0	0	1	-360	360;
	5308	8347	0.00131	0.01262	0	1414	0	0	0	0	1	-360	360;
	7148	5610	0.00262	0.019669	0	433	0	0	0	0	1	-360	360;
	7148	2421	0.00082	0.00526	0	376	0	0	0	0	1	-360	360;
	7148	2421	0.000841	0.005151	0	357	0	0	0	0	1	-360	360;
	7148	2421	0.000669	0.005081	0	376	0	0	0	0	1	-360	360;
	7148	6989	0.001919	0.012669	0	376	0	0	0	0	1	-360	360;
	44	1545	0.002659	0.02	0	0	0	0	0	0	1	-360	360;
	44	3541	0.003031	0.02343	0	472	0	0	0	0	1	-360	360;
	6231	4710	0.01176	0.059159	0	300	0	0	0	0	1	-360	360;
	6231	8367	0.002469	0.017531	0	395	0	0	0	0	1	-360	360;
	6231	2288	0.000669	0.005729	0	0	0	0	0	0	1	-360	360;
	8291	8473	0.000781	0.00876	0	281	0	0	0	0	1	-360	360;
	8291	5455	0.00062	0.00482	0	453	0	0	0	0	1	-360	360;
	8291	7635	0.001349	0.012599	0	433	0	0	0	0	1	-360	360;
	6842	8473	0.000979	0.0082	0	0	0	0	0	0	1	-360	360;
	6842	883	0.008091	0.028831	0	0	0	0	0	0	1	-360	360;
	6842	2288	0.000711	0.005969	0	414	0	0	0	0	1	-360	360;
	3969	7726	0.00145	0.00857	0	0	0	0	0	0	1	-360	360;
	1090	1129	0.00869	0.03768	0	300	0	0	0	0	1	-360	360;
	1090	3070	0.001719	0.00837	0	357	0	0	0	0	1	-360	360;
	8515	5365	0.01582	0.067349	0	338	0	0	0	0	1	-360	360;
	8515	5365	0.01457	0.06469	0	300	0	0	0	0	1	-360	360;
	8515	4529	0.021711	0.090539	0	281	0	0	0	0	1	-360	360;
	8515	3364	0.00019	0.000729	0	300	0	0	0	0	1	-360	360;
	935	5383	0.001271	0.00638	0	300	0	0	0	0	1	-360	360;
	920	6570	3.9e-05	0.0002	0	300	0	0	0	0	1	-360	360;
	6664	4426	0.000281	0.00236	0	529	0	0	0	0	1	-360	360;
	4339	2183	0.00019	0.00329	0	1644	0	0	0	0	1	-360	360;
	7513	4000	0.00035	0.00324	0	0	0	0	0	0	1	-360	360;
	6570	8542	0.00049	0.00349	0	319	0	0	0	0	1	-360	360;
	6570	8542	0.000469	0.00331	0	319	0	0	0	0	1	-360	360;
	6570	1662	0.00313	0.01324	0	0	0	0	0	0	1	-360	360;
	6570	9119	0.001031	0.006719	0	376	0	0	0	0	1	-360	360;
	6570	1151	0.00038	0.00401	0	0	0	0	0	0	1	-360	360;
	502	3589	0.00037	0.002849	0	395	0	0	0	0	1	-360	360;
	8473	1040	0.00032	0.00256	0	376	0	0	0	0	1	-360	360;
	8473	145	0.00037	0.00118	0	281	0	0	0	0	1	-360	360;
	8473	4267	0.00068	0.001151	0	0	0	0	0	0	1	-360	360;
	4936	5179	0.00168	0.009841	0	395	0	0	0	0	1	-360	360;
	5317	8854	0.000599	0.002401	0	300	0	0	0	0	1	-360	360;
	9018	4	0.002169	0.00911	0	300	0	0	0	0	1	-360	360;
	9018	4829	0.002849	0.01282	0	357	0	0	0	0	1	-360	360;
	6253	5383	0.001229	0.01207	0	453	0	0	0	0	1	-360	360;
	6253	7507	0.001289	0.01089	0	510	0	0	0	0	1	-360	360;
	8887	1644	0.00093	0.01036	0	756	0	0	0	0	1	-360	360;
	8887	3483	0.00083	0.01174	0	723	0	0	0	0	1	-360	360;
	6639	7840	0.00118	0.009469	0	0	0	0	0	0	1	-360	360;
	6639	7840	0.001151	0.009651	0	0	0	0	0	0	1	-360	360;
	5461	8109	0.00101	0.01327	0	1480	0	0	0	0	1	-360	360;
	6291	4189	0.00364	0.02486	0	472	0	0	0	0	1	-360	360;
	8405	4747	0.00056	0.00425	0	433	0	0	0	0	1	-360	360;
	8405	4580	0.00369	0.02681	0	433	0	0	0	0	1	-360	360;
	338	7824	0.000531	0.00276	0	300	0	0	0	0	1	-360	360;
	2468	7824	0.000521	0.00281	0	281	0	0	0	0	1	-360	360;
	5257	7824	0.000419	0.00232	0	0	0	0	0	0	1	-360	360;
	2189	7824	0.000409	0.00237	0	0	0	0	0	0	1	-360	360;
	8458	8267	0.00014	0.00193	0	1414	0	0	0	0	1	-360	360;
	2676	5987	0.003211	0.01481	0	319	0	0	0	0	1	-360	360;
	1187	5712	0.00381	0.01175	0	0	0	0	0	0	1	-360	360;
	1187	6521	0.003781	0.02676	0	319	0	0	0	0	1	-360	360;
	2360	8636	0.000919	0.005461	0	453	0	0	0	0	1	-360	360;
	2360	8829	0.000271	0.001211	0	281	0	0	0	0	1	-360	360;
	2360	4235	0.002229	0.017401	0	472	0	0	0	0	1	-360	360;
	2360	4235	0.00243	0.017659	0	510	0	0	0	0	1	-360	360;
	2360	4235	0.002531	0.01705	0	491	0	0	0	0	1	-360	360;
	8818	5099	0.012531	0.05451	0	0	0	0	0	0	1	-360	360;
	5781	4734	0.00087	0.01212	0	1249	0	0	0	0	1	-360	360;
	5781	905	0.0006	0.00802	0	1611	0	0	0	0	1	-360	360;
	5781	8334	0.00072	0.00667	0	1743	0	0	0	0	1	-360	360;
	5781	2093	0.00234	0.02626	0	1315	0	0	0	0	1	-360	360;
	6772	6989	0.00038	0.002969	0	414	0	0	0	0	1	-360	360;
	7885	3912	0.00024	0.00045	0	0	0	0	0	0	1	-360	360;
	6730	3912	0.000229	0.000409	0	0	0	0	0	0	1	-360	360;
	4301	3912	0.000281	0.000401	0	0	0	0	0	0	1	-360	360;
	5362	5610	0.001211	0.005271	0	0	0	0	0	0	1	-360	360;
	5362	350	0.000281	0.00107	0	338	0	0	0	0	1	-360	360;
	5362	6763	0.00025	0.001091	0	0	0	0	0	0	1	-360	360;
	3855	1973	0.00019	0.00081	0	529	0	0	0	0	1	-360	360;
	3855	1973	0.00045	0.00136	0	491	0	0	0	0	1	-360	360;
	3481	554	0.003461	0.023289	0	0	0	0	0	0	1	-360	360;
	5460	8222	0.0013	0.01609	0	1743	0	0	0	0	1	-360	360;
	5460	8222	0.00128	0.0164	0	1611	0	0	0	0	1	-360	360;
	5460	2721	0.00089	0.01082	0	1480	0	0	0	0	1	-360	360;
	5460	2721	0.00087	0.01102	0	1545	0	0	0	0	1	-360	360;
	3786	8578	0.003349	0.02514	0	433	0	0	0	0	1	-360	360;
	490	2653	0.0003	0.00061	0	0	0	0	0	0	1	-360	360;
	148	7069	0.000289	0.001211	0	414	0	0	0	0	1	-360	360;
	7256	8743	0.01301	0.044091	0	338	0	0	0	0	1	-360	360;
	7256	1742	0.02162	0.065969	0	300	0	0	0	0	1	-360	360;
	8930	314	0.006461	0.05275	0	414	0	0	0	0	1	-360	360;
	6954	7974	0.0002	0.000271	0	0	0	0	0	0	1	-360	360;
	3325	7974	0.00018	0.00026	0	0	0	0	0	0	1	-360	360;
	5627	1742	0.00032	0.00188	0	0	0	0	0	0	1	-360	360;
	5610	6989	0.00105	0.006591	0	376	0	0	0	0	1	-360	360;
	21	6163	0.00013	0.00055	0	0	0	0	0	0	1	-360	360;
	3749	5365	0.006151	0.04013	0	529	0	0	0	0	1	-360	360;
	3749	6982	0.007159	0.040591	0	433	0	0	0	0	1	-360	360;
	2291	3133	0.00874	0.048461	0	300	0	0	0	0	1	-360	360;
	2291	4738	0.004159	0.01606	0	0	0	0	0	0	1	-360	360;
	2291	3656	0.002919	0.01912	0	433	0	0	0	0	1	-360	360;
	314	4300	0.00094	0.00394	0	319	0	0	0	0	1	-360	360;
	314	5789	0.000901	0.00395	0	0	0	0	0	0	1	-360	360;
	314	2467	0.003159	0.017789	0	472	0	0	0	0	1	-360	360;
	314	2467	0.002419	0.018331	0	0	0	0	0	0	1	-360	360;
	314	5256	0.00375	0.018281	0	319	0	0	0	0	1	-360	360;
	3200	2653	0.00025	0.001031	0	319	0	0	0	0	1	-360	360;
	4679	4852	0.00264	0.01238	0	319	0	0	0	0	1	-360	360;
	4679	4852	0.001289	0.00795	0	357	0	0	0	0	1	-360	360;
	4679	1151	0.00087	0.003081	0	376	0	0	0	0	1	-360	360;
	8568	5215	0.000401	0.00224	0	319	0	0	0	0	1	-360	360;
	58	221	0.0003	0.00385	0	0	0	0	0	0	1	-360	360;
	6807	851	0.00126	0.008219	0	433	0	0	0	0	1	-360	360;
	150	5099	0.01905	0.055729	0	0	0	0	0	0	1	-360	360;
	150	3558	0.00201	0.00444	0	0	0	0	0	0	1	-360	360;
	7209	2934	0.00101	0.00324	0	0	0	0	0	0	1	-360	360;
	1465	4787	0.00105	0.00677	0	1118	0	0	0	0	1	-360	360;
	1465	3830	0.00039	0.00502	0	1052	0	0	0	0	1	-360	360;
	1465	5648	0.0004	0.00492	0	854	0	0	0	0	1	-360	360;
	1465	4000	0.00051	0.00617	0	1249	0	0	0	0	1	-360	360;
	7537	3485	0.0003	0.00314	0	1480	0	0	0	0	1	-360	360;
	7537	3830	0.00051	0.00668	0	1545	0	0	0	0	1	-360	360;
	7537	5648	0.0005	0.00678	0	1677	0	0	0	0	1	-360	360;
	1262	2424	0.002909	0.017	0	0	0	0	0	0	1	-360	360;
	1262	1156	0.00245	0.011281	0	376	0	0	0	0	1	-360	360;
	1262	4914	0.0027	0.01532	0	357	0	0	0	0	1	-360	360;
	1262	4914	0.00105	0.011151	0	0	0	0	0	0	1	-360	360;
	1262	3834	0.00036	0.002669	0	0	0	0	0	0	1	-360	360;
	1262	556	0.0018	0.011169	0	357	0	0	0	0	1	-360	360;
	1262	3019	0.000711	0.007771	0	0	0	0	0	0	1	-360	360;
	1262	7923	0.00036	0.00257	0	0	0	0	0	0	1	-360	360;
	1551	4103	0.00201	0.01195	0	510	0	0	0	0	1	-360	360;
	1551	7862	0.001091	0.00731	0	472	0	0	0	0	1	-360	360;
	1551	5458	0.00105	0.007591	0	529	0	0	0	0	1	-360	360;
	1551	1838	0.00299	0.02256	0	376	0	0	0	0	1	-360	360;
	3070	2888	0.00036	0.00288	0	0	0	0	0	0	1	-360	360;
	4874	4504	0.00589	0.045599	0	453	0	0	0	0	1	-360	360;
	4874	3579	0.00157	0.0127	0	376	0	0	0	0	1	-360	360;
	4874	3579	0.00168	0.011969	0	0	0	0	0	0	1	-360	360;
	4874	7903	0.00289	0.021031	0	338	0	0	0	0	1	-360	360;
	6368	1102	0.00395	0.025211	0	472	0	0	0	0	1	-360	360;
	6368	9217	0.002771	0.021659	0	376	0	0	0	0	1	-360	360;
	6368	5653	0.002979	0.016901	0	453	0	0	0	0	1	-360	360;
	8478	5907	0.007419	0.034021	0	0	0	0	0	0	1	-360	360;
	5241	6921	0.00056	0.00813	0	789	0	0	0	0	1	-360	360;
	4	1813	0.00361	0.015211	0	338	0	0	0	0	1	-360	360;
	8950	7049	0.000409	0.003331	0	529	0	0	0	0	1	-360	360;
	6846	8788	0.004091	0.019599	0	376	0	0	0	0	1	-360	360;
	6846	6734	0.00294	0.012841	0	395	0	0	0	0	1	-360	360;
	4426	5616	0.020771	0.080479	0	453	0	0	0	0	1	-360	360;
	4426	1965	0.00662	0.0513	0	453	0	0	0	0	1	-360	360;
	4426	1767	0.00607	0.04581	0	433	0	0	0	0	1	-360	360;
	3037	1102	0.000669	0.0042	0	376	0	0	0	0	1	-360	360;
	3037	8485	0.001271	0.007419	0	376	0	0	0	0	1	-360	360;
	3037	5653	0.001599	0.012271	0	472	0	0	0	0	1	-360	360;
	5395	4661	0.009599	0.05024	0	319	0	0	0	0	1	-360	360;
	5395	520	0.00149	0.009219	0	338	0	0	0	0	1	-360	360;
	5395	6816	0.01488	0.06724	0	300	0	0	0	0	1	-360	360;
	5395	5918	0.003599	0.024031	0	319	0	0	0	0	1	-360	360;
	5395	5918	0.00519	0.024	0	338	0	0	0	0	1	-360	360;
	8636	1380	0.003461	0.013349	0	300	0	0	0	0	1	-360	360;
	3436	4239	0.00105	0.01236	0	0	0	0	0	0	1	-360	360;
	3436	4239	0.00199	0.012331	0	395	0	0	0	0	1	-360	360;
	3436	3656	0.00711	0.044711	0	338	0	0	0	0	1	-360	360;
	6203	7124	0.006729	0.02543	0	0	0	0	0	0	1	-360	360;
	9059	7124	0.005771	0.02999	0	0	0	0	0	0	1	-360	360;
	5994	96	0.0048	0.021771	0	0	0	0	0	0	1	-360	360;
	5049	306	0.00286	0.02212	0	433	0	0	0	0	1	-360	360;
	718	8397	0.00351	0.02175	0	414	0	0	0	0	1	-360	360;
	718	6982	0.008401	0.0448	0	472	0	0	0	0	1	-360	360;
	7202	8807	0.0018	0.01187	0	0	0	0	0	0	1	-360	360;
	7202	8807	0.001831	0.01163	0	0	0	0	0	0	1	-360	360;
	8592	2995	0.00075	0.00538	0	0	0	0	0	0	1	-360	360;
	6581	823	2e-05	0.00033	0	0	0	0	0	0	1	-360	360;
	6581	1680	1e-05	0.00034	0	1710	0	0	0	0	1	-360	360;
	6199	953	0.00023	0.00258	0	723	0	0	0	0	1	-360	360;
	6199	1817	0.00023	0.00252	0	657	0	0	0	0	1	-360	360;
	3231	1504	0.002031	0.008531	0	0	0	0	0	0	1	-360	360;
	3231	7226	0.005349	0.020789	0	319	0	0	0	0	1	-360	360;
	2928	7865	0.0002	0.00082	0	395	0	0	0	0	1	-360	360;
	3916	6891	0.002091	0.011219	0	0	0	0	0	0	1	-360	360;
	6053	171	0.00262	0.016979	0	433	0	0	0	0	1	-360	360;
	6053	171	0.002669	0.01663	0	491	0	0	0	0	1	-360	360;
	8829	3044	0.00025	0.00118	0	414	0	0	0	0	1	-360	360;
	1552	1194	0.00056	0.00715	0	1545	0	0	0	0	1	-360	360;
	3912	3740	0.0007	0.00455	0	338	0	0	0	0	1	-360	360;
	3912	2695	0.000331	0.001711	0	281	0	0	0	0	1	-360	360;
	3912	2695	0.00011	0.000781	0	453	0	0	0	0	1	-360	360;
	3912	8877	0.00111	0.00664	0	338	0	0	0	0	1	-360	360;
	7520	8222	0.00066	0.00838	0	1447	0	0	0	0	1	-360	360;
	7520	8222	0.00064	0.00854	0	1447	0	0	0	0	1	-360	360;
	2421	3610	0.003091	0.017091	0	472	0	0	0	0	1	-360	360;
	2421	1414	0.00643	0.03687	0	529	0	0	0	0	1	-360	360;
	8879	3610	0.00275	0.01937	0	0	0	0	0	0	1	-360	360;
	8397	5918	0.00357	0.021289	0	414	0	0	0	0	1	-360	360;
	1914	6982	0.000289	0.002789	0	0	0	0	0	0	1	-360	360;
	7873	6638	0.002531	0.01345	0	281	0	0	0	0	1	-360	360;
	4562	7702	0.00162	0.012159	0	414	0	0	0	0	1	-360	360;
	4562	7702	0.00262	0.01168	0	319	0	0	0	0	1	-360	360;
	8835	1156	0.00099	0.006081	0	0	0	0	0	0	1	-360	360;
	8835	1233	0.000909	0.005979	0	0	0	0	0	0	1	-360	360;
	3390	408	0.00011	0.0008	0	0	0	0	0	0	1	-360	360;
	3390	7231	0.00011	0.00068	0	0	0	0	0	0	1	-360	360;
	3390	5993	0.00207	0.008341	0	395	0	0	0	0	1	-360	360;
	1910	2550	0.00019	0.001031	0	0	0	0	0	0	1	-360	360;
	3975	7353	0.014039	0.0667	0	0	0	0	0	0	1	-360	360;
	3126	7495	0.004961	0.03043	0	338	0	0	0	0	1	-360	360;
	2689	5753	0.00056	0.003531	0	0	0	0	0	0	1	-360	360;
	1201	4235	0.001461	0.00889	0	414	0	0	0	0	1	-360	360;
	7626	7076	6e-05	0.000419	0	395	0	0	0	0	1	-360	360;
	4823	4060	0.00139	0.005169	0	0	0	0	0	0	1	-360	360;
	4141	1311	0.00069	0.00825	0	657	0	0	0	0	1	-360	360;
	7124	4566	0.004969	0.030781	0	0	0	0	0	0	1	-360	360;
	7124	8873	0.00012	0.000969	0	491	0	0	0	0	1	-360	360;
	6684	5383	0.001091	0.010159	0	472	0	0	0	0	1	-360	360;
	8151	5383	0.000969	0.00887	0	453	0	0	0	0	1	-360	360;
	4357	6619	0.00099	0.006789	0	414	0	0	0	0	1	-360	360;
	8191	8765	0.001031	0.00586	0	281	0	0	0	0	1	-360	360;
	7466	5469	0.00401	0.03345	0	529	0	0	0	0	1	-360	360;
	3401	5351	9.1e-05	0.000349	0	0	0	0	0	0	1	-360	360;
	3760	225	0.00125	0.00745	0	491	0	0	0	0	1	-360	360;
	53	4031	0.003711	0.024219	0	414	0	0	0	0	1	-360	360;
	8222	6921	0.00169	0.02156	0	1578	0	0	0	0	1	-360	360;
	8222	6921	0.00171	0.02114	0	1677	0	0	0	0	1	-360	360;
	6556	4239	0.00088	0.00562	0	433	0	0	0	0	1	-360	360;
	6556	4239	0.0008	0.00618	0	472	0	0	0	0	1	-360	360;
	2940	7178	0.00161	0.01014	0	0	0	0	0	0	1	-360	360;
	3928	2563	0.00032	0.002031	0	433	0	0	0	0	1	-360	360;
	3928	6521	0.00011	0.000659	0	395	0	0	0	0	1	-360	360;
	1923	3830	0.00042	0.00505	0	986	0	0	0	0	1	-360	360;
	1923	5648	0.00043	0.00494	0	1151	0	0	0	0	1	-360	360;
	2815	4914	0.00038	0.000711	0	0	0	0	0	0	1	-360	360;
	2815	4914	0.00037	0.000659	0	0	0	0	0	0	1	-360	360;
	4125	1644	0.00248	0.03096	0	1743	0	0	0	0	1	-360	360;
	163	3498	0.00375	0.019729	0	395	0	0	0	0	1	-360	360;
	163	2430	0.00518	0.020021	0	338	0	0	0	0	1	-360	360;
	7943	9108	0.000771	0.003711	0	395	0	0	0	0	1	-360	360;
	7943	1478	0.00076	0.003789	0	357	0	0	0	0	1	-360	360;
	2308	3541	0.000781	0.00574	0	0	0	0	0	0	1	-360	360;
	3535	7582	0.00089	0.00562	0	0	0	0	0	0	1	-360	360;
	2019	8542	0.000331	0.001401	0	0	0	0	0	0	1	-360	360;
	5174	8542	0.000401	0.00137	0	0	0	0	0	0	1	-360	360;
	7945	5215	0.00051	0.003349	0	0	0	0	0	0	1	-360	360;
	2057	9130	0.000461	0.00168	0	0	0	0	0	0	1	-360	360;
	2057	9130	0.000469	0.00164	0	0	0	0	0	0	1	-360	360;
	4661	6816	0.006081	0.03114	0	357	0	0	0	0	1	-360	360;
	7519	776	0.005831	0.02218	0	395	0	0	0	0	1	-360	360;
	8989	3435	0.00013	0.000461	0	0	0	0	0	0	1	-360	360;
	7797	2020	0.00013	0.000461	0	0	0	0	0	0	1	-360	360;
	871	4031	0.00262	0.020289	0	414	0	0	0	0	1	-360	360;
	871	2878	0.00164	0.01194	0	453	0	0	0	0	1	-360	360;
	7353	9164	0.006479	0.02845	0	414	0	0	0	0	1	-360	360;
	7974	2012	0.00294	0.014419	0	395	0	0	0	0	1	-360	360;
	7974	2012	0.00238	0.01463	0	395	0	0	0	0	1	-360	360;
	7974	1750	0.002219	0.00993	0	376	0	0	0	0	1	-360	360;
	1401	2303	0.008651	0.04886	0	300	0	0	0	0	1	-360	360;
	4504	7903	0.00464	0.030979	0	0	0	0	0	0	1	-360	360;
	5837	449	0.00092	0.00889	0	1447	0	0	0	0	1	-360	360;
	5837	449	0.00093	0.00869	0	1348	0	0	0	0	1	-360	360;
	5837	1817	0.0022	0.03485	0	1578	0	0	0	0	1	-360	360;
	5837	2458	0.00074	0.00941	0	1743	0	0	0	0	1	-360	360;
	5837	6427	0.00087	0.00884	0	1578	0	0	0	0	1	-360	360;
	5837	1730	0.00139	0.01662	0	0	0	0	0	0	1	-360	360;
	9217	594	0.002781	0.022979	0	300	0	0	0	0	1	-360	360;
	6351	5393	0.000521	0.00407	0	529	0	0	0	0	1	-360	360;
	6194	7702	0.008479	0.032841	0	0	0	0	0	0	1	-360	360;
	2503	7491	0.005669	0.02901	0	0	0	0	0	0	1	-360	360;
	4031	2872	0.001031	0.004099	0	0	0	0	0	0	1	-360	360;
	4031	8043	0.00301	0.018719	0	376	0	0	0	0	1	-360	360;
	1813	2866	0.00257	0.01713	0	338	0	0	0	0	1	-360	360;
	5365	5417	0.010461	0.03836	0	0	0	0	0	0	1	-360	360;
	5365	4100	0.001719	0.011531	0	414	0	0	0	0	1	-360	360;
	5365	1556	9.1e-05	0.00056	0	300	0	0	0	0	1	-360	360;
	3894	3865	0.00476	0.025289	0	0	0	0	0	0	1	-360	360;
	128	1554	0.000219	0.00164	0	319	0	0	0	0	1	-360	360;
	198	7994	0.002159	0.01163	0	0	0	0	0	0	1	-360	360;
	198	1554	0.00295	0.01539	0	300	0	0	0	0	1	-360	360;
	2985	3133	0.007789	0.033039	0	300	0	0	0	0	1	-360	360;
	7316	3071	0.00026	0.0013	0	0	0	0	0	0	1	-360	360;
	6486	8992	0.00162	0.013229	0	338	0	0	0	0	1	-360	360;
	8485	594	0.007919	0.04194	0	338	0	0	0	0	1	-360	360;
	682	8477	0.001229	0.00401	0	300	0	0	0	0	1	-360	360;
	5146	5488	0.008021	0.042831	0	338	0	0	0	0	1	-360	360;
	6638	776	0.007711	0.03006	0	338	0	0	0	0	1	-360	360;
	8788	789	0.00261	0.015469	0	319	0	0	0	0	1	-360	360;
	6332	1556	0.00139	0.00536	0	0	0	0	0	0	1	-360	360;
	6332	1556	0.001419	0.00525	0	0	0	0	0	0	1	-360	360;
	1883	7702	0.006289	0.032271	0	0	0	0	0	0	1	-360	360;
	1883	2132	0.011031	0.05625	0	0	0	0	0	0	1	-360	360;
	171	2341	0.001841	0.01324	0	472	0	0	0	0	1	-360	360;
	6630	10	0.000599	0.00186	0	453	0	0	0	0	1	-360	360;
	4484	4239	0.00036	0.001591	0	433	0	0	0	0	1	-360	360;
	3044	4235	0.003849	0.027651	0	453	0	0	0	0	1	-360	360;
	851	4454	0.00186	0.00874	0	491	0	0	0	0	1	-360	360;
	3485	2183	0.00025	0.00285	0	1480	0	0	0	0	1	-360	360;
	1156	7923	0.00207	0.008841	0	319	0	0	0	0	1	-360	360;
	1156	1233	6e-05	0.00039	0	453	0	0	0	0	1	-360	360;
	8913	4852	0.002159	0.008909	0	357	0	0	0	0	1	-360	360;
	4039	1233	0.000219	0.00039	0	0	0	0	0	0	1	-360	360;
	4914	2653	0.00011	0.000521	0	300	0	0	0	0	1	-360	360;
	4914	2653	0.00011	0.00049	0	300	0	0	0	0	1	-360	360;
	4914	556	0.000771	0.003789	0	338	0	0	0	0	1	-360	360;
	4914	3019	0.000229	0.002849	0	0	0	0	0	0	1	-360	360;
	7994	5256	0.00445	0.027599	0	0	0	0	0	0	1	-360	360;
	1709	3346	0.00218	0.02	0	1381	0	0	0	0	1	-360	360;
	1709	8468	0.00074	0.00865	0	1118	0	0	0	0	1	-360	360;
	7702	1234	0.002531	0.019419	0	300	0	0	0	0	1	-360	360;
	7702	8035	0.003	0.01631	0	0	0	0	0	0	1	-360	360;
	8677	3331	0.008461	0.025081	0	0	0	0	0	0	1	-360	360;
	8189	3643	0.00382	0.013669	0	395	0	0	0	0	1	-360	360;
	3133	8477	0.002581	0.01814	0	414	0	0	0	0	1	-360	360;
	6744	6101	0.00201	0.012219	0	433	0	0	0	0	1	-360	360;
	8542	4313	0.00087	0.001599	0	0	0	0	0	0	1	-360	360;
	769	5233	0.00455	0.02394	0	319	0	0	0	0	1	-360	360;
	1965	5926	0.005651	0.041771	0	357	0	0	0	0	1	-360	360;
	9014	3999	0.00069	0.001271	0	0	0	0	0	0	1	-360	360;
	1917	8406	0.00068	0.00269	0	300	0	0	0	0	1	-360	360;
	1917	2021	0.00036	0.00261	0	376	0	0	0	0	1	-360	360;
	1917	8293	0.001669	0.012031	0	319	0	0	0	0	1	-360	360;
	4738	3656	0.009669	0.03513	0	0	0	0	0	0	1	-360	360;
	4738	8477	0.001099	0.00862	0	0	0	0	0	0	1	-360	360;
	7998	1644	0.0005	0.00711	0	657	0	0	0	0	1	-360	360;
	6734	1015	0.003919	0.017781	0	395	0	0	0	0	1	-360	360;
	6734	8992	0.00161	0.019781	0	0	0	0	0	0	1	-360	360;
	6734	789	0.0007	0.01256	0	510	0	0	0	0	1	-360	360;
	5004	5393	0.000409	0.0018	0	0	0	0	0	0	1	-360	360;
	1708	3906	2e-05	0.00043	0	0	0	0	0	0	1	-360	360;
	9150	3906	3e-05	0.0004	0	1710	0	0	0	0	1	-360	360;
	280	9158	0.00968	0.056659	0	357	0	0	0	0	1	-360	360;
	8043	3656	0.003969	0.02925	0	0	0	0	0	0	1	-360	360;
	5712	8992	0.00219	0.01468	0	414	0	0	0	0	1	-360	360;
	5712	8466	0.000401	0.003159	0	376	0	0	0	0	1	-360	360;
	5712	6880	0.000531	0.003	0	376	0	0	0	0	1	-360	360;
	5215	7266	0.00136	0.00576	0	376	0	0	0	0	1	-360	360;
	8807	8992	0.000849	0.006289	0	357	0	0	0	0	1	-360	360;
	8807	8992	0.000841	0.006409	0	357	0	0	0	0	1	-360	360;
	8807	8992	0.00045	0.00506	0	0	0	0	0	0	1	-360	360;
	218	2088	0.003669	0.02988	0	395	0	0	0	0	1	-360	360;
	520	5918	0.002	0.01568	0	414	0	0	0	0	1	-360	360;
	2653	2128	0.00031	0.0008	0	319	0	0	0	0	1	-360	360;
	1216	5233	0.00043	0.00038	0	0	0	0	0	0	1	-360	360;
	8980	5993	0.00043	0.002229	0	0	0	0	0	0	1	-360	360;
	5212	5993	0.00049	0.00218	0	0	0	0	0	0	1	-360	360;
	3656	4060	0.00282	0.01844	0	529	0	0	0	0	1	-360	360;
	883	3331	0.00462	0.014849	0	0	0	0	0	0	1	-360	360;
	5144	3346	0.00263	0.04076	0	1447	0	0	0	0	1	-360	360;
	3498	2563	0.01062	0.056719	0	357	0	0	0	0	1	-360	360;
	1896	8200	0.0001	0.00105	0	657	0	0	0	0	1	-360	360;
	9130	8864	0.00239	0.017341	0	376	0	0	0	0	1	-360	360;
	3503	2510	0.001341	0.010581	0	433	0	0	0	0	1	-360	360;
	3503	6146	0.00101	0.007599	0	472	0	0	0	0	1	-360	360;
	401	1584	0.00011	0.000289	0	395	0	0	0	0	1	-360	360;
	7328	6921	0.00146	0.01769	0	0	0	0	0	0	1	-360	360;
	7328	6921	0.0013	0.02004	0	1644	0	0	0	0	1	-360	360;
	6816	5573	0.007909	0.028349	0	319	0	0	0	0	1	-360	360;
	7396	8564	0.00012	0.00049	0	0	0	0	0	0	1	-360	360;
	7396	8564	0.00012	0.000479	0	0	0	0	0	0	1	-360	360;
	6921	432	0.00076	0.01018	0	1513	0	0	0	0	1	-360	360;
	6921	432	0.0009	0.00865	0	1677	0	0	0	0	1	-360	360;
	6921	432	0.00096	0.01294	0	1545	0	0	0	0	1	-360	360;
	6921	432	0.0004	0.00756	0	1611	0	0	0	0	1	-360	360;
	6921	432	0.00039	0.00771	0	1447	0	0	0	0	1	-360	360;
	6921	8200	0.00163	0.02291	0	1545	0	0	0	0	1	-360	360;
	2949	3121	0.00636	0.03855	0	395	0	0	0	0	1	-360	360;
	2949	3121	0.00639	0.03863	0	414	0	0	0	0	1	-360	360;
	953	1730	0.00113	0.01362	0	690	0	0	0	0	1	-360	360;
	1767	892	0.00861	0.05318	0	529	0	0	0	0	1	-360	360;
	4674	4889	0.0012	0.004669	0	338	0	0	0	0	1	-360	360;
	4674	7076	0.00018	0.001091	0	414	0	0	0	0	1	-360	360;
	1234	8035	0.001401	0.00612	0	357	0	0	0	0	1	-360	360;
	115	225	0.00068	0.00387	0	491	0	0	0	0	1	-360	360;
	5720	8578	0.0043	0.033591	0	414	0	0	0	0	1	-360	360;
	3579	5469	0.003039	0.01911	0	491	0	0	0	0	1	-360	360;
	3246	124	6e-05	0.00111	0	853	0	0	0	0	1	-360	360;
	8846	4783	5e-05	0.001091	0	853	0	0	0	0	1	-360	360;
	2393	3306	3.9e-05	0.00089	0	853	0	0	0	0	1	-360	360;
	1895	4480	3.9e-05	0.000901	0	853	0	0	0	0	1	-360	360;
	678	4454	0.000169	0.00111	0	510	0	0	0	0	1	-360	360;
	1002	4852	0.000331	0.0027	0	0	0	0	0	0	1	-360	360;
	6429	2012	0.00026	0.00369	0	0	0	0	0	0	1	-360	360;
	4889	7076	0.000341	0.00274	0	0	0	0	0	0	1	-360	360;
	5341	7491	0.00038	0.00506	0	414	0	0	0	0	1	-360	360;
	2563	6521	0.00012	0.000831	0	376	0	0	0	0	1	-360	360;
	1111	594	0.003031	0.0123	0	0	0	0	0	0	1	-360	360;
	4747	7791	0.001521	0.00774	0	395	0	0	0	0	1	-360	360;
	3906	3022	0.00059	0.00552	0	1611	0	0	0	0	1	-360	360;
	3906	1194	0.00053	0.00625	0	1545	0	0	0	0	1	-360	360;
	5993	8804	0.00143	0.01076	0	529	0	0	0	0	1	-360	360;
	1798	8487	0.00056	0.00593	0	0	0	0	0	0	1	-360	360;
	2467	7178	0.00205	0.01461	0	414	0	0	0	0	1	-360	360;
	2132	2303	0.01068	0.054521	0	0	0	0	0	0	1	-360	360;
	8992	8466	0.00194	0.011039	0	433	0	0	0	0	1	-360	360;
	8992	6880	0.00131	0.010099	0	433	0	0	0	0	1	-360	360;
	8992	2995	0.001289	0.014229	0	319	0	0	0	0	1	-360	360;
	805	3221	0.00455	0.027031	0	395	0	0	0	0	1	-360	360;
	805	3327	0.00239	0.01414	0	510	0	0	0	0	1	-360	360;
	805	8310	0.004039	0.029669	0	491	0	0	0	0	1	-360	360;
	805	2088	0.001099	0.007771	0	529	0	0	0	0	1	-360	360;
	5383	2021	0.00082	0.00762	0	414	0	0	0	0	1	-360	360;
	5383	4114	0.000719	0.0033	0	0	0	0	0	0	1	-360	360;
	5383	5519	0.000599	0.00361	0	300	0	0	0	0	1	-360	360;
	5383	8293	0.000521	0.004039	0	376	0	0	0	0	1	-360	360;
	5383	3435	0.000901	0.0088	0	433	0	0	0	0	1	-360	360;
	7507	6990	0.000781	0.00306	0	0	0	0	0	0	1	-360	360;
	7507	2406	0.000419	0.00399	0	0	0	0	0	0	1	-360	360;
	7507	2020	0.000841	0.009841	0	281	0	0	0	0	1	-360	360;
	6478	6146	0.004021	0.03064	0	414	0	0	0	0	1	-360	360;
	4852	2695	0.0023	0.009031	0	338	0	0	0	0	1	-360	360;
	4852	2695	0.002081	0.009469	0	376	0	0	0	0	1	-360	360;
	4852	5502	0.00164	0.008831	0	0	0	0	0	0	1	-360	360;
	4852	8877	0.00094	0.00488	0	319	0	0	0	0	1	-360	360;
	4852	9119	0.000901	0.00551	0	453	0	0	0	0	1	-360	360;
	4852	1151	0.00051	0.005961	0	0	0	0	0	0	1	-360	360;
	2012	1750	0.00113	0.004349	0	433	0	0	0	0	1	-360	360;
	1478	6616	0.002081	0.013031	0	395	0	0	0	0	1	-360	360;
	809	4594	0.00116	0.01438	0	0	0	0	0	0	1	-360	360;
	8656	4729	0.00013	0.001031	0	0	0	0	0	0	1	-360	360;
	4100	615	0.00036	0.002659	0	414	0	0	0	0	1	-360	360;
	4060	5393	0.000771	0.004539	0	529	0	0	0	0	1	-360	360;
	9203	2426	0.000695	0.049098	0	0	0	0	0.982143	0	1	-360	360;
	9203	2426	0.000355	0.024893	0	591	0	0	0.93617	0	1	-360	360;
	118	2426	0.000365	0.024372	0	591	0	0	0.979227	0	1	-360	360;
	3145	7770	0.000411	0.023652	0	591	0	0	0.93617	0	1	-360	360;
	7267	4918	0.000386	0.025673	0	591	0	0	0.93617	0	1	-360	360;
	7267	4918	0.000344	0.029336	0	591	0	0	0.93617	0	1	-360	360;
	839	4506	0.000653	0.052104	0	0	0	0	0.982143	0	1	-360	360;
	839	4506	0.000767	0.043531	0	0	0	0	0.982143	0	1	-360	360;
	3445	6114	0.000323	0.028314	0	591	0	0	0.93617	0	1	-360	360;
	3445	6114	0.000323	0.025999	0	591	0	0	0.93617	0	1	-360	360;
	3445	7264	0.000365	0.025478	0	591	0	0	0.93617	0	1	-360	360;
	891	6306	0.00086	0.039291	0	0	0	0	0.982143	0	1	-360	360;
	720	1711	0.000355	0.026586	0	591	0	0	0.93617	0	1	-360	360;
	720	1711	0.000375	0.026313	0	591	0	0	0.93617	0	1	-360	360;
	6206	3240	0.000344	0.023329	0	591	0	0	0.979227	0	1	-360	360;
	6206	3240	0.000302	0.026375	0	591	0	0	0.979227	0	1	-360	360;
	9174	2924	0.000788	0.04323	0	0	0	0	0.982143	0	1	-360	360;
	9174	2924	0.000767	0.04407	0	0	0	0	0.982143	0	1	-360	360;
	1355	5856	0.000767	0.043562	0	0	0	0	0.982143	0	1	-360	360;
	1355	5856	0.000788	0.041281	0	0	0	0	0.982143	0	1	-360	360;
	1355	5856	0.000892	0.040524	0	0	0	0	0.982143	0	1	-360	360;
	3205	4683	0.000975	0.043562	0	0	0	0	0.982143	0	1	-360	360;
	3205	4683	0.000757	0.042649	0	0	0	0	0.982143	0	1	-360	360;
	5350	2361	0.000261	0.030762	0	591	0	0	0.93617	0	1	-360	360;
	5350	2361	0.000261	0.030088	0	591	0	0	0.93617	0	1	-360	360;
	5016	6112	0.000323	0.026447	0	591	0	0	0.979227	0	1	-360	360;
	5016	8860	0.000323	0.026885	0	591	0	0	0.979227	0	1	-360	360;
	7471	549	0.000323	0.02748	0	591	0	0	0.93617	0	1	-360	360;
	7471	549	0.000292	0.028158	0	591	0	0	0.93617	0	1	-360	360;
	549	5002	0	0.009197	0	567	0	0	0	0.072386	1	-360	360;
	5490	3817	0.000736	0.044526	0	0	0	0	0.982143	0	1	-360	360;
	2854	9112	0.00086	0.039488	0	0	0	0	0.982143	0	1	-360	360;
	2918	6532	0.000365	0.024372	0	591	0	0	0.93617	0	1	-360	360;
	2918	6532	0.000375	0.023986	0	591	0	0	0.93617	0	1	-360	360;
	4494	7284	0.000386	0.025673	0	591	0	0	0.93617	0	1	-360	360;
	2656	1775	0.000809	0.042318	0	0	0	0	0.982143	0	1	-360	360;
	3519	5641	0.000705	0.042038	0	0	0	0	0.982143	0	1	-360	360;
	3543	367	0.000334	0.028022	0	591	0	0	0.93617	0	1	-360	360;
	3543	1833	0.000353	0.022403	0	591	0	0	0.982143	0	1	-360	360;
	2372	367	0.000313	0.026354	0	591	0	0	0.93617	0	1	-360	360;
	2372	367	0.000323	0.025478	0	591	0	0	0.93617	0	1	-360	360;
	1876	7752	0.000344	0.024622	0	591	0	0	0.979227	0	1	-360	360;
	113	7752	0.000355	0.024038	0	591	0	0	0.979227	0	1	-360	360;
	3499	7289	0.000881	0.038244	0	0	0	0	0.982143	0	1	-360	360;
	3499	7289	0.000757	0.040959	0	0	0	0	0.982143	0	1	-360	360;
	3082	2083	0.000354	0.028765	0	591	0	0	0.93617	0	1	-360	360;
	6224	2083	0.000354	0.027114	0	591	0	0	0.93617	0	1	-360	360;
	2174	2083	0.000695	0.04748	0	0	0	0	0.982143	0	1	-360	360;
	4623	1609	0.000323	0.028314	0	591	0	0	0.93617	0	1	-360	360;
	4623	1609	0.000323	0.026354	0	591	0	0	0.93617	0	1	-360	360;
	4623	1609	0.000323	0.025696	0	591	0	0	0.93617	0	1	-360	360;
	5589	1001	0.000355	0.027305	0	591	0	0	0.979227	0	1	-360	360;
	5589	1001	0.000355	0.026666	0	591	0	0	0.979227	0	1	-360	360;
	5589	1001	0.000923	0.037456	0	0	0	0	0.982143	0	1	-360	360;
	5589	1001	0.000788	0.035217	0	0	0	0	0.982143	0	1	-360	360;
	1198	7473	0.000323	0.026739	0	591	0	0	0.93617	0	1	-360	360;
	1198	7473	0.000323	0.02749	0	591	0	0	0.93617	0	1	-360	360;
	2479	5482	0.000313	0.027324	0	591	0	0	0.93617	0	1	-360	360;
	2479	5482	0.000313	0.028074	0	591	0	0	0.93617	0	1	-360	360;
	7831	5525	0.000601	0.024984	0	0	0	0	0.982143	0	1	-360	360;
	7831	5525	0.000715	0.022797	0	0	0	0	0.982143	0	1	-360	360;
	7831	5525	0.000344	0.024893	0	591	0	0	0.93617	0	1	-360	360;
	4970	7955	0.000365	0.024163	0	591	0	0	0.979227	0	1	-360	360;
	4970	7955	0.000344	0.02385	0	591	0	0	0.979227	0	1	-360	360;
	5007	7367	0.000757	0.040182	0	0	0	0	0.982143	0	1	-360	360;
	757	3758	0.000323	0.028937	0	591	0	0	0.979227	0	1	-360	360;
	3697	6952	0.000354	0.026958	0	591	0	0	0.93617	0	1	-360	360;
	3697	6952	0.000778	0.045656	0	0	0	0	0.982143	0	1	-360	360;
	3697	6952	0.000695	0.048932	0	0	0	0	0.982143	0	1	-360	360;
	3697	6952	0.000302	0.025926	0	591	0	0	0.93617	0	1	-360	360;
	6738	8180	0.000355	0.027328	0	591	0	0	0.93617	0	1	-360	360;
	6738	8180	0.000365	0.027156	0	591	0	0	0.93617	0	1	-360	360;
	9222	5935	0.000355	0.024299	0	591	0	0	0.979227	0	1	-360	360;
	7988	804	0.000375	0.026313	0	591	0	0	0.93617	0	1	-360	360;
	7988	804	0.000375	0.025605	0	591	0	0	0.93617	0	1	-360	360;
	3239	1539	0.000334	0.026291	0	591	0	0	0.979227	0	1	-360	360;
	3239	1539	0.000705	0.046735	0	0	0	0	0.982143	0	1	-360	360;
	2056	4084	0.000313	0.026875	0	591	0	0	0.979227	0	1	-360	360;
	2056	4084	0.000302	0.027376	0	591	0	0	0.979227	0	1	-360	360;
	4598	5110	0.000664	0.044868	0	0	0	0	0.982143	0	1	-360	360;
	2129	5110	0.000809	0.044526	0	0	0	0	0.982143	0	1	-360	360;
	5658	5550	0.00086	0.034253	0	0	0	0	0.982143	0	1	-360	360;
	5658	5550	0.000355	0.024372	0	591	0	0	0.93617	0	1	-360	360;
	2848	1526	0.000365	0.023788	0	591	0	0	0.93617	0	1	-360	360;
	4541	7464	0.000375	0.023204	0	591	0	0	0.979227	0	1	-360	360;
	4541	7464	0.000302	0.026593	0	591	0	0	0.979227	0	1	-360	360;
	1754	3857	0.000323	0.029256	0	591	0	0	0.93617	0	1	-360	360;
	1754	3857	0.000334	0.029986	0	591	0	0	0.93617	0	1	-360	360;
	1754	3857	0.000355	0.030979	0	591	0	0	0.93617	0	1	-360	360;
	2518	4544	0.000302	0.025707	0	591	0	0	0.93617	0	1	-360	360;
	2518	4544	0.000344	0.025405	0	591	0	0	0.93617	0	1	-360	360;
	3069	6115	0.00012	0.008908	0	491	0	0	0	-0.072388	1	-360	360;
	5067	7641	0.000365	0.023715	0	591	0	0	0.93617	0	1	-360	360;
	6036	8670	0.000365	0.023517	0	591	0	0	0.979227	0	1	-360	360;
	6036	5522	0.000334	0.026812	0	591	0	0	0.979227	0	1	-360	360;
	8486	8791	0.000695	0.047253	0	0	0	0	0.982143	0	1	-360	360;
	7530	8791	0.000778	0.043759	0	0	0	0	0.982143	0	1	-360	360;
	7522	7571	0.000313	0.030636	0	591	0	0	0.93617	0	1	-360	360;
	7522	7571	0.000313	0.028286	0	591	0	0	0.93617	0	1	-360	360;
	7522	7571	0.000344	0.028115	0	591	0	0	0.93617	0	1	-360	360;
	2526	1483	0.000902	0.039425	0	0	0	0	0.982143	0	1	-360	360;
	2365	8316	0.000355	0.024237	0	591	0	0	0.93617	0	1	-360	360;
	8672	1860	0.000375	0.023986	0	591	0	0	0.93617	0	1	-360	360;
	8672	1860	0.000375	0.023861	0	591	0	0	0.93617	0	1	-360	360;
	1494	583	0.000819	0.04153	0	0	0	0	0.982143	0	1	-360	360;
	1494	583	0.000809	0.042701	0	0	0	0	0.982143	0	1	-360	360;
	1502	2795	0.000334	0.028022	0	591	0	0	0.93617	0	1	-360	360;
	8195	933	0.000632	0.048237	0	0	0	0	0.982143	0	1	-360	360;
	8195	933	0.000632	0.045159	0	0	0	0	0.982143	0	1	-360	360;
	8195	933	0.000715	0.044039	0	0	0	0	0.982143	0	1	-360	360;
	3918	6802	0.00074	0.044479	0	0	0	0	0.986547	0	1	-360	360;
	3918	4025	0.00074	0.043811	0	0	0	0	0.986547	0	1	-360	360;
	4435	2079	0.000375	0.02385	0	591	0	0	0.93617	0	1	-360	360;
	4435	2079	0.000365	0.023267	0	591	0	0	0.93617	0	1	-360	360;
	4435	2079	0.000323	0.026739	0	591	0	0	0.93617	0	1	-360	360;
	1081	7361	0.000684	0.047025	0	0	0	0	0.982143	0	1	-360	360;
	9101	2177	0.000313	0.02748	0	591	0	0	0.93617	0	1	-360	360;
	6555	8475	0.000323	0.028554	0	591	0	0	0.93617	0	1	-360	360;
	6555	8475	0.000302	0.026437	0	591	0	0	0.93617	0	1	-360	360;
	7056	7577	0.000365	0.025551	0	591	0	0	0.93617	0	1	-360	360;
	7056	7577	0.000344	0.024893	0	591	0	0	0.93617	0	1	-360	360;
	7056	9051	0.000375	0.024508	0	591	0	0	0.979227	0	1	-360	360;
	6889	6785	0.000778	0.041157	0	0	0	0	0.982143	0	1	-360	360;
	6889	6785	0.000923	0.03702	0	0	0	0	0.982143	0	1	-360	360;
	6889	6785	0.000302	0.026593	0	591	0	0	0.979227	0	1	-360	360;
	1027	7148	0.000344	0.029667	0	591	0	0	0.93617	0	1	-360	360;
	1027	44	0.000334	0.02748	0	591	0	0	0.93617	0	1	-360	360;
	5308	7148	0.000323	0.030636	0	591	0	0	0.89674	0	1	-360	360;
	5308	44	0.000323	0.028526	0	591	0	0	0.93617	0	1	-360	360;
	9213	6231	0.000344	0.025332	0	591	0	0	0.979227	0	1	-360	360;
	9213	8291	0.000334	0.024674	0	591	0	0	0.979227	0	1	-360	360;
	9213	6842	0.000344	0.02457	0	591	0	0	0.979227	0	1	-360	360;
	3608	6664	0.000365	0.023788	0	591	0	0	0.93617	0	1	-360	360;
	4339	6570	0.000365	0.023601	0	591	0	0	0.93617	0	1	-360	360;
	4339	6570	0.000344	0.02652	0	591	0	0	0.93617	0	1	-360	360;
	7513	6570	0.000344	0.027272	0	591	0	0	0.93617	0	1	-360	360;
	7513	6570	0.000334	0.027563	0	591	0	0	0.93617	0	1	-360	360;
	8887	6639	0.000313	0.030636	0	591	0	0	0.979227	0	1	-360	360;
	8887	6639	0.000313	0.028286	0	591	0	0	0.979227	0	1	-360	360;
	5461	4816	0.000881	0.036803	0	0	0	0	0.982143	0	1	-360	360;
	5461	4816	0.000902	0.036668	0	0	0	0	0.982143	0	1	-360	360;
	5461	4816	0.000881	0.034014	0	0	0	0	0.982143	0	1	-360	360;
	5460	3786	0.000365	0.026027	0	591	0	0	0.93617	0	1	-360	360;
	5460	3786	0.000386	0.025673	0	591	0	0	0.93617	0	1	-360	360;
	7256	4491	0.00012	0.011186	0	586	0	0	0	0.084507	1	-360	360;
	6153	58	0.00017	0.006789	0	1216	0	0	0	-0.086984	1	-360	360;
	6153	6807	0.000664	0.047107	0	0	0	0	0.982143	0	1	-360	360;
	1465	1262	0.000313	0.028314	0	591	0	0	0.93617	0	1	-360	360;
	1465	1262	0.000334	0.025707	0	591	0	0	0.93617	0	1	-360	360;
	1465	1551	0.000344	0.025478	0	591	0	0	0.93617	0	1	-360	360;
	7537	1551	0.000344	0.027236	0	591	0	0	0.93617	0	1	-360	360;
	7537	1551	0.000375	0.026518	0	591	0	0	0.93617	0	1	-360	360;
	6901	4874	0.000386	0.02646	0	591	0	0	0.93617	0	1	-360	360;
	5241	4	0.000788	0.040317	0	0	0	0	0.982143	0	1	-360	360;
	7520	561	0.000323	0.025863	0	591	0	0	0.979227	0	1	-360	360;
	749	4324	1e-05	0.022863	0	948	0	0	0	0.04898	1	-360	360;
	4141	7124	0.000334	0.030306	0	591	0	0	0.93617	0	1	-360	360;
	4141	7124	0.000302	0.030636	0	591	0	0	0.93617	0	1	-360	360;
	7466	3649	0.00011	0.010476	0	472	0	0	0	-0.06138	1	-360	360;
	8222	6556	0.000344	0.024841	0	591	0	0	0.979227	0	1	-360	360;
	8222	6556	0.000365	0.025029	0	591	0	0	0.979227	0	1	-360	360;
	8222	6556	0.000355	0.024101	0	591	0	0	0.979227	0	1	-360	360;
	4125	163	0.000344	0.02385	0	591	0	0	0.93617	0	1	-360	360;
	4125	163	0.000411	0.02313	0	591	0	0	0.93617	0	1	-360	360;
	4125	163	0.000829	0.042764	0	0	0	0	0.982143	0	1	-360	360;
	5837	9217	0.000705	0.04748	0	0	0	0	0.982143	0	1	-360	360;
	5837	1448	0.000323	0.029986	0	591	0	0	0.93617	0	1	-360	360;
	5837	1448	0.000302	0.029952	0	591	0	0	0.93617	0	1	-360	360;
	8267	5365	0.000746	0.041064	0	0	0	0	0.982143	0	1	-360	360;
	8267	5365	0.000881	0.042214	0	0	0	0	0.982143	0	1	-360	360;
	2359	8005	0.00086	0.041209	0	0	0	0	0.982143	0	1	-360	360;
	2359	8005	0.000912	0.038669	0	0	0	0	0.982143	0	1	-360	360;
	8109	171	0.000375	0.026244	0	591	0	0	0.979227	0	1	-360	360;
	8109	171	0.000375	0.025457	0	591	0	0	0.979227	0	1	-360	360;
	3485	1156	0.000302	0.029096	0	591	0	0	0.93617	0	1	-360	360;
	4734	2229	0.000302	0.029667	0	591	0	0	0.93617	0	1	-360	360;
	1709	7702	0.000334	0.027563	0	591	0	0	0.979227	0	1	-360	360;
	1709	7702	0.000323	0.028074	0	591	0	0	0.979227	0	1	-360	360;
	905	8189	0.000736	0.041064	0	0	0	0	0.982143	0	1	-360	360;
	8334	8189	0.001234	0.040109	0	0	0	0	0.982143	0	1	-360	360;
	8334	8189	0.000757	0.043209	0	0	0	0	0.982143	0	1	-360	360;
	7998	6734	0.000323	0.026586	0	854	0	0	0.93617	0	1	-360	360;
	4196	280	0.000788	0.040296	0	0	0	0	0.982143	0	1	-360	360;
	4196	280	0.000912	0.040514	0	0	0	0	0.982143	0	1	-360	360;
	6475	3503	0.000798	0.042183	0	0	0	0	0.982143	0	1	-360	360;
	6475	3503	0.000778	0.043147	0	0	0	0	0.982143	0	1	-360	360;
	7328	4235	0.000323	0.027793	0	591	0	0	0.93617	0	1	-360	360;
	7328	4235	0.000302	0.027918	0	591	0	0	0.93617	0	1	-360	360;
	7328	4235	0.000302	0.025926	0	591	0	0	0.93617	0	1	-360	360;
	6921	2949	0.000344	0.024956	0	591	0	0	0.979227	0	1	-360	360;
	6921	2949	0.000355	0.024435	0	591	0	0	0.979227	0	1	-360	360;
	3830	306	0.000386	0.026392	0	591	0	0	0.93617	0	1	-360	360;
	5648	306	0.000365	0.025525	0	591	0	0	0.93617	0	1	-360	360;
	1730	115	0.000344	0.029986	0	591	0	0	0.93617	0	1	-360	360;
	1730	115	0.000323	0.029838	0	591	0	0	0.93617	0	1	-360	360;
	2721	5720	0.000323	0.02748	0	591	0	0	0.93617	0	1	-360	360;
	2721	5720	0.000313	0.028001	0	591	0	0	0.93617	0	1	-360	360;
	432	3246	0.000179	0.012988	0	1151	0	0	0.990991	0	1	-360	360;
	432	8846	0.000198	0.012724	0	1151	0	0	0.990991	0	1	-360	360;
	432	2393	0.000207	0.012469	0	1151	0	0	0.990991	0	1	-360	360;
	432	1895	0.000207	0.012206	0	1151	0	0	0.990991	0	1	-360	360;
	432	3412	0.000217	0.012912	0	1151	0	0	0.990991	0	1	-360	360;
	432	5586	0.000217	0.012627	0	1151	0	0	0.990991	0	1	-360	360;
	432	3112	0.000198	0.014317	0	1151	0	0	0.990991	0	1	-360	360;
	432	3112	0.000189	0.014592	0	1151	0	0	0.990991	0	1	-360	360;
	432	7857	0.000354	0.02725	0	591	0	0	0.979227	0	1	-360	360;
	960	8497	0.000628	0.042428	0	0	0	0	1.009174	0	1	-360	360;
	960	8497	0.000726	0.044039	0	0	0	0	0.982143	0	1	-360	360;
	3577	5907	0.000365	0.024893	0	591	0	0	0.979227	0	1	-360	360;
	3577	5907	0.000438	0.02457	0	591	0	0	0.979227	0	1	-360	360;
	3577	5907	0.000365	0.023986	0	591	0	0	0.979227	0	1	-360	360;
	7164	5918	0.000375	0.023267	0	591	0	0	0.979227	0	1	-360	360;
	7164	5918	0.000323	0.026375	0	591	0	0	0.93617	0	1	-360	360;
	1798	2467	0.000334	0.029735	0	591	0	0	0.93617	0	1	-360	360;
	1798	2467	0.000313	0.030237	0	591	0	0	0.93617	0	1	-360	360;
	3346	2132	0.000767	0.044049	0	0	0	0	0.982143	0	1	-360	360;
	3346	2132	0.000302	0.026072	0	591	0	0	0.93617	0	1	-360	360;
	3346	2132	0.00085	0.039892	0	0	0	0	0.982143	0	1	-360	360;
	3483	8992	0.000344	0.024956	0	591	0	0	0.979227	0	1	-360	360;
	8886	8992	0.000334	0.026666	0	591	0	0	0.93617	0	1	-360	360;
	2093	805	0.000365	0.023788	0	591	0	0	0.93617	0	1	-360	360;
	2093	805	0.000365	0.023465	0	591	0	0	0.93617	0	1	-360	360;
	221	1541	0.000334	0.026447	0	591	0	0	0.93617	0	1	-360	360;
	8347	5383	0.000323	0.027042	0	591	0	0	0.93617	0	1	-360	360;
	8347	5383	0.000313	0.02748	0	591	0	0	0.93617	0	1	-360	360;
	8347	5383	0.000302	0.028314	0	591	0	0	0.93617	0	1	-360	360;
	8347	7507	0.000323	0.025853	0	591	0	0	0.93617	0	1	-360	360;
	8347	7507	0.000365	0.026197	0	591	0	0	0.93617	0	1	-360	360;
	8347	7507	0.000365	0.025165	0	591	0	0	0.93617	0	1	-360	360;
	4000	4852	0.000355	0.024237	0	591	0	0	0.979227	0	1	-360	360;
	4000	2012	0.000375	0.024247	0	591	0	0	0.93617	0	1	-360	360;
	2183	4852	0.000365	0.023267	0	591	0	0	0.979227	0	1	-360	360;
	2183	2012	0.000342	0.023139	0	591	0	0	0.982143	0	1	-360	360;
	1642	7098	0.000344	0.026666	0	591	0	0	0.979227	0	1	-360	360;
	1642	687	0.000334	0.027021	0	591	0	0	0.93617	0	1	-360	360;
	809	7098	0.000302	0.027615	0	591	0	0	0.979227	0	1	-360	360;
	809	687	0.000313	0.026135	0	591	0	0	0.93617	0	1	-360	360;
	4594	8804	0.000365	0.025905	0	591	0	0	0.93617	0	1	-360	360;
	4594	8804	0.000355	0.025238	0	591	0	0	0.93617	0	1	-360	360;
	2919	4215	0.000334	0.026814	0	591	0	0	0.93617	0	1	-360	360;
];

%%-----  OPF Data  -----%%
%% generator cost data
%	1	startup	shutdown	n	x1	y1	...	xn	yn
%	2	startup	shutdown	n	c(n-1)	...	c0
mpc.gencost = [
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
];
