function mpc = case13659pegase
%CASE13659PEGASE  Power flow data for European system with step-up transformers.
%   Please see CASEFORMAT for details on the case file format.
%
%   This case accurately represents the size and complexity of the European
%   high voltage transmission network. The network contains 13,659 buses,
%   4,092 generators, and 20,467 branches and it operates at 750, 400, 380,
%   330, 220, 154, 150, 120, and 110 kV. Low voltage buses ranging from 
%   27,000 to 400 V are used to model step-up transformers which connect 
%   generators to the high-voltage network. Please note that the data are
%   fictitious and do not correspond to real world data. They can thus be
%   used to validate methods and tools but should not be used for operation
%   and planning of the European grid.
%
%   The data stems from the Pan European Grid Advanced Simulation and State
%   Estimation (PEGASE) project, part of the 7th Framework Program of the
%   European Union (https://www.fp7-pegase.com/).
%
%   When publishing results based on this data, please cite:
%
%     C. Josz, S. Fliscounakis, J. Maeght, and P. Panciatici, "AC Power Flow
%     Data in MATPOWER and QCQP Format: iTesla, RTE Snapshots, and PEGASE"
%     https://arxiv.org/abs/1603.01533
%
%     S. Fliscounakis, P. Panciatici, F. Capitanescu, and L. Wehenkel,
%     "Contingency ranking with respect to overloads in very large power
%     systems taking into account uncertainty, preventive and corrective
%     actions", Power Systems, IEEE Trans. on, (28)4:4909-4917, 2013.
%     https://doi.org/10.1109/TPWRS.2013.2251015
%
%   Remarks:
%
%   1. Line flow limits are the current flow limits found in PEGASE data.
%
%   2. PEGASE data contains asymmetric shunt conductance and susceptance in
%   the PI transmission line model of branches. Thus total line charging
%   susceptance of branches is set to 0 p.u. and the nodal representation
%   of shunt condutance and susceptance is used. As a result, power flow
%   equations are left unchanged compared with original PEGASE data.
%   However, line flow constraints in the optimal flow problem are
%   modified.
%
%   3. Identical linear costs are used for all generators to form a loss
%   minimizing OPF objective function.
%
%   4. Since some parts of the network are aggregated, some generators
%   (e.g. with negative PMIN) represent aggregations of multiple loads
%   and generators.
%
%   Contacts:
%     Cédric Josz, Stéphane Fliscounakis, Jean Maeght, Patrick Panciatici
%     firstname.lastname@rte-france.com
%     Réseau de Transport d'Electricité (French Transmission System Operator)
%     Département Expertise Système, Immeuble "Le Colbert"
%     9 rue de la Porte de Buc, 78000 Versailles Cedex, France
%
%   March 4th, 2016

%   MATPOWER
%   Copyright (c) 2016 by Cédric Josz, Stéphane Fliscounakis, Jean Maeght,
%   and Patrick Panciatici
%   Licensed under the Creative Commons Attribution 4.0 International license,
%   https://creativecommons.org/licenses/by/4.0/

%% MATPOWER Case Format : Version 2
mpc.version = '2';

%%-----  Power Flow Data  -----%%
%% system MVA base
mpc.baseMVA = 100;

%% bus data
%	bus_i	type	Pd	Qd	Gs	Bs	area	Vm	Va	baseKV	zone	Vmax	Vmin
mpc.bus = [
	1	3	0	0	0	0	0	1.031695	0	20	6	1.1	0.9;
	2	1	72.68	45	0	6.61	0	1.00238	-9.035194	150	9	1.1	0.9;
	3	1	378.4	98.5	0	16.61	0	0.989665	-7.514519	220	6	1.1	0.9;
	4	1	134	19.1	0	0.08	0	1.005027	12.325176	154	1	1.1	0.9;
	5	1	151	48.8	0	4.69	0	0.98998	7.445755	220	5	1.1	0.9;
	6	1	171.41	23.4	0	2.1	0	1.004053	22.504675	220	5	1.1	0.9;
	7	1	45.95	14	0	13.34	0	0.978685	-21.610492	220	6	1.1	0.9;
	8	1	343.4	53.1	0.70508	36.832433	0	1.023941	-13.3046	380	6	1.1	0.9;
	9	1	-0	-0	0	75.076713	0	0.995208	14.929241	380	5	1.1	0.9;
	10	1	31.03	10.3	0	0.65	0	0.935403	-13.105656	110	21	1.1	0.9;
	11	1	0	-0	0	3.41	0	1.038279	3.372868	154	1	1.1	0.9;
	12	2	0	0	0	0	0	1.004054	96.865589	20	3	1.1	0.9;
	13	1	14.3	2	0	1.91	0	1.046187	21.142072	154	1	1.1	0.9;
	14	1	134	24.7	0	12.44	0	1.028056	6.025887	220	5	1.1	0.9;
	15	1	0	0	0	0	0	0.966985	-9.629605	150	9	1.1	0.9;
	16	1	-6.11	2	0	0.47	0	1.048294	-4.845447	150	9	1.1	0.9;
	17	1	-22.74	-8.32	0	6.5	0	1.01052	-21.436889	220	6	1.1	0.9;
	18	2	0	0	0	0	0	1.043087	11.638661	20	14	1.1	0.9;
	19	1	59.67	26.7	0	12.84	0	1.029741	-21.109094	220	4	1.1	0.9;
	20	1	30.6	4.3	0	5.15	0	0.995795	32.041235	154	1	1.1	0.9;
	21	1	0	-0	0.234457	-0.195381	0	1.024139	-12.441425	380	6	1.1	0.9;
	22	1	67.67	12.1	0	2.18	0	1.012996	31.441278	220	3	1.1	0.9;
	23	1	0	0	0.295874	20.603438	0	1.023227	43.176177	380	3	1.1	0.9;
	24	1	25.33	7.5	0	2.39	0	0.977442	-13.211902	110	12	1.1	0.9;
	25	1	1.4	0.2	0	4.93	0	1.01187	32.552868	154	1	1.1	0.9;
	26	1	161.2	39.3	0	30.71	0	1.030926	8.56355	220	5	1.1	0.9;
	27	1	0	-0	0	0.85	0	1.032196	11.548192	220	5	1.1	0.9;
	28	1	19.62	7.1	0	1.15	0	0.994911	-5.413569	110	19	1.1	0.9;
	29	2	0	0	0	0	0	1.033319	1.207015	20	18	1.1	0.9;
	30	1	31.94	8.8	0	0.73	0	1.014217	-5.598531	110	22	1.1	0.9;
	31	1	0	-0	0.235277	0.313936	0	1.029519	36.693892	220	3	1.1	0.9;
	32	1	167.3	54.8	0	110.54	0	1.008442	25.900391	380	5	1.1	0.9;
	33	2	0	0	0	0	0	1.018631	-11.528817	20	4	1.1	0.9;
	34	1	-0	0	0	14.82	0	1.050355	40.500598	380	13	1.1	0.9;
	35	1	0	0	0	0	0	1.009463	-9.554155	150	9	1.1	0.9;
	36	1	152.9	49.4	0	27.65	0	0.978505	14.001978	380	5	1.1	0.9;
	37	1	-0	0	0	0.56	0	1.027858	-7.880552	150	9	1.1	0.9;
	38	1	69.68	43.2	0	6.3	0	0.914844	-11.630868	150	9	1.1	0.9;
	39	2	0	0	0	0	0	1.073284	1.2557	20	9	1.1	0.9;
	40	1	0	-0	0	37.54	0	1.002842	-15.045231	220	4	1.1	0.9;
	41	1	0	-0	0	0.25	0	1.020848	-10.580356	220	4	1.1	0.9;
	42	2	0	0	0	0	0	0.985213	28.561806	11	1	1.1	0.9;
	43	1	13.51	4.9	0	1.8	0	1.021293	-2.929459	110	19	1.1	0.9;
	44	1	10.51	3.5	0	0.31	0	0.990534	-11.414462	110	12	1.1	0.9;
	45	2	0	0	0	0	0	1.012099	-18.860818	20	4	1.1	0.9;
	46	1	-0	0	0.12	15.34	0	1.039103	0.763772	220	19	1.1	0.9;
	47	1	9.21	2.6	0	2	0	0.9757	-6.904131	110	18	1.1	0.9;
	48	1	42.75	12.8	0	13.3	0	1.049416	58.695823	220	8	1.1	0.9;
	49	1	-0	0	0.019959	0.320041	0	0.998825	-10.375539	150	10	1.1	0.9;
	50	1	16.02	4.54	0	0.1	0	0.983186	-8.374635	110	23	1.1	0.9;
	51	2	0	0	0	0	0	1.011416	44.498643	14.4	1	1.1	0.9;
	52	1	27.9	4	0	4.23	0	0.988125	16.109775	154	1	1.1	0.9;
	53	1	20.82	7.9	0	26.89	0	1.016129	-21.455497	380	4	1.1	0.9;
	54	2	0	0	0	0	0	1.021005	21.987746	20	14	1.1	0.9;
	55	1	0	0	0	7.27	0	1.036171	8.177694	220	5	1.1	0.9;
	56	1	10.11	3.7	0	0.81	0	1.04206	-2.933811	110	19	1.1	0.9;
	57	1	10.01	2.7	0	1.14	0	0.954805	-6.844584	110	18	1.1	0.9;
	58	1	6.31	-0	0	0.67	0	0.981092	-5.140628	110	18	1.1	0.9;
	59	1	2.9	0.9	0	0.6	0	1.038133	-3.37057	110	15	1.1	0.9;
	60	1	-0	0	0.318913	17.760815	0	1.006337	-6.766633	220	6	1.1	0.9;
	61	1	-0	-0	0.549745	0.432068	0	1.047066	36.05057	380	2	1.1	0.9;
	62	2	0	0	0	0	0	1.058268	15.875889	20	5	1.1	0.9;
	63	2	0	0	0	0	0	1.058268	15.875889	20	5	1.1	0.9;
	64	1	151.8	48.3	0	5.53	0	1.009945	-20.345542	220	4	1.1	0.9;
	65	1	127.6	31.8	0	20.51	0	0.99193	25.086276	150	8	1.1	0.9;
	66	2	0	0	0	0	0	1.034768	8.492541	20	22	1.1	0.9;
	67	1	0	0	0	0.84	0	0.975466	-6.422787	110	21	1.1	0.9;
	68	2	0	0	0	0	0	1.035023	-10.739655	20	4	1.1	0.9;
	69	2	0	0	0	0	0	1.035023	-10.739655	20	4	1.1	0.9;
	70	1	110.5	-26.8	0	16.37	0	1.007417	-5.990692	220	6	1.1	0.9;
	71	1	-0	-0	0	0	0	1.0206	13.215214	20	5	1.1	0.9;
	72	1	-0	-0	0	0	0	1.0206	13.215214	20	5	1.1	0.9;
	73	1	0	0	0	1.23	0	0.993601	33.018065	220	3	1.1	0.9;
	74	1	0	-0	0	0	0	1.024981	18.183447	20	5	1.1	0.9;
	75	1	0	-0	0	0	0	1.024981	18.183447	20	5	1.1	0.9;
	76	1	0	-0	0	0	0	1.024981	18.183447	20	5	1.1	0.9;
	77	2	0	0	0	0	0	1.038344	11.554722	20	5	1.1	0.9;
	78	2	0	0	0	0	0	1.027383	17.740146	20	5	1.1	0.9;
	79	2	0	0	0	0	0	1.016716	24.455607	20	5	1.1	0.9;
	80	1	140	79.7	0	6.5	0	1.009229	11.978416	220	3	1.1	0.9;
	81	1	21.6	3	0	3.54	0	0.990476	15.061407	154	1	1.1	0.9;
	82	1	0	-0	0	2.31	0	1.005323	7.725281	380	5	1.1	0.9;
	83	1	119.2	20.7	0	3.31	0	1.01756	5.059048	220	5	1.1	0.9;
	84	1	26.43	9.9	0	9.57	0	0.989474	-14.141367	110	22	1.1	0.9;
	85	1	0.24	-1.66	0	0.65	0	1.024209	3.944647	110	18	1.1	0.9;
	86	2	0	0	0	0	0	1.023881	-4.354948	20	6	1.1	0.9;
	87	1	-0	0	0	0	0	1.014629	-9.004047	20	6	1.1	0.9;
	88	1	-0	0	0	0	0	1.014629	-9.004047	20	6	1.1	0.9;
	89	1	0	0	0	58.57	0	1.067567	73.239254	380	3	1.1	0.9;
	90	1	-0	0	0	1.34	0	0.965658	-5.045088	110	18	1.1	0.9;
	91	1	187.9	110.2	0	21.25	0	1.016938	32.836298	380	3	1.1	0.9;
	92	1	14.92	5.5	0	1.31	0	1.05599	-0.961794	110	15	1.1	0.9;
	93	1	-18.52	6.1	0	0.79	0	0.929361	-10.259554	150	9	1.1	0.9;
	94	2	0	0	0	0	0	1.022611	22.294732	11	1	1.1	0.9;
	95	1	4.61	1.7	0	2.32	0	0.989635	-2.082376	110	19	1.1	0.9;
	96	2	0	0	0	0	0	1.043225	-4.501166	20	4	1.1	0.9;
	97	2	0	0	0	0	0	1.043225	-4.501166	20	4	1.1	0.9;
	98	2	0	0	0	0	0	1.043225	-4.501166	20	4	1.1	0.9;
	99	2	0	0	0	0	0	1.031885	-12.651409	20	4	1.1	0.9;
	100	2	0	0	0	0	0	1.031885	-12.651409	20	4	1.1	0.9;
	101	2	0	0	0	0	0	1.043225	-4.501166	20	4	1.1	0.9;
	102	2	0	0	0	0	0	1.043225	-4.501166	20	4	1.1	0.9;
	103	1	62.87	20.5	0	1.15	0	0.989544	23.826254	150	8	1.1	0.9;
	104	2	0	0	0	0	0	1.052257	44.731305	20	3	1.1	0.9;
	105	1	13.5	14.8	0	132.435168	0	1.141691	0.929631	380	9	1.2	0.8;
	106	1	126.3	42.4	0	55.24	0	1.033923	30.21213	150	8	1.1	0.9;
	107	2	0	0	0	0	0	1.028153	78.700831	20	3	1.1	0.9;
	108	1	6.91	-4.6	0	5.75	0	1.02024	-13.484638	220	4	1.1	0.9;
	109	1	15.72	5.2	0	0.11	0	0.99755	-13.611914	110	12	1.1	0.9;
	110	2	0	0	0	0	0	1.051233	87.799236	20	3	1.1	0.9;
	111	1	16.12	4.3	0	0.49	0	1.040632	-1.700832	110	23	1.1	0.9;
	112	1	38.74	6.1	0	1.17	0	1.014811	-25.695432	220	4	1.1	0.9;
	113	1	22.13	10.7	0	14.21	0	1.049481	58.705166	220	8	1.1	0.9;
	114	2	0	0	0	0	0	1.030832	-9.752509	20	4	1.1	0.9;
	115	2	0	0	0	0	0	1.019	-2.975058	20	10	1.1	0.9;
	116	1	0	-0	0	0.73	0	1.01569	-23.117496	220	4	1.1	0.9;
	117	1	0	0	0	0.83	0	0.98256	1.655497	110	18	1.1	0.9;
	118	1	138.8	136.7	0	8.76	0	1.005586	39.420522	380	3	1.1	0.9;
	119	1	137.3	66.4	0.174767	14.17407	0	1.047225	-3.097806	220	23	1.1	0.9;
	120	1	67.76	19.18	0	1.47	0	0.992206	88.681174	150	1	1.1	0.9;
	121	2	0	0	0	0	0	1.045359	10.101302	20	14	1.1	0.9;
	122	1	0	-0	0	6.62	0	0.989992	37.681252	220	3	1.1	0.9;
	123	1	-0	-0	0	7.22	0	1.015598	45.599895	380	1	1.1	0.9;
	124	1	1	0.3	0	0.19	0	1.046082	-0.185974	110	19	1.1	0.9;
	125	1	20.9	3	0	5.74	0	0.979547	43.485266	154	1	1.1	0.9;
	126	1	0	0	0	0.12	0	1.054143	-2.947678	150	9	1.1	0.9;
	127	1	21.52	5.1	0	0.5	0	1.015831	7.375377	110	18	1.1	0.9;
	128	1	71.98	28.8	0	11	0	0.992515	-4.018187	220	6	1.1	0.9;
	129	1	0	0	0.205132	4.470744	0	0.992133	32.281513	380	2	1.1	0.9;
	130	1	363.25	115.39	0	3.95	0	1.032584	1.551181	220	5	1.1	0.9;
	131	1	192.9	73.3	0	20.75	0	1.01593	31.223648	380	3	1.1	0.9;
	132	2	0	0	0	0	0	0.996117	32.042491	11	1	1.1	0.9;
	133	2	0	0	0	0	0	1.042317	14.274354	20	14	1.1	0.9;
	134	2	0	0	0	0	0	1.044129	-4.738961	20	4	1.1	0.9;
	135	2	0	0	0	0	0	1.030832	-9.752509	20	4	1.1	0.9;
	136	2	0	0	0	0	0	1.049674	87.363329	20	3	1.1	0.9;
	137	1	17.52	6.3	0	0.26	0	1.033716	0.043737	110	19	1.1	0.9;
	138	1	33.14	12	0	0.65	0	0.998323	-1.616561	110	19	1.1	0.9;
	139	1	40.44	25	0	9.82	0	0.982498	-8.529449	150	9	1.1	0.9;
	140	1	193.17	53.37	0	4.58	0	1.007093	11.733465	220	5	1.1	0.9;
	141	1	90.7	25.5	0	3.11	0	1.086033	24.723865	220	3	1.2	0.8;
	142	1	176.3	-58.9	0	24.77	0	1.010147	0.04267	380	6	1.1	0.9;
	143	1	35.94	22.3	0	0	0	1.041468	-6.86067	150	9	1.1	0.9;
	144	2	0	0	0	0	0	1.034631	4.683321	20	14	1.1	0.9;
	145	2	0	0	0	0	0	1.040587	38.218788	20	5	1.1	0.9;
	146	2	0	0	0	0	0	1.044313	37.077846	20	5	1.1	0.9;
	147	2	0	0	0	0	0	1.044313	37.077846	20	5	1.1	0.9;
	148	1	143.9	123.2	0	72.64	0	1.01726	31.78152	380	3	1.1	0.9;
	149	1	-0	-0	0	3.77	0	0.998552	-23.031405	220	6	1.1	0.9;
	150	1	-0	8.4	0	15.72	0	1.035657	-6.436529	150	9	1.1	0.9;
	151	2	0	0	0	0	0	1.023215	19.248221	13.8	1	1.1	0.9;
	152	1	51.36	-8.8	0	2.15	0	0.957631	-29.999659	220	6	1.1	0.9;
	153	2	0	0	0	0	0	1.027348	16.396901	20	14	1.1	0.9;
	154	2	0	0	0	0	0	1.05252	75.282812	20	3	1.1	0.9;
	155	1	-0	-0	0	82.55	0	1.018112	-4.652605	380	6	1.1	0.9;
	156	2	0	0	0	0	0	1.034597	19.022639	20	6	1.1	0.9;
	157	1	322.7	11.9	0	10.28	0	0.997201	-11.924333	380	6	1.1	0.9;
	158	1	303.6	79.8	0.019533	1.690467	0	1.030319	13.564017	220	5	1.1	0.9;
	159	1	3.8	1.3	0	0.4	0	1.003488	6.243881	110	18	1.1	0.9;
	160	1	137	19.5	0	4.9	0	1.005167	12.788654	154	1	1.1	0.9;
	161	2	0	0	0	0	0	1.034597	19.022639	20	6	1.1	0.9;
	162	2	0	0	0	0	0	1.034597	19.022639	20	6	1.1	0.9;
	163	1	0	0	0	0	0	0.998875	-9.905754	150	9	1.1	0.9;
	164	1	61.07	36	0	3.09	0	1.048414	5.258031	220	11	1.1	0.9;
	165	2	0	0	0	0	0	1.030832	-9.752509	20	4	1.1	0.9;
	166	1	-0	0	0	3.24	0	1.069303	18.198045	220	3	1.1	0.9;
	167	1	0	0	0	6.520303	0	1.025278	15.598401	380	5	1.1	0.9;
	168	1	55.7	7.9	0	4.21	0	0.982808	36.169129	154	1	1.1	0.9;
	169	1	0	0	0	4.21	0	1.043196	10.569078	220	5	1.1	0.9;
	170	1	1.8	0.7	0	0.3	0	0.963316	-5.114898	110	19	1.1	0.9;
	171	1	0	-0	0	0.71	0	1.004078	-8.960914	150	9	1.1	0.9;
	172	1	2.5	0	0	0.96	0	0.954773	-7.079414	110	18	1.1	0.9;
	173	1	0	0	0	0.09	0	1.071275	1.379824	150	9	1.1	0.9;
	174	1	38.94	11	0	0.64	0	0.982878	-10.227511	110	1	1.1	0.9;
	175	1	45.15	14.1	0	0.19	0	1.055172	-2.479395	110	15	1.1	0.9;
	176	1	0	-0	0	5.75	0	0.995048	32.610865	220	3	1.1	0.9;
	177	1	0	0	0.019004	4.360996	0	1.045733	7.66394	220	11	1.1	0.9;
	178	1	0	-0	0.566322	-0.241138	0	1.066307	38.390318	220	5	1.1	0.9;
	179	2	0	0	0	0	0	1.023144	-9.538309	20	4	1.1	0.9;
	180	1	17.22	5.7	0	1.22	0	0.938533	-9.006406	110	21	1.1	0.9;
	181	1	-0	-0	0	0	0	1.020557	9.624602	20	14	1.1	0.9;
	182	2	0	0	0	0	0	1.054937	4.624127	20	20	1.1	0.9;
	183	1	0	0	0	4.3	0	1.005985	0.980119	220	6	1.1	0.9;
	184	1	23.73	12.7	0	0.52	0	1.018182	14.939378	220	5	1.1	0.9;
	185	1	121.9	7.4	0	1.34	0	1.01375	15.934135	220	3	1.1	0.9;
	186	1	0	-0	0.135375	47.750675	0	1.043206	-19.578893	380	10	1.1	0.9;
	187	1	47.35	16.4	0	4.08	0	1.0253	-13.383559	110	22	1.1	0.9;
	188	1	-0	0	0	0	0	1.021419	-22.29244	20	4	1.1	0.9;
	189	1	-0	-0	0	21.39	0	1.043162	-16.102529	380	1	1.1	0.9;
	190	1	38.14	12.5	0	6.28	0	1.000545	26.619584	150	8	1.1	0.9;
	191	1	54.2	7.8	0	94.9	0	0.973826	31.884786	400	1	1.1	0.9;
	192	1	130.6	84.9	0	4.22	0	1.007461	6.385629	220	7	1.1	0.9;
	193	1	7.91	2.6	0	0.74	0	0.982641	-11.113148	110	12	1.1	0.9;
	194	1	42.1	74.61	0	37.58	0	0.983941	88.980027	150	1	1.1	0.9;
	195	1	-75.38	7.8	0	29.9	0	1.058902	-14.107366	380	10	1.1	0.9;
	196	2	0	0	0	0	0	1.021915	-9.376379	20	6	1.1	0.9;
	197	2	0	0	0	0	0	1.005341	-1.941399	20	4	1.1	0.9;
	198	1	41.05	14.3	0	11.53	0	1.002832	0.458665	220	6	1.1	0.9;
	199	1	-0	-0	0	4.14	0	1.012917	48.51202	220	8	1.1	0.9;
	200	1	100.9	49.3	0.020067	4.329933	0	0.982261	24.304858	220	13	1.1	0.9;
	201	1	-0	0	0.116699	1.102475	0	1.029747	-0.920263	220	6	1.1	0.9;
	202	1	0	0	0	0	0	1.024001	31.504337	150	8	1.1	0.9;
	203	1	-0	0	0	2.03	0	1.026171	18.685276	220	5	1.1	0.9;
	204	1	-0	0	0	98.53	0	1.059773	46.330163	380	14	1.1	0.9;
	205	1	2.8	264.7	0.529803	56.537931	0	1.032579	-12.222514	380	4	1.1	0.9;
	206	1	38.34	5.6	0	7.64	0	1.028327	20.28167	220	5	1.1	0.9;
	207	2	0	0	0	0	0	1.016461	26.125945	20	5	1.1	0.9;
	208	1	269.6	65.7	0	0	0	1.094867	0.194636	120	20	1.2	0.8;
	209	1	-8.71	-19.54	0	0	0	1.045186	43.271639	150	8	1.1	0.9;
	210	1	160	3.3	0.076728	5.693272	0	1.041193	12.289677	220	7	1.1	0.9;
	211	1	32.34	12	0	6.1	0	1.029639	21.583329	220	5	1.1	0.9;
	212	1	100.1	45.1	0	1.11	0	1.006064	-11.74724	380	6	1.1	0.9;
	213	1	12.31	3.9	0.076917	5.053083	0	1.047813	32.292346	220	5	1.1	0.9;
	214	1	131.75	24.35	0	4.48	0	1.028798	0.187665	154	1	1.1	0.9;
	215	1	0	0	0	107.29	0	0.965812	24.322013	400	1	1.1	0.9;
	216	1	0	-0	0	20.96	0	1.037755	-15.333664	220	4	1.1	0.9;
	217	1	0	0	0	0.56	0	1.000388	0.372092	150	7	1.1	0.9;
	218	1	0	0	0	1.29	0	0.994997	-14.242668	110	12	1.1	0.9;
	219	1	0	-0	0	1.7	0	0.994318	-4.863546	110	21	1.1	0.9;
	220	2	0	0	0	0	0	1.055654	39.981354	20	3	1.1	0.9;
	221	2	0	0	0	0	0	1.055654	39.981354	20	3	1.1	0.9;
	222	2	0	0	0	0	0	1.051105	39.960989	20	3	1.1	0.9;
	223	2	0	0	0	0	0	1.055654	39.981354	20	3	1.1	0.9;
	224	2	0	0	0	0	0	1.051105	39.960989	20	3	1.1	0.9;
	225	2	0	0	0	0	0	1.051105	39.960989	20	3	1.1	0.9;
	226	1	9.21	3.3	0	1.2	0	1.009411	7.858173	110	19	1.1	0.9;
	227	1	0	-0	0.488616	70.580637	0	0.992431	20.155304	380	5	1.1	0.9;
	228	1	166	60.3	0	0.57	0	1.006088	31.382093	220	3	1.1	0.9;
	229	2	0	0	0	0	0	1.034441	34.869174	20	10	1.1	0.9;
	230	1	-0	0	0	0.1	0	1.029856	-6.864459	150	9	1.1	0.9;
	231	1	79.4	11.3	0	4.92	0	1.010683	26.224607	154	1	1.1	0.9;
	232	1	0	0	0	0.16	0	1.097619	-4.744858	150	9	1.1	0.9;
	233	1	-0	-0	0	66.65	0	1.041056	18.76151	380	6	1.1	0.9;
	234	1	-597.44	-326.86	0	0	0	1.038236	23.896352	27	3	1.1	0.9;
	235	1	147.76	32.73	0	8.34	0	1.0473	19.001366	220	5	1.1	0.9;
	236	1	0	-0	0	5.17	0	1.021324	26.421146	150	8	1.1	0.9;
	237	1	26.8	-3.7	0	51.261678	0	1.080332	-2.610374	380	9	1.1	0.9;
	238	1	-0	0	0.176128	3.073227	0	1.031593	36.052809	220	3	1.1	0.9;
	239	1	257.5	131.1	0	1.08	0	1.029114	25.227158	220	3	1.1	0.9;
	240	1	16.7	2.4	0	6.09	0	0.994895	38.566247	154	1	1.1	0.9;
	241	2	0	0	0	0	0	1.037622	28.736858	20	14	1.1	0.9;
	242	1	-0	-0	0	1.91	0	1.05291	-6.33502	150	9	1.1	0.9;
	243	2	0	0	0	0	0	1.053647	50.613884	20	3	1.1	0.9;
	244	1	100.8	13.2	0	5.24	0	1.083523	3.965034	220	5	1.1	0.9;
	245	1	111.7	10.96	0	4.22	0	1.005919	-19.363692	220	4	1.1	0.9;
	246	1	90.38	-5.05	0	8.49	0	1.03892	41.627356	220	13	1.1	0.9;
	247	1	370.5	-17	0	9.91	0	1.041176	1.637578	220	5	1.1	0.9;
	248	2	0	0	0	0	0	1.026279	27.085932	20	10	1.1	0.9;
	249	1	41.05	-46	0.056872	19.253128	0	1.056058	3.848789	220	11	1.1	0.9;
	250	2	0	0	0	0	0	1.012099	-18.860818	20	4	1.1	0.9;
	251	1	9.91	3.2	0	0.67	0	0.984854	-15.499059	110	12	1.1	0.9;
	252	1	-0.69	-5.46	0	5.07	0	1.049653	-5.686791	150	9	1.1	0.9;
	253	1	55.46	8.4	0	5	0	1.015243	-27.279483	220	4	1.1	0.9;
	254	1	94.9	36.2	0	16.46	0	0.962719	-15.482689	220	10	1.1	0.9;
	255	1	81.09	48	0	0.03	0	1.017585	-8.578565	150	9	1.1	0.9;
	256	1	0	0	0	113.7	0	1.085688	7.234194	380	18	1.2	0.8;
	257	1	210.4	159.2	0.100208	131.579896	0	1.018032	-22.27659	380	4	1.1	0.9;
	258	2	0	0	0	0	0	0.955439	48.700819	10.6	1	1.1	0.9;
	259	2	0	0	0	0	0	1.03425	8.859346	20	14	1.1	0.9;
	260	1	19.32	26	0	0.69	0	1.034249	65.998285	220	3	1.1	0.9;
	261	1	29.83	7.3	0	6.49	0	1.024115	3.443986	220	5	1.1	0.9;
	262	1	-0	-0	0	2.42	0	1.035138	19.25936	220	5	1.1	0.9;
	263	1	20.9	3	0	4.11	0	1.007769	17.657973	154	1	1.1	0.9;
	264	1	19.92	6.6	0	0.56	0	0.98382	-7.354011	110	21	1.1	0.9;
	265	1	57.16	2	0	1.48	0	0.978872	-12.878729	110	22	1.1	0.9;
	266	2	0	0	0	0	0	1.005915	0.12691	20	18	1.1	0.9;
	267	1	36.84	13.9	0	3.3	0	1.05264	10.6209	220	5	1.1	0.9;
	268	1	-15.04	-5.12	0	0.43	0	0.992499	-4.854035	110	18	1.1	0.9;
	269	1	9.21	6	0.060012	0.149988	0	0.994437	-0.345088	110	19	1.1	0.9;
	270	2	0	0	0	0	0	1.003342	23.89027	10.6	1	1.1	0.9;
	271	2	0	0	0	0	0	1.053159	33.930403	20	5	1.1	0.9;
	272	1	-0	0	0.114032	0.684974	0	1.050896	93.317943	380	3	1.1	0.9;
	273	2	0	0	0	0	0	1.051456	33.495027	20	5	1.1	0.9;
	274	2	0	0	0	0	0	1.051088	33.437371	20	5	1.1	0.9;
	275	2	0	0	0	0	0	1.051088	33.437371	20	5	1.1	0.9;
	276	2	0	0	0	0	0	1.050379	33.293543	20	5	1.1	0.9;
	277	1	119.9	3.7	0	3.33	0	1.021851	51.649183	220	3	1.1	0.9;
	278	2	0	0	0	0	0	0.960771	44.51306	21	1	1.1	0.9;
	279	1	-0	-0	0.172756	-0.134366	0	1.03301	39.005968	150	8	1.1	0.9;
	280	1	-0	0	0.039023	0.210733	0	1.024903	-19.493228	220	6	1.1	0.9;
	281	1	73.14	21.55	0	8.52	0	1.027446	20.01219	220	5	1.1	0.9;
	282	1	-0	-0	0	12.48	0	1.024187	3.450586	220	5	1.1	0.9;
	283	2	0	0	0	0	0	1.000621	19.05802	11	1	1.1	0.9;
	284	1	0	-0	0.401208	101.317571	0	1.093767	71.95923	380	3	1.2	0.8;
	285	1	39.74	11.8	0	1.44	0	1.015091	14.638222	220	5	1.1	0.9;
	286	1	44.35	5.1	0	4	0	1.032642	-19.048559	220	4	1.1	0.9;
	287	1	86.8	28.4	0	1.34	0	0.988052	25.416707	150	8	1.1	0.9;
	288	1	50.56	5.9	0	10.58	0	1.007117	-15.468653	220	4	1.1	0.9;
	289	2	0	0	0	0	0	1.040289	49.685314	20	10	1.1	0.9;
	290	1	0	0	0	4.73	0	1.017529	-23.903561	220	4	1.1	0.9;
	291	1	110.9	32.7	0	15.4	0	1.010616	26.29288	220	14	1.1	0.9;
	292	1	18.2	2.6	0	12.38	0	0.996347	39.809508	154	1	1.1	0.9;
	293	1	0	0	0.181549	14.690889	0	1.050988	-0.863124	220	18	1.1	0.9;
	294	2	0	0	0	0	0	0.999187	25.525754	14.4	1	1.1	0.9;
	295	1	46.15	18.2	0	1.29	0	1.020708	32.06907	150	8	1.1	0.9;
	296	2	0	0	0	0	0	0.989117	33.465932	6.3	1	1.1	0.9;
	297	1	55.46	15.7	0	13.46	0	1.029403	-20.962021	220	4	1.1	0.9;
	298	1	34.04	12.3	0	0.23	0	1.038932	0.095836	110	19	1.1	0.9;
	299	2	0	0	0	0	0	0.993468	44.441311	11	1	1.1	0.9;
	300	1	321.98	84.4	0	1.34	0	1.018049	17.631655	220	5	1.1	0.9;
	301	1	42.55	8.2	0	2.06	0	1.0319	6.852205	220	5	1.1	0.9;
	302	1	33.44	10	0	1.36	0	1.012136	-3.292816	110	3	1.1	0.9;
	303	1	-0	0	0.019646	0.570354	0	0.995957	-11.169142	110	12	1.1	0.9;
	304	2	0	0	0	0	0	1.037997	-15.804853	20	4	1.1	0.9;
	305	2	0	0	0	0	0	1.030407	32.510933	20	5	1.1	0.9;
	306	2	0	0	0	0	0	1.030407	32.510933	20	5	1.1	0.9;
	307	2	0	0	0	0	0	1.049321	25.075911	20	5	1.1	0.9;
	308	2	0	0	0	0	0	1.049321	25.075911	20	5	1.1	0.9;
	309	2	0	0	0	0	0	1.05655	34.123914	20	5	1.1	0.9;
	310	2	0	0	0	0	0	1.05655	34.123914	20	5	1.1	0.9;
	311	2	0	0	0	0	0	1.05655	34.123914	20	5	1.1	0.9;
	312	2	0	0	0	0	0	1.05655	34.123914	20	5	1.1	0.9;
	313	1	-24.99	-4.94	0	6.99	0	1.08829	70.695055	220	3	1.2	0.8;
	314	1	95.69	24.54	0	2.85	0	1.016819	21.364532	220	13	1.1	0.9;
	315	2	0	0	0	0	0	1.021202	7.512556	6.3	1	1.1	0.9;
	316	1	37.34	10.9	0	0.63	0	0.980529	-11.005188	110	1	1.1	0.9;
	317	1	340	93.8	0.230291	74.777686	0	1.050755	40.337803	380	13	1.1	0.9;
	318	1	-0	0	0	2.04	0	1.053731	-3.038033	150	9	1.1	0.9;
	319	1	155.6	54.7	0	2.21	0	1.025917	5.589988	220	5	1.1	0.9;
	320	1	105.2	0.1	0	7.36	0	1.051242	28.233863	220	5	1.1	0.9;
	321	1	69.83	15.12	0	2.23	0	1.028278	16.095823	220	5	1.1	0.9;
	322	1	-0	0	1.107693	6.867601	0	1.052181	12.761421	380	7	1.1	0.9;
	323	2	0	0	0	0	0	1.016993	30.047857	20	5	1.1	0.9;
	324	2	0	0	0	0	0	1.024179	7.053021	11	1	1.1	0.9;
	325	2	0	0	0	0	0	0.996355	28.771079	6.3	1	1.1	0.9;
	326	1	-0	-0	0.197696	3.784535	0	1.009826	7.243886	380	5	1.1	0.9;
	327	1	176.2	20.8	0.019425	4.590288	0	1.021452	91.885907	150	1	1.1	0.9;
	328	2	0	0	0	0	0	1.023168	23.385838	20	10	1.1	0.9;
	329	1	27.03	8.8	0	1.84	0	1.016848	2.25686	110	21	1.1	0.9;
	330	1	0	-0	0	0.28	0	0.982988	1.603555	150	6	1.1	0.9;
	331	2	0	0	0	0	0	1.006233	16.146442	13.8	1	1.1	0.9;
	332	2	0	0	0	0	0	0.996766	22.457835	6.3	1	1.1	0.9;
	333	1	-0	0	0	1.15	0	1.040075	10.160115	220	5	1.1	0.9;
	334	1	0	-0	0	0	0	1.017664	-15.016249	20	4	1.1	0.9;
	335	1	14.62	4.8	0	1.02	0	0.996984	-8.982181	110	12	1.1	0.9;
	336	2	0	0	0	0	0	1.010862	-10.418134	20	4	1.1	0.9;
	337	1	0	-0	0	0	0	1.017664	-15.016249	20	4	1.1	0.9;
	338	2	0	0	0	0	0	1.00038	-5.809998	20	4	1.1	0.9;
	339	1	-0	0	0	0	0	1.008279	-10.478046	20	4	1.1	0.9;
	340	1	45.05	-2.2	0	3.96	0	1.018948	-16.326346	220	4	1.1	0.9;
	341	1	0	-0	0	2.55	0	1.020054	5.015345	220	7	1.1	0.9;
	342	1	-23.43	57.4	0	0	0	1.059806	27.195352	150	2	1.1	0.9;
	343	1	426.6	93.7	0	80.94	0	1.026432	2.663199	380	6	1.1	0.9;
	344	1	152.07	22.52	0	7.87	0	0.99088	-19.172603	220	4	1.1	0.9;
	345	2	0	0	0	0	0	1.030395	-6.336516	20	4	1.1	0.9;
	346	2	0	0	0	0	0	1.030395	-6.336516	20	4	1.1	0.9;
	347	1	0	0	0	0.7	0	1.016788	-2.028229	150	9	1.1	0.9;
	348	1	2.5	3.8	0	0.15	0	0.951099	-5.317252	110	18	1.1	0.9;
	349	1	30.53	10.9	0.589029	2.678252	0	1.02049	11.872201	220	19	1.1	0.9;
	350	1	34.04	4.4	0	1.25	0	0.968384	-15.748675	110	22	1.1	0.9;
	351	2	0	0	0	0	0	1.024484	-6.060208	20	6	1.1	0.9;
	352	1	45.85	7	0	1.05	0	1.013376	-26.727823	220	4	1.1	0.9;
	353	1	-0	0	0	30.97	0	1.021444	63.427508	380	3	1.1	0.9;
	354	2	0	0	0	0	0	1.029166	13.522836	20	14	1.1	0.9;
	355	1	46.05	16.9	0.03829	0.29171	0	1.044995	-0.245344	110	19	1.1	0.9;
	356	1	0	0	0	0	0	1.028324	-8.261232	150	9	1.1	0.9;
	357	1	0	0	0	3.95	0	1.048037	86.241564	220	3	1.1	0.9;
	358	2	0	0	0	0	0	1.050725	42.166745	20	10	1.1	0.9;
	359	1	-876.96	-325.87	0	0	0	1.051074	31.370584	27	3	1.1	0.9;
	360	2	0	0	0	0	0	1.028641	-17.676734	20	4	1.1	0.9;
	361	2	0	0	0	0	0	1.028641	-17.676734	20	4	1.1	0.9;
	362	2	0	0	0	0	0	1.028641	-17.676734	20	4	1.1	0.9;
	363	2	0	0	0	0	0	1.028641	-17.676734	20	4	1.1	0.9;
	364	2	0	0	0	0	0	1.00564	-10.626942	20	4	1.1	0.9;
	365	2	0	0	0	0	0	0.997585	4.636327	15.75	1	1.1	0.9;
	366	1	167.9	18.2	0	5.97	0	1.064626	24.975015	220	3	1.1	0.9;
	367	1	0	-0	0	13.49	0	1.021538	43.092184	150	8	1.1	0.9;
	368	2	0	0	0	0	0	1.028641	-17.676734	20	4	1.1	0.9;
	369	2	0	0	0	0	0	1.028641	-17.676734	20	4	1.1	0.9;
	370	1	-0	-0	0	0.68	0	1.003889	29.490383	150	8	1.1	0.9;
	371	1	44.35	14.8	0	1.94	0	0.977321	-8.562505	110	21	1.1	0.9;
	372	1	131.9	31.1	0.457733	14.047628	0	1.044634	1.324756	220	1	1.1	0.9;
	373	1	0	-0	0	0.72	0	0.98721	-1.170875	110	18	1.1	0.9;
	374	1	224.2	73.9	0.039171	84.090829	0	1.026096	38.079369	380	13	1.1	0.9;
	375	1	-0	0	0	26.3	0	1.002164	-15.041903	220	4	1.1	0.9;
	376	1	128.9	73.7	0	25.37	0	1.031976	45.925047	380	3	1.1	0.9;
	377	1	0	-0	0	0	0	0.993504	-13.39494	20	22	1.1	0.9;
	378	1	0	-0	0	0	0	0.993504	-13.39494	20	22	1.1	0.9;
	379	1	-0	-0	0	13.85	0	1.044274	36.097767	380	3	1.1	0.9;
	380	1	118.4	16.9	0	8.02	0	0.987495	35.201778	154	1	1.1	0.9;
	381	1	0	0	0	0	0	0.989227	-9.859665	20	22	1.1	0.9;
	382	2	0	0	0	0	0	0.990596	-4.984835	20	22	1.1	0.9;
	383	2	0	0	0	0	0	1.012482	-8.65265	20	22	1.1	0.9;
	384	1	0	0	0	0	0	0.989227	-9.859665	20	22	1.1	0.9;
	385	1	-0	-0	0	2.66	0	1.039191	75.245398	220	3	1.1	0.9;
	386	1	-0	0	0	4.06	0	1.036947	7.315715	220	1	1.1	0.9;
	387	1	0	0	0.135493	12.74062	0	1.056459	-2.614124	380	21	1.1	0.9;
	388	1	-0	0	0.276837	2.240894	0	1.061652	11.069394	380	16	1.1	0.9;
	389	1	-0	-0	0	7.09	0	1.024692	-4.329904	220	6	1.1	0.9;
	390	1	113.9	24.5	0	6.72	0	1.011342	-18.466433	220	4	1.1	0.9;
	391	1	88	12.6	0	22.25	0	1.024048	-0.786284	154	1	1.1	0.9;
	392	2	0	0	0	0	0	1.030636	1.593662	6.3	1	1.1	0.9;
	393	1	42.67	12.1	0	14.84	0	1.022918	-10.902946	150	9	1.1	0.9;
	394	2	0	0	0	0	0	1.0326	4.106738	20	14	1.1	0.9;
	395	2	0	0	0	0	0	1.028283	16.586422	20	14	1.1	0.9;
	396	2	0	0	0	0	0	1.029542	-3.096329	20	6	1.1	0.9;
	397	1	0	-0	0	1.29	0	1.043272	-18.000155	380	10	1.1	0.9;
	398	1	-0	-0	0	9.41	0	1.064278	4.191179	220	18	1.2	0.8;
	399	1	160.8	10.6	0.039515	21.960485	0	1.013719	35.800215	220	13	1.1	0.9;
	400	2	0	0	0	0	0	0.979015	43.42213	10.3	1	1.1	0.9;
	401	1	126.8	21.4	0	11.43	0	1.015344	-26.64921	220	4	1.1	0.9;
	402	1	48.8	6.9	0	3.11	0	1.00423	25.494521	154	1	1.1	0.9;
	403	1	38.74	9	0	0.57	0	1.020174	-19.057619	220	4	1.1	0.9;
	404	1	350.4	29.3	0	19.31	0	1.034858	-0.227793	220	11	1.1	0.9;
	405	2	0	0	0	0	0	1.00773	-6.62746	20	2	1.1	0.9;
	406	2	0	0	0	0	0	1.00773	-6.62746	20	2	1.1	0.9;
	407	1	103.8	2.7	0	3.33	0	1.011745	-24.157084	220	4	1.1	0.9;
	408	2	0	0	0	0	0	1.043492	-15.322718	20	4	1.1	0.9;
	409	1	-0.13	-1	0	0	0	1.071256	58.769025	220	8	1.1	0.9;
	410	1	0	-0	0	0.66	0	0.999874	-13.902635	110	22	1.1	0.9;
	411	1	0	0	0	6.34	0	1.042875	93.821172	380	1	1.1	0.9;
	412	1	96.7	26.8	0	0	0	1.045353	27.754007	150	2	1.1	0.9;
	413	1	0	0	0	19.54	0	1.067563	-10.167624	150	9	1.1	0.9;
	414	1	-0	-0	0	14.87	0	1.040225	-4.06361	220	23	1.1	0.9;
	415	1	0	0	0	0	0	0.988701	-12.066079	150	9	1.1	0.9;
	416	1	19.02	6	0	0.25	0	0.995801	-3.096316	110	21	1.1	0.9;
	417	1	102.1	25.3	0	1.11	0	0.994428	26.055624	150	8	1.1	0.9;
	418	1	-0	0	0	2.44	0	1.049988	90.865555	220	3	1.1	0.9;
	419	1	0	0	0	0	0	1.020452	-7.005782	150	9	1.1	0.9;
	420	1	113.3	30.2	0.039525	2.140475	0	1.004358	-4.040243	110	18	1.1	0.9;
	421	1	8.31	3	0	0.99	0	1.030113	-1.594177	110	19	1.1	0.9;
	422	1	119.6	29.6	0	5.34	0	1.035806	18.542315	220	5	1.1	0.9;
	423	2	0	0	0	0	0	1.015919	3.580325	6.3	1	1.1	0.9;
	424	1	-0	0	0	18.07	0	1.034615	39.288516	380	8	1.1	0.9;
	425	1	49.9	144	1.091586	80.060431	0	1.079277	10.025709	380	18	1.2	0.8;
	426	1	27.33	11.9	0	1.17	0	1.026746	18.911778	220	5	1.1	0.9;
	427	1	191.2	-3.4	0	19.21	0	1.029484	-21.160424	220	4	1.1	0.9;
	428	1	-0	-0	0.028663	18.318528	0	1.026931	93.803559	380	1	1.1	0.9;
	429	1	101.6	14.5	0	1.94	0	0.998912	29.803254	154	1	1.1	0.9;
	430	1	177.1	66.9	0	71.03	0	1.060659	10.322664	380	16	1.1	0.9;
	431	1	0	-0	0	0.3	0	1.069341	1.460666	150	9	1.1	0.9;
	432	2	0	0	0	0	0	1.016602	20.114087	6.3	1	1.1	0.9;
	433	2	0	0	0	0	0	1.011489	10.204543	20	23	1.1	0.9;
	434	1	0	-0	0.038538	-0.028903	0	1.036946	-2.421418	220	6	1.1	0.9;
	435	1	53.46	18.7	0	1.64	0	1.031463	4.496802	220	20	1.1	0.9;
	436	1	0	-0	0	12.73	0	1.023478	31.01211	220	14	1.1	0.9;
	437	1	344.7	-31.1	0	39.129841	0	1.023529	3.172519	380	6	1.1	0.9;
	438	2	0	0	0	0	0	1.016716	24.455607	20	5	1.1	0.9;
	439	1	-0	-0	0	1.98	0	1.017888	79.240726	220	3	1.1	0.9;
	440	1	85.1	67	0	0.18	0	1.076677	0.280361	150	9	1.1	0.9;
	441	1	0	0	0	0	0	1.024987	12.673591	20	5	1.1	0.9;
	442	2	0	0	0	0	0	1.043357	16.76639	20	5	1.1	0.9;
	443	1	-0	0	0	0.08	0	1.003413	0.732744	220	6	1.1	0.9;
	444	1	189.3	76.7	0	3.24	0	0.996316	47.58349	220	3	1.1	0.9;
	445	1	23.73	9.4	0	1.45	0	1.034044	31.640506	150	8	1.1	0.9;
	446	1	10.01	-21	0	8.59	0	1.017123	-4.668405	220	11	1.1	0.9;
	447	1	-0	-0	0	0	0	1.0206	13.215214	20	5	1.1	0.9;
	448	2	0	0	0	0	0	1.038795	17.135422	20	5	1.1	0.9;
	449	1	228.1	131.8	0.95374	3.934956	0	1.02775	-18.795505	380	4	1.1	0.9;
	450	1	-0	-0	0	1.05	0	1.021068	-0.6916	150	9	1.1	0.9;
	451	1	-14.85	-4.95	0	0.37	0	1.002719	-1.665851	110	18	1.1	0.9;
	452	1	20.9	3	0	2.65	0	1.013021	33.465429	154	1	1.1	0.9;
	453	1	431.7	121.6	0	1.14	0	0.984948	-4.258651	220	21	1.1	0.9;
	454	2	0	0	0	0	0	0.973833	16.627296	20	5	1.1	0.9;
	455	1	17.62	7.7	0	0.59	0	0.988031	-12.652652	110	22	1.1	0.9;
	456	1	0	0	0	109.27	0	1.046057	-14.752144	380	4	1.1	0.9;
	457	1	103	12	0	0.63	0	1.029934	-3.834541	220	5	1.1	0.9;
	458	2	0	0	0	0	0	1.009892	4.868208	20	9	1.1	0.9;
	459	1	0	0	0	4.76	0	1.029914	11.923984	220	5	1.1	0.9;
	460	1	0	0	0	25.85	0	1.078154	0.918362	380	1	1.2	0.8;
	461	1	0	0	0	0	0	0.868495	-15.380873	150	9	1.2	0.8;
	462	1	0	-0	0.060508	3.079577	0	0.997808	37.195998	220	3	1.1	0.9;
	463	2	0	0	0	0	0	1.126642	-0.913545	380	9	1.2	0.8;
	464	1	42.05	17.2	0	1.31	0	1.023637	24.670209	150	8	1.1	0.9;
	465	1	-0.16	-2.91	0	0.43	0	1.031718	39.18761	110	8	1.1	0.9;
	466	1	0	-0	0	0.06	0	1.007693	31.850349	154	1	1.1	0.9;
	467	1	163.1	46.3	0	10.6	0	1.009138	14.501366	220	5	1.1	0.9;
	468	1	0	-0	0	2.13	0	1.020059	-8.568677	220	10	1.1	0.9;
	469	1	-0	-0	0	2.81	0	1.067283	22.816615	220	3	1.1	0.9;
	470	2	0	0	0	0	0	1.046905	39.305577	20	2	1.1	0.9;
	471	1	0	0	0.049626	-1.990412	0	1.044907	35.76351	380	2	1.1	0.9;
	472	1	-0	-0	0	1.05	0	1.000606	-9.570545	150	9	1.1	0.9;
	473	1	153.3	21.8	0	18.25	0	1.027549	1.721949	154	1	1.1	0.9;
	474	1	-0	0	0	141.93	0	0.970629	34.677428	400	1	1.1	0.9;
	475	1	67.17	-23.9	0	0.93	0	1.050845	90.640831	220	3	1.1	0.9;
	476	2	0	0	0	0	0	1.029484	52.40343	20	10	1.1	0.9;
	477	2	0	0	0	0	0	1.043087	11.638661	20	14	1.1	0.9;
	478	1	-0	0	0.114166	11.294861	0	1.060985	79.464606	380	3	1.1	0.9;
	479	1	382.4	86.3	0	22.29	0	1.028161	4.473394	220	11	1.1	0.9;
	480	2	0	0	0	0	0	1.033425	29.680614	20	6	1.1	0.9;
	481	2	0	0	0	0	0	1.033425	29.680614	20	6	1.1	0.9;
	482	2	0	0	0	0	0	1.033425	29.680614	20	6	1.1	0.9;
	483	2	0	0	0	0	0	1.033425	29.680614	20	6	1.1	0.9;
	484	1	33.5	4.8	0	5.48	0	1.016462	21.52717	154	1	1.1	0.9;
	485	2	0	0	0	0	0	1.059619	6.357453	20	18	1.1	0.9;
	486	1	0.01	-0	0	0	0	1.071666	58.820002	220	8	1.1	0.9;
	487	1	0	0	0	12.29	0	1.014191	15.517546	380	3	1.1	0.9;
	488	1	291.5	51.3	0.244436	12.636488	0	1.004279	49.93024	220	13	1.1	0.9;
	489	1	24.23	7.1	0	1.64	0	0.988321	-4.477618	110	21	1.1	0.9;
	490	1	108.6	2.6	0	31.8	0	1.04238	36.640438	380	3	1.1	0.9;
	491	1	4	2.4	0	0.03	0	0.981642	-7.831663	150	9	1.1	0.9;
	492	2	0	0	0	0	0	1.041985	31.793879	20	10	1.1	0.9;
	493	1	30.03	10.9	0	1.45	0	1.002503	-5.375842	110	19	1.1	0.9;
	494	1	0	0	0	45.39	0	1.118986	0.152255	380	15	1.2	0.8;
	495	1	524.5	43	0	6.85	0	0.991902	-13.539498	220	5	1.1	0.9;
	496	2	0	0	0	0	0	1.00876	0.250829	20	18	1.1	0.9;
	497	1	129.9	18.5	0	0.38	0	1.015717	16.675593	154	1	1.1	0.9;
	498	1	-0	0	0	1.92	0	0.955363	-6.864987	110	18	1.1	0.9;
	499	1	4.4	0	0	0.51	0	1.008611	2.339129	110	18	1.1	0.9;
	500	1	9.22	1.2	0	0.8	0	0.981895	-2.975305	110	19	1.1	0.9;
	501	1	0	0	0.059058	0.120942	0	1.016425	22.882064	220	5	1.1	0.9;
	502	1	-0	0	0	0.14	0	1.022789	-2.776784	220	6	1.1	0.9;
	503	1	-0	0	0	0.36	0	0.963995	-4.73909	110	19	1.1	0.9;
	504	1	0	-0	0	0.81	0	0.989389	-4.565793	110	21	1.1	0.9;
	505	1	35.41	20.3	0	10.07	0	0.99315	0.886893	220	6	1.1	0.9;
	506	1	3.2	1.2	0	0.69	0	1.017162	-1.318256	110	19	1.1	0.9;
	507	1	372.8	147.5	0	27.73	0	1.018067	25.301446	150	8	1.1	0.9;
	508	2	0	0	0	0	0	1.044776	37.116836	20	5	1.1	0.9;
	509	2	0	0	0	0	0	1.044388	36.589282	20	5	1.1	0.9;
	510	2	0	0	0	0	0	1.044096	37.018553	20	5	1.1	0.9;
	511	1	-0	-0	0	0	0	1.002012	13.542185	20	5	1.1	0.9;
	512	1	-0	-0	0	0	0	1.002012	13.542185	20	5	1.1	0.9;
	513	1	-0	-0	0	0	0	1.002012	13.542185	20	5	1.1	0.9;
	514	2	0	0	0	0	0	1.044511	37.099046	20	5	1.1	0.9;
	515	1	-0	-0	0	0	0	1.002012	13.542185	20	5	1.1	0.9;
	516	2	0	0	0	0	0	1.041748	29.3468	20	5	1.1	0.9;
	517	1	-0	0	0.080024	-0.240071	0	1.014573	-5.773397	220	11	1.1	0.9;
	518	1	197.5	48.3	0	0.71	0	1.028486	8.959933	220	5	1.1	0.9;
	519	1	-0	0	0	0.22	0	1.005897	20.492913	154	1	1.1	0.9;
	520	1	-0	-0	0	15.15	0	1.02668	35.22961	150	8	1.1	0.9;
	521	1	110.4	26.2	0	5.62	0	1.024265	6.009421	220	5	1.1	0.9;
	522	1	-0.15	-1.93	0	0	0	1.016294	49.478494	220	8	1.1	0.9;
	523	1	-0	0	1.138771	10.122844	0	1.055944	31.159791	380	5	1.1	0.9;
	524	1	-0	-0	0	14.6	0	1.026439	-4.895454	220	12	1.1	0.9;
	525	1	0	0	0	0	0	0.986332	25.334582	150	8	1.1	0.9;
	526	1	-4.86	-2.15	0	0.31	0	0.988925	-2.520654	110	18	1.1	0.9;
	527	1	141.1	2.1	0	4.78	0	1.034921	6.905166	220	7	1.1	0.9;
	528	2	0	0	0	0	0	1.020703	22.380203	20	5	1.1	0.9;
	529	1	52.96	32.8	0	14.14	0	1.023322	-11.234733	150	9	1.1	0.9;
	530	1	149.2	-74.6	0.227896	49.239582	0	1.06537	47.218455	380	13	1.1	0.9;
	531	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	532	2	0	0	0	0	0	1.036699	34.476926	20	10	1.1	0.9;
	533	2	0	0	0	0	0	1.016377	7.336958	20	22	1.1	0.9;
	534	1	10.81	5.1	0	1.09	0	0.931414	-8.249444	110	18	1.1	0.9;
	535	1	-0	0	0	2.37	0	1.01686	13.039574	220	1	1.1	0.9;
	536	1	35.54	11.5	0	3.65	0	1.023154	38.809154	110	8	1.1	0.9;
	537	1	-0	-0	0.230877	22.907222	0	1.049545	42.876879	380	13	1.1	0.9;
	538	1	48.93	14.52	0	2.85	0	0.991967	89.112631	150	1	1.1	0.9;
	539	2	0	0	0	0	0	0.993492	29.317066	15	1	1.1	0.9;
	540	1	0	-0	0	26.52	0	1.102789	-15.259978	150	9	1.2	0.8;
	541	1	10.01	4	0	8.11	0	0.995314	-3.374716	220	6	1.1	0.9;
	542	1	0	0	0	0.82	0	1.013699	-0.827325	150	9	1.1	0.9;
	543	1	109.7	15	0	30.97	0	0.988843	-3.474628	220	5	1.1	0.9;
	544	1	65.17	-51.5	0	4.85	0	1.023056	6.428381	220	5	1.1	0.9;
	545	1	66.97	14.1	0	5.58	0	0.977758	4.623964	220	3	1.1	0.9;
	546	1	0	0	0	7.11	0	0.998467	-3.301764	220	6	1.1	0.9;
	547	1	-0	-0	0	3.78	0	0.998593	42.474186	220	8	1.1	0.9;
	548	1	56.76	23.2	0	0.49	0	1.020988	24.486279	150	8	1.1	0.9;
	549	1	0	-0	0	0.7	0	1.04807	86.165538	220	3	1.1	0.9;
	550	1	10.81	3.9	0	1.64	0	1.086334	1.463858	110	15	1.1	0.9;
	551	1	0	-0	0	0	0	0.995963	-9.132106	150	9	1.1	0.9;
	552	2	0	0	0	0	0	0.99834	24.249316	13.8	1	1.1	0.9;
	553	1	373.7	125.6	0.341257	71.244578	0	1.072559	36.179181	380	14	1.1	0.9;
	554	1	0	-0	0	0.16	0	0.980028	-8.586818	150	9	1.1	0.9;
	555	1	0	-0	0	2.27	0	1.023552	-0.978719	154	1	1.1	0.9;
	556	1	0	-0	0	1.61	0	1.02302	79.89045	220	3	1.1	0.9;
	557	1	-0	0	0	35.09	0	1.012671	-2.175683	380	6	1.1	0.9;
	558	1	-0.18	-1	0	0	0	1.020181	36.372744	150	8	1.1	0.9;
	559	2	0	0	0	0	0	0.954584	49.806962	11	1	1.1	0.9;
	560	1	9.91	3.3	0	6.72	0	1.009937	-13.160441	110	22	1.1	0.9;
	561	1	221.35	51.8	0	8.81	0	0.981754	18.128137	220	13	1.1	0.9;
	562	1	23.83	7.9	0	0.72	0	0.968054	-6.793963	110	21	1.1	0.9;
	563	1	-95.71	-2.8	0	7.25	0	0.997104	-5.319336	220	10	1.1	0.9;
	564	2	0	0	0	0	0	1.024499	-3.834468	20	20	1.1	0.9;
	565	1	44.05	-0	0	0	0	0.983771	-0.029864	150	7	1.1	0.9;
	566	1	16.52	3	0.03836	11.50164	0	1.050612	5.201724	220	7	1.1	0.9;
	567	1	-0	-0	0	0	0	1.037403	12.654648	20	7	1.1	0.9;
	568	1	-0	-0	0	19.63	0	0.995622	27.676502	220	3	1.1	0.9;
	569	1	0	-0	0	0.3	0	1.012073	-8.225001	150	9	1.1	0.9;
	570	1	0	-0	0	105.84	0	1.01609	-22.212248	380	4	1.1	0.9;
	571	1	-0	0	0	28.09	0	1.060426	48.406404	380	3	1.1	0.9;
	572	2	0	0	0	0	0	1.004314	14.241295	13.8	1	1.1	0.9;
	573	1	-0	0	0.258771	12.056323	0	0.988321	6.881414	220	19	1.1	0.9;
	574	1	-0	0	0.059384	0.280616	0	1.011211	-13.487557	150	10	1.1	0.9;
	575	2	0	0	0	0	0	1.016739	22.326375	20	5	1.1	0.9;
	576	2	0	0	0	0	0	1.016739	22.326375	20	5	1.1	0.9;
	577	2	0	0	0	0	0	1.016739	22.326375	20	5	1.1	0.9;
	578	1	34.94	21.6	0	0.18	0	0.982939	-5.438658	150	9	1.1	0.9;
	579	1	134.1	34.2	0	0.75	0	0.982134	23.783698	150	8	1.1	0.9;
	580	1	17.72	5.9	0	0.32	0	1.028821	-3.48117	220	21	1.1	0.9;
	581	2	0	0	0	0	0	1.016739	22.326375	20	5	1.1	0.9;
	582	1	-0	-0	1.560224	77.024822	0	1.07923	68.996986	380	3	1.1	0.9;
	583	1	0	0	0	7.85	0	1.033308	40.656079	380	8	1.1	0.9;
	584	1	-0.51	-20.69	0	2.07	0	1.018412	11.984544	220	3	1.1	0.9;
	585	2	0	0	0	0	0	1.041629	38.466941	20	2	1.1	0.9;
	586	1	0	0	0	0	0	0.912043	-12.081031	150	9	1.1	0.9;
	587	1	77.79	-40.8	0	12.06	0	1.037667	14.939315	220	3	1.1	0.9;
	588	1	0	-0	0	2.09	0	1.012735	9.265213	220	5	1.1	0.9;
	589	2	0	0	0	0	0	1.00755	-13.681446	20	4	1.1	0.9;
	590	1	5.01	2	0	0.13	0	0.999923	-3.039737	220	6	1.1	0.9;
	591	1	-19.66	-10.83	0	0.06	0	0.997618	-7.177191	110	18	1.1	0.9;
	592	1	56.5	36.29	0	3.08	0	1.003147	15.307317	220	5	1.1	0.9;
	593	1	-0	-0	0	0	0	1.020596	-16.730831	20	4	1.1	0.9;
	594	1	-0	-0	0	0	0	1.020596	-16.730831	20	4	1.1	0.9;
	595	1	-0	-0	0.090774	7.330444	0	1.055576	-5.790807	220	18	1.2	0.8;
	596	1	3.81	0.24	0	0.23	0	1.008982	-3.14743	110	18	1.1	0.9;
	597	1	254.7	79.4	0	9.43	0	1.004651	-16.584297	380	6	1.1	0.9;
	598	1	20.9	3	0	2.54	0	0.985019	22.550521	154	1	1.1	0.9;
	599	1	77.99	30.8	0	2.41	0	1.01493	29.607123	150	8	1.1	0.9;
	600	1	-0.15	-1	0	0.38	0	1.032362	39.964616	110	8	1.1	0.9;
	601	1	129.1	7.4	0	0.04	0	1.029876	19.391484	220	5	1.1	0.9;
	602	1	126.3	17.9	0	2.59	0	1.015668	18.412381	154	1	1.1	0.9;
	603	1	151.5	7.9	0	5	0	1.041573	-0.003639	220	5	1.1	0.9;
	604	1	-0	0	0.019593	1.150407	0	1.012186	-11.629863	110	22	1.1	0.9;
	605	1	26.4	3.8	0	4.65	0	0.982106	40.293349	154	1	1.1	0.9;
	606	1	0	-0	0.139186	6.210814	0	1.02619	24.279876	220	5	1.1	0.9;
	607	1	20.52	10.3	0	0.04	0	1.024665	-10.538885	150	9	1.1	0.9;
	608	1	4.71	0.7	0	11.39	0	1.019224	-24.665042	220	4	1.1	0.9;
	609	1	70.68	13.5	0	6.96	0	1.040938	22.649998	220	5	1.1	0.9;
	610	2	0	0	0	0	0	1.019723	-1.552655	20	18	1.1	0.9;
	611	1	234.3	105.9	0.020294	5.479706	0	0.985982	40.427243	220	13	1.1	0.9;
	612	1	24.43	8.7	0	0.17	0	1.023098	-4.178904	110	15	1.1	0.9;
	613	1	-0	0	0	0	0	1.007942	-10.569275	20	4	1.1	0.9;
	614	2	0	0	0	0	0	1.031902	6.119938	20	6	1.1	0.9;
	615	2	0	0	0	0	0	1.031902	6.119938	20	6	1.1	0.9;
	616	1	0	0	0	0	0	1.050465	43.901371	380	8	1.1	0.9;
	617	1	46.15	11.2	0	10.16	0	1.030589	4.0661	220	5	1.1	0.9;
	618	1	33.74	20.9	0	1.1	0	0.998875	-9.905754	150	9	1.1	0.9;
	619	2	0	0	0	0	0	1.031502	4.663459	20	6	1.1	0.9;
	620	1	35.84	12.5	0	0.26	0	1.022299	0.336959	110	19	1.1	0.9;
	621	2	0	0	0	0	0	1.003942	39.409116	11	1	1.1	0.9;
	622	1	-0	0	0.646492	4.682923	0	1.020094	47.941792	380	3	1.1	0.9;
	623	2	0	0	0	0	0	1.045359	10.101302	20	14	1.1	0.9;
	624	1	-0	-0	0	2.72	0	1.020239	32.099905	380	1	1.1	0.9;
	625	1	-0	-0	0	0	0	0.995856	-9.114615	150	9	1.1	0.9;
	626	2	0	0	0	0	0	1.035771	2.999441	20	14	1.1	0.9;
	627	1	123.1	6.5	0	3.25	0	1.037098	11.171028	220	5	1.1	0.9;
	628	1	-0	-0	0	1.24	0	1.064436	-8.048474	380	1	1.1	0.9;
	629	1	83.9	11.9	0	1.14	0	1.006004	20.84448	154	1	1.1	0.9;
	630	1	0.1	-0	0	0.57	0	0.980004	1.020887	110	18	1.1	0.9;
	631	1	26.93	-22.4	0	4.95	0	1.027192	-22.090057	220	4	1.1	0.9;
	632	1	29.03	9	0	0.56	0	1.029976	-3.543373	110	21	1.1	0.9;
	633	1	188.42	106.93	0	4.89	0	1.026706	33.46792	220	13	1.1	0.9;
	634	1	0	0	0	68.575723	0	1.024739	32.215254	380	5	1.1	0.9;
	635	2	0	0	0	0	0	1.023511	0.594852	11	1	1.1	0.9;
	636	1	-0	0	0	0.69	0	1.004986	-8.852574	150	9	1.1	0.9;
	637	2	0	0	0	0	0	1.04616	-9.114978	20	4	1.1	0.9;
	638	1	0	-0	0.295061	0.304116	0	1.024059	44.846055	380	3	1.1	0.9;
	639	1	9.7	1.4	0	4.89	0	1.011245	27.629302	154	1	1.1	0.9;
	640	1	176.2	-46.9	0	5.47	0	1.004462	-15.792016	220	6	1.1	0.9;
	641	1	33.14	11.6	0	0.22	0	1.057792	0.42714	110	19	1.1	0.9;
	642	1	18.3	8.2	0	35.421268	0	1.12305	1.040238	380	9	1.2	0.8;
	643	1	69.08	24	0	1	0	1.003618	-2.822895	220	6	1.1	0.9;
	644	1	-0	-0	0.057082	3.172432	0	1.049525	86.868297	220	3	1.1	0.9;
	645	1	-0	0	0.22164	16.9353	0	1.0826	86.434281	220	3	1.2	0.8;
	646	2	0	0	0	0	0	1.030395	-6.336516	20	4	1.1	0.9;
	647	1	-0	-0	0.199768	27.888516	0	1.020186	-24.812762	380	10	1.1	0.9;
	648	2	0	0	0	0	0	1.007781	59.664038	20	10	1.1	0.9;
	649	2	0	0	0	0	0	1.030395	-6.336516	20	4	1.1	0.9;
	650	1	44.85	7.1	0	5.65	0	0.994655	-18.823114	220	4	1.1	0.9;
	651	2	0	0	0	0	0	1.028351	-2.802846	20	23	1.1	0.9;
	652	2	0	0	0	0	0	1.028351	-2.802846	20	23	1.1	0.9;
	653	1	-12.61	4.1	0	1.83	0	1.006947	-5.91582	150	9	1.1	0.9;
	654	2	0	0	0	0	0	1.014507	26.136069	20	6	1.1	0.9;
	655	1	0	0	0	2.45	0	1.027725	18.930085	220	5	1.1	0.9;
	656	1	36.74	10.3	0	8.1	0	1.030053	-17.261924	220	4	1.1	0.9;
	657	1	100.53	43.37	0	0	0	1.014108	0.87733	110	24	1.1	0.9;
	658	1	2.2	1.3	0.019376	1.050624	0	0.995621	14.139898	110	18	1.1	0.9;
	659	1	0	0	0	0	0	0.974088	-8.221786	150	9	1.1	0.9;
	660	2	0	0	0	0	0	1.022689	-16.891323	20	4	1.1	0.9;
	661	1	-0	-0	0	29.42	0	1.010252	22.297225	380	5	1.1	0.9;
	662	1	42.04	43	0	0.02	0	1.080621	-3.774626	150	9	1.1	0.9;
	663	1	0	0	0	22.71	0	1.061402	49.738372	380	1	1.1	0.9;
	664	1	104.4	27.6	0.039809	1.850191	0	0.995363	-14.400857	110	22	1.1	0.9;
	665	1	0	-0	0.362004	3.698497	0	0.999024	20.683055	380	5	1.1	0.9;
	666	2	0	0	0	0	0	1.018739	-6.896129	20	9	1.1	0.9;
	667	1	23.7	3.4	0	8.13	0	1.010467	35.110882	154	1	1.1	0.9;
	668	1	274.32	5.66	0	13.24	0	1.031848	-1.87464	220	5	1.1	0.9;
	669	1	2.5	0	0	0.64	0	1.018382	7.768221	110	18	1.1	0.9;
	670	1	1.8	0.7	0	0.18	0	0.980165	3.066285	110	19	1.1	0.9;
	671	2	0	0	0	0	0	0.999061	42.441878	11	1	1.1	0.9;
	672	1	-0	0	0.189164	35.274722	0	1.027993	3.626388	380	11	1.1	0.9;
	673	1	97	40.3	0	16.79	0	1.030324	22.492733	220	3	1.1	0.9;
	674	1	0	-0	0	0.88	0	0.994789	-3.452739	110	18	1.1	0.9;
	675	1	0	0	0	0	0	1.017174	-12.280888	150	9	1.1	0.9;
	676	2	0	0	0	0	0	1.018066	-18.544118	20	4	1.1	0.9;
	677	1	15.22	5.1	0	1.34	0	0.976124	0.454545	110	18	1.1	0.9;
	678	1	0	0	0	0	0	1.014364	-8.724986	150	9	1.1	0.9;
	679	1	-0	-0	0.31285	13.426672	0	1.12555	-0.331163	380	9	1.2	0.8;
	680	1	108.5	35.3	0	49.63	0	1.038504	35.779711	380	3	1.1	0.9;
	681	2	0	0	0	0	0	1.00657	25.464337	20	24	1.1	0.9;
	682	1	24.33	7.6	0	1.18	0	0.985728	2.216033	110	18	1.1	0.9;
	683	1	44.75	0.9	0	6.67	0	1.034215	12.306912	220	7	1.1	0.9;
	684	2	0	0	0	0	0	1.035618	-0.462639	20	6	1.1	0.9;
	685	1	0	0	0	0	0	1.039652	-6.424256	150	9	1.1	0.9;
	686	1	40.5	5.7	0	3.9	0	1.004715	15.162759	154	1	1.1	0.9;
	687	1	-0	-0	0	27.85	0	1.0114	31.187658	380	3	1.1	0.9;
	688	1	9.01	6.8	0	0.44	0	0.963906	-6.15131	110	18	1.1	0.9;
	689	1	81.09	48	0	0.03	0	1.01888	-8.431592	150	9	1.1	0.9;
	690	1	0	0	0	0.43	0	1.054345	-5.313694	150	9	1.1	0.9;
	691	1	11.01	5.5	0	1.12	0	0.991725	-13.734054	110	22	1.1	0.9;
	692	1	-0	0	0	1.47	0	1.012011	5.276376	220	6	1.1	0.9;
	693	2	0	0	0	0	0	1.038999	38.995592	20	2	1.1	0.9;
	694	1	336.2	131.6	0	64.01	0	1.025884	34.061261	150	8	1.1	0.9;
	695	2	0	0	0	0	0	1.010516	39.993345	11	1	1.1	0.9;
	696	1	-22.95	-3.72	0	14.34	0	1.021643	-18.319826	220	4	1.1	0.9;
	697	1	0	0	0	0.97	0	0.985966	-13.767739	110	1	1.1	0.9;
	698	2	0	0	0	0	0	1.018527	-13.24267	20	4	1.1	0.9;
	699	1	0	0	0	9.41	0	1.043232	77.757181	380	3	1.1	0.9;
	700	1	67.3	94.8	0	7.96	0	0.971314	29.785076	220	3	1.1	0.9;
	701	1	-0	0	0	70.37	0	1.022096	-14.512809	380	4	1.1	0.9;
	702	1	0	0	0	33.23	0	1.083955	72.533534	380	3	1.2	0.8;
	703	1	16.62	3.1	0	1.14	0	1.005226	6.168419	110	18	1.1	0.9;
	704	1	5.81	1.1	0	4.02	0	1.039757	11.331045	110	19	1.1	0.9;
	705	1	0	-0	0	6.31	0	1.047819	-5.173347	220	5	1.1	0.9;
	706	1	-0	-0	0	17.44	0	1.028453	-20.661313	220	4	1.1	0.9;
	707	1	113.4	11.9	0	5.4	0	1.058829	21.8732	220	3	1.1	0.9;
	708	1	0	-0	0	1.05	0	0.977623	-2.554024	110	18	1.1	0.9;
	709	2	0	0	0	0	0	1.051727	31.564136	20	5	1.1	0.9;
	710	1	23.26	3.01	0	9.29	0	0.996598	8.383316	220	5	1.1	0.9;
	711	1	57.26	15.6	0	3.43	0	1.014587	-22.560626	220	4	1.1	0.9;
	712	1	0	0	0.17102	25.167483	0	1.047436	81.329599	220	3	1.1	0.9;
	713	1	7	1	0	1.3	0	0.997077	26.794575	154	1	1.1	0.9;
	714	1	7.01	2	0	0.44	0	0.998982	27.909753	150	8	1.1	0.9;
	715	2	0	0	0	0	0	1.007844	10.25283	20	18	1.1	0.9;
	716	1	87.8	12.5	0	0.38	0	1.015981	20.565424	154	1	1.1	0.9;
	717	2	0	0	0	0	0	1.027572	24.287546	11	1	1.1	0.9;
	718	1	6.61	2.92	0	1.17	0	1.020115	30.202806	220	14	1.1	0.9;
	719	1	115.9	14.7	0.019966	2.9	0	1.013826	-19.850393	220	4	1.1	0.9;
	720	1	20.52	4	0	2.77	0	1.029133	7.492457	220	5	1.1	0.9;
	721	2	0	0	0	0	0	0.992442	9.617717	20	18	1.1	0.9;
	722	1	0	-0	0.366153	32.651054	0	1.026106	61.149064	380	3	1.1	0.9;
	723	1	68.18	22	0	2.42	0	0.94297	-17.488149	110	22	1.1	0.9;
	724	2	0	0	0	0	0	1.045213	9.250128	20	14	1.1	0.9;
	725	1	23.23	7.6	0	0.76	0	0.95067	-7.990035	110	21	1.1	0.9;
	726	1	0	0	0	0	0	0.998821	-8.907903	150	9	1.1	0.9;
	727	1	181.7	51	0	73.49	0	1.068269	74.52173	380	3	1.1	0.9;
	728	1	-0	0	0	10.04	0	1.018461	36.492173	150	8	1.1	0.9;
	729	1	40	5.7	0	2.06	0	1.000858	34.472609	154	1	1.1	0.9;
	730	2	0	0	0	0	0	1.01458	18.108546	20	6	1.1	0.9;
	731	2	0	0	0	0	0	1.01458	18.108546	20	6	1.1	0.9;
	732	2	0	0	0	0	0	1.01458	18.108546	20	6	1.1	0.9;
	733	1	287.31	-5.58	0	19.9	0	0.990189	21.27187	380	13	1.1	0.9;
	734	2	0	0	0	0	0	1.01458	18.108546	20	6	1.1	0.9;
	735	1	144.4	35.5	0	7.66	0	1.027209	14.05509	220	5	1.1	0.9;
	736	1	-0	0	0	1.49	0	0.977249	0.569527	110	18	1.1	0.9;
	737	1	0	0	0	0.25	0	1.008405	-3.834733	110	3	1.1	0.9;
	738	1	0	-0	0	1.96	0	1.02769	23.00179	220	5	1.1	0.9;
	739	1	4.1	-607	0.290423	85.788781	0	1.065043	50.496306	380	3	1.1	0.9;
	740	1	18.62	6.2	0	1.07	0	0.970367	-6.743137	110	21	1.1	0.9;
	741	1	0	0	0	1.32	0	0.999635	35.983555	154	1	1.1	0.9;
	742	1	-0	0	0	0.4	0	1.058599	20.351147	380	5	1.1	0.9;
	743	1	23.43	8.5	0	0.33	0	1.040076	-3.544332	110	19	1.1	0.9;
	744	1	23.92	7.75	0	1.91	0	0.989367	-2.478771	110	18	1.1	0.9;
	745	1	77.29	7	0	17.6	0	1.018947	-33.068169	220	4	1.1	0.9;
	746	2	0	0	0	0	0	0.994194	24.952464	11	1	1.1	0.9;
	747	1	0	-0	0	1.5	0	1.013353	5.579718	220	6	1.1	0.9;
	748	1	0	-0	0.385462	139.767007	0	1.082894	85.120497	380	3	1.2	0.8;
	749	2	0	0	0	0	0	1.009791	14.416544	20	5	1.1	0.9;
	750	2	0	0	0	0	0	1.009791	14.416544	20	5	1.1	0.9;
	751	2	0	0	0	0	0	1.009791	14.416544	20	5	1.1	0.9;
	752	1	0	-0	0	13.16	0	1.00664	31.182408	380	3	1.1	0.9;
	753	2	0	0	0	0	0	1.018866	14.228745	20	10	1.1	0.9;
	754	2	0	0	0	0	0	1.009683	1.076815	20	18	1.1	0.9;
	755	1	9.61	3.1	0	0.77	0	1.004347	-13.085564	110	12	1.1	0.9;
	756	1	-195.22	-19	0.078938	7.621062	0	1.026914	7.66077	220	11	1.1	0.9;
	757	1	27.03	9	0	0.98	0	1.016003	2.506825	110	21	1.1	0.9;
	758	1	7.71	16.7	0	3.59	0	0.988701	-12.066079	150	9	1.1	0.9;
	759	1	602.9	201.4	0.118737	39.321052	0	1.029997	24.456838	380	14	1.1	0.9;
	760	1	207.2	9.4	0	8.77	0	1.014144	-23.295024	220	4	1.1	0.9;
	761	1	118.5	20.4	0	2.57	0	1.036321	1.925379	220	5	1.1	0.9;
	762	1	14.72	5.3	0	0.76	0	1.008774	-3.324558	110	19	1.1	0.9;
	763	1	100.5	37	0	14.95	0	1.070453	1.995467	380	18	1.2	0.8;
	764	1	11.61	7.2	0	28.77	0	1.024756	-11.276001	150	9	1.1	0.9;
	765	2	0	0	0	0	0	1.032985	-6.930528	20	6	1.1	0.9;
	766	1	189.9	74.7	0	26.2	0	1.019596	33.38308	150	8	1.1	0.9;
	767	1	0	-0	0	0.09	0	1.007718	37.209288	154	1	1.1	0.9;
	768	2	0	0	0	0	0	1.044527	23.855553	20	5	1.1	0.9;
	769	2	0	0	0	0	0	1.044527	23.855553	20	5	1.1	0.9;
	770	1	0	-0	0	0	0	1.039156	19.470405	20	5	1.1	0.9;
	771	1	0	-0	0	0	0	1.039156	19.470405	20	5	1.1	0.9;
	772	1	36.54	27.4	0	0.15	0	0.97928	37.684032	220	3	1.1	0.9;
	773	2	0	0	0	0	0	1.016081	21.284924	20	5	1.1	0.9;
	774	1	0	0	0	0	0	1.007722	14.132282	20	5	1.1	0.9;
	775	2	0	0	0	0	0	1.003456	18.777669	20	5	1.1	0.9;
	776	1	0	-0	0	0	0	1.039156	19.470405	20	5	1.1	0.9;
	777	1	277.7	45.4	0	0.7	0	1.018902	-18.885965	220	4	1.1	0.9;
	778	2	0	0	0	0	0	1.02868	47.75243	20	10	1.1	0.9;
	779	2	0	0	0	0	0	0.9829	25.033677	10	1	1.1	0.9;
	780	1	-0	0	0	28.16	0	1.022808	12.457089	380	1	1.1	0.9;
	781	1	25	3.6	0	4.73	0	1.046208	1.751552	154	1	1.1	0.9;
	782	1	-0	-0	0	47.4	0	0.975786	5.725108	380	1	1.1	0.9;
	783	2	0	0	0	0	0	1.027518	-17.089784	20	4	1.1	0.9;
	784	1	-0	0	0	1.98	0	1.008445	33.79762	220	3	1.1	0.9;
	785	1	16.52	3.8	0	0.98	0	0.95552	-7.1101	110	18	1.1	0.9;
	786	2	0	0	0	0	0	1.013792	13.048881	11	1	1.1	0.9;
	787	1	118.1	16.8	0	2.94	0	0.9983	26.918478	154	1	1.1	0.9;
	788	1	11.51	3.8	0	1.21	0	0.983092	-11.324453	110	12	1.1	0.9;
	789	1	87.8	-1.4	0.079073	36.047349	0	1.040763	-20.963956	380	10	1.1	0.9;
	790	1	27.53	17	0	0.76	0	0.954084	-7.518624	150	9	1.1	0.9;
	791	1	21.02	13	0	1.17	0	0.970954	-14.134822	150	9	1.1	0.9;
	792	1	57.96	18.9	0	14.91	0	0.996611	26.942681	150	8	1.1	0.9;
	793	1	166.9	86.1	0	4.06	0	0.981471	34.867888	220	3	1.1	0.9;
	794	1	-5.51	1.8	0	0.91	0	0.954505	-7.782367	150	9	1.1	0.9;
	795	1	86.5	29	0	1.45	0	1.066578	53.394043	380	3	1.1	0.9;
	796	1	2	0.7	0	0.02	0	0.978625	-3.324992	110	19	1.1	0.9;
	797	1	-0	-0	0	35.78	0	1.011512	30.216177	380	3	1.1	0.9;
	798	2	0	0	0	0	0	1.015954	3.75914	20	9	1.1	0.9;
	799	1	72	10.3	0	6.24	0	0.990011	1.446443	154	1	1.1	0.9;
	800	2	0	0	0	0	0	1.040587	38.218788	20	5	1.1	0.9;
	801	1	79.69	39.6	0	18.83	0	1.022097	-24.539739	380	10	1.1	0.9;
	802	1	-0	0	0	8.48	0	1.063006	0.575668	220	18	1.2	0.8;
	803	1	0	-0	0	29.89	0	1.099039	-14.346049	150	9	1.2	0.8;
	804	1	30.03	14	0	0.61	0	0.999004	-3.134908	220	6	1.1	0.9;
	805	1	30.93	8.9	0	0.45	0	1.036024	47.46783	110	8	1.1	0.9;
	806	1	0	-0	0	5.15	0	1.032716	17.680245	220	5	1.1	0.9;
	807	1	0	-0	0	10	0	1.02396	-18.053463	220	4	1.1	0.9;
	808	1	17.72	5.9	0	0.52	0	0.939327	-13.271181	110	21	1.1	0.9;
	809	2	0	0	0	0	0	1.023128	-17.667071	20	4	1.1	0.9;
	810	1	180.8	90.8	0	2.42	0	0.960507	-15.710242	150	10	1.1	0.9;
	811	2	0	0	0	0	0	1.022192	20.047374	20	5	1.1	0.9;
	812	1	0	-0	0	1.97	0	1.02116	9.225534	220	5	1.1	0.9;
	813	1	0	0	0	7.35	0	1.010866	-10.017249	220	11	1.1	0.9;
	814	2	0	0	0	0	0	1.01249	-3.900476	20	4	1.1	0.9;
	815	2	0	0	0	0	0	1.026207	-11.793103	20	4	1.1	0.9;
	816	1	16.52	9	0	1.17	0	1.029733	7.74677	220	5	1.1	0.9;
	817	1	-0	-0	0	12.14	0	1.001872	5.9468	220	19	1.1	0.9;
	818	1	-0.18	-3.57	0	0	0	1.021546	43.092206	150	8	1.1	0.9;
	819	2	0	0	0	0	0	1.048632	39.237883	20	2	1.1	0.9;
	820	1	0	-0	0	0.32	0	1.025508	18.162907	220	5	1.1	0.9;
	821	1	0	-0	0	0.48	0	1.015947	22.790189	220	14	1.1	0.9;
	822	1	-0	-0	0	0.55	0	0.969735	-9.642283	150	9	1.1	0.9;
	823	1	58.56	-0.1	0	7.38	0	1.026078	-19.617451	220	5	1.1	0.9;
	824	2	0	0	0	0	0	1.018739	-6.896129	20	9	1.1	0.9;
	825	1	47.55	-1.41	0	0.76	0	0.956703	-9.070653	110	18	1.1	0.9;
	826	1	-0	-0	0	14.5	0	1.045563	-9.81487	380	1	1.1	0.9;
	827	1	20.22	7.3	0	0.59	0	1.045872	-0.227578	110	19	1.1	0.9;
	828	2	0	0	0	0	0	1.01121	21.596284	20	10	1.1	0.9;
	829	1	-0	0	0	35.364539	0	1.089984	2.669964	380	20	1.2	0.8;
	830	1	47.95	29.7	0	1.59	0	0.998601	-10.373841	150	9	1.1	0.9;
	831	1	120.63	56.2	0	3	0	1.012387	35.628558	220	3	1.1	0.9;
	832	1	27.13	8.9	0	1.37	0	1.022107	3.77318	110	19	1.1	0.9;
	833	1	10.71	3.5	0	0.82	0	0.991892	-13.719477	110	12	1.1	0.9;
	834	1	82.89	26.8	0	0.65	0	0.998019	-5.508598	110	19	1.1	0.9;
	835	1	-0	-0	0	0.05	0	1.055149	5.178292	220	5	1.1	0.9;
	836	2	0	0	0	0	0	1.015461	-14.262307	20	4	1.1	0.9;
	837	1	0	-0	0	0	0	1.005017	-18.861992	20	4	1.1	0.9;
	838	1	108.6	38.7	0	0	0	1.001252	3.505451	110	18	1.1	0.9;
	839	2	0	0	0	0	0	1.0079	-15.676436	20	4	1.1	0.9;
	840	1	0	0	0	0	0	1.006753	-20.322256	20	4	1.1	0.9;
	841	2	0	0	0	0	0	1.036065	-16.582357	20	4	1.1	0.9;
	842	1	-0.13	-1	0	0	0	1.070032	58.678343	220	8	1.1	0.9;
	843	1	166	-8.8	0.03964	5.26027	0	1.019188	45.616813	220	13	1.1	0.9;
	844	2	0	0	0	0	0	1.059427	39.6916	20	14	1.1	0.9;
	845	2	0	0	0	0	0	1.059427	39.6916	20	14	1.1	0.9;
	846	2	0	0	0	0	0	1.059427	39.6916	20	14	1.1	0.9;
	847	1	17.32	6.6	0	16.52	0	1.00721	-24.676363	380	4	1.1	0.9;
	848	1	0	-0	0	0.17	0	0.984773	-5.596227	110	18	1.1	0.9;
	849	2	0	0	0	0	0	1.028153	78.700831	20	3	1.1	0.9;
	850	1	-0	0	0	31.19	0	1.158733	0.598452	380	23	1.2	0.8;
	851	1	48.8	6.9	0	0.12	0	1.033793	0.608319	154	1	1.1	0.9;
	852	1	51.16	18.2	0	5.23	0	1.03046	8.194629	220	7	1.1	0.9;
	853	1	0	-0	0.33357	58.172025	0	1.093065	68.734796	380	3	1.2	0.8;
	854	1	151.2	72.4	0	4.57	0	1.01458	31.505458	380	3	1.1	0.9;
	855	2	0	0	0	0	0	1.046298	0.868346	20	2	1.1	0.9;
	856	1	9.51	3.2	0	0.34	0	1.019323	-2.576944	110	21	1.1	0.9;
	857	1	3	0.2	0.230913	0.257194	0	1.039468	26.93922	220	13	1.1	0.9;
	858	2	0	0	0	0	0	1.021858	-2.534901	20	6	1.1	0.9;
	859	2	0	0	0	0	0	1.037973	11.471538	20	14	1.1	0.9;
	860	1	91.1	33.5	0	1.8	0	0.991857	-3.017578	220	6	1.1	0.9;
	861	2	0	0	0	0	0	1.018125	-0.191544	20	2	1.1	0.9;
	862	2	0	0	0	0	0	1.018125	-0.191544	20	2	1.1	0.9;
	863	1	27.03	9.8	0	1.01	0	1.04402	-4.498864	110	19	1.1	0.9;
	864	2	0	0	0	0	0	1.018125	-0.191544	20	2	1.1	0.9;
	865	2	0	0	0	0	0	1.018125	-0.191544	20	2	1.1	0.9;
	866	1	300.1	126.3	0.454323	17.960328	0	1.060805	81.893945	380	3	1.1	0.9;
	867	1	0	-0	0	0	0	0.987063	-10.738282	20	2	1.1	0.9;
	868	1	0	-0	0	0.56	0	0.977307	-6.322138	110	21	1.1	0.9;
	869	2	0	0	0	0	0	1.0326	4.106738	20	14	1.1	0.9;
	870	1	0	0	0	0.27	0	0.995628	27.676458	220	3	1.1	0.9;
	871	1	0	0	0	29.2	0	1.054603	34.712387	380	8	1.1	0.9;
	872	1	67.57	17	0	5.75	0	1.013659	21.90913	220	5	1.1	0.9;
	873	2	0	0	0	0	0	0.994173	92.075115	20	3	1.1	0.9;
	874	2	0	0	0	0	0	1.006137	-0.690843	20	2	1.1	0.9;
	875	1	47.55	5.3	0	7.59	0	1.014427	-23.263068	220	4	1.1	0.9;
	876	1	49.91	23.81	0	0.17	0	0.999588	10.969176	110	18	1.1	0.9;
	877	1	33.34	15.7	0	64.15	0	1.03178	-20.492791	380	4	1.1	0.9;
	878	1	0	-0	0	0.82	0	0.99162	-5.359677	150	9	1.1	0.9;
	879	1	-0	-0	0	37.635138	0	0.984454	4.505019	220	24	1.1	0.9;
	880	1	-0	0	0	8.72	0	1.011853	29.867832	380	3	1.1	0.9;
	881	1	140.6	42.6	0	5.71	0	1.019436	14.909228	220	5	1.1	0.9;
	882	1	110.8	-8.8	0	3.15	0	1.031652	-19.696271	220	4	1.1	0.9;
	883	1	-0	0	0.370251	146.807945	0	1.084383	6.111699	380	18	1.2	0.8;
	884	1	48.9	28.6	0	98.806507	0	1.132703	1.05655	380	9	1.2	0.8;
	885	1	33.24	4.4	0	24.54	0	0.990361	17.913132	380	5	1.1	0.9;
	886	1	23.4	3.3	0	1.91	0	1.002846	19.38659	154	1	1.1	0.9;
	887	1	0	0	0	19.56	0	1.026473	48.956035	380	3	1.1	0.9;
	888	2	0	0	0	0	0	1.055337	0.681282	20	9	1.1	0.9;
	889	1	150.9	10.4	0	4.77	0	1.063608	-3.825825	220	5	1.1	0.9;
	890	1	0	0	0	5.67	0	1.014413	14.29141	220	5	1.1	0.9;
	891	1	0	-0	0	0.91	0	0.973214	-28.659481	150	10	1.1	0.9;
	892	1	-0	0	0.094944	0.903407	0	1.004579	18.223427	380	7	1.1	0.9;
	893	1	-0	0	0	0.52	0	1.017581	-1.849378	110	3	1.1	0.9;
	894	1	163.5	44.9	0	1.17	0	1.034239	9.365432	220	5	1.1	0.9;
	895	1	18.5	3.7	0	0.01	0	0.966895	-14.773838	150	9	1.1	0.9;
	896	1	0	0	0.019388	0.670612	0	1.061137	18.700301	220	5	1.1	0.9;
	897	1	-0	-0	0.233419	2.686581	0	1.027683	16.404202	220	5	1.1	0.9;
	898	2	0	0	0	0	0	1.058434	6.52887	20	5	1.1	0.9;
	899	1	-0.92	-4.99	0	1.86	0	0.98985	-3.297046	110	18	1.1	0.9;
	900	1	0.9	74.1	0.47996	33.190034	0	1.015391	-21.569641	380	4	1.1	0.9;
	901	2	0	0	0	0	0	1.00657	25.464337	20	24	1.1	0.9;
	902	1	72.88	14.4	0	2.96	0	1.03356	4.612966	220	5	1.1	0.9;
	903	1	15.42	5.8	0	0.57	0	0.995983	3.433584	110	18	1.1	0.9;
	904	2	0	0	0	0	0	1.028707	4.250081	20	18	1.1	0.9;
	905	1	21.42	8.1	0	7.38	0	1.004501	-24.623949	220	4	1.1	0.9;
	906	1	0	-0	0	11.57	0	1.065604	86.499709	380	3	1.1	0.9;
	907	2	0	0	0	0	0	1.028707	4.250081	20	18	1.1	0.9;
	908	1	36.1	5.2	0	11.12	0	0.988945	27.024845	154	1	1.1	0.9;
	909	1	111.4	36.5	0	70.21	0	1.019902	16.981536	380	3	1.1	0.9;
	910	2	0	0	0	0	0	1.127961	6.657565	20	9	1.2	0.8;
	911	1	0	0	0	0	0	0.936045	-11.578054	150	9	1.1	0.9;
	912	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	913	1	0	0	0	0	0	0.970063	-11.531386	150	9	1.1	0.9;
	914	1	116.8	60.3	0	9.09	0	1.002176	-13.481147	150	9	1.1	0.9;
	915	1	0	0	0.122539	8.948096	0	0.977393	-7.174319	220	10	1.1	0.9;
	916	2	0	0	0	0	0	0.957523	42.691932	15	1	1.1	0.9;
	917	1	177.2	32.3	0	6.57	0	1.061641	9.414892	220	3	1.1	0.9;
	918	1	23.8	3.4	0	6.09	0	0.999527	26.548017	154	1	1.1	0.9;
	919	1	-0	-0	0	17.07	0	1.022407	34.785858	380	3	1.1	0.9;
	920	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	921	2	0	0	0	0	0	1.018066	-18.544118	20	4	1.1	0.9;
	922	1	40.04	16	0	0.52	0	0.996132	-1.925907	110	18	1.1	0.9;
	923	1	-0	0	0	0.23	0	1.094819	-4.683784	150	9	1.1	0.9;
	924	1	-0	0	0	36.493989	0	0.975171	15.860336	380	5	1.1	0.9;
	925	2	0	0	0	0	0	1.032265	43.171125	20	10	1.1	0.9;
	926	2	0	0	0	0	0	1.005341	-1.941399	20	4	1.1	0.9;
	927	2	0	0	0	0	0	1.007757	23.409003	11	1	1.1	0.9;
	928	1	0.2	0.9	0	0.04	0	1.075021	46.081147	220	3	1.1	0.9;
	929	1	7.91	2.5	0.058378	1.241622	0	1.033176	-1.797358	110	19	1.1	0.9;
	930	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	931	1	0	-0	0	1.18	0	0.993788	43.151154	154	1	1.1	0.9;
	932	2	0	0	0	0	0	1.023156	37.450201	20	10	1.1	0.9;
	933	1	181.41	16.52	0	4.07	0	1.035445	41.576564	220	13	1.1	0.9;
	934	1	179.9	-18.1	0.142401	12.4776	0	0.99149	-2.552543	220	5	1.1	0.9;
	935	1	102	26.6	0	2.31	0	1.014305	63.297249	220	3	1.1	0.9;
	936	1	45.32	18.5	0	0.74	0	1.009981	13.559945	220	5	1.1	0.9;
	937	2	0	0	0	0	0	1.026822	-0.476726	20	18	1.1	0.9;
	938	2	0	0	0	0	0	1.019144	17.562907	20	14	1.1	0.9;
	939	1	15.52	6.3	0	0.6	0	1.027438	24.929731	150	8	1.1	0.9;
	940	1	192.3	27.4	0	30.64	0	0.981378	4.54737	400	1	1.1	0.9;
	941	1	103	40.2	0	1.13	0	1.007452	18.316005	220	5	1.1	0.9;
	942	1	26.33	9.5	0	1.85	0	1.009333	-1.804582	110	19	1.1	0.9;
	943	1	-0.59	-0.83	0	5.79	0	1.05086	0.067818	150	9	1.1	0.9;
	944	1	48.05	15.8	0	6.08	0	0.996221	91.123997	150	1	1.1	0.9;
	945	1	0	-0	0.038966	3.071034	0	1.03579	79.927541	220	3	1.1	0.9;
	946	1	40.94	19.2	0	5.2	0	1.030451	-20.972645	220	4	1.1	0.9;
	947	1	17.22	4.4	0	1.71	0	1.021721	42.723795	110	8	1.1	0.9;
	948	1	62.46	7.15	0	10.06	0	1.00207	91.511646	150	1	1.1	0.9;
	949	1	-0	0	0	5.12	0	1.020449	16.959464	220	3	1.1	0.9;
	950	1	205.6	74	0	5.56	0	1.008763	4.218298	220	7	1.1	0.9;
	951	1	9.7	1.4	0	6.56	0	0.983636	43.579066	154	1	1.1	0.9;
	952	1	-0.15	-1.93	0	0	0	1.016274	49.480744	220	8	1.1	0.9;
	953	1	3.42	2.68	0	1.46	0	1.022132	3.908599	110	18	1.1	0.9;
	954	2	0	0	0	0	0	0.961175	47.68032	13.8	1	1.1	0.9;
	955	1	0	0	0	11.68	0	1.042596	13.310389	220	5	1.1	0.9;
	956	2	0	0	0	0	0	1.044097	40.791392	20	10	1.1	0.9;
	957	1	-0	-0	0	4.74	0	1.020566	-15.909789	380	4	1.1	0.9;
	958	2	0	0	0	0	0	1.016735	24.349916	11	1	1.1	0.9;
	959	1	65.27	30.9	0	3.1	0	1.005259	33.167214	220	3	1.1	0.9;
	960	1	0	0	0.058795	0.011205	0	1.016898	-5.759659	220	10	1.1	0.9;
	961	1	258.4	66.7	0	5.87	0	1.026224	11.981429	220	5	1.1	0.9;
	962	1	-0	0	0.187983	73.848631	0	1.044446	35.667348	380	2	1.1	0.9;
	963	1	9.91	3.3	0	0.66	0	0.986552	-16.814946	110	22	1.1	0.9;
	964	1	40.24	16	0	17.25	0	1.032136	55.592104	380	3	1.1	0.9;
	965	1	-0	0	0	0.6	0	1.040995	33.204924	380	5	1.1	0.9;
	966	1	14.82	0	0	4.07	0	1.050778	5.142571	220	7	1.1	0.9;
	967	1	-0	0	0.12	12.89	0	0.984956	-5.174055	220	21	1.1	0.9;
	968	1	38.04	11.7	0	0.18	0	0.989399	-5.35447	110	3	1.1	0.9;
	969	1	0	-0	0	7.2	0	1.029744	-0.176276	220	19	1.1	0.9;
	970	1	-5.61	-2.79	0	150.93	0	1.035791	23.926406	380	5	1.1	0.9;
	971	1	0	0	0	2.76	0	1.021972	-27.551339	220	11	1.1	0.9;
	972	1	48.8	6.9	0	5.14	0	0.977721	20.997949	154	1	1.1	0.9;
	973	1	-0	-0	0	29.29	0	1.035607	-8.450105	150	9	1.1	0.9;
	974	1	0	0	0	0	0	1.01079	-5.231958	150	9	1.1	0.9;
	975	2	0	0	0	0	0	1.042545	18.989939	20	5	1.1	0.9;
	976	2	0	0	0	0	0	1.070058	2.935613	20	5	1.1	0.9;
	977	2	0	0	0	0	0	1.070058	2.935613	20	5	1.1	0.9;
	978	2	0	0	0	0	0	1.070058	2.935613	20	5	1.1	0.9;
	979	1	0	-0	0	12.75	0	1.032776	38.238973	380	3	1.1	0.9;
	980	2	0	0	0	0	0	1.023796	19.440568	20	5	1.1	0.9;
	981	2	0	0	0	0	0	1.018701	14.159627	20	5	1.1	0.9;
	982	2	0	0	0	0	0	1.018701	14.159627	20	5	1.1	0.9;
	983	2	0	0	0	0	0	1.018701	14.159627	20	5	1.1	0.9;
	984	2	0	0	0	0	0	1.050192	6.863532	20	5	1.1	0.9;
	985	2	0	0	0	0	0	1.050192	6.863532	20	5	1.1	0.9;
	986	2	0	0	0	0	0	1.050192	6.863532	20	5	1.1	0.9;
	987	2	0	0	0	0	0	1.050192	6.863532	20	5	1.1	0.9;
	988	1	109.3	35.4	0	1.91	0	1.026665	32.869784	150	8	1.1	0.9;
	989	1	30.83	10.8	0	2.59	0	0.996033	-7.242929	110	18	1.1	0.9;
	990	1	232.1	89.8	0	0.09	0	0.964816	9.741815	220	5	1.1	0.9;
	991	1	-0	-0	0	3.94	0	1.029738	9.51492	220	1	1.1	0.9;
	992	1	79.59	50.5	0	14.31	0	1.035723	34.688997	380	3	1.1	0.9;
	993	1	0	-0	0	14.54	0	1.045616	-9.82138	380	1	1.1	0.9;
	994	1	-0	0	0.038933	0.331067	0	1.031548	18.747231	220	5	1.1	0.9;
	995	2	0	0	0	0	0	1.028283	16.586422	20	14	1.1	0.9;
	996	1	169.8	55.5	0	3.3	0	0.987703	34.817014	220	3	1.1	0.9;
	997	1	60.07	28	0	4.13	0	0.981668	-22.889071	220	6	1.1	0.9;
	998	1	-0	0	0	47.77	0	1.034713	19.620965	380	6	1.1	0.9;
	999	1	42.35	13.1	0.05007	3.85993	0	1.012803	-19.145251	220	4	1.1	0.9;
	1000	1	0	-0	0	5.6	0	1.052691	28.640951	220	5	1.1	0.9;
	1001	1	-3.4	-1.3	0	2.27	0	1.018883	4.573924	220	7	1.1	0.9;
	1002	2	0	0	0	0	0	1.026117	25.086822	20	6	1.1	0.9;
	1003	1	22.42	8.2	0	0.53	0	1.008785	-5.83331	110	3	1.1	0.9;
	1004	2	0	0	0	0	0	1.020732	7.56493	11	1	1.1	0.9;
	1005	1	6.11	2.2	0	0.88	0	0.955499	-2.810333	110	19	1.1	0.9;
	1006	1	83.6	11.9	0	3.53	0	0.994079	27.998088	154	1	1.1	0.9;
	1007	2	0	0	0	0	0	1.009902	7.124298	11	1	1.1	0.9;
	1008	1	-0	0	0.226936	-0.189114	0	1.053245	-7.911014	120	6	1.1	0.9;
	1009	1	0	-0	0	93.75	0	1.049527	-12.606328	380	4	1.1	0.9;
	1010	2	0	0	0	0	0	0.962614	39.578299	15.75	1	1.1	0.9;
	1011	1	0	-0	0.12	6.65	0	1.015816	0.326553	220	21	1.1	0.9;
	1012	1	96.41	26.5	0	0.08	0	1.001313	-10.409452	220	4	1.1	0.9;
	1013	2	0	0	0	0	0	1.001252	-14.682022	20	10	1.1	0.9;
	1014	1	12.4	6.89	0	0.63	0	1.059988	0.149576	150	9	1.1	0.9;
	1015	1	-0	0	0.466209	1.440918	0	1.027137	-6.465563	380	6	1.1	0.9;
	1016	1	0	0	0.180903	10.599297	0	0.995069	-1.871876	220	10	1.1	0.9;
	1017	1	0	-0	0	0	0	1.022632	4.381728	20	14	1.1	0.9;
	1018	1	36.54	8.4	0	6.58	0	1.026849	-25.366304	220	4	1.1	0.9;
	1019	1	62.7	9	0	3.22	0	1.038108	3.293597	154	1	1.1	0.9;
	1020	1	12.71	0.8	0	0.51	0	0.967495	-6.844112	110	18	1.1	0.9;
	1021	1	20.32	6.7	0	0.32	0	0.97453	-11.79808	110	12	1.1	0.9;
	1022	1	0	0	0	0.48	0	0.99761	-7.841022	150	10	1.1	0.9;
	1023	1	0	-0	0	7.84	0	1.122861	-0.191999	380	9	1.2	0.8;
	1024	1	-0	-0	0	3.14	0	1.043672	13.295119	220	5	1.1	0.9;
	1025	1	156.9	122.5	0	17.71	0	1.028176	19.200425	380	3	1.1	0.9;
	1026	1	114.8	22.8	0	5.66	0	1.007223	-16.523743	220	4	1.1	0.9;
	1027	1	83.99	21.2	0	5.36	0	1.062394	3.693466	220	5	1.1	0.9;
	1028	1	0	0	0	9.21	0	0.984812	-23.030576	220	6	1.1	0.9;
	1029	1	0	0	0.038515	61.381485	0	1.058308	-14.57333	380	4	1.1	0.9;
	1030	1	76.38	5.1	0	3.97	0	1.022228	-17.722424	220	4	1.1	0.9;
	1031	2	0	0	0	0	0	1.050891	3.953704	20	18	1.1	0.9;
	1032	1	423.4	215.2	0.474782	27.554131	0	1.050032	25.624075	380	14	1.1	0.9;
	1033	1	110.4	41.9	0	1.88	0	0.990296	27.156206	150	8	1.1	0.9;
	1034	1	-0	0	0.058557	-0.008557	0	0.993023	13.966977	110	18	1.1	0.9;
	1035	1	0	-0	0.471052	0.847084	0	1.058705	18.398371	380	14	1.1	0.9;
	1036	1	0	-0	0	113.86	0	0.981526	32.68627	400	1	1.1	0.9;
	1037	1	29.63	10.8	0	2.39	0	1.014304	-3.233621	110	19	1.1	0.9;
	1038	1	4.5	-5.6	0	0.09	0	1.013009	-1.260469	110	18	1.1	0.9;
	1039	1	135.11	41	0	5.31	0	1.025298	4.11833	220	5	1.1	0.9;
	1040	1	-0.54	-0.99	0	1.13	0	1.058232	-0.595237	220	18	1.2	0.8;
	1041	1	0	0	0	23.550494	0	0.922503	1.030887	380	5	1.1	0.9;
	1042	1	0	0	0	1.49	0	1.026918	3.536883	220	7	1.1	0.9;
	1043	2	0	0	0	0	0	1.010998	1.300317	20	10	1.1	0.9;
	1044	2	0	0	0	0	0	1.019396	-15.910525	20	4	1.1	0.9;
	1045	1	-0	-0	0	0	0	1.015757	-19.850393	20	4	1.1	0.9;
	1046	2	0	0	0	0	0	1.014635	-15.220088	20	4	1.1	0.9;
	1047	1	-0	0	0	0.78	0	1.030726	27.099334	150	8	1.1	0.9;
	1048	2	0	0	0	0	0	1.024095	1.173322	11	1	1.1	0.9;
	1049	1	104.6	14.9	0	7.35	0	0.999693	30.670692	154	1	1.1	0.9;
	1050	2	0	0	0	0	0	1.048227	33.464235	20	14	1.1	0.9;
	1051	1	30.83	6.6	0	3.19	0	0.990482	-14.344882	110	22	1.1	0.9;
	1052	1	-0	0	0	34.82	0	1.126918	-1.53611	380	9	1.2	0.8;
	1053	1	400.3	189.9	0	0.53	0	1.043296	27.768258	220	5	1.1	0.9;
	1054	1	34.94	-32.3	0.418254	37.168883	0	1.120384	9.693407	380	19	1.2	0.8;
	1055	2	0	0	0	0	0	1.043492	-15.322718	20	4	1.1	0.9;
	1056	1	-0	0	0	2.17	0	0.995341	38.515307	154	1	1.1	0.9;
	1057	2	0	0	0	0	0	1.055479	63.485669	20	3	1.1	0.9;
	1058	1	20.62	6.5	0	0.72	0	0.988428	-15.328352	110	12	1.1	0.9;
	1059	2	0	0	0	0	0	0.98379	29.90696	10.6	1	1.1	0.9;
	1060	1	-0	0	0	61.48	0	1.017165	-22.642005	380	4	1.1	0.9;
	1061	1	0	0	0	11.12	0	1.028126	44.738341	380	3	1.1	0.9;
	1062	1	58.36	10.7	0	8.41	0	1.006682	35.480523	110	8	1.1	0.9;
	1063	1	109.23	66.36	0	5.44	0	1.022947	24.940474	220	13	1.1	0.9;
	1064	1	0	-0	0	4.58	0	0.997082	42.307024	154	1	1.1	0.9;
	1065	1	45.56	-1.51	0	19.94	0	0.988217	88.927697	150	1	1.1	0.9;
	1066	1	8.11	2	0	1.95	0	0.99727	-6.171382	110	18	1.1	0.9;
	1067	1	11.01	3.3	0.178397	1.531336	0	0.996576	-13.394939	110	22	1.1	0.9;
	1068	1	1.3	59.6	0.29248	128.206266	0	1.032844	-14.822365	380	4	1.1	0.9;
	1069	1	22.22	13.7	0	0.12	0	1.009755	-7.535974	150	9	1.1	0.9;
	1070	2	0	0	0	0	0	1.01964	-10.829105	20	10	1.1	0.9;
	1071	2	0	0	0	0	0	1.031397	6.88995	20	14	1.1	0.9;
	1072	2	0	0	0	0	0	1.065506	1.014062	20	9	1.1	0.9;
	1073	2	0	0	0	0	0	1.037189	3.219879	20	2	1.1	0.9;
	1074	2	0	0	0	0	0	1.061099	-1.375053	20	5	1.1	0.9;
	1075	1	0	0	0	2.83	0	1.05286	28.639443	220	5	1.1	0.9;
	1076	1	5.01	2	0	2.99	0	1.044639	-4.02157	220	23	1.1	0.9;
	1077	2	0	0	0	0	0	1.015556	16.601372	20	5	1.1	0.9;
	1078	1	6.51	4.1	0	1.27	0	1.036704	0.145279	150	9	1.1	0.9;
	1079	2	0	0	0	0	0	1.037501	2.321125	20	2	1.1	0.9;
	1080	1	90.4	15.2	0	0.07	0	0.994214	15.255805	220	6	1.1	0.9;
	1081	1	2.5	1.3	0	0.07	0	0.953342	-6.246594	110	18	1.1	0.9;
	1082	1	69.58	16	0	3.03	0	1.02325	3.383664	220	5	1.1	0.9;
	1083	2	0	0	0	0	0	1.035499	26.862046	20	6	1.1	0.9;
	1084	1	-0	-0	0	5.2	0	1.045184	-18.36115	220	6	1.1	0.9;
	1085	2	0	0	0	0	0	0.98883	21.250891	19	1	1.1	0.9;
	1086	1	16.52	7.8	0.116495	10.493505	0	1.064193	2.672742	220	5	1.1	0.9;
	1087	1	0	-0	0	18.63	0	1.027715	5.739639	220	5	1.1	0.9;
	1088	2	0	0	0	0	0	1.015045	0.518146	11	1	1.1	0.9;
	1089	2	0	0	0	0	0	1.039754	25.994186	20	14	1.1	0.9;
	1090	2	0	0	0	0	0	1.026413	2.766853	20	2	1.1	0.9;
	1091	1	2.5	1.3	0	1.55	0	0.920705	-7.854239	110	18	1.1	0.9;
	1092	2	0	0	0	0	0	1.039754	25.994186	20	14	1.1	0.9;
	1093	2	0	0	0	0	0	1.039754	25.994186	20	14	1.1	0.9;
	1094	1	-0	0	0	4.15	0	1.021323	0.157034	220	7	1.1	0.9;
	1095	1	1.9	0.6	0	0.38	0	0.996303	-3.626208	110	18	1.1	0.9;
	1096	2	0	0	0	0	0	1.030566	-6.658147	20	9	1.1	0.9;
	1097	1	132.1	32.4	0	58.29	0	1.027843	-17.473372	380	4	1.1	0.9;
	1098	1	28.9	4.2	0	2.09	0	1.017668	17.689564	154	1	1.1	0.9;
	1099	1	36.74	8.9	0	0.63	0	0.989962	5.425776	110	18	1.1	0.9;
	1100	1	-0	0	0.060092	103.572441	0	1.013156	16.173778	380	5	1.1	0.9;
	1101	1	0	0	0	0	0	0.970954	-14.134822	150	9	1.1	0.9;
	1102	1	-0	-0	0	9.58	0	1.012771	10.113246	220	6	1.1	0.9;
	1103	1	0	0	0	1.5	0	1.033938	7.498563	110	18	1.1	0.9;
	1104	2	0	0	0	0	0	1.043492	-15.322718	20	4	1.1	0.9;
	1105	1	-0	0	0	0.08	0	1.051619	-6.41508	150	9	1.1	0.9;
	1106	1	0	-0	0.038914	0.380814	0	1.030345	-0.824449	220	6	1.1	0.9;
	1107	1	223.4	107.8	0.23317	7.325407	0	1.027071	22.872057	220	13	1.1	0.9;
	1108	1	0	-0	0.120942	-0.040706	0	1.003921	-2.768952	220	6	1.1	0.9;
	1109	1	-0	0	0	7.85	0	0.988123	9.386666	220	5	1.1	0.9;
	1110	1	0	0	0	0	0	0.92416	-12.234897	150	9	1.1	0.9;
	1111	1	-0	-0	0	0.45	0	1.000291	0.966435	150	9	1.1	0.9;
	1112	1	0	0	0	36.92	0	1.013882	44.861905	380	8	1.1	0.9;
	1113	1	152.57	40	0	6.12	0	1.016356	21.039881	220	5	1.1	0.9;
	1114	1	-0	-0	0	1.23	0	1.02902	25.380845	220	13	1.1	0.9;
	1115	1	46.15	20.9	0	1.4	0	1.004379	-14.102643	110	22	1.1	0.9;
	1116	1	343.9	141.9	0	5.55	0	1.022419	14.92737	220	5	1.1	0.9;
	1117	1	349.6	96.3	0	105.72	0	1.020247	-15.259037	380	6	1.1	0.9;
	1118	1	-14.87	-4.17	0	0.58	0	0.982088	1.918784	110	18	1.1	0.9;
	1119	1	-0	-0	0	1.54	0	1.030206	34.64258	154	1	1.1	0.9;
	1120	1	-55.14	-27.14	0	0	0	1.025958	34.069666	150	8	1.1	0.9;
	1121	1	45.05	19.2	0	0.21	0	0.987687	2.389894	110	18	1.1	0.9;
	1122	1	142.2	29.6	0.020088	4.449912	0	1.002251	15.884286	220	5	1.1	0.9;
	1123	2	0	0	0	0	0	1.038747	-8.642337	20	4	1.1	0.9;
	1124	1	-0	-0	0	1.44	0	1.029801	7.874787	154	1	1.1	0.9;
	1125	2	0	0	0	0	0	1.018739	-6.896129	20	9	1.1	0.9;
	1126	1	-0	0	0.019802	4.940198	0	1.016001	22.347716	220	5	1.1	0.9;
	1127	1	-0	0	0.510485	8.332978	0	1.059631	33.365886	380	5	1.1	0.9;
	1128	1	0	0	0.019363	0.980637	0	1.036479	43.646869	380	3	1.1	0.9;
	1129	1	0	0	0.040037	-3.7e-05	0	0.996569	-17.371589	220	10	1.1	0.9;
	1130	1	34.04	11	0	0.49	0	0.988367	-17.850399	110	22	1.1	0.9;
	1131	1	0	0	0	7.28	0	1.022658	8.751926	220	3	1.1	0.9;
	1132	1	0	0	0	0	0	0.986182	-8.108772	150	9	1.1	0.9;
	1133	1	-26	-14.72	0	0	0	1.033532	31.669766	150	8	1.1	0.9;
	1134	1	0	0	0	4.03	0	1.02575	-6.012669	220	10	1.1	0.9;
	1135	1	0	-0	0	0.79	0	1.049746	7.525607	220	1	1.1	0.9;
	1136	1	-0	-0	0	2.47	0	0.979901	31.362904	220	3	1.1	0.9;
	1137	1	18.02	6.4	0	0.79	0	0.950327	-6.911106	110	18	1.1	0.9;
	1138	2	0	0	0	0	0	0.971901	45.235162	15.75	1	1.1	0.9;
	1139	1	-0	0	0	6.63	0	1.044684	23.296473	220	5	1.1	0.9;
	1140	1	0	0	0	0.65	0	1.01876	2.813545	150	7	1.1	0.9;
	1141	1	52.06	-0	0.683815	82.738646	0	1.054949	20.005722	380	13	1.1	0.9;
	1142	1	295.3	41.9	0	0	0	1.053847	28.546328	150	2	1.1	0.9;
	1143	1	9.7	1.4	0	4.22	0	0.993025	40.432571	154	1	1.1	0.9;
	1144	1	0	0	0	1.653516	0	0.982128	10.906524	220	24	1.1	0.9;
	1145	1	-0	0	0	1.86	0	1.021827	60.814659	110	8	1.1	0.9;
	1146	1	0	-0	0.039425	2.320575	0	1.004752	-16.633178	110	22	1.1	0.9;
	1147	1	2	114.6	0	16.15	0	1.007501	-12.066805	220	4	1.1	0.9;
	1148	2	0	0	0	0	0	1.020087	-16.009175	20	4	1.1	0.9;
	1149	1	235.2	41.1	0	10.01	0	1.004523	39.504964	380	3	1.1	0.9;
	1150	2	0	0	0	0	0	0.970854	29.381452	15.75	1	1.1	0.9;
	1151	1	9.71	3	0	0.95	0	0.957655	-8.78103	110	21	1.1	0.9;
	1152	1	200.4	60.5	0.080512	8.569488	0	1.028416	4.295836	220	5	1.1	0.9;
	1153	1	-0	-0	0	7.26	0	1.037027	21.722482	220	5	1.1	0.9;
	1154	1	312.2	28	0	48.6	0	0.98901	-3.370618	220	5	1.1	0.9;
	1155	1	-0	0	0	32.02	0	1.016314	33.74591	380	8	1.1	0.9;
	1156	1	222.5	262.8	1.155017	156.145612	0	1.053231	-10.776146	380	4	1.1	0.9;
	1157	1	24.53	8.6	0	7.01	0	1.012197	-23.005874	220	4	1.1	0.9;
	1158	1	0	0	0	58.081854	0	1.030199	31.958924	380	5	1.1	0.9;
	1159	1	0	0	0	0	0	1.017792	-5.319138	150	9	1.1	0.9;
	1160	1	17.83	-9.5	0	2.79	0	1.01649	6.344203	220	7	1.1	0.9;
	1161	1	0	0	0	66.22	0	1.033769	4.919429	400	1	1.1	0.9;
	1162	1	68.08	13.6	0.039413	27.260587	0	1.068779	5.406898	220	5	1.1	0.9;
	1163	1	287.7	16.3	0	12.86	0	1.009126	-23.142825	380	4	1.1	0.9;
	1164	1	25.27	5.71	0	0.67	0	1.069974	3.588654	110	19	1.1	0.9;
	1165	1	234.7	-12	0	23.35	0	1.049403	-6.829279	380	17	1.1	0.9;
	1166	1	333.1	17.5	0	18.86	0	0.996409	-17.049556	380	6	1.1	0.9;
	1167	1	-0	0	0	8.73	0	1.033993	79.013264	220	3	1.1	0.9;
	1168	1	0	-0	0	0.22	0	1.040321	28.79688	220	5	1.1	0.9;
	1169	1	12.6	1.8	0	1.89	0	1.005897	24.832965	154	1	1.1	0.9;
	1170	1	15.02	6	0	6.06	0	1.020687	4.488546	220	3	1.1	0.9;
	1171	1	-0	0	0	22.93	0	1.099354	-14.35941	150	9	1.2	0.8;
	1172	1	-16.84	-7.96	0	53.85	0	1.025808	70.814725	380	3	1.1	0.9;
	1173	1	-0	-0	1.147464	-0.494081	0	1.04707	28.776876	380	5	1.1	0.9;
	1174	1	32.1	4.5	0	8.93	0	0.998348	22.168889	154	1	1.1	0.9;
	1175	1	38.64	23.9	0	0.94	0	0.999171	-8.851257	150	9	1.1	0.9;
	1176	1	31.94	7.7	0	1.11	0	1.017158	-5.196209	110	22	1.1	0.9;
	1177	1	167.8	77	0	12.66	0	1.00878	14.885564	380	3	1.1	0.9;
	1178	1	0	0	0	0.01	0	1.003491	27.632662	154	1	1.1	0.9;
	1179	1	113.6	122.7	0	8.74	0	1.010326	40.703694	380	3	1.1	0.9;
	1180	2	0	0	0	0	0	0.999061	42.441878	11	1	1.1	0.9;
	1181	1	0.01	0.01	0.358775	5.700953	0	0.999224	-22.970532	220	6	1.1	0.9;
	1182	2	0	0	0	0	0	1.056688	67.221009	20	3	1.1	0.9;
	1183	2	0	0	0	0	0	1.056688	67.221009	20	3	1.1	0.9;
	1184	1	15.02	10	0	0.55	0	1.037346	-2.833684	110	23	1.1	0.9;
	1185	1	-17.17	-7.1	0	57.020535	0	1.011067	18.293892	380	5	1.1	0.9;
	1186	1	0	0	0	24.4	0	0.985304	-3.668301	380	6	1.1	0.9;
	1187	1	0	-0	0	56.7	0	0.996718	2.387562	400	1	1.1	0.9;
	1188	1	-0	-0	0.03986	2.050105	0	1.010639	-8.661899	220	6	1.1	0.9;
	1189	2	0	0	0	0	0	1.029009	32.575777	20	5	1.1	0.9;
	1190	1	0	-0	0	8.17514	0	1.025637	30.618892	380	5	1.1	0.9;
	1191	1	-0	-0	0	2.7	0	1.026575	-1.348248	150	9	1.1	0.9;
	1192	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	1193	1	-0	-0	0.135375	32.572175	0	1.080724	1.741895	380	18	1.2	0.8;
	1194	1	0	0	0.349307	35.121782	0	1.017625	5.374756	380	7	1.1	0.9;
	1195	1	-0	-0	0	2.5	0	1.022798	25.191563	220	3	1.1	0.9;
	1196	1	-0	-0	0	0.64	0	0.994841	-14.026789	110	1	1.1	0.9;
	1197	1	-0	-0	0.05873	0.55127	0	1.057724	18.27348	380	14	1.1	0.9;
	1198	1	-0	0	0	1.11	0	1.008562	-8.919415	150	9	1.1	0.9;
	1199	1	159.2	59.9	0	37.33	0	1.055006	11.790162	380	16	1.1	0.9;
	1200	1	13.9	1.9	0	7.37	0	0.987606	46.494111	154	1	1.1	0.9;
	1201	1	62.77	0.7	0.078743	2.090943	0	0.993849	89.074217	150	1	1.1	0.9;
	1202	2	0	0	0	0	0	1.045213	9.250128	20	14	1.1	0.9;
	1203	1	0	-0	0.019624	56.1	0	1.031185	-14.185965	380	4	1.1	0.9;
	1204	1	18.5	2.7	0	0.11	0	1.012343	31.295134	154	1	1.1	0.9;
	1205	1	-0	-0	0	0.47	0	0.955981	-5.03467	110	18	1.1	0.9;
	1206	1	34.44	4.4	0.10038	1.41962	0	0.979015	11.019297	220	5	1.1	0.9;
	1207	1	0	0	0	5.46	0	1.034431	-1.078525	154	1	1.1	0.9;
	1208	2	0	0	0	0	0	1.010291	28.950447	20	18	1.1	0.9;
	1209	1	190	43.3	0	12.94	0	1.021946	16.613523	220	3	1.1	0.9;
	1210	1	49.95	7.5	0	8.86	0	1.014877	-26.886832	220	4	1.1	0.9;
	1211	1	10.01	3.3	0	0.24	0	0.999457	-8.828526	110	12	1.1	0.9;
	1212	1	45.05	50.5	0.039751	5.610187	0	0.979058	89.609227	150	1	1.1	0.9;
	1213	1	-28.93	9.5	0	0.38	0	0.989029	0.920029	150	9	1.1	0.9;
	1214	1	21.72	4.8	0	1.26	0	0.960525	-8.071165	110	18	1.1	0.9;
	1215	1	0	-0	0.574719	1.882173	0	1.036118	33.13745	380	5	1.1	0.9;
	1216	1	0	0	0	0.28	0	1.099831	-4.401904	150	9	1.2	0.8;
	1217	1	242.08	59.98	0	5.6	0	1.026955	10.877686	220	5	1.1	0.9;
	1218	2	0	0	0	0	0	0.989939	94.137681	20	3	1.1	0.9;
	1219	1	0	-0	0.058902	0.151098	0	1.012402	-5.282433	110	22	1.1	0.9;
	1220	1	40.44	13.3	0	0.94	0	0.970825	-5.981165	110	21	1.1	0.9;
	1221	1	111.4	15.9	0	2.36	0	0.991182	17.232684	154	1	1.1	0.9;
	1222	1	1.3	0.6	0	1.23	0	0.954047	-6.852207	110	18	1.1	0.9;
	1223	1	43.45	26.9	0	3.56	0	1.090479	-4.891767	150	9	1.1	0.9;
	1224	1	0	-0	0	20.07	0	1.017267	70.448807	220	3	1.1	0.9;
	1225	1	25.13	14	0	0.04	0	1.01747	24.913441	220	3	1.1	0.9;
	1226	1	13.82	5	0	0.11	0	1.023929	-0.379788	110	19	1.1	0.9;
	1227	1	94.2	5.5	0	6.51	0	1.033419	-19.31843	220	4	1.1	0.9;
	1228	1	108.89	-21.76	0	3.72	0	1.025589	17.0989	220	5	1.1	0.9;
	1229	1	-0	0	0	42.212329	0	0.980304	12.228304	380	5	1.1	0.9;
	1230	1	-0	0	0	0	0	1.016217	20.317809	20	5	1.1	0.9;
	1231	1	-0	0	0	0	0	1.016217	20.317809	20	5	1.1	0.9;
	1232	1	-0	0	0	0	0	1.016217	20.317809	20	5	1.1	0.9;
	1233	1	0	-0	0.295925	2.563236	0	1.027014	-3.607711	380	6	1.1	0.9;
	1234	1	42.25	13.9	0	0.95	0	0.999152	-3.843739	110	21	1.1	0.9;
	1235	1	75.68	18.4	0.050059	6.269941	0	1.012217	-15.655535	220	4	1.1	0.9;
	1236	1	0	0	0	25.28	0	1.033007	35.671238	150	8	1.1	0.9;
	1237	1	0	0	0	0	0	1.038979	-13.937788	20	4	1.1	0.9;
	1238	2	0	0	0	0	0	1.003807	31.8753	11	1	1.1	0.9;
	1239	1	0	-0	0	2	0	1.010888	23.125198	220	6	1.1	0.9;
	1240	1	-0	0	0	1.47	0	1.024087	-0.54486	150	9	1.1	0.9;
	1241	1	394.9	77	0	44.93	0	1.00962	-16.927071	380	6	1.1	0.9;
	1242	1	192.5	78.3	0	6.65	0	1.016551	19.011695	220	5	1.1	0.9;
	1243	1	13.82	5	0	0.63	0	1.04846	-0.047812	110	19	1.1	0.9;
	1244	1	23.03	14.3	0	6.61	0	1.001857	-9.078486	150	9	1.1	0.9;
	1245	2	0	0	0	0	0	0.977562	26.05322	13.8	1	1.1	0.9;
	1246	1	-2.85	-7.24	0	0.07	0	1.025868	35.252167	220	3	1.1	0.9;
	1247	1	51.86	19.2	0	1.07	0	0.997624	-3.342561	220	6	1.1	0.9;
	1248	1	17.42	11.9	0.293436	14.435105	0	1.037647	0.367988	220	19	1.1	0.9;
	1249	1	3.9	1.4	0	0.78	0	1.023959	-3.085896	110	19	1.1	0.9;
	1250	1	8.91	3.8	0	0.38	0	0.984611	-5.611707	110	18	1.1	0.9;
	1251	1	-0	0	1.414448	10.672156	0	1.047302	10.700556	380	5	1.1	0.9;
	1252	1	452.5	113.2	0.059992	16.270008	0	1.034819	1.718515	220	5	1.1	0.9;
	1253	1	-0	-0	0	1.77	0	1.024092	-4.135271	220	6	1.1	0.9;
	1254	1	0	0	0	1.32	0	0.983565	-8.783815	150	10	1.1	0.9;
	1255	1	-0	-0	0.234854	36.333138	0	1.064545	-1.597101	380	21	1.1	0.9;
	1256	1	0	-0	0	1.81	0	1.025328	17.022252	154	1	1.1	0.9;
	1257	2	0	0	0	0	0	1.03036	-6.512869	20	4	1.1	0.9;
	1258	1	153.5	21.9	0	7.79	0	1.034351	3.127292	154	1	1.1	0.9;
	1259	1	25.63	7.9	0	0.77	0	1.016681	43.976275	110	8	1.1	0.9;
	1260	1	0	-0	0	59.04	0	1.026315	15.249077	380	6	1.1	0.9;
	1261	1	-0	0	0	0.15	0	1.01688	13.461135	220	5	1.1	0.9;
	1262	2	0	0	0	0	0	1.048632	39.237883	20	2	1.1	0.9;
	1263	2	0	0	0	0	0	1.017546	-7.215576	20	4	1.1	0.9;
	1264	2	0	0	0	0	0	1.002671	41.912372	15.75	1	1.1	0.9;
	1265	1	0	0	0	0.05	0	1.032343	4.317268	220	5	1.1	0.9;
	1266	1	82.69	20.1	0	1.36	0	1.023208	3.35577	220	5	1.1	0.9;
	1267	1	146.9	60.6	0	26.49	0	0.992997	-3.956098	220	6	1.1	0.9;
	1268	1	-0	0	0.13389	0.01611	0	1.050784	25.734127	220	5	1.1	0.9;
	1269	1	0	0	0	18.004129	0	1.031159	14.124866	380	5	1.1	0.9;
	1270	1	0	0	0	0.23	0	0.964568	-8.270907	150	9	1.1	0.9;
	1271	1	0	0	0	0.05	0	1.023486	19.424705	220	5	1.1	0.9;
	1272	1	26.53	7.1	0	11.02	0	1.003523	-3.772589	220	6	1.1	0.9;
	1273	2	0	0	0	0	0	1.044317	35.206283	20	5	1.1	0.9;
	1274	2	0	0	0	0	0	1.036065	-16.582357	20	4	1.1	0.9;
	1275	1	125.9	-0.5	0	0.21	0	1.036266	21.408499	220	5	1.1	0.9;
	1276	1	12.61	4.5	0	2.63	0	1.049499	-1.562398	110	15	1.1	0.9;
	1277	1	6.3	0.9	0	3.82	0	1.014588	19.116104	154	1	1.1	0.9;
	1278	2	0	0	0	0	0	1.055337	0.681282	20	9	1.1	0.9;
	1279	1	13.21	2.2	0	10.05	0	0.998779	-14.538696	110	22	1.1	0.9;
	1280	1	48.8	6.9	0	6.16	0	1.011658	18.310559	154	1	1.1	0.9;
	1281	1	22.02	13.2	0	0.03	0	1.09787	-4.489169	150	9	1.1	0.9;
	1282	1	0	0	0.997827	10.627573	0	1.02307	35.928234	380	2	1.1	0.9;
	1283	1	0	0	0	89.51	0	1.080673	-3.230149	380	9	1.1	0.9;
	1284	1	24.53	7.9	0	0.63	0	1.031863	-3.799969	110	19	1.1	0.9;
	1285	1	0.8	18.6	0.197474	3.171894	0	1.011397	-10.092906	220	4	1.1	0.9;
	1286	1	-0	-0	0.019316	0.720684	0	1.059588	2.630814	110	15	1.1	0.9;
	1287	2	0	0	0	0	0	1.052164	7.333527	20	5	1.1	0.9;
	1288	2	0	0	0	0	0	1.024762	17.972121	20	14	1.1	0.9;
	1289	1	-0	0	0	13.89	0	1.011456	38.709997	380	8	1.1	0.9;
	1290	2	0	0	0	0	0	1.019482	16.777108	20	5	1.1	0.9;
	1291	1	34.44	8.5	0	1.11	0	1.033392	9.500925	220	5	1.1	0.9;
	1292	1	-0	-0	0.126343	6.455031	0	0.988535	-1.526728	220	19	1.1	0.9;
	1293	1	15.22	7.6	0	1.56	0	0.978392	5.399845	110	18	1.1	0.9;
	1294	1	48.95	14.9	0	4.04	0	1.024075	6.410383	220	5	1.1	0.9;
	1295	1	10.01	5	0	0.11	0	0.987567	-22.138945	220	6	1.1	0.9;
	1296	1	8.01	4.1	0	2.25	0	0.964203	7.059293	110	18	1.1	0.9;
	1297	1	-0	-0	0	110.25	0	0.978053	24.541317	400	1	1.1	0.9;
	1298	2	0	0	0	0	0	1.052164	7.333527	20	5	1.1	0.9;
	1299	1	0	0	0.038839	-0.029129	0	1.033373	-18.785108	220	6	1.1	0.9;
	1300	1	112.9	19.3	0	17.34	0	0.99014	24.93683	150	8	1.1	0.9;
	1301	1	43.75	12.4	0	9.35	0	1.032787	-18.511488	220	4	1.1	0.9;
	1302	1	30.03	15	0	0.81	0	1.036835	-1.697863	110	23	1.1	0.9;
	1303	1	1.1	61.2	0.524876	46.291763	0	1.040676	-9.782077	380	4	1.1	0.9;
	1304	2	0	0	0	0	0	1.037973	11.471538	20	14	1.1	0.9;
	1305	2	0	0	0	0	0	1.022099	17.218586	20	14	1.1	0.9;
	1306	2	0	0	0	0	0	1.023144	-9.538309	20	4	1.1	0.9;
	1307	1	64	9.1	0	12.23	0	0.977439	38.340653	154	1	1.1	0.9;
	1308	2	0	0	0	0	0	1.127182	5.800414	20	9	1.2	0.8;
	1309	1	474.3	203.4	0	47.469681	0	1.013598	-10.219682	380	6	1.1	0.9;
	1310	2	0	0	0	0	0	1.043619	-8.436397	20	4	1.1	0.9;
	1311	1	77.59	12.3	0	5.1	0	1.013231	-26.409956	220	4	1.1	0.9;
	1312	1	40.34	25	0	1.14	0	0.955741	-10.11486	150	9	1.1	0.9;
	1313	1	0	0	0	17.82055	0	1.025483	1.905603	220	20	1.1	0.9;
	1314	1	0	-0	0	22.75	0	0.994242	7.715079	220	5	1.1	0.9;
	1315	1	-0	0	0	6.66	0	1.02791	-15.503278	220	4	1.1	0.9;
	1316	1	2.5	1.3	0	2.33	0	0.993604	-4.183455	110	18	1.1	0.9;
	1317	1	0	-0	0	9.46	0	1.037231	5.419113	220	5	1.1	0.9;
	1318	1	-0	0	0	1.93	0	1.003098	-4.680224	150	9	1.1	0.9;
	1319	1	-12.51	4.1	0	0.09	0	1.005178	-10.412822	150	9	1.1	0.9;
	1320	1	0	-0	0	31.47	0	1.063111	64.17397	380	3	1.1	0.9;
	1321	1	150.1	35	0	3.55	0	1.009846	4.913756	220	6	1.1	0.9;
	1322	1	50.06	16.3	0	0.26	0	1.022037	-4.256844	110	21	1.1	0.9;
	1323	2	0	0	0	0	0	1.128032	6.903218	20	9	1.2	0.8;
	1324	2	0	0	0	0	0	1.055337	0.681282	20	9	1.1	0.9;
	1325	1	26.03	9.8	0	0.82	0	0.983431	-1.771365	110	18	1.1	0.9;
	1326	1	-0	-0	0	34.26	0	1.066503	7.436005	380	1	1.1	0.9;
	1327	1	-0	0	0	3.37	0	1.016711	6.114524	220	7	1.1	0.9;
	1328	1	30.33	11	0	1.43	0	1.024026	0.390871	110	19	1.1	0.9;
	1329	2	0	0	0	0	0	1.009821	29.021748	6.3	1	1.1	0.9;
	1330	2	0	0	0	0	0	1.024544	18.974877	20	14	1.1	0.9;
	1331	1	0.2	0.1	0	0.5	0	1.005198	-7.161553	110	12	1.1	0.9;
	1332	1	189.1	-11.9	0	27.462397	0	1.005039	-20.950342	380	6	1.1	0.9;
	1333	1	7.41	1.8	0	0.3	0	1.005156	-3.750223	110	19	1.1	0.9;
	1334	1	90.7	5.8	0	4.1	0	1.029972	-21.872489	220	4	1.1	0.9;
	1335	2	0	0	0	0	0	1.0608	42.720858	20	3	1.1	0.9;
	1336	1	4.81	2.2	0	0.62	0	0.95238	-7.290511	110	18	1.1	0.9;
	1337	1	69.08	22.6	0	0.54	0	0.989854	24.705399	150	8	1.1	0.9;
	1338	2	0	0	0	0	0	1.02611	9.397426	20	18	1.1	0.9;
	1339	2	0	0	0	0	0	1.02611	9.397426	20	18	1.1	0.9;
	1340	1	0.5	-271.9	0	3.12	0	1.040865	12.654119	380	3	1.1	0.9;
	1341	1	-0	0	0	0.38	0	1.08342	0.64159	150	9	1.1	0.9;
	1342	2	0	0	0	0	0	1.02611	9.397426	20	18	1.1	0.9;
	1343	2	0	0	0	0	0	1.02611	9.397426	20	18	1.1	0.9;
	1344	2	0	0	0	0	0	1.00657	25.464337	20	24	1.1	0.9;
	1345	1	108.7	-12.15	0	37.11	0	1.039121	17.644693	380	5	1.1	0.9;
	1346	1	66.27	13.6	0	0.67	0	1.034028	8.40309	220	5	1.1	0.9;
	1347	1	-13.91	4.6	0	0	0	0.965753	-9.552607	150	9	1.1	0.9;
	1348	2	0	0	0	0	0	1.02611	9.397426	20	18	1.1	0.9;
	1349	2	0	0	0	0	0	1.031144	-5.517648	20	18	1.1	0.9;
	1350	1	244.2	152.7	0	0	0	1.01497	24.847751	150	2	1.1	0.9;
	1351	1	19.42	11.6	0	0	0	1.015493	16.769648	110	24	1.1	0.9;
	1352	1	65.77	16.1	0	1.66	0	1.009437	-18.638291	220	4	1.1	0.9;
	1353	1	-0	0	0	51.29	0	1.051369	-2.55004	380	11	1.1	0.9;
	1354	1	0	0	0	0	0	1.026675	25.143797	150	2	1.1	0.9;
	1355	1	166.6	8.1	0	10.11	0	1.007341	13.840393	380	3	1.1	0.9;
	1356	1	0.01	-0.01	0	108.728602	0	1.058598	20.351152	380	5	1.1	0.9;
	1357	1	32.94	20.4	0	8.95	0	0.88398	-14.103043	150	9	1.2	0.8;
	1358	1	42.55	17.1	0	1.67	0	1.014293	-18.380601	220	4	1.1	0.9;
	1359	2	0	0	0	0	0	1.008948	-4.613451	20	10	1.1	0.9;
	1360	1	0	-0	0	12.74	0	1.003115	-6.90371	220	12	1.1	0.9;
	1361	1	88.5	16.5	0	146.44	0	1.013428	18.272141	380	5	1.1	0.9;
	1362	1	45.05	9.9	0	0.24	0	0.992611	-14.416547	110	22	1.1	0.9;
	1363	1	190.91	55.46	0	14.77	0	0.994236	40.473758	220	13	1.1	0.9;
	1364	1	1.3	0.4	0	0.64	0	0.981067	1.057278	110	18	1.1	0.9;
	1365	1	0	-0	0	49.26	0	0.964474	34.963089	400	1	1.1	0.9;
	1366	1	0.4	1.1	0	1.99	0	1.017076	57.811893	380	3	1.1	0.9;
	1367	2	0	0	0	0	0	1.031502	0.465782	20	6	1.1	0.9;
	1368	1	62.27	-2.1	0	8.01	0	1.008419	-14.737948	220	4	1.1	0.9;
	1369	2	0	0	0	0	0	1.032265	43.171125	20	10	1.1	0.9;
	1370	1	100.9	2.9	0.029966	6.760034	0	1.027241	-22.307748	220	4	1.1	0.9;
	1371	1	-0	-0	0.099501	68.573595	0	1.078275	5.990479	380	18	1.2	0.8;
	1372	2	0	0	0	0	0	1.018066	-18.544118	20	4	1.1	0.9;
	1373	1	0	-0	1.049905	7.793635	0	1.084648	34.31875	380	5	1.2	0.8;
	1374	1	0	0	0	0	0	1.017753	-8.026498	150	9	1.1	0.9;
	1375	1	6.71	2.2	0.058526	1.221474	0	1.006855	-8.554992	110	12	1.1	0.9;
	1376	2	0	0	0	0	0	1.027131	-1.211905	20	18	1.1	0.9;
	1377	1	0	-0	0.05054	43.31946	0	1.014793	-22.29244	380	4	1.1	0.9;
	1378	1	0	-0	0.211519	1.118481	0	1.038249	11.383163	220	7	1.1	0.9;
	1379	1	95.58	-14	0	4.48	0	1.033035	14.850214	220	13	1.1	0.9;
	1380	1	-0	-0	0.009683	0.180317	0	1.011107	-2.897343	110	18	1.1	0.9;
	1381	2	0	0	0	0	0	1.030174	-14.576007	20	4	1.1	0.9;
	1382	2	0	0	0	0	0	1.013179	14.179628	20	18	1.1	0.9;
	1383	1	-0	-0	0.181549	5.810889	0	1.029442	-1.667248	220	18	1.1	0.9;
	1384	1	21.12	7.7	0	1.48	0	0.94375	-2.230118	110	19	1.1	0.9;
	1385	1	14.72	4.6	0	7.88	0	1.033997	4.729511	220	5	1.1	0.9;
	1386	1	0	-0	0	0.84	0	1.011473	-16.503789	220	4	1.1	0.9;
	1387	1	62.7	9	0	2.94	0	0.994546	32.929681	154	1	1.1	0.9;
	1388	1	13.8	461.3	1.731993	405.420277	0	1.033278	-13.106385	380	4	1.1	0.9;
	1389	2	0	0	0	0	0	1.03558	-5.743469	20	6	1.1	0.9;
	1390	2	0	0	0	0	0	1.03558	-5.743469	20	6	1.1	0.9;
	1391	2	0	0	0	0	0	1.03558	-5.743469	20	6	1.1	0.9;
	1392	1	0	0	0	0.31	0	1.0184	-6.821964	110	1	1.1	0.9;
	1393	1	-0	0	1.111071	86.974909	0	1.090026	71.311631	380	3	1.2	0.8;
	1394	2	0	0	0	0	0	1.03858	30.747075	20	10	1.1	0.9;
	1395	1	-0	0	0.933942	3.100629	0	1.027954	-6.383546	380	6	1.1	0.9;
	1396	1	17.52	4.6	0	11.82	0	1.030911	-12.637403	220	6	1.1	0.9;
	1397	1	15.98	0.86	0	5.96	0	1.011721	-25.197742	220	4	1.1	0.9;
	1398	1	0	-0	0	1.18	0	0.993773	-11.20036	150	10	1.1	0.9;
	1399	1	-0	0	0	8.62	0	1.014538	-25.40241	220	4	1.1	0.9;
	1400	1	32.34	10.6	0	0.64	0	0.957686	-7.394537	110	21	1.1	0.9;
	1401	2	0	0	0	0	0	1.015161	-11.543898	20	4	1.1	0.9;
	1402	1	-0	0	0	8.19	0	0.996571	-13.145129	150	9	1.1	0.9;
	1403	1	18.12	5.6	0.019734	2.520266	0	0.996751	-1.363677	110	18	1.1	0.9;
	1404	1	38.34	51.8	0.039559	5.52033	0	1.010936	2.529477	220	6	1.1	0.9;
	1405	1	171	69.2	0.020113	28.479887	0	0.990787	28.512684	380	13	1.1	0.9;
	1406	1	0	0	0	25	0	1.029731	25.844994	150	8	1.1	0.9;
	1407	2	0	0	0	0	0	1.042865	51.92649	20	3	1.1	0.9;
	1408	2	0	0	0	0	0	1.042865	51.92649	20	3	1.1	0.9;
	1409	1	-0	-0	0	7.66	0	1.003865	-9.5568	220	11	1.1	0.9;
	1410	2	0	0	0	0	0	1.045666	52.068402	20	3	1.1	0.9;
	1411	2	0	0	0	0	0	1.045666	52.068402	20	3	1.1	0.9;
	1412	2	0	0	0	0	0	1.045666	52.068402	20	3	1.1	0.9;
	1413	2	0	0	0	0	0	1.045666	52.068402	20	3	1.1	0.9;
	1414	2	0	0	0	0	0	1.042865	51.92649	20	3	1.1	0.9;
	1415	2	0	0	0	0	0	1.042865	51.92649	20	3	1.1	0.9;
	1416	2	0	0	0	0	0	1.045666	52.068402	20	3	1.1	0.9;
	1417	2	0	0	0	0	0	1.042865	51.92649	20	3	1.1	0.9;
	1418	1	0	0	0	1.8	0	1.018252	-5.565917	220	1	1.1	0.9;
	1419	2	0	0	0	0	0	1.03758	22.028378	20	10	1.1	0.9;
	1420	1	0	-0	0.059141	23.820859	0	1.046338	-4.535742	220	5	1.1	0.9;
	1421	2	0	0	0	0	0	0.996766	22.457835	6.3	1	1.1	0.9;
	1422	1	-0	0	0.098001	0.521999	0	1.037267	10.219486	220	5	1.1	0.9;
	1423	2	0	0	0	0	0	1.033534	-10.544863	20	4	1.1	0.9;
	1424	1	24.03	8	0	0.33	0	0.953323	-8.205349	110	21	1.1	0.9;
	1425	1	0	-0	0.060427	0.599644	0	0.997713	33.606331	220	3	1.1	0.9;
	1426	1	-0	0	0	0.81	0	1.043663	17.125447	220	5	1.1	0.9;
	1427	2	0	0	0	0	0	1.046359	90.322887	20	3	1.1	0.9;
	1428	1	-0	0	0	45.31	0	1.092968	-0.268371	380	9	1.1	0.9;
	1429	1	0	-0	0	1.51	0	1.020081	-4.546391	150	9	1.1	0.9;
	1430	1	-5.01	0	0	3.54	0	0.991566	-24.906856	220	6	1.1	0.9;
	1431	1	0	-0	0	11.4	0	1.017696	24.858482	220	3	1.1	0.9;
	1432	1	0.2	0.4	0	0.73	0	1.020393	17.568548	220	3	1.1	0.9;
	1433	2	0	0	0	0	0	1.033648	19.100507	20	5	1.1	0.9;
	1434	2	0	0	0	0	0	1.033648	19.100507	20	5	1.1	0.9;
	1435	2	0	0	0	0	0	1.048683	24.799749	20	5	1.1	0.9;
	1436	2	0	0	0	0	0	1.048683	24.799749	20	5	1.1	0.9;
	1437	2	0	0	0	0	0	1.033648	19.100507	20	5	1.1	0.9;
	1438	2	0	0	0	0	0	1.033648	19.100507	20	5	1.1	0.9;
	1439	1	-0	-0	0	0	0	1.040698	17.14354	20	5	1.1	0.9;
	1440	2	0	0	0	0	0	1.04081	21.523284	20	5	1.1	0.9;
	1441	2	0	0	0	0	0	1.048683	24.799749	20	5	1.1	0.9;
	1442	1	-0	-0	0	0	0	1.040698	17.14354	20	5	1.1	0.9;
	1443	1	136.06	40.4	0	12.08	0	0.992777	-19.070079	220	4	1.1	0.9;
	1444	1	0	-0	0	1.07	0	1.013902	-8.72938	150	9	1.1	0.9;
	1445	1	44.55	-0	0	4.78	0	1.047149	23.914598	220	5	1.1	0.9;
	1446	1	-0	0	0	0	0	0.993422	-9.294213	150	9	1.1	0.9;
	1447	2	0	0	0	0	0	1.030174	-14.576007	20	4	1.1	0.9;
	1448	2	0	0	0	0	0	1.030174	-14.576007	20	4	1.1	0.9;
	1449	2	0	0	0	0	0	1.030174	-14.576007	20	4	1.1	0.9;
	1450	2	0	0	0	0	0	1.030174	-14.576007	20	4	1.1	0.9;
	1451	2	0	0	0	0	0	1.021906	-21.605261	20	4	1.1	0.9;
	1452	1	-0	-0	0	2.35	0	0.99768	60.908647	220	1	1.1	0.9;
	1453	2	0	0	0	0	0	1.021906	-21.605261	20	4	1.1	0.9;
	1454	2	0	0	0	0	0	1.021906	-21.605261	20	4	1.1	0.9;
	1455	2	0	0	0	0	0	1.009892	4.868208	20	9	1.1	0.9;
	1456	1	8.2	3.3	0	8.38	0	1.055477	33.90241	150	8	1.1	0.9;
	1457	1	227.9	36.1	0.060304	13.069696	0	1.007444	-18.937606	220	4	1.1	0.9;
	1458	2	0	0	0	0	0	1.046359	90.322887	20	3	1.1	0.9;
	1459	1	0	0	0	0.32	0	0.97474	5.213673	110	18	1.1	0.9;
	1460	1	-0	-0	0	0.93	0	1.013161	-26.710349	220	4	1.1	0.9;
	1461	1	27.9	4	0	0.29	0	1.004067	12.245503	154	1	1.1	0.9;
	1462	2	0	0	0	0	0	1.036065	-16.582357	20	4	1.1	0.9;
	1463	1	241.4	-2.5	0.224624	31.280331	0	1.013499	-24.806539	380	10	1.1	0.9;
	1464	1	201.4	53	0	9.67	0	1.022649	0.085134	220	5	1.1	0.9;
	1465	1	-5.91	-2.33	0	30.388612	0	0.977422	12.295759	380	5	1.1	0.9;
	1466	1	13.82	5	0	0.45	0	1.008172	-3.314025	110	19	1.1	0.9;
	1467	1	0	-0	0.113517	5.325402	0	1.035848	92.435397	220	3	1.1	0.9;
	1468	1	178.9	9.1	0	15.71	0	1.024015	-15.769618	220	4	1.1	0.9;
	1469	2	0	0	0	0	0	1.01541	24.000055	6.3	1	1.1	0.9;
	1470	2	0	0	0	0	0	1.06527	-0.439984	20	5	1.1	0.9;
	1471	2	0	0	0	0	0	1.06527	-0.439984	20	5	1.1	0.9;
	1472	1	-34.54	2.7	0.078778	9.951222	0	1.010036	-8.472091	220	4	1.1	0.9;
	1473	2	0	0	0	0	0	1.06527	-0.439984	20	5	1.1	0.9;
	1474	2	0	0	0	0	0	1.06527	-0.439984	20	5	1.1	0.9;
	1475	2	0	0	0	0	0	1.06527	-0.439984	20	5	1.1	0.9;
	1476	2	0	0	0	0	0	1.06527	-0.439984	20	5	1.1	0.9;
	1477	2	0	0	0	0	0	1.063432	5.558066	20	5	1.1	0.9;
	1478	2	0	0	0	0	0	1.063432	5.558066	20	5	1.1	0.9;
	1479	2	0	0	0	0	0	1.063432	5.558066	20	5	1.1	0.9;
	1480	1	0	-0	0	22.91	0	1.018379	-11.027788	150	9	1.1	0.9;
	1481	1	0	-0	0	0	0	1.007081	13.329704	20	5	1.1	0.9;
	1482	1	0	-0	0	0	0	1.007081	13.329704	20	5	1.1	0.9;
	1483	1	0	-0	0	0	0	1.007081	13.329704	20	5	1.1	0.9;
	1484	2	0	0	0	0	0	1.060798	-1.229323	20	5	1.1	0.9;
	1485	2	0	0	0	0	0	1.060798	-1.229323	20	5	1.1	0.9;
	1486	2	0	0	0	0	0	1.060798	-1.229323	20	5	1.1	0.9;
	1487	1	0	-0	0	14.23	0	1.003412	0.732753	220	6	1.1	0.9;
	1488	2	0	0	0	0	0	1.003171	37.39829	6.3	1	1.1	0.9;
	1489	1	382.5	57.9	0	8.16	0	1.030584	-3.714139	220	5	1.1	0.9;
	1490	1	-0	-0	0.173429	1.295476	0	1.032515	-3.516223	220	10	1.1	0.9;
	1491	1	0	-0	0	0.55	0	1.042169	34.593975	380	2	1.1	0.9;
	1492	1	3.7	-0	0	0.64	0	1.008724	2.168065	110	18	1.1	0.9;
	1493	1	34.64	-0	0	0.91	0	1.019002	9.546797	220	5	1.1	0.9;
	1494	1	31.94	8.8	0	3.79	0	1.027605	20.158794	220	5	1.1	0.9;
	1495	1	12.11	4.4	0	0.95	0	0.984285	-17.22273	110	22	1.1	0.9;
	1496	1	2.5	0.6	0	0.61	0	0.979544	-5.959906	110	18	1.1	0.9;
	1497	1	34.2	4.9	0	9.08	0	1.013843	19.287793	154	1	1.1	0.9;
	1498	1	0	0	0	11.44	0	1.033336	3.342046	154	1	1.1	0.9;
	1499	1	0	0	0.540258	-0.140201	0	1.00705	25.577191	380	5	1.1	0.9;
	1500	1	40.5	5.7	0	3.2	0	1.001203	30.380734	154	1	1.1	0.9;
	1501	1	47.05	19.3	0	0.04	0	1.018456	22.124078	150	8	1.1	0.9;
	1502	1	1.9	0.8	0	0.65	0	0.983972	-5.211149	110	18	1.1	0.9;
	1503	1	27.8	-2.03	0	0.5	0	0.98652	-3.592922	110	18	1.1	0.9;
	1504	1	0	0	0	28.8	0	1.136154	0.822895	380	1	1.2	0.8;
	1505	2	0	0	0	0	0	1.025426	-15.273241	20	4	1.1	0.9;
	1506	1	28.19	-23.24	0.226891	11.120378	0	1.046538	34.171044	220	13	1.1	0.9;
	1507	1	0	0	0	3.79	0	1.018281	16.808324	220	5	1.1	0.9;
	1508	1	67.47	32.6	0	13.11	0	1.043518	54.330098	220	8	1.1	0.9;
	1509	1	84.99	19.6	0	4.32	0	1.023648	3.396089	220	5	1.1	0.9;
	1510	1	83.6	11.9	0	6.56	0	1.017682	29.642897	154	1	1.1	0.9;
	1511	1	0	-0	0	0	0	1.018239	-19.841296	20	4	1.1	0.9;
	1512	1	-0	-0	0	29.27	0	1.012964	-12.27724	380	4	1.1	0.9;
	1513	1	0	0	0	0	0	0.963689	-9.802775	150	9	1.1	0.9;
	1514	2	0	0	0	0	0	1.029166	13.522836	20	14	1.1	0.9;
	1515	1	0	0	0	0	0	1.056035	-6.130078	150	9	1.1	0.9;
	1516	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	1517	2	0	0	0	0	0	1.033944	81.658005	20	3	1.1	0.9;
	1518	2	0	0	0	0	0	1.033944	81.658005	20	3	1.1	0.9;
	1519	2	0	0	0	0	0	1.033944	81.5989	20	3	1.1	0.9;
	1520	2	0	0	0	0	0	1.033944	81.5989	20	3	1.1	0.9;
	1521	2	0	0	0	0	0	1.055774	84.490729	20	3	1.1	0.9;
	1522	1	-0	-0	0	11.96	0	0.995923	22.396868	380	3	1.1	0.9;
	1523	1	0	0	0	62.76	0	1.080104	-3.304065	380	9	1.1	0.9;
	1524	1	0	-0	0	75.38	0	1.053202	-14.326704	380	4	1.1	0.9;
	1525	1	0.01	0	0	0	0	0.995889	-9.119979	150	9	1.1	0.9;
	1526	2	0	0	0	0	0	1.014332	26.817672	10.5	1	1.1	0.9;
	1527	1	84.89	32.4	0	7.77	0	1.019936	7.010383	220	7	1.1	0.9;
	1528	1	50.06	17	0	1.61	0	1.004165	-2.237835	110	21	1.1	0.9;
	1529	1	215.2	-7.6	0	1.24	0	1.03218	-19.625729	220	4	1.1	0.9;
	1530	1	89.6	20	0	9.07	0	1.014984	-26.876942	220	4	1.1	0.9;
	1531	1	0	-0	0.332671	-0.258744	0	1.076788	36.804416	150	8	1.1	0.9;
	1532	2	0	0	0	0	0	1.012556	-8.920589	20	10	1.1	0.9;
	1533	1	-0	0	0	9.67	0	1.010467	48.767303	380	3	1.1	0.9;
	1534	2	0	0	0	0	0	1.053776	40.435567	20	3	1.1	0.9;
	1535	1	279.4	22.9	0	15.24	0	1.069767	18.174762	220	3	1.1	0.9;
	1536	1	35.24	11.6	0	1.72	0	0.995702	-9.051189	110	12	1.1	0.9;
	1537	1	0	0	0	0	0	1.063704	66.981987	20	8	1.1	0.9;
	1538	1	-0	-0	0	8.72	0	1.01392	8.701747	220	1	1.1	0.9;
	1539	1	1.6	-0	0	0.44	0	1.009185	2.56099	110	18	1.1	0.9;
	1540	1	83.3	47.9	0	6.19	0	1.006756	-15.1197	150	9	1.1	0.9;
	1541	1	0	-0	0	68.81	0	1.027472	15.203439	380	6	1.1	0.9;
	1542	1	21.92	7.2	0	3.14	0	1.065223	1.152691	110	19	1.1	0.9;
	1543	1	403.9	-0.1	0	143.1	0	1.036277	-15.753777	380	4	1.1	0.9;
	1544	1	0	0	0	0	0	1.063704	66.981987	20	8	1.1	0.9;
	1545	1	0	0	0	0	0	1.063704	66.981987	20	8	1.1	0.9;
	1546	2	0	0	0	0	0	1.051916	71.34914	20	8	1.1	0.9;
	1547	1	0	0	0	0	0	0.966992	-14.766387	150	9	1.1	0.9;
	1548	1	0	0	0	0.1	0	1.026748	18.911757	220	5	1.1	0.9;
	1549	1	-0	0	0	4.74	0	1.055824	25.80864	220	5	1.1	0.9;
	1550	1	0.2	1.3	0.355669	10.384433	0	1.009685	30.34199	380	3	1.1	0.9;
	1551	1	48.8	6.9	0	7.95	0	1.029094	15.098629	154	1	1.1	0.9;
	1552	2	0	0	0	0	0	1.043138	41.982485	20	10	1.1	0.9;
	1553	2	0	0	0	0	0	1.027348	16.396901	20	14	1.1	0.9;
	1554	1	277.86	58.95	0	182.880644	0	1.032971	19.158498	380	5	1.1	0.9;
	1555	1	-0	-0	0.239685	3.640263	0	1.01419	-0.817409	380	6	1.1	0.9;
	1556	2	0	0	0	0	0	1.069569	35.273027	20	14	1.1	0.9;
	1557	2	0	0	0	0	0	1.069569	35.273027	20	14	1.1	0.9;
	1558	2	0	0	0	0	0	1.069569	35.273027	20	14	1.1	0.9;
	1559	2	0	0	0	0	0	1.113563	6.086123	20	20	1.2	0.8;
	1560	2	0	0	0	0	0	1.059142	21.764381	20	14	1.1	0.9;
	1561	2	0	0	0	0	0	1.066067	59.272605	20	10	1.1	0.9;
	1562	2	0	0	0	0	0	0.995346	36.752154	6.3	1	1.1	0.9;
	1563	1	-0	-0	0.058263	-0.008263	0	1.033443	20.750736	220	5	1.1	0.9;
	1564	1	-30.03	9.9	0	0	0	0.982474	0.471501	150	9	1.1	0.9;
	1565	1	42.15	14.2	0	1.71	0	1.050774	-0.576722	110	19	1.1	0.9;
	1566	1	162.1	43.2	0	3.32	0	1.04051	19.367506	220	3	1.1	0.9;
	1567	1	12.01	5.6	0	39.29	0	1.003765	-20.136541	380	4	1.1	0.9;
	1568	1	0	-0	0	3.75	0	0.962628	-7.419291	110	21	1.1	0.9;
	1569	1	39.1	5.5	0	3.82	0	1.01263	30.120443	154	1	1.1	0.9;
	1570	1	-7.61	2.5	0	0.13	0	1.03946	-1.166173	150	9	1.1	0.9;
	1571	1	56.06	18.7	0	4.24	0	1.009978	9.05433	220	5	1.1	0.9;
	1572	1	-24.85	-3.56	0	1.31	0	1.016271	7.799267	110	18	1.1	0.9;
	1573	1	30.93	-46.6	0	1.62	0	1.018638	-8.917122	220	6	1.1	0.9;
	1574	1	-0	-0	0.710079	-0.032146	0	1.026966	21.39997	380	5	1.1	0.9;
	1575	1	28.53	11.3	0	3.25	0	1.049448	33.310799	150	8	1.1	0.9;
	1576	1	0	0	0	43.02	0	1.107591	6.852032	380	1	1.2	0.8;
	1577	2	0	0	0	0	0	1.026206	-2.974992	20	6	1.1	0.9;
	1578	1	-0	0	0	0.16	0	1.037603	28.404087	220	5	1.1	0.9;
	1579	1	0	-0	0	4	0	1.00483	17.539976	154	1	1.1	0.9;
	1580	1	46.25	21	0.116588	2.262843	0	1.033691	0.886126	110	19	1.1	0.9;
	1581	1	0	0	0	0	0	1.007312	-8.549741	150	9	1.1	0.9;
	1582	1	0	0	0.039225	0.280775	0	1.027792	25.041453	220	5	1.1	0.9;
	1583	1	0	0	0	2.05	0	1.024352	14.858869	220	5	1.1	0.9;
	1584	1	0	0	0	0.32	0	0.982878	-5.663054	110	18	1.1	0.9;
	1585	1	13.31	8.1	0.455742	65.649204	0	1.072416	54.322097	380	13	1.1	0.9;
	1586	2	0	0	0	0	0	1.055063	40.300168	20	3	1.1	0.9;
	1587	2	0	0	0	0	0	1.055063	40.300168	20	3	1.1	0.9;
	1588	1	0	-0	0	1.83	0	1.035624	-7.38671	150	9	1.1	0.9;
	1589	1	192.4	63.9	0	10	0	1.034925	31.775405	150	8	1.1	0.9;
	1590	1	1.8	0.6	0	0.52	0	0.948702	-2.365664	110	19	1.1	0.9;
	1591	1	-0	-0	0	0.27	0	1.012625	-16.317081	220	4	1.1	0.9;
	1592	1	-0	-0	0	0	0	1.014963	35.494625	380	8	1.1	0.9;
	1593	1	-0	-0	0	59.43	0	1.033167	35.997496	380	8	1.1	0.9;
	1594	2	0	0	0	0	0	1.004901	24.768889	10.5	1	1.1	0.9;
	1595	1	-0	0	0	0.4	0	1.050827	-6.680648	150	9	1.1	0.9;
	1596	1	-54.06	13.1	0	14.23	0	1.021094	-7.217557	220	10	1.1	0.9;
	1597	1	227.34	236.13	0	66.84	0	1.020965	38.434761	380	13	1.1	0.9;
	1598	1	0	0	0	1.1	0	1.013796	21.907219	220	5	1.1	0.9;
	1599	2	0	0	0	0	0	1.030843	4.085708	6.3	1	1.1	0.9;
	1600	1	10.91	3.6	0	0.4	0	1.005344	-8.260142	110	12	1.1	0.9;
	1601	1	20.9	3	0	5.46	0	0.994517	24.909354	154	1	1.1	0.9;
	1602	1	59.27	5.1	0	26.03	0	1.030956	-10.923484	380	4	1.1	0.9;
	1603	1	7.01	1	0	0.92	0	1.004339	0.968887	150	7	1.1	0.9;
	1604	1	-40.55	9.3	0	3.05	0	1.027648	-11.776071	220	4	1.1	0.9;
	1605	1	70.28	58.9	0	7.33	0	1.009428	40.291154	380	3	1.1	0.9;
	1606	2	0	0	0	0	0	1.127948	6.872839	20	9	1.2	0.8;
	1607	1	-20.94	-6.16	0	11.99	0	1.001877	13.978057	220	5	1.1	0.9;
	1608	1	0	-0	0	6.12	0	0.969986	37.585796	400	1	1.1	0.9;
	1609	1	0	-0	0	12.28	0	1.013688	41.65503	380	3	1.1	0.9;
	1610	1	22.22	13.8	0	3.11	0	1.030589	-7.348342	150	9	1.1	0.9;
	1611	1	244.8	119.3	0	3.78	0	1.016514	14.597678	220	3	1.1	0.9;
	1612	1	11.41	36.1	0.236629	9.002529	0	1.024652	-11.689645	220	4	1.1	0.9;
	1613	1	-0	-0	0.098235	8.531765	0	1.029447	4.381728	220	7	1.1	0.9;
	1614	1	-9	-10.19	0	0	0	1.02904	42.831627	150	8	1.1	0.9;
	1615	1	-0	-0	0.039398	30.749631	0	1.003602	94.235585	380	1	1.1	0.9;
	1616	1	0	0	0.116733	7.623267	0	1.011301	-5.438701	220	12	1.1	0.9;
	1617	1	64.21	5.59	0	2.57	0	1.018117	-24.674837	220	4	1.1	0.9;
	1618	1	-0	-0	0	0	0	0.98032	29.982954	400	1	1.1	0.9;
	1619	1	0	-0	0.086633	0.514111	0	1.07975	11.116625	220	18	1.2	0.8;
	1620	1	-0	-0	0.475577	28.946605	0	1.043596	20.487381	380	3	1.1	0.9;
	1621	1	190.47	55.03	0	6.7	0	1.003202	8.213746	220	5	1.1	0.9;
	1622	1	22.5	3.2	0	5.11	0	1.006471	24.363227	154	1	1.1	0.9;
	1623	1	0	-0	0.540344	-0.140268	0	1.024326	15.431611	380	6	1.1	0.9;
	1624	1	7.8	50.9	0	87.765989	0	1.110428	4.110338	380	9	1.2	0.8;
	1625	1	115.2	319.2	0	290.91	0	1.031122	-16.175337	380	4	1.1	0.9;
	1626	1	118.2	-33	0	12.02	0	0.999166	6.82436	220	3	1.1	0.9;
	1627	2	0	0	0	0	0	1.050142	-0.087597	20	23	1.1	0.9;
	1628	1	103.9	14.8	0	1.91	0	1.018402	14.715203	154	1	1.1	0.9;
	1629	2	0	0	0	0	0	1.026298	27.062483	20	10	1.1	0.9;
	1630	1	1.4	0.6	0	0.41	0	0.962206	-0.774182	110	18	1.1	0.9;
	1631	2	0	0	0	0	0	1.052654	21.51137	20	5	1.1	0.9;
	1632	1	16.22	2	0	0.71	0	1.031924	11.492212	220	5	1.1	0.9;
	1633	1	-0	-0	0	0.05	0	1.03559	5.457408	220	5	1.1	0.9;
	1634	1	0	-0	0.526084	0.520916	0	1.098165	34.420929	380	14	1.2	0.8;
	1635	1	18.92	9	0.116297	-0.026297	0	1.026075	0.540039	220	21	1.1	0.9;
	1636	2	0	0	0	0	0	1.033848	35.240634	20	10	1.1	0.9;
	1637	1	-0	0	0	1.78	0	1.005728	-22.029535	220	6	1.1	0.9;
	1638	1	0	-0	0	23.36	0	1.039203	24.660347	380	7	1.1	0.9;
	1639	1	45.9	36.8	0	55.224909	0	1.080579	-2.979957	380	9	1.1	0.9;
	1640	2	0	0	0	0	0	0.994858	32.812983	15	1	1.1	0.9;
	1641	2	0	0	0	0	0	1.01282	1.386847	20	18	1.1	0.9;
	1642	2	0	0	0	0	0	1.012131	24.048464	11	1	1.1	0.9;
	1643	1	0	0	0.210571	8.306088	0	1.026591	-0.512107	220	19	1.1	0.9;
	1644	2	0	0	0	0	0	1.042135	37.953176	11	1	1.1	0.9;
	1645	1	39.04	-5	0	1.31	0	1.00771	62.900668	220	3	1.1	0.9;
	1646	1	-62.07	-37	0.267849	0.328679	0	1.038782	4.48626	220	11	1.1	0.9;
	1647	1	32.84	0	0	1.41	0	1.037487	24.088495	220	5	1.1	0.9;
	1648	1	20.02	15	0	1.56	0	0.987725	33.846731	220	3	1.1	0.9;
	1649	1	141.4	123	0	8.52	0	1.008538	29.987076	380	3	1.1	0.9;
	1650	1	0	0	0	0.32	0	1.039688	-10.445191	150	9	1.1	0.9;
	1651	1	0	-0	0	24.6	0	1.028392	52.702197	380	3	1.1	0.9;
	1652	1	0	-0	0	46.01	0	0.948903	26.443293	400	1	1.1	0.9;
	1653	1	12.41	3.9	0	2.23	0	1.029276	26.312713	220	5	1.1	0.9;
	1654	1	55.7	7.9	0	2.62	0	1.029758	-0.312213	154	1	1.1	0.9;
	1655	1	-0	-0	0	1.91	0	1.024295	-24.350532	380	10	1.1	0.9;
	1656	2	0	0	0	0	0	1.062519	23.472046	20	5	1.1	0.9;
	1657	1	65.67	24	0	9.95	0	1.031004	8.006228	220	5	1.1	0.9;
	1658	1	-0	0	0	14.19	0	1.000694	-7.123335	220	6	1.1	0.9;
	1659	1	3.7	0.6	0	0.01	0	0.993012	40.431099	154	1	1.1	0.9;
	1660	1	24.13	12.7	0	0.1	0	0.986622	-8.004861	110	18	1.1	0.9;
	1661	2	0	0	0	0	0	1.051456	33.495027	20	5	1.1	0.9;
	1662	2	0	0	0	0	0	1.027716	27.965045	20	5	1.1	0.9;
	1663	1	-0	0	0	0	0	1.022433	23.470814	20	5	1.1	0.9;
	1664	1	0.1	0.5	0	0.48	0	0.995602	27.675294	220	3	1.1	0.9;
	1665	1	170.2	60.8	0	16.27	0	1.031097	34.909723	380	3	1.1	0.9;
	1666	1	0	0	0	2.22	0	1.017928	7.402342	220	1	1.1	0.9;
	1667	1	-0	0	0	47.06	0	1.015463	32.987394	150	8	1.1	0.9;
	1668	1	0	0	0	16.04	0	1.007044	32.287282	380	2	1.1	0.9;
	1669	1	26.43	16.4	0	0.02	0	0.997706	-9.08371	150	9	1.1	0.9;
	1670	1	-0	-0	0	14.13	0	1.036348	43.634336	380	3	1.1	0.9;
	1671	1	-39.81	-10.9	0	3.37	0	1.036105	-7.389707	150	9	1.1	0.9;
	1672	1	0	0	0	0	0	1.023538	-11.335143	150	9	1.1	0.9;
	1673	1	-0	-0	0	89.06	0	1.032768	-14.901123	380	4	1.1	0.9;
	1674	1	116.2	16.5	0	7.8	0	1.027028	1.510678	154	1	1.1	0.9;
	1675	2	0	0	0	0	0	1.050379	33.293543	20	5	1.1	0.9;
	1676	1	-9.51	-3.32	0	0	0	0.989504	-26.003725	220	6	1.1	0.9;
	1677	1	79.89	29	0	1.42	0	1.03233	-3.90576	110	19	1.1	0.9;
	1678	1	-0	0	0	0.1	0	1.026263	-4.032324	110	19	1.1	0.9;
	1679	1	0	-0	0	2.3	0	1.024381	4.28593	220	5	1.1	0.9;
	1680	1	129.1	5.8	0	3.59	0	1.023576	-14.051827	220	4	1.1	0.9;
	1681	1	0	0	0.019971	0.640029	0	0.998143	-10.564266	150	10	1.1	0.9;
	1682	1	-128.74	38	0.155644	13.673267	0	1.043684	-12.356638	220	4	1.1	0.9;
	1683	1	92.7	-11.6	0.116434	0.152972	0	1.016748	-0.472433	110	18	1.1	0.9;
	1684	1	-2.35	-2.06	0	3.38	0	1.046981	83.09022	220	3	1.1	0.9;
	1685	2	0	0	0	0	0	1.010291	28.950447	20	18	1.1	0.9;
	1686	1	86.6	23.4	0	0.39	0	1.049391	28.038123	220	5	1.1	0.9;
	1687	1	43.75	16.8	0	0.42	0	1.029649	-5.32138	110	3	1.1	0.9;
	1688	1	0.1	-19.3	0	0.25	0	1.010338	7.101232	220	3	1.1	0.9;
	1689	2	0	0	0	0	0	1.043738	28.126856	20	5	1.1	0.9;
	1690	1	25.1	3.6	0	6.2	0	1.019192	18.214373	154	1	1.1	0.9;
	1691	1	-0	-0	0.232698	1.796085	0	1.030474	19.396926	380	3	1.1	0.9;
	1692	1	150.4	15.3	0	8.09	0	1.007571	-16.567631	220	4	1.1	0.9;
	1693	1	62.7	9	0	3.18	0	0.997319	27.375421	154	1	1.1	0.9;
	1694	1	67.58	21.2	0	5.71	0	1.043979	24.689649	220	5	1.1	0.9;
	1695	1	30.93	6.6	0	9.09	0	1.048376	15.262753	220	5	1.1	0.9;
	1696	1	96.41	18.1	0	2.04	0	1.029257	65.057244	220	3	1.1	0.9;
	1697	1	14.62	5.6	0	0.25	0	0.996164	2.095783	110	18	1.1	0.9;
	1698	1	37.3	5.4	0	0.12	0	0.987249	13.896936	154	1	1.1	0.9;
	1699	1	13.21	9.2	0	0.69	0	0.976428	-3.019212	110	18	1.1	0.9;
	1700	1	181.5	21.2	0	64.45	0	1.034556	18.42067	380	6	1.1	0.9;
	1701	2	0	0	0	0	0	1.036966	15.051213	20	14	1.1	0.9;
	1702	1	129.6	13.3	0	13.98	0	1.030863	14.68831	380	5	1.1	0.9;
	1703	1	5.5	0.8	0	1.19	0	1.006316	26.135381	154	1	1.1	0.9;
	1704	1	-0	-0	0	1	0	1.08513	1.171749	150	9	1.1	0.9;
	1705	1	-0	-0	0.181549	8.450889	0	1.035803	-3.943509	220	18	1.1	0.9;
	1706	1	25.23	10.6	0	1.44	0	0.980003	-5.079566	110	18	1.1	0.9;
	1707	1	-0	0	0	30.644542	0	0.974808	13.364665	380	5	1.1	0.9;
	1708	1	-0	-0	0.052811	29.652975	0	1.052451	2.236915	380	19	1.1	0.9;
	1709	1	124.8	7.9	0	48.99	0	1.006512	-13.785134	380	4	1.1	0.9;
	1710	1	114.1	23.5	0	1.08	0	1.030846	9.19717	220	5	1.1	0.9;
	1711	1	0	-0	0	46.28	0	1.127285	-1.495208	380	9	1.2	0.8;
	1712	1	41.8	5.9	0	1.75	0	1.015535	16.842725	154	1	1.1	0.9;
	1713	1	-0	0	0	0.72	0	1.06612	-4.564989	150	9	1.1	0.9;
	1714	1	-0.19	-1	0	0	0	1.005239	28.687581	150	8	1.1	0.9;
	1715	1	27.13	16.8	0	0.3	0	0.961566	-8.310096	150	9	1.1	0.9;
	1716	1	70.18	28.7	0	0.8	0	1.022288	26.268365	150	8	1.1	0.9;
	1717	1	13.9	1.9	0	0.38	0	1.006656	17.569983	154	1	1.1	0.9;
	1718	1	78.49	48.6	0	1.05	0	0.984449	-9.992776	150	9	1.1	0.9;
	1719	1	-15.82	5.2	0	1.76	0	1.01352	-5.029621	150	9	1.1	0.9;
	1720	1	-0.59	-2.55	0.04975	111.131798	0	1.085024	10.7034	380	18	1.2	0.8;
	1721	1	-0	0	0	35.18	0	1.150964	0.206027	380	1	1.2	0.8;
	1722	1	29.8	4.3	0	2.04	0	1.007874	18.273992	154	1	1.1	0.9;
	1723	1	105	412	0	37.57	0	1.017683	43.672345	380	3	1.1	0.9;
	1724	1	-0	0	0	0.34	0	1.03251	-0.902957	150	9	1.1	0.9;
	1725	1	0	-0	0	4.93	0	1.032164	20.701685	220	5	1.1	0.9;
	1726	2	0	0	0	0	0	1.058676	15.850616	20	5	1.1	0.9;
	1727	1	30.83	13.2	0	1.12	0	0.961076	-15.147703	110	22	1.1	0.9;
	1728	1	1.4	0.5	0	1.87	0	0.975077	-12.260891	110	12	1.1	0.9;
	1729	1	-0	0	0.169548	2.78813	0	1.074971	35.609269	380	14	1.1	0.9;
	1730	1	230.5	91	0	57.67	0	1.022999	25.475204	150	8	1.1	0.9;
	1731	2	0	0	0	0	0	1.028604	-12.690591	20	4	1.1	0.9;
	1732	2	0	0	0	0	0	1.03036	-6.512869	20	4	1.1	0.9;
	1733	1	26.4	3.8	0	0.7	0	1.000553	15.509241	154	1	1.1	0.9;
	1734	2	0	0	0	0	0	1.068377	26.459749	20	14	1.1	0.9;
	1735	2	0	0	0	0	0	1.068377	26.459749	20	14	1.1	0.9;
	1736	1	0	-0	0.134584	4.575416	0	1.042308	10.509848	220	7	1.1	0.9;
	1737	1	-0	-0	0	0.21	0	1.096519	-4.58332	150	9	1.1	0.9;
	1738	1	-0.18	-1	0	0	0	1.020198	36.372942	150	8	1.1	0.9;
	1739	1	53.16	17.7	0	0.48	0	0.978756	-8.89294	110	21	1.1	0.9;
	1740	1	7.41	2.7	0	1.39	0	0.975704	5.743947	110	19	1.1	0.9;
	1741	1	0	0	0	2.25	0	1.023788	-1.119868	154	1	1.1	0.9;
	1742	1	0	-0.01	0	0	0	1.00201	-7.836312	150	9	1.1	0.9;
	1743	1	-0	0	0	81.52	0	1.019118	-22.616319	380	4	1.1	0.9;
	1744	1	259.3	91.5	0.06021	6.83979	0	1.014828	22.549373	220	5	1.1	0.9;
	1745	1	71.78	17.9	0	1.96	0	1.039129	13.113694	220	5	1.1	0.9;
	1746	1	4.3	0.6	0	4.43	0	0.997864	19.705031	154	1	1.1	0.9;
	1747	1	-0	-0	0	14.29	0	1.02833	44.732893	380	3	1.1	0.9;
	1748	1	-0	-0	1.059066	6.756337	0	1.081511	24.720091	380	5	1.1	0.9;
	1749	2	0	0	0	0	0	1.009892	4.868208	20	9	1.1	0.9;
	1750	2	0	0	0	0	0	1.061222	57.751527	20	8	1.1	0.9;
	1751	2	0	0	0	0	0	1.061222	57.751527	20	8	1.1	0.9;
	1752	1	60.37	35.4	0	14.88	0	1.042878	84.226782	220	3	1.1	0.9;
	1753	1	103.6	16.01	0	11.22	0	1.055148	5.178293	220	5	1.1	0.9;
	1754	1	-0	0	0	0	0	0.986144	-28.039313	220	10	1.1	0.9;
	1755	1	23.53	4	0	9.39	0	1.016839	-13.978622	220	4	1.1	0.9;
	1756	1	-30.13	-22.3	0.040189	13.549811	0	1.004472	-10.478046	220	4	1.1	0.9;
	1757	2	0	0	0	0	0	1.024762	17.972121	20	14	1.1	0.9;
	1758	1	226.3	51.7	0	3.86	0	1.071437	24.004582	220	3	1.1	0.9;
	1759	1	-0	-0	0.056124	49.56323	0	1.080236	84.291341	380	3	1.2	0.8;
	1760	1	44.6	6.4	0	0.17	0	1.026126	7.522777	154	1	1.1	0.9;
	1761	1	-4.57	-1.31	0	1.12	0	1.011735	-3.055011	110	18	1.1	0.9;
	1762	1	2	1	0	0.61	0	0.988344	-4.51079	110	21	1.1	0.9;
	1763	2	0	0	0	0	0	1.022099	17.218586	20	14	1.1	0.9;
	1764	2	0	0	0	0	0	1.015505	-3.112218	20	22	1.1	0.9;
	1765	1	-0	0	0	1.17	0	1.047457	18.999824	220	5	1.1	0.9;
	1766	1	141.26	30.6	0	18.01	0	1.005802	18.488347	380	5	1.1	0.9;
	1767	1	39.44	24.9	0	4.18	0	0.983671	-7.624866	150	10	1.1	0.9;
	1768	1	-0	0	0.710067	0.007863	0	1.02702	21.390252	380	5	1.1	0.9;
	1769	1	-0	-0	0	6.56	0	1.002913	-8.664264	110	12	1.1	0.9;
	1770	2	0	0	0	0	0	1.040009	86.062395	20	3	1.1	0.9;
	1771	1	186.6	155.8	0	28.99	0	1.008369	47.486691	380	3	1.1	0.9;
	1772	2	0	0	0	0	0	1.040009	86.062395	20	3	1.1	0.9;
	1773	1	-0	0	0	0.4	0	0.982965	-6.219482	110	18	1.1	0.9;
	1774	2	0	0	0	0	0	1.043619	-8.436397	20	4	1.1	0.9;
	1775	1	20.62	8.2	0	6.62	0	1.024001	31.504337	150	8	1.1	0.9;
	1776	1	138.2	40.3	0	1.15	0	1.055355	-2.286835	110	15	1.1	0.9;
	1777	1	25.13	-61.6	0	3.68	0	0.992335	-2.275176	220	5	1.1	0.9;
	1778	1	-0	0	0	13.02	0	1.017294	-9.17725	220	6	1.1	0.9;
	1779	1	0	0	0.039102	6.540898	0	1.03133	64.946605	220	3	1.1	0.9;
	1780	1	-0	-0	0	39.46	0	1.126128	-1.959378	380	9	1.2	0.8;
	1781	1	50.86	36.6	0	0	0	1.01137	41.620925	150	8	1.1	0.9;
	1782	1	279.5	25.8	0	20.3	0	1.008994	-13.540968	220	6	1.1	0.9;
	1783	1	-0	0	0	10.73	0	1.03164	8.485177	220	5	1.1	0.9;
	1784	1	0	-0	0.256113	11.74089	0	1.005587	10.408982	220	19	1.1	0.9;
	1785	1	224.9	-32.5	0	63.28	0	1.014214	1.008491	380	6	1.1	0.9;
	1786	1	173	-21.3	0	5.93	0	1.022087	-22.528665	220	5	1.1	0.9;
	1787	1	15.42	2.2	0	1.46	0	0.977544	-15.597058	110	22	1.1	0.9;
	1788	1	-18.69	-7.9	0	2.09	0	1.026706	3.473674	220	7	1.1	0.9;
	1789	1	151.2	96.4	0.302056	38.856615	0	0.998882	24.025243	380	3	1.1	0.9;
	1790	1	112.8	27.6	0	35.18	0	0.997327	-7.348981	220	6	1.1	0.9;
	1791	2	0	0	0	0	0	1.016081	21.284924	20	5	1.1	0.9;
	1792	1	1.8	0.7	0	0.17	0	1.019347	-2.997978	110	19	1.1	0.9;
	1793	1	119.6	26.9	0.050934	6.129066	0	0.994612	-18.861992	220	4	1.1	0.9;
	1794	2	0	0	0	0	0	1.024544	18.974877	20	14	1.1	0.9;
	1795	1	-0	-0	0	0.23	0	1.020774	-8.3858	150	9	1.1	0.9;
	1796	2	0	0	0	0	0	1.008485	2.759897	20	10	1.1	0.9;
	1797	1	7.01	2.3	0	0.69	0	0.993529	-10.710423	110	12	1.1	0.9;
	1798	1	130.4	44.33	0	5.22	0	0.998219	11.690674	220	5	1.1	0.9;
	1799	1	149.9	53.2	0	1.07	0	0.95548	0.062075	110	18	1.1	0.9;
	1800	1	31.74	6.3	0	0.5	0	0.992174	-3.580764	110	18	1.1	0.9;
	1801	2	0	0	0	0	0	1.058149	41.053759	20	3	1.1	0.9;
	1802	1	0	0	0	0.48	0	0.990018	-6.93099	150	9	1.1	0.9;
	1803	2	0	0	0	0	0	1.018554	-1.801383	20	18	1.1	0.9;
	1804	1	-0.2	-1	0	0	0	1.018603	25.307117	150	8	1.1	0.9;
	1805	2	0	0	0	0	0	1.023695	-0.879651	20	18	1.1	0.9;
	1806	1	0	0	0	28.5	0	1.035794	12.685893	380	7	1.1	0.9;
	1807	2	0	0	0	0	0	1.023695	-0.879651	20	18	1.1	0.9;
	1808	2	0	0	0	0	0	1.018554	-1.801383	20	18	1.1	0.9;
	1809	1	-0	-0	0	1.56	0	0.987681	-22.129778	220	6	1.1	0.9;
	1810	1	0	-0	0	107.39	0	1.061635	-9.161951	380	4	1.1	0.9;
	1811	1	0	0	0	0	0	0.98112	0.528152	150	9	1.1	0.9;
	1812	2	0	0	0	0	0	1.031971	-3.542743	20	6	1.1	0.9;
	1813	2	0	0	0	0	0	1.012457	-4.968966	20	23	1.1	0.9;
	1814	1	202	34.1	0	1.94	0	1.031498	-19.60853	220	4	1.1	0.9;
	1815	2	0	0	0	0	0	1.007614	14.656778	20	5	1.1	0.9;
	1816	2	0	0	0	0	0	1.042445	10.433315	20	23	1.1	0.9;
	1817	1	42.75	15.2	0	0.54	0	1.046727	-2.648573	110	15	1.1	0.9;
	1818	1	19.02	9.1	0	0.52	0	0.991625	-3.871698	110	18	1.1	0.9;
	1819	1	0	-0.2	0	0.15	0	1.043584	8.515333	110	18	1.1	0.9;
	1820	1	71.58	48.2	0.288552	130.762047	0	1.013812	18.281622	380	3	1.1	0.9;
	1821	2	0	0	0	0	0	1.011416	44.498643	14.4	1	1.1	0.9;
	1822	2	0	0	0	0	0	1.035162	21.902936	13.8	1	1.1	0.9;
	1823	1	-0.19	-78.58	0	2.31	0	1.046766	42.710463	380	3	1.1	0.9;
	1824	1	9.61	5.9	0	1.11	0	1.01493	-10.961932	150	9	1.1	0.9;
	1825	1	0	0	0	0.02	0	1.025109	-10.517266	150	9	1.1	0.9;
	1826	1	18.82	6.2	0	0.65	0	0.936118	-10.724498	110	21	1.1	0.9;
	1827	1	22.32	7.3	0	0.72	0	0.982566	-11.168107	110	12	1.1	0.9;
	1828	1	124.1	17.6	0	6.03	0	0.993076	11.776433	154	1	1.1	0.9;
	1829	1	180.9	33.2	0	29.17	0	1.053882	4.374174	220	5	1.1	0.9;
	1830	1	23.4	3.3	0	6.88	0	1.00348	19.229454	154	1	1.1	0.9;
	1831	2	0	0	0	0	0	1.043059	9.285297	20	23	1.1	0.9;
	1832	1	26.63	9.7	0	0.37	0	0.989657	-1.540359	110	19	1.1	0.9;
	1833	1	6.11	-0	0	2.51	0	1.019509	-20.798895	220	4	1.1	0.9;
	1834	2	0	0	0	0	0	0.998251	7.060817	20	18	1.1	0.9;
	1835	1	181.1	78.7	0.040042	0.959958	0	0.995787	20.301147	220	5	1.1	0.9;
	1836	2	0	0	0	0	0	1.006718	-12.119029	20	4	1.1	0.9;
	1837	1	0	-0	0	0.3	0	1.019633	-8.384664	150	9	1.1	0.9;
	1838	1	19.02	8	0	2.05	0	1.001725	-3.617498	110	18	1.1	0.9;
	1839	2	0	0	0	0	0	1.019231	22.327173	20	24	1.1	0.9;
	1840	2	0	0	0	0	0	1.033947	-9.228621	20	6	1.1	0.9;
	1841	1	0	0	0	0	0	0.974052	-15.620125	150	9	1.1	0.9;
	1842	1	30.53	10	0	0.64	0	0.950524	-8.554178	110	21	1.1	0.9;
	1843	1	42.95	26.6	0	2.95	0	0.923452	-12.26721	150	9	1.1	0.9;
	1844	1	479	99.2	0	10.68	0	1.027844	0.686907	220	5	1.1	0.9;
	1845	2	0	0	0	0	0	0.962614	39.578299	15.75	1	1.1	0.9;
	1846	1	-19.83	-6.01	0	0.38	0	0.99762	-1.846002	110	18	1.1	0.9;
	1847	1	-0	-0	0	3.47	0	1.00818	22.120561	220	6	1.1	0.9;
	1848	1	75.28	19.2	0	1.9	0	1.028496	44.093107	110	8	1.1	0.9;
	1849	1	191.7	6.8	0	11.58	0	0.981635	-15.100371	220	10	1.1	0.9;
	1850	1	0	0	0	4.34	0	1.032021	6.627548	220	5	1.1	0.9;
	1851	2	0	0	0	0	0	1.058231	41.885187	20	10	1.1	0.9;
	1852	1	8.51	6.4	0	0.39	0	0.973955	-7.167886	110	18	1.1	0.9;
	1853	2	0	0	0	0	0	1.031737	11.756042	20	5	1.1	0.9;
	1854	1	479.3	127.6	0.039325	50.220675	0	1.047805	17.179536	380	14	1.1	0.9;
	1855	1	71.38	10.3	0	4.56	0	0.984485	-4.527908	220	6	1.1	0.9;
	1856	1	16.7	2.4	0	0.87	0	1.025593	-0.389646	154	1	1.1	0.9;
	1857	1	17.72	3.8	0	0.94	0	0.955374	-4.828749	110	18	1.1	0.9;
	1858	1	-0	-0	0	0.2	0	0.942852	-11.019109	150	9	1.1	0.9;
	1859	1	37.94	15.2	0	35.31	0	1.008235	8.191752	380	5	1.1	0.9;
	1860	1	18.2	2.6	0	1.51	0	1.013247	13.29657	154	1	1.1	0.9;
	1861	1	18.72	7.7	0.019611	1.710389	0	1.010307	-11.825094	110	22	1.1	0.9;
	1862	1	-0	0	0	2.79	0	1.013327	-1.436043	220	19	1.1	0.9;
	1863	1	13.82	3.6	0	0.72	0	0.981273	-7.052661	110	18	1.1	0.9;
	1864	1	32.64	11.8	0	2.07	0	1.021349	-2.688037	110	19	1.1	0.9;
	1865	1	0.2	0.8	0	0.05	0	0.996575	38.818652	220	3	1.1	0.9;
	1866	1	-0	0	0	0.08	0	1.054745	-6.214548	150	9	1.1	0.9;
	1867	1	0	-0	0.238928	0.940804	0	1.007945	-21.264125	220	6	1.1	0.9;
	1868	2	0	0	0	0	0	0.989649	6.296698	13.8	1	1.1	0.9;
	1869	2	0	0	0	0	0	1.017464	-4.981811	20	10	1.1	0.9;
	1870	1	11.41	5.1	0	1.04	0	1.017517	3.740543	110	18	1.1	0.9;
	1871	1	0	0	0	0	0	0.990878	-13.512721	150	9	1.1	0.9;
	1872	1	118.9	129.2	0	10.49	0	1.009026	29.741759	380	3	1.1	0.9;
	1873	1	37.14	11.6	0	1.8	0	1.024193	-3.16787	110	19	1.1	0.9;
	1874	1	37.69	-19.49	0	0.11	0	1.026773	-4.248659	110	19	1.1	0.9;
	1875	1	0	0	0	0.98	0	1.039974	-6.001274	220	5	1.1	0.9;
	1876	2	0	0	0	0	0	1.024324	-10.377063	20	4	1.1	0.9;
	1877	2	0	0	0	0	0	1.024324	-10.377063	20	4	1.1	0.9;
	1878	2	0	0	0	0	0	1.024324	-10.377063	20	4	1.1	0.9;
	1879	2	0	0	0	0	0	1.024324	-10.377063	20	4	1.1	0.9;
	1880	2	0	0	0	0	0	1.024324	-10.377063	20	4	1.1	0.9;
	1881	2	0	0	0	0	0	1.024324	-10.377063	20	4	1.1	0.9;
	1882	2	0	0	0	0	0	1.015006	-6.446837	20	4	1.1	0.9;
	1883	2	0	0	0	0	0	1.015006	-6.446837	20	4	1.1	0.9;
	1884	2	0	0	0	0	0	1.015006	-6.446837	20	4	1.1	0.9;
	1885	1	0	0	0	50.16	0	1.026524	-6.800114	380	6	1.1	0.9;
	1886	1	105.8	15.1	0	4.91	0	1.005046	28.526638	154	1	1.1	0.9;
	1887	1	100.6	15.3	0	1.28	0	0.96929	22.580591	150	8	1.1	0.9;
	1888	1	0	0	0	0	0	1.02223	-11.03513	150	9	1.1	0.9;
	1889	2	0	0	0	0	0	1.022889	18.879559	20	14	1.1	0.9;
	1890	1	0	0	0	4.93	0	0.99663	-4.736881	110	18	1.1	0.9;
	1891	1	-9.61	3.2	0	0.75	0	1.016347	-11.53137	150	9	1.1	0.9;
	1892	1	-0	0	1.160978	-0.444847	0	1.059249	20.524844	380	5	1.1	0.9;
	1893	1	-0	-0	0	0.59	0	1.042632	13.310034	220	5	1.1	0.9;
	1894	1	-0	-0	0.647865	2.624819	0	1.04959	20.547418	380	3	1.1	0.9;
	1895	2	0	0	0	0	0	1.029146	3.188044	20	20	1.1	0.9;
	1896	2	0	0	0	0	0	1.012991	3.357398	20	18	1.1	0.9;
	1897	1	13.9	1.9	0	1.02	0	1.00589	17.856342	154	1	1.1	0.9;
	1898	1	0	-0	0.232141	72.996222	0	1.060486	9.337743	380	16	1.1	0.9;
	1899	1	29.03	9.4	0	5.11	0	1.022122	38.733805	110	8	1.1	0.9;
	1900	1	108.4	-2.4	0	80.12	0	1.040198	-18.927563	380	10	1.1	0.9;
	1901	1	7.61	0	0	6.6	0	0.923978	-8.123126	110	18	1.1	0.9;
	1902	1	-0	-0	0.058904	3.471096	0	1.023748	14.863019	220	5	1.1	0.9;
	1903	2	0	0	0	0	0	1.031159	7.793526	11	1	1.1	0.9;
	1904	1	5.41	2.81	0	1.9	0	0.999453	-5.145639	150	9	1.1	0.9;
	1905	1	-0	-0	0	58.3	0	0.983412	14.056335	380	5	1.1	0.9;
	1906	2	0	0	0	0	0	0.999838	14.362863	20	18	1.1	0.9;
	1907	2	0	0	0	0	0	1.007227	24.551226	11	1	1.1	0.9;
	1908	1	-0	-0	0	2.13	0	1.033278	-18.937067	220	4	1.1	0.9;
	1909	1	1.91	3.95	0	0.02	0	0.984339	88.942359	150	1	1.1	0.9;
	1910	1	-0	-0	0	1.87	0	1.00975	7.1093	220	3	1.1	0.9;
	1911	1	0	-0	0	12.2	0	1.015708	-26.173758	220	4	1.1	0.9;
	1912	1	34.9	5	0	6.35	0	1.011379	19.937648	154	1	1.1	0.9;
	1913	2	0	0	0	0	0	1.042096	35.548796	20	14	1.1	0.9;
	1914	2	0	0	0	0	0	1.042096	35.548796	20	14	1.1	0.9;
	1915	1	30.53	13.8	0	2.19	0	1.024484	26.245019	150	8	1.1	0.9;
	1916	1	-0	0	0	0.46	0	0.948801	6.067726	110	18	1.1	0.9;
	1917	1	28.93	11.9	0	3.84	0	1.03672	25.549676	150	8	1.1	0.9;
	1918	1	6.41	2.3	0	0.85	0	0.996071	-1.070853	110	19	1.1	0.9;
	1919	1	291.9	91.9	0.040252	1.189748	0	1.021438	22.294536	220	5	1.1	0.9;
	1920	1	-3.29	-1.32	0	190.704588	0	1.022555	21.016577	380	5	1.1	0.9;
	1921	1	538.3	90.8	0	21.34	0	1.011947	-10.122615	380	6	1.1	0.9;
	1922	1	2.1	0.3	0	2.83	0	1.002616	29.478033	154	1	1.1	0.9;
	1923	1	100.3	4.8	0	0.15	0	0.968083	-11.943757	110	21	1.1	0.9;
	1924	1	60.17	30.6	0	0.12	0	1.017203	24.892446	220	3	1.1	0.9;
	1925	1	-0	0	0.27075	68.09435	0	1.077741	1.680409	380	18	1.2	0.8;
	1926	1	211.8	87.9	0.119912	0.300074	0	0.984054	-2.85841	110	18	1.1	0.9;
	1927	1	0	-0	0	6.93	0	1.069883	58.677401	220	8	1.1	0.9;
	1928	2	0	0	0	0	0	0.999593	27.016686	20	5	1.1	0.9;
	1929	1	0	-0	0.352132	27.727388	0	0.974224	7.613845	380	11	1.1	0.9;
	1930	1	40.85	14.5	0	0.59	0	1.039102	17.583714	220	5	1.1	0.9;
	1931	1	0	0	0.233608	-0.194673	0	1.031453	-10.473783	380	6	1.1	0.9;
	1932	2	0	0	0	0	0	1.016993	30.047809	20	5	1.1	0.9;
	1933	1	-0	-0	0	0.88	0	0.963828	-9.831601	150	9	1.1	0.9;
	1934	1	-0	0	0.527187	-0.130034	0	1.036112	19.191894	380	6	1.1	0.9;
	1935	2	0	0	0	0	0	1.048808	20.576441	20	5	1.1	0.9;
	1936	2	0	0	0	0	0	1.048808	20.576441	20	5	1.1	0.9;
	1937	1	-0	0	0.156456	-1.292742	0	1.028386	33.436937	380	2	1.1	0.9;
	1938	1	0	0	0	0	0	1.007899	-8.721694	150	9	1.1	0.9;
	1939	1	0	-0	0.245037	32.163304	0	1.024597	3.62755	380	7	1.1	0.9;
	1940	2	0	0	0	0	0	1.061603	44.385823	20	8	1.1	0.9;
	1941	1	46.55	28.8	0	1.08	0	1.029562	-6.831163	150	9	1.1	0.9;
	1942	1	100.9	14.4	0	0.06	0	1.004021	26.366061	154	1	1.1	0.9;
	1943	1	-0	0	0	0.29	0	0.998756	-10.351408	150	9	1.1	0.9;
	1944	1	0	0	0.077274	2.812726	0	1.045322	15.214941	220	5	1.1	0.9;
	1945	1	10.11	8.9	0	0.12	0	0.953415	-6.243675	110	18	1.1	0.9;
	1946	1	36.44	12	0	1.39	0	1.014073	1.963772	110	21	1.1	0.9;
	1947	1	0	0	0	0	0	1.007442	-12.200575	150	9	1.1	0.9;
	1948	1	18.02	6	0	0.41	0	0.962245	-6.704835	110	21	1.1	0.9;
	1949	2	0	0	0	0	0	1.00597	27.345397	154	1	1.1	0.9;
	1950	1	60.37	54.7	0	0.62	0	0.990306	33.144365	220	3	1.1	0.9;
	1951	1	2.91	-1.45	0	5.1	0	1.035714	-6.509155	150	9	1.1	0.9;
	1952	1	89.5	12.8	0	5.08	0	1.01008	26.93927	154	1	1.1	0.9;
	1953	1	-0	0	0	0	0	1.023321	-15.334384	20	4	1.1	0.9;
	1954	1	-0	0	0	0	0	1.023321	-15.334384	20	4	1.1	0.9;
	1955	2	0	0	0	0	0	1.026243	6.043442	20	6	1.1	0.9;
	1956	1	93.9	13.4	0	5.87	0	0.980421	27.054435	154	1	1.1	0.9;
	1957	2	0	0	0	0	0	1.019066	-10.738952	20	4	1.1	0.9;
	1958	1	10.81	3.5	0	0.98	0	0.991285	-10.566094	110	12	1.1	0.9;
	1959	2	0	0	0	0	0	1.023372	9.393826	11	1	1.1	0.9;
	1960	2	0	0	0	0	0	1.019	-2.975058	20	10	1.1	0.9;
	1961	2	0	0	0	0	0	0.996564	21.128006	11	1	1.1	0.9;
	1962	1	39.04	12	0.019755	3.180245	0	0.98563	89.067595	150	1	1.1	0.9;
	1963	2	0	0	0	0	0	1.015227	-3.436884	20	10	1.1	0.9;
	1964	1	0	0	0	14.34	0	1.08603	72.320383	380	3	1.2	0.8;
	1965	1	14.75	-0.07	0	3.07	0	0.980905	89.087518	150	1	1.1	0.9;
	1966	2	0	0	0	0	0	1.051329	90.483342	20	3	1.1	0.9;
	1967	2	0	0	0	0	0	1.051329	90.483342	20	3	1.1	0.9;
	1968	1	0	0	0	0	0	1.001857	-9.078486	150	9	1.1	0.9;
	1969	2	0	0	0	0	0	1.01993	9.094543	20	22	1.1	0.9;
	1970	1	20.32	6.7	0	1.11	0	0.990503	-9.05799	110	12	1.1	0.9;
	1971	1	0	0	0	0	0	1.006782	-15.116007	150	9	1.1	0.9;
	1972	2	0	0	0	0	0	1.029166	13.522836	20	14	1.1	0.9;
	1973	1	-0	-0	0.019477	11.600523	0	1.022205	-9.782228	220	4	1.1	0.9;
	1974	1	-0	-0	0	4.2	0	1.01575	49.478918	220	8	1.1	0.9;
	1975	1	2.4	0.7	0	10.69	0	1.022085	-18.016691	220	4	1.1	0.9;
	1976	1	30.93	7.6	0	4.69	0	1.024032	3.456517	220	5	1.1	0.9;
	1977	1	0	-0	0	3.36	0	1.046561	-20.234799	380	10	1.1	0.9;
	1978	1	52.96	15.2	0	0.56	0	1.012723	-6.930307	110	3	1.1	0.9;
	1979	1	0	-6	0	0	0	1.032253	-8.120405	150	9	1.1	0.9;
	1980	1	0	0	0	2.14	0	1.027181	16.66973	220	5	1.1	0.9;
	1981	1	196.7	44.2	0	0.1	0	1.002482	10.277586	220	5	1.1	0.9;
	1982	1	185.5	62.3	0.02989	30.86011	0	1.017507	-21.662554	380	4	1.1	0.9;
	1983	1	39.25	71.7	0	33.75	0	1.011147	29.27305	380	3	1.1	0.9;
	1984	1	11.91	2.9	0	5.27	0	1.033515	14.852151	220	5	1.1	0.9;
	1985	2	0	0	0	0	0	0.961956	47.51733	13.8	1	1.1	0.9;
	1986	1	0	-0	0.176224	0.082937	0	1.025692	6.014725	220	20	1.1	0.9;
	1987	1	-22.07	-1.97	0	13.85	0	1.060928	81.98539	220	3	1.1	0.9;
	1988	2	0	0	0	0	0	1.027807	-10.041134	20	4	1.1	0.9;
	1989	1	69.3	9.9	0	0.03	0	1.010825	30.668863	154	1	1.1	0.9;
	1990	1	-1401.62	-585.3	0	0	0	1.056256	29.122283	27	3	1.1	0.9;
	1991	1	-0	0	0	12.98	0	1.021956	-10.101004	150	9	1.1	0.9;
	1992	1	157.9	26	0	3.74	0	1.011131	4.479816	220	7	1.1	0.9;
	1993	1	0	-0	0	4.15	0	0.976411	-22.233533	150	10	1.1	0.9;
	1994	1	15.52	5.2	0	0.29	0	0.966913	-9.382322	110	21	1.1	0.9;
	1995	2	0	0	0	0	0	0.989323	97.08776	20	3	1.1	0.9;
	1996	1	138.2	153.5	0.473187	29.995678	0	1.030347	-20.964804	220	4	1.1	0.9;
	1997	2	0	0	0	0	0	1.044265	38.235788	20	5	1.1	0.9;
	1998	1	62.97	57.7	0.459488	11.506239	0	1.04895	37.800634	220	13	1.1	0.9;
	1999	2	0	0	0	0	0	1.058091	39.373985	20	14	1.1	0.9;
	2000	2	0	0	0	0	0	1.058091	39.373985	20	14	1.1	0.9;
	2001	2	0	0	0	0	0	1.038561	17.439863	20	14	1.1	0.9;
	2002	1	5.11	0.9	0	0.37	0	1.001121	-3.440015	110	18	1.1	0.9;
	2003	2	0	0	0	0	0	1.0385	19.998666	20	14	1.1	0.9;
	2004	2	0	0	0	0	0	1.0385	19.998666	20	14	1.1	0.9;
	2005	2	0	0	0	0	0	1.0385	19.998666	20	14	1.1	0.9;
	2006	2	0	0	0	0	0	1.053599	21.197213	20	14	1.1	0.9;
	2007	1	144.7	20.3	0	6.35	0	0.998972	2.227611	220	5	1.1	0.9;
	2008	1	-0	0	0	0.3	0	1.032008	18.368955	220	5	1.1	0.9;
	2009	1	7.81	2.6	0	0.3	0	0.995927	-1.680715	110	19	1.1	0.9;
	2010	1	4.81	1.5	0.039642	1.450358	0	0.986291	10.017412	110	18	1.1	0.9;
	2011	2	0	0	0	0	0	1.058091	39.373985	20	14	1.1	0.9;
	2012	2	0	0	0	0	0	1.040793	26.940223	20	5	1.1	0.9;
	2013	2	0	0	0	0	0	1.040793	26.940223	20	5	1.1	0.9;
	2014	2	0	0	0	0	0	1.040793	26.940223	20	5	1.1	0.9;
	2015	1	42.95	19.8	0	3.85	0	1.014279	-24.903336	220	4	1.1	0.9;
	2016	1	132.2	32.8	0.020137	12.049863	0	1.023472	8.423098	220	5	1.1	0.9;
	2017	1	0	-0	0	8.53	0	1.005826	-2.715534	220	6	1.1	0.9;
	2018	1	-4.89	-2.35	0	0.13	0	1.008644	-0.910004	110	18	1.1	0.9;
	2019	1	106.1	-6.8	0	22.23	0	1.015484	31.289514	380	3	1.1	0.9;
	2020	2	0	0	0	0	0	1.040793	26.940223	20	5	1.1	0.9;
	2021	1	-0	-0	0.284748	10.36271	0	1.069232	53.740618	380	3	1.1	0.9;
	2022	1	0	0	0	1.85	0	0.976329	-6.685752	110	21	1.1	0.9;
	2023	1	22.83	7.5	0	0.36	0	0.978493	-11.442138	110	12	1.1	0.9;
	2024	1	-0	-0	0	121.27	0	1.03921	-8.713669	380	22	1.1	0.9;
	2025	1	20.82	7.9	0	5.62	0	1.011495	-23.598553	220	4	1.1	0.9;
	2026	1	-8.31	2.7	0	0.18	0	0.981238	-8.016126	150	9	1.1	0.9;
	2027	1	193.2	121.7	0	11.94	0	1.01658	19.6589	220	3	1.1	0.9;
	2028	1	39.54	14.3	0	0.23	0	1.020551	-1.791743	110	19	1.1	0.9;
	2029	1	45.95	7.5	0	8.39	0	1.017947	-17.668188	220	4	1.1	0.9;
	2030	1	178	33.8	0	2.73	0	1.015266	3.283199	220	7	1.1	0.9;
	2031	1	0	-0	0	4.88	0	1.045752	2.640675	220	1	1.1	0.9;
	2032	1	-0	-0	0	3.85	0	1.021846	5.833793	220	5	1.1	0.9;
	2033	1	-0	0	0.039816	0.010184	0	1.033157	16.61434	220	5	1.1	0.9;
	2034	1	204.6	53.2	0	7.4	0	0.988998	-10.568899	220	4	1.1	0.9;
	2035	1	0	-0	0.31	25.37	0	1.030155	-0.095505	220	19	1.1	0.9;
	2036	1	110.1	2	0	13.85	0	1.030761	-20.945532	220	4	1.1	0.9;
	2037	1	-0	-0	0	2.25	0	0.991062	20.818561	154	1	1.1	0.9;
	2038	1	53.86	-15.3	0	2.87	0	1.030023	-18.736805	220	4	1.1	0.9;
	2039	1	89.3	12.8	0	4.33	0	1.007012	18.351122	154	1	1.1	0.9;
	2040	1	171.96	40.73	0	23.47	0	1.027588	18.766692	380	13	1.1	0.9;
	2041	1	-0	-0	0	4.56	0	1.016874	18.929174	220	5	1.1	0.9;
	2042	2	0	0	0	0	0	1.027071	15.95651	154	1	1.1	0.9;
	2043	1	-0	-0	0.287012	9.86039	0	1.051863	6.942059	220	11	1.1	0.9;
	2044	1	238.4	210.9	0	0	0	1.053612	29.620375	220	2	1.1	0.9;
	2045	1	0	0	0	-0.73	0	0.996044	89.673422	150	1	1.1	0.9;
	2046	1	10.91	3	0	0.88	0	0.930348	-9.655565	110	18	1.1	0.9;
	2047	2	0	0	0	0	0	1.05948	38.842339	20	8	1.1	0.9;
	2048	2	0	0	0	0	0	1.047307	6.762881	11	1	1.1	0.9;
	2049	1	-0	-0	0	27.56	0	1.056048	54.150215	220	3	1.1	0.9;
	2050	1	-0	-0	0	9.49	0	1.058485	15.237211	220	5	1.1	0.9;
	2051	1	0	-0	0	6.65	0	1.034192	-10.845742	150	9	1.1	0.9;
	2052	1	302.1	216	0.944202	84.972507	0	1.033742	-16.886752	380	4	1.1	0.9;
	2053	2	0	0	0	0	0	1.039038	39.160421	20	2	1.1	0.9;
	2054	1	20.52	6.7	0	0.13	0	1.014955	-9.020008	110	12	1.1	0.9;
	2055	1	0	0	0	2.3	0	1.009821	31.22273	380	3	1.1	0.9;
	2056	1	0	0	0	0	0	1.03472	7.897462	20	7	1.1	0.9;
	2057	1	80.09	26.3	0	0.37	0	0.987847	-10.788419	110	12	1.1	0.9;
	2058	1	57.16	19.7	0	22.17	0	1.028259	18.983679	220	5	1.1	0.9;
	2059	1	199.2	22.8	0	27.54	0	1.018557	-14.385235	220	4	1.1	0.9;
	2060	1	18.02	4.8	0	0.43	0	0.956784	-8.620578	110	18	1.1	0.9;
	2061	1	-0	-0	0	0.37	0	0.977222	-12.558727	110	1	1.1	0.9;
	2062	1	20.82	36	0	0.97	0	1.011552	-11.845244	220	4	1.1	0.9;
	2063	2	0	0	0	0	0	1.013208	-8.664527	20	10	1.1	0.9;
	2064	1	4	172.7	1.18784	144.149627	0	1.024743	-10.705998	380	4	1.1	0.9;
	2065	1	-0	-0	0	3.38	0	1.032593	8.105588	220	5	1.1	0.9;
	2066	2	0	0	0	0	0	1.032856	3.083067	20	18	1.1	0.9;
	2067	2	0	0	0	0	0	1.099592	65.127951	380	8	1.2	0.8;
	2068	1	0	0	0	1.44	0	1.046448	-13.015813	220	4	1.1	0.9;
	2069	2	0	0	0	0	0	1.019153	9.002657	11	1	1.1	0.9;
	2070	1	35.24	11.4	0	1.84	0	0.987889	-10.352859	110	12	1.1	0.9;
	2071	2	0	0	0	0	0	0.999187	25.525754	14.4	1	1.1	0.9;
	2072	1	109.6	36.4	0	13.24	0	1.013096	32.553236	150	8	1.1	0.9;
	2073	1	162	83.6	0	5.32	0	1.001714	3.036512	220	7	1.1	0.9;
	2074	1	22.73	6.4	0	2.1	0	1.006549	30.289156	150	8	1.1	0.9;
	2075	1	-11.03	-5.11	0	29.703251	0	0.982274	15.279996	380	5	1.1	0.9;
	2076	1	42.95	26.6	0	3.11	0	1.091249	-4.873838	150	9	1.1	0.9;
	2077	1	0	0	0	0	0	1.011712	-8.997116	20	6	1.1	0.9;
	2078	1	0	0	0	0	0	1.011712	-8.997116	20	6	1.1	0.9;
	2079	1	50.06	16	0	0.65	0	0.970356	-6.496627	110	21	1.1	0.9;
	2080	2	0	0	0	0	0	1.043665	14.930439	20	14	1.1	0.9;
	2081	1	0	0	0	0	0	1.02378	-11.347601	150	9	1.1	0.9;
	2082	2	0	0	0	0	0	1.034376	-4.403708	20	6	1.1	0.9;
	2083	1	-0	-0	0	17.6	0	0.990663	15.887389	220	6	1.1	0.9;
	2084	2	0	0	0	0	0	1.033042	-4.165701	20	7	1.1	0.9;
	2085	2	0	0	0	0	0	1.03432	-4.425322	20	6	1.1	0.9;
	2086	2	0	0	0	0	0	1.03432	-4.425322	20	6	1.1	0.9;
	2087	1	0	-0	0	0	0	1.01167	-8.995127	20	6	1.1	0.9;
	2088	1	0	-0	0	0	0	1.01167	-8.995127	20	6	1.1	0.9;
	2089	1	-0	-0	0	6.4	0	1.038677	-3.733513	220	6	1.1	0.9;
	2090	2	0	0	0	0	0	1.030138	25.689742	20	5	1.1	0.9;
	2091	2	0	0	0	0	0	1.030138	25.689742	20	5	1.1	0.9;
	2092	1	18.12	-11.6	0	14.14	0	0.989306	-22.035267	220	6	1.1	0.9;
	2093	2	0	0	0	0	0	1.030138	25.689742	20	5	1.1	0.9;
	2094	1	0	0	0	0.98	0	0.982509	-2.834787	110	18	1.1	0.9;
	2095	1	-0	-0	0.15	6.49	0	0.978202	-5.335353	220	21	1.1	0.9;
	2096	1	39.2	5.5	0	7.25	0	0.987179	19.019867	154	1	1.1	0.9;
	2097	2	0	0	0	0	0	1.024484	-6.060208	20	6	1.1	0.9;
	2098	1	117.13	33.86	0	6.04	0	1.01394	6.023647	220	7	1.1	0.9;
	2099	1	-0	0	0	0.77	0	0.966304	-5.236521	110	19	1.1	0.9;
	2100	1	30.63	19	0	2.33	0	0.868495	-15.380873	150	9	1.2	0.8;
	2101	1	-0	0	0	5.51	0	1.023051	34.780864	150	8	1.1	0.9;
	2102	2	0	0	0	0	0	1.012679	-1.92622	20	18	1.1	0.9;
	2103	1	30.03	10	0	0.08	0	0.976063	-6.122796	110	21	1.1	0.9;
	2104	1	-0	0	0.038912	2.611088	0	1.035226	13.284427	220	5	1.1	0.9;
	2105	1	82.19	23.4	0	17.91	0	1.021039	-15.662962	380	4	1.1	0.9;
	2106	2	0	0	0	0	0	0.962508	28.579218	15	1	1.1	0.9;
	2107	2	0	0	0	0	0	1.022583	-10.419342	20	4	1.1	0.9;
	2108	1	100.1	45.6	0	0.23	0	0.990372	-19.213624	220	6	1.1	0.9;
	2109	1	627.8	293.4	0.698702	7.516861	0	1.033996	28.695012	220	13	1.1	0.9;
	2110	1	0	0	0	0	0	1.027323	-4.289105	150	9	1.1	0.9;
	2111	1	-28.21	-6	0	0.05	0	1.027234	16.676769	220	5	1.1	0.9;
	2112	1	0	-0	0	19.36	0	1.036635	-21.205229	220	6	1.1	0.9;
	2113	1	-0	0	0.175191	0.074008	0	1.036582	80.128108	220	3	1.1	0.9;
	2114	1	144.9	87.7	0.213866	36.224477	0	0.998509	12.552323	380	3	1.1	0.9;
	2115	1	79.59	18.8	0	6.78	0	1.02775	7.156287	220	5	1.1	0.9;
	2116	1	-29.03	-4	0.191321	4.088679	0	1.040298	14.848896	220	7	1.1	0.9;
	2117	1	64	9.1	0	9.94	0	0.997409	27.546763	154	1	1.1	0.9;
	2118	1	0	0	0	2.54	0	1.019995	-8.583845	220	10	1.1	0.9;
	2119	2	0	0	0	0	0	1.048702	40.981802	20	14	1.1	0.9;
	2120	1	117.5	-3	0	5.38	0	1.025293	-19.876992	220	4	1.1	0.9;
	2121	1	193.8	27.3	0.462162	21.334122	0	1.029357	20.473844	220	13	1.1	0.9;
	2122	2	0	0	0	0	0	0.993105	34.379244	10.6	1	1.1	0.9;
	2123	1	1.1	2.5	0.263627	23.900622	0	0.994655	21.997919	380	3	1.1	0.9;
	2124	1	-1.85	-1.63	0	31.424588	0	1.030794	17.671729	380	5	1.1	0.9;
	2125	1	76.18	13.2	0	7.4	0	1.016921	-12.573705	220	4	1.1	0.9;
	2126	1	-0	-0	0	0.94	0	1.019395	8.024298	110	18	1.1	0.9;
	2127	1	0.7	0.2	0	0.22	0	1.004219	-8.787097	110	12	1.1	0.9;
	2128	1	0	-0	0	11.44	0	0.978447	49.743461	220	3	1.1	0.9;
	2129	1	-0	-0	0	103.07	0	0.96807	40.282968	400	1	1.1	0.9;
	2130	1	-0	-0	0	56.53	0	1.055919	-12.348545	380	4	1.1	0.9;
	2131	2	0	0	0	0	0	0.988831	38.418014	10.5	1	1.1	0.9;
	2132	1	0	0	0.119399	8.830601	0	0.990911	-11.398259	220	22	1.1	0.9;
	2133	1	13.82	5	0	0.47	0	0.987324	-2.421606	110	19	1.1	0.9;
	2134	1	-0	0	0	12.57	0	1.015672	-26.779891	220	4	1.1	0.9;
	2135	2	0	0	0	0	0	1.035502	32.502703	20	5	1.1	0.9;
	2136	2	0	0	0	0	0	1.035502	32.502796	20	5	1.1	0.9;
	2137	2	0	0	0	0	0	1.035502	32.502796	20	5	1.1	0.9;
	2138	2	0	0	0	0	0	1.035502	32.502703	20	5	1.1	0.9;
	2139	2	0	0	0	0	0	1.003313	37.135409	11	1	1.1	0.9;
	2140	1	142.8	-6.4	0	12.36	0	1.008457	-16.144565	220	4	1.1	0.9;
	2141	1	-0	-0	0	1.84	0	1.019922	17.052804	220	5	1.1	0.9;
	2142	1	6.71	3.3	0	1.31	0	0.99793	-4.757094	110	18	1.1	0.9;
	2143	2	0	0	0	0	0	1.017546	-7.215576	20	4	1.1	0.9;
	2144	2	0	0	0	0	0	1.017546	-7.215576	20	4	1.1	0.9;
	2145	1	-0	-0	0	149.14	0	1.067958	2.705618	380	19	1.1	0.9;
	2146	2	0	0	0	0	0	1.017546	-7.215576	20	4	1.1	0.9;
	2147	2	0	0	0	0	0	1.019095	-7.960677	20	4	1.1	0.9;
	2148	2	0	0	0	0	0	1.046243	23.151951	20	5	1.1	0.9;
	2149	2	0	0	0	0	0	1.046243	23.151951	20	5	1.1	0.9;
	2150	1	0.3	-34.1	0	0.25	0	1.030941	29.790909	220	3	1.1	0.9;
	2151	1	-0	-0	0	1.12	0	1.026477	-3.548893	150	9	1.1	0.9;
	2152	1	196.6	28.1	0	0.52	0	0.990583	0.504532	154	1	1.1	0.9;
	2153	2	0	0	0	0	0	1.01017	15.666854	20	5	1.1	0.9;
	2154	2	0	0	0	0	0	1.04545	17.664921	20	5	1.1	0.9;
	2155	2	0	0	0	0	0	1.030117	31.43919	20	5	1.1	0.9;
	2156	2	0	0	0	0	0	1.049836	11.857105	20	5	1.1	0.9;
	2157	2	0	0	0	0	0	1.030852	14.833561	20	5	1.1	0.9;
	2158	2	0	0	0	0	0	1.030117	31.43919	20	5	1.1	0.9;
	2159	1	339.1	89	0	80.71	0	1.011607	-13.743323	380	6	1.1	0.9;
	2160	1	0	-0	0	4.31	0	1.035657	77.742691	220	3	1.1	0.9;
	2161	1	-0	0	0	1.18	0	1.007469	-1.437468	150	9	1.1	0.9;
	2162	1	-0	-0	0	28.31	0	0.965839	24.000238	400	1	1.1	0.9;
	2163	1	0	-0	0	0	0	1.019372	-15.361499	20	4	1.1	0.9;
	2164	1	0	-0	0	0	0	1.019372	-15.361499	20	4	1.1	0.9;
	2165	1	22.02	9	0	0.24	0	1.041679	26.199412	150	8	1.1	0.9;
	2166	2	0	0	0	0	0	1.02436	-12.195994	20	4	1.1	0.9;
	2167	2	0	0	0	0	0	1.022192	20.047374	20	5	1.1	0.9;
	2168	1	7.41	2.7	0	0.98	0	1.003941	-4.688236	110	19	1.1	0.9;
	2169	1	-0	-0	0	0.1	0	1.001706	-10.006824	150	9	1.1	0.9;
	2170	1	-10.51	3.5	0	0.8	0	1.01079	-5.231958	150	9	1.1	0.9;
	2171	1	173.6	72.9	0	2.5	0	1.006201	-19.200996	220	4	1.1	0.9;
	2172	2	0	0	0	0	0	1.04914	42.65185	20	2	1.1	0.9;
	2173	1	-0	-0	0	0	0	0.993439	-9.302864	150	9	1.1	0.9;
	2174	1	72.1	71.8	0	22	0	1.014044	47.531028	380	3	1.1	0.9;
	2175	1	135	19.3	0	3.31	0	1.00349	27.619503	154	1	1.1	0.9;
	2176	2	0	0	0	0	0	1.054354	23.084216	20	3	1.1	0.9;
	2177	2	0	0	0	0	0	1.054354	23.084216	20	3	1.1	0.9;
	2178	2	0	0	0	0	0	1.054354	23.084216	20	3	1.1	0.9;
	2179	2	0	0	0	0	0	1.054354	23.084216	20	3	1.1	0.9;
	2180	2	0	0	0	0	0	0.985554	42.387961	14.4	1	1.1	0.9;
	2181	1	115.2	38.9	0	2.96	0	1.009209	5.096099	220	7	1.1	0.9;
	2182	1	0	-0	0	17.24	0	1.051605	44.804322	380	3	1.1	0.9;
	2183	2	0	0	0	0	0	1.015185	-4.986575	20	4	1.1	0.9;
	2184	1	-10.86	-5.93	0	12.54	0	1.035121	19.25997	220	5	1.1	0.9;
	2185	1	0	0	0	5.38	0	0.91885	-12.405973	150	9	1.1	0.9;
	2186	2	0	0	0	0	0	1.05594	38.382088	20	14	1.1	0.9;
	2187	1	5.81	2	0	0.2	0	0.97676	-5.773301	110	21	1.1	0.9;
	2188	1	35.3	5	0	0.89	0	1.004763	17.696683	154	1	1.1	0.9;
	2189	1	0	8.4	0	15.21	0	1.033472	-6.583857	150	9	1.1	0.9;
	2190	2	0	0	0	0	0	1.033542	1.716976	20	6	1.1	0.9;
	2191	1	-0	-0	0	0	0	1.004923	27.003315	150	2	1.1	0.9;
	2192	2	0	0	0	0	0	1.015919	3.580325	6.3	1	1.1	0.9;
	2193	1	2.5	1.3	0	0.85	0	0.994926	3.247305	110	18	1.1	0.9;
	2194	1	43.8	20.7	0	133.678646	0	1.07972	-3.354891	380	9	1.1	0.9;
	2195	1	0	0	0.261953	32.381259	0	1.071315	-1.316775	380	21	1.1	0.9;
	2196	1	129.5	51.8	0	107.08	0	1.025646	6.556766	380	6	1.1	0.9;
	2197	2	0	0	0	0	0	1.024678	4.335626	20	20	1.1	0.9;
	2198	2	0	0	0	0	0	1.033735	-14.68008	20	4	1.1	0.9;
	2199	1	-0	-0	0	0.63	0	1.003276	4.030994	110	18	1.1	0.9;
	2200	1	0	0	0	0	0	0.971722	-11.45905	150	9	1.1	0.9;
	2201	1	0	0	1.025363	33.6215	0	1.048558	30.218377	380	3	1.1	0.9;
	2202	1	0	0	0	0	0	0.979797	29.825191	400	1	1.1	0.9;
	2203	1	73.58	3.24	0	52.35	0	1.058269	14.205436	220	5	1.1	0.9;
	2204	1	0	0	0.27075	74.70435	0	1.08089	6.292575	380	18	1.2	0.8;
	2205	2	0	0	0	0	0	1.010845	-15.053498	20	4	1.1	0.9;
	2206	1	125.4	45.8	0	0	0	1.05891	5.576137	220	5	1.1	0.9;
	2207	1	8	1.1	0	3.65	0	0.997627	15.442845	154	1	1.1	0.9;
	2208	1	116.7	64	0	10.41	0	0.984148	-22.31072	220	10	1.1	0.9;
	2209	1	0	-0	0	44.56	0	1.122892	-2.324184	380	23	1.2	0.8;
	2210	2	0	0	0	0	0	1.024762	17.972121	20	14	1.1	0.9;
	2211	1	0	-0	0.233775	-0.194812	0	1.029792	-11.449739	380	6	1.1	0.9;
	2212	1	25.13	9.4	0.039154	0.480846	0	1.030503	26.412296	220	5	1.1	0.9;
	2213	1	7.61	2.5	0	0.72	0	0.962326	5.494068	110	18	1.1	0.9;
	2214	1	300.2	-3.1	0	29.94	0	1.005604	-25.891381	380	4	1.1	0.9;
	2215	2	0	0	0	0	0	1.003014	30.196256	11.5	1	1.1	0.9;
	2216	1	-0	0	0.220588	73.145368	0	1.085672	8.755816	380	18	1.2	0.8;
	2217	1	0	-0	0	7.99	0	1.037968	10.118901	220	5	1.1	0.9;
	2218	1	20.42	8.1	0	35.15	0	1.031688	13.151537	380	5	1.1	0.9;
	2219	1	-21.62	-3.9	0	4.01	0	1.030191	9.268782	220	7	1.1	0.9;
	2220	1	0	0	0	4.96	0	1.018429	14.959406	220	5	1.1	0.9;
	2221	1	0	-0	0.471052	0.677084	0	1.058715	18.400691	380	14	1.1	0.9;
	2222	2	0	0	0	0	0	1.019405	-14.419189	20	10	1.1	0.9;
	2223	2	0	0	0	0	0	0.996935	21.950327	10.6	1	1.1	0.9;
	2224	1	-0.58	-53.03	0	6.85	0	1.035309	-11.752272	220	5	1.1	0.9;
	2225	1	-0	-0	0.846571	39.359754	0	1.031496	10.828188	380	7	1.1	0.9;
	2226	1	20.9	3	0	4.18	0	1.02273	18.752443	154	1	1.1	0.9;
	2227	1	-0	0	0	2.62	0	1.010154	33.063453	154	1	1.1	0.9;
	2228	1	-0.5	-1	0	0.68	0	1.013501	-1.43555	220	19	1.1	0.9;
	2229	1	175.2	82.4	0	0.21	0	1.009721	-8.747532	220	6	1.1	0.9;
	2230	1	77.38	24.1	0	2.67	0	1.029803	4.001005	220	5	1.1	0.9;
	2231	1	56.56	3.5	0	0.63	0	0.99691	-4.56571	110	18	1.1	0.9;
	2232	1	140.4	49.4	0	4.5	0	1.005956	32.719778	150	8	1.1	0.9;
	2233	1	16.7	2.4	0	3.57	0	1.006552	27.185653	154	1	1.1	0.9;
	2234	1	156.9	-4	0	2.48	0	1.02641	5.55336	220	5	1.1	0.9;
	2235	1	60.07	27	0	0.48	0	1.027955	-2.520583	110	23	1.1	0.9;
	2236	2	0	0	0	0	0	1.062896	-0.392816	20	5	1.1	0.9;
	2237	2	0	0	0	0	0	0.991989	39.317054	13.8	1	1.1	0.9;
	2238	1	69.08	21.5	0	1.65	0	1.033687	4.712833	220	5	1.1	0.9;
	2239	1	0	0	0	0	0	1.034072	-10.492546	150	9	1.1	0.9;
	2240	1	0	0	0.113849	6.775126	0	1.065871	-7.624412	380	10	1.1	0.9;
	2241	1	329.4	12.9	0	34.36	0	1.030878	-0.616064	380	6	1.1	0.9;
	2242	2	0	0	0	0	0	1.024544	18.974877	20	14	1.1	0.9;
	2243	1	12.6	1.8	0	1.96	0	0.976528	39.092265	154	1	1.1	0.9;
	2244	1	34.04	4.4	0	0.22	0	0.99004	-14.379986	110	22	1.1	0.9;
	2245	1	-0	0	0	2.18	0	0.97777	-8.392336	220	22	1.1	0.9;
	2246	2	0	0	0	0	0	1.005775	-8.537815	20	22	1.1	0.9;
	2247	2	0	0	0	0	0	1.019723	-1.552655	20	18	1.1	0.9;
	2248	2	0	0	0	0	0	1.019723	-1.552655	20	18	1.1	0.9;
	2249	2	0	0	0	0	0	1.019723	-1.552655	20	18	1.1	0.9;
	2250	1	32.04	19.8	0	0.72	0	1.058721	-0.102635	150	9	1.1	0.9;
	2251	1	-0	0	0	1.38	0	1.029067	19.246507	220	5	1.1	0.9;
	2252	2	0	0	0	0	0	1.019723	-1.552655	20	18	1.1	0.9;
	2253	2	0	0	0	0	0	1.052369	0.152148	20	18	1.1	0.9;
	2254	1	-0	0	0	16.56	0	0.993075	42.619453	380	2	1.1	0.9;
	2255	1	0	-0	0	0.19	0	1.016806	-0.631712	150	9	1.1	0.9;
	2256	1	146.6	3.3	0	18.83	0	1.047504	13.147125	220	3	1.1	0.9;
	2257	1	-0	-0	0.228851	8.919291	0	1.046454	93.889744	380	3	1.1	0.9;
	2258	1	-4.14	-1.83	0	0.46	0	1.0026	-1.481173	110	18	1.1	0.9;
	2259	1	-0	-0	0	3.24	0	1.003601	15.328145	220	5	1.1	0.9;
	2260	2	0	0	0	0	0	1.001531	8.966652	15.75	1	1.1	0.9;
	2261	1	30.53	10.4	0	0.56	0	1.03277	12.252895	110	19	1.1	0.9;
	2262	1	0	-0	0	0	0	1.023482	0.234518	110	3	1.1	0.9;
	2263	1	0	-0	0	21.01	0	1.006414	-4.037915	380	7	1.1	0.9;
	2264	1	78.5	137.8	0	3.01	0	1.028156	82.753284	220	3	1.1	0.9;
	2265	1	104.5	-56.4	0	23.22	0	1.033153	19.747498	380	3	1.1	0.9;
	2266	2	0	0	0	0	0	1.006784	17.600399	154	1	1.1	0.9;
	2267	1	0.9	0.3	0	1.48	0	1.043069	-2.696955	110	19	1.1	0.9;
	2268	2	0	0	0	0	0	1.037301	24.456545	11	1	1.1	0.9;
	2269	1	-0.16	-11.88	0	16.12	0	1.012368	45.757755	380	8	1.1	0.9;
	2270	2	0	0	0	0	0	0.99902	31.838047	10.5	1	1.1	0.9;
	2271	1	28.33	44.8	0	0.51	0	1.051812	30.353844	220	5	1.1	0.9;
	2272	1	-0	-0	0	3.01	0	1.067986	87.275457	380	3	1.1	0.9;
	2273	1	-0	-0	0	288	0	0.977674	13.494685	400	1	1.1	0.9;
	2274	1	71.77	13.57	0	3.84	0	0.991864	0.773119	220	5	1.1	0.9;
	2275	1	76.29	45.7	0	10.89	0	1.046417	67.606979	380	3	1.1	0.9;
	2276	1	51.96	13.4	0	0.65	0	1.02565	45.340429	110	8	1.1	0.9;
	2277	1	58.87	36.4	0	3.38	0	0.953175	-4.97832	150	9	1.1	0.9;
	2278	2	0	0	0	0	0	1.018046	26.61034	6.3	1	1.1	0.9;
	2279	1	-0	0	0	0.48	0	1.010213	-20.852	220	4	1.1	0.9;
	2280	1	38.74	14.1	0	12.78	0	1.014992	-26.755976	220	4	1.1	0.9;
	2281	1	-0	-0	0	4.63	0	1.008045	29.243998	380	3	1.1	0.9;
	2282	2	0	0	0	0	0	1.022192	20.047374	20	5	1.1	0.9;
	2283	2	0	0	0	0	0	1.022192	20.047374	20	5	1.1	0.9;
	2284	2	0	0	0	0	0	1.022192	20.047374	20	5	1.1	0.9;
	2285	2	0	0	0	0	0	1.022192	20.047374	20	5	1.1	0.9;
	2286	2	0	0	0	0	0	1.000084	9.446263	20	23	1.1	0.9;
	2287	2	0	0	0	0	0	1.000084	9.446263	20	23	1.1	0.9;
	2288	1	45.75	4.2	0	4.9	0	1.019235	-32.702004	220	4	1.1	0.9;
	2289	1	-0	-0	0	0	0	1.000465	12.056714	20	5	1.1	0.9;
	2290	1	137.2	16.1	0	5.26	0	1.011343	3.613636	220	7	1.1	0.9;
	2291	1	-0	0	0.04006	0.11994	0	1.028027	23.037833	220	5	1.1	0.9;
	2292	2	0	0	0	0	0	1.024089	38.406443	20	10	1.1	0.9;
	2293	1	-0	0	0.224579	1.042208	0	1.074566	35.543789	380	14	1.1	0.9;
	2294	1	73.78	-0	0	2.69	0	1.035703	23.827158	380	5	1.1	0.9;
	2295	1	25.13	15.6	0	4.45	0	1.047679	-11.39683	150	9	1.1	0.9;
	2296	1	0	0	0	6.55	0	1.039935	9.269401	220	5	1.1	0.9;
	2297	1	164.8	42.5	0	32.849681	0	1.014053	-16.759949	380	6	1.1	0.9;
	2298	1	274.1	166.5	0	0.56	0	1.049907	29.51908	150	2	1.1	0.9;
	2299	1	18.5	2.6	0	3.05	0	1.002675	33.754879	154	1	1.1	0.9;
	2300	1	149.8	63.1	0	34.02	0	1.028161	19.306285	220	6	1.1	0.9;
	2301	2	0	0	0	0	0	1.033551	29.956876	20	10	1.1	0.9;
	2302	2	0	0	0	0	0	1.058231	41.885187	20	10	1.1	0.9;
	2303	2	0	0	0	0	0	1.005693	-4.241388	20	22	1.1	0.9;
	2304	1	0	-0	0	14.45	0	0.99958	33.087313	380	2	1.1	0.9;
	2305	1	14.72	5.3	0	0.47	0	1.026161	17.223291	110	19	1.1	0.9;
	2306	1	152.1	194.4	0.039528	79.820472	0	1.028005	-16.294789	380	4	1.1	0.9;
	2307	2	0	0	0	0	0	0.994621	93.255028	20	3	1.1	0.9;
	2308	2	0	0	0	0	0	1.001855	42.457602	15.75	1	1.1	0.9;
	2309	1	20.32	2.5	0	5.2	0	0.925212	-7.995152	110	18	1.1	0.9;
	2310	1	0	0	0	0	0	1.012673	-2.175699	380	6	1.1	0.9;
	2311	1	4	0.6	0	3.36	0	0.996151	22.304241	154	1	1.1	0.9;
	2312	1	-0.22	-0.4	0	6.67	0	1.030371	22.492325	220	3	1.1	0.9;
	2313	1	0	-0	0	7.74	0	1.026872	64.663279	220	3	1.1	0.9;
	2314	1	164.2	-10.7	0	11.44	0	1.045287	11.803747	220	3	1.1	0.9;
	2315	1	-0.02	-0.51	0	1.61	0	0.992463	89.041689	150	1	1.1	0.9;
	2316	1	73.18	22.3	0	8.77	0	0.988006	9.38833	220	5	1.1	0.9;
	2317	1	75.88	11.2	0	4.58	0	1.01362	-27.364061	220	4	1.1	0.9;
	2318	2	0	0	0	0	0	1.025724	-14.139236	20	4	1.1	0.9;
	2319	1	123.5	-49.4	0	2.9	0	1.035932	17.18209	220	3	1.1	0.9;
	2320	1	41.95	8.5	0	1.05	0	1.037448	20.429836	220	5	1.1	0.9;
	2321	2	0	0	0	0	0	1.025724	-14.139236	20	4	1.1	0.9;
	2322	2	0	0	0	0	0	1.025724	-14.139236	20	4	1.1	0.9;
	2323	1	0	0	0	20.95	0	1.007398	55.223677	380	3	1.1	0.9;
	2324	2	0	0	0	0	0	1.015556	16.601372	20	5	1.1	0.9;
	2325	2	0	0	0	0	0	1.015556	16.601372	20	5	1.1	0.9;
	2326	1	0	-0	0	0.87	0	0.986681	-9.116329	150	9	1.1	0.9;
	2327	1	20.82	7.9	0	1.07	0	1.016009	-25.270754	220	4	1.1	0.9;
	2328	2	0	0	0	0	0	1.038216	15.918534	20	5	1.1	0.9;
	2329	2	0	0	0	0	0	1.038216	15.918534	20	5	1.1	0.9;
	2330	2	0	0	0	0	0	1.038216	15.918534	20	5	1.1	0.9;
	2331	2	0	0	0	0	0	1.038216	15.918534	20	5	1.1	0.9;
	2332	2	0	0	0	0	0	1.022889	18.879559	20	14	1.1	0.9;
	2333	2	0	0	0	0	0	1.015556	16.601372	20	5	1.1	0.9;
	2334	1	5.31	2.3	0	1.03	0	0.967527	7.346013	110	18	1.1	0.9;
	2335	1	-0	-0	0	0	0	1.014893	-9.156289	20	4	1.1	0.9;
	2336	1	-0.18	-1.06	0	0	0	1.015412	36.155986	150	8	1.1	0.9;
	2337	1	0	0	0.119008	29.381167	0	1.023977	94.20243	380	1	1.1	0.9;
	2338	1	-8.41	2.7	0	0.83	0	1.050498	-5.356327	150	9	1.1	0.9;
	2339	1	7.41	2.7	0	1.28	0	1.0677	-2.868433	110	19	1.1	0.9;
	2340	1	70.08	11.1	0	9.84	0	1.0146	-26.896818	220	4	1.1	0.9;
	2341	2	0	0	0	0	0	0.9846	26.113191	6.3	1	1.1	0.9;
	2342	1	0	0	0	36.99	0	1.027637	71.016364	380	3	1.1	0.9;
	2343	1	51.66	8	0	4.51	0	1.014803	-26.035801	220	4	1.1	0.9;
	2344	1	33.6	4.8	0	5.24	0	1.021485	33.002348	154	1	1.1	0.9;
	2345	2	0	0	0	0	0	1.035781	10.208476	20	18	1.1	0.9;
	2346	1	33.24	11.5	0	1.61	0	1.045938	0.131246	110	19	1.1	0.9;
	2347	2	0	0	0	0	0	1.043619	-8.436397	20	4	1.1	0.9;
	2348	1	-0	0	0.039624	66.502129	0	1.032069	32.464552	380	5	1.1	0.9;
	2349	1	95.71	25.9	0	11.38	0	1.029855	3.966773	220	5	1.1	0.9;
	2350	1	-33.1	-14.35	0	94.55	0	1.063743	32.498552	380	5	1.1	0.9;
	2351	2	0	0	0	0	0	1.015932	-1.816542	20	18	1.1	0.9;
	2352	1	60.27	9.4	0	0.62	0	1.01263	-27.102961	220	4	1.1	0.9;
	2353	1	15.34	-1.3	0	0.71	0	0.998554	89.915574	150	1	1.1	0.9;
	2354	1	124.5	59.2	0	77.23	0	1.034241	21.237398	380	3	1.1	0.9;
	2355	1	13.31	4.8	0	0.2	0	1.039543	0.151979	110	19	1.1	0.9;
	2356	1	103.7	14.9	0	3.04	0	1.022416	2.327808	220	20	1.1	0.9;
	2357	2	0	0	0	0	0	0.958146	44.224727	6.3	1	1.1	0.9;
	2358	1	33.04	15.7	0	1.5	0	1.035093	-4.093903	110	23	1.1	0.9;
	2359	1	216.8	40.5	0	5.91	0	1.056528	16.073477	220	3	1.1	0.9;
	2360	1	-0	0	0	0.73	0	1.05106	28.445589	220	1	1.1	0.9;
	2361	1	22.02	11.9	0	0.79	0	0.981068	-8.315885	110	23	1.1	0.9;
	2362	1	-0	0	0.207485	26.23319	0	0.996546	-0.46454	380	7	1.1	0.9;
	2363	1	0	0	0	45.19	0	1.102599	-3.769686	380	1	1.2	0.8;
	2364	1	15.22	7.6	0	0.6	0	0.968379	9.315915	110	18	1.1	0.9;
	2365	1	257.2	-26.8	0	5.2	0	1.010427	25.129998	220	3	1.1	0.9;
	2366	1	62.07	14.7	0	4.84	0	1.028299	7.415194	220	5	1.1	0.9;
	2367	1	20.9	3	0	1.49	0	1.009453	14.280459	154	1	1.1	0.9;
	2368	2	0	0	0	0	0	1.046905	39.305577	20	2	1.1	0.9;
	2369	2	0	0	0	0	0	1.03654	20.486651	20	5	1.1	0.9;
	2370	2	0	0	0	0	0	1.03654	20.486651	20	5	1.1	0.9;
	2371	2	0	0	0	0	0	1.03654	20.486651	20	5	1.1	0.9;
	2372	2	0	0	0	0	0	1.05433	78.959661	20	3	1.1	0.9;
	2373	2	0	0	0	0	0	1.05433	78.959661	20	3	1.1	0.9;
	2374	1	22.02	7.5	0	0.59	0	0.990002	-4.642297	110	21	1.1	0.9;
	2375	1	3.67	-1.86	0	3.89	0	1.034078	9.468305	220	5	1.1	0.9;
	2376	1	85.39	31.8	0.831419	202.607702	0	1.035818	19.272279	380	3	1.1	0.9;
	2377	2	0	0	0	0	0	1.05433	78.959661	20	3	1.1	0.9;
	2378	1	86.1	40.6	0	7.39	0	1.091952	-4.839993	150	9	1.1	0.9;
	2379	1	125.4	17.8	0	3.11	0	1.034179	0.653604	154	1	1.1	0.9;
	2380	1	20.02	6.6	0	0.74	0	0.951872	-8.263509	110	21	1.1	0.9;
	2381	1	34.04	-0	0	2.4	0	1.01137	-20.329813	220	4	1.1	0.9;
	2382	1	-0.75	-5.36	0	10.46	0	1.07309	71.208471	380	3	1.1	0.9;
	2383	1	91.7	17	0	2.48	0	1.010623	-20.466016	220	4	1.1	0.9;
	2384	1	-0	-0	0.259637	9.774901	0	1.05688	2.348638	380	19	1.1	0.9;
	2385	1	-0	-0	0.181549	8.990889	0	1.089079	12.928353	220	18	1.2	0.8;
	2386	1	46.45	19.1	0	6.66	0	1.042431	26.262064	150	8	1.1	0.9;
	2387	1	3.3	1.1	0	1.12	0	1.000873	-13.941582	110	22	1.1	0.9;
	2388	1	0	0	0	14.82	0	1.026764	35.146887	150	8	1.1	0.9;
	2389	2	0	0	0	0	0	1.00773	-6.62746	20	2	1.1	0.9;
	2390	1	-28.13	-0.83	0	0.1	0	1.058617	17.901573	220	5	1.1	0.9;
	2391	1	4.68	88.21	0	6.79	0	0.99298	89.197377	150	1	1.1	0.9;
	2392	1	74.98	19.8	0	6.47	0	1.026376	19.236355	220	5	1.1	0.9;
	2393	1	55.26	-11.1	0	1.45	0	1.020078	16.65542	220	3	1.1	0.9;
	2394	1	-0	-0	0.188054	-1.648746	0	1.006316	32.241424	380	2	1.1	0.9;
	2395	1	-0	-0	0	140.96	0	1.044734	-14.530759	380	4	1.1	0.9;
	2396	1	9.21	3.3	0	0.6	0	1.032306	1.092988	110	19	1.1	0.9;
	2397	2	0	0	0	0	0	1.01541	24.000055	6.3	1	1.1	0.9;
	2398	1	-0	0	0	15.6	0	1.099009	-14.343804	150	9	1.2	0.8;
	2399	1	-0	-0	0	21.06	0	1.006179	-4.430856	380	7	1.1	0.9;
	2400	1	7.41	2.7	0	1.24	0	0.958965	-1.951824	110	19	1.1	0.9;
	2401	1	22.28	5.67	0	1.2	0	1.0264	-0.200844	110	19	1.1	0.9;
	2402	1	-0	0	0	86.56	0	0.993935	6.959387	220	24	1.1	0.9;
	2403	1	415.5	52.5	0.709256	29.248637	0	1.01934	-7.045149	380	6	1.1	0.9;
	2404	1	-0	-0	0.05988	-0.02988	0	0.999094	-10.340628	150	10	1.1	0.9;
	2405	1	0	0	0.37285	19.231353	0	1.058808	-0.196718	380	21	1.1	0.9;
	2406	1	-0	-0	0	0.94	0	0.940227	-11.17719	150	9	1.1	0.9;
	2407	1	72.18	31.9	0	27.25	0	1.010562	-8.673231	220	6	1.1	0.9;
	2408	2	0	0	0	0	0	1.036065	-16.582357	20	4	1.1	0.9;
	2409	1	0	0	0	58.96	0	1.032898	35.871615	380	8	1.1	0.9;
	2410	2	0	0	0	0	0	1.034871	5.967043	20	14	1.1	0.9;
	2411	2	0	0	0	0	0	1.020161	16.222206	20	5	1.1	0.9;
	2412	1	36.5	5.2	0	1.48	0	1.005589	18.229827	154	1	1.1	0.9;
	2413	1	83.19	25.8	0	0.33	0	1.035366	21.028956	220	5	1.1	0.9;
	2414	1	28.74	1.23	0	2.15	0	0.953878	-6.48994	110	18	1.1	0.9;
	2415	1	3.9	1.1	0	0.19	0	0.980354	-5.292423	110	18	1.1	0.9;
	2416	1	0	0	0.090774	3.510444	0	1.049098	-1.034667	220	18	1.1	0.9;
	2417	1	12.41	4.4	0	0.34	0	1.019402	-2.700194	110	15	1.1	0.9;
	2418	1	207.9	9.2	0.050393	57.239607	0	1.015136	-23.064206	380	4	1.1	0.9;
	2419	1	14.72	5.3	0	1.72	0	1.034936	19.380328	110	19	1.1	0.9;
	2420	1	4.3	1.8	0	0.29	0	1.014957	-4.634794	110	21	1.1	0.9;
	2421	1	5.01	2	0	0.6	0	1.014035	-6.298865	110	23	1.1	0.9;
	2422	2	0	0	0	0	0	1.03091	3.948229	6.3	1	1.1	0.9;
	2423	1	47.45	18.7	0	1.36	0	1.030882	31.218703	150	8	1.1	0.9;
	2424	1	0	-0	1.151541	-0.347329	0	1.056671	13.173411	380	5	1.1	0.9;
	2425	1	62.16	8	0	2.99	0	1.030156	18.368221	220	5	1.1	0.9;
	2426	1	0	0	0	0.4	0	1.025121	-0.421457	150	9	1.1	0.9;
	2427	1	0	0	0	12.18	0	1.020357	44.124293	380	3	1.1	0.9;
	2428	2	0	0	0	0	0	0.994234	37.670381	13.8	1	1.1	0.9;
	2429	2	0	0	0	0	0	1.040642	8.532255	20	14	1.1	0.9;
	2430	2	0	0	0	0	0	1.008982	42.351688	15	1	1.1	0.9;
	2431	1	92.61	36.3	0	4.51	0	0.97922	35.097859	220	3	1.1	0.9;
	2432	1	0	0	0.574318	0.092492	0	1.032532	16.666699	380	5	1.1	0.9;
	2433	1	-15.38	-4.94	0	33.04827	0	0.994714	14.910429	380	5	1.1	0.9;
	2434	2	0	0	0	0	0	1.035431	26.859737	20	6	1.1	0.9;
	2435	2	0	0	0	0	0	1.052787	84.55589	20	3	1.1	0.9;
	2436	2	0	0	0	0	0	1.052787	84.55589	20	3	1.1	0.9;
	2437	2	0	0	0	0	0	1.052787	84.55589	20	3	1.1	0.9;
	2438	2	0	0	0	0	0	1.052787	84.55589	20	3	1.1	0.9;
	2439	2	0	0	0	0	0	1.052787	84.55589	20	3	1.1	0.9;
	2440	2	0	0	0	0	0	1.059619	48.79156	20	3	1.1	0.9;
	2441	1	92.1	29.5	0	3.14	0	0.992448	18.912893	220	5	1.1	0.9;
	2442	1	2.5	0.6	0	0.51	0	0.985787	-4.747483	110	18	1.1	0.9;
	2443	1	4.91	1.4	0	1.32	0	0.968592	-8.970276	110	18	1.1	0.9;
	2444	1	53.56	17.2	0	1.76	0	0.999525	-7.285883	110	21	1.1	0.9;
	2445	1	48.25	6.4	0.039744	0.680256	0	1.002746	-5.871327	110	3	1.1	0.9;
	2446	2	0	0	0	0	0	1.031971	-3.542743	20	6	1.1	0.9;
	2447	2	0	0	0	0	0	1.031971	-3.542743	20	6	1.1	0.9;
	2448	2	0	0	0	0	0	1.031971	-3.542743	20	6	1.1	0.9;
	2449	1	-0	-0	0	16.62	0	0.996275	-13.083407	380	4	1.1	0.9;
	2450	1	0	-0	0.019574	0.340426	0	1.010365	-1.148729	110	18	1.1	0.9;
	2451	1	-14.32	4.7	0	0.35	0	0.988455	-9.696508	150	9	1.1	0.9;
	2452	1	157.64	115.5	0.462433	27.713907	0	1.062716	50.740342	220	13	1.1	0.9;
	2453	1	0	-0	0.683135	2.569185	0	1.060805	38.8513	380	3	1.1	0.9;
	2454	1	0	0	0.019235	0.030765	0	1.047172	32.227056	220	5	1.1	0.9;
	2455	1	-0.2	-1	0	0	0	1.018412	31.685509	150	8	1.1	0.9;
	2456	1	-0	0	0	57.29	0	1.051156	7.856494	380	16	1.1	0.9;
	2457	1	8.01	4.9	0	0	0	0.987908	-7.451534	150	9	1.1	0.9;
	2458	2	0	0	0	0	0	1.020659	-5.050211	20	10	1.1	0.9;
	2459	2	0	0	0	0	0	1.037973	11.471538	20	14	1.1	0.9;
	2460	2	0	0	0	0	0	1.011637	15.19334	13.8	1	1.1	0.9;
	2461	1	38.04	29	0.057188	20.082812	0	1.057377	6.088258	220	11	1.1	0.9;
	2462	2	0	0	0	0	0	1.033042	-4.165701	20	7	1.1	0.9;
	2463	1	0	0	0	0	0	0.985529	-4.986197	150	9	1.1	0.9;
	2464	2	0	0	0	0	0	1.011692	-19.302078	20	4	1.1	0.9;
	2465	1	34.44	14.4	0	48.56	0	1.019953	-21.391835	380	4	1.1	0.9;
	2466	1	68.38	42.4	0	1.39	0	1.013533	-4.833141	150	9	1.1	0.9;
	2467	1	-17.18	-4.35	0	44.421484	0	1.005681	16.085441	380	5	1.1	0.9;
	2468	1	19.92	6.6	0	1.32	0	0.989597	-8.414096	110	21	1.1	0.9;
	2469	1	0	0	0.135375	28.222175	0	1.072267	-1.044825	380	18	1.2	0.8;
	2470	1	16.02	5.3	0	0.67	0	0.994222	-11.370405	110	12	1.1	0.9;
	2471	1	0	-0	0	3.65	0	0.977418	-27.229276	220	10	1.1	0.9;
	2472	1	-0	-0	0	1.2	0	1.06633	0.67803	150	9	1.1	0.9;
	2473	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	2474	1	0	0	0.115864	-0.096553	0	1.06026	-1.965756	110	15	1.1	0.9;
	2475	1	9.71	3.5	0	1.44	0	1.033543	-0.572386	110	19	1.1	0.9;
	2476	1	0	0	0	11.14	0	1.053994	-17.781594	220	6	1.1	0.9;
	2477	1	0	-0	0	62.36	0	0.975014	23.320703	400	1	1.1	0.9;
	2478	1	-0	0	0	5.09	0	1.069032	17.910132	220	3	1.1	0.9;
	2479	1	10.11	2.5	0	0.47	0	0.995942	-4.331716	110	18	1.1	0.9;
	2480	2	0	0	0	0	0	1.078638	63.716182	20	3	1.1	0.9;
	2481	2	0	0	0	0	0	1.078638	63.716182	20	3	1.1	0.9;
	2482	2	0	0	0	0	0	1.02424	84.887043	20	3	1.1	0.9;
	2483	1	0	-0	0	6.68	0	1.026073	-2.579002	220	5	1.1	0.9;
	2484	1	0	-0.1	0.08029	-0.060218	0	1.017615	25.421095	220	6	1.1	0.9;
	2485	2	0	0	0	0	0	1.015531	4.163384	20	9	1.1	0.9;
	2486	2	0	0	0	0	0	1.04434	27.891337	20	14	1.1	0.9;
	2487	2	0	0	0	0	0	1.04434	27.891337	20	14	1.1	0.9;
	2488	2	0	0	0	0	0	1.055727	25.429279	20	14	1.1	0.9;
	2489	2	0	0	0	0	0	1.055727	25.429279	20	14	1.1	0.9;
	2490	2	0	0	0	0	0	1.04434	27.891337	20	14	1.1	0.9;
	2491	1	0	0	0	0	0	0.966548	-5.661815	150	9	1.1	0.9;
	2492	1	0	0	0	1.86	0	1.03632	5.804646	220	5	1.1	0.9;
	2493	1	-0	-0	0.511956	7.711812	0	1.055063	34.721991	380	8	1.1	0.9;
	2494	1	-0	0	0.235883	0.243431	0	1.023762	-7.315358	380	6	1.1	0.9;
	2495	1	39.24	24.3	0	0.07	0	1.016491	-11.127263	150	9	1.1	0.9;
	2496	1	0	0	0	0	0	1.03472	7.897462	20	7	1.1	0.9;
	2497	1	58.46	6.6	0	10.13	0	1.027844	11.224092	220	5	1.1	0.9;
	2498	1	5.41	1.8	0	0.72	0	0.99908	-9.995212	110	12	1.1	0.9;
	2499	1	11.51	2.3	0	0.66	0	1.001143	3.67274	110	18	1.1	0.9;
	2500	1	0	0	0	0	0	1.030227	6.942059	20	23	1.1	0.9;
	2501	1	0	0	0	0	0	1.030227	6.942059	20	23	1.1	0.9;
	2502	1	0	0	0	0	0	1.030227	6.942059	20	23	1.1	0.9;
	2503	1	85.8	13.4	0	3.23	0	1.023024	-25.867914	220	4	1.1	0.9;
	2504	1	0	-0	0	3.13	0	1.014237	-26.217572	220	4	1.1	0.9;
	2505	1	0	0	0.12	1.51	0	1.000523	-2.320988	220	21	1.1	0.9;
	2506	2	0	0	0	0	0	0.990208	39.96266	13.8	1	1.1	0.9;
	2507	1	0	-0	0.020109	133.82196	0	1.031983	12.039209	380	5	1.1	0.9;
	2508	1	199.4	-19.9	0	6.68	0	1.041513	-3.711399	220	5	1.1	0.9;
	2509	1	-1.5	-41.98	0	36.76	0	1.056029	48.463068	380	3	1.1	0.9;
	2510	2	0	0	0	0	0	1.026207	-11.793103	20	4	1.1	0.9;
	2511	1	0	0	0.232565	17.73498	0	1.0377	-0.080977	220	18	1.1	0.9;
	2512	1	80.09	20	0	3.08	0	1.016536	42.941822	220	3	1.1	0.9;
	2513	2	0	0	0	0	0	1.033542	1.716976	20	6	1.1	0.9;
	2514	2	0	0	0	0	0	1.033552	1.72662	20	6	1.1	0.9;
	2515	2	0	0	0	0	0	1.033552	1.72662	20	6	1.1	0.9;
	2516	2	0	0	0	0	0	1.033552	1.72662	20	6	1.1	0.9;
	2517	1	0	-0	0	220.46	0	1.081376	9.242874	380	18	1.2	0.8;
	2518	2	0	0	0	0	0	1.026319	-4.093349	20	6	1.1	0.9;
	2519	1	0	0	0	1.15	0	1.039767	-3.494237	220	6	1.1	0.9;
	2520	1	328.9	30.4	0	66.79	0	1.025946	-14.450303	380	6	1.1	0.9;
	2521	1	11.11	6.9	0	1.35	0	1.024048	-7.803527	150	9	1.1	0.9;
	2522	1	45.75	20.1	0	0.41	0	1.009827	3.156346	110	18	1.1	0.9;
	2523	1	0	0	0	0.28	0	0.991127	-1.59697	380	5	1.1	0.9;
	2524	1	25.43	14.7	0	7.79	0	1.057969	-0.595757	220	18	1.2	0.8;
	2525	1	122.2	68.9	0	0.58	0	1.036911	18.978271	380	3	1.1	0.9;
	2526	2	0	0	0	0	0	1.01592	3.179112	0.4	1	1.1	0.9;
	2527	1	26.03	8.7	0	0.29	0	0.986413	-2.111327	110	18	1.1	0.9;
	2528	1	333.5	-1.4	0	8.37	0	1.010953	-17.43421	220	5	1.1	0.9;
	2529	1	73.98	22.5	0	1.15	0	1.043181	18.536382	220	5	1.1	0.9;
	2530	1	3	1.5	0	0.4	0	1.023175	-5.240961	110	23	1.1	0.9;
	2531	1	-0	0	0	2.55	0	1.006174	-13.997849	220	10	1.1	0.9;
	2532	1	-0	0	0	0.54	0	0.999111	-3.504719	110	21	1.1	0.9;
	2533	2	0	0	0	0	0	1.014104	26.485556	20	6	1.1	0.9;
	2534	2	0	0	0	0	0	1.046641	37.417065	20	5	1.1	0.9;
	2535	1	12.6	1.8	0	3.8	0	1.004397	16.235341	154	1	1.1	0.9;
	2536	1	-0	0	0	11.49	0	0.993039	48.679905	220	3	1.1	0.9;
	2537	1	226.4	-13	0	15.2	0	1.009987	-18.52003	220	4	1.1	0.9;
	2538	1	-0	-0	0	1.64	0	1.050999	4.005332	220	5	1.1	0.9;
	2539	2	0	0	0	0	0	1.000445	95.959998	380	1	1.1	0.9;
	2540	2	0	0	0	0	0	1.046641	37.417065	20	5	1.1	0.9;
	2541	1	101.3	17.6	0	0.12	0	1.010105	-20.870277	220	4	1.1	0.9;
	2542	1	-0.18	-3.85	0	0	0	1.017997	42.29957	150	8	1.1	0.9;
	2543	2	0	0	0	0	0	1.046641	37.417065	20	5	1.1	0.9;
	2544	1	5.51	2	0	0.79	0	0.993408	-5.133948	110	19	1.1	0.9;
	2545	1	-1.8	-4.8	0	0.44	0	1.05382	-3.028079	150	9	1.1	0.9;
	2546	1	16.52	5.5	0	1.59	0	1.009455	-6.474424	110	22	1.1	0.9;
	2547	1	28.63	10	0	1.34	0	0.994668	-2.235494	110	19	1.1	0.9;
	2548	1	0	-0.8	0	0.46	0	1.036036	25.403437	110	18	1.1	0.9;
	2549	1	0	0	0	0	0	0.999524	-9.834457	150	9	1.1	0.9;
	2550	1	169.8	-19.6	0	53.619841	0	1.027588	-1.88516	380	6	1.1	0.9;
	2551	1	18.4	10.1	0.672256	21.491354	0	1.097221	2.395495	380	9	1.2	0.8;
	2552	1	32.94	10.1	0	0.41	0	1.014815	43.027357	110	8	1.1	0.9;
	2553	1	-0	0	0.353382	20.075147	0	1.013365	43.78577	380	8	1.1	0.9;
	2554	1	0	-0	0	2.19	0	1.072967	24.077735	220	3	1.1	0.9;
	2555	1	-19.32	6.4	0	0	0	1.043042	-1.38939	150	9	1.1	0.9;
	2556	1	-0	-0	0	6.8	0	1.017559	41.986243	380	3	1.1	0.9;
	2557	1	166.1	64.9	0.039665	2.130335	0	1.010888	-16.54274	220	4	1.1	0.9;
	2558	1	-0	-0	1.139646	11.552147	0	1.053981	30.656361	380	5	1.1	0.9;
	2559	2	0	0	0	0	0	1.029344	21.964639	20	10	1.1	0.9;
	2560	1	71.18	23.8	0	0.07	0	0.98751	-4.706892	110	21	1.1	0.9;
	2561	2	0	0	0	0	0	1.024067	20.713541	20	10	1.1	0.9;
	2562	1	13.31	7.9	0	0.45	0	0.986502	-4.442456	110	18	1.1	0.9;
	2563	1	12.31	0.2	0.039377	3.700623	0	1.00122	-3.581791	110	18	1.1	0.9;
	2564	1	-0.63	-0.83	0	22.772633	0	0.986561	22.995822	380	5	1.1	0.9;
	2565	1	0	-0	0	1.46	0	1.027161	2.397403	154	1	1.1	0.9;
	2566	2	0	0	0	0	0	1.002755	40.303215	11	1	1.1	0.9;
	2567	1	8.3	1.2	0	4.96	0	1.018406	22.88777	154	1	1.1	0.9;
	2568	1	-0	0	0	0.67	0	1.029368	-2.3572	150	9	1.1	0.9;
	2569	2	0	0	0	0	0	1.030434	0.665089	20	23	1.1	0.9;
	2570	1	8.51	3.2	0	2.93	0	0.990186	-18.851277	220	4	1.1	0.9;
	2571	2	0	0	0	0	0	1.046641	37.417065	20	5	1.1	0.9;
	2572	2	0	0	0	0	0	1.014104	26.485556	20	6	1.1	0.9;
	2573	2	0	0	0	0	0	1.014104	26.485556	20	6	1.1	0.9;
	2574	1	-0	-0	0	0.05	0	1.031914	8.483142	220	5	1.1	0.9;
	2575	1	23.03	10.5	0	3.96	0	1.000415	-3.000786	220	6	1.1	0.9;
	2576	1	112.6	16.1	0	3.38	0	1.011272	29.276423	154	1	1.1	0.9;
	2577	2	0	0	0	0	0	1.046641	37.417065	20	5	1.1	0.9;
	2578	1	-4.17	-2.16	0	0.66	0	1.003204	-1.650544	110	18	1.1	0.9;
	2579	1	25.43	8.3	0	1.35	0	0.929204	-10.397849	110	21	1.1	0.9;
	2580	1	-3.55	-1.43	0	11.839519	0	1.009101	27.36724	380	5	1.1	0.9;
	2581	1	166.2	55.1	0.117826	7.712174	0	1.027592	28.817146	220	5	1.1	0.9;
	2582	1	-0	-0	0	239.779558	0	1.007085	5.065757	750	20	1.1	0.9;
	2583	1	0	-0	0	0	0	0.989033	-4.933794	20	18	1.1	0.9;
	2584	2	0	0	0	0	0	1.031301	-10.495578	20	10	1.1	0.9;
	2585	1	52.46	11.3	0	0.27	0	1.03974	48.885471	110	8	1.1	0.9;
	2586	1	99.51	1.5	0	32.89	0	1.038712	59.525463	380	3	1.1	0.9;
	2587	1	44.65	8.8	0	20.17	0	1.034957	4.77046	220	5	1.1	0.9;
	2588	1	0	-0	0.116549	-0.087411	0	1.031813	4.43475	380	6	1.1	0.9;
	2589	1	13.9	1.9	0	3.95	0	1.029269	36.020984	154	1	1.1	0.9;
	2590	1	88.92	19.79	0	2.81	0	1.008801	22.021191	220	5	1.1	0.9;
	2591	1	34.14	-2.8	0.239003	18.790831	0	1.024105	8.131913	380	6	1.1	0.9;
	2592	2	0	0	0	0	0	1.014299	9.180056	20	22	1.1	0.9;
	2593	1	63.66	9.7	0	0	0	1.063957	30.159406	220	2	1.1	0.9;
	2594	1	335.3	-101.3	0	32.69	0	1.037812	4.529303	380	11	1.1	0.9;
	2595	1	108.91	33.7	0	56.24	0	1.042511	18.111486	380	5	1.1	0.9;
	2596	1	-29.73	9.8	0	2.31	0	1.009993	-5.164253	150	9	1.1	0.9;
	2597	1	-0	-0	0.198366	-0.008774	0	1.026857	15.275337	380	6	1.1	0.9;
	2598	1	0	0	0	0.06	0	0.998698	2.211508	220	19	1.1	0.9;
	2599	1	14.62	2	0	10.89	0	1.088032	11.809018	220	18	1.2	0.8;
	2600	1	-0	0	0	3.33	0	1.030403	18.225281	220	3	1.1	0.9;
	2601	1	0	0	0.467058	2.950246	0	1.027954	-13.585392	380	6	1.1	0.9;
	2602	1	-0	0	0.438658	64.65877	0	1.015817	42.07164	380	3	1.1	0.9;
	2603	2	0	0	0	0	0	1.004901	24.768889	10.5	1	1.1	0.9;
	2604	1	-0.65	-1.81	0	1.52	0	1.086546	-3.616335	150	9	1.1	0.9;
	2605	1	0	0	0	0.38	0	1.036591	27.54814	380	3	1.1	0.9;
	2606	2	0	0	0	0	0	1.033036	-3.138558	20	6	1.1	0.9;
	2607	2	0	0	0	0	0	1.033036	-3.138558	20	6	1.1	0.9;
	2608	1	56.46	22	0	41.96	0	1.014963	-11.37104	220	6	1.1	0.9;
	2609	1	114.2	16.3	0	57.91	0	0.98271	5.157436	400	1	1.1	0.9;
	2610	2	0	0	0	0	0	1.017828	-6.986287	20	4	1.1	0.9;
	2611	1	5.01	13	0	30.48	0	1.015852	3.041504	220	11	1.1	0.9;
	2612	1	17.82	6.7	0	0.09	0	1.054932	-1.045505	110	15	1.1	0.9;
	2613	1	-0	-0	0	0.51	0	1.024786	-10.50186	150	9	1.1	0.9;
	2614	1	34.14	21.1	0	0.35	0	0.963689	-9.802775	150	9	1.1	0.9;
	2615	1	0	-0	0	3.33	0	1.021143	7.911593	220	5	1.1	0.9;
	2616	1	8.21	2.5	0	0.12	0	1.006569	-1.11117	110	18	1.1	0.9;
	2617	1	44.23	22.19	0	4.35	0	0.996039	26.252459	220	13	1.1	0.9;
	2618	1	62.37	25.5	0	3.98	0	1.032639	25.002791	150	8	1.1	0.9;
	2619	1	113.27	27.67	0	3.61	0	1.025318	25.079708	220	13	1.1	0.9;
	2620	1	0.4	-170.2	0	3.04	0	1.044314	12.69888	380	3	1.1	0.9;
	2621	1	112	19.7	0	38.04	0	1.028919	7.369747	220	5	1.1	0.9;
	2622	1	-0	0	0	0.62	0	0.96688	-9.635467	150	9	1.1	0.9;
	2623	1	0	-0	0	25.22	0	1.052837	28.078372	220	1	1.1	0.9;
	2624	2	0	0	0	0	0	1.045696	14.309815	20	18	1.1	0.9;
	2625	2	0	0	0	0	0	1.039006	38.999932	20	2	1.1	0.9;
	2626	1	30.53	15.1	0.061431	0.888569	0	0.978212	-2.836179	110	19	1.1	0.9;
	2627	2	0	0	0	0	0	1.042062	24.663334	20	18	1.1	0.9;
	2628	1	123.9	17.6	0	11.56	0	1.02517	17.079665	154	1	1.1	0.9;
	2629	1	0	0	0	3.07	0	0.989897	48.462007	154	1	1.1	0.9;
	2630	1	154.8	-47.1	0	4.18	0	1.017958	-24.080211	220	4	1.1	0.9;
	2631	1	22.02	21.8	0	10.29	0	1.025806	-5.755511	220	10	1.1	0.9;
	2632	1	5.11	2.4	0	1.34	0	0.992408	13.822029	110	18	1.1	0.9;
	2633	1	7.81	5.8	0.030123	9.949877	0	1.009876	-24.041959	220	4	1.1	0.9;
	2634	1	-0	0	0.019998	0.030002	0	1.033067	16.604846	220	5	1.1	0.9;
	2635	2	0	0	0	0	0	1.028604	-12.690591	20	4	1.1	0.9;
	2636	1	25.73	8.9	0	1.29	0	0.982579	1.681904	110	18	1.1	0.9;
	2637	1	0.2	0.5	0	0.42	0	1.047829	86.05075	220	3	1.1	0.9;
	2638	1	0	-0	0	1.94	0	0.97736	31.54763	220	3	1.1	0.9;
	2639	1	-41.65	13.7	0	0.3	0	1.037572	-0.519639	150	9	1.1	0.9;
	2640	1	5.51	2	0	0.95	0	1.040335	20.591441	110	19	1.1	0.9;
	2641	1	0	-0	0	1.53	0	0.992684	17.756538	220	5	1.1	0.9;
	2642	2	0	0	0	0	0	1.030658	49.902414	20	10	1.1	0.9;
	2643	1	369.5	16.7	0	10.6	0	0.979525	7.924919	220	6	1.1	0.9;
	2644	1	14	2	0	3.32	0	0.995624	19.508108	154	1	1.1	0.9;
	2645	2	0	0	0	0	0	1.013484	22.942549	11.5	1	1.1	0.9;
	2646	2	0	0	0	0	0	1.037142	24.334502	20	5	1.1	0.9;
	2647	2	0	0	0	0	0	1.0136	-3.104621	20	10	1.1	0.9;
	2648	1	-0	0	0	19.01	0	1.025366	45.298404	380	3	1.1	0.9;
	2649	1	177.4	41.7	0	2.98	0	0.979734	-12.99789	150	10	1.1	0.9;
	2650	2	0	0	0	0	0	1.0608	42.720858	20	3	1.1	0.9;
	2651	2	0	0	0	0	0	1.0608	42.720858	20	3	1.1	0.9;
	2652	2	0	0	0	0	0	1.058549	46.919585	20	3	1.1	0.9;
	2653	2	0	0	0	0	0	1.058549	46.919585	20	3	1.1	0.9;
	2654	2	0	0	0	0	0	1.058549	46.919585	20	3	1.1	0.9;
	2655	2	0	0	0	0	0	1.058549	46.919585	20	3	1.1	0.9;
	2656	2	0	0	0	0	0	1.058549	46.919585	20	3	1.1	0.9;
	2657	2	0	0	0	0	0	1.058549	46.919585	20	3	1.1	0.9;
	2658	1	-19.62	28.6	0	1.69	0	1.05595	-6.533405	150	9	1.1	0.9;
	2659	1	55.26	23.2	0	8.01	0	1.031923	-1.472759	220	18	1.1	0.9;
	2660	1	209.2	23.5	0	6.44	0	1.051111	1.167386	220	5	1.1	0.9;
	2661	1	0	0	0	0.06	0	1.024083	-4.130597	220	6	1.1	0.9;
	2662	2	0	0	0	0	0	1.055337	0.681282	20	9	1.1	0.9;
	2663	1	11.71	3.8	0	2.96	0	1.011735	11.840184	220	5	1.1	0.9;
	2664	1	0	0	0	0	0	1.016944	-15.006207	20	4	1.1	0.9;
	2665	1	0	0	0	0	0	1.016944	-15.006207	20	4	1.1	0.9;
	2666	2	0	0	0	0	0	1.027807	-10.041134	20	4	1.1	0.9;
	2667	2	0	0	0	0	0	1.027807	-10.041134	20	4	1.1	0.9;
	2668	1	-0	-0	0	0.74	0	0.991249	25.858246	150	8	1.1	0.9;
	2669	2	0	0	0	0	0	1.027807	-10.041134	20	4	1.1	0.9;
	2670	2	0	0	0	0	0	1.024891	10.992359	20	18	1.1	0.9;
	2671	1	15.57	9.59	0	0.41	0	1.01802	8.028494	110	18	1.1	0.9;
	2672	1	94.3	13.8	0	5.3	0	1.015027	-26.995567	220	4	1.1	0.9;
	2673	1	83.29	30.1	0	4.68	0	0.976247	-1.700019	110	19	1.1	0.9;
	2674	1	223.9	72.6	0	2.04	0	1.037377	-7.9846	110	1	1.1	0.9;
	2675	1	0	0	0	0	0	0.999171	-8.851257	150	9	1.1	0.9;
	2676	2	0	0	0	0	0	1.012735	-8.652706	20	4	1.1	0.9;
	2677	1	0	0	0	6.262609	0	1.025349	15.599134	380	5	1.1	0.9;
	2678	1	1.2	0.8	0.114435	4.545565	0	1.055434	0.517348	150	9	1.1	0.9;
	2679	1	-0	-0	0	0.13	0	0.995829	-3.792126	110	18	1.1	0.9;
	2680	1	-0	0	0	0.57	0	1.020153	-10.106982	150	9	1.1	0.9;
	2681	2	0	0	0	0	0	1.043665	14.930439	20	14	1.1	0.9;
	2682	1	50.26	16.4	0	0.23	0	0.985714	25.1639	150	8	1.1	0.9;
	2683	1	-0	-0	0	50.2	0	0.984403	20.182088	400	1	1.1	0.9;
	2684	1	0.2	108.3	0.167319	83.060568	0	1.07544	81.399784	380	3	1.2	0.8;
	2685	1	27.33	5	0	6.77	0	1.036282	13.112105	220	5	1.1	0.9;
	2686	1	-0	-0	0	0	0	1.021606	-18.065799	20	4	1.1	0.9;
	2687	2	0	0	0	0	0	1.034314	-10.021136	20	4	1.1	0.9;
	2688	2	0	0	0	0	0	1.021208	-7.595433	20	4	1.1	0.9;
	2689	1	2.2	1.1	0	1.57	0	0.996623	-13.551816	110	22	1.1	0.9;
	2690	2	0	0	0	0	0	1.034314	-10.021136	20	4	1.1	0.9;
	2691	2	0	0	0	0	0	1.01939	-11.773065	20	4	1.1	0.9;
	2692	1	-0.75	-0.93	0	0.26	0	1.013327	-8.475961	110	12	1.1	0.9;
	2693	2	0	0	0	0	0	1.021055	-14.38773	20	4	1.1	0.9;
	2694	2	0	0	0	0	0	1.021055	-14.38773	20	4	1.1	0.9;
	2695	2	0	0	0	0	0	1.030241	-13.589685	20	4	1.1	0.9;
	2696	1	182.32	35.4	0.524537	0.022886	0	1.049033	30.517749	220	5	1.1	0.9;
	2697	1	57.16	-0	0	6.58	0	1.025817	39.935591	110	8	1.1	0.9;
	2698	2	0	0	0	0	0	1.03058	15.553508	20	14	1.1	0.9;
	2699	1	0	-0	0.925827	-0.732946	0	1.059225	86.44158	380	8	1.1	0.9;
	2700	2	0	0	0	0	0	1.024762	17.972121	20	14	1.1	0.9;
	2701	1	28.63	17.8	0	0.06	0	1.008172	-8.50998	150	9	1.1	0.9;
	2702	2	0	0	0	0	0	1.013686	-8.216227	20	4	1.1	0.9;
	2703	2	0	0	0	0	0	1.031603	9.382948	20	5	1.1	0.9;
	2704	1	29.4	4.2	0	2	0	0.991906	20.94822	154	1	1.1	0.9;
	2705	2	0	0	0	0	0	1.031603	9.382948	20	5	1.1	0.9;
	2706	1	32.34	14.1	0	0.24	0	1.000198	-1.38798	110	18	1.1	0.9;
	2707	1	22.22	5.9	0	10.28	0	1.028806	-20.398353	220	4	1.1	0.9;
	2708	2	0	0	0	0	0	1.066367	6.221208	20	5	1.1	0.9;
	2709	1	1.4	0.5	0	0.89	0	1.063696	0.616025	110	19	1.1	0.9;
	2710	1	20.82	-0	0	0.14	0	1.055275	37.598344	220	5	1.1	0.9;
	2711	2	0	0	0	0	0	1.042669	14.686826	20	5	1.1	0.9;
	2712	2	0	0	0	0	0	1.016984	26.026364	11.5	1	1.1	0.9;
	2713	1	53.96	4.9	0	4.21	0	1.01779	-32.964038	220	4	1.1	0.9;
	2714	1	0.5	21.5	0.190226	30.657331	0	1.062807	-8.18081	380	4	1.1	0.9;
	2715	1	0	0	0.202076	48.319777	0	1.049679	-19.588649	380	10	1.1	0.9;
	2716	1	-0	0	0.334631	25.761141	0	1.082729	86.239056	380	3	1.2	0.8;
	2717	2	0	0	0	0	0	0.996677	47.723917	14.4	1	1.1	0.9;
	2718	1	0.2	1.4	0	0.49	0	1.003112	32.826245	220	3	1.1	0.9;
	2719	1	-0	-0	0	35.14	0	1.031244	39.254249	380	8	1.1	0.9;
	2720	1	26.43	7.7	0	7.88	0	1.016666	-11.633661	110	22	1.1	0.9;
	2721	2	0	0	0	0	0	1.029248	22.566355	11	1	1.1	0.9;
	2722	1	0	-0	0	3.15	0	1.019828	-10.992526	150	9	1.1	0.9;
	2723	1	16.02	5.3	0	0.67	0	0.983665	-13.090398	110	12	1.1	0.9;
	2724	1	17.5	502.7	0.319765	0.651294	0	1.099472	10.206014	380	19	1.2	0.8;
	2725	1	25	3.6	0	2.17	0	0.976233	42.311602	154	1	1.1	0.9;
	2726	1	0	0	0.057077	-0.047564	0	1.049529	86.878119	220	3	1.1	0.9;
	2727	1	-11.95	-3.92	0	0.18	0	1.013271	-0.962724	110	18	1.1	0.9;
	2728	1	99.4	35.8	0.605414	57.517845	0	1.043768	39.249593	380	3	1.1	0.9;
	2729	1	0	-0	0	6.21	0	1.040621	17.754771	220	5	1.1	0.9;
	2730	1	39.54	14.3	0	0.91	0	1.030248	-2.052544	110	19	1.1	0.9;
	2731	1	62.17	0.7	0	5.98	0	1.027632	-14.918109	220	4	1.1	0.9;
	2732	1	52.06	14.3	0	1.59	0	0.980785	-17.6427	220	4	1.1	0.9;
	2733	1	91.1	54.8	0	30.59	0	1.028095	35.299342	150	8	1.1	0.9;
	2734	2	0	0	0	0	0	1.025803	97.507727	20	3	1.1	0.9;
	2735	1	-0	0	0.241629	-0.061629	0	1.03532	8.357264	220	5	1.1	0.9;
	2736	1	33.64	3.9	0	3.37	0	1.026937	0.980466	220	7	1.1	0.9;
	2737	1	0	0	0	0	0	1.084215	30.391991	220	5	1.1	0.9;
	2738	1	10.51	3.5	0	0.66	0	0.980974	-14.106135	110	12	1.1	0.9;
	2739	1	-0	-0	0	13.33	0	1.010421	31.816906	380	3	1.1	0.9;
	2740	1	197.5	-31	0	17.34	0	1.038514	4.74121	220	5	1.1	0.9;
	2741	1	-0	-0	0	21.97	0	1.02475	27.384949	380	3	1.1	0.9;
	2742	1	-0	-0	0.361262	2.649018	0	1.021263	23.715333	380	6	1.1	0.9;
	2743	1	0	-0	0	0.3	0	1.021295	7.910264	220	5	1.1	0.9;
	2744	1	49.25	4.9	0	1.06	0	1.003913	7.649642	110	18	1.1	0.9;
	2745	1	0	0	0.126343	10.705031	0	1.078554	4.187041	220	19	1.2	0.8;
	2746	1	192.11	45.5	0	68.04	0	0.996714	19.141779	380	5	1.1	0.9;
	2747	1	108.3	62	0	36.65	0	1.017811	4.765995	220	7	1.1	0.9;
	2748	1	192.3	27.4	0	120.83	0	0.964803	22.614166	400	1	1.1	0.9;
	2749	1	0	0	0	21.18	0	1.020315	57.22313	380	3	1.1	0.9;
	2750	1	64.37	38.3	0.291893	16.726756	0	1.03288	23.275029	220	3	1.1	0.9;
	2751	1	-0	-0	0	2.16	0	1.003553	-12.572953	110	22	1.1	0.9;
	2752	1	45.6	6.5	0	95.24	0	0.951282	29.908619	400	1	1.1	0.9;
	2753	1	14.8	6.6	0	84.333648	0	1.121048	-0.477263	380	9	1.2	0.8;
	2754	1	177.4	85.5	0.040271	8.399796	0	0.982851	19.049233	220	13	1.1	0.9;
	2755	1	51.86	13.9	0	0.67	0	0.972195	-12.166593	110	1	1.1	0.9;
	2756	2	0	0	0	0	0	0.998215	12.734047	20	18	1.1	0.9;
	2757	1	-0	0	0	45.35	0	1.052606	2.644204	380	1	1.1	0.9;
	2758	1	0	0	0	4.29	0	1.022049	-13.383678	380	6	1.1	0.9;
	2759	2	0	0	0	0	0	0.990464	21.09396	15.75	1	1.1	0.9;
	2760	2	0	0	0	0	0	1.058355	41.76159	20	10	1.1	0.9;
	2761	1	-0	-0	0	0.07	0	1.025265	-10.510786	150	9	1.1	0.9;
	2762	1	98.6	14	0	0.31	0	1.004778	27.105564	154	1	1.1	0.9;
	2763	1	-0	-0	0	30.78	0	1.015003	31.694995	150	8	1.1	0.9;
	2764	1	0	0	0	4.76	0	1.048333	57.693081	380	1	1.1	0.9;
	2765	1	-0	-0	0	13.03	0	1.028071	36.87365	154	1	1.1	0.9;
	2766	1	-11.44	-3.29	0	0.04	0	1.012041	5.279989	220	6	1.1	0.9;
	2767	2	0	0	0	0	0	1.003847	10.087109	20	18	1.1	0.9;
	2768	1	0	0	0	1.19	0	1.063889	11.960574	220	5	1.1	0.9;
	2769	1	0	-0	0.113836	13.934623	0	1.051211	1.426245	220	6	1.1	0.9;
	2770	1	44.75	4.1	0	3.12	0	1.021137	-31.689864	220	4	1.1	0.9;
	2771	1	4.4	99.7	0	40.45	0	1.042717	32.921323	380	3	1.1	0.9;
	2772	1	-0	-0	0	0.27	0	1.042842	84.150417	220	3	1.1	0.9;
	2773	1	0	0	0	0	0	0.969729	-14.732562	150	9	1.1	0.9;
	2774	1	19.42	-0.8	0	0.78	0	1.034866	65.764397	220	3	1.1	0.9;
	2775	1	21.82	10.3	0	32.61	0	1.050756	-12.792959	380	4	1.1	0.9;
	2776	1	32.2	4.6	0	2.54	0	1.03265	18.464929	154	1	1.1	0.9;
	2777	2	0	0	0	0	0	1.009892	4.868208	20	9	1.1	0.9;
	2778	1	0	0	0	4.96	0	0.99744	33.84332	220	3	1.1	0.9;
	2779	1	159.8	45	0	5.6	0	1.008162	19.967949	220	3	1.1	0.9;
	2780	1	0	-0	0	2.1	0	1.006374	26.856693	380	3	1.1	0.9;
	2781	1	-0	-0	0	0.76	0	1.049701	7.525621	220	1	1.1	0.9;
	2782	1	115.1	10.4	0	6.12	0	1.020405	-1.824814	220	11	1.1	0.9;
	2783	2	0	0	0	0	0	1.040793	26.940223	20	5	1.1	0.9;
	2784	2	0	0	0	0	0	1.019419	6.474548	11	1	1.1	0.9;
	2785	2	0	0	0	0	0	1.003847	10.087109	20	18	1.1	0.9;
	2786	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	2787	2	0	0	0	0	0	1.017845	-4.792355	20	10	1.1	0.9;
	2788	1	21.62	13.4	0	1.79	0	1.017792	-5.319138	150	9	1.1	0.9;
	2789	2	0	0	0	0	0	1.011998	0.419036	20	18	1.1	0.9;
	2790	2	0	0	0	0	0	1.031726	-2.251951	20	18	1.1	0.9;
	2791	1	0	0	0	10.51	0	1.019413	41.884948	380	3	1.1	0.9;
	2792	1	0	0	0	51.66	0	1.086757	63.706535	380	8	1.1	0.9;
	2793	1	-0	0	0	0	0	1.020424	-20.573487	20	4	1.1	0.9;
	2794	1	76.28	2.9	0	7.28	0	1.021416	-21.165213	220	4	1.1	0.9;
	2795	1	23.13	8.8	0	0.95	0	1.003987	-6.326655	110	22	1.1	0.9;
	2796	2	0	0	0	0	0	1.022889	18.879559	20	14	1.1	0.9;
	2797	1	-0	-0	0.01938	0.48062	0	1.061902	18.798237	220	5	1.1	0.9;
	2798	1	-0	0	0	1.94	0	0.996885	-3.778126	110	21	1.1	0.9;
	2799	2	0	0	0	0	0	1.040793	26.940223	20	5	1.1	0.9;
	2800	1	3.2	1.5	0	1.81	0	0.968535	5.04654	110	18	1.1	0.9;
	2801	1	-0	-0	0	0.27	0	1.047904	84.142942	220	3	1.1	0.9;
	2802	1	179.5	85.21	0.04028	0.53979	0	0.982717	19.046316	220	13	1.1	0.9;
	2803	2	0	0	0	0	0	1.015042	-0.37136	20	18	1.1	0.9;
	2804	1	248.7	8.1	0	12.46	0	1.021573	-22.183058	220	5	1.1	0.9;
	2805	1	-20.6	-9.4	0	1.12	0	1.02996	-10.060779	150	9	1.1	0.9;
	2806	1	46.05	16.7	0	0.49	0	0.973439	-3.167151	110	19	1.1	0.9;
	2807	1	28.83	10.4	0	0.26	0	0.987767	-5.2323	110	21	1.1	0.9;
	2808	1	0	-0	0	0.86	0	1.013167	-26.447449	220	4	1.1	0.9;
	2809	1	-0	0	0	0	0	1.021419	-22.29244	20	4	1.1	0.9;
	2810	2	0	0	0	0	0	1.040793	26.940223	20	5	1.1	0.9;
	2811	2	0	0	0	0	0	1.022087	-3.367486	20	6	1.1	0.9;
	2812	1	1.8	4.9	0	92.20984	0	1.117558	-2.51737	380	9	1.2	0.8;
	2813	1	210.1	66.2	0	59.36	0	1.024123	3.471914	220	5	1.1	0.9;
	2814	1	241.9	38.8	0	12.46	0	0.956074	-16.770188	220	10	1.1	0.9;
	2815	1	23.33	9.1	0	15.36	0	1.026048	31.606612	150	8	1.1	0.9;
	2816	2	0	0	0	0	0	0.992395	35.295433	11	1	1.1	0.9;
	2817	1	0	-0	0	12.41	0	1.030487	33.829091	380	1	1.1	0.9;
	2818	1	0	0	0.135493	29.07062	0	1.085877	-0.824855	380	21	1.2	0.8;
	2819	2	0	0	0	0	0	1.031736	41.040722	20	10	1.1	0.9;
	2820	1	-0	0	0	0.14	0	1.046339	-4.535747	220	5	1.1	0.9;
	2821	1	39.64	9.6	0	1.8	0	1.030407	4.051965	220	5	1.1	0.9;
	2822	1	19.32	6.4	0	3.74	0	1.015223	-8.707967	110	12	1.1	0.9;
	2823	1	12.41	3.9	0	1.14	0	1.013354	-0.836591	110	19	1.1	0.9;
	2824	1	41.8	5.9	0	2.76	0	1.010914	19.579521	154	1	1.1	0.9;
	2825	1	188.1	26.8	0	5.85	0	1.033805	-1.24014	154	1	1.1	0.9;
	2826	1	-0	0	0	1.66	0	1.029765	-2.314086	150	9	1.1	0.9;
	2827	1	0	0	0	0	0	1.031122	31.502598	150	8	1.1	0.9;
	2828	2	0	0	0	0	0	1.017638	-6.936102	20	4	1.1	0.9;
	2829	1	147.4	-55.9	0	1.12	0	1.01075	-22.689555	220	4	1.1	0.9;
	2830	1	-0	0	0	1.83	0	1.040594	-6.46768	150	9	1.1	0.9;
	2831	2	0	0	0	0	0	0.989469	40.488685	11.5	1	1.1	0.9;
	2832	1	70.18	44.7	0	8.09	0	0.99781	-5.781696	150	9	1.1	0.9;
	2833	1	50.26	5.5	0	1.3	0	1.000328	-3.366901	110	18	1.1	0.9;
	2834	1	69.78	20.3	0	10.26	0	1.032192	45.944685	110	8	1.1	0.9;
	2835	1	61.87	-12.4	0	0.83	0	0.992019	-3.212205	220	6	1.1	0.9;
	2836	2	0	0	0	0	0	1.061581	6.921348	20	5	1.1	0.9;
	2837	1	-0	-0	0.03986	2.100105	0	1.010638	-8.662097	220	6	1.1	0.9;
	2838	2	0	0	0	0	0	1.064384	12.999659	20	5	1.1	0.9;
	2839	1	89.5	33.9	0	1.52	0	0.995336	33.541224	220	3	1.1	0.9;
	2840	2	0	0	0	0	0	1.027968	24.05088	20	5	1.1	0.9;
	2841	2	0	0	0	0	0	1.027968	24.05088	20	5	1.1	0.9;
	2842	2	0	0	0	0	0	1.061581	6.921348	20	5	1.1	0.9;
	2843	1	-0	-0	0.427339	57.506029	0	1.043628	20.270583	380	3	1.1	0.9;
	2844	2	0	0	0	0	0	1.070991	7.604863	20	5	1.1	0.9;
	2845	2	0	0	0	0	0	1.064384	12.999659	20	5	1.1	0.9;
	2846	1	0	-0	0.121579	0.048816	0	0.988251	-9.325342	220	10	1.1	0.9;
	2847	1	25.73	7.3	0	2.82	0	1.024475	10.258345	220	5	1.1	0.9;
	2848	1	-0	-0	0.113571	-0.085178	0	1.068587	-12.198559	380	10	1.1	0.9;
	2849	1	0	0	0	102.579963	0	1.047582	31.453215	380	5	1.1	0.9;
	2850	1	54	29.1	0	204.291522	0	1.079367	-3.384889	380	9	1.1	0.9;
	2851	1	-0	-0	0	0.4	0	1.005848	0.866704	150	9	1.1	0.9;
	2852	2	0	0	0	0	0	1.013948	12.809021	11	1	1.1	0.9;
	2853	1	0	-0	0	0	0	1.097846	65.011121	380	8	1.2	0.8;
	2854	1	139.1	60.4	0	1.98	0	0.996972	32.605519	220	3	1.1	0.9;
	2855	1	0	-0	0	3.72	0	1.028082	-7.262597	110	22	1.1	0.9;
	2856	1	-61.76	-31.85	0	0	0	1.077186	0.891513	27	9	1.1	0.9;
	2857	1	54.66	0	0	2.97	0	1.013816	4.686922	220	7	1.1	0.9;
	2858	2	0	0	0	0	0	1.044939	29.387061	20	5	1.1	0.9;
	2859	1	-40.44	1.5	0	14.2	0	1.014283	-10.876603	220	10	1.1	0.9;
	2860	1	82.1	11.7	0	3.76	0	1.006131	12.49776	154	1	1.1	0.9;
	2861	1	0	-0	0	0.26	0	0.993839	-3.903185	110	18	1.1	0.9;
	2862	1	-0	0	0.039441	0.080559	0	0.999754	-4.31743	110	18	1.1	0.9;
	2863	1	302.5	65.1	0	8.87	0	1.035883	9.357223	220	5	1.1	0.9;
	2864	1	0	0	0	0	0	0.957345	-9.979812	150	9	1.1	0.9;
	2865	1	-0	-0	0	0.85	0	1.011456	-3.024872	110	18	1.1	0.9;
	2866	1	41.45	4.9	0	18.43	0	1.029821	11.74799	220	5	1.1	0.9;
	2867	1	0	0	0	18.99	0	1.037736	21.469997	220	13	1.1	0.9;
	2868	1	0	-0	0	0.34	0	1.088981	9.07002	380	1	1.2	0.8;
	2869	1	95.91	-15.4	0	1.51	0	1.020944	-1.99455	220	6	1.1	0.9;
	2870	2	0	0	0	0	0	1.030832	-9.752509	20	4	1.1	0.9;
	2871	2	0	0	0	0	0	1.044924	29.406092	20	5	1.1	0.9;
	2872	1	119	-315.3	0	41.89	0	1.030825	34.272754	380	3	1.1	0.9;
	2873	1	0	0	0.29374	2.585217	0	1.027005	1.029856	220	7	1.1	0.9;
	2874	1	-0	0	0.039591	0.130409	0	1.005971	11.672077	110	18	1.1	0.9;
	2875	1	72.08	21.3	0	7.19	0	1.032338	9.184881	220	5	1.1	0.9;
	2876	1	-0	0	0	9.98	0	1.038018	5.797275	220	5	1.1	0.9;
	2877	2	0	0	0	0	0	1.029166	13.522836	20	14	1.1	0.9;
	2878	1	3.9	1	0	3.81	0	1.038131	4.912288	220	5	1.1	0.9;
	2879	1	8.11	2.9	0	0.52	0	0.992496	-1.791255	110	19	1.1	0.9;
	2880	2	0	0	0	0	0	1.028088	-13.031645	20	4	1.1	0.9;
	2881	2	0	0	0	0	0	1.028088	-13.031645	20	4	1.1	0.9;
	2882	1	89.1	2.3	0	1.68	0	0.989155	-22.138997	220	6	1.1	0.9;
	2883	1	16.02	4	0	5.56	0	1.048247	-8.989717	220	11	1.1	0.9;
	2884	1	-0	-0	0.181549	5.390889	0	1.021395	2.988226	220	18	1.1	0.9;
	2885	1	110.9	15.8	0	30.2	0	1.01021	4.316508	400	1	1.1	0.9;
	2886	1	90.1	12.7	0	10.61	0	1.006203	33.698504	220	3	1.1	0.9;
	2887	1	0	0	0.330327	-0.256921	0	1.071397	40.134574	150	8	1.1	0.9;
	2888	1	25.83	9.4	0	0.56	0	1.024582	0.545384	110	19	1.1	0.9;
	2889	1	134	19.1	0	1.55	0	1.0227	0.37135	154	1	1.1	0.9;
	2890	1	0	-0	0	0.66	0	1.020294	26.013319	220	3	1.1	0.9;
	2891	1	153.64	29.2	0	7.65	0	1.001933	0.520394	220	6	1.1	0.9;
	2892	1	-0	0	0.575637	1.661442	0	1.034652	32.708036	380	5	1.1	0.9;
	2893	1	6.31	3.8	0	0.36	0	1.008604	-3.447075	110	18	1.1	0.9;
	2894	2	0	0	0	0	0	1.045516	41.754111	20	10	1.1	0.9;
	2895	2	0	0	0	0	0	1.126501	-1.915268	20	9	1.2	0.8;
	2896	1	28.83	9.6	0	0.8	0	0.976336	-9.661402	110	21	1.1	0.9;
	2897	1	0	0	0	14.4	0	1.036344	43.587167	380	3	1.1	0.9;
	2898	1	297.7	258.6	0	65.78	0	1.036237	19.120016	380	6	1.1	0.9;
	2899	1	178.4	69.1	0	14.2	0	1.005957	-18.039574	220	6	1.1	0.9;
	2900	1	-0	-0	0.32223	108.252095	0	1.0554	5.332672	380	19	1.1	0.9;
	2901	2	0	0	0	0	0	1.039232	7.281291	20	6	1.1	0.9;
	2902	2	0	0	0	0	0	1.039232	7.281291	20	6	1.1	0.9;
	2903	1	3.2	1.7	0	0.05	0	1.023458	8.422084	220	5	1.1	0.9;
	2904	1	8.51	3.2	0	0.96	0	1.016396	18.8779	220	5	1.1	0.9;
	2905	1	0	0	0	0.74	0	0.969903	-8.40306	150	9	1.1	0.9;
	2906	2	0	0	0	0	0	1.026436	2.785646	11	1	1.1	0.9;
	2907	1	83.59	21.6	0	25.51	0	1.014613	-26.808033	220	4	1.1	0.9;
	2908	1	-0.65	-1.44	0	0.16	0	1.039793	-6.423412	150	9	1.1	0.9;
	2909	1	19.22	6.9	0	1.16	0	1.085162	1.341951	110	15	1.1	0.9;
	2910	1	29.3	4.2	0	5.26	0	0.996191	18.211728	154	1	1.1	0.9;
	2911	1	39.84	24.6	0	13.99	0	1.004491	-12.436802	150	9	1.1	0.9;
	2912	2	0	0	0	0	0	1.020176	20.468258	11.5	1	1.1	0.9;
	2913	1	0	-0	0.154629	1.135371	0	1.051252	7.897462	380	16	1.1	0.9;
	2914	1	0	0	0	5.39	0	1.019441	-3.771239	220	12	1.1	0.9;
	2915	1	0	0	0	0	0	0.965144	-10.373275	150	9	1.1	0.9;
	2916	1	0	0	0	2.99	0	1.021714	-0.968124	220	1	1.1	0.9;
	2917	1	-0	-0	0	0.06	0	1.04687	5.192474	220	7	1.1	0.9;
	2918	1	52.66	21.5	0	0.04	0	1.019459	22.105369	150	8	1.1	0.9;
	2919	1	0	0	0	7.196658	0	0.981514	20.291106	380	5	1.1	0.9;
	2920	1	120.2	20.5	0	19.94	0	0.989503	-3.363256	220	5	1.1	0.9;
	2921	1	-0	0	0	1.31	0	1.043272	-18.000164	380	10	1.1	0.9;
	2922	1	-0	-0	0	16.08	0	1.111192	0.559337	380	9	1.2	0.8;
	2923	1	115.1	23.5	0	3.78	0	1.018011	-30.733354	220	4	1.1	0.9;
	2924	1	0	-0	0	0.31	0	0.978502	5.404407	110	18	1.1	0.9;
	2925	1	-0.61	-1.33	0	0.43	0	1.054212	-0.245142	110	15	1.1	0.9;
	2926	1	108.7	10.3	0.087306	1.402394	0	1.01782	13.626458	220	13	1.1	0.9;
	2927	1	14.4	2	0	4.17	0	0.995079	17.378652	154	1	1.1	0.9;
	2928	1	98.6	14	0	12.86	0	1.001808	17.707788	154	1	1.1	0.9;
	2929	1	19.82	7.3	0	0.25	0	1.008479	-10.420579	110	22	1.1	0.9;
	2930	1	0	0	0	0	0	1.043805	-6.740275	150	9	1.1	0.9;
	2931	1	50.16	18	0	0.1	0	1.051506	0.116562	110	19	1.1	0.9;
	2932	1	1.4	-2	0	5.05	0	1.022012	92.365113	150	1	1.1	0.9;
	2933	2	0	0	0	0	0	0.989831	32.095099	6.3	1	1.1	0.9;
	2934	2	0	0	0	0	0	1.033975	-14.23689	20	6	1.1	0.9;
	2935	1	-0	-0	0	2.88	0	1.040896	75.624353	220	3	1.1	0.9;
	2936	1	0	-0	0	149.047157	0	1.026327	21.205564	380	5	1.1	0.9;
	2937	1	101.2	2.1	0	38.43	0	1.0563	11.740795	380	5	1.1	0.9;
	2938	1	-0	0	0.060224	0.419776	0	1.025389	30.951659	220	14	1.1	0.9;
	2939	1	86.9	30.3	0	1.32	0	1.041371	-2.347999	110	15	1.1	0.9;
	2940	1	50.56	4.6	0	1.74	0	1.019671	-32.938285	220	4	1.1	0.9;
	2941	1	32.54	10.6	0	1.22	0	0.909658	-12.472146	110	21	1.1	0.9;
	2942	1	-0	-0	0	0.23	0	0.991779	-19.092039	220	6	1.1	0.9;
	2943	1	88.3	2.8	0	5.21	0	1.022404	-12.286706	220	4	1.1	0.9;
	2944	1	7	1	0	1.54	0	0.994725	33.286745	154	1	1.1	0.9;
	2945	1	10.21	-3.2	0	5.15	0	1.042125	-13.973553	220	4	1.1	0.9;
	2946	1	125.1	44	0	6.04	0	0.978089	13.519151	220	5	1.1	0.9;
	2947	1	-0	0	0	0.29	0	1.000392	-8.664731	150	9	1.1	0.9;
	2948	1	163.7	94.1	0	18.55	0	1.006167	19.203386	380	3	1.1	0.9;
	2949	1	30.93	2.8	0	2.11	0	1.030428	-21.332673	220	4	1.1	0.9;
	2950	2	0	0	0	0	0	1.006699	27.063644	154	1	1.1	0.9;
	2951	1	121.4	20.2	0	16.61	0	1.032138	11.548616	220	5	1.1	0.9;
	2952	1	-0	-0	0	6.01	0	1.059984	46.329026	380	14	1.1	0.9;
	2953	1	-0	-0	0.096544	0.473456	0	1.04137	33.25698	380	5	1.1	0.9;
	2954	1	0	0	0.037641	3.052359	0	1.052423	4.564159	220	18	1.1	0.9;
	2955	2	0	0	0	0	0	1.054489	6.09839	20	20	1.1	0.9;
	2956	1	-0	0	0.126829	38.306912	0	1.015553	5.973983	380	7	1.1	0.9;
	2957	1	199.86	56.93	0	46.83	0	1.030714	4.086976	220	5	1.1	0.9;
	2958	1	218.7	-9.1	0	49	0	1.011483	-11.28309	220	5	1.1	0.9;
	2959	1	-0	-0	0.019608	2.610392	0	1.016528	-8.99771	220	4	1.1	0.9;
	2960	1	110.09	53.32	0	0	0	0.966387	0.261157	110	18	1.1	0.9;
	2961	1	-5.64	-2.84	0	34.641229	0	1.021332	21.250094	380	5	1.1	0.9;
	2962	1	65.87	31.4	0	3.46	0	1.116078	70.148814	220	3	1.2	0.8;
	2963	2	0	0	0	0	0	1.035406	21.74444	20	5	1.1	0.9;
	2964	2	0	0	0	0	0	1.036949	12.275543	20	7	1.1	0.9;
	2965	1	3.3	313.7	0.705985	162.230901	0	1.018688	-18.692878	380	4	1.1	0.9;
	2966	1	54.26	19.5	0	0.05	0	1.016599	6.103548	220	7	1.1	0.9;
	2967	1	-0	-0	0.117169	2.552359	0	1.053421	28.464675	380	14	1.1	0.9;
	2968	2	0	0	0	0	0	1.03481	9.302046	20	18	1.1	0.9;
	2969	2	0	0	0	0	0	1.03481	9.302046	20	18	1.1	0.9;
	2970	2	0	0	0	0	0	1.030988	2.394145	20	9	1.1	0.9;
	2971	1	-3.7	-8	0	19.02	0	1.01275	1.829725	380	7	1.1	0.9;
	2972	1	-0	0	0	5.19	0	1.044543	-2.917968	220	23	1.1	0.9;
	2973	2	0	0	0	0	0	1.03481	9.302046	20	18	1.1	0.9;
	2974	2	0	0	0	0	0	1.03481	9.302046	20	18	1.1	0.9;
	2975	1	-0	0	0	38.32	0	1.006502	23.545534	380	5	1.1	0.9;
	2976	2	0	0	0	0	0	1.00657	25.464337	20	24	1.1	0.9;
	2977	1	-0	-0	0	3.95	0	0.967193	48.296353	220	3	1.1	0.9;
	2978	2	0	0	0	0	0	1.027523	-21.832316	20	6	1.1	0.9;
	2979	2	0	0	0	0	0	1.027523	-21.832316	20	6	1.1	0.9;
	2980	2	0	0	0	0	0	1.027523	-21.832316	20	6	1.1	0.9;
	2981	2	0	0	0	0	0	1.024756	1.450408	20	6	1.1	0.9;
	2982	2	0	0	0	0	0	1.024756	1.450408	20	6	1.1	0.9;
	2983	1	0	0	0	6.22	0	1.043072	93.88678	380	1	1.1	0.9;
	2984	2	0	0	0	0	0	1.024833	-17.429341	20	6	1.1	0.9;
	2985	1	-0	-0	0.865147	27.25011	0	1.030141	39.290649	380	2	1.1	0.9;
	2986	1	0	-0	0.090774	10.830444	0	1.047584	-0.741816	220	18	1.1	0.9;
	2987	1	529.9	23.3	0	57.66	0	1.014067	-13.863803	380	6	1.1	0.9;
	2988	1	10.01	3	0	0.84	0	0.92396	-11.623321	110	21	1.1	0.9;
	2989	1	12.81	4.4	0	0.47	0	1.008539	3.062301	110	18	1.1	0.9;
	2990	2	0	0	0	0	0	1.016588	24.927519	10.5	1	1.1	0.9;
	2991	2	0	0	0	0	0	1.032132	-3.910762	20	6	1.1	0.9;
	2992	1	60.67	37.6	0	0.92	0	1.012722	-8.876216	150	9	1.1	0.9;
	2993	1	-0	0	0.207683	8.945585	0	1.127982	-0.363224	380	9	1.2	0.8;
	2994	2	0	0	0	0	0	0.985642	34.609787	11	1	1.1	0.9;
	2995	1	5.71	2	0	1.03	0	0.950784	-0.403765	110	18	1.1	0.9;
	2996	1	11	3.15	0	0.63	0	0.948638	-1.282439	110	18	1.1	0.9;
	2997	1	53.36	7.1	0	6.02	0	1.010472	-18.912844	220	4	1.1	0.9;
	2998	1	-0.2	-1	0	0	0	1.018613	25.307234	150	8	1.1	0.9;
	2999	1	0	-0	0	3.61	0	0.991684	-11.480713	220	1	1.1	0.9;
	3000	1	4	0.6	0	3.46	0	0.996058	20.763579	154	1	1.1	0.9;
	3001	1	-0.19	-3.76	0	0	0	1.028103	35.299364	150	8	1.1	0.9;
	3002	1	4.4	2.2	0	1.79	0	0.993999	-13.385176	110	22	1.1	0.9;
	3003	1	-0	0	0	2.33	0	1.00749	31.832433	154	1	1.1	0.9;
	3004	1	19.82	6	0	1.68	0	0.961256	-8.218502	110	21	1.1	0.9;
	3005	1	8.07	1.23	0	3.91	0	1.014074	90.565376	150	1	1.1	0.9;
	3006	1	-0	0	0	1.81	0	0.993474	-9.319885	150	9	1.1	0.9;
	3007	1	92.3	14.6	0	1.14	0	1.013906	-25.46217	220	4	1.1	0.9;
	3008	1	30.33	11	0	2.29	0	1.039582	1.171175	110	19	1.1	0.9;
	3009	1	79.09	18.7	0	10.8	0	1.028638	7.443496	220	5	1.1	0.9;
	3010	1	-0	0	0	56.686515	0	0.99539	20.289767	380	5	1.1	0.9;
	3011	1	0	0	0	0	0	1.050792	2.691479	20	2	1.1	0.9;
	3012	2	0	0	0	0	0	1.024412	23.818834	20	2	1.1	0.9;
	3013	2	0	0	0	0	0	1.041662	8.678673	20	2	1.1	0.9;
	3014	1	16.22	4	0	0	0	1.0963	-4.854397	150	9	1.1	0.9;
	3015	2	0	0	0	0	0	1.033334	3.429229	20	2	1.1	0.9;
	3016	1	8.91	8.9	0	0.58	0	0.954853	-4.912864	110	18	1.1	0.9;
	3017	2	0	0	0	0	0	1.014104	26.485556	20	6	1.1	0.9;
	3018	1	107.3	14.2	0	20.23	0	1.026992	13.603011	220	5	1.1	0.9;
	3019	1	-0	0	0	11.63	0	1.060003	79.42593	380	3	1.1	0.9;
	3020	1	0	-0	0	25.72	0	1.053608	28.092308	220	1	1.1	0.9;
	3021	2	0	0	0	0	0	1.014104	26.485556	20	6	1.1	0.9;
	3022	2	0	0	0	0	0	1.014104	26.485556	20	6	1.1	0.9;
	3023	1	78.69	56.7	0	10.47	0	1.001872	0.589968	220	6	1.1	0.9;
	3024	1	23.2	3.3	0	3	0	1.005994	30.332283	154	1	1.1	0.9;
	3025	1	-0	-0	0	4.52	0	1.019586	29.439072	380	3	1.1	0.9;
	3026	1	137.8	64.5	0	18.2	0	1.009419	16.230861	380	3	1.1	0.9;
	3027	1	4.69	1.89	0	2.45	0	1.012463	-5.993975	150	9	1.1	0.9;
	3028	1	101.5	-18.3	0	8.47	0	1.031926	9.191503	220	5	1.1	0.9;
	3029	2	0	0	0	0	0	1.043665	14.930439	20	14	1.1	0.9;
	3030	1	-0	-0	0	1.2	0	0.974652	-25.839438	150	10	1.1	0.9;
	3031	2	0	0	0	0	0	1.018242	17.423399	20	10	1.1	0.9;
	3032	1	20.02	6.8	0.01945	0.80055	0	1.0352	-3.530694	110	19	1.1	0.9;
	3033	1	29.4	4.2	0	5.52	0	1.004662	26.192512	154	1	1.1	0.9;
	3034	1	-0	0	0	7.63	0	1.033962	21.349108	220	5	1.1	0.9;
	3035	1	129.5	38.3	0	5.42	0	0.987509	-15.826529	220	4	1.1	0.9;
	3036	1	-0	-0	0	0.44	0	1.041806	4.567962	220	20	1.1	0.9;
	3037	1	126.98	12.83	0	9.13	0	1.019828	21.769526	220	13	1.1	0.9;
	3038	1	130.9	37.8	0	0.27	0	1.015717	14.479145	220	3	1.1	0.9;
	3039	1	6.11	0	0	0.92	0	1.012677	-20.123731	220	4	1.1	0.9;
	3040	1	0	-0	0	0.3	0	1.049836	-6.49771	150	9	1.1	0.9;
	3041	1	17.62	5.8	0	0.48	0	0.920298	-11.220957	110	21	1.1	0.9;
	3042	1	101.9	24.3	0	8.53	0	1.032426	18.042041	220	5	1.1	0.9;
	3043	1	439	254.8	0	148.04	0	1.07085	72.829731	380	3	1.1	0.9;
	3044	1	0	0	0.338538	-0.263307	0	1.067115	48.55262	380	8	1.1	0.9;
	3045	1	0	0	0	116.62	0	0.963367	24.301846	400	1	1.1	0.9;
	3046	1	-2.23	-1.27	0	3.64	0	1.027641	-16.751919	220	4	1.1	0.9;
	3047	1	-0	0	0.181549	4.450889	0	1.047872	-1.209101	220	18	1.1	0.9;
	3048	1	0	-0	0	1.01	0	1.085973	-3.616211	150	9	1.1	0.9;
	3049	1	0	0	0.29	13.47	0	1.034082	31.667909	380	2	1.1	0.9;
	3050	2	0	0	0	0	0	1.033768	14.596978	20	14	1.1	0.9;
	3051	1	6.01	2.2	0	1.91	0	0.977685	10.299801	110	18	1.1	0.9;
	3052	1	-0	0	0	3.6	0	1.020565	30.713834	380	3	1.1	0.9;
	3053	1	-0	0	0	1.12	0	1.003915	-5.029519	150	9	1.1	0.9;
	3054	1	51.96	32.2	0	0.84	0	0.834005	-17.59375	150	9	1.2	0.8;
	3055	1	304.2	114.1	0	26.84	0	1.011139	-15.286107	380	6	1.1	0.9;
	3056	1	-0	-0	0	1.02	0	1.081964	13.95024	220	5	1.1	0.9;
	3057	1	12.21	4.4	0	1.03	0	1.072158	0.502654	110	15	1.1	0.9;
	3058	2	0	0	0	0	0	1.021353	-16.057707	20	6	1.1	0.9;
	3059	1	0	-0	0	0	0	1.010498	-17.064676	20	4	1.1	0.9;
	3060	1	6.61	2.4	0.019921	0.740079	0	1.008604	-1.974044	110	19	1.1	0.9;
	3061	2	0	0	0	0	0	1.041842	41.954443	20	3	1.1	0.9;
	3062	2	0	0	0	0	0	1.041842	41.954443	20	3	1.1	0.9;
	3063	2	0	0	0	0	0	1.05544	46.172493	20	3	1.1	0.9;
	3064	1	15.82	9.8	0	4.86	0	1.051779	-6.398929	150	9	1.1	0.9;
	3065	2	0	0	0	0	0	1.048637	18.752224	20	5	1.1	0.9;
	3066	1	0	-0	0	10.34	0	0.983472	31.050625	400	1	1.1	0.9;
	3067	1	20.92	7.7	0	1.21	0	1.042065	14.434339	220	5	1.1	0.9;
	3068	2	0	0	0	0	0	1.043939	36.504273	20	5	1.1	0.9;
	3069	1	29.93	9.9	0.117792	0.22184	0	1.023764	-3.675899	110	19	1.1	0.9;
	3070	1	8.6	-1.48	0	0	0	1.023596	21.694368	150	2	1.1	0.9;
	3071	1	120.1	30.5	0	5.26	0	1.0132	-13.74777	220	17	1.1	0.9;
	3072	2	0	0	0	0	0	1.023058	6.860658	20	6	1.1	0.9;
	3073	1	-0	0	0	1.33	0	1.01869	3.847783	220	1	1.1	0.9;
	3074	2	0	0	0	0	0	1.023058	6.860658	20	6	1.1	0.9;
	3075	1	37.6	5.4	0	2.24	0	1.030804	-0.760296	154	1	1.1	0.9;
	3076	1	114.2	34.4	0	10.29	0	1.019117	-22.580163	220	4	1.1	0.9;
	3077	1	396.7	62.8	0	0	0	1.062348	-1.257926	120	20	1.1	0.9;
	3078	1	-0	-0	0	0	0	1.002012	13.542185	20	5	1.1	0.9;
	3079	1	106.4	24.2	0	0.11	0	1.020527	-4.053254	110	19	1.1	0.9;
	3080	1	95.1	35	0.040461	2.159654	0	0.983365	-23.074447	220	6	1.1	0.9;
	3081	1	-1397.75	-894.7	0.416616	123.667654	0	1.028955	49.734107	380	3	1.1	0.9;
	3082	2	0	0	0	0	0	0.959658	41.263569	10.6	1	1.1	0.9;
	3083	1	171.1	49.1	0.058585	15.741415	0	1.068316	1.198768	220	5	1.1	0.9;
	3084	1	40.65	13.2	0	1.27	0	1.008605	-6.77169	110	22	1.1	0.9;
	3085	1	50.86	-23	0	16.36	0	1.028133	8.275168	220	3	1.1	0.9;
	3086	1	-0	-0	0	0.3	0	1.027606	28.817047	220	5	1.1	0.9;
	3087	1	-15.8	-2.8	0	7.6	0	1.066871	-5.364063	150	9	1.1	0.9;
	3088	1	-0	-0	0	69.27	0	0.948425	51.609641	400	1	1.1	0.9;
	3089	1	0	0	0	0.04	0	0.996361	-17.3888	220	10	1.1	0.9;
	3090	1	-0	-0	0	32.802252	0	0.989513	14.286196	380	5	1.1	0.9;
	3091	1	-0	-0	0	1.29	0	1.00244	-9.914239	150	9	1.1	0.9;
	3092	1	0	0	0.056531	4.133469	0	1.054244	8.966818	220	18	1.2	0.8;
	3093	2	0	0	0	0	0	1.062094	84.081646	20	3	1.1	0.9;
	3094	2	0	0	0	0	0	1.048559	90.431059	20	3	1.1	0.9;
	3095	2	0	0	0	0	0	1.050106	73.063142	20	3	1.1	0.9;
	3096	2	0	0	0	0	0	1.048559	90.431059	20	3	1.1	0.9;
	3097	2	0	0	0	0	0	1.050106	73.063142	20	3	1.1	0.9;
	3098	2	0	0	0	0	0	1.050106	73.063142	20	3	1.1	0.9;
	3099	1	14.22	6.3	0	0.38	0	1.034913	19.583318	220	5	1.1	0.9;
	3100	2	0	0	0	0	0	1.062073	84.08696	20	3	1.1	0.9;
	3101	2	0	0	0	0	0	1.062094	84.081646	20	3	1.1	0.9;
	3102	1	156.5	31.2	0.536484	34.842734	0	1.02653	21.638239	380	6	1.1	0.9;
	3103	2	0	0	0	0	0	1.040259	34.3106	20	5	1.1	0.9;
	3104	2	0	0	0	0	0	1.040259	34.3106	20	5	1.1	0.9;
	3105	1	-0	0	0	0.18	0	1.015795	22.756039	220	5	1.1	0.9;
	3106	1	-14.27	-6.6	0	0.91	0	1.060487	48.408157	380	3	1.1	0.9;
	3107	2	0	0	0	0	0	1.023574	3.735821	20	2	1.1	0.9;
	3108	2	0	0	0	0	0	1.023574	3.735821	20	2	1.1	0.9;
	3109	2	0	0	0	0	0	1.0234	3.527787	20	2	1.1	0.9;
	3110	1	0	-0	0	9.9	0	1.026902	3.580533	220	7	1.1	0.9;
	3111	2	0	0	0	0	0	1.0234	3.527787	20	2	1.1	0.9;
	3112	2	0	0	0	0	0	1.030367	0.220716	20	2	1.1	0.9;
	3113	2	0	0	0	0	0	1.030443	-1.752265	20	2	1.1	0.9;
	3114	2	0	0	0	0	0	1.00571	1.688469	20	2	1.1	0.9;
	3115	1	-19.79	-21.11	0.153249	65.355488	0	1.096855	60.617	380	3	1.2	0.8;
	3116	1	-0	-0	0	5.95	0	1.009609	-19.661174	220	4	1.1	0.9;
	3117	1	-0	0	0	23.66	0	0.99314	-9.90363	150	9	1.1	0.9;
	3118	1	55.46	18.5	0	2.23	0	0.963262	-9.644491	110	21	1.1	0.9;
	3119	1	0	0	0	5.14	0	1.035147	65.857496	220	3	1.1	0.9;
	3120	1	7.29	-0.73	0	0.73	0	1.034839	0.308162	110	19	1.1	0.9;
	3121	2	0	0	0	0	0	1.057682	40.579523	20	3	1.1	0.9;
	3122	2	0	0	0	0	0	1.057682	40.579523	20	3	1.1	0.9;
	3123	2	0	0	0	0	0	1.050841	90.684517	20	8	1.1	0.9;
	3124	2	0	0	0	0	0	1.050841	90.684517	20	8	1.1	0.9;
	3125	1	34.14	11.4	0	0.8	0	0.960854	-4.989884	110	18	1.1	0.9;
	3126	1	297.2	93.2	0	23.74	0	0.990358	-22.741184	380	6	1.1	0.9;
	3127	2	0	0	0	0	0	1.050579	35.320834	20	8	1.1	0.9;
	3128	2	0	0	0	0	0	1.045612	86.470349	20	8	1.1	0.9;
	3129	1	137.3	-2.8	0	16.39	0	1.005101	0.818924	220	6	1.1	0.9;
	3130	2	0	0	0	0	0	1.045612	86.470349	20	8	1.1	0.9;
	3131	1	29.03	9.6	0	0.53	0	0.948542	-8.803724	110	21	1.1	0.9;
	3132	1	213	39.6	0	20.63	0	1.012469	-22.467823	380	4	1.1	0.9;
	3133	1	-0	-0	0.352533	27.156463	0	0.972032	7.622339	380	11	1.1	0.9;
	3134	2	0	0	0	0	0	1.055592	83.040122	20	3	1.1	0.9;
	3135	2	0	0	0	0	0	1.055592	83.040122	20	3	1.1	0.9;
	3136	2	0	0	0	0	0	1.055592	83.040122	20	3	1.1	0.9;
	3137	2	0	0	0	0	0	1.055592	83.040122	20	3	1.1	0.9;
	3138	1	11.01	3.6	0	0.84	0	0.975876	-11.404679	110	12	1.1	0.9;
	3139	2	0	0	0	0	0	1.033768	14.596978	20	14	1.1	0.9;
	3140	1	-0	-0	0.177225	2.972313	0	1.013004	-6.672823	150	10	1.1	0.9;
	3141	2	0	0	0	0	0	1.03603	78.293887	20	3	1.1	0.9;
	3142	2	0	0	0	0	0	1.03603	78.293887	20	3	1.1	0.9;
	3143	2	0	0	0	0	0	1.034933	81.348467	20	3	1.1	0.9;
	3144	1	17.52	10.4	0.039249	-0.019249	0	0.996071	-7.239259	110	18	1.1	0.9;
	3145	2	0	0	0	0	0	1.034547	81.296442	20	3	1.1	0.9;
	3146	2	0	0	0	0	0	1.028604	-12.690591	20	4	1.1	0.9;
	3147	1	74.98	26.9	0	2.96	0	1.021721	15.54218	220	5	1.1	0.9;
	3148	2	0	0	0	0	0	1.017868	-15.794755	20	4	1.1	0.9;
	3149	1	6.01	4.1	0	0.19	0	0.964033	-6.241528	110	18	1.1	0.9;
	3150	1	-0	-7.5	0	0	0	1.004597	-15.207244	150	9	1.1	0.9;
	3151	1	44.55	16.1	0	1.18	0	0.923941	-1.894529	110	19	1.1	0.9;
	3152	2	0	0	0	0	0	1.05813	5.10364	20	18	1.1	0.9;
	3153	1	60.07	45	0	2.53	0	0.986936	36.296305	220	3	1.1	0.9;
	3154	2	0	0	0	0	0	1.041223	4.865379	20	18	1.1	0.9;
	3155	1	0	0	0	0.3	0	1.001884	-10.143098	150	9	1.1	0.9;
	3156	1	-0	0	0.17315	3.835328	0	1.012113	93.904214	380	1	1.1	0.9;
	3157	1	0	0	0	0	0	0.947288	-10.924648	150	9	1.1	0.9;
	3158	1	23.7	3.4	0	2.51	0	1.00507	17.0212	154	1	1.1	0.9;
	3159	1	-0	-0	0	47.343052	0	1.064147	34.611639	380	13	1.1	0.9;
	3160	1	3.9	0.9	0	0.3	0	0.98633	-2.201471	110	18	1.1	0.9;
	3161	1	0	-0	0	0.41	0	0.963134	-9.751258	150	9	1.1	0.9;
	3162	1	0	-0	0	1.6	0	0.994543	-3.850432	110	21	1.1	0.9;
	3163	1	20.32	6.7	0	0.27	0	0.983836	-11.332269	110	12	1.1	0.9;
	3164	2	0	0	0	0	0	1.01814	-7.851847	20	4	1.1	0.9;
	3165	2	0	0	0	0	0	1.029459	27.896063	20	10	1.1	0.9;
	3166	2	0	0	0	0	0	1.014154	-7.89767	20	10	1.1	0.9;
	3167	2	0	0	0	0	0	1.022861	4.217398	20	6	1.1	0.9;
	3168	2	0	0	0	0	0	1.022861	4.217398	20	6	1.1	0.9;
	3169	2	0	0	0	0	0	1.022861	4.217398	20	6	1.1	0.9;
	3170	1	0	-0	0	2.67	0	1.065913	74.319622	380	3	1.1	0.9;
	3171	1	142.8	19.5	0	10.53	0	1.010695	28.760676	220	3	1.1	0.9;
	3172	1	78.79	27.4	0	2.37	0	0.991482	-25.02317	220	6	1.1	0.9;
	3173	1	88.7	25.5	0	3.07	0	1.022913	-16.933536	220	4	1.1	0.9;
	3174	1	128.7	25.6	0	1.32	0	1.031132	-7.749155	110	1	1.1	0.9;
	3175	1	0	0	0	0	0	1.006756	-15.1197	150	9	1.1	0.9;
	3176	2	0	0	0	0	0	1.03058	15.553508	20	14	1.1	0.9;
	3177	1	143.1	18.9	0	3.42	0	1.020476	55.380688	110	8	1.1	0.9;
	3178	2	0	0	0	0	0	0.962508	28.579218	15	1	1.1	0.9;
	3179	2	0	0	0	0	0	0.969605	29.504132	19	1	1.1	0.9;
	3180	1	-0	0	0	2.91	0	1.051606	16.05419	220	5	1.1	0.9;
	3181	1	-0	0	0	1.71	0	1.021283	7.910363	220	5	1.1	0.9;
	3182	1	-0	0	0	0.45	0	1.036145	30.869082	220	5	1.1	0.9;
	3183	1	0	-0	0.148489	25.143031	0	1.033937	31.845751	380	2	1.1	0.9;
	3184	2	0	0	0	0	0	0.993105	34.379244	10.6	1	1.1	0.9;
	3185	1	-0	-0	0	1.68	0	1.064935	-7.972259	220	5	1.1	0.9;
	3186	2	0	0	0	0	0	1.048808	68.937466	20	3	1.1	0.9;
	3187	1	0	0	0.927753	28.525838	0	1.032021	3.020299	380	6	1.1	0.9;
	3188	1	63.67	39.5	0	0.38	0	1.06045	-6.223335	150	9	1.1	0.9;
	3189	1	125.5	22.5	0	9.17	0	1.024334	-19.631844	220	4	1.1	0.9;
	3190	1	13.9	1.9	0	0.62	0	1.009114	34.555955	154	1	1.1	0.9;
	3191	1	43.75	16.7	0	2.83	0	1.026175	-15.073177	220	4	1.1	0.9;
	3192	1	10.71	17.7	0	10.58	0	1.023187	36.942852	150	8	1.1	0.9;
	3193	1	0	-0	0.115814	4.423488	0	1.043327	79.313508	220	3	1.1	0.9;
	3194	1	65.77	27.3	0	0.49	0	1.042014	29.50224	220	5	1.1	0.9;
	3195	1	0	-0	0	0.57	0	0.969914	7.06577	110	18	1.1	0.9;
	3196	1	69.08	24	0	4.79	0	0.99471	-3.303459	220	6	1.1	0.9;
	3197	1	66.87	21.8	0	12.3	0	0.996125	26.875266	150	8	1.1	0.9;
	3198	1	-0	-0	0	144.15	0	0.966153	37.549398	400	1	1.1	0.9;
	3199	1	0	-0	0	2.76	0	1.058579	11.117197	220	5	1.1	0.9;
	3200	1	55.6	7.9	0	5.4	0	1.013174	29.812789	154	1	1.1	0.9;
	3201	2	0	0	0	0	0	1.034282	8.113127	11	1	1.1	0.9;
	3202	1	-0	-0	0	0.24	0	1.039109	-1.110008	150	9	1.1	0.9;
	3203	1	33.5	4.8	0	6.37	0	0.989027	33.433286	154	1	1.1	0.9;
	3204	1	47.46	11.54	0	14.15	0	0.992836	90.007676	150	1	1.1	0.9;
	3205	1	110.3	28.7	0	0.65	0	1.021971	-18.847648	220	4	1.1	0.9;
	3206	2	0	0	0	0	0	1.018631	-11.528817	20	4	1.1	0.9;
	3207	1	1.8	0.7	0	0.22	0	1.08011	1.741777	110	19	1.2	0.8;
	3208	1	162.9	93.6	0.040274	1.169794	0	0.982791	19.076224	220	13	1.1	0.9;
	3209	1	-5.15	-16.03	0	0	0	1.040256	43.019104	150	8	1.1	0.9;
	3210	2	0	0	0	0	0	1.016876	22.252733	20	6	1.1	0.9;
	3211	1	100.7	-4.2	0	1.57	0	1.024174	-20.53005	220	5	1.1	0.9;
	3212	1	0	0.01	0.434555	9.815977	0	1.114099	0.152522	380	21	1.2	0.8;
	3213	2	0	0	0	0	0	1.016868	22.255371	20	6	1.1	0.9;
	3214	2	0	0	0	0	0	1.016876	22.252733	20	6	1.1	0.9;
	3215	2	0	0	0	0	0	1.016876	22.252733	20	6	1.1	0.9;
	3216	2	0	0	0	0	0	1.016876	22.252733	20	6	1.1	0.9;
	3217	2	0	0	0	0	0	1.063297	57.37246	20	3	1.1	0.9;
	3218	1	-0	-8	0	0	0	1.001825	-12.546618	150	9	1.1	0.9;
	3219	2	0	0	0	0	0	1.016868	22.255371	20	6	1.1	0.9;
	3220	2	0	0	0	0	0	1.016868	22.255371	20	6	1.1	0.9;
	3221	1	23.03	8.3	0	1.14	0	0.957465	-0.98431	110	19	1.1	0.9;
	3222	1	97	18.7	0	21.46	0	1.009445	4.392906	220	7	1.1	0.9;
	3223	1	0	-0	0	29.88	0	1.069984	-1.843245	380	1	1.2	0.8;
	3224	2	0	0	0	0	0	1.034075	49.875441	20	10	1.1	0.9;
	3225	2	0	0	0	0	0	1.034314	-10.021136	20	4	1.1	0.9;
	3226	1	-0	0	0	0	0	1.03499	-19.617451	20	5	1.1	0.9;
	3227	1	0	0	0	7.75	0	1.002095	34.235111	220	3	1.1	0.9;
	3228	2	0	0	0	0	0	1.027518	-17.089784	20	4	1.1	0.9;
	3229	1	305.6	207.3	0	52.37	0	1.014562	32.096227	380	3	1.1	0.9;
	3230	1	0	-0	0	20.5	0	1.055461	-13.247245	380	4	1.1	0.9;
	3231	1	-0	0	0	1.55	0	0.981328	-8.077724	150	9	1.1	0.9;
	3232	1	1.9	0.7	0	0.42	0	0.961557	-5.026887	110	19	1.1	0.9;
	3233	2	0	0	0	0	0	1.034314	-10.021136	20	4	1.1	0.9;
	3234	2	0	0	0	0	0	1.014559	-1.055158	20	18	1.1	0.9;
	3235	1	27.9	4	0	2.16	0	0.996619	17.584415	154	1	1.1	0.9;
	3236	2	0	0	0	0	0	1.000449	37.338871	13.8	1	1.1	0.9;
	3237	1	-0	-0	0	0.39	0	1.025252	2.340914	154	1	1.1	0.9;
	3238	1	0	-0	0.385832	0.591127	0	1.040701	35.174448	380	2	1.1	0.9;
	3239	1	-0	0	0.197033	22.516441	0	0.985476	31.316276	380	2	1.1	0.9;
	3240	1	-0	0	0	2.59	0	1.054736	8.161866	220	1	1.1	0.9;
	3241	1	0	0	0	0.33	0	1.068493	1.28496	150	9	1.1	0.9;
	3242	1	-0	0	0	0	0	1.03499	-19.617451	20	5	1.1	0.9;
	3243	1	73.68	15.4	0	3.73	0	1.012445	4.699137	220	7	1.1	0.9;
	3244	1	-0	0	0	26.22	0	1.051753	14.759663	380	1	1.1	0.9;
	3245	1	-0	0	0	0	0	1.03499	-19.617451	20	5	1.1	0.9;
	3246	1	141.3	32.4	0	7.29	0	1.023718	3.428847	220	5	1.1	0.9;
	3247	1	34.04	21.1	0	0	0	0.993427	-9.324273	150	9	1.1	0.9;
	3248	2	0	0	0	0	0	1.037973	11.471538	20	14	1.1	0.9;
	3249	1	25.33	8.9	0	5.26	0	1.009126	-1.40707	110	18	1.1	0.9;
	3250	1	91.9	5.5	0	7.02	0	1.0082	-1.533549	110	18	1.1	0.9;
	3251	1	0	0	0.058997	1.431003	0	1.054383	5.160988	220	5	1.1	0.9;
	3252	1	-0	-0	0	15.57	0	1.032515	74.532787	380	3	1.1	0.9;
	3253	1	-0	-0	0	2.08	0	0.999445	44.319644	154	1	1.1	0.9;
	3254	1	73.18	22.2	0	0.32	0	0.982412	-1.783489	110	19	1.1	0.9;
	3255	2	0	0	0	0	0	1.058355	41.76159	20	10	1.1	0.9;
	3256	1	11.61	5.8	0	0.43	0	1.024264	-3.679291	110	15	1.1	0.9;
	3257	1	30.6	4.3	0	7.74	0	1.017591	16.020058	154	1	1.1	0.9;
	3258	1	0	0	0	0.75	0	0.986619	-13.452821	110	1	1.1	0.9;
	3259	1	1.4	0.5	0	0.63	0	1.023766	-2.523865	110	19	1.1	0.9;
	3260	1	50.65	56.5	0	6.86	0	1.045984	36.938005	380	3	1.1	0.9;
	3261	1	0	-0	0	3.82	0	1.010819	90.346267	150	1	1.1	0.9;
	3262	1	208.88	60.8	0	6.68	0	1.016747	23.522042	220	5	1.1	0.9;
	3263	1	-6.7	-7.5	0	1.48	0	1.051248	-5.705729	150	9	1.1	0.9;
	3264	1	4.3	1.4	0	1.59	0	0.997223	-9.751241	110	12	1.1	0.9;
	3265	2	0	0	0	0	0	1.05787	-5.405768	20	4	1.1	0.9;
	3266	1	74.68	46.3	0	4.98	0	1.001352	-15.450636	150	9	1.1	0.9;
	3267	1	-0	0	0	15.02	0	1.030691	13.607772	220	7	1.1	0.9;
	3268	1	0	0	0.097763	8.512237	0	1.033013	18.139955	220	5	1.1	0.9;
	3269	1	13.04	7.4	0	5.33	0	1.028324	-8.261232	150	9	1.1	0.9;
	3270	2	0	0	0	0	0	0.98472	-1.315224	20	18	1.1	0.9;
	3271	1	-0	-0	0	1.32	0	1.041611	-6.848028	150	9	1.1	0.9;
	3272	1	71.88	51.3	0	24.75	0	1.051683	13.028435	380	16	1.1	0.9;
	3273	1	15.02	5	0	0.39	0	0.996835	-3.084223	110	21	1.1	0.9;
	3274	1	30.83	4.4	0	0.46	0	0.989931	-17.785937	110	22	1.1	0.9;
	3275	1	93.2	13.6	0	14.9	0	1.013363	-27.00617	220	4	1.1	0.9;
	3276	1	162.7	23.2	0	1.7	0	1.023041	-1.733209	154	1	1.1	0.9;
	3277	1	402.1	110.2	0	0	0	1.035066	25.308502	150	2	1.1	0.9;
	3278	1	-0	0	0	13.21	0	1.022381	6.286187	220	3	1.1	0.9;
	3279	1	82.39	12.3	0	9.02	0	1.014258	-26.848945	220	4	1.1	0.9;
	3280	2	0	0	0	0	0	1.041353	7.871779	20	22	1.1	0.9;
	3281	1	628.5	91.8	0	19.83	0	0.988563	-22.912606	380	6	1.1	0.9;
	3282	2	0	0	0	0	0	1.039025	12.06939	20	6	1.1	0.9;
	3283	2	0	0	0	0	0	1.039025	12.06939	20	6	1.1	0.9;
	3284	2	0	0	0	0	0	1.039025	12.06939	20	6	1.1	0.9;
	3285	1	-0	0	0	7.94	0	1.028692	13.638716	220	5	1.1	0.9;
	3286	1	88.7	-0.1	0	2.83	0	1.063544	3.36992	220	5	1.1	0.9;
	3287	1	21.22	8.3	0	19.1	0	1.027757	27.328033	150	8	1.1	0.9;
	3288	1	-27.25	-9.38	0	18.78	0	1.063031	70.16259	380	3	1.1	0.9;
	3289	1	45.75	3.3	0	0.37	0	0.990129	-5.911249	110	3	1.1	0.9;
	3290	1	32.04	6.6	0	0.44	0	1.028836	6.244416	220	5	1.1	0.9;
	3291	1	-1.8	-12.01	0.290423	159.398781	0	1.064525	50.433275	380	3	1.1	0.9;
	3292	2	0	0	0	0	0	1.035771	2.999441	20	14	1.1	0.9;
	3293	1	8.41	3	0	1.2	0	0.986524	-6.192619	110	18	1.1	0.9;
	3294	2	0	0	0	0	0	0.974759	32.169943	15.75	1	1.1	0.9;
	3295	1	0	0	0	0	0	1.049444	-6.481726	150	9	1.1	0.9;
	3296	2	0	0	0	0	0	1.041629	38.466941	20	2	1.1	0.9;
	3297	2	0	0	0	0	0	1.048358	31.36576	20	5	1.1	0.9;
	3298	2	0	0	0	0	0	1.048358	31.36576	20	5	1.1	0.9;
	3299	2	0	0	0	0	0	1.048358	31.36576	20	5	1.1	0.9;
	3300	1	13.9	1.9	0	0.03	0	1.006064	27.970425	154	1	1.1	0.9;
	3301	1	55.46	16.9	0	1.92	0	1.001167	22.914164	220	5	1.1	0.9;
	3302	1	-0	-0	0	0	0	1.027866	16.538428	20	5	1.1	0.9;
	3303	1	-0	-0	0	0	0	1.027866	16.538428	20	5	1.1	0.9;
	3304	1	-0	-0	0	0	0	1.027866	16.538428	20	5	1.1	0.9;
	3305	1	-0	-0	0	0	0	1.027866	16.538428	20	5	1.1	0.9;
	3306	2	0	0	0	0	0	1.048358	31.36576	20	5	1.1	0.9;
	3307	2	0	0	0	0	0	1.048358	31.36576	20	5	1.1	0.9;
	3308	2	0	0	0	0	0	1.054031	9.216302	20	5	1.1	0.9;
	3309	2	0	0	0	0	0	1.054031	9.216302	20	5	1.1	0.9;
	3310	1	39.64	20.9	0	0.65	0	1.029961	20.347416	220	5	1.1	0.9;
	3311	1	-0	0	0	0	0	0.992991	15.869625	20	5	1.1	0.9;
	3312	1	8.41	2.8	0	0.67	0	1.007186	-12.965964	110	12	1.1	0.9;
	3313	1	0	0	0	0	0	0.993675	90.655341	150	1	1.1	0.9;
	3314	1	0	-0	0	3.3	0	0.989667	4.466255	220	1	1.1	0.9;
	3315	1	0	-0	0	4.16	0	1.009646	18.952981	154	1	1.1	0.9;
	3316	1	-0	0	0	0	0	0.992991	15.869625	20	5	1.1	0.9;
	3317	1	0	0	0	101.48	0	0.951009	24.174308	400	1	1.1	0.9;
	3318	1	24.23	6.6	0	1.05	0	0.998359	-14.319232	110	22	1.1	0.9;
	3319	1	114.5	25.2	0	17.57	0	1.03403	4.734221	220	5	1.1	0.9;
	3320	1	0	-0	0	0.4	0	1.021206	16.043182	220	5	1.1	0.9;
	3321	1	66.77	19.8	0	0	0	1.010822	15.191231	110	24	1.1	0.9;
	3322	1	-0	-0	0	28.53	0	1.04126	36.379724	380	1	1.1	0.9;
	3323	2	0	0	0	0	0	1.019144	17.562907	20	14	1.1	0.9;
	3324	1	-0.2	-1.33	0	0	0	0.990299	27.156239	150	8	1.1	0.9;
	3325	2	0	0	0	0	0	1.03955	19.355753	20	5	1.1	0.9;
	3326	2	0	0	0	0	0	1.03955	19.355753	20	5	1.1	0.9;
	3327	2	0	0	0	0	0	1.039449	19.583781	20	5	1.1	0.9;
	3328	2	0	0	0	0	0	1.039449	19.583781	20	5	1.1	0.9;
	3329	2	0	0	0	0	0	1.017638	-6.936102	20	4	1.1	0.9;
	3330	2	0	0	0	0	0	1.062419	33.93163	20	2	1.1	0.9;
	3331	1	0	-0	0	0.5	0	1.012867	-2.151125	380	6	1.1	0.9;
	3332	1	27.93	6.4	0	0.89	0	0.967733	-4.938733	110	18	1.1	0.9;
	3333	1	0	-0	0	0.35	0	1.079271	0.453859	150	9	1.1	0.9;
	3334	2	0	0	0	0	0	1.031677	20.632573	20	5	1.1	0.9;
	3335	2	0	0	0	0	0	1.031677	20.632573	20	5	1.1	0.9;
	3336	2	0	0	0	0	0	1.031677	20.632573	20	5	1.1	0.9;
	3337	2	0	0	0	0	0	1.031677	20.632573	20	5	1.1	0.9;
	3338	2	0	0	0	0	0	0.997985	14.815692	20	5	1.1	0.9;
	3339	2	0	0	0	0	0	1.031677	20.632573	20	5	1.1	0.9;
	3340	1	0	-0	0	0.76	0	1.05284	27.865674	220	1	1.1	0.9;
	3341	1	103.4	27	0	1.41	0	1.020439	-0.4141	220	5	1.1	0.9;
	3342	1	12.71	4.9	0	0.83	0	0.953092	-6.861044	110	18	1.1	0.9;
	3343	2	0	0	0	0	0	1.044129	-4.738961	20	4	1.1	0.9;
	3344	2	0	0	0	0	0	1.062747	39.318661	20	8	1.1	0.9;
	3345	1	50.36	-36.2	0	1.25	0	1.031794	-20.193093	220	4	1.1	0.9;
	3346	1	0	-0	0	0	0	1.013615	22.347716	20	5	1.1	0.9;
	3347	1	24.33	15	0	0.31	0	1.01361	-2.270783	150	9	1.1	0.9;
	3348	2	0	0	0	0	0	1.02922	2.854759	11	1	1.1	0.9;
	3349	1	96.91	13.9	0	3.98	0	1.02814	19.066088	220	5	1.1	0.9;
	3350	1	-0	-0	0	3.28	0	1.078042	3.076046	220	5	1.1	0.9;
	3351	2	0	0	0	0	0	1.049756	20.130595	20	5	1.1	0.9;
	3352	2	0	0	0	0	0	1.035421	27.235973	20	6	1.1	0.9;
	3353	1	-0	-0	0	12.77	0	1.054494	82.561024	380	3	1.1	0.9;
	3354	1	33.04	5.6	0	14.55	0	1.02386	-15.308152	220	4	1.1	0.9;
	3355	1	0	-0	0	426.57	0	1.042858	18.670383	380	6	1.1	0.9;
	3356	1	-0	0	0	2.92	0	1.004221	-1.833701	220	21	1.1	0.9;
	3357	1	53.3	7.6	0	10.7	0	0.997701	34.849551	154	1	1.1	0.9;
	3358	1	-0	-0	0	0.5	0	1.008545	-7.681908	150	9	1.1	0.9;
	3359	1	208.24	11.9	0	24.11	0	1.054031	28.07277	220	2	1.1	0.9;
	3360	2	0	0	0	0	0	1.024095	1.435812	11	1	1.1	0.9;
	3361	2	0	0	0	0	0	1.026025	16.98036	20	14	1.1	0.9;
	3362	1	118.4	7.1	0	11.11	0	1.029604	-21.160918	220	4	1.1	0.9;
	3363	1	-0	0	0	7.03	0	1.01349	41.769888	150	8	1.1	0.9;
	3364	1	84.89	6.1	0.058904	3.850913	0	0.990936	89.584804	150	1	1.1	0.9;
	3365	2	0	0	0	0	0	1.060605	43.468335	20	3	1.1	0.9;
	3366	2	0	0	0	0	0	1.060605	43.468335	20	3	1.1	0.9;
	3367	2	0	0	0	0	0	1.060605	43.468335	20	3	1.1	0.9;
	3368	2	0	0	0	0	0	1.060605	43.468335	20	3	1.1	0.9;
	3369	2	0	0	0	0	0	1.000957	10.706535	11	1	1.1	0.9;
	3370	1	179.4	25.6	0	0.42	0	1.004978	16.730285	154	1	1.1	0.9;
	3371	2	0	0	0	0	0	1.002465	24.966083	11	1	1.1	0.9;
	3372	1	54.46	11.9	0	3.93	0	1.014893	-23.853656	220	4	1.1	0.9;
	3373	1	0	0	0	0	0	0.979844	-8.592076	150	9	1.1	0.9;
	3374	1	225.9	50	0.03945	54.16055	0	1.047328	23.171061	380	14	1.1	0.9;
	3375	1	52.96	8.2	0	5.68	0	1.014647	-26.875031	220	4	1.1	0.9;
	3376	1	-0	-0	0.040027	17.005662	0	1.014113	19.61214	380	5	1.1	0.9;
	3377	1	215.5	31.3	0	14.23	0	1.025023	7.688637	220	5	1.1	0.9;
	3378	1	16.9	3.85	0	25.09	0	0.993261	-1.484639	220	5	1.1	0.9;
	3379	1	-0	0	0	0.81	0	1.010215	-20.850901	220	4	1.1	0.9;
	3380	1	-0	-0	0.471697	3.616573	0	1.029748	23.417775	380	6	1.1	0.9;
	3381	1	0	-0	0	5.71	0	0.99291	4.092686	220	1	1.1	0.9;
	3382	1	-0	-0	0	2.59	0	1.024269	33.869817	220	3	1.1	0.9;
	3383	1	37.94	15.4	0	117.402115	0	1.00224	7.728333	380	5	1.1	0.9;
	3384	2	0	0	0	0	0	1.015185	-4.986575	20	4	1.1	0.9;
	3385	2	0	0	0	0	0	0.988654	21.137466	10	1	1.1	0.9;
	3386	1	0	0	0	4.24	0	0.993023	-22.088253	220	6	1.1	0.9;
	3387	1	5.5	0.8	0	1.57	0	1.007717	20.315673	154	1	1.1	0.9;
	3388	1	-15.92	5.2	0	0.3	0	1.052287	-6.250178	150	9	1.1	0.9;
	3389	1	178.5	65.8	0	93.42	0	1.058741	-0.618359	380	20	1.1	0.9;
	3390	1	0	-0	0.113993	4.365006	0	1.05585	4.663945	220	19	1.1	0.9;
	3391	1	110.7	26.9	0.097905	10.552095	0	1.038039	28.384829	220	5	1.1	0.9;
	3392	1	0	-0	0	2.53	0	1.050098	2.549535	154	1	1.1	0.9;
	3393	1	-0	-0	0	8.48	0	1.022635	59.948045	380	3	1.1	0.9;
	3394	1	0	-0	0	28.65	0	1.015601	-16.259088	380	1	1.1	0.9;
	3395	1	-0	-0	0	14.525048	0	0.979972	13.126418	380	5	1.1	0.9;
	3396	2	0	0	0	0	0	1.017601	-9.8258	20	10	1.1	0.9;
	3397	1	153.3	21.8	0	10	0	1.041041	-0.264253	154	1	1.1	0.9;
	3398	1	0	0	0	0	0	1.008563	-8.920839	150	9	1.1	0.9;
	3399	1	131.1	18.7	0	6.53	0	1.036457	-1.746896	154	1	1.1	0.9;
	3400	1	17.12	3.8	0	1	0	0.990509	-4.575961	110	18	1.1	0.9;
	3401	1	101.7	17.6	0	13.14	0	1.027046	14.54611	220	5	1.1	0.9;
	3402	1	-41.25	13.5	0	0.48	0	1.047034	-4.588658	150	9	1.1	0.9;
	3403	2	0	0	0	0	0	1.017078	-0.332817	20	10	1.1	0.9;
	3404	2	0	0	0	0	0	1.024544	18.974877	20	14	1.1	0.9;
	3405	1	1.4	0.5	0	0.35	0	1.02403	-2.72839	110	19	1.1	0.9;
	3406	1	0	-0	0	2.9	0	1.038025	5.013821	220	5	1.1	0.9;
	3407	1	-0	0	0	5.43	0	1.017179	-1.184569	220	1	1.1	0.9;
	3408	1	32.94	13	0	4.04	0	1.042118	32.286272	150	8	1.1	0.9;
	3409	2	0	0	0	0	0	1.016693	-11.720572	20	4	1.1	0.9;
	3410	2	0	0	0	0	0	1.031553	24.597119	20	5	1.1	0.9;
	3411	1	52.26	44.2	0.181549	9.190889	0	1.052352	7.80343	220	18	1.1	0.9;
	3412	2	0	0	0	0	0	1.031898	-9.317211	20	6	1.1	0.9;
	3413	1	0	0	0	9.85	0	1.025829	-8.115294	220	6	1.1	0.9;
	3414	2	0	0	0	0	0	1.031898	-9.317211	20	6	1.1	0.9;
	3415	2	0	0	0	0	0	1.031898	-9.317211	20	6	1.1	0.9;
	3416	2	0	0	0	0	0	1.031898	-9.317211	20	6	1.1	0.9;
	3417	1	122.1	16.7	0	27.05	0	1.027663	19.154658	380	3	1.1	0.9;
	3418	2	0	0	0	0	0	1.023338	0.199966	20	6	1.1	0.9;
	3419	2	0	0	0	0	0	1.023338	0.199966	20	6	1.1	0.9;
	3420	1	24.13	8.7	0.083411	0.134843	0	1.037361	0.630863	220	19	1.1	0.9;
	3421	1	0.1	1	0	0.06	0	0.987799	33.84916	220	3	1.1	0.9;
	3422	1	57.56	24	0	1.01	0	0.993355	-4.276195	220	6	1.1	0.9;
	3423	1	-0	0	0	4.02	0	1.014991	-2.708292	220	22	1.1	0.9;
	3424	1	55.54	15.06	0	6.28	0	0.974881	5.633736	220	5	1.1	0.9;
	3425	1	0	0	0	0.79	0	0.993608	33.340436	220	3	1.1	0.9;
	3426	1	20.12	12.5	0	2.83	0	0.980816	-7.972115	150	9	1.1	0.9;
	3427	1	173.7	23	0	18.4	0	1.014031	3.520682	220	7	1.1	0.9;
	3428	1	20.82	-0	0	0.13	0	1.055428	37.715156	220	5	1.1	0.9;
	3429	1	200.5	85	0.454272	18.990368	0	1.060842	81.899171	380	3	1.1	0.9;
	3430	1	80.29	28.4	0	8.67	0	1.014486	0.895655	220	20	1.1	0.9;
	3431	1	72.4	10.3	0	8.67	0	0.996638	31.225259	154	1	1.1	0.9;
	3432	2	0	0	0	0	0	1.01809	15.340169	11	1	1.1	0.9;
	3433	1	0	0	0	0.52	0	1.001679	-9.728984	150	9	1.1	0.9;
	3434	1	-0	-0	0	0	0	1.039787	-14.573329	20	4	1.1	0.9;
	3435	1	3.1	-5.2	0	0.86	0	1.014242	3.601829	110	18	1.1	0.9;
	3436	1	69.38	43	0	3.22	0	0.970063	-11.531386	150	9	1.1	0.9;
	3437	1	0	-0	0	3.99	0	0.994775	39.043524	154	1	1.1	0.9;
	3438	1	-0	0	0	0.16	0	0.970327	-8.450148	150	9	1.1	0.9;
	3439	1	0	0	0.094628	0.435372	0	1.050577	5.678636	220	21	1.1	0.9;
	3440	1	27.23	8.9	0	0.72	0	0.963419	-11.658198	110	12	1.1	0.9;
	3441	1	37.24	5.3	0	14.78	0	1.020428	-20.356315	220	4	1.1	0.9;
	3442	1	0	-0	0	0.96	0	0.92804	-9.859715	110	18	1.1	0.9;
	3443	1	-0	-0	0	3.79	0	0.997356	34.995095	154	1	1.1	0.9;
	3444	1	0	0	0	1.69	0	1.03559	5.45741	220	5	1.1	0.9;
	3445	1	117.13	52.4	0	7.51	0	0.987953	-12.008273	150	10	1.1	0.9;
	3446	1	30.6	4.3	0	2.82	0	1.031452	-2.009965	154	1	1.1	0.9;
	3447	1	64.8	99.1	0	4.28	0	1.046044	37.014055	380	3	1.1	0.9;
	3448	1	-0	0	0	91.163618	0	1.061419	31.294242	380	5	1.1	0.9;
	3449	1	-0	-0	0	9.14	0	0.996656	89.663571	150	1	1.1	0.9;
	3450	1	0	-0	0.237418	13.951937	0	1.024316	-15.494313	220	4	1.1	0.9;
	3451	2	0	0	0	0	0	1.032321	27.942406	20	5	1.1	0.9;
	3452	1	93.9	51.9	0.120858	3.729285	0	0.994252	33.32759	220	3	1.1	0.9;
	3453	1	30.6	4.3	0	7.3	0	0.975542	41.132827	154	1	1.1	0.9;
	3454	1	5.51	2	0	1.3	0	1.022093	-1.298193	110	19	1.1	0.9;
	3455	1	253.4	57.4	0.039029	5.000971	0	1.036969	8.807607	220	5	1.1	0.9;
	3456	2	0	0	0	0	0	1.022332	-0.918362	20	6	1.1	0.9;
	3457	1	14.32	5.6	0	88.82	0	1.022727	-22.069659	380	4	1.1	0.9;
	3458	1	-0.54	-1	0	0.81	0	1.051628	2.2953	110	15	1.1	0.9;
	3459	1	14	1.9	0	1.33	0	0.966758	19.018606	154	1	1.1	0.9;
	3460	1	-0	0	0	7.15	0	0.982006	31.373859	220	3	1.1	0.9;
	3461	1	70.58	60.4	0.277205	6.192096	0	1.021236	-10.829774	220	4	1.1	0.9;
	3462	1	3.4	123.8	0	64.94	0	1.010426	10.003085	380	3	1.1	0.9;
	3463	2	0	0	0	0	0	1.063388	34.050037	20	10	1.1	0.9;
	3464	1	262.1	13.9	0	10.23	0	1.012147	29.950952	380	3	1.1	0.9;
	3465	1	-0	0	0.379958	52.33965	0	1.024265	-24.35039	380	10	1.1	0.9;
	3466	1	373.8	55.6	0.039614	6.550386	0	1.025014	6.614618	220	5	1.1	0.9;
	3467	1	10.51	1.3	0	0.18	0	0.987196	-9.010973	110	18	1.1	0.9;
	3468	2	0	0	0	0	0	1.028283	16.586422	20	14	1.1	0.9;
	3469	2	0	0	0	0	0	1.034629	40.948139	20	10	1.1	0.9;
	3470	1	523.3	137.2	0	65.61	0	1.006421	-16.70944	380	6	1.1	0.9;
	3471	2	0	0	0	0	0	1.060112	4.394131	20	20	1.1	0.9;
	3472	1	44.55	16.8	0	2.31	0	1.027981	8.098574	220	5	1.1	0.9;
	3473	1	197.4	52.4	0.057873	1.402127	0	1.059962	21.272506	220	5	1.1	0.9;
	3474	1	0	0	0	0	0	1.03472	7.897462	20	7	1.1	0.9;
	3475	1	119.4	23.6	0.15878	12.06122	0	1.012584	13.76329	220	5	1.1	0.9;
	3476	1	-0	0	0.27075	102.13435	0	1.080752	1.771085	380	18	1.2	0.8;
	3477	1	0	-0	0	0.47	0	1.069676	18.142683	220	3	1.1	0.9;
	3478	1	73.18	23.4	0	0.17	0	0.978974	-1.90048	110	19	1.1	0.9;
	3479	1	-0	-0	0	50	0	0.993763	25.0541	150	8	1.1	0.9;
	3480	1	68.38	6.2	0.039608	2.410392	0	1.010314	-12.093155	220	4	1.1	0.9;
	3481	2	0	0	0	0	0	1.040138	32.45384	20	17	1.1	0.9;
	3482	2	0	0	0	0	0	1.00657	25.464337	20	24	1.1	0.9;
	3483	1	0	0	0	0.57	0	0.96217	-9.934797	150	9	1.1	0.9;
	3484	1	291.2	35.4	0	21.87	0	1.026623	-2.535644	220	5	1.1	0.9;
	3485	2	0	0	0	0	0	1.023602	23.130323	20	6	1.1	0.9;
	3486	1	0	0	0	2.31	0	1.006606	31.420169	220	3	1.1	0.9;
	3487	2	0	0	0	0	0	1.023603	23.127902	20	6	1.1	0.9;
	3488	2	0	0	0	0	0	0.975077	24.924484	10.6	1	1.1	0.9;
	3489	2	0	0	0	0	0	1.023602	23.130323	20	6	1.1	0.9;
	3490	2	0	0	0	0	0	1.023602	23.130323	20	6	1.1	0.9;
	3491	2	0	0	0	0	0	1.037557	26.827154	20	5	1.1	0.9;
	3492	1	0	-8.6	0	297.94	0	1.036951	-14.261938	380	4	1.1	0.9;
	3493	1	-0	0	0	0.64	0	1.036353	8.668859	220	5	1.1	0.9;
	3494	2	0	0	0	0	0	1.036517	-13.684117	20	4	1.1	0.9;
	3495	2	0	0	0	0	0	1.036517	-13.684117	20	4	1.1	0.9;
	3496	2	0	0	0	0	0	1.036517	-13.684117	20	4	1.1	0.9;
	3497	1	-0	-0	0	8.66	0	1.027593	45.337586	220	3	1.1	0.9;
	3498	1	0	0	0	10.21	0	1.012828	30.519623	380	3	1.1	0.9;
	3499	2	0	0	0	0	0	0.998951	38.095582	9	1	1.1	0.9;
	3500	1	-0	0	0.254277	1.046433	0	0.982498	31.482622	380	2	1.1	0.9;
	3501	1	2.43	0.49	0	0.17	0	0.996236	-3.764714	110	18	1.1	0.9;
	3502	1	52.86	13	0	6.04	0	1.050951	24.745156	220	5	1.1	0.9;
	3503	1	45.6	142	0	3.17	0	1.023975	33.808211	220	3	1.1	0.9;
	3504	1	0	0	0	18.25	0	1.016408	34.721433	380	3	1.1	0.9;
	3505	1	-0	-0	0	0.19	0	0.96416	-9.80938	150	9	1.1	0.9;
	3506	1	-0	0	1.106166	14.878817	0	1.065819	38.30131	380	5	1.1	0.9;
	3507	1	69.7	9.9	0	0.12	0	1.005241	12.365918	154	1	1.1	0.9;
	3508	2	0	0	0	0	0	1.01212	1.557585	20	18	1.1	0.9;
	3509	1	34.24	16.5	0	0	0	1.071621	58.815915	220	8	1.1	0.9;
	3510	1	301.9	-103.2	0	0	0	1.106046	27.40938	150	2	1.2	0.8;
	3511	1	20.3	3.7	0	0.06	0	1.00261	29.643643	154	1	1.1	0.9;
	3512	1	-0	-0	0	3.62	0	1.043163	0.620392	150	9	1.1	0.9;
	3513	1	-0	-0	0	6.96	0	0.977145	43.93916	154	1	1.1	0.9;
	3514	1	98.36	-0.47	0	5.44	0	1.032796	24.345595	220	13	1.1	0.9;
	3515	2	0	0	0	0	0	1.045647	-10.158279	20	4	1.1	0.9;
	3516	1	0	0	0	1.55	0	1.031876	-8.151731	150	9	1.1	0.9;
	3517	2	0	0	0	0	0	1.005112	-1.198328	20	18	1.1	0.9;
	3518	1	488.7	118.4	0	75.45	0	1.021184	-3.795612	380	6	1.1	0.9;
	3519	1	262.5	18.9	0	39.85	0	1.016509	21.315301	380	6	1.1	0.9;
	3520	2	0	0	0	0	0	1.046665	36.809947	20	5	1.1	0.9;
	3521	2	0	0	0	0	0	1.046665	36.809947	20	5	1.1	0.9;
	3522	1	30.74	9.6	0	2.75	0	1.026609	5.560793	220	5	1.1	0.9;
	3523	2	0	0	0	0	0	1.046665	36.809947	20	5	1.1	0.9;
	3524	2	0	0	0	0	0	1.046665	36.809947	20	5	1.1	0.9;
	3525	2	0	0	0	0	0	1.046665	36.809947	20	5	1.1	0.9;
	3526	2	0	0	0	0	0	1.046665	36.809947	20	5	1.1	0.9;
	3527	1	179.6	3	0.059781	6.140219	0	1.00273	9.696275	220	5	1.1	0.9;
	3528	2	0	0	0	0	0	1.031098	31.248826	20	5	1.1	0.9;
	3529	2	0	0	0	0	0	1.031098	31.248826	20	5	1.1	0.9;
	3530	2	0	0	0	0	0	1.031098	31.248826	20	5	1.1	0.9;
	3531	1	-0	0	0	145.98	0	1.019147	-15.161399	380	4	1.1	0.9;
	3532	1	6.81	23.8	0.040939	2.299061	0	0.971988	-12.365085	220	10	1.1	0.9;
	3533	1	3.3	1.9	0	0.25	0	1.035099	12.647063	110	19	1.1	0.9;
	3534	1	-5.22	-3.77	0	0.91	0	1.009	-3.635217	110	18	1.1	0.9;
	3535	2	0	0	0	0	0	1.043665	14.930439	20	14	1.1	0.9;
	3536	1	0	0	0	32.55	0	1.083932	72.529665	380	3	1.2	0.8;
	3537	1	1.5	0.4	0	0.89	0	0.970818	-10.152793	110	18	1.1	0.9;
	3538	1	314.65	87.51	0	42.61	0	1.052634	28.316735	380	14	1.1	0.9;
	3539	1	-0	0.5	0	1.82	0	0.935866	-7.822814	110	18	1.1	0.9;
	3540	1	-0	0	0	11.65	0	1.022458	16.340229	220	5	1.1	0.9;
	3541	1	-0	0	0.039385	0.140615	0	1.016192	22.827293	220	5	1.1	0.9;
	3542	1	0	-0	0	1.12	0	1.008608	-8.684052	150	9	1.1	0.9;
	3543	2	0	0	0	0	0	1.051307	2.565569	20	18	1.1	0.9;
	3544	1	0	-0	0	1.31	0	1.023752	-4.194406	150	9	1.1	0.9;
	3545	1	103.5	21.7	0	8.88	0	1.004518	-19.968461	220	4	1.1	0.9;
	3546	1	76.58	14	0	0	0	0.981021	28.816632	220	3	1.1	0.9;
	3547	1	-0	-0	0	8.94	0	1.02717	5.529453	220	5	1.1	0.9;
	3548	1	40.44	12.8	0	1.04	0	0.919642	-11.307959	110	21	1.1	0.9;
	3549	1	35.04	16	0	12.14	0	1.047536	-3.624886	220	23	1.1	0.9;
	3550	1	-0	-0	0	0.21	0	1.036635	21.536239	220	5	1.1	0.9;
	3551	2	0	0	0	0	0	1.045549	56.658865	20	10	1.1	0.9;
	3552	1	0	0	0	0	0	0.982474	0.471501	150	9	1.1	0.9;
	3553	2	0	0	0	0	0	1.048632	39.237883	20	2	1.1	0.9;
	3554	1	4.2	0.6	0	4.73	0	0.993719	14.993796	154	1	1.1	0.9;
	3555	1	50.96	6.8	0	9.13	0	1.032921	46.915222	110	8	1.1	0.9;
	3556	1	-2.65	-1.17	0	65.704714	0	0.99438	26.585157	380	5	1.1	0.9;
	3557	2	0	0	0	0	0	1.066429	42.054161	20	5	1.1	0.9;
	3558	2	0	0	0	0	0	1.06254	42.153021	20	5	1.1	0.9;
	3559	1	-0	-0	0.039181	2.830819	0	1.028344	10.931458	220	5	1.1	0.9;
	3560	1	0	-0	0	11.69	0	1.017793	13.597911	220	5	1.1	0.9;
	3561	1	-0.21	-1.98	0	0	0	1.009695	30.943238	150	8	1.1	0.9;
	3562	1	26.6	3.8	0	6.43	0	1.011684	22.030289	154	1	1.1	0.9;
	3563	1	0	0	0	2.43	0	1.021046	-19.851631	220	6	1.1	0.9;
	3564	1	2	0.5	0	1.13	0	0.995074	-7.669658	110	18	1.1	0.9;
	3565	1	33.5	4.8	0	1.52	0	1.001548	34.654477	154	1	1.1	0.9;
	3566	1	79.09	-36.9	0	4.58	0	1.033327	38.101688	380	3	1.1	0.9;
	3567	2	0	0	0	0	0	0.992476	33.395862	11	1	1.1	0.9;
	3568	1	0	0	0	13.39	0	1.03441	8.755896	220	7	1.1	0.9;
	3569	2	0	0	0	0	0	1.056202	-3.897592	20	5	1.1	0.9;
	3570	1	0	0	0	0	0	1.025816	-8.26672	20	5	1.1	0.9;
	3571	2	0	0	0	0	0	1.043771	22.935303	20	5	1.1	0.9;
	3572	1	0	-0	0.176084	1.383916	0	1.03612	30.869394	220	5	1.1	0.9;
	3573	1	42.2	6	0	5.22	0	0.994784	23.512593	154	1	1.1	0.9;
	3574	1	48.8	6.9	0	3.23	0	0.984061	32.874687	154	1	1.1	0.9;
	3575	1	0	0	0	0	0	1.025816	-8.26672	20	5	1.1	0.9;
	3576	1	74.18	22.6	0	0.31	0	1.036937	17.554732	220	3	1.1	0.9;
	3577	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	3578	1	0	0	0	0	0	0.99193	25.086276	150	8	1.1	0.9;
	3579	1	39.34	-10.7	0	0	0	1.003099	31.233662	150	2	1.1	0.9;
	3580	2	0	0	0	0	0	1.054138	25.098015	20	3	1.1	0.9;
	3581	1	-0	0	0	59.04	0	1.087145	1.037104	380	19	1.2	0.8;
	3582	2	0	0	0	0	0	1.054138	25.098015	20	3	1.1	0.9;
	3583	1	127.2	40.8	0	0.2	0	1.021252	4.698139	220	5	1.1	0.9;
	3584	1	0	0	0	9.36	0	0.967704	-29.098732	220	10	1.1	0.9;
	3585	2	0	0	0	0	0	1.037254	4.793274	20	6	1.1	0.9;
	3586	2	0	0	0	0	0	1.037254	4.793274	20	6	1.1	0.9;
	3587	2	0	0	0	0	0	1.033975	-14.23689	20	6	1.1	0.9;
	3588	2	0	0	0	0	0	1.033975	-14.23689	20	6	1.1	0.9;
	3589	1	36.34	3.3	0	2.16	0	1.019797	-29.648677	220	4	1.1	0.9;
	3590	1	65.07	40.3	0	5.54	0	1.009359	-12.831934	150	9	1.1	0.9;
	3591	1	0	0	0.038998	2.651002	0	1.030391	4.513136	220	7	1.1	0.9;
	3592	1	0	-0	0	84.7	0	1.043945	-12.163947	380	22	1.1	0.9;
	3593	2	0	0	0	0	0	1.017275	5.427076	11.5	1	1.1	0.9;
	3594	1	11.91	3.8	0	0.46	0	0.977983	1.255828	110	18	1.1	0.9;
	3595	1	119.2	32.1	0	4.14	0	1.003565	-17.947547	220	4	1.1	0.9;
	3596	1	0	0	0	18.62	0	1.015526	5.219003	220	7	1.1	0.9;
	3597	1	48.75	16.2	0	0.77	0	0.965234	-12.021644	110	21	1.1	0.9;
	3598	1	-14.9	-4.81	0	0.45	0	0.964011	0.683582	110	18	1.1	0.9;
	3599	1	0	-0	0	36.17	0	1.037334	-14.082496	380	4	1.1	0.9;
	3600	1	41.8	5.9	0	4.74	0	0.979489	33.786715	154	1	1.1	0.9;
	3601	1	133.7	10.4	0.039058	3.540942	0	1.015973	11.841244	220	13	1.1	0.9;
	3602	1	115.87	29.41	0	7.92	0	1.028369	17.782814	220	5	1.1	0.9;
	3603	1	174.1	-519.7	0	54.02	0	1.041768	42.609741	380	3	1.1	0.9;
	3604	1	0	0	0	18.83	0	1.010833	-9.547443	380	12	1.1	0.9;
	3605	1	0	-0	0	29.79	0	1.033512	60.138665	380	3	1.1	0.9;
	3606	1	18.2	2.6	0	1.75	0	1.012302	18.270868	154	1	1.1	0.9;
	3607	2	0	0	0	0	0	1.039371	7.907608	20	22	1.1	0.9;
	3608	2	0	0	0	0	0	1.029635	34.5439	20	5	1.1	0.9;
	3609	1	180.4	71.6	0	81.042938	0	1.013143	25.793225	380	5	1.1	0.9;
	3610	2	0	0	0	0	0	1.068762	42.463865	20	3	1.1	0.9;
	3611	2	0	0	0	0	0	1.068762	42.463865	20	3	1.1	0.9;
	3612	1	-0	-0	0.019731	0.480269	0	0.996474	-3.242785	110	18	1.1	0.9;
	3613	1	0	0	0	6.66	0	1.060493	15.281688	220	3	1.1	0.9;
	3614	1	-0	-0	0	0	0	1.020596	-16.730831	20	4	1.1	0.9;
	3615	1	42.25	12.2	0.149786	10.010142	0	1.021029	-18.0658	220	4	1.1	0.9;
	3616	1	-5.29	-1.91	0	42.657632	0	0.984779	17.794275	380	5	1.1	0.9;
	3617	2	0	0	0	0	0	1.024499	-3.834468	20	20	1.1	0.9;
	3618	1	-0	0	0.058188	3.780311	0	0.993659	32.238381	380	2	1.1	0.9;
	3619	2	0	0	0	0	0	1.043657	3.510271	20	2	1.1	0.9;
	3620	2	0	0	0	0	0	1.043657	3.510271	20	2	1.1	0.9;
	3621	2	0	0	0	0	0	1.030254	1.875231	20	2	1.1	0.9;
	3622	1	-0	0	0	0	0	1.001093	-2.661671	20	2	1.1	0.9;
	3623	1	33.5	4.8	0	3.47	0	1.003215	16.411463	154	1	1.1	0.9;
	3624	1	11.2	1.6	0	3.05	0	0.969701	20.289409	154	1	1.1	0.9;
	3625	2	0	0	0	0	0	1.022039	16.134374	20	2	1.1	0.9;
	3626	1	18.32	14.1	0	3.12	0	0.997548	-6.641209	150	9	1.1	0.9;
	3627	1	238.45	46.86	0	125.534496	0	1.011158	10.497232	380	5	1.1	0.9;
	3628	1	59.77	16.7	0	0.64	0	1.04549	24.880973	220	5	1.1	0.9;
	3629	1	-10.41	-3.3	0	4.11	0	1.03845	-0.981126	150	9	1.1	0.9;
	3630	1	0	0	0	0	0	1.029562	-6.831163	150	9	1.1	0.9;
	3631	2	0	0	0	0	0	1.050006	39.688536	20	3	1.1	0.9;
	3632	1	77.39	20.7	0	0.46	0	0.990319	-5.145593	110	3	1.1	0.9;
	3633	2	0	0	0	0	0	1.024572	-13.995277	20	4	1.1	0.9;
	3634	1	1.3	0	0	0.58	0	0.985279	-0.58714	110	18	1.1	0.9;
	3635	1	-0	-0	0.108686	4.920557	0	1.063829	73.160942	380	3	1.1	0.9;
	3636	1	107	43	0	8.34	0	1.001427	20.99492	220	3	1.1	0.9;
	3637	1	0	0	0.28908	8.031674	0	1.064717	0.810879	220	18	1.2	0.8;
	3638	1	51.56	2.1	0	0.18	0	1.047953	30.002755	220	5	1.1	0.9;
	3639	1	15.12	9.4	0	0.69	0	1.00818	-9.064966	150	9	1.1	0.9;
	3640	1	70.28	16.2	0	0.12	0	1.002688	-1.315468	110	18	1.1	0.9;
	3641	2	0	0	0	0	0	1.03809	81.248194	20	3	1.1	0.9;
	3642	2	0	0	0	0	0	1.03809	81.248194	20	3	1.1	0.9;
	3643	1	118.8	39.5	0	2.61	0	1.014073	-16.400343	220	4	1.1	0.9;
	3644	2	0	0	0	0	0	1.033768	14.596978	20	14	1.1	0.9;
	3645	1	24.33	15	0	2.11	0	0.986182	-8.108772	150	9	1.1	0.9;
	3646	2	0	0	0	0	0	1.01964	-10.829105	20	10	1.1	0.9;
	3647	2	0	0	0	0	0	1.055981	85.333831	20	3	1.1	0.9;
	3648	1	278.35	90.4	0	12.1	0	0.983213	-3.233177	220	6	1.1	0.9;
	3649	1	0	0	0.078072	-0.078072	0	1.024836	0.177514	150	9	1.1	0.9;
	3650	1	0	-0	0	0	0	1.03285	5.192474	20	14	1.1	0.9;
	3651	2	0	0	0	0	0	1.037131	84.517509	20	3	1.1	0.9;
	3652	1	72.68	45	0	8.25	0	1.002482	-9.023078	150	9	1.1	0.9;
	3653	1	38.44	13.2	0	1.73	0	1.014031	-11.505239	110	22	1.1	0.9;
	3654	1	27.33	5	0	0.15	0	0.998271	89.04279	150	1	1.1	0.9;
	3655	2	0	0	0	0	0	1.048808	68.937466	20	3	1.1	0.9;
	3656	2	0	0	0	0	0	1.037131	84.517509	20	3	1.1	0.9;
	3657	1	0	0	0	10.78	0	1.027845	44.66359	380	3	1.1	0.9;
	3658	1	0	-0	0	0.23	0	1.071505	-2.602415	110	23	1.1	0.9;
	3659	1	-0.6	-0.5	0.039759	1.07012	0	0.987125	-10.738282	110	12	1.1	0.9;
	3660	1	-4.88	5.51	0	1.38	0	1.006782	-15.116007	150	9	1.1	0.9;
	3661	1	-0	-0	0	3.86	0	1.055822	28.547053	220	1	1.1	0.9;
	3662	1	0	0	0.059074	-0.019074	0	1.016459	22.356581	220	5	1.1	0.9;
	3663	2	0	0	0	0	0	1.028091	-15.249181	20	10	1.1	0.9;
	3664	1	-0	-0	0	0	0	1.014379	-16.62541	20	4	1.1	0.9;
	3665	1	-0	-0	0	0	0	1.014379	-16.62541	20	4	1.1	0.9;
	3666	1	28.53	11.3	0	6.81	0	1.032339	31.528476	150	8	1.1	0.9;
	3667	1	25.77	-36	0	9.85	0	1.048078	39.856974	220	13	1.1	0.9;
	3668	1	0	0	0	0	0	1.026038	-10.378066	20	4	1.1	0.9;
	3669	1	69.5	9.9	0	98.63	0	0.964668	33.685037	400	1	1.1	0.9;
	3670	1	0	-0	0	2.75	0	1.037142	8.484027	220	7	1.1	0.9;
	3671	2	0	0	0	0	0	1.038581	90.625252	20	3	1.1	0.9;
	3672	2	0	0	0	0	0	1.038581	90.625252	20	3	1.1	0.9;
	3673	1	-0	-0	0	0	0	1.014379	-16.62541	20	4	1.1	0.9;
	3674	2	0	0	0	0	0	1.021115	-13.977656	20	4	1.1	0.9;
	3675	2	0	0	0	0	0	1.025842	-13.216265	20	4	1.1	0.9;
	3676	1	17.02	5.9	0	0.52	0	1.007888	-3.514467	110	19	1.1	0.9;
	3677	1	6.51	1.4	0	0.54	0	0.987529	-2.60863	110	18	1.1	0.9;
	3678	1	1.8	0.7	0	0.59	0	1.029714	0.22642	110	19	1.1	0.9;
	3679	1	-88.6	-24.2	0	5.66	0	0.990688	-24.46203	220	10	1.1	0.9;
	3680	2	0	0	0	0	0	1.014154	-7.89767	20	10	1.1	0.9;
	3681	1	7.71	3.3	0	0.56	0	0.991422	-17.579363	110	22	1.1	0.9;
	3682	2	0	0	0	0	0	1.023493	7.689974	6.3	1	1.1	0.9;
	3683	2	0	0	0	0	0	1.03058	15.553508	20	14	1.1	0.9;
	3684	1	0	-0	0	0.93	0	1.014246	-26.213478	220	4	1.1	0.9;
	3685	1	-0	-0	0	1.27	0	0.977248	8.573334	110	18	1.1	0.9;
	3686	2	0	0	0	0	0	1.003179	38.552189	13.8	1	1.1	0.9;
	3687	1	0	-0	0	68.93	0	1.036572	-17.33246	380	4	1.1	0.9;
	3688	1	25.43	7.1	0	0.22	0	0.993874	-4.949615	110	18	1.1	0.9;
	3689	1	78.59	21.4	0.10048	1.61952	0	0.986667	-5.290434	110	21	1.1	0.9;
	3690	1	42.25	10.6	0.019696	23.820304	0	0.991862	90.288614	150	1	1.1	0.9;
	3691	1	-28.96	-8.62	0	3.08	0	1.037397	-18.746774	220	6	1.1	0.9;
	3692	1	0	0	0	1.16	0	1.026617	-7.205824	150	9	1.1	0.9;
	3693	1	0	0	0.336549	-0.261761	0	1.060844	39.586385	150	8	1.1	0.9;
	3694	1	152.18	23.99	0	14.75	0	0.994502	22.334869	220	13	1.1	0.9;
	3695	2	0	0	0	0	0	1.047752	29.881973	20	5	1.1	0.9;
	3696	2	0	0	0	0	0	1.047752	29.881973	20	5	1.1	0.9;
	3697	2	0	0	0	0	0	1.047752	29.881973	20	5	1.1	0.9;
	3698	2	0	0	0	0	0	1.047752	29.881973	20	5	1.1	0.9;
	3699	2	0	0	0	0	0	1.047752	29.881973	20	5	1.1	0.9;
	3700	2	0	0	0	0	0	1.047752	29.881973	20	5	1.1	0.9;
	3701	2	0	0	0	0	0	1.047752	29.881973	20	5	1.1	0.9;
	3702	1	62.97	17.4	0	0.3	0	1.05008	28.114519	220	5	1.1	0.9;
	3703	1	27.05	22.7	0	14.74	0	1.052063	-6.604376	150	9	1.1	0.9;
	3704	1	0	0	0	87.06	0	1.018219	-4.488219	380	6	1.1	0.9;
	3705	1	-0	0	0	-0.3	0	0.999493	-22.903814	220	6	1.1	0.9;
	3706	1	34.34	12.2	0	0.34	0	0.981241	-4.030762	110	19	1.1	0.9;
	3707	1	9.51	3.1	0	0.52	0	0.980046	-3.760338	110	19	1.1	0.9;
	3708	1	-0	-0	0	1.61	0	0.927075	-12.034628	150	9	1.1	0.9;
	3709	1	-15.72	5.2	0	1.02	0	0.966332	-8.051077	150	9	1.1	0.9;
	3710	2	0	0	0	0	0	1.023144	-9.538309	20	4	1.1	0.9;
	3711	1	2.3	0.8	0	0.59	0	1.028269	-1.177972	110	19	1.1	0.9;
	3712	2	0	0	0	0	0	1.027328	-11.116661	20	4	1.1	0.9;
	3713	1	13.9	1.9	0	6.37	0	1.001857	35.141479	154	1	1.1	0.9;
	3714	2	0	0	0	0	0	1.027328	-11.116661	20	4	1.1	0.9;
	3715	1	0	-0	0.1083	35.537525	0	1.074416	-1.545428	380	21	1.2	0.8;
	3716	1	-0	-0	0	0.36	0	1.066366	1.241813	150	9	1.1	0.9;
	3717	2	0	0	0	0	0	1.020566	-15.965907	20	6	1.1	0.9;
	3718	2	0	0	0	0	0	1.020566	-15.965907	20	6	1.1	0.9;
	3719	1	39.04	24.2	0	13.22	0	1.02378	-11.347601	150	9	1.1	0.9;
	3720	2	0	0	0	0	0	1.020566	-15.965907	20	6	1.1	0.9;
	3721	1	235.1	54.6	0.020343	41.409657	0	1.002241	13.029504	220	14	1.1	0.9;
	3722	1	0	-0	0.356558	0.802869	0	1.017794	22.007307	380	6	1.1	0.9;
	3723	1	-0	-0	0	0.98	0	1.023167	30.670615	220	14	1.1	0.9;
	3724	2	0	0	0	0	0	1.052406	15.378216	20	14	1.1	0.9;
	3725	1	253.77	84.6	0	20.71	0	1.027461	15.874106	220	5	1.1	0.9;
	3726	1	28.13	14.2	0	0.49	0	0.974623	0.373543	110	18	1.1	0.9;
	3727	1	-9.51	3.1	0	0.83	0	0.971118	-5.043555	150	9	1.1	0.9;
	3728	2	0	0	0	0	0	1.041161	15.301508	20	3	1.1	0.9;
	3729	2	0	0	0	0	0	1.041161	15.301508	20	3	1.1	0.9;
	3730	1	-0	-0	0	68.01	0	0.994136	2.381519	400	1	1.1	0.9;
	3731	1	-0	0	0.039158	1.040842	0	1.025789	37.325871	220	3	1.1	0.9;
	3732	1	0	-0	0.690439	2.503402	0	1.056986	40.11833	380	3	1.1	0.9;
	3733	1	-0	0	0	2.08	0	1.043418	-6.197364	150	9	1.1	0.9;
	3734	1	14.72	5.4	0	0.72	0	1.012816	18.331432	220	5	1.1	0.9;
	3735	1	25.63	12.7	0	8.42	0	1.023095	14.051895	220	5	1.1	0.9;
	3736	1	48.8	6.9	0	9.44	0	0.991819	41.572318	154	1	1.1	0.9;
	3737	1	16.32	3	0.078645	0.171355	0	1.01362	13.891163	220	5	1.1	0.9;
	3738	2	0	0	0	0	0	1.024613	34.717073	11	1	1.1	0.9;
	3739	1	0	0	0	3.86	0	0.988327	-11.84474	220	1	1.1	0.9;
	3740	1	-23.13	-8.4	0	1.01	0	1.048922	-4.155024	110	19	1.1	0.9;
	3741	1	0	-0	0	5.09	0	0.998862	-13.022323	150	9	1.1	0.9;
	3742	1	187.19	1.8	0	20.47	0	1.016197	16.501286	220	3	1.1	0.9;
	3743	1	115.3	-8.8	0	16.22	0	1.074827	17.162744	220	3	1.2	0.8;
	3744	1	-0	0	0.176135	2.543221	0	1.031927	36.078435	220	3	1.1	0.9;
	3745	1	-0	0	0	11.54	0	1.0605	6.755748	380	1	1.1	0.9;
	3746	1	226.8	53.2	0	0	0	0.994397	45.599039	220	3	1.1	0.9;
	3747	1	14.92	4.4	0	1.32	0	0.9722	-5.625129	110	18	1.1	0.9;
	3748	1	499.8	110.3	0	4.21	0	1.031194	34.071272	220	3	1.1	0.9;
	3749	1	-0	-0	0	29.89	0	1.080084	-12.975193	150	9	1.2	0.8;
	3750	1	-0	-0	0	31.53	0	0.991611	-11.55884	220	12	1.1	0.9;
	3751	1	0	0	0	7.61	0	1.010898	21.592721	220	3	1.1	0.9;
	3752	1	-0	-0	0	23.88	0	1.012062	-17.461214	380	6	1.1	0.9;
	3753	1	-0	0	0	1.46	0	1.048282	32.763091	150	8	1.1	0.9;
	3754	2	0	0	0	0	0	1.04459	35.949142	20	5	1.1	0.9;
	3755	1	14.74	14.7	0	7.86	0	1.056035	-6.130078	150	9	1.1	0.9;
	3756	1	69.7	9.9	0	2.89	0	0.99107	11.907024	154	1	1.1	0.9;
	3757	2	0	0	0	0	0	1.04459	35.949142	20	5	1.1	0.9;
	3758	1	-0	0	0	1.01	0	1.001985	28.995572	150	8	1.1	0.9;
	3759	1	0	0	0.118268	1.771732	0	0.996863	-6.046226	220	12	1.1	0.9;
	3760	1	112.1	16	0	0.89	0	1.035602	1.219734	154	1	1.1	0.9;
	3761	1	0	0	0	6.91	0	1.018383	-22.418329	380	4	1.1	0.9;
	3762	1	64.4	9.1	0	1.61	0	1.002985	17.450922	154	1	1.1	0.9;
	3763	2	0	0	0	0	0	1.048358	31.36576	20	5	1.1	0.9;
	3764	1	-0	-0	0.058857	2.061143	0	1.023597	2.686954	220	7	1.1	0.9;
	3765	1	9.7	1.4	0	2.74	0	0.991176	37.934917	154	1	1.1	0.9;
	3766	1	10.61	3.5	0	1.61	0	1.000539	-8.686801	110	12	1.1	0.9;
	3767	1	5.51	2	0	0.8	0	0.975165	-3.346759	110	19	1.1	0.9;
	3768	1	0	-0	1.107803	9.437513	0	1.065932	38.264905	380	5	1.1	0.9;
	3769	1	-0	-0	0	1.39	0	0.982575	-7.71788	150	9	1.1	0.9;
	3770	1	203.1	65.2	0	0.64	0	1.032419	5.391576	220	5	1.1	0.9;
	3771	2	0	0	0	0	0	1.002675	8.668171	20	18	1.1	0.9;
	3772	1	166.7	87.6	0	60.34	0	1.008525	14.844609	380	3	1.1	0.9;
	3773	1	29.93	-4.2	0	1.47	0	1.007386	62.981368	220	3	1.1	0.9;
	3774	1	15.12	5.4	0	0.63	0	1.048023	-2.69037	110	15	1.1	0.9;
	3775	1	655.8	72.8	0	10.76	0	1.008333	17.437603	380	3	1.1	0.9;
	3776	1	29.33	9.6	0	0.6	0	0.954138	-8.142046	110	21	1.1	0.9;
	3777	1	-0	0	0	44.86	0	1.006789	-9.025473	380	6	1.1	0.9;
	3778	1	42.95	26.6	0	3.89	0	0.859522	-15.937191	150	9	1.2	0.8;
	3779	1	98.3	25.9	0	3.37	0	0.995427	-25.228844	220	6	1.1	0.9;
	3780	1	33.04	12.1	0	0.55	0	0.984021	-14.70471	110	22	1.1	0.9;
	3781	2	0	0	0	0	0	1.045647	-10.158279	20	4	1.1	0.9;
	3782	2	0	0	0	0	0	1.045647	-10.158279	20	4	1.1	0.9;
	3783	1	45.05	-4.5	0	1.57	0	1.031984	-4.06316	110	23	1.1	0.9;
	3784	1	37.6	5.4	0	4.56	0	0.993915	19.983753	154	1	1.1	0.9;
	3785	1	-0	0	0	1.72	0	1.03532	20.235996	220	5	1.1	0.9;
	3786	1	87.8	492.1	1.132565	159.277787	0	1.056058	-9.799792	380	4	1.1	0.9;
	3787	2	0	0	0	0	0	1.045647	-10.158279	20	4	1.1	0.9;
	3788	1	0	0	0	6.12	0	1.031603	19.976423	220	5	1.1	0.9;
	3789	1	-0	0	0	0.19	0	1.02417	5.104899	220	5	1.1	0.9;
	3790	1	37.94	7.5	0	11.89	0	1.028141	7.209486	220	5	1.1	0.9;
	3791	1	123.1	3.9	0	59.08	0	1.05732	15.310719	380	16	1.1	0.9;
	3792	2	0	0	0	0	0	1.041629	38.466941	20	2	1.1	0.9;
	3793	1	70.18	30.5	0	0.48	0	0.994331	27.571319	220	3	1.1	0.9;
	3794	1	46.65	15.1	0	9.79	0	1.025166	38.583294	110	8	1.1	0.9;
	3795	1	0	0	0.168574	70.609522	0	1.06046	-9.926529	380	22	1.2	0.8;
	3796	1	0	-0	0.079037	0.070722	0	1.024926	-4.261057	220	6	1.1	0.9;
	3797	2	0	0	0	0	0	0.97953	39.88697	20	1	1.1	0.9;
	3798	1	-0	-0	0	0.39	0	0.987713	-4.372476	110	18	1.1	0.9;
	3799	1	0.2	0.2	0	0.3	0	1.010942	25.120376	220	3	1.1	0.9;
	3800	2	0	0	0	0	0	1.034738	85.578957	20	3	1.1	0.9;
	3801	1	32.24	4.8	0	0.85	0	1.018264	-24.474933	220	4	1.1	0.9;
	3802	1	71.47	24	0.058852	26.531148	0	1.029199	7.499056	220	5	1.1	0.9;
	3803	1	207.1	50.4	0	4.93	0	1.004636	10.522343	220	5	1.1	0.9;
	3804	1	0	-0	0	1.38	0	0.987933	33.848491	220	3	1.1	0.9;
	3805	1	-0.72	0	0	69.582294	0	1.037622	6.656996	380	5	1.1	0.9;
	3806	1	0	0	0	1.28	0	1.024055	78.99349	220	3	1.1	0.9;
	3807	1	-10.81	35.4	0.198284	15.121287	0	1.017888	-21.976636	220	4	1.1	0.9;
	3808	1	0	-0	0	64.39	0	0.973113	20.532216	400	1	1.1	0.9;
	3809	1	-0	-0	0.118987	1.05076	0	1.007265	-3.961623	150	10	1.1	0.9;
	3810	2	0	0	0	0	0	1.043637	16.74073	20	5	1.1	0.9;
	3811	2	0	0	0	0	0	1.037557	26.827154	20	5	1.1	0.9;
	3812	2	0	0	0	0	0	1.037557	26.827154	20	5	1.1	0.9;
	3813	1	62.67	7.7	0.039244	5.490567	0	0.990954	87.918132	150	1	1.1	0.9;
	3814	2	0	0	0	0	0	0.990319	27.296201	20	5	1.1	0.9;
	3815	2	0	0	0	0	0	1.038225	16.178789	20	5	1.1	0.9;
	3816	2	0	0	0	0	0	1.038225	16.178789	20	5	1.1	0.9;
	3817	1	150	104.3	0	9.21	0	1.002497	5.433967	220	3	1.1	0.9;
	3818	1	7.21	4.4	0.236348	9.252891	0	1.052683	-2.999217	150	9	1.1	0.9;
	3819	1	-0	0	0	2.73	0	0.987464	34.509556	220	3	1.1	0.9;
	3820	2	0	0	0	0	0	1.043637	16.74073	20	5	1.1	0.9;
	3821	1	30.6	4.3	0	8.67	0	1.000644	13.734227	154	1	1.1	0.9;
	3822	1	70.68	10.7	0	1.24	0	1.014986	-26.081164	220	4	1.1	0.9;
	3823	2	0	0	0	0	0	1.046905	39.305577	20	2	1.1	0.9;
	3824	1	223.5	48	0	12.74	0	0.9933	-21.265596	380	6	1.1	0.9;
	3825	1	20.92	1.1	0	1.18	0	0.998239	-16.359764	110	22	1.1	0.9;
	3826	1	16.82	5.5	0	0.11	0	0.998667	-8.872941	110	12	1.1	0.9;
	3827	1	303.8	85.4	0	3.63	0	0.953224	-17.045665	220	10	1.1	0.9;
	3828	1	13.21	4.4	0	1.17	0	0.959197	-14.465123	110	22	1.1	0.9;
	3829	1	-0	0	0	0.12	0	0.966718	-9.612211	150	9	1.1	0.9;
	3830	1	0	0	0	11.28	0	1.015509	-13.056484	220	6	1.1	0.9;
	3831	1	295.8	90.9	0.0593	5.9207	0	1.052987	3.642006	220	5	1.1	0.9;
	3832	1	123.24	23.65	0	6.44	0	1.028263	14.542009	220	5	1.1	0.9;
	3833	1	18.72	6.6	0	0.51	0	1.008916	-9.196952	110	22	1.1	0.9;
	3834	1	0	-0	0.252685	32.19909	0	1.001521	1.510049	220	19	1.1	0.9;
	3835	1	10.61	3.8	0	0.96	0	1.057805	-3.449537	110	19	1.1	0.9;
	3836	1	27.93	10.1	0	0.44	0	1.020046	-0.776938	110	19	1.1	0.9;
	3837	1	-0	-0	0	100.092366	0	1.023558	-4.113173	380	6	1.1	0.9;
	3838	2	0	0	0	0	0	1.004212	-7.923304	20	2	1.1	0.9;
	3839	1	-0	0	0	29.87	0	1.081422	65.714745	380	3	1.1	0.9;
	3840	1	0	-0	0	0.47	0	0.996192	-6.458913	220	12	1.1	0.9;
	3841	1	-0	0	0	23.22	0	0.971137	21.534292	400	1	1.1	0.9;
	3842	2	0	0	0	0	0	0.991765	36.62363	6.3	1	1.1	0.9;
	3843	1	394.9	45.9	0	19.569841	0	1.002541	-8.381312	380	6	1.1	0.9;
	3844	1	-0	-0	0.118846	3.070866	0	1.011328	-24.089806	220	6	1.1	0.9;
	3845	2	0	0	0	0	0	1.021005	21.987746	20	14	1.1	0.9;
	3846	1	107.4	37.3	0.118459	11.111541	0	1.025029	-15.361499	220	4	1.1	0.9;
	3847	1	2	1	0	0.17	0	0.973572	-6.558893	110	21	1.1	0.9;
	3848	2	0	0	0	0	0	0.943602	40.868468	10.6	1	1.1	0.9;
	3849	1	0	0	0	0	0	0.923452	-12.26721	150	9	1.1	0.9;
	3850	1	-0	-0	0	0.45	0	1.044453	29.691902	220	5	1.1	0.9;
	3851	1	-0	0	0	0.85	0	1.036804	-4.012085	150	9	1.1	0.9;
	3852	1	-5.89	-3.59	0	0.51	0	0.997494	-4.67416	110	18	1.1	0.9;
	3853	2	0	0	0	0	0	0.999522	92.907532	20	3	1.1	0.9;
	3854	2	0	0	0	0	0	1.015954	3.75914	20	9	1.1	0.9;
	3855	1	149.6	21.3	0	2.09	0	1.024245	2.286767	154	1	1.1	0.9;
	3856	1	681.5	57.7	0	66.36	0	1.007272	-15.666097	380	6	1.1	0.9;
	3857	2	0	0	0	0	0	1.015006	-6.446837	20	4	1.1	0.9;
	3858	1	-0	-0	0	0.24	0	1.055068	-6.456273	150	9	1.1	0.9;
	3859	1	0	0	0	0.06	0	1.001697	-10.005429	150	9	1.1	0.9;
	3860	1	-0	-1.2	0	0.41	0	1.020631	58.183176	110	8	1.1	0.9;
	3861	1	83.6	11.9	0	3.51	0	1.011146	4.249813	154	1	1.1	0.9;
	3862	1	0	-0	0	3.08	0	1.018959	-11.528438	380	4	1.1	0.9;
	3863	1	0	0	0	0	0	1.002114	-15.311194	150	9	1.1	0.9;
	3864	1	149.9	-0	0	28.38	0	1.04125	-15.47147	380	4	1.1	0.9;
	3865	1	151.7	27.6	0	6.49	0	1.03117	4.253385	220	5	1.1	0.9;
	3866	2	0	0	0	0	0	1.032762	-8.550207	20	4	1.1	0.9;
	3867	2	0	0	0	0	0	1.032762	-8.550207	20	4	1.1	0.9;
	3868	1	-0.2	-4.97	0	0	0	1.021788	34.479059	150	8	1.1	0.9;
	3869	1	22.32	4	0	0.45	0	1.025944	-5.850714	110	3	1.1	0.9;
	3870	1	117.4	12.1	0	1.14	0	1.005382	32.91492	220	3	1.1	0.9;
	3871	1	118	37.8	0	0	0	1.001258	3.503927	110	18	1.1	0.9;
	3872	2	0	0	0	0	0	0.972026	38.478023	6.3	1	1.1	0.9;
	3873	1	0	0	0	0	0	0.955741	-10.11486	150	9	1.1	0.9;
	3874	1	47.75	18.8	0	4.42	0	1.020898	26.410473	150	8	1.1	0.9;
	3875	1	189.7	51.99	0	30.21	0	1.004746	-16.601417	380	6	1.1	0.9;
	3876	1	349.6	96.7	0.950871	42.896884	0	1.015194	-3.247167	380	6	1.1	0.9;
	3877	2	0	0	0	0	0	1.015185	-4.986575	20	4	1.1	0.9;
	3878	2	0	0	0	0	0	1.027132	31.963051	20	5	1.1	0.9;
	3879	1	96.5	-43	0	0.91	0	1.039223	17.370871	220	3	1.1	0.9;
	3880	2	0	0	0	0	0	1.007227	24.551226	11	1	1.1	0.9;
	3881	1	104.8	22	0	6.99	0	1.004302	30.940698	220	3	1.1	0.9;
	3882	1	-0	0	0	0.27	0	1.022362	13.99707	110	19	1.1	0.9;
	3883	1	-0	8.4	0	12.47	0	1.033444	-6.578835	150	9	1.1	0.9;
	3884	1	90.5	27.8	0	7.85	0	1.025704	9.945964	220	7	1.1	0.9;
	3885	1	0	0	0	0	0	1.031155	-14.651392	20	4	1.1	0.9;
	3886	1	-0	0	0	0	0	1.023287	-16.294788	20	4	1.1	0.9;
	3887	1	32.04	11	0	0.57	0	1.04812	-1.694954	110	15	1.1	0.9;
	3888	1	48.8	6.9	0	5.48	0	1.007881	34.678063	154	1	1.1	0.9;
	3889	1	-0	-0	0.098631	1.841369	0	1.010373	12.216216	220	5	1.1	0.9;
	3890	1	13.02	14.2	0	15.78	0	1.021497	43.08635	150	8	1.1	0.9;
	3891	1	0	0	0	10.022192	0	0.924209	5.916211	380	5	1.1	0.9;
	3892	1	-0	0	0	56.39	0	1.023447	33.261963	380	8	1.1	0.9;
	3893	2	0	0	0	0	0	1.025595	-10.129182	20	4	1.1	0.9;
	3894	2	0	0	0	0	0	1.030843	4.085708	6.3	1	1.1	0.9;
	3895	1	151.2	81.3	0	50.19	0	1.02364	36.454073	380	3	1.1	0.9;
	3896	1	43.35	26.9	0.078763	7.571237	0	1.022852	-10.893	150	9	1.1	0.9;
	3897	1	-0	-0	0	4.68	0	1.021979	42.562188	380	3	1.1	0.9;
	3898	1	32.54	11.7	0	1.02	0	1.031562	0.150715	110	15	1.1	0.9;
	3899	2	0	0	0	0	0	1.029166	13.522836	20	14	1.1	0.9;
	3900	1	-0.13	-1	0	0	0	1.071255	58.769096	220	8	1.1	0.9;
	3901	1	26.63	-1.3	0.058336	1.721664	0	1.025753	-6.323962	110	1	1.1	0.9;
	3902	1	0	-0	0	30.05	0	1.025331	0.441862	220	21	1.1	0.9;
	3903	2	0	0	0	0	0	0.988537	92.030569	20	3	1.1	0.9;
	3904	2	0	0	0	0	0	1.054388	47.373035	20	3	1.1	0.9;
	3905	1	0.1	152.7	0	233.41	0	1.013951	-16.487085	380	4	1.1	0.9;
	3906	2	0	0	0	0	0	1.019357	22.943488	20	5	1.1	0.9;
	3907	2	0	0	0	0	0	1.019357	22.943488	20	5	1.1	0.9;
	3908	2	0	0	0	0	0	1.019357	22.943488	20	5	1.1	0.9;
	3909	2	0	0	0	0	0	1.019357	22.943488	20	5	1.1	0.9;
	3910	2	0	0	0	0	0	1.019357	22.943488	20	5	1.1	0.9;
	3911	2	0	0	0	0	0	1.019357	22.943488	20	5	1.1	0.9;
	3912	2	0	0	0	0	0	1.034597	19.022639	20	6	1.1	0.9;
	3913	2	0	0	0	0	0	1.008149	95.140553	20	3	1.1	0.9;
	3914	2	0	0	0	0	0	1.054388	47.373035	20	3	1.1	0.9;
	3915	2	0	0	0	0	0	1.042317	14.274354	20	14	1.1	0.9;
	3916	2	0	0	0	0	0	1.028994	35.401004	20	5	1.1	0.9;
	3917	1	85.59	20.72	0	142.79	0	1.014686	19.975313	380	5	1.1	0.9;
	3918	1	36.19	10.75	0	0.58	0	0.976747	2.601981	110	19	1.1	0.9;
	3919	1	-61.67	-0.9	0	12.33	0	0.977157	-20.730202	150	10	1.1	0.9;
	3920	1	29.5	4.3	0	21.5	0	1.019428	17.336764	154	1	1.1	0.9;
	3921	1	-0	-0	0.173354	13.795539	0	1.039625	75.551848	220	3	1.1	0.9;
	3922	2	0	0	0	0	0	1.064335	2.396339	20	20	1.1	0.9;
	3923	1	-0	-0	0.12635	9.466325	0	1.105017	-0.734092	380	21	1.2	0.8;
	3924	1	-0	0	0	0	0	0.995921	-9.125342	150	9	1.1	0.9;
	3925	1	-0	-0	0	4.52	0	0.967568	-8.292964	220	12	1.1	0.9;
	3926	2	0	0	0	0	0	0.984593	34.337885	15.75	1	1.1	0.9;
	3927	1	0	-0	0	0.6	0	0.975764	-8.269344	150	9	1.1	0.9;
	3928	1	36.24	5.7	0	15.06	0	1.013428	-26.985431	220	4	1.1	0.9;
	3929	2	0	0	0	0	0	0.995077	34.268512	11	1	1.1	0.9;
	3930	1	148.5	21.2	0	1.26	0	1.017816	14.353588	154	1	1.1	0.9;
	3931	1	0	0	0	0	0	1.057526	-6.024854	150	9	1.1	0.9;
	3932	1	0	0	0.038726	11.471274	0	1.040088	77.795929	220	3	1.1	0.9;
	3933	1	-9.82	-3.4	0	0.21	0	1.01076	-0.654704	110	18	1.1	0.9;
	3934	2	0	0	0	0	0	0.997643	-5.902274	20	2	1.1	0.9;
	3935	1	0	0	0	56.27	0	1.029049	23.334025	150	8	1.1	0.9;
	3936	1	27.9	4	0	2.58	0	1.013604	14.122748	154	1	1.1	0.9;
	3937	1	-11.77	-3.81	0	0.23	0	1.015977	-2.619697	110	18	1.1	0.9;
	3938	1	65.47	26.7	0	18.6	0	1.025102	22.870461	150	8	1.1	0.9;
	3939	2	0	0	0	0	0	0.980209	32.361809	15.75	1	1.1	0.9;
	3940	1	55.56	26.5	0	1.09	0	0.989944	-2.693359	220	5	1.1	0.9;
	3941	1	118.4	16.9	0	41.52	0	1.046885	2.545883	154	1	1.2	0.8;
	3942	1	34.54	8.6	0	5.05	0	1.031813	8.024995	220	5	1.1	0.9;
	3943	2	0	0	0	0	0	1.046877	41.968115	20	10	1.1	0.9;
	3944	1	12.21	1	0	0.96	0	0.983686	-5.932747	110	18	1.1	0.9;
	3945	1	-0.63	-1.17	0	1.92	0	0.940212	-11.535774	150	9	1.1	0.9;
	3946	1	0	0	0	3.72	0	1.039621	15.763593	220	5	1.1	0.9;
	3947	2	0	0	0	0	0	1.058835	5.385711	20	20	1.1	0.9;
	3948	2	0	0	0	0	0	0.9646	54.13357	11	1	1.1	0.9;
	3949	1	159.3	13.5	0	62.36	0	1.016389	18.498708	380	6	1.1	0.9;
	3950	1	9.21	6.6	0.059997	0.120003	0	0.994916	-0.136436	110	19	1.1	0.9;
	3951	1	20.12	7.6	0	0.89	0	1.032756	21.238112	220	5	1.1	0.9;
	3952	1	263.3	21.9	0.933649	45.680861	0	1.039539	40.11205	380	13	1.1	0.9;
	3953	1	90.4	13.8	0	0.35	0	1.051684	27.644505	220	1	1.1	0.9;
	3954	1	64.87	25.6	0	14.89	0	1.027371	31.740459	150	8	1.1	0.9;
	3955	1	8.51	2.8	0	1.35	0	1.002954	-7.59108	110	12	1.1	0.9;
	3956	1	59.9	8.5	0	1.27	0	0.993705	27.251449	154	1	1.1	0.9;
	3957	1	16.52	6	0	0.55	0	1.014973	-3.123764	110	19	1.1	0.9;
	3958	1	0.9	49.1	0.530904	186.067075	0	1.020294	-21.912903	380	4	1.1	0.9;
	3959	1	40.5	5.7	0	4.13	0	0.973285	31.740546	154	1	1.1	0.9;
	3960	1	142.2	33.7	0	6.58	0	1.005221	50.42174	220	3	1.1	0.9;
	3961	1	136.6	29	0	2.05	0	1.019851	10.317448	220	5	1.1	0.9;
	3962	2	0	0	0	0	0	1.127961	6.657565	20	9	1.2	0.8;
	3963	1	0	0	0	0.47	0	0.983047	-6.236385	110	18	1.1	0.9;
	3964	1	34.9	5	0	2.1	0	0.992968	17.488668	154	1	1.1	0.9;
	3965	1	98.6	28.3	0	4.76	0	1.048896	58.644273	220	8	1.1	0.9;
	3966	1	85.39	14.3	0	0.35	0	1.037509	21.307632	220	5	1.1	0.9;
	3967	1	57.21	-7.12	0	6.86	0	1.045866	32.653982	220	13	1.1	0.9;
	3968	2	0	0	0	0	0	0.988831	38.418014	10.5	1	1.1	0.9;
	3969	1	103.1	3.1	0	61.7	0	1.050617	-2.825003	380	11	1.1	0.9;
	3970	1	4.7	1.8	0	0.17	0	0.96994	7.065764	110	18	1.1	0.9;
	3971	2	0	0	0	0	0	1.019231	22.327173	20	24	1.1	0.9;
	3972	1	1.5	0.3	0	0.47	0	0.996945	-3.705224	110	18	1.1	0.9;
	3973	1	0	-0	0	56.49	0	0.992697	22.713113	400	1	1.1	0.9;
	3974	1	16.76	-0.51	0	0.35	0	1.021843	-3.697729	110	19	1.1	0.9;
	3975	1	26.63	7.6	0	0.1	0	1.045434	6.466231	110	18	1.1	0.9;
	3976	1	0	0	0	0	0	1.008172	-8.50998	150	9	1.1	0.9;
	3977	1	35.34	21.9	0	3.03	0	0.940462	-11.519394	150	9	1.1	0.9;
	3978	1	2.9	0.8	0	0.68	0	0.971768	-4.92158	110	18	1.1	0.9;
	3979	1	-9.41	3.1	0	0.87	0	0.997595	-6.3622	150	9	1.1	0.9;
	3980	1	-28.13	-0.81	0	0.1	0	1.058636	17.902405	220	5	1.1	0.9;
	3981	1	0	-0	0	14.23	0	1.032156	4.284671	220	5	1.1	0.9;
	3982	1	0.3	1.3	0	0.28	0	1.002071	34.234835	220	3	1.1	0.9;
	3983	1	189.4	25.8	0	3.62	0	1.026894	-22.417254	220	4	1.1	0.9;
	3984	2	0	0	0	0	0	1.028414	-6.287287	20	4	1.1	0.9;
	3985	1	0	0	0	1.77	0	0.993185	32.487862	220	3	1.1	0.9;
	3986	1	129.5	26	0	0	0	0.999796	46.140349	220	3	1.1	0.9;
	3987	1	0	0	0	0	0	1.0014	-12.664749	150	9	1.1	0.9;
	3988	1	55.16	19.8	0	135.46	0	1.081845	64.335223	380	3	1.1	0.9;
	3989	1	0	0	0	0.87	0	0.976827	-2.952988	110	19	1.1	0.9;
	3990	2	0	0	0	0	0	1.014542	2.91424	20	18	1.1	0.9;
	3991	1	3	0.1	0	0.42	0	0.978995	-5.776233	110	18	1.1	0.9;
	3992	1	54.56	16.7	0	0.91	0	0.983142	-2.827838	110	18	1.1	0.9;
	3993	1	16.7	2.4	0	2.21	0	1.009288	12.462553	154	1	1.1	0.9;
	3994	1	0	0	0	0.39	0	1.032143	11.548575	220	5	1.1	0.9;
	3995	1	0	-0	0	1.94	0	1.044996	-3.065495	220	5	1.1	0.9;
	3996	1	100.7	17.5	0	6.54	0	1.021209	34.440022	150	8	1.1	0.9;
	3997	2	0	0	0	0	0	1.033735	-14.68008	20	4	1.1	0.9;
	3998	1	274.1	69.3	0.098089	7.591911	0	1.027268	24.902116	220	5	1.1	0.9;
	3999	1	8.91	2.7	0	1.01	0	0.991303	-5.167444	110	18	1.1	0.9;
	4000	2	0	0	0	0	0	1.038747	-8.642337	20	4	1.1	0.9;
	4001	1	10.11	3.7	0	0.52	0	0.982862	-1.448349	110	19	1.1	0.9;
	4002	1	-0	0	0.078843	-0.028843	0	1.04374	17.148861	220	5	1.1	0.9;
	4003	1	1	128	0.717304	71.222097	0	1.016745	-21.316913	380	4	1.1	0.9;
	4004	2	0	0	0	0	0	1.043665	14.930439	20	14	1.1	0.9;
	4005	1	0	-0	0	12.17	0	1.027589	5.680135	220	5	1.1	0.9;
	4006	1	24.63	8.1	0	0.93	0	0.966005	-6.185631	110	21	1.1	0.9;
	4007	1	69.08	68	0	2.93	0	0.99605	-7.338816	220	6	1.1	0.9;
	4008	1	31.33	9.6	0	1.86	0	1.022605	43.459728	110	8	1.1	0.9;
	4009	2	0	0	0	0	0	1.016106	0.556804	11	1	1.1	0.9;
	4010	1	390.8	79.5	0	20.79	0	1.025555	-13.384562	380	6	1.1	0.9;
	4011	2	0	0	0	0	0	1.045295	35.245517	20	5	1.1	0.9;
	4012	1	0	0	0	0.39	0	1.034685	29.873888	220	5	1.1	0.9;
	4013	2	0	0	0	0	0	1.024096	61.613978	20	10	1.1	0.9;
	4014	1	327.4	125.3	0	34.36	0	1.048336	10.601474	380	16	1.1	0.9;
	4015	1	72	10.3	0	5.5	0	1.00785	17.28578	154	1	1.1	0.9;
	4016	1	27.53	11	0	2.49	0	0.94325	-17.454758	110	22	1.1	0.9;
	4017	1	0	0	0	5.37	0	1.02334	36.751236	154	1	1.1	0.9;
	4018	1	6.91	2.3	0	0.74	0	0.987618	-10.736749	110	12	1.1	0.9;
	4019	1	97	18.6	0	1.25	0	1.017989	42.299548	150	8	1.1	0.9;
	4020	2	0	0	0	0	0	1.035469	23.613483	20	5	1.1	0.9;
	4021	2	0	0	0	0	0	1.035469	23.613483	20	5	1.1	0.9;
	4022	2	0	0	0	0	0	1.048029	9.440891	20	5	1.1	0.9;
	4023	1	32.94	11.8	0	0.83	0	1.021457	0.097153	110	19	1.1	0.9;
	4024	2	0	0	0	0	0	1.049669	23.260983	20	5	1.1	0.9;
	4025	2	0	0	0	0	0	1.049669	23.260983	20	5	1.1	0.9;
	4026	2	0	0	0	0	0	1.045658	29.191718	20	5	1.1	0.9;
	4027	2	0	0	0	0	0	1.045658	29.191718	20	5	1.1	0.9;
	4028	1	-0	0	0	0	0	1.047384	5.160988	20	5	1.1	0.9;
	4029	1	-0	0	0	0	0	1.047384	5.160988	20	5	1.1	0.9;
	4030	1	205.8	66.8	0	3.65	0	1.044254	35.795297	380	3	1.1	0.9;
	4031	1	0	-0	1.139177	12.19252	0	1.053588	30.508882	380	5	1.1	0.9;
	4032	1	-0	0	0	0.25	0	0.995355	0.982197	150	9	1.1	0.9;
	4033	1	-0	0	0.040549	-0.040549	0	0.976468	-19.018776	150	10	1.1	0.9;
	4034	1	76.6	10.9	0	2.96	0	1.019562	6.13255	154	1	1.1	0.9;
	4035	1	22.02	2.2	0	2.4	0	1.01212	-14.62095	110	22	1.1	0.9;
	4036	1	0	0	0	1.28	0	0.994207	-7.173177	150	9	1.1	0.9;
	4037	1	81.19	31.4	0	2.04	0	0.978444	-23.811182	220	6	1.1	0.9;
	4038	1	0	-0	0.0785	7.691125	0	1.017934	-20.325334	220	6	1.1	0.9;
	4039	2	0	0	0	0	0	1.040404	13.619935	20	5	1.1	0.9;
	4040	2	0	0	0	0	0	1.040404	13.619935	20	5	1.1	0.9;
	4041	2	0	0	0	0	0	1.066829	5.050047	20	5	1.1	0.9;
	4042	2	0	0	0	0	0	1.040404	13.619935	20	5	1.1	0.9;
	4043	2	0	0	0	0	0	1.048055	45.108406	20	3	1.1	0.9;
	4044	2	0	0	0	0	0	1.048055	45.108406	20	3	1.1	0.9;
	4045	2	0	0	0	0	0	1.022502	19.424113	20	5	1.1	0.9;
	4046	1	-0	-0	0	31.79	0	1.037914	-10.199077	150	9	1.1	0.9;
	4047	2	0	0	0	0	0	1.048211	53.23324	20	3	1.1	0.9;
	4048	2	0	0	0	0	0	1.049221	26.031718	20	5	1.1	0.9;
	4049	2	0	0	0	0	0	1.041409	32.91812	20	3	1.1	0.9;
	4050	1	0	-0	0	8.74	0	0.980747	40.789038	154	1	1.1	0.9;
	4051	1	8.91	4.9	0	4.58	0	1.02423	25.85169	220	13	1.1	0.9;
	4052	2	0	0	0	0	0	1.001216	-0.162465	20	18	1.1	0.9;
	4053	2	0	0	0	0	0	1.035102	5.926523	20	6	1.1	0.9;
	4054	2	0	0	0	0	0	1.035102	5.926523	20	6	1.1	0.9;
	4055	1	166.7	87.6	0	4.22	0	1.029081	8.308248	220	3	1.1	0.9;
	4056	2	0	0	0	0	0	1.035102	5.926523	20	6	1.1	0.9;
	4057	2	0	0	0	0	0	1.028397	-15.822938	20	6	1.1	0.9;
	4058	2	0	0	0	0	0	1.028397	-15.822938	20	6	1.1	0.9;
	4059	2	0	0	0	0	0	1.028397	-15.822938	20	6	1.1	0.9;
	4060	2	0	0	0	0	0	1.028397	-15.822938	20	6	1.1	0.9;
	4061	2	0	0	0	0	0	1.028397	-15.822938	20	6	1.1	0.9;
	4062	2	0	0	0	0	0	1.028397	-15.822938	20	6	1.1	0.9;
	4063	2	0	0	0	0	0	1.035771	2.999441	20	14	1.1	0.9;
	4064	1	46.55	9.1	0	0.68	0	0.95793	-5.946482	110	18	1.1	0.9;
	4065	1	7.41	2.7	0	0.56	0	1.059908	0.316663	110	19	1.1	0.9;
	4066	2	0	0	0	0	0	1.013219	-5.540854	20	4	1.1	0.9;
	4067	2	0	0	0	0	0	1.013219	-5.540854	20	4	1.1	0.9;
	4068	2	0	0	0	0	0	1.013219	-5.540854	20	4	1.1	0.9;
	4069	2	0	0	0	0	0	1.013219	-5.540854	20	4	1.1	0.9;
	4070	1	3.8	1.3	0	0.11	0	0.974324	5.178453	110	18	1.1	0.9;
	4071	1	75.08	4	0.693752	69.270779	0	1.038223	21.742293	380	13	1.1	0.9;
	4072	1	31.28	5.34	0	8.75	0	0.992435	89.610855	150	1	1.1	0.9;
	4073	1	0	0	0	21.27	0	0.978133	-14.543133	150	9	1.1	0.9;
	4074	2	0	0	0	0	0	1.027264	14.46311	20	14	1.1	0.9;
	4075	1	0	-0	0.038874	12.290844	0	1.017898	93.587736	150	1	1.1	0.9;
	4076	1	330.9	51.9	0	0	0	1.082295	2.293173	120	20	1.2	0.8;
	4077	1	0	-0	0.057699	0.382301	0	1.080027	3.379229	220	5	1.1	0.9;
	4078	1	-0	0	0.020032	0.069968	0	1.02787	23.027691	220	5	1.1	0.9;
	4079	1	25.13	15.6	0	0.04	0	0.979844	-8.592076	150	9	1.1	0.9;
	4080	1	179	50.5	0	3.08	0	1.008626	-19.001337	220	4	1.1	0.9;
	4081	1	21.22	5.3	0	1.25	0	0.985866	-4.420012	110	18	1.1	0.9;
	4082	1	42.25	4.4	0	9.58843	0	0.934743	6.163013	220	16	1.1	0.9;
	4083	2	0	0	0	0	0	1.023361	29.828652	20	5	1.1	0.9;
	4084	2	0	0	0	0	0	1.023418	29.920304	20	5	1.1	0.9;
	4085	2	0	0	0	0	0	1.009821	29.021748	6.3	1	1.1	0.9;
	4086	1	40.14	8.8	0	0.82	0	1.054801	25.441411	220	5	1.1	0.9;
	4087	2	0	0	0	0	0	1.023418	29.920304	20	5	1.1	0.9;
	4088	1	453.23	139.16	0	25.584498	0	0.992612	23.761381	380	5	1.1	0.9;
	4089	1	-0	0	0	70.11	0	1.009928	-15.578396	380	4	1.1	0.9;
	4090	1	-0	-0	0	4.76	0	1.028835	11.150593	220	5	1.1	0.9;
	4091	1	0	0	0	0.3	0	1.020153	-7.766804	150	9	1.1	0.9;
	4092	1	0.2	77.4	0.112312	70.516407	0	1.068646	82.428858	220	3	1.2	0.8;
	4093	1	-0	0	0.080375	-0.060375	0	0.991175	-8.351367	150	10	1.1	0.9;
	4094	1	-12.33	-2.65	0	69.323028	0	1.038965	18.786912	380	5	1.1	0.9;
	4095	1	-0	0	0.234854	45.08542	0	1.081382	4.111262	380	21	1.2	0.8;
	4096	1	7.11	4.4	0	0.97	0	0.975223	-5.552476	150	9	1.1	0.9;
	4097	1	4.2	1.1	0	0.73	0	0.979572	1.145757	110	18	1.1	0.9;
	4098	1	-0	0	0	0.54	0	1.029964	-7.491249	150	9	1.1	0.9;
	4099	1	192.3	84.6	0	38.41	0	1.01787	4.463063	220	20	1.1	0.9;
	4100	1	-57	-24.9	0	8.1	0	0.961663	-9.848287	150	9	1.1	0.9;
	4101	1	0	0	0	0	0	1.00818	-9.064966	150	9	1.1	0.9;
	4102	2	0	0	0	0	0	1.026027	-7.002651	20	4	1.1	0.9;
	4103	1	17.72	5.9	0	0.17	0	0.975776	-8.68967	110	21	1.1	0.9;
	4104	1	157.2	14.5	0.039215	12.000785	0	1.017398	17.982008	220	13	1.1	0.9;
	4105	1	37.73	13.84	0	0.09	0	0.992471	17.736413	220	5	1.1	0.9;
	4106	1	30.03	10	0	0.66	0	0.959874	-6.874073	110	21	1.1	0.9;
	4107	1	46.35	16.8	0	8.35	0	1.034789	0.003882	220	19	1.1	0.9;
	4108	1	0	-0	0	4.21	0	1.020027	22.344668	220	5	1.1	0.9;
	4109	1	235.3	75	0.158009	38.121991	0	1.027725	4.463602	220	11	1.1	0.9;
	4110	1	134.1	24.8	0	13.73	0	1.029048	6.157125	220	5	1.1	0.9;
	4111	1	18.8	8.5	0	75.218793	0	1.130975	-0.154886	380	9	1.2	0.8;
	4112	1	0	-0	0	0.62	0	1.042169	34.593973	380	2	1.1	0.9;
	4113	1	59.32	-20.32	0	4.86	0	1.008635	-21.730038	220	6	1.1	0.9;
	4114	1	-0	0	0	0.37	0	1.028891	18.463948	220	5	1.1	0.9;
	4115	1	33.36	20	0	0	0	1.103517	-15.031183	150	9	1.2	0.8;
	4116	2	0	0	0	0	0	0.96498	32.264881	400	1	1.1	0.9;
	4117	1	69.28	23.1	0	4.86	0	1.005204	-14.148568	110	22	1.1	0.9;
	4118	1	24.13	15	0	3.86	0	1.042943	-7.588989	150	9	1.1	0.9;
	4119	1	29.93	0	0	0.46	0	1.022047	14.401509	220	5	1.1	0.9;
	4120	2	0	0	0	0	0	1.04914	42.65185	20	2	1.1	0.9;
	4121	1	43.55	12	0.119294	3.440706	0	1.026554	18.183448	220	5	1.1	0.9;
	4122	1	13.82	5	0	0.24	0	1.033358	12.070407	110	19	1.1	0.9;
	4123	1	0	0	0	14.2	0	0.988687	-22.211081	220	6	1.1	0.9;
	4124	1	46.2	6.6	0	0.12	0	1.012789	31.444759	154	1	1.1	0.9;
	4125	1	-7.47	-6.45	0	1.08	0	0.992828	-5.901404	110	18	1.1	0.9;
	4126	2	0	0	0	0	0	1.013384	-6.332412	20	10	1.1	0.9;
	4127	1	205.9	38.3	0	6.54	0	1.032447	15.783395	220	3	1.1	0.9;
	4128	1	282.5	17.4	0	20.04	0	1.001526	-10.113049	380	6	1.1	0.9;
	4129	1	-0	0	0	24.73	0	1.071869	-2.308956	380	1	1.2	0.8;
	4130	2	0	0	0	0	0	0.992143	-13.076815	20	10	1.1	0.9;
	4131	1	57.66	61	0	6.46	0	0.997814	-14.774542	220	17	1.1	0.9;
	4132	1	0	0	0	2.48	0	1.011581	3.186067	220	1	1.1	0.9;
	4133	2	0	0	0	0	0	1.03058	15.553508	20	14	1.1	0.9;
	4134	1	0	-0	0	3.2	0	1.0376	28.404116	220	5	1.1	0.9;
	4135	2	0	0	0	0	0	1.036966	15.051213	20	14	1.1	0.9;
	4136	2	0	0	0	0	0	1.060489	63.22043	20	3	1.1	0.9;
	4137	2	0	0	0	0	0	1.025746	97.376396	20	3	1.1	0.9;
	4138	2	0	0	0	0	0	1.025746	97.376396	20	3	1.1	0.9;
	4139	2	0	0	0	0	0	1.025746	97.376396	20	3	1.1	0.9;
	4140	2	0	0	0	0	0	1.025803	97.507727	20	3	1.1	0.9;
	4141	2	0	0	0	0	0	1.025803	97.507727	20	3	1.1	0.9;
	4142	1	0	0	0	2.82	0	0.971681	-8.335279	220	1	1.1	0.9;
	4143	2	0	0	0	0	0	1.025803	97.507727	20	3	1.1	0.9;
	4144	2	0	0	0	0	0	1.027882	90.395237	20	3	1.1	0.9;
	4145	2	0	0	0	0	0	1.057682	40.579523	20	3	1.1	0.9;
	4146	1	-0	-0	0	0.64	0	0.994482	-11.263795	150	10	1.1	0.9;
	4147	1	151	45.7	0	3.02	0	1.062128	13.395699	220	3	1.1	0.9;
	4148	1	23.43	14.5	0	1.05	0	1.029293	-6.89657	150	9	1.1	0.9;
	4149	1	227	5	0	8.46	0	0.999603	-16.851733	220	4	1.1	0.9;
	4150	1	-0	-0	0.115534	0.36335	0	1.056565	2.913111	110	23	1.1	0.9;
	4151	2	0	0	0	0	0	1.008035	95.91491	20	3	1.1	0.9;
	4152	2	0	0	0	0	0	1.015318	96.999592	20	3	1.1	0.9;
	4153	1	0	0	0	0	0	1.030882	31.218703	150	8	1.1	0.9;
	4154	2	0	0	0	0	0	1.016616	-4.658699	20	4	1.1	0.9;
	4155	2	0	0	0	0	0	1.02424	84.887043	20	3	1.1	0.9;
	4156	2	0	0	0	0	0	1.02424	84.887043	20	3	1.1	0.9;
	4157	2	0	0	0	0	0	1.026566	74.36307	20	3	1.1	0.9;
	4158	2	0	0	0	0	0	1.026566	74.36307	20	3	1.1	0.9;
	4159	2	0	0	0	0	0	1.041381	77.156696	20	3	1.1	0.9;
	4160	2	0	0	0	0	0	1.026566	74.36307	20	3	1.1	0.9;
	4161	2	0	0	0	0	0	1.041381	77.156696	20	3	1.1	0.9;
	4162	2	0	0	0	0	0	0.997106	93.076376	20	3	1.1	0.9;
	4163	2	0	0	0	0	0	1.027512	78.665189	20	3	1.1	0.9;
	4164	2	0	0	0	0	0	1.027512	78.665189	20	3	1.1	0.9;
	4165	1	-0	0	0.019473	0.030527	0	1.059061	5.657484	220	5	1.1	0.9;
	4166	2	0	0	0	0	0	1.061581	6.921348	20	5	1.1	0.9;
	4167	2	0	0	0	0	0	1.008948	-4.613451	20	10	1.1	0.9;
	4168	1	133.7	59	0	2.48	0	0.977908	32.002635	220	3	1.1	0.9;
	4169	1	-0	0	0.37905	21.97435	0	1.081804	7.236687	380	18	1.2	0.8;
	4170	1	-0	-0	0	0.43	0	0.999986	-3.034835	220	6	1.1	0.9;
	4171	2	0	0	0	0	0	1.016461	26.125945	20	5	1.1	0.9;
	4172	1	255.8	144.8	0	52.33	0	1.015162	17.478726	380	3	1.1	0.9;
	4173	1	241	32.3	0	34.74	0	1.029349	-20.926602	220	4	1.1	0.9;
	4174	1	203.8	61	0	0.6	0	1.034716	26.159333	150	2	1.1	0.9;
	4175	1	-9.87	-3.76	0	0.21	0	0.992998	-2.167744	110	18	1.1	0.9;
	4176	1	0	0	0	0	0	1.038979	-13.937788	20	4	1.1	0.9;
	4177	1	-0	0	0	5.21	0	1.000365	-6.940098	220	6	1.1	0.9;
	4178	1	0	0	0	2.87	0	1.054815	30.42484	220	5	1.1	0.9;
	4179	1	10.91	3.6	0	1.21	0	0.997573	-8.894797	110	12	1.1	0.9;
	4180	1	46.25	32.2	0	1.5	0	1.033244	65.697923	220	3	1.1	0.9;
	4181	2	0	0	0	0	0	1.021987	8.5782	20	6	1.1	0.9;
	4182	2	0	0	0	0	0	1.061099	-1.375053	20	5	1.1	0.9;
	4183	2	0	0	0	0	0	1.061099	-1.375053	20	5	1.1	0.9;
	4184	1	51.96	9.4	0	42.51	0	1.053208	-14.335191	380	4	1.1	0.9;
	4185	2	0	0	0	0	0	1.061099	-1.375053	20	5	1.1	0.9;
	4186	1	0	0	0	3.43	0	1.025849	0.594473	110	19	1.1	0.9;
	4187	2	0	0	0	0	0	1.061099	-1.375053	20	5	1.1	0.9;
	4188	1	0	0	0	0.66	0	0.991689	-5.052309	110	18	1.1	0.9;
	4189	1	63.47	30.2	0	81.36	0	1.031654	-10.43315	380	4	1.1	0.9;
	4190	2	0	0	0	0	0	0.985642	34.609787	11	1	1.1	0.9;
	4191	2	0	0	0	0	0	1.03058	15.553508	20	14	1.1	0.9;
	4192	1	0	0	0	0	0	0.971128	-15.83459	150	9	1.1	0.9;
	4193	2	0	0	0	0	0	1.03425	8.859346	20	14	1.1	0.9;
	4194	1	-0	0	0	32.046949	0	0.979837	12.086573	380	5	1.1	0.9;
	4195	2	0	0	0	0	0	1.049436	-0.226153	20	2	1.1	0.9;
	4196	2	0	0	0	0	0	1.019692	25.075054	11	1	1.1	0.9;
	4197	2	0	0	0	0	0	1.005717	-1.442301	20	2	1.1	0.9;
	4198	1	37.24	7.1	0	1.07	0	1.017124	-13.230084	220	4	1.1	0.9;
	4199	1	-669.84	-285.73	0	0	0	0.986732	64.355646	27	3	1.1	0.9;
	4200	2	0	0	0	0	0	1.070991	7.604863	20	5	1.1	0.9;
	4201	1	34.04	12.3	0	0.25	0	1.073256	-2.406789	110	19	1.1	0.9;
	4202	2	0	0	0	0	0	1.001419	42.731033	15.75	1	1.1	0.9;
	4203	2	0	0	0	0	0	1.046905	39.305577	20	2	1.1	0.9;
	4204	1	-0	0	0.700348	17.865558	0	1.026932	-12.933112	380	6	1.1	0.9;
	4205	1	19.92	7.4	0	17.28	0	1.012014	-5.390975	220	5	1.1	0.9;
	4206	2	0	0	0	0	0	1.031165	-7.839877	20	4	1.1	0.9;
	4207	2	0	0	0	0	0	1.030656	3.405373	20	6	1.1	0.9;
	4208	2	0	0	0	0	0	1.017341	23.332363	15	1	1.1	0.9;
	4209	2	0	0	0	0	0	1.023144	-9.538309	20	4	1.1	0.9;
	4210	1	61.97	14.3	0	8.38	0	1.029437	8.424708	220	7	1.1	0.9;
	4211	1	0	-0	0	26.18	0	1.050322	43.725557	380	8	1.1	0.9;
	4212	1	0	-0.01	0	92.38	0	1.023935	3.432539	220	5	1.1	0.9;
	4213	1	-0	0	0	0.34	0	1.054004	-6.505281	150	9	1.1	0.9;
	4214	2	0	0	0	0	0	1.030656	3.405373	20	6	1.1	0.9;
	4215	1	-0	0	0.039814	0.000186	0	1.033173	16.619807	220	5	1.1	0.9;
	4216	1	230.7	19.3	0	4.56	0	1.011721	-3.14251	220	5	1.1	0.9;
	4217	2	0	0	0	0	0	1.018631	-11.528817	20	4	1.1	0.9;
	4218	1	17.7	-333.3	0.134221	89.130826	0	1.049394	44.617258	380	3	1.1	0.9;
	4219	1	70.08	10	0	33.04	0	1.05329	-4.134432	380	11	1.1	0.9;
	4220	1	0	-0	0	0.26	0	1.016739	22.52849	220	5	1.1	0.9;
	4221	1	43.15	11.1	0	0.4	0	1.027135	-4.081656	110	19	1.1	0.9;
	4222	1	-0	-0	0	-4.4045	0	1.106791	7.64715	380	1	1.2	0.8;
	4223	1	-0	-0	0	3.9	0	1.040272	1.405241	220	1	1.1	0.9;
	4224	1	16.72	6.1	0	0.41	0	0.923737	-1.526105	110	19	1.1	0.9;
	4225	1	-0	0	0	137.29	0	1.000923	24.053391	380	5	1.1	0.9;
	4226	2	0	0	0	0	0	1.027461	-17.241264	20	4	1.1	0.9;
	4227	1	-0	0	0.26	6.5	0	1.014959	-1.062228	220	21	1.1	0.9;
	4228	1	103	28.7	0	2.88	0	1.019954	15.839816	220	5	1.1	0.9;
	4229	1	0	0	0	35.93	0	1.047325	-7.65973	380	1	1.1	0.9;
	4230	1	17.5	502.7	0.319748	0.671307	0	1.099512	10.203054	380	19	1.2	0.8;
	4231	1	10.91	3.6	0	0.55	0	1.024031	-6.855341	110	12	1.1	0.9;
	4232	1	-0	0	0	0.75	0	0.959513	-9.813499	150	9	1.1	0.9;
	4233	1	-0	-0	0	2.32	0	1.006575	31.414598	220	3	1.1	0.9;
	4234	2	0	0	0	0	0	0.990977	91.853031	20	3	1.1	0.9;
	4235	2	0	0	0	0	0	1.011745	17.062335	13.8	1	1.1	0.9;
	4236	1	41.95	26	0	12.42	0	1.022953	-11.439051	150	9	1.1	0.9;
	4237	2	0	0	0	0	0	0.994601	31.575691	11	1	1.1	0.9;
	4238	1	-0	-0	0	0	0	1.020557	9.624602	20	14	1.1	0.9;
	4239	1	0	-0	0	0	0	0.988795	89.074217	20	3	1.1	0.9;
	4240	2	0	0	0	0	0	0.992643	93.936649	20	3	1.1	0.9;
	4241	1	-21.38	-13.23	0	2.52	0	1.043238	-2.647032	110	21	1.1	0.9;
	4242	2	0	0	0	0	0	1.022616	65.87032	20	3	1.1	0.9;
	4243	1	215.7	60.28	0	97.42	0	1.001805	26.013339	380	13	1.1	0.9;
	4244	1	-632.1	-447	0	6.4	0	1.027589	42.943439	380	3	1.1	0.9;
	4245	2	0	0	0	0	0	1.063358	20.842529	20	14	1.1	0.9;
	4246	1	0	-0	0	11.03	0	1.046251	24.813234	220	5	1.1	0.9;
	4247	1	16.22	5.7	0	4.68	0	1.075059	0.395845	110	15	1.1	0.9;
	4248	1	104.6	40.7	0	1.42	0	1.02437	-26.534792	220	4	1.1	0.9;
	4249	1	94.3	13.4	0	0.9	0	0.987295	23.502088	154	1	1.1	0.9;
	4250	1	0	0	0	0	0	1.016491	-11.127263	150	9	1.1	0.9;
	4251	1	4.81	1.2	0	0.13	0	0.972913	-28.926073	150	10	1.1	0.9;
	4252	2	0	0	0	0	0	1.050901	18.201002	20	7	1.1	0.9;
	4253	2	0	0	0	0	0	1.077217	12.558971	20	2	1.1	0.9;
	4254	2	0	0	0	0	0	1.077217	12.558971	20	2	1.1	0.9;
	4255	1	-0	0	0	32.63	0	0.998069	-6.301264	220	12	1.1	0.9;
	4256	2	0	0	0	0	0	1.011374	-9.719553	20	10	1.1	0.9;
	4257	1	81.19	24.4	0	0.6	0	1.028997	-11.598749	220	4	1.1	0.9;
	4258	1	16.72	5.5	0	0.66	0	0.998306	-13.674263	110	12	1.1	0.9;
	4259	2	0	0	0	0	0	1.010516	39.993345	11	1	1.1	0.9;
	4260	1	-0.45	-0.03	0	0.46	0	1.003361	2.504232	110	18	1.1	0.9;
	4261	2	0	0	0	0	0	0.993468	44.441311	11	1	1.1	0.9;
	4262	1	0.1	-0.9	0	0.91	0	1.029319	65.07377	220	3	1.1	0.9;
	4263	1	82.6	11.7	0	0.38	0	1.034894	-0.627901	154	1	1.1	0.9;
	4264	2	0	0	0	0	0	1.029166	13.522836	20	14	1.1	0.9;
	4265	1	-0.13	-1	0	0	0	1.02267	55.287657	110	8	1.1	0.9;
	4266	1	-0	0	0	70.905048	0	0.999811	25.88942	380	5	1.1	0.9;
	4267	2	0	0	0	0	0	0.98379	29.90696	10.6	1	1.1	0.9;
	4268	1	93.5	42.5	0	133.56	0	1.040643	-15.185394	380	4	1.1	0.9;
	4269	1	0	0	0	22.54	0	1.017835	3.93041	380	7	1.1	0.9;
	4270	1	-0	0	0	0.68	0	1.002594	-8.605101	150	9	1.1	0.9;
	4271	1	20.9	3	0	2.5	0	0.983757	23.92604	154	1	1.1	0.9;
	4272	1	4.4	1.5	0	0.58	0	0.961186	-0.841341	110	18	1.1	0.9;
	4273	1	-0	-0	0	0.19	0	1.034197	12.040084	220	5	1.1	0.9;
	4274	2	0	0	0	0	0	1.063453	20.858871	20	14	1.1	0.9;
	4275	1	112.51	16.51	0	1.63	0	1.068148	22.122368	220	5	1.1	0.9;
	4276	1	9.31	3	0	0.08	0	1.017443	-8.966772	110	12	1.1	0.9;
	4277	1	30.33	11	0	0.56	0	0.97714	-3.397693	110	19	1.1	0.9;
	4278	1	-0	0	0	0.22	0	0.982175	-5.537714	110	18	1.1	0.9;
	4279	1	97.82	20.59	0	6.67	0	1.022	14.080066	220	5	1.1	0.9;
	4280	1	-0	0	0.526076	0.520923	0	1.098116	34.412051	380	14	1.2	0.8;
	4281	1	65.47	26.7	0	0.6	0	1.029441	24.255246	150	8	1.1	0.9;
	4282	2	0	0	0	0	0	1.003847	10.087109	20	18	1.1	0.9;
	4283	1	53.86	15.4	0.019415	0.900585	0	1.020394	-7.926604	110	22	1.1	0.9;
	4284	2	0	0	0	0	0	1.058465	72.88019	20	3	1.1	0.9;
	4285	2	0	0	0	0	0	1.058465	72.88019	20	3	1.1	0.9;
	4286	2	0	0	0	0	0	1.058465	72.88019	20	3	1.1	0.9;
	4287	2	0	0	0	0	0	1.058465	72.88019	20	3	1.1	0.9;
	4288	2	0	0	0	0	0	1.058465	72.88019	20	3	1.1	0.9;
	4289	1	-0	-0	0	0.93	0	0.95762	-7.505197	110	21	1.1	0.9;
	4290	2	0	0	0	0	0	1.058465	72.88019	20	3	1.1	0.9;
	4291	1	5.81	1.3	0	0.83	0	0.961115	-7.106834	110	21	1.1	0.9;
	4292	1	-0	-0	0	2.47	0	1.030191	-20.990522	220	4	1.1	0.9;
	4293	1	13.9	1.9	0	1.5	0	1.009078	32.145889	154	1	1.1	0.9;
	4294	1	-0	0	0	0.42	0	1.0586	20.351145	380	5	1.1	0.9;
	4295	1	0	-0	0.040033	-0.020033	0	1.008299	-3.139528	110	21	1.1	0.9;
	4296	1	26.83	13	0	0.36	0	0.98083	15.703111	220	6	1.1	0.9;
	4297	1	17.35	-0.8	0	1.21	0	0.979476	-3.862218	110	18	1.1	0.9;
	4298	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	4299	1	137	22.8	0.311343	10.796493	0	1.02699	20.317809	220	5	1.1	0.9;
	4300	1	-0	-0	0	4.11	0	1.023774	-11.345098	150	9	1.1	0.9;
	4301	2	0	0	0	0	0	1.017638	-6.936102	20	4	1.1	0.9;
	4302	1	0	-0	0	21.88	0	1.031992	-0.42865	380	6	1.1	0.9;
	4303	1	-0	-0	0.153209	9.466791	0	1.04343	12.362636	220	7	1.1	0.9;
	4304	1	27.03	9	0	0.2	0	0.96385	-8.844238	110	21	1.1	0.9;
	4305	1	-0	-0	0	5.51	0	1.032706	34.192137	220	3	1.1	0.9;
	4306	1	629	184.3	0.059643	27.270357	0	1.03513	18.037725	380	14	1.1	0.9;
	4307	1	0	0	0	12.22	0	1.032787	34.953885	380	3	1.1	0.9;
	4308	1	0	0	0	0	0	0.859522	-15.937191	150	9	1.2	0.8;
	4309	2	0	0	0	0	0	1.041629	38.466941	20	2	1.1	0.9;
	4310	2	0	0	0	0	0	1.059635	3.783182	20	9	1.1	0.9;
	4311	1	-0	-0	0.223411	1.033133	0	1.073959	-3.242981	380	12	1.2	0.8;
	4312	1	17.72	7.3	0	0.13	0	0.96058	-5.579673	110	18	1.1	0.9;
	4313	1	5.81	1.9	0	3.45	0	0.998654	-9.322995	110	12	1.1	0.9;
	4314	2	0	0	0	0	0	1.017273	-4.353956	20	4	1.1	0.9;
	4315	1	-0	0	0	3.9	0	1.054119	25.676484	220	5	1.1	0.9;
	4316	1	101.2	45.2	0.039705	18.570221	0	0.996114	26.265826	220	13	1.1	0.9;
	4317	1	31.84	5.5	0	0.25	0	0.974827	5.219098	110	18	1.1	0.9;
	4318	1	193.9	221.1	1.827014	194.933614	0	1.067942	38.350166	380	13	1.1	0.9;
	4319	1	17.02	6.2	0	0.87	0	1.022689	2.947261	110	19	1.1	0.9;
	4320	1	48.8	6.9	0	14.55	0	1.025787	17.322493	154	1	1.1	0.9;
	4321	1	139.6	68.4	0	0	0	1.042801	40.432398	150	8	1.1	0.9;
	4322	1	87.73	23.17	0	2.68	0	1.010966	27.093735	220	13	1.1	0.9;
	4323	1	435.2	1.3	0	3.25	0	1.015646	-23.160808	220	4	1.1	0.9;
	4324	2	0	0	0	0	0	1.016415	0.018865	20	18	1.1	0.9;
	4325	1	339.7	106.8	0.460071	4.885777	0	1.04515	31.525887	220	13	1.1	0.9;
	4326	2	0	0	0	0	0	1.028438	-19.446471	20	4	1.1	0.9;
	4327	1	-0	0	0	0.77	0	1.053918	0.437976	150	9	1.1	0.9;
	4328	1	-0	0	0	2.03	0	1.027693	19.154508	380	3	1.1	0.9;
	4329	1	87.6	27.4	0	0.99	0	1.016631	11.070367	110	19	1.1	0.9;
	4330	1	108.8	33.2	0.358716	45.62985	0	1.016453	-24.792837	380	10	1.1	0.9;
	4331	2	0	0	0	0	0	0.992476	33.395862	11	1	1.1	0.9;
	4332	1	0	0	0	22.7	0	1.021799	-18.657358	380	4	1.1	0.9;
	4333	1	146.9	2	0.058721	4.921066	0	1.001355	87.992692	150	1	1.1	0.9;
	4334	1	0	0	0	0	0	1.001352	-15.450636	150	9	1.1	0.9;
	4335	1	80.8	11.6	0	6.96	0	0.998851	33.48563	154	1	1.1	0.9;
	4336	1	33.34	11	0	1.16	0	0.908524	-12.499757	110	21	1.1	0.9;
	4337	2	0	0	0	0	0	1.020298	7.297911	11.5	1	1.1	0.9;
	4338	1	-0	-0	0.685544	54.443768	0	1.033139	21.557753	380	3	1.1	0.9;
	4339	2	0	0	0	0	0	1.027032	8.891931	20	14	1.1	0.9;
	4340	1	2.2	1.1	0	0.67	0	1.042471	24.42381	220	5	1.1	0.9;
	4341	1	12.81	3.9	0	0.66	0	1.062686	0.108075	110	15	1.1	0.9;
	4342	1	-2.35	-2.06	0	18.45	0	1.046959	83.086248	220	3	1.1	0.9;
	4343	2	0	0	0	0	0	1.017273	-4.353956	20	4	1.1	0.9;
	4344	2	0	0	0	0	0	1.02868	47.75243	20	10	1.1	0.9;
	4345	2	0	0	0	0	0	1.017273	-4.353956	20	4	1.1	0.9;
	4346	1	30.23	10	0	0.31	0	1.010548	-2.153324	110	21	1.1	0.9;
	4347	2	0	0	0	0	0	1.033534	-10.544863	20	4	1.1	0.9;
	4348	1	-0	0	0	90.194994	0	1.081404	3.81443	380	20	1.2	0.8;
	4349	1	-0	-0	0.100924	-0.050924	0	1.018458	16.838869	220	5	1.1	0.9;
	4350	1	0	-0	0	1.37	0	1.029633	7.746571	220	5	1.1	0.9;
	4351	1	-0	0	0	0.64	0	1.01891	-9.219346	220	6	1.1	0.9;
	4352	1	74.38	8.8	0	6.1	0	1.031515	19.736436	220	5	1.1	0.9;
	4353	2	0	0	0	0	0	1.012422	6.125903	11	1	1.1	0.9;
	4354	2	0	0	0	0	0	1.021005	21.987746	20	14	1.1	0.9;
	4355	1	0	-0	0	27.69	0	1.014876	35.516812	380	8	1.1	0.9;
	4356	1	-0	-0	0	1.57	0	0.99339	-1.893366	220	6	1.1	0.9;
	4357	1	0	0	0.176454	40.472414	0	1.022735	8.384385	380	11	1.1	0.9;
	4358	1	0	0	0	0.9	0	0.961398	-7.399041	110	21	1.1	0.9;
	4359	1	-0	0	0	1.54	0	1.014444	-26.228611	220	4	1.1	0.9;
	4360	1	-0	-0	0.57572	1.921376	0	1.034319	32.619872	380	5	1.1	0.9;
	4361	1	28.73	9.9	0.038964	7.091036	0	1.010026	-1.361937	110	18	1.1	0.9;
	4362	1	78.59	25.6	0	1	0	0.953397	-6.241067	110	18	1.1	0.9;
	4363	1	6.61	2.4	0	0.61	0	0.993713	-7.604821	110	18	1.1	0.9;
	4364	1	-0	0	0	2.37	0	1.011131	34.830175	154	1	1.1	0.9;
	4365	1	41.15	5.7	0	4.11	0	0.990699	-4.540509	110	18	1.1	0.9;
	4366	2	0	0	0	0	0	0.999838	14.362863	20	18	1.1	0.9;
	4367	2	0	0	0	0	0	1.037481	12.714126	20	18	1.1	0.9;
	4368	1	-0	0	0.70365	1.442944	0	1.022351	-16.833971	380	6	1.1	0.9;
	4369	1	116.93	33.6	0	3.35	0	1.033639	4.691084	220	5	1.1	0.9;
	4370	1	221	146.8	0	8.35	0	1.02692	24.279743	220	5	1.1	0.9;
	4371	2	0	0	0	0	0	1.031705	39.244941	20	10	1.1	0.9;
	4372	1	-0	0	0.176371	0.133024	0	1.028903	35.575869	220	3	1.1	0.9;
	4373	1	96.41	4.7	0.019462	10.010539	0	1.018131	13.098639	220	13	1.1	0.9;
	4374	1	-0	0	0	0.86	0	0.966228	-9.704952	150	9	1.1	0.9;
	4375	1	-0	0	0	0.91	0	1.042435	10.508967	220	7	1.1	0.9;
	4376	2	0	0	0	0	0	1.05097	72.632097	20	3	1.1	0.9;
	4377	1	-0	-0	0	0	0	1.020557	9.624602	20	14	1.1	0.9;
	4378	2	0	0	0	0	0	1.05097	72.632097	20	3	1.1	0.9;
	4379	2	0	0	0	0	0	1.05097	72.632097	20	3	1.1	0.9;
	4380	2	0	0	0	0	0	1.05097	72.632097	20	3	1.1	0.9;
	4381	1	118.4	137	0	0	0	1.03193	23.513612	150	2	1.1	0.9;
	4382	1	-0	0	0	50	0	1.01901	4.536634	380	7	1.1	0.9;
	4383	1	26.93	8	0	0.27	0	0.989359	-1.825	110	18	1.1	0.9;
	4384	2	0	0	0	0	0	1.043087	11.638661	20	14	1.1	0.9;
	4385	2	0	0	0	0	0	1.056129	3.755647	20	5	1.1	0.9;
	4386	2	0	0	0	0	0	1.056129	3.755647	20	5	1.1	0.9;
	4387	2	0	0	0	0	0	1.056129	3.755647	20	5	1.1	0.9;
	4388	1	2.2	0.8	0	1.31	0	0.96614	1.586225	110	18	1.1	0.9;
	4389	1	118.4	16.9	0	2.57	0	1.027342	2.330121	154	1	1.1	0.9;
	4390	1	0	-0	0	4.21	0	0.979195	-8.157227	220	1	1.1	0.9;
	4391	1	-0	-0	0	3.84	0	1.000762	26.627945	150	8	1.1	0.9;
	4392	2	0	0	0	0	0	1.034872	3.134864	11	1	1.1	0.9;
	4393	2	0	0	0	0	0	1.034646	-0.807488	20	2	1.1	0.9;
	4394	1	25.93	8	0	1.24	0	1.006945	-3.256973	110	21	1.1	0.9;
	4395	1	-0	-0	0	0.5	0	1.053877	-6.619644	150	9	1.1	0.9;
	4396	1	-0.54	-0.77	0.181549	1.270889	0	1.069904	5.784464	220	18	1.2	0.8;
	4397	1	0	0	0	0	0	1.025298	4.11833	220	5	1.1	0.9;
	4398	1	0	0	0	8.21	0	1.017859	30.171289	380	3	1.1	0.9;
	4399	1	0	-0	0.224497	70.910941	0	1.016319	3.592341	380	7	1.1	0.9;
	4400	2	0	0	0	0	0	1.04876	45.73683	20	10	1.1	0.9;
	4401	1	181.9	-68.8	0.240365	23.269711	0	0.990365	-24.372042	220	10	1.1	0.9;
	4402	1	1.4	0.2	0	5	0	0.993799	38.541357	154	1	1.1	0.9;
	4403	2	0	0	0	0	0	1.029323	-3.215687	20	6	1.1	0.9;
	4404	1	22.5	9.5	0	29.37436	0	1.126478	-1.587109	380	9	1.2	0.8;
	4405	2	0	0	0	0	0	1.038344	11.554722	20	5	1.1	0.9;
	4406	1	0	-0	0	1.11	0	0.980316	12.22824	380	5	1.1	0.9;
	4407	1	84.09	52.09	0	0	0	1.025451	-11.135864	150	9	1.1	0.9;
	4408	1	0	0	0	0.82	0	1.018686	19.883168	154	1	1.1	0.9;
	4409	1	-14.85	-4.99	0	0.6	0	0.996887	-1.878265	110	18	1.1	0.9;
	4410	1	0	0	0	1.62	0	1.001018	29.463147	154	1	1.1	0.9;
	4411	2	0	0	0	0	0	1.017582	-1.17871	20	10	1.1	0.9;
	4412	1	493.4	53.5	0	58.09	0	1.01959	-20.768164	380	4	1.1	0.9;
	4413	1	8.71	1.6	0	1.23	0	1.022516	9.26728	110	19	1.1	0.9;
	4414	1	38.64	6	0	3.26	0	1.018843	-26.434948	220	4	1.1	0.9;
	4415	1	71.28	6.9	0	1.93	0	1.016672	-25.102813	220	4	1.1	0.9;
	4416	1	-15.25	-6.42	0	43.956703	0	1.00644	12.404584	380	5	1.1	0.9;
	4417	1	-0	0	0	61.58	0	0.982329	32.021815	400	1	1.1	0.9;
	4418	1	55.86	8.8	0	4.2	0	1.013605	-27.369953	220	4	1.1	0.9;
	4419	1	119.7	3.7	0	0.88	0	1.012569	-27.123986	220	4	1.1	0.9;
	4420	1	272.3	58.5	0	23.57	0	1.009664	-14.472462	380	4	1.1	0.9;
	4421	1	566.1	54.2	0	44.559841	0	0.998549	-19.873954	380	6	1.1	0.9;
	4422	2	0	0	0	0	0	1.062896	-0.392816	20	5	1.1	0.9;
	4423	2	0	0	0	0	0	0.984601	26.113187	6.3	1	1.1	0.9;
	4424	2	0	0	0	0	0	1.062896	-0.392816	20	5	1.1	0.9;
	4425	2	0	0	0	0	0	1.062896	-0.392816	20	5	1.1	0.9;
	4426	2	0	0	0	0	0	1.029009	32.575777	20	5	1.1	0.9;
	4427	2	0	0	0	0	0	1.029009	32.575777	20	5	1.1	0.9;
	4428	1	-0	0	0	0	0	0.992666	12.175779	20	5	1.1	0.9;
	4429	1	-0	0	0	0	0	0.992666	12.175779	20	5	1.1	0.9;
	4430	2	0	0	0	0	0	1.033479	18.956983	20	5	1.1	0.9;
	4431	1	-0	0	0	0	0	0.992666	12.175779	20	5	1.1	0.9;
	4432	2	0	0	0	0	0	1.030407	32.510933	20	5	1.1	0.9;
	4433	2	0	0	0	0	0	1.030407	32.510933	20	5	1.1	0.9;
	4434	2	0	0	0	0	0	1.030407	32.510933	20	5	1.1	0.9;
	4435	1	89.6	22.7	0	4.06	0	1.029551	20.015518	220	5	1.1	0.9;
	4436	1	-0	0	0	4.27	0	1.013595	9.488984	220	5	1.1	0.9;
	4437	1	0	-0	0	2.83	0	1.011826	21.695695	220	5	1.1	0.9;
	4438	1	92.9	27.7	0	5.22	0	0.990734	-2.662187	220	5	1.1	0.9;
	4439	1	0	0	0	39.58	0	0.987781	23.062889	380	5	1.1	0.9;
	4440	2	0	0	0	0	0	1.042317	14.274354	20	14	1.1	0.9;
	4441	1	100.7	-23.3	0	2.97	0	1.019093	5.814914	220	7	1.1	0.9;
	4442	1	0	-0	0	0.51	0	1.01963	-10.13352	150	9	1.1	0.9;
	4443	2	0	0	0	0	0	1.128032	6.903218	20	9	1.2	0.8;
	4444	1	4	1.3	0	0.62	0	0.97207	-13.128881	110	12	1.1	0.9;
	4445	1	30.33	10.5	0	0.33	0	1.038188	12.744727	220	5	1.1	0.9;
	4446	1	164	42.3	0	8.07	0	1.001673	-20.498177	220	4	1.1	0.9;
	4447	1	14.02	7.6	0	0.33	0	1.003507	6.361923	110	18	1.1	0.9;
	4448	2	0	0	0	0	0	1.127182	5.800414	20	9	1.2	0.8;
	4449	2	0	0	0	0	0	1.035406	21.74444	20	5	1.1	0.9;
	4450	1	-0	0	0.298417	1.071319	0	1.00848	1.492828	220	6	1.1	0.9;
	4451	2	0	0	0	0	0	1.026147	8.391166	20	6	1.1	0.9;
	4452	2	0	0	0	0	0	1.044746	-1.000699	20	2	1.1	0.9;
	4453	2	0	0	0	0	0	1.026147	8.391166	20	6	1.1	0.9;
	4454	2	0	0	0	0	0	1.026722	7.499091	20	6	1.1	0.9;
	4455	2	0	0	0	0	0	1.025977	8.588106	20	6	1.1	0.9;
	4456	2	0	0	0	0	0	1.026722	7.499091	20	6	1.1	0.9;
	4457	2	0	0	0	0	0	1.026722	7.499091	20	6	1.1	0.9;
	4458	2	0	0	0	0	0	1.031695	0	20	6	1.1	0.9;
	4459	1	0	0	0	0.24	0	1.084051	0.578262	150	9	1.1	0.9;
	4460	2	0	0	0	0	0	1.02748	-20.019397	20	6	1.1	0.9;
	4461	2	0	0	0	0	0	1.026147	8.391166	20	6	1.1	0.9;
	4462	1	567.97	130.73	0	1.09	0	1.037705	1.765872	220	5	1.1	0.9;
	4463	1	300.83	65.8	0.232481	3.525953	0	1.034718	27.446584	220	13	1.1	0.9;
	4464	1	64.67	24.5	0	1.46	0	1.036487	1.705969	220	5	1.1	0.9;
	4465	1	0	-0	0	0.93	0	1.008461	-8.488351	150	9	1.1	0.9;
	4466	1	49.15	12.8	0.120373	25.479751	0	1.014786	-14.560212	220	4	1.1	0.9;
	4467	1	0	-0	0	3.64	0	0.994359	-11.554653	220	1	1.1	0.9;
	4468	1	0	-0	0.116426	5.513574	0	1.027351	14.976779	220	7	1.1	0.9;
	4469	1	159.9	53.2	0	7.99	0	1.027797	5.798068	220	5	1.1	0.9;
	4470	2	0	0	0	0	0	1.057682	40.579523	20	3	1.1	0.9;
	4471	1	15.22	4.2	0	1.35	0	0.994801	-10.572755	110	1	1.1	0.9;
	4472	1	0	0	0.038802	-0.038802	0	1.030888	22.528343	220	3	1.1	0.9;
	4473	2	0	0	0	0	0	1.037029	34.381584	20	10	1.1	0.9;
	4474	1	-0	0	0	1.23	0	0.999778	19.659402	154	1	1.1	0.9;
	4475	1	28.63	12.1	0	0.77	0	0.998935	-13.055396	110	22	1.1	0.9;
	4476	1	11.21	3.7	0	0.49	0	1.020321	-7.162256	110	12	1.1	0.9;
	4477	1	4.1	1.5	0	0.52	0	0.997236	-1.693587	110	19	1.1	0.9;
	4478	1	69.78	15.3	0	1.34	0	1.021495	-26.750182	220	4	1.1	0.9;
	4479	1	30.73	6.3	0.079865	1.130135	0	0.987011	11.817348	220	5	1.1	0.9;
	4480	2	0	0	0	0	0	1.031189	-1.476188	20	20	1.1	0.9;
	4481	2	0	0	0	0	0	1.025116	-16.986526	20	6	1.1	0.9;
	4482	1	128.11	64.9	0	19.1	0	1.040694	34.393917	380	2	1.1	0.9;
	4483	2	0	0	0	0	0	1.054159	6.298568	20	20	1.1	0.9;
	4484	2	0	0	0	0	0	1.022819	-18.100331	20	6	1.1	0.9;
	4485	1	-0	-0	0	6.1	0	0.982602	-7.019601	220	12	1.1	0.9;
	4486	1	380.9	85.1	0	57.449841	0	1.020213	2.674933	380	6	1.1	0.9;
	4487	1	0	-0	0	4	0	1.014397	-8.790084	150	9	1.1	0.9;
	4488	2	0	0	0	0	0	0.982004	27.243336	11	1	1.1	0.9;
	4489	1	65.07	40.3	0	5.43	0	1.001803	-12.546618	150	9	1.1	0.9;
	4490	1	125.2	105.4	0	76.73	0	1.025377	37.825766	380	3	1.1	0.9;
	4491	1	0	-0	0.06	11.38	0	1.032325	1.535915	220	21	1.1	0.9;
	4492	1	2.3	0.8	0	0.8	0	0.984733	-11.498484	110	12	1.1	0.9;
	4493	1	30.63	11.6	0	52.28	0	1.026169	-17.870685	380	4	1.1	0.9;
	4494	1	0	-0	0	2.65	0	1.031757	29.291677	220	1	1.1	0.9;
	4495	2	0	0	0	0	0	1.038747	-8.642337	20	4	1.1	0.9;
	4496	1	0	-0	0	0.39	0	1.025271	-7.114844	150	9	1.1	0.9;
	4497	1	-0	0	0	8.33	0	1.013359	-15.130671	220	4	1.1	0.9;
	4498	1	-0.19	-73.31	0	2.36	0	1.046775	42.710459	380	3	1.1	0.9;
	4499	2	0	0	0	0	0	1.019231	22.327173	20	24	1.1	0.9;
	4500	1	7	1	0	4.69	0	0.998219	19.035279	154	1	1.1	0.9;
	4501	2	0	0	0	0	0	1.028707	4.250081	20	18	1.1	0.9;
	4502	1	0	0	0	0	0	1.033547	32.215254	220	5	1.1	0.9;
	4503	1	36.4	356.1	0.040327	38.479673	0	1.020461	-10.378065	380	4	1.1	0.9;
	4504	1	0	0	0.019653	0.620347	0	1.030095	17.157141	220	5	1.1	0.9;
	4505	2	0	0	0	0	0	1.028414	-6.287287	20	4	1.1	0.9;
	4506	1	57.86	70.5	0.549606	217.978967	0	1.013391	26.007892	380	3	1.1	0.9;
	4507	1	0	0	0	4.36	0	1.006727	63.076151	220	3	1.1	0.9;
	4508	1	99.9	20.1	0	17.41	0	1.018251	31.685041	150	8	1.1	0.9;
	4509	2	0	0	0	0	0	1.061212	43.548638	20	2	1.1	0.9;
	4510	1	0	-0	0.181549	15.300889	0	1.057573	-0.339331	220	18	1.2	0.8;
	4511	1	242.16	73.64	0	12.79	0	1.027331	22.416711	220	5	1.1	0.9;
	4512	2	0	0	0	0	0	1.028707	4.250081	20	18	1.1	0.9;
	4513	2	0	0	0	0	0	0.990908	15.83898	20	18	1.1	0.9;
	4514	1	0	-0	0	0.33	0	0.996914	3.406076	110	18	1.1	0.9;
	4515	1	115.5	71.5	0	150.47	0	1.018221	13.018444	380	6	1.1	0.9;
	4516	1	0	-0	0	3.13	0	1.020049	4.294635	220	7	1.1	0.9;
	4517	2	0	0	0	0	0	0.971901	45.235162	15.75	1	1.1	0.9;
	4518	1	-0	-0	0	3.25	0	1.022517	-12.397149	220	4	1.1	0.9;
	4519	1	104.78	20.92	0	6.17	0	1.004938	11.405014	220	5	1.1	0.9;
	4520	1	0.7	270.7	0.334521	197.981232	0	1.072885	81.684684	380	3	1.2	0.8;
	4521	1	58.97	11.1	0	1.05	0	1.031572	19.723822	220	5	1.1	0.9;
	4522	1	0	-0	0	2.48	0	1.017389	-26.420667	220	4	1.1	0.9;
	4523	1	0	-0	0.348552	2.288904	0	1.047656	5.772198	380	20	1.1	0.9;
	4524	1	3.5	1.4	0	0.83	0	0.962181	-11.221239	110	18	1.1	0.9;
	4525	1	285.7	31.9	0.193667	29.236333	0	1.030304	9.376935	220	7	1.1	0.9;
	4526	2	0	0	0	0	0	0.969146	47.259579	13.8	1	1.1	0.9;
	4527	1	-14.9	-4.6	0	0.14	0	0.986146	2.115148	110	18	1.1	0.9;
	4528	1	56.06	17.1	0.11629	9.96371	0	1.031064	19.661973	220	5	1.1	0.9;
	4529	1	278.3	174.6	0.084108	5.365892	0	0.972301	13.963846	220	5	1.1	0.9;
	4530	1	-0	0	0	16.06	0	1.032296	-9.89235	380	4	1.1	0.9;
	4531	1	56.86	35.2	0	2.29	0	1.018762	-7.534283	150	9	1.1	0.9;
	4532	1	7.61	2.5	0	1.2	0	0.988864	-4.94572	110	21	1.1	0.9;
	4533	1	0	0	0	0.03	0	1.078042	3.076045	220	5	1.1	0.9;
	4534	2	0	0	0	0	0	1.011489	10.204543	20	23	1.1	0.9;
	4535	2	0	0	0	0	0	1.011489	10.204543	20	23	1.1	0.9;
	4536	1	20.72	6.3	0	0.37	0	1.044785	-3.114997	110	19	1.1	0.9;
	4537	2	0	0	0	0	0	1.011489	10.204543	20	23	1.1	0.9;
	4538	1	91.7	13	0	2.25	0	1.023441	-1.107233	154	1	1.1	0.9;
	4539	2	0	0	0	0	0	1.043357	16.76639	20	5	1.1	0.9;
	4540	2	0	0	0	0	0	1.026425	15.756693	20	5	1.1	0.9;
	4541	1	0	0	0	1.83	0	1.016652	-0.629448	150	9	1.1	0.9;
	4542	1	19.82	7.2	0	0.36	0	0.996006	-5.352503	110	19	1.1	0.9;
	4543	1	-0	0	0	24.9092	0	0.997607	-13.44051	380	5	1.1	0.9;
	4544	1	49.86	14.1	0	11.33	0	1.030425	-20.977234	220	4	1.1	0.9;
	4545	1	-0	-0	0	0	0	1.010169	13.76329	20	5	1.1	0.9;
	4546	1	-0	-0	0	0	0	1.010169	13.76329	20	5	1.1	0.9;
	4547	2	0	0	0	0	0	1.040215	27.967812	20	5	1.1	0.9;
	4548	2	0	0	0	0	0	1.040215	27.967812	20	5	1.1	0.9;
	4549	2	0	0	0	0	0	1.040215	27.967812	20	5	1.1	0.9;
	4550	2	0	0	0	0	0	1.040215	27.967812	20	5	1.1	0.9;
	4551	2	0	0	0	0	0	1.040215	27.967812	20	5	1.1	0.9;
	4552	2	0	0	0	0	0	1.040215	27.967812	20	5	1.1	0.9;
	4553	2	0	0	0	0	0	1.01638	20.943598	20	5	1.1	0.9;
	4554	2	0	0	0	0	0	1.040215	27.967812	20	5	1.1	0.9;
	4555	1	0	0	0	0	0	1.026869	30.816332	150	8	1.1	0.9;
	4556	1	80.09	-1.8	0	4.09	0	1.020128	43.998857	380	3	1.1	0.9;
	4557	1	40.34	7.8	0.019977	4	0	1.018198	-20.573487	220	4	1.1	0.9;
	4558	1	8.31	3	0	0.87	0	0.990232	-4.199522	110	19	1.1	0.9;
	4559	1	0	-0	0.056825	3.423175	0	1.064625	6.448089	220	11	1.1	0.9;
	4560	1	68.68	19.1	0	0.1	0	0.995572	2.058235	110	18	1.1	0.9;
	4561	1	0	-0	0	0.21	0	0.949812	-10.547801	150	9	1.1	0.9;
	4562	1	19.02	11.5	0	0	0	0.996004	-9.138869	150	9	1.1	0.9;
	4563	2	0	0	0	0	0	1.019144	17.562907	20	14	1.1	0.9;
	4564	1	149	99.8	0	9.27	0	1.017686	23.465014	220	5	1.1	0.9;
	4565	1	123.8	17.6	0	4.03	0	1.014404	18.336432	154	1	1.1	0.9;
	4566	1	0	-0	0.019918	2.4	0	1.010193	-10.142395	220	4	1.1	0.9;
	4567	1	205	24.6	0.176296	18.672881	0	0.996683	91.121691	150	1	1.1	0.9;
	4568	1	-0	0	0.912692	19.247452	0	1.089177	-1.019557	380	9	1.1	0.9;
	4569	1	136.7	19.3	0	37.5	0	1.016432	13.735608	380	6	1.1	0.9;
	4570	1	10.61	51.2	0.231109	8.234819	0	1.028256	38.376915	380	3	1.1	0.9;
	4571	1	59.98	15.54	0	0.49	0	0.995186	-1.259042	110	18	1.1	0.9;
	4572	2	0	0	0	0	0	1.04927	45.669864	20	3	1.1	0.9;
	4573	2	0	0	0	0	0	1.043665	14.930439	20	14	1.1	0.9;
	4574	1	8.31	2.5	0	0.8	0	1.001175	0.04192	110	21	1.1	0.9;
	4575	1	-0	-8.2	0.120069	9.499931	0	1.011461	-19.855402	220	4	1.1	0.9;
	4576	2	0	0	0	0	0	1.02421	5.097202	20	6	1.1	0.9;
	4577	2	0	0	0	0	0	1.02421	5.097202	20	6	1.1	0.9;
	4578	2	0	0	0	0	0	1.02421	5.097202	20	6	1.1	0.9;
	4579	1	-16.5	3.6	0	1.54	0	1.025301	-10.504175	150	9	1.1	0.9;
	4580	2	0	0	0	0	0	1.02421	5.097202	20	6	1.1	0.9;
	4581	1	184.91	32.8	0	8.35	0	1.037006	12.001931	220	5	1.1	0.9;
	4582	1	11.11	3.6	0	3.36	0	0.997762	-9.31689	110	12	1.1	0.9;
	4583	1	0	-0	0	0.49	0	1.027652	1.721221	154	1	1.1	0.9;
	4584	1	-0	0	0	75	0	1.016467	30.168856	150	8	1.1	0.9;
	4585	2	0	0	0	0	0	1.014507	26.136069	20	6	1.1	0.9;
	4586	1	39.64	9.8	0	1.08	0	0.976931	-18.3327	110	22	1.1	0.9;
	4587	1	13.84	1.85	0	0.2	0	0.981444	-1.785435	110	19	1.1	0.9;
	4588	1	-0	-0	0	1.98	0	0.974931	-5.066414	150	9	1.1	0.9;
	4589	1	-0	0	0.22	21.48	0	1.010439	-1.609642	220	19	1.1	0.9;
	4590	2	0	0	0	0	0	1.000808	-3.647252	20	18	1.1	0.9;
	4591	1	0	0	0	0	0	1.013455	-8.739663	150	9	1.1	0.9;
	4592	1	-0	0	0	1.58	0	1.043782	-6.737508	150	9	1.1	0.9;
	4593	1	132.9	40.5	0	6.92	0	1.001733	30.653349	220	3	1.1	0.9;
	4594	2	0	0	0	0	0	1.027264	14.46311	20	14	1.1	0.9;
	4595	1	39.34	11.2	0	0.85	0	1.004092	29.986156	150	8	1.1	0.9;
	4596	1	180.2	46.4	0	9.7	0	1.015556	-17.99129	220	4	1.1	0.9;
	4597	1	23.63	7.8	0	0.62	0	1.009547	-9.013371	110	12	1.1	0.9;
	4598	2	0	0	0	0	0	1.032985	-6.930528	20	6	1.1	0.9;
	4599	1	-0	-0	0	62.76	0	1.017889	-9.346836	380	6	1.1	0.9;
	4600	1	0	0	0	0.54	0	0.946679	-10.89634	150	9	1.1	0.9;
	4601	1	31.94	12.6	0	10.67	0	1.031122	31.502598	150	8	1.1	0.9;
	4602	2	0	0	0	0	0	1.032985	-6.930528	20	6	1.1	0.9;
	4603	2	0	0	0	0	0	1.032985	-6.930528	20	6	1.1	0.9;
	4604	2	0	0	0	0	0	1.032985	-6.930528	20	6	1.1	0.9;
	4605	1	-0	0	0.058973	0.731027	0	1.045802	17.677058	220	5	1.1	0.9;
	4606	1	9.51	2.8	0	0.5	0	0.986696	-2.173717	110	18	1.1	0.9;
	4607	2	0	0	0	0	0	1.026025	92.46745	20	3	1.1	0.9;
	4608	1	14.12	7.7	0	1.42	0	1.028083	8.918743	220	5	1.1	0.9;
	4609	1	-20.92	2.5	0	16.42	0	1.01663	-5.817088	220	10	1.1	0.9;
	4610	1	107	34.7	0	1.54	0	1.030733	16.163563	220	5	1.1	0.9;
	4611	2	0	0	0	0	0	1.015919	3.580325	6.3	1	1.1	0.9;
	4612	2	0	0	0	0	0	1.033649	3.915318	20	18	1.1	0.9;
	4613	1	75.58	30.3	0	34.66	0	1.014648	-11.384785	220	6	1.1	0.9;
	4614	2	0	0	0	0	0	1.039055	4.342555	11	1	1.1	0.9;
	4615	1	-0	0	0.058106	46.651578	0	1.030556	74.673405	220	3	1.1	0.9;
	4616	2	0	0	0	0	0	1.026027	-7.002651	20	4	1.1	0.9;
	4617	2	0	0	0	0	0	1.041665	2.506214	20	2	1.1	0.9;
	4618	2	0	0	0	0	0	1.024278	13.272885	20	2	1.1	0.9;
	4619	2	0	0	0	0	0	1.024278	13.272885	20	2	1.1	0.9;
	4620	1	14.01	6.5	0	0.09	0	1.0394	27.763701	220	5	1.1	0.9;
	4621	1	127.8	15.2	0	9.34	0	1.029352	19.23476	220	5	1.1	0.9;
	4622	2	0	0	0	0	0	1.054046	31.44229	20	10	1.1	0.9;
	4623	1	-0	0	0	207.85	0	1.041519	-15.897906	380	4	1.1	0.9;
	4624	1	0	-0	0	54.23	0	0.970562	35.740322	400	1	1.1	0.9;
	4625	1	104.5	31.8	0	0.71	0	0.992529	-4.178433	220	6	1.1	0.9;
	4626	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	4627	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	4628	1	398	11.5	0	7.749841	0	1.017578	-4.847153	380	6	1.1	0.9;
	4629	1	15.92	12	0	6.79	0	0.949924	12.732443	220	16	1.1	0.9;
	4630	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	4631	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	4632	1	86.2	19.9	0	5.43	0	1.028847	7.452884	220	5	1.1	0.9;
	4633	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	4634	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	4635	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	4636	1	-0	0	0.354092	-0.245076	0	1.047643	-4.357333	220	5	1.1	0.9;
	4637	1	-0	0	0	0.14	0	1.020805	17.057331	220	5	1.1	0.9;
	4638	1	-0	0	0.343964	110.768913	0	1.049155	27.915801	380	5	1.1	0.9;
	4639	1	0	0	0	15.04	0	1.084959	51.860986	220	3	1.2	0.8;
	4640	1	146.7	20.7	0	62.64	0	1.037956	35.654623	150	8	1.1	0.9;
	4641	1	98.01	16.6	0	0	0	0.936442	-4.804	110	18	1.1	0.9;
	4642	1	-0	0	0	104.11	0	1.06598	-3.21968	380	22	1.2	0.8;
	4643	2	0	0	0	0	0	1.012893	26.368849	11	1	1.1	0.9;
	4644	2	0	0	0	0	0	1.011577	-11.964199	20	4	1.1	0.9;
	4645	1	2.3	0.8	0	0.42	0	0.985531	-15.362323	110	12	1.1	0.9;
	4646	1	0	0	0	0	0	1.017233	-16.531897	20	4	1.1	0.9;
	4647	1	0	0	0	0	0	1.017233	-16.531897	20	4	1.1	0.9;
	4648	2	0	0	0	0	0	1.032267	-6.990671	20	4	1.1	0.9;
	4649	1	0	-0	0	0	0	1.034774	-11.388611	20	4	1.1	0.9;
	4650	1	0	-0	0	0	0	1.034774	-11.388611	20	4	1.1	0.9;
	4651	1	0	-0	0	0	0	1.034774	-11.388611	20	4	1.1	0.9;
	4652	2	0	0	0	0	0	1.037564	83.849724	20	3	1.1	0.9;
	4653	1	-0	0	0	47.99	0	1.025357	33.219063	380	8	1.1	0.9;
	4654	1	0	0	0	5.21	0	0.997462	-17.125727	220	10	1.1	0.9;
	4655	2	0	0	0	0	0	1.042492	18.795962	20	6	1.1	0.9;
	4656	2	0	0	0	0	0	1.042492	18.795962	20	6	1.1	0.9;
	4657	2	0	0	0	0	0	1.026117	25.086822	20	6	1.1	0.9;
	4658	2	0	0	0	0	0	1.026117	25.086822	20	6	1.1	0.9;
	4659	1	-0	-0	0	1.43	0	0.976593	-5.482851	150	9	1.1	0.9;
	4660	1	63.07	25.8	0	18.63	0	1.024738	22.795258	150	8	1.1	0.9;
	4661	1	73.58	30.1	0	6.02	0	1.023633	22.672364	150	8	1.1	0.9;
	4662	2	0	0	0	0	0	1.055425	81.529284	20	3	1.1	0.9;
	4663	2	0	0	0	0	0	1.055425	81.529284	20	3	1.1	0.9;
	4664	2	0	0	0	0	0	1.053529	75.167292	20	3	1.1	0.9;
	4665	2	0	0	0	0	0	1.053529	75.167292	20	3	1.1	0.9;
	4666	2	0	0	0	0	0	1.055479	63.485669	20	3	1.1	0.9;
	4667	1	165.2	64.5	0.040447	7.159553	0	0.991538	17.560846	220	5	1.1	0.9;
	4668	2	0	0	0	0	0	0.998951	38.095582	9	1	1.1	0.9;
	4669	1	341.2	46.6	0.229484	12.678325	0	1.046994	35.737826	220	13	1.1	0.9;
	4670	2	0	0	0	0	0	1.043049	13.489159	20	18	1.1	0.9;
	4671	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	4672	1	76.38	24.8	0	3.24	0	1.029878	17.097079	220	5	1.1	0.9;
	4673	1	-0	0	0	0	0	1.014846	-20.345542	20	4	1.1	0.9;
	4674	1	9.01	1.1	0	1.26	0	0.991989	-14.512462	110	12	1.1	0.9;
	4675	2	0	0	0	0	0	1.036966	15.051213	20	14	1.1	0.9;
	4676	2	0	0	0	0	0	1.059275	72.522304	20	3	1.1	0.9;
	4677	2	0	0	0	0	0	1.059275	72.522304	20	3	1.1	0.9;
	4678	2	0	0	0	0	0	1.059275	72.522304	20	3	1.1	0.9;
	4679	2	0	0	0	0	0	1.052163	86.031538	20	3	1.1	0.9;
	4680	2	0	0	0	0	0	1.052163	86.031538	20	3	1.1	0.9;
	4681	1	0	-0	0	0.6	0	1.01164	-7.354502	150	9	1.1	0.9;
	4682	1	-0	-0	0.039193	0.350807	0	1.010452	12.228523	220	5	1.1	0.9;
	4683	2	0	0	0	0	0	1.047117	89.983578	20	3	1.1	0.9;
	4684	1	177.3	36.4	0	0.7	0	1.028781	8.833625	220	5	1.1	0.9;
	4685	1	81.2	89.3	0	15.55	0	1.030939	38.422518	380	3	1.1	0.9;
	4686	2	0	0	0	0	0	1.028648	-14.669346	20	10	1.1	0.9;
	4687	1	-0	-0	0.233641	-0.194701	0	1.034213	1.500755	380	6	1.1	0.9;
	4688	1	0	-0	0	3.58	0	1.010532	22.946012	220	6	1.1	0.9;
	4689	1	-0	0	0.242346	2.448045	0	1.018486	23.539639	380	6	1.1	0.9;
	4690	2	0	0	0	0	0	1.041748	29.3468	20	5	1.1	0.9;
	4691	2	0	0	0	0	0	1.037993	28.716517	20	5	1.1	0.9;
	4692	2	0	0	0	0	0	1.026746	3.870557	20	20	1.1	0.9;
	4693	1	-0	-0	0	0	0	1.01339	-8.814025	20	6	1.1	0.9;
	4694	2	0	0	0	0	0	1.041748	29.3468	20	5	1.1	0.9;
	4695	2	0	0	0	0	0	1.038587	-4.268876	20	6	1.1	0.9;
	4696	1	-0	-0	0	0	0	1.01339	-8.814025	20	6	1.1	0.9;
	4697	2	0	0	0	0	0	1.039251	-4.888975	20	10	1.1	0.9;
	4698	2	0	0	0	0	0	1.037993	28.716517	20	5	1.1	0.9;
	4699	2	0	0	0	0	0	1.037993	28.716517	20	5	1.1	0.9;
	4700	1	251.3	42.6	0	5.47	0	1.019762	17.054946	220	5	1.1	0.9;
	4701	1	133.9	32.4	0	0	0	0.99113	-5.121845	110	21	1.1	0.9;
	4702	2	0	0	0	0	0	1.041748	29.3468	20	5	1.1	0.9;
	4703	2	0	0	0	0	0	1.017828	-6.986287	20	4	1.1	0.9;
	4704	2	0	0	0	0	0	0.939142	40.851376	13.8	1	1.1	0.9;
	4705	1	374.1	61.4	0.281144	7.339142	0	1.010022	-18.844281	220	4	1.1	0.9;
	4706	1	-38.84	-17	0.07874	3.110945	0	1.017598	3.454906	220	6	1.1	0.9;
	4707	2	0	0	0	0	0	1.041748	29.3468	20	5	1.1	0.9;
	4708	1	-6.58	-1.67	0	1.14	0	1.022811	3.980805	110	18	1.1	0.9;
	4709	2	0	0	0	0	0	1.033648	19.100507	20	5	1.1	0.9;
	4710	2	0	0	0	0	0	1.0447	34.637859	20	14	1.1	0.9;
	4711	1	-0	-0	0.239102	7.690711	0	0.990691	-5.824455	220	12	1.1	0.9;
	4712	1	16.42	5.9	0	0.17	0	1.024664	-3.951182	110	19	1.1	0.9;
	4713	1	-0	0	0	6.19	0	1.024941	-4.246073	220	6	1.1	0.9;
	4714	1	-5.06	-1.84	0	187.314047	0	1.0463	17.884755	380	5	1.1	0.9;
	4715	2	0	0	0	0	0	1.018631	-11.528817	20	4	1.1	0.9;
	4716	1	132	27	0	4.66	0	1.008588	11.184185	220	5	1.1	0.9;
	4717	1	89	-1.8	0	6.15	0	1.033766	7.242446	220	3	1.1	0.9;
	4718	2	0	0	0	0	0	1.045544	14.333931	20	18	1.1	0.9;
	4719	1	-0	0	0	1.18	0	0.977434	12.295699	380	5	1.1	0.9;
	4720	1	0	0	0	0	0	0.967808	-5.298475	150	9	1.1	0.9;
	4721	1	0	-0	0	3.03	0	1.019917	31.140334	380	3	1.1	0.9;
	4722	2	0	0	0	0	0	1.048656	21.505608	20	5	1.1	0.9;
	4723	2	0	0	0	0	0	1.048656	21.505608	20	5	1.1	0.9;
	4724	2	0	0	0	0	0	1.048656	21.505608	20	5	1.1	0.9;
	4725	2	0	0	0	0	0	1.048656	21.505608	20	5	1.1	0.9;
	4726	2	0	0	0	0	0	1.048656	21.505608	20	5	1.1	0.9;
	4727	2	0	0	0	0	0	1.048656	21.505608	20	5	1.1	0.9;
	4728	2	0	0	0	0	0	1.048656	21.505608	20	5	1.1	0.9;
	4729	2	0	0	0	0	0	1.048656	21.505608	20	5	1.1	0.9;
	4730	1	20.9	-171.5	0	0.21	0	1.064094	37.912708	220	5	1.1	0.9;
	4731	1	113	56.5	0	4.12	0	1.001491	-13.031617	150	9	1.1	0.9;
	4732	1	84.19	17.3	0	6.2	0	1.045332	19.973133	220	5	1.1	0.9;
	4733	2	0	0	0	0	0	1.043092	17.519475	20	5	1.1	0.9;
	4734	1	-0	0	0	17.65	0	1.013038	26.570765	380	3	1.1	0.9;
	4735	1	81.79	51.1	0	3.06	0	0.970438	49.384588	220	3	1.1	0.9;
	4736	1	142.4	87.8	0.4624	9.223933	0	1.062761	50.754899	220	13	1.1	0.9;
	4737	1	0	0	0	0	0	0.997804	1.009459	150	9	1.1	0.9;
	4738	2	0	0	0	0	0	1.043092	17.519475	20	5	1.1	0.9;
	4739	1	0	-0	0	2.19	0	1.014647	18.500995	154	1	1.1	0.9;
	4740	1	0	-0	0	45.7	0	1.017165	12.422282	380	3	1.1	0.9;
	4741	1	0	0	0	1.44	0	1.031913	8.483144	220	5	1.1	0.9;
	4742	1	0	0	0	1.87	0	1.024295	-24.350529	380	10	1.1	0.9;
	4743	1	0	0	0	3.71	0	0.998781	42.499423	220	8	1.1	0.9;
	4744	1	-11.71	3.8	0	0.13	0	1.002639	-10.11886	150	9	1.1	0.9;
	4745	2	0	0	0	0	0	1.04834	88.719592	20	3	1.1	0.9;
	4746	2	0	0	0	0	0	1.026568	1.551222	6.3	1	1.1	0.9;
	4747	1	-0	0	0	53.64	0	1.13198	-0.09724	380	15	1.2	0.8;
	4748	1	83.6	11.9	0	5.03	0	1.012174	29.444069	154	1	1.1	0.9;
	4749	1	87	29.6	0	4.95	0	1.009602	33.524231	220	3	1.1	0.9;
	4750	1	0	0	0.2166	10.9704	0	1.101029	-0.788511	380	1	1.2	0.8;
	4751	2	0	0	0	0	0	1.046359	90.322887	20	3	1.1	0.9;
	4752	2	0	0	0	0	0	1.032344	9.093698	20	18	1.1	0.9;
	4753	2	0	0	0	0	0	1.046359	90.322887	20	3	1.1	0.9;
	4754	1	9.21	3.5	0.037778	0.662222	0	1.059913	0.520524	110	19	1.1	0.9;
	4755	1	0	-0	0	23.92	0	1.017262	9.223881	380	1	1.1	0.9;
	4756	1	-0	-0	0.019362	1.140638	0	1.036476	43.598697	380	3	1.1	0.9;
	4757	2	0	0	0	0	0	1.027461	-17.241264	20	4	1.1	0.9;
	4758	1	0	0	0	0	0	0.963378	-9.81008	150	9	1.1	0.9;
	4759	1	0	-0	0	4.62	0	0.977808	-27.135545	220	10	1.1	0.9;
	4760	1	0	-0	0	0.23	0	1.099212	-4.426011	150	9	1.1	0.9;
	4761	1	-0	-0	0	0	0	1.020557	9.624602	20	14	1.1	0.9;
	4762	1	-0	-0	0	21.82	0	1.02719	47.01676	380	3	1.1	0.9;
	4763	2	0	0	0	0	0	1.024544	18.974877	20	14	1.1	0.9;
	4764	2	0	0	0	0	0	1.030832	-9.752509	20	4	1.1	0.9;
	4765	1	176.7	25.6	0	2.01	0	1.007992	-20.693888	220	6	1.1	0.9;
	4766	1	0	0	0	0	0	1.008391	-8.4759	150	9	1.1	0.9;
	4767	1	-0	-0	0	5.12	0	1.004933	30.37899	380	3	1.1	0.9;
	4768	1	-0	-0	0	2.13	0	1.01753	29.670296	154	1	1.1	0.9;
	4769	1	-0	0	0.668421	4.480834	0	1.082014	9.78954	380	18	1.2	0.8;
	4770	2	0	0	0	0	0	1.023397	-17.3231	20	6	1.1	0.9;
	4771	2	0	0	0	0	0	1.022819	-4.100145	20	6	1.1	0.9;
	4772	2	0	0	0	0	0	1.023397	-17.3231	20	6	1.1	0.9;
	4773	2	0	0	0	0	0	1.023397	-17.3231	20	6	1.1	0.9;
	4774	2	0	0	0	0	0	1.026206	-2.974992	20	6	1.1	0.9;
	4775	2	0	0	0	0	0	0.993383	32.126146	11	1	1.1	0.9;
	4776	1	165.55	72.35	0.039777	8.820223	0	1.025857	24.243802	220	5	1.1	0.9;
	4777	2	0	0	0	0	0	1.050901	18.201002	20	7	1.1	0.9;
	4778	1	0	-0	0.417783	38.745174	0	1.05955	-4.540924	380	11	1.1	0.9;
	4779	1	-0	0	0	17.75	0	1.036248	-15.392834	380	4	1.1	0.9;
	4780	1	0	-0	0	0.13	0	1.033888	21.355633	220	5	1.1	0.9;
	4781	1	0	-0	0	0	0	1.039156	19.470405	20	5	1.1	0.9;
	4782	1	-0	-0	0.294847	40.776932	0	1.039848	12.278319	380	7	1.1	0.9;
	4783	1	66.67	33.3	0.059554	3.910446	0	1.010141	-0.21068	220	7	1.1	0.9;
	4784	1	-0	-0	0	0	0	1.053999	28.580111	150	2	1.1	0.9;
	4785	1	-0	-0	0	2.08	0	1.028555	23.328029	150	8	1.1	0.9;
	4786	2	0	0	0	0	0	1.000057	26.10357	11	1	1.1	0.9;
	4787	1	10.61	3.8	0	0.66	0	0.991747	4.829091	110	19	1.1	0.9;
	4788	1	1.8	0.7	0	0.85	0	1.019129	-0.189136	110	19	1.1	0.9;
	4789	1	-0	0	0	9.66	0	1.026375	-29.790551	220	4	1.1	0.9;
	4790	1	180.7	71.8	0	2.46	0	1.005495	20.92235	220	5	1.1	0.9;
	4791	2	0	0	0	0	0	1.031502	4.663459	20	6	1.1	0.9;
	4792	2	0	0	0	0	0	1.031502	4.663459	20	6	1.1	0.9;
	4793	2	0	0	0	0	0	1.011478	-0.639289	20	10	1.1	0.9;
	4794	1	-0	0	0.118845	-0.089134	0	1.015411	-1.60616	380	6	1.1	0.9;
	4795	1	0	-0	0.470875	11.647224	0	1.027302	3.535768	380	6	1.1	0.9;
	4796	2	0	0	0	0	0	1.034961	34.914591	11	1	1.1	0.9;
	4797	1	178.62	14.44	0	136.23	0	1.096479	34.187198	380	14	1.2	0.8;
	4798	1	-0	-0	0	16.34	0	1.03203	35.148877	380	3	1.1	0.9;
	4799	1	27.8	4	0	9.38	0	0.990537	37.906966	154	1	1.1	0.9;
	4800	1	-0.57	-0.91	0	0.09	0	1.025342	0.44206	220	21	1.1	0.9;
	4801	1	11.01	6	0	0.04	0	1.002965	-9.338003	150	9	1.1	0.9;
	4802	1	0	-0	0.113921	102.294559	0	1.073833	35.411481	380	14	1.1	0.9;
	4803	1	0.6	28.7	0.298623	22.851147	0	1.022039	-17.229409	220	4	1.1	0.9;
	4804	1	0	-0	0	84.85	0	0.961385	28.501592	400	1	1.1	0.9;
	4805	1	33.14	11.6	0	4.51	0	1.029498	19.437343	220	5	1.1	0.9;
	4806	1	5.51	2	0	0.39	0	1.022695	16.114493	110	19	1.1	0.9;
	4807	1	0	-0	0	30.55	0	1.099343	-14.359026	150	9	1.2	0.8;
	4808	1	-0	-0	0	1.08	0	1.053701	-5.383885	150	9	1.1	0.9;
	4809	2	0	0	0	0	0	1.040009	86.062395	20	3	1.1	0.9;
	4810	2	0	0	0	0	0	1.040009	86.062395	20	3	1.1	0.9;
	4811	2	0	0	0	0	0	1.040009	86.062395	20	3	1.1	0.9;
	4812	1	-0	0	0	0.8	0	1.00402	-8.963428	150	9	1.1	0.9;
	4813	1	22.12	13.7	0.113069	1.716931	0	1.079671	-3.428225	150	9	1.1	0.9;
	4814	2	0	0	0	0	0	1.06019	67.358874	20	3	1.1	0.9;
	4815	1	332.9	99.6	0	0.53	0	1.004083	30.252992	150	8	1.1	0.9;
	4816	1	33.24	11.1	0	0.27	0	0.962276	-9.765757	110	21	1.1	0.9;
	4817	1	72.08	29.5	0	1.06	0	1.028471	23.568416	150	8	1.1	0.9;
	4818	1	34.9	5	0	11.17	0	0.985623	34.062103	154	1	1.1	0.9;
	4819	1	-0	0	0	3.25	0	1.028138	17.58447	220	3	1.1	0.9;
	4820	1	0	0	0	10.72	0	0.967719	40.671168	400	1	1.1	0.9;
	4821	1	13.9	1.9	0	3.4	0	1.02665	-3.135251	154	1	1.1	0.9;
	4822	1	0	0	0.573452	-0.326818	0	1.057748	38.103723	220	5	1.1	0.9;
	4823	2	0	0	0	0	0	1.014332	26.817672	10.5	1	1.1	0.9;
	4824	1	0	0	0	1.84	0	0.996149	-9.146124	150	9	1.1	0.9;
	4825	1	0	0	0	7.75	0	1.052729	7.447866	380	1	1.1	0.9;
	4826	1	149.3	-2	0	20.77	0	1.038011	34.413286	380	2	1.1	0.9;
	4827	1	0	0	0.039049	0.520475	0	0.993635	13.821924	110	18	1.1	0.9;
	4828	1	-0	-0	0	5.18	0	1.069027	17.921154	220	3	1.1	0.9;
	4829	1	-10.11	3.3	0	0.53	0	1.0342	0.003159	150	9	1.1	0.9;
	4830	1	0	0	0	34.49	0	1.0461	10.881862	380	1	1.1	0.9;
	4831	1	5.61	2.2	0.195552	45.454448	0	1.02548	6.087271	380	7	1.1	0.9;
	4832	1	-0.2	-1.64	0	0	0	1.01757	30.154601	150	8	1.1	0.9;
	4833	1	31.23	11.3	0	0.33	0	1.014184	-5.288127	110	19	1.1	0.9;
	4834	1	0	-0	0	0.05	0	1.09564	-4.636709	150	9	1.1	0.9;
	4835	1	59.7	26.3	0	37.680934	0	1.078887	-2.523292	380	9	1.1	0.9;
	4836	1	50.56	16.6	0	2.02	0	0.97179	-11.726386	110	12	1.1	0.9;
	4837	2	0	0	0	0	0	1.026243	6.043442	20	6	1.1	0.9;
	4838	1	0	0	0	0	0	1.01493	-10.961932	150	9	1.1	0.9;
	4839	1	120.4	39.5	0	47.2	0	1.012555	12.021501	380	3	1.1	0.9;
	4840	1	0	0	0	0.83	0	1.063837	-4.873875	150	9	1.1	0.9;
	4841	1	-0	0	0	2.22	0	1.024755	36.25485	220	3	1.1	0.9;
	4842	1	78.19	27	0.058553	4.621206	0	1.01309	92.614678	150	1	1.1	0.9;
	4843	2	0	0	0	0	0	1.026243	6.043442	20	6	1.1	0.9;
	4844	2	0	0	0	0	0	1.026243	6.043442	20	6	1.1	0.9;
	4845	2	0	0	0	0	0	1.026243	6.043442	20	6	1.1	0.9;
	4846	2	0	0	0	0	0	1.026243	6.043442	20	6	1.1	0.9;
	4847	2	0	0	0	0	0	1.026243	6.043442	20	6	1.1	0.9;
	4848	1	70.48	16.4	0.040466	4.649534	0	0.998611	13.238352	220	5	1.1	0.9;
	4849	1	-0	-0	0	0.91	0	1.025189	10.345214	220	5	1.1	0.9;
	4850	1	219.22	70.1	0.685037	32.787679	0	1.0435	33.060974	220	13	1.1	0.9;
	4851	1	-0	-0	0	2.27	0	1.020012	17.086449	220	5	1.1	0.9;
	4852	1	14.72	4.6	0	0.33	0	0.996243	-1.910178	110	18	1.1	0.9;
	4853	1	2.3	0.8	0	0.6	0	0.955883	-1.464253	110	19	1.1	0.9;
	4854	1	13.62	8.4	0	2.67	0	1.041421	-3.450344	150	9	1.1	0.9;
	4855	2	0	0	0	0	0	1.060956	44.39088	20	8	1.1	0.9;
	4856	2	0	0	0	0	0	1.060956	44.39088	20	8	1.1	0.9;
	4857	2	0	0	0	0	0	1.062536	46.73809	20	8	1.1	0.9;
	4858	2	0	0	0	0	0	1.062536	46.73809	20	8	1.1	0.9;
	4859	2	0	0	0	0	0	1.059942	43.353318	20	8	1.1	0.9;
	4860	1	-10.59	-5.25	0	3.59	0	1.025165	64.621612	220	3	1.1	0.9;
	4861	2	0	0	0	0	0	1.05948	38.842339	20	8	1.1	0.9;
	4862	2	0	0	0	0	0	1.05948	38.842339	20	8	1.1	0.9;
	4863	1	34.04	11	0	0.96	0	0.997923	-14.24707	110	22	1.1	0.9;
	4864	2	0	0	0	0	0	1.059942	43.353318	20	8	1.1	0.9;
	4865	1	120.5	37.5	0.039658	3.410342	0	1.020612	-10.811239	220	4	1.1	0.9;
	4866	1	67.4	9.6	0	4.68	0	0.992519	17.53406	154	1	1.1	0.9;
	4867	1	48.8	6.9	0	4.64	0	1.016484	18.987706	154	1	1.1	0.9;
	4868	1	236.2	116.4	0	16.22	0	0.980935	-5.899348	220	6	1.1	0.9;
	4869	1	36.84	14.1	0.048813	44.261187	0	1.046186	-11.388611	380	4	1.1	0.9;
	4870	2	0	0	0	0	0	1.025271	20.88058	20	5	1.1	0.9;
	4871	1	0	0	0	0.76	0	1.067428	22.702865	220	3	1.1	0.9;
	4872	1	379.5	113.3	0.06172	22.14828	0	0.991109	16.535153	220	14	1.1	0.9;
	4873	1	46.85	16.6	0	2.08	0	0.986934	-15.934332	150	10	1.1	0.9;
	4874	1	2	0.6	0	0.53	0	0.993737	-4.899565	110	21	1.1	0.9;
	4875	1	18.02	6	0	0.98	0	1.005551	-5.201173	110	21	1.1	0.9;
	4876	1	38.54	49.6	0.276138	0.922896	0	1.019487	-18.538846	220	4	1.1	0.9;
	4877	1	379.9	73	0	59.58	0	1.012173	-17.563659	380	6	1.1	0.9;
	4878	1	83.29	24.1	0	1.23	0	1.025261	12.50452	220	5	1.1	0.9;
	4879	1	-0	-0	0	0	0	0.991951	-1.701584	110	18	1.1	0.9;
	4880	1	0	-0	0.040009	346.717844	0	1.00528	17.378627	380	5	1.1	0.9;
	4881	1	38.04	20	0	0.03	0	1.013234	5.571068	220	6	1.1	0.9;
	4882	1	0	-0	0	0.68	0	1.012673	-2.175699	380	6	1.1	0.9;
	4883	1	0	0	0.114098	0.674918	0	1.05045	93.240906	380	3	1.1	0.9;
	4884	1	1.3	0.4	0	0.24	0	0.986687	-3.572455	110	18	1.1	0.9;
	4885	1	20.32	8.9	0	1.29	0	0.96819	-4.809229	110	18	1.1	0.9;
	4886	1	-0	-0	0	5.22	0	1.019326	37.994898	380	8	1.1	0.9;
	4887	2	0	0	0	0	0	1.01898	27.302632	20	6	1.1	0.9;
	4888	2	0	0	0	0	0	1.01898	27.302632	20	6	1.1	0.9;
	4889	1	-0	0	0	1.02	0	1.027266	14.054547	220	5	1.1	0.9;
	4890	1	0	0	0	0	0	1.033858	-10.849314	150	9	1.1	0.9;
	4891	1	-0	0	0	5.14	0	1.01396	22.693254	220	1	1.1	0.9;
	4892	2	0	0	0	0	0	1.024711	-14.053311	20	4	1.1	0.9;
	4893	2	0	0	0	0	0	1.024711	-14.053311	20	4	1.1	0.9;
	4894	1	-0	-0	0.185125	56.232473	0	1.075595	5.528103	380	18	1.2	0.8;
	4895	1	36.54	27.4	0	0.16	0	0.98599	37.658867	220	3	1.1	0.9;
	4896	2	0	0	0	0	0	1.024711	-14.053311	20	4	1.1	0.9;
	4897	2	0	0	0	0	0	1.024711	-14.053311	20	4	1.1	0.9;
	4898	1	-0	-0	0	0.82	0	0.997952	-4.752719	110	18	1.1	0.9;
	4899	1	-0	0	0	6.1	0	1.027074	3.213132	380	6	1.1	0.9;
	4900	1	0	0	0	7.68	0	1.014709	31.598072	380	3	1.1	0.9;
	4901	1	-0	-0	0	0	0	1.000465	12.056714	20	5	1.1	0.9;
	4902	2	0	0	0	0	0	1.045936	19.626816	20	5	1.1	0.9;
	4903	2	0	0	0	0	0	1.045936	19.626816	20	5	1.1	0.9;
	4904	2	0	0	0	0	0	1.058505	17.418274	20	5	1.1	0.9;
	4905	2	0	0	0	0	0	1.058505	17.418274	20	5	1.1	0.9;
	4906	1	155.8	38.4	0	3.77	0	0.99184	26.455293	150	8	1.1	0.9;
	4907	1	0	0	0.089212	-0.009474	0	1.043661	17.14354	220	5	1.1	0.9;
	4908	1	-0	0	0.090774	0.080444	0	1.056764	-0.4217	220	18	1.1	0.9;
	4909	1	3.8	1.3	0	0.99	0	0.982654	-4.138319	110	18	1.1	0.9;
	4910	2	0	0	0	0	0	1.040138	32.45384	20	17	1.1	0.9;
	4911	1	143.79	106.96	0	99.84	0	1.042194	18.137057	380	13	1.1	0.9;
	4912	1	-0	0	0	0.19	0	0.966719	-9.648099	150	9	1.1	0.9;
	4913	1	3.4	1.4	0	0.62	0	0.979117	-2.898026	110	18	1.1	0.9;
	4914	1	-0	-0	0	0	0	1.050455	44.07457	380	8	1.1	0.9;
	4915	1	0	0	0.039148	4.840852	0	1.009725	-8.960644	220	22	1.1	0.9;
	4916	1	0	-0	0	1.58	0	0.995163	19.315344	220	5	1.1	0.9;
	4917	2	0	0	0	0	0	0.989071	33.252478	6.3	1	1.1	0.9;
	4918	2	0	0	0	0	0	0.943436	50.691637	13.8	1	1.1	0.9;
	4919	1	0.5	0.5	0	0.74	0	1.011297	2.419482	110	18	1.1	0.9;
	4920	1	156.2	24.5	0	2.22	0	1.017154	35.686623	150	8	1.1	0.9;
	4921	1	322.1	131.9	0.241556	9.808768	0	1.017104	29.53828	220	14	1.1	0.9;
	4922	2	0	0	0	0	0	1.04826	-8.707987	20	10	1.1	0.9;
	4923	1	15.22	2.5	0	0.63	0	0.935533	-8.058668	110	18	1.1	0.9;
	4924	2	0	0	0	0	0	1.025724	-14.139236	20	4	1.1	0.9;
	4925	2	0	0	0	0	0	1.017601	-9.8258	20	10	1.1	0.9;
	4926	2	0	0	0	0	0	1.036433	-0.581009	20	2	1.1	0.9;
	4927	1	77.49	21.4	0	0.05	0	1.026041	18.668727	220	5	1.1	0.9;
	4928	2	0	0	0	0	0	1.04876	45.73683	20	10	1.1	0.9;
	4929	1	-0	0	0	13.02	0	1.02234	-17.877403	220	4	1.1	0.9;
	4930	1	13.9	1.9	0	5.68	0	1.016405	28.05316	154	1	1.1	0.9;
	4931	2	0	0	0	0	0	1.037501	2.321125	20	2	1.1	0.9;
	4932	2	0	0	0	0	0	1.037501	2.321125	20	2	1.1	0.9;
	4933	1	50.76	17.5	0	8.97	0	1.002517	0.43526	220	6	1.1	0.9;
	4934	2	0	0	0	0	0	1.037973	11.471538	20	14	1.1	0.9;
	4935	1	50.06	19.7	0	0.84	0	1.04852	-4.374565	110	23	1.1	0.9;
	4936	1	8.31	3	0	0.69	0	0.953356	-2.428855	110	19	1.1	0.9;
	4937	1	115	-4.5	0	15.05	0	1.033465	-19.461167	220	4	1.1	0.9;
	4938	2	0	0	0	0	0	1.027032	8.891931	20	14	1.1	0.9;
	4939	2	0	0	0	0	0	1.03611	8.921216	20	23	1.1	0.9;
	4940	1	47.15	15.2	0	0.17	0	0.964812	-11.436007	110	12	1.1	0.9;
	4941	1	-0	0	0	5.7	0	1.008296	29.565359	380	3	1.1	0.9;
	4942	2	0	0	0	0	0	1.038474	82.561969	20	3	1.1	0.9;
	4943	2	0	0	0	0	0	1.038474	82.561969	20	3	1.1	0.9;
	4944	1	0	-0	0	30.28	0	1.055417	39.131981	380	3	1.1	0.9;
	4945	2	0	0	0	0	0	1.042258	67.377104	20	3	1.1	0.9;
	4946	2	0	0	0	0	0	1.042258	67.377104	20	3	1.1	0.9;
	4947	1	4.61	1.7	0	0.19	0	1.017976	-2.931441	110	19	1.1	0.9;
	4948	1	-0	-0	0	19.57	0	1.02073	-26.16821	220	4	1.1	0.9;
	4949	2	0	0	0	0	0	1.01601	-0.145046	20	10	1.1	0.9;
	4950	1	-0.2	-1.64	0	0	0	1.01757	30.154601	150	8	1.1	0.9;
	4951	1	-0	-0	0.060584	0.479416	0	1.018559	23.073382	220	14	1.1	0.9;
	4952	2	0	0	0	0	0	1.03281	29.803723	20	10	1.1	0.9;
	4953	1	91.8	18.6	0.05911	1.87089	0	1.030072	19.446828	220	5	1.1	0.9;
	4954	1	88.8	-3.2	0	13.63	0	1.051449	3.601764	220	5	1.1	0.9;
	4955	1	93.7	29.9	0	2.1	0	1.008586	35.815607	110	8	1.1	0.9;
	4956	1	-0	0	0	4.79	0	1.0168	34.173569	220	3	1.1	0.9;
	4957	1	-0	-0	0	0.67	0	1.001915	-10.138862	150	9	1.1	0.9;
	4958	2	0	0	0	0	0	1.025082	-16.616549	20	6	1.1	0.9;
	4959	2	0	0	0	0	0	1.024274	-9.95923	20	4	1.1	0.9;
	4960	1	27.13	9	0	0.42	0	0.991219	-5.133534	110	21	1.1	0.9;
	4961	1	0	0	0	0.54	0	1.002673	-10.168847	150	9	1.1	0.9;
	4962	1	-0	-0	0	0	0	1.021606	-18.065799	20	4	1.1	0.9;
	4963	2	0	0	0	0	0	1.019736	17.906487	20	6	1.1	0.9;
	4964	2	0	0	0	0	0	1.019736	17.906487	20	6	1.1	0.9;
	4965	1	103.1	14.7	0	4.43	0	1.005748	25.955736	154	1	1.1	0.9;
	4966	1	-0	-0	0	17.31	0	1.034413	4.784259	220	5	1.1	0.9;
	4967	2	0	0	0	0	0	1.013686	-8.216227	20	4	1.1	0.9;
	4968	1	-0	-0	0	0	0	1.021606	-18.065799	20	4	1.1	0.9;
	4969	1	23.23	6.3	0	0.66	0	1.008908	13.390577	220	5	1.1	0.9;
	4970	1	115	16.4	0	3.94	0	1.028476	2.26797	154	1	1.1	0.9;
	4971	1	234.2	33.3	0	0.88	0	1.035056	1.123153	154	1	1.1	0.9;
	4972	1	140.93	18.71	0	12.73	0	1.034173	26.119223	220	13	1.1	0.9;
	4973	1	-0.76	-0.86	0	6.15	0	1.003912	-6.9049	220	12	1.1	0.9;
	4974	1	-0	-0	0	13.94	0	1.038699	17.007141	220	5	1.1	0.9;
	4975	1	108.7	26.3	0	2.06	0	1.011379	5.553053	220	7	1.1	0.9;
	4976	1	0	0	0	0	0	0.995518	-13.198494	150	9	1.1	0.9;
	4977	1	0	-0	0	16.4	0	0.997271	-10.738639	380	1	1.1	0.9;
	4978	1	52.06	17	0	0.24	0	0.985966	-4.894497	110	21	1.1	0.9;
	4979	1	29.43	10.7	0	0.2	0	0.984412	-2.691034	110	19	1.1	0.9;
	4980	1	-0	-0	0	0	0	1.033951	32.215255	220	5	1.1	0.9;
	4981	1	19.42	7	0	0.6	0	0.972225	-21.169989	150	10	1.1	0.9;
	4982	2	0	0	0	0	0	1.01458	18.108546	20	6	1.1	0.9;
	4983	2	0	0	0	0	0	0.995619	18.792566	6.3	1	1.1	0.9;
	4984	1	78.09	27	0	0.83	0	1.01978	-13.417385	110	22	1.1	0.9;
	4985	2	0	0	0	0	0	1.073911	4.277217	20	20	1.1	0.9;
	4986	1	14.72	5.3	0	0.58	0	1.015712	-3.141636	110	19	1.1	0.9;
	4987	2	0	0	0	0	0	1.020265	-12.139526	20	4	1.1	0.9;
	4988	2	0	0	0	0	0	1.022099	17.218586	20	14	1.1	0.9;
	4989	1	0	0	0	1.78	0	0.980508	-10.597685	150	9	1.1	0.9;
	4990	1	0	-0	0	0.24	0	1.065886	-4.560143	150	9	1.1	0.9;
	4991	1	1.9	0.7	0	0.19	0	1.012934	-4.995932	110	19	1.1	0.9;
	4992	1	6.71	2.4	0	0.2	0	1.015836	-3.196121	110	19	1.1	0.9;
	4993	1	11.51	4.2	0	3.64	0	0.99786	-1.75859	110	19	1.1	0.9;
	4994	1	24.23	6.4	0	3.9	0	1.014314	9.813801	220	5	1.1	0.9;
	4995	1	173.1	4.6	0	2.57	0	1.061594	9.387569	220	3	1.1	0.9;
	4996	1	11.5	1.7	0	3.11	0	0.966778	18.754357	154	1	1.1	0.9;
	4997	1	-0	-0	0.171753	9.136873	0	1.046062	93.758969	380	3	1.1	0.9;
	4998	2	0	0	0	0	0	1.019231	22.327173	20	24	1.1	0.9;
	4999	1	0	0	1.191402	16.470917	0	1.002629	28.069895	380	5	1.1	0.9;
	5000	2	0	0	0	0	0	1.008736	24.556899	11	1	1.1	0.9;
	5001	1	146.9	26.6	0	11.95	0	1.014507	-26.601129	220	4	1.1	0.9;
	5002	1	0	0	0	17.92	0	1.02876	-17.22899	380	4	1.1	0.9;
	5003	1	90.2	5.8	0	18.69	0	1.039585	62.096018	380	3	1.1	0.9;
	5004	1	77.19	18.6	0	2.22	0	1.007368	14.66858	220	5	1.1	0.9;
	5005	1	-0	0	0.113531	6.215391	0	1.035825	92.435601	220	3	1.1	0.9;
	5006	1	39.05	-11.82	0	1.12	0	1.030812	9.540266	220	7	1.1	0.9;
	5007	2	0	0	0	0	0	1.035406	21.74444	20	5	1.1	0.9;
	5008	2	0	0	0	0	0	1.035406	21.74444	20	5	1.1	0.9;
	5009	2	0	0	0	0	0	1.035406	21.74444	20	5	1.1	0.9;
	5010	1	179.8	5.67	0	12.83	0	0.989794	-3.083252	220	5	1.1	0.9;
	5011	1	314.5	83.9	0.710505	9.997517	0	1.029063	15.621241	220	5	1.1	0.9;
	5012	2	0	0	0	0	0	1.032132	-3.910762	20	6	1.1	0.9;
	5013	2	0	0	0	0	0	1.032132	-3.910762	20	6	1.1	0.9;
	5014	1	0	0	0	0.04	0	1.009216	-1.923247	110	19	1.1	0.9;
	5015	1	140.1	10	0.0386	8.2714	0	1.041524	28.972318	220	13	1.1	0.9;
	5016	1	9.7	1.4	0	4.51	0	1.000468	19.917128	154	1	1.1	0.9;
	5017	2	0	0	0	0	0	1.032238	-4.146012	20	6	1.1	0.9;
	5018	2	0	0	0	0	0	1.032238	-4.146012	20	6	1.1	0.9;
	5019	1	0	0	0	80.21	0	1.059485	4.79611	380	19	1.1	0.9;
	5020	1	0	0	0.29461	7.926671	0	1.057905	7.307607	220	18	1.2	0.8;
	5021	1	54.96	12.2	0	7.05	0	1.016921	-26.232106	220	4	1.1	0.9;
	5022	1	0	0	0	0.74	0	1.003914	29.484531	150	8	1.1	0.9;
	5023	2	0	0	0	0	0	1.052394	33.692026	20	10	1.1	0.9;
	5024	1	5.11	1.3	0	0.67	0	0.98053	-5.939116	110	18	1.1	0.9;
	5025	2	0	0	0	0	0	1.050987	40.998977	20	10	1.1	0.9;
	5026	1	-7.63	-4.3	0	94.012267	0	0.993198	26.13824	380	5	1.1	0.9;
	5027	1	271.43	74.55	0	78.4	0	1.043019	20.310505	380	14	1.1	0.9;
	5028	1	0	-0	0	6.02	0	1.034668	17.194328	220	5	1.1	0.9;
	5029	2	0	0	0	0	0	1.005112	-1.198328	20	18	1.1	0.9;
	5030	1	-0	0	0	2.24	0	1.021232	26.399346	150	8	1.1	0.9;
	5031	1	0	0	0	0	0	1.038979	-13.937788	20	4	1.1	0.9;
	5032	1	137.4	23.9	0	7.26	0	1.032656	9.024375	220	5	1.1	0.9;
	5033	1	20.82	9.5	0	1.16	0	0.956433	1.124618	110	18	1.1	0.9;
	5034	2	0	0	0	0	0	1.019932	23.482162	20	10	1.1	0.9;
	5035	1	15.63	4.8	0	1.21	0	0.949685	-7.51879	110	18	1.1	0.9;
	5036	2	0	0	0	0	0	1.038011	31.920163	20	10	1.1	0.9;
	5037	2	0	0	0	0	0	1.030988	2.394145	20	9	1.1	0.9;
	5038	2	0	0	0	0	0	1.043059	9.285297	20	23	1.1	0.9;
	5039	1	-0	-0	0	33.22	0	1.032774	20.844648	380	13	1.1	0.9;
	5040	1	-18.8	-5.9	0	6.23	0	1.03749	-6.683276	150	9	1.1	0.9;
	5041	1	25.63	51.6	0	2.79	0	1.019746	-12.030982	220	4	1.1	0.9;
	5042	1	23.13	8.8	0	1.42	0	1.018079	-13.446069	110	22	1.1	0.9;
	5043	1	4.2	0.6	0	4.02	0	0.994733	38.659408	154	1	1.1	0.9;
	5044	1	76.6	10.9	0	9.96	0	1.031478	18.798826	154	1	1.1	0.9;
	5045	1	0	0	0	1.38	0	1.047292	-2.403002	150	9	1.1	0.9;
	5046	1	-0	0	0.039816	0.010184	0	1.033166	16.613981	220	5	1.1	0.9;
	5047	1	119.1	60.5	0	2	0	1.020262	-10.694863	150	9	1.1	0.9;
	5048	1	43.75	17.3	0	7.73	0	1.019623	31.518845	150	8	1.1	0.9;
	5049	1	209.5	95	0	24.39	0	1.017537	36.998934	150	8	1.1	0.9;
	5050	2	0	0	0	0	0	0.987146	20.677409	20	5	1.1	0.9;
	5051	2	0	0	0	0	0	1.035412	11.75856	20	5	1.1	0.9;
	5052	2	0	0	0	0	0	1.035412	11.75856	20	5	1.1	0.9;
	5053	2	0	0	0	0	0	1.035412	11.75856	20	5	1.1	0.9;
	5054	2	0	0	0	0	0	1.05667	-2.722877	20	5	1.1	0.9;
	5055	2	0	0	0	0	0	1.05667	-2.722877	20	5	1.1	0.9;
	5056	2	0	0	0	0	0	1.05667	-2.722877	20	5	1.1	0.9;
	5057	2	0	0	0	0	0	1.05667	-2.722877	20	5	1.1	0.9;
	5058	2	0	0	0	0	0	1.049254	8.363188	20	2	1.1	0.9;
	5059	2	0	0	0	0	0	1.083018	11.899545	20	2	1.1	0.9;
	5060	2	0	0	0	0	0	1.083018	11.899545	20	2	1.1	0.9;
	5061	2	0	0	0	0	0	0.994858	32.812983	15	1	1.1	0.9;
	5062	2	0	0	0	0	0	1.056742	8.416383	20	2	1.1	0.9;
	5063	2	0	0	0	0	0	1.056742	8.416383	20	2	1.1	0.9;
	5064	2	0	0	0	0	0	1.056742	8.416383	20	2	1.1	0.9;
	5065	2	0	0	0	0	0	1.049254	8.363188	20	2	1.1	0.9;
	5066	1	7.71	2.2	0	0.81	0	1.000777	-7.436122	110	22	1.1	0.9;
	5067	1	1.5	0.5	0	0.59	0	1.019774	-9.023427	110	12	1.1	0.9;
	5068	1	3.8	1.3	0	0.65	0	0.998339	6.169569	110	18	1.1	0.9;
	5069	2	0	0	0	0	0	0.995797	13.295176	20	5	1.1	0.9;
	5070	2	0	0	0	0	0	0.999202	27.044594	20	5	1.1	0.9;
	5071	2	0	0	0	0	0	0.995797	13.295176	20	5	1.1	0.9;
	5072	2	0	0	0	0	0	0.995797	13.295176	20	5	1.1	0.9;
	5073	2	0	0	0	0	0	0.995797	13.295176	20	5	1.1	0.9;
	5074	2	0	0	0	0	0	0.995797	13.295176	20	5	1.1	0.9;
	5075	1	-0	0	0	0.27	0	1.015605	-7.523526	110	1	1.1	0.9;
	5076	2	0	0	0	0	0	1.049756	20.130595	20	5	1.1	0.9;
	5077	2	0	0	0	0	0	1.049756	20.130595	20	5	1.1	0.9;
	5078	2	0	0	0	0	0	1.049756	20.130595	20	5	1.1	0.9;
	5079	1	123.1	51.8	0	1.68	0	0.997034	-2.86835	110	21	1.1	0.9;
	5080	1	18.72	5.5	0	1.15	0	1.020965	-7.890322	110	22	1.1	0.9;
	5081	1	18.02	-15.7	0	1.03	0	1.045031	-0.976333	110	23	1.1	0.9;
	5082	2	0	0	0	0	0	1.029323	-7.468239	20	6	1.1	0.9;
	5083	1	0	0	0	0	0	1.003491	-10.282006	150	9	1.1	0.9;
	5084	2	0	0	0	0	0	1.06339	20.87235	20	14	1.1	0.9;
	5085	2	0	0	0	0	0	1.041566	-8.637491	20	6	1.1	0.9;
	5086	2	0	0	0	0	0	1.041566	-8.637491	20	6	1.1	0.9;
	5087	1	23.93	8.7	0	0.52	0	1.023477	-4.753253	110	19	1.1	0.9;
	5088	1	-0	-0	0	64.67	0	0.95519	43.97653	400	1	1.1	0.9;
	5089	2	0	0	0	0	0	1.039883	-8.632286	20	6	1.1	0.9;
	5090	2	0	0	0	0	0	1.039883	-8.632286	20	6	1.1	0.9;
	5091	2	0	0	0	0	0	1.041566	-8.637491	20	6	1.1	0.9;
	5092	2	0	0	0	0	0	1.041566	-8.637491	20	6	1.1	0.9;
	5093	1	22.12	7.9	0	0.7	0	0.987249	-1.003496	110	19	1.1	0.9;
	5094	1	-3.4	-0.5	0	0.23	0	0.997504	-17.119751	220	10	1.1	0.9;
	5095	1	124.9	104	0	4.31	0	0.990391	33.947064	220	3	1.1	0.9;
	5096	1	-0	-0	0.236401	-0.027001	0	1.028111	19.436473	220	6	1.1	0.9;
	5097	1	142.5	23.6	0	4.24	0	1.020692	9.229442	220	5	1.1	0.9;
	5098	1	22.02	13.7	0	0.44	0	0.964149	-15.086645	150	9	1.1	0.9;
	5099	2	0	0	0	0	0	1.015374	8.632692	20	18	1.1	0.9;
	5100	1	3.9	-35.8	0	4.71	0	1.019657	-9.251983	220	10	1.1	0.9;
	5101	1	0	-0	0	37.1	0	1.031272	39.234703	380	8	1.1	0.9;
	5102	1	4.61	1.4	0	0.83	0	0.980589	1.375059	110	18	1.1	0.9;
	5103	1	16.32	7.6	0	3.15	0	1.018365	-12.87646	220	4	1.1	0.9;
	5104	1	36.34	22.5	0	2.39	0	0.996198	-5.152317	150	9	1.1	0.9;
	5105	1	55.66	4.5	0	1.33	0	1.027523	18.894424	220	5	1.1	0.9;
	5106	1	-0	0	0	75	0	1.005261	28.683345	150	8	1.1	0.9;
	5107	1	-0	-0	0	12.24	0	1.015718	-26.178986	220	4	1.1	0.9;
	5108	1	12.51	4.1	0	0.71	0	0.982029	-11.79148	110	12	1.1	0.9;
	5109	2	0	0	0	0	0	1.048748	7.516157	20	14	1.1	0.9;
	5110	1	15.92	5.2	0	0.58	0	0.997358	-13.357973	110	12	1.1	0.9;
	5111	2	0	0	0	0	0	0.996117	32.042491	11	1	1.1	0.9;
	5112	1	473.4	113.1	0.058293	10.341707	0	1.032187	20.798472	220	5	1.1	0.9;
	5113	1	12.11	4.7	0.020109	0.529891	0	0.985135	-2.653308	110	19	1.1	0.9;
	5114	1	226.3	61.1	0	0	0	1.036024	25.906454	150	2	1.1	0.9;
	5115	1	-0	-0	0.243936	107.503604	0	1.027659	8.695729	380	7	1.1	0.9;
	5116	2	0	0	0	0	0	1.013887	22.6309	11.5	1	1.1	0.9;
	5117	1	-0	0	0	4.8	0	1.123141	-2.325257	380	23	1.2	0.8;
	5118	1	99.3	23.3	0	1.88	0	1.030789	9.190472	220	5	1.1	0.9;
	5119	1	100.5	32	0	1.89	0	1.028814	31.142822	150	8	1.1	0.9;
	5120	1	138.5	19.3	0	5.26	0	1.041517	19.441981	220	5	1.1	0.9;
	5121	1	0	-0	0	2.34	0	1.024166	5.104934	220	5	1.1	0.9;
	5122	2	0	0	0	0	0	1.030744	-0.382174	20	5	1.1	0.9;
	5123	2	0	0	0	0	0	1.043216	17.315526	20	5	1.1	0.9;
	5124	1	94.2	21.6	0	2.88	0	1.017243	-25.297769	220	4	1.1	0.9;
	5125	1	0	0	0	7.06	0	1.069866	58.675575	220	8	1.1	0.9;
	5126	2	0	0	0	0	0	1.045141	25.131975	20	5	1.1	0.9;
	5127	2	0	0	0	0	0	1.045141	25.131975	20	5	1.1	0.9;
	5128	2	0	0	0	0	0	1.030744	-0.382174	20	5	1.1	0.9;
	5129	2	0	0	0	0	0	1.030744	-0.382174	20	5	1.1	0.9;
	5130	2	0	0	0	0	0	1.030744	-0.382174	20	5	1.1	0.9;
	5131	2	0	0	0	0	0	1.030744	-0.382174	20	5	1.1	0.9;
	5132	1	0	-0	0	1.93	0	1.031995	18.369092	220	5	1.1	0.9;
	5133	2	0	0	0	0	0	1.030744	-0.382174	20	5	1.1	0.9;
	5134	2	0	0	0	0	0	1.035888	21.491029	20	5	1.1	0.9;
	5135	2	0	0	0	0	0	1.026027	-7.002651	20	4	1.1	0.9;
	5136	1	0	-0	0	3.49	0	1.006075	9.474182	220	19	1.1	0.9;
	5137	1	-0	-0	0	0.6	0	1.015209	24.887597	150	2	1.1	0.9;
	5138	2	0	0	0	0	0	1.024572	-13.995277	20	4	1.1	0.9;
	5139	1	0	-0	0.081141	2.229144	0	0.995796	-3.341236	220	6	1.1	0.9;
	5140	1	0	0	0.491069	77.396812	0	1.039721	3.497796	380	11	1.1	0.9;
	5141	1	23.63	16.5	0	8.39	0	1.023365	36.901836	150	8	1.1	0.9;
	5142	1	0	0	0	43.87	0	1.07771	2.6181	380	18	1.2	0.8;
	5143	2	0	0	0	0	0	1.033735	-14.68008	20	4	1.1	0.9;
	5144	2	0	0	0	0	0	1.017192	-7.434926	20	22	1.1	0.9;
	5145	2	0	0	0	0	0	1.01785	-2.67748	20	22	1.1	0.9;
	5146	2	0	0	0	0	0	1.01785	-2.67748	20	22	1.1	0.9;
	5147	2	0	0	0	0	0	1.01785	-2.67748	20	22	1.1	0.9;
	5148	2	0	0	0	0	0	1.024966	-4.448719	20	22	1.1	0.9;
	5149	1	65.64	-0.98	0	11.55	0	1.007011	-25.858855	220	4	1.1	0.9;
	5150	1	-0	-0	0.54037	57.959712	0	1.006731	25.484118	380	5	1.1	0.9;
	5151	2	0	0	0	0	0	1.015505	-3.112218	20	22	1.1	0.9;
	5152	1	-0.64	-3.53	0	8.84	0	1.073042	71.1915	380	3	1.1	0.9;
	5153	1	123.8	41.4	0.039325	4.960675	0	1.007003	25.797244	220	13	1.1	0.9;
	5154	1	102.7	39	0	135.53	0	1.0811	64.133926	380	3	1.1	0.9;
	5155	2	0	0	0	0	0	1.044772	4.847391	20	18	1.1	0.9;
	5156	1	-0	-0	0	3.48	0	1.064971	0.399978	110	15	1.1	0.9;
	5157	1	0	0	0	42.71	0	1.055228	32.49598	380	1	1.1	0.9;
	5158	2	0	0	0	0	0	0.98522	42.733784	14.4	1	1.1	0.9;
	5159	2	0	0	0	0	0	1.033735	-14.68008	20	4	1.1	0.9;
	5160	1	-0	0	0	68.053762	0	1.037151	6.605163	380	5	1.1	0.9;
	5161	1	7.01	2.3	0	0.81	0	1.016177	-5.241408	110	12	1.1	0.9;
	5162	2	0	0	0	0	0	1.008964	6.659925	6.3	1	1.1	0.9;
	5163	2	0	0	0	0	0	1.033735	-14.68008	20	4	1.1	0.9;
	5164	1	-0	-0	0	0	0	1.040039	-14.484778	20	4	1.1	0.9;
	5165	1	-0	-0	0	0	0	1.040039	-14.484778	20	4	1.1	0.9;
	5166	2	0	0	0	0	0	1.017024	16.872664	20	18	1.1	0.9;
	5167	1	-0	-0	0	0	0	1.040039	-14.484778	20	4	1.1	0.9;
	5168	1	0	0	0	0	0	1.017812	-14.560212	20	4	1.1	0.9;
	5169	2	0	0	0	0	0	1.040788	-10.098416	20	4	1.1	0.9;
	5170	1	0	0	0	0	0	1.017812	-14.560212	20	4	1.1	0.9;
	5171	1	0	0	0	0	0	1.017812	-14.560212	20	4	1.1	0.9;
	5172	1	29.93	10.8	0	0.46	0	1.027506	-4.211385	110	19	1.1	0.9;
	5173	1	0	-0	0.041301	0.708772	0	1.016815	92.165392	220	1	1.1	0.9;
	5174	1	33.34	13.2	0	0.76	0	1.052139	28.580107	220	5	1.1	0.9;
	5175	1	0	-0	0	0.9	0	0.98591	-3.871199	110	18	1.1	0.9;
	5176	1	0	0	0	0	0	1.052015	-6.377282	150	9	1.1	0.9;
	5177	2	0	0	0	0	0	1.023095	-15.63004	20	4	1.1	0.9;
	5178	2	0	0	0	0	0	1.023095	-15.63004	20	4	1.1	0.9;
	5179	1	0	0	0	2.47	0	1.031323	8.512702	220	1	1.1	0.9;
	5180	1	118.5	21	0	2.19	0	0.986332	25.334582	150	8	1.1	0.9;
	5181	1	7.81	2.8	0	0.77	0	1.027796	-1.101464	110	19	1.1	0.9;
	5182	1	0	0	0	0	0	0.925713	-12.097092	150	9	1.1	0.9;
	5183	1	159.1	-28.7	0	2.31	0	1.068162	14.967776	220	3	1.1	0.9;
	5184	1	8.61	3.1	0	0.48	0	1.044865	-2.820297	110	19	1.1	0.9;
	5185	1	0	-0	0	0.49	0	1.025971	5.864981	220	5	1.1	0.9;
	5186	1	-0.48	0.1	0	0.91	0	1.039505	-2.381498	220	18	1.1	0.9;
	5187	2	0	0	0	0	0	1.021005	21.987746	20	14	1.1	0.9;
	5188	2	0	0	0	0	0	1.036966	15.051213	20	14	1.1	0.9;
	5189	1	0	0	0	4.17	0	1.015752	49.476558	220	8	1.1	0.9;
	5190	1	97	-52.1	0	3.95	0	1.019429	16.445362	220	3	1.1	0.9;
	5191	1	6.01	3	0	0.05	0	0.999899	-3.041462	220	6	1.1	0.9;
	5192	2	0	0	0	0	0	1.044001	-0.389429	20	23	1.1	0.9;
	5193	1	0	0	0	0.09	0	0.979776	-5.820604	110	21	1.1	0.9;
	5194	1	-0	-0	0	3.34	0	1.029423	6.341302	220	5	1.1	0.9;
	5195	1	0	0	0	0.98	0	0.994848	33.393544	220	3	1.1	0.9;
	5196	2	0	0	0	0	0	1.028002	7.184716	20	23	1.1	0.9;
	5197	1	0	0	0	14.09776	0	0.98388	13.009881	380	5	1.1	0.9;
	5198	1	0	-0	0	83.96	0	1.040272	-16.192477	380	4	1.1	0.9;
	5199	1	0	-0	0	1.56	0	0.979851	-8.587269	150	9	1.1	0.9;
	5200	1	-0	0	0	48.46	0	1.02816	16.70904	380	6	1.1	0.9;
	5201	1	4.2	1.2	0	30.55	0	1.040106	-14.883171	380	10	1.1	0.9;
	5202	2	0	0	0	0	0	0.99834	24.249316	13.8	1	1.1	0.9;
	5203	1	-0	0	0.115596	2.11367	0	1.040622	20.300906	110	19	1.1	0.9;
	5204	1	0	0	0	0	0	1.019678	-23.024367	20	4	1.1	0.9;
	5205	1	-0	-0	0.136871	2.413044	0	0.981836	1.606593	220	19	1.1	0.9;
	5206	1	0	-0	0	3.91	0	1.01681	-14.76301	220	4	1.1	0.9;
	5207	2	0	0	0	0	0	1.047444	6.684923	20	6	1.1	0.9;
	5208	2	0	0	0	0	0	1.047444	6.684923	20	6	1.1	0.9;
	5209	1	0	0	0.174825	7.755175	0	1.036607	7.408678	220	7	1.1	0.9;
	5210	2	0	0	0	0	0	1.029935	1.106697	11	1	1.1	0.9;
	5211	1	387.1	138.9	0.239691	22.220245	0	1.023026	30.671726	220	14	1.1	0.9;
	5212	2	0	0	0	0	0	1.021725	-10.876959	20	4	1.1	0.9;
	5213	1	-0	-0	0	0	0	1.019792	-15.494313	20	4	1.1	0.9;
	5214	1	0	0	0	18.31	0	1.024726	45.414901	380	3	1.1	0.9;
	5215	1	-0	-0	0	0	0	1.019792	-15.494313	20	4	1.1	0.9;
	5216	1	-0	-0	0	0	0	1.019792	-15.494313	20	4	1.1	0.9;
	5217	1	362.2	-10.5	0	19.86	0	0.994437	15.297275	220	6	1.1	0.9;
	5218	1	0	-0	0	1.33	0	1.020144	-4.540428	150	9	1.1	0.9;
	5219	1	31.23	12.9	0	0.36	0	1.009761	-21.023168	220	6	1.1	0.9;
	5220	1	87.2	31.8	0	5.96	0	0.988915	-3.063974	220	5	1.1	0.9;
	5221	1	41.65	13.6	0	0.31	0	0.97237	-6.977885	110	21	1.1	0.9;
	5222	1	-0	-0	0	2.51	0	1.047099	32.217182	220	5	1.1	0.9;
	5223	2	0	0	0	0	0	1.054235	39.510254	20	14	1.1	0.9;
	5224	1	4.1	0.9	0	0.09	0	0.99685	-7.213455	110	18	1.1	0.9;
	5225	2	0	0	0	0	0	1.037142	24.334502	20	5	1.1	0.9;
	5226	2	0	0	0	0	0	1.037142	24.334502	20	5	1.1	0.9;
	5227	1	152.5	-0.1	0	2.66	0	1.013023	31.438752	220	3	1.1	0.9;
	5228	2	0	0	0	0	0	1.037142	24.334502	20	5	1.1	0.9;
	5229	2	0	0	0	0	0	1.037142	24.334502	20	5	1.1	0.9;
	5230	1	31.43	10.6	0	1.61	0	1.021677	7.174807	220	5	1.1	0.9;
	5231	1	6.31	1.3	0	0.43	0	0.929626	-8.481962	110	18	1.1	0.9;
	5232	1	392.8	70.6	0	13.309841	0	1.012195	-6.68477	380	6	1.1	0.9;
	5233	1	0	-0	0	5.34	0	1.019848	22.321936	154	1	1.1	0.9;
	5234	1	19.72	9.4	0.038541	3.531459	0	1.078665	3.17763	220	5	1.1	0.9;
	5235	1	5.21	1.5	0	1.1	0	0.951897	-5.655079	110	18	1.1	0.9;
	5236	1	0	-0	0	53.268666	0	1.018217	0.541576	380	5	1.1	0.9;
	5237	1	30.6	4.3	0	8.4	0	1.012824	28.22011	154	1	1.1	0.9;
	5238	1	215.3	37.7	0	14.36	0	1.024109	7.923737	220	7	1.1	0.9;
	5239	1	19.04	7.9	0	4.53	0	1.027323	-4.289105	150	9	1.1	0.9;
	5240	1	-0	0	0.161328	3.849004	0	0.988395	-21.929683	220	6	1.1	0.9;
	5241	1	8.31	3	0	0.24	0	0.986972	4.191813	110	19	1.1	0.9;
	5242	2	0	0	0	0	0	1.022616	65.87032	20	3	1.1	0.9;
	5243	2	0	0	0	0	0	1.003094	94.404188	20	3	1.1	0.9;
	5244	1	0	0	0	0.53	0	0.981817	-5.453509	150	9	1.1	0.9;
	5245	2	0	0	0	0	0	0.993793	93.000413	20	3	1.1	0.9;
	5246	2	0	0	0	0	0	1.022616	65.87032	20	3	1.1	0.9;
	5247	1	24.43	8.7	0	0.46	0	1.024461	-4.101161	110	15	1.1	0.9;
	5248	1	1.4	0.5	0	0.85	0	1.000505	-2.078746	110	19	1.1	0.9;
	5249	1	-38.7	-12.99	0	13.393621	0	1.028566	25.71228	380	5	1.1	0.9;
	5250	1	-0	0	0	11.06	0	1.024655	48.619208	380	3	1.1	0.9;
	5251	1	277.72	7.35	0	3.58	0	1.030893	-12.424281	220	5	1.1	0.9;
	5252	2	0	0	0	0	0	1.017273	-4.353956	20	4	1.1	0.9;
	5253	1	-0	0	0.393538	-0.195154	0	1.047882	-4.305097	220	5	1.1	0.9;
	5254	2	0	0	0	0	0	1.024094	0.934183	11	1	1.1	0.9;
	5255	1	-10.61	3.5	0	1.3	0	0.959894	-10.038785	150	9	1.1	0.9;
	5256	1	353	20.7	0.099975	37.760025	0	1.014605	-21.99726	380	4	1.1	0.9;
	5257	1	-0	-0	0.019671	18.610329	0	1.02887	49.205594	380	3	1.1	0.9;
	5258	1	115	34.8	0.03998	3.70001	0	1.009593	-17.836874	220	4	1.1	0.9;
	5259	1	-0	0	0.040038	0.009962	0	0.996526	-17.372997	220	10	1.1	0.9;
	5260	2	0	0	0	0	0	1.028088	-13.031645	20	4	1.1	0.9;
	5261	2	0	0	0	0	0	1.017273	-4.353956	20	4	1.1	0.9;
	5262	2	0	0	0	0	0	1.017273	-4.353956	20	4	1.1	0.9;
	5263	1	0	0	0	78.22	0	1.062003	-12.552832	380	4	1.1	0.9;
	5264	1	-0	-0	0	1.6	0	0.978669	-7.861808	150	9	1.1	0.9;
	5265	1	37.54	18.3	0	0.22	0	1.027054	22.947359	220	5	1.1	0.9;
	5266	1	46	6.6	0	3.11	0	1.013859	18.447376	154	1	1.1	0.9;
	5267	1	-0	0	0	0.49	0	0.957532	-11.597741	110	18	1.1	0.9;
	5268	1	13.9	1.9	0	4.47	0	0.995085	12.187928	154	1	1.1	0.9;
	5269	1	-0	0	0	5.85	0	0.921793	-12.253668	150	9	1.1	0.9;
	5270	2	0	0	0	0	0	1.029323	-3.215687	20	6	1.1	0.9;
	5271	2	0	0	0	0	0	1.029323	-3.215687	20	6	1.1	0.9;
	5272	1	102.7	77.5	0	4.89	0	1.015467	-22.677921	220	4	1.1	0.9;
	5273	1	-17.8	-3.68	0.109667	41.234333	0	1.110897	70.705716	380	3	1.2	0.8;
	5274	2	0	0	0	0	0	1.023803	-2.474109	20	6	1.1	0.9;
	5275	2	0	0	0	0	0	1.023803	-2.474109	20	6	1.1	0.9;
	5276	2	0	0	0	0	0	1.023803	-2.474109	20	6	1.1	0.9;
	5277	2	0	0	0	0	0	1.023803	-2.474109	20	6	1.1	0.9;
	5278	2	0	0	0	0	0	1.029323	-3.215687	20	6	1.1	0.9;
	5279	2	0	0	0	0	0	1.029323	-3.215687	20	6	1.1	0.9;
	5280	1	31.43	6.6	0	3.32	0	1.029562	10.160061	220	5	1.1	0.9;
	5281	2	0	0	0	0	0	0.98306	43.226184	14.4	1	1.1	0.9;
	5282	1	0	-0	0	2.09	0	1.038028	5.797119	220	5	1.1	0.9;
	5283	1	11.91	3	0	1.22	0	1.000423	5.18281	110	18	1.1	0.9;
	5284	1	0	0	0	0.06	0	1.025331	0.441857	220	21	1.1	0.9;
	5285	1	68.78	13	0	10.45	0	1.029928	-21.084434	220	4	1.1	0.9;
	5286	2	0	0	0	0	0	1.023526	23.555243	20	2	1.1	0.9;
	5287	2	0	0	0	0	0	1.023526	23.555243	20	2	1.1	0.9;
	5288	1	129.2	13.4	0.04047	13.81953	0	1.015912	22.790588	220	14	1.1	0.9;
	5289	2	0	0	0	0	0	1.00746	39.171104	11	1	1.1	0.9;
	5290	1	0	-0	0	0.38	0	1.045177	-4.513312	150	9	1.1	0.9;
	5291	1	0	0	0	0.23	0	1.028189	8.896679	220	5	1.1	0.9;
	5292	2	0	0	0	0	0	1.050901	18.201002	20	7	1.1	0.9;
	5293	1	86.6	14.9	0	1.34	0	1.029206	11.095284	220	5	1.1	0.9;
	5294	1	0	0	0	0.74	0	1.057852	-6.001761	150	9	1.1	0.9;
	5295	1	124.3	39.4	0	6.43	0	1.039938	4.268956	380	20	1.1	0.9;
	5296	1	0	0	0.117165	11.642126	0	1.028062	-8.182777	220	6	1.1	0.9;
	5297	1	194.3	27.9	0	3.53	0	1.11636	70.15342	220	3	1.2	0.8;
	5298	1	95.1	-19.36	0	12.46	0	1.042365	38.732833	220	13	1.1	0.9;
	5299	1	0	-0	0	58.144073	0	0.96454	-13.395209	380	5	1.1	0.9;
	5300	1	-28.13	-0.82	0	0.1	0	1.058631	17.901103	220	5	1.1	0.9;
	5301	2	0	0	0	0	0	1.021316	-4.288887	20	6	1.1	0.9;
	5302	2	0	0	0	0	0	1.017834	1.692627	20	6	1.1	0.9;
	5303	2	0	0	0	0	0	1.024054	-16.964783	20	6	1.1	0.9;
	5304	2	0	0	0	0	0	1.024054	-16.964783	20	6	1.1	0.9;
	5305	2	0	0	0	0	0	1.03032	28.518041	20	6	1.1	0.9;
	5306	2	0	0	0	0	0	1.03032	28.518041	20	6	1.1	0.9;
	5307	2	0	0	0	0	0	1.030262	28.566856	20	6	1.1	0.9;
	5308	2	0	0	0	0	0	1.030262	28.566856	20	6	1.1	0.9;
	5309	2	0	0	0	0	0	1.033768	-2.92291	20	6	1.1	0.9;
	5310	2	0	0	0	0	0	1.033768	-2.92291	20	6	1.1	0.9;
	5311	2	0	0	0	0	0	1.057062	4.322623	20	18	1.1	0.9;
	5312	1	217.8	60.5	0.059037	2.290963	0	1.02263	10.160081	220	5	1.1	0.9;
	5313	1	-0	0	0	4.48	0	1.023815	28.570167	154	1	1.1	0.9;
	5314	2	0	0	0	0	0	0.984098	4.477364	15.75	1	1.1	0.9;
	5315	1	152.5	14.4	0	24.61	0	0.990143	1.402472	380	5	1.1	0.9;
	5316	2	0	0	0	0	0	1.002922	32.371982	6.3	1	1.1	0.9;
	5317	2	0	0	0	0	0	1.034048	-9.733404	20	4	1.1	0.9;
	5318	1	9.21	3.3	0	1.13	0	0.965112	-4.13419	110	19	1.1	0.9;
	5319	2	0	0	0	0	0	1.071081	3.910791	20	9	1.1	0.9;
	5320	1	0	0	0	13.51	0	0.999808	-10.542975	380	1	1.1	0.9;
	5321	1	108.43	31.33	0	0.88	0	1.023031	24.952149	220	13	1.1	0.9;
	5322	2	0	0	0	0	0	0.99879	3.61899	20	2	1.1	0.9;
	5323	2	0	0	0	0	0	1.011203	-2.333775	20	4	1.1	0.9;
	5324	2	0	0	0	0	0	1.011203	-2.333775	20	4	1.1	0.9;
	5325	2	0	0	0	0	0	1.011203	-2.333775	20	4	1.1	0.9;
	5326	2	0	0	0	0	0	1.011203	-2.333775	20	4	1.1	0.9;
	5327	2	0	0	0	0	0	1.011203	-2.333775	20	4	1.1	0.9;
	5328	2	0	0	0	0	0	1.011203	-2.333775	20	4	1.1	0.9;
	5329	1	11.31	4	0.038744	1.291256	0	1.056832	2.726838	110	15	1.1	0.9;
	5330	2	0	0	0	0	0	1.029009	32.575777	20	5	1.1	0.9;
	5331	1	0	0	0	61.52	0	1.019753	-24.135214	380	10	1.1	0.9;
	5332	2	0	0	0	0	0	1.028002	7.184716	20	23	1.1	0.9;
	5333	2	0	0	0	0	0	1.03063	1.723631	11	1	1.1	0.9;
	5334	1	35.14	16.1	0	0.45	0	1.02129	29.063688	150	8	1.1	0.9;
	5335	2	0	0	0	0	0	1.033768	14.596978	20	14	1.1	0.9;
	5336	1	58.97	27.1	0.197746	0.841691	0	1.015825	-25.064254	220	4	1.1	0.9;
	5337	1	21.12	8.3	0	4.26	0	1.026869	30.816332	150	8	1.1	0.9;
	5338	1	-0	-0	0	53.84	0	1.067711	-6.48805	380	12	1.2	0.8;
	5339	1	-0	-0	0	4.08	0	1.025694	-6.021941	220	10	1.1	0.9;
	5340	1	0	0	0	0	0	1.001756	-10.041769	150	9	1.1	0.9;
	5341	2	0	0	0	0	0	1.065506	1.014062	20	9	1.1	0.9;
	5342	2	0	0	0	0	0	1.019736	17.906487	20	6	1.1	0.9;
	5343	2	0	0	0	0	0	1.000957	10.706535	11	1	1.1	0.9;
	5344	1	70.48	33.9	0.112194	3.176505	0	1.051774	91.644565	220	3	1.1	0.9;
	5345	1	-0	0	0	0.39	0	1.035666	-7.386962	150	9	1.1	0.9;
	5346	1	34.34	12.5	0	0.23	0	1.011158	-5.202412	110	19	1.1	0.9;
	5347	2	0	0	0	0	0	1.043665	14.930439	20	14	1.1	0.9;
	5348	1	-0	-0	0	93.58	0	0.993916	5.014657	400	1	1.1	0.9;
	5349	1	24.63	16.7	0.080106	8.659894	0	1.005644	-16.679897	220	4	1.1	0.9;
	5350	2	0	0	0	0	0	1.021475	28.018024	20	10	1.1	0.9;
	5351	1	21.52	7.6	0	0.18	0	0.952457	-5.260666	110	18	1.1	0.9;
	5352	2	0	0	0	0	0	1.007679	-4.961668	20	2	1.1	0.9;
	5353	1	-0	0	0	11.98	0	1.053771	81.870655	380	3	1.1	0.9;
	5354	2	0	0	0	0	0	1.039621	10.691121	20	18	1.1	0.9;
	5355	1	-12.3	-10.6	0	3.54	0	1.084788	-3.606185	150	9	1.1	0.9;
	5356	2	0	0	0	0	0	1.041223	4.865379	20	18	1.1	0.9;
	5357	1	-0	-0	0	0	0	1.016884	-1.962084	20	18	1.1	0.9;
	5358	1	-0	0	0	15.23	0	1.030402	-27.749342	220	4	1.1	0.9;
	5359	2	0	0	0	0	0	1.054892	7.069138	20	2	1.1	0.9;
	5360	1	0	0	0	0	0	1.023166	4.463602	20	23	1.1	0.9;
	5361	1	0	0	0	0	0	1.023166	4.463602	20	23	1.1	0.9;
	5362	1	0	0	0	0	0	1.023166	4.463602	20	23	1.1	0.9;
	5363	1	0	0	0	2.61	0	0.979016	44.462885	154	1	1.1	0.9;
	5364	1	0	0	0	10.82	0	1.017741	41.666685	380	3	1.1	0.9;
	5365	2	0	0	0	0	0	1.03611	8.921216	20	23	1.1	0.9;
	5366	2	0	0	0	0	0	1.03611	8.921216	20	23	1.1	0.9;
	5367	1	0	0	0	12.4	0	1.010305	12.204616	220	5	1.1	0.9;
	5368	2	0	0	0	0	0	1.023161	-12.168584	20	6	1.1	0.9;
	5369	2	0	0	0	0	0	1.023161	-12.168584	20	6	1.1	0.9;
	5370	2	0	0	0	0	0	1.023161	-12.168584	20	6	1.1	0.9;
	5371	2	0	0	0	0	0	1.023161	-12.168584	20	6	1.1	0.9;
	5372	1	0	0	0	7.1	0	1.038719	-0.876477	220	18	1.1	0.9;
	5373	1	30.33	11	0	0.58	0	1.072551	-2.385083	110	19	1.1	0.9;
	5374	1	1.3	0.4	0	1.7	0	1.012325	4.838884	110	18	1.1	0.9;
	5375	1	-0	0	0.058227	-0.008227	0	1.061002	18.625585	220	5	1.1	0.9;
	5376	1	-5.01	-1.81	0	15.16	0	1.005051	-5.039137	220	11	1.1	0.9;
	5377	1	0	-5.5	0.039787	12.430213	0	1.019643	-16.147917	220	4	1.1	0.9;
	5378	1	42.55	17.6	0	17.28	0	1.011669	-5.415058	220	5	1.1	0.9;
	5379	1	32.24	11.7	0	0.1	0	1.025372	-4.086848	110	19	1.1	0.9;
	5380	1	-0	-0	0.709066	-1.750701	0	1.040126	59.677576	380	3	1.1	0.9;
	5381	1	1.5	0.8	0	0.55	0	0.98156	-6.430337	110	18	1.1	0.9;
	5382	2	0	0	0	0	0	1.019736	17.906487	20	6	1.1	0.9;
	5383	1	-0	0	0	57.87	0	1.050924	32.263423	380	1	1.1	0.9;
	5384	2	0	0	0	0	0	0.962211	47.074823	6.3	1	1.1	0.9;
	5385	2	0	0	0	0	0	1.018125	-0.191544	20	2	1.1	0.9;
	5386	1	-0	0	0	4.86	0	1.009475	-1.839005	220	5	1.1	0.9;
	5387	1	0	-0	0	4.68	0	1.021639	-15.781645	220	4	1.1	0.9;
	5388	2	0	0	0	0	0	1.033359	20.594354	20	5	1.1	0.9;
	5389	1	-0	0	0	11.87	0	1.070228	-0.629475	380	1	1.2	0.8;
	5390	1	61.4	8.8	0	3.53	0	1.024683	17.0095	154	1	1.1	0.9;
	5391	2	0	0	0	0	0	1.043492	-15.322718	20	4	1.1	0.9;
	5392	1	-0	0	0	0.04	0	0.99768	-9.077286	150	9	1.1	0.9;
	5393	1	20.72	5	0	6.64	0	1.023492	3.369288	220	5	1.1	0.9;
	5394	1	114.8	121.5	0	24.25	0	1.00901	29.747481	380	3	1.1	0.9;
	5395	1	138.4	39.5	0.118033	13.031967	0	1.029146	16.996023	220	5	1.1	0.9;
	5396	1	131	1.3	0	120.45	0	1.025143	3.709433	220	5	1.1	0.9;
	5397	2	0	0	0	0	0	1.01809	15.340169	11	1	1.1	0.9;
	5398	1	0	-0	0	2.8	0	1.065863	-4.559857	150	9	1.1	0.9;
	5399	1	0	0	0	0	0	1.019678	-23.024367	20	4	1.1	0.9;
	5400	1	0	0	0	0	0	1.019678	-23.024367	20	4	1.1	0.9;
	5401	2	0	0	0	0	0	1.028342	-18.47093	20	4	1.1	0.9;
	5402	1	-0	-0	1.294898	-0.826149	0	1.046795	41.135123	380	2	1.1	0.9;
	5403	2	0	0	0	0	0	1.018098	29.808617	154	1	1.1	0.9;
	5404	1	0	0	0.573222	0.203366	0	1.03936	34.351848	380	5	1.1	0.9;
	5405	1	0	0	0	0	0	1.019678	-23.024367	20	4	1.1	0.9;
	5406	2	0	0	0	0	0	1.007636	96.580359	20	3	1.1	0.9;
	5407	1	50.2	7.1	0	8.14	0	1.002675	25.619018	154	1	1.1	0.9;
	5408	2	0	0	0	0	0	0.996117	32.042491	11	1	1.1	0.9;
	5409	1	295.4	42	0	0.83	0	1.026365	-15.128429	220	4	1.1	0.9;
	5410	1	0	-0	0	0.74	0	1.034682	-7.443278	150	9	1.1	0.9;
	5411	2	0	0	0	0	0	0.991559	90.791769	20	3	1.1	0.9;
	5412	1	141.5	70.8	0	4.92	0	1.005549	11.254419	380	3	1.1	0.9;
	5413	1	0	-0	0	13.93	0	1.042695	35.810054	380	3	1.1	0.9;
	5414	1	-0	-0	0	2.04	0	1.084115	1.981553	110	19	1.2	0.8;
	5415	1	28.03	6.5	0	6.32	0	1.030851	8.542661	220	5	1.1	0.9;
	5416	2	0	0	0	0	0	1.024669	12.739456	20	18	1.1	0.9;
	5417	1	22.83	9.7	0	5.07	0	0.923632	-8.085026	110	18	1.1	0.9;
	5418	2	0	0	0	0	0	0.978998	26.658997	10.5	1	1.1	0.9;
	5419	1	132.3	135.6	0.358583	4.131181	0	1.008248	-11.333077	220	4	1.1	0.9;
	5420	2	0	0	0	0	0	1.015531	4.163384	20	9	1.1	0.9;
	5421	1	95	10.43	0	89.57	0	1.048791	17.954078	380	13	1.1	0.9;
	5422	1	5.51	2.2	0	0.59	0	0.988176	-11.073965	110	22	1.1	0.9;
	5423	1	-0	-0	0	9.51	0	1.010477	-20.953094	220	6	1.1	0.9;
	5424	1	24.83	7.6	0	0.23	0	0.967393	-6.460925	110	18	1.1	0.9;
	5425	1	1.7	0.3	0	0.21	0	0.974289	-7.210672	110	18	1.1	0.9;
	5426	1	-0	-0	0	0.02	0	0.926207	-12.070366	150	9	1.1	0.9;
	5427	1	2.5	1.3	0	0.12	0	0.964213	-6.205379	110	18	1.1	0.9;
	5428	1	0	-0	0	16.73	0	1.041265	2.248937	220	5	1.1	0.9;
	5429	1	28.93	10.5	0	0.88	0	1.000285	-3.353486	110	19	1.1	0.9;
	5430	1	182.3	65.9	0	62.49	0	1.080601	0.695571	380	20	1.2	0.8;
	5431	1	52.79	2.64	0	0	0	0.936417	-4.834215	110	18	1.1	0.9;
	5432	1	-0	0.01	0	0	0	1.001979	-7.828523	150	9	1.1	0.9;
	5433	1	-0	0	0	70.71	0	1.017043	-22.798983	380	4	1.1	0.9;
	5434	1	-0	0	0.1	5.73	0	1.007403	-6.121058	220	21	1.1	0.9;
	5435	1	-0	0	0	8.99	0	1.033263	-17.887254	220	4	1.1	0.9;
	5436	1	-17.62	5.8	0	0.23	0	1.023839	-7.475146	150	9	1.1	0.9;
	5437	1	-0	0	0.172737	-0.134351	0	1.033003	38.705858	150	8	1.1	0.9;
	5438	1	223.4	143.1	0.280648	4.559514	0	1.006737	-14.697123	220	4	1.1	0.9;
	5439	2	0	0	0	0	0	1.037973	11.471538	20	14	1.1	0.9;
	5440	1	-0	-0	0	1	0	0.99939	28.980276	154	1	1.1	0.9;
	5441	1	16.4	2.3	0	3.43	0	1.008389	18.986342	154	1	1.1	0.9;
	5442	1	-0	0	0	0.56	0	1.019446	-6.719975	150	9	1.1	0.9;
	5443	1	-0	-0	0.473225	52.235363	0	1.031638	2.949928	380	6	1.1	0.9;
	5444	2	0	0	0	0	0	1.017679	10.590328	20	22	1.1	0.9;
	5445	1	-32.71	-13.76	0	135.135902	0	1.030228	9.043792	380	5	1.1	0.9;
	5446	1	74.58	12.3	0.039447	21.250553	0	1.028368	5.87542	220	5	1.1	0.9;
	5447	1	16.62	-10.3	0	0.1	0	1.032604	-19.07465	220	6	1.1	0.9;
	5448	1	-4.86	-2.12	0	0.64	0	0.995449	-2.100907	110	18	1.1	0.9;
	5449	1	0	0	0	0.43	0	1.034882	4.848076	220	5	1.1	0.9;
	5450	1	0	-0	0	13.3	0	1.038472	4.437466	220	11	1.1	0.9;
	5451	1	0	-0	0	1.53	0	1.021105	-0.682824	150	9	1.1	0.9;
	5452	1	-0	-0	0	1.12	0	1.013387	-2.934412	220	12	1.1	0.9;
	5453	1	120.7	24.3	0	4.51	0	0.999883	-7.177022	220	6	1.1	0.9;
	5454	1	2.5	-0	0	0.57	0	1.010163	2.244387	110	18	1.1	0.9;
	5455	2	0	0	0	0	0	1.018325	9.145355	20	23	1.1	0.9;
	5456	2	0	0	0	0	0	1.009683	1.076815	20	18	1.1	0.9;
	5457	1	-0	0	0.019585	4.410415	0	1.023874	-12.218924	220	4	1.1	0.9;
	5458	1	32.1	4.5	0	2.53	0	0.999498	29.48892	154	1	1.1	0.9;
	5459	1	0	-0.01	0	1.2	0	1.017517	24.916531	220	3	1.1	0.9;
	5460	1	-0	-0	0	7.4	0	1.025654	-0.361056	220	1	1.1	0.9;
	5461	1	0	0	0.040055	2.829945	0	0.994387	-11.289887	150	10	1.1	0.9;
	5462	1	-0.19	-0.87	0	0	0	1.007118	30.130445	150	8	1.1	0.9;
	5463	2	0	0	0	0	0	1.054031	9.216302	20	5	1.1	0.9;
	5464	2	0	0	0	0	0	1.011467	18.152933	11	1	1.1	0.9;
	5465	1	24.03	8	0	0.75	0	1.002052	0.724345	110	21	1.1	0.9;
	5466	2	0	0	0	0	0	1.021987	8.5782	20	6	1.1	0.9;
	5467	2	0	0	0	0	0	1.021987	8.5782	20	6	1.1	0.9;
	5468	2	0	0	0	0	0	1.021987	8.5782	20	6	1.1	0.9;
	5469	1	115.3	16.4	0	2.45	0	1.009591	18.431924	154	1	1.1	0.9;
	5470	1	85.1	-122.4	0	6.67	0	1.00951	14.111952	380	3	1.1	0.9;
	5471	1	0	0	0	4.67	0	1.038952	-3.685101	220	1	1.1	0.9;
	5472	1	179.4	58.3	0	0.64	0	1.032584	4.617096	220	5	1.1	0.9;
	5473	2	0	0	0	0	0	1.034048	-9.733404	20	4	1.1	0.9;
	5474	1	104	51.2	0.020067	3.329933	0	0.982293	24.305163	220	13	1.1	0.9;
	5475	1	0	-0	0	0.87	0	1.015359	-20.625876	220	6	1.1	0.9;
	5476	2	0	0	0	0	0	1.058287	5.521142	20	5	1.1	0.9;
	5477	1	65.1	26.4	0	0.58	0	0.999998	-7.478578	110	23	1.1	0.9;
	5478	1	0	0	0.683996	-0.541497	0	1.065239	82.186971	380	8	1.1	0.9;
	5479	1	0	-0	0.0588	-0.049	0	1.020383	19.713555	220	3	1.1	0.9;
	5480	1	-0	-0	0.090774	7.550444	0	1.03358	4.727162	220	18	1.1	0.9;
	5481	1	16.52	7.1	0	1.1	0	0.995881	-14.296589	110	22	1.1	0.9;
	5482	1	165.3	89.2	0	13.47	0	1.026776	33.821305	380	3	1.1	0.9;
	5483	1	-32.04	10.3	0	82.52	0	1.043257	-18.000081	380	10	1.1	0.9;
	5484	2	0	0	0	0	0	0.956073	24.685075	13.8	1	1.1	0.9;
	5485	1	2.3	0.8	0	0.21	0	0.954523	-1.873735	110	19	1.1	0.9;
	5486	1	-0	-0	0.090774	3.980444	0	1.038771	7.75297	220	18	1.1	0.9;
	5487	1	-24.87	-3.7	0	0.35	0	1.023609	8.435988	110	18	1.1	0.9;
	5488	1	-0	-0	0	0.11	0	1.002597	-1.485346	110	18	1.1	0.9;
	5489	1	51.36	-0.7	0	0.57	0	1.032625	-19.504367	220	4	1.1	0.9;
	5490	2	0	0	0	0	0	1.028982	-17.79027	20	4	1.1	0.9;
	5491	1	-0	-0	0	0.09	0	0.997947	-10.410664	150	9	1.1	0.9;
	5492	1	45.15	4.1	0	2.61	0	1.013587	-30.743973	220	4	1.1	0.9;
	5493	1	42.48	10.36	0	0.37	0	1.047324	29.965681	220	5	1.1	0.9;
	5494	2	0	0	0	0	0	0.991002	22.274075	10.6	1	1.1	0.9;
	5495	1	18.42	6	0	0.93	0	0.942228	-8.841529	110	21	1.1	0.9;
	5496	1	0.3	1.3	0.038604	0.101396	0	1.043448	79.328992	220	3	1.1	0.9;
	5497	1	40.44	20.5	0	0.85	0	1.068116	5.359204	220	5	1.1	0.9;
	5498	2	0	0	0	0	0	1.041198	4.140848	20	3	1.1	0.9;
	5499	1	-0	-0	0	13.03	0	1.033521	9.078689	220	5	1.1	0.9;
	5500	1	-0	-0	0.708649	1.518986	0	1.034725	32.734841	380	5	1.1	0.9;
	5501	1	122.3	45.9	0	4.03	0	1.012878	6.54106	220	7	1.1	0.9;
	5502	1	71.38	10.9	0.039268	3.160549	0	1.001698	92.057182	150	1	1.1	0.9;
	5503	2	0	0	0	0	0	1.032124	21.999225	20	10	1.1	0.9;
	5504	1	93.9	10.95	0	4.98	0	1.03277	-19.320799	220	4	1.1	0.9;
	5505	1	24.03	7.9	0	1.08	0	0.985939	-2.873709	110	21	1.1	0.9;
	5506	1	34.14	11.8	0.039086	0.950914	0	1.026291	-4.491374	110	19	1.1	0.9;
	5507	1	39.44	5.9	0	1.92	0	1.014219	-26.03395	220	4	1.1	0.9;
	5508	1	194.6	48.3	0	0.85	0	1.000405	-4.7006	110	21	1.1	0.9;
	5509	1	-0	-0	0.038572	0.211071	0	1.039944	-3.443809	220	6	1.1	0.9;
	5510	2	0	0	0	0	0	1.05632	32.386122	20	14	1.1	0.9;
	5511	1	69.28	44.1	0.39706	-0.66706	0	1.031703	-3.136077	220	21	1.1	0.9;
	5512	2	0	0	0	0	0	1.014266	27.581576	20	10	1.1	0.9;
	5513	1	7.61	0.5	0	0.59	0	0.966548	-0.111776	110	18	1.1	0.9;
	5514	2	0	0	0	0	0	1.05632	32.386122	20	14	1.1	0.9;
	5515	1	0	0	0.345747	7.051877	0	1.04787	36.346539	380	3	1.1	0.9;
	5516	1	-24.93	11.8	0	3.51	0	1.046421	-13.009884	220	4	1.1	0.9;
	5517	2	0	0	0	0	0	1.03284	28.489594	20	10	1.1	0.9;
	5518	1	35.54	11.7	0	0.19	0	0.990205	-4.49642	110	21	1.1	0.9;
	5519	1	4	1.3	0	0.56	0	0.981793	-11.284467	110	12	1.1	0.9;
	5520	1	83.49	17.4	0	19.32	0	1.057585	11.125317	220	5	1.1	0.9;
	5521	1	0	-0	0.115739	8.593551	0	1.037504	75.247519	220	3	1.1	0.9;
	5522	1	181.1	25.8	0	1.1	0	0.993993	29.495121	154	1	1.1	0.9;
	5523	2	0	0	0	0	0	1.037993	28.716517	20	5	1.1	0.9;
	5524	1	71.08	16.2	0	11.58	0	1.025446	11.298124	220	3	1.1	0.9;
	5525	1	338.22	104.11	0	91.11	0	0.997105	20.026989	380	13	1.1	0.9;
	5526	1	119.1	61.4	0	1.9	0	1.042313	-6.262534	150	9	1.1	0.9;
	5527	1	37.84	23.4	0	0.02	0	1.002937	-4.683269	150	9	1.1	0.9;
	5528	2	0	0	0	0	0	1.009486	16.888084	20	5	1.1	0.9;
	5529	1	4	0.1	0	0.09	0	1.053958	-6.513606	150	9	1.1	0.9;
	5530	2	0	0	0	0	0	1.037993	28.716517	20	5	1.1	0.9;
	5531	1	-0	0	0	1.04	0	1.017681	78.940855	220	3	1.1	0.9;
	5532	1	18.02	6.5	0	0.64	0	1.025267	-2.273267	110	19	1.1	0.9;
	5533	1	46.9	6.7	0	0.07	0	1.015563	19.169778	154	1	1.1	0.9;
	5534	2	0	0	0	0	0	1.020332	23.152992	20	10	1.1	0.9;
	5535	1	33.04	21.6	0	26.59	0	1.038002	-21.381787	220	6	1.1	0.9;
	5536	1	42.25	5.9	0	13.07	0	1.019573	-23.233961	220	4	1.1	0.9;
	5537	1	19.82	10.9	0	0.05	0	1.031681	8.016407	220	5	1.1	0.9;
	5538	1	-0	0	0.157271	12.642047	0	1.018608	-18.078221	220	6	1.1	0.9;
	5539	1	34.14	13.5	0	0	0	1.020643	26.34871	150	8	1.1	0.9;
	5540	2	0	0	0	0	0	1.045359	9.917398	20	23	1.1	0.9;
	5541	1	2.2	-48.8	0	39.69	0	1.04929	32.879646	380	3	1.1	0.9;
	5542	2	0	0	0	0	0	1.010203	8.259748	11	1	1.1	0.9;
	5543	1	0	0	0	30.09	0	1.052316	-5.470021	380	1	1.1	0.9;
	5544	1	-0	0	0	0.46	0	0.944329	-10.949935	150	9	1.1	0.9;
	5545	1	0	0	0	2.46	0	1.033232	20.726267	220	5	1.1	0.9;
	5546	2	0	0	0	0	0	1.026207	-11.793103	20	4	1.1	0.9;
	5547	1	201.4	47.5	0	0	0	1.004093	46.303458	220	3	1.1	0.9;
	5548	2	0	0	0	0	0	1.030988	2.394145	20	9	1.1	0.9;
	5549	1	-0	0	0	5.25	0	1.022765	36.791666	150	8	1.1	0.9;
	5550	1	77.19	25.4	0	4.13	0	0.991249	31.489724	220	3	1.1	0.9;
	5551	1	-0	0	0	6.96	0	1.033986	-0.883818	220	1	1.1	0.9;
	5552	1	-0	-0	0	0.57	0	1.097818	-4.717769	150	9	1.1	0.9;
	5553	1	8.41	2.5	0	1.18	0	0.987058	-8.963027	110	18	1.1	0.9;
	5554	1	52.46	19	0	0.34	0	1.0273	-4.462847	110	19	1.1	0.9;
	5555	1	162.4	14.4	0.019919	12.920081	0	0.993552	21.325826	220	13	1.1	0.9;
	5556	1	-0	-0	0	10.31	0	1.030728	9.614364	154	1	1.1	0.9;
	5557	1	46.35	86.9	0	10.32	0	1.045313	37.497254	380	3	1.1	0.9;
	5558	1	27.8	4	0	3.34	0	0.968669	19.556191	154	1	1.1	0.9;
	5559	1	-0	-0	0	26.98	0	0.992104	9.370388	220	5	1.1	0.9;
	5560	1	80.56	20.85	0	10.33	0	0.994311	40.473761	220	13	1.1	0.9;
	5561	1	23.93	8.7	0	0.48	0	1.041016	-3.593429	110	19	1.1	0.9;
	5562	1	0	0	0	9.09	0	1.02065	42.452202	380	3	1.1	0.9;
	5563	1	0	0	0	0	0	1.03097	-5.06415	150	9	1.1	0.9;
	5564	1	24.63	15.3	0	1.28	0	1.033422	-6.60149	150	9	1.1	0.9;
	5565	1	0	-0	0.060373	1.049627	0	0.990674	-12.457063	150	10	1.1	0.9;
	5566	1	105.4	92.5	0.117734	22.942266	0	1.029668	-16.520796	220	4	1.1	0.9;
	5567	1	63.67	20.2	0	6.13	0	1.039509	27.773539	220	5	1.1	0.9;
	5568	1	-1.9	-28.1	0	100.56	0	1.054612	-13.021657	380	4	1.1	0.9;
	5569	1	-0	-0	0.019548	15.600452	0	1.038162	75.883553	380	3	1.1	0.9;
	5570	1	0	-0	0	0.96	0	0.985969	-5.347667	150	9	1.1	0.9;
	5571	1	2.8	0.6	0	0.62	0	1.047477	-2.055486	110	15	1.1	0.9;
	5572	1	272.5	127.2	0.039174	55.140826	0	1.028225	45.555177	380	13	1.1	0.9;
	5573	1	0	0	0	17.64	0	1.053988	-13.855785	380	4	1.1	0.9;
	5574	1	39	5.5	0	18.09	0	0.984528	42.045502	154	1	1.1	0.9;
	5575	2	0	0	0	0	0	1.031736	41.040722	20	10	1.1	0.9;
	5576	2	0	0	0	0	0	0.976475	43.406918	15.5	1	1.1	0.9;
	5577	1	2.3	0.8	0.079208	2.270792	0	1.035016	0.8172	110	15	1.1	0.9;
	5578	1	0	-0	0	2.97	0	0.999568	-12.992092	110	12	1.1	0.9;
	5579	1	-0	0	0.181549	6.920889	0	1.059024	-2.926118	220	18	1.2	0.8;
	5580	1	23.3	3.3	0	9.29	0	1.006276	16.36648	154	1	1.1	0.9;
	5581	2	0	0	0	0	0	1.063714	6.510197	20	9	1.1	0.9;
	5582	1	39.34	24.4	0	0	0	1.009463	-9.554155	150	9	1.1	0.9;
	5583	1	10.11	5.6	0	0.89	0	0.987706	-3.303202	110	18	1.1	0.9;
	5584	1	-0.5	-0.17	0	0.21	0	0.979206	-6.232471	110	18	1.1	0.9;
	5585	1	-0	-0	0	0	0	1.019792	-15.494313	20	4	1.1	0.9;
	5586	1	5.01	2	0	0.84	0	1.025107	-5.101065	110	23	1.1	0.9;
	5587	2	0	0	0	0	0	1.041875	20.867491	20	7	1.1	0.9;
	5588	1	-0	0	0.548295	7.403222	0	1.079463	38.243221	380	14	1.2	0.8;
	5589	1	90.6	15.8	0	7.99	0	1.062436	4.140276	220	5	1.1	0.9;
	5590	1	-9.75	-2.74	0	0.41	0	1.026131	4.125771	110	18	1.1	0.9;
	5591	1	159.9	10.6	0	69.6	0	1.01737	-3.57985	380	6	1.1	0.9;
	5592	2	0	0	0	0	0	1.00876	0.250829	20	18	1.1	0.9;
	5593	2	0	0	0	0	0	1.014635	-0.980319	20	18	1.1	0.9;
	5594	2	0	0	0	0	0	1.028002	7.184716	20	23	1.1	0.9;
	5595	1	226	55.79	0	18.2	0	1.014621	-26.468886	220	4	1.1	0.9;
	5596	1	130.5	24.7	0	0.4	0	1.030843	7.951587	220	5	1.1	0.9;
	5597	1	16.32	2.8	0	1.41	0	0.992031	-1.689804	110	18	1.1	0.9;
	5598	1	77.49	17.3	0	0.1	0	1.017435	13.5457	220	5	1.1	0.9;
	5599	1	0	-0	0	6.32	0	0.998677	-23.026413	220	6	1.1	0.9;
	5600	1	35.66	-9.39	0	7.89	0	1.043652	0.642216	220	5	1.1	0.9;
	5601	1	44.45	10.9	0	2.23	0	1.030483	4.072037	220	5	1.1	0.9;
	5602	1	-0	-0	0	35	0	1.019321	21.930962	150	8	1.1	0.9;
	5603	1	5.31	1.8	0	0.72	0	1.002126	-8.787185	110	12	1.1	0.9;
	5604	1	105.6	70	0.237412	7.212157	0	1.029928	-11.421758	220	4	1.1	0.9;
	5605	1	-0	0	0	0	0	1.029028	28.384829	20	5	1.1	0.9;
	5606	1	-0	0	0	0	0	1.029028	28.384829	20	5	1.1	0.9;
	5607	2	0	0	0	0	0	1.025746	32.862676	20	5	1.1	0.9;
	5608	2	0	0	0	0	0	1.045989	24.800863	20	5	1.1	0.9;
	5609	1	-0	0	0	0	0	1.029028	28.384829	20	5	1.1	0.9;
	5610	2	0	0	0	0	0	1.025746	32.862676	20	5	1.1	0.9;
	5611	2	0	0	0	0	0	1.047982	25.815744	20	5	1.1	0.9;
	5612	2	0	0	0	0	0	1.047982	25.815744	20	5	1.1	0.9;
	5613	2	0	0	0	0	0	1.045989	24.800863	20	5	1.1	0.9;
	5614	2	0	0	0	0	0	1.047806	26.509997	20	5	1.1	0.9;
	5615	1	0	0	0	0	0	0.990005	-9.648284	150	9	1.1	0.9;
	5616	1	0	0	0	0.72	0	1.048472	3.178718	220	5	1.1	0.9;
	5617	1	-0	-0	0	9.47	0	1.077244	-11.176288	220	11	1.1	0.9;
	5618	1	40.5	5.7	0	5.41	0	1.003124	25.734388	154	1	1.1	0.9;
	5619	2	0	0	0	0	0	1.049918	7.582329	20	5	1.1	0.9;
	5620	1	18.82	6.8	0	0.59	0	0.952596	-2.388742	110	19	1.1	0.9;
	5621	1	0	0	0	0.67	0	0.946511	-10.737395	150	9	1.1	0.9;
	5622	2	0	0	0	0	0	1.03975	24.63487	20	5	1.1	0.9;
	5623	1	-0	-0	0	97.323975	0	0.988243	16.963551	380	5	1.1	0.9;
	5624	1	-0	-0	0	0	0	1.019792	-15.494313	20	4	1.1	0.9;
	5625	1	130.1	19.3	0	19.24	0	0.978962	7.657504	220	6	1.1	0.9;
	5626	1	14.22	6.1	0	0.18	0	0.989695	-1.825556	110	18	1.1	0.9;
	5627	2	0	0	0	0	0	1.026746	3.870557	20	20	1.1	0.9;
	5628	2	0	0	0	0	0	1.020137	-11.623454	20	22	1.1	0.9;
	5629	2	0	0	0	0	0	1.020137	-11.623454	20	22	1.1	0.9;
	5630	2	0	0	0	0	0	0.976217	-9.446382	20	22	1.1	0.9;
	5631	2	0	0	0	0	0	1.009821	-5.315519	20	22	1.1	0.9;
	5632	2	0	0	0	0	0	1.015993	-9.375302	20	22	1.1	0.9;
	5633	1	180.4	119.2	0	53.22	0	1.001096	16.815156	380	3	1.1	0.9;
	5634	2	0	0	0	0	0	1.009821	-5.315519	20	22	1.1	0.9;
	5635	1	44.15	10.6	0	1.62	0	0.99051	-17.628892	110	22	1.1	0.9;
	5636	1	36.74	22.7	0	0.46	0	1.001756	-10.041769	150	9	1.1	0.9;
	5637	2	0	0	0	0	0	1.043748	36.747443	20	2	1.1	0.9;
	5638	1	89.1	27.5	0	1.01	0	1.014728	-11.054108	110	22	1.1	0.9;
	5639	1	33.04	10.9	0	0.97	0	0.949125	-8.184163	110	21	1.1	0.9;
	5640	2	0	0	0	0	0	1.06339	20.87235	20	14	1.1	0.9;
	5641	2	0	0	0	0	0	1.043087	11.638661	20	14	1.1	0.9;
	5642	1	44.35	12.8	0.049456	6.940544	0	1.025062	-16.730832	220	4	1.1	0.9;
	5643	1	-0	-0	0.117943	0.021542	0	1.028403	19.42828	220	6	1.1	0.9;
	5644	1	6.41	2.3	0	0.97	0	0.963377	-3.10766	110	19	1.1	0.9;
	5645	2	0	0	0	0	0	1.003179	38.552189	13.8	1	1.1	0.9;
	5646	1	123.9	32.6	0	1.52	0	1.01326	-17.047746	220	4	1.1	0.9;
	5647	1	12.91	4.7	0	0.39	0	1.031808	17.017475	110	19	1.1	0.9;
	5648	1	16.7	2.4	0	2.73	0	1.002793	28.630967	154	1	1.1	0.9;
	5649	1	-0	0	0	5.36	0	1.01842	36.689147	150	8	1.1	0.9;
	5650	1	6.81	2.3	0	1.88	0	0.98149	-7.303136	110	18	1.1	0.9;
	5651	1	123.09	9.2	0	0.92	0	1.034087	30.349767	220	13	1.1	0.9;
	5652	1	0	0	0	0	0	1.015014	-12.066594	150	9	1.1	0.9;
	5653	2	0	0	0	0	0	1.01212	1.557585	20	18	1.1	0.9;
	5654	2	0	0	0	0	0	1.063453	20.858871	20	14	1.1	0.9;
	5655	1	73.08	15.7	0.039626	1.900281	0	0.989306	90.493667	150	1	1.1	0.9;
	5656	1	-0	0	0	6.39	0	1.014415	46.434829	380	1	1.1	0.9;
	5657	2	0	0	0	0	0	1.048637	18.752224	20	5	1.1	0.9;
	5658	2	0	0	0	0	0	1.015919	3.580325	6.3	1	1.1	0.9;
	5659	1	-0	0	0.224609	28.762184	0	1.080485	11.132062	220	18	1.2	0.8;
	5660	2	0	0	0	0	0	1.05544	46.172493	20	3	1.1	0.9;
	5661	1	85.5	118.1	0	5.09	0	0.987955	35.791506	220	3	1.1	0.9;
	5662	1	-38.73	-15.32	0	97.23	0	1.024984	3.613638	220	5	1.1	0.9;
	5663	1	28.03	8.8	0	1.05	0	0.998779	-8.829037	110	12	1.1	0.9;
	5664	1	0	0	0	11.68	0	1.068809	5.549085	220	5	1.1	0.9;
	5665	1	0	-0	0	0.13	0	1.035449	-4.107003	150	9	1.1	0.9;
	5666	1	0	-0	0	2.07	0	1.069948	21.128396	220	3	1.1	0.9;
	5667	1	-0	-0	0	0.23	0	1.047105	87.088421	220	3	1.1	0.9;
	5668	2	0	0	0	0	0	1.028707	4.250081	20	18	1.1	0.9;
	5669	1	28.03	10.2	0	0.83	0	1.045843	-0.415561	110	19	1.1	0.9;
	5670	2	0	0	0	0	0	1.028707	4.250081	20	18	1.1	0.9;
	5671	1	0	0	0	8.07	0	1.021296	30.056732	380	3	1.1	0.9;
	5672	2	0	0	0	0	0	1.03058	15.553508	20	14	1.1	0.9;
	5673	1	-18.32	6	0	0.32	0	1.01606	-7.943038	150	9	1.1	0.9;
	5674	2	0	0	0	0	0	1.043874	15.587474	20	18	1.1	0.9;
	5675	1	-0	-0	0	0.73	0	1.016124	2.135408	110	21	1.1	0.9;
	5676	2	0	0	0	0	0	1.025082	-16.616549	20	6	1.1	0.9;
	5677	1	-0	-0	0	2.26	0	1.008845	14.340823	220	5	1.1	0.9;
	5678	1	201.06	19.46	0	2.87	0	1.036144	1.509807	220	5	1.1	0.9;
	5679	2	0	0	0	0	0	0.996318	28.762446	6.3	1	1.1	0.9;
	5680	1	-8.43	-7.98	0	0	0	0.990521	25.757347	150	8	1.1	0.9;
	5681	2	0	0	0	0	0	1.000057	26.10357	11	1	1.1	0.9;
	5682	1	0	-0	0	4.43	0	1.010703	-2.718775	220	1	1.1	0.9;
	5683	1	0	-0	0.574645	1.662232	0	1.036415	33.220718	380	5	1.1	0.9;
	5684	1	12.8	1.8	0	6	0	0.99565	24.118259	154	1	1.1	0.9;
	5685	1	50.06	23	0	0.24	0	0.977662	-23.879736	220	6	1.1	0.9;
	5686	1	-2835.7	-818.9	0	56.48	0	1.037302	19.028226	380	3	1.1	0.9;
	5687	1	77	-239.9	0	4.21	0	1.035473	31.04393	220	3	1.1	0.9;
	5688	1	63.67	21.3	0	3.69	0	1.020772	7.039712	220	7	1.1	0.9;
	5689	1	51.26	-17.2	0	100.22	0	1.037277	18.603106	380	6	1.1	0.9;
	5690	1	11.61	4.3	0	1	0	1.048282	-1.885985	110	15	1.1	0.9;
	5691	1	102.1	27.8	0.039298	0.180702	0	1.034135	20.861206	220	5	1.1	0.9;
	5692	1	31.73	3.8	0	1.3	0	1.015709	7.345958	110	18	1.1	0.9;
	5693	1	0	-0	0	9.19	0	1.043404	54.291731	220	8	1.1	0.9;
	5694	1	-1.09	1.5	0.077254	1.542746	0	1.053797	-3.818548	150	9	1.1	0.9;
	5695	2	0	0	0	0	0	1.034646	-0.807488	20	2	1.1	0.9;
	5696	1	-28.13	-0.82	0	0.09	0	1.058617	17.901877	220	5	1.1	0.9;
	5697	2	0	0	0	0	0	1.002466	-10.296121	20	4	1.1	0.9;
	5698	1	-0	0	0	0	0	1.005121	-15.045231	20	4	1.1	0.9;
	5699	1	-0	0	0	0.44	0	0.988269	-4.625892	110	18	1.1	0.9;
	5700	1	56.26	20.1	0	0.92	0	1.012126	-1.815901	110	15	1.1	0.9;
	5701	1	4.4	2.2	0	0.41	0	1.015564	-11.698808	110	22	1.1	0.9;
	5702	2	0	0	0	0	0	1.027418	-10.310547	20	4	1.1	0.9;
	5703	1	-0	-0	0	0	0	1.03789	-14.752144	20	4	1.1	0.9;
	5704	2	0	0	0	0	0	1.053462	40.300443	20	14	1.1	0.9;
	5705	2	0	0	0	0	0	1.053459	40.302912	20	14	1.1	0.9;
	5706	2	0	0	0	0	0	1.056281	36.699329	20	14	1.1	0.9;
	5707	2	0	0	0	0	0	1.056281	36.699329	20	14	1.1	0.9;
	5708	2	0	0	0	0	0	1.063449	20.85668	20	14	1.1	0.9;
	5709	1	0	0	0.019798	0.280202	0	1.016072	22.381001	220	5	1.1	0.9;
	5710	1	94.3	-27.5	0	2.34	0	1.004588	32.779584	220	3	1.1	0.9;
	5711	1	-49.05	1	0	32.5	0	1.049141	7.531626	220	11	1.1	0.9;
	5712	1	0	-0	0.161274	1.883044	0	1.036026	35.070334	380	2	1.1	0.9;
	5713	1	-10.41	3.4	0	0.99	0	1.028846	-5.79278	150	9	1.1	0.9;
	5714	2	0	0	0	0	0	1.041293	27.016012	20	10	1.1	0.9;
	5715	2	0	0	0	0	0	1.019539	-2.214354	20	10	1.1	0.9;
	5716	2	0	0	0	0	0	1.006233	16.146442	13.8	1	1.1	0.9;
	5717	1	4.61	1.7	0	0.49	0	1.00657	-1.131032	110	19	1.1	0.9;
	5718	2	0	0	0	0	0	1.043748	36.747443	20	2	1.1	0.9;
	5719	1	0	0	0.090774	6.319044	0	1.024367	-3.01161	220	18	1.1	0.9;
	5720	1	19.62	-3.1	0	10.98	0	1.025855	-3.600731	220	6	1.1	0.9;
	5721	1	-0	0	0	0.05	0	1.033065	16.599351	220	5	1.1	0.9;
	5722	1	0	-0	0	4.77	0	1.005471	13.874385	220	5	1.1	0.9;
	5723	2	0	0	0	0	0	1.042317	14.274354	20	14	1.1	0.9;
	5724	1	0	0	0	14.27	0	1.066969	53.426857	380	3	1.1	0.9;
	5725	1	-272.2	6.6	0	29.33	0	1.022383	5.236844	380	6	1.1	0.9;
	5726	1	33.04	12.1	0	1.19	0	0.9717	-14.549319	110	22	1.1	0.9;
	5727	1	23.43	8.4	0	0.32	0	1.056156	-2.309831	110	15	1.1	0.9;
	5728	2	0	0	0	0	0	1.044132	13.300463	20	3	1.1	0.9;
	5729	2	0	0	0	0	0	1.044132	13.300463	20	3	1.1	0.9;
	5730	2	0	0	0	0	0	1.044132	13.300463	20	3	1.1	0.9;
	5731	2	0	0	0	0	0	1.044132	13.300463	20	3	1.1	0.9;
	5732	2	0	0	0	0	0	1.034671	8.05531	20	3	1.1	0.9;
	5733	1	121.5	29.6	0	2.49	0	0.973928	22.692838	150	8	1.1	0.9;
	5734	1	5.11	2.3	0	0.48	0	0.951297	-7.389214	110	18	1.1	0.9;
	5735	1	-0	0	0.057583	0.542417	0	1.034875	1.842217	220	21	1.1	0.9;
	5736	1	-0	0	0	0.59	0	0.968092	-10.363327	150	9	1.1	0.9;
	5737	1	20.12	8.6	0	0.18	0	0.990695	-4.359908	110	18	1.1	0.9;
	5738	1	211.1	93.1	0	0	0	0.948716	-0.128776	110	18	1.1	0.9;
	5739	2	0	0	0	0	0	1.038438	0.065623	20	9	1.1	0.9;
	5740	1	-0	0	0.166443	-0.391647	0	1.049045	7.495402	220	3	1.1	0.9;
	5741	2	0	0	0	0	0	1.045359	10.101302	20	14	1.1	0.9;
	5742	2	0	0	0	0	0	0.973258	21.632267	15.75	1	1.1	0.9;
	5743	1	-0	-0	0	69.75	0	0.978727	35.787613	400	1	1.1	0.9;
	5744	1	5.31	1.9	0.009974	0.81	0	0.988018	-6.490165	110	18	1.1	0.9;
	5745	1	-0	0	0	8.16	0	1.022914	15.763597	220	13	1.1	0.9;
	5746	2	0	0	0	0	0	1.037284	8.367307	20	23	1.1	0.9;
	5747	1	0	0	0	0	0	1.03026	3.848789	20	23	1.1	0.9;
	5748	1	0	0	0	0	0	1.03026	3.848789	20	23	1.1	0.9;
	5749	1	0	-0	0.22295	6.573498	0	1.065613	46.834612	380	14	1.1	0.9;
	5750	2	0	0	0	0	0	1.028771	2.144472	20	6	1.1	0.9;
	5751	2	0	0	0	0	0	1.022435	-15.897794	20	6	1.1	0.9;
	5752	1	-3.3	-0.7	0	0.7	0	1.085474	1.430716	150	9	1.1	0.9;
	5753	1	-0	0.01	0	19.73	0	1.033793	9.21995	220	5	1.1	0.9;
	5754	1	143.9	85.2	0	3.1	0	1.040399	19.36982	220	3	1.1	0.9;
	5755	1	-0	-0	0	0	0	1.020557	9.624602	20	14	1.1	0.9;
	5756	1	16.32	5.3	0	1.04	0	0.946362	-8.81689	110	21	1.1	0.9;
	5757	1	-0	-0	0	2.68	0	1.016262	22.326539	220	5	1.1	0.9;
	5758	1	0	-0	0.086633	0.534111	0	1.079673	11.115093	220	18	1.2	0.8;
	5759	1	0	0	0	22.74	0	1.020477	-11.022564	150	9	1.1	0.9;
	5760	1	66.67	16.9	0.039171	7.290829	0	1.038988	17.120629	220	5	1.1	0.9;
	5761	2	0	0	0	0	0	1.054937	4.624127	20	20	1.1	0.9;
	5762	1	20.32	6.7	0	1	0	0.964932	-6.955073	110	21	1.1	0.9;
	5763	1	-11.8	-5.47	0	88.050735	0	1.024798	30.073669	380	5	1.1	0.9;
	5764	2	0	0	0	0	0	1.050901	18.201002	20	7	1.1	0.9;
	5765	2	0	0	0	0	0	1.030566	-6.658147	20	9	1.1	0.9;
	5766	1	67.78	10.4	0	7.18	0	1.016135	-25.646727	220	4	1.1	0.9;
	5767	1	-0	-0	0.232844	2.415665	0	1.054759	28.608378	380	14	1.1	0.9;
	5768	1	39.34	10.3	0	0.33	0	1.0003	2.358032	110	18	1.1	0.9;
	5769	2	0	0	0	0	0	1.025333	3.180104	20	6	1.1	0.9;
	5770	2	0	0	0	0	0	1.025019	-12.264635	20	6	1.1	0.9;
	5771	2	0	0	0	0	0	1.025019	-12.264635	20	6	1.1	0.9;
	5772	2	0	0	0	0	0	1.025333	3.180104	20	6	1.1	0.9;
	5773	1	-0	0	0.343686	0.573595	0	1.056553	39.273401	380	3	1.1	0.9;
	5774	1	11.11	4	0	0.15	0	0.981267	-2.841394	110	19	1.1	0.9;
	5775	1	227.4	40.3	0	0.27	0	1.025402	-4.180541	110	21	1.1	0.9;
	5776	2	0	0	0	0	0	1.039438	-12.890359	20	4	1.1	0.9;
	5777	2	0	0	0	0	0	1.039438	-12.890359	20	4	1.1	0.9;
	5778	1	-2.3	-3.34	0	8.15	0	1.012762	-16.416743	220	4	1.1	0.9;
	5779	1	89.3	25.1	0	2.98	0	1.015081	16.352467	220	5	1.1	0.9;
	5780	1	75.78	46.9	0	3.07	0	1.021718	-10.711519	150	9	1.1	0.9;
	5781	2	0	0	0	0	0	1.039438	-12.890359	20	4	1.1	0.9;
	5782	2	0	0	0	0	0	1.039438	-12.890359	20	4	1.1	0.9;
	5783	1	29.73	0	0	1.34	0	0.990994	-16.127751	110	22	1.1	0.9;
	5784	1	124.9	77	0	23.05	0	1.026131	33.839963	380	3	1.1	0.9;
	5785	1	105.2	24.2	0	0.25	0	1.050514	24.699304	220	5	1.1	0.9;
	5786	1	2.4	1	0.058589	0.961411	0	1.055489	5.337934	220	5	1.1	0.9;
	5787	1	-0	-0	0	0.85	0	1.01752	-4.62634	150	9	1.1	0.9;
	5788	1	0	0	0	14.83	0	0.993679	90.244767	150	1	1.1	0.9;
	5789	1	0	-0	0.27075	43.13435	0	1.082373	9.508502	380	18	1.2	0.8;
	5790	1	37.94	23.5	0	0.85	0	1.008391	-8.4759	150	9	1.1	0.9;
	5791	1	-0	-0	0	0.78	0	1.045831	-3.308284	150	9	1.1	0.9;
	5792	1	0	-0	0	0.89	0	1.068636	22.055101	220	3	1.1	0.9;
	5793	1	-0	-0	0	0.81	0	1.016584	-8.474181	110	12	1.1	0.9;
	5794	1	-0	-0	0	15.55	0	0.976391	21.084011	400	1	1.1	0.9;
	5795	1	101.1	6.6	0	5.66	0	1.014298	-18.383692	220	4	1.1	0.9;
	5796	1	83.19	11.4	0	9.84	0	1.00485	-15.564435	220	4	1.1	0.9;
	5797	2	0	0	0	0	0	1.042461	6.951342	20	22	1.1	0.9;
	5798	1	27.9	4	0	0.02	0	1.01945	6.119322	154	1	1.1	0.9;
	5799	1	62.4	123.4	0.588182	22.609849	0	1.020095	18.921466	220	24	1.1	0.9;
	5800	1	1.1	-0	0	26.91	0	1.007767	-9.473275	220	22	1.1	0.9;
	5801	1	90.5	12.8	0	7.61	0	1.006593	32.135135	154	1	1.1	0.9;
	5802	1	0	0	0	4.14	0	1.025117	16.028526	220	3	1.1	0.9;
	5803	1	-0	-0	0	9.05	0	1.01388	8.701007	220	1	1.1	0.9;
	5804	2	0	0	0	0	0	1.040138	32.45384	20	17	1.1	0.9;
	5805	2	0	0	0	0	0	1.024788	-12.464012	20	4	1.1	0.9;
	5806	1	20.02	15	0	2.11	0	0.987722	33.845836	220	3	1.1	0.9;
	5807	1	74.28	-11.3	0.0391	2.880675	0	1.0266	-8.228562	220	6	1.1	0.9;
	5808	1	0	-0	0.096503	3.353497	0	1.031067	14.367272	220	7	1.1	0.9;
	5809	1	0.4	0.9	0	0.38	0	1.03095	60.569539	380	3	1.1	0.9;
	5810	1	90.4	53.4	0.243268	0.817549	0	1.020064	16.538428	220	5	1.1	0.9;
	5811	1	5.11	3.8	0	1.06	0	0.951832	-5.449631	110	18	1.1	0.9;
	5812	1	-0	-0	0	8.54	0	1.027005	59.333781	380	3	1.1	0.9;
	5813	1	310.8	162.6	0	19.47	0	0.966594	-29.247887	220	10	1.1	0.9;
	5814	1	-0	0	0	0	0	0.950338	10.805302	110	16	1.1	0.9;
	5815	1	-0.19	-0.92	0	0	0	1.025886	34.061284	150	8	1.1	0.9;
	5816	2	0	0	0	0	0	1.035914	35.115322	20	10	1.1	0.9;
	5817	1	0	-0	0	7.4	0	1.035449	9.412244	220	5	1.1	0.9;
	5818	1	25	3.6	0	13	0	1.019498	21.474685	154	1	1.1	0.9;
	5819	1	0	-0	0.079227	0.08058	0	1.023278	-2.708677	220	6	1.1	0.9;
	5820	1	-0	-0	0	1.11	0	1.013406	36.020088	154	1	1.1	0.9;
	5821	1	-0	-0	0	58.09	0	1.135601	0.615015	380	15	1.2	0.8;
	5822	1	0	0	0	5.62	0	1.023112	34.781938	150	8	1.1	0.9;
	5823	1	42.35	7.2	0	1.47	0	1.012684	-20.110839	220	4	1.1	0.9;
	5824	1	19.42	11.8	0	2.6	0	1.023298	6.634271	220	5	1.1	0.9;
	5825	2	0	0	0	0	0	1.021452	28.019771	20	10	1.1	0.9;
	5826	2	0	0	0	0	0	1.013783	30.401512	11	1	1.1	0.9;
	5827	1	127.8	37.6	0.238595	7.981012	0	1.006104	-16.109583	380	6	1.1	0.9;
	5828	1	1.3	0	0	0.57	0	0.993436	-3.814152	110	18	1.1	0.9;
	5829	1	4.2	0.9	0	0.12	0	0.971881	-8.50254	110	18	1.1	0.9;
	5830	2	0	0	0	0	0	1.030828	-0.367949	20	6	1.1	0.9;
	5831	2	0	0	0	0	0	1.020161	16.222206	20	5	1.1	0.9;
	5832	2	0	0	0	0	0	1.020161	16.222206	20	5	1.1	0.9;
	5833	1	220.7	-14.8	0	125.52	0	1.038306	-15.533961	380	4	1.1	0.9;
	5834	1	3.9	1.4	0	0.2	0	1.049794	0.17822	110	19	1.1	0.9;
	5835	1	-0.18	-2.27	0	0	0	1.050845	44.075856	380	8	1.1	0.9;
	5836	1	-0	-0	0	0.84	0	1.026245	18.684578	220	5	1.1	0.9;
	5837	1	5.11	1.1	0	0.33	0	0.958686	-5.800294	110	18	1.1	0.9;
	5838	2	0	0	0	0	0	1.032652	22.317624	20	5	1.1	0.9;
	5839	1	218.5	93.7	0.058873	0.241127	0	1.031387	20.467452	220	5	1.1	0.9;
	5840	1	70.18	19.5	0	3.59	0	1.069097	22.417608	220	5	1.1	0.9;
	5841	1	2.51	-1.64	0	6.26	0	1.024467	-0.601057	220	7	1.1	0.9;
	5842	2	0	0	0	0	0	1.024479	-0.687261	20	18	1.1	0.9;
	5843	1	-0.18	-1.42	0	0	0	1.024282	31.503415	150	8	1.1	0.9;
	5844	1	148	26.1	0	9.44	0	0.9664	4.444682	220	5	1.1	0.9;
	5845	1	159.5	42	0	2.3	0	1.017777	15.68236	220	5	1.1	0.9;
	5846	1	-2.09	-1.43	0	48.397495	0	0.983943	11.381087	380	5	1.1	0.9;
	5847	2	0	0	0	0	0	1.037296	-1.6783	20	3	1.1	0.9;
	5848	1	0	-0	0	28.93	0	1.06333	-7.700665	380	1	1.2	0.8;
	5849	2	0	0	0	0	0	1.031737	11.756042	20	5	1.1	0.9;
	5850	2	0	0	0	0	0	1.035207	-6.870994	20	4	1.1	0.9;
	5851	1	13.9	1.9	0	3.72	0	0.988353	22.737619	154	1	1.1	0.9;
	5852	1	0	0	0.359411	37.9409	0	1.054531	2.335273	380	19	1.1	0.9;
	5853	1	130.5	18.6	0	4.56	0	1.01089	27.117676	154	1	1.1	0.9;
	5854	1	-24.63	159.9	0.141735	64.738265	0	1.012572	-23.024367	380	4	1.1	0.9;
	5855	1	28.63	7.3	0	4.36	0	1.005005	35.477455	110	8	1.1	0.9;
	5856	2	0	0	0	0	0	1.056042	38.515188	20	14	1.1	0.9;
	5857	1	77.19	26.5	0	6.2	0	1.0313	29.629994	220	5	1.1	0.9;
	5858	2	0	0	0	0	0	1.056042	38.515188	20	14	1.1	0.9;
	5859	1	14.15	3.8	0	0.47	0	0.992724	-4.117129	110	18	1.1	0.9;
	5860	1	0	-0	0.039173	0.830827	0	1.018571	-5.556974	220	10	1.1	0.9;
	5861	1	90.5	12.8	0	101.1	0	0.979237	6.572803	400	1	1.1	0.9;
	5862	1	22.73	5.8	0	0.51	0	1.014975	-2.046475	110	3	1.1	0.9;
	5863	1	84.79	12.5	0	2.55	0	1.012632	-27.064346	220	4	1.1	0.9;
	5864	2	0	0	0	0	0	1.043738	-8.975119	20	4	1.1	0.9;
	5865	1	76.6	10.9	0	1.33	0	0.994547	1.423679	154	1	1.1	0.9;
	5866	1	7.01	1.8	0	35.44	0	1.037913	-21.405342	380	10	1.1	0.9;
	5867	1	262.29	-6.21	0.038838	0.691162	0	1.035807	30.09497	220	13	1.1	0.9;
	5868	2	0	0	0	0	0	1.06019	67.358874	20	3	1.1	0.9;
	5869	1	34.64	21.4	0	0.02	0	0.985529	-4.986197	150	9	1.1	0.9;
	5870	1	119.9	17.1	0	8.99	0	1.015465	17.28757	154	1	1.1	0.9;
	5871	1	0	-0	0	0.66	0	1.066444	-4.571592	150	9	1.1	0.9;
	5872	1	36.34	10.4	0	2.67	0	1.03628	29.217017	220	5	1.1	0.9;
	5873	1	-7.91	2.6	0	1.77	0	0.959562	-9.810601	150	9	1.1	0.9;
	5874	1	148.9	23.5	0.019355	6.560645	0	1.029619	20.011321	220	13	1.1	0.9;
	5875	1	26.13	9.7	0	2.37	0	1.034951	19.251286	220	5	1.1	0.9;
	5876	1	-0	0	0	0.99	0	1.002041	-7.8441	150	9	1.1	0.9;
	5877	1	110.8	36.9	0	23.2	0	1.02758	22.650628	220	5	1.1	0.9;
	5878	1	25.23	11.9	0	18.72	0	1.018599	-22.400698	380	4	1.1	0.9;
	5879	1	18.2	2.6	0	6.72	0	0.985272	39.219636	154	1	1.1	0.9;
	5880	2	0	0	0	0	0	1.021915	-9.376379	20	6	1.1	0.9;
	5881	2	0	0	0	0	0	1.035485	43.144579	20	8	1.1	0.9;
	5882	2	0	0	0	0	0	1.052667	56.594903	20	8	1.1	0.9;
	5883	2	0	0	0	0	0	1.052667	56.594903	20	8	1.1	0.9;
	5884	2	0	0	0	0	0	1.052667	56.594903	20	8	1.1	0.9;
	5885	2	0	0	0	0	0	1.021915	-9.376379	20	6	1.1	0.9;
	5886	2	0	0	0	0	0	1.021915	-9.376379	20	6	1.1	0.9;
	5887	2	0	0	0	0	0	1.052481	55.678412	20	8	1.1	0.9;
	5888	1	171.18	28.49	0	0	0	1.060323	-2.554398	220	5	1.1	0.9;
	5889	2	0	0	0	0	0	1.052481	55.678412	20	8	1.1	0.9;
	5890	1	-0.18	-1	0	0	0	1.020215	36.373139	150	8	1.1	0.9;
	5891	1	76.6	10.9	0	7.8	0	1.028495	1.853439	154	1	1.1	0.9;
	5892	1	-0	0	0.649603	14.162287	0	1.131076	7.162518	220	19	1.2	0.8;
	5893	1	22.93	8.6	0.05769	0.49231	0	1.035388	12.710696	110	19	1.1	0.9;
	5894	1	92.3	13.1	0	2.09	0	0.998753	12.721985	154	1	1.1	0.9;
	5895	1	144.8	39.7	0	12.64	0	1.01471	-26.77008	220	4	1.1	0.9;
	5896	2	0	0	0	0	0	1.024375	-1.976838	20	10	1.1	0.9;
	5897	2	0	0	0	0	0	1.054462	48.756838	20	3	1.1	0.9;
	5898	2	0	0	0	0	0	1.054462	48.756838	20	3	1.1	0.9;
	5899	1	0	-0	0.037721	21.232279	0	1.047153	89.840725	220	3	1.1	0.9;
	5900	2	0	0	0	0	0	1.05445	48.75455	20	3	1.1	0.9;
	5901	2	0	0	0	0	0	1.05445	48.75455	20	3	1.1	0.9;
	5902	2	0	0	0	0	0	1.054462	48.756838	20	3	1.1	0.9;
	5903	1	449.4	14.6	0	87.31	0	1.012912	-23.281628	380	4	1.1	0.9;
	5904	1	0	0	0	0.2	0	0.963348	-6.444186	110	21	1.1	0.9;
	5905	2	0	0	0	0	0	0.994923	49.296703	14.4	1	1.1	0.9;
	5906	1	-0	-0	0	0	0	1.025904	-18.853098	150	6	1.1	0.9;
	5907	1	0	0	0	0	0	0.975223	-5.552476	150	9	1.1	0.9;
	5908	1	394.9	-17.3	0	30.22	0	1.01868	-7.446296	380	6	1.1	0.9;
	5909	1	-0	-0	0	14.28	0	1.015628	-22.718473	380	4	1.1	0.9;
	5910	1	158.8	105.3	0	21.76	0	1.017767	42.2857	150	8	1.1	0.9;
	5911	2	0	0	0	0	0	1.037214	32.974337	20	10	1.1	0.9;
	5912	1	0	0	0	154.25	0	1.020286	29.155848	380	3	1.1	0.9;
	5913	1	-0	-0	0.096569	0.453431	0	1.041343	33.230577	380	5	1.1	0.9;
	5914	1	-0	0	0	2.67	0	1.036424	29.228101	220	5	1.1	0.9;
	5915	1	1.7	-5.9	0.290764	53.977697	0	1.051929	-13.750141	380	4	1.1	0.9;
	5916	1	0	0	0	0	0	1.031135	4.071247	380	6	1.1	0.9;
	5917	1	93.7	28.5	0.019768	0.250232	0	1.012556	-5.179216	110	19	1.1	0.9;
	5918	1	20.92	10.3	0	0.68	0	0.956119	-6.683602	110	18	1.1	0.9;
	5919	1	67.27	15.9	0	2.65	0	1.009506	-11.672246	220	4	1.1	0.9;
	5920	1	0	0	0	0	0	1.090479	-4.891767	150	9	1.1	0.9;
	5921	1	-0	-0	0	6.42	0	1.028879	18.453528	220	5	1.1	0.9;
	5922	1	265.7	113.5	0	1.85	0	1.005029	17.484263	220	5	1.1	0.9;
	5923	1	54.96	31.7	0	8.58	0	1.043233	84.231609	220	3	1.1	0.9;
	5924	2	0	0	0	0	0	1.009892	4.868208	20	9	1.1	0.9;
	5925	1	5.51	2	0	0.46	0	0.972084	-6.601209	110	21	1.1	0.9;
	5926	1	58.56	15.9	0	3.75	0	1.017545	6.135969	220	6	1.1	0.9;
	5927	1	-0	-0	0	0.3	0	1.028214	5.959624	380	6	1.1	0.9;
	5928	2	0	0	0	0	0	1.044924	29.406092	20	5	1.1	0.9;
	5929	2	0	0	0	0	0	1.032126	34.816975	20	5	1.1	0.9;
	5930	2	0	0	0	0	0	1.032126	34.816975	20	5	1.1	0.9;
	5931	2	0	0	0	0	0	1.0234	3.527787	20	2	1.1	0.9;
	5932	2	0	0	0	0	0	1.032126	34.816975	20	5	1.1	0.9;
	5933	2	0	0	0	0	0	1.032126	34.816975	20	5	1.1	0.9;
	5934	2	0	0	0	0	0	1.032126	34.816975	20	5	1.1	0.9;
	5935	2	0	0	0	0	0	1.03284	33.745209	20	5	1.1	0.9;
	5936	2	0	0	0	0	0	1.03284	33.745209	20	5	1.1	0.9;
	5937	1	0	0	0	1.18	0	1.030929	7.980983	220	5	1.1	0.9;
	5938	1	37.24	40.5	0	67.07	0	1.019496	15.814985	380	6	1.1	0.9;
	5939	1	0	-0	0	19.37	0	1.030376	55.872614	380	3	1.1	0.9;
	5940	1	-0	-0	0	0	0	1.020557	9.624602	20	14	1.1	0.9;
	5941	1	8.9	1.3	0	3.55	0	1.015548	14.593293	154	1	1.1	0.9;
	5942	1	76.6	10.9	0	3.24	0	1.015926	14.190338	154	1	1.1	0.9;
	5943	1	1.9	0.7	0	1.34	0	1.021785	5.75129	110	19	1.1	0.9;
	5944	1	0	0	0	0.57	0	0.964065	-6.238285	110	18	1.1	0.9;
	5945	1	0	0	0.233271	18.215327	0	1.058336	25.706117	380	14	1.1	0.9;
	5946	1	-0	-0	0	23.74	0	1.025378	12.036667	220	3	1.1	0.9;
	5947	1	0	0	0	17.54	0	1.061006	27.420501	220	1	1.1	0.9;
	5948	1	-0	-0	0	0.89	0	1.016554	19.011609	220	5	1.1	0.9;
	5949	1	-0	-0	0	107.36	0	0.998206	29.448628	400	1	1.1	0.9;
	5950	1	27.57	-33.38	0	13.08	0	1.048029	39.867702	220	13	1.1	0.9;
	5951	1	-0	0	0.038161	34.951839	0	1.057237	76.895536	380	3	1.1	0.9;
	5952	1	0	0	0	1.12	0	0.981012	-5.330129	110	18	1.1	0.9;
	5953	1	32.84	20.3	0	7.23	0	1.049406	-5.456139	150	9	1.1	0.9;
	5954	1	24.93	22.9	0	2.47	0	1.007312	-8.549741	150	9	1.1	0.9;
	5955	1	-0	-0	0	0	0	1.013531	-19.654143	20	4	1.1	0.9;
	5956	1	-0	-0	0	0	0	1.013531	-19.654143	20	4	1.1	0.9;
	5957	1	-0	-0	0	0	0	1.013531	-19.654143	20	4	1.1	0.9;
	5958	1	-0	-0	0	0	0	1.013531	-19.654143	20	4	1.1	0.9;
	5959	1	347.7	96.6	0	101.73	0	1.045284	8.595428	380	16	1.1	0.9;
	5960	1	123.1	23.5	0.039429	0.000571	0	1.023365	19.406767	220	5	1.1	0.9;
	5961	2	0	0	0	0	0	1.022689	-16.891323	20	4	1.1	0.9;
	5962	1	74.78	-31.4	0	1.74	0	1.016779	36.467514	110	8	1.1	0.9;
	5963	2	0	0	0	0	0	1.040642	8.532255	20	14	1.1	0.9;
	5964	2	0	0	0	0	0	1.008348	30.337205	14.4	1	1.1	0.9;
	5965	2	0	0	0	0	0	0.985995	10.771118	15	1	1.1	0.9;
	5966	1	0	0	0	0.91	0	1.011585	-20.232733	220	4	1.1	0.9;
	5967	1	0	0	0	0	0	1.097943	64.987411	380	8	1.2	0.8;
	5968	1	44.35	19.3	0	3.25	0	1.046992	24.651278	220	5	1.1	0.9;
	5969	1	0	-0	0	5.66	0	1.034951	-1.196557	154	1	1.1	0.9;
	5970	1	6.31	-0	0	0.42	0	1.010211	-20.853988	220	4	1.1	0.9;
	5971	1	2.5	-0	0	0.8	0	1.000938	3.696158	110	18	1.1	0.9;
	5972	1	131.35	63.6	0	8.55	0	1.036314	28.432012	220	1	1.1	0.9;
	5973	2	0	0	0	0	0	1.066829	5.050047	20	5	1.1	0.9;
	5974	1	14.72	5.3	0	0.49	0	1.064411	1.408548	110	19	1.1	0.9;
	5975	1	111.4	15.9	0	1.04	0	1.014887	-2.059817	154	1	1.1	0.9;
	5976	1	-0	-0	0	3.51	0	1.027966	-0.065388	150	9	1.1	0.9;
	5977	2	0	0	0	0	0	1.028438	-19.446471	20	4	1.1	0.9;
	5978	2	0	0	0	0	0	1.02806	-6.75496	20	6	1.1	0.9;
	5979	2	0	0	0	0	0	1.028283	16.586422	20	14	1.1	0.9;
	5980	1	18.22	2.3	0	0.97	0	0.968446	-9.259647	110	18	1.1	0.9;
	5981	1	79.29	40.1	0.040286	34.579857	0	1.015899	23.470813	380	5	1.1	0.9;
	5982	2	0	0	0	0	0	1.031159	7.793526	11	1	1.1	0.9;
	5983	1	35.44	22	0	0.61	0	0.986074	-5.521684	150	9	1.1	0.9;
	5984	2	0	0	0	0	0	1.016616	-4.658699	20	4	1.1	0.9;
	5985	1	0	-0	0	3.64	0	1.002412	27.878654	154	1	1.1	0.9;
	5986	1	4.2	2	0	0.21	0	0.971925	7.112503	110	18	1.1	0.9;
	5987	2	0	0	0	0	0	1.041616	36.679648	20	5	1.1	0.9;
	5988	2	0	0	0	0	0	1.041616	36.679648	20	5	1.1	0.9;
	5989	2	0	0	0	0	0	1.039304	37.982317	20	5	1.1	0.9;
	5990	2	0	0	0	0	0	1.039304	37.982317	20	5	1.1	0.9;
	5991	1	-0	0	0	1.01	0	1.078833	-3.907281	150	9	1.1	0.9;
	5992	2	0	0	0	0	0	1.015185	-4.986575	20	4	1.1	0.9;
	5993	1	125	107.2	0.202178	53.177543	0	1.031888	35.715294	380	3	1.1	0.9;
	5994	1	0	-0	0	0.53	0	1.023642	0.458031	110	3	1.1	0.9;
	5995	1	0	-0.1	0.080273	-0.060205	0	1.017687	25.51753	220	6	1.1	0.9;
	5996	1	2.5	-0	0	0.11	0	0.995065	-3.415633	110	18	1.1	0.9;
	5997	1	0	0	0.119927	0.900061	0	1.029204	31.870214	220	14	1.1	0.9;
	5998	1	0	0	0	23.27	0	1.044885	42.101051	380	8	1.1	0.9;
	5999	2	0	0	0	0	0	1.03143	0.621932	20	6	1.1	0.9;
	6000	1	-0	-0	0.16	31.35	0	1.003181	0.104403	380	7	1.1	0.9;
	6001	1	0	0	0	1.51	0	1.071599	-4.221692	150	9	1.1	0.9;
	6002	2	0	0	0	0	0	1.027152	23.121261	20	5	1.1	0.9;
	6003	2	0	0	0	0	0	1.027152	23.121261	20	5	1.1	0.9;
	6004	1	0	-0	0	6.9	0	1.018643	-25.676655	220	4	1.1	0.9;
	6005	2	0	0	0	0	0	1.027152	23.121261	20	5	1.1	0.9;
	6006	2	0	0	0	0	0	1.027152	23.121261	20	5	1.1	0.9;
	6007	2	0	0	0	0	0	1.027152	23.121261	20	5	1.1	0.9;
	6008	2	0	0	0	0	0	1.027152	23.121261	20	5	1.1	0.9;
	6009	2	0	0	0	0	0	1.016602	20.114087	6.3	1	1.1	0.9;
	6010	1	113	57.3	0	12.82	0	1.016233	-10.469824	150	9	1.1	0.9;
	6011	1	135.9	4	0.392133	6.065901	0	1.032195	20.523086	220	5	1.1	0.9;
	6012	1	0	0	0.117038	0.062468	0	1.02714	64.681878	220	3	1.1	0.9;
	6013	1	62.7	9	0	8.02	0	1.027412	1.646975	154	1	1.1	0.9;
	6014	2	0	0	0	0	0	1.03143	0.621932	20	6	1.1	0.9;
	6015	1	-0	-0	0	0.88	0	1.028327	8.752555	220	1	1.1	0.9;
	6016	2	0	0	0	0	0	0.991806	18.024083	20	18	1.1	0.9;
	6017	2	0	0	0	0	0	1.050579	35.320834	20	8	1.1	0.9;
	6018	1	195.2	92.2	0	11.67	0	0.981626	15.778389	220	6	1.1	0.9;
	6019	1	204	29	0	2.28	0	1.027596	-0.273797	154	1	1.1	0.9;
	6020	1	0	-0	0	28.29	0	1.10805	-0.569645	380	1	1.2	0.8;
	6021	1	0	-0	0	0.05	0	0.996578	38.818649	220	3	1.1	0.9;
	6022	1	40.24	24.9	0	1.81	0	1.009788	-5.34131	150	9	1.1	0.9;
	6023	2	0	0	0	0	0	1.045612	86.470349	20	8	1.1	0.9;
	6024	1	225.8	123	0.305983	9.235014	0	0.998597	-10.549173	220	6	1.1	0.9;
	6025	2	0	0	0	0	0	1.026207	-11.793103	20	4	1.1	0.9;
	6026	1	65.27	6	0	3.89	0	1.022727	-30.889809	220	4	1.1	0.9;
	6027	1	47.65	7.1	0	7.3	0	1.010815	-20.730474	220	4	1.1	0.9;
	6028	1	0	-0	0	4.05	0	0.982613	18.688338	400	1	1.1	0.9;
	6029	1	61.67	28.3	0	0.1	0	0.955097	-16.549447	220	10	1.1	0.9;
	6030	1	173	11.3	0	8.41	0	1.063003	10.873519	220	3	1.1	0.9;
	6031	1	0	-0	0	1.23	0	1.06091	18.606651	220	5	1.1	0.9;
	6032	1	128	72.9	0	14.95	0	1.030262	45.324578	380	3	1.1	0.9;
	6033	1	5.31	2	0	0.17	0	0.970512	4.936778	110	18	1.1	0.9;
	6034	2	0	0	0	0	0	1.009892	4.868208	20	9	1.1	0.9;
	6035	1	40.34	10.4	0	7.02	0	1.022485	41.156373	110	8	1.1	0.9;
	6036	1	0.5	12.4	0	0.25	0	1.030326	29.794757	220	3	1.1	0.9;
	6037	1	4.2	0.6	0	2.84	0	1.010682	19.216431	154	1	1.1	0.9;
	6038	2	0	0	0	0	0	0.981513	33.636923	15.75	1	1.1	0.9;
	6039	2	0	0	0	0	0	1.054609	40.170592	20	14	1.1	0.9;
	6040	2	0	0	0	0	0	1.054609	40.170592	20	14	1.1	0.9;
	6041	1	-0	-0	0	0.34	0	0.985319	0.658331	150	9	1.1	0.9;
	6042	1	-0	0	0	5.16	0	1.021057	32.465877	154	1	1.1	0.9;
	6043	1	13.01	3.8	0	3.59	0	1.034131	30.32841	220	5	1.1	0.9;
	6044	1	0	0	0.009025	43.164575	0	1.042458	13.473477	380	16	1.1	0.9;
	6045	1	64.77	6.4	0	3.84	0	1.003423	-17.153993	220	4	1.1	0.9;
	6046	1	-0	-0	0	84.33	0	1.008604	11.47174	380	3	1.1	0.9;
	6047	1	9.91	5.4	0	4.6	0	1.036671	5.406937	220	5	1.1	0.9;
	6048	2	0	0	0	0	0	1.055592	83.040122	20	3	1.1	0.9;
	6049	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	6050	1	2.2	0.5	0	2.31	0	0.97203	-23.552605	150	10	1.1	0.9;
	6051	1	0	0	0	18.49	0	1.01017	51.775874	380	3	1.1	0.9;
	6052	1	-0	0	0.039884	3.820116	0	1.009109	14.405897	220	5	1.1	0.9;
	6053	1	-0	0	0	19.2	0	1.063491	70.230354	380	3	1.1	0.9;
	6054	1	-0.54	-1	0	4.8	0	1.046321	-3.291866	150	9	1.1	0.9;
	6055	2	0	0	0	0	0	1.030356	27.865848	20	6	1.1	0.9;
	6056	1	0	0	0.18226	20.898116	0	1.003143	62.493742	220	3	1.1	0.9;
	6057	1	0	0	0	1.18	0	0.987794	23.062821	380	5	1.1	0.9;
	6058	2	0	0	0	0	0	1.036943	14.917752	20	7	1.1	0.9;
	6059	1	230.42	-7.87	0	5.13	0	1.01934	-18.297835	220	4	1.1	0.9;
	6060	1	44.05	14.6	0	0.6	0	0.927915	-10.328649	110	21	1.1	0.9;
	6061	1	-0	0	0.039589	50.801364	0	1.009529	21.188144	380	5	1.1	0.9;
	6062	1	-0	-0	0	1.48	0	1.008772	22.488054	220	6	1.1	0.9;
	6063	1	39.14	7.7	0	11.68	0	1.024416	10.242461	220	5	1.1	0.9;
	6064	1	-0	-0	0.060014	-0.060014	0	0.990322	11.749369	220	5	1.1	0.9;
	6065	1	132.4	-10.4	0	3.02	0	1.057121	10.685511	220	3	1.1	0.9;
	6066	2	0	0	0	0	0	1.019403	-19.675877	20	10	1.1	0.9;
	6067	1	0	0.01	0	2.1	0	1.015409	36.155958	150	8	1.1	0.9;
	6068	2	0	0	0	0	0	1.037973	11.471538	20	14	1.1	0.9;
	6069	1	39.84	17.6	0.231215	2.096955	0	1.015065	-1.760043	110	21	1.1	0.9;
	6070	1	25.43	7.9	0	1.03	0	1.014621	42.928119	110	8	1.1	0.9;
	6071	1	145.8	42.4	0	2.34	0	1.003711	-22.214236	220	6	1.1	0.9;
	6072	2	0	0	0	0	0	1.00419	31.464758	11	1	1.1	0.9;
	6073	1	-0	-0	0	0.09	0	1.046339	-4.535753	220	5	1.1	0.9;
	6074	1	52.6	7.5	0	3.66	0	0.999472	31.886733	154	1	1.1	0.9;
	6075	1	0	0	0	49.86	0	1.075296	2.671869	380	19	1.1	0.9;
	6076	2	0	0	0	0	0	1.000808	-3.647252	20	18	1.1	0.9;
	6077	1	13.64	9	0	1.7	0	1.015014	-12.066594	150	9	1.1	0.9;
	6078	1	137.8	22.9	0	6.55	0	0.999152	-2.431726	220	6	1.1	0.9;
	6079	1	0	0	0	20.86	0	1.015055	32.061061	380	3	1.1	0.9;
	6080	1	0	0	0	11.85	0	1.004169	6.716692	220	3	1.1	0.9;
	6081	1	-0.62	-0	0	75.073626	0	1.011307	8.711375	380	5	1.1	0.9;
	6082	1	0	-0	0	30.64	0	1.063182	4.144091	380	1	1.1	0.9;
	6083	1	11.11	4.1	0	0.8	0	1.037822	28.36056	220	5	1.1	0.9;
	6084	1	310.2	4.5	0	8.51	0	0.988625	9.51287	220	5	1.1	0.9;
	6085	2	0	0	0	0	0	1.061212	43.548638	20	2	1.1	0.9;
	6086	1	128.18	35.03	0	1.76	0	0.992135	88.958859	150	1	1.1	0.9;
	6087	2	0	0	0	0	0	1.044747	17.319249	20	5	1.1	0.9;
	6088	2	0	0	0	0	0	1.016616	-4.658699	20	4	1.1	0.9;
	6089	2	0	0	0	0	0	1.042545	18.989939	20	5	1.1	0.9;
	6090	2	0	0	0	0	0	1.030904	22.114026	20	5	1.1	0.9;
	6091	2	0	0	0	0	0	1.030904	22.114026	20	5	1.1	0.9;
	6092	2	0	0	0	0	0	1.032747	21.521572	20	5	1.1	0.9;
	6093	2	0	0	0	0	0	1.044747	17.319249	20	5	1.1	0.9;
	6094	1	6.71	1.3	0	0.34	0	1.014655	7.023289	110	18	1.1	0.9;
	6095	1	-0	8.5	0	0.67	0	1.080121	-12.378342	150	9	1.2	0.8;
	6096	2	0	0	0	0	0	1.022821	44.842821	20	10	1.1	0.9;
	6097	2	0	0	0	0	0	1.042545	18.989939	20	5	1.1	0.9;
	6098	2	0	0	0	0	0	1.042545	18.989939	20	5	1.1	0.9;
	6099	1	91.4	-88.2	0	0.29	0	1.023281	24.830016	220	3	1.1	0.9;
	6100	1	0	-0	0	1.41	0	1.013701	-8.764256	110	12	1.1	0.9;
	6101	1	-0	-0	0	0.67	0	1.033145	20.10233	220	5	1.1	0.9;
	6102	1	-0	0	0	2.69	0	1.029399	77.694517	220	3	1.1	0.9;
	6103	2	0	0	0	0	0	1.082967	1.032196	20	9	1.1	0.9;
	6104	1	110.7	94.3	0	6.43	0	0.992505	32.70807	220	3	1.1	0.9;
	6105	1	153.4	86.8	0	53.98	0	1.045772	-14.51788	380	4	1.1	0.9;
	6106	1	0.2	109.4	0	0	0	1.076107	86.31696	380	3	1.2	0.8;
	6107	2	0	0	0	0	0	1.043087	11.638661	20	14	1.1	0.9;
	6108	1	121.4	40	0	11.16	0	1.012364	11.991469	380	3	1.1	0.9;
	6109	1	175.2	70.7	0	109.1	0	1.013488	15.812501	380	6	1.1	0.9;
	6110	1	0	0	0	0	0	0.994612	-8.532772	150	9	1.1	0.9;
	6111	1	29.03	10.1	0	2.41	0	0.998443	-1.488477	110	19	1.1	0.9;
	6112	1	-0	-0	0	4.37	0	1.123449	-0.180616	380	9	1.2	0.8;
	6113	1	27.03	69.9	0.230833	9.787258	0	1.038344	25.364547	220	13	1.1	0.9;
	6114	1	0	0	0	14.62	0	1.010168	12.38255	380	3	1.1	0.9;
	6115	1	325.2	38.1	0	19.99	0	1.012394	-17.98108	380	4	1.1	0.9;
	6116	1	0	0	0	0	0	1.0171	-4.093262	150	9	1.1	0.9;
	6117	1	113.4	61.9	0	1.51	0	1.02244	38.825086	380	3	1.1	0.9;
	6118	1	200.19	90.59	0	24.07	0	0.999051	-22.980681	220	6	1.1	0.9;
	6119	1	62.07	20	0	0.25	0	0.953757	-7.790324	110	21	1.1	0.9;
	6120	1	0	0	0	12.88	0	0.984959	-4.457638	220	10	1.1	0.9;
	6121	2	0	0	0	0	0	1.029542	-3.096329	20	6	1.1	0.9;
	6122	1	7.91	3.8	0	0.22	0	1.008844	-3.098739	110	18	1.1	0.9;
	6123	2	0	0	0	0	0	1.021005	21.987746	20	14	1.1	0.9;
	6124	1	0	-0	0.019597	0.530403	0	0.998791	-11.061968	110	12	1.1	0.9;
	6125	1	34.9	5	0	1.72	0	1.010111	14.513609	154	1	1.1	0.9;
	6126	1	0	-0	0	0.71	0	1.015624	-8.566527	150	9	1.1	0.9;
	6127	1	-0	0	0	0	0	1.010949	-10.142394	20	4	1.1	0.9;
	6128	1	-0	-0	1.098741	87.634731	0	1.090805	71.912751	380	3	1.2	0.8;
	6129	1	144.5	40.9	0	17.35	0	1.0104	15.56306	220	7	1.1	0.9;
	6130	2	0	0	0	0	0	1.019692	25.075054	11	1	1.1	0.9;
	6131	1	20.22	5.7	0	3.25	0	1.014679	-17.550577	220	4	1.1	0.9;
	6132	2	0	0	0	0	0	1.033224	23.231932	11	1	1.1	0.9;
	6133	1	-0	-0	0.349725	3.347992	0	1.027165	-6.569477	380	6	1.1	0.9;
	6134	1	0	-0	0.058236	0.481764	0	1.005425	-4.989476	110	18	1.1	0.9;
	6135	2	0	0	0	0	0	1.031189	-1.476188	20	20	1.1	0.9;
	6136	1	195.6	-1	0	3.19	0	1.017021	-18.281991	220	4	1.1	0.9;
	6137	1	29.3	90.5	0.475938	17.413385	0	1.019529	-16.098943	380	4	1.1	0.9;
	6138	2	0	0	0	0	0	1.030727	91.348864	20	3	1.1	0.9;
	6139	2	0	0	0	0	0	1.030727	91.348864	20	3	1.1	0.9;
	6140	2	0	0	0	0	0	1.030727	91.348864	20	3	1.1	0.9;
	6141	2	0	0	0	0	0	1.028568	11.503252	20	2	1.1	0.9;
	6142	1	-42.25	2.8	0	0.7	0	1.01831	-6.363543	110	1	1.1	0.9;
	6143	1	26.73	9.6	0	2.19	0	0.98312	-3.040745	110	19	1.1	0.9;
	6144	1	-0	0	0	5.06	0	1.028983	7.369286	220	5	1.1	0.9;
	6145	1	-0	-0.3	0.119513	-0.089635	0	1.013987	-14.578261	380	6	1.1	0.9;
	6146	1	-0	-0	0.090774	6.020444	0	1.053174	4.050736	220	18	1.1	0.9;
	6147	1	59.17	19.3	0	1.16	0	0.980264	24.515922	150	8	1.1	0.9;
	6148	1	200.65	30.72	0	9.72	0	1.019143	2.631633	220	5	1.1	0.9;
	6149	1	-0	0	0	42.131932	0	1.04485	18.305539	380	24	1.1	0.9;
	6150	1	12.71	2.5	0	0.83	0	0.977987	-5.025449	110	18	1.1	0.9;
	6151	1	0	-0	0.497694	-0.387095	0	1.074706	42.983853	380	8	1.1	0.9;
	6152	1	0	0	0	2.06	0	1.013998	19.384747	154	1	1.1	0.9;
	6153	1	0	0	0	0	0	0.95811	-5.43376	150	9	1.1	0.9;
	6154	1	1110.6	290.8	0.535741	178.433223	0	1.068588	31.972979	380	14	1.1	0.9;
	6155	2	0	0	0	0	0	1.018029	-17.367562	20	4	1.1	0.9;
	6156	2	0	0	0	0	0	1.024861	-2.735628	20	10	1.1	0.9;
	6157	1	-3.2	-1.05	0	48.471563	0	1.027114	22.163566	380	5	1.1	0.9;
	6158	1	9.01	4.9	0	0.64	0	1.065116	30.409943	220	5	1.1	0.9;
	6159	1	-0	-0	0	0	0	1.015642	-21.99726	20	4	1.1	0.9;
	6160	1	-0	-0	0	0	0	1.015642	-21.99726	20	4	1.1	0.9;
	6161	2	0	0	0	0	0	1.035942	7.607952	20	2	1.1	0.9;
	6162	2	0	0	0	0	0	1.035942	7.607952	20	2	1.1	0.9;
	6163	2	0	0	0	0	0	1.024003	-9.536137	20	4	1.1	0.9;
	6164	1	50.3	7.2	0	8.06	0	1.001961	33.696499	154	1	1.1	0.9;
	6165	1	-0	0	0	0	0	1.014678	-19.145251	20	4	1.1	0.9;
	6166	1	-0	0	0	0	0	1.014678	-19.145251	20	4	1.1	0.9;
	6167	2	0	0	0	0	0	1.024289	-14.563347	20	4	1.1	0.9;
	6168	1	-0	0	0	0	0	1.014678	-19.145251	20	4	1.1	0.9;
	6169	1	172.5	48.7	0	54.92	0	1.046319	35.394029	380	14	1.1	0.9;
	6170	1	-0	0	0	0	0	1.014678	-19.145251	20	4	1.1	0.9;
	6171	2	0	0	0	0	0	1.011745	17.062335	13.8	1	1.1	0.9;
	6172	1	54.32	7.98	0	13.58	0	0.99209	89.123449	150	1	1.1	0.9;
	6173	1	318.3	76.8	0.140637	63.274356	0	1.026557	7.921498	380	7	1.1	0.9;
	6174	1	53.86	18.7	0	3.48	0	1.003336	-13.963068	110	22	1.1	0.9;
	6175	1	5.11	-0.1	0	3.73	0	0.99484	-3.165135	110	18	1.1	0.9;
	6176	1	0	0	0	0	0	0.976527	-15.468409	150	9	1.1	0.9;
	6177	1	52.46	5.2	0	11.9	0	1.018651	-12.512781	220	4	1.1	0.9;
	6178	1	5.81	1.7	0	0.42	0	1.024234	16.336837	220	5	1.1	0.9;
	6179	1	-0	-0	0	0	0	1.018329	-16.147917	20	4	1.1	0.9;
	6180	1	-0	0	0	39.36	0	1.026563	26.257444	380	3	1.1	0.9;
	6181	1	67.07	34	0.436529	7.949365	0	1.021802	6.666844	220	11	1.1	0.9;
	6182	1	0	-0	0	10.1	0	1.127406	-0.281581	380	15	1.2	0.8;
	6183	1	0	-0	0	0	0	1.050679	44.063233	380	8	1.1	0.9;
	6184	1	-2.29	0.3	0.077775	2.712225	0	1.044427	-4.588031	150	9	1.1	0.9;
	6185	1	0	0	0.01953	10.17047	0	1.027975	4.169052	220	7	1.1	0.9;
	6186	1	-0	0	0	0.79	0	1.075847	0.359401	150	9	1.1	0.9;
	6187	1	149	19.8	0	23.89	0	1.029931	14.928203	380	5	1.1	0.9;
	6188	1	-806.3	-50.1	0	4.19	0	1.07128	88.404374	380	3	1.1	0.9;
	6189	1	-0	0	0	17.29	0	1.035738	7.298143	220	5	1.1	0.9;
	6190	1	-0.17	-8.06	0	0.83	0	1.024006	37.242884	110	8	1.1	0.9;
	6191	1	0	0	0.1161	64.50319	0	1.028754	93.767882	380	1	1.1	0.9;
	6192	2	0	0	0	0	0	1.031157	34.96917	11	1	1.1	0.9;
	6193	2	0	0	0	0	0	1.025991	-18.980906	20	6	1.1	0.9;
	6194	2	0	0	0	0	0	1.025991	-18.980906	20	6	1.1	0.9;
	6195	2	0	0	0	0	0	1.019411	-14.420581	20	10	1.1	0.9;
	6196	1	-0	-0	0	7.6	0	1.026879	64.663185	220	3	1.1	0.9;
	6197	1	27.63	9.2	0	0.58	0	0.978193	-8.678554	110	21	1.1	0.9;
	6198	1	6.73	14.7	0	11.51	0	1.033645	-10.792881	150	9	1.1	0.9;
	6199	2	0	0	0	0	0	1.021456	23.220478	11	1	1.1	0.9;
	6200	2	0	0	0	0	0	0.98149	26.04182	6.3	1	1.1	0.9;
	6201	2	0	0	0	0	0	1.048523	41.360645	20	5	1.1	0.9;
	6202	2	0	0	0	0	0	1.048523	41.360645	20	5	1.1	0.9;
	6203	2	0	0	0	0	0	1.050193	41.315696	20	5	1.1	0.9;
	6204	2	0	0	0	0	0	1.050193	41.315696	20	5	1.1	0.9;
	6205	1	0	0	0	1.44	0	1.013177	2.793231	110	18	1.1	0.9;
	6206	1	-20.56	-3.88	0	1.72	0	1.010737	-10.615253	220	4	1.1	0.9;
	6207	1	0	-0	0	122.87	0	1.037904	12.680695	380	3	1.1	0.9;
	6208	2	0	0	0	0	0	1.062519	23.472046	20	5	1.1	0.9;
	6209	1	119.3	106.3	0.058752	6.33104	0	0.996523	91.383121	150	1	1.1	0.9;
	6210	2	0	0	0	0	0	1.045359	10.101302	20	14	1.1	0.9;
	6211	1	57.76	16	0	2.39	0	0.980902	-17.246984	220	4	1.1	0.9;
	6212	2	0	0	0	0	0	1.021013	29.620439	380	5	1.1	0.9;
	6213	2	0	0	0	0	0	1.029044	-14.386411	20	6	1.1	0.9;
	6214	1	11.01	3.1	0	0.14	0	1.01222	-5.332586	110	19	1.1	0.9;
	6215	2	0	0	0	0	0	1.023729	-10.001413	20	6	1.1	0.9;
	6216	1	14.26	-1.13	0	0.65	0	0.992242	-4.42905	110	18	1.1	0.9;
	6217	1	0	0	0	8.95	0	1.064596	12.664928	220	5	1.1	0.9;
	6218	2	0	0	0	0	0	1.029044	-14.386411	20	6	1.1	0.9;
	6219	2	0	0	0	0	0	1.025649	-20.066719	20	6	1.1	0.9;
	6220	2	0	0	0	0	0	1.023566	-9.987944	20	6	1.1	0.9;
	6221	1	0	0	0	0	0	1.013445	-14.636462	20	6	1.1	0.9;
	6222	1	19.52	6.4	0	0.03	0	1.049638	-6.515397	150	9	1.1	0.9;
	6223	1	0	-0	0	42.02	0	0.976033	36.171732	400	1	1.1	0.9;
	6224	1	-0	-0	0	8.08	0	1.012976	-20.015955	220	4	1.1	0.9;
	6225	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	6226	2	0	0	0	0	0	1.062007	32.438041	20	14	1.1	0.9;
	6227	2	0	0	0	0	0	1.050331	30.533232	20	14	1.1	0.9;
	6228	1	190.7	52	0	20.57	0	1.023396	15.789487	220	5	1.1	0.9;
	6229	1	44.35	14.8	0	0.36	0	0.980794	-8.763251	110	21	1.1	0.9;
	6230	1	-0	-0	0	93.61	0	1.01982	-15.989494	380	4	1.1	0.9;
	6231	1	269.22	97.24	0	0	0	1.018103	0.252161	110	24	1.1	0.9;
	6232	1	188.48	128.3	0	18.06	0	1.001314	0.650862	220	6	1.1	0.9;
	6233	1	72.18	49.6	0	37.52	0	1.021333	42.692724	150	8	1.1	0.9;
	6234	1	24.93	9.6	0	0.09	0	1.059812	21.257384	220	5	1.1	0.9;
	6235	1	24.23	5.5	0	1.18	0	1.017659	-13.828547	110	22	1.1	0.9;
	6236	1	33.54	20.8	0	2.42	0	0.86014	-15.783699	150	9	1.2	0.8;
	6237	1	0	-0	0	3.24	0	1.024959	59.105911	220	8	1.1	0.9;
	6238	1	57.06	15.6	0	2.03	0	1.014402	-27.500291	220	4	1.1	0.9;
	6239	1	-10.71	3.5	0	0.58	0	1.0484	-3.629317	150	9	1.1	0.9;
	6240	1	47.15	15.7	0	5.41	0	1.0128	11.596616	220	5	1.1	0.9;
	6241	1	0	-0	0	0.56	0	1.021588	8.255339	110	18	1.1	0.9;
	6242	1	105	18	0	5.41	0	1.016154	-27.023897	220	4	1.1	0.9;
	6243	1	44.55	27.6	0	1.86	0	1.031416	-7.138703	150	9	1.1	0.9;
	6244	1	2.7	0.4	0	4.05	0	1.014131	22.41328	154	1	1.1	0.9;
	6245	1	-0	0	0	6.76	0	1.013743	41.891899	220	3	1.1	0.9;
	6246	1	100.42	58.37	0	40.35	0	1.019224	34.644067	380	13	1.1	0.9;
	6247	1	9.51	5.9	0	0.7	0	1.014677	-8.666869	150	9	1.1	0.9;
	6248	1	0	-0	0.019898	2.400102	0	1.004043	-14.131992	150	10	1.1	0.9;
	6249	1	-0	0	0	0.54	0	1.033678	0.121424	150	9	1.1	0.9;
	6250	1	-0	0	0	7.91	0	1.011682	29.949869	380	3	1.1	0.9;
	6251	1	0	0	0	0	0	1.013533	-4.833141	150	9	1.1	0.9;
	6252	1	-0	0	0	0	0	1.017479	-21.662554	20	4	1.1	0.9;
	6253	1	-11.81	-4.02	0	0.37	0	1.001763	-1.609966	110	18	1.1	0.9;
	6254	2	0	0	0	0	0	1.051727	31.564136	20	5	1.1	0.9;
	6255	1	-0	-0	0	0.74	0	0.991931	48.589507	154	1	1.1	0.9;
	6256	1	-0	-0	0	0.88	0	1.024815	-7.144323	150	9	1.1	0.9;
	6257	1	61.4	8.8	0	4.24	0	1.000676	17.215156	154	1	1.1	0.9;
	6258	1	70.78	19.6	0	7.9	0	1.027842	20.26667	220	5	1.1	0.9;
	6259	1	303.7	247.8	0	8.79	0	0.977081	48.792581	220	3	1.1	0.9;
	6260	1	24.93	8.2	0	0.6	0	0.950103	-7.728147	110	21	1.1	0.9;
	6261	1	145.7	83.6	0.135375	53.822175	0	1.067116	5.990973	380	18	1.2	0.8;
	6262	2	0	0	0	0	0	1.055881	11.701666	20	18	1.1	0.9;
	6263	1	66.97	45.1	0	2.04	0	0.927991	5.596329	110	18	1.1	0.9;
	6264	1	184.7	71.5	0	95.42	0	1.005094	28.687557	150	8	1.1	0.9;
	6265	1	70.78	28	0	1.49	0	1.015174	15.585606	220	5	1.1	0.9;
	6266	1	35.49	11.53	0	1.31	0	1.013273	-26.453738	220	4	1.1	0.9;
	6267	1	5.71	2.1	0	0.31	0	1.060164	0.669784	110	19	1.1	0.9;
	6268	1	-0	-0	0	60.39	0	1.015951	12.417775	380	3	1.1	0.9;
	6269	1	-0	0	0	1.63	0	0.991926	-11.44705	150	10	1.1	0.9;
	6270	1	5.5	0.8	0	9.28	0	1.013145	33.554518	154	1	1.1	0.9;
	6271	1	9.61	62.4	0.353037	29.405802	0	1.015168	-6.517912	220	4	1.1	0.9;
	6272	2	0	0	0	0	0	1.003919	32.082214	11	1	1.1	0.9;
	6273	1	11.91	4.3	0	0.47	0	1.033074	12.968949	110	19	1.1	0.9;
	6274	1	-0	0	0	11.4	0	1.041697	19.55122	220	5	1.1	0.9;
	6275	1	35.84	14.1	0	50.11	0	1.029062	31.074575	150	8	1.1	0.9;
	6276	2	0	0	0	0	0	1.060659	20.876727	20	5	1.1	0.9;
	6277	2	0	0	0	0	0	1.060659	20.876838	20	5	1.1	0.9;
	6278	2	0	0	0	0	0	1.024788	-12.464012	20	4	1.1	0.9;
	6279	2	0	0	0	0	0	1.060659	20.876727	20	5	1.1	0.9;
	6280	2	0	0	0	0	0	1.022087	-3.367486	20	6	1.1	0.9;
	6281	1	22.22	6	0	0.43	0	1.005114	2.765793	110	18	1.1	0.9;
	6282	1	0	0	0	8.45	0	0.993797	0.50435	220	6	1.1	0.9;
	6283	1	214.4	52.9	0	2.59	0	1.031273	9.700247	220	7	1.1	0.9;
	6284	1	124.8	23.5	0	5.84	0	0.955228	7.418913	220	16	1.1	0.9;
	6285	1	0	-0	0	14.49	0	1.085917	72.214157	380	3	1.2	0.8;
	6286	1	-253.24	-29.59	0	10.61	0	1.012925	94.491383	150	1	1.1	0.9;
	6287	1	0.5	0.3	0	0.07	0	1.047273	-2.403889	150	9	1.1	0.9;
	6288	1	-0	0	0	1.09	0	1.069053	21.801014	220	3	1.1	0.9;
	6289	1	0	0	0	0	0	1.00234	-7.016126	150	9	1.1	0.9;
	6290	1	197.7	67.65	0	13.58	0	1.024233	8.83079	220	5	1.1	0.9;
	6291	2	0	0	0	0	0	1.040138	32.45384	20	17	1.1	0.9;
	6292	1	-0	-0	0	7.3	0	1.035813	3.435348	154	1	1.1	0.9;
	6293	1	80.59	7.3	0	18.92	0	1.024188	-27.15003	220	4	1.1	0.9;
	6294	1	199.6	24.4	0	10.6	0	1.016786	-18.076679	220	4	1.1	0.9;
	6295	1	-0.18	-1	0	0	0	1.015829	32.990745	150	8	1.1	0.9;
	6296	2	0	0	0	0	0	1.007268	1.91796	20	18	1.1	0.9;
	6297	1	165.3	7.4	0	28.32	0	0.990524	1.302692	380	5	1.1	0.9;
	6298	2	0	0	0	0	0	1.016891	10.696134	20	5	1.1	0.9;
	6299	2	0	0	0	0	0	1.016891	10.696134	20	5	1.1	0.9;
	6300	1	26.63	9.7	0	0.75	0	1.043493	-1.226779	110	19	1.1	0.9;
	6301	1	46.55	-18.9	0	14.78	0	1.018897	7.480558	220	3	1.1	0.9;
	6302	1	-0	-0	0	3.56	0	1.02662	-3.516785	150	9	1.1	0.9;
	6303	1	81.49	25.1	0	0.36	0	1.007406	14.333514	220	5	1.1	0.9;
	6304	1	2.5	0.6	0	1.06	0	1.003487	5.205063	110	18	1.1	0.9;
	6305	2	0	0	0	0	0	1.090507	11.169589	10.5	1	1.2	0.8;
	6306	1	0	-0	0	2.27	0	1.033802	9.219584	220	5	1.1	0.9;
	6307	1	30.53	9	0	1.9	0	1.032903	55.757518	110	8	1.1	0.9;
	6308	1	90.5	12.8	0	15.77	0	1.03125	3.474686	154	1	1.1	0.9;
	6309	1	2.4	0.9	0	1.18	0	1.05329	-2.80874	110	19	1.1	0.9;
	6310	1	42.45	-10.2	0	10.02	0	1.054716	-3.993237	220	23	1.1	0.9;
	6311	1	-0	-0	0	0.3	0	0.991127	-1.596972	380	5	1.1	0.9;
	6312	1	131.7	32.9	0	108.56	0	1.041387	-13.269592	380	4	1.1	0.9;
	6313	1	33.5	4.8	0	3.33	0	1.005819	17.433285	154	1	1.1	0.9;
	6314	2	0	0	0	0	0	0.984601	26.113186	6.3	1	1.1	0.9;
	6315	1	60.07	-4.7	0	0.25	0	1.060094	-2.00787	110	15	1.1	0.9;
	6316	1	0	-0	0	0	0	1.018632	-12.251956	220	11	1.1	0.9;
	6317	2	0	0	0	0	0	1.033042	-4.165701	20	7	1.1	0.9;
	6318	1	8.91	2.5	0	0.68	0	0.981871	1.723445	110	18	1.1	0.9;
	6319	1	11.41	-3.7	0	0.1	0	1.012982	-1.274503	110	18	1.1	0.9;
	6320	1	133.4	29.4	0	10.87	0	1.031351	9.130329	220	5	1.1	0.9;
	6321	1	26.13	2.7	0	1.7	0	1.040814	16.893773	220	5	1.1	0.9;
	6322	1	-113.33	45.2	0	67.52	0	1.031589	-10.695606	380	4	1.1	0.9;
	6323	1	-9.01	21.8	0	4.58	0	0.967808	-5.298475	150	9	1.1	0.9;
	6324	1	-0	0	0	3.81	0	1.022288	-13.543676	220	4	1.1	0.9;
	6325	2	0	0	0	0	0	1.043067	14.485157	20	5	1.1	0.9;
	6326	2	0	0	0	0	0	1.043067	14.485157	20	5	1.1	0.9;
	6327	2	0	0	0	0	0	1.043067	14.485157	20	5	1.1	0.9;
	6328	2	0	0	0	0	0	1.043067	14.485157	20	5	1.1	0.9;
	6329	2	0	0	0	0	0	1.043067	14.485157	20	5	1.1	0.9;
	6330	1	51.06	16.5	0	2.37	0	1.022386	38.742084	110	8	1.1	0.9;
	6331	2	0	0	0	0	0	1.035207	-6.870994	20	4	1.1	0.9;
	6332	1	-0	-0	0	16.38	0	0.993205	-6.773238	220	6	1.1	0.9;
	6333	1	50.36	24.3	0	3.64	0	1.025584	59.039755	220	8	1.1	0.9;
	6334	1	-0	-0	0	28.03	0	1.039628	-10.444432	150	9	1.1	0.9;
	6335	2	0	0	0	0	0	1.031942	-8.549229	20	4	1.1	0.9;
	6336	2	0	0	0	0	0	1.031942	-8.549229	20	4	1.1	0.9;
	6337	2	0	0	0	0	0	1.031942	-8.549229	20	4	1.1	0.9;
	6338	1	19.52	11	0	1.11	0	1.007144	-3.728197	220	6	1.1	0.9;
	6339	1	0	-0	0	18.33	0	1.081416	1.199061	380	1	1.2	0.8;
	6340	1	366.6	57.4	0	60.78	0	1.008398	-23.683296	380	4	1.1	0.9;
	6341	2	0	0	0	0	0	1.045213	9.250128	20	14	1.1	0.9;
	6342	1	10.81	3.6	0	0.26	0	0.992877	-13.88975	110	12	1.1	0.9;
	6343	1	0	0	0.054465	9.927347	0	1.023125	-2.909088	220	18	1.1	0.9;
	6344	1	0	0	0	12.64	0	1.054585	82.616556	380	3	1.1	0.9;
	6345	1	-0	0	0	1.58	0	0.967119	-9.617679	150	9	1.1	0.9;
	6346	2	0	0	0	0	0	1.024479	-0.687261	20	18	1.1	0.9;
	6347	2	0	0	0	0	0	1.02409	0.237238	11	1	1.1	0.9;
	6348	1	96.2	37.3	0	0.19	0	1.005722	-6.70794	110	23	1.1	0.9;
	6349	1	269.9	12.4	0	4.09	0	0.973116	-28.905172	150	10	1.1	0.9;
	6350	1	-0	0	0.040436	-0.040436	0	0.992738	17.76388	220	5	1.1	0.9;
	6351	2	0	0	0	0	0	1.055778	36.671382	20	8	1.1	0.9;
	6352	2	0	0	0	0	0	1.055778	36.671382	20	8	1.1	0.9;
	6353	2	0	0	0	0	0	1.070329	49.112584	20	8	1.1	0.9;
	6354	2	0	0	0	0	0	1.070329	49.112584	20	8	1.1	0.9;
	6355	1	0	-0	0	7.13	0	1.0331	60.584093	380	3	1.1	0.9;
	6356	1	-0	0	0.120861	4.779355	0	0.992712	-2.759259	220	10	1.1	0.9;
	6357	2	0	0	0	0	0	1.048705	42.664255	20	8	1.1	0.9;
	6358	2	0	0	0	0	0	1.048705	42.664255	20	8	1.1	0.9;
	6359	2	0	0	0	0	0	0.995077	34.268512	11	1	1.1	0.9;
	6360	2	0	0	0	0	0	1.043227	47.072253	20	8	1.1	0.9;
	6361	2	0	0	0	0	0	1.070329	49.112584	20	8	1.1	0.9;
	6362	1	0	-0	0.569185	79.269885	0	1.052114	81.244934	380	3	1.1	0.9;
	6363	2	0	0	0	0	0	1.024375	-1.976838	20	10	1.1	0.9;
	6364	1	-0	0	0.135375	8.900675	0	1.070028	2.099566	380	18	1.2	0.8;
	6365	1	193.1	69.9	0	28.91	0	1.017617	-17.187976	380	4	1.1	0.9;
	6366	1	0	-0	0.699934	14.325886	0	1.026296	33.499485	380	8	1.1	0.9;
	6367	1	92.4	22	0.019713	6.930287	0	1.029058	-20.283666	220	4	1.1	0.9;
	6368	2	0	0	0	0	0	1.032321	27.942453	20	5	1.1	0.9;
	6369	2	0	0	0	0	0	1.014698	-14.460554	20	4	1.1	0.9;
	6370	2	0	0	0	0	0	1.014698	-14.460554	20	4	1.1	0.9;
	6371	2	0	0	0	0	0	1.014698	-14.460554	20	4	1.1	0.9;
	6372	2	0	0	0	0	0	1.014698	-14.460554	20	4	1.1	0.9;
	6373	1	-0	0	0	4.03	0	1.072201	58.84567	220	8	1.1	0.9;
	6374	1	18.62	6.8	0	0.92	0	0.955487	-1.753626	110	19	1.1	0.9;
	6375	1	14.32	7.2	0	3.62	0	1.022492	7.188929	220	7	1.1	0.9;
	6376	1	39.94	13.3	0	0.85	0	0.943205	-12.984503	110	21	1.1	0.9;
	6377	1	43.55	14.3	0	1.63	0	0.979898	-5.266097	110	21	1.1	0.9;
	6378	1	0	-0	0	3.4	0	0.977474	-26.890479	150	10	1.1	0.9;
	6379	1	68.48	-10.6	0	2.65	0	0.983454	-24.020982	150	10	1.1	0.9;
	6380	1	-0	0	0	18.8	0	1.026259	45.39552	380	3	1.1	0.9;
	6381	1	57	39.9	0	153.693557	0	1.127748	0.999187	380	9	1.2	0.8;
	6382	1	31.2	4.4	0	1.61	0	1.000814	14.198741	154	1	1.1	0.9;
	6383	2	0	0	0	0	0	1.003111	19.775229	11	1	1.1	0.9;
	6384	1	-0	-0	0	31.35	0	1.013502	-11.0947	220	6	1.1	0.9;
	6385	1	24.13	29.4	0	146.51	0	1.077575	72.579655	380	3	1.2	0.8;
	6386	1	-0	-0	0.116759	0.732701	0	1.037068	80.040751	220	3	1.1	0.9;
	6387	1	0	-0	0	0.57	0	0.959541	-5.714015	110	18	1.1	0.9;
	6388	1	163.63	23	0	5.11	0	1.012813	-1.040485	220	6	1.1	0.9;
	6389	1	165.8	44	0	6.15	0	1.062234	2.581972	220	5	1.1	0.9;
	6390	1	-0.73	-1.93	0	0.36	0	0.991721	-5.970433	220	12	1.1	0.9;
	6391	2	0	0	0	0	0	1.02421	5.097202	20	6	1.1	0.9;
	6392	1	65.07	15	0	1.37	0	1.022789	3.322967	220	5	1.1	0.9;
	6393	1	22.12	6.5	0	0.94	0	1.000284	-9.091433	110	12	1.1	0.9;
	6394	1	235	58.9	0	14.47	0	1.013352	12.865452	220	5	1.1	0.9;
	6395	1	-0	0	0	104.15	0	0.974247	28.9452	400	1	1.1	0.9;
	6396	1	3.7	0.6	0	5.81	0	1.017635	26.849439	154	1	1.1	0.9;
	6397	1	5.11	2.3	0	0.94	0	0.980798	11.596392	110	18	1.1	0.9;
	6398	2	0	0	0	0	0	1.044939	29.387061	20	5	1.1	0.9;
	6399	1	20.82	8.2	0	2.96	0	1.001688	-25.282589	220	4	1.1	0.9;
	6400	1	220.79	13.21	0.039607	8.600393	0	1.024694	18.017866	220	5	1.1	0.9;
	6401	2	0	0	0	0	0	1.052205	80.628473	20	3	1.1	0.9;
	6402	2	0	0	0	0	0	1.052231	80.621613	20	3	1.1	0.9;
	6403	1	0	-0	0	18.66	0	1.008601	-11.667879	380	1	1.1	0.9;
	6404	2	0	0	0	0	0	1.052205	80.628473	20	3	1.1	0.9;
	6405	2	0	0	0	0	0	1.047514	68.881226	20	3	1.1	0.9;
	6406	2	0	0	0	0	0	1.040945	-2.019998	20	2	1.1	0.9;
	6407	2	0	0	0	0	0	1.052231	80.621613	20	3	1.1	0.9;
	6408	2	0	0	0	0	0	1.047514	68.881226	20	3	1.1	0.9;
	6409	2	0	0	0	0	0	1.054213	39.514325	20	14	1.1	0.9;
	6410	1	1.3	-0	0	1.48	0	0.991877	-1.750776	110	18	1.1	0.9;
	6411	1	-13.71	4.5	0	0.08	0	0.976407	-5.464167	150	9	1.1	0.9;
	6412	1	8.3	1.2	0	1.6	0	0.996489	34.446399	154	1	1.1	0.9;
	6413	1	23.23	7.6	0	0.88	0	0.997389	-8.959946	110	12	1.1	0.9;
	6414	1	-1235.46	-43	0	90.6	0	1.064947	81.922707	380	3	1.1	0.9;
	6415	1	0	-0	0.255719	9.892154	0	1.061435	-0.283673	220	18	1.2	0.8;
	6416	1	306.5	57.7	0.411758	4.816868	0	1.051836	-4.896212	220	5	1.1	0.9;
	6417	2	0	0	0	0	0	1.054686	33.79222	20	3	1.1	0.9;
	6418	2	0	0	0	0	0	1.054686	33.79222	20	3	1.1	0.9;
	6419	1	72.28	6.6	0	1.42	0	1.02386	-27.409382	220	4	1.1	0.9;
	6420	2	0	0	0	0	0	0.994392	93.663435	20	3	1.1	0.9;
	6421	2	0	0	0	0	0	0.994839	90.156106	20	3	1.1	0.9;
	6422	2	0	0	0	0	0	1.054686	33.79222	20	3	1.1	0.9;
	6423	1	0	0	0	0	0	0.986477	-6.386478	150	9	1.1	0.9;
	6424	1	-0	0	0.038061	1.491939	0	1.045691	-0.691332	220	18	1.1	0.9;
	6425	1	-0	0	0.174934	1.86394	0	0.999966	90.949741	150	1	1.1	0.9;
	6426	1	37.6	5.4	0	5.37	0	0.976525	41.441914	154	1	1.1	0.9;
	6427	1	-0	-0	0	0	0	1.035283	-13.067891	20	4	1.1	0.9;
	6428	1	-0	-0	0	0	0	1.035283	-13.067891	20	4	1.1	0.9;
	6429	2	0	0	0	0	0	1.021473	-8.614663	20	4	1.1	0.9;
	6430	1	-0	-0	0	0	0	1.018329	-16.147917	20	4	1.1	0.9;
	6431	2	0	0	0	0	0	1.021729	-11.056037	20	4	1.1	0.9;
	6432	1	-0	0	0	0	0	1.013977	-15.655534	20	4	1.1	0.9;
	6433	1	-0	0	0	0	0	1.013977	-15.655534	20	4	1.1	0.9;
	6434	1	0	-0	0	0.51	0	0.972952	7.217167	110	18	1.1	0.9;
	6435	1	-0	-0	0	9.57	0	1.056869	30.897354	380	1	1.1	0.9;
	6436	2	0	0	0	0	0	1.063714	6.510197	20	9	1.1	0.9;
	6437	1	-67.31	-44.04	0	0	0	0.99382	25.664776	150	8	1.1	0.9;
	6438	1	0	0	0	3.16	0	1.014147	-16.760348	380	6	1.1	0.9;
	6439	1	35.24	12.7	0	3.4	0	0.996102	-10.239186	110	22	1.1	0.9;
	6440	2	0	0	0	0	0	1.000784	-5.3448	20	18	1.1	0.9;
	6441	1	17.82	5.9	0	0.64	0	0.985655	-5.031986	110	21	1.1	0.9;
	6442	1	33.74	11.8	0	1.66	0	1.033882	-5.074867	110	3	1.1	0.9;
	6443	1	60.97	19.7	0	2.22	0	0.969413	-12.083758	110	12	1.1	0.9;
	6444	1	16.7	2.4	0	3.18	0	0.981496	25.250342	154	1	1.1	0.9;
	6445	1	10.91	3.6	0	0.26	0	0.977303	-11.041494	110	12	1.1	0.9;
	6446	1	-0	0	0.26	6.56	0	1.094333	4.418276	220	19	1.2	0.8;
	6447	1	-0	0	0	20.48	0	1.034404	39.871817	380	3	1.1	0.9;
	6448	1	-0	-0	0	72.64	0	1.033845	-11.409379	380	4	1.1	0.9;
	6449	1	0	0	0	1.23	0	0.999565	4.245344	110	18	1.1	0.9;
	6450	1	62.07	20.7	0	6.92	0	1.012671	-4.325832	220	21	1.1	0.9;
	6451	2	0	0	0	0	0	1.004119	-8.128564	20	2	1.1	0.9;
	6452	2	0	0	0	0	0	1.004063	-5.264305	20	2	1.1	0.9;
	6453	1	0	0	0.135375	53.752175	0	1.079919	1.350644	380	18	1.2	0.8;
	6454	2	0	0	0	0	0	1.027739	-17.648712	20	6	1.1	0.9;
	6455	2	0	0	0	0	0	1.012552	-3.840911	20	2	1.1	0.9;
	6456	2	0	0	0	0	0	1.012552	-3.840911	20	2	1.1	0.9;
	6457	1	0	0	0.27735	33.218875	0	1.091484	80.814919	380	3	1.2	0.8;
	6458	1	0	0	0.278651	111.707791	0	1.088116	75.264549	380	3	1.2	0.8;
	6459	1	0	0	0	91.64	0	1.105642	4.175279	380	20	1.2	0.8;
	6460	1	193	17.6	0	4.74	0	1.011873	-19.783189	220	4	1.1	0.9;
	6461	1	0	0	0.194155	-0.145616	0	1.031847	4.740375	380	6	1.1	0.9;
	6462	1	266	62.8	0	1.95	0	1.011513	-20.230893	220	4	1.1	0.9;
	6463	1	-0	0	0.224562	1.022222	0	1.074544	35.547599	380	14	1.1	0.9;
	6464	1	-0	-0	0	8.33	0	1.053801	7.308064	220	3	1.1	0.9;
	6465	2	0	0	0	0	0	1.050901	18.201002	20	7	1.1	0.9;
	6466	1	0	0	0	1.02	0	1.081967	13.950182	220	5	1.1	0.9;
	6467	1	35.24	12.3	0	0.3	0	0.995728	-5.989622	110	3	1.1	0.9;
	6468	1	0	0	0	0	0	0.961566	-8.310096	150	9	1.1	0.9;
	6469	1	-15.02	4.9	0	0.26	0	0.997804	1.009459	150	9	1.1	0.9;
	6470	1	99.1	-98.6	0	6.83	0	1.02067	-32.687496	220	4	1.1	0.9;
	6471	2	0	0	0	0	0	1.028916	-13.292219	20	6	1.1	0.9;
	6472	1	-0	0	0	51.43	0	1.041317	-13.35377	380	4	1.1	0.9;
	6473	1	0	0	0	0	0	1.050245	29.580149	150	2	1.1	0.9;
	6474	1	25.33	12.7	0.181549	6.830889	0	1.039387	-2.383874	220	18	1.1	0.9;
	6475	1	45.05	13.2	0	44.93	0	1.021028	-7.741887	220	5	1.1	0.9;
	6476	2	0	0	0	0	0	1.031498	24.749019	20	10	1.1	0.9;
	6477	2	0	0	0	0	0	1.022447	17.861755	20	10	1.1	0.9;
	6478	1	-0	0	0	6.72	0	1.057304	18.734409	220	3	1.1	0.9;
	6479	1	19.4	2.8	0	2.53	0	0.976882	21.673666	154	1	1.1	0.9;
	6480	1	-0	-0	0	13.44	0	1.018888	36.987079	150	8	1.1	0.9;
	6481	1	0	0	0	0	0	1.04035	8.357264	20	5	1.1	0.9;
	6482	1	303.6	47.5	0	14.96	0	1.013426	-0.97101	380	6	1.1	0.9;
	6483	1	30.03	9.7	0	4.1	0	1.009941	36.737144	110	8	1.1	0.9;
	6484	1	117.4	12.5	0	5.16	0	1.010329	15.30931	220	7	1.1	0.9;
	6485	1	-0	-0	0	5.12	0	1.021989	-10.739532	220	22	1.1	0.9;
	6486	1	-0	-0	0	0.05	0	1.055149	5.178292	220	5	1.1	0.9;
	6487	2	0	0	0	0	0	1.028283	16.586422	20	14	1.1	0.9;
	6488	1	-0	-0	0	28.168167	0	0.921315	3.650781	380	5	1.1	0.9;
	6489	1	0	0	0	0	0	1.017525	-4.63816	150	9	1.1	0.9;
	6490	1	0	0	0	3.01	0	1.012571	-11.292378	220	5	1.1	0.9;
	6491	2	0	0	0	0	0	1.04554	8.581299	10.5	1	1.1	0.9;
	6492	1	32.84	9.3	0	0.76	0	0.976911	-5.807222	110	18	1.1	0.9;
	6493	1	77.19	47.8	0	2.08	0	0.985548	-7.292905	150	9	1.1	0.9;
	6494	1	0	0	0	0.33	0	1.035156	25.307552	150	2	1.1	0.9;
	6495	1	0	0	0	0.51	0	1.067715	11.555887	220	3	1.2	0.8;
	6496	1	-0	0	0	3.22	0	0.989037	34.007972	220	3	1.1	0.9;
	6497	1	485.5	-63.38	0	34.92	0	1.0248	-18.844784	220	5	1.1	0.9;
	6498	2	0	0	0	0	0	1.016868	22.255371	20	6	1.1	0.9;
	6499	1	296.8	77.7	0	0	0	1.037008	26.476248	150	2	1.1	0.9;
	6500	2	0	0	0	0	0	1.034935	24.924619	20	10	1.1	0.9;
	6501	1	-0	-0	0.03995	11.24005	0	1.016451	-16.531897	220	4	1.1	0.9;
	6502	1	11.21	4.4	0	0.25	0	1.038846	-2.91737	110	15	1.1	0.9;
	6503	1	50.06	0	0	0	0	0.971031	-11.133184	150	7	1.1	0.9;
	6504	1	39	5.5	0	1.3	0	1.014673	20.715481	154	1	1.1	0.9;
	6505	1	-0	-0	0	0.68	0	1.049745	7.525677	220	1	1.1	0.9;
	6506	1	19.52	5.6	0	1.46	0	1.012636	30.857104	150	8	1.1	0.9;
	6507	2	0	0	0	0	0	1.009892	4.868208	20	9	1.1	0.9;
	6508	1	-0	0	0	0.8	0	1.023937	3.43247	220	5	1.1	0.9;
	6509	1	142.5	523.5	0	34.96	0	1.0402	18.602574	380	6	1.1	0.9;
	6510	1	97	-539.3	0	50.89	0	1.041758	42.616765	380	3	1.1	0.9;
	6511	1	60.07	45	0	0.12	0	1.074626	46.062649	220	3	1.1	0.9;
	6512	1	0	0	0	102.31354	0	1.011748	20.601504	380	5	1.1	0.9;
	6513	1	-0	-0	0	420.1	0	1.040904	18.608226	380	6	1.1	0.9;
	6514	1	0	-0	0.17	11.68	0	1.002206	-2.183095	220	21	1.1	0.9;
	6515	1	5.51	2	0	0.69	0	1.057132	-3.660314	110	19	1.1	0.9;
	6516	1	13.82	4.5	0	0.25	0	0.995974	-9.032462	110	12	1.1	0.9;
	6517	1	141.2	33.5	0.118053	1.231623	0	0.988296	89.600207	150	1	1.1	0.9;
	6518	1	1.3	0.4	0	1.43	0	0.988704	12.931136	110	18	1.1	0.9;
	6519	1	31.54	12.5	0.050337	32.519663	0	1.016781	-22.881448	380	4	1.1	0.9;
	6520	1	5.21	1	0	0.94	0	1.002202	-5.307223	110	18	1.1	0.9;
	6521	1	16.52	4.8	0	0.33	0	0.992433	-2.243554	110	18	1.1	0.9;
	6522	1	0	-0	0.039415	12.270585	0	1.020074	42.138042	380	3	1.1	0.9;
	6523	1	0	-0	0.452023	8.088382	0	1.008427	-8.814025	380	6	1.1	0.9;
	6524	1	-0	0	0	9.25	0	1.022036	-11.413747	380	4	1.1	0.9;
	6525	1	27.43	13.7	0.21794	-0.012536	0	1.087524	1.371618	380	21	1.2	0.8;
	6526	1	28.13	10.2	0	1.73	0	0.961031	-1.92733	110	19	1.1	0.9;
	6527	2	0	0	0	0	0	1.035162	21.902936	13.8	1	1.1	0.9;
	6528	1	32.8	4.7	0	8.02	0	1.008983	30.24237	154	1	1.1	0.9;
	6529	1	140.2	11.2	0	9.1	0	1.014558	-27.115835	220	4	1.1	0.9;
	6530	1	98.2	19.1	0	10.92	0	1.026433	18.625558	220	5	1.1	0.9;
	6531	1	1	1	0	0.27	0	1.04041	-1.359387	110	23	1.1	0.9;
	6532	1	-0	-0	0	0.03	0	1.021486	33.002338	154	1	1.1	0.9;
	6533	1	52.6	110.6	0	2.72	0	1.020436	16.327592	220	7	1.1	0.9;
	6534	1	0	0	0	1	0	0.966258	9.890834	220	5	1.1	0.9;
	6535	1	-0	-0	0	30.98	0	0.985471	14.787943	400	1	1.1	0.9;
	6536	1	0	0	0	2.06	0	1.010153	-8.125449	150	9	1.1	0.9;
	6537	1	55.7	7.9	0	2.29	0	0.982509	25.800974	154	1	1.1	0.9;
	6538	1	413	-353	0	8.88	0	1.067699	3.591344	400	1	1.1	0.9;
	6539	2	0	0	0	0	0	1.042166	-4.652525	20	9	1.1	0.9;
	6540	1	-14.83	-3.56	0	0.06	0	1.021537	9.115608	110	18	1.1	0.9;
	6541	2	0	0	0	0	0	1.024267	0.497112	20	18	1.1	0.9;
	6542	2	0	0	0	0	0	1.024267	0.497112	20	18	1.1	0.9;
	6543	2	0	0	0	0	0	1.024267	0.497112	20	18	1.1	0.9;
	6544	2	0	0	0	0	0	1.023818	-10.471889	20	4	1.1	0.9;
	6545	2	0	0	0	0	0	1.019094	-5.889827	20	18	1.1	0.9;
	6546	1	-0	0	0	39.29	0	1.106402	-3.071016	380	9	1.2	0.8;
	6547	2	0	0	0	0	0	1.023818	-10.471889	20	4	1.1	0.9;
	6548	2	0	0	0	0	0	1.023818	-10.471889	20	4	1.1	0.9;
	6549	1	22.42	4.8	0	0.53	0	0.998877	-5.06743	110	3	1.1	0.9;
	6550	1	0	-0	0	14.69	0	1.028709	44.833215	380	3	1.1	0.9;
	6551	1	0	-0	0.057073	0.012927	0	1.053314	-6.50391	150	9	1.1	0.9;
	6552	1	88.5	27.4	0	0.04	0	0.974467	-6.132079	110	21	1.1	0.9;
	6553	1	5.71	1.6	0	0.48	0	0.989926	-3.94964	110	18	1.1	0.9;
	6554	1	-0	0	0	2.06	0	0.987546	-24.919637	220	6	1.1	0.9;
	6555	1	-0	0	0	7.22	0	0.99451	-18.776597	220	4	1.1	0.9;
	6556	1	0	0	0	0	0	1.018762	-7.534283	150	9	1.1	0.9;
	6557	2	0	0	0	0	0	1.023603	23.127902	20	6	1.1	0.9;
	6558	1	-0	0	0	0.54	0	1.008549	32.542056	154	1	1.1	0.9;
	6559	1	227.6	37.8	0	39.33	0	1.018674	-17.209554	380	6	1.1	0.9;
	6560	1	0	-0	0	6.45	0	1.016289	-11.291227	220	1	1.1	0.9;
	6561	1	37.44	-28.3	0	9.28	0	1.02297	32.327798	220	3	1.1	0.9;
	6562	1	0	-0	0	31.6	0	1.01419	29.738593	380	3	1.1	0.9;
	6563	2	0	0	0	0	0	0.998951	38.095582	9	1	1.1	0.9;
	6564	1	0	0	0	9.72	0	1.018282	43.66889	380	3	1.1	0.9;
	6565	1	49.45	30.6	0	1.57	0	0.936045	-11.578054	150	9	1.1	0.9;
	6566	2	0	0	0	0	0	1.014865	17.912579	20	5	1.1	0.9;
	6567	2	0	0	0	0	0	1.014865	17.912579	20	5	1.1	0.9;
	6568	1	24.93	15.4	0	5.48	0	0.969729	-14.732562	150	9	1.1	0.9;
	6569	1	-0.2	-1.64	0	0	0	1.01757	30.154601	150	8	1.1	0.9;
	6570	1	-0	-0	0.287402	10.710498	0	1.057905	79.411231	380	3	1.1	0.9;
	6571	1	0	-0	0	2.13	0	0.963608	-9.787433	150	9	1.1	0.9;
	6572	1	0	0	0.5735	-0.316856	0	1.057634	37.991837	220	5	1.1	0.9;
	6573	1	0	0	0	2.7	0	1.017763	-5.85612	220	1	1.1	0.9;
	6574	1	-0	-0	0	53.24	0	1.0129	-10.929789	220	6	1.1	0.9;
	6575	1	-0	0	0.039781	0.460219	0	1.038086	23.154703	220	5	1.1	0.9;
	6576	1	0	0	0	0	0	1.051779	-6.398929	150	9	1.1	0.9;
	6577	1	69.78	25	0	14.24	0	1.022528	15.642489	220	5	1.1	0.9;
	6578	1	-0	0	0	18.32	0	1.041551	41.353637	380	8	1.1	0.9;
	6579	1	1.1	0.4	0.019452	1.410548	0	1.026102	-1.724689	110	18	1.1	0.9;
	6580	1	-0	0	0	0.24	0	1.013214	-9.554082	380	6	1.1	0.9;
	6581	1	71.88	-2.2	0	0.07	0	1.053766	7.292607	220	3	1.1	0.9;
	6582	1	-0	0	0	5.9	0	1.031077	-19.605096	220	4	1.1	0.9;
	6583	1	0	0	0	12.29	0	0.968466	-12.007066	220	22	1.1	0.9;
	6584	2	0	0	0	0	0	0.956299	35.305913	14	1	1.1	0.9;
	6585	2	0	0	0	0	0	0.989649	6.296698	13.8	1	1.1	0.9;
	6586	1	-0	0	0	1.75	0	1.018409	-3.066024	220	5	1.1	0.9;
	6587	1	-0	0	0	18.07	0	1.042629	65.129501	380	3	1.1	0.9;
	6588	1	15.32	18.2	0	27.62	0	1.023758	-12.955313	220	4	1.1	0.9;
	6589	1	-0	-0	0	55.517991	0	1.023448	30.391991	380	5	1.1	0.9;
	6590	1	336.3	104.4	0	0	0	1.039999	25.6544	150	2	1.1	0.9;
	6591	1	0	0	0	2.91	0	1.009992	35.445834	220	3	1.1	0.9;
	6592	1	0	-0	0.304554	30.256205	0	1.0651	-7.973645	380	10	1.1	0.9;
	6593	2	0	0	0	0	0	1.047126	22.873876	20	5	1.1	0.9;
	6594	1	-0	-0	0	0.7	0	0.943618	7.025726	110	18	1.1	0.9;
	6595	1	-0	-0	0.587783	334.58331	0	1.088261	72.590393	380	3	1.2	0.8;
	6596	1	80.09	20	0	6.98	0	1.023753	69.81304	220	3	1.1	0.9;
	6597	1	199.3	53.7	0	12.54	0	1.005645	39.813001	380	3	1.1	0.9;
	6598	1	65.6	-382	0.202178	57.724057	0	1.022364	58.584365	380	3	1.1	0.9;
	6599	1	-37.34	-13.33	0	251.33835	0	1.005609	28.049125	380	5	1.1	0.9;
	6600	1	-0	0	0.078724	6.200957	0	1.018103	-17.745392	220	6	1.1	0.9;
	6601	2	0	0	0	0	0	1.012099	-18.860818	20	4	1.1	0.9;
	6602	2	0	0	0	0	0	1.021005	21.987746	20	14	1.1	0.9;
	6603	1	20.72	6.8	0	0.69	0	0.979933	-12.309951	110	12	1.1	0.9;
	6604	1	170.1	-9.3	0	5.97	0	1.030791	-15.072577	220	5	1.1	0.9;
	6605	1	95.51	17.1	0	8.5	0	1.013631	12.900828	220	5	1.1	0.9;
	6606	2	0	0	0	0	0	1.024151	24.403529	20	5	1.1	0.9;
	6607	2	0	0	0	0	0	1.024151	24.403529	20	5	1.1	0.9;
	6608	2	0	0	0	0	0	1.024151	24.403529	20	5	1.1	0.9;
	6609	2	0	0	0	0	0	1.024151	24.403529	20	5	1.1	0.9;
	6610	2	0	0	0	0	0	1.024151	24.403529	20	5	1.1	0.9;
	6611	2	0	0	0	0	0	1.024151	24.403529	20	5	1.1	0.9;
	6612	2	0	0	0	0	0	1.024151	24.403529	20	5	1.1	0.9;
	6613	1	219.6	62.5	0.039623	5.500377	0	1.030105	26.548864	220	5	1.1	0.9;
	6614	1	-0	-0	0	43.44	0	1.019216	-14.137173	380	4	1.1	0.9;
	6615	1	55.98	18.47	0	9.45	0	0.991156	88.859311	150	1	1.1	0.9;
	6616	2	0	0	0	0	0	1.029779	8.046458	10.5	1	1.1	0.9;
	6617	1	75.38	21.8	0	0.63	0	1.02896	-11.606592	220	4	1.1	0.9;
	6618	1	74.98	46.4	0	2.74	0	1.03097	-5.06415	150	9	1.1	0.9;
	6619	1	0	-0	0	12.9	0	1.006922	-16.346989	380	6	1.1	0.9;
	6620	2	0	0	0	0	0	1.058114	43.382414	20	10	1.1	0.9;
	6621	2	0	0	0	0	0	1.059543	4.22325	20	5	1.1	0.9;
	6622	2	0	0	0	0	0	1.030338	22.622295	20	5	1.1	0.9;
	6623	2	0	0	0	0	0	1.040509	24.175107	20	5	1.1	0.9;
	6624	2	0	0	0	0	0	1.040509	24.175107	20	5	1.1	0.9;
	6625	1	0	-0	0	0	0	1.023453	0.234493	110	3	1.1	0.9;
	6626	1	-0	0	0.236449	45.072959	0	1.013779	-20.268901	220	6	1.1	0.9;
	6627	1	7	1	0	3.58	0	0.982047	26.53851	154	1	1.1	0.9;
	6628	2	0	0	0	0	0	1.041212	-15.377355	20	4	1.1	0.9;
	6629	1	12.91	4	0	1.66	0	0.978503	-1.75321	110	19	1.1	0.9;
	6630	2	0	0	0	0	0	1.040509	24.175107	20	5	1.1	0.9;
	6631	2	0	0	0	0	0	1.04659	30.049071	20	5	1.1	0.9;
	6632	1	160	69.2	0	0.19	0	1.029284	8.268155	220	5	1.1	0.9;
	6633	1	-0	0	0	1.1	0	0.976221	-5.459556	150	9	1.1	0.9;
	6634	1	97.5	13.9	0	29.65	0	0.975002	33.150591	400	1	1.1	0.9;
	6635	1	-0	-0	0	0.94	0	1.016554	19.011605	220	5	1.1	0.9;
	6636	1	349.6	22	0	10.29	0	1.013938	-10.330003	380	6	1.1	0.9;
	6637	1	91.1	36.1	0	1.01	0	0.979234	35.087735	220	3	1.1	0.9;
	6638	1	114.2	16.3	0	2.85	0	1.006554	28.610342	154	1	1.1	0.9;
	6639	2	0	0	0	0	0	1.031885	-12.651409	20	4	1.1	0.9;
	6640	2	0	0	0	0	0	1.031885	-12.651409	20	4	1.1	0.9;
	6641	2	0	0	0	0	0	1.031885	-12.651409	20	4	1.1	0.9;
	6642	2	0	0	0	0	0	1.031885	-12.651409	20	4	1.1	0.9;
	6643	1	41.05	31.8	0.039111	1.370667	0	1.023367	-19.593969	220	6	1.1	0.9;
	6644	1	268.56	29.72	0	6.5	0	0.979542	15.358319	220	13	1.1	0.9;
	6645	1	3.2	8.3	0	0.38	0	0.964386	-8.067074	150	9	1.1	0.9;
	6646	1	48.48	8.7	0	10.14	0	1.05203	14.372042	220	5	1.1	0.9;
	6647	1	113.6	16.2	0	1.83	0	1.009227	21.977145	154	1	1.1	0.9;
	6648	1	15.32	9.5	0	5.33	0	0.986477	-6.386478	150	9	1.1	0.9;
	6649	2	0	0	0	0	0	0.955439	48.700819	10.6	1	1.1	0.9;
	6650	2	0	0	0	0	0	1.036065	-16.582357	20	4	1.1	0.9;
	6651	2	0	0	0	0	0	1.04914	42.65185	20	2	1.1	0.9;
	6652	1	5.11	1.9	0	2.31	0	0.969632	6.466075	110	18	1.1	0.9;
	6653	1	405.5	247.2	1.369347	107.995934	0	1.069454	39.838153	380	13	1.1	0.9;
	6654	2	0	0	0	0	0	1.043738	-8.975119	20	4	1.1	0.9;
	6655	1	3.2	0.4	0	0.65	0	1.009163	2.151893	110	18	1.1	0.9;
	6656	1	-0	-0	0.199592	35.580613	0	1.019587	36.854451	380	3	1.1	0.9;
	6657	1	7.71	1.3	0	2.44	0	0.987522	-8.995861	110	18	1.1	0.9;
	6658	1	-14.82	-5.03	0	0.15	0	1.012203	-2.922462	110	18	1.1	0.9;
	6659	1	-15.88	-4.9	0	0.29	0	1.005863	-11.937443	110	22	1.1	0.9;
	6660	1	0	0	0	14.72	0	1.021265	4.517407	220	7	1.1	0.9;
	6661	1	0	-0	0	150.649311	0	1.002222	19.480877	380	5	1.1	0.9;
	6662	1	0	0	0	38.49	0	1.039618	-15.811939	380	4	1.1	0.9;
	6663	2	0	0	0	0	0	1.045666	52.068402	20	3	1.1	0.9;
	6664	1	-0	-0	0	2.32	0	1.043316	-18.000334	380	10	1.1	0.9;
	6665	1	0	-0	0	0	0	1.005017	-18.861992	20	4	1.1	0.9;
	6666	1	0	-0	0	14.64	0	1.02671	14.58169	220	5	1.1	0.9;
	6667	2	0	0	0	0	0	1.034595	28.964131	20	10	1.1	0.9;
	6668	1	21.12	13.1	0	0.18	0	0.966548	-5.661815	150	9	1.1	0.9;
	6669	1	2.8	0.5	0	2.37	0	0.996844	-4.724158	110	18	1.1	0.9;
	6670	2	0	0	0	0	0	1.042317	14.274354	20	14	1.1	0.9;
	6671	1	0	0	0	0	0	1.009689	30.943205	150	8	1.1	0.9;
	6672	1	20.02	9	0	0.67	0	0.999657	-7.602703	110	23	1.1	0.9;
	6673	1	0	-0	0	0	0	1.005017	-18.861992	20	4	1.1	0.9;
	6674	1	-0	0	0	0.3	0	1.032007	18.368937	220	5	1.1	0.9;
	6675	1	-0	0	0	124.2	0	1.021005	19.411039	380	5	1.1	0.9;
	6676	1	0	-0	0	1.33	0	1.029293	-6.917664	150	9	1.1	0.9;
	6677	1	-8.71	2.9	0	0.42	0	0.959452	-9.679922	150	9	1.1	0.9;
	6678	1	179.8	76.5	0	8.05	0	1.018742	15.409657	220	7	1.1	0.9;
	6679	1	119	21.9	0	0.95	0	1.061515	21.336436	220	5	1.1	0.9;
	6680	2	0	0	0	0	0	0.994234	37.670381	13.8	1	1.1	0.9;
	6681	1	48.45	16	0	0.47	0	0.955296	-7.17774	110	21	1.1	0.9;
	6682	1	138.2	21.3	0	2.33	0	1.029181	-21.531084	220	4	1.1	0.9;
	6683	1	93	66	0	2.25	0	0.97968	37.704964	220	3	1.1	0.9;
	6684	1	6.41	2.3	0	0.78	0	1.041238	-3.488157	110	19	1.1	0.9;
	6685	1	17.02	5.6	0	0.19	0	1.016267	-9.006082	110	12	1.1	0.9;
	6686	2	0	0	0	0	0	0.995022	23.934845	11	1	1.1	0.9;
	6687	1	13.92	6.3	0	0.37	0	0.92232	-8.785007	110	18	1.1	0.9;
	6688	1	101	20.6	0	54.5	0	1.012285	-13.758635	380	6	1.1	0.9;
	6689	2	0	0	0	0	0	1.034597	19.022639	20	6	1.1	0.9;
	6690	2	0	0	0	0	0	1.034597	19.022639	20	6	1.1	0.9;
	6691	2	0	0	0	0	0	1.034597	19.022639	20	6	1.1	0.9;
	6692	1	130.5	25.7	0	2.27	0	1.038983	12.821442	220	5	1.1	0.9;
	6693	2	0	0	0	0	0	1.034532	19.101591	20	6	1.1	0.9;
	6694	1	-9.87	-3.77	0	0.42	0	0.995478	-2.011399	110	18	1.1	0.9;
	6695	1	-0	-0	0.20225	3.148312	0	1.011145	22.97461	220	6	1.1	0.9;
	6696	1	0	0	0	0	0	1.023631	36.965167	150	8	1.1	0.9;
	6697	2	0	0	0	0	0	1.034532	19.101591	20	6	1.1	0.9;
	6698	1	0	0	0	0	0	1.026038	-10.378066	20	4	1.1	0.9;
	6699	1	77.59	-6.1	0.039393	2.780607	0	1.047642	1.687712	220	5	1.1	0.9;
	6700	1	92.4	13.1	0	98.17	0	0.970463	21.317703	400	1	1.1	0.9;
	6701	1	-0	-0	0	0.31	0	1.012648	-12.139976	150	9	1.1	0.9;
	6702	1	0	0	0.165633	-1.064787	0	1.025467	4.507504	220	3	1.1	0.9;
	6703	1	11.91	4.3	0	1.68	0	0.9897	-3.56235	110	19	1.1	0.9;
	6704	1	73.28	9.4	0	24.65	0	1.056943	60.232085	380	3	1.1	0.9;
	6705	1	47.95	16.3	0	3.46	0	0.994609	-3.609617	220	6	1.1	0.9;
	6706	1	27.9	4	0	2.96	0	1.021366	16.94013	154	1	1.1	0.9;
	6707	1	-31.26	-13.2	0	2.81	0	1.00577	-22.314337	220	6	1.1	0.9;
	6708	1	8.51	2.8	0	1.16	0	0.979823	-11.301293	110	12	1.1	0.9;
	6709	2	0	0	0	0	0	1.021634	-5.880743	20	4	1.1	0.9;
	6710	1	311.4	43	0	37.55	0	1.009255	-15.636886	380	6	1.1	0.9;
	6711	1	2.2	0.6	0	0.04	0	0.972797	-28.709094	150	10	1.1	0.9;
	6712	1	100.7	12.8	0.038981	0.661019	0	1.021983	15.493568	220	13	1.1	0.9;
	6713	2	0	0	0	0	0	1.022626	30.799151	20	10	1.1	0.9;
	6714	1	19.52	6.4	0	0.03	0	1.05121	-6.447524	150	9	1.1	0.9;
	6715	1	11.01	3.9	0	7.98	0	1.036613	19.194472	220	5	1.1	0.9;
	6716	1	36.28	-45.18	0	0.84	0	1.053996	37.457424	220	13	1.1	0.9;
	6717	2	0	0	0	0	0	0.996757	22.457888	6.3	1	1.1	0.9;
	6718	1	13.9	1.9	0	4.25	0	1.004956	28.372364	154	1	1.1	0.9;
	6719	2	0	0	0	0	0	1.028667	-4.741858	20	6	1.1	0.9;
	6720	1	-0	-0	0	2.07	0	1.025277	-5.231393	110	22	1.1	0.9;
	6721	2	0	0	0	0	0	1.028667	-4.741858	20	6	1.1	0.9;
	6722	1	-0	0	0	0.52	0	1.050523	7.201001	110	18	1.2	0.8;
	6723	1	-0	0	0	0.16	0	0.995804	-4.792814	110	18	1.1	0.9;
	6724	1	672.7	107.6	0.115285	0.563929	0	1.043088	25.157234	150	2	1.1	0.9;
	6725	1	12.21	7.9	0	0.56	0	0.978253	12.237049	110	18	1.1	0.9;
	6726	2	0	0	0	0	0	1.016479	49.02148	20	8	1.1	0.9;
	6727	1	0	-0	0	75.38	0	1.017758	-22.793685	380	4	1.1	0.9;
	6728	1	5.77	2.68	0	1.72	0	1.034504	-7.558168	150	9	1.1	0.9;
	6729	2	0	0	0	0	0	1.058114	43.382414	20	10	1.1	0.9;
	6730	1	11.2	1.6	0	9.91	0	1.034133	19.549478	154	1	1.1	0.9;
	6731	2	0	0	0	0	0	1.043942	17.353218	20	5	1.1	0.9;
	6732	1	-0	-0	0	3.16	0	0.999144	-9.605939	110	22	1.1	0.9;
	6733	1	183.8	2.2	0	2.98	0	1.030889	-21.232311	220	4	1.1	0.9;
	6734	2	0	0	0	0	0	1.005341	-1.941399	20	4	1.1	0.9;
	6735	1	20.32	-24.8	0	7.45	0	1.016536	32.890325	220	3	1.1	0.9;
	6736	1	-1.73	-1.71	0	22.069454	0	0.994332	26.200415	380	5	1.1	0.9;
	6737	1	84.59	42.82	0	0.86	0	1.042294	37.004115	220	13	1.1	0.9;
	6738	1	170.9	39	0	11.25	0	1.0297	20.595183	220	5	1.1	0.9;
	6739	1	-0	-0	0	0.98	0	0.97387	1.322165	110	18	1.1	0.9;
	6740	1	0	-0	0	4.25	0	1.006743	63.028253	220	3	1.1	0.9;
	6741	1	0	0	0	52.958443	0	0.989853	20.337179	380	5	1.1	0.9;
	6742	1	0	-0	0	1.27	0	1.043229	0.510669	110	19	1.1	0.9;
	6743	1	19.12	7.4	0	21.97	0	1.014159	-22.67014	380	4	1.1	0.9;
	6744	1	0	-0	0	37.74	0	0.998584	90.169777	150	1	1.1	0.9;
	6745	1	37.94	18.9	0.436494	1.944442	0	1.090977	1.75798	380	21	1.2	0.8;
	6746	1	70.28	10.3	0	13.73	0	1.009314	-24.91449	220	4	1.1	0.9;
	6747	2	0	0	0	0	0	1.016595	10.054092	15.75	1	1.1	0.9;
	6748	1	0	0	0	4.36	0	1.049409	-5.379399	150	9	1.1	0.9;
	6749	2	0	0	0	0	0	1.046665	36.809947	20	5	1.1	0.9;
	6750	1	10.11	3.7	0	2.15	0	1.057831	-3.45976	110	19	1.1	0.9;
	6751	1	-0	0	0	1.69	0	1.027117	-8.230885	380	6	1.1	0.9;
	6752	1	10.11	-0	0	0.32	0	1.007116	2.912353	110	18	1.1	0.9;
	6753	1	0	-0	0	4.45	0	0.992412	40.794888	154	1	1.1	0.9;
	6754	2	0	0	0	0	0	1.073284	1.2557	20	9	1.1	0.9;
	6755	1	8.3	1.2	0	2.32	0	1.009101	17.686434	154	1	1.1	0.9;
	6756	1	-0	-0	0	16.93	0	1.027052	-12.831896	380	6	1.1	0.9;
	6757	2	0	0	0	0	0	1.043178	23.372652	20	5	1.1	0.9;
	6758	2	0	0	0	0	0	1.043178	23.372652	20	5	1.1	0.9;
	6759	2	0	0	0	0	0	1.043178	23.372652	20	5	1.1	0.9;
	6760	2	0	0	0	0	0	1.043178	23.372652	20	5	1.1	0.9;
	6761	2	0	0	0	0	0	1.043178	23.372652	20	5	1.1	0.9;
	6762	2	0	0	0	0	0	1.043178	23.372652	20	5	1.1	0.9;
	6763	2	0	0	0	0	0	1.043178	23.372652	20	5	1.1	0.9;
	6764	2	0	0	0	0	0	1.041274	37.583261	20	3	1.1	0.9;
	6765	2	0	0	0	0	0	1.042697	41.570544	20	3	1.1	0.9;
	6766	1	42.55	17.6	0	1.2	0	1.080871	3.730594	220	5	1.1	0.9;
	6767	1	135	25.7	0	43.01	0	1.023409	-7.389363	380	6	1.1	0.9;
	6768	1	37.94	51.4	0	0.55	0	1.050333	36.708894	380	3	1.1	0.9;
	6769	2	0	0	0	0	0	1.034441	34.869174	20	10	1.1	0.9;
	6770	1	1.9	0.6	0	1.06	0	0.973487	-13.777021	110	12	1.1	0.9;
	6771	1	-0	0	0	1.5	0	1.017201	14.001757	154	1	1.1	0.9;
	6772	1	-0	-0	0.109969	70.058359	0	1.088891	65.01165	380	3	1.2	0.8;
	6773	2	0	0	0	0	0	1.048358	31.36576	20	5	1.1	0.9;
	6774	2	0	0	0	0	0	1.005097	-2.034262	20	2	1.1	0.9;
	6775	1	26.03	16.1	0	0	0	1.020824	-8.385682	150	9	1.1	0.9;
	6776	1	40.04	15	0	1.59	0	0.992389	48.592226	220	3	1.1	0.9;
	6777	1	0	0	0	80.389436	0	0.990041	13.13447	380	5	1.1	0.9;
	6778	1	-0	-0	0.039145	0.050855	0	1.028164	13.427248	220	5	1.1	0.9;
	6779	1	162.1	73	0	0.3	0	1.007318	-7.256131	110	23	1.1	0.9;
	6780	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	6781	1	44.6	6.4	0	2.31	0	1.012828	23.567387	154	1	1.1	0.9;
	6782	1	7	1	0	0.87	0	0.995095	12.412171	154	1	1.1	0.9;
	6783	1	194.1	27.6	0	3.26	0	1.01937	14.556113	154	1	1.1	0.9;
	6784	2	0	0	0	0	0	1.014368	23.036796	6.3	1	1.1	0.9;
	6785	2	0	0	0	0	0	1.005097	-2.034262	20	2	1.1	0.9;
	6786	1	16.7	2.4	0	14.37	0	1.024439	14.22927	154	1	1.1	0.9;
	6787	1	18.2	2.6	0	1.59	0	1.029022	2.955197	154	1	1.1	0.9;
	6788	1	21.62	7.8	0	1.43	0	0.967902	-5.517482	110	19	1.1	0.9;
	6789	2	0	0	0	0	0	1.005097	-2.034262	20	2	1.1	0.9;
	6790	1	-0	-0	0	3.21	0	1.021786	-21.93928	220	4	1.1	0.9;
	6791	1	111.1	15.7	0	4.13	0	1.009835	16.142687	154	1	1.1	0.9;
	6792	2	0	0	0	0	0	1.024356	22.048022	20	5	1.1	0.9;
	6793	2	0	0	0	0	0	1.024356	22.048022	20	5	1.1	0.9;
	6794	2	0	0	0	0	0	1.017963	20.751946	20	5	1.1	0.9;
	6795	2	0	0	0	0	0	1.017963	20.751946	20	5	1.1	0.9;
	6796	1	137.8	64.5	0	5.41	0	1.009503	16.289821	380	3	1.1	0.9;
	6797	2	0	0	0	0	0	1.014343	23.052749	20	5	1.1	0.9;
	6798	2	0	0	0	0	0	1.014343	23.052749	20	5	1.1	0.9;
	6799	2	0	0	0	0	0	1.032132	-0.658602	20	9	1.1	0.9;
	6800	1	0	0	0	0	0	1.055797	-6.166252	150	9	1.1	0.9;
	6801	2	0	0	0	0	0	1.040566	26.548404	20	5	1.1	0.9;
	6802	2	0	0	0	0	0	1.040566	26.548404	20	5	1.1	0.9;
	6803	1	-0	0	0	0.11	0	0.999928	-8.812998	110	12	1.1	0.9;
	6804	1	75.58	-7.5	0	8.08	0	1.011878	14.800866	380	3	1.1	0.9;
	6805	1	280.7	53.7	0.038299	42.191701	0	1.046245	19.550585	380	13	1.1	0.9;
	6806	2	0	0	0	0	0	1.021328	-12.907544	20	4	1.1	0.9;
	6807	2	0	0	0	0	0	1.021328	-12.907544	20	4	1.1	0.9;
	6808	2	0	0	0	0	0	1.021328	-12.907544	20	4	1.1	0.9;
	6809	1	0	-0	0	0	0	1.012736	-19.855402	20	4	1.1	0.9;
	6810	1	0	-0	0	0	0	1.012736	-19.855402	20	4	1.1	0.9;
	6811	1	0	-0	0	0	0	1.012736	-19.855402	20	4	1.1	0.9;
	6812	1	0	-0	0	0	0	1.012736	-19.855402	20	4	1.1	0.9;
	6813	1	17.92	7.2	0	0.8	0	0.995335	-4.631918	110	21	1.1	0.9;
	6814	2	0	0	0	0	0	1.021328	-12.907544	20	4	1.1	0.9;
	6815	2	0	0	0	0	0	1.021328	-12.907544	20	4	1.1	0.9;
	6816	1	-0	0	0	1.59	0	1.01453	29.775973	154	1	1.1	0.9;
	6817	1	1.7	0.6	0	0.51	0	0.990631	-4.053386	110	18	1.1	0.9;
	6818	1	0	-0	0	1.45	0	1.047319	-3.179328	150	9	1.1	0.9;
	6819	1	3.7	1.3	0	0.9	0	1.035139	-3.879299	110	19	1.1	0.9;
	6820	2	0	0	0	0	0	1.045213	9.250128	20	14	1.1	0.9;
	6821	1	18.22	8.6	0	34.7	0	1.008047	-17.105722	380	4	1.1	0.9;
	6822	2	0	0	0	0	0	0.991398	31.360075	15.75	1	1.1	0.9;
	6823	1	0	-0	0.334497	0.149596	0	1.019862	29.08678	380	3	1.1	0.9;
	6824	1	-0	-0	0	3.72	0	1.029912	29.766421	220	3	1.1	0.9;
	6825	1	0	0	0.058975	11.000854	0	1.027243	48.935361	380	3	1.1	0.9;
	6826	1	41.75	13.2	0	3.36	0	1.016012	-8.320606	110	22	1.1	0.9;
	6827	1	5.01	2	0	0.77	0	1.022241	-5.274029	110	23	1.1	0.9;
	6828	1	-0	0	0	8.1	0	1.017129	-26.819281	220	4	1.1	0.9;
	6829	1	-0	-0	0	165.63	0	1.040686	-17.485492	380	4	1.1	0.9;
	6830	1	0	-0	0	0.84	0	0.969413	5.081044	110	18	1.1	0.9;
	6831	1	0	0	0	5.63	0	1.045272	2.414404	154	1	1.1	0.9;
	6832	1	28.83	9.6	0	0.85	0	0.975583	-8.871913	110	21	1.1	0.9;
	6833	1	4.61	1.7	0	0.88	0	1.050095	-2.695471	110	19	1.1	0.9;
	6834	1	93.9	-0.4	0	3.66	0	1.03336	9.2202	220	5	1.1	0.9;
	6835	1	68.38	-24.4	0.195436	7.063423	0	1.033688	-11.140358	220	4	1.1	0.9;
	6836	2	0	0	0	0	0	0.990475	21.129805	15.75	1	1.1	0.9;
	6837	1	27.13	16.8	0	0.28	0	1.066325	1.237267	150	9	1.1	0.9;
	6838	1	16.52	3.2	0	0.9	0	0.956799	-6.8869	110	18	1.1	0.9;
	6839	1	100.6	15.7	0.019802	5.520198	0	1.041567	2.843854	220	5	1.1	0.9;
	6840	1	87.9	75	1.087555	58.933642	0	1.082871	70.71862	380	3	1.1	0.9;
	6841	1	204.2	41.1	0.059977	12.500023	0	1.014509	-22.794624	220	4	1.1	0.9;
	6842	2	0	0	0	0	0	1.034464	34.885068	20	10	1.1	0.9;
	6843	1	13.21	4.4	0	0.38	0	0.944554	-8.604833	110	21	1.1	0.9;
	6844	1	390.5	62.5	0	0	0	1.039281	25.115606	150	2	1.1	0.9;
	6845	1	0	-0	0.304942	0.015882	0	0.99738	13.433482	220	6	1.1	0.9;
	6846	2	0	0	0	0	0	1.029146	3.188044	20	20	1.1	0.9;
	6847	2	0	0	0	0	0	1.024018	19.935265	20	10	1.1	0.9;
	6848	1	-0.58	-2.28	0	0.56	0	1.015475	-3.187287	110	19	1.1	0.9;
	6849	1	11.41	5.1	0	0.15	0	1.014697	-0.819702	110	18	1.1	0.9;
	6850	1	0.3	97.4	0.556326	52.18058	0	1.068548	77.186837	380	3	1.1	0.9;
	6851	1	16.62	5.5	0	1.34	0	0.980722	-5.618401	110	21	1.1	0.9;
	6852	1	91.2	56.1	0	8.3	0	0.993093	-2.60065	220	6	1.1	0.9;
	6853	1	0	0	0	1.06	0	0.940314	-11.534037	150	9	1.1	0.9;
	6854	1	124	29.2	0	0.53	0	1.024423	19.04153	220	5	1.1	0.9;
	6855	1	0.2	0.8	0	0.04	0	1.075022	46.081152	220	3	1.1	0.9;
	6856	1	0	-0	0.442888	6.15938	0	1.085669	11.310275	380	18	1.2	0.8;
	6857	1	-0	0	0	0.66	0	1.027676	-4.853286	220	1	1.1	0.9;
	6858	1	25.13	8.2	0	0.89	0	1.036576	5.994521	110	21	1.1	0.9;
	6859	1	102.7	17.8	0	48.05	0	1.035838	9.663321	220	5	1.1	0.9;
	6860	1	13.82	1.4	0	0.7	0	1.026477	14.511622	220	5	1.1	0.9;
	6861	2	0	0	0	0	0	1.027518	-17.089784	20	4	1.1	0.9;
	6862	2	0	0	0	0	0	1.039025	12.06939	20	6	1.1	0.9;
	6863	1	-0	0	0	0.6	0	0.986618	-2.650237	110	19	1.1	0.9;
	6864	2	0	0	0	0	0	0.984601	26.113187	6.3	1	1.1	0.9;
	6865	2	0	0	0	0	0	1.036049	1.565792	20	2	1.1	0.9;
	6866	1	84.09	1	0	32.3	0	1.022247	8.824908	380	11	1.1	0.9;
	6867	1	142.6	50.6	0.197772	31.051671	0	1.014038	-16.164128	220	4	1.1	0.9;
	6868	1	0	-0	0	23.94	0	1.050781	93.280802	380	3	1.1	0.9;
	6869	1	186.6	56.6	0	30.24	0	1.076798	3.635802	380	20	1.2	0.8;
	6870	2	0	0	0	0	0	1.023144	-9.538309	20	4	1.1	0.9;
	6871	2	0	0	0	0	0	0.984599	43.311569	14.4	1	1.1	0.9;
	6872	1	16.02	5.1	0	0.63	0	1.008741	36.754291	110	8	1.1	0.9;
	6873	1	-0	0	0	8.67	0	1.039224	12.699314	380	1	1.1	0.9;
	6874	1	0	-0	0	43.7	0	0.977566	23.48751	400	1	1.1	0.9;
	6875	2	0	0	0	0	0	1.024647	-2.171451	20	6	1.1	0.9;
	6876	2	0	0	0	0	0	1.024647	-2.171451	20	6	1.1	0.9;
	6877	2	0	0	0	0	0	1.024647	-2.171451	20	6	1.1	0.9;
	6878	1	0.8	45.3	0.545381	16.735815	0	1.014043	-17.464868	380	4	1.1	0.9;
	6879	2	0	0	0	0	0	1.066491	41.851274	20	5	1.1	0.9;
	6880	2	0	0	0	0	0	1.066564	41.739182	20	5	1.1	0.9;
	6881	1	-0	8.3	0	10.19	0	1.102706	-15.26137	150	9	1.2	0.8;
	6882	1	-0	0	0	50	0	0.989077	23.626111	150	8	1.1	0.9;
	6883	1	29.38	-14.09	0	1.33	0	1.024983	-1.567143	110	19	1.1	0.9;
	6884	1	10.81	3.9	0	0.61	0	1.039455	-2.790506	110	15	1.1	0.9;
	6885	2	0	0	0	0	0	1.127948	6.872839	20	9	1.2	0.8;
	6886	2	0	0	0	0	0	1.023096	23.468478	20	10	1.1	0.9;
	6887	1	124.3	42	0.039324	3.310676	0	1.007022	25.795589	220	13	1.1	0.9;
	6888	1	15.82	7.2	0	1.9	0	0.98601	-3.642105	110	18	1.1	0.9;
	6889	1	0	0	0.147967	92.214944	0	1.088824	9.099117	380	18	1.2	0.8;
	6890	1	25.33	8.8	0	1.39	0	0.998448	-7.205564	110	22	1.1	0.9;
	6891	1	0	-0	0	3.4	0	0.965152	-9.606927	150	9	1.1	0.9;
	6892	1	-0	0	0	0	0	1.028513	-22.307748	20	4	1.1	0.9;
	6893	1	-0	0	0	0	0	1.007942	-10.569275	20	4	1.1	0.9;
	6894	1	8.81	3.2	0	1.12	0	0.964842	-2.150668	110	19	1.1	0.9;
	6895	1	0	-0	0.178742	3.270839	0	1.040858	19.470405	380	5	1.1	0.9;
	6896	1	136.8	4.8	0	10	0	1.042756	13.157799	220	3	1.1	0.9;
	6897	1	19.72	7.1	0	0.28	0	1.012562	-1.688332	110	15	1.1	0.9;
	6898	1	-0	0	0.058489	3.001511	0	1.021496	-8.698463	220	4	1.1	0.9;
	6899	1	19.62	7.4	0	0.76	0	1.063802	1.596336	110	15	1.1	0.9;
	6900	1	0	0	0	20.81	0	1.034333	-18.758757	380	4	1.1	0.9;
	6901	1	59.2	8.4	0	2.26	0	1.004217	26.321861	154	1	1.1	0.9;
	6902	1	54.76	33.9	0	0.46	0	0.999524	-9.834457	150	9	1.1	0.9;
	6903	2	0	0	0	0	0	1.042446	6.957021	20	22	1.1	0.9;
	6904	1	-0	0	0.06009	0.19991	0	0.97934	5.474588	110	18	1.1	0.9;
	6905	1	21.42	11.8	0	28.85	0	0.995712	14.84413	380	5	1.1	0.9;
	6906	1	195.4	76.84	0	8.59	0	1.001417	19.79822	220	5	1.1	0.9;
	6907	1	0	0	0	0	0	1.030589	-7.348342	150	9	1.1	0.9;
	6908	1	14.22	5.1	0	0.39	0	1.044426	-2.371415	110	15	1.1	0.9;
	6909	2	0	0	0	0	0	1.023155	-9.729643	20	6	1.1	0.9;
	6910	2	0	0	0	0	0	1.023155	-9.729643	20	6	1.1	0.9;
	6911	1	57.46	17.1	0	1.61	0	1.019724	22.195041	220	5	1.1	0.9;
	6912	2	0	0	0	0	0	1.038747	-8.642337	20	4	1.1	0.9;
	6913	1	-0	0	0	23.37	0	1.035579	-21.142018	220	6	1.1	0.9;
	6914	1	392	177.7	0.181549	10.880889	0	1.027597	3.370206	220	18	1.1	0.9;
	6915	1	-0	-0	0	46.25	0	1.028275	-14.834395	380	4	1.1	0.9;
	6916	1	97	57.4	0	93.82	0	1.018879	16.2204	380	3	1.1	0.9;
	6917	1	-2.91	-0.6	0	0.02	0	0.996141	-3.969887	220	6	1.1	0.9;
	6918	2	0	0	0	0	0	1.073133	28.039751	20	14	1.1	0.9;
	6919	1	-0	-0	0	4.51	0	1.010871	90.224838	150	1	1.1	0.9;
	6920	1	-0	0	0	6.11	0	0.976017	-27.442601	150	10	1.1	0.9;
	6921	1	92.1	38	0	4.93	0	0.926333	5.238493	220	16	1.1	0.9;
	6922	1	0	-0	0.178535	0.81114	0	1.018554	4.748595	220	20	1.1	0.9;
	6923	1	0	0	0	0	0	1.021718	-10.711519	150	9	1.1	0.9;
	6924	1	4.91	1.6	0	4.15	0	1.018374	18.572044	220	5	1.1	0.9;
	6925	1	5.81	3.8	0	0.94	0	0.963581	-5.373919	110	18	1.1	0.9;
	6926	1	2	0.6	0	0.53	0	0.947279	-7.863736	110	21	1.1	0.9;
	6927	1	66.97	18.6	0	18.5	0	1.027734	22.667054	220	5	1.1	0.9;
	6928	1	3.1	0.1	0	1.34	0	1.028182	8.896762	220	5	1.1	0.9;
	6929	1	0	-0	0	2.38	0	1.035431	27.220462	220	5	1.1	0.9;
	6930	2	0	0	0	0	0	1.028641	-17.676734	20	4	1.1	0.9;
	6931	2	0	0	0	0	0	1.028641	-17.676734	20	4	1.1	0.9;
	6932	1	-0	0	0	2.42	0	1.039723	-1.23119	150	9	1.1	0.9;
	6933	1	17.7	-334.5	0.131772	55.711176	0	1.051894	44.335758	380	3	1.1	0.9;
	6934	1	-0	0	0	0	0	1.05053	43.952891	380	8	1.1	0.9;
	6935	1	-0	-0	0.135375	4.892175	0	1.075692	2.382142	380	18	1.2	0.8;
	6936	1	0	-0	0	0.41	0	0.967292	-3.807226	110	19	1.1	0.9;
	6937	1	0	0	0	25.368606	0	1.017088	18.99512	380	5	1.1	0.9;
	6938	1	-0	-0	0	13.59	0	1.008996	-16.587104	380	1	1.1	0.9;
	6939	2	0	0	0	0	0	0.994858	32.812983	15	1	1.1	0.9;
	6940	1	110.6	20.4	0	26.9	0	0.996842	-7.284318	220	6	1.1	0.9;
	6941	1	0	-0	0.099501	43.903596	0	1.080967	7.377587	380	18	1.2	0.8;
	6942	2	0	0	0	0	0	1.017532	25.788753	6.3	1	1.1	0.9;
	6943	1	101.89	25.72	0	5.11	0	1.029906	18.030995	220	5	1.1	0.9;
	6944	2	0	0	0	0	0	1.054031	9.216302	20	5	1.1	0.9;
	6945	1	0.4	0.3	0.039289	1.170711	0	0.996333	11.356692	110	18	1.1	0.9;
	6946	1	-0.15	-0.9	0	0	0	1.039744	48.885504	110	8	1.1	0.9;
	6947	1	17.12	6.3	0	0.51	0	0.96407	0.772625	110	18	1.1	0.9;
	6948	2	0	0	0	0	0	1.031397	6.88995	20	14	1.1	0.9;
	6949	1	0	-0	0	2.03	0	1.006367	26.85524	380	3	1.1	0.9;
	6950	1	13.77	18.73	0	0.46	0	1.080244	-12.388095	150	9	1.2	0.8;
	6951	1	2.2	-0	0	4.81	0	0.968179	-13.778323	220	10	1.1	0.9;
	6952	2	0	0	0	0	0	0.955439	48.700819	10.6	1	1.1	0.9;
	6953	1	58.97	18.1	0	0.32	0	0.969766	4.808904	110	18	1.1	0.9;
	6954	1	0	-0	0.181549	17.630889	0	1.027988	-1.986489	220	18	1.1	0.9;
	6955	1	93.3	27.8	0	6.1	0	1.027171	18.782805	220	5	1.1	0.9;
	6956	1	0	-0	0	1.49	0	1.020802	17.057351	220	5	1.1	0.9;
	6957	1	264	40.1	0	9.489841	0	1.007395	-5.72667	380	6	1.1	0.9;
	6958	1	0	-0	0	3.65	0	1.018519	26.913419	154	1	1.1	0.9;
	6959	1	615.8	286.7	0	101.04	0	1.017566	30.154577	150	8	1.1	0.9;
	6960	1	120.1	2.5	0	6.6	0	1.015516	-17.663901	220	4	1.1	0.9;
	6961	1	0	0	0	0	0	1.025451	-11.135864	150	9	1.1	0.9;
	6962	1	101.1	19.8	0	3.62	0	1.003289	-20.416251	220	4	1.1	0.9;
	6963	2	0	0	0	0	0	1.029893	-6.546635	20	4	1.1	0.9;
	6964	1	-0	0	0	1.4	0	1.048554	86.41124	220	3	1.1	0.9;
	6965	1	59.17	14	0	4.07	0	0.979077	4.930838	220	3	1.1	0.9;
	6966	1	-0	0	0	1.67	0	1.010708	19.950213	154	1	1.1	0.9;
	6967	2	0	0	0	0	0	1.054031	9.216302	20	5	1.1	0.9;
	6968	1	10.97	-6.31	0	1.14	0	0.920281	-11.399862	110	18	1.1	0.9;
	6969	2	0	0	0	0	0	1.054031	9.216302	20	5	1.1	0.9;
	6970	1	-0	-0	0	5.57	0	1.063753	11.961612	220	5	1.1	0.9;
	6971	1	0	-0	0.34708	0.510766	0	1.048304	39.358328	380	3	1.1	0.9;
	6972	1	12.51	3.3	0	1.7	0	0.97159	-8.552571	110	18	1.1	0.9;
	6973	1	63.5	-381.2	0.514139	207.080426	0	1.065197	53.171221	380	3	1.1	0.9;
	6974	1	65.07	30	0	0.99	0	1.070029	0.064588	110	15	1.1	0.9;
	6975	1	173.4	31	0	1.91	0	1.009578	3.843937	220	7	1.1	0.9;
	6976	2	0	0	0	0	0	1.050684	91.443424	20	3	1.1	0.9;
	6977	1	-0.15	-1	0	0.42	0	1.032362	39.964728	110	8	1.1	0.9;
	6978	1	31.94	10.5	0	1.02	0	0.936583	-8.21497	110	21	1.1	0.9;
	6979	1	14.92	5.1	0	0.65	0	0.983945	-6.05191	110	18	1.1	0.9;
	6980	1	33.04	10.1	0	0.43	0	1.019273	16.993216	220	5	1.1	0.9;
	6981	1	121.9	-0.2	0	12.83	0	1.012504	22.41158	220	3	1.1	0.9;
	6982	1	-0	-0	0	1.04	0	1.004156	-13.474558	110	12	1.1	0.9;
	6983	1	8.41	2.6	0	1.77	0	1.019528	43.583518	110	8	1.1	0.9;
	6984	1	0	-0	0.078399	0.891201	0	1.02293	-17.553442	220	6	1.1	0.9;
	6985	2	0	0	0	0	0	1.029991	3.188016	20	6	1.1	0.9;
	6986	1	13.9	1.9	0	3.89	0	1.000178	13.361626	154	1	1.1	0.9;
	6987	2	0	0	0	0	0	1.029991	3.188016	20	6	1.1	0.9;
	6988	1	-0	-0	0	0.6	0	1.035766	1.983491	110	3	1.1	0.9;
	6989	1	0	-0	0.3205	28.133148	0	1.069722	2.806287	380	18	1.2	0.8;
	6990	2	0	0	0	0	0	1.027131	-1.211905	20	18	1.1	0.9;
	6991	1	24.93	13.7	0.155083	-0.105083	0	1.026001	0.562799	220	21	1.1	0.9;
	6992	2	0	0	0	0	0	1.034688	8.184132	20	18	1.1	0.9;
	6993	1	87.5	25.9	0.098609	24.561391	0	1.019217	-20.70702	220	4	1.1	0.9;
	6994	2	0	0	0	0	0	1.003942	39.409116	11	1	1.1	0.9;
	6995	1	-0	0	0.039707	7.680293	0	1.020015	-15.016249	220	4	1.1	0.9;
	6996	2	0	0	0	0	0	1.0182	22.6787	20	6	1.1	0.9;
	6997	1	63.27	28.9	0	2.19	0	1.020504	28.639428	150	8	1.1	0.9;
	6998	2	0	0	0	0	0	1.033534	-10.544863	20	4	1.1	0.9;
	6999	1	153.2	31.5	0	12.56	0	1.047214	5.338776	220	11	1.1	0.9;
	7000	1	0	-0	0.506518	-0.393958	0	1.068832	52.609074	380	8	1.1	0.9;
	7001	1	97.5	13.9	0	1.24	0	1.02173	-0.10958	154	1	1.1	0.9;
	7002	1	-0	0	0	0.01	0	0.981801	-7.8197	150	9	1.1	0.9;
	7003	1	0	-0	0	3.52	0	1.000892	44.245377	380	8	1.1	0.9;
	7004	1	2.5	1.5	0	0.83	0	0.971372	-5.891139	110	18	1.1	0.9;
	7005	2	0	0	0	0	0	1.04062	33.045458	20	10	1.1	0.9;
	7006	1	40.14	24.9	0	5.71	0	0.963378	-9.81008	150	9	1.1	0.9;
	7007	1	21.02	14.8	0	0.1	0	1.006897	-8.792843	150	9	1.1	0.9;
	7008	1	-0	-0	0.573295	-0.286693	0	1.058048	38.30669	220	5	1.1	0.9;
	7009	2	0	0	0	0	0	1.023432	25.118071	10.5	1	1.1	0.9;
	7010	1	69.78	22.4	0	2.72	0	0.958921	-8.905383	110	21	1.1	0.9;
	7011	1	-0	0	0	6.26	0	1.030485	29.794078	220	3	1.1	0.9;
	7012	1	89.9	40.2	0	0.37	0	1.023179	44.762006	380	3	1.1	0.9;
	7013	1	71.98	27.1	0	7.2	0	0.986883	15.785624	380	5	1.1	0.9;
	7014	1	146.9	-5.1	0	0	0	1.088419	29.405227	220	3	1.2	0.8;
	7015	1	0	0	0	0	0	0.998601	-10.373841	150	9	1.1	0.9;
	7016	1	9.11	3	0	1.24	0	0.999384	-8.318633	110	12	1.1	0.9;
	7017	1	-0	0	0	29.91	0	1.032373	-8.516164	150	9	1.1	0.9;
	7018	1	336.9	81	0	19.53	0	1.011587	23.671261	220	3	1.1	0.9;
	7019	1	11.41	3.8	0	0.36	0	1.057103	7.110478	110	18	1.2	0.8;
	7020	2	0	0	0	0	0	1.033214	-6.888934	20	6	1.1	0.9;
	7021	1	-0	0	0	81.29	0	1.047209	-16.441064	380	4	1.1	0.9;
	7022	1	0	0	0	11.82	0	0.995912	22.400101	380	3	1.1	0.9;
	7023	1	-9.6	-2.5	0	2.15	0	1.103415	-1.649909	150	9	1.2	0.8;
	7024	1	0	-0	0	0	0	1.098193	65.001322	380	8	1.2	0.8;
	7025	1	0	0	0	0	0	1.020824	-8.385682	150	9	1.1	0.9;
	7026	1	0	-0	0	2.62	0	1.027345	33.608464	380	1	1.1	0.9;
	7027	1	0	-0	0	0.34	0	1.046939	-6.613174	150	9	1.1	0.9;
	7028	2	0	0	0	0	0	1.043087	11.638661	20	14	1.1	0.9;
	7029	1	-0	0	0.252818	66.30157	0	1.072481	-0.655471	380	21	1.2	0.8;
	7030	1	3.2	1.2	0	0.3	0	0.984025	-1.613209	110	19	1.1	0.9;
	7031	2	0	0	0	0	0	1.012542	1.713537	20	18	1.1	0.9;
	7032	1	20.12	4.9	0	1.61	0	0.993843	-5.877347	110	18	1.1	0.9;
	7033	1	-0	0	0	2.03	0	0.978182	-5.962125	110	21	1.1	0.9;
	7034	1	9.71	3.2	0	0.26	0	0.992755	-14.102923	110	12	1.1	0.9;
	7035	1	118.6	43.4	0	0.8	0	1.012334	7.544783	220	3	1.1	0.9;
	7036	1	66.67	0	0	0.11	0	1.007989	4.843111	220	7	1.1	0.9;
	7037	1	210.8	-0.7	0	4.14	0	1.033519	-19.312338	220	4	1.1	0.9;
	7038	1	39.84	6.4	0	34.58	0	1.043944	-14.401464	380	10	1.1	0.9;
	7039	1	0	-0	0	0	0	1.022632	4.381728	20	14	1.1	0.9;
	7040	1	69.98	21.3	0.039275	3.360725	0	1.011322	-3.298846	110	18	1.1	0.9;
	7041	1	151	7.1	0	4.24	0	1.011233	-18.289423	220	4	1.1	0.9;
	7042	1	0	-0	0	40.13	0	1.030455	21.683383	380	3	1.1	0.9;
	7043	1	11.2	1.6	0	4.75	0	1.005654	21.004212	154	1	1.1	0.9;
	7044	1	141.36	53.5	0	2.76	0	0.982381	-22.951765	220	6	1.1	0.9;
	7045	2	0	0	0	0	0	1.041985	31.793879	20	10	1.1	0.9;
	7046	1	41.25	13.5	0	0.39	0	0.959203	-6.211295	110	21	1.1	0.9;
	7047	1	0	0	0	0.15	0	1.012145	-2.02475	110	21	1.1	0.9;
	7048	1	112.4	13	0	3.99	0	1.025542	-7.188187	220	5	1.1	0.9;
	7049	1	270.5	77.8	0	81.25	0	1.067375	2.210525	380	20	1.1	0.9;
	7050	1	110.2	15.7	0	2.3	0	1.011297	28.945748	154	1	1.1	0.9;
	7051	1	0	-0	0.574676	1.582208	0	1.036449	33.240018	380	5	1.1	0.9;
	7052	2	0	0	0	0	0	0.99693	24.855313	10.6	1	1.1	0.9;
	7053	1	-0	-0	0	0	0	1.002012	13.542185	20	5	1.1	0.9;
	7054	1	-0	-0	0	0	0	1.002012	13.542185	20	5	1.1	0.9;
	7055	2	0	0	0	0	0	1.046478	33.359815	20	3	1.1	0.9;
	7056	2	0	0	0	0	0	1.028102	1.989595	20	18	1.1	0.9;
	7057	2	0	0	0	0	0	1.040138	32.45384	20	17	1.1	0.9;
	7058	1	-0	-0	0	0.06	0	1.085466	-20.194962	120	6	1.2	0.8;
	7059	1	-0.15	-1.4	0	0.01	0	1.018502	51.344505	110	8	1.1	0.9;
	7060	1	-0	-0	0	0.83	0	1.003292	-9.321509	150	9	1.1	0.9;
	7061	1	183.03	15.7	0	1.83	0	1.005676	-22.163188	220	6	1.1	0.9;
	7062	1	-0	0	0.009921	-0.009921	0	1.01975	13.979643	220	5	1.1	0.9;
	7063	2	0	0	0	0	0	1.058676	15.850616	20	5	1.1	0.9;
	7064	2	0	0	0	0	0	1.015531	4.163384	20	9	1.1	0.9;
	7065	1	-0	-0	0.670574	55.699129	0	1.089137	9.040939	380	21	1.2	0.8;
	7066	2	0	0	0	0	0	0.994601	31.575691	11	1	1.1	0.9;
	7067	1	-0	0	0	46.7	0	1.014955	35.533143	380	8	1.1	0.9;
	7068	2	0	0	0	0	0	1.026372	24.396837	11	1	1.1	0.9;
	7069	1	0	0	0.176261	1.013116	0	1.029226	35.370949	220	3	1.1	0.9;
	7070	2	0	0	0	0	0	1.002967	-14.170517	20	4	1.1	0.9;
	7071	2	0	0	0	0	0	1.058114	43.382414	20	10	1.1	0.9;
	7072	2	0	0	0	0	0	1.036699	34.476926	20	10	1.1	0.9;
	7073	2	0	0	0	0	0	1.034768	8.492541	20	22	1.1	0.9;
	7074	1	0	0	0	3.38	0	1.047226	85.684611	220	3	1.1	0.9;
	7075	2	0	0	0	0	0	1.042846	0.690036	20	18	1.1	0.9;
	7076	1	0	0	0	0	0	1.031735	-7.53735	150	9	1.1	0.9;
	7077	1	171.9	39.1	0	4.23	0	1.000639	23.86914	220	5	1.1	0.9;
	7078	1	-0	0	0.122915	25.570901	0	1.024673	19.777124	380	7	1.1	0.9;
	7079	2	0	0	0	0	0	1.024254	10.943171	20	6	1.1	0.9;
	7080	2	0	0	0	0	0	1.020703	22.380203	20	5	1.1	0.9;
	7081	2	0	0	0	0	0	1.020703	22.380203	20	5	1.1	0.9;
	7082	2	0	0	0	0	0	1.020703	22.380203	20	5	1.1	0.9;
	7083	2	0	0	0	0	0	1.020703	22.380203	20	5	1.1	0.9;
	7084	2	0	0	0	0	0	1.007614	14.656778	20	5	1.1	0.9;
	7085	2	0	0	0	0	0	1.020703	22.380203	20	5	1.1	0.9;
	7086	1	19.22	6.4	0	31.78	0	1.015038	-11.388711	220	6	1.1	0.9;
	7087	2	0	0	0	0	0	1.007614	14.656778	20	5	1.1	0.9;
	7088	1	-0	0	0	0	0	1.013595	22.381001	20	5	1.1	0.9;
	7089	1	5.11	2.5	0	0.57	0	0.925516	-8.685239	110	18	1.1	0.9;
	7090	1	0	0	0	0.81	0	1.024048	35.943886	154	1	1.1	0.9;
	7091	1	34.14	21.1	0	0	0	1.102719	-15.262746	150	9	1.2	0.8;
	7092	1	-0	0	0.11	4.95	0	1.033129	-0.064448	380	6	1.1	0.9;
	7093	1	-10.28	-2.68	0	0.9	0	1.016327	7.817231	110	18	1.1	0.9;
	7094	1	19.82	6.5	0	1.47	0	0.970506	-5.960988	110	18	1.1	0.9;
	7095	1	121.6	49.3	0	10	0	0.972641	29.929237	220	3	1.1	0.9;
	7096	1	151.1	66.9	0	0.41	0	1.009754	12.380416	380	3	1.1	0.9;
	7097	1	25.23	15.6	0	0	0	1.008563	-8.920839	150	9	1.1	0.9;
	7098	1	54.96	16.9	0	3.37	0	1.006535	-14.099903	110	22	1.1	0.9;
	7099	1	0	0	0.340119	-0.264537	0	1.055931	33.95161	150	8	1.1	0.9;
	7100	1	-0	0	0	31.14	0	1.041105	-14.651391	380	4	1.1	0.9;
	7101	1	173.2	-7	0.495294	14.757893	0	0.992834	59.276406	220	13	1.1	0.9;
	7102	1	0	0	0	1.36	0	0.925181	-12.121466	150	9	1.1	0.9;
	7103	1	-0	-0	0.172715	-0.134334	0	1.033165	38.744307	150	8	1.1	0.9;
	7104	1	347.3	105.6	0.230364	23.317628	0	1.049732	40.027444	380	13	1.1	0.9;
	7105	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	7106	1	58.06	4.1	0.231528	17.426707	0	1.055771	-2.521165	220	23	1.1	0.9;
	7107	1	90.5	12.8	0	0.78	0	1.024203	1.49438	154	1	1.1	0.9;
	7108	1	0.2	0.5	0.688989	47.290598	0	1.021296	19.962313	380	3	1.1	0.9;
	7109	1	0	-0	0	0.43	0	1.001948	-7.820734	150	9	1.1	0.9;
	7110	2	0	0	0	0	0	1.020566	-15.965907	20	6	1.1	0.9;
	7111	2	0	0	0	0	0	0.995435	4.377334	20	9	1.1	0.9;
	7112	1	45.55	-59.9	0	30.04	0	1.036262	9.976334	220	5	1.1	0.9;
	7113	2	0	0	0	0	0	1.034235	2.214597	20	6	1.1	0.9;
	7114	2	0	0	0	0	0	1.014725	27.348397	11.5	1	1.1	0.9;
	7115	2	0	0	0	0	0	1.040138	32.45384	20	17	1.1	0.9;
	7116	2	0	0	0	0	0	1.042317	14.274354	20	14	1.1	0.9;
	7117	1	125.7	17.9	0	0.2	0	1.015873	20.575127	154	1	1.1	0.9;
	7118	1	127.9	-34	0	19.86	0	1.021335	-32.216893	220	4	1.1	0.9;
	7119	1	-0	0	0	1.92	0	0.99907	17.635445	154	1	1.1	0.9;
	7120	1	78.69	19.6	0	0.87	0	1.006299	-4.259923	110	3	1.1	0.9;
	7121	1	0	0	0	12.69	0	1.012124	39.690958	380	3	1.1	0.9;
	7122	2	0	0	0	0	0	1.03036	-6.512869	20	4	1.1	0.9;
	7123	1	261.2	105.6	0	28.53	0	1.059752	82.219607	380	3	1.1	0.9;
	7124	1	0	-0	0	29.32	0	0.999402	23.601458	380	3	1.1	0.9;
	7125	1	0	-0	0.09	7.22	0	1.013229	-2.688715	220	21	1.1	0.9;
	7126	1	0.1	2.2	0	1.22	0	1.013777	35.825827	220	3	1.1	0.9;
	7127	1	-5.88	-3.76	0	0.39	0	0.995485	-2.095523	110	18	1.1	0.9;
	7128	1	-0	0	0	0	0	1.02227	36.7768	150	8	1.1	0.9;
	7129	1	0	-0	0	4.86	0	1.028248	5.309757	220	5	1.1	0.9;
	7130	1	0	0	0	0	0	0.86014	-15.783699	150	9	1.2	0.8;
	7131	1	169	56.4	0	8.06	0	1.024163	46.575972	380	3	1.1	0.9;
	7132	1	-0	0	0	9.74	0	1.035363	9.414958	220	5	1.1	0.9;
	7133	2	0	0	0	0	0	1.044129	-4.738961	20	4	1.1	0.9;
	7134	1	50.56	18	0	0.11	0	1.026317	18.877265	220	5	1.1	0.9;
	7135	1	2.2	0.5	0	2.88	0	0.981282	-20.456784	150	10	1.1	0.9;
	7136	1	0	0	0	8.1	0	1.011074	22.948932	220	6	1.1	0.9;
	7137	2	0	0	0	0	0	1.062747	39.318661	20	8	1.1	0.9;
	7138	1	0	-0	0	0.28	0	0.99524	-1.245843	110	18	1.1	0.9;
	7139	2	0	0	0	0	0	1.034631	4.683321	20	14	1.1	0.9;
	7140	1	65.5	9.3	0	5.31	0	0.964795	30.02488	154	1	1.1	0.9;
	7141	1	-0	-0	0	0	0	1.020596	-16.730831	20	4	1.1	0.9;
	7142	2	0	0	0	0	0	1.033768	14.596978	20	14	1.1	0.9;
	7143	2	0	0	0	0	0	1.017775	-9.583337	20	4	1.1	0.9;
	7144	1	-0	-0	0	0	0	1.024103	-14.185965	20	4	1.1	0.9;
	7145	1	15.42	3.9	0	0.79	0	1.012621	36.142762	110	8	1.1	0.9;
	7146	2	0	0	0	0	0	1.017024	16.872664	20	18	1.1	0.9;
	7147	1	262.6	31.8	0	8.13	0	0.972792	-17.770017	220	6	1.1	0.9;
	7148	2	0	0	0	0	0	0.962614	39.578299	15.75	1	1.1	0.9;
	7149	2	0	0	0	0	0	0.969146	47.259579	13.8	1	1.1	0.9;
	7150	1	0	-0	0	0.95	0	1.058684	17.914192	220	5	1.1	0.9;
	7151	1	-0	-0	0	2.74	0	0.994222	33.464801	220	3	1.1	0.9;
	7152	1	230.9	42.6	0	12.62	0	1.008396	5.114162	220	7	1.1	0.9;
	7153	1	25.73	9.3	0	0.12	0	0.978315	-3.345335	110	19	1.1	0.9;
	7154	1	5.11	1.7	0	0.48	0	1.000441	-8.732796	110	12	1.1	0.9;
	7155	1	0	0	0	3.48	0	1.013461	26.007513	380	3	1.1	0.9;
	7156	2	0	0	0	0	0	1.019539	-2.214354	20	10	1.1	0.9;
	7157	1	135.9	24.7	0.078499	1.141501	0	1.007191	11.825783	220	5	1.1	0.9;
	7158	1	5.11	-0	0	4.97	0	0.953148	-6.753445	110	18	1.1	0.9;
	7159	2	0	0	0	0	0	0.989224	44.932169	154	1	1.1	0.9;
	7160	1	-0	0	0	1.26	0	1.00838	-9.715089	150	9	1.1	0.9;
	7161	1	0	0	0.51746	3.458783	0	1.050533	36.718919	380	3	1.1	0.9;
	7162	1	-0	0	0	3.08	0	1.068034	87.318363	380	3	1.1	0.9;
	7163	2	0	0	0	0	0	1.025208	-17.00361	20	6	1.1	0.9;
	7164	2	0	0	0	0	0	1.054937	4.624127	20	20	1.1	0.9;
	7165	1	45.45	7.1	0	1.05	0	1.023258	0.195954	110	3	1.1	0.9;
	7166	1	186.9	26.6	0	11.96	0	0.979125	5.224833	400	1	1.1	0.9;
	7167	1	-0	-0	0	0	0	0.984641	13.821924	20	18	1.1	0.9;
	7168	2	0	0	0	0	0	1.045359	10.101302	20	14	1.1	0.9;
	7169	1	16.92	5.6	0	0.2	0	0.987683	-14.465453	110	12	1.1	0.9;
	7170	2	0	0	0	0	0	0.989707	35.937223	11	1	1.1	0.9;
	7171	1	185	43.6	0	6.3	0	1.025863	-20.929808	220	4	1.1	0.9;
	7172	1	25.1	3.6	0	1.74	0	0.997885	10.689103	154	1	1.1	0.9;
	7173	2	0	0	0	0	0	1.127961	6.657565	20	9	1.2	0.8;
	7174	1	-0	-0	0	39.21	0	1.034257	31.986033	380	1	1.1	0.9;
	7175	1	21.62	7.8	0	0.68	0	1.021239	0.375074	110	19	1.1	0.9;
	7176	1	0	0	0	0.34	0	0.974695	-8.316245	150	9	1.1	0.9;
	7177	1	-7.61	2.5	0	0.5	0	1.043125	0.745455	150	9	1.1	0.9;
	7178	2	0	0	0	0	0	1.037281	-8.202796	20	4	1.1	0.9;
	7179	2	0	0	0	0	0	1.004576	18.337057	20	5	1.1	0.9;
	7180	1	329.7	32.4	0	4.5	0	1.012085	-31.074125	220	4	1.1	0.9;
	7181	1	32.24	8.9	0.039602	1.700398	0	1.013408	-18.555259	220	4	1.1	0.9;
	7182	1	55.7	7.9	0	0.1	0	1.005335	12.373439	154	1	1.1	0.9;
	7183	2	0	0	0	0	0	1.026147	8.391166	20	6	1.1	0.9;
	7184	1	-0	0	0.12	12.76	0	0.999967	-2.971427	220	21	1.1	0.9;
	7185	1	-0	-0	0	2.8	0	0.994416	41.167332	154	1	1.1	0.9;
	7186	2	0	0	0	0	0	0.962211	47.074823	6.3	1	1.1	0.9;
	7187	1	-0	0	0	0.88	0	1.003062	-0.954541	110	3	1.1	0.9;
	7188	1	1.9	0	0	0.5	0	0.986782	-9.062218	110	18	1.1	0.9;
	7189	1	-0	0	0	121.29	0	1.032126	26.472433	380	5	1.1	0.9;
	7190	1	96	33.3	0	16.89	0	0.992569	-4.021009	220	6	1.1	0.9;
	7191	1	17.12	5.6	0	1.12	0	0.996019	-9.034279	110	12	1.1	0.9;
	7192	1	185.6	62.3	0.118766	18.421234	0	1.011629	21.659763	220	5	1.1	0.9;
	7193	1	175.8	70.7	0	19.08	0	1.018093	-18.599118	220	4	1.1	0.9;
	7194	1	125.16	29.2	0	2.69	0	1.031969	17.968997	220	5	1.1	0.9;
	7195	1	50.06	9.9	0	11.56	0	1.033222	9.023837	220	5	1.1	0.9;
	7196	1	220.6	83.4	0	17.73	0	1.042863	4.73746	380	20	1.1	0.9;
	7197	1	0	-0	0	1.37	0	1.028985	11.966174	220	1	1.1	0.9;
	7198	1	0	0	0	0.95	0	1.033128	-19.384685	220	4	1.1	0.9;
	7199	1	-0	-0	0	1.09	0	1.050633	-5.746947	150	9	1.1	0.9;
	7200	2	0	0	0	0	0	1.026147	8.391166	20	6	1.1	0.9;
	7201	1	-0	-0	0	0.45	0	1.042328	4.557376	220	20	1.1	0.9;
	7202	1	-24.03	7.9	0	0.25	0	1.021843	-4.315515	150	9	1.1	0.9;
	7203	1	14.92	-8.73	0	34.66	0	1.029341	7.526066	220	5	1.1	0.9;
	7204	1	-0	0	0	0.31	0	1.003398	-10.072574	150	9	1.1	0.9;
	7205	1	10.51	4.7	0	1.54	0	0.983301	1.387697	110	18	1.1	0.9;
	7206	1	71.6	10.2	0	58.47	0	0.97201	10.933304	400	1	1.1	0.9;
	7207	1	13.11	4.9	0.038998	12.341002	0	1.047655	20.267424	220	5	1.1	0.9;
	7208	1	15.82	9.8	0	7.61	0	1.055797	-6.166252	150	9	1.1	0.9;
	7209	1	-0	-0	0	145.62	0	0.960568	24.073792	400	1	1.1	0.9;
	7210	1	-0	-0	0	1.17	0	1.04686	27.640783	150	8	1.1	0.9;
	7211	1	-0	0	0	18.54	0	1.016765	34.240821	380	3	1.1	0.9;
	7212	1	0	0	0	2.57	0	1.041205	12.505171	380	1	1.1	0.9;
	7213	2	0	0	0	0	0	1.014795	23.808394	6.3	1	1.1	0.9;
	7214	1	0	0	0	26	0	1.018912	10.542087	380	1	1.1	0.9;
	7215	1	0	0	0	94.945914	0	1.026561	-12.445623	380	17	1.1	0.9;
	7216	1	195.3	87.6	0	9.53	0	0.987802	33.849156	220	3	1.1	0.9;
	7217	1	3.8	1.1	0	10.69	0	1.044684	-14.421225	380	10	1.1	0.9;
	7218	1	46.2	6.6	0	1.24	0	0.991802	1.432003	154	1	1.1	0.9;
	7219	1	-0	-0	0	98.2	0	1.031691	16.4694	380	5	1.1	0.9;
	7220	1	40.04	10	0	23.59	0	1.00867	-17.214347	380	6	1.1	0.9;
	7221	1	37	8.7	0.075047	50.85006	0	1.042167	34.593988	380	2	1.1	0.9;
	7222	1	2.5	1.3	0	0.16	0	0.995355	-3.371498	110	18	1.1	0.9;
	7223	1	-0	-0	0.049332	27.640668	0	1.03301	-15.651629	380	4	1.1	0.9;
	7224	1	0	0	0	0	0	0.986074	-5.521684	150	9	1.1	0.9;
	7225	1	-19.22	6.3	0	1.02	0	0.981434	-8.220208	150	9	1.1	0.9;
	7226	1	22.42	7.3	0	0.77	0	0.999121	-13.152442	110	12	1.1	0.9;
	7227	1	-0	0	0	15.76	0	1.009223	-5.815173	220	12	1.1	0.9;
	7228	2	0	0	0	0	0	1.053647	50.613884	20	3	1.1	0.9;
	7229	1	124.5	29.6	0	15.4	0	1.039036	18.150633	220	5	1.1	0.9;
	7230	1	11.61	2.5	0	0.42	0	1.023464	-6.251097	110	3	1.1	0.9;
	7231	1	-0	0	1.258518	11.87411	0	1.043674	25.996716	380	7	1.1	0.9;
	7232	1	56.76	31.1	0	0.37	0	1.045468	24.966619	220	5	1.1	0.9;
	7233	2	0	0	0	0	0	1.038999	38.995592	20	2	1.1	0.9;
	7234	1	0	0	0	5.5	0	0.926718	-12.042081	150	9	1.1	0.9;
	7235	1	59.67	25.6	0	11.54	0	0.995493	32.22324	220	3	1.1	0.9;
	7236	2	0	0	0	0	0	1.007821	10.245998	20	18	1.1	0.9;
	7237	1	109.9	33.6	0	3.72	0	1.016923	16.575462	220	5	1.1	0.9;
	7238	1	0	-0	0	0.83	0	1.03456	18.420655	380	6	1.1	0.9;
	7239	1	36.84	14	0.020146	0.269854	0	0.982727	-1.783019	110	19	1.1	0.9;
	7240	2	0	0	0	0	0	1.043764	37.173377	20	5	1.1	0.9;
	7241	1	0.1	35.9	0.29124	18.2273	0	1.039685	-13.261568	380	4	1.1	0.9;
	7242	2	0	0	0	0	0	1.027264	14.46311	20	14	1.1	0.9;
	7243	2	0	0	0	0	0	0.958146	44.224727	6.3	1	1.1	0.9;
	7244	1	329.7	220.4	0	46.92	0	1.006661	14.689505	380	3	1.1	0.9;
	7245	1	-0	-0	0	20.79	0	1.003591	25.364085	380	3	1.1	0.9;
	7246	1	72.18	44.9	0	6.34	0	1.019342	37.26415	150	8	1.1	0.9;
	7247	1	-0	0	0.057201	0.402799	0	1.040082	26.616011	110	18	1.1	0.9;
	7248	1	11.2	1.6	0	4	0	0.968489	23.525802	154	1	1.1	0.9;
	7249	2	0	0	0	0	0	1.005915	0.12691	20	18	1.1	0.9;
	7250	2	0	0	0	0	0	1.020233	20.547626	11	1	1.1	0.9;
	7251	1	67.37	31.4	0.039096	2.300904	0	1.030935	11.629416	220	5	1.1	0.9;
	7252	2	0	0	0	0	0	1.037622	28.736858	20	14	1.1	0.9;
	7253	1	66.77	11.1	0	2.7	0	1.013702	-24.444263	220	4	1.1	0.9;
	7254	1	52.76	5.9	0.059373	5.140627	0	1.017332	-16.047972	220	4	1.1	0.9;
	7255	1	27.9	4	0	5.34	0	1.025494	17.288026	154	1	1.1	0.9;
	7256	1	-0	-0	0	0.99	0	1.031224	29.555502	220	1	1.1	0.9;
	7257	2	0	0	0	0	0	1.03161	11.84227	11	1	1.1	0.9;
	7258	1	1.8	91.2	0.703792	89.782606	0	1.028323	-11.00057	380	4	1.1	0.9;
	7259	1	0	-0	0	1.94	0	1.022177	-18.166654	220	4	1.1	0.9;
	7260	1	-119.4	-359.4	0	39.39	0	1.067355	5.985102	400	1	1.1	0.9;
	7261	2	0	0	0	0	0	0.984601	26.113186	6.3	1	1.1	0.9;
	7262	2	0	0	0	0	0	0.991024	-5.15543	20	22	1.1	0.9;
	7263	1	12.91	7.1	0	0.84	0	0.956518	-5.487802	110	18	1.1	0.9;
	7264	1	62.97	68	0	13.59	0	0.927744	7.78192	220	16	1.1	0.9;
	7265	1	29.13	18	0	0.06	0	0.990005	-9.648284	150	9	1.1	0.9;
	7266	1	-0	-0	0	0.29	0	0.997837	-9.08488	150	9	1.1	0.9;
	7267	2	0	0	0	0	0	0.996117	32.042491	11	1	1.1	0.9;
	7268	1	-0	-0	0	29.32	0	1.02094	37.983713	380	8	1.1	0.9;
	7269	1	0	0	0	0	0	1.024048	-7.803527	150	9	1.1	0.9;
	7270	1	110.7	78.7	0	5.27	0	1.002139	-14.467046	220	17	1.1	0.9;
	7271	1	292	81.8	0.275203	1.863598	0	1.028435	-14.827023	220	4	1.1	0.9;
	7272	1	-0	-0	0	16.57	0	1.044486	33.222726	380	1	1.1	0.9;
	7273	1	0	-0	0.471049	0.717086	0	1.058674	18.383846	380	14	1.1	0.9;
	7274	1	0	-0	0	171.844376	0	1.055938	11.437944	380	5	1.1	0.9;
	7275	1	130	128	0	7.73	0	1.011321	14.23252	220	3	1.1	0.9;
	7276	1	13.9	1.9	0	4.71	0	1.005505	17.153822	154	1	1.1	0.9;
	7277	1	0	0	0	0	0	1.032532	-0.066659	150	9	1.1	0.9;
	7278	1	268.2	-257.9	0	42.56	0	1.03347	34.98236	380	3	1.1	0.9;
	7279	1	0	0	0	0	0	0.90264	-12.097367	150	9	1.1	0.9;
	7280	1	-15.89	-6.42	0	38.662969	0	0.936041	10.852081	380	5	1.1	0.9;
	7281	1	-0	0	0.181549	24.230889	0	1.051879	-6.358399	220	18	1.2	0.8;
	7282	2	0	0	0	0	0	1.041421	-15.253203	20	5	1.1	0.9;
	7283	1	-0	0	0	0	0	1.03499	-19.617451	20	5	1.1	0.9;
	7284	1	-0	0	0	0	0	1.03499	-19.617451	20	5	1.1	0.9;
	7285	1	0	0	0	0.31	0	0.964991	-9.63915	150	9	1.1	0.9;
	7286	1	-0	0	0	0	0	1.03499	-19.617451	20	5	1.1	0.9;
	7287	1	-0	0	0	0	0	1.03499	-19.617451	20	5	1.1	0.9;
	7288	1	-0	-0	0.198366	-0.008774	0	1.026857	15.275337	380	6	1.1	0.9;
	7289	2	0	0	0	0	0	1.033768	14.596978	20	14	1.1	0.9;
	7290	1	4.81	1.6	0	0.53	0	0.986547	-15.19694	110	12	1.1	0.9;
	7291	2	0	0	0	0	0	1.00571	1.688469	20	2	1.1	0.9;
	7292	1	0	-0	0	2.34	0	1.012463	27.539428	154	1	1.1	0.9;
	7293	1	0	0	0	16.87	0	1.034238	-1.125368	154	1	1.1	0.9;
	7294	1	88.4	-5.9	0	3.32	0	1.029767	-14.92546	220	4	1.1	0.9;
	7295	1	10.11	1.6	0	0.66	0	1.003923	-3.319875	110	18	1.1	0.9;
	7296	1	9.31	0.8	0	0.9	0	1.001715	-1.126851	110	3	1.1	0.9;
	7297	2	0	0	0	0	0	1.034048	-9.733404	20	4	1.1	0.9;
	7298	1	32.5	30.5	0	265.721081	0	1.112059	-0.97463	380	9	1.2	0.8;
	7299	1	-22.8	-16.4	0	0.19	0	1.052695	-2.176957	110	21	1.1	0.9;
	7300	1	-4.66	1.11	0	0.19	0	1.012727	-2.819685	110	18	1.1	0.9;
	7301	1	11.41	3.8	0	1.01	0	1.017988	3.774778	110	18	1.1	0.9;
	7302	1	-4.89	-2.4	0	0.25	0	0.991789	-2.334341	110	18	1.1	0.9;
	7303	1	0	0	0.090774	3.730444	0	1.041313	-3.543238	220	18	1.1	0.9;
	7304	2	0	0	0	0	0	1.03036	-6.512869	20	4	1.1	0.9;
	7305	1	501.1	10.7	0	18.07	0	1.010597	-16.590721	380	6	1.1	0.9;
	7306	1	58.36	42.7	0	1.29	0	0.939342	5.315842	110	18	1.1	0.9;
	7307	1	-0.12	-1.74	0	0	0	1.100006	65.129088	380	8	1.2	0.8;
	7308	1	-0	-0	0.232315	-0.193596	0	1.0267	1.766412	380	6	1.1	0.9;
	7309	1	190.7	13.3	0	4.61	0	0.986662	-26.829415	220	6	1.1	0.9;
	7310	1	-0	0	0.077493	3.252507	0	1.012714	-7.579063	110	12	1.1	0.9;
	7311	1	0	0	0	0.06	0	1.032903	-3.01883	220	21	1.1	0.9;
	7312	1	110	6.83	0	13.51	0	1.005455	17.595107	220	14	1.1	0.9;
	7313	1	-0	-0	0.090774	2.110444	0	1.028058	-1.961635	220	18	1.1	0.9;
	7314	1	-28.14	-0.76	0	0.1	0	1.058635	17.90213	220	5	1.1	0.9;
	7315	1	53.86	33.4	0	12.95	0	1.005468	-3.810857	220	6	1.1	0.9;
	7316	1	27.03	13	0	2.32	0	1.006171	0.370886	220	6	1.1	0.9;
	7317	2	0	0	0	0	0	1.029166	13.522836	20	14	1.1	0.9;
	7318	1	82.49	24.5	0	0.62	0	1.033283	46.248802	110	8	1.1	0.9;
	7319	1	-0	0	0	4.34	0	1.022762	24.657166	220	5	1.1	0.9;
	7320	1	17.87	-8.59	0	0.21	0	1.06258	21.454963	220	5	1.1	0.9;
	7321	1	-0	0	0.120181	7.369819	0	1.018001	13.215213	220	5	1.1	0.9;
	7322	1	-0	0	0.287116	4.880736	0	1.05612	81.744333	380	3	1.1	0.9;
	7323	1	-58.77	18.7	0	4.23	0	0.982662	-9.442097	150	10	1.1	0.9;
	7324	1	-0	0	0	40.291745	0	1.002311	14.090329	380	5	1.1	0.9;
	7325	1	0	-0	0	23.76	0	1.012062	38.612315	380	8	1.1	0.9;
	7326	1	0	0	0.27075	87.43435	0	1.080819	8.237326	380	18	1.2	0.8;
	7327	2	0	0	0	0	0	1.020032	8.963657	11	1	1.1	0.9;
	7328	1	0	-0	0	45.86	0	1.05717	53.317776	380	13	1.1	0.9;
	7329	2	0	0	0	0	0	0.992436	31.623055	11	1	1.1	0.9;
	7330	1	-0	0	0	11.27	0	1.0366	38.640875	380	3	1.1	0.9;
	7331	1	5.11	2.5	0	0.52	0	0.955313	-6.820872	110	18	1.1	0.9;
	7332	1	5.11	3.8	0	0.88	0	0.961011	-6.504868	110	18	1.1	0.9;
	7333	2	0	0	0	0	0	1.042865	51.92649	20	3	1.1	0.9;
	7334	2	0	0	0	0	0	1.033239	35.570464	20	10	1.1	0.9;
	7335	1	0	0	0	1.15	0	0.962918	-9.819129	150	9	1.1	0.9;
	7336	1	70.08	35	0	1.45	0	1.010435	22.936576	220	6	1.1	0.9;
	7337	1	27.13	9.8	0	0.17	0	0.96354	-4.229015	110	19	1.1	0.9;
	7338	2	0	0	0	0	0	0.999789	98.512855	20	3	1.1	0.9;
	7339	1	0	0	0	8.155532	0	0.992745	24.126033	380	5	1.1	0.9;
	7340	2	0	0	0	0	0	0.938623	24.03983	10.5	1	1.1	0.9;
	7341	1	0.1	-9.6	0	1.3	0	1.023025	8.747596	220	3	1.1	0.9;
	7342	2	0	0	0	0	0	1.044097	40.791392	20	10	1.1	0.9;
	7343	1	159.27	38.31	0	2.18	0	1.019982	14.451272	220	5	1.1	0.9;
	7344	1	0	-0	0.03846	5.16154	0	1.042851	-3.461128	220	1	1.1	0.9;
	7345	2	0	0	0	0	0	1.030841	-16.24512	20	4	1.1	0.9;
	7346	2	0	0	0	0	0	1.027328	-11.116661	20	4	1.1	0.9;
	7347	1	-0	0	0	29.76	0	0.979537	-10.799228	220	22	1.1	0.9;
	7348	1	-0	-0	0	2.81	0	0.990307	46.2101	154	1	1.1	0.9;
	7349	2	0	0	0	0	0	1.040292	-9.24091	20	4	1.1	0.9;
	7350	2	0	0	0	0	0	1.040292	-9.24091	20	4	1.1	0.9;
	7351	2	0	0	0	0	0	1.036065	-16.582357	20	4	1.1	0.9;
	7352	2	0	0	0	0	0	1.029893	-6.546635	20	4	1.1	0.9;
	7353	1	162.6	36.4	0	26.03	0	1.002965	-20.322256	220	4	1.1	0.9;
	7354	2	0	0	0	0	0	1.024274	-9.95923	20	4	1.1	0.9;
	7355	1	130.1	40.5	0.112754	0.266039	0	1.040326	2.497708	110	19	1.1	0.9;
	7356	2	0	0	0	0	0	1.038438	0.065623	20	9	1.1	0.9;
	7357	1	38.1	5.5	0	3.32	0	1.005748	16.065016	154	1	1.1	0.9;
	7358	1	-12.92	-1.53	0	0.96	0	1.018386	8.832269	110	18	1.1	0.9;
	7359	1	30.33	18.8	0	1.35	0	0.965144	-10.373275	150	9	1.1	0.9;
	7360	1	0	0	0	1.31	0	1.019462	-18.827651	220	4	1.1	0.9;
	7361	1	0	0	0	0	0	1.04035	8.357264	20	5	1.1	0.9;
	7362	1	174.7	35	0	2.14	0	1.008054	30.667304	150	8	1.1	0.9;
	7363	1	14.32	4.4	0	0.74	0	0.992259	-8.136274	110	22	1.1	0.9;
	7364	1	18.9	2.7	0	2.81	0	1.009651	13.007592	154	1	1.1	0.9;
	7365	2	0	0	0	0	0	1.037973	11.471538	20	14	1.1	0.9;
	7366	1	93.1	43.7	0	11.87	0	1.02223	-11.03513	150	9	1.1	0.9;
	7367	1	0	0	0	13.25	0	1.014062	-14.687051	380	1	1.1	0.9;
	7368	1	-0	0	0.038705	1.011295	0	1.060019	0.350257	110	15	1.1	0.9;
	7369	2	0	0	0	0	0	1.005303	-3.689085	20	2	1.1	0.9;
	7370	2	0	0	0	0	0	1.007216	-1.420425	20	2	1.1	0.9;
	7371	2	0	0	0	0	0	1.043492	-15.322718	20	4	1.1	0.9;
	7372	2	0	0	0	0	0	1.007216	-1.420425	20	2	1.1	0.9;
	7373	1	20.32	10.1	0	0.14	0	0.950412	-5.365339	110	18	1.1	0.9;
	7374	2	0	0	0	0	0	1.005303	-3.689085	20	2	1.1	0.9;
	7375	2	0	0	0	0	0	1.005303	-3.689085	20	2	1.1	0.9;
	7376	2	0	0	0	0	0	1.005303	-3.689085	20	2	1.1	0.9;
	7377	2	0	0	0	0	0	1.001397	94.4747	20	3	1.1	0.9;
	7378	2	0	0	0	0	0	1.004934	-5.075181	20	2	1.1	0.9;
	7379	2	0	0	0	0	0	1.004934	-5.075181	20	2	1.1	0.9;
	7380	1	0	8.8	0	18.87	0	1.076757	-9.360128	150	9	1.2	0.8;
	7381	2	0	0	0	0	0	0.976475	43.406918	15.5	1	1.1	0.9;
	7382	1	24.33	10	0	84.57	0	1.019801	-22.440546	380	4	1.1	0.9;
	7383	1	303.1	75.5	0	22.74	0	1.006546	-11.697125	380	6	1.1	0.9;
	7384	1	-0	0	0	12.01	0	1.049348	-12.689887	380	1	1.1	0.9;
	7385	1	11.41	2.5	0	0.22	0	0.986576	2.287295	110	18	1.1	0.9;
	7386	2	0	0	0	0	0	1.058287	5.521142	20	5	1.1	0.9;
	7387	1	0	0	0.073897	1.024333	0	1.005095	92.428959	220	1	1.1	0.9;
	7388	2	0	0	0	0	0	1.038881	6.451637	20	22	1.1	0.9;
	7389	1	0	0	0	39.95	0	1.00316	23.070314	380	5	1.1	0.9;
	7390	1	4.61	1.3	0	0.53	0	0.984518	-6.510386	110	18	1.1	0.9;
	7391	2	0	0	0	0	0	1.025019	-12.264635	20	6	1.1	0.9;
	7392	2	0	0	0	0	0	1.055337	0.681282	20	9	1.1	0.9;
	7393	1	0	-0	0	9.59	0	1.06184	80.335931	380	3	1.1	0.9;
	7394	1	-0	-0	0	12.96	0	1.02091	-7.970929	380	6	1.1	0.9;
	7395	1	-0	-0	0.181549	18.620889	0	1.043549	3.22096	220	18	1.1	0.9;
	7396	1	21.32	7.9	0.081251	7.368749	0	1.011306	13.237271	220	5	1.1	0.9;
	7397	1	0	0	0	5.173673	0	0.971662	4.274204	220	24	1.1	0.9;
	7398	1	190	72.7	0	30.78	0	0.993179	-4.309492	220	6	1.1	0.9;
	7399	1	43.45	26.9	0	6.37	0	1.049444	-6.481726	150	9	1.1	0.9;
	7400	1	31.74	17.4	0	0.89	0	1.0494	25.484796	220	5	1.1	0.9;
	7401	1	0	0	0.263214	3.036617	0	0.992535	-0.937169	220	19	1.1	0.9;
	7402	1	62.7	9	0	4.93	0	1.007335	16.489323	154	1	1.1	0.9;
	7403	1	0	0	0	0.06	0	0.988449	-4.637233	110	21	1.1	0.9;
	7404	1	295.2	39.3	0	7.3	0	1.014322	-16.885153	220	4	1.1	0.9;
	7405	1	9.51	0.5	0	0.04	0	1.025141	-10.524427	150	9	1.1	0.9;
	7406	1	0	0	0.11885	-0.089137	0	1.015343	-1.615945	380	6	1.1	0.9;
	7407	1	-0	0	0.029104	22.310896	0	1.049351	-13.067891	380	4	1.1	0.9;
	7408	1	25.33	14.2	0	4.05	0	1.029262	7.522386	220	5	1.1	0.9;
	7409	1	160.6	51.2	0	46.39	0	1.064417	0.577932	380	20	1.1	0.9;
	7410	1	0	0	0	0	0	0.974625	-8.32398	150	9	1.1	0.9;
	7411	1	22.32	7.3	0	0.46	0	0.987278	-11.720559	110	12	1.1	0.9;
	7412	1	0	-0	0.670766	-0.531023	0	1.07921	54.070288	380	8	1.2	0.8;
	7413	1	35.64	22	0	4.46	0	0.92416	-12.234897	150	9	1.1	0.9;
	7414	1	0	0	0.404001	4.446629	0	1.010408	22.827203	380	5	1.1	0.9;
	7415	1	3.8	1.3	0	0.78	0	0.995248	-4.048539	110	18	1.1	0.9;
	7416	2	0	0	0	0	0	1.003342	23.89027	10.6	1	1.1	0.9;
	7417	1	-19.07	-6.06	0	16.68	0	1.004325	-5.281581	220	11	1.1	0.9;
	7418	1	50.06	25	0	1.21	0	0.979437	-10.944319	110	12	1.1	0.9;
	7419	1	36.44	22.6	0	2.32	0	0.978181	-9.34911	150	9	1.1	0.9;
	7420	1	-0	-0	0	22.93	0	1.103483	-15.025503	150	9	1.2	0.8;
	7421	1	49.8	7.1	0	1.4	0	1.013945	18.92959	154	1	1.1	0.9;
	7422	1	5.11	2.5	0	1.5	0	0.930277	-7.416239	110	18	1.1	0.9;
	7423	1	148.3	12.1	0.039183	7.280613	0	1.017402	17.981879	220	13	1.1	0.9;
	7424	1	19.12	2.8	0	0.97	0	0.975147	-7.162476	110	18	1.1	0.9;
	7425	1	-0	0	0.037961	6.562039	0	1.047892	7.785203	220	11	1.1	0.9;
	7426	2	0	0	0	0	0	1.030874	24.601888	20	5	1.1	0.9;
	7427	1	-0	0	0	0.04	0	1.032836	-10.828842	150	9	1.1	0.9;
	7428	2	0	0	0	0	0	1.049495	12.649819	20	5	1.1	0.9;
	7429	1	3.7	1.3	0	0.98	0	0.927616	-1.873772	110	19	1.1	0.9;
	7430	2	0	0	0	0	0	0.991002	22.274075	10.6	1	1.1	0.9;
	7431	1	80.39	10.8	0	0.35	0	1.027954	19.079635	220	5	1.1	0.9;
	7432	1	53.26	29.5	0	41.35	0	1.023631	36.965167	150	8	1.1	0.9;
	7433	1	6.61	3.3	0	2.67	0	0.993646	-13.107606	110	22	1.1	0.9;
	7434	1	-0	-0	0	-0.46	0	1.009631	-16.516918	380	6	1.1	0.9;
	7435	1	7.41	2.7	0	1.08	0	1.000291	1.096095	110	19	1.1	0.9;
	7436	1	-14.81	-0.52	0	0.13	0	1.045147	8.735036	110	18	1.1	0.9;
	7437	1	87.38	-90.15	0	16.05	0	1.027684	6.13627	220	3	1.1	0.9;
	7438	1	155.25	23.3	0	3.74	0	0.980437	-5.788214	220	6	1.1	0.9;
	7439	1	-0	-0	0	4.74	0	1.019063	27.899716	154	1	1.1	0.9;
	7440	1	71.58	57.6	0.233839	24.774621	0	1.033627	-11.381318	380	4	1.1	0.9;
	7441	1	0	-0	0.278965	11.049132	0	1.059512	17.119921	380	16	1.1	0.9;
	7442	1	-29.96	-21.31	0	0	0	1.035887	43.5707	150	8	1.1	0.9;
	7443	1	0	-0	0	5.02	0	1.022832	-6.905618	220	10	1.1	0.9;
	7444	1	-0	-0	0	11.55	0	0.983585	27.331059	380	1	1.1	0.9;
	7445	1	0	0	0	0	0	1.001491	-13.031617	150	9	1.1	0.9;
	7446	1	-2.27	-1.81	0	42.76	0	1.027329	-11.117382	380	4	1.1	0.9;
	7447	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	7448	1	0	0	0	38.24	0	0.965783	34.653221	400	1	1.1	0.9;
	7449	2	0	0	0	0	0	0.990475	21.129805	15.75	1	1.1	0.9;
	7450	1	0	-0	0	9.34	0	1.041428	31.297695	220	5	1.1	0.9;
	7451	1	117.7	115	0	12.29	0	1.033239	19.752155	380	3	1.1	0.9;
	7452	2	0	0	0	0	0	1.035023	-10.739655	20	4	1.1	0.9;
	7453	1	0	-0	0.091693	56.645921	0	1.00276	92.867423	380	1	1.1	0.9;
	7454	1	99.4	-347.1	0	31.86	0	1.039213	59.560635	380	3	1.1	0.9;
	7455	1	10.81	1.9	0	1.6	0	0.970971	8.339669	110	18	1.1	0.9;
	7456	1	0	-0	0	9.06	0	1.026603	8.287587	220	3	1.1	0.9;
	7457	1	37	5.3	0	1.74	0	0.995497	16.901519	154	1	1.1	0.9;
	7458	1	39.94	13	0	2.47	0	1.034461	-2.846498	220	5	1.1	0.9;
	7459	1	122.9	19.3	0	68.84	0	1.048234	8.13578	380	16	1.1	0.9;
	7460	1	-55.83	-31.99	0	2.88	0	1.043896	46.673508	110	8	1.1	0.9;
	7461	1	213.1	-12.3	0	30.87	0	1.026298	3.000376	380	6	1.1	0.9;
	7462	1	94.71	15.6	0.040271	2.879729	0	0.996721	18.292833	220	5	1.1	0.9;
	7463	1	139.9	1.1	0	5.58	0	1.001698	-25.302194	220	4	1.1	0.9;
	7464	1	305.04	77.16	0	17.43	0	1.023246	-4.014324	220	5	1.1	0.9;
	7465	1	41.35	25.6	0	9.9	0	1.023776	-5.098586	150	9	1.1	0.9;
	7466	1	0	-0	0	11.9	0	1.027457	0.392065	154	1	1.1	0.9;
	7467	1	0	0	0	3.41	0	1.006837	21.320742	220	6	1.1	0.9;
	7468	1	2.47	-0.62	0	2.54	0	0.996113	91.68458	150	1	1.1	0.9;
	7469	1	35.64	22	0	5.67	0	0.926519	-12.060065	150	9	1.1	0.9;
	7470	2	0	0	0	0	0	1.060213	11.478706	20	5	1.1	0.9;
	7471	1	46.25	13.1	0	1.29	0	1.030405	-19.337916	220	4	1.1	0.9;
	7472	1	0	0	0	1.23	0	1.02918	34.311008	220	3	1.1	0.9;
	7473	2	0	0	0	0	0	1.015521	7.836059	11	1	1.1	0.9;
	7474	2	0	0	0	0	0	1.015682	32.523561	20	10	1.1	0.9;
	7475	2	0	0	0	0	0	1.043764	37.173377	20	5	1.1	0.9;
	7476	2	0	0	0	0	0	1.015161	-11.543898	20	4	1.1	0.9;
	7477	2	0	0	0	0	0	1.015161	-11.543898	20	4	1.1	0.9;
	7478	1	48.95	11.6	0	2	0	1.023844	10.180784	220	5	1.1	0.9;
	7479	1	23.4	8.38	0	1.07	0	1.01549	-16.739813	220	4	1.1	0.9;
	7480	1	163.1	23.2	0	3.36	0	1.026848	-2.172567	154	1	1.1	0.9;
	7481	1	-15.91	-6.11	0	0.21	0	0.986796	-8.205739	110	23	1.1	0.9;
	7482	1	0	0	0	0.66	0	1.012673	-2.175699	380	6	1.1	0.9;
	7483	1	123.9	46.5	0.039444	12.660556	0	1.006165	19.267453	220	13	1.1	0.9;
	7484	1	17.82	7.3	0	0.14	0	0.987767	-4.687808	110	18	1.1	0.9;
	7485	1	66.47	22.1	0	0.46	0	1.000671	-7.208347	110	21	1.1	0.9;
	7486	1	-0	-0	0	2.74	0	0.98001	31.417055	220	3	1.1	0.9;
	7487	2	0	0	0	0	0	1.031445	23.624078	11	1	1.1	0.9;
	7488	1	-0	0	0	1.23	0	1.004387	30.940233	220	3	1.1	0.9;
	7489	2	0	0	0	0	0	0.969605	29.504132	19	1	1.1	0.9;
	7490	2	0	0	0	0	0	1.028453	-12.903077	20	4	1.1	0.9;
	7491	1	-0	0	0	4.86	0	1.022918	-6.867161	220	10	1.1	0.9;
	7492	1	572.3	159.7	0	19.88	0	1.0082	-3.889166	380	6	1.1	0.9;
	7493	1	-0	0	0	7.84	0	1.024005	3.431686	220	5	1.1	0.9;
	7494	1	236.2	45.1	0	2.88	0	0.976666	29.531177	220	3	1.1	0.9;
	7495	1	12.31	3	0	0.98	0	0.966785	-7.214159	110	18	1.1	0.9;
	7496	1	-0	-0	0	0.16	0	1.066456	-4.571785	150	9	1.1	0.9;
	7497	1	0	-0	0	0.93	0	0.971858	-3.874575	110	18	1.1	0.9;
	7498	2	0	0	0	0	0	1.056333	76.025167	20	3	1.1	0.9;
	7499	2	0	0	0	0	0	1.074058	4.909711	20	20	1.1	0.9;
	7500	1	52.96	32.8	0	5.79	0	1.004576	-15.207244	150	9	1.1	0.9;
	7501	2	0	0	0	0	0	1.050494	10.971923	20	5	1.1	0.9;
	7502	2	0	0	0	0	0	1.044704	16.811399	20	5	1.1	0.9;
	7503	1	16.92	-0	0	0.57	0	1.042464	13.256191	220	5	1.1	0.9;
	7504	2	0	0	0	0	0	1.044704	16.811399	20	5	1.1	0.9;
	7505	2	0	0	0	0	0	1.033648	19.100507	20	5	1.1	0.9;
	7506	1	0	-0	0.234321	-0.195267	0	1.03025	4.714884	380	6	1.1	0.9;
	7507	2	0	0	0	0	0	1.050494	10.971923	20	5	1.1	0.9;
	7508	2	0	0	0	0	0	1.036203	12.985596	20	5	1.1	0.9;
	7509	1	-0	0	0	0.19	0	1.036636	21.536243	220	5	1.1	0.9;
	7510	1	26.35	30	0	1.82	0	1.026508	-9.702337	150	9	1.1	0.9;
	7511	2	0	0	0	0	0	1.016367	7.342385	20	22	1.1	0.9;
	7512	2	0	0	0	0	0	1.041748	29.3468	20	5	1.1	0.9;
	7513	2	0	0	0	0	0	1.050204	7.926973	20	5	1.1	0.9;
	7514	1	0	0	0	1.66	0	1.020185	-9.133094	150	9	1.1	0.9;
	7515	1	-0	-0	1.175023	84.993965	0	1.015134	26.112226	380	5	1.1	0.9;
	7516	1	0	0	0.714478	10.088417	0	1.050319	7.519611	220	3	1.1	0.9;
	7517	1	229.9	48.9	0	8.36	0	1.009036	-4.581287	220	5	1.1	0.9;
	7518	1	292.8	114.2	0	22.01	0	1.011944	-15.886414	380	6	1.1	0.9;
	7519	1	-0	-0	0	17.81	0	1.04144	33.928864	380	1	1.1	0.9;
	7520	1	1.63	8.8	0	0.16	0	0.966985	-9.629605	150	9	1.1	0.9;
	7521	1	0.2	1.6	0	0.23	0	0.997786	37.196075	220	3	1.1	0.9;
	7522	1	-0	0	0	1.33	0	0.99887	-8.910757	150	9	1.1	0.9;
	7523	2	0	0	0	0	0	1.046478	33.359815	20	3	1.1	0.9;
	7524	2	0	0	0	0	0	1.046478	33.359815	20	3	1.1	0.9;
	7525	1	0	0	0	0.32	0	1.051255	-11.661899	150	9	1.1	0.9;
	7526	2	0	0	0	0	0	1.046478	33.359815	20	3	1.1	0.9;
	7527	1	0	0	0.037754	6.272246	0	1.075024	46.081157	220	3	1.1	0.9;
	7528	2	0	0	0	0	0	1.046478	33.359815	20	3	1.1	0.9;
	7529	1	18.52	6.1	0	1.24	0	0.927084	-10.698244	110	21	1.1	0.9;
	7530	1	82.39	12.8	0.039396	0.060453	0	1.011686	94.331837	150	1	1.1	0.9;
	7531	2	0	0	0	0	0	1.048637	18.752224	20	5	1.1	0.9;
	7532	1	0	0	0	0	0	1.016071	-5.447511	150	9	1.1	0.9;
	7533	1	-0	-0	0	0.23	0	1.009464	-1.490677	110	18	1.1	0.9;
	7534	1	48.35	62.8	0.460556	14.875393	0	1.048061	37.584997	220	13	1.1	0.9;
	7535	2	0	0	0	0	0	1.04554	8.581299	10.5	1	1.1	0.9;
	7536	1	21.12	7.7	0	0.52	0	1.033157	12.58568	110	19	1.1	0.9;
	7537	1	0	0	0.094757	-0.276956	0	1.001278	5.844525	220	19	1.1	0.9;
	7538	1	-0	-0	0.252685	6.907589	0	0.974828	-2.630234	220	19	1.1	0.9;
	7539	2	0	0	0	0	0	1.026207	-11.793103	20	4	1.1	0.9;
	7540	2	0	0	0	0	0	1.016705	-7.501421	20	22	1.1	0.9;
	7541	1	55.26	-0	0	1.05	0	1.024052	-20.491624	220	4	1.1	0.9;
	7542	1	-11.81	-12	0	19.39	0	1.046479	-13.228828	220	4	1.1	0.9;
	7543	1	-0	-0	0.390917	96.684236	0	1.079263	63.417816	380	3	1.1	0.9;
	7544	2	0	0	0	0	0	1.031898	-9.317211	20	6	1.1	0.9;
	7545	1	0	-0	0	35.06	0	1.023406	33.215649	380	8	1.1	0.9;
	7546	1	0	-0	0	0.48	0	0.944594	-10.932082	150	9	1.1	0.9;
	7547	2	0	0	0	0	0	1.045375	17.213843	20	5	1.1	0.9;
	7548	1	12.01	0.7	0	0.5	0	1.03676	80.007231	220	3	1.1	0.9;
	7549	1	-0	-0	0	0.45	0	0.973721	-10.511526	110	12	1.1	0.9;
	7550	1	82.6	11.7	0	1.75	0	1.033803	2.408169	154	1	1.1	0.9;
	7551	2	0	0	0	0	0	1.044493	17.115248	20	7	1.1	0.9;
	7552	1	106.8	28.8	0.0502	4.2998	0	1.01034	-19.654143	220	4	1.1	0.9;
	7553	1	3	1.7	0	0.53	0	1.050063	1.790288	110	23	1.1	0.9;
	7554	1	182.18	36.4	0	7.79	0	0.966012	-26.182023	220	6	1.1	0.9;
	7555	1	0	0	0	8.19	0	1.03287	9.234842	220	5	1.1	0.9;
	7556	2	0	0	0	0	0	0.979015	43.42213	10.3	1	1.1	0.9;
	7557	2	0	0	0	0	0	1.017287	26.369278	20	5	1.1	0.9;
	7558	2	0	0	0	0	0	1.017287	26.369278	20	5	1.1	0.9;
	7559	2	0	0	0	0	0	1.017287	26.369278	20	5	1.1	0.9;
	7560	2	0	0	0	0	0	1.017287	26.369278	20	5	1.1	0.9;
	7561	2	0	0	0	0	0	1.050891	3.953704	20	18	1.1	0.9;
	7562	2	0	0	0	0	0	1.026207	28.927907	20	5	1.1	0.9;
	7563	2	0	0	0	0	0	1.026207	28.927907	20	5	1.1	0.9;
	7564	2	0	0	0	0	0	1.048808	20.576441	20	5	1.1	0.9;
	7565	1	0	0	0.44806	95.507515	0	1.028074	9.122801	380	11	1.1	0.9;
	7566	1	175.4	28.8	0	2.03	0	0.967886	-29.550722	150	10	1.1	0.9;
	7567	1	0	0	0	10.13	0	1.031559	4.308699	220	5	1.1	0.9;
	7568	2	0	0	0	0	0	1.000449	37.338871	13.8	1	1.1	0.9;
	7569	1	39	5.5	0	3.4	0	0.9959	31.309077	154	1	1.1	0.9;
	7570	1	45.05	15.3	0	12.29	0	0.991874	-3.881717	220	6	1.1	0.9;
	7571	1	0	-0	0	0.18	0	1.029122	21.17436	220	5	1.1	0.9;
	7572	1	0	-0	0	2.18	0	1.017489	3.82958	220	1	1.1	0.9;
	7573	1	-0	0.2	0	0.12	0	1.017516	24.916537	220	3	1.1	0.9;
	7574	1	-22.2	-5.79	0	5.28	0	1.085005	72.392093	380	3	1.2	0.8;
	7575	1	0	-0	0.090774	3.340444	0	1.037518	8.166974	220	18	1.1	0.9;
	7576	1	153.9	11.5	0.565068	6.959861	0	1.036632	21.536273	220	5	1.1	0.9;
	7577	2	0	0	0	0	0	0.977562	26.05322	13.8	1	1.1	0.9;
	7578	1	10.8	1.6	0	2.88	0	0.966425	19.011458	154	1	1.1	0.9;
	7579	1	107.9	17.6	0	1.03	0	0.994266	24.822965	150	8	1.1	0.9;
	7580	1	24.33	5.8	0	13.48	0	1.032352	-19.474633	220	4	1.1	0.9;
	7581	1	-0	-0	0	0.53	0	1.006747	-6.692315	110	21	1.1	0.9;
	7582	1	67.98	9.2	0	7.77	0	1.021195	-22.64557	220	4	1.1	0.9;
	7583	1	-0	-0	0.117944	0.021542	0	1.028403	19.42587	220	6	1.1	0.9;
	7584	1	11.31	4.6	0.038844	2.201156	0	1.020593	8.251081	110	18	1.1	0.9;
	7585	2	0	0	0	0	0	0.998951	38.095582	9	1	1.1	0.9;
	7586	1	33.24	11.1	0	1.14	0	0.95133	-12.17191	110	21	1.1	0.9;
	7587	1	0	-0	0	0	0	1.018239	-19.841296	20	4	1.1	0.9;
	7588	1	35.14	12.6	0	0.97	0	1.053124	0.544081	110	15	1.1	0.9;
	7589	1	7.41	2.7	0	0.8	0	0.962323	-5.059859	110	19	1.1	0.9;
	7590	1	76.08	17.1	0	5.78	0	1.00847	2.517223	220	20	1.1	0.9;
	7591	1	35.94	26	0	8.07	0	1.037428	-2.276022	220	1	1.1	0.9;
	7592	1	85.09	13.4	0	0.23	0	1.014219	-26.230001	220	4	1.1	0.9;
	7593	2	0	0	0	0	0	0.999494	10.749804	11	1	1.1	0.9;
	7594	1	18.8	2.7	0	0.31	0	1.012367	19.105535	154	1	1.1	0.9;
	7595	1	-0	-0	0	2.84	0	1.029975	-17.836256	220	1	1.1	0.9;
	7596	1	-0	0	0	4.99	0	1.030725	8.292435	220	1	1.1	0.9;
	7597	1	11.71	3.9	0	0.77	0	0.986301	-14.390471	110	12	1.1	0.9;
	7598	2	0	0	0	0	0	1.039863	93.944189	20	3	1.1	0.9;
	7599	2	0	0	0	0	0	1.056697	84.200005	20	3	1.1	0.9;
	7600	2	0	0	0	0	0	1.034532	19.101591	20	6	1.1	0.9;
	7601	1	158.79	39.45	0	63.31	0	0.993956	27.916175	380	13	1.1	0.9;
	7602	1	347.61	74.37	0	0	0	1.052113	29.000359	220	2	1.1	0.9;
	7603	1	-0	-7.5	0	0	0	1.017601	-8.578565	150	9	1.1	0.9;
	7604	1	-0	0	0	0.2	0	1.033074	20.041137	220	5	1.1	0.9;
	7605	1	246.8	36.3	0	16.01	0	1.019087	3.567469	220	5	1.1	0.9;
	7606	1	-0	0	0	0.89	0	1.035028	23.912435	220	5	1.1	0.9;
	7607	1	12.51	4.2	0.020558	0.199442	0	0.973132	-9.209535	110	21	1.1	0.9;
	7608	1	-11.01	3.6	0	0.55	0	1.02769	0.17135	150	9	1.1	0.9;
	7609	1	-0	0	0.117914	10.521738	0	1.019595	11.893135	220	3	1.1	0.9;
	7610	1	12.11	-0	0	0.6	0	1.010451	-20.541975	220	4	1.1	0.9;
	7611	1	68.68	21	0	0.56	0	0.983582	-5.301166	110	21	1.1	0.9;
	7612	1	23.03	8.3	0	0.09	0	1.021547	-3.918499	110	19	1.1	0.9;
	7613	2	0	0	0	0	0	1.012556	-8.920589	20	10	1.1	0.9;
	7614	1	58.5	8.3	0	4	0	0.996842	30.308668	154	1	1.1	0.9;
	7615	1	29.53	25.1	0	0.8	0	0.900287	-12.257818	150	9	1.1	0.9;
	7616	1	13.9	1.9	0	3.05	0	1.007201	17.355456	154	1	1.1	0.9;
	7617	1	71.88	44.5	0	0.04	0	1.024793	-7.141759	150	9	1.1	0.9;
	7618	1	4.61	1.7	0	0.49	0	0.961786	-4.769176	110	19	1.1	0.9;
	7619	1	8.31	3	0	1.37	0	1.009246	-1.433025	110	19	1.1	0.9;
	7620	1	27.73	17.2	0	0.36	0	0.957345	-9.979812	150	9	1.1	0.9;
	7621	1	75.58	26.1	0.15448	6.28552	0	1.033359	10.826321	220	7	1.1	0.9;
	7622	1	14.5	3.1	0.154498	-0.265705	0	1.030095	1.211663	380	9	1.1	0.9;
	7623	1	1.9	1.4	0	1.07	0	0.967258	-0.3052	110	18	1.1	0.9;
	7624	1	44.95	18.7	0	5.36	0	1.032722	23.254549	220	3	1.1	0.9;
	7625	2	0	0	0	0	0	1.040138	32.45384	20	17	1.1	0.9;
	7626	1	0	-0	0.294914	17.471387	0	0.986904	36.297583	380	2	1.1	0.9;
	7627	1	102.3	14.5	0	22.12	0	1.026631	0.05062	154	1	1.1	0.9;
	7628	2	0	0	0	0	0	1.052257	44.731305	20	3	1.1	0.9;
	7629	1	8.21	0.5	0.115193	3.024006	0	1.019372	-2.020811	110	18	1.1	0.9;
	7630	1	53.46	15.4	0	0.45	0	0.992585	32.412682	220	3	1.1	0.9;
	7631	1	-0	0	0	14.04	0	1.032145	20.7029	220	5	1.1	0.9;
	7632	1	264.3	111.6	0	34.93	0	1.053421	57.48218	380	3	1.1	0.9;
	7633	1	14.22	4.7	0	2.05	0	0.989151	-11.706984	110	12	1.1	0.9;
	7634	2	0	0	0	0	0	1.04616	-9.114978	20	4	1.1	0.9;
	7635	1	5.51	17.7	0	0.58	0	0.952203	-7.669787	150	9	1.1	0.9;
	7636	2	0	0	0	0	0	1.059142	21.764381	20	14	1.1	0.9;
	7637	2	0	0	0	0	0	1.059142	21.764381	20	14	1.1	0.9;
	7638	1	-0	-0	0.07741	7.45259	0	1.041676	14.357601	220	5	1.1	0.9;
	7639	2	0	0	0	0	0	1.005118	2.132581	11	1	1.1	0.9;
	7640	2	0	0	0	0	0	1.020355	-19.386832	380	4	1.1	0.9;
	7641	2	0	0	0	0	0	0.98379	29.90696	10.6	1	1.1	0.9;
	7642	1	27.6	4	0	6.91	0	1.014614	27.544046	154	1	1.1	0.9;
	7643	2	0	0	0	0	0	0.989847	8.461746	10.5	1	1.1	0.9;
	7644	1	0	-0	0	49.26	0	0.964474	34.963089	400	1	1.1	0.9;
	7645	2	0	0	0	0	0	1.030879	24.610942	20	5	1.1	0.9;
	7646	1	0	0	0	2.59	0	1.046867	5.19198	220	7	1.1	0.9;
	7647	1	-0	0	0	29.01	0	0.964216	32.646923	400	1	1.1	0.9;
	7648	2	0	0	0	0	0	1.030879	24.610942	20	5	1.1	0.9;
	7649	2	0	0	0	0	0	1.030879	24.610942	20	5	1.1	0.9;
	7650	1	20.9	3	0	3.58	0	1.004253	18.360619	154	1	1.1	0.9;
	7651	1	-0	0	0	0.05	0	0.982403	0.459837	150	9	1.1	0.9;
	7652	1	-0	0	0	1.06	0	1.025091	37.934609	150	8	1.1	0.9;
	7653	1	0	-0	0	0.22	0	1.021282	26.179888	150	8	1.1	0.9;
	7654	2	0	0	0	0	0	1.065374	52.949406	20	10	1.1	0.9;
	7655	1	-0	0	0	0.4	0	1.027345	79.006607	220	3	1.1	0.9;
	7656	1	-4.88	-2.79	0	0.38	0	0.980995	-7.049246	110	18	1.1	0.9;
	7657	1	101.2	48.9	0	7.78	0	1.0014	-12.664749	150	9	1.1	0.9;
	7658	1	1.4	0.2	0	0.81	0	0.985093	44.213476	154	1	1.1	0.9;
	7659	1	0	-0	0	0.18	0	0.995622	2.064857	110	18	1.1	0.9;
	7660	1	69.18	38.1	0	28.66	0	1.011162	-11.417722	220	6	1.1	0.9;
	7661	1	0	0	0	0	0	1.009788	-5.34131	150	9	1.1	0.9;
	7662	2	0	0	0	0	0	0.990464	21.09396	15.75	1	1.1	0.9;
	7663	1	43.95	8.5	0	29.29	0	1.021775	34.479028	150	8	1.1	0.9;
	7664	1	91	36	0	5.47	0	1.006123	14.025015	220	5	1.1	0.9;
	7665	1	0	0	0	0.18	0	1.015795	22.756038	220	5	1.1	0.9;
	7666	1	0	0	0	14.79	0	1.066932	53.431939	380	3	1.1	0.9;
	7667	2	0	0	0	0	0	1.038747	-8.642337	20	4	1.1	0.9;
	7668	2	0	0	0	0	0	1.018739	-6.896129	20	9	1.1	0.9;
	7669	1	-17.8	-3.78	0	0.08	0	1.020153	9.00136	110	18	1.1	0.9;
	7670	2	0	0	0	0	0	1.043163	36.844687	20	3	1.1	0.9;
	7671	2	0	0	0	0	0	1.043163	36.844687	20	3	1.1	0.9;
	7672	2	0	0	0	0	0	1.043163	36.844687	20	3	1.1	0.9;
	7673	1	0	-0	0	2.24	0	0.997952	42.31077	154	1	1.1	0.9;
	7674	2	0	0	0	0	0	1.039038	39.160421	20	2	1.1	0.9;
	7675	2	0	0	0	0	0	1.038621	36.677136	20	3	1.1	0.9;
	7676	2	0	0	0	0	0	1.038621	36.677136	20	3	1.1	0.9;
	7677	2	0	0	0	0	0	1.012381	36.32396	11	1	1.1	0.9;
	7678	1	-0	-0	0.278403	6.279598	0	1.059011	14.603372	380	16	1.1	0.9;
	7679	1	0	0	0	0.23	0	1.039579	-17.294304	380	10	1.1	0.9;
	7680	1	-0	-0	0	1.06	0	0.996007	40.621651	154	1	1.1	0.9;
	7681	2	0	0	0	0	0	0.948759	58.410077	14	1	1.1	0.9;
	7682	1	18.12	11.2	0	2.96	0	0.97384	-5.930731	150	9	1.1	0.9;
	7683	1	2.2	0.5	0	0.06	0	1.022827	-6.906314	220	10	1.1	0.9;
	7684	1	72.18	44.7	0	2.72	0	1.025682	-7.286792	150	9	1.1	0.9;
	7685	1	66.97	41.5	0	0.39	0	1.023538	-11.335143	150	9	1.1	0.9;
	7686	1	36.44	12.5	0	0.59	0	0.978047	-5.585121	110	21	1.1	0.9;
	7687	1	0	-0	0	1.18	0	1.036304	5.804838	220	5	1.1	0.9;
	7688	1	69.18	19.7	0	6.99	0	1.014024	31.07903	150	8	1.1	0.9;
	7689	2	0	0	0	0	0	1.030566	-6.658147	20	9	1.1	0.9;
	7690	1	108.31	14.23	0	8.47	0	1.026803	33.480301	220	13	1.1	0.9;
	7691	1	117.5	40	0	3.69	0	1.057106	10.677187	220	3	1.1	0.9;
	7692	1	46.05	1	0.058848	3.14096	0	0.998297	91.50797	150	1	1.1	0.9;
	7693	1	2.2	0.5	0	1.84	0	0.978541	-19.945206	150	10	1.1	0.9;
	7694	1	66.47	14.14	0	0.17	0	0.974594	-3.800413	110	19	1.1	0.9;
	7695	1	143.2	5.9	0	5.79	0	1.014554	-17.012577	220	4	1.1	0.9;
	7696	1	1.8	0.6	0	0.43	0	1.003899	29.486608	150	8	1.1	0.9;
	7697	1	78.5	11.2	0	1.65	0	1.000697	34.52941	154	1	1.1	0.9;
	7698	1	0	-0	0.059142	1.940858	0	1.028653	5.929535	220	5	1.1	0.9;
	7699	1	28.93	0.2	0	0	0	1.005266	46.404335	220	3	1.1	0.9;
	7700	2	0	0	0	0	0	1.007844	10.25283	20	18	1.1	0.9;
	7701	1	0	0	0	53.04	0	1.065243	-6.294863	380	12	1.2	0.8;
	7702	1	19.82	12.3	0	0.06	0	1.0171	-4.093262	150	9	1.1	0.9;
	7703	1	63.5	9.1	0	51.96	0	1.034971	-1.011495	154	1	1.1	0.9;
	7704	1	0.39	-0.24	0	0.15	0	0.994884	-3.860114	110	18	1.1	0.9;
	7705	2	0	0	0	0	0	1.005763	29.922079	11.5	1	1.1	0.9;
	7706	1	0	-0	0	26.46055	0	1.025326	5.65858	220	20	1.1	0.9;
	7707	2	0	0	0	0	0	0.954584	49.806962	11	1	1.1	0.9;
	7708	1	0	0	0.096104	1.033896	0	1.031582	-3.692888	220	12	1.1	0.9;
	7709	1	339.9	70.5	0	28.91	0	1.031224	8.621811	220	5	1.1	0.9;
	7710	1	7	1	0	8.71	0	1.01677	27.238345	154	1	1.1	0.9;
	7711	1	0	0	0	3.06	0	0.995217	14.929118	380	5	1.1	0.9;
	7712	2	0	0	0	0	0	1.033848	35.240634	20	10	1.1	0.9;
	7713	1	0	-0	0	165.34	0	1.023683	-10.396377	380	4	1.1	0.9;
	7714	1	-40.34	18.3	0.235962	27.853028	0	1.011968	-8.193842	220	4	1.1	0.9;
	7715	1	25	3.6	0	1.15	0	1.025386	8.904631	154	1	1.1	0.9;
	7716	2	0	0	0	0	0	1.02806	-6.75496	20	6	1.1	0.9;
	7717	2	0	0	0	0	0	1.02806	-6.75496	20	6	1.1	0.9;
	7718	1	-0	-0	0	0.86	0	0.993956	-21.157564	220	6	1.1	0.9;
	7719	2	0	0	0	0	0	1.02806	-6.75496	20	6	1.1	0.9;
	7720	1	-0	-0	0	62.49	0	0.973567	19.26737	400	1	1.1	0.9;
	7721	2	0	0	0	0	0	1.02806	-6.75496	20	6	1.1	0.9;
	7722	2	0	0	0	0	0	1.02806	-6.75496	20	6	1.1	0.9;
	7723	1	31.2	4.4	0	0.35	0	1.005522	27.151982	154	1	1.1	0.9;
	7724	1	-0	-0	0	2.08	0	1.038631	53.209448	220	8	1.1	0.9;
	7725	1	16.02	4.7	0	0.94	0	1.018206	51.157996	110	8	1.1	0.9;
	7726	1	55.26	11	0	7.68	0	1.027951	5.82431	220	5	1.1	0.9;
	7727	1	0	-0	0.432051	21.130097	0	1.000718	32.41785	380	2	1.1	0.9;
	7728	1	-0	-0	0	0.48	0	1.086195	-3.617645	150	9	1.1	0.9;
	7729	1	-0	-0	0.936444	-0.728345	0	1.093017	66.981987	380	8	1.2	0.8;
	7730	1	0	-0	0	1.3	0	1.033879	20.32366	220	5	1.1	0.9;
	7731	2	0	0	0	0	0	1.015733	2.857525	6.3	1	1.1	0.9;
	7732	2	0	0	0	0	0	1.050328	7.929133	20	5	1.1	0.9;
	7733	1	-0	0	0.115812	4.47349	0	1.043338	79.315336	220	3	1.1	0.9;
	7734	2	0	0	0	0	0	1.04043	49.086444	20	10	1.1	0.9;
	7735	1	267.1	57.4	0	13.64	0	1.012081	-5.974014	380	6	1.1	0.9;
	7736	2	0	0	0	0	0	1.064843	50.81976	20	14	1.1	0.9;
	7737	1	40.04	20	0	0.05	0	0.99141	-24.920151	220	6	1.1	0.9;
	7738	1	415.8	126.6	0.183094	18.897593	0	0.996569	19.168156	220	14	1.1	0.9;
	7739	1	-0	0	0	4.87	0	1.039146	-7.96239	150	9	1.1	0.9;
	7740	2	0	0	0	0	0	1.045213	9.250128	20	14	1.1	0.9;
	7741	1	180.2	24.7	0.230883	6.557217	0	1.029393	20.464953	220	13	1.1	0.9;
	7742	2	0	0	0	0	0	1.022262	22.386094	11	1	1.1	0.9;
	7743	2	0	0	0	0	0	0.993125	29.628125	15	1	1.1	0.9;
	7744	1	-0.47	0.1	0	0.35	0	0.991606	-5.011662	110	18	1.1	0.9;
	7745	1	89.6	13.1	0	8.14	0	1.014605	-26.887876	220	4	1.1	0.9;
	7746	1	0	0	0	6.4	0	1.00234	-7.016126	150	9	1.1	0.9;
	7747	1	185.6	26.4	0	32.78	0	1.041754	-0.054874	154	1	1.1	0.9;
	7748	1	-0	0	0.264114	2.943924	0	1.031224	9.624601	220	7	1.1	0.9;
	7749	1	97.06	-0.33	0	1.24	0	1.032821	24.353123	220	13	1.1	0.9;
	7750	2	0	0	0	0	0	1.008982	42.351688	15	1	1.1	0.9;
	7751	2	0	0	0	0	0	1.008348	30.337205	14.4	1	1.1	0.9;
	7752	1	-0	-0	0	7.85	0	1.014457	46.592195	220	8	1.1	0.9;
	7753	1	1.1	0	0	1.53	0	0.990596	-14.119287	110	22	1.1	0.9;
	7754	2	0	0	0	0	0	1.032344	9.093698	20	18	1.1	0.9;
	7755	1	-0	-0	0	29.25	0	1.069965	-1.842784	380	1	1.2	0.8;
	7756	1	0	0	0	3.86	0	1.015916	42.071087	380	3	1.1	0.9;
	7757	1	0	-0	0	30.04	0	1.081813	65.890792	380	3	1.1	0.9;
	7758	1	-0	0	0	4.55	0	0.987774	6.666713	220	5	1.1	0.9;
	7759	1	46.95	54.6	0	29.46	0	0.970708	-10.509264	150	9	1.1	0.9;
	7760	1	21.12	7.7	0	0.7	0	1.031909	-3.777508	110	19	1.1	0.9;
	7761	1	111	36.6	0	2.21	0	1.036191	16.7363	220	5	1.1	0.9;
	7762	1	0	0	0	0	0	0.970166	-9.580952	150	9	1.1	0.9;
	7763	1	185.2	26.4	0	12.41	0	1.000824	17.437935	154	1	1.1	0.9;
	7764	1	41.2	56.52	0	0	0	0.947723	4.470254	110	18	1.1	0.9;
	7765	1	0	0	0.039358	8.480642	0	1.023288	75.989607	220	3	1.1	0.9;
	7766	1	-0	0	0	15.14	0	1.020274	26.013456	220	3	1.1	0.9;
	7767	1	-0	-0	0	6.93	0	1.015998	31.069444	380	3	1.1	0.9;
	7768	2	0	0	0	0	0	1.011648	24.408754	11	1	1.1	0.9;
	7769	1	0	0	0	138.24	0	0.967461	24.197449	400	1	1.1	0.9;
	7770	1	0	-0	0	34.13	0	1.037193	-13.548015	380	4	1.1	0.9;
	7771	1	0	-0	0.225625	26.287525	0	1.063673	-0.208683	380	21	1.1	0.9;
	7772	1	288.7	117.3	0	183.57	0	1.016119	-9.542933	380	6	1.1	0.9;
	7773	1	55.4	7.9	0	0.12	0	1.018408	19.83996	154	1	1.1	0.9;
	7774	1	-12.61	4.1	0	0.33	0	1.019517	-6.861741	150	9	1.1	0.9;
	7775	1	0	-0	0	0.23	0	0.990546	-8.428479	150	10	1.1	0.9;
	7776	2	0	0	0	0	0	1.032321	27.942453	20	5	1.1	0.9;
	7777	2	0	0	0	0	0	1.009853	14.713647	13.8	1	1.1	0.9;
	7778	1	83.59	-7.2	0	12.95	0	0.991779	-19.695979	220	10	1.1	0.9;
	7779	1	17.52	6.6	0	3.79	0	1.026053	18.466754	220	5	1.1	0.9;
	7780	1	-0	0	0	1.79	0	1.028263	33.70683	380	2	1.1	0.9;
	7781	1	-0	-0	0	0.57	0	1.012112	8.292067	110	18	1.1	0.9;
	7782	1	-0	-0	0.090774	1.480444	0	1.016164	-3.688379	220	18	1.1	0.9;
	7783	1	19.4	2.8	0	4.73	0	0.984846	37.960461	154	1	1.1	0.9;
	7784	2	0	0	0	0	0	1.023144	-9.538309	20	4	1.1	0.9;
	7785	1	-16.62	8.7	0	7.68	0	0.991494	-4.787477	220	10	1.1	0.9;
	7786	1	0	-0	0.05	4.54	0	0.985348	-3.762933	220	21	1.1	0.9;
	7787	1	64.27	17.5	0	0.43	0	1.011021	9.09354	220	5	1.1	0.9;
	7788	1	0	0	0	3.36	0	1.028111	-1.849562	154	1	1.1	0.9;
	7789	1	0	-0	0	0.81	0	1.020697	60.764946	110	8	1.1	0.9;
	7790	1	-0	0	0.030084	3.709916	0	1.004708	-15.20422	220	4	1.1	0.9;
	7791	1	5.5	0.8	0	7.28	0	1.00454	36.826069	154	1	1.1	0.9;
	7792	1	0	0	0.573568	0.23309	0	1.039218	34.334547	380	5	1.1	0.9;
	7793	1	-0	-0	0	3.28	0	1.008722	13.375916	220	5	1.1	0.9;
	7794	1	28.93	8.3	0	0.42	0	1.018005	31.659728	150	8	1.1	0.9;
	7795	1	-0	-0	0	0.34	0	0.992301	-5.634869	110	3	1.1	0.9;
	7796	1	419.5	114.1	0	55.629522	0	1.012958	20.531904	380	6	1.1	0.9;
	7797	1	229	77.1	0	36.732397	0	1.025822	-6.866112	380	6	1.1	0.9;
	7798	1	78	11.1	0	1.33	0	1.024428	6.959359	154	1	1.1	0.9;
	7799	1	7	1	0	4.5	0	1.014249	27.578122	154	1	1.1	0.9;
	7800	1	0	0	0	53.34	0	0.969691	31.843897	400	1	1.1	0.9;
	7801	1	-0	-0	0	55	0	1.023124	25.472115	150	8	1.1	0.9;
	7802	1	-0	-0	0	0.43	0	0.944949	-10.831545	150	9	1.1	0.9;
	7803	1	-0	-0	0	1.54	0	0.912138	-12.07238	150	9	1.1	0.9;
	7804	1	58.16	16.5	0	6.24	0	1.029984	17.936173	220	5	1.1	0.9;
	7805	1	173.6	105.6	0	21.26	0	1.005544	13.352738	380	3	1.1	0.9;
	7806	1	0	0	0	2.77	0	0.961678	-15.621054	220	10	1.1	0.9;
	7807	1	152.73	-17.2	0	45.85	0	1.024781	6.070788	220	5	1.1	0.9;
	7808	2	0	0	0	0	0	1.031165	22.577641	20	10	1.1	0.9;
	7809	1	79.79	49.4	0	2.24	0	1.012421	-10.166509	150	9	1.1	0.9;
	7810	1	110.6	38.1	0	0.98	0	1.015703	25.158766	150	8	1.1	0.9;
	7811	1	11.11	3	0.019465	22.220535	0	1.05855	17.887152	220	5	1.1	0.9;
	7812	2	0	0	0	0	0	1.026025	16.98036	20	14	1.1	0.9;
	7813	1	66.07	12.5	0	2.08	0	1.032564	-19.50784	220	4	1.1	0.9;
	7814	2	0	0	0	0	0	1.064355	32.276383	20	14	1.1	0.9;
	7815	1	40.5	5.7	0	5.29	0	1.034227	18.922301	154	1	1.1	0.9;
	7816	1	0	0	0.173707	0.774894	0	1.040591	20.715061	110	19	1.1	0.9;
	7817	1	118.5	18.3	0.038959	6.891041	0	1.028099	32.52124	220	13	1.1	0.9;
	7818	1	93.59	77.1	0	91.89	0	1.023323	44.777999	380	3	1.1	0.9;
	7819	1	0	-0	0	0	0	1.027195	-15.651629	20	4	1.1	0.9;
	7820	1	0	-0	0	0	0	1.027195	-15.651629	20	4	1.1	0.9;
	7821	1	0	-0	0	0	0	1.027195	-15.651629	20	4	1.1	0.9;
	7822	1	0	-0	0	0	0	1.027195	-15.651629	20	4	1.1	0.9;
	7823	1	-13.32	20.6	0	1.16	0	1.076588	-9.350023	150	9	1.2	0.8;
	7824	1	0	-0	0	0	0	1.005017	-18.861992	20	4	1.1	0.9;
	7825	2	0	0	0	0	0	1.027844	-11.185494	20	4	1.1	0.9;
	7826	1	21.02	8	0	13.14	0	1.047585	-13.395079	380	4	1.1	0.9;
	7827	1	120.94	25.23	0	5.9	0	1.002971	29.68317	154	1	1.1	0.9;
	7828	1	13.2	3.25	0	5.66	0	0.989515	88.943489	150	1	1.1	0.9;
	7829	1	48.8	6.9	0	1.59	0	0.996861	27.639759	154	1	1.1	0.9;
	7830	1	408.3	84.5	0	18.02	0	1.010023	-7.96254	380	6	1.1	0.9;
	7831	1	-0	-0	0	2.96	0	1.028315	92.327608	220	1	1.1	0.9;
	7832	1	13.21	4.4	0	0.86	0	0.986743	-17.679196	110	22	1.1	0.9;
	7833	2	0	0	0	0	0	1.027461	-17.241264	20	4	1.1	0.9;
	7834	1	-0	-0	1.063676	6.592665	0	1.081314	24.164392	380	5	1.1	0.9;
	7835	1	127.3	90.5	0	6.8	0	1.010769	30.193461	380	3	1.1	0.9;
	7836	1	0	0	0	0	0	1.022953	-11.439051	150	9	1.1	0.9;
	7837	1	0.8	0.5	0	0	0	0.993187	-7.06435	150	9	1.1	0.9;
	7838	1	-0	-0	0	24.7	0	1.036605	5.426006	220	5	1.1	0.9;
	7839	1	48.05	28.8	0	0.02	0	1.01143	-10.415049	150	9	1.1	0.9;
	7840	1	-0	-0	0	17.83	0	1.024756	7.179138	380	1	1.1	0.9;
	7841	2	0	0	0	0	0	1.006944	-5.921798	20	2	1.1	0.9;
	7842	2	0	0	0	0	0	1.005097	-2.034262	20	2	1.1	0.9;
	7843	1	243.45	104.28	0	4.66	0	0.978579	18.445171	220	13	1.1	0.9;
	7844	2	0	0	0	0	0	1.048358	31.36576	20	5	1.1	0.9;
	7845	1	-0	-0	0	4.39	0	1.020755	30.541323	380	3	1.1	0.9;
	7846	2	0	0	0	0	0	1.048358	31.36576	20	5	1.1	0.9;
	7847	2	0	0	0	0	0	1.006137	-0.690843	20	2	1.1	0.9;
	7848	2	0	0	0	0	0	1.005097	-2.034262	20	2	1.1	0.9;
	7849	2	0	0	0	0	0	1.041549	2.824814	20	23	1.1	0.9;
	7850	1	0	0	0	55.700331	0	0.994055	13.044026	220	24	1.1	0.9;
	7851	2	0	0	0	0	0	1.005097	-2.034262	20	2	1.1	0.9;
	7852	1	39.34	24.3	0	3.06	0	1.017465	-8.354209	150	9	1.1	0.9;
	7853	1	0	0	0	3.62	0	1.033509	29.0844	220	1	1.1	0.9;
	7854	1	57.06	11.9	0	1.68	0	1.027445	-20.926376	220	4	1.1	0.9;
	7855	1	0	0	0	0.62	0	0.967513	-5.643907	150	9	1.1	0.9;
	7856	1	0	0	0	9.43	0	1.017543	13.164281	220	5	1.1	0.9;
	7857	2	0	0	0	0	0	1.014368	23.036796	6.3	1	1.1	0.9;
	7858	2	0	0	0	0	0	1.037973	11.471538	20	14	1.1	0.9;
	7859	2	0	0	0	0	0	1.022099	17.218586	20	14	1.1	0.9;
	7860	1	0	0	0	2.77	0	1.006175	27.984073	154	1	1.1	0.9;
	7861	1	-0	-0	0	9.1	0	1.039829	-6.000279	220	5	1.1	0.9;
	7862	1	-0	-0	0	1.52	0	1.008156	0.533035	150	9	1.1	0.9;
	7863	1	186.2	3.7	0.039901	0.100099	0	0.989328	-5.037058	110	21	1.1	0.9;
	7864	1	-0	-0	0.233672	-0.194726	0	1.024341	-12.447897	380	6	1.1	0.9;
	7865	1	20.9	3	0	2.4	0	0.986315	45.881122	154	1	1.1	0.9;
	7866	1	241.62	52.78	0	0.82	0	1.013108	15.40527	220	5	1.1	0.9;
	7867	1	-0	-0	0.057655	27.001954	0	1.051051	80.227836	380	3	1.1	0.9;
	7868	1	-0	0	0	0.64	0	1.027313	23.13711	154	1	1.1	0.9;
	7869	1	-0	-0	0.29412	8.4749	0	1.031202	40.220313	380	3	1.1	0.9;
	7870	1	5.21	1.7	0	0.87	0	1.008153	-6.482489	110	12	1.1	0.9;
	7871	1	0	0	0	0	0	1.031416	-7.138703	150	9	1.1	0.9;
	7872	1	-0	0	0	36.42	0	1.106096	-3.070014	380	9	1.2	0.8;
	7873	1	6.61	2.4	0	0.81	0	0.986459	-2.169128	110	18	1.1	0.9;
	7874	1	11.91	4.3	0	0.66	0	1.003978	-4.980697	110	19	1.1	0.9;
	7875	2	0	0	0	0	0	1.055337	0.681282	20	9	1.1	0.9;
	7876	1	0.9	27.4	0.158479	16.541141	0	1.011225	-11.321651	220	4	1.1	0.9;
	7877	2	0	0	0	0	0	0.996117	32.042491	11	1	1.1	0.9;
	7878	1	69.78	26	0	5.18	0	1.035434	8.556606	220	5	1.1	0.9;
	7879	1	0	0	0	3.79	0	1.010494	19.277469	154	1	1.1	0.9;
	7880	1	-0	0	0.226101	1.121004	0	1.073183	36.32563	380	14	1.1	0.9;
	7881	1	147	56.3	0	0.51	0	1.009545	29.617666	150	8	1.1	0.9;
	7882	1	0	0	0	0	0	1.047679	-11.39683	150	9	1.1	0.9;
	7883	1	16.7	2.4	0	0.02	0	1.008263	28.002972	154	1	1.1	0.9;
	7884	1	-0	0	0	25.13	0	1.040535	35.122896	380	2	1.1	0.9;
	7885	1	2.8	1	0	0.18	0	0.988474	-3.626576	110	19	1.1	0.9;
	7886	1	0	-0	0	3.96	0	1.033255	9.023406	220	5	1.1	0.9;
	7887	1	349.3	144.3	0	0.54	0	1.010224	13.928751	220	5	1.1	0.9;
	7888	1	56.96	1.94	0	26.48	0	1.008603	6.034417	380	5	1.1	0.9;
	7889	1	-0	0	0	0.92	0	0.995215	14.929193	380	5	1.1	0.9;
	7890	2	0	0	0	0	0	1.024544	18.974877	20	14	1.1	0.9;
	7891	1	0	-0	0.174657	-0.135844	0	1.018138	42.343965	150	8	1.1	0.9;
	7892	1	143.3	71.4	0	11.16	0	1.005547	11.257255	380	3	1.1	0.9;
	7893	1	104.5	14.9	0	0.67	0	1.006809	18.334783	154	1	1.1	0.9;
	7894	1	33.34	20.7	0	0.02	0	1.026448	-3.53911	150	9	1.1	0.9;
	7895	1	0	0	0	2.04	0	1.076103	-2.193069	110	19	1.1	0.9;
	7896	1	0.5	17.9	0.151979	42.926015	0	1.063731	-8.12425	380	4	1.1	0.9;
	7897	1	-0	0	0	6.07	0	1.016823	7.32684	220	7	1.1	0.9;
	7898	1	152.8	38	0	1.46	0	1.002692	26.834688	150	8	1.1	0.9;
	7899	2	0	0	0	0	0	1.034118	0.51016	20	18	1.1	0.9;
	7900	1	233.8	33.4	0	1.22	0	1.025324	1.838916	154	1	1.1	0.9;
	7901	2	0	0	0	0	0	1.037128	1.366278	20	18	1.1	0.9;
	7902	2	0	0	0	0	0	1.037128	1.366278	20	18	1.1	0.9;
	7903	2	0	0	0	0	0	1.016718	-0.774001	20	18	1.1	0.9;
	7904	2	0	0	0	0	0	1.016718	-0.774001	20	18	1.1	0.9;
	7905	2	0	0	0	0	0	1.042846	0.690036	20	18	1.1	0.9;
	7906	1	26.4	3.8	0	9.18	0	1.01006	29.755173	154	1	1.1	0.9;
	7907	1	58.5	8.3	0	1.2	0	0.993402	10.337394	154	1	1.1	0.9;
	7908	2	0	0	0	0	0	1.016718	-0.774001	20	18	1.1	0.9;
	7909	1	0.01	0	0	0	0	1.079773	-12.967146	150	9	1.2	0.8;
	7910	1	69.5	9.8	0	1.91	0	1.011179	13.966873	154	1	1.1	0.9;
	7911	1	103.3	24.1	0	9.41	0	1.015261	-19.554209	220	4	1.1	0.9;
	7912	1	-4.86	-2.12	0	0.26	0	0.994456	-2.15907	110	18	1.1	0.9;
	7913	1	-0	-0	0	11.59	0	1.02601	13.667023	380	1	1.1	0.9;
	7914	1	104.7	-12.8	0	2.25	0	1.040499	5.769009	220	7	1.1	0.9;
	7915	1	0	0	0	2.72	0	1.034903	75.355671	380	3	1.1	0.9;
	7916	1	116.4	-8.8	0	8.39	0	1.074984	17.200912	220	3	1.2	0.8;
	7917	1	138.01	34.55	0	8.44	0	1.048186	30.077687	220	5	1.1	0.9;
	7918	2	0	0	0	0	0	0.994601	31.575691	11	1	1.1	0.9;
	7919	1	-0.17	-0.84	0	0	0	1.017657	42.270931	150	8	1.1	0.9;
	7920	1	2.8	1	0	0.95	0	1.029471	-2.690244	110	19	1.1	0.9;
	7921	2	0	0	0	0	0	1.023478	24.695749	10.5	1	1.1	0.9;
	7922	1	-0	0	0	9.77	0	1.005907	26.581496	380	3	1.1	0.9;
	7923	1	53.56	33.2	0	3.77	0	0.986867	-7.038889	150	9	1.1	0.9;
	7924	1	8.91	1.8	0	1.78	0	0.984641	-5.606786	110	18	1.1	0.9;
	7925	2	0	0	0	0	0	1.032265	43.171125	20	10	1.1	0.9;
	7926	1	9.61	1.8	0	2.84	0	0.982765	1.187555	110	18	1.1	0.9;
	7927	1	3.33	2.95	0	1.5	0	0.955017	-5.135866	110	18	1.1	0.9;
	7928	1	0	-0	0	1.06	0	1.009223	1.639395	150	7	1.1	0.9;
	7929	2	0	0	0	0	0	1.034412	41.712007	20	10	1.1	0.9;
	7930	2	0	0	0	0	0	1.129641	7.22366	20	9	1.2	0.8;
	7931	2	0	0	0	0	0	1.026444	23.18548	20	10	1.1	0.9;
	7932	2	0	0	0	0	0	1.028453	-12.903077	20	4	1.1	0.9;
	7933	2	0	0	0	0	0	0.991806	18.024083	20	18	1.1	0.9;
	7934	1	-0	0	0	1.82	0	1.032454	-19.078332	220	6	1.1	0.9;
	7935	1	82.09	39.8	0	2.35	0	0.979359	-20.534001	150	10	1.1	0.9;
	7936	1	45.58	-1.95	0	1.8	0	1.015733	89.969438	150	1	1.1	0.9;
	7937	1	-0	0	0	1.43	0	1.024372	-3.025544	150	9	1.1	0.9;
	7938	1	-0	0	0	5.27	0	1.029117	21.174398	220	5	1.1	0.9;
	7939	2	0	0	0	0	0	1.019144	17.562907	20	14	1.1	0.9;
	7940	1	-0	0	0	1.84	0	0.930776	-1.864544	110	19	1.1	0.9;
	7941	1	6.01	2.2	0.03841	1.04159	0	1.037259	10.990708	110	19	1.1	0.9;
	7942	1	-0	-0	0	9.09	0	1.033674	9.102988	220	5	1.1	0.9;
	7943	2	0	0	0	0	0	1.028453	-12.903077	20	4	1.1	0.9;
	7944	1	211.3	69.3	0	7.96	0	0.975641	-22.657708	220	6	1.1	0.9;
	7945	1	0	-0	0	1.77	0	1.043815	-4.396276	220	1	1.1	0.9;
	7946	2	0	0	0	0	0	1.027132	31.963051	20	5	1.1	0.9;
	7947	1	-0	-0	0	0.12	0	1.065187	-4.78518	150	9	1.1	0.9;
	7948	1	-0	0	0.039529	24.367428	0	1.056758	10.699713	380	5	1.1	0.9;
	7949	2	0	0	0	0	0	1.02542	12.734487	11	1	1.1	0.9;
	7950	1	-0	-0	0	16.92	0	1.123995	-0.160436	380	1	1.2	0.8;
	7951	2	0	0	0	0	0	0.987171	19.06229	20	18	1.1	0.9;
	7952	1	25.23	3.3	0	35.45	0	1.030003	-10.973638	380	4	1.1	0.9;
	7953	1	-0	-0	0	91.66	0	0.962686	21.941446	400	1	1.1	0.9;
	7954	1	0	0	0.05035	5.60965	0	0.998859	-19.342822	220	4	1.1	0.9;
	7955	1	67.5	9.6	0	26.5	0	1.034538	4.92857	400	1	1.1	0.9;
	7956	1	1	0.3	0	0.46	0	1.032639	0.267894	110	19	1.1	0.9;
	7957	1	0	-0	0	1.408086	0	0.999409	1.749705	220	20	1.1	0.9;
	7958	1	106.6	-18.8	0	4.1	0	1.025319	-11.644989	220	5	1.1	0.9;
	7959	1	3.3	1.1	0	0.25	0	1.02166	-5.451912	110	22	1.1	0.9;
	7960	1	0	-0	0	20.81	0	1.012874	43.808504	380	8	1.1	0.9;
	7961	1	1.3	-0	0	0.41	0	1.026149	4.103817	110	18	1.1	0.9;
	7962	2	0	0	0	0	0	1.03013	4.624843	6.3	1	1.1	0.9;
	7963	1	0	0	0	4.42	0	1.048245	57.713043	380	1	1.1	0.9;
	7964	1	8.91	2.5	0	0.7	0	0.969247	-5.056037	110	18	1.1	0.9;
	7965	2	0	0	0	0	0	1.042865	51.92649	20	3	1.1	0.9;
	7966	2	0	0	0	0	0	1.042865	51.92649	20	3	1.1	0.9;
	7967	2	0	0	0	0	0	1.043738	-8.975119	20	4	1.1	0.9;
	7968	1	0	0	0	90.33	0	1.015817	-18.078109	380	6	1.1	0.9;
	7969	1	0	0	0	18.02	0	1.024873	3.61754	220	5	1.1	0.9;
	7970	2	0	0	0	0	0	1.045647	-10.158279	20	4	1.1	0.9;
	7971	1	-0	-0	0	11.04	0	1.028847	5.912393	220	5	1.1	0.9;
	7972	1	0	-0	0	91.78	0	0.945368	40.225431	400	1	1.1	0.9;
	7973	1	15.72	9.7	0	1.61	0	1.009202	-9.329154	150	9	1.1	0.9;
	7974	1	30.93	5.8	0	4.45	0	1.062085	2.562194	220	5	1.1	0.9;
	7975	1	-0	0	1.16103	-0.644889	0	1.059168	20.496879	380	5	1.1	0.9;
	7976	2	0	0	0	0	0	1.012031	-8.893616	20	10	1.1	0.9;
	7977	1	144	56.6	0	5.61	0	0.991636	-4.411495	220	6	1.1	0.9;
	7978	1	-0	0	0.059659	26.507268	0	1.015664	93.242149	380	1	1.1	0.9;
	7979	2	0	0	0	0	0	1.035023	-10.739655	20	4	1.1	0.9;
	7980	1	-0	-0	0	30.55	0	1.102262	-15.183704	150	9	1.2	0.8;
	7981	2	0	0	0	0	0	1.035023	-10.739655	20	4	1.1	0.9;
	7982	1	0	0	0	5.75	0	1.013418	23.824726	220	1	1.1	0.9;
	7983	1	0	-0	0.573252	0.193341	0	1.039291	34.333574	380	5	1.1	0.9;
	7984	1	0	0	0.209041	8.836502	0	1.034943	4.996743	220	18	1.1	0.9;
	7985	1	148.2	31.2	0.040188	16.499812	0	1.037837	-0.48399	220	5	1.1	0.9;
	7986	1	11.81	2.4	0	0.16	0	0.979495	5.576426	110	18	1.1	0.9;
	7987	1	0	0	2.296299	56.527719	0	1.061678	13.951101	380	5	1.1	0.9;
	7988	2	0	0	0	0	0	1.035023	-10.739655	20	4	1.1	0.9;
	7989	2	0	0	0	0	0	1.035023	-10.739655	20	4	1.1	0.9;
	7990	1	0	0	0.425362	83.915128	0	1.048569	30.055695	380	3	1.1	0.9;
	7991	1	102.3	14.6	0	1.1	0	1.014936	-2.091201	154	1	1.1	0.9;
	7992	2	0	0	0	0	0	1.041485	29.444117	20	5	1.1	0.9;
	7993	2	0	0	0	0	0	1.041485	29.444117	20	5	1.1	0.9;
	7994	1	-0	0	0	3.21	0	0.990892	-16.757203	220	6	1.1	0.9;
	7995	2	0	0	0	0	0	1.029709	20.473808	20	5	1.1	0.9;
	7996	1	-0	-0	0	4.69	0	0.999047	-9.556209	220	4	1.1	0.9;
	7997	1	195.74	-32.87	0	0.83	0	1.099689	80.197228	380	3	1.2	0.8;
	7998	2	0	0	0	0	0	0.9865	27.337996	20	5	1.1	0.9;
	7999	1	139.77	41.14	0	15.57	0	1.022426	18.257301	220	13	1.1	0.9;
	8000	2	0	0	0	0	0	1.029709	20.473808	20	5	1.1	0.9;
	8001	2	0	0	0	0	0	1.029709	20.473808	20	5	1.1	0.9;
	8002	1	-0	0	0	6.89	0	0.995026	-9.217762	220	22	1.1	0.9;
	8003	1	574.3	94.6	0	16.63	0	0.993351	-7.087304	380	6	1.1	0.9;
	8004	1	20.12	7.9	0	0.08	0	1.038218	-2.96549	110	15	1.1	0.9;
	8005	1	111.5	24.8	0	9.65	0	1.014429	-26.565142	220	4	1.1	0.9;
	8006	1	0	0	0	16.84	0	0.989942	-11.563854	220	22	1.1	0.9;
	8007	1	0	-0	0	0.58	0	1.034063	9.790369	220	5	1.1	0.9;
	8008	1	342.7	82.2	0	75.02	0	1.058182	13.863396	380	16	1.1	0.9;
	8009	1	17.62	5.5	0.038856	1.361144	0	1.016709	-11.6258	110	22	1.1	0.9;
	8010	2	0	0	0	0	0	0.981441	27.374251	6.3	1	1.1	0.9;
	8011	1	92.59	8.37	0	16.58	0	1.048779	17.891441	380	13	1.1	0.9;
	8012	1	-0	-0	0	0.2	0	0.964063	-6.217216	110	18	1.1	0.9;
	8013	1	17.62	7.7	0	10.73	0	0.987716	-14.359155	110	22	1.1	0.9;
	8014	1	0	0	0	0	0	1.017465	-8.354209	150	9	1.1	0.9;
	8015	1	0	-0	0	50.54	0	0.995728	-9.194286	220	22	1.1	0.9;
	8016	1	499	191.1	0	60.73	0	1.005033	-16.192033	380	6	1.1	0.9;
	8017	2	0	0	0	0	0	1.052257	44.731305	20	3	1.1	0.9;
	8018	1	67.68	6.2	0	6.83	0	1.020785	-32.057931	220	4	1.1	0.9;
	8019	1	73.18	13.1	0.078753	5.501247	0	1.028516	-14.223496	220	4	1.1	0.9;
	8020	1	456	123.8	0.23482	16.044101	0	1.019656	21.733827	220	13	1.1	0.9;
	8021	2	0	0	0	0	0	1.052257	44.731305	20	3	1.1	0.9;
	8022	2	0	0	0	0	0	1.052257	44.731305	20	3	1.1	0.9;
	8023	2	0	0	0	0	0	0.9595	30.616672	10.6	1	1.1	0.9;
	8024	1	22.12	8	0	0.14	0	0.974629	-2.300917	110	19	1.1	0.9;
	8025	1	18.22	6	0	0.14	0	1.013755	-8.966796	110	12	1.1	0.9;
	8026	1	-0	0	0	61.88	0	0.967843	38.675053	400	1	1.1	0.9;
	8027	1	34.24	-19.6	0	3.59	0	0.968655	-15.322053	150	10	1.1	0.9;
	8028	1	4.4	2.2	0	1.19	0	1.012501	-10.498035	110	22	1.1	0.9;
	8029	1	0	-0	0	0	0	1.022632	4.381728	20	14	1.1	0.9;
	8030	1	-0	-0	0	1.98	0	1.03685	5.364208	220	5	1.1	0.9;
	8031	1	471.79	140.6	0.266552	16.151011	0	1.041947	32.905618	220	13	1.1	0.9;
	8032	1	30.43	10.1	0	1.07	0	1.073663	4.638114	110	19	1.2	0.8;
	8033	2	0	0	0	0	0	1.010732	24.191523	11	1	1.1	0.9;
	8034	1	-0	-0	0	1.18	0	1.065239	30.419096	220	5	1.1	0.9;
	8035	1	150.2	25.5	0	42.49	0	1.02239	28.921949	380	3	1.1	0.9;
	8036	1	14.52	5.3	0	0.49	0	1.021827	-0.432854	110	15	1.1	0.9;
	8037	1	0	0	0	4.11	0	1.038408	-1.18914	154	1	1.1	0.9;
	8038	1	236.33	59.35	0	18.71	0	1.035923	15.651998	220	5	1.1	0.9;
	8039	2	0	0	0	0	0	1.016716	24.455607	20	5	1.1	0.9;
	8040	1	49.65	12.3	0	2.2	0	1.032725	8.156247	220	5	1.1	0.9;
	8041	1	0	-0	0	58.79	0	1.033791	12.544574	380	7	1.1	0.9;
	8042	1	13.92	4.1	0	2.27	0	0.967135	9.358512	110	18	1.1	0.9;
	8043	1	0	-0	0	0.36	0	1.048207	30.077507	220	5	1.1	0.9;
	8044	1	154.9	18.7	0.039758	6.560242	0	1.003644	29.588773	220	3	1.1	0.9;
	8045	1	-0	0	0	0.82	0	1.036466	79.982353	220	3	1.1	0.9;
	8046	1	51.8	7.4	0	3.41	0	1.007687	16.544078	154	1	1.1	0.9;
	8047	1	0	-0	0.09832	1.38168	0	1.013027	13.869066	220	5	1.1	0.9;
	8048	1	0	-0	0	0.6	0	1.002448	-8.572679	110	1	1.1	0.9;
	8049	1	68.08	42.2	0	4.4	0	1.015869	-10.791845	150	9	1.1	0.9;
	8050	1	85.8	12.2	0	1.7	0	1.016762	17.505942	154	1	1.1	0.9;
	8051	1	-0.53	-0.91	0.090774	7.680444	0	1.064407	-2.034851	220	18	1.2	0.8;
	8052	1	91.9	13.1	0	7.35	0	1.017285	27.012865	154	1	1.1	0.9;
	8053	2	0	0	0	0	0	1.028453	-12.903077	20	4	1.1	0.9;
	8054	1	12.21	4.4	0	2.77	0	0.937011	-2.220462	110	19	1.1	0.9;
	8055	2	0	0	0	0	0	1.028453	-12.903077	20	4	1.1	0.9;
	8056	2	0	0	0	0	0	1.028453	-12.903077	20	4	1.1	0.9;
	8057	1	0	0	0	0.94	0	0.989447	3.221942	110	18	1.1	0.9;
	8058	1	11.71	2.2	0	0.48	0	1.018751	-6.51886	110	3	1.1	0.9;
	8059	1	6.71	3	0	1.3	0	1.030513	-3.211646	110	15	1.1	0.9;
	8060	2	0	0	0	0	0	0.939142	40.851376	13.8	1	1.1	0.9;
	8061	1	-0	-0	0	7.08	0	1.009419	14.139585	380	3	1.1	0.9;
	8062	1	47.65	29.5	0	4.19	0	0.912043	-12.081031	150	9	1.1	0.9;
	8063	2	0	0	0	0	0	1.029166	13.522836	20	14	1.1	0.9;
	8064	1	33.04	8.1	0	1.55	0	0.992477	-17.499986	110	22	1.1	0.9;
	8065	1	-0	-0	0.019551	0.530449	0	1.054956	5.244845	220	5	1.1	0.9;
	8066	1	10.01	3	0	0.47	0	0.979131	-6.584456	110	21	1.1	0.9;
	8067	2	0	0	0	0	0	1.021005	21.987746	20	14	1.1	0.9;
	8068	1	30.03	12.2	0	0.93	0	1.022252	26.043187	150	8	1.1	0.9;
	8069	1	-0	0	0	0.92	0	0.997732	-13.333126	110	12	1.1	0.9;
	8070	1	-0	-0	0	24.052379	0	0.986909	14.960718	380	5	1.1	0.9;
	8071	1	15.12	5	0	0.69	0	0.997776	-13.235724	110	12	1.1	0.9;
	8072	1	23.93	8.7	0	0.42	0	1.017257	-3.442803	110	19	1.1	0.9;
	8073	1	2.22	7.8	0	19.16	0	0.978554	-8.141372	150	9	1.1	0.9;
	8074	1	21.22	6.7	0	0.2	0	0.973556	-6.561174	110	21	1.1	0.9;
	8075	1	0	0	0	17.79	0	1.032389	35.40681	380	3	1.1	0.9;
	8076	1	88.93	16.75	0	6.75	0	1.039236	44.914119	220	13	1.1	0.9;
	8077	1	0	-0	0.343895	2.96342	0	1.055478	38.52234	380	3	1.1	0.9;
	8078	1	50.26	16.5	0	0.69	0	0.928894	-8.400611	110	21	1.1	0.9;
	8079	1	-0	-0	0	10.43	0	1.035694	16.395031	380	1	1.1	0.9;
	8080	1	0	0	0	0	0	1.023776	-5.098586	150	9	1.1	0.9;
	8081	1	0	-0	0	15.17	0	1.011802	45.75926	380	8	1.1	0.9;
	8082	1	0	-0	0	2.66	0	0.99158	-15.458284	220	10	1.1	0.9;
	8083	1	-0	-0	0	1.56	0	1.021745	-7.001016	150	9	1.1	0.9;
	8084	1	31.13	12	0	1.03	0	1.026855	-3.498214	110	15	1.1	0.9;
	8085	1	-0	0	0.19855	29.549	0	1.04669	-18.503591	380	10	1.1	0.9;
	8086	1	21.62	7.8	0	1.21	0	0.969435	-2.584123	110	19	1.1	0.9;
	8087	1	0	0	0.038832	0.271168	0	1.035327	16.80885	220	5	1.1	0.9;
	8088	2	0	0	0	0	0	1.055956	46.135912	20	8	1.1	0.9;
	8089	2	0	0	0	0	0	1.017878	-10.871595	20	10	1.1	0.9;
	8090	2	0	0	0	0	0	1.055956	46.135912	20	8	1.1	0.9;
	8091	2	0	0	0	0	0	0.951565	38.529485	13.8	1	1.1	0.9;
	8092	2	0	0	0	0	0	1.027348	16.396901	20	14	1.1	0.9;
	8093	1	5.11	1.8	0	0.62	0	1.028437	11.212731	110	19	1.1	0.9;
	8094	1	0	-0	0	1.59	0	1.001052	28.442883	380	3	1.1	0.9;
	8095	1	0	0	0	3.59	0	1.00298	10.352439	220	5	1.1	0.9;
	8096	2	0	0	0	0	0	1.015954	3.75914	20	9	1.1	0.9;
	8097	2	0	0	0	0	0	1.043138	41.982485	20	10	1.1	0.9;
	8098	1	82.89	41.3	0	36.2	0	1.036356	36.321766	150	8	1.1	0.9;
	8099	2	0	0	0	0	0	1.028604	-12.690591	20	4	1.1	0.9;
	8100	1	-0	0	0	6.29	0	1.006677	-15.127499	150	9	1.1	0.9;
	8101	1	-0	0	0	0.45	0	0.97105	7.093885	110	18	1.1	0.9;
	8102	1	23.53	8.5	0	0.64	0	1.004221	-3.48294	110	19	1.1	0.9;
	8103	1	0	-0	0	1.01	0	1.021286	7.910246	220	5	1.1	0.9;
	8104	1	-0	-0.1	0.080282	-0.060211	0	1.017654	25.468999	220	6	1.1	0.9;
	8105	1	0	-0	0	2.94	0	1.059015	5.650918	220	5	1.1	0.9;
	8106	1	99.7	6	0	2.06	0	1.032111	19.618252	220	5	1.1	0.9;
	8107	1	0	-0	0	59.52	0	1.025876	-12.172104	380	6	1.1	0.9;
	8108	1	220	68.3	0	3.09	0	1.017196	5.79313	220	7	1.1	0.9;
	8109	2	0	0	0	0	0	1.010466	11.179882	20	18	1.1	0.9;
	8110	2	0	0	0	0	0	1.055415	20.933966	20	14	1.1	0.9;
	8111	2	0	0	0	0	0	1.049781	32.053958	20	14	1.1	0.9;
	8112	2	0	0	0	0	0	1.055415	20.933966	20	14	1.1	0.9;
	8113	2	0	0	0	0	0	1.051968	36.354642	20	14	1.1	0.9;
	8114	2	0	0	0	0	0	1.049781	32.053958	20	14	1.1	0.9;
	8115	1	26.63	8.6	0	4.44	0	0.934933	-8.893156	110	18	1.1	0.9;
	8116	2	0	0	0	0	0	1.051968	36.354642	20	14	1.1	0.9;
	8117	1	16.22	5.3	0	0.25	0	0.993316	-11.194562	110	12	1.1	0.9;
	8118	1	130.7	89.1	0	0.08	0	1.016798	24.876978	220	3	1.1	0.9;
	8119	1	0	-0	0	88.66	0	0.99229	29.110566	400	1	1.1	0.9;
	8120	1	-0	0	0	0.17	0	0.966664	-9.644625	150	9	1.1	0.9;
	8121	2	0	0	0	0	0	0.989111	40.491338	11.5	1	1.1	0.9;
	8122	2	0	0	0	0	0	0.984601	26.113186	6.3	1	1.1	0.9;
	8123	2	0	0	0	0	0	1.031098	31.248826	20	5	1.1	0.9;
	8124	1	3.2	1	0	0.64	0	1.003398	5.466678	110	18	1.1	0.9;
	8125	2	0	0	0	0	0	1.020649	22.846825	20	6	1.1	0.9;
	8126	2	0	0	0	0	0	1.031098	31.248826	20	5	1.1	0.9;
	8127	2	0	0	0	0	0	1.020649	22.846825	20	6	1.1	0.9;
	8128	2	0	0	0	0	0	1.020649	22.846825	20	6	1.1	0.9;
	8129	1	1.9	0.7	0	0.48	0	0.971373	-3.469829	110	19	1.1	0.9;
	8130	2	0	0	0	0	0	1.046665	36.809947	20	5	1.1	0.9;
	8131	1	0	8.5	0	12.99	0	1.043422	-6.379294	150	9	1.1	0.9;
	8132	1	-0.19	-14.8	0	3.08	0	1.023685	43.657157	220	3	1.1	0.9;
	8133	1	-0	-0	0	0.17	0	0.999092	-12.216942	110	22	1.1	0.9;
	8134	1	0	-0	0.080305	-0.240916	0	1.033347	0.325046	220	11	1.1	0.9;
	8135	1	0	0	0	0.33	0	1.0361	25.930527	150	2	1.1	0.9;
	8136	1	0	-0	0	0.05	0	0.970328	-8.450182	150	9	1.1	0.9;
	8137	1	-0	0	0	173.790366	0	0.991126	-1.596968	380	5	1.1	0.9;
	8138	1	55.4	7.9	0	6.35	0	1.007204	28.534975	154	1	1.1	0.9;
	8139	1	118.6	41.5	0	0.58	0	1.013364	25.066455	150	8	1.1	0.9;
	8140	1	13.9	1.9	0	3.36	0	1.013087	29.534987	154	1	1.1	0.9;
	8141	2	0	0	0	0	0	1.026372	24.396837	11	1	1.1	0.9;
	8142	1	19.32	6	0	1.38	0	0.997551	-14.582917	110	12	1.1	0.9;
	8143	2	0	0	0	0	0	1.029542	-3.096329	20	6	1.1	0.9;
	8144	1	20.82	7.9	0	18	0	1.007809	-24.274205	220	4	1.1	0.9;
	8145	2	0	0	0	0	0	1.055063	40.300168	20	3	1.1	0.9;
	8146	2	0	0	0	0	0	1.053776	40.435567	20	3	1.1	0.9;
	8147	2	0	0	0	0	0	1.050978	39.760414	20	3	1.1	0.9;
	8148	2	0	0	0	0	0	1.050978	39.760414	20	3	1.1	0.9;
	8149	2	0	0	0	0	0	1.053776	40.435567	20	3	1.1	0.9;
	8150	1	19.52	7	0	1.73	0	1.044994	-2.361946	110	15	1.1	0.9;
	8151	2	0	0	0	0	0	1.051219	38.74676	20	3	1.1	0.9;
	8152	1	275.3	55.2	0.019768	17.370232	0	1.000571	-9.357387	220	11	1.1	0.9;
	8153	2	0	0	0	0	0	1.050978	39.760414	20	3	1.1	0.9;
	8154	2	0	0	0	0	0	1.051219	38.74676	20	3	1.1	0.9;
	8155	1	168.5	54.8	0	5.34	0	1.03518	30.602846	150	8	1.1	0.9;
	8156	1	172.8	11.3	0	8.61	0	1.063032	10.867657	220	3	1.1	0.9;
	8157	2	0	0	0	0	0	0.996935	21.950327	10.6	1	1.1	0.9;
	8158	1	0	0	0	14.53	0	1.069519	18.108418	220	3	1.1	0.9;
	8159	2	0	0	0	0	0	1.040138	32.45384	20	17	1.1	0.9;
	8160	1	8.3	1.2	0	4.2	0	0.996411	16.073394	154	1	1.1	0.9;
	8161	1	-0	0	0.06177	8.002243	0	0.998943	95.769564	380	1	1.1	0.9;
	8162	2	0	0	0	0	0	1.054462	48.756838	20	3	1.1	0.9;
	8163	2	0	0	0	0	0	0.992395	35.295433	11	1	1.1	0.9;
	8164	1	-0	0	0	12.61	0	1.019889	6.724749	220	1	1.1	0.9;
	8165	1	0	0	0	2.6	0	1.020439	35.495297	154	1	1.1	0.9;
	8166	1	-0	-0	0	2.95	0	1.02045	25.177237	154	1	1.1	0.9;
	8167	2	0	0	0	0	0	1.086148	2.994871	20	18	1.1	0.9;
	8168	1	0	0	0	0.17	0	0.972825	-28.705322	150	10	1.1	0.9;
	8169	1	83.4	11.9	0	1.4	0	0.990939	17.193294	154	1	1.1	0.9;
	8170	1	0	0	0	20.96	0	0.994021	-3.834903	380	1	1.1	0.9;
	8171	1	0	-0	0	0.74	0	1.036322	-8.15799	150	9	1.1	0.9;
	8172	1	34.14	12.4	0	0.4	0	1.040945	-3.367	110	19	1.1	0.9;
	8173	2	0	0	0	0	0	1.054462	48.756838	20	3	1.1	0.9;
	8174	1	23.31	-6.67	0	2.15	0	0.99236	88.032583	150	1	1.1	0.9;
	8175	1	-0	-0	0	7.17	0	1.054662	30.426567	220	5	1.1	0.9;
	8176	1	0	0	0.019018	15.280983	0	1.072959	74.686739	380	3	1.1	0.9;
	8177	1	0.3	10.5	0.059307	6.200693	0	1.021373	-15.288951	220	4	1.1	0.9;
	8178	1	76.6	10.9	0	3.49	0	1.007637	20.568351	154	1	1.1	0.9;
	8179	2	0	0	0	0	0	0.997286	47.055099	14.4	1	1.1	0.9;
	8180	1	0	0	0	0.59	0	0.993632	-5.154689	110	18	1.1	0.9;
	8181	1	48.8	6.9	0	6.29	0	1.014524	16.614362	154	1	1.1	0.9;
	8182	1	0	-0	0	2.42	0	1.055819	25.808549	220	5	1.1	0.9;
	8183	1	225.2	35.2	0	11.99	0	1.044682	-6.242353	220	5	1.1	0.9;
	8184	1	38.04	3.4	0	3.43	0	1.028869	-21.802606	220	4	1.1	0.9;
	8185	1	0	0	0	0	0	1.013946	-14.578261	20	6	1.1	0.9;
	8186	1	0	-0	0.181549	16.110889	0	1.037828	-1.327147	220	18	1.1	0.9;
	8187	1	301.6	122.3	0.116416	34.512688	0	1.044665	38.20133	380	14	1.1	0.9;
	8188	1	6.51	3.7	0	0.24	0	0.970886	-7.359378	110	18	1.1	0.9;
	8189	1	158.3	-20.5	0	2.07	0	1.017415	11.420798	220	3	1.1	0.9;
	8190	1	0	0	0	0	0	1.06045	-6.223335	150	9	1.1	0.9;
	8191	2	0	0	0	0	0	1.050901	18.201002	20	7	1.1	0.9;
	8192	1	513.4	199.8	0	90.272366	0	1.022252	-1.200424	380	6	1.1	0.9;
	8193	2	0	0	0	0	0	1.013271	19.830863	10.5	1	1.1	0.9;
	8194	1	152.7	6	0	0.36	0	1.070002	-2.899764	120	20	1.2	0.8;
	8195	1	0	0	0	80	0	1.059478	-5.258058	380	12	1.1	0.9;
	8196	1	13.72	4.7	0	0.22	0	0.9786	-5.693268	110	18	1.1	0.9;
	8197	1	-0	0	0	0.53	0	1.026084	-2.579099	220	5	1.1	0.9;
	8198	1	-4.43	-3.56	0	0.26	0	0.963234	-6.344118	110	18	1.1	0.9;
	8199	1	0	0	0	0	0	1.013946	-14.578261	20	6	1.1	0.9;
	8200	1	0	-0	0	0	0	1.018239	-19.841296	20	4	1.1	0.9;
	8201	1	-0	0	0	1.36	0	0.974036	-11.281582	110	21	1.1	0.9;
	8202	2	0	0	0	0	0	0.959478	45.430548	11	1	1.1	0.9;
	8203	1	-0	0	0	11.52	0	1.05014	25.794001	220	5	1.1	0.9;
	8204	1	105.5	31.1	0	9.63	0	0.986885	-15.737631	220	4	1.1	0.9;
	8205	2	0	0	0	0	0	1.042184	0.453797	20	18	1.1	0.9;
	8206	2	0	0	0	0	0	1.01282	1.386847	20	18	1.1	0.9;
	8207	1	27.46	5.87	0	0.8	0	0.96659	-5.094534	110	18	1.1	0.9;
	8208	1	50.56	8.1	0	8.02	0	1.029519	20.554688	220	5	1.1	0.9;
	8209	1	29.13	13.5	0	3.62	0	0.958018	-29.201992	220	6	1.1	0.9;
	8210	1	21.12	13.1	0	0.36	0	0.947288	-10.924648	150	9	1.1	0.9;
	8211	1	25.03	18.8	0	1.46	0	0.987721	33.847662	220	3	1.1	0.9;
	8212	1	0	0	0	0	0	1.042943	-7.588989	150	9	1.1	0.9;
	8213	1	1.38	-1.15	0	0.5	0	1.07209	1.461296	110	19	1.2	0.8;
	8214	1	0	0	0	4.06	0	0.963725	-9.838418	150	9	1.1	0.9;
	8215	1	-0	0	0	0.35	0	1.008023	7.053534	110	18	1.1	0.9;
	8216	1	64.33	-57.14	0	8.9	0	1.015887	34.293408	220	13	1.1	0.9;
	8217	1	0	-0	0	5.98	0	1.049377	3.839235	220	5	1.1	0.9;
	8218	2	0	0	0	0	0	1.041811	28.30558	20	5	1.1	0.9;
	8219	2	0	0	0	0	0	1.041811	28.30558	20	5	1.1	0.9;
	8220	2	0	0	0	0	0	1.041811	28.30558	20	5	1.1	0.9;
	8221	2	0	0	0	0	0	1.041811	28.30558	20	5	1.1	0.9;
	8222	2	0	0	0	0	0	1.041811	28.30558	20	5	1.1	0.9;
	8223	2	0	0	0	0	0	1.041811	28.30558	20	5	1.1	0.9;
	8224	2	0	0	0	0	0	1.041811	28.30558	20	5	1.1	0.9;
	8225	1	72.78	45.1	0	11.93	0	0.990878	-13.512721	150	9	1.1	0.9;
	8226	1	61.67	16.3	0	0.05	0	1.047959	30.05409	220	5	1.1	0.9;
	8227	1	-0	0	0	1.1	0	0.975802	-29.038137	220	10	1.1	0.9;
	8228	1	45.05	40	0	1.79	0	0.989489	31.955589	220	3	1.1	0.9;
	8229	1	26.43	21.7	0.081155	5.149134	0	0.982612	-23.764535	220	6	1.1	0.9;
	8230	1	0	-0	0.181156	1.859101	0	0.998349	-1.002676	220	10	1.1	0.9;
	8231	1	6.31	2.5	0	0.81	0	0.973304	-5.681095	110	18	1.1	0.9;
	8232	1	-0	-0	0	115.512343	0	0.988357	17.315704	380	5	1.1	0.9;
	8233	1	55.7	7.9	0	5.77	0	0.983939	28.475593	154	1	1.1	0.9;
	8234	1	0	0	0	18.66	0	1.040974	41.75659	380	3	1.1	0.9;
	8235	1	0	0	0	0	0	0.984449	-9.992776	150	9	1.1	0.9;
	8236	1	45.05	5	0	33.8	0	1.037105	2.47416	380	11	1.1	0.9;
	8237	1	96.59	7.28	0	2.52	0	1.020725	23.17778	220	5	1.1	0.9;
	8238	1	-0.14	-1.51	0	1.85	0	0.989115	-13.585363	110	22	1.1	0.9;
	8239	1	-0.21	-1.04	0	0	0	0.998394	26.560836	150	8	1.1	0.9;
	8240	1	198.71	60.3	0.232451	12.675977	0	1.032922	24.930845	220	13	1.1	0.9;
	8241	1	-0	0	0	0.58	0	1.049466	-3.512776	220	23	1.1	0.9;
	8242	1	0	0	0	19.73	0	1.028214	5.959625	380	6	1.1	0.9;
	8243	1	-0	-0	0.125321	170.820083	0	0.989854	4.973943	380	5	1.1	0.9;
	8244	1	-27.43	9	0	0.28	0	0.974047	-4.969632	150	9	1.1	0.9;
	8245	1	30.03	9.9	0	0.22	0	0.987861	-10.819537	110	12	1.1	0.9;
	8246	2	0	0	0	0	0	1.024094	0.934183	11	1	1.1	0.9;
	8247	1	185.7	18.4	0	23.68	0	1.001207	-14.751255	220	10	1.1	0.9;
	8248	1	0	0	0	0	0	1.041516	-6.856455	150	9	1.1	0.9;
	8249	1	0	0	0.077486	0.462514	0	1.066577	30.679964	220	5	1.1	0.9;
	8250	1	0	-0	0	6.53	0	1.011444	46.587511	380	1	1.1	0.9;
	8251	1	124.8	34.7	0	3.66	0	1.039863	20.752002	220	5	1.1	0.9;
	8252	1	75.2	10.7	0	3.89	0	0.993565	42.531711	154	1	1.1	0.9;
	8253	2	0	0	0	0	0	1.010291	28.950447	20	18	1.1	0.9;
	8254	1	89.7	7.9	0.23657	2.722572	0	1.013443	-15.402449	220	6	1.1	0.9;
	8255	1	64.27	20.1	0	5.4	0	0.991767	-4.43535	220	6	1.1	0.9;
	8256	1	-0	0	0.040084	2.749916	0	1.017229	-19.936916	220	4	1.1	0.9;
	8257	1	25.03	1.4	0	0.26	0	0.979379	-5.99893	110	18	1.1	0.9;
	8258	1	43.25	15.7	0	0.37	0	1.025358	-4.388867	110	19	1.1	0.9;
	8259	1	139.7	9.7	0	12.61	0	0.972704	-13.911507	220	10	1.1	0.9;
	8260	1	67.27	1.5	0	16.46	0	1.025058	30.614922	220	3	1.1	0.9;
	8261	1	0	-0	0	41.27	0	1.024585	32.972112	380	8	1.1	0.9;
	8262	1	54.96	17.6	0	0.17	0	1.001838	-13.924947	110	22	1.1	0.9;
	8263	1	6.21	2.2	0	0.46	0	1.018306	-3.275258	110	19	1.1	0.9;
	8264	1	43.45	7	0.078324	13.181257	0	1.020098	6.744104	220	6	1.1	0.9;
	8265	1	25.13	8.8	0	0.44	0	0.979331	-8.607713	110	21	1.1	0.9;
	8266	1	77.89	11.2	0.038961	16.120779	0	1.002425	90.543113	150	1	1.1	0.9;
	8267	1	153.5	46	0	0	0	1.062023	13.37906	220	3	1.1	0.9;
	8268	1	198.1	33.9	0	22.82	0	1.020179	-26.326218	220	4	1.1	0.9;
	8269	1	-0	0	0	6.1	0	1.022265	29.224087	380	3	1.1	0.9;
	8270	1	0	0	0	64.07	0	0.982114	29.866731	400	1	1.1	0.9;
	8271	1	0	0	0	71.79	0	1.08802	1.365864	380	21	1.2	0.8;
	8272	1	34.74	16.3	0	33.38	0	1.006525	-15.938718	380	4	1.1	0.9;
	8273	1	52.26	19.6	0	2.64	0	0.994009	18.407286	220	5	1.1	0.9;
	8274	2	0	0	0	0	0	1.014321	25.016584	20	5	1.1	0.9;
	8275	2	0	0	0	0	0	1.014321	25.016584	20	5	1.1	0.9;
	8276	2	0	0	0	0	0	1.014321	25.016584	20	5	1.1	0.9;
	8277	1	14.12	5.2	0	1.44	0	1.030431	21.089276	220	5	1.1	0.9;
	8278	1	-0	0	0	0	0	1.016217	20.317809	20	5	1.1	0.9;
	8279	1	-0	-0	0	0	0	1.021606	-18.065799	20	4	1.1	0.9;
	8280	1	41.8	5.9	0	5.81	0	0.967096	31.751104	154	1	1.1	0.9;
	8281	1	0	0	0	0	0	1.064717	0.810879	220	18	1.2	0.8;
	8282	1	195.1	-14.2	0	12.89	0	1.023094	-17.062795	220	5	1.1	0.9;
	8283	1	-0	0	0.468397	8.309186	0	1.031827	1.976857	380	6	1.1	0.9;
	8284	1	-0	0	0	6.13	0	1.041322	11.945773	380	1	1.1	0.9;
	8285	1	-0	0	0	0.06	0	1.054641	7.773991	220	3	1.1	0.9;
	8286	1	40.24	14.4	0	1.52	0	0.989646	-1.533454	220	6	1.1	0.9;
	8287	1	23.33	4	0.180843	0.479438	0	0.988464	12.175779	220	5	1.1	0.9;
	8288	1	131.6	33.8	0	6.55	0	1.014387	-20.165566	220	4	1.1	0.9;
	8289	1	3.7	1.3	0	0.91	0	0.999876	-0.254373	110	21	1.1	0.9;
	8290	2	0	0	0	0	0	1.01208	8.274182	10.5	1	1.1	0.9;
	8291	1	11.81	4.2	0	1.07	0	0.988122	-11.257963	110	1	1.1	0.9;
	8292	1	-0	-0	0	0.59	0	1.037284	23.086	220	5	1.1	0.9;
	8293	1	31.94	4.4	0	0.29	0	0.99154	-14.522454	110	22	1.1	0.9;
	8294	1	20.9	3	0	5	0	0.973884	41.032516	154	1	1.1	0.9;
	8295	1	0	0	0	0.91	0	1.019144	-3.508016	110	3	1.1	0.9;
	8296	1	29.03	18	0	0.02	0	1.052651	-3.861667	150	9	1.1	0.9;
	8297	1	119.6	10.6	0	0.9	0	0.997834	-23.10233	220	6	1.1	0.9;
	8298	1	0	0	0.037238	0.032071	0	1.085507	-20.189527	120	6	1.2	0.8;
	8299	2	0	0	0	0	0	1.011227	39.453645	11	1	1.1	0.9;
	8300	1	32.24	11.7	0	0.85	0	1.042262	-4.737566	110	19	1.1	0.9;
	8301	2	0	0	0	0	0	1.029941	-14.211447	20	4	1.1	0.9;
	8302	1	0	0	0	0.51	0	1.007431	-12.204666	150	9	1.1	0.9;
	8303	1	30.5	4.3	0	2.99	0	1.001671	17.650724	154	1	1.1	0.9;
	8304	1	0.4	0.3	0	1.46	0	0.984603	-7.0462	110	18	1.1	0.9;
	8305	1	0	-0	0	32.63	0	1.035746	36.598507	380	3	1.1	0.9;
	8306	1	352.6	193.5	0	39.3	0	1.018658	15.730309	380	6	1.1	0.9;
	8307	1	-0	-0	0	8.78	0	1.036877	21.208834	220	5	1.1	0.9;
	8308	2	0	0	0	0	0	1.020649	22.846825	20	6	1.1	0.9;
	8309	1	183.07	44.65	0	2.77	0	1.046169	24.949362	220	5	1.1	0.9;
	8310	1	20.72	-2.5	0	0.7	0	0.999064	-4.33545	110	18	1.1	0.9;
	8311	1	15.52	4.1	0	0.63	0	0.996232	-10.906345	110	12	1.1	0.9;
	8312	1	89.51	-12.51	0	7.04	0	1.006809	89.114721	150	1	1.1	0.9;
	8313	1	-0	-0	0	21.85	0	1.016279	47.706943	380	3	1.1	0.9;
	8314	1	0	0	0.056559	13.942868	0	1.067887	88.947852	380	3	1.1	0.9;
	8315	1	-0	0	0	5.05	0	0.974937	-12.82364	220	12	1.1	0.9;
	8316	1	0	-0	0	2.57	0	1.033029	76.227372	220	3	1.1	0.9;
	8317	2	0	0	0	0	0	1.060899	41.530587	20	14	1.1	0.9;
	8318	1	337.6	-15.4	0	6.27	0	0.984251	3.036721	220	5	1.1	0.9;
	8319	2	0	0	0	0	0	1.060899	41.530587	20	14	1.1	0.9;
	8320	2	0	0	0	0	0	1.060899	41.530587	20	14	1.1	0.9;
	8321	2	0	0	0	0	0	0.998025	27.572524	154	1	1.1	0.9;
	8322	1	141.2	45.4	0	7.55	0	1.035815	19.548683	220	5	1.1	0.9;
	8323	2	0	0	0	0	0	1.03284	33.745209	20	5	1.1	0.9;
	8324	1	20.12	8.6	0	0.17	0	0.991644	-4.340533	110	18	1.1	0.9;
	8325	1	19.42	6.3	0	0.48	0	0.972389	-8.434989	110	18	1.1	0.9;
	8326	2	0	0	0	0	0	1.024762	17.972121	20	14	1.1	0.9;
	8327	1	0.5	-0	0.049772	8.070228	0	1.018135	-15.006207	220	4	1.1	0.9;
	8328	1	42.45	14	0	0.44	0	0.963709	-7.73773	110	21	1.1	0.9;
	8329	1	4.61	0.9	0	0.54	0	0.990854	-6.458169	110	18	1.1	0.9;
	8330	1	-0	0	0	1.67	0	0.964984	-9.807744	150	9	1.1	0.9;
	8331	1	-0	-0	0	0.61	0	1.04975	-4.156713	110	19	1.1	0.9;
	8332	1	-0	-0	0	0.8	0	1.025435	0.441157	220	21	1.1	0.9;
	8333	1	37.14	4.02	0	3.32	0	1.025297	28.126862	220	13	1.1	0.9;
	8334	1	0	-0	0	42.8	0	1.046256	4.635137	220	21	1.1	0.9;
	8335	1	34.9	5	0	5.14	0	1.03589	-1.009428	154	1	1.1	0.9;
	8336	1	161.3	10.04	0	4.38	0	1.072398	2.270159	220	5	1.1	0.9;
	8337	1	31.73	5.4	0.039366	4.720634	0	1.002439	-3.657013	110	18	1.1	0.9;
	8338	1	-0	0	0	8.13	0	1.017961	18.629278	220	5	1.1	0.9;
	8339	1	0	-0	0	2.69	0	1.015664	-1.705792	150	9	1.1	0.9;
	8340	1	0	-0	0	1.26	0	0.999009	27.91604	150	8	1.1	0.9;
	8341	1	38.64	13.8	0	0.11	0	1.002863	-1.281886	110	18	1.1	0.9;
	8342	2	0	0	0	0	0	1.043619	-8.436397	20	4	1.1	0.9;
	8343	2	0	0	0	0	0	1.028736	-6.917679	20	4	1.1	0.9;
	8344	1	0	0	0	1.55	0	0.970215	-11.106049	220	22	1.1	0.9;
	8345	1	-0	0	0	4.28	0	1.039095	5.770455	220	5	1.1	0.9;
	8346	1	-0	0	0	4.96	0	0.997426	42.314603	154	1	1.1	0.9;
	8347	1	13.72	4.5	0	0.16	0	0.987647	-10.834649	110	12	1.1	0.9;
	8348	1	71.98	28.8	0	4.88	0	0.990578	-0.513213	220	6	1.1	0.9;
	8349	1	371.2	73.5	0	39.869841	0	1.009482	-17.383207	380	6	1.1	0.9;
	8350	1	-0	0	0	0	0	1.033951	32.215255	220	5	1.1	0.9;
	8351	1	0	0	0	4.27	0	1.018539	22.129856	150	8	1.1	0.9;
	8352	2	0	0	0	0	0	1.03161	11.84227	11	1	1.1	0.9;
	8353	1	-928	-491.4	0	0.73	0	1.057947	39.34569	380	3	1.1	0.9;
	8354	2	0	0	0	0	0	1.028736	-6.917679	20	4	1.1	0.9;
	8355	1	174	47.7	0.177451	1.142124	0	1.010475	-13.381402	110	22	1.1	0.9;
	8356	1	175.6	25	0	20.78	0	0.964924	32.309672	400	1	1.1	0.9;
	8357	2	0	0	0	0	0	1.028736	-6.917679	20	4	1.1	0.9;
	8358	1	52.96	50.3	0	0.71	0	1.018305	-6.370445	110	1	1.1	0.9;
	8359	2	0	0	0	0	0	1.028038	4.18117	6.3	1	1.1	0.9;
	8360	1	0	0	0	1.81	0	1.044436	29.692046	220	5	1.1	0.9;
	8361	1	222.6	62.6	0	30.19	0	1.034693	37.528372	380	3	1.1	0.9;
	8362	1	0	-0	0	5.45	0	1.034369	-8.017111	150	9	1.1	0.9;
	8363	1	16.7	2.4	0	3.27	0	0.995774	31.504265	154	1	1.1	0.9;
	8364	1	210.7	30	0	116.98	0	0.998121	2.826748	400	1	1.1	0.9;
	8365	1	-0	-0	0	9.34	0	1.007013	-13.166796	150	9	1.1	0.9;
	8366	1	-0	0	0.446415	2.766588	0	1.080721	8.164837	380	18	1.2	0.8;
	8367	2	0	0	0	0	0	1.024544	18.974877	20	14	1.1	0.9;
	8368	1	0	-0	0	0.49	0	1.077945	-3.550211	150	9	1.1	0.9;
	8369	2	0	0	0	0	0	1.032797	5.614702	20	18	1.1	0.9;
	8370	2	0	0	0	0	0	1.040138	32.45384	20	17	1.1	0.9;
	8371	1	-0	0	0	2.23	0	1.03379	40.550643	380	2	1.1	0.9;
	8372	1	-0	0	0	4.95	0	1.03277	28.800169	220	1	1.1	0.9;
	8373	2	0	0	0	0	0	1.034215	9.728222	20	18	1.1	0.9;
	8374	1	99.4	4.3	0	3.69	0	1.016256	3.307379	220	5	1.1	0.9;
	8375	1	9.8	1.4	0	0.06	0	1.006049	20.814447	154	1	1.1	0.9;
	8376	1	176.4	17.2	0	18.37	0	0.974212	6.709329	220	6	1.1	0.9;
	8377	1	-0	-0	0	1.87	0	1.030205	-20.991033	220	4	1.1	0.9;
	8378	2	0	0	0	0	0	1.034215	9.728222	20	18	1.1	0.9;
	8379	1	0	0	0	22.54	0	0.999357	23.53834	380	3	1.1	0.9;
	8380	2	0	0	0	0	0	1.032797	5.614702	20	18	1.1	0.9;
	8381	1	-0	0	0	3.23	0	1.022856	-10.882134	150	9	1.1	0.9;
	8382	1	-0	0	0	0.44	0	1.066458	-4.513735	150	9	1.1	0.9;
	8383	1	190.5	70.9	0	21.3	0	1.016769	-19.113768	380	4	1.1	0.9;
	8384	1	38.34	12.5	0	1.59	0	0.996109	-3.698386	220	6	1.1	0.9;
	8385	2	0	0	0	0	0	1.023034	24.738464	11	1	1.1	0.9;
	8386	2	0	0	0	0	0	1.029709	20.473808	20	5	1.1	0.9;
	8387	2	0	0	0	0	0	1.00657	25.464337	20	24	1.1	0.9;
	8388	2	0	0	0	0	0	0.988817	21.215061	19	1	1.1	0.9;
	8389	2	0	0	0	0	0	1.026207	-11.793103	20	4	1.1	0.9;
	8390	1	-0	-0	0	0.36	0	1.055126	28.025271	220	1	1.1	0.9;
	8391	1	0	-0	0	8.92	0	1.025807	48.325393	380	3	1.1	0.9;
	8392	1	30.03	10	0	0.26	0	0.973403	-6.343919	110	21	1.1	0.9;
	8393	2	0	0	0	0	0	1.066894	2.513611	11	1	1.1	0.9;
	8394	2	0	0	0	0	0	0.962614	39.578299	15.75	1	1.1	0.9;
	8395	1	222.7	80.5	0	0	0	1.037317	25.872032	150	2	1.1	0.9;
	8396	2	0	0	0	0	0	1.020502	-7.965953	20	4	1.1	0.9;
	8397	2	0	0	0	0	0	1.021376	22.718616	20	5	1.1	0.9;
	8398	1	0	0	0	0	0	1.019587	-11.095675	150	9	1.1	0.9;
	8399	1	-0	0	0	2.16	0	0.989523	-4.645698	220	7	1.1	0.9;
	8400	1	0	-0	0	32.28	0	1.162819	1.018543	380	1	1.2	0.8;
	8401	1	0	-0	0	0.95	0	1.016604	-26.138576	220	4	1.1	0.9;
	8402	2	0	0	0	0	0	1.021376	22.718616	20	5	1.1	0.9;
	8403	1	-0	0	0	0	0	1.011027	-12.592409	20	4	1.1	0.9;
	8404	1	-0	0	1.044962	67.708163	0	1.03806	17.376838	380	5	1.1	0.9;
	8405	1	-0	-0	0	19.08	0	1.079461	-12.959093	150	9	1.2	0.8;
	8406	1	69.08	-49.6	0	3.42	0	1.031682	39.188537	110	8	1.1	0.9;
	8407	1	0	-0	0	0.94	0	1.027651	79.0077	220	3	1.1	0.9;
	8408	1	54.2	7.8	0	21.9	0	1.024365	-1.048273	154	1	1.1	0.9;
	8409	2	0	0	0	0	0	1.026822	-0.476726	20	18	1.1	0.9;
	8410	1	43.25	19.5	0.157072	16.902196	0	1.024407	-16.057704	220	4	1.1	0.9;
	8411	1	12.6	1.8	0	3.18	0	1.00377	23.582695	154	1	1.1	0.9;
	8412	1	0	0	0	0.35	0	1.009152	14.501185	220	5	1.1	0.9;
	8413	2	0	0	0	0	0	1.00358	37.337153	11	1	1.1	0.9;
	8414	1	4.4	1.1	0	0.41	0	1.019924	-8.595055	220	10	1.1	0.9;
	8415	2	0	0	0	0	0	1.013179	14.179628	20	18	1.1	0.9;
	8416	1	-0	0	0	0.85	0	0.982395	-5.498542	110	18	1.1	0.9;
	8417	1	-0	-0	0	0	0	0.994896	-3.136363	110	18	1.1	0.9;
	8418	1	40.95	13.4	0	51.25	0	0.993652	25.056908	150	8	1.1	0.9;
	8419	1	6.11	-0	0	1.17	0	1.012685	-20.115544	220	4	1.1	0.9;
	8420	1	0	-0	0	4.64	0	1.004597	91.103926	150	1	1.1	0.9;
	8421	2	0	0	0	0	0	1.019405	-14.419189	20	10	1.1	0.9;
	8422	1	124.6	15.9	0	2.29	0	1.013128	4.797777	220	7	1.1	0.9;
	8423	2	0	0	0	0	0	1.022087	-3.367486	20	6	1.1	0.9;
	8424	2	0	0	0	0	0	1.022087	-3.367486	20	6	1.1	0.9;
	8425	2	0	0	0	0	0	1.004419	18.615053	20	5	1.1	0.9;
	8426	1	-0	0	0	2.47	0	1.039204	27.724089	220	5	1.1	0.9;
	8427	1	0	0	0	0	0	1.025816	-8.26672	20	5	1.1	0.9;
	8428	1	0	-0	0	1.38	0	1.011997	-3.144432	220	5	1.1	0.9;
	8429	1	12.6	1.8	0	3.66	0	1.012148	30.893197	154	1	1.1	0.9;
	8430	1	0	0	0	0	0	1.025816	-8.26672	20	5	1.1	0.9;
	8431	1	-0	-0	0	0.65	0	1.026184	-11.181267	150	9	1.1	0.9;
	8432	1	0	-0	0	7.67	0	0.946254	7.550305	220	16	1.1	0.9;
	8433	2	0	0	0	0	0	1.022285	8.641504	11.5	1	1.1	0.9;
	8434	1	-0	-0	0	0.35	0	1.039004	17.120517	220	5	1.1	0.9;
	8435	1	0	0	0	0	0	1.025816	-8.26672	20	5	1.1	0.9;
	8436	2	0	0	0	0	0	1.007757	23.409003	11	1	1.1	0.9;
	8437	1	167.6	-9	0	0.39	0	0.975399	-27.696451	150	10	1.1	0.9;
	8438	1	15.82	4	0	0.7	0	0.973801	-28.252014	150	10	1.1	0.9;
	8439	2	0	0	0	0	0	1.045666	52.068402	20	3	1.1	0.9;
	8440	1	0	0	0	0	0	1.050455	44.07457	380	1	1.1	0.9;
	8441	2	0	0	0	0	0	1.023837	21.039483	20	5	1.1	0.9;
	8442	1	-0	0	0	0	0	1.013511	22.52849	20	5	1.1	0.9;
	8443	1	0	0	0	0	0	1.021734	-10.957359	150	9	1.1	0.9;
	8444	2	0	0	0	0	0	1.051219	38.74676	20	3	1.1	0.9;
	8445	2	0	0	0	0	0	1.039546	19.356134	20	5	1.1	0.9;
	8446	1	-7.91	2.6	0	1.72	0	1.039998	0.106626	150	9	1.1	0.9;
	8447	2	0	0	0	0	0	1.03758	22.028378	20	10	1.1	0.9;
	8448	1	0	-0	0.248071	0.263274	0	0.991247	18.669171	220	6	1.1	0.9;
	8449	1	-0	0	0	10.45	0	1.007791	36.976632	154	1	1.1	0.9;
	8450	1	32.04	14.9	0.019626	3.350374	0	1.013902	-2.752831	110	18	1.1	0.9;
	8451	2	0	0	0	0	0	0.992069	93.520525	20	3	1.1	0.9;
	8452	1	50.46	-10.8	0	42.99	0	1.099484	-2.340691	380	1	1.2	0.8;
	8453	2	0	0	0	0	0	0.972158	41.061568	21	1	1.1	0.9;
	8454	1	-0	-0	0.173635	2.565304	0	1.044067	35.638952	220	3	1.1	0.9;
	8455	1	-0	0	0	0.14	0	0.974207	-7.20883	110	18	1.1	0.9;
	8456	2	0	0	0	0	0	1.024324	-10.377063	20	4	1.1	0.9;
	8457	2	0	0	0	0	0	1.02611	9.397426	20	18	1.1	0.9;
	8458	1	0	0	0	4.12	0	1.001359	33.84971	220	3	1.1	0.9;
	8459	2	0	0	0	0	0	1.02611	9.397426	20	18	1.1	0.9;
	8460	1	190.6	51.6	0.215561	8.284439	0	1.035351	19.619226	220	5	1.1	0.9;
	8461	1	5.21	1.3	0	24.85	0	1.026561	-23.946661	380	10	1.1	0.9;
	8462	1	0	-0	0	93.78	0	0.956424	44.925894	400	1	1.1	0.9;
	8463	1	13.01	2.4	0	0.42	0	0.990698	-4.186367	110	18	1.1	0.9;
	8464	1	7	1	0	4.33	0	1.031881	17.767373	154	1	1.1	0.9;
	8465	1	0	0	0	0	0	1.000535	-9.681265	150	9	1.1	0.9;
	8466	1	343.1	109.5	0	41.33	0	1.035784	8.880294	380	16	1.1	0.9;
	8467	1	72	10.3	0	5.95	0	0.976034	39.492744	154	1	1.1	0.9;
	8468	1	0	-0	0	2.02	0	1.040393	84.017948	220	3	1.1	0.9;
	8469	1	96.81	19.4	0.059957	4.650043	0	1.013769	-16.62541	220	4	1.1	0.9;
	8470	2	0	0	0	0	0	1.020956	21.259248	20	6	1.1	0.9;
	8471	2	0	0	0	0	0	1.020956	21.259248	20	6	1.1	0.9;
	8472	1	145.7	-38	0.228661	28.338977	0	1.057594	44.556346	380	13	1.1	0.9;
	8473	2	0	0	0	0	0	1.020956	21.259248	20	6	1.1	0.9;
	8474	2	0	0	0	0	0	1.020956	21.259248	20	6	1.1	0.9;
	8475	1	0.9	28.3	0.288059	37.479951	0	1.058803	-12.07492	380	4	1.1	0.9;
	8476	2	0	0	0	0	0	1.000621	19.05802	11	1	1.1	0.9;
	8477	1	0	0	0.697821	27.39675	0	1.038248	-7.751154	380	17	1.1	0.9;
	8478	1	-0.18	-1	0	0	0	1.025548	33.220843	380	8	1.1	0.9;
	8479	2	0	0	0	0	0	1.029635	34.5439	20	5	1.1	0.9;
	8480	1	-91.6	-4.9	0.151656	24.068758	0	0.986955	-9.508168	220	10	1.1	0.9;
	8481	1	50.2	7.1	0	7.26	0	1.023818	16.867894	154	1	1.1	0.9;
	8482	1	87.7	22.2	0	7.35	0	1.024954	10.62805	220	7	1.1	0.9;
	8483	2	0	0	0	0	0	1.001419	42.731033	15.75	1	1.1	0.9;
	8484	1	0	0	0.02947	3.16053	0	1.022094	-9.15629	220	4	1.1	0.9;
	8485	1	0	0	0	0.68	0	1.017452	31.964092	154	1	1.1	0.9;
	8486	1	155.5	38.3	0	52.84	0	0.988967	23.628918	150	8	1.1	0.9;
	8487	1	53.86	17.7	0	0.33	0	0.945208	-8.532705	110	21	1.1	0.9;
	8488	2	0	0	0	0	0	1.042782	34.930401	11	1	1.1	0.9;
	8489	1	183.1	33.9	0	42.16	0	1.013726	-20.719686	380	4	1.1	0.9;
	8490	2	0	0	0	0	0	1.015531	4.163384	20	9	1.1	0.9;
	8491	1	-0	-0	0	0.86	0	0.906068	-11.914453	150	9	1.1	0.9;
	8492	1	-46.15	15.1	0	0.64	0	1.029156	0.076633	150	9	1.1	0.9;
	8493	2	0	0	0	0	0	1.024481	5.581028	11	1	1.1	0.9;
	8494	1	-4.73	4.23	0	0.42	0	1.033437	0.247819	110	23	1.1	0.9;
	8495	1	0	-0	0	3.32	0	1.017873	42.270591	150	8	1.1	0.9;
	8496	1	244.09	48.92	0	5.96	0	1.028917	19.248459	220	5	1.1	0.9;
	8497	1	-0	-0	0	132.047128	0	1.008448	20.598984	380	5	1.1	0.9;
	8498	1	222	90.5	0	0.16	0	0.98794	11.541561	220	5	1.1	0.9;
	8499	2	0	0	0	0	0	0.962508	28.579218	15	1	1.1	0.9;
	8500	1	-0	0	0.079222	9.330583	0	1.008519	-20.573227	220	6	1.1	0.9;
	8501	1	-0	0	0	8.78	0	1.041942	0.135841	154	1	1.1	0.9;
	8502	1	0	0	0	0	0	1.031135	4.071247	380	6	1.1	0.9;
	8503	1	-0	-0	0	0.19	0	1.001218	-5.42202	110	18	1.1	0.9;
	8504	1	-0	-0	0	8.58	0	1.033556	40.640833	380	8	1.1	0.9;
	8505	1	0	0	0.460713	22.766073	0	1.040967	10.592542	220	7	1.1	0.9;
	8506	1	26.93	9.5	0	7.65	0	1.041842	5.172442	220	7	1.1	0.9;
	8507	1	23.7	3.4	0	2.1	0	1.00832	26.187636	154	1	1.1	0.9;
	8508	1	129.4	18.4	0	24.68	0	0.985987	19.67774	400	1	1.1	0.9;
	8509	1	-0	0	0	15.97	0	1.041506	1.898415	220	11	1.1	0.9;
	8510	1	0	-0	0.29	-0.42	0	0.998909	45.144777	380	2	1.1	0.9;
	8511	1	47.3	6.7	0	2.54	0	1.001839	15.738688	154	1	1.1	0.9;
	8512	1	20.52	7.6	0	2.15	0	0.986787	-3.561117	110	18	1.1	0.9;
	8513	1	-0	-0	0	0	0	0.999015	14.71774	20	5	1.1	0.9;
	8514	1	0.9	0.1	0	8.74	0	1.009518	20.920501	154	1	1.1	0.9;
	8515	1	0	0	0	0	0	0.992463	89.041689	150	1	1.1	0.9;
	8516	1	-0	-0	0	0	0	0.999015	14.71774	20	5	1.1	0.9;
	8517	1	122.5	38.9	0	0.84	0	1.009875	21.502969	220	5	1.1	0.9;
	8518	1	18.42	1.2	0	0.12	0	1.007078	-3.740809	220	6	1.1	0.9;
	8519	2	0	0	0	0	0	1.026207	-11.793103	20	4	1.1	0.9;
	8520	1	-0	-0	0.47104	0.487094	0	1.058816	18.414025	380	14	1.1	0.9;
	8521	1	105.9	29.5	0.237858	25.422142	0	1.02718	17.886722	220	5	1.1	0.9;
	8522	1	78.39	21.2	0	0.97	0	1.033815	9.21886	220	5	1.1	0.9;
	8523	1	0	-0	0	0.2	0	1.023605	-10.708096	150	9	1.1	0.9;
	8524	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	8525	1	2.2	-0.48	0	3.47	0	0.998818	-9.928904	150	9	1.1	0.9;
	8526	1	0	-0	0	12.5	0	1.088895	-1.018806	380	9	1.1	0.9;
	8527	1	18.52	11.5	0	4.41	0	1.033485	-8.016898	150	9	1.1	0.9;
	8528	1	0	-0	0	0.1	0	1.004025	-9.891566	150	9	1.1	0.9;
	8529	1	0	-0	0.03816	38.09184	0	1.057292	76.888885	380	3	1.1	0.9;
	8530	2	0	0	0	0	0	1.024655	-11.090856	20	4	1.1	0.9;
	8531	1	0	-0	0	0	0	1.019474	-15.630869	20	4	1.1	0.9;
	8532	1	0	0	0	0	0	1.033485	-8.016898	150	9	1.1	0.9;
	8533	1	0	-0	0	0	0	1.019474	-15.630869	20	4	1.1	0.9;
	8534	1	0	-0	0	0	0	1.019474	-15.630869	20	4	1.1	0.9;
	8535	1	0	-0	0	0	0	1.019474	-15.630869	20	4	1.1	0.9;
	8536	1	19.82	5.8	0	4.99	0	1.016639	-11.771752	110	22	1.1	0.9;
	8537	2	0	0	0	0	0	1.032132	-3.910762	20	6	1.1	0.9;
	8538	2	0	0	0	0	0	1.015227	-3.436884	20	10	1.1	0.9;
	8539	1	-0	-0	0	0.4	0	1.002179	20.283789	154	1	1.1	0.9;
	8540	1	20.72	6.8	0	0.65	0	0.953343	-7.264906	110	21	1.1	0.9;
	8541	1	10.01	-30	0	0.74	0	1.016042	-11.166981	220	6	1.1	0.9;
	8542	1	55.26	34.2	0	0.51	0	0.963231	-9.827573	150	9	1.1	0.9;
	8543	1	145.7	122.2	0	4.5	0	1.02762	17.221656	220	3	1.1	0.9;
	8544	2	0	0	0	0	0	1.012697	95.47097	20	3	1.1	0.9;
	8545	2	0	0	0	0	0	1.012641	95.734391	20	3	1.1	0.9;
	8546	2	0	0	0	0	0	1.012641	95.734391	20	3	1.1	0.9;
	8547	2	0	0	0	0	0	1.028244	96.39747	20	3	1.1	0.9;
	8548	2	0	0	0	0	0	1.028244	96.39747	20	3	1.1	0.9;
	8549	2	0	0	0	0	0	1.028456	96.447506	20	3	1.1	0.9;
	8550	2	0	0	0	0	0	1.028456	96.447506	20	3	1.1	0.9;
	8551	1	0	-0	0	3.45	0	1.006366	-15.137227	150	9	1.1	0.9;
	8552	2	0	0	0	0	0	1.050589	87.905314	20	3	1.1	0.9;
	8553	1	28.83	-3.8	0	7.4	0	1.013905	-12.380472	110	22	1.1	0.9;
	8554	1	177.5	72.5	0	46.35	0	1.020669	36.645797	150	8	1.1	0.9;
	8555	2	0	0	0	0	0	1.000449	37.338871	13.8	1	1.1	0.9;
	8556	1	0	-0	0	132.84	0	1.019976	-17.782538	380	4	1.1	0.9;
	8557	1	39.84	13.8	0.058009	12.81166	0	1.01659	91.688856	150	1	1.1	0.9;
	8558	1	2	1	0	0.3	0	0.978984	-5.102198	110	21	1.1	0.9;
	8559	2	0	0	0	0	0	0.99654	26.163496	11	1	1.1	0.9;
	8560	1	56.53	15.4	0	0.12	0	1.061997	16.341333	110	24	1.1	0.9;
	8561	2	0	0	0	0	0	0.992794	44.481603	11	1	1.1	0.9;
	8562	1	-0	0	0	0	0	1.013824	-24.041959	20	4	1.1	0.9;
	8563	2	0	0	0	0	0	1.020659	-5.050211	20	10	1.1	0.9;
	8564	1	-0	-0	0	0.71	0	1.071229	-5.679029	150	9	1.1	0.9;
	8565	1	38.54	12	0	16.52	0	1.014963	-26.893813	220	4	1.1	0.9;
	8566	1	0	-0	0	8.17	0	1.054641	7.773993	220	3	1.1	0.9;
	8567	1	0	-0	0	0.41	0	0.962886	-4.988351	110	19	1.1	0.9;
	8568	1	-0	0	0	0.67	0	1.009413	-9.559143	150	9	1.1	0.9;
	8569	2	0	0	0	0	0	1.043912	38.237051	20	5	1.1	0.9;
	8570	2	0	0	0	0	0	1.044271	38.24025	20	5	1.1	0.9;
	8571	2	0	0	0	0	0	1.043938	38.255003	20	5	1.1	0.9;
	8572	1	0	0	0.03922	0.76078	0	1.03074	18.715448	220	5	1.1	0.9;
	8573	2	0	0	0	0	0	1.040642	8.532255	20	14	1.1	0.9;
	8574	1	-0	0	0.056963	3.583037	0	1.074386	35.546719	380	14	1.1	0.9;
	8575	1	25.2	18.5	0.616465	136.433067	0	1.17806	3.460132	380	9	1.2	0.8;
	8576	2	0	0	0	0	0	1.04434	27.891337	20	14	1.1	0.9;
	8577	2	0	0	0	0	0	1.039495	25.225942	20	14	1.1	0.9;
	8578	2	0	0	0	0	0	1.039495	25.225942	20	14	1.1	0.9;
	8579	1	-0	-0	0.078399	4.4512	0	1.018943	6.634561	220	6	1.1	0.9;
	8580	1	22.4	3.1	0	3.4	0	1.00711	28.112169	154	1	1.1	0.9;
	8581	2	0	0	0	0	0	1.053599	21.197213	20	14	1.1	0.9;
	8582	2	0	0	0	0	0	1.053599	21.197213	20	14	1.1	0.9;
	8583	1	-0	0	0	93.07	0	1.052531	42.677501	380	8	1.1	0.9;
	8584	1	27.9	4	0	3.43	0	1.022818	35.669345	154	1	1.1	0.9;
	8585	1	0	-0	0	3.66	0	1.03042	-18.615687	220	4	1.1	0.9;
	8586	1	14.02	5.5	0	3.3	0	1.028806	19.359171	220	5	1.1	0.9;
	8587	1	26.03	10.6	0	2.14	0	1.029304	47.335419	110	8	1.1	0.9;
	8588	1	0	-0	0.116781	-0.097317	0	1.033776	23.366887	220	3	1.1	0.9;
	8589	1	0	0	0	45.02	0	1.018061	42.315747	150	8	1.1	0.9;
	8590	1	106.6	15.2	0	0.34	0	1.016112	20.652649	154	1	1.1	0.9;
	8591	1	0	0	0	0	0	1.002176	-13.481147	150	9	1.1	0.9;
	8592	1	0	0	0.040225	12.359775	0	0.996578	38.818651	220	3	1.1	0.9;
	8593	1	0	0	0	0	0	1.050151	43.874516	380	8	1.1	0.9;
	8594	1	16.72	5.5	0	0.89	0	0.990952	-14.211234	110	12	1.1	0.9;
	8595	2	0	0	0	0	0	1.003942	39.409116	11	1	1.1	0.9;
	8596	1	7.21	3.3	0	0.91	0	1.022763	-3.774086	110	15	1.1	0.9;
	8597	1	-0	-0	0	0.19	0	1.005676	-8.882179	150	9	1.1	0.9;
	8598	2	0	0	0	0	0	1.031898	-9.317211	20	6	1.1	0.9;
	8599	1	-0	-7.2	0	0	0	1.037505	-6.683276	150	9	1.1	0.9;
	8600	1	64.4	9.1	0	2.29	0	1.007712	17.817524	154	1	1.1	0.9;
	8601	1	196.3	-10.4	0	8.97	0	1.023757	15.821231	220	7	1.1	0.9;
	8602	1	33.24	11.1	0	1.56	0	0.951154	-12.421936	110	21	1.1	0.9;
	8603	1	0.1	-19.1	0	0.26	0	0.977623	30.982888	220	3	1.1	0.9;
	8604	2	0	0	0	0	0	1.052591	40.424368	20	3	1.1	0.9;
	8605	1	186.1	14.8	0	17.28	0	1.070477	22.577139	220	5	1.1	0.9;
	8606	1	17.52	6.5	0	0.74	0	1.025701	-3.549152	110	15	1.1	0.9;
	8607	1	-0	0	0	0.69	0	1.011387	-3.579207	110	21	1.1	0.9;
	8608	1	181.9	92.6	0	4.15	0	1.027615	8.124619	220	3	1.1	0.9;
	8609	2	0	0	0	0	0	1.052591	40.424368	20	3	1.1	0.9;
	8610	2	0	0	0	0	0	1.052591	40.424368	20	3	1.1	0.9;
	8611	1	42.55	14	0	7.52	0	1.010002	-12.911741	110	12	1.1	0.9;
	8612	1	0	0	0	9.84	0	1.037571	3.363381	220	3	1.1	0.9;
	8613	1	0	0	0	0.89	0	1.03523	7.175845	220	1	1.1	0.9;
	8614	1	25.63	6.5	0	0.52	0	1.035503	30.793495	220	5	1.1	0.9;
	8615	1	80.99	30.1	0	29.8	0	1.002828	0.450085	220	6	1.1	0.9;
	8616	1	94.1	39	0	27.61	0	1.042887	29.173811	380	3	1.1	0.9;
	8617	1	-0	0	0	4.34	0	1.012367	-8.878211	150	9	1.1	0.9;
	8618	1	-57.59	-30.52	0	2.45	0	1.04551	49.61108	110	8	1.1	0.9;
	8619	1	61.67	17.4	0	0.29	0	1.024857	-4.385691	110	19	1.1	0.9;
	8620	1	73.83	16.41	0	15.77	0	1.043895	3.24249	220	5	1.1	0.9;
	8621	1	-0	0	0	1.75	0	1.019496	22.104164	150	8	1.1	0.9;
	8622	1	0	0	0	33.03	0	1.015499	-22.709466	380	4	1.1	0.9;
	8623	1	0	0	0	30.48	0	1.062888	3.794097	400	1	1.1	0.9;
	8624	1	0	0	0	34.98	0	0.970458	32.668903	400	1	1.1	0.9;
	8625	1	26.43	6.6	0.01945	0.32055	0	1.016183	-11.264275	110	22	1.1	0.9;
	8626	1	74.08	10.7	0	13.55	0	1.014766	-26.891027	220	4	1.1	0.9;
	8627	1	29.03	18	0	0.02	0	1.003491	-10.282006	150	9	1.1	0.9;
	8628	1	0	0	0	0	0	0.980816	-7.972115	150	9	1.1	0.9;
	8629	1	-0	-0	0	3.98	0	1.026653	-20.337648	220	5	1.1	0.9;
	8630	1	0	0	0	1.44	0	0.983511	-8.815823	150	10	1.1	0.9;
	8631	1	0	0	0	1.39	0	1.020498	-7.004186	150	9	1.1	0.9;
	8632	2	0	0	0	0	0	1.128032	6.903218	20	9	1.2	0.8;
	8633	1	41.8	5.9	0	7.64	0	1.012585	28.941869	154	1	1.1	0.9;
	8634	1	0	-0	0	1.74	0	1.049966	90.439526	220	3	1.1	0.9;
	8635	2	0	0	0	0	0	0.968831	49.279117	11	1	1.1	0.9;
	8636	1	53.06	17.4	0	0.11	0	0.982346	-5.664036	110	21	1.1	0.9;
	8637	2	0	0	0	0	0	1.043665	14.930439	20	14	1.1	0.9;
	8638	1	22.3	3.1	0	0.54	0	1.023494	16.826765	154	1	1.1	0.9;
	8639	2	0	0	0	0	0	1.026319	-4.093349	20	6	1.1	0.9;
	8640	1	78.49	18.5	0	6.91	0	1.011847	-21.235399	220	4	1.1	0.9;
	8641	1	-0	0	0	37.02	0	0.965176	19.389202	400	1	1.1	0.9;
	8642	1	-0	-0	0	5.75	0	1.078046	7.07184	220	5	1.1	0.9;
	8643	1	0	-0	0	0.38	0	1.021366	-0.687341	150	9	1.1	0.9;
	8644	2	0	0	0	0	0	0.979015	43.42213	10.3	1	1.1	0.9;
	8645	1	0	-0	0	0	0	1.059723	-11.874145	150	9	1.1	0.9;
	8646	1	-0	-0	0	3.6	0	0.999314	44.425993	380	8	1.1	0.9;
	8647	2	0	0	0	0	0	1.052654	21.51137	20	5	1.1	0.9;
	8648	2	0	0	0	0	0	1.052654	21.51137	20	5	1.1	0.9;
	8649	1	-0	-0	0	15.14	0	1.051779	-5.83511	380	1	1.1	0.9;
	8650	1	0	0	0	0.65	0	1.080534	-3.780007	150	9	1.1	0.9;
	8651	2	0	0	0	0	0	1.024067	20.713541	20	10	1.1	0.9;
	8652	1	114.4	34.2	0.019661	4.460339	0	1.016464	14.858583	220	5	1.1	0.9;
	8653	1	-0	0	0.038809	40.691191	0	1.0434	77.613508	380	3	1.1	0.9;
	8654	2	0	0	0	0	0	0.984601	26.113187	6.3	1	1.1	0.9;
	8655	2	0	0	0	0	0	1.019231	22.327173	20	24	1.1	0.9;
	8656	2	0	0	0	0	0	1.012276	40.254811	11	1	1.1	0.9;
	8657	1	5.21	1.7	0	1.37	0	1.013385	-8.195491	110	12	1.1	0.9;
	8658	2	0	0	0	0	0	1.061235	43.554839	20	3	1.1	0.9;
	8659	1	15.12	7.5	0	0.33	0	1.013817	35.753916	220	3	1.1	0.9;
	8660	1	71.88	21.7	0	9.63	0	1.064477	3.338461	220	5	1.1	0.9;
	8661	1	-0	0	0	5.96	0	1.026147	-15.127436	220	4	1.1	0.9;
	8662	1	-0	0	0	1.4	0	1.058505	-0.088728	150	9	1.1	0.9;
	8663	1	57.06	13.3	0	19.18	0	1.029605	-21.145475	220	4	1.1	0.9;
	8664	1	160.7	40.3	0	3.04	0	0.997958	26.556133	150	8	1.1	0.9;
	8665	2	0	0	0	0	0	1.061235	43.554839	20	3	1.1	0.9;
	8666	1	26.03	-14.1	0	0.84	0	1.034807	-0.14975	110	23	1.1	0.9;
	8667	1	0	-0	0	0.55	0	1.02466	-8.239737	150	9	1.1	0.9;
	8668	2	0	0	0	0	0	1.060483	72.601224	20	3	1.1	0.9;
	8669	1	40.34	15.9	0	1.31	0	1.030836	31.47751	150	8	1.1	0.9;
	8670	1	-7.97	6.3	0	2.73	0	1.004627	-12.540227	150	9	1.1	0.9;
	8671	2	0	0	0	0	0	1.060483	72.601224	20	3	1.1	0.9;
	8672	1	106.7	33.6	0	3.28	0	1.044925	2.499152	110	19	1.1	0.9;
	8673	1	80.09	23.4	0.03907	0.62093	0	1.01216	7.425482	110	18	1.1	0.9;
	8674	1	-0	-0	0	0.18	0	1.044074	-6.730644	150	9	1.1	0.9;
	8675	2	0	0	0	0	0	1.031719	9.474961	20	5	1.1	0.9;
	8676	1	208.06	36.6	0	3.55	0	1.015423	15.004573	220	5	1.1	0.9;
	8677	1	209.8	102.6	0	24.31	0	0.985195	5.739458	220	3	1.1	0.9;
	8678	1	160	50	0	23.19	0	0.996665	-7.315796	220	6	1.1	0.9;
	8679	2	0	0	0	0	0	1.031719	9.474961	20	5	1.1	0.9;
	8680	2	0	0	0	0	0	1.031719	9.474961	20	5	1.1	0.9;
	8681	1	113.5	11.5	0	5.53	0	0.99442	-19.142553	220	4	1.1	0.9;
	8682	1	64.27	22.9	0	3.06	0	1.03304	-2.188961	110	19	1.1	0.9;
	8683	1	-0	0	0	11.87	0	1.073694	71.260391	380	3	1.1	0.9;
	8684	1	43.25	15.2	0.03938	0.18062	0	0.998764	-1.713308	110	18	1.1	0.9;
	8685	1	144.5	67.8	0	2.43	0	0.993084	32.616818	220	3	1.1	0.9;
	8686	2	0	0	0	0	0	1.027264	14.46311	20	14	1.1	0.9;
	8687	2	0	0	0	0	0	0.992442	9.617717	20	18	1.1	0.9;
	8688	1	62.57	22.7	0	0.47	0	1.012567	-5.484707	110	19	1.1	0.9;
	8689	1	-0	0	0	0.86	0	1.034054	-4.198425	150	9	1.1	0.9;
	8690	1	89.8	5.9	0.039177	15.480617	0	1.010868	93.234354	150	1	1.1	0.9;
	8691	1	99.5	28.1	0	6.23	0	1.022737	38.788607	380	3	1.1	0.9;
	8692	1	24.23	9.3	0.23794	6.201545	0	1.010946	-13.506969	220	4	1.1	0.9;
	8693	1	22.22	5.6	0.019675	0.800325	0	1.0179	3.935052	380	7	1.1	0.9;
	8694	1	148.2	65.3	0	10.51	0	1.00949	12.391585	380	3	1.1	0.9;
	8695	1	-0	-0	0.076637	2.763363	0	1.039945	14.048373	220	7	1.1	0.9;
	8696	1	0	0	0	0	0	1.023631	36.965167	150	8	1.1	0.9;
	8697	1	-0	-0	0.038657	0.031343	0	1.014343	-8.671556	110	12	1.1	0.9;
	8698	1	0	-0	0	6.35	0	1.031641	-4.160427	220	23	1.1	0.9;
	8699	1	-0	-0	0.228692	3.068952	0	1.024166	94.721888	380	1	1.1	0.9;
	8700	1	3.5	0.9	0	1.12	0	0.995748	-5.907958	110	18	1.1	0.9;
	8701	2	0	0	0	0	0	1.033214	-6.888934	20	6	1.1	0.9;
	8702	2	0	0	0	0	0	1.033214	-6.888934	20	6	1.1	0.9;
	8703	1	17.12	1.3	0	0.99	0	0.988671	-4.542598	110	18	1.1	0.9;
	8704	2	0	0	0	0	0	1.038438	0.065623	20	9	1.1	0.9;
	8705	1	90.5	12.9	0	2.99	0	1.000675	17.35251	154	1	1.1	0.9;
	8706	2	0	0	0	0	0	1.033214	-6.86015	20	6	1.1	0.9;
	8707	2	0	0	0	0	0	1.033214	-6.86015	20	6	1.1	0.9;
	8708	1	0	-0	0	0.29	0	1.010994	-1.162397	380	6	1.1	0.9;
	8709	2	0	0	0	0	0	0.982549	44.896548	11	1	1.1	0.9;
	8710	1	157.7	39.7	0.039543	53.100457	0	1.052339	14.454599	380	5	1.1	0.9;
	8711	2	0	0	0	0	0	1.044461	26.985271	20	5	1.1	0.9;
	8712	2	0	0	0	0	0	1.044461	26.985271	20	5	1.1	0.9;
	8713	2	0	0	0	0	0	1.044893	26.926518	20	5	1.1	0.9;
	8714	2	0	0	0	0	0	1.044893	26.926518	20	5	1.1	0.9;
	8715	1	198.6	63	1.058228	72.887005	0	1.102004	60.082984	380	3	1.2	0.8;
	8716	2	0	0	0	0	0	1.019095	-7.960677	20	4	1.1	0.9;
	8717	2	0	0	0	0	0	1.019095	-7.960677	20	4	1.1	0.9;
	8718	1	24.04	1.68	0	4.02	0	1.028086	15.549208	220	5	1.1	0.9;
	8719	2	0	0	0	0	0	1.022039	16.134374	20	2	1.1	0.9;
	8720	1	83.1	26.1	0	23.95	0	1.035548	-4.408789	220	23	1.1	0.9;
	8721	1	-0	-0	0	4.53	0	1.030956	18.955624	220	5	1.1	0.9;
	8722	1	241.4	116.9	0.713357	23.885167	0	1.015204	-22.86801	220	4	1.1	0.9;
	8723	1	408.9	77.3	0	13.58	0	1.010537	3.784995	220	7	1.1	0.9;
	8724	1	0	-0	0	2.75	0	1.033069	20.041193	220	5	1.1	0.9;
	8725	2	0	0	0	0	0	1.027819	-9.442028	20	4	1.1	0.9;
	8726	2	0	0	0	0	0	1.000784	-5.3448	20	18	1.1	0.9;
	8727	1	18.52	6.1	0	0.09	0	0.976129	-6.118497	110	21	1.1	0.9;
	8728	1	48.8	6.9	0	8.71	0	1.007795	20.535138	154	1	1.1	0.9;
	8729	1	0	0	0.170406	3.942552	0	1.033742	40.550873	380	2	1.1	0.9;
	8730	1	25.83	12.5	0	0.2	0	1.02783	19.279135	220	6	1.1	0.9;
	8731	2	0	0	0	0	0	1.0598	43.409328	20	3	1.1	0.9;
	8732	2	0	0	0	0	0	1.0598	43.409328	20	3	1.1	0.9;
	8733	2	0	0	0	0	0	1.0598	43.409328	20	3	1.1	0.9;
	8734	2	0	0	0	0	0	1.0598	43.409328	20	3	1.1	0.9;
	8735	2	0	0	0	0	0	1.0598	43.409328	20	3	1.1	0.9;
	8736	1	36.04	-8	0	32.73	0	1.042484	-8.329333	220	11	1.1	0.9;
	8737	1	0	-0	0.495152	-0.385118	0	1.077098	41.368237	380	8	1.1	0.9;
	8738	1	-0	0	0	0.45	0	0.993908	39.29719	154	1	1.1	0.9;
	8739	1	4.61	1.7	0	0.9	0	0.999104	-5.464238	110	19	1.1	0.9;
	8740	2	0	0	0	0	0	0.99389	98.19335	20	3	1.1	0.9;
	8741	2	0	0	0	0	0	0.988223	95.517768	20	3	1.1	0.9;
	8742	1	7.01	2.3	0	0.61	0	0.983111	-11.388547	110	12	1.1	0.9;
	8743	1	0	0	0	2.37	0	1.046536	-20.234575	380	10	1.1	0.9;
	8744	2	0	0	0	0	0	1.031165	-7.839877	20	4	1.1	0.9;
	8745	2	0	0	0	0	0	1.031165	-7.839877	20	4	1.1	0.9;
	8746	2	0	0	0	0	0	1.031165	-7.839877	20	4	1.1	0.9;
	8747	2	0	0	0	0	0	1.031165	-7.839877	20	4	1.1	0.9;
	8748	1	-0	0	0	50.27	0	1.110806	0.955681	380	9	1.2	0.8;
	8749	1	-0	-0	0.25	20.36	0	1.113386	5.807349	220	19	1.2	0.8;
	8750	1	191.96	38.57	0	122.89	0	1.057631	18.249732	380	14	1.1	0.9;
	8751	1	-0	0	0	219.65	0	0.970261	37.372194	400	1	1.1	0.9;
	8752	1	20.82	7.9	0	6.73	0	1.00853	-24.207466	220	4	1.1	0.9;
	8753	2	0	0	0	0	0	1.04914	42.65185	20	2	1.1	0.9;
	8754	1	-0	-0	0	1.35	0	1.068881	5.548581	220	5	1.1	0.9;
	8755	2	0	0	0	0	0	1.031165	-7.839877	20	4	1.1	0.9;
	8756	1	11.41	0.7	0	0.24	0	1.001498	6.023218	110	18	1.1	0.9;
	8757	1	-8.81	2.9	0	0.65	0	0.953282	-7.517012	150	9	1.1	0.9;
	8758	1	-0	-0	0.341219	69.634607	0	1.058786	43.17313	380	8	1.1	0.9;
	8759	2	0	0	0	0	0	1.032762	-8.550207	20	4	1.1	0.9;
	8760	1	-0	-0	0.538544	40.619042	0	1.041023	3.783242	380	11	1.1	0.9;
	8761	1	105.8	51.8	0	5.31	0	1.002114	-15.311194	150	9	1.1	0.9;
	8762	1	0.04	-0.05	0	20.75	0	1.026254	25.235043	220	14	1.1	0.9;
	8763	2	0	0	0	0	0	1.009892	4.868208	20	9	1.1	0.9;
	8764	2	0	0	0	0	0	1.030035	1.49544	11	1	1.1	0.9;
	8765	1	-0	0	0	9.72	0	1.019803	31.216132	380	3	1.1	0.9;
	8766	1	92.1	45	0	0.4	0	0.962993	-26.453508	220	6	1.1	0.9;
	8767	1	0	-0	0	65.82	0	1.096125	7.622014	380	1	1.2	0.8;
	8768	1	-0	-0	0	23.15	0	1.019648	37.982078	380	8	1.1	0.9;
	8769	1	27.53	17	0	0.86	0	1.007899	-8.721694	150	9	1.1	0.9;
	8770	2	0	0	0	0	0	1.048632	39.237883	20	2	1.1	0.9;
	8771	1	0	-0	0	10.52	0	1.030564	13.40168	380	1	1.1	0.9;
	8772	2	0	0	0	0	0	1.023158	23.406873	20	10	1.1	0.9;
	8773	2	0	0	0	0	0	1.042956	23.602891	20	14	1.1	0.9;
	8774	2	0	0	0	0	0	1.03036	-6.512869	20	4	1.1	0.9;
	8775	1	-0	-0	0	35.66	0	1.126856	-1.563266	380	9	1.2	0.8;
	8776	2	0	0	0	0	0	0.96764	46.597759	11	1	1.1	0.9;
	8777	1	0.1	-19.3	0	0.27	0	1.010273	7.100671	220	3	1.1	0.9;
	8778	1	0	-0	0	76.15	0	0.974291	22.734352	400	1	1.1	0.9;
	8779	1	-0	-0	0	5.5	0	1.001376	-10.405306	220	4	1.1	0.9;
	8780	1	9.01	3	0	0.51	0	1.033474	-4.183218	110	23	1.1	0.9;
	8781	1	0	0	0	0.37	0	0.995823	-9.109252	150	9	1.1	0.9;
	8782	1	-0.16	-1.92	0	0.45	0	1.032918	46.113189	110	8	1.1	0.9;
	8783	1	-0	0	0	0.05	0	1.051409	-6.429455	150	9	1.1	0.9;
	8784	1	61.27	11	0	50.16	0	1.026352	44.518276	380	3	1.1	0.9;
	8785	1	0	0	0	3.24	0	1.023876	37.245368	110	8	1.1	0.9;
	8786	1	23.03	7.6	0	0.78	0	0.970007	-14.00338	110	12	1.1	0.9;
	8787	1	48.8	6.9	0	12.07	0	0.98838	37.047366	154	1	1.1	0.9;
	8788	1	0	0	0	2.22	0	1.027371	13.364096	220	5	1.1	0.9;
	8789	1	0	0	0	2.06	0	0.982997	-8.492112	150	9	1.1	0.9;
	8790	1	23.03	9.2	0.058901	1.221099	0	1.005639	-1.201359	110	19	1.1	0.9;
	8791	2	0	0	0	0	0	1.011737	-19.669579	20	4	1.1	0.9;
	8792	1	-0	-0	0	0.19	0	1.012784	-1.969846	110	21	1.1	0.9;
	8793	2	0	0	0	0	0	1.032132	-0.658602	20	9	1.1	0.9;
	8794	1	0	0	0	0.84	0	1.045009	-6.785844	150	9	1.1	0.9;
	8795	1	22.32	13.8	0	2.96	0	1.031735	-7.53735	150	9	1.1	0.9;
	8796	1	-18.52	6.1	0	0.26	0	1.019317	-6.296032	150	9	1.1	0.9;
	8797	2	0	0	0	0	0	1.024762	17.972121	20	14	1.1	0.9;
	8798	1	21.62	9.8	0	0.14	0	0.991864	-4.264715	110	18	1.1	0.9;
	8799	1	43.55	-20.5	0	2.53	0	1.019944	-18.109195	220	4	1.1	0.9;
	8800	1	165.9	154.6	0	15.17	0	1.00888	-23.681101	380	4	1.1	0.9;
	8801	1	-57.46	-30.3	0	5.4	0	1.030578	-17.961902	220	4	1.1	0.9;
	8802	1	0	0	0	0.53	0	1.026693	-20.337925	220	5	1.1	0.9;
	8803	2	0	0	0	0	0	1.050841	90.684517	20	8	1.1	0.9;
	8804	1	20.02	6.8	0	0.48	0	1.021294	52.586969	110	8	1.1	0.9;
	8805	1	49.76	15	0	1.22	0	0.932912	-10.326204	110	21	1.1	0.9;
	8806	2	0	0	0	0	0	1.057091	77.565872	20	3	1.1	0.9;
	8807	2	0	0	0	0	0	1.057091	77.565872	20	3	1.1	0.9;
	8808	1	117.5	16.7	0	3.35	0	1.015754	19.136214	154	1	1.1	0.9;
	8809	2	0	0	0	0	0	1.057091	77.565872	20	3	1.1	0.9;
	8810	2	0	0	0	0	0	1.015185	-4.986575	20	4	1.1	0.9;
	8811	1	-0.2	-0.98	0	0	0	1.021214	26.174979	150	8	1.1	0.9;
	8812	1	60.57	39	0.585626	26.746283	0	1.005241	4.405336	220	19	1.1	0.9;
	8813	1	79.49	22.7	0	9.26	0	1.014813	31.102856	150	8	1.1	0.9;
	8814	1	6.61	1.9	0	0.67	0	0.967949	-7.321869	110	18	1.1	0.9;
	8815	1	-0	0	0	3.23	0	1.041813	19.321741	220	3	1.1	0.9;
	8816	2	0	0	0	0	0	1.036206	7.975321	15.75	1	1.1	0.9;
	8817	1	-0	-0	0	42.02	0	0.974981	35.747123	400	1	1.1	0.9;
	8818	1	5.61	2	0	0.96	0	0.938316	1.779955	110	19	1.1	0.9;
	8819	1	0	-0	0	0	0	1.032746	44.077076	150	8	1.1	0.9;
	8820	2	0	0	0	0	0	1.043492	-15.322718	20	4	1.1	0.9;
	8821	1	412.1	79.3	0	7.73	0	1.068183	2.586748	220	5	1.1	0.9;
	8822	1	4.11	1.4	0	0.4	0	1.012401	-7.614122	110	12	1.1	0.9;
	8823	1	19.32	7	0	0.39	0	1.021393	14.872458	110	19	1.1	0.9;
	8824	2	0	0	0	0	0	1.027152	23.121261	20	5	1.1	0.9;
	8825	1	0	0	0	1.43	0	1.057455	-6.036434	150	9	1.1	0.9;
	8826	1	47.65	19.4	0	0.79	0	1.027024	25.049298	150	8	1.1	0.9;
	8827	1	0	-0	0	1.73	0	1.037238	38.50001	110	8	1.1	0.9;
	8828	1	35.14	14.3	0	7.31	0	0.993697	-13.728539	110	22	1.1	0.9;
	8829	1	0	-0	0	1.14	0	1.025185	37.927142	150	8	1.1	0.9;
	8830	2	0	0	0	0	0	0.993492	29.317066	15	1	1.1	0.9;
	8831	1	15.02	9	0	0.01	0	1.008279	-9.719471	150	9	1.1	0.9;
	8832	1	13.31	2.8	0	0.88	0	0.998808	-3.513311	110	18	1.1	0.9;
	8833	1	61.36	23.66	0	0.16	0	0.995792	-4.79195	110	18	1.1	0.9;
	8834	1	-0	0	0	14.03	0	1.169501	2.546668	380	1	1.2	0.8;
	8835	1	25.73	9.3	0	0.59	0	0.975424	-3.142251	110	19	1.1	0.9;
	8836	1	-0	0	0	30.94	0	1.013955	15.110149	380	3	1.1	0.9;
	8837	2	0	0	0	0	0	0.988686	21.121822	10	1	1.1	0.9;
	8838	2	0	0	0	0	0	1.027152	23.121261	20	5	1.1	0.9;
	8839	2	0	0	0	0	0	1.055981	85.333831	20	3	1.1	0.9;
	8840	1	-0	0	0	-0.44	0	1.009626	-16.517753	380	6	1.1	0.9;
	8841	2	0	0	0	0	0	1.04706	-6.69158	20	4	1.1	0.9;
	8842	2	0	0	0	0	0	1.04706	-6.69158	20	4	1.1	0.9;
	8843	2	0	0	0	0	0	1.04706	-6.69158	20	4	1.1	0.9;
	8844	2	0	0	0	0	0	1.04706	-6.69158	20	4	1.1	0.9;
	8845	1	-0	0	0	33.08	0	1.03174	38.509818	380	8	1.1	0.9;
	8846	2	0	0	0	0	0	1.017859	24.98212	15	1	1.1	0.9;
	8847	1	0	0	0	0.5	0	0.967066	-14.754981	150	9	1.1	0.9;
	8848	1	0	0	0.304178	52.494371	0	0.976629	28.348694	380	2	1.1	0.9;
	8849	1	-14.58	-5.96	0	4.76	0	1.009499	4.528531	220	6	1.1	0.9;
	8850	1	-20.82	6.8	0	0.91	0	1.010576	-0.995047	150	9	1.1	0.9;
	8851	2	0	0	0	0	0	1.045213	9.250128	20	14	1.1	0.9;
	8852	2	0	0	0	0	0	1.024544	18.974877	20	14	1.1	0.9;
	8853	1	47.6	41.3	0	295.019105	0	1.142288	0.575038	380	9	1.2	0.8;
	8854	1	27.9	4	0	1.75	0	1.025034	17.055153	154	1	1.1	0.9;
	8855	1	2.7	0.9	0	1.49	0	0.989482	-10.614166	110	12	1.1	0.9;
	8856	1	0	-0	0	4.17	0	1.026603	13.234905	220	1	1.1	0.9;
	8857	2	0	0	0	0	0	1.004314	14.241295	13.8	1	1.1	0.9;
	8858	1	1.8	0.7	0	0.32	0	0.977994	-2.830646	110	19	1.1	0.9;
	8859	1	0	0	0	-1.896325	0	1.11383	8.533461	380	1	1.2	0.8;
	8860	2	0	0	0	0	0	1.045495	-3.944766	20	18	1.1	0.9;
	8861	2	0	0	0	0	0	1.045495	-3.944766	20	18	1.1	0.9;
	8862	1	0	0	0	1.25	0	0.985693	-4.971174	150	9	1.1	0.9;
	8863	1	160.2	89.9	0	0	0	1.021072	-0.687203	150	9	1.1	0.9;
	8864	2	0	0	0	0	0	1.052369	0.152148	20	18	1.1	0.9;
	8865	1	-0	0	0.361	33.735	0	1.039352	-14.878706	380	10	1.1	0.9;
	8866	2	0	0	0	0	0	1.052369	0.152148	20	18	1.1	0.9;
	8867	2	0	0	0	0	0	1.052369	0.152148	20	18	1.1	0.9;
	8868	1	5.81	1.4	0	0.2	0	0.988916	-1.928572	110	18	1.1	0.9;
	8869	1	6.01	1.5	0	0.73	0	0.983323	-5.042196	110	18	1.1	0.9;
	8870	1	111.5	147.5	0	26.54	0	1.034942	38.071181	380	3	1.1	0.9;
	8871	1	0	-0	0.072134	52.200985	0	1.011795	93.536574	380	1	1.1	0.9;
	8872	1	-0	-0	0	26.36	0	1.04969	43.665298	380	8	1.1	0.9;
	8873	2	0	0	0	0	0	1.028771	2.144472	20	6	1.1	0.9;
	8874	1	-0	0	0	22.6	0	0.95615	40.908516	400	1	1.1	0.9;
	8875	1	133.2	24.8	0	74.43	0	1.020138	10.175758	380	6	1.1	0.9;
	8876	1	-0	0	0	1.29	0	1.007733	27.889509	154	1	1.1	0.9;
	8877	1	27.9	4	0	3.97	0	0.984173	22.434717	154	1	1.1	0.9;
	8878	1	353	123.4	0.038547	42.441453	0	1.055387	32.270698	380	14	1.1	0.9;
	8879	1	26.23	8.6	0	6.14	0	0.998148	-8.920058	110	12	1.1	0.9;
	8880	1	0	5.9	0.0392	4.7208	0	1.034225	-11.583408	220	4	1.1	0.9;
	8881	2	0	0	0	0	0	1.008788	8.521149	11.5	1	1.1	0.9;
	8882	1	0	0	0	17.336728	0	1.015119	17.913244	220	24	1.1	0.9;
	8883	1	1.8	72	0.520071	44.685501	0	1.034988	-12.026258	380	4	1.1	0.9;
	8884	2	0	0	0	0	0	1.00657	25.464337	20	24	1.1	0.9;
	8885	2	0	0	0	0	0	1.037131	84.517509	20	3	1.1	0.9;
	8886	1	0	-0	0	0.45	0	1.027925	-7.904402	150	9	1.1	0.9;
	8887	2	0	0	0	0	0	1.010141	8.299456	11	1	1.1	0.9;
	8888	1	0	-0	0	10.05	0	1.040866	76.768598	380	3	1.1	0.9;
	8889	1	0	0	0	0	0	1.041421	-3.450344	150	9	1.1	0.9;
	8890	1	0	0	0	3.87	0	1.015899	42.071153	380	3	1.1	0.9;
	8891	1	-0	0	0.135493	28.29062	0	1.099261	-0.957248	380	21	1.2	0.8;
	8892	1	-0	-0	0	9	0	0.992276	28.977626	400	1	1.1	0.9;
	8893	2	0	0	0	0	0	0.994601	31.575691	11	1	1.1	0.9;
	8894	2	0	0	0	0	0	1.018066	-18.544118	20	4	1.1	0.9;
	8895	1	0.3	2	0	10.34	0	1.021922	25.65862	220	14	1.1	0.9;
	8896	2	0	0	0	0	0	0.996866	24.883283	10.6	1	1.1	0.9;
	8897	2	0	0	0	0	0	1.053937	40.267539	20	3	1.1	0.9;
	8898	2	0	0	0	0	0	1.039942	37.156863	20	3	1.1	0.9;
	8899	1	0	0	0	0	0	1.09911	-4.414302	150	9	1.1	0.9;
	8900	2	0	0	0	0	0	1.033776	8.182079	20	18	1.1	0.9;
	8901	2	0	0	0	0	0	1.010028	16.739138	20	5	1.1	0.9;
	8902	2	0	0	0	0	0	1.010028	16.739138	20	5	1.1	0.9;
	8903	2	0	0	0	0	0	1.010028	16.739138	20	5	1.1	0.9;
	8904	2	0	0	0	0	0	1.050006	39.688536	20	3	1.1	0.9;
	8905	2	0	0	0	0	0	1.050006	39.688536	20	3	1.1	0.9;
	8906	2	0	0	0	0	0	1.010028	16.739138	20	5	1.1	0.9;
	8907	2	0	0	0	0	0	1.050006	39.688536	20	3	1.1	0.9;
	8908	2	0	0	0	0	0	1.053937	40.267539	20	3	1.1	0.9;
	8909	2	0	0	0	0	0	1.053937	40.267539	20	3	1.1	0.9;
	8910	2	0	0	0	0	0	1.053937	40.267539	20	3	1.1	0.9;
	8911	2	0	0	0	0	0	1.033947	-9.228621	20	6	1.1	0.9;
	8912	2	0	0	0	0	0	1.019231	22.327173	20	24	1.1	0.9;
	8913	1	0	-0	0	0.14	0	1.030967	-6.801792	150	9	1.1	0.9;
	8914	2	0	0	0	0	0	1.024089	38.406443	20	10	1.1	0.9;
	8915	1	93.12	4.99	0	0.18	0	1.025031	16.970735	220	5	1.1	0.9;
	8916	2	0	0	0	0	0	0.984601	26.113186	6.3	1	1.1	0.9;
	8917	1	0	-0	0.059132	2.540868	0	1.014698	-2.72048	110	18	1.1	0.9;
	8918	1	1.3	3.8	0	0.09	0	0.954917	-5.07993	110	18	1.1	0.9;
	8919	1	0.7	1.5	0.119605	-0.089704	0	1.013094	-14.636462	380	6	1.1	0.9;
	8920	1	128.6	27.8	0	36.2	0	0.99609	-7.183362	220	6	1.1	0.9;
	8921	2	0	0	0	0	0	1.055881	11.701666	20	18	1.1	0.9;
	8922	1	-0	0	0	24.64	0	1.034538	60.163707	380	3	1.1	0.9;
	8923	1	176.75	51.7	0	39.79	0	0.982386	15.630758	380	13	1.1	0.9;
	8924	1	78.59	16	0	1.87	0	1.029315	-6.933752	220	5	1.1	0.9;
	8925	1	13.9	1.9	0	0.54	0	1.02373	8.679936	154	1	1.1	0.9;
	8926	1	45.28	15.27	0	7.88	0	1.027104	19.850995	220	5	1.1	0.9;
	8927	1	0	-0	0	0.05	0	0.966455	-9.685045	150	9	1.1	0.9;
	8928	1	10.11	2.5	0	0.18	0	1.013738	-0.854652	110	18	1.1	0.9;
	8929	1	0	-0	0	1.18	0	1.031022	26.637798	220	5	1.1	0.9;
	8930	1	55.7	7.9	0	11.74	0	1.016312	17.084852	154	1	1.1	0.9;
	8931	1	-0	-0	0.118007	-0.299556	0	1.034915	4.996779	220	18	1.1	0.9;
	8932	1	0	-0	0	1.79	0	0.977877	-8.635776	150	9	1.1	0.9;
	8933	2	0	0	0	0	0	1.011801	13.983947	11	1	1.1	0.9;
	8934	1	82.55	30.58	0	4.26	0	1.048895	58.647529	220	8	1.1	0.9;
	8935	1	221.25	72.56	0	8.23	0	1.018391	13.251652	220	5	1.1	0.9;
	8936	1	-0	0	0.038332	7.391668	0	1.052932	42.986036	220	3	1.1	0.9;
	8937	2	0	0	0	0	0	1.025724	-14.139236	20	4	1.1	0.9;
	8938	2	0	0	0	0	0	1.025724	-14.139236	20	4	1.1	0.9;
	8939	2	0	0	0	0	0	1.025724	-14.139236	20	4	1.1	0.9;
	8940	1	16.12	5.8	0	0.91	0	1.084411	0.591993	150	9	1.1	0.9;
	8941	2	0	0	0	0	0	1.019281	-1.243469	20	4	1.1	0.9;
	8942	2	0	0	0	0	0	1.016891	10.696134	20	5	1.1	0.9;
	8943	2	0	0	0	0	0	1.019281	-1.243469	20	4	1.1	0.9;
	8944	2	0	0	0	0	0	1.019281	-1.243469	20	4	1.1	0.9;
	8945	2	0	0	0	0	0	1.017845	-4.792355	20	10	1.1	0.9;
	8946	1	3.7	1.4	0	0.63	0	1.063102	1.190758	110	15	1.1	0.9;
	8947	1	55.8	7.9	0	11.5	0	0.998	10.001814	154	1	1.1	0.9;
	8948	1	21.32	7.3	0	0.6	0	1.028952	55.568264	110	8	1.1	0.9;
	8949	1	0	0	0	1.47	0	0.985807	-5.363422	150	9	1.1	0.9;
	8950	1	118.4	16.9	0	0.02	0	1.037844	-1.254072	154	1	1.1	0.9;
	8951	1	-0.2	-1	0	0	0	1.018463	25.30556	150	8	1.1	0.9;
	8952	2	0	0	0	0	0	1.037965	14.047316	20	5	1.1	0.9;
	8953	2	0	0	0	0	0	1.037965	14.047316	20	5	1.1	0.9;
	8954	2	0	0	0	0	0	1.037965	14.047316	20	5	1.1	0.9;
	8955	2	0	0	0	0	0	1.037965	14.047316	20	5	1.1	0.9;
	8956	2	0	0	0	0	0	1.062565	2.294369	20	5	1.1	0.9;
	8957	1	60.57	18.9	0	1.46	0	1.026453	5.550109	220	5	1.1	0.9;
	8958	2	0	0	0	0	0	0.999428	8.526955	20	22	1.1	0.9;
	8959	1	0	0	0.242897	-0.202414	0	1.013143	25.58747	380	6	1.1	0.9;
	8960	1	0	-0	0.038985	8.051015	0	1.025141	64.616553	220	3	1.1	0.9;
	8961	2	0	0	0	0	0	1.015042	-0.37136	20	18	1.1	0.9;
	8962	2	0	0	0	0	0	1.037965	14.047316	20	5	1.1	0.9;
	8963	2	0	0	0	0	0	1.038216	15.918534	20	5	1.1	0.9;
	8964	2	0	0	0	0	0	1.030057	8.282162	20	6	1.1	0.9;
	8965	2	0	0	0	0	0	1.030057	8.282162	20	6	1.1	0.9;
	8966	2	0	0	0	0	0	1.030057	8.282162	20	6	1.1	0.9;
	8967	2	0	0	0	0	0	1.030057	8.282162	20	6	1.1	0.9;
	8968	1	0	0	0.495568	-0.385442	0	1.092019	46.227822	150	8	1.2	0.8;
	8969	2	0	0	0	0	0	1.030988	2.394145	20	9	1.1	0.9;
	8970	1	60.67	5.9	0	5.32	0	1.020105	-23.013394	220	4	1.1	0.9;
	8971	1	109	28.3	0	0.55	0	1.034786	-3.940241	110	19	1.1	0.9;
	8972	1	42.55	6.5	0	3.7	0	1.020798	-25.911872	220	4	1.1	0.9;
	8973	1	2.2	0.5	0	0.5	0	0.951938	-28.831094	150	10	1.1	0.9;
	8974	1	58.8	8.4	0	5.06	0	1.006345	31.488089	154	1	1.1	0.9;
	8975	1	-0	-0	0	37.02	0	1.02966	70.488194	380	3	1.1	0.9;
	8976	2	0	0	0	0	0	1.011998	0.419036	20	18	1.1	0.9;
	8977	1	-0	0	0	0.87	0	1.003375	-10.294034	150	9	1.1	0.9;
	8978	1	162	4.8	0.019957	4.120043	0	1.001585	-16.688433	220	4	1.1	0.9;
	8979	1	76.89	35.1	0	1.9	0	1.046682	27.644328	150	8	1.1	0.9;
	8980	2	0	0	0	0	0	0.9829	25.033677	10	1	1.1	0.9;
	8981	1	11.51	3.8	0	7.14	0	1.015719	-12.837155	110	12	1.1	0.9;
	8982	1	205.3	27.8	0	8.22	0	1.015701	-33.369225	220	4	1.1	0.9;
	8983	1	116.2	16.5	0	2.4	0	0.998637	27.042734	154	1	1.1	0.9;
	8984	1	39	5.5	0	7.15	0	1.001162	15.958225	154	1	1.1	0.9;
	8985	1	0	0	0.009936	0.8	0	0.978934	-6.243605	110	18	1.1	0.9;
	8986	1	15.82	22	0.156398	7.542701	0	1.019717	-5.292666	220	4	1.1	0.9;
	8987	1	-0	-0	0	3.15	0	1.028252	17.788941	220	5	1.1	0.9;
	8988	1	0	-0	0	0.61	0	1.023959	-7.603703	150	9	1.1	0.9;
	8989	1	8.71	5.4	0	0.84	0	0.965388	-11.747495	150	9	1.1	0.9;
	8990	1	39.04	24.8	0	7.47	0	1.017656	42.270911	150	8	1.1	0.9;
	8991	1	15.32	5	0	0.76	0	1.001041	0.110249	110	21	1.1	0.9;
	8992	1	60.67	33.2	0	0.31	0	1.022687	6.310749	220	5	1.1	0.9;
	8993	1	0	0	0	2.57	0	1.026732	81.3069	220	3	1.1	0.9;
	8994	2	0	0	0	0	0	1.063361	65.621633	20	3	1.1	0.9;
	8995	1	-0	-0	0	0.67	0	1.027669	-4.85404	220	1	1.1	0.9;
	8996	1	67.98	54.6	0	3.39	0	0.966187	48.309261	220	3	1.1	0.9;
	8997	1	0	0	0	0	0	1.026038	-10.378066	20	4	1.1	0.9;
	8998	1	95.9	26.2	0	16.21	0	0.994355	-3.968896	220	6	1.1	0.9;
	8999	2	0	0	0	0	0	1.020283	-12.097744	20	4	1.1	0.9;
	9000	1	-0	-0.2	0	0.69	0	1.016745	6.269013	110	18	1.1	0.9;
	9001	1	160.2	88.7	0	0	0	1.016622	-0.631522	150	9	1.1	0.9;
	9002	1	-0	-0	0.25	2.3	0	0.990777	1.505378	220	19	1.1	0.9;
	9003	1	-0	0	0	2.45	0	1.021185	16.043363	220	5	1.1	0.9;
	9004	1	81.09	12.5	0	2.4	0	1.031854	8.82489	220	5	1.1	0.9;
	9005	1	0	-0	0	2.02	0	1.005176	18.349832	380	1	1.1	0.9;
	9006	1	0	-0	0.125869	4.727092	0	1.029944	4.331875	220	3	1.1	0.9;
	9007	1	119.1	22.3	0	2.71	0	0.992817	-23.197745	220	6	1.1	0.9;
	9008	1	383.6	-37.9	0	21.3	0	1.01039	-16.760711	380	6	1.1	0.9;
	9009	1	0	-0	2.332213	65.292136	0	1.028369	21.402713	380	5	1.1	0.9;
	9010	1	0	0	0	0	0	1.002482	-9.023078	150	9	1.1	0.9;
	9011	1	0	0	0	0	0	1.012722	-8.876216	150	9	1.1	0.9;
	9012	1	-0	0	0	4.96	0	1.030954	8.56318	220	5	1.1	0.9;
	9013	1	-0	0	0.202178	26.042403	0	1.047034	39.205614	380	3	1.1	0.9;
	9014	1	0	-0	0	0.32	0	0.989274	-8.954617	150	9	1.1	0.9;
	9015	1	67.28	-0.33	0	8.4	0	1.028789	19.384691	220	13	1.1	0.9;
	9016	2	0	0	0	0	0	0.96764	46.597759	11	1	1.1	0.9;
	9017	1	152.6	61.8	0	2.35	0	1.015328	36.145136	150	8	1.1	0.9;
	9018	1	-0	-0	0.039334	1.160333	0	1.013542	14.132282	220	5	1.1	0.9;
	9019	1	124.8	-14.7	0.03948	0.92052	0	1.015247	35.207066	220	13	1.1	0.9;
	9020	1	-0	0	0	9.38	0	1.058375	31.68334	380	1	1.1	0.9;
	9021	1	-0	0	0	0.29	0	0.96673	-9.641623	150	9	1.1	0.9;
	9022	1	45.75	27.5	0	4.37	0	1.029962	8.797799	220	7	1.1	0.9;
	9023	1	-6.31	-3.18	0	7.28	0	1.004322	63.287127	220	3	1.1	0.9;
	9024	2	0	0	0	0	0	1.02301	5.762905	13.8	1	1.1	0.9;
	9025	1	289.1	84.2	0	17.34	0	1.010136	-20.682871	220	4	1.1	0.9;
	9026	1	-0	-0	0	2.51	0	1.023087	15.977582	154	1	1.1	0.9;
	9027	1	16.72	5.5	0	0.65	0	0.996095	-11.375796	110	12	1.1	0.9;
	9028	2	0	0	0	0	0	1.044704	16.811399	20	5	1.1	0.9;
	9029	1	18.42	5.3	0	0.39	0	0.994316	-3.892453	110	18	1.1	0.9;
	9030	1	91	28.4	0	10.57	0	1.026796	5.578213	220	5	1.1	0.9;
	9031	1	4.3	1.5	0.038806	4.611194	0	1.046528	-13.149141	220	4	1.1	0.9;
	9032	1	0	-0	0	20.11	0	1.032811	40.060195	380	3	1.1	0.9;
	9033	2	0	0	0	0	0	1.046254	9.292953	20	6	1.1	0.9;
	9034	2	0	0	0	0	0	1.046254	9.292953	20	6	1.1	0.9;
	9035	2	0	0	0	0	0	1.01993	9.094543	20	22	1.1	0.9;
	9036	1	223.5	31.8	0	3.21	0	0.986493	0.727562	154	1	1.1	0.9;
	9037	1	0	0	0.05792	0.38208	0	1.028051	-4.835427	220	22	1.1	0.9;
	9038	1	21.02	7	0	0.42	0	0.950221	-8.591731	110	21	1.1	0.9;
	9039	1	59.37	12.7	0	22.18	0	1.016291	-26.376564	220	4	1.1	0.9;
	9040	1	-0	0	0	1.01	0	1.019938	-6.64007	150	9	1.1	0.9;
	9041	2	0	0	0	0	0	1.045359	10.101302	20	14	1.1	0.9;
	9042	1	63.07	13.4	0	15.52	0	1.015858	-26.579869	220	4	1.1	0.9;
	9043	1	0	0	0.361	25.935	0	1.044271	-13.760443	380	10	1.1	0.9;
	9044	1	209.9	38	0	5.26	0	0.965189	-23.699355	150	10	1.1	0.9;
	9045	2	0	0	0	0	0	1.060659	20.876838	20	5	1.1	0.9;
	9046	1	-2.91	-0.63	0	0.01	0	0.996141	-3.969869	220	6	1.1	0.9;
	9047	2	0	0	0	0	0	1.021906	-21.605261	20	4	1.1	0.9;
	9048	1	165.4	130.5	0	28.06	0	0.999464	16.144937	380	3	1.1	0.9;
	9049	2	0	0	0	0	0	0.968879	49.23979	11	1	1.1	0.9;
	9050	1	-218.34	-16.8	0.040234	16.999765	0	0.997133	-6.657802	220	10	1.1	0.9;
	9051	2	0	0	0	0	0	1.065482	13.792426	20	5	1.1	0.9;
	9052	1	19.32	7	0	0.65	0	1.009249	-1.513395	110	19	1.1	0.9;
	9053	1	2.6	0.5	0	0.14	0	0.997675	4.122377	110	18	1.1	0.9;
	9054	2	0	0	0	0	0	0.978477	21.737802	15	1	1.1	0.9;
	9055	1	17.52	6.1	0	0.53	0	1.029631	2.875298	110	19	1.1	0.9;
	9056	2	0	0	0	0	0	1.030388	1.487688	6.3	1	1.1	0.9;
	9057	1	126.7	46.1	0	5.58	0	1.022427	14.359655	220	3	1.1	0.9;
	9058	1	154.2	22	0	12.1	0	1.034663	-1.23355	154	1	1.1	0.9;
	9059	1	-4.82	-0.88	0	224.85	0	1.044536	18.08141	380	5	1.1	0.9;
	9060	2	0	0	0	0	0	1.065482	13.792426	20	5	1.1	0.9;
	9061	2	0	0	0	0	0	1.040642	8.532255	20	14	1.1	0.9;
	9062	1	74.98	10.5	0	12.96	0	1.017583	13.582942	220	5	1.1	0.9;
	9063	1	83.79	33.1	0	65.16	0	1.029324	30.921278	150	8	1.1	0.9;
	9064	1	0	-0	0	3.78	0	1.028171	33.067028	150	8	1.1	0.9;
	9065	1	14.02	5.6	0	2.34	0	1.039602	2.071227	220	5	1.1	0.9;
	9066	1	0	0	0.559639	19.414186	0	1.033897	4.794087	220	20	1.1	0.9;
	9067	1	0	-0	0	0.5	0	1.067926	11.58949	220	3	1.2	0.8;
	9068	1	-7.62	-2.52	0	33.185607	0	1.010119	29.593937	380	5	1.1	0.9;
	9069	1	0	0	0.36128	3.039004	0	1.021099	23.62848	380	6	1.1	0.9;
	9070	1	253.2	78.1	0	10.47	0	0.990743	11.266312	220	6	1.1	0.9;
	9071	1	48.8	6.9	0	3.06	0	0.982518	14.362004	154	1	1.1	0.9;
	9072	1	117	19.2	0	5.38	0	1.05886	21.876947	220	3	1.1	0.9;
	9073	1	-0	0	0	2.49	0	1.002952	-4.67636	150	9	1.1	0.9;
	9074	2	0	0	0	0	0	0.991243	91.841886	20	3	1.1	0.9;
	9075	2	0	0	0	0	0	1.044756	52.061269	20	3	1.1	0.9;
	9076	1	7.61	2.5	0	0.57	0	0.981999	-11.531273	110	1	1.1	0.9;
	9077	1	0	-0	0	2.18	0	1.021447	-10.622729	150	9	1.1	0.9;
	9078	1	13.72	4.5	0	0.91	0	0.976712	-11.424113	110	12	1.1	0.9;
	9079	1	0	-0	0	3.88	0	1.022048	-13.383541	380	6	1.1	0.9;
	9080	1	88.8	35.1	0	1.24	0	1.023597	31.928632	150	8	1.1	0.9;
	9081	2	0	0	0	0	0	1.042184	0.453797	20	18	1.1	0.9;
	9082	1	-0	-0	0	0.22	0	1.02606	18.466684	220	5	1.1	0.9;
	9083	1	231.9	27.6	0	0.36	0	1.039067	6.437524	220	5	1.1	0.9;
	9084	2	0	0	0	0	0	1.017679	10.590328	20	22	1.1	0.9;
	9085	1	-11.79	-22.64	0.093433	0.71958	0	1.099883	80.244023	380	3	1.2	0.8;
	9086	1	0	-0	0	1.28	0	0.937547	-11.56742	150	9	1.1	0.9;
	9087	1	-0	-0	0	58.996104	0	1.000613	17.219878	380	5	1.1	0.9;
	9088	2	0	0	0	0	0	1.039009	81.664701	20	3	1.1	0.9;
	9089	1	0	-0	0	0.8	0	1.02192	11.038401	220	1	1.1	0.9;
	9090	1	0	-0	0	1.14	0	1.00818	-8.516263	150	9	1.1	0.9;
	9091	1	88	70.3	0	36.55	0	1.022581	44.312899	380	3	1.1	0.9;
	9092	1	26.9	3.8	0	4.69	0	0.992338	26.531617	154	1	1.1	0.9;
	9093	1	26.43	5.5	0	1.03	0	0.993992	-16.264092	110	22	1.1	0.9;
	9094	1	-0	-0	0	0.98	0	0.978769	-8.124843	150	9	1.1	0.9;
	9095	1	0	0	0	0	0	1.004627	-12.540227	150	9	1.1	0.9;
	9096	1	0	-0	0	11.5	0	1.030819	21.136828	220	5	1.1	0.9;
	9097	1	49.25	12	0	7.28	0	1.030531	4.062415	220	5	1.1	0.9;
	9098	1	-0	-0	0.680765	0.501061	0	1.045325	35.806305	380	2	1.1	0.9;
	9099	1	227.6	71	0	8.59	0	1.023203	9.520695	220	3	1.1	0.9;
	9100	1	16.51	0	0	0.13	0	1.09631	-4.852232	150	9	1.1	0.9;
	9101	1	70.58	28.9	0.089366	0.450423	0	0.990956	-4.933793	110	18	1.1	0.9;
	9102	2	0	0	0	0	0	1.04328	33.92678	20	14	1.1	0.9;
	9103	1	85.33	14.98	0	4.44	0	1.025547	-0.136832	154	1	1.1	0.9;
	9104	1	0	-0	0	49.59	0	1.038741	26.945356	380	13	1.1	0.9;
	9105	1	0	-0	0	10.75	0	1.036304	-17.70087	220	4	1.1	0.9;
	9106	2	0	0	0	0	0	1.04328	33.92678	20	14	1.1	0.9;
	9107	2	0	0	0	0	0	1.04328	33.92678	20	14	1.1	0.9;
	9108	1	-0	-0	0	1.13	0	1.015908	42.071101	380	3	1.1	0.9;
	9109	1	0	0	0	0	0	1.03472	7.897462	20	7	1.1	0.9;
	9110	2	0	0	0	0	0	0.964368	44.185324	11	1	1.1	0.9;
	9111	1	-0	0	0	0.94	0	1.025478	-11.132082	150	9	1.1	0.9;
	9112	1	-0	0	0.058041	0.191959	0	1.019044	-0.260439	110	18	1.1	0.9;
	9113	1	-0	0	0	10.44	0	1.048151	0.703407	154	1	1.1	0.9;
	9114	1	88.41	34.42	0	3.45	0	1.028437	-16.535053	220	4	1.1	0.9;
	9115	1	-0	-0	0	2.44	0	1.028389	-21.95793	220	4	1.1	0.9;
	9116	1	0	-0	0.27075	34.52435	0	1.082787	9.081419	380	18	1.2	0.8;
	9117	1	-0	0	0	86.73	0	0.998796	-9.18697	380	5	1.1	0.9;
	9118	1	0	-0	0	0.3	0	0.99123	-4.961218	110	18	1.1	0.9;
	9119	1	-0.6	-1.58	0	0.22	0	1.087273	1.338482	380	21	1.2	0.8;
	9120	2	0	0	0	0	0	1.013208	-8.664527	20	10	1.1	0.9;
	9121	1	0	0	0	1.07	0	0.970403	-0.115981	110	18	1.1	0.9;
	9122	1	380.53	113.37	0	115	0	1.045872	23.94196	380	14	1.1	0.9;
	9123	2	0	0	0	0	0	1.032856	3.083067	20	18	1.1	0.9;
	9124	1	164.7	44.5	0	3.67	0	1.020008	0.167283	220	5	1.1	0.9;
	9125	1	128.8	29.3	0	26.74	0	1.030762	4.104283	220	5	1.1	0.9;
	9126	1	-0	-0	0	9.9	0	0.977664	-27.479082	220	10	1.1	0.9;
	9127	1	-0.01	-0	0	26.52	0	1.103591	-15.033378	150	9	1.2	0.8;
	9128	1	0	-0	0.16245	33.4651	0	1.082841	3.289791	380	18	1.2	0.8;
	9129	1	-0	-0	0	36.48	0	1.126202	-1.919233	380	9	1.2	0.8;
	9130	1	11.61	2.5	0	3.84	0	0.986673	-1.277013	110	19	1.1	0.9;
	9131	1	8.91	5.1	0	1.06	0	0.939246	-6.113394	110	18	1.1	0.9;
	9132	1	47.55	29.4	0	5.79	0	0.993879	-6.791088	150	9	1.1	0.9;
	9133	1	198.4	32.7	0	31.82	0	0.99674	13.168518	220	6	1.1	0.9;
	9134	1	45.05	12.7	0	1.73	0	0.998994	-9.854786	110	22	1.1	0.9;
	9135	1	12.81	4.7	0	0.64	0	0.988493	-4.159622	110	21	1.1	0.9;
	9136	1	0	0	0	0	0	1.005209	-3.078095	150	9	1.1	0.9;
	9137	2	0	0	0	0	0	1.033542	1.716976	20	6	1.1	0.9;
	9138	2	0	0	0	0	0	1.001252	-14.682022	20	10	1.1	0.9;
	9139	1	46.56	9.25	0	2.37	0	0.99927	-3.461027	110	18	1.1	0.9;
	9140	1	83.19	21.8	0	3.96	0	1.073901	24.12996	220	3	1.1	0.9;
	9141	2	0	0	0	0	0	1.031723	0.213022	20	6	1.1	0.9;
	9142	2	0	0	0	0	0	1.031723	0.213022	20	6	1.1	0.9;
	9143	1	91.3	18.1	0	3.05	0	1.027737	5.791688	220	5	1.1	0.9;
	9144	2	0	0	0	0	0	1.033534	-10.544863	20	4	1.1	0.9;
	9145	1	112.17	-11.06	0	13.68	0	1.023214	15.365927	220	13	1.1	0.9;
	9146	1	-0	0	0	2.28	0	1.029829	2.299759	154	1	1.1	0.9;
	9147	1	-0	0	0	96.62	0	1.045655	3.07576	154	1	1.1	0.9;
	9148	2	0	0	0	0	0	1.035363	12.970591	20	18	1.1	0.9;
	9149	2	0	0	0	0	0	1.021005	21.987746	20	14	1.1	0.9;
	9150	1	3.7	2.6	0	4.45	0	1.034355	10.2688	220	5	1.1	0.9;
	9151	2	0	0	0	0	0	0.943602	40.868468	10.6	1	1.1	0.9;
	9152	2	0	0	0	0	0	1.058618	17.704675	20	5	1.1	0.9;
	9153	1	125.9	-21.2	0	18.41	0	1.042367	64.847481	380	3	1.1	0.9;
	9154	1	12.11	4.4	0	1.26	0	0.987632	-15.556879	110	22	1.1	0.9;
	9155	1	18.62	6.1	0	2	0	1.049805	25.547514	220	5	1.1	0.9;
	9156	1	441.2	200.7	0	42.12	0	1.005991	13.874731	380	3	1.1	0.9;
	9157	1	122.8	37.2	0	37.13	0	1.019242	21.932927	150	8	1.1	0.9;
	9158	1	2.3	0.3	0	0.29	0	0.983962	-5.083708	110	18	1.1	0.9;
	9159	1	153.7	21.9	0	17.71	0	1.048472	0.586861	154	1	1.1	0.9;
	9160	1	15.22	4.9	0	0.18	0	0.98332	-5.104188	110	21	1.1	0.9;
	9161	1	77.49	24.1	0.039371	4.630629	0	1.002301	-3.67191	110	18	1.1	0.9;
	9162	1	28.03	58.8	0.463629	66.853939	0	1.060728	31.642799	380	13	1.1	0.9;
	9163	1	52.9	7.6	0	4.15	0	0.997036	26.425104	154	1	1.1	0.9;
	9164	1	0	0	0	0.19	0	0.902549	-12.121472	150	9	1.1	0.9;
	9165	1	-0.28	-5.54	0	0	0	1.072384	58.895099	220	8	1.1	0.9;
	9166	2	0	0	0	0	0	1.00081	22.506847	10.5	1	1.1	0.9;
	9167	1	-0	-0	0	18.64	0	1.041957	41.367634	380	8	1.1	0.9;
	9168	2	0	0	0	0	0	1.020547	8.547903	11	1	1.1	0.9;
	9169	2	0	0	0	0	0	1.014559	-1.055158	20	18	1.1	0.9;
	9170	1	100.1	44.5	0	3.28	0	0.998493	-23.037485	220	6	1.1	0.9;
	9171	1	0	0	0	0	0	1.026448	-3.53911	150	9	1.1	0.9;
	9172	1	-0	0	0.05765	26.341958	0	1.051242	80.128427	380	3	1.1	0.9;
	9173	1	19.12	11.9	0	1.58	0	1.020921	-7.474467	150	9	1.1	0.9;
	9174	1	51.96	22.8	0	0.6	0	0.98776	-4.689833	110	18	1.1	0.9;
	9175	1	91.8	8.3	0	2.44	0	1.013499	-26.262852	220	4	1.1	0.9;
	9176	1	0	-0	0	1.51	0	1.0329	65.528598	220	3	1.1	0.9;
	9177	1	0	-0	0	7.1	0	1.034423	26.077071	220	13	1.1	0.9;
	9178	1	0	-0	0.116795	5.242671	0	1.035644	60.585744	380	3	1.1	0.9;
	9179	2	0	0	0	0	0	1.033768	14.596978	20	14	1.1	0.9;
	9180	1	0	-0	0	3.19	0	0.996606	23.721288	154	1	1.1	0.9;
	9181	1	47.65	20.4	0	3.63	0	1.028249	15.997162	220	5	1.1	0.9;
	9182	2	0	0	0	0	0	1.016648	20.309778	20	5	1.1	0.9;
	9183	1	-0	-0	0.060538	4.359462	0	1.029866	4.939178	220	5	1.1	0.9;
	9184	2	0	0	0	0	0	1.020161	16.222206	20	5	1.1	0.9;
	9185	1	186.5	49.9	0	13.17	0	1.02712	-21.419208	220	4	1.1	0.9;
	9186	1	0	0	0	33.66	0	0.986167	31.910384	380	2	1.1	0.9;
	9187	1	90.5	12.8	0	2.68	0	0.996851	27.182321	154	1	1.1	0.9;
	9188	2	0	0	0	0	0	1.031838	-9.201101	20	6	1.1	0.9;
	9189	2	0	0	0	0	0	1.031838	-9.201101	20	6	1.1	0.9;
	9190	2	0	0	0	0	0	1.031838	-9.201101	20	6	1.1	0.9;
	9191	2	0	0	0	0	0	1.031838	-9.201101	20	6	1.1	0.9;
	9192	1	0	-0	0	0	0	0.989033	-4.933794	20	18	1.1	0.9;
	9193	1	5.21	1.7	0.077745	9.272255	0	1.011027	-5.468014	220	12	1.1	0.9;
	9194	1	20.02	8	0	5.4	0	1.018491	3.932101	220	3	1.1	0.9;
	9195	2	0	0	0	0	0	0.980209	32.361809	15.75	1	1.1	0.9;
	9196	1	0	0	0.274712	52.929035	0	1.036884	-20.880107	380	10	1.1	0.9;
	9197	1	-0	-0	0	3.41	0	1.010484	22.571023	220	6	1.1	0.9;
	9198	1	0	-0	0	2.37	0	1.023253	-17.309898	220	4	1.1	0.9;
	9199	1	84.59	-18.5	0	64.57	0	1.026686	-7.216911	380	6	1.1	0.9;
	9200	1	139.1	19.8	0	16.14	0	0.98717	3.88052	400	1	1.1	0.9;
	9201	1	-0.57	-1	0	2.27	0	1.053989	-6.618869	150	9	1.1	0.9;
	9202	1	-0	-0	0	0.44	0	1.02763	-5.734248	110	3	1.1	0.9;
	9203	1	49.96	15.6	0	2.66	0	0.98692	-23.433598	220	6	1.1	0.9;
	9204	2	0	0	0	0	0	1.056762	11.950776	10.5	1	1.1	0.9;
	9205	1	1.7	0.6	0	0.56	0	1.016688	-3.141725	110	19	1.1	0.9;
	9206	2	0	0	0	0	0	1.039006	38.999932	20	2	1.1	0.9;
	9207	1	127.9	3.7	0	41.47	0	1.03014	15.494692	380	6	1.1	0.9;
	9208	1	31.33	6.4	0	1.96	0	0.919168	-11.232303	110	18	1.1	0.9;
	9209	1	0	-0	0	2.13	0	1.055456	5.239412	220	5	1.1	0.9;
	9210	1	0	0	0	0	0	0.963667	-7.149819	150	9	1.1	0.9;
	9211	1	-4.94	-3.67	0	0.11	0	1.010289	-0.72505	110	18	1.1	0.9;
	9212	1	-0	-7.5	0	0	0	1.023343	-11.234733	150	9	1.1	0.9;
	9213	2	0	0	0	0	0	1.044317	35.206283	20	5	1.1	0.9;
	9214	2	0	0	0	0	0	1.011094	17.983711	20	5	1.1	0.9;
	9215	2	0	0	0	0	0	1.05775	16.377573	20	5	1.1	0.9;
	9216	2	0	0	0	0	0	1.019482	16.777108	20	5	1.1	0.9;
	9217	1	-0	-0	0	2.05	0	0.961704	-15.615679	220	10	1.1	0.9;
	9218	2	0	0	0	0	0	1.052164	7.333527	20	5	1.1	0.9;
	9219	2	0	0	0	0	0	1.043764	37.173377	20	5	1.1	0.9;
	9220	2	0	0	0	0	0	1.043764	37.173377	20	5	1.1	0.9;
	9221	1	-0	0	0	1.66	0	1.029753	16.35285	220	1	1.1	0.9;
	9222	2	0	0	0	0	0	1.044317	35.206283	20	5	1.1	0.9;
	9223	2	0	0	0	0	0	1.017385	23.570466	6.3	1	1.1	0.9;
	9224	2	0	0	0	0	0	1.044317	35.206283	20	5	1.1	0.9;
	9225	2	0	0	0	0	0	1.048036	22.181451	20	5	1.1	0.9;
	9226	1	95.71	19.6	0	20.2	0	0.986945	-15.393544	220	4	1.1	0.9;
	9227	2	0	0	0	0	0	0.989111	40.491338	11.5	1	1.1	0.9;
	9228	2	0	0	0	0	0	1.058394	46.810832	20	3	1.1	0.9;
	9229	2	0	0	0	0	0	1.058394	46.810832	20	3	1.1	0.9;
	9230	2	0	0	0	0	0	1.058394	46.810832	20	3	1.1	0.9;
	9231	2	0	0	0	0	0	1.058394	46.810832	20	3	1.1	0.9;
	9232	2	0	0	0	0	0	1.058394	46.810832	20	3	1.1	0.9;
	9233	2	0	0	0	0	0	1.058394	46.810832	20	3	1.1	0.9;
	9234	1	-0	0	0.090774	7.310444	0	1.061463	-0.278195	220	18	1.2	0.8;
	9235	1	-0	-0	0	1.05	0	0.999436	-9.864132	150	9	1.1	0.9;
	9236	2	0	0	0	0	0	1.032132	-0.658602	20	9	1.1	0.9;
	9237	2	0	0	0	0	0	1.0608	42.720858	20	3	1.1	0.9;
	9238	1	-0	-0	0.135375	10.042175	0	1.057471	5.471328	380	18	1.1	0.9;
	9239	1	0	-0	0.115496	-0.096247	0	1.036493	14.764614	110	18	1.1	0.9;
	9240	1	0	-0	0	0	0	1.019372	-15.361499	20	4	1.1	0.9;
	9241	1	0	-0	0	0	0	1.019372	-15.361499	20	4	1.1	0.9;
	9242	1	-0.2	-1.64	0	0	0	1.01757	30.154601	150	8	1.1	0.9;
	9243	1	0	-0	0	0	0	1.019372	-15.361499	20	4	1.1	0.9;
	9244	2	0	0	0	0	0	1.019266	-10.730307	20	4	1.1	0.9;
	9245	1	128.8	70	0	1.87	0	1.005781	34.978072	220	3	1.1	0.9;
	9246	2	0	0	0	0	0	1.030174	-14.576007	20	4	1.1	0.9;
	9247	2	0	0	0	0	0	1.030174	-14.576007	20	4	1.1	0.9;
	9248	1	0	0	0	1.99	0	0.960595	-9.912117	150	9	1.1	0.9;
	9249	1	116.1	30.2	0	4.38	0	0.995268	26.354915	150	8	1.1	0.9;
	9250	2	0	0	0	0	0	1.063388	34.050037	20	10	1.1	0.9;
	9251	1	13.9	1.9	0	2.05	0	1.036666	0.044572	154	1	1.1	0.9;
	9252	1	23	3.2	0	3.05	0	0.979058	22.311772	154	1	1.1	0.9;
	9253	1	9.7	1.4	0	7.89	0	1.016044	35.103553	154	1	1.1	0.9;
	9254	1	0	-0	0	0.68	0	0.971063	-9.097095	110	18	1.1	0.9;
	9255	1	0	0	0	0	0	1.014677	-8.666869	150	9	1.1	0.9;
	9256	2	0	0	0	0	0	1.030939	24.620583	20	5	1.1	0.9;
	9257	2	0	0	0	0	0	1.030939	24.620583	20	5	1.1	0.9;
	9258	2	0	0	0	0	0	1.030939	24.620583	20	5	1.1	0.9;
	9259	1	-0	-0	0.180933	0.399274	0	0.981849	-16.787288	150	10	1.1	0.9;
	9260	2	0	0	0	0	0	1.030874	24.601888	20	5	1.1	0.9;
	9261	2	0	0	0	0	0	1.030874	24.601888	20	5	1.1	0.9;
	9262	1	-0	0	0.038902	0.921098	0	1.03964	28.678963	220	5	1.1	0.9;
	9263	1	-0	0	0.097251	0.812749	0	1.059544	18.04116	220	5	1.1	0.9;
	9264	1	42.75	4	0	5.41	0	1.022708	-13.666083	220	4	1.1	0.9;
	9265	2	0	0	0	0	0	1.042062	24.663334	20	18	1.1	0.9;
	9266	1	0	-0	0.080014	-0.604253	0	1.000982	-22.656101	220	6	1.1	0.9;
	9267	1	92.76	-46.04	0	1.24	0	1.009946	3.890945	220	7	1.1	0.9;
	9268	1	245.51	46.8	0	4.89	0	1.003137	32.826235	220	3	1.1	0.9;
	9269	1	-0	0	0	8.26	0	1.063275	21.172593	380	5	1.1	0.9;
	9270	1	0	-0	0.039331	1.250334	0	1.021841	-1.962085	110	18	1.1	0.9;
	9271	1	7.21	2.4	0	1.2	0	0.968621	-5.442615	110	18	1.1	0.9;
	9272	1	0	0	0	2.27	0	1.043316	-18.000315	380	10	1.1	0.9;
	9273	2	0	0	0	0	0	1.015185	-4.986575	20	4	1.1	0.9;
	9274	1	279.5	104.7	0.438922	19.496498	0	1.055807	25.808993	220	5	1.1	0.9;
	9275	1	46.55	15.5	0	0.4	0	0.940948	-12.203927	110	21	1.1	0.9;
	9276	1	22.73	5.3	0	0.54	0	0.99528	-3.853654	110	18	1.1	0.9;
	9277	1	-0	0	0	3.25	0	1.042719	49.486952	110	8	1.1	0.9;
	9278	1	143.9	123.3	0.211368	60.936732	0	1.017398	31.813706	380	3	1.1	0.9;
	9279	2	0	0	0	0	0	1.046178	92.710304	20	3	1.1	0.9;
	9280	2	0	0	0	0	0	1.044756	52.061269	20	3	1.1	0.9;
	9281	2	0	0	0	0	0	1.073284	1.256236	20	9	1.1	0.9;
	9282	2	0	0	0	0	0	1.0136	-3.104621	20	10	1.1	0.9;
	9283	2	0	0	0	0	0	1.028417	-6.105187	20	6	1.1	0.9;
	9284	1	-0	-0	0	3.8	0	1.021743	-12.549569	110	12	1.1	0.9;
	9285	1	0	0	0.171996	8.536225	0	1.050718	4.154259	220	19	1.1	0.9;
	9286	1	32.04	15	0	0.15	0	1.009879	-7.120499	110	23	1.1	0.9;
	9287	1	55.46	14.1	0	0.92	0	1.026202	-20.972467	220	4	1.1	0.9;
	9288	2	0	0	0	0	0	1.012542	1.713537	20	18	1.1	0.9;
	9289	1	21.72	7.5	0.03975	1.86025	0	1.015096	12.342701	220	5	1.1	0.9;
	9290	1	0	-0	0	0.68	0	0.948888	-8.6077	110	21	1.1	0.9;
	9291	1	0.6	-0	0	0.6	0	0.992079	-2.215235	110	18	1.1	0.9;
	9292	2	0	0	0	0	0	1.046548	6.871525	11	1	1.1	0.9;
	9293	1	-0	0	0	0.23	0	1.065591	-5.496083	150	9	1.1	0.9;
	9294	2	0	0	0	0	0	1.061222	57.751527	20	8	1.1	0.9;
	9295	1	44.95	12.8	0	2.61	0	1.005322	29.905154	150	8	1.1	0.9;
	9296	1	430.8	107.9	0	11.64	0	0.960317	4.257884	220	5	1.1	0.9;
	9297	2	0	0	0	0	0	1.027348	16.396901	20	14	1.1	0.9;
	9298	1	9.01	6.4	0.040942	0.169058	0	0.964189	-8.826746	110	21	1.1	0.9;
	9299	1	74.68	10.8	0	5.82	0	1.014641	-26.811678	220	4	1.1	0.9;
	9300	1	105.5	15.1	0	3.5	0	1.002311	15.234306	154	1	1.1	0.9;
	9301	2	0	0	0	0	0	1.024891	10.992359	20	18	1.1	0.9;
	9302	1	13.41	3.4	0	0.72	0	0.980522	-6.641514	110	18	1.1	0.9;
	9303	1	24.03	8	0	0	0	1.053223	-7.960288	120	6	1.1	0.9;
	9304	1	12.94	4.9	0	0.15	0	0.971014	7.088734	110	18	1.1	0.9;
	9305	1	0	0	0	0	0	0.979908	-26.763384	220	10	1.1	0.9;
	9306	1	0	0	0	3.28	0	1.029753	-4.984284	220	1	1.1	0.9;
	9307	1	-0	0	0.11875	6.100938	0	1.011289	-24.599894	220	6	1.1	0.9;
	9308	1	13.31	4.4	0	1.51	0	0.988705	-8.008281	110	21	1.1	0.9;
	9309	1	5.51	2	0	0.66	0	1.051067	-0.144768	110	19	1.1	0.9;
	9310	1	-0	-0	0	1.72	0	1.030924	7.981131	220	5	1.1	0.9;
	9311	1	105.5	24.7	0	58.37	0	0.998366	-7.281753	220	6	1.1	0.9;
	9312	2	0	0	0	0	0	1.003305	-5.479609	20	4	1.1	0.9;
	9313	1	-0	0	0	6.34	0	0.996195	3.691746	400	1	1.1	0.9;
	9314	2	0	0	0	0	0	1.012031	-8.893616	20	10	1.1	0.9;
	9315	1	-0	0	0	0.35	0	1.027608	28.817016	220	5	1.1	0.9;
	9316	1	0	0	0	0	0	1.003585	-15.239367	150	9	1.1	0.9;
	9317	2	0	0	0	0	0	1.028283	16.586422	20	14	1.1	0.9;
	9318	1	325	75.8	0.080802	2.289198	0	1.020464	24.448718	220	14	1.1	0.9;
	9319	1	0	-0	0.223046	1.974129	0	1.083095	86.444403	380	3	1.2	0.8;
	9320	1	3.1	1.5	0.235344	1.033686	0	1.029358	47.037025	220	13	1.1	0.9;
	9321	2	0	0	0	0	0	1.033776	8.182079	20	18	1.1	0.9;
	9322	2	0	0	0	0	0	1.021055	-14.38773	20	4	1.1	0.9;
	9323	2	0	0	0	0	0	1.021055	-14.38773	20	4	1.1	0.9;
	9324	1	251	100.5	0.15739	9.43261	0	1.027226	29.406206	220	5	1.1	0.9;
	9325	1	11.61	7.2	0	16.18	0	0.976527	-15.468409	150	9	1.1	0.9;
	9326	1	11.31	3.7	0	0.98	0	0.976265	-11.501599	110	12	1.1	0.9;
	9327	2	0	0	0	0	0	1.015617	-6.080033	20	4	1.1	0.9;
	9328	1	118.4	16.9	0	27.1	0	1.021862	23.744834	154	1	1.1	0.9;
	9329	1	0	-0	0	0.52	0	1.071079	-4.189882	150	9	1.1	0.9;
	9330	1	246.2	119.1	0	31.72	0	0.973939	-29.205879	220	10	1.1	0.9;
	9331	1	75.43	10.49	0	7.91	0	1.062456	21.477259	220	5	1.1	0.9;
	9332	1	0	-0	0.284218	27.033151	0	1.051053	1.270275	220	6	1.1	0.9;
	9333	1	15.62	5.3	0	2.17	0	1.022666	55.287627	110	8	1.1	0.9;
	9334	1	159.9	47.4	0	3.65	0	1.028712	27.886448	220	5	1.1	0.9;
	9335	1	14.31	0.87	0	3.79	0	1.035217	25.954532	220	13	1.1	0.9;
	9336	1	20.9	3	0	1.96	0	1.012806	13.180306	154	1	1.1	0.9;
	9337	1	-0	0	0	3.86	0	1.013217	-10.429283	220	1	1.1	0.9;
	9338	1	-0	-0	0.079833	-0.059875	0	1.005645	-5.875986	220	6	1.1	0.9;
	9339	1	27.9	4	0	6.41	0	1.031495	-1.850016	154	1	1.1	0.9;
	9340	1	22.02	7.7	0	0.78	0	1.019096	-14.08161	110	22	1.1	0.9;
	9341	1	0	0	0	0.41	0	1.036382	8.668558	220	5	1.1	0.9;
	9342	2	0	0	0	0	0	1.024572	-13.995277	20	4	1.1	0.9;
	9343	2	0	0	0	0	0	1.029323	-3.215687	20	6	1.1	0.9;
	9344	1	16.52	3.3	0	0.37	0	0.998231	-17.176137	110	22	1.1	0.9;
	9345	2	0	0	0	0	0	1.028588	4.701247	20	18	1.1	0.9;
	9346	2	0	0	0	0	0	1.028588	4.701247	20	18	1.1	0.9;
	9347	2	0	0	0	0	0	1.028588	4.701247	20	18	1.1	0.9;
	9348	2	0	0	0	0	0	1.028588	4.701247	20	18	1.1	0.9;
	9349	2	0	0	0	0	0	1.028588	4.701247	20	18	1.1	0.9;
	9350	2	0	0	0	0	0	1.028588	4.701247	20	18	1.1	0.9;
	9351	2	0	0	0	0	0	1.028588	4.701247	20	18	1.1	0.9;
	9352	2	0	0	0	0	0	1.028588	4.701247	20	18	1.1	0.9;
	9353	1	74.32	11.08	0	2.08	0	1.009063	89.858413	150	1	1.1	0.9;
	9354	2	0	0	0	0	0	1.028283	16.586422	20	14	1.1	0.9;
	9355	1	80.09	19	0	1.88	0	1.019668	-2.710611	110	15	1.1	0.9;
	9356	1	278.1	95.2	0.037933	113.352067	0	1.068482	38.645676	380	13	1.1	0.9;
	9357	2	0	0	0	0	0	1.03036	-6.512869	20	4	1.1	0.9;
	9358	1	11.61	7.6	0	6.21	0	1.017583	-3.765671	220	5	1.1	0.9;
	9359	2	0	0	0	0	0	1.00657	25.464337	20	24	1.1	0.9;
	9360	1	154.7	48.2	0	5.49	0	1.015913	33.726089	150	8	1.1	0.9;
	9361	1	4	0.1	0	0.02	0	0.902529	-12.124125	150	9	1.1	0.9;
	9362	1	0	0	0	27.4	0	1.024298	-14.442769	220	4	1.1	0.9;
	9363	1	-0	-0	0	249.594232	0	1.110446	8.690534	750	24	1.2	0.8;
	9364	2	0	0	0	0	0	0.970854	29.381452	15.75	1	1.1	0.9;
	9365	1	44.05	18	0	0.88	0	1.027017	25.396175	150	8	1.1	0.9;
	9366	1	-0	0	0	1.66	0	1.027114	-8.230884	380	6	1.1	0.9;
	9367	1	62.7	9	0	1.55	0	0.996384	33.202949	154	1	1.1	0.9;
	9368	1	7.91	2.6	0	1.72	0	1.010068	-8.536738	110	12	1.1	0.9;
	9369	1	0	-0	0	0.04	0	0.992214	21.028002	154	1	1.1	0.9;
	9370	2	0	0	0	0	0	1.073284	1.256236	20	9	1.1	0.9;
	9371	1	0	-0	0	0.33	0	0.967068	-14.765518	150	9	1.1	0.9;
	9372	1	76.48	19.8	0	5.41	0	0.997458	13.119639	220	5	1.1	0.9;
	9373	1	2.5	1.3	0	1.01	0	0.984446	-0.108883	110	18	1.1	0.9;
	9374	2	0	0	0	0	0	1.042317	14.274354	20	14	1.1	0.9;
	9375	1	16.42	5.9	0	0.25	0	1.024292	-3.924291	110	19	1.1	0.9;
	9376	2	0	0	0	0	0	1.022192	20.047374	20	5	1.1	0.9;
	9377	2	0	0	0	0	0	1.022192	20.047374	20	5	1.1	0.9;
	9378	2	0	0	0	0	0	1.022192	20.047374	20	5	1.1	0.9;
	9379	2	0	0	0	0	0	1.022192	20.047374	20	5	1.1	0.9;
	9380	1	126.9	80.5	0	46.82	0	1.042328	18.688066	380	6	1.1	0.9;
	9381	1	10.44	2.73	0	0.26	0	0.972093	-7.21041	110	18	1.1	0.9;
	9382	1	57.86	23.2	0	2.08	0	0.961374	-28.168655	220	6	1.1	0.9;
	9383	2	0	0	0	0	0	1.022192	20.047374	20	5	1.1	0.9;
	9384	2	0	0	0	0	0	1.022192	20.047374	20	5	1.1	0.9;
	9385	1	39.64	11.9	0	2.86	0	0.996081	-14.048563	110	22	1.1	0.9;
	9386	1	36.34	15.2	0	0.3	0	1.02934	19.998755	220	5	1.1	0.9;
	9387	2	0	0	0	0	0	1.023156	37.450201	20	10	1.1	0.9;
	9388	1	170.4	81.9	0	59.64	0	1.047493	-7.198916	380	17	1.1	0.9;
	9389	2	0	0	0	0	0	0.985384	-0.999006	20	18	1.1	0.9;
	9390	1	-119.63	-2.7	0.122326	29.298062	0	0.977593	-10.732565	220	10	1.1	0.9;
	9391	1	67.67	16.4	0	3.7	0	1.034617	20.900749	220	5	1.1	0.9;
	9392	2	0	0	0	0	0	1.058231	41.885187	20	10	1.1	0.9;
	9393	2	0	0	0	0	0	1.012697	95.47097	20	3	1.1	0.9;
	9394	1	43.75	2.2	0	1.36	0	1.017521	32.87514	220	3	1.1	0.9;
	9395	2	0	0	0	0	0	1.020595	7.622879	11	1	1.1	0.9;
	9396	1	246.9	20.3	0	4.52	0	0.987038	34.043606	220	3	1.1	0.9;
	9397	1	0	0	0.674563	11.497864	0	1.067202	42.60671	380	3	1.1	0.9;
	9398	1	70.08	43.4	0	1.83	0	0.974625	-8.32398	150	9	1.1	0.9;
	9399	1	-0	-0	0	5.96	0	0.929197	-11.904661	150	9	1.1	0.9;
	9400	1	5.11	2.5	0	0.95	0	0.976688	-4.938655	110	18	1.1	0.9;
	9401	1	174.3	11.8	0	0.51	0	1.093949	24.561183	220	3	1.2	0.8;
	9402	1	17.25	4.26	0	0.21	0	0.964466	-5.279309	110	18	1.1	0.9;
	9403	1	-0	-0	0	2.26	0	1.019764	22.127907	150	8	1.1	0.9;
	9404	1	-0.18	-2.28	0	0	0	1.050897	44.076116	380	8	1.1	0.9;
	9405	1	0	-0	0	1.08	0	1.025041	-11.218198	150	9	1.1	0.9;
	9406	1	-13.26	-7.13	0	23.16	0	0.9926	-22.495997	220	6	1.1	0.9;
	9407	1	10.21	3.8	0	4.6	0	1.04016	31.253607	220	5	1.1	0.9;
	9408	2	0	0	0	0	0	1.074058	4.909711	20	20	1.1	0.9;
	9409	1	83.19	22.9	0	28.5	0	0.99566	17.494551	380	5	1.1	0.9;
	9410	1	93.1	43.7	0	2.7	0	1.021734	-10.957359	150	9	1.1	0.9;
	9411	1	0	-0	0.098479	0.231521	0	1.031195	12.084859	220	5	1.1	0.9;
	9412	1	0	0	0	6.03	0	1.027321	0.48095	220	5	1.1	0.9;
	9413	2	0	0	0	0	0	0.984601	26.113187	6.3	1	1.1	0.9;
	9414	1	1.4	0.5	0	0.47	0	1.004293	-8.36677	110	12	1.1	0.9;
	9415	1	-0	0	0.223468	1.103777	0	1.053522	92.905505	220	3	1.2	0.8;
	9416	1	0	-0	0	0	0	1.018239	-19.841296	20	4	1.1	0.9;
	9417	2	0	0	0	0	0	1.017845	-4.792355	20	10	1.1	0.9;
	9418	1	0	-0	0	0	0	0.986151	-7.335049	150	7	1.1	0.9;
	9419	1	69.78	35.9	0.712912	6.505612	0	1.021554	5.304081	220	19	1.1	0.9;
	9420	1	28.63	11	0	1.27	0	0.98623	-10.553231	110	22	1.1	0.9;
	9421	2	0	0	0	0	0	1.030419	1.643213	6.3	1	1.1	0.9;
	9422	2	0	0	0	0	0	1.011998	0.419036	20	18	1.1	0.9;
	9423	1	0	-0	0	0.64	0	1.036304	-8.162625	150	9	1.1	0.9;
	9424	1	-85.69	-2	0	21.28	0	1.038332	12.07119	220	7	1.1	0.9;
	9425	1	0	0	0	0	0	1.026665	32.869784	150	8	1.1	0.9;
	9426	1	-23.09	-46.44	0	0	0	1.036268	17.18762	220	3	1.1	0.9;
	9427	2	0	0	0	0	0	1.022889	18.879559	20	14	1.1	0.9;
	9428	1	47	6.7	0	6.12	0	1.00488	20.110868	154	1	1.1	0.9;
	9429	1	59.9	8.5	0	7.81	0	1.025297	17.011969	154	1	1.1	0.9;
	9430	1	163.3	70.3	0	5.2	0	1.021867	37.577612	150	8	1.1	0.9;
	9431	1	43.95	15.5	0	6.36	0	1.016597	-22.538977	220	4	1.1	0.9;
	9432	1	193.8	18	0	82.73	0	1.032675	4.70045	380	6	1.1	0.9;
	9433	1	1	-2.3	0	6.33	0	1.060742	0.198664	150	9	1.1	0.9;
	9434	1	77.19	23	0	63.4	0	1.04353	14.87029	380	16	1.1	0.9;
	9435	1	54.96	13.2	0	2.95	0	1.016767	-8.269117	110	22	1.1	0.9;
	9436	1	27.13	8.9	0	0.27	0	0.958996	-8.031065	110	21	1.1	0.9;
	9437	1	127.1	44.7	0.040242	4.329758	0	0.993276	18.960031	220	5	1.1	0.9;
	9438	2	0	0	0	0	0	1.022995	4.68997	20	6	1.1	0.9;
	9439	2	0	0	0	0	0	1.035797	10.227937	20	18	1.1	0.9;
	9440	2	0	0	0	0	0	1.022995	4.68997	20	6	1.1	0.9;
	9441	2	0	0	0	0	0	1.022995	4.68997	20	6	1.1	0.9;
	9442	2	0	0	0	0	0	1.022995	4.68997	20	6	1.1	0.9;
	9443	2	0	0	0	0	0	1.022995	4.68997	20	6	1.1	0.9;
	9444	1	0	-0	0	5.95	0	1.022579	22.997302	220	3	1.1	0.9;
	9445	1	8.31	3	0	0.56	0	1.041055	2.513615	110	19	1.1	0.9;
	9446	1	0	-0	0	11.67	0	1.014887	-26.770924	220	4	1.1	0.9;
	9447	1	20.52	13.3	0	0.17	0	1.031297	8.46293	220	5	1.1	0.9;
	9448	1	159.4	28.4	0	15.17	0	1.031816	-19.546233	220	4	1.1	0.9;
	9449	1	-9.78	-3.29	0	0.12	0	1.017607	-2.523511	110	18	1.1	0.9;
	9450	1	-0	0	0.056243	48.433131	0	1.077878	83.760207	380	3	1.2	0.8;
	9451	1	-0	0	0.119552	0.120336	0	1.001151	-10.708247	150	10	1.1	0.9;
	9452	1	-0.01	0	0	0	0	1.071012	58.767541	220	8	1.1	0.9;
	9453	1	71.68	14.8	0	13.66	0	0.998614	-23.017212	220	6	1.1	0.9;
	9454	1	-0	-0	0.060112	0.249888	0	0.980664	5.730494	110	18	1.1	0.9;
	9455	1	22.23	8	0	5.7	0	1.028904	-3.473096	220	21	1.1	0.9;
	9456	1	11.1	1.6	0	3.18	0	1.002908	25.414397	154	1	1.1	0.9;
	9457	2	0	0	0	0	0	0.984599	43.311569	14.4	1	1.1	0.9;
	9458	1	9.41	13.8	0	65.59	0	1.018556	8.310178	380	6	1.1	0.9;
	9459	1	3.2	-0.4	0	1.76	0	0.994575	-7.679701	110	18	1.1	0.9;
	9460	1	1.83	-5.69	0	5.49	0	1.084998	72.398199	380	3	1.2	0.8;
	9461	2	0	0	0	0	0	0.985642	34.609787	11	1	1.1	0.9;
	9462	1	25	3.6	0	3.86	0	0.983264	22.644761	154	1	1.1	0.9;
	9463	1	-0	-0	0	0.45	0	1.0159	-12.399971	150	9	1.1	0.9;
	9464	2	0	0	0	0	0	1.017638	-6.936102	20	4	1.1	0.9;
	9465	1	-0	-0	0	2.55	0	1.03099	77.708152	220	3	1.1	0.9;
	9466	2	0	0	0	0	0	1.018726	25.16552	11.5	1	1.1	0.9;
	9467	2	0	0	0	0	0	1.05433	78.959661	20	3	1.1	0.9;
	9468	1	0	0	0.249386	4.993711	0	1.002557	39.380309	380	3	1.1	0.9;
	9469	1	200.4	115.2	0	6.18	0	0.99394	-3.056318	220	7	1.1	0.9;
	9470	1	0	0	0	9.93	0	1.024403	6.31329	220	5	1.1	0.9;
	9471	2	0	0	0	0	0	1.027968	24.05088	20	5	1.1	0.9;
	9472	2	0	0	0	0	0	1.027968	24.05088	20	5	1.1	0.9;
	9473	2	0	0	0	0	0	1.03654	20.486651	20	5	1.1	0.9;
	9474	2	0	0	0	0	0	1.03654	20.486651	20	5	1.1	0.9;
	9475	2	0	0	0	0	0	1.066367	6.221208	20	5	1.1	0.9;
	9476	2	0	0	0	0	0	1.066367	6.221208	20	5	1.1	0.9;
	9477	2	0	0	0	0	0	1.042669	14.686826	20	5	1.1	0.9;
	9478	2	0	0	0	0	0	1.042101	20.088214	20	5	1.1	0.9;
	9479	2	0	0	0	0	0	1.042101	20.088214	20	5	1.1	0.9;
	9480	2	0	0	0	0	0	1.042101	20.088214	20	5	1.1	0.9;
	9481	2	0	0	0	0	0	1.042101	20.088214	20	5	1.1	0.9;
	9482	2	0	0	0	0	0	1.042101	20.088214	20	5	1.1	0.9;
	9483	1	0	0	0	1.99	0	1.038639	53.208541	220	8	1.1	0.9;
	9484	1	141.2	34.8	0	2.37	0	1.012004	15.477863	220	5	1.1	0.9;
	9485	1	-0	-0	0	23.58	0	0.987691	-10.438707	150	9	1.1	0.9;
	9486	2	0	0	0	0	0	1.05433	78.959661	20	3	1.1	0.9;
	9487	1	0	0	0.192988	25.798353	0	1.024988	40.267048	380	3	1.1	0.9;
	9488	1	-0	0	0.090774	1.940444	0	1.057403	6.33416	220	18	1.1	0.9;
	9489	1	0	-0	0	29.2	0	1.015585	-16.260044	380	1	1.1	0.9;
	9490	1	3.7	1.5	0.057734	0.922266	0	1.066603	2.162499	110	15	1.1	0.9;
	9491	1	2.1	122.3	0.724149	28.756715	0	1.020145	-18.154827	380	4	1.1	0.9;
	9492	1	-0	-0	0.713731	106.320569	0	1.010233	18.034291	380	3	1.1	0.9;
	9493	1	-0	-0	0.057981	20.043081	0	1.005977	92.476825	380	1	1.1	0.9;
	9494	1	18.36	-0.95	0	3.83	0	0.997129	90.192284	150	1	1.1	0.9;
	9495	1	0	-0	0	23.8	0	1.023452	23.736503	154	1	1.1	0.9;
	9496	1	-0.12	-1.74	0	0	0	1.098589	65.002382	380	8	1.2	0.8;
	9497	1	84.59	12.9	0	3.66	0	1.016832	-25.444147	220	4	1.1	0.9;
	9498	1	-0	0	0	17.82	0	1.072853	-2.234716	380	1	1.2	0.8;
	9499	1	174.6	48.9	0	7.46	0	1.003975	13.514344	380	3	1.1	0.9;
	9500	1	0	-0	0.297804	2.218894	0	1.001232	93.284203	380	1	1.1	0.9;
	9501	1	9.51	4.6	0	0.16	0	0.973768	-8.365086	150	9	1.1	0.9;
	9502	1	0	-0	0	8.2	0	1.03958	-14.080105	380	1	1.1	0.9;
	9503	1	0.9	0.5	0	0.9	0	1.016166	-7.266251	110	1	1.1	0.9;
	9504	2	0	0	0	0	0	0.991989	39.317054	13.8	1	1.1	0.9;
	9505	1	0	-0	0	24.38	0	1.044769	-13.80956	380	4	1.1	0.9;
	9506	1	-0	0	0.703357	4.645536	0	1.024599	47.939847	380	3	1.1	0.9;
	9507	1	0	-0	0	3.02	0	1.005416	32.987634	154	1	1.1	0.9;
	9508	2	0	0	0	0	0	1.017078	-0.332817	20	10	1.1	0.9;
	9509	1	-0	0	0	13.86	0	1.022066	-17.640449	220	4	1.1	0.9;
	9510	1	25.13	9	0	1.46	0	1.066218	-0.226637	110	15	1.1	0.9;
	9511	1	0	0	0.104186	55.436651	0	1.068565	-12.213472	380	10	1.1	0.9;
	9512	2	0	0	0	0	0	1.036065	-16.582357	20	4	1.1	0.9;
	9513	2	0	0	0	0	0	1.059317	37.132576	20	8	1.1	0.9;
	9514	1	-0	-0	0	2.64	0	0.997717	60.875011	220	1	1.1	0.9;
	9515	2	0	0	0	0	0	1.055847	34.276348	20	8	1.1	0.9;
	9516	1	4.61	2.6	0.038572	0.121428	0	1.025395	0.662428	110	19	1.1	0.9;
	9517	1	13.21	3.5	0	0.32	0	1.057335	6.795081	110	18	1.2	0.8;
	9518	1	-0.19	-4.38	0	0	0	1.055487	33.902432	150	8	1.1	0.9;
	9519	1	49.25	30.9	0.097351	12.97086	0	0.99636	-17.388796	220	10	1.1	0.9;
	9520	2	0	0	0	0	0	1.026025	16.98036	20	14	1.1	0.9;
	9521	2	0	0	0	0	0	1.042077	7.956295	20	14	1.1	0.9;
	9522	2	0	0	0	0	0	1.059317	37.132576	20	8	1.1	0.9;
	9523	2	0	0	0	0	0	1.053646	52.143481	20	8	1.1	0.9;
	9524	1	0	-0	0.26929	0.080552	0	1.013572	-9.004048	380	6	1.1	0.9;
	9525	2	0	0	0	0	0	1.053646	52.143481	20	8	1.1	0.9;
	9526	1	41.31	5.96	0	0.89	0	0.99505	88.716582	150	1	1.1	0.9;
	9527	2	0	0	0	0	0	1.045516	41.754111	20	10	1.1	0.9;
	9528	1	41.8	5.9	0	10.95	0	1.008317	28.003801	154	1	1.1	0.9;
	9529	1	141.5	20.1	0	5.35	0	0.999104	17.264377	154	1	1.1	0.9;
	9530	1	0	-0	0	0.22	0	1.013601	9.488938	220	5	1.1	0.9;
	9531	1	-0	-0	0	0.36	0	1.016597	-0.634223	150	9	1.1	0.9;
	9532	1	88.2	26.7	0	8.91	0	1.040599	2.681523	220	5	1.1	0.9;
	9533	1	20.02	9.6	0	0.17	0	1.018963	-9.214164	150	9	1.1	0.9;
	9534	1	68.28	39	0.19855	79.84755	0	1.04647	-16.339563	380	10	1.1	0.9;
	9535	1	13.92	4.3	0	1.1	0	0.959436	4.956416	110	18	1.1	0.9;
	9536	1	-0.18	-1	0	0	0	1.055662	33.904007	150	8	1.1	0.9;
	9537	1	20.32	8.6	0	0.94	0	0.951715	-11.635825	110	18	1.1	0.9;
	9538	1	0	-0	0.133519	0.166481	0	1.056779	25.791482	220	5	1.1	0.9;
	9539	1	-14.73	-9.85	0	0.51	0	0.997784	-3.020761	110	18	1.1	0.9;
	9540	2	0	0	0	0	0	1.052481	55.678412	20	8	1.1	0.9;
	9541	1	-0.19	-1.94	0	0	0	1.024164	36.959582	150	8	1.1	0.9;
	9542	2	0	0	0	0	0	1.011374	-9.719553	20	10	1.1	0.9;
	9543	1	20.32	-22	0	5.35	0	1.017654	1.64075	220	7	1.1	0.9;
	9544	1	0	-0	0	2.3	0	1.050629	24.719094	220	5	1.1	0.9;
	9545	1	30.43	12	0	0.42	0	0.988917	-2.520979	110	18	1.1	0.9;
	9546	1	24.83	9	0	0.24	0	1.0229	12.548502	110	19	1.1	0.9;
	9547	2	0	0	0	0	0	1.034547	81.296442	20	3	1.1	0.9;
	9548	1	0	0	0	0.74	0	0.99902	13.532872	154	1	1.1	0.9;
	9549	2	0	0	0	0	0	1.034547	81.296442	20	3	1.1	0.9;
	9550	2	0	0	0	0	0	1.034547	81.296442	20	3	1.1	0.9;
	9551	2	0	0	0	0	0	1.034547	81.296442	20	3	1.1	0.9;
	9552	2	0	0	0	0	0	1.010522	6.873688	11.5	1	1.1	0.9;
	9553	2	0	0	0	0	0	0.948759	58.410077	14	1	1.1	0.9;
	9554	2	0	0	0	0	0	0.962462	44.453484	6.3	1	1.1	0.9;
	9555	1	100.1	48.9	0	0.11	0	1.070982	-4.197914	150	9	1.1	0.9;
	9556	2	0	0	0	0	0	1.012871	-6.113908	20	10	1.1	0.9;
	9557	1	-3.51	-0.52	0	0.19	0	1.005257	-1.228322	110	18	1.1	0.9;
	9558	1	446.5	105.6	0	38.83	0	0.997177	-10.309572	380	6	1.1	0.9;
	9559	1	42.85	14.3	0	0.1	0	1.012641	-13.548608	110	22	1.1	0.9;
	9560	1	0	0	0	31.19	0	1.049458	-11.424885	150	9	1.1	0.9;
	9561	1	120.69	18.1	0	15.46	0	1.06763	-2.574176	220	5	1.1	0.9;
	9562	1	37.6	21.67	0	0	0	1.102248	-15.189017	150	9	1.2	0.8;
	9563	1	0	0	0	0	0	1.044427	-4.588031	150	9	1.1	0.9;
	9564	1	27.9	4	0	1.86	0	0.999619	24.383536	154	1	1.1	0.9;
	9565	1	34.9	5	0	1.74	0	1.041529	1.66662	154	1	1.1	0.9;
	9566	1	20.72	7.5	0	0.33	0	1.013163	-3.278369	110	19	1.1	0.9;
	9567	1	349.2	61.3	0.233208	36.974905	0	1.058148	83.691686	380	3	1.1	0.9;
	9568	1	68.48	22.4	0	4.38	0	0.988028	25.497372	150	8	1.1	0.9;
	9569	1	0.9	0.3	0	0.12	0	0.98349	-8.818575	150	10	1.1	0.9;
	9570	1	0	-0	0	0	0	1.01167	-8.995127	20	6	1.1	0.9;
	9571	1	27.43	-36.4	0	2.2	0	1.020003	19.679578	220	3	1.1	0.9;
	9572	1	33.24	11.1	0	0.62	0	0.934854	-13.687619	110	21	1.1	0.9;
	9573	2	0	0	0	0	0	1.022149	-8.624199	20	22	1.1	0.9;
	9574	2	0	0	0	0	0	1.022149	-8.624199	20	22	1.1	0.9;
	9575	1	4.91	1.4	0	0.49	0	0.953841	-6.772254	110	18	1.1	0.9;
	9576	1	92.4	13.1	0	0.66	0	1.038148	1.704093	154	1	1.1	0.9;
	9577	1	1.1	-0	0	0.28	0	1.00474	6.183504	110	18	1.1	0.9;
	9578	1	26.33	8.7	0	0.58	0	1.002904	-10.549962	110	12	1.1	0.9;
	9579	2	0	0	0	0	0	1.05252	75.282812	20	3	1.1	0.9;
	9580	1	7.61	1.3	0	0.58	0	0.955017	-6.204751	110	18	1.1	0.9;
	9581	1	0	-0	0	0.76	0	0.900053	-13.491159	150	9	1.1	0.9;
	9582	1	75.18	50.09	0.678725	45.75816	0	1.113909	0.125762	380	21	1.2	0.8;
	9583	1	20.1	2.9	0	10.11	0	1.001356	24.721284	154	1	1.1	0.9;
	9584	2	0	0	0	0	0	1.056808	73.571638	20	3	1.1	0.9;
	9585	2	0	0	0	0	0	1.056808	73.571638	20	3	1.1	0.9;
	9586	2	0	0	0	0	0	1.02748	-20.019397	20	6	1.1	0.9;
	9587	2	0	0	0	0	0	1.02748	-20.019397	20	6	1.1	0.9;
	9588	2	0	0	0	0	0	1.030361	8.088164	20	6	1.1	0.9;
	9589	2	0	0	0	0	0	1.030361	8.088164	20	6	1.1	0.9;
	9590	2	0	0	0	0	0	1.030361	8.088164	20	6	1.1	0.9;
	9591	2	0	0	0	0	0	1.030361	8.088164	20	6	1.1	0.9;
	9592	1	-37.81	-21.64	0	83.358794	0	0.962631	-15.697201	380	5	1.1	0.9;
	9593	1	0	0	0	4.61	0	1.024271	16.34188	220	5	1.1	0.9;
	9594	1	3.7	1.3	0	0.2	0	1.042476	-3.280872	110	19	1.1	0.9;
	9595	1	-0	-0	0	17.75	0	1.015991	31.274209	380	3	1.1	0.9;
	9596	1	105.5	35.3	0	1.23	0	1.027355	25.185208	150	8	1.1	0.9;
	9597	2	0	0	0	0	0	1.044678	11.395331	20	7	1.1	0.9;
	9598	1	216	56.2	0	6.98	0	1.007853	-21.50971	220	4	1.1	0.9;
	9599	1	9.08	3.94	0	0.12	0	1.066739	16.847814	110	24	1.1	0.9;
	9600	1	0	0	0	0	0	0.929893	-10.863302	150	9	1.1	0.9;
	9601	1	3.7	0.9	0	0	0	1.008448	19.375399	154	1	1.1	0.9;
	9602	2	0	0	0	0	0	1.034298	11.483502	20	23	1.1	0.9;
	9603	1	328.9	-9.3	0	28.37	0	1.041502	-15.945667	380	4	1.1	0.9;
	9604	1	0	0	0	0	0	1.03472	7.897462	20	7	1.1	0.9;
	9605	1	0	0	0	0	0	1.030227	6.942059	20	23	1.1	0.9;
	9606	1	0	0	0	0	0	1.030227	6.942059	20	23	1.1	0.9;
	9607	1	43.15	14.4	0	1.73	0	0.990146	-1.905831	220	6	1.1	0.9;
	9608	1	43.15	14.8	0	17.41	0	1.002895	0.544676	220	6	1.1	0.9;
	9609	1	0	0	0	1.2	0	1.046846	3.598889	220	1	1.1	0.9;
	9610	1	-15.12	5	0	0.25	0	1.04738	-3.632232	150	9	1.1	0.9;
	9611	1	0	0	0	32.43	0	1.033419	9.504467	220	5	1.1	0.9;
	9612	1	0	-0	0	6.28	0	1.001529	-10.388139	220	4	1.1	0.9;
	9613	2	0	0	0	0	0	1.00657	25.464337	20	24	1.1	0.9;
	9614	2	0	0	0	0	0	1.022099	17.218586	20	14	1.1	0.9;
	9615	1	114.8	33.1	0	2.03	0	1.058811	-2.126192	110	15	1.1	0.9;
	9616	2	0	0	0	0	0	1.021579	4.43582	20	6	1.1	0.9;
	9617	2	0	0	0	0	0	1.023225	-19.634643	20	6	1.1	0.9;
	9618	2	0	0	0	0	0	1.023225	-19.634643	20	6	1.1	0.9;
	9619	2	0	0	0	0	0	1.025226	0.949671	20	6	1.1	0.9;
	9620	1	-0	0	0	24.57	0	1.073765	1.494745	380	1	1.1	0.9;
	9621	1	0	0	0	26.84	0	1.052427	27.707429	380	1	1.1	0.9;
	9622	2	0	0	0	0	0	1.025226	0.949671	20	6	1.1	0.9;
	9623	2	0	0	0	0	0	1.024756	1.450408	20	6	1.1	0.9;
	9624	1	51.76	21	0	0.57	0	1.007587	-6.12031	110	3	1.1	0.9;
	9625	1	35.34	37.8	0	9.28	0	1.009734	25.198035	220	3	1.1	0.9;
	9626	1	15.14	11.3	0	13.16	0	1.017174	-12.280888	150	9	1.1	0.9;
	9627	2	0	0	0	0	0	1.055007	43.318634	20	2	1.1	0.9;
	9628	2	0	0	0	0	0	1.012126	5.963328	11	1	1.1	0.9;
	9629	1	-0	-0	0	3.48	0	1.007963	29.122932	154	1	1.1	0.9;
	9630	1	8.91	2.4	0	0.6	0	0.988371	-6.145154	110	18	1.1	0.9;
	9631	1	26.63	9.4	0	0.14	0	0.999983	-3.374059	110	18	1.1	0.9;
	9632	1	-6.84	-0.61	0	0.49	0	1.043269	8.473763	110	18	1.1	0.9;
	9633	1	153.8	23.4	0	87.73	0	1.045599	28.560936	380	5	1.1	0.9;
	9634	1	-0	-0	0	10.15	0	1.023319	37.705322	150	8	1.1	0.9;
	9635	1	0	0	0	0	0	0.986867	-7.038889	150	9	1.1	0.9;
	9636	1	-22.27	-8.9	0.095871	32.326698	0	1.05028	78.087019	380	3	1.1	0.9;
	9637	1	24.73	6.4	0	0.64	0	1.023003	0.661159	110	3	1.1	0.9;
	9638	2	0	0	0	0	0	0.975244	29.57444	6.3	1	1.1	0.9;
	9639	1	0	-0	0	1.76	0	1.012101	27.54054	154	1	1.1	0.9;
	9640	1	0	0	0	34.29	0	0.993752	2.699763	400	1	1.1	0.9;
	9641	1	124.8	45.6	0	3.12	0	1.018927	41.487619	380	3	1.1	0.9;
	9642	1	0	0	0	0.47	0	1.012858	-2.152579	380	6	1.1	0.9;
	9643	1	9.61	20.2	0.232189	10.586184	0	1.035926	25.971708	220	13	1.1	0.9;
	9644	1	1.3	-0	0	0.37	0	0.953241	-5.357245	110	18	1.1	0.9;
	9645	1	0	0	0	0	0	0.950398	-10.506843	150	9	1.1	0.9;
	9646	1	-0	0	0.050928	29.839072	0	1.007774	-19.841296	380	4	1.1	0.9;
	9647	1	8.81	2.9	0	0.53	0	1.001437	-8.846246	110	12	1.1	0.9;
	9648	1	109.1	15	0	8.82	0	0.99941	-27.313147	220	4	1.1	0.9;
	9649	1	52.36	7.4	0	26.39	0	1.030691	-20.97126	220	4	1.1	0.9;
	9650	1	16.52	2.1	0	0.72	0	0.993064	-8.173443	110	22	1.1	0.9;
	9651	1	10.61	3.5	0	1.3	0	0.986558	-10.760497	110	12	1.1	0.9;
	9652	1	69.98	23.1	0	2.52	0	1.029936	12.989549	110	19	1.1	0.9;
	9653	1	25.83	7.2	0	0.88	0	1.032345	39.964899	110	8	1.1	0.9;
	9654	1	77.58	-6.54	0	10.66	0	1.038957	41.625628	220	13	1.1	0.9;
	9655	2	0	0	0	0	0	1.064074	38.489396	20	14	1.1	0.9;
	9656	2	0	0	0	0	0	1.064118	38.498525	20	14	1.1	0.9;
	9657	1	0	-0	0	16.66	0	1.003385	-6.85368	220	12	1.1	0.9;
	9658	1	174	49.6	0	14.84	0	0.994041	8.719222	220	3	1.1	0.9;
	9659	2	0	0	0	0	0	1.043178	23.372652	20	5	1.1	0.9;
	9660	1	15.22	7.2	0	1.81	0	1.018955	-12.057193	220	4	1.1	0.9;
	9661	1	137.4	81.7	0	6.1	0	1.014781	12.155669	380	3	1.1	0.9;
	9662	2	0	0	0	0	0	1.043665	14.930439	20	14	1.1	0.9;
	9663	1	-0	0	0	40.86	0	1.041265	-14.533241	380	4	1.1	0.9;
	9664	1	65.77	5.2	0.058568	5.441193	0	0.997685	89.671137	150	1	1.1	0.9;
	9665	1	483.8	105.2	0	0	0	1.063379	0.197545	120	20	1.1	0.9;
	9666	1	14.02	6	0	0.57	0	1.036226	-3.103964	110	21	1.1	0.9;
	9667	1	0	-0	0	11.72	0	1.06565	86.49366	380	3	1.1	0.9;
	9668	1	76.6	10.9	0	12.63	0	0.969807	36.7392	154	1	1.1	0.9;
	9669	1	41.14	15.25	0	3.96	0	0.989079	-4.171297	220	5	1.1	0.9;
	9670	1	0	-0	0	1.04	0	1.052384	28.271347	220	1	1.1	0.9;
	9671	1	13.31	3.5	0	0.52	0	0.988269	3.695187	110	18	1.1	0.9;
	9672	1	-0	0	0	1.35	0	1.00409	-9.883933	150	9	1.1	0.9;
	9673	1	0	-0	0	11.08	0	1.046806	67.561933	380	3	1.1	0.9;
	9674	1	0	-0	0	3.06	0	1.030674	13.215171	220	5	1.1	0.9;
	9675	1	69.7	9.9	0	4.66	0	0.989765	14.047486	154	1	1.1	0.9;
	9676	1	60.7	8.7	0	3.23	0	1.005126	29.093859	154	1	1.1	0.9;
	9677	1	0	-0	0	10.67	0	1.016878	13.461153	220	5	1.1	0.9;
	9678	1	0	0	0	0	0	1.025237	-7.644418	150	9	1.1	0.9;
	9679	2	0	0	0	0	0	1.023313	25.625812	20	5	1.1	0.9;
	9680	1	0.4	191.7	0.172516	29.716237	0	1.047106	69.880946	220	3	1.1	0.9;
	9681	2	0	0	0	0	0	1.036584	20.620161	20	5	1.1	0.9;
	9682	2	0	0	0	0	0	1.036459	20.605907	20	5	1.1	0.9;
	9683	1	0	0	0	7.88	0	0.999988	-7.149899	220	6	1.1	0.9;
	9684	1	62.7	9	0	5.31	0	1.007	30.406198	154	1	1.1	0.9;
	9685	1	-37.33	-11.88	0	145.18	0	1.038011	34.090519	380	5	1.1	0.9;
	9686	1	18.82	1.7	0	6.24	0	1.021223	-29.489741	220	4	1.1	0.9;
	9687	2	0	0	0	0	0	1.129641	7.22366	20	9	1.2	0.8;
	9688	1	0	-0	0.019638	0.520362	0	1.014344	-3.12179	110	21	1.1	0.9;
	9689	1	0	-0	0.058048	0.431627	0	1.053142	-0.118026	110	23	1.1	0.9;
	9690	2	0	0	0	0	0	1.031301	-10.495578	20	10	1.1	0.9;
	9691	1	35.04	12.9	0	0.46	0	0.948316	-8.013982	110	21	1.1	0.9;
	9692	1	62.63	16.3	0	2.03	0	1.044742	19.902807	220	5	1.1	0.9;
	9693	2	0	0	0	0	0	1.003834	31.873789	11	1	1.1	0.9;
	9694	1	143.8	46.45	0	3.26	0	1.03953	23.274277	220	5	1.1	0.9;
	9695	2	0	0	0	0	0	1.046243	23.151951	20	5	1.1	0.9;
	9696	2	0	0	0	0	0	1.046243	23.151951	20	5	1.1	0.9;
	9697	2	0	0	0	0	0	1.046243	23.151951	20	5	1.1	0.9;
	9698	2	0	0	0	0	0	1.046243	23.151951	20	5	1.1	0.9;
	9699	2	0	0	0	0	0	1.030153	24.072406	20	5	1.1	0.9;
	9700	2	0	0	0	0	0	1.048966	21.235269	20	5	1.1	0.9;
	9701	1	0	-0	0.083053	-0.249159	0	1.041304	4.262776	220	11	1.1	0.9;
	9702	2	0	0	0	0	0	1.030153	24.072406	20	5	1.1	0.9;
	9703	1	0	-0	0.224628	-0.177831	0	1.080545	11.129943	220	18	1.2	0.8;
	9704	1	117.6	16.7	0	10.55	0	1.049012	0.593857	154	1	1.1	0.9;
	9705	1	-0	0	0	4.18	0	1.039834	43.452094	220	13	1.1	0.9;
	9706	2	0	0	0	0	0	1.007614	14.656778	20	5	1.1	0.9;
	9707	1	-0	0	0	0.31	0	1.050647	32.165732	220	5	1.1	0.9;
	9708	1	0	0	0	26.92	0	1.069992	-3.402535	380	1	1.2	0.8;
	9709	1	61.65	10.24	0	3.48	0	0.999569	14.240658	220	5	1.1	0.9;
	9710	1	176.7	41.2	0	10.66	0	1.06271	27.708558	220	5	1.1	0.9;
	9711	1	20.9	3	0	2.59	0	1.008309	11.773919	154	1	1.1	0.9;
	9712	1	-0	0	0.039696	0.220304	0	0.996162	-17.299436	110	22	1.1	0.9;
	9713	2	0	0	0	0	0	1.017828	-6.986287	20	4	1.1	0.9;
	9714	1	4.81	1	0	0.69	0	0.989153	3.178231	110	18	1.1	0.9;
	9715	1	0	0	0	0	0	0.900287	-12.257818	150	9	1.1	0.9;
	9716	1	-0	-0	0	0.27	0	1.020928	-8.381911	150	9	1.1	0.9;
	9717	1	27.63	17.1	0	5.89	0	0.95811	-5.43376	150	9	1.1	0.9;
	9718	1	38.04	18	0	0.15	0	1.003301	0.519687	220	6	1.1	0.9;
	9719	1	21.72	7.2	0	0.76	0	0.94532	-8.705616	110	21	1.1	0.9;
	9720	1	48.8	6.9	0	3.56	0	1.025138	8.534452	154	1	1.1	0.9;
	9721	1	-5.92	-2.24	0	1.43	0	0.986396	-5.096526	110	18	1.1	0.9;
	9722	2	0	0	0	0	0	0.988205	29.491637	15	1	1.1	0.9;
	9723	1	-0	-0	0	0.69	0	0.994631	-3.825994	220	6	1.1	0.9;
	9724	1	15.32	5	0	1.3	0	1.012631	-8.55661	110	12	1.1	0.9;
	9725	1	4.61	1.7	0	0.36	0	0.976121	-3.668858	110	19	1.1	0.9;
	9726	1	-9.32	-10.58	0	11.38	0	1.028943	29.162114	220	3	1.1	0.9;
	9727	1	-0	0	0.060179	66.686182	0	1.00283	93.080719	380	1	1.1	0.9;
	9728	2	0	0	0	0	0	1.02351	0.368636	11	1	1.1	0.9;
	9729	1	30.83	80	0.470063	21.77828	0	1.033874	-19.115405	220	4	1.1	0.9;
	9730	1	0	0	0	1.68	0	1.058833	15.23437	220	5	1.1	0.9;
	9731	1	0	0	0.197132	5.602151	0	1.011264	-9.485571	220	4	1.1	0.9;
	9732	1	-0	-0	0	19.57	0	1.008401	51.637644	380	3	1.1	0.9;
	9733	1	29.73	11	0	0.61	0	0.959031	-14.016518	110	22	1.1	0.9;
	9734	1	199.6	27.1	0	33.58	0	1.012198	-19.054945	380	4	1.1	0.9;
	9735	1	48.8	6.9	0	0.44	0	1.017276	20.695588	154	1	1.1	0.9;
	9736	1	-0	0	0.060423	0.269647	0	0.996013	33.473733	220	3	1.1	0.9;
	9737	1	-0.2	-1.64	0	0	0	1.01757	30.154601	150	8	1.1	0.9;
	9738	2	0	0	0	0	0	1.039371	7.907608	20	22	1.1	0.9;
	9739	1	74.58	16.6	0	2.02	0	1.010151	-20.700361	220	4	1.1	0.9;
	9740	1	0	0	0	10.68	0	1.026065	-18.352217	220	5	1.1	0.9;
	9741	1	0	0	0.090774	3.610444	0	1.054135	-0.225548	220	18	1.1	0.9;
	9742	2	0	0	0	0	0	1.028604	-12.690591	20	4	1.1	0.9;
	9743	1	10.31	4.6	0	0.64	0	1.021205	-3.900416	110	15	1.1	0.9;
	9744	1	0	-0	0	35.98	0	1.010863	24.030251	380	3	1.1	0.9;
	9745	1	261.7	104.6	0	29.76	0	1.059759	82.219705	380	3	1.1	0.9;
	9746	1	132.9	13.6	0	3.44	0	0.994238	-25.837919	220	4	1.1	0.9;
	9747	1	0	0	0	3.87	0	1.014425	3.598052	220	6	1.1	0.9;
	9748	2	0	0	0	0	0	0.982004	27.243336	11	1	1.1	0.9;
	9749	1	207.7	63	0.100827	232.399999	0	1.007061	23.578268	380	5	1.1	0.9;
	9750	1	-0	0	0	0.42	0	0.960577	-0.88261	110	18	1.1	0.9;
	9751	2	0	0	0	0	0	1.057943	47.733865	20	3	1.1	0.9;
	9752	1	406.3	92.5	0	66.92	0	1.020153	18.72273	380	6	1.1	0.9;
	9753	2	0	0	0	0	0	1.057896	47.783354	20	3	1.1	0.9;
	9754	2	0	0	0	0	0	1.057682	40.579523	20	3	1.1	0.9;
	9755	1	40.5	5.7	0	2.11	0	0.997177	11.986244	154	1	1.1	0.9;
	9756	2	0	0	0	0	0	1.0136	-3.104621	20	10	1.1	0.9;
	9757	2	0	0	0	0	0	1.058111	40.947033	20	3	1.1	0.9;
	9758	2	0	0	0	0	0	1.058111	40.947033	20	3	1.1	0.9;
	9759	2	0	0	0	0	0	1.057682	40.579523	20	3	1.1	0.9;
	9760	1	0	0	0.153311	-0.114983	0	1.045285	5.023031	380	6	1.1	0.9;
	9761	2	0	0	0	0	0	1.058111	40.947033	20	3	1.1	0.9;
	9762	2	0	0	0	0	0	1.058111	40.947033	20	3	1.1	0.9;
	9763	2	0	0	0	0	0	1.058111	40.947033	20	3	1.1	0.9;
	9764	2	0	0	0	0	0	1.058111	40.947033	20	3	1.1	0.9;
	9765	1	20.02	6	0	1.96	0	0.974853	-5.721699	110	21	1.1	0.9;
	9766	2	0	0	0	0	0	1.024995	14.159948	20	14	1.1	0.9;
	9767	1	0	0	0	0	0	1.051219	0.292959	150	9	1.1	0.9;
	9768	1	85.9	12.2	0	2.77	0	1.040206	-0.472104	154	1	1.1	0.9;
	9769	1	-0	0	0	98.36	0	0.975376	18.346988	400	1	1.1	0.9;
	9770	1	109.5	22.8	0	7.06	0	0.927968	-7.699274	110	18	1.1	0.9;
	9771	1	0	0	0	0	0	1.016944	-15.006207	20	4	1.1	0.9;
	9772	1	0	0	0	0	0	1.016944	-15.006207	20	4	1.1	0.9;
	9773	1	-0	0	0.048383	193.021617	0	1.054652	-13.052855	380	4	1.1	0.9;
	9774	1	0	0	0	0.51	0	1.019397	-6.35206	110	3	1.1	0.9;
	9775	2	0	0	0	0	0	1.033768	14.596978	20	14	1.1	0.9;
	9776	1	66.67	27.8	0	0.51	0	1.013783	7.123359	220	7	1.1	0.9;
	9777	1	-0	0	0	0	0	1.013824	-24.041959	20	4	1.1	0.9;
	9778	2	0	0	0	0	0	1.00995	-19.451519	20	4	1.1	0.9;
	9779	1	70.68	11.9	0	4.34	0	1.005122	2.268732	220	20	1.1	0.9;
	9780	1	0	0	0	0	0	1.002327	-2.090017	150	9	1.1	0.9;
	9781	2	0	0	0	0	0	1.026025	92.46745	20	3	1.1	0.9;
	9782	1	-0	-0	0	0	0	1.05076	44.070116	380	8	1.1	0.9;
	9783	1	0	0	0	32.21	0	1.073513	-3.322838	380	12	1.2	0.8;
	9784	1	-7.1	-4.02	0	1.05	0	0.958458	-6.733708	110	18	1.1	0.9;
	9785	2	0	0	0	0	0	1.037564	83.849724	20	3	1.1	0.9;
	9786	1	0	0	0	0	0	1.002609	-15.370282	150	9	1.1	0.9;
	9787	1	42.55	12.5	0	1.26	0	1.051011	-4.984581	220	5	1.1	0.9;
	9788	1	10.33	3.3	0	4.65	0	1.036442	14.080919	220	5	1.1	0.9;
	9789	1	-0	-0	0	21.03	0	1.071124	7.551911	380	1	1.1	0.9;
	9790	1	-0	0	0	0.82	0	1.001263	-15.294204	220	1	1.1	0.9;
	9791	1	45.95	16.7	0	0.34	0	1.040113	-3.422467	110	19	1.1	0.9;
	9792	1	0	0	0	0.35	0	0.983446	-5.444689	110	18	1.1	0.9;
	9793	2	0	0	0	0	0	1.028091	-15.249181	20	10	1.1	0.9;
	9794	1	219.6	36.4	0	17.89	0	1.020096	5.857127	220	5	1.1	0.9;
	9795	1	309	74.5	0	3.69	0	1.018303	-8.158586	220	5	1.1	0.9;
	9796	1	0	0	0	1.32	0	1.047026	-3.76473	150	9	1.1	0.9;
	9797	2	0	0	0	0	0	1.016107	1.01711	11	1	1.1	0.9;
	9798	1	0	0	0	0	0	1.020262	-10.694863	150	9	1.1	0.9;
	9799	1	54.93	16.5	0	6.75	0	1.047712	26.451878	220	5	1.1	0.9;
	9800	1	12.6	1.8	0	1.98	0	0.996947	24.602605	154	1	1.1	0.9;
	9801	1	11.91	1.8	0	0.37	0	1.017241	17.950982	220	13	1.1	0.9;
	9802	1	0	-0	0	5.48	0	1.018419	36.677185	150	8	1.1	0.9;
	9803	1	115.6	16.4	0	2.41	0	1.005632	18.008822	154	1	1.1	0.9;
	9804	2	0	0	0	0	0	0.975244	29.57444	6.3	1	1.1	0.9;
	9805	1	-0	-0	0.212252	0.661817	0	1.003281	32.058023	380	2	1.1	0.9;
	9806	1	89.2	42.7	0	13.74	0	1.052015	-6.377282	150	9	1.1	0.9;
	9807	1	0.1	4.4	0	0.19	0	0.971396	7.124961	110	18	1.1	0.9;
	9808	2	0	0	0	0	0	1.03058	15.553508	20	14	1.1	0.9;
	9809	2	0	0	0	0	0	1.024762	17.972121	20	14	1.1	0.9;
	9810	1	119.5	44.5	0	3.61	0	1.021156	36.648632	150	8	1.1	0.9;
	9811	1	46.08	-47.15	0	11.39	0	1.036478	27.36973	220	13	1.1	0.9;
	9812	1	12.01	3.9	0	0.94	0	0.972329	-5.639687	110	21	1.1	0.9;
	9813	1	0	-0	0	37.35	0	1.08103	1.353606	380	18	1.2	0.8;
	9814	1	12.11	5.4	0	0.63	0	0.958547	-0.989991	110	18	1.1	0.9;
	9815	2	0	0	0	0	0	1.022031	39.678638	20	10	1.1	0.9;
	9816	1	7.61	6.3	0	1.33	0	0.943794	-6.340562	110	18	1.1	0.9;
	9817	1	72.48	44.9	0	1.75	0	1.002609	-15.370282	150	9	1.1	0.9;
	9818	2	0	0	0	0	0	1.001006	93.944307	20	3	1.1	0.9;
	9819	1	15.6	2.2	0	2.31	0	1.002876	26.824286	154	1	1.1	0.9;
	9820	1	0	-0	0.019675	28.750325	0	1.022525	77.641197	220	3	1.1	0.9;
	9821	1	0	-0	0	0.38	0	0.993408	-9.287355	150	9	1.1	0.9;
	9822	1	-20.82	-8.8	0	9.24	0	1.017364	-18.571503	220	4	1.1	0.9;
	9823	1	139.8	-23.6	0	1.91	0	1.036569	7.552584	220	3	1.1	0.9;
	9824	1	61.07	11.1	0	4.8	0	1.015024	15.20421	220	7	1.1	0.9;
	9825	1	-1.6	-0.9	0	0.38	0	0.985774	-6.101869	110	3	1.1	0.9;
	9826	1	32.54	11.7	0	1.17	0	1.024347	-4.066216	110	15	1.1	0.9;
	9827	1	18.02	5.9	0	0.3	0	1.001754	-10.577335	110	12	1.1	0.9;
	9828	1	0	0	0.13	10.79	0	0.997842	-3.148065	220	21	1.1	0.9;
	9829	1	24.03	5.6	0	0.23	0	1.006842	-3.990202	110	3	1.1	0.9;
	9830	2	0	0	0	0	0	1.030843	4.085708	6.3	1	1.1	0.9;
	9831	1	715.8	24.1	0.019472	-0.019472	0	1.026675	25.143797	150	2	1.1	0.9;
	9832	1	0	0	0	0	0	0.970315	-8.449974	150	9	1.1	0.9;
	9833	1	4.2	0.6	0	3.9	0	1.020982	16.877854	154	1	1.1	0.9;
	9834	1	-0	-0	0.27075	91.79435	0	1.078577	3.402619	380	18	1.2	0.8;
	9835	1	0	-0	0	2.18	0	1.033802	9.219664	220	5	1.1	0.9;
	9836	2	0	0	0	0	0	0.984262	23.826085	11	1	1.1	0.9;
	9837	1	-6.71	-2.4	0	0.48	0	1.012771	-0.442279	110	19	1.1	0.9;
	9838	1	24.38	5.91	0	0.41	0	0.988957	-2.562924	110	18	1.1	0.9;
	9839	1	166.4	50.4	0	1.23	0	1.023538	3.417449	220	5	1.1	0.9;
	9840	1	0	-0	0	0.4	0	1.049156	27.915793	380	5	1.1	0.9;
	9841	1	0	-0	0.536088	13.763043	0	1.019755	-1.22221	380	6	1.1	0.9;
	9842	1	-0	0	0	9.74	0	1.026718	18.644754	220	5	1.1	0.9;
	9843	1	-4.91	-2.53	0	120.558401	0	0.992695	4.921387	380	5	1.1	0.9;
	9844	1	0	-0	0	7.52	0	1.013429	-19.964795	220	4	1.1	0.9;
	9845	1	-0	0	0.039558	2.150442	0	1.018435	-0.87376	110	19	1.1	0.9;
	9846	1	0	0	0	38.84	0	1.168959	2.542881	380	9	1.2	0.8;
	9847	1	78.99	16.2	0	1.97	0	1.034787	8.45732	220	5	1.1	0.9;
	9848	1	0	-0	0	79.119351	0	1.084023	4.623015	380	20	1.2	0.8;
	9849	1	244.12	95.29	0	7.55	0	1.014311	18.977675	220	5	1.1	0.9;
	9850	1	0	-0	0	2.09	0	1.028661	23.341446	150	8	1.1	0.9;
	9851	1	-0	-0	0	41.75	0	1.023579	33.717002	150	8	1.1	0.9;
	9852	2	0	0	0	0	0	1.029637	45.618005	20	10	1.1	0.9;
	9853	2	0	0	0	0	0	1.042135	37.953176	11	1	1.1	0.9;
	9854	1	27.9	4	0	4.18	0	0.983802	37.273795	154	1	1.1	0.9;
	9855	2	0	0	0	0	0	1.015406	15.331048	6.3	1	1.1	0.9;
	9856	2	0	0	0	0	0	1.047943	24.026261	20	3	1.1	0.9;
	9857	1	-0	0	0	1.28	0	0.989798	-4.249455	220	1	1.1	0.9;
	9858	1	5.91	1.7	0	8.75	0	1.017609	-21.781337	220	4	1.1	0.9;
	9859	1	267.1	57.4	0	21.33	0	0.999864	-7.119187	380	6	1.1	0.9;
	9860	2	0	0	0	0	0	0.998215	12.734047	20	18	1.1	0.9;
	9861	1	6.81	2.5	0	1	0	1.073302	1.352587	110	15	1.1	0.9;
	9862	2	0	0	0	0	0	1.033735	-14.68008	20	4	1.1	0.9;
	9863	1	0	-0	0.039649	8.710351	0	1.022452	48.259654	380	3	1.1	0.9;
	9864	1	-0	0	0	34.44	0	1.045894	10.932839	380	1	1.1	0.9;
	9865	1	18.42	6	0	0.72	0	0.954852	-8.03732	110	21	1.1	0.9;
	9866	1	-0	0	0	13.79	0	1.015482	34.721993	380	3	1.1	0.9;
	9867	1	-0	-0	0	6.57	0	1.00884	28.509216	380	5	1.1	0.9;
	9868	1	53.66	19.5	0	1.26	0	1.018342	-5.238465	110	19	1.1	0.9;
	9869	1	-0	0	0	2.94	0	1.035871	12.613244	380	1	1.1	0.9;
	9870	2	0	0	0	0	0	1.033735	-14.68008	20	4	1.1	0.9;
	9871	2	0	0	0	0	0	1.033788	21.861155	13.8	1	1.1	0.9;
	9872	1	34.35	3.62	0	0.97	0	0.977849	-5.747813	110	19	1.1	0.9;
	9873	2	0	0	0	0	0	1.040642	8.532255	20	14	1.1	0.9;
	9874	1	207.2	28.6	0	13.85	0	1.014206	-8.158765	220	11	1.1	0.9;
	9875	2	0	0	0	0	0	1.058355	41.76159	20	10	1.1	0.9;
	9876	1	0	-0	0.009917	0.92	0	0.981799	-5.916291	110	18	1.1	0.9;
	9877	1	12.61	7.8	0	0	0	1.0363	-8.161166	150	9	1.1	0.9;
	9878	1	0	0	0	342.15	0	1.045756	-16.19126	380	4	1.1	0.9;
	9879	1	-0	0	0	5.12	0	1.019351	37.983767	380	8	1.1	0.9;
	9880	2	0	0	0	0	0	1.03481	9.302046	20	18	1.1	0.9;
	9881	2	0	0	0	0	0	1.026644	-8.343517	20	4	1.1	0.9;
	9882	1	8.3	1.2	0	1.07	0	1.000383	15.746042	154	1	1.1	0.9;
	9883	1	27.53	9.9	0	85.25	0	1.029959	-17.695067	380	4	1.1	0.9;
	9884	2	0	0	0	0	0	1.015531	4.163384	20	9	1.1	0.9;
	9885	2	0	0	0	0	0	1.034215	9.728222	20	18	1.1	0.9;
	9886	1	99	23.2	0	10.16	0	1.020132	-17.027986	220	4	1.1	0.9;
	9887	2	0	0	0	0	0	1.034215	9.728222	20	18	1.1	0.9;
	9888	1	60.27	21.3	0	1.87	0	0.999681	-3.315766	110	18	1.1	0.9;
	9889	2	0	0	0	0	0	0.97134	45.507431	15.75	1	1.1	0.9;
	9890	2	0	0	0	0	0	1.027947	90.287595	20	3	1.1	0.9;
	9891	1	-0	0	0	0.34	0	1.040062	-16.220503	380	4	1.1	0.9;
	9892	1	61.37	23.2	0	0.22	0	1.024213	6.206273	220	5	1.1	0.9;
	9893	1	-0	-0	0	2.18	0	1.028581	5.914906	220	5	1.1	0.9;
	9894	1	34.14	21.1	0	0.04	0	0.978211	-7.893051	150	9	1.1	0.9;
	9895	1	283.2	101.8	0.230421	43.471474	0	1.019814	13.132837	380	7	1.1	0.9;
	9896	2	0	0	0	0	0	1.035163	2.77244	20	18	1.1	0.9;
	9897	2	0	0	0	0	0	1.016633	-9.935635	20	4	1.1	0.9;
	9898	1	-1377.93	-320	0	36.89	0	1.045649	20.864366	380	3	1.1	0.9;
	9899	1	-0	-0	0.235946	0.173378	0	1.023729	-7.335996	380	6	1.1	0.9;
	9900	2	0	0	0	0	0	1.016633	-9.935635	20	4	1.1	0.9;
	9901	2	0	0	0	0	0	1.022889	18.879559	20	14	1.1	0.9;
	9902	2	0	0	0	0	0	0.951565	38.529485	13.8	1	1.1	0.9;
	9903	1	74.68	46.3	0	12.92	0	1.019587	-11.095675	150	9	1.1	0.9;
	9904	1	-0	-0	0.018975	3.031025	0	1.075446	2.691479	380	19	1.1	0.9;
	9905	1	0	-0	0	20.44	0	0.99587	23.917917	400	1	1.1	0.9;
	9906	2	0	0	0	0	0	0.991405	49.871472	11	1	1.1	0.9;
	9907	1	18.2	2.6	0	12.22	0	1.019315	28.739796	154	1	1.1	0.9;
	9908	1	140.1	25.5	0	10.78	0	1.014569	-26.829214	220	4	1.1	0.9;
	9909	2	0	0	0	0	0	1.035421	27.235973	20	6	1.1	0.9;
	9910	1	11.91	4.4	0	0.13	0	1.054211	-1.099056	110	15	1.1	0.9;
	9911	2	0	0	0	0	0	1.035421	27.235973	20	6	1.1	0.9;
	9912	2	0	0	0	0	0	1.035421	27.235973	20	6	1.1	0.9;
	9913	1	-0	-0	0	8.9	0	1.06299	30.147149	220	5	1.1	0.9;
	9914	2	0	0	0	0	0	1.052369	0.152148	20	18	1.1	0.9;
	9915	1	-6.77	-2.13	0	0.47	0	1.026231	4.096914	110	18	1.1	0.9;
	9916	1	0	-0	0.057943	2.072057	0	1.045591	14.974901	220	5	1.1	0.9;
	9917	1	50.96	17	0	0.31	0	0.998017	-7.448126	110	21	1.1	0.9;
	9918	1	0	-0	0	4.09	0	0.965057	-9.801871	150	9	1.1	0.9;
	9919	1	154.9	25	0	72.89	0	1.072464	0.866855	380	20	1.2	0.8;
	9920	1	46.35	9.1	0	18.2	0	1.033741	4.638633	220	5	1.1	0.9;
	9921	1	43.35	16.2	0	0.77	0	0.970219	-7.386157	110	18	1.1	0.9;
	9922	2	0	0	0	0	0	1.029161	20.548569	20	5	1.1	0.9;
	9923	2	0	0	0	0	0	1.01017	15.666854	20	5	1.1	0.9;
	9924	1	79.49	17	0	29.21	0	1.108643	49.347411	220	3	1.2	0.8;
	9925	1	-73.98	-29.8	0	17.77	0	0.988226	-21.973367	220	10	1.1	0.9;
	9926	2	0	0	0	0	0	1.009486	16.888084	20	5	1.1	0.9;
	9927	1	0	0	0	1.56	0	1.019031	36.638239	154	1	1.1	0.9;
	9928	1	0	-0	0	0	0	1.00569	12.216215	20	5	1.1	0.9;
	9929	1	0	-0	0	0	0	1.00569	12.216215	20	5	1.1	0.9;
	9930	1	0	-0	0	0	0	1.00569	12.216215	20	5	1.1	0.9;
	9931	1	45.45	28.1	0	0.16	0	0.944702	-10.819415	150	9	1.1	0.9;
	9932	2	0	0	0	0	0	1.029382	20.080326	20	5	1.1	0.9;
	9933	2	0	0	0	0	0	1.029161	20.548569	20	5	1.1	0.9;
	9934	1	253.4	81.7	0	16.85	0	1.001995	2.088521	220	20	1.1	0.9;
	9935	1	4.81	1.3	0	0.85	0	0.957824	-5.219353	110	18	1.1	0.9;
	9936	1	4.2	0.7	0	0.16	0	0.970769	4.966177	110	18	1.1	0.9;
	9937	2	0	0	0	0	0	1.01208	8.274182	10.5	1	1.1	0.9;
	9938	1	-0	-0	0	7.43	0	1.029313	26.316753	220	5	1.1	0.9;
	9939	1	55.76	19.9	0	1.14	0	1.005022	6.351004	110	18	1.1	0.9;
	9940	1	0	0	0	0	0	1.012035	-12.726866	150	9	1.1	0.9;
	9941	1	8.91	1.3	0	0.82	0	0.930948	-7.371611	110	18	1.1	0.9;
	9942	1	135.84	15.97	0	4.57	0	1.036318	8.669231	220	5	1.1	0.9;
	9943	1	-0	-0	0	0.05	0	0.966804	-9.648323	150	9	1.1	0.9;
	9944	1	270.14	57.56	0	4.32	0	0.96942	29.960607	220	3	1.1	0.9;
	9945	1	76.79	30.7	0	17.7	0	0.996771	-3.509392	220	6	1.1	0.9;
	9946	2	0	0	0	0	0	1.038225	16.178789	20	5	1.1	0.9;
	9947	1	20.02	10	0	8.08	0	1.000254	-3.054389	220	6	1.1	0.9;
	9948	1	0.5	159.8	0.113242	99.355631	0	1.068984	73.067046	380	3	1.1	0.9;
	9949	1	47.35	15.5	0	1.41	0	1.020529	-14.481379	110	22	1.1	0.9;
	9950	1	55.7	7.9	0	14.44	0	0.98626	29.963187	154	1	1.1	0.9;
	9951	1	0	0	0	15.06	0	1.018021	0.638927	380	1	1.1	0.9;
	9952	1	-0	-0	0	27.51	0	1.051195	-11.661125	150	9	1.1	0.9;
	9953	2	0	0	0	0	0	1.017638	-6.936102	20	4	1.1	0.9;
	9954	1	32.24	10.1	0	0.81	0	0.979395	-2.770904	110	18	1.1	0.9;
	9955	1	59.47	21.4	0.078398	11.851602	0	1.0523	4.801942	220	5	1.1	0.9;
	9956	1	106	24.6	0	5.9	0	1.028162	4.080023	220	7	1.1	0.9;
	9957	2	0	0	0	0	0	1.026081	32.715698	20	5	1.1	0.9;
	9958	2	0	0	0	0	0	1.026081	32.715698	20	5	1.1	0.9;
	9959	2	0	0	0	0	0	1.026081	32.715698	20	5	1.1	0.9;
	9960	2	0	0	0	0	0	1.127948	6.872839	20	9	1.2	0.8;
	9961	2	0	0	0	0	0	1.016605	25.542934	20	5	1.1	0.9;
	9962	1	-0	-0	0	0.3	0	0.978034	-7.906363	150	9	1.1	0.9;
	9963	2	0	0	0	0	0	1.016461	26.125945	20	5	1.1	0.9;
	9964	2	0	0	0	0	0	1.016605	25.542934	20	5	1.1	0.9;
	9965	1	162.7	23.2	0	4.81	0	1.012245	18.343627	154	1	1.1	0.9;
	9966	1	-0	0	0.198354	-0.018766	0	1.026844	15.277514	380	6	1.1	0.9;
	9967	1	36.84	14.4	0	5.09	0	1.028796	30.892401	150	8	1.1	0.9;
	9968	1	68.78	28.2	0	0.26	0	1.016251	21.980602	150	8	1.1	0.9;
	9969	1	0	-0	0	4.36	0	1.039376	5.373829	220	5	1.1	0.9;
	9970	1	326.9	-82.3	0	0	0	1.044895	25.556743	150	2	1.1	0.9;
	9971	1	55.46	18.5	0	1.11	0	0.950074	-11.205307	110	21	1.1	0.9;
	9972	2	0	0	0	0	0	1.017859	24.98212	15	1	1.1	0.9;
	9973	1	46.2	28.47	0	4.19	0	0.992001	57.207789	220	13	1.1	0.9;
	9974	2	0	0	0	0	0	1.044939	29.387061	20	5	1.1	0.9;
	9975	2	0	0	0	0	0	0.957116	25.356265	13.8	1	1.1	0.9;
	9976	2	0	0	0	0	0	1.029146	3.188044	20	20	1.1	0.9;
	9977	1	-0	0	0	0.48	0	0.944686	-10.875693	150	9	1.1	0.9;
	9978	1	0	-0	0	8.12	0	0.98221	31.994448	220	3	1.1	0.9;
	9979	1	0	0	0	0	0	1.063704	66.981987	20	8	1.1	0.9;
	9980	1	-0.12	-1.74	0	0	0	1.098598	65.002423	380	8	1.2	0.8;
	9981	1	-0	0	0.243713	4.864064	0	0.969283	27.641116	380	2	1.1	0.9;
	9982	1	3.8	2.8	0	0.77	0	0.999953	-3.032403	110	18	1.1	0.9;
	9983	1	-0.15	-1.93	0	0	0	1.016284	49.478439	220	8	1.1	0.9;
	9984	2	0	0	0	0	0	1.026243	6.043442	20	6	1.1	0.9;
	9985	1	13.94	9.5	0	3.53	0	1.032237	-8.120404	150	9	1.1	0.9;
	9986	1	184.89	-21.1	0	39.26	0	1.030293	-20.991696	220	4	1.1	0.9;
	9987	1	47.75	11	0	0.05	0	1.033881	21.336662	220	5	1.1	0.9;
	9988	1	0	0	0.346595	-0.269574	0	1.039039	32.352018	150	8	1.1	0.9;
	9989	1	78.74	17.8	0	0.58	0	0.993356	89.245985	150	1	1.1	0.9;
	9990	1	0	-0	0	2.64	0	1.001746	17.611273	154	1	1.1	0.9;
	9991	1	122.8	20.8	0	0.47	0	1.013705	-26.311523	220	4	1.1	0.9;
	9992	1	98	21.7	0	6	0	1.031639	20.734057	220	5	1.1	0.9;
	9993	1	252.9	95.6	0	1.22	0	1.009996	-16.605815	220	4	1.1	0.9;
	9994	1	29.03	11.7	0.059952	1.440048	0	0.99045	-4.367166	110	21	1.1	0.9;
	9995	2	0	0	0	0	0	0.993383	32.126146	11	1	1.1	0.9;
	9996	1	-0	0	0	1.04	0	0.991988	0.772011	220	5	1.1	0.9;
	9997	2	0	0	0	0	0	1.016106	0.556804	11	1	1.1	0.9;
	9998	1	-0	-0	0.059401	1.100599	0	1.024978	18.038243	220	5	1.1	0.9;
	9999	1	0	-0	0	2.08	0	1.019717	13.94859	220	5	1.1	0.9;
	10000	1	0.3	0.5	0.911638	17.535694	0	1.010073	29.952253	380	3	1.1	0.9;
	10001	2	0	0	0	0	0	1.024544	18.974877	20	14	1.1	0.9;
	10002	1	50.46	36	0	0.57	0	0.9864	-4.126615	110	18	1.1	0.9;
	10003	1	0	-0	0	0.25	0	1.011474	16.380507	154	1	1.1	0.9;
	10004	2	0	0	0	0	0	1.015185	-4.986575	20	4	1.1	0.9;
	10005	1	-0	0	0.11	2.05	0	1.015961	-2.994662	220	21	1.1	0.9;
	10006	2	0	0	0	0	0	1.016735	24.349916	11	1	1.1	0.9;
	10007	2	0	0	0	0	0	1.043764	37.173377	20	5	1.1	0.9;
	10008	1	372.4	121.1	0.020615	7.899385	0	0.990736	9.872108	220	5	1.1	0.9;
	10009	1	71.08	44	0	3.28	0	1.014364	-8.724986	150	9	1.1	0.9;
	10010	2	0	0	0	0	0	1.012893	26.368849	11	1	1.1	0.9;
	10011	1	0	0	0	0	0	1.045246	19.956837	220	5	1.1	0.9;
	10012	2	0	0	0	0	0	1.01515	14.648403	11	1	1.1	0.9;
	10013	1	-16.82	5.5	0	0.63	0	0.929504	-10.424558	150	9	1.1	0.9;
	10014	1	16.52	5.5	0	3.5	0	1.031548	-12.552935	110	22	1.1	0.9;
	10015	1	86.2	-30.7	0	8.84	0	1.026585	23.218908	220	3	1.1	0.9;
	10016	1	0	0	0	0	0	0.964803	-5.53487	150	9	1.1	0.9;
	10017	2	0	0	0	0	0	1.017078	-0.332817	20	10	1.1	0.9;
	10018	1	-0	0	0	74.37	0	0.961347	34.727912	400	1	1.1	0.9;
	10019	1	1.9	9.2	0	7.57	0	0.995355	4.44153	220	20	1.1	0.9;
	10020	1	118.4	18.6	0.038961	10.701039	0	1.028019	32.505211	220	13	1.1	0.9;
	10021	1	-0	-0	0	7.58	0	1.035188	16.78973	220	5	1.1	0.9;
	10022	1	-0	0	0	0	0	1.023287	-16.294788	20	4	1.1	0.9;
	10023	2	0	0	0	0	0	1.019357	22.943488	20	5	1.1	0.9;
	10024	1	43.15	22.8	0.039222	0.010778	0	0.986837	13.352735	110	18	1.1	0.9;
	10025	1	-0	0	0	11.39	0	1.00692	0.387233	220	6	1.1	0.9;
	10026	1	0	-0	0.039254	45.960746	0	1.028324	47.055857	380	3	1.1	0.9;
	10027	1	-0	-0	0	3.37	0	1.025886	4.522198	220	3	1.1	0.9;
	10028	1	6.41	2.1	0	0.51	0	0.991697	-10.534217	110	12	1.1	0.9;
	10029	1	1.2	0.4	0	0.6	0	1.070509	1.019613	110	19	1.2	0.8;
	10030	1	-0	-0	0	17.8	0	1.010526	37.963011	380	8	1.1	0.9;
	10031	1	37.87	11.59	0	1.88	0	1.018093	-25.999068	220	4	1.1	0.9;
	10032	1	54.9	-96.8	0	16.42	0	1.016587	-25.957681	220	4	1.1	0.9;
	10033	1	-0	-0	0	3.67	0	1.024594	92.45199	220	1	1.1	0.9;
	10034	2	0	0	0	0	0	1.006944	-5.921798	20	2	1.1	0.9;
	10035	1	0	-0	0	0.86	0	1.013269	7.216318	220	1	1.1	0.9;
	10036	1	0	0	0	0	0	1.009755	-7.535974	150	9	1.1	0.9;
	10037	2	0	0	0	0	0	1.017464	-4.981811	20	10	1.1	0.9;
	10038	1	0	0	0	0	0	0.985548	-7.292905	150	9	1.1	0.9;
	10039	1	18.92	6.6	0	1.4	0	0.983619	-6.734185	110	18	1.1	0.9;
	10040	1	-0	0	0.241227	-0.201023	0	1.013012	25.491888	380	6	1.1	0.9;
	10041	1	-0	0	0	7.49	0	1.067387	4.878862	400	1	1.1	0.9;
	10042	1	-0	0	0	0	0	0.996077	-9.142556	150	9	1.1	0.9;
	10043	1	1.8	0.6	0	0.75	0	0.985622	-13.402712	110	12	1.1	0.9;
	10044	1	-0	0	0.048779	59.541221	0	1.051887	-14.484778	380	4	1.1	0.9;
	10045	1	161.28	54	0	2.12	0	0.99473	33.46147	220	3	1.1	0.9;
	10046	1	25.23	8.3	0	0.61	0	0.965162	-7.575228	110	21	1.1	0.9;
	10047	1	63.87	55.1	0	14.84	0	1.026388	44.52921	380	3	1.1	0.9;
	10048	1	-0	0	0	1.36	0	1.046742	25.054968	220	5	1.1	0.9;
	10049	1	11.41	5.1	0	1	0	0.922112	-7.609271	110	18	1.1	0.9;
	10050	2	0	0	0	0	0	1.02494	28.997481	20	5	1.1	0.9;
	10051	1	67.47	6.1	0	1.46	0	1.025335	-26.564951	220	4	1.1	0.9;
	10052	2	0	0	0	0	0	1.019503	8.365451	6.3	1	1.1	0.9;
	10053	1	-0.12	-1.31	0	0	0	1.098295	64.988854	380	8	1.2	0.8;
	10054	1	25.73	15.9	0	0.74	0	0.974088	-8.221786	150	9	1.1	0.9;
	10055	1	119.3	-2.8	0	0.87	0	1.001416	34.084126	220	3	1.1	0.9;
	10056	1	0	0	0	3.13	0	1.030573	26.350367	220	5	1.1	0.9;
	10057	2	0	0	0	0	0	1.003179	38.552189	13.8	1	1.1	0.9;
	10058	1	109	20.4	0.195847	7.494153	0	1.012718	13.884467	220	5	1.1	0.9;
	10059	1	-12.49	-5.83	0	6.173378	0	1.040992	33.204952	380	5	1.1	0.9;
	10060	2	0	0	0	0	0	1.032321	27.942453	20	5	1.1	0.9;
	10061	1	0	-0	0	5.59	0	1.014454	20.742832	154	1	1.1	0.9;
	10062	1	135.61	26.86	0	1.93	0	1.038246	16.785299	220	5	1.1	0.9;
	10063	1	-53.06	-100	0	0.06	0	1.051738	1.280044	220	6	1.1	0.9;
	10064	2	0	0	0	0	0	1.033602	30.076827	20	5	1.1	0.9;
	10065	2	0	0	0	0	0	1.033601	30.077016	20	5	1.1	0.9;
	10066	1	83.6	11.9	0	1.69	0	1.025422	3.194796	154	1	1.1	0.9;
	10067	1	0	0	0	50.14	0	0.947551	51.235294	400	1	1.1	0.9;
	10068	2	0	0	0	0	0	1.044317	35.206283	20	5	1.1	0.9;
	10069	1	182.5	47	0	33.69	0	1.011202	-12.139109	380	6	1.1	0.9;
	10070	1	8.31	2.7	0	0.72	0	1.011199	-8.344907	110	12	1.1	0.9;
	10071	1	176.8	39.4	0	28.24	0	1.047627	12.568215	380	5	1.1	0.9;
	10072	1	31.03	10	0	0.13	0	0.975088	-6.482619	110	21	1.1	0.9;
	10073	2	0	0	0	0	0	1.028279	-5.259984	20	6	1.1	0.9;
	10074	2	0	0	0	0	0	1.028279	-5.260179	20	6	1.1	0.9;
	10075	2	0	0	0	0	0	1.027739	-17.648712	20	6	1.1	0.9;
	10076	2	0	0	0	0	0	1.027739	-17.648712	20	6	1.1	0.9;
	10077	2	0	0	0	0	0	1.027739	-17.648712	20	6	1.1	0.9;
	10078	2	0	0	0	0	0	1.027739	-17.648712	20	6	1.1	0.9;
	10079	1	8.01	2.6	0	1.11	0	1.001371	-7.617414	110	12	1.1	0.9;
	10080	2	0	0	0	0	0	1.027739	-17.648712	20	6	1.1	0.9;
	10081	1	-37.84	-19.23	0	28.136589	0	0.964931	-3.531377	380	5	1.1	0.9;
	10082	1	0.3	-9	0	1.96	0	1.026513	30.597405	220	3	1.1	0.9;
	10083	1	20	2.9	0	2.33	0	0.994374	10.486339	154	1	1.1	0.9;
	10084	2	0	0	0	0	0	1.028916	-13.292219	20	6	1.1	0.9;
	10085	1	0	0	0	6.97	0	1.027149	-28.47052	220	4	1.1	0.9;
	10086	1	0	-0	0	4.4	0	1.003768	-12.883511	150	9	1.1	0.9;
	10087	1	115.2	55.1	0	30.49	0	1.040379	38.521981	380	3	1.1	0.9;
	10088	1	-9.86	-2.8	0	0.2	0	0.94887	-0.107604	110	18	1.1	0.9;
	10089	1	0	0	0	0.93	0	1.012856	14.386679	154	1	1.1	0.9;
	10090	1	122.5	8.7	0	15.85	0	1.028895	7.374044	220	5	1.1	0.9;
	10091	1	-0	-0	0.200856	-0.150642	0	1.006932	-8.997116	380	6	1.1	0.9;
	10092	2	0	0	0	0	0	1.060112	4.394131	20	20	1.1	0.9;
	10093	1	25.23	47.6	0	4.21	0	1.018243	34.475442	220	3	1.1	0.9;
	10094	2	0	0	0	0	0	0.990464	21.09396	15.75	1	1.1	0.9;
	10095	1	18.72	7.7	0	1.09	0	0.968494	-13.945945	110	22	1.1	0.9;
	10096	2	0	0	0	0	0	1.032797	5.614702	20	18	1.1	0.9;
	10097	1	-0	-0	0	0.88	0	1.009787	-8.141186	150	9	1.1	0.9;
	10098	1	18.41	41.8	0	0.17	0	0.979064	-10.93172	110	12	1.1	0.9;
	10099	1	-0	-0	0.110962	39.477531	0	1.091334	82.213028	380	3	1.2	0.8;
	10100	2	0	0	0	0	0	1.036949	12.275543	20	7	1.1	0.9;
	10101	2	0	0	0	0	0	1.019231	22.327173	20	24	1.1	0.9;
	10102	2	0	0	0	0	0	1.034412	41.712007	20	10	1.1	0.9;
	10103	1	11.71	3.7	0	0.8	0	1.014977	44.252308	110	8	1.1	0.9;
	10104	1	0	0	0	0.66	0	1.086357	-3.618297	150	9	1.1	0.9;
	10105	1	-0	0	0	96.89	0	1.037061	-15.042474	380	4	1.1	0.9;
	10106	1	6.71	1	0	0.39	0	0.984354	-5.000535	110	18	1.1	0.9;
	10107	1	0	0	0	0	0	1.011712	-8.997116	20	6	1.1	0.9;
	10108	1	21.62	9.8	0	0.14	0	0.993002	-4.228433	110	18	1.1	0.9;
	10109	1	52.9	7.6	0	2.3	0	0.984093	18.898155	154	1	1.1	0.9;
	10110	2	0	0	0	0	0	1.012031	-8.893616	20	10	1.1	0.9;
	10111	1	0	0	0	0	0	0.880509	-14.642882	150	9	1.2	0.8;
	10112	1	113.6	44.7	0	1.91	0	0.996028	20.31987	220	5	1.1	0.9;
	10113	1	33.34	8.6	0	3.93	0	1.040357	46.518975	110	8	1.1	0.9;
	10114	2	0	0	0	0	0	0.991929	29.398008	6.3	1	1.1	0.9;
	10115	2	0	0	0	0	0	1.041212	-15.377355	20	4	1.1	0.9;
	10116	2	0	0	0	0	0	1.041212	-15.377355	20	4	1.1	0.9;
	10117	2	0	0	0	0	0	1.041212	-15.377355	20	4	1.1	0.9;
	10118	1	-0	0	0.079002	1.100749	0	0.996379	-9.059184	220	22	1.1	0.9;
	10119	1	-10.36	-4.66	0	0.58	0	1.031368	-5.222436	110	3	1.1	0.9;
	10120	2	0	0	0	0	0	1.033534	-10.544863	20	4	1.1	0.9;
	10121	1	47.3	6.7	0	4.05	0	0.981228	25.659354	154	1	1.1	0.9;
	10122	1	-0	0	0	1.16	0	1.032084	-6.737315	150	9	1.1	0.9;
	10123	1	-92.2	-244.4	0	7.23	0	1.073678	6.788925	400	1	1.1	0.9;
	10124	1	0	0	0.117756	10.14187	0	1.043894	12.654648	380	16	1.1	0.9;
	10125	1	20.9	3	0	7.14	0	0.999681	15.764613	154	1	1.1	0.9;
	10126	1	117	20.4	0	22.34	0	1.016633	-26.276997	220	4	1.1	0.9;
	10127	1	136.1	18.8	0	6.55	0	1.011334	15.123433	220	7	1.1	0.9;
	10128	1	76.38	6.8	0.039079	2.110691	0	0.9986	90.164736	150	1	1.1	0.9;
	10129	1	0	-0	0	2.15	0	0.981485	0.334141	150	9	1.1	0.9;
	10130	1	160.4	63.9	0.533365	14.325161	0	1.031372	-11.35432	220	4	1.1	0.9;
	10131	1	67.28	-15.37	0.23431	15.624504	0	1.018079	18.919741	220	13	1.1	0.9;
	10132	2	0	0	0	0	0	1.046665	36.809947	20	5	1.1	0.9;
	10133	1	0	0	0	0.71	0	0.966583	-5.616871	150	9	1.1	0.9;
	10134	2	0	0	0	0	0	1.046665	36.809947	20	5	1.1	0.9;
	10135	2	0	0	0	0	0	1.046665	36.809947	20	5	1.1	0.9;
	10136	1	3.2	1.1	0	0.56	0	1.007265	-8.463478	110	12	1.1	0.9;
	10137	1	30.83	16.5	0	3.83	0	1.017428	42.294322	150	8	1.1	0.9;
	10138	1	-0	0	0	10.46	0	1.046181	1.313814	220	5	1.1	0.9;
	10139	1	43.7	6.2	0	1.73	0	1.002858	15.589186	154	1	1.1	0.9;
	10140	1	350	129.4	0	33.52	0	1.020735	7.8322	220	7	1.1	0.9;
	10141	2	0	0	0	0	0	1.014507	26.136069	20	6	1.1	0.9;
	10142	2	0	0	0	0	0	1.014507	26.136069	20	6	1.1	0.9;
	10143	2	0	0	0	0	0	1.014507	26.136069	20	6	1.1	0.9;
	10144	1	0	-0	0	1.13	0	1.047938	-5.803473	150	9	1.1	0.9;
	10145	1	0	-0	0	1.6	0	1.030478	-7.645913	150	9	1.1	0.9;
	10146	1	56.66	15.6	0	7.45	0	1.002892	-20.493605	220	4	1.1	0.9;
	10147	1	0.2	1.8	0	0.43	0	1.031214	76.497249	220	3	1.1	0.9;
	10148	1	0	-0	0	12.42	0	1.013214	-9.554079	380	6	1.1	0.9;
	10149	1	-0	-0	0.237121	0.012399	0	1.018017	36.086297	220	3	1.1	0.9;
	10150	1	35.44	5.3	0	2.24	0	1.013983	-26.722266	220	4	1.1	0.9;
	10151	2	0	0	0	0	0	1.009173	19.132388	20	5	1.1	0.9;
	10152	1	0	0	0	0.49	0	0.942539	-11.042543	150	9	1.1	0.9;
	10153	1	22.8	3.2	0	9.98	0	1.01379	35.557931	154	1	1.1	0.9;
	10154	1	18.72	7.9	0	1.67	0	0.999957	-4.276216	110	18	1.1	0.9;
	10155	1	0	0	0	0	0	1.013445	-14.636462	20	6	1.1	0.9;
	10156	2	0	0	0	0	0	0.978477	21.737802	15	1	1.1	0.9;
	10157	1	61.77	10.2	0	4.01	0	1.012921	-22.921055	220	4	1.1	0.9;
	10158	2	0	0	0	0	0	1.051307	2.565569	20	18	1.1	0.9;
	10159	1	18.12	10	0	1.03	0	1.025128	10.343603	220	5	1.1	0.9;
	10160	1	-0	-0	0	116.64	0	1.018665	23.13173	380	6	1.1	0.9;
	10161	1	334.1	103.5	0	14.78	0	0.969348	-29.133437	220	10	1.1	0.9;
	10162	1	86.49	29	0	1.48	0	1.066536	53.399189	380	3	1.1	0.9;
	10163	1	27.86	14.21	0	0	0	1.099095	-14.350196	150	9	1.2	0.8;
	10164	2	0	0	0	0	0	1.009173	19.132388	20	5	1.1	0.9;
	10165	2	0	0	0	0	0	1.028579	38.422797	20	10	1.1	0.9;
	10166	1	5.04	6.1	0	8.68	0	1.100646	-4.336342	150	9	1.2	0.8;
	10167	1	20.02	9	0	0.15	0	0.999717	-3.057798	220	6	1.1	0.9;
	10168	1	0	0	0	4.07	0	1.022517	12.676727	220	1	1.1	0.9;
	10169	1	389.8	-68.1	0	5.03	0	0.978291	-29.075324	220	10	1.1	0.9;
	10170	1	56.56	18.4	0	3.29	0	1.02764	29.676461	220	5	1.1	0.9;
	10171	2	0	0	0	0	0	1.046641	37.417065	20	5	1.1	0.9;
	10172	2	0	0	0	0	0	1.046641	37.417065	20	5	1.1	0.9;
	10173	2	0	0	0	0	0	1.046641	37.417065	20	5	1.1	0.9;
	10174	2	0	0	0	0	0	1.046641	37.417065	20	5	1.1	0.9;
	10175	1	-0	0	0	0.69	0	1.007976	22.201821	220	6	1.1	0.9;
	10176	1	137.7	32.6	0	5.17	0	1.013014	-18.282288	220	4	1.1	0.9;
	10177	2	0	0	0	0	0	1.011416	44.498643	14.4	1	1.1	0.9;
	10178	2	0	0	0	0	0	1.014865	17.912579	20	5	1.1	0.9;
	10179	1	94.4	13.5	0	5.37	0	1.011217	30.729911	154	1	1.1	0.9;
	10180	1	-0	0	0.232388	3.566026	0	1.055178	-0.20412	220	23	1.1	0.9;
	10181	1	-0	-0	0	0	0	1.015832	3.038481	220	1	1.1	0.9;
	10182	1	-0	0	0.03892	0.39108	0	1.059753	5.769076	220	5	1.1	0.9;
	10183	1	0	-0	0	27.8	0	1.006741	-19.775592	380	4	1.1	0.9;
	10184	1	-0	-0	0	7.51	0	0.998885	2.232778	220	19	1.1	0.9;
	10185	1	-0	0	0	0.32	0	1.045273	-4.513781	150	9	1.1	0.9;
	10186	2	0	0	0	0	0	1.031159	7.793526	11	1	1.1	0.9;
	10187	1	83.36	2.6	0	1.06	0	1.034934	20.694892	220	13	1.1	0.9;
	10188	1	92.1	43.8	0	3.69	0	1.030469	33.297086	220	3	1.1	0.9;
	10189	1	-0	0	0	0.75	0	0.981597	1.798554	110	18	1.1	0.9;
	10190	1	0.6	0.1	0	0.38	0	0.994035	-3.862533	110	18	1.1	0.9;
	10191	1	36.04	22.3	0	4.15	0	0.925713	-12.097092	150	9	1.1	0.9;
	10192	2	0	0	0	0	0	1.059619	48.79156	20	3	1.1	0.9;
	10193	1	16.22	8.9	0	2.2	0	1.033897	9.294569	220	5	1.1	0.9;
	10194	2	0	0	0	0	0	1.048908	44.904436	20	3	1.1	0.9;
	10195	2	0	0	0	0	0	1.048908	44.904436	20	3	1.1	0.9;
	10196	1	53.06	75.8	0	51.93	0	1.03659	27.548145	380	3	1.1	0.9;
	10197	2	0	0	0	0	0	1.052604	52.245403	20	3	1.1	0.9;
	10198	2	0	0	0	0	0	1.052604	52.245403	20	3	1.1	0.9;
	10199	1	19.4	2.8	0	9.78	0	1.006354	16.054798	154	1	1.1	0.9;
	10200	1	0	0	0	0.34	0	0.996074	40.620445	154	1	1.1	0.9;
	10201	2	0	0	0	0	0	1.043753	9.584078	20	5	1.1	0.9;
	10202	2	0	0	0	0	0	1.043753	9.584078	20	5	1.1	0.9;
	10203	1	195.5	67.9	0	6.17	0	1.000061	30.564736	220	3	1.1	0.9;
	10204	1	89.1	20.2	0.05992	24.59008	0	1.034597	8.187737	220	5	1.1	0.9;
	10205	1	0	-0	0	53.97	0	1.087147	63.648812	380	8	1.2	0.8;
	10206	1	-0	-0	0	0.51	0	1.004349	-20.93293	380	6	1.1	0.9;
	10207	1	125.2	1.2	0	15.12	0	1.014965	-26.867436	220	4	1.1	0.9;
	10208	1	-0	0	0	0.39	0	1.027154	79.001432	220	3	1.1	0.9;
	10209	1	112.1	36.3	0	6.4	0	1.020951	30.196596	150	8	1.1	0.9;
	10210	2	0	0	0	0	0	1.04659	30.049071	20	5	1.1	0.9;
	10211	2	0	0	0	0	0	0.979798	33.852101	15.75	1	1.1	0.9;
	10212	1	0	0	0	0.12	0	0.954713	-7.719623	110	21	1.1	0.9;
	10213	1	13.31	4.4	0	0.33	0	0.997519	-11.087305	110	12	1.1	0.9;
	10214	1	0	-0	0	4.54	0	1.036388	5.802341	220	5	1.1	0.9;
	10215	1	36.24	12.8	0	2.99	0	0.957608	-1.330355	110	19	1.1	0.9;
	10216	1	89.7	17.5	0	4.06	0	0.994006	-8.864006	220	6	1.1	0.9;
	10217	1	63.97	12.8	0	0.17	0	1.03253	19.967515	220	5	1.1	0.9;
	10218	2	0	0	0	0	0	1.042184	0.453797	20	18	1.1	0.9;
	10219	1	5.41	67.5	0.113378	75.955519	0	1.068014	62.182795	380	3	1.1	0.9;
	10220	1	-0	0	0	11.7	0	1.010311	-13.538958	220	4	1.1	0.9;
	10221	1	-0	-0	0	3.42	0	1.020205	30.212831	220	14	1.1	0.9;
	10222	2	0	0	0	0	0	1.008414	-5.891971	20	4	1.1	0.9;
	10223	2	0	0	0	0	0	1.011478	-0.639289	20	10	1.1	0.9;
	10224	1	-0	0	0	46.83	0	1.049709	35.041807	150	8	1.1	0.9;
	10225	1	190.3	84.7	0	7.15	0	0.994316	21.023844	220	3	1.1	0.9;
	10226	1	30.04	8.6	0	21.45	0	1.009689	30.943205	150	8	1.1	0.9;
	10227	1	91.9	41.9	1.426313	164.822572	0	1.063548	7.183251	380	19	1.1	0.9;
	10228	1	2.3	0.8	0	0.06	0	0.979022	-3.295551	110	19	1.1	0.9;
	10229	1	0	-0	0	0.61	0	0.987981	35.791209	220	3	1.1	0.9;
	10230	1	51.16	12	0	6.02	0	1.015815	-14.527358	220	4	1.1	0.9;
	10231	1	2.9	-0	0	0.62	0	1.016353	6.24117	110	18	1.1	0.9;
	10232	2	0	0	0	0	0	1.059543	4.22325	20	5	1.1	0.9;
	10233	1	23.43	7.5	0	0.1	0	0.985117	-3.673542	110	18	1.1	0.9;
	10234	1	10.01	2	0	0.61	0	1.000945	0.457571	150	7	1.1	0.9;
	10235	1	-0	-0	0	5.97	0	1.032403	79.8948	220	3	1.1	0.9;
	10236	2	0	0	0	0	0	1.00571	1.688469	20	2	1.1	0.9;
	10237	2	0	0	0	0	0	1.008736	24.556899	11	1	1.1	0.9;
	10238	2	0	0	0	0	0	0.99879	3.61899	20	2	1.1	0.9;
	10239	2	0	0	0	0	0	0.989503	7.040273	13.8	1	1.1	0.9;
	10240	1	0	0	0.386669	50.203331	0	1.019574	-2.005894	220	22	1.1	0.9;
	10241	2	0	0	0	0	0	1.017394	1.999553	20	2	1.1	0.9;
	10242	2	0	0	0	0	0	1.017394	1.999553	20	2	1.1	0.9;
	10243	1	4.2	0.6	0	7.92	0	1.009077	33.192465	154	1	1.1	0.9;
	10244	1	73.7	10.5	0	4.55	0	0.9962	13.133301	154	1	1.1	0.9;
	10245	1	84.59	21.7	0	7.42	0	1.025394	37.285169	110	8	1.1	0.9;
	10246	2	0	0	0	0	0	1.024572	-13.995277	20	4	1.1	0.9;
	10247	1	22.42	2.1	0	0.19	0	0.990947	88.845295	150	1	1.1	0.9;
	10248	1	18.5	3.8	0	0.38	0	1.013093	34.914411	154	1	1.1	0.9;
	10249	2	0	0	0	0	0	1.050901	18.201002	20	7	1.1	0.9;
	10250	2	0	0	0	0	0	1.050579	35.320834	20	8	1.1	0.9;
	10251	1	-0	-0	0	0.3	0	1.051255	28.233713	220	5	1.1	0.9;
	10252	1	0	0	0	3.06	0	1.048361	-3.801602	150	9	1.1	0.9;
	10253	2	0	0	0	0	0	1.055981	85.333831	20	3	1.1	0.9;
	10254	2	0	0	0	0	0	1.055981	85.333831	20	3	1.1	0.9;
	10255	1	8.31	2.7	0	0.69	0	0.97789	-11.444181	110	12	1.1	0.9;
	10256	1	-0	0	0.149251	110.015393	0	1.058657	5.689686	380	18	1.1	0.9;
	10257	1	0	0	0	494.41	0	1.095813	7.157977	750	1	1.2	0.8;
	10258	1	52.56	12.8	0	2.87	0	1.030172	4.037839	220	5	1.1	0.9;
	10259	1	126.3	17.9	0	4.1	0	0.987249	19.471878	154	1	1.1	0.9;
	10260	1	226.8	53.3	0	3.26	0	0.990413	12.1358	220	5	1.1	0.9;
	10261	1	14.92	5.4	0	0.62	0	1.024702	-2.505392	110	19	1.1	0.9;
	10262	1	0	0	0	0	0	0.963231	-9.827573	150	9	1.1	0.9;
	10263	1	-0	0	0	1.76	0	1.022401	-10.668986	150	9	1.1	0.9;
	10264	2	0	0	0	0	0	1.028604	-12.690591	20	4	1.1	0.9;
	10265	1	235.8	23.3	0.020076	8.959924	0	0.995455	35.464296	220	13	1.1	0.9;
	10266	1	0	0	0	51.27	0	0.976287	33.683826	400	1	1.1	0.9;
	10267	2	0	0	0	0	0	1.04126	77.085737	20	3	1.1	0.9;
	10268	2	0	0	0	0	0	1.04126	77.085737	20	3	1.1	0.9;
	10269	2	0	0	0	0	0	1.041091	75.637106	20	3	1.1	0.9;
	10270	2	0	0	0	0	0	1.041091	75.637106	20	3	1.1	0.9;
	10271	1	-0	-0	0.297126	6.812395	0	1.019358	40.783179	380	3	1.1	0.9;
	10272	1	1.7	145.9	0.703396	57.442914	0	1.035344	-9.440561	380	4	1.1	0.9;
	10273	1	24.63	2.1	0	1.26	0	1.035769	30.299859	220	13	1.1	0.9;
	10274	1	19.4	2.8	0	2.51	0	1.003251	19.398693	154	1	1.1	0.9;
	10275	1	-0	0	0	42.94	0	1.006461	-4.530383	380	1	1.1	0.9;
	10276	2	0	0	0	0	0	0.943602	40.868468	10.6	1	1.1	0.9;
	10277	1	0	-0	0	3.15	0	1.002743	2.439835	154	1	1.1	0.9;
	10278	1	16.72	6.4	0.205819	3.97706	0	1.09625	2.009539	110	15	1.2	0.8;
	10279	1	0	-0	0	0	0	0.993457	-9.311514	150	9	1.1	0.9;
	10280	1	0	-0	0.379382	23.898001	0	1.008306	49.448109	380	3	1.1	0.9;
	10281	2	0	0	0	0	0	1.014364	-5.03166	20	4	1.1	0.9;
	10282	1	0	-0	0.038829	0.291171	0	1.035329	16.816351	220	5	1.1	0.9;
	10283	2	0	0	0	0	0	1.014154	-7.89767	20	10	1.1	0.9;
	10284	1	-0	-0	0	2.77	0	0.978318	-28.587274	220	10	1.1	0.9;
	10285	2	0	0	0	0	0	1.00107	32.866746	11	1	1.1	0.9;
	10286	1	19.52	6.4	0	0.33	0	1.003423	-8.957484	110	12	1.1	0.9;
	10287	2	0	0	0	0	0	1.010115	96.580077	20	3	1.1	0.9;
	10288	1	324.9	37.3	0	10.15	0	1.010918	16.282898	380	3	1.1	0.9;
	10289	1	21.82	6.3	0	1.14	0	1.001807	-5.363974	110	18	1.1	0.9;
	10290	1	13	8.8	0	78.85984	0	1.09297	-3.415131	380	9	1.2	0.8;
	10291	1	-175.8	-19.6	0	14.58	0	1.004793	-14.589756	220	10	1.1	0.9;
	10292	1	220.98	91	0	1.16	0	1.04987	-0.490168	220	5	1.1	0.9;
	10293	1	6.91	1.9	0	1.05	0	1.018221	7.924863	110	18	1.1	0.9;
	10294	1	0	0	0.042333	99.938012	0	0.969274	14.532451	380	5	1.1	0.9;
	10295	2	0	0	0	0	0	1.003014	30.196256	11.5	1	1.1	0.9;
	10296	1	1.9	-0	0	1.32	0	1.014051	2.972553	110	18	1.1	0.9;
	10297	1	-0	-0	0	0.41	0	1.09824	-4.476134	150	9	1.1	0.9;
	10298	2	0	0	0	0	0	0.935194	31.202294	6.6	1	1.1	0.9;
	10299	1	17.22	5.1	0	0.99	0	1.023849	8.252664	220	5	1.1	0.9;
	10300	1	12.91	5.1	0	0.16	0	1.009017	-1.484532	110	18	1.1	0.9;
	10301	1	-0	-0	0.058249	2.901751	0	1.036288	14.771989	220	5	1.1	0.9;
	10302	2	0	0	0	0	0	1.016385	22.987603	11	1	1.1	0.9;
	10303	1	120	26	0	22.01	0	1.033317	9.058139	220	5	1.1	0.9;
	10304	1	0	0	0	21.19	0	1.013444	32.585317	380	2	1.1	0.9;
	10305	1	18.12	5.9	0	1.52	0	1.009653	27.767279	150	8	1.1	0.9;
	10306	1	32.14	13.1	0	1.72	0	1.033281	25.592347	150	8	1.1	0.9;
	10307	1	129.7	27.1	0	0.36	0	1.012491	-5.726948	110	19	1.1	0.9;
	10308	1	125.2	8.5	0	0.25	0	1.037198	14.833926	220	3	1.1	0.9;
	10309	1	41.95	14.8	0.361242	114.216909	0	1.001779	27.441266	380	5	1.1	0.9;
	10310	1	20.82	0	0	0.17	0	1.05522	37.830617	220	5	1.1	0.9;
	10311	2	0	0	0	0	0	1.052378	80.413565	20	3	1.1	0.9;
	10312	1	-0	-0	0.038703	2.301297	0	1.05303	-11.87903	220	4	1.1	0.9;
	10313	1	0	-0	0	27.42	0	1.011702	30.813924	380	3	1.1	0.9;
	10314	1	0	0	0	0.09	0	0.996509	34.480354	154	1	1.1	0.9;
	10315	2	0	0	0	0	0	1.018631	-11.528817	20	4	1.1	0.9;
	10316	1	0	0	0	4.93	0	1.01557	17.067482	154	1	1.1	0.9;
	10317	1	0	-0	0	0	0	1.019745	-19.936916	20	4	1.1	0.9;
	10318	1	-0	-0	0	0.47	0	1.01529	6.788901	110	18	1.1	0.9;
	10319	1	-20.12	6.6	0	3.24	0	1.040656	-3.583822	150	9	1.1	0.9;
	10320	1	-0	0	0.276841	2.12089	0	1.061637	11.080165	380	16	1.1	0.9;
	10321	1	-0	0	0	1.02	0	1.018706	-19.205323	220	4	1.1	0.9;
	10322	1	128.1	42.8	0	25.57	0	1.021753	-15.894846	380	17	1.1	0.9;
	10323	1	-0	0	0.176283	0.063098	0	1.026148	5.975825	220	20	1.1	0.9;
	10324	2	0	0	0	0	0	1.058149	41.053759	20	3	1.1	0.9;
	10325	1	41.25	48.4	0	16.52	0	1.018895	-9.219198	220	6	1.1	0.9;
	10326	1	6.31	5.1	0	0.08	0	0.954248	-5.118194	110	18	1.1	0.9;
	10327	1	32.74	5.3	0	3.07	0	1.054552	17.252723	220	5	1.1	0.9;
	10328	2	0	0	0	0	0	1.053529	75.167292	20	3	1.1	0.9;
	10329	2	0	0	0	0	0	1.058149	41.053759	20	3	1.1	0.9;
	10330	2	0	0	0	0	0	1.052406	15.378216	20	14	1.1	0.9;
	10331	1	142.9	25.8	0	24.18	0	1.022138	15.571452	220	5	1.1	0.9;
	10332	1	708.5	228.6	0.175428	67.413556	0	1.048184	29.187518	380	14	1.1	0.9;
	10333	1	80.19	23.6	0	19.21	0	0.997862	-7.341464	220	6	1.1	0.9;
	10334	1	62.67	14.6	0	0.73	0	0.996879	-10.051669	110	22	1.1	0.9;
	10335	1	-0	0	0.05667	52.542775	0	1.062527	90.470423	380	3	1.1	0.9;
	10336	1	-0	0	0	15.47	0	1.026118	35.20302	150	8	1.1	0.9;
	10337	2	0	0	0	0	0	1.019549	96.321283	20	3	1.1	0.9;
	10338	1	25.33	70.8	0.233534	37.945119	0	1.046953	57.054821	380	13	1.1	0.9;
	10339	2	0	0	0	0	0	1.041748	29.3468	20	5	1.1	0.9;
	10340	1	62.17	16.9	0.030139	25.439861	0	1.019083	-15.334385	220	4	1.1	0.9;
	10341	1	0	-0	0	0.36	0	1.024571	19.065775	220	5	1.1	0.9;
	10342	1	18.82	6.8	0	2.15	0	1.014006	-4.787416	110	19	1.1	0.9;
	10343	2	0	0	0	0	0	1.02039	94.325798	20	3	1.1	0.9;
	10344	1	-13.82	4.6	0	0.45	0	1.014212	-8.402985	150	9	1.1	0.9;
	10345	1	0	0	0	2.2	0	1.024295	-24.350579	380	10	1.1	0.9;
	10346	1	7.61	2.5	0	0.39	0	1.012238	7.437218	110	18	1.1	0.9;
	10347	1	0	0	0	63.75	0	1.046748	-10.254125	380	22	1.1	0.9;
	10348	2	0	0	0	0	0	1.006137	-0.690843	20	2	1.1	0.9;
	10349	1	152.2	53.2	0	33.26	0	1.014264	20.268969	380	3	1.1	0.9;
	10350	1	76.6	10.9	0	2.3	0	0.983108	25.395341	154	1	1.1	0.9;
	10351	1	-0	-0	0	27.92	0	0.962955	20.095211	400	1	1.1	0.9;
	10352	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	10353	1	236.6	33.7	0	0.14	0	1.03465	-0.98238	154	1	1.1	0.9;
	10354	1	-0	0	0.172197	2.206503	0	1.052483	35.961266	220	3	1.1	0.9;
	10355	2	0	0	0	0	0	1.056125	27.444238	20	3	1.1	0.9;
	10356	1	10.81	3.6	0	0.08	0	0.997086	-13.477822	110	12	1.1	0.9;
	10357	2	0	0	0	0	0	1.058355	41.76159	20	10	1.1	0.9;
	10358	1	88.4	51.1	0.164758	133.03229	0	1.033029	15.551902	380	3	1.1	0.9;
	10359	1	0	0	0	30.43	0	0.975691	-10.350743	150	9	1.1	0.9;
	10360	1	220.6	70.7	0.038442	120.281558	0	1.053457	38.488901	380	13	1.1	0.9;
	10361	2	0	0	0	0	0	1.020123	-10.115022	20	4	1.1	0.9;
	10362	2	0	0	0	0	0	0.992211	93.728319	20	3	1.1	0.9;
	10363	2	0	0	0	0	0	1.006137	-0.690843	20	2	1.1	0.9;
	10364	1	0	-0	0	0.66	0	1.020444	-7.003967	150	9	1.1	0.9;
	10365	2	0	0	0	0	0	0.996564	21.128006	11	1	1.1	0.9;
	10366	1	296.1	82.3	0	0	0	1.057947	29.561246	150	2	1.1	0.9;
	10367	1	-0	0	0.116788	4.382677	0	1.035619	60.598709	380	3	1.1	0.9;
	10368	2	0	0	0	0	0	0.992945	23.548377	11	1	1.1	0.9;
	10369	1	22.3	3.1	0	4.33	0	0.973171	26.287986	154	1	1.1	0.9;
	10370	1	-0	-0	0	0.29	0	1.022683	-8.218139	150	9	1.1	0.9;
	10371	1	-0	-0	0	57.66	0	1.054139	-12.955601	380	4	1.1	0.9;
	10372	1	109.5	33.6	0	5.75	0	1.022566	32.312225	220	3	1.1	0.9;
	10373	2	0	0	0	0	0	1.029653	5.192863	20	23	1.1	0.9;
	10374	2	0	0	0	0	0	1.022804	17.173209	20	2	1.1	0.9;
	10375	2	0	0	0	0	0	1.022804	17.173209	20	2	1.1	0.9;
	10376	2	0	0	0	0	0	0.927975	33.115277	8.5	1	1.1	0.9;
	10377	2	0	0	0	0	0	1.027231	14.312058	20	2	1.1	0.9;
	10378	2	0	0	0	0	0	1.027231	14.312058	20	2	1.1	0.9;
	10379	1	102	26.6	0	2.27	0	1.014621	63.360606	220	3	1.1	0.9;
	10380	1	35.94	22.3	0	1.63	0	0.971722	-11.45905	150	9	1.1	0.9;
	10381	1	4.61	1.7	0	0.34	0	1.051356	-0.511838	110	19	1.1	0.9;
	10382	1	0	-0	0.203413	0.430864	0	1.019966	29.105238	380	3	1.1	0.9;
	10383	1	0	0	0	16.39	0	1.001192	1.338221	220	6	1.1	0.9;
	10384	2	0	0	0	0	0	1.045359	10.101302	20	14	1.1	0.9;
	10385	1	20.02	8.5	0	1.68	0	0.949279	-0.499054	110	18	1.1	0.9;
	10386	2	0	0	0	0	0	1.024669	12.739456	20	18	1.1	0.9;
	10387	1	-0	-0	0	7.46	0	1.065095	-6.236122	380	12	1.2	0.8;
	10388	2	0	0	0	0	0	1.034464	34.885068	20	10	1.1	0.9;
	10389	1	0	-0	0	0.11	0	1.08082	-3.765935	150	9	1.1	0.9;
	10390	2	0	0	0	0	0	1.058213	27.963025	20	2	1.1	0.9;
	10391	1	316	102.9	0.24333	12.017364	0	1.018502	21.645619	220	14	1.1	0.9;
	10392	1	282.5	56.2	0	8.6	0	0.998615	-13.124885	380	6	1.1	0.9;
	10393	1	0	-0	0	1.13	0	0.971541	5.051864	110	18	1.1	0.9;
	10394	1	0	0	0	0	0	1.020311	-14.530379	220	4	1.1	0.9;
	10395	1	-0	-0	0	1.49	0	1.011667	-13.100581	110	22	1.1	0.9;
	10396	1	0.1	-0	0	0.39	0	1.045532	-3.335566	110	19	1.1	0.9;
	10397	1	-0	-0	0	0.16	0	1.030037	17.642709	154	1	1.1	0.9;
	10398	1	301.6	87.4	0.05701	67.292991	0	1.087045	31.959882	380	14	1.1	0.9;
	10399	1	142.8	20.3	0	6.77	0	1.05128	2.217499	220	5	1.1	0.9;
	10400	1	117.5	91	0.375992	45.737717	0	1.008407	28.740661	380	3	1.1	0.9;
	10401	2	0	0	0	0	0	1.029941	-14.211447	20	4	1.1	0.9;
	10402	2	0	0	0	0	0	1.029941	-14.211447	20	4	1.1	0.9;
	10403	1	31.94	10	0	0.18	0	0.996631	4.042487	110	18	1.1	0.9;
	10404	2	0	0	0	0	0	1.029941	-14.211447	20	4	1.1	0.9;
	10405	2	0	0	0	0	0	1.029941	-14.211447	20	4	1.1	0.9;
	10406	2	0	0	0	0	0	1.04163	29.614847	20	2	1.1	0.9;
	10407	1	7.51	3.1	0	0	0	1.056158	27.931459	220	5	1.1	0.9;
	10408	2	0	0	0	0	0	1.042166	-4.652525	20	9	1.1	0.9;
	10409	2	0	0	0	0	0	1.012872	19.301017	15	1	1.1	0.9;
	10410	2	0	0	0	0	0	1.034215	9.728222	20	18	1.1	0.9;
	10411	1	149.53	34.9	0	8.9	0	0.996493	12.979776	220	5	1.1	0.9;
	10412	1	3.8	0.6	0	0.73	0	0.972849	9.043882	110	18	1.1	0.9;
	10413	1	0	-0	0	6.17	0	1.068439	10.944825	220	18	1.1	0.9;
	10414	1	29.43	10.7	0	0.99	0	1.063742	1.726953	110	19	1.1	0.9;
	10415	1	0	-0	0	13.03	0	1.126602	-0.385846	380	9	1.2	0.8;
	10416	1	-38.01	-22.73	0	0	0	1.04632	32.563114	150	8	1.1	0.9;
	10417	2	0	0	0	0	0	1.002675	8.668171	20	18	1.1	0.9;
	10418	2	0	0	0	0	0	1.0326	4.106738	20	14	1.1	0.9;
	10419	1	-0	0	0	0.28	0	0.998209	-5.886993	110	3	1.1	0.9;
	10420	1	0	-0	0	0	0	1.035607	16.604846	20	5	1.1	0.9;
	10421	1	5.71	3	0	0.18	0	0.993109	-2.156818	110	18	1.1	0.9;
	10422	2	0	0	0	0	0	1.039546	19.356134	20	5	1.1	0.9;
	10423	2	0	0	0	0	0	1.020956	21.259248	20	6	1.1	0.9;
	10424	1	-0	-0	0	0	0	1.010169	13.76329	20	5	1.1	0.9;
	10425	1	-0	-0	0	0	0	1.010169	13.76329	20	5	1.1	0.9;
	10426	2	0	0	0	0	0	1.015046	18.361995	20	5	1.1	0.9;
	10427	2	0	0	0	0	0	1.015046	18.361995	20	5	1.1	0.9;
	10428	2	0	0	0	0	0	1.046548	6.871525	11	1	1.1	0.9;
	10429	1	-0	-0	0	0	0	1.010169	13.76329	20	5	1.1	0.9;
	10430	1	-0	-0	0	0	0	1.010169	13.76329	20	5	1.1	0.9;
	10431	1	33.34	5	0	0.38	0	1.048057	30.057407	220	5	1.1	0.9;
	10432	2	0	0	0	0	0	0.998217	13.335097	20	5	1.1	0.9;
	10433	2	0	0	0	0	0	0.997543	24.058125	11	1	1.1	0.9;
	10434	2	0	0	0	0	0	1.029635	34.5439	20	5	1.1	0.9;
	10435	2	0	0	0	0	0	1.029635	34.5439	20	5	1.1	0.9;
	10436	2	0	0	0	0	0	0.997985	14.815692	20	5	1.1	0.9;
	10437	2	0	0	0	0	0	1.042028	12.694917	20	5	1.1	0.9;
	10438	2	0	0	0	0	0	1.042028	12.694917	20	5	1.1	0.9;
	10439	2	0	0	0	0	0	0.998217	13.335097	20	5	1.1	0.9;
	10440	1	0.9	0.3	0	2.19	0	1.017812	-8.584974	110	12	1.1	0.9;
	10441	1	-0	0	0	1.04	0	1.00604	24.82908	154	1	1.1	0.9;
	10442	2	0	0	0	0	0	1.041381	77.156696	20	3	1.1	0.9;
	10443	2	0	0	0	0	0	1.038225	16.178789	20	5	1.1	0.9;
	10444	1	102.3	19	0	1.84	0	1.008987	13.241803	220	5	1.1	0.9;
	10445	1	-0	-0	0	0	0	1.014379	-16.62541	20	4	1.1	0.9;
	10446	1	-0	0	0.365093	66.535145	0	1.011732	16.560007	380	3	1.1	0.9;
	10447	2	0	0	0	0	0	1.044747	17.319249	20	5	1.1	0.9;
	10448	2	0	0	0	0	0	1.044129	-4.738961	20	4	1.1	0.9;
	10449	1	29.23	9.4	0	0.68	0	1.008104	35.763493	110	8	1.1	0.9;
	10450	1	244	97.8	0.721852	64.518559	0	1.013812	-18.422761	380	4	1.1	0.9;
	10451	1	150.2	51.6	0	2.79	0	1.020135	22.317919	150	8	1.1	0.9;
	10452	2	0	0	0	0	0	1.014607	15.799125	11.5	1	1.1	0.9;
	10453	2	0	0	0	0	0	0.994929	92.815452	20	3	1.1	0.9;
	10454	1	-0	0	0.094706	0.435294	0	1.049983	5.255239	220	21	1.1	0.9;
	10455	1	99.4	22.2	0	2.16	0	1.015763	-26.684165	220	4	1.1	0.9;
	10456	2	0	0	0	0	0	1.009038	19.41053	20	5	1.1	0.9;
	10457	1	51	7.3	0	6.06	0	1.011409	27.376862	154	1	1.1	0.9;
	10458	1	64.2	26.4	0	1.05	0	1.031857	-4.668184	110	23	1.1	0.9;
	10459	1	32.34	16.8	0	1.42	0	0.989404	-3.967336	110	18	1.1	0.9;
	10460	1	-0	-0	0	0.39	0	0.974543	-8.287979	150	9	1.1	0.9;
	10461	2	0	0	0	0	0	1.035338	-7.144697	20	4	1.1	0.9;
	10462	2	0	0	0	0	0	1.035338	-7.144697	20	4	1.1	0.9;
	10463	1	8.48	-1.82	0	0.37	0	1.021625	18.08826	220	5	1.1	0.9;
	10464	1	46.75	24.4	0	3.66	0	1.022503	-30.8059	220	4	1.1	0.9;
	10465	1	0	0	0	0	0	1.098505	65.011053	380	1	1.2	0.8;
	10466	2	0	0	0	0	0	1.008964	6.659925	11	1	1.1	0.9;
	10467	1	293	-45.9	0	33.199841	0	1.023901	2.236436	380	6	1.1	0.9;
	10468	1	-0	0	0	10.89	0	1.028121	7.01221	220	5	1.1	0.9;
	10469	2	0	0	0	0	0	1.021005	21.987746	20	14	1.1	0.9;
	10470	2	0	0	0	0	0	1.044747	17.319249	20	5	1.1	0.9;
	10471	1	6.01	2	0	0.17	0	1.038898	6.306789	110	21	1.1	0.9;
	10472	2	0	0	0	0	0	1.061235	43.554839	20	3	1.1	0.9;
	10473	2	0	0	0	0	0	0.962508	28.579218	15	1	1.1	0.9;
	10474	2	0	0	0	0	0	1.061235	43.554839	20	3	1.1	0.9;
	10475	2	0	0	0	0	0	1.061235	43.554839	20	3	1.1	0.9;
	10476	2	0	0	0	0	0	1.061235	43.554839	20	3	1.1	0.9;
	10477	1	0	0	0	0	0	1.033645	-10.792881	150	9	1.1	0.9;
	10478	1	89.8	19.7	0	0.36	0	1.070343	-2.942936	120	20	1.2	0.8;
	10479	2	0	0	0	0	0	1.060605	43.468335	20	3	1.1	0.9;
	10480	2	0	0	0	0	0	1.060605	43.468335	20	3	1.1	0.9;
	10481	1	73.58	26.8	0	3.76	0	1.033317	28.755712	220	1	1.1	0.9;
	10482	1	-8.01	-3.24	0	0.32	0	0.986425	-8.184485	110	23	1.1	0.9;
	10483	1	-0	-0	0	0	0	1.024181	-13.991173	20	4	1.1	0.9;
	10484	1	-0	-0	0	0	0	1.024181	-13.991173	20	4	1.1	0.9;
	10485	1	117.5	36.4	0.119432	3.760284	0	1.001889	12.056714	220	5	1.1	0.9;
	10486	1	-0	-0	0	0	0	1.024181	-13.991173	20	4	1.1	0.9;
	10487	1	0	-0	0	10.34	0	0.983472	31.050625	400	1	1.1	0.9;
	10488	2	0	0	0	0	0	1.031942	-8.549229	20	4	1.1	0.9;
	10489	2	0	0	0	0	0	1.031942	-8.549229	20	4	1.1	0.9;
	10490	1	0	-0	0	40.44	0	0.962065	32.684828	400	1	1.1	0.9;
	10491	1	-0	-0	0.57356	0.223096	0	1.039235	34.339114	380	5	1.1	0.9;
	10492	1	0	0	0	25	0	1.018124	25.300043	150	8	1.1	0.9;
	10493	1	122.8	20.4	0	0.77	0	1.0002	34.065331	220	3	1.1	0.9;
	10494	1	-0	0	0	2.48	0	1.022296	15.568947	220	5	1.1	0.9;
	10495	1	0	0	0	1.74	0	1.022369	-10.670954	150	9	1.1	0.9;
	10496	1	-0	0	0	5.58	0	1.001607	28.753258	380	3	1.1	0.9;
	10497	1	-0	0	0	9.26	0	1.043292	-5.133111	220	5	1.1	0.9;
	10498	1	34.74	9.1	0.059151	0.790849	0	1.000583	-4.148606	110	18	1.1	0.9;
	10499	1	64.47	40	0	5.82	0	0.987679	-2.410947	220	6	1.1	0.9;
	10500	1	35.84	11.8	0	0.94	0	0.943691	-8.960229	110	21	1.1	0.9;
	10501	2	0	0	0	0	0	0.993105	34.379244	10.6	1	1.1	0.9;
	10502	1	-0	-0	0.037175	146.582825	0	1.096229	69.590629	380	3	1.2	0.8;
	10503	1	-1083.9	-433.7	0	2.72	0	1.046045	21.235676	380	3	1.1	0.9;
	10504	1	0.1	-9.6	0	1.28	0	1.023031	8.747727	220	3	1.1	0.9;
	10505	2	0	0	0	0	0	1.058114	43.382414	20	10	1.1	0.9;
	10506	1	0	0	0	129.47	0	1.023916	-20.729397	380	4	1.1	0.9;
	10507	2	0	0	0	0	0	1.016731	-10.513954	20	4	1.1	0.9;
	10508	2	0	0	0	0	0	1.009642	-17.065869	20	4	1.1	0.9;
	10509	1	0	0	0	0	0	1.010487	-18.296961	20	4	1.1	0.9;
	10510	1	0	0	0	0	0	1.010487	-18.296961	20	4	1.1	0.9;
	10511	1	51.7	7.4	0	2.07	0	1.011931	29.719396	154	1	1.1	0.9;
	10512	1	-0	0	0	0	0	1.017479	-21.662554	20	4	1.1	0.9;
	10513	1	77.91	5.03	0	19.71	0	1.015995	27.945435	220	13	1.1	0.9;
	10514	1	-0	0	0	10.22	0	1.011219	48.828665	380	3	1.1	0.9;
	10515	1	27.43	15.7	0.229924	-0.132023	0	1.026552	0.558438	220	21	1.1	0.9;
	10516	1	-0	0	0	0	0	1.023287	-16.294788	20	4	1.1	0.9;
	10517	1	10.21	3.3	0	0.62	0	0.996447	-9.896793	110	12	1.1	0.9;
	10518	1	-0	-0	0	0	0	0.999015	14.71774	20	5	1.1	0.9;
	10519	1	44.05	16	0	1.63	0	0.986622	-5.371359	110	21	1.1	0.9;
	10520	1	-0	-0	0	0	0	1.039787	-14.573329	20	4	1.1	0.9;
	10521	2	0	0	0	0	0	1.031898	-9.317211	20	6	1.1	0.9;
	10522	2	0	0	0	0	0	1.031898	-9.317211	20	6	1.1	0.9;
	10523	1	0	0	0.240143	4.170791	0	0.969715	27.552049	380	2	1.1	0.9;
	10524	1	1	0.2	0	0.52	0	1.047212	-3.686325	110	19	1.1	0.9;
	10525	1	290.8	30.5	0	29.529841	0	0.999054	-14.356299	380	6	1.1	0.9;
	10526	1	-0	0	1.064216	6.752235	0	1.081283	24.107685	380	5	1.1	0.9;
	10527	2	0	0	0	0	0	1.029166	13.522836	20	14	1.1	0.9;
	10528	1	16.7	2.4	0	3.56	0	1.008758	25.321078	154	1	1.1	0.9;
	10529	1	-18.52	6.1	0	1.01	0	1.01949	-6.454647	150	9	1.1	0.9;
	10530	2	0	0	0	0	0	1.043738	28.126856	20	5	1.1	0.9;
	10531	1	4.75	-0.16	0	0.61	0	0.996554	-2.148476	110	18	1.1	0.9;
	10532	1	10.81	3.9	0	1.39	0	1.044115	-2.191384	110	15	1.1	0.9;
	10533	1	7.21	2.4	0	1.67	0	0.990048	-14.661845	110	12	1.1	0.9;
	10534	2	0	0	0	0	0	1.01601	-0.145046	20	10	1.1	0.9;
	10535	2	0	0	0	0	0	1.126501	-1.915268	20	9	1.2	0.8;
	10536	1	0	0	0.234325	6.034729	0	1.031886	-0.395607	380	6	1.1	0.9;
	10537	2	0	0	0	0	0	1.019357	22.943488	20	5	1.1	0.9;
	10538	2	0	0	0	0	0	1.019357	22.943488	20	5	1.1	0.9;
	10539	1	0	0	0	1.77	0	1.009538	-0.472346	110	18	1.1	0.9;
	10540	1	0	-0	0	0.11	0	0.989531	-9.664587	150	9	1.1	0.9;
	10541	1	11.2	1.6	0	3.74	0	1.011803	30.349504	154	1	1.1	0.9;
	10542	1	-0	0	0	4.6	0	0.995039	-1.015237	220	15	1.1	0.9;
	10543	1	0	0	0	0	0	0.982939	-5.438658	150	9	1.1	0.9;
	10544	1	0	0	0	7.17	0	1.014699	-19.147266	220	4	1.1	0.9;
	10545	1	456.6	227.7	0	0.61	0	1.015085	78.810625	220	3	1.1	0.9;
	10546	1	99.9	19.9	0	6.19	0	0.98902	25.688069	150	8	1.1	0.9;
	10547	1	72.68	1.1	0	6.06	0	1.014595	-27.559944	220	4	1.1	0.9;
	10548	1	0	0	0.08	4.38	0	1.028279	-14.143423	380	6	1.1	0.9;
	10549	1	-0	0	0	1.8	0	0.997397	29.197028	154	1	1.1	0.9;
	10550	1	32.1	4.5	0	2.82	0	1.040624	1.360588	154	1	1.1	0.9;
	10551	1	6.31	1.3	0	0.52	0	0.995284	-3.635619	110	18	1.1	0.9;
	10552	1	21.92	14.1	0.100168	0.259832	0	0.993841	-4.134498	110	21	1.1	0.9;
	10553	1	-0	-0	0	31.34	0	1.026938	-11.97068	380	6	1.1	0.9;
	10554	2	0	0	0	0	0	1.001419	42.731033	15.75	1	1.1	0.9;
	10555	2	0	0	0	0	0	1.058617	17.70472	20	5	1.1	0.9;
	10556	2	0	0	0	0	0	1.058617	17.704764	20	5	1.1	0.9;
	10557	2	0	0	0	0	0	0.962211	47.074823	6.3	1	1.1	0.9;
	10558	2	0	0	0	0	0	1.058617	17.704764	20	5	1.1	0.9;
	10559	1	-241.32	-48.72	0	16.62	0	1.057043	-2.110383	220	23	1.1	0.9;
	10560	1	48.05	23	0	0.78	0	0.991924	-6.64751	110	23	1.1	0.9;
	10561	1	8.3	1.2	0	3.78	0	0.970525	38.020423	154	1	1.1	0.9;
	10562	1	25.23	10.3	0	2.69	0	1.025974	23.061161	150	8	1.1	0.9;
	10563	1	86.7	34.6	0	2.73	0	0.986721	-24.607242	220	6	1.1	0.9;
	10564	1	-0	0	0	2.21	0	1.016733	22.521914	220	5	1.1	0.9;
	10565	1	134.05	13.3	0	25.19	0	1.003706	0.548715	220	6	1.1	0.9;
	10566	1	6.61	1.6	0	2.13	0	0.986044	2.946158	110	18	1.1	0.9;
	10567	2	0	0	0	0	0	1.033196	35.634446	20	10	1.1	0.9;
	10568	1	20.9	3	0	0.17	0	1.005894	17.491888	154	1	1.1	0.9;
	10569	1	0	-0	0.226098	0.861006	0	1.07312	36.327806	380	14	1.1	0.9;
	10570	1	19.22	3.8	0	0.31	0	1.017623	7.990727	110	18	1.1	0.9;
	10571	1	-0	-0	0	0.68	0	0.936276	-8.848857	110	18	1.1	0.9;
	10572	1	0	-0	0	3.98	0	0.993855	48.71734	154	1	1.1	0.9;
	10573	1	0	-0	0	6.22	0	1.06341	18.471219	220	3	1.1	0.9;
	10574	1	89.7	40.5	0	38.13	0	1.010892	6.188269	220	7	1.1	0.9;
	10575	2	0	0	0	0	0	1.058835	5.385711	20	20	1.1	0.9;
	10576	1	-0	0.2	0	2.19	0	1.011815	23.670484	220	3	1.1	0.9;
	10577	1	17.72	6.3	0	0.33	0	0.92263	-8.762423	110	18	1.1	0.9;
	10578	1	0.3	16.8	0.272933	10.6453	0	1.019245	-7.81403	220	4	1.1	0.9;
	10579	1	-0	-0	0	5	0	1.07044	5.391057	220	5	1.1	0.9;
	10580	2	0	0	0	0	0	1.003171	37.39829	6.3	1	1.1	0.9;
	10581	1	0	0	0	0	0	1.03472	7.897462	20	7	1.1	0.9;
	10582	1	0	0	0	1.4	0	1.012451	27.828179	154	1	1.1	0.9;
	10583	1	-0	-0	0	0.37	0	0.997256	-4.698247	110	18	1.1	0.9;
	10584	1	-0	-0	0.181549	13.990889	0	1.037717	-0.24122	220	18	1.1	0.9;
	10585	1	144	31.3	0	3.18	0	1.017725	15.676084	220	5	1.1	0.9;
	10586	1	-0	-0	0	17.37	0	0.979511	26.106648	400	1	1.1	0.9;
	10587	1	3.1	6.8	0.410979	101.610403	0	1.175754	3.227378	380	9	1.2	0.8;
	10588	2	0	0	0	0	0	1.000057	26.10357	11	1	1.1	0.9;
	10589	2	0	0	0	0	0	1.019231	22.327173	20	24	1.1	0.9;
	10590	1	-0	0	0	0.62	0	0.962754	-9.860801	150	9	1.1	0.9;
	10591	1	-0	0	0	3.86	0	1.028915	25.353044	150	8	1.1	0.9;
	10592	1	-0.18	-1	0	0	0	1.021257	36.651298	150	8	1.1	0.9;
	10593	1	0	-0	0.099501	119.783595	0	1.07537	4.059371	380	18	1.2	0.8;
	10594	1	-0	-0	0	0.05	0	1.030072	19.446825	220	5	1.1	0.9;
	10595	1	7.61	2.5	0	0.93	0	0.986497	-8.940272	110	18	1.1	0.9;
	10596	1	0	0	0	21.62	0	1.035323	-10.516182	150	9	1.1	0.9;
	10597	1	388.7	97.7	0	41.94	0	1.000973	-13.484082	380	6	1.1	0.9;
	10598	2	0	0	0	0	0	0.996611	21.111934	11	1	1.1	0.9;
	10599	1	-0	-0	0	1.99	0	0.955736	-16.492467	220	10	1.1	0.9;
	10600	1	0	0	0	0	0	0.978181	-9.34911	150	9	1.1	0.9;
	10601	1	106	15.1	0	0.16	0	1.002352	11.816729	154	1	1.1	0.9;
	10602	1	34.14	11.5	0	0.26	0	0.981048	-5.225564	110	21	1.1	0.9;
	10603	1	0	0	0	2.77	0	1.031271	76.497104	220	3	1.1	0.9;
	10604	1	11.21	3.7	0	0.92	0	0.995801	-13.455391	110	12	1.1	0.9;
	10605	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	10606	1	27.93	17.3	0	0	0	1.020452	-7.005782	150	9	1.1	0.9;
	10607	1	3.4	1.2	0	0.19	0	1.012867	-4.992838	110	19	1.1	0.9;
	10608	1	0	0	0	0.76	0	0.965984	-7.512472	110	21	1.1	0.9;
	10609	1	16.7	2.4	0	1.21	0	0.976871	21.844087	154	1	1.1	0.9;
	10610	1	-0	-0	0	1.33	0	0.987043	33.947682	220	3	1.1	0.9;
	10611	1	7.61	2.5	0	0.25	0	0.973563	-5.937481	110	18	1.1	0.9;
	10612	2	0	0	0	0	0	1.014542	2.91424	20	18	1.1	0.9;
	10613	2	0	0	0	0	0	1.018329	1.468316	0.4	1	1.1	0.9;
	10614	1	7.41	2.7	0	0.61	0	0.998681	-5.302422	110	19	1.1	0.9;
	10615	1	0	-0	0	0	0	1.012277	-17.836874	20	4	1.1	0.9;
	10616	1	0	0	0	2.37	0	1.052535	-6.418371	150	9	1.1	0.9;
	10617	2	0	0	0	0	0	1.00657	25.464337	20	24	1.1	0.9;
	10618	1	233.2	39.4	0	104.569841	0	1.009027	2.688479	380	6	1.1	0.9;
	10619	2	0	0	0	0	0	1.031098	31.248826	20	5	1.1	0.9;
	10620	1	37.94	9.8	0	2.26	0	1.013068	-14.640825	220	4	1.1	0.9;
	10621	2	0	0	0	0	0	1.016891	10.696134	20	5	1.1	0.9;
	10622	2	0	0	0	0	0	1.032126	34.816975	20	5	1.1	0.9;
	10623	1	0	0	0	0	0	1.002425	-9.120945	150	9	1.1	0.9;
	10624	1	220.8	68.6	0	1.97	0	0.954754	-16.903313	220	10	1.1	0.9;
	10625	2	0	0	0	0	0	1.016891	10.696134	20	5	1.1	0.9;
	10626	1	0	-0	0	4.05	0	1.029312	22.383604	220	5	1.1	0.9;
	10627	2	0	0	0	0	0	1.040259	34.3106	20	5	1.1	0.9;
	10628	1	101.6	4.5	0	3.59	0	1.020601	-17.036733	220	4	1.1	0.9;
	10629	2	0	0	0	0	0	1.038747	-8.642337	20	4	1.1	0.9;
	10630	1	392.25	92.02	0	9.45	0	1.012281	-3.154738	220	5	1.1	0.9;
	10631	1	-8.15	-5.23	0	0	0	1.026824	34.117804	150	8	1.1	0.9;
	10632	1	117.73	51.5	0.537393	7.902028	0	1.00775	1.043576	220	6	1.1	0.9;
	10633	2	0	0	0	0	0	1.043665	14.930439	20	14	1.1	0.9;
	10634	1	3.8	2.5	0	0.74	0	0.995515	4.358377	110	18	1.1	0.9;
	10635	2	0	0	0	0	0	1.047752	29.881973	20	5	1.1	0.9;
	10636	1	82.39	-0	0	0.15	0	1.036427	21.464192	220	5	1.1	0.9;
	10637	1	0	0	0.019449	-0.019449	0	1.019707	44.368784	150	8	1.1	0.9;
	10638	1	96.7	23.7	0	39.1	0	1.049538	58.287287	380	3	1.1	0.9;
	10639	1	-0	-0	0	35.98	0	1.034322	-7.989117	380	12	1.1	0.9;
	10640	2	0	0	0	0	0	1.07025	2.362941	20	9	1.1	0.9;
	10641	1	185	26.3	0	11.78	0	1.027326	1.583262	154	1	1.1	0.9;
	10642	1	-0	-0	0	3.42	0	1.033422	20.189076	220	5	1.1	0.9;
	10643	2	0	0	0	0	0	1.056442	84.300227	20	3	1.1	0.9;
	10644	1	7.61	3.8	0	0.82	0	1.018021	3.740756	110	18	1.1	0.9;
	10645	1	-0	-0	0	2.12	0	1.038413	24.293754	220	5	1.1	0.9;
	10646	1	-0	0	0	2	0	1.010826	23.125552	220	6	1.1	0.9;
	10647	2	0	0	0	0	0	1.045549	56.658865	20	10	1.1	0.9;
	10648	2	0	0	0	0	0	1.050191	8.166057	20	5	1.1	0.9;
	10649	2	0	0	0	0	0	1.050191	8.166057	20	5	1.1	0.9;
	10650	2	0	0	0	0	0	0.958146	44.224727	6.3	1	1.1	0.9;
	10651	1	0	-0	0.506185	-0.393699	0	1.069459	51.885185	380	8	1.1	0.9;
	10652	2	0	0	0	0	0	0.96764	46.597759	11	1	1.1	0.9;
	10653	2	0	0	0	0	0	1.035469	23.613483	20	5	1.1	0.9;
	10654	1	0	-0	0.090774	6.600444	0	1.049993	-2.183536	220	18	1.1	0.9;
	10655	1	18.62	6.7	0	0.83	0	1.038751	-2.94006	110	15	1.1	0.9;
	10656	2	0	0	0	0	0	1.035773	21.901715	20	5	1.1	0.9;
	10657	2	0	0	0	0	0	0.996408	16.94258	20	5	1.1	0.9;
	10658	2	0	0	0	0	0	0.996408	16.94258	20	5	1.1	0.9;
	10659	1	-0	0	0	0	0	0.992675	12.17578	20	5	1.1	0.9;
	10660	2	0	0	0	0	0	1.050191	8.166057	20	5	1.1	0.9;
	10661	1	79.39	31.3	0	2.23	0	1.031908	31.49392	150	8	1.1	0.9;
	10662	1	-0	-0	0	8.38	0	1.039231	6.804049	220	1	1.1	0.9;
	10663	1	0	-0	0	10.8	0	1.044343	1.455005	154	1	1.1	0.9;
	10664	1	-0	0	0.223135	10.383351	0	1.063137	8.658596	220	18	1.2	0.8;
	10665	1	18.12	6	0	0.7	0	0.954782	-5.432343	110	18	1.1	0.9;
	10666	1	7.41	2.7	0	0.56	0	0.965625	-5.153597	110	19	1.1	0.9;
	10667	1	181.1	25.8	0	42.18	0	1.034195	-0.043883	154	1	1.1	0.9;
	10668	2	0	0	0	0	0	1.043771	22.935303	20	5	1.1	0.9;
	10669	2	0	0	0	0	0	1.043771	22.935303	20	5	1.1	0.9;
	10670	2	0	0	0	0	0	1.009038	19.41053	20	5	1.1	0.9;
	10671	2	0	0	0	0	0	1.058434	6.52887	20	5	1.1	0.9;
	10672	2	0	0	0	0	0	1.058434	6.52887	20	5	1.1	0.9;
	10673	1	46.8	6.7	0	25.55	0	0.959077	36.812848	400	1	1.1	0.9;
	10674	2	0	0	0	0	0	1.058434	6.52887	20	5	1.1	0.9;
	10675	2	0	0	0	0	0	1.058434	6.52887	20	5	1.1	0.9;
	10676	2	0	0	0	0	0	1.058434	6.52887	20	5	1.1	0.9;
	10677	1	-0	-0	0	20.08	0	1.082511	-4.68964	380	12	1.2	0.8;
	10678	2	0	0	0	0	0	1.013602	39.787074	11	1	1.1	0.9;
	10679	1	11.91	4.4	0	0.16	0	1.054397	-1.085701	110	15	1.1	0.9;
	10680	1	-0	0	0	46.28	0	1.046743	-13.79286	380	4	1.1	0.9;
	10681	1	-0	-0	0	0	0	0.999015	14.71774	20	5	1.1	0.9;
	10682	1	3.4	1.1	0	0.55	0	0.976152	2.665581	110	18	1.1	0.9;
	10683	1	0	0	0	26.5	0	1.03011	23.136345	220	5	1.1	0.9;
	10684	1	25.83	5.7	0	0.41	0	0.990428	-1.815452	110	18	1.1	0.9;
	10685	1	-0	0	0	21.32	0	1.022564	-10.647834	150	9	1.1	0.9;
	10686	1	-0	-0	0	49.904544	0	1.04661	5.291874	380	20	1.1	0.9;
	10687	1	-0	0	0	15.49	0	1.015797	19.144387	380	1	1.1	0.9;
	10688	1	11.71	3.8	0	2.39	0	0.999929	-8.815144	110	12	1.1	0.9;
	10689	1	114.7	4.1	0.038987	12.311013	0	1.022926	17.060146	220	13	1.1	0.9;
	10690	1	-0	-0	0	0	0	0.999015	14.71774	20	5	1.1	0.9;
	10691	1	2.7	0.8	0	1.09	0	0.989001	-4.438276	110	18	1.1	0.9;
	10692	1	41.85	15.2	0	0.21	0	1.02229	13.638837	110	19	1.1	0.9;
	10693	2	0	0	0	0	0	1.017738	25.492678	6.3	1	1.1	0.9;
	10694	1	0	0	0	2.63	0	1.03362	30.239326	150	8	1.1	0.9;
	10695	1	36.84	13.3	0	1.05	0	1.009341	3.831939	110	19	1.1	0.9;
	10696	2	0	0	0	0	0	1.027264	14.46311	20	14	1.1	0.9;
	10697	1	97.41	-4.48	0	16.55	0	1.029567	36.282901	220	13	1.1	0.9;
	10698	1	-0	-0	0.135375	37.092175	0	1.072042	-2.027427	380	18	1.2	0.8;
	10699	1	-0	0	0	1.42	0	1.025635	80.371453	220	3	1.1	0.9;
	10700	1	-0	-0	0	0	0	0.999015	14.71774	20	5	1.1	0.9;
	10701	1	0	-0	0	0	0	0.998771	-8.905049	150	9	1.1	0.9;
	10702	1	36.04	12	0	0.11	0	0.955457	-7.69714	110	21	1.1	0.9;
	10703	1	-0	-0	0	2.25	0	0.99653	23.710857	154	1	1.1	0.9;
	10704	1	2.5	94.6	0.543767	2.75707	0	1.013587	-17.471914	380	4	1.1	0.9;
	10705	1	0.4	0.1	0	1.39	0	1.004591	-4.12331	110	18	1.1	0.9;
	10706	1	82.89	32.6	0	45.84	0	1.028713	30.959458	150	8	1.1	0.9;
	10707	1	-0	0	0	5.38	0	0.991507	-2.827643	220	6	1.1	0.9;
	10708	1	0	0	0	0	0	0.979788	-12.77143	150	9	1.1	0.9;
	10709	1	214.8	42.7	0.36439	68.73561	0	1.04422	-9.526922	380	17	1.1	0.9;
	10710	1	-0	0	0	1.32	0	1.045246	19.956837	220	5	1.1	0.9;
	10711	1	256	69.9	0	43.56	0	1.011822	-19.587728	380	4	1.1	0.9;
	10712	2	0	0	0	0	0	1.023574	3.735821	20	2	1.1	0.9;
	10713	2	0	0	0	0	0	1.041679	-0.382764	20	2	1.1	0.9;
	10714	1	102.6	12	0	66.849841	0	1.006499	-20.670075	380	6	1.1	0.9;
	10715	2	0	0	0	0	0	1.043049	13.489159	20	18	1.1	0.9;
	10716	1	11.8	1.7	0	3.85	0	0.995755	15.594071	154	1	1.1	0.9;
	10717	1	-0	-0	0	16.07	0	1.133829	-1.146555	380	23	1.2	0.8;
	10718	1	-0	-0	0.12635	23.056325	0	1.098845	-1.768818	380	21	1.2	0.8;
	10719	2	0	0	0	0	0	1.045442	-2.002576	20	2	1.1	0.9;
	10720	2	0	0	0	0	0	1.045442	-2.002576	20	2	1.1	0.9;
	10721	2	0	0	0	0	0	1.041679	-0.382764	20	2	1.1	0.9;
	10722	2	0	0	0	0	0	1.041679	-0.382764	20	2	1.1	0.9;
	10723	1	29.53	2.4	0	0.35	0	1.013697	-6.977467	110	3	1.1	0.9;
	10724	1	-0	-0	0	0.09	0	1.011827	21.695682	220	5	1.1	0.9;
	10725	1	48.95	21.4	0.057981	1.162019	0	1.044784	-3.156384	110	19	1.1	0.9;
	10726	1	-0	-0	0	106.24	0	0.982381	33.01246	400	1	1.1	0.9;
	10727	1	210.9	48.3	0	9.75	0	1.058551	26.994293	220	1	1.1	0.9;
	10728	1	0	-0	0	2.1	0	1.02296	44.620529	380	3	1.1	0.9;
	10729	1	-0	-0	0	0.81	0	1.017365	-4.056944	150	9	1.1	0.9;
	10730	1	0	-0	0	0.15	0	1.001044	-2.929985	220	6	1.1	0.9;
	10731	1	132	26	0	27.18	0	0.994363	17.183871	380	5	1.1	0.9;
	10732	1	36.94	62.6	0	0.12	0	0.98617	-9.055253	150	9	1.1	0.9;
	10733	1	18.22	6.3	0	0.48	0	0.972246	-5.618279	110	21	1.1	0.9;
	10734	1	-0	0	0	2.33	0	1.022678	25.196328	220	3	1.1	0.9;
	10735	1	-0	-0	0	0.41	0	1.024278	16.341816	220	5	1.1	0.9;
	10736	1	69.7	9.9	0	1.52	0	1.030348	-1.015593	154	1	1.1	0.9;
	10737	1	34.94	8.6	0.078291	0.921709	0	1.029091	11.050068	220	5	1.1	0.9;
	10738	2	0	0	0	0	0	1.04914	42.65185	20	2	1.1	0.9;
	10739	1	-0	-0	0	6.54	0	1.0031	91.66228	150	1	1.1	0.9;
	10740	1	120.6	35.1	0	20.052	0	1.021614	-4.069285	380	6	1.1	0.9;
	10741	1	-0	0	0.038738	-0.038738	0	1.031622	-6.333057	220	10	1.1	0.9;
	10742	2	0	0	0	0	0	1.052378	80.413565	20	3	1.1	0.9;
	10743	2	0	0	0	0	0	1.01964	-10.829105	20	10	1.1	0.9;
	10744	1	0	-0	0.205481	71.847328	0	1.163482	2.07534	380	15	1.2	0.8;
	10745	1	-0	0	0.019259	0.270741	0	1.018903	-11.984477	110	22	1.1	0.9;
	10746	1	258.79	42.82	0	9.28	0	1.029795	21.603625	220	5	1.1	0.9;
	10747	2	0	0	0	0	0	1.053529	75.167292	20	3	1.1	0.9;
	10748	2	0	0	0	0	0	1.055981	85.333831	20	3	1.1	0.9;
	10749	1	25.13	14	0	0.04	0	1.017471	24.91358	220	3	1.1	0.9;
	10750	2	0	0	0	0	0	1.019549	96.321283	20	3	1.1	0.9;
	10751	1	45.75	15.8	0	3.42	0	0.979584	2.951039	110	19	1.1	0.9;
	10752	2	0	0	0	0	0	1.02039	94.325798	20	3	1.1	0.9;
	10753	2	0	0	0	0	0	1.060516	63.207291	20	3	1.1	0.9;
	10754	2	0	0	0	0	0	1.060516	63.207291	20	3	1.1	0.9;
	10755	2	0	0	0	0	0	1.019549	96.321283	20	3	1.1	0.9;
	10756	2	0	0	0	0	0	1.019549	96.321283	20	3	1.1	0.9;
	10757	2	0	0	0	0	0	1.008788	8.521149	11.5	1	1.1	0.9;
	10758	1	0	-0	0	0.06	0	1.013968	18.460501	154	1	1.1	0.9;
	10759	1	0	0	0	2.98	0	1.02531	14.551431	220	5	1.1	0.9;
	10760	1	0	0	0	8.41	0	1.008795	-8.614735	150	9	1.1	0.9;
	10761	1	5.51	2.1	0	1.92	0	1.016881	18.245981	220	5	1.1	0.9;
	10762	2	0	0	0	0	0	1.071277	43.713777	20	3	1.1	0.9;
	10763	2	0	0	0	0	0	1.071277	43.713777	20	3	1.1	0.9;
	10764	2	0	0	0	0	0	1.068762	42.463865	20	3	1.1	0.9;
	10765	2	0	0	0	0	0	1.071277	43.713777	20	3	1.1	0.9;
	10766	1	23.83	8.6	0	3.08	0	1.047426	-0.530089	110	19	1.1	0.9;
	10767	1	37.44	0	0	0.67	0	1.054401	30.288494	220	5	1.1	0.9;
	10768	1	169	70.8	0	5.23	0	1.003424	13.295032	220	5	1.1	0.9;
	10769	1	0	-0	0	61.36	0	1.02179	-15.445748	380	4	1.1	0.9;
	10770	1	31.03	8.9	0	0.49	0	0.994485	-7.687301	110	18	1.1	0.9;
	10771	1	-0	0	0	0.61	0	0.972803	-9.022804	110	18	1.1	0.9;
	10772	1	-0	0	0	0	0	1.021346	-22.881448	20	4	1.1	0.9;
	10773	1	-0	0	0	0	0	1.021346	-22.881448	20	4	1.1	0.9;
	10774	1	264.33	86.69	0	91.94	0	1.049971	16.509432	380	14	1.1	0.9;
	10775	2	0	0	0	0	0	1.034738	85.578957	20	3	1.1	0.9;
	10776	1	-0	-0	0	0	0	1.014379	-16.62541	20	4	1.1	0.9;
	10777	1	16.7	2.4	0	3.98	0	1.001185	37.535252	154	1	1.1	0.9;
	10778	1	-0	0	0	0	0	1.021346	-22.881448	20	4	1.1	0.9;
	10779	1	16.52	6.3	0	0.23	0	1.073663	7.212197	110	18	1.2	0.8;
	10780	1	75.56	13.04	0	8.77	0	1.012788	-24.277248	220	4	1.1	0.9;
	10781	1	-0	-0	0	3.9	0	1.018238	31.754693	154	1	1.1	0.9;
	10782	1	87.5	21.3	0	7.52	0	1.030283	4.050461	220	5	1.1	0.9;
	10783	2	0	0	0	0	0	1.030285	40.859534	11	1	1.1	0.9;
	10784	2	0	0	0	0	0	1.010203	8.259748	11	1	1.1	0.9;
	10785	2	0	0	0	0	0	1.040138	32.45384	20	17	1.1	0.9;
	10786	1	0	0	0	4.49	0	1.001399	27.844323	380	3	1.1	0.9;
	10787	1	127.3	26.7	0	7.02	0	0.972834	6.994576	220	5	1.1	0.9;
	10788	2	0	0	0	0	0	1.03058	15.553508	20	14	1.1	0.9;
	10789	1	110.1	29.4	0	1.11	0	0.978084	-5.505378	110	21	1.1	0.9;
	10790	1	26.33	5.3	0	4.31	0	1.027498	5.66065	220	5	1.1	0.9;
	10791	1	12.71	4.2	0	0.65	0	0.973626	-7.095233	110	18	1.1	0.9;
	10792	1	-40.95	-28.4	0	7.29	0	0.982866	-21.140797	150	10	1.1	0.9;
	10793	1	0	0	0.077725	20.901706	0	1.031135	4.071247	380	6	1.1	0.9;
	10794	1	88.2	21	0	2.22	0	0.986382	37.678673	220	3	1.1	0.9;
	10795	1	0	0	0	14.29	0	1.050489	14.314816	220	5	1.1	0.9;
	10796	1	223.31	61.3	0	4.51	0	1.043249	28.121526	220	1	1.1	0.9;
	10797	1	208.9	29.8	0	1.42	0	1.023736	-1.23217	154	1	1.1	0.9;
	10798	1	0	-0.01	0.23465	19.69065	0	1.087269	1.338438	380	21	1.2	0.8;
	10799	1	6.31	1.3	0	0.12	0	0.98542	-4.881531	110	18	1.1	0.9;
	10800	1	70.48	22.5	0	4.82	0	1.017676	-12.774326	220	4	1.1	0.9;
	10801	1	0	0	0	0	0	1.014647	-12.087921	150	9	1.1	0.9;
	10802	1	1.5	0.4	0	0.15	0	1.006447	10.354507	220	5	1.1	0.9;
	10803	2	0	0	0	0	0	1.047752	29.881973	20	5	1.1	0.9;
	10804	2	0	0	0	0	0	1.047752	29.881973	20	5	1.1	0.9;
	10805	1	-0	-0	0	0.6	0	0.971251	-9.48586	150	9	1.1	0.9;
	10806	1	-0	-0	0	7.35	0	1.01559	45.608342	380	1	1.1	0.9;
	10807	1	50.3	7.2	0	3.22	0	1.02238	33.263905	154	1	1.1	0.9;
	10808	2	0	0	0	0	0	1.023144	-9.538309	20	4	1.1	0.9;
	10809	2	0	0	0	0	0	1.030656	3.405373	20	6	1.1	0.9;
	10810	1	34.9	5	0	2.95	0	0.988125	29.987284	154	1	1.1	0.9;
	10811	2	0	0	0	0	0	1.048656	21.505608	20	5	1.1	0.9;
	10812	1	10.01	4.3	0	5.64	0	1.002307	4.494366	110	18	1.1	0.9;
	10813	1	-0	-0	0	1.81	0	1.023355	6.426723	220	5	1.1	0.9;
	10814	2	0	0	0	0	0	0.976475	43.406918	15.5	1	1.1	0.9;
	10815	1	-0.2	-0.83	0	0	0	1.021355	26.400741	150	8	1.1	0.9;
	10816	1	53.06	19.2	0	0.41	0	1.041533	-3.179597	110	19	1.1	0.9;
	10817	1	22.02	5.1	0	0.69	0	1.005066	-10.192072	110	1	1.1	0.9;
	10818	1	-0	-0	0	0.63	0	0.998652	13.237944	220	5	1.1	0.9;
	10819	1	-0	-0	0	14.06	0	0.958395	15.135397	400	1	1.1	0.9;
	10820	1	41.5	5.9	0	8.21	0	1.008397	27.531499	154	1	1.1	0.9;
	10821	1	105.67	0.56	0	18.05	0	0.987947	15.904596	380	5	1.1	0.9;
	10822	1	0	0	0	0.28	0	1.003446	-10.066789	150	9	1.1	0.9;
	10823	1	0	0	0	0	0	0.953175	-4.97832	150	9	1.1	0.9;
	10824	1	-0	-0	0.059551	4.150374	0	0.983355	-10.131998	220	22	1.1	0.9;
	10825	2	0	0	0	0	0	0.988205	29.491637	15	1	1.1	0.9;
	10826	1	36.34	45	0.594241	4.650845	0	1.005475	21.389486	330	24	1.1	0.9;
	10827	1	0	-0	0	4.04	0	1.038434	10.255024	220	1	1.1	0.9;
	10828	1	-0	-0	0	1.55	0	1.005207	32.006799	154	1	1.1	0.9;
	10829	2	0	0	0	0	0	1.03064	1.46507	6.3	1	1.1	0.9;
	10830	1	-0	0	0	1.68	0	0.987919	-7.45049	150	9	1.1	0.9;
	10831	1	-0	0	0	12.79	0	1.098505	65.011053	380	8	1.2	0.8;
	10832	1	0	0	0	4.09	0	1.022588	-10.912578	150	9	1.1	0.9;
	10833	1	-0	-0	0	0	0	1.020557	9.624602	20	14	1.1	0.9;
	10834	1	16.62	5.5	0	1.68	0	1.009718	-10.337767	110	12	1.1	0.9;
	10835	2	0	0	0	0	0	0.990208	39.96266	13.8	1	1.1	0.9;
	10836	1	-0	-0	0	0.04	0	1.029181	34.311007	220	3	1.1	0.9;
	10837	1	142	27.8	0	17.77	0	1.02339	-27.351175	220	4	1.1	0.9;
	10838	1	153.3	27.6	0	23.2	0	1.022449	29.009397	380	3	1.1	0.9;
	10839	1	55.66	22.4	0	2.11	0	1.016054	37.442406	110	8	1.1	0.9;
	10840	1	-0	-0	0	87.69	0	1.021683	-16.312902	380	4	1.1	0.9;
	10841	2	0	0	0	0	0	1.012871	-6.113908	20	10	1.1	0.9;
	10842	1	0	0	0	5.68	0	1.002744	9.696079	220	5	1.1	0.9;
	10843	1	41.45	17.3	0	0.93	0	0.980965	-5.333412	110	18	1.1	0.9;
	10844	1	100.3	17.2	0	6.63	0	0.996299	-23.298039	220	4	1.1	0.9;
	10845	2	0	0	0	0	0	1.050901	18.201002	20	7	1.1	0.9;
	10846	1	168.9	113	0	123.25	0	1.016226	41.638187	380	3	1.1	0.9;
	10847	2	0	0	0	0	0	1.127948	6.872839	20	9	1.2	0.8;
	10848	1	30.03	18.6	0	0.47	0	1.017525	-4.63816	150	9	1.1	0.9;
	10849	1	187	108.2	0.459362	5.986338	0	1.045058	31.533985	220	13	1.1	0.9;
	10850	1	-0	-0	0.179185	0.330679	0	1.007631	33.145748	220	3	1.1	0.9;
	10851	1	0	0	0	0	0	1.024793	-7.141759	150	9	1.1	0.9;
	10852	1	-0	0	0	1.13	0	1.03446	9.76169	220	5	1.1	0.9;
	10853	1	59.67	6	0	3.6	0	0.993185	-24.339386	220	6	1.1	0.9;
	10854	1	19.02	5.4	0	0.63	0	1.024449	39.046761	110	8	1.1	0.9;
	10855	1	0	-0.8	0	0.39	0	1.030884	1.632423	110	3	1.1	0.9;
	10856	1	94.5	-27.8	0	5.52	0	1.027891	-20.15245	220	4	1.1	0.9;
	10857	1	46.35	28.7	0	16.37	0	1.005209	-3.078095	150	9	1.1	0.9;
	10858	1	0	0	0	127.01	0	0.985439	14.751794	400	1	1.1	0.9;
	10859	1	50.06	-30.4	0.175758	2.103299	0	1.046888	-1.764346	110	23	1.1	0.9;
	10860	1	0	-0	0	1.03	0	0.875714	-14.918506	150	9	1.2	0.8;
	10861	1	0	0	0.156722	26.853278	0	1.024702	5.558519	380	7	1.1	0.9;
	10862	2	0	0	0	0	0	1.017582	-1.17871	20	10	1.1	0.9;
	10863	1	0	-0	0	0.68	0	0.994421	-11.286237	150	10	1.1	0.9;
	10864	1	-0	0	0	1.49	0	1.051216	90.952815	220	3	1.1	0.9;
	10865	2	0	0	0	0	0	1.030566	-6.658147	20	9	1.1	0.9;
	10866	1	33.64	13.7	0	0.75	0	1.016275	-26.563495	220	4	1.1	0.9;
	10867	1	23.43	8.5	0	0.28	0	0.9937	-4.748994	110	21	1.1	0.9;
	10868	1	16.22	5	0	0.18	0	0.993792	-1.394335	110	18	1.1	0.9;
	10869	1	0	0	0.150434	36.284268	0	1.039547	93.884512	380	1	1.1	0.9;
	10870	1	106.4	19	0	4.65	0	1.014981	-18.306895	220	4	1.1	0.9;
	10871	1	-0	0	0.099359	90.374619	0	1.078549	2.923312	380	18	1.2	0.8;
	10872	1	18.02	12.2	0	0.53	0	0.960372	-0.352042	110	18	1.1	0.9;
	10873	1	415.5	103.4	0	37.13	0	1.006471	-4.581022	380	6	1.1	0.9;
	10874	1	168.3	-57.1	0	16.319522	0	1.01123	-2.950698	380	6	1.1	0.9;
	10875	1	8.11	2.7	0	2.03	0	1.013565	-8.752258	110	12	1.1	0.9;
	10876	1	-23.83	7.8	0	2.3	0	1.036364	0.609659	150	9	1.1	0.9;
	10877	1	65.67	25.6	0	2.34	0	1.013398	18.391176	220	5	1.1	0.9;
	10878	1	-0	-0	0	7.3	0	1.015022	-12.727541	220	4	1.1	0.9;
	10879	1	123.1	20.9	0	25.95	0	1.004452	15.490712	380	3	1.1	0.9;
	10880	1	4.2	0.6	0	2.98	0	0.99896	35.843784	154	1	1.1	0.9;
	10881	2	0	0	0	0	0	1.002675	8.668171	20	18	1.1	0.9;
	10882	1	1.8	0.6	0	0.79	0	1.013081	3.304943	110	18	1.1	0.9;
	10883	1	411.3	168	0	0	0	1.029528	26.573846	150	2	1.1	0.9;
	10884	1	-0	0	0	120.17	0	1.016682	-22.865962	380	4	1.1	0.9;
	10885	1	0	-0	0	37.39	0	0.974842	19.011133	400	1	1.1	0.9;
	10886	1	162.5	48.4	0	5.73	0	0.983405	-17.874163	220	4	1.1	0.9;
	10887	1	0	0	0	7.21	0	1.050632	28.441322	220	1	1.1	0.9;
	10888	1	0	-0	0.526584	7.230435	0	1.018307	-5.887967	220	4	1.1	0.9;
	10889	1	0	-0	0.039495	0.960505	0	1.015268	-1.232652	220	19	1.1	0.9;
	10890	1	116.8	59	0	1.63	0	0.995518	-13.198494	150	9	1.1	0.9;
	10891	2	0	0	0	0	0	1.020801	18.620106	11	1	1.1	0.9;
	10892	1	-0	-0	0	13.33	0	1.006256	26.531463	400	1	1.1	0.9;
	10893	1	0	-0	0	0.46	0	0.990944	-9.604254	150	9	1.1	0.9;
	10894	1	-0	-0	0	19.61	0	1.014555	-22.095801	380	4	1.1	0.9;
	10895	1	18.32	5.3	0	7.32	0	1.018347	4.40155	220	5	1.1	0.9;
	10896	1	121.3	20.7	0	2.39	0	1.028201	18.476555	220	5	1.1	0.9;
	10897	1	12.71	2.5	0	0.56	0	0.957636	-5.912977	110	18	1.1	0.9;
	10898	2	0	0	0	0	0	1.045295	35.245517	20	5	1.1	0.9;
	10899	1	-0	0	0	22.76	0	1.017068	57.812163	380	3	1.1	0.9;
	10900	1	5.44	2.14	0	0.93	0	1.02095	3.804651	110	18	1.1	0.9;
	10901	2	0	0	0	0	0	1.041629	38.466941	20	2	1.1	0.9;
	10902	1	-0	-0	0	36.77	0	1.017885	42.294482	150	8	1.1	0.9;
	10903	1	0	0	0	17.05	0	1.038112	42.316461	380	3	1.1	0.9;
	10904	1	111.9	118.3	0	11.48	0	1.02681	37.543364	380	3	1.1	0.9;
	10905	1	-0	0	0	3.24	0	1.031605	79.170309	220	3	1.1	0.9;
	10906	1	-0	0	0.030087	2.639913	0	1.007106	-18.296961	220	4	1.1	0.9;
	10907	2	0	0	0	0	0	1.017638	-6.936102	20	4	1.1	0.9;
	10908	2	0	0	0	0	0	1.021858	-2.534901	20	6	1.1	0.9;
	10909	1	34.9	5	0	3.05	0	1.003916	16.769714	154	1	1.1	0.9;
	10910	1	0	-0	0	92.73	0	1.061466	10.803282	380	16	1.1	0.9;
	10911	1	15.96	2.18	0	3.58	0	0.982556	88.950654	150	1	1.1	0.9;
	10912	1	108.8	28.1	0	30.9	0	1.018445	36.653084	150	8	1.1	0.9;
	10913	2	0	0	0	0	0	1.063361	65.621633	20	3	1.1	0.9;
	10914	1	227.49	61.6	0	13.7	0	1.000943	33.854292	220	3	1.1	0.9;
	10915	1	49.45	16.5	0	4.13	0	1.004033	-14.136758	110	22	1.1	0.9;
	10916	1	223.1	31.8	0	23.58	0	1.034574	-1.223747	154	1	1.1	0.9;
	10917	1	-0	-0	0	3.49	0	1.07271	19.588991	220	3	1.1	0.9;
	10918	2	0	0	0	0	0	1.046905	39.305577	20	2	1.1	0.9;
	10919	1	0	0	0.135493	90.112445	0	1.089202	7.894075	380	19	1.2	0.8;
	10920	1	9.01	3.3	0	0.62	0	1.01303	-0.115586	110	19	1.1	0.9;
	10921	1	96.2	23.3	0	2.05	0	1.078292	24.313687	220	3	1.1	0.9;
	10922	1	0	0	0	83.200565	0	1.060681	4.426558	380	20	1.1	0.9;
	10923	1	-7.41	-1.6	0	2	0	1.021614	6.608806	220	7	1.1	0.9;
	10924	1	0	-0	0	37.99	0	1.026724	-18.974379	380	4	1.1	0.9;
	10925	1	10.11	3.7	0	0.76	0	0.98864	-1.800447	110	19	1.1	0.9;
	10926	1	-0	-0	0	1.77	0	1.044116	-1.932699	150	9	1.1	0.9;
	10927	2	0	0	0	0	0	1.03324	-5.012105	20	4	1.1	0.9;
	10928	2	0	0	0	0	0	1.03324	-5.012105	20	4	1.1	0.9;
	10929	1	178.7	31.3	0	5.6	0	1.013059	-26.461759	220	4	1.1	0.9;
	10930	2	0	0	0	0	0	1.03324	-5.012105	20	4	1.1	0.9;
	10931	1	45.45	7.4	0.039206	1.230596	0	0.992578	88.955071	150	1	1.1	0.9;
	10932	2	0	0	0	0	0	0.989744	5.658378	15	1	1.1	0.9;
	10933	1	240.63	27.93	0.122637	7.217363	0	1.013242	-8.26672	220	5	1.1	0.9;
	10934	1	-11.14	-3.38	0	2.35	0	1.015902	4.400211	220	7	1.1	0.9;
	10935	2	0	0	0	0	0	1.026025	16.98036	20	14	1.1	0.9;
	10936	1	15.72	5.2	0	0.76	0	0.948098	-8.048171	110	21	1.1	0.9;
	10937	1	1.5	0.4	0	0.66	0	0.966707	0.298451	110	18	1.1	0.9;
	10938	2	0	0	0	0	0	1.034618	3.213705	11	1	1.1	0.9;
	10939	1	13.72	8.5	0	19.46	0	1.094178	-13.789895	150	9	1.2	0.8;
	10940	2	0	0	0	0	0	0.963524	49.352991	13.8	1	1.1	0.9;
	10941	1	-18.95	-6.93	0	33.134947	0	0.987164	22.924227	380	5	1.1	0.9;
	10942	1	-0	0	0	20.52	0	1.044422	11.906623	380	7	1.1	0.9;
	10943	1	251.47	90.86	0	14.54	0	1.012774	11.698457	220	5	1.1	0.9;
	10944	1	0	0	0.061301	0.038699	0	1.004017	15.367102	220	5	1.1	0.9;
	10945	2	0	0	0	0	0	0.987313	21.230551	3.3	1	1.1	0.9;
	10946	2	0	0	0	0	0	1.032433	27.631736	20	10	1.1	0.9;
	10947	1	6.41	2.3	0	0.93	0	0.964957	-4.188803	110	19	1.1	0.9;
	10948	1	59.67	27.1	0	43.03	0	1.014833	32.681428	150	8	1.1	0.9;
	10949	1	67.07	35.5	0	32.32	0	0.975998	-29.032562	220	10	1.1	0.9;
	10950	1	47.95	29.7	0	0.23	0	0.950398	-10.506843	150	9	1.1	0.9;
	10951	2	0	0	0	0	0	0.979833	25.164622	11	1	1.1	0.9;
	10952	1	4.5	1.1	0	12.56	0	1.040375	-20.27869	380	10	1.1	0.9;
	10953	1	-0.18	-3.54	0	0	0	1.021504	43.086372	150	8	1.1	0.9;
	10954	2	0	0	0	0	0	1.029835	-14.681025	20	4	1.1	0.9;
	10955	1	0	0	0	0.04	0	1.009407	30.297446	154	1	1.1	0.9;
	10956	1	0	0	0	0.04	0	0.966842	-9.620847	150	9	1.1	0.9;
	10957	1	-1316.59	-129.9	0.227758	22.625182	0	1.042113	78.257418	380	3	1.1	0.9;
	10958	1	0.4	0.1	0	0.93	0	1.013008	-2.721836	110	19	1.1	0.9;
	10959	1	44.35	10.8	0	8.72	0	1.030177	4.029448	220	5	1.1	0.9;
	10960	1	4.41	2.7	0	1.58	0	0.965702	-5.056087	110	18	1.1	0.9;
	10961	1	-0	-0	0	5.97	0	1.031993	19.778436	220	5	1.1	0.9;
	10962	1	0.2	0.7	0	0.04	0	0.996576	38.818644	220	3	1.1	0.9;
	10963	2	0	0	0	0	0	1.056762	11.950776	10.5	1	1.1	0.9;
	10964	1	70.08	35	0	0.33	0	1.004147	-22.172201	220	6	1.1	0.9;
	10965	1	198.6	282.2	0.11228	170.076433	0	1.07444	72.668763	380	3	1.2	0.8;
	10966	1	-0	0	0	4.29	0	1.012121	9.272005	220	5	1.1	0.9;
	10967	1	0	0	0	43.34	0	1.049435	44.907943	380	3	1.1	0.9;
	10968	1	0	-0	0	4.14	0	1.035837	30.303383	220	13	1.1	0.9;
	10969	1	-6.79	-1.44	0	6.66	0	0.996139	-3.970212	220	6	1.1	0.9;
	10970	1	139.4	21.2	0	6.21	0	0.995719	11.936633	220	5	1.1	0.9;
	10971	2	0	0	0	0	0	0.992395	35.295433	11	1	1.1	0.9;
	10972	1	10.11	3.7	0	0.25	0	1.001455	-1.412213	110	19	1.1	0.9;
	10973	2	0	0	0	0	0	1.015065	20.651679	11	1	1.1	0.9;
	10974	1	4.4	2.2	0	0.29	0	1.000402	-14.170487	110	22	1.1	0.9;
	10975	1	160.9	22.9	0	41.42	0	0.965497	21.450186	400	1	1.1	0.9;
	10976	1	100.6	49.6	0.12159	-0.101325	0	0.977661	5.932781	110	18	1.1	0.9;
	10977	1	-0	-0	0	0	0	1.015642	-21.99726	20	4	1.1	0.9;
	10978	2	0	0	0	0	0	1.017582	-1.17871	20	10	1.1	0.9;
	10979	1	136.2	4.8	0	7.58	0	1.042714	13.161129	220	3	1.1	0.9;
	10980	1	-2.3	-0.8	0	0.33	0	1.056556	-2.749188	110	19	1.1	0.9;
	10981	1	0	-0	0.009549	0.550451	0	1.057896	29.562711	150	2	1.1	0.9;
	10982	1	-0	0	0	1.86	0	0.989913	-15.56772	220	10	1.1	0.9;
	10983	2	0	0	0	0	0	1.024499	-3.834468	20	20	1.1	0.9;
	10984	1	-0	0	0	3.51	0	1.006745	-15.282702	220	4	1.1	0.9;
	10985	2	0	0	0	0	0	1.023338	0.199966	20	6	1.1	0.9;
	10986	1	-0	-0	0	0	0	1.039787	-14.573329	20	4	1.1	0.9;
	10987	2	0	0	0	0	0	1.018866	14.228745	20	10	1.1	0.9;
	10988	1	22.3	3.1	0	8.81	0	1.021238	18.63529	154	1	1.1	0.9;
	10989	1	4	0.8	0	17.65	0	1.036108	5.806299	220	5	1.1	0.9;
	10990	1	56.26	8.5	0	5.32	0	1.014119	-26.1744	220	4	1.1	0.9;
	10991	1	-10.06	-3.91	0	14.08	0	1.004444	63.292218	220	3	1.1	0.9;
	10992	1	46.35	14.9	0	20.62	0	1.033683	-19.351445	220	4	1.1	0.9;
	10993	1	0.1	0	0	5.01	0	1.035529	26.984517	220	13	1.1	0.9;
	10994	2	0	0	0	0	0	0.964368	44.185324	11	1	1.1	0.9;
	10995	1	0	0	0	2.28	0	1.00521	30.518109	380	3	1.1	0.9;
	10996	1	0	0	0	38.570506	0	0.992371	15.325908	380	5	1.1	0.9;
	10997	1	1.4	100	0	21.43	0	1.011697	51.097732	220	3	1.1	0.9;
	10998	2	0	0	0	0	0	1.05787	-5.405768	20	4	1.1	0.9;
	10999	1	0	-0	0.090774	3.150444	0	1.056895	7.793415	220	18	1.2	0.8;
	11000	1	3.8	0.6	0	0.68	0	0.965871	1.76722	110	18	1.1	0.9;
	11001	1	9.01	4	0	1.39	0	1.04564	-1.473285	110	23	1.1	0.9;
	11002	1	-14.89	-4.46	0	0.3	0	0.960912	0.645663	110	18	1.1	0.9;
	11003	1	97.97	31.8	0	5.02	0	1.01601	18.41194	220	5	1.1	0.9;
	11004	1	0	-0	0.49396	13.698454	0	0.981057	20.248958	380	5	1.1	0.9;
	11005	2	0	0	0	0	0	1.042317	14.274354	20	14	1.1	0.9;
	11006	1	62.7	9	0	4.6	0	1.00026	19.167902	154	1	1.1	0.9;
	11007	1	-0	-0	0.15	5.08	0	0.985204	-4.275915	220	21	1.1	0.9;
	11008	1	0	-0	0.07	3.65	0	1.034643	-5.112265	220	1	1.1	0.9;
	11009	1	107.2	45.6	0	5.28	0	1.023224	9.526501	220	3	1.1	0.9;
	11010	2	0	0	0	0	0	1.064335	2.396339	20	20	1.1	0.9;
	11011	1	-8.66	-6.01	0	0	0	1.02543	42.897015	150	8	1.1	0.9;
	11012	1	0	0	0	1.88	0	1.016333	31.296283	380	2	1.1	0.9;
	11013	1	125.4	17.8	0	7.64	0	0.99812	29.873325	154	1	1.1	0.9;
	11014	1	211	59.3	0	3.46	0	0.997828	-4.430017	110	21	1.1	0.9;
	11015	1	0	-0	0	42.361623	0	0.993099	14.527357	380	5	1.1	0.9;
	11016	1	-0	-0	0	13.62	0	1.018455	3.690325	220	7	1.1	0.9;
	11017	1	-0	0	0	31.242294	0	1.030905	9.647854	380	5	1.1	0.9;
	11018	1	26.23	8.4	0	0.96	0	0.99801	-10.108962	110	12	1.1	0.9;
	11019	1	13.82	5	0	0.43	0	1.130738	7.126476	220	19	1.2	0.8;
	11020	2	0	0	0	0	0	0.991929	29.398008	6.3	1	1.1	0.9;
	11021	1	43.15	15.4	0	0.36	0	1.085458	1.301643	110	15	1.1	0.9;
	11022	1	20.62	6.7	0	0.71	0	0.964159	-6.496671	110	21	1.1	0.9;
	11023	2	0	0	0	0	0	1.031695	0	20	6	1.1	0.9;
	11024	1	-0	-0	0	1.98	0	0.963863	-9.047548	150	9	1.1	0.9;
	11025	2	0	0	0	0	0	1.031695	0	20	6	1.1	0.9;
	11026	1	-0	-0	0	17.166964	0	0.982822	15.001687	380	5	1.1	0.9;
	11027	1	130.5	17.2	0	5.65	0	0.99836	-22.714373	220	6	1.1	0.9;
	11028	2	0	0	0	0	0	1.025977	8.588106	20	6	1.1	0.9;
	11029	1	40.85	15.7	0	2.08	0	1.023866	16.659339	220	5	1.1	0.9;
	11030	2	0	0	0	0	0	1.021495	11.964278	20	18	1.1	0.9;
	11031	1	186.8	22.9	0	9.94	0	1.023514	-27.422646	220	4	1.1	0.9;
	11032	2	0	0	0	0	0	1.04071	74.319379	20	3	1.1	0.9;
	11033	1	346.42	89.04	0	43.55	0	1.038753	13.420474	380	5	1.1	0.9;
	11034	1	-0	-0	0	0.53	0	1.071071	-4.191594	150	9	1.1	0.9;
	11035	1	72.48	44.9	0	3.86	0	0.989728	-13.568538	150	9	1.1	0.9;
	11036	1	29.65	7.18	0	4.51	0	1.013145	93.290213	150	1	1.1	0.9;
	11037	2	0	0	0	0	0	1.029671	48.863044	20	10	1.1	0.9;
	11038	1	0	0	0	0	0	0.926519	-12.060065	150	9	1.1	0.9;
	11039	2	0	0	0	0	0	1.058835	5.385711	20	20	1.1	0.9;
	11040	1	281.3	40.9	0	33.56	0	1.014113	-10.382927	220	11	1.1	0.9;
	11041	1	20.92	9.35	0	1.56	0	0.99966	4.257778	110	18	1.1	0.9;
	11042	1	-0	-0	0	5.55	0	1.017096	-26.064396	220	1	1.1	0.9;
	11043	1	101.48	19.53	0	4.41	0	1.019788	21.776211	220	13	1.1	0.9;
	11044	1	0	-0	0	0.31	0	1.032592	-10.840409	150	9	1.1	0.9;
	11045	1	-0	-0	0	22.68	0	1.027749	35.91349	380	13	1.1	0.9;
	11046	1	159.98	76.4	0	22.64	0	1.022149	51.967191	220	3	1.1	0.9;
	11047	2	0	0	0	0	0	0.956299	35.305913	14	1	1.1	0.9;
	11048	1	0	0	0	3.1	0	1.057421	-9.850777	220	1	1.1	0.9;
	11049	2	0	0	0	0	0	1.038747	-8.642337	20	4	1.1	0.9;
	11050	1	34.44	12.1	0	0.87	0	1.048761	-1.634623	110	15	1.1	0.9;
	11051	2	0	0	0	0	0	1.019231	22.327173	20	24	1.1	0.9;
	11052	1	90.5	12.8	0	1.17	0	1.000581	2.016131	154	1	1.1	0.9;
	11053	1	8.98	-3.79	0	2.86	0	0.998763	89.131252	150	1	1.1	0.9;
	11054	1	34.64	12.4	0	2.66	0	0.965663	-3.946676	110	19	1.1	0.9;
	11055	1	0	-0	0	20.54	0	0.990963	29.255516	380	1	1.1	0.9;
	11056	1	218	18.6	0	36.29	0	1.041941	12.254953	380	16	1.1	0.9;
	11057	1	23.03	8.3	0	1.05	0	1.01539	-4.024409	110	19	1.1	0.9;
	11058	1	531.7	-12.8	0	0	0	1.043702	24.901421	150	2	1.1	0.9;
	11059	2	0	0	0	0	0	0.993383	32.126146	11	1	1.1	0.9;
	11060	1	-0	0	0	0.81	0	1.03274	40.004966	110	8	1.1	0.9;
	11061	1	0	0	0	0	0	1.020343	-11.016395	150	9	1.1	0.9;
	11062	1	13.9	1.9	0	2.23	0	1.016593	33.004723	154	1	1.1	0.9;
	11063	1	0	-0	0.058317	0.581683	0	1.011448	-5.405014	220	12	1.1	0.9;
	11064	1	-0	0	0	1.18	0	0.95216	-28.786392	150	10	1.1	0.9;
	11065	2	0	0	0	0	0	1.014542	2.91424	20	18	1.1	0.9;
	11066	1	-0	0	0	13.9	0	1.041703	17.122885	220	5	1.1	0.9;
	11067	1	0	-0	0	0.73	0	1.065465	-5.509476	150	9	1.1	0.9;
	11068	1	-0	-0	0	48.75	0	1.084957	7.346331	380	19	1.2	0.8;
	11069	1	126.6	28.7	0	15.02	0	1.015628	-26.653114	220	4	1.1	0.9;
	11070	1	-0	0	0	12.4	0	1.049335	-12.691784	380	1	1.1	0.9;
	11071	1	0	-0	0	2.12	0	1.017049	-8.947511	220	6	1.1	0.9;
	11072	1	29.03	6.1	0	0.44	0	1.021208	38.056929	110	8	1.1	0.9;
	11073	1	-0	0	0	0.5	0	1.028931	19.248313	220	5	1.1	0.9;
	11074	1	-0	0	0	6.69	0	1.042079	21.032687	220	5	1.1	0.9;
	11075	1	45.95	15.4	0	0.6	0	0.997963	-4.774298	110	18	1.1	0.9;
	11076	2	0	0	0	0	0	1.011132	1.192229	20	18	1.1	0.9;
	11077	1	136.9	11.4	0	3.05	0	1.025183	2.063528	220	5	1.1	0.9;
	11078	1	0	-0	0	0.17	0	1.025593	17.098854	220	5	1.1	0.9;
	11079	1	32.1	4.5	0	3.75	0	1.004312	24.594716	154	1	1.1	0.9;
	11080	1	300.1	73.7	0.038936	1.881064	0	1.032381	14.273943	220	5	1.1	0.9;
	11081	1	5.01	1.8	0	3.68	0	1.030478	28.680093	220	5	1.1	0.9;
	11082	1	127.6	45.3	0	57.43	0	1.019106	25.890575	150	8	1.1	0.9;
	11083	1	-0	0	0	68.73	0	1.021648	-12.117867	380	4	1.1	0.9;
	11084	1	0	0	0	0	0	1.023166	4.463602	20	23	1.1	0.9;
	11085	2	0	0	0	0	0	1.038916	11.288829	20	23	1.1	0.9;
	11086	2	0	0	0	0	0	1.038916	11.288829	20	23	1.1	0.9;
	11087	2	0	0	0	0	0	1.038916	11.288829	20	23	1.1	0.9;
	11088	2	0	0	0	0	0	1.038916	11.288829	20	23	1.1	0.9;
	11089	1	-0	-0	0	0.87	0	0.950334	-10.842503	150	9	1.1	0.9;
	11090	2	0	0	0	0	0	1.042693	10.439455	20	23	1.1	0.9;
	11091	1	10.11	4.4	0	0.29	0	0.986965	-2.623333	110	18	1.1	0.9;
	11092	1	50.26	19.3	0	2.1	0	1.031473	33.288433	220	3	1.1	0.9;
	11093	1	-0	0	0	0.71	0	0.974554	-6.12814	110	21	1.1	0.9;
	11094	1	-0.66	-2.79	0	2.49	0	1.074857	-3.74166	150	9	1.1	0.9;
	11095	2	0	0	0	0	0	1.026425	15.756693	20	5	1.1	0.9;
	11096	2	0	0	0	0	0	1.024096	61.613978	20	10	1.1	0.9;
	11097	1	11.21	3.7	0	0.9	0	0.99605	-9.027342	110	12	1.1	0.9;
	11098	1	-0.2	-1	0	0	0	1.042796	26.265727	150	8	1.1	0.9;
	11099	2	0	0	0	0	0	1.026957	18.072383	20	5	1.1	0.9;
	11100	1	58.77	16.1	0	4.11	0	1.007503	-16.273277	220	6	1.1	0.9;
	11101	1	116.5	33.4	0	0.28	0	1.019883	12.179054	110	19	1.1	0.9;
	11102	2	0	0	0	0	0	1.035958	27.387404	20	5	1.1	0.9;
	11103	1	30.43	10	0	0.41	0	0.958617	-8.042541	110	21	1.1	0.9;
	11104	2	0	0	0	0	0	1.035958	27.387404	20	5	1.1	0.9;
	11105	1	16.42	5.3	0	0.43	0	0.990461	-2.437161	110	18	1.1	0.9;
	11106	2	0	0	0	0	0	1.053822	27.723432	20	5	1.1	0.9;
	11107	2	0	0	0	0	0	1.035958	27.387404	20	5	1.1	0.9;
	11108	2	0	0	0	0	0	1.053822	27.723432	20	5	1.1	0.9;
	11109	2	0	0	0	0	0	1.053822	27.723432	20	5	1.1	0.9;
	11110	2	0	0	0	0	0	1.01638	20.943598	20	5	1.1	0.9;
	11111	2	0	0	0	0	0	1.053822	27.723432	20	5	1.1	0.9;
	11112	1	18.02	4	0	3.3	0	1.034476	-0.01541	220	11	1.1	0.9;
	11113	1	0.5	0	0	0.84	0	0.948716	-7.342315	110	18	1.1	0.9;
	11114	1	0	-0	0	2.13	0	1.029326	26.316485	220	5	1.1	0.9;
	11115	2	0	0	0	0	0	1.019144	17.562907	20	14	1.1	0.9;
	11116	1	119	31.7	0	1.94	0	1.013758	-18.448945	220	4	1.1	0.9;
	11117	1	-0	-0	0	0.62	0	1.047639	85.933932	220	3	1.1	0.9;
	11118	2	0	0	0	0	0	1.042829	13.668895	20	5	1.1	0.9;
	11119	2	0	0	0	0	0	1.042829	13.668895	20	5	1.1	0.9;
	11120	2	0	0	0	0	0	1.035471	18.239403	20	5	1.1	0.9;
	11121	2	0	0	0	0	0	1.035471	18.239403	20	5	1.1	0.9;
	11122	2	0	0	0	0	0	1.055315	14.241147	20	5	1.1	0.9;
	11123	1	119.7	43.4	0	0.92	0	1.012226	7.516602	220	3	1.1	0.9;
	11124	2	0	0	0	0	0	1.042829	13.668895	20	5	1.1	0.9;
	11125	2	0	0	0	0	0	1.042829	13.668895	20	5	1.1	0.9;
	11126	1	5.51	2.2	0	1.43	0	0.975009	-15.542069	110	22	1.1	0.9;
	11127	1	87	30.8	0	2.43	0	0.96051	-7.992356	110	21	1.1	0.9;
	11128	2	0	0	0	0	0	1.043665	14.930439	20	14	1.1	0.9;
	11129	2	0	0	0	0	0	1.027132	31.963051	20	5	1.1	0.9;
	11130	2	0	0	0	0	0	1.035471	18.239403	20	5	1.1	0.9;
	11131	2	0	0	0	0	0	1.035771	2.999441	20	14	1.1	0.9;
	11132	2	0	0	0	0	0	1.028204	5.666749	20	6	1.1	0.9;
	11133	2	0	0	0	0	0	1.028204	5.666749	20	6	1.1	0.9;
	11134	1	13.11	2.8	0.134137	1.835863	0	1.040116	6.4776	110	21	1.1	0.9;
	11135	1	8.31	3	0	1.27	0	0.949807	-0.06613	110	19	1.1	0.9;
	11136	1	24.03	-56.8	0	2.69	0	1.020968	24.881656	220	3	1.1	0.9;
	11137	2	0	0	0	0	0	1.028204	5.666749	20	6	1.1	0.9;
	11138	2	0	0	0	0	0	1.016735	24.349916	11	1	1.1	0.9;
	11139	1	26.82	24.2	0	2.89	0	0.979788	-12.77143	150	9	1.1	0.9;
	11140	1	115.4	31	0	28.85	0	1.033483	-19.436094	220	4	1.1	0.9;
	11141	1	-0	0	0	0	0	1.020252	-22.27659	20	4	1.1	0.9;
	11142	1	-0	0	0	0	0	1.020252	-22.27659	20	4	1.1	0.9;
	11143	1	-0	0	0	0	0	1.020252	-22.27659	20	4	1.1	0.9;
	11144	1	-0	0	0	0	0	1.020252	-22.27659	20	4	1.1	0.9;
	11145	2	0	0	0	0	0	1.000808	-3.647252	20	18	1.1	0.9;
	11146	1	0	-0	0	9.84	0	1.032343	4.317275	220	5	1.1	0.9;
	11147	1	6.31	2.5	0	0.44	0	0.920261	-7.851533	110	18	1.1	0.9;
	11148	1	41.8	5.9	0	2.54	0	1.006718	31.655284	154	1	1.1	0.9;
	11149	1	70.48	17.6	0	3.92	0	0.976309	-26.254285	220	10	1.1	0.9;
	11150	1	-0	-0	0	0.15	0	0.997555	-9.051308	150	9	1.1	0.9;
	11151	1	0	-0	0.205709	31.487194	0	1.042487	12.714635	380	3	1.1	0.9;
	11152	1	-0	0	0	1.69	0	1.002501	0.438398	220	6	1.1	0.9;
	11153	1	-0	0	0.078682	0.390988	0	1.023525	-7.363087	380	6	1.1	0.9;
	11154	2	0	0	0	0	0	1.07025	2.362941	20	9	1.1	0.9;
	11155	2	0	0	0	0	0	1.027264	14.46311	20	14	1.1	0.9;
	11156	1	136.95	87.7	0	7.84	0	0.992653	32.856803	220	3	1.1	0.9;
	11157	1	-0	-0	0	0.84	0	1.01561	-12.050754	150	9	1.1	0.9;
	11158	2	0	0	0	0	0	0.991765	36.62363	6.3	1	1.1	0.9;
	11159	1	-0	-0	0	14.94	0	1.049883	28.076782	220	1	1.1	0.9;
	11160	1	90.5	12.8	0	6.78	0	0.998522	27.102964	154	1	1.1	0.9;
	11161	1	32.74	10.5	0	1.78	0	1.0158	9.843152	220	5	1.1	0.9;
	11162	1	164.41	101.14	0	61.4	0	1.058677	31.63551	380	13	1.1	0.9;
	11163	1	78.89	45.6	0	0.04	0	1.024339	19.049223	220	5	1.1	0.9;
	11164	1	0	-0	0.058918	0.141082	0	1.000841	-3.260781	110	18	1.1	0.9;
	11165	1	0	0	0.165195	-0.128485	0	1.070567	30.667523	150	8	1.1	0.9;
	11166	1	-0	0	0	1.15	0	1.023141	-27.708185	220	4	1.1	0.9;
	11167	2	0	0	0	0	0	1.040259	34.3106	20	5	1.1	0.9;
	11168	1	10.31	3.8	0.090914	4.82939	0	1.003521	-17.064676	220	4	1.1	0.9;
	11169	1	0	-0	0	2.82	0	1.031203	65.423612	220	3	1.1	0.9;
	11170	2	0	0	0	0	0	0.962614	39.578299	15.75	1	1.1	0.9;
	11171	1	91.6	20.3	0	0	0	1.00831	-2.707293	220	6	1.1	0.9;
	11172	2	0	0	0	0	0	1.040259	34.3106	20	5	1.1	0.9;
	11173	1	5.5	0.8	0	1.9	0	0.99898	15.395595	154	1	1.1	0.9;
	11174	2	0	0	0	0	0	1.040259	34.3106	20	5	1.1	0.9;
	11175	2	0	0	0	0	0	1.040259	34.3106	20	5	1.1	0.9;
	11176	1	-0	-0	0	1.86	0	1.033872	21.357215	220	5	1.1	0.9;
	11177	2	0	0	0	0	0	1.01515	14.648403	11	1	1.1	0.9;
	11178	1	-12.61	11.6	0	0.02	0	1.045197	-4.510954	150	9	1.1	0.9;
	11179	1	0	-0	0.29	6.18	0	1.033898	34.422243	380	2	1.1	0.9;
	11180	1	429.5	163.1	0	28.67	0	1.06407	5.836436	220	11	1.1	0.9;
	11181	1	110.1	44.6	0.04004	2.98997	0	0.998612	-7.023033	220	6	1.1	0.9;
	11182	1	68.78	35	0	34.52	0	1.015748	23.459867	380	5	1.1	0.9;
	11183	1	0	0	0.494726	0.384626	0	1.030771	24.629866	220	5	1.1	0.9;
	11184	1	0	0	0	138.75	0	0.984017	31.265821	400	1	1.1	0.9;
	11185	1	0	-0	0	26.93	0	1.09594	-2.304833	380	9	1.2	0.8;
	11186	1	90.3	5.7	0	19.04	0	1.039325	61.949237	380	3	1.1	0.9;
	11187	1	124.1	-30.3	0	14.26	0	1.015857	-18.592394	220	4	1.1	0.9;
	11188	1	109.1	15.6	0	2.47	0	1.00546	18.561202	154	1	1.1	0.9;
	11189	1	-0	-0	0	33.48	0	1.055623	15.228766	380	1	1.1	0.9;
	11190	1	0	-0	0	0.1	0	1.040076	10.160103	220	5	1.1	0.9;
	11191	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	11192	1	14.92	5.4	0	0.97	0	1.015221	-1.987936	110	19	1.1	0.9;
	11193	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	11194	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	11195	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	11196	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	11197	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	11198	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	11199	1	29.13	11.3	0	0.16	0	0.975835	5.335155	110	18	1.1	0.9;
	11200	1	20.62	7.5	0	0.79	0	0.964611	-4.941161	110	19	1.1	0.9;
	11201	2	0	0	0	0	0	1.020879	-2.731653	20	6	1.1	0.9;
	11202	1	61.97	38.4	0	7.42	0	1.09911	-4.414302	150	9	1.1	0.9;
	11203	2	0	0	0	0	0	1.05667	-2.722877	20	5	1.1	0.9;
	11204	2	0	0	0	0	0	1.020879	-2.731653	20	6	1.1	0.9;
	11205	1	255.3	78.7	0	6.5	0	0.976943	30.990047	220	3	1.1	0.9;
	11206	1	33.94	11.4	0	0	0	1.029596	7.742794	220	5	1.1	0.9;
	11207	1	41.8	5.9	0	3.29	0	1.007227	17.02309	154	1	1.1	0.9;
	11208	1	117	15.2	0	2.87	0	1.028472	33.988037	220	3	1.1	0.9;
	11209	1	-0	-0	0.179063	9.690781	0	0.989842	-9.859665	220	22	1.1	0.9;
	11210	1	-0	0	0.216888	13.908297	0	1.134218	1.147159	380	9	1.2	0.8;
	11211	2	0	0	0	0	0	1.059581	78.307427	20	3	1.1	0.9;
	11212	2	0	0	0	0	0	1.049265	79.575419	20	3	1.1	0.9;
	11213	1	142.3	19.2	0.079316	13.420684	0	1.01937	-12.967405	220	4	1.1	0.9;
	11214	2	0	0	0	0	0	1.013527	-7.417807	20	10	1.1	0.9;
	11215	1	-0	0	0.090774	5.169044	0	1.047231	8.451346	220	18	1.1	0.9;
	11216	1	0	-0	0.11613	2.442903	0	1.031611	-6.339819	220	10	1.1	0.9;
	11217	1	4.81	1.6	0	0.19	0	0.99989	-8.802342	110	12	1.1	0.9;
	11218	1	-0.21	-2.37	0	0	0	1.018256	31.685067	150	8	1.1	0.9;
	11219	2	0	0	0	0	0	1.047402	52.472138	20	3	1.1	0.9;
	11220	2	0	0	0	0	0	1.045943	92.728902	20	3	1.1	0.9;
	11221	2	0	0	0	0	0	1.03058	15.553508	20	14	1.1	0.9;
	11222	1	37.14	2.7	0	2.33	0	1.033676	19.488235	220	5	1.1	0.9;
	11223	1	256.8	-61.8	0	36.73	0	1.023924	15.352257	380	6	1.1	0.9;
	11224	1	16.52	5.5	0	1.15	0	1.012524	-14.686613	110	22	1.1	0.9;
	11225	2	0	0	0	0	0	1.039009	81.664701	20	3	1.1	0.9;
	11226	2	0	0	0	0	0	1.028091	-15.249181	20	10	1.1	0.9;
	11227	2	0	0	0	0	0	1.044939	29.387061	20	5	1.1	0.9;
	11228	1	93.4	27.8	0	29.52	0	1.019273	36.364634	150	8	1.1	0.9;
	11229	2	0	0	0	0	0	1.009652	16.297978	20	5	1.1	0.9;
	11230	2	0	0	0	0	0	1.045545	9.62853	20	14	1.1	0.9;
	11231	2	0	0	0	0	0	1.044939	29.387061	20	5	1.1	0.9;
	11232	1	42.66	-22.22	0	0.66	0	1.03189	29.322101	220	1	1.1	0.9;
	11233	2	0	0	0	0	0	1.041114	89.192565	20	3	1.1	0.9;
	11234	1	-0	-0	0	0	0	1.020254	-23.064206	20	4	1.1	0.9;
	11235	2	0	0	0	0	0	1.05622	67.539298	20	3	1.1	0.9;
	11236	2	0	0	0	0	0	1.041114	89.192565	20	3	1.1	0.9;
	11237	2	0	0	0	0	0	1.015411	-5.476788	20	4	1.1	0.9;
	11238	2	0	0	0	0	0	1.05622	67.539298	20	3	1.1	0.9;
	11239	2	0	0	0	0	0	1.027155	-13.349627	20	4	1.1	0.9;
	11240	2	0	0	0	0	0	1.063297	57.37246	20	3	1.1	0.9;
	11241	2	0	0	0	0	0	1.063297	57.37246	20	3	1.1	0.9;
	11242	2	0	0	0	0	0	1.063297	57.37246	20	3	1.1	0.9;
	11243	1	67.07	25.2	0	8.11	0	1.029481	-21.176174	220	4	1.1	0.9;
	11244	2	0	0	0	0	0	1.015411	-5.476788	20	4	1.1	0.9;
	11245	1	63.07	7.1	0.039913	1.170065	0	1.004639	94.458816	150	1	1.1	0.9;
	11246	1	0	-0	0	0	0	1.028646	7.51961	20	3	1.1	0.9;
	11247	1	0	0	0	35.53	0	1.030205	22.436567	380	6	1.1	0.9;
	11248	1	20.82	7.9	0	42.36	0	1.033234	-15.461113	380	4	1.1	0.9;
	11249	1	0	-0	0.24033	46.858516	0	1.083816	-4.534997	380	1	1.2	0.8;
	11250	2	0	0	0	0	0	1.008948	-4.613451	20	10	1.1	0.9;
	11251	2	0	0	0	0	0	1.028648	-14.669346	20	10	1.1	0.9;
	11252	1	0	-0	0	52.89	0	1.093657	-0.10105	380	9	1.1	0.9;
	11253	1	66.07	21.6	0	0.75	0	0.981547	24.947698	150	8	1.1	0.9;
	11254	1	0	0	0.037655	48.632345	0	1.073833	64.707336	380	3	1.1	0.9;
	11255	1	-0	0	0	5.24	0	1.025855	17.158047	154	1	1.1	0.9;
	11256	2	0	0	0	0	0	1.03657	5.732096	20	6	1.1	0.9;
	11257	2	0	0	0	0	0	1.036997	8.018504	11	1	1.1	0.9;
	11258	1	-17.32	5.7	0	1.52	0	0.963667	-7.149819	150	9	1.1	0.9;
	11259	2	0	0	0	0	0	1.03657	5.732096	20	6	1.1	0.9;
	11260	1	36.94	15.2	0.196151	15.653849	0	1.025173	10.163799	220	7	1.1	0.9;
	11261	1	-0	-0	0	0.09	0	1.024433	-0.497913	150	9	1.1	0.9;
	11262	1	0	-0	0	0.44	0	1.024545	36.252791	154	1	1.1	0.9;
	11263	1	0	0	0	1.59	0	1.047483	87.056246	220	3	1.1	0.9;
	11264	2	0	0	0	0	0	1.03058	15.553508	20	14	1.1	0.9;
	11265	2	0	0	0	0	0	1.024762	17.972121	20	14	1.1	0.9;
	11266	1	0	0	0	0	0	1.091952	-4.839993	150	9	1.1	0.9;
	11267	1	49.45	6.5	0	3.53	0	1.029045	-16.851924	220	4	1.1	0.9;
	11268	1	-0	0	0	48.77	0	1.04893	0.483335	380	1	1.1	0.9;
	11269	1	90.2	30.3	0.039983	2.510017	0	1.00751	-6.732268	110	21	1.1	0.9;
	11270	1	12.71	5.1	0	15.31	0	1.015548	-22.604629	380	4	1.1	0.9;
	11271	1	95.6	13.6	0	21.47	0	1.018385	17.026556	154	1	1.1	0.9;
	11272	1	-0	-0	0	5.28	0	1.009712	31.100358	380	3	1.1	0.9;
	11273	2	0	0	0	0	0	0.96764	46.597759	11	1	1.1	0.9;
	11274	1	0	-0	0	96.48	0	1.001049	25.288942	380	5	1.1	0.9;
	11275	1	-0	-0	0.160142	3.503537	0	1.024574	2.187824	220	3	1.1	0.9;
	11276	1	-0	0	0	9.54	0	1.027666	10.98452	220	5	1.1	0.9;
	11277	1	25.13	15.6	0	4.51	0	1.03616	-10.170176	150	9	1.1	0.9;
	11278	1	-14.96	-4.7	0	0.04	0	1.008907	14.345628	220	5	1.1	0.9;
	11279	1	-0	0	0	10.69	0	0.991622	-5.970926	220	12	1.1	0.9;
	11280	1	10.91	2.5	0	1.34	0	0.971409	-8.469243	110	18	1.1	0.9;
	11281	2	0	0	0	0	0	1.03858	23.545079	11	1	1.1	0.9;
	11282	1	18.9	2.7	0	5.84	0	1.022038	27.354939	154	1	1.1	0.9;
	11283	1	0	-0	0.13352	0.05648	0	1.056779	25.810444	220	5	1.1	0.9;
	11284	1	-0	-0	0	0.06	0	1.053091	-3.843098	150	9	1.1	0.9;
	11285	2	0	0	0	0	0	1.032344	9.093698	20	18	1.1	0.9;
	11286	1	-0	-0	0	19.62	0	1.033669	33.596335	380	1	1.1	0.9;
	11287	1	66.17	27.4	0	14.12	0	1.026297	18.651014	220	5	1.1	0.9;
	11288	1	0	-0	0	0	0	1.019745	-19.936916	20	4	1.1	0.9;
	11289	2	0	0	0	0	0	1.035773	21.901715	20	5	1.1	0.9;
	11290	1	11.91	3.5	0	0.43	0	1.001925	28.994671	150	8	1.1	0.9;
	11291	1	-0	-0	0	0.81	0	0.960396	-8.243085	150	9	1.1	0.9;
	11292	1	51.96	6.5	0	5.14	0	1.001636	-3.640685	110	18	1.1	0.9;
	11293	1	0	0	0	0	0	1.025682	-7.286792	150	9	1.1	0.9;
	11294	2	0	0	0	0	0	1.01249	-3.900476	20	4	1.1	0.9;
	11295	1	48.65	16.1	0	7.14	0	1.032164	20.172702	220	5	1.1	0.9;
	11296	1	0	0	0	1.57	0	1.07219	19.888883	220	3	1.1	0.9;
	11297	1	131.3	19.1	0.039686	14.000314	0	1.019164	45.629261	220	13	1.1	0.9;
	11298	1	0.2	88.5	0	2.21	0	1.005342	94.829384	150	1	1.1	0.9;
	11299	2	0	0	0	0	0	1.065506	1.014062	20	9	1.1	0.9;
	11300	1	2.2	0.5	0	0.58	0	1.006151	-14.005277	220	10	1.1	0.9;
	11301	1	60.07	36	0	0.06	0	0.995276	-10.528333	150	9	1.1	0.9;
	11302	1	126.2	-12.5	0	10.61	0	1.09598	24.803662	220	3	1.2	0.8;
	11303	1	0	0	0	0	0	0.954084	-7.518624	150	9	1.1	0.9;
	11304	1	27.93	10.1	0	0.43	0	0.973019	-2.59664	110	19	1.1	0.9;
	11305	2	0	0	0	0	0	1.028153	78.700831	20	3	1.1	0.9;
	11306	1	15.12	4.9	0	0.68	0	0.971726	-12.437265	110	12	1.1	0.9;
	11307	2	0	0	0	0	0	1.04834	88.719592	20	3	1.1	0.9;
	11308	2	0	0	0	0	0	1.03671	88.578657	20	3	1.1	0.9;
	11309	2	0	0	0	0	0	1.035773	21.901715	20	5	1.1	0.9;
	11310	1	0	0	0	1.08	0	1.015777	34.284954	220	3	1.1	0.9;
	11311	2	0	0	0	0	0	0.982004	27.243336	11	1	1.1	0.9;
	11312	2	0	0	0	0	0	1.049674	87.363329	20	3	1.1	0.9;
	11313	1	0	0	0	4.17	0	1.012944	48.511806	220	8	1.1	0.9;
	11314	1	85.9	93.8	0	3.93	0	1.045933	36.252184	380	3	1.1	0.9;
	11315	2	0	0	0	0	0	1.024544	18.974877	20	14	1.1	0.9;
	11316	1	-0	0	0.243444	14.86713	0	0.994483	-3.900492	220	6	1.1	0.9;
	11317	1	-0	0	0	2.66	0	1.05478	8.215389	220	1	1.1	0.9;
	11318	1	70.78	43.8	0	2.12	0	0.880509	-14.642882	150	9	1.2	0.8;
	11319	1	140.6	20.1	0	13.81	0	1.040709	-0.335258	154	1	1.1	0.9;
	11320	1	17.02	5.5	0	1.02	0	0.979694	-6.226744	110	21	1.1	0.9;
	11321	1	28.63	5.6	0	1.28	0	1.019059	-14.535086	110	22	1.1	0.9;
	11322	1	220.2	83.2	0	0.61	0	1.002359	-6.963059	220	12	1.1	0.9;
	11323	1	39.64	4.4	0	0	0	1.000147	1.048084	110	24	1.1	0.9;
	11324	1	65.98	64.1	0	29.14	0	1.044735	35.829517	380	3	1.1	0.9;
	11325	2	0	0	0	0	0	1.041409	32.91812	20	3	1.1	0.9;
	11326	1	46.05	22	0	2.95	0	1.003473	-3.77501	220	6	1.1	0.9;
	11327	2	0	0	0	0	0	1.024267	0.497112	20	18	1.1	0.9;
	11328	2	0	0	0	0	0	1.050901	18.201002	20	7	1.1	0.9;
	11329	2	0	0	0	0	0	0.988112	23.208642	6.3	1	1.1	0.9;
	11330	2	0	0	0	0	0	1.049221	26.031718	20	5	1.1	0.9;
	11331	1	20.82	6.7	0	2.25	0	1.005689	36.422342	110	8	1.1	0.9;
	11332	2	0	0	0	0	0	1.077226	12.556118	20	2	1.1	0.9;
	11333	2	0	0	0	0	0	1.077226	12.556118	20	2	1.1	0.9;
	11334	1	0	0	0	122.13	0	1.031939	26.447767	380	5	1.1	0.9;
	11335	1	137	33.2	0	3.82	0	1.01186	2.974183	220	7	1.1	0.9;
	11336	2	0	0	0	0	0	1.023695	-0.879651	20	18	1.1	0.9;
	11337	1	7	1	0	1.51	0	1.02527	-3.478109	154	1	1.1	0.9;
	11338	1	65.57	44.6	0	5.95	0	0.929893	-10.863302	150	9	1.1	0.9;
	11339	1	-0	-0	0	0.36	0	1.022049	-0.726718	150	9	1.1	0.9;
	11340	1	0	-0	0	55	0	1.019263	25.887486	150	8	1.1	0.9;
	11341	1	35.74	11.1	0.038318	0.271682	0	1.052147	0.192688	110	19	1.1	0.9;
	11342	1	31.84	9.2	0	5.23	0	1.02579	7.134146	220	5	1.1	0.9;
	11343	1	-0	0	0	0.97	0	1.0051	-10.459769	150	9	1.1	0.9;
	11344	2	0	0	0	0	0	1.052164	7.333527	20	5	1.1	0.9;
	11345	2	0	0	0	0	0	1.035102	5.926523	20	6	1.1	0.9;
	11346	1	-0	0	0	0	0	1.013977	-15.655534	20	4	1.1	0.9;
	11347	2	0	0	0	0	0	1.021627	-19.648509	20	6	1.1	0.9;
	11348	2	0	0	0	0	0	1.034643	0.304854	20	6	1.1	0.9;
	11349	1	0	-0	0.162749	0.137938	0	0.997344	13.393742	220	6	1.1	0.9;
	11350	1	-0	0	0	0	0	1.013977	-15.655534	20	4	1.1	0.9;
	11351	2	0	0	0	0	0	1.025019	-12.264635	20	6	1.1	0.9;
	11352	2	0	0	0	0	0	1.025019	-12.264635	20	6	1.1	0.9;
	11353	2	0	0	0	0	0	1.034643	0.304854	20	6	1.1	0.9;
	11354	1	24.23	4.7	0	0.75	0	0.984275	-6.031468	110	18	1.1	0.9;
	11355	2	0	0	0	0	0	1.012747	-6.745565	20	4	1.1	0.9;
	11356	1	58.5	8.3	0	4.93	0	0.991129	19.563971	154	1	1.1	0.9;
	11357	1	0	-0	0	5.53	0	1.005377	4.629549	110	18	1.1	0.9;
	11358	1	122.7	-29.8	0	7.11	0	1.017916	24.245969	220	3	1.1	0.9;
	11359	1	10.9	1.6	0	1.5	0	1.004714	20.616677	154	1	1.1	0.9;
	11360	2	0	0	0	0	0	1.086148	2.994871	20	18	1.1	0.9;
	11361	2	0	0	0	0	0	1.085902	3.862201	20	18	1.1	0.9;
	11362	2	0	0	0	0	0	1.085902	3.862201	20	18	1.1	0.9;
	11363	2	0	0	0	0	0	1.02099	1.431006	20	18	1.1	0.9;
	11364	2	0	0	0	0	0	1.06019	67.358874	20	3	1.1	0.9;
	11365	1	0	0	0	0	0	1.028915	25.353044	150	8	1.1	0.9;
	11366	1	15.9	116.3	0	5.1	0	1.051422	54.686199	220	3	1.1	0.9;
	11367	2	0	0	0	0	0	1.06019	67.358874	20	3	1.1	0.9;
	11368	2	0	0	0	0	0	1.06019	67.358874	20	3	1.1	0.9;
	11369	2	0	0	0	0	0	1.06019	67.358874	20	3	1.1	0.9;
	11370	2	0	0	0	0	0	1.06019	67.358874	20	3	1.1	0.9;
	11371	1	524.5	126	0	7.45	0	1.062178	11.667815	220	5	1.1	0.9;
	11372	1	0	0	0	0.13	0	0.994918	-20.94236	220	6	1.1	0.9;
	11373	1	76.6	10.9	0	2.41	0	1.012764	18.020749	154	1	1.1	0.9;
	11374	1	2.2	0.5	0	0.21	0	0.97735	-26.907108	150	10	1.1	0.9;
	11375	1	8.03	5.5	0	0.12	0	1.034951	-6.540792	150	9	1.1	0.9;
	11376	1	17.52	5.7	0	0.28	0	1.052654	6.569125	110	18	1.2	0.8;
	11377	2	0	0	0	0	0	1.005433	2.576087	20	10	1.1	0.9;
	11378	1	-0	0	0.212593	6.327407	0	1.059916	14.412289	380	16	1.1	0.9;
	11379	1	-0	-0	0.818168	126.193818	0	1.181	3.660071	380	9	1.2	0.8;
	11380	1	10.61	1	0	1.7	0	0.980128	0.905508	110	18	1.1	0.9;
	11381	1	0	-0	0.059523	1.960477	0	1.006263	-13.98211	220	10	1.1	0.9;
	11382	1	0	-0	0.30117	48.396871	0	1.00314	34.191295	380	2	1.1	0.9;
	11383	1	0	0	0	0.15	0	1.052148	-6.603627	150	9	1.1	0.9;
	11384	1	0	-0	0	25.67	0	1.053186	6.507084	380	1	1.1	0.9;
	11385	1	11.01	3.6	0	0.5	0	0.96718	-6.283516	110	21	1.1	0.9;
	11386	1	78.89	25.6	0	1.6	0	1.025113	18.480146	220	5	1.1	0.9;
	11387	1	0	0	0	1.98	0	1.033489	17.191314	220	3	1.1	0.9;
	11388	1	11.91	3.9	0	2	0	1.001041	-8.121981	110	12	1.1	0.9;
	11389	1	3.4	0.5	0	3.28	0	1.011249	20.694257	154	1	1.1	0.9;
	11390	1	47.15	15.5	0	9.69	0	1.03413	-16.318991	220	5	1.1	0.9;
	11391	1	0	0	0	13.4	0	1.073637	71.243837	380	3	1.1	0.9;
	11392	1	0	-0.6	0	0.5	0	1.020228	59.501729	110	8	1.1	0.9;
	11393	2	0	0	0	0	0	1.016106	0.556804	11	1	1.1	0.9;
	11394	1	48.15	14.8	0	2.16	0	1.015825	42.943198	110	8	1.1	0.9;
	11395	1	223.6	51.8	0	18.45	0	1.047997	3.553626	220	5	1.1	0.9;
	11396	1	12.6	1.8	0	2.03	0	1.023472	8.50061	154	1	1.1	0.9;
	11397	1	0	0	0	0	0	1.042313	-6.262534	150	9	1.1	0.9;
	11398	1	27.23	10.1	0	0	0	1.015813	3.035457	220	3	1.1	0.9;
	11399	1	170.89	50.38	0	6.44	0	1.020219	2.145142	220	5	1.1	0.9;
	11400	1	-0	-0	1.051502	8.672363	0	1.082023	35.200613	380	5	1.2	0.8;
	11401	2	0	0	0	0	0	1.050328	7.929133	20	5	1.1	0.9;
	11402	1	153	21.5	0	28.85	0	1.048807	12.651704	380	5	1.1	0.9;
	11403	2	0	0	0	0	0	1.039038	39.160421	20	2	1.1	0.9;
	11404	1	7.61	2.5	0	1.67	0	0.959868	-6.582302	110	18	1.1	0.9;
	11405	1	124.5	101.4	0	0.53	0	1.015046	16.817691	220	5	1.1	0.9;
	11406	2	0	0	0	0	0	1.050841	90.684517	20	8	1.1	0.9;
	11407	1	-0	-0	0	13.33	0	1.006256	26.531463	400	1	1.1	0.9;
	11408	2	0	0	0	0	0	1.013014	63.774723	20	8	1.1	0.9;
	11409	2	0	0	0	0	0	1.013014	63.774723	20	8	1.1	0.9;
	11410	2	0	0	0	0	0	1.030034	63.293305	20	8	1.1	0.9;
	11411	2	0	0	0	0	0	1.035618	-0.462639	20	6	1.1	0.9;
	11412	2	0	0	0	0	0	1.035618	-0.462639	20	6	1.1	0.9;
	11413	1	0	-0	0.11	2.57	0	0.998061	-2.568006	220	21	1.1	0.9;
	11414	2	0	0	0	0	0	1.035618	-0.462639	20	6	1.1	0.9;
	11415	1	26.13	8.5	0	0.37	0	0.972366	-6.443004	110	21	1.1	0.9;
	11416	2	0	0	0	0	0	1.035618	-0.462639	20	6	1.1	0.9;
	11417	1	0	0	0.115837	3.304163	0	1.042395	5.320062	220	7	1.1	0.9;
	11418	2	0	0	0	0	0	1.051278	31.614896	20	5	1.1	0.9;
	11419	2	0	0	0	0	0	1.051278	31.614941	20	5	1.1	0.9;
	11420	1	30.03	9.6	0	4.38	0	1.002435	-22.866408	220	6	1.1	0.9;
	11421	1	-0	-0	0	4.9	0	1.009157	2.687917	380	6	1.1	0.9;
	11422	1	0	-0	0	55.79	0	0.976924	30.187846	400	1	1.1	0.9;
	11423	1	0	-0	0.1	-0.29	0	1.013272	-1.439863	220	19	1.1	0.9;
	11424	1	-0	0	0	1.33	0	1.021625	18.108773	220	5	1.1	0.9;
	11425	1	0	0	0	0.46	0	1.002261	-10.056932	150	10	1.1	0.9;
	11426	1	7.81	2.8	0	0.4	0	0.961189	-4.928205	110	19	1.1	0.9;
	11427	1	97.5	13.9	0	2.42	0	0.997223	15.348082	154	1	1.1	0.9;
	11428	1	0	-0	0	3.46	0	0.991226	-23.290634	220	6	1.1	0.9;
	11429	1	12.91	4.7	0	1.09	0	1.052317	-2.667381	110	19	1.1	0.9;
	11430	1	139.9	-26.2	0	5.42	0	0.97744	5.937442	220	6	1.1	0.9;
	11431	1	-0	0	0.039069	0.260931	0	1.012224	7.433648	110	18	1.1	0.9;
	11432	1	0	0	0	1.53	0	1.010935	29.835678	154	1	1.1	0.9;
	11433	1	9.61	3.9	0.058856	0.431144	0	0.996251	-7.108586	110	18	1.1	0.9;
	11434	1	-0	0	0	0.14	0	1.018338	-8.533019	150	9	1.1	0.9;
	11435	1	23.93	5.8	0	0.64	0	1.028279	-3.670442	110	21	1.1	0.9;
	11436	1	0	0	0.349217	20.455616	0	1.030579	39.766653	380	3	1.1	0.9;
	11437	1	15.22	3.9	0	0.89	0	0.981251	-5.256502	110	18	1.1	0.9;
	11438	1	-0	0	0	30.46	0	1.014925	-26.692737	220	4	1.1	0.9;
	11439	1	-0	0	0	0.22	0	0.974858	5.221454	110	18	1.1	0.9;
	11440	2	0	0	0	0	0	0.984601	26.113188	6.3	1	1.1	0.9;
	11441	1	0	-0	0	56.53	0	1.110503	1.036394	380	9	1.2	0.8;
	11442	1	-0	-0	1.139672	11.312126	0	1.054224	30.71588	380	5	1.1	0.9;
	11443	1	169.5	27.7	0	3.04	0	0.988541	17.0878	220	5	1.1	0.9;
	11444	1	-9.31	3	0	0.02	0	0.977862	-8.626469	150	9	1.1	0.9;
	11445	1	-0	0	0	43	0	1.034682	-22.435771	380	10	1.1	0.9;
	11446	1	115.1	16.4	0	7.67	0	1.038413	11.908771	220	5	1.1	0.9;
	11447	1	-0	0	0	3.21	0	1.031657	79.191379	220	3	1.1	0.9;
	11448	1	209.1	1.7	0	18.49	0	1.056599	16.07956	220	3	1.1	0.9;
	11449	1	22.83	3.8	0	0.26	0	1.001069	5.906574	110	18	1.1	0.9;
	11450	1	1.8	0.5	0	0.79	0	0.956504	0.050308	110	18	1.1	0.9;
	11451	1	34.9	5	0	2.53	0	1.027023	17.407089	154	1	1.1	0.9;
	11452	1	-0	0	0	10.33	0	1.004233	-6.701569	220	6	1.1	0.9;
	11453	1	149.1	-18.3	0	10.76	0	1.005946	5.461847	220	3	1.1	0.9;
	11454	2	0	0	0	0	0	1.013471	23.648482	20	5	1.1	0.9;
	11455	2	0	0	0	0	0	1.013471	23.648482	20	5	1.1	0.9;
	11456	2	0	0	0	0	0	1.001613	16.787108	20	5	1.1	0.9;
	11457	1	-0	-0	0	0	0	1.000465	12.056714	20	5	1.1	0.9;
	11458	1	-0	-0	0	0	0	1.000465	12.056714	20	5	1.1	0.9;
	11459	1	-0	-0	0	0	0	1.000465	12.056714	20	5	1.1	0.9;
	11460	1	584.3	166.3	0	48.159681	0	1.012362	-10.049221	380	6	1.1	0.9;
	11461	1	0	0	0	0	0	1.063704	66.981987	20	8	1.1	0.9;
	11462	1	0	0	0	4.14	0	1.018804	38.015302	380	8	1.1	0.9;
	11463	1	0	0	0	0	0	1.063704	66.981987	20	8	1.1	0.9;
	11464	1	0	0	0	0	0	1.063704	66.981987	20	8	1.1	0.9;
	11465	1	77.7	11.1	0	4.49	0	1.007297	16.512842	154	1	1.1	0.9;
	11466	1	0	-0	0.039875	12.830062	0	1.009763	-12.59241	220	4	1.1	0.9;
	11467	2	0	0	0	0	0	1.040642	8.532255	20	14	1.1	0.9;
	11468	1	12.11	-4	0.23333	12.135003	0	1.037667	-14.462311	220	4	1.1	0.9;
	11469	1	127.2	-10.7	0	7.8	0	1.026209	-27.036398	220	4	1.1	0.9;
	11470	1	179	22.9	0	25.28	0	1.011959	-23.601992	380	4	1.1	0.9;
	11471	2	0	0	0	0	0	1.026413	2.766853	20	2	1.1	0.9;
	11472	2	0	0	0	0	0	1.035329	26.785392	20	6	1.1	0.9;
	11473	2	0	0	0	0	0	1.017601	-9.8258	20	10	1.1	0.9;
	11474	2	0	0	0	0	0	1.036049	1.565792	20	2	1.1	0.9;
	11475	2	0	0	0	0	0	1.035259	26.774039	20	6	1.1	0.9;
	11476	1	0	0	0	0	0	0.834005	-17.59375	150	9	1.2	0.8;
	11477	2	0	0	0	0	0	1.035329	26.785392	20	6	1.1	0.9;
	11478	2	0	0	0	0	0	1.035259	26.774039	20	6	1.1	0.9;
	11479	2	0	0	0	0	0	1.035499	26.862046	20	6	1.1	0.9;
	11480	2	0	0	0	0	0	1.035431	26.859737	20	6	1.1	0.9;
	11481	2	0	0	0	0	0	1.019623	1.497251	20	2	1.1	0.9;
	11482	2	0	0	0	0	0	1.022039	16.134374	20	2	1.1	0.9;
	11483	2	0	0	0	0	0	1.063714	6.510197	20	9	1.1	0.9;
	11484	2	0	0	0	0	0	1.037469	11.940478	20	5	1.1	0.9;
	11485	1	0	-0	0	0.52	0	0.978613	-8.136031	150	9	1.1	0.9;
	11486	1	-0	0	0	1.77	0	0.993188	-7.06422	150	9	1.1	0.9;
	11487	2	0	0	0	0	0	1.042317	14.274354	20	14	1.1	0.9;
	11488	1	37.04	15.2	0.174005	27.214996	0	1.043769	-17.010151	380	10	1.1	0.9;
	11489	1	32.44	20.1	0	2.37	0	1.051219	0.292959	150	9	1.1	0.9;
	11490	1	137.2	41.9	0.101088	13.207291	0	1.036739	-15.108104	380	10	1.1	0.9;
	11491	1	0	0	0.090774	13.090445	0	1.050366	-6.017105	220	18	1.2	0.8;
	11492	2	0	0	0	0	0	0.970854	29.381452	15.75	1	1.1	0.9;
	11493	1	-3.2	-7.9	0	1.95	0	1.051816	27.634038	220	2	1.1	0.9;
	11494	1	-0	-0	0.135493	30.26062	0	1.069655	-1.580413	380	21	1.1	0.9;
	11495	1	52.76	17.2	0	38.38	0	0.994132	-4.107772	220	6	1.1	0.9;
	11496	2	0	0	0	0	0	1.062896	-0.392816	20	5	1.1	0.9;
	11497	2	0	0	0	0	0	1.029009	32.575777	20	5	1.1	0.9;
	11498	2	0	0	0	0	0	1.029009	32.575777	20	5	1.1	0.9;
	11499	2	0	0	0	0	0	1.029009	32.575777	20	5	1.1	0.9;
	11500	2	0	0	0	0	0	1.029009	32.575777	20	5	1.1	0.9;
	11501	1	193.3	42	0.119745	3.480255	0	1.025734	-19.195718	220	4	1.1	0.9;
	11502	2	0	0	0	0	0	1.017349	20.721078	11	1	1.1	0.9;
	11503	1	134.7	17.5	0	4.53	0	1.001734	-25.89184	220	4	1.1	0.9;
	11504	2	0	0	0	0	0	1.022885	22.331648	20	10	1.1	0.9;
	11505	1	110.3	10.7	0.117264	16.622052	0	1.024319	-2.010067	220	6	1.1	0.9;
	11506	2	0	0	0	0	0	1.062896	-0.392816	20	5	1.1	0.9;
	11507	2	0	0	0	0	0	1.062896	-0.392816	20	5	1.1	0.9;
	11508	1	0	-0	0.009953	1.310047	0	0.988502	2.309053	110	18	1.1	0.9;
	11509	2	0	0	0	0	0	1.013195	9.237936	11	1	1.1	0.9;
	11510	2	0	0	0	0	0	1.03324	-5.012105	20	4	1.1	0.9;
	11511	1	0	0	0	1.01	0	1.019793	-7.561936	150	9	1.1	0.9;
	11512	1	10.11	4.3	0	0.63	0	0.985835	-4.027629	110	18	1.1	0.9;
	11513	1	0	0	0.215164	46.233143	0	1.080848	2.641129	380	18	1.2	0.8;
	11514	1	-9.78	-3.66	0	0.72	0	0.973933	-8.952943	110	18	1.1	0.9;
	11515	1	0	-0	0	2.06	0	1.056507	27.96286	220	5	1.1	0.9;
	11516	1	8.71	2.5	0	0.66	0	0.969418	-8.813263	110	18	1.1	0.9;
	11517	1	-0	0	0.101122	0.748878	0	0.984712	15.209887	380	5	1.1	0.9;
	11518	1	0	-0	0.302291	2.117657	0	1.060327	8.879279	220	3	1.1	0.9;
	11519	1	-0	0	0	27.39	0	1.036174	36.321501	150	8	1.1	0.9;
	11520	2	0	0	0	0	0	1.023361	29.828652	20	5	1.1	0.9;
	11521	2	0	0	0	0	0	1.024274	-9.95923	20	4	1.1	0.9;
	11522	1	185.8	31.8	0	3.73	0	1.020685	7.63502	220	5	1.1	0.9;
	11523	2	0	0	0	0	0	1.031502	0.465782	20	6	1.1	0.9;
	11524	2	0	0	0	0	0	1.031502	0.465782	20	6	1.1	0.9;
	11525	2	0	0	0	0	0	1.031502	0.465782	20	6	1.1	0.9;
	11526	2	0	0	0	0	0	1.031502	0.465782	20	6	1.1	0.9;
	11527	2	0	0	0	0	0	1.03122	1.037828	20	6	1.1	0.9;
	11528	2	0	0	0	0	0	1.03122	1.037828	20	6	1.1	0.9;
	11529	2	0	0	0	0	0	1.007216	-1.420425	20	2	1.1	0.9;
	11530	2	0	0	0	0	0	1.03122	1.037828	20	6	1.1	0.9;
	11531	1	0	0	0	0	0	1.03616	-10.170176	150	9	1.1	0.9;
	11532	1	0	-0	0	1.55	0	1.017123	21.904571	220	5	1.1	0.9;
	11533	1	0	0	0	0.86	0	1.000846	-7.407948	110	21	1.1	0.9;
	11534	1	0	-0	0	15.21	0	1.029093	49.644716	380	3	1.1	0.9;
	11535	1	12.6	1.8	0	4.19	0	0.970932	21.818127	154	1	1.1	0.9;
	11536	2	0	0	0	0	0	1.023361	29.828652	20	5	1.1	0.9;
	11537	1	-0	0	0	3.98	0	1.052535	7.6299	220	1	1.1	0.9;
	11538	1	-3.53	-4.62	0	0	0	0.990321	23.849769	150	8	1.1	0.9;
	11539	2	0	0	0	0	0	1.056762	11.950776	10.5	1	1.1	0.9;
	11540	1	40.65	16	0	0.36	0	0.987018	-4.48064	110	18	1.1	0.9;
	11541	1	316.5	100.8	0.239735	5.64021	0	1.006736	55.801921	220	13	1.1	0.9;
	11542	1	-0	0	0	13.08	0	0.996405	-7.172261	220	6	1.1	0.9;
	11543	1	-0	0	0	0	0	1.050293	43.935202	380	8	1.1	0.9;
	11544	1	-0	-0	0	2.36	0	1.034667	29.874033	220	5	1.1	0.9;
	11545	1	-0	0	0	15.04	0	1.034377	-13.734738	380	1	1.1	0.9;
	11546	1	0	0	0	5.59	0	0.995183	32.889399	220	3	1.1	0.9;
	11547	1	-0	-0	0.463131	37.913355	0	1.048459	-12.77196	380	10	1.1	0.9;
	11548	1	0	0	0	0	0	1.036105	-7.389707	150	9	1.1	0.9;
	11549	2	0	0	0	0	0	0.995145	31.935435	11	1	1.1	0.9;
	11550	1	258.3	166	0	16.31	0	0.988946	17.967208	220	6	1.1	0.9;
	11551	2	0	0	0	0	0	1.054159	6.298568	20	20	1.1	0.9;
	11552	1	-0	0	0	0.33	0	0.964189	-9.676227	150	9	1.1	0.9;
	11553	1	11.41	0.3	0	0.2	0	0.988905	-7.944058	110	18	1.1	0.9;
	11554	1	39.14	13.2	0	2.92	0	0.970038	-5.739908	110	18	1.1	0.9;
	11555	1	229.3	32.7	0	10.27	0	1.011924	12.571042	154	1	1.1	0.9;
	11556	1	88.5	10.8	0	9.54	0	1.025664	3.941893	220	20	1.1	0.9;
	11557	1	2.8	0.5	0	1.26	0	0.994823	-7.7742	110	18	1.1	0.9;
	11558	2	0	0	0	0	0	1.019231	22.327173	20	24	1.1	0.9;
	11559	1	-0	-0	0	0.26	0	1.022691	-10.833665	150	9	1.1	0.9;
	11560	1	-0	0	0	0	0	1.021419	-22.29244	20	4	1.1	0.9;
	11561	1	0	0	0	77.597016	0	1.05421	6.879607	380	24	1.1	0.9;
	11562	1	0	0	0	0	0	0.978211	-7.893051	150	9	1.1	0.9;
	11563	1	17.02	2	0	3.65	0	1.029712	9.125452	220	7	1.1	0.9;
	11564	2	0	0	0	0	0	1.033947	-9.228621	20	6	1.1	0.9;
	11565	1	254	74.3	0	54.48	0	1.047277	7.939752	380	16	1.1	0.9;
	11566	1	-0	0	0	6.51	0	0.985612	-18.102016	220	4	1.1	0.9;
	11567	1	-0	0	0	8.61	0	1.019305	44.637887	220	8	1.1	0.9;
	11568	2	0	0	0	0	0	1.032238	-4.146012	20	6	1.1	0.9;
	11569	2	0	0	0	0	0	1.032238	-4.146012	20	6	1.1	0.9;
	11570	2	0	0	0	0	0	1.026319	-4.093349	20	6	1.1	0.9;
	11571	1	0.3	-97.2	0	0.07	0	1.037776	30.45979	220	3	1.1	0.9;
	11572	1	73.58	48.6	0	2.87	0	0.952007	-28.718282	150	10	1.1	0.9;
	11573	1	0	0	0.295058	0.244119	0	1.024079	44.848437	380	3	1.1	0.9;
	11574	1	-0	-0	0	7.99	0	1.052808	14.230733	220	5	1.1	0.9;
	11575	1	0	0	0	0	0	1.029478	-2.341076	150	9	1.1	0.9;
	11576	1	47.55	17.8	0	0.32	0	1.027086	0.614256	220	5	1.1	0.9;
	11577	2	0	0	0	0	0	0.990908	15.83898	20	18	1.1	0.9;
	11578	2	0	0	0	0	0	1.016999	60.332496	20	10	1.1	0.9;
	11579	2	0	0	0	0	0	1.050987	40.998977	20	10	1.1	0.9;
	11580	2	0	0	0	0	0	1.023809	10.838684	20	6	1.1	0.9;
	11581	2	0	0	0	0	0	0.995022	23.934845	11	1	1.1	0.9;
	11582	1	142.4	20	0	2.42	0	1.033386	9.187824	220	5	1.1	0.9;
	11583	1	162.5	42.7	0	12.79	0	1.001509	9.954559	220	5	1.1	0.9;
	11584	2	0	0	0	0	0	1.00419	31.464758	11	1	1.1	0.9;
	11585	1	41.88	3.02	0	5.65	0	1.004923	90.798314	150	1	1.1	0.9;
	11586	2	0	0	0	0	0	1.052394	33.692026	20	10	1.1	0.9;
	11587	1	0	-0	0	4.52	0	1.030635	2.430468	154	1	1.1	0.9;
	11588	2	0	0	0	0	0	1.029344	21.964639	20	10	1.1	0.9;
	11589	1	0	0	0.090774	2.080444	0	1.05536	4.225642	220	18	1.1	0.9;
	11590	1	125.3	7.7	0.019458	14.570542	0	1.022638	24.068742	220	13	1.1	0.9;
	11591	1	427.9	145.8	0	40.72	0	1.021461	-8.576014	380	6	1.1	0.9;
	11592	1	36.94	11	0	0.52	0	1.023532	17.869884	220	5	1.1	0.9;
	11593	1	0	0	0	8.77	0	1.013843	-20.803128	220	6	1.1	0.9;
	11594	1	9.7	1.4	0	11.34	0	1.01472	35.673432	154	1	1.1	0.9;
	11595	1	338.7	52.5	0	11.36	0	1.039128	10.819615	220	5	1.1	0.9;
	11596	1	-0	0	0.15322	5.97678	0	1.039377	14.867516	220	7	1.1	0.9;
	11597	1	3.2	0.6	0	0.82	0	0.988782	-5.594179	110	18	1.1	0.9;
	11598	1	-0	-0	0	25.73	0	1.061414	49.736419	380	1	1.1	0.9;
	11599	2	0	0	0	0	0	1.045359	9.917398	20	23	1.1	0.9;
	11600	2	0	0	0	0	0	1.031054	10.435949	20	5	1.1	0.9;
	11601	1	52.46	2.9	0	0.35	0	1.02841	11.034704	220	5	1.1	0.9;
	11602	2	0	0	0	0	0	1.045359	9.917398	20	23	1.1	0.9;
	11603	2	0	0	0	0	0	1.022352	20.659989	20	5	1.1	0.9;
	11604	2	0	0	0	0	0	1.022352	20.659989	20	5	1.1	0.9;
	11605	2	0	0	0	0	0	1.031054	10.435949	20	5	1.1	0.9;
	11606	2	0	0	0	0	0	1.031054	10.435949	20	5	1.1	0.9;
	11607	1	0	-0	0	0.15	0	1.02032	9.008458	110	18	1.1	0.9;
	11608	1	0	0	0	7.78	0	1.031085	-14.460229	220	5	1.1	0.9;
	11609	1	9.21	3.4	0	2.63	0	1.015791	22.756082	220	5	1.1	0.9;
	11610	1	30.35	30.5	0	0.76	0	1.017753	-8.026498	150	9	1.1	0.9;
	11611	2	0	0	0	0	0	1.043059	9.285297	20	23	1.1	0.9;
	11612	1	7	1	0	4.05	0	0.98336	22.995978	154	1	1.1	0.9;
	11613	1	0	0	0	1.55	0	1.052439	-6.564757	150	9	1.1	0.9;
	11614	1	77.2	11	0	9.22	0	0.975047	41.518085	154	1	1.1	0.9;
	11615	1	0	-0	0	0.33	0	1.018322	-0.328638	110	18	1.1	0.9;
	11616	1	0	-0	0	7.73	0	1.004155	27.356143	380	3	1.1	0.9;
	11617	2	0	0	0	0	0	1.009173	19.132388	20	5	1.1	0.9;
	11618	1	-0	8.3	0	10.19	0	1.102278	-15.190395	150	9	1.2	0.8;
	11619	2	0	0	0	0	0	1.041353	7.871779	20	22	1.1	0.9;
	11620	1	0	0	0	0	0	0.940462	-11.519394	150	9	1.1	0.9;
	11621	1	-0	0	0	0	0	0.992991	15.869625	20	5	1.1	0.9;
	11622	1	-0	0	0	0	0	0.992991	15.869625	20	5	1.1	0.9;
	11623	2	0	0	0	0	0	1.009173	19.132388	20	5	1.1	0.9;
	11624	2	0	0	0	0	0	1.009173	19.132388	20	5	1.1	0.9;
	11625	2	0	0	0	0	0	1.014865	17.912579	20	5	1.1	0.9;
	11626	2	0	0	0	0	0	1.014865	17.912579	20	5	1.1	0.9;
	11627	1	-0.18	-1.31	0	0	0	1.017817	42.272152	150	8	1.1	0.9;
	11628	1	35.14	7.6	0	6.49	0	1.020622	-11.444475	110	22	1.1	0.9;
	11629	1	0	-0	0	2.74	0	1.04378	11.735527	380	1	1.1	0.9;
	11630	2	0	0	0	0	0	0.991989	39.317054	13.8	1	1.1	0.9;
	11631	2	0	0	0	0	0	1.050944	40.491716	20	3	1.1	0.9;
	11632	2	0	0	0	0	0	1.0508	40.476149	20	3	1.1	0.9;
	11633	2	0	0	0	0	0	1.04927	45.669864	20	3	1.1	0.9;
	11634	2	0	0	0	0	0	1.050944	40.491716	20	3	1.1	0.9;
	11635	2	0	0	0	0	0	1.0508	40.476149	20	3	1.1	0.9;
	11636	1	160.3	89.7	0	47.28	0	1.00649	19.103021	380	3	1.1	0.9;
	11637	1	38.15	23.6	0	0	0	1.053823	0.430007	150	9	1.1	0.9;
	11638	2	0	0	0	0	0	1.012679	-1.92622	20	18	1.1	0.9;
	11639	1	-0	0	0.12	20.1	0	1.01915	23.7232	380	6	1.1	0.9;
	11640	1	-0	0	0	0.97	0	1.031648	8.37821	220	5	1.1	0.9;
	11641	2	0	0	0	0	0	1.00746	39.171104	11	1	1.1	0.9;
	11642	1	88.4	25.2	0	12.9	0	1.029112	7.476738	220	5	1.1	0.9;
	11643	2	0	0	0	0	0	1.035771	2.999441	20	14	1.1	0.9;
	11644	1	0.3	0.2	0	4.7	0	1.03531	20.027062	380	3	1.1	0.9;
	11645	1	36.84	14.1	0	33.55	0	1.045347	-15.357554	380	4	1.1	0.9;
	11646	1	60.07	20	0	0.17	0	0.955328	-7.482884	110	21	1.1	0.9;
	11647	1	-0	-0	0	65.24	0	0.969264	32.129316	400	1	1.1	0.9;
	11648	2	0	0	0	0	0	1.029166	13.522836	20	14	1.1	0.9;
	11649	2	0	0	0	0	0	1.023418	29.920304	20	5	1.1	0.9;
	11650	1	0	0	0.232942	35.845683	0	1.030951	60.569581	380	3	1.1	0.9;
	11651	1	17.52	9.1	0	2.26	0	0.987257	-3.558123	110	18	1.1	0.9;
	11652	2	0	0	0	0	0	1.015374	8.632692	20	18	1.1	0.9;
	11653	1	-0.67	-2.78	0	11.64	0	1.051675	-1.334511	220	23	1.1	0.9;
	11654	2	0	0	0	0	0	1.027264	14.46311	20	14	1.1	0.9;
	11655	1	285.2	65.3	0	11.28	0	0.977412	-27.214747	220	10	1.1	0.9;
	11656	1	-0	0	0	20.11	0	1.041642	19.877159	380	1	1.1	0.9;
	11657	1	1.5	87.9	0.529771	7.297956	0	1.020195	-15.690329	380	4	1.1	0.9;
	11658	1	-14.91	-5	0	0.57	0	0.992732	2.429388	110	18	1.1	0.9;
	11659	1	0	0	0.733021	192.153934	0	1.068375	74.040783	380	3	1.1	0.9;
	11660	1	0	-0	0	42.81	0	1.032044	38.537454	380	8	1.1	0.9;
	11661	1	115.8	29.3	0	4.31	0	1.007654	15.013591	220	7	1.1	0.9;
	11662	1	56.06	17	0	2.4	0	1.035273	21.970264	220	5	1.1	0.9;
	11663	1	106.8	22.2	0	0.97	0	1.019769	-3.333231	110	21	1.1	0.9;
	11664	1	0	0	0	0	0	1.004491	-12.436802	150	9	1.1	0.9;
	11665	2	0	0	0	0	0	0.947933	57.203162	11	1	1.1	0.9;
	11666	2	0	0	0	0	0	1.032102	-0.219036	20	6	1.1	0.9;
	11667	2	0	0	0	0	0	1.032102	-0.219036	20	6	1.1	0.9;
	11668	2	0	0	0	0	0	1.032102	-0.219036	20	6	1.1	0.9;
	11669	2	0	0	0	0	0	1.032102	-0.219036	20	6	1.1	0.9;
	11670	2	0	0	0	0	0	1.032102	-0.219036	20	6	1.1	0.9;
	11671	1	0	0	0.361295	2.568993	0	1.021237	23.717974	380	6	1.1	0.9;
	11672	2	0	0	0	0	0	1.018847	32.305408	11	1	1.1	0.9;
	11673	1	-35.74	38.6	0	4.65	0	1.02462	-12.151463	220	4	1.1	0.9;
	11674	1	100.9	26.7	0	118.53	0	1.021472	42.962213	380	3	1.1	0.9;
	11675	1	38.44	2	0	0.56	0	1.033912	12.93783	220	5	1.1	0.9;
	11676	1	-0	-0	0	2.12	0	1.010939	25.120796	220	3	1.1	0.9;
	11677	2	0	0	0	0	0	1.058213	27.963025	20	2	1.1	0.9;
	11678	1	17.92	5.8	0	0.84	0	1.036103	23.985216	220	5	1.1	0.9;
	11679	1	0	0	0.181549	21.720889	0	1.059879	-2.236576	220	18	1.2	0.8;
	11680	2	0	0	0	0	0	1.026027	-7.002651	20	4	1.1	0.9;
	11681	1	0	0	0	0	0	1.012673	-2.175699	380	6	1.1	0.9;
	11682	1	-0	0	0.090774	1.980444	0	1.0437	6.318844	220	18	1.1	0.9;
	11683	1	0	0	0	3.86	0	1.014036	28.98618	154	1	1.1	0.9;
	11684	1	125.4	17.8	0	4.02	0	1.027671	29.425746	154	1	1.1	0.9;
	11685	1	156	60.7	0	1.52	0	0.989283	27.206785	150	8	1.1	0.9;
	11686	2	0	0	0	0	0	1.015505	-3.112218	20	22	1.1	0.9;
	11687	2	0	0	0	0	0	1.01946	-9.574196	20	22	1.1	0.9;
	11688	1	488.7	184.7	0.35476	78.854076	0	1.047817	24.207232	380	14	1.1	0.9;
	11689	2	0	0	0	0	0	1.005775	-8.537815	20	22	1.1	0.9;
	11690	2	0	0	0	0	0	1.005775	-8.537815	20	22	1.1	0.9;
	11691	2	0	0	0	0	0	1.005775	-8.537815	20	22	1.1	0.9;
	11692	2	0	0	0	0	0	1.005775	-8.537815	20	22	1.1	0.9;
	11693	2	0	0	0	0	0	1.005775	-8.537815	20	22	1.1	0.9;
	11694	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	11695	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	11696	1	111.6	8.9	0	16.19	0	1.079269	5.904616	220	5	1.1	0.9;
	11697	1	5.11	1.3	0	0.8	0	1.018627	-2.342489	110	18	1.1	0.9;
	11698	1	0	-0	0	8.63	0	1.02038	17.569195	220	3	1.1	0.9;
	11699	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	11700	1	0	0	0	0.22	0	1.021285	26.180239	150	8	1.1	0.9;
	11701	1	0	0	0	59.88	0	1.127132	-1.539016	380	9	1.2	0.8;
	11702	1	0	-0	0.34087	0.574879	0	1.045492	35.801591	380	2	1.1	0.9;
	11703	1	443.8	172.1	0	52.54	0	1.059179	30.508917	380	14	1.1	0.9;
	11704	2	0	0	0	0	0	1.047549	4.244102	20	18	1.1	0.9;
	11705	2	0	0	0	0	0	1.003732	-4.459321	20	4	1.1	0.9;
	11706	1	3.3	1.1	0	0.28	0	0.997366	-14.326802	110	22	1.1	0.9;
	11707	1	0	0	0	3.28	0	1.049541	0.642038	154	1	1.1	0.9;
	11708	1	0	0	0	0	0	1.015869	-10.791845	150	9	1.1	0.9;
	11709	1	-0	-0	0	0	0	1.014893	-9.156289	20	4	1.1	0.9;
	11710	1	-0	-0	0	0.24	0	1.001826	2.228946	154	1	1.1	0.9;
	11711	1	0	0	0	79.5	0	1.024531	-10.409141	380	1	1.1	0.9;
	11712	1	21.72	7.2	0	0.64	0	0.958243	-7.97843	110	21	1.1	0.9;
	11713	2	0	0	0	0	0	1.0182	22.6787	20	6	1.1	0.9;
	11714	2	0	0	0	0	0	1.0182	22.6787	20	6	1.1	0.9;
	11715	1	0	-0	0	48.74	0	0.967345	5.815652	380	1	1.1	0.9;
	11716	2	0	0	0	0	0	1.024861	-2.735628	20	10	1.1	0.9;
	11717	2	0	0	0	0	0	1.042492	18.795962	20	6	1.1	0.9;
	11718	1	69.48	19.7	0	22.1	0	1.02973	-21.007744	220	4	1.1	0.9;
	11719	1	0	0	0.2527	79.80215	0	1.039836	-14.881275	380	10	1.1	0.9;
	11720	1	30.03	10	0	1.11	0	0.991694	-4.403598	110	21	1.1	0.9;
	11721	1	-0	0	0	21.6	0	1.023018	33.158363	380	3	1.1	0.9;
	11722	1	237.9	68.5	0	5.12	0	1.01975	9.799079	220	5	1.1	0.9;
	11723	2	0	0	0	0	0	1.050494	10.971923	20	5	1.1	0.9;
	11724	1	-0.2	-0.96	0	0	0	1.014816	31.102888	150	8	1.1	0.9;
	11725	1	115.7	10.7	0	19.1	0	1.013341	-26.927209	220	4	1.1	0.9;
	11726	1	234	13.1	0	51.03	0	1.025187	10.345307	220	5	1.1	0.9;
	11727	1	0	0	0	0.59	0	0.96359	-0.680295	110	18	1.1	0.9;
	11728	1	0	-0	0	1.57	0	0.924746	-12.203993	150	9	1.1	0.9;
	11729	2	0	0	0	0	0	1.036966	15.051213	20	14	1.1	0.9;
	11730	1	33.14	20.5	0	0.54	0	1.019724	-7.10214	150	9	1.1	0.9;
	11731	1	110.2	15.7	0	1.11	0	0.997872	29.769503	154	1	1.1	0.9;
	11732	1	217.7	164.6	0	11.88	0	1.012944	32.566079	380	3	1.1	0.9;
	11733	1	-0	0	0.12	15.18	0	0.999881	-2.01284	220	21	1.1	0.9;
	11734	1	-1272.84	-649.5	0	0	0	1.041776	36.108702	27	3	1.1	0.9;
	11735	1	253.7	85	1.799271	133.197518	0	1.07361	4.31055	380	20	1.2	0.8;
	11736	2	0	0	0	0	0	1.046239	4.338704	20	23	1.1	0.9;
	11737	1	233.5	151	0	35.31	0	1.033684	-16.072132	380	10	1.1	0.9;
	11738	1	35.14	1.8	0	6.12	0	1.033098	-19.377215	220	4	1.1	0.9;
	11739	1	4.61	1.7	0	0.47	0	1.024838	0.506407	110	19	1.1	0.9;
	11740	1	120.1	17.1	0	0.08	0	1.016088	20.739625	154	1	1.1	0.9;
	11741	1	-0	0	0	0.98	0	1.036683	47.83205	110	8	1.1	0.9;
	11742	1	-0	-0	0	107.2	0	0.967445	33.308335	400	1	1.1	0.9;
	11743	1	62.7	9	0	5.83	0	1.009308	25.594526	154	1	1.1	0.9;
	11744	1	0	-8.1	0	0	0	0.997571	-6.64121	150	9	1.1	0.9;
	11745	2	0	0	0	0	0	1.039251	-4.888975	20	10	1.1	0.9;
	11746	1	-0	0	0	0.27	0	1.056194	25.665777	220	5	1.1	0.9;
	11747	1	12.61	7.8	0	2.42	0	0.964803	-5.53487	150	9	1.1	0.9;
	11748	1	-0	0	0	0.27	0	1.014672	-3.125729	110	19	1.1	0.9;
	11749	2	0	0	0	0	0	1.017828	-6.986287	20	4	1.1	0.9;
	11750	1	400.4	23.6	0.577149	133.344695	0	1.040527	0.879566	380	11	1.1	0.9;
	11751	1	18.22	6.3	0	0.35	0	0.995316	-1.997087	110	18	1.1	0.9;
	11752	2	0	0	0	0	0	1.017287	26.369278	20	5	1.1	0.9;
	11753	1	11.91	3.3	0	6.84	0	1.014637	14.638791	220	5	1.1	0.9;
	11754	1	-0	-0	0	1.91	0	0.976898	39.547729	154	1	1.1	0.9;
	11755	2	0	0	0	0	0	1.019643	-4.885706	20	22	1.1	0.9;
	11756	2	0	0	0	0	0	1.019643	-4.885706	20	22	1.1	0.9;
	11757	2	0	0	0	0	0	1.019643	-4.885706	20	22	1.1	0.9;
	11758	2	0	0	0	0	0	1.037142	24.334502	20	5	1.1	0.9;
	11759	2	0	0	0	0	0	1.037142	24.334502	20	5	1.1	0.9;
	11760	2	0	0	0	0	0	1.037142	24.334502	20	5	1.1	0.9;
	11761	2	0	0	0	0	0	1.018631	-11.528817	20	4	1.1	0.9;
	11762	2	0	0	0	0	0	1.037142	24.334502	20	5	1.1	0.9;
	11763	2	0	0	0	0	0	1.037142	24.334502	20	5	1.1	0.9;
	11764	1	0	-0	0	9.68	0	1.030705	-20.918416	220	4	1.1	0.9;
	11765	1	-0	-0	0	0.35	0	1.027351	-0.814842	150	9	1.1	0.9;
	11766	1	0	0	0	0.52	0	0.999629	27.315451	154	1	1.1	0.9;
	11767	1	0	-0	0.458938	18.186674	0	1.060664	31.154385	380	13	1.1	0.9;
	11768	2	0	0	0	0	0	1.023993	4.449412	11	1	1.1	0.9;
	11769	1	-0	-0	0	0.36	0	1.008615	-8.918557	150	9	1.1	0.9;
	11770	1	0	-0	0	0.76	0	1.021249	-4.603249	110	22	1.1	0.9;
	11771	1	0	-0	0	0.37	0	0.981726	-7.828011	150	9	1.1	0.9;
	11772	1	135.7	31.6	0	26.65	0	0.999599	-7.186284	220	6	1.1	0.9;
	11773	2	0	0	0	0	0	1.008948	-4.613451	20	10	1.1	0.9;
	11774	1	-0	0	0	16.77	0	1.015224	15.706003	220	5	1.1	0.9;
	11775	1	14.92	4.9	0	0.65	0	1.008464	-9.273691	110	12	1.1	0.9;
	11776	1	81.99	32.2	0	7.3	0	1.017793	-25.568793	220	4	1.1	0.9;
	11777	1	13.9	1.9	0	6.58	0	1.004419	17.422	154	1	1.1	0.9;
	11778	1	0	0	0	0	0	0.965388	-11.747495	150	9	1.1	0.9;
	11779	1	3.2	1.5	0	0.95	0	0.990025	-5.482919	110	18	1.1	0.9;
	11780	1	-0	-0	0	0.35	0	0.918259	-12.437657	150	9	1.1	0.9;
	11781	2	0	0	0	0	0	1.03053	1.384954	11	1	1.1	0.9;
	11782	1	36.84	14.1	0	17.84	0	1.03795	-12.780799	380	4	1.1	0.9;
	11783	1	-0	-0	0	7.92	0	1.017814	42.27213	150	8	1.1	0.9;
	11784	1	40.04	15.1	0	1.05	0	0.953672	-7.51553	110	21	1.1	0.9;
	11785	1	149.63	59	0	4.99	0	0.992936	-4.042052	220	6	1.1	0.9;
	11786	2	0	0	0	0	0	1.058114	43.382414	20	10	1.1	0.9;
	11787	1	28.53	11.5	0	0.72	0	0.979531	-5.995895	110	21	1.1	0.9;
	11788	1	16.7	2.4	0	2.64	0	1.031335	18.168562	154	1	1.1	0.9;
	11789	2	0	0	0	0	0	0.996117	32.042491	11	1	1.1	0.9;
	11790	1	-0	-0	0	8.99	0	1.02416	36.959559	150	8	1.1	0.9;
	11791	1	9.41	2.3	0	1.18	0	0.980914	1.532206	110	18	1.1	0.9;
	11792	1	0	0	0	2.78	0	1.01769	-5.855082	220	1	1.1	0.9;
	11793	1	0	0	0	2.59	0	0.967603	48.289835	220	3	1.1	0.9;
	11794	1	1.8	0.7	0	0.27	0	0.984643	-1.523613	110	19	1.1	0.9;
	11795	1	14.22	4.7	0	5.58	0	0.990096	-3.188831	220	10	1.1	0.9;
	11796	2	0	0	0	0	0	1.015006	-6.446837	20	4	1.1	0.9;
	11797	2	0	0	0	0	0	1.027461	-17.241264	20	4	1.1	0.9;
	11798	1	-0	0	0	4.32	0	1.095454	-13.630803	150	9	1.2	0.8;
	11799	1	0	-0	0	0.11	0	0.997412	-7.185192	110	18	1.1	0.9;
	11800	1	11.41	4.4	0	0.12	0	0.986973	-7.976535	110	18	1.1	0.9;
	11801	1	76.88	16.4	0	6.61	0	1.040262	2.637511	220	5	1.1	0.9;
	11802	1	-0	-0	0	0	0	1.020557	9.624602	20	14	1.1	0.9;
	11803	2	0	0	0	0	0	1.060489	63.22043	20	3	1.1	0.9;
	11804	1	-0	0	0	21.6	0	1.005867	26.566969	380	3	1.1	0.9;
	11805	2	0	0	0	0	0	1.030832	-9.752509	20	4	1.1	0.9;
	11806	2	0	0	0	0	0	0.990977	91.853031	20	3	1.1	0.9;
	11807	1	0	0	0	4.28	0	1.004043	-7.27619	220	1	1.1	0.9;
	11808	1	8.31	3	0	0.3	0	1.024183	12.272913	110	19	1.1	0.9;
	11809	1	156.6	16.9	0	12.26	0	1.025276	-26.834417	220	4	1.1	0.9;
	11810	2	0	0	0	0	0	1.023397	-17.3231	20	6	1.1	0.9;
	11811	2	0	0	0	0	0	1.026579	1.653291	20	6	1.1	0.9;
	11812	1	254.9	71.4	0	0.7	0	1.000643	27.976596	380	3	1.1	0.9;
	11813	2	0	0	0	0	0	1.024687	-5.034434	20	6	1.1	0.9;
	11814	2	0	0	0	0	0	1.024687	-5.034434	20	6	1.1	0.9;
	11815	2	0	0	0	0	0	1.026579	1.653291	20	6	1.1	0.9;
	11816	2	0	0	0	0	0	1.026579	1.653291	20	6	1.1	0.9;
	11817	2	0	0	0	0	0	1.029323	-3.215687	20	6	1.1	0.9;
	11818	2	0	0	0	0	0	1.029323	-3.215687	20	6	1.1	0.9;
	11819	2	0	0	0	0	0	1.024687	-5.034434	20	6	1.1	0.9;
	11820	2	0	0	0	0	0	1.024687	-5.034434	20	6	1.1	0.9;
	11821	1	12.91	4.2	0	0.35	0	0.986648	-14.509154	110	12	1.1	0.9;
	11822	2	0	0	0	0	0	1.050901	18.201002	20	7	1.1	0.9;
	11823	1	-0	-0	0	215.65	0	1.041349	33.378779	380	5	1.1	0.9;
	11824	1	-0	-0	0	24.17	0	1.031653	60.53182	380	3	1.1	0.9;
	11825	2	0	0	0	0	0	1.024059	-1.195476	154	1	1.1	0.9;
	11826	1	-0	-0	0	0	0	1.034698	-10.504371	150	9	1.1	0.9;
	11827	1	-0.17	-1.82	0	0	0	1.050857	44.08115	380	8	1.1	0.9;
	11828	1	83.99	6	0	65.12	0	1.013971	-22.771884	380	4	1.1	0.9;
	11829	1	-0	-0	0	4.22	0	1.039817	13.277797	220	5	1.1	0.9;
	11830	1	37.84	12.3	0.039146	9.70064	0	1.023783	-20.67892	220	6	1.1	0.9;
	11831	1	-0	0	0	5.73	0	1.022983	34.764528	150	8	1.1	0.9;
	11832	1	12.11	5.1	0	6.68	0	0.924454	-8.009866	110	18	1.1	0.9;
	11833	1	27.9	4	0	4.85	0	0.994977	26.39214	154	1	1.1	0.9;
	11834	1	70.38	27.5	0	0.77	0	0.986767	-2.140551	110	3	1.1	0.9;
	11835	1	93.8	15.9	0	0	0	1.065242	11.327355	110	24	1.1	0.9;
	11836	1	0	-0	0	0.07	0	0.971979	7.130049	110	18	1.1	0.9;
	11837	1	123.8	40.2	0.078704	3.951296	0	1.026312	17.731987	220	5	1.1	0.9;
	11838	1	-0	-0	0	0	0	1.01282	-20.654825	20	6	1.1	0.9;
	11839	1	0	-0	0.348979	0.528572	0	1.023827	61.091818	110	8	1.1	0.9;
	11840	1	-880.44	-449.03	0	0	0	1.040515	28.323211	27	3	1.1	0.9;
	11841	1	3.3	1.1	0.009871	0.810129	0	0.985038	-14.340693	110	22	1.1	0.9;
	11842	1	-0	-0	0	61.795782	0	1.010591	16.730632	380	5	1.1	0.9;
	11843	1	25	3.6	0	2.96	0	1.013919	18.923306	154	1	1.1	0.9;
	11844	1	0	0	0	0	0	0.97384	-5.930731	150	9	1.1	0.9;
	11845	1	0	0	0	66.29	0	1.041528	18.789335	380	6	1.1	0.9;
	11846	1	42.35	13.5	0	0.32	0	0.980712	-2.997082	110	18	1.1	0.9;
	11847	1	19.02	6	0	1.45	0	0.994501	-3.334975	110	21	1.1	0.9;
	11848	1	53.56	18.8	0	1.45	0	1.031466	4.488829	220	20	1.1	0.9;
	11849	2	0	0	0	0	0	1.040009	86.062395	20	3	1.1	0.9;
	11850	2	0	0	0	0	0	0.955439	48.700819	10.6	1	1.1	0.9;
	11851	2	0	0	0	0	0	1.056682	5.170602	20	18	1.1	0.9;
	11852	1	0	0	0	0.01	0	1.004296	17.471584	154	1	1.1	0.9;
	11853	2	0	0	0	0	0	1.057062	4.322623	20	18	1.1	0.9;
	11854	2	0	0	0	0	0	1.056682	5.170602	20	18	1.1	0.9;
	11855	1	28.63	8.8	0	0.67	0	0.997329	-9.773309	110	22	1.1	0.9;
	11856	2	0	0	0	0	0	1.021695	19.217324	13.8	1	1.1	0.9;
	11857	1	14.32	2.2	0	1.26	0	0.99325	-13.883981	110	22	1.1	0.9;
	11858	1	69.18	9.5	0	3.67	0	1.069542	22.417993	220	5	1.1	0.9;
	11859	2	0	0	0	0	0	1.009821	29.021748	6.3	1	1.1	0.9;
	11860	1	170.8	70.4	0	7.23	0	0.97939	6.577418	220	3	1.1	0.9;
	11861	1	-0	-0	0	9.54	0	1.022284	15.569104	220	5	1.1	0.9;
	11862	1	-19.32	6.4	0	0.23	0	0.98112	0.528152	150	9	1.1	0.9;
	11863	1	0	0	0	0	0	1.060995	0.247818	150	23	1.1	0.9;
	11864	1	36.9	5.3	0	3.7	0	1.002496	19.130689	154	1	1.1	0.9;
	11865	1	135.2	70.6	0	9.37	0	0.971128	-15.83459	150	9	1.1	0.9;
	11866	1	56.76	13.4	0	1.18	0	1.028521	7.435606	220	5	1.1	0.9;
	11867	1	0	0	0	3.22	0	1.013456	26.007542	380	3	1.1	0.9;
	11868	1	41.75	7.4	0	18.54	0	1.016223	-26.428264	220	4	1.1	0.9;
	11869	1	-0	-0	0	44.66	0	1.039579	-17.294301	380	10	1.1	0.9;
	11870	1	174.2	54.3	0	23.93	0	1.009571	-19.616145	380	6	1.1	0.9;
	11871	1	3.8	1.3	0	1.11	0	0.995342	-4.221904	110	18	1.1	0.9;
	11872	1	0	-0	0	0.48	0	0.98329	1.387261	110	18	1.1	0.9;
	11873	1	0	0	0	16.37	0	1.006753	32.438243	380	2	1.1	0.9;
	11874	1	68.48	42.7	0	6.37	0	1.032056	23.237993	220	3	1.1	0.9;
	11875	2	0	0	0	0	0	1.016616	-4.658699	20	4	1.1	0.9;
	11876	1	14.2	2	0	1.34	0	0.98849	14.04829	154	1	1.1	0.9;
	11877	1	6.21	1.5	0	1.42	0	0.987181	-4.824838	110	18	1.1	0.9;
	11878	1	44.65	-1.6	0.019811	15.410189	0	1.010274	-12.645316	220	4	1.1	0.9;
	11879	1	130	18.5	0	0.62	0	1.023918	1.897357	154	1	1.1	0.9;
	11880	2	0	0	0	0	0	1.010466	11.179882	20	18	1.1	0.9;
	11881	1	6.91	-11.1	0	0.67	0	1.025497	53.824202	110	8	1.1	0.9;
	11882	1	35.94	22.3	0	0	0	1.041516	-6.856455	150	9	1.1	0.9;
	11883	1	-0.67	-3.25	0	0	0	1.099209	-14.353653	150	9	1.2	0.8;
	11884	1	92.9	13.2	0	4.15	0	1.012574	31.34297	154	1	1.1	0.9;
	11885	1	35.14	11	0	0.48	0	1.000041	-13.895043	110	22	1.1	0.9;
	11886	2	0	0	0	0	0	1.013429	96.570197	20	3	1.1	0.9;
	11887	1	-0	0	0	0	0	1.008279	-10.478046	20	4	1.1	0.9;
	11888	1	0	0	1.052336	8.691699	0	1.081636	33.907239	380	5	1.2	0.8;
	11889	2	0	0	0	0	0	1.061603	44.385823	20	8	1.1	0.9;
	11890	1	9.7	1.4	0	5.11	0	0.981663	21.412796	154	1	1.1	0.9;
	11891	2	0	0	0	0	0	1.038438	0.065623	20	9	1.1	0.9;
	11892	2	0	0	0	0	0	1.017287	26.369278	20	5	1.1	0.9;
	11893	1	0	-0	0	17.69	0	1.034739	39.279987	380	8	1.1	0.9;
	11894	2	0	0	0	0	0	1.023338	0.199966	20	6	1.1	0.9;
	11895	2	0	0	0	0	0	1.010998	1.300317	20	10	1.1	0.9;
	11896	1	316.1	45	0	15.2	0	0.985586	4.453073	400	1	1.1	0.9;
	11897	2	0	0	0	0	0	1.061603	44.385823	20	8	1.1	0.9;
	11898	2	0	0	0	0	0	1.062536	46.73809	20	8	1.1	0.9;
	11899	1	0	-0	0	0.32	0	0.993049	0.964393	150	9	1.1	0.9;
	11900	2	0	0	0	0	0	0.964368	44.185324	11	1	1.1	0.9;
	11901	1	0	-0	0.469372	91.718414	0	1.031288	1.850677	380	6	1.1	0.9;
	11902	1	-0	-0	0.22542	1.27215	0	1.074011	71.405171	380	3	1.1	0.9;
	11903	2	0	0	0	0	0	1.03013	4.624843	6.3	1	1.1	0.9;
	11904	1	-0	-0	0.229027	1.548687	0	1.028716	93.98767	380	1	1.1	0.9;
	11905	1	0	0	0	5.56	0	1.024817	27.384645	380	3	1.1	0.9;
	11906	1	2.7	-0.1	0.038314	0.201686	0	1.053308	-2.057219	110	21	1.1	0.9;
	11907	1	13.21	6.6	0	1.19	0	0.954989	-16.611965	110	22	1.1	0.9;
	11908	1	-0	-0	0.096728	2.243272	0	1.026883	14.232028	220	7	1.1	0.9;
	11909	2	0	0	0	0	0	1.028568	11.503252	20	2	1.1	0.9;
	11910	1	132.1	13.2	0	6.9	0	1.001011	-3.053282	220	6	1.1	0.9;
	11911	1	25	3.6	0	1.31	0	1.000373	27.679835	154	1	1.1	0.9;
	11912	1	0	0	0	2.71	0	1.032162	11.548429	220	5	1.1	0.9;
	11913	2	0	0	0	0	0	1.033359	20.594354	20	5	1.1	0.9;
	11914	2	0	0	0	0	0	1.033359	20.594354	20	5	1.1	0.9;
	11915	1	82.1	29.1	0	0.4	0	1.004773	-7.512043	110	23	1.1	0.9;
	11916	2	0	0	0	0	0	1.026957	18.072383	20	5	1.1	0.9;
	11917	1	0	0	0	16.11	0	1.031923	19.780218	220	5	1.1	0.9;
	11918	1	-0	-0	0	1.74	0	1.027586	-0.028263	154	1	1.1	0.9;
	11919	2	0	0	0	0	0	1.026957	18.072383	20	5	1.1	0.9;
	11920	2	0	0	0	0	0	0.998951	38.095582	9	1	1.1	0.9;
	11921	1	59.9	8.5	0	0.88	0	1.023854	16.871011	154	1	1.1	0.9;
	11922	2	0	0	0	0	0	1.029382	20.080326	20	5	1.1	0.9;
	11923	1	68.18	16.1	0	3.35	0	1.027524	7.135692	220	5	1.1	0.9;
	11924	2	0	0	0	0	0	1.0608	42.720858	20	3	1.1	0.9;
	11925	1	71.98	34.1	0	7.65	0	1.059073	29.479384	220	5	1.1	0.9;
	11926	1	35.84	13	0	2.26	0	1.022428	-2.963232	110	19	1.1	0.9;
	11927	2	0	0	0	0	0	1.058046	7.877809	380	5	1.1	0.9;
	11928	1	-0	0	0	0.03	0	0.976931	30.182064	400	1	1.1	0.9;
	11929	1	0	0	0	4.13	0	1.02281	-10.941228	150	9	1.1	0.9;
	11930	1	87.9	29.3	0	1.21	0	1.042018	29.370588	220	5	1.1	0.9;
	11931	1	59.27	13.3	0.039187	3.630407	0	1.032947	12.673591	220	5	1.1	0.9;
	11932	1	56.7	8	0	4.42	0	1.002928	17.35183	154	1	1.1	0.9;
	11933	1	0	0	0	5.46	0	1.024842	6.070005	220	5	1.1	0.9;
	11934	1	-0	-0	0	30.39	0	1.010399	38.6722	380	8	1.1	0.9;
	11935	2	0	0	0	0	0	0.994234	37.670381	13.8	1	1.1	0.9;
	11936	1	19.02	6.3	0	0.2	0	0.970991	4.942762	110	18	1.1	0.9;
	11937	1	200.6	28.6	0	0.29	0	1.024614	2.680146	154	1	1.1	0.9;
	11938	1	73.98	-7.4	0.705428	109.0541	0	1.021685	-10.456797	380	6	1.1	0.9;
	11939	1	29.23	-49.9	0	5.49	0	1.012907	2.260233	220	6	1.1	0.9;
	11940	1	111.1	35.6	0	0.53	0	0.978898	-6.597551	110	3	1.1	0.9;
	11941	1	-0	0	0.038239	0.171761	0	1.07251	0.28715	110	15	1.1	0.9;
	11942	1	0	-0	0	26.46	0	1.009609	-16.526291	380	6	1.1	0.9;
	11943	1	-0	-0	0	1.25	0	1.00842	28.258293	154	1	1.1	0.9;
	11944	1	164.4	23.4	0	10.01	0	1.036963	-1.571198	154	1	1.1	0.9;
	11945	2	0	0	0	0	0	1.012893	26.368849	11	1	1.1	0.9;
	11946	2	0	0	0	0	0	1.033214	-6.888934	20	6	1.1	0.9;
	11947	1	-0.01	-0	0	0	0	1.071016	58.767652	220	8	1.1	0.9;
	11948	1	80.49	7.3	0	1.81	0	1.019544	-32.956369	220	4	1.1	0.9;
	11949	1	-0	-0	0	9.49	0	1.033686	13.392829	380	1	1.1	0.9;
	11950	2	0	0	0	0	0	1.04459	35.949142	20	5	1.1	0.9;
	11951	1	-0	0	0.170749	0.347709	0	1.054686	-2.922157	150	9	1.1	0.9;
	11952	1	-0	0	0	23.91	0	1.040569	13.793762	380	1	1.1	0.9;
	11953	2	0	0	0	0	0	0.951565	38.529485	13.8	1	1.1	0.9;
	11954	1	8.91	2.5	0	0.47	0	0.959267	-6.51368	110	18	1.1	0.9;
	11955	1	-0	0	0	8.1	0	1.063275	21.172664	380	5	1.1	0.9;
	11956	1	-0	0	0	1.4	0	0.984374	88.942692	150	1	1.1	0.9;
	11957	1	34.44	27	0	0.34	0	0.90264	-12.097367	150	9	1.1	0.9;
	11958	1	0	0	0	0.23	0	1.007569	-12.183288	150	9	1.1	0.9;
	11959	2	0	0	0	0	0	1.033214	-6.86015	20	6	1.1	0.9;
	11960	1	19.4	2.8	0	3.46	0	0.989121	29.435974	154	1	1.1	0.9;
	11961	1	0	-0	0.039166	8.300834	0	1.025491	37.337345	220	3	1.1	0.9;
	11962	2	0	0	0	0	0	1.033214	-6.86015	20	6	1.1	0.9;
	11963	1	47.4	6.7	0	3.82	0	1.00389	18.263979	154	1	1.1	0.9;
	11964	1	67.6	9.6	0	2.95	0	0.991781	14.889742	154	1	1.1	0.9;
	11965	1	-0	-0	0	2.67	0	1.033623	30.239446	150	8	1.1	0.9;
	11966	1	242.4	46.7	0	75.67	0	1.011867	1.688451	380	6	1.1	0.9;
	11967	2	0	0	0	0	0	1.019503	8.365451	6.3	1	1.1	0.9;
	11968	2	0	0	0	0	0	1.023144	-9.538309	20	4	1.1	0.9;
	11969	1	-0	0	0	50	0	1.034068	30.209322	150	8	1.1	0.9;
	11970	2	0	0	0	0	0	0.988112	23.208642	6.3	1	1.1	0.9;
	11971	1	10.01	2	0	0.73	0	0.983835	1.517642	150	7	1.1	0.9;
	11972	1	16.7	2.4	0	15.54	0	1.00288	28.368105	154	1	1.1	0.9;
	11973	1	56.26	18.5	0	1.01	0	0.999369	-13.617092	110	12	1.1	0.9;
	11974	1	11.01	4	0	0.73	0	0.966317	0.966505	110	19	1.1	0.9;
	11975	2	0	0	0	0	0	1.015227	-3.436884	20	10	1.1	0.9;
	11976	2	0	0	0	0	0	1.04826	-8.707987	20	10	1.1	0.9;
	11977	1	169.93	42.88	0	9.84	0	1.02486	-11.460772	220	17	1.1	0.9;
	11978	2	0	0	0	0	0	1.054388	47.373035	20	3	1.1	0.9;
	11979	2	0	0	0	0	0	1.054388	47.373035	20	3	1.1	0.9;
	11980	2	0	0	0	0	0	1.054388	47.373035	20	3	1.1	0.9;
	11981	1	0	0	0	0	0	1.020921	-7.474467	150	9	1.1	0.9;
	11982	2	0	0	0	0	0	1.04876	45.73683	20	10	1.1	0.9;
	11983	1	303.5	56.2	0	21.07	0	1.026753	-20.195844	220	4	1.1	0.9;
	11984	1	3.8	1.9	0	0.72	0	1.009654	2.894147	110	18	1.1	0.9;
	11985	1	27.53	12.1	0	0.88	0	1.000468	-9.336623	110	22	1.1	0.9;
	11986	2	0	0	0	0	0	1.01601	-0.145046	20	10	1.1	0.9;
	11987	2	0	0	0	0	0	0.971901	45.235162	15.75	1	1.1	0.9;
	11988	2	0	0	0	0	0	1.037973	11.471538	20	14	1.1	0.9;
	11989	1	111.36	27.44	0	2.59	0	0.989988	-18.685834	220	4	1.1	0.9;
	11990	2	0	0	0	0	0	1.002465	24.966083	11	1	1.1	0.9;
	11991	1	35.1	5	0	1.57	0	1.003002	18.950589	154	1	1.1	0.9;
	11992	1	-0	-0	0	4.47	0	0.966158	-9.712391	150	9	1.1	0.9;
	11993	1	-0	-0	0	1.28	0	0.978141	-14.535595	150	9	1.1	0.9;
	11994	1	21.52	5.1	0	0.25	0	1.01023	-1.501018	110	18	1.1	0.9;
	11995	1	0	0	0	0.15	0	1.053059	-2.989742	150	9	1.1	0.9;
	11996	2	0	0	0	0	0	1.032132	-0.658602	20	9	1.1	0.9;
	11997	1	125.3	34.6	0	4.36	0	1.024048	34.917994	150	8	1.1	0.9;
	11998	1	-0	0	0.113823	1.255147	0	1.072837	36.250253	380	14	1.1	0.9;
	11999	1	0	-0	0	0.12	0	1.012583	31.342815	154	1	1.1	0.9;
	12000	1	-0	0	0	29.73	0	1.051128	37.961676	380	3	1.1	0.9;
	12001	2	0	0	0	0	0	1.016606	10.005767	15.75	1	1.1	0.9;
	12002	1	0	0	0.233623	-0.194686	0	1.031386	-10.502639	380	6	1.1	0.9;
	12003	1	-0	-0	0.097109	0.522891	0	1.040737	28.962102	220	5	1.1	0.9;
	12004	1	2.1	396.2	0.407504	115.720413	0	1.045892	-13.188115	380	4	1.1	0.9;
	12005	1	0	0	0	0	0	1.016233	-10.469824	150	9	1.1	0.9;
	12006	1	-34.04	-2	0	197.23	0	1.021912	-13.38294	380	6	1.1	0.9;
	12007	1	-0	-0	0	1.28	0	0.994863	-7.726774	150	9	1.1	0.9;
	12008	1	76.59	16.2	0	21.54	0	1.014403	-26.633426	220	4	1.1	0.9;
	12009	1	0	0	0	0	0	1.017509	4.48626	20	23	1.1	0.9;
	12010	1	0	0	0	0	0	1.017509	4.48626	20	23	1.1	0.9;
	12011	1	0	0	0	0	0	1.017509	4.48626	20	23	1.1	0.9;
	12012	1	-0	0	0.099501	82.133595	0	1.074409	0.241246	380	18	1.2	0.8;
	12013	2	0	0	0	0	0	1.008914	18.33376	20	10	1.1	0.9;
	12014	1	-0	0	0	1.2	0	1.023938	3.432435	220	5	1.1	0.9;
	12015	1	0	-0	0.090774	5.709044	0	1.056086	-3.056772	220	18	1.2	0.8;
	12016	2	0	0	0	0	0	1.065343	53.072846	20	10	1.1	0.9;
	12017	1	0	0	0	1.22	0	0.999005	27.908744	150	8	1.1	0.9;
	12018	1	0	-0	0	0	0	1.010498	-17.064676	20	4	1.1	0.9;
	12019	1	31.63	7.5	0	21.91	0	1.032873	-19.373029	220	4	1.1	0.9;
	12020	2	0	0	0	0	0	1.015252	-6.735079	20	4	1.1	0.9;
	12021	2	0	0	0	0	0	1.018676	-12.462271	20	4	1.1	0.9;
	12022	1	-0	0	0	0.39	0	0.998709	-8.901449	150	9	1.1	0.9;
	12023	1	-12.15	-0.84	0	0	0	1.041308	-1.973007	220	6	1.1	0.9;
	12024	2	0	0	0	0	0	1.034048	-9.733404	20	4	1.1	0.9;
	12025	2	0	0	0	0	0	1.030499	6.658168	20	6	1.1	0.9;
	12026	1	23.03	14.3	0	16.88	0	1.033858	-10.849314	150	9	1.1	0.9;
	12027	2	0	0	0	0	0	1.042665	-9.557361	20	4	1.1	0.9;
	12028	1	2.8	14.6	0	113.059828	0	1.123219	-0.417507	380	9	1.2	0.8;
	12029	1	-0	0	0.706809	0.700443	0	1.024385	-8.261011	380	6	1.1	0.9;
	12030	1	22.12	8	0	0.53	0	1.014648	-4.044158	110	19	1.1	0.9;
	12031	1	-0	0	0	0.79	0	1.055555	-3.757386	110	19	1.1	0.9;
	12032	1	13.11	3.6	0	0.38	0	1.017145	17.949482	220	13	1.1	0.9;
	12033	1	10.81	3.6	0	0.38	0	0.987995	-14.036111	110	12	1.1	0.9;
	12034	1	26.93	9.6	0	1.41	0	1.006456	-4.02582	110	21	1.1	0.9;
	12035	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	12036	1	41.8	5.9	0	10.67	0	1.002452	16.388576	154	1	1.1	0.9;
	12037	1	0	0	0	0	0	1.002937	-4.683269	150	9	1.1	0.9;
	12038	1	0	-0	0.189785	19.193314	0	0.992378	34.134663	380	2	1.1	0.9;
	12039	1	0	-0	0	6.59	0	1.049848	19.032506	220	3	1.1	0.9;
	12040	1	72.38	23	0	0.28	0	1.02822	-20.958707	220	4	1.1	0.9;
	12041	1	-0	0	0.211407	6.615121	0	0.994127	33.837786	380	2	1.1	0.9;
	12042	2	0	0	0	0	0	1.014428	-7.519045	20	10	1.1	0.9;
	12043	1	73.58	-13.1	0	7.58	0	1.036533	7.566778	220	3	1.1	0.9;
	12044	1	0	-0	0	2.97	0	1.046552	-20.234692	380	10	1.1	0.9;
	12045	2	0	0	0	0	0	1.013722	-11.978977	20	4	1.1	0.9;
	12046	1	0	-0	0	3.33	0	1.023058	-10.869221	150	9	1.1	0.9;
	12047	1	0	-0	0.150629	-0.150629	0	1.006941	-8.995126	380	6	1.1	0.9;
	12048	1	99.6	22.9	0	15.02	0	1.015467	-26.709784	220	4	1.1	0.9;
	12049	2	0	0	0	0	0	1.073911	4.277217	20	20	1.1	0.9;
	12050	2	0	0	0	0	0	1.008348	30.337205	14.4	1	1.1	0.9;
	12051	1	54.96	62.6	0.296391	5.573008	0	1.011292	29.334797	380	3	1.1	0.9;
	12052	1	120.2	30.7	0	9.98	0	1.008412	69.340675	220	3	1.1	0.9;
	12053	1	-0.2	-1	0	0	0	1.025116	22.799171	150	8	1.1	0.9;
	12054	1	0	0	0	0.32	0	1.036891	21.208732	220	5	1.1	0.9;
	12055	1	0	-0	0	11.38	0	1.018472	13.591184	220	5	1.1	0.9;
	12056	1	2	1	0	1	0	0.956568	-7.820575	110	21	1.1	0.9;
	12057	1	0	-0	0	0.84	0	1.027437	79.001946	220	3	1.1	0.9;
	12058	1	-0	0	0.563882	82.938191	0	1.048954	68.758885	380	3	1.1	0.9;
	12059	1	-0.18	-1	0	0	0	1.021269	36.651421	150	8	1.1	0.9;
	12060	2	0	0	0	0	0	0.996411	28.78407	6.3	1	1.1	0.9;
	12061	1	20.2	2.9	0	5.89	0	1.000158	26.820337	154	1	1.1	0.9;
	12062	1	-134.35	-6.5	0	3.38	0	0.998503	-10.411516	150	10	1.1	0.9;
	12063	1	165.9	49.6	0	6.13	0	1.016437	51.339973	220	3	1.1	0.9;
	12064	1	169.6	98.3	0	27.8	0	1.005625	13.356599	380	3	1.1	0.9;
	12065	2	0	0	0	0	0	1.031165	22.577641	20	10	1.1	0.9;
	12066	1	0.18	87.5	0	3.94	0	0.993675	90.655341	150	1	1.1	0.9;
	12067	1	11.8	1.7	0	6.98	0	1.015757	31.653843	154	1	1.1	0.9;
	12068	1	202	28.7	0	0.36	0	1.016555	19.710237	154	1	1.1	0.9;
	12069	2	0	0	0	0	0	1.035406	21.74444	20	5	1.1	0.9;
	12070	2	0	0	0	0	0	1.035406	21.74444	20	5	1.1	0.9;
	12071	2	0	0	0	0	0	1.035406	21.74444	20	5	1.1	0.9;
	12072	2	0	0	0	0	0	1.051326	28.77397	20	14	1.1	0.9;
	12073	2	0	0	0	0	0	1.035406	21.74444	20	5	1.1	0.9;
	12074	2	0	0	0	0	0	1.035406	21.74444	20	5	1.1	0.9;
	12075	2	0	0	0	0	0	1.035406	21.74444	20	5	1.1	0.9;
	12076	2	0	0	0	0	0	1.035406	21.74444	20	5	1.1	0.9;
	12077	1	3.8	3.8	0	0.95	0	0.950409	-5.384362	110	18	1.1	0.9;
	12078	1	83.39	16.5	0	16.43	0	1.024668	10.263437	220	5	1.1	0.9;
	12079	1	7.5	173.3	0.962911	57.817695	0	1.017661	-10.180091	380	4	1.1	0.9;
	12080	1	0	0	0	65.93	0	1.021811	49.289996	380	8	1.1	0.9;
	12081	1	0	-0	0.038816	1.131184	0	1.018208	-5.092604	110	18	1.1	0.9;
	12082	1	207.9	25.4	0	1.75	0	1.014547	-18.70804	220	4	1.1	0.9;
	12083	1	2.5	0.5	0	0.8	0	0.987102	3.009904	110	18	1.1	0.9;
	12084	2	0	0	0	0	0	1.012034	5.351331	20	2	1.1	0.9;
	12085	2	0	0	0	0	0	1.012034	5.351331	20	2	1.1	0.9;
	12086	1	-0	-0	0	10.28	0	1.033277	23.511773	220	3	1.1	0.9;
	12087	1	63.77	39.5	0	1.12	0	0.970315	-8.449974	150	9	1.1	0.9;
	12088	2	0	0	0	0	0	1.014269	27.689496	20	10	1.1	0.9;
	12089	1	0	-0	0	1.57	0	1.044104	2.133328	154	1	1.1	0.9;
	12090	1	5.31	2.3	0	0.93	0	0.969719	-6.483101	110	18	1.1	0.9;
	12091	2	0	0	0	0	0	1.005112	-1.198328	20	18	1.1	0.9;
	12092	1	0	0	0	0	0	1.049093	7.513515	220	1	1.1	0.9;
	12093	2	0	0	0	0	0	1.056125	27.444238	20	3	1.1	0.9;
	12094	1	-0	-0	0.041633	77.581311	0	0.969713	18.803207	380	5	1.1	0.9;
	12095	2	0	0	0	0	0	1.024647	-2.171451	20	6	1.1	0.9;
	12096	1	-0	0	0	24.67	0	0.976458	41.660917	400	1	1.1	0.9;
	12097	2	0	0	0	0	0	1.026207	-11.793103	20	4	1.1	0.9;
	12098	1	135.7	19.3	0	3.58	0	1.00499	31.934591	154	1	1.1	0.9;
	12099	2	0	0	0	0	0	1.030988	2.394145	20	9	1.1	0.9;
	12100	1	92.7	-5.6	0.2527	35.620175	0	1.04154	-14.32672	380	10	1.1	0.9;
	12101	1	0	0	0.019665	0.420335	0	1.029639	17.04104	220	5	1.1	0.9;
	12102	2	0	0	0	0	0	1.023818	-10.471889	20	4	1.1	0.9;
	12103	1	65.5	9.3	0	7.21	0	1.005395	30.514439	154	1	1.1	0.9;
	12104	1	26.33	7.6	0	0.32	0	1.006974	2.951136	110	18	1.1	0.9;
	12105	1	0	0	0	0.55	0	0.987737	46.492873	154	1	1.1	0.9;
	12106	2	0	0	0	0	0	1.024267	0.497112	20	18	1.1	0.9;
	12107	1	12.11	5.5	0	1.19	0	1.002453	-9.411953	110	22	1.1	0.9;
	12108	1	0	0	0	0	0	1.037317	25.872032	150	2	1.1	0.9;
	12109	1	102.8	30.8	0	2.66	0	1.042633	20.102708	380	3	1.1	0.9;
	12110	1	-0	0	0	54.41	0	0.978701	25.524635	400	1	1.1	0.9;
	12111	1	0	-0	0	0.32	0	0.982065	-7.80271	150	9	1.1	0.9;
	12112	1	131.9	50.6	0	79.02	0	1.014425	-9.789718	380	6	1.1	0.9;
	12113	1	-0	0	0	1.5	0	1.012055	-4.801845	150	9	1.1	0.9;
	12114	1	0	0	0	5.67	0	0.978251	10.243079	220	1	1.1	0.9;
	12115	1	1.9	1.3	0	0.43	0	1.040444	6.867695	110	18	1.1	0.9;
	12116	2	0	0	0	0	0	1.031695	0	20	6	1.1	0.9;
	12117	1	-0	-0	0	0.31	0	1.029233	-10.064406	150	9	1.1	0.9;
	12118	1	118.1	20.2	0	1.47	0	1.012268	-26.760459	220	4	1.1	0.9;
	12119	1	370.2	107.7	0	81.59	0	1.013515	-9.02175	380	6	1.1	0.9;
	12120	1	0	0	0	0	0	1.053349	-2.98314	150	9	1.1	0.9;
	12121	2	0	0	0	0	0	1.049756	20.130595	20	5	1.1	0.9;
	12122	1	0	-0	0	0	0	1.025009	18.463948	20	5	1.1	0.9;
	12123	2	0	0	0	0	0	1.003288	23.155407	20	5	1.1	0.9;
	12124	2	0	0	0	0	0	1.025271	20.88058	20	5	1.1	0.9;
	12125	1	5.11	2.5	0	0.19	0	1.044881	6.491094	110	18	1.1	0.9;
	12126	2	0	0	0	0	0	1.025372	28.364246	20	5	1.1	0.9;
	12127	2	0	0	0	0	0	1.025372	28.364246	20	5	1.1	0.9;
	12128	1	44.75	27.7	0	1.14	0	1.001622	-9.968038	150	9	1.1	0.9;
	12129	1	0	-0	0	8.13	0	1.007399	27.999318	380	3	1.1	0.9;
	12130	1	69.28	23.8	0	2.56	0	1.002279	-13.779687	110	22	1.1	0.9;
	12131	2	0	0	0	0	0	1.039883	-8.632286	20	6	1.1	0.9;
	12132	2	0	0	0	0	0	1.039883	-8.632286	20	6	1.1	0.9;
	12133	2	0	0	0	0	0	1.033722	29.674924	20	6	1.1	0.9;
	12134	2	0	0	0	0	0	1.033722	29.674924	20	6	1.1	0.9;
	12135	2	0	0	0	0	0	1.033722	29.674924	20	6	1.1	0.9;
	12136	2	0	0	0	0	0	1.033722	29.674924	20	6	1.1	0.9;
	12137	2	0	0	0	0	0	1.033736	29.645181	20	6	1.1	0.9;
	12138	2	0	0	0	0	0	1.033736	29.645181	20	6	1.1	0.9;
	12139	2	0	0	0	0	0	1.033736	29.645181	20	6	1.1	0.9;
	12140	2	0	0	0	0	0	1.033736	29.645181	20	6	1.1	0.9;
	12141	1	216.8	30.7	0	58.2	0	1.05459	7.846898	380	16	1.1	0.9;
	12142	1	0	0	0.710082	-0.042148	0	1.026998	21.381155	380	5	1.1	0.9;
	12143	1	0.4	182.7	0.233255	122.325059	0	1.037456	-12.983195	380	4	1.1	0.9;
	12144	1	0	0	0	8.88	0	0.972666	20.278257	400	1	1.1	0.9;
	12145	1	24.12	9.8	0	0.04	0	1.021211	26.174949	150	8	1.1	0.9;
	12146	1	196.1	32.1	0	3.86	0	0.978549	1.896927	220	5	1.1	0.9;
	12147	1	9.61	2.5	0	0.37	0	0.976996	-5.755046	110	18	1.1	0.9;
	12148	1	3	-0	0.236576	6.812853	0	1.012529	-13.185268	380	6	1.1	0.9;
	12149	1	0	-0	0	1.56	0	1.008452	19.375763	154	1	1.1	0.9;
	12150	1	39.04	13.1	0	3.22	0	1.010382	15.656999	220	5	1.1	0.9;
	12151	1	-0.24	-23.15	0	26.49	0	1.05433	45.681206	380	3	1.1	0.9;
	12152	2	0	0	0	0	0	1.030656	3.405373	20	6	1.1	0.9;
	12153	1	0	-0	0	69.23	0	1.137373	-1.114463	380	23	1.2	0.8;
	12154	1	19.4	2.8	0	6.62	0	1.011305	28.622228	154	1	1.1	0.9;
	12155	1	-0	0	0	10.21	0	1.017379	-22.651908	380	4	1.1	0.9;
	12156	1	69.38	22.5	0	0.67	0	0.994871	-3.848423	110	21	1.1	0.9;
	12157	1	7	1	0	4	0	0.975069	16.05223	154	1	1.1	0.9;
	12158	1	255.9	128.3	0	54.58	0	1.020063	-14.094216	380	17	1.1	0.9;
	12159	1	0	0	0	0	0	1.038979	-13.937788	20	4	1.1	0.9;
	12160	1	-0	-0	0	65.73	0	1.021838	49.280767	380	8	1.1	0.9;
	12161	1	0	0	1.147648	-0.504228	0	1.047022	28.781319	380	5	1.1	0.9;
	12162	1	316.7	46.1	0	46.23	0	0.999462	-21.242915	380	6	1.1	0.9;
	12163	2	0	0	0	0	0	1.05667	-2.722877	20	5	1.1	0.9;
	12164	2	0	0	0	0	0	1.05667	-2.722877	20	5	1.1	0.9;
	12165	1	0	-0	0	24.67	0	1.10093	-2.183793	380	1	1.2	0.8;
	12166	2	0	0	0	0	0	1.030179	32.417925	20	5	1.1	0.9;
	12167	2	0	0	0	0	0	1.030179	32.417925	20	5	1.1	0.9;
	12168	2	0	0	0	0	0	1.030179	32.417925	20	5	1.1	0.9;
	12169	2	0	0	0	0	0	1.030179	32.417925	20	5	1.1	0.9;
	12170	2	0	0	0	0	0	1.05667	-2.722877	20	5	1.1	0.9;
	12171	2	0	0	0	0	0	1.05667	-2.722877	20	5	1.1	0.9;
	12172	2	0	0	0	0	0	1.030179	32.417925	20	5	1.1	0.9;
	12173	2	0	0	0	0	0	1.04071	74.319379	20	3	1.1	0.9;
	12174	1	15.62	2.8	0	0.71	0	0.952652	-11.963627	110	18	1.1	0.9;
	12175	2	0	0	0	0	0	1.035888	21.491029	20	5	1.1	0.9;
	12176	1	120.6	35.7	0	15.36	0	1.005453	-5.630958	380	6	1.1	0.9;
	12177	1	21.82	6.3	0	2.01	0	1.00191	-5.355721	110	18	1.1	0.9;
	12178	2	0	0	0	0	0	1.014358	8.671133	11.5	1	1.1	0.9;
	12179	1	15.62	5.7	0	0.17	0	1.017635	-3.080045	110	19	1.1	0.9;
	12180	1	107.9	35.9	0	5.48	0	1.006681	30.125372	150	8	1.1	0.9;
	12181	1	40.05	20	0	7.45	0	0.998766	-23.003981	220	6	1.1	0.9;
	12182	1	0	-0	0	25.28	0	0.979669	-3.90823	380	6	1.1	0.9;
	12183	1	0	0	0	1.6	0	0.988243	13.342353	110	18	1.1	0.9;
	12184	1	57.66	6.1	0	1.77	0	1.025229	54.199763	110	8	1.1	0.9;
	12185	1	0	0	0.059718	0.860282	0	1.000192	-4.465417	150	10	1.1	0.9;
	12186	1	-0	-0	0	0.62	0	1.057634	18.249721	380	14	1.1	0.9;
	12187	1	114.42	23.58	0	46.99	0	1.019178	34.651774	380	13	1.1	0.9;
	12188	1	14.32	5.5	0	3.15	0	0.95809	-16.391366	110	22	1.1	0.9;
	12189	2	0	0	0	0	0	1.007227	24.551226	11	1	1.1	0.9;
	12190	1	9.41	5.6	0	0.03	0	1.023431	-10.714615	150	9	1.1	0.9;
	12191	2	0	0	0	0	0	1.013974	-8.858315	20	22	1.1	0.9;
	12192	2	0	0	0	0	0	1.013974	-8.858315	20	22	1.1	0.9;
	12193	2	0	0	0	0	0	1.015043	-15.09113	20	22	1.1	0.9;
	12194	2	0	0	0	0	0	1.015043	-15.09113	20	22	1.1	0.9;
	12195	2	0	0	0	0	0	1.048748	7.516157	20	14	1.1	0.9;
	12196	1	-0	0	0	36.43	0	1.087414	8.898313	380	1	1.2	0.8;
	12197	1	-0	-0	0	17.3	0	1.040006	41.084857	380	3	1.1	0.9;
	12198	1	-0.12	-1.31	0	0	0	1.098531	65.002694	380	8	1.2	0.8;
	12199	2	0	0	0	0	0	0.991398	31.360075	15.75	1	1.1	0.9;
	12200	2	0	0	0	0	0	1.031553	24.597119	20	5	1.1	0.9;
	12201	1	-0	-0	0	0.3	0	0.995093	5.914505	110	18	1.1	0.9;
	12202	1	0	0	0	2.56	0	1.023421	5.136164	220	3	1.1	0.9;
	12203	1	16.02	5.1	0	8.6	0	1.024504	38.533334	110	8	1.1	0.9;
	12204	2	0	0	0	0	0	0.975834	33.412223	18	1	1.1	0.9;
	12205	1	-29.84	-2.96	0	0.28	0	1.039675	26.268321	110	18	1.1	0.9;
	12206	1	-0	0	0	1.06	0	0.994875	-4.178758	110	18	1.1	0.9;
	12207	1	0	0	0	6.52	0	1.032194	21.496822	220	5	1.1	0.9;
	12208	2	0	0	0	0	0	1.043874	15.587474	20	18	1.1	0.9;
	12209	1	394.41	96.3	0.233064	1.955491	0	1.033818	27.836778	220	13	1.1	0.9;
	12210	1	112.7	14.5	0	1.24	0	1.036271	5.075112	220	5	1.1	0.9;
	12211	2	0	0	0	0	0	1.026027	-7.002651	20	4	1.1	0.9;
	12212	1	82.99	25.9	0	0.23	0	1.024841	18.337972	220	5	1.1	0.9;
	12213	1	5.01	2	0	0.6	0	1.038536	1.12257	110	23	1.1	0.9;
	12214	1	-0	-0	0	5.81	0	1.006583	-15.130947	150	9	1.1	0.9;
	12215	1	-0	0	0	0.18	0	0.996897	23.818152	154	1	1.1	0.9;
	12216	2	0	0	0	0	0	1.021005	21.987746	20	14	1.1	0.9;
	12217	2	0	0	0	0	0	1.036966	15.051213	20	14	1.1	0.9;
	12218	1	0	0	0	14.35	0	1.098453	64.998015	380	8	1.2	0.8;
	12219	2	0	0	0	0	0	1.043942	17.353218	20	5	1.1	0.9;
	12220	1	31.03	10.3	0	0.89	0	0.973261	-9.531782	110	21	1.1	0.9;
	12221	2	0	0	0	0	0	1.029545	3.003491	11	1	1.1	0.9;
	12222	2	0	0	0	0	0	1.02494	28.997481	20	5	1.1	0.9;
	12223	1	-0	0	0	15	0	1.08404	51.888439	220	3	1.2	0.8;
	12224	1	-0	0	0	0.05	0	1.095691	-4.631591	150	9	1.1	0.9;
	12225	1	59.67	9.2	0	18.03	0	1.019604	-24.436994	220	4	1.1	0.9;
	12226	1	13.92	3.8	0	1.26	0	1.000971	-4.328082	110	18	1.1	0.9;
	12227	2	0	0	0	0	0	1.043087	11.638661	20	14	1.1	0.9;
	12228	1	0	-0	0	0.93	0	1.01227	-2.402886	150	9	1.1	0.9;
	12229	1	0	-0	0.361302	2.888987	0	1.021096	23.640191	380	6	1.1	0.9;
	12230	1	0	-0	0	8.18	0	0.970372	27.607329	400	1	1.1	0.9;
	12231	1	99.3	23.6	0	1.77	0	0.998346	-4.38145	110	21	1.1	0.9;
	12232	2	0	0	0	0	0	1.037993	28.716517	20	5	1.1	0.9;
	12233	1	-0	0	0	0.15	0	1.050133	-3.105156	150	9	1.1	0.9;
	12234	1	-0	-0	0	0.14	0	1.050186	25.602515	220	5	1.1	0.9;
	12235	1	33.44	11	0	4	0	0.990868	-2.50336	110	21	1.1	0.9;
	12236	1	11.2	1.6	0	2.09	0	0.996281	12.615694	154	1	1.1	0.9;
	12237	1	10.31	2.3	0	1.11	0	0.970208	5.158798	110	18	1.1	0.9;
	12238	1	43.45	15.9	0.0196	1.8004	0	1.01	12.196035	220	5	1.1	0.9;
	12239	1	86.7	41.5	0	18.57	0	1.012035	-12.726866	150	9	1.1	0.9;
	12240	2	0	0	0	0	0	1.022424	-10.029386	20	4	1.1	0.9;
	12241	1	-0	0	1.373975	28.45227	0	1.065676	17.770112	380	5	1.1	0.9;
	12242	1	26.73	10.8	0	0.43	0	0.989904	-7.663746	110	18	1.1	0.9;
	12243	1	-0	-0	0	0.36	0	0.965855	-6.047065	110	18	1.1	0.9;
	12244	1	0	0	0	0	0	1.029293	-6.89657	150	9	1.1	0.9;
	12245	2	0	0	0	0	0	1.037281	-8.202796	20	4	1.1	0.9;
	12246	1	0	-0	0	0	0	1.034774	-11.388611	20	4	1.1	0.9;
	12247	1	13.92	8.9	0	2.65	0	0.97655	2.339227	110	18	1.1	0.9;
	12248	2	0	0	0	0	0	0.988817	21.215061	19	1	1.1	0.9;
	12249	1	19.32	7	0	0.44	0	1.004014	-3.506061	110	19	1.1	0.9;
	12250	1	-0	-0	0.531454	0.176647	0	1.026831	21.377342	380	5	1.1	0.9;
	12251	1	-0	0	0	1.12	0	0.97451	22.850553	400	1	1.1	0.9;
	12252	2	0	0	0	0	0	1.063358	20.842529	20	14	1.1	0.9;
	12253	1	0	-0	0	0	0	1.050524	44.07973	380	8	1.1	0.9;
	12254	1	18.02	6.5	0	2.19	0	1.075269	1.500112	110	19	1.2	0.8;
	12255	2	0	0	0	0	0	1.062762	29.233304	20	14	1.1	0.9;
	12256	1	313.75	96.8	0	5.67	0	0.98827	-24.136428	220	6	1.1	0.9;
	12257	1	74.78	26.4	0	2.42	0	0.996144	-14.280798	110	22	1.1	0.9;
	12258	1	-0	0	0.278602	31.723346	0	1.095609	8.932766	380	19	1.2	0.8;
	12259	2	0	0	0	0	0	1.06345	20.85668	20	14	1.1	0.9;
	12260	1	0	-0	0	3.03	0	1.025901	4.522465	220	3	1.1	0.9;
	12261	1	41.45	56.2	0	6.55	0	1.00971	-11.697978	220	17	1.1	0.9;
	12262	1	-0	0	0	0.29	0	1.000845	34.550179	154	1	1.1	0.9;
	12263	1	-0	-8	0	0	0	1.009381	-12.831934	150	9	1.1	0.9;
	12264	1	82.39	14.4	0	13.37	0	1.027923	5.769471	220	5	1.1	0.9;
	12265	1	0	-0	0.057346	7.362654	0	1.041994	0.526847	220	1	1.1	0.9;
	12266	2	0	0	0	0	0	1.034282	8.113127	11	1	1.1	0.9;
	12267	1	89.44	63.66	0	2.6	0	1.02525	28.139664	220	13	1.1	0.9;
	12268	1	8.91	5.1	0	0.42	0	1.045266	6.578998	110	18	1.1	0.9;
	12269	1	0	0	0	9.41	0	1.062465	80.362016	380	3	1.1	0.9;
	12270	1	145.2	21.4	0	1.09	0	1.013884	6.141471	220	5	1.1	0.9;
	12271	1	174.7	31	0.23738	27.44968	0	1.057978	83.728387	380	3	1.1	0.9;
	12272	1	0	-0	0	33.298535	0	1.026362	14.401429	380	5	1.1	0.9;
	12273	1	637.92	139.57	0	0	0	1.004873	26.961875	150	2	1.1	0.9;
	12274	1	257.58	-303.11	0.290737	64.242856	0	1.006483	26.909077	380	3	1.1	0.9;
	12275	2	0	0	0	0	0	1.042396	85.13469	20	3	1.1	0.9;
	12276	2	0	0	0	0	0	1.042396	85.13469	20	3	1.1	0.9;
	12277	2	0	0	0	0	0	1.00107	32.866746	11	1	1.1	0.9;
	12278	2	0	0	0	0	0	1.047117	89.983578	20	3	1.1	0.9;
	12279	2	0	0	0	0	0	1.047117	89.983578	20	3	1.1	0.9;
	12280	2	0	0	0	0	0	1.042396	85.13469	20	3	1.1	0.9;
	12281	2	0	0	0	0	0	1.047117	89.983578	20	3	1.1	0.9;
	12282	2	0	0	0	0	0	1.028283	16.586422	20	14	1.1	0.9;
	12283	1	-0	0	0	0.3	0	1.013768	-6.885216	110	23	1.1	0.9;
	12284	1	29.75	27.1	0.077948	5.312052	0	1.034072	-10.492546	150	9	1.1	0.9;
	12285	2	0	0	0	0	0	0.991765	36.62363	6.3	1	1.1	0.9;
	12286	2	0	0	0	0	0	0.959658	41.263569	10.6	1	1.1	0.9;
	12287	2	0	0	0	0	0	0.993125	29.628125	15	1	1.1	0.9;
	12288	1	133.7	19	0	4.05	0	1.002315	11.835726	154	1	1.1	0.9;
	12289	1	74.18	14.6	0	11.71	0	1.022965	37.031937	110	8	1.1	0.9;
	12290	1	44.05	14.4	0	0.92	0	0.947278	-8.004469	110	21	1.1	0.9;
	12291	2	0	0	0	0	0	0.990464	21.09396	15.75	1	1.1	0.9;
	12292	1	12.6	1.8	0	10.9	0	1.024657	15.946191	154	1	1.1	0.9;
	12293	1	19.2	9.3	0	128.079082	0	1.115485	1.405788	380	9	1.2	0.8;
	12294	1	80.29	-4.1	0	0.43	0	1.016064	32.747434	220	3	1.1	0.9;
	12295	1	29.83	14.4	0	8.89	0	1.050283	28.393059	220	1	1.1	0.9;
	12296	1	19.4	2.8	0	1.61	0	1.012516	20.730823	154	1	1.1	0.9;
	12297	1	2.5	1.3	0	0.62	0	0.988782	-2.439471	110	18	1.1	0.9;
	12298	1	0	0	0.12	8.51	0	0.980116	-4.823997	220	21	1.1	0.9;
	12299	1	192.1	33.4	0.039803	6.200197	0	1.037269	23.086136	220	5	1.1	0.9;
	12300	1	191.2	16.4	0	17.99	0	1.001772	-12.11369	380	4	1.1	0.9;
	12301	1	14.86	2.99	0	1.24	0	1.014152	3.583716	110	18	1.1	0.9;
	12302	2	0	0	0	0	0	1.028425	-14.940962	20	6	1.1	0.9;
	12303	2	0	0	0	0	0	1.028425	-14.940962	20	6	1.1	0.9;
	12304	2	0	0	0	0	0	1.028425	-14.940962	20	6	1.1	0.9;
	12305	1	155.1	39.9	0	6.37	0	1.025731	-28.633876	220	4	1.1	0.9;
	12306	2	0	0	0	0	0	1.026363	2.051273	20	6	1.1	0.9;
	12307	2	0	0	0	0	0	1.028425	-14.940962	20	6	1.1	0.9;
	12308	1	0	0	0	3.77	0	1.018396	3.799696	220	1	1.1	0.9;
	12309	1	0	-0	0	8.82	0	1.011255	29.949849	380	3	1.1	0.9;
	12310	2	0	0	0	0	0	1.054937	4.624127	20	20	1.1	0.9;
	12311	1	-0	0	0	0.72	0	0.991283	25.812593	150	8	1.1	0.9;
	12312	1	50.96	17	0	2.66	0	1.00568	-6.75619	110	21	1.1	0.9;
	12313	1	-0	0	0.135493	21.11062	0	1.070945	1.199007	380	21	1.2	0.8;
	12314	1	2.9	1	0	0.39	0	1.022969	-2.790609	110	19	1.1	0.9;
	12315	1	27.9	4	0	2.23	0	1.008747	26.26701	154	1	1.1	0.9;
	12316	1	2.8	1	0	0.74	0	0.998315	-1.39854	110	19	1.1	0.9;
	12317	1	0	0	0	0	0	1.019678	-23.024367	20	4	1.1	0.9;
	12318	1	-0	-0	0	3.97	0	1.052137	7.420861	220	1	1.1	0.9;
	12319	2	0	0	0	0	0	1.021414	-20.989235	20	4	1.1	0.9;
	12320	1	126.54	29.1	0	12.55	0	1.022961	3.336952	220	5	1.1	0.9;
	12321	2	0	0	0	0	0	1.050901	18.201002	20	7	1.1	0.9;
	12322	1	-0	-0	0.213044	4.836956	0	1.027056	17.711863	220	7	1.1	0.9;
	12323	1	-0	0	0	0	0	1.021419	-22.29244	20	4	1.1	0.9;
	12324	1	6.31	2.5	0	0.16	0	0.986522	-4.034109	110	18	1.1	0.9;
	12325	2	0	0	0	0	0	1.021414	-20.989235	20	4	1.1	0.9;
	12326	1	12.01	4	0	0.72	0	0.935752	-10.35524	110	21	1.1	0.9;
	12327	1	69.7	9.9	0	6.07	0	1.006613	17.383224	154	1	1.1	0.9;
	12328	1	66.4	9.4	0	4.35	0	1.016692	13.997892	154	1	1.1	0.9;
	12329	1	77.99	19.3	0	3.04	0	1.025747	22.553164	220	5	1.1	0.9;
	12330	2	0	0	0	0	0	1.030356	27.865848	20	6	1.1	0.9;
	12331	1	0	0	0	1.23	0	1.001642	19.795922	220	5	1.1	0.9;
	12332	2	0	0	0	0	0	1.030356	27.865848	20	6	1.1	0.9;
	12333	2	0	0	0	0	0	1.030356	27.865848	20	6	1.1	0.9;
	12334	2	0	0	0	0	0	1.030356	27.865848	20	6	1.1	0.9;
	12335	1	81.19	23.4	0	52.6	0	1.057014	60.122944	380	3	1.1	0.9;
	12336	1	-0	-0	0	42.32	0	1.102395	0.903089	380	1	1.2	0.8;
	12337	2	0	0	0	0	0	1.030375	28.469876	20	6	1.1	0.9;
	12338	2	0	0	0	0	0	1.030375	28.469876	20	6	1.1	0.9;
	12339	1	0	0	0.223313	34.22321	0	1.078678	-4.744253	380	12	1.2	0.8;
	12340	1	14.72	5.3	0	0.91	0	1.003295	-3.652509	110	19	1.1	0.9;
	12341	2	0	0	0	0	0	1.034048	-9.733404	20	4	1.1	0.9;
	12342	1	0	-0	0	0	0	1.012736	-19.855402	20	4	1.1	0.9;
	12343	1	216.3	-6.9	0	2.44	0	1.026073	-20.572924	220	5	1.1	0.9;
	12344	1	15.02	7	0	0.67	0	1.023729	-5.215442	110	23	1.1	0.9;
	12345	1	0	-0	0.318467	8.72115	0	1.008569	-7.665426	380	6	1.1	0.9;
	12346	2	0	0	0	0	0	1.059619	6.357453	20	18	1.1	0.9;
	12347	1	-0	-0	0	33.61	0	1.066503	7.436668	380	1	1.1	0.9;
	12348	2	0	0	0	0	0	1.057338	4.561597	20	18	1.1	0.9;
	12349	1	60.87	5	0.176606	-0.077172	0	1.005646	-3.164565	110	18	1.1	0.9;
	12350	1	0	0	0.019758	4.380242	0	1.010129	-10.716048	220	4	1.1	0.9;
	12351	1	0	0	0	2.28	0	0.958855	-9.877767	150	9	1.1	0.9;
	12352	1	-0	-0	0	0.22	0	1.018483	51.345355	110	8	1.1	0.9;
	12353	1	0	-0	0	17.92	0	1.04731	-17.410105	380	1	1.1	0.9;
	12354	1	0	-0	0	1.32	0	0.986477	-4.89785	110	21	1.1	0.9;
	12355	1	0	-0	0.310722	0.351065	0	0.989862	18.147647	220	6	1.1	0.9;
	12356	1	60.2	8.6	0	1.79	0	0.99277	18.046203	154	1	1.1	0.9;
	12357	1	-0	-0	0	61.61	0	1.047215	2.332733	154	1	1.2	0.8;
	12358	2	0	0	0	0	0	1.016383	11.964626	11	1	1.1	0.9;
	12359	2	0	0	0	0	0	1.00107	32.866746	11	1	1.1	0.9;
	12360	1	231.2	103.5	0	10.91	0	1.007927	11.058227	220	3	1.1	0.9;
	12361	1	115.03	2.21	0	10.42	0	1.015736	34.286404	220	13	1.1	0.9;
	12362	1	96.8	76.1	0.245933	14.261622	0	0.999618	53.047091	380	3	1.1	0.9;
	12363	1	213.6	50.1	0	0	0	0.994361	45.601994	220	3	1.1	0.9;
	12364	1	286.3	94.1	0.468335	42.319235	0	1.025692	38.011993	380	13	1.1	0.9;
	12365	1	107.9	15.3	0	99.52	0	0.995947	2.30069	400	1	1.1	0.9;
	12366	1	4.2	0.6	0	2.04	0	1.006419	29.654203	154	1	1.1	0.9;
	12367	2	0	0	0	0	0	0.981513	33.636923	15.75	1	1.1	0.9;
	12368	1	100.8	32.3	0	7.16	0	1.023184	33.690697	150	8	1.1	0.9;
	12369	1	-0	-0	0	37.57	0	1.013904	44.840594	380	8	1.1	0.9;
	12370	1	70.88	43.9	0	3.65	0	1.002425	-9.120945	150	9	1.1	0.9;
	12371	1	33.74	7.4	0.039535	0.630465	0	1.004982	5.145557	110	18	1.1	0.9;
	12372	1	58.26	16.9	0	0.95	0	1.023157	55.244731	110	8	1.1	0.9;
	12373	2	0	0	0	0	0	1.035914	35.115322	20	10	1.1	0.9;
	12374	1	8.21	2.7	0	0.35	0	0.989849	-10.807776	110	12	1.1	0.9;
	12375	2	0	0	0	0	0	1.007198	24.820792	11	1	1.1	0.9;
	12376	1	-0	-0	0	4.05	0	1.035137	3.669478	220	5	1.1	0.9;
	12377	1	18.42	7	0.038628	0.661372	0	1.018693	-8.555675	110	12	1.1	0.9;
	12378	1	79.39	24.7	0	198.67	0	1.027094	-8.230817	380	6	1.1	0.9;
	12379	1	15.82	9.5	0	0.02	0	1.044824	-6.793188	150	9	1.1	0.9;
	12380	1	0	0	0	0.54	0	0.986027	-2.287209	110	18	1.1	0.9;
	12381	2	0	0	0	0	0	1.021475	28.018024	20	10	1.1	0.9;
	12382	1	0	0	0	0.28	0	0.994581	-7.684207	110	18	1.1	0.9;
	12383	2	0	0	0	0	0	1.017202	25.964665	11.5	1	1.1	0.9;
	12384	1	101.9	20	0	22.64	0	1.033833	-19.153366	220	4	1.1	0.9;
	12385	1	40.75	12.5	0	2.61	0	1.027912	18.761456	220	5	1.1	0.9;
	12386	1	0	-0	0	14.38	0	1.038916	4.735605	220	5	1.1	0.9;
	12387	2	0	0	0	0	0	1.031726	-2.251951	20	18	1.1	0.9;
	12388	2	0	0	0	0	0	1.017464	-4.981811	20	10	1.1	0.9;
	12389	1	64.57	23.1	0	1.88	0	0.985038	-5.07072	110	19	1.1	0.9;
	12390	2	0	0	0	0	0	1.039621	10.691121	20	18	1.1	0.9;
	12391	1	-0	-0	0	19.08	0	1.060227	-11.884948	150	9	1.1	0.9;
	12392	1	2.8	1	0	0.56	0	1.014056	-1.175376	110	19	1.1	0.9;
	12393	2	0	0	0	0	0	1.090507	11.169589	10.5	1	1.2	0.8;
	12394	1	33.84	11.1	0	1.57	0	1.004203	-9.712444	110	12	1.1	0.9;
	12395	1	-0	0	0.233802	1.535165	0	1.026656	-6.600104	380	6	1.1	0.9;
	12396	1	-0.24	-25.61	0	11.32	0	1.137371	48.602264	220	3	1.2	0.8;
	12397	1	169.9	41.9	0.039317	0.820683	0	1.033675	9.664326	220	5	1.1	0.9;
	12398	2	0	0	0	0	0	0.988831	38.418014	10.5	1	1.1	0.9;
	12399	1	41.05	12.5	0.019459	0.140541	0	1.020782	-1.666242	110	21	1.1	0.9;
	12400	2	0	0	0	0	0	1.024479	-0.687261	20	18	1.1	0.9;
	12401	2	0	0	0	0	0	1.019049	8.433799	11	1	1.1	0.9;
	12402	1	18.52	6.1	0	0.84	0	0.997819	-4.854372	110	21	1.1	0.9;
	12403	1	29.13	9.6	0	0.76	0	1.000872	-5.558833	110	21	1.1	0.9;
	12404	1	11.01	3.3	0	0.67	0	0.995006	-14.28861	110	22	1.1	0.9;
	12405	1	-0	-0	0.168001	0.399333	0	1.051918	5.627699	220	21	1.1	0.9;
	12406	1	0	-0	0.056554	12.862872	0	1.068044	88.930216	380	3	1.1	0.9;
	12407	1	188.2	23.4	0	17.95	0	1.057605	8.235054	220	3	1.1	0.9;
	12408	1	68.6	9.8	0	117.06	0	0.976355	30.641085	400	1	1.1	0.9;
	12409	1	19.72	6.5	0	0.2	0	0.994272	-4.708058	110	21	1.1	0.9;
	12410	1	0	-0	0	1.15	0	0.958773	0.244074	110	18	1.1	0.9;
	12411	2	0	0	0	0	0	1.035207	-6.870994	20	4	1.1	0.9;
	12412	1	-0	0	0	2.11	0	1.040313	28.796962	220	5	1.1	0.9;
	12413	1	-0	-0	0.037576	3.542424	0	1.053914	4.841595	220	18	1.1	0.9;
	12414	1	-0	0	0	-0.17	0	0.988935	-26.085217	220	6	1.1	0.9;
	12415	2	0	0	0	0	0	1.028414	-6.287287	20	4	1.1	0.9;
	12416	1	0	0	0	1.67	0	1.029697	16.341634	220	1	1.1	0.9;
	12417	1	0	0	0	6.501239	0	0.984319	14.271938	380	5	1.1	0.9;
	12418	1	15.12	5	0	0.53	0	0.961184	-6.931245	110	21	1.1	0.9;
	12419	2	0	0	0	0	0	1.013208	-8.664527	20	10	1.1	0.9;
	12420	1	50.2	7.1	0	1.78	0	1.044939	2.648844	154	1	1.1	0.9;
	12421	2	0	0	0	0	0	0.981334	37.783033	6.3	1	1.1	0.9;
	12422	1	-0.19	-17.71	0	28.98	0	1.020663	37.928025	380	8	1.1	0.9;
	12423	1	374.4	85.5	0.458551	99.79698	0	1.053522	-4.967752	380	11	1.1	0.9;
	12424	1	133	18.9	0	18.79	0	1.001285	17.521015	154	1	1.1	0.9;
	12425	1	-0	0	0	0.1	0	1.001641	-9.981403	150	9	1.1	0.9;
	12426	1	8.81	5.5	0	2.83	0	1.012396	-9.882591	150	9	1.1	0.9;
	12427	1	35.04	12	0	1.07	0	0.951436	-8.337769	110	21	1.1	0.9;
	12428	2	0	0	0	0	0	1.027725	-17.372828	20	4	1.1	0.9;
	12429	2	0	0	0	0	0	1.027725	-17.372828	20	4	1.1	0.9;
	12430	2	0	0	0	0	0	1.027725	-17.372828	20	4	1.1	0.9;
	12431	2	0	0	0	0	0	1.027725	-17.372828	20	4	1.1	0.9;
	12432	1	-0	-0	0.343783	27.652613	0	1.063824	4.899786	380	20	1.1	0.9;
	12433	2	0	0	0	0	0	1.027725	-17.372828	20	4	1.1	0.9;
	12434	2	0	0	0	0	0	1.013219	-5.540854	20	4	1.1	0.9;
	12435	1	380.8	62.8	0	7.11	0	1.033007	6.963586	220	5	1.1	0.9;
	12436	1	59.07	23.4	0	21.88	0	1.022086	26.499257	150	8	1.1	0.9;
	12437	1	139.3	19.9	0	0.56	0	1.019936	14.700597	154	1	1.1	0.9;
	12438	1	0	0	0	0.01	0	1.008454	28.268379	154	1	1.1	0.9;
	12439	1	31.44	12.4	0	2.89	0	1.033316	30.22478	150	8	1.1	0.9;
	12440	2	0	0	0	0	0	1.043738	-8.975119	20	4	1.1	0.9;
	12441	2	0	0	0	0	0	1.004719	2.040837	11	1	1.1	0.9;
	12442	1	332.8	92.2	0	29.46	0	0.98672	-22.441078	220	6	1.1	0.9;
	12443	1	0	-0	0	0.66	0	0.967037	-8.252401	150	9	1.1	0.9;
	12444	1	-7.6	28.6	0	0.19	0	1.00773	0.837784	150	9	1.1	0.9;
	12445	1	33.64	4.5	0	0.4	0	1.058596	0.438284	110	19	1.1	0.9;
	12446	1	-0	0	0	1.1	0	1.041946	-1.815076	110	15	1.1	0.9;
	12447	2	0	0	0	0	0	1.028667	-4.741858	20	6	1.1	0.9;
	12448	2	0	0	0	0	0	1.036966	15.051213	20	14	1.1	0.9;
	12449	1	286.1	89.3	0	8.89	0	1.017949	10.152132	220	5	1.1	0.9;
	12450	1	111.08	-44.35	0	7.94	0	1.025585	34.455378	220	13	1.1	0.9;
	12451	2	0	0	0	0	0	1.024669	12.739456	20	18	1.1	0.9;
	12452	2	0	0	0	0	0	1.035311	42.433038	20	8	1.1	0.9;
	12453	2	0	0	0	0	0	1.035255	42.395082	20	8	1.1	0.9;
	12454	2	0	0	0	0	0	1.016573	45.994905	20	8	1.1	0.9;
	12455	2	0	0	0	0	0	1.028667	-4.741858	20	6	1.1	0.9;
	12456	1	-26.03	-0.84	0	0.1	0	1.058626	17.900462	220	5	1.1	0.9;
	12457	1	0	-0	0.039626	0.610374	0	1.014358	41.933926	220	3	1.1	0.9;
	12458	1	-0	-0	0	0.85	0	1.01073	-8.580613	150	9	1.1	0.9;
	12459	2	0	0	0	0	0	1.04826	-8.707987	20	10	1.1	0.9;
	12460	1	35.84	10.9	0	0.11	0	1.062485	0.136857	110	15	1.1	0.9;
	12461	2	0	0	0	0	0	1.057713	44.127182	20	3	1.1	0.9;
	12462	2	0	0	0	0	0	1.057713	44.127182	20	3	1.1	0.9;
	12463	1	13.91	10.1	0	0.31	0	0.971587	5.047365	110	18	1.1	0.9;
	12464	2	0	0	0	0	0	1.057713	44.127182	20	3	1.1	0.9;
	12465	2	0	0	0	0	0	1.05445	48.75455	20	3	1.1	0.9;
	12466	2	0	0	0	0	0	1.057713	44.127182	20	3	1.1	0.9;
	12467	2	0	0	0	0	0	1.057713	44.127182	20	3	1.1	0.9;
	12468	1	24.22	-1.06	0	2.92	0	0.991937	88.98538	150	1	1.1	0.9;
	12469	2	0	0	0	0	0	1.05445	48.75455	20	3	1.1	0.9;
	12470	2	0	0	0	0	0	1.05445	48.75455	20	3	1.1	0.9;
	12471	1	0	0	0	0	0	1.033422	-6.60149	150	9	1.1	0.9;
	12472	1	-0	-0	0.582185	0.836226	0	1.040643	19.952471	380	5	1.1	0.9;
	12473	1	55.56	34.4	0	2.57	0	1.039652	-6.424256	150	9	1.1	0.9;
	12474	1	277.9	143.1	0	0.69	0	1.000416	28.059075	380	3	1.1	0.9;
	12475	1	0	0	0.332077	31.17749	0	1.009859	39.894462	380	3	1.1	0.9;
	12476	1	-0	0	0.550522	8.61491	0	1.01083	28.808468	380	5	1.1	0.9;
	12477	2	0	0	0	0	0	1.037214	32.974337	20	10	1.1	0.9;
	12478	1	51.06	7.4	0	2.49	0	1.012634	-19.834858	220	4	1.1	0.9;
	12479	1	0	-0	0	2.02	0	0.983018	-8.488121	150	9	1.1	0.9;
	12480	1	-0.59	-0.86	0	0.14	0	1.015791	-1.704574	150	9	1.1	0.9;
	12481	1	24.13	10.8	0	15.2	0	1.014445	32.429678	150	8	1.1	0.9;
	12482	1	-0	-0	0.112964	-0.054136	0	1.077814	0.7293	110	15	1.1	0.9;
	12483	1	38.84	8.5	0	7.31	0	1.056191	25.665794	220	5	1.1	0.9;
	12484	1	0	-0	0	1.97	0	1.046801	25.041884	220	5	1.1	0.9;
	12485	1	216.2	75.3	0	2.74	0	0.983182	7.024133	220	5	1.1	0.9;
	12486	1	-0	0	0	19.54	0	0.994915	24.02232	400	1	1.1	0.9;
	12487	1	0	-0	0.073728	5.986272	0	1.024241	-6.779685	220	10	1.1	0.9;
	12488	1	0	0	0.208917	0.724607	0	1.12345	-0.14804	380	9	1.2	0.8;
	12489	1	0	-0	0	2.78	0	0.992433	-3.280361	220	21	1.1	0.9;
	12490	1	-0	-0	0	0.73	0	1.023616	-1.513773	110	21	1.1	0.9;
	12491	1	19.4	2.8	0	3.63	0	1.033243	1.205486	154	1	1.1	0.9;
	12492	1	299.5	48.7	0.098259	18.211741	0	1.028436	20.306678	220	5	1.1	0.9;
	12493	1	55.96	30	0.438961	55.322489	0	1.079318	2.779714	380	21	1.2	0.8;
	12494	2	0	0	0	0	0	1.01993	9.094543	20	22	1.1	0.9;
	12495	1	0	0	0	8.35	0	1.0268	-20.255767	220	5	1.1	0.9;
	12496	1	10.5	7.32	0	0.24	0	0.972421	7.134096	110	18	1.1	0.9;
	12497	1	0	-0	0	2.46	0	1.030865	18.670476	220	5	1.1	0.9;
	12498	1	-0	-0	0	3.01	0	1.071574	22.564331	220	5	1.1	0.9;
	12499	2	0	0	0	0	0	1.040719	26.306193	11	1	1.1	0.9;
	12500	2	0	0	0	0	0	1.050684	91.443424	20	3	1.1	0.9;
	12501	1	-0	-0	0	1.15	0	1.032889	46.113563	110	8	1.1	0.9;
	12502	1	18.02	5.9	0.059068	2.260932	0	1.004154	-3.879517	110	21	1.1	0.9;
	12503	1	-0	-0	0	0.42	0	0.985206	1.958171	110	18	1.1	0.9;
	12504	1	-0.17	-1.82	0	0	0	1.050851	44.081079	380	8	1.1	0.9;
	12505	2	0	0	0	0	0	1.029991	3.188016	20	6	1.1	0.9;
	12506	2	0	0	0	0	0	1.029991	3.188016	20	6	1.1	0.9;
	12507	2	0	0	0	0	0	1.015733	2.857525	0.4	1	1.1	0.9;
	12508	1	0	-0	0	0.13	0	1.001199	-2.907812	220	6	1.1	0.9;
	12509	2	0	0	0	0	0	1.060659	20.876727	20	5	1.1	0.9;
	12510	1	36.54	22.6	0	0.04	0	1.043805	-6.740275	150	9	1.1	0.9;
	12511	1	-0	0	0.040091	29.754956	0	1.032077	9.94197	380	5	1.1	0.9;
	12512	1	44.45	7.68	0	17.83	0	1.003067	26.71029	380	5	1.1	0.9;
	12513	1	-0	-0	0.047978	5.881618	0	1.049549	58.720219	220	8	1.1	0.9;
	12514	1	371.6	100.4	0.231295	3.706891	0	1.033467	32.233232	220	13	1.1	0.9;
	12515	1	0	0	0	21.62	0	1.05922	-11.863331	150	9	1.1	0.9;
	12516	1	0	0	0.087808	3.044992	0	1.02408	7.904971	220	18	1.1	0.9;
	12517	2	0	0	0	0	0	1.024254	10.943171	20	6	1.1	0.9;
	12518	2	0	0	0	0	0	1.028417	-6.105187	20	6	1.1	0.9;
	12519	2	0	0	0	0	0	1.057091	77.565872	20	3	1.1	0.9;
	12520	1	-0.19	-1	0	0	0	1.021469	26.422336	150	8	1.1	0.9;
	12521	2	0	0	0	0	0	1.022689	-16.891323	20	4	1.1	0.9;
	12522	2	0	0	0	0	0	1.022689	-16.891323	20	4	1.1	0.9;
	12523	2	0	0	0	0	0	1.022689	-16.891323	20	4	1.1	0.9;
	12524	1	0	0	0	1.35	0	0.983198	-2.811042	110	18	1.1	0.9;
	12525	2	0	0	0	0	0	1.034235	2.214597	20	6	1.1	0.9;
	12526	2	0	0	0	0	0	1.034235	2.214597	20	6	1.1	0.9;
	12527	1	98.6	27.1	0	42.45	0	1.024706	22.682557	150	8	1.1	0.9;
	12528	2	0	0	0	0	0	1.040642	8.532255	20	14	1.1	0.9;
	12529	2	0	0	0	0	0	1.024151	24.403529	20	5	1.1	0.9;
	12530	1	0	0	0	13.01	0	0.963945	-9.066656	220	10	1.1	0.9;
	12531	1	235.97	34.52	0	7.15	0	1.062163	20.979236	380	5	1.1	0.9;
	12532	1	23.93	7.9	0	0.88	0	0.947189	-8.737106	110	21	1.1	0.9;
	12533	1	9.51	3.4	0	0.73	0	1.044117	0.887619	110	15	1.1	0.9;
	12534	1	-0	0	0.123474	4.29472	0	1.042522	-4.869386	220	1	1.1	0.9;
	12535	1	38.84	14.2	0	3.5	0	1.040037	14.871755	220	5	1.1	0.9;
	12536	1	236.44	-98.55	0	1.31	0	1.04179	74.863317	220	3	1.1	0.9;
	12537	2	0	0	0	0	0	1.035364	12.970449	20	18	1.1	0.9;
	12538	1	-14.89	-4.43	0	0.63	0	0.992161	2.241109	110	18	1.1	0.9;
	12539	1	0	0	0	15.91	0	1.006626	-6.636873	380	1	1.1	0.9;
	12540	1	10.11	0	0	1.92	0	0.942721	-7.656569	110	18	1.1	0.9;
	12541	2	0	0	0	0	0	1.02993	1.561794	11	1	1.1	0.9;
	12542	2	0	0	0	0	0	1.028438	-19.446471	20	4	1.1	0.9;
	12543	1	97.6	33.9	0	6.47	0	1.012068	24.123921	220	3	1.1	0.9;
	12544	1	6.41	2.2	0	52.07	0	1.051432	-10.828911	380	4	1.1	0.9;
	12545	1	-0.59	-2.63	0	0.14	0	0.964406	-4.211993	110	19	1.1	0.9;
	12546	1	-0	0	0	0.67	0	0.985531	-4.991115	150	9	1.1	0.9;
	12547	1	37.44	9.1	0	7.87	0	1.023772	3.420383	220	5	1.1	0.9;
	12548	1	-8.02	-15.14	0	0	0	1.033735	42.729299	150	8	1.1	0.9;
	12549	1	-0	-0	0	3.76	0	1.015104	-0.273623	220	6	1.1	0.9;
	12550	1	0	-0	0	8.78	0	1.019828	44.646555	220	8	1.1	0.9;
	12551	2	0	0	0	0	0	1.034631	4.683321	20	14	1.1	0.9;
	12552	1	93.6	-43.5	0	33.51	0	1.027627	5.677018	220	5	1.1	0.9;
	12553	1	140.7	10.4	0	2.51	0	1.012708	-19.712251	220	4	1.1	0.9;
	12554	1	13.51	5.6	0	0.86	0	1.020861	-3.956799	110	15	1.1	0.9;
	12555	2	0	0	0	0	0	1.003834	31.873789	11	1	1.1	0.9;
	12556	2	0	0	0	0	0	1.029835	-14.681025	20	4	1.1	0.9;
	12557	1	8.3	1.2	0	2.86	0	0.992801	39.577248	154	1	1.1	0.9;
	12558	2	0	0	0	0	0	0.971908	40.431219	11	1	1.1	0.9;
	12559	1	101.7	59.8	0.098539	36.661461	0	1.04358	17.126486	220	5	1.1	0.9;
	12560	1	104.3	12.3	0	8.29	0	1.032136	33.351366	220	3	1.1	0.9;
	12561	1	89.08	26.9	0	4.68	0	1.02606	16.511466	220	5	1.1	0.9;
	12562	1	0	-0	0	0	0	1.017664	-15.016249	20	4	1.1	0.9;
	12563	2	0	0	0	0	0	1.020396	12.801356	20	18	1.1	0.9;
	12564	1	-0	-0	0	1.82	0	1.001293	17.356053	154	1	1.1	0.9;
	12565	1	13.72	8.5	0	0.06	0	1.057526	-6.024854	150	9	1.1	0.9;
	12566	1	0	0	0	4.5	0	1.030112	17.57744	220	3	1.1	0.9;
	12567	1	-0	-0	0	5.98	0	1.032903	-3.018826	220	21	1.1	0.9;
	12568	1	84	19.4	0.010011	4.66	0	0.990229	11.73922	220	5	1.1	0.9;
	12569	1	20.9	3	0	7.77	0	1.011983	20.514966	154	1	1.1	0.9;
	12570	2	0	0	0	0	0	1.022447	17.861755	20	10	1.1	0.9;
	12571	1	0	-3.7	0	0.25	0	1.021812	59.451188	110	8	1.1	0.9;
	12572	2	0	0	0	0	0	0.989707	35.937223	11	1	1.1	0.9;
	12573	1	0	0	0	45.57454	0	1.060539	4.452559	380	20	1.1	0.9;
	12574	1	0	0	0	3.32	0	0.976396	-27.177556	220	6	1.1	0.9;
	12575	1	0	0	0	0	0	1.04035	8.357264	20	5	1.1	0.9;
	12576	1	0	0	0	0	0	1.04035	8.357264	20	5	1.1	0.9;
	12577	1	76.99	-0.7	0	1.06	0	1.017564	-19.177951	220	4	1.1	0.9;
	12578	1	0	0	0	0	0	1.04035	8.357264	20	5	1.1	0.9;
	12579	2	0	0	0	0	0	0.960771	44.51306	21	1	1.1	0.9;
	12580	1	0	0	0	0	0	1.04035	8.357264	20	5	1.1	0.9;
	12581	1	0	0	0	0	0	1.04035	8.357264	20	5	1.1	0.9;
	12582	2	0	0	0	0	0	1.049495	12.649819	20	5	1.1	0.9;
	12583	2	0	0	0	0	0	1.049495	12.649819	20	5	1.1	0.9;
	12584	1	0	0	0.117152	0.132373	0	1.025973	5.996733	220	20	1.1	0.9;
	12585	1	0	0	0	0	0	1.04035	8.357264	20	5	1.1	0.9;
	12586	2	0	0	0	0	0	1.058287	5.521142	20	5	1.1	0.9;
	12587	2	0	0	0	0	0	1.016947	23.125552	380	6	1.1	0.9;
	12588	1	12.31	4.1	0	0.32	0	0.980901	1.580971	110	18	1.1	0.9;
	12589	2	0	0	0	0	0	1.058287	5.521142	20	5	1.1	0.9;
	12590	2	0	0	0	0	0	1.058287	5.521142	20	5	1.1	0.9;
	12591	2	0	0	0	0	0	1.058287	5.521142	20	5	1.1	0.9;
	12592	2	0	0	0	0	0	1.058287	5.521142	20	5	1.1	0.9;
	12593	1	0	0	0	0.31	0	0.991603	-12.182028	110	1	1.1	0.9;
	12594	1	81.89	19.9	0	5.1	0	1.029513	3.924035	220	5	1.1	0.9;
	12595	1	72.68	24.3	0	3.12	0	1.031042	8.597105	220	5	1.1	0.9;
	12596	1	-0	-0	0	0.85	0	0.975425	9.544549	110	18	1.1	0.9;
	12597	1	187.64	31.94	0	4.75	0	1.03197	34.05665	220	13	1.1	0.9;
	12598	1	92.4	6.3	0	4.32	0	0.99025	-4.371984	220	6	1.1	0.9;
	12599	2	0	0	0	0	0	1.033945	-7.536483	20	22	1.1	0.9;
	12600	2	0	0	0	0	0	1.033945	-7.536483	20	22	1.1	0.9;
	12601	2	0	0	0	0	0	1.033945	-7.536483	20	22	1.1	0.9;
	12602	1	0	0	0	0	0	1.012421	-10.166509	150	9	1.1	0.9;
	12603	2	0	0	0	0	0	0.947933	57.203162	11	1	1.1	0.9;
	12604	1	0	0	0	0.64	0	0.944619	-7.941005	110	21	1.1	0.9;
	12605	1	111.4	15.9	0	19.05	0	1.02557	17.194572	154	1	1.1	0.9;
	12606	2	0	0	0	0	0	1.020332	23.152992	20	10	1.1	0.9;
	12607	2	0	0	0	0	0	1.026221	1.816827	20	6	1.1	0.9;
	12608	1	10	1.4	0	5.74	0	1.011671	23.560034	154	1	1.1	0.9;
	12609	1	20.9	3	0	3.48	0	1.003783	14.429739	154	1	1.1	0.9;
	12610	2	0	0	0	0	0	1.026221	1.816827	20	6	1.1	0.9;
	12611	1	5.01	1.8	0	0.77	0	1.026964	17.858001	220	5	1.1	0.9;
	12612	2	0	0	0	0	0	1.030988	2.394145	20	9	1.1	0.9;
	12613	1	21.22	9.9	0	4.64	0	0.99345	-24.884337	220	4	1.1	0.9;
	12614	1	-0	0	0	15	0	1.045898	82.995161	220	3	1.1	0.9;
	12615	1	-0	-0	0	0.63	0	0.993712	-1.750779	110	19	1.1	0.9;
	12616	2	0	0	0	0	0	0.996117	32.042491	11	1	1.1	0.9;
	12617	1	-0	-0	0	1.42	0	0.979247	-8.554353	150	9	1.1	0.9;
	12618	1	0	0	0.176855	3.419803	0	0.96442	-0.202797	220	19	1.1	0.9;
	12619	1	0	0	0	0	0	0.996198	-5.152317	150	9	1.1	0.9;
	12620	1	60.9	8.7	0	1.34	0	1.010475	30.420994	154	1	1.1	0.9;
	12621	1	562.3	159.1	0.245655	7.565523	0	1.006881	26.429157	220	14	1.1	0.9;
	12622	1	0	0	2.226096	44.76755	0	1.074567	88.264357	380	3	1.1	0.9;
	12623	1	0	-0	0	2.64	0	1.001746	17.611273	154	1	1.1	0.9;
	12624	1	22.02	7.7	0	0.17	0	1.018821	-14.589104	110	22	1.1	0.9;
	12625	2	0	0	0	0	0	0.948759	58.410078	14	1	1.1	0.9;
	12626	1	214.8	30.6	0	3.65	0	1.042524	1.0069	154	1	1.1	0.9;
	12627	1	221.9	86.3	0	2.18	0	1.003651	31.884957	150	8	1.1	0.9;
	12628	1	0	0	0	0	0	0.964149	-15.086645	150	9	1.1	0.9;
	12629	1	0	0	0	18.22	0	1.024655	7.233326	380	1	1.1	0.9;
	12630	1	-0	0	0	0.88	0	1.055954	27.990261	220	1	1.1	0.9;
	12631	1	0	0	0	237.54	0	1.048783	-15.463026	380	4	1.1	0.9;
	12632	1	0	-0	0.379262	96.912184	0	1.079049	-0.263696	380	21	1.1	0.9;
	12633	1	-0	-0	0	7.06	0	1.014079	18.851636	330	24	1.1	0.9;
	12634	1	172.6	24.6	0	11.39	0	1.036563	-1.739082	154	1	1.1	0.9;
	12635	1	1.9	1.3	0	0.92	0	1.012593	3.023648	110	18	1.1	0.9;
	12636	2	0	0	0	0	0	1.007821	10.245998	20	18	1.1	0.9;
	12637	1	15.32	3.5	0	0.69	0	0.998377	6.098094	110	18	1.1	0.9;
	12638	1	38.04	15	0	0.47	0	0.995589	-3.184629	110	21	1.1	0.9;
	12639	1	15.42	4.4	0	10.84	0	0.991849	-13.731605	110	22	1.1	0.9;
	12640	1	0	-0	0.019788	3.740212	0	1.014372	-1.282515	220	19	1.1	0.9;
	12641	2	0	0	0	0	0	0.995435	4.377334	20	9	1.1	0.9;
	12642	1	-18.27	-12.99	0	0	0	1.022369	22.437418	150	8	1.1	0.9;
	12643	1	-0	0	0.235924	0.173397	0	1.023722	-7.333653	380	6	1.1	0.9;
	12644	1	0	0	0	1.71	0	0.962725	-7.715632	110	21	1.1	0.9;
	12645	1	40.54	3.6	0	1.88	0	1.026651	-27.256651	220	4	1.1	0.9;
	12646	2	0	0	0	0	0	1.042693	10.439455	20	23	1.1	0.9;
	12647	2	0	0	0	0	0	1.029333	-7.499675	20	6	1.1	0.9;
	12648	2	0	0	0	0	0	1.029333	-7.499675	20	6	1.1	0.9;
	12649	1	13.92	6.9	0	0.79	0	0.990644	-1.876429	110	18	1.1	0.9;
	12650	2	0	0	0	0	0	1.029333	-7.499675	20	6	1.1	0.9;
	12651	2	0	0	0	0	0	1.029323	-7.468239	20	6	1.1	0.9;
	12652	2	0	0	0	0	0	1.029323	-7.468239	20	6	1.1	0.9;
	12653	2	0	0	0	0	0	1.029333	-7.499675	20	6	1.1	0.9;
	12654	2	0	0	0	0	0	1.029323	-7.468239	20	6	1.1	0.9;
	12655	1	0	0.01	0	22.04	0	1.004761	-16.59957	380	6	1.1	0.9;
	12656	2	0	0	0	0	0	1.04659	30.049071	20	5	1.1	0.9;
	12657	2	0	0	0	0	0	1.017738	25.492678	6.3	1	1.1	0.9;
	12658	2	0	0	0	0	0	0.956073	24.685075	13.8	1	1.1	0.9;
	12659	2	0	0	0	0	0	1.040793	26.940223	20	5	1.1	0.9;
	12660	1	72.4	10.3	0	2.54	0	1.016117	19.693513	154	1	1.1	0.9;
	12661	1	59.87	19.9	0	0.32	0	0.993913	-7.711171	110	21	1.1	0.9;
	12662	1	293.2	87.4	0	26.36	0	1.031338	-16.305128	220	5	1.1	0.9;
	12663	1	9.91	3.3	0	0.82	0	0.997924	-10.001384	110	22	1.1	0.9;
	12664	2	0	0	0	0	0	1.037469	11.940478	20	5	1.1	0.9;
	12665	2	0	0	0	0	0	1.038308	18.388307	20	7	1.1	0.9;
	12666	1	0	0	0	72.57192	0	1.075158	0.186	380	20	1.2	0.8;
	12667	1	1.6	0.8	0	0.79	0	0.978996	-5.170122	110	18	1.1	0.9;
	12668	1	27.83	-2.6	0	6.07	0	1.017623	-24.44662	220	4	1.1	0.9;
	12669	1	0	0	0	6.29	0	1.067807	11.076245	220	18	1.1	0.9;
	12670	1	9.7	1.4	0	2.96	0	1.016854	15.686717	154	1	1.1	0.9;
	12671	1	-0	-0	0	8.89	0	1.027688	40.037008	380	3	1.1	0.9;
	12672	1	13.62	5.2	0.03955	0.19045	0	1.003775	-1.133771	110	18	1.1	0.9;
	12673	2	0	0	0	0	0	1.040793	26.940223	20	5	1.1	0.9;
	12674	1	27.53	17	0	0.03	0	1.029478	-2.341076	150	9	1.1	0.9;
	12675	1	317.9	55.4	0	37.61	0	1.004938	-21.014507	380	6	1.1	0.9;
	12676	1	0	0	0	0.51	0	0.974521	-8.282724	150	9	1.1	0.9;
	12677	1	-33.74	11.1	0	1.87	0	1.041687	-2.946438	150	9	1.1	0.9;
	12678	2	0	0	0	0	0	1.090507	11.169589	10.5	1	1.2	0.8;
	12679	1	-0	-0	0	0	0	1.020254	-23.064206	20	4	1.1	0.9;
	12680	1	12.31	7.4	0.316013	6.664212	0	1.069511	1.473714	150	9	1.1	0.9;
	12681	2	0	0	0	0	0	1.014635	-0.980319	20	18	1.1	0.9;
	12682	1	0	0	0	80.66	0	1.024518	-10.41164	380	1	1.1	0.9;
	12683	1	0	-0	0	0.43	0	1.054144	25.676299	220	5	1.1	0.9;
	12684	1	171.3	13.9	0	59.09	0	1.072691	72.716931	380	3	1.1	0.9;
	12685	1	3.8	3.8	0	0.08	0	0.954262	-5.140695	110	18	1.1	0.9;
	12686	1	72.4	10.3	0	6.82	0	0.996188	29.118836	154	1	1.1	0.9;
	12687	2	0	0	0	0	0	1.01812	15.465461	11	1	1.1	0.9;
	12688	2	0	0	0	0	0	1.001419	42.731033	15.75	1	1.1	0.9;
	12689	1	65.2	9.3	0	3.89	0	1.025956	7.450641	154	1	1.1	0.9;
	12690	2	0	0	0	0	0	1.040793	26.940223	20	5	1.1	0.9;
	12691	1	245.51	9.17	0	24.25	0	0.989794	-3.305032	220	5	1.1	0.9;
	12692	2	0	0	0	0	0	1.015529	-18.474164	20	4	1.1	0.9;
	12693	1	26.93	9.6	0	1.09	0	0.963088	-5.788161	110	19	1.1	0.9;
	12694	1	0	-0	0	4.82	0	1.015506	-2.151015	220	6	1.1	0.9;
	12695	1	0	0	0	1.96	0	1.024295	-24.350547	380	10	1.1	0.9;
	12696	1	-0	-0	0	6.62	0	1.038067	5.796634	220	5	1.1	0.9;
	12697	2	0	0	0	0	0	1.027132	31.963051	20	5	1.1	0.9;
	12698	1	-0	0	0	2.05	0	0.987855	34.356338	220	3	1.1	0.9;
	12699	2	0	0	0	0	0	1.027132	31.963051	20	5	1.1	0.9;
	12700	1	-0	-0	0.117033	-0.117033	0	1.024885	0.182742	150	9	1.1	0.9;
	12701	2	0	0	0	0	0	1.041939	17.231675	20	5	1.1	0.9;
	12702	2	0	0	0	0	0	1.027132	31.963051	20	5	1.1	0.9;
	12703	1	0	0	0.35622	3.44315	0	1.01143	-19.86799	380	6	1.1	0.9;
	12704	2	0	0	0	0	0	1.041939	17.231675	20	5	1.1	0.9;
	12705	2	0	0	0	0	0	1.03975	24.63487	20	5	1.1	0.9;
	12706	2	0	0	0	0	0	1.03975	24.63487	20	5	1.1	0.9;
	12707	1	0	-0	0.347111	22.210024	0	1.02861	33.499831	380	2	1.1	0.9;
	12708	2	0	0	0	0	0	1.017238	-14.795645	20	22	1.1	0.9;
	12709	2	0	0	0	0	0	1.011489	10.204543	20	23	1.1	0.9;
	12710	1	0	0	0	0	0	1.00238	-9.035194	150	9	1.1	0.9;
	12711	2	0	0	0	0	0	1.007801	-12.641343	20	22	1.1	0.9;
	12712	1	0	-0	0	0	0	0.997405	-17.375683	20	22	1.1	0.9;
	12713	2	0	0	0	0	0	1.009886	-8.195055	20	22	1.1	0.9;
	12714	2	0	0	0	0	0	1.017238	-14.795645	20	22	1.1	0.9;
	12715	1	0	-0	0.297442	67.102132	0	1.020382	-2.841001	380	6	1.1	0.9;
	12716	1	347.1	87.2	0	88.35	0	1.030243	11.953288	380	5	1.1	0.9;
	12717	2	0	0	0	0	0	1.043087	11.638661	20	14	1.1	0.9;
	12718	1	55.7	7.9	0	0.76	0	1.01571	22.558348	154	1	1.1	0.9;
	12719	1	77.99	18.5	0.059094	5.220755	0	0.992973	90.962411	150	1	1.1	0.9;
	12720	1	159.4	-7.3	0	3.3	0	1.088271	29.372504	220	3	1.2	0.8;
	12721	1	-0	0	0	12.69	0	0.993038	-9.265022	220	12	1.1	0.9;
	12722	1	101.8	20.2	0	0.91	0	1.000556	26.386651	150	8	1.1	0.9;
	12723	2	0	0	0	0	0	1.015411	-5.476788	20	4	1.1	0.9;
	12724	1	0	-0	0	10.83	0	1.028402	2.843072	220	5	1.1	0.9;
	12725	1	-0	0	0	1.35	0	1.042121	-1.28407	150	9	1.1	0.9;
	12726	2	0	0	0	0	0	1.035338	-7.144697	20	4	1.1	0.9;
	12727	2	0	0	0	0	0	1.029542	-3.096329	20	6	1.1	0.9;
	12728	2	0	0	0	0	0	1.03058	15.553508	20	14	1.1	0.9;
	12729	1	0	-0	0	0.45	0	0.953443	-7.581698	110	21	1.1	0.9;
	12730	2	0	0	0	0	0	1.017293	7.59717	11	1	1.1	0.9;
	12731	1	-0	0	0	5.72	0	0.996162	-2.611738	220	21	1.1	0.9;
	12732	1	-0	0	0	13.3	0	1.013494	12.768835	380	1	1.1	0.9;
	12733	1	76.69	23.3	0	0.3	0	1.029781	18.149388	220	3	1.1	0.9;
	12734	1	-70.45	-35.14	0	0	0	1.021614	43.10077	150	8	1.1	0.9;
	12735	1	-8.4	-17.5	0.058341	1.591659	0	1.008169	-10.617063	110	12	1.1	0.9;
	12736	2	0	0	0	0	0	0.957116	25.356265	13.8	1	1.1	0.9;
	12737	1	-26.03	-2.9	0.157541	15.981844	0	1.016051	5.508506	380	6	1.1	0.9;
	12738	2	0	0	0	0	0	1.027968	24.05088	20	5	1.1	0.9;
	12739	1	0	-0	0	12.68	0	1.034435	9.724651	220	5	1.1	0.9;
	12740	1	0	0	0	13.49	0	0.983226	-12.964461	150	9	1.1	0.9;
	12741	1	-0	-0	0	0	0	1.020254	-23.064206	20	4	1.1	0.9;
	12742	2	0	0	0	0	0	1.035958	27.387404	20	5	1.1	0.9;
	12743	1	-0	0	0.049279	37.690721	0	1.04543	-13.937788	380	4	1.1	0.9;
	12744	1	253.4	111.3	0.415544	10.153713	0	1.049407	3.6767	220	5	1.1	0.9;
	12745	1	136	36.1	0	6.07	0	0.998919	12.925679	220	5	1.1	0.9;
	12746	1	57.56	21.1	0	1.46	0	0.990115	-3.684995	220	6	1.1	0.9;
	12747	2	0	0	0	0	0	1.035958	27.387404	20	5	1.1	0.9;
	12748	2	0	0	0	0	0	0.9846	26.113194	6.3	1	1.1	0.9;
	12749	1	-0	-0	0	4.26	0	1.020995	3.543798	220	5	1.1	0.9;
	12750	1	277.1	37	0	21.03	0	1.031586	6.722493	220	5	1.1	0.9;
	12751	1	98.61	50.3	0	0.17	0	0.961075	-15.738133	150	10	1.1	0.9;
	12752	1	0	-0	0.172712	3.166074	0	1.046972	-19.838782	380	10	1.1	0.9;
	12753	1	-0	0	0	0.54	0	0.927015	-12.042266	150	9	1.1	0.9;
	12754	1	0	-0	0	117.47	0	1.015754	-20.966355	380	4	1.1	0.9;
	12755	1	-0	0	0	2	0	1.053726	-3.038855	150	9	1.1	0.9;
	12756	1	0	0	0.090774	8.200444	0	1.026235	6.371926	220	18	1.1	0.9;
	12757	1	74.18	27.3	0	25.25	0	1.001945	0.575302	220	6	1.1	0.9;
	12758	2	0	0	0	0	0	1.0608	42.720858	20	3	1.1	0.9;
	12759	1	0	0	0	2.64	0	0.991622	29.430959	154	1	1.1	0.9;
	12760	2	0	0	0	0	0	1.028736	-6.917679	20	4	1.1	0.9;
	12761	2	0	0	0	0	0	1.028736	-6.917679	20	4	1.1	0.9;
	12762	1	282	21.5	0	21.3	0	0.996936	-19.146593	380	6	1.1	0.9;
	12763	2	0	0	0	0	0	1.028736	-6.917679	20	4	1.1	0.9;
	12764	2	0	0	0	0	0	1.051947	4.007575	20	2	1.1	0.9;
	12765	2	0	0	0	0	0	1.051947	4.007575	20	2	1.1	0.9;
	12766	1	0	-0	0	1.2	0	1.011954	-10.391865	150	9	1.1	0.9;
	12767	1	-20.32	6.7	0	0.2	0	0.965766	-9.555001	150	9	1.1	0.9;
	12768	1	62.27	23.8	0	0.13	0	1.015423	34.258867	220	3	1.1	0.9;
	12769	1	130.2	88.8	0	0.08	0	1.016798	24.877895	220	3	1.1	0.9;
	12770	1	-0	-0	0.153185	22.892346	0	1.177153	3.211364	380	9	1.2	0.8;
	12771	1	-0	0	0	15.67	0	1.083886	-4.648509	380	1	1.2	0.8;
	12772	2	0	0	0	0	0	1.04706	-6.69158	20	4	1.1	0.9;
	12773	1	163.2	31.9	0	4.9	0	1.0196	34.038724	150	8	1.1	0.9;
	12774	1	29.13	9.5	0	0.98	0	0.98453	-5.62162	110	18	1.1	0.9;
	12775	2	0	0	0	0	0	0.968879	49.23979	11	1	1.1	0.9;
	12776	1	25.18	3.37	0	1.01	0	0.973294	0.169843	110	18	1.1	0.9;
	12777	2	0	0	0	0	0	1.010323	-7.916943	20	4	1.1	0.9;
	12778	1	0	-0	0	0	0	1.038504	-13.052855	20	4	1.1	0.9;
	12779	1	0	-0	0	0	0	1.038504	-13.052855	20	4	1.1	0.9;
	12780	1	0	-0	0	0	0	1.038504	-13.052855	20	4	1.1	0.9;
	12781	1	0	-0	0	0	0	1.038504	-13.052855	20	4	1.1	0.9;
	12782	2	0	0	0	0	0	1.033617	-8.646596	20	4	1.1	0.9;
	12783	1	62.7	9	0	8.23	0	1.005883	16.469663	154	1	1.1	0.9;
	12784	1	42.25	16.6	0	0	0	1.048201	32.736039	150	8	1.1	0.9;
	12785	1	0	0	0	142.05	0	1.043675	20.973559	380	14	1.1	0.9;
	12786	2	0	0	0	0	0	1.03036	-6.512869	20	4	1.1	0.9;
	12787	1	20.9	3	0	2.09	0	1.040854	1.201737	154	1	1.1	0.9;
	12788	2	0	0	0	0	0	1.045385	33.348896	20	3	1.1	0.9;
	12789	2	0	0	0	0	0	1.045385	33.348896	20	3	1.1	0.9;
	12790	2	0	0	0	0	0	1.045385	33.348896	20	3	1.1	0.9;
	12791	2	0	0	0	0	0	1.045385	33.348896	20	3	1.1	0.9;
	12792	1	10.31	4.3	0	0.67	0	0.95061	-5.613582	110	18	1.1	0.9;
	12793	1	0	0	0	0	0	1.091249	-4.873838	150	9	1.1	0.9;
	12794	2	0	0	0	0	0	1.041485	26.637585	20	10	1.1	0.9;
	12795	1	163.5	35.6	0	4.59	0	1.020907	12.759721	220	5	1.1	0.9;
	12796	1	4.4	2.2	0	1.09	0	1.016079	-14.335683	110	22	1.1	0.9;
	12797	1	162.49	45.11	0	5.38	0	1.001956	10.351303	220	5	1.1	0.9;
	12798	1	0	0	0	21.22	0	1.036256	30.880029	220	13	1.1	0.9;
	12799	2	0	0	0	0	0	1.028648	-14.669346	20	10	1.1	0.9;
	12800	1	44.55	7.5	0	0.6	0	1.027731	13.470944	220	5	1.1	0.9;
	12801	1	125.2	17.8	0	2.5	0	1.013388	13.908401	154	1	1.1	0.9;
	12802	2	0	0	0	0	0	1.027025	78.256658	20	3	1.1	0.9;
	12803	2	0	0	0	0	0	1.056333	76.025167	20	3	1.1	0.9;
	12804	1	-16.42	5.4	0	0.67	0	1.035658	-8.418453	150	9	1.1	0.9;
	12805	1	5.51	1.3	0	0.09	0	1.035412	27.217512	220	5	1.1	0.9;
	12806	2	0	0	0	0	0	1.053741	76.235894	20	3	1.1	0.9;
	12807	2	0	0	0	0	0	1.053741	76.235894	20	3	1.1	0.9;
	12808	2	0	0	0	0	0	1.053741	76.235894	20	3	1.1	0.9;
	12809	1	0	0	0	61.44	0	0.975659	17.56195	400	1	1.1	0.9;
	12810	1	0	-0	0	67.98	0	1.017062	-22.849854	380	4	1.1	0.9;
	12811	1	62.7	9	0	1.12	0	1.026645	1.873835	154	1	1.1	0.9;
	12812	1	-0	-0	0	0	0	1.040039	-14.484778	20	4	1.1	0.9;
	12813	1	129.8	67.2	0	2.57	0	1.003585	-15.239367	150	9	1.1	0.9;
	12814	2	0	0	0	0	0	1.051326	28.77397	20	14	1.1	0.9;
	12815	2	0	0	0	0	0	1.046995	7.343111	11	1	1.1	0.9;
	12816	2	0	0	0	0	0	1.003342	23.89027	10.6	1	1.1	0.9;
	12817	2	0	0	0	0	0	1.062539	23.499848	20	5	1.1	0.9;
	12818	2	0	0	0	0	0	1.062539	23.499848	20	5	1.1	0.9;
	12819	2	0	0	0	0	0	1.022889	18.879559	20	14	1.1	0.9;
	12820	1	226.7	39.3	0	0.99	0	1.030274	-20.752051	220	4	1.1	0.9;
	12821	1	-0.12	-1.74	0	0	0	1.098911	64.999347	380	8	1.2	0.8;
	12822	2	0	0	0	0	0	1.045359	10.101302	20	14	1.1	0.9;
	12823	1	-0	0	0.297267	36.522277	0	1.01789	21.771841	380	6	1.1	0.9;
	12824	1	34.54	15.4	0	29.82	0	1.031793	-18.98361	380	4	1.1	0.9;
	12825	2	0	0	0	0	0	1.055315	14.241147	20	5	1.1	0.9;
	12826	1	45.45	14.9	0	1.12	0	0.953427	-6.513376	110	21	1.1	0.9;
	12827	1	1.3	0.5	0	0.39	0	0.971607	2.277859	110	18	1.1	0.9;
	12828	1	9.91	3.3	0	0.98	0	0.981384	-14.232345	110	12	1.1	0.9;
	12829	2	0	0	0	0	0	0.935194	31.202294	6.6	1	1.1	0.9;
	12830	1	0	0	0.262086	66.559018	0	1.063125	3.494752	380	19	1.1	0.9;
	12831	1	69.88	19.3	0	1.63	0	1.018065	44.19048	110	8	1.1	0.9;
	12832	1	37.94	15.6	0.435676	1.98509	0	1.091105	1.871396	380	21	1.2	0.8;
	12833	2	0	0	0	0	0	1.073133	28.039751	20	14	1.1	0.9;
	12834	1	62.77	5.7	0	2.41	0	1.014279	-23.738766	220	4	1.1	0.9;
	12835	1	0	0	0	28.523869	0	1.021754	8.163576	380	5	1.1	0.9;
	12836	1	-0	0	0	13	0	0.988099	-12.498604	150	9	1.1	0.9;
	12837	1	-0	0	0	1.31	0	1.046789	3.602815	220	1	1.1	0.9;
	12838	1	0	0	0.039248	13.490752	0	1.020484	4.216846	220	7	1.1	0.9;
	12839	1	334.5	135.2	0.040246	7.609754	0	0.989721	7.405739	220	5	1.1	0.9;
	12840	2	0	0	0	0	0	1.052369	0.152148	20	18	1.1	0.9;
	12841	1	0	0	0	2.29	0	1.017459	18.918066	220	5	1.1	0.9;
	12842	1	0	0	0	0	0	0.993879	-6.791088	150	9	1.1	0.9;
	12843	2	0	0	0	0	0	1.052369	0.152148	20	18	1.1	0.9;
	12844	1	-0	0	0	29.01	0	0.964216	32.646923	400	1	1.1	0.9;
	12845	1	0	0	0	90.93	0	0.967312	16.29545	400	1	1.1	0.9;
	12846	1	-0	-0	0	4.26	0	0.965204	-27.491586	220	6	1.1	0.9;
	12847	2	0	0	0	0	0	1.023718	-17.827354	20	4	1.1	0.9;
	12848	1	34.9	5	0	7.79	0	0.997919	15.815205	154	1	1.1	0.9;
	12849	1	131.78	18.51	0.059161	58.650838	0	1.028575	5.914981	220	5	1.1	0.9;
	12850	1	188.6	164.9	0	21.03	0	1.004372	14.523885	380	3	1.1	0.9;
	12851	1	-20.38	-7.04	0	0.65	0	1.017441	-2.682822	110	18	1.1	0.9;
	12852	1	168.13	36.31	0	88.12	0	0.989531	22.273953	380	13	1.1	0.9;
	12853	1	0	-0	0	0.65	0	1.034362	-10.828436	150	9	1.1	0.9;
	12854	1	-0.15	-152.82	0.394814	24.589755	0	1.016881	46.122544	380	3	1.1	0.9;
	12855	2	0	0	0	0	0	1.057579	34.734619	20	14	1.1	0.9;
	12856	2	0	0	0	0	0	1.052369	0.152148	20	18	1.1	0.9;
	12857	1	42.65	18.2	0	0	0	1.031795	9.603746	220	5	1.1	0.9;
	12858	1	-0	-0	0	0	0	1.003512	-19.342822	20	4	1.1	0.9;
	12859	2	0	0	0	0	0	1.055881	11.701666	20	18	1.1	0.9;
	12860	1	0	-0	0.52227	11.818844	0	1.093729	21.721068	220	18	1.2	0.8;
	12861	1	0	0	0	2.81	0	1.03889	12.478006	380	1	1.1	0.9;
	12862	1	155.05	39.56	0	6.27	0	1.022406	14.490154	220	5	1.1	0.9;
	12863	2	0	0	0	0	0	1.017634	-12.373197	20	4	1.1	0.9;
	12864	2	0	0	0	0	0	0.999314	-14.571261	20	4	1.1	0.9;
	12865	2	0	0	0	0	0	1.029957	9.250663	15.75	1	1.1	0.9;
	12866	1	0	-0	0.037734	20.842266	0	1.070357	88.048615	380	3	1.1	0.9;
	12867	1	0	-0	0	20.75	0	1.021041	-10.966008	150	9	1.1	0.9;
	12868	1	5.5	0.8	0	0.9	0	1.013846	22.363012	154	1	1.1	0.9;
	12869	2	0	0	0	0	0	1.028204	5.666749	20	6	1.1	0.9;
	12870	1	-0	-0	0	6.52	0	1.060308	15.402866	220	3	1.1	0.9;
	12871	1	0	0	0	5.01	0	1.00034	8.32903	220	5	1.1	0.9;
	12872	1	20.02	15	0	0.14	0	0.98683	33.936217	220	3	1.1	0.9;
	12873	2	0	0	0	0	0	1.014795	23.808394	6.3	1	1.1	0.9;
	12874	2	0	0	0	0	0	1.042461	6.951342	20	22	1.1	0.9;
	12875	1	89.3	10	0	9.62	0	1.011483	33.497968	220	3	1.1	0.9;
	12876	1	0	0	0	0.3	0	1.025536	-11.129474	150	9	1.1	0.9;
	12877	1	82.59	23.4	0	6.65	0	1.005955	-16.299988	220	4	1.1	0.9;
	12878	2	0	0	0	0	0	1.060659	20.876838	20	5	1.1	0.9;
	12879	1	0	0	0	2.67	0	1.00235	0.684638	220	6	1.1	0.9;
	12880	2	0	0	0	0	0	1.065482	13.792426	20	5	1.1	0.9;
	12881	1	0	0	0	3.1	0	1.018533	-11.038857	150	9	1.1	0.9;
	12882	2	0	0	0	0	0	1.065482	13.792426	20	5	1.1	0.9;
	12883	2	0	0	0	0	0	1.065482	13.792426	20	5	1.1	0.9;
	12884	2	0	0	0	0	0	1.065482	13.792426	20	5	1.1	0.9;
	12885	1	-0	0	0	0.15	0	0.982559	0.477885	150	9	1.1	0.9;
	12886	1	0	0	0	1.01	0	0.993337	-4.053104	110	18	1.1	0.9;
	12887	2	0	0	0	0	0	1.019063	-9.896101	20	4	1.1	0.9;
	12888	1	40.34	15.9	0	6.02	0	1.019803	31.688356	150	8	1.1	0.9;
	12889	1	-8.21	-2.4	0.039039	0.870721	0	1.024414	-19.55614	220	6	1.1	0.9;
	12890	1	-0	0	0	11.22	0	1.010994	-1.162395	380	6	1.1	0.9;
	12891	1	3.7	1.3	0	0.48	0	1.014972	-1.093742	110	19	1.1	0.9;
	12892	1	124.5	-6	0	1.91	0	1.025891	-20.35324	220	4	1.1	0.9;
	12893	1	0	-0	0	3.9	0	1.029487	19.232681	220	5	1.1	0.9;
	12894	1	0	-0	0.19108	3.37892	0	1.050831	28.477811	220	1	1.1	0.9;
	12895	1	0	0	0	0	0	0.944702	-10.819415	150	9	1.1	0.9;
	12896	2	0	0	0	0	0	0.995077	34.268512	11	1	1.1	0.9;
	12897	1	-0	-0	0	89.82	0	0.980806	17.58249	400	1	1.1	0.9;
	12898	1	-20.12	6.6	0	0.85	0	1.043059	-1.392239	150	9	1.1	0.9;
	12899	1	29.43	0.7	0.056538	0.952885	0	1.050082	4.029033	110	3	1.1	0.9;
	12900	2	0	0	0	0	0	1.007268	1.91796	20	18	1.1	0.9;
	12901	1	0	-0	0	0.02	0	1.031096	27.214607	150	8	1.1	0.9;
	12902	2	0	0	0	0	0	1.030509	2.528611	20	18	1.1	0.9;
	12903	1	23.53	8.5	0	1.28	0	1.010415	-0.925141	110	19	1.1	0.9;
	12904	2	0	0	0	0	0	1.019094	-5.889827	20	18	1.1	0.9;
	12905	1	15.32	5	0	0.49	0	1.008763	-9.147175	110	12	1.1	0.9;
	12906	2	0	0	0	0	0	1.020161	16.222206	20	5	1.1	0.9;
	12907	2	0	0	0	0	0	1.020161	16.222206	20	5	1.1	0.9;
	12908	2	0	0	0	0	0	1.020161	16.222206	20	5	1.1	0.9;
	12909	1	3.5	2.7	0	0.39	0	0.964867	-6.136956	110	18	1.1	0.9;
	12910	2	0	0	0	0	0	1.020161	16.222206	20	5	1.1	0.9;
	12911	2	0	0	0	0	0	1.045359	10.101302	20	14	1.1	0.9;
	12912	2	0	0	0	0	0	1.020161	16.222206	20	5	1.1	0.9;
	12913	1	0	-0	0	2.13	0	1.022559	4.844849	220	5	1.1	0.9;
	12914	2	0	0	0	0	0	1.021452	28.019771	20	10	1.1	0.9;
	12915	2	0	0	0	0	0	1.033735	-14.68008	20	4	1.1	0.9;
	12916	2	0	0	0	0	0	1.020161	16.222206	20	5	1.1	0.9;
	12917	2	0	0	0	0	0	1.020161	16.222206	20	5	1.1	0.9;
	12918	1	126.3	109.2	0	54.97	0	1.024162	34.668761	380	3	1.1	0.9;
	12919	1	100.4	1.4	0	5.71	0	1.007061	3.290321	220	20	1.1	0.9;
	12920	2	0	0	0	0	0	1.038461	24.211106	20	5	1.1	0.9;
	12921	2	0	0	0	0	0	1.038461	24.211106	20	5	1.1	0.9;
	12922	1	-0	-0	0	0.04	0	1.025069	-10.520076	150	9	1.1	0.9;
	12923	1	12.11	-1.7	0.059355	18.920645	0	1.011292	-10.569275	220	4	1.1	0.9;
	12924	1	-0	0	0	3.79	0	1.049763	-5.507334	150	9	1.1	0.9;
	12925	2	0	0	0	0	0	1.038461	24.211153	20	5	1.1	0.9;
	12926	1	66.27	19.3	0	12.94	0	1.033586	9.148129	220	5	1.1	0.9;
	12927	1	25	3.6	0	3.77	0	0.998271	35.539323	154	1	1.1	0.9;
	12928	1	35.04	20	0	2.49	0	0.992257	-3.145402	220	21	1.1	0.9;
	12929	2	0	0	0	0	0	1.037296	-1.6783	20	3	1.1	0.9;
	12930	2	0	0	0	0	0	1.041198	4.140848	20	3	1.1	0.9;
	12931	1	8.11	1.4	0	1.47	0	0.969086	-9.139692	110	18	1.1	0.9;
	12932	1	0	0	0	0	0	0.997706	-9.08371	150	9	1.1	0.9;
	12933	1	61.97	38.4	0.505136	4.390101	0	1.059268	-3.944719	150	9	1.1	0.9;
	12934	2	0	0	0	0	0	1.035207	-6.870994	20	4	1.1	0.9;
	12935	1	-0	-0	0	0.93	0	1.009492	30.233366	154	1	1.1	0.9;
	12936	1	105.8	53.6	0	6.27	0	0.974052	-15.620125	150	9	1.1	0.9;
	12937	2	0	0	0	0	0	1.038461	24.211153	20	5	1.1	0.9;
	12938	1	0	0	0	1.72	0	1.027116	-8.230896	380	6	1.1	0.9;
	12939	1	0	-0	0	31.05	0	1.038957	23.917076	380	7	1.1	0.9;
	12940	1	57.46	15.3	0	4.37	0	1.013165	-25.196592	220	4	1.1	0.9;
	12941	1	-0	0	0	25.67	0	0.971452	-14.646159	150	9	1.1	0.9;
	12942	1	45.35	28.1	0	1.9	0	1.016071	-5.447511	150	9	1.1	0.9;
	12943	1	0	-0	0	0.47	0	0.964592	-4.465921	110	19	1.1	0.9;
	12944	1	-0	0	0	0.12	0	1.050586	-3.068048	150	9	1.1	0.9;
	12945	1	279.7	4.6	0	3.77	0	1.021335	-22.198903	220	5	1.1	0.9;
	12946	2	0	0	0	0	0	0.998951	38.095582	9	1	1.1	0.9;
	12947	1	0	0	0	0	0	0.99781	-5.781696	150	9	1.1	0.9;
	12948	1	21.02	12.4	0	2.96	0	1.017569	-13.372717	220	4	1.1	0.9;
	12949	2	0	0	0	0	0	1.002664	30.778713	11	1	1.1	0.9;
	12950	1	61.98	39.2	0	5.1	0	0.994612	-8.532772	150	9	1.1	0.9;
	12951	1	81.79	10.8	0	10.84	0	1.022459	64.220655	220	3	1.1	0.9;
	12952	1	30.03	15	0	2.49	0	1.002492	-3.462358	110	18	1.1	0.9;
	12953	1	19.02	3.8	0	4.55	0	0.926522	-7.872363	110	18	1.1	0.9;
	12954	1	54.66	18.6	0	3.04	0	1.019249	-32.394295	220	4	1.1	0.9;
	12955	1	43.05	25	0	0.02	0	0.997165	-9.074411	150	9	1.1	0.9;
	12956	1	20.62	10.7	0	0.14	0	1.053349	-2.98314	150	9	1.1	0.9;
	12957	1	76.6	10.9	0	6.82	0	1.00654	29.808981	154	1	1.1	0.9;
	12958	1	-0.21	-1.04	0	0	0	0.998386	26.560735	150	8	1.1	0.9;
	12959	1	71.76	11.37	0	0	0	1.059079	9.801529	110	24	1.1	0.9;
	12960	1	18.62	3.6	0	1.14	0	0.988154	-1.80095	110	18	1.1	0.9;
	12961	1	58.77	19.5	0.234954	12.084205	0	1.033542	73.267064	380	3	1.1	0.9;
	12962	2	0	0	0	0	0	1.129641	7.22366	20	9	1.2	0.8;
	12963	1	201.8	71.2	0.233209	14.925376	0	1.027061	22.856557	220	13	1.1	0.9;
	12964	1	22.42	29.5	0	0.08	0	1.020343	-11.016395	150	9	1.1	0.9;
	12965	2	0	0	0	0	0	1.043227	47.072253	20	8	1.1	0.9;
	12966	2	0	0	0	0	0	0.977562	26.05322	13.8	1	1.1	0.9;
	12967	1	9.7	1.4	0	4.91	0	1.008283	18.258946	154	1	1.1	0.9;
	12968	2	0	0	0	0	0	1.053159	33.930358	20	5	1.1	0.9;
	12969	2	0	0	0	0	0	1.053581	52.246384	20	8	1.1	0.9;
	12970	2	0	0	0	0	0	1.053581	52.246384	20	8	1.1	0.9;
	12971	1	25.73	9.3	0	1.49	0	1.014823	-3.066083	110	19	1.1	0.9;
	12972	1	0	0	0	0	0	1.053823	0.430007	150	9	1.1	0.9;
	12973	2	0	0	0	0	0	1.024018	19.935265	20	10	1.1	0.9;
	12974	1	120	17.1	0	0.04	0	1.025296	-0.17102	154	1	1.1	0.9;
	12975	1	0	-0	0	63.59	0	0.970638	21.202444	400	1	1.1	0.9;
	12976	1	-0	-0	0.019913	-0.019913	0	1.002278	-10.054248	150	10	1.1	0.9;
	12977	1	-0	0	0	30.31	0	1.010415	38.669	380	8	1.1	0.9;
	12978	1	7.71	2.4	0	6.93	0	0.992128	10.128688	220	5	1.1	0.9;
	12979	1	-0	0	0	15.04	0	1.009289	-3.506967	220	6	1.1	0.9;
	12980	1	45.25	19.1	0	1.89	0	1.004575	30.879385	220	3	1.1	0.9;
	12981	1	0	0	0	0.17	0	0.991872	-1.713117	110	18	1.1	0.9;
	12982	1	7.61	6.3	0	0.1	0	0.985512	-3.634382	110	18	1.1	0.9;
	12983	2	0	0	0	0	0	0.989744	5.658378	15	1	1.1	0.9;
	12984	2	0	0	0	0	0	1.046953	2.429703	20	2	1.1	0.9;
	12985	1	199.1	23.3	0.030559	2.07	0	0.992255	-19.517601	220	4	1.1	0.9;
	12986	2	0	0	0	0	0	1.046953	2.429703	20	2	1.1	0.9;
	12987	1	292.4	-228.7	0	106.08	0	1.045673	42.716267	380	3	1.1	0.9;
	12988	1	1.8	0.8	0	0.98	0	0.992757	3.56089	110	18	1.1	0.9;
	12989	2	0	0	0	0	0	1.021989	24.453866	11	1	1.1	0.9;
	12990	1	8.81	2.2	0	4.78	0	0.987219	-14.841108	110	22	1.1	0.9;
	12991	1	26.23	9.5	0	0.65	0	1.046517	-0.884756	110	19	1.1	0.9;
	12992	1	11.51	3.8	0	0.53	0	0.962534	-6.995462	110	21	1.1	0.9;
	12993	2	0	0	0	0	0	1.036433	-0.581009	20	2	1.1	0.9;
	12994	2	0	0	0	0	0	1.036433	-0.581009	20	2	1.1	0.9;
	12995	1	-0	-0	0	15.18	0	1.05204	28.126902	220	1	1.1	0.9;
	12996	1	128.8	54.5	0	24.74	0	0.968256	9.335661	220	6	1.1	0.9;
	12997	1	22.83	8	0	3.71	0	1.049159	13.512116	220	5	1.1	0.9;
	12998	1	0	0	0	17.77	0	0.990019	19.974179	220	13	1.1	0.9;
	12999	1	-29.73	23.6	0	11.75	0	0.976439	-19.028349	150	10	1.1	0.9;
	13000	1	126.21	-7.73	0	6.2	0	1.032761	8.736741	220	7	1.1	0.9;
	13001	1	-0	-0	0	0.36	0	1.041792	7.031154	220	5	1.1	0.9;
	13002	2	0	0	0	0	0	0.960023	9.298617	15.75	1	1.1	0.9;
	13003	1	0	-0	0.2364	-0.027	0	1.028106	19.439072	220	6	1.1	0.9;
	13004	1	23.53	9.6	0.020379	3.689621	0	0.993886	57.027969	220	13	1.1	0.9;
	13005	1	179.6	25.6	0	149.28	0	0.996732	2.354009	400	1	1.1	0.9;
	13006	1	3.8	-12.7	0	1.42	0	1.016378	6.321351	110	18	1.1	0.9;
	13007	2	0	0	0	0	0	1.041919	21.315953	20	5	1.1	0.9;
	13008	2	0	0	0	0	0	1.041919	21.315953	20	5	1.1	0.9;
	13009	2	0	0	0	0	0	1.041919	21.315953	20	5	1.1	0.9;
	13010	2	0	0	0	0	0	1.090507	11.169589	10.5	1	1.2	0.8;
	13011	1	-0	0	0.11625	2.64375	0	1.042906	29.386549	220	5	1.1	0.9;
	13012	2	0	0	0	0	0	1.033471	35.764202	20	10	1.1	0.9;
	13013	1	-0	0	0	6.59	0	0.996231	-13.15907	150	9	1.1	0.9;
	13014	1	-0	0	0.041073	2.799464	0	0.984345	-2.661671	110	19	1.1	0.9;
	13015	1	14.72	6.7	0	0.32	0	0.986078	-4.509414	110	18	1.1	0.9;
	13016	1	80.59	5.2	0	7.77	0	1.021103	22.611795	220	3	1.1	0.9;
	13017	1	151.56	-314.2	0	60.91	0	1.04338	41.777724	380	3	1.1	0.9;
	13018	1	487.2	116.7	0	56.29	0	1.008126	-14.668138	380	6	1.1	0.9;
	13019	1	31.03	8.9	0	0.14	0	1.046382	6.543925	110	18	1.1	0.9;
	13020	1	-0	-0	0	23.68	0	1.050334	93.202502	380	3	1.1	0.9;
	13021	1	31.21	-13.87	0	0.32	0	1.048113	-0.97616	110	19	1.1	0.9;
	13022	1	22.02	4.4	0	4.83	0	1.03864	21.497055	220	5	1.1	0.9;
	13023	1	47.95	17.2	0	7.91	0	0.987771	-2.171556	220	6	1.1	0.9;
	13024	1	-0	-0	0	0	0	1.018329	-16.147917	20	4	1.1	0.9;
	13025	1	-0	0	0	0.42	0	1.023493	8.422588	220	5	1.1	0.9;
	13026	1	-0	-0	0	0	0	1.007971	-15.20422	20	4	1.1	0.9;
	13027	2	0	0	0	0	0	1.024979	-15.140439	20	4	1.1	0.9;
	13028	1	373	131.1	0.038265	4.041735	0	1.050184	25.602523	220	5	1.1	0.9;
	13029	2	0	0	0	0	0	1.0018	-10.471434	20	4	1.1	0.9;
	13030	1	-0	-0	0	0	0	1.007971	-15.20422	20	4	1.1	0.9;
	13031	1	203.6	19.81	0	15.46	0	1.065203	3.392771	220	5	1.1	0.9;
	13032	2	0	0	0	0	0	1.011811	-11.502098	20	4	1.1	0.9;
	13033	1	285.2	61.4	0	26.04	0	0.988838	9.016669	220	5	1.1	0.9;
	13034	1	101.2	-2.9	0	0.51	0	1.026116	-2.752551	220	5	1.1	0.9;
	13035	1	0.3	1.5	0	0.1	0	1.037066	30.465618	220	3	1.1	0.9;
	13036	1	0	-0	0.27075	22.54135	0	1.068833	-3.112735	380	18	1.2	0.8;
	13037	2	0	0	0	0	0	1.067719	1.976449	11	1	1.1	0.9;
	13038	1	0	0	0	1.55	0	0.957912	-9.968502	150	9	1.1	0.9;
	13039	1	89.9	29.3	0	6.77	0	1.010465	27.854841	150	8	1.1	0.9;
	13040	1	12.02	3.6	0	0.47	0	0.966671	-6.553134	110	18	1.1	0.9;
	13041	2	0	0	0	0	0	1.040642	8.532255	20	14	1.1	0.9;
	13042	1	82.19	15.6	0	0.43	0	1.015517	32.752016	220	3	1.1	0.9;
	13043	2	0	0	0	0	0	1.010028	16.739138	20	5	1.1	0.9;
	13044	1	15.7	2.2	0	2.88	0	1.03659	21.196154	154	1	1.1	0.9;
	13045	1	37.34	12.1	0.078064	5.711936	0	1.03216	30.377663	220	5	1.1	0.9;
	13046	2	0	0	0	0	0	1.010028	16.739138	20	5	1.1	0.9;
	13047	2	0	0	0	0	0	1.007128	-2.88458	20	2	1.1	0.9;
	13048	2	0	0	0	0	0	1.007128	-2.88458	20	2	1.1	0.9;
	13049	2	0	0	0	0	0	1.007128	-2.88458	20	2	1.1	0.9;
	13050	1	0	-0	0	0.78	0	1.01602	34.837954	154	1	1.1	0.9;
	13051	1	0	-0	0.25741	6.654316	0	1.060215	0.364906	220	18	1.2	0.8;
	13052	2	0	0	0	0	0	1.007128	-2.88458	20	2	1.1	0.9;
	13053	2	0	0	0	0	0	1.007128	-2.88458	20	2	1.1	0.9;
	13054	2	0	0	0	0	0	1.007128	-2.88458	20	2	1.1	0.9;
	13055	2	0	0	0	0	0	1.010028	16.739138	20	5	1.1	0.9;
	13056	1	0	0	0	0	0	1.063704	66.981987	20	8	1.1	0.9;
	13057	2	0	0	0	0	0	1.007679	-4.961668	20	2	1.1	0.9;
	13058	2	0	0	0	0	0	1.007679	-4.961668	20	2	1.1	0.9;
	13059	2	0	0	0	0	0	1.010028	16.739138	20	5	1.1	0.9;
	13060	2	0	0	0	0	0	1.010028	16.739138	20	5	1.1	0.9;
	13061	1	221.92	101.15	0.230538	107.407491	0	1.052488	29.654092	380	13	1.1	0.9;
	13062	2	0	0	0	0	0	1.040259	34.3106	20	5	1.1	0.9;
	13063	1	0	0	0	0.39	0	0.980954	-7.906638	150	9	1.1	0.9;
	13064	2	0	0	0	0	0	1.010028	16.739138	20	5	1.1	0.9;
	13065	1	21.9	3.1	0	4.94	0	1.005379	17.277761	154	1	1.1	0.9;
	13066	2	0	0	0	0	0	1.040259	34.3106	20	5	1.1	0.9;
	13067	1	-0	0	0	44.95	0	1.0058	30.296047	380	5	1.1	0.9;
	13068	1	75.3	7.1	0	0.59	0	1.036027	-4.088525	110	23	1.1	0.9;
	13069	2	0	0	0	0	0	1.023818	-10.471889	20	4	1.1	0.9;
	13070	1	-0	-0	0	1.05	0	1.001957	29.002971	150	8	1.1	0.9;
	13071	1	91.6	27.1	0	10.95	0	1.025033	7.696957	220	5	1.1	0.9;
	13072	1	21.32	7.2	0	1.2	0	1.058962	-2.160344	110	19	1.1	0.9;
	13073	1	26.33	10.3	0	1.11	0	0.923946	-8.905161	110	18	1.1	0.9;
	13074	1	0	-0	0	2.24	0	1.020714	26.358811	150	8	1.1	0.9;
	13075	2	0	0	0	0	0	1.021202	7.512555	6.3	1	1.1	0.9;
	13076	1	-0	0	0	12.42	0	1.003152	-15.344503	150	9	1.1	0.9;
	13077	1	0	0	0	1.52	0	1.00346	8.210209	220	5	1.1	0.9;
	13078	1	113.7	22.3	0	15.55	0	1.029571	7.597342	220	5	1.1	0.9;
	13079	2	0	0	0	0	0	1.065314	53.15385	20	10	1.1	0.9;
	13080	2	0	0	0	0	0	1.034597	19.022639	20	6	1.1	0.9;
	13081	2	0	0	0	0	0	1.024595	25.456155	20	10	1.1	0.9;
	13082	2	0	0	0	0	0	1.003807	31.8753	11	1	1.1	0.9;
	13083	1	0	-0	0	1.91	0	1.025108	6.300939	220	5	1.1	0.9;
	13084	1	20.02	12.4	0	0.03	0	0.966992	-14.766387	150	9	1.1	0.9;
	13085	1	23.7	3.4	0	4.27	0	1.013872	19.392415	154	1	1.1	0.9;
	13086	2	0	0	0	0	0	0.991806	18.024083	20	18	1.1	0.9;
	13087	2	0	0	0	0	0	0.975077	24.924484	10.6	1	1.1	0.9;
	13088	1	-0.94	-6.39	0	1.76	0	0.983289	-8.188197	150	9	1.1	0.9;
	13089	1	1.5	0.3	0	1.54	0	1.003783	-4.176779	110	18	1.1	0.9;
	13090	2	0	0	0	0	0	0.970854	29.381452	15.75	1	1.1	0.9;
	13091	2	0	0	0	0	0	1.022149	-8.624199	20	22	1.1	0.9;
	13092	1	0	-0	0	18.82	0	1.035406	9.681664	220	5	1.1	0.9;
	13093	2	0	0	0	0	0	1.023603	23.127902	20	6	1.1	0.9;
	13094	1	0	0	0	0	0	0.952203	-7.669787	150	9	1.1	0.9;
	13095	2	0	0	0	0	0	1.034376	-4.403708	20	6	1.1	0.9;
	13096	2	0	0	0	0	0	0.9595	30.616672	10.6	1	1.1	0.9;
	13097	1	208.71	62.99	0	3.45	0	1.024594	18.007302	220	5	1.1	0.9;
	13098	1	1.3	-0	0	0.06	0	0.954239	-5.131616	110	18	1.1	0.9;
	13099	1	0	0	0	0	0	1.063704	66.981987	20	8	1.1	0.9;
	13100	1	19.4	2.8	0	11.38	0	0.995233	38.646991	154	1	1.1	0.9;
	13101	2	0	0	0	0	0	1.034597	19.022639	20	6	1.1	0.9;
	13102	1	-0	0	0	0.2	0	0.966688	-9.645178	150	9	1.1	0.9;
	13103	1	-0	-0	0.16572	-0.637052	0	1.012712	32.305565	380	2	1.1	0.9;
	13104	1	0	0	0	0	0	0.914844	-11.630868	150	9	1.1	0.9;
	13105	1	-0.68	-2.44	0	0.02	0	1.040629	-6.467869	150	9	1.1	0.9;
	13106	1	7.41	2.4	0	2.23	0	1.015439	-7.646483	110	12	1.1	0.9;
	13107	1	0	0	0.178531	0.831142	0	1.018571	4.743217	220	20	1.1	0.9;
	13108	2	0	0	0	0	0	1.034532	19.101591	20	6	1.1	0.9;
	13109	1	6.01	2.2	0	0.82	0	1.026591	0.238487	110	19	1.1	0.9;
	13110	1	410.3	91.6	0	13.25	0	1.024898	-7.521787	220	5	1.1	0.9;
	13111	1	69.18	35.1	0	12.59	0	1.030429	-16.387658	220	5	1.1	0.9;
	13112	2	0	0	0	0	0	0.994692	24.255635	11	1	1.1	0.9;
	13113	1	20.22	7.3	0	0.31	0	0.978245	-5.718324	110	19	1.1	0.9;
	13114	1	-0.57	-1	0	0	0	1.052152	-6.603595	150	9	1.1	0.9;
	13115	1	-0	0	0	2.29	0	1.017687	-20.359003	220	6	1.1	0.9;
	13116	1	2	0.3	0	1.02	0	0.98896	-8.641435	110	18	1.1	0.9;
	13117	1	207.4	44.3	0.237333	18.722	0	1.021606	-15.901143	220	4	1.1	0.9;
	13118	1	0	-0	0	4.6	0	1.02049	-0.396649	220	15	1.1	0.9;
	13119	1	0	0	0	39.83	0	1.108354	0.984062	380	1	1.2	0.8;
	13120	1	0	-0	0	3.95	0	0.977424	-27.223914	220	10	1.1	0.9;
	13121	1	50.96	15.7	0	1.52	0	1.017519	44.942025	110	8	1.1	0.9;
	13122	2	0	0	0	0	0	1.035838	-5.172678	20	7	1.1	0.9;
	13123	2	0	0	0	0	0	1.034935	24.924619	20	10	1.1	0.9;
	13124	1	0	0	0	0	0	1.063704	66.981987	20	8	1.1	0.9;
	13125	1	-0	0	0	6.26	0	0.989365	-4.13673	220	5	1.1	0.9;
	13126	1	50.96	16.7	0	0.42	0	0.961079	-6.675736	110	21	1.1	0.9;
	13127	1	0	0	0	0.48	0	1.069676	18.143892	220	3	1.1	0.9;
	13128	1	27.7	4	0	13.66	0	1.02407	-1.128997	154	1	1.1	0.9;
	13129	1	131.5	12.9	0	18.7	0	1.029526	-21.179796	220	4	1.1	0.9;
	13130	1	-18.6	-6.72	0	152.656532	0	1.019598	6.987203	380	5	1.1	0.9;
	13131	2	0	0	0	0	0	1.015521	7.836059	11	1	1.1	0.9;
	13132	1	-0	0	0.468803	1.398864	0	1.028169	-8.006973	380	6	1.1	0.9;
	13133	2	0	0	0	0	0	1.015531	4.163384	20	9	1.1	0.9;
	13134	1	95.61	8.6	0	3.48	0	1.029418	-21.61392	220	4	1.1	0.9;
	13135	1	426.6	32.4	0.157467	3.3319	0	1.019336	-25.362726	220	4	1.1	0.9;
	13136	1	-0	-0	1.152118	-0.467789	0	1.056468	13.149601	380	5	1.1	0.9;
	13137	2	0	0	0	0	0	1.058231	41.885187	20	10	1.1	0.9;
	13138	1	-0	-0	0	1.87	0	1.04054	2.144138	220	5	1.1	0.9;
	13139	1	0	-0	0.090774	21.420444	0	1.058848	6.682119	220	18	1.1	0.9;
	13140	1	0	0	0.339749	-0.264249	0	1.062954	40.091654	150	8	1.1	0.9;
	13141	1	8.11	3.8	0	0.26	0	0.960402	-0.350918	110	18	1.1	0.9;
	13142	1	-0	0	0	1.03	0	1.035139	9.656078	220	5	1.1	0.9;
	13143	2	0	0	0	0	0	1.016606	10.005767	15.75	1	1.1	0.9;
	13144	1	0	-0	0	1.47	0	1.020976	7.632703	220	5	1.1	0.9;
	13145	2	0	0	0	0	0	1.029219	47.514576	20	10	1.1	0.9;
	13146	1	9.21	3.4	0	1.48	0	1.036832	1.60787	110	19	1.1	0.9;
	13147	2	0	0	0	0	0	1.034688	8.184132	20	18	1.1	0.9;
	13148	1	-0.16	-1.44	0	0.42	0	1.02447	39.046287	110	8	1.1	0.9;
	13149	1	0	-0	0	49.35	0	1.053388	-5.242446	380	1	1.1	0.9;
	13150	1	35.64	10.7	0	5.94	0	0.924064	-8.114251	110	18	1.1	0.9;
	13151	1	-0	-0	0.1444	24.6772	0	1.08983	7.912429	380	19	1.2	0.8;
	13152	1	0	-0	0	-1.394261	0	0.983713	25.689119	380	5	1.1	0.9;
	13153	1	-0	-0	0	28.22	0	1.038118	-21.317066	220	6	1.1	0.9;
	13154	1	4.61	1.5	0	1.23	0	0.991214	-14.218835	110	12	1.1	0.9;
	13155	1	15.22	2.5	0	0.69	0	1.002707	2.493238	110	18	1.1	0.9;
	13156	1	22.12	8	0	1.51	0	0.975452	-4.937173	110	19	1.1	0.9;
	13157	1	18.32	6.6	0	0.45	0	1.041751	1.738367	110	23	1.1	0.9;
	13158	2	0	0	0	0	0	1.036955	14.906904	20	7	1.1	0.9;
	13159	2	0	0	0	0	0	0.974729	26.313842	6.3	1	1.1	0.9;
	13160	2	0	0	0	0	0	1.015029	3.043049	20	22	1.1	0.9;
	13161	1	-0	-0	0	0.86	0	1.01713	-4.112133	150	9	1.1	0.9;
	13162	1	278.3	33.6	0	25.9	0	1.017956	13.697984	220	5	1.1	0.9;
	13163	2	0	0	0	0	0	1.026957	18.072383	20	5	1.1	0.9;
	13164	1	0	-0	0.224798	32.902035	0	1.068363	-3.573969	380	12	1.2	0.8;
	13165	2	0	0	0	0	0	1.031159	7.793526	11	1	1.1	0.9;
	13166	1	-0.2	-1	0	0	0	1.018573	25.306793	150	8	1.1	0.9;
	13167	1	87.7	26	0	47.58	0	1.032754	9.236724	220	5	1.1	0.9;
	13168	1	0	-0	0	1.25	0	1.015159	31.900987	154	1	1.1	0.9;
	13169	1	141.52	50.48	0	66.28	0	1.044224	13.79089	380	13	1.1	0.9;
	13170	1	0	-0	0	0	0	1.050488	44.077076	380	8	1.1	0.9;
	13171	1	-0	0	0.188566	35.384296	0	1.043123	11.567567	380	3	1.1	0.9;
	13172	1	3.1	1.6	0	0.09	0	1.024989	6.612752	220	5	1.1	0.9;
	13173	1	36.04	22.3	0	2.37	0	1.013455	-8.739663	150	9	1.1	0.9;
	13174	2	0	0	0	0	0	1.043087	11.638661	20	14	1.1	0.9;
	13175	1	-0.22	-3.34	0	0	0	1.010986	27.852289	150	8	1.1	0.9;
	13176	1	5.51	2.2	0	3.41	0	0.99653	-9.692613	110	22	1.1	0.9;
	13177	1	0	-0	0.051721	0.858279	0	0.975181	15.869625	380	5	1.1	0.9;
	13178	1	-0	0	0	1.72	0	1.000925	34.567738	154	1	1.1	0.9;
	13179	1	-0	-0	0	0.98	0	1.025843	35.252312	220	3	1.1	0.9;
	13180	1	0	0	0.11521	58.265737	0	1.024306	14.515875	380	7	1.1	0.9;
	13181	1	91	21	0	11.23	0	1.028706	7.443704	220	5	1.1	0.9;
	13182	2	0	0	0	0	0	0.975956	41.478844	154	1	1.1	0.9;
	13183	1	-0	-0	0	0.7	0	1.066454	-4.703726	150	9	1.1	0.9;
	13184	1	98.7	14	0	1.01	0	1.006143	20.873808	154	1	1.1	0.9;
	13185	1	33.54	12.2	0	2.41	0	1.025093	13.584701	110	19	1.1	0.9;
	13186	1	32.1	4.5	0	2.23	0	1.004296	17.471585	154	1	1.1	0.9;
	13187	1	-6.81	2.2	0	0.57	0	1.026643	-1.232792	150	9	1.1	0.9;
	13188	2	0	0	0	0	0	1.051387	32.912535	20	5	1.1	0.9;
	13189	2	0	0	0	0	0	1.051387	32.912535	20	5	1.1	0.9;
	13190	2	0	0	0	0	0	1.051387	32.912535	20	5	1.1	0.9;
	13191	2	0	0	0	0	0	1.020032	8.963657	11	1	1.1	0.9;
	13192	2	0	0	0	0	0	1.038082	40.430696	20	2	1.1	0.9;
	13193	2	0	0	0	0	0	1.003171	37.39829	6.3	1	1.1	0.9;
	13194	1	-0	-0	0.038554	0.011446	0	1.018948	-8.90721	110	12	1.1	0.9;
	13195	2	0	0	0	0	0	0.98379	29.90696	10.6	1	1.1	0.9;
	13196	1	-0	0	0	4.8	0	1.03765	17.609515	220	3	1.1	0.9;
	13197	1	59.67	13.1	0	2.26	0	0.995164	34.73145	110	8	1.1	0.9;
	13198	1	0	-0	1.04964	7.653846	0	1.083585	35.443136	380	5	1.2	0.8;
	13199	1	10.81	1.3	0	2.23	0	0.975439	-10.224225	110	18	1.1	0.9;
	13200	2	0	0	0	0	0	0.99654	26.163496	11	1	1.1	0.9;
	13201	1	93.1	-15.12	0	1.04	0	1.010954	89.347655	150	1	1.1	0.9;
	13202	1	-0	-0	0.019887	0.140113	0	0.995533	-14.126603	110	22	1.1	0.9;
	13203	1	29.33	5.6	0	6.56	0	1.025445	7.76184	220	5	1.1	0.9;
	13204	1	3.4	1.1	0	2.14	0	0.985641	2.113381	110	18	1.1	0.9;
	13205	1	7.41	2.7	0	3.34	0	1.042369	0.793884	110	19	1.1	0.9;
	13206	1	66.57	0.56	0	7.63	0	1.028784	19.379012	220	13	1.1	0.9;
	13207	1	0	0	0	0	0	1.023166	4.463602	20	23	1.1	0.9;
	13208	1	50.96	16.8	0	1.03	0	0.942791	-9.095892	110	21	1.1	0.9;
	13209	2	0	0	0	0	0	1.029835	-14.681025	20	4	1.1	0.9;
	13210	1	-0	-0	0.040026	0.439974	0	1.005238	-7.194433	150	10	1.1	0.9;
	13211	1	122.3	17.5	0	1.51	0	1.025167	2.023484	154	1	1.1	0.9;
	13212	2	0	0	0	0	0	1.00107	32.866746	11	1	1.1	0.9;
	13213	1	93.9	21.6	0	2.05	0	1.071669	23.992312	220	3	1.1	0.9;
	13214	1	113.4	19.7	0.215748	8.474252	0	1.018996	14.242218	220	5	1.1	0.9;
	13215	1	-0.56	-0	0	108.21	0	1.0629	21.173824	380	5	1.1	0.9;
	13216	1	0	-7.5	0	0	0	1.018896	-8.431592	150	9	1.1	0.9;
	13217	1	0	-0	0	2.47	0	0.988894	-26.08969	220	6	1.1	0.9;
	13218	2	0	0	0	0	0	1.008348	30.337205	14.4	1	1.1	0.9;
	13219	1	88	36.2	0	60	0	1.040094	-20.271382	380	10	1.1	0.9;
	13220	1	0	0	0	26.15	0	1.040914	13.793	380	1	1.1	0.9;
	13221	1	-0	-0	0.04975	68.521798	0	1.066121	2.762902	380	18	1.1	0.9;
	13222	2	0	0	0	0	0	0.943436	50.691637	13.8	1	1.1	0.9;
	13223	2	0	0	0	0	0	1.021005	21.987746	20	14	1.1	0.9;
	13224	1	37.94	3.4	0	11.17	0	1.024598	-28.939594	220	4	1.1	0.9;
	13225	1	0	-0	0	3.72	0	0.998123	-23.060675	220	6	1.1	0.9;
	13226	2	0	0	0	0	0	1.00746	39.171104	11	1	1.1	0.9;
	13227	1	4.2	1.4	0	0.98	0	1.013676	-8.338005	110	12	1.1	0.9;
	13228	1	19.92	6.6	0	0.18	0	0.975711	-8.687888	110	21	1.1	0.9;
	13229	1	-0	0	0	6.285833	0	0.979466	-1.074571	220	20	1.1	0.9;
	13230	1	0	-0	0.055698	10.563585	0	1.076036	84.574972	220	3	1.2	0.8;
	13231	1	0	-0	0	43.7	0	0.977566	23.48751	400	1	1.1	0.9;
	13232	2	0	0	0	0	0	1.010998	1.300317	20	10	1.1	0.9;
	13233	1	1.6	88.5	0.467225	47.990646	0	1.03133	-13.963675	380	4	1.1	0.9;
	13234	1	3.7	0.2	0	0.05	0	1.030067	19.445758	220	5	1.1	0.9;
	13235	2	0	0	0	0	0	1.058114	43.382414	20	10	1.1	0.9;
	13236	1	20.32	6.3	0	0.05	0	0.996666	-4.723793	110	18	1.1	0.9;
	13237	1	-0	0	0	3.44	0	1.025943	5.865231	220	5	1.1	0.9;
	13238	2	0	0	0	0	0	0.9646	54.13357	11	1	1.1	0.9;
	13239	1	27.93	8.9	0	0.47	0	0.98836	-4.6058	110	18	1.1	0.9;
	13240	1	516.7	144.3	0	2.32	0	1.007537	-12.433936	220	4	1.1	0.9;
	13241	1	51.56	18.7	0	0.18	0	1.024332	-4.383178	110	19	1.1	0.9;
	13242	1	0	0	0	7.89	0	1.014468	46.591225	220	8	1.1	0.9;
	13243	1	-0	-0	0	16.23	0	0.982731	29.96695	220	3	1.1	0.9;
	13244	1	0	-0	0.33861	-0.263363	0	1.066937	48.44975	380	8	1.1	0.9;
	13245	1	0	0	0	4.39	0	0.977843	-27.152454	220	10	1.1	0.9;
	13246	1	28.63	17.7	0	0.1	0	0.970166	-9.580952	150	9	1.1	0.9;
	13247	1	-0	-0	0.256503	101.229042	0	1.046477	-20.234317	380	10	1.1	0.9;
	13248	1	0	0	0	0.4	0	1.022708	-7.370065	150	9	1.1	0.9;
	13249	1	0	0	0	9.64	0	1.022649	44.459971	380	3	1.1	0.9;
	13250	1	100.7	20.5	0.07943	3.98057	0	1.00925	13.329705	220	5	1.1	0.9;
	13251	1	44.05	21	0	0.03	0	0.992728	-23.206002	220	6	1.1	0.9;
	13252	1	-0	-0	0	0.25	0	1.041896	26.213331	150	8	1.1	0.9;
	13253	1	-0	-0	0	2.78	0	1.006717	-12.883457	220	10	1.1	0.9;
	13254	1	925.91	147.97	0	0	0	1.023428	21.664024	150	2	1.1	0.9;
	13255	1	7.71	2.2	0.115287	1.163535	0	1.03042	-8.082898	110	22	1.1	0.9;
	13256	1	0	0	0.167278	198.780601	0	1.096387	69.583402	380	3	1.2	0.8;
	13257	1	23.43	8.4	0	0.17	0	1.056069	-2.371819	110	15	1.1	0.9;
	13258	1	0	0	0	35.28	0	0.96993	17.803297	400	1	1.1	0.9;
	13259	1	261.5	95	0	49.05	0	1.019183	16.868099	380	3	1.1	0.9;
	13260	1	0	0	0	1.29	0	0.991102	-5.123003	110	21	1.1	0.9;
	13261	1	-0	-0	0	0	0	1.015642	-21.99726	20	4	1.1	0.9;
	13262	1	0	0	0	9.679984	0	1.011422	-0.473281	220	20	1.1	0.9;
	13263	1	298.4	122.5	0.233011	3.125533	0	1.032574	26.636821	220	13	1.1	0.9;
	13264	1	-0	-0	0	0	0	1.003373	-19.517601	20	4	1.1	0.9;
	13265	2	0	0	0	0	0	1.000157	-14.767921	20	4	1.1	0.9;
	13266	1	-0	-0	0	0	0	1.003373	-19.517601	20	4	1.1	0.9;
	13267	1	0	0	0.019352	0.700648	0	1.014637	-8.280321	110	12	1.1	0.9;
	13268	2	0	0	0	0	0	1.016188	-15.391233	20	4	1.1	0.9;
	13269	1	0	-0	0	0	0	1.019745	-19.936916	20	4	1.1	0.9;
	13270	2	0	0	0	0	0	0.981334	37.783033	6.3	1	1.1	0.9;
	13271	1	274.3	269.4	1.130161	137.739703	0	1.03295	-8.426784	380	4	1.1	0.9;
	13272	1	-0	0	0	44.96	0	1.052781	-9.309841	380	22	1.1	0.9;
	13273	1	82.29	29.6	0	1.86	0	0.981474	-5.943609	110	21	1.1	0.9;
	13274	1	0	0	0.097487	3.732513	0	1.034196	12.040102	220	5	1.1	0.9;
	13275	1	81.29	26.5	0	7.04	0	0.987786	25.477106	150	8	1.1	0.9;
	13276	1	60.57	30.7	0.461662	2.094517	0	1.032979	13.492326	110	19	1.1	0.9;
	13277	1	-0	-0	0	0.51	0	0.980943	-7.908747	150	9	1.1	0.9;
	13278	1	-0	-0	0.040017	0.899983	0	0.994083	-17.375682	110	22	1.1	0.9;
	13279	1	-0	0	0.050138	2.879862	0	1.016914	-15.630869	220	4	1.1	0.9;
	13280	1	-0	0	0	1.49	0	1.025593	37.336465	220	3	1.1	0.9;
	13281	1	4.1	1.5	0	1.01	0	1.015635	-1.587105	110	19	1.1	0.9;
	13282	1	53.36	44.2	0	22.89	0	1.02421	52.49799	380	3	1.1	0.9;
	13283	1	18.52	11.1	0	0.07	0	0.961964	-9.865409	150	9	1.1	0.9;
	13284	1	0	0	0	0	0	0.965766	-9.555001	150	9	1.1	0.9;
	13285	2	0	0	0	0	0	1.019281	-1.243469	20	4	1.1	0.9;
	13286	1	92.9	48.4	0	5.76	0	1.016314	7.334762	220	7	1.1	0.9;
	13287	1	44.95	21.2	0.040111	39.389889	0	1.021338	-13.991173	380	4	1.1	0.9;
	13288	1	0	-0	0	0.87	0	1.027732	2.440628	154	1	1.1	0.9;
	13289	1	88.5	24.5	0	2.23	0	1.031349	4.288277	220	5	1.1	0.9;
	13290	1	2	1	0	0.8	0	1.025409	-3.188199	110	21	1.1	0.9;
	13291	1	-8.49	-17.09	0	0	0	0.998718	26.916364	150	8	1.1	0.9;
	13292	2	0	0	0	0	0	1.042317	14.274354	20	14	1.1	0.9;
	13293	1	4.2	2	0	0.42	0	0.995043	-3.754412	110	18	1.1	0.9;
	13294	1	-0	0	0	14.64	0	1.051966	25.688886	220	5	1.1	0.9;
	13295	1	79.39	16.8	0.039541	2.620459	0	1.009613	-10.15781	220	4	1.1	0.9;
	13296	1	69.28	31.3	0	32.76	0	1.014222	32.604509	150	8	1.1	0.9;
	13297	1	11.5	1.7	0	3.08	0	1.011334	21.793145	154	1	1.1	0.9;
	13298	2	0	0	0	0	0	1.019411	-14.420581	20	10	1.1	0.9;
	13299	1	8.7	1.2	0	2.82	0	0.997725	19.123501	154	1	1.1	0.9;
	13300	1	0	-0	0	50.210627	0	0.984318	15.137295	380	5	1.1	0.9;
	13301	1	-0	-0	0	4.08	0	0.961087	-15.659318	220	10	1.1	0.9;
	13302	1	41.15	10.3	0	2.41	0	1.034077	8.370153	220	5	1.1	0.9;
	13303	2	0	0	0	0	0	1.045359	10.101302	20	14	1.1	0.9;
	13304	1	0	-0	0.233761	-0.194801	0	1.029838	-11.418316	380	6	1.1	0.9;
	13305	1	52.48	6.43	0	1.47	0	1.010396	89.157471	150	1	1.1	0.9;
	13306	1	-0	0	0	3.08	0	1.054559	75.762051	220	3	1.1	0.9;
	13307	1	79.29	46.6	0.293438	5.195468	0	1.033156	-14.381315	220	4	1.1	0.9;
	13308	1	119.02	12.54	0	23.45	0	1.026587	18.649448	220	5	1.1	0.9;
	13309	1	0	-0	0	31.57	0	1.021082	63.452163	380	3	1.1	0.9;
	13310	2	0	0	0	0	0	1.025649	-20.066719	20	6	1.1	0.9;
	13311	2	0	0	0	0	0	1.055904	-17.760173	20	6	1.1	0.9;
	13312	2	0	0	0	0	0	0.995346	36.752154	6.3	1	1.1	0.9;
	13313	2	0	0	0	0	0	1.023809	10.838684	20	6	1.1	0.9;
	13314	1	102.6	14.7	0	2.97	0	1.023188	2.537974	220	20	1.1	0.9;
	13315	1	-0	0	0.048364	5.54206	0	1.056296	-0.708052	220	18	1.1	0.9;
	13316	2	0	0	0	0	0	1.034532	19.101591	20	6	1.1	0.9;
	13317	1	73.68	24	0	0.46	0	0.993556	25.389197	150	8	1.1	0.9;
	13318	1	23.03	14.2	0	1.24	0	1.032532	-0.066659	150	9	1.1	0.9;
	13319	1	27.13	16.8	0	1.34	0	1.014647	-12.087921	150	9	1.1	0.9;
	13320	1	145.3	3.8	0	41.054794	0	1.010963	-0.712232	380	6	1.1	0.9;
	13321	1	0	-0	0.270022	107.094537	0	1.078833	1.728689	380	19	1.1	0.9;
	13322	2	0	0	0	0	0	1.043049	13.489159	20	18	1.1	0.9;
	13323	1	-0	0	0	0	0	1.021346	-22.881448	20	4	1.1	0.9;
	13324	1	0	0	0	3.02	0	1.023892	8.254887	220	5	1.1	0.9;
	13325	1	0	-0	0.039122	1.170659	0	1.021789	-19.844351	220	6	1.1	0.9;
	13326	1	70.78	19.6	0	8.1	0	1.026967	22.605928	220	5	1.1	0.9;
	13327	2	0	0	0	0	0	1.117446	4.922187	20	9	1.2	0.8;
	13328	1	15.63	3.27	0	0.77	0	0.98627	-4.659541	110	18	1.1	0.9;
	13329	1	-0	0	0.280385	9.908955	0	1.059724	-0.446506	380	21	1.1	0.9;
	13330	2	0	0	0	0	0	1.011478	-0.639289	20	10	1.1	0.9;
	13331	1	0	-0	0	5.78	0	0.972672	16.077471	220	1	1.1	0.9;
	13332	1	110	3.4	0	11.95	0	1.027321	5.578339	220	5	1.1	0.9;
	13333	2	0	0	0	0	0	1.065343	53.072846	20	10	1.1	0.9;
	13334	1	91.4	18.1	0	4.89	0	1.026971	5.501097	220	5	1.1	0.9;
	13335	1	6.41	4.3	0	0.23	0	0.971843	-7.304062	110	18	1.1	0.9;
	13336	1	152.18	21.63	0	1.91	0	1.037306	27.360815	220	5	1.1	0.9;
	13337	1	0	0	0	249.03	0	0.975965	28.510882	400	1	1.1	0.9;
	13338	2	0	0	0	0	0	1.000057	26.10357	11	1	1.1	0.9;
	13339	1	80.49	30	0	5.89	0	0.984937	7.581128	220	3	1.1	0.9;
	13340	1	-0	0	0	0.49	0	0.973224	-4.991666	150	9	1.1	0.9;
	13341	1	-0	0	0.218331	2.960118	0	0.995041	33.268248	380	2	1.1	0.9;
	13342	1	0.2	0.1	0	0.28	0	1.063145	-2.590835	110	19	1.1	0.9;
	13343	1	0	0	0	0	0	0.993675	90.655341	150	1	1.1	0.9;
	13344	1	7.91	2.7	0	0.01	0	0.972809	-28.706724	150	10	1.1	0.9;
	13345	2	0	0	0	0	0	1.025019	-12.264635	20	6	1.1	0.9;
	13346	1	132.2	49.1	0	69.43	0	1.041768	12.293615	380	16	1.1	0.9;
	13347	1	54.3	7.8	0	4.44	0	1.024854	17.026834	154	1	1.1	0.9;
	13348	1	0	0	0	2.42	0	1.046536	-20.234594	380	10	1.1	0.9;
	13349	1	-0.1	68.5	0	17.24	0	1.037314	-19.889053	220	4	1.1	0.9;
	13350	1	0	-0	0.243398	-0.202832	0	1.012731	25.47515	380	6	1.1	0.9;
	13351	1	0	0	0	0	0	1.017233	-16.531897	20	4	1.1	0.9;
	13352	1	12.61	6.3	0	14.98	0	1.048814	28.18734	220	1	1.1	0.9;
	13353	1	-19.83	-13.51	0	0.31	0	1.024685	-1.981769	110	21	1.1	0.9;
	13354	1	-22.37	-8.9	0.151672	11.87164	0	1.107706	75.952864	380	3	1.2	0.8;
	13355	1	-0	0	0	1.03	0	1.066282	1.23233	150	9	1.1	0.9;
	13356	1	8.31	3	0	0.52	0	1.068417	1.099548	110	19	1.1	0.9;
	13357	1	78.01	4.25	0	12.32	0	1.016026	27.942514	220	13	1.1	0.9;
	13358	1	-0	0	0	0.57	0	1.009595	-9.543523	150	9	1.1	0.9;
	13359	1	0	0	0	6.27	0	1.01437	6.320018	380	1	1.1	0.9;
	13360	1	-0	0	0	0.71	0	1.06669	-4.514841	150	9	1.1	0.9;
	13361	1	0	0	0	4.15	0	1.029371	9.370477	220	1	1.1	0.9;
	13362	1	-0	-0	0	4.74	0	1.000539	-7.914068	220	22	1.1	0.9;
	13363	2	0	0	0	0	0	1.062565	2.294369	20	5	1.1	0.9;
	13364	1	0	0	0	0	0	1.019724	-7.10214	150	9	1.1	0.9;
	13365	1	21.88	4.73	0	-8.06	0	0.995583	89.682428	150	1	1.1	0.9;
	13366	1	206.4	21	0	8.55	0	0.973767	-22.709919	150	10	1.1	0.9;
	13367	1	36.84	14.1	0	3.8	0	1.00814	-12.831745	220	4	1.1	0.9;
	13368	1	0	0	0	0	0	0.88398	-14.103043	150	9	1.2	0.8;
	13369	1	14.81	0.6	0	0.68	0	1.030892	1.639451	110	3	1.1	0.9;
	13370	2	0	0	0	0	0	1.024788	-12.464012	20	4	1.1	0.9;
	13371	2	0	0	0	0	0	1.027307	-18.397327	20	4	1.1	0.9;
	13372	1	0	-0	0	2.15	0	1.050644	32.165751	220	5	1.1	0.9;
	13373	1	121.8	17.4	0	1.01	0	1.019313	14.657917	154	1	1.1	0.9;
	13374	1	-0	-0	0	0	0	1.020254	-23.064206	20	4	1.1	0.9;
	13375	1	0	-0	0	0.44	0	1.029418	77.694305	220	3	1.1	0.9;
	13376	1	27.93	11.1	0.217305	30.057967	0	1.089497	1.416675	380	21	1.2	0.8;
	13377	1	0	-0	0	0	0	1.028646	7.51961	20	3	1.1	0.9;
	13378	1	11.35	3.27	0	1.47	0	1.009925	-3.074401	110	18	1.1	0.9;
	13379	2	0	0	0	0	0	1.052254	11.947216	20	3	1.1	0.9;
	13380	1	291.6	47	0	2.38	0	1.039079	-1.89058	220	5	1.1	0.9;
	13381	1	42.45	-1.8	0.058845	0.360963	0	0.99558	90.692442	150	1	1.1	0.9;
	13382	1	0	0	0	2	0	1.04189	2.242319	220	5	1.1	0.9;
	13383	2	0	0	0	0	0	1.063297	57.37246	20	3	1.1	0.9;
	13384	1	492.4	162	0.039293	50.430707	0	1.037466	28.004671	380	14	1.1	0.9;
	13385	2	0	0	0	0	0	1.030828	-0.367949	20	6	1.1	0.9;
	13386	2	0	0	0	0	0	1.030828	-0.367949	20	6	1.1	0.9;
	13387	2	0	0	0	0	0	1.030828	-0.367949	20	6	1.1	0.9;
	13388	2	0	0	0	0	0	1.055605	26.96637	20	3	1.1	0.9;
	13389	2	0	0	0	0	0	1.055605	26.96637	20	3	1.1	0.9;
	13390	2	0	0	0	0	0	1.016891	10.696134	20	5	1.1	0.9;
	13391	2	0	0	0	0	0	1.016891	10.696134	20	5	1.1	0.9;
	13392	2	0	0	0	0	0	1.043178	23.372652	20	5	1.1	0.9;
	13393	2	0	0	0	0	0	1.016891	10.696134	20	5	1.1	0.9;
	13394	2	0	0	0	0	0	1.016891	10.696134	20	5	1.1	0.9;
	13395	2	0	0	0	0	0	1.016891	10.696134	20	5	1.1	0.9;
	13396	1	27.9	4	0	4.34	0	1.005221	16.315529	154	1	1.1	0.9;
	13397	2	0	0	0	0	0	1.055605	26.96637	20	3	1.1	0.9;
	13398	2	0	0	0	0	0	1.055605	26.96637	20	3	1.1	0.9;
	13399	2	0	0	0	0	0	1.055605	26.96637	20	3	1.1	0.9;
	13400	1	16.52	6.6	0	0.65	0	0.99704	-13.288962	110	22	1.1	0.9;
	13401	1	-41.57	-23.4	0	0	0	1.019406	37.270568	150	8	1.1	0.9;
	13402	1	-0	-0	0	15.26	0	0.999512	33.001947	380	2	1.1	0.9;
	13403	1	-0	0	0	0.59	0	1.040995	33.204925	380	5	1.1	0.9;
	13404	1	8.91	2.5	0	0.31	0	0.99762	-4.347464	110	18	1.1	0.9;
	13405	1	-0	0	0.211714	1.318286	0	1.054724	33.113517	220	5	1.1	0.9;
	13406	1	-0	0	0	11.35	0	1.004024	-18.321279	220	6	1.1	0.9;
	13407	2	0	0	0	0	0	0.965886	44.832426	11	1	1.1	0.9;
	13408	1	-0	0	0.118317	122.437372	0	1.040657	19.42178	380	5	1.1	0.9;
	13409	2	0	0	0	0	0	1.015954	3.75914	20	9	1.1	0.9;
	13410	1	133.5	24.1	0	2.92	0	1.013727	36.461333	110	8	1.1	0.9;
	13411	2	0	0	0	0	0	1.043619	-8.436397	20	4	1.1	0.9;
	13412	2	0	0	0	0	0	1.027264	14.46311	20	14	1.1	0.9;
	13413	1	8.3	1.2	0	0.53	0	1.016594	19.314094	154	1	1.1	0.9;
	13414	1	0	0	0	1.41	0	1.048204	32.7209	150	8	1.1	0.9;
	13415	1	-0	-0	0	5.05	0	0.983779	-12.14246	220	22	1.1	0.9;
	13416	1	46.9	6.7	0	3.45	0	0.975073	40.56328	154	1	1.1	0.9;
	13417	1	242.4	5.1	0	4.93	0	1.012371	-11.441989	220	6	1.1	0.9;
	13418	1	0	-0	0	0.55	0	0.944968	-10.854722	150	9	1.1	0.9;
	13419	2	0	0	0	0	0	1.043942	17.353218	20	5	1.1	0.9;
	13420	1	6.41	2.3	0	1.51	0	0.963614	-4.868439	110	19	1.1	0.9;
	13421	2	0	0	0	0	0	1.02494	28.997481	20	5	1.1	0.9;
	13422	1	24.33	9.9	0	0.76	0	0.98586	-4.164195	110	18	1.1	0.9;
	13423	2	0	0	0	0	0	1.02494	28.997481	20	5	1.1	0.9;
	13424	2	0	0	0	0	0	1.02494	28.997481	20	5	1.1	0.9;
	13425	2	0	0	0	0	0	1.025625	-16.674993	20	6	1.1	0.9;
	13426	2	0	0	0	0	0	1.03448	21.013586	20	5	1.1	0.9;
	13427	1	-0	-0	0	0	0	1.027866	16.538428	20	5	1.1	0.9;
	13428	1	-0	-0	0	20.53	0	1.00143	-6.748273	220	6	1.1	0.9;
	13429	1	-0	-0	0	0.59	0	1.053805	0.42706	150	9	1.1	0.9;
	13430	1	50.2	7.1	0	3.52	0	0.998645	10.064174	154	1	1.1	0.9;
	13431	1	88.1	36	0	27.89	0	1.029674	25.846398	150	8	1.1	0.9;
	13432	1	-0	0	0	5.88	0	1.027952	46.673663	380	3	1.1	0.9;
	13433	2	0	0	0	0	0	1.021414	-20.989235	20	4	1.1	0.9;
	13434	1	0	0	0	0	0	1.019678	-23.024367	20	4	1.1	0.9;
	13435	1	-0	0	0	0.52	0	0.929129	-11.916099	150	9	1.1	0.9;
	13436	1	125	9.6	0	0.25	0	1.03722	14.842216	220	3	1.1	0.9;
	13437	2	0	0	0	0	0	1.021414	-20.989235	20	4	1.1	0.9;
	13438	1	33.5	4.8	0	1.98	0	1.003202	27.868067	154	1	1.1	0.9;
	13439	2	0	0	0	0	0	1.021414	-20.989235	20	4	1.1	0.9;
	13440	2	0	0	0	0	0	1.014614	-15.196146	20	4	1.1	0.9;
	13441	1	55.9	102.9	0	23.9	0	1.021508	58.607652	380	3	1.1	0.9;
	13442	1	0	0	0	4.5	0	0.989357	45.563421	154	1	1.1	0.9;
	13443	1	0	-0	0	0.58	0	1.017775	79.055094	220	3	1.1	0.9;
	13444	1	0	-0	0	3.65	0	1.023486	19.424707	220	5	1.1	0.9;
	13445	1	-0.57	-1	0	3.29	0	1.053846	-6.619406	150	9	1.1	0.9;
	13446	1	3.8	2.5	0	0.84	0	1.009746	8.102362	110	18	1.1	0.9;
	13447	2	0	0	0	0	0	1.011745	17.062335	13.8	1	1.1	0.9;
	13448	1	-0	0	0	6.33	0	1.016787	21.906986	220	5	1.1	0.9;
	13449	1	119	97.8	0	0.14	0	0.992617	33.294937	220	3	1.1	0.9;
	13450	1	7	1	0	0.78	0	1.01278	22.188029	154	1	1.1	0.9;
	13451	1	2.2	0.5	0	0.13	0	1.025682	-6.023946	220	10	1.1	0.9;
	13452	1	-0	-0	0	1.77	0	1.0037	-10.261267	150	9	1.1	0.9;
	13453	1	181.1	25.8	0	0.83	0	1.025028	7.263324	154	1	1.1	0.9;
	13454	1	223.3	-7.6	0	19.97	0	1.008844	51.084468	380	3	1.1	0.9;
	13455	1	140.8	44.9	0.586408	62.421067	0	1.043627	7.79121	380	5	1.1	0.9;
	13456	2	0	0	0	0	0	1.005915	0.12691	20	18	1.1	0.9;
	13457	1	0	0	0	2.83	0	0.991364	18.825266	154	1	1.1	0.9;
	13458	1	112.4	16.1	0	3.37	0	1.019071	7.96358	154	1	1.1	0.9;
	13459	1	11.31	13.8	0	1.11	0	1.025237	-7.644418	150	9	1.1	0.9;
	13460	1	-678.72	-333.49	0	0	0	1.042565	24.855063	27	3	1.1	0.9;
	13461	1	-0	-0.3	0.055108	0.334077	0	1.077884	7.357907	110	18	1.2	0.8;
	13462	1	-0	-0	0	1.12	0	1.00691	-22.518312	220	4	1.1	0.9;
	13463	1	87.4	35.5	0	16.41	0	1.046177	8.247773	380	16	1.1	0.9;
	13464	1	285.2	83.5	0	5.6	0	0.958925	-16.006713	220	10	1.1	0.9;
	13465	1	-0	0	0	0.31	0	1.050646	32.165732	220	5	1.1	0.9;
	13466	2	0	0	0	0	0	1.062896	-0.392816	20	5	1.1	0.9;
	13467	2	0	0	0	0	0	0.989708	1.882164	20	18	1.1	0.9;
	13468	2	0	0	0	0	0	1.024375	-1.976838	20	10	1.1	0.9;
	13469	1	0	0	0	0	0	1.04035	8.357264	20	5	1.1	0.9;
	13470	1	35.14	12.7	0	1.42	0	0.972308	-2.718329	110	19	1.1	0.9;
	13471	1	0	-0	0	19.71	0	1.033064	16.599356	220	5	1.1	0.9;
	13472	1	-0	-0	0.080832	6.709376	0	1.000964	-2.993742	220	6	1.1	0.9;
	13473	1	-0	0	0	47.1	0	0.97666	29.263732	400	1	1.1	0.9;
	13474	1	-0	-0	0	2.257591	0	1.014746	17.839526	220	24	1.1	0.9;
	13475	1	25.03	12	0	2.76	0	0.998052	-0.623483	220	6	1.1	0.9;
	13476	1	152.7	81	0	17.54	0	1.005595	14.268483	380	3	1.1	0.9;
	13477	1	124	24.4	2.308381	123.32112	0	1.036114	25.111328	380	5	1.1	0.9;
	13478	1	36.74	14.1	0	16.98	0	1.00787	-25.080034	380	4	1.1	0.9;
	13479	1	-0	0	0.079602	10.290299	0	1.019038	-1.856624	220	6	1.1	0.9;
	13480	1	0	0	0	37.112979	0	0.998235	2.780849	380	5	1.1	0.9;
	13481	1	-0	-0	0	1.24	0	1.03499	9.763907	220	5	1.1	0.9;
	13482	2	0	0	0	0	0	1.062896	-0.392816	20	5	1.1	0.9;
	13483	1	0	0	0	4.13	0	1.001346	26.686522	150	8	1.1	0.9;
	13484	1	32.1	4.5	0	4.34	0	1.013127	27.002676	154	1	1.1	0.9;
	13485	1	-0	-0	0	15.15	0	1.051699	-5.861304	380	1	1.1	0.9;
	13486	2	0	0	0	0	0	1.062896	-0.392816	20	5	1.1	0.9;
	13487	2	0	0	0	0	0	1.03143	0.621932	20	6	1.1	0.9;
	13488	1	13.9	1.9	0	5.87	0	0.990162	46.115411	154	1	1.1	0.9;
	13489	2	0	0	0	0	0	1.061099	-1.375053	20	5	1.1	0.9;
	13490	1	-5.15	-2.04	0	154.352274	0	1.055526	12.967489	380	5	1.1	0.9;
	13491	2	0	0	0	0	0	1.026081	32.715698	20	5	1.1	0.9;
	13492	1	175.7	56.2	0	0.13	0	1.033125	23.713599	220	5	1.1	0.9;
	13493	2	0	0	0	0	0	0.97953	39.88697	20	1	1.1	0.9;
	13494	1	-0	0	0	32.64	0	0.997884	-4.981739	380	1	1.1	0.9;
	13495	1	38.84	9.6	0	0.88	0	1.023284	53.641998	110	8	1.1	0.9;
	13496	2	0	0	0	0	0	1.044924	29.406092	20	5	1.1	0.9;
	13497	2	0	0	0	0	0	1.044924	29.406092	20	5	1.1	0.9;
	13498	1	50.56	8.7	0	0.19	0	1.050728	28.162224	220	5	1.1	0.9;
	13499	1	39.5	5.6	0	8.26	0	1.037216	-1.577146	154	1	1.1	0.9;
	13500	2	0	0	0	0	0	1.044939	29.387061	20	5	1.1	0.9;
	13501	1	13.9	1.9	0	2.11	0	1.029091	2.944245	154	1	1.1	0.9;
	13502	2	0	0	0	0	0	1.044939	29.387061	20	5	1.1	0.9;
	13503	1	0	-0	0.471538	4.526699	0	1.037295	5.560158	380	6	1.1	0.9;
	13504	1	-0	-0	0	7.12	0	1.015643	-12.653637	380	17	1.1	0.9;
	13505	2	0	0	0	0	0	1.044924	29.406092	20	5	1.1	0.9;
	13506	2	0	0	0	0	0	1.044924	29.406092	20	5	1.1	0.9;
	13507	2	0	0	0	0	0	1.044924	29.406092	20	5	1.1	0.9;
	13508	1	21.72	13.4	0	2.38	0	1.002327	-2.090017	150	9	1.1	0.9;
	13509	1	169.04	33.9	0	5.48	0	0.979714	-25.778478	220	6	1.1	0.9;
	13510	2	0	0	0	0	0	1.016107	1.117381	11	1	1.1	0.9;
	13511	1	0	0	0	25.04	0	1.043446	0.547925	154	1	1.1	0.9;
	13512	2	0	0	0	0	0	1.030407	32.510933	20	5	1.1	0.9;
	13513	1	0	0	0	0	0	0.979705	-26.824454	220	10	1.1	0.9;
	13514	2	0	0	0	0	0	1.000449	37.338871	13.8	1	1.1	0.9;
	13515	1	16.7	2.4	0	2.38	0	0.99786	15.372787	154	1	1.1	0.9;
	13516	1	431.5	111.5	0	36.709841	0	1.004347	-20.932921	380	6	1.1	0.9;
	13517	1	158.4	31	0.079228	3.990772	0	1.042521	24.431977	220	5	1.1	0.9;
	13518	1	0	0	0	0	0	0.989728	-13.568538	150	9	1.1	0.9;
	13519	1	0	0	0	0	0	1.01361	-2.270783	150	9	1.1	0.9;
	13520	2	0	0	0	0	0	1.061099	-1.375053	20	5	1.1	0.9;
	13521	1	89.59	96.2	0	48.28	0	1.044479	35.825794	380	3	1.1	0.9;
	13522	1	-0	0	0.24842	10.168602	0	0.94985	0.99867	220	19	1.1	0.9;
	13523	1	66.57	23.6	0	0.14	0	1.053911	0.255931	110	19	1.1	0.9;
	13524	1	35.44	11.4	0	1.15	0	1.01268	36.766084	110	8	1.1	0.9;
	13525	2	0	0	0	0	0	1.025333	3.180104	20	6	1.1	0.9;
	13526	2	0	0	0	0	0	1.062073	84.08696	20	3	1.1	0.9;
	13527	1	15.22	5.1	0	0.36	0	1.068425	7.142111	110	18	1.2	0.8;
	13528	1	0	0	0	0	0	1.009202	-9.329154	150	9	1.1	0.9;
	13529	1	0	0	0.574367	0.152454	0	1.03256	16.651965	380	5	1.1	0.9;
	13530	2	0	0	0	0	0	1.042446	6.957021	20	22	1.1	0.9;
	13531	1	-0	0	0	0	0	1.028513	-22.307748	20	4	1.1	0.9;
	13532	1	-0	-0	0	0	0	1.003512	-19.342822	20	4	1.1	0.9;
	13533	1	-0	-0	0	0	0	1.003512	-19.342822	20	4	1.1	0.9;
	13534	1	5.5	0.8	0	8.36	0	1.015051	32.251738	154	1	1.1	0.9;
	13535	1	-0	-0	0	0	0	1.003512	-19.342822	20	4	1.1	0.9;
	13536	1	0	-0	0	17.63	0	1.040042	41.08762	380	3	1.1	0.9;
	13537	1	2.2	1.1	0	1.31	0	1.015472	-11.700252	110	22	1.1	0.9;
	13538	1	-0	0	0	0	0	1.008279	-10.478046	20	4	1.1	0.9;
	13539	1	83.19	37.3	0	28.16	0	1.023383	42.683688	380	3	1.1	0.9;
	13540	1	0	0	0	0.55	0	0.983404	1.560689	150	1	1.1	0.9;
	13541	1	99.5	9.2	0	27.55	0	1.025518	6.138633	220	5	1.1	0.9;
	13542	1	60.07	-0	0	0.04	0	1.01555	-10.855307	150	9	1.1	0.9;
	13543	1	3.04	-13.76	0	4.17	0	0.983074	-8.186805	150	9	1.1	0.9;
	13544	1	54.96	9.42	0	5.92	0	1.033426	15.128762	220	5	1.1	0.9;
	13545	1	0	-0	0	81.15	0	1.026198	-16.155637	380	4	1.1	0.9;
	13546	1	-0	0	0.12	1.55	0	1.025258	-3.745037	220	21	1.1	0.9;
	13547	1	311.4	310.5	0.18614	104.074947	0	1.040263	28.661705	380	3	1.1	0.9;
	13548	1	1.6	5.3	0.196409	7.603591	0	1.021257	7.312963	220	7	1.1	0.9;
	13549	2	0	0	0	0	0	0.996935	21.950327	10.6	1	1.1	0.9;
	13550	2	0	0	0	0	0	1.016735	24.349916	11	1	1.1	0.9;
	13551	2	0	0	0	0	0	0.984601	26.113188	6.3	1	1.1	0.9;
	13552	1	94.9	40.3	0	17.2	0	1.018662	-33.087457	220	4	1.1	0.9;
	13553	1	0	0	0	0	0	0.982498	-8.529449	150	9	1.1	0.9;
	13554	1	0	0	0	0	0	0.993427	-9.324273	150	9	1.1	0.9;
	13555	1	244.6	96.3	0	7.45	0	1.007864	11.047273	220	3	1.1	0.9;
	13556	2	0	0	0	0	0	0.994601	31.575691	11	1	1.1	0.9;
	13557	1	9.91	2	0	0.23	0	1.008244	-3.184184	110	18	1.1	0.9;
	13558	1	100.5	19.3	0	13.51	0	1.021067	36.65418	150	8	1.1	0.9;
	13559	1	-0	0	0.168013	0.409324	0	1.051912	5.608241	220	21	1.1	0.9;
	13560	2	0	0	0	0	0	1.033239	35.570464	20	10	1.1	0.9;
	13561	1	43.15	26.7	0	0.11	0	1.000535	-9.681265	150	9	1.1	0.9;
	13562	1	6.01	3	0	0.43	0	1.039357	-2.051947	110	23	1.1	0.9;
	13563	1	125.4	7.6	0.019458	7.490542	0	1.022641	24.05693	220	13	1.1	0.9;
	13564	1	10.11	2.8	0	1.01	0	1.005815	-2.908572	110	18	1.1	0.9;
	13565	1	66.07	15.6	0	5.57	0	1.024082	10.196492	220	5	1.1	0.9;
	13566	1	489.8	162	0.187289	98.738033	0	0.992072	-4.62442	380	5	1.1	0.9;
	13567	1	20.32	6.7	0	1.27	0	0.996488	-8.999786	110	12	1.1	0.9;
	13568	1	-0	0	0	1.28	0	1.033292	16.596766	220	5	1.1	0.9;
	13569	2	0	0	0	0	0	1.036065	-16.582357	20	4	1.1	0.9;
	13570	1	4.61	0.8	0	2.46	0	1.013683	-0.730425	110	19	1.1	0.9;
	13571	1	4.4	2.2	0	0.28	0	0.997383	-14.35601	110	22	1.1	0.9;
	13572	1	-0	0	0.077289	3.152711	0	1.030297	20.234769	220	5	1.1	0.9;
	13573	1	2.43	1.35	0	0.1	0	0.995147	19.31397	220	5	1.1	0.9;
	13574	1	29.7	4.3	0	13.67	0	1.025659	22.029811	154	1	1.1	0.9;
	13575	1	156.3	34.9	0	125.56	0	1.059983	-1.662639	380	20	1.1	0.9;
	13576	1	50.06	44	0	52.57	0	1.036622	7.110319	220	11	1.1	0.9;
	13577	2	0	0	0	0	0	1.034597	19.022639	20	6	1.1	0.9;
	13578	1	-28.33	9.3	0	0.06	0	1.001456	-8.619342	150	9	1.1	0.9;
	13579	1	-0	-0	0	0.35	0	1.071454	1.370559	150	9	1.1	0.9;
	13580	2	0	0	0	0	0	1.043178	23.372652	20	5	1.1	0.9;
	13581	1	62.57	-61	0	55.14	0	1.034873	4.848207	220	5	1.1	0.9;
	13582	2	0	0	0	0	0	1.043178	23.372652	20	5	1.1	0.9;
	13583	1	0	0	0.484386	7.703962	0	0.996638	13.542185	380	5	1.1	0.9;
	13584	2	0	0	0	0	0	1.030744	-0.382174	20	5	1.1	0.9;
	13585	2	0	0	0	0	0	1.04545	17.664921	20	5	1.1	0.9;
	13586	1	17.92	23.3	0	0	0	1.032358	23.741403	150	2	1.1	0.9;
	13587	1	-0	0	0	2.72	0	1.006208	20.83367	154	1	1.1	0.9;
	13588	1	-0	-0	0	1.15	0	1.032503	-3.384672	110	21	1.1	0.9;
	13589	2	0	0	0	0	0	1.04545	17.664921	20	5	1.1	0.9;
	13590	2	0	0	0	0	0	0.962614	39.578299	15.75	1	1.1	0.9;
	13591	1	15.82	3.7	0	0.22	0	1.000661	-5.473621	110	18	1.1	0.9;
	13592	2	0	0	0	0	0	1.051118	25.29251	20	14	1.1	0.9;
	13593	1	35.64	6.4	0.038822	1.491178	0	1.023635	0.441918	110	3	1.1	0.9;
	13594	2	0	0	0	0	0	1.045375	17.213843	20	5	1.1	0.9;
	13595	1	18.71	0.64	0	1.58	0	1.005996	-15.154422	110	22	1.1	0.9;
	13596	1	-0	0	0.059381	2.260413	0	1.01508	-20.654825	220	6	1.1	0.9;
	13597	2	0	0	0	0	0	1.045375	17.213843	20	5	1.1	0.9;
	13598	1	282.2	268.3	0	0.28	0	0.995589	33.909349	220	3	1.1	0.9;
	13599	1	14.42	3.8	0	1.24	0	0.91947	-10.968821	110	18	1.1	0.9;
	13600	1	-0	-0	0.176054	-0.106712	0	1.029652	34.356781	220	3	1.1	0.9;
	13601	2	0	0	0	0	0	1.050175	5.20925	20	5	1.1	0.9;
	13602	1	137.9	25.6	0	8.24	0	1.004626	-15.516883	220	4	1.1	0.9;
	13603	1	15.42	6.6	0	2.46	0	0.996849	-12.475537	110	22	1.1	0.9;
	13604	2	0	0	0	0	0	1.04545	17.664921	20	5	1.1	0.9;
	13605	1	-0	-0	0.162419	14.137581	0	0.990055	14.71774	380	5	1.1	0.9;
	13606	1	11.71	3.8	0	0.41	0	0.988767	-1.899151	110	18	1.1	0.9;
	13607	2	0	0	0	0	0	1.020265	-12.139526	20	4	1.1	0.9;
	13608	1	23.43	7.3	0	0.89	0	1.01076	44.146078	110	8	1.1	0.9;
	13609	2	0	0	0	0	0	1.030852	14.833561	20	5	1.1	0.9;
	13610	1	16.52	5.5	0	0.75	0	1.018465	-5.640469	110	22	1.1	0.9;
	13611	1	226.1	88.8	0	122.78	0	1.016297	30.173068	150	8	1.1	0.9;
	13612	1	0	0	0	13.84	0	0.98248	33.260777	400	1	1.1	0.9;
	13613	1	160.2	22.8	0	14.42	0	1.015753	16.87996	154	1	1.1	0.9;
	13614	1	-0	0	0	2.29	0	1.037073	30.465607	220	3	1.1	0.9;
	13615	1	11.44	2.38	0	0.78	0	0.989728	-4.296953	110	18	1.1	0.9;
	13616	1	12.11	4.4	0	2.35	0	0.937702	-1.59787	110	19	1.1	0.9;
	13617	2	0	0	0	0	0	0.991929	29.398008	6.3	1	1.1	0.9;
	13618	1	11.5	6.16	0	1.79	0	1.06601	-5.428594	150	9	1.1	0.9;
	13619	1	0	0	0.457274	6.947991	0	1.066104	33.9921	380	14	1.1	0.9;
	13620	2	0	0	0	0	0	1.034935	24.924619	20	10	1.1	0.9;
	13621	1	-0	-0	0	10.019046	0	0.994725	24.542662	380	5	1.1	0.9;
	13622	1	64.57	10	0	4.8	0	1.003125	-3.610696	110	18	1.1	0.9;
	13623	1	1.3	0	0	1.1	0	0.946151	-7.796967	110	18	1.1	0.9;
	13624	1	41.6	5.9	0	2.5	0	1.016603	17.55807	154	1	1.1	0.9;
	13625	1	103.6	4.7	0	36.68	0	1.030275	-20.984998	220	4	1.1	0.9;
	13626	1	187.4	72.8	0	34.95	0	1.013634	32.140775	150	8	1.1	0.9;
	13627	1	0	-0	0.048364	5.66206	0	1.051765	-1.298833	220	18	1.1	0.9;
	13628	1	39.34	13.3	0	0.67	0	0.948174	-7.95012	110	21	1.1	0.9;
	13629	1	139.8	-25.8	0	22.95	0	1.056217	14.209439	220	3	1.1	0.9;
	13630	1	26.4	3.8	0	7.48	0	1.030233	17.606169	154	1	1.1	0.9;
	13631	1	0	-0	0	6.58	0	1.035312	38.187653	380	3	1.1	0.9;
	13632	1	22.02	7.1	0	1.38	0	1.00299	1.259242	110	21	1.1	0.9;
	13633	2	0	0	0	0	0	1.03143	0.621932	20	6	1.1	0.9;
	13634	1	0	0	0	0.04	0	1.034538	19.040969	154	1	1.1	0.9;
	13635	1	-0	-0	0	5.98	0	1.020998	22.939343	220	3	1.1	0.9;
	13636	1	10.11	3.3	0	0.92	0	0.958613	-0.321963	110	18	1.1	0.9;
	13637	1	0	0	0	0	0	1.049406	-5.456139	150	9	1.1	0.9;
	13638	1	0	0	0.090774	11.940444	0	1.033627	4.718668	220	18	1.1	0.9;
	13639	2	0	0	0	0	0	0.980011	18.705058	20	18	1.1	0.9;
	13640	1	-0	0	0	6.95	0	1.001397	93.26676	380	1	1.1	0.9;
	13641	1	38.34	12.6	0	0.58	0	1.012393	-8.865545	110	12	1.1	0.9;
	13642	1	83.5	41.5	0	0	0	1.007442	-12.200575	150	9	1.1	0.9;
	13643	2	0	0	0	0	0	0.995077	34.268512	11	1	1.1	0.9;
	13644	1	11.41	3.7	0	0.63	0	0.979074	-10.789922	110	12	1.1	0.9;
	13645	1	-0	0	0.039946	0.25004	0	1.002037	0.742162	220	6	1.1	0.9;
	13646	1	0	-0	0.101868	70.579862	0	1.101093	10.056103	380	19	1.2	0.8;
	13647	1	157	44.2	0	22.63	0	1.048531	28.220134	220	1	1.1	0.9;
	13648	2	0	0	0	0	0	1.03143	0.621932	20	6	1.1	0.9;
	13649	1	2.1	0.8	0	0.42	0	0.993195	-1.493042	110	18	1.1	0.9;
	13650	1	-121.89	-52.61	0	0.48	0	1.038563	26.451559	220	13	1.1	0.9;
	13651	2	0	0	0	0	0	1.042166	-4.652525	20	9	1.1	0.9;
	13652	1	246.5	-1.9	0	31.37	0	1.01608	-6.699546	380	6	1.1	0.9;
	13653	1	0	0	0	1.27	0	1.02042	-1.807734	110	19	1.1	0.9;
	13654	2	0	0	0	0	0	1.079579	42.660061	380	13	1.2	0.8;
	13655	2	0	0	0	0	0	0.992821	91.559603	20	3	1.1	0.9;
	13656	1	0	-0	0.340862	0.564885	0	1.045517	35.805985	380	2	1.1	0.9;
	13657	1	-0	-0	0	3.66	0	0.962266	-9.928975	150	9	1.1	0.9;
	13658	2	0	0	0	0	0	1.038011	31.920163	20	10	1.1	0.9;
	13659	1	-0	0	0	0.36	0	1.038656	21.496915	220	5	1.1	0.9;
];

%% generator data
%	bus	Pg	Qg	Qmax	Qmin	Vg	mBase	status	Pmax	Pmin	Pc1	Pc2	Qc1min	Qc1max	Qc2min	Qc2max	ramp_agc	ramp_10	ramp_30	ramp_q	apf
mpc.gen = [
	1	42.33	15.72	40	-10	1.031695	100	1	100	28.22	0	0	0	0	0	0	0	0	0	0	0;
	12	238.58	-4.09	210	-50	1.004054	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	18	38.02	11.21	30	-10	1.043087	100	1	100	25.35	0	0	0	0	0	0	0	0	0	0	0;
	29	34.78	14.67	30	-10	1.033319	100	1	100	23.19	0	0	0	0	0	0	0	0	0	0	0;
	33	25.35	8.5	40	-10	1.018631	100	1	100	16.9	0	0	0	0	0	0	0	0	0	0	0;
	39	214.6	55.92	210	-50	1.073284	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	42	16.1	-4.12	20	-20	0.985213	100	1	100	10.73	0	0	0	0	0	0	0	0	0	0	0;
	45	192.55	1.87	160	-40	1.012099	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	51	161	0.83	90	-90	1.011416	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	54	37.08	5.51	30	-10	1.021005	100	1	100	24.72	0	0	0	0	0	0	0	0	0	0	0;
	62	617.7	178.65	520	-120	1.058268	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	63	617.7	178.65	520	-120	1.058268	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	66	179.12	14.03	160	-40	1.034768	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	68	53.72	11.91	40	-10	1.035023	100	1	200	35.81	0	0	0	0	0	0	0	0	0	0	0;
	69	53.72	11.91	40	-10	1.035023	100	1	200	35.81	0	0	0	0	0	0	0	0	0	0	0;
	77	24.65	6.97	30	-10	1.038344	100	1	100	16.43	0	0	0	0	0	0	0	0	0	0	0;
	78	54.48	10.37	30	-10	1.027383	100	1	200	36.32	0	0	0	0	0	0	0	0	0	0	0;
	79	18.13	3.85	30	-10	1.016716	100	1	100	12.09	0	0	0	0	0	0	0	0	0	0	0;
	86	152.7	32.92	80	-20	1.023881	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	94	17.64	3.05	20	-10	1.022611	100	1	100	11.76	0	0	0	0	0	0	0	0	0	0	0;
	96	49.42	5.32	40	-10	1.043225	100	1	100	32.95	0	0	0	0	0	0	0	0	0	0	0;
	97	49.42	5.32	40	-10	1.043225	100	1	100	32.95	0	0	0	0	0	0	0	0	0	0	0;
	98	49.42	5.32	40	-10	1.043225	100	1	100	32.95	0	0	0	0	0	0	0	0	0	0	0;
	99	23.85	6.18	40	-10	1.031885	100	1	100	15.9	0	0	0	0	0	0	0	0	0	0	0;
	100	23.85	6.18	40	-10	1.031885	100	1	100	15.9	0	0	0	0	0	0	0	0	0	0	0;
	101	49.42	5.32	40	-10	1.043225	100	1	100	32.95	0	0	0	0	0	0	0	0	0	0	0;
	102	49.42	5.32	40	-10	1.043225	100	1	100	32.95	0	0	0	0	0	0	0	0	0	0	0;
	104	87.95	47.73	60	-20	1.052257	100	1	200	58.63	0	0	0	0	0	0	0	0	0	0	0;
	107	68.26	9.75	60	-20	1.028153	100	1	200	45.51	0	0	0	0	0	0	0	0	0	0	0;
	110	91.47	10.42	60	-20	1.051233	100	1	200	60.98	0	0	0	0	0	0	0	0	0	0	0;
	114	42.35	6.03	40	-10	1.030832	100	1	100	28.23	0	0	0	0	0	0	0	0	0	0	0;
	115	22.35	5.33	30	-10	1.019	100	1	100	14.9	0	0	0	0	0	0	0	0	0	0	0;
	121	36.03	16.08	30	-10	1.045359	100	1	100	24.02	0	0	0	0	0	0	0	0	0	0	0;
	132	6.52	0.05	10	-10	0.996117	100	1	100	4.35	0	0	0	0	0	0	0	0	0	0	0;
	133	36.77	13.93	30	-10	1.042317	100	1	100	24.51	0	0	0	0	0	0	0	0	0	0	0;
	134	45.54	5.69	40	-10	1.044129	100	1	100	30.36	0	0	0	0	0	0	0	0	0	0	0;
	135	42.35	6.03	40	-10	1.030832	100	1	100	28.23	0	0	0	0	0	0	0	0	0	0	0;
	136	50.83	3.96	60	-20	1.049674	100	1	200	33.89	0	0	0	0	0	0	0	0	0	0	0;
	144	17.8	10.21	30	-10	1.034631	100	1	100	11.87	0	0	0	0	0	0	0	0	0	0	0;
	145	612.01	227.35	520	-120	1.040587	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	146	612.04	259.97	520	-120	1.044313	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	147	612.04	259.97	520	-120	1.044313	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	151	204	30.47	150	-60	1.023215	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	153	36.13	4.17	30	-10	1.027348	100	1	100	24.09	0	0	0	0	0	0	0	0	0	0	0;
	154	766.83	2.68	520	-120	1.05252	100	1	1600	500	0	0	0	0	0	0	0	0	0	0	0;
	156	49.34	9.15	40	-10	1.034597	100	1	100	32.89	0	0	0	0	0	0	0	0	0	0	0;
	161	49.34	9.15	40	-10	1.034597	100	1	100	32.89	0	0	0	0	0	0	0	0	0	0	0;
	162	49.34	9.15	40	-10	1.034597	100	1	100	32.89	0	0	0	0	0	0	0	0	0	0	0;
	165	42.35	6.03	40	-10	1.030832	100	1	100	28.23	0	0	0	0	0	0	0	0	0	0	0;
	179	95.02	8.09	60	-20	1.023144	100	1	200	63.35	0	0	0	0	0	0	0	0	0	0	0;
	182	34.96	12.97	30	-10	1.054937	100	1	100	23.31	0	0	0	0	0	0	0	0	0	0	0;
	196	81.07	20.71	60	-20	1.021915	100	1	200	54.05	0	0	0	0	0	0	0	0	0	0	0;
	197	245.73	1.48	160	-40	1.005341	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	207	27.95	4.02	30	-10	1.016461	100	1	100	18.63	0	0	0	0	0	0	0	0	0	0	0;
	220	97.72	38.87	60	-20	1.055654	100	1	200	65.15	0	0	0	0	0	0	0	0	0	0	0;
	221	97.72	38.87	60	-20	1.055654	100	1	200	65.15	0	0	0	0	0	0	0	0	0	0	0;
	222	97.66	41.74	60	-20	1.051105	100	1	200	65.11	0	0	0	0	0	0	0	0	0	0	0;
	223	97.72	38.87	60	-20	1.055654	100	1	200	65.15	0	0	0	0	0	0	0	0	0	0	0;
	224	97.66	41.74	60	-20	1.051105	100	1	200	65.11	0	0	0	0	0	0	0	0	0	0	0;
	225	97.66	41.74	60	-20	1.051105	100	1	200	65.11	0	0	0	0	0	0	0	0	0	0	0;
	229	21.03	9.77	30	-10	1.034441	100	1	100	14.02	0	0	0	0	0	0	0	0	0	0	0;
	241	601.87	179.7	520	-120	1.037622	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	243	32.43	17.87	30	-10	1.053647	100	1	100	21.62	0	0	0	0	0	0	0	0	0	0	0;
	248	26.38	8.7	30	-10	1.026279	100	1	100	17.59	0	0	0	0	0	0	0	0	0	0	0;
	250	192.55	1.87	160	-40	1.012099	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	258	8.2	-2.56	10	-10	0.955439	100	1	100	5.47	0	0	0	0	0	0	0	0	0	0	0;
	259	38.51	8.85	30	-10	1.03425	100	1	100	25.67	0	0	0	0	0	0	0	0	0	0	0;
	266	28.48	6.95	30	-10	1.005915	100	1	100	18.99	0	0	0	0	0	0	0	0	0	0	0;
	270	18.4	-1.01	20	-20	1.003342	100	1	100	12.27	0	0	0	0	0	0	0	0	0	0	0;
	271	626.63	195.6	520	-120	1.053159	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	273	626.63	190.1	520	-120	1.051456	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	274	626.64	188.59	520	-120	1.051088	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	275	626.64	188.59	520	-120	1.051088	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	276	626.64	187.14	520	-120	1.050379	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	278	627	28.92	320	-190	0.960771	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	283	8	0.73	10	-10	1.000621	100	1	100	5.33	0	0	0	0	0	0	0	0	0	0	0;
	289	36.9	11.76	30	-10	1.040289	100	1	100	24.6	0	0	0	0	0	0	0	0	0	0	0;
	294	92.8	3.25	50	-50	0.999187	100	1	200	61.87	0	0	0	0	0	0	0	0	0	0	0;
	296	21	0.7	10	-10	0.989117	100	1	100	14	0	0	0	0	0	0	0	0	0	0	0;
	299	158.9	35.36	100	-30	0.993468	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	304	39.98	8.7	30	-10	1.037997	100	1	100	26.65	0	0	0	0	0	0	0	0	0	0	0;
	305	27.73	2	30	-10	1.030407	100	1	100	18.49	0	0	0	0	0	0	0	0	0	0	0;
	306	27.73	2	30	-10	1.030407	100	1	100	18.49	0	0	0	0	0	0	0	0	0	0	0;
	307	17.66	6.95	30	-10	1.049321	100	1	100	11.77	0	0	0	0	0	0	0	0	0	0	0;
	308	17.66	6.95	30	-10	1.049321	100	1	100	11.77	0	0	0	0	0	0	0	0	0	0	0;
	309	32.08	3.64	30	-10	1.05655	100	1	100	21.39	0	0	0	0	0	0	0	0	0	0	0;
	310	32.08	3.64	30	-10	1.05655	100	1	100	21.39	0	0	0	0	0	0	0	0	0	0	0;
	311	32.08	3.64	30	-10	1.05655	100	1	100	21.39	0	0	0	0	0	0	0	0	0	0	0;
	312	32.08	3.64	30	-10	1.05655	100	1	100	21.39	0	0	0	0	0	0	0	0	0	0	0;
	315	2.06	0.62	10	-10	1.021202	100	1	100	1.37	0	0	0	0	0	0	0	0	0	0	0;
	323	419.13	110.83	520	-120	1.016993	100	1	900	200	0	0	0	0	0	0	0	0	0	0	0;
	324	5	1.91	10	-10	1.024179	100	1	100	3.33	0	0	0	0	0	0	0	0	0	0	0;
	325	4.8	-0.12	10	-10	0.996355	100	1	100	3.2	0	0	0	0	0	0	0	0	0	0	0;
	328	54.54	27.56	40	-10	1.023168	100	1	200	36.36	0	0	0	0	0	0	0	0	0	0	0;
	331	27.62	3.89	30	-20	1.006233	100	1	100	18.41	0	0	0	0	0	0	0	0	0	0	0;
	332	2	-0	10	-10	0.996766	100	1	100	1.33	0	0	0	0	0	0	0	0	0	0	0;
	336	16.08	0.01	10	-10	1.010862	100	1	100	10.72	0	0	0	0	0	0	0	0	0	0	0;
	338	16.08	-0.17	10	-10	1.00038	100	1	100	10.72	0	0	0	0	0	0	0	0	0	0	0;
	345	333.1	57.13	210	-50	1.030395	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	346	333.1	57.13	210	-50	1.030395	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	351	167.59	39.08	210	-50	1.024484	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	354	36.42	7.96	30	-10	1.029166	100	1	100	24.28	0	0	0	0	0	0	0	0	0	0	0;
	358	165.24	104.62	210	-50	1.050725	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	360	84.12	16.88	60	-20	1.028641	100	1	200	56.08	0	0	0	0	0	0	0	0	0	0	0;
	361	84.12	16.88	60	-20	1.028641	100	1	200	56.08	0	0	0	0	0	0	0	0	0	0	0;
	362	84.12	16.88	60	-20	1.028641	100	1	200	56.08	0	0	0	0	0	0	0	0	0	0	0;
	363	84.12	16.88	60	-20	1.028641	100	1	200	56.08	0	0	0	0	0	0	0	0	0	0	0;
	364	5.25	0.21	10	-10	1.00564	100	1	100	3.5	0	0	0	0	0	0	0	0	0	0	0;
	365	95.2	36.17	110	-30	0.997585	100	1	200	63.47	0	0	0	0	0	0	0	0	0	0	0;
	368	84.12	16.88	60	-20	1.028641	100	1	200	56.08	0	0	0	0	0	0	0	0	0	0	0;
	369	84.12	16.88	60	-20	1.028641	100	1	200	56.08	0	0	0	0	0	0	0	0	0	0	0;
	382	102.03	10.64	60	-20	0.990596	100	1	300	68.02	0	0	0	0	0	0	0	0	0	0	0;
	383	102.24	30.32	60	-20	1.012482	100	1	300	68.16	0	0	0	0	0	0	0	0	0	0	0;
	392	1.03	0.09	10	-10	1.030636	100	1	100	0.69	0	0	0	0	0	0	0	0	0	0	0;
	394	37.9	13.47	30	-10	1.0326	100	1	100	25.27	0	0	0	0	0	0	0	0	0	0	0;
	395	37.23	4.01	30	-10	1.028283	100	1	100	24.82	0	0	0	0	0	0	0	0	0	0	0;
	396	85.28	21.31	60	-20	1.029542	100	1	200	56.85	0	0	0	0	0	0	0	0	0	0	0;
	400	94.5	8.92	50	-40	0.979015	100	1	200	63	0	0	0	0	0	0	0	0	0	0	0;
	405	33.64	6.9	30	-10	1.00773	100	1	100	22.43	0	0	0	0	0	0	0	0	0	0	0;
	406	33.64	6.9	30	-10	1.00773	100	1	100	22.43	0	0	0	0	0	0	0	0	0	0	0;
	408	34.92	9.74	30	-10	1.043492	100	1	100	23.28	0	0	0	0	0	0	0	0	0	0	0;
	423	0.99	-0.05	10	-10	1.015919	100	1	100	0.66	0	0	0	0	0	0	0	0	0	0	0;
	432	4.5	0.47	10	-10	1.016602	100	1	100	3	0	0	0	0	0	0	0	0	0	0	0;
	433	30.04	7.71	30	-10	1.011489	100	1	100	20.03	0	0	0	0	0	0	0	0	0	0	0;
	438	18.13	3.85	30	-10	1.016716	100	1	100	12.09	0	0	0	0	0	0	0	0	0	0	0;
	442	31.35	11.33	30	-10	1.043357	100	1	100	20.9	0	0	0	0	0	0	0	0	0	0	0;
	448	25.22	6.35	20	-10	1.038795	100	1	100	16.81	0	0	0	0	0	0	0	0	0	0	0;
	454	8.35	-1.53	10	-10	0.973833	100	1	100	5.57	0	0	0	0	0	0	0	0	0	0	0;
	458	31.28	-0.92	30	-10	1.009892	100	1	100	20.85	0	0	0	0	0	0	0	0	0	0	0;
	463	-19.98	270.34	100000	-100000	1.126642	100	1	100	-26.64	0	0	0	0	0	0	0	0	0	0	0;
	470	74.8	42.23	60	-20	1.046905	100	1	200	49.87	0	0	0	0	0	0	0	0	0	0	0;
	476	214.05	76.67	210	-50	1.029484	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	477	38.02	11.21	30	-10	1.043087	100	1	100	25.35	0	0	0	0	0	0	0	0	0	0	0;
	480	89.2	18.25	60	-20	1.033425	100	1	200	59.47	0	0	0	0	0	0	0	0	0	0	0;
	481	89.2	18.25	60	-20	1.033425	100	1	200	59.47	0	0	0	0	0	0	0	0	0	0	0;
	482	89.2	18.25	60	-20	1.033425	100	1	200	59.47	0	0	0	0	0	0	0	0	0	0	0;
	483	89.2	18.25	60	-20	1.033425	100	1	200	59.47	0	0	0	0	0	0	0	0	0	0	0;
	485	325.03	139.06	210	-50	1.059619	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	492	273.28	112.89	210	-50	1.041985	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	496	64.28	29.97	60	-20	1.00876	100	1	200	42.85	0	0	0	0	0	0	0	0	0	0	0;
	508	857.64	276.43	520	-120	1.044776	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	509	857.69	282.91	520	-120	1.044388	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	510	857.66	271.49	520	-120	1.044096	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	514	857.64	273.87	520	-120	1.044511	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	516	32.57	7.29	30	-10	1.041748	100	1	100	21.71	0	0	0	0	0	0	0	0	0	0	0;
	528	23.66	2.76	30	-10	1.020703	100	1	100	15.77	0	0	0	0	0	0	0	0	0	0	0;
	531	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	532	258.95	65.97	210	-50	1.036699	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	533	164.92	6.68	160	-40	1.016377	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	539	152	33.4	90	-40	0.993492	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	552	30.6	-1.46	20	-20	0.99834	100	1	100	20.4	0	0	0	0	0	0	0	0	0	0	0;
	559	51.5	-7.54	30	-30	0.954584	100	1	200	34.33	0	0	0	0	0	0	0	0	0	0	0;
	564	21.77	6.35	30	-10	1.024499	100	1	100	14.51	0	0	0	0	0	0	0	0	0	0	0;
	572	7	0.9	10	-10	1.004314	100	1	100	4.67	0	0	0	0	0	0	0	0	0	0	0;
	575	18.09	1.85	30	-10	1.016739	100	1	100	12.06	0	0	0	0	0	0	0	0	0	0	0;
	576	18.09	1.85	30	-10	1.016739	100	1	100	12.06	0	0	0	0	0	0	0	0	0	0	0;
	577	18.09	1.85	30	-10	1.016739	100	1	100	12.06	0	0	0	0	0	0	0	0	0	0	0;
	581	18.09	1.85	30	-10	1.016739	100	1	100	12.06	0	0	0	0	0	0	0	0	0	0	0;
	585	72.61	29.56	60	-20	1.041629	100	1	200	48.41	0	0	0	0	0	0	0	0	0	0	0;
	589	13.38	0.68	10	-10	1.00755	100	1	100	8.92	0	0	0	0	0	0	0	0	0	0	0;
	610	31.95	15.26	30	-10	1.019723	100	1	100	21.3	0	0	0	0	0	0	0	0	0	0	0;
	614	34.81	16.31	40	-10	1.031902	100	1	100	23.21	0	0	0	0	0	0	0	0	0	0	0;
	615	34.81	16.31	40	-10	1.031902	100	1	100	23.21	0	0	0	0	0	0	0	0	0	0	0;
	619	42.5	5.74	40	-10	1.031502	100	1	100	28.33	0	0	0	0	0	0	0	0	0	0	0;
	621	5.8	-0.78	10	-10	1.003942	100	1	100	3.87	0	0	0	0	0	0	0	0	0	0	0;
	623	36.03	16.08	30	-10	1.045359	100	1	100	24.02	0	0	0	0	0	0	0	0	0	0	0;
	626	43.29	23.54	60	-20	1.035771	100	1	100	28.86	0	0	0	0	0	0	0	0	0	0	0;
	635	0.13	-0.1	10	-10	1.023511	100	1	100	0.09	0	0	0	0	0	0	0	0	0	0	0;
	637	25.22	5.17	30	-10	1.04616	100	1	100	16.81	0	0	0	0	0	0	0	0	0	0	0;
	646	333.1	57.13	210	-50	1.030395	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	648	322.05	10.4	210	-50	1.007781	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	649	333.1	57.13	210	-50	1.030395	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	651	188.27	-7.75	210	-50	1.028351	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	652	188.27	-7.75	210	-50	1.028351	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	654	92.4	2.72	60	-20	1.014507	100	1	200	61.6	0	0	0	0	0	0	0	0	0	0	0;
	660	33.21	5.47	30	-10	1.022689	100	1	100	22.14	0	0	0	0	0	0	0	0	0	0	0;
	666	31.3	-0.77	30	-10	1.018739	100	1	100	20.87	0	0	0	0	0	0	0	0	0	0	0;
	671	25.2	-0.32	20	-10	0.999061	100	1	100	16.8	0	0	0	0	0	0	0	0	0	0	0;
	676	274.65	30.64	160	-40	1.018066	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	681	84.68	7.94	60	-20	1.00657	100	1	200	56.45	0	0	0	0	0	0	0	0	0	0	0;
	684	68.89	19.23	60	-20	1.035618	100	1	200	45.93	0	0	0	0	0	0	0	0	0	0	0;
	693	209.96	80.03	160	-40	1.038999	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	695	8.4	-0.84	10	-10	1.010516	100	1	100	5.6	0	0	0	0	0	0	0	0	0	0	0;
	698	26.72	4.41	20	-10	1.018527	100	1	100	17.81	0	0	0	0	0	0	0	0	0	0	0;
	709	626.68	236.63	520	-120	1.051727	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	715	23.9	3.61	30	-10	1.007844	100	1	100	15.93	0	0	0	0	0	0	0	0	0	0	0;
	717	23.12	5.58	10	0	1.027572	100	1	100	15.41	0	0	0	0	0	0	0	0	0	0	0;
	721	74.35	10.28	60	-20	0.992442	100	1	200	49.57	0	0	0	0	0	0	0	0	0	0	0;
	724	35.5	10.89	30	-10	1.045213	100	1	100	23.67	0	0	0	0	0	0	0	0	0	0	0;
	730	98.46	12.34	60	-20	1.01458	100	1	200	65.64	0	0	0	0	0	0	0	0	0	0	0;
	731	98.46	12.34	60	-20	1.01458	100	1	200	65.64	0	0	0	0	0	0	0	0	0	0	0;
	732	98.46	12.34	60	-20	1.01458	100	1	200	65.64	0	0	0	0	0	0	0	0	0	0	0;
	734	98.46	12.34	60	-20	1.01458	100	1	200	65.64	0	0	0	0	0	0	0	0	0	0	0;
	746	16	0.19	10	-10	0.994194	100	1	100	10.67	0	0	0	0	0	0	0	0	0	0	0;
	749	22.58	9.34	30	-10	1.009791	100	1	100	15.05	0	0	0	0	0	0	0	0	0	0	0;
	750	22.58	9.34	30	-10	1.009791	100	1	100	15.05	0	0	0	0	0	0	0	0	0	0	0;
	751	22.58	9.34	30	-10	1.009791	100	1	100	15.05	0	0	0	0	0	0	0	0	0	0	0;
	753	20.64	7.21	30	-10	1.018866	100	1	100	13.76	0	0	0	0	0	0	0	0	0	0	0;
	754	64.65	13.16	60	-20	1.009683	100	1	200	43.1	0	0	0	0	0	0	0	0	0	0	0;
	765	78.37	30.34	60	-20	1.032985	100	1	200	52.25	0	0	0	0	0	0	0	0	0	0	0;
	768	46.74	7.22	30	-10	1.044527	100	1	100	31.16	0	0	0	0	0	0	0	0	0	0	0;
	769	46.74	7.22	30	-10	1.044527	100	1	100	31.16	0	0	0	0	0	0	0	0	0	0	0;
	773	30.13	8.92	30	-10	1.016081	100	1	100	20.09	0	0	0	0	0	0	0	0	0	0	0;
	775	26.34	0.75	20	-10	1.003456	100	1	100	17.56	0	0	0	0	0	0	0	0	0	0	0;
	778	18.74	6.56	30	-10	1.02868	100	1	100	12.49	0	0	0	0	0	0	0	0	0	0	0;
	779	21.6	0.33	10	-10	0.9829	100	1	100	14.4	0	0	0	0	0	0	0	0	0	0	0;
	783	16.28	8.5	30	-10	1.027518	100	1	100	10.85	0	0	0	0	0	0	0	0	0	0	0;
	786	31.29	-0.48	30	-30	1.013792	100	1	100	20.86	0	0	0	0	0	0	0	0	0	0	0;
	798	30.9	1.92	30	-10	1.015954	100	1	100	20.6	0	0	0	0	0	0	0	0	0	0	0;
	800	612.01	227.35	520	-120	1.040587	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	809	31.3	3.8	20	-10	1.023128	100	1	100	20.87	0	0	0	0	0	0	0	0	0	0	0;
	811	31.72	4.48	30	-10	1.022192	100	1	100	21.15	0	0	0	0	0	0	0	0	0	0	0;
	814	951.82	32.74	520	-120	1.01249	100	1	2000	600	0	0	0	0	0	0	0	0	0	0	0;
	815	92.69	14.48	60	-20	1.026207	100	1	200	61.79	0	0	0	0	0	0	0	0	0	0	0;
	819	219.68	138.48	160	-40	1.048632	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	824	31.3	-0.77	30	-10	1.018739	100	1	100	20.87	0	0	0	0	0	0	0	0	0	0	0;
	828	224.42	24.73	210	-50	1.01121	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	836	19.15	4.15	20	-10	1.015461	100	1	100	12.77	0	0	0	0	0	0	0	0	0	0	0;
	839	7.54	0.74	10	-10	1.0079	100	1	100	5.03	0	0	0	0	0	0	0	0	0	0	0;
	841	96.03	17.22	60	-20	1.036065	100	1	200	64.02	0	0	0	0	0	0	0	0	0	0	0;
	844	59.02	10.61	40	-10	1.059427	100	1	200	39.35	0	0	0	0	0	0	0	0	0	0	0;
	845	59.02	10.61	40	-10	1.059427	100	1	200	39.35	0	0	0	0	0	0	0	0	0	0	0;
	846	59.02	10.61	40	-10	1.059427	100	1	200	39.35	0	0	0	0	0	0	0	0	0	0	0;
	849	68.26	9.75	60	-20	1.028153	100	1	200	45.51	0	0	0	0	0	0	0	0	0	0	0;
	855	39.67	13.3	30	-10	1.046298	100	1	100	26.45	0	0	0	0	0	0	0	0	0	0	0;
	858	257.94	36.73	160	-40	1.021858	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	859	36.91	13.33	30	-10	1.037973	100	1	100	24.61	0	0	0	0	0	0	0	0	0	0	0;
	861	30.28	3.51	30	-10	1.018125	100	1	100	20.19	0	0	0	0	0	0	0	0	0	0	0;
	862	30.28	3.51	30	-10	1.018125	100	1	100	20.19	0	0	0	0	0	0	0	0	0	0	0;
	864	30.28	3.51	30	-10	1.018125	100	1	100	20.19	0	0	0	0	0	0	0	0	0	0	0;
	865	30.28	3.51	30	-10	1.018125	100	1	100	20.19	0	0	0	0	0	0	0	0	0	0	0;
	869	37.9	13.47	30	-10	1.0326	100	1	100	25.27	0	0	0	0	0	0	0	0	0	0	0;
	873	37.97	8.2	40	-10	0.994173	100	1	100	25.31	0	0	0	0	0	0	0	0	0	0	0;
	874	40.23	5.24	30	-10	1.006137	100	1	100	26.82	0	0	0	0	0	0	0	0	0	0	0;
	888	38.16	4.12	30	-10	1.055337	100	1	100	25.44	0	0	0	0	0	0	0	0	0	0	0;
	898	35.97	5.34	30	-10	1.058434	100	1	100	23.98	0	0	0	0	0	0	0	0	0	0	0;
	901	84.68	7.94	60	-20	1.00657	100	1	200	56.45	0	0	0	0	0	0	0	0	0	0	0;
	904	32.52	9.41	30	-10	1.028707	100	1	100	21.68	0	0	0	0	0	0	0	0	0	0	0;
	907	32.52	9.41	30	-10	1.028707	100	1	100	21.68	0	0	0	0	0	0	0	0	0	0	0;
	910	88.32	56.36	60	-20	1.127961	100	1	200	58.88	0	0	0	0	0	0	0	0	0	0	0;
	912	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	916	159.8	-8.66	100	-40	0.957523	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	920	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	921	274.65	30.64	160	-40	1.018066	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	925	266.84	38.78	210	-50	1.032265	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	926	245.73	1.48	160	-40	1.005341	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	927	43.5	1.21	30	-30	1.007757	100	1	100	29	0	0	0	0	0	0	0	0	0	0	0;
	930	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	932	19.57	7.03	30	-10	1.023156	100	1	100	13.05	0	0	0	0	0	0	0	0	0	0	0;
	937	24.63	12.17	30	-10	1.026822	100	1	100	16.42	0	0	0	0	0	0	0	0	0	0	0;
	938	28.83	4.62	30	-10	1.019144	100	1	100	19.22	0	0	0	0	0	0	0	0	0	0	0;
	954	60.3	-12.67	30	-30	0.961175	100	1	200	40.2	0	0	0	0	0	0	0	0	0	0	0;
	956	20.94	5.59	30	-10	1.044097	100	1	100	13.96	0	0	0	0	0	0	0	0	0	0	0;
	958	31.5	1.72	30	-20	1.016735	100	1	100	21	0	0	0	0	0	0	0	0	0	0	0;
	975	16.74	3.04	30	-10	1.042545	100	1	100	11.16	0	0	0	0	0	0	0	0	0	0	0;
	976	16.62	7.24	30	-10	1.070058	100	1	100	11.08	0	0	0	0	0	0	0	0	0	0	0;
	977	16.62	7.24	30	-10	1.070058	100	1	100	11.08	0	0	0	0	0	0	0	0	0	0	0;
	978	16.62	7.24	30	-10	1.070058	100	1	100	11.08	0	0	0	0	0	0	0	0	0	0	0;
	980	30.87	9.69	30	-10	1.023796	100	1	100	20.58	0	0	0	0	0	0	0	0	0	0	0;
	981	38.33	9.23	30	-10	1.018701	100	1	100	25.55	0	0	0	0	0	0	0	0	0	0	0;
	982	38.33	9.23	30	-10	1.018701	100	1	100	25.55	0	0	0	0	0	0	0	0	0	0	0;
	983	38.33	9.23	30	-10	1.018701	100	1	100	25.55	0	0	0	0	0	0	0	0	0	0	0;
	984	19	4.14	30	-10	1.050192	100	1	100	12.67	0	0	0	0	0	0	0	0	0	0	0;
	985	19	4.14	30	-10	1.050192	100	1	100	12.67	0	0	0	0	0	0	0	0	0	0	0;
	986	19	4.14	30	-10	1.050192	100	1	100	12.67	0	0	0	0	0	0	0	0	0	0	0;
	987	19	4.14	30	-10	1.050192	100	1	100	12.67	0	0	0	0	0	0	0	0	0	0	0;
	995	37.23	4.01	30	-10	1.028283	100	1	100	24.82	0	0	0	0	0	0	0	0	0	0	0;
	1002	223.3	11.76	160	-40	1.026117	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	1004	4.2	1.44	10	-10	1.020732	100	1	100	2.8	0	0	0	0	0	0	0	0	0	0	0;
	1007	15	2.18	10	-10	1.009902	100	1	100	10	0	0	0	0	0	0	0	0	0	0	0;
	1010	86.25	-3.04	60	-80	0.962614	100	1	200	57.5	0	0	0	0	0	0	0	0	0	0	0;
	1013	35.97	9.16	30	-10	1.001252	100	1	100	23.98	0	0	0	0	0	0	0	0	0	0	0;
	1031	89.66	46.76	60	-20	1.050891	100	1	200	59.77	0	0	0	0	0	0	0	0	0	0	0;
	1043	50.92	11.1	40	-10	1.010998	100	1	200	33.95	0	0	0	0	0	0	0	0	0	0	0;
	1044	5.25	0.74	10	-10	1.019396	100	1	100	3.5	0	0	0	0	0	0	0	0	0	0	0;
	1046	8.54	0.6	10	-10	1.014635	100	1	100	5.69	0	0	0	0	0	0	0	0	0	0	0;
	1048	0.34	-0.09	10	-10	1.024095	100	1	100	0.23	0	0	0	0	0	0	0	0	0	0	0;
	1050	348.55	140.56	210	-50	1.048227	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	1055	34.92	9.74	30	-10	1.043492	100	1	100	23.28	0	0	0	0	0	0	0	0	0	0	0;
	1057	52.81	50.16	60	-20	1.055479	100	1	200	35.21	0	0	0	0	0	0	0	0	0	0	0;
	1059	36.8	-2.54	30	-30	0.98379	100	1	100	24.53	0	0	0	0	0	0	0	0	0	0	0;
	1070	23.01	6.63	30	-10	1.01964	100	1	100	15.34	0	0	0	0	0	0	0	0	0	0	0;
	1071	23.6	10.13	30	-10	1.031397	100	1	100	15.73	0	0	0	0	0	0	0	0	0	0	0;
	1072	89.91	48.94	60	-20	1.065506	100	1	200	59.94	0	0	0	0	0	0	0	0	0	0	0;
	1073	39.81	14.61	30	-10	1.037189	100	1	100	26.54	0	0	0	0	0	0	0	0	0	0	0;
	1074	80.74	25.03	60	-20	1.061099	100	1	200	53.83	0	0	0	0	0	0	0	0	0	0	0;
	1077	23.3	5.11	30	-10	1.015556	100	1	100	15.53	0	0	0	0	0	0	0	0	0	0	0;
	1079	36.7	9.55	30	-10	1.037501	100	1	100	24.47	0	0	0	0	0	0	0	0	0	0	0;
	1083	206.32	45.91	160	-40	1.035499	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	1085	261.4	61.69	100	-30	0.98883	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	1088	0.07	-0.16	10	-10	1.015045	100	1	100	0.05	0	0	0	0	0	0	0	0	0	0	0;
	1089	26.24	8.12	30	-10	1.039754	100	1	100	17.49	0	0	0	0	0	0	0	0	0	0	0;
	1090	19.85	-1.9	30	-10	1.026413	100	1	100	13.23	0	0	0	0	0	0	0	0	0	0	0;
	1092	26.24	8.12	30	-10	1.039754	100	1	100	17.49	0	0	0	0	0	0	0	0	0	0	0;
	1093	26.24	8.12	30	-10	1.039754	100	1	100	17.49	0	0	0	0	0	0	0	0	0	0	0;
	1096	37.29	8.19	30	-10	1.030566	100	1	100	24.86	0	0	0	0	0	0	0	0	0	0	0;
	1104	34.92	9.74	30	-10	1.043492	100	1	100	23.28	0	0	0	0	0	0	0	0	0	0	0;
	1123	100.41	11.92	60	-20	1.038747	100	1	300	66.94	0	0	0	0	0	0	0	0	0	0	0;
	1125	31.3	-0.77	30	-10	1.018739	100	1	100	20.87	0	0	0	0	0	0	0	0	0	0	0;
	1138	290	-23.36	100	-100	0.971901	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	1148	9	0.72	10	-10	1.020087	100	1	100	6	0	0	0	0	0	0	0	0	0	0	0;
	1150	240.6	37.13	100	-30	0.970854	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	1180	25.2	-0.32	20	-10	0.999061	100	1	100	16.8	0	0	0	0	0	0	0	0	0	0	0;
	1182	23.37	7.44	30	-10	1.056688	100	1	100	15.58	0	0	0	0	0	0	0	0	0	0	0;
	1183	23.37	7.44	30	-10	1.056688	100	1	100	15.58	0	0	0	0	0	0	0	0	0	0	0;
	1189	27.96	5.11	30	-10	1.029009	100	1	100	18.64	0	0	0	0	0	0	0	0	0	0	0;
	1192	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	1202	35.5	10.89	30	-10	1.045213	100	1	100	23.67	0	0	0	0	0	0	0	0	0	0	0;
	1208	20	-2.63	30	-10	1.010291	100	1	100	13.33	0	0	0	0	0	0	0	0	0	0	0;
	1218	181.73	12.69	160	-40	0.989939	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	1238	6.55	-0.21	10	-10	1.003807	100	1	100	4.37	0	0	0	0	0	0	0	0	0	0	0;
	1245	17.6	0.07	20	-20	0.977562	100	1	100	11.73	0	0	0	0	0	0	0	0	0	0	0;
	1257	56.25	11.12	40	-10	1.03036	100	1	200	37.5	0	0	0	0	0	0	0	0	0	0	0;
	1262	219.68	138.48	160	-40	1.048632	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	1263	241.27	5.04	160	-40	1.017546	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	1264	270	-15.65	100	-100	1.002671	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	1273	17.95	10.38	30	-10	1.044317	100	1	100	11.97	0	0	0	0	0	0	0	0	0	0	0;
	1274	96.03	17.22	60	-20	1.036065	100	1	200	64.02	0	0	0	0	0	0	0	0	0	0	0;
	1278	38.16	4.12	30	-10	1.055337	100	1	100	25.44	0	0	0	0	0	0	0	0	0	0	0;
	1287	27.88	10.37	30	-10	1.052164	100	1	100	18.59	0	0	0	0	0	0	0	0	0	0	0;
	1288	27.2	3.82	30	-10	1.024762	100	1	100	18.13	0	0	0	0	0	0	0	0	0	0	0;
	1290	20.55	5.71	30	-10	1.019482	100	1	100	13.7	0	0	0	0	0	0	0	0	0	0	0;
	1298	27.88	10.37	30	-10	1.052164	100	1	100	18.59	0	0	0	0	0	0	0	0	0	0	0;
	1304	36.91	13.33	30	-10	1.037973	100	1	100	24.61	0	0	0	0	0	0	0	0	0	0	0;
	1305	24.84	4.52	30	-10	1.022099	100	1	100	16.56	0	0	0	0	0	0	0	0	0	0	0;
	1306	95.02	8.09	60	-20	1.023144	100	1	200	63.35	0	0	0	0	0	0	0	0	0	0	0;
	1308	264.57	207.6	210	-50	1.127182	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	1310	81.31	8.06	60	-20	1.043619	100	1	200	54.21	0	0	0	0	0	0	0	0	0	0	0;
	1323	353.02	203.37	210	-50	1.128032	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	1324	38.16	4.12	30	-10	1.055337	100	1	100	25.44	0	0	0	0	0	0	0	0	0	0	0;
	1329	5	2.04	10	-10	1.009821	100	1	100	3.33	0	0	0	0	0	0	0	0	0	0	0;
	1330	36.11	3.87	30	-10	1.024544	100	1	100	24.07	0	0	0	0	0	0	0	0	0	0	0;
	1335	95.51	37.64	60	-20	1.0608	100	1	200	63.67	0	0	0	0	0	0	0	0	0	0	0;
	1338	25.65	3.95	30	-10	1.02611	100	1	100	17.1	0	0	0	0	0	0	0	0	0	0	0;
	1339	25.65	3.95	30	-10	1.02611	100	1	100	17.1	0	0	0	0	0	0	0	0	0	0	0;
	1342	25.65	3.95	30	-10	1.02611	100	1	100	17.1	0	0	0	0	0	0	0	0	0	0	0;
	1343	25.65	3.95	30	-10	1.02611	100	1	100	17.1	0	0	0	0	0	0	0	0	0	0	0;
	1344	84.68	7.94	60	-20	1.00657	100	1	200	56.45	0	0	0	0	0	0	0	0	0	0	0;
	1348	25.65	3.95	30	-10	1.02611	100	1	100	17.1	0	0	0	0	0	0	0	0	0	0	0;
	1349	32.01	22.27	30	-10	1.031144	100	1	100	21.34	0	0	0	0	0	0	0	0	0	0	0;
	1359	31.57	7.72	30	-10	1.008948	100	1	100	21.05	0	0	0	0	0	0	0	0	0	0	0;
	1367	48.33	15.81	40	-10	1.031502	100	1	100	32.22	0	0	0	0	0	0	0	0	0	0	0;
	1369	266.84	38.78	210	-50	1.032265	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	1372	274.65	30.64	160	-40	1.018066	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	1376	24.65	14.16	30	-10	1.027131	100	1	100	16.43	0	0	0	0	0	0	0	0	0	0	0;
	1381	55.46	13.9	40	-10	1.030174	100	1	200	36.97	0	0	0	0	0	0	0	0	0	0	0;
	1382	21.45	5.94	30	-10	1.013179	100	1	100	14.3	0	0	0	0	0	0	0	0	0	0	0;
	1389	220.02	99.9	210	-50	1.03558	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	1390	220.02	99.9	210	-50	1.03558	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	1391	220.02	99.9	210	-50	1.03558	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	1394	255.53	96.98	210	-50	1.03858	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	1401	262.4	19.68	160	-40	1.015161	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	1407	87.88	32.08	60	-20	1.042865	100	1	200	58.59	0	0	0	0	0	0	0	0	0	0	0;
	1408	87.88	32.08	60	-20	1.042865	100	1	200	58.59	0	0	0	0	0	0	0	0	0	0	0;
	1410	91.17	37.8	60	-20	1.045666	100	1	200	60.78	0	0	0	0	0	0	0	0	0	0	0;
	1411	91.17	37.8	60	-20	1.045666	100	1	200	60.78	0	0	0	0	0	0	0	0	0	0	0;
	1412	91.17	37.8	60	-20	1.045666	100	1	200	60.78	0	0	0	0	0	0	0	0	0	0	0;
	1413	91.17	37.8	60	-20	1.045666	100	1	200	60.78	0	0	0	0	0	0	0	0	0	0	0;
	1414	87.88	32.08	60	-20	1.042865	100	1	200	58.59	0	0	0	0	0	0	0	0	0	0	0;
	1415	87.88	32.08	60	-20	1.042865	100	1	200	58.59	0	0	0	0	0	0	0	0	0	0	0;
	1416	91.17	37.8	60	-20	1.045666	100	1	200	60.78	0	0	0	0	0	0	0	0	0	0	0;
	1417	87.88	32.08	60	-20	1.042865	100	1	200	58.59	0	0	0	0	0	0	0	0	0	0	0;
	1419	22.22	6.81	30	-10	1.03758	100	1	100	14.81	0	0	0	0	0	0	0	0	0	0	0;
	1421	2	-0	10	-10	0.996766	100	1	100	1.33	0	0	0	0	0	0	0	0	0	0	0;
	1423	83.78	15.02	60	-20	1.033534	100	1	200	55.85	0	0	0	0	0	0	0	0	0	0	0;
	1427	91.88	4.87	60	-20	1.046359	100	1	200	61.25	0	0	0	0	0	0	0	0	0	0	0;
	1433	18.76	6.31	30	-10	1.033648	100	1	100	12.51	0	0	0	0	0	0	0	0	0	0	0;
	1434	18.76	6.31	30	-10	1.033648	100	1	100	12.51	0	0	0	0	0	0	0	0	0	0	0;
	1435	32.35	4.05	30	-10	1.048683	100	1	100	21.57	0	0	0	0	0	0	0	0	0	0	0;
	1436	32.35	4.05	30	-10	1.048683	100	1	100	21.57	0	0	0	0	0	0	0	0	0	0	0;
	1437	18.76	6.31	30	-10	1.033648	100	1	100	12.51	0	0	0	0	0	0	0	0	0	0	0;
	1438	18.76	6.31	30	-10	1.033648	100	1	100	12.51	0	0	0	0	0	0	0	0	0	0	0;
	1440	44.64	3.39	30	-10	1.04081	100	1	100	29.76	0	0	0	0	0	0	0	0	0	0	0;
	1441	32.35	4.05	30	-10	1.048683	100	1	100	21.57	0	0	0	0	0	0	0	0	0	0	0;
	1447	55.46	13.9	40	-10	1.030174	100	1	200	36.97	0	0	0	0	0	0	0	0	0	0	0;
	1448	55.46	13.9	40	-10	1.030174	100	1	200	36.97	0	0	0	0	0	0	0	0	0	0	0;
	1449	55.46	13.9	40	-10	1.030174	100	1	200	36.97	0	0	0	0	0	0	0	0	0	0	0;
	1450	55.46	13.9	40	-10	1.030174	100	1	200	36.97	0	0	0	0	0	0	0	0	0	0	0;
	1451	48.3	8.12	40	-10	1.021906	100	1	100	32.2	0	0	0	0	0	0	0	0	0	0	0;
	1453	48.3	8.12	40	-10	1.021906	100	1	100	32.2	0	0	0	0	0	0	0	0	0	0	0;
	1454	48.3	8.12	40	-10	1.021906	100	1	100	32.2	0	0	0	0	0	0	0	0	0	0	0;
	1455	31.28	-0.92	30	-10	1.009892	100	1	100	20.85	0	0	0	0	0	0	0	0	0	0	0;
	1458	91.88	4.87	60	-20	1.046359	100	1	200	61.25	0	0	0	0	0	0	0	0	0	0	0;
	1462	96.03	17.22	60	-20	1.036065	100	1	200	64.02	0	0	0	0	0	0	0	0	0	0	0;
	1469	7	0.35	10	-10	1.01541	100	1	100	4.67	0	0	0	0	0	0	0	0	0	0	0;
	1470	90.1	33.82	60	-20	1.06527	100	1	200	60.07	0	0	0	0	0	0	0	0	0	0	0;
	1471	90.1	33.82	60	-20	1.06527	100	1	200	60.07	0	0	0	0	0	0	0	0	0	0	0;
	1473	90.1	33.82	60	-20	1.06527	100	1	200	60.07	0	0	0	0	0	0	0	0	0	0	0;
	1474	90.1	33.82	60	-20	1.06527	100	1	200	60.07	0	0	0	0	0	0	0	0	0	0	0;
	1475	90.1	33.82	60	-20	1.06527	100	1	200	60.07	0	0	0	0	0	0	0	0	0	0	0;
	1476	90.1	33.82	60	-20	1.06527	100	1	200	60.07	0	0	0	0	0	0	0	0	0	0	0;
	1477	18.08	8.32	30	-10	1.063432	100	1	100	12.05	0	0	0	0	0	0	0	0	0	0	0;
	1478	18.08	8.32	30	-10	1.063432	100	1	100	12.05	0	0	0	0	0	0	0	0	0	0	0;
	1479	18.08	8.32	30	-10	1.063432	100	1	100	12.05	0	0	0	0	0	0	0	0	0	0	0;
	1484	30.8	11.3	30	-10	1.060798	100	1	100	20.53	0	0	0	0	0	0	0	0	0	0	0;
	1485	30.8	11.3	30	-10	1.060798	100	1	100	20.53	0	0	0	0	0	0	0	0	0	0	0;
	1486	30.8	11.3	30	-10	1.060798	100	1	100	20.53	0	0	0	0	0	0	0	0	0	0	0;
	1488	1.2	0.03	10	-10	1.003171	100	1	100	0.8	0	0	0	0	0	0	0	0	0	0	0;
	1505	21.35	3.74	20	-10	1.025426	100	1	100	14.23	0	0	0	0	0	0	0	0	0	0	0;
	1514	36.42	7.96	30	-10	1.029166	100	1	100	24.28	0	0	0	0	0	0	0	0	0	0	0;
	1516	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	1517	51.4	10.71	60	-20	1.033944	100	1	200	34.27	0	0	0	0	0	0	0	0	0	0	0;
	1518	51.4	10.71	60	-20	1.033944	100	1	200	34.27	0	0	0	0	0	0	0	0	0	0	0;
	1519	50.15	10.7	60	-20	1.033944	100	1	200	33.43	0	0	0	0	0	0	0	0	0	0	0;
	1520	50.15	10.7	60	-20	1.033944	100	1	200	33.43	0	0	0	0	0	0	0	0	0	0	0;
	1521	73.5	31.87	60	-20	1.055774	100	1	200	49	0	0	0	0	0	0	0	0	0	0	0;
	1526	22	10.36	20	0	1.014332	100	1	100	14.67	0	0	0	0	0	0	0	0	0	0	0;
	1532	20.21	8.41	30	-10	1.012556	100	1	100	13.47	0	0	0	0	0	0	0	0	0	0	0;
	1534	97.52	42.26	60	-20	1.053776	100	1	200	65.01	0	0	0	0	0	0	0	0	0	0	0;
	1546	152.79	-6.51	80	-20	1.051916	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	1552	369.99	86.96	210	-50	1.043138	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	1553	36.13	4.17	30	-10	1.027348	100	1	100	24.09	0	0	0	0	0	0	0	0	0	0	0;
	1556	31.58	5.23	30	-10	1.069569	100	1	100	21.05	0	0	0	0	0	0	0	0	0	0	0;
	1557	31.58	5.23	30	-10	1.069569	100	1	100	21.05	0	0	0	0	0	0	0	0	0	0	0;
	1558	31.58	5.23	30	-10	1.069569	100	1	100	21.05	0	0	0	0	0	0	0	0	0	0	0;
	1559	400.01	197.96	210	-50	1.113563	100	1	900	200	0	0	0	0	0	0	0	0	0	0	0;
	1560	32.57	6.62	30	-10	1.059142	100	1	100	21.71	0	0	0	0	0	0	0	0	0	0	0;
	1561	204.89	166.08	210	-50	1.066067	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	1562	3.5	-0.25	10	-10	0.995346	100	1	100	2.33	0	0	0	0	0	0	0	0	0	0	0;
	1577	37.45	15.75	40	-10	1.026206	100	1	100	24.97	0	0	0	0	0	0	0	0	0	0	0;
	1586	97.82	33.17	60	-20	1.055063	100	1	200	65.21	0	0	0	0	0	0	0	0	0	0	0;
	1587	97.82	33.17	60	-20	1.055063	100	1	200	65.21	0	0	0	0	0	0	0	0	0	0	0;
	1594	44.8	-1.18	50	-40	1.004901	100	1	100	29.87	0	0	0	0	0	0	0	0	0	0	0;
	1599	3.84	0.45	10	-10	1.030843	100	1	100	2.56	0	0	0	0	0	0	0	0	0	0	0;
	1606	82.46	55.28	60	-20	1.127948	100	1	200	54.97	0	0	0	0	0	0	0	0	0	0	0;
	1627	219.46	64.31	210	-50	1.050142	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	1629	26.28	8.7	30	-10	1.026298	100	1	100	17.52	0	0	0	0	0	0	0	0	0	0	0;
	1631	276.46	124.36	160	-40	1.052654	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	1636	20.79	7.32	30	-10	1.033848	100	1	100	13.86	0	0	0	0	0	0	0	0	0	0	0;
	1640	10.5	-0.35	10	-10	0.994858	100	1	100	7	0	0	0	0	0	0	0	0	0	0	0;
	1641	26.33	10.66	30	-10	1.01282	100	1	100	17.55	0	0	0	0	0	0	0	0	0	0	0;
	1642	12.15	4.2	10	0	1.012131	100	1	100	8.1	0	0	0	0	0	0	0	0	0	0	0;
	1644	13.95	4.91	10	0	1.042135	100	1	100	9.3	0	0	0	0	0	0	0	0	0	0	0;
	1656	685.9	149.42	520	-120	1.062519	100	1	1400	400	0	0	0	0	0	0	0	0	0	0	0;
	1661	626.63	190.1	520	-120	1.051456	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	1662	26.06	3.99	20	-10	1.027716	100	1	100	17.37	0	0	0	0	0	0	0	0	0	0	0;
	1675	626.64	187.14	520	-120	1.050379	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	1685	20	-2.63	30	-10	1.010291	100	1	100	13.33	0	0	0	0	0	0	0	0	0	0	0;
	1689	754.17	238.65	520	-120	1.043738	100	1	1600	500	0	0	0	0	0	0	0	0	0	0	0;
	1701	40.49	8.06	30	-10	1.036966	100	1	100	26.99	0	0	0	0	0	0	0	0	0	0	0;
	1726	617.75	183.04	520	-120	1.058676	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	1731	41.8	9.39	40	-10	1.028604	100	1	100	27.87	0	0	0	0	0	0	0	0	0	0	0;
	1732	56.25	11.12	40	-10	1.03036	100	1	200	37.5	0	0	0	0	0	0	0	0	0	0	0;
	1734	156.87	105.55	160	-40	1.068377	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	1735	156.87	105.55	160	-40	1.068377	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	1749	31.28	-0.92	30	-10	1.009892	100	1	100	20.85	0	0	0	0	0	0	0	0	0	0	0;
	1750	336.9	100.14	210	-50	1.061222	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	1751	336.9	100.14	210	-50	1.061222	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	1757	27.2	3.82	30	-10	1.024762	100	1	100	18.13	0	0	0	0	0	0	0	0	0	0	0;
	1763	24.84	4.52	30	-10	1.022099	100	1	100	16.56	0	0	0	0	0	0	0	0	0	0	0;
	1764	18.67	5.94	30	-10	1.015505	100	1	100	12.45	0	0	0	0	0	0	0	0	0	0	0;
	1770	97.12	8.38	60	-20	1.040009	100	1	200	64.75	0	0	0	0	0	0	0	0	0	0	0;
	1772	97.12	8.38	60	-20	1.040009	100	1	200	64.75	0	0	0	0	0	0	0	0	0	0	0;
	1774	81.31	8.06	60	-20	1.043619	100	1	200	54.21	0	0	0	0	0	0	0	0	0	0	0;
	1791	30.13	8.92	30	-10	1.016081	100	1	100	20.09	0	0	0	0	0	0	0	0	0	0	0;
	1794	36.11	3.87	30	-10	1.024544	100	1	100	24.07	0	0	0	0	0	0	0	0	0	0	0;
	1796	234.71	26.51	160	-40	1.008485	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	1801	98.05	38.32	60	-20	1.058149	100	1	200	65.37	0	0	0	0	0	0	0	0	0	0	0;
	1803	27.65	14.97	30	-10	1.018554	100	1	100	18.43	0	0	0	0	0	0	0	0	0	0	0;
	1805	30.02	16.69	30	-10	1.023695	100	1	100	20.01	0	0	0	0	0	0	0	0	0	0	0;
	1807	30.02	16.69	30	-10	1.023695	100	1	100	20.01	0	0	0	0	0	0	0	0	0	0	0;
	1808	27.65	14.97	30	-10	1.018554	100	1	100	18.43	0	0	0	0	0	0	0	0	0	0	0;
	1812	81.88	21.37	60	-20	1.031971	100	1	200	54.59	0	0	0	0	0	0	0	0	0	0	0;
	1813	37.23	9.09	30	-10	1.012457	100	1	100	24.82	0	0	0	0	0	0	0	0	0	0	0;
	1815	23.96	4.79	30	-10	1.007614	100	1	100	15.97	0	0	0	0	0	0	0	0	0	0	0;
	1816	24.94	11.53	30	-10	1.042445	100	1	100	16.63	0	0	0	0	0	0	0	0	0	0	0;
	1821	161	0.83	90	-90	1.011416	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	1822	204	41.74	150	-60	1.035162	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	1831	28.97	5.47	30	-10	1.043059	100	1	100	19.31	0	0	0	0	0	0	0	0	0	0	0;
	1834	14.99	2.45	10	-10	0.998251	100	1	100	9.99	0	0	0	0	0	0	0	0	0	0	0;
	1836	3.58	0.38	10	-10	1.006718	100	1	100	2.39	0	0	0	0	0	0	0	0	0	0	0;
	1839	72.47	11.44	60	-20	1.019231	100	1	200	48.31	0	0	0	0	0	0	0	0	0	0	0;
	1840	321.41	105.18	210	-50	1.033947	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	1845	86.25	-3.04	60	-80	0.962614	100	1	200	57.5	0	0	0	0	0	0	0	0	0	0	0;
	1851	322.34	84.39	210	-50	1.058231	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	1853	37.8	19.78	30	-10	1.031737	100	1	100	25.2	0	0	0	0	0	0	0	0	0	0	0;
	1868	114	6.43	70	-50	0.989649	100	1	300	76	0	0	0	0	0	0	0	0	0	0	0;
	1869	57.12	17.95	40	-10	1.017464	100	1	200	38.08	0	0	0	0	0	0	0	0	0	0	0;
	1876	55.55	12.64	40	-10	1.024324	100	1	200	37.03	0	0	0	0	0	0	0	0	0	0	0;
	1877	55.55	12.64	40	-10	1.024324	100	1	200	37.03	0	0	0	0	0	0	0	0	0	0	0;
	1878	55.55	12.64	40	-10	1.024324	100	1	200	37.03	0	0	0	0	0	0	0	0	0	0	0;
	1879	55.55	12.64	40	-10	1.024324	100	1	200	37.03	0	0	0	0	0	0	0	0	0	0	0;
	1880	55.55	12.64	40	-10	1.024324	100	1	200	37.03	0	0	0	0	0	0	0	0	0	0	0;
	1881	55.55	12.64	40	-10	1.024324	100	1	200	37.03	0	0	0	0	0	0	0	0	0	0	0;
	1882	46.26	7.53	40	-10	1.015006	100	1	100	30.84	0	0	0	0	0	0	0	0	0	0	0;
	1883	46.26	7.53	40	-10	1.015006	100	1	100	30.84	0	0	0	0	0	0	0	0	0	0	0;
	1884	46.26	7.53	40	-10	1.015006	100	1	100	30.84	0	0	0	0	0	0	0	0	0	0	0;
	1889	33.97	5.68	30	-10	1.022889	100	1	100	22.65	0	0	0	0	0	0	0	0	0	0	0;
	1895	23.5	5.12	30	-10	1.029146	100	1	100	15.67	0	0	0	0	0	0	0	0	0	0	0;
	1896	39.93	11.77	30	-10	1.012991	100	1	100	26.62	0	0	0	0	0	0	0	0	0	0	0;
	1903	5.2	0.53	10	-10	1.031159	100	1	100	3.47	0	0	0	0	0	0	0	0	0	0	0;
	1906	25.01	5.98	30	-10	0.999838	100	1	100	16.67	0	0	0	0	0	0	0	0	0	0	0;
	1907	18.1	0.8	10	-10	1.007227	100	1	100	12.07	0	0	0	0	0	0	0	0	0	0	0;
	1913	81.62	16.81	60	-20	1.042096	100	1	200	54.41	0	0	0	0	0	0	0	0	0	0	0;
	1914	81.62	16.81	60	-20	1.042096	100	1	200	54.41	0	0	0	0	0	0	0	0	0	0	0;
	1928	16.54	-1.5	10	-10	0.999593	100	1	100	11.03	0	0	0	0	0	0	0	0	0	0	0;
	1932	419.12	110.83	520	-120	1.016993	100	1	900	200	0	0	0	0	0	0	0	0	0	0	0;
	1935	17.96	1.67	30	-10	1.048808	100	1	100	11.97	0	0	0	0	0	0	0	0	0	0	0;
	1936	17.96	1.67	30	-10	1.048808	100	1	100	11.97	0	0	0	0	0	0	0	0	0	0	0;
	1940	206.37	113.42	160	-40	1.061603	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	1949	-30.21	0.29	10	-10	1.00597	100	1	100	-40.28	0	0	0	0	0	0	0	0	0	0	0;
	1955	65.21	12.85	60	-20	1.026243	100	1	200	43.47	0	0	0	0	0	0	0	0	0	0	0;
	1957	12.81	0.42	10	-10	1.019066	100	1	100	8.54	0	0	0	0	0	0	0	0	0	0	0;
	1959	10.75	0.78	10	-10	1.023372	100	1	100	7.17	0	0	0	0	0	0	0	0	0	0	0;
	1960	22.35	5.33	30	-10	1.019	100	1	100	14.9	0	0	0	0	0	0	0	0	0	0	0;
	1961	17.5	-0.6	20	-20	0.996564	100	1	100	11.67	0	0	0	0	0	0	0	0	0	0	0;
	1963	68.3	12.27	60	-20	1.015227	100	1	200	45.53	0	0	0	0	0	0	0	0	0	0	0;
	1966	50.34	11.93	60	-20	1.051329	100	1	200	33.56	0	0	0	0	0	0	0	0	0	0	0;
	1967	50.34	11.93	60	-20	1.051329	100	1	200	33.56	0	0	0	0	0	0	0	0	0	0	0;
	1969	66.65	5.48	60	-20	1.01993	100	1	200	44.43	0	0	0	0	0	0	0	0	0	0	0;
	1972	36.42	7.96	30	-10	1.029166	100	1	100	24.28	0	0	0	0	0	0	0	0	0	0	0;
	1985	60.3	-12.73	30	-30	0.961956	100	1	200	40.2	0	0	0	0	0	0	0	0	0	0	0;
	1988	25.73	6.33	30	-10	1.027807	100	1	100	17.15	0	0	0	0	0	0	0	0	0	0	0;
	1995	305.54	12.74	210	-50	0.989323	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	1997	863.34	248.84	520	-120	1.044265	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	1999	35.65	6.01	30	-10	1.058091	100	1	100	23.77	0	0	0	0	0	0	0	0	0	0	0;
	2000	35.65	6.01	30	-10	1.058091	100	1	100	23.77	0	0	0	0	0	0	0	0	0	0	0;
	2001	39.12	14.36	30	-10	1.038561	100	1	100	26.08	0	0	0	0	0	0	0	0	0	0	0;
	2003	30.7	16.74	30	-10	1.0385	100	1	100	20.47	0	0	0	0	0	0	0	0	0	0	0;
	2004	30.7	16.74	30	-10	1.0385	100	1	100	20.47	0	0	0	0	0	0	0	0	0	0	0;
	2005	30.7	16.74	30	-10	1.0385	100	1	100	20.47	0	0	0	0	0	0	0	0	0	0	0;
	2006	29.05	10.97	30	-10	1.053599	100	1	100	19.37	0	0	0	0	0	0	0	0	0	0	0;
	2011	35.65	6.01	30	-10	1.058091	100	1	100	23.77	0	0	0	0	0	0	0	0	0	0	0;
	2012	54.32	19.31	40	-10	1.040793	100	1	200	36.21	0	0	0	0	0	0	0	0	0	0	0;
	2013	54.32	19.31	40	-10	1.040793	100	1	200	36.21	0	0	0	0	0	0	0	0	0	0	0;
	2014	54.32	19.31	40	-10	1.040793	100	1	200	36.21	0	0	0	0	0	0	0	0	0	0	0;
	2020	54.32	19.31	40	-10	1.040793	100	1	200	36.21	0	0	0	0	0	0	0	0	0	0	0;
	2042	-143.08	-12.98	30	-50	1.027071	100	1	100	-190.77	0	0	0	0	0	0	0	0	0	0	0;
	2047	279.23	123.34	160	-40	1.05948	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	2048	22	2.53	20	-10	1.047307	100	1	100	14.67	0	0	0	0	0	0	0	0	0	0	0;
	2053	293.92	111.78	210	-50	1.039038	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	2063	25.89	9.83	40	-10	1.013208	100	1	100	17.26	0	0	0	0	0	0	0	0	0	0	0;
	2066	59.5	9.44	60	-20	1.032856	100	1	200	39.67	0	0	0	0	0	0	0	0	0	0	0;
	2067	-599.99	1981.94	100000	-100000	1.099592	100	1	100	-799.99	0	0	0	0	0	0	0	0	0	0	0;
	2069	172.2	40.07	100	-30	1.019153	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	2071	92.8	3.25	50	-50	0.999187	100	1	200	61.87	0	0	0	0	0	0	0	0	0	0	0;
	2080	90.48	33.58	60	-20	1.043665	100	1	200	60.32	0	0	0	0	0	0	0	0	0	0	0;
	2082	58.57	21.97	40	-10	1.034376	100	1	200	39.05	0	0	0	0	0	0	0	0	0	0	0;
	2084	312.49	55.02	210	-50	1.033042	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	2085	58.57	21.99	40	-10	1.03432	100	1	200	39.05	0	0	0	0	0	0	0	0	0	0	0;
	2086	58.57	21.99	40	-10	1.03432	100	1	200	39.05	0	0	0	0	0	0	0	0	0	0	0;
	2090	279.76	45.2	160	-40	1.030138	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	2091	279.76	45.2	160	-40	1.030138	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	2093	279.76	45.2	160	-40	1.030138	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	2097	167.59	39.08	210	-50	1.024484	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	2102	20.38	9.61	30	-10	1.012679	100	1	100	13.59	0	0	0	0	0	0	0	0	0	0	0;
	2106	165	33.72	100	-20	0.962508	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	2107	19.1	2.95	20	-10	1.022583	100	1	100	12.73	0	0	0	0	0	0	0	0	0	0	0;
	2119	616.37	99.15	520	-120	1.048702	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	2122	25.2	-3.93	20	-20	0.993105	100	1	100	16.8	0	0	0	0	0	0	0	0	0	0	0;
	2131	44.1	5.46	50	-20	0.988831	100	1	100	29.4	0	0	0	0	0	0	0	0	0	0	0;
	2135	629.75	200.05	520	-120	1.035502	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	2136	629.77	200.05	520	-120	1.035502	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	2137	629.77	200.05	520	-120	1.035502	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	2138	629.75	200.05	520	-120	1.035502	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	2139	57	2.92	30	-30	1.003313	100	1	200	38	0	0	0	0	0	0	0	0	0	0	0;
	2143	241.27	5.04	160	-40	1.017546	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	2144	241.27	5.04	160	-40	1.017546	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	2146	241.27	5.04	160	-40	1.017546	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	2147	259.51	14.89	160	-40	1.019095	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	2148	33.94	6.74	30	-10	1.046243	100	1	100	22.63	0	0	0	0	0	0	0	0	0	0	0;
	2149	33.94	6.74	30	-10	1.046243	100	1	100	22.63	0	0	0	0	0	0	0	0	0	0	0;
	2153	29.22	4.97	30	-10	1.01017	100	1	100	19.48	0	0	0	0	0	0	0	0	0	0	0;
	2154	22.27	8.99	30	-10	1.04545	100	1	100	14.85	0	0	0	0	0	0	0	0	0	0	0;
	2155	24.48	2.37	30	-10	1.030117	100	1	100	16.32	0	0	0	0	0	0	0	0	0	0	0;
	2156	31.3	11.67	30	-10	1.049836	100	1	100	20.87	0	0	0	0	0	0	0	0	0	0	0;
	2157	22	9.29	30	-10	1.030852	100	1	100	14.67	0	0	0	0	0	0	0	0	0	0	0;
	2158	24.48	2.37	30	-10	1.030117	100	1	100	16.32	0	0	0	0	0	0	0	0	0	0	0;
	2166	21.45	2.81	20	-10	1.02436	100	1	100	14.3	0	0	0	0	0	0	0	0	0	0	0;
	2167	31.72	4.48	30	-10	1.022192	100	1	100	21.15	0	0	0	0	0	0	0	0	0	0	0;
	2172	74.78	41.98	60	-20	1.04914	100	1	200	49.85	0	0	0	0	0	0	0	0	0	0	0;
	2176	82.46	49.93	60	-20	1.054354	100	1	200	54.97	0	0	0	0	0	0	0	0	0	0	0;
	2177	82.46	49.93	60	-20	1.054354	100	1	200	54.97	0	0	0	0	0	0	0	0	0	0	0;
	2178	82.46	49.93	60	-20	1.054354	100	1	200	54.97	0	0	0	0	0	0	0	0	0	0	0;
	2179	82.46	49.93	60	-20	1.054354	100	1	200	54.97	0	0	0	0	0	0	0	0	0	0	0;
	2180	132.3	-12.95	80	-80	0.985554	100	1	300	88.2	0	0	0	0	0	0	0	0	0	0	0;
	2183	35.95	6.39	40	-10	1.015185	100	1	100	23.97	0	0	0	0	0	0	0	0	0	0	0;
	2186	189.36	46.97	160	-40	1.05594	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	2190	43.55	17.11	40	-10	1.033542	100	1	100	29.03	0	0	0	0	0	0	0	0	0	0	0;
	2192	0.99	-0.05	10	-10	1.015919	100	1	100	0.66	0	0	0	0	0	0	0	0	0	0	0;
	2197	200.04	44.85	210	-50	1.024678	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	2198	96.9	14.27	60	-20	1.033735	100	1	200	64.6	0	0	0	0	0	0	0	0	0	0	0;
	2205	7.94	0.62	10	-10	1.010845	100	1	100	5.29	0	0	0	0	0	0	0	0	0	0	0;
	2210	27.2	3.82	30	-10	1.024762	100	1	100	18.13	0	0	0	0	0	0	0	0	0	0	0;
	2215	27.37	3.79	30	-20	1.003014	100	1	100	18.25	0	0	0	0	0	0	0	0	0	0	0;
	2222	25.41	10.96	30	-10	1.019405	100	1	100	16.94	0	0	0	0	0	0	0	0	0	0	0;
	2223	23.04	1.33	20	-20	0.996935	100	1	100	15.36	0	0	0	0	0	0	0	0	0	0	0;
	2236	53.99	18.23	40	-10	1.062896	100	1	200	35.99	0	0	0	0	0	0	0	0	0	0	0;
	2237	44.8	3.17	40	-20	0.991989	100	1	100	29.87	0	0	0	0	0	0	0	0	0	0	0;
	2242	36.11	3.87	30	-10	1.024544	100	1	100	24.07	0	0	0	0	0	0	0	0	0	0	0;
	2246	24.03	8.37	30	-10	1.005775	100	1	100	16.02	0	0	0	0	0	0	0	0	0	0	0;
	2247	31.95	15.26	30	-10	1.019723	100	1	100	21.3	0	0	0	0	0	0	0	0	0	0	0;
	2248	31.95	15.26	30	-10	1.019723	100	1	100	21.3	0	0	0	0	0	0	0	0	0	0	0;
	2249	31.95	15.26	30	-10	1.019723	100	1	100	21.3	0	0	0	0	0	0	0	0	0	0	0;
	2252	31.95	15.26	30	-10	1.019723	100	1	100	21.3	0	0	0	0	0	0	0	0	0	0	0;
	2253	30.01	15.35	30	-10	1.052369	100	1	100	20.01	0	0	0	0	0	0	0	0	0	0	0;
	2260	136	40.17	110	-30	1.001531	100	1	300	90.67	0	0	0	0	0	0	0	0	0	0	0;
	2266	-31.54	-6.13	10	-20	1.006784	100	1	100	-42.05	0	0	0	0	0	0	0	0	0	0	0;
	2268	15	5.29	10	0	1.037301	100	1	100	10	0	0	0	0	0	0	0	0	0	0	0;
	2270	14	-0.92	10	-10	0.99902	100	1	100	9.33	0	0	0	0	0	0	0	0	0	0	0;
	2278	5.2	0.52	10	-10	1.018046	100	1	100	3.47	0	0	0	0	0	0	0	0	0	0	0;
	2282	31.72	4.48	30	-10	1.022192	100	1	100	21.15	0	0	0	0	0	0	0	0	0	0	0;
	2283	31.72	4.48	30	-10	1.022192	100	1	100	21.15	0	0	0	0	0	0	0	0	0	0	0;
	2284	31.72	4.48	30	-10	1.022192	100	1	100	21.15	0	0	0	0	0	0	0	0	0	0	0;
	2285	31.72	4.48	30	-10	1.022192	100	1	100	21.15	0	0	0	0	0	0	0	0	0	0	0;
	2286	170.85	16.98	160	-40	1.000084	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	2287	170.85	16.98	160	-40	1.000084	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	2292	22.73	8.07	30	-10	1.024089	100	1	100	15.15	0	0	0	0	0	0	0	0	0	0	0;
	2301	219.03	73.16	210	-50	1.033551	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	2302	322.34	84.39	210	-50	1.058231	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	2303	131.51	26.68	70	-20	1.005693	100	1	300	87.67	0	0	0	0	0	0	0	0	0	0	0;
	2307	24.45	4.95	30	-10	0.994621	100	1	100	16.3	0	0	0	0	0	0	0	0	0	0	0;
	2308	290	-14.76	100	-100	1.001855	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	2318	56.84	10.43	40	-10	1.025724	100	1	200	37.89	0	0	0	0	0	0	0	0	0	0	0;
	2321	56.84	10.43	40	-10	1.025724	100	1	200	37.89	0	0	0	0	0	0	0	0	0	0	0;
	2322	56.84	10.43	40	-10	1.025724	100	1	200	37.89	0	0	0	0	0	0	0	0	0	0	0;
	2324	23.3	5.11	30	-10	1.015556	100	1	100	15.53	0	0	0	0	0	0	0	0	0	0	0;
	2325	23.3	5.11	30	-10	1.015556	100	1	100	15.53	0	0	0	0	0	0	0	0	0	0	0;
	2328	34.3	7.71	30	-10	1.038216	100	1	100	22.87	0	0	0	0	0	0	0	0	0	0	0;
	2329	34.3	7.71	30	-10	1.038216	100	1	100	22.87	0	0	0	0	0	0	0	0	0	0	0;
	2330	34.3	7.71	30	-10	1.038216	100	1	100	22.87	0	0	0	0	0	0	0	0	0	0	0;
	2331	34.3	7.71	30	-10	1.038216	100	1	100	22.87	0	0	0	0	0	0	0	0	0	0	0;
	2332	33.97	5.68	30	-10	1.022889	100	1	100	22.65	0	0	0	0	0	0	0	0	0	0	0;
	2333	23.3	5.11	30	-10	1.015556	100	1	100	15.53	0	0	0	0	0	0	0	0	0	0	0;
	2341	2.5	0.07	10	-10	0.9846	100	1	100	1.67	0	0	0	0	0	0	0	0	0	0	0;
	2345	300.01	75.79	160	-40	1.035781	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	2347	81.31	8.06	60	-20	1.043619	100	1	200	54.21	0	0	0	0	0	0	0	0	0	0	0;
	2351	9.98	3.74	10	-10	1.015932	100	1	100	6.65	0	0	0	0	0	0	0	0	0	0	0;
	2357	6.38	-2.2	10	-10	0.958146	100	1	100	4.25	0	0	0	0	0	0	0	0	0	0	0;
	2368	74.8	42.23	60	-20	1.046905	100	1	200	49.87	0	0	0	0	0	0	0	0	0	0	0;
	2369	32.57	6.88	30	-10	1.03654	100	1	100	21.71	0	0	0	0	0	0	0	0	0	0	0;
	2370	32.57	6.88	30	-10	1.03654	100	1	100	21.71	0	0	0	0	0	0	0	0	0	0	0;
	2371	32.57	6.88	30	-10	1.03654	100	1	100	21.71	0	0	0	0	0	0	0	0	0	0	0;
	2372	84.3	7.63	60	-20	1.05433	100	1	200	56.2	0	0	0	0	0	0	0	0	0	0	0;
	2373	84.3	7.63	60	-20	1.05433	100	1	200	56.2	0	0	0	0	0	0	0	0	0	0	0;
	2377	84.3	7.63	60	-20	1.05433	100	1	200	56.2	0	0	0	0	0	0	0	0	0	0	0;
	2389	33.64	6.9	30	-10	1.00773	100	1	100	22.43	0	0	0	0	0	0	0	0	0	0	0;
	2397	7	0.35	10	-10	1.01541	100	1	100	4.67	0	0	0	0	0	0	0	0	0	0	0;
	2408	96.03	17.22	60	-20	1.036065	100	1	200	64.02	0	0	0	0	0	0	0	0	0	0	0;
	2410	16.07	9.31	30	-10	1.034871	100	1	100	10.71	0	0	0	0	0	0	0	0	0	0	0;
	2411	17.21	5.28	30	-10	1.020161	100	1	100	11.47	0	0	0	0	0	0	0	0	0	0	0;
	2422	3.37	0.4	10	-10	1.03091	100	1	100	2.25	0	0	0	0	0	0	0	0	0	0	0;
	2428	55.25	2.21	40	-40	0.994234	100	1	200	36.83	0	0	0	0	0	0	0	0	0	0	0;
	2429	26.62	13.25	30	-10	1.040642	100	1	100	17.75	0	0	0	0	0	0	0	0	0	0	0;
	2430	159.8	20.91	100	-40	1.008982	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	2434	206.32	45.73	160	-40	1.035431	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	2435	85.21	4.77	60	-20	1.052787	100	1	200	56.81	0	0	0	0	0	0	0	0	0	0	0;
	2436	85.21	4.77	60	-20	1.052787	100	1	200	56.81	0	0	0	0	0	0	0	0	0	0	0;
	2437	85.21	4.77	60	-20	1.052787	100	1	200	56.81	0	0	0	0	0	0	0	0	0	0	0;
	2438	85.21	4.77	60	-20	1.052787	100	1	200	56.81	0	0	0	0	0	0	0	0	0	0	0;
	2439	85.21	4.77	60	-20	1.052787	100	1	200	56.81	0	0	0	0	0	0	0	0	0	0	0;
	2440	25.32	7.7	30	-10	1.059619	100	1	100	16.88	0	0	0	0	0	0	0	0	0	0	0;
	2446	81.88	21.37	60	-20	1.031971	100	1	200	54.59	0	0	0	0	0	0	0	0	0	0	0;
	2447	81.88	21.37	60	-20	1.031971	100	1	200	54.59	0	0	0	0	0	0	0	0	0	0	0;
	2448	81.88	21.37	60	-20	1.031971	100	1	200	54.59	0	0	0	0	0	0	0	0	0	0	0;
	2458	18.62	7.28	30	-10	1.020659	100	1	100	12.41	0	0	0	0	0	0	0	0	0	0	0;
	2459	36.91	13.33	30	-10	1.037973	100	1	100	24.61	0	0	0	0	0	0	0	0	0	0	0;
	2460	21.1	5.27	20	-10	1.011637	100	1	100	14.07	0	0	0	0	0	0	0	0	0	0	0;
	2462	312.49	55.02	210	-50	1.033042	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	2464	5.26	0.06	10	-10	1.011692	100	1	100	3.51	0	0	0	0	0	0	0	0	0	0	0;
	2473	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	2480	851.1	450.38	520	-120	1.078638	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	2481	851.1	450.38	520	-120	1.078638	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	2482	76.7	4.41	60	-20	1.02424	100	1	200	51.13	0	0	0	0	0	0	0	0	0	0	0;
	2485	34.31	1.99	30	-10	1.015531	100	1	100	22.87	0	0	0	0	0	0	0	0	0	0	0;
	2486	31.12	9.56	30	-10	1.04434	100	1	100	20.75	0	0	0	0	0	0	0	0	0	0	0;
	2487	31.12	9.56	30	-10	1.04434	100	1	100	20.75	0	0	0	0	0	0	0	0	0	0	0;
	2488	20.97	7.8	30	-10	1.055727	100	1	100	13.98	0	0	0	0	0	0	0	0	0	0	0;
	2489	20.97	7.8	30	-10	1.055727	100	1	100	13.98	0	0	0	0	0	0	0	0	0	0	0;
	2490	31.12	9.56	30	-10	1.04434	100	1	100	20.75	0	0	0	0	0	0	0	0	0	0	0;
	2506	47.25	-1.74	20	-20	0.990208	100	1	100	31.5	0	0	0	0	0	0	0	0	0	0	0;
	2510	92.69	14.48	60	-20	1.026207	100	1	200	61.79	0	0	0	0	0	0	0	0	0	0	0;
	2513	43.55	17.11	40	-10	1.033542	100	1	100	29.03	0	0	0	0	0	0	0	0	0	0	0;
	2514	43.55	17.08	40	-10	1.033552	100	1	100	29.03	0	0	0	0	0	0	0	0	0	0	0;
	2515	43.55	17.08	40	-10	1.033552	100	1	100	29.03	0	0	0	0	0	0	0	0	0	0	0;
	2516	43.55	17.08	40	-10	1.033552	100	1	100	29.03	0	0	0	0	0	0	0	0	0	0	0;
	2518	53.01	8.19	40	-10	1.026319	100	1	200	35.34	0	0	0	0	0	0	0	0	0	0	0;
	2526	0.91	-0.06	10	-10	1.01592	100	1	100	0.6	0	0	0	0	0	0	0	0	0	0	0;
	2533	94.74	2.75	60	-20	1.014104	100	1	200	63.16	0	0	0	0	0	0	0	0	0	0	0;
	2534	90.45	21.15	60	-20	1.046641	100	1	200	60.3	0	0	0	0	0	0	0	0	0	0	0;
	2539	595	-374.87	100000	-100000	1.000445	100	1	1200	300	0	0	0	0	0	0	0	0	0	0	0;
	2540	90.45	21.15	60	-20	1.046641	100	1	200	60.3	0	0	0	0	0	0	0	0	0	0	0;
	2543	90.45	21.15	60	-20	1.046641	100	1	200	60.3	0	0	0	0	0	0	0	0	0	0	0;
	2559	23.61	14.88	30	-10	1.029344	100	1	100	15.74	0	0	0	0	0	0	0	0	0	0	0;
	2561	23.81	8.18	30	-10	1.024067	100	1	100	15.87	0	0	0	0	0	0	0	0	0	0	0;
	2566	48.5	-4.15	30	-30	1.002755	100	1	100	32.33	0	0	0	0	0	0	0	0	0	0	0;
	2569	159.17	-15.32	160	-40	1.030434	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	2571	90.45	21.15	60	-20	1.046641	100	1	200	60.3	0	0	0	0	0	0	0	0	0	0	0;
	2572	94.74	2.75	60	-20	1.014104	100	1	200	63.16	0	0	0	0	0	0	0	0	0	0	0;
	2573	94.74	2.75	60	-20	1.014104	100	1	200	63.16	0	0	0	0	0	0	0	0	0	0	0;
	2577	90.45	21.15	60	-20	1.046641	100	1	200	60.3	0	0	0	0	0	0	0	0	0	0	0;
	2584	198.91	5.92	210	-50	1.031301	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	2592	200.67	-1.54	160	-40	1.014299	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	2603	44.8	-1.18	50	-40	1.004901	100	1	100	29.87	0	0	0	0	0	0	0	0	0	0	0;
	2606	338.05	109.11	210	-50	1.033036	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	2607	338.05	109.11	210	-50	1.033036	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	2610	55.22	8.96	40	-10	1.017828	100	1	200	36.81	0	0	0	0	0	0	0	0	0	0	0;
	2624	285.75	2.4	210	-50	1.045696	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	2625	209.96	80.01	160	-40	1.039006	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	2627	261.96	25.03	210	-50	1.042062	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	2635	41.8	9.39	40	-10	1.028604	100	1	100	27.87	0	0	0	0	0	0	0	0	0	0	0;
	2642	55.81	12.44	40	-10	1.030658	100	1	200	37.21	0	0	0	0	0	0	0	0	0	0	0;
	2645	36.72	5.04	30	-20	1.013484	100	1	100	24.48	0	0	0	0	0	0	0	0	0	0	0;
	2646	50.33	11.56	40	-10	1.037142	100	1	200	33.55	0	0	0	0	0	0	0	0	0	0	0;
	2647	50.94	19.18	40	-10	1.0136	100	1	200	33.96	0	0	0	0	0	0	0	0	0	0	0;
	2650	95.51	37.64	60	-20	1.0608	100	1	200	63.67	0	0	0	0	0	0	0	0	0	0	0;
	2651	95.51	37.64	60	-20	1.0608	100	1	200	63.67	0	0	0	0	0	0	0	0	0	0	0;
	2652	97.9	33	60	-20	1.058549	100	1	200	65.27	0	0	0	0	0	0	0	0	0	0	0;
	2653	97.9	33	60	-20	1.058549	100	1	200	65.27	0	0	0	0	0	0	0	0	0	0	0;
	2654	97.9	33	60	-20	1.058549	100	1	200	65.27	0	0	0	0	0	0	0	0	0	0	0;
	2655	97.9	33	60	-20	1.058549	100	1	200	65.27	0	0	0	0	0	0	0	0	0	0	0;
	2656	97.9	33	60	-20	1.058549	100	1	200	65.27	0	0	0	0	0	0	0	0	0	0	0;
	2657	97.9	33	60	-20	1.058549	100	1	200	65.27	0	0	0	0	0	0	0	0	0	0	0;
	2662	38.16	4.12	30	-10	1.055337	100	1	100	25.44	0	0	0	0	0	0	0	0	0	0	0;
	2666	25.73	6.33	30	-10	1.027807	100	1	100	17.15	0	0	0	0	0	0	0	0	0	0	0;
	2667	25.73	6.33	30	-10	1.027807	100	1	100	17.15	0	0	0	0	0	0	0	0	0	0	0;
	2669	25.73	6.33	30	-10	1.027807	100	1	100	17.15	0	0	0	0	0	0	0	0	0	0	0;
	2670	79.79	-1.82	60	-20	1.024891	100	1	200	53.19	0	0	0	0	0	0	0	0	0	0	0;
	2676	34.14	3.85	30	-10	1.012735	100	1	100	22.76	0	0	0	0	0	0	0	0	0	0	0;
	2681	90.48	33.58	60	-20	1.043665	100	1	200	60.32	0	0	0	0	0	0	0	0	0	0	0;
	2687	37.32	7.31	30	-10	1.034314	100	1	100	24.88	0	0	0	0	0	0	0	0	0	0	0;
	2688	40.16	4.54	30	-10	1.021208	100	1	100	26.77	0	0	0	0	0	0	0	0	0	0	0;
	2690	37.32	7.31	30	-10	1.034314	100	1	100	24.88	0	0	0	0	0	0	0	0	0	0	0;
	2691	5.26	0.5	10	-10	1.01939	100	1	100	3.51	0	0	0	0	0	0	0	0	0	0	0;
	2693	19.91	7.13	30	-10	1.021055	100	1	100	13.27	0	0	0	0	0	0	0	0	0	0	0;
	2694	19.91	7.13	30	-10	1.021055	100	1	100	13.27	0	0	0	0	0	0	0	0	0	0	0;
	2695	46.2	9.44	30	-10	1.030241	100	1	100	30.8	0	0	0	0	0	0	0	0	0	0	0;
	2698	36.8	6.62	30	-10	1.03058	100	1	100	24.53	0	0	0	0	0	0	0	0	0	0	0;
	2700	27.2	3.82	30	-10	1.024762	100	1	100	18.13	0	0	0	0	0	0	0	0	0	0	0;
	2702	16.67	4.68	30	-10	1.013686	100	1	100	11.11	0	0	0	0	0	0	0	0	0	0	0;
	2703	31.08	5.31	30	-10	1.031603	100	1	100	20.72	0	0	0	0	0	0	0	0	0	0	0;
	2705	31.08	5.31	30	-10	1.031603	100	1	100	20.72	0	0	0	0	0	0	0	0	0	0	0;
	2708	26.98	3.71	30	-10	1.066367	100	1	100	17.99	0	0	0	0	0	0	0	0	0	0	0;
	2711	27.51	11.78	30	-10	1.042669	100	1	100	18.34	0	0	0	0	0	0	0	0	0	0	0;
	2712	49.1	1.57	30	-20	1.016984	100	1	100	32.73	0	0	0	0	0	0	0	0	0	0	0;
	2717	107.01	-5.63	50	-50	0.996677	100	1	300	71.34	0	0	0	0	0	0	0	0	0	0	0;
	2721	12.6	5.12	10	0	1.029248	100	1	100	8.4	0	0	0	0	0	0	0	0	0	0	0;
	2734	78.21	5.29	60	-20	1.025803	100	1	200	52.14	0	0	0	0	0	0	0	0	0	0	0;
	2756	22.5	7.8	30	-10	0.998215	100	1	100	15	0	0	0	0	0	0	0	0	0	0	0;
	2759	240.6	61.33	100	-30	0.990464	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	2760	311.23	83.72	210	-50	1.058355	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	2767	36.22	11.19	30	-10	1.003847	100	1	100	24.15	0	0	0	0	0	0	0	0	0	0	0;
	2777	31.28	-0.92	30	-10	1.009892	100	1	100	20.85	0	0	0	0	0	0	0	0	0	0	0;
	2783	54.32	19.31	40	-10	1.040793	100	1	200	36.21	0	0	0	0	0	0	0	0	0	0	0;
	2784	4.2	1.57	10	-10	1.019419	100	1	100	2.8	0	0	0	0	0	0	0	0	0	0	0;
	2785	36.22	11.19	30	-10	1.003847	100	1	100	24.15	0	0	0	0	0	0	0	0	0	0	0;
	2786	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	2787	57.27	17.52	40	-10	1.017845	100	1	200	38.18	0	0	0	0	0	0	0	0	0	0	0;
	2789	74.86	19.42	60	-20	1.011998	100	1	200	49.91	0	0	0	0	0	0	0	0	0	0	0;
	2790	24.98	14.2	30	-10	1.031726	100	1	100	16.65	0	0	0	0	0	0	0	0	0	0	0;
	2796	33.97	5.68	30	-10	1.022889	100	1	100	22.65	0	0	0	0	0	0	0	0	0	0	0;
	2799	54.32	19.31	40	-10	1.040793	100	1	200	36.21	0	0	0	0	0	0	0	0	0	0	0;
	2803	27.38	10.62	30	-10	1.015042	100	1	100	18.25	0	0	0	0	0	0	0	0	0	0	0;
	2810	54.32	19.31	40	-10	1.040793	100	1	200	36.21	0	0	0	0	0	0	0	0	0	0	0;
	2811	257.89	39.7	210	-50	1.022087	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	2816	15	-0.86	10	-10	0.992395	100	1	100	10	0	0	0	0	0	0	0	0	0	0	0;
	2819	26.19	8.11	30	-10	1.031736	100	1	100	17.46	0	0	0	0	0	0	0	0	0	0	0;
	2828	93.04	14.9	60	-20	1.017638	100	1	200	62.03	0	0	0	0	0	0	0	0	0	0	0;
	2831	23	-6.2	20	-20	0.989469	100	1	100	15.33	0	0	0	0	0	0	0	0	0	0	0;
	2836	18.58	13.67	30	-10	1.061581	100	1	100	12.39	0	0	0	0	0	0	0	0	0	0	0;
	2838	21.72	6.98	30	-10	1.064384	100	1	100	14.48	0	0	0	0	0	0	0	0	0	0	0;
	2840	32.97	4.46	30	-10	1.027968	100	1	100	21.98	0	0	0	0	0	0	0	0	0	0	0;
	2841	32.97	4.46	30	-10	1.027968	100	1	100	21.98	0	0	0	0	0	0	0	0	0	0	0;
	2842	18.58	13.67	30	-10	1.061581	100	1	100	12.39	0	0	0	0	0	0	0	0	0	0	0;
	2844	21.05	5.06	30	-10	1.070991	100	1	100	14.03	0	0	0	0	0	0	0	0	0	0	0;
	2845	21.72	6.98	30	-10	1.064384	100	1	100	14.48	0	0	0	0	0	0	0	0	0	0	0;
	2852	30	-0.55	30	-30	1.013948	100	1	100	20	0	0	0	0	0	0	0	0	0	0	0;
	2858	32.64	6.68	30	-10	1.044939	100	1	100	21.76	0	0	0	0	0	0	0	0	0	0	0;
	2870	42.35	6.03	40	-10	1.030832	100	1	100	28.23	0	0	0	0	0	0	0	0	0	0	0;
	2871	32.64	6.67	30	-10	1.044924	100	1	100	21.76	0	0	0	0	0	0	0	0	0	0	0;
	2877	36.42	7.96	30	-10	1.029166	100	1	100	24.28	0	0	0	0	0	0	0	0	0	0	0;
	2880	286.96	43.74	160	-40	1.028088	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	2881	286.96	43.74	160	-40	1.028088	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	2894	370.08	97.24	210	-50	1.045516	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	2895	209.49	209.89	210	-50	1.126501	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	2901	328.17	108.89	210	-50	1.039232	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	2902	328.17	108.89	210	-50	1.039232	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	2906	9.1	0.73	10	-10	1.026436	100	1	100	6.07	0	0	0	0	0	0	0	0	0	0	0;
	2912	19.55	6.06	30	-10	1.020176	100	1	100	13.03	0	0	0	0	0	0	0	0	0	0	0;
	2933	3.78	0.53	10	-10	0.989831	100	1	100	2.52	0	0	0	0	0	0	0	0	0	0	0;
	2934	225.92	112.91	210	-50	1.033975	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	2950	-142.46	-0.39	10	-10	1.006699	100	1	100	-189.95	0	0	0	0	0	0	0	0	0	0	0;
	2955	32.06	6.83	30	-10	1.054489	100	1	100	21.37	0	0	0	0	0	0	0	0	0	0	0;
	2963	34.37	6.67	30	-10	1.035406	100	1	100	22.91	0	0	0	0	0	0	0	0	0	0	0;
	2964	35.56	3.76	20	-10	1.036949	100	1	100	23.71	0	0	0	0	0	0	0	0	0	0	0;
	2968	36	6.82	30	-10	1.03481	100	1	100	24	0	0	0	0	0	0	0	0	0	0	0;
	2969	36	6.82	30	-10	1.03481	100	1	100	24	0	0	0	0	0	0	0	0	0	0	0;
	2970	16.71	0.01	30	-10	1.030988	100	1	100	11.14	0	0	0	0	0	0	0	0	0	0	0;
	2973	36	6.82	30	-10	1.03481	100	1	100	24	0	0	0	0	0	0	0	0	0	0	0;
	2974	36	6.82	30	-10	1.03481	100	1	100	24	0	0	0	0	0	0	0	0	0	0	0;
	2976	84.68	7.94	60	-20	1.00657	100	1	200	56.45	0	0	0	0	0	0	0	0	0	0	0;
	2978	35.84	15.66	40	-10	1.027523	100	1	100	23.89	0	0	0	0	0	0	0	0	0	0	0;
	2979	35.84	15.66	40	-10	1.027523	100	1	100	23.89	0	0	0	0	0	0	0	0	0	0	0;
	2980	35.84	15.66	40	-10	1.027523	100	1	100	23.89	0	0	0	0	0	0	0	0	0	0	0;
	2981	44.71	9.09	40	-10	1.024756	100	1	100	29.81	0	0	0	0	0	0	0	0	0	0	0;
	2982	44.71	9.09	40	-10	1.024756	100	1	100	29.81	0	0	0	0	0	0	0	0	0	0	0;
	2984	31.25	10.3	40	-10	1.024833	100	1	100	20.83	0	0	0	0	0	0	0	0	0	0	0;
	2990	159.8	8.06	100	-50	1.016588	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	2991	74.36	21.23	60	-20	1.032132	100	1	200	49.57	0	0	0	0	0	0	0	0	0	0	0;
	2994	50	-1.48	30	-20	0.985642	100	1	200	33.33	0	0	0	0	0	0	0	0	0	0	0;
	3012	200.42	8.27	160	-40	1.024412	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3013	299.46	57.87	160	-40	1.041662	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	3015	199.83	54.25	160	-40	1.033334	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	3017	94.74	2.75	60	-20	1.014104	100	1	200	63.16	0	0	0	0	0	0	0	0	0	0	0;
	3021	94.74	2.75	60	-20	1.014104	100	1	200	63.16	0	0	0	0	0	0	0	0	0	0	0;
	3022	94.74	2.75	60	-20	1.014104	100	1	200	63.16	0	0	0	0	0	0	0	0	0	0	0;
	3029	90.48	33.58	60	-20	1.043665	100	1	200	60.32	0	0	0	0	0	0	0	0	0	0	0;
	3031	37.21	7.27	30	-10	1.018242	100	1	100	24.81	0	0	0	0	0	0	0	0	0	0	0;
	3050	87.35	17.85	60	-20	1.033768	100	1	200	58.23	0	0	0	0	0	0	0	0	0	0	0;
	3058	46.56	9.04	30	-10	1.021353	100	1	100	31.04	0	0	0	0	0	0	0	0	0	0	0;
	3061	27.88	21.29	30	-10	1.041842	100	1	100	18.59	0	0	0	0	0	0	0	0	0	0	0;
	3062	27.88	21.29	30	-10	1.041842	100	1	100	18.59	0	0	0	0	0	0	0	0	0	0	0;
	3063	29.35	12.59	30	-10	1.05544	100	1	100	19.57	0	0	0	0	0	0	0	0	0	0	0;
	3065	279.13	130.91	210	-50	1.048637	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	3068	857.74	280.43	520	-120	1.043939	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	3072	328.39	49.41	210	-50	1.023058	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	3074	328.39	49.41	210	-50	1.023058	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	3082	55.25	0.75	30	-30	0.959658	100	1	200	36.83	0	0	0	0	0	0	0	0	0	0	0;
	3093	200.87	149.12	210	-50	1.062094	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3094	194.99	63.31	210	-50	1.048559	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	3095	366.07	56.62	210	-50	1.050106	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	3096	194.99	63.31	210	-50	1.048559	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	3097	366.07	56.62	210	-50	1.050106	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	3098	366.07	56.62	210	-50	1.050106	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	3100	200.87	149.13	210	-50	1.062073	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3101	200.87	149.12	210	-50	1.062094	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3103	84.19	8.48	60	-20	1.040259	100	1	200	56.13	0	0	0	0	0	0	0	0	0	0	0;
	3104	84.19	8.48	60	-20	1.040259	100	1	200	56.13	0	0	0	0	0	0	0	0	0	0	0;
	3107	33.38	14.38	30	-10	1.023574	100	1	100	22.25	0	0	0	0	0	0	0	0	0	0	0;
	3108	33.38	14.38	30	-10	1.023574	100	1	100	22.25	0	0	0	0	0	0	0	0	0	0	0;
	3109	33.37	14.47	30	-10	1.0234	100	1	100	22.25	0	0	0	0	0	0	0	0	0	0	0;
	3111	33.37	14.47	30	-10	1.0234	100	1	100	22.25	0	0	0	0	0	0	0	0	0	0	0;
	3112	25.03	21.04	30	-10	1.030367	100	1	100	16.69	0	0	0	0	0	0	0	0	0	0	0;
	3113	30.07	11.9	30	-10	1.030443	100	1	100	20.05	0	0	0	0	0	0	0	0	0	0	0;
	3114	24.37	4.5	30	-10	1.00571	100	1	100	16.25	0	0	0	0	0	0	0	0	0	0	0;
	3121	95.62	39.51	60	-20	1.057682	100	1	200	63.75	0	0	0	0	0	0	0	0	0	0	0;
	3122	95.62	39.51	60	-20	1.057682	100	1	200	63.75	0	0	0	0	0	0	0	0	0	0	0;
	3123	382.46	57.23	210	-50	1.050841	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	3124	382.46	57.23	210	-50	1.050841	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	3127	163.94	189.92	210	-50	1.050579	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	3128	383.05	40.91	210	-50	1.045612	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	3130	383.05	40.91	210	-50	1.045612	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	3134	82.3	25.62	60	-20	1.055592	100	1	200	54.87	0	0	0	0	0	0	0	0	0	0	0;
	3135	82.3	25.62	60	-20	1.055592	100	1	200	54.87	0	0	0	0	0	0	0	0	0	0	0;
	3136	82.3	25.62	60	-20	1.055592	100	1	200	54.87	0	0	0	0	0	0	0	0	0	0	0;
	3137	82.3	25.62	60	-20	1.055592	100	1	200	54.87	0	0	0	0	0	0	0	0	0	0	0;
	3139	87.35	17.85	60	-20	1.033768	100	1	200	58.23	0	0	0	0	0	0	0	0	0	0	0;
	3141	20.56	10.34	30	-10	1.03603	100	1	100	13.71	0	0	0	0	0	0	0	0	0	0	0;
	3142	20.56	10.34	30	-10	1.03603	100	1	100	13.71	0	0	0	0	0	0	0	0	0	0	0;
	3143	32.85	10.54	30	-10	1.034933	100	1	100	21.9	0	0	0	0	0	0	0	0	0	0	0;
	3145	90.32	6.14	60	-20	1.034547	100	1	200	60.21	0	0	0	0	0	0	0	0	0	0	0;
	3146	41.8	9.39	40	-10	1.028604	100	1	100	27.87	0	0	0	0	0	0	0	0	0	0	0;
	3148	7.36	0.9	10	-10	1.017868	100	1	100	4.91	0	0	0	0	0	0	0	0	0	0	0;
	3152	335.08	109.13	210	-50	1.05813	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	3154	89.73	17.85	60	-20	1.041223	100	1	200	59.82	0	0	0	0	0	0	0	0	0	0	0;
	3164	5.26	-0.96	10	-10	1.01814	100	1	100	3.51	0	0	0	0	0	0	0	0	0	0	0;
	3165	219.49	53.36	210	-50	1.029459	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3166	20.91	7.74	30	-10	1.014154	100	1	100	13.94	0	0	0	0	0	0	0	0	0	0	0;
	3167	202.63	63.88	160	-40	1.022861	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3168	202.63	63.88	160	-40	1.022861	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3169	202.63	63.88	160	-40	1.022861	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3176	36.8	6.62	30	-10	1.03058	100	1	100	24.53	0	0	0	0	0	0	0	0	0	0	0;
	3178	165	33.72	100	-20	0.962508	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	3179	261.4	37.15	100	-30	0.969605	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	3184	25.2	-3.93	20	-20	0.993105	100	1	100	16.8	0	0	0	0	0	0	0	0	0	0	0;
	3186	94.28	43.72	60	-20	1.048808	100	1	200	62.85	0	0	0	0	0	0	0	0	0	0	0;
	3201	155.5	12.42	100	-70	1.034282	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	3206	25.35	8.5	40	-10	1.018631	100	1	100	16.9	0	0	0	0	0	0	0	0	0	0	0;
	3210	60.07	-4.22	60	-20	1.016876	100	1	200	40.05	0	0	0	0	0	0	0	0	0	0	0;
	3213	60.07	-4.22	60	-20	1.016868	100	1	200	40.05	0	0	0	0	0	0	0	0	0	0	0;
	3214	60.07	-4.22	60	-20	1.016876	100	1	200	40.05	0	0	0	0	0	0	0	0	0	0	0;
	3215	60.07	-4.22	60	-20	1.016876	100	1	200	40.05	0	0	0	0	0	0	0	0	0	0	0;
	3216	60.07	-4.22	60	-20	1.016876	100	1	200	40.05	0	0	0	0	0	0	0	0	0	0	0;
	3217	83.51	27.65	60	-20	1.063297	100	1	200	55.67	0	0	0	0	0	0	0	0	0	0	0;
	3219	60.07	-4.22	60	-20	1.016868	100	1	200	40.05	0	0	0	0	0	0	0	0	0	0	0;
	3220	60.07	-4.22	60	-20	1.016868	100	1	200	40.05	0	0	0	0	0	0	0	0	0	0	0;
	3224	233.96	-20.31	210	-50	1.034075	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3225	37.32	7.31	30	-10	1.034314	100	1	100	24.88	0	0	0	0	0	0	0	0	0	0	0;
	3228	16.28	8.5	30	-10	1.027518	100	1	100	10.85	0	0	0	0	0	0	0	0	0	0	0;
	3233	37.32	7.31	30	-10	1.034314	100	1	100	24.88	0	0	0	0	0	0	0	0	0	0	0;
	3234	22.36	9.77	30	-10	1.014559	100	1	100	14.91	0	0	0	0	0	0	0	0	0	0	0;
	3236	31	0.14	20	-20	1.000449	100	1	100	20.67	0	0	0	0	0	0	0	0	0	0	0;
	3248	36.91	13.33	30	-10	1.037973	100	1	100	24.61	0	0	0	0	0	0	0	0	0	0	0;
	3255	311.23	83.72	210	-50	1.058355	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	3265	994.84	302.65	520	-120	1.05787	100	1	2000	600	0	0	0	0	0	0	0	0	0	0	0;
	3270	9.94	1.23	10	-10	0.98472	100	1	100	6.63	0	0	0	0	0	0	0	0	0	0	0;
	3280	237.72	5.93	160	-40	1.041353	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3282	86.61	21.7	60	-20	1.039025	100	1	200	57.74	0	0	0	0	0	0	0	0	0	0	0;
	3283	86.61	21.7	60	-20	1.039025	100	1	200	57.74	0	0	0	0	0	0	0	0	0	0	0;
	3284	86.61	21.7	60	-20	1.039025	100	1	200	57.74	0	0	0	0	0	0	0	0	0	0	0;
	3292	43.29	23.54	60	-20	1.035771	100	1	100	28.86	0	0	0	0	0	0	0	0	0	0	0;
	3294	228.6	9.57	190	-100	0.974759	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3296	72.61	29.56	60	-20	1.041629	100	1	200	48.41	0	0	0	0	0	0	0	0	0	0	0;
	3297	46.38	17.23	40	-10	1.048358	100	1	100	30.92	0	0	0	0	0	0	0	0	0	0	0;
	3298	46.38	17.23	40	-10	1.048358	100	1	100	30.92	0	0	0	0	0	0	0	0	0	0	0;
	3299	46.38	17.23	40	-10	1.048358	100	1	100	30.92	0	0	0	0	0	0	0	0	0	0	0;
	3306	46.38	17.23	40	-10	1.048358	100	1	100	30.92	0	0	0	0	0	0	0	0	0	0	0;
	3307	46.38	17.23	40	-10	1.048358	100	1	100	30.92	0	0	0	0	0	0	0	0	0	0	0;
	3308	38.42	21.93	30	-10	1.054031	100	1	100	25.61	0	0	0	0	0	0	0	0	0	0	0;
	3309	38.42	21.93	30	-10	1.054031	100	1	100	25.61	0	0	0	0	0	0	0	0	0	0	0;
	3323	28.83	4.62	30	-10	1.019144	100	1	100	19.22	0	0	0	0	0	0	0	0	0	0	0;
	3325	24.72	4.34	30	-10	1.03955	100	1	100	16.48	0	0	0	0	0	0	0	0	0	0	0;
	3326	24.72	4.34	30	-10	1.03955	100	1	100	16.48	0	0	0	0	0	0	0	0	0	0	0;
	3327	26.71	4.35	30	-10	1.039449	100	1	100	17.81	0	0	0	0	0	0	0	0	0	0	0;
	3328	26.71	4.35	30	-10	1.039449	100	1	100	17.81	0	0	0	0	0	0	0	0	0	0	0;
	3329	93.04	14.9	60	-20	1.017638	100	1	200	62.03	0	0	0	0	0	0	0	0	0	0	0;
	3330	21.12	7.89	20	-10	1.062419	100	1	100	14.08	0	0	0	0	0	0	0	0	0	0	0;
	3334	22.13	5.02	30	-10	1.031677	100	1	100	14.75	0	0	0	0	0	0	0	0	0	0	0;
	3335	22.13	5.02	30	-10	1.031677	100	1	100	14.75	0	0	0	0	0	0	0	0	0	0	0;
	3336	22.13	5.02	30	-10	1.031677	100	1	100	14.75	0	0	0	0	0	0	0	0	0	0	0;
	3337	22.13	5.02	30	-10	1.031677	100	1	100	14.75	0	0	0	0	0	0	0	0	0	0	0;
	3338	24.86	5.7	30	-10	0.997985	100	1	100	16.57	0	0	0	0	0	0	0	0	0	0	0;
	3339	22.13	5.02	30	-10	1.031677	100	1	100	14.75	0	0	0	0	0	0	0	0	0	0	0;
	3343	45.54	5.69	40	-10	1.044129	100	1	100	30.36	0	0	0	0	0	0	0	0	0	0	0;
	3344	172.98	103.13	160	-40	1.062747	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	3348	5.21	0.17	10	-10	1.02922	100	1	100	3.47	0	0	0	0	0	0	0	0	0	0	0;
	3351	19.26	2.33	30	-10	1.049756	100	1	100	12.84	0	0	0	0	0	0	0	0	0	0	0;
	3352	81.02	15.21	60	-20	1.035421	100	1	200	54.01	0	0	0	0	0	0	0	0	0	0	0;
	3360	0.43	-0.09	10	-10	1.024095	100	1	100	0.29	0	0	0	0	0	0	0	0	0	0	0;
	3361	25.75	4.04	30	-10	1.026025	100	1	100	17.17	0	0	0	0	0	0	0	0	0	0	0;
	3365	95.45	36.47	60	-20	1.060605	100	1	200	63.63	0	0	0	0	0	0	0	0	0	0	0;
	3366	95.45	36.47	60	-20	1.060605	100	1	200	63.63	0	0	0	0	0	0	0	0	0	0	0;
	3367	95.45	36.47	60	-20	1.060605	100	1	200	63.63	0	0	0	0	0	0	0	0	0	0	0;
	3368	95.45	36.47	60	-20	1.060605	100	1	200	63.63	0	0	0	0	0	0	0	0	0	0	0;
	3369	158.9	55.92	100	-30	1.000957	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	3371	19.75	-0.1	20	-20	1.002465	100	1	100	13.17	0	0	0	0	0	0	0	0	0	0	0;
	3384	35.95	6.39	40	-10	1.015185	100	1	100	23.97	0	0	0	0	0	0	0	0	0	0	0;
	3385	8.75	-0.1	10	-10	0.988654	100	1	100	5.83	0	0	0	0	0	0	0	0	0	0	0;
	3396	35.69	7.89	30	-10	1.017601	100	1	100	23.79	0	0	0	0	0	0	0	0	0	0	0;
	3403	67.68	4.49	60	-20	1.017078	100	1	200	45.12	0	0	0	0	0	0	0	0	0	0	0;
	3404	36.11	3.87	30	-10	1.024544	100	1	100	24.07	0	0	0	0	0	0	0	0	0	0	0;
	3409	15.52	0.02	10	-10	1.016693	100	1	100	10.35	0	0	0	0	0	0	0	0	0	0	0;
	3410	29.92	13.61	30	-10	1.031553	100	1	100	19.95	0	0	0	0	0	0	0	0	0	0	0;
	3412	86.41	23.5	60	-20	1.031898	100	1	200	57.61	0	0	0	0	0	0	0	0	0	0	0;
	3414	86.41	23.5	60	-20	1.031898	100	1	200	57.61	0	0	0	0	0	0	0	0	0	0	0;
	3415	86.41	23.5	60	-20	1.031898	100	1	200	57.61	0	0	0	0	0	0	0	0	0	0	0;
	3416	86.41	23.5	60	-20	1.031898	100	1	200	57.61	0	0	0	0	0	0	0	0	0	0	0;
	3418	87.19	27.18	60	-20	1.023338	100	1	200	58.13	0	0	0	0	0	0	0	0	0	0	0;
	3419	87.19	27.18	60	-20	1.023338	100	1	200	58.13	0	0	0	0	0	0	0	0	0	0	0;
	3432	29.4	8.8	40	-30	1.01809	100	1	100	19.6	0	0	0	0	0	0	0	0	0	0	0;
	3451	617.46	92.14	520	-120	1.032321	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	3456	32.69	0.72	40	-10	1.022332	100	1	100	21.79	0	0	0	0	0	0	0	0	0	0	0;
	3463	221.97	103.8	210	-50	1.063388	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3468	37.23	4.01	30	-10	1.028283	100	1	100	24.82	0	0	0	0	0	0	0	0	0	0	0;
	3469	73.92	33.18	210	-50	1.034629	100	1	200	49.28	0	0	0	0	0	0	0	0	0	0	0;
	3471	37.66	9.24	30	-10	1.060112	100	1	100	25.11	0	0	0	0	0	0	0	0	0	0	0;
	3481	35.74	6.84	30	-10	1.040138	100	1	100	23.83	0	0	0	0	0	0	0	0	0	0	0;
	3482	84.68	7.94	60	-20	1.00657	100	1	200	56.45	0	0	0	0	0	0	0	0	0	0	0;
	3485	47.94	3.29	40	-10	1.023602	100	1	100	31.96	0	0	0	0	0	0	0	0	0	0	0;
	3487	47.94	3.29	40	-10	1.023603	100	1	100	31.96	0	0	0	0	0	0	0	0	0	0	0;
	3488	15.73	-3.4	10	-20	0.975077	100	1	100	10.49	0	0	0	0	0	0	0	0	0	0	0;
	3489	47.94	3.29	40	-10	1.023602	100	1	100	31.96	0	0	0	0	0	0	0	0	0	0	0;
	3490	47.94	3.29	40	-10	1.023602	100	1	100	31.96	0	0	0	0	0	0	0	0	0	0	0;
	3491	38.05	10.9	30	-10	1.037557	100	1	100	25.37	0	0	0	0	0	0	0	0	0	0	0;
	3494	391.23	71.04	210	-50	1.036517	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	3495	391.23	71.04	210	-50	1.036517	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	3496	391.23	71.04	210	-50	1.036517	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	3499	31.3	-0.89	20	-20	0.998951	100	1	100	20.87	0	0	0	0	0	0	0	0	0	0	0;
	3508	24.86	7.42	30	-10	1.01212	100	1	100	16.57	0	0	0	0	0	0	0	0	0	0	0;
	3515	80.4	12.89	60	-20	1.045647	100	1	200	53.6	0	0	0	0	0	0	0	0	0	0	0;
	3517	31.79	7.29	30	-10	1.005112	100	1	100	21.19	0	0	0	0	0	0	0	0	0	0	0;
	3520	33.68	6.27	30	-10	1.046665	100	1	100	22.45	0	0	0	0	0	0	0	0	0	0	0;
	3521	33.68	6.27	30	-10	1.046665	100	1	100	22.45	0	0	0	0	0	0	0	0	0	0	0;
	3523	33.68	6.27	30	-10	1.046665	100	1	100	22.45	0	0	0	0	0	0	0	0	0	0	0;
	3524	33.68	6.27	30	-10	1.046665	100	1	100	22.45	0	0	0	0	0	0	0	0	0	0	0;
	3525	33.68	6.27	30	-10	1.046665	100	1	100	22.45	0	0	0	0	0	0	0	0	0	0	0;
	3526	33.68	6.27	30	-10	1.046665	100	1	100	22.45	0	0	0	0	0	0	0	0	0	0	0;
	3528	81.96	34.23	60	-20	1.031098	100	1	200	54.64	0	0	0	0	0	0	0	0	0	0	0;
	3529	81.96	34.23	60	-20	1.031098	100	1	200	54.64	0	0	0	0	0	0	0	0	0	0	0;
	3530	81.96	34.23	60	-20	1.031098	100	1	200	54.64	0	0	0	0	0	0	0	0	0	0	0;
	3535	90.48	33.58	60	-20	1.043665	100	1	200	60.32	0	0	0	0	0	0	0	0	0	0	0;
	3543	29.65	15.88	30	-10	1.051307	100	1	100	19.77	0	0	0	0	0	0	0	0	0	0	0;
	3551	209.93	1.9	210	-50	1.045549	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3553	219.68	138.48	160	-40	1.048632	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3557	863.42	301.07	520	-120	1.066429	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	3558	863.4	229.87	520	-120	1.06254	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	3567	50	-0.52	30	-20	0.992476	100	1	200	33.33	0	0	0	0	0	0	0	0	0	0	0;
	3569	38.22	19.69	30	-10	1.056202	100	1	100	25.48	0	0	0	0	0	0	0	0	0	0	0;
	3571	31.45	10.87	30	-10	1.043771	100	1	100	20.97	0	0	0	0	0	0	0	0	0	0	0;
	3577	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	3580	23.53	19.57	30	-10	1.054138	100	1	100	15.69	0	0	0	0	0	0	0	0	0	0	0;
	3582	23.53	19.57	30	-10	1.054138	100	1	100	15.69	0	0	0	0	0	0	0	0	0	0	0;
	3585	257.98	85.38	210	-50	1.037254	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	3586	257.98	85.38	210	-50	1.037254	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	3587	225.92	112.91	210	-50	1.033975	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3588	225.92	112.91	210	-50	1.033975	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3593	43.18	13.4	30	-10	1.017275	100	1	100	28.79	0	0	0	0	0	0	0	0	0	0	0;
	3607	200.19	-0.55	160	-40	1.039371	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3608	19.98	0.98	30	-10	1.029635	100	1	100	13.32	0	0	0	0	0	0	0	0	0	0	0;
	3610	332.36	180.64	210	-50	1.068762	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	3611	332.36	180.64	210	-50	1.068762	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	3617	21.77	6.35	30	-10	1.024499	100	1	100	14.51	0	0	0	0	0	0	0	0	0	0	0;
	3619	30.04	7.07	30	-10	1.043657	100	1	100	20.03	0	0	0	0	0	0	0	0	0	0	0;
	3620	30.04	7.07	30	-10	1.043657	100	1	100	20.03	0	0	0	0	0	0	0	0	0	0	0;
	3621	29.76	14.06	20	-10	1.030254	100	1	100	19.84	0	0	0	0	0	0	0	0	0	0	0;
	3625	29.84	3.47	30	-10	1.022039	100	1	100	19.89	0	0	0	0	0	0	0	0	0	0	0;
	3631	79.99	47.72	60	-20	1.050006	100	1	200	53.33	0	0	0	0	0	0	0	0	0	0	0;
	3633	26.75	6.78	40	-10	1.024572	100	1	100	17.83	0	0	0	0	0	0	0	0	0	0	0;
	3641	17.27	6.34	30	-10	1.03809	100	1	100	11.51	0	0	0	0	0	0	0	0	0	0	0;
	3642	17.27	6.34	30	-10	1.03809	100	1	100	11.51	0	0	0	0	0	0	0	0	0	0	0;
	3644	87.35	17.85	60	-20	1.033768	100	1	200	58.23	0	0	0	0	0	0	0	0	0	0	0;
	3646	23.01	6.63	30	-10	1.01964	100	1	100	15.34	0	0	0	0	0	0	0	0	0	0	0;
	3647	81.35	28.89	60	-20	1.055981	100	1	200	54.23	0	0	0	0	0	0	0	0	0	0	0;
	3651	96.29	19	60	-20	1.037131	100	1	200	64.19	0	0	0	0	0	0	0	0	0	0	0;
	3655	94.28	43.72	60	-20	1.048808	100	1	200	62.85	0	0	0	0	0	0	0	0	0	0	0;
	3656	96.29	19	60	-20	1.037131	100	1	200	64.19	0	0	0	0	0	0	0	0	0	0	0;
	3663	99.7	4.57	60	-20	1.028091	100	1	200	66.47	0	0	0	0	0	0	0	0	0	0	0;
	3671	93.11	0.33	60	-20	1.038581	100	1	200	62.07	0	0	0	0	0	0	0	0	0	0	0;
	3672	93.11	0.33	60	-20	1.038581	100	1	200	62.07	0	0	0	0	0	0	0	0	0	0	0;
	3674	4.27	0.56	10	-10	1.021115	100	1	100	2.85	0	0	0	0	0	0	0	0	0	0	0;
	3675	5.27	0.07	10	-10	1.025842	100	1	100	3.51	0	0	0	0	0	0	0	0	0	0	0;
	3680	20.91	7.74	30	-10	1.014154	100	1	100	13.94	0	0	0	0	0	0	0	0	0	0	0;
	3682	17	6.28	20	-10	1.023493	100	1	100	11.33	0	0	0	0	0	0	0	0	0	0	0;
	3683	36.8	6.62	30	-10	1.03058	100	1	100	24.53	0	0	0	0	0	0	0	0	0	0	0;
	3686	53.13	-3.75	40	-40	1.003179	100	1	200	35.42	0	0	0	0	0	0	0	0	0	0	0;
	3695	54.83	12.78	40	-10	1.047752	100	1	200	36.55	0	0	0	0	0	0	0	0	0	0	0;
	3696	54.83	12.78	40	-10	1.047752	100	1	200	36.55	0	0	0	0	0	0	0	0	0	0	0;
	3697	54.83	12.78	40	-10	1.047752	100	1	200	36.55	0	0	0	0	0	0	0	0	0	0	0;
	3698	54.83	12.78	40	-10	1.047752	100	1	200	36.55	0	0	0	0	0	0	0	0	0	0	0;
	3699	54.83	12.78	40	-10	1.047752	100	1	200	36.55	0	0	0	0	0	0	0	0	0	0	0;
	3700	54.83	12.78	40	-10	1.047752	100	1	200	36.55	0	0	0	0	0	0	0	0	0	0	0;
	3701	54.83	12.78	40	-10	1.047752	100	1	200	36.55	0	0	0	0	0	0	0	0	0	0	0;
	3710	95.02	8.09	60	-20	1.023144	100	1	200	63.35	0	0	0	0	0	0	0	0	0	0	0;
	3712	36.6	6.27	30	-10	1.027328	100	1	100	24.4	0	0	0	0	0	0	0	0	0	0	0;
	3714	36.6	6.27	30	-10	1.027328	100	1	100	24.4	0	0	0	0	0	0	0	0	0	0	0;
	3717	91.41	18.38	60	-20	1.020566	100	1	200	60.94	0	0	0	0	0	0	0	0	0	0	0;
	3718	91.41	18.38	60	-20	1.020566	100	1	200	60.94	0	0	0	0	0	0	0	0	0	0	0;
	3720	91.41	18.38	60	-20	1.020566	100	1	200	60.94	0	0	0	0	0	0	0	0	0	0	0;
	3724	586.88	407.04	520	-120	1.052406	100	1	1200	300	0	0	0	0	0	0	0	0	0	0	0;
	3728	74.79	37.77	60	-20	1.041161	100	1	200	49.86	0	0	0	0	0	0	0	0	0	0	0;
	3729	74.79	37.77	60	-20	1.041161	100	1	200	49.86	0	0	0	0	0	0	0	0	0	0	0;
	3738	25	6.1	10	0	1.024613	100	1	100	16.67	0	0	0	0	0	0	0	0	0	0	0;
	3754	285.52	89.66	210	-50	1.04459	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	3757	285.52	89.66	210	-50	1.04459	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	3763	46.38	17.23	40	-10	1.048358	100	1	100	30.92	0	0	0	0	0	0	0	0	0	0	0;
	3771	26.56	10.95	30	-10	1.002675	100	1	100	17.71	0	0	0	0	0	0	0	0	0	0	0;
	3781	80.4	12.89	60	-20	1.045647	100	1	200	53.6	0	0	0	0	0	0	0	0	0	0	0;
	3782	80.4	12.89	60	-20	1.045647	100	1	200	53.6	0	0	0	0	0	0	0	0	0	0	0;
	3787	80.4	12.89	60	-20	1.045647	100	1	200	53.6	0	0	0	0	0	0	0	0	0	0	0;
	3792	72.61	29.56	60	-20	1.041629	100	1	200	48.41	0	0	0	0	0	0	0	0	0	0	0;
	3797	344.6	7.8	180	-140	0.97953	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	3800	69.42	1.26	60	-20	1.034738	100	1	200	46.28	0	0	0	0	0	0	0	0	0	0	0;
	3810	22.28	12.48	30	-10	1.043637	100	1	100	14.85	0	0	0	0	0	0	0	0	0	0	0;
	3811	38.05	10.9	30	-10	1.037557	100	1	100	25.37	0	0	0	0	0	0	0	0	0	0	0;
	3812	38.05	10.9	30	-10	1.037557	100	1	100	25.37	0	0	0	0	0	0	0	0	0	0	0;
	3814	7.28	-1.45	10	-10	0.990319	100	1	100	4.85	0	0	0	0	0	0	0	0	0	0	0;
	3815	26.31	9.5	30	-10	1.038225	100	1	100	17.54	0	0	0	0	0	0	0	0	0	0	0;
	3816	26.31	9.5	30	-10	1.038225	100	1	100	17.54	0	0	0	0	0	0	0	0	0	0	0;
	3820	22.28	12.48	30	-10	1.043637	100	1	100	14.85	0	0	0	0	0	0	0	0	0	0	0;
	3823	74.8	42.23	60	-20	1.046905	100	1	200	49.87	0	0	0	0	0	0	0	0	0	0	0;
	3838	26.29	7.55	30	-10	1.004212	100	1	100	17.53	0	0	0	0	0	0	0	0	0	0	0;
	3842	3.2	1.16	10	-10	0.991765	100	1	100	2.13	0	0	0	0	0	0	0	0	0	0	0;
	3845	37.08	5.51	30	-10	1.021005	100	1	100	24.72	0	0	0	0	0	0	0	0	0	0	0;
	3848	9	-0.95	10	-10	0.943602	100	1	100	6	0	0	0	0	0	0	0	0	0	0	0;
	3853	33.75	8.82	40	-10	0.999522	100	1	100	22.5	0	0	0	0	0	0	0	0	0	0	0;
	3854	30.9	1.92	30	-10	1.015954	100	1	100	20.6	0	0	0	0	0	0	0	0	0	0	0;
	3857	46.26	7.53	40	-10	1.015006	100	1	100	30.84	0	0	0	0	0	0	0	0	0	0	0;
	3866	239.85	38.05	160	-40	1.032762	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3867	239.85	38.05	160	-40	1.032762	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3872	6	1	10	-10	0.972026	100	1	100	4	0	0	0	0	0	0	0	0	0	0	0;
	3877	35.95	6.39	40	-10	1.015185	100	1	100	23.97	0	0	0	0	0	0	0	0	0	0	0;
	3878	27.67	4.45	30	-10	1.027132	100	1	100	18.45	0	0	0	0	0	0	0	0	0	0	0;
	3880	18.1	0.8	10	-10	1.007227	100	1	100	12.07	0	0	0	0	0	0	0	0	0	0	0;
	3893	16.08	-1.58	10	-10	1.025595	100	1	100	10.72	0	0	0	0	0	0	0	0	0	0	0;
	3894	3.84	0.45	10	-10	1.030843	100	1	100	2.56	0	0	0	0	0	0	0	0	0	0	0;
	3899	36.42	7.96	30	-10	1.029166	100	1	100	24.28	0	0	0	0	0	0	0	0	0	0	0;
	3903	49.08	7.63	60	-20	0.988537	100	1	100	32.72	0	0	0	0	0	0	0	0	0	0	0;
	3904	93.77	48.62	60	-20	1.054388	100	1	200	62.51	0	0	0	0	0	0	0	0	0	0	0;
	3906	28.92	12.55	40	-10	1.019357	100	1	100	19.28	0	0	0	0	0	0	0	0	0	0	0;
	3907	28.92	12.55	40	-10	1.019357	100	1	100	19.28	0	0	0	0	0	0	0	0	0	0	0;
	3908	28.92	12.55	40	-10	1.019357	100	1	100	19.28	0	0	0	0	0	0	0	0	0	0	0;
	3909	28.92	12.55	40	-10	1.019357	100	1	100	19.28	0	0	0	0	0	0	0	0	0	0	0;
	3910	28.92	12.55	40	-10	1.019357	100	1	100	19.28	0	0	0	0	0	0	0	0	0	0	0;
	3911	28.92	12.55	40	-10	1.019357	100	1	100	19.28	0	0	0	0	0	0	0	0	0	0	0;
	3912	49.34	9.15	40	-10	1.034597	100	1	100	32.89	0	0	0	0	0	0	0	0	0	0	0;
	3913	52.7	8.59	60	-20	1.008149	100	1	200	35.13	0	0	0	0	0	0	0	0	0	0	0;
	3914	93.77	48.62	60	-20	1.054388	100	1	200	62.51	0	0	0	0	0	0	0	0	0	0	0;
	3915	36.77	13.93	30	-10	1.042317	100	1	100	24.51	0	0	0	0	0	0	0	0	0	0	0;
	3916	28.12	1.08	30	-10	1.028994	100	1	100	18.75	0	0	0	0	0	0	0	0	0	0	0;
	3922	100.29	29.56	60	-20	1.064335	100	1	300	66.86	0	0	0	0	0	0	0	0	0	0	0;
	3926	230	25.76	140	-50	0.984593	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3929	18.32	-0.94	10	-10	0.995077	100	1	100	12.21	0	0	0	0	0	0	0	0	0	0	0;
	3934	30.46	6.53	20	-10	0.997643	100	1	100	20.31	0	0	0	0	0	0	0	0	0	0	0;
	3939	210	15.07	130	-60	0.980209	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3943	174.02	89.48	210	-50	1.046877	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	3947	30.07	6.48	30	-10	1.058835	100	1	100	20.05	0	0	0	0	0	0	0	0	0	0	0;
	3948	58.8	-10.54	30	-30	0.9646	100	1	200	39.2	0	0	0	0	0	0	0	0	0	0	0;
	3962	88.32	56.36	60	-20	1.127961	100	1	200	58.88	0	0	0	0	0	0	0	0	0	0	0;
	3968	44.1	5.46	50	-20	0.988831	100	1	100	29.4	0	0	0	0	0	0	0	0	0	0	0;
	3971	72.47	11.44	60	-20	1.019231	100	1	200	48.31	0	0	0	0	0	0	0	0	0	0	0;
	3984	227.37	10.01	160	-40	1.028414	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	3990	27.19	6.05	30	-10	1.014542	100	1	100	18.13	0	0	0	0	0	0	0	0	0	0	0;
	3997	96.9	14.27	60	-20	1.033735	100	1	200	64.6	0	0	0	0	0	0	0	0	0	0	0;
	4000	100.41	11.92	60	-20	1.038747	100	1	300	66.94	0	0	0	0	0	0	0	0	0	0	0;
	4004	90.48	33.58	60	-20	1.043665	100	1	200	60.32	0	0	0	0	0	0	0	0	0	0	0;
	4009	0.08	-0.15	10	-10	1.016106	100	1	100	0.05	0	0	0	0	0	0	0	0	0	0	0;
	4011	25.24	9.06	30	-10	1.045295	100	1	100	16.83	0	0	0	0	0	0	0	0	0	0	0;
	4013	200.33	74.16	210	-50	1.024096	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	4020	28.05	6.93	30	-10	1.035469	100	1	100	18.7	0	0	0	0	0	0	0	0	0	0	0;
	4021	28.05	6.93	30	-10	1.035469	100	1	100	18.7	0	0	0	0	0	0	0	0	0	0	0;
	4022	38.97	3.58	30	-10	1.048029	100	1	100	25.98	0	0	0	0	0	0	0	0	0	0	0;
	4024	27.42	7.69	30	-10	1.049669	100	1	100	18.28	0	0	0	0	0	0	0	0	0	0	0;
	4025	27.42	7.69	30	-10	1.049669	100	1	100	18.28	0	0	0	0	0	0	0	0	0	0	0;
	4026	23.99	10.2	30	-10	1.045658	100	1	100	15.99	0	0	0	0	0	0	0	0	0	0	0;
	4027	23.99	10.2	30	-10	1.045658	100	1	100	15.99	0	0	0	0	0	0	0	0	0	0	0;
	4039	30.91	13.15	30	-10	1.040404	100	1	100	20.61	0	0	0	0	0	0	0	0	0	0	0;
	4040	30.91	13.15	30	-10	1.040404	100	1	100	20.61	0	0	0	0	0	0	0	0	0	0	0;
	4041	17.82	3.95	30	-10	1.066829	100	1	100	11.88	0	0	0	0	0	0	0	0	0	0	0;
	4042	30.91	13.15	30	-10	1.040404	100	1	100	20.61	0	0	0	0	0	0	0	0	0	0	0;
	4043	28.62	19.42	30	-10	1.048055	100	1	100	19.08	0	0	0	0	0	0	0	0	0	0	0;
	4044	28.62	19.42	30	-10	1.048055	100	1	100	19.08	0	0	0	0	0	0	0	0	0	0	0;
	4045	39.58	7.79	30	-10	1.022502	100	1	100	26.39	0	0	0	0	0	0	0	0	0	0	0;
	4047	36.43	14.37	30	-10	1.048211	100	1	100	24.29	0	0	0	0	0	0	0	0	0	0	0;
	4048	26.97	7.32	30	-10	1.049221	100	1	100	17.98	0	0	0	0	0	0	0	0	0	0	0;
	4049	29.57	20.51	30	-10	1.041409	100	1	100	19.71	0	0	0	0	0	0	0	0	0	0	0;
	4052	42.63	9.46	30	-10	1.001216	100	1	100	28.42	0	0	0	0	0	0	0	0	0	0	0;
	4053	90.32	31.5	60	-20	1.035102	100	1	200	60.21	0	0	0	0	0	0	0	0	0	0	0;
	4054	90.32	31.5	60	-20	1.035102	100	1	200	60.21	0	0	0	0	0	0	0	0	0	0	0;
	4056	90.32	31.5	60	-20	1.035102	100	1	200	60.21	0	0	0	0	0	0	0	0	0	0	0;
	4057	86.85	27.93	60	-20	1.028397	100	1	200	57.9	0	0	0	0	0	0	0	0	0	0	0;
	4058	86.85	27.93	60	-20	1.028397	100	1	200	57.9	0	0	0	0	0	0	0	0	0	0	0;
	4059	86.85	27.93	60	-20	1.028397	100	1	200	57.9	0	0	0	0	0	0	0	0	0	0	0;
	4060	86.85	27.93	60	-20	1.028397	100	1	200	57.9	0	0	0	0	0	0	0	0	0	0	0;
	4061	86.85	27.93	60	-20	1.028397	100	1	200	57.9	0	0	0	0	0	0	0	0	0	0	0;
	4062	86.85	27.93	60	-20	1.028397	100	1	200	57.9	0	0	0	0	0	0	0	0	0	0	0;
	4063	43.29	23.54	60	-20	1.035771	100	1	100	28.86	0	0	0	0	0	0	0	0	0	0	0;
	4066	49.88	7.11	40	-10	1.013219	100	1	100	33.25	0	0	0	0	0	0	0	0	0	0	0;
	4067	49.88	7.11	40	-10	1.013219	100	1	100	33.25	0	0	0	0	0	0	0	0	0	0	0;
	4068	49.88	7.11	40	-10	1.013219	100	1	100	33.25	0	0	0	0	0	0	0	0	0	0	0;
	4069	49.88	7.11	40	-10	1.013219	100	1	100	33.25	0	0	0	0	0	0	0	0	0	0	0;
	4074	31.92	6.07	30	-10	1.027264	100	1	100	21.28	0	0	0	0	0	0	0	0	0	0	0;
	4083	277.24	62.37	160	-40	1.023361	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4084	277.21	61.8	160	-40	1.023418	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4085	5	2.04	10	-10	1.009821	100	1	100	3.33	0	0	0	0	0	0	0	0	0	0	0;
	4087	277.21	61.8	160	-40	1.023418	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4102	319.18	33.68	210	-50	1.026027	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	4116	-143.87	-1.92	20	-20	0.96498	100	1	100	-191.83	0	0	0	0	0	0	0	0	0	0	0;
	4120	74.78	41.98	60	-20	1.04914	100	1	200	49.85	0	0	0	0	0	0	0	0	0	0	0;
	4126	34.43	7.93	30	-10	1.013384	100	1	100	22.95	0	0	0	0	0	0	0	0	0	0	0;
	4130	223.9	28.79	160	-40	0.992143	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	4133	36.8	6.62	30	-10	1.03058	100	1	100	24.53	0	0	0	0	0	0	0	0	0	0	0;
	4135	40.49	8.06	30	-10	1.036966	100	1	100	26.99	0	0	0	0	0	0	0	0	0	0	0;
	4136	59.71	48.45	60	-20	1.060489	100	1	200	39.81	0	0	0	0	0	0	0	0	0	0	0;
	4137	78.19	5.27	60	-20	1.025746	100	1	200	52.13	0	0	0	0	0	0	0	0	0	0	0;
	4138	78.19	5.27	60	-20	1.025746	100	1	200	52.13	0	0	0	0	0	0	0	0	0	0	0;
	4139	78.19	5.27	60	-20	1.025746	100	1	200	52.13	0	0	0	0	0	0	0	0	0	0	0;
	4140	78.21	5.29	60	-20	1.025803	100	1	200	52.14	0	0	0	0	0	0	0	0	0	0	0;
	4141	78.21	5.29	60	-20	1.025803	100	1	200	52.14	0	0	0	0	0	0	0	0	0	0	0;
	4143	78.21	5.29	60	-20	1.025803	100	1	200	52.14	0	0	0	0	0	0	0	0	0	0	0;
	4144	76.33	5.57	60	-20	1.027882	100	1	200	50.89	0	0	0	0	0	0	0	0	0	0	0;
	4145	95.62	39.51	60	-20	1.057682	100	1	200	63.75	0	0	0	0	0	0	0	0	0	0	0;
	4151	29.25	3.13	40	-10	1.008035	100	1	100	19.5	0	0	0	0	0	0	0	0	0	0	0;
	4152	32.29	8.39	40	-10	1.015318	100	1	100	21.53	0	0	0	0	0	0	0	0	0	0	0;
	4154	32.72	7.26	30	-10	1.016616	100	1	100	21.81	0	0	0	0	0	0	0	0	0	0	0;
	4155	76.7	4.41	60	-20	1.02424	100	1	200	51.13	0	0	0	0	0	0	0	0	0	0	0;
	4156	76.7	4.41	60	-20	1.02424	100	1	200	51.13	0	0	0	0	0	0	0	0	0	0	0;
	4157	97.15	3.14	60	-20	1.026566	100	1	200	64.77	0	0	0	0	0	0	0	0	0	0	0;
	4158	97.15	3.14	60	-20	1.026566	100	1	200	64.77	0	0	0	0	0	0	0	0	0	0	0;
	4159	69.46	4.37	60	-20	1.041381	100	1	200	46.31	0	0	0	0	0	0	0	0	0	0	0;
	4160	97.15	3.14	60	-20	1.026566	100	1	200	64.77	0	0	0	0	0	0	0	0	0	0	0;
	4161	69.46	4.37	60	-20	1.041381	100	1	200	46.31	0	0	0	0	0	0	0	0	0	0	0;
	4162	31.64	8.73	40	-10	0.997106	100	1	100	21.09	0	0	0	0	0	0	0	0	0	0	0;
	4163	73.94	10.77	60	-20	1.027512	100	1	200	49.29	0	0	0	0	0	0	0	0	0	0	0;
	4164	73.94	10.77	60	-20	1.027512	100	1	200	49.29	0	0	0	0	0	0	0	0	0	0	0;
	4166	18.58	13.67	30	-10	1.061581	100	1	100	12.39	0	0	0	0	0	0	0	0	0	0	0;
	4167	31.57	7.72	30	-10	1.008948	100	1	100	21.05	0	0	0	0	0	0	0	0	0	0	0;
	4171	27.95	4.02	30	-10	1.016461	100	1	100	18.63	0	0	0	0	0	0	0	0	0	0	0;
	4181	39.44	10.06	40	-10	1.021987	100	1	100	26.29	0	0	0	0	0	0	0	0	0	0	0;
	4182	80.74	25.03	60	-20	1.061099	100	1	200	53.83	0	0	0	0	0	0	0	0	0	0	0;
	4183	80.74	25.03	60	-20	1.061099	100	1	200	53.83	0	0	0	0	0	0	0	0	0	0	0;
	4185	80.74	25.03	60	-20	1.061099	100	1	200	53.83	0	0	0	0	0	0	0	0	0	0	0;
	4187	80.74	25.03	60	-20	1.061099	100	1	200	53.83	0	0	0	0	0	0	0	0	0	0	0;
	4190	50	-1.48	30	-20	0.985642	100	1	200	33.33	0	0	0	0	0	0	0	0	0	0	0;
	4191	36.8	6.62	30	-10	1.03058	100	1	100	24.53	0	0	0	0	0	0	0	0	0	0	0;
	4193	38.51	8.85	30	-10	1.03425	100	1	100	25.67	0	0	0	0	0	0	0	0	0	0	0;
	4195	300.49	77.66	210	-50	1.049436	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	4196	5	0.52	10	-10	1.019692	100	1	100	3.33	0	0	0	0	0	0	0	0	0	0	0;
	4197	360.13	84.68	210	-50	1.005717	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	4200	21.05	5.06	30	-10	1.070991	100	1	100	14.03	0	0	0	0	0	0	0	0	0	0	0;
	4202	300	-14.29	100	-100	1.001419	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	4203	74.8	42.23	60	-20	1.046905	100	1	200	49.87	0	0	0	0	0	0	0	0	0	0	0;
	4206	50.29	10.48	40	-10	1.031165	100	1	200	33.53	0	0	0	0	0	0	0	0	0	0	0;
	4207	91.34	21.91	60	-20	1.030656	100	1	200	60.89	0	0	0	0	0	0	0	0	0	0	0;
	4208	151.81	0.39	100	-50	1.017341	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	4209	95.02	8.09	60	-20	1.023144	100	1	200	63.35	0	0	0	0	0	0	0	0	0	0	0;
	4214	91.34	21.91	60	-20	1.030656	100	1	200	60.89	0	0	0	0	0	0	0	0	0	0	0;
	4217	25.35	8.5	40	-10	1.018631	100	1	100	16.9	0	0	0	0	0	0	0	0	0	0	0;
	4226	262.9	48.53	160	-40	1.027461	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4234	45.38	12.31	60	-20	0.990977	100	1	100	30.25	0	0	0	0	0	0	0	0	0	0	0;
	4235	12.3	0.28	10	-10	1.011745	100	1	100	8.2	0	0	0	0	0	0	0	0	0	0	0;
	4237	6.52	-0.06	10	-10	0.994601	100	1	100	4.35	0	0	0	0	0	0	0	0	0	0	0;
	4240	67	8.72	40	-10	0.992643	100	1	200	44.67	0	0	0	0	0	0	0	0	0	0	0;
	4242	72.17	16.24	60	-20	1.022616	100	1	200	48.11	0	0	0	0	0	0	0	0	0	0	0;
	4245	227.21	72.31	210	-50	1.063358	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	4252	34.86	4.81	30	-10	1.050901	100	1	100	23.24	0	0	0	0	0	0	0	0	0	0	0;
	4253	165.7	155.15	160	-40	1.077217	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	4254	165.7	155.15	160	-40	1.077217	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	4256	22.38	13.57	30	-10	1.011374	100	1	100	14.92	0	0	0	0	0	0	0	0	0	0	0;
	4259	8.4	-0.84	10	-10	1.010516	100	1	100	5.6	0	0	0	0	0	0	0	0	0	0	0;
	4261	158.9	35.36	100	-30	0.993468	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	4264	36.42	7.96	30	-10	1.029166	100	1	100	24.28	0	0	0	0	0	0	0	0	0	0	0;
	4267	36.8	-2.54	30	-30	0.98379	100	1	100	24.53	0	0	0	0	0	0	0	0	0	0	0;
	4274	227.2	72.57	210	-50	1.063453	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	4282	36.22	11.19	30	-10	1.003847	100	1	100	24.15	0	0	0	0	0	0	0	0	0	0	0;
	4284	95.25	8.91	60	-20	1.058465	100	1	200	63.5	0	0	0	0	0	0	0	0	0	0	0;
	4285	95.25	8.91	60	-20	1.058465	100	1	200	63.5	0	0	0	0	0	0	0	0	0	0	0;
	4286	95.25	8.91	60	-20	1.058465	100	1	200	63.5	0	0	0	0	0	0	0	0	0	0	0;
	4287	95.25	8.91	60	-20	1.058465	100	1	200	63.5	0	0	0	0	0	0	0	0	0	0	0;
	4288	95.25	8.91	60	-20	1.058465	100	1	200	63.5	0	0	0	0	0	0	0	0	0	0	0;
	4290	95.25	8.91	60	-20	1.058465	100	1	200	63.5	0	0	0	0	0	0	0	0	0	0	0;
	4298	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	4301	93.04	14.9	60	-20	1.017638	100	1	200	62.03	0	0	0	0	0	0	0	0	0	0	0;
	4309	72.61	29.56	60	-20	1.041629	100	1	200	48.41	0	0	0	0	0	0	0	0	0	0	0;
	4310	380.17	55.84	210	-50	1.059635	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	4314	48.81	10.42	40	-10	1.017273	100	1	100	32.54	0	0	0	0	0	0	0	0	0	0	0;
	4324	27.81	13.1	30	-10	1.016415	100	1	100	18.54	0	0	0	0	0	0	0	0	0	0	0;
	4326	29.43	7.09	30	-10	1.028438	100	1	100	19.62	0	0	0	0	0	0	0	0	0	0	0;
	4331	50	-0.52	30	-20	0.992476	100	1	200	33.33	0	0	0	0	0	0	0	0	0	0	0;
	4337	20.16	7.15	20	-10	1.020298	100	1	100	13.44	0	0	0	0	0	0	0	0	0	0	0;
	4339	37.67	5.32	20	-10	1.027032	100	1	100	25.12	0	0	0	0	0	0	0	0	0	0	0;
	4343	48.81	10.42	40	-10	1.017273	100	1	100	32.54	0	0	0	0	0	0	0	0	0	0	0;
	4344	18.74	6.56	30	-10	1.02868	100	1	100	12.49	0	0	0	0	0	0	0	0	0	0	0;
	4345	48.81	10.42	40	-10	1.017273	100	1	100	32.54	0	0	0	0	0	0	0	0	0	0	0;
	4347	83.78	15.02	60	-20	1.033534	100	1	200	55.85	0	0	0	0	0	0	0	0	0	0	0;
	4353	50.6	9.87	30	-20	1.012422	100	1	200	33.73	0	0	0	0	0	0	0	0	0	0	0;
	4354	37.08	5.51	30	-10	1.021005	100	1	100	24.72	0	0	0	0	0	0	0	0	0	0	0;
	4366	25.01	5.98	30	-10	0.999838	100	1	100	16.67	0	0	0	0	0	0	0	0	0	0	0;
	4367	350.88	-13.06	210	-50	1.037481	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	4371	305.43	40.11	210	-50	1.031705	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	4376	328.01	53.95	210	-50	1.05097	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	4378	328.01	53.95	210	-50	1.05097	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	4379	328.01	53.95	210	-50	1.05097	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	4380	328.01	53.95	210	-50	1.05097	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	4384	38.02	11.21	30	-10	1.043087	100	1	100	25.35	0	0	0	0	0	0	0	0	0	0	0;
	4385	18.92	10.7	30	-10	1.056129	100	1	100	12.61	0	0	0	0	0	0	0	0	0	0	0;
	4386	18.92	10.7	30	-10	1.056129	100	1	100	12.61	0	0	0	0	0	0	0	0	0	0	0;
	4387	18.92	10.7	30	-10	1.056129	100	1	100	12.61	0	0	0	0	0	0	0	0	0	0	0;
	4392	29.4	6.58	40	-20	1.034872	100	1	100	19.6	0	0	0	0	0	0	0	0	0	0	0;
	4393	62.61	24.83	60	-20	1.034646	100	1	200	41.74	0	0	0	0	0	0	0	0	0	0	0;
	4400	277.26	31.82	210	-50	1.04876	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4403	57.51	17.72	40	-10	1.029323	100	1	200	38.34	0	0	0	0	0	0	0	0	0	0	0;
	4405	24.65	6.97	30	-10	1.038344	100	1	100	16.43	0	0	0	0	0	0	0	0	0	0	0;
	4411	39.43	5.93	30	-10	1.017582	100	1	100	26.29	0	0	0	0	0	0	0	0	0	0	0;
	4422	53.99	18.23	40	-10	1.062896	100	1	200	35.99	0	0	0	0	0	0	0	0	0	0	0;
	4423	2.5	0.07	10	-10	0.984601	100	1	100	1.67	0	0	0	0	0	0	0	0	0	0	0;
	4424	53.99	18.23	40	-10	1.062896	100	1	200	35.99	0	0	0	0	0	0	0	0	0	0	0;
	4425	53.99	18.23	40	-10	1.062896	100	1	200	35.99	0	0	0	0	0	0	0	0	0	0	0;
	4426	27.96	5.11	30	-10	1.029009	100	1	100	18.64	0	0	0	0	0	0	0	0	0	0	0;
	4427	27.96	5.11	30	-10	1.029009	100	1	100	18.64	0	0	0	0	0	0	0	0	0	0	0;
	4430	17.09	6.06	30	-10	1.033479	100	1	100	11.39	0	0	0	0	0	0	0	0	0	0	0;
	4432	27.73	2	30	-10	1.030407	100	1	100	18.49	0	0	0	0	0	0	0	0	0	0	0;
	4433	27.73	2	30	-10	1.030407	100	1	100	18.49	0	0	0	0	0	0	0	0	0	0	0;
	4434	27.73	2	30	-10	1.030407	100	1	100	18.49	0	0	0	0	0	0	0	0	0	0	0;
	4440	36.77	13.93	30	-10	1.042317	100	1	100	24.51	0	0	0	0	0	0	0	0	0	0	0;
	4443	353.02	203.37	210	-50	1.128032	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	4448	264.57	207.6	210	-50	1.127182	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4449	34.37	6.67	30	-10	1.035406	100	1	100	22.91	0	0	0	0	0	0	0	0	0	0	0;
	4451	47.36	7.24	40	-10	1.026147	100	1	100	31.57	0	0	0	0	0	0	0	0	0	0	0;
	4452	200.58	41.06	210	-50	1.044746	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	4453	47.36	7.24	40	-10	1.026147	100	1	100	31.57	0	0	0	0	0	0	0	0	0	0	0;
	4454	39.82	7.25	40	-10	1.026722	100	1	100	26.55	0	0	0	0	0	0	0	0	0	0	0;
	4455	58.52	7.84	40	-10	1.025977	100	1	200	39.01	0	0	0	0	0	0	0	0	0	0	0;
	4456	39.82	7.25	40	-10	1.026722	100	1	100	26.55	0	0	0	0	0	0	0	0	0	0	0;
	4457	39.82	7.25	40	-10	1.026722	100	1	100	26.55	0	0	0	0	0	0	0	0	0	0	0;
	4458	42.33	15.72	40	-10	1.031695	100	1	100	28.22	0	0	0	0	0	0	0	0	0	0	0;
	4460	52.53	15.88	40	-10	1.02748	100	1	200	35.02	0	0	0	0	0	0	0	0	0	0	0;
	4461	47.36	7.24	40	-10	1.026147	100	1	100	31.57	0	0	0	0	0	0	0	0	0	0	0;
	4470	95.62	39.51	60	-20	1.057682	100	1	200	63.75	0	0	0	0	0	0	0	0	0	0	0;
	4473	188.54	108.59	210	-50	1.037029	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	4480	17.64	4.13	30	-10	1.031189	100	1	100	11.76	0	0	0	0	0	0	0	0	0	0	0;
	4481	33.27	9.43	40	-10	1.025116	100	1	100	22.18	0	0	0	0	0	0	0	0	0	0	0;
	4483	32.98	7.31	30	-10	1.054159	100	1	100	21.99	0	0	0	0	0	0	0	0	0	0	0;
	4484	33.28	7.19	40	-10	1.022819	100	1	100	22.19	0	0	0	0	0	0	0	0	0	0	0;
	4488	3.65	0.04	10	-10	0.982004	100	1	100	2.43	0	0	0	0	0	0	0	0	0	0	0;
	4495	100.41	11.92	60	-20	1.038747	100	1	300	66.94	0	0	0	0	0	0	0	0	0	0	0;
	4499	72.47	11.44	60	-20	1.019231	100	1	200	48.31	0	0	0	0	0	0	0	0	0	0	0;
	4501	32.52	9.41	30	-10	1.028707	100	1	100	21.68	0	0	0	0	0	0	0	0	0	0	0;
	4505	227.37	10.01	160	-40	1.028414	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	4509	550.03	481.13	520	-120	1.061212	100	1	1200	300	0	0	0	0	0	0	0	0	0	0	0;
	4512	32.52	9.41	30	-10	1.028707	100	1	100	21.68	0	0	0	0	0	0	0	0	0	0	0;
	4513	20.33	6.46	30	-10	0.990908	100	1	100	13.55	0	0	0	0	0	0	0	0	0	0	0;
	4517	290	-23.36	100	-100	0.971901	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4526	85	-13.61	60	-60	0.969146	100	1	200	56.67	0	0	0	0	0	0	0	0	0	0	0;
	4534	30.04	7.71	30	-10	1.011489	100	1	100	20.03	0	0	0	0	0	0	0	0	0	0	0;
	4535	30.04	7.71	30	-10	1.011489	100	1	100	20.03	0	0	0	0	0	0	0	0	0	0	0;
	4537	30.04	7.71	30	-10	1.011489	100	1	100	20.03	0	0	0	0	0	0	0	0	0	0	0;
	4539	31.35	11.33	30	-10	1.043357	100	1	100	20.9	0	0	0	0	0	0	0	0	0	0	0;
	4540	22.02	10.78	30	-10	1.026425	100	1	100	14.68	0	0	0	0	0	0	0	0	0	0	0;
	4547	33.09	8.73	30	-10	1.040215	100	1	100	22.06	0	0	0	0	0	0	0	0	0	0	0;
	4548	33.09	8.73	30	-10	1.040215	100	1	100	22.06	0	0	0	0	0	0	0	0	0	0	0;
	4549	33.09	8.73	30	-10	1.040215	100	1	100	22.06	0	0	0	0	0	0	0	0	0	0	0;
	4550	33.09	8.73	30	-10	1.040215	100	1	100	22.06	0	0	0	0	0	0	0	0	0	0	0;
	4551	33.09	8.73	30	-10	1.040215	100	1	100	22.06	0	0	0	0	0	0	0	0	0	0	0;
	4552	33.09	8.73	30	-10	1.040215	100	1	100	22.06	0	0	0	0	0	0	0	0	0	0	0;
	4553	28.95	9.55	30	-10	1.01638	100	1	100	19.3	0	0	0	0	0	0	0	0	0	0	0;
	4554	33.09	8.73	30	-10	1.040215	100	1	100	22.06	0	0	0	0	0	0	0	0	0	0	0;
	4563	28.83	4.62	30	-10	1.019144	100	1	100	19.22	0	0	0	0	0	0	0	0	0	0	0;
	4572	32.46	19.3	30	-10	1.04927	100	1	100	21.64	0	0	0	0	0	0	0	0	0	0	0;
	4573	90.48	33.58	60	-20	1.043665	100	1	200	60.32	0	0	0	0	0	0	0	0	0	0	0;
	4576	82.6	2.11	60	-20	1.02421	100	1	200	55.07	0	0	0	0	0	0	0	0	0	0	0;
	4577	82.6	2.11	60	-20	1.02421	100	1	200	55.07	0	0	0	0	0	0	0	0	0	0	0;
	4578	82.6	2.11	60	-20	1.02421	100	1	200	55.07	0	0	0	0	0	0	0	0	0	0	0;
	4580	82.6	2.11	60	-20	1.02421	100	1	200	55.07	0	0	0	0	0	0	0	0	0	0	0;
	4585	92.4	2.72	60	-20	1.014507	100	1	200	61.6	0	0	0	0	0	0	0	0	0	0	0;
	4590	28.8	6.92	30	-10	1.000808	100	1	100	19.2	0	0	0	0	0	0	0	0	0	0	0;
	4594	31.92	6.07	30	-10	1.027264	100	1	100	21.28	0	0	0	0	0	0	0	0	0	0	0;
	4598	78.37	30.34	60	-20	1.032985	100	1	200	52.25	0	0	0	0	0	0	0	0	0	0	0;
	4602	78.37	30.34	60	-20	1.032985	100	1	200	52.25	0	0	0	0	0	0	0	0	0	0	0;
	4603	78.37	30.34	60	-20	1.032985	100	1	200	52.25	0	0	0	0	0	0	0	0	0	0	0;
	4604	78.37	30.34	60	-20	1.032985	100	1	200	52.25	0	0	0	0	0	0	0	0	0	0	0;
	4607	23.07	4.45	30	-10	1.026025	100	1	100	15.38	0	0	0	0	0	0	0	0	0	0	0;
	4611	0.99	-0.05	10	-10	1.015919	100	1	100	0.66	0	0	0	0	0	0	0	0	0	0	0;
	4612	290.07	141.78	210	-50	1.033649	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4614	22.5	5.53	10	0	1.039055	100	1	100	15	0	0	0	0	0	0	0	0	0	0	0;
	4616	319.18	33.68	210	-50	1.026027	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	4617	39.78	18.08	30	-10	1.041665	100	1	100	26.52	0	0	0	0	0	0	0	0	0	0	0;
	4618	20	3.75	30	-10	1.024278	100	1	100	13.33	0	0	0	0	0	0	0	0	0	0	0;
	4619	20	3.75	30	-10	1.024278	100	1	100	13.33	0	0	0	0	0	0	0	0	0	0	0;
	4622	162.61	110.24	210	-50	1.054046	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	4626	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	4627	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	4630	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	4631	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	4633	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	4634	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	4635	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	4643	10.36	0.68	10	-10	1.012893	100	1	100	6.91	0	0	0	0	0	0	0	0	0	0	0;
	4644	14.76	0.22	10	-10	1.011577	100	1	100	9.84	0	0	0	0	0	0	0	0	0	0	0;
	4648	21.3	1.07	20	-10	1.032267	100	1	100	14.2	0	0	0	0	0	0	0	0	0	0	0;
	4652	83.69	18.55	60	-20	1.037564	100	1	200	55.79	0	0	0	0	0	0	0	0	0	0	0;
	4655	223.02	66.59	160	-40	1.042492	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	4656	223.02	66.59	160	-40	1.042492	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	4657	223.3	11.76	160	-40	1.026117	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	4658	223.3	11.76	160	-40	1.026117	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	4662	47.08	25.26	60	-20	1.055425	100	1	100	31.39	0	0	0	0	0	0	0	0	0	0	0;
	4663	47.08	25.26	60	-20	1.055425	100	1	100	31.39	0	0	0	0	0	0	0	0	0	0	0;
	4664	85.34	16.39	60	-20	1.053529	100	1	200	56.89	0	0	0	0	0	0	0	0	0	0	0;
	4665	85.34	16.39	60	-20	1.053529	100	1	200	56.89	0	0	0	0	0	0	0	0	0	0	0;
	4666	52.81	50.16	60	-20	1.055479	100	1	200	35.21	0	0	0	0	0	0	0	0	0	0	0;
	4668	31.3	-0.89	20	-20	0.998951	100	1	100	20.87	0	0	0	0	0	0	0	0	0	0	0;
	4670	330.71	4.92	210	-50	1.043049	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	4671	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	4675	40.49	8.06	30	-10	1.036966	100	1	100	26.99	0	0	0	0	0	0	0	0	0	0	0;
	4676	67.83	2.97	60	-20	1.059275	100	1	200	45.22	0	0	0	0	0	0	0	0	0	0	0;
	4677	67.83	2.97	60	-20	1.059275	100	1	200	45.22	0	0	0	0	0	0	0	0	0	0	0;
	4678	67.83	2.97	60	-20	1.059275	100	1	200	45.22	0	0	0	0	0	0	0	0	0	0	0;
	4679	86.9	4.11	60	-20	1.052163	100	1	200	57.93	0	0	0	0	0	0	0	0	0	0	0;
	4680	86.9	4.11	60	-20	1.052163	100	1	200	57.93	0	0	0	0	0	0	0	0	0	0	0;
	4683	79.77	4.15	60	-20	1.047117	100	1	200	53.18	0	0	0	0	0	0	0	0	0	0	0;
	4686	50.96	2.69	60	-20	1.028648	100	1	200	33.97	0	0	0	0	0	0	0	0	0	0	0;
	4690	32.57	7.29	30	-10	1.041748	100	1	100	21.71	0	0	0	0	0	0	0	0	0	0	0;
	4691	89.93	16.64	60	-20	1.037993	100	1	200	59.95	0	0	0	0	0	0	0	0	0	0	0;
	4692	219.98	31.44	210	-50	1.026746	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	4694	32.57	7.29	30	-10	1.041748	100	1	100	21.71	0	0	0	0	0	0	0	0	0	0	0;
	4695	253.65	108.56	140	-40	1.038587	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4697	60.81	1.23	60	-20	1.039251	100	1	200	40.54	0	0	0	0	0	0	0	0	0	0	0;
	4698	89.93	16.64	60	-20	1.037993	100	1	200	59.95	0	0	0	0	0	0	0	0	0	0	0;
	4699	89.93	16.64	60	-20	1.037993	100	1	200	59.95	0	0	0	0	0	0	0	0	0	0	0;
	4702	32.57	7.29	30	-10	1.041748	100	1	100	21.71	0	0	0	0	0	0	0	0	0	0	0;
	4703	55.22	8.96	40	-10	1.017828	100	1	200	36.81	0	0	0	0	0	0	0	0	0	0	0;
	4704	45	-3.79	30	-30	0.939142	100	1	100	30	0	0	0	0	0	0	0	0	0	0	0;
	4707	32.57	7.29	30	-10	1.041748	100	1	100	21.71	0	0	0	0	0	0	0	0	0	0	0;
	4709	18.76	6.31	30	-10	1.033648	100	1	100	12.51	0	0	0	0	0	0	0	0	0	0	0;
	4710	350.89	113.2	210	-50	1.0447	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	4715	25.35	8.5	40	-10	1.018631	100	1	100	16.9	0	0	0	0	0	0	0	0	0	0	0;
	4718	287.65	2.21	210	-50	1.045544	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4722	91.94	33.76	60	-20	1.048656	100	1	200	61.29	0	0	0	0	0	0	0	0	0	0	0;
	4723	91.94	33.76	60	-20	1.048656	100	1	200	61.29	0	0	0	0	0	0	0	0	0	0	0;
	4724	91.94	33.76	60	-20	1.048656	100	1	200	61.29	0	0	0	0	0	0	0	0	0	0	0;
	4725	91.94	33.76	60	-20	1.048656	100	1	200	61.29	0	0	0	0	0	0	0	0	0	0	0;
	4726	91.94	33.76	60	-20	1.048656	100	1	200	61.29	0	0	0	0	0	0	0	0	0	0	0;
	4727	91.94	33.76	60	-20	1.048656	100	1	200	61.29	0	0	0	0	0	0	0	0	0	0	0;
	4728	91.94	33.76	60	-20	1.048656	100	1	200	61.29	0	0	0	0	0	0	0	0	0	0	0;
	4729	91.94	33.76	60	-20	1.048656	100	1	200	61.29	0	0	0	0	0	0	0	0	0	0	0;
	4733	26.6	23.03	30	-10	1.043092	100	1	100	17.73	0	0	0	0	0	0	0	0	0	0	0;
	4738	26.6	23.03	30	-10	1.043092	100	1	100	17.73	0	0	0	0	0	0	0	0	0	0	0;
	4745	56.07	3.64	60	-20	1.04834	100	1	200	37.38	0	0	0	0	0	0	0	0	0	0	0;
	4746	1.46	0.07	10	-10	1.026568	100	1	100	0.98	0	0	0	0	0	0	0	0	0	0	0;
	4751	91.88	4.87	60	-20	1.046359	100	1	200	61.25	0	0	0	0	0	0	0	0	0	0	0;
	4752	88.85	23.49	60	-20	1.032344	100	1	200	59.23	0	0	0	0	0	0	0	0	0	0	0;
	4753	91.88	4.87	60	-20	1.046359	100	1	200	61.25	0	0	0	0	0	0	0	0	0	0	0;
	4757	262.9	48.53	160	-40	1.027461	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4763	36.11	3.87	30	-10	1.024544	100	1	100	24.07	0	0	0	0	0	0	0	0	0	0	0;
	4764	42.35	6.03	40	-10	1.030832	100	1	100	28.23	0	0	0	0	0	0	0	0	0	0	0;
	4770	58.52	23.33	40	-10	1.023397	100	1	200	39.01	0	0	0	0	0	0	0	0	0	0	0;
	4771	37.44	17.23	40	-10	1.022819	100	1	100	24.96	0	0	0	0	0	0	0	0	0	0	0;
	4772	58.52	23.33	40	-10	1.023397	100	1	200	39.01	0	0	0	0	0	0	0	0	0	0	0;
	4773	58.52	23.33	40	-10	1.023397	100	1	200	39.01	0	0	0	0	0	0	0	0	0	0	0;
	4774	37.45	15.75	40	-10	1.026206	100	1	100	24.97	0	0	0	0	0	0	0	0	0	0	0;
	4775	50	-2.27	20	-20	0.993383	100	1	100	33.33	0	0	0	0	0	0	0	0	0	0	0;
	4777	34.86	4.81	30	-10	1.050901	100	1	100	23.24	0	0	0	0	0	0	0	0	0	0	0;
	4786	62.5	10.73	20	0	1.000057	100	1	200	41.67	0	0	0	0	0	0	0	0	0	0	0;
	4791	42.5	5.74	40	-10	1.031502	100	1	100	28.33	0	0	0	0	0	0	0	0	0	0	0;
	4792	42.5	5.74	40	-10	1.031502	100	1	100	28.33	0	0	0	0	0	0	0	0	0	0	0;
	4793	32.51	6.89	30	-10	1.011478	100	1	100	21.67	0	0	0	0	0	0	0	0	0	0	0;
	4796	15	5.62	10	0	1.034961	100	1	100	10	0	0	0	0	0	0	0	0	0	0	0;
	4809	97.12	8.38	60	-20	1.040009	100	1	200	64.75	0	0	0	0	0	0	0	0	0	0	0;
	4810	97.12	8.38	60	-20	1.040009	100	1	200	64.75	0	0	0	0	0	0	0	0	0	0	0;
	4811	97.12	8.38	60	-20	1.040009	100	1	200	64.75	0	0	0	0	0	0	0	0	0	0	0;
	4814	90.17	24.56	60	-20	1.06019	100	1	200	60.11	0	0	0	0	0	0	0	0	0	0	0;
	4823	22	10.36	20	0	1.014332	100	1	100	14.67	0	0	0	0	0	0	0	0	0	0	0;
	4837	65.21	12.85	60	-20	1.026243	100	1	200	43.47	0	0	0	0	0	0	0	0	0	0	0;
	4843	65.21	12.85	60	-20	1.026243	100	1	200	43.47	0	0	0	0	0	0	0	0	0	0	0;
	4844	65.21	12.85	60	-20	1.026243	100	1	200	43.47	0	0	0	0	0	0	0	0	0	0	0;
	4845	65.21	12.85	60	-20	1.026243	100	1	200	43.47	0	0	0	0	0	0	0	0	0	0	0;
	4846	65.21	12.85	60	-20	1.026243	100	1	200	43.47	0	0	0	0	0	0	0	0	0	0	0;
	4847	65.21	12.85	60	-20	1.026243	100	1	200	43.47	0	0	0	0	0	0	0	0	0	0	0;
	4855	288.69	133.24	160	-40	1.060956	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4856	288.69	133.24	160	-40	1.060956	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4857	256.4	117.4	160	-40	1.062536	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4858	256.4	117.4	160	-40	1.062536	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4859	222.47	102.94	160	-40	1.059942	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	4861	279.23	123.34	160	-40	1.05948	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4862	279.23	123.34	160	-40	1.05948	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4864	222.47	102.94	160	-40	1.059942	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	4870	18.74	3.61	30	-10	1.025271	100	1	100	12.49	0	0	0	0	0	0	0	0	0	0	0;
	4887	331.89	-5.79	210	-50	1.01898	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	4888	331.89	-5.79	210	-50	1.01898	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	4892	280.8	38.38	160	-40	1.024711	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4893	280.8	38.38	160	-40	1.024711	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4896	280.8	38.38	160	-40	1.024711	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4897	280.8	38.38	160	-40	1.024711	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4902	22.8	8.85	30	-10	1.045936	100	1	100	15.2	0	0	0	0	0	0	0	0	0	0	0;
	4903	22.8	8.85	30	-10	1.045936	100	1	100	15.2	0	0	0	0	0	0	0	0	0	0	0;
	4904	27.65	6.22	30	-10	1.058505	100	1	100	18.43	0	0	0	0	0	0	0	0	0	0	0;
	4905	27.65	6.22	30	-10	1.058505	100	1	100	18.43	0	0	0	0	0	0	0	0	0	0	0;
	4910	35.74	6.84	30	-10	1.040138	100	1	100	23.83	0	0	0	0	0	0	0	0	0	0	0;
	4917	21	0.7	10	-10	0.989071	100	1	100	14	0	0	0	0	0	0	0	0	0	0	0;
	4918	39.1	-2.22	20	-20	0.943436	100	1	100	26.07	0	0	0	0	0	0	0	0	0	0	0;
	4922	47.29	6.72	40	-10	1.04826	100	1	100	31.53	0	0	0	0	0	0	0	0	0	0	0;
	4924	56.84	10.43	40	-10	1.025724	100	1	200	37.89	0	0	0	0	0	0	0	0	0	0	0;
	4925	35.69	7.89	30	-10	1.017601	100	1	100	23.79	0	0	0	0	0	0	0	0	0	0	0;
	4926	19.96	23.09	30	-10	1.036433	100	1	100	13.31	0	0	0	0	0	0	0	0	0	0	0;
	4928	277.26	31.82	210	-50	1.04876	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	4931	36.7	9.55	30	-10	1.037501	100	1	100	24.47	0	0	0	0	0	0	0	0	0	0	0;
	4932	36.7	9.55	30	-10	1.037501	100	1	100	24.47	0	0	0	0	0	0	0	0	0	0	0;
	4934	36.91	13.33	30	-10	1.037973	100	1	100	24.61	0	0	0	0	0	0	0	0	0	0	0;
	4938	37.67	5.32	20	-10	1.027032	100	1	100	25.12	0	0	0	0	0	0	0	0	0	0	0;
	4939	38.53	9.19	30	-10	1.03611	100	1	100	25.69	0	0	0	0	0	0	0	0	0	0	0;
	4942	23.63	8.01	30	-10	1.038474	100	1	100	15.75	0	0	0	0	0	0	0	0	0	0	0;
	4943	23.63	8.01	30	-10	1.038474	100	1	100	15.75	0	0	0	0	0	0	0	0	0	0	0;
	4945	24.77	15.08	30	-10	1.042258	100	1	100	16.51	0	0	0	0	0	0	0	0	0	0	0;
	4946	24.77	15.08	30	-10	1.042258	100	1	100	16.51	0	0	0	0	0	0	0	0	0	0	0;
	4949	48.42	9.78	40	-10	1.01601	100	1	100	32.28	0	0	0	0	0	0	0	0	0	0	0;
	4952	249.55	64.24	210	-50	1.03281	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	4958	50.82	16.11	40	-10	1.025082	100	1	200	33.88	0	0	0	0	0	0	0	0	0	0	0;
	4959	677.41	27.58	520	-120	1.024274	100	1	1400	400	0	0	0	0	0	0	0	0	0	0	0;
	4963	57.59	10.86	40	-10	1.019736	100	1	200	38.39	0	0	0	0	0	0	0	0	0	0	0;
	4964	57.59	10.86	40	-10	1.019736	100	1	200	38.39	0	0	0	0	0	0	0	0	0	0	0;
	4967	16.67	4.68	30	-10	1.013686	100	1	100	11.11	0	0	0	0	0	0	0	0	0	0	0;
	4982	98.46	12.34	60	-20	1.01458	100	1	200	65.64	0	0	0	0	0	0	0	0	0	0	0;
	4983	7	0.65	10	-10	0.995619	100	1	100	4.67	0	0	0	0	0	0	0	0	0	0	0;
	4985	37.96	13.4	30	-10	1.073911	100	1	100	25.31	0	0	0	0	0	0	0	0	0	0	0;
	4987	38.04	7.18	30	-10	1.020265	100	1	100	25.36	0	0	0	0	0	0	0	0	0	0	0;
	4988	24.84	4.52	30	-10	1.022099	100	1	100	16.56	0	0	0	0	0	0	0	0	0	0	0;
	4998	72.47	11.44	60	-20	1.019231	100	1	200	48.31	0	0	0	0	0	0	0	0	0	0	0;
	5000	2.94	0.27	10	-10	1.008736	100	1	100	1.96	0	0	0	0	0	0	0	0	0	0	0;
	5007	34.37	6.67	30	-10	1.035406	100	1	100	22.91	0	0	0	0	0	0	0	0	0	0	0;
	5008	34.37	6.67	30	-10	1.035406	100	1	100	22.91	0	0	0	0	0	0	0	0	0	0	0;
	5009	34.37	6.67	30	-10	1.035406	100	1	100	22.91	0	0	0	0	0	0	0	0	0	0	0;
	5012	74.36	21.23	60	-20	1.032132	100	1	200	49.57	0	0	0	0	0	0	0	0	0	0	0;
	5013	74.36	21.23	60	-20	1.032132	100	1	200	49.57	0	0	0	0	0	0	0	0	0	0	0;
	5017	69.35	21.13	60	-20	1.032238	100	1	200	46.23	0	0	0	0	0	0	0	0	0	0	0;
	5018	69.35	21.13	60	-20	1.032238	100	1	200	46.23	0	0	0	0	0	0	0	0	0	0	0;
	5023	229.56	74.98	210	-50	1.052394	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5025	21.69	5.19	30	-10	1.050987	100	1	100	14.46	0	0	0	0	0	0	0	0	0	0	0;
	5029	31.79	7.29	30	-10	1.005112	100	1	100	21.19	0	0	0	0	0	0	0	0	0	0	0;
	5034	256.62	49.7	210	-50	1.019932	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	5036	26.41	7.43	30	-10	1.038011	100	1	100	17.61	0	0	0	0	0	0	0	0	0	0	0;
	5037	16.71	0.01	30	-10	1.030988	100	1	100	11.14	0	0	0	0	0	0	0	0	0	0	0;
	5038	28.97	5.47	30	-10	1.043059	100	1	100	19.31	0	0	0	0	0	0	0	0	0	0	0;
	5050	17	0.3	10	-10	0.987146	100	1	100	11.33	0	0	0	0	0	0	0	0	0	0	0;
	5051	37.85	8.46	30	-10	1.035412	100	1	100	25.23	0	0	0	0	0	0	0	0	0	0	0;
	5052	37.85	8.46	30	-10	1.035412	100	1	100	25.23	0	0	0	0	0	0	0	0	0	0	0;
	5053	37.85	8.46	30	-10	1.035412	100	1	100	25.23	0	0	0	0	0	0	0	0	0	0	0;
	5054	17.6	22.53	30	-10	1.05667	100	1	100	11.73	0	0	0	0	0	0	0	0	0	0	0;
	5055	17.6	22.53	30	-10	1.05667	100	1	100	11.73	0	0	0	0	0	0	0	0	0	0	0;
	5056	17.6	22.53	30	-10	1.05667	100	1	100	11.73	0	0	0	0	0	0	0	0	0	0	0;
	5057	17.6	22.53	30	-10	1.05667	100	1	100	11.73	0	0	0	0	0	0	0	0	0	0	0;
	5058	349.51	198.35	210	-50	1.049254	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	5059	210.14	188.39	210	-50	1.083018	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5060	210.14	188.39	210	-50	1.083018	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5061	10.5	-0.35	10	-10	0.994858	100	1	100	7	0	0	0	0	0	0	0	0	0	0	0;
	5062	279.8	200.76	210	-50	1.056742	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	5063	279.8	200.76	210	-50	1.056742	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	5064	279.8	200.76	210	-50	1.056742	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	5065	349.51	198.35	210	-50	1.049254	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	5069	18.8	7.03	30	-10	0.995797	100	1	100	12.53	0	0	0	0	0	0	0	0	0	0	0;
	5070	15.94	-1.51	10	-10	0.999202	100	1	100	10.63	0	0	0	0	0	0	0	0	0	0	0;
	5071	18.8	7.03	30	-10	0.995797	100	1	100	12.53	0	0	0	0	0	0	0	0	0	0	0;
	5072	18.8	7.03	30	-10	0.995797	100	1	100	12.53	0	0	0	0	0	0	0	0	0	0	0;
	5073	18.8	7.03	30	-10	0.995797	100	1	100	12.53	0	0	0	0	0	0	0	0	0	0	0;
	5074	18.8	7.03	30	-10	0.995797	100	1	100	12.53	0	0	0	0	0	0	0	0	0	0	0;
	5076	19.26	2.33	30	-10	1.049756	100	1	100	12.84	0	0	0	0	0	0	0	0	0	0	0;
	5077	19.26	2.33	30	-10	1.049756	100	1	100	12.84	0	0	0	0	0	0	0	0	0	0	0;
	5078	19.26	2.33	30	-10	1.049756	100	1	100	12.84	0	0	0	0	0	0	0	0	0	0	0;
	5082	85.48	16.54	60	-20	1.029323	100	1	200	56.99	0	0	0	0	0	0	0	0	0	0	0;
	5084	227.2	71.82	210	-50	1.06339	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5085	83.48	36.7	60	-20	1.041566	100	1	200	55.65	0	0	0	0	0	0	0	0	0	0	0;
	5086	83.48	36.7	60	-20	1.041566	100	1	200	55.65	0	0	0	0	0	0	0	0	0	0	0;
	5089	83.47	36.43	60	-20	1.039883	100	1	200	55.65	0	0	0	0	0	0	0	0	0	0	0;
	5090	83.47	36.43	60	-20	1.039883	100	1	200	55.65	0	0	0	0	0	0	0	0	0	0	0;
	5091	83.48	36.7	60	-20	1.041566	100	1	200	55.65	0	0	0	0	0	0	0	0	0	0	0;
	5092	83.48	36.7	60	-20	1.041566	100	1	200	55.65	0	0	0	0	0	0	0	0	0	0	0;
	5099	29.83	8.08	30	-10	1.015374	100	1	100	19.89	0	0	0	0	0	0	0	0	0	0	0;
	5109	21.17	9.91	30	-10	1.048748	100	1	100	14.11	0	0	0	0	0	0	0	0	0	0	0;
	5111	6.52	0.05	10	-10	0.996117	100	1	100	4.35	0	0	0	0	0	0	0	0	0	0	0;
	5116	32.34	-0.25	30	-20	1.013887	100	1	100	21.56	0	0	0	0	0	0	0	0	0	0	0;
	5122	19.12	15.11	30	-10	1.030744	100	1	100	12.75	0	0	0	0	0	0	0	0	0	0	0;
	5123	33.69	12.78	30	-10	1.043216	100	1	100	22.46	0	0	0	0	0	0	0	0	0	0	0;
	5126	38.57	10.56	30	-10	1.045141	100	1	100	25.71	0	0	0	0	0	0	0	0	0	0	0;
	5127	38.57	10.56	30	-10	1.045141	100	1	100	25.71	0	0	0	0	0	0	0	0	0	0	0;
	5128	19.12	15.11	30	-10	1.030744	100	1	100	12.75	0	0	0	0	0	0	0	0	0	0	0;
	5129	19.12	15.11	30	-10	1.030744	100	1	100	12.75	0	0	0	0	0	0	0	0	0	0	0;
	5130	19.12	15.11	30	-10	1.030744	100	1	100	12.75	0	0	0	0	0	0	0	0	0	0	0;
	5131	19.12	15.11	30	-10	1.030744	100	1	100	12.75	0	0	0	0	0	0	0	0	0	0	0;
	5133	19.12	15.11	30	-10	1.030744	100	1	100	12.75	0	0	0	0	0	0	0	0	0	0	0;
	5134	30.93	8.23	30	-10	1.035888	100	1	100	20.62	0	0	0	0	0	0	0	0	0	0	0;
	5135	319.18	33.68	210	-50	1.026027	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	5138	26.75	6.78	40	-10	1.024572	100	1	100	17.83	0	0	0	0	0	0	0	0	0	0	0;
	5143	96.9	14.27	60	-20	1.033735	100	1	200	64.6	0	0	0	0	0	0	0	0	0	0	0;
	5144	35.99	7.92	30	-10	1.017192	100	1	100	23.99	0	0	0	0	0	0	0	0	0	0	0;
	5145	18.68	3.68	30	-10	1.01785	100	1	100	12.45	0	0	0	0	0	0	0	0	0	0	0;
	5146	18.68	3.68	30	-10	1.01785	100	1	100	12.45	0	0	0	0	0	0	0	0	0	0	0;
	5147	18.68	3.68	30	-10	1.01785	100	1	100	12.45	0	0	0	0	0	0	0	0	0	0	0;
	5148	30.26	9.8	30	-10	1.024966	100	1	100	20.17	0	0	0	0	0	0	0	0	0	0	0;
	5151	18.67	5.94	30	-10	1.015505	100	1	100	12.45	0	0	0	0	0	0	0	0	0	0	0;
	5155	147.93	31.02	160	-40	1.044772	100	1	300	98.62	0	0	0	0	0	0	0	0	0	0	0;
	5158	141.75	-12.6	80	-80	0.98522	100	1	300	94.5	0	0	0	0	0	0	0	0	0	0	0;
	5159	96.9	14.27	60	-20	1.033735	100	1	200	64.6	0	0	0	0	0	0	0	0	0	0	0;
	5162	10	2.6	10	-10	1.008964	100	1	100	6.67	0	0	0	0	0	0	0	0	0	0	0;
	5163	96.9	14.27	60	-20	1.033735	100	1	200	64.6	0	0	0	0	0	0	0	0	0	0	0;
	5166	44.9	-0.09	60	-20	1.017024	100	1	100	29.93	0	0	0	0	0	0	0	0	0	0	0;
	5169	19.9	1.8	20	-10	1.040788	100	1	100	13.27	0	0	0	0	0	0	0	0	0	0	0;
	5177	25.46	7.25	30	-10	1.023095	100	1	100	16.97	0	0	0	0	0	0	0	0	0	0	0;
	5178	25.46	7.25	30	-10	1.023095	100	1	100	16.97	0	0	0	0	0	0	0	0	0	0	0;
	5187	37.08	5.51	30	-10	1.021005	100	1	100	24.72	0	0	0	0	0	0	0	0	0	0	0;
	5188	40.49	8.06	30	-10	1.036966	100	1	100	26.99	0	0	0	0	0	0	0	0	0	0	0;
	5192	180.86	44.37	160	-40	1.044001	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	5196	32.36	7.22	30	-10	1.028002	100	1	100	21.57	0	0	0	0	0	0	0	0	0	0	0;
	5202	30.6	-1.46	20	-20	0.99834	100	1	100	20.4	0	0	0	0	0	0	0	0	0	0	0;
	5207	332.21	116.49	210	-50	1.047444	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	5208	332.21	116.49	210	-50	1.047444	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	5210	0.51	0.04	10	-10	1.029935	100	1	100	0.34	0	0	0	0	0	0	0	0	0	0	0;
	5212	67.5	7.59	40	-10	1.021725	100	1	200	45	0	0	0	0	0	0	0	0	0	0	0;
	5223	359.06	76.72	210	-50	1.054235	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	5225	50.33	11.56	40	-10	1.037142	100	1	200	33.55	0	0	0	0	0	0	0	0	0	0	0;
	5226	50.33	11.56	40	-10	1.037142	100	1	200	33.55	0	0	0	0	0	0	0	0	0	0	0;
	5228	50.33	11.56	40	-10	1.037142	100	1	200	33.55	0	0	0	0	0	0	0	0	0	0	0;
	5229	50.33	11.56	40	-10	1.037142	100	1	200	33.55	0	0	0	0	0	0	0	0	0	0	0;
	5242	72.17	16.24	60	-20	1.022616	100	1	200	48.11	0	0	0	0	0	0	0	0	0	0	0;
	5243	56.17	4.09	60	-20	1.003094	100	1	200	37.45	0	0	0	0	0	0	0	0	0	0	0;
	5245	46.8	8.97	60	-20	0.993793	100	1	100	31.2	0	0	0	0	0	0	0	0	0	0	0;
	5246	72.17	16.24	60	-20	1.022616	100	1	200	48.11	0	0	0	0	0	0	0	0	0	0	0;
	5252	48.81	10.42	40	-10	1.017273	100	1	100	32.54	0	0	0	0	0	0	0	0	0	0	0;
	5254	0.26	-0.09	10	-10	1.024094	100	1	100	0.17	0	0	0	0	0	0	0	0	0	0	0;
	5260	286.96	43.74	160	-40	1.028088	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	5261	48.81	10.42	40	-10	1.017273	100	1	100	32.54	0	0	0	0	0	0	0	0	0	0	0;
	5262	48.81	10.42	40	-10	1.017273	100	1	100	32.54	0	0	0	0	0	0	0	0	0	0	0;
	5270	57.51	17.72	40	-10	1.029323	100	1	200	38.34	0	0	0	0	0	0	0	0	0	0	0;
	5271	57.51	17.72	40	-10	1.029323	100	1	200	38.34	0	0	0	0	0	0	0	0	0	0	0;
	5274	55.04	14.65	40	-10	1.023803	100	1	200	36.69	0	0	0	0	0	0	0	0	0	0	0;
	5275	55.04	14.65	40	-10	1.023803	100	1	200	36.69	0	0	0	0	0	0	0	0	0	0	0;
	5276	55.04	14.65	40	-10	1.023803	100	1	200	36.69	0	0	0	0	0	0	0	0	0	0	0;
	5277	55.04	14.65	40	-10	1.023803	100	1	200	36.69	0	0	0	0	0	0	0	0	0	0	0;
	5278	57.51	17.72	40	-10	1.029323	100	1	200	38.34	0	0	0	0	0	0	0	0	0	0	0;
	5279	57.51	17.72	40	-10	1.029323	100	1	200	38.34	0	0	0	0	0	0	0	0	0	0	0;
	5281	155	-15.46	90	-90	0.98306	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	5286	70.12	2.87	60	-20	1.023526	100	1	200	46.75	0	0	0	0	0	0	0	0	0	0	0;
	5287	70.12	2.87	60	-20	1.023526	100	1	200	46.75	0	0	0	0	0	0	0	0	0	0	0;
	5289	8.2	-0.77	10	-10	1.00746	100	1	100	5.47	0	0	0	0	0	0	0	0	0	0	0;
	5292	34.86	4.81	30	-10	1.050901	100	1	100	23.24	0	0	0	0	0	0	0	0	0	0	0;
	5301	50.71	5.07	40	-10	1.021316	100	1	200	33.81	0	0	0	0	0	0	0	0	0	0	0;
	5302	52.8	0.59	40	-10	1.017834	100	1	200	35.2	0	0	0	0	0	0	0	0	0	0	0;
	5303	43.28	11.6	40	-10	1.024054	100	1	100	28.85	0	0	0	0	0	0	0	0	0	0	0;
	5304	43.28	11.6	40	-10	1.024054	100	1	100	28.85	0	0	0	0	0	0	0	0	0	0	0;
	5305	39.88	8.02	40	-10	1.03032	100	1	100	26.59	0	0	0	0	0	0	0	0	0	0	0;
	5306	39.88	8.02	40	-10	1.03032	100	1	100	26.59	0	0	0	0	0	0	0	0	0	0	0;
	5307	39.88	7.99	40	-10	1.030262	100	1	100	26.59	0	0	0	0	0	0	0	0	0	0	0;
	5308	39.88	7.99	40	-10	1.030262	100	1	100	26.59	0	0	0	0	0	0	0	0	0	0	0;
	5309	58.06	14.52	40	-10	1.033768	100	1	200	38.71	0	0	0	0	0	0	0	0	0	0	0;
	5310	58.06	14.52	40	-10	1.033768	100	1	200	38.71	0	0	0	0	0	0	0	0	0	0	0;
	5311	233.91	82.52	210	-50	1.057062	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5314	93.45	15.84	90	-60	0.984098	100	1	200	62.3	0	0	0	0	0	0	0	0	0	0	0;
	5316	3.19	-0.17	10	-10	1.002922	100	1	100	2.12	0	0	0	0	0	0	0	0	0	0	0;
	5317	77.25	12.22	60	-20	1.034048	100	1	200	51.5	0	0	0	0	0	0	0	0	0	0	0;
	5319	377.38	119.4	210	-50	1.071081	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	5322	24.81	-3.88	30	-10	0.99879	100	1	100	16.54	0	0	0	0	0	0	0	0	0	0	0;
	5323	87.75	8.05	60	-20	1.011203	100	1	200	58.5	0	0	0	0	0	0	0	0	0	0	0;
	5324	87.75	8.05	60	-20	1.011203	100	1	200	58.5	0	0	0	0	0	0	0	0	0	0	0;
	5325	87.75	8.05	60	-20	1.011203	100	1	200	58.5	0	0	0	0	0	0	0	0	0	0	0;
	5326	87.75	8.05	60	-20	1.011203	100	1	200	58.5	0	0	0	0	0	0	0	0	0	0	0;
	5327	87.75	8.05	60	-20	1.011203	100	1	200	58.5	0	0	0	0	0	0	0	0	0	0	0;
	5328	87.75	8.05	60	-20	1.011203	100	1	200	58.5	0	0	0	0	0	0	0	0	0	0	0;
	5330	27.96	5.11	30	-10	1.029009	100	1	100	18.64	0	0	0	0	0	0	0	0	0	0	0;
	5332	32.36	7.22	30	-10	1.028002	100	1	100	21.57	0	0	0	0	0	0	0	0	0	0	0;
	5333	1.12	0.09	10	-10	1.03063	100	1	100	0.75	0	0	0	0	0	0	0	0	0	0	0;
	5335	87.35	17.85	60	-20	1.033768	100	1	200	58.23	0	0	0	0	0	0	0	0	0	0	0;
	5341	89.91	48.94	60	-20	1.065506	100	1	200	59.94	0	0	0	0	0	0	0	0	0	0	0;
	5342	57.59	10.86	40	-10	1.019736	100	1	200	38.39	0	0	0	0	0	0	0	0	0	0	0;
	5343	158.9	55.92	100	-30	1.000957	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	5347	90.48	33.58	60	-20	1.043665	100	1	200	60.32	0	0	0	0	0	0	0	0	0	0	0;
	5350	19.27	11.16	30	-10	1.021475	100	1	100	12.85	0	0	0	0	0	0	0	0	0	0	0;
	5352	30.32	6.61	30	-10	1.007679	100	1	100	20.21	0	0	0	0	0	0	0	0	0	0	0;
	5354	224.68	-13.42	210	-50	1.039621	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5356	89.73	17.85	60	-20	1.041223	100	1	200	59.82	0	0	0	0	0	0	0	0	0	0	0;
	5359	14.52	2.12	10	-10	1.054892	100	1	100	9.68	0	0	0	0	0	0	0	0	0	0	0;
	5365	38.53	9.19	30	-10	1.03611	100	1	100	25.69	0	0	0	0	0	0	0	0	0	0	0;
	5366	38.53	9.19	30	-10	1.03611	100	1	100	25.69	0	0	0	0	0	0	0	0	0	0	0;
	5368	83.89	24.93	60	-20	1.023161	100	1	200	55.93	0	0	0	0	0	0	0	0	0	0	0;
	5369	83.89	24.93	60	-20	1.023161	100	1	200	55.93	0	0	0	0	0	0	0	0	0	0	0;
	5370	83.89	24.93	60	-20	1.023161	100	1	200	55.93	0	0	0	0	0	0	0	0	0	0	0;
	5371	83.89	24.93	60	-20	1.023161	100	1	200	55.93	0	0	0	0	0	0	0	0	0	0	0;
	5382	57.59	10.86	40	-10	1.019736	100	1	200	38.39	0	0	0	0	0	0	0	0	0	0	0;
	5384	6.38	-0.43	10	-10	0.962211	100	1	100	4.25	0	0	0	0	0	0	0	0	0	0	0;
	5385	30.28	3.51	30	-10	1.018125	100	1	100	20.19	0	0	0	0	0	0	0	0	0	0	0;
	5388	39.03	10.16	30	-10	1.033359	100	1	100	26.02	0	0	0	0	0	0	0	0	0	0	0;
	5391	34.92	9.74	30	-10	1.043492	100	1	100	23.28	0	0	0	0	0	0	0	0	0	0	0;
	5397	29.4	8.8	40	-30	1.01809	100	1	100	19.6	0	0	0	0	0	0	0	0	0	0	0;
	5401	38.85	7.6	30	-10	1.028342	100	1	100	25.9	0	0	0	0	0	0	0	0	0	0	0;
	5403	-128.54	-23.93	40	-50	1.018098	100	1	100	-171.38	0	0	0	0	0	0	0	0	0	0	0;
	5406	41.92	5.34	40	-10	1.007636	100	1	100	27.95	0	0	0	0	0	0	0	0	0	0	0;
	5408	6.52	0.05	10	-10	0.996117	100	1	100	4.35	0	0	0	0	0	0	0	0	0	0	0;
	5411	34.83	7.05	40	-10	0.991559	100	1	100	23.22	0	0	0	0	0	0	0	0	0	0	0;
	5416	33.15	0.93	30	-10	1.024669	100	1	100	22.1	0	0	0	0	0	0	0	0	0	0	0;
	5418	8.33	2	10	-10	0.978998	100	1	100	5.55	0	0	0	0	0	0	0	0	0	0	0;
	5420	34.31	1.99	30	-10	1.015531	100	1	100	22.87	0	0	0	0	0	0	0	0	0	0	0;
	5439	36.91	13.33	30	-10	1.037973	100	1	100	24.61	0	0	0	0	0	0	0	0	0	0	0;
	5444	97.68	6.22	60	-20	1.017679	100	1	200	65.12	0	0	0	0	0	0	0	0	0	0	0;
	5455	112	11.12	60	-20	1.018325	100	1	300	74.67	0	0	0	0	0	0	0	0	0	0	0;
	5456	64.65	13.16	60	-20	1.009683	100	1	200	43.1	0	0	0	0	0	0	0	0	0	0	0;
	5463	38.42	21.93	30	-10	1.054031	100	1	100	25.61	0	0	0	0	0	0	0	0	0	0	0;
	5464	84	2.79	60	-50	1.011467	100	1	200	56	0	0	0	0	0	0	0	0	0	0	0;
	5466	39.44	10.06	40	-10	1.021987	100	1	100	26.29	0	0	0	0	0	0	0	0	0	0	0;
	5467	39.44	10.06	40	-10	1.021987	100	1	100	26.29	0	0	0	0	0	0	0	0	0	0	0;
	5468	39.44	10.06	40	-10	1.021987	100	1	100	26.29	0	0	0	0	0	0	0	0	0	0	0;
	5473	77.25	12.22	60	-20	1.034048	100	1	200	51.5	0	0	0	0	0	0	0	0	0	0	0;
	5476	42.41	17.68	60	-20	1.058287	100	1	100	28.27	0	0	0	0	0	0	0	0	0	0	0;
	5484	40	4.52	20	-20	0.956073	100	1	100	26.67	0	0	0	0	0	0	0	0	0	0	0;
	5490	21.35	3.86	20	-10	1.028982	100	1	100	14.23	0	0	0	0	0	0	0	0	0	0	0;
	5494	25.6	-1.25	20	-20	0.991002	100	1	100	17.07	0	0	0	0	0	0	0	0	0	0	0;
	5498	32.94	16.6	30	-10	1.041198	100	1	100	21.96	0	0	0	0	0	0	0	0	0	0	0;
	5503	174.4	15	210	-50	1.032124	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	5510	89.26	20.01	60	-20	1.05632	100	1	200	59.51	0	0	0	0	0	0	0	0	0	0	0;
	5512	27.76	14.88	30	-10	1.014266	100	1	100	18.51	0	0	0	0	0	0	0	0	0	0	0;
	5514	89.26	20.01	60	-20	1.05632	100	1	200	59.51	0	0	0	0	0	0	0	0	0	0	0;
	5517	219.5	66.99	210	-50	1.03284	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5523	89.93	16.64	60	-20	1.037993	100	1	200	59.95	0	0	0	0	0	0	0	0	0	0	0;
	5528	38.95	5.12	30	-10	1.009486	100	1	100	25.97	0	0	0	0	0	0	0	0	0	0	0;
	5530	89.93	16.64	60	-20	1.037993	100	1	200	59.95	0	0	0	0	0	0	0	0	0	0	0;
	5534	228.19	49.03	210	-50	1.020332	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5540	31.59	4.82	30	-10	1.045359	100	1	100	21.06	0	0	0	0	0	0	0	0	0	0	0;
	5542	40	-1.34	30	-30	1.010203	100	1	100	26.67	0	0	0	0	0	0	0	0	0	0	0;
	5546	92.69	14.48	60	-20	1.026207	100	1	200	61.79	0	0	0	0	0	0	0	0	0	0	0;
	5548	16.71	0.01	30	-10	1.030988	100	1	100	11.14	0	0	0	0	0	0	0	0	0	0	0;
	5575	26.19	8.11	30	-10	1.031736	100	1	100	17.46	0	0	0	0	0	0	0	0	0	0	0;
	5576	168.75	16.36	140	-130	0.976475	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	5581	381.8	21.67	210	-50	1.063714	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	5587	417.43	19.88	260	-60	1.041875	100	1	900	200	0	0	0	0	0	0	0	0	0	0	0;
	5592	64.28	29.97	60	-20	1.00876	100	1	200	42.85	0	0	0	0	0	0	0	0	0	0	0;
	5593	22.87	9.78	30	-10	1.014635	100	1	100	15.25	0	0	0	0	0	0	0	0	0	0	0;
	5594	32.36	7.22	30	-10	1.028002	100	1	100	21.57	0	0	0	0	0	0	0	0	0	0	0;
	5607	34.32	1.14	20	-10	1.025746	100	1	100	22.88	0	0	0	0	0	0	0	0	0	0	0;
	5608	22.77	10.64	30	-10	1.045989	100	1	100	15.18	0	0	0	0	0	0	0	0	0	0	0;
	5610	34.32	1.14	20	-10	1.025746	100	1	100	22.88	0	0	0	0	0	0	0	0	0	0	0;
	5611	25.33	9.63	30	-10	1.047982	100	1	100	16.89	0	0	0	0	0	0	0	0	0	0	0;
	5612	25.33	9.63	30	-10	1.047982	100	1	100	16.89	0	0	0	0	0	0	0	0	0	0	0;
	5613	22.77	10.64	30	-10	1.045989	100	1	100	15.18	0	0	0	0	0	0	0	0	0	0	0;
	5614	31.68	9.78	30	-10	1.047806	100	1	100	21.12	0	0	0	0	0	0	0	0	0	0	0;
	5619	21.53	3.36	30	-10	1.049918	100	1	100	14.35	0	0	0	0	0	0	0	0	0	0	0;
	5622	34.27	11.44	30	-10	1.03975	100	1	100	22.85	0	0	0	0	0	0	0	0	0	0	0;
	5627	219.98	31.44	210	-50	1.026746	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5628	23.9	13.28	30	-10	1.020137	100	1	100	15.93	0	0	0	0	0	0	0	0	0	0	0;
	5629	23.9	13.28	30	-10	1.020137	100	1	100	15.93	0	0	0	0	0	0	0	0	0	0	0;
	5630	10.76	-0.13	10	-10	0.976217	100	1	100	7.17	0	0	0	0	0	0	0	0	0	0	0;
	5631	30.93	5.5	30	-10	1.009821	100	1	100	20.62	0	0	0	0	0	0	0	0	0	0	0;
	5632	16.28	6.64	30	-10	1.015993	100	1	100	10.85	0	0	0	0	0	0	0	0	0	0	0;
	5634	30.93	5.5	30	-10	1.009821	100	1	100	20.62	0	0	0	0	0	0	0	0	0	0	0;
	5637	214.01	124.9	160	-40	1.043748	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5640	227.2	71.82	210	-50	1.06339	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5641	38.02	11.21	30	-10	1.043087	100	1	100	25.35	0	0	0	0	0	0	0	0	0	0	0;
	5645	53.13	-3.75	40	-40	1.003179	100	1	200	35.42	0	0	0	0	0	0	0	0	0	0	0;
	5653	24.86	7.42	30	-10	1.01212	100	1	100	16.57	0	0	0	0	0	0	0	0	0	0	0;
	5654	227.2	72.57	210	-50	1.063453	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5657	279.13	130.91	210	-50	1.048637	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	5658	0.99	-0.05	10	-10	1.015919	100	1	100	0.66	0	0	0	0	0	0	0	0	0	0	0;
	5660	29.35	12.59	30	-10	1.05544	100	1	100	19.57	0	0	0	0	0	0	0	0	0	0	0;
	5668	32.52	9.41	30	-10	1.028707	100	1	100	21.68	0	0	0	0	0	0	0	0	0	0	0;
	5670	32.52	9.41	30	-10	1.028707	100	1	100	21.68	0	0	0	0	0	0	0	0	0	0	0;
	5672	36.8	6.62	30	-10	1.03058	100	1	100	24.53	0	0	0	0	0	0	0	0	0	0	0;
	5674	382.77	9.72	210	-50	1.043874	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	5676	50.82	16.11	40	-10	1.025082	100	1	200	33.88	0	0	0	0	0	0	0	0	0	0	0;
	5679	4.8	-0.12	10	-10	0.996318	100	1	100	3.2	0	0	0	0	0	0	0	0	0	0	0;
	5681	62.5	10.73	20	0	1.000057	100	1	200	41.67	0	0	0	0	0	0	0	0	0	0	0;
	5695	62.61	24.83	60	-20	1.034646	100	1	200	41.74	0	0	0	0	0	0	0	0	0	0	0;
	5697	8.06	0.42	10	-10	1.002466	100	1	100	5.37	0	0	0	0	0	0	0	0	0	0	0;
	5702	6.32	-0.32	10	-10	1.027418	100	1	100	4.21	0	0	0	0	0	0	0	0	0	0	0;
	5704	359.93	62.09	210	-50	1.053462	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	5705	359.94	62.42	210	-50	1.053459	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	5706	246.54	77.85	210	-50	1.056281	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5707	246.54	77.85	210	-50	1.056281	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5708	227.21	72.6	210	-50	1.063449	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5714	362.55	145.9	210	-50	1.041293	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	5715	38.18	10.06	30	-10	1.019539	100	1	100	25.45	0	0	0	0	0	0	0	0	0	0	0;
	5716	27.62	3.89	30	-20	1.006233	100	1	100	18.41	0	0	0	0	0	0	0	0	0	0	0;
	5718	214.01	124.9	160	-40	1.043748	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5723	36.77	13.93	30	-10	1.042317	100	1	100	24.51	0	0	0	0	0	0	0	0	0	0	0;
	5728	97.89	27.09	60	-20	1.044132	100	1	200	65.26	0	0	0	0	0	0	0	0	0	0	0;
	5729	97.89	27.09	60	-20	1.044132	100	1	200	65.26	0	0	0	0	0	0	0	0	0	0	0;
	5730	97.89	27.09	60	-20	1.044132	100	1	200	65.26	0	0	0	0	0	0	0	0	0	0	0;
	5731	97.89	27.09	60	-20	1.044132	100	1	200	65.26	0	0	0	0	0	0	0	0	0	0	0;
	5732	87.9	19.61	60	-20	1.034671	100	1	200	58.6	0	0	0	0	0	0	0	0	0	0	0;
	5739	35.02	1.75	30	-10	1.038438	100	1	100	23.35	0	0	0	0	0	0	0	0	0	0	0;
	5741	36.03	16.08	30	-10	1.045359	100	1	100	24.02	0	0	0	0	0	0	0	0	0	0	0;
	5742	209.95	44.18	130	-20	0.973258	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5746	32.37	5.79	20	-10	1.037284	100	1	100	21.58	0	0	0	0	0	0	0	0	0	0	0;
	5750	52.06	9.48	40	-10	1.028771	100	1	200	34.71	0	0	0	0	0	0	0	0	0	0	0;
	5751	47.57	7.97	40	-10	1.022435	100	1	100	31.71	0	0	0	0	0	0	0	0	0	0	0;
	5761	34.96	12.97	30	-10	1.054937	100	1	100	23.31	0	0	0	0	0	0	0	0	0	0	0;
	5764	34.86	4.81	30	-10	1.050901	100	1	100	23.24	0	0	0	0	0	0	0	0	0	0	0;
	5765	37.29	8.19	30	-10	1.030566	100	1	100	24.86	0	0	0	0	0	0	0	0	0	0	0;
	5769	53.07	7.74	40	-10	1.025333	100	1	200	35.38	0	0	0	0	0	0	0	0	0	0	0;
	5770	40.44	13.81	40	-10	1.025019	100	1	100	26.96	0	0	0	0	0	0	0	0	0	0	0;
	5771	40.44	13.81	40	-10	1.025019	100	1	100	26.96	0	0	0	0	0	0	0	0	0	0	0;
	5772	53.07	7.74	40	-10	1.025333	100	1	200	35.38	0	0	0	0	0	0	0	0	0	0	0;
	5776	351.74	73.14	210	-50	1.039438	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	5777	351.74	73.14	210	-50	1.039438	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	5781	351.74	73.14	210	-50	1.039438	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	5782	351.74	73.14	210	-50	1.039438	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	5797	235.87	0.58	210	-50	1.042461	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	5804	35.74	6.84	30	-10	1.040138	100	1	100	23.83	0	0	0	0	0	0	0	0	0	0	0;
	5805	31.29	6.48	30	-10	1.024788	100	1	100	20.86	0	0	0	0	0	0	0	0	0	0	0;
	5816	313.39	68.14	210	-50	1.035914	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	5825	19.27	11.16	30	-10	1.021452	100	1	100	12.85	0	0	0	0	0	0	0	0	0	0	0;
	5826	15.75	3.03	20	-10	1.013783	100	1	100	10.5	0	0	0	0	0	0	0	0	0	0	0;
	5830	62.37	25.07	60	-20	1.030828	100	1	200	41.58	0	0	0	0	0	0	0	0	0	0	0;
	5831	17.21	5.28	30	-10	1.020161	100	1	100	11.47	0	0	0	0	0	0	0	0	0	0	0;
	5832	17.21	5.28	30	-10	1.020161	100	1	100	11.47	0	0	0	0	0	0	0	0	0	0	0;
	5838	518.28	-15.34	520	-120	1.032652	100	1	1100	300	0	0	0	0	0	0	0	0	0	0	0;
	5842	29.89	15.39	30	-10	1.024479	100	1	100	19.93	0	0	0	0	0	0	0	0	0	0	0;
	5847	36.88	19.22	30	-10	1.037296	100	1	100	24.59	0	0	0	0	0	0	0	0	0	0	0;
	5849	37.8	19.78	30	-10	1.031737	100	1	100	25.2	0	0	0	0	0	0	0	0	0	0	0;
	5850	99.68	14.71	60	-20	1.035207	100	1	200	66.45	0	0	0	0	0	0	0	0	0	0	0;
	5856	159.8	43.16	160	-40	1.056042	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	5858	159.8	43.16	160	-40	1.056042	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	5864	45.28	10.84	40	-10	1.043738	100	1	100	30.19	0	0	0	0	0	0	0	0	0	0	0;
	5868	90.17	24.56	60	-20	1.06019	100	1	200	60.11	0	0	0	0	0	0	0	0	0	0	0;
	5880	81.07	20.71	60	-20	1.021915	100	1	200	54.05	0	0	0	0	0	0	0	0	0	0	0;
	5881	269.27	88.74	160	-40	1.035485	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	5882	269.15	67.63	160	-40	1.052667	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	5883	269.15	67.63	160	-40	1.052667	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	5884	269.15	67.63	160	-40	1.052667	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	5885	81.07	20.71	60	-20	1.021915	100	1	200	54.05	0	0	0	0	0	0	0	0	0	0	0;
	5886	81.07	20.71	60	-20	1.021915	100	1	200	54.05	0	0	0	0	0	0	0	0	0	0	0;
	5887	256.21	65.38	160	-40	1.052481	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	5889	256.21	65.38	160	-40	1.052481	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	5896	56.36	7.67	40	-10	1.024375	100	1	200	37.57	0	0	0	0	0	0	0	0	0	0	0;
	5897	87.39	49.11	60	-20	1.054462	100	1	200	58.26	0	0	0	0	0	0	0	0	0	0	0;
	5898	87.39	49.11	60	-20	1.054462	100	1	200	58.26	0	0	0	0	0	0	0	0	0	0	0;
	5900	87.39	49.12	60	-20	1.05445	100	1	200	58.26	0	0	0	0	0	0	0	0	0	0	0;
	5901	87.39	49.12	60	-20	1.05445	100	1	200	58.26	0	0	0	0	0	0	0	0	0	0	0;
	5902	87.39	49.11	60	-20	1.054462	100	1	200	58.26	0	0	0	0	0	0	0	0	0	0	0;
	5905	130.5	-4.01	50	-50	0.994923	100	1	300	87	0	0	0	0	0	0	0	0	0	0	0;
	5911	25.74	8.25	30	-10	1.037214	100	1	100	17.16	0	0	0	0	0	0	0	0	0	0	0;
	5924	31.28	-0.92	30	-10	1.009892	100	1	100	20.85	0	0	0	0	0	0	0	0	0	0	0;
	5928	32.64	6.67	30	-10	1.044924	100	1	100	21.76	0	0	0	0	0	0	0	0	0	0	0;
	5929	35.06	4.02	30	-10	1.032126	100	1	100	23.37	0	0	0	0	0	0	0	0	0	0	0;
	5930	35.06	4.02	30	-10	1.032126	100	1	100	23.37	0	0	0	0	0	0	0	0	0	0	0;
	5931	33.37	14.47	30	-10	1.0234	100	1	100	22.25	0	0	0	0	0	0	0	0	0	0	0;
	5932	35.06	4.02	30	-10	1.032126	100	1	100	23.37	0	0	0	0	0	0	0	0	0	0	0;
	5933	35.06	4.02	30	-10	1.032126	100	1	100	23.37	0	0	0	0	0	0	0	0	0	0	0;
	5934	35.06	4.02	30	-10	1.032126	100	1	100	23.37	0	0	0	0	0	0	0	0	0	0	0;
	5935	25.6	3.9	30	-10	1.03284	100	1	100	17.07	0	0	0	0	0	0	0	0	0	0	0;
	5936	25.6	3.9	30	-10	1.03284	100	1	100	17.07	0	0	0	0	0	0	0	0	0	0	0;
	5961	33.21	5.47	30	-10	1.022689	100	1	100	22.14	0	0	0	0	0	0	0	0	0	0	0;
	5963	26.62	13.25	30	-10	1.040642	100	1	100	17.75	0	0	0	0	0	0	0	0	0	0	0;
	5964	135	10.29	70	-20	1.008348	100	1	300	90	0	0	0	0	0	0	0	0	0	0	0;
	5965	175.1	24.83	110	-80	0.985995	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	5973	17.82	3.95	30	-10	1.066829	100	1	100	11.88	0	0	0	0	0	0	0	0	0	0	0;
	5977	29.43	7.09	30	-10	1.028438	100	1	100	19.62	0	0	0	0	0	0	0	0	0	0	0;
	5978	79.8	19.8	60	-20	1.02806	100	1	200	53.2	0	0	0	0	0	0	0	0	0	0	0;
	5979	37.23	4.01	30	-10	1.028283	100	1	100	24.82	0	0	0	0	0	0	0	0	0	0	0;
	5982	5.2	0.53	10	-10	1.031159	100	1	100	3.47	0	0	0	0	0	0	0	0	0	0	0;
	5984	32.72	7.26	30	-10	1.016616	100	1	100	21.81	0	0	0	0	0	0	0	0	0	0	0;
	5987	612.07	246.66	520	-120	1.041616	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	5988	612.07	246.66	520	-120	1.041616	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	5989	612.02	218.37	520	-120	1.039304	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	5990	612.02	218.37	520	-120	1.039304	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	5992	35.95	6.39	40	-10	1.015185	100	1	100	23.97	0	0	0	0	0	0	0	0	0	0	0;
	5999	50.34	15.84	40	-10	1.03143	100	1	200	33.56	0	0	0	0	0	0	0	0	0	0	0;
	6002	63.6	31.38	60	-20	1.027152	100	1	200	42.4	0	0	0	0	0	0	0	0	0	0	0;
	6003	63.6	31.38	60	-20	1.027152	100	1	200	42.4	0	0	0	0	0	0	0	0	0	0	0;
	6005	63.6	31.38	60	-20	1.027152	100	1	200	42.4	0	0	0	0	0	0	0	0	0	0	0;
	6006	63.6	31.38	60	-20	1.027152	100	1	200	42.4	0	0	0	0	0	0	0	0	0	0	0;
	6007	63.6	31.38	60	-20	1.027152	100	1	200	42.4	0	0	0	0	0	0	0	0	0	0	0;
	6008	63.6	31.38	60	-20	1.027152	100	1	200	42.4	0	0	0	0	0	0	0	0	0	0	0;
	6009	4.5	0.47	10	-10	1.016602	100	1	100	3	0	0	0	0	0	0	0	0	0	0	0;
	6014	50.34	15.84	40	-10	1.03143	100	1	200	33.56	0	0	0	0	0	0	0	0	0	0	0;
	6016	33.19	5.48	30	-10	0.991806	100	1	100	22.13	0	0	0	0	0	0	0	0	0	0	0;
	6017	163.94	189.92	210	-50	1.050579	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	6023	383.05	40.91	210	-50	1.045612	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	6025	92.69	14.48	60	-20	1.026207	100	1	200	61.79	0	0	0	0	0	0	0	0	0	0	0;
	6034	31.28	-0.92	30	-10	1.009892	100	1	100	20.85	0	0	0	0	0	0	0	0	0	0	0;
	6038	210	16.99	130	-60	0.981513	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6039	89.25	12.95	60	-20	1.054609	100	1	200	59.5	0	0	0	0	0	0	0	0	0	0	0;
	6040	89.25	12.95	60	-20	1.054609	100	1	200	59.5	0	0	0	0	0	0	0	0	0	0	0;
	6048	82.3	25.62	60	-20	1.055592	100	1	200	54.87	0	0	0	0	0	0	0	0	0	0	0;
	6049	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	6055	63.72	11.02	40	-10	1.030356	100	1	200	42.48	0	0	0	0	0	0	0	0	0	0	0;
	6058	424.18	1.6	260	-60	1.036943	100	1	900	200	0	0	0	0	0	0	0	0	0	0	0;
	6066	393.28	141.99	210	-50	1.019403	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	6068	36.91	13.33	30	-10	1.037973	100	1	100	24.61	0	0	0	0	0	0	0	0	0	0	0;
	6072	9.1	-0.32	10	-10	1.00419	100	1	100	6.07	0	0	0	0	0	0	0	0	0	0	0;
	6076	28.8	6.92	30	-10	1.000808	100	1	100	19.2	0	0	0	0	0	0	0	0	0	0	0;
	6085	550.03	481.13	520	-120	1.061212	100	1	1200	300	0	0	0	0	0	0	0	0	0	0	0;
	6087	26.81	10.09	30	-10	1.044747	100	1	100	17.87	0	0	0	0	0	0	0	0	0	0	0;
	6088	32.72	7.26	30	-10	1.016616	100	1	100	21.81	0	0	0	0	0	0	0	0	0	0	0;
	6089	16.74	3.04	30	-10	1.042545	100	1	100	11.16	0	0	0	0	0	0	0	0	0	0	0;
	6090	30.08	5.21	30	-10	1.030904	100	1	100	20.05	0	0	0	0	0	0	0	0	0	0	0;
	6091	30.08	5.21	30	-10	1.030904	100	1	100	20.05	0	0	0	0	0	0	0	0	0	0	0;
	6092	38.69	6.26	30	-10	1.032747	100	1	100	25.79	0	0	0	0	0	0	0	0	0	0	0;
	6093	26.81	10.09	30	-10	1.044747	100	1	100	17.87	0	0	0	0	0	0	0	0	0	0	0;
	6096	37.93	15.04	30	-10	1.022821	100	1	100	25.29	0	0	0	0	0	0	0	0	0	0	0;
	6097	16.74	3.04	30	-10	1.042545	100	1	100	11.16	0	0	0	0	0	0	0	0	0	0	0;
	6098	16.74	3.04	30	-10	1.042545	100	1	100	11.16	0	0	0	0	0	0	0	0	0	0	0;
	6103	379.71	199.24	210	-50	1.082967	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	6107	38.02	11.21	30	-10	1.043087	100	1	100	25.35	0	0	0	0	0	0	0	0	0	0	0;
	6121	85.28	21.31	60	-20	1.029542	100	1	200	56.85	0	0	0	0	0	0	0	0	0	0	0;
	6123	37.08	5.51	30	-10	1.021005	100	1	100	24.72	0	0	0	0	0	0	0	0	0	0	0;
	6130	5	0.52	10	-10	1.019692	100	1	100	3.33	0	0	0	0	0	0	0	0	0	0	0;
	6132	15	4.65	10	0	1.033224	100	1	100	10	0	0	0	0	0	0	0	0	0	0	0;
	6135	17.64	4.13	30	-10	1.031189	100	1	100	11.76	0	0	0	0	0	0	0	0	0	0	0;
	6138	667.17	206.58	520	-120	1.030727	100	1	1400	400	0	0	0	0	0	0	0	0	0	0	0;
	6139	667.17	206.58	520	-120	1.030727	100	1	1400	400	0	0	0	0	0	0	0	0	0	0	0;
	6140	667.17	206.58	520	-120	1.030727	100	1	1400	400	0	0	0	0	0	0	0	0	0	0	0;
	6141	938.69	9.52	520	-120	1.028568	100	1	1900	600	0	0	0	0	0	0	0	0	0	0	0;
	6155	31.3	3.62	20	-10	1.018029	100	1	100	20.87	0	0	0	0	0	0	0	0	0	0	0;
	6156	31.47	4.94	30	-10	1.024861	100	1	100	20.98	0	0	0	0	0	0	0	0	0	0	0;
	6161	64.84	8.42	60	-20	1.035942	100	1	200	43.23	0	0	0	0	0	0	0	0	0	0	0;
	6162	64.84	8.42	60	-20	1.035942	100	1	200	43.23	0	0	0	0	0	0	0	0	0	0	0;
	6163	4.37	0.22	10	-10	1.024003	100	1	100	2.91	0	0	0	0	0	0	0	0	0	0	0;
	6167	20.75	4.27	20	-10	1.024289	100	1	100	13.83	0	0	0	0	0	0	0	0	0	0	0;
	6171	12.3	0.28	10	-10	1.011745	100	1	100	8.2	0	0	0	0	0	0	0	0	0	0	0;
	6192	17.55	5.77	10	0	1.031157	100	1	100	11.7	0	0	0	0	0	0	0	0	0	0	0;
	6193	254.08	105.54	160	-40	1.025991	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	6194	254.08	105.54	160	-40	1.025991	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	6195	25.41	10.98	30	-10	1.019411	100	1	100	16.94	0	0	0	0	0	0	0	0	0	0	0;
	6199	27.9	5.62	10	0	1.021456	100	1	100	18.6	0	0	0	0	0	0	0	0	0	0	0;
	6200	7.7	0.04	10	-10	0.98149	100	1	100	5.13	0	0	0	0	0	0	0	0	0	0	0;
	6201	685.04	205.8	520	-120	1.048523	100	1	1400	400	0	0	0	0	0	0	0	0	0	0	0;
	6202	685.04	205.8	520	-120	1.048523	100	1	1400	400	0	0	0	0	0	0	0	0	0	0	0;
	6203	685.05	215.92	520	-120	1.050193	100	1	1400	400	0	0	0	0	0	0	0	0	0	0	0;
	6204	685.05	215.92	520	-120	1.050193	100	1	1400	400	0	0	0	0	0	0	0	0	0	0	0;
	6208	685.9	149.42	520	-120	1.062519	100	1	1400	400	0	0	0	0	0	0	0	0	0	0	0;
	6210	36.03	16.08	30	-10	1.045359	100	1	100	24.02	0	0	0	0	0	0	0	0	0	0	0;
	6212	179.95	-207.96	100000	-100000	1.021013	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	6213	41.19	12.09	40	-10	1.029044	100	1	100	27.46	0	0	0	0	0	0	0	0	0	0	0;
	6215	67.23	14.17	40	-10	1.023729	100	1	200	44.82	0	0	0	0	0	0	0	0	0	0	0;
	6218	41.19	12.09	40	-10	1.029044	100	1	100	27.46	0	0	0	0	0	0	0	0	0	0	0;
	6219	33.39	17.82	40	-10	1.025649	100	1	100	22.26	0	0	0	0	0	0	0	0	0	0	0;
	6220	67.23	14.3	40	-10	1.023566	100	1	200	44.82	0	0	0	0	0	0	0	0	0	0	0;
	6225	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	6226	350.16	118.4	210	-50	1.062007	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	6227	364.79	160.47	210	-50	1.050331	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	6254	626.68	236.63	520	-120	1.051727	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	6262	243.32	23.39	210	-50	1.055881	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6272	-0.15	1.91	10	-10	1.003919	100	1	100	-0.2	0	0	0	0	0	0	0	0	0	0	0;
	6276	284.39	98.98	210	-50	1.060659	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	6277	284.4	98.98	210	-50	1.060659	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	6278	31.29	6.48	30	-10	1.024788	100	1	100	20.86	0	0	0	0	0	0	0	0	0	0	0;
	6279	284.39	98.98	210	-50	1.060659	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	6280	257.89	39.7	210	-50	1.022087	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	6291	35.74	6.84	30	-10	1.040138	100	1	100	23.83	0	0	0	0	0	0	0	0	0	0	0;
	6296	49.83	6.62	60	-20	1.007268	100	1	100	33.22	0	0	0	0	0	0	0	0	0	0	0;
	6298	29.68	6.45	30	-10	1.016891	100	1	100	19.79	0	0	0	0	0	0	0	0	0	0	0;
	6299	29.68	6.45	30	-10	1.016891	100	1	100	19.79	0	0	0	0	0	0	0	0	0	0	0;
	6305	100	41.35	90	-50	1.090507	100	1	300	66.67	0	0	0	0	0	0	0	0	0	0	0;
	6314	2.5	0.07	10	-10	0.984601	100	1	100	1.67	0	0	0	0	0	0	0	0	0	0	0;
	6317	312.49	55.02	210	-50	1.033042	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	6325	22.09	10.91	30	-10	1.043067	100	1	100	14.73	0	0	0	0	0	0	0	0	0	0	0;
	6326	22.09	10.91	30	-10	1.043067	100	1	100	14.73	0	0	0	0	0	0	0	0	0	0	0;
	6327	22.09	10.91	30	-10	1.043067	100	1	100	14.73	0	0	0	0	0	0	0	0	0	0	0;
	6328	22.09	10.91	30	-10	1.043067	100	1	100	14.73	0	0	0	0	0	0	0	0	0	0	0;
	6329	22.09	10.91	30	-10	1.043067	100	1	100	14.73	0	0	0	0	0	0	0	0	0	0	0;
	6331	99.68	14.71	60	-20	1.035207	100	1	200	66.45	0	0	0	0	0	0	0	0	0	0	0;
	6335	34.04	7.33	40	-10	1.031942	100	1	100	22.69	0	0	0	0	0	0	0	0	0	0	0;
	6336	34.04	7.33	40	-10	1.031942	100	1	100	22.69	0	0	0	0	0	0	0	0	0	0	0;
	6337	34.04	7.33	40	-10	1.031942	100	1	100	22.69	0	0	0	0	0	0	0	0	0	0	0;
	6341	35.5	10.89	30	-10	1.045213	100	1	100	23.67	0	0	0	0	0	0	0	0	0	0	0;
	6346	29.89	15.39	30	-10	1.024479	100	1	100	19.93	0	0	0	0	0	0	0	0	0	0	0;
	6347	0.02	-0.09	10	-10	1.02409	100	1	100	0.01	0	0	0	0	0	0	0	0	0	0	0;
	6351	289.73	138.75	160	-40	1.055778	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	6352	289.73	138.75	160	-40	1.055778	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	6353	201.27	91.44	160	-40	1.070329	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6354	201.27	91.44	160	-40	1.070329	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6357	206.08	85.48	160	-40	1.048705	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6358	206.08	85.48	160	-40	1.048705	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6359	18.32	-0.94	10	-10	0.995077	100	1	100	12.21	0	0	0	0	0	0	0	0	0	0	0;
	6360	215.66	145.16	160	-40	1.043227	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6361	201.27	91.44	160	-40	1.070329	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6363	56.36	7.67	40	-10	1.024375	100	1	200	37.57	0	0	0	0	0	0	0	0	0	0	0;
	6368	617.47	92.14	520	-120	1.032321	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	6369	267.35	24.57	160	-40	1.014698	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	6370	267.35	24.57	160	-40	1.014698	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	6371	267.35	24.57	160	-40	1.014698	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	6372	267.35	24.57	160	-40	1.014698	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	6383	37	0.77	30	-30	1.003111	100	1	100	24.67	0	0	0	0	0	0	0	0	0	0	0;
	6391	82.6	2.11	60	-20	1.02421	100	1	200	55.07	0	0	0	0	0	0	0	0	0	0	0;
	6398	32.64	6.68	30	-10	1.044939	100	1	100	21.76	0	0	0	0	0	0	0	0	0	0	0;
	6401	34.22	10.13	30	-10	1.052205	100	1	100	22.81	0	0	0	0	0	0	0	0	0	0	0;
	6402	34.22	10.13	30	-10	1.052231	100	1	100	22.81	0	0	0	0	0	0	0	0	0	0	0;
	6404	34.22	10.13	30	-10	1.052205	100	1	100	22.81	0	0	0	0	0	0	0	0	0	0	0;
	6405	35.55	14.31	30	-10	1.047514	100	1	100	23.7	0	0	0	0	0	0	0	0	0	0	0;
	6406	242.59	1.87	210	-50	1.040945	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6407	34.22	10.13	30	-10	1.052231	100	1	100	22.81	0	0	0	0	0	0	0	0	0	0	0;
	6408	35.55	14.31	30	-10	1.047514	100	1	100	23.7	0	0	0	0	0	0	0	0	0	0	0;
	6409	359.06	76.92	210	-50	1.054213	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	6417	320.2	176.92	210	-50	1.054686	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	6418	320.2	176.92	210	-50	1.054686	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	6420	46.29	9.49	60	-20	0.994392	100	1	100	30.86	0	0	0	0	0	0	0	0	0	0	0;
	6421	44.03	4.68	60	-20	0.994839	100	1	100	29.35	0	0	0	0	0	0	0	0	0	0	0;
	6422	320.2	176.92	210	-50	1.054686	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	6429	13.29	-1.25	10	-10	1.021473	100	1	100	8.86	0	0	0	0	0	0	0	0	0	0	0;
	6431	21	3.83	20	-10	1.021729	100	1	100	14	0	0	0	0	0	0	0	0	0	0	0;
	6436	381.8	21.67	210	-50	1.063714	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	6440	15.84	6.64	30	-10	1.000784	100	1	100	10.56	0	0	0	0	0	0	0	0	0	0	0;
	6451	25.45	8.23	30	-10	1.004119	100	1	100	16.97	0	0	0	0	0	0	0	0	0	0	0;
	6452	25.39	2.92	30	-10	1.004063	100	1	100	16.93	0	0	0	0	0	0	0	0	0	0	0;
	6454	46.7	16.39	40	-10	1.027739	100	1	100	31.13	0	0	0	0	0	0	0	0	0	0	0;
	6455	40.16	6.22	30	-10	1.012552	100	1	100	26.77	0	0	0	0	0	0	0	0	0	0	0;
	6456	40.16	6.22	30	-10	1.012552	100	1	100	26.77	0	0	0	0	0	0	0	0	0	0	0;
	6465	34.86	4.81	30	-10	1.050901	100	1	100	23.24	0	0	0	0	0	0	0	0	0	0	0;
	6471	57.67	14.65	40	-10	1.028916	100	1	200	38.45	0	0	0	0	0	0	0	0	0	0	0;
	6476	260.62	111.59	210	-50	1.031498	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	6477	20.63	6.57	30	-10	1.022447	100	1	100	13.75	0	0	0	0	0	0	0	0	0	0	0;
	6487	37.23	4.01	30	-10	1.028283	100	1	100	24.82	0	0	0	0	0	0	0	0	0	0	0;
	6491	138.8	17.14	110	-30	1.04554	100	1	300	92.53	0	0	0	0	0	0	0	0	0	0	0;
	6498	60.07	-4.22	60	-20	1.016868	100	1	200	40.05	0	0	0	0	0	0	0	0	0	0	0;
	6500	277.81	77.81	210	-50	1.034935	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	6507	31.28	-0.92	30	-10	1.009892	100	1	100	20.85	0	0	0	0	0	0	0	0	0	0	0;
	6527	204	41.74	150	-60	1.035162	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6539	16.79	7.28	30	-10	1.042166	100	1	100	11.19	0	0	0	0	0	0	0	0	0	0	0;
	6541	39.88	15.15	30	-10	1.024267	100	1	100	26.59	0	0	0	0	0	0	0	0	0	0	0;
	6542	39.88	15.15	30	-10	1.024267	100	1	100	26.59	0	0	0	0	0	0	0	0	0	0	0;
	6543	39.88	15.15	30	-10	1.024267	100	1	100	26.59	0	0	0	0	0	0	0	0	0	0	0;
	6544	93.64	6.04	60	-20	1.023818	100	1	200	62.43	0	0	0	0	0	0	0	0	0	0	0;
	6545	24.99	21.34	30	-10	1.019094	100	1	100	16.66	0	0	0	0	0	0	0	0	0	0	0;
	6547	93.64	6.04	60	-20	1.023818	100	1	200	62.43	0	0	0	0	0	0	0	0	0	0	0;
	6548	93.64	6.04	60	-20	1.023818	100	1	200	62.43	0	0	0	0	0	0	0	0	0	0	0;
	6557	47.94	3.29	40	-10	1.023603	100	1	100	31.96	0	0	0	0	0	0	0	0	0	0	0;
	6563	31.3	-0.89	20	-20	0.998951	100	1	100	20.87	0	0	0	0	0	0	0	0	0	0	0;
	6566	34.65	5.41	30	-10	1.014865	100	1	100	23.1	0	0	0	0	0	0	0	0	0	0	0;
	6567	34.65	5.41	30	-10	1.014865	100	1	100	23.1	0	0	0	0	0	0	0	0	0	0	0;
	6584	154.5	-14.56	80	-70	0.956299	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	6585	114	6.43	70	-50	0.989649	100	1	300	76	0	0	0	0	0	0	0	0	0	0	0;
	6593	38.27	1.62	30	-10	1.047126	100	1	100	25.51	0	0	0	0	0	0	0	0	0	0	0;
	6601	192.55	1.87	160	-40	1.012099	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	6602	37.08	5.51	30	-10	1.021005	100	1	100	24.72	0	0	0	0	0	0	0	0	0	0	0;
	6606	88.06	35.93	60	-20	1.024151	100	1	200	58.71	0	0	0	0	0	0	0	0	0	0	0;
	6607	88.06	35.93	60	-20	1.024151	100	1	200	58.71	0	0	0	0	0	0	0	0	0	0	0;
	6608	88.06	35.93	60	-20	1.024151	100	1	200	58.71	0	0	0	0	0	0	0	0	0	0	0;
	6609	88.06	35.93	60	-20	1.024151	100	1	200	58.71	0	0	0	0	0	0	0	0	0	0	0;
	6610	88.06	35.93	60	-20	1.024151	100	1	200	58.71	0	0	0	0	0	0	0	0	0	0	0;
	6611	88.06	35.93	60	-20	1.024151	100	1	200	58.71	0	0	0	0	0	0	0	0	0	0	0;
	6612	88.06	35.93	60	-20	1.024151	100	1	200	58.71	0	0	0	0	0	0	0	0	0	0	0;
	6616	138.8	-5.3	110	-30	1.029779	100	1	300	92.53	0	0	0	0	0	0	0	0	0	0	0;
	6620	323.35	77.76	210	-50	1.058114	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	6621	23.65	10.04	30	-10	1.059543	100	1	100	15.77	0	0	0	0	0	0	0	0	0	0	0;
	6622	52.64	8.15	30	-10	1.030338	100	1	200	35.09	0	0	0	0	0	0	0	0	0	0	0;
	6623	30.67	11.2	30	-10	1.040509	100	1	100	20.45	0	0	0	0	0	0	0	0	0	0	0;
	6624	30.67	11.2	30	-10	1.040509	100	1	100	20.45	0	0	0	0	0	0	0	0	0	0	0;
	6628	301.72	83.89	210	-50	1.041212	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	6630	30.67	11.2	30	-10	1.040509	100	1	100	20.45	0	0	0	0	0	0	0	0	0	0	0;
	6631	38.52	8.15	30	-10	1.04659	100	1	100	25.68	0	0	0	0	0	0	0	0	0	0	0;
	6639	23.85	6.18	40	-10	1.031885	100	1	100	15.9	0	0	0	0	0	0	0	0	0	0	0;
	6640	23.85	6.18	40	-10	1.031885	100	1	100	15.9	0	0	0	0	0	0	0	0	0	0	0;
	6641	23.85	6.18	40	-10	1.031885	100	1	100	15.9	0	0	0	0	0	0	0	0	0	0	0;
	6642	23.85	6.18	40	-10	1.031885	100	1	100	15.9	0	0	0	0	0	0	0	0	0	0	0;
	6649	8.2	-2.56	10	-10	0.955439	100	1	100	5.47	0	0	0	0	0	0	0	0	0	0	0;
	6650	96.03	17.22	60	-20	1.036065	100	1	200	64.02	0	0	0	0	0	0	0	0	0	0	0;
	6651	74.78	41.98	60	-20	1.04914	100	1	200	49.85	0	0	0	0	0	0	0	0	0	0	0;
	6654	45.28	10.84	40	-10	1.043738	100	1	100	30.19	0	0	0	0	0	0	0	0	0	0	0;
	6663	91.17	37.8	60	-20	1.045666	100	1	200	60.78	0	0	0	0	0	0	0	0	0	0	0;
	6667	203.46	82.67	210	-50	1.034595	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6670	36.77	13.93	30	-10	1.042317	100	1	100	24.51	0	0	0	0	0	0	0	0	0	0	0;
	6680	55.25	2.21	40	-40	0.994234	100	1	200	36.83	0	0	0	0	0	0	0	0	0	0	0;
	6686	25.44	-2.35	20	-20	0.995022	100	1	100	16.96	0	0	0	0	0	0	0	0	0	0	0;
	6689	49.34	9.15	40	-10	1.034597	100	1	100	32.89	0	0	0	0	0	0	0	0	0	0	0;
	6690	49.34	9.15	40	-10	1.034597	100	1	100	32.89	0	0	0	0	0	0	0	0	0	0	0;
	6691	49.34	9.15	40	-10	1.034597	100	1	100	32.89	0	0	0	0	0	0	0	0	0	0	0;
	6693	50.34	9.17	40	-10	1.034532	100	1	200	33.56	0	0	0	0	0	0	0	0	0	0	0;
	6697	50.34	9.17	40	-10	1.034532	100	1	200	33.56	0	0	0	0	0	0	0	0	0	0	0;
	6709	16.12	0.49	10	-10	1.021634	100	1	100	10.75	0	0	0	0	0	0	0	0	0	0	0;
	6713	57.54	22.74	40	-10	1.022626	100	1	200	38.36	0	0	0	0	0	0	0	0	0	0	0;
	6717	2	-0	10	-10	0.996757	100	1	100	1.33	0	0	0	0	0	0	0	0	0	0	0;
	6719	68.89	5.63	60	-20	1.028667	100	1	200	45.93	0	0	0	0	0	0	0	0	0	0	0;
	6721	68.89	5.63	60	-20	1.028667	100	1	200	45.93	0	0	0	0	0	0	0	0	0	0	0;
	6726	32.05	4.77	20	-10	1.016479	100	1	100	21.37	0	0	0	0	0	0	0	0	0	0	0;
	6729	323.35	77.76	210	-50	1.058114	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	6731	22.46	14.75	30	-10	1.043942	100	1	100	14.97	0	0	0	0	0	0	0	0	0	0	0;
	6734	245.73	1.48	160	-40	1.005341	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6747	169.6	14.07	130	-80	1.016595	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	6749	33.68	6.27	30	-10	1.046665	100	1	100	22.45	0	0	0	0	0	0	0	0	0	0	0;
	6754	214.6	55.92	210	-50	1.073284	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6757	33.84	9.43	30	-10	1.043178	100	1	100	22.56	0	0	0	0	0	0	0	0	0	0	0;
	6758	33.84	9.43	30	-10	1.043178	100	1	100	22.56	0	0	0	0	0	0	0	0	0	0	0;
	6759	33.84	9.43	30	-10	1.043178	100	1	100	22.56	0	0	0	0	0	0	0	0	0	0	0;
	6760	33.84	9.43	30	-10	1.043178	100	1	100	22.56	0	0	0	0	0	0	0	0	0	0	0;
	6761	33.84	9.43	30	-10	1.043178	100	1	100	22.56	0	0	0	0	0	0	0	0	0	0	0;
	6762	33.84	9.43	30	-10	1.043178	100	1	100	22.56	0	0	0	0	0	0	0	0	0	0	0;
	6763	33.84	9.43	30	-10	1.043178	100	1	100	22.56	0	0	0	0	0	0	0	0	0	0	0;
	6764	86.6	49.99	60	-20	1.041274	100	1	200	57.73	0	0	0	0	0	0	0	0	0	0	0;
	6765	95.35	51.27	60	-20	1.042697	100	1	200	63.57	0	0	0	0	0	0	0	0	0	0	0;
	6769	21.03	9.77	30	-10	1.034441	100	1	100	14.02	0	0	0	0	0	0	0	0	0	0	0;
	6773	46.38	17.23	40	-10	1.048358	100	1	100	30.92	0	0	0	0	0	0	0	0	0	0	0;
	6774	33.61	7.79	30	-10	1.005097	100	1	100	22.41	0	0	0	0	0	0	0	0	0	0	0;
	6780	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	6784	8.4	0.33	10	-10	1.014368	100	1	100	5.6	0	0	0	0	0	0	0	0	0	0	0;
	6785	33.61	7.79	30	-10	1.005097	100	1	100	22.41	0	0	0	0	0	0	0	0	0	0	0;
	6789	33.61	7.79	30	-10	1.005097	100	1	100	22.41	0	0	0	0	0	0	0	0	0	0	0;
	6792	23.13	4.04	30	-10	1.024356	100	1	100	15.42	0	0	0	0	0	0	0	0	0	0	0;
	6793	23.13	4.04	30	-10	1.024356	100	1	100	15.42	0	0	0	0	0	0	0	0	0	0	0;
	6794	21.19	8.66	30	-10	1.017963	100	1	100	14.13	0	0	0	0	0	0	0	0	0	0	0;
	6795	21.19	8.66	30	-10	1.017963	100	1	100	14.13	0	0	0	0	0	0	0	0	0	0	0;
	6797	34.86	9.16	30	-10	1.014343	100	1	100	23.24	0	0	0	0	0	0	0	0	0	0	0;
	6798	34.86	9.16	30	-10	1.014343	100	1	100	23.24	0	0	0	0	0	0	0	0	0	0	0;
	6799	35.01	1.48	30	-10	1.032132	100	1	100	23.34	0	0	0	0	0	0	0	0	0	0	0;
	6801	20.77	8.67	30	-10	1.040566	100	1	100	13.85	0	0	0	0	0	0	0	0	0	0	0;
	6802	20.77	8.67	30	-10	1.040566	100	1	100	13.85	0	0	0	0	0	0	0	0	0	0	0;
	6806	41.79	9.53	40	-10	1.021328	100	1	100	27.86	0	0	0	0	0	0	0	0	0	0	0;
	6807	41.79	9.53	40	-10	1.021328	100	1	100	27.86	0	0	0	0	0	0	0	0	0	0	0;
	6808	41.79	9.53	40	-10	1.021328	100	1	100	27.86	0	0	0	0	0	0	0	0	0	0	0;
	6814	41.79	9.53	40	-10	1.021328	100	1	100	27.86	0	0	0	0	0	0	0	0	0	0	0;
	6815	41.79	9.53	40	-10	1.021328	100	1	100	27.86	0	0	0	0	0	0	0	0	0	0	0;
	6820	35.5	10.89	30	-10	1.045213	100	1	100	23.67	0	0	0	0	0	0	0	0	0	0	0;
	6822	210	37.99	140	-20	0.991398	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6836	240.6	61.27	100	-30	0.990475	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6842	21.03	9.75	30	-10	1.034464	100	1	100	14.02	0	0	0	0	0	0	0	0	0	0	0;
	6846	23.5	5.12	30	-10	1.029146	100	1	100	15.67	0	0	0	0	0	0	0	0	0	0	0;
	6847	25.08	6.95	30	-10	1.024018	100	1	100	16.72	0	0	0	0	0	0	0	0	0	0	0;
	6861	16.28	8.5	30	-10	1.027518	100	1	100	10.85	0	0	0	0	0	0	0	0	0	0	0;
	6862	86.61	21.7	60	-20	1.039025	100	1	200	57.74	0	0	0	0	0	0	0	0	0	0	0;
	6864	2.5	0.07	10	-10	0.984601	100	1	100	1.67	0	0	0	0	0	0	0	0	0	0	0;
	6865	24.84	13.51	30	-10	1.036049	100	1	100	16.56	0	0	0	0	0	0	0	0	0	0	0;
	6870	95.02	8.09	60	-20	1.023144	100	1	200	63.35	0	0	0	0	0	0	0	0	0	0	0;
	6871	157.5	-11.97	80	-80	0.984599	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	6875	95.01	14.99	60	-20	1.024647	100	1	200	63.34	0	0	0	0	0	0	0	0	0	0	0;
	6876	95.01	14.99	60	-20	1.024647	100	1	200	63.34	0	0	0	0	0	0	0	0	0	0	0;
	6877	95.01	14.99	60	-20	1.024647	100	1	200	63.34	0	0	0	0	0	0	0	0	0	0	0;
	6879	863.43	303.77	520	-120	1.066491	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	6880	863.43	305.56	520	-120	1.066564	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	6885	82.46	55.28	60	-20	1.127948	100	1	200	54.97	0	0	0	0	0	0	0	0	0	0	0;
	6886	55.94	27.58	40	-10	1.023096	100	1	200	37.29	0	0	0	0	0	0	0	0	0	0	0;
	6903	236.37	0.58	210	-50	1.042446	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6909	329.87	46.94	210	-50	1.023155	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	6910	329.87	46.94	210	-50	1.023155	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	6912	100.41	11.92	60	-20	1.038747	100	1	300	66.94	0	0	0	0	0	0	0	0	0	0	0;
	6918	226.17	144.77	210	-50	1.073133	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6930	84.12	16.88	60	-20	1.028641	100	1	200	56.08	0	0	0	0	0	0	0	0	0	0	0;
	6931	84.12	16.88	60	-20	1.028641	100	1	200	56.08	0	0	0	0	0	0	0	0	0	0	0;
	6939	10.5	-0.35	10	-10	0.994858	100	1	100	7	0	0	0	0	0	0	0	0	0	0	0;
	6942	5.2	0.51	10	-10	1.017532	100	1	100	3.47	0	0	0	0	0	0	0	0	0	0	0;
	6944	38.42	21.93	30	-10	1.054031	100	1	100	25.61	0	0	0	0	0	0	0	0	0	0	0;
	6948	23.6	10.13	30	-10	1.031397	100	1	100	15.73	0	0	0	0	0	0	0	0	0	0	0;
	6952	8.2	-2.56	10	-10	0.955439	100	1	100	5.47	0	0	0	0	0	0	0	0	0	0	0;
	6963	37.55	7.2	30	-10	1.029893	100	1	100	25.03	0	0	0	0	0	0	0	0	0	0	0;
	6967	38.42	21.93	30	-10	1.054031	100	1	100	25.61	0	0	0	0	0	0	0	0	0	0	0;
	6969	38.42	21.93	30	-10	1.054031	100	1	100	25.61	0	0	0	0	0	0	0	0	0	0	0;
	6976	31.2	5.73	30	-10	1.050684	100	1	100	20.8	0	0	0	0	0	0	0	0	0	0	0;
	6985	77.81	13.09	60	-20	1.029991	100	1	200	51.87	0	0	0	0	0	0	0	0	0	0	0;
	6987	77.81	13.09	60	-20	1.029991	100	1	200	51.87	0	0	0	0	0	0	0	0	0	0	0;
	6990	24.65	14.16	30	-10	1.027131	100	1	100	16.43	0	0	0	0	0	0	0	0	0	0	0;
	6992	32.2	7.05	30	-10	1.034688	100	1	100	21.47	0	0	0	0	0	0	0	0	0	0	0;
	6994	5.8	-0.78	10	-10	1.003942	100	1	100	3.87	0	0	0	0	0	0	0	0	0	0	0;
	6996	223.39	-18.71	160	-40	1.0182	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	6998	83.78	15.02	60	-20	1.033534	100	1	200	55.85	0	0	0	0	0	0	0	0	0	0	0;
	7005	381.4	117.77	210	-50	1.04062	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	7009	15.19	3.2	20	-10	1.023432	100	1	100	10.13	0	0	0	0	0	0	0	0	0	0	0;
	7020	78.71	19.55	60	-20	1.033214	100	1	200	52.47	0	0	0	0	0	0	0	0	0	0	0;
	7028	38.02	11.21	30	-10	1.043087	100	1	100	25.35	0	0	0	0	0	0	0	0	0	0	0;
	7031	24.28	7.01	30	-10	1.012542	100	1	100	16.19	0	0	0	0	0	0	0	0	0	0	0;
	7045	273.28	112.89	210	-50	1.041985	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	7052	23.2	-0.19	20	-20	0.99693	100	1	100	15.47	0	0	0	0	0	0	0	0	0	0	0;
	7055	88.45	51.85	60	-20	1.046478	100	1	200	58.97	0	0	0	0	0	0	0	0	0	0	0;
	7056	31.96	10.69	30	-10	1.028102	100	1	100	21.31	0	0	0	0	0	0	0	0	0	0	0;
	7057	35.74	6.84	30	-10	1.040138	100	1	100	23.83	0	0	0	0	0	0	0	0	0	0	0;
	7063	617.75	183.04	520	-120	1.058676	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	7064	34.31	1.99	30	-10	1.015531	100	1	100	22.87	0	0	0	0	0	0	0	0	0	0	0;
	7066	6.52	-0.06	10	-10	0.994601	100	1	100	4.35	0	0	0	0	0	0	0	0	0	0	0;
	7068	16.1	3.37	20	-10	1.026372	100	1	100	10.73	0	0	0	0	0	0	0	0	0	0	0;
	7070	4.56	0.3	10	-10	1.002967	100	1	100	3.04	0	0	0	0	0	0	0	0	0	0	0;
	7071	323.35	77.76	210	-50	1.058114	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	7072	258.95	65.97	210	-50	1.036699	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	7073	179.12	14.03	160	-40	1.034768	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	7075	24.89	5.96	30	-10	1.042846	100	1	100	16.59	0	0	0	0	0	0	0	0	0	0	0;
	7079	54.05	11.46	40	-10	1.024254	100	1	200	36.03	0	0	0	0	0	0	0	0	0	0	0;
	7080	23.66	2.76	30	-10	1.020703	100	1	100	15.77	0	0	0	0	0	0	0	0	0	0	0;
	7081	23.66	2.76	30	-10	1.020703	100	1	100	15.77	0	0	0	0	0	0	0	0	0	0	0;
	7082	23.66	2.76	30	-10	1.020703	100	1	100	15.77	0	0	0	0	0	0	0	0	0	0	0;
	7083	23.66	2.76	30	-10	1.020703	100	1	100	15.77	0	0	0	0	0	0	0	0	0	0	0;
	7084	23.96	4.79	30	-10	1.007614	100	1	100	15.97	0	0	0	0	0	0	0	0	0	0	0;
	7085	23.66	2.76	30	-10	1.020703	100	1	100	15.77	0	0	0	0	0	0	0	0	0	0	0;
	7087	23.96	4.79	30	-10	1.007614	100	1	100	15.97	0	0	0	0	0	0	0	0	0	0	0;
	7105	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	7110	91.41	18.38	60	-20	1.020566	100	1	200	60.94	0	0	0	0	0	0	0	0	0	0	0;
	7111	177.84	-23.15	160	-40	0.995435	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	7113	224.13	67.46	160	-40	1.034235	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	7114	48.11	1.87	30	-20	1.014725	100	1	100	32.07	0	0	0	0	0	0	0	0	0	0	0;
	7115	35.74	6.84	30	-10	1.040138	100	1	100	23.83	0	0	0	0	0	0	0	0	0	0	0;
	7116	36.77	13.93	30	-10	1.042317	100	1	100	24.51	0	0	0	0	0	0	0	0	0	0	0;
	7122	56.25	11.12	40	-10	1.03036	100	1	200	37.5	0	0	0	0	0	0	0	0	0	0	0;
	7133	45.54	5.69	40	-10	1.044129	100	1	100	30.36	0	0	0	0	0	0	0	0	0	0	0;
	7137	172.98	103.13	160	-40	1.062747	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	7139	17.8	10.21	30	-10	1.034631	100	1	100	11.87	0	0	0	0	0	0	0	0	0	0	0;
	7142	87.35	17.85	60	-20	1.033768	100	1	200	58.23	0	0	0	0	0	0	0	0	0	0	0;
	7143	8.74	0.02	10	-10	1.017775	100	1	100	5.83	0	0	0	0	0	0	0	0	0	0	0;
	7146	44.9	-0.09	60	-20	1.017024	100	1	100	29.93	0	0	0	0	0	0	0	0	0	0	0;
	7148	86.25	-3.04	60	-80	0.962614	100	1	200	57.5	0	0	0	0	0	0	0	0	0	0	0;
	7149	85	-13.61	60	-60	0.969146	100	1	200	56.67	0	0	0	0	0	0	0	0	0	0	0;
	7156	38.18	10.06	30	-10	1.019539	100	1	100	25.45	0	0	0	0	0	0	0	0	0	0	0;
	7159	13	-2.26	10	-10	0.989224	100	1	100	8.67	0	0	0	0	0	0	0	0	0	0	0;
	7163	32.25	9.28	40	-10	1.025208	100	1	100	21.5	0	0	0	0	0	0	0	0	0	0	0;
	7164	34.96	12.97	30	-10	1.054937	100	1	100	23.31	0	0	0	0	0	0	0	0	0	0	0;
	7168	36.03	16.08	30	-10	1.045359	100	1	100	24.02	0	0	0	0	0	0	0	0	0	0	0;
	7170	112	-3.86	80	-60	0.989707	100	1	300	74.67	0	0	0	0	0	0	0	0	0	0	0;
	7173	88.32	56.36	60	-20	1.127961	100	1	200	58.88	0	0	0	0	0	0	0	0	0	0	0;
	7178	30.24	6.83	30	-10	1.037281	100	1	100	20.16	0	0	0	0	0	0	0	0	0	0	0;
	7179	95.28	12.14	50	-20	1.004576	100	1	200	63.52	0	0	0	0	0	0	0	0	0	0	0;
	7183	47.36	7.24	40	-10	1.026147	100	1	100	31.57	0	0	0	0	0	0	0	0	0	0	0;
	7186	6.38	-0.43	10	-10	0.962211	100	1	100	4.25	0	0	0	0	0	0	0	0	0	0	0;
	7200	47.36	7.24	40	-10	1.026147	100	1	100	31.57	0	0	0	0	0	0	0	0	0	0	0;
	7213	12.6	0.36	10	-10	1.014795	100	1	100	8.4	0	0	0	0	0	0	0	0	0	0	0;
	7228	32.43	17.87	30	-10	1.053647	100	1	100	21.62	0	0	0	0	0	0	0	0	0	0	0;
	7233	209.96	80.03	160	-40	1.038999	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	7236	23.91	3.61	30	-10	1.007821	100	1	100	15.94	0	0	0	0	0	0	0	0	0	0	0;
	7240	35.29	10.61	30	-10	1.043764	100	1	100	23.53	0	0	0	0	0	0	0	0	0	0	0;
	7242	31.92	6.07	30	-10	1.027264	100	1	100	21.28	0	0	0	0	0	0	0	0	0	0	0;
	7243	6.38	-2.2	10	-10	0.958146	100	1	100	4.25	0	0	0	0	0	0	0	0	0	0	0;
	7249	28.48	6.95	30	-10	1.005915	100	1	100	18.99	0	0	0	0	0	0	0	0	0	0	0;
	7250	19.82	3.12	20	-10	1.020233	100	1	100	13.21	0	0	0	0	0	0	0	0	0	0	0;
	7252	601.87	179.7	520	-120	1.037622	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	7257	25	4.2	20	-10	1.03161	100	1	100	16.67	0	0	0	0	0	0	0	0	0	0	0;
	7261	2.5	0.07	10	-10	0.984601	100	1	100	1.67	0	0	0	0	0	0	0	0	0	0	0;
	7262	100.89	15.27	60	-20	0.991024	100	1	300	67.26	0	0	0	0	0	0	0	0	0	0	0;
	7267	6.52	0.05	10	-10	0.996117	100	1	100	4.35	0	0	0	0	0	0	0	0	0	0	0;
	7282	6.96	1.26	10	-10	1.041421	100	1	100	4.64	0	0	0	0	0	0	0	0	0	0	0;
	7289	87.35	17.85	60	-20	1.033768	100	1	200	58.23	0	0	0	0	0	0	0	0	0	0	0;
	7291	24.37	4.5	30	-10	1.00571	100	1	100	16.25	0	0	0	0	0	0	0	0	0	0	0;
	7297	77.25	12.22	60	-20	1.034048	100	1	200	51.5	0	0	0	0	0	0	0	0	0	0	0;
	7304	56.25	11.12	40	-10	1.03036	100	1	200	37.5	0	0	0	0	0	0	0	0	0	0	0;
	7317	36.42	7.96	30	-10	1.029166	100	1	100	24.28	0	0	0	0	0	0	0	0	0	0	0;
	7327	158.9	39.46	100	-30	1.020032	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	7329	38	-2.61	20	-20	0.992436	100	1	100	25.33	0	0	0	0	0	0	0	0	0	0	0;
	7333	87.88	32.08	60	-20	1.042865	100	1	200	58.59	0	0	0	0	0	0	0	0	0	0	0;
	7334	218.81	74.1	210	-50	1.033239	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	7338	311.43	2.71	210	-50	0.999789	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	7340	159.8	16.39	100	-50	0.938623	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	7342	20.94	5.59	30	-10	1.044097	100	1	100	13.96	0	0	0	0	0	0	0	0	0	0	0;
	7345	3.7	0.68	10	-10	1.030841	100	1	100	2.47	0	0	0	0	0	0	0	0	0	0	0;
	7346	36.6	6.27	30	-10	1.027328	100	1	100	24.4	0	0	0	0	0	0	0	0	0	0	0;
	7349	35.22	5.06	30	-10	1.040292	100	1	100	23.48	0	0	0	0	0	0	0	0	0	0	0;
	7350	35.22	5.06	30	-10	1.040292	100	1	100	23.48	0	0	0	0	0	0	0	0	0	0	0;
	7351	96.03	17.22	60	-20	1.036065	100	1	200	64.02	0	0	0	0	0	0	0	0	0	0	0;
	7352	37.55	7.2	30	-10	1.029893	100	1	100	25.03	0	0	0	0	0	0	0	0	0	0	0;
	7354	677.41	27.58	520	-120	1.024274	100	1	1400	400	0	0	0	0	0	0	0	0	0	0	0;
	7356	35.02	1.75	30	-10	1.038438	100	1	100	23.35	0	0	0	0	0	0	0	0	0	0	0;
	7365	36.91	13.33	30	-10	1.037973	100	1	100	24.61	0	0	0	0	0	0	0	0	0	0	0;
	7369	32.75	4.43	30	-10	1.005303	100	1	100	21.83	0	0	0	0	0	0	0	0	0	0	0;
	7370	33.64	5.18	30	-10	1.007216	100	1	100	22.43	0	0	0	0	0	0	0	0	0	0	0;
	7371	34.92	9.74	30	-10	1.043492	100	1	100	23.28	0	0	0	0	0	0	0	0	0	0	0;
	7372	33.64	5.18	30	-10	1.007216	100	1	100	22.43	0	0	0	0	0	0	0	0	0	0	0;
	7374	32.75	4.43	30	-10	1.005303	100	1	100	21.83	0	0	0	0	0	0	0	0	0	0	0;
	7375	32.75	4.43	30	-10	1.005303	100	1	100	21.83	0	0	0	0	0	0	0	0	0	0	0;
	7376	32.75	4.43	30	-10	1.005303	100	1	100	21.83	0	0	0	0	0	0	0	0	0	0	0;
	7377	29.93	6.22	40	-10	1.001397	100	1	100	19.95	0	0	0	0	0	0	0	0	0	0	0;
	7378	30.28	3.94	30	-10	1.004934	100	1	100	20.19	0	0	0	0	0	0	0	0	0	0	0;
	7379	30.28	3.94	30	-10	1.004934	100	1	100	20.19	0	0	0	0	0	0	0	0	0	0	0;
	7381	168.75	16.36	140	-130	0.976475	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	7386	42.41	17.68	60	-20	1.058287	100	1	100	28.27	0	0	0	0	0	0	0	0	0	0	0;
	7388	475.26	-51.2	520	-120	1.038881	100	1	1000	300	0	0	0	0	0	0	0	0	0	0	0;
	7391	40.44	13.81	40	-10	1.025019	100	1	100	26.96	0	0	0	0	0	0	0	0	0	0	0;
	7392	38.16	4.12	30	-10	1.055337	100	1	100	25.44	0	0	0	0	0	0	0	0	0	0	0;
	7416	18.4	-1.01	20	-20	1.003342	100	1	100	12.27	0	0	0	0	0	0	0	0	0	0	0;
	7426	279.66	54.13	210	-50	1.030874	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	7428	38.56	8.03	30	-10	1.049495	100	1	100	25.71	0	0	0	0	0	0	0	0	0	0	0;
	7430	25.6	-1.25	20	-20	0.991002	100	1	100	17.07	0	0	0	0	0	0	0	0	0	0	0;
	7447	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	7449	240.6	61.27	100	-30	0.990475	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	7452	53.72	11.91	40	-10	1.035023	100	1	200	35.81	0	0	0	0	0	0	0	0	0	0	0;
	7470	841.6	252.59	520	-120	1.060213	100	1	1700	500	0	0	0	0	0	0	0	0	0	0	0;
	7473	23.5	6.96	20	-10	1.015521	100	1	100	15.67	0	0	0	0	0	0	0	0	0	0	0;
	7474	34.17	11.11	30	-10	1.015682	100	1	100	22.78	0	0	0	0	0	0	0	0	0	0	0;
	7475	35.29	10.61	30	-10	1.043764	100	1	100	23.53	0	0	0	0	0	0	0	0	0	0	0;
	7476	262.4	19.68	160	-40	1.015161	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	7477	262.4	19.68	160	-40	1.015161	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	7487	45	9.88	20	0	1.031445	100	1	100	30	0	0	0	0	0	0	0	0	0	0	0;
	7489	261.4	37.15	100	-30	0.969605	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	7490	31.89	4.43	30	-10	1.028453	100	1	100	21.26	0	0	0	0	0	0	0	0	0	0	0;
	7498	78.62	8.98	60	-20	1.056333	100	1	200	52.41	0	0	0	0	0	0	0	0	0	0	0;
	7499	97.57	36.99	60	-20	1.074058	100	1	200	65.05	0	0	0	0	0	0	0	0	0	0	0;
	7501	25.54	8.34	30	-10	1.050494	100	1	100	17.03	0	0	0	0	0	0	0	0	0	0	0;
	7502	18.52	11.72	30	-10	1.044704	100	1	100	12.35	0	0	0	0	0	0	0	0	0	0	0;
	7504	18.52	11.72	30	-10	1.044704	100	1	100	12.35	0	0	0	0	0	0	0	0	0	0	0;
	7505	18.76	6.31	30	-10	1.033648	100	1	100	12.51	0	0	0	0	0	0	0	0	0	0	0;
	7507	25.54	8.34	30	-10	1.050494	100	1	100	17.03	0	0	0	0	0	0	0	0	0	0	0;
	7508	27.56	15.15	30	-10	1.036203	100	1	100	18.37	0	0	0	0	0	0	0	0	0	0	0;
	7511	164.92	6.67	160	-40	1.016367	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	7512	32.57	7.29	30	-10	1.041748	100	1	100	21.71	0	0	0	0	0	0	0	0	0	0	0;
	7513	20.93	2.5	30	-10	1.050204	100	1	100	13.95	0	0	0	0	0	0	0	0	0	0	0;
	7523	88.45	51.85	60	-20	1.046478	100	1	200	58.97	0	0	0	0	0	0	0	0	0	0	0;
	7524	88.45	51.85	60	-20	1.046478	100	1	200	58.97	0	0	0	0	0	0	0	0	0	0	0;
	7526	88.45	51.85	60	-20	1.046478	100	1	200	58.97	0	0	0	0	0	0	0	0	0	0	0;
	7528	88.45	51.85	60	-20	1.046478	100	1	200	58.97	0	0	0	0	0	0	0	0	0	0	0;
	7531	279.13	130.91	210	-50	1.048637	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	7535	138.8	17.14	110	-30	1.04554	100	1	300	92.53	0	0	0	0	0	0	0	0	0	0	0;
	7539	92.69	14.48	60	-20	1.026207	100	1	200	61.79	0	0	0	0	0	0	0	0	0	0	0;
	7540	37.03	8.4	30	-10	1.016705	100	1	100	24.69	0	0	0	0	0	0	0	0	0	0	0;
	7544	86.41	23.5	60	-20	1.031898	100	1	200	57.61	0	0	0	0	0	0	0	0	0	0	0;
	7547	20.37	8.85	30	-10	1.045375	100	1	100	13.58	0	0	0	0	0	0	0	0	0	0	0;
	7551	98.62	18.17	60	-20	1.044493	100	1	200	65.75	0	0	0	0	0	0	0	0	0	0	0;
	7556	94.5	8.92	50	-40	0.979015	100	1	200	63	0	0	0	0	0	0	0	0	0	0	0;
	7557	40.51	5.96	30	-10	1.017287	100	1	100	27.01	0	0	0	0	0	0	0	0	0	0	0;
	7558	40.51	5.96	30	-10	1.017287	100	1	100	27.01	0	0	0	0	0	0	0	0	0	0	0;
	7559	40.51	5.96	30	-10	1.017287	100	1	100	27.01	0	0	0	0	0	0	0	0	0	0	0;
	7560	40.51	5.96	30	-10	1.017287	100	1	100	27.01	0	0	0	0	0	0	0	0	0	0	0;
	7561	89.66	46.76	60	-20	1.050891	100	1	200	59.77	0	0	0	0	0	0	0	0	0	0	0;
	7562	22.13	3.21	30	-10	1.026207	100	1	100	14.75	0	0	0	0	0	0	0	0	0	0	0;
	7563	22.13	3.21	30	-10	1.026207	100	1	100	14.75	0	0	0	0	0	0	0	0	0	0	0;
	7564	17.96	1.67	30	-10	1.048808	100	1	100	11.97	0	0	0	0	0	0	0	0	0	0	0;
	7568	31	0.14	20	-20	1.000449	100	1	100	20.67	0	0	0	0	0	0	0	0	0	0	0;
	7577	17.6	0.07	20	-20	0.977562	100	1	100	11.73	0	0	0	0	0	0	0	0	0	0	0;
	7585	31.3	-0.89	20	-20	0.998951	100	1	100	20.87	0	0	0	0	0	0	0	0	0	0	0;
	7593	172.2	56.31	100	-30	0.999494	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	7598	77.03	9.29	60	-20	1.039863	100	1	200	51.35	0	0	0	0	0	0	0	0	0	0	0;
	7599	92.1	33.68	60	-20	1.056697	100	1	200	61.4	0	0	0	0	0	0	0	0	0	0	0;
	7600	50.34	9.17	40	-10	1.034532	100	1	200	33.56	0	0	0	0	0	0	0	0	0	0	0;
	7613	20.21	8.41	30	-10	1.012556	100	1	100	13.47	0	0	0	0	0	0	0	0	0	0	0;
	7625	35.74	6.84	30	-10	1.040138	100	1	100	23.83	0	0	0	0	0	0	0	0	0	0	0;
	7628	87.95	47.73	60	-20	1.052257	100	1	200	58.63	0	0	0	0	0	0	0	0	0	0	0;
	7634	25.22	5.17	30	-10	1.04616	100	1	100	16.81	0	0	0	0	0	0	0	0	0	0	0;
	7636	32.57	6.62	30	-10	1.059142	100	1	100	21.71	0	0	0	0	0	0	0	0	0	0	0;
	7637	32.57	6.62	30	-10	1.059142	100	1	100	21.71	0	0	0	0	0	0	0	0	0	0	0;
	7639	7.32	2.26	10	-10	1.005118	100	1	100	4.88	0	0	0	0	0	0	0	0	0	0	0;
	7640	-1344.4	409.68	100000	-100000	1.020355	100	1	100	-1792.53	0	0	0	0	0	0	0	0	0	0	0;
	7641	36.8	-2.54	30	-30	0.98379	100	1	100	24.53	0	0	0	0	0	0	0	0	0	0	0;
	7643	138.8	-2.57	110	-30	0.989847	100	1	300	92.53	0	0	0	0	0	0	0	0	0	0	0;
	7645	279.66	54.1	210	-50	1.030879	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	7648	279.66	54.1	210	-50	1.030879	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	7649	279.66	54.1	210	-50	1.030879	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	7654	204.59	110.18	210	-50	1.065374	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	7662	240.6	61.33	100	-30	0.990464	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	7667	100.41	11.92	60	-20	1.038747	100	1	300	66.94	0	0	0	0	0	0	0	0	0	0	0;
	7668	31.3	-0.77	30	-10	1.018739	100	1	100	20.87	0	0	0	0	0	0	0	0	0	0	0;
	7670	81.04	47.11	60	-20	1.043163	100	1	200	54.03	0	0	0	0	0	0	0	0	0	0	0;
	7671	81.04	47.11	60	-20	1.043163	100	1	200	54.03	0	0	0	0	0	0	0	0	0	0	0;
	7672	81.04	47.11	60	-20	1.043163	100	1	200	54.03	0	0	0	0	0	0	0	0	0	0	0;
	7674	293.92	111.78	210	-50	1.039038	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	7675	72.73	50.32	60	-20	1.038621	100	1	200	48.49	0	0	0	0	0	0	0	0	0	0	0;
	7676	72.73	50.32	60	-20	1.038621	100	1	200	48.49	0	0	0	0	0	0	0	0	0	0	0;
	7677	21.5	5.52	10	0	1.012381	100	1	100	14.33	0	0	0	0	0	0	0	0	0	0	0;
	7681	167.4	-8.63	90	-70	0.948759	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	7689	37.29	8.19	30	-10	1.030566	100	1	100	24.86	0	0	0	0	0	0	0	0	0	0	0;
	7700	23.9	3.61	30	-10	1.007844	100	1	100	15.93	0	0	0	0	0	0	0	0	0	0	0;
	7705	27.37	1.37	30	-20	1.005763	100	1	100	18.25	0	0	0	0	0	0	0	0	0	0	0;
	7707	51.5	-7.54	30	-30	0.954584	100	1	200	34.33	0	0	0	0	0	0	0	0	0	0	0;
	7712	20.79	7.32	30	-10	1.033848	100	1	100	13.86	0	0	0	0	0	0	0	0	0	0	0;
	7716	79.8	19.8	60	-20	1.02806	100	1	200	53.2	0	0	0	0	0	0	0	0	0	0	0;
	7717	79.8	19.8	60	-20	1.02806	100	1	200	53.2	0	0	0	0	0	0	0	0	0	0	0;
	7719	79.8	19.8	60	-20	1.02806	100	1	200	53.2	0	0	0	0	0	0	0	0	0	0	0;
	7721	79.8	19.8	60	-20	1.02806	100	1	200	53.2	0	0	0	0	0	0	0	0	0	0	0;
	7722	79.8	19.8	60	-20	1.02806	100	1	200	53.2	0	0	0	0	0	0	0	0	0	0	0;
	7731	0.83	-0.07	10	-10	1.015733	100	1	100	0.55	0	0	0	0	0	0	0	0	0	0	0;
	7732	19.93	2.39	30	-10	1.050328	100	1	100	13.29	0	0	0	0	0	0	0	0	0	0	0;
	7734	31.6	11.6	30	-10	1.04043	100	1	100	21.07	0	0	0	0	0	0	0	0	0	0	0;
	7736	362.93	192.92	210	-50	1.064843	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	7740	35.5	10.89	30	-10	1.045213	100	1	100	23.67	0	0	0	0	0	0	0	0	0	0	0;
	7742	30	5.12	10	0	1.022262	100	1	100	20	0	0	0	0	0	0	0	0	0	0	0;
	7743	159.8	33.64	90	-40	0.993125	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	7750	159.8	20.91	100	-40	1.008982	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	7751	135	10.29	70	-20	1.008348	100	1	300	90	0	0	0	0	0	0	0	0	0	0	0;
	7754	88.85	23.49	60	-20	1.032344	100	1	200	59.23	0	0	0	0	0	0	0	0	0	0	0;
	7768	9	0.68	10	-10	1.011648	100	1	100	6	0	0	0	0	0	0	0	0	0	0	0;
	7776	617.47	92.14	520	-120	1.032321	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	7777	17.5	5.36	20	-10	1.009853	100	1	100	11.67	0	0	0	0	0	0	0	0	0	0	0;
	7784	95.02	8.09	60	-20	1.023144	100	1	200	63.35	0	0	0	0	0	0	0	0	0	0	0;
	7808	224.46	15.93	210	-50	1.031165	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	7812	25.75	4.04	30	-10	1.026025	100	1	100	17.17	0	0	0	0	0	0	0	0	0	0	0;
	7814	212.35	111.47	160	-40	1.064355	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	7825	20.25	1.86	20	-10	1.027844	100	1	100	13.5	0	0	0	0	0	0	0	0	0	0	0;
	7833	262.9	48.53	160	-40	1.027461	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	7841	25.27	3.21	30	-10	1.006944	100	1	100	16.85	0	0	0	0	0	0	0	0	0	0	0;
	7842	33.61	7.79	30	-10	1.005097	100	1	100	22.41	0	0	0	0	0	0	0	0	0	0	0;
	7844	46.38	17.23	40	-10	1.048358	100	1	100	30.92	0	0	0	0	0	0	0	0	0	0	0;
	7846	46.38	17.23	40	-10	1.048358	100	1	100	30.92	0	0	0	0	0	0	0	0	0	0	0;
	7847	40.23	5.24	30	-10	1.006137	100	1	100	26.82	0	0	0	0	0	0	0	0	0	0	0;
	7848	33.61	7.79	30	-10	1.005097	100	1	100	22.41	0	0	0	0	0	0	0	0	0	0	0;
	7849	269.64	16.27	210	-50	1.041549	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	7851	33.61	7.79	30	-10	1.005097	100	1	100	22.41	0	0	0	0	0	0	0	0	0	0	0;
	7857	8.4	0.33	10	-10	1.014368	100	1	100	5.6	0	0	0	0	0	0	0	0	0	0	0;
	7858	36.91	13.33	30	-10	1.037973	100	1	100	24.61	0	0	0	0	0	0	0	0	0	0	0;
	7859	24.84	4.52	30	-10	1.022099	100	1	100	16.56	0	0	0	0	0	0	0	0	0	0	0;
	7875	38.16	4.12	30	-10	1.055337	100	1	100	25.44	0	0	0	0	0	0	0	0	0	0	0;
	7877	6.52	0.05	10	-10	0.996117	100	1	100	4.35	0	0	0	0	0	0	0	0	0	0	0;
	7890	36.11	3.87	30	-10	1.024544	100	1	100	24.07	0	0	0	0	0	0	0	0	0	0	0;
	7899	31.98	14.6	30	-10	1.034118	100	1	100	21.32	0	0	0	0	0	0	0	0	0	0	0;
	7901	39.83	14.7	30	-10	1.037128	100	1	100	26.55	0	0	0	0	0	0	0	0	0	0	0;
	7902	39.83	14.7	30	-10	1.037128	100	1	100	26.55	0	0	0	0	0	0	0	0	0	0	0;
	7903	26.6	10.01	30	-10	1.016718	100	1	100	17.73	0	0	0	0	0	0	0	0	0	0	0;
	7904	26.6	10.01	30	-10	1.016718	100	1	100	17.73	0	0	0	0	0	0	0	0	0	0	0;
	7905	24.89	5.96	30	-10	1.042846	100	1	100	16.59	0	0	0	0	0	0	0	0	0	0	0;
	7908	26.6	10.01	30	-10	1.016718	100	1	100	17.73	0	0	0	0	0	0	0	0	0	0	0;
	7918	6.52	-0.06	10	-10	0.994601	100	1	100	4.35	0	0	0	0	0	0	0	0	0	0	0;
	7921	14	3.19	20	-10	1.023478	100	1	100	9.33	0	0	0	0	0	0	0	0	0	0	0;
	7925	266.84	38.78	210	-50	1.032265	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	7929	321.29	105.87	210	-50	1.034412	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	7930	366.21	205.07	210	-50	1.129641	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	7931	27.81	6.76	30	-10	1.026444	100	1	100	18.54	0	0	0	0	0	0	0	0	0	0	0;
	7932	31.89	4.43	30	-10	1.028453	100	1	100	21.26	0	0	0	0	0	0	0	0	0	0	0;
	7933	33.19	5.48	30	-10	0.991806	100	1	100	22.13	0	0	0	0	0	0	0	0	0	0	0;
	7939	28.83	4.62	30	-10	1.019144	100	1	100	19.22	0	0	0	0	0	0	0	0	0	0	0;
	7943	31.89	4.43	30	-10	1.028453	100	1	100	21.26	0	0	0	0	0	0	0	0	0	0	0;
	7946	27.67	4.45	30	-10	1.027132	100	1	100	18.45	0	0	0	0	0	0	0	0	0	0	0;
	7949	30	2.24	30	-20	1.02542	100	1	100	20	0	0	0	0	0	0	0	0	0	0	0;
	7951	39.96	4.13	30	-10	0.987171	100	1	100	26.64	0	0	0	0	0	0	0	0	0	0	0;
	7962	3.02	0.33	10	-10	1.03013	100	1	100	2.02	0	0	0	0	0	0	0	0	0	0	0;
	7965	87.88	32.08	60	-20	1.042865	100	1	200	58.59	0	0	0	0	0	0	0	0	0	0	0;
	7966	87.88	32.08	60	-20	1.042865	100	1	200	58.59	0	0	0	0	0	0	0	0	0	0	0;
	7967	45.28	10.84	40	-10	1.043738	100	1	100	30.19	0	0	0	0	0	0	0	0	0	0	0;
	7970	80.4	12.89	60	-20	1.045647	100	1	200	53.6	0	0	0	0	0	0	0	0	0	0	0;
	7976	30.24	9.06	30	-10	1.012031	100	1	100	20.16	0	0	0	0	0	0	0	0	0	0	0;
	7979	53.72	11.91	40	-10	1.035023	100	1	200	35.81	0	0	0	0	0	0	0	0	0	0	0;
	7981	53.72	11.91	40	-10	1.035023	100	1	200	35.81	0	0	0	0	0	0	0	0	0	0	0;
	7988	53.72	11.91	40	-10	1.035023	100	1	200	35.81	0	0	0	0	0	0	0	0	0	0	0;
	7989	53.72	11.91	40	-10	1.035023	100	1	200	35.81	0	0	0	0	0	0	0	0	0	0	0;
	7992	34.6	7.46	30	-10	1.041485	100	1	100	23.07	0	0	0	0	0	0	0	0	0	0	0;
	7993	34.6	7.46	30	-10	1.041485	100	1	100	23.07	0	0	0	0	0	0	0	0	0	0	0;
	7995	24.22	5.78	30	-10	1.029709	100	1	100	16.15	0	0	0	0	0	0	0	0	0	0	0;
	7998	3.4	-0.83	10	-10	0.9865	100	1	100	2.27	0	0	0	0	0	0	0	0	0	0	0;
	8000	24.22	5.78	30	-10	1.029709	100	1	100	16.15	0	0	0	0	0	0	0	0	0	0	0;
	8001	24.22	5.78	30	-10	1.029709	100	1	100	16.15	0	0	0	0	0	0	0	0	0	0	0;
	8010	18.5	0.5	10	-10	0.981441	100	1	100	12.33	0	0	0	0	0	0	0	0	0	0	0;
	8017	87.95	47.73	60	-20	1.052257	100	1	200	58.63	0	0	0	0	0	0	0	0	0	0	0;
	8021	87.95	47.73	60	-20	1.052257	100	1	200	58.63	0	0	0	0	0	0	0	0	0	0	0;
	8022	87.95	47.73	60	-20	1.052257	100	1	200	58.63	0	0	0	0	0	0	0	0	0	0	0;
	8023	43.6	4.09	30	-30	0.9595	100	1	100	29.07	0	0	0	0	0	0	0	0	0	0	0;
	8033	75.6	-2.95	50	-40	1.010732	100	1	200	50.4	0	0	0	0	0	0	0	0	0	0	0;
	8039	18.13	3.85	30	-10	1.016716	100	1	100	12.09	0	0	0	0	0	0	0	0	0	0	0;
	8053	31.89	4.43	30	-10	1.028453	100	1	100	21.26	0	0	0	0	0	0	0	0	0	0	0;
	8055	31.89	4.43	30	-10	1.028453	100	1	100	21.26	0	0	0	0	0	0	0	0	0	0	0;
	8056	31.89	4.43	30	-10	1.028453	100	1	100	21.26	0	0	0	0	0	0	0	0	0	0	0;
	8060	45	-3.79	30	-30	0.939142	100	1	100	30	0	0	0	0	0	0	0	0	0	0	0;
	8063	36.42	7.96	30	-10	1.029166	100	1	100	24.28	0	0	0	0	0	0	0	0	0	0	0;
	8067	37.08	5.51	30	-10	1.021005	100	1	100	24.72	0	0	0	0	0	0	0	0	0	0	0;
	8088	201.01	93.92	160	-40	1.055956	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	8089	28.04	8.11	30	-10	1.017878	100	1	100	18.69	0	0	0	0	0	0	0	0	0	0	0;
	8090	201.01	93.92	160	-40	1.055956	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	8091	64.18	-6.8	40	-40	0.951565	100	1	200	42.79	0	0	0	0	0	0	0	0	0	0	0;
	8092	36.13	4.17	30	-10	1.027348	100	1	100	24.09	0	0	0	0	0	0	0	0	0	0	0;
	8096	30.9	1.92	30	-10	1.015954	100	1	100	20.6	0	0	0	0	0	0	0	0	0	0	0;
	8097	369.99	86.96	210	-50	1.043138	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	8099	41.8	9.39	40	-10	1.028604	100	1	100	27.87	0	0	0	0	0	0	0	0	0	0	0;
	8109	24.99	2.31	30	-10	1.010466	100	1	100	16.66	0	0	0	0	0	0	0	0	0	0	0;
	8110	34.69	8.62	30	-10	1.055415	100	1	100	23.13	0	0	0	0	0	0	0	0	0	0	0;
	8111	36.88	11.22	30	-10	1.049781	100	1	100	24.59	0	0	0	0	0	0	0	0	0	0	0;
	8112	34.69	8.62	30	-10	1.055415	100	1	100	23.13	0	0	0	0	0	0	0	0	0	0	0;
	8113	37.48	8.43	30	-10	1.051968	100	1	100	24.99	0	0	0	0	0	0	0	0	0	0	0;
	8114	36.88	11.22	30	-10	1.049781	100	1	100	24.59	0	0	0	0	0	0	0	0	0	0	0;
	8116	37.48	8.43	30	-10	1.051968	100	1	100	24.99	0	0	0	0	0	0	0	0	0	0	0;
	8121	23	-6.16	20	-20	0.989111	100	1	100	15.33	0	0	0	0	0	0	0	0	0	0	0;
	8122	2.5	0.07	10	-10	0.984601	100	1	100	1.67	0	0	0	0	0	0	0	0	0	0	0;
	8123	81.96	34.23	60	-20	1.031098	100	1	200	54.64	0	0	0	0	0	0	0	0	0	0	0;
	8125	88.86	16.51	60	-20	1.020649	100	1	200	59.24	0	0	0	0	0	0	0	0	0	0	0;
	8126	81.96	34.23	60	-20	1.031098	100	1	200	54.64	0	0	0	0	0	0	0	0	0	0	0;
	8127	88.86	16.51	60	-20	1.020649	100	1	200	59.24	0	0	0	0	0	0	0	0	0	0	0;
	8128	88.86	16.51	60	-20	1.020649	100	1	200	59.24	0	0	0	0	0	0	0	0	0	0	0;
	8130	33.68	6.27	30	-10	1.046665	100	1	100	22.45	0	0	0	0	0	0	0	0	0	0	0;
	8141	16.1	3.37	20	-10	1.026372	100	1	100	10.73	0	0	0	0	0	0	0	0	0	0	0;
	8143	85.28	21.31	60	-20	1.029542	100	1	200	56.85	0	0	0	0	0	0	0	0	0	0	0;
	8145	97.82	33.17	60	-20	1.055063	100	1	200	65.21	0	0	0	0	0	0	0	0	0	0	0;
	8146	97.52	42.26	60	-20	1.053776	100	1	200	65.01	0	0	0	0	0	0	0	0	0	0	0;
	8147	97.74	41.58	60	-20	1.050978	100	1	200	65.16	0	0	0	0	0	0	0	0	0	0	0;
	8148	97.74	41.58	60	-20	1.050978	100	1	200	65.16	0	0	0	0	0	0	0	0	0	0	0;
	8149	97.52	42.26	60	-20	1.053776	100	1	200	65.01	0	0	0	0	0	0	0	0	0	0	0;
	8151	97.77	42.1	60	-20	1.051219	100	1	200	65.18	0	0	0	0	0	0	0	0	0	0	0;
	8153	97.74	41.58	60	-20	1.050978	100	1	200	65.16	0	0	0	0	0	0	0	0	0	0	0;
	8154	97.77	42.1	60	-20	1.051219	100	1	200	65.18	0	0	0	0	0	0	0	0	0	0	0;
	8157	23.04	1.33	20	-20	0.996935	100	1	100	15.36	0	0	0	0	0	0	0	0	0	0	0;
	8159	35.74	6.84	30	-10	1.040138	100	1	100	23.83	0	0	0	0	0	0	0	0	0	0	0;
	8162	87.39	49.11	60	-20	1.054462	100	1	200	58.26	0	0	0	0	0	0	0	0	0	0	0;
	8163	15	-0.86	10	-10	0.992395	100	1	100	10	0	0	0	0	0	0	0	0	0	0	0;
	8167	274.92	131.33	210	-50	1.086148	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	8173	87.39	49.11	60	-20	1.054462	100	1	200	58.26	0	0	0	0	0	0	0	0	0	0	0;
	8179	96.96	-6.25	50	-50	0.997286	100	1	200	64.64	0	0	0	0	0	0	0	0	0	0	0;
	8191	34.86	4.81	30	-10	1.050901	100	1	100	23.24	0	0	0	0	0	0	0	0	0	0	0;
	8193	6.51	0.2	10	-10	1.013271	100	1	100	4.34	0	0	0	0	0	0	0	0	0	0	0;
	8202	60	-6.94	30	-30	0.959478	100	1	200	40	0	0	0	0	0	0	0	0	0	0	0;
	8205	28.35	17.5	30	-10	1.042184	100	1	100	18.9	0	0	0	0	0	0	0	0	0	0	0;
	8206	26.33	10.66	30	-10	1.01282	100	1	100	17.55	0	0	0	0	0	0	0	0	0	0	0;
	8218	33.1	8.3	30	-10	1.041811	100	1	100	22.07	0	0	0	0	0	0	0	0	0	0	0;
	8219	33.1	8.3	30	-10	1.041811	100	1	100	22.07	0	0	0	0	0	0	0	0	0	0	0;
	8220	33.1	8.3	30	-10	1.041811	100	1	100	22.07	0	0	0	0	0	0	0	0	0	0	0;
	8221	33.1	8.3	30	-10	1.041811	100	1	100	22.07	0	0	0	0	0	0	0	0	0	0	0;
	8222	33.1	8.3	30	-10	1.041811	100	1	100	22.07	0	0	0	0	0	0	0	0	0	0	0;
	8223	33.1	8.3	30	-10	1.041811	100	1	100	22.07	0	0	0	0	0	0	0	0	0	0	0;
	8224	33.1	8.3	30	-10	1.041811	100	1	100	22.07	0	0	0	0	0	0	0	0	0	0	0;
	8246	0.26	-0.09	10	-10	1.024094	100	1	100	0.17	0	0	0	0	0	0	0	0	0	0	0;
	8253	20	-2.63	30	-10	1.010291	100	1	100	13.33	0	0	0	0	0	0	0	0	0	0	0;
	8274	61.36	4.64	40	-10	1.014321	100	1	200	40.91	0	0	0	0	0	0	0	0	0	0	0;
	8275	61.36	4.64	40	-10	1.014321	100	1	200	40.91	0	0	0	0	0	0	0	0	0	0	0;
	8276	61.36	4.64	40	-10	1.014321	100	1	200	40.91	0	0	0	0	0	0	0	0	0	0	0;
	8290	100	10.54	90	-50	1.01208	100	1	300	66.67	0	0	0	0	0	0	0	0	0	0	0;
	8299	10.5	-0.76	10	-10	1.011227	100	1	100	7	0	0	0	0	0	0	0	0	0	0	0;
	8301	65.58	12.54	60	-20	1.029941	100	1	200	43.72	0	0	0	0	0	0	0	0	0	0	0;
	8308	88.86	16.51	60	-20	1.020649	100	1	200	59.24	0	0	0	0	0	0	0	0	0	0	0;
	8317	45.63	24.24	40	-10	1.060899	100	1	100	30.42	0	0	0	0	0	0	0	0	0	0	0;
	8319	45.63	24.24	40	-10	1.060899	100	1	100	30.42	0	0	0	0	0	0	0	0	0	0	0;
	8320	45.63	24.24	40	-10	1.060899	100	1	100	30.42	0	0	0	0	0	0	0	0	0	0	0;
	8321	-42.07	-6.87	40	-70	0.998025	100	1	100	-56.09	0	0	0	0	0	0	0	0	0	0	0;
	8323	25.6	3.9	30	-10	1.03284	100	1	100	17.07	0	0	0	0	0	0	0	0	0	0	0;
	8326	27.2	3.82	30	-10	1.024762	100	1	100	18.13	0	0	0	0	0	0	0	0	0	0	0;
	8342	81.31	8.06	60	-20	1.043619	100	1	200	54.21	0	0	0	0	0	0	0	0	0	0	0;
	8343	58.19	6.99	40	-10	1.028736	100	1	200	38.79	0	0	0	0	0	0	0	0	0	0	0;
	8352	25	4.2	20	-10	1.03161	100	1	100	16.67	0	0	0	0	0	0	0	0	0	0	0;
	8354	58.19	6.99	40	-10	1.028736	100	1	200	38.79	0	0	0	0	0	0	0	0	0	0	0;
	8357	58.19	6.99	40	-10	1.028736	100	1	200	38.79	0	0	0	0	0	0	0	0	0	0	0;
	8359	6.97	0.62	10	-10	1.028038	100	1	100	4.65	0	0	0	0	0	0	0	0	0	0	0;
	8367	36.11	3.87	30	-10	1.024544	100	1	100	24.07	0	0	0	0	0	0	0	0	0	0	0;
	8369	33.32	9.67	30	-10	1.032797	100	1	100	22.21	0	0	0	0	0	0	0	0	0	0	0;
	8370	35.74	6.84	30	-10	1.040138	100	1	100	23.83	0	0	0	0	0	0	0	0	0	0	0;
	8373	36	6.45	30	-10	1.034215	100	1	100	24	0	0	0	0	0	0	0	0	0	0	0;
	8378	36	6.45	30	-10	1.034215	100	1	100	24	0	0	0	0	0	0	0	0	0	0	0;
	8380	33.32	9.67	30	-10	1.032797	100	1	100	22.21	0	0	0	0	0	0	0	0	0	0	0;
	8385	30	5.04	10	0	1.023034	100	1	100	20	0	0	0	0	0	0	0	0	0	0	0;
	8386	24.22	5.78	30	-10	1.029709	100	1	100	16.15	0	0	0	0	0	0	0	0	0	0	0;
	8387	84.68	7.94	60	-20	1.00657	100	1	200	56.45	0	0	0	0	0	0	0	0	0	0	0;
	8388	261.4	61.75	100	-30	0.988817	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	8389	92.69	14.48	60	-20	1.026207	100	1	200	61.79	0	0	0	0	0	0	0	0	0	0	0;
	8393	15.4	0.45	20	-10	1.066894	100	1	100	10.27	0	0	0	0	0	0	0	0	0	0	0;
	8394	86.25	-3.04	60	-80	0.962614	100	1	200	57.5	0	0	0	0	0	0	0	0	0	0	0;
	8396	4.28	0.36	10	-10	1.020502	100	1	100	2.85	0	0	0	0	0	0	0	0	0	0	0;
	8397	33.53	16.02	30	-10	1.021376	100	1	100	22.35	0	0	0	0	0	0	0	0	0	0	0;
	8402	33.53	16.02	30	-10	1.021376	100	1	100	22.35	0	0	0	0	0	0	0	0	0	0	0;
	8409	24.63	12.17	30	-10	1.026822	100	1	100	16.42	0	0	0	0	0	0	0	0	0	0	0;
	8413	40	-0.67	20	-20	1.00358	100	1	100	26.67	0	0	0	0	0	0	0	0	0	0	0;
	8415	21.45	5.94	30	-10	1.013179	100	1	100	14.3	0	0	0	0	0	0	0	0	0	0	0;
	8421	25.41	10.96	30	-10	1.019405	100	1	100	16.94	0	0	0	0	0	0	0	0	0	0	0;
	8423	257.89	39.7	210	-50	1.022087	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	8424	257.89	39.7	210	-50	1.022087	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	8425	10.36	-1.27	10	-10	1.004419	100	1	100	6.91	0	0	0	0	0	0	0	0	0	0	0;
	8433	55.8	16.15	40	-30	1.022285	100	1	200	37.2	0	0	0	0	0	0	0	0	0	0	0;
	8436	43.5	1.21	30	-30	1.007757	100	1	100	29	0	0	0	0	0	0	0	0	0	0	0;
	8439	91.17	37.8	60	-20	1.045666	100	1	200	60.78	0	0	0	0	0	0	0	0	0	0	0;
	8441	15.64	-1.08	10	-10	1.023837	100	1	100	10.43	0	0	0	0	0	0	0	0	0	0	0;
	8444	97.77	42.1	60	-20	1.051219	100	1	200	65.18	0	0	0	0	0	0	0	0	0	0	0;
	8445	24.72	4.34	30	-10	1.039546	100	1	100	16.48	0	0	0	0	0	0	0	0	0	0	0;
	8447	22.22	6.81	30	-10	1.03758	100	1	100	14.81	0	0	0	0	0	0	0	0	0	0	0;
	8451	40.75	3.68	60	-20	0.992069	100	1	100	27.17	0	0	0	0	0	0	0	0	0	0	0;
	8453	323.05	-7.03	120	-130	0.972158	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	8456	55.55	12.64	40	-10	1.024324	100	1	200	37.03	0	0	0	0	0	0	0	0	0	0	0;
	8457	25.65	3.95	30	-10	1.02611	100	1	100	17.1	0	0	0	0	0	0	0	0	0	0	0;
	8459	25.65	3.95	30	-10	1.02611	100	1	100	17.1	0	0	0	0	0	0	0	0	0	0	0;
	8470	66.31	16.47	60	-20	1.020956	100	1	200	44.21	0	0	0	0	0	0	0	0	0	0	0;
	8471	66.31	16.47	60	-20	1.020956	100	1	200	44.21	0	0	0	0	0	0	0	0	0	0	0;
	8473	66.31	16.47	60	-20	1.020956	100	1	200	44.21	0	0	0	0	0	0	0	0	0	0	0;
	8474	66.31	16.47	60	-20	1.020956	100	1	200	44.21	0	0	0	0	0	0	0	0	0	0	0;
	8476	8	0.73	10	-10	1.000621	100	1	100	5.33	0	0	0	0	0	0	0	0	0	0	0;
	8479	19.98	0.98	30	-10	1.029635	100	1	100	13.32	0	0	0	0	0	0	0	0	0	0	0;
	8483	300	-14.29	100	-100	1.001419	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	8488	11.25	5.25	10	0	1.042782	100	1	100	7.5	0	0	0	0	0	0	0	0	0	0	0;
	8490	34.31	1.99	30	-10	1.015531	100	1	100	22.87	0	0	0	0	0	0	0	0	0	0	0;
	8493	4	1.81	10	-10	1.024481	100	1	100	2.67	0	0	0	0	0	0	0	0	0	0	0;
	8499	165	33.72	100	-20	0.962508	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	8519	92.69	14.48	60	-20	1.026207	100	1	200	61.79	0	0	0	0	0	0	0	0	0	0	0;
	8524	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	8530	19.4	2.93	20	-10	1.024655	100	1	100	12.93	0	0	0	0	0	0	0	0	0	0	0;
	8537	74.36	21.23	60	-20	1.032132	100	1	200	49.57	0	0	0	0	0	0	0	0	0	0	0;
	8538	68.3	12.27	60	-20	1.015227	100	1	200	45.53	0	0	0	0	0	0	0	0	0	0	0;
	8544	63.93	6.28	60	-20	1.012697	100	1	200	42.62	0	0	0	0	0	0	0	0	0	0	0;
	8545	69.46	6.43	60	-20	1.012641	100	1	200	46.31	0	0	0	0	0	0	0	0	0	0	0;
	8546	69.46	6.43	60	-20	1.012641	100	1	200	46.31	0	0	0	0	0	0	0	0	0	0	0;
	8547	68.58	4.97	60	-20	1.028244	100	1	200	45.72	0	0	0	0	0	0	0	0	0	0	0;
	8548	68.58	4.97	60	-20	1.028244	100	1	200	45.72	0	0	0	0	0	0	0	0	0	0	0;
	8549	68.02	5.04	60	-20	1.028456	100	1	200	45.35	0	0	0	0	0	0	0	0	0	0	0;
	8550	68.02	5.04	60	-20	1.028456	100	1	200	45.35	0	0	0	0	0	0	0	0	0	0	0;
	8552	81.86	7.76	60	-20	1.050589	100	1	200	54.57	0	0	0	0	0	0	0	0	0	0	0;
	8555	31	0.14	20	-20	1.000449	100	1	100	20.67	0	0	0	0	0	0	0	0	0	0	0;
	8559	57.5	6.09	30	-30	0.99654	100	1	200	38.33	0	0	0	0	0	0	0	0	0	0	0;
	8561	172.2	36.24	100	-30	0.992794	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	8563	18.62	7.28	30	-10	1.020659	100	1	100	12.41	0	0	0	0	0	0	0	0	0	0	0;
	8569	863.32	247.01	520	-120	1.043912	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	8570	863.33	248.79	520	-120	1.044271	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	8571	863.3	246.84	520	-120	1.043938	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	8573	26.62	13.25	30	-10	1.040642	100	1	100	17.75	0	0	0	0	0	0	0	0	0	0	0;
	8576	31.12	9.56	30	-10	1.04434	100	1	100	20.75	0	0	0	0	0	0	0	0	0	0	0;
	8577	21.89	8.72	30	-10	1.039495	100	1	100	14.59	0	0	0	0	0	0	0	0	0	0	0;
	8578	21.89	8.72	30	-10	1.039495	100	1	100	14.59	0	0	0	0	0	0	0	0	0	0	0;
	8581	29.05	10.97	30	-10	1.053599	100	1	100	19.37	0	0	0	0	0	0	0	0	0	0	0;
	8582	29.05	10.97	30	-10	1.053599	100	1	100	19.37	0	0	0	0	0	0	0	0	0	0	0;
	8595	5.8	-0.78	10	-10	1.003942	100	1	100	3.87	0	0	0	0	0	0	0	0	0	0	0;
	8598	86.41	23.5	60	-20	1.031898	100	1	200	57.61	0	0	0	0	0	0	0	0	0	0	0;
	8604	97.67	41.19	60	-20	1.052591	100	1	200	65.11	0	0	0	0	0	0	0	0	0	0	0;
	8609	97.67	41.19	60	-20	1.052591	100	1	200	65.11	0	0	0	0	0	0	0	0	0	0	0;
	8610	97.67	41.19	60	-20	1.052591	100	1	200	65.11	0	0	0	0	0	0	0	0	0	0	0;
	8632	353.02	203.37	210	-50	1.128032	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	8635	172.2	6.25	100	-30	0.968831	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	8637	90.48	33.58	60	-20	1.043665	100	1	200	60.32	0	0	0	0	0	0	0	0	0	0	0;
	8639	53.01	8.19	40	-10	1.026319	100	1	200	35.34	0	0	0	0	0	0	0	0	0	0	0;
	8644	94.5	8.92	50	-40	0.979015	100	1	200	63	0	0	0	0	0	0	0	0	0	0	0;
	8647	276.46	124.36	160	-40	1.052654	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	8648	276.46	124.36	160	-40	1.052654	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	8651	23.81	8.18	30	-10	1.024067	100	1	100	15.87	0	0	0	0	0	0	0	0	0	0	0;
	8654	2.5	0.07	10	-10	0.984601	100	1	100	1.67	0	0	0	0	0	0	0	0	0	0	0;
	8655	72.47	11.44	60	-20	1.019231	100	1	200	48.31	0	0	0	0	0	0	0	0	0	0	0;
	8656	12	-0.74	10	-10	1.012276	100	1	100	8	0	0	0	0	0	0	0	0	0	0	0;
	8658	95.38	41.05	60	-20	1.061235	100	1	200	63.59	0	0	0	0	0	0	0	0	0	0	0;
	8665	95.38	41.05	60	-20	1.061235	100	1	200	63.59	0	0	0	0	0	0	0	0	0	0	0;
	8668	28.31	1.9	30	-10	1.060483	100	1	100	18.87	0	0	0	0	0	0	0	0	0	0	0;
	8671	28.31	1.9	30	-10	1.060483	100	1	100	18.87	0	0	0	0	0	0	0	0	0	0	0;
	8675	31.55	5.32	30	-10	1.031719	100	1	100	21.03	0	0	0	0	0	0	0	0	0	0	0;
	8679	31.55	5.32	30	-10	1.031719	100	1	100	21.03	0	0	0	0	0	0	0	0	0	0	0;
	8680	31.55	5.32	30	-10	1.031719	100	1	100	21.03	0	0	0	0	0	0	0	0	0	0	0;
	8686	31.92	6.07	30	-10	1.027264	100	1	100	21.28	0	0	0	0	0	0	0	0	0	0	0;
	8687	74.35	10.28	60	-20	0.992442	100	1	200	49.57	0	0	0	0	0	0	0	0	0	0	0;
	8701	78.71	19.55	60	-20	1.033214	100	1	200	52.47	0	0	0	0	0	0	0	0	0	0	0;
	8702	78.71	19.55	60	-20	1.033214	100	1	200	52.47	0	0	0	0	0	0	0	0	0	0	0;
	8704	35.02	1.75	30	-10	1.038438	100	1	100	23.35	0	0	0	0	0	0	0	0	0	0	0;
	8706	78.71	19.51	60	-20	1.033214	100	1	200	52.47	0	0	0	0	0	0	0	0	0	0	0;
	8707	78.71	19.51	60	-20	1.033214	100	1	200	52.47	0	0	0	0	0	0	0	0	0	0	0;
	8709	23.8	-3.12	20	-20	0.982549	100	1	100	15.87	0	0	0	0	0	0	0	0	0	0	0;
	8711	626.78	218.85	520	-120	1.044461	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	8712	626.78	218.85	520	-120	1.044461	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	8713	626.79	221.4	520	-120	1.044893	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	8714	626.79	221.4	520	-120	1.044893	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	8716	259.51	14.89	160	-40	1.019095	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	8717	259.51	14.89	160	-40	1.019095	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	8719	29.84	3.47	30	-10	1.022039	100	1	100	19.89	0	0	0	0	0	0	0	0	0	0	0;
	8725	23.08	3.01	20	-10	1.027819	100	1	100	15.39	0	0	0	0	0	0	0	0	0	0	0;
	8726	15.84	6.64	30	-10	1.000784	100	1	100	10.56	0	0	0	0	0	0	0	0	0	0	0;
	8731	94.21	42.75	60	-20	1.0598	100	1	200	62.81	0	0	0	0	0	0	0	0	0	0	0;
	8732	94.21	42.75	60	-20	1.0598	100	1	200	62.81	0	0	0	0	0	0	0	0	0	0	0;
	8733	94.21	42.75	60	-20	1.0598	100	1	200	62.81	0	0	0	0	0	0	0	0	0	0	0;
	8734	94.21	42.75	60	-20	1.0598	100	1	200	62.81	0	0	0	0	0	0	0	0	0	0	0;
	8735	94.21	42.75	60	-20	1.0598	100	1	200	62.81	0	0	0	0	0	0	0	0	0	0	0;
	8740	260.74	7.53	160	-40	0.99389	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	8741	274.28	12.54	160	-40	0.988223	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	8744	50.29	10.48	40	-10	1.031165	100	1	200	33.53	0	0	0	0	0	0	0	0	0	0	0;
	8745	50.29	10.48	40	-10	1.031165	100	1	200	33.53	0	0	0	0	0	0	0	0	0	0	0;
	8746	50.29	10.48	40	-10	1.031165	100	1	200	33.53	0	0	0	0	0	0	0	0	0	0	0;
	8747	50.29	10.48	40	-10	1.031165	100	1	200	33.53	0	0	0	0	0	0	0	0	0	0	0;
	8753	74.78	41.98	60	-20	1.04914	100	1	200	49.85	0	0	0	0	0	0	0	0	0	0	0;
	8755	50.29	10.48	40	-10	1.031165	100	1	200	33.53	0	0	0	0	0	0	0	0	0	0	0;
	8759	239.85	38.05	160	-40	1.032762	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	8763	31.28	-0.92	30	-10	1.009892	100	1	100	20.85	0	0	0	0	0	0	0	0	0	0	0;
	8764	3.41	0.22	10	-10	1.030035	100	1	100	2.28	0	0	0	0	0	0	0	0	0	0	0;
	8770	219.68	138.48	160	-40	1.048632	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	8772	55.14	27.56	40	-10	1.023158	100	1	200	36.76	0	0	0	0	0	0	0	0	0	0	0;
	8773	289.8	143.26	160	-40	1.042956	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	8774	56.25	11.12	40	-10	1.03036	100	1	200	37.5	0	0	0	0	0	0	0	0	0	0	0;
	8776	40	-3.06	20	-20	0.96764	100	1	100	26.67	0	0	0	0	0	0	0	0	0	0	0;
	8791	0	0.58	10	-10	1.011737	100	1	100	-0	0	0	0	0	0	0	0	0	0	0	0;
	8793	35.01	1.48	30	-10	1.032132	100	1	100	23.34	0	0	0	0	0	0	0	0	0	0	0;
	8797	27.2	3.82	30	-10	1.024762	100	1	100	18.13	0	0	0	0	0	0	0	0	0	0	0;
	8803	382.46	57.23	210	-50	1.050841	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	8806	96.75	44.18	60	-20	1.057091	100	1	200	64.5	0	0	0	0	0	0	0	0	0	0	0;
	8807	96.75	44.18	60	-20	1.057091	100	1	200	64.5	0	0	0	0	0	0	0	0	0	0	0;
	8809	96.75	44.18	60	-20	1.057091	100	1	200	64.5	0	0	0	0	0	0	0	0	0	0	0;
	8810	35.95	6.39	40	-10	1.015185	100	1	100	23.97	0	0	0	0	0	0	0	0	0	0	0;
	8816	188.7	18.36	120	-80	1.036206	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	8820	34.92	9.74	30	-10	1.043492	100	1	100	23.28	0	0	0	0	0	0	0	0	0	0	0;
	8824	63.6	31.38	60	-20	1.027152	100	1	200	42.4	0	0	0	0	0	0	0	0	0	0	0;
	8830	152	33.4	90	-40	0.993492	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	8837	8.75	-0.1	10	-10	0.988686	100	1	100	5.83	0	0	0	0	0	0	0	0	0	0	0;
	8838	63.6	31.38	60	-20	1.027152	100	1	200	42.4	0	0	0	0	0	0	0	0	0	0	0;
	8839	81.35	28.89	60	-20	1.055981	100	1	200	54.23	0	0	0	0	0	0	0	0	0	0	0;
	8841	364.88	71.7	210	-50	1.04706	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	8842	364.88	71.7	210	-50	1.04706	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	8843	364.88	71.7	210	-50	1.04706	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	8844	364.88	71.7	210	-50	1.04706	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	8846	153	12.38	100	-40	1.017859	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	8851	35.5	10.89	30	-10	1.045213	100	1	100	23.67	0	0	0	0	0	0	0	0	0	0	0;
	8852	36.11	3.87	30	-10	1.024544	100	1	100	24.07	0	0	0	0	0	0	0	0	0	0	0;
	8857	7	0.9	10	-10	1.004314	100	1	100	4.67	0	0	0	0	0	0	0	0	0	0	0;
	8860	25.07	19.14	30	-10	1.045495	100	1	100	16.71	0	0	0	0	0	0	0	0	0	0	0;
	8861	25.07	19.14	30	-10	1.045495	100	1	100	16.71	0	0	0	0	0	0	0	0	0	0	0;
	8864	30.01	15.35	30	-10	1.052369	100	1	100	20.01	0	0	0	0	0	0	0	0	0	0	0;
	8866	30.01	15.35	30	-10	1.052369	100	1	100	20.01	0	0	0	0	0	0	0	0	0	0	0;
	8867	30.01	15.35	30	-10	1.052369	100	1	100	20.01	0	0	0	0	0	0	0	0	0	0	0;
	8873	52.06	9.48	40	-10	1.028771	100	1	200	34.71	0	0	0	0	0	0	0	0	0	0	0;
	8881	10.5	2.44	10	-10	1.008788	100	1	100	7	0	0	0	0	0	0	0	0	0	0	0;
	8884	84.68	7.94	60	-20	1.00657	100	1	200	56.45	0	0	0	0	0	0	0	0	0	0	0;
	8885	96.29	19	60	-20	1.037131	100	1	200	64.19	0	0	0	0	0	0	0	0	0	0	0;
	8887	40	-1.33	30	-30	1.010141	100	1	100	26.67	0	0	0	0	0	0	0	0	0	0	0;
	8893	6.52	-0.06	10	-10	0.994601	100	1	100	4.35	0	0	0	0	0	0	0	0	0	0	0;
	8894	274.65	30.64	160	-40	1.018066	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	8896	23.2	-0.2	20	-20	0.996866	100	1	100	15.47	0	0	0	0	0	0	0	0	0	0	0;
	8897	80.13	43.5	60	-20	1.053937	100	1	200	53.42	0	0	0	0	0	0	0	0	0	0	0;
	8898	80.07	50.15	60	-20	1.039942	100	1	200	53.38	0	0	0	0	0	0	0	0	0	0	0;
	8900	29.7	6.22	30	-10	1.033776	100	1	100	19.8	0	0	0	0	0	0	0	0	0	0	0;
	8901	24.3	3.82	30	-10	1.010028	100	1	100	16.2	0	0	0	0	0	0	0	0	0	0	0;
	8902	24.3	3.82	30	-10	1.010028	100	1	100	16.2	0	0	0	0	0	0	0	0	0	0	0;
	8903	24.3	3.82	30	-10	1.010028	100	1	100	16.2	0	0	0	0	0	0	0	0	0	0	0;
	8904	79.99	47.72	60	-20	1.050006	100	1	200	53.33	0	0	0	0	0	0	0	0	0	0	0;
	8905	79.99	47.72	60	-20	1.050006	100	1	200	53.33	0	0	0	0	0	0	0	0	0	0	0;
	8906	24.3	3.82	30	-10	1.010028	100	1	100	16.2	0	0	0	0	0	0	0	0	0	0	0;
	8907	79.99	47.72	60	-20	1.050006	100	1	200	53.33	0	0	0	0	0	0	0	0	0	0	0;
	8908	80.13	43.5	60	-20	1.053937	100	1	200	53.42	0	0	0	0	0	0	0	0	0	0	0;
	8909	80.13	43.5	60	-20	1.053937	100	1	200	53.42	0	0	0	0	0	0	0	0	0	0	0;
	8910	80.13	43.5	60	-20	1.053937	100	1	200	53.42	0	0	0	0	0	0	0	0	0	0	0;
	8911	321.41	105.18	210	-50	1.033947	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	8912	72.47	11.44	60	-20	1.019231	100	1	200	48.31	0	0	0	0	0	0	0	0	0	0	0;
	8914	22.73	8.07	30	-10	1.024089	100	1	100	15.15	0	0	0	0	0	0	0	0	0	0	0;
	8916	2.5	0.07	10	-10	0.984601	100	1	100	1.67	0	0	0	0	0	0	0	0	0	0	0;
	8921	243.32	23.39	210	-50	1.055881	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	8933	15	-0.04	20	-20	1.011801	100	1	100	10	0	0	0	0	0	0	0	0	0	0	0;
	8937	56.84	10.43	40	-10	1.025724	100	1	200	37.89	0	0	0	0	0	0	0	0	0	0	0;
	8938	56.84	10.43	40	-10	1.025724	100	1	200	37.89	0	0	0	0	0	0	0	0	0	0	0;
	8939	56.84	10.43	40	-10	1.025724	100	1	200	37.89	0	0	0	0	0	0	0	0	0	0	0;
	8941	51.75	8.57	40	-10	1.019281	100	1	200	34.5	0	0	0	0	0	0	0	0	0	0	0;
	8942	29.68	6.45	30	-10	1.016891	100	1	100	19.79	0	0	0	0	0	0	0	0	0	0	0;
	8943	51.75	8.57	40	-10	1.019281	100	1	200	34.5	0	0	0	0	0	0	0	0	0	0	0;
	8944	51.75	8.57	40	-10	1.019281	100	1	200	34.5	0	0	0	0	0	0	0	0	0	0	0;
	8945	57.27	17.52	40	-10	1.017845	100	1	200	38.18	0	0	0	0	0	0	0	0	0	0	0;
	8952	34.31	5.82	30	-10	1.037965	100	1	100	22.87	0	0	0	0	0	0	0	0	0	0	0;
	8953	34.31	5.82	30	-10	1.037965	100	1	100	22.87	0	0	0	0	0	0	0	0	0	0	0;
	8954	34.31	5.82	30	-10	1.037965	100	1	100	22.87	0	0	0	0	0	0	0	0	0	0	0;
	8955	34.31	5.82	30	-10	1.037965	100	1	100	22.87	0	0	0	0	0	0	0	0	0	0	0;
	8956	26.01	11.35	30	-10	1.062565	100	1	100	17.34	0	0	0	0	0	0	0	0	0	0	0;
	8958	769.34	-82.54	520	-120	0.999428	100	1	1600	500	0	0	0	0	0	0	0	0	0	0	0;
	8961	27.38	10.62	30	-10	1.015042	100	1	100	18.25	0	0	0	0	0	0	0	0	0	0	0;
	8962	34.31	5.82	30	-10	1.037965	100	1	100	22.87	0	0	0	0	0	0	0	0	0	0	0;
	8963	34.3	7.71	30	-10	1.038216	100	1	100	22.87	0	0	0	0	0	0	0	0	0	0	0;
	8964	77.39	15.14	60	-20	1.030057	100	1	200	51.59	0	0	0	0	0	0	0	0	0	0	0;
	8965	77.39	15.14	60	-20	1.030057	100	1	200	51.59	0	0	0	0	0	0	0	0	0	0	0;
	8966	77.39	15.14	60	-20	1.030057	100	1	200	51.59	0	0	0	0	0	0	0	0	0	0	0;
	8967	77.39	15.14	60	-20	1.030057	100	1	200	51.59	0	0	0	0	0	0	0	0	0	0	0;
	8969	16.71	0.01	30	-10	1.030988	100	1	100	11.14	0	0	0	0	0	0	0	0	0	0	0;
	8976	74.86	19.42	60	-20	1.011998	100	1	200	49.91	0	0	0	0	0	0	0	0	0	0	0;
	8980	21.6	0.33	10	-10	0.9829	100	1	100	14.4	0	0	0	0	0	0	0	0	0	0	0;
	8994	78.96	32.05	60	-20	1.063361	100	1	200	52.64	0	0	0	0	0	0	0	0	0	0	0;
	8999	22.68	3.66	20	-10	1.020283	100	1	100	15.12	0	0	0	0	0	0	0	0	0	0	0;
	9016	40	-3.06	20	-20	0.96764	100	1	100	26.67	0	0	0	0	0	0	0	0	0	0	0;
	9024	204	85.48	150	-60	1.02301	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	9028	18.52	11.72	30	-10	1.044704	100	1	100	12.35	0	0	0	0	0	0	0	0	0	0	0;
	9033	332.24	91.88	210	-50	1.046254	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	9034	332.24	91.88	210	-50	1.046254	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	9035	66.65	5.48	60	-20	1.01993	100	1	200	44.43	0	0	0	0	0	0	0	0	0	0	0;
	9041	36.03	16.08	30	-10	1.045359	100	1	100	24.02	0	0	0	0	0	0	0	0	0	0	0;
	9045	284.4	98.98	210	-50	1.060659	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	9047	48.3	8.12	40	-10	1.021906	100	1	100	32.2	0	0	0	0	0	0	0	0	0	0	0;
	9049	158.9	5.88	100	-30	0.968879	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	9051	284.51	138.58	210	-50	1.065482	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	9054	159.8	34.27	100	-40	0.978477	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	9056	1.16	0.09	10	-10	1.030388	100	1	100	0.77	0	0	0	0	0	0	0	0	0	0	0;
	9060	284.51	138.58	210	-50	1.065482	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	9061	26.62	13.25	30	-10	1.040642	100	1	100	17.75	0	0	0	0	0	0	0	0	0	0	0;
	9074	43.84	6.21	60	-20	0.991243	100	1	100	29.23	0	0	0	0	0	0	0	0	0	0	0;
	9075	34.2	13.71	30	-10	1.044756	100	1	100	22.8	0	0	0	0	0	0	0	0	0	0	0;
	9081	28.35	17.5	30	-10	1.042184	100	1	100	18.9	0	0	0	0	0	0	0	0	0	0	0;
	9084	97.68	6.22	60	-20	1.017679	100	1	200	65.12	0	0	0	0	0	0	0	0	0	0	0;
	9088	34.65	8.09	30	-10	1.039009	100	1	100	23.1	0	0	0	0	0	0	0	0	0	0	0;
	9102	26.91	8.03	30	-10	1.04328	100	1	100	17.94	0	0	0	0	0	0	0	0	0	0	0;
	9106	26.91	8.03	30	-10	1.04328	100	1	100	17.94	0	0	0	0	0	0	0	0	0	0	0;
	9107	26.91	8.03	30	-10	1.04328	100	1	100	17.94	0	0	0	0	0	0	0	0	0	0	0;
	9110	23.58	-5	20	-20	0.964368	100	1	100	15.72	0	0	0	0	0	0	0	0	0	0	0;
	9120	25.89	9.83	40	-10	1.013208	100	1	100	17.26	0	0	0	0	0	0	0	0	0	0	0;
	9123	59.5	9.44	60	-20	1.032856	100	1	200	39.67	0	0	0	0	0	0	0	0	0	0	0;
	9137	43.55	17.11	40	-10	1.033542	100	1	100	29.03	0	0	0	0	0	0	0	0	0	0	0;
	9138	35.97	9.16	30	-10	1.001252	100	1	100	23.98	0	0	0	0	0	0	0	0	0	0	0;
	9141	58.49	10.33	40	-10	1.031723	100	1	200	38.99	0	0	0	0	0	0	0	0	0	0	0;
	9142	58.49	10.33	40	-10	1.031723	100	1	200	38.99	0	0	0	0	0	0	0	0	0	0	0;
	9144	83.78	15.02	60	-20	1.033534	100	1	200	55.85	0	0	0	0	0	0	0	0	0	0	0;
	9148	649.29	-32.35	520	-120	1.035363	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	9149	37.08	5.51	30	-10	1.021005	100	1	100	24.72	0	0	0	0	0	0	0	0	0	0	0;
	9151	9	-0.95	10	-10	0.943602	100	1	100	6	0	0	0	0	0	0	0	0	0	0	0;
	9152	858.21	148.21	520	-120	1.058618	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	9166	17.5	0.21	10	-10	1.00081	100	1	100	11.67	0	0	0	0	0	0	0	0	0	0	0;
	9168	4.8	1.5	10	-10	1.020547	100	1	100	3.2	0	0	0	0	0	0	0	0	0	0	0;
	9169	22.36	9.77	30	-10	1.014559	100	1	100	14.91	0	0	0	0	0	0	0	0	0	0	0;
	9179	87.35	17.85	60	-20	1.033768	100	1	200	58.23	0	0	0	0	0	0	0	0	0	0	0;
	9182	0	4.21	40	-10	1.016648	100	1	100	-0	0	0	0	0	0	0	0	0	0	0	0;
	9184	17.21	5.28	30	-10	1.020161	100	1	100	11.47	0	0	0	0	0	0	0	0	0	0	0;
	9188	88.9	23.56	60	-20	1.031838	100	1	200	59.27	0	0	0	0	0	0	0	0	0	0	0;
	9189	88.9	23.56	60	-20	1.031838	100	1	200	59.27	0	0	0	0	0	0	0	0	0	0	0;
	9190	88.9	23.56	60	-20	1.031838	100	1	200	59.27	0	0	0	0	0	0	0	0	0	0	0;
	9191	88.9	23.56	60	-20	1.031838	100	1	200	59.27	0	0	0	0	0	0	0	0	0	0	0;
	9195	210	15.07	130	-60	0.980209	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	9204	111.5	6.79	90	-60	1.056762	100	1	300	74.33	0	0	0	0	0	0	0	0	0	0	0;
	9206	209.96	80.01	160	-40	1.039006	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	9213	17.95	10.38	30	-10	1.044317	100	1	100	11.97	0	0	0	0	0	0	0	0	0	0	0;
	9214	39	5.27	30	-10	1.011094	100	1	100	26	0	0	0	0	0	0	0	0	0	0	0;
	9215	40.06	12.17	30	-10	1.05775	100	1	100	26.71	0	0	0	0	0	0	0	0	0	0	0;
	9216	20.55	5.71	30	-10	1.019482	100	1	100	13.7	0	0	0	0	0	0	0	0	0	0	0;
	9218	27.88	10.37	30	-10	1.052164	100	1	100	18.59	0	0	0	0	0	0	0	0	0	0	0;
	9219	35.29	10.61	30	-10	1.043764	100	1	100	23.53	0	0	0	0	0	0	0	0	0	0	0;
	9220	35.29	10.61	30	-10	1.043764	100	1	100	23.53	0	0	0	0	0	0	0	0	0	0	0;
	9222	17.95	10.38	30	-10	1.044317	100	1	100	11.97	0	0	0	0	0	0	0	0	0	0	0;
	9223	3.38	0.57	10	-10	1.017385	100	1	100	2.25	0	0	0	0	0	0	0	0	0	0	0;
	9224	17.95	10.38	30	-10	1.044317	100	1	100	11.97	0	0	0	0	0	0	0	0	0	0	0;
	9225	31.08	1.51	30	-10	1.048036	100	1	100	20.72	0	0	0	0	0	0	0	0	0	0	0;
	9227	23	-6.16	20	-20	0.989111	100	1	100	15.33	0	0	0	0	0	0	0	0	0	0	0;
	9228	95.42	32.99	60	-20	1.058394	100	1	200	63.61	0	0	0	0	0	0	0	0	0	0	0;
	9229	95.42	32.99	60	-20	1.058394	100	1	200	63.61	0	0	0	0	0	0	0	0	0	0	0;
	9230	95.42	32.99	60	-20	1.058394	100	1	200	63.61	0	0	0	0	0	0	0	0	0	0	0;
	9231	95.42	32.99	60	-20	1.058394	100	1	200	63.61	0	0	0	0	0	0	0	0	0	0	0;
	9232	95.42	32.99	60	-20	1.058394	100	1	200	63.61	0	0	0	0	0	0	0	0	0	0	0;
	9233	95.42	32.99	60	-20	1.058394	100	1	200	63.61	0	0	0	0	0	0	0	0	0	0	0;
	9236	35.01	1.48	30	-10	1.032132	100	1	100	23.34	0	0	0	0	0	0	0	0	0	0	0;
	9237	95.51	37.64	60	-20	1.0608	100	1	200	63.67	0	0	0	0	0	0	0	0	0	0	0;
	9244	34.62	2.89	20	-10	1.019266	100	1	100	23.08	0	0	0	0	0	0	0	0	0	0	0;
	9246	55.46	13.9	40	-10	1.030174	100	1	200	36.97	0	0	0	0	0	0	0	0	0	0	0;
	9247	55.46	13.9	40	-10	1.030174	100	1	200	36.97	0	0	0	0	0	0	0	0	0	0	0;
	9250	221.97	103.8	210	-50	1.063388	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	9256	279.67	54.61	210	-50	1.030939	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	9257	279.67	54.61	210	-50	1.030939	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	9258	279.67	54.61	210	-50	1.030939	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	9260	279.66	54.13	210	-50	1.030874	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	9261	279.66	54.13	210	-50	1.030874	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	9265	261.96	25.03	210	-50	1.042062	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	9273	35.95	6.39	40	-10	1.015185	100	1	100	23.97	0	0	0	0	0	0	0	0	0	0	0;
	9279	84.72	12.1	60	-20	1.046178	100	1	200	56.48	0	0	0	0	0	0	0	0	0	0	0;
	9280	34.2	13.71	30	-10	1.044756	100	1	100	22.8	0	0	0	0	0	0	0	0	0	0	0;
	9281	214.65	55.92	210	-50	1.073284	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	9282	50.94	19.18	40	-10	1.0136	100	1	200	33.96	0	0	0	0	0	0	0	0	0	0	0;
	9283	280.13	69.09	160	-40	1.028417	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	9288	24.28	7.01	30	-10	1.012542	100	1	100	16.19	0	0	0	0	0	0	0	0	0	0	0;
	9292	30	3.15	20	-10	1.046548	100	1	100	20	0	0	0	0	0	0	0	0	0	0	0;
	9294	336.9	100.14	210	-50	1.061222	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	9297	36.13	4.17	30	-10	1.027348	100	1	100	24.09	0	0	0	0	0	0	0	0	0	0	0;
	9301	79.79	-1.82	60	-20	1.024891	100	1	200	53.19	0	0	0	0	0	0	0	0	0	0	0;
	9312	9.02	-0.08	10	-10	1.003305	100	1	100	6.01	0	0	0	0	0	0	0	0	0	0	0;
	9314	30.24	9.06	30	-10	1.012031	100	1	100	20.16	0	0	0	0	0	0	0	0	0	0	0;
	9317	37.23	4.01	30	-10	1.028283	100	1	100	24.82	0	0	0	0	0	0	0	0	0	0	0;
	9321	29.7	6.22	30	-10	1.033776	100	1	100	19.8	0	0	0	0	0	0	0	0	0	0	0;
	9322	19.91	7.13	30	-10	1.021055	100	1	100	13.27	0	0	0	0	0	0	0	0	0	0	0;
	9323	19.91	7.13	30	-10	1.021055	100	1	100	13.27	0	0	0	0	0	0	0	0	0	0	0;
	9327	5.27	-0.03	10	-10	1.015617	100	1	100	3.51	0	0	0	0	0	0	0	0	0	0	0;
	9342	26.75	6.78	40	-10	1.024572	100	1	100	17.83	0	0	0	0	0	0	0	0	0	0	0;
	9343	57.51	17.72	40	-10	1.029323	100	1	200	38.34	0	0	0	0	0	0	0	0	0	0	0;
	9345	36.24	9.52	30	-10	1.028588	100	1	100	24.16	0	0	0	0	0	0	0	0	0	0	0;
	9346	36.24	9.52	30	-10	1.028588	100	1	100	24.16	0	0	0	0	0	0	0	0	0	0	0;
	9347	36.24	9.52	30	-10	1.028588	100	1	100	24.16	0	0	0	0	0	0	0	0	0	0	0;
	9348	36.24	9.52	30	-10	1.028588	100	1	100	24.16	0	0	0	0	0	0	0	0	0	0	0;
	9349	36.24	9.52	30	-10	1.028588	100	1	100	24.16	0	0	0	0	0	0	0	0	0	0	0;
	9350	36.24	9.52	30	-10	1.028588	100	1	100	24.16	0	0	0	0	0	0	0	0	0	0	0;
	9351	36.24	9.52	30	-10	1.028588	100	1	100	24.16	0	0	0	0	0	0	0	0	0	0	0;
	9352	36.24	9.52	30	-10	1.028588	100	1	100	24.16	0	0	0	0	0	0	0	0	0	0	0;
	9354	37.23	4.01	30	-10	1.028283	100	1	100	24.82	0	0	0	0	0	0	0	0	0	0	0;
	9357	56.25	11.12	40	-10	1.03036	100	1	200	37.5	0	0	0	0	0	0	0	0	0	0	0;
	9359	84.68	7.94	60	-20	1.00657	100	1	200	56.45	0	0	0	0	0	0	0	0	0	0	0;
	9364	240.6	37.13	100	-30	0.970854	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	9370	214.65	55.92	210	-50	1.073284	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	9374	36.77	13.93	30	-10	1.042317	100	1	100	24.51	0	0	0	0	0	0	0	0	0	0	0;
	9376	31.72	4.48	30	-10	1.022192	100	1	100	21.15	0	0	0	0	0	0	0	0	0	0	0;
	9377	31.72	4.48	30	-10	1.022192	100	1	100	21.15	0	0	0	0	0	0	0	0	0	0	0;
	9378	31.72	4.48	30	-10	1.022192	100	1	100	21.15	0	0	0	0	0	0	0	0	0	0	0;
	9379	31.72	4.48	30	-10	1.022192	100	1	100	21.15	0	0	0	0	0	0	0	0	0	0	0;
	9383	31.72	4.48	30	-10	1.022192	100	1	100	21.15	0	0	0	0	0	0	0	0	0	0	0;
	9384	31.72	4.48	30	-10	1.022192	100	1	100	21.15	0	0	0	0	0	0	0	0	0	0	0;
	9387	19.57	7.03	30	-10	1.023156	100	1	100	13.05	0	0	0	0	0	0	0	0	0	0	0;
	9389	9.94	1.09	10	-10	0.985384	100	1	100	6.63	0	0	0	0	0	0	0	0	0	0	0;
	9392	322.34	84.39	210	-50	1.058231	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	9393	63.93	6.28	60	-20	1.012697	100	1	200	42.62	0	0	0	0	0	0	0	0	0	0	0;
	9395	60	3.34	30	-20	1.020595	100	1	200	40	0	0	0	0	0	0	0	0	0	0	0;
	9408	97.57	36.99	60	-20	1.074058	100	1	200	65.05	0	0	0	0	0	0	0	0	0	0	0;
	9413	2.5	0.07	10	-10	0.984601	100	1	100	1.67	0	0	0	0	0	0	0	0	0	0	0;
	9417	57.27	17.52	40	-10	1.017845	100	1	200	38.18	0	0	0	0	0	0	0	0	0	0	0;
	9421	0.91	0.07	10	-10	1.030419	100	1	100	0.61	0	0	0	0	0	0	0	0	0	0	0;
	9422	74.86	19.42	60	-20	1.011998	100	1	200	49.91	0	0	0	0	0	0	0	0	0	0	0;
	9427	33.97	5.68	30	-10	1.022889	100	1	100	22.65	0	0	0	0	0	0	0	0	0	0	0;
	9438	68.26	20.93	60	-20	1.022995	100	1	200	45.51	0	0	0	0	0	0	0	0	0	0	0;
	9439	300.01	75.96	160	-40	1.035797	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	9440	68.26	20.93	60	-20	1.022995	100	1	200	45.51	0	0	0	0	0	0	0	0	0	0	0;
	9441	68.26	20.93	60	-20	1.022995	100	1	200	45.51	0	0	0	0	0	0	0	0	0	0	0;
	9442	68.26	20.93	60	-20	1.022995	100	1	200	45.51	0	0	0	0	0	0	0	0	0	0	0;
	9443	68.26	20.93	60	-20	1.022995	100	1	200	45.51	0	0	0	0	0	0	0	0	0	0	0;
	9457	157.5	-11.97	80	-80	0.984599	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	9461	50	-1.48	30	-20	0.985642	100	1	200	33.33	0	0	0	0	0	0	0	0	0	0	0;
	9464	93.04	14.9	60	-20	1.017638	100	1	200	62.03	0	0	0	0	0	0	0	0	0	0	0;
	9466	25.84	4.48	20	-10	1.018726	100	1	100	17.23	0	0	0	0	0	0	0	0	0	0	0;
	9467	84.3	7.63	60	-20	1.05433	100	1	200	56.2	0	0	0	0	0	0	0	0	0	0	0;
	9471	32.97	4.46	30	-10	1.027968	100	1	100	21.98	0	0	0	0	0	0	0	0	0	0	0;
	9472	32.97	4.46	30	-10	1.027968	100	1	100	21.98	0	0	0	0	0	0	0	0	0	0	0;
	9473	32.57	6.88	30	-10	1.03654	100	1	100	21.71	0	0	0	0	0	0	0	0	0	0	0;
	9474	32.57	6.88	30	-10	1.03654	100	1	100	21.71	0	0	0	0	0	0	0	0	0	0	0;
	9475	26.98	3.71	30	-10	1.066367	100	1	100	17.99	0	0	0	0	0	0	0	0	0	0	0;
	9476	26.98	3.71	30	-10	1.066367	100	1	100	17.99	0	0	0	0	0	0	0	0	0	0	0;
	9477	27.51	11.78	30	-10	1.042669	100	1	100	18.34	0	0	0	0	0	0	0	0	0	0	0;
	9478	26.86	3.11	30	-10	1.042101	100	1	100	17.91	0	0	0	0	0	0	0	0	0	0	0;
	9479	26.86	3.11	30	-10	1.042101	100	1	100	17.91	0	0	0	0	0	0	0	0	0	0	0;
	9480	26.86	3.11	30	-10	1.042101	100	1	100	17.91	0	0	0	0	0	0	0	0	0	0	0;
	9481	26.86	3.11	30	-10	1.042101	100	1	100	17.91	0	0	0	0	0	0	0	0	0	0	0;
	9482	26.86	3.11	30	-10	1.042101	100	1	100	17.91	0	0	0	0	0	0	0	0	0	0	0;
	9486	84.3	7.63	60	-20	1.05433	100	1	200	56.2	0	0	0	0	0	0	0	0	0	0	0;
	9504	44.8	3.17	40	-20	0.991989	100	1	100	29.87	0	0	0	0	0	0	0	0	0	0	0;
	9508	67.68	4.49	60	-20	1.017078	100	1	200	45.12	0	0	0	0	0	0	0	0	0	0	0;
	9512	96.03	17.22	60	-20	1.036065	100	1	200	64.02	0	0	0	0	0	0	0	0	0	0	0;
	9513	216.13	123.79	160	-40	1.059317	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	9515	243.54	114.28	160	-40	1.055847	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	9520	25.75	4.04	30	-10	1.026025	100	1	100	17.17	0	0	0	0	0	0	0	0	0	0	0;
	9521	35.84	16.6	30	-10	1.042077	100	1	100	23.89	0	0	0	0	0	0	0	0	0	0	0;
	9522	216.13	123.79	160	-40	1.059317	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	9523	249.82	72.04	160	-40	1.053646	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	9525	249.82	72.04	160	-40	1.053646	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	9527	370.08	97.24	210	-50	1.045516	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	9540	256.21	65.38	160	-40	1.052481	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	9542	22.38	13.57	30	-10	1.011374	100	1	100	14.92	0	0	0	0	0	0	0	0	0	0	0;
	9547	90.32	6.14	60	-20	1.034547	100	1	200	60.21	0	0	0	0	0	0	0	0	0	0	0;
	9549	90.32	6.14	60	-20	1.034547	100	1	200	60.21	0	0	0	0	0	0	0	0	0	0	0;
	9550	90.32	6.14	60	-20	1.034547	100	1	200	60.21	0	0	0	0	0	0	0	0	0	0	0;
	9551	90.32	6.14	60	-20	1.034547	100	1	200	60.21	0	0	0	0	0	0	0	0	0	0	0;
	9552	24.7	5.64	20	-20	1.010522	100	1	100	16.47	0	0	0	0	0	0	0	0	0	0	0;
	9553	167.4	-8.63	90	-70	0.948759	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	9554	2.12	-0.62	10	-10	0.962462	100	1	100	1.41	0	0	0	0	0	0	0	0	0	0	0;
	9556	95.83	32.34	60	-20	1.012871	100	1	200	63.89	0	0	0	0	0	0	0	0	0	0	0;
	9573	101	24.39	60	-20	1.022149	100	1	300	67.33	0	0	0	0	0	0	0	0	0	0	0;
	9574	101	24.39	60	-20	1.022149	100	1	300	67.33	0	0	0	0	0	0	0	0	0	0	0;
	9579	766.83	2.68	520	-120	1.05252	100	1	1600	500	0	0	0	0	0	0	0	0	0	0	0;
	9584	518.63	-13.96	520	-120	1.056808	100	1	1100	300	0	0	0	0	0	0	0	0	0	0	0;
	9585	518.63	-13.96	520	-120	1.056808	100	1	1100	300	0	0	0	0	0	0	0	0	0	0	0;
	9586	52.53	15.88	40	-10	1.02748	100	1	200	35.02	0	0	0	0	0	0	0	0	0	0	0;
	9587	52.53	15.88	40	-10	1.02748	100	1	200	35.02	0	0	0	0	0	0	0	0	0	0	0;
	9588	40.17	5.3	40	-10	1.030361	100	1	100	26.78	0	0	0	0	0	0	0	0	0	0	0;
	9589	40.17	5.3	40	-10	1.030361	100	1	100	26.78	0	0	0	0	0	0	0	0	0	0	0;
	9590	40.17	5.3	40	-10	1.030361	100	1	100	26.78	0	0	0	0	0	0	0	0	0	0	0;
	9591	40.17	5.3	40	-10	1.030361	100	1	100	26.78	0	0	0	0	0	0	0	0	0	0	0;
	9597	184.49	7.07	210	-50	1.044678	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	9602	76.5	12.8	40	-10	1.034298	100	1	200	51	0	0	0	0	0	0	0	0	0	0	0;
	9613	84.68	7.94	60	-20	1.00657	100	1	200	56.45	0	0	0	0	0	0	0	0	0	0	0;
	9614	24.84	4.52	30	-10	1.022099	100	1	100	16.56	0	0	0	0	0	0	0	0	0	0	0;
	9616	47.18	14.99	40	-10	1.021579	100	1	100	31.45	0	0	0	0	0	0	0	0	0	0	0;
	9617	52.42	24.94	40	-10	1.023225	100	1	200	34.95	0	0	0	0	0	0	0	0	0	0	0;
	9618	52.42	24.94	40	-10	1.023225	100	1	200	34.95	0	0	0	0	0	0	0	0	0	0	0;
	9619	55.06	16.94	40	-10	1.025226	100	1	200	36.71	0	0	0	0	0	0	0	0	0	0	0;
	9622	55.06	16.94	40	-10	1.025226	100	1	200	36.71	0	0	0	0	0	0	0	0	0	0	0;
	9623	44.71	9.09	40	-10	1.024756	100	1	100	29.81	0	0	0	0	0	0	0	0	0	0	0;
	9627	898.67	496.37	520	-120	1.055007	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	9628	50.6	9.92	30	-20	1.012126	100	1	200	33.73	0	0	0	0	0	0	0	0	0	0	0;
	9638	18.5	2.5	10	-10	0.975244	100	1	100	12.33	0	0	0	0	0	0	0	0	0	0	0;
	9655	932.42	349.72	520	-120	1.064074	100	1	1900	600	0	0	0	0	0	0	0	0	0	0	0;
	9656	932.56	349.6	520	-120	1.064118	100	1	1900	600	0	0	0	0	0	0	0	0	0	0	0;
	9659	33.84	9.43	30	-10	1.043178	100	1	100	22.56	0	0	0	0	0	0	0	0	0	0	0;
	9662	90.48	33.58	60	-20	1.043665	100	1	200	60.32	0	0	0	0	0	0	0	0	0	0	0;
	9679	874.56	115.53	520	-120	1.023313	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	9681	862.85	226.06	520	-120	1.036584	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	9682	862.85	223.65	520	-120	1.036459	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	9687	366.21	205.07	210	-50	1.129641	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	9690	198.91	5.92	210	-50	1.031301	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	9693	8.93	-0.3	10	-10	1.003834	100	1	100	5.95	0	0	0	0	0	0	0	0	0	0	0;
	9695	33.94	6.74	30	-10	1.046243	100	1	100	22.63	0	0	0	0	0	0	0	0	0	0	0;
	9696	33.94	6.74	30	-10	1.046243	100	1	100	22.63	0	0	0	0	0	0	0	0	0	0	0;
	9697	33.94	6.74	30	-10	1.046243	100	1	100	22.63	0	0	0	0	0	0	0	0	0	0	0;
	9698	33.94	6.74	30	-10	1.046243	100	1	100	22.63	0	0	0	0	0	0	0	0	0	0	0;
	9699	39.23	8.3	30	-10	1.030153	100	1	100	26.15	0	0	0	0	0	0	0	0	0	0	0;
	9700	30.78	2.48	30	-10	1.048966	100	1	100	20.52	0	0	0	0	0	0	0	0	0	0	0;
	9702	39.23	8.3	30	-10	1.030153	100	1	100	26.15	0	0	0	0	0	0	0	0	0	0	0;
	9706	23.96	4.79	30	-10	1.007614	100	1	100	15.97	0	0	0	0	0	0	0	0	0	0	0;
	9713	55.22	8.96	40	-10	1.017828	100	1	200	36.81	0	0	0	0	0	0	0	0	0	0	0;
	9722	152	11.5	90	-90	0.988205	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	9728	0.06	-0.1	10	-10	1.02351	100	1	100	0.04	0	0	0	0	0	0	0	0	0	0	0;
	9738	200.19	-0.55	160	-40	1.039371	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	9742	41.8	9.39	40	-10	1.028604	100	1	100	27.87	0	0	0	0	0	0	0	0	0	0	0;
	9748	3.65	0.04	10	-10	0.982004	100	1	100	2.43	0	0	0	0	0	0	0	0	0	0	0;
	9751	37.85	19.77	30	-10	1.057943	100	1	100	25.23	0	0	0	0	0	0	0	0	0	0	0;
	9753	37.86	19.74	30	-10	1.057896	100	1	100	25.24	0	0	0	0	0	0	0	0	0	0	0;
	9754	95.62	39.51	60	-20	1.057682	100	1	200	63.75	0	0	0	0	0	0	0	0	0	0	0;
	9756	50.94	19.18	40	-10	1.0136	100	1	200	33.96	0	0	0	0	0	0	0	0	0	0	0;
	9757	95.65	38.29	60	-20	1.058111	100	1	200	63.77	0	0	0	0	0	0	0	0	0	0	0;
	9758	95.65	38.29	60	-20	1.058111	100	1	200	63.77	0	0	0	0	0	0	0	0	0	0	0;
	9759	95.62	39.51	60	-20	1.057682	100	1	200	63.75	0	0	0	0	0	0	0	0	0	0	0;
	9761	95.65	38.29	60	-20	1.058111	100	1	200	63.77	0	0	0	0	0	0	0	0	0	0	0;
	9762	95.65	38.29	60	-20	1.058111	100	1	200	63.77	0	0	0	0	0	0	0	0	0	0	0;
	9763	95.65	38.29	60	-20	1.058111	100	1	200	63.77	0	0	0	0	0	0	0	0	0	0	0;
	9764	95.65	38.29	60	-20	1.058111	100	1	200	63.77	0	0	0	0	0	0	0	0	0	0	0;
	9766	45	6.33	30	-10	1.024995	100	1	100	30	0	0	0	0	0	0	0	0	0	0	0;
	9775	87.35	17.85	60	-20	1.033768	100	1	200	58.23	0	0	0	0	0	0	0	0	0	0	0;
	9778	10.77	0.4	10	-10	1.00995	100	1	100	7.18	0	0	0	0	0	0	0	0	0	0	0;
	9781	23.07	4.45	30	-10	1.026025	100	1	100	15.38	0	0	0	0	0	0	0	0	0	0	0;
	9785	83.69	18.55	60	-20	1.037564	100	1	200	55.79	0	0	0	0	0	0	0	0	0	0	0;
	9793	99.7	4.57	60	-20	1.028091	100	1	200	66.47	0	0	0	0	0	0	0	0	0	0	0;
	9797	0.18	-0.15	10	-10	1.016107	100	1	100	0.12	0	0	0	0	0	0	0	0	0	0	0;
	9804	18.5	2.5	10	-10	0.975244	100	1	100	12.33	0	0	0	0	0	0	0	0	0	0	0;
	9808	36.8	6.62	30	-10	1.03058	100	1	100	24.53	0	0	0	0	0	0	0	0	0	0	0;
	9809	27.2	3.82	30	-10	1.024762	100	1	100	18.13	0	0	0	0	0	0	0	0	0	0	0;
	9815	36.28	12.54	30	-10	1.022031	100	1	100	24.19	0	0	0	0	0	0	0	0	0	0	0;
	9818	41.94	8.5	40	-10	1.001006	100	1	100	27.96	0	0	0	0	0	0	0	0	0	0	0;
	9830	3.84	0.45	10	-10	1.030843	100	1	100	2.56	0	0	0	0	0	0	0	0	0	0	0;
	9836	12.8	1.08	10	-10	0.984262	100	1	100	8.53	0	0	0	0	0	0	0	0	0	0	0;
	9852	238.84	0.87	210	-50	1.029637	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	9853	13.95	4.91	10	0	1.042135	100	1	100	9.3	0	0	0	0	0	0	0	0	0	0	0;
	9855	8.8	0.68	10	-10	1.015406	100	1	100	5.87	0	0	0	0	0	0	0	0	0	0	0;
	9856	95.16	47.93	60	-20	1.047943	100	1	200	63.44	0	0	0	0	0	0	0	0	0	0	0;
	9860	22.5	7.8	30	-10	0.998215	100	1	100	15	0	0	0	0	0	0	0	0	0	0	0;
	9862	96.9	14.27	60	-20	1.033735	100	1	200	64.6	0	0	0	0	0	0	0	0	0	0	0;
	9870	96.9	14.27	60	-20	1.033735	100	1	200	64.6	0	0	0	0	0	0	0	0	0	0	0;
	9871	201.8	37.41	150	-70	1.033788	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	9873	26.62	13.25	30	-10	1.040642	100	1	100	17.75	0	0	0	0	0	0	0	0	0	0	0;
	9875	311.23	83.72	210	-50	1.058355	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	9880	36	6.82	30	-10	1.03481	100	1	100	24	0	0	0	0	0	0	0	0	0	0	0;
	9881	4.78	-0.59	10	-10	1.026644	100	1	100	3.19	0	0	0	0	0	0	0	0	0	0	0;
	9884	34.31	1.99	30	-10	1.015531	100	1	100	22.87	0	0	0	0	0	0	0	0	0	0	0;
	9885	36	6.45	30	-10	1.034215	100	1	100	24	0	0	0	0	0	0	0	0	0	0	0;
	9887	36	6.45	30	-10	1.034215	100	1	100	24	0	0	0	0	0	0	0	0	0	0	0;
	9889	300	-22.92	100	-100	0.97134	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	9890	74.22	5.53	60	-20	1.027947	100	1	200	49.48	0	0	0	0	0	0	0	0	0	0	0;
	9896	39.74	12.1	30	-10	1.035163	100	1	100	26.49	0	0	0	0	0	0	0	0	0	0	0;
	9897	18.61	5.39	30	-10	1.016633	100	1	100	12.41	0	0	0	0	0	0	0	0	0	0	0;
	9900	18.61	5.39	30	-10	1.016633	100	1	100	12.41	0	0	0	0	0	0	0	0	0	0	0;
	9901	33.97	5.68	30	-10	1.022889	100	1	100	22.65	0	0	0	0	0	0	0	0	0	0	0;
	9902	64.18	-6.8	40	-40	0.951565	100	1	200	42.79	0	0	0	0	0	0	0	0	0	0	0;
	9906	140	-6.9	70	-70	0.991405	100	1	300	93.33	0	0	0	0	0	0	0	0	0	0	0;
	9909	81.02	15.21	60	-20	1.035421	100	1	200	54.01	0	0	0	0	0	0	0	0	0	0	0;
	9911	81.02	15.21	60	-20	1.035421	100	1	200	54.01	0	0	0	0	0	0	0	0	0	0	0;
	9912	81.02	15.21	60	-20	1.035421	100	1	200	54.01	0	0	0	0	0	0	0	0	0	0	0;
	9914	30.01	15.35	30	-10	1.052369	100	1	100	20.01	0	0	0	0	0	0	0	0	0	0	0;
	9922	32.93	4.74	30	-10	1.029161	100	1	100	21.95	0	0	0	0	0	0	0	0	0	0	0;
	9923	29.22	4.97	30	-10	1.01017	100	1	100	19.48	0	0	0	0	0	0	0	0	0	0	0;
	9926	38.95	5.12	30	-10	1.009486	100	1	100	25.97	0	0	0	0	0	0	0	0	0	0	0;
	9932	28.89	4.69	30	-10	1.029382	100	1	100	19.26	0	0	0	0	0	0	0	0	0	0	0;
	9933	32.93	4.74	30	-10	1.029161	100	1	100	21.95	0	0	0	0	0	0	0	0	0	0	0;
	9937	100	10.54	90	-50	1.01208	100	1	300	66.67	0	0	0	0	0	0	0	0	0	0	0;
	9946	26.31	9.5	30	-10	1.038225	100	1	100	17.54	0	0	0	0	0	0	0	0	0	0	0;
	9953	93.04	14.9	60	-20	1.017638	100	1	200	62.03	0	0	0	0	0	0	0	0	0	0	0;
	9957	20.57	3.1	30	-10	1.026081	100	1	100	13.71	0	0	0	0	0	0	0	0	0	0	0;
	9958	20.57	3.1	30	-10	1.026081	100	1	100	13.71	0	0	0	0	0	0	0	0	0	0	0;
	9959	20.57	3.1	30	-10	1.026081	100	1	100	13.71	0	0	0	0	0	0	0	0	0	0	0;
	9960	82.46	55.28	60	-20	1.127948	100	1	200	54.97	0	0	0	0	0	0	0	0	0	0	0;
	9961	23.42	3.99	30	-10	1.016605	100	1	100	15.61	0	0	0	0	0	0	0	0	0	0	0;
	9963	27.95	4.02	30	-10	1.016461	100	1	100	18.63	0	0	0	0	0	0	0	0	0	0	0;
	9964	23.42	3.99	30	-10	1.016605	100	1	100	15.61	0	0	0	0	0	0	0	0	0	0	0;
	9972	153	12.38	100	-40	1.017859	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	9974	32.64	6.68	30	-10	1.044939	100	1	100	21.76	0	0	0	0	0	0	0	0	0	0	0;
	9975	40	4.84	20	-20	0.957116	100	1	100	26.67	0	0	0	0	0	0	0	0	0	0	0;
	9976	23.5	5.12	30	-10	1.029146	100	1	100	15.67	0	0	0	0	0	0	0	0	0	0	0;
	9984	65.21	12.85	60	-20	1.026243	100	1	200	43.47	0	0	0	0	0	0	0	0	0	0	0;
	9995	50	-2.27	20	-20	0.993383	100	1	100	33.33	0	0	0	0	0	0	0	0	0	0	0;
	9997	0.08	-0.15	10	-10	1.016106	100	1	100	0.05	0	0	0	0	0	0	0	0	0	0	0;
	10001	36.11	3.87	30	-10	1.024544	100	1	100	24.07	0	0	0	0	0	0	0	0	0	0	0;
	10004	35.95	6.39	40	-10	1.015185	100	1	100	23.97	0	0	0	0	0	0	0	0	0	0	0;
	10006	31.5	1.72	30	-20	1.016735	100	1	100	21	0	0	0	0	0	0	0	0	0	0	0;
	10007	35.29	10.61	30	-10	1.043764	100	1	100	23.53	0	0	0	0	0	0	0	0	0	0	0;
	10010	10.36	0.68	10	-10	1.012893	100	1	100	6.91	0	0	0	0	0	0	0	0	0	0	0;
	10012	10	0.56	10	-10	1.01515	100	1	100	6.67	0	0	0	0	0	0	0	0	0	0	0;
	10017	67.68	4.49	60	-20	1.017078	100	1	200	45.12	0	0	0	0	0	0	0	0	0	0	0;
	10023	28.92	12.55	40	-10	1.019357	100	1	100	19.28	0	0	0	0	0	0	0	0	0	0	0;
	10034	25.27	3.21	30	-10	1.006944	100	1	100	16.85	0	0	0	0	0	0	0	0	0	0	0;
	10037	57.12	17.95	40	-10	1.017464	100	1	200	38.08	0	0	0	0	0	0	0	0	0	0	0;
	10050	35.86	4.09	30	-10	1.02494	100	1	100	23.91	0	0	0	0	0	0	0	0	0	0	0;
	10052	5.2	1.64	10	-10	1.019503	100	1	100	3.47	0	0	0	0	0	0	0	0	0	0	0;
	10057	53.13	-3.75	40	-40	1.003179	100	1	200	35.42	0	0	0	0	0	0	0	0	0	0	0;
	10060	617.47	92.14	520	-120	1.032321	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	10064	858.79	258.85	520	-120	1.033602	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	10065	858.83	258.85	520	-120	1.033601	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	10068	17.95	10.38	30	-10	1.044317	100	1	100	11.97	0	0	0	0	0	0	0	0	0	0	0;
	10073	44.08	14.37	40	-10	1.028279	100	1	100	29.39	0	0	0	0	0	0	0	0	0	0	0;
	10074	44.08	14.37	40	-10	1.028279	100	1	100	29.39	0	0	0	0	0	0	0	0	0	0	0;
	10075	46.7	16.39	40	-10	1.027739	100	1	100	31.13	0	0	0	0	0	0	0	0	0	0	0;
	10076	46.7	16.39	40	-10	1.027739	100	1	100	31.13	0	0	0	0	0	0	0	0	0	0	0;
	10077	46.7	16.39	40	-10	1.027739	100	1	100	31.13	0	0	0	0	0	0	0	0	0	0	0;
	10078	46.7	16.39	40	-10	1.027739	100	1	100	31.13	0	0	0	0	0	0	0	0	0	0	0;
	10080	46.7	16.39	40	-10	1.027739	100	1	100	31.13	0	0	0	0	0	0	0	0	0	0	0;
	10084	57.67	14.65	40	-10	1.028916	100	1	200	38.45	0	0	0	0	0	0	0	0	0	0	0;
	10092	37.66	9.24	30	-10	1.060112	100	1	100	25.11	0	0	0	0	0	0	0	0	0	0	0;
	10094	240.6	61.33	100	-30	0.990464	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	10096	33.32	9.67	30	-10	1.032797	100	1	100	22.21	0	0	0	0	0	0	0	0	0	0	0;
	10100	35.56	3.76	20	-10	1.036949	100	1	100	23.71	0	0	0	0	0	0	0	0	0	0	0;
	10101	72.47	11.44	60	-20	1.019231	100	1	200	48.31	0	0	0	0	0	0	0	0	0	0	0;
	10102	321.29	105.87	210	-50	1.034412	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	10110	30.24	9.06	30	-10	1.012031	100	1	100	20.16	0	0	0	0	0	0	0	0	0	0	0;
	10114	2.26	-1.2	10	-10	0.991929	100	1	100	1.51	0	0	0	0	0	0	0	0	0	0	0;
	10115	301.72	83.89	210	-50	1.041212	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	10116	301.72	83.89	210	-50	1.041212	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	10117	301.72	83.89	210	-50	1.041212	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	10120	83.78	15.02	60	-20	1.033534	100	1	200	55.85	0	0	0	0	0	0	0	0	0	0	0;
	10132	33.68	6.27	30	-10	1.046665	100	1	100	22.45	0	0	0	0	0	0	0	0	0	0	0;
	10134	33.68	6.27	30	-10	1.046665	100	1	100	22.45	0	0	0	0	0	0	0	0	0	0	0;
	10135	33.68	6.27	30	-10	1.046665	100	1	100	22.45	0	0	0	0	0	0	0	0	0	0	0;
	10141	92.4	2.72	60	-20	1.014507	100	1	200	61.6	0	0	0	0	0	0	0	0	0	0	0;
	10142	92.4	2.72	60	-20	1.014507	100	1	200	61.6	0	0	0	0	0	0	0	0	0	0	0;
	10143	92.4	2.72	60	-20	1.014507	100	1	200	61.6	0	0	0	0	0	0	0	0	0	0	0;
	10151	33.06	9.48	30	-10	1.009173	100	1	100	22.04	0	0	0	0	0	0	0	0	0	0	0;
	10156	159.8	34.27	100	-40	0.978477	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	10158	29.65	15.88	30	-10	1.051307	100	1	100	19.77	0	0	0	0	0	0	0	0	0	0	0;
	10164	33.06	9.48	30	-10	1.009173	100	1	100	22.04	0	0	0	0	0	0	0	0	0	0	0;
	10165	367.04	57.82	210	-50	1.028579	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	10171	90.45	21.15	60	-20	1.046641	100	1	200	60.3	0	0	0	0	0	0	0	0	0	0	0;
	10172	90.45	21.15	60	-20	1.046641	100	1	200	60.3	0	0	0	0	0	0	0	0	0	0	0;
	10173	90.45	21.15	60	-20	1.046641	100	1	200	60.3	0	0	0	0	0	0	0	0	0	0	0;
	10174	90.45	21.15	60	-20	1.046641	100	1	200	60.3	0	0	0	0	0	0	0	0	0	0	0;
	10177	161	0.83	90	-90	1.011416	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	10178	34.65	5.41	30	-10	1.014865	100	1	100	23.1	0	0	0	0	0	0	0	0	0	0	0;
	10186	5.2	0.53	10	-10	1.031159	100	1	100	3.47	0	0	0	0	0	0	0	0	0	0	0;
	10192	25.32	7.7	30	-10	1.059619	100	1	100	16.88	0	0	0	0	0	0	0	0	0	0	0;
	10194	25.07	18.26	30	-10	1.048908	100	1	100	16.71	0	0	0	0	0	0	0	0	0	0	0;
	10195	25.07	18.26	30	-10	1.048908	100	1	100	16.71	0	0	0	0	0	0	0	0	0	0	0;
	10197	22.99	16.83	30	-10	1.052604	100	1	100	15.33	0	0	0	0	0	0	0	0	0	0	0;
	10198	22.99	16.83	30	-10	1.052604	100	1	100	15.33	0	0	0	0	0	0	0	0	0	0	0;
	10201	26.78	11.89	30	-10	1.043753	100	1	100	17.85	0	0	0	0	0	0	0	0	0	0	0;
	10202	26.78	11.89	30	-10	1.043753	100	1	100	17.85	0	0	0	0	0	0	0	0	0	0	0;
	10210	38.52	8.15	30	-10	1.04659	100	1	100	25.68	0	0	0	0	0	0	0	0	0	0	0;
	10211	228.6	14.34	190	-100	0.979798	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	10218	28.35	17.5	30	-10	1.042184	100	1	100	18.9	0	0	0	0	0	0	0	0	0	0	0;
	10222	33.12	3.26	20	-10	1.008414	100	1	100	22.08	0	0	0	0	0	0	0	0	0	0	0;
	10223	32.51	6.89	30	-10	1.011478	100	1	100	21.67	0	0	0	0	0	0	0	0	0	0	0;
	10232	23.65	10.04	30	-10	1.059543	100	1	100	15.77	0	0	0	0	0	0	0	0	0	0	0;
	10236	24.37	4.5	30	-10	1.00571	100	1	100	16.25	0	0	0	0	0	0	0	0	0	0	0;
	10237	2.94	0.27	10	-10	1.008736	100	1	100	1.96	0	0	0	0	0	0	0	0	0	0	0;
	10238	24.81	-3.88	30	-10	0.99879	100	1	100	16.54	0	0	0	0	0	0	0	0	0	0	0;
	10239	113.64	7.1	70	-50	0.989503	100	1	300	75.76	0	0	0	0	0	0	0	0	0	0	0;
	10241	19.5	-2.56	30	-10	1.017394	100	1	100	13	0	0	0	0	0	0	0	0	0	0	0;
	10242	19.5	-2.56	30	-10	1.017394	100	1	100	13	0	0	0	0	0	0	0	0	0	0	0;
	10246	26.75	6.78	40	-10	1.024572	100	1	100	17.83	0	0	0	0	0	0	0	0	0	0	0;
	10249	34.86	4.81	30	-10	1.050901	100	1	100	23.24	0	0	0	0	0	0	0	0	0	0	0;
	10250	163.94	189.92	210	-50	1.050579	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	10253	81.35	28.89	60	-20	1.055981	100	1	200	54.23	0	0	0	0	0	0	0	0	0	0	0;
	10254	81.35	28.89	60	-20	1.055981	100	1	200	54.23	0	0	0	0	0	0	0	0	0	0	0;
	10264	41.8	9.39	40	-10	1.028604	100	1	100	27.87	0	0	0	0	0	0	0	0	0	0	0;
	10267	98.47	3.73	60	-20	1.04126	100	1	200	65.65	0	0	0	0	0	0	0	0	0	0	0;
	10268	98.47	3.73	60	-20	1.04126	100	1	200	65.65	0	0	0	0	0	0	0	0	0	0	0;
	10269	57.3	3.1	60	-20	1.041091	100	1	200	38.2	0	0	0	0	0	0	0	0	0	0	0;
	10270	57.3	3.1	60	-20	1.041091	100	1	200	38.2	0	0	0	0	0	0	0	0	0	0	0;
	10276	9	-0.95	10	-10	0.943602	100	1	100	6	0	0	0	0	0	0	0	0	0	0	0;
	10281	34.01	4.15	30	-10	1.014364	100	1	100	22.67	0	0	0	0	0	0	0	0	0	0	0;
	10283	20.91	7.74	30	-10	1.014154	100	1	100	13.94	0	0	0	0	0	0	0	0	0	0	0;
	10285	6.1	0.22	10	-10	1.00107	100	1	100	4.07	0	0	0	0	0	0	0	0	0	0	0;
	10287	29.01	1.68	20	-10	1.010115	100	1	100	19.34	0	0	0	0	0	0	0	0	0	0	0;
	10295	27.37	3.79	30	-20	1.003014	100	1	100	18.25	0	0	0	0	0	0	0	0	0	0	0;
	10298	34.96	5.47	30	-30	0.935194	100	1	100	23.31	0	0	0	0	0	0	0	0	0	0	0;
	10302	36.1	4.8	30	-20	1.016385	100	1	100	24.07	0	0	0	0	0	0	0	0	0	0	0;
	10311	25.68	12.42	30	-10	1.052378	100	1	100	17.12	0	0	0	0	0	0	0	0	0	0	0;
	10315	25.35	8.5	40	-10	1.018631	100	1	100	16.9	0	0	0	0	0	0	0	0	0	0	0;
	10324	98.05	38.32	60	-20	1.058149	100	1	200	65.37	0	0	0	0	0	0	0	0	0	0	0;
	10328	85.34	16.39	60	-20	1.053529	100	1	200	56.89	0	0	0	0	0	0	0	0	0	0	0;
	10329	98.05	38.32	60	-20	1.058149	100	1	200	65.37	0	0	0	0	0	0	0	0	0	0	0;
	10330	586.88	407.04	520	-120	1.052406	100	1	1200	300	0	0	0	0	0	0	0	0	0	0	0;
	10337	72.99	3.91	60	-20	1.019549	100	1	200	48.66	0	0	0	0	0	0	0	0	0	0	0;
	10339	32.57	7.29	30	-10	1.041748	100	1	100	21.71	0	0	0	0	0	0	0	0	0	0	0;
	10343	57.4	4	60	-20	1.02039	100	1	200	38.27	0	0	0	0	0	0	0	0	0	0	0;
	10348	40.23	5.24	30	-10	1.006137	100	1	100	26.82	0	0	0	0	0	0	0	0	0	0	0;
	10352	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	10355	91.55	46.19	60	-20	1.056125	100	1	200	61.03	0	0	0	0	0	0	0	0	0	0	0;
	10357	311.23	83.72	210	-50	1.058355	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	10361	6.78	-0.39	10	-10	1.020123	100	1	100	4.52	0	0	0	0	0	0	0	0	0	0	0;
	10362	55.89	7.86	60	-20	0.992211	100	1	200	37.26	0	0	0	0	0	0	0	0	0	0	0;
	10363	40.23	5.24	30	-10	1.006137	100	1	100	26.82	0	0	0	0	0	0	0	0	0	0	0;
	10365	17.5	-0.6	20	-20	0.996564	100	1	100	11.67	0	0	0	0	0	0	0	0	0	0	0;
	10368	82.95	6.41	60	-50	0.992945	100	1	200	55.3	0	0	0	0	0	0	0	0	0	0	0;
	10373	99.8	-17.67	110	-30	1.029653	100	1	200	66.53	0	0	0	0	0	0	0	0	0	0	0;
	10374	314.64	49.18	210	-50	1.022804	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	10375	314.64	49.18	210	-50	1.022804	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	10376	56.1	11.17	40	-40	0.927975	100	1	200	37.4	0	0	0	0	0	0	0	0	0	0	0;
	10377	209.47	121.67	210	-50	1.027231	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	10378	209.47	121.67	210	-50	1.027231	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	10384	36.03	16.08	30	-10	1.045359	100	1	100	24.02	0	0	0	0	0	0	0	0	0	0	0;
	10386	33.15	0.93	30	-10	1.024669	100	1	100	22.1	0	0	0	0	0	0	0	0	0	0	0;
	10388	21.03	9.75	30	-10	1.034464	100	1	100	14.02	0	0	0	0	0	0	0	0	0	0	0;
	10390	63.63	44.69	60	-20	1.058213	100	1	200	42.42	0	0	0	0	0	0	0	0	0	0	0;
	10401	65.58	12.54	60	-20	1.029941	100	1	200	43.72	0	0	0	0	0	0	0	0	0	0	0;
	10402	65.58	12.54	60	-20	1.029941	100	1	200	43.72	0	0	0	0	0	0	0	0	0	0	0;
	10404	65.58	12.54	60	-20	1.029941	100	1	200	43.72	0	0	0	0	0	0	0	0	0	0	0;
	10405	65.58	12.54	60	-20	1.029941	100	1	200	43.72	0	0	0	0	0	0	0	0	0	0	0;
	10406	34.22	12.75	20	-10	1.04163	100	1	100	22.81	0	0	0	0	0	0	0	0	0	0	0;
	10408	16.79	7.28	30	-10	1.042166	100	1	100	11.19	0	0	0	0	0	0	0	0	0	0	0;
	10409	6.3	0.47	10	-10	1.012872	100	1	100	4.2	0	0	0	0	0	0	0	0	0	0	0;
	10410	36	6.45	30	-10	1.034215	100	1	100	24	0	0	0	0	0	0	0	0	0	0	0;
	10417	26.56	10.95	30	-10	1.002675	100	1	100	17.71	0	0	0	0	0	0	0	0	0	0	0;
	10418	37.9	13.47	30	-10	1.0326	100	1	100	25.27	0	0	0	0	0	0	0	0	0	0	0;
	10422	24.72	4.34	30	-10	1.039546	100	1	100	16.48	0	0	0	0	0	0	0	0	0	0	0;
	10423	66.31	16.47	60	-20	1.020956	100	1	200	44.21	0	0	0	0	0	0	0	0	0	0	0;
	10426	38.16	5.63	30	-10	1.015046	100	1	100	25.44	0	0	0	0	0	0	0	0	0	0	0;
	10427	38.16	5.63	30	-10	1.015046	100	1	100	25.44	0	0	0	0	0	0	0	0	0	0	0;
	10428	30	3.15	20	-10	1.046548	100	1	100	20	0	0	0	0	0	0	0	0	0	0	0;
	10432	12.66	5.6	30	-10	0.998217	100	1	100	8.44	0	0	0	0	0	0	0	0	0	0	0;
	10433	7.46	-0.12	10	-10	0.997543	100	1	100	4.97	0	0	0	0	0	0	0	0	0	0	0;
	10434	19.98	0.98	30	-10	1.029635	100	1	100	13.32	0	0	0	0	0	0	0	0	0	0	0;
	10435	19.98	0.98	30	-10	1.029635	100	1	100	13.32	0	0	0	0	0	0	0	0	0	0	0;
	10436	24.86	5.7	30	-10	0.997985	100	1	100	16.57	0	0	0	0	0	0	0	0	0	0	0;
	10437	15.97	11.98	30	-10	1.042028	100	1	100	10.65	0	0	0	0	0	0	0	0	0	0	0;
	10438	15.97	11.98	30	-10	1.042028	100	1	100	10.65	0	0	0	0	0	0	0	0	0	0	0;
	10439	12.66	5.6	30	-10	0.998217	100	1	100	8.44	0	0	0	0	0	0	0	0	0	0	0;
	10442	69.46	4.37	60	-20	1.041381	100	1	200	46.31	0	0	0	0	0	0	0	0	0	0	0;
	10443	26.31	9.5	30	-10	1.038225	100	1	100	17.54	0	0	0	0	0	0	0	0	0	0	0;
	10447	26.81	10.09	30	-10	1.044747	100	1	100	17.87	0	0	0	0	0	0	0	0	0	0	0;
	10448	45.54	5.69	40	-10	1.044129	100	1	100	30.36	0	0	0	0	0	0	0	0	0	0	0;
	10452	-0.54	-3.75	20	-20	1.014607	100	1	100	-0.72	0	0	0	0	0	0	0	0	0	0	0;
	10453	38.91	13.48	40	-10	0.994929	100	1	100	25.94	0	0	0	0	0	0	0	0	0	0	0;
	10456	39.24	8.33	30	-10	1.009038	100	1	100	26.16	0	0	0	0	0	0	0	0	0	0	0;
	10461	276.1	40.94	160	-40	1.035338	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	10462	276.1	40.94	160	-40	1.035338	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	10466	10	2.6	10	-10	1.008964	100	1	100	6.67	0	0	0	0	0	0	0	0	0	0	0;
	10469	37.08	5.51	30	-10	1.021005	100	1	100	24.72	0	0	0	0	0	0	0	0	0	0	0;
	10470	26.81	10.09	30	-10	1.044747	100	1	100	17.87	0	0	0	0	0	0	0	0	0	0	0;
	10472	95.38	41.05	60	-20	1.061235	100	1	200	63.59	0	0	0	0	0	0	0	0	0	0	0;
	10473	165	33.72	100	-20	0.962508	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	10474	95.38	41.05	60	-20	1.061235	100	1	200	63.59	0	0	0	0	0	0	0	0	0	0	0;
	10475	95.38	41.05	60	-20	1.061235	100	1	200	63.59	0	0	0	0	0	0	0	0	0	0	0;
	10476	95.38	41.05	60	-20	1.061235	100	1	200	63.59	0	0	0	0	0	0	0	0	0	0	0;
	10479	95.45	36.47	60	-20	1.060605	100	1	200	63.63	0	0	0	0	0	0	0	0	0	0	0;
	10480	95.45	36.47	60	-20	1.060605	100	1	200	63.63	0	0	0	0	0	0	0	0	0	0	0;
	10488	34.04	7.33	40	-10	1.031942	100	1	100	22.69	0	0	0	0	0	0	0	0	0	0	0;
	10489	34.04	7.33	40	-10	1.031942	100	1	100	22.69	0	0	0	0	0	0	0	0	0	0	0;
	10501	25.2	-3.93	20	-20	0.993105	100	1	100	16.8	0	0	0	0	0	0	0	0	0	0	0;
	10505	323.35	77.76	210	-50	1.058114	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	10507	5.27	-0.06	10	-10	1.016731	100	1	100	3.51	0	0	0	0	0	0	0	0	0	0	0;
	10508	10.89	-0.15	10	-10	1.009642	100	1	100	7.26	0	0	0	0	0	0	0	0	0	0	0;
	10521	86.41	23.5	60	-20	1.031898	100	1	200	57.61	0	0	0	0	0	0	0	0	0	0	0;
	10522	86.41	23.5	60	-20	1.031898	100	1	200	57.61	0	0	0	0	0	0	0	0	0	0	0;
	10527	36.42	7.96	30	-10	1.029166	100	1	100	24.28	0	0	0	0	0	0	0	0	0	0	0;
	10530	754.17	238.65	520	-120	1.043738	100	1	1600	500	0	0	0	0	0	0	0	0	0	0	0;
	10534	48.42	9.78	40	-10	1.01601	100	1	100	32.28	0	0	0	0	0	0	0	0	0	0	0;
	10535	209.49	209.89	210	-50	1.126501	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	10537	28.92	12.55	40	-10	1.019357	100	1	100	19.28	0	0	0	0	0	0	0	0	0	0	0;
	10538	28.92	12.55	40	-10	1.019357	100	1	100	19.28	0	0	0	0	0	0	0	0	0	0	0;
	10554	300	-14.29	100	-100	1.001419	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	10555	858.22	148.21	520	-120	1.058617	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	10556	858.23	148.21	520	-120	1.058617	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	10557	6.38	-0.43	10	-10	0.962211	100	1	100	4.25	0	0	0	0	0	0	0	0	0	0	0;
	10558	858.23	148.21	520	-120	1.058617	100	1	1800	500	0	0	0	0	0	0	0	0	0	0	0;
	10567	224.32	74.3	210	-50	1.033196	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	10575	30.07	6.48	30	-10	1.058835	100	1	100	20.05	0	0	0	0	0	0	0	0	0	0	0;
	10580	1.2	0.03	10	-10	1.003171	100	1	100	0.8	0	0	0	0	0	0	0	0	0	0	0;
	10588	62.5	10.73	20	0	1.000057	100	1	200	41.67	0	0	0	0	0	0	0	0	0	0	0;
	10589	72.47	11.44	60	-20	1.019231	100	1	200	48.31	0	0	0	0	0	0	0	0	0	0	0;
	10598	9.8	-0.3	10	-10	0.996611	100	1	100	6.53	0	0	0	0	0	0	0	0	0	0	0;
	10605	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	10612	27.19	6.05	30	-10	1.014542	100	1	100	18.13	0	0	0	0	0	0	0	0	0	0	0;
	10613	0.46	-0.11	10	-10	1.018329	100	1	100	0.31	0	0	0	0	0	0	0	0	0	0	0;
	10617	84.68	7.94	60	-20	1.00657	100	1	200	56.45	0	0	0	0	0	0	0	0	0	0	0;
	10619	81.96	34.23	60	-20	1.031098	100	1	200	54.64	0	0	0	0	0	0	0	0	0	0	0;
	10621	29.68	6.45	30	-10	1.016891	100	1	100	19.79	0	0	0	0	0	0	0	0	0	0	0;
	10622	35.06	4.02	30	-10	1.032126	100	1	100	23.37	0	0	0	0	0	0	0	0	0	0	0;
	10625	29.68	6.45	30	-10	1.016891	100	1	100	19.79	0	0	0	0	0	0	0	0	0	0	0;
	10627	84.19	8.48	60	-20	1.040259	100	1	200	56.13	0	0	0	0	0	0	0	0	0	0	0;
	10629	100.41	11.92	60	-20	1.038747	100	1	300	66.94	0	0	0	0	0	0	0	0	0	0	0;
	10633	90.48	33.58	60	-20	1.043665	100	1	200	60.32	0	0	0	0	0	0	0	0	0	0	0;
	10635	54.83	12.78	40	-10	1.047752	100	1	200	36.55	0	0	0	0	0	0	0	0	0	0	0;
	10640	188.36	79.01	160	-40	1.07025	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	10643	92.08	33.53	60	-20	1.056442	100	1	200	61.39	0	0	0	0	0	0	0	0	0	0	0;
	10647	209.93	1.9	210	-50	1.045549	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	10648	26.04	3.61	30	-10	1.050191	100	1	100	17.36	0	0	0	0	0	0	0	0	0	0	0;
	10649	26.04	3.61	30	-10	1.050191	100	1	100	17.36	0	0	0	0	0	0	0	0	0	0	0;
	10650	6.38	-2.2	10	-10	0.958146	100	1	100	4.25	0	0	0	0	0	0	0	0	0	0	0;
	10652	40	-3.06	20	-20	0.96764	100	1	100	26.67	0	0	0	0	0	0	0	0	0	0	0;
	10653	28.05	6.93	30	-10	1.035469	100	1	100	18.7	0	0	0	0	0	0	0	0	0	0	0;
	10656	34.38	8.21	30	-10	1.035773	100	1	100	22.92	0	0	0	0	0	0	0	0	0	0	0;
	10657	44.7	5.86	30	-10	0.996408	100	1	100	29.8	0	0	0	0	0	0	0	0	0	0	0;
	10658	44.7	5.86	30	-10	0.996408	100	1	100	29.8	0	0	0	0	0	0	0	0	0	0	0;
	10660	26.04	3.61	30	-10	1.050191	100	1	100	17.36	0	0	0	0	0	0	0	0	0	0	0;
	10668	31.45	10.87	30	-10	1.043771	100	1	100	20.97	0	0	0	0	0	0	0	0	0	0	0;
	10669	31.45	10.87	30	-10	1.043771	100	1	100	20.97	0	0	0	0	0	0	0	0	0	0	0;
	10670	39.24	8.33	30	-10	1.009038	100	1	100	26.16	0	0	0	0	0	0	0	0	0	0	0;
	10671	35.97	5.34	30	-10	1.058434	100	1	100	23.98	0	0	0	0	0	0	0	0	0	0	0;
	10672	35.97	5.34	30	-10	1.058434	100	1	100	23.98	0	0	0	0	0	0	0	0	0	0	0;
	10674	35.97	5.34	30	-10	1.058434	100	1	100	23.98	0	0	0	0	0	0	0	0	0	0	0;
	10675	35.97	5.34	30	-10	1.058434	100	1	100	23.98	0	0	0	0	0	0	0	0	0	0	0;
	10676	35.97	5.34	30	-10	1.058434	100	1	100	23.98	0	0	0	0	0	0	0	0	0	0	0;
	10678	11.68	-0.75	10	-10	1.013602	100	1	100	7.79	0	0	0	0	0	0	0	0	0	0	0;
	10693	5	0.46	10	-10	1.017738	100	1	100	3.33	0	0	0	0	0	0	0	0	0	0	0;
	10696	31.92	6.07	30	-10	1.027264	100	1	100	21.28	0	0	0	0	0	0	0	0	0	0	0;
	10712	33.38	14.38	30	-10	1.023574	100	1	100	22.25	0	0	0	0	0	0	0	0	0	0	0;
	10713	25.04	7.29	30	-10	1.041679	100	1	100	16.69	0	0	0	0	0	0	0	0	0	0	0;
	10715	330.71	4.92	210	-50	1.043049	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	10719	22.5	15.45	30	-10	1.045442	100	1	100	15	0	0	0	0	0	0	0	0	0	0	0;
	10720	22.5	15.45	30	-10	1.045442	100	1	100	15	0	0	0	0	0	0	0	0	0	0	0;
	10721	25.04	7.29	30	-10	1.041679	100	1	100	16.69	0	0	0	0	0	0	0	0	0	0	0;
	10722	25.04	7.29	30	-10	1.041679	100	1	100	16.69	0	0	0	0	0	0	0	0	0	0	0;
	10738	74.78	41.98	60	-20	1.04914	100	1	200	49.85	0	0	0	0	0	0	0	0	0	0	0;
	10742	25.68	12.42	30	-10	1.052378	100	1	100	17.12	0	0	0	0	0	0	0	0	0	0	0;
	10743	23.01	6.63	30	-10	1.01964	100	1	100	15.34	0	0	0	0	0	0	0	0	0	0	0;
	10747	85.34	16.39	60	-20	1.053529	100	1	200	56.89	0	0	0	0	0	0	0	0	0	0	0;
	10748	81.35	28.89	60	-20	1.055981	100	1	200	54.23	0	0	0	0	0	0	0	0	0	0	0;
	10750	72.99	3.91	60	-20	1.019549	100	1	200	48.66	0	0	0	0	0	0	0	0	0	0	0;
	10752	57.4	4	60	-20	1.02039	100	1	200	38.27	0	0	0	0	0	0	0	0	0	0	0;
	10753	59.71	48.41	60	-20	1.060516	100	1	200	39.81	0	0	0	0	0	0	0	0	0	0	0;
	10754	59.71	48.41	60	-20	1.060516	100	1	200	39.81	0	0	0	0	0	0	0	0	0	0	0;
	10755	72.99	3.91	60	-20	1.019549	100	1	200	48.66	0	0	0	0	0	0	0	0	0	0	0;
	10756	72.99	3.91	60	-20	1.019549	100	1	200	48.66	0	0	0	0	0	0	0	0	0	0	0;
	10757	10.5	2.44	10	-10	1.008788	100	1	100	7	0	0	0	0	0	0	0	0	0	0	0;
	10762	332.24	184.77	210	-50	1.071277	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	10763	332.24	184.77	210	-50	1.071277	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	10764	332.36	180.64	210	-50	1.068762	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	10765	332.24	184.77	210	-50	1.071277	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	10775	69.42	1.26	60	-20	1.034738	100	1	200	46.28	0	0	0	0	0	0	0	0	0	0	0;
	10783	67.5	10.49	20	0	1.030285	100	1	200	45	0	0	0	0	0	0	0	0	0	0	0;
	10784	40	-1.34	30	-30	1.010203	100	1	100	26.67	0	0	0	0	0	0	0	0	0	0	0;
	10785	35.74	6.84	30	-10	1.040138	100	1	100	23.83	0	0	0	0	0	0	0	0	0	0	0;
	10788	36.8	6.62	30	-10	1.03058	100	1	100	24.53	0	0	0	0	0	0	0	0	0	0	0;
	10803	54.83	12.78	40	-10	1.047752	100	1	200	36.55	0	0	0	0	0	0	0	0	0	0	0;
	10804	54.83	12.78	40	-10	1.047752	100	1	200	36.55	0	0	0	0	0	0	0	0	0	0	0;
	10808	95.02	8.09	60	-20	1.023144	100	1	200	63.35	0	0	0	0	0	0	0	0	0	0	0;
	10809	91.34	21.91	60	-20	1.030656	100	1	200	60.89	0	0	0	0	0	0	0	0	0	0	0;
	10811	91.94	33.76	60	-20	1.048656	100	1	200	61.29	0	0	0	0	0	0	0	0	0	0	0;
	10814	168.75	16.36	140	-130	0.976475	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	10825	152	11.5	90	-90	0.988205	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	10829	0.93	0.09	10	-10	1.03064	100	1	100	0.62	0	0	0	0	0	0	0	0	0	0	0;
	10835	47.25	-1.74	20	-20	0.990208	100	1	100	31.5	0	0	0	0	0	0	0	0	0	0	0;
	10841	95.83	32.34	60	-20	1.012871	100	1	200	63.89	0	0	0	0	0	0	0	0	0	0	0;
	10845	34.86	4.81	30	-10	1.050901	100	1	100	23.24	0	0	0	0	0	0	0	0	0	0	0;
	10847	82.46	55.28	60	-20	1.127948	100	1	200	54.97	0	0	0	0	0	0	0	0	0	0	0;
	10862	39.43	5.93	30	-10	1.017582	100	1	100	26.29	0	0	0	0	0	0	0	0	0	0	0;
	10865	37.29	8.19	30	-10	1.030566	100	1	100	24.86	0	0	0	0	0	0	0	0	0	0	0;
	10881	26.56	10.95	30	-10	1.002675	100	1	100	17.71	0	0	0	0	0	0	0	0	0	0	0;
	10891	15.59	-0.61	30	-20	1.020801	100	1	100	10.4	0	0	0	0	0	0	0	0	0	0	0;
	10898	25.24	9.06	30	-10	1.045295	100	1	100	16.83	0	0	0	0	0	0	0	0	0	0	0;
	10901	72.61	29.56	60	-20	1.041629	100	1	200	48.41	0	0	0	0	0	0	0	0	0	0	0;
	10907	93.04	14.9	60	-20	1.017638	100	1	200	62.03	0	0	0	0	0	0	0	0	0	0	0;
	10908	257.94	36.73	160	-40	1.021858	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	10913	78.96	32.05	60	-20	1.063361	100	1	200	52.64	0	0	0	0	0	0	0	0	0	0	0;
	10918	74.8	42.23	60	-20	1.046905	100	1	200	49.87	0	0	0	0	0	0	0	0	0	0	0;
	10927	289.29	40.05	160	-40	1.03324	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	10928	289.29	40.05	160	-40	1.03324	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	10930	289.29	40.05	160	-40	1.03324	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	10932	150	8.11	120	-70	0.989744	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	10935	25.75	4.04	30	-10	1.026025	100	1	100	17.17	0	0	0	0	0	0	0	0	0	0	0;
	10938	21.76	4.41	20	-10	1.034618	100	1	100	14.51	0	0	0	0	0	0	0	0	0	0	0;
	10940	40.5	-3.48	20	-20	0.963524	100	1	100	27	0	0	0	0	0	0	0	0	0	0	0;
	10945	7	-0.2	10	-10	0.987313	100	1	100	4.67	0	0	0	0	0	0	0	0	0	0	0;
	10946	234.93	76.89	210	-50	1.032433	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	10951	50.8	2.36	40	-30	0.979833	100	1	200	33.87	0	0	0	0	0	0	0	0	0	0	0;
	10954	180.96	38.95	160	-40	1.029835	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	10963	111.5	6.79	90	-60	1.056762	100	1	300	74.33	0	0	0	0	0	0	0	0	0	0	0;
	10971	15	-0.86	10	-10	0.992395	100	1	100	10	0	0	0	0	0	0	0	0	0	0	0;
	10973	89.04	1.21	60	-40	1.015065	100	1	200	59.36	0	0	0	0	0	0	0	0	0	0	0;
	10978	39.43	5.93	30	-10	1.017582	100	1	100	26.29	0	0	0	0	0	0	0	0	0	0	0;
	10983	21.77	6.35	30	-10	1.024499	100	1	100	14.51	0	0	0	0	0	0	0	0	0	0	0;
	10985	87.19	27.18	60	-20	1.023338	100	1	200	58.13	0	0	0	0	0	0	0	0	0	0	0;
	10987	20.64	7.21	30	-10	1.018866	100	1	100	13.76	0	0	0	0	0	0	0	0	0	0	0;
	10994	23.58	-5	20	-20	0.964368	100	1	100	15.72	0	0	0	0	0	0	0	0	0	0	0;
	10998	994.84	302.65	520	-120	1.05787	100	1	2000	600	0	0	0	0	0	0	0	0	0	0	0;
	11005	36.77	13.93	30	-10	1.042317	100	1	100	24.51	0	0	0	0	0	0	0	0	0	0	0;
	11010	100.29	29.56	60	-20	1.064335	100	1	300	66.86	0	0	0	0	0	0	0	0	0	0	0;
	11020	2.26	-1.2	10	-10	0.991929	100	1	100	1.51	0	0	0	0	0	0	0	0	0	0	0;
	11023	42.33	15.72	40	-10	1.031695	100	1	100	28.22	0	0	0	0	0	0	0	0	0	0	0;
	11025	42.33	15.72	40	-10	1.031695	100	1	100	28.22	0	0	0	0	0	0	0	0	0	0	0;
	11028	58.52	7.84	40	-10	1.025977	100	1	200	39.01	0	0	0	0	0	0	0	0	0	0	0;
	11030	99.72	-12.35	60	-20	1.021495	100	1	200	66.48	0	0	0	0	0	0	0	0	0	0	0;
	11032	800.47	18.87	520	-120	1.04071	100	1	1700	500	0	0	0	0	0	0	0	0	0	0	0;
	11037	158.79	48.73	210	-50	1.029671	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	11039	30.07	6.48	30	-10	1.058835	100	1	100	20.05	0	0	0	0	0	0	0	0	0	0	0;
	11047	154.5	-14.56	80	-70	0.956299	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	11049	100.41	11.92	60	-20	1.038747	100	1	300	66.94	0	0	0	0	0	0	0	0	0	0	0;
	11051	72.47	11.44	60	-20	1.019231	100	1	200	48.31	0	0	0	0	0	0	0	0	0	0	0;
	11059	50	-2.27	20	-20	0.993383	100	1	100	33.33	0	0	0	0	0	0	0	0	0	0	0;
	11065	27.19	6.05	30	-10	1.014542	100	1	100	18.13	0	0	0	0	0	0	0	0	0	0	0;
	11076	19.96	5.53	30	-10	1.011132	100	1	100	13.31	0	0	0	0	0	0	0	0	0	0	0;
	11085	32.44	9.61	30	-10	1.038916	100	1	100	21.63	0	0	0	0	0	0	0	0	0	0	0;
	11086	32.44	9.61	30	-10	1.038916	100	1	100	21.63	0	0	0	0	0	0	0	0	0	0	0;
	11087	32.44	9.61	30	-10	1.038916	100	1	100	21.63	0	0	0	0	0	0	0	0	0	0	0;
	11088	32.44	9.61	30	-10	1.038916	100	1	100	21.63	0	0	0	0	0	0	0	0	0	0	0;
	11090	23.94	10.89	30	-10	1.042693	100	1	100	15.96	0	0	0	0	0	0	0	0	0	0	0;
	11095	22.02	10.78	30	-10	1.026425	100	1	100	14.68	0	0	0	0	0	0	0	0	0	0	0;
	11096	200.33	74.16	210	-50	1.024096	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	11099	35.6	15.17	30	-10	1.026957	100	1	100	23.73	0	0	0	0	0	0	0	0	0	0	0;
	11102	33.73	12.57	30	-10	1.035958	100	1	100	22.49	0	0	0	0	0	0	0	0	0	0	0;
	11104	33.73	12.57	30	-10	1.035958	100	1	100	22.49	0	0	0	0	0	0	0	0	0	0	0;
	11106	30.35	8.38	30	-10	1.053822	100	1	100	20.23	0	0	0	0	0	0	0	0	0	0	0;
	11107	33.73	12.57	30	-10	1.035958	100	1	100	22.49	0	0	0	0	0	0	0	0	0	0	0;
	11108	30.35	8.38	30	-10	1.053822	100	1	100	20.23	0	0	0	0	0	0	0	0	0	0	0;
	11109	30.35	8.38	30	-10	1.053822	100	1	100	20.23	0	0	0	0	0	0	0	0	0	0	0;
	11110	28.95	9.55	30	-10	1.01638	100	1	100	19.3	0	0	0	0	0	0	0	0	0	0	0;
	11111	30.35	8.38	30	-10	1.053822	100	1	100	20.23	0	0	0	0	0	0	0	0	0	0	0;
	11115	28.83	4.62	30	-10	1.019144	100	1	100	19.22	0	0	0	0	0	0	0	0	0	0	0;
	11118	23.6	12.35	30	-10	1.042829	100	1	100	15.73	0	0	0	0	0	0	0	0	0	0	0;
	11119	23.6	12.35	30	-10	1.042829	100	1	100	15.73	0	0	0	0	0	0	0	0	0	0	0;
	11120	25.54	10.25	30	-10	1.035471	100	1	100	17.03	0	0	0	0	0	0	0	0	0	0	0;
	11121	25.54	10.25	30	-10	1.035471	100	1	100	17.03	0	0	0	0	0	0	0	0	0	0	0;
	11122	39.54	11.66	30	-10	1.055315	100	1	100	26.36	0	0	0	0	0	0	0	0	0	0	0;
	11124	23.6	12.35	30	-10	1.042829	100	1	100	15.73	0	0	0	0	0	0	0	0	0	0	0;
	11125	23.6	12.35	30	-10	1.042829	100	1	100	15.73	0	0	0	0	0	0	0	0	0	0	0;
	11128	90.48	33.58	60	-20	1.043665	100	1	200	60.32	0	0	0	0	0	0	0	0	0	0	0;
	11129	27.67	4.45	30	-10	1.027132	100	1	100	18.45	0	0	0	0	0	0	0	0	0	0	0;
	11130	25.54	10.25	30	-10	1.035471	100	1	100	17.03	0	0	0	0	0	0	0	0	0	0	0;
	11131	43.29	23.54	60	-20	1.035771	100	1	100	28.86	0	0	0	0	0	0	0	0	0	0	0;
	11132	90.23	10.46	60	-20	1.028204	100	1	200	60.15	0	0	0	0	0	0	0	0	0	0	0;
	11133	90.23	10.46	60	-20	1.028204	100	1	200	60.15	0	0	0	0	0	0	0	0	0	0	0;
	11137	90.23	10.46	60	-20	1.028204	100	1	200	60.15	0	0	0	0	0	0	0	0	0	0	0;
	11138	31.5	1.72	30	-20	1.016735	100	1	100	21	0	0	0	0	0	0	0	0	0	0	0;
	11145	28.8	6.92	30	-10	1.000808	100	1	100	19.2	0	0	0	0	0	0	0	0	0	0	0;
	11154	188.36	79.01	160	-40	1.07025	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	11155	31.92	6.07	30	-10	1.027264	100	1	100	21.28	0	0	0	0	0	0	0	0	0	0	0;
	11158	3.2	1.16	10	-10	0.991765	100	1	100	2.13	0	0	0	0	0	0	0	0	0	0	0;
	11167	84.19	8.48	60	-20	1.040259	100	1	200	56.13	0	0	0	0	0	0	0	0	0	0	0;
	11170	86.25	-3.04	60	-80	0.962614	100	1	200	57.5	0	0	0	0	0	0	0	0	0	0	0;
	11172	84.19	8.48	60	-20	1.040259	100	1	200	56.13	0	0	0	0	0	0	0	0	0	0	0;
	11174	84.19	8.48	60	-20	1.040259	100	1	200	56.13	0	0	0	0	0	0	0	0	0	0	0;
	11175	84.19	8.48	60	-20	1.040259	100	1	200	56.13	0	0	0	0	0	0	0	0	0	0	0;
	11177	10	0.56	10	-10	1.01515	100	1	100	6.67	0	0	0	0	0	0	0	0	0	0	0;
	11191	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	11193	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	11194	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	11195	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	11196	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	11197	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	11198	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	11201	318.17	43.77	210	-50	1.020879	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	11203	17.6	22.53	30	-10	1.05667	100	1	100	11.73	0	0	0	0	0	0	0	0	0	0	0;
	11204	318.17	43.77	210	-50	1.020879	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	11211	33.79	6.99	30	-10	1.059581	100	1	100	22.53	0	0	0	0	0	0	0	0	0	0	0;
	11212	33.59	11.77	30	-10	1.049265	100	1	100	22.39	0	0	0	0	0	0	0	0	0	0	0;
	11214	26.84	7.89	30	-10	1.013527	100	1	100	17.89	0	0	0	0	0	0	0	0	0	0	0;
	11219	78.58	36.25	60	-20	1.047402	100	1	200	52.39	0	0	0	0	0	0	0	0	0	0	0;
	11220	84.71	11.94	60	-20	1.045943	100	1	200	56.47	0	0	0	0	0	0	0	0	0	0	0;
	11221	36.8	6.62	30	-10	1.03058	100	1	100	24.53	0	0	0	0	0	0	0	0	0	0	0;
	11225	34.65	8.09	30	-10	1.039009	100	1	100	23.1	0	0	0	0	0	0	0	0	0	0	0;
	11226	99.7	4.57	60	-20	1.028091	100	1	200	66.47	0	0	0	0	0	0	0	0	0	0	0;
	11227	32.64	6.68	30	-10	1.044939	100	1	100	21.76	0	0	0	0	0	0	0	0	0	0	0;
	11229	34.8	5.13	30	-10	1.009652	100	1	100	23.2	0	0	0	0	0	0	0	0	0	0	0;
	11230	1.84	0.43	10	-10	1.045545	100	1	100	1.23	0	0	0	0	0	0	0	0	0	0	0;
	11231	32.64	6.68	30	-10	1.044939	100	1	100	21.76	0	0	0	0	0	0	0	0	0	0	0;
	11233	61.58	-0.59	60	-20	1.041114	100	1	200	41.05	0	0	0	0	0	0	0	0	0	0	0;
	11235	57.74	17.07	60	-20	1.05622	100	1	200	38.49	0	0	0	0	0	0	0	0	0	0	0;
	11236	61.58	-0.59	60	-20	1.041114	100	1	200	41.05	0	0	0	0	0	0	0	0	0	0	0;
	11237	27.72	3.43	30	-10	1.015411	100	1	100	18.48	0	0	0	0	0	0	0	0	0	0	0;
	11238	57.74	17.07	60	-20	1.05622	100	1	200	38.49	0	0	0	0	0	0	0	0	0	0	0;
	11239	5.17	0.23	10	-10	1.027155	100	1	100	3.45	0	0	0	0	0	0	0	0	0	0	0;
	11240	83.51	27.65	60	-20	1.063297	100	1	200	55.67	0	0	0	0	0	0	0	0	0	0	0;
	11241	83.51	27.65	60	-20	1.063297	100	1	200	55.67	0	0	0	0	0	0	0	0	0	0	0;
	11242	83.51	27.65	60	-20	1.063297	100	1	200	55.67	0	0	0	0	0	0	0	0	0	0	0;
	11244	27.72	3.43	30	-10	1.015411	100	1	100	18.48	0	0	0	0	0	0	0	0	0	0	0;
	11250	31.57	7.72	30	-10	1.008948	100	1	100	21.05	0	0	0	0	0	0	0	0	0	0	0;
	11251	50.96	2.69	60	-20	1.028648	100	1	200	33.97	0	0	0	0	0	0	0	0	0	0	0;
	11256	328.2	86.92	210	-50	1.03657	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	11257	45.95	11.1	20	0	1.036997	100	1	100	30.63	0	0	0	0	0	0	0	0	0	0	0;
	11259	328.2	86.92	210	-50	1.03657	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	11264	36.8	6.62	30	-10	1.03058	100	1	100	24.53	0	0	0	0	0	0	0	0	0	0	0;
	11265	27.2	3.82	30	-10	1.024762	100	1	100	18.13	0	0	0	0	0	0	0	0	0	0	0;
	11273	40	-3.06	20	-20	0.96764	100	1	100	26.67	0	0	0	0	0	0	0	0	0	0	0;
	11281	15	4.39	10	0	1.03858	100	1	100	10	0	0	0	0	0	0	0	0	0	0	0;
	11285	88.85	23.49	60	-20	1.032344	100	1	200	59.23	0	0	0	0	0	0	0	0	0	0	0;
	11289	34.38	8.21	30	-10	1.035773	100	1	100	22.92	0	0	0	0	0	0	0	0	0	0	0;
	11294	951.82	32.74	520	-120	1.01249	100	1	2000	600	0	0	0	0	0	0	0	0	0	0	0;
	11299	89.91	48.94	60	-20	1.065506	100	1	200	59.94	0	0	0	0	0	0	0	0	0	0	0;
	11305	68.26	9.75	60	-20	1.028153	100	1	200	45.51	0	0	0	0	0	0	0	0	0	0	0;
	11307	56.07	3.64	60	-20	1.04834	100	1	200	37.38	0	0	0	0	0	0	0	0	0	0	0;
	11308	88.6	0.84	60	-20	1.03671	100	1	200	59.07	0	0	0	0	0	0	0	0	0	0	0;
	11309	34.38	8.21	30	-10	1.035773	100	1	100	22.92	0	0	0	0	0	0	0	0	0	0	0;
	11311	3.65	0.04	10	-10	0.982004	100	1	100	2.43	0	0	0	0	0	0	0	0	0	0	0;
	11312	50.83	3.96	60	-20	1.049674	100	1	200	33.89	0	0	0	0	0	0	0	0	0	0	0;
	11315	36.11	3.87	30	-10	1.024544	100	1	100	24.07	0	0	0	0	0	0	0	0	0	0	0;
	11325	29.57	20.51	30	-10	1.041409	100	1	100	19.71	0	0	0	0	0	0	0	0	0	0	0;
	11327	39.88	15.15	30	-10	1.024267	100	1	100	26.59	0	0	0	0	0	0	0	0	0	0	0;
	11328	34.86	4.81	30	-10	1.050901	100	1	100	23.24	0	0	0	0	0	0	0	0	0	0	0;
	11329	8.4	-0.55	10	-10	0.988112	100	1	100	5.6	0	0	0	0	0	0	0	0	0	0	0;
	11330	26.97	7.32	30	-10	1.049221	100	1	100	17.98	0	0	0	0	0	0	0	0	0	0	0;
	11332	165.71	155.15	160	-40	1.077226	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	11333	165.71	155.15	160	-40	1.077226	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	11336	30.02	16.69	30	-10	1.023695	100	1	100	20.01	0	0	0	0	0	0	0	0	0	0	0;
	11344	27.88	10.37	30	-10	1.052164	100	1	100	18.59	0	0	0	0	0	0	0	0	0	0	0;
	11345	90.32	31.5	60	-20	1.035102	100	1	200	60.21	0	0	0	0	0	0	0	0	0	0	0;
	11347	43.48	23.86	40	-10	1.021627	100	1	100	28.99	0	0	0	0	0	0	0	0	0	0	0;
	11348	39.64	12.04	40	-10	1.034643	100	1	100	26.43	0	0	0	0	0	0	0	0	0	0	0;
	11351	40.44	13.81	40	-10	1.025019	100	1	100	26.96	0	0	0	0	0	0	0	0	0	0	0;
	11352	40.44	13.81	40	-10	1.025019	100	1	100	26.96	0	0	0	0	0	0	0	0	0	0	0;
	11353	39.64	12.04	40	-10	1.034643	100	1	100	26.43	0	0	0	0	0	0	0	0	0	0	0;
	11355	33.92	4.57	30	-10	1.012747	100	1	100	22.61	0	0	0	0	0	0	0	0	0	0	0;
	11360	274.92	131.33	210	-50	1.086148	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	11361	354.27	134.78	210	-50	1.085902	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	11362	354.27	134.78	210	-50	1.085902	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	11363	269.77	119.39	210	-50	1.02099	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	11364	90.17	24.56	60	-20	1.06019	100	1	200	60.11	0	0	0	0	0	0	0	0	0	0	0;
	11367	90.17	24.56	60	-20	1.06019	100	1	200	60.11	0	0	0	0	0	0	0	0	0	0	0;
	11368	90.17	24.56	60	-20	1.06019	100	1	200	60.11	0	0	0	0	0	0	0	0	0	0	0;
	11369	90.17	24.56	60	-20	1.06019	100	1	200	60.11	0	0	0	0	0	0	0	0	0	0	0;
	11370	90.17	24.56	60	-20	1.06019	100	1	200	60.11	0	0	0	0	0	0	0	0	0	0	0;
	11377	275.46	31.72	160	-40	1.005433	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	11393	0.08	-0.15	10	-10	1.016106	100	1	100	0.05	0	0	0	0	0	0	0	0	0	0	0;
	11401	19.93	2.39	30	-10	1.050328	100	1	100	13.29	0	0	0	0	0	0	0	0	0	0	0;
	11403	293.92	111.78	210	-50	1.039038	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	11406	382.46	57.23	210	-50	1.050841	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	11408	169.22	17.53	160	-40	1.013014	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	11409	169.22	17.53	160	-40	1.013014	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	11410	84.71	3.25	50	-20	1.030034	100	1	200	56.47	0	0	0	0	0	0	0	0	0	0	0;
	11411	68.89	19.23	60	-20	1.035618	100	1	200	45.93	0	0	0	0	0	0	0	0	0	0	0;
	11412	68.89	19.23	60	-20	1.035618	100	1	200	45.93	0	0	0	0	0	0	0	0	0	0	0;
	11414	68.89	19.23	60	-20	1.035618	100	1	200	45.93	0	0	0	0	0	0	0	0	0	0	0;
	11416	68.89	19.23	60	-20	1.035618	100	1	200	45.93	0	0	0	0	0	0	0	0	0	0	0;
	11418	638.57	231.71	520	-120	1.051278	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	11419	638.58	231.71	520	-120	1.051278	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	11440	2.5	0.07	10	-10	0.984601	100	1	100	1.67	0	0	0	0	0	0	0	0	0	0	0;
	11454	28.52	8.46	30	-10	1.013471	100	1	100	19.01	0	0	0	0	0	0	0	0	0	0	0;
	11455	28.52	8.46	30	-10	1.013471	100	1	100	19.01	0	0	0	0	0	0	0	0	0	0	0;
	11456	30.24	3.01	20	-10	1.001613	100	1	100	20.16	0	0	0	0	0	0	0	0	0	0	0;
	11467	26.62	13.25	30	-10	1.040642	100	1	100	17.75	0	0	0	0	0	0	0	0	0	0	0;
	11471	19.85	-1.9	30	-10	1.026413	100	1	100	13.23	0	0	0	0	0	0	0	0	0	0	0;
	11472	206.33	45.76	160	-40	1.035329	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	11473	35.69	7.89	30	-10	1.017601	100	1	100	23.79	0	0	0	0	0	0	0	0	0	0	0;
	11474	24.84	13.51	30	-10	1.036049	100	1	100	16.56	0	0	0	0	0	0	0	0	0	0	0;
	11475	206.33	45.6	160	-40	1.035259	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	11477	206.33	45.76	160	-40	1.035329	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	11478	206.33	45.6	160	-40	1.035259	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	11479	206.32	45.91	160	-40	1.035499	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	11480	206.32	45.73	160	-40	1.035431	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	11481	28.04	16.4	30	-10	1.019623	100	1	100	18.69	0	0	0	0	0	0	0	0	0	0	0;
	11482	29.84	3.47	30	-10	1.022039	100	1	100	19.89	0	0	0	0	0	0	0	0	0	0	0;
	11483	381.8	21.67	210	-50	1.063714	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	11484	20.43	6.2	30	-10	1.037469	100	1	100	13.62	0	0	0	0	0	0	0	0	0	0	0;
	11487	36.77	13.93	30	-10	1.042317	100	1	100	24.51	0	0	0	0	0	0	0	0	0	0	0;
	11492	240.6	37.13	100	-30	0.970854	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	11496	53.99	18.23	40	-10	1.062896	100	1	200	35.99	0	0	0	0	0	0	0	0	0	0	0;
	11497	27.96	5.11	30	-10	1.029009	100	1	100	18.64	0	0	0	0	0	0	0	0	0	0	0;
	11498	27.96	5.11	30	-10	1.029009	100	1	100	18.64	0	0	0	0	0	0	0	0	0	0	0;
	11499	27.96	5.11	30	-10	1.029009	100	1	100	18.64	0	0	0	0	0	0	0	0	0	0	0;
	11500	27.96	5.11	30	-10	1.029009	100	1	100	18.64	0	0	0	0	0	0	0	0	0	0	0;
	11502	11.05	2.35	20	-10	1.017349	100	1	100	7.37	0	0	0	0	0	0	0	0	0	0	0;
	11504	55.79	12.28	40	-10	1.022885	100	1	200	37.19	0	0	0	0	0	0	0	0	0	0	0;
	11506	53.99	18.23	40	-10	1.062896	100	1	200	35.99	0	0	0	0	0	0	0	0	0	0	0;
	11507	53.99	18.23	40	-10	1.062896	100	1	200	35.99	0	0	0	0	0	0	0	0	0	0	0;
	11509	70	2.13	60	-40	1.013195	100	1	200	46.67	0	0	0	0	0	0	0	0	0	0	0;
	11510	289.29	40.05	160	-40	1.03324	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	11520	277.24	62.37	160	-40	1.023361	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	11521	677.41	27.58	520	-120	1.024274	100	1	1400	400	0	0	0	0	0	0	0	0	0	0	0;
	11523	48.33	15.81	40	-10	1.031502	100	1	100	32.22	0	0	0	0	0	0	0	0	0	0	0;
	11524	48.33	15.81	40	-10	1.031502	100	1	100	32.22	0	0	0	0	0	0	0	0	0	0	0;
	11525	48.33	15.81	40	-10	1.031502	100	1	100	32.22	0	0	0	0	0	0	0	0	0	0	0;
	11526	48.33	15.81	40	-10	1.031502	100	1	100	32.22	0	0	0	0	0	0	0	0	0	0	0;
	11527	55.69	15.93	40	-10	1.03122	100	1	200	37.13	0	0	0	0	0	0	0	0	0	0	0;
	11528	55.69	15.93	40	-10	1.03122	100	1	200	37.13	0	0	0	0	0	0	0	0	0	0	0;
	11529	33.64	5.18	30	-10	1.007216	100	1	100	22.43	0	0	0	0	0	0	0	0	0	0	0;
	11530	55.69	15.93	40	-10	1.03122	100	1	200	37.13	0	0	0	0	0	0	0	0	0	0	0;
	11536	277.24	62.37	160	-40	1.023361	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	11539	111.5	6.79	90	-60	1.056762	100	1	300	74.33	0	0	0	0	0	0	0	0	0	0	0;
	11549	88.2	-9.06	50	-50	0.995145	100	1	200	58.8	0	0	0	0	0	0	0	0	0	0	0;
	11551	32.98	7.31	30	-10	1.054159	100	1	100	21.99	0	0	0	0	0	0	0	0	0	0	0;
	11558	72.47	11.44	60	-20	1.019231	100	1	200	48.31	0	0	0	0	0	0	0	0	0	0	0;
	11564	321.41	105.18	210	-50	1.033947	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	11568	69.35	21.13	60	-20	1.032238	100	1	200	46.23	0	0	0	0	0	0	0	0	0	0	0;
	11569	69.35	21.13	60	-20	1.032238	100	1	200	46.23	0	0	0	0	0	0	0	0	0	0	0;
	11570	53.01	8.19	40	-10	1.026319	100	1	200	35.34	0	0	0	0	0	0	0	0	0	0	0;
	11577	20.33	6.46	30	-10	0.990908	100	1	100	13.55	0	0	0	0	0	0	0	0	0	0	0;
	11578	28.41	6.81	30	-10	1.016999	100	1	100	18.94	0	0	0	0	0	0	0	0	0	0	0;
	11579	21.69	5.19	30	-10	1.050987	100	1	100	14.46	0	0	0	0	0	0	0	0	0	0	0;
	11580	54.06	11.61	40	-10	1.023809	100	1	200	36.04	0	0	0	0	0	0	0	0	0	0	0;
	11581	25.44	-2.35	20	-20	0.995022	100	1	100	16.96	0	0	0	0	0	0	0	0	0	0	0;
	11584	9.1	-0.32	10	-10	1.00419	100	1	100	6.07	0	0	0	0	0	0	0	0	0	0	0;
	11586	229.56	74.98	210	-50	1.052394	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	11588	23.61	14.88	30	-10	1.029344	100	1	100	15.74	0	0	0	0	0	0	0	0	0	0	0;
	11599	31.59	4.82	30	-10	1.045359	100	1	100	21.06	0	0	0	0	0	0	0	0	0	0	0;
	11600	39.86	5.41	30	-10	1.031054	100	1	100	26.57	0	0	0	0	0	0	0	0	0	0	0;
	11602	31.59	4.82	30	-10	1.045359	100	1	100	21.06	0	0	0	0	0	0	0	0	0	0	0;
	11603	28.46	8.42	30	-10	1.022352	100	1	100	18.97	0	0	0	0	0	0	0	0	0	0	0;
	11604	28.46	8.42	30	-10	1.022352	100	1	100	18.97	0	0	0	0	0	0	0	0	0	0	0;
	11605	39.86	5.41	30	-10	1.031054	100	1	100	26.57	0	0	0	0	0	0	0	0	0	0	0;
	11606	39.86	5.41	30	-10	1.031054	100	1	100	26.57	0	0	0	0	0	0	0	0	0	0	0;
	11611	28.97	5.47	30	-10	1.043059	100	1	100	19.31	0	0	0	0	0	0	0	0	0	0	0;
	11617	33.06	9.48	30	-10	1.009173	100	1	100	22.04	0	0	0	0	0	0	0	0	0	0	0;
	11619	237.72	5.93	160	-40	1.041353	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	11623	33.06	9.48	30	-10	1.009173	100	1	100	22.04	0	0	0	0	0	0	0	0	0	0	0;
	11624	33.06	9.48	30	-10	1.009173	100	1	100	22.04	0	0	0	0	0	0	0	0	0	0	0;
	11625	34.65	5.41	30	-10	1.014865	100	1	100	23.1	0	0	0	0	0	0	0	0	0	0	0;
	11626	34.65	5.41	30	-10	1.014865	100	1	100	23.1	0	0	0	0	0	0	0	0	0	0	0;
	11630	44.8	3.17	40	-20	0.991989	100	1	100	29.87	0	0	0	0	0	0	0	0	0	0	0;
	11631	28.74	18.23	30	-10	1.050944	100	1	100	19.16	0	0	0	0	0	0	0	0	0	0	0;
	11632	28.49	18.25	30	-10	1.0508	100	1	100	18.99	0	0	0	0	0	0	0	0	0	0	0;
	11633	32.46	19.3	30	-10	1.04927	100	1	100	21.64	0	0	0	0	0	0	0	0	0	0	0;
	11634	28.74	18.23	30	-10	1.050944	100	1	100	19.16	0	0	0	0	0	0	0	0	0	0	0;
	11635	28.49	18.25	30	-10	1.0508	100	1	100	18.99	0	0	0	0	0	0	0	0	0	0	0;
	11638	20.38	9.61	30	-10	1.012679	100	1	100	13.59	0	0	0	0	0	0	0	0	0	0	0;
	11641	8.2	-0.77	10	-10	1.00746	100	1	100	5.47	0	0	0	0	0	0	0	0	0	0	0;
	11643	43.29	23.54	60	-20	1.035771	100	1	100	28.86	0	0	0	0	0	0	0	0	0	0	0;
	11648	36.42	7.96	30	-10	1.029166	100	1	100	24.28	0	0	0	0	0	0	0	0	0	0	0;
	11649	277.21	61.8	160	-40	1.023418	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	11652	29.83	8.08	30	-10	1.015374	100	1	100	19.89	0	0	0	0	0	0	0	0	0	0	0;
	11654	31.92	6.07	30	-10	1.027264	100	1	100	21.28	0	0	0	0	0	0	0	0	0	0	0;
	11665	153.3	-6.31	80	-60	0.947933	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	11666	57.11	19.46	60	-20	1.032102	100	1	200	38.07	0	0	0	0	0	0	0	0	0	0	0;
	11667	57.11	19.46	60	-20	1.032102	100	1	200	38.07	0	0	0	0	0	0	0	0	0	0	0;
	11668	57.11	19.46	60	-20	1.032102	100	1	200	38.07	0	0	0	0	0	0	0	0	0	0	0;
	11669	57.11	19.46	60	-20	1.032102	100	1	200	38.07	0	0	0	0	0	0	0	0	0	0	0;
	11670	57.11	19.46	60	-20	1.032102	100	1	200	38.07	0	0	0	0	0	0	0	0	0	0	0;
	11672	0.02	1.34	10	-10	1.018847	100	1	100	0.01	0	0	0	0	0	0	0	0	0	0	0;
	11677	63.63	44.69	60	-20	1.058213	100	1	200	42.42	0	0	0	0	0	0	0	0	0	0	0;
	11680	319.18	33.68	210	-50	1.026027	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	11686	18.67	5.94	30	-10	1.015505	100	1	100	12.45	0	0	0	0	0	0	0	0	0	0	0;
	11687	38.91	13.65	30	-10	1.01946	100	1	100	25.94	0	0	0	0	0	0	0	0	0	0	0;
	11689	24.03	8.37	30	-10	1.005775	100	1	100	16.02	0	0	0	0	0	0	0	0	0	0	0;
	11690	24.03	8.37	30	-10	1.005775	100	1	100	16.02	0	0	0	0	0	0	0	0	0	0	0;
	11691	24.03	8.37	30	-10	1.005775	100	1	100	16.02	0	0	0	0	0	0	0	0	0	0	0;
	11692	24.03	8.37	30	-10	1.005775	100	1	100	16.02	0	0	0	0	0	0	0	0	0	0	0;
	11693	24.03	8.37	30	-10	1.005775	100	1	100	16.02	0	0	0	0	0	0	0	0	0	0	0;
	11694	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	11695	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	11699	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	11704	299.72	120.65	160	-40	1.047549	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	11705	13.26	-0.68	10	-10	1.003732	100	1	100	8.84	0	0	0	0	0	0	0	0	0	0	0;
	11713	223.39	-18.71	160	-40	1.0182	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	11714	223.39	-18.71	160	-40	1.0182	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	11716	31.47	4.94	30	-10	1.024861	100	1	100	20.98	0	0	0	0	0	0	0	0	0	0	0;
	11717	223.02	66.59	160	-40	1.042492	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	11723	25.54	8.34	30	-10	1.050494	100	1	100	17.03	0	0	0	0	0	0	0	0	0	0	0;
	11729	40.49	8.06	30	-10	1.036966	100	1	100	26.99	0	0	0	0	0	0	0	0	0	0	0;
	11736	99.76	14.18	60	-20	1.046239	100	1	200	66.51	0	0	0	0	0	0	0	0	0	0	0;
	11745	60.81	1.23	60	-20	1.039251	100	1	200	40.54	0	0	0	0	0	0	0	0	0	0	0;
	11749	55.22	8.96	40	-10	1.017828	100	1	200	36.81	0	0	0	0	0	0	0	0	0	0	0;
	11752	40.51	5.96	30	-10	1.017287	100	1	100	27.01	0	0	0	0	0	0	0	0	0	0	0;
	11755	40.97	7.11	40	-10	1.019643	100	1	100	27.31	0	0	0	0	0	0	0	0	0	0	0;
	11756	40.97	7.11	40	-10	1.019643	100	1	100	27.31	0	0	0	0	0	0	0	0	0	0	0;
	11757	40.97	7.11	40	-10	1.019643	100	1	100	27.31	0	0	0	0	0	0	0	0	0	0	0;
	11758	50.33	11.56	40	-10	1.037142	100	1	200	33.55	0	0	0	0	0	0	0	0	0	0	0;
	11759	50.33	11.56	40	-10	1.037142	100	1	200	33.55	0	0	0	0	0	0	0	0	0	0	0;
	11760	50.33	11.56	40	-10	1.037142	100	1	200	33.55	0	0	0	0	0	0	0	0	0	0	0;
	11761	25.35	8.5	40	-10	1.018631	100	1	100	16.9	0	0	0	0	0	0	0	0	0	0	0;
	11762	50.33	11.56	40	-10	1.037142	100	1	200	33.55	0	0	0	0	0	0	0	0	0	0	0;
	11763	50.33	11.56	40	-10	1.037142	100	1	200	33.55	0	0	0	0	0	0	0	0	0	0	0;
	11768	50.4	3.24	30	-30	1.023993	100	1	200	33.6	0	0	0	0	0	0	0	0	0	0	0;
	11773	31.57	7.72	30	-10	1.008948	100	1	100	21.05	0	0	0	0	0	0	0	0	0	0	0;
	11781	0.96	0.09	10	-10	1.03053	100	1	100	0.64	0	0	0	0	0	0	0	0	0	0	0;
	11786	323.35	77.76	210	-50	1.058114	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	11789	6.52	0.05	10	-10	0.996117	100	1	100	4.35	0	0	0	0	0	0	0	0	0	0	0;
	11796	46.26	7.53	40	-10	1.015006	100	1	100	30.84	0	0	0	0	0	0	0	0	0	0	0;
	11797	262.9	48.53	160	-40	1.027461	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	11803	59.71	48.45	60	-20	1.060489	100	1	200	39.81	0	0	0	0	0	0	0	0	0	0	0;
	11805	42.35	6.03	40	-10	1.030832	100	1	100	28.23	0	0	0	0	0	0	0	0	0	0	0;
	11806	45.38	12.31	60	-20	0.990977	100	1	100	30.25	0	0	0	0	0	0	0	0	0	0	0;
	11810	58.52	23.33	40	-10	1.023397	100	1	200	39.01	0	0	0	0	0	0	0	0	0	0	0;
	11811	57.02	14.41	40	-10	1.026579	100	1	200	38.01	0	0	0	0	0	0	0	0	0	0	0;
	11813	22.46	14.34	40	-10	1.024687	100	1	100	14.97	0	0	0	0	0	0	0	0	0	0	0;
	11814	22.46	14.34	40	-10	1.024687	100	1	100	14.97	0	0	0	0	0	0	0	0	0	0	0;
	11815	57.02	14.41	40	-10	1.026579	100	1	200	38.01	0	0	0	0	0	0	0	0	0	0	0;
	11816	57.02	14.41	40	-10	1.026579	100	1	200	38.01	0	0	0	0	0	0	0	0	0	0	0;
	11817	57.51	17.72	40	-10	1.029323	100	1	200	38.34	0	0	0	0	0	0	0	0	0	0	0;
	11818	57.51	17.72	40	-10	1.029323	100	1	200	38.34	0	0	0	0	0	0	0	0	0	0	0;
	11819	22.46	14.34	40	-10	1.024687	100	1	100	14.97	0	0	0	0	0	0	0	0	0	0	0;
	11820	22.46	14.34	40	-10	1.024687	100	1	100	14.97	0	0	0	0	0	0	0	0	0	0	0;
	11822	34.86	4.81	30	-10	1.050901	100	1	100	23.24	0	0	0	0	0	0	0	0	0	0	0;
	11825	-172.6	2.92	20	-20	1.024059	100	1	100	-230.13	0	0	0	0	0	0	0	0	0	0	0;
	11849	97.12	8.38	60	-20	1.040009	100	1	200	64.75	0	0	0	0	0	0	0	0	0	0	0;
	11850	8.2	-2.56	10	-10	0.955439	100	1	100	5.47	0	0	0	0	0	0	0	0	0	0	0;
	11851	300	86.58	210	-50	1.056682	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	11853	233.91	82.52	210	-50	1.057062	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	11854	300	86.58	210	-50	1.056682	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	11856	201.8	24.53	150	-70	1.021695	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	11859	5	2.04	10	-10	1.009821	100	1	100	3.33	0	0	0	0	0	0	0	0	0	0	0;
	11875	32.72	7.26	30	-10	1.016616	100	1	100	21.81	0	0	0	0	0	0	0	0	0	0	0;
	11880	24.99	2.31	30	-10	1.010466	100	1	100	16.66	0	0	0	0	0	0	0	0	0	0	0;
	11886	28.39	6.61	40	-10	1.013429	100	1	100	18.93	0	0	0	0	0	0	0	0	0	0	0;
	11889	206.37	113.42	160	-40	1.061603	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	11891	35.02	1.75	30	-10	1.038438	100	1	100	23.35	0	0	0	0	0	0	0	0	0	0	0;
	11892	40.51	5.96	30	-10	1.017287	100	1	100	27.01	0	0	0	0	0	0	0	0	0	0	0;
	11894	87.19	27.18	60	-20	1.023338	100	1	200	58.13	0	0	0	0	0	0	0	0	0	0	0;
	11895	50.92	11.1	40	-10	1.010998	100	1	200	33.95	0	0	0	0	0	0	0	0	0	0	0;
	11897	206.37	113.42	160	-40	1.061603	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	11898	256.4	117.4	160	-40	1.062536	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	11900	23.58	-5	20	-20	0.964368	100	1	100	15.72	0	0	0	0	0	0	0	0	0	0	0;
	11903	3.02	0.33	10	-10	1.03013	100	1	100	2.02	0	0	0	0	0	0	0	0	0	0	0;
	11909	938.69	9.52	520	-120	1.028568	100	1	1900	600	0	0	0	0	0	0	0	0	0	0	0;
	11913	39.03	10.16	30	-10	1.033359	100	1	100	26.02	0	0	0	0	0	0	0	0	0	0	0;
	11914	39.03	10.16	30	-10	1.033359	100	1	100	26.02	0	0	0	0	0	0	0	0	0	0	0;
	11916	35.6	15.17	30	-10	1.026957	100	1	100	23.73	0	0	0	0	0	0	0	0	0	0	0;
	11919	35.6	15.17	30	-10	1.026957	100	1	100	23.73	0	0	0	0	0	0	0	0	0	0	0;
	11920	31.3	-0.89	20	-20	0.998951	100	1	100	20.87	0	0	0	0	0	0	0	0	0	0	0;
	11922	28.89	4.69	30	-10	1.029382	100	1	100	19.26	0	0	0	0	0	0	0	0	0	0	0;
	11924	95.51	37.64	60	-20	1.0608	100	1	200	63.67	0	0	0	0	0	0	0	0	0	0	0;
	11927	992	834.49	100000	-100000	1.058046	100	1	2000	600	0	0	0	0	0	0	0	0	0	0	0;
	11935	55.25	2.21	40	-40	0.994234	100	1	200	36.83	0	0	0	0	0	0	0	0	0	0	0;
	11945	10.36	0.68	10	-10	1.012893	100	1	100	6.91	0	0	0	0	0	0	0	0	0	0	0;
	11946	78.71	19.55	60	-20	1.033214	100	1	200	52.47	0	0	0	0	0	0	0	0	0	0	0;
	11950	285.52	89.66	210	-50	1.04459	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	11953	64.18	-6.8	40	-40	0.951565	100	1	200	42.79	0	0	0	0	0	0	0	0	0	0	0;
	11959	78.71	19.51	60	-20	1.033214	100	1	200	52.47	0	0	0	0	0	0	0	0	0	0	0;
	11962	78.71	19.51	60	-20	1.033214	100	1	200	52.47	0	0	0	0	0	0	0	0	0	0	0;
	11967	5.2	1.64	10	-10	1.019503	100	1	100	3.47	0	0	0	0	0	0	0	0	0	0	0;
	11968	95.02	8.09	60	-20	1.023144	100	1	200	63.35	0	0	0	0	0	0	0	0	0	0	0;
	11970	8.4	-0.55	10	-10	0.988112	100	1	100	5.6	0	0	0	0	0	0	0	0	0	0	0;
	11975	68.3	12.27	60	-20	1.015227	100	1	200	45.53	0	0	0	0	0	0	0	0	0	0	0;
	11976	47.29	6.72	40	-10	1.04826	100	1	100	31.53	0	0	0	0	0	0	0	0	0	0	0;
	11978	93.77	48.62	60	-20	1.054388	100	1	200	62.51	0	0	0	0	0	0	0	0	0	0	0;
	11979	93.77	48.62	60	-20	1.054388	100	1	200	62.51	0	0	0	0	0	0	0	0	0	0	0;
	11980	93.77	48.62	60	-20	1.054388	100	1	200	62.51	0	0	0	0	0	0	0	0	0	0	0;
	11982	277.26	31.82	210	-50	1.04876	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	11986	48.42	9.78	40	-10	1.01601	100	1	100	32.28	0	0	0	0	0	0	0	0	0	0	0;
	11987	290	-23.36	100	-100	0.971901	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	11988	36.91	13.33	30	-10	1.037973	100	1	100	24.61	0	0	0	0	0	0	0	0	0	0	0;
	11990	19.75	-0.1	20	-20	1.002465	100	1	100	13.17	0	0	0	0	0	0	0	0	0	0	0;
	11996	35.01	1.48	30	-10	1.032132	100	1	100	23.34	0	0	0	0	0	0	0	0	0	0	0;
	12001	168	14.06	130	-80	1.016606	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	12013	156.71	14.75	90	-20	1.008914	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	12016	214.52	110.59	210	-50	1.065343	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12020	26.22	3.21	30	-10	1.015252	100	1	100	17.48	0	0	0	0	0	0	0	0	0	0	0;
	12021	43.5	8.1	30	-10	1.018676	100	1	100	29	0	0	0	0	0	0	0	0	0	0	0;
	12024	77.25	12.22	60	-20	1.034048	100	1	200	51.5	0	0	0	0	0	0	0	0	0	0	0;
	12025	53.45	18.98	40	-10	1.030499	100	1	200	35.63	0	0	0	0	0	0	0	0	0	0	0;
	12027	20.4	2.66	20	-10	1.042665	100	1	100	13.6	0	0	0	0	0	0	0	0	0	0	0;
	12035	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	12042	21.72	6.92	30	-10	1.014428	100	1	100	14.48	0	0	0	0	0	0	0	0	0	0	0;
	12045	40.16	7.07	30	-10	1.013722	100	1	100	26.77	0	0	0	0	0	0	0	0	0	0	0;
	12049	37.96	13.4	30	-10	1.073911	100	1	100	25.31	0	0	0	0	0	0	0	0	0	0	0;
	12050	135	10.29	70	-20	1.008348	100	1	300	90	0	0	0	0	0	0	0	0	0	0	0;
	12060	4.8	-0.11	10	-10	0.996411	100	1	100	3.2	0	0	0	0	0	0	0	0	0	0	0;
	12065	224.46	15.93	210	-50	1.031165	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12069	34.37	6.67	30	-10	1.035406	100	1	100	22.91	0	0	0	0	0	0	0	0	0	0	0;
	12070	34.37	6.67	30	-10	1.035406	100	1	100	22.91	0	0	0	0	0	0	0	0	0	0	0;
	12071	34.37	6.67	30	-10	1.035406	100	1	100	22.91	0	0	0	0	0	0	0	0	0	0	0;
	12072	96.57	34.4	60	-20	1.051326	100	1	200	64.38	0	0	0	0	0	0	0	0	0	0	0;
	12073	34.37	6.67	30	-10	1.035406	100	1	100	22.91	0	0	0	0	0	0	0	0	0	0	0;
	12074	34.37	6.67	30	-10	1.035406	100	1	100	22.91	0	0	0	0	0	0	0	0	0	0	0;
	12075	34.37	6.67	30	-10	1.035406	100	1	100	22.91	0	0	0	0	0	0	0	0	0	0	0;
	12076	34.37	6.67	30	-10	1.035406	100	1	100	22.91	0	0	0	0	0	0	0	0	0	0	0;
	12084	60.08	4.3	60	-20	1.012034	100	1	200	40.05	0	0	0	0	0	0	0	0	0	0	0;
	12085	60.08	4.3	60	-20	1.012034	100	1	200	40.05	0	0	0	0	0	0	0	0	0	0	0;
	12088	28.66	14.89	30	-10	1.014269	100	1	100	19.11	0	0	0	0	0	0	0	0	0	0	0;
	12091	31.79	7.29	30	-10	1.005112	100	1	100	21.19	0	0	0	0	0	0	0	0	0	0	0;
	12093	91.55	46.19	60	-20	1.056125	100	1	200	61.03	0	0	0	0	0	0	0	0	0	0	0;
	12095	95.01	14.99	60	-20	1.024647	100	1	200	63.34	0	0	0	0	0	0	0	0	0	0	0;
	12097	92.69	14.48	60	-20	1.026207	100	1	200	61.79	0	0	0	0	0	0	0	0	0	0	0;
	12099	16.71	0.01	30	-10	1.030988	100	1	100	11.14	0	0	0	0	0	0	0	0	0	0	0;
	12102	93.64	6.04	60	-20	1.023818	100	1	200	62.43	0	0	0	0	0	0	0	0	0	0	0;
	12106	39.88	15.15	30	-10	1.024267	100	1	100	26.59	0	0	0	0	0	0	0	0	0	0	0;
	12116	42.33	15.72	40	-10	1.031695	100	1	100	28.22	0	0	0	0	0	0	0	0	0	0	0;
	12121	19.26	2.33	30	-10	1.049756	100	1	100	12.84	0	0	0	0	0	0	0	0	0	0	0;
	12123	4.5	-0.83	10	-10	1.003288	100	1	100	3	0	0	0	0	0	0	0	0	0	0	0;
	12124	18.74	3.61	30	-10	1.025271	100	1	100	12.49	0	0	0	0	0	0	0	0	0	0	0;
	12126	29.14	3.85	30	-10	1.025372	100	1	100	19.43	0	0	0	0	0	0	0	0	0	0	0;
	12127	29.14	3.85	30	-10	1.025372	100	1	100	19.43	0	0	0	0	0	0	0	0	0	0	0;
	12131	83.47	36.43	60	-20	1.039883	100	1	200	55.65	0	0	0	0	0	0	0	0	0	0	0;
	12132	83.47	36.43	60	-20	1.039883	100	1	200	55.65	0	0	0	0	0	0	0	0	0	0	0;
	12133	90.97	23.14	60	-20	1.033722	100	1	200	60.65	0	0	0	0	0	0	0	0	0	0	0;
	12134	90.97	23.14	60	-20	1.033722	100	1	200	60.65	0	0	0	0	0	0	0	0	0	0	0;
	12135	90.97	23.14	60	-20	1.033722	100	1	200	60.65	0	0	0	0	0	0	0	0	0	0	0;
	12136	90.97	23.14	60	-20	1.033722	100	1	200	60.65	0	0	0	0	0	0	0	0	0	0	0;
	12137	90.94	17.93	60	-20	1.033736	100	1	200	60.63	0	0	0	0	0	0	0	0	0	0	0;
	12138	90.94	17.93	60	-20	1.033736	100	1	200	60.63	0	0	0	0	0	0	0	0	0	0	0;
	12139	90.94	17.93	60	-20	1.033736	100	1	200	60.63	0	0	0	0	0	0	0	0	0	0	0;
	12140	90.94	17.93	60	-20	1.033736	100	1	200	60.63	0	0	0	0	0	0	0	0	0	0	0;
	12152	91.34	21.91	60	-20	1.030656	100	1	200	60.89	0	0	0	0	0	0	0	0	0	0	0;
	12163	17.6	22.53	30	-10	1.05667	100	1	100	11.73	0	0	0	0	0	0	0	0	0	0	0;
	12164	17.6	22.53	30	-10	1.05667	100	1	100	11.73	0	0	0	0	0	0	0	0	0	0	0;
	12166	30.64	2.43	30	-10	1.030179	100	1	100	20.43	0	0	0	0	0	0	0	0	0	0	0;
	12167	30.64	2.43	30	-10	1.030179	100	1	100	20.43	0	0	0	0	0	0	0	0	0	0	0;
	12168	30.64	2.43	30	-10	1.030179	100	1	100	20.43	0	0	0	0	0	0	0	0	0	0	0;
	12169	30.64	2.43	30	-10	1.030179	100	1	100	20.43	0	0	0	0	0	0	0	0	0	0	0;
	12170	17.6	22.53	30	-10	1.05667	100	1	100	11.73	0	0	0	0	0	0	0	0	0	0	0;
	12171	17.6	22.53	30	-10	1.05667	100	1	100	11.73	0	0	0	0	0	0	0	0	0	0	0;
	12172	30.64	2.43	30	-10	1.030179	100	1	100	20.43	0	0	0	0	0	0	0	0	0	0	0;
	12173	800.47	18.87	520	-120	1.04071	100	1	1700	500	0	0	0	0	0	0	0	0	0	0	0;
	12175	30.93	8.23	30	-10	1.035888	100	1	100	20.62	0	0	0	0	0	0	0	0	0	0	0;
	12178	49.1	2.32	40	-30	1.014358	100	1	100	32.73	0	0	0	0	0	0	0	0	0	0	0;
	12189	18.1	0.8	10	-10	1.007227	100	1	100	12.07	0	0	0	0	0	0	0	0	0	0	0;
	12191	23.7	5.77	30	-10	1.013974	100	1	100	15.8	0	0	0	0	0	0	0	0	0	0	0;
	12192	23.7	5.77	30	-10	1.013974	100	1	100	15.8	0	0	0	0	0	0	0	0	0	0	0;
	12193	18.91	10.94	30	-10	1.015043	100	1	100	12.61	0	0	0	0	0	0	0	0	0	0	0;
	12194	18.91	10.94	30	-10	1.015043	100	1	100	12.61	0	0	0	0	0	0	0	0	0	0	0;
	12195	21.17	9.91	30	-10	1.048748	100	1	100	14.11	0	0	0	0	0	0	0	0	0	0	0;
	12199	210	37.99	140	-20	0.991398	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12200	29.92	13.61	30	-10	1.031553	100	1	100	19.95	0	0	0	0	0	0	0	0	0	0	0;
	12204	298.7	3.31	190	-120	0.975834	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	12208	382.77	9.72	210	-50	1.043874	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	12211	319.18	33.68	210	-50	1.026027	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	12216	37.08	5.51	30	-10	1.021005	100	1	100	24.72	0	0	0	0	0	0	0	0	0	0	0;
	12217	40.49	8.06	30	-10	1.036966	100	1	100	26.99	0	0	0	0	0	0	0	0	0	0	0;
	12219	22.46	14.75	30	-10	1.043942	100	1	100	14.97	0	0	0	0	0	0	0	0	0	0	0;
	12221	3.13	0.12	10	-10	1.029545	100	1	100	2.09	0	0	0	0	0	0	0	0	0	0	0;
	12222	35.86	4.09	30	-10	1.02494	100	1	100	23.91	0	0	0	0	0	0	0	0	0	0	0;
	12227	38.02	11.21	30	-10	1.043087	100	1	100	25.35	0	0	0	0	0	0	0	0	0	0	0;
	12232	89.93	16.64	60	-20	1.037993	100	1	200	59.95	0	0	0	0	0	0	0	0	0	0	0;
	12240	51.04	7.37	30	-10	1.022424	100	1	200	34.03	0	0	0	0	0	0	0	0	0	0	0;
	12245	30.24	6.83	30	-10	1.037281	100	1	100	20.16	0	0	0	0	0	0	0	0	0	0	0;
	12248	261.4	61.75	100	-30	0.988817	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	12252	227.21	72.31	210	-50	1.063358	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12255	323.89	98.77	210	-50	1.062762	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	12259	227.21	72.6	210	-50	1.06345	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12266	155.5	12.42	100	-70	1.034282	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	12275	83.29	7.43	60	-20	1.042396	100	1	200	55.53	0	0	0	0	0	0	0	0	0	0	0;
	12276	83.29	7.43	60	-20	1.042396	100	1	200	55.53	0	0	0	0	0	0	0	0	0	0	0;
	12277	6.1	0.22	10	-10	1.00107	100	1	100	4.07	0	0	0	0	0	0	0	0	0	0	0;
	12278	79.77	4.15	60	-20	1.047117	100	1	200	53.18	0	0	0	0	0	0	0	0	0	0	0;
	12279	79.77	4.15	60	-20	1.047117	100	1	200	53.18	0	0	0	0	0	0	0	0	0	0	0;
	12280	83.29	7.43	60	-20	1.042396	100	1	200	55.53	0	0	0	0	0	0	0	0	0	0	0;
	12281	79.77	4.15	60	-20	1.047117	100	1	200	53.18	0	0	0	0	0	0	0	0	0	0	0;
	12282	37.23	4.01	30	-10	1.028283	100	1	100	24.82	0	0	0	0	0	0	0	0	0	0	0;
	12285	3.2	1.16	10	-10	0.991765	100	1	100	2.13	0	0	0	0	0	0	0	0	0	0	0;
	12286	55.25	0.75	30	-30	0.959658	100	1	200	36.83	0	0	0	0	0	0	0	0	0	0	0;
	12287	159.8	33.64	90	-40	0.993125	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	12291	240.6	61.33	100	-30	0.990464	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12302	40.71	13.63	40	-10	1.028425	100	1	100	27.14	0	0	0	0	0	0	0	0	0	0	0;
	12303	40.71	13.63	40	-10	1.028425	100	1	100	27.14	0	0	0	0	0	0	0	0	0	0	0;
	12304	40.71	13.63	40	-10	1.028425	100	1	100	27.14	0	0	0	0	0	0	0	0	0	0	0;
	12306	37.35	7.36	40	-10	1.026363	100	1	100	24.9	0	0	0	0	0	0	0	0	0	0	0;
	12307	40.71	13.63	40	-10	1.028425	100	1	100	27.14	0	0	0	0	0	0	0	0	0	0	0;
	12310	34.96	12.97	30	-10	1.054937	100	1	100	23.31	0	0	0	0	0	0	0	0	0	0	0;
	12319	52.29	8.76	40	-10	1.021414	100	1	200	34.86	0	0	0	0	0	0	0	0	0	0	0;
	12321	34.86	4.81	30	-10	1.050901	100	1	100	23.24	0	0	0	0	0	0	0	0	0	0	0;
	12325	52.29	8.76	40	-10	1.021414	100	1	200	34.86	0	0	0	0	0	0	0	0	0	0	0;
	12330	63.72	11.02	40	-10	1.030356	100	1	200	42.48	0	0	0	0	0	0	0	0	0	0	0;
	12332	63.72	11.02	40	-10	1.030356	100	1	200	42.48	0	0	0	0	0	0	0	0	0	0	0;
	12333	63.72	11.02	40	-10	1.030356	100	1	200	42.48	0	0	0	0	0	0	0	0	0	0	0;
	12334	63.72	11.02	40	-10	1.030356	100	1	200	42.48	0	0	0	0	0	0	0	0	0	0	0;
	12337	39.88	8.05	40	-10	1.030375	100	1	100	26.59	0	0	0	0	0	0	0	0	0	0	0;
	12338	39.88	8.05	40	-10	1.030375	100	1	100	26.59	0	0	0	0	0	0	0	0	0	0	0;
	12341	77.25	12.22	60	-20	1.034048	100	1	200	51.5	0	0	0	0	0	0	0	0	0	0	0;
	12346	325.03	139.06	210	-50	1.059619	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	12348	289.92	105.36	210	-50	1.057338	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	12358	10	0.51	10	-10	1.016383	100	1	100	6.67	0	0	0	0	0	0	0	0	0	0	0;
	12359	6.1	0.22	10	-10	1.00107	100	1	100	4.07	0	0	0	0	0	0	0	0	0	0	0;
	12367	210	16.99	130	-60	0.981513	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12373	313.39	68.14	210	-50	1.035914	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	12375	126	-10.22	90	-60	1.007198	100	1	300	84	0	0	0	0	0	0	0	0	0	0	0;
	12381	19.27	11.16	30	-10	1.021475	100	1	100	12.85	0	0	0	0	0	0	0	0	0	0	0;
	12383	52.88	1.71	30	-20	1.017202	100	1	200	35.25	0	0	0	0	0	0	0	0	0	0	0;
	12387	24.98	14.2	30	-10	1.031726	100	1	100	16.65	0	0	0	0	0	0	0	0	0	0	0;
	12388	57.12	17.95	40	-10	1.017464	100	1	200	38.08	0	0	0	0	0	0	0	0	0	0	0;
	12390	224.68	-13.42	210	-50	1.039621	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12393	100	41.35	90	-50	1.090507	100	1	300	66.67	0	0	0	0	0	0	0	0	0	0	0;
	12398	44.1	5.46	50	-20	0.988831	100	1	100	29.4	0	0	0	0	0	0	0	0	0	0	0;
	12400	29.89	15.39	30	-10	1.024479	100	1	100	19.93	0	0	0	0	0	0	0	0	0	0	0;
	12401	99	2.05	50	-40	1.019049	100	1	200	66	0	0	0	0	0	0	0	0	0	0	0;
	12411	99.68	14.71	60	-20	1.035207	100	1	200	66.45	0	0	0	0	0	0	0	0	0	0	0;
	12415	227.37	10.01	160	-40	1.028414	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12419	25.89	9.83	40	-10	1.013208	100	1	100	17.26	0	0	0	0	0	0	0	0	0	0	0;
	12421	5	-1.11	10	-10	0.981334	100	1	100	3.33	0	0	0	0	0	0	0	0	0	0	0;
	12428	59.66	11.27	40	-10	1.027725	100	1	200	39.77	0	0	0	0	0	0	0	0	0	0	0;
	12429	59.66	11.27	40	-10	1.027725	100	1	200	39.77	0	0	0	0	0	0	0	0	0	0	0;
	12430	59.66	11.27	40	-10	1.027725	100	1	200	39.77	0	0	0	0	0	0	0	0	0	0	0;
	12431	59.66	11.27	40	-10	1.027725	100	1	200	39.77	0	0	0	0	0	0	0	0	0	0	0;
	12433	59.66	11.27	40	-10	1.027725	100	1	200	39.77	0	0	0	0	0	0	0	0	0	0	0;
	12434	49.88	7.11	40	-10	1.013219	100	1	100	33.25	0	0	0	0	0	0	0	0	0	0	0;
	12440	45.28	10.84	40	-10	1.043738	100	1	100	30.19	0	0	0	0	0	0	0	0	0	0	0;
	12441	7.32	2.25	10	-10	1.004719	100	1	100	4.88	0	0	0	0	0	0	0	0	0	0	0;
	12447	68.89	5.63	60	-20	1.028667	100	1	200	45.93	0	0	0	0	0	0	0	0	0	0	0;
	12448	40.49	8.06	30	-10	1.036966	100	1	100	26.99	0	0	0	0	0	0	0	0	0	0	0;
	12451	33.15	0.93	30	-10	1.024669	100	1	100	22.1	0	0	0	0	0	0	0	0	0	0	0;
	12452	240.2	86.73	160	-40	1.035311	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12453	240.2	86.86	160	-40	1.035255	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12454	230.51	50.58	160	-40	1.016573	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12455	68.89	5.63	60	-20	1.028667	100	1	200	45.93	0	0	0	0	0	0	0	0	0	0	0;
	12459	47.29	6.72	40	-10	1.04826	100	1	100	31.53	0	0	0	0	0	0	0	0	0	0	0;
	12461	88.02	46.54	60	-20	1.057713	100	1	200	58.68	0	0	0	0	0	0	0	0	0	0	0;
	12462	88.02	46.54	60	-20	1.057713	100	1	200	58.68	0	0	0	0	0	0	0	0	0	0	0;
	12464	88.02	46.54	60	-20	1.057713	100	1	200	58.68	0	0	0	0	0	0	0	0	0	0	0;
	12465	87.39	49.12	60	-20	1.05445	100	1	200	58.26	0	0	0	0	0	0	0	0	0	0	0;
	12466	88.02	46.54	60	-20	1.057713	100	1	200	58.68	0	0	0	0	0	0	0	0	0	0	0;
	12467	88.02	46.54	60	-20	1.057713	100	1	200	58.68	0	0	0	0	0	0	0	0	0	0	0;
	12469	87.39	49.12	60	-20	1.05445	100	1	200	58.26	0	0	0	0	0	0	0	0	0	0	0;
	12470	87.39	49.12	60	-20	1.05445	100	1	200	58.26	0	0	0	0	0	0	0	0	0	0	0;
	12477	25.74	8.25	30	-10	1.037214	100	1	100	17.16	0	0	0	0	0	0	0	0	0	0	0;
	12494	66.65	5.48	60	-20	1.01993	100	1	200	44.43	0	0	0	0	0	0	0	0	0	0	0;
	12499	12.15	5.1	10	0	1.040719	100	1	100	8.1	0	0	0	0	0	0	0	0	0	0	0;
	12500	31.2	5.73	30	-10	1.050684	100	1	100	20.8	0	0	0	0	0	0	0	0	0	0	0;
	12505	77.81	13.09	60	-20	1.029991	100	1	200	51.87	0	0	0	0	0	0	0	0	0	0	0;
	12506	77.81	13.09	60	-20	1.029991	100	1	200	51.87	0	0	0	0	0	0	0	0	0	0	0;
	12507	0.83	-0.07	10	-10	1.015733	100	1	100	0.55	0	0	0	0	0	0	0	0	0	0	0;
	12509	284.39	98.98	210	-50	1.060659	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	12517	54.05	11.46	40	-10	1.024254	100	1	200	36.03	0	0	0	0	0	0	0	0	0	0	0;
	12518	280.13	69.09	160	-40	1.028417	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	12519	96.75	44.18	60	-20	1.057091	100	1	200	64.5	0	0	0	0	0	0	0	0	0	0	0;
	12521	33.21	5.47	30	-10	1.022689	100	1	100	22.14	0	0	0	0	0	0	0	0	0	0	0;
	12522	33.21	5.47	30	-10	1.022689	100	1	100	22.14	0	0	0	0	0	0	0	0	0	0	0;
	12523	33.21	5.47	30	-10	1.022689	100	1	100	22.14	0	0	0	0	0	0	0	0	0	0	0;
	12525	224.13	67.46	160	-40	1.034235	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12526	224.13	67.46	160	-40	1.034235	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12528	26.62	13.25	30	-10	1.040642	100	1	100	17.75	0	0	0	0	0	0	0	0	0	0	0;
	12529	88.06	35.93	60	-20	1.024151	100	1	200	58.71	0	0	0	0	0	0	0	0	0	0	0;
	12537	649.26	-32.35	520	-120	1.035364	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	12541	0.76	0.04	10	-10	1.02993	100	1	100	0.51	0	0	0	0	0	0	0	0	0	0	0;
	12542	29.43	7.09	30	-10	1.028438	100	1	100	19.62	0	0	0	0	0	0	0	0	0	0	0;
	12551	17.8	10.21	30	-10	1.034631	100	1	100	11.87	0	0	0	0	0	0	0	0	0	0	0;
	12555	8.93	-0.3	10	-10	1.003834	100	1	100	5.95	0	0	0	0	0	0	0	0	0	0	0;
	12556	180.96	38.95	160	-40	1.029835	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	12558	6.03	-2.37	10	-10	0.971908	100	1	100	4.02	0	0	0	0	0	0	0	0	0	0	0;
	12563	355.56	-18.65	210	-50	1.020396	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	12570	20.63	6.57	30	-10	1.022447	100	1	100	13.75	0	0	0	0	0	0	0	0	0	0	0;
	12572	112	-3.86	80	-60	0.989707	100	1	300	74.67	0	0	0	0	0	0	0	0	0	0	0;
	12579	627	28.92	320	-190	0.960771	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	12582	38.56	8.03	30	-10	1.049495	100	1	100	25.71	0	0	0	0	0	0	0	0	0	0	0;
	12583	38.56	8.03	30	-10	1.049495	100	1	100	25.71	0	0	0	0	0	0	0	0	0	0	0;
	12586	42.41	17.68	60	-20	1.058287	100	1	100	28.27	0	0	0	0	0	0	0	0	0	0	0;
	12587	-310.36	-575.04	100000	-100000	1.016947	100	1	100	-413.82	0	0	0	0	0	0	0	0	0	0	0;
	12589	42.41	17.68	60	-20	1.058287	100	1	100	28.27	0	0	0	0	0	0	0	0	0	0	0;
	12590	42.41	17.68	60	-20	1.058287	100	1	100	28.27	0	0	0	0	0	0	0	0	0	0	0;
	12591	42.41	17.68	60	-20	1.058287	100	1	100	28.27	0	0	0	0	0	0	0	0	0	0	0;
	12592	42.41	17.68	60	-20	1.058287	100	1	100	28.27	0	0	0	0	0	0	0	0	0	0	0;
	12599	52.46	9.17	60	-20	1.033945	100	1	200	34.97	0	0	0	0	0	0	0	0	0	0	0;
	12600	52.46	9.17	60	-20	1.033945	100	1	200	34.97	0	0	0	0	0	0	0	0	0	0	0;
	12601	52.46	9.17	60	-20	1.033945	100	1	200	34.97	0	0	0	0	0	0	0	0	0	0	0;
	12603	153.3	-6.31	80	-60	0.947933	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	12606	228.19	49.03	210	-50	1.020332	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12607	61.89	15.66	40	-10	1.026221	100	1	200	41.26	0	0	0	0	0	0	0	0	0	0	0;
	12610	61.89	15.66	40	-10	1.026221	100	1	200	41.26	0	0	0	0	0	0	0	0	0	0	0;
	12612	16.71	0.01	30	-10	1.030988	100	1	100	11.14	0	0	0	0	0	0	0	0	0	0	0;
	12616	6.52	0.05	10	-10	0.996117	100	1	100	4.35	0	0	0	0	0	0	0	0	0	0	0;
	12625	167.4	-8.63	90	-70	0.948759	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	12636	23.91	3.61	30	-10	1.007821	100	1	100	15.94	0	0	0	0	0	0	0	0	0	0	0;
	12641	177.84	-23.15	160	-40	0.995435	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	12646	23.94	10.89	30	-10	1.042693	100	1	100	15.96	0	0	0	0	0	0	0	0	0	0	0;
	12647	85.48	16.57	60	-20	1.029333	100	1	200	56.99	0	0	0	0	0	0	0	0	0	0	0;
	12648	85.48	16.57	60	-20	1.029333	100	1	200	56.99	0	0	0	0	0	0	0	0	0	0	0;
	12650	85.48	16.57	60	-20	1.029333	100	1	200	56.99	0	0	0	0	0	0	0	0	0	0	0;
	12651	85.48	16.54	60	-20	1.029323	100	1	200	56.99	0	0	0	0	0	0	0	0	0	0	0;
	12652	85.48	16.54	60	-20	1.029323	100	1	200	56.99	0	0	0	0	0	0	0	0	0	0	0;
	12653	85.48	16.57	60	-20	1.029333	100	1	200	56.99	0	0	0	0	0	0	0	0	0	0	0;
	12654	85.48	16.54	60	-20	1.029323	100	1	200	56.99	0	0	0	0	0	0	0	0	0	0	0;
	12656	38.52	8.15	30	-10	1.04659	100	1	100	25.68	0	0	0	0	0	0	0	0	0	0	0;
	12657	5	0.46	10	-10	1.017738	100	1	100	3.33	0	0	0	0	0	0	0	0	0	0	0;
	12658	40	4.52	20	-20	0.956073	100	1	100	26.67	0	0	0	0	0	0	0	0	0	0	0;
	12659	54.32	19.31	40	-10	1.040793	100	1	200	36.21	0	0	0	0	0	0	0	0	0	0	0;
	12664	20.43	6.2	30	-10	1.037469	100	1	100	13.62	0	0	0	0	0	0	0	0	0	0	0;
	12665	419.21	7.57	260	-60	1.038308	100	1	900	200	0	0	0	0	0	0	0	0	0	0	0;
	12673	54.32	19.31	40	-10	1.040793	100	1	200	36.21	0	0	0	0	0	0	0	0	0	0	0;
	12678	100	41.35	90	-50	1.090507	100	1	300	66.67	0	0	0	0	0	0	0	0	0	0	0;
	12681	22.87	9.78	30	-10	1.014635	100	1	100	15.25	0	0	0	0	0	0	0	0	0	0	0;
	12687	30.45	8.8	40	-30	1.01812	100	1	100	20.3	0	0	0	0	0	0	0	0	0	0	0;
	12688	300	-14.29	100	-100	1.001419	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	12690	54.32	19.31	40	-10	1.040793	100	1	200	36.21	0	0	0	0	0	0	0	0	0	0	0;
	12692	18.9	0.51	10	-10	1.015529	100	1	100	12.6	0	0	0	0	0	0	0	0	0	0	0;
	12697	27.67	4.45	30	-10	1.027132	100	1	100	18.45	0	0	0	0	0	0	0	0	0	0	0;
	12699	27.67	4.45	30	-10	1.027132	100	1	100	18.45	0	0	0	0	0	0	0	0	0	0	0;
	12701	34.08	12.69	30	-10	1.041939	100	1	100	22.72	0	0	0	0	0	0	0	0	0	0	0;
	12702	27.67	4.45	30	-10	1.027132	100	1	100	18.45	0	0	0	0	0	0	0	0	0	0	0;
	12704	34.08	12.69	30	-10	1.041939	100	1	100	22.72	0	0	0	0	0	0	0	0	0	0	0;
	12705	34.27	11.44	30	-10	1.03975	100	1	100	22.85	0	0	0	0	0	0	0	0	0	0	0;
	12706	34.27	11.44	30	-10	1.03975	100	1	100	22.85	0	0	0	0	0	0	0	0	0	0	0;
	12708	15.86	9.49	30	-10	1.017238	100	1	100	10.57	0	0	0	0	0	0	0	0	0	0	0;
	12709	30.04	7.71	30	-10	1.011489	100	1	100	20.03	0	0	0	0	0	0	0	0	0	0	0;
	12711	31.76	6.93	20	-10	1.007801	100	1	100	21.17	0	0	0	0	0	0	0	0	0	0	0;
	12713	32.17	5.12	30	-10	1.009886	100	1	100	21.45	0	0	0	0	0	0	0	0	0	0	0;
	12714	15.86	9.49	30	-10	1.017238	100	1	100	10.57	0	0	0	0	0	0	0	0	0	0	0;
	12717	38.02	11.21	30	-10	1.043087	100	1	100	25.35	0	0	0	0	0	0	0	0	0	0	0;
	12723	27.72	3.43	30	-10	1.015411	100	1	100	18.48	0	0	0	0	0	0	0	0	0	0	0;
	12726	276.1	40.94	160	-40	1.035338	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	12727	85.28	21.31	60	-20	1.029542	100	1	200	56.85	0	0	0	0	0	0	0	0	0	0	0;
	12728	36.8	6.62	30	-10	1.03058	100	1	100	24.53	0	0	0	0	0	0	0	0	0	0	0;
	12730	70	0.5	40	-30	1.017293	100	1	200	46.67	0	0	0	0	0	0	0	0	0	0	0;
	12736	40	4.84	20	-20	0.957116	100	1	100	26.67	0	0	0	0	0	0	0	0	0	0	0;
	12738	32.97	4.46	30	-10	1.027968	100	1	100	21.98	0	0	0	0	0	0	0	0	0	0	0;
	12742	33.73	12.57	30	-10	1.035958	100	1	100	22.49	0	0	0	0	0	0	0	0	0	0	0;
	12747	33.73	12.57	30	-10	1.035958	100	1	100	22.49	0	0	0	0	0	0	0	0	0	0	0;
	12748	2.5	0.07	10	-10	0.9846	100	1	100	1.67	0	0	0	0	0	0	0	0	0	0	0;
	12758	95.51	37.64	60	-20	1.0608	100	1	200	63.67	0	0	0	0	0	0	0	0	0	0	0;
	12760	58.19	6.99	40	-10	1.028736	100	1	200	38.79	0	0	0	0	0	0	0	0	0	0	0;
	12761	58.19	6.99	40	-10	1.028736	100	1	200	38.79	0	0	0	0	0	0	0	0	0	0	0;
	12763	58.19	6.99	40	-10	1.028736	100	1	200	38.79	0	0	0	0	0	0	0	0	0	0	0;
	12764	69.92	41.17	60	-20	1.051947	100	1	200	46.61	0	0	0	0	0	0	0	0	0	0	0;
	12765	69.92	41.17	60	-20	1.051947	100	1	200	46.61	0	0	0	0	0	0	0	0	0	0	0;
	12772	364.88	71.7	210	-50	1.04706	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	12775	158.9	5.88	100	-30	0.968879	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	12777	25.64	1.96	20	-10	1.010323	100	1	100	17.09	0	0	0	0	0	0	0	0	0	0	0;
	12782	21.35	0.38	20	-10	1.033617	100	1	100	14.23	0	0	0	0	0	0	0	0	0	0	0;
	12786	56.25	11.12	40	-10	1.03036	100	1	200	37.5	0	0	0	0	0	0	0	0	0	0	0;
	12788	74.67	51.83	60	-20	1.045385	100	1	200	49.78	0	0	0	0	0	0	0	0	0	0	0;
	12789	74.67	51.83	60	-20	1.045385	100	1	200	49.78	0	0	0	0	0	0	0	0	0	0	0;
	12790	74.67	51.83	60	-20	1.045385	100	1	200	49.78	0	0	0	0	0	0	0	0	0	0	0;
	12791	74.67	51.83	60	-20	1.045385	100	1	200	49.78	0	0	0	0	0	0	0	0	0	0	0;
	12794	331.23	145.12	210	-50	1.041485	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	12799	50.96	2.69	60	-20	1.028648	100	1	200	33.97	0	0	0	0	0	0	0	0	0	0	0;
	12802	77.24	16.07	60	-20	1.027025	100	1	200	51.49	0	0	0	0	0	0	0	0	0	0	0;
	12803	78.62	8.98	60	-20	1.056333	100	1	200	52.41	0	0	0	0	0	0	0	0	0	0	0;
	12806	97.7	2.58	60	-20	1.053741	100	1	200	65.13	0	0	0	0	0	0	0	0	0	0	0;
	12807	97.7	2.58	60	-20	1.053741	100	1	200	65.13	0	0	0	0	0	0	0	0	0	0	0;
	12808	97.7	2.58	60	-20	1.053741	100	1	200	65.13	0	0	0	0	0	0	0	0	0	0	0;
	12814	96.57	34.4	60	-20	1.051326	100	1	200	64.38	0	0	0	0	0	0	0	0	0	0	0;
	12815	31	6.15	10	0	1.046995	100	1	100	20.67	0	0	0	0	0	0	0	0	0	0	0;
	12816	18.4	-1.01	20	-20	1.003342	100	1	100	12.27	0	0	0	0	0	0	0	0	0	0	0;
	12817	685.9	148.95	520	-120	1.062539	100	1	1400	400	0	0	0	0	0	0	0	0	0	0	0;
	12818	685.9	148.95	520	-120	1.062539	100	1	1400	400	0	0	0	0	0	0	0	0	0	0	0;
	12819	33.97	5.68	30	-10	1.022889	100	1	100	22.65	0	0	0	0	0	0	0	0	0	0	0;
	12822	36.03	16.08	30	-10	1.045359	100	1	100	24.02	0	0	0	0	0	0	0	0	0	0	0;
	12825	39.54	11.66	30	-10	1.055315	100	1	100	26.36	0	0	0	0	0	0	0	0	0	0	0;
	12829	34.96	5.47	30	-30	0.935194	100	1	100	23.31	0	0	0	0	0	0	0	0	0	0	0;
	12833	226.17	144.77	210	-50	1.073133	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12840	30.01	15.35	30	-10	1.052369	100	1	100	20.01	0	0	0	0	0	0	0	0	0	0	0;
	12843	30.01	15.35	30	-10	1.052369	100	1	100	20.01	0	0	0	0	0	0	0	0	0	0	0;
	12847	10.89	0.26	10	-10	1.023718	100	1	100	7.26	0	0	0	0	0	0	0	0	0	0	0;
	12855	623.11	488.64	520	-120	1.057579	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	12856	30.01	15.35	30	-10	1.052369	100	1	100	20.01	0	0	0	0	0	0	0	0	0	0	0;
	12859	243.32	23.39	210	-50	1.055881	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12863	5.26	0.32	10	-10	1.017634	100	1	100	3.51	0	0	0	0	0	0	0	0	0	0	0;
	12864	18.6	0.67	10	-10	0.999314	100	1	100	12.4	0	0	0	0	0	0	0	0	0	0	0;
	12865	172.8	-7.44	130	-80	1.029957	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	12869	90.23	10.46	60	-20	1.028204	100	1	200	60.15	0	0	0	0	0	0	0	0	0	0	0;
	12873	12.6	0.36	10	-10	1.014795	100	1	100	8.4	0	0	0	0	0	0	0	0	0	0	0;
	12874	235.87	0.58	210	-50	1.042461	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12878	284.4	98.98	210	-50	1.060659	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	12880	284.51	138.58	210	-50	1.065482	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	12882	284.51	138.58	210	-50	1.065482	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	12883	284.51	138.58	210	-50	1.065482	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	12884	284.51	138.58	210	-50	1.065482	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	12887	5.27	0.39	10	-10	1.019063	100	1	100	3.51	0	0	0	0	0	0	0	0	0	0	0;
	12896	18.32	-0.94	10	-10	0.995077	100	1	100	12.21	0	0	0	0	0	0	0	0	0	0	0;
	12900	49.83	6.62	60	-20	1.007268	100	1	100	33.22	0	0	0	0	0	0	0	0	0	0	0;
	12902	29.54	8.16	20	-10	1.030509	100	1	100	19.69	0	0	0	0	0	0	0	0	0	0	0;
	12904	24.99	21.34	30	-10	1.019094	100	1	100	16.66	0	0	0	0	0	0	0	0	0	0	0;
	12906	17.21	5.28	30	-10	1.020161	100	1	100	11.47	0	0	0	0	0	0	0	0	0	0	0;
	12907	17.21	5.28	30	-10	1.020161	100	1	100	11.47	0	0	0	0	0	0	0	0	0	0	0;
	12908	17.21	5.28	30	-10	1.020161	100	1	100	11.47	0	0	0	0	0	0	0	0	0	0	0;
	12910	17.21	5.28	30	-10	1.020161	100	1	100	11.47	0	0	0	0	0	0	0	0	0	0	0;
	12911	36.03	16.08	30	-10	1.045359	100	1	100	24.02	0	0	0	0	0	0	0	0	0	0	0;
	12912	17.21	5.28	30	-10	1.020161	100	1	100	11.47	0	0	0	0	0	0	0	0	0	0	0;
	12914	19.27	11.16	30	-10	1.021452	100	1	100	12.85	0	0	0	0	0	0	0	0	0	0	0;
	12915	96.9	14.27	60	-20	1.033735	100	1	200	64.6	0	0	0	0	0	0	0	0	0	0	0;
	12916	17.21	5.28	30	-10	1.020161	100	1	100	11.47	0	0	0	0	0	0	0	0	0	0	0;
	12917	17.21	5.28	30	-10	1.020161	100	1	100	11.47	0	0	0	0	0	0	0	0	0	0	0;
	12920	617.34	197.21	520	-120	1.038461	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	12921	617.34	197.21	520	-120	1.038461	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	12925	617.35	197.21	520	-120	1.038461	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	12929	36.88	19.22	30	-10	1.037296	100	1	100	24.59	0	0	0	0	0	0	0	0	0	0	0;
	12930	32.94	16.6	30	-10	1.041198	100	1	100	21.96	0	0	0	0	0	0	0	0	0	0	0;
	12934	99.68	14.71	60	-20	1.035207	100	1	200	66.45	0	0	0	0	0	0	0	0	0	0	0;
	12937	617.35	197.21	520	-120	1.038461	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	12946	31.3	-0.89	20	-20	0.998951	100	1	100	20.87	0	0	0	0	0	0	0	0	0	0	0;
	12949	32.72	3.89	30	-20	1.002664	100	1	100	21.81	0	0	0	0	0	0	0	0	0	0	0;
	12962	366.21	205.07	210	-50	1.129641	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	12965	215.66	145.16	160	-40	1.043227	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12966	17.6	0.07	20	-20	0.977562	100	1	100	11.73	0	0	0	0	0	0	0	0	0	0	0;
	12968	626.62	195.6	520	-120	1.053159	100	1	1300	400	0	0	0	0	0	0	0	0	0	0	0;
	12969	249.81	71.56	160	-40	1.053581	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12970	249.81	71.56	160	-40	1.053581	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	12973	25.08	6.95	30	-10	1.024018	100	1	100	16.72	0	0	0	0	0	0	0	0	0	0	0;
	12983	150	8.11	120	-70	0.989744	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	12984	29.77	17.28	30	-10	1.046953	100	1	100	19.85	0	0	0	0	0	0	0	0	0	0	0;
	12986	29.77	17.28	30	-10	1.046953	100	1	100	19.85	0	0	0	0	0	0	0	0	0	0	0;
	12989	33.35	5.71	10	0	1.021989	100	1	100	22.23	0	0	0	0	0	0	0	0	0	0	0;
	12993	19.96	23.09	30	-10	1.036433	100	1	100	13.31	0	0	0	0	0	0	0	0	0	0	0;
	12994	19.96	23.09	30	-10	1.036433	100	1	100	13.31	0	0	0	0	0	0	0	0	0	0	0;
	13002	172.8	-48.69	130	-80	0.960023	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	13007	32.99	2.77	30	-10	1.041919	100	1	100	21.99	0	0	0	0	0	0	0	0	0	0	0;
	13008	32.99	2.77	30	-10	1.041919	100	1	100	21.99	0	0	0	0	0	0	0	0	0	0	0;
	13009	32.99	2.77	30	-10	1.041919	100	1	100	21.99	0	0	0	0	0	0	0	0	0	0	0;
	13010	100	41.35	90	-50	1.090507	100	1	300	66.67	0	0	0	0	0	0	0	0	0	0	0;
	13012	248.87	83.85	210	-50	1.033471	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	13027	19.3	4.46	20	-10	1.024979	100	1	100	12.87	0	0	0	0	0	0	0	0	0	0	0;
	13029	10.86	0.1	10	-10	1.0018	100	1	100	7.24	0	0	0	0	0	0	0	0	0	0	0;
	13032	16.44	0.07	10	-10	1.011811	100	1	100	10.96	0	0	0	0	0	0	0	0	0	0	0;
	13037	12.95	0.5	20	-10	1.067719	100	1	100	8.63	0	0	0	0	0	0	0	0	0	0	0;
	13041	26.62	13.25	30	-10	1.040642	100	1	100	17.75	0	0	0	0	0	0	0	0	0	0	0;
	13043	24.3	3.82	30	-10	1.010028	100	1	100	16.2	0	0	0	0	0	0	0	0	0	0	0;
	13046	24.3	3.82	30	-10	1.010028	100	1	100	16.2	0	0	0	0	0	0	0	0	0	0	0;
	13047	21.61	4.62	30	-10	1.007128	100	1	100	14.41	0	0	0	0	0	0	0	0	0	0	0;
	13048	21.61	4.62	30	-10	1.007128	100	1	100	14.41	0	0	0	0	0	0	0	0	0	0	0;
	13049	21.61	4.62	30	-10	1.007128	100	1	100	14.41	0	0	0	0	0	0	0	0	0	0	0;
	13052	21.61	4.62	30	-10	1.007128	100	1	100	14.41	0	0	0	0	0	0	0	0	0	0	0;
	13053	21.61	4.62	30	-10	1.007128	100	1	100	14.41	0	0	0	0	0	0	0	0	0	0	0;
	13054	21.61	4.62	30	-10	1.007128	100	1	100	14.41	0	0	0	0	0	0	0	0	0	0	0;
	13055	24.3	3.82	30	-10	1.010028	100	1	100	16.2	0	0	0	0	0	0	0	0	0	0	0;
	13057	30.32	6.61	30	-10	1.007679	100	1	100	20.21	0	0	0	0	0	0	0	0	0	0	0;
	13058	30.32	6.61	30	-10	1.007679	100	1	100	20.21	0	0	0	0	0	0	0	0	0	0	0;
	13059	24.3	3.82	30	-10	1.010028	100	1	100	16.2	0	0	0	0	0	0	0	0	0	0	0;
	13060	24.3	3.82	30	-10	1.010028	100	1	100	16.2	0	0	0	0	0	0	0	0	0	0	0;
	13062	84.19	8.48	60	-20	1.040259	100	1	200	56.13	0	0	0	0	0	0	0	0	0	0	0;
	13064	24.3	3.82	30	-10	1.010028	100	1	100	16.2	0	0	0	0	0	0	0	0	0	0	0;
	13066	84.19	8.48	60	-20	1.040259	100	1	200	56.13	0	0	0	0	0	0	0	0	0	0	0;
	13069	93.64	6.04	60	-20	1.023818	100	1	200	62.43	0	0	0	0	0	0	0	0	0	0	0;
	13075	2.06	0.62	10	-10	1.021202	100	1	100	1.37	0	0	0	0	0	0	0	0	0	0	0;
	13079	223.23	111.08	210	-50	1.065314	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	13080	49.34	9.15	40	-10	1.034597	100	1	100	32.89	0	0	0	0	0	0	0	0	0	0	0;
	13081	35.55	16.64	30	-10	1.024595	100	1	100	23.7	0	0	0	0	0	0	0	0	0	0	0;
	13082	6.55	-0.21	10	-10	1.003807	100	1	100	4.37	0	0	0	0	0	0	0	0	0	0	0;
	13086	33.19	5.48	30	-10	0.991806	100	1	100	22.13	0	0	0	0	0	0	0	0	0	0	0;
	13087	15.73	-3.4	10	-20	0.975077	100	1	100	10.49	0	0	0	0	0	0	0	0	0	0	0;
	13090	240.6	37.13	100	-30	0.970854	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	13091	101	24.39	60	-20	1.022149	100	1	300	67.33	0	0	0	0	0	0	0	0	0	0	0;
	13093	47.94	3.29	40	-10	1.023603	100	1	100	31.96	0	0	0	0	0	0	0	0	0	0	0;
	13095	58.57	21.97	40	-10	1.034376	100	1	200	39.05	0	0	0	0	0	0	0	0	0	0	0;
	13096	43.6	4.09	30	-30	0.9595	100	1	100	29.07	0	0	0	0	0	0	0	0	0	0	0;
	13101	49.34	9.15	40	-10	1.034597	100	1	100	32.89	0	0	0	0	0	0	0	0	0	0	0;
	13108	50.34	9.17	40	-10	1.034532	100	1	200	33.56	0	0	0	0	0	0	0	0	0	0	0;
	13112	14	0.14	10	-10	0.994692	100	1	100	9.33	0	0	0	0	0	0	0	0	0	0	0;
	13122	38.79	7.11	30	-10	1.035838	100	1	100	25.86	0	0	0	0	0	0	0	0	0	0	0;
	13123	277.81	77.81	210	-50	1.034935	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	13131	23.5	6.96	20	-10	1.015521	100	1	100	15.67	0	0	0	0	0	0	0	0	0	0	0;
	13133	34.31	1.99	30	-10	1.015531	100	1	100	22.87	0	0	0	0	0	0	0	0	0	0	0;
	13137	322.34	84.39	210	-50	1.058231	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	13143	168	14.06	130	-80	1.016606	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	13145	257.97	-14.76	210	-50	1.029219	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	13147	32.2	7.05	30	-10	1.034688	100	1	100	21.47	0	0	0	0	0	0	0	0	0	0	0;
	13158	424.18	1.63	260	-60	1.036955	100	1	900	200	0	0	0	0	0	0	0	0	0	0	0;
	13159	4.2	-0.51	10	-10	0.974729	100	1	100	2.8	0	0	0	0	0	0	0	0	0	0	0;
	13160	43.17	6.95	30	-10	1.015029	100	1	100	28.78	0	0	0	0	0	0	0	0	0	0	0;
	13163	35.6	15.17	30	-10	1.026957	100	1	100	23.73	0	0	0	0	0	0	0	0	0	0	0;
	13165	5.2	0.53	10	-10	1.031159	100	1	100	3.47	0	0	0	0	0	0	0	0	0	0	0;
	13174	38.02	11.21	30	-10	1.043087	100	1	100	25.35	0	0	0	0	0	0	0	0	0	0	0;
	13182	-12.51	-2.52	10	-10	0.975956	100	1	100	-16.68	0	0	0	0	0	0	0	0	0	0	0;
	13188	274.11	117.05	160	-40	1.051387	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	13189	274.11	117.05	160	-40	1.051387	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	13190	274.11	117.05	160	-40	1.051387	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	13191	158.9	39.46	100	-30	1.020032	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	13192	952.11	346.75	520	-120	1.038082	100	1	2000	600	0	0	0	0	0	0	0	0	0	0	0;
	13193	1.2	0.03	10	-10	1.003171	100	1	100	0.8	0	0	0	0	0	0	0	0	0	0	0;
	13195	36.8	-2.54	30	-30	0.98379	100	1	100	24.53	0	0	0	0	0	0	0	0	0	0	0;
	13200	57.5	6.09	30	-30	0.99654	100	1	200	38.33	0	0	0	0	0	0	0	0	0	0	0;
	13209	180.96	38.95	160	-40	1.029835	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	13212	6.1	0.22	10	-10	1.00107	100	1	100	4.07	0	0	0	0	0	0	0	0	0	0	0;
	13218	135	10.29	70	-20	1.008348	100	1	300	90	0	0	0	0	0	0	0	0	0	0	0;
	13222	39.1	-2.22	20	-20	0.943436	100	1	100	26.07	0	0	0	0	0	0	0	0	0	0	0;
	13223	37.08	5.51	30	-10	1.021005	100	1	100	24.72	0	0	0	0	0	0	0	0	0	0	0;
	13226	8.2	-0.77	10	-10	1.00746	100	1	100	5.47	0	0	0	0	0	0	0	0	0	0	0;
	13232	50.92	11.1	40	-10	1.010998	100	1	200	33.95	0	0	0	0	0	0	0	0	0	0	0;
	13235	323.35	77.76	210	-50	1.058114	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	13238	58.8	-10.54	30	-30	0.9646	100	1	200	39.2	0	0	0	0	0	0	0	0	0	0	0;
	13265	10.26	0.48	10	-10	1.000157	100	1	100	6.84	0	0	0	0	0	0	0	0	0	0	0;
	13268	16.6	0.7	10	-10	1.016188	100	1	100	11.07	0	0	0	0	0	0	0	0	0	0	0;
	13270	5	-1.11	10	-10	0.981334	100	1	100	3.33	0	0	0	0	0	0	0	0	0	0	0;
	13285	51.75	8.57	40	-10	1.019281	100	1	200	34.5	0	0	0	0	0	0	0	0	0	0	0;
	13292	36.77	13.93	30	-10	1.042317	100	1	100	24.51	0	0	0	0	0	0	0	0	0	0	0;
	13298	25.41	10.98	30	-10	1.019411	100	1	100	16.94	0	0	0	0	0	0	0	0	0	0	0;
	13303	36.03	16.08	30	-10	1.045359	100	1	100	24.02	0	0	0	0	0	0	0	0	0	0	0;
	13310	33.39	17.82	40	-10	1.025649	100	1	100	22.26	0	0	0	0	0	0	0	0	0	0	0;
	13311	33.45	6.28	40	-10	1.055904	100	1	100	22.3	0	0	0	0	0	0	0	0	0	0	0;
	13312	3.5	-0.25	10	-10	0.995346	100	1	100	2.33	0	0	0	0	0	0	0	0	0	0	0;
	13313	54.06	11.61	40	-10	1.023809	100	1	200	36.04	0	0	0	0	0	0	0	0	0	0	0;
	13316	50.34	9.17	40	-10	1.034532	100	1	200	33.56	0	0	0	0	0	0	0	0	0	0	0;
	13322	330.71	4.92	210	-50	1.043049	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	13327	378.98	209.33	210	-50	1.117446	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	13330	32.51	6.89	30	-10	1.011478	100	1	100	21.67	0	0	0	0	0	0	0	0	0	0	0;
	13333	214.52	110.59	210	-50	1.065343	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	13338	62.5	10.73	20	0	1.000057	100	1	200	41.67	0	0	0	0	0	0	0	0	0	0	0;
	13345	40.44	13.81	40	-10	1.025019	100	1	100	26.96	0	0	0	0	0	0	0	0	0	0	0;
	13363	26.01	11.35	30	-10	1.062565	100	1	100	17.34	0	0	0	0	0	0	0	0	0	0	0;
	13370	31.29	6.48	30	-10	1.024788	100	1	100	20.86	0	0	0	0	0	0	0	0	0	0	0;
	13371	20.4	3.27	20	-10	1.027307	100	1	100	13.6	0	0	0	0	0	0	0	0	0	0	0;
	13379	423.72	187.67	230	-60	1.052254	100	1	900	200	0	0	0	0	0	0	0	0	0	0	0;
	13383	83.51	27.65	60	-20	1.063297	100	1	200	55.67	0	0	0	0	0	0	0	0	0	0	0;
	13385	62.37	25.07	60	-20	1.030828	100	1	200	41.58	0	0	0	0	0	0	0	0	0	0	0;
	13386	62.37	25.07	60	-20	1.030828	100	1	200	41.58	0	0	0	0	0	0	0	0	0	0	0;
	13387	62.37	25.07	60	-20	1.030828	100	1	200	41.58	0	0	0	0	0	0	0	0	0	0	0;
	13388	82.89	46.38	60	-20	1.055605	100	1	200	55.26	0	0	0	0	0	0	0	0	0	0	0;
	13389	82.89	46.38	60	-20	1.055605	100	1	200	55.26	0	0	0	0	0	0	0	0	0	0	0;
	13390	29.68	6.45	30	-10	1.016891	100	1	100	19.79	0	0	0	0	0	0	0	0	0	0	0;
	13391	29.68	6.45	30	-10	1.016891	100	1	100	19.79	0	0	0	0	0	0	0	0	0	0	0;
	13392	33.84	9.43	30	-10	1.043178	100	1	100	22.56	0	0	0	0	0	0	0	0	0	0	0;
	13393	29.68	6.45	30	-10	1.016891	100	1	100	19.79	0	0	0	0	0	0	0	0	0	0	0;
	13394	29.68	6.45	30	-10	1.016891	100	1	100	19.79	0	0	0	0	0	0	0	0	0	0	0;
	13395	29.68	6.45	30	-10	1.016891	100	1	100	19.79	0	0	0	0	0	0	0	0	0	0	0;
	13397	82.89	46.38	60	-20	1.055605	100	1	200	55.26	0	0	0	0	0	0	0	0	0	0	0;
	13398	82.89	46.38	60	-20	1.055605	100	1	200	55.26	0	0	0	0	0	0	0	0	0	0	0;
	13399	82.89	46.38	60	-20	1.055605	100	1	200	55.26	0	0	0	0	0	0	0	0	0	0	0;
	13407	25.2	-7.17	20	-20	0.965886	100	1	100	16.8	0	0	0	0	0	0	0	0	0	0	0;
	13409	30.9	1.92	30	-10	1.015954	100	1	100	20.6	0	0	0	0	0	0	0	0	0	0	0;
	13411	81.31	8.06	60	-20	1.043619	100	1	200	54.21	0	0	0	0	0	0	0	0	0	0	0;
	13412	31.92	6.07	30	-10	1.027264	100	1	100	21.28	0	0	0	0	0	0	0	0	0	0	0;
	13419	22.46	14.75	30	-10	1.043942	100	1	100	14.97	0	0	0	0	0	0	0	0	0	0	0;
	13421	35.86	4.09	30	-10	1.02494	100	1	100	23.91	0	0	0	0	0	0	0	0	0	0	0;
	13423	35.86	4.09	30	-10	1.02494	100	1	100	23.91	0	0	0	0	0	0	0	0	0	0	0;
	13424	35.86	4.09	30	-10	1.02494	100	1	100	23.91	0	0	0	0	0	0	0	0	0	0	0;
	13425	27.44	5.18	40	-10	1.025625	100	1	100	18.29	0	0	0	0	0	0	0	0	0	0	0;
	13426	63.42	10.31	40	-10	1.03448	100	1	200	42.28	0	0	0	0	0	0	0	0	0	0	0;
	13433	52.29	8.76	40	-10	1.021414	100	1	200	34.86	0	0	0	0	0	0	0	0	0	0	0;
	13437	52.29	8.76	40	-10	1.021414	100	1	200	34.86	0	0	0	0	0	0	0	0	0	0	0;
	13439	52.29	8.76	40	-10	1.021414	100	1	200	34.86	0	0	0	0	0	0	0	0	0	0	0;
	13440	33.96	3.76	20	-10	1.014614	100	1	100	22.64	0	0	0	0	0	0	0	0	0	0	0;
	13447	12.3	0.28	10	-10	1.011745	100	1	100	8.2	0	0	0	0	0	0	0	0	0	0	0;
	13456	28.48	6.95	30	-10	1.005915	100	1	100	18.99	0	0	0	0	0	0	0	0	0	0	0;
	13466	53.99	18.23	40	-10	1.062896	100	1	200	35.99	0	0	0	0	0	0	0	0	0	0	0;
	13467	14.92	-0.69	10	-10	0.989708	100	1	100	9.95	0	0	0	0	0	0	0	0	0	0	0;
	13468	56.36	7.67	40	-10	1.024375	100	1	200	37.57	0	0	0	0	0	0	0	0	0	0	0;
	13482	53.99	18.23	40	-10	1.062896	100	1	200	35.99	0	0	0	0	0	0	0	0	0	0	0;
	13486	53.99	18.23	40	-10	1.062896	100	1	200	35.99	0	0	0	0	0	0	0	0	0	0	0;
	13487	50.34	15.84	40	-10	1.03143	100	1	200	33.56	0	0	0	0	0	0	0	0	0	0	0;
	13489	80.74	25.03	60	-20	1.061099	100	1	200	53.83	0	0	0	0	0	0	0	0	0	0	0;
	13491	20.57	3.1	30	-10	1.026081	100	1	100	13.71	0	0	0	0	0	0	0	0	0	0	0;
	13493	344.6	7.8	180	-140	0.97953	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	13496	32.64	6.67	30	-10	1.044924	100	1	100	21.76	0	0	0	0	0	0	0	0	0	0	0;
	13497	32.64	6.67	30	-10	1.044924	100	1	100	21.76	0	0	0	0	0	0	0	0	0	0	0;
	13500	32.64	6.68	30	-10	1.044939	100	1	100	21.76	0	0	0	0	0	0	0	0	0	0	0;
	13502	32.64	6.68	30	-10	1.044939	100	1	100	21.76	0	0	0	0	0	0	0	0	0	0	0;
	13505	32.64	6.67	30	-10	1.044924	100	1	100	21.76	0	0	0	0	0	0	0	0	0	0	0;
	13506	32.64	6.67	30	-10	1.044924	100	1	100	21.76	0	0	0	0	0	0	0	0	0	0	0;
	13507	32.64	6.67	30	-10	1.044924	100	1	100	21.76	0	0	0	0	0	0	0	0	0	0	0;
	13510	0.2	-0.15	10	-10	1.016107	100	1	100	0.14	0	0	0	0	0	0	0	0	0	0	0;
	13512	27.73	2	30	-10	1.030407	100	1	100	18.49	0	0	0	0	0	0	0	0	0	0	0;
	13514	31	0.14	20	-20	1.000449	100	1	100	20.67	0	0	0	0	0	0	0	0	0	0	0;
	13520	80.74	25.03	60	-20	1.061099	100	1	200	53.83	0	0	0	0	0	0	0	0	0	0	0;
	13525	53.07	7.74	40	-10	1.025333	100	1	200	35.38	0	0	0	0	0	0	0	0	0	0	0;
	13526	200.87	149.13	210	-50	1.062073	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	13530	236.37	0.58	210	-50	1.042446	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	13549	23.04	1.33	20	-20	0.996935	100	1	100	15.36	0	0	0	0	0	0	0	0	0	0	0;
	13550	31.5	1.72	30	-20	1.016735	100	1	100	21	0	0	0	0	0	0	0	0	0	0	0;
	13551	2.5	0.07	10	-10	0.984601	100	1	100	1.67	0	0	0	0	0	0	0	0	0	0	0;
	13556	6.52	-0.06	10	-10	0.994601	100	1	100	4.35	0	0	0	0	0	0	0	0	0	0	0;
	13560	218.81	74.1	210	-50	1.033239	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	13569	96.03	17.22	60	-20	1.036065	100	1	200	64.02	0	0	0	0	0	0	0	0	0	0	0;
	13577	49.34	9.15	40	-10	1.034597	100	1	100	32.89	0	0	0	0	0	0	0	0	0	0	0;
	13580	33.84	9.43	30	-10	1.043178	100	1	100	22.56	0	0	0	0	0	0	0	0	0	0	0;
	13582	33.84	9.43	30	-10	1.043178	100	1	100	22.56	0	0	0	0	0	0	0	0	0	0	0;
	13584	19.12	15.11	30	-10	1.030744	100	1	100	12.75	0	0	0	0	0	0	0	0	0	0	0;
	13585	22.27	8.99	30	-10	1.04545	100	1	100	14.85	0	0	0	0	0	0	0	0	0	0	0;
	13589	22.27	8.99	30	-10	1.04545	100	1	100	14.85	0	0	0	0	0	0	0	0	0	0	0;
	13590	86.25	-3.04	60	-80	0.962614	100	1	200	57.5	0	0	0	0	0	0	0	0	0	0	0;
	13592	326.27	128.78	210	-50	1.051118	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	13594	20.37	8.85	30	-10	1.045375	100	1	100	13.58	0	0	0	0	0	0	0	0	0	0	0;
	13597	20.37	8.85	30	-10	1.045375	100	1	100	13.58	0	0	0	0	0	0	0	0	0	0	0;
	13601	21.74	6.85	30	-10	1.050175	100	1	100	14.49	0	0	0	0	0	0	0	0	0	0	0;
	13604	22.27	8.99	30	-10	1.04545	100	1	100	14.85	0	0	0	0	0	0	0	0	0	0	0;
	13607	38.04	7.18	30	-10	1.020265	100	1	100	25.36	0	0	0	0	0	0	0	0	0	0	0;
	13609	22	9.29	30	-10	1.030852	100	1	100	14.67	0	0	0	0	0	0	0	0	0	0	0;
	13617	2.26	-1.2	10	-10	0.991929	100	1	100	1.51	0	0	0	0	0	0	0	0	0	0	0;
	13620	277.81	77.81	210	-50	1.034935	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	13633	50.34	15.84	40	-10	1.03143	100	1	200	33.56	0	0	0	0	0	0	0	0	0	0	0;
	13639	29.92	0.94	20	-10	0.980011	100	1	100	19.95	0	0	0	0	0	0	0	0	0	0	0;
	13643	18.32	-0.94	10	-10	0.995077	100	1	100	12.21	0	0	0	0	0	0	0	0	0	0	0;
	13648	50.34	15.84	40	-10	1.03143	100	1	200	33.56	0	0	0	0	0	0	0	0	0	0	0;
	13651	16.79	7.28	30	-10	1.042166	100	1	100	11.19	0	0	0	0	0	0	0	0	0	0	0;
	13654	-0	27.02	560	-190	1.079579	100	1	100	-0	0	0	0	0	0	0	0	0	0	0	0;
	13655	20.46	6.24	30	-10	0.992821	100	1	100	13.64	0	0	0	0	0	0	0	0	0	0	0;
	13658	26.41	7.43	30	-10	1.038011	100	1	100	17.61	0	0	0	0	0	0	0	0	0	0	0;
];

%% branch data
%	fbus	tbus	r	x	b	rateA	rateB	rateC	ratio	angle	status	angmin	angmax
mpc.branch = [
	7519	4482	0.0006	0.00616	0	0	0	0	0	0	1	-360	360;
	7519	11823	0.0005	0.0074	0	0	0	0	0	0	1	-360	360;
	3073	4516	0.00107	0.00626	0	0	0	0	0	0	1	-360	360;
	3073	4706	0.000841	0.0055	0	0	0	0	0	0	1	-360	360;
	5157	1032	0.00143	0.01563	0	0	0	0	0	0	1	-360	360;
	5157	6653	0.00146	0.01692	0	0	0	0	0	0	1	-360	360;
	628	7896	7e-05	0.00057	0	0	0	0	0	0	1	-360	360;
	628	6592	6e-05	0.00056	0	0	0	0	0	0	1	-360	360;
	11792	10888	0.003349	0.01656	0	0	0	0	0	0	1	-360	360;
	11792	4609	0.003409	0.016211	0	0	0	0	0	0	1	-360	360;
	6573	10888	0.003479	0.01586	0	0	0	0	0	0	1	-360	360;
	6573	4609	0.003099	0.017969	0	0	0	0	0	0	1	-360	360;
	782	13300	0.00127	0.01842	0	0	0	0	0	0	1	-360	360;
	782	1186	0.00126	0.01847	0	0	0	0	0	0	1	-360	360;
	11715	924	0.0013	0.01808	0	0	0	0	0	0	1	-360	360;
	11715	12182	0.00131	0.01765	0	0	0	0	0	0	1	-360	360;
	10257	2582	0.00065	0.01311	0	0	0	0	0	0	1	-360	360;
	10257	9363	0.00067	0.00974	0	0	0	0	0	0	1	-360	360;
	7595	11501	0.00236	0.01418	0	0	0	0	0	0	1	-360	360;
	7595	11390	0.002169	0.015909	0	0	0	0	0	0	1	-360	360;
	6873	1806	0.00027	0.00254	0	0	0	0	0	0	1	-360	360;
	6873	11151	0.00027	0.00259	0	0	0	0	0	0	1	-360	360;
	11949	1806	0.00033	0.00302	0	0	0	0	0	0	1	-360	360;
	11949	1269	0.00029	0.00312	0	0	0	0	0	0	1	-360	360;
	2623	3359	0.000289	0.00137	0	0	0	0	0	0	1	-360	360;
	2623	11159	0.000219	0.00263	0	0	0	0	0	0	1	-360	360;
	3020	3359	0.000289	0.001341	0	0	0	0	0	0	1	-360	360;
	3020	12995	0.000229	0.00257	0	0	0	0	0	0	1	-360	360;
	9670	3359	0.000841	0.00689	0	0	0	0	0	0	1	-360	360;
	9670	2360	0.000979	0.00586	0	0	0	0	0	0	1	-360	360;
	8390	3359	0.00014	0.0012	0	0	0	0	0	0	1	-360	360;
	8390	11515	0.000211	0.001539	0	0	0	0	0	0	1	-360	360;
	12630	3359	0.000401	0.004901	0	0	0	0	0	0	1	-360	360;
	12630	11515	0.000219	0.00151	0	0	0	0	0	0	1	-360	360;
	3340	3359	0.000729	0.004539	0	0	0	0	0	0	1	-360	360;
	3340	3953	0.00062	0.00482	0	0	0	0	0	0	1	-360	360;
	7444	8848	0.00052	0.00407	0	0	0	0	0	0	1	-360	360;
	7444	5026	0.00034	0.00502	0	0	0	0	0	0	1	-360	360;
	11055	8848	0.00065	0.00698	0	0	0	0	0	0	1	-360	360;
	11055	13067	0.00059	0.00806	0	0	0	0	0	0	1	-360	360;
	9857	13125	0.00063	0.00464	0	0	0	0	0	0	1	-360	360;
	9857	12598	0.000599	0.005099	0	0	0	0	0	0	1	-360	360;
	7256	7011	0.00064	0.00307	0	0	0	0	0	0	1	-360	360;
	7256	11232	0.000659	0.003	0	0	0	0	0	0	1	-360	360;
	12732	9895	0.00053	0.00528	0	0	0	0	0	0	1	-360	360;
	12732	4416	0.00056	0.0053	0	0	0	0	0	0	1	-360	360;
	4494	7011	0.0018	0.007919	0	0	0	0	0	0	1	-360	360;
	4494	10481	0.001651	0.00862	0	0	0	0	0	0	1	-360	360;
	7913	9895	0.00045	0.00545	0	0	0	0	0	0	1	-360	360;
	7913	1269	0.00052	0.00476	0	0	0	0	0	0	1	-360	360;
	3258	10043	0.01286	0.037421	0	0	0	0	0	0	1	-360	360;
	3258	8238	0.032397	0.123	0	0	0	0	0	0	1	-360	360;
	697	10043	0.032851	0.069727	0	0	0	0	0	0	1	-360	360;
	697	11841	0.059942	0.112802	0	0	0	0	0	0	1	-360	360;
	11629	10942	0.00011	0.00109	0	0	0	0	0	0	1	-360	360;
	11629	13171	0.00011	0.00107	0	0	0	0	0	0	1	-360	360;
	9089	11260	0.000461	0.00395	0	0	0	0	0	0	1	-360	360;
	9089	7609	0.000599	0.003789	0	0	0	0	0	0	1	-360	360;
	13119	3581	0.00173	0.01542	0	0	0	0	0	0	1	-360	360;
	13119	6381	0.00166	0.02083	0	0	0	0	0	0	1	-360	360;
	5075	13106	0.006314	0.018488	0	0	0	0	0	0	1	-360	360;
	5075	9503	0.015727	0.037669	0	0	0	0	0	0	1	-360	360;
	11048	10312	0.00374	0.02095	0	0	0	0	0	0	1	-360	360;
	11048	3185	0.00511	0.019461	0	0	0	0	0	0	1	-360	360;
	1538	4109	0.011281	0.050771	0	0	0	0	0	0	1	-360	360;
	1538	3721	0.0115	0.049719	0	0	0	0	0	0	1	-360	360;
	5803	4109	0.011719	0.048659	0	0	0	0	0	0	1	-360	360;
	5803	3721	0.01193	0.047599	0	0	0	0	0	0	1	-360	360;
	189	4268	0.00064	0.00715	0	0	0	0	0	0	1	-360	360;
	189	11488	0.00089	0.00712	0	0	0	0	0	0	1	-360	360;
	7384	1009	0.00048	0.00555	0	0	0	0	0	0	1	-360	360;
	7384	11547	0.00047	0.00566	0	0	0	0	0	0	1	-360	360;
	11070	1009	0.00046	0.00524	0	0	0	0	0	0	1	-360	360;
	11070	11547	0.00052	0.00514	0	0	0	0	0	0	1	-360	360;
	2061	6603	0.00438	0.010686	0	0	0	0	0	0	1	-360	360;
	2061	9733	0.031347	0.062653	0	0	0	0	0	0	1	-360	360;
	9502	7241	0.00029	0.00298	0	0	0	0	0	0	1	-360	360;
	9502	8865	0.0003	0.00291	0	0	0	0	0	0	1	-360	360;
	12336	13321	0.00155	0.01529	0	0	0	0	0	0	1	-360	360;
	12336	494	0.00148	0.0152	0	0	0	0	0	0	1	-360	360;
	13540	11971	0.000169	0.00096	0	0	0	0	0	0	1	-360	360;
	13540	330	0.000182	0.000942	0	0	0	0	0	0	1	-360	360;
	8250	622	0.00016	0.00184	0	0	0	0	0	0	1	-360	360;
	8250	8646	0.00022	0.00287	0	0	0	0	0	0	1	-360	360;
	5656	9506	0.00015	0.00187	0	0	0	0	0	0	1	-360	360;
	5656	7003	0.00022	0.00266	0	0	0	0	0	0	1	-360	360;
	8767	13151	0.00138	0.0174	0	0	0	0	0	0	1	-360	360;
	8767	256	0.00272	0.0288	0	0	0	0	0	0	1	-360	360;
	11952	7565	0.00065	0.01024	0	0	0	0	0	0	1	-360	360;
	11952	8750	0.00073	0.01004	0	0	0	0	0	0	1	-360	360;
	13220	7565	0.00075	0.00983	0	0	0	0	0	0	1	-360	360;
	13220	8750	0.00076	0.00963	0	0	0	0	0	0	1	-360	360;
	10465	10831	0.00042	0.02666	0	0	0	0	0	0	1	-360	360;
	8079	10358	0.00029	0.0026	0	0	0	0	0	0	1	-360	360;
	8079	8404	0.00022	0.00303	0	0	0	0	0	0	1	-360	360;
	7197	5524	0.001	0.006159	0	0	0	0	0	0	1	-360	360;
	7197	11931	0.00113	0.00657	0	0	0	0	0	0	1	-360	360;
	1721	12153	0.0013	0.01277	0	0	0	0	0	0	1	-360	360;
	1721	10744	0.00163	0.01808	0	0	0	0	0	0	1	-360	360;
	8400	850	0.00047	0.00406	0	0	0	0	0	0	1	-360	360;
	8400	8575	0.00235	0.02343	0	0	0	0	0	0	1	-360	360;
	9020	2201	0.00036	0.00445	0	0	0	0	0	0	1	-360	360;
	9020	2350	0.00017	0.0025	0	0	0	0	0	0	1	-360	360;
	6435	2201	0.00036	0.00436	0	0	0	0	0	0	1	-360	360;
	6435	3448	0.00017	0.00255	0	0	0	0	0	0	1	-360	360;
	3223	4642	0.00096	0.00995	0	0	0	0	0	0	1	-360	360;
	3223	9919	0.00152	0.01949	0	0	0	0	0	0	1	-360	360;
	7755	4642	0.00094	0.01014	0	0	0	0	0	0	1	-360	360;
	7755	9919	0.00149	0.01986	0	0	0	0	0	0	1	-360	360;
	4129	4642	0.00107	0.01243	0	0	0	0	0	0	1	-360	360;
	4129	3715	0.00102	0.01034	0	0	0	0	0	0	1	-360	360;
	5471	7106	0.0047	0.02061	0	0	0	0	0	0	1	-360	360;
	5471	5434	0.00739	0.040531	0	0	0	0	0	0	1	-360	360;
	8834	12770	0.00049	0.00611	0	0	0	0	0	0	1	-360	360;
	8834	10744	0.00035	0.00431	0	0	0	0	0	0	1	-360	360;
	7831	1467	0.00261	0.01344	0	0	0	0	0	0	1	-360	360;
	7831	5173	0.003521	0.019591	0	0	0	0	0	0	1	-360	360;
	10033	5005	0.0025	0.01419	0	0	0	0	0	0	1	-360	360;
	10033	7387	0.00287	0.0237	0	0	0	0	0	0	1	-360	360;
	9789	9848	0.00088	0.00956	0	0	0	0	0	0	1	-360	360;
	9789	430	0.00095	0.00877	0	0	0	0	0	0	1	-360	360;
	2916	1094	0.00332	0.020651	0	0	0	0	0	0	1	-360	360;
	2916	2869	0.003969	0.018781	0	0	0	0	0	0	1	-360	360;
	4142	3925	0.002849	0.014909	0	0	0	0	0	0	1	-360	360;
	4142	2245	0.00439	0.02451	0	0	0	0	0	0	1	-360	360;
	123	12854	0.0002	0.0018	0	0	0	0	0	0	1	-360	360;
	123	12369	0.00025	0.00261	0	0	0	0	0	0	1	-360	360;
	10806	12854	0.0002	0.00176	0	0	0	0	0	0	1	-360	360;
	10806	1112	0.00025	0.00255	0	0	0	0	0	0	1	-360	360;
	12593	9027	0.012488	0.047413	0	0	0	0	0	0	1	-360	360;
	12593	455	0.010479	0.027711	0	0	0	0	0	0	1	-360	360;
	3745	10922	0.00044	0.00428	0	0	0	0	0	0	1	-360	360;
	3745	1898	0.00036	0.00475	0	0	0	0	0	0	1	-360	360;
	6938	7640	0.00058	0.00534	0	0	0	0	0	0	1	-360	360;
	6938	4543	0.00054	0.00589	0	0	0	0	0	0	1	-360	360;
	4229	2024	0.00206	0.02057	0	0	0	0	0	0	1	-360	360;
	4229	9388	0.00077	0.00872	0	0	0	0	0	0	1	-360	360;
	5543	13575	0.00161	0.01963	0	0	0	0	0	0	1	-360	360;
	5543	9388	0.00075	0.00888	0	0	0	0	0	0	1	-360	360;
	2764	10638	0.00016	0.00152	0	0	0	0	0	0	1	-360	360;
	2764	10338	0.00019	0.00163	0	0	0	0	0	0	1	-360	360;
	7963	10638	0.00015	0.0016	0	0	0	0	0	0	1	-360	360;
	7963	10338	0.00015	0.00183	0	0	0	0	0	0	1	-360	360;
	5383	4802	0.00121	0.01426	0	0	0	0	0	0	1	-360	360;
	5383	8035	0.00116	0.01456	0	0	0	0	0	0	1	-360	360;
	663	204	0.00068	0.00727	0	0	0	0	0	0	1	-360	360;
	663	6973	0.0007	0.00734	0	0	0	0	0	0	1	-360	360;
	11598	204	0.00076	0.00712	0	0	0	0	0	0	1	-360	360;
	11598	6973	0.0008	0.00719	0	0	0	0	0	0	1	-360	360;
	2983	2257	0.0002	0.0022	0	0	0	0	0	0	1	-360	360;
	2983	10869	0.0002	0.0022	0	0	0	0	0	0	1	-360	360;
	411	4997	0.000205	0.00215	0	0	0	0	0	0	1	-360	360;
	411	10869	0.000205	0.00215	0	0	0	0	0	0	1	-360	360;
	8649	12423	0.00053	0.00589	0	0	0	0	0	0	1	-360	360;
	8649	1165	0.00046	0.00675	0	0	0	0	0	0	1	-360	360;
	13485	12423	0.00054	0.00577	0	0	0	0	0	0	1	-360	360;
	13485	1165	0.00045	0.00625	0	0	0	0	0	0	1	-360	360;
	2363	2209	0.00275	0.02752	0	0	0	0	0	0	1	-360	360;
	2363	11249	0.00184	0.01515	0	0	0	0	0	0	1	-360	360;
	3314	10019	0.00438	0.02432	0	0	0	0	0	0	1	-360	360;
	3314	879	0.0043	0.01949	0	0	0	0	0	0	1	-360	360;
	12347	10227	0.00033	0.00269	0	0	0	0	0	0	1	-360	360;
	12347	2517	0.00243	0.01904	0	0	0	0	0	0	1	-360	360;
	1326	10227	0.00034	0.00263	0	0	0	0	0	0	1	-360	360;
	1326	2517	0.00248	0.01864	0	0	0	0	0	0	1	-360	360;
	9864	13169	0.00122	0.01249	0	0	0	0	0	0	1	-360	360;
	9864	5959	0.00106	0.01031	0	0	0	0	0	0	1	-360	360;
	4830	13169	0.00119	0.01273	0	0	0	0	0	0	1	-360	360;
	4830	5959	0.00108	0.01009	0	0	0	0	0	0	1	-360	360;
	12318	6464	0.001039	0.013711	0	0	0	0	0	0	1	-360	360;
	12318	7516	0.00114	0.011409	0	0	0	0	0	0	1	-360	360;
	11537	8566	0.001021	0.013969	0	0	0	0	0	0	1	-360	360;
	11537	7516	0.001169	0.011159	0	0	0	0	0	0	1	-360	360;
	13343	12066	0	0.000271	0	0	0	0	0	0	1	-360	360;
	3313	12066	0	0.000271	0	0	0	0	0	0	1	-360	360;
	1196	218	0.009372	0.035702	0	0	0	0	0	0	1	-360	360;
	1196	3002	0.029157	0.103686	0	0	0	0	0	0	1	-360	360;
	780	8041	0.00065	0.00673	0	0	0	0	0	0	1	-360	360;
	780	6114	0.00066	0.00658	0	0	0	0	0	0	1	-360	360;
	5179	3568	0.000969	0.006021	0	0	0	0	0	0	1	-360	360;
	5179	3085	0.00099	0.005901	0	0	0	0	0	0	1	-360	360;
	1666	7897	0.00088	0.006599	0	0	0	0	0	0	1	-360	360;
	1666	6301	0.00086	0.006729	0	0	0	0	0	0	1	-360	360;
	12861	4782	9e-05	0.00077	0	0	0	0	0	0	1	-360	360;
	12861	6207	9e-05	0.00078	0	0	0	0	0	0	1	-360	360;
	7212	4782	8e-05	0.00082	0	0	0	0	0	0	1	-360	360;
	7212	11151	8e-05	0.00076	0	0	0	0	0	0	1	-360	360;
	9869	8041	9e-05	0.00074	0	0	0	0	0	0	1	-360	360;
	9869	6207	9e-05	0.00073	0	0	0	0	0	0	1	-360	360;
	6015	3568	0.00062	0.003781	0	0	0	0	0	0	1	-360	360;
	6015	1131	0.000599	0.0035	0	0	0	0	0	0	1	-360	360;
	10035	7897	0.00064	0.00274	0	0	0	0	0	0	1	-360	360;
	10035	1910	0.000651	0.00268	0	0	0	0	0	0	1	-360	360;
	9951	1939	0.00055	0.00685	0	0	0	0	0	0	1	-360	360;
	9951	557	0.00061	0.00639	0	0	0	0	0	0	1	-360	360;
	8771	8041	0.00042	0.00523	0	0	0	0	0	0	1	-360	360;
	8771	12272	0.00045	0.00605	0	0	0	0	0	0	1	-360	360;
	8284	4782	0.0003	0.00228	0	0	0	0	0	0	1	-360	360;
	8284	13171	0.00026	0.00264	0	0	0	0	0	0	1	-360	360;
	4891	7738	0.00536	0.03568	0	0	0	0	0	0	1	-360	360;
	4891	9177	0.006091	0.03549	0	0	0	0	0	0	1	-360	360;
	7982	7738	0.006419	0.033271	0	0	0	0	0	0	1	-360	360;
	7982	2109	0.005789	0.035901	0	0	0	0	0	0	1	-360	360;
	10181	2611	0	0.0002	0	0	0	0	0	0	1	-360	360;
	10181	11398	0	0.0002	0	0	0	0	0	0	1	-360	360;
	13331	7738	0.008401	0.04011	0	0	0	0	0	0	1	-360	360;
	13331	4629	0.007831	0.041531	0	0	0	0	0	0	1	-360	360;
	11042	971	0.00575	0.046461	0	0	0	0	0	0	1	-360	360;
	11042	9307	0.006711	0.04426	0	0	0	0	0	0	1	-360	360;
	5320	12300	0.00036	0.00485	0	0	0	0	0	0	1	-360	360;
	5320	9117	0.00042	0.00415	0	0	0	0	0	0	1	-360	360;
	10275	2399	9e-05	0.00069	0	0	0	0	0	0	1	-360	360;
	10275	10873	4e-05	0.00036	0	0	0	0	0	0	1	-360	360;
	11545	3592	0.00055	0.00569	0	0	0	0	0	0	1	-360	360;
	11545	10322	0.0007	0.00767	0	0	0	0	0	0	1	-360	360;
	8859	13646	0.0013	0.01299	0	0	0	0	0	0	1	-360	360;
	4222	12258	0.00092	0.01031	0	0	0	0	0	0	1	-360	360;
	1576	13646	0.00158	0.01496	0	0	0	0	0	0	1	-360	360;
	1576	1624	0.00123	0.01309	0	0	0	0	0	0	1	-360	360;
	5947	9710	0.000599	0.00274	0	0	0	0	0	0	1	-360	360;
	5947	10727	0.0005	0.00405	0	0	0	0	0	0	1	-360	360;
	3661	11493	0.003349	0.02445	0	0	0	0	0	0	1	-360	360;
	3661	11925	0.003281	0.02493	0	0	0	0	0	0	1	-360	360;
	5848	5338	0.00087	0.01097	0	0	0	0	0	0	1	-360	360;
	5848	13272	0.00161	0.01464	0	0	0	0	0	0	1	-360	360;
	8440	4914	0	0.00046	0	0	0	0	0	0	1	-360	360;
	6339	5430	0.00118	0.01223	0	0	0	0	0	0	1	-360	360;
	6339	9813	0.00045	0.00364	0	0	0	0	0	0	1	-360	360;
	3407	2782	0.00714	0.03393	0	0	0	0	0	0	1	-360	360;
	3407	13262	0.00655	0.036849	0	0	0	0	0	0	1	-360	360;
	11656	11688	0.00088	0.00722	0	0	0	0	0	0	1	-360	360;
	11656	9434	0.00071	0.00835	0	0	0	0	0	0	1	-360	360;
	9621	11688	0.00084	0.00705	0	0	0	0	0	0	1	-360	360;
	9621	9162	0.00102	0.00799	0	0	0	0	0	0	1	-360	360;
	9337	8152	0.003781	0.029969	0	0	0	0	0	0	1	-360	360;
	9337	11977	0.00386	0.02969	0	0	0	0	0	0	1	-360	360;
	9790	640	0.00076	0.00463	0	0	0	0	0	0	1	-360	360;
	9790	4131	0.000711	0.004841	0	0	0	0	0	0	1	-360	360;
	7596	13576	0.00631	0.034469	0	0	0	0	0	0	1	-360	360;
	7596	11009	0.006169	0.03512	0	0	0	0	0	0	1	-360	360;
	4467	8002	0.002961	0.017651	0	0	0	0	0	0	1	-360	360;
	4467	4131	0.00299	0.02462	0	0	0	0	0	0	1	-360	360;
	386	13576	0.002789	0.011659	0	0	0	0	0	0	1	-360	360;
	386	9823	0.00249	0.013211	0	0	0	0	0	0	1	-360	360;
	10827	13576	0.002669	0.016031	0	0	0	0	0	0	1	-360	360;
	10827	6896	0.00261	0.014841	0	0	0	0	0	0	1	-360	360;
	8613	13576	0.00074	0.004599	0	0	0	0	0	0	1	-360	360;
	8613	4717	0.000719	0.00468	0	0	0	0	0	0	1	-360	360;
	2868	6889	3.5e-05	0.000355	0	0	0	0	0	0	1	-360	360;
	2868	7065	3.5e-05	0.000355	0	0	0	0	0	0	1	-360	360;
	2031	372	0.00511	0.02949	0	0	0	0	0	0	1	-360	360;
	2031	8334	0.007719	0.044271	0	0	0	0	0	0	1	-360	360;
	4223	372	0.004281	0.022469	0	0	0	0	0	0	1	-360	360;
	4223	4491	0.005961	0.03382	0	0	0	0	0	0	1	-360	360;
	7174	6169	0.00073	0.01034	0	0	0	0	0	0	1	-360	360;
	7174	10838	0.00076	0.00889	0	0	0	0	0	0	1	-360	360;
	3739	3750	0.005159	0.025789	0	0	0	0	0	0	1	-360	360;
	3739	13415	0.00549	0.027711	0	0	0	0	0	0	1	-360	360;
	2999	3750	0.00614	0.01764	0	0	0	0	0	0	1	-360	360;
	2999	2132	0.003229	0.01788	0	0	0	0	0	0	1	-360	360;
	12353	7021	0.00053	0.00587	0	0	0	0	0	0	1	-360	360;
	12353	8085	0.00055	0.00666	0	0	0	0	0	0	1	-360	360;
	8372	9726	0.00326	0.01388	0	0	0	0	0	0	1	-360	360;
	8372	5972	0.003081	0.01488	0	0	0	0	0	0	1	-360	360;
	4977	2449	0.00051	0.00614	0	0	0	0	0	0	1	-360	360;
	4977	9117	0.00043	0.00406	0	0	0	0	0	0	1	-360	360;
	11807	9193	0.00761	0.046031	0	0	0	0	0	0	1	-360	360;
	11807	13362	0.00351	0.01632	0	0	0	0	0	0	1	-360	360;
	7367	2297	0.00031	0.00495	0	0	0	0	0	0	1	-360	360;
	7367	13504	0.00043	0.00485	0	0	0	0	0	0	1	-360	360;
	13361	210	0.00426	0.027099	0	0	0	0	0	0	1	-360	360;
	13361	8579	0.00437	0.024969	0	0	0	0	0	0	1	-360	360;
	8856	11596	0.005	0.03932	0	0	0	0	0	0	1	-360	360;
	8856	10943	0.006039	0.036789	0	0	0	0	0	0	1	-360	360;
	12165	8452	0.0009	0.00893	0	0	0	0	0	0	1	-360	360;
	12165	10718	0.00286	0.02156	0	0	0	0	0	0	1	-360	360;
	991	210	0.00418	0.022729	0	0	0	0	0	0	1	-360	360;
	991	8264	0.00405	0.02231	0	0	0	0	0	0	1	-360	360;
	10687	7078	0.00013	0.00675	0	0	0	0	0	0	1	-360	360;
	10687	1766	0.00013	0.00688	0	0	0	0	0	0	1	-360	360;
	13494	6000	0.00078	0.01201	0	0	0	0	0	0	1	-360	360;
	13494	9558	0.00075	0.01254	0	0	0	0	0	0	1	-360	360;
	3322	4944	0.00071	0.00722	0	0	0	0	0	0	1	-360	360;
	3322	6366	0.00084	0.0074	0	0	0	0	0	0	1	-360	360;
	6082	13221	0.00105	0.0081	0	0	0	0	0	0	1	-360	360;
	6082	11561	0.00203	0.01565	0	0	0	0	0	0	1	-360	360;
	12539	2263	0.00035	0.00581	0	0	0	0	0	0	1	-360	360;
	12539	3777	0.00045	0.00534	0	0	0	0	0	0	1	-360	360;
	12308	11016	8.1e-05	0.001331	0	0	0	0	0	0	1	-360	360;
	12308	9194	0.00025	0.001591	0	0	0	0	0	0	1	-360	360;
	10168	3267	0.00401	0.026219	0	0	0	0	0	0	1	-360	360;
	10168	10943	0.004669	0.027521	0	0	0	0	0	0	1	-360	360;
	460	12666	0.00115	0.01265	0	0	0	0	0	0	1	-360	360;
	460	1925	0.00112	0.0129	0	0	0	0	0	0	1	-360	360;
	11384	10686	0.00123	0.01589	0	0	0	0	0	0	1	-360	360;
	11384	11561	0.00054	0.00484	0	0	0	0	0	0	1	-360	360;
	5551	2914	0.00401	0.03961	0	0	0	0	0	0	1	-360	360;
	5551	12265	0.002909	0.01981	0	0	0	0	0	0	1	-360	360;
	1418	8986	0.002211	0.014419	0	0	0	0	0	0	1	-360	360;
	1418	4609	0.002169	0.013349	0	0	0	0	0	0	1	-360	360;
	5389	12666	0.00055	0.00508	0	0	0	0	0	0	1	-360	360;
	5389	1255	0.00046	0.00595	0	0	0	0	0	0	1	-360	360;
	4132	10934	0.00201	0.00988	0	0	0	0	0	0	1	-360	360;
	4132	3129	0.003401	0.019099	0	0	0	0	0	0	1	-360	360;
	2817	992	0.00028	0.00308	0	0	0	0	0	0	1	-360	360;
	2817	8261	0.00038	0.00307	0	0	0	0	0	0	1	-360	360;
	7950	6182	0.00044	0.00577	0	0	0	0	0	0	1	-360	360;
	7950	9582	0.00102	0.01295	0	0	0	0	0	0	1	-360	360;
	8170	2362	0.00085	0.01005	0	0	0	0	0	0	1	-360	360;
	8170	8003	0.00073	0.0097	0	0	0	0	0	0	1	-360	360;
	11189	10774	0.00139	0.0126	0	0	0	0	0	0	1	-360	360;
	11189	8008	0.00117	0.01389	0	0	0	0	0	0	1	-360	360;
	9620	2145	0.00076	0.00908	0	0	0	0	0	0	1	-360	360;
	9620	12632	0.00137	0.01338	0	0	0	0	0	0	1	-360	360;
	12114	3721	0.006	0.0402	0	0	0	0	0	0	1	-360	360;
	12114	6284	0.005719	0.03906	0	0	0	0	0	0	1	-360	360;
	3244	10774	0.0007	0.01065	0	0	0	0	0	0	1	-360	360;
	3244	3272	0.00066	0.01062	0	0	0	0	0	0	1	-360	360;
	6020	494	0.00153	0.01695	0	0	0	0	0	0	1	-360	360;
	6020	8891	0.00083	0.00925	0	0	0	0	0	0	1	-360	360;
	12682	7713	0.00033	0.00101	0	0	0	0	0	0	1	-360	360;
	11711	7713	0.00028	0.00104	0	0	0	0	0	0	1	-360	360;
	1504	884	0.00139	0.0125	0	0	0	0	0	0	1	-360	360;
	1504	5821	0.00119	0.01236	0	0	0	0	0	0	1	-360	360;
	3381	12919	0.010099	0.05513	0	0	0	0	0	0	1	-360	360;
	3381	879	0.00464	0.027021	0	0	0	0	0	0	1	-360	360;
	7853	6824	0.002211	0.012031	0	0	0	0	0	0	1	-360	360;
	7853	5972	0.002479	0.0115	0	0	0	0	0	0	1	-360	360;
	8048	4231	0.030281	0.058314	0	0	0	0	0	0	1	-360	360;
	8048	174	0.028174	0.054256	0	0	0	0	0	0	1	-360	360;
	1392	4231	0.01114	0.026231	0	0	0	0	0	0	1	-360	360;
	1392	3084	0.018157	0.045248	0	0	0	0	0	0	1	-360	360;
	9306	524	0.00369	0.017159	0	0	0	0	0	0	1	-360	360;
	9306	11008	0.004909	0.03018	0	0	0	0	0	0	1	-360	360;
	12771	10677	0.00408	0.0056	0	0	0	0	0	0	1	-360	360;
	12771	11249	0.00301	0.00925	0	0	0	0	0	0	1	-360	360;
	6857	524	0.001599	0.00489	0	0	0	0	0	0	1	-360	360;
	6857	9037	0.00031	0.00175	0	0	0	0	0	0	1	-360	360;
	8995	524	0.00163	0.004789	0	0	0	0	0	0	1	-360	360;
	8995	9037	0.00031	0.001789	0	0	0	0	0	0	1	-360	360;
	4390	4711	0.00988	0.063229	0	0	0	0	0	0	1	-360	360;
	4390	2245	0.00114	0.00638	0	0	0	0	0	0	1	-360	360;
	993	10347	0.00065	0.00755	0	0	0	0	0	0	1	-360	360;
	993	10709	0.00052	0.00503	0	0	0	0	0	0	1	-360	360;
	826	10347	0.00063	0.0077	0	0	0	0	0	0	1	-360	360;
	826	10709	0.00053	0.00493	0	0	0	0	0	0	1	-360	360;
	3394	4877	0.00055	0.00683	0	0	0	0	0	0	1	-360	360;
	3394	12158	0.00118	0.01134	0	0	0	0	0	0	1	-360	360;
	9489	4877	0.00056	0.00668	0	0	0	0	0	0	1	-360	360;
	9489	12158	0.0012	0.0111	0	0	0	0	0	0	1	-360	360;
	9708	13164	0.0098	0.01047	0	0	0	0	0	0	1	-360	360;
	9708	4642	0.00134	0.01215	0	0	0	0	0	0	1	-360	360;
	9498	13164	0.00092	0.01198	0	0	0	0	0	0	1	-360	360;
	9498	3715	0.00057	0.00616	0	0	0	0	0	0	1	-360	360;
	12196	10919	0.00108	0.01773	0	0	0	0	0	0	1	-360	360;
	12196	5789	0.00064	0.01047	0	0	0	0	0	0	1	-360	360;
	535	11908	0.002211	0.013341	0	0	0	0	0	0	1	-360	360;
	535	140	0.001901	0.014229	0	0	0	0	0	0	1	-360	360;
	7945	1076	0.00193	0.011771	0	0	0	0	0	0	1	-360	360;
	7945	12534	0.00251	0.014909	0	0	0	0	0	0	1	-360	360;
	9005	892	6e-05	0.00084	0	0	0	0	0	0	1	-360	360;
	9005	1766	6e-05	0.00092	0	0	0	0	0	0	1	-360	360;
	12416	8601	0.00138	0.010349	0	0	0	0	0	0	1	-360	360;
	12416	10021	0.00156	0.009151	0	0	0	0	0	0	1	-360	360;
	9221	8601	0.001349	0.010539	0	0	0	0	0	0	1	-360	360;
	9221	10021	0.001591	0.00895	0	0	0	0	0	0	1	-360	360;
	8515	2315	0	0.000271	0	0	0	0	0	0	1	-360	360;
	6403	6614	0.00057	0.0071	0	0	0	0	0	0	1	-360	360;
	6403	9117	0.00057	0.00697	0	0	0	0	0	0	1	-360	360;
	9514	6056	0.000969	0.00743	0	0	0	0	0	0	1	-360	360;
	9514	7101	0.00099	0.007271	0	0	0	0	0	0	1	-360	360;
	1452	6056	0.00088	0.00824	0	0	0	0	0	0	1	-360	360;
	1452	7101	0.00086	0.008401	0	0	0	0	0	0	1	-360	360;
	5682	5452	0.00293	0.01664	0	0	0	0	0	0	1	-360	360;
	5682	11733	0.008719	0.05249	0	0	0	0	0	0	1	-360	360;
	4825	2456	0.00057	0.00627	0	0	0	0	0	0	1	-360	360;
	4825	11561	0.00088	0.00878	0	0	0	0	0	0	1	-360	360;
	13359	326	0.00033	0.00421	0	0	0	0	0	0	1	-360	360;
	13359	12737	0.00084	0.00353	0	0	0	0	0	0	1	-360	360;
	624	11012	0.0001	0.00119	0	0	0	0	0	0	1	-360	360;
	624	8261	0.0001	0.0013	0	0	0	0	0	0	1	-360	360;
	7026	7780	9e-05	0.00121	0	0	0	0	0	0	1	-360	360;
	7026	6366	0.0001	0.00134	0	0	0	0	0	0	1	-360	360;
	4755	4357	0.00063	0.01067	0	0	0	0	0	0	1	-360	360;
	4755	3462	0.00073	0.00959	0	0	0	0	0	0	1	-360	360;
	10662	5711	0.002021	0.030271	0	0	0	0	0	0	1	-360	360;
	10662	7437	0.00218	0.02786	0	0	0	0	0	0	1	-360	360;
	7214	4357	0.00086	0.00781	0	0	0	0	0	0	1	-360	360;
	7214	6268	0.00092	0.00666	0	0	0	0	0	0	1	-360	360;
	8164	5711	0.003969	0.0272	0	0	0	0	0	0	1	-360	360;
	8164	8677	0.00311	0.029729	0	0	0	0	0	0	1	-360	360;
	12629	4357	0.00055	0.00837	0	0	0	0	0	0	1	-360	360;
	12629	4831	0.00053	0.0084	0	0	0	0	0	0	1	-360	360;
	7840	4357	0.00053	0.00853	0	0	0	0	0	0	1	-360	360;
	7840	4831	0.00052	0.00778	0	0	0	0	0	0	1	-360	360;
	11317	5711	0.0017	0.008159	0	0	0	0	0	0	1	-360	360;
	11317	11518	0.001729	0.00799	0	0	0	0	0	0	1	-360	360;
	3240	5711	0.00176	0.00781	0	0	0	0	0	0	1	-360	360;
	3240	11518	0.00157	0.008849	0	0	0	0	0	0	1	-360	360;
	1135	5711	0.000219	0.002229	0	0	0	0	0	0	1	-360	360;
	1135	7516	0.000229	0.00218	0	0	0	0	0	0	1	-360	360;
	2781	5711	0.000229	0.00213	0	0	0	0	0	0	1	-360	360;
	2781	7516	0.000211	0.002419	0	0	0	0	0	0	1	-360	360;
	6505	5711	0.00019	0.00255	0	0	0	0	0	0	1	-360	360;
	6505	7516	0.000211	0.0025	0	0	0	0	0	0	1	-360	360;
	12092	5711	0	0.0002	0	0	0	0	0	0	1	-360	360;
	12092	5740	0	0.0002	0	0	0	0	0	0	1	-360	360;
	2757	11750	0.00067	0.01058	0	0	0	0	0	0	1	-360	360;
	2757	10922	0.00065	0.01078	0	0	0	0	0	0	1	-360	360;
	5460	404	0.009281	0.05299	0	0	0	0	0	0	1	-360	360;
	5460	13262	0.009419	0.057219	0	0	0	0	0	0	1	-360	360;
	7572	3427	0.00055	0.00505	0	0	0	0	0	0	1	-360	360;
	7572	9194	9.9e-05	0.00164	0	0	0	0	0	0	1	-360	360;
	11268	11750	0.00064	0.00998	0	0	0	0	0	0	1	-360	360;
	11268	3389	0.00216	0.02943	0	0	0	0	0	0	1	-360	360;
	7272	3049	0.00039	0.00474	0	0	0	0	0	0	1	-360	360;
	7272	2493	0.00045	0.00467	0	0	0	0	0	0	1	-360	360;
	11286	3049	0.00052	0.00567	0	0	0	0	0	0	1	-360	360;
	11286	2409	0.00054	0.00667	0	0	0	0	0	0	1	-360	360;
	6560	6485	0.008909	0.042349	0	0	0	0	0	0	1	-360	360;
	6560	12261	0.00713	0.03293	0	0	0	0	0	0	1	-360	360;
	13149	2024	0.0021	0.02015	0	0	0	0	0	0	1	-360	360;
	13149	13575	0.00177	0.02084	0	0	0	0	0	0	1	-360	360;
	9609	249	0.001461	0.00781	0	0	0	0	0	0	1	-360	360;
	9609	8612	0.001651	0.007651	0	0	0	0	0	0	1	-360	360;
	12837	249	0.00168	0.0075	0	0	0	0	0	0	1	-360	360;
	12837	8612	0.001719	0.007341	0	0	0	0	0	0	1	-360	360;
	11653	10180	0.00626	0.020771	0	0	0	0	0	0	1	-360	360;
	11653	10180	0.00638	0.02032	0	0	0	0	0	0	1	-360	360;
	111	1302	0.047983	0.076364	0	0	0	0	0	0	1	-360	360;
	111	2235	0.010868	0.039083	0	0	0	0	0	0	1	-360	360;
	6531	5081	0.021851	0.032364	0	0	0	0	0	0	1	-360	360;
	10859	5081	0.075008	0.121636	0	0	0	0	0	0	1	-360	360;
	5081	7553	0.049273	0.084942	0	0	0	0	0	0	1	-360	360;
	8666	8494	0.058521	0.110017	0	0	0	0	0	0	1	-360	360;
	11001	9689	0.012992	0.052719	0	0	0	0	0	0	1	-360	360;
	9689	11001	0.012727	0.053752	0	0	0	0	0	0	1	-360	360;
	6310	3549	0.013341	0.054169	0	0	0	0	0	0	1	-360	360;
	7106	6310	0.014961	0.083521	0	0	0	0	0	0	1	-360	360;
	10458	2358	0.023653	0.033603	0	0	0	0	0	0	1	-360	360;
	10458	2358	0.024099	0.032876	0	0	0	0	0	0	1	-360	360;
	2358	11001	0.024934	0.109579	0	0	0	0	0	0	1	-360	360;
	2358	11001	0.024397	0.111653	0	0	0	0	0	0	1	-360	360;
	2358	8780	0.03414	0.051058	0	0	0	0	0	0	1	-360	360;
	1302	6531	0.018893	0.029083	0	0	0	0	0	0	1	-360	360;
	3549	8241	0.000581	0.003081	0	0	0	0	0	0	1	-360	360;
	3549	8241	0.0005	0.002789	0	0	0	0	0	0	1	-360	360;
	3549	2972	0.009281	0.03489	0	0	0	0	0	0	1	-360	360;
	8720	3549	0.006081	0.034	0	0	0	0	0	0	1	-360	360;
	8720	3549	0.00595	0.034651	0	0	0	0	0	0	1	-360	360;
	2421	10458	0.05019	0.076339	0	0	0	0	0	0	1	-360	360;
	10458	5586	0.05724	0.101868	0	0	0	0	0	0	1	-360	360;
	12153	10717	0.0013	0.01496	0	0	0	0	0	0	1	-360	360;
	2209	12153	0.00108	0.01115	0	0	0	0	0	0	1	-360	360;
	12153	850	0.00223	0.02267	0	0	0	0	0	0	1	-360	360;
	11653	2972	0.008229	0.03325	0	0	0	0	0	0	1	-360	360;
	11653	8698	0.018159	0.06526	0	0	0	0	0	0	1	-360	360;
	10859	13562	0.02257	0.034777	0	0	0	0	0	0	1	-360	360;
	10859	3783	0.032339	0.103017	0	0	0	0	0	0	1	-360	360;
	10859	3783	0.031678	0.105041	0	0	0	0	0	0	1	-360	360;
	10859	1184	0.028331	0.050182	0	0	0	0	0	0	1	-360	360;
	10859	1302	0.047314	0.063678	0	0	0	0	0	0	1	-360	360;
	7106	10559	0.0045	0.02618	0	0	0	0	0	0	1	-360	360;
	7106	8720	0.01913	0.10681	0	0	0	0	0	0	1	-360	360;
	12213	8666	0.059273	0.08086	0	0	0	0	0	0	1	-360	360;
	2530	12344	0.008041	0.014298	0	0	0	0	0	0	1	-360	360;
	6827	2530	0.047802	0.073562	0	0	0	0	0	0	1	-360	360;
	5477	2421	0.049636	0.058818	0	0	0	0	0	0	1	-360	360;
	13562	2235	0.04738	0.075992	0	0	0	0	0	0	1	-360	360;
	8780	13068	0.035603	0.045331	0	0	0	0	0	0	1	-360	360;
	4150	7553	0.024529	0.031678	0	0	0	0	0	0	1	-360	360;
	13157	4150	0.036851	0.047521	0	0	0	0	0	0	1	-360	360;
	9286	12283	0.004058	0.011041	0	0	0	0	0	0	1	-360	360;
	6779	9286	0.012818	0.022554	0	0	0	0	0	0	1	-360	360;
	6672	2361	0.079256	0.118479	0	0	0	0	0	0	1	-360	360;
	6779	6672	0.012554	0.022967	0	0	0	0	0	0	1	-360	360;
	12283	11915	0.007917	0.029628	0	0	0	0	0	0	1	-360	360;
	12283	11915	0.007744	0.027438	0	0	0	0	0	0	1	-360	360;
	2209	5117	0.00039	0.00461	0	0	0	0	0	0	1	-360	360;
	10559	414	0.014039	0.071669	0	0	0	0	0	0	1	-360	360;
	10559	414	0.01251	0.081219	0	0	0	0	0	0	1	-360	360;
	414	8698	0.0045	0.016831	0	0	0	0	0	0	1	-360	360;
	8720	414	0.00261	0.009401	0	0	0	0	0	0	1	-360	360;
	10560	3783	0.08024	0.135628	0	0	0	0	0	0	1	-360	360;
	6348	10560	0.043587	0.057314	0	0	0	0	0	0	1	-360	360;
	10559	119	0.00326	0.01895	0	0	0	0	0	0	1	-360	360;
	1076	119	0.005409	0.02675	0	0	0	0	0	0	1	-360	360;
	12344	4935	0.083537	0.116554	0	0	0	0	0	0	1	-360	360;
	3658	4935	0.012463	0.054793	0	0	0	0	0	0	1	-360	360;
	1184	13068	0.044992	0.07586	0	0	0	0	0	0	1	-360	360;
	10482	2361	0.032777	0.050628	0	0	0	0	0	0	1	-360	360;
	5586	6827	0.060314	0.092818	0	0	0	0	0	0	1	-360	360;
	13157	12213	0.02462	0.034967	0	0	0	0	0	0	1	-360	360;
	11915	5477	0.026603	0.03038	0	0	0	0	0	0	1	-360	360;
	50	7481	0.01595	0.022281	0	0	0	0	0	0	1	-360	360;
	10482	7481	0.018083	0.021843	0	0	0	0	0	0	1	-360	360;
	119	8720	0.0132	0.06293	0	0	0	0	0	0	1	-360	360;
	119	8720	0.01345	0.06162	0	0	0	0	0	0	1	-360	360;
	6779	5477	0.010612	0.023017	0	0	0	0	0	0	1	-360	360;
	1409	11040	0.01601	0.06706	0	0	0	0	0	0	1	-360	360;
	1409	8152	0.00362	0.01457	0	0	0	0	0	0	1	-360	360;
	2043	6999	0.010281	0.04007	0	0	0	0	0	0	1	-360	360;
	2043	13576	0.01549	0.06756	0	0	0	0	0	0	1	-360	360;
	2594	7565	0.00073	0.01141	0	0	0	0	0	0	1	-360	360;
	2594	11750	0.00108	0.01698	0	0	0	0	0	0	1	-360	360;
	4109	756	0.0022	0.027409	0	0	0	0	0	0	1	-360	360;
	4109	756	0.002151	0.02794	0	0	0	0	0	0	1	-360	360;
	4109	11112	0.00445	0.02582	0	0	0	0	0	0	1	-360	360;
	4109	404	0.005099	0.02426	0	0	0	0	0	0	1	-360	360;
	4109	2461	0.019539	0.100021	0	0	0	0	0	0	1	-360	360;
	4109	2461	0.01914	0.101979	0	0	0	0	0	0	1	-360	360;
	7565	3133	0.00158	0.02551	0	0	0	0	0	0	1	-360	360;
	7565	6866	0.00021	0.00296	0	0	0	0	0	0	1	-360	360;
	7565	8236	0.00157	0.02091	0	0	0	0	0	0	1	-360	360;
	1929	6866	0.00162	0.02163	0	0	0	0	0	0	1	-360	360;
	11180	6999	0.008521	0.03463	0	0	0	0	0	0	1	-360	360;
	11180	13576	0.00463	0.078531	0	0	0	0	0	0	1	-360	360;
	11180	13576	0.004531	0.072711	0	0	0	0	0	0	1	-360	360;
	11180	164	0.00575	0.02876	0	0	0	0	0	0	1	-360	360;
	11180	4559	0.00411	0.01826	0	0	0	0	0	0	1	-360	360;
	11180	2461	0.008091	0.03594	0	0	0	0	0	0	1	-360	360;
	5617	2883	0.011031	0.060581	0	0	0	0	0	0	1	-360	360;
	5617	8736	0.01326	0.07163	0	0	0	0	0	0	1	-360	360;
	7417	2611	0.007031	0.069341	0	0	0	0	0	0	1	-360	360;
	7417	8152	0.00238	0.031081	0	0	0	0	0	0	1	-360	360;
	6999	13576	0.010831	0.057479	0	0	0	0	0	0	1	-360	360;
	6999	164	0.001849	0.01024	0	0	0	0	0	0	1	-360	360;
	11040	9874	0.017581	0.07664	0	0	0	0	0	0	1	-360	360;
	11040	9874	0.017909	0.07501	0	0	0	0	0	0	1	-360	360;
	11040	8736	0.01312	0.05724	0	0	0	0	0	0	1	-360	360;
	11040	8736	0.01337	0.056021	0	0	0	0	0	0	1	-360	360;
	11040	813	0.007961	0.0347	0	0	0	0	0	0	1	-360	360;
	7425	13576	0.01181	0.044729	0	0	0	0	0	0	1	-360	360;
	7425	177	0.00457	0.018729	0	0	0	0	0	0	1	-360	360;
	12423	4778	0.0012	0.0141	0	0	0	0	0	0	1	-360	360;
	12423	4778	0.00107	0.01598	0	0	0	0	0	0	1	-360	360;
	12423	4219	0.00046	0.00666	0	0	0	0	0	0	1	-360	360;
	12423	3969	0.001	0.0181	0	0	0	0	0	0	1	-360	360;
	1646	5450	9.1e-05	0.00139	0	0	0	0	0	0	1	-360	360;
	1646	5450	0.00013	0.00089	0	0	0	0	0	0	1	-360	360;
	11112	404	0.000281	0.001289	0	0	0	0	0	0	1	-360	360;
	8760	5140	0.00119	0.01964	0	0	0	0	0	0	1	-360	360;
	8760	5140	0.00117	0.02001	0	0	0	0	0	0	1	-360	360;
	2611	6181	0.00238	0.032341	0	0	0	0	0	0	1	-360	360;
	2611	6181	0.002331	0.02994	0	0	0	0	0	0	1	-360	360;
	2611	5376	0.005479	0.07155	0	0	0	0	0	0	1	-360	360;
	2782	404	0.008091	0.036841	0	0	0	0	0	0	1	-360	360;
	8152	5376	0.003081	0.03707	0	0	0	0	0	0	1	-360	360;
	8152	813	0.01086	0.05936	0	0	0	0	0	0	1	-360	360;
	4219	1353	0.00063	0.00975	0	0	0	0	0	0	1	-360	360;
	13576	479	0.00143	0.025771	0	0	0	0	0	0	1	-360	360;
	13576	479	0.001401	0.026271	0	0	0	0	0	0	1	-360	360;
	13576	177	0.00631	0.03037	0	0	0	0	0	0	1	-360	360;
	2883	8736	0.00325	0.01388	0	0	0	0	0	0	1	-360	360;
	479	5450	0.015461	0.0828	0	0	0	0	0	0	1	-360	360;
	479	5450	0.017521	0.08114	0	0	0	0	0	0	1	-360	360;
	8236	11750	0.00036	0.00554	0	0	0	0	0	0	1	-360	360;
	3969	11750	0.0018	0.02089	0	0	0	0	0	0	1	-360	360;
	1353	11750	0.00183	0.02045	0	0	0	0	0	0	1	-360	360;
	5140	672	0.00052	0.01489	0	0	0	0	0	0	1	-360	360;
	5140	672	0.00051	0.01518	0	0	0	0	0	0	1	-360	360;
	8509	8736	0.020669	0.110081	0	0	0	0	0	0	1	-360	360;
	8509	8736	0.02024	0.1122	0	0	0	0	0	0	1	-360	360;
	446	404	0.01089	0.05145	0	0	0	0	0	0	1	-360	360;
	446	404	0.01111	0.050401	0	0	0	0	0	0	1	-360	360;
	4559	2461	0.00493	0.026719	0	0	0	0	0	0	1	-360	360;
	5711	249	0.005169	0.067461	0	0	0	0	0	0	1	-360	360;
	5711	249	0.00506	0.068729	0	0	0	0	0	0	1	-360	360;
	2723	10043	0.011397	0.04438	0	0	0	0	0	0	1	-360	360;
	11775	12394	0.022504	0.082851	0	0	0	0	0	0	1	-360	360;
	11775	5793	0.015793	0.061587	0	0	0	0	0	0	1	-360	360;
	8315	3750	0.012659	0.05405	0	0	0	0	0	0	1	-360	360;
	1058	8142	0.016149	0.057248	0	0	0	0	0	0	1	-360	360;
	218	8142	0.038612	0.122471	0	0	0	0	0	0	1	-360	360;
	8142	10533	0.044273	0.108959	0	0	0	0	0	0	1	-360	360;
	13106	8657	0.038876	0.139587	0	0	0	0	0	0	1	-360	360;
	13106	9368	0.10805	0.191992	0	0	0	0	0	0	1	-360	360;
	4476	13106	0.035802	0.084231	0	0	0	0	0	0	1	-360	360;
	4940	7549	0.00614	0.02205	0	0	0	0	0	0	1	-360	360;
	4940	3440	0.003587	0.01395	0	0	0	0	0	0	1	-360	360;
	7549	11388	0.019554	0.05981	0	0	0	0	0	0	1	-360	360;
	3440	4836	0.038066	0.142281	0	0	0	0	0	0	1	-360	360;
	11973	6342	0.007421	0.02181	0	0	0	0	0	0	1	-360	360;
	6342	12033	0.007562	0.021339	0	0	0	0	0	0	1	-360	360;
	11973	7034	0.007388	0.026537	0	0	0	0	0	0	1	-360	360;
	7034	7169	0.00724	0.02705	0	0	0	0	0	0	1	-360	360;
	109	10356	0.004893	0.018273	0	0	0	0	0	0	1	-360	360;
	10356	8071	0.004777	0.018612	0	0	0	0	0	0	1	-360	360;
	11973	9284	0.015289	0.041463	0	0	0	0	0	0	1	-360	360;
	11973	9284	0.01143	0.042091	0	0	0	0	0	0	1	-360	360;
	109	11973	0.006793	0.014182	0	0	0	0	0	0	1	-360	360;
	8071	11973	0.012967	0.048471	0	0	0	0	0	0	1	-360	360;
	8981	9284	0.009281	0.029554	0	0	0	0	0	0	1	-360	360;
	8594	9284	0.041248	0.116479	0	0	0	0	0	0	1	-360	360;
	5578	9284	0.071479	0.137678	0	0	0	0	0	0	1	-360	360;
	3312	9284	0.020612	0.075893	0	0	0	0	0	0	1	-360	360;
	9284	755	0.05214	0.100421	0	0	0	0	0	0	1	-360	360;
	6982	9284	0.047942	0.179157	0	0	0	0	0	0	1	-360	360;
	10639	3604	0.0016	0.01402	0	0	0	0	0	0	1	-360	360;
	8611	8981	0.009149	0.027992	0	0	0	0	0	0	1	-360	360;
	12735	8611	0.031107	0.116281	0	0	0	0	0	0	1	-360	360;
	12735	8071	0.026463	0.091273	0	0	0	0	0	0	1	-360	360;
	3659	12735	0.023562	0.075017	0	0	0	0	0	0	1	-360	360;
	8311	12735	0.015702	0.050017	0	0	0	0	0	0	1	-360	360;
	6443	4836	0.004802	0.017231	0	0	0	0	0	0	1	-360	360;
	6443	1728	0.065579	0.208826	0	0	0	0	0	0	1	-360	360;
	6443	5108	0.049917	0.112843	0	0	0	0	0	0	1	-360	360;
	6443	193	0.047793	0.095603	0	0	0	0	0	0	1	-360	360;
	10834	4836	0.043421	0.168942	0	0	0	0	0	0	1	-360	360;
	11388	4836	0.029364	0.093017	0	0	0	0	0	0	1	-360	360;
	4179	5603	0.023521	0.086562	0	0	0	0	0	0	1	-360	360;
	1375	5603	0.017909	0.058463	0	0	0	0	0	0	1	-360	360;
	5519	788	0.008198	0.030248	0	0	0	0	0	0	1	-360	360;
	788	4492	0.015579	0.053744	0	0	0	0	0	0	1	-360	360;
	788	8742	0.030471	0.09738	0	0	0	0	0	0	1	-360	360;
	4018	788	0.017157	0.066	0	0	0	0	0	0	1	-360	360;
	1797	6603	0.017628	0.065868	0	0	0	0	0	0	1	-360	360;
	1797	6100	0.022686	0.072273	0	0	0	0	0	0	1	-360	360;
	3163	12374	0.010322	0.03719	0	0	0	0	0	0	1	-360	360;
	12374	11018	0.010802	0.038281	0	0	0	0	0	0	1	-360	360;
	7290	10533	0.019298	0.073041	0	0	0	0	0	0	1	-360	360;
	4645	7290	0.011678	0.035273	0	0	0	0	0	0	1	-360	360;
	7411	6603	0.019719	0.057256	0	0	0	0	0	0	1	-360	360;
	251	1058	0.026421	0.086471	0	0	0	0	0	0	1	-360	360;
	12033	2738	0.014281	0.045471	0	0	0	0	0	0	1	-360	360;
	1375	10286	0.004529	0.013686	0	0	0	0	0	0	1	-360	360;
	10286	11018	0.014603	0.058463	0	0	0	0	0	0	1	-360	360;
	9827	8117	0.007702	0.020876	0	0	0	0	0	0	1	-360	360;
	9827	5067	0.010281	0.03957	0	0	0	0	0	0	1	-360	360;
	8117	7411	0.00686	0.023661	0	0	0	0	0	0	1	-360	360;
	5519	7418	0.02457	0.09557	0	0	0	0	0	0	1	-360	360;
	24	1728	0.03024	0.085372	0	0	0	0	0	0	1	-360	360;
	24	10834	0.04157	0.14924	0	0	0	0	0	0	1	-360	360;
	8786	24	0.028132	0.097008	0	0	0	0	0	0	1	-360	360;
	24	12828	0.038661	0.123091	0	0	0	0	0	0	1	-360	360;
	2498	12905	0.019364	0.077413	0	0	0	0	0	0	1	-360	360;
	12905	5793	0.009562	0.036182	0	0	0	0	0	0	1	-360	360;
	10639	8195	0.00147	0.01643	0	0	0	0	0	0	1	-360	360;
	1021	9078	0.005719	0.016826	0	0	0	0	0	0	1	-360	360;
	9078	13644	0.007843	0.027041	0	0	0	0	0	0	1	-360	360;
	9078	1827	0.016826	0.045628	0	0	0	0	0	0	1	-360	360;
	9078	3138	0.02238	0.062777	0	0	0	0	0	0	1	-360	360;
	1728	1021	0.012959	0.041248	0	0	0	0	0	0	1	-360	360;
	13644	6445	0.004702	0.017562	0	0	0	0	0	0	1	-360	360;
	11306	13644	0.023331	0.080446	0	0	0	0	0	0	1	-360	360;
	2127	1375	0.00505	0.018653	0	0	0	0	0	0	1	-360	360;
	6393	2127	0.008562	0.025686	0	0	0	0	0	0	1	-360	360;
	5110	4258	0.026124	0.083149	0	0	0	0	0	0	1	-360	360;
	4258	6982	0.02038	0.04605	0	0	0	0	0	0	1	-360	360;
	11322	9657	0.000169	0.00106	0	0	0	0	0	0	1	-360	360;
	1360	11322	0.00043	0.003901	0	0	0	0	0	0	1	-360	360;
	11217	6803	0.005298	0.021	0	0	0	0	0	0	1	-360	360;
	6803	10688	0.001322	0.004289	0	0	0	0	0	0	1	-360	360;
	9414	1600	0.005471	0.016554	0	0	0	0	0	0	1	-360	360;
	9414	7154	0.022198	0.064653	0	0	0	0	0	0	1	-360	360;
	7016	3955	0.027653	0.099256	0	0	0	0	0	0	1	-360	360;
	3766	7016	0.044099	0.152107	0	0	0	0	0	0	1	-360	360;
	4313	4582	0.017603	0.047711	0	0	0	0	0	0	1	-360	360;
	4313	6393	0.024198	0.06562	0	0	0	0	0	0	1	-360	360;
	12339	5338	0.00165	0.01428	0	0	0	0	0	0	1	-360	360;
	10677	12339	0.00131	0.01467	0	0	0	0	0	0	1	-360	360;
	8657	13227	0.040802	0.129909	0	0	0	0	0	0	1	-360	360;
	1331	3955	0.005058	0.019686	0	0	0	0	0	0	1	-360	360;
	3955	1970	0.075521	0.145463	0	0	0	0	0	0	1	-360	360;
	1331	5161	0.024033	0.088488	0	0	0	0	0	0	1	-360	360;
	11063	7227	0.001599	0.007401	0	0	0	0	0	0	1	-360	360;
	1827	12394	0.025289	0.082934	0	0	0	0	0	0	1	-360	360;
	10213	10834	0.012421	0.035058	0	0	0	0	0	0	1	-360	360;
	4711	3925	0.00855	0.03774	0	0	0	0	0	0	1	-360	360;
	8594	7597	0.010686	0.03019	0	0	0	0	0	0	1	-360	360;
	6124	9027	0.010331	0.03438	0	0	0	0	0	0	1	-360	360;
	9027	9578	0.01581	0.050339	0	0	0	0	0	0	1	-360	360;
	11217	7154	0.005661	0.022744	0	0	0	0	0	0	1	-360	360;
	10440	12377	0.000636	0.002512	0	0	0	0	0	0	1	-360	360;
	5793	10440	0.005281	0.017521	0	0	0	0	0	0	1	-360	360;
	10440	2822	0.002826	0.00781	0	0	0	0	0	0	1	-360	360;
	1600	7310	0.018066	0.056876	0	0	0	0	0	0	1	-360	360;
	7310	6100	0.038537	0.144033	0	0	0	0	0	0	1	-360	360;
	7310	1769	0.049669	0.141083	0	0	0	0	0	0	1	-360	360;
	7310	10688	0.058628	0.17286	0	0	0	0	0	0	1	-360	360;
	7310	8822	0.00043	0.001273	0	0	0	0	0	0	1	-360	360;
	7310	10070	0.025917	0.093058	0	0	0	0	0	0	1	-360	360;
	2470	6124	0.02295	0.065	0	0	0	0	0	0	1	-360	360;
	8697	6100	0.001504	0.00486	0	0	0	0	0	0	1	-360	360;
	8697	6100	0.002182	0.006678	0	0	0	0	0	0	1	-360	360;
	4973	1360	0.002979	0.02218	0	0	0	0	0	0	1	-360	360;
	4973	1360	0.003021	0.022531	0	0	0	0	0	0	1	-360	360;
	3659	8311	0.020843	0.058843	0	0	0	0	0	0	1	-360	360;
	7418	3659	0.009331	0.027413	0	0	0	0	0	0	1	-360	360;
	3659	10098	0.009182	0.024769	0	0	0	0	0	0	1	-360	360;
	7708	524	0.00275	0.01488	0	0	0	0	0	0	1	-360	360;
	9783	4311	0.0001	0.00082	0	0	0	0	0	0	1	-360	360;
	9193	1360	0.01476	0.08713	0	0	0	0	0	0	1	-360	360;
	9193	7227	0.00212	0.011469	0	0	0	0	0	0	1	-360	360;
	10688	9647	0.013521	0.049959	0	0	0	0	0	0	1	-360	360;
	9647	1375	0.017793	0.07124	0	0	0	0	0	0	1	-360	360;
	12721	3750	0.01481	0.086021	0	0	0	0	0	0	1	-360	360;
	7227	12721	0.01686	0.07207	0	0	0	0	0	0	1	-360	360;
	11018	8245	0.012331	0.034413	0	0	0	0	0	0	1	-360	360;
	8347	8245	0.000777	0.00276	0	0	0	0	0	0	1	-360	360;
	4674	218	0.028421	0.08657	0	0	0	0	0	0	1	-360	360;
	2738	6770	0.023521	0.081132	0	0	0	0	0	0	1	-360	360;
	4255	3759	0.000901	0.005	0	0	0	0	0	0	1	-360	360;
	3759	4485	0.00263	0.01551	0	0	0	0	0	0	1	-360	360;
	10517	4597	0.017463	0.071669	0	0	0	0	0	0	1	-360	360;
	10517	2070	0.016826	0.065512	0	0	0	0	0	0	1	-360	360;
	4582	5663	0.019587	0.076612	0	0	0	0	0	0	1	-360	360;
	1958	2498	0.022537	0.071769	0	0	0	0	0	0	1	-360	360;
	4645	251	0.013256	0.04005	0	0	0	0	0	0	1	-360	360;
	13154	4674	0.046397	0.13586	0	0	0	0	0	0	1	-360	360;
	13154	833	0.026471	0.105893	0	0	0	0	0	0	1	-360	360;
	8069	10604	0.025504	0.08119	0	0	0	0	0	0	1	-360	360;
	7226	8069	0.027603	0.10738	0	0	0	0	0	0	1	-360	360;
	10533	5578	0.084843	0.156942	0	0	0	0	0	0	1	-360	360;
	7633	4492	0.030521	0.115653	0	0	0	0	0	0	1	-360	360;
	7169	11821	0.004	0.010851	0	0	0	0	0	0	1	-360	360;
	11821	7597	0.016017	0.043421	0	0	0	0	0	0	1	-360	360;
	7633	2723	0.027091	0.087488	0	0	0	0	0	0	1	-360	360;
	10875	10136	0.02243	0.047793	0	0	0	0	0	0	1	-360	360;
	11388	10136	0.016851	0.051537	0	0	0	0	0	0	1	-360	360;
	12377	12394	0.01638	0.053901	0	0	0	0	0	0	1	-360	360;
	12377	12394	0.012298	0.045521	0	0	0	0	0	0	1	-360	360;
	12394	10028	0.025017	0.048653	0	0	0	0	0	0	1	-360	360;
	5067	13194	0.00186	0.007174	0	0	0	0	0	0	1	-360	360;
	4276	13194	0.001397	0.005372	0	0	0	0	0	0	1	-360	360;
	3138	6445	0.010777	0.043479	0	0	0	0	0	0	1	-360	360;
	9326	3138	0.016413	0.047628	0	0	0	0	0	0	1	-360	360;
	2054	8025	0.004	0.011281	0	0	0	0	0	0	1	-360	360;
	6100	8025	0.004041	0.015099	0	0	0	0	0	0	1	-360	360;
	5067	6685	0.004636	0.017579	0	0	0	0	0	0	1	-360	360;
	6685	6100	0.006529	0.025091	0	0	0	0	0	0	1	-360	360;
	9578	5067	0.01781	0.052339	0	0	0	0	0	0	1	-360	360;
	2054	4276	0.003744	0.010149	0	0	0	0	0	0	1	-360	360;
	9724	6100	0.006248	0.023702	0	0	0	0	0	0	1	-360	360;
	9724	9368	0.035107	0.106207	0	0	0	0	0	0	1	-360	360;
	9724	13267	0.014769	0.045182	0	0	0	0	0	0	1	-360	360;
	9724	2692	0.004322	0.012198	0	0	0	0	0	0	1	-360	360;
	9724	10070	0.015678	0.054083	0	0	0	0	0	0	1	-360	360;
	44	10213	0.011331	0.023636	0	0	0	0	0	0	1	-360	360;
	1360	9657	0.000591	0.002969	0	0	0	0	0	0	1	-360	360;
	1360	9657	0.000281	0.0025	0	0	0	0	0	0	1	-360	360;
	7227	9657	0.00781	0.039159	0	0	0	0	0	0	1	-360	360;
	9657	1616	0.00419	0.02431	0	0	0	0	0	0	1	-360	360;
	524	9657	0.014479	0.06989	0	0	0	0	0	0	1	-360	360;
	524	9657	0.013419	0.07594	0	0	0	0	0	0	1	-360	360;
	5338	7701	0.00208	0.02183	0	0	0	0	0	0	1	-360	360;
	10604	5578	0.028083	0.103397	0	0	0	0	0	0	1	-360	360;
	8742	3163	0.005793	0.017802	0	0	0	0	0	0	1	-360	360;
	13227	13267	0.016124	0.049322	0	0	0	0	0	0	1	-360	360;
	44	5108	0.018471	0.041802	0	0	0	0	0	0	1	-360	360;
	1970	1536	0.031719	0.066198	0	0	0	0	0	0	1	-360	360;
	303	2470	0.017554	0.056124	0	0	0	0	0	0	1	-360	360;
	8879	1769	0.011207	0.037917	0	0	0	0	0	0	1	-360	360;
	3750	4255	0.02261	0.105521	0	0	0	0	0	0	1	-360	360;
	5578	7226	0.002397	0.013603	0	0	0	0	0	0	1	-360	360;
	7226	755	0.021901	0.055818	0	0	0	0	0	0	1	-360	360;
	5110	5578	0.007719	0.030066	0	0	0	0	0	0	1	-360	360;
	5578	3312	0.017711	0.05	0	0	0	0	0	0	1	-360	360;
	833	5578	0.035322	0.070802	0	0	0	0	0	0	1	-360	360;
	6770	8786	0.017281	0.050777	0	0	0	0	0	0	1	-360	360;
	3264	7633	0.04324	0.168223	0	0	0	0	0	0	1	-360	360;
	8822	3264	0.022678	0.082017	0	0	0	0	0	0	1	-360	360;
	3264	4018	0.026	0.079537	0	0	0	0	0	0	1	-360	360;
	7870	10079	0.027347	0.080372	0	0	0	0	0	0	1	-360	360;
	5161	7870	0.023364	0.074372	0	0	0	0	0	0	1	-360	360;
	4231	4476	0.011421	0.02757	0	0	0	0	0	0	1	-360	360;
	7227	4255	0.011039	0.047169	0	0	0	0	0	0	1	-360	360;
	1616	4255	0.015159	0.0732	0	0	0	0	0	0	1	-360	360;
	3840	4255	0.00119	0.005729	0	0	0	0	0	0	1	-360	360;
	11279	4255	0.00182	0.060349	0	0	0	0	0	0	1	-360	360;
	10688	5663	0.00795	0.032132	0	0	0	0	0	0	1	-360	360;
	1211	10688	0.006372	0.021934	0	0	0	0	0	0	1	-360	360;
	3766	10688	0.015992	0.036041	0	0	0	0	0	0	1	-360	360;
	10688	6393	0.022843	0.087157	0	0	0	0	0	0	1	-360	360;
	11097	4179	0.00495	0.014942	0	0	0	0	0	0	1	-360	360;
	1536	11097	0.000901	0.00238	0	0	0	0	0	0	1	-360	360;
	5663	11097	0.008942	0.027339	0	0	0	0	0	0	1	-360	360;
	5663	4179	0.00476	0.017587	0	0	0	0	0	0	1	-360	360;
	10079	4179	0.037339	0.128802	0	0	0	0	0	0	1	-360	360;
	1536	7191	0.00081	0.00219	0	0	0	0	0	0	1	-360	360;
	1536	6516	0.005132	0.018438	0	0	0	0	0	0	1	-360	360;
	6413	5663	0.005934	0.023107	0	0	0	0	0	0	1	-360	360;
	6413	5663	0.005802	0.021397	0	0	0	0	0	0	1	-360	360;
	6413	335	0.000463	0.001421	0	0	0	0	0	0	1	-360	360;
	6413	8879	0.000909	0.002521	0	0	0	0	0	0	1	-360	360;
	5663	6516	0.009347	0.025364	0	0	0	0	0	0	1	-360	360;
	13567	7191	0.00062	0.002017	0	0	0	0	0	0	1	-360	360;
	335	13567	0.000917	0.002364	0	0	0	0	0	0	1	-360	360;
	13567	3766	0.009248	0.016446	0	0	0	0	0	0	1	-360	360;
	8879	3826	0.00295	0.008	0	0	0	0	0	0	1	-360	360;
	3766	8879	0.003339	0.012612	0	0	0	0	0	0	1	-360	360;
	3766	3826	0.00319	0.012504	0	0	0	0	0	0	1	-360	360;
	1211	3766	0.007826	0.030521	0	0	0	0	0	0	1	-360	360;
	10387	7701	0.00063	0.00702	0	0	0	0	0	0	1	-360	360;
	7701	8195	0.00184	0.02018	0	0	0	0	0	0	1	-360	360;
	2057	8855	0.003818	0.010661	0	0	0	0	0	0	1	-360	360;
	11018	2057	0.008678	0.025198	0	0	0	0	0	0	1	-360	360;
	8347	2057	0.008843	0.024661	0	0	0	0	0	0	1	-360	360;
	193	10028	0.023347	0.041909	0	0	0	0	0	0	1	-360	360;
	12828	7597	0.018091	0.055347	0	0	0	0	0	0	1	-360	360;
	9651	1958	0.03443	0.093347	0	0	0	0	0	0	1	-360	360;
	9651	2070	0.041033	0.111248	0	0	0	0	0	0	1	-360	360;
	4444	11306	0.015843	0.05686	0	0	0	0	0	0	1	-360	360;
	6770	4444	0.020397	0.064959	0	0	0	0	0	0	1	-360	360;
	303	7633	0.017893	0.054959	0	0	0	0	0	0	1	-360	360;
	1375	8855	0.029496	0.09295	0	0	0	0	0	0	1	-360	360;
	13641	12377	0.011579	0.046397	0	0	0	0	0	0	1	-360	360;
	11279	4711	0.00099	0.0042	0	0	0	0	0	0	1	-360	360;
	11279	4711	0.00076	0.00449	0	0	0	0	0	0	1	-360	360;
	10255	2023	0.006587	0.016868	0	0	0	0	0	0	1	-360	360;
	2023	8855	0.016876	0.064909	0	0	0	0	0	0	1	-360	360;
	10255	7418	0.036612	0.115372	0	0	0	0	0	0	1	-360	360;
	4597	5793	0.009421	0.030826	0	0	0	0	0	0	1	-360	360;
	13641	5793	0.008041	0.030438	0	0	0	0	0	0	1	-360	360;
	7418	10098	0.000744	0.002182	0	0	0	0	0	0	1	-360	360;
	11279	6390	0.000901	0.003979	0	0	0	0	0	0	1	-360	360;
	11279	4485	0.01212	0.05176	0	0	0	0	0	0	1	-360	360;
	9783	8195	0.00256	0.02765	0	0	0	0	0	0	1	-360	360;
	2822	10875	0.00224	0.007256	0	0	0	0	0	0	1	-360	360;
	13164	8195	0.0009	0.00943	0	0	0	0	0	0	1	-360	360;
	4597	13641	0.007744	0.024273	0	0	0	0	0	0	1	-360	360;
	10875	13641	0.006372	0.017992	0	0	0	0	0	0	1	-360	360;
	2070	11388	0.048554	0.142669	0	0	0	0	0	0	1	-360	360;
	2692	13267	0.010802	0.030488	0	0	0	0	0	0	1	-360	360;
	9326	6708	0.015496	0.047397	0	0	0	0	0	0	1	-360	360;
	11018	8855	0.010207	0.030826	0	0	0	0	0	0	1	-360	360;
	8855	6708	0.026504	0.077876	0	0	0	0	0	0	1	-360	360;
	4482	962	0.00073	0.00905	0	0	0	0	0	0	1	-360	360;
	3183	7221	0.00192	0.02296	0	0	0	0	0	0	1	-360	360;
	2593	3359	0.01431	0.05899	0	0	0	0	0	0	1	-360	360;
	3359	2044	0.02607	0.07168	0	0	0	0	0	0	1	-360	360;
	8848	9981	0.00054	0.00465	0	0	0	0	0	0	1	-360	360;
	8848	3239	0.00126	0.01124	0	0	0	0	0	0	1	-360	360;
	8848	10523	0.00048	0.00527	0	0	0	0	0	0	1	-360	360;
	8848	9186	0.00082	0.00938	0	0	0	0	0	0	1	-360	360;
	4381	13586	0.00028	0.003502	0	0	0	0	0	0	1	-360	360;
	4381	5114	0.691467	4.414711	0	0	0	0	0	0	1	-360	360;
	4381	1350	0.02004	0.0916	0	0	0	0	0	0	1	-360	360;
	4381	342	0.841911	4.768444	0	0	0	0	0	0	1	-360	360;
	4381	9970	0.026	0.08948	0	0	0	0	0	0	1	-360	360;
	4381	13254	0.002631	0.011889	0	0	0	0	0	0	1	-360	360;
	4381	8395	0.7232	1.944	0	0	0	0	0	0	1	-360	360;
	13586	1350	0.156169	0.546756	0	0	0	0	0	0	1	-360	360;
	13586	9970	0.168809	0.580622	0	0	0	0	0	0	1	-360	360;
	13586	13254	0.019222	0.068209	0	0	0	0	0	0	1	-360	360;
	7727	2304	0.00039	0.00386	0	0	0	0	0	0	1	-360	360;
	7727	1668	0.00096	0.01138	0	0	0	0	0	0	1	-360	360;
	7727	13402	0.00037	0.00325	0	0	0	0	0	0	1	-360	360;
	7727	11873	0.00029	0.00339	0	0	0	0	0	0	1	-360	360;
	5114	3277	0.009458	0.158871	0	0	0	0	0	0	1	-360	360;
	3277	6494	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9970	3277	0.312951	7.130667	0	0	0	0	0	0	1	-360	360;
	11058	3277	0.247649	1.075911	0	0	0	0	0	0	1	-360	360;
	6499	3277	0.06056	0.18612	0	0	0	0	0	0	1	-360	360;
	12273	3277	2.142489	4.347733	0	0	0	0	0	0	1	-360	360;
	10883	3277	0.07132	1.068889	0	0	0	0	0	0	1	-360	360;
	412	3277	0.02268	0.08652	0	0	0	0	0	0	1	-360	360;
	3510	3277	2.007778	4.494222	0	0	0	0	0	0	1	-360	360;
	6844	3277	0.006178	0.028938	0	0	0	0	0	0	1	-360	360;
	6590	3277	0.027711	0.260951	0	0	0	0	0	0	1	-360	360;
	5114	8135	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6724	5114	0.00692	0.337378	0	0	0	0	0	0	1	-360	360;
	342	5114	0.025991	0.10048	0	0	0	0	0	0	1	-360	360;
	9970	5114	0.083711	0.8648	0	0	0	0	0	0	1	-360	360;
	11058	5114	0.447378	1.862089	0	0	0	0	0	0	1	-360	360;
	6499	5114	0.062169	0.153191	0	0	0	0	0	0	1	-360	360;
	12273	5114	2.492533	4.725778	0	0	0	0	0	0	1	-360	360;
	10883	5114	0.020271	0.08244	0	0	0	0	0	0	1	-360	360;
	412	5114	0.11416	1.777556	0	0	0	0	0	0	1	-360	360;
	8395	5114	0.574978	2.992578	0	0	0	0	0	0	1	-360	360;
	3510	5114	1.887067	4.964444	0	0	0	0	0	0	1	-360	360;
	6844	5114	0.250778	0.817422	0	0	0	0	0	0	1	-360	360;
	6590	5114	0.007049	0.029609	0	0	0	0	0	0	1	-360	360;
	8135	6724	0.003049	0.020218	0	0	0	0	0	0	1	-360	360;
	6494	6724	0.002911	0.020262	0	0	0	0	0	0	1	-360	360;
	7221	4112	4e-05	0.00041	0	0	0	0	0	0	1	-360	360;
	7221	962	0.00087	0.01087	0	0	0	0	0	0	1	-360	360;
	7221	1491	4e-05	0.00046	0	0	0	0	0	0	1	-360	360;
	7221	962	0.00095	0.00924	0	0	0	0	0	0	1	-360	360;
	2593	7602	0.528017	2.636157	0	0	0	0	0	0	1	-360	360;
	2593	2044	0.032031	0.128659	0	0	0	0	0	0	1	-360	360;
	2304	11382	0.00044	0.0032	0	0	0	0	0	0	1	-360	360;
	2304	3618	0.00046	0.00489	0	0	0	0	0	0	1	-360	360;
	13402	11382	0.0003	0.00352	0	0	0	0	0	0	1	-360	360;
	13402	129	0.00051	0.00454	0	0	0	0	0	0	1	-360	360;
	4826	962	0.00068	0.00746	0	0	0	0	0	0	1	-360	360;
	4826	12707	0.00093	0.01022	0	0	0	0	0	0	1	-360	360;
	12707	1937	4e-05	0.0004	0	0	0	0	0	0	1	-360	360;
	12707	11873	0.00078	0.00698	0	0	0	0	0	0	1	-360	360;
	1668	2394	4e-05	0.00043	0	0	0	0	0	0	1	-360	360;
	1668	10304	0.00025	0.00309	0	0	0	0	0	0	1	-360	360;
	1282	11382	0.00047	0.00538	0	0	0	0	0	0	1	-360	360;
	2985	11382	0.00046	0.00548	0	0	0	0	0	0	1	-360	360;
	2985	8729	0.00016	0.00173	0	0	0	0	0	0	1	-360	360;
	2985	9186	0.00166	0.01471	0	0	0	0	0	0	1	-360	360;
	1282	11382	0.00051	0.00512	0	0	0	0	0	0	1	-360	360;
	9805	11873	0.00021	0.00248	0	0	0	0	0	0	1	-360	360;
	342	6724	0.004391	0.021689	0	0	0	0	0	0	1	-360	360;
	3500	9186	0.00024	0.00289	0	0	0	0	0	0	1	-360	360;
	1350	5137	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1350	13254	0.01788	0.100471	0	0	0	0	0	0	1	-360	360;
	5137	4174	0.006458	0.039982	0	0	0	0	0	0	1	-360	360;
	342	9970	0.106151	0.993333	0	0	0	0	0	0	1	-360	360;
	342	8395	0.699778	3.232311	0	0	0	0	0	0	1	-360	360;
	13254	9970	0.257058	0.43592	0	0	0	0	0	0	1	-360	360;
	8395	9970	0.002769	0.011449	0	0	0	0	0	0	1	-360	360;
	962	5712	0.00034	0.00305	0	0	0	0	0	0	1	-360	360;
	962	10304	0.00123	0.01255	0	0	0	0	0	0	1	-360	360;
	962	13656	6e-05	0.00065	0	0	0	0	0	0	1	-360	360;
	962	11702	6e-05	0.00063	0	0	0	0	0	0	1	-360	360;
	962	9098	6e-05	0.00055	0	0	0	0	0	0	1	-360	360;
	962	7884	0.00079	0.00968	0	0	0	0	0	0	1	-360	360;
	962	61	7e-05	0.00079	0	0	0	0	0	0	1	-360	360;
	1142	4784	0	0.000222	0	0	0	0	0	0	1	-360	360;
	11058	1142	2.064178	7.545333	0	0	0	0	0	0	1	-360	360;
	9831	1142	0.153711	0.482489	0	0	0	0	0	0	1	-360	360;
	12273	1142	0.015809	0.12368	0	0	0	0	0	0	1	-360	360;
	3510	1142	0.057978	0.363422	0	0	0	0	0	0	1	-360	360;
	6844	1142	0.6932	2.196889	0	0	0	0	0	0	1	-360	360;
	6590	1142	0.560356	2.395867	0	0	0	0	0	0	1	-360	360;
	7884	3238	0.00024	0.00213	0	0	0	0	0	0	1	-360	360;
	7884	11179	0.00093	0.00884	0	0	0	0	0	0	1	-360	360;
	13254	3070	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5402	8729	0.00018	0.00204	0	0	0	0	0	0	1	-360	360;
	8371	8729	0.00018	0.00208	0	0	0	0	0	0	1	-360	360;
	10366	10981	0	0.000222	0	0	0	0	0	0	1	-360	360;
	11058	10366	0.02068	0.093391	0	0	0	0	0	0	1	-360	360;
	6499	10366	0.777778	7.408	0	0	0	0	0	0	1	-360	360;
	10366	9831	0.00884	0.07292	0	0	0	0	0	0	1	-360	360;
	412	10366	0.131671	8.295556	0	0	0	0	0	0	1	-360	360;
	10366	2298	0.082462	0.568222	0	0	0	0	0	0	1	-360	360;
	10981	2298	0.0042	0.02624	0	0	0	0	0	0	1	-360	360;
	12041	12038	0.00031	0.00317	0	0	0	0	0	0	1	-360	360;
	12041	11382	0.00043	0.00498	0	0	0	0	0	0	1	-360	360;
	13341	11382	0.00042	0.00508	0	0	0	0	0	0	1	-360	360;
	11058	6499	0.110671	0.773067	0	0	0	0	0	0	1	-360	360;
	11058	9831	0.00912	0.037609	0	0	0	0	0	0	1	-360	360;
	11058	12273	0.098049	0.598711	0	0	0	0	0	0	1	-360	360;
	11058	10883	0.019031	1.160356	0	0	0	0	0	0	1	-360	360;
	11058	412	0.024071	0.915733	0	0	0	0	0	0	1	-360	360;
	3510	11058	1.484267	4.479111	0	0	0	0	0	0	1	-360	360;
	6844	11058	0.011578	0.063271	0	0	0	0	0	0	1	-360	360;
	6590	11058	0.654889	1.839111	0	0	0	0	0	0	1	-360	360;
	11058	2298	0.022049	0.10076	0	0	0	0	0	0	1	-360	360;
	6499	9831	0.459867	2.604711	0	0	0	0	0	0	1	-360	360;
	6499	12273	2.320222	6.088	0	0	0	0	0	0	1	-360	360;
	6499	10883	0.006369	0.048089	0	0	0	0	0	0	1	-360	360;
	6499	412	0.006631	0.0432	0	0	0	0	0	0	1	-360	360;
	3510	6499	2.567244	6.767111	0	0	0	0	0	0	1	-360	360;
	6844	6499	0.309889	1.214178	0	0	0	0	0	0	1	-360	360;
	6590	6499	0.009191	0.043289	0	0	0	0	0	0	1	-360	360;
	6499	2298	0.691022	5.239556	0	0	0	0	0	0	1	-360	360;
	7602	2044	0.002669	0.01718	0	0	0	0	0	0	1	-360	360;
	12038	7626	0.00074	0.00694	0	0	0	0	0	0	1	-360	360;
	12038	11382	0.00068	0.008	0	0	0	0	0	0	1	-360	360;
	9831	1354	0	0.000222	0	0	0	0	0	0	1	-360	360;
	12273	9831	0.00524	0.030631	0	0	0	0	0	0	1	-360	360;
	10883	9831	0.33932	3.814667	0	0	0	0	0	0	1	-360	360;
	412	9831	0.22696	2.614933	0	0	0	0	0	0	1	-360	360;
	3510	9831	0.190671	0.719378	0	0	0	0	0	0	1	-360	360;
	6844	9831	1.050222	2.105556	0	0	0	0	0	0	1	-360	360;
	6590	9831	1.431644	4.294578	0	0	0	0	0	0	1	-360	360;
	2298	9831	0.022618	0.161089	0	0	0	0	0	0	1	-360	360;
	7626	2254	0.00088	0.00895	0	0	0	0	0	0	1	-360	360;
	12273	2191	0	0.000222	0	0	0	0	0	0	1	-360	360;
	10883	12273	1.827778	7.768	0	0	0	0	0	0	1	-360	360;
	3510	12273	0.020849	0.153382	0	0	0	0	0	0	1	-360	360;
	6844	12273	0.156231	0.656267	0	0	0	0	0	0	1	-360	360;
	6590	12273	0.176498	0.891556	0	0	0	0	0	0	1	-360	360;
	11382	3239	0.00067	0.00777	0	0	0	0	0	0	1	-360	360;
	10883	412	0.004578	0.061889	0	0	0	0	0	0	1	-360	360;
	10883	4174	0.01348	0.05932	0	0	0	0	0	0	1	-360	360;
	6844	10883	0.066391	2.802622	0	0	0	0	0	0	1	-360	360;
	6590	10883	0.194258	2.220444	0	0	0	0	0	0	1	-360	360;
	10883	2298	0.008151	0.072822	0	0	0	0	0	0	1	-360	360;
	412	4174	0.016502	0.055382	0	0	0	0	0	0	1	-360	360;
	6844	412	0.0806	4.633778	0	0	0	0	0	0	1	-360	360;
	6590	412	0.831556	3.729644	0	0	0	0	0	0	1	-360	360;
	412	2298	0.007951	0.074151	0	0	0	0	0	0	1	-360	360;
	2254	8510	0.0004	0.00362	0	0	0	0	0	0	1	-360	360;
	10304	13103	0.00015	0.00177	0	0	0	0	0	0	1	-360	360;
	8395	12108	0	0.000222	0	0	0	0	0	0	1	-360	360;
	471	9098	2e-05	0.00022	0	0	0	0	0	0	1	-360	360;
	6844	3510	0.02864	0.172902	0	0	0	0	0	0	1	-360	360;
	3510	6590	0.017591	0.1404	0	0	0	0	0	0	1	-360	360;
	6844	6590	0.00752	0.031711	0	0	0	0	0	0	1	-360	360;
	2298	6473	0	0.000222	0	0	0	0	0	0	1	-360	360;
	12971	3259	0.032587	0.058388	0	0	0	0	0	0	1	-360	360;
	929	3259	0.029752	0.070686	0	0	0	0	0	0	1	-360	360;
	6446	5892	0.00699	0.04244	0	0	0	0	0	0	1	-360	360;
	5892	11019	0.0008	0.006081	0	0	0	0	0	0	1	-360	360;
	5892	8749	0.004669	0.041919	0	0	0	0	0	0	1	-360	360;
	5892	8749	0.00476	0.04106	0	0	0	0	0	0	1	-360	360;
	10227	2900	0.00146	0.0163	0	0	0	0	0	0	1	-360	360;
	10227	2900	0.00098	0.01485	0	0	0	0	0	0	1	-360	360;
	10227	2900	0.00143	0.01661	0	0	0	0	0	0	1	-360	360;
	2145	10227	0.00209	0.0241	0	0	0	0	0	0	1	-360	360;
	2145	10227	0.00273	0.02677	0	0	0	0	0	0	1	-360	360;
	2145	10227	0.00279	0.02623	0	0	0	0	0	0	1	-360	360;
	8032	1164	0.007479	0.032207	0	0	0	0	0	0	1	-360	360;
	8032	3008	0.04319	0.16186	0	0	0	0	0	0	1	-360	360;
	2305	4806	0.007256	0.029702	0	0	0	0	0	0	1	-360	360;
	4806	8823	0.006488	0.036041	0	0	0	0	0	0	1	-360	360;
	5561	6684	0.028653	0.066364	0	0	0	0	0	0	1	-360	360;
	10396	5561	0.010463	0.027446	0	0	0	0	0	0	1	-360	360;
	2035	4589	0.012409	0.071919	0	0	0	0	0	0	1	-360	360;
	2035	4589	0.012659	0.07045	0	0	0	0	0	0	1	-360	360;
	4589	12640	0.003419	0.02344	0	0	0	0	0	0	1	-360	360;
	4589	10889	0.002469	0.01486	0	0	0	0	0	0	1	-360	360;
	4589	7538	0.00682	0.03368	0	0	0	0	0	0	1	-360	360;
	4589	1292	0.02036	0.09743	0	0	0	0	0	0	1	-360	360;
	12971	5532	0.048	0.080331	0	0	0	0	0	0	1	-360	360;
	12971	4992	0.00262	0.008083	0	0	0	0	0	0	1	-360	360;
	12971	9566	0.004107	0.023083	0	0	0	0	0	0	1	-360	360;
	12971	11748	0.003207	0.015719	0	0	0	0	0	0	1	-360	360;
	12971	1037	0.004727	0.02143	0	0	0	0	0	0	1	-360	360;
	12971	6848	0.004041	0.012702	0	0	0	0	0	0	1	-360	360;
	12971	3676	0.017802	0.037777	0	0	0	0	0	0	1	-360	360;
	4201	7895	0.002934	0.014273	0	0	0	0	0	0	1	-360	360;
	5373	4201	0.006231	0.032174	0	0	0	0	0	0	1	-360	360;
	8818	4224	0.026876	0.085248	0	0	0	0	0	0	1	-360	360;
	8818	1740	0.050636	0.099223	0	0	0	0	0	0	1	-360	360;
	1384	4853	0.045347	0.092	0	0	0	0	0	0	1	-360	360;
	10215	4853	0.005231	0.013504	0	0	0	0	0	0	1	-360	360;
	13420	8567	0.024752	0.057446	0	0	0	0	0	0	1	-360	360;
	13420	2099	0.015083	0.065182	0	0	0	0	0	0	1	-360	360;
	13420	10947	0.03986	0.070686	0	0	0	0	0	0	1	-360	360;
	7589	13420	0.036231	0.064248	0	0	0	0	0	0	1	-360	360;
	13420	503	0.004537	0.022041	0	0	0	0	0	0	1	-360	360;
	13470	11054	0.052603	0.157802	0	0	0	0	0	0	1	-360	360;
	10947	11054	0.005562	0.024083	0	0	0	0	0	0	1	-360	360;
	10947	11054	0.005446	0.024554	0	0	0	0	0	0	1	-360	360;
	5318	11054	0.009107	0.051661	0	0	0	0	0	0	1	-360	360;
	11054	6936	0.00419	0.009868	0	0	0	0	0	0	1	-360	360;
	6629	11054	0.024281	0.13338	0	0	0	0	0	0	1	-360	360;
	6629	11054	0.02376	0.123496	0	0	0	0	0	0	1	-360	360;
	7337	11054	0.005331	0.020752	0	0	0	0	0	0	1	-360	360;
	6703	7885	0.001107	0.004603	0	0	0	0	0	0	1	-360	360;
	7885	3706	0.005628	0.030702	0	0	0	0	0	0	1	-360	360;
	1037	11926	0.024628	0.130893	0	0	0	0	0	0	1	-360	360;
	1037	11926	0.024132	0.133463	0	0	0	0	0	0	1	-360	360;
	8682	11926	0.01419	0.07062	0	0	0	0	0	0	1	-360	360;
	8682	11926	0.014471	0.069182	0	0	0	0	0	0	1	-360	360;
	9872	12389	0.008372	0.045231	0	0	0	0	0	0	1	-360	360;
	6788	9872	0.024628	0.097983	0	0	0	0	0	0	1	-360	360;
	9872	13113	0.002471	0.009331	0	0	0	0	0	0	1	-360	360;
	500	2673	0.025702	0.144587	0	0	0	0	0	0	1	-360	360;
	6143	500	0.003421	0.019752	0	0	0	0	0	0	1	-360	360;
	5184	56	0.004562	0.017174	0	0	0	0	0	0	1	-360	360;
	56	1284	0.018868	0.083661	0	0	0	0	0	0	1	-360	360;
	56	2267	0.01019	0.027917	0	0	0	0	0	0	1	-360	360;
	13072	5184	0.019471	0.064058	0	0	0	0	0	0	1	-360	360;
	9130	2400	0.053446	0.118207	0	0	0	0	0	0	1	-360	360;
	2400	8054	0.049521	0.123438	0	0	0	0	0	0	1	-360	360;
	3834	7401	0.00914	0.042039	0	0	0	0	0	0	1	-360	360;
	7401	1292	0.004711	0.021591	0	0	0	0	0	0	1	-360	360;
	2028	13653	0	0.000893	0	0	0	0	0	0	1	-360	360;
	6883	2028	0.00405	0.021132	0	0	0	0	0	0	1	-360	360;
	6883	2028	0.004587	0.020711	0	0	0	0	0	0	1	-360	360;
	13653	5429	0.044182	0.091041	0	0	0	0	0	0	1	-360	360;
	13653	11192	0.022653	0.047421	0	0	0	0	0	0	1	-360	360;
	13653	10958	0.058876	0.108463	0	0	0	0	0	0	1	-360	360;
	670	5241	0.013174	0.026992	0	0	0	0	0	0	1	-360	360;
	10751	670	0.001273	0.002818	0	0	0	0	0	0	1	-360	360;
	2419	2305	0.009471	0.048769	0	0	0	0	0	0	1	-360	360;
	8688	5087	0.017388	0.04138	0	0	0	0	0	0	1	-360	360;
	5087	10725	0.017339	0.054661	0	0	0	0	0	0	1	-360	360;
	7895	6750	0.025033	0.137942	0	0	0	0	0	0	1	-360	360;
	7895	6750	0.024521	0.140653	0	0	0	0	0	0	1	-360	360;
	6750	10524	0.021198	0.056041	0	0	0	0	0	0	1	-360	360;
	6750	12031	0.007174	0.035	0	0	0	0	0	0	1	-360	360;
	3835	6750	0.010826	0.029992	0	0	0	0	0	0	1	-360	360;
	3835	6750	0.012876	0.028058	0	0	0	0	0	0	1	-360	360;
	3454	7619	0.034248	0.143124	0	0	0	0	0	0	1	-360	360;
	8790	7619	0.015421	0.066364	0	0	0	0	0	0	1	-360	360;
	7589	8567	0.014512	0.035901	0	0	0	0	0	0	1	-360	360;
	9285	2035	0.00863	0.054419	0	0	0	0	0	0	1	-360	360;
	9285	2035	0.008461	0.055479	0	0	0	0	0	0	1	-360	360;
	704	6273	0.013256	0.073248	0	0	0	0	0	0	1	-360	360;
	13276	6273	0.004033	0.018174	0	0	0	0	0	0	1	-360	360;
	929	5532	0.016008	0.027066	0	0	0	0	0	0	1	-360	360;
	7618	5318	0.013537	0.070083	0	0	0	0	0	0	1	-360	360;
	11426	7618	0.004744	0.024876	0	0	0	0	0	0	1	-360	360;
	4992	3974	0.014364	0.037438	0	0	0	0	0	0	1	-360	360;
	12254	3207	0.010364	0.020314	0	0	0	0	0	0	1	-360	360;
	3207	5414	0.007289	0.018207	0	0	0	0	0	0	1	-360	360;
	12316	2133	0.011744	0.052446	0	0	0	0	0	0	1	-360	360;
	10920	12316	0.015438	0.062876	0	0	0	0	0	0	1	-360	360;
	13321	3581	0.00157	0.01612	0	0	0	0	0	0	1	-360	360;
	13321	3581	0.00154	0.01643	0	0	0	0	0	0	1	-360	360;
	5373	7895	0.001967	0.008438	0	0	0	0	0	0	1	-360	360;
	7895	2339	0.01162	0.033653	0	0	0	0	0	0	1	-360	360;
	1643	1248	0.006901	0.04661	0	0	0	0	0	0	1	-360	360;
	1643	1248	0.00782	0.04568	0	0	0	0	0	0	1	-360	360;
	4107	1248	0.005401	0.0343	0	0	0	0	0	0	1	-360	360;
	969	1248	0.011419	0.07275	0	0	0	0	0	0	1	-360	360;
	6446	8749	0.01044	0.05919	0	0	0	0	0	0	1	-360	360;
	13205	1542	0.050314	0.112132	0	0	0	0	0	0	1	-360	360;
	13205	1542	0.057017	0.109893	0	0	0	0	0	0	1	-360	360;
	1542	6267	0.007	0.028231	0	0	0	0	0	0	1	-360	360;
	1542	12445	0.010719	0.040868	0	0	0	0	0	0	1	-360	360;
	1542	5669	0.04443	0.10976	0	0	0	0	0	0	1	-360	360;
	1542	827	0.021372	0.121033	0	0	0	0	0	0	1	-360	360;
	1542	1243	0.020579	0.095331	0	0	0	0	0	0	1	-360	360;
	4186	13205	0.02386	0.06986	0	0	0	0	0	0	1	-360	360;
	13109	4186	0.062901	0.105628	0	0	0	0	0	0	1	-360	360;
	4186	3008	0.035686	0.149314	0	0	0	0	0	0	1	-360	360;
	9516	4186	0.00181	0.006876	0	0	0	0	0	0	1	-360	360;
	2888	4186	0.002628	0.01462	0	0	0	0	0	0	1	-360	360;
	2888	4186	0.002983	0.014331	0	0	0	0	0	0	1	-360	360;
	620	4186	0.006157	0.03114	0	0	0	0	0	0	1	-360	360;
	2879	10925	0.022	0.043066	0	0	0	0	0	0	1	-360	360;
	2009	2879	0.013438	0.033479	0	0	0	0	0	0	1	-360	360;
	2879	12615	0.013488	0.028893	0	0	0	0	0	0	1	-360	360;
	4993	13570	0.114058	0.230033	0	0	0	0	0	0	1	-360	360;
	13570	1328	0.062901	0.133777	0	0	0	0	0	0	1	-360	360;
	13570	2823	0.041868	0.106554	0	0	0	0	0	0	1	-360	360;
	4986	12249	0.052347	0.07786	0	0	0	0	0	0	1	-360	360;
	12249	8102	0.003099	0.006174	0	0	0	0	0	0	1	-360	360;
	1249	10766	0.04357	0.108678	0	0	0	0	0	0	1	-360	360;
	1249	12030	0.020893	0.043289	0	0	0	0	0	0	1	-360	360;
	6684	10725	0.035289	0.072058	0	0	0	0	0	0	1	-360	360;
	3008	13205	0.050579	0.074298	0	0	0	0	0	0	1	-360	360;
	3120	13205	0.042934	0.087917	0	0	0	0	0	0	1	-360	360;
	1164	10414	0.036711	0.086397	0	0	0	0	0	0	1	-360	360;
	10414	12254	0.053124	0.100066	0	0	0	0	0	0	1	-360	360;
	10414	5974	0.022959	0.041727	0	0	0	0	0	0	1	-360	360;
	5834	2346	0.00476	0.023868	0	0	0	0	0	0	1	-360	360;
	11341	5834	0.005529	0.011182	0	0	0	0	0	0	1	-360	360;
	11068	1054	0.00183	0.02821	0	0	0	0	0	0	1	-360	360;
	10919	11068	0.00198	0.02074	0	0	0	0	0	0	1	-360	360;
	13146	8672	0.022917	0.094893	0	0	0	0	0	0	1	-360	360;
	13146	8672	0.023347	0.092868	0	0	0	0	0	0	1	-360	360;
	8672	4319	0.030364	0.073256	0	0	0	0	0	0	1	-360	360;
	8672	9445	0.004562	0.020132	0	0	0	0	0	0	1	-360	360;
	8672	9445	0.010562	0.020686	0	0	0	0	0	0	1	-360	360;
	1328	8672	0.01838	0.07314	0	0	0	0	0	0	1	-360	360;
	8672	7355	0.005628	0.029702	0	0	0	0	0	0	1	-360	360;
	8672	7355	0.005777	0.029273	0	0	0	0	0	0	1	-360	360;
	4023	8672	0.014529	0.236802	0	0	0	0	0	0	1	-360	360;
	3711	2475	0.005818	0.032182	0	0	0	0	0	0	1	-360	360;
	3711	6883	0.003636	0.021843	0	0	0	0	0	0	1	-360	360;
	5019	2900	0.00318	0.02624	0	0	0	0	0	0	1	-360	360;
	5019	10919	0.00442	0.03649	0	0	0	0	0	0	1	-360	360;
	6742	2355	0.005099	0.027868	0	0	0	0	0	0	1	-360	360;
	6742	2475	0.00995	0.049397	0	0	0	0	0	0	1	-360	360;
	6742	2475	0.01014	0.048397	0	0	0	0	0	0	1	-360	360;
	6742	2401	0.007917	0.04757	0	0	0	0	0	0	1	-360	360;
	6742	137	0.004008	0.022182	0	0	0	0	0	0	1	-360	360;
	6742	298	0.006603	0.039124	0	0	0	0	0	0	1	-360	360;
	12254	8213	0.00657	0.028281	0	0	0	0	0	0	1	-360	360;
	10766	12254	0.071471	0.124661	0	0	0	0	0	0	1	-360	360;
	10766	12254	0.069983	0.127058	0	0	0	0	0	0	1	-360	360;
	12254	5414	0.00743	0.018182	0	0	0	0	0	0	1	-360	360;
	13072	11429	0.004653	0.024339	0	0	0	0	0	0	1	-360	360;
	10725	13072	0.007157	0.03462	0	0	0	0	0	0	1	-360	360;
	10725	13072	0.007298	0.033909	0	0	0	0	0	0	1	-360	360;
	9904	6075	0.00024	0.00281	0	0	0	0	0	0	1	-360	360;
	7874	10614	0.010711	0.056603	0	0	0	0	0	0	1	-360	360;
	5506	7874	0.064273	0.068058	0	0	0	0	0	0	1	-360	360;
	11200	8129	0.013917	0.06676	0	0	0	0	0	0	1	-360	360;
	11200	12693	0.034298	0.076091	0	0	0	0	0	0	1	-360	360;
	6703	4558	0.00676	0.044942	0	0	0	0	0	0	1	-360	360;
	6703	6143	0.025579	0.109438	0	0	0	0	0	0	1	-360	360;
	6703	6143	0.026091	0.107198	0	0	0	0	0	0	1	-360	360;
	1864	4947	0.00319	0.01543	0	0	0	0	0	0	1	-360	360;
	4947	3957	0.003669	0.014628	0	0	0	0	0	0	1	-360	360;
	4122	2261	0.001653	0.007331	0	0	0	0	0	0	1	-360	360;
	704	4122	0.009661	0.04095	0	0	0	0	0	0	1	-360	360;
	2145	13321	0.00089	0.00981	0	0	0	0	0	0	1	-360	360;
	2145	13321	0.00087	0.01	0	0	0	0	0	0	1	-360	360;
	13321	6075	0.00114	0.01867	0	0	0	0	0	0	1	-360	360;
	7175	4001	0.034661	0.089231	0	0	0	0	0	0	1	-360	360;
	1580	7175	0.009463	0.018603	0	0	0	0	0	0	1	-360	360;
	4065	10381	0.032777	0.061248	0	0	0	0	0	0	1	-360	360;
	10029	4065	0.029182	0.043207	0	0	0	0	0	0	1	-360	360;
	9868	8688	0.012826	0.022587	0	0	0	0	0	0	1	-360	360;
	2806	13470	0.01519	0.065017	0	0	0	0	0	0	1	-360	360;
	11304	13470	0.002107	0.009	0	0	0	0	0	0	1	-360	360;
	13470	6629	0.010099	0.046587	0	0	0	0	0	0	1	-360	360;
	298	2355	0.001934	0.009008	0	0	0	0	0	0	1	-360	360;
	9130	10925	0.041537	0.116843	0	0	0	0	0	0	1	-360	360;
	6788	2099	0.013091	0.050041	0	0	0	0	0	0	1	-360	360;
	1565	10381	0.003058	0.005579	0	0	0	0	0	0	1	-360	360;
	2396	13146	0.008554	0.035397	0	0	0	0	0	0	1	-360	360;
	13616	7940	0.020314	0.102727	0	0	0	0	0	0	1	-360	360;
	7429	7940	0.005008	0.025364	0	0	0	0	0	0	1	-360	360;
	8054	7940	0.042182	0.161132	0	0	0	0	0	0	1	-360	360;
	11423	1862	0.00014	0.000581	0	0	0	0	0	0	1	-360	360;
	5014	3060	0.001339	0.007438	0	0	0	0	0	0	1	-360	360;
	3533	2261	0.004223	0.009347	0	0	0	0	0	0	1	-360	360;
	2261	8093	0.016314	0.034562	0	0	0	0	0	0	1	-360	360;
	13276	2261	0.009331	0.045339	0	0	0	0	0	0	1	-360	360;
	3533	5893	0.00038	0.002769	0	0	0	0	0	0	1	-360	360;
	13276	3533	0.007066	0.043107	0	0	0	0	0	0	1	-360	360;
	2626	2806	0.003124	0.017231	0	0	0	0	0	0	1	-360	360;
	3678	7956	0.020512	0.048562	0	0	0	0	0	0	1	-360	360;
	3120	7956	0.016934	0.03314	0	0	0	0	0	0	1	-360	360;
	13342	5373	0.011083	0.030769	0	0	0	0	0	0	1	-360	360;
	704	9652	0.024314	0.099174	0	0	0	0	0	0	1	-360	360;
	704	9652	0.024769	0.097058	0	0	0	0	0	0	1	-360	360;
	704	9652	0.025231	0.09495	0	0	0	0	0	0	1	-360	360;
	704	7536	0.012917	0.049612	0	0	0	0	0	0	1	-360	360;
	704	4787	0.048058	0.135157	0	0	0	0	0	0	1	-360	360;
	704	5893	0.009372	0.048876	0	0	0	0	0	0	1	-360	360;
	704	7941	0.00538	0.031752	0	0	0	0	0	0	1	-360	360;
	704	7941	0.005273	0.029397	0	0	0	0	0	0	1	-360	360;
	704	13185	0.044562	0.09762	0	0	0	0	0	0	1	-360	360;
	9652	5647	0.017876	0.097091	0	0	0	0	0	0	1	-360	360;
	9652	5203	0.037207	0.166008	0	0	0	0	0	0	1	-360	360;
	10919	13151	0.00018	0.00154	0	0	0	0	0	0	1	-360	360;
	10919	13151	0.00023	0.00194	0	0	0	0	0	0	1	-360	360;
	1784	573	0.005669	0.040521	0	0	0	0	0	0	1	-360	360;
	1784	349	0.001581	0.012159	0	0	0	0	0	0	1	-360	360;
	1784	349	0.001409	0.013781	0	0	0	0	0	0	1	-360	360;
	1784	5136	0.00111	0.0075	0	0	0	0	0	0	1	-360	360;
	1784	817	0.0115	0.07662	0	0	0	0	0	0	1	-360	360;
	3151	4224	0.002661	0.01286	0	0	0	0	0	0	1	-360	360;
	2475	6883	0.010793	0.046942	0	0	0	0	0	0	1	-360	360;
	1384	13616	0.023826	0.100992	0	0	0	0	0	0	1	-360	360;
	13616	11135	0.014603	0.07986	0	0	0	0	0	0	1	-360	360;
	13616	7429	0.026132	0.099818	0	0	0	0	0	0	1	-360	360;
	11794	7030	0.002289	0.012554	0	0	0	0	0	0	1	-360	360;
	11794	9130	0.00743	0.030719	0	0	0	0	0	0	1	-360	360;
	12314	4986	0.014653	0.029174	0	0	0	0	0	0	1	-360	360;
	8682	12314	0.016388	0.04286	0	0	0	0	0	0	1	-360	360;
	11192	13281	0.056248	0.122041	0	0	0	0	0	0	1	-360	360;
	506	13281	0.024901	0.049099	0	0	0	0	0	0	1	-360	360;
	6819	10524	0.026554	0.061851	0	0	0	0	0	0	1	-360	360;
	4993	9845	0.073678	0.305793	0	0	0	0	0	0	1	-360	360;
	4477	4993	0.029083	0.060017	0	0	0	0	0	0	1	-360	360;
	4993	95	0.021893	0.134521	0	0	0	0	0	0	1	-360	360;
	1873	8072	0.018983	0.03914	0	0	0	0	0	0	1	-360	360;
	8072	8263	0.024529	0.041612	0	0	0	0	0	0	1	-360	360;
	10342	7920	0.039421	0.085686	0	0	0	0	0	0	1	-360	360;
	10766	7920	0.039421	0.085769	0	0	0	0	0	0	1	-360	360;
	10607	10342	0.00876	0.016901	0	0	0	0	0	0	1	-360	360;
	5917	10607	0.006967	0.018248	0	0	0	0	0	0	1	-360	360;
	12340	3957	0.025099	0.107901	0	0	0	0	0	0	1	-360	360;
	5429	12340	0.02605	0.103438	0	0	0	0	0	0	1	-360	360;
	7589	3232	0.012719	0.033719	0	0	0	0	0	0	1	-360	360;
	3232	11426	0.022273	0.043661	0	0	0	0	0	0	1	-360	360;
	3079	7612	0.00124	0.006653	0	0	0	0	0	0	1	-360	360;
	3069	7612	0.001587	0.007636	0	0	0	0	0	0	1	-360	360;
	8971	5554	0.006207	0.032331	0	0	0	0	0	0	1	-360	360;
	8971	5554	0.006083	0.032942	0	0	0	0	0	0	1	-360	360;
	5554	5506	0.001314	0.005446	0	0	0	0	0	0	1	-360	360;
	6267	641	0.003851	0.01662	0	0	0	0	0	0	1	-360	360;
	4329	11808	0.005124	0.031934	0	0	0	0	0	0	1	-360	360;
	13276	11808	0.005008	0.029562	0	0	0	0	0	0	1	-360	360;
	863	12031	0.048331	0.094372	0	0	0	0	0	0	1	-360	360;
	863	3740	0.015678	0.073826	0	0	0	0	0	0	1	-360	360;
	3836	1226	0.005579	0.019174	0	0	0	0	0	0	1	-360	360;
	506	3836	0.029397	0.077132	0	0	0	0	0	0	1	-360	360;
	2401	1226	0.002901	0.006802	0	0	0	0	0	0	1	-360	360;
	2401	137	0.004769	0.020826	0	0	0	0	0	0	1	-360	360;
	2401	12891	0.031388	0.06305	0	0	0	0	0	0	1	-360	360;
	2401	5717	0.032587	0.068752	0	0	0	0	0	0	1	-360	360;
	2401	1918	0.042802	0.083983	0	0	0	0	0	0	1	-360	360;
	5647	5203	0.01214	0.069124	0	0	0	0	0	0	1	-360	360;
	4001	2673	0.008388	0.017033	0	0	0	0	0	0	1	-360	360;
	10947	12943	0.008521	0.046653	0	0	0	0	0	0	1	-360	360;
	10947	12545	0.003331	0.013818	0	0	0	0	0	0	1	-360	360;
	3834	573	0.016729	0.090401	0	0	0	0	0	0	1	-360	360;
	3834	8812	0.01526	0.081479	0	0	0	0	0	0	1	-360	360;
	3834	12618	0.00905	0.068219	0	0	0	0	0	0	1	-360	360;
	3834	2745	0.01205	0.08575	0	0	0	0	0	0	1	-360	360;
	3834	10184	0.007841	0.04113	0	0	0	0	0	0	1	-360	360;
	124	355	0.002413	0.010686	0	0	0	0	0	0	1	-360	360;
	1243	124	0.005041	0.022769	0	0	0	0	0	0	1	-360	360;
	3420	46	0.00126	0.007531	0	0	0	0	0	0	1	-360	360;
	11341	2931	0.00176	0.007603	0	0	0	0	0	0	1	-360	360;
	11341	2931	0.00157	0.00862	0	0	0	0	0	0	1	-360	360;
	13109	3678	0.02743	0.059157	0	0	0	0	0	0	1	-360	360;
	2396	11739	0.028992	0.060893	0	0	0	0	0	0	1	-360	360;
	5181	3454	0.012942	0.051851	0	0	0	0	0	0	1	-360	360;
	1565	5181	0.071157	0.104149	0	0	0	0	0	0	1	-360	360;
	1677	9791	0.002017	0.010686	0	0	0	0	0	0	1	-360	360;
	4536	9791	0.00195	0.009364	0	0	0	0	0	0	1	-360	360;
	4536	9791	0.001992	0.009174	0	0	0	0	0	0	1	-360	360;
	1873	834	0.027017	0.074521	0	0	0	0	0	0	1	-360	360;
	834	10342	0.025107	0.072339	0	0	0	0	0	0	1	-360	360;
	6374	5485	0.002942	0.006248	0	0	0	0	0	0	1	-360	360;
	5485	5620	0.006983	0.028678	0	0	0	0	0	0	1	-360	360;
	2673	8086	0.049636	0.077364	0	0	0	0	0	0	1	-360	360;
	13014	8086	0.071752	0.121198	0	0	0	0	0	0	1	-360	360;
	11304	8024	0.00219	0.009174	0	0	0	0	0	0	1	-360	360;
	8858	11304	0.018719	0.067983	0	0	0	0	0	0	1	-360	360;
	13276	7536	0.006339	0.025397	0	0	0	0	0	0	1	-360	360;
	10614	28	0.011876	0.064967	0	0	0	0	0	0	1	-360	360;
	9845	12392	0.004603	0.024603	0	0	0	0	0	0	1	-360	360;
	2547	9845	0.037934	0.102868	0	0	0	0	0	0	1	-360	360;
	4319	10695	0.020818	0.067711	0	0	0	0	0	0	1	-360	360;
	3974	1677	0.020223	0.048471	0	0	0	0	0	0	1	-360	360;
	1677	5506	0.00705	0.034314	0	0	0	0	0	0	1	-360	360;
	1677	3032	0.015388	0.074008	0	0	0	0	0	0	1	-360	360;
	1677	3032	0.015066	0.075463	0	0	0	0	0	0	1	-360	360;
	1873	43	0.060331	0.128752	0	0	0	0	0	0	1	-360	360;
	43	10261	0.037752	0.085843	0	0	0	0	0	0	1	-360	360;
	43	929	0.060471	0.109992	0	0	0	0	0	0	1	-360	360;
	8263	43	0.021959	0.050058	0	0	0	0	0	0	1	-360	360;
	10342	5346	0.005347	0.026198	0	0	0	0	0	0	1	-360	360;
	5917	5346	0.005058	0.015769	0	0	0	0	0	0	1	-360	360;
	1384	1590	0.010256	0.053223	0	0	0	0	0	0	1	-360	360;
	5248	12392	0.01476	0.082157	0	0	0	0	0	0	1	-360	360;
	6819	11057	0.047843	0.112033	0	0	0	0	0	0	1	-360	360;
	1832	6111	0.013909	0.036793	0	0	0	0	0	0	1	-360	360;
	6111	4477	0.030289	0.064818	0	0	0	0	0	0	1	-360	360;
	6111	1580	0.026132	0.123132	0	0	0	0	0	0	1	-360	360;
	6111	1580	0.026628	0.120562	0	0	0	0	0	0	1	-360	360;
	5113	6111	0.016727	0.063959	0	0	0	0	0	0	1	-360	360;
	9725	7694	0.002413	0.010248	0	0	0	0	0	0	1	-360	360;
	9725	6143	0.009	0.040843	0	0	0	0	0	0	1	-360	360;
	641	12445	0.008421	0.018488	0	0	0	0	0	0	1	-360	360;
	4277	6863	0.014917	0.034322	0	0	0	0	0	0	1	-360	360;
	4277	3989	0.00743	0.035942	0	0	0	0	0	0	1	-360	360;
	7694	4277	0.003851	0.016	0	0	0	0	0	0	1	-360	360;
	7153	4277	0.00162	0.008008	0	0	0	0	0	0	1	-360	360;
	5093	1832	0.014372	0.045793	0	0	0	0	0	0	1	-360	360;
	7435	5093	0.038678	0.089967	0	0	0	0	0	0	1	-360	360;
	11429	6833	0.009504	0.021727	0	0	0	0	0	0	1	-360	360;
	6833	2267	0.018364	0.096521	0	0	0	0	0	0	1	-360	360;
	7355	9445	0.00195	0.010669	0	0	0	0	0	0	1	-360	360;
	9445	9055	0.021182	0.046826	0	0	0	0	0	0	1	-360	360;
	10751	11135	0.031752	0.131826	0	0	0	0	0	0	1	-360	360;
	8024	6629	0.002917	0.01238	0	0	0	0	0	0	1	-360	360;
	28	2544	0.016157	0.035826	0	0	0	0	0	0	1	-360	360;
	28	4542	0.003727	0.020529	0	0	0	0	0	0	1	-360	360;
	28	8739	0.019992	0.079231	0	0	0	0	0	0	1	-360	360;
	4558	2544	0.023653	0.094587	0	0	0	0	0	0	1	-360	360;
	4542	2168	0.009322	0.049058	0	0	0	0	0	0	1	-360	360;
	8682	2168	0.026727	0.146752	0	0	0	0	0	0	1	-360	360;
	7537	817	0.000531	0.002919	0	0	0	0	0	0	1	-360	360;
	832	9055	0.021992	0.059397	0	0	0	0	0	0	1	-360	360;
	832	7435	0.044413	0.10176	0	0	0	0	0	0	1	-360	360;
	5943	832	0.043223	0.088562	0	0	0	0	0	0	1	-360	360;
	5241	4787	0.006992	0.014066	0	0	0	0	0	0	1	-360	360;
	2419	7816	0.007256	0.028802	0	0	0	0	0	0	1	-360	360;
	2419	7816	0.006471	0.032636	0	0	0	0	0	0	1	-360	360;
	2419	13185	0.043198	0.156397	0	0	0	0	0	0	1	-360	360;
	3706	3707	0.016248	0.04014	0	0	0	0	0	0	1	-360	360;
	2035	1643	0.00206	0.01393	0	0	0	0	0	0	1	-360	360;
	2035	1643	0.002021	0.01419	0	0	0	0	0	0	1	-360	360;
	2035	46	0.00324	0.02076	0	0	0	0	0	0	1	-360	360;
	2035	46	0.00318	0.021159	0	0	0	0	0	0	1	-360	360;
	1873	10342	0.031603	0.070331	0	0	0	0	0	0	1	-360	360;
	3069	1873	0.003132	0.014554	0	0	0	0	0	0	1	-360	360;
	9375	1873	0.006669	0.032091	0	0	0	0	0	0	1	-360	360;
	503	12943	0.008934	0.046298	0	0	0	0	0	0	1	-360	360;
	1284	7760	0.002942	0.012008	0	0	0	0	0	0	1	-360	360;
	8213	5974	0.010942	0.062207	0	0	0	0	0	0	1	-360	360;
	8258	8619	0.002529	0.011777	0	0	0	0	0	0	1	-360	360;
	8258	743	0.015107	0.042769	0	0	0	0	0	0	1	-360	360;
	10980	13342	0.00705	0.022099	0	0	0	0	0	0	1	-360	360;
	8835	5644	0.038231	0.105893	0	0	0	0	0	0	1	-360	360;
	5774	8835	0.005901	0.021289	0	0	0	0	0	0	1	-360	360;
	170	7589	0.013636	0.022686	0	0	0	0	0	0	1	-360	360;
	10261	929	0.023826	0.059364	0	0	0	0	0	0	1	-360	360;
	10342	5172	0.016091	0.045967	0	0	0	0	0	0	1	-360	360;
	5172	6309	0.017033	0.053008	0	0	0	0	0	0	1	-360	360;
	1005	5644	0.017744	0.09686	0	0	0	0	0	0	1	-360	360;
	8619	4221	0.003636	0.018818	0	0	0	0	0	0	1	-360	360;
	8619	4221	0.003562	0.019174	0	0	0	0	0	0	1	-360	360;
	9594	8172	0.008256	0.021909	0	0	0	0	0	0	1	-360	360;
	8172	11429	0.026917	0.061636	0	0	0	0	0	0	1	-360	360;
	8331	6515	0.022702	0.057149	0	0	0	0	0	0	1	-360	360;
	6515	2339	0.023983	0.070124	0	0	0	0	0	0	1	-360	360;
	5318	3767	0.035231	0.078314	0	0	0	0	0	0	1	-360	360;
	3767	13014	0.02462	0.053603	0	0	0	0	0	0	1	-360	360;
	7030	7239	0.008091	0.03214	0	0	0	0	0	0	1	-360	360;
	6894	4936	0.016793	0.081388	0	0	0	0	0	0	1	-360	360;
	4936	1590	0.008636	0.047388	0	0	0	0	0	0	1	-360	360;
	4991	10342	0.009364	0.018207	0	0	0	0	0	0	1	-360	360;
	10342	12030	0.023331	0.063727	0	0	0	0	0	0	1	-360	360;
	10342	4833	0.014983	0.033579	0	0	0	0	0	0	1	-360	360;
	4536	8971	0.004636	0.019678	0	0	0	0	0	0	1	-360	360;
	4536	8971	0.004727	0.019273	0	0	0	0	0	0	1	-360	360;
	4329	226	0.018446	0.089686	0	0	0	0	0	0	1	-360	360;
	13276	4329	0.011529	0.056587	0	0	0	0	0	0	1	-360	360;
	3478	9130	0.00862	0.021587	0	0	0	0	0	0	1	-360	360;
	3478	4587	0.005702	0.013017	0	0	0	0	0	0	1	-360	360;
	9052	942	0.048256	0.078099	0	0	0	0	0	0	1	-360	360;
	942	6300	0.065529	0.159802	0	0	0	0	0	0	1	-360	360;
	8790	942	0.02681	0.110702	0	0	0	0	0	0	1	-360	360;
	95	2547	0.04538	0.142983	0	0	0	0	0	0	1	-360	360;
	1565	1864	0.024992	0.097777	0	0	0	0	0	0	1	-360	360;
	12991	1864	0.020256	0.109372	0	0	0	0	0	0	1	-360	360;
	1864	762	0.024636	0.095653	0	0	0	0	0	0	1	-360	360;
	3151	7429	0.003364	0.017669	0	0	0	0	0	0	1	-360	360;
	5414	13356	0.018463	0.039124	0	0	0	0	0	0	1	-360	360;
	13356	9309	0.021636	0.060793	0	0	0	0	0	0	1	-360	360;
	10666	170	0.01786	0.039711	0	0	0	0	0	0	1	-360	360;
	573	5136	0.00588	0.03186	0	0	0	0	0	0	1	-360	360;
	10215	10751	0.045058	0.177512	0	0	0	0	0	0	1	-360	360;
	10751	3918	0.003686	0.017711	0	0	0	0	0	0	1	-360	360;
	10751	3918	0.003612	0.018058	0	0	0	0	0	0	1	-360	360;
	10751	3918	0.007876	0.021066	0	0	0	0	0	0	1	-360	360;
	3221	10751	0.03186	0.159347	0	0	0	0	0	0	1	-360	360;
	13014	5248	0.034901	0.071347	0	0	0	0	0	0	1	-360	360;
	2640	7816	0.006339	0.033281	0	0	0	0	0	0	1	-360	360;
	2640	7816	0.006207	0.033917	0	0	0	0	0	0	1	-360	360;
	5203	7816	0.006207	0.034983	0	0	0	0	0	0	1	-360	360;
	9546	13276	0.006289	0.024661	0	0	0	0	0	0	1	-360	360;
	9546	11101	0.002554	0.01376	0	0	0	0	0	0	1	-360	360;
	1328	11739	0.003579	0.01486	0	0	0	0	0	0	1	-360	360;
	8129	2626	0.01495	0.02843	0	0	0	0	0	0	1	-360	360;
	2267	10816	0.011727	0.028934	0	0	0	0	0	0	1	-360	360;
	10816	10980	0.013843	0.048909	0	0	0	0	0	0	1	-360	360;
	2673	2626	0.032388	0.080711	0	0	0	0	0	0	1	-360	360;
	2673	2626	0.011719	0.074793	0	0	0	0	0	0	1	-360	360;
	2626	8858	0.00057	0.002612	0	0	0	0	0	0	1	-360	360;
	3405	1792	0.012554	0.026314	0	0	0	0	0	0	1	-360	360;
	8682	3405	0.022554	0.049397	0	0	0	0	0	0	1	-360	360;
	1328	4023	0.005438	0.021669	0	0	0	0	0	0	1	-360	360;
	2384	2145	0.00079	0.00721	0	0	0	0	0	0	1	-360	360;
	2384	5852	0.00023	0.00228	0	0	0	0	0	0	1	-360	360;
	11341	13523	0.002248	0.012083	0	0	0	0	0	0	1	-360	360;
	11341	13523	0.002198	0.012314	0	0	0	0	0	0	1	-360	360;
	10766	2709	0.014727	0.064099	0	0	0	0	0	0	1	-360	360;
	355	10766	0.016364	0.040901	0	0	0	0	0	0	1	-360	360;
	5669	10766	0.024471	0.063686	0	0	0	0	0	0	1	-360	360;
	12258	13646	0.00037	0.00429	0	0	0	0	0	0	1	-360	360;
	12830	13646	0.00201	0.02375	0	0	0	0	0	0	1	-360	360;
	13646	4230	9e-05	0.00093	0	0	0	0	0	0	1	-360	360;
	13646	2724	9e-05	0.00095	0	0	0	0	0	0	1	-360	360;
	8812	13522	0.004969	0.034979	0	0	0	0	0	0	1	-360	360;
	8812	13522	0.00468	0.033031	0	0	0	0	0	0	1	-360	360;
	13522	9419	0.00076	0.007909	0	0	0	0	0	0	1	-360	360;
	13522	7538	0.020539	0.080461	0	0	0	0	0	0	1	-360	360;
	2673	3989	0.024562	0.099231	0	0	0	0	0	0	1	-360	360;
	2673	13014	0.103397	0.231174	0	0	0	0	0	0	1	-360	360;
	2673	95	0.063843	0.141661	0	0	0	0	0	0	1	-360	360;
	12258	1054	0.00059	0.00906	0	0	0	0	0	0	1	-360	360;
	8812	9419	0.001979	0.02175	0	0	0	0	0	0	1	-360	360;
	8812	5205	0.005281	0.040039	0	0	0	0	0	0	1	-360	360;
	8812	817	0.012281	0.069219	0	0	0	0	0	0	1	-360	360;
	8812	10184	0.00668	0.04144	0	0	0	0	0	0	1	-360	360;
	9419	9002	0.00725	0.03801	0	0	0	0	0	0	1	-360	360;
	5852	2900	0.00225	0.02188	0	0	0	0	0	0	1	-360	360;
	8749	2745	0.00525	0.03749	0	0	0	0	0	0	1	-360	360;
	8749	2745	0.00605	0.03069	0	0	0	0	0	0	1	-360	360;
	5414	4754	0.054793	0.13405	0	0	0	0	0	0	1	-360	360;
	2709	5414	0.019099	0.076488	0	0	0	0	0	0	1	-360	360;
	5414	10029	0.035686	0.057678	0	0	0	0	0	0	1	-360	360;
	9868	10307	0.00481	0.020579	0	0	0	0	0	0	1	-360	360;
	9868	10307	0.004909	0.020149	0	0	0	0	0	0	1	-360	360;
	9868	6309	0.010843	0.055512	0	0	0	0	0	0	1	-360	360;
	9868	6309	0.012289	0.054397	0	0	0	0	0	0	1	-360	360;
	4833	9868	0.003562	0.018124	0	0	0	0	0	0	1	-360	360;
	12903	6863	0.038017	0.081934	0	0	0	0	0	0	1	-360	360;
	12903	1580	0.020149	0.083124	0	0	0	0	0	0	1	-360	360;
	12903	9837	0.009554	0.040959	0	0	0	0	0	0	1	-360	360;
	4023	4788	0.008331	0.043322	0	0	0	0	0	0	1	-360	360;
	4788	2823	0.02114	0.114479	0	0	0	0	0	0	1	-360	360;
	1565	4754	0.017488	0.03838	0	0	0	0	0	0	1	-360	360;
	10307	6214	0.00381	0.019256	0	0	0	0	0	0	1	-360	360;
	5917	6214	0.001397	0.005802	0	0	0	0	0	0	1	-360	360;
	3254	9130	0.004719	0.024603	0	0	0	0	0	0	1	-360	360;
	3254	9130	0.00462	0.025058	0	0	0	0	0	0	1	-360	360;
	4587	9130	0.005413	0.027876	0	0	0	0	0	0	1	-360	360;
	9130	10215	0.032041	0.196413	0	0	0	0	0	0	1	-360	360;
	3950	9130	0.002198	0.022248	0	0	0	0	0	0	1	-360	360;
	9130	269	0.00262	0.018463	0	0	0	0	0	0	1	-360	360;
	9130	6894	0.030347	0.112149	0	0	0	0	0	0	1	-360	360;
	5620	1005	0.014512	0.05981	0	0	0	0	0	0	1	-360	360;
	1792	12179	0.005438	0.009124	0	0	0	0	0	0	1	-360	360;
	12640	1862	0.005169	0.022219	0	0	0	0	0	0	1	-360	360;
	1037	9566	0.007793	0.029769	0	0	0	0	0	0	1	-360	360;
	1037	11748	0.006091	0.028653	0	0	0	0	0	0	1	-360	360;
	6143	3707	0.02876	0.060388	0	0	0	0	0	0	1	-360	360;
	2346	2730	0.02443	0.09576	0	0	0	0	0	0	1	-360	360;
	8682	2730	0.009083	0.03857	0	0	0	0	0	0	1	-360	360;
	13014	6936	0.048157	0.08462	0	0	0	0	0	0	1	-360	360;
	7239	3254	0.002512	0.013826	0	0	0	0	0	0	1	-360	360;
	12179	9205	0.012298	0.026826	0	0	0	0	0	0	1	-360	360;
	12389	12693	0.022744	0.133521	0	0	0	0	0	0	1	-360	360;
	7760	2267	0.022554	0.088876	0	0	0	0	0	0	1	-360	360;
	9052	12891	0.019231	0.033802	0	0	0	0	0	0	1	-360	360;
	10725	9594	0.010017	0.022711	0	0	0	0	0	0	1	-360	360;
	10725	743	0.00219	0.011843	0	0	0	0	0	0	1	-360	360;
	8331	8300	0.019281	0.064488	0	0	0	0	0	0	1	-360	360;
	3740	8300	0.02119	0.091727	0	0	0	0	0	0	1	-360	360;
	6300	13021	0.003099	0.013893	0	0	0	0	0	0	1	-360	360;
	1565	13021	0.004132	0.021959	0	0	0	0	0	0	1	-360	360;
	1565	13021	0.004041	0.02238	0	0	0	0	0	0	1	-360	360;
	1565	12991	0.002471	0.010901	0	0	0	0	0	0	1	-360	360;
	6075	12830	0.002	0.017	0	0	0	0	0	0	1	-360	360;
	1708	12830	0.0016	0.01758	0	0	0	0	0	0	1	-360	360;
	12389	493	0.027397	0.111917	0	0	0	0	0	0	1	-360	360;
	12389	13113	0.008752	0.044802	0	0	0	0	0	0	1	-360	360;
	1037	9205	0.045727	0.07386	0	0	0	0	0	0	1	-360	360;
	10215	6526	0.044281	0.119579	0	0	0	0	0	0	1	-360	360;
	1918	6526	0.057612	0.11224	0	0	0	0	0	0	1	-360	360;
	6374	6526	0.063314	0.107273	0	0	0	0	0	0	1	-360	360;
	5917	4991	0.006769	0.017909	0	0	0	0	0	0	1	-360	360;
	10228	6143	0.003273	0.01157	0	0	0	0	0	0	1	-360	360;
	10228	796	0.000248	0.001331	0	0	0	0	0	0	1	-360	360;
	3918	11974	0.010603	0.045727	0	0	0	0	0	0	1	-360	360;
	11974	6374	0.016273	0.089678	0	0	0	0	0	0	1	-360	360;
	5717	10972	0.008397	0.022777	0	0	0	0	0	0	1	-360	360;
	10972	2009	0.013421	0.036314	0	0	0	0	0	0	1	-360	360;
	13276	5893	0.009628	0.050752	0	0	0	0	0	0	1	-360	360;
	8739	493	0.011777	0.046843	0	0	0	0	0	0	1	-360	360;
	5506	493	0.014446	0.059967	0	0	0	0	0	0	1	-360	360;
	10215	3221	0.005174	0.029562	0	0	0	0	0	0	1	-360	360;
	2339	3835	0.069256	0.110017	0	0	0	0	0	0	1	-360	360;
	2133	4979	0.00524	0.022124	0	0	0	0	0	0	1	-360	360;
	4221	13241	0.005083	0.019421	0	0	0	0	0	0	1	-360	360;
	13241	1874	0.004099	0.013868	0	0	0	0	0	0	1	-360	360;
	13156	10666	0.04181	0.081397	0	0	0	0	0	0	1	-360	360;
	13156	6788	0.041446	0.067174	0	0	0	0	0	0	1	-360	360;
	46	3390	0.008841	0.0587	0	0	0	0	0	0	1	-360	360;
	2640	5203	0.011876	0.064	0	0	0	0	0	0	1	-360	360;
	8054	3151	0.037579	0.182612	0	0	0	0	0	0	1	-360	360;
	226	10695	0.025752	0.129421	0	0	0	0	0	0	1	-360	360;
	4979	5113	0.00543	0.011421	0	0	0	0	0	0	1	-360	360;
	5113	5774	0.003727	0.010066	0	0	0	0	0	0	1	-360	360;
	6309	10396	0.015868	0.053702	0	0	0	0	0	0	1	-360	360;
	3060	13156	0.026388	0.120603	0	0	0	0	0	0	1	-360	360;
	3079	3069	0.002372	0.010669	0	0	0	0	0	0	1	-360	360;
	4221	1874	0.001289	0.005041	0	0	0	0	0	0	1	-360	360;
	4221	5379	0.001438	0.006421	0	0	0	0	0	0	1	-360	360;
	4221	1678	0.001471	0.006289	0	0	0	0	0	0	1	-360	360;
	5852	1708	0.00105	0.01056	0	0	0	0	0	0	1	-360	360;
	4107	46	0.01137	0.056789	0	0	0	0	0	0	1	-360	360;
	969	46	0.00519	0.028151	0	0	0	0	0	0	1	-360	360;
	2346	8682	0.030091	0.150033	0	0	0	0	0	0	1	-360	360;
	8790	138	0.011207	0.023851	0	0	0	0	0	0	1	-360	360;
	138	12615	0.073347	0.103488	0	0	0	0	0	0	1	-360	360;
	7941	1740	0.064149	0.131041	0	0	0	0	0	0	1	-360	360;
	2228	1862	0.002341	0.009719	0	0	0	0	0	0	1	-360	360;
	8102	421	0.051231	0.126554	0	0	0	0	0	0	1	-360	360;
	9309	421	0.031967	0.079413	0	0	0	0	0	0	1	-360	360;
	7153	796	0.00024	0.001	0	0	0	0	0	0	1	-360	360;
	8093	4413	0.034008	0.069388	0	0	0	0	0	0	1	-360	360;
	762	1466	0.003	0.01243	0	0	0	0	0	0	1	-360	360;
	1466	10958	0.036281	0.073678	0	0	0	0	0	0	1	-360	360;
	3069	4712	0.003587	0.017231	0	0	0	0	0	0	1	-360	360;
	2598	10184	0.00012	0.00062	0	0	0	0	0	0	1	-360	360;
	10920	9837	0.006909	0.039231	0	0	0	0	0	0	1	-360	360;
	2888	9516	0.003314	0.007017	0	0	0	0	0	0	1	-360	360;
	13185	13276	0.011289	0.063661	0	0	0	0	0	0	1	-360	360;
	13185	13276	0.01105	0.058942	0	0	0	0	0	0	1	-360	360;
	11101	13276	0.007711	0.039413	0	0	0	0	0	0	1	-360	360;
	827	355	0.004017	0.013669	0	0	0	0	0	0	1	-360	360;
	6848	11057	0.040521	0.08386	0	0	0	0	0	0	1	-360	360;
	3676	1333	0.024579	0.061231	0	0	0	0	0	0	1	-360	360;
	4413	5943	0.067992	0.151463	0	0	0	0	0	0	1	-360	360;
	7337	12545	0.003579	0.011942	0	0	0	0	0	0	1	-360	360;
	2888	620	0.00776	0.021587	0	0	0	0	0	0	1	-360	360;
	13185	10692	0.003298	0.009463	0	0	0	0	0	0	1	-360	360;
	3882	10692	0.005157	0.01376	0	0	0	0	0	0	1	-360	360;
	8823	3882	0.009934	0.034479	0	0	0	0	0	0	1	-360	360;
	5379	9375	0.002008	0.010802	0	0	0	0	0	0	1	-360	360;
	1678	4712	0.001967	0.011008	0	0	0	0	0	0	1	-360	360;
	950	3596	0.003219	0.01439	0	0	0	0	0	0	1	-360	360;
	950	2747	0.00239	0.01426	0	0	0	0	0	0	1	-360	360;
	12838	9543	0.00143	0.013591	0	0	0	0	0	0	1	-360	360;
	12838	10574	0.004669	0.05795	0	0	0	0	0	0	1	-360	360;
	12838	2219	0.00412	0.043151	0	0	0	0	0	0	1	-360	360;
	12838	4516	0.000521	0.00612	0	0	0	0	0	0	1	-360	360;
	7928	1603	0.005849	0.02768	0	0	0	0	0	0	1	-360	360;
	1140	7928	0.012742	0.049311	0	0	0	0	0	0	1	-360	360;
	2181	10574	0.00206	0.01061	0	0	0	0	0	0	1	-360	360;
	2181	2857	0.001591	0.01	0	0	0	0	0	0	1	-360	360;
	1806	2225	0.00098	0.00874	0	0	0	0	0	0	1	-360	360;
	1806	8041	0.00045	0.00548	0	0	0	0	0	0	1	-360	360;
	6975	8723	0.000719	0.003349	0	0	0	0	0	0	1	-360	360;
	6975	3222	0.000919	0.00687	0	0	0	0	0	0	1	-360	360;
	6975	9267	0.000219	0.0015	0	0	0	0	0	0	1	-360	360;
	4783	3764	0.00355	0.019669	0	0	0	0	0	0	1	-360	360;
	4783	1094	0.00389	0.02201	0	0	0	0	0	0	1	-360	360;
	4783	9469	0.00251	0.01455	0	0	0	0	0	0	1	-360	360;
	6484	6678	0.00825	0.036781	0	0	0	0	0	0	1	-360	360;
	6484	6129	0.000581	0.00412	0	0	0	0	0	0	1	-360	360;
	9895	13180	0.00103	0.01236	0	0	0	0	0	0	1	-360	360;
	9895	8041	0.0014	0.01768	0	0	0	0	0	0	1	-360	360;
	7914	527	0.00088	0.01169	0	0	0	0	0	0	1	-360	360;
	5808	210	0.001711	0.01164	0	0	0	0	0	0	1	-360	360;
	5808	6533	0.00268	0.015031	0	0	0	0	0	0	1	-360	360;
	5808	11908	0.0018	0.01132	0	0	0	0	0	0	1	-360	360;
	1613	6185	0.00051	0.00268	0	0	0	0	0	0	1	-360	360;
	1613	3591	0.000651	0.003289	0	0	0	0	0	0	1	-360	360;
	2956	6173	0.0012	0.01137	0	0	0	0	0	0	1	-360	360;
	2956	4399	0.00168	0.02275	0	0	0	0	0	0	1	-360	360;
	6660	8723	0.002331	0.0225	0	0	0	0	0	0	1	-360	360;
	6660	8723	0.002081	0.0255	0	0	0	0	0	0	1	-360	360;
	6660	1001	0.000781	0.004979	0	0	0	0	0	0	1	-360	360;
	6660	3110	0.00514	0.06699	0	0	0	0	0	0	1	-360	360;
	10942	5115	0.00088	0.00804	0	0	0	0	0	0	1	-360	360;
	10942	4782	0.00032	0.00311	0	0	0	0	0	0	1	-360	360;
	10942	322	0.00015	0.00197	0	0	0	0	0	0	1	-360	360;
	11260	3596	0.004719	0.023831	0	0	0	0	0	0	1	-360	360;
	11260	3222	0.00669	0.029979	0	0	0	0	0	0	1	-360	360;
	13548	2098	0.001031	0.01225	0	0	0	0	0	0	1	-360	360;
	13548	7897	0.00105	0.00937	0	0	0	0	0	0	1	-360	360;
	13548	1327	0.00201	0.01293	0	0	0	0	0	0	1	-360	360;
	2873	1094	0.002	0.012229	0	0	0	0	0	0	1	-360	360;
	2873	2736	8.1e-05	0.000521	0	0	0	0	0	0	1	-360	360;
	2873	9543	0.00113	0.010969	0	0	0	0	0	0	1	-360	360;
	13180	12939	0.00195	0.02538	0	0	0	0	0	0	1	-360	360;
	13180	4782	0.00164	0.01684	0	0	0	0	0	0	1	-360	360;
	10140	9424	0.006289	0.06895	0	0	0	0	0	0	1	-360	360;
	10140	13286	0.001169	0.010229	0	0	0	0	0	0	1	-360	360;
	10140	8505	0.00805	0.060409	0	0	0	0	0	0	1	-360	360;
	10140	8505	0.00789	0.061581	0	0	0	0	0	0	1	-360	360;
	10140	10574	0.007159	0.04164	0	0	0	0	0	0	1	-360	360;
	10140	10574	0.007289	0.040729	0	0	0	0	0	0	1	-360	360;
	10140	4525	0.004531	0.02731	0	0	0	0	0	0	1	-360	360;
	10140	1527	0.004521	0.033581	0	0	0	0	0	0	1	-360	360;
	2098	4975	0.0005	0.005081	0	0	0	0	0	0	1	-360	360;
	2098	1160	0.00132	0.006521	0	0	0	0	0	0	1	-360	360;
	7748	9424	0.004401	0.030531	0	0	0	0	0	0	1	-360	360;
	7748	2219	0.000469	0.003409	0	0	0	0	0	0	1	-360	360;
	4382	5115	0.00239	0.02364	0	0	0	0	0	0	1	-360	360;
	4382	10861	0.00034	0.0041	0	0	0	0	0	0	1	-360	360;
	4382	4399	0.00027	0.00327	0	0	0	0	0	0	1	-360	360;
	4382	4399	0.00026	0.00333	0	0	0	0	0	0	1	-360	360;
	4382	1194	0.00059	0.00574	0	0	0	0	0	0	1	-360	360;
	8506	11417	0.000789	0.003711	0	0	0	0	0	0	1	-360	360;
	8506	11016	0.00419	0.04988	0	0	0	0	0	0	1	-360	360;
	8506	7646	0.001651	0.010919	0	0	0	0	0	0	1	-360	360;
	8482	4525	0.007281	0.03162	0	0	0	0	0	0	1	-360	360;
	8482	3267	0.006831	0.03505	0	0	0	0	0	0	1	-360	360;
	5115	8041	0.00108	0.01258	0	0	0	0	0	0	1	-360	360;
	5115	10861	0.00174	0.0229	0	0	0	0	0	0	1	-360	360;
	5115	1194	0.00223	0.02125	0	0	0	0	0	0	1	-360	360;
	3596	8723	0.0023	0.014081	0	0	0	0	0	0	1	-360	360;
	3596	2747	0.00213	0.026831	0	0	0	0	0	0	1	-360	360;
	3596	8422	0.001039	0.00549	0	0	0	0	0	0	1	-360	360;
	3596	3243	0.00301	0.016711	0	0	0	0	0	0	1	-360	360;
	3596	1992	0.000849	0.005229	0	0	0	0	0	0	1	-360	360;
	3596	2290	0.002781	0.020281	0	0	0	0	0	0	1	-360	360;
	3596	1327	0.002159	0.01389	0	0	0	0	0	0	1	-360	360;
	217	11971	0.003502	0.031631	0	0	0	0	0	0	1	-360	360;
	1603	10234	0.003929	0.024729	0	0	0	0	0	0	1	-360	360;
	10234	217	0.00076	0.005609	0	0	0	0	0	0	1	-360	360;
	3764	6185	0.002039	0.01249	0	0	0	0	0	0	1	-360	360;
	7231	12939	0.00036	0.00572	0	0	0	0	0	0	1	-360	360;
	7231	1638	0.00028	0.00448	0	0	0	0	0	0	1	-360	360;
	4303	9424	0.001091	0.01663	0	0	0	0	0	0	1	-360	360;
	4303	4525	0.011909	0.080909	0	0	0	0	0	0	1	-360	360;
	4303	210	0.000401	0.00232	0	0	0	0	0	0	1	-360	360;
	4303	210	0.000349	0.00263	0	0	0	0	0	0	1	-360	360;
	2116	8695	0.000531	0.0063	0	0	0	0	0	0	1	-360	360;
	2116	683	0.005531	0.02938	0	0	0	0	0	0	1	-360	360;
	2116	11596	0.00019	0.002521	0	0	0	0	0	0	1	-360	360;
	8695	9424	0.00075	0.010909	0	0	0	0	0	0	1	-360	360;
	12322	6129	0.007	0.03088	0	0	0	0	0	0	1	-360	360;
	12322	6533	0.002349	0.009281	0	0	0	0	0	0	1	-360	360;
	9424	7621	0.003219	0.02125	0	0	0	0	0	0	1	-360	360;
	9424	7621	0.004461	0.027961	0	0	0	0	0	0	1	-360	360;
	1378	6283	0.00105	0.007901	0	0	0	0	0	0	1	-360	360;
	9824	6678	0.002461	0.01019	0	0	0	0	0	0	1	-360	360;
	9824	6129	0.00605	0.028409	0	0	0	0	0	0	1	-360	360;
	2966	1327	7e-05	0.00039	0	0	0	0	0	0	1	-360	360;
	8723	2747	0.0047	0.02175	0	0	0	0	0	0	1	-360	360;
	10923	6375	0.00118	0.00719	0	0	0	0	0	0	1	-360	360;
	10923	341	0.002349	0.01887	0	0	0	0	0	0	1	-360	360;
	2971	2263	0.00104	0.01314	0	0	0	0	0	0	1	-360	360;
	2971	4399	0.0004	0.004	0	0	0	0	0	0	1	-360	360;
	13286	9776	0.001289	0.006229	0	0	0	0	0	0	1	-360	360;
	13286	5238	0.00555	0.035419	0	0	0	0	0	0	1	-360	360;
	13286	5238	0.00381	0.0328	0	0	0	0	0	0	1	-360	360;
	6678	8601	0.00093	0.006841	0	0	0	0	0	0	1	-360	360;
	6678	8601	0.000909	0.006979	0	0	0	0	0	0	1	-360	360;
	4441	8108	0.001091	0.0082	0	0	0	0	0	0	1	-360	360;
	4441	5688	0.002081	0.020419	0	0	0	0	0	0	1	-360	360;
	6283	5006	0.00082	0.00276	0	0	0	0	0	0	1	-360	360;
	6283	4525	0.00162	0.007021	0	0	0	0	0	0	1	-360	360;
	11563	8505	0.002831	0.015021	0	0	0	0	0	0	1	-360	360;
	11563	192	0.006781	0.03026	0	0	0	0	0	0	1	-360	360;
	2225	4782	0.00041	0.00523	0	0	0	0	0	0	1	-360	360;
	2225	6173	0.00062	0.0098	0	0	0	0	0	0	1	-360	360;
	2225	6173	0.00083	0.00975	0	0	0	0	0	0	1	-360	360;
	5238	7897	0.00131	0.01564	0	0	0	0	0	0	1	-360	360;
	5238	10574	0.002039	0.026831	0	0	0	0	0	0	1	-360	360;
	5238	5688	0.001031	0.00945	0	0	0	0	0	0	1	-360	360;
	5238	5501	0.001789	0.015211	0	0	0	0	0	0	1	-360	360;
	2747	10574	0.003021	0.03739	0	0	0	0	0	0	1	-360	360;
	2747	566	0.004211	0.04031	0	0	0	0	0	0	1	-360	360;
	2747	5209	0.004979	0.029409	0	0	0	0	0	0	1	-360	360;
	2747	5209	0.00488	0.029979	0	0	0	0	0	0	1	-360	360;
	2747	3427	0.002211	0.02399	0	0	0	0	0	0	1	-360	360;
	2747	3427	0.00231	0.023021	0	0	0	0	0	0	1	-360	360;
	683	4525	0.00719	0.046841	0	0	0	0	0	0	1	-360	360;
	1736	8505	0.00155	0.00968	0	0	0	0	0	0	1	-360	360;
	1736	8505	0.0015	0.01051	0	0	0	0	0	0	1	-360	360;
	1736	4375	0.00169	0.013401	0	0	0	0	0	0	1	-360	360;
	1736	3670	0.001659	0.012409	0	0	0	0	0	0	1	-360	360;
	3884	4525	0.005091	0.0167	0	0	0	0	0	0	1	-360	360;
	3884	3267	0.009919	0.04649	0	0	0	0	0	0	1	-360	360;
	2917	7646	9.1e-05	0.000531	0	0	0	0	0	0	1	-360	360;
	8505	10574	0.00982	0.045331	0	0	0	0	0	0	1	-360	360;
	8505	13000	0.003039	0.02964	0	0	0	0	0	0	1	-360	360;
	2736	5841	0.004159	0.02051	0	0	0	0	0	0	1	-360	360;
	5006	4525	0.00087	0.00469	0	0	0	0	0	0	1	-360	360;
	8108	852	0.0017	0.021039	0	0	0	0	0	0	1	-360	360;
	4782	322	0.00019	0.0032	0	0	0	0	0	0	1	-360	360;
	3568	4210	0.006979	0.03975	0	0	0	0	0	0	1	-360	360;
	3568	852	0.000521	0.004081	0	0	0	0	0	0	1	-360	360;
	3568	1527	0.00382	0.02214	0	0	0	0	0	0	1	-360	360;
	1939	6173	0.00204	0.02487	0	0	0	0	0	0	1	-360	360;
	9543	6185	0.00312	0.018469	0	0	0	0	0	0	1	-360	360;
	4210	4525	0.005479	0.038841	0	0	0	0	0	0	1	-360	360;
	192	10574	0.000789	0.007729	0	0	0	0	0	0	1	-360	360;
	8399	9469	0.001789	0.01618	0	0	0	0	0	0	1	-360	360;
	8399	9469	0.00175	0.016479	0	0	0	0	0	0	1	-360	360;
	9469	5841	0.00375	0.025099	0	0	0	0	0	0	1	-360	360;
	6503	9418	0.010778	0.035609	0	0	0	0	0	0	1	-360	360;
	9418	6503	0.0096	0.04036	0	0	0	0	0	0	1	-360	360;
	11661	6129	0.001031	0.016919	0	0	0	0	0	0	1	-360	360;
	11661	6129	0.00101	0.01725	0	0	0	0	0	0	1	-360	360;
	10574	7152	0.00124	0.013169	0	0	0	0	0	0	1	-360	360;
	10574	7152	0.00119	0.01356	0	0	0	0	0	0	1	-360	360;
	10574	5501	0.00119	0.013281	0	0	0	0	0	0	1	-360	360;
	6375	10934	0.00306	0.02405	0	0	0	0	0	0	1	-360	360;
	6375	7621	0.002469	0.018081	0	0	0	0	0	0	1	-360	360;
	11335	2030	0.0018	0.014651	0	0	0	0	0	0	1	-360	360;
	11335	3427	0.001081	0.010219	0	0	0	0	0	0	1	-360	360;
	2030	11016	0.000669	0.003521	0	0	0	0	0	0	1	-360	360;
	4525	527	0.001039	0.0142	0	0	0	0	0	0	1	-360	360;
	4525	9022	0.002771	0.019919	0	0	0	0	0	0	1	-360	360;
	4975	7152	0.00295	0.013401	0	0	0	0	0	0	1	-360	360;
	7152	3222	0.00411	0.02488	0	0	0	0	0	0	1	-360	360;
	7152	7036	0.00061	0.007211	0	0	0	0	0	0	1	-360	360;
	1527	1160	0.002219	0.011081	0	0	0	0	0	0	1	-360	360;
	8693	4269	6e-05	0.00066	0	0	0	0	0	0	1	-360	360;
	6185	3670	0.00411	0.026081	0	0	0	0	0	0	1	-360	360;
	4831	4399	0.00168	0.02222	0	0	0	0	0	0	1	-360	360;
	4831	4269	0.00049	0.0061	0	0	0	0	0	0	1	-360	360;
	3222	9267	0.00082	0.00575	0	0	0	0	0	0	1	-360	360;
	11596	3267	0.001409	0.01918	0	0	0	0	0	0	1	-360	360;
	8422	3243	0.001841	0.011159	0	0	0	0	0	0	1	-360	360;
	6000	4269	0.00109	0.01044	0	0	0	0	0	0	1	-360	360;
	7078	1638	0.00088	0.01608	0	0	0	0	0	0	1	-360	360;
	6129	10127	0.000591	0.007781	0	0	0	0	0	0	1	-360	360;
	6129	4468	0.001901	0.0247	0	0	0	0	0	0	1	-360	360;
	11417	9956	0.00382	0.018401	0	0	0	0	0	0	1	-360	360;
	11016	9956	0.004169	0.02775	0	0	0	0	0	0	1	-360	360;
	11016	3427	0.001849	0.01889	0	0	0	0	0	0	1	-360	360;
	3267	4468	0.00076	0.01036	0	0	0	0	0	0	1	-360	360;
	2857	1001	0.00207	0.009849	0	0	0	0	0	0	1	-360	360;
	966	566	0.001849	0.013841	0	0	0	0	0	0	1	-360	360;
	966	566	0.00243	0.015771	0	0	0	0	0	0	1	-360	360;
	4399	2362	0.00128	0.01149	0	0	0	0	0	0	1	-360	360;
	3110	1042	0.00057	0.005961	0	0	0	0	0	0	1	-360	360;
	3110	1788	0.00218	0.015099	0	0	0	0	0	0	1	-360	360;
	566	7646	0.00131	0.008771	0	0	0	0	0	0	1	-360	360;
	1042	1788	0.00163	0.008729	0	0	0	0	0	0	1	-360	360;
	1992	2073	0.004831	0.02588	0	0	0	0	0	0	1	-360	360;
	341	4516	0.00062	0.00857	0	0	0	0	0	0	1	-360	360;
	10127	8601	0.002281	0.0362	0	0	0	0	0	0	1	-360	360;
	9022	13000	0.00132	0.01036	0	0	0	0	0	0	1	-360	360;
	2073	3427	0.00281	0.017099	0	0	0	0	0	0	1	-360	360;
	3427	2290	0.00424	0.0258	0	0	0	0	0	0	1	-360	360;
	1032	11688	0.00036	0.00313	0	0	0	0	0	0	1	-360	360;
	8878	13384	0.00112	0.0114	0	0	0	0	0	0	1	-360	360;
	8878	13619	0.00046	0.007	0	0	0	0	0	0	1	-360	360;
	8878	553	0.00113	0.01516	0	0	0	0	0	0	1	-360	360;
	13384	10332	0.00138	0.01478	0	0	0	0	0	0	1	-360	360;
	13384	759	0.00107	0.01447	0	0	0	0	0	0	1	-360	360;
	8895	12621	0.012651	0.08939	0	0	0	0	0	0	1	-360	360;
	8895	8762	0.00957	0.05424	0	0	0	0	0	0	1	-360	360;
	1854	5027	0.0017	0.01823	0	0	0	0	0	0	1	-360	360;
	1854	8750	0.00092	0.00892	0	0	0	0	0	0	1	-360	360;
	1854	10774	0.00063	0.00829	0	0	0	0	0	0	1	-360	360;
	10332	11703	0.00052	0.00757	0	0	0	0	0	0	1	-360	360;
	11703	4797	0.00189	0.02546	0	0	0	0	0	0	1	-360	360;
	11703	6154	0.00044	0.00602	0	0	0	0	0	0	1	-360	360;
	204	8187	0.00138	0.01477	0	0	0	0	0	0	1	-360	360;
	8187	6169	0.00055	0.00597	0	0	0	0	0	0	1	-360	360;
	553	5588	0.00047	0.00684	0	0	0	0	0	0	1	-360	360;
	12621	4921	0.0038	0.026331	0	0	0	0	0	0	1	-360	360;
	10332	3538	0.001	0.00937	0	0	0	0	0	0	1	-360	360;
	10332	553	0.00203	0.01819	0	0	0	0	0	0	1	-360	360;
	8762	9318	0.002419	0.015021	0	0	0	0	0	0	1	-360	360;
	8762	9318	0.00274	0.014719	0	0	0	0	0	0	1	-360	360;
	8762	5288	0.014599	0.077461	0	0	0	0	0	0	1	-360	360;
	8762	10391	0.012461	0.05613	0	0	0	0	0	0	1	-360	360;
	8762	4921	0.014581	0.083849	0	0	0	0	0	0	1	-360	360;
	10398	4797	0.0006	0.00665	0	0	0	0	0	0	1	-360	360;
	10398	4797	0.00057	0.00856	0	0	0	0	0	0	1	-360	360;
	10398	8750	0.00234	0.02648	0	0	0	0	0	0	1	-360	360;
	3538	5767	0.00019	0.00167	0	0	0	0	0	0	1	-360	360;
	3538	2967	0.00019	0.00164	0	0	0	0	0	0	1	-360	360;
	8750	12186	3e-05	0.00047	0	0	0	0	0	0	1	-360	360;
	8750	1197	3e-05	0.00048	0	0	0	0	0	0	1	-360	360;
	9122	5945	0.00097	0.01084	0	0	0	0	0	0	1	-360	360;
	8750	8520	8e-05	0.00073	0	0	0	0	0	0	1	-360	360;
	8750	7273	8e-05	0.0006	0	0	0	0	0	0	1	-360	360;
	8750	2221	7e-05	0.00067	0	0	0	0	0	0	1	-360	360;
	8750	1035	7e-05	0.00066	0	0	0	0	0	0	1	-360	360;
	5288	821	0.00145	0.00707	0	0	0	0	0	0	1	-360	360;
	5288	4951	0.00164	0.00693	0	0	0	0	0	0	1	-360	360;
	553	10569	8e-05	0.00084	0	0	0	0	0	0	1	-360	360;
	553	7880	0.0001	0.00083	0	0	0	0	0	0	1	-360	360;
	553	11998	0.0001	0.00081	0	0	0	0	0	0	1	-360	360;
	4802	1729	0.00019	0.00215	0	0	0	0	0	0	1	-360	360;
	204	5749	0.00037	0.00292	0	0	0	0	0	0	1	-360	360;
	204	2952	0.00033	0.00332	0	0	0	0	0	0	1	-360	360;
	5211	5997	0.002031	0.01382	0	0	0	0	0	0	1	-360	360;
	5211	3723	0.001979	0.014091	0	0	0	0	0	0	1	-360	360;
	4797	4280	4e-05	0.00052	0	0	0	0	0	0	1	-360	360;
	4797	1634	4e-05	0.00054	0	0	0	0	0	0	1	-360	360;
	5211	2938	0.00136	0.0067	0	0	0	0	0	0	1	-360	360;
	4802	6463	0.0001	0.00078	0	0	0	0	0	0	1	-360	360;
	4802	2293	0.00011	0.00076	0	0	0	0	0	0	1	-360	360;
	5027	4802	0.00376	0.04872	0	0	0	0	0	0	1	-360	360;
	4802	8574	0.0002	0.00258	0	0	0	0	0	0	1	-360	360;
	4802	553	0.00089	0.0087	0	0	0	0	0	0	1	-360	360;
	436	5211	0.01282	0.053159	0	0	0	0	0	0	1	-360	360;
	436	10221	0.002219	0.01038	0	0	0	0	0	0	1	-360	360;
	436	5211	0.01307	0.05205	0	0	0	0	0	0	1	-360	360;
	436	4921	0.00449	0.01863	0	0	0	0	0	0	1	-360	360;
	204	6154	0.00215	0.0248	0	0	0	0	0	0	1	-360	360;
	3374	9122	0.00192	0.02113	0	0	0	0	0	0	1	-360	360;
	3374	11688	0.00165	0.01704	0	0	0	0	0	0	1	-360	360;
	4797	6169	0.00168	0.02469	0	0	0	0	0	0	1	-360	360;
	4797	6154	0.00222	0.02324	0	0	0	0	0	0	1	-360	360;
	9122	759	0.00173	0.01682	0	0	0	0	0	0	1	-360	360;
	9122	12785	0.00174	0.01872	0	0	0	0	0	0	1	-360	360;
	9122	3538	0.00194	0.02194	0	0	0	0	0	0	1	-360	360;
	7738	4872	0.013831	0.056841	0	0	0	0	0	0	1	-360	360;
	7738	4872	0.014091	0.055599	0	0	0	0	0	0	1	-360	360;
	5288	291	0.009081	0.05263	0	0	0	0	0	0	1	-360	360;
	5288	7312	0.007401	0.05661	0	0	0	0	0	0	1	-360	360;
	11688	12785	0.00164	0.01627	0	0	0	0	0	0	1	-360	360;
	10391	3721	0.022909	0.124591	0	0	0	0	0	0	1	-360	360;
	4306	12785	0.00074	0.00791	0	0	0	0	0	0	1	-360	360;
	4306	10774	0.00136	0.01726	0	0	0	0	0	0	1	-360	360;
	291	5211	0.011211	0.07026	0	0	0	0	0	0	1	-360	360;
	291	5211	0.01143	0.06882	0	0	0	0	0	0	1	-360	360;
	12785	6154	0.00544	0.06455	0	0	0	0	0	0	1	-360	360;
	4872	3721	0.016289	0.066719	0	0	0	0	0	0	1	-360	360;
	4872	3721	0.016591	0.065271	0	0	0	0	0	0	1	-360	360;
	8750	10774	0.00077	0.0096	0	0	0	0	0	0	1	-360	360;
	8750	10774	0.00093	0.00905	0	0	0	0	0	0	1	-360	360;
	3721	7312	0.021729	0.15268	0	0	0	0	0	0	1	-360	360;
	718	10221	0.00064	0.002919	0	0	0	0	0	0	1	-360	360;
	10221	4921	0.002211	0.009169	0	0	0	0	0	0	1	-360	360;
	5531	13443	0.00014	0.001729	0	0	0	0	0	0	1	-360	360;
	439	13443	0.0002	0.00281	0	0	0	0	0	0	1	-360	360;
	3806	7655	0.000711	0.004289	0	0	0	0	0	0	1	-360	360;
	3806	10208	0.0007	0.00438	0	0	0	0	0	0	1	-360	360;
	5531	3806	0.00032	0.00437	0	0	0	0	0	0	1	-360	360;
	7655	8407	7e-05	0.000401	0	0	0	0	0	0	1	-360	360;
	10208	12057	7e-05	0.000401	0	0	0	0	0	0	1	-360	360;
	1167	8407	0.00189	0.0083	0	0	0	0	0	0	1	-360	360;
	1167	12057	0.00168	0.009401	0	0	0	0	0	0	1	-360	360;
	5531	10545	0.000151	0.00169	0	0	0	0	0	0	1	-360	360;
	9099	11009	0	0.0002	0	0	0	0	0	0	1	-360	360;
	9099	11009	0.001	0.120531	0	0	0	0	0	0	1	-360	360;
	13629	9099	4.583678	10.16446	0	0	0	0	0	0	1	-360	360;
	9099	8267	8.430371	11.9126	0	0	0	0	0	0	1	-360	360;
	9099	4147	8.911157	11.64008	0	0	0	0	0	0	1	-360	360;
	11448	9099	7.455785	17.24773	0	0	0	0	0	0	1	-360	360;
	2359	9099	9.746074	17.22748	0	0	0	0	0	0	1	-360	360;
	8608	9099	0.796777	2.654959	0	0	0	0	0	0	1	-360	360;
	4055	9099	0.960248	2.559091	0	0	0	0	0	0	1	-360	360;
	8156	9099	5.215909	9.009917	0	0	0	0	0	0	1	-360	360;
	6030	9099	5.097107	9.164256	0	0	0	0	0	0	1	-360	360;
	917	9099	19.22748	20.93388	0	0	0	0	0	0	1	-360	360;
	4995	9099	17.18905	24.75	0	0	0	0	0	0	1	-360	360;
	9099	12407	17.2064	22.05785	0	0	0	0	0	0	1	-360	360;
	12043	9099	0.227459	1.377252	0	0	0	0	0	0	1	-360	360;
	9823	9099	0.13382	0.643926	0	0	0	0	0	0	1	-360	360;
	6896	9099	0.211219	1.206674	0	0	0	0	0	0	1	-360	360;
	10979	9099	0.230165	1.120413	0	0	0	0	0	0	1	-360	360;
	1535	9099	29.19835	48.54959	0	0	0	0	0	0	1	-360	360;
	5183	9099	0.13231	0.939504	0	0	0	0	0	0	1	-360	360;
	7916	9099	0.842851	2.338843	0	0	0	0	0	0	1	-360	360;
	3743	9099	0.820764	2.471488	0	0	0	0	0	0	1	-360	360;
	4717	9099	0.553244	1.973719	0	0	0	0	0	0	1	-360	360;
	4717	9099	0.01405	0.0548	0	0	0	0	0	0	1	-360	360;
	9099	2256	1.071157	2.739256	0	0	0	0	0	0	1	-360	360;
	9099	2256	0.00774	0.04075	0	0	0	0	0	0	1	-360	360;
	7691	9099	16.64566	21.19835	0	0	0	0	0	0	1	-360	360;
	6065	9099	16.22913	16.64318	0	0	0	0	0	0	1	-360	360;
	13629	11009	10.87128	19.07149	0	0	0	0	0	0	1	-360	360;
	11009	8267	15.62583	25.67355	0	0	0	0	0	0	1	-360	360;
	11009	4147	16.95331	27.19835	0	0	0	0	0	0	1	-360	360;
	11448	11009	16.6593	34.37603	0	0	0	0	0	0	1	-360	360;
	2359	11009	19.6814	37.93595	0	0	0	0	0	0	1	-360	360;
	8608	11009	1.881012	4.967975	0	0	0	0	0	0	1	-360	360;
	4055	11009	1.902851	5.765909	0	0	0	0	0	0	1	-360	360;
	8156	11009	10.14587	20.68595	0	0	0	0	0	0	1	-360	360;
	6030	11009	10.93058	18.98595	0	0	0	0	0	0	1	-360	360;
	917	11009	36.07851	50.20248	0	0	0	0	0	0	1	-360	360;
	4995	11009	40.72727	46.32438	0	0	0	0	0	0	1	-360	360;
	11009	12407	36.24587	46.68595	0	0	0	0	0	0	1	-360	360;
	12043	11009	0.444855	2.8	0	0	0	0	0	0	1	-360	360;
	9823	11009	0.273946	1.391054	0	0	0	0	0	0	1	-360	360;
	6896	11009	0.514545	2.160744	0	0	0	0	0	0	1	-360	360;
	10979	11009	0.450992	2.507851	0	0	0	0	0	0	1	-360	360;
	5183	11009	0.292665	1.871012	0	0	0	0	0	0	1	-360	360;
	7916	11009	1.795062	4.826033	0	0	0	0	0	0	1	-360	360;
	3743	11009	1.551219	5.755991	0	0	0	0	0	0	1	-360	360;
	4717	11009	1.089959	4.426653	0	0	0	0	0	0	1	-360	360;
	11009	2256	0.00757	0.037729	0	0	0	0	0	0	1	-360	360;
	11009	2256	2.158264	6.029752	0	0	0	0	0	0	1	-360	360;
	7691	11009	35.04132	44.83471	0	0	0	0	0	0	1	-360	360;
	6065	11009	30.36983	39.79132	0	0	0	0	0	0	1	-360	360;
	7867	6362	0.00058	0.00554	0	0	0	0	0	0	1	-360	360;
	7867	8653	0.00116	0.01169	0	0	0	0	0	0	1	-360	360;
	9172	6362	0.00052	0.00627	0	0	0	0	0	0	1	-360	360;
	9172	8653	0.00118	0.01144	0	0	0	0	0	0	1	-360	360;
	7867	9172	0.001	0.223012	0	0	0	0	0	0	1	-360	360;
	7867	6362	0.05072	0.617493	0	0	0	0	0	0	1	-360	360;
	7867	8653	0.05373	0.600208	0	0	0	0	0	0	1	-360	360;
	478	7867	0.381143	1.437258	0	0	0	0	0	0	1	-360	360;
	9172	6362	0.04991	0.630229	0	0	0	0	0	0	1	-360	360;
	9172	8653	0.05131	0.624993	0	0	0	0	0	0	1	-360	360;
	478	9172	0.389938	1.410526	0	0	0	0	0	0	1	-360	360;
	6587	5003	0.00059	0.00463	0	0	0	0	0	0	1	-360	360;
	6587	2275	0.00047	0.0038	0	0	0	0	0	0	1	-360	360;
	9153	11186	0.0006	0.00453	0	0	0	0	0	0	1	-360	360;
	9153	5003	0.03442	0.298892	0	0	0	0	0	0	1	-360	360;
	9153	11186	0.03377	0.305229	0	0	0	0	0	0	1	-360	360;
	9153	9673	0.00047	0.00372	0	0	0	0	0	0	1	-360	360;
	6562	880	0.00036	0.00341	0	0	0	0	0	0	1	-360	360;
	5912	6562	0.00164	0.01639	0	0	0	0	0	0	1	-360	360;
	707	9072	0	0.0002	0	0	0	0	0	0	1	-360	360;
	707	9072	0.001	0.13669	0	0	0	0	0	0	1	-360	360;
	707	11302	2.103099	5.086777	0	0	0	0	0	0	1	-360	360;
	707	9140	0.004781	0.02089	0	0	0	0	0	0	1	-360	360;
	707	7014	46.67975	63.95455	0	0	0	0	0	0	1	-360	360;
	707	12720	45.98967	55.45455	0	0	0	0	0	0	1	-360	360;
	707	9401	1.337128	3.255992	0	0	0	0	0	0	1	-360	360;
	707	1566	0.09056	0.365041	0	0	0	0	0	0	1	-360	360;
	707	5754	0.085081	0.477665	0	0	0	0	0	0	1	-360	360;
	707	5754	0.008	0.040669	0	0	0	0	0	0	1	-360	360;
	9072	11302	2.446281	4.115289	0	0	0	0	0	0	1	-360	360;
	9072	141	0.00488	0.020469	0	0	0	0	0	0	1	-360	360;
	9072	7014	44.43595	63.26653	0	0	0	0	0	0	1	-360	360;
	9072	12720	47.34298	50.52893	0	0	0	0	0	0	1	-360	360;
	9072	9401	1.382355	2.978719	0	0	0	0	0	0	1	-360	360;
	9072	1566	0.07226	0.393864	0	0	0	0	0	0	1	-360	360;
	9072	1566	0.007831	0.04144	0	0	0	0	0	0	1	-360	360;
	9072	5754	0.08989	0.428161	0	0	0	0	0	0	1	-360	360;
	5482	11721	0.00055	0.00435	0	0	0	0	0	0	1	-360	360;
	3229	11721	0.00128	0.00706	0	0	0	0	0	0	1	-360	360;
	4798	8075	0.00053	0.00407	0	0	0	0	0	0	1	-360	360;
	5993	8075	0.00056	0.00478	0	0	0	0	0	0	1	-360	360;
	6510	5482	0.208393	1.089612	0	0	0	0	0	0	1	-360	360;
	3603	5482	0.199571	1.1759	0	0	0	0	0	0	1	-360	360;
	1665	5482	0.02232	0.346918	0	0	0	0	0	0	1	-360	360;
	4490	5482	0.858172	1.866898	0	0	0	0	0	0	1	-360	360;
	2872	5482	0.00031	0.00216	0	0	0	0	0	0	1	-360	360;
	7278	5482	0.001	0.274529	0	0	0	0	0	0	1	-360	360;
	13017	5482	4.431302	8.136427	0	0	0	0	0	0	1	-360	360;
	10335	12622	0.36115	2.404917	0	0	0	0	0	0	1	-360	360;
	10335	272	0.319377	2.230125	0	0	0	0	0	0	1	-360	360;
	10335	4883	0.325519	2.183657	0	0	0	0	0	0	1	-360	360;
	7997	10335	1.726316	5.458795	0	0	0	0	0	0	1	-360	360;
	10335	6414	2.428047	7.275623	0	0	0	0	0	0	1	-360	360;
	10335	12406	0.00056	0.00716	0	0	0	0	0	0	1	-360	360;
	10335	12406	0.527043	3.451801	0	0	0	0	0	0	1	-360	360;
	10335	8314	0.515582	3.196122	0	0	0	0	0	0	1	-360	360;
	10335	9085	1.334072	5.856163	0	0	0	0	0	0	1	-360	360;
	10335	6868	0.0011	0.01318	0	0	0	0	0	0	1	-360	360;
	10335	13020	0.00108	0.01344	0	0	0	0	0	0	1	-360	360;
	10335	8314	0.00063	0.00702	0	0	0	0	0	0	1	-360	360;
	5899	1467	0.246612	1.335723	0	0	0	0	0	0	1	-360	360;
	5899	1467	0.008229	0.04962	0	0	0	0	0	0	1	-360	360;
	5899	5005	0.219628	1.572521	0	0	0	0	0	0	1	-360	360;
	5899	5923	4.089256	11.14855	0	0	0	0	0	0	1	-360	360;
	5899	1752	0.00943	0.04726	0	0	0	0	0	0	1	-360	360;
	5899	1752	3.689876	10.15083	0	0	0	0	0	0	1	-360	360;
	5899	644	0.202581	1.497417	0	0	0	0	0	0	1	-360	360;
	5899	2726	0.1973	1.490124	0	0	0	0	0	0	1	-360	360;
	5899	418	0.0028	0.013271	0	0	0	0	0	0	1	-360	360;
	5899	5344	0.12369	1.125744	0	0	0	0	0	0	1	-360	360;
	5899	2264	1.335496	4.761364	0	0	0	0	0	0	1	-360	360;
	5899	4342	1.15	4.888636	0	0	0	0	0	0	1	-360	360;
	5899	1684	1.052913	5.261983	0	0	0	0	0	0	1	-360	360;
	5899	5005	0.00932	0.04863	0	0	0	0	0	0	1	-360	360;
	5899	5923	0.00961	0.046281	0	0	0	0	0	0	1	-360	360;
	5899	8634	0.0025	0.015031	0	0	0	0	0	0	1	-360	360;
	7074	11117	0.0008	0.00519	0	0	0	0	0	0	1	-360	360;
	7074	1752	0.00612	0.03036	0	0	0	0	0	0	1	-360	360;
	11117	2637	0.00038	0.00243	0	0	0	0	0	0	1	-360	360;
	6964	549	0.000909	0.005091	0	0	0	0	0	0	1	-360	360;
	6964	644	0.00187	0.00944	0	0	0	0	0	0	1	-360	360;
	549	2637	0.00043	0.00238	0	0	0	0	0	0	1	-360	360;
	12866	6188	0.00013	0.00176	0	0	0	0	0	0	1	-360	360;
	12866	6188	0.00013	0.00179	0	0	0	0	0	0	1	-360	360;
	10280	1533	0.00018	0.00189	0	0	0	0	0	0	1	-360	360;
	10280	10514	0.0002	0.00185	0	0	0	0	0	0	1	-360	360;
	1533	2174	0.00039	0.00344	0	0	0	0	0	0	1	-360	360;
	10514	8313	0.0004	0.00337	0	0	0	0	0	0	1	-360	360;
	10280	13454	0.00027	0.00274	0	0	0	0	0	0	1	-360	360;
	10280	1771	0.00056	0.00437	0	0	0	0	0	0	1	-360	360;
	10280	9732	0.00026	0.00279	0	0	0	0	0	0	1	-360	360;
	7699	3986	0.001349	0.00763	0	0	0	0	0	0	1	-360	360;
	7699	3746	0.00374	0.02218	0	0	0	0	0	0	1	-360	360;
	7699	5547	0.000349	0.00186	0	0	0	0	0	0	1	-360	360;
	7699	3746	0.004651	0.02025	0	0	0	0	0	0	1	-360	360;
	7699	5547	0.00037	0.00182	0	0	0	0	0	0	1	-360	360;
	376	6032	0.001	0.209093	0	0	0	0	0	0	1	-360	360;
	376	6510	0.00062	0.00513	0	0	0	0	0	0	1	-360	360;
	376	3603	0.04427	0.493158	0	0	0	0	0	0	1	-360	360;
	376	2174	2.319321	5.378532	0	0	0	0	0	0	1	-360	360;
	7131	376	2.762188	4.809072	0	0	0	0	0	0	1	-360	360;
	13017	376	24.69598	23.43698	0	0	0	0	0	0	1	-360	360;
	376	6550	1.383864	4.675693	0	0	0	0	0	0	1	-360	360;
	964	376	0.470817	2.500277	0	0	0	0	0	0	1	-360	360;
	1651	376	0.00099	0.00946	0	0	0	0	0	0	1	-360	360;
	13282	376	0.05862	0.781925	0	0	0	0	0	0	1	-360	360;
	6032	3603	0.04335	0.502652	0	0	0	0	0	0	1	-360	360;
	6032	3603	0.00063	0.00502	0	0	0	0	0	0	1	-360	360;
	6032	2174	2.364751	5.268767	0	0	0	0	0	0	1	-360	360;
	7131	6032	2.814266	4.706787	0	0	0	0	0	0	1	-360	360;
	13017	6032	25.15305	22.92729	0	0	0	0	0	0	1	-360	360;
	6032	13432	0.00032	0.00223	0	0	0	0	0	0	1	-360	360;
	6032	6550	1.353809	4.329363	0	0	0	0	0	0	1	-360	360;
	964	6032	0.460582	2.315097	0	0	0	0	0	0	1	-360	360;
	13282	6032	0.0574	0.796953	0	0	0	0	0	0	1	-360	360;
	11793	2977	0.00439	0.016341	0	0	0	0	0	0	1	-360	360;
	385	8316	0.004039	0.019031	0	0	0	0	0	0	1	-360	360;
	385	12536	0.001331	0.007669	0	0	0	0	0	0	1	-360	360;
	8316	10603	0.001091	0.005159	0	0	0	0	0	0	1	-360	360;
	12536	2935	0.00131	0.00782	0	0	0	0	0	0	1	-360	360;
	2935	3932	0.00449	0.022021	0	0	0	0	0	0	1	-360	360;
	6510	3603	0	0.0002	0	0	0	0	0	0	1	-360	360;
	6510	3603	0.001	0.229342	0	0	0	0	0	0	1	-360	360;
	1665	6510	0.159093	0.880956	0	0	0	0	0	0	1	-360	360;
	4490	6510	0.11009	0.785942	0	0	0	0	0	0	1	-360	360;
	2872	6510	0.00133	0.01275	0	0	0	0	0	0	1	-360	360;
	7278	6510	0.08106	0.897645	0	0	0	0	0	0	1	-360	360;
	13017	6510	0.0009	0.00658	0	0	0	0	0	0	1	-360	360;
	13017	6510	1.066967	3.005956	0	0	0	0	0	0	1	-360	360;
	2897	6510	0.00037	0.00335	0	0	0	0	0	0	1	-360	360;
	1665	3603	0.14919	0.968975	0	0	0	0	0	0	1	-360	360;
	4490	3603	0.107472	0.83054	0	0	0	0	0	0	1	-360	360;
	3603	2174	6.591066	12.31994	0	0	0	0	0	0	1	-360	360;
	7278	3603	0.089861	0.931163	0	0	0	0	0	0	1	-360	360;
	7278	3603	0.00151	0.0125	0	0	0	0	0	0	1	-360	360;
	7131	3603	7.50831	11.99515	0	0	0	0	0	0	1	-360	360;
	13017	3603	57.05332	52.93698	0	0	0	0	0	0	1	-360	360;
	13017	3603	1.218213	2.697091	0	0	0	0	0	0	1	-360	360;
	13017	3603	0.00091	0.00644	0	0	0	0	0	0	1	-360	360;
	6550	3603	5.391066	9.229224	0	0	0	0	0	0	1	-360	360;
	964	3603	1.899723	5.672438	0	0	0	0	0	0	1	-360	360;
	1670	3603	0.00034	0.00341	0	0	0	0	0	0	1	-360	360;
	13282	3603	0.372853	1.916066	0	0	0	0	0	0	1	-360	360;
	260	4180	0.001	0.398657	0	0	0	0	0	0	1	-360	360;
	260	3119	0.00157	0.00875	0	0	0	0	0	0	1	-360	360;
	2774	4180	0.013419	0.18632	0	0	0	0	0	0	1	-360	360;
	2774	3119	0.001771	0.00857	0	0	0	0	0	0	1	-360	360;
	5912	10382	3e-05	0.00035	0	0	0	0	0	0	1	-360	360;
	5912	6823	4e-05	0.00047	0	0	0	0	0	0	1	-360	360;
	11448	11302	6.406818	19.24731	0	0	0	0	0	0	1	-360	360;
	2359	11302	5.408471	18.64401	0	0	0	0	0	0	1	-360	360;
	366	11302	4.777893	12.85021	0	0	0	0	0	0	1	-360	360;
	11302	1535	7.316529	22.04545	0	0	0	0	0	0	1	-360	360;
	11302	1535	0.00518	0.053229	0	0	0	0	0	0	1	-360	360;
	7014	11302	1.28	5.819421	0	0	0	0	0	0	1	-360	360;
	12720	11302	1.306178	5.47314	0	0	0	0	0	0	1	-360	360;
	11302	9401	0.001	0.244421	0	0	0	0	0	0	1	-360	360;
	1566	11302	0.326384	2.031736	0	0	0	0	0	0	1	-360	360;
	5754	11302	0.34876	2.295661	0	0	0	0	0	0	1	-360	360;
	11302	9401	0.000401	0.00324	0	0	0	0	0	0	1	-360	360;
	272	12622	1.103116	4.165166	0	0	0	0	0	0	1	-360	360;
	4883	12622	0.982756	4.720568	0	0	0	0	0	0	1	-360	360;
	7997	12622	17.77701	32.87327	0	0	0	0	0	0	1	-360	360;
	12622	6414	11.94321	24.78324	0	0	0	0	0	0	1	-360	360;
	12406	12622	0.13187	1.326939	0	0	0	0	0	0	1	-360	360;
	8314	12622	0.129169	1.352909	0	0	0	0	0	0	1	-360	360;
	9085	12622	15.47438	32.5367	0	0	0	0	0	0	1	-360	360;
	12866	12622	0.00019	0.00211	0	0	0	0	0	0	1	-360	360;
	12866	12622	0.0002	0.00206	0	0	0	0	0	0	1	-360	360;
	11263	5667	0.00013	0.001771	0	0	0	0	0	0	1	-360	360;
	9820	7765	0.008081	0.041651	0	0	0	0	0	0	1	-360	360;
	9820	7765	0.168091	0.83657	0	0	0	0	0	0	1	-360	360;
	9820	7765	0.008229	0.040771	0	0	0	0	0	0	1	-360	360;
	7733	7765	2.293802	5.98719	0	0	0	0	0	0	1	-360	360;
	3193	7765	2.825	5.516736	0	0	0	0	0	0	1	-360	360;
	7765	6386	0.731178	1.985165	0	0	0	0	0	0	1	-360	360;
	11447	9820	0.001349	0.01345	0	0	0	0	0	0	1	-360	360;
	945	11447	0.00051	0.00643	0	0	0	0	0	0	1	-360	360;
	10905	9820	0.00137	0.013159	0	0	0	0	0	0	1	-360	360;
	945	10905	0.0005	0.00655	0	0	0	0	0	0	1	-360	360;
	13354	9636	0.001	0.425381	0	0	0	0	0	0	1	-360	360;
	9506	9636	4.659141	9.407202	0	0	0	0	0	0	1	-360	360;
	1172	9636	1.022853	3.286357	0	0	0	0	0	0	1	-360	360;
	9636	5569	1.18795	4.384626	0	0	0	0	0	0	1	-360	360;
	9636	8176	1.213781	4.303324	0	0	0	0	0	0	1	-360	360;
	6825	9636	8.581024	12.40582	0	0	0	0	0	0	1	-360	360;
	10957	9636	0.203857	1.400623	0	0	0	0	0	0	1	-360	360;
	9636	5569	0.00082	0.0099	0	0	0	0	0	0	1	-360	360;
	10957	9636	0.00056	0.0066	0	0	0	0	0	0	1	-360	360;
	6570	9636	0.00031	0.00634	0	0	0	0	0	0	1	-360	360;
	9506	13354	4.02687	9.790858	0	0	0	0	0	0	1	-360	360;
	1172	13354	1.000554	3.350416	0	0	0	0	0	0	1	-360	360;
	13354	8176	0.00081	0.0101	0	0	0	0	0	0	1	-360	360;
	13354	5569	1.418213	3.650623	0	0	0	0	0	0	1	-360	360;
	13354	8176	1.263504	4.137327	0	0	0	0	0	0	1	-360	360;
	6825	13354	8.415512	12.67105	0	0	0	0	0	0	1	-360	360;
	10957	13354	0.172188	1.486496	0	0	0	0	0	0	1	-360	360;
	7733	9820	0.008531	0.04201	0	0	0	0	0	0	1	-360	360;
	7733	9820	0.19088	1.317025	0	0	0	0	0	0	1	-360	360;
	3193	9820	0.20351	1.242748	0	0	0	0	0	0	1	-360	360;
	9820	10603	0.003419	0.02226	0	0	0	0	0	0	1	-360	360;
	9820	6386	0.05611	0.44343	0	0	0	0	0	0	1	-360	360;
	3193	9820	0.00869	0.041091	0	0	0	0	0	0	1	-360	360;
	9820	6102	0.003349	0.02268	0	0	0	0	0	0	1	-360	360;
	13306	9820	0.00176	0.018651	0	0	0	0	0	0	1	-360	360;
	11905	2741	0.00019	0.00245	0	0	0	0	0	0	1	-360	360;
	11905	2741	0.00019	0.00227	0	0	0	0	0	0	1	-360	360;
	7915	5569	0.00018	0.00209	0	0	0	0	0	0	1	-360	360;
	3170	8176	0.00017	0.00213	0	0	0	0	0	0	1	-360	360;
	9863	622	0.00024	0.0025	0	0	0	0	0	0	1	-360	360;
	9506	1172	0.897576	2.903878	0	0	0	0	0	0	1	-360	360;
	6825	9506	0.01432	0.301302	0	0	0	0	0	0	1	-360	360;
	8391	9506	0.00025	0.00245	0	0	0	0	0	0	1	-360	360;
	9506	10957	3.381094	7.270776	0	0	0	0	0	0	1	-360	360;
	687	4490	0.00142	0.01435	0	0	0	0	0	0	1	-360	360;
	687	1550	0.00023	0.00181	0	0	0	0	0	0	1	-360	360;
	10313	4490	0.00139	0.01463	0	0	0	0	0	0	1	-360	360;
	10313	10000	0.00023	0.00178	0	0	0	0	0	0	1	-360	360;
	1665	4798	0.00047	0.00384	0	0	0	0	0	0	1	-360	360;
	1665	4307	0.0005	0.00403	0	0	0	0	0	0	1	-360	360;
	7278	4307	0.00025	0.00236	0	0	0	0	0	0	1	-360	360;
	1665	4490	0.838573	2.37223	0	0	0	0	0	0	1	-360	360;
	7278	1665	0.078193	0.509841	0	0	0	0	0	0	1	-360	360;
	13017	1665	5.654848	7.590028	0	0	0	0	0	0	1	-360	360;
	12197	4490	0.0009	0.0069	0	0	0	0	0	0	1	-360	360;
	13017	12197	0.0002	0.00149	0	0	0	0	0	0	1	-360	360;
	13536	4490	0.00092	0.00675	0	0	0	0	0	0	1	-360	360;
	13017	13536	0.0002	0.00145	0	0	0	0	0	0	1	-360	360;
	7278	4490	0.413608	1.666759	0	0	0	0	0	0	1	-360	360;
	13017	4490	0.03562	0.272507	0	0	0	0	0	0	1	-360	360;
	7162	906	0.0001	0.00144	0	0	0	0	0	0	1	-360	360;
	12866	7162	9e-05	0.00129	0	0	0	0	0	0	1	-360	360;
	2272	9667	0.0001	0.00133	0	0	0	0	0	0	1	-360	360;
	12866	2272	9e-05	0.00132	0	0	0	0	0	0	1	-360	360;
	906	6344	0.00063	0.00676	0	0	0	0	0	0	1	-360	360;
	9667	3353	0.00065	0.00662	0	0	0	0	0	0	1	-360	360;
	6362	6344	0.00017	0.00236	0	0	0	0	0	0	1	-360	360;
	6362	3353	0.00017	0.00219	0	0	0	0	0	0	1	-360	360;
	1758	13213	0.0017	0.007469	0	0	0	0	0	0	1	-360	360;
	13213	9140	0.001039	0.005599	0	0	0	0	0	0	1	-360	360;
	1758	10921	0.001729	0.00731	0	0	0	0	0	0	1	-360	360;
	10921	141	0.001021	0.0057	0	0	0	0	0	0	1	-360	360;
	1758	2554	0.00168	0.007159	0	0	0	0	0	0	1	-360	360;
	2554	9140	0.00112	0.00495	0	0	0	0	0	0	1	-360	360;
	13213	10921	0.001	0.364917	0	0	0	0	0	0	1	-360	360;
	1758	13213	0.106669	0.821178	0	0	0	0	0	0	1	-360	360;
	13213	9140	0.0977	0.935248	0	0	0	0	0	0	1	-360	360;
	13213	141	0.10481	0.756446	0	0	0	0	0	0	1	-360	360;
	1758	10921	0.094539	0.895083	0	0	0	0	0	0	1	-360	360;
	10921	9140	0.09563	0.921384	0	0	0	0	0	0	1	-360	360;
	10921	141	0.085479	0.810062	0	0	0	0	0	0	1	-360	360;
	6362	5353	0.0002	0.0021	0	0	0	0	0	0	1	-360	360;
	5353	9567	0.00051	0.00613	0	0	0	0	0	0	1	-360	360;
	10699	8993	0.00062	0.006331	0	0	0	0	0	0	1	-360	360;
	8993	2264	0.00075	0.00981	0	0	0	0	0	0	1	-360	360;
	6362	2342	0.00105	0.01369	0	0	0	0	0	0	1	-360	360;
	2342	722	0.00132	0.01291	0	0	0	0	0	0	1	-360	360;
	6362	8975	0.00103	0.01395	0	0	0	0	0	0	1	-360	360;
	8975	11650	0.00134	0.01263	0	0	0	0	0	0	1	-360	360;
	6362	8653	0.269647	1.156579	0	0	0	0	0	0	1	-360	360;
	478	6362	0.06211	0.718075	0	0	0	0	0	0	1	-360	360;
	6362	866	0.00062	0.005	0	0	0	0	0	0	1	-360	360;
	7322	6362	0.00031	0.00248	0	0	0	0	0	0	1	-360	360;
	6362	3429	0.0006	0.0051	0	0	0	0	0	0	1	-360	360;
	1167	3932	0.007331	0.044461	0	0	0	0	0	0	1	-360	360;
	1167	3932	0.00718	0.0453	0	0	0	0	0	0	1	-360	360;
	10699	556	0.0002	0.00211	0	0	0	0	0	0	1	-360	360;
	439	556	0.000659	0.002909	0	0	0	0	0	0	1	-360	360;
	9732	10899	0.00088	0.0079	0	0	0	0	0	0	1	-360	360;
	13454	2749	0.00089	0.00774	0	0	0	0	0	0	1	-360	360;
	8313	6380	0.00089	0.00706	0	0	0	0	0	0	1	-360	360;
	2174	2648	0.0009	0.0069	0	0	0	0	0	0	1	-360	360;
	7131	2174	0.527957	1.890097	0	0	0	0	0	0	1	-360	360;
	13017	2174	6.434972	8.855956	0	0	0	0	0	0	1	-360	360;
	6550	2174	0.02292	0.469398	0	0	0	0	0	0	1	-360	360;
	964	2174	0.03034	0.81482	0	0	0	0	0	0	1	-360	360;
	13282	2174	0.225173	1.343006	0	0	0	0	0	0	1	-360	360;
	7767	4900	0.0001	0.00122	0	0	0	0	0	0	1	-360	360;
	11732	4900	0.00025	0.00221	0	0	0	0	0	0	1	-360	360;
	131	10838	2.696676	13.13573	0	0	0	0	0	0	1	-360	360;
	131	8035	3.125	11.9903	0	0	0	0	0	0	1	-360	360;
	5912	131	0.00054	0.00664	0	0	0	0	0	0	1	-360	360;
	11812	131	0.899169	5.002355	0	0	0	0	0	0	1	-360	360;
	12474	131	0.06172	1.437119	0	0	0	0	0	0	1	-360	360;
	91	131	0.521247	2.927216	0	0	0	0	0	0	1	-360	360;
	11732	131	0.00397	0.614771	0	0	0	0	0	0	1	-360	360;
	12274	131	0.956579	5.569321	0	0	0	0	0	0	1	-360	360;
	3895	131	0.071011	1.246122	0	0	0	0	0	0	1	-360	360;
	1789	131	22.47576	40.81994	0	0	0	0	0	0	1	-360	360;
	919	131	0.00141	0.00778	0	0	0	0	0	0	1	-360	360;
	1747	2648	0.00021	0.00179	0	0	0	0	0	0	1	-360	360;
	6550	6380	0.00021	0.00175	0	0	0	0	0	0	1	-360	360;
	7454	5003	0.00051	0.00411	0	0	0	0	0	0	1	-360	360;
	2586	11186	0.00052	0.00402	0	0	0	0	0	0	1	-360	360;
	5003	11186	0.00439	0.248677	0	0	0	0	0	0	1	-360	360;
	7733	3193	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7733	3193	0.001	0.289959	0	0	0	0	0	0	1	-360	360;
	7733	5496	9.9e-05	0.000789	0	0	0	0	0	0	1	-360	360;
	7733	6386	0.13343	0.727438	0	0	0	0	0	0	1	-360	360;
	3193	6386	0.11368	0.773802	0	0	0	0	0	0	1	-360	360;
	752	10995	0.0001	0.0011	0	0	0	0	0	0	1	-360	360;
	7211	752	0.00069	0.00512	0	0	0	0	0	0	1	-360	360;
	2739	2055	0.0001	0.00102	0	0	0	0	0	0	1	-360	360;
	3504	2739	0.0007	0.005	0	0	0	0	0	0	1	-360	360;
	8269	5671	0.00031	0.00399	0	0	0	0	0	0	1	-360	360;
	10838	8269	9e-05	0.00103	0	0	0	0	0	0	1	-360	360;
	5671	7845	0.00021	0.00231	0	0	0	0	0	0	1	-360	360;
	7845	3052	8e-05	0.00082	0	0	0	0	0	0	1	-360	360;
	3052	4721	0.00016	0.00203	0	0	0	0	0	0	1	-360	360;
	4721	8765	3e-05	0.00036	0	0	0	0	0	0	1	-360	360;
	91	8765	0.00061	0.00766	0	0	0	0	0	0	1	-360	360;
	10838	8035	0	0.0002	0	0	0	0	0	0	1	-360	360;
	10838	8035	0.01257	0.43678	0	0	0	0	0	0	1	-360	360;
	11812	10838	22.73892	49	0	0	0	0	0	0	1	-360	360;
	12474	10838	4.844044	19.18698	0	0	0	0	0	0	1	-360	360;
	91	10838	0.001	1.752701	0	0	0	0	0	0	1	-360	360;
	11732	10838	0.908172	6.096537	0	0	0	0	0	0	1	-360	360;
	10349	10838	4.358657	9.756232	0	0	0	0	0	0	1	-360	360;
	12274	10838	25.29224	66.18213	0	0	0	0	0	0	1	-360	360;
	3895	10838	1.071399	5.397922	0	0	0	0	0	0	1	-360	360;
	1789	10838	0.155789	0.985526	0	0	0	0	0	0	1	-360	360;
	11812	8035	23.51524	49.62604	0	0	0	0	0	0	1	-360	360;
	12474	8035	5.179155	18.97368	0	0	0	0	0	0	1	-360	360;
	91	8035	0.0144	1.921053	0	0	0	0	0	0	1	-360	360;
	11732	8035	0.997091	6.027978	0	0	0	0	0	0	1	-360	360;
	10349	8035	4.636704	9.781162	0	0	0	0	0	0	1	-360	360;
	12274	8035	29.33518	59.12049	0	0	0	0	0	0	1	-360	360;
	3895	8035	1.098269	5.695983	0	0	0	0	0	0	1	-360	360;
	1789	8035	0.00085	0.00853	0	0	0	0	0	0	1	-360	360;
	1789	8035	0.15653	1.070014	0	0	0	0	0	0	1	-360	360;
	8888	8653	0.00015	0.00148	0	0	0	0	0	0	1	-360	360;
	699	8653	0.00014	0.00167	0	0	0	0	0	0	1	-360	360;
	478	8653	0.525672	1.551039	0	0	0	0	0	0	1	-360	360;
	9465	2160	0.00411	0.01861	0	0	0	0	0	0	1	-360	360;
	6102	9465	0.000969	0.005841	0	0	0	0	0	0	1	-360	360;
	2160	3932	0.004581	0.02156	0	0	0	0	0	0	1	-360	360;
	3775	9492	0.00026	0.00323	0	0	0	0	0	0	1	-360	360;
	3775	9492	0.00026	0.00329	0	0	0	0	0	0	1	-360	360;
	272	6868	6e-05	0.00065	0	0	0	0	0	0	1	-360	360;
	6868	2257	0.00038	0.00377	0	0	0	0	0	0	1	-360	360;
	4883	13020	6e-05	0.00066	0	0	0	0	0	0	1	-360	360;
	13020	4997	0.00038	0.00369	0	0	0	0	0	0	1	-360	360;
	272	4883	0.01122	0.71669	0	0	0	0	0	0	1	-360	360;
	7997	272	25.10803	39.16205	0	0	0	0	0	0	1	-360	360;
	272	6414	32.07964	43.87258	0	0	0	0	0	0	1	-360	360;
	272	12406	0.427673	3.90554	0	0	0	0	0	0	1	-360	360;
	272	8314	0.418573	3.979224	0	0	0	0	0	0	1	-360	360;
	9085	272	23.26662	36.54363	0	0	0	0	0	0	1	-360	360;
	7997	4883	25.59141	38.34626	0	0	0	0	0	0	1	-360	360;
	4883	6414	32.6849	42.93906	0	0	0	0	0	0	1	-360	360;
	4883	12406	0.409481	3.684488	0	0	0	0	0	0	1	-360	360;
	4883	8314	0.464072	3.610803	0	0	0	0	0	0	1	-360	360;
	9085	4883	23.70568	35.76593	0	0	0	0	0	0	1	-360	360;
	1467	5005	0	0.0002	0	0	0	0	0	0	1	-360	360;
	1467	5005	0.001	0.19151	0	0	0	0	0	0	1	-360	360;
	1467	5923	17.56446	20.66736	0	0	0	0	0	0	1	-360	360;
	1467	1752	13.87645	21.76446	0	0	0	0	0	0	1	-360	360;
	1467	644	1.145248	3.637603	0	0	0	0	0	0	1	-360	360;
	1467	2726	1.372004	3.186983	0	0	0	0	0	0	1	-360	360;
	5344	1467	0.119159	0.998905	0	0	0	0	0	0	1	-360	360;
	1467	2264	6.052479	9.575826	0	0	0	0	0	0	1	-360	360;
	4342	1467	4.474793	10.74483	0	0	0	0	0	0	1	-360	360;
	1684	1467	4.534711	10.47087	0	0	0	0	0	0	1	-360	360;
	5005	5923	18.49236	22.54545	0	0	0	0	0	0	1	-360	360;
	5005	1752	13.69401	22.82645	0	0	0	0	0	0	1	-360	360;
	5005	644	1.333636	3.866942	0	0	0	0	0	0	1	-360	360;
	5005	2726	1.179442	4.148554	0	0	0	0	0	0	1	-360	360;
	5344	5005	0.114281	1.035289	0	0	0	0	0	0	1	-360	360;
	5005	2264	5.617769	11.75372	0	0	0	0	0	0	1	-360	360;
	4342	5005	4.939669	11.00083	0	0	0	0	0	0	1	-360	360;
	1684	5005	4.903306	10.93058	0	0	0	0	0	0	1	-360	360;
	11272	2055	2e-05	0.00021	0	0	0	0	0	0	1	-360	360;
	11272	4941	0.00028	0.00263	0	0	0	0	0	0	1	-360	360;
	4767	10995	2e-05	0.00023	0	0	0	0	0	0	1	-360	360;
	4767	10496	0.00027	0.00268	0	0	0	0	0	0	1	-360	360;
	1758	9140	0.02601	0.665517	0	0	0	0	0	0	1	-360	360;
	1758	141	0.028591	0.528326	0	0	0	0	0	0	1	-360	360;
	2019	3464	0.01088	0.212832	0	0	0	0	0	0	1	-360	360;
	3498	3464	0.00032	0.00331	0	0	0	0	0	0	1	-360	360;
	6250	3464	0.00031	0.00306	0	0	0	0	0	0	1	-360	360;
	1172	353	0.00057	0.01002	0	0	0	0	0	0	1	-360	360;
	1172	3252	0.00053	0.00558	0	0	0	0	0	0	1	-360	360;
	1172	5569	22.58241	30.56787	0	0	0	0	0	0	1	-360	360;
	1172	8176	22.12119	31.16759	0	0	0	0	0	0	1	-360	360;
	6825	1172	1.627632	4.119529	0	0	0	0	0	0	1	-360	360;
	12961	1172	0.0003	0.00319	0	0	0	0	0	0	1	-360	360;
	10957	1172	0.00232	0.519501	0	0	0	0	0	0	1	-360	360;
	1172	13309	0.00058	0.00981	0	0	0	0	0	0	1	-360	360;
	13441	13309	0.00065	0.00643	0	0	0	0	0	0	1	-360	360;
	6598	353	0.00063	0.00655	0	0	0	0	0	0	1	-360	360;
	2872	7278	0.001	0.215229	0	0	0	0	0	0	1	-360	360;
	2872	2019	0.00102	0.00911	0	0	0	0	0	0	1	-360	360;
	7278	9595	0.00094	0.00901	0	0	0	0	0	0	1	-360	360;
	13017	7278	3.359072	5.253809	0	0	0	0	0	0	1	-360	360;
	1550	797	0.00019	0.00221	0	0	0	0	0	0	1	-360	360;
	2019	1550	0.00039	0.00355	0	0	0	0	0	0	1	-360	360;
	1550	1649	0.0001	0.00073	0	0	0	0	0	0	1	-360	360;
	10000	880	0.00018	0.00226	0	0	0	0	0	0	1	-360	360;
	12309	10000	0.00038	0.00368	0	0	0	0	0	0	1	-360	360;
	10000	10400	0.00047	0.00386	0	0	0	0	0	0	1	-360	360;
	9140	141	0.001	0.309525	0	0	0	0	0	0	1	-360	360;
	4180	3119	0.000711	0.00393	0	0	0	0	0	0	1	-360	360;
	4180	11169	0.001219	0.00569	0	0	0	0	0	0	1	-360	360;
	4262	9176	0.00125	0.005581	0	0	0	0	0	0	1	-360	360;
	3119	9176	0.0007	0.00401	0	0	0	0	0	0	1	-360	360;
	13017	7131	0.0599	0.313109	0	0	0	0	0	0	1	-360	360;
	7131	4762	0.00051	0.00517	0	0	0	0	0	0	1	-360	360;
	7131	6550	0.92403	2.090097	0	0	0	0	0	0	1	-360	360;
	7131	964	0.05633	0.587403	0	0	0	0	0	0	1	-360	360;
	7131	13282	0.333109	1.178601	0	0	0	0	0	0	1	-360	360;
	439	10545	0.0002	0.002599	0	0	0	0	0	0	1	-360	360;
	3025	5912	6e-05	0.00066	0	0	0	0	0	0	1	-360	360;
	5912	797	0.00203	0.01921	0	0	0	0	0	0	1	-360	360;
	5912	12274	0.00132	0.01562	0	0	0	0	0	0	1	-360	360;
	5912	2741	0.00039	0.00541	0	0	0	0	0	0	1	-360	360;
	5912	12274	0.0013	0.01592	0	0	0	0	0	0	1	-360	360;
	5912	9744	0.00084	0.00852	0	0	0	0	0	0	1	-360	360;
	3025	4398	0.00016	0.0017	0	0	0	0	0	0	1	-360	360;
	11824	7454	0.00092	0.00761	0	0	0	0	0	0	1	-360	360;
	11824	722	0.00062	0.00468	0	0	0	0	0	0	1	-360	360;
	8922	2586	0.00094	0.00746	0	0	0	0	0	0	1	-360	360;
	8922	11650	0.00063	0.00458	0	0	0	0	0	0	1	-360	360;
	7454	2586	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7454	2586	0.001	0.227327	0	0	0	0	0	0	1	-360	360;
	7454	5380	3e-05	0.00034	0	0	0	0	0	0	1	-360	360;
	964	7454	0.00052	0.00501	0	0	0	0	0	0	1	-360	360;
	5939	2586	0.00059	0.00491	0	0	0	0	0	0	1	-360	360;
	357	1752	0.006229	0.029719	0	0	0	0	0	0	1	-360	360;
	357	644	0.00194	0.00924	0	0	0	0	0	0	1	-360	360;
	4507	3773	0.00189	0.00769	0	0	0	0	0	0	1	-360	360;
	4507	10991	0.0037	0.017021	0	0	0	0	0	0	1	-360	360;
	6740	1645	0.00168	0.008719	0	0	0	0	0	0	1	-360	360;
	6740	9023	0.003771	0.016669	0	0	0	0	0	0	1	-360	360;
	12951	11169	0.004669	0.024729	0	0	0	0	0	0	1	-360	360;
	1696	4262	0	0.0002	0	0	0	0	0	0	1	-360	360;
	1696	4262	0.00099	0.00607	0	0	0	0	0	0	1	-360	360;
	12951	1696	0.00476	0.02624	0	0	0	0	0	0	1	-360	360;
	1752	8468	0.00051	0.00286	0	0	0	0	0	0	1	-360	360;
	1752	8468	0.0005	0.002909	0	0	0	0	0	0	1	-360	360;
	5923	1752	0	0.0002	0	0	0	0	0	0	1	-360	360;
	5923	1752	0.001	0.18874	0	0	0	0	0	0	1	-360	360;
	5923	644	0.704421	2.986983	0	0	0	0	0	0	1	-360	360;
	5923	2726	0.680207	2.970661	0	0	0	0	0	0	1	-360	360;
	5344	5923	4.389256	10.44091	0	0	0	0	0	0	1	-360	360;
	5923	2264	0.00861	0.432252	0	0	0	0	0	0	1	-360	360;
	4342	5923	0.006521	0.039849	0	0	0	0	0	0	1	-360	360;
	4342	5923	0.450599	2.525826	0	0	0	0	0	0	1	-360	360;
	1684	5923	0.51345	2.178099	0	0	0	0	0	0	1	-360	360;
	5923	2801	0.000159	0.002091	0	0	0	0	0	0	1	-360	360;
	1752	644	0.547603	3.122934	0	0	0	0	0	0	1	-360	360;
	1752	2726	0.63281	2.590496	0	0	0	0	0	0	1	-360	360;
	5344	1752	4.541736	8.227686	0	0	0	0	0	0	1	-360	360;
	1752	2264	0.00687	0.424525	0	0	0	0	0	0	1	-360	360;
	4342	1752	0.394752	2.341736	0	0	0	0	0	0	1	-360	360;
	1684	1752	0.00638	0.04063	0	0	0	0	0	0	1	-360	360;
	1684	1752	0.459607	1.981798	0	0	0	0	0	0	1	-360	360;
	1752	2772	0.000159	0.00213	0	0	0	0	0	0	1	-360	360;
	3773	1645	0.001	0.319917	0	0	0	0	0	0	1	-360	360;
	4860	3773	0.074539	0.567996	0	0	0	0	0	0	1	-360	360;
	4860	1645	0.054219	0.438326	0	0	0	0	0	0	1	-360	360;
	2382	6053	0.00049	0.0046	0	0	0	0	0	0	1	-360	360;
	12058	6053	0.00063	0.00676	0	0	0	0	0	0	1	-360	360;
	2382	3288	0.02777	0.440042	0	0	0	0	0	0	1	-360	360;
	5152	3288	0.02704	0.444668	0	0	0	0	0	0	1	-360	360;
	5152	3288	0.00048	0.00469	0	0	0	0	0	0	1	-360	360;
	12058	3288	0.00062	0.00626	0	0	0	0	0	0	1	-360	360;
	2382	5152	0	0.0002	0	0	0	0	0	0	1	-360	360;
	2382	5152	0.001	0.247292	0	0	0	0	0	0	1	-360	360;
	2382	8683	3e-05	0.00035	0	0	0	0	0	0	1	-360	360;
	11902	2382	0.00014	0.00118	0	0	0	0	0	0	1	-360	360;
	5152	11391	3e-05	0.00034	0	0	0	0	0	0	1	-360	360;
	6595	8683	0.00072	0.00903	0	0	0	0	0	0	1	-360	360;
	6595	11391	0.00082	0.00885	0	0	0	0	0	0	1	-360	360;
	10899	1366	0.00012	0.00093	0	0	0	0	0	0	1	-360	360;
	10147	10603	0.000831	0.004229	0	0	0	0	0	0	1	-360	360;
	13375	6102	0.000841	0.00414	0	0	0	0	0	0	1	-360	360;
	12961	8888	0.00049	0.00607	0	0	0	0	0	0	1	-360	360;
	10957	3252	0.00048	0.00567	0	0	0	0	0	0	1	-360	360;
	10957	699	0.00047	0.00578	0	0	0	0	0	0	1	-360	360;
	7997	9085	6e-05	0.00069	0	0	0	0	0	0	1	-360	360;
	7997	6414	0.198269	1.245429	0	0	0	0	0	0	1	-360	360;
	7997	12406	28.23615	48.46122	0	0	0	0	0	0	1	-360	360;
	7997	8314	32.00069	47.49169	0	0	0	0	0	0	1	-360	360;
	7997	9085	0.001	0.339079	0	0	0	0	0	0	1	-360	360;
	5569	8176	0.001	0.23876	0	0	0	0	0	0	1	-360	360;
	10957	5569	7.242382	14.44598	0	0	0	0	0	0	1	-360	360;
	10957	8176	7.08795	14.71814	0	0	0	0	0	0	1	-360	360;
	13629	8158	0.00493	0.059419	0	0	0	0	0	0	1	-360	360;
	13629	8158	0.01006	0.07863	0	0	0	0	0	0	1	-360	360;
	13629	8267	0.001	0.434587	0	0	0	0	0	0	1	-360	360;
	13629	4147	0.001	0.434132	0	0	0	0	0	0	1	-360	360;
	13629	6581	9.525206	20.12748	0	0	0	0	0	0	1	-360	360;
	13629	11448	0.490393	3.121901	0	0	0	0	0	0	1	-360	360;
	13629	2359	0.324483	3.440909	0	0	0	0	0	0	1	-360	360;
	13629	8608	7.110331	13.27211	0	0	0	0	0	0	1	-360	360;
	13629	4055	8.147314	13.64236	0	0	0	0	0	0	1	-360	360;
	13629	8156	1.873657	4.730165	0	0	0	0	0	0	1	-360	360;
	13629	6030	1.953223	4.544628	0	0	0	0	0	0	1	-360	360;
	13629	917	0.893533	3.267562	0	0	0	0	0	0	1	-360	360;
	13629	4995	0.947417	3.197521	0	0	0	0	0	0	1	-360	360;
	13629	12407	1.735227	5.478926	0	0	0	0	0	0	1	-360	360;
	12043	13629	49.21281	47.44215	0	0	0	0	0	0	1	-360	360;
	9823	13629	21.70248	30.12603	0	0	0	0	0	0	1	-360	360;
	13629	6896	9.99835	24.78719	0	0	0	0	0	0	1	-360	360;
	13629	10979	10.26198	26.88017	0	0	0	0	0	0	1	-360	360;
	13629	1535	0.625785	4.196281	0	0	0	0	0	0	1	-360	360;
	5183	13629	0.531426	3.034091	0	0	0	0	0	0	1	-360	360;
	7916	13629	3.245868	10.84669	0	0	0	0	0	0	1	-360	360;
	3743	13629	3.282025	10.98388	0	0	0	0	0	0	1	-360	360;
	13629	2256	0.002919	0.01451	0	0	0	0	0	0	1	-360	360;
	13629	2256	0.001	0.67564	0	0	0	0	0	0	1	-360	360;
	13629	7691	0.227169	1.409215	0	0	0	0	0	0	1	-360	360;
	13629	6065	0.192229	1.194917	0	0	0	0	0	0	1	-360	360;
	13629	2256	0.002979	0.0142	0	0	0	0	0	0	1	-360	360;
	7548	8045	0.000151	0.0015	0	0	0	0	0	0	1	-360	360;
	7548	6386	0.000151	0.001461	0	0	0	0	0	0	1	-360	360;
	9492	2123	0.00136	0.0167	0	0	0	0	0	0	1	-360	360;
	8267	4147	0	0.0002	0	0	0	0	0	0	1	-360	360;
	8267	4147	0.001	0.256632	0	0	0	0	0	0	1	-360	360;
	6581	8267	6.78595	11.8655	0	0	0	0	0	0	1	-360	360;
	11448	8267	1.097169	5.085744	0	0	0	0	0	0	1	-360	360;
	2359	8267	0.904029	5.181818	0	0	0	0	0	0	1	-360	360;
	8608	8267	4.170248	8.645247	0	0	0	0	0	0	1	-360	360;
	4055	8267	4.773967	8.884503	0	0	0	0	0	0	1	-360	360;
	8156	8267	0.874421	3.172727	0	0	0	0	0	0	1	-360	360;
	6030	8267	1.05436	2.929339	0	0	0	0	0	0	1	-360	360;
	917	8267	0.513368	2.495455	0	0	0	0	0	0	1	-360	360;
	4995	8267	0.604525	2.218802	0	0	0	0	0	0	1	-360	360;
	12407	8267	1.308616	3.033264	0	0	0	0	0	0	1	-360	360;
	12043	8267	39.69008	38.26033	0	0	0	0	0	0	1	-360	360;
	9823	8267	17.11942	22.4938	0	0	0	0	0	0	1	-360	360;
	6896	8267	20.47789	32.19835	0	0	0	0	0	0	1	-360	360;
	10979	8267	21.35331	30.98554	0	0	0	0	0	0	1	-360	360;
	1535	8267	1.604091	6.167149	0	0	0	0	0	0	1	-360	360;
	5183	8267	1.137851	5.037397	0	0	0	0	0	0	1	-360	360;
	7916	8267	6.625413	15.09483	0	0	0	0	0	0	1	-360	360;
	3743	8267	6.196074	16.5874	0	0	0	0	0	0	1	-360	360;
	4717	8267	59.24587	56.3595	0	0	0	0	0	0	1	-360	360;
	8267	2256	0.086409	1.077231	0	0	0	0	0	0	1	-360	360;
	7691	8267	0.080841	0.916612	0	0	0	0	0	0	1	-360	360;
	6065	8267	0.07195	0.701488	0	0	0	0	0	0	1	-360	360;
	6581	4147	5.952479	12.55207	0	0	0	0	0	0	1	-360	360;
	11448	4147	1.208636	4.790289	0	0	0	0	0	0	1	-360	360;
	2359	4147	0.897293	4.88781	0	0	0	0	0	0	1	-360	360;
	8608	4147	4.399793	8.436157	0	0	0	0	0	0	1	-360	360;
	4055	4147	4.466736	9.775413	0	0	0	0	0	0	1	-360	360;
	8156	4147	0.924855	3.415083	0	0	0	0	0	0	1	-360	360;
	6030	4147	1.008368	3.168595	0	0	0	0	0	0	1	-360	360;
	917	4147	0.519401	2.533471	0	0	0	0	0	0	1	-360	360;
	4995	4147	0.611653	2.252893	0	0	0	0	0	0	1	-360	360;
	12407	4147	1.180868	3.490496	0	0	0	0	0	0	1	-360	360;
	12043	4147	38.08471	41.4876	0	0	0	0	0	0	1	-360	360;
	9823	4147	19.69132	22.41116	0	0	0	0	0	0	1	-360	360;
	6896	4147	21.63017	31.43802	0	0	0	0	0	0	1	-360	360;
	10979	4147	20.91529	32.94215	0	0	0	0	0	0	1	-360	360;
	1535	4147	1.695537	6.03657	0	0	0	0	0	0	1	-360	360;
	5183	4147	1.302851	4.551033	0	0	0	0	0	0	1	-360	360;
	7916	4147	5.718182	16.36116	0	0	0	0	0	0	1	-360	360;
	3743	4147	6.683678	15.9062	0	0	0	0	0	0	1	-360	360;
	4147	2256	0.10661	0.987583	0	0	0	0	0	0	1	-360	360;
	7691	4147	0.079729	0.948512	0	0	0	0	0	0	1	-360	360;
	6065	4147	0.06449	0.807169	0	0	0	0	0	0	1	-360	360;
	6065	4147	0.002151	0.02656	0	0	0	0	0	0	1	-360	360;
	9492	7042	0.00094	0.01246	0	0	0	0	0	0	1	-360	360;
	7042	7108	0.00109	0.00971	0	0	0	0	0	0	1	-360	360;
	9492	4338	0.00099	0.01086	0	0	0	0	0	0	1	-360	360;
	9492	4338	0.00092	0.0127	0	0	0	0	0	0	1	-360	360;
	4338	7108	0.00097	0.01101	0	0	0	0	0	0	1	-360	360;
	644	2726	0	0.0002	0	0	0	0	0	0	1	-360	360;
	644	2726	0.001	0.236901	0	0	0	0	0	0	1	-360	360;
	5344	644	1.033202	3.18595	0	0	0	0	0	0	1	-360	360;
	2264	644	0.07336	0.952996	0	0	0	0	0	0	1	-360	360;
	4342	644	4.823554	9.755785	0	0	0	0	0	0	1	-360	360;
	1684	644	4.344835	10.74711	0	0	0	0	0	0	1	-360	360;
	11263	644	0.000789	0.01039	0	0	0	0	0	0	1	-360	360;
	5344	2726	0.877293	3.230992	0	0	0	0	0	0	1	-360	360;
	2264	2726	0.08095	0.839008	0	0	0	0	0	0	1	-360	360;
	4342	2726	4.635744	9.702067	0	0	0	0	0	0	1	-360	360;
	1684	2726	4.775207	9.229752	0	0	0	0	0	0	1	-360	360;
	9595	3498	0.00017	0.0018	0	0	0	0	0	0	1	-360	360;
	12309	6250	0.00017	0.00184	0	0	0	0	0	0	1	-360	360;
	2281	4941	6e-05	0.00055	0	0	0	0	0	0	1	-360	360;
	8094	10496	5e-05	0.00051	0	0	0	0	0	0	1	-360	360;
	10367	9178	4.9e-05	0.000194	0	0	0	0	0	0	1	-360	360;
	10367	9178	0.001	0.502417	0	0	0	0	0	0	1	-360	360;
	10367	722	0.11597	1.150069	0	0	0	0	0	0	1	-360	360;
	10367	6355	0.00027	0.00201	0	0	0	0	0	0	1	-360	360;
	10367	478	12.69044	26.69945	0	0	0	0	0	0	1	-360	360;
	8529	10367	5.039543	12.54224	0	0	0	0	0	0	1	-360	360;
	5951	10367	4.861565	13.04917	0	0	0	0	0	0	1	-360	360;
	9178	722	0.131427	1.127078	0	0	0	0	0	0	1	-360	360;
	9178	3605	0.00034	0.00324	0	0	0	0	0	0	1	-360	360;
	9178	478	14.34765	23.73407	0	0	0	0	0	0	1	-360	360;
	8529	9178	4.937535	12.78878	0	0	0	0	0	0	1	-360	360;
	5951	9178	4.758934	13.29571	0	0	0	0	0	0	1	-360	360;
	5344	418	0.00189	0.01011	0	0	0	0	0	0	1	-360	360;
	5344	10864	0.00193	0.009901	0	0	0	0	0	0	1	-360	360;
	475	10864	0.00095	0.004461	0	0	0	0	0	0	1	-360	360;
	475	8634	0.000841	0.00505	0	0	0	0	0	0	1	-360	360;
	9415	5344	0.00074	0.00837	0	0	0	0	0	0	1	-360	360;
	5344	2264	3.807645	6.613017	0	0	0	0	0	0	1	-360	360;
	4342	5344	0.512211	2.208058	0	0	0	0	0	0	1	-360	360;
	1684	5344	0.432727	2.575207	0	0	0	0	0	0	1	-360	360;
	5344	475	0.00551	0.13675	0	0	0	0	0	0	1	-360	360;
	12058	9673	0.00018	0.00165	0	0	0	0	0	0	1	-360	360;
	12058	2275	0.00017	0.00168	0	0	0	0	0	0	1	-360	360;
	4398	7767	0.00022	0.00207	0	0	0	0	0	0	1	-360	360;
	8285	8566	6e-05	0.000531	0	0	0	0	0	0	1	-360	360;
	6581	6464	6e-05	0.000521	0	0	0	0	0	0	1	-360	360;
	6581	8608	2.538636	7.69814	0	0	0	0	0	0	1	-360	360;
	6581	4055	2.881818	7.882025	0	0	0	0	0	0	1	-360	360;
	6581	8156	0.648822	3.687603	0	0	0	0	0	0	1	-360	360;
	6581	6030	0.662004	3.615496	0	0	0	0	0	0	1	-360	360;
	6581	917	0.347417	1.771488	0	0	0	0	0	0	1	-360	360;
	6581	4995	0.302727	2.122521	0	0	0	0	0	0	1	-360	360;
	6581	12407	0.13874	1.055682	0	0	0	0	0	0	1	-360	360;
	12043	6581	32.21488	46.94215	0	0	0	0	0	0	1	-360	360;
	9823	6581	16.64897	25.36157	0	0	0	0	0	0	1	-360	360;
	6581	2256	31.40496	45.72727	0	0	0	0	0	0	1	-360	360;
	6581	7691	1.96531	5.971488	0	0	0	0	0	0	1	-360	360;
	6581	6065	1.628843	5.349793	0	0	0	0	0	0	1	-360	360;
	8566	12407	0.003831	0.04362	0	0	0	0	0	0	1	-360	360;
	6464	12407	0.003901	0.042711	0	0	0	0	0	0	1	-360	360;
	11812	12474	0	0.0002	0	0	0	0	0	0	1	-360	360;
	11812	12474	0.03143	0.812673	0	0	0	0	0	0	1	-360	360;
	10786	11812	6e-05	0.00062	0	0	0	0	0	0	1	-360	360;
	91	11812	5.540582	14.44598	0	0	0	0	0	0	1	-360	360;
	11732	11812	0.171863	2.799169	0	0	0	0	0	0	1	-360	360;
	10349	11812	31.73753	62.63988	0	0	0	0	0	0	1	-360	360;
	12274	11812	0.0037	0.670762	0	0	0	0	0	0	1	-360	360;
	3895	11812	4.69723	16.67659	0	0	0	0	0	0	1	-360	360;
	1789	11812	13.8608	37.26524	0	0	0	0	0	0	1	-360	360;
	8094	12474	6e-05	0.00063	0	0	0	0	0	0	1	-360	360;
	91	12474	0.842036	5.10651	0	0	0	0	0	0	1	-360	360;
	11732	12474	0.001	0.767382	0	0	0	0	0	0	1	-360	360;
	12274	12474	0.0122	0.920499	0	0	0	0	0	0	1	-360	360;
	3895	12474	0.831094	4.707341	0	0	0	0	0	0	1	-360	360;
	1789	12474	13.1392	39.18767	0	0	0	0	0	0	1	-360	360;
	11616	10786	0.00022	0.0023	0	0	0	0	0	0	1	-360	360;
	12129	2281	0.00022	0.00213	0	0	0	0	0	0	1	-360	360;
	12269	6414	0.00034	0.003	0	0	0	0	0	0	1	-360	360;
	478	12269	0.00022	0.00172	0	0	0	0	0	0	1	-360	360;
	7393	6414	0.00034	0.00294	0	0	0	0	0	0	1	-360	360;
	3019	7393	0.00022	0.00168	0	0	0	0	0	0	1	-360	360;
	12406	6414	25.47853	44.51731	0	0	0	0	0	0	1	-360	360;
	8314	6414	24.93629	45.35734	0	0	0	0	0	0	1	-360	360;
	9085	6414	0.138622	1.351247	0	0	0	0	0	0	1	-360	360;
	4342	2264	1.420021	4.04938	0	0	0	0	0	0	1	-360	360;
	1684	2264	1.226446	4.635537	0	0	0	0	0	0	1	-360	360;
	2264	8468	0.000771	0.0083	0	0	0	0	0	0	1	-360	360;
	3986	12363	0.00449	0.023349	0	0	0	0	0	0	1	-360	360;
	935	10379	0.001	0.313037	0	0	0	0	0	0	1	-360	360;
	935	12951	0.025331	0.364938	0	0	0	0	0	0	1	-360	360;
	935	12951	0.004219	0.018211	0	0	0	0	0	0	1	-360	360;
	10379	12951	0.02581	0.357169	0	0	0	0	0	0	1	-360	360;
	10379	12951	0.00413	0.01686	0	0	0	0	0	0	1	-360	360;
	722	3393	0.00018	0.00157	0	0	0	0	0	0	1	-360	360;
	478	722	4.622922	11.33102	0	0	0	0	0	0	1	-360	360;
	8529	722	1.337465	6.217936	0	0	0	0	0	0	1	-360	360;
	5951	722	1.319252	5.765305	0	0	0	0	0	0	1	-360	360;
	11650	5812	0.00017	0.0016	0	0	0	0	0	0	1	-360	360;
	11650	6355	0.00017	0.00163	0	0	0	0	0	0	1	-360	360;
	11650	5809	2e-05	0.00022	0	0	0	0	0	0	1	-360	360;
	10991	9023	0	0.0002	0	0	0	0	0	0	1	-360	360;
	10991	9023	0.01018	0.46657	0	0	0	0	0	0	1	-360	360;
	8960	10991	0.00887	0.046021	0	0	0	0	0	0	1	-360	360;
	6196	10991	0.007479	0.02551	0	0	0	0	0	0	1	-360	360;
	2313	9023	0.00762	0.024979	0	0	0	0	0	0	1	-360	360;
	10899	3393	0.00034	0.00277	0	0	0	0	0	0	1	-360	360;
	2749	5812	0.00035	0.00271	0	0	0	0	0	0	1	-360	360;
	6598	3605	0.00146	0.01134	0	0	0	0	0	0	1	-360	360;
	4342	1684	0	0.0002	0	0	0	0	0	0	1	-360	360;
	4342	1684	0.001	0.250041	0	0	0	0	0	0	1	-360	360;
	4342	12614	0.00031	0.00232	0	0	0	0	0	0	1	-360	360;
	4860	8960	0	0.0002	0	0	0	0	0	0	1	-360	360;
	4860	8960	0.001	0.319463	0	0	0	0	0	0	1	-360	360;
	4860	12951	0.002479	0.011479	0	0	0	0	0	0	1	-360	360;
	4860	1779	1.778616	5.164256	0	0	0	0	0	0	1	-360	360;
	3119	4860	0.00644	0.026229	0	0	0	0	0	0	1	-360	360;
	6012	4860	0.342769	1.509876	0	0	0	0	0	0	1	-360	360;
	8960	12951	0.002521	0.01124	0	0	0	0	0	0	1	-360	360;
	8960	1779	1.29219	4.678099	0	0	0	0	0	0	1	-360	360;
	6012	8960	0.236529	1.420661	0	0	0	0	0	0	1	-360	360;
	11448	2359	0	0.0002	0	0	0	0	0	0	1	-360	360;
	11448	2359	0.05505	0.67124	0	0	0	0	0	0	1	-360	360;
	366	11448	4.467975	11.21446	0	0	0	0	0	0	1	-360	360;
	11448	8156	34.72521	52.45248	0	0	0	0	0	0	1	-360	360;
	11448	6030	36.34711	50.63223	0	0	0	0	0	0	1	-360	360;
	11448	917	22.76033	33.1281	0	0	0	0	0	0	1	-360	360;
	11448	4995	21.48554	36.70661	0	0	0	0	0	0	1	-360	360;
	11448	12407	39.50826	49.23967	0	0	0	0	0	0	1	-360	360;
	6896	11448	17.04917	42.27893	0	0	0	0	0	0	1	-360	360;
	10979	11448	16.04855	40.82025	0	0	0	0	0	0	1	-360	360;
	11448	1535	0.053771	1.208946	0	0	0	0	0	0	1	-360	360;
	11448	1535	0.003281	0.016341	0	0	0	0	0	0	1	-360	360;
	5183	11448	0.825227	5.014463	0	0	0	0	0	0	1	-360	360;
	11448	7014	4.278926	11.33285	0	0	0	0	0	0	1	-360	360;
	11448	12720	3.46157	11.98905	0	0	0	0	0	0	1	-360	360;
	7916	11448	4.532025	17.77417	0	0	0	0	0	0	1	-360	360;
	7916	11448	0.009169	0.03831	0	0	0	0	0	0	1	-360	360;
	3743	11448	4.582645	17.99917	0	0	0	0	0	0	1	-360	360;
	11448	9401	4.507231	15.28285	0	0	0	0	0	0	1	-360	360;
	11448	2256	0.001	1.211157	0	0	0	0	0	0	1	-360	360;
	11448	7691	7.098967	16.7219	0	0	0	0	0	0	1	-360	360;
	11448	6065	6.249587	14.57107	0	0	0	0	0	0	1	-360	360;
	366	2359	3.945041	12.27665	0	0	0	0	0	0	1	-360	360;
	2359	8156	34.19215	51.58471	0	0	0	0	0	0	1	-360	360;
	2359	6030	31.52686	50.73347	0	0	0	0	0	0	1	-360	360;
	2359	917	18.00289	35.73554	0	0	0	0	0	0	1	-360	360;
	2359	4995	21.65083	34.30165	0	0	0	0	0	0	1	-360	360;
	2359	12407	35.85331	52.29959	0	0	0	0	0	0	1	-360	360;
	6896	2359	18.05847	51.41322	0	0	0	0	0	0	1	-360	360;
	10979	2359	21.28512	43.97521	0	0	0	0	0	0	1	-360	360;
	2359	1535	0.036771	0.970744	0	0	0	0	0	0	1	-360	360;
	5183	2359	0.78843	6.091736	0	0	0	0	0	0	1	-360	360;
	2359	7014	3.228512	12.74483	0	0	0	0	0	0	1	-360	360;
	2359	12720	3.551653	11.03347	0	0	0	0	0	0	1	-360	360;
	7916	2359	5.941736	18.63595	0	0	0	0	0	0	1	-360	360;
	3743	2359	4.915083	20.96901	0	0	0	0	0	0	1	-360	360;
	3743	2359	0.01039	0.037539	0	0	0	0	0	0	1	-360	360;
	2359	9401	4.096281	15.0905	0	0	0	0	0	0	1	-360	360;
	2359	2256	0.001	1.352211	0	0	0	0	0	0	1	-360	360;
	2359	7691	6.635537	16.54979	0	0	0	0	0	0	1	-360	360;
	2359	6065	5.82438	14.40847	0	0	0	0	0	0	1	-360	360;
	5951	3019	0.00046	0.00466	0	0	0	0	0	0	1	-360	360;
	8529	478	0.03625	0.46376	0	0	0	0	0	0	1	-360	360;
	8529	478	0.00045	0.00475	0	0	0	0	0	0	1	-360	360;
	5951	478	0.04062	0.4091	0	0	0	0	0	0	1	-360	360;
	11448	2256	0.006271	0.0373	0	0	0	0	0	0	1	-360	360;
	945	8045	0.000331	0.00331	0	0	0	0	0	0	1	-360	360;
	945	6386	0.000341	0.00324	0	0	0	0	0	0	1	-360	360;
	945	2113	0.00013	0.00132	0	0	0	0	0	0	1	-360	360;
	8608	4055	0.001	0.1802	0	0	0	0	0	0	1	-360	360;
	8608	8156	0.04812	0.726715	0	0	0	0	0	0	1	-360	360;
	8608	8156	0.005169	0.036401	0	0	0	0	0	0	1	-360	360;
	8608	6030	0.04006	0.78624	0	0	0	0	0	0	1	-360	360;
	917	8608	0.559318	2.376033	0	0	0	0	0	0	1	-360	360;
	4995	8608	0.574607	2.433264	0	0	0	0	0	0	1	-360	360;
	8608	12407	0.449463	2.310744	0	0	0	0	0	0	1	-360	360;
	12043	8608	0.18062	1.392789	0	0	0	0	0	0	1	-360	360;
	9823	8608	0.092211	0.750041	0	0	0	0	0	0	1	-360	360;
	6896	8608	6.405372	12.55372	0	0	0	0	0	0	1	-360	360;
	10979	8608	5.734504	14.28161	0	0	0	0	0	0	1	-360	360;
	5183	8608	15.87066	21.82231	0	0	0	0	0	0	1	-360	360;
	7916	8608	45.78719	60.88223	0	0	0	0	0	0	1	-360	360;
	3743	8608	52.13017	54.55165	0	0	0	0	0	0	1	-360	360;
	4717	8608	0.615	2.63781	0	0	0	0	0	0	1	-360	360;
	8608	2256	16.26798	23.5124	0	0	0	0	0	0	1	-360	360;
	8608	7691	0.818636	3.652686	0	0	0	0	0	0	1	-360	360;
	8608	6065	0.670434	2.968182	0	0	0	0	0	0	1	-360	360;
	4055	8156	0.041711	0.912376	0	0	0	0	0	0	1	-360	360;
	4055	6030	0.04611	0.822479	0	0	0	0	0	0	1	-360	360;
	4055	6030	0.005271	0.03561	0	0	0	0	0	0	1	-360	360;
	917	4055	0.605434	2.534091	0	0	0	0	0	0	1	-360	360;
	4995	4055	0.542851	2.99876	0	0	0	0	0	0	1	-360	360;
	4055	12407	0.520909	2.569628	0	0	0	0	0	0	1	-360	360;
	12043	4055	0.177789	1.458264	0	0	0	0	0	0	1	-360	360;
	9823	4055	0.09881	0.801281	0	0	0	0	0	0	1	-360	360;
	6896	4055	6.124793	15.47397	0	0	0	0	0	0	1	-360	360;
	10979	4055	5.91157	16.17149	0	0	0	0	0	0	1	-360	360;
	5183	4055	14.54876	25.34091	0	0	0	0	0	0	1	-360	360;
	7916	4055	54.67355	60.11777	0	0	0	0	0	0	1	-360	360;
	4717	4055	0.802355	2.58657	0	0	0	0	0	0	1	-360	360;
	4055	2256	19.94897	25.18388	0	0	0	0	0	0	1	-360	360;
	4055	7691	1.027479	3.393802	0	0	0	0	0	0	1	-360	360;
	4055	6065	0.843471	3.042975	0	0	0	0	0	0	1	-360	360;
	13017	10967	0.00078	0.00782	0	0	0	0	0	0	1	-360	360;
	13017	6550	7.575485	11.77562	0	0	0	0	0	0	1	-360	360;
	13017	964	1.23705	3.353947	0	0	0	0	0	0	1	-360	360;
	13017	13282	3.429501	6.941136	0	0	0	0	0	0	1	-360	360;
	13017	2182	0.00076	0.00797	0	0	0	0	0	0	1	-360	360;
	12151	2182	0.00025	0.00233	0	0	0	0	0	0	1	-360	360;
	3291	10967	0.00179	0.01397	0	0	0	0	0	0	1	-360	360;
	11732	91	0.173643	1.472438	0	0	0	0	0	0	1	-360	360;
	91	12274	5.182133	18.94529	0	0	0	0	0	0	1	-360	360;
	91	3895	0.188989	1.114889	0	0	0	0	0	0	1	-360	360;
	1789	91	1.860596	8.193906	0	0	0	0	0	0	1	-360	360;
	366	469	0.00424	0.029211	0	0	0	0	0	0	1	-360	360;
	366	1535	6.176859	17.47149	0	0	0	0	0	0	1	-360	360;
	5183	366	10.52665	27.33264	0	0	0	0	0	0	1	-360	360;
	366	7014	0.202	1.377087	0	0	0	0	0	0	1	-360	360;
	366	12720	0.19876	1.343988	0	0	0	0	0	0	1	-360	360;
	366	12720	0.00545	0.042341	0	0	0	0	0	0	1	-360	360;
	7916	366	7.300207	16.26798	0	0	0	0	0	0	1	-360	360;
	3743	366	6.816736	17.84773	0	0	0	0	0	0	1	-360	360;
	366	9401	3.997727	8.729958	0	0	0	0	0	0	1	-360	360;
	366	2256	15.84711	34.44008	0	0	0	0	0	0	1	-360	360;
	366	1566	30.61777	53.27893	0	0	0	0	0	0	1	-360	360;
	366	5754	36.28512	62.88017	0	0	0	0	0	0	1	-360	360;
	91	3895	0.00082	0.01007	0	0	0	0	0	0	1	-360	360;
	10917	11296	0.0005	0.004169	0	0	0	0	0	0	1	-360	360;
	7916	10917	0.00525	0.033151	0	0	0	0	0	0	1	-360	360;
	11296	5666	0.001969	0.01718	0	0	0	0	0	0	1	-360	360;
	5666	6288	0.001289	0.009229	0	0	0	0	0	0	1	-360	360;
	6288	5792	0.00043	0.00349	0	0	0	0	0	0	1	-360	360;
	5792	4871	0.000979	0.008901	0	0	0	0	0	0	1	-360	360;
	4871	469	0.000219	0.00155	0	0	0	0	0	0	1	-360	360;
	10288	10446	0.00025	0.00319	0	0	0	0	0	0	1	-360	360;
	10288	10446	0.00024	0.00295	0	0	0	0	0	0	1	-360	360;
	6550	1061	0.00022	0.00171	0	0	0	0	0	0	1	-360	360;
	10047	1061	0.0005	0.00382	0	0	0	0	0	0	1	-360	360;
	1747	3657	0.0002	0.00192	0	0	0	0	0	0	1	-360	360;
	10047	3657	0.0005	0.00383	0	0	0	0	0	0	1	-360	360;
	8156	6030	0	0.0002	0	0	0	0	0	0	1	-360	360;
	8156	6030	0.001	0.268678	0	0	0	0	0	0	1	-360	360;
	917	8156	0.042159	0.956198	0	0	0	0	0	0	1	-360	360;
	917	8156	0.002341	0.022341	0	0	0	0	0	0	1	-360	360;
	4995	8156	0.04555	0.93969	0	0	0	0	0	0	1	-360	360;
	8156	12407	0.077909	1.08686	0	0	0	0	0	0	1	-360	360;
	6495	8156	0.0003	0.003289	0	0	0	0	0	0	1	-360	360;
	12043	8156	1.527831	6.098554	0	0	0	0	0	0	1	-360	360;
	9823	8156	0.958926	2.68595	0	0	0	0	0	0	1	-360	360;
	6896	8156	28.84298	48.1157	0	0	0	0	0	0	1	-360	360;
	10979	8156	33.55372	46.46901	0	0	0	0	0	0	1	-360	360;
	1535	8156	47.90702	63.74587	0	0	0	0	0	0	1	-360	360;
	5183	8156	28.68802	30.77066	0	0	0	0	0	0	1	-360	360;
	4717	8156	4.526653	11.10372	0	0	0	0	0	0	1	-360	360;
	8156	2256	6.754959	11.13905	0	0	0	0	0	0	1	-360	360;
	8156	7691	0.105229	1.236012	0	0	0	0	0	0	1	-360	360;
	8156	6065	0.085781	1.020186	0	0	0	0	0	0	1	-360	360;
	917	6030	0.044289	0.919483	0	0	0	0	0	0	1	-360	360;
	4995	6030	0.047781	0.901963	0	0	0	0	0	0	1	-360	360;
	6030	12407	0.076341	1.006446	0	0	0	0	0	0	1	-360	360;
	9067	6030	0.0003	0.00336	0	0	0	0	0	0	1	-360	360;
	12043	6030	1.621322	5.733678	0	0	0	0	0	0	1	-360	360;
	9823	6030	0.903202	2.852273	0	0	0	0	0	0	1	-360	360;
	6896	6030	30.65289	49.86157	0	0	0	0	0	0	1	-360	360;
	10979	6030	32.78719	47.26653	0	0	0	0	0	0	1	-360	360;
	1535	6030	44.17355	62.70041	0	0	0	0	0	0	1	-360	360;
	5183	6030	25.07438	35.58471	0	0	0	0	0	0	1	-360	360;
	4717	6030	5.060744	9.753098	0	0	0	0	0	0	1	-360	360;
	6030	2256	6.015083	12.60103	0	0	0	0	0	0	1	-360	360;
	6030	7691	0.099409	1.193017	0	0	0	0	0	0	1	-360	360;
	6030	7691	0.002021	0.01812	0	0	0	0	0	0	1	-360	360;
	6030	6065	0.09157	0.961405	0	0	0	0	0	0	1	-360	360;
	6825	8391	0.00032	0.00387	0	0	0	0	0	0	1	-360	360;
	6825	5257	0.00038	0.00378	0	0	0	0	0	0	1	-360	360;
	5250	9863	0.00032	0.00394	0	0	0	0	0	0	1	-360	360;
	5250	887	0.00038	0.0037	0	0	0	0	0	0	1	-360	360;
	6825	10957	6.036911	10.78324	0	0	0	0	0	0	1	-360	360;
	9492	10446	0.0008	0.00945	0	0	0	0	0	0	1	-360	360;
	10446	909	0.00079	0.00828	0	0	0	0	0	0	1	-360	360;
	9492	10446	0.00078	0.00964	0	0	0	0	0	0	1	-360	360;
	10446	909	0.00081	0.00811	0	0	0	0	0	0	1	-360	360;
	917	4995	0	0.0002	0	0	0	0	0	0	1	-360	360;
	917	4995	0.001	0.26281	0	0	0	0	0	0	1	-360	360;
	917	12407	0.001	0.461798	0	0	0	0	0	0	1	-360	360;
	917	12407	0.00164	0.020729	0	0	0	0	0	0	1	-360	360;
	12043	917	7.774793	17.76901	0	0	0	0	0	0	1	-360	360;
	9823	917	3.756405	9.209917	0	0	0	0	0	0	1	-360	360;
	917	1535	27.82645	45.00826	0	0	0	0	0	0	1	-360	360;
	5183	917	20.96901	31.09711	0	0	0	0	0	0	1	-360	360;
	4717	917	18.12293	35.08471	0	0	0	0	0	0	1	-360	360;
	917	2256	3.360124	8.657025	0	0	0	0	0	0	1	-360	360;
	917	7691	0.14569	1.294938	0	0	0	0	0	0	1	-360	360;
	917	6065	0.125919	1.049607	0	0	0	0	0	0	1	-360	360;
	4995	12407	0.001	0.471508	0	0	0	0	0	0	1	-360	360;
	4995	12407	0.001599	0.02112	0	0	0	0	0	0	1	-360	360;
	12043	4995	8.624174	16.76653	0	0	0	0	0	0	1	-360	360;
	9823	4995	4.365909	9.241321	0	0	0	0	0	0	1	-360	360;
	4995	1535	29.57231	44.15909	0	0	0	0	0	0	1	-360	360;
	5183	4995	22.67562	29.81818	0	0	0	0	0	0	1	-360	360;
	4717	4995	22.21281	29.92975	0	0	0	0	0	0	1	-360	360;
	4995	2256	3.288017	9.179959	0	0	0	0	0	0	1	-360	360;
	4995	7691	0.15644	1.276653	0	0	0	0	0	0	1	-360	360;
	4995	6065	0.124669	1.119298	0	0	0	0	0	0	1	-360	360;
	12043	12407	6.748554	16.31901	0	0	0	0	0	0	1	-360	360;
	9823	12407	3.853099	7.976446	0	0	0	0	0	0	1	-360	360;
	1535	12407	53.92562	59.04132	0	0	0	0	0	0	1	-360	360;
	5183	12407	37.32851	39.04752	0	0	0	0	0	0	1	-360	360;
	4717	12407	16.76198	30.60331	0	0	0	0	0	0	1	-360	360;
	12407	2256	6.070868	12.11942	0	0	0	0	0	0	1	-360	360;
	7691	12407	0.29624	1.60312	0	0	0	0	0	0	1	-360	360;
	6065	12407	0.293884	1.293946	0	0	0	0	0	0	1	-360	360;
	11732	12274	0.15633	3.578947	0	0	0	0	0	0	1	-360	360;
	11732	3895	0.242957	1.518213	0	0	0	0	0	0	1	-360	360;
	11732	1789	8.069251	24.53047	0	0	0	0	0	0	1	-360	360;
	11732	9866	0.00029	0.00357	0	0	0	0	0	0	1	-360	360;
	3498	7835	0.00012	0.00131	0	0	0	0	0	0	1	-360	360;
	12043	9823	0	0.0002	0	0	0	0	0	0	1	-360	360;
	12043	9823	0.001	0.168909	0	0	0	0	0	0	1	-360	360;
	12043	6896	0.837231	3.56405	0	0	0	0	0	0	1	-360	360;
	12043	10979	0.690682	3.987397	0	0	0	0	0	0	1	-360	360;
	12043	10979	0.00988	0.05469	0	0	0	0	0	0	1	-360	360;
	5183	12043	5.606818	11.56529	0	0	0	0	0	0	1	-360	360;
	7916	12043	15.6624	27.83058	0	0	0	0	0	0	1	-360	360;
	3743	12043	16.76116	26.42355	0	0	0	0	0	0	1	-360	360;
	12043	4717	0.039521	0.57	0	0	0	0	0	0	1	-360	360;
	12043	2256	25.64463	32.3843	0	0	0	0	0	0	1	-360	360;
	12043	7691	12.49401	21.31612	0	0	0	0	0	0	1	-360	360;
	12043	6065	10.90826	18.38058	0	0	0	0	0	0	1	-360	360;
	9823	6896	0.367355	2.263636	0	0	0	0	0	0	1	-360	360;
	9823	10979	0.394566	2.152479	0	0	0	0	0	0	1	-360	360;
	5183	9823	2.95062	6.117354	0	0	0	0	0	0	1	-360	360;
	7916	9823	8.41281	14.42025	0	0	0	0	0	0	1	-360	360;
	3743	9823	8.672314	14.29339	0	0	0	0	0	0	1	-360	360;
	9823	4717	0.0192	0.321384	0	0	0	0	0	0	1	-360	360;
	9823	2256	12.78037	18.27831	0	0	0	0	0	0	1	-360	360;
	9823	7691	6.830371	10.80145	0	0	0	0	0	0	1	-360	360;
	9823	6065	5.092149	10.9593	0	0	0	0	0	0	1	-360	360;
	7108	10349	0.00043	0.00522	0	0	0	0	0	0	1	-360	360;
	1789	10349	0.00184	0.01667	0	0	0	0	0	0	1	-360	360;
	1789	10349	0.248089	2.761427	0	0	0	0	0	0	1	-360	360;
	7245	7922	0.00037	0.00464	0	0	0	0	0	0	1	-360	360;
	7245	8379	0.00059	0.00694	0	0	0	0	0	0	1	-360	360;
	12406	12866	0.00027	0.00348	0	0	0	0	0	0	1	-360	360;
	8314	12866	0.00027	0.00354	0	0	0	0	0	0	1	-360	360;
	12406	8314	0.02114	0.839889	0	0	0	0	0	0	1	-360	360;
	9085	12406	30.32756	43.46053	0	0	0	0	0	0	1	-360	360;
	9085	8314	30.89958	42.53601	0	0	0	0	0	0	1	-360	360;
	6896	10979	0	0.0002	0	0	0	0	0	0	1	-360	360;
	6896	10979	0.001	0.238595	0	0	0	0	0	0	1	-360	360;
	5183	6896	0.256632	1.964752	0	0	0	0	0	0	1	-360	360;
	7916	6896	0.63343	3.574793	0	0	0	0	0	0	1	-360	360;
	3743	6896	0.709752	3.284711	0	0	0	0	0	0	1	-360	360;
	3743	6896	0.01324	0.090909	0	0	0	0	0	0	1	-360	360;
	4717	6896	0.000591	1.201426	0	0	0	0	0	0	1	-360	360;
	6896	2256	1.821963	7.541736	0	0	0	0	0	0	1	-360	360;
	5183	10979	0.210826	2.191942	0	0	0	0	0	0	1	-360	360;
	7916	10979	0.746074	3.056198	0	0	0	0	0	0	1	-360	360;
	3743	10979	0.740041	3.16219	0	0	0	0	0	0	1	-360	360;
	4717	10979	0.000841	1.29095	0	0	0	0	0	0	1	-360	360;
	10979	2256	2.154545	7.118182	0	0	0	0	0	0	1	-360	360;
	7922	2780	8e-05	0.00105	0	0	0	0	0	0	1	-360	360;
	12129	12274	0.00025	0.00186	0	0	0	0	0	0	1	-360	360;
	1535	8158	9.1e-05	0.000581	0	0	0	0	0	0	1	-360	360;
	11616	12274	0.00023	0.00211	0	0	0	0	0	0	1	-360	360;
	6949	11804	8e-05	0.00107	0	0	0	0	0	0	1	-360	360;
	6949	12274	2e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	3477	2478	0.00063	0.00406	0	0	0	0	0	0	1	-360	360;
	3477	1535	9.1e-05	0.00056	0	0	0	0	0	0	1	-360	360;
	13127	4828	0.00064	0.003969	0	0	0	0	0	0	1	-360	360;
	1535	13127	9.1e-05	0.00055	0	0	0	0	0	0	1	-360	360;
	12274	2780	2e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	11804	7124	0.00091	0.01097	0	0	0	0	0	0	1	-360	360;
	2478	3613	0.00711	0.045841	0	0	0	0	0	0	1	-360	360;
	4828	12870	0.00724	0.044849	0	0	0	0	0	0	1	-360	360;
	3895	12274	5.952838	17.29294	0	0	0	0	0	0	1	-360	360;
	1789	12274	20.08726	50.82479	0	0	0	0	0	0	1	-360	360;
	5183	1535	4.890702	15.41012	0	0	0	0	0	0	1	-360	360;
	7014	1535	5.639256	15.40868	0	0	0	0	0	0	1	-360	360;
	12720	1535	4.76281	15.7	0	0	0	0	0	0	1	-360	360;
	7916	1535	21.2624	53.66529	0	0	0	0	0	0	1	-360	360;
	3743	1535	25.16942	46.14463	0	0	0	0	0	0	1	-360	360;
	1535	9401	5.023554	17.83409	0	0	0	0	0	0	1	-360	360;
	1535	2256	0.483492	3.46095	0	0	0	0	0	0	1	-360	360;
	166	1535	0.000211	0.0012	0	0	0	0	0	0	1	-360	360;
	1535	7691	9.888017	20.17066	0	0	0	0	0	0	1	-360	360;
	1535	6065	9.581818	15.8126	0	0	0	0	0	0	1	-360	360;
	2123	1522	0.00015	0.00151	0	0	0	0	0	0	1	-360	360;
	7124	7022	0.00044	0.00443	0	0	0	0	0	0	1	-360	360;
	3613	13629	0.00301	0.01868	0	0	0	0	0	0	1	-360	360;
	12870	13629	0.00268	0.02118	0	0	0	0	0	0	1	-360	360;
	8379	1522	0.00044	0.00433	0	0	0	0	0	0	1	-360	360;
	7022	2123	0.00015	0.00148	0	0	0	0	0	0	1	-360	360;
	5183	7014	36.67975	64.23967	0	0	0	0	0	0	1	-360	360;
	5183	12720	34.68595	57.82851	0	0	0	0	0	0	1	-360	360;
	7916	5183	0.001	0.997107	0	0	0	0	0	0	1	-360	360;
	3743	5183	0.001	0.990558	0	0	0	0	0	0	1	-360	360;
	3743	5183	0.00389	0.03206	0	0	0	0	0	0	1	-360	360;
	5183	4717	2.394421	10.2281	0	0	0	0	0	0	1	-360	360;
	5183	2256	0.001	0.942996	0	0	0	0	0	0	1	-360	360;
	5183	7691	7.162397	16.07149	0	0	0	0	0	0	1	-360	360;
	5183	6065	6.970248	12.65785	0	0	0	0	0	0	1	-360	360;
	6656	9032	0.00041	0.00489	0	0	0	0	0	0	1	-360	360;
	12987	9032	0.00035	0.00415	0	0	0	0	0	0	1	-360	360;
	3895	6447	0.00041	0.00478	0	0	0	0	0	0	1	-360	360;
	12987	6447	0.00036	0.00406	0	0	0	0	0	0	1	-360	360;
	6656	10903	0.00069	0.00736	0	0	0	0	0	0	1	-360	360;
	9866	6656	0.00035	0.00355	0	0	0	0	0	0	1	-360	360;
	3504	6656	0.0003	0.0037	0	0	0	0	0	0	1	-360	360;
	3895	8234	0.00079	0.00639	0	0	0	0	0	0	1	-360	360;
	1789	3895	10.65028	23.35249	0	0	0	0	0	0	1	-360	360;
	7211	3895	0.00035	0.00375	0	0	0	0	0	0	1	-360	360;
	919	3895	0.00035	0.00368	0	0	0	0	0	0	1	-360	360;
	7014	12720	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7014	12720	0.001	0.307397	0	0	0	0	0	0	1	-360	360;
	7916	7014	30.07851	44.4938	0	0	0	0	0	0	1	-360	360;
	3743	7014	28.13843	48.89876	0	0	0	0	0	0	1	-360	360;
	7014	9401	1.021095	4.339256	0	0	0	0	0	0	1	-360	360;
	7014	2256	19.04773	44.46694	0	0	0	0	0	0	1	-360	360;
	7014	1566	11.71198	25.83058	0	0	0	0	0	0	1	-360	360;
	7014	5754	13.4062	28.03306	0	0	0	0	0	0	1	-360	360;
	7916	12720	25.77686	44.3843	0	0	0	0	0	0	1	-360	360;
	3743	12720	29.27686	39.66322	0	0	0	0	0	0	1	-360	360;
	12720	9401	0.90343	4.251446	0	0	0	0	0	0	1	-360	360;
	12720	2256	15.67913	46.04752	0	0	0	0	0	0	1	-360	360;
	12720	1566	10.06777	25.78512	0	0	0	0	0	0	1	-360	360;
	12720	5754	11.7936	27.51653	0	0	0	0	0	0	1	-360	360;
	6933	8234	0.00037	0.00319	0	0	0	0	0	0	1	-360	360;
	4218	10903	0.00037	0.0032	0	0	0	0	0	0	1	-360	360;
	4762	13432	6e-05	0.00057	0	0	0	0	0	0	1	-360	360;
	13282	4762	0.00081	0.00793	0	0	0	0	0	0	1	-360	360;
	7916	3743	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7916	3743	0.001	0.218595	0	0	0	0	0	0	1	-360	360;
	7916	4717	5.508884	16.92169	0	0	0	0	0	0	1	-360	360;
	7916	2256	0.417335	3.096901	0	0	0	0	0	0	1	-360	360;
	7916	7691	35.56405	45.20868	0	0	0	0	0	0	1	-360	360;
	7916	6065	26.29545	47.22107	0	0	0	0	0	0	1	-360	360;
	3743	4717	5.134091	18.53492	0	0	0	0	0	0	1	-360	360;
	3743	2256	0.373512	3.536364	0	0	0	0	0	0	1	-360	360;
	3743	7691	31.99174	51.7624	0	0	0	0	0	0	1	-360	360;
	3743	6065	26.61157	47.82645	0	0	0	0	0	0	1	-360	360;
	9401	2256	22.02479	61.94007	0	0	0	0	0	0	1	-360	360;
	1566	9401	0.20395	1.135248	0	0	0	0	0	0	1	-360	360;
	5754	9401	0.207314	1.339256	0	0	0	0	0	0	1	-360	360;
	4717	2256	15.82955	28.76446	0	0	0	0	0	0	1	-360	360;
	4717	7691	27.39876	40.44215	0	0	0	0	0	0	1	-360	360;
	4717	6065	23.59711	35.69835	0	0	0	0	0	0	1	-360	360;
	7691	2256	1.210248	3.369008	0	0	0	0	0	0	1	-360	360;
	6065	2256	0.914421	3.375826	0	0	0	0	0	0	1	-360	360;
	6196	1779	0.005729	0.02032	0	0	0	0	0	0	1	-360	360;
	2313	1779	0.005841	0.019909	0	0	0	0	0	0	1	-360	360;
	6012	1779	0.084781	0.493471	0	0	0	0	0	0	1	-360	360;
	8529	5951	0	0.0002	0	0	0	0	0	0	1	-360	360;
	8529	5951	0.001	0.19356	0	0	0	0	0	0	1	-360	360;
	8529	12058	0.00192	0.01356	0	0	0	0	0	0	1	-360	360;
	5951	12058	0.00172	0.01537	0	0	0	0	0	0	1	-360	360;
	2741	6180	0.00035	0.00346	0	0	0	0	0	0	1	-360	360;
	1566	5754	0	0.0002	0	0	0	0	0	0	1	-360	360;
	1566	5754	0.001	0.159719	0	0	0	0	0	0	1	-360	360;
	1566	8815	0.000341	0.00199	0	0	0	0	0	0	1	-360	360;
	8815	12039	0.00219	0.012901	0	0	0	0	0	0	1	-360	360;
	12039	6478	0.002531	0.013539	0	0	0	0	0	0	1	-360	360;
	6478	10573	0.002281	0.01263	0	0	0	0	0	0	1	-360	360;
	10573	166	0.00232	0.01393	0	0	0	0	0	0	1	-360	360;
	1747	2897	0.00043	0.00418	0	0	0	0	0	0	1	-360	360;
	964	6550	0.113102	0.936981	0	0	0	0	0	0	1	-360	360;
	6550	1670	0.00043	0.00419	0	0	0	0	0	0	1	-360	360;
	13282	6550	0.124889	1.006302	0	0	0	0	0	0	1	-360	360;
	3746	12363	0	0.0002	0	0	0	0	0	0	1	-360	360;
	3746	12363	0.003651	0.01288	0	0	0	0	0	0	1	-360	360;
	12363	5547	0.004021	0.01951	0	0	0	0	0	0	1	-360	360;
	964	1651	0.00057	0.00402	0	0	0	0	0	0	1	-360	360;
	5939	13282	0.00064	0.00448	0	0	0	0	0	0	1	-360	360;
	964	13282	0.00646	0.473407	0	0	0	0	0	0	1	-360	360;
	4756	2897	8e-05	0.00061	0	0	0	0	0	0	1	-360	360;
	1128	1670	6e-05	0.00065	0	0	0	0	0	0	1	-360	360;
	6012	6196	0.00012	0.000461	0	0	0	0	0	0	1	-360	360;
	6012	2313	0.00013	0.00045	0	0	0	0	0	0	1	-360	360;
	7691	6065	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7691	6065	0.001	0.249814	0	0	0	0	0	0	1	-360	360;
	12684	3043	0.0001	0.00075	0	0	0	0	0	0	1	-360	360;
	12684	10965	0.0038	0.11597	0	0	0	0	0	0	1	-360	360;
	12684	10965	0.0001	0.00081	0	0	0	0	0	0	1	-360	360;
	9948	3043	0.0001	0.00094	0	0	0	0	0	0	1	-360	360;
	9948	3043	9e-05	0.00107	0	0	0	0	0	0	1	-360	360;
	89	9948	7e-05	0.00085	0	0	0	0	0	0	1	-360	360;
	89	9948	8e-05	0.00081	0	0	0	0	0	0	1	-360	360;
	3043	10965	0.00018	0.00171	0	0	0	0	0	0	1	-360	360;
	89	3635	0.00017	0.00191	0	0	0	0	0	0	1	-360	360;
	89	3635	0.00016	0.00194	0	0	0	0	0	0	1	-360	360;
	89	11659	0.0005	0.00317	0	0	0	0	0	0	1	-360	360;
	89	11659	0.00048	0.00322	0	0	0	0	0	0	1	-360	360;
	10965	6385	0.00427	0.19732	0	0	0	0	0	0	1	-360	360;
	10965	6385	0.0001	0.00122	0	0	0	0	0	0	1	-360	360;
	10965	6385	0.0001	0.00119	0	0	0	0	0	0	1	-360	360;
	6385	6595	0.0024	0.02048	0	0	0	0	0	0	1	-360	360;
	6385	284	0.00042	0.00474	0	0	0	0	0	0	1	-360	360;
	5297	2962	0	0.0002	0	0	0	0	0	0	1	-360	360;
	5297	2962	0.003409	0.129831	0	0	0	0	0	0	1	-360	360;
	5297	313	0.00605	0.024031	0	0	0	0	0	0	1	-360	360;
	2962	313	0.005919	0.0245	0	0	0	0	0	0	1	-360	360;
	4127	9057	0.406364	1.478988	0	0	0	0	0	0	1	-360	360;
	4127	587	0.12912	0.95312	0	0	0	0	0	0	1	-360	360;
	4127	7275	0.23062	0.90281	0	0	0	0	0	0	1	-360	360;
	1611	4127	0.05107	0.366901	0	0	0	0	0	0	1	-360	360;
	4127	587	0.00207	0.0168	0	0	0	0	0	0	1	-360	360;
	4127	2319	0.00119	0.009659	0	0	0	0	0	0	1	-360	360;
	3879	4127	0.120151	0.700041	0	0	0	0	0	0	1	-360	360;
	4127	8543	0.032581	0.326756	0	0	0	0	0	0	1	-360	360;
	5470	8061	2e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	5470	1355	0.00025	0.00344	0	0	0	0	0	0	1	-360	360;
	5470	1355	0.01671	0.288767	0	0	0	0	0	0	1	-360	360;
	5470	9499	0.01954	0.281191	0	0	0	0	0	0	1	-360	360;
	9156	5470	0.00356	0.222701	0	0	0	0	0	0	1	-360	360;
	9156	5470	0.00271	0.199571	0	0	0	0	0	0	1	-360	360;
	6804	8061	0.00016	0.00253	0	0	0	0	0	0	1	-360	360;
	9156	8061	0.00018	0.00214	0	0	0	0	0	0	1	-360	360;
	9057	587	0.002091	0.0138	0	0	0	0	0	0	1	-360	360;
	9057	587	0.291467	1.202521	0	0	0	0	0	0	1	-360	360;
	7275	9057	0.204719	0.794277	0	0	0	0	0	0	1	-360	360;
	1611	9057	0.026401	0.214814	0	0	0	0	0	0	1	-360	360;
	1611	9057	0.00295	0.01256	0	0	0	0	0	0	1	-360	360;
	6301	11453	0.359525	1.423657	0	0	0	0	0	0	1	-360	360;
	3085	11453	0.32907	1.339607	0	0	0	0	0	0	1	-360	360;
	3085	11453	0.00469	0.03913	0	0	0	0	0	0	1	-360	360;
	3817	11453	0.000841	0.00544	0	0	0	0	0	0	1	-360	360;
	3817	11453	0.014581	0.16557	0	0	0	0	0	0	1	-360	360;
	7035	11453	1.038884	1.650227	0	0	0	0	0	0	1	-360	360;
	11123	11453	0.863926	1.96155	0	0	0	0	0	0	1	-360	360;
	80	587	0.005081	0.02426	0	0	0	0	0	0	1	-360	360;
	2314	80	0.470062	1.363636	0	0	0	0	0	0	1	-360	360;
	80	12360	0.035969	0.35219	0	0	0	0	0	0	1	-360	360;
	80	13555	0.0033	0.01864	0	0	0	0	0	0	1	-360	360;
	80	13555	0.018021	0.1825	0	0	0	0	0	0	1	-360	360;
	7437	1170	0.002651	0.027659	0	0	0	0	0	0	1	-360	360;
	7456	7437	0.00226	0.04061	0	0	0	0	0	0	1	-360	360;
	7437	12202	0.000711	0.0098	0	0	0	0	0	0	1	-360	360;
	3278	7437	0.002211	0.037599	0	0	0	0	0	0	1	-360	360;
	6046	3462	0.0016	0.01356	0	0	0	0	0	0	1	-360	360;
	6046	3462	0.00165	0.01328	0	0	0	0	0	0	1	-360	360;
	9156	1355	0.01336	0.25865	0	0	0	0	0	0	1	-360	360;
	9156	1355	0.00015	0.00253	0	0	0	0	0	0	1	-360	360;
	10358	2376	0.00225	0.02106	0	0	0	0	0	0	1	-360	360;
	6207	2376	0.00426	0.03506	0	0	0	0	0	0	1	-360	360;
	9898	2376	0.00043	0.00581	0	0	0	0	0	0	1	-360	360;
	7275	5802	0.001151	0.009909	0	0	0	0	0	0	1	-360	360;
	1611	5802	0.000961	0.00751	0	0	0	0	0	0	1	-360	360;
	1209	5802	0.00032	0.00324	0	0	0	0	0	0	1	-360	360;
	10358	2376	0.00245	0.02191	0	0	0	0	0	0	1	-360	360;
	1620	2376	0.00037	0.00606	0	0	0	0	0	0	1	-360	360;
	10196	2376	0.0029	0.0289	0	0	0	0	0	0	1	-360	360;
	10308	587	0.000151	0.00156	0	0	0	0	0	0	1	-360	360;
	13436	587	0.00014	0.00144	0	0	0	0	0	0	1	-360	360;
	7275	587	0.026841	0.406942	0	0	0	0	0	0	1	-360	360;
	1611	587	0.025909	0.304649	0	0	0	0	0	0	1	-360	360;
	3772	1177	0	0.0002	0	0	0	0	0	0	1	-360	360;
	3772	1177	0.001	0.18678	0	0	0	0	0	0	1	-360	360;
	7892	3772	0.989958	2.675	0	0	0	0	0	0	1	-360	360;
	5412	3772	0.994529	2.704017	0	0	0	0	0	0	1	-360	360;
	6108	3772	0.154938	0.902078	0	0	0	0	0	0	1	-360	360;
	4839	3772	0.141247	1.002008	0	0	0	0	0	0	1	-360	360;
	13476	3772	0.01709	0.303539	0	0	0	0	0	0	1	-360	360;
	3772	10879	0.00081	0.00826	0	0	0	0	0	0	1	-360	360;
	3772	6046	0.00174	0.01199	0	0	0	0	0	0	1	-360	360;
	3772	9048	0.00136	0.012	0	0	0	0	0	0	1	-360	360;
	7892	1177	1.135249	2.306648	0	0	0	0	0	0	1	-360	360;
	5412	1177	1.077147	2.483726	0	0	0	0	0	0	1	-360	360;
	6108	1177	0.140319	0.988089	0	0	0	0	0	0	1	-360	360;
	4839	1177	0.15108	0.940582	0	0	0	0	0	0	1	-360	360;
	13476	1177	0.00045	0.00442	0	0	0	0	0	0	1	-360	360;
	13476	1177	0.01578	0.332528	0	0	0	0	0	0	1	-360	360;
	6916	1177	0.00032	0.00308	0	0	0	0	0	0	1	-360	360;
	487	6804	0.00018	0.0025	0	0	0	0	0	0	1	-360	360;
	4172	6804	0.03152	0.366053	0	0	0	0	0	0	1	-360	360;
	2314	5524	0.00251	0.13469	0	0	0	0	0	0	1	-360	360;
	2314	6301	0.00943	0.04787	0	0	0	0	0	0	1	-360	360;
	2314	12360	0.002781	0.02494	0	0	0	0	0	0	1	-360	360;
	2314	12360	0.22876	0.857376	0	0	0	0	0	0	1	-360	360;
	2314	13555	0.113521	0.454752	0	0	0	0	0	0	1	-360	360;
	10358	11151	0.00159	0.01603	0	0	0	0	0	0	1	-360	360;
	5524	3085	0.00924	0.04881	0	0	0	0	0	0	1	-360	360;
	5524	13555	0.00643	0.027271	0	0	0	0	0	0	1	-360	360;
	10358	6207	0.00183	0.01368	0	0	0	0	0	0	1	-360	360;
	12064	7805	0	0.0002	0	0	0	0	0	0	1	-360	360;
	12064	7805	0.001	0.191392	0	0	0	0	0	0	1	-360	360;
	4740	12064	0.00081	0.00981	0	0	0	0	0	0	1	-360	360;
	12064	7244	0.090907	0.571413	0	0	0	0	0	0	1	-360	360;
	12064	7244	0.00058	0.00647	0	0	0	0	0	0	1	-360	360;
	12064	13259	0.441572	1.595429	0	0	0	0	0	0	1	-360	360;
	12064	7096	0.0463	0.493158	0	0	0	0	0	0	1	-360	360;
	12064	8694	0.053	0.456427	0	0	0	0	0	0	1	-360	360;
	8694	7805	0.00048	0.00707	0	0	0	0	0	0	1	-360	360;
	7805	7244	0.0007	0.00626	0	0	0	0	0	0	1	-360	360;
	7805	7244	0.099273	0.491122	0	0	0	0	0	0	1	-360	360;
	7805	13259	0.411558	1.466205	0	0	0	0	0	0	1	-360	360;
	7805	7096	0.05226	0.428179	0	0	0	0	0	0	1	-360	360;
	7805	8694	0.04979	0.473137	0	0	0	0	0	0	1	-360	360;
	10503	11644	0.00018	0.00218	0	0	0	0	0	0	1	-360	360;
	5633	1025	0.344619	1.353116	0	0	0	0	0	0	1	-360	360;
	5633	3417	0.581787	1.764958	0	0	0	0	0	0	1	-360	360;
	5633	11636	0.075831	0.835526	0	0	0	0	0	0	1	-360	360;
	5633	11636	0.00095	0.0105	0	0	0	0	0	0	1	-360	360;
	5633	2948	0.081281	0.782202	0	0	0	0	0	0	1	-360	360;
	5633	10879	0.06719	0.554377	0	0	0	0	0	0	1	-360	360;
	5633	9048	0.00409	0.271641	0	0	0	0	0	0	1	-360	360;
	5633	9048	0.00041	0.00351	0	0	0	0	0	0	1	-360	360;
	11698	949	0.00138	0.008789	0	0	0	0	0	0	1	-360	360;
	11698	4819	0.002331	0.01832	0	0	0	0	0	0	1	-360	360;
	1025	3417	0	0.0002	0	0	0	0	0	0	1	-360	360;
	1025	3417	0.001	0.311717	0	0	0	0	0	0	1	-360	360;
	2265	1025	0.00043	0.00408	0	0	0	0	0	0	1	-360	360;
	1025	11636	0.083899	0.525478	0	0	0	0	0	0	1	-360	360;
	1025	2948	0.07124	0.605921	0	0	0	0	0	0	1	-360	360;
	1025	11644	0.00013	0.00147	0	0	0	0	0	0	1	-360	360;
	1025	10879	2.055194	4.798753	0	0	0	0	0	0	1	-360	360;
	13259	1025	0.00048	0.00587	0	0	0	0	0	0	1	-360	360;
	9048	1025	0.81115	2.560526	0	0	0	0	0	0	1	-360	360;
	1432	11698	0.00095	0.004169	0	0	0	0	0	0	1	-360	360;
	3636	11698	0.01012	0.04631	0	0	0	0	0	0	1	-360	360;
	4328	3417	0.00013	0.0014	0	0	0	0	0	0	1	-360	360;
	1691	3417	0.00013	0.00143	0	0	0	0	0	0	1	-360	360;
	3417	11636	0.091447	0.726385	0	0	0	0	0	0	1	-360	360;
	3417	11636	0.00093	0.00869	0	0	0	0	0	0	1	-360	360;
	3417	2948	0.089307	0.727078	0	0	0	0	0	0	1	-360	360;
	3417	10879	3.590651	5.504778	0	0	0	0	0	0	1	-360	360;
	4172	3417	0.00047	0.00544	0	0	0	0	0	0	1	-360	360;
	9048	3417	1.2741	3.747645	0	0	0	0	0	0	1	-360	360;
	2354	6180	0.00142	0.01566	0	0	0	0	0	0	1	-360	360;
	2948	9744	0.00082	0.00794	0	0	0	0	0	0	1	-360	360;
	7892	5412	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7892	5412	0.001	0.149882	0	0	0	0	0	0	1	-360	360;
	6108	7892	0.088033	0.626579	0	0	0	0	0	0	1	-360	360;
	6108	7892	0.001	0.00866	0	0	0	0	0	0	1	-360	360;
	4839	7892	0.090852	0.619418	0	0	0	0	0	0	1	-360	360;
	13476	7892	0.87133	2.390028	0	0	0	0	0	0	1	-360	360;
	6108	5412	0.087632	0.633012	0	0	0	0	0	0	1	-360	360;
	4839	5412	0.097832	0.57669	0	0	0	0	0	0	1	-360	360;
	13476	5412	0.857064	2.463089	0	0	0	0	0	0	1	-360	360;
	5412	6046	0.0004	0.0039	0	0	0	0	0	0	1	-360	360;
	6301	3085	0.001	0.486508	0	0	0	0	0	0	1	-360	360;
	3817	6301	0.003909	0.034711	0	0	0	0	0	0	1	-360	360;
	3817	6301	0.497789	2.055971	0	0	0	0	0	0	1	-360	360;
	6301	7035	0.357087	1.677376	0	0	0	0	0	0	1	-360	360;
	6301	11123	0.349194	1.694876	0	0	0	0	0	0	1	-360	360;
	3817	3085	0.438409	2.013802	0	0	0	0	0	0	1	-360	360;
	3085	7035	0.307107	1.517831	0	0	0	0	0	0	1	-360	360;
	3085	11123	0.354174	1.443988	0	0	0	0	0	0	1	-360	360;
	12733	3576	0.00274	0.144159	0	0	0	0	0	0	1	-360	360;
	12733	2600	0.00039	0.00175	0	0	0	0	0	0	1	-360	360;
	3576	13196	0.000401	0.001711	0	0	0	0	0	0	1	-360	360;
	2354	7451	0.06537	0.54912	0	0	0	0	0	0	1	-360	360;
	2354	2265	0.00098	0.00852	0	0	0	0	0	0	1	-360	360;
	2354	2265	0.076039	0.582729	0	0	0	0	0	0	1	-360	360;
	2354	2843	0.00185	0.01418	0	0	0	0	0	0	1	-360	360;
	12109	2354	0.790374	2.170014	0	0	0	0	0	0	1	-360	360;
	2393	949	0.00188	0.0087	0	0	0	0	0	0	1	-360	360;
	2393	5190	0.042461	0.437066	0	0	0	0	0	0	1	-360	360;
	3742	5190	0.00068	0.008531	0	0	0	0	0	0	1	-360	360;
	5633	11636	0.00107	0.01108	0	0	0	0	0	0	1	-360	360;
	5633	10879	0.00092	0.00947	0	0	0	0	0	0	1	-360	360;
	6108	4839	0	0.0002	0	0	0	0	0	0	1	-360	360;
	6108	4839	0.001	0.196378	0	0	0	0	0	0	1	-360	360;
	6108	13476	0.102652	0.915512	0	0	0	0	0	0	1	-360	360;
	4839	13476	0.110602	0.871537	0	0	0	0	0	0	1	-360	360;
	8836	4839	0.00153	0.01371	0	0	0	0	0	0	1	-360	360;
	13171	4839	0.00296	0.02271	0	0	0	0	0	0	1	-360	360;
	7451	2265	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7451	2265	0.001	0.229661	0	0	0	0	0	0	1	-360	360;
	7451	2843	0.00076	0.0066	0	0	0	0	0	0	1	-360	360;
	12109	7451	0.134051	0.802355	0	0	0	0	0	0	1	-360	360;
	12109	2265	0.136413	0.960942	0	0	0	0	0	0	1	-360	360;
	12566	13196	0.0027	0.02195	0	0	0	0	0	0	1	-360	360;
	12566	4819	0.001211	0.004961	0	0	0	0	0	0	1	-360	360;
	3038	10308	0.010091	0.211612	0	0	0	0	0	0	1	-360	360;
	3038	13436	0.008349	0.215682	0	0	0	0	0	0	1	-360	360;
	1611	3038	0.000401	0.001781	0	0	0	0	0	0	1	-360	360;
	10308	13436	0.00076	0.205461	0	0	0	0	0	0	1	-360	360;
	13259	9898	0.00086	0.0138	0	0	0	0	0	0	1	-360	360;
	4172	1620	0.00101	0.01174	0	0	0	0	0	0	1	-360	360;
	11636	2948	0	0.0002	0	0	0	0	0	0	1	-360	360;
	11636	2948	0.001	0.17286	0	0	0	0	0	0	1	-360	360;
	11636	10879	0.917036	2.770706	0	0	0	0	0	0	1	-360	360;
	9048	11636	0.299162	1.508587	0	0	0	0	0	0	1	-360	360;
	2948	10879	0.973546	2.445291	0	0	0	0	0	0	1	-360	360;
	9048	2948	0.335817	1.383795	0	0	0	0	0	0	1	-360	360;
	1131	7035	0.00114	0.00625	0	0	0	0	0	0	1	-360	360;
	7341	1131	0.00069	0.00343	0	0	0	0	0	0	1	-360	360;
	10504	1131	0.000669	0.00349	0	0	0	0	0	0	1	-360	360;
	7609	1131	0.00264	0.01612	0	0	0	0	0	0	1	-360	360;
	4740	6207	0.001	0.01303	0	0	0	0	0	0	1	-360	360;
	1340	6207	0.00018	0.00112	0	0	0	0	0	0	1	-360	360;
	8777	1910	0.000771	0.0027	0	0	0	0	0	0	1	-360	360;
	1688	1910	0.000719	0.003039	0	0	0	0	0	0	1	-360	360;
	1910	11123	0.00138	0.005409	0	0	0	0	0	0	1	-360	360;
	2620	11151	0.00017	0.0011	0	0	0	0	0	0	1	-360	360;
	1611	7275	0.01424	0.207211	0	0	0	0	0	0	1	-360	360;
	12360	7275	0.00288	0.023031	0	0	0	0	0	0	1	-360	360;
	13476	9156	0.00059	0.00738	0	0	0	0	0	0	1	-360	360;
	6114	7096	2e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	4740	9661	0.00034	0.00243	0	0	0	0	0	0	1	-360	360;
	9156	9499	0.00939	0.239481	0	0	0	0	0	0	1	-360	360;
	9156	9499	0.0003	0.00385	0	0	0	0	0	0	1	-360	360;
	9194	12202	0.00082	0.0117	0	0	0	0	0	0	1	-360	360;
	9194	1170	0.000831	0.010711	0	0	0	0	0	0	1	-360	360;
	6796	3026	0	0.0002	0	0	0	0	0	0	1	-360	360;
	6796	3026	0.001	0.424958	0	0	0	0	0	0	1	-360	360;
	6796	4172	0.00035	0.00315	0	0	0	0	0	0	1	-360	360;
	6796	4172	0.01362	0.396288	0	0	0	0	0	0	1	-360	360;
	6796	9156	0.088213	0.739543	0	0	0	0	0	0	1	-360	360;
	3026	4172	0.01388	0.388199	0	0	0	0	0	0	1	-360	360;
	3026	9156	0.0012	0.01148	0	0	0	0	0	0	1	-360	360;
	3026	9156	0.086413	0.754086	0	0	0	0	0	0	1	-360	360;
	487	13259	0.00049	0.00475	0	0	0	0	0	0	1	-360	360;
	584	8189	0.002669	0.01443	0	0	0	0	0	0	1	-360	360;
	5190	949	0.00338	0.01488	0	0	0	0	0	0	1	-360	360;
	3879	13196	0.00101	0.005841	0	0	0	0	0	0	1	-360	360;
	9898	2843	0.00034	0.00381	0	0	0	0	0	0	1	-360	360;
	12109	2843	0.00017	0.00223	0	0	0	0	0	0	1	-360	360;
	8543	2600	0.00161	0.013789	0	0	0	0	0	0	1	-360	360;
	2843	1894	0.00032	0.00234	0	0	0	0	0	0	1	-360	360;
	9048	10879	0.01246	0.249882	0	0	0	0	0	0	1	-360	360;
	7244	4172	0.00073	0.00838	0	0	0	0	0	0	1	-360	360;
	7244	13259	0.01019	0.13124	0	0	0	0	0	0	1	-360	360;
	7244	13259	0.00078	0.00873	0	0	0	0	0	0	1	-360	360;
	7244	7096	0.463179	1.173823	0	0	0	0	0	0	1	-360	360;
	7244	8694	0.428629	1.321537	0	0	0	0	0	0	1	-360	360;
	9156	4172	0.099591	0.647687	0	0	0	0	0	0	1	-360	360;
	1820	4172	0.00019	0.00245	0	0	0	0	0	0	1	-360	360;
	13259	7096	1.591828	3.383587	0	0	0	0	0	0	1	-360	360;
	13259	8694	1.558587	3.254017	0	0	0	0	0	0	1	-360	360;
	3817	7035	1.410971	2.609298	0	0	0	0	0	0	1	-360	360;
	3817	11123	1.456178	2.474793	0	0	0	0	0	0	1	-360	360;
	7035	11123	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7035	11123	0.001	0.139341	0	0	0	0	0	0	1	-360	360;
	7096	8694	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7096	8694	0.001	0.196801	0	0	0	0	0	0	1	-360	360;
	11387	2319	0.00025	0.00232	0	0	0	0	0	0	1	-360	360;
	8543	11387	0.000719	0.005461	0	0	0	0	0	0	1	-360	360;
	2319	9426	0.00012	0.00069	0	0	0	0	0	0	1	-360	360;
	9571	2319	0.003099	0.11999	0	0	0	0	0	0	1	-360	360;
	12360	13555	0	0.0002	0	0	0	0	0	0	1	-360	360;
	12360	13555	0.001	0.13145	0	0	0	0	0	0	1	-360	360;
	3879	8543	0.10799	0.554256	0	0	0	0	0	0	1	-360	360;
	12271	12622	0.00077	0.00927	0	0	0	0	0	0	1	-360	360;
	9567	12622	0.00078	0.00908	0	0	0	0	0	0	1	-360	360;
	7123	6414	0.00038	0.0045	0	0	0	0	0	0	1	-360	360;
	6414	9450	0.00147	0.0125	0	0	0	0	0	0	1	-360	360;
	9745	6414	0.00039	0.00441	0	0	0	0	0	0	1	-360	360;
	6414	1759	0.00202	0.01654	0	0	0	0	0	0	1	-360	360;
	9567	12271	0	0.0002	0	0	0	0	0	0	1	-360	360;
	9567	12271	0.00147	0.112341	0	0	0	0	0	0	1	-360	360;
	9567	7123	0.00069	0.00564	0	0	0	0	0	0	1	-360	360;
	12271	9745	0.00068	0.00574	0	0	0	0	0	0	1	-360	360;
	7123	9745	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7123	9745	0.00088	0.07777	0	0	0	0	0	0	1	-360	360;
	866	7123	0.00062	0.00515	0	0	0	0	0	0	1	-360	360;
	3429	9745	0.0007	0.00505	0	0	0	0	0	0	1	-360	360;
	866	3429	0	0.0002	0	0	0	0	0	0	1	-360	360;
	866	3429	0.00076	0.06481	0	0	0	0	0	0	1	-360	360;
	4472	673	0.00011	0.00149	0	0	0	0	0	0	1	-360	360;
	8588	2750	9.9e-05	0.000979	0	0	0	0	0	0	1	-360	360;
	7095	7494	0.01788	0.227831	0	0	0	0	0	0	1	-360	360;
	7095	7494	0.004669	0.015969	0	0	0	0	0	0	1	-360	360;
	9978	7095	0.011229	0.03868	0	0	0	0	0	0	1	-360	360;
	3460	7095	0.01001	0.043841	0	0	0	0	0	0	1	-360	360;
	7095	700	0.002151	0.01032	0	0	0	0	0	0	1	-360	360;
	6245	8592	0.003281	0.017901	0	0	0	0	0	0	1	-360	360;
	3497	6245	0.005271	0.026909	0	0	0	0	0	0	1	-360	360;
	6245	12457	0.00044	0.002169	0	0	0	0	0	0	1	-360	360;
	6245	12457	0.00045	0.00213	0	0	0	0	0	0	1	-360	360;
	4556	9091	0.452251	2.097438	0	0	0	0	0	0	1	-360	360;
	9091	1723	0.247659	1.380679	0	0	0	0	0	0	1	-360	360;
	1771	9091	0.436378	1.290097	0	0	0	0	0	0	1	-360	360;
	7818	9091	0.218089	1.220914	0	0	0	0	0	0	1	-360	360;
	7012	9091	0.0612	1.024169	0	0	0	0	0	0	1	-360	360;
	8784	9091	0.06168	0.959765	0	0	0	0	0	0	1	-360	360;
	8784	9091	0.0011	0.00838	0	0	0	0	0	0	1	-360	360;
	10047	9091	0.05459	0.962258	0	0	0	0	0	0	1	-360	360;
	7818	9091	0.00126	0.01278	0	0	0	0	0	0	1	-360	360;
	7472	10836	5e-05	0.000229	0	0	0	0	0	0	1	-360	360;
	6735	7472	0.001331	0.009331	0	0	0	0	0	0	1	-360	360;
	7472	13600	6e-05	0.00031	0	0	0	0	0	0	1	-360	360;
	6735	9394	0.00155	0.00805	0	0	0	0	0	0	1	-360	360;
	9394	228	0.351674	1.414483	0	0	0	0	0	0	1	-360	360;
	6735	9394	0.239545	1.416095	0	0	0	0	0	0	1	-360	360;
	7624	2750	0.00011	0.00088	0	0	0	0	0	0	1	-360	360;
	2312	673	0.00012	0.00064	0	0	0	0	0	0	1	-360	360;
	12086	673	0.002349	0.01088	0	0	0	0	0	0	1	-360	360;
	12086	673	0.0023	0.011099	0	0	0	0	0	0	1	-360	360;
	673	1209	0.004771	0.03713	0	0	0	0	0	0	1	-360	360;
	11874	2750	0.00056	0.0013	0	0	0	0	0	0	1	-360	360;
	10015	2750	0.001091	0.007409	0	0	0	0	0	0	1	-360	360;
	10015	2750	0.00168	0.007479	0	0	0	0	0	0	1	-360	360;
	2027	2750	0.003151	0.01876	0	0	0	0	0	0	1	-360	360;
	1664	3793	0.005719	0.255764	0	0	0	0	0	0	1	-360	360;
	568	1664	0.000599	0.00294	0	0	0	0	0	0	1	-360	360;
	568	3793	0.00062	0.00288	0	0	0	0	0	0	1	-360	360;
	8260	7011	0.001169	0.012591	0	0	0	0	0	0	1	-360	360;
	8260	7011	0.00131	0.01119	0	0	0	0	0	0	1	-360	360;
	2150	7011	0.000159	0.00137	0	0	0	0	0	0	1	-360	360;
	6036	7011	0.000159	0.001341	0	0	0	0	0	0	1	-360	360;
	12560	11092	0.00051	0.0025	0	0	0	0	0	0	1	-360	360;
	6561	10372	0.00031	0.001289	0	0	0	0	0	0	1	-360	360;
	12560	10372	0.05163	0.55124	0	0	0	0	0	0	1	-360	360;
	11208	10372	0.201229	1.043864	0	0	0	0	0	0	1	-360	360;
	11358	10015	1.027128	2.385331	0	0	0	0	0	0	1	-360	360;
	11358	239	0.288306	1.382335	0	0	0	0	0	0	1	-360	360;
	10225	11358	0.05949	0.340661	0	0	0	0	0	0	1	-360	360;
	11358	3751	0.007211	0.030409	0	0	0	0	0	0	1	-360	360;
	12769	11358	0.899153	2.660744	0	0	0	0	0	0	1	-360	360;
	8118	11358	0.809979	2.92686	0	0	0	0	0	0	1	-360	360;
	11358	7766	0.00251	0.011211	0	0	0	0	0	0	1	-360	360;
	3636	7018	0.08695	0.512231	0	0	0	0	0	0	1	-360	360;
	3636	13016	0.006289	0.693946	0	0	0	0	0	0	1	-360	360;
	2365	3636	0.140669	0.495269	0	0	0	0	0	0	1	-360	360;
	9625	3636	0.00882	0.041831	0	0	0	0	0	0	1	-360	360;
	10196	2605	2e-05	0.00025	0	0	0	0	0	0	1	-360	360;
	7527	6511	0.000159	0.00074	0	0	0	0	0	0	1	-360	360;
	1425	9736	0.004469	0.139591	0	0	0	0	0	0	1	-360	360;
	1425	3452	0.001039	0.006211	0	0	0	0	0	0	1	-360	360;
	1950	13449	0.00813	0.186659	0	0	0	0	0	0	1	-360	360;
	1950	3452	0.00101	0.00625	0	0	0	0	0	0	1	-360	360;
	8659	831	8.1e-05	0.000781	0	0	0	0	0	0	1	-360	360;
	8659	10149	0.0003	0.002031	0	0	0	0	0	0	1	-360	360;
	7126	831	9.1e-05	0.000719	0	0	0	0	0	0	1	-360	360;
	7126	3731	0.00086	0.005581	0	0	0	0	0	0	1	-360	360;
	6564	1723	0.00062	0.00605	0	0	0	0	0	0	1	-360	360;
	3227	831	0.00181	0.009599	0	0	0	0	0	0	1	-360	360;
	9245	831	0.00099	0.00481	0	0	0	0	0	0	1	-360	360;
	831	6591	0.00038	0.00218	0	0	0	0	0	0	1	-360	360;
	5394	1983	0.00031	0.00342	0	0	0	0	0	0	1	-360	360;
	10400	1983	0.00083	0.00929	0	0	0	0	0	0	1	-360	360;
	4734	1983	0.00074	0.00793	0	0	0	0	0	0	1	-360	360;
	4506	4734	0.00016	0.00166	0	0	0	0	0	0	1	-360	360;
	1983	12051	0	0.0002	0	0	0	0	0	0	1	-360	360;
	1983	12051	0.01471	0.443012	0	0	0	0	0	0	1	-360	360;
	1872	12051	0.00037	0.00337	0	0	0	0	0	0	1	-360	360;
	9978	4168	0.001461	0.006711	0	0	0	0	0	0	1	-360	360;
	2638	4168	0.001479	0.00656	0	0	0	0	0	0	1	-360	360;
	10576	7018	0.00068	0.0113	0	0	0	0	0	0	1	-360	360;
	7155	4506	0.00019	0.00196	0	0	0	0	0	0	1	-360	360;
	11867	4506	0.00019	0.00198	0	0	0	0	0	0	1	-360	360;
	1872	7835	0.00886	0.294079	0	0	0	0	0	0	1	-360	360;
	1872	7835	0.00028	0.00275	0	0	0	0	0	0	1	-360	360;
	5394	7835	0.01321	0.296717	0	0	0	0	0	0	1	-360	360;
	12294	13042	0.001	0.169901	0	0	0	0	0	0	1	-360	360;
	12294	228	0.03111	0.255	0	0	0	0	0	0	1	-360	360;
	6735	12294	0.00057	0.002849	0	0	0	0	0	0	1	-360	360;
	13042	228	0.03143	0.234731	0	0	0	0	0	0	1	-360	360;
	6735	13042	0.000581	0.002789	0	0	0	0	0	0	1	-360	360;
	12698	7216	0.00157	0.007	0	0	0	0	0	0	1	-360	360;
	12698	996	0.00131	0.00636	0	0	0	0	0	0	1	-360	360;
	10914	7216	0.001669	0.00724	0	0	0	0	0	0	1	-360	360;
	7216	9396	0.00663	0.249339	0	0	0	0	0	0	1	-360	360;
	7216	1648	0.000169	0.00032	0	0	0	0	0	0	1	-360	360;
	7216	5806	0.000151	0.00038	0	0	0	0	0	0	1	-360	360;
	7216	3804	0.000841	0.009581	0	0	0	0	0	0	1	-360	360;
	7216	793	0.0355	0.360558	0	0	0	0	0	0	1	-360	360;
	7216	996	0.027539	0.325145	0	0	0	0	0	0	1	-360	360;
	7216	5661	0.025271	0.339442	0	0	0	0	0	0	1	-360	360;
	7216	8211	0.000169	0.000219	0	0	0	0	0	0	1	-360	360;
	3486	228	0.000401	0.0018	0	0	0	0	0	0	1	-360	360;
	6735	3486	0.00205	0.01162	0	0	0	0	0	0	1	-360	360;
	3881	3486	0.000651	0.0045	0	0	0	0	0	0	1	-360	360;
	4233	228	0.00038	0.00187	0	0	0	0	0	0	1	-360	360;
	6735	4233	0.002091	0.01139	0	0	0	0	0	0	1	-360	360;
	3881	4233	0.00064	0.004169	0	0	0	0	0	0	1	-360	360;
	6735	228	0.03551	0.421116	0	0	0	0	0	0	1	-360	360;
	6735	1246	0.090159	0.624236	0	0	0	0	0	0	1	-360	360;
	6735	10093	0.025919	0.457603	0	0	0	0	0	0	1	-360	360;
	6735	3748	0.140581	0.992087	0	0	0	0	0	0	1	-360	360;
	6735	4956	0.00082	0.007331	0	0	0	0	0	0	1	-360	360;
	12086	6981	0.008469	0.038599	0	0	0	0	0	0	1	-360	360;
	12086	9444	0.003729	0.012081	0	0	0	0	0	0	1	-360	360;
	10225	10015	0.071281	0.33186	0	0	0	0	0	0	1	-360	360;
	10225	10015	0.011919	0.055651	0	0	0	0	0	0	1	-360	360;
	10015	13635	0.0038	0.01182	0	0	0	0	0	0	1	-360	360;
	4506	1820	0.00233	0.0191	0	0	0	0	0	0	1	-360	360;
	4506	9278	0.00258	0.02125	0	0	0	0	0	0	1	-360	360;
	4506	1820	0.00246	0.01795	0	0	0	0	0	0	1	-360	360;
	4506	13547	0.00231	0.02199	0	0	0	0	0	0	1	-360	360;
	10400	4506	0.00105	0.0114	0	0	0	0	0	0	1	-360	360;
	4506	148	0.00268	0.02035	0	0	0	0	0	0	1	-360	360;
	3227	2839	0.00124	0.005711	0	0	0	0	0	0	1	-360	360;
	3227	13598	0.0002	0.0022	0	0	0	0	0	0	1	-360	360;
	3227	3982	0.00019	0.00225	0	0	0	0	0	0	1	-360	360;
	3227	10055	0.001169	0.005159	0	0	0	0	0	0	1	-360	360;
	3227	10045	0.001271	0.005591	0	0	0	0	0	0	1	-360	360;
	3227	4841	0.00182	0.01218	0	0	0	0	0	0	1	-360	360;
	3227	10493	0.00107	0.005789	0	0	0	0	0	0	1	-360	360;
	9468	118	0.00026	0.00275	0	0	0	0	0	0	1	-360	360;
	9468	1149	0.0001	0.00093	0	0	0	0	0	0	1	-360	360;
	13539	4244	0.00019	0.00166	0	0	0	0	0	0	1	-360	360;
	13539	4244	0.00019	0.00163	0	0	0	0	0	0	1	-360	360;
	5541	7990	0.00126	0.00953	0	0	0	0	0	0	1	-360	360;
	5541	11324	0.00131	0.00994	0	0	0	0	0	0	1	-360	360;
	2771	7990	0.00128	0.00933	0	0	0	0	0	0	1	-360	360;
	2771	13521	0.00134	0.00972	0	0	0	0	0	0	1	-360	360;
	5541	2771	0.001	0.180201	0	0	0	0	0	0	1	-360	360;
	5541	11324	1.365928	3.472299	0	0	0	0	0	0	1	-360	360;
	5541	13521	0.186177	1.029224	0	0	0	0	0	0	1	-360	360;
	992	5541	0.98608	2.33608	0	0	0	0	0	0	1	-360	360;
	5541	11314	0.487888	1.618975	0	0	0	0	0	0	1	-360	360;
	5541	6768	0.850277	2.290235	0	0	0	0	0	0	1	-360	360;
	2771	11324	1.29162	3.486565	0	0	0	0	0	0	1	-360	360;
	2771	13521	0.201801	0.8991	0	0	0	0	0	0	1	-360	360;
	992	2771	0.842729	2.35651	0	0	0	0	0	0	1	-360	360;
	2771	11314	0.415547	1.630194	0	0	0	0	0	0	1	-360	360;
	2771	6768	0.830956	1.995914	0	0	0	0	0	0	1	-360	360;
	700	13243	0.005729	0.021281	0	0	0	0	0	0	1	-360	360;
	700	13243	0.005081	0.029331	0	0	0	0	0	0	1	-360	360;
	10203	13243	0.00576	0.035781	0	0	0	0	0	0	1	-360	360;
	7494	13243	0.002021	0.00976	0	0	0	0	0	0	1	-360	360;
	4593	13243	0.00563	0.03313	0	0	0	0	0	0	1	-360	360;
	12918	5993	0.001	0.260568	0	0	0	0	0	0	1	-360	360;
	10087	12918	0.127098	1.298546	0	0	0	0	0	0	1	-360	360;
	8361	12918	0.04182	0.986219	0	0	0	0	0	0	1	-360	360;
	8361	12918	0.00129	0.01393	0	0	0	0	0	0	1	-360	360;
	3229	12918	0.00082	0.00915	0	0	0	0	0	0	1	-360	360;
	490	12918	0.00119	0.01287	0	0	0	0	0	0	1	-360	360;
	10087	5993	0.118677	1.418975	0	0	0	0	0	0	1	-360	360;
	10087	5993	0.00148	0.01585	0	0	0	0	0	0	1	-360	360;
	8361	5993	0.04574	0.913504	0	0	0	0	0	0	1	-360	360;
	680	5993	0.00116	0.01192	0	0	0	0	0	0	1	-360	360;
	10400	1649	0.06698	0.909972	0	0	0	0	0	0	1	-360	360;
	10400	1649	0.00042	0.00331	0	0	0	0	0	0	1	-360	360;
	1626	6080	0.002591	0.01299	0	0	0	0	0	0	1	-360	360;
	6080	3278	0.01169	0.067849	0	0	0	0	0	0	1	-360	360;
	10734	5459	0.00074	0.00368	0	0	0	0	0	0	1	-360	360;
	10734	239	0.00064	0.0042	0	0	0	0	0	0	1	-360	360;
	10734	7766	0.0018	0.01139	0	0	0	0	0	0	1	-360	360;
	1195	5459	0.00082	0.003401	0	0	0	0	0	0	1	-360	360;
	1195	239	0.00069	0.00381	0	0	0	0	0	0	1	-360	360;
	1195	7766	0.00189	0.01094	0	0	0	0	0	0	1	-360	360;
	5459	7573	0.000151	0.000711	0	0	0	0	0	0	1	-360	360;
	5459	1924	0.00013	0.000789	0	0	0	0	0	0	1	-360	360;
	5459	10749	5e-05	0.00024	0	0	0	0	0	0	1	-360	360;
	5459	1225	5e-05	0.00025	0	0	0	0	0	0	1	-360	360;
	12769	5459	0.00012	0.000711	0	0	0	0	0	0	1	-360	360;
	8118	5459	0.00011	0.000719	0	0	0	0	0	0	1	-360	360;
	959	6104	0.003091	0.015719	0	0	0	0	0	0	1	-360	360;
	6104	12768	0.018159	0.39124	0	0	0	0	0	0	1	-360	360;
	6104	9268	0.169419	0.883946	0	0	0	0	0	0	1	-360	360;
	6104	5095	0.656054	1.65126	0	0	0	0	0	0	1	-360	360;
	6104	7630	0.452913	1.939091	0	0	0	0	0	0	1	-360	360;
	6104	12875	0.055969	0.74907	0	0	0	0	0	0	1	-360	360;
	5550	6104	0.549483	1.897438	0	0	0	0	0	0	1	-360	360;
	5687	6104	0.483574	1.959731	0	0	0	0	0	0	1	-360	360;
	6104	11156	0.279731	1.818967	0	0	0	0	0	0	1	-360	360;
	6104	11156	0.006521	0.02749	0	0	0	0	0	0	1	-360	360;
	10904	8870	0.07687	1.255956	0	0	0	0	0	0	1	-360	360;
	10904	8870	0.00068	0.00577	0	0	0	0	0	0	1	-360	360;
	6561	10188	0.00812	0.12314	0	0	0	0	0	0	1	-360	360;
	22	6561	0.00157	0.007031	0	0	0	0	0	0	1	-360	360;
	5227	6561	0.001599	0.00688	0	0	0	0	0	0	1	-360	360;
	6561	12560	0.00113	0.00724	0	0	0	0	0	0	1	-360	360;
	6561	12560	0.001271	0.007091	0	0	0	0	0	0	1	-360	360;
	959	12875	0.000919	0.005211	0	0	0	0	0	0	1	-360	360;
	10914	9245	0.00199	0.010219	0	0	0	0	0	0	1	-360	360;
	10914	7486	0.005409	0.02475	0	0	0	0	0	0	1	-360	360;
	10914	1136	0.00482	0.02805	0	0	0	0	0	0	1	-360	360;
	10914	4749	0.028479	0.302273	0	0	0	0	0	0	1	-360	360;
	10914	9268	0.11506	0.606198	0	0	0	0	0	0	1	-360	360;
	10914	9245	0.025271	0.321983	0	0	0	0	0	0	1	-360	360;
	10914	8458	0.00194	0.010091	0	0	0	0	0	0	1	-360	360;
	10914	4749	0.00314	0.019719	0	0	0	0	0	0	1	-360	360;
	12475	10271	0.00042	0.00384	0	0	0	0	0	0	1	-360	360;
	12475	1149	0.00031	0.00229	0	0	0	0	0	0	1	-360	360;
	12475	1605	0.00024	0.00235	0	0	0	0	0	0	1	-360	360;
	12475	2728	0.00103	0.01072	0	0	0	0	0	0	1	-360	360;
	3503	12560	0.00275	0.01311	0	0	0	0	0	0	1	-360	360;
	3503	4305	0.001771	0.01225	0	0	0	0	0	0	1	-360	360;
	462	8592	0.001	0.385207	0	0	0	0	0	0	1	-360	360;
	462	8592	0.003349	0.022711	0	0	0	0	0	0	1	-360	360;
	462	7521	0.00032	0.001531	0	0	0	0	0	0	1	-360	360;
	462	8936	0.359525	1.472438	0	0	0	0	0	0	1	-360	360;
	462	7527	1.632355	4.693388	0	0	0	0	0	0	1	-360	360;
	462	11961	0.08494	0.651364	0	0	0	0	0	0	1	-360	360;
	462	3731	0.052581	0.671529	0	0	0	0	0	0	1	-360	360;
	12457	462	0.12794	1.157789	0	0	0	0	0	0	1	-360	360;
	996	462	0.002461	0.014531	0	0	0	0	0	0	1	-360	360;
	11205	9944	0.002659	0.013271	0	0	0	0	0	0	1	-360	360;
	11205	9944	0.002711	0.01299	0	0	0	0	0	0	1	-360	360;
	6855	7527	3.9e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	928	7527	5e-05	0.00031	0	0	0	0	0	0	1	-360	360;
	5214	10026	0.00089	0.00698	0	0	0	0	0	0	1	-360	360;
	5214	7818	0.00028	0.00271	0	0	0	0	0	0	1	-360	360;
	2427	4556	0.00025	0.00273	0	0	0	0	0	0	1	-360	360;
	2427	1723	0.00027	0.00238	0	0	0	0	0	0	1	-360	360;
	2427	7818	0.00027	0.00277	0	0	0	0	0	0	1	-360	360;
	4556	1723	0.00808	0.60428	0	0	0	0	0	0	1	-360	360;
	1771	4556	0.01086	0.386371	0	0	0	0	0	0	1	-360	360;
	4556	7818	0.03494	0.767659	0	0	0	0	0	0	1	-360	360;
	4556	7012	0.0265	0.810665	0	0	0	0	0	0	1	-360	360;
	8784	4556	0.524993	2.841898	0	0	0	0	0	0	1	-360	360;
	10047	4556	0.475568	2.62313	0	0	0	0	0	0	1	-360	360;
	12769	239	0.02113	0.408595	0	0	0	0	0	0	1	-360	360;
	8118	239	0.022469	0.384442	0	0	0	0	0	0	1	-360	360;
	13547	2201	0.00045	0.0038	0	0	0	0	0	0	1	-360	360;
	13547	2201	0.00047	0.00374	0	0	0	0	0	0	1	-360	360;
	12768	11310	0.00019	0.00086	0	0	0	0	0	0	1	-360	360;
	9268	12768	0.542397	1.40281	0	0	0	0	0	0	1	-360	360;
	5095	12768	1.299236	2.713017	0	0	0	0	0	0	1	-360	360;
	12768	7630	1.066591	3.481612	0	0	0	0	0	0	1	-360	360;
	12768	12875	0.276529	1.159153	0	0	0	0	0	0	1	-360	360;
	5550	12768	1.192211	3.154339	0	0	0	0	0	0	1	-360	360;
	5687	12768	1.210785	2.840289	0	0	0	0	0	0	1	-360	360;
	11156	12768	0.804979	2.837397	0	0	0	0	0	0	1	-360	360;
	1246	13179	9.1e-05	0.00055	0	0	0	0	0	0	1	-360	360;
	10093	1246	0.009909	0.377748	0	0	0	0	0	0	1	-360	360;
	1246	3748	0.914793	3.247727	0	0	0	0	0	0	1	-360	360;
	4372	13179	0.000349	0.002159	0	0	0	0	0	0	1	-360	360;
	10093	11310	0.00132	0.00624	0	0	0	0	0	0	1	-360	360;
	10093	13179	0.00088	0.005159	0	0	0	0	0	0	1	-360	360;
	8454	3748	0.001349	0.01051	0	0	0	0	0	0	1	-360	360;
	784	2886	0.000539	0.003099	0	0	0	0	0	0	1	-360	360;
	784	4956	0.00207	0.01182	0	0	0	0	0	0	1	-360	360;
	10188	12560	0.000539	0.003159	0	0	0	0	0	0	1	-360	360;
	10188	4305	0.004081	0.0247	0	0	0	0	0	0	1	-360	360;
	11136	6099	0.000531	0.003521	0	0	0	0	0	0	1	-360	360;
	11676	3799	0.000539	0.00345	0	0	0	0	0	0	1	-360	360;
	2365	11676	0.005169	0.02325	0	0	0	0	0	0	1	-360	360;
	11205	1136	0.000669	0.004099	0	0	0	0	0	0	1	-360	360;
	11205	8603	0.000599	0.00344	0	0	0	0	0	0	1	-360	360;
	3460	11205	0.00139	0.008919	0	0	0	0	0	0	1	-360	360;
	11205	7486	0.00069	0.00419	0	0	0	0	0	0	1	-360	360;
	2638	11205	0.001539	0.008099	0	0	0	0	0	0	1	-360	360;
	9268	3870	0.00062	0.00356	0	0	0	0	0	0	1	-360	360;
	3870	12875	0.001091	0.006479	0	0	0	0	0	0	1	-360	360;
	13249	10728	6e-05	0.00052	0	0	0	0	0	0	1	-360	360;
	13249	11674	0.00023	0.00485	0	0	0	0	0	0	1	-360	360;
	7818	10728	7e-05	0.00051	0	0	0	0	0	0	1	-360	360;
	10962	8592	5e-05	0.000229	0	0	0	0	0	0	1	-360	360;
	1865	8592	8.1e-05	0.0003	0	0	0	0	0	0	1	-360	360;
	6021	8592	6e-05	0.000341	0	0	0	0	0	0	1	-360	360;
	8936	8592	0.339112	2.39938	0	0	0	0	0	0	1	-360	360;
	7527	8592	0.982252	4.492975	0	0	0	0	0	0	1	-360	360;
	8592	11961	0.299979	2.042583	0	0	0	0	0	0	1	-360	360;
	8592	3731	0.216508	1.7125	0	0	0	0	0	0	1	-360	360;
	122	8592	0.002841	0.01551	0	0	0	0	0	0	1	-360	360;
	12457	8592	0.00326	1.128223	0	0	0	0	0	0	1	-360	360;
	6683	8592	0.00232	0.016961	0	0	0	0	0	0	1	-360	360;
	3153	8592	0.003979	0.02513	0	0	0	0	0	0	1	-360	360;
	10794	8592	0.002281	0.017289	0	0	0	0	0	0	1	-360	360;
	1771	1723	0.111143	0.652313	0	0	0	0	0	0	1	-360	360;
	23	1723	0.00045	0.00399	0	0	0	0	0	0	1	-360	360;
	7818	1723	0.01009	0.588213	0	0	0	0	0	0	1	-360	360;
	7012	1723	0.0029	0.60196	0	0	0	0	0	0	1	-360	360;
	8784	1723	0.468573	2.303532	0	0	0	0	0	0	1	-360	360;
	10047	1723	0.401032	2.262812	0	0	0	0	0	0	1	-360	360;
	7818	1723	0.00054	0.00451	0	0	0	0	0	0	1	-360	360;
	23	1723	0.0004	0.00439	0	0	0	0	0	0	1	-360	360;
	3382	3748	0.00132	0.0077	0	0	0	0	0	0	1	-360	360;
	3382	12875	0.002289	0.013909	0	0	0	0	0	0	1	-360	360;
	7330	4570	0.00015	0.00153	0	0	0	0	0	0	1	-360	360;
	7330	9013	0.00022	0.00174	0	0	0	0	0	0	1	-360	360;
	7330	13631	0.00031	0.00276	0	0	0	0	0	0	1	-360	360;
	10093	3748	0.498802	2.231818	0	0	0	0	0	0	1	-360	360;
	10093	12875	0.00243	0.01461	0	0	0	0	0	0	1	-360	360;
	10093	4956	0.0013	0.005919	0	0	0	0	0	0	1	-360	360;
	7527	8936	2.422314	7.164876	0	0	0	0	0	0	1	-360	360;
	7527	8936	0.005271	0.03874	0	0	0	0	0	0	1	-360	360;
	8936	11961	0.157781	1.324793	0	0	0	0	0	0	1	-360	360;
	8936	11961	0.00695	0.055591	0	0	0	0	0	0	1	-360	360;
	8936	3731	0.211777	1.52969	0	0	0	0	0	0	1	-360	360;
	12457	8936	0.1823	1.686529	0	0	0	0	0	0	1	-360	360;
	3081	6522	2.525346	8.326177	0	0	0	0	0	0	1	-360	360;
	3081	6522	0.00086	0.00871	0	0	0	0	0	0	1	-360	360;
	6522	2602	0.13018	1.288158	0	0	0	0	0	0	1	-360	360;
	6522	5257	5.25734	15.96122	0	0	0	0	0	0	1	-360	360;
	6522	10026	0.196787	1.606648	0	0	0	0	0	0	1	-360	360;
	12854	2602	0.00124	0.01282	0	0	0	0	0	0	1	-360	360;
	22	5227	0	0.0002	0	0	0	0	0	0	1	-360	360;
	22	5227	0.227149	1.244607	0	0	0	0	0	0	1	-360	360;
	22	2886	0.024581	0.402459	0	0	0	0	0	0	1	-360	360;
	3881	22	0.04388	0.657934	0	0	0	0	0	0	1	-360	360;
	10203	22	0.12525	0.756384	0	0	0	0	0	0	1	-360	360;
	4593	22	0.18987	0.937128	0	0	0	0	0	0	1	-360	360;
	22	12980	0.00163	0.007031	0	0	0	0	0	0	1	-360	360;
	5227	2886	0.456736	1.546157	0	0	0	0	0	0	1	-360	360;
	3881	5227	0.03062	0.528388	0	0	0	0	0	0	1	-360	360;
	10203	5227	0.192099	0.773678	0	0	0	0	0	0	1	-360	360;
	4593	5227	0.027081	0.45345	0	0	0	0	0	0	1	-360	360;
	4593	5227	0.002211	0.01299	0	0	0	0	0	0	1	-360	360;
	2977	8996	0.00344	0.015909	0	0	0	0	0	0	1	-360	360;
	1771	7818	0.123539	0.685277	0	0	0	0	0	0	1	-360	360;
	1771	7012	0.096593	0.575748	0	0	0	0	0	0	1	-360	360;
	1771	8784	0.225512	1.104709	0	0	0	0	0	0	1	-360	360;
	1771	10047	0.17135	1.111427	0	0	0	0	0	0	1	-360	360;
	1771	7818	0.00117	0.00907	0	0	0	0	0	0	1	-360	360;
	5686	2525	3e-05	0.00039	0	0	0	0	0	0	1	-360	360;
	909	5686	0.00079	0.0065	0	0	0	0	0	0	1	-360	360;
	6916	5686	0.00091	0.00755	0	0	0	0	0	0	1	-360	360;
	909	5686	0.00081	0.00636	0	0	0	0	0	0	1	-360	360;
	6916	5686	0.00086	0.00818	0	0	0	0	0	0	1	-360	360;
	3497	7527	0.003979	0.02364	0	0	0	0	0	0	1	-360	360;
	7527	11961	6.941322	16.0093	0	0	0	0	0	0	1	-360	360;
	7527	3731	6.73657	15.66095	0	0	0	0	0	0	1	-360	360;
	12457	7527	0.185229	2.220661	0	0	0	0	0	0	1	-360	360;
	3081	2602	5.786496	13.8054	0	0	0	0	0	0	1	-360	360;
	3081	5257	0.04375	0.648878	0	0	0	0	0	0	1	-360	360;
	3081	887	0.0008	0.00847	0	0	0	0	0	0	1	-360	360;
	3081	10026	0.19205	1.80741	0	0	0	0	0	0	1	-360	360;
	3081	10026	0.00157	0.01084	0	0	0	0	0	0	1	-360	360;
	6598	3081	0.00155	0.01203	0	0	0	0	0	0	1	-360	360;
	3081	6051	0.00062	0.00656	0	0	0	0	0	0	1	-360	360;
	3081	2602	0.00203	0.0219	0	0	0	0	0	0	1	-360	360;
	3081	11534	0.00017	0.00135	0	0	0	0	0	0	1	-360	360;
	7151	10045	0.000341	0.001729	0	0	0	0	0	0	1	-360	360;
	5095	7151	0.00101	0.00926	0	0	0	0	0	0	1	-360	360;
	3452	7151	0.00055	0.00262	0	0	0	0	0	0	1	-360	360;
	2839	10045	0.018781	0.265517	0	0	0	0	0	0	1	-360	360;
	7235	2839	0.0298	0.392211	0	0	0	0	0	0	1	-360	360;
	2839	3425	0.00057	0.002841	0	0	0	0	0	0	1	-360	360;
	3452	2839	0.00943	0.27814	0	0	0	0	0	0	1	-360	360;
	7235	10045	0.027961	0.338967	0	0	0	0	0	0	1	-360	360;
	10045	5195	0.000581	0.002781	0	0	0	0	0	0	1	-360	360;
	3452	10045	0.010831	0.239876	0	0	0	0	0	0	1	-360	360;
	7235	8228	0.00261	0.01243	0	0	0	0	0	0	1	-360	360;
	3819	9396	0.00214	0.01187	0	0	0	0	0	0	1	-360	360;
	793	9396	0.04931	0.648058	0	0	0	0	0	0	1	-360	360;
	793	9396	0.001711	0.010781	0	0	0	0	0	0	1	-360	360;
	9396	6496	0.00068	0.003271	0	0	0	0	0	0	1	-360	360;
	996	9396	0.07599	0.788884	0	0	0	0	0	0	1	-360	360;
	5661	9396	0.05989	0.967314	0	0	0	0	0	0	1	-360	360;
	10610	9396	0.00036	0.002719	0	0	0	0	0	0	1	-360	360;
	1179	1605	0.00022	0.00201	0	0	0	0	0	0	1	-360	360;
	8361	1179	0.04617	0.493712	0	0	0	0	0	0	1	-360	360;
	1179	10846	0.00031	0.00302	0	0	0	0	0	0	1	-360	360;
	10087	8361	0.001	0.254681	0	0	0	0	0	0	1	-360	360;
	10087	2728	0.00034	0.00288	0	0	0	0	0	0	1	-360	360;
	4030	13521	8e-05	0.00085	0	0	0	0	0	0	1	-360	360;
	4030	13521	8e-05	0.00087	0	0	0	0	0	0	1	-360	360;
	4030	13521	7e-05	0.00089	0	0	0	0	0	0	1	-360	360;
	4735	8996	0.808079	2.998967	0	0	0	0	0	0	1	-360	360;
	444	8996	0.11606	0.766157	0	0	0	0	0	0	1	-360	360;
	6259	8996	0.044669	0.42969	0	0	0	0	0	0	1	-360	360;
	6259	8996	0.0045	0.023841	0	0	0	0	0	0	1	-360	360;
	5946	3742	0.009969	0.07739	0	0	0	0	0	0	1	-360	360;
	7609	5946	0.01431	0.10638	0	0	0	0	0	0	1	-360	360;
	5946	7456	0.00774	0.06957	0	0	0	0	0	0	1	-360	360;
	185	3742	0.016669	0.26188	0	0	0	0	0	0	1	-360	360;
	13016	3742	0.009969	0.061409	0	0	0	0	0	0	1	-360	360;
	185	3742	0.001729	0.00995	0	0	0	0	0	0	1	-360	360;
	1820	8836	0.0015	0.01398	0	0	0	0	0	0	1	-360	360;
	6916	1820	0.00243	0.01942	0	0	0	0	0	0	1	-360	360;
	12543	6981	0.00455	0.02705	0	0	0	0	0	0	1	-360	360;
	7766	12543	0.004331	0.01676	0	0	0	0	0	0	1	-360	360;
	12543	2779	0.24907	0.891963	0	0	0	0	0	0	1	-360	360;
	12543	6981	0.033831	0.228099	0	0	0	0	0	0	1	-360	360;
	444	4735	0.344318	0.959132	0	0	0	0	0	0	1	-360	360;
	4735	6259	0.087211	0.561364	0	0	0	0	0	0	1	-360	360;
	4735	2128	0.003909	0.019531	0	0	0	0	0	0	1	-360	360;
	9641	1149	0.01509	0.413102	0	0	0	0	0	0	1	-360	360;
	9641	2791	9e-05	0.001	0	0	0	0	0	0	1	-360	360;
	9641	6117	0.00094	0.00943	0	0	0	0	0	0	1	-360	360;
	9641	6597	0.03361	0.58162	0	0	0	0	0	0	1	-360	360;
	6079	3229	0.0001	0.00107	0	0	0	0	0	0	1	-360	360;
	6079	148	0.00084	0.00882	0	0	0	0	0	0	1	-360	360;
	3497	444	0.00501	0.021599	0	0	0	0	0	0	1	-360	360;
	444	6259	0.04149	0.219607	0	0	0	0	0	0	1	-360	360;
	444	2536	0.000781	0.00744	0	0	0	0	0	0	1	-360	360;
	8677	11860	0.011789	0.277645	0	0	0	0	0	0	1	-360	360;
	8677	11860	0.0052	0.02574	0	0	0	0	0	0	1	-360	360;
	11860	9658	0.031669	0.507603	0	0	0	0	0	0	1	-360	360;
	11860	9658	0.00438	0.025461	0	0	0	0	0	0	1	-360	360;
	1626	11860	0.27186	0.993388	0	0	0	0	0	0	1	-360	360;
	118	1149	0.0037	0.327618	0	0	0	0	0	0	1	-360	360;
	1149	6597	0.01026	0.269931	0	0	0	0	0	0	1	-360	360;
	568	870	0.000281	0.0023	0	0	0	0	0	0	1	-360	360;
	11136	1431	0.000539	0.005021	0	0	0	0	0	0	1	-360	360;
	11136	7018	0.251054	1.054483	0	0	0	0	0	0	1	-360	360;
	9625	11136	0.00539	0.022409	0	0	0	0	0	0	1	-360	360;
	2779	1209	0.18813	0.545888	0	0	0	0	0	0	1	-360	360;
	2779	2027	0.02382	0.200979	0	0	0	0	0	0	1	-360	360;
	2779	2027	0.00375	0.019289	0	0	0	0	0	0	1	-360	360;
	2779	3751	0.003151	0.01874	0	0	0	0	0	0	1	-360	360;
	2779	6981	0.02419	0.20719	0	0	0	0	0	0	1	-360	360;
	6965	8677	0.005479	0.022531	0	0	0	0	0	0	1	-360	360;
	6965	545	0.008581	0.133581	0	0	0	0	0	0	1	-360	360;
	10229	5661	0.000849	0.00436	0	0	0	0	0	0	1	-360	360;
	5557	4944	0.0004	0.00418	0	0	0	0	0	0	1	-360	360;
	5557	490	0.00023	0.00253	0	0	0	0	0	0	1	-360	360;
	7488	3881	0.000659	0.006909	0	0	0	0	0	0	1	-360	360;
	5557	3447	0.14356	0.922161	0	0	0	0	0	0	1	-360	360;
	5557	3260	0.05061	0.949861	0	0	0	0	0	0	1	-360	360;
	3546	10045	0.001669	0.19101	0	0	0	0	0	0	1	-360	360;
	5710	3546	0.008901	0.19594	0	0	0	0	0	0	1	-360	360;
	5710	10045	0.008849	0.207376	0	0	0	0	0	0	1	-360	360;
	854	3229	0.04068	0.314813	0	0	0	0	0	0	1	-360	360;
	854	148	0.001	0.358151	0	0	0	0	0	0	1	-360	360;
	854	148	0.00029	0.00339	0	0	0	0	0	0	1	-360	360;
	854	9278	0.001	0.381247	0	0	0	0	0	0	1	-360	360;
	5394	1872	0	0.0002	0	0	0	0	0	0	1	-360	360;
	5394	1872	0.001	0.261032	0	0	0	0	0	0	1	-360	360;
	7216	3421	5e-05	0.000409	0	0	0	0	0	0	1	-360	360;
	10610	7216	0.00036	0.00411	0	0	0	0	0	0	1	-360	360;
	6637	2431	0	0.0002	0	0	0	0	0	0	1	-360	360;
	6637	2431	0.06963	0.647459	0	0	0	0	0	0	1	-360	360;
	793	6637	0.00138	0.00794	0	0	0	0	0	0	1	-360	360;
	996	6637	0.04905	0.618285	0	0	0	0	0	0	1	-360	360;
	10794	6637	0.01481	0.65845	0	0	0	0	0	0	1	-360	360;
	6683	6637	0.057229	0.834525	0	0	0	0	0	0	1	-360	360;
	122	2431	0.00276	0.034271	0	0	0	0	0	0	1	-360	360;
	996	2431	0.045901	0.629938	0	0	0	0	0	0	1	-360	360;
	10794	2431	0.05236	0.804298	0	0	0	0	0	0	1	-360	360;
	6683	2431	0.010979	0.705971	0	0	0	0	0	0	1	-360	360;
	5394	3229	0.00111	0.00985	0	0	0	0	0	0	1	-360	360;
	5257	2602	12.66967	25.61427	0	0	0	0	0	0	1	-360	360;
	2602	10026	0.7509	2.64626	0	0	0	0	0	0	1	-360	360;
	11961	3731	0.001	0.522603	0	0	0	0	0	0	1	-360	360;
	11961	3731	0	0.0002	0	0	0	0	0	0	1	-360	360;
	12457	11961	1.015517	3.518182	0	0	0	0	0	0	1	-360	360;
	5661	11961	0.00274	0.01674	0	0	0	0	0	0	1	-360	360;
	13280	11961	0.001031	0.006669	0	0	0	0	0	0	1	-360	360;
	12457	3731	0.827397	3.941942	0	0	0	0	0	0	1	-360	360;
	5364	2602	0.00018	0.00147	0	0	0	0	0	0	1	-360	360;
	7756	2602	0.00025	0.00253	0	0	0	0	0	0	1	-360	360;
	5562	2602	0.00045	0.00372	0	0	0	0	0	0	1	-360	360;
	1609	2602	0.00019	0.00142	0	0	0	0	0	0	1	-360	360;
	8890	2602	0.00017	0.00161	0	0	0	0	0	0	1	-360	360;
	6776	2536	0.00031	0.00387	0	0	0	0	0	0	1	-360	360;
	9268	10850	0.00057	0.00261	0	0	0	0	0	0	1	-360	360;
	9268	2718	0.000591	0.002591	0	0	0	0	0	0	1	-360	360;
	3819	996	0.001599	0.00775	0	0	0	0	0	0	1	-360	360;
	793	996	0.014031	0.614215	0	0	0	0	0	0	1	-360	360;
	793	5661	0.012341	0.556384	0	0	0	0	0	0	1	-360	360;
	793	5661	0.001599	0.00738	0	0	0	0	0	0	1	-360	360;
	12850	2525	0.077687	1.036288	0	0	0	0	0	0	1	-360	360;
	12850	909	0.00057	0.00542	0	0	0	0	0	0	1	-360	360;
	12850	909	0.001	0.513781	0	0	0	0	0	0	1	-360	360;
	12850	2114	0.01036	0.396828	0	0	0	0	0	0	1	-360	360;
	12850	2114	0.00067	0.00594	0	0	0	0	0	0	1	-360	360;
	12850	6916	0.897853	2.161981	0	0	0	0	0	0	1	-360	360;
	2114	6268	0.0015	0.01442	0	0	0	0	0	0	1	-360	360;
	6916	6268	0.00139	0.01387	0	0	0	0	0	0	1	-360	360;
	909	2525	0.001	0.795637	0	0	0	0	0	0	1	-360	360;
	2114	2525	0.230298	1.27597	0	0	0	0	0	0	1	-360	360;
	6916	2525	0.05565	0.673331	0	0	0	0	0	0	1	-360	360;
	13539	23	0.00019	0.00214	0	0	0	0	0	0	1	-360	360;
	13539	23	0.00019	0.00198	0	0	0	0	0	0	1	-360	360;
	7818	7012	0.07358	0.966343	0	0	0	0	0	0	1	-360	360;
	8784	7818	0.06384	1.43054	0	0	0	0	0	0	1	-360	360;
	10047	7818	0.076281	1.166136	0	0	0	0	0	0	1	-360	360;
	8784	7012	0.210457	1.788158	0	0	0	0	0	0	1	-360	360;
	10047	7012	0.185533	1.718213	0	0	0	0	0	0	1	-360	360;
	7818	7012	2e-05	0.00024	0	0	0	0	0	0	1	-360	360;
	490	5784	0.01408	0.37133	0	0	0	0	0	0	1	-360	360;
	680	5784	0.01647	0.363712	0	0	0	0	0	0	1	-360	360;
	680	5784	0.00068	0.00621	0	0	0	0	0	0	1	-360	360;
	5784	148	0.00079	0.00807	0	0	0	0	0	0	1	-360	360;
	2728	8361	0.00052	0.00514	0	0	0	0	0	0	1	-360	360;
	148	3229	0.111198	0.57367	0	0	0	0	0	0	1	-360	360;
	9278	3229	0.126101	0.495623	0	0	0	0	0	0	1	-360	360;
	6259	2128	0.0032	0.016281	0	0	0	0	0	0	1	-360	360;
	6259	2128	0.003659	0.01407	0	0	0	0	0	0	1	-360	360;
	6259	2536	0.002281	0.010039	0	0	0	0	0	0	1	-360	360;
	6259	2536	0.001919	0.011669	0	0	0	0	0	0	1	-360	360;
	1431	7018	0.00236	0.02263	0	0	0	0	0	0	1	-360	360;
	568	1431	0.009849	0.054531	0	0	0	0	0	0	1	-360	360;
	2027	1209	0.02949	0.13405	0	0	0	0	0	0	1	-360	360;
	2027	1209	0.005021	0.030341	0	0	0	0	0	0	1	-360	360;
	5257	11534	0.00076	0.00675	0	0	0	0	0	0	1	-360	360;
	5257	10026	0.568047	3.572576	0	0	0	0	0	0	1	-360	360;
	2778	6496	0.003419	0.014531	0	0	0	0	0	0	1	-360	360;
	2886	2778	0.002979	0.016461	0	0	0	0	0	0	1	-360	360;
	9268	4749	0.09382	0.44438	0	0	0	0	0	0	1	-360	360;
	11208	4749	0.003789	0.02306	0	0	0	0	0	0	1	-360	360;
	3897	5562	0.00014	0.00108	0	0	0	0	0	0	1	-360	360;
	3897	13539	0.00015	0.00119	0	0	0	0	0	0	1	-360	360;
	2791	13539	0.00031	0.00232	0	0	0	0	0	0	1	-360	360;
	2556	2791	0.00025	0.00193	0	0	0	0	0	0	1	-360	360;
	13539	11674	0.137791	1.05367	0	0	0	0	0	0	1	-360	360;
	13539	10846	0.00054	0.00395	0	0	0	0	0	0	1	-360	360;
	8044	3881	0.00113	0.012581	0	0	0	0	0	0	1	-360	360;
	8044	568	0.00238	0.02086	0	0	0	0	0	0	1	-360	360;
	8044	10203	0.127331	1.369545	0	0	0	0	0	0	1	-360	360;
	8044	3171	0.699669	2.452893	0	0	0	0	0	0	1	-360	360;
	13339	9658	0.00331	0.025159	0	0	0	0	0	0	1	-360	360;
	909	2114	0.01975	0.825554	0	0	0	0	0	0	1	-360	360;
	6916	909	0.507639	2.270152	0	0	0	0	0	0	1	-360	360;
	8677	9658	0.20336	0.874897	0	0	0	0	0	0	1	-360	360;
	8677	545	0.00732	0.030659	0	0	0	0	0	0	1	-360	360;
	1626	8677	0.075401	0.292748	0	0	0	0	0	0	1	-360	360;
	1626	8677	0.00314	0.02136	0	0	0	0	0	0	1	-360	360;
	13547	10196	0.00044	0.00325	0	0	0	0	0	0	1	-360	360;
	73	3452	0.00076	0.00364	0	0	0	0	0	0	1	-360	360;
	73	8685	0.00075	0.00474	0	0	0	0	0	0	1	-360	360;
	9268	5095	3.557851	5.439876	0	0	0	0	0	0	1	-360	360;
	9268	7630	4.011157	6.327273	0	0	0	0	0	0	1	-360	360;
	9268	12875	0.006	0.327025	0	0	0	0	0	0	1	-360	360;
	5550	9268	3.78905	6.419628	0	0	0	0	0	0	1	-360	360;
	5687	9268	3.833471	5.760331	0	0	0	0	0	0	1	-360	360;
	5710	9268	0.000841	0.00362	0	0	0	0	0	0	1	-360	360;
	9268	2854	0.03	0.279855	0	0	0	0	0	0	1	-360	360;
	9268	2854	0.00107	0.004849	0	0	0	0	0	0	1	-360	360;
	9268	11156	3.092975	5.176033	0	0	0	0	0	0	1	-360	360;
	9268	11156	0.02688	0.291116	0	0	0	0	0	0	1	-360	360;
	9268	11156	0.00188	0.0132	0	0	0	0	0	0	1	-360	360;
	9571	2027	0.001469	0.00724	0	0	0	0	0	0	1	-360	360;
	9571	2027	0.0015	0.007081	0	0	0	0	0	0	1	-360	360;
	4956	7069	0.0017	0.00806	0	0	0	0	0	0	1	-360	360;
	2886	3744	0.00369	0.016091	0	0	0	0	0	0	1	-360	360;
	12560	10354	0.00301	0.0177	0	0	0	0	0	0	1	-360	360;
	9013	6971	5e-05	0.00053	0	0	0	0	0	0	1	-360	360;
	4944	5773	6e-05	0.0005	0	0	0	0	0	0	1	-360	360;
	4944	8353	4e-05	0.00047	0	0	0	0	0	0	1	-360	360;
	12875	238	0.00174	0.016521	0	0	0	0	0	0	1	-360	360;
	12000	8077	0.00019	0.00196	0	0	0	0	0	0	1	-360	360;
	2728	9397	0.00074	0.00584	0	0	0	0	0	0	1	-360	360;
	12000	2453	0.00019	0.00183	0	0	0	0	0	0	1	-360	360;
	9013	3732	0.00018	0.00186	0	0	0	0	0	0	1	-360	360;
	8784	10047	0	0.0002	0	0	0	0	0	0	1	-360	360;
	8784	10047	0.001	0.413968	0	0	0	0	0	0	1	-360	360;
	8784	7818	0.00215	0.02177	0	0	0	0	0	0	1	-360	360;
	10225	6981	0.008659	0.0422	0	0	0	0	0	0	1	-360	360;
	118	7121	0.00029	0.00252	0	0	0	0	0	0	1	-360	360;
	118	6597	0.0343	0.612763	0	0	0	0	0	0	1	-360	360;
	3748	4305	0.00036	0.00195	0	0	0	0	0	0	1	-360	360;
	3748	4305	0.000409	0.00213	0	0	0	0	0	0	1	-360	360;
	7235	3985	0.00118	0.008719	0	0	0	0	0	0	1	-360	360;
	5550	7235	0.0053	0.023409	0	0	0	0	0	0	1	-360	360;
	5687	7235	0.005401	0.022909	0	0	0	0	0	0	1	-360	360;
	8685	7235	0.000901	0.00412	0	0	0	0	0	0	1	-360	360;
	7235	3452	0.05139	0.435785	0	0	0	0	0	0	1	-360	360;
	5095	7630	0.01661	0.339194	0	0	0	0	0	0	1	-360	360;
	5095	12875	2.441736	4.517975	0	0	0	0	0	0	1	-360	360;
	5550	5095	0.03336	0.374174	0	0	0	0	0	0	1	-360	360;
	5687	5095	0.033219	0.37814	0	0	0	0	0	0	1	-360	360;
	5095	6591	0.003151	0.017599	0	0	0	0	0	0	1	-360	360;
	11156	5095	0.02231	0.393079	0	0	0	0	0	0	1	-360	360;
	9487	5364	0.0005	0.00519	0	0	0	0	0	0	1	-360	360;
	9487	12671	0.00024	0.00227	0	0	0	0	0	0	1	-360	360;
	9487	7121	0.00049	0.00529	0	0	0	0	0	0	1	-360	360;
	9487	7869	0.00036	0.00305	0	0	0	0	0	0	1	-360	360;
	6824	8260	0.00264	0.014781	0	0	0	0	0	0	1	-360	360;
	10082	8260	0.002539	0.01357	0	0	0	0	0	0	1	-360	360;
	3171	8260	0.0053	0.021599	0	0	0	0	0	0	1	-360	360;
	9726	8260	0.003091	0.01568	0	0	0	0	0	0	1	-360	360;
	13614	8260	0.00274	0.0127	0	0	0	0	0	0	1	-360	360;
	3171	8260	0.04224	0.3325	0	0	0	0	0	0	1	-360	360;
	8616	7990	0.00086	0.00873	0	0	0	0	0	0	1	-360	360;
	13547	7990	0.00206	0.01561	0	0	0	0	0	0	1	-360	360;
	5413	13521	0.00032	0.00271	0	0	0	0	0	0	1	-360	360;
	680	5413	0.00051	0.00524	0	0	0	0	0	0	1	-360	360;
	379	11324	0.00032	0.00265	0	0	0	0	0	0	1	-360	360;
	490	379	0.0005	0.00534	0	0	0	0	0	0	1	-360	360;
	11324	13521	0	0.0002	0	0	0	0	0	0	1	-360	360;
	11324	13521	0.163629	1.021814	0	0	0	0	0	0	1	-360	360;
	992	11324	0.001	0.350312	0	0	0	0	0	0	1	-360	360;
	11314	11324	0.0571	0.699515	0	0	0	0	0	0	1	-360	360;
	5515	11324	0.00022	0.00214	0	0	0	0	0	0	1	-360	360;
	6768	11324	0.05584	0.597071	0	0	0	0	0	0	1	-360	360;
	3260	13521	0.00044	0.00329	0	0	0	0	0	0	1	-360	360;
	992	13521	0.0684	0.765166	0	0	0	0	0	0	1	-360	360;
	992	13521	0.0005	0.00377	0	0	0	0	0	0	1	-360	360;
	11314	13521	0.02916	0.635783	0	0	0	0	0	0	1	-360	360;
	7161	13521	0.00021	0.00218	0	0	0	0	0	0	1	-360	360;
	6768	13521	0.076648	0.893421	0	0	0	0	0	0	1	-360	360;
	9978	11546	0.003581	0.01713	0	0	0	0	0	0	1	-360	360;
	3460	176	0.003651	0.01676	0	0	0	0	0	0	1	-360	360;
	6117	8691	0	0.0002	0	0	0	0	0	0	1	-360	360;
	6117	8691	0.001	0.363179	0	0	0	0	0	0	1	-360	360;
	6117	2728	0.03201	0.610942	0	0	0	0	0	0	1	-360	360;
	8691	2728	0.03493	0.703809	0	0	0	0	0	0	1	-360	360;
	8691	4570	0.00038	0.00378	0	0	0	0	0	0	1	-360	360;
	1626	9658	0.662562	1.966736	0	0	0	0	0	0	1	-360	360;
	1626	9658	0.007461	0.03539	0	0	0	0	0	0	1	-360	360;
	6916	2114	1.057548	2.384834	0	0	0	0	0	0	1	-360	360;
	2886	11546	0.00313	0.015591	0	0	0	0	0	0	1	-360	360;
	2886	176	0.00325	0.01495	0	0	0	0	0	0	1	-360	360;
	3881	2886	0.174211	0.885269	0	0	0	0	0	0	1	-360	360;
	10203	2886	0.05432	0.507293	0	0	0	0	0	0	1	-360	360;
	4593	2886	0.315744	1.143306	0	0	0	0	0	0	1	-360	360;
	11208	12560	0.106531	0.728988	0	0	0	0	0	0	1	-360	360;
	4944	2728	0.00046	0.00355	0	0	0	0	0	0	1	-360	360;
	9013	2728	0.00046	0.00356	0	0	0	0	0	0	1	-360	360;
	3985	7630	0.00056	0.00351	0	0	0	0	0	0	1	-360	360;
	11156	3985	0.001271	0.00699	0	0	0	0	0	0	1	-360	360;
	7630	12875	2.229752	5.528926	0	0	0	0	0	0	1	-360	360;
	5550	7630	0.186169	1.134959	0	0	0	0	0	0	1	-360	360;
	5687	7630	0.191419	0.99374	0	0	0	0	0	0	1	-360	360;
	11156	7630	0.001	0.342025	0	0	0	0	0	0	1	-360	360;
	13631	8870	8e-05	0.00071	0	0	0	0	0	0	1	-360	360;
	12000	8870	0.00047	0.00487	0	0	0	0	0	0	1	-360	360;
	5550	12875	2.630992	4.595455	0	0	0	0	0	0	1	-360	360;
	5687	12875	2.602893	4.194835	0	0	0	0	0	0	1	-360	360;
	5710	12875	0.002849	0.0122	0	0	0	0	0	0	1	-360	360;
	11156	12875	1.630455	4.689463	0	0	0	0	0	0	1	-360	360;
	4570	8870	0.136641	1.051662	0	0	0	0	0	0	1	-360	360;
	979	8870	0.00035	0.00282	0	0	0	0	0	0	1	-360	360;
	8616	13547	0.00092	0.0092	0	0	0	0	0	0	1	-360	360;
	3447	3260	0	0.0002	0	0	0	0	0	0	1	-360	360;
	3447	3260	0.02074	0.459238	0	0	0	0	0	0	1	-360	360;
	3447	12000	0.00025	0.00224	0	0	0	0	0	0	1	-360	360;
	11208	4305	0.000651	0.0037	0	0	0	0	0	0	1	-360	360;
	5550	5687	0.001	0.238182	0	0	0	0	0	0	1	-360	360;
	5550	11156	0.168469	0.943719	0	0	0	0	0	0	1	-360	360;
	5687	11156	0.1577	0.919711	0	0	0	0	0	0	1	-360	360;
	9444	7018	0.0033	0.023531	0	0	0	0	0	0	1	-360	360;
	13635	7018	0.00336	0.023031	0	0	0	0	0	0	1	-360	360;
	13016	7018	0.050961	0.353843	0	0	0	0	0	0	1	-360	360;
	2365	7018	0.07906	0.386591	0	0	0	0	0	0	1	-360	360;
	2365	7018	0.00224	0.02005	0	0	0	0	0	0	1	-360	360;
	9625	7018	0.001979	0.018841	0	0	0	0	0	0	1	-360	360;
	4841	31	0.0003	0.002669	0	0	0	0	0	0	1	-360	360;
	2602	2556	0.00017	0.00166	0	0	0	0	0	0	1	-360	360;
	1609	6597	0.00057	0.00619	0	0	0	0	0	0	1	-360	360;
	8890	9108	8e-05	0.00082	0	0	0	0	0	0	1	-360	360;
	996	5661	0.001	0.452087	0	0	0	0	0	0	1	-360	360;
	996	10794	0.028289	0.776901	0	0	0	0	0	0	1	-360	360;
	996	6683	0.027469	0.75064	0	0	0	0	0	0	1	-360	360;
	5661	3153	0.001409	0.00714	0	0	0	0	0	0	1	-360	360;
	7766	9726	0.00776	0.04488	0	0	0	0	0	0	1	-360	360;
	979	3566	0.00028	0.0023	0	0	0	0	0	0	1	-360	360;
	4685	979	0.00015	0.00155	0	0	0	0	0	0	1	-360	360;
	6683	772	0.000229	0.00113	0	0	0	0	0	0	1	-360	360;
	10794	4895	0.00024	0.001099	0	0	0	0	0	0	1	-360	360;
	9444	13016	0.000651	0.005031	0	0	0	0	0	0	1	-360	360;
	13635	13016	0.000581	0.0057	0	0	0	0	0	0	1	-360	360;
	2365	13016	0.395351	1.062603	0	0	0	0	0	0	1	-360	360;
	680	12000	0.00073	0.00574	0	0	0	0	0	0	1	-360	360;
	8305	9013	0.00074	0.00562	0	0	0	0	0	0	1	-360	360;
	1149	6597	0.00021	0.00217	0	0	0	0	0	0	1	-360	360;
	12769	8118	0.001	0.253636	0	0	0	0	0	0	1	-360	360;
	11436	4685	0.00078	0.00691	0	0	0	0	0	0	1	-360	360;
	8305	9278	0.00135	0.01007	0	0	0	0	0	0	1	-360	360;
	3881	10203	0.048271	0.517335	0	0	0	0	0	0	1	-360	360;
	3881	10203	0.001901	0.01187	0	0	0	0	0	0	1	-360	360;
	3881	4593	0.02499	0.413099	0	0	0	0	0	0	1	-360	360;
	3881	4593	0.00186	0.01099	0	0	0	0	0	0	1	-360	360;
	490	680	0.00351	0.296032	0	0	0	0	0	0	1	-360	360;
	3425	13449	0.000169	0.000789	0	0	0	0	0	0	1	-360	360;
	3452	3425	0.000211	0.001031	0	0	0	0	0	0	1	-360	360;
	5195	9736	0.00038	0.001909	0	0	0	0	0	0	1	-360	360;
	3452	5195	0.000219	0.001	0	0	0	0	0	0	1	-360	360;
	992	11314	0.0114	0.491198	0	0	0	0	0	0	1	-360	360;
	992	6768	0.0111	0.44741	0	0	0	0	0	0	1	-360	360;
	10203	4593	0.065849	0.525558	0	0	0	0	0	0	1	-360	360;
	10203	12980	0.00095	0.0053	0	0	0	0	0	0	1	-360	360;
	8685	3452	0.00139	0.008281	0	0	0	0	0	0	1	-360	360;
	10055	10493	0.001	0.26688	0	0	0	0	0	0	1	-360	360;
	7766	2890	0.00036	0.00286	0	0	0	0	0	0	1	-360	360;
	11156	2854	0.02231	0.26281	0	0	0	0	0	0	1	-360	360;
	11156	2854	0.001669	0.01064	0	0	0	0	0	0	1	-360	360;
	7818	11573	3e-05	0.00031	0	0	0	0	0	0	1	-360	360;
	7818	638	3e-05	0.0003	0	0	0	0	0	0	1	-360	360;
	11674	10846	0.0004	0.00637	0	0	0	0	0	0	1	-360	360;
	11436	7869	0.00018	0.0019	0	0	0	0	0	0	1	-360	360;
	11436	12671	0.00032	0.00267	0	0	0	0	0	0	1	-360	360;
	5515	11314	0.00023	0.00202	0	0	0	0	0	0	1	-360	360;
	11314	6768	0.001	0.382043	0	0	0	0	0	0	1	-360	360;
	11571	13614	0.00011	0.00075	0	0	0	0	0	0	1	-360	360;
	13035	13614	0.00013	0.000531	0	0	0	0	0	0	1	-360	360;
	10794	6683	0.010909	0.62314	0	0	0	0	0	0	1	-360	360;
	2128	2536	0.005969	0.037081	0	0	0	0	0	0	1	-360	360;
	6598	13441	0	0.0002	0	0	0	0	0	0	1	-360	360;
	6598	13441	0.001	0.39696	0	0	0	0	0	0	1	-360	360;
	6598	12362	0.192888	1.242729	0	0	0	0	0	0	1	-360	360;
	13441	2323	0.00078	0.00629	0	0	0	0	0	0	1	-360	360;
	13441	12362	0.215693	1.63338	0	0	0	0	0	0	1	-360	360;
	2323	12362	0.00048	0.00397	0	0	0	0	0	0	1	-360	360;
	6051	12362	0.00049	0.00388	0	0	0	0	0	0	1	-360	360;
	10610	12872	0.000211	0.00113	0	0	0	0	0	0	1	-360	360;
	7161	6768	3e-05	0.00035	0	0	0	0	0	0	1	-360	360;
	568	3171	0.007409	0.042281	0	0	0	0	0	0	1	-360	360;
	5479	9571	0.00012	0.0007	0	0	0	0	0	0	1	-360	360;
	148	9278	0	0.0002	0	0	0	0	0	0	1	-360	360;
	148	9278	0.001	0.284017	0	0	0	0	0	0	1	-360	360;
	12987	12151	1.997645	4.178116	0	0	0	0	0	0	1	-360	360;
	12987	4498	0.00015	0.00152	0	0	0	0	0	0	1	-360	360;
	12987	1823	0.00014	0.00141	0	0	0	0	0	0	1	-360	360;
	12987	6933	0.00103	0.00914	0	0	0	0	0	0	1	-360	360;
	12987	4218	0.00105	0.00896	0	0	0	0	0	0	1	-360	360;
	3291	12987	0.334702	1.492936	0	0	0	0	0	0	1	-360	360;
	3291	12987	0.00142	0.01428	0	0	0	0	0	0	1	-360	360;
	739	12987	0.340893	1.460526	0	0	0	0	0	0	1	-360	360;
	739	12987	0.00139	0.01455	0	0	0	0	0	0	1	-360	360;
	582	6840	0.0004	0.00376	0	0	0	0	0	0	1	-360	360;
	582	7757	0.00101	0.01129	0	0	0	0	0	0	1	-360	360;
	582	6704	0.00076	0.0089	0	0	0	0	0	0	1	-360	360;
	582	6840	0.0004	0.00368	0	0	0	0	0	0	1	-360	360;
	582	3839	0.00099	0.01151	0	0	0	0	0	0	1	-360	360;
	582	1320	0.00075	0.00824	0	0	0	0	0	0	1	-360	360;
	11659	6850	0.00156	0.01568	0	0	0	0	0	0	1	-360	360;
	11659	6850	0.00152	0.01599	0	0	0	0	0	0	1	-360	360;
	712	10235	0.00539	0.01612	0	0	0	0	0	0	1	-360	360;
	712	10235	0.0048	0.018271	0	0	0	0	0	0	1	-360	360;
	12052	1224	0.008729	0.034159	0	0	0	0	0	0	1	-360	360;
	12052	1224	0.00855	0.03482	0	0	0	0	0	0	1	-360	360;
	6106	2716	0.00146	0.00651	0	0	0	0	0	0	1	-360	360;
	748	2716	0.134252	0.771745	0	0	0	0	0	0	1	-360	360;
	748	2716	0.00072	0.00587	0	0	0	0	0	0	1	-360	360;
	748	2716	0.00071	0.006	0	0	0	0	0	0	1	-360	360;
	9319	2716	0.00013	0.00132	0	0	0	0	0	0	1	-360	360;
	2684	2716	0.322853	1.362742	0	0	0	0	0	0	1	-360	360;
	4092	13230	0.161841	1.112975	0	0	0	0	0	0	1	-360	360;
	4092	13230	0.00356	0.022409	0	0	0	0	0	0	1	-360	360;
	645	13230	0.211508	1.021529	0	0	0	0	0	0	1	-360	360;
	645	13230	0.00838	0.038479	0	0	0	0	0	0	1	-360	360;
	712	13230	0.912748	1.922025	0	0	0	0	0	0	1	-360	360;
	11046	12063	0.005211	0.016901	0	0	0	0	0	0	1	-360	360;
	11046	12063	0.219566	0.646591	0	0	0	0	0	0	1	-360	360;
	11046	12063	0.00531	0.016539	0	0	0	0	0	0	1	-360	360;
	7632	10638	0.140519	0.391918	0	0	0	0	0	0	1	-360	360;
	6973	7632	0.00131	0.01484	0	0	0	0	0	0	1	-360	360;
	7632	6704	0.05123	0.434093	0	0	0	0	0	0	1	-360	360;
	7632	6704	0.0006	0.00673	0	0	0	0	0	0	1	-360	360;
	7632	12335	0.05191	0.445222	0	0	0	0	0	0	1	-360	360;
	5154	7632	0.736704	3.325	0	0	0	0	0	0	1	-360	360;
	3988	7632	2.622299	8.981301	0	0	0	0	0	0	1	-360	360;
	7574	9460	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7574	9460	0.001	0.291627	0	0	0	0	0	0	1	-360	360;
	3536	7574	0.00016	0.00138	0	0	0	0	0	0	1	-360	360;
	6285	7574	0.00017	0.00135	0	0	0	0	0	0	1	-360	360;
	702	9460	0.00017	0.00133	0	0	0	0	0	0	1	-360	360;
	1964	9460	0.00017	0.0013	0	0	0	0	0	0	1	-360	360;
	11659	3536	0.00189	0.01446	0	0	0	0	0	0	1	-360	360;
	6285	1393	0.00074	0.00698	0	0	0	0	0	0	1	-360	360;
	11659	702	0.00193	0.01416	0	0	0	0	0	0	1	-360	360;
	1964	6128	0.00073	0.00711	0	0	0	0	0	0	1	-360	360;
	3291	12151	0.136731	0.910873	0	0	0	0	0	0	1	-360	360;
	3291	12151	0.00128	0.0134	0	0	0	0	0	0	1	-360	360;
	739	12151	0.13394	0.92874	0	0	0	0	0	0	1	-360	360;
	8715	3115	0.271641	0.707964	0	0	0	0	0	0	1	-360	360;
	8715	3115	0.00037	0.0037	0	0	0	0	0	0	1	-360	360;
	8715	3115	0.00038	0.00363	0	0	0	0	0	0	1	-360	360;
	5154	3115	0.00156	0.01473	0	0	0	0	0	0	1	-360	360;
	3988	3115	0.00159	0.01442	0	0	0	0	0	0	1	-360	360;
	12223	2049	0.01049	0.04336	0	0	0	0	0	0	1	-360	360;
	4639	2049	0.010271	0.04418	0	0	0	0	0	0	1	-360	360;
	11046	2049	0.00881	0.037021	0	0	0	0	0	0	1	-360	360;
	11366	2049	0.004409	0.01682	0	0	0	0	0	0	1	-360	360;
	11366	2049	0.00432	0.017151	0	0	0	0	0	0	1	-360	360;
	10997	2049	0.01025	0.042341	0	0	0	0	0	0	1	-360	360;
	5521	3921	0.05386	0.203789	0	0	0	0	0	0	1	-360	360;
	5521	3921	0.005	0.017849	0	0	0	0	0	0	1	-360	360;
	5521	3921	0.004909	0.01824	0	0	0	0	0	0	1	-360	360;
	3921	4615	0.442645	0.98938	0	0	0	0	0	0	1	-360	360;
	3921	4615	0.00789	0.025831	0	0	0	0	0	0	1	-360	360;
	3921	4615	0.00774	0.02638	0	0	0	0	0	0	1	-360	360;
	9450	748	0.00141	0.01209	0	0	0	0	0	0	1	-360	360;
	9450	6457	0.104079	0.42973	0	0	0	0	0	0	1	-360	360;
	1393	6840	0.00139	0.01016	0	0	0	0	0	0	1	-360	360;
	6128	6840	0.00142	0.00994	0	0	0	0	0	0	1	-360	360;
	3106	571	9e-05	0.00064	0	0	0	0	0	0	1	-360	360;
	3106	6933	0.195942	0.820914	0	0	0	0	0	0	1	-360	360;
	3106	4218	0.199778	0.804155	0	0	0	0	0	0	1	-360	360;
	739	3106	0.02836	0.37876	0	0	0	0	0	0	1	-360	360;
	571	6933	0.00109	0.00814	0	0	0	0	0	0	1	-360	360;
	739	571	0.00054	0.00424	0	0	0	0	0	0	1	-360	360;
	4520	748	0.00263	0.02613	0	0	0	0	0	0	1	-360	360;
	748	10099	0.00166	0.01885	0	0	0	0	0	0	1	-360	360;
	748	2684	0.847576	2.035042	0	0	0	0	0	0	1	-360	360;
	1759	748	0.00094	0.00794	0	0	0	0	0	0	1	-360	360;
	4092	1987	0.014081	0.055831	0	0	0	0	0	0	1	-360	360;
	4092	645	2.006198	3.572314	0	0	0	0	0	0	1	-360	360;
	4092	645	0.0122	0.05982	0	0	0	0	0	0	1	-360	360;
	4092	712	0.30564	1.254566	0	0	0	0	0	0	1	-360	360;
	4092	712	0.0235	0.076229	0	0	0	0	0	0	1	-360	360;
	4092	4615	0.029969	0.12889	0	0	0	0	0	0	1	-360	360;
	4092	4615	0.02932	0.119341	0	0	0	0	0	0	1	-360	360;
	9680	6596	0.00699	0.02886	0	0	0	0	0	0	1	-360	360;
	6596	1224	0.004909	0.014719	0	0	0	0	0	0	1	-360	360;
	3960	10997	0.005591	0.01814	0	0	0	0	0	0	1	-360	360;
	3960	10997	0.00569	0.01776	0	0	0	0	0	0	1	-360	360;
	10638	6704	0.03624	0.415741	0	0	0	0	0	0	1	-360	360;
	10638	12335	0.03359	0.482777	0	0	0	0	0	0	1	-360	360;
	10638	12335	0.00088	0.01001	0	0	0	0	0	0	1	-360	360;
	10638	1320	0.001	0.00981	0	0	0	0	0	0	1	-360	360;
	5154	10638	0.706579	3.541136	0	0	0	0	0	0	1	-360	360;
	3988	10638	2.102562	10.3982	0	0	0	0	0	0	1	-360	360;
	712	1987	0.00962	0.03813	0	0	0	0	0	0	1	-360	360;
	11254	7543	0.04235	0.255499	0	0	0	0	0	0	1	-360	360;
	11254	10219	0.03604	0.287562	0	0	0	0	0	0	1	-360	360;
	11254	10219	0.00069	0.00676	0	0	0	0	0	0	1	-360	360;
	11254	6772	0.02303	0.242479	0	0	0	0	0	0	1	-360	360;
	6595	11254	0.002	0.02189	0	0	0	0	0	0	1	-360	360;
	7757	3988	0.00063	0.00569	0	0	0	0	0	0	1	-360	360;
	3839	5154	0.00064	0.00557	0	0	0	0	0	0	1	-360	360;
	7543	10219	0.001	0.193123	0	0	0	0	0	0	1	-360	360;
	7543	6772	0.01778	0.156988	0	0	0	0	0	0	1	-360	360;
	7543	6772	0.00104	0.00843	0	0	0	0	0	0	1	-360	360;
	3291	7543	0.00096	0.01465	0	0	0	0	0	0	1	-360	360;
	6595	7543	0.003	0.02517	0	0	0	0	0	0	1	-360	360;
	10219	6772	0.462929	1.374931	0	0	0	0	0	0	1	-360	360;
	853	10219	0.00244	0.01729	0	0	0	0	0	0	1	-360	360;
	739	10219	0.00094	0.01356	0	0	0	0	0	0	1	-360	360;
	12223	9924	0.01526	0.05118	0	0	0	0	0	0	1	-360	360;
	4639	9924	0.01555	0.05011	0	0	0	0	0	0	1	-360	360;
	645	712	0.26188	0.583802	0	0	0	0	0	0	1	-360	360;
	4520	11659	0.0036	0.04	0	0	0	0	0	0	1	-360	360;
	4520	727	0.00363	0.03446	0	0	0	0	0	0	1	-360	360;
	4520	2684	0.00099	0.01049	0	0	0	0	0	0	1	-360	360;
	13256	6772	0.00332	0.03036	0	0	0	0	0	0	1	-360	360;
	5724	2021	0.00027	0.00309	0	0	0	0	0	0	1	-360	360;
	7666	2021	0.00031	0.00303	0	0	0	0	0	0	1	-360	360;
	5521	4615	0.027099	0.21312	0	0	0	0	0	0	1	-360	360;
	5521	4615	0.00357	0.011021	0	0	0	0	0	0	1	-360	360;
	5521	4615	0.00351	0.01126	0	0	0	0	0	0	1	-360	360;
	11659	727	0.00037	0.00418	0	0	0	0	0	0	1	-360	360;
	6056	9680	0.04124	0.192349	0	0	0	0	0	0	1	-360	360;
	6056	9680	0.015669	0.064711	0	0	0	0	0	0	1	-360	360;
	6056	9680	0.016	0.06349	0	0	0	0	0	0	1	-360	360;
	9680	1224	0.012281	0.03676	0	0	0	0	0	0	1	-360	360;
	11046	277	0.02043	0.19226	0	0	0	0	0	0	1	-360	360;
	11046	277	0.00287	0.01095	0	0	0	0	0	0	1	-360	360;
	11046	277	0.00281	0.011169	0	0	0	0	0	0	1	-360	360;
	6457	10099	0.00054	0.00596	0	0	0	0	0	0	1	-360	360;
	6458	6457	0.00134	0.01244	0	0	0	0	0	0	1	-360	360;
	1393	6128	0.001	0.265693	0	0	0	0	0	0	1	-360	360;
	13256	1393	0.15358	0.524508	0	0	0	0	0	0	1	-360	360;
	10502	1393	0.613823	2.118352	0	0	0	0	0	0	1	-360	360;
	10502	1393	0.00081	0.0077	0	0	0	0	0	0	1	-360	360;
	1393	5154	0.00228	0.02083	0	0	0	0	0	0	1	-360	360;
	13256	6128	0.168047	0.592382	0	0	0	0	0	0	1	-360	360;
	13256	6128	0.0008	0.00785	0	0	0	0	0	0	1	-360	360;
	10502	6128	0.698199	2.371468	0	0	0	0	0	0	1	-360	360;
	6128	3988	0.00232	0.02039	0	0	0	0	0	0	1	-360	360;
	3291	8715	0.00177	0.01939	0	0	0	0	0	0	1	-360	360;
	739	8715	0.00173	0.01976	0	0	0	0	0	0	1	-360	360;
	13256	10502	0	0.0002	0	0	0	0	0	0	1	-360	360;
	13256	10502	0.001	0.213047	0	0	0	0	0	0	1	-360	360;
	13256	3988	0.00194	0.01469	0	0	0	0	0	0	1	-360	360;
	853	13256	0.00123	0.00838	0	0	0	0	0	0	1	-360	360;
	13256	5273	0.00121	0.0116	0	0	0	0	0	0	1	-360	360;
	13256	6595	0.00346	0.03846	0	0	0	0	0	0	1	-360	360;
	10502	5154	0.00198	0.01438	0	0	0	0	0	0	1	-360	360;
	10502	6595	0.00338	0.0392	0	0	0	0	0	0	1	-360	360;
	10502	284	0.00248	0.02446	0	0	0	0	0	0	1	-360	360;
	6933	4218	0.001	0.26367	0	0	0	0	0	0	1	-360	360;
	739	6933	0.580512	1.382895	0	0	0	0	0	0	1	-360	360;
	2509	6933	0.00094	0.00807	0	0	0	0	0	0	1	-360	360;
	6973	4218	0.00232	0.01638	0	0	0	0	0	0	1	-360	360;
	3291	4218	0.00217	0.0131	0	0	0	0	0	0	1	-360	360;
	739	4218	0.568421	1.409488	0	0	0	0	0	0	1	-360	360;
	2512	8132	0.005331	0.017341	0	0	0	0	0	0	1	-360	360;
	6973	12335	0.00199	0.02218	0	0	0	0	0	0	1	-360	360;
	6973	5154	0.00163	0.01513	0	0	0	0	0	0	1	-360	360;
	6973	3988	0.00159	0.01541	0	0	0	0	0	0	1	-360	360;
	6973	2509	0.00136	0.00959	0	0	0	0	0	0	1	-360	360;
	6973	5724	0.00042	0.00496	0	0	0	0	0	0	1	-360	360;
	6973	7666	0.00041	0.00505	0	0	0	0	0	0	1	-360	360;
	11046	10997	0.016591	0.06199	0	0	0	0	0	0	1	-360	360;
	6704	12335	0	0.0002	0	0	0	0	0	0	1	-360	360;
	6704	12335	0.001	0.404868	0	0	0	0	0	0	1	-360	360;
	5154	6704	0.130547	2.317729	0	0	0	0	0	0	1	-360	360;
	3988	6704	0.431627	6.284488	0	0	0	0	0	0	1	-360	360;
	5154	12335	0.127403	2.421884	0	0	0	0	0	0	1	-360	360;
	3988	12335	0.397583	6.979917	0	0	0	0	0	0	1	-360	360;
	2684	6458	0.00356	0.03374	0	0	0	0	0	0	1	-360	360;
	5154	3988	0.00164	0.14108	0	0	0	0	0	0	1	-360	360;
	6595	6458	0.00058	0.00658	0	0	0	0	0	0	1	-360	360;
	6595	6458	0.00065	0.00645	0	0	0	0	0	0	1	-360	360;
	5273	284	0.079079	0.405817	0	0	0	0	0	0	1	-360	360;
	5273	284	0.00139	0.01268	0	0	0	0	0	0	1	-360	360;
	3291	739	0	0.0002	0	0	0	0	0	0	1	-360	360;
	3291	739	0.001	0.270111	0	0	0	0	0	0	1	-360	360;
	12396	9924	0.185091	0.839215	0	0	0	0	0	0	1	-360	360;
	12396	9924	0.004651	0.039669	0	0	0	0	0	0	1	-360	360;
	12396	9924	0.00476	0.042841	0	0	0	0	0	0	1	-360	360;
	6595	284	0.0018	0.02033	0	0	0	0	0	0	1	-360	360;
	795	10162	0.001	0.335748	0	0	0	0	0	0	1	-360	360;
	5724	795	0.0001	0.00068	0	0	0	0	0	0	1	-360	360;
	7666	10162	0.0001	0.00067	0	0	0	0	0	0	1	-360	360;
	7187	13593	0.026107	0.07605	0	0	0	0	0	0	1	-360	360;
	11834	7187	0.019727	0.062124	0	0	0	0	0	0	1	-360	360;
	7296	13593	0.026603	0.074438	0	0	0	0	0	0	1	-360	360;
	11834	7296	0.020107	0.060851	0	0	0	0	0	0	1	-360	360;
	3869	9774	0.014017	0.040471	0	0	0	0	0	0	1	-360	360;
	9774	1978	0.012488	0.045289	0	0	0	0	0	0	1	-360	360;
	9202	8058	0.014281	0.039612	0	0	0	0	0	0	1	-360	360;
	8058	10723	0.010826	0.032686	0	0	0	0	0	0	1	-360	360;
	6549	9825	0.012554	0.049893	0	0	0	0	0	0	1	-360	360;
	9825	11940	0.00705	0.023174	0	0	0	0	0	0	1	-360	360;
	6442	7230	0.018818	0.054132	0	0	0	0	0	0	1	-360	360;
	1978	7230	0.032769	0.043661	0	0	0	0	0	0	1	-360	360;
	8295	6549	0.015322	0.049653	0	0	0	0	0	0	1	-360	360;
	3632	968	0.002413	0.006917	0	0	0	0	0	0	1	-360	360;
	968	7795	0.010537	0.025917	0	0	0	0	0	0	1	-360	360;
	6442	3869	0.009124	0.032314	0	0	0	0	0	0	1	-360	360;
	6442	9202	0.008934	0.032934	0	0	0	0	0	0	1	-360	360;
	13593	5994	0	0.000413	0	0	0	0	0	0	1	-360	360;
	6988	5994	0.015562	0.044669	0	0	0	0	0	0	1	-360	360;
	9637	5994	0.013	0.049983	0	0	0	0	0	0	1	-360	360;
	13593	7165	0.014347	0.041182	0	0	0	0	0	0	1	-360	360;
	13593	7165	0.014628	0.040322	0	0	0	0	0	0	1	-360	360;
	2445	1003	0.017446	0.029752	0	0	0	0	0	0	1	-360	360;
	2445	9624	0.031579	0.070992	0	0	0	0	0	0	1	-360	360;
	10419	2445	0.007843	0.018793	0	0	0	0	0	0	1	-360	360;
	6467	2445	0.007992	0.018413	0	0	0	0	0	0	1	-360	360;
	8612	11275	0.001271	0.016781	0	0	0	0	0	0	1	-360	360;
	8612	11275	0.00124	0.015539	0	0	0	0	0	0	1	-360	360;
	1003	6442	0.033777	0.06419	0	0	0	0	0	0	1	-360	360;
	6442	1687	0.014347	0.028397	0	0	0	0	0	0	1	-360	360;
	9624	6442	0.023331	0.052182	0	0	0	0	0	0	1	-360	360;
	6442	10119	0.009446	0.022636	0	0	0	0	0	0	1	-360	360;
	10723	1978	0.006992	0.025488	0	0	0	0	0	0	1	-360	360;
	10119	1687	0.003339	0.008851	0	0	0	0	0	0	1	-360	360;
	6988	12899	0.016471	0.06081	0	0	0	0	0	0	1	-360	360;
	10855	9637	0.020149	0.058727	0	0	0	0	0	0	1	-360	360;
	7165	893	0.013992	0.054	0	0	0	0	0	0	1	-360	360;
	7165	5862	0.013686	0.05	0	0	0	0	0	0	1	-360	360;
	7165	6625	0.000223	0.000826	0	0	0	0	0	0	1	-360	360;
	7165	2262	0.000248	0.00081	0	0	0	0	0	0	1	-360	360;
	13369	12899	0.036512	0.098843	0	0	0	0	0	0	1	-360	360;
	893	302	0.012876	0.037438	0	0	0	0	0	0	1	-360	360;
	5862	302	0.013124	0.036636	0	0	0	0	0	0	1	-360	360;
	3289	10419	0.013909	0.033041	0	0	0	0	0	0	1	-360	360;
	7795	6467	0.014182	0.032372	0	0	0	0	0	0	1	-360	360;
	7516	12260	0.00474	0.027521	0	0	0	0	0	0	1	-360	360;
	7516	10027	0.00538	0.026969	0	0	0	0	0	0	1	-360	360;
	7120	11940	0.020372	0.06276	0	0	0	0	0	0	1	-360	360;
	13369	10855	0	0.000413	0	0	0	0	0	0	1	-360	360;
	9006	8612	0.0053	0.02831	0	0	0	0	0	0	1	-360	360;
	9006	8612	0.00601	0.02774	0	0	0	0	0	0	1	-360	360;
	12260	9006	0.00113	0.006229	0	0	0	0	0	0	1	-360	360;
	10027	9006	0.001151	0.00611	0	0	0	0	0	0	1	-360	360;
	3632	3289	0.014364	0.029529	0	0	0	0	0	0	1	-360	360;
	7120	3632	0.009017	0.019727	0	0	0	0	0	0	1	-360	360;
	7120	3632	0.00919	0.019314	0	0	0	0	0	0	1	-360	360;
	10027	6702	9.1e-05	0.000219	0	0	0	0	0	0	1	-360	360;
	12260	6702	9.1e-05	0.000229	0	0	0	0	0	0	1	-360	360;
	8295	7120	0.00205	0.00562	0	0	0	0	0	0	1	-360	360;
	302	8295	0.016	0.049818	0	0	0	0	0	0	1	-360	360;
	302	8295	0.014331	0.054347	0	0	0	0	0	0	1	-360	360;
	9829	7120	0.00738	0.021149	0	0	0	0	0	0	1	-360	360;
	302	9829	0.011256	0.027769	0	0	0	0	0	0	1	-360	360;
	737	7120	0.007207	0.019587	0	0	0	0	0	0	1	-360	360;
	302	737	0.01195	0.024826	0	0	0	0	0	0	1	-360	360;
	8266	4567	0.00796	0.04504	0	0	0	0	0	0	1	-360	360;
	8266	10128	0.00844	0.068649	0	0	0	0	0	0	1	-360	360;
	8266	8557	0.021551	0.0842	0	0	0	0	0	0	1	-360	360;
	6172	9494	0.002809	0.01632	0	0	0	0	0	0	1	-360	360;
	6172	538	0.00028	0.00084	0	0	0	0	0	0	1	-360	360;
	6172	538	0.00032	0.000822	0	0	0	0	0	0	1	-360	360;
	6172	4072	0.007711	0.054271	0	0	0	0	0	0	1	-360	360;
	6172	6615	0.002298	0.017129	0	0	0	0	0	0	1	-360	360;
	13305	13201	0.004711	0.026049	0	0	0	0	0	0	1	-360	360;
	13305	7936	0.008462	0.03792	0	0	0	0	0	0	1	-360	360;
	1615	2539	0.00095	0.01117	0	0	0	0	0	0	1	-360	360;
	1615	7453	0.00142	0.01429	0	0	0	0	0	0	1	-360	360;
	8420	11585	0.000569	0.005382	0	0	0	0	0	0	1	-360	360;
	4842	2932	0.00984	0.04424	0	0	0	0	0	0	1	-360	360;
	4842	6286	0.0166	0.105449	0	0	0	0	0	0	1	-360	360;
	1201	10931	0.007209	0.034969	0	0	0	0	0	0	1	-360	360;
	1201	2391	0.023951	0.07004	0	0	0	0	0	0	1	-360	360;
	13201	3261	0.002258	0.01644	0	0	0	0	0	0	1	-360	360;
	1212	1965	0.00948	0.02892	0	0	0	0	0	0	1	-360	360;
	1212	3690	0.01224	0.04384	0	0	0	0	0	0	1	-360	360;
	3364	3690	0.016	0.062369	0	0	0	0	0	0	1	-360	360;
	3364	3204	0.020422	0.082729	0	0	0	0	0	0	1	-360	360;
	3364	3204	0.020009	0.08436	0	0	0	0	0	0	1	-360	360;
	6517	3204	0.010702	0.030338	0	0	0	0	0	0	1	-360	360;
	6517	3204	0.010902	0.02968	0	0	0	0	0	0	1	-360	360;
	8174	3813	0.00848	0.040809	0	0	0	0	0	0	1	-360	360;
	8174	2315	0.012862	0.05772	0	0	0	0	0	0	1	-360	360;
	8161	2539	0.00047	0.00546	0	0	0	0	0	0	1	-360	360;
	11298	11245	0.002871	0.02108	0	0	0	0	0	0	1	-360	360;
	11298	6286	0.004849	0.02916	0	0	0	0	0	0	1	-360	360;
	11298	6286	0.004751	0.029711	0	0	0	0	0	0	1	-360	360;
	7530	6286	0.001569	0.005489	0	0	0	0	0	0	1	-360	360;
	9727	9500	0.0003	0.00316	0	0	0	0	0	0	1	-360	360;
	9727	7453	0.00114	0.01207	0	0	0	0	0	0	1	-360	360;
	9727	8871	0.00147	0.01724	0	0	0	0	0	0	1	-360	360;
	9727	13640	0.00039	0.00552	0	0	0	0	0	0	1	-360	360;
	9494	4072	0.004991	0.036511	0	0	0	0	0	0	1	-360	360;
	9494	13365	0.007302	0.041071	0	0	0	0	0	0	1	-360	360;
	2337	8699	0.00022	0.00305	0	0	0	0	0	0	1	-360	360;
	2337	428	0.00074	0.00834	0	0	0	0	0	0	1	-360	360;
	2337	6191	0.00126	0.0174	0	0	0	0	0	0	1	-360	360;
	6209	4567	0.00216	0.012578	0	0	0	0	0	0	1	-360	360;
	6209	944	0.001329	0.0106	0	0	0	0	0	0	1	-360	360;
	6209	10128	0.006049	0.063742	0	0	0	0	0	0	1	-360	360;
	7828	12468	0.006769	0.02648	0	0	0	0	0	0	1	-360	360;
	7828	11956	0.016022	0.047791	0	0	0	0	0	0	1	-360	360;
	7828	4072	0.02884	0.096209	0	0	0	0	0	0	1	-360	360;
	10911	11956	0.00656	0.024578	0	0	0	0	0	0	1	-360	360;
	10911	1965	0.00388	0.017342	0	0	0	0	0	0	1	-360	360;
	10911	1962	0.01772	0.064978	0	0	0	0	0	0	1	-360	360;
	4567	948	0.01312	0.052809	0	0	0	0	0	0	1	-360	360;
	4567	944	0.001738	0.002889	0	0	0	0	0	0	1	-360	360;
	948	10739	0.001169	0.003489	0	0	0	0	0	0	1	-360	360;
	6086	11053	0.00368	0.035022	0	0	0	0	0	0	1	-360	360;
	6086	9989	0.001391	0.010338	0	0	0	0	0	0	1	-360	360;
	6086	6744	0.004222	0.030831	0	0	0	0	0	0	1	-360	360;
	4333	8312	0.004271	0.036089	0	0	0	0	0	0	1	-360	360;
	4333	11053	0.00596	0.04024	0	0	0	0	0	0	1	-360	360;
	5655	7468	0.011529	0.046698	0	0	0	0	0	0	1	-360	360;
	5655	3690	0.02252	0.087338	0	0	0	0	0	0	1	-360	360;
	9664	2045	0.004071	0.035591	0	0	0	0	0	0	1	-360	360;
	9664	6744	0.004138	0.0386	0	0	0	0	0	0	1	-360	360;
	4075	8690	0.010911	0.027231	0	0	0	0	0	0	1	-360	360;
	120	9526	0.00256	0.024622	0	0	0	0	0	0	1	-360	360;
	120	6744	0.005129	0.040671	0	0	0	0	0	0	1	-360	360;
	12468	10931	0.005462	0.032671	0	0	0	0	0	0	1	-360	360;
	12468	2391	0.002871	0.014209	0	0	0	0	0	0	1	-360	360;
	12468	2391	0.002929	0.013911	0	0	0	0	0	0	1	-360	360;
	12468	10247	0.003422	0.011009	0	0	0	0	0	0	1	-360	360;
	9526	11053	0.00212	0.016769	0	0	0	0	0	0	1	-360	360;
	9493	7453	0.00125	0.01387	0	0	0	0	0	0	1	-360	360;
	12719	3690	0.004289	0.017071	0	0	0	0	0	0	1	-360	360;
	12719	7692	0.0084	0.024738	0	0	0	0	0	0	1	-360	360;
	8690	6286	0.01044	0.039671	0	0	0	0	0	0	1	-360	360;
	8690	6286	0.010218	0.040418	0	0	0	0	0	0	1	-360	360;
	8690	5502	0.008351	0.02432	0	0	0	0	0	0	1	-360	360;
	2045	3449	0.00188	0.006711	0	0	0	0	0	0	1	-360	360;
	2045	13365	0.00132	0.00352	0	0	0	0	0	0	1	-360	360;
	10869	6191	0.00113	0.01322	0	0	0	0	0	0	1	-360	360;
	10869	2539	0.00098	0.01301	0	0	0	0	0	0	1	-360	360;
	2932	7936	0.005271	0.05104	0	0	0	0	0	0	1	-360	360;
	2932	11036	0.017458	0.109631	0	0	0	0	0	0	1	-360	360;
	2932	327	0.002831	0.01016	0	0	0	0	0	0	1	-360	360;
	2932	327	0.001978	0.010902	0	0	0	0	0	0	1	-360	360;
	428	6191	0.00059	0.00764	0	0	0	0	0	0	1	-360	360;
	1909	11956	0.000231	0.000769	0	0	0	0	0	0	1	-360	360;
	8312	3261	0.002809	0.02468	0	0	0	0	0	0	1	-360	360;
	8312	11585	0.009942	0.050222	0	0	0	0	0	0	1	-360	360;
	6191	7978	0.00117	0.01756	0	0	0	0	0	0	1	-360	360;
	6191	11904	9e-05	0.0017	0	0	0	0	0	0	1	-360	360;
	3261	9353	0.002569	0.02344	0	0	0	0	0	0	1	-360	360;
	3261	6919	0.00144	0.011369	0	0	0	0	0	0	1	-360	360;
	13381	12066	0.002129	0.019689	0	0	0	0	0	0	1	-360	360;
	6286	11036	0.00756	0.048378	0	0	0	0	0	0	1	-360	360;
	6286	7468	0.031138	0.104969	0	0	0	0	0	0	1	-360	360;
	7978	8871	0.00043	0.00623	0	0	0	0	0	0	1	-360	360;
	11053	3654	0.000769	0.005769	0	0	0	0	0	0	1	-360	360;
	1065	6744	0.01464	0.043018	0	0	0	0	0	0	1	-360	360;
	1065	194	0.006609	0.014329	0	0	0	0	0	0	1	-360	360;
	3156	8871	0.00021	0.00252	0	0	0	0	0	0	1	-360	360;
	6425	6744	0.002689	0.019791	0	0	0	0	0	0	1	-360	360;
	6425	6744	0.002738	0.01936	0	0	0	0	0	0	1	-360	360;
	6425	6744	0.001338	0.010018	0	0	0	0	0	0	1	-360	360;
	9989	6744	0.001698	0.012871	0	0	0	0	0	0	1	-360	360;
	2353	6744	0.00384	0.028342	0	0	0	0	0	0	1	-360	360;
	4072	12066	0.00616	0.033618	0	0	0	0	0	0	1	-360	360;
	4072	1962	0.01488	0.045698	0	0	0	0	0	0	1	-360	360;
	9500	13640	4e-05	0.00051	0	0	0	0	0	0	1	-360	360;
	2391	3813	0.028751	0.090409	0	0	0	0	0	0	1	-360	360;
	2391	2315	0.00092	0.004609	0	0	0	0	0	0	1	-360	360;
	5788	3690	0.005151	0.012329	0	0	0	0	0	0	1	-360	360;
	9353	6919	0.002129	0.018	0	0	0	0	0	0	1	-360	360;
	327	8557	0.008929	0.041471	0	0	0	0	0	0	1	-360	360;
	327	3005	0.0222	0.091822	0	0	0	0	0	0	1	-360	360;
	5502	7692	0.005671	0.021169	0	0	0	0	0	0	1	-360	360;
	6919	3005	0.00956	0.035551	0	0	0	0	0	0	1	-360	360;
	3204	12066	0.003169	0.021911	0	0	0	0	0	0	1	-360	360;
	3204	12066	0.00324	0.021471	0	0	0	0	0	0	1	-360	360;
	3204	6744	0.01168	0.055391	0	0	0	0	0	0	1	-360	360;
	3204	6744	0.011902	0.054209	0	0	0	0	0	0	1	-360	360;
	6744	194	0.01908	0.062769	0	0	0	0	0	0	1	-360	360;
	2315	6615	0.001591	0.011098	0	0	0	0	0	0	1	-360	360;
	9598	1833	0.003289	0.019091	0	0	0	0	0	0	1	-360	360;
	9598	2381	0.004169	0.025349	0	0	0	0	0	0	1	-360	360;
	9598	6027	0.00439	0.023909	0	0	0	0	0	0	1	-360	360;
	6915	6230	0.00156	0.01956	0	0	0	0	0	0	1	-360	360;
	6230	10840	0.00029	0.00391	0	0	0	0	0	0	1	-360	360;
	6230	8556	0.00179	0.02522	0	0	0	0	0	0	1	-360	360;
	2105	6230	0.00079	0.00756	0	0	0	0	0	0	1	-360	360;
	6230	11657	0.00027	0.00281	0	0	0	0	0	0	1	-360	360;
	6230	11657	0.00026	0.00286	0	0	0	0	0	0	1	-360	360;
	3205	403	0.000281	0.00213	0	0	0	0	0	0	1	-360	360;
	3205	7360	0.000419	0.00299	0	0	0	0	0	0	1	-360	360;
	5504	7037	0.001659	0.009091	0	0	0	0	0	0	1	-360	360;
	5504	9105	0.00439	0.03324	0	0	0	0	0	0	1	-360	360;
	8722	12834	0.000841	0.00612	0	0	0	0	0	0	1	-360	360;
	8722	290	0.003341	0.022841	0	0	0	0	0	0	1	-360	360;
	8722	5149	0.00581	0.04232	0	0	0	0	0	0	1	-360	360;
	8722	2025	0.00495	0.02919	0	0	0	0	0	0	1	-360	360;
	8722	8144	0.01789	0.07795	0	0	0	0	0	0	1	-360	360;
	8722	7253	0.00186	0.014289	0	0	0	0	0	0	1	-360	360;
	8722	3372	0.002281	0.014401	0	0	0	0	0	0	1	-360	360;
	6617	4257	0.00031	0.00132	0	0	0	0	0	0	1	-360	360;
	6617	10130	0.001349	0.0077	0	0	0	0	0	0	1	-360	360;
	8288	8640	0.00586	0.02574	0	0	0	0	0	0	1	-360	360;
	8288	8801	0.00556	0.020211	0	0	0	0	0	0	1	-360	360;
	7580	12019	0.00026	0.001159	0	0	0	0	0	0	1	-360	360;
	12019	9729	0.000599	0.002419	0	0	0	0	0	0	1	-360	360;
	4522	10455	0.0005	0.00386	0	0	0	0	0	0	1	-360	360;
	4522	8972	0.00107	0.008289	0	0	0	0	0	0	1	-360	360;
	4522	8972	0.001091	0.00811	0	0	0	0	0	0	1	-360	360;
	10866	4522	0.00038	0.00255	0	0	0	0	0	0	1	-360	360;
	7580	9448	0.000271	0.000979	0	0	0	0	0	0	1	-360	360;
	3375	2340	7e-05	0.00076	0	0	0	0	0	0	1	-360	360;
	3375	401	0.000521	0.003901	0	0	0	0	0	0	1	-360	360;
	6915	12143	0.0016	0.01839	0	0	0	0	0	0	1	-360	360;
	8410	1975	0.007651	0.058781	0	0	0	0	0	0	1	-360	360;
	8410	9114	0.00394	0.02182	0	0	0	0	0	0	1	-360	360;
	8410	5778	0.003461	0.03005	0	0	0	0	0	0	1	-360	360;
	8410	6867	0.00337	0.01831	0	0	0	0	0	0	1	-360	360;
	11809	6293	0.000581	0.00194	0	0	0	0	0	0	1	-360	360;
	10837	6293	0.000539	0.001719	0	0	0	0	0	0	1	-360	360;
	1388	1303	0.00249	0.03075	0	0	0	0	0	0	1	-360	360;
	1303	6472	0.0014	0.01409	0	0	0	0	0	0	1	-360	360;
	11503	9648	0.008	0.050031	0	0	0	0	0	0	1	-360	360;
	11503	8144	0.002211	0.01562	0	0	0	0	0	0	1	-360	360;
	12645	5358	0.00281	0.021831	0	0	0	0	0	0	1	-360	360;
	12645	11469	0.000711	0.004919	0	0	0	0	0	0	1	-360	360;
	4198	12948	0.001841	0.00738	0	0	0	0	0	0	1	-360	360;
	4198	5103	0.002849	0.008979	0	0	0	0	0	0	1	-360	360;
	12305	5358	0.00232	0.014099	0	0	0	0	0	0	1	-360	360;
	12305	5358	0.00237	0.0138	0	0	0	0	0	0	1	-360	360;
	12305	10085	0.00081	0.004591	0	0	0	0	0	0	1	-360	360;
	12305	9686	0.002219	0.010469	0	0	0	0	0	0	1	-360	360;
	2317	2672	0.001081	0.00851	0	0	0	0	0	0	1	-360	360;
	2672	11725	0.00087	0.006969	0	0	0	0	0	0	1	-360	360;
	2672	11725	0.00088	0.006831	0	0	0	0	0	0	1	-360	360;
	2672	6828	0.000349	0.00237	0	0	0	0	0	0	1	-360	360;
	1030	9509	0.00232	0.00924	0	0	0	0	0	0	1	-360	360;
	1030	5566	0.00812	0.03669	0	0	0	0	0	0	1	-360	360;
	3116	9822	0.003521	0.02825	0	0	0	0	0	0	1	-360	360;
	3116	6962	0.00264	0.017651	0	0	0	0	0	0	1	-360	360;
	11187	9509	0.00486	0.02149	0	0	0	0	0	0	1	-360	360;
	11187	9886	0.006831	0.034669	0	0	0	0	0	0	1	-360	360;
	11187	10340	0.017211	0.108151	0	0	0	0	0	0	1	-360	360;
	11187	12082	0.00063	0.00475	0	0	0	0	0	0	1	-360	360;
	11187	12082	0.00062	0.004841	0	0	0	0	0	0	1	-360	360;
	1388	3599	0.0009	0.01092	0	0	0	0	0	0	1	-360	360;
	3599	10105	0.00124	0.01096	0	0	0	0	0	0	1	-360	360;
	4478	4248	0.003211	0.00176	0	0	0	0	0	0	1	-360	360;
	4478	11469	0.001521	0.008031	0	0	0	0	0	0	1	-360	360;
	7896	2714	0.0005	0.00441	0	0	0	0	0	0	1	-360	360;
	7896	2130	0.00247	0.02573	0	0	0	0	0	0	1	-360	360;
	7896	1810	0.00042	0.00367	0	0	0	0	0	0	1	-360	360;
	10888	6271	0.0023	0.01644	0	0	0	0	0	0	1	-360	360;
	10888	6271	0.00226	0.01676	0	0	0	0	0	0	1	-360	360;
	10888	6271	0.0023	0.0162	0	0	0	0	0	0	1	-360	360;
	10888	6271	0.00225	0.015	0	0	0	0	0	0	1	-360	360;
	3354	8661	0.007281	0.035729	0	0	0	0	0	0	1	-360	360;
	3191	8661	0.001919	0.00945	0	0	0	0	0	0	1	-360	360;
	5604	1680	0.005211	0.02981	0	0	0	0	0	0	1	-360	360;
	5604	10130	0.000659	0.00586	0	0	0	0	0	0	1	-360	360;
	5604	10130	0.000591	0.00664	0	0	0	0	0	0	1	-360	360;
	5604	2537	0.011969	0.07424	0	0	0	0	0	0	1	-360	360;
	2317	4418	0.000531	0.003919	0	0	0	0	0	0	1	-360	360;
	6059	4929	0.001521	0.00718	0	0	0	0	0	0	1	-360	360;
	6059	4929	0.001211	0.00613	0	0	0	0	0	0	1	-360	360;
	6059	7041	0.00676	0.03725	0	0	0	0	0	0	1	-360	360;
	6059	10176	0.00399	0.018091	0	0	0	0	0	0	1	-360	360;
	9739	7610	0.000341	0.00339	0	0	0	0	0	0	1	-360	360;
	9739	6027	0.001271	0.0102	0	0	0	0	0	0	1	-360	360;
	9739	6027	0.00125	0.010401	0	0	0	0	0	0	1	-360	360;
	7118	745	0.00261	0.016651	0	0	0	0	0	0	1	-360	360;
	13552	745	0.000219	0.002409	0	0	0	0	0	0	1	-360	360;
	9886	6960	0.00601	0.02814	0	0	0	0	0	0	1	-360	360;
	6960	10230	0.008969	0.03787	0	0	0	0	0	0	1	-360	360;
	6960	6294	0.00232	0.01094	0	0	0	0	0	0	1	-360	360;
	1388	10105	0.00196	0.02382	0	0	0	0	0	0	1	-360	360;
	1388	10506	0.00422	0.03598	0	0	0	0	0	0	1	-360	360;
	1388	10506	0.0043	0.03522	0	0	0	0	0	0	1	-360	360;
	12544	1388	0.00208	0.01691	0	0	0	0	0	0	1	-360	360;
	1388	3905	0.00544	0.05439	0	0	0	0	0	0	1	-360	360;
	1388	205	0.00239	0.02598	0	0	0	0	0	0	1	-360	360;
	1388	570	0.00396	0.03096	0	0	0	0	0	0	1	-360	360;
	1388	570	0.00353	0.03509	0	0	0	0	0	0	1	-360	360;
	5387	13117	0.000531	0.003729	0	0	0	0	0	0	1	-360	360;
	5387	13117	0.00038	0.0025	0	0	0	0	0	0	1	-360	360;
	5387	13117	0.00039	0.00244	0	0	0	0	0	0	1	-360	360;
	5387	2029	0.00836	0.044521	0	0	0	0	0	0	1	-360	360;
	2306	1524	0.00264	0.03029	0	0	0	0	0	0	1	-360	360;
	2306	9878	0.00396	0.04589	0	0	0	0	0	0	1	-360	360;
	12613	9746	0.00411	0.021729	0	0	0	0	0	0	1	-360	360;
	12613	10844	0.00607	0.0285	0	0	0	0	0	0	1	-360	360;
	11948	2940	0.00075	0.0047	0	0	0	0	0	0	1	-360	360;
	6470	2940	0.00114	0.007969	0	0	0	0	0	0	1	-360	360;
	4530	4189	0.00036	0.00324	0	0	0	0	0	0	1	-360	360;
	4530	13271	0.00094	0.00874	0	0	0	0	0	0	1	-360	360;
	3545	10628	0.00949	0.05145	0	0	0	0	0	0	1	-360	360;
	3545	7193	0.011719	0.0337	0	0	0	0	0	0	1	-360	360;
	4446	3545	0.0022	0.012331	0	0	0	0	0	0	1	-360	360;
	6137	6365	0.00038	0.00448	0	0	0	0	0	0	1	-360	360;
	6137	8556	0.00098	0.01126	0	0	0	0	0	0	1	-360	360;
	53	8622	0.00055	0.00621	0	0	0	0	0	0	1	-360	360;
	53	2965	0.00163	0.01287	0	0	0	0	0	0	1	-360	360;
	9185	13349	0.0018	0.01886	0	0	0	0	0	0	1	-360	360;
	9185	13349	0.0047	0.024969	0	0	0	0	0	0	1	-360	360;
	9185	3807	0.009711	0.059669	0	0	0	0	0	0	1	-360	360;
	9185	7582	0.005271	0.03451	0	0	0	0	0	0	1	-360	360;
	1301	1682	0.016169	0.11936	0	0	0	0	0	0	1	-360	360;
	1301	1908	0.00226	0.014789	0	0	0	0	0	0	1	-360	360;
	12834	2015	0.00151	0.01081	0	0	0	0	0	0	1	-360	360;
	1029	6829	0.00228	0.0217	0	0	0	0	0	0	1	-360	360;
	1029	5263	0.00137	0.01593	0	0	0	0	0	0	1	-360	360;
	12004	5573	0.00099	0.0121	0	0	0	0	0	0	1	-360	360;
	12004	3230	0	0.01881	0	0	0	0	0	0	1	-360	360;
	12004	456	0.00143	0.0163	0	0	0	0	0	0	1	-360	360;
	12004	10680	0.00057	0.00562	0	0	0	0	0	0	1	-360	360;
	12004	10680	0.00055	0.00572	0	0	0	0	0	0	1	-360	360;
	12004	9505	0.00057	0.00658	0	0	0	0	0	0	1	-360	360;
	12004	7826	0.00075	0.0082	0	0	0	0	0	0	1	-360	360;
	12004	1156	0.00253	0.02153	0	0	0	0	0	0	1	-360	360;
	290	12668	0.00236	0.011219	0	0	0	0	0	0	1	-360	360;
	4623	7770	0.00173	0.0154	0	0	0	0	0	0	1	-360	360;
	4623	3958	0.00385	0.04172	0	0	0	0	0	0	1	-360	360;
	4623	3958	0.00377	0.04251	0	0	0	0	0	0	1	-360	360;
	4623	2130	0.00199	0.02121	0	0	0	0	0	0	1	-360	360;
	4623	1625	0.00236	0.01933	0	0	0	0	0	0	1	-360	360;
	4623	1625	0.00241	0.01891	0	0	0	0	0	0	1	-360	360;
	3279	5001	0.000521	0.004729	0	0	0	0	0	0	1	-360	360;
	3279	7745	0.00094	0.005581	0	0	0	0	0	0	1	-360	360;
	6821	1567	0.00134	0.01324	0	0	0	0	0	0	1	-360	360;
	6821	4089	0.00073	0.00638	0	0	0	0	0	0	1	-360	360;
	2064	3531	0.00248	0.02774	0	0	0	0	0	0	1	-360	360;
	2064	701	0.00152	0.013	0	0	0	0	0	0	1	-360	360;
	2064	701	0.00155	0.01272	0	0	0	0	0	0	1	-360	360;
	2064	12079	0.00117	0.01301	0	0	0	0	0	0	1	-360	360;
	2064	4503	0.00197	0.02054	0	0	0	0	0	0	1	-360	360;
	2064	13287	0.00157	0.01909	0	0	0	0	0	0	1	-360	360;
	4418	6529	0.00107	0.008669	0	0	0	0	0	0	1	-360	360;
	6529	2134	0.000469	0.00314	0	0	0	0	0	0	1	-360	360;
	1692	1026	0.000409	0.00305	0	0	0	0	0	0	1	-360	360;
	1755	1026	0.00569	0.032979	0	0	0	0	0	0	1	-360	360;
	5124	13135	0.00087	0.006771	0	0	0	0	0	0	1	-360	360;
	5124	12668	0.002151	0.014039	0	0	0	0	0	0	1	-360	360;
	9299	9908	9.9e-05	0.00043	0	0	0	0	0	0	1	-360	360;
	9446	9299	0.000151	0.0005	0	0	0	0	0	0	1	-360	360;
	11501	3189	0.001531	0.00812	0	0	0	0	0	0	1	-360	360;
	11501	3189	0.00113	0.00857	0	0	0	0	0	0	1	-360	360;
	6242	10547	0.00226	0.01868	0	0	0	0	0	0	1	-360	360;
	6242	6828	0.000401	0.00239	0	0	0	0	0	0	1	-360	360;
	6211	2732	0.001169	0.008349	0	0	0	0	0	0	1	-360	360;
	6211	9226	0.00324	0.02312	0	0	0	0	0	0	1	-360	360;
	905	6399	0.006031	0.034789	0	0	0	0	0	0	1	-360	360;
	905	1157	0.0103	0.05269	0	0	0	0	0	0	1	-360	360;
	4268	10105	0.00129	0.01494	0	0	0	0	0	0	1	-360	360;
	4268	10105	0.00126	0.01523	0	0	0	0	0	0	1	-360	360;
	7259	7193	0.001031	0.00601	0	0	0	0	0	0	1	-360	360;
	7259	7193	0.00105	0.00588	0	0	0	0	0	0	1	-360	360;
	3786	8475	0.00197	0.01719	0	0	0	0	0	0	1	-360	360;
	3786	2775	0.00114	0.01237	0	0	0	0	0	0	1	-360	360;
	3786	5833	0.00314	0.02868	0	0	0	0	0	0	1	-360	360;
	3786	1543	0.00294	0.03326	0	0	0	0	0	0	1	-360	360;
	3786	1543	0.00288	0.0308	0	0	0	0	0	0	1	-360	360;
	3786	1156	0.00088	0.02821	0	0	0	0	0	0	1	-360	360;
	6115	9734	0.00152	0.01702	0	0	0	0	0	0	1	-360	360;
	6115	10704	0.00022	0.00209	0	0	0	0	0	0	1	-360	360;
	2557	9993	0.00087	0.005	0	0	0	0	0	0	1	-360	360;
	2557	5778	0.000711	0.00439	0	0	0	0	0	0	1	-360	360;
	2557	5778	0.000729	0.004289	0	0	0	0	0	0	1	-360	360;
	6841	9431	0.001031	0.00549	0	0	0	0	0	0	1	-360	360;
	6841	760	0.003219	0.01512	0	0	0	0	0	0	1	-360	360;
	6841	875	0.007271	0.031419	0	0	0	0	0	0	1	-360	360;
	6841	4803	0.02145	0.085781	0	0	0	0	0	0	1	-360	360;
	8978	10984	0.002841	0.017771	0	0	0	0	0	0	1	-360	360;
	8978	4149	0.003331	0.017289	0	0	0	0	0	0	1	-360	360;
	2059	3354	0.01524	0.064091	0	0	0	0	0	0	1	-360	360;
	7404	7479	0.000831	0.0052	0	0	0	0	0	0	1	-360	360;
	7479	340	0.00164	0.0103	0	0	0	0	0	0	1	-360	360;
	5823	6224	0.00139	0.00864	0	0	0	0	0	0	1	-360	360;
	5823	9844	0.00143	0.00764	0	0	0	0	0	0	1	-360	360;
	5823	8419	0.000229	0.00126	0	0	0	0	0	0	1	-360	360;
	5823	8419	0.000211	0.00143	0	0	0	0	0	0	1	-360	360;
	1377	2214	0.0008	0.01072	0	0	0	0	0	0	1	-360	360;
	1377	6743	0.00067	0.00661	0	0	0	0	0	0	1	-360	360;
	1377	4003	0.0003	0.00427	0	0	0	0	0	0	1	-360	360;
	1377	4003	0.00068	0.00648	0	0	0	0	0	0	1	-360	360;
	2280	12048	0.0002	0.001289	0	0	0	0	0	0	1	-360	360;
	2280	5895	0.00013	0.000961	0	0	0	0	0	0	1	-360	360;
	8383	6365	0.00118	0.01285	0	0	0	0	0	0	1	-360	360;
	8383	7640	0.00062	0.00577	0	0	0	0	0	0	1	-360	360;
	13625	1996	9.1e-05	0.000289	0	0	0	0	0	0	1	-360	360;
	13625	1996	9.1e-05	0.000271	0	0	0	0	0	0	1	-360	360;
	13625	9649	0.0003	0.002719	0	0	0	0	0	0	1	-360	360;
	13625	9986	0.00049	0.002091	0	0	0	0	0	0	1	-360	360;
	13625	19	0.000159	0.001461	0	0	0	0	0	0	1	-360	360;
	631	8970	0.00301	0.02281	0	0	0	0	0	0	1	-360	360;
	4323	8970	0.0027	0.019979	0	0	0	0	0	0	1	-360	360;
	2036	946	0.00019	0.000539	0	0	0	0	0	0	1	-360	360;
	946	9986	0.000169	0.00061	0	0	0	0	0	0	1	-360	360;
	6462	7911	0.00163	0.00939	0	0	0	0	0	0	1	-360	360;
	6462	7911	0.001159	0.009031	0	0	0	0	0	0	1	-360	360;
	6462	5966	9.9e-05	0.00037	0	0	0	0	0	0	1	-360	360;
	8184	1370	0.003789	0.01805	0	0	0	0	0	0	1	-360	360;
	8184	13134	0.00069	0.00389	0	0	0	0	0	0	1	-360	360;
	6312	11645	0.00201	0.01845	0	0	0	0	0	0	1	-360	360;
	9878	11645	0.00121	0.00902	0	0	0	0	0	0	1	-360	360;
	6662	2775	0.00155	0.01573	0	0	0	0	0	0	1	-360	360;
	6662	3864	0.00031	0.00274	0	0	0	0	0	0	1	-360	360;
	6662	5833	0.00047	0.00563	0	0	0	0	0	0	1	-360	360;
	6662	5198	0.00066	0.00721	0	0	0	0	0	0	1	-360	360;
	10992	7037	0.00326	0.016711	0	0	0	0	0	0	1	-360	360;
	10992	4173	0.002479	0.01262	0	0	0	0	0	0	1	-360	360;
	10992	4173	0.00243	0.01286	0	0	0	0	0	0	1	-360	360;
	10992	6367	0.004521	0.02306	0	0	0	0	0	0	1	-360	360;
	10992	1814	0.00211	0.00899	0	0	0	0	0	0	1	-360	360;
	10992	286	0.001289	0.005651	0	0	0	0	0	0	1	-360	360;
	10992	4937	0.002479	0.009021	0	0	0	0	0	0	1	-360	360;
	10992	9448	0.00151	0.010521	0	0	0	0	0	0	1	-360	360;
	7876	5919	0.00101	0.004151	0	0	0	0	0	0	1	-360	360;
	7876	1147	0.015599	0.079	0	0	0	0	0	0	1	-360	360;
	7876	4466	0.013961	0.067591	0	0	0	0	0	0	1	-360	360;
	7876	1756	0.007099	0.043169	0	0	0	0	0	0	1	-360	360;
	3457	7382	0.00317	0.02987	0	0	0	0	0	0	1	-360	360;
	3457	3958	0.00081	0.0088	0	0	0	0	0	0	1	-360	360;
	3457	3958	0.00098	0.00961	0	0	0	0	0	0	1	-360	360;
	3457	257	0.00194	0.02225	0	0	0	0	0	0	1	-360	360;
	13307	7294	0.001031	0.006349	0	0	0	0	0	0	1	-360	360;
	13307	656	0.00668	0.0297	0	0	0	0	0	0	1	-360	360;
	7041	3643	0.00439	0.02236	0	0	0	0	0	0	1	-360	360;
	3643	9362	0.00249	0.01349	0	0	0	0	0	0	1	-360	360;
	2949	6682	0.001969	0.01205	0	0	0	0	0	0	1	-360	360;
	2949	6733	0.000461	0.003031	0	0	0	0	0	0	1	-360	360;
	5272	407	0.00444	0.023479	0	0	0	0	0	0	1	-360	360;
	5272	6790	0.00175	0.01288	0	0	0	0	0	0	1	-360	360;
	5272	6790	0.001711	0.01312	0	0	0	0	0	0	1	-360	360;
	2214	847	0.0008	0.00784	0	0	0	0	0	0	1	-360	360;
	2214	13478	0.00051	0.00612	0	0	0	0	0	0	1	-360	360;
	4789	10464	0.00137	0.00905	0	0	0	0	0	0	1	-360	360;
	4789	7118	0.002771	0.021409	0	0	0	0	0	0	1	-360	360;
	4789	7118	0.00314	0.020979	0	0	0	0	0	0	1	-360	360;
	4789	6026	0.002	0.01118	0	0	0	0	0	0	1	-360	360;
	11809	1018	0.00119	0.00875	0	0	0	0	0	0	1	-360	360;
	11809	10051	0.00031	0.00236	0	0	0	0	0	0	1	-360	360;
	11809	11469	0.000831	0.00539	0	0	0	0	0	0	1	-360	360;
	407	10780	0.004349	0.02394	0	0	0	0	0	0	1	-360	360;
	1709	12300	0.0007	0.00782	0	0	0	0	0	0	1	-360	360;
	1709	2449	0.00046	0.00432	0	0	0	0	0	0	1	-360	360;
	1709	4420	0.00049	0.00617	0	0	0	0	0	0	1	-360	360;
	1709	1543	0.00127	0.01514	0	0	0	0	0	0	1	-360	360;
	5206	2731	0.00232	0.01843	0	0	0	0	0	0	1	-360	360;
	5206	2140	0.00488	0.032151	0	0	0	0	0	0	1	-360	360;
	1996	5285	0.000229	0.00136	0	0	0	0	0	0	1	-360	360;
	1996	9986	0.0003	0.001531	0	0	0	0	0	0	1	-360	360;
	1996	4544	0.00026	0.00126	0	0	0	0	0	0	1	-360	360;
	1996	8663	0.00056	0.002091	0	0	0	0	0	0	1	-360	360;
	6472	4268	0.00103	0.0093	0	0	0	0	0	0	1	-360	360;
	6472	3492	0.00183	0.01689	0	0	0	0	0	0	1	-360	360;
	10312	7542	0.00324	0.01095	0	0	0	0	0	0	1	-360	360;
	352	12118	0.001419	0.008539	0	0	0	0	0	0	1	-360	360;
	352	5001	0.00068	0.00438	0	0	0	0	0	0	1	-360	360;
	4189	11083	0.00178	0.01651	0	0	0	0	0	0	1	-360	360;
	4189	6322	0.00151	0.01827	0	0	0	0	0	0	1	-360	360;
	4189	6322	0.00148	0.01862	0	0	0	0	0	0	1	-360	360;
	4268	3492	0.00201	0.02546	0	0	0	0	0	0	1	-360	360;
	4268	3492	0.00197	0.02357	0	0	0	0	0	0	1	-360	360;
	5507	5001	0.001091	0.00775	0	0	0	0	0	0	1	-360	360;
	5507	1399	0.000841	0.006669	0	0	0	0	0	0	1	-360	360;
	1311	5766	0.00151	0.012081	0	0	0	0	0	0	1	-360	360;
	1311	10929	0.00032	0.002521	0	0	0	0	0	0	1	-360	360;
	3461	2125	0.006091	0.03019	0	0	0	0	0	0	1	-360	360;
	3461	11213	0.003651	0.016461	0	0	0	0	0	0	1	-360	360;
	3461	4865	0.00076	0.00425	0	0	0	0	0	0	1	-360	360;
	3461	4865	0.00087	0.004031	0	0	0	0	0	0	1	-360	360;
	2343	5766	0.000651	0.004729	0	0	0	0	0	0	1	-360	360;
	2343	2504	0.00045	0.00337	0	0	0	0	0	0	1	-360	360;
	4173	706	0.001539	0.00839	0	0	0	0	0	0	1	-360	360;
	4173	706	0.001919	0.010479	0	0	0	0	0	0	1	-360	360;
	706	12040	0.001169	0.007031	0	0	0	0	0	0	1	-360	360;
	2707	706	0.00739	0.034081	0	0	0	0	0	0	1	-360	360;
	7171	706	0.00164	0.00964	0	0	0	0	0	0	1	-360	360;
	8585	706	0.00206	0.012	0	0	0	0	0	0	1	-360	360;
	7854	706	0.000961	0.00531	0	0	0	0	0	0	1	-360	360;
	5198	9891	6e-05	0.00029	0	0	0	0	0	0	1	-360	360;
	1908	7037	0.001919	0.013151	0	0	0	0	0	0	1	-360	360;
	1227	7037	0.00019	0.00107	0	0	0	0	0	0	1	-360	360;
	1227	7037	0.00018	0.001091	0	0	0	0	0	0	1	-360	360;
	8272	3531	0.00147	0.01714	0	0	0	0	0	0	1	-360	360;
	3905	3531	0.00143	0.01632	0	0	0	0	0	0	1	-360	360;
	3905	3531	0.00142	0.01416	0	0	0	0	0	0	1	-360	360;
	3905	3531	0.00145	0.01386	0	0	0	0	0	0	1	-360	360;
	3531	13287	0.00073	0.00732	0	0	0	0	0	0	1	-360	360;
	1692	1755	0.009039	0.05989	0	0	0	0	0	0	1	-360	360;
	1692	8177	0.007919	0.04975	0	0	0	0	0	0	1	-360	360;
	9993	1386	0.000521	0.005841	0	0	0	0	0	0	1	-360	360;
	9993	1591	0.000521	0.0028	0	0	0	0	0	0	1	-360	360;
	9603	1543	0.0007	0.00744	0	0	0	0	0	0	1	-360	360;
	9603	5915	0.00159	0.0148	0	0	0	0	0	0	1	-360	360;
	253	6238	0.000659	0.005151	0	0	0	0	0	0	1	-360	360;
	10547	6238	0.000729	0.004901	0	0	0	0	0	0	1	-360	360;
	6743	5854	0.00064	0.00772	0	0	0	0	0	0	1	-360	360;
	11069	9042	0.00018	0.00093	0	0	0	0	0	0	1	-360	360;
	11069	12048	0.00019	0.00118	0	0	0	0	0	0	1	-360	360;
	10866	11069	0.000409	0.002341	0	0	0	0	0	0	1	-360	360;
	9175	1460	0.00087	0.006591	0	0	0	0	0	0	1	-360	360;
	9175	4415	0.00162	0.0098	0	0	0	0	0	0	1	-360	360;
	205	7770	0.00096	0.00849	0	0	0	0	0	0	1	-360	360;
	9039	10126	0.000341	0.001081	0	0	0	0	0	0	1	-360	360;
	9039	10126	0.0003	0.001219	0	0	0	0	0	0	1	-360	360;
	9039	11868	9.9e-05	0.00076	0	0	0	0	0	0	1	-360	360;
	9039	11868	0.00011	0.00074	0	0	0	0	0	0	1	-360	360;
	2288	7118	0.00151	0.009039	0	0	0	0	0	0	1	-360	360;
	2288	8982	0.00356	0.02331	0	0	0	0	0	0	1	-360	360;
	11764	2036	0.000281	0.00169	0	0	0	0	0	0	1	-360	360;
	11764	9649	0.000169	0.001531	0	0	0	0	0	0	1	-360	360;
	11764	12820	0.00038	0.00326	0	0	0	0	0	0	1	-360	360;
	1567	9646	0.00136	0.01738	0	0	0	0	0	0	1	-360	360;
	9431	760	0.00186	0.0133	0	0	0	0	0	0	1	-360	360;
	9431	760	0.00182	0.01356	0	0	0	0	0	0	1	-360	360;
	2630	9431	0.002531	0.015919	0	0	0	0	0	0	1	-360	360;
	9431	9858	0.00289	0.015969	0	0	0	0	0	0	1	-360	360;
	1612	2943	0.001771	0.01107	0	0	0	0	0	0	1	-360	360;
	6177	2943	0.007969	0.04238	0	0	0	0	0	0	1	-360	360;
	4173	297	0.00013	0.00062	0	0	0	0	0	0	1	-360	360;
	4173	427	0.0002	0.00201	0	0	0	0	0	0	1	-360	360;
	701	13545	0.00095	0.01024	0	0	0	0	0	0	1	-360	360;
	1625	13545	0.00553	0.04909	0	0	0	0	0	0	1	-360	360;
	4596	9509	0.009229	0.03693	0	0	0	0	0	0	1	-360	360;
	5566	9509	0.00451	0.028469	0	0	0	0	0	0	1	-360	360;
	9509	8799	0.003591	0.01906	0	0	0	0	0	0	1	-360	360;
	9509	10844	0.010159	0.05169	0	0	0	0	0	0	1	-360	360;
	112	3007	0.00039	0.003099	0	0	0	0	0	0	1	-360	360;
	112	10032	0.000729	0.0048	0	0	0	0	0	0	1	-360	360;
	10450	9734	0.00141	0.01296	0	0	0	0	0	0	1	-360	360;
	5419	5041	0.001969	0.019229	0	0	0	0	0	0	1	-360	360;
	5419	13240	0.001081	0.0105	0	0	0	0	0	0	1	-360	360;
	5419	13240	0.00105	0.010711	0	0	0	0	0	0	1	-360	360;
	8272	4089	0.00052	0.00449	0	0	0	0	0	0	1	-360	360;
	4929	9362	0.01024	0.056349	0	0	0	0	0	0	1	-360	360;
	4929	2171	0.004599	0.023659	0	0	0	0	0	0	1	-360	360;
	4929	4705	0.010021	0.057409	0	0	0	0	0	0	1	-360	360;
	4929	4803	0.00275	0.01525	0	0	0	0	0	0	1	-360	360;
	1009	4869	0.00107	0.01122	0	0	0	0	0	0	1	-360	360;
	1009	11083	0.00242	0.02336	0	0	0	0	0	0	1	-360	360;
	1009	5263	0.00165	0.01303	0	0	0	0	0	0	1	-360	360;
	1009	5263	0.00155	0.01332	0	0	0	0	0	0	1	-360	360;
	10220	8692	0.00087	0.005469	0	0	0	0	0	0	1	-360	360;
	10220	8692	0.002281	0.011909	0	0	0	0	0	0	1	-360	360;
	10220	1368	0.00106	0.01076	0	0	0	0	0	0	1	-360	360;
	10220	7790	0.001409	0.01618	0	0	0	0	0	0	1	-360	360;
	10220	3480	0.00306	0.015521	0	0	0	0	0	0	1	-360	360;
	5646	7404	0.000969	0.006831	0	0	0	0	0	0	1	-360	360;
	5646	340	0.002531	0.016961	0	0	0	0	0	0	1	-360	360;
	1457	390	0.01207	0.077901	0	0	0	0	0	0	1	-360	360;
	1457	10130	0.014091	0.080581	0	0	0	0	0	0	1	-360	360;
	1457	2537	0.003289	0.01481	0	0	0	0	0	0	1	-360	360;
	1457	245	0.003349	0.0145	0	0	0	0	0	0	1	-360	360;
	1315	5566	0.005651	0.026479	0	0	0	0	0	0	1	-360	360;
	1315	11213	0.01193	0.064669	0	0	0	0	0	0	1	-360	360;
	9746	7463	0.004151	0.017281	0	0	0	0	0	0	1	-360	360;
	8018	10464	0.00213	0.014091	0	0	0	0	0	0	1	-360	360;
	6367	2707	0.00113	0.00651	0	0	0	0	0	0	1	-360	360;
	6367	2707	0.001159	0.00637	0	0	0	0	0	0	1	-360	360;
	6367	9448	0.00611	0.03288	0	0	0	0	0	0	1	-360	360;
	13233	1203	0.00012	0.00156	0	0	0	0	0	0	1	-360	360;
	13233	1203	0.00013	0.00153	0	0	0	0	0	0	1	-360	360;
	13233	9883	0.00285	0.03071	0	0	0	0	0	0	1	-360	360;
	8692	4149	0.00676	0.03787	0	0	0	0	0	0	1	-360	360;
	8692	13367	0.004461	0.023521	0	0	0	0	0	0	1	-360	360;
	6340	900	0.00064	0.0078	0	0	0	0	0	0	1	-360	360;
	6340	900	0.00065	0.00764	0	0	0	0	0	0	1	-360	360;
	6340	847	0.00066	0.00602	0	0	0	0	0	0	1	-360	360;
	6340	6519	0.00149	0.01224	0	0	0	0	0	0	1	-360	360;
	6340	1163	0.00061	0.00582	0	0	0	0	0	0	1	-360	360;
	5358	10085	0.001461	0.00744	0	0	0	0	0	0	1	-360	360;
	5358	11469	0.00386	0.02406	0	0	0	0	0	0	1	-360	360;
	5358	13224	0.00306	0.01555	0	0	0	0	0	0	1	-360	360;
	5358	13224	0.003	0.01586	0	0	0	0	0	0	1	-360	360;
	5349	6045	0.003271	0.020651	0	0	0	0	0	0	1	-360	360;
	6045	11168	0.000849	0.00657	0	0	0	0	0	0	1	-360	360;
	2352	4419	0.000211	0.001169	0	0	0	0	0	0	1	-360	360;
	2352	11725	0.000469	0.00345	0	0	0	0	0	0	1	-360	360;
	7241	205	0.00126	0.0145	0	0	0	0	0	0	1	-360	360;
	2794	8640	0.008219	0.04925	0	0	0	0	0	0	1	-360	360;
	2794	11738	0.006771	0.04132	0	0	0	0	0	0	1	-360	360;
	10870	11116	0.000841	0.004539	0	0	0	0	0	0	1	-360	360;
	10870	5566	0.00701	0.03161	0	0	0	0	0	0	1	-360	360;
	10870	6294	0.000781	0.004271	0	0	0	0	0	0	1	-360	360;
	10870	1358	0.00049	0.0032	0	0	0	0	0	0	1	-360	360;
	5107	5595	0.000331	0.00199	0	0	0	0	0	0	1	-360	360;
	1911	5595	0.00032	0.002031	0	0	0	0	0	0	1	-360	360;
	5595	8005	9.1e-05	0.00057	0	0	0	0	0	0	1	-360	360;
	7813	1227	0.001651	0.006789	0	0	0	0	0	0	1	-360	360;
	7813	4937	0.001461	0.007229	0	0	0	0	0	0	1	-360	360;
	5919	13367	0.00437	0.023979	0	0	0	0	0	0	1	-360	360;
	3046	696	0.00781	0.04595	0	0	0	0	0	0	1	-360	360;
	3046	9114	0.0015	0.006531	0	0	0	0	0	0	1	-360	360;
	8377	12040	0.001031	0.006469	0	0	0	0	0	0	1	-360	360;
	4003	4412	0.00183	0.01686	0	0	0	0	0	0	1	-360	360;
	4003	900	0.0002	0.00294	0	0	0	0	0	0	1	-360	360;
	4003	900	0.00021	0.00288	0	0	0	0	0	0	1	-360	360;
	4003	7382	0.00171	0.01889	0	0	0	0	0	0	1	-360	360;
	10506	3958	0.00069	0.00565	0	0	0	0	0	0	1	-360	360;
	10506	3958	0.0007	0.00553	0	0	0	0	0	0	1	-360	360;
	10371	5568	0.00011	0.00115	0	0	0	0	0	0	1	-360	360;
	10371	9773	0.00012	0.00126	0	0	0	0	0	0	1	-360	360;
	10371	13271	0.00326	0.02919	0	0	0	0	0	0	1	-360	360;
	10371	7407	0.00092	0.00876	0	0	0	0	0	0	1	-360	360;
	8401	10126	0.00087	0.005409	0	0	0	0	0	0	1	-360	360;
	8401	10032	0.00131	0.00701	0	0	0	0	0	0	1	-360	360;
	10578	9731	0.002531	0.018781	0	0	0	0	0	0	1	-360	360;
	10578	12923	0.00894	0.056909	0	0	0	0	0	0	1	-360	360;
	8019	7294	0.00187	0.010659	0	0	0	0	0	0	1	-360	360;
	11267	7294	0.00286	0.01538	0	0	0	0	0	0	1	-360	360;
	7294	7271	0.0008	0.00444	0	0	0	0	0	0	1	-360	360;
	1443	11566	0.00545	0.03936	0	0	0	0	0	0	1	-360	360;
	10886	11566	0.00119	0.009031	0	0	0	0	0	0	1	-360	360;
	64	8256	0.00295	0.0162	0	0	0	0	0	0	1	-360	360;
	64	9025	0.001581	0.0087	0	0	0	0	0	0	1	-360	360;
	64	3615	0.007729	0.025401	0	0	0	0	0	0	1	-360	360;
	11438	2907	0.00013	0.001461	0	0	0	0	0	0	1	-360	360;
	11438	8565	0.000341	0.02219	0	0	0	0	0	0	1	-360	360;
	11438	5895	0.00018	0.001211	0	0	0	0	0	0	1	-360	360;
	11438	11776	0.003099	0.0147	0	0	0	0	0	0	1	-360	360;
	11438	11776	0.003159	0.014401	0	0	0	0	0	0	1	-360	360;
	10126	6266	0.00281	0.01305	0	0	0	0	0	0	1	-360	360;
	10126	13135	0.002281	0.01075	0	0	0	0	0	0	1	-360	360;
	10126	13135	0.00232	0.010521	0	0	0	0	0	0	1	-360	360;
	10126	5021	0.000151	0.000841	0	0	0	0	0	0	1	-360	360;
	10157	3807	0.002771	0.012091	0	0	0	0	0	0	1	-360	360;
	10157	8144	0.00613	0.03005	0	0	0	0	0	0	1	-360	360;
	2125	11213	0.003219	0.02486	0	0	0	0	0	0	1	-360	360;
	9042	11868	0.000211	0.001331	0	0	0	0	0	0	1	-360	360;
	12544	2714	0.00185	0.01917	0	0	0	0	0	0	1	-360	360;
	1814	1227	0.00082	0.00381	0	0	0	0	0	0	1	-360	360;
	11878	11466	0.00075	0.004531	0	0	0	0	0	0	1	-360	360;
	11878	11466	0.000729	0.00419	0	0	0	0	0	0	1	-360	360;
	11878	40	0.02626	0.15576	0	0	0	0	0	0	1	-360	360;
	11878	10878	0.011469	0.03937	0	0	0	0	0	0	1	-360	360;
	10455	12048	0.00011	0.00061	0	0	0	0	0	0	1	-360	360;
	10455	1530	0.000531	0.00313	0	0	0	0	0	0	1	-360	360;
	10455	8972	0.001211	0.00864	0	0	0	0	0	0	1	-360	360;
	253	6828	0.00094	0.006289	0	0	0	0	0	0	1	-360	360;
	1060	5909	0.001	0.00886	0	0	0	0	0	0	1	-360	360;
	9886	4596	0.001979	0.010469	0	0	0	0	0	0	1	-360	360;
	9886	4596	0.0017	0.00855	0	0	0	0	0	0	1	-360	360;
	4596	2997	0.00757	0.04624	0	0	0	0	0	0	1	-360	360;
	4596	5795	0.001719	0.00794	0	0	0	0	0	0	1	-360	360;
	4596	5795	0.00176	0.007781	0	0	0	0	0	0	1	-360	360;
	4566	1285	0.00126	0.0053	0	0	0	0	0	0	1	-360	360;
	4566	1756	0.00444	0.022789	0	0	0	0	0	0	1	-360	360;
	701	3492	0.00163	0.0145	0	0	0	0	0	0	1	-360	360;
	9886	390	0.003281	0.02124	0	0	0	0	0	0	1	-360	360;
	9886	5795	0.004979	0.025271	0	0	0	0	0	0	1	-360	360;
	13117	3441	0.0072	0.055289	0	0	0	0	0	0	1	-360	360;
	13117	6993	0.00787	0.04582	0	0	0	0	0	0	1	-360	360;
	13117	8177	0.003961	0.017711	0	0	0	0	0	0	1	-360	360;
	3441	5536	0.00756	0.04826	0	0	0	0	0	0	1	-360	360;
	288	4466	0.003401	0.02086	0	0	0	0	0	0	1	-360	360;
	288	5796	0.001341	0.00669	0	0	0	0	0	0	1	-360	360;
	11243	13129	6e-05	0.000581	0	0	0	0	0	0	1	-360	360;
	11243	8663	9.1e-05	0.00074	0	0	0	0	0	0	1	-360	360;
	10837	11031	0.000531	0.00175	0	0	0	0	0	0	1	-360	360;
	1602	7446	0.00116	0.01159	0	0	0	0	0	0	1	-360	360;
	1602	6322	0.00038	0.00457	0	0	0	0	0	0	1	-360	360;
	4419	11725	0.000729	0.00405	0	0	0	0	0	0	1	-360	360;
	403	10321	0.000331	0.001919	0	0	0	0	0	0	1	-360	360;
	5878	6900	0.00086	0.01029	0	0	0	0	0	0	1	-360	360;
	10884	5878	0.00027	0.00332	0	0	0	0	0	0	1	-360	360;
	4948	8268	0.00012	0.00156	0	0	0	0	0	0	1	-360	360;
	4948	8268	0.00012	0.001531	0	0	0	0	0	0	1	-360	360;
	12954	2923	0.00256	0.014211	0	0	0	0	0	0	1	-360	360;
	12954	6470	0.00049	0.00338	0	0	0	0	0	0	1	-360	360;
	8752	2025	0.007531	0.04444	0	0	0	0	0	0	1	-360	360;
	8752	8144	0.007409	0.03226	0	0	0	0	0	0	1	-360	360;
	4557	11983	0.001789	0.012841	0	0	0	0	0	0	1	-360	360;
	4557	6027	0.001831	0.01257	0	0	0	0	0	0	1	-360	360;
	2327	5336	0.000169	0.00143	0	0	0	0	0	0	1	-360	360;
	2327	10032	0.00088	0.005099	0	0	0	0	0	0	1	-360	360;
	4412	10924	0.0004	0.00479	0	0	0	0	0	0	1	-360	360;
	4412	10924	0.00052	0.00481	0	0	0	0	0	0	1	-360	360;
	4412	2052	0.00135	0.01431	0	0	0	0	0	0	1	-360	360;
	4412	1163	0.00051	0.00549	0	0	0	0	0	0	1	-360	360;
	1018	10051	0.00105	0.007969	0	0	0	0	0	0	1	-360	360;
	1018	1334	0.00339	0.020531	0	0	0	0	0	0	1	-360	360;
	1018	4248	0.00169	0.009331	0	0	0	0	0	0	1	-360	360;
	5258	6131	0.00374	0.01905	0	0	0	0	0	0	1	-360	360;
	5258	7695	0.00487	0.026341	0	0	0	0	0	0	1	-360	360;
	3801	12225	0.00057	0.004581	0	0	0	0	0	0	1	-360	360;
	3801	12668	0.000409	0.00301	0	0	0	0	0	0	1	-360	360;
	2068	5516	9.1e-05	0.00076	0	0	0	0	0	0	1	-360	360;
	2068	9031	0.003919	0.01682	0	0	0	0	0	0	1	-360	360;
	8475	1524	0.00117	0.0131	0	0	0	0	0	0	1	-360	360;
	2059	7254	0.008271	0.02787	0	0	0	0	0	0	1	-360	360;
	2059	1612	0.016599	0.0572	0	0	0	0	0	0	1	-360	360;
	2059	1468	0.00999	0.057	0	0	0	0	0	0	1	-360	360;
	2059	1468	0.01132	0.05586	0	0	0	0	0	0	1	-360	360;
	10704	6878	4e-05	0.00045	0	0	0	0	0	0	1	-360	360;
	7254	7404	0.003281	0.01038	0	0	0	0	0	0	1	-360	360;
	6324	108	0.003711	0.01725	0	0	0	0	0	0	1	-360	360;
	6324	9264	0.00069	0.00232	0	0	0	0	0	0	1	-360	360;
	11468	2945	0.004151	0.02436	0	0	0	0	0	0	1	-360	360;
	7542	2945	0.006781	0.028729	0	0	0	0	0	0	1	-360	360;
	5107	10032	0.000331	0.00151	0	0	0	0	0	0	1	-360	360;
	1911	10032	0.000331	0.001479	0	0	0	0	0	0	1	-360	360;
	5433	12810	0.0001	0.00144	0	0	0	0	0	0	1	-360	360;
	5433	8622	0.00027	0.00221	0	0	0	0	0	0	1	-360	360;
	1743	6727	0.0001	0.00141	0	0	0	0	0	0	1	-360	360;
	1743	3958	0.00054	0.00546	0	0	0	0	0	0	1	-360	360;
	10884	12810	7e-05	0.00053	0	0	0	0	0	0	1	-360	360;
	6727	10884	6e-05	0.0006	0	0	0	0	0	0	1	-360	360;
	11031	6419	0.000341	0.00245	0	0	0	0	0	0	1	-360	360;
	11031	11469	0.000841	0.006719	0	0	0	0	0	0	1	-360	360;
	11718	297	0.000401	0.00164	0	0	0	0	0	0	1	-360	360;
	760	10780	0.004409	0.032531	0	0	0	0	0	0	1	-360	360;
	631	1617	0.001961	0.013581	0	0	0	0	0	0	1	-360	360;
	2707	7171	0.006659	0.03437	0	0	0	0	0	0	1	-360	360;
	5970	2279	9.1e-05	0.0007	0	0	0	0	0	0	1	-360	360;
	5970	3379	0.00051	0.004211	0	0	0	0	0	0	1	-360	360;
	9991	2504	0.00043	0.0027	0	0	0	0	0	0	1	-360	360;
	9991	3684	0.00038	0.00306	0	0	0	0	0	0	1	-360	360;
	12892	2120	0.001461	0.009479	0	0	0	0	0	0	1	-360	360;
	12892	10856	0.001841	0.011081	0	0	0	0	0	0	1	-360	360;
	1368	10984	0.00094	0.007169	0	0	0	0	0	0	1	-360	360;
	8800	11828	0.00082	0.00846	0	0	0	0	0	0	1	-360	360;
	8800	10884	0.00029	0.00325	0	0	0	0	0	0	1	-360	360;
	11725	5863	0.000781	0.0062	0	0	0	0	0	0	1	-360	360;
	11725	3928	8.1e-05	0.00187	0	0	0	0	0	0	1	-360	360;
	11725	3275	8.1e-05	0.001909	0	0	0	0	0	0	1	-360	360;
	10044	1524	0.00113	0.00956	0	0	0	0	0	0	1	-360	360;
	10044	1524	0.00115	0.00935	0	0	0	0	0	0	1	-360	360;
	13279	6501	0.00161	0.02036	0	0	0	0	0	0	1	-360	360;
	13279	6501	0.00157	0.02075	0	0	0	0	0	0	1	-360	360;
	10990	5766	0.00101	0.007401	0	0	0	0	0	0	1	-360	360;
	10990	10929	0.00087	0.00711	0	0	0	0	0	0	1	-360	360;
	5256	8622	0.00135	0.01442	0	0	0	0	0	0	1	-360	360;
	5256	2965	0.00176	0.01441	0	0	0	0	0	0	1	-360	360;
	2630	10780	0.00682	0.03187	0	0	0	0	0	0	1	-360	360;
	11828	1982	0.00109	0.00987	0	0	0	0	0	0	1	-360	360;
	6829	11828	0.00273	0.02829	0	0	0	0	0	0	1	-360	360;
	2503	4414	0.001581	0.011289	0	0	0	0	0	0	1	-360	360;
	2503	5001	0.001669	0.013099	0	0	0	0	0	0	1	-360	360;
	2541	2279	9.9e-05	0.00069	0	0	0	0	0	0	1	-360	360;
	2541	3379	9.9e-05	0.000669	0	0	0	0	0	0	1	-360	360;
	7180	5492	0.000591	0.00382	0	0	0	0	0	0	1	-360	360;
	7180	13224	0.00524	0.02276	0	0	0	0	0	0	1	-360	360;
	5492	9686	0.002719	0.01143	0	0	0	0	0	0	1	-360	360;
	1068	12631	0.00294	0.02744	0	0	0	0	0	0	1	-360	360;
	1068	10840	0.0011	0.01339	0	0	0	0	0	0	1	-360	360;
	1068	8556	0.00193	0.01695	0	0	0	0	0	0	1	-360	360;
	1068	7100	0.00134	0.01054	0	0	0	0	0	0	1	-360	360;
	8883	1068	0.00245	0.02064	0	0	0	0	0	0	1	-360	360;
	1975	9822	0.001841	0.00901	0	0	0	0	0	0	1	-360	360;
	1975	9822	0.00187	0.00882	0	0	0	0	0	0	1	-360	360;
	3864	1156	0.00197	0.02312	0	0	0	0	0	0	1	-360	360;
	6365	10840	0.00078	0.00887	0	0	0	0	0	0	1	-360	360;
	12553	7911	0.003771	0.0288	0	0	0	0	0	0	1	-360	360;
	12553	12478	0.000909	0.00737	0	0	0	0	0	0	1	-360	360;
	2052	10924	0.00086	0.00783	0	0	0	0	0	0	1	-360	360;
	2052	10924	0.00095	0.00855	0	0	0	0	0	0	1	-360	360;
	449	10924	0.00012	0.00157	0	0	0	0	0	0	1	-360	360;
	449	10924	0.00013	0.00154	0	0	0	0	0	0	1	-360	360;
	7471	9448	0.00089	0.004771	0	0	0	0	0	0	1	-360	360;
	7471	2038	0.002	0.00881	0	0	0	0	0	0	1	-360	360;
	9362	10340	0.007711	0.035331	0	0	0	0	0	0	1	-360	360;
	10340	10130	0.005919	0.030419	0	0	0	0	0	0	1	-360	360;
	10340	2537	0.0053	0.031719	0	0	0	0	0	0	1	-360	360;
	6266	2808	0.00025	0.001289	0	0	0	0	0	0	1	-360	360;
	2537	2997	0.0062	0.040039	0	0	0	0	0	0	1	-360	360;
	7446	7952	0.0008	0.00707	0	0	0	0	0	0	1	-360	360;
	7446	7258	0.0002	0.00207	0	0	0	0	0	0	1	-360	360;
	7446	6524	0.00036	0.00322	0	0	0	0	0	0	1	-360	360;
	1604	8880	0.0023	0.0157	0	0	0	0	0	0	1	-360	360;
	1604	6177	0.003479	0.021159	0	0	0	0	0	0	1	-360	360;
	5833	5002	0.00107	0.01013	0	0	0	0	0	0	1	-360	360;
	2036	9986	0.00025	0.000909	0	0	0	0	0	0	1	-360	360;
	2036	4544	0.00037	0.00188	0	0	0	0	0	0	1	-360	360;
	13602	5796	0.0003	0.00145	0	0	0	0	0	0	1	-360	360;
	13602	5438	0.001419	0.007539	0	0	0	0	0	0	1	-360	360;
	12631	2395	0.0012	0.0122	0	0	0	0	0	0	1	-360	360;
	12631	3687	0.00397	0.02936	0	0	0	0	0	0	1	-360	360;
	12631	9878	0.00092	0.01112	0	0	0	0	0	0	1	-360	360;
	12631	10884	0.00387	0.0452	0	0	0	0	0	0	1	-360	360;
	12631	1810	0.00432	0.03156	0	0	0	0	0	0	1	-360	360;
	3905	1625	0.00407	0.03937	0	0	0	0	0	0	1	-360	360;
	3905	1625	0.00415	0.03856	0	0	0	0	0	0	1	-360	360;
	13349	7463	0.00869	0.05343	0	0	0	0	0	0	1	-360	360;
	13349	3076	0.0068	0.036961	0	0	0	0	0	0	1	-360	360;
	13349	3807	0.01805	0.095469	0	0	0	0	0	0	1	-360	360;
	11116	5795	0.00045	0.00206	0	0	0	0	0	0	1	-360	360;
	696	10544	0.0095	0.037159	0	0	0	0	0	0	1	-360	360;
	696	7353	0.02106	0.0885	0	0	0	0	0	0	1	-360	360;
	10840	8556	0.00157	0.02318	0	0	0	0	0	0	1	-360	360;
	10840	957	0.00032	0.00282	0	0	0	0	0	0	1	-360	360;
	10840	7223	0.00165	0.01605	0	0	0	0	0	0	1	-360	360;
	7911	719	0.00107	0.00714	0	0	0	0	0	0	1	-360	360;
	7911	719	0.000961	0.00555	0	0	0	0	0	0	1	-360	360;
	7911	12577	0.00074	0.00461	0	0	0	0	0	0	1	-360	360;
	7911	10321	0.00076	0.00451	0	0	0	0	0	0	1	-360	360;
	7911	5966	0.001419	0.01093	0	0	0	0	0	0	1	-360	360;
	7911	12478	0.003729	0.024599	0	0	0	0	0	0	1	-360	360;
	3492	10272	0.00233	0.01992	0	0	0	0	0	0	1	-360	360;
	3492	10272	0.00238	0.01949	0	0	0	0	0	0	1	-360	360;
	3492	7021	0.00217	0.0253	0	0	0	0	0	0	1	-360	360;
	3492	7021	0.00212	0.02579	0	0	0	0	0	0	1	-360	360;
	3492	1625	0.00458	0.04138	0	0	0	0	0	0	1	-360	360;
	5566	807	0.01555	0.089841	0	0	0	0	0	0	1	-360	360;
	9198	5566	0.00274	0.0168	0	0	0	0	0	0	1	-360	360;
	5566	6294	0.005831	0.032211	0	0	0	0	0	0	1	-360	360;
	5566	6136	0.006669	0.031341	0	0	0	0	0	0	1	-360	360;
	11468	7542	0.01081	0.0538	0	0	0	0	0	0	1	-360	360;
	11468	1468	0.005581	0.02457	0	0	0	0	0	0	1	-360	360;
	11468	1468	0.00569	0.024039	0	0	0	0	0	0	1	-360	360;
	1352	10906	0.002081	0.01294	0	0	0	0	0	0	1	-360	360;
	1352	7181	0.001599	0.00919	0	0	0	0	0	0	1	-360	360;
	7640	8556	0.00083	0.00768	0	0	0	0	0	0	1	-360	360;
	3189	11983	0.004979	0.024099	0	0	0	0	0	0	1	-360	360;
	3189	11983	0.005081	0.023591	0	0	0	0	0	0	1	-360	360;
	5568	2395	0.00412	0.03917	0	0	0	0	0	0	1	-360	360;
	2395	7100	0.00086	0.00798	0	0	0	0	0	0	1	-360	360;
	4779	2395	0.00087	0.00961	0	0	0	0	0	0	1	-360	360;
	12143	2395	0.00095	0.01034	0	0	0	0	0	0	1	-360	360;
	6105	2395	0.00211	0.0196	0	0	0	0	0	0	1	-360	360;
	10544	7353	0.00963	0.05281	0	0	0	0	0	0	1	-360	360;
	1529	5489	0.00018	0.001349	0	0	0	0	0	0	1	-360	360;
	1529	9448	0.000961	0.006021	0	0	0	0	0	0	1	-360	360;
	1529	9448	0.000979	0.00589	0	0	0	0	0	0	1	-360	360;
	1617	4415	0.000461	0.002729	0	0	0	0	0	0	1	-360	360;
	1397	12940	0.005039	0.01901	0	0	0	0	0	0	1	-360	360;
	1397	6746	0.00812	0.028919	0	0	0	0	0	0	1	-360	360;
	9646	10183	0.00106	0.01005	0	0	0	0	0	0	1	-360	360;
	8556	9883	0.00074	0.0081	0	0	0	0	0	0	1	-360	360;
	11983	9025	0.012961	0.03605	0	0	0	0	0	0	1	-360	360;
	11983	9025	0.00844	0.0378	0	0	0	0	0	0	1	-360	360;
	11983	7541	0.000591	0.004169	0	0	0	0	0	0	1	-360	360;
	6294	8799	0.003091	0.012289	0	0	0	0	0	0	1	-360	360;
	6682	6733	0.000969	0.0053	0	0	0	0	0	0	1	-360	360;
	3450	3846	0.00161	0.012919	0	0	0	0	0	0	1	-360	360;
	3450	3846	0.003219	0.01795	0	0	0	0	0	0	1	-360	360;
	3450	5377	0.004331	0.025	0	0	0	0	0	0	1	-360	360;
	3450	5377	0.004409	0.02445	0	0	0	0	0	0	1	-360	360;
	7118	8982	0.00425	0.02014	0	0	0	0	0	0	1	-360	360;
	7118	10085	0.009531	0.060979	0	0	0	0	0	0	1	-360	360;
	7118	2713	0.002409	0.0135	0	0	0	0	0	0	1	-360	360;
	1512	12079	0.0006	0.00662	0	0	0	0	0	0	1	-360	360;
	1512	4089	0.00173	0.01701	0	0	0	0	0	0	1	-360	360;
	9362	1680	0.004289	0.018531	0	0	0	0	0	0	1	-360	360;
	9362	10130	0.006919	0.039789	0	0	0	0	0	0	1	-360	360;
	9362	10130	0.007591	0.035539	0	0	0	0	0	0	1	-360	360;
	9362	10176	0.00961	0.05225	0	0	0	0	0	0	1	-360	360;
	5409	2731	0.000581	0.004599	0	0	0	0	0	0	1	-360	360;
	5409	2731	0.00057	0.00469	0	0	0	0	0	0	1	-360	360;
	5409	2731	0.00087	0.003159	0	0	0	0	0	0	1	-360	360;
	2383	7610	0.00018	0.001419	0	0	0	0	0	0	1	-360	360;
	2383	9822	0.00195	0.018841	0	0	0	0	0	0	1	-360	360;
	108	12948	0.00293	0.01293	0	0	0	0	0	0	1	-360	360;
	2171	4705	0.001969	0.010599	0	0	0	0	0	0	1	-360	360;
	10207	1210	3.9e-05	0.000591	0	0	0	0	0	0	1	-360	360;
	10207	608	0.00364	0.02431	0	0	0	0	0	0	1	-360	360;
	10207	2134	0.00114	0.007211	0	0	0	0	0	0	1	-360	360;
	719	6224	0.001271	0.008659	0	0	0	0	0	0	1	-360	360;
	719	9844	0.00105	0.00668	0	0	0	0	0	0	1	-360	360;
	5568	9773	0.00011	0.00122	0	0	0	0	0	0	1	-360	360;
	5568	5263	0.00117	0.0124	0	0	0	0	0	0	1	-360	360;
	5568	6105	0.00168	0.01891	0	0	0	0	0	0	1	-360	360;
	1147	11466	0.00132	0.008281	0	0	0	0	0	0	1	-360	360;
	1147	12350	0.004531	0.01924	0	0	0	0	0	0	1	-360	360;
	1147	1756	0.007841	0.041909	0	0	0	0	0	0	1	-360	360;
	2052	456	0.00113	0.01235	0	0	0	0	0	0	1	-360	360;
	2052	456	0.00128	0.0121	0	0	0	0	0	0	1	-360	360;
	2052	12743	0.0015	0.01792	0	0	0	0	0	0	1	-360	360;
	6733	13134	0.00111	0.005281	0	0	0	0	0	0	1	-360	360;
	6733	13134	0.00113	0.005169	0	0	0	0	0	0	1	-360	360;
	8801	216	0.00563	0.03255	0	0	0	0	0	0	1	-360	360;
	1682	216	0.005289	0.022539	0	0	0	0	0	0	1	-360	360;
	216	9105	0.007159	0.05086	0	0	0	0	0	0	1	-360	360;
	216	5435	0.009669	0.05649	0	0	0	0	0	0	1	-360	360;
	216	9031	0.0092	0.052169	0	0	0	0	0	0	1	-360	360;
	6312	7440	0.00197	0.01594	0	0	0	0	0	0	1	-360	360;
	6312	9878	0.00343	0.02573	0	0	0	0	0	0	1	-360	360;
	6312	12143	0.00151	0.01485	0	0	0	0	0	0	1	-360	360;
	6312	7407	0.00092	0.01042	0	0	0	0	0	0	1	-360	360;
	6312	11782	0.00029	0.00336	0	0	0	0	0	0	1	-360	360;
	2120	6582	0.001711	0.01255	0	0	0	0	0	0	1	-360	360;
	2120	2140	0.005	0.03962	0	0	0	0	0	0	1	-360	360;
	1203	1673	0.0014	0.01687	0	0	0	0	0	0	1	-360	360;
	1203	1673	0.00067	0.00678	0	0	0	0	0	0	1	-360	360;
	1203	1673	0.00065	0.00692	0	0	0	0	0	0	1	-360	360;
	8883	1203	0.00137	0.01524	0	0	0	0	0	0	1	-360	360;
	5642	999	0.00424	0.02376	0	0	0	0	0	0	1	-360	360;
	5642	8469	0.005591	0.031711	0	0	0	0	0	0	1	-360	360;
	5642	3173	0.00138	0.00649	0	0	0	0	0	0	1	-360	360;
	5642	3173	0.00107	0.00544	0	0	0	0	0	0	1	-360	360;
	7440	6448	3e-05	0.00041	0	0	0	0	0	0	1	-360	360;
	7440	6448	3e-05	0.00041	0	0	0	0	0	0	1	-360	360;
	7440	7258	0.00044	0.00342	0	0	0	0	0	0	1	-360	360;
	10906	3595	0.003841	0.01737	0	0	0	0	0	0	1	-360	360;
	10886	2732	0.001771	0.01337	0	0	0	0	0	0	1	-360	360;
	711	3807	0.00139	0.007349	0	0	0	0	0	0	1	-360	360;
	711	8144	0.00625	0.03463	0	0	0	0	0	0	1	-360	360;
	7463	6399	0.00055	0.00255	0	0	0	0	0	0	1	-360	360;
	9648	5149	0.00489	0.043331	0	0	0	0	0	0	1	-360	360;
	12877	4466	0.00145	0.013099	0	0	0	0	0	0	1	-360	360;
	12877	1793	0.004039	0.02881	0	0	0	0	0	0	1	-360	360;
	3191	7271	0.002219	0.00988	0	0	0	0	0	0	1	-360	360;
	7745	4414	0.001919	0.0153	0	0	0	0	0	0	1	-360	360;
	12824	1982	0.00189	0.01319	0	0	0	0	0	0	1	-360	360;
	7592	2504	0.00011	0.000659	0	0	0	0	0	0	1	-360	360;
	7592	3684	0.00011	0.00064	0	0	0	0	0	0	1	-360	360;
	7592	4359	0.000211	0.00111	0	0	0	0	0	0	1	-360	360;
	9986	11718	0.000469	0.001669	0	0	0	0	0	0	1	-360	360;
	8626	1210	6e-05	0.000849	0	0	0	0	0	0	1	-360	360;
	10450	10711	0.00051	0.00597	0	0	0	0	0	0	1	-360	360;
	10450	10183	0.00119	0.01087	0	0	0	0	0	0	1	-360	360;
	8489	10450	0.00219	0.01922	0	0	0	0	0	0	1	-360	360;
	8622	3958	0.00051	0.00465	0	0	0	0	0	0	1	-360	360;
	13135	10031	0.00055	0.0067	0	0	0	0	0	0	1	-360	360;
	3595	1235	0.005099	0.027349	0	0	0	0	0	0	1	-360	360;
	4705	4080	0.000599	0.003961	0	0	0	0	0	0	1	-360	360;
	4705	4803	0.005461	0.032719	0	0	0	0	0	0	1	-360	360;
	11989	4149	0.00349	0.02256	0	0	0	0	0	0	1	-360	360;
	11989	2570	0.00132	0.00861	0	0	0	0	0	0	1	-360	360;
	8484	6898	0.0045	0.01838	0	0	0	0	0	0	1	-360	360;
	8484	1973	0.004289	0.019719	0	0	0	0	0	0	1	-360	360;
	12118	4359	0.001409	0.00881	0	0	0	0	0	0	1	-360	360;
	1673	9663	0.00212	0.01857	0	0	0	0	0	0	1	-360	360;
	1673	11248	0.00049	0.00574	0	0	0	0	0	0	1	-360	360;
	1673	1097	0.00184	0.021	0	0	0	0	0	0	1	-360	360;
	3846	6995	0.00593	0.037409	0	0	0	0	0	0	1	-360	360;
	3173	6131	0.00514	0.024169	0	0	0	0	0	0	1	-360	360;
	5573	4184	0.00041	0.00492	0	0	0	0	0	0	1	-360	360;
	5198	4184	0.00296	0.02677	0	0	0	0	0	0	1	-360	360;
	7193	807	0.0072	0.03406	0	0	0	0	0	0	1	-360	360;
	5001	10150	0.00043	0.00276	0	0	0	0	0	0	1	-360	360;
	5001	11776	0.001789	0.013469	0	0	0	0	0	0	1	-360	360;
	7171	10856	0.00163	0.011091	0	0	0	0	0	0	1	-360	360;
	7171	9287	0.00036	0.002651	0	0	0	0	0	0	1	-360	360;
	13129	19	0.00013	0.00132	0	0	0	0	0	0	1	-360	360;
	13129	427	0.000169	0.001659	0	0	0	0	0	0	1	-360	360;
	6501	7404	0.00136	0.012831	0	0	0	0	0	0	1	-360	360;
	6501	7404	0.00139	0.01256	0	0	0	0	0	0	1	-360	360;
	6501	7695	0.00126	0.009521	0	0	0	0	0	0	1	-360	360;
	6501	7695	0.00143	0.009331	0	0	0	0	0	0	1	-360	360;
	5377	6501	0.00164	0.012961	0	0	0	0	0	0	1	-360	360;
	5377	6501	0.001599	0.012	0	0	0	0	0	0	1	-360	360;
	10130	4257	0.001271	0.00524	0	0	0	0	0	0	1	-360	360;
	4869	13271	0.00226	0.02465	0	0	0	0	0	0	1	-360	360;
	3362	5285	0.00032	0.0015	0	0	0	0	0	0	1	-360	360;
	3362	8663	0.00032	0.00112	0	0	0	0	0	0	1	-360	360;
	4779	7223	0.00075	0.00817	0	0	0	0	0	0	1	-360	360;
	12008	2907	0.00037	0.0027	0	0	0	0	0	0	1	-360	360;
	2907	9908	3.9e-05	0.000219	0	0	0	0	0	0	1	-360	360;
	3076	875	0.012781	0.05762	0	0	0	0	0	0	1	-360	360;
	8565	1530	0.00036	0.001289	0	0	0	0	0	0	1	-360	360;
	8880	1682	0.006771	0.039349	0	0	0	0	0	0	1	-360	360;
	4493	11248	0.00292	0.02831	0	0	0	0	0	0	1	-360	360;
	4493	2965	0.00134	0.01229	0	0	0	0	0	0	1	-360	360;
	12008	8005	0.00011	0.00063	0	0	0	0	0	0	1	-360	360;
	10620	4466	0.0003	0.003331	0	0	0	0	0	0	1	-360	360;
	10620	5438	0.001289	0.01555	0	0	0	0	0	0	1	-360	360;
	9908	8626	0.00011	0.0017	0	0	0	0	0	0	1	-360	360;
	9908	7745	0.00043	0.00249	0	0	0	0	0	0	1	-360	360;
	1612	6835	0.002651	0.016771	0	0	0	0	0	0	1	-360	360;
	5103	1612	0.0082	0.025081	0	0	0	0	0	0	1	-360	360;
	5021	10031	0.00037	0.002781	0	0	0	0	0	0	1	-360	360;
	2029	7193	0.010719	0.054719	0	0	0	0	0	0	1	-360	360;
	4446	7193	0.006469	0.04074	0	0	0	0	0	0	1	-360	360;
	11083	13271	0.00101	0.01274	0	0	0	0	0	0	1	-360	360;
	4466	8681	9.1e-05	0.05076	0	0	0	0	0	0	1	-360	360;
	4466	5349	0.00588	0.03938	0	0	0	0	0	0	1	-360	360;
	4466	5438	0.003591	0.022099	0	0	0	0	0	0	1	-360	360;
	4466	11168	0.00555	0.032521	0	0	0	0	0	0	1	-360	360;
	10044	456	0.00116	0.01347	0	0	0	0	0	0	1	-360	360;
	10044	456	0.00113	0.01247	0	0	0	0	0	0	1	-360	360;
	7382	257	0.00107	0.00871	0	0	0	0	0	0	1	-360	360;
	7382	6519	0.00078	0.00609	0	0	0	0	0	0	1	-360	360;
	3687	9883	0.00075	0.00604	0	0	0	0	0	0	1	-360	360;
	7181	9822	0.00224	0.011091	0	0	0	0	0	0	1	-360	360;
	5863	10150	0.001961	0.013969	0	0	0	0	0	0	1	-360	360;
	1833	7541	0.001271	0.00761	0	0	0	0	0	0	1	-360	360;
	9773	9878	0.00374	0.03581	0	0	0	0	0	0	1	-360	360;
	9773	9878	0.00385	0.03284	0	0	0	0	0	0	1	-360	360;
	9773	9878	0.00392	0.03216	0	0	0	0	0	0	1	-360	360;
	9773	13271	0.00307	0.03131	0	0	0	0	0	0	1	-360	360;
	11466	40	0.027	0.131289	0	0	0	0	0	0	1	-360	360;
	9264	8019	0.005539	0.0195	0	0	0	0	0	0	1	-360	360;
	7404	340	0.003099	0.017849	0	0	0	0	0	0	1	-360	360;
	7404	7695	0.002831	0.014099	0	0	0	0	0	0	1	-360	360;
	5915	3230	0	0.01368	0	0	0	0	0	0	1	-360	360;
	6177	6835	0.01082	0.079719	0	0	0	0	0	0	1	-360	360;
	1060	3958	0.00102	0.00941	0	0	0	0	0	0	1	-360	360;
	1060	877	0.00183	0.01819	0	0	0	0	0	0	1	-360	360;
	1060	570	0.00028	0.00346	0	0	0	0	0	0	1	-360	360;
	5766	3684	0.0012	0.00982	0	0	0	0	0	0	1	-360	360;
	3958	3761	0.00027	0.0032	0	0	0	0	0	0	1	-360	360;
	3958	11270	0.00064	0.00679	0	0	0	0	0	0	1	-360	360;
	3958	12155	0.00047	0.00442	0	0	0	0	0	0	1	-360	360;
	608	12940	0.00493	0.017599	0	0	0	0	0	0	1	-360	360;
	608	6004	0.00207	0.0153	0	0	0	0	0	0	1	-360	360;
	10680	456	0.00094	0.0092	0	0	0	0	0	0	1	-360	360;
	10680	456	0.00095	0.00901	0	0	0	0	0	0	1	-360	360;
	1334	3983	0.00101	0.00719	0	0	0	0	0	0	1	-360	360;
	5489	9729	0.000419	0.00331	0	0	0	0	0	0	1	-360	360;
	8982	2713	0.002099	0.015599	0	0	0	0	0	0	1	-360	360;
	9878	12824	0.00163	0.01283	0	0	0	0	0	0	1	-360	360;
	9878	10884	0.00445	0.03788	0	0	0	0	0	0	1	-360	360;
	9878	6829	0.00092	0.00703	0	0	0	0	0	0	1	-360	360;
	375	40	9.1e-05	0.000961	0	0	0	0	0	0	1	-360	360;
	10886	375	0.00387	0.028599	0	0	0	0	0	0	1	-360	360;
	375	7714	0.00949	0.06625	0	0	0	0	0	0	1	-360	360;
	375	6555	0.00414	0.035331	0	0	0	0	0	0	1	-360	360;
	375	7714	0.009781	0.065581	0	0	0	0	0	0	1	-360	360;
	1973	40	0.02818	0.116099	0	0	0	0	0	0	1	-360	360;
	40	3035	0.00324	0.01912	0	0	0	0	0	0	1	-360	360;
	4332	9883	0.00135	0.01609	0	0	0	0	0	0	1	-360	360;
	9025	8256	0.00294	0.0182	0	0	0	0	0	0	1	-360	360;
	12985	8681	0.000711	0.00649	0	0	0	0	0	0	1	-360	360;
	650	12985	0.002651	0.01375	0	0	0	0	0	0	1	-360	360;
	1793	8681	0.00105	0.008469	0	0	0	0	0	0	1	-360	360;
	1227	6582	0.002229	0.016461	0	0	0	0	0	0	1	-360	360;
	1227	6582	0.002281	0.01613	0	0	0	0	0	0	1	-360	360;
	882	6582	0.003341	0.012651	0	0	0	0	0	0	1	-360	360;
	882	6582	0.003271	0.012901	0	0	0	0	0	0	1	-360	360;
	9491	12754	0.00101	0.01218	0	0	0	0	0	0	1	-360	360;
	9491	10894	0.0009	0.0095	0	0	0	0	0	0	1	-360	360;
	12754	6878	0.00098	0.01241	0	0	0	0	0	0	1	-360	360;
	8018	2770	0.001219	0.00951	0	0	0	0	0	0	1	-360	360;
	8018	6470	0.001789	0.01382	0	0	0	0	0	0	1	-360	360;
	8018	6470	0.00175	0.014081	0	0	0	0	0	0	1	-360	360;
	11140	4937	9.1e-05	0.00087	0	0	0	0	0	0	1	-360	360;
	11140	4937	9.1e-05	0.00086	0	0	0	0	0	0	1	-360	360;
	11140	12384	0.000229	0.00236	0	0	0	0	0	0	1	-360	360;
	999	7552	0.00125	0.00557	0	0	0	0	0	0	1	-360	360;
	3076	7582	0.00757	0.04013	0	0	0	0	0	0	1	-360	360;
	2418	257	0.00157	0.01604	0	0	0	0	0	0	1	-360	360;
	2418	257	0.00154	0.01634	0	0	0	0	0	0	1	-360	360;
	257	2965	0.00169	0.02383	0	0	0	0	0	0	1	-360	360;
	2465	257	0.00068	0.0089	0	0	0	0	0	0	1	-360	360;
	2633	6746	0.010289	0.036651	0	0	0	0	0	0	1	-360	360;
	2633	6746	0.00563	0.03395	0	0	0	0	0	0	1	-360	360;
	4497	8327	0.003591	0.02763	0	0	0	0	0	0	1	-360	360;
	4497	8327	0.00351	0.028159	0	0	0	0	0	0	1	-360	360;
	8327	6995	0.00143	0.00838	0	0	0	0	0	0	1	-360	360;
	11213	6294	0.01213	0.06156	0	0	0	0	0	0	1	-360	360;
	9025	7552	0.004789	0.022849	0	0	0	0	0	0	1	-360	360;
	777	4876	0.00019	0.00211	0	0	0	0	0	0	1	-360	360;
	777	12577	0.000331	0.002349	0	0	0	0	0	0	1	-360	360;
	777	7360	9.1e-05	0.000539	0	0	0	0	0	0	1	-360	360;
	9505	9663	0.00088	0.00776	0	0	0	0	0	0	1	-360	360;
	6829	6900	0.00039	0.00368	0	0	0	0	0	0	1	-360	360;
	9731	2959	0.00275	0.01268	0	0	0	0	0	0	1	-360	360;
	9731	13295	0.00205	0.00874	0	0	0	0	0	0	1	-360	360;
	9731	1756	0.00082	0.00519	0	0	0	0	0	0	1	-360	360;
	3807	1157	0.00519	0.02355	0	0	0	0	0	0	1	-360	360;
	4803	9858	0.016711	0.08874	0	0	0	0	0	0	1	-360	360;
	10856	6460	0.005099	0.022659	0	0	0	0	0	0	1	-360	360;
	10856	11738	0.001469	0.01213	0	0	0	0	0	0	1	-360	360;
	10856	11738	0.001659	0.01189	0	0	0	0	0	0	1	-360	360;
	1443	6555	0.000521	0.00374	0	0	0	0	0	0	1	-360	360;
	1443	7353	0.006461	0.048349	0	0	0	0	0	0	1	-360	360;
	12754	10711	0.00293	0.0257	0	0	0	0	0	0	1	-360	360;
	12754	5903	0.00178	0.0157	0	0	0	0	0	0	1	-360	360;
	8489	12754	0.00115	0.01332	0	0	0	0	0	0	1	-360	360;
	12754	3132	0.00078	0.00752	0	0	0	0	0	0	1	-360	360;
	116	4323	3.9e-05	0.00037	0	0	0	0	0	0	1	-360	360;
	11948	6470	0.0012	0.006729	0	0	0	0	0	0	1	-360	360;
	41	4865	0.000211	0.001659	0	0	0	0	0	0	1	-360	360;
	9497	12225	0.001659	0.011919	0	0	0	0	0	0	1	-360	360;
	9497	10032	0.002	0.0133	0	0	0	0	0	0	1	-360	360;
	6419	11469	0.00105	0.00837	0	0	0	0	0	0	1	-360	360;
	4149	7790	0.00138	0.014979	0	0	0	0	0	0	1	-360	360;
	8377	9986	0	0.0003	0	0	0	0	0	0	1	-360	360;
	4292	9986	0	0.0003	0	0	0	0	0	0	1	-360	360;
	4292	7854	0.00164	0.00782	0	0	0	0	0	0	1	-360	360;
	5349	344	0.007099	0.04044	0	0	0	0	0	0	1	-360	360;
	340	1468	0.00181	0.00914	0	0	0	0	0	0	1	-360	360;
	2105	10769	0.00027	0.00247	0	0	0	0	0	0	1	-360	360;
	10769	7258	0.00266	0.03206	0	0	0	0	0	0	1	-360	360;
	5778	1386	0.000591	0.005081	0	0	0	0	0	0	1	-360	360;
	5778	6867	0.002229	0.016099	0	0	0	0	0	0	1	-360	360;
	5915	7826	0.00205	0.02585	0	0	0	0	0	0	1	-360	360;
	11470	5854	0.00101	0.0091	0	0	0	0	0	0	1	-360	360;
	11470	5903	0.001	0.00871	0	0	0	0	0	0	1	-360	360;
	5457	11673	0.00112	0.00895	0	0	0	0	0	0	1	-360	360;
	5457	11673	0.001099	0.008289	0	0	0	0	0	0	1	-360	360;
	5457	4518	0.001289	0.0095	0	0	0	0	0	0	1	-360	360;
	5457	4518	0.00126	0.00968	0	0	0	0	0	0	1	-360	360;
	6322	7952	0.00065	0.00723	0	0	0	0	0	0	1	-360	360;
	2770	6026	0.001781	0.012669	0	0	0	0	0	0	1	-360	360;
	5833	1543	0.00086	0.00956	0	0	0	0	0	0	1	-360	360;
	5833	1156	0.00248	0.02429	0	0	0	0	0	0	1	-360	360;
	5833	6614	0.00151	0.01699	0	0	0	0	0	0	1	-360	360;
	8585	5435	0.003159	0.01537	0	0	0	0	0	0	1	-360	360;
	12079	4503	0.00075	0.00683	0	0	0	0	0	0	1	-360	360;
	12079	4089	0.00223	0.02524	0	0	0	0	0	0	1	-360	360;
	5854	5903	0.00191	0.0208	0	0	0	0	0	0	1	-360	360;
	2418	5854	0.00074	0.00784	0	0	0	0	0	0	1	-360	360;
	2418	5854	0.00073	0.00799	0	0	0	0	0	0	1	-360	360;
	2015	5336	0.000789	0.00482	0	0	0	0	0	0	1	-360	360;
	3007	2015	0.000789	0.004409	0	0	0	0	0	0	1	-360	360;
	4080	4803	0.005461	0.03161	0	0	0	0	0	0	1	-360	360;
	2340	7745	6e-05	0.00087	0	0	0	0	0	0	1	-360	360;
	4575	9844	0.00593	0.040961	0	0	0	0	0	0	1	-360	360;
	4575	7353	0.003771	0.030841	0	0	0	0	0	0	1	-360	360;
	9446	401	0.00038	0.001461	0	0	0	0	0	0	1	-360	360;
	401	6004	0.00149	0.00938	0	0	0	0	0	0	1	-360	360;
	401	6004	0.001521	0.00918	0	0	0	0	0	0	1	-360	360;
	2959	6898	0.00407	0.015469	0	0	0	0	0	0	1	-360	360;
	7253	2015	0.00082	0.006219	0	0	0	0	0	0	1	-360	360;
	3822	2808	0.00161	0.00888	0	0	0	0	0	0	1	-360	360;
	3822	11776	0.00118	0.006461	0	0	0	0	0	0	1	-360	360;
	10031	8972	0.00082	0.009091	0	0	0	0	0	0	1	-360	360;
	6867	6588	0.01893	0.0848	0	0	0	0	0	0	1	-360	360;
	6867	6588	0.01313	0.082409	0	0	0	0	0	0	1	-360	360;
	6867	1591	0.000281	0.00149	0	0	0	0	0	0	1	-360	360;
	7713	4503	0.00019	0.00199	0	0	0	0	0	0	1	-360	360;
	4503	7713	0.0002	0.00188	0	0	0	0	0	0	1	-360	360;
	245	2537	0.003159	0.025331	0	0	0	0	0	0	1	-360	360;
	11267	2038	0.00445	0.01886	0	0	0	0	0	0	1	-360	360;
	8469	1235	0.00461	0.02205	0	0	0	0	0	0	1	-360	360;
	5795	12082	0.00061	0.00494	0	0	0	0	0	0	1	-360	360;
	5795	1358	0.00082	0.00462	0	0	0	0	0	0	1	-360	360;
	3345	12820	0.000331	0.003271	0	0	0	0	0	0	1	-360	360;
	3345	11738	0.00056	0.004081	0	0	0	0	0	0	1	-360	360;
	9115	1370	0.00163	0.007521	0	0	0	0	0	0	1	-360	360;
	9115	13134	0.001659	0.00736	0	0	0	0	0	0	1	-360	360;
	2731	2140	0.00719	0.049211	0	0	0	0	0	0	1	-360	360;
	10929	2808	6e-05	0.00039	0	0	0	0	0	0	1	-360	360;
	1472	8779	0.00288	0.018	0	0	0	0	0	0	1	-360	360;
	1472	9612	0.00288	0.0182	0	0	0	0	0	0	1	-360	360;
	1472	7714	0.00111	0.008531	0	0	0	0	0	0	1	-360	360;
	1472	6271	0.00307	0.022289	0	0	0	0	0	0	1	-360	360;
	4420	6614	0.00091	0.00868	0	0	0	0	0	0	1	-360	360;
	5903	13478	0.00098	0.0118	0	0	0	0	0	0	1	-360	360;
	5903	10894	0.00053	0.00591	0	0	0	0	0	0	1	-360	360;
	5903	3132	0.00086	0.00991	0	0	0	0	0	0	1	-360	360;
	10230	4865	0.00545	0.034099	0	0	0	0	0	0	1	-360	360;
	2923	3589	0.00095	0.006091	0	0	0	0	0	0	1	-360	360;
	5198	1156	0.00305	0.0272	0	0	0	0	0	0	1	-360	360;
	1227	7198	0.000281	0.001841	0	0	0	0	0	0	1	-360	360;
	1227	9448	0.004531	0.02451	0	0	0	0	0	0	1	-360	360;
	6460	2140	0.00618	0.02961	0	0	0	0	0	0	1	-360	360;
	11673	5041	0.00064	0.004461	0	0	0	0	0	0	1	-360	360;
	1285	6206	0.00143	0.00513	0	0	0	0	0	0	1	-360	360;
	1285	1756	0.0052	0.021841	0	0	0	0	0	0	1	-360	360;
	882	7198	0.00138	0.008669	0	0	0	0	0	0	1	-360	360;
	9448	882	0.00314	0.013169	0	0	0	0	0	0	1	-360	360;
	6206	3480	0.002479	0.01331	0	0	0	0	0	0	1	-360	360;
	3275	3928	7e-05	0.00194	0	0	0	0	0	0	1	-360	360;
	13552	6470	0.001159	0.00931	0	0	0	0	0	0	1	-360	360;
	7854	9287	0.00062	0.003669	0	0	0	0	0	0	1	-360	360;
	3615	7542	0.015969	0.056711	0	0	0	0	0	0	1	-360	360;
	7542	5516	0.0085	0.03719	0	0	0	0	0	0	1	-360	360;
	3983	1370	0.001531	0.00732	0	0	0	0	0	0	1	-360	360;
	3983	1370	0.00205	0.01062	0	0	0	0	0	0	1	-360	360;
	7258	7952	0.00091	0.01056	0	0	0	0	0	0	1	-360	360;
	1156	12743	0.00167	0.02069	0	0	0	0	0	0	1	-360	360;
	9198	6294	0.002719	0.016219	0	0	0	0	0	0	1	-360	360;
	12143	6448	0.002	0.02376	0	0	0	0	0	0	1	-360	360;
	12143	6448	0.00226	0.02328	0	0	0	0	0	0	1	-360	360;
	11168	650	0.004151	0.02119	0	0	0	0	0	0	1	-360	360;
	13295	12350	0.001841	0.010909	0	0	0	0	0	0	1	-360	360;
	6294	6136	0.000669	0.00399	0	0	0	0	0	0	1	-360	360;
	957	11657	8e-05	0.00098	0	0	0	0	0	0	1	-360	360;
	4876	7360	0.0002	0.0025	0	0	0	0	0	0	1	-360	360;
	7996	7714	0.006219	0.031599	0	0	0	0	0	0	1	-360	360;
	7996	2034	0.00438	0.023169	0	0	0	0	0	0	1	-360	360;
	8986	6271	0.00581	0.03088	0	0	0	0	0	0	1	-360	360;
	8986	6271	0.0065	0.02637	0	0	0	0	0	0	1	-360	360;
	6829	877	0.00198	0.02208	0	0	0	0	0	0	1	-360	360;
	4497	1235	0.00093	0.006659	0	0	0	0	0	0	1	-360	360;
	4497	1235	0.000909	0.006789	0	0	0	0	0	0	1	-360	360;
	5915	1543	0.00244	0.01919	0	0	0	0	0	0	1	-360	360;
	3372	12668	0.00194	0.01549	0	0	0	0	0	0	1	-360	360;
	2381	5966	0.0002	0.001479	0	0	0	0	0	0	1	-360	360;
	6588	11673	0.001091	0.009461	0	0	0	0	0	0	1	-360	360;
	11673	9660	0.003081	0.00899	0	0	0	0	0	0	1	-360	360;
	9660	2062	0.00357	0.0138	0	0	0	0	0	0	1	-360	360;
	3589	11166	0.001229	0.00986	0	0	0	0	0	0	1	-360	360;
	2829	13462	0.001211	0.00913	0	0	0	0	0	0	1	-360	360;
	7258	11782	0.00116	0.01053	0	0	0	0	0	0	1	-360	360;
	6828	8268	0.0017	0.009781	0	0	0	0	0	0	1	-360	360;
	6828	8268	0.001521	0.011091	0	0	0	0	0	0	1	-360	360;
	286	656	0.0043	0.0233	0	0	0	0	0	0	1	-360	360;
	3862	6524	0.00019	0.0016	0	0	0	0	0	0	1	-360	360;
	6993	12225	0.00838	0.072479	0	0	0	0	0	0	1	-360	360;
	6993	1399	0.00787	0.049211	0	0	0	0	0	0	1	-360	360;
	650	344	0.00557	0.03376	0	0	0	0	0	0	1	-360	360;
	6224	10146	0.01213	0.06564	0	0	0	0	0	0	1	-360	360;
	10800	4518	0.003531	0.01086	0	0	0	0	0	0	1	-360	360;
	10800	10878	0.0158	0.057031	0	0	0	0	0	0	1	-360	360;
	1097	2965	0.00289	0.0244	0	0	0	0	0	0	1	-360	360;
	2279	6027	0.000521	0.00413	0	0	0	0	0	0	1	-360	360;
	3379	6027	0.000531	0.004039	0	0	0	0	0	0	1	-360	360;
	344	2570	0.00431	0.02226	0	0	0	0	0	0	1	-360	360;
	4359	6004	0.00163	0.00899	0	0	0	0	0	0	1	-360	360;
	6829	1810	0.00274	0.02428	0	0	0	0	0	0	1	-360	360;
	9226	8779	0.0093	0.056979	0	0	0	0	0	0	1	-360	360;
	9226	12923	0.011281	0.080039	0	0	0	0	0	0	1	-360	360;
	9226	8204	0.00449	0.02376	0	0	0	0	0	0	1	-360	360;
	3761	12155	0.00014	0.00148	0	0	0	0	0	0	1	-360	360;
	5536	12225	0.0055	0.033659	0	0	0	0	0	0	1	-360	360;
	6555	7954	0.00249	0.02082	0	0	0	0	0	0	1	-360	360;
	2465	2965	0.00164	0.02252	0	0	0	0	0	0	1	-360	360;
	8419	3039	0.000961	0.0047	0	0	0	0	0	0	1	-360	360;
	8419	3039	0.000979	0.004599	0	0	0	0	0	0	1	-360	360;
	11270	570	0.00057	0.00515	0	0	0	0	0	0	1	-360	360;
	12225	10032	0.00143	0.011651	0	0	0	0	0	0	1	-360	360;
	2134	6004	0.00126	0.009211	0	0	0	0	0	0	1	-360	360;
	8779	1012	9.1e-05	0.000211	0	0	0	0	0	0	1	-360	360;
	9612	1012	0.000211	0.000711	0	0	0	0	0	0	1	-360	360;
	9612	8204	0.01095	0.07062	0	0	0	0	0	0	1	-360	360;
	3035	8204	0.00125	0.007281	0	0	0	0	0	0	1	-360	360;
	3035	8204	0.001	0.0045	0	0	0	0	0	0	1	-360	360;
	7954	7353	0.00282	0.026419	0	0	0	0	0	0	1	-360	360;
	13224	9686	0.003479	0.01456	0	0	0	0	0	0	1	-360	360;
	7714	6271	0.005521	0.03418	0	0	0	0	0	0	1	-360	360;
	7714	6271	0.00563	0.033461	0	0	0	0	0	0	1	-360	360;
	7353	6962	0.00095	0.0072	0	0	0	0	0	0	1	-360	360;
	7353	10146	0.00093	0.007159	0	0	0	0	0	0	1	-360	360;
	2034	6271	0.00988	0.058651	0	0	0	0	0	0	1	-360	360;
	9729	12384	3.9e-05	0.00044	0	0	0	0	0	0	1	-360	360;
	9729	12384	3.9e-05	0.00043	0	0	0	0	0	0	1	-360	360;
	8226	7917	0.00018	0.000781	0	0	0	0	0	0	1	-360	360;
	6321	11066	0.002961	0.01669	0	0	0	0	0	0	1	-360	360;
	8434	5760	0.00056	0.00432	0	0	0	0	0	0	1	-360	360;
	9867	12476	0.0002	0.00246	0	0	0	0	0	0	1	-360	360;
	9867	6599	0.0003	0.00377	0	0	0	0	0	0	1	-360	360;
	5046	13471	0.00014	0.00057	0	0	0	0	0	0	1	-360	360;
	4215	13471	0.00013	0.000729	0	0	0	0	0	0	1	-360	360;
	2033	13471	0.00012	0.000581	0	0	0	0	0	0	1	-360	360;
	2634	13471	0.00011	0.00064	0	0	0	0	0	0	1	-360	360;
	5721	13471	0.00019	0.00056	0	0	0	0	0	0	1	-360	360;
	9083	13001	0.00062	0.004909	0	0	0	0	0	0	1	-360	360;
	13300	11517	7e-05	0.00076	0	0	0	0	0	0	1	-360	360;
	13300	13605	0.00066	0.00622	0	0	0	0	0	0	1	-360	360;
	13300	3010	0.00054	0.00752	0	0	0	0	0	0	1	-360	360;
	13300	36	0.00105	0.00877	0	0	0	0	0	0	1	-360	360;
	924	13177	8e-05	0.00091	0	0	0	0	0	0	1	-360	360;
	924	6741	0.00056	0.00719	0	0	0	0	0	0	1	-360	360;
	3475	8047	0.000531	0.002039	0	0	0	0	0	0	1	-360	360;
	3475	1798	0.008659	0.03769	0	0	0	0	0	0	1	-360	360;
	3475	13250	0.003849	0.021099	0	0	0	0	0	0	1	-360	360;
	3475	12568	0.00688	0.035969	0	0	0	0	0	0	1	-360	360;
	3475	10058	0.006729	0.03713	0	0	0	0	0	0	1	-360	360;
	3475	4969	0.00111	0.006651	0	0	0	0	0	0	1	-360	360;
	3475	936	0.0023	0.008909	0	0	0	0	0	0	1	-360	360;
	3475	3737	0.000841	0.00313	0	0	0	0	0	0	1	-360	360;
	3475	7793	0.003521	0.017479	0	0	0	0	0	0	1	-360	360;
	4397	1039	3.9e-05	0.00025	0	0	0	0	0	0	1	-360	360;
	4369	3319	0.000159	0.000729	0	0	0	0	0	0	1	-360	360;
	12797	10943	0.005281	0.032099	0	0	0	0	0	0	1	-360	360;
	12797	10943	0.005651	0.029419	0	0	0	0	0	0	1	-360	360;
	1266	5393	0.000271	0.00036	0	0	0	0	0	0	1	-360	360;
	12926	10303	0.00044	0.00143	0	0	0	0	0	0	1	-360	360;
	12926	5753	0.000159	0.00074	0	0	0	0	0	0	1	-360	360;
	3246	2813	0.00019	0.000591	0	0	0	0	0	0	1	-360	360;
	7493	4212	0.00019	0.00087	0	0	0	0	0	0	1	-360	360;
	9692	7207	0.002591	0.013659	0	0	0	0	0	0	1	-360	360;
	9692	10710	0.00305	0.01168	0	0	0	0	0	0	1	-360	360;
	606	4776	0.000159	0.000599	0	0	0	0	0	0	1	-360	360;
	606	4776	0.000159	0.00062	0	0	0	0	0	0	1	-360	360;
	4732	7207	0.00101	0.004229	0	0	0	0	0	0	1	-360	360;
	4732	6274	0.00319	0.018651	0	0	0	0	0	0	1	-360	360;
	4732	10710	0.000729	0.00337	0	0	0	0	0	0	1	-360	360;
	1647	1694	0.002961	0.012831	0	0	0	0	0	0	1	-360	360;
	1647	11678	0.000591	0.00339	0	0	0	0	0	0	1	-360	360;
	13172	3466	0.00024	0.001219	0	0	0	0	0	0	1	-360	360;
	12795	4279	0.006219	0.036341	0	0	0	0	0	0	1	-360	360;
	12795	8788	0.001841	0.01193	0	0	0	0	0	0	1	-360	360;
	4805	10642	0.001409	0.01113	0	0	0	0	0	0	1	-360	360;
	4805	426	0.00151	0.00768	0	0	0	0	0	0	1	-360	360;
	4805	4435	0.00599	0.03313	0	0	0	0	0	0	1	-360	360;
	4805	4953	0.001771	0.010151	0	0	0	0	0	0	1	-360	360;
	543	1154	0.00074	0.00311	0	0	0	0	0	0	1	-360	360;
	543	1154	0.000729	0.003159	0	0	0	0	0	0	1	-360	360;
	4667	11443	0.005271	0.01906	0	0	0	0	0	0	1	-360	360;
	4667	7938	0.00906	0.053151	0	0	0	0	0	0	1	-360	360;
	4667	2641	0.00405	0.0153	0	0	0	0	0	0	1	-360	360;
	9942	7878	0.00055	0.003961	0	0	0	0	0	0	1	-360	360;
	9942	7878	0.00055	0.003729	0	0	0	0	0	0	1	-360	360;
	9942	2296	0.00107	0.007789	0	0	0	0	0	0	1	-360	360;
	9942	2296	0.00112	0.00775	0	0	0	0	0	0	1	-360	360;
	9942	2296	0.001031	0.00775	0	0	0	0	0	0	1	-360	360;
	9942	3493	0.00062	0.003219	0	0	0	0	0	0	1	-360	360;
	9942	1346	0.00068	0.002659	0	0	0	0	0	0	1	-360	360;
	13517	738	0.00193	0.012479	0	0	0	0	0	0	1	-360	360;
	13517	4340	0.001159	0.007229	0	0	0	0	0	0	1	-360	360;
	13517	4246	0.00137	0.008729	0	0	0	0	0	0	1	-360	360;
	13517	4246	0.001401	0.00855	0	0	0	0	0	0	1	-360	360;
	13517	4246	0.001331	0.008521	0	0	0	0	0	0	1	-360	360;
	10331	3147	0.00031	0.00081	0	0	0	0	0	0	1	-360	360;
	10331	11861	0.000341	0.00119	0	0	0	0	0	0	1	-360	360;
	10331	6577	0.00031	0.001281	0	0	0	0	0	0	1	-360	360;
	10331	6228	0.00094	0.0035	0	0	0	0	0	0	1	-360	360;
	12595	7709	7e-05	0.000651	0	0	0	0	0	0	1	-360	360;
	7969	5396	0.00019	0.00124	0	0	0	0	0	0	1	-360	360;
	7969	4212	0.000331	0.002539	0	0	0	0	0	0	1	-360	360;
	4543	5236	0.0016	0.02113	0	0	0	0	0	0	1	-360	360;
	11390	12662	0.000909	0.004711	0	0	0	0	0	0	1	-360	360;
	11390	12662	0.00088	0.004969	0	0	0	0	0	0	1	-360	360;
	5	12839	0.001841	0.0088	0	0	0	0	0	0	1	-360	360;
	5	11276	0.01013	0.04238	0	0	0	0	0	0	1	-360	360;
	9749	4225	0.00059	0.00802	0	0	0	0	0	0	1	-360	360;
	9749	4225	0.00054	0.0079	0	0	0	0	0	0	1	-360	360;
	9749	5763	0.00171	0.02128	0	0	0	0	0	0	1	-360	360;
	9749	5763	0.00174	0.02084	0	0	0	0	0	0	1	-360	360;
	9749	9633	0.00086	0.01433	0	0	0	0	0	0	1	-360	360;
	9749	9633	0.00087	0.01404	0	0	0	0	0	0	1	-360	360;
	9749	5623	0.00126	0.01447	0	0	0	0	0	0	1	-360	360;
	9749	8232	0.00122	0.01509	0	0	0	0	0	0	1	-360	360;
	9749	8232	0.00154	0.019	0	0	0	0	0	0	1	-360	360;
	7207	7194	0.00474	0.03564	0	0	0	0	0	0	1	-360	360;
	7207	8676	0.006	0.047849	0	0	0	0	0	0	1	-360	360;
	7207	7631	0.008039	0.03375	0	0	0	0	0	0	1	-360	360;
	10011	10710	2.1e-05	0.000211	0	0	0	0	0	0	1	-360	360;
	12238	5367	0.00045	0.002341	0	0	0	0	0	0	1	-360	360;
	3966	13022	0.000669	0.004281	0	0	0	0	0	0	1	-360	360;
	13659	13022	0.00068	0.00419	0	0	0	0	0	0	1	-360	360;
	4462	5428	0.000729	0.00431	0	0	0	0	0	0	1	-360	360;
	4462	5428	0.000591	0.004651	0	0	0	0	0	0	1	-360	360;
	4462	13138	0.00099	0.005081	0	0	0	0	0	0	1	-360	360;
	2234	13237	0.00212	0.013031	0	0	0	0	0	0	1	-360	360;
	2234	5194	0.002159	0.01275	0	0	0	0	0	0	1	-360	360;
	12849	5446	0.00025	0.00095	0	0	0	0	0	0	1	-360	360;
	12849	5446	0.000271	0.00089	0	0	0	0	0	0	1	-360	360;
	12849	4005	0.001961	0.012409	0	0	0	0	0	0	1	-360	360;
	12849	12264	0.00044	0.00207	0	0	0	0	0	0	1	-360	360;
	12849	12552	0.000469	0.00143	0	0	0	0	0	0	1	-360	360;
	12849	9611	0.001271	0.01406	0	0	0	0	0	0	1	-360	360;
	12849	9611	0.00124	0.013021	0	0	0	0	0	0	1	-360	360;
	12849	7971	0.000409	0.00239	0	0	0	0	0	0	1	-360	360;
	12849	319	0.00094	0.005	0	0	0	0	0	0	1	-360	360;
	12849	4469	0.000219	0.0007	0	0	0	0	0	0	1	-360	360;
	12849	9893	6e-05	0.00025	0	0	0	0	0	0	1	-360	360;
	12849	7698	5e-05	0.000229	0	0	0	0	0	0	1	-360	360;
	5537	3942	0.00019	0.000909	0	0	0	0	0	0	1	-360	360;
	3251	8065	0.001711	0.00782	0	0	0	0	0	0	1	-360	360;
	3251	9955	0.002651	0.01176	0	0	0	0	0	0	1	-360	360;
	11386	655	0.001409	0.010599	0	0	0	0	0	0	1	-360	360;
	11386	7779	0.001419	0.010219	0	0	0	0	0	0	1	-360	360;
	1190	5763	0.00013	0.00212	0	0	0	0	0	0	1	-360	360;
	1190	1158	0.00026	0.00313	0	0	0	0	0	0	1	-360	360;
	6613	3262	0.0047	0.0318	0	0	0	0	0	0	1	-360	360;
	6613	9938	0.00136	0.00768	0	0	0	0	0	0	1	-360	360;
	6613	8426	0.005669	0.027789	0	0	0	0	0	0	1	-360	360;
	9794	9150	0.004409	0.033091	0	0	0	0	0	0	1	-360	360;
	9794	2032	0.01056	0.044469	0	0	0	0	0	0	1	-360	360;
	9794	10468	0.01963	0.07736	0	0	0	0	0	0	1	-360	360;
	9794	10895	0.01157	0.04694	0	0	0	0	0	0	1	-360	360;
	8915	1228	0.000479	0.00255	0	0	0	0	0	0	1	-360	360;
	11078	1228	0.000469	0.00236	0	0	0	0	0	0	1	-360	360;
	1695	2203	0.01237	0.067039	0	0	0	0	0	0	1	-360	360;
	1695	12559	0.0132	0.056229	0	0	0	0	0	0	1	-360	360;
	5839	6011	0.00026	0.001711	0	0	0	0	0	0	1	-360	360;
	5839	6011	0.000271	0.001669	0	0	0	0	0	0	1	-360	360;
	9998	6400	0.000211	0.000401	0	0	0	0	0	0	1	-360	360;
	3099	8460	0.00105	0.005159	0	0	0	0	0	0	1	-360	360;
	8287	4479	0.00212	0.00931	0	0	0	0	0	0	1	-360	360;
	902	9920	0.000151	0.00068	0	0	0	0	0	0	1	-360	360;
	7886	7195	0.00019	0.00082	0	0	0	0	0	0	1	-360	360;
	5786	9955	0.003219	0.01405	0	0	0	0	0	0	1	-360	360;
	9788	13544	0.007271	0.033159	0	0	0	0	0	0	1	-360	360;
	9788	11829	0.006659	0.03132	0	0	0	0	0	0	1	-360	360;
	5026	4088	0.00057	0.00668	0	0	0	0	0	0	1	-360	360;
	5026	10309	0.00057	0.00758	0	0	0	0	0	0	1	-360	360;
	5026	10309	0.00039	0.00607	0	0	0	0	0	0	1	-360	360;
	5026	2348	0.00155	0.0179	0	0	0	0	0	0	1	-360	360;
	5026	6736	0.00036	0.00485	0	0	0	0	0	0	1	-360	360;
	8307	2413	0.000581	0.004229	0	0	0	0	0	0	1	-360	360;
	8307	12054	0.00056	0.00431	0	0	0	0	0	0	1	-360	360;
	8307	8496	0.003099	0.020531	0	0	0	0	0	0	1	-360	360;
	8307	4621	0.00118	0.016919	0	0	0	0	0	0	1	-360	360;
	8307	7730	0.00188	0.01407	0	0	0	0	0	0	1	-360	360;
	8307	2392	0.00545	0.04082	0	0	0	0	0	0	1	-360	360;
	10642	655	0.00214	0.01639	0	0	0	0	0	0	1	-360	360;
	10642	6955	0.002479	0.01549	0	0	0	0	0	0	1	-360	360;
	2621	11642	0.000159	0.001229	0	0	0	0	0	0	1	-360	360;
	2621	3790	0.00026	0.00162	0	0	0	0	0	0	1	-360	360;
	2621	10090	0.000159	0.00113	0	0	0	0	0	0	1	-360	360;
	2621	13078	0.000461	0.003219	0	0	0	0	0	0	1	-360	360;
	2621	6144	0.000159	0.001229	0	0	0	0	0	0	1	-360	360;
	2291	738	0.000229	0.00139	0	0	0	0	0	0	1	-360	360;
	4078	738	0.0002	0.00156	0	0	0	0	0	0	1	-360	360;
	1919	6911	0.00145	0.00824	0	0	0	0	0	0	1	-360	360;
	1919	738	0.00069	0.005	0	0	0	0	0	0	1	-360	360;
	1251	5445	0.00036	0.00455	0	0	0	0	0	0	1	-360	360;
	1251	5445	0.00036	0.00453	0	0	0	0	0	0	1	-360	360;
	12241	11033	0.00066	0.00997	0	0	0	0	0	0	1	-360	360;
	12241	11033	0.00064	0.01016	0	0	0	0	0	0	1	-360	360;
	9209	1753	9.9e-05	0.00044	0	0	0	0	0	0	1	-360	360;
	1753	6486	6e-05	0.000469	0	0	0	0	0	0	1	-360	360;
	1753	835	3.9e-05	0.000341	0	0	0	0	0	0	1	-360	360;
	1753	1829	0.002531	0.01943	0	0	0	0	0	0	1	-360	360;
	1753	5428	0.00945	0.06418	0	0	0	0	0	0	1	-360	360;
	13498	320	0.000591	0.002831	0	0	0	0	0	0	1	-360	360;
	10251	320	0.00087	0.00387	0	0	0	0	0	0	1	-360	360;
	1041	6488	0.0015	0.01785	0	0	0	0	0	0	1	-360	360;
	1041	6488	0.00147	0.0182	0	0	0	0	0	0	1	-360	360;
	1777	934	0.00151	0.0113	0	0	0	0	0	0	1	-360	360;
	1777	934	0.001479	0.011521	0	0	0	0	0	0	1	-360	360;
	1777	12691	0.00174	0.010341	0	0	0	0	0	0	1	-360	360;
	1777	12691	0.001659	0.01206	0	0	0	0	0	0	1	-360	360;
	12078	2847	0.00069	0.000479	0	0	0	0	0	0	1	-360	360;
	12078	11726	0.00044	0.00139	0	0	0	0	0	0	1	-360	360;
	1489	13380	0.002039	0.01601	0	0	0	0	0	0	1	-360	360;
	1489	10933	0.01349	0.05662	0	0	0	0	0	0	1	-360	360;
	1489	457	0.00082	0.00438	0	0	0	0	0	0	1	-360	360;
	1489	457	0.0008	0.004461	0	0	0	0	0	0	1	-360	360;
	1489	7458	0.001091	0.00888	0	0	0	0	0	0	1	-360	360;
	1489	7458	0.00111	0.007039	0	0	0	0	0	0	1	-360	360;
	3628	12484	0.000849	0.00537	0	0	0	0	0	0	1	-360	360;
	9062	13162	0.00061	0.00275	0	0	0	0	0	0	1	-360	360;
	9068	11274	0.00141	0.0188	0	0	0	0	0	0	1	-360	360;
	9068	9685	0.00121	0.01427	0	0	0	0	0	0	1	-360	360;
	3391	12003	0.001711	0.007159	0	0	0	0	0	0	1	-360	360;
	3391	13011	0.00545	0.02381	0	0	0	0	0	0	1	-360	360;
	3391	6083	0.000771	0.004331	0	0	0	0	0	0	1	-360	360;
	3391	3998	0.00414	0.03282	0	0	0	0	0	0	1	-360	360;
	3391	9262	0.00264	0.011521	0	0	0	0	0	0	1	-360	360;
	3391	9334	0.00261	0.011	0	0	0	0	0	0	1	-360	360;
	3391	12412	0.00239	0.00919	0	0	0	0	0	0	1	-360	360;
	3391	4134	0.000539	0.00226	0	0	0	0	0	0	1	-360	360;
	3391	10056	0.00325	0.017979	0	0	0	0	0	0	1	-360	360;
	1766	6661	0.00076	0.0071	0	0	0	0	0	0	1	-360	360;
	1168	12412	0.000651	0.00343	0	0	0	0	0	0	1	-360	360;
	3286	6389	0.001219	0.00812	0	0	0	0	0	0	1	-360	360;
	3286	1162	0.00243	0.01439	0	0	0	0	0	0	1	-360	360;
	13566	9117	0.0006	0.00839	0	0	0	0	0	0	1	-360	360;
	13566	9117	0.00059	0.00856	0	0	0	0	0	0	1	-360	360;
	13566	8243	0.00087	0.01328	0	0	0	0	0	0	1	-360	360;
	13566	8243	0.00073	0.01595	0	0	0	0	0	0	1	-360	360;
	4205	5378	0.00163	0.002039	0	0	0	0	0	0	1	-360	360;
	4205	5378	0.001659	0.002	0	0	0	0	0	0	1	-360	360;
	83	8374	0.00339	0.019669	0	0	0	0	0	0	1	-360	360;
	83	3377	0.00239	0.017281	0	0	0	0	0	0	1	-360	360;
	9532	11801	0.00032	0.00112	0	0	0	0	0	0	1	-360	360;
	9532	8620	0.0018	0.01401	0	0	0	0	0	0	1	-360	360;
	9532	8620	0.00212	0.012469	0	0	0	0	0	0	1	-360	360;
	11601	4090	0.00063	0.00411	0	0	0	0	0	0	1	-360	360;
	9407	6043	0.009729	0.03505	0	0	0	0	0	0	1	-360	360;
	9407	5222	0.01007	0.03056	0	0	0	0	0	0	1	-360	360;
	7189	4031	0.00042	0.00623	0	0	0	0	0	0	1	-360	360;
	7189	523	0.00036	0.00722	0	0	0	0	0	0	1	-360	360;
	7189	9633	0.00053	0.0078	0	0	0	0	0	0	1	-360	360;
	7189	2746	0.00109	0.0173	0	0	0	0	0	0	1	-360	360;
	7189	8497	0.00071	0.01144	0	0	0	0	0	0	1	-360	360;
	7189	5623	0.00157	0.01884	0	0	0	0	0	0	1	-360	360;
	7189	5249	0.00032	0.00384	0	0	0	0	0	0	1	-360	360;
	11334	2558	0.00041	0.00649	0	0	0	0	0	0	1	-360	360;
	11334	11442	0.0004	0.00658	0	0	0	0	0	0	1	-360	360;
	11334	9633	0.00054	0.00763	0	0	0	0	0	0	1	-360	360;
	11334	2746	0.00107	0.01602	0	0	0	0	0	0	1	-360	360;
	11334	8497	0.0007	0.01165	0	0	0	0	0	0	1	-360	360;
	11334	8232	0.0016	0.01844	0	0	0	0	0	0	1	-360	360;
	11334	5249	0.00032	0.00376	0	0	0	0	0	0	1	-360	360;
	6766	244	0.00212	0.01207	0	0	0	0	0	0	1	-360	360;
	4533	3350	5e-05	0.00037	0	0	0	0	0	0	1	-360	360;
	9150	12997	0.00355	0.024651	0	0	0	0	0	0	1	-360	360;
	9124	12744	0.00675	0.04145	0	0	0	0	0	0	1	-360	360;
	8047	10444	0.00195	0.014341	0	0	0	0	0	0	1	-360	360;
	12320	6392	0.00018	0.00043	0	0	0	0	0	0	1	-360	360;
	12320	4212	0.00037	0.000969	0	0	0	0	0	0	1	-360	360;
	890	1902	0.00257	0.017169	0	0	0	0	0	0	1	-360	360;
	890	7887	0.00076	0.00399	0	0	0	0	0	0	1	-360	360;
	890	7887	0.00075	0.00406	0	0	0	0	0	0	1	-360	360;
	890	4974	0.003901	0.023961	0	0	0	0	0	0	1	-360	360;
	890	1116	0.00249	0.019211	0	0	0	0	0	0	1	-360	360;
	5393	4212	0.00025	0.001159	0	0	0	0	0	0	1	-360	360;
	1920	13477	0.00195	0.02924	0	0	0	0	0	0	1	-360	360;
	1920	5981	0.00152	0.01527	0	0	0	0	0	0	1	-360	360;
	1920	11182	0.00149	0.01558	0	0	0	0	0	0	1	-360	360;
	1920	1554	0.00253	0.0276	0	0	0	0	0	0	1	-360	360;
	1920	1554	0.00263	0.02648	0	0	0	0	0	0	1	-360	360;
	1920	3376	0.00077	0.0092	0	0	0	0	0	0	1	-360	360;
	1920	6157	0.00112	0.01705	0	0	0	0	0	0	1	-360	360;
	8521	12611	0.00226	0.01101	0	0	0	0	0	0	1	-360	360;
	8521	8718	0.004531	0.029039	0	0	0	0	0	0	1	-360	360;
	8521	11774	0.002081	0.030409	0	0	0	0	0	0	1	-360	360;
	8521	5779	0.0053	0.03257	0	0	0	0	0	0	1	-360	360;
	8521	4848	0.00513	0.041901	0	0	0	0	0	0	1	-360	360;
	8521	6956	0.001461	0.01056	0	0	0	0	0	0	1	-360	360;
	5497	1162	0.0005	0.002591	0	0	0	0	0	0	1	-360	360;
	8992	1294	0.00055	0.0033	0	0	0	0	0	0	1	-360	360;
	5888	9561	0.002289	0.01763	0	0	0	0	0	0	1	-360	360;
	5888	3083	0.0075	0.043531	0	0	0	0	0	0	1	-360	360;
	8322	4974	0.00138	0.019961	0	0	0	0	0	0	1	-360	360;
	8322	1139	0.00126	0.018729	0	0	0	0	0	0	1	-360	360;
	3319	13581	0.00038	0.001331	0	0	0	0	0	0	1	-360	360;
	3319	4966	0.00019	0.001461	0	0	0	0	0	0	1	-360	360;
	300	6955	0.002151	0.013401	0	0	0	0	0	0	1	-360	360;
	300	11424	0.000789	0.005669	0	0	0	0	0	0	1	-360	360;
	4638	9840	3e-05	0.00042	0	0	0	0	0	0	1	-360	360;
	4638	970	0.00079	0.00899	0	0	0	0	0	0	1	-360	360;
	4638	970	0.0008	0.00879	0	0	0	0	0	0	1	-360	360;
	4638	6512	0.00334	0.03393	0	0	0	0	0	0	1	-360	360;
	4638	2350	0.00063	0.00789	0	0	0	0	0	0	1	-360	360;
	4638	3448	0.0006	0.00842	0	0	0	0	0	0	1	-360	360;
	12483	11746	0.00011	0.00125	0	0	0	0	0	0	1	-360	360;
	12483	9538	0.00012	0.001091	0	0	0	0	0	0	1	-360	360;
	12483	11283	9.9e-05	0.00125	0	0	0	0	0	0	1	-360	360;
	12483	609	0.00294	0.02443	0	0	0	0	0	0	1	-360	360;
	12483	4086	0.00144	0.011151	0	0	0	0	0	0	1	-360	360;
	12483	3502	0.001831	0.010849	0	0	0	0	0	0	1	-360	360;
	12483	3502	0.001599	0.01055	0	0	0	0	0	0	1	-360	360;
	12483	4315	0.00144	0.019591	0	0	0	0	0	0	1	-360	360;
	8282	1420	0.0133	0.09551	0	0	0	0	0	0	1	-360	360;
	8282	6497	0.009591	0.073961	0	0	0	0	0	0	1	-360	360;
	6488	3891	0.0007	0.00671	0	0	0	0	0	0	1	-360	360;
	6488	3891	0.00072	0.00658	0	0	0	0	0	0	1	-360	360;
	3378	1154	0.001901	0.00964	0	0	0	0	0	0	1	-360	360;
	3378	5010	0.001711	0.01189	0	0	0	0	0	0	1	-360	360;
	4637	6956	0.00025	0.00187	0	0	0	0	0	0	1	-360	360;
	6158	8034	0.00043	0.002211	0	0	0	0	0	0	1	-360	360;
	8292	12299	0.000401	0.00243	0	0	0	0	0	0	1	-360	360;
	11443	7462	0.00336	0.017	0	0	0	0	0	0	1	-360	360;
	10579	1162	0.005521	0.031031	0	0	0	0	0	0	1	-360	360;
	6466	3056	9.9e-05	0.00026	0	0	0	0	0	0	1	-360	360;
	9009	3917	0.00045	0.00718	0	0	0	0	0	0	1	-360	360;
	9009	6675	0.00044	0.00665	0	0	0	0	0	0	1	-360	360;
	9009	11015	0.00103	0.01516	0	0	0	0	0	0	1	-360	360;
	9009	10996	0.00101	0.01546	0	0	0	0	0	0	1	-360	360;
	7339	32	0.00045	0.00608	0	0	0	0	0	0	1	-360	360;
	7339	13621	0.0004	0.00501	0	0	0	0	0	0	1	-360	360;
	8926	13308	0.001729	0.01337	0	0	0	0	0	0	1	-360	360;
	8926	9096	0.00264	0.01489	0	0	0	0	0	0	1	-360	360;
	8926	1242	0.002341	0.018961	0	0	0	0	0	0	1	-360	360;
	8926	8338	0.001841	0.01363	0	0	0	0	0	0	1	-360	360;
	2058	11287	0.001219	0.00949	0	0	0	0	0	0	1	-360	360;
	2058	13308	0.002271	0.01151	0	0	0	0	0	0	1	-360	360;
	2058	8721	0.010781	0.05775	0	0	0	0	0	0	1	-360	360;
	2058	10761	0.00251	0.013919	0	0	0	0	0	0	1	-360	360;
	2058	203	0.00101	0.007331	0	0	0	0	0	0	1	-360	360;
	3572	6043	0.00061	0.003909	0	0	0	0	0	0	1	-360	360;
	5857	9324	0.004289	0.019331	0	0	0	0	0	0	1	-360	360;
	5857	6043	0.0008	0.004401	0	0	0	0	0	0	1	-360	360;
	5857	7077	0.008099	0.03519	0	0	0	0	0	0	1	-360	360;
	5857	11544	0.00063	0.004591	0	0	0	0	0	0	1	-360	360;
	4012	11544	0.000651	0.00456	0	0	0	0	0	0	1	-360	360;
	3572	8614	0.00114	0.00581	0	0	0	0	0	0	1	-360	360;
	3572	3182	0.001211	0.00538	0	0	0	0	0	0	1	-360	360;
	961	6290	0.00418	0.030159	0	0	0	0	0	0	1	-360	360;
	961	12535	0.00324	0.02413	0	0	0	0	0	0	1	-360	360;
	961	12535	0.003169	0.024591	0	0	0	0	0	0	1	-360	360;
	6530	13308	0.000219	0.00088	0	0	0	0	0	0	1	-360	360;
	6530	13308	0.000289	0.00095	0	0	0	0	0	0	1	-360	360;
	1294	55	0.0035	0.022211	0	0	0	0	0	0	1	-360	360;
	1294	13237	0.00313	0.02256	0	0	0	0	0	0	1	-360	360;
	8660	13031	0.000831	0.00162	0	0	0	0	0	0	1	-360	360;
	4605	12559	0.002099	0.01068	0	0	0	0	0	0	1	-360	360;
	1426	12559	0.00224	0.00976	0	0	0	0	0	0	1	-360	360;
	12997	2050	0.00169	0.01231	0	0	0	0	0	0	1	-360	360;
	9707	13372	0.00011	0.00076	0	0	0	0	0	0	1	-360	360;
	13465	13372	0.00011	0.00068	0	0	0	0	0	0	1	-360	360;
	10170	9324	0.005479	0.021909	0	0	0	0	0	0	1	-360	360;
	10170	13045	0.003331	0.017039	0	0	0	0	0	0	1	-360	360;
	7974	6389	0.000479	0.001271	0	0	0	0	0	0	1	-360	360;
	10941	4266	0.00109	0.01629	0	0	0	0	0	0	1	-360	360;
	10941	4439	4e-05	0.00063	0	0	0	0	0	0	1	-360	360;
	806	9181	0.0053	0.024531	0	0	0	0	0	0	1	-360	360;
	806	3268	0.004659	0.021919	0	0	0	0	0	0	1	-360	360;
	806	5395	0.001659	0.01043	0	0	0	0	0	0	1	-360	360;
	7319	6084	0.006961	0.053539	0	0	0	0	0	0	1	-360	360;
	2696	7917	0.00036	0.00288	0	0	0	0	0	0	1	-360	360;
	2696	7917	0.00038	0.003	0	0	0	0	0	0	1	-360	360;
	7987	7274	0.00048	0.00643	0	0	0	0	0	0	1	-360	360;
	7987	7274	0.00051	0.0066	0	0	0	0	0	0	1	-360	360;
	7987	7274	0.00046	0.00729	0	0	0	0	0	0	1	-360	360;
	7987	7948	0.0006	0.00868	0	0	0	0	0	0	1	-360	360;
	7987	11402	0.00068	0.01076	0	0	0	0	0	0	1	-360	360;
	7987	10071	0.00066	0.01096	0	0	0	0	0	0	1	-360	360;
	4445	6692	0.00101	0.005099	0	0	0	0	0	0	1	-360	360;
	823	12495	0.003901	0.02437	0	0	0	0	0	0	1	-360	360;
	823	3211	0.00276	0.016461	0	0	0	0	0	0	1	-360	360;
	823	11608	0.00588	0.047081	0	0	0	0	0	0	1	-360	360;
	6350	2641	3.9e-05	0.000219	0	0	0	0	0	0	1	-360	360;
	4119	12862	0.00131	0.00538	0	0	0	0	0	0	1	-360	360;
	4094	1554	0.00679	0.04679	0	0	0	0	0	0	1	-360	360;
	4094	13408	0.00138	0.01466	0	0	0	0	0	0	1	-360	360;
	13471	13568	0.003531	0.01724	0	0	0	0	0	0	1	-360	360;
	13471	2016	0.01324	0.06936	0	0	0	0	0	0	1	-360	360;
	13471	12559	0.00363	0.01743	0	0	0	0	0	0	1	-360	360;
	13471	12559	0.00257	0.01662	0	0	0	0	0	0	1	-360	360;
	13471	7396	0.01399	0.056289	0	0	0	0	0	0	1	-360	360;
	13471	1507	0.00343	0.01751	0	0	0	0	0	0	1	-360	360;
	13471	1980	0.00301	0.015409	0	0	0	0	0	0	1	-360	360;
	13471	2259	0.00857	0.033091	0	0	0	0	0	0	1	-360	360;
	9018	10058	0.003341	0.01689	0	0	0	0	0	0	1	-360	360;
	11206	4350	3.9e-05	0.000219	0	0	0	0	0	0	1	-360	360;
	10767	8175	0.00087	0.00714	0	0	0	0	0	0	1	-360	360;
	6895	13408	0.00011	0.001	0	0	0	0	0	0	1	-360	360;
	2390	7811	0.000219	0.00101	0	0	0	0	0	0	1	-360	360;
	5696	7811	0.000219	0.001031	0	0	0	0	0	0	1	-360	360;
	12456	7811	0.000271	0.00101	0	0	0	0	0	0	1	-360	360;
	5300	7811	0.000271	0.000979	0	0	0	0	0	0	1	-360	360;
	3980	7811	0.000289	0.00107	0	0	0	0	0	0	1	-360	360;
	7314	7811	0.000289	0.00105	0	0	0	0	0	0	1	-360	360;
	6389	8821	0.001591	0.012289	0	0	0	0	0	0	1	-360	360;
	3067	7638	0.000271	0.001901	0	0	0	0	0	0	1	-360	360;
	3067	9916	0.00186	0.010711	0	0	0	0	0	0	1	-360	360;
	1499	5150	1e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	5948	1242	0.000169	0.000271	0	0	0	0	0	0	1	-360	360;
	6635	1242	0.000169	0.00031	0	0	0	0	0	0	1	-360	360;
	2425	8460	0.00526	0.03787	0	0	0	0	0	0	1	-360	360;
	11923	2115	0.000229	0.000599	0	0	0	0	0	0	1	-360	360;
	8249	8034	0.00089	0.004091	0	0	0	0	0	0	1	-360	360;
	6575	12299	0.00093	0.004	0	0	0	0	0	0	1	-360	360;
	13011	12412	0.00289	0.013099	0	0	0	0	0	0	1	-360	360;
	3341	9412	0.002599	0.016479	0	0	0	0	0	0	1	-360	360;
	668	603	0.00369	0.02712	0	0	0	0	0	0	1	-360	360;
	668	5664	0.0062	0.04864	0	0	0	0	0	0	1	-360	360;
	668	495	0.008021	0.059031	0	0	0	0	0	0	1	-360	360;
	668	10138	0.004979	0.0295	0	0	0	0	0	0	1	-360	360;
	12101	5395	0.001159	0.005281	0	0	0	0	0	0	1	-360	360;
	6970	11371	0.00361	0.01824	0	0	0	0	0	0	1	-360	360;
	6970	2768	0.001521	0.01075	0	0	0	0	0	0	1	-360	360;
	6970	6217	0.00868	0.041711	0	0	0	0	0	0	1	-360	360;
	3725	7804	0.01118	0.05986	0	0	0	0	0	0	1	-360	360;
	3725	11066	0.003229	0.02543	0	0	0	0	0	0	1	-360	360;
	3725	11066	0.00218	0.01869	0	0	0	0	0	0	1	-360	360;
	3725	11066	0.001901	0.021909	0	0	0	0	0	0	1	-360	360;
	3725	11753	0.00974	0.043969	0	0	0	0	0	0	1	-360	360;
	3725	10585	0.003461	0.01814	0	0	0	0	0	0	1	-360	360;
	3725	2220	0.00487	0.027669	0	0	0	0	0	0	1	-360	360;
	11642	1657	0.000651	0.003961	0	0	0	0	0	0	1	-360	360;
	12857	8007	0.00169	0.00887	0	0	0	0	0	0	1	-360	360;
	4464	13138	0.001711	0.01336	0	0	0	0	0	0	1	-360	360;
	2238	1385	0.00031	0.000539	0	0	0	0	0	0	1	-360	360;
	2904	2041	0.0027	0.01157	0	0	0	0	0	0	1	-360	360;
	5220	934	0.000909	0.006901	0	0	0	0	0	0	1	-360	360;
	5220	1154	0.001729	0.01214	0	0	0	0	0	0	1	-360	360;
	609	5112	0.00614	0.03106	0	0	0	0	0	0	1	-360	360;
	609	11662	0.00418	0.02388	0	0	0	0	0	0	1	-360	360;
	4741	1783	0.00238	0.017771	0	0	0	0	0	0	1	-360	360;
	4741	2574	8.1e-05	0.00063	0	0	0	0	0	0	1	-360	360;
	10813	544	0.001599	0.01618	0	0	0	0	0	0	1	-360	360;
	6240	3018	0.00532	0.026961	0	0	0	0	0	0	1	-360	360;
	6240	3018	0.00361	0.02814	0	0	0	0	0	0	1	-360	360;
	6240	8095	0.002469	0.01025	0	0	0	0	0	0	1	-360	360;
	9324	2581	0.005229	0.02675	0	0	0	0	0	0	1	-360	360;
	9324	2581	0.00512	0.02726	0	0	0	0	0	0	1	-360	360;
	9324	11544	0.00355	0.01694	0	0	0	0	0	0	1	-360	360;
	6927	10683	0.00139	0.007659	0	0	0	0	0	0	1	-360	360;
	6927	5877	0.000169	0.000669	0	0	0	0	0	0	1	-360	360;
	12800	3285	0.00111	0.00712	0	0	0	0	0	0	1	-360	360;
	1275	7576	0.00031	0.001901	0	0	0	0	0	0	1	-360	360;
	7509	7576	0.000281	0.002091	0	0	0	0	0	0	1	-360	360;
	6178	9593	0.000281	0.001651	0	0	0	0	0	0	1	-360	360;
	10735	9593	0.000271	0.00169	0	0	0	0	0	0	1	-360	360;
	8522	2863	0.001521	0.00843	0	0	0	0	0	0	1	-360	360;
	8522	894	0.000841	0.0062	0	0	0	0	0	0	1	-360	360;
	5598	3560	0.00019	0.00126	0	0	0	0	0	0	1	-360	360;
	1261	9677	0.000219	0.00143	0	0	0	0	0	0	1	-360	360;
	5236	11927	0.00121	0.01796	0	0	0	0	0	0	1	-360	360;
	5236	11927	0.0014	0.01512	0	0	0	0	0	0	1	-360	360;
	9561	3083	0.008409	0.063651	0	0	0	0	0	0	1	-360	360;
	9561	12662	0.00918	0.07344	0	0	0	0	0	0	1	-360	360;
	9561	889	0.004219	0.0345	0	0	0	0	0	0	1	-360	360;
	9561	889	0.0042	0.031651	0	0	0	0	0	0	1	-360	360;
	6589	11823	0.00121	0.01381	0	0	0	0	0	0	1	-360	360;
	6589	10309	0.00107	0.01326	0	0	0	0	0	0	1	-360	360;
	7804	8460	0.003401	0.02607	0	0	0	0	0	0	1	-360	360;
	5415	26	0.000289	0.00138	0	0	0	0	0	0	1	-360	360;
	7503	955	0.001031	0.006039	0	0	0	0	0	0	1	-360	360;
	1893	955	0.00105	0.005909	0	0	0	0	0	0	1	-360	360;
	12449	11161	0.001091	0.007901	0	0	0	0	0	0	1	-360	360;
	12449	1621	0.00332	0.024021	0	0	0	0	0	0	1	-360	360;
	12449	4436	0.00344	0.021419	0	0	0	0	0	0	1	-360	360;
	12449	459	0.00095	0.010419	0	0	0	0	0	0	1	-360	360;
	12449	459	0.00093	0.01062	0	0	0	0	0	0	1	-360	360;
	12449	10966	0.00474	0.025409	0	0	0	0	0	0	1	-360	360;
	617	2957	0.000271	0.000841	0	0	0	0	0	0	1	-360	360;
	10959	9125	0.000271	0.00119	0	0	0	0	0	0	1	-360	360;
	10959	2230	0.000289	0.00076	0	0	0	0	0	0	1	-360	360;
	11400	2350	0.00031	0.00454	0	0	0	0	0	0	1	-360	360;
	11888	3448	0.00032	0.0044	0	0	0	0	0	0	1	-360	360;
	13198	2350	0.00028	0.00494	0	0	0	0	0	0	1	-360	360;
	1373	3448	0.00028	0.00508	0	0	0	0	0	0	1	-360	360;
	13181	2366	0.000531	0.00093	0	0	0	0	0	0	1	-360	360;
	13181	7203	0.0003	0.001031	0	0	0	0	0	0	1	-360	360;
	3805	13455	0.00139	0.02163	0	0	0	0	0	0	1	-360	360;
	3805	13130	0.00164	0.02321	0	0	0	0	0	0	1	-360	360;
	5160	13455	0.00136	0.02205	0	0	0	0	0	0	1	-360	360;
	5160	13130	0.0016	0.02367	0	0	0	0	0	0	1	-360	360;
	2538	12744	0.002091	0.01174	0	0	0	0	0	0	1	-360	360;
	2538	12744	0.001719	0.00961	0	0	0	0	0	0	1	-360	360;
	8217	1464	0.012341	0.076781	0	0	0	0	0	0	1	-360	360;
	8217	12744	0.00176	0.009419	0	0	0	0	0	0	1	-360	360;
	1578	4134	0.00031	0.00182	0	0	0	0	0	0	1	-360	360;
	10821	3917	0.00386	0.02348	0	0	0	0	0	0	1	-360	360;
	10821	7013	0.00017	0.00104	0	0	0	0	0	0	1	-360	360;
	5197	1229	0.00061	0.00808	0	0	0	0	0	0	1	-360	360;
	5197	5623	0.00059	0.0067	0	0	0	0	0	0	1	-360	360;
	3395	1465	0.00059	0.00748	0	0	0	0	0	0	1	-360	360;
	3395	8232	0.00061	0.0063	0	0	0	0	0	0	1	-360	360;
	544	1679	0.001031	0.008789	0	0	0	0	0	0	1	-360	360;
	544	5121	0.00051	0.006169	0	0	0	0	0	0	1	-360	360;
	544	12913	0.0005	0.00619	0	0	0	0	0	0	1	-360	360;
	2615	11522	0.00049	0.00376	0	0	0	0	0	0	1	-360	360;
	2615	11522	0.000539	0.003229	0	0	0	0	0	0	1	-360	360;
	2615	5824	0.001031	0.00882	0	0	0	0	0	0	1	-360	360;
	2615	13324	0.000711	0.00463	0	0	0	0	0	0	1	-360	360;
	2615	3181	0.000711	0.004539	0	0	0	0	0	0	1	-360	360;
	5845	3540	0.00206	0.015341	0	0	0	0	0	0	1	-360	360;
	5845	3540	0.002021	0.01564	0	0	0	0	0	0	1	-360	360;
	8924	13110	0.00094	0.005781	0	0	0	0	0	0	1	-360	360;
	8924	1420	0.002219	0.016969	0	0	0	0	0	0	1	-360	360;
	11081	2581	0.00238	0.01187	0	0	0	0	0	0	1	-360	360;
	11081	4134	0.00707	0.03095	0	0	0	0	0	0	1	-360	360;
	32	9685	0.00256	0.03401	0	0	0	0	0	0	1	-360	360;
	32	3556	0.00203	0.02416	0	0	0	0	0	0	1	-360	360;
	32	6741	0.001	0.01459	0	0	0	0	0	0	1	-360	360;
	32	6212	0.00103	0.00987	0	0	0	0	0	0	1	-360	360;
	32	7515	0.00038	0.00483	0	0	0	0	0	0	1	-360	360;
	32	7515	0.00041	0.00398	0	0	0	0	0	0	1	-360	360;
	11274	3556	0.00167	0.01721	0	0	0	0	0	0	1	-360	360;
	11274	3010	0.00098	0.01487	0	0	0	0	0	0	1	-360	360;
	11274	13621	0.00042	0.00342	0	0	0	0	0	0	1	-360	360;
	11274	13621	0.00019	0.00248	0	0	0	0	0	0	1	-360	360;
	11274	6212	0.00105	0.00966	0	0	0	0	0	0	1	-360	360;
	11274	5150	0.00043	0.00473	0	0	0	0	0	0	1	-360	360;
	11274	5150	0.00041	0.00397	0	0	0	0	0	0	1	-360	360;
	2677	7219	0.00051	0.00514	0	0	0	0	0	0	1	-360	360;
	167	7219	0.00053	0.00503	0	0	0	0	0	0	1	-360	360;
	11371	5520	0.00455	0.022771	0	0	0	0	0	0	1	-360	360;
	11371	5520	0.00557	0.02055	0	0	0	0	0	0	1	-360	360;
	11371	5520	0.00339	0.02026	0	0	0	0	0	0	1	-360	360;
	7251	13274	0.00237	0.01068	0	0	0	0	0	0	1	-360	360;
	7251	11276	0.0037	0.019479	0	0	0	0	0	0	1	-360	360;
	8935	11583	0.01407	0.056849	0	0	0	0	0	0	1	-360	360;
	8935	6666	0.001669	0.015409	0	0	0	0	0	0	1	-360	360;
	8935	6666	0.001711	0.015091	0	0	0	0	0	0	1	-360	360;
	12716	5445	0.00172	0.01619	0	0	0	0	0	0	1	-360	360;
	12716	11033	0.00029	0.00447	0	0	0	0	0	0	1	-360	360;
	12716	11033	0.00033	0.00387	0	0	0	0	0	0	1	-360	360;
	12716	6081	0.00079	0.01154	0	0	0	0	0	0	1	-360	360;
	12716	6081	0.00077	0.01177	0	0	0	0	0	0	1	-360	360;
	12716	6187	0.00127	0.01498	0	0	0	0	0	0	1	-360	360;
	2529	235	0.00214	0.012599	0	0	0	0	0	0	1	-360	360;
	6234	3473	0.000159	0.00125	0	0	0	0	0	0	1	-360	360;
	5446	7726	0.000211	0.000669	0	0	0	0	0	0	1	-360	360;
	1087	12264	0.000341	0.001169	0	0	0	0	0	0	1	-360	360;
	1087	13332	0.00024	0.001469	0	0	0	0	0	0	1	-360	360;
	1087	4469	0.00018	0.000669	0	0	0	0	0	0	1	-360	360;
	3609	2580	0.00033	0.00683	0	0	0	0	0	0	1	-360	360;
	3609	5981	0.00121	0.01211	0	0	0	0	0	0	1	-360	360;
	3609	11182	0.00123	0.01185	0	0	0	0	0	0	1	-360	360;
	3609	6212	0.00058	0.0097	0	0	0	0	0	0	1	-360	360;
	3609	7515	0.00103	0.01292	0	0	0	0	0	0	1	-360	360;
	3609	5150	0.00116	0.01266	0	0	0	0	0	0	1	-360	360;
	10683	5877	0.0013	0.006841	0	0	0	0	0	0	1	-360	360;
	4349	1507	6e-05	0.000349	0	0	0	0	0	0	1	-360	360;
	3865	7567	0.000169	0.0007	0	0	0	0	0	0	1	-360	360;
	12696	2876	0.000169	0.000719	0	0	0	0	0	0	1	-360	360;
	1265	11146	0.000271	0.001031	0	0	0	0	0	0	1	-360	360;
	1798	5367	0.008271	0.036841	0	0	0	0	0	0	1	-360	360;
	7878	3942	0.002969	0.02299	0	0	0	0	0	0	1	-360	360;
	9	6777	0.00124	0.01273	0	0	0	0	0	0	1	-360	360;
	9	11015	0.00106	0.00989	0	0	0	0	0	0	1	-360	360;
	9	7711	7e-05	0.00029	0	0	0	0	0	0	1	-360	360;
	9	7889	9e-05	0.00079	0	0	0	0	0	0	1	-360	360;
	9	2936	0.00196	0.02049	0	0	0	0	0	0	1	-360	360;
	9	6905	0.00158	0.01519	0	0	0	0	0	0	1	-360	360;
	9087	3917	0.00172	0.01811	0	0	0	0	0	0	1	-360	360;
	9087	10996	0.00096	0.01121	0	0	0	0	0	0	1	-360	360;
	9087	2936	0.00147	0.01707	0	0	0	0	0	0	1	-360	360;
	11842	6777	0.00129	0.0129	0	0	0	0	0	0	1	-360	360;
	11842	6675	0.00179	0.01737	0	0	0	0	0	0	1	-360	360;
	11842	2936	0.00131	0.01935	0	0	0	0	0	0	1	-360	360;
	2951	1217	0.003901	0.01614	0	0	0	0	0	0	1	-360	360;
	2951	1217	0.002591	0.014771	0	0	0	0	0	0	1	-360	360;
	2951	1217	0.00224	0.0173	0	0	0	0	0	0	1	-360	360;
	2951	11912	0.00012	0.000849	0	0	0	0	0	0	1	-360	360;
	2951	5097	0.00394	0.03039	0	0	0	0	0	0	1	-360	360;
	2951	5293	0.00189	0.01	0	0	0	0	0	0	1	-360	360;
	2951	3994	0.00032	0.00257	0	0	0	0	0	0	1	-360	360;
	2951	3994	0.00038	0.00224	0	0	0	0	0	0	1	-360	360;
	2951	12485	0.00751	0.03955	0	0	0	0	0	0	1	-360	360;
	2951	1632	0.00086	0.006531	0	0	0	0	0	0	1	-360	360;
	2951	27	0.00087	0.006591	0	0	0	0	0	0	1	-360	360;
	13110	1420	0.003031	0.02276	0	0	0	0	0	0	1	-360	360;
	13110	2958	0.006331	0.03911	0	0	0	0	0	0	1	-360	360;
	13110	6475	0.00605	0.02986	0	0	0	0	0	0	1	-360	360;
	13110	9795	0.003919	0.01612	0	0	0	0	0	0	1	-360	360;
	13110	7048	0.002039	0.0108	0	0	0	0	0	0	1	-360	360;
	13110	10497	0.006289	0.026271	0	0	0	0	0	0	1	-360	360;
	4225	4088	0.00042	0.00547	0	0	0	0	0	0	1	-360	360;
	4225	4088	0.00043	0.00536	0	0	0	0	0	0	1	-360	360;
	4225	227	0.00147	0.01373	0	0	0	0	0	0	1	-360	360;
	4225	227	0.00131	0.01555	0	0	0	0	0	0	1	-360	360;
	4225	2348	0.00221	0.02217	0	0	0	0	0	0	1	-360	360;
	4225	1158	0.00224	0.0217	0	0	0	0	0	0	1	-360	360;
	13581	9920	0.00095	0.003289	0	0	0	0	0	0	1	-360	360;
	13581	4966	0.000229	0.00082	0	0	0	0	0	0	1	-360	360;
	13581	2587	0.00068	0.00306	0	0	0	0	0	0	1	-360	360;
	13581	5472	0.00049	0.00256	0	0	0	0	0	0	1	-360	360;
	13581	5449	0.000531	0.0022	0	0	0	0	0	0	1	-360	360;
	13581	10989	0.00136	0.00956	0	0	0	0	0	0	1	-360	360;
	13581	10989	0.0012	0.010729	0	0	0	0	0	0	1	-360	360;
	13581	10989	0.00105	0.00861	0	0	0	0	0	0	1	-360	360;
	13581	3444	0.000591	0.006969	0	0	0	0	0	0	1	-360	360;
	4005	12552	0.00025	0.0008	0	0	0	0	0	0	1	-360	360;
	4005	319	0.00113	0.00632	0	0	0	0	0	0	1	-360	360;
	665	3010	0.0003	0.00275	0	0	0	0	0	0	1	-360	360;
	7464	1829	0.00836	0.051659	0	0	0	0	0	0	1	-360	360;
	7464	1829	0.01338	0.073469	0	0	0	0	0	0	1	-360	360;
	7464	2508	0.0027	0.02089	0	0	0	0	0	0	1	-360	360;
	7464	6475	0.003419	0.02636	0	0	0	0	0	0	1	-360	360;
	7464	9358	0.009531	0.033711	0	0	0	0	0	0	1	-360	360;
	3506	11823	0.0004	0.00707	0	0	0	0	0	0	1	-360	360;
	3768	11823	0.00035	0.00701	0	0	0	0	0	0	1	-360	360;
	11925	9913	0.00194	0.01751	0	0	0	0	0	0	1	-360	360;
	11925	9913	0.00156	0.01936	0	0	0	0	0	0	1	-360	360;
	6777	1361	0.00078	0.00941	0	0	0	0	0	0	1	-360	360;
	6777	1361	0.00074	0.00969	0	0	0	0	0	0	1	-360	360;
	6777	5846	0.00042	0.00532	0	0	0	0	0	0	1	-360	360;
	6777	5846	0.0004	0.00473	0	0	0	0	0	0	1	-360	360;
	6777	5846	0.0006	0.00499	0	0	0	0	0	0	1	-360	360;
	2296	3455	0.00086	0.00638	0	0	0	0	0	0	1	-360	360;
	2296	3455	0.000969	0.00625	0	0	0	0	0	0	1	-360	360;
	2296	13302	0.002781	0.01725	0	0	0	0	0	0	1	-360	360;
	5589	1162	0.005031	0.028081	0	0	0	0	0	0	1	-360	360;
	7975	1356	1e-05	0.00021	0	0	0	0	0	0	1	-360	360;
	1892	1356	1e-05	0.00025	0	0	0	0	0	0	1	-360	360;
	742	1356	2e-05	0.00031	0	0	0	0	0	0	1	-360	360;
	4294	1356	3e-05	0.0004	0	0	0	0	0	0	1	-360	360;
	6679	9331	0.000841	0.00474	0	0	0	0	0	0	1	-360	360;
	6679	9331	0.00082	0.00487	0	0	0	0	0	0	1	-360	360;
	9996	2274	0.00186	0.011979	0	0	0	0	0	0	1	-360	360;
	12212	7779	0.00056	0.003021	0	0	0	0	0	0	1	-360	360;
	9082	7779	0.00055	0.003081	0	0	0	0	0	0	1	-360	360;
	13605	13583	0.00114	0.01063	0	0	0	0	0	0	1	-360	360;
	3831	8642	0.003719	0.02295	0	0	0	0	0	0	1	-360	360;
	3831	10399	0.006669	0.04288	0	0	0	0	0	0	1	-360	360;
	4216	8428	0.00243	0.0198	0	0	0	0	0	0	1	-360	360;
	4216	5428	0.00845	0.04339	0	0	0	0	0	0	1	-360	360;
	3735	12862	0.00219	0.015169	0	0	0	0	0	0	1	-360	360;
	3735	3018	0.00357	0.0302	0	0	0	0	0	0	1	-360	360;
	7343	11066	0.005849	0.032419	0	0	0	0	0	0	1	-360	360;
	13077	1621	0.00406	0.01688	0	0	0	0	0	0	1	-360	360;
	10818	4848	0.00113	0.0065	0	0	0	0	0	0	1	-360	360;
	1786	6497	0.0068	0.04613	0	0	0	0	0	0	1	-360	360;
	1786	2804	0.00337	0.02011	0	0	0	0	0	0	1	-360	360;
	11861	10494	0.00011	0.0005	0	0	0	0	0	0	1	-360	360;
	10112	7192	0.003711	0.018521	0	0	0	0	0	0	1	-360	360;
	10112	1835	0.000271	0.00145	0	0	0	0	0	0	1	-360	360;
	12331	6906	0.00326	0.01824	0	0	0	0	0	0	1	-360	360;
	13250	10444	0.00118	0.00712	0	0	0	0	0	0	1	-360	360;
	13250	10058	0.00482	0.02232	0	0	0	0	0	0	1	-360	360;
	10782	10258	0.00013	0.000469	0	0	0	0	0	0	1	-360	360;
	10782	9125	0.000229	0.00075	0	0	0	0	0	0	1	-360	360;
	11595	2863	0.004591	0.035021	0	0	0	0	0	0	1	-360	360;
	11595	2050	0.005469	0.03924	0	0	0	0	0	0	1	-360	360;
	11595	2050	0.00536	0.0403	0	0	0	0	0	0	1	-360	360;
	11595	169	0.00424	0.033031	0	0	0	0	0	0	1	-360	360;
	8137	6311	2e-05	0.00024	0	0	0	0	0	0	1	-360	360;
	8137	2523	1e-05	0.00021	0	0	0	0	0	0	1	-360	360;
	8137	5445	0.00286	0.03653	0	0	0	0	0	0	1	-360	360;
	8137	6297	0.00117	0.01183	0	0	0	0	0	0	1	-360	360;
	8137	5315	0.00103	0.01336	0	0	0	0	0	0	1	-360	360;
	8137	13480	0.00156	0.01969	0	0	0	0	0	0	1	-360	360;
	8137	9592	0.00449	0.05772	0	0	0	0	0	0	1	-360	360;
	8137	5299	0.00259	0.03469	0	0	0	0	0	0	1	-360	360;
	1420	6073	0.000219	0.00118	0	0	0	0	0	0	1	-360	360;
	1420	2820	6e-05	0.00095	0	0	0	0	0	0	1	-360	360;
	1420	4636	6e-05	0.000651	0	0	0	0	0	0	1	-360	360;
	1420	5253	8.1e-05	0.000841	0	0	0	0	0	0	1	-360	360;
	1420	6416	0.00274	0.020831	0	0	0	0	0	0	1	-360	360;
	1420	8183	0.004781	0.027729	0	0	0	0	0	0	1	-360	360;
	1420	8183	0.002849	0.02093	0	0	0	0	0	0	1	-360	360;
	1420	9795	0.00513	0.03014	0	0	0	0	0	0	1	-360	360;
	1420	10497	0.00113	0.00651	0	0	0	0	0	0	1	-360	360;
	1420	7861	0.00125	0.00707	0	0	0	0	0	0	1	-360	360;
	10896	7779	0.003349	0.01463	0	0	0	0	0	0	1	-360	360;
	10896	2184	0.00245	0.0122	0	0	0	0	0	0	1	-360	360;
	1844	1829	0.00587	0.036521	0	0	0	0	0	0	1	-360	360;
	1844	1829	0.00569	0.03774	0	0	0	0	0	0	1	-360	360;
	1844	1829	0.006031	0.03545	0	0	0	0	0	0	1	-360	360;
	1844	11576	0.000531	0.00301	0	0	0	0	0	0	1	-360	360;
	1844	9412	0.00076	0.006099	0	0	0	0	0	0	1	-360	360;
	1126	10564	0.008841	0.028919	0	0	0	0	0	0	1	-360	360;
	1126	5757	0.00169	0.007091	0	0	0	0	0	0	1	-360	360;
	1126	5709	0.001031	0.00361	0	0	0	0	0	0	1	-360	360;
	1126	4437	0.00813	0.030961	0	0	0	0	0	0	1	-360	360;
	7726	9143	0.00013	0.00068	0	0	0	0	0	0	1	-360	360;
	5396	5662	0.000211	0.000729	0	0	0	0	0	0	1	-360	360;
	5396	5662	0.000169	0.000669	0	0	0	0	0	0	1	-360	360;
	5396	4212	0.000669	0.00251	0	0	0	0	0	0	1	-360	360;
	5396	4212	0.000711	0.00238	0	0	0	0	0	0	1	-360	360;
	5396	1679	0.000229	0.00238	0	0	0	0	0	0	1	-360	360;
	5396	5121	0.0007	0.0065	0	0	0	0	0	0	1	-360	360;
	5396	12913	0.00061	0.006031	0	0	0	0	0	0	1	-360	360;
	9470	7807	0.000281	0.0018	0	0	0	0	0	0	1	-360	360;
	9470	5824	0.000211	0.002409	0	0	0	0	0	0	1	-360	360;
	1707	1905	0.00056	0.00773	0	0	0	0	0	0	1	-360	360;
	1707	36	0.0023	0.01581	0	0	0	0	0	0	1	-360	360;
	10943	5367	0.005039	0.032151	0	0	0	0	0	0	1	-360	360;
	10943	140	0.00364	0.01713	0	0	0	0	0	0	1	-360	360;
	10943	2663	0.002169	0.012409	0	0	0	0	0	0	1	-360	360;
	518	4608	0.00113	0.004719	0	0	0	0	0	0	1	-360	360;
	518	9611	0.00113	0.004979	0	0	0	0	0	0	1	-360	360;
	4608	6928	0.00264	0.01174	0	0	0	0	0	0	1	-360	360;
	3803	11583	0.00494	0.03181	0	0	0	0	0	0	1	-360	360;
	3803	3018	0.003711	0.023961	0	0	0	0	0	0	1	-360	360;
	6084	13033	0.001409	0.01101	0	0	0	0	0	0	1	-360	360;
	6084	2316	0.001169	0.00593	0	0	0	0	0	0	1	-360	360;
	6084	2316	0.00118	0.00587	0	0	0	0	0	0	1	-360	360;
	6084	5559	0.00264	0.012469	0	0	0	0	0	0	1	-360	360;
	6084	5559	0.002349	0.01414	0	0	0	0	0	0	1	-360	360;
	5375	6031	0.00014	0.0007	0	0	0	0	0	0	1	-360	360;
	9685	7792	3e-05	0.00054	0	0	0	0	0	0	1	-360	360;
	9685	10491	3e-05	0.00055	0	0	0	0	0	0	1	-360	360;
	9685	5404	4e-05	0.00058	0	0	0	0	0	0	1	-360	360;
	9685	7983	4e-05	0.00054	0	0	0	0	0	0	1	-360	360;
	9685	6212	0.00171	0.02197	0	0	0	0	0	0	1	-360	360;
	9685	6212	0.00164	0.02064	0	0	0	0	0	0	1	-360	360;
	9685	6599	0.00136	0.01756	0	0	0	0	0	0	1	-360	360;
	9685	634	0.00097	0.00942	0	0	0	0	0	0	1	-360	360;
	9685	634	0.00086	0.01109	0	0	0	0	0	0	1	-360	360;
	9685	634	0.00085	0.0113	0	0	0	0	0	0	1	-360	360;
	2212	10056	0.000979	0.0058	0	0	0	0	0	0	1	-360	360;
	3556	4999	0.00052	0.00593	0	0	0	0	0	0	1	-360	360;
	3556	4999	0.00046	0.00672	0	0	0	0	0	0	1	-360	360;
	7192	4790	0.001151	0.00636	0	0	0	0	0	0	1	-360	360;
	7192	4790	0.001031	0.007211	0	0	0	0	0	0	1	-360	360;
	7192	6906	0.002341	0.02286	0	0	0	0	0	0	1	-360	360;
	7192	1835	0.002281	0.02518	0	0	0	0	0	0	1	-360	360;
	7192	5757	0.005479	0.02114	0	0	0	0	0	0	1	-360	360;
	7192	4299	0.011159	0.08856	0	0	0	0	0	0	1	-360	360;
	7192	8517	0.000669	0.00249	0	0	0	0	0	0	1	-360	360;
	7192	11609	0.00193	0.01643	0	0	0	0	0	0	1	-360	360;
	7192	4437	0.00024	0.0017	0	0	0	0	0	0	1	-360	360;
	3788	281	0.010159	0.065849	0	0	0	0	0	0	1	-360	360;
	710	10802	0.00294	0.02131	0	0	0	0	0	0	1	-360	360;
	710	12871	0.018841	0.07469	0	0	0	0	0	0	1	-360	360;
	710	9296	0.015599	0.052169	0	0	0	0	0	0	1	-360	360;
	11583	4519	0.005831	0.03969	0	0	0	0	0	0	1	-360	360;
	11583	3527	0.00087	0.00661	0	0	0	0	0	0	1	-360	360;
	11583	8095	0.00637	0.024659	0	0	0	0	0	0	1	-360	360;
	9391	4435	0.0027	0.020419	0	0	0	0	0	0	1	-360	360;
	9391	5691	0.000409	0.00314	0	0	0	0	0	0	1	-360	360;
	9391	13022	0.001669	0.01301	0	0	0	0	0	0	1	-360	360;
	9391	13022	0.00164	0.01325	0	0	0	0	0	0	1	-360	360;
	1185	6675	0.00089	0.00556	0	0	0	0	0	0	1	-360	360;
	1185	6061	0.00186	0.01809	0	0	0	0	0	0	1	-360	360;
	1185	4880	0.00381	0.02853	0	0	0	0	0	0	1	-360	360;
	8038	3401	0.012419	0.072151	0	0	0	0	0	0	1	-360	360;
	8038	1984	0.0042	0.02556	0	0	0	0	0	0	1	-360	360;
	8038	4090	0.010771	0.0493	0	0	0	0	0	0	1	-360	360;
	8038	2729	0.010159	0.04995	0	0	0	0	0	0	1	-360	360;
	8038	12978	0.013169	0.05113	0	0	0	0	0	0	1	-360	360;
	12552	10790	0.000469	0.001159	0	0	0	0	0	0	1	-360	360;
	12552	9030	0.000229	0.00105	0	0	0	0	0	0	1	-360	360;
	4966	1385	0.000271	0.001271	0	0	0	0	0	0	1	-360	360;
	2975	6741	0.00099	0.01194	0	0	0	0	0	0	1	-360	360;
	2975	6661	0.00077	0.01184	0	0	0	0	0	0	1	-360	360;
	2975	7515	0.00034	0.00426	0	0	0	0	0	0	1	-360	360;
	7389	3010	0.00101	0.01168	0	0	0	0	0	0	1	-360	360;
	7389	6661	0.00086	0.01046	0	0	0	0	0	0	1	-360	360;
	7389	5150	0.00035	0.00417	0	0	0	0	0	0	1	-360	360;
	3790	2115	0.000159	0.00069	0	0	0	0	0	0	1	-360	360;
	11917	8496	0.000781	0.00488	0	0	0	0	0	0	1	-360	360;
	11917	6101	0.000781	0.00511	0	0	0	0	0	0	1	-360	360;
	11917	4521	0.00064	0.003539	0	0	0	0	0	0	1	-360	360;
	11917	4521	0.00057	0.003781	0	0	0	0	0	0	1	-360	360;
	11917	4352	0.000531	0.00111	0	0	0	0	0	0	1	-360	360;
	11917	10961	0.000521	0.00113	0	0	0	0	0	0	1	-360	360;
	11917	8724	0.00176	0.00843	0	0	0	0	0	0	1	-360	360;
	11917	13022	0.002461	0.01975	0	0	0	0	0	0	1	-360	360;
	2441	9437	0.000591	0.00105	0	0	0	0	0	0	1	-360	360;
	2903	2016	0.00012	0.00064	0	0	0	0	0	0	1	-360	360;
	3086	2581	0.000581	0.004419	0	0	0	0	0	0	1	-360	360;
	9315	2581	0.000651	0.004331	0	0	0	0	0	0	1	-360	360;
	6911	4108	0.001289	0.008289	0	0	0	0	0	0	1	-360	360;
	7274	13490	0.00233	0.03502	0	0	0	0	0	0	1	-360	360;
	7274	13490	0.00238	0.03781	0	0	0	0	0	0	1	-360	360;
	7274	2937	0.00052	0.00605	0	0	0	0	0	0	1	-360	360;
	7274	7948	0.00023	0.00275	0	0	0	0	0	0	1	-360	360;
	7274	11927	0.00134	0.01343	0	0	0	0	0	0	1	-360	360;
	7274	11927	0.00116	0.01552	0	0	0	0	0	0	1	-360	360;
	11696	1027	0.004169	0.027919	0	0	0	0	0	0	1	-360	360;
	11696	1162	0.00063	0.008	0	0	0	0	0	0	1	-360	360;
	11696	2660	0.004841	0.028581	0	0	0	0	0	0	1	-360	360;
	11696	7985	0.007419	0.05787	0	0	0	0	0	0	1	-360	360;
	6265	3540	0.002719	0.020331	0	0	0	0	0	0	1	-360	360;
	3320	9003	0.000789	0.005	0	0	0	0	0	0	1	-360	360;
	13572	281	0.00238	0.013219	0	0	0	0	0	0	1	-360	360;
	13572	4528	0.003771	0.02701	0	0	0	0	0	0	1	-360	360;
	6860	6666	0.001271	0.00939	0	0	0	0	0	0	1	-360	360;
	13308	9842	0.000841	0.00131	0	0	0	0	0	0	1	-360	360;
	12435	5520	0.005711	0.043039	0	0	0	0	0	0	1	-360	360;
	12435	5520	0.00562	0.041901	0	0	0	0	0	0	1	-360	360;
	12561	4228	0.00356	0.019781	0	0	0	0	0	0	1	-360	360;
	12561	235	0.004841	0.03149	0	0	0	0	0	0	1	-360	360;
	1082	12547	0.00057	0.00107	0	0	0	0	0	0	1	-360	360;
	10090	7203	0.000539	0.00213	0	0	0	0	0	0	1	-360	360;
	655	6955	0.0007	0.00425	0	0	0	0	0	0	1	-360	360;
	2065	8040	0.0012	0.009531	0	0	0	0	0	0	1	-360	360;
	2065	3942	0.00206	0.015901	0	0	0	0	0	0	1	-360	360;
	5291	6928	0.00063	0.00312	0	0	0	0	0	0	1	-360	360;
	1768	2936	3e-05	0.00041	0	0	0	0	0	0	1	-360	360;
	1574	2936	2e-05	0.00043	0	0	0	0	0	0	1	-360	360;
	12142	2936	3e-05	0.00039	0	0	0	0	0	0	1	-360	360;
	12250	2936	3e-05	0.00038	0	0	0	0	0	0	1	-360	360;
	2008	5132	0.0008	0.004151	0	0	0	0	0	0	1	-360	360;
	6674	5132	0.000901	0.00381	0	0	0	0	0	0	1	-360	360;
	4438	934	0.000591	0.004341	0	0	0	0	0	0	1	-360	360;
	4438	934	0.000591	0.00405	0	0	0	0	0	0	1	-360	360;
	7324	6661	0.00218	0.0275	0	0	0	0	0	0	1	-360	360;
	7324	4416	0.00184	0.01677	0	0	0	0	0	0	1	-360	360;
	1100	1361	0.00058	0.00807	0	0	0	0	0	0	1	-360	360;
	1100	1361	0.00056	0.00823	0	0	0	0	0	0	1	-360	360;
	1100	7219	0.00123	0.01146	0	0	0	0	0	0	1	-360	360;
	1100	7219	0.0011	0.01298	0	0	0	0	0	0	1	-360	360;
	1100	5846	0.00156	0.01585	0	0	0	0	0	0	1	-360	360;
	1100	4194	0.00226	0.02781	0	0	0	0	0	0	1	-360	360;
	5520	267	0.00614	0.02831	0	0	0	0	0	0	1	-360	360;
	5520	3199	0.00513	0.03401	0	0	0	0	0	0	1	-360	360;
	9969	6047	0.013219	0.0597	0	0	0	0	0	0	1	-360	360;
	8345	10214	0.012969	0.06087	0	0	0	0	0	0	1	-360	360;
	12745	10970	0.00437	0.032831	0	0	0	0	0	0	1	-360	360;
	12745	3540	0.005651	0.033229	0	0	0	0	0	0	1	-360	360;
	5600	7985	0.008969	0.03625	0	0	0	0	0	0	1	-360	360;
	5600	11395	0.012409	0.05645	0	0	0	0	0	0	1	-360	360;
	7320	9331	0.00038	0.002271	0	0	0	0	0	0	1	-360	360;
	7576	6011	0.00138	0.01089	0	0	0	0	0	0	1	-360	360;
	7576	6011	0.001539	0.012521	0	0	0	0	0	0	1	-360	360;
	7576	11295	0.00205	0.014229	0	0	0	0	0	0	1	-360	360;
	7576	10636	0.00026	0.00164	0	0	0	0	0	0	1	-360	360;
	7576	3550	0.000289	0.00169	0	0	0	0	0	0	1	-360	360;
	7779	11074	0.003349	0.051211	0	0	0	0	0	0	1	-360	360;
	8208	6738	0.00038	0.001479	0	0	0	0	0	0	1	-360	360;
	2271	8175	0.00082	0.00625	0	0	0	0	0	0	1	-360	360;
	4610	4974	0.00275	0.015581	0	0	0	0	0	0	1	-360	360;
	8802	8629	0.000909	0.00739	0	0	0	0	0	0	1	-360	360;
	5445	12511	0.00082	0.01038	0	0	0	0	0	0	1	-360	360;
	5445	11017	0.00093	0.01017	0	0	0	0	0	0	1	-360	360;
	5445	13480	0.00137	0.01431	0	0	0	0	0	0	1	-360	360;
	5445	2218	0.00215	0.02021	0	0	0	0	0	0	1	-360	360;
	1829	6290	0.01189	0.058841	0	0	0	0	0	0	1	-360	360;
	9843	3627	0.00104	0.01241	0	0	0	0	0	0	1	-360	360;
	9843	3627	0.00092	0.014	0	0	0	0	0	0	1	-360	360;
	9843	3383	0.00055	0.00845	0	0	0	0	0	0	1	-360	360;
	9843	3383	0.00054	0.00861	0	0	0	0	0	0	1	-360	360;
	9843	6297	0.00105	0.01057	0	0	0	0	0	0	1	-360	360;
	9843	5315	0.00089	0.01122	0	0	0	0	0	0	1	-360	360;
	9843	10081	0.00223	0.02584	0	0	0	0	0	0	1	-360	360;
	13380	7458	0.00101	0.007841	0	0	0	0	0	0	1	-360	360;
	13380	7458	0.001081	0.008099	0	0	0	0	0	0	1	-360	360;
	705	10497	0.011081	0.068479	0	0	0	0	0	0	1	-360	360;
	13490	2424	2e-05	0.00033	0	0	0	0	0	0	1	-360	360;
	13490	13136	1e-05	0.00029	0	0	0	0	0	0	1	-360	360;
	13490	12835	0.00071	0.00892	0	0	0	0	0	0	1	-360	360;
	13490	13130	0.00136	0.01526	0	0	0	0	0	0	1	-360	360;
	8642	11395	0.005401	0.036021	0	0	0	0	0	0	1	-360	360;
	4529	2946	0.00437	0.033719	0	0	0	0	0	0	1	-360	360;
	4529	1507	0.005151	0.034919	0	0	0	0	0	0	1	-360	360;
	426	7134	0.00038	0.00139	0	0	0	0	0	0	1	-360	360;
	426	1548	0.00037	0.0015	0	0	0	0	0	0	1	-360	360;
	426	6955	0.00069	0.004461	0	0	0	0	0	0	1	-360	360;
	2863	7605	0.00688	0.040591	0	0	0	0	0	0	1	-360	360;
	2863	169	0.001341	0.008349	0	0	0	0	0	0	1	-360	360;
	2863	169	0.001229	0.009669	0	0	0	0	0	0	1	-360	360;
	3194	8360	0.00137	0.006039	0	0	0	0	0	0	1	-360	360;
	11029	9593	0.001919	0.01261	0	0	0	0	0	0	1	-360	360;
	11029	1228	0.00163	0.0095	0	0	0	0	0	0	1	-360	360;
	13162	13033	0.00488	0.0497	0	0	0	0	0	0	1	-360	360;
	13162	6605	0.001289	0.0087	0	0	0	0	0	0	1	-360	360;
	13162	5011	0.00211	0.01524	0	0	0	0	0	0	1	-360	360;
	13162	5011	0.00206	0.015539	0	0	0	0	0	0	1	-360	360;
	13162	5011	0.00276	0.01631	0	0	0	0	0	0	1	-360	360;
	13162	9593	0.00305	0.019521	0	0	0	0	0	0	1	-360	360;
	13162	3560	0.000409	0.0023	0	0	0	0	0	0	1	-360	360;
	13162	9677	0.00036	0.00268	0	0	0	0	0	0	1	-360	360;
	8182	9274	0.00032	0.00049	0	0	0	0	0	0	1	-360	360;
	1549	9274	0.00013	0.000349	0	0	0	0	0	0	1	-360	360;
	2580	6212	0.0002	0.00327	0	0	0	0	0	0	1	-360	360;
	1153	3034	0.000919	0.00561	0	0	0	0	0	0	1	-360	360;
	1153	3034	0.00094	0.00549	0	0	0	0	0	0	1	-360	360;
	1153	11003	0.00175	0.01976	0	0	0	0	0	0	1	-360	360;
	1153	9849	0.001729	0.019331	0	0	0	0	0	0	1	-360	360;
	3262	9938	0.00406	0.02581	0	0	0	0	0	0	1	-360	360;
	3262	1113	0.00526	0.025961	0	0	0	0	0	0	1	-360	360;
	4714	1554	0.00179	0.02098	0	0	0	0	0	0	1	-360	360;
	4714	8710	0.00162	0.02148	0	0	0	0	0	0	1	-360	360;
	4714	13408	0.00492	0.04991	0	0	0	0	0	0	1	-360	360;
	4714	1356	0.00109	0.01361	0	0	0	0	0	0	1	-360	360;
	4714	1356	0.00097	0.0153	0	0	0	0	0	0	1	-360	360;
	4714	7219	0.00322	0.03022	0	0	0	0	0	0	1	-360	360;
	2203	10795	0.032831	0.074151	0	0	0	0	0	0	1	-360	360;
	2203	10008	0.01411	0.058711	0	0	0	0	0	0	1	-360	360;
	2203	12559	0.02725	0.102979	0	0	0	0	0	0	1	-360	360;
	2203	4581	0.012229	0.04049	0	0	0	0	0	0	1	-360	360;
	2203	3832	0.01994	0.066719	0	0	0	0	0	0	1	-360	360;
	2203	6290	0.018591	0.07462	0	0	0	0	0	0	1	-360	360;
	2203	6646	0.008979	0.03251	0	0	0	0	0	0	1	-360	360;
	2203	6217	0.016531	0.086031	0	0	0	0	0	0	1	-360	360;
	12750	55	0.003591	0.02137	0	0	0	0	0	0	1	-360	360;
	12750	4110	0.00124	0.008	0	0	0	0	0	0	1	-360	360;
	12750	4110	0.001271	0.007831	0	0	0	0	0	0	1	-360	360;
	12750	301	0.000469	0.00294	0	0	0	0	0	0	1	-360	360;
	12750	6189	0.000841	0.003289	0	0	0	0	0	0	1	-360	360;
	3406	2740	0.00219	0.016539	0	0	0	0	0	0	1	-360	360;
	3406	8030	0.002901	0.020901	0	0	0	0	0	0	1	-360	360;
	11190	333	0.0002	0.001211	0	0	0	0	0	0	1	-360	360;
	2111	1980	9.1e-05	0.000479	0	0	0	0	0	0	1	-360	360;
	10260	10411	0.005461	0.029469	0	0	0	0	0	0	1	-360	360;
	10260	4848	0.001979	0.011289	0	0	0	0	0	0	1	-360	360;
	1765	235	0.0023	0.01289	0	0	0	0	0	0	1	-360	360;
	8957	3522	0.000159	0.00037	0	0	0	0	0	0	1	-360	360;
	13033	5559	0.002789	0.01562	0	0	0	0	0	0	1	-360	360;
	8586	8106	0.002531	0.012331	0	0	0	0	0	0	1	-360	360;
	8586	11222	0.00357	0.023289	0	0	0	0	0	0	1	-360	360;
	8586	2392	0.00076	0.006039	0	0	0	0	0	0	1	-360	360;
	5875	2184	0.000469	0.00075	0	0	0	0	0	0	1	-360	360;
	262	2184	0.00044	0.00069	0	0	0	0	0	0	1	-360	360;
	4279	5028	0.00561	0.035521	0	0	0	0	0	0	1	-360	360;
	12839	1314	0.004539	0.03468	0	0	0	0	0	0	1	-360	360;
	12839	1314	0.00175	0.0067	0	0	0	0	0	0	1	-360	360;
	12839	4994	0.007461	0.031419	0	0	0	0	0	0	1	-360	360;
	4790	4916	0.00224	0.012909	0	0	0	0	0	0	1	-360	360;
	7709	26	0.0003	0.000961	0	0	0	0	0	0	1	-360	360;
	7709	26	0.00032	0.001409	0	0	0	0	0	0	1	-360	360;
	7709	2217	0.002031	0.0157	0	0	0	0	0	0	1	-360	360;
	7709	9004	0.00038	0.00306	0	0	0	0	0	0	1	-360	360;
	7709	7132	0.00118	0.0083	0	0	0	0	0	0	1	-360	360;
	7709	7942	0.001229	0.00787	0	0	0	0	0	0	1	-360	360;
	8106	8251	0.00301	0.01538	0	0	0	0	0	0	1	-360	360;
	5032	7942	0.00063	0.00114	0	0	0	0	0	0	1	-360	360;
	5817	7132	0.00064	0.00111	0	0	0	0	0	0	1	-360	360;
	603	5664	0.004919	0.03837	0	0	0	0	0	0	1	-360	360;
	9181	5011	0.001831	0.008349	0	0	0	0	0	0	1	-360	360;
	2946	592	0.00536	0.03188	0	0	0	0	0	0	1	-360	360;
	4273	13274	0.00038	0.001461	0	0	0	0	0	0	1	-360	360;
	4273	13274	0.000289	0.001521	0	0	0	0	0	0	1	-360	360;
	13274	3559	0.00305	0.016841	0	0	0	0	0	0	1	-360	360;
	13274	11931	0.001781	0.012729	0	0	0	0	0	0	1	-360	360;
	2919	11004	0.00041	0.00461	0	0	0	0	0	0	1	-360	360;
	2919	11004	0.0004	0.0047	0	0	0	0	0	0	1	-360	360;
	6011	6400	0.00213	0.01401	0	0	0	0	0	0	1	-360	360;
	6011	3310	0.001271	0.00887	0	0	0	0	0	0	1	-360	360;
	6011	4121	0.00462	0.02743	0	0	0	0	0	0	1	-360	360;
	3734	10877	0.001539	0.00776	0	0	0	0	0	0	1	-360	360;
	10526	13215	0.00025	0.00478	0	0	0	0	0	0	1	-360	360;
	7834	13215	0.00025	0.00487	0	0	0	0	0	0	1	-360	360;
	11955	13215	0.00025	0.00435	0	0	0	0	0	0	1	-360	360;
	9269	13215	0.00026	0.00426	0	0	0	0	0	0	1	-360	360;
	1748	13215	0.00025	0.00479	0	0	0	0	0	0	1	-360	360;
	6063	13565	0.000169	0.00076	0	0	0	0	0	0	1	-360	360;
	6063	11726	0.00039	0.00126	0	0	0	0	0	0	1	-360	360;
	3484	1162	0.011229	0.074659	0	0	0	0	0	0	1	-360	360;
	3484	2660	0.014521	0.058081	0	0	0	0	0	0	1	-360	360;
	3484	13034	0.00064	0.003919	0	0	0	0	0	0	1	-360	360;
	3484	7517	0.014789	0.07768	0	0	0	0	0	0	1	-360	360;
	3484	10138	0.00938	0.05355	0	0	0	0	0	0	1	-360	360;
	3484	2483	0.00124	0.004289	0	0	0	0	0	0	1	-360	360;
	7605	12749	0.00968	0.04388	0	0	0	0	0	0	1	-360	360;
	7605	247	0.008531	0.067151	0	0	0	0	0	0	1	-360	360;
	7605	10895	0.00731	0.04106	0	0	0	0	0	0	1	-360	360;
	10768	8652	0.004349	0.02555	0	0	0	0	0	0	1	-360	360;
	10768	13214	0.00599	0.03462	0	0	0	0	0	0	1	-360	360;
	13031	1162	0.002961	0.021591	0	0	0	0	0	0	1	-360	360;
	13031	1162	0.002901	0.02206	0	0	0	0	0	0	1	-360	360;
	13031	8821	0.002979	0.016969	0	0	0	0	0	0	1	-360	360;
	8374	6148	0.00312	0.020021	0	0	0	0	0	0	1	-360	360;
	12547	2813	0.000219	0.000909	0	0	0	0	0	0	1	-360	360;
	1976	2813	0.000219	0.00093	0	0	0	0	0	0	1	-360	360;
	11866	3009	0.000151	0.000289	0	0	0	0	0	0	1	-360	360;
	7237	12150	0.00563	0.043599	0	0	0	0	0	0	1	-360	360;
	7237	6956	0.000969	0.00612	0	0	0	0	0	0	1	-360	360;
	3028	5280	0.005021	0.020719	0	0	0	0	0	0	1	-360	360;
	3028	6859	0.00988	0.041039	0	0	0	0	0	0	1	-360	360;
	6834	12739	0.003159	0.02011	0	0	0	0	0	0	1	-360	360;
	6834	12739	0.003219	0.01969	0	0	0	0	0	0	1	-360	360;
	10431	7917	0.00049	0.00238	0	0	0	0	0	0	1	-360	360;
	10431	7917	0.000479	0.002419	0	0	0	0	0	0	1	-360	360;
	12146	8318	0.00264	0.019729	0	0	0	0	0	0	1	-360	360;
	12146	8318	0.002581	0.02011	0	0	0	0	0	0	1	-360	360;
	9437	4916	0.00037	0.002831	0	0	0	0	0	0	1	-360	360;
	9437	8273	0.00213	0.012789	0	0	0	0	0	0	1	-360	360;
	3401	7229	0.018669	0.092281	0	0	0	0	0	0	1	-360	360;
	1271	13444	3.9e-05	0.000401	0	0	0	0	0	0	1	-360	360;
	5960	13444	6e-05	0.00044	0	0	0	0	0	0	1	-360	360;
	9987	3034	6e-05	0.0005	0	0	0	0	0	0	1	-360	360;
	8251	3785	0.001271	0.007031	0	0	0	0	0	0	1	-360	360;
	8251	11074	0.00055	0.003159	0	0	0	0	0	0	1	-360	360;
	8251	11074	0.00055	0.00338	0	0	0	0	0	0	1	-360	360;
	8251	11074	0.00105	0.004771	0	0	0	0	0	0	1	-360	360;
	8251	2320	0.003289	0.015039	0	0	0	0	0	0	1	-360	360;
	1053	1000	0.00036	0.004349	0	0	0	0	0	0	1	-360	360;
	5174	1000	0.00038	0.003669	0	0	0	0	0	0	1	-360	360;
	9331	3473	0.00118	0.00701	0	0	0	0	0	0	1	-360	360;
	9331	3473	0.001229	0.00686	0	0	0	0	0	0	1	-360	360;
	9331	8605	0.00301	0.0217	0	0	0	0	0	0	1	-360	360;
	9331	8605	0.003409	0.01924	0	0	0	0	0	0	1	-360	360;
	9331	8605	0.003409	0.01938	0	0	0	0	0	0	1	-360	360;
	11532	13448	0.002719	0.02131	0	0	0	0	0	0	1	-360	360;
	9730	2050	0.00295	0.01956	0	0	0	0	0	0	1	-360	360;
	7462	8273	0.001091	0.00831	0	0	0	0	0	0	1	-360	360;
	4275	8605	0.00111	0.008211	0	0	0	0	0	0	1	-360	360;
	7917	5567	0.002469	0.015271	0	0	0	0	0	0	1	-360	360;
	7917	11930	0.00212	0.01601	0	0	0	0	0	0	1	-360	360;
	7917	3638	0.00037	0.0028	0	0	0	0	0	0	1	-360	360;
	7917	8175	0.00162	0.01238	0	0	0	0	0	0	1	-360	360;
	7917	8175	0.001909	0.01162	0	0	0	0	0	0	1	-360	360;
	7917	8175	0.001909	0.010729	0	0	0	0	0	0	1	-360	360;
	7917	5493	0.00089	0.00526	0	0	0	0	0	0	1	-360	360;
	7917	8043	0.00087	0.00536	0	0	0	0	0	0	1	-360	360;
	7917	8360	0.001521	0.01176	0	0	0	0	0	0	1	-360	360;
	13334	3547	0.00014	0.000591	0	0	0	0	0	0	1	-360	360;
	1902	1583	0.003531	0.028781	0	0	0	0	0	0	1	-360	360;
	3998	12207	0.004599	0.03657	0	0	0	0	0	0	1	-360	360;
	3998	12329	0.00307	0.019159	0	0	0	0	0	0	1	-360	360;
	3998	1582	0.000581	0.004461	0	0	0	0	0	0	1	-360	360;
	3998	10056	0.00164	0.01161	0	0	0	0	0	0	1	-360	360;
	7232	10048	0.00057	0.003281	0	0	0	0	0	0	1	-360	360;
	11342	10204	0.020419	0.063669	0	0	0	0	0	0	1	-360	360;
	10626	4511	0.01424	0.04751	0	0	0	0	0	0	1	-360	360;
	3042	422	0.00157	0.008849	0	0	0	0	0	0	1	-360	360;
	3042	3602	0.01218	0.07524	0	0	0	0	0	0	1	-360	360;
	2316	1109	0.00037	0.00151	0	0	0	0	0	0	1	-360	360;
	2821	9097	0.000219	0.000531	0	0	0	0	0	0	1	-360	360;
	5810	1980	0.002849	0.0137	0	0	0	0	0	0	1	-360	360;
	1361	13477	0.00116	0.01692	0	0	0	0	0	0	1	-360	360;
	1361	13477	0.00133	0.01462	0	0	0	0	0	0	1	-360	360;
	1361	2936	0.00063	0.00821	0	0	0	0	0	0	1	-360	360;
	1361	2936	0.00064	0.00804	0	0	0	0	0	0	1	-360	360;
	8243	7888	0.00126	0.01272	0	0	0	0	0	0	1	-360	360;
	8243	13408	0.00337	0.0494	0	0	0	0	0	0	1	-360	360;
	8243	10294	0.0014	0.0214	0	0	0	0	0	0	1	-360	360;
	8243	10294	0.00137	0.02182	0	0	0	0	0	0	1	-360	360;
	8243	13130	0.00212	0.02063	0	0	0	0	0	0	1	-360	360;
	12724	11399	0.00186	0.01045	0	0	0	0	0	0	1	-360	360;
	12724	11077	0.00426	0.021771	0	0	0	0	0	0	1	-360	360;
	12724	11077	0.002659	0.020539	0	0	0	0	0	0	1	-360	360;
	12724	6148	0.00606	0.049919	0	0	0	0	0	0	1	-360	360;
	13477	6157	0.00109	0.01192	0	0	0	0	0	0	1	-360	360;
	2184	5760	0.013021	0.049219	0	0	0	0	0	0	1	-360	360;
	2184	6955	0.00236	0.01239	0	0	0	0	0	0	1	-360	360;
	2184	6955	0.0027	0.015169	0	0	0	0	0	0	1	-360	360;
	6661	4880	0.00285	0.03283	0	0	0	0	0	0	1	-360	360;
	6661	4880	0.00529	0.03969	0	0	0	0	0	0	1	-360	360;
	1905	4880	0.00254	0.03721	0	0	0	0	0	0	1	-360	360;
	5367	3889	0.000159	0.00069	0	0	0	0	0	0	1	-360	360;
	5367	3889	0.000229	0.000979	0	0	0	0	0	0	1	-360	360;
	5367	3889	0.00024	0.00082	0	0	0	0	0	0	1	-360	360;
	5367	4682	0.000169	0.00075	0	0	0	0	0	0	1	-360	360;
	5367	7157	0.00368	0.01756	0	0	0	0	0	0	1	-360	360;
	5367	2663	0.002599	0.020659	0	0	0	0	0	0	1	-360	360;
	10021	8087	0.00018	0.00064	0	0	0	0	0	0	1	-360	360;
	10021	10282	0.000159	0.000771	0	0	0	0	0	0	1	-360	360;
	10021	8652	0.00357	0.01981	0	0	0	0	0	0	1	-360	360;
	10021	13214	0.006021	0.03205	0	0	0	0	0	0	1	-360	360;
	1944	9916	0.001781	0.01164	0	0	0	0	0	0	1	-360	360;
	1944	10301	0.002781	0.018831	0	0	0	0	0	0	1	-360	360;
	3602	8987	0.000349	0.00281	0	0	0	0	0	0	1	-360	360;
	130	4954	0.00313	0.02444	0	0	0	0	0	0	1	-360	360;
	130	4954	0.00368	0.02074	0	0	0	0	0	0	1	-360	360;
	3034	4780	9.1e-05	0.001539	0	0	0	0	0	0	1	-360	360;
	3034	3951	0.00251	0.01111	0	0	0	0	0	0	1	-360	360;
	3034	9992	0.003081	0.017961	0	0	0	0	0	0	1	-360	360;
	3034	9849	0.005341	0.030419	0	0	0	0	0	0	1	-360	360;
	3034	11176	9.9e-05	0.00126	0	0	0	0	0	0	1	-360	360;
	13444	1494	0.002169	0.017901	0	0	0	0	0	0	1	-360	360;
	3522	9030	0.00011	0.00038	0	0	0	0	0	0	1	-360	360;
	2016	13025	0.00212	0.00494	0	0	0	0	0	0	1	-360	360;
	2016	11395	0.018901	0.077349	0	0	0	0	0	0	1	-360	360;
	11574	7811	0.00957	0.0395	0	0	0	0	0	0	1	-360	360;
	11574	10204	0.02324	0.063469	0	0	0	0	0	0	1	-360	360;
	11405	4851	0.000531	0.004331	0	0	0	0	0	0	1	-360	360;
	2797	896	0.002581	0.006229	0	0	0	0	0	0	1	-360	360;
	9447	1783	0.00031	0.002211	0	0	0	0	0	0	1	-360	360;
	1152	10204	0.01087	0.06114	0	0	0	0	0	0	1	-360	360;
	1152	8620	0.011581	0.061099	0	0	0	0	0	0	1	-360	360;
	11161	4436	0.002099	0.016271	0	0	0	0	0	0	1	-360	360;
	735	4889	0.00093	0.005461	0	0	0	0	0	0	1	-360	360;
	735	7638	0.009591	0.037039	0	0	0	0	0	0	1	-360	360;
	735	1116	0.002591	0.017211	0	0	0	0	0	0	1	-360	360;
	735	11080	0.00164	0.013651	0	0	0	0	0	0	1	-360	360;
	3946	4974	0.003219	0.014719	0	0	0	0	0	0	1	-360	360;
	3946	3180	0.00401	0.022229	0	0	0	0	0	0	1	-360	360;
	3946	10301	0.00126	0.010211	0	0	0	0	0	0	1	-360	360;
	6052	1607	0.009169	0.03993	0	0	0	0	0	0	1	-360	360;
	6052	5677	0.001349	0.006081	0	0	0	0	0	0	1	-360	360;
	2507	12511	0.00156	0.01931	0	0	0	0	0	0	1	-360	360;
	2507	11017	0.00152	0.01968	0	0	0	0	0	0	1	-360	360;
	2507	11402	0.00094	0.01433	0	0	0	0	0	0	1	-360	360;
	2507	10071	0.00092	0.01461	0	0	0	0	0	0	1	-360	360;
	2507	1356	0.00137	0.02012	0	0	0	0	0	0	1	-360	360;
	2507	1356	0.00134	0.02051	0	0	0	0	0	0	1	-360	360;
	5664	6586	0.00314	0.042151	0	0	0	0	0	0	1	-360	360;
	5664	8754	0.000651	0.00494	0	0	0	0	0	0	1	-360	360;
	5664	10138	0.003961	0.02426	0	0	0	0	0	0	1	-360	360;
	1215	2348	0.00013	0.0015	0	0	0	0	0	0	1	-360	360;
	4360	1158	0.00013	0.00147	0	0	0	0	0	0	1	-360	360;
	7051	2348	0.00011	0.00172	0	0	0	0	0	0	1	-360	360;
	2892	1158	0.00012	0.00166	0	0	0	0	0	0	1	-360	360;
	5683	2348	0.00012	0.00168	0	0	0	0	0	0	1	-360	360;
	5500	1158	0.00012	0.00172	0	0	0	0	0	0	1	-360	360;
	1075	1000	0.00093	0.005669	0	0	0	0	0	0	1	-360	360;
	4178	8175	0.00105	0.00506	0	0	0	0	0	0	1	-360	360;
	8277	9096	0.00113	0.006539	0	0	0	0	0	0	1	-360	360;
	7571	7938	0.00037	0.002919	0	0	0	0	0	0	1	-360	360;
	1027	10399	0.005169	0.03738	0	0	0	0	0	0	1	-360	360;
	5596	1346	0.00125	0.006669	0	0	0	0	0	0	1	-360	360;
	9341	3493	0.001281	0.00661	0	0	0	0	0	0	1	-360	360;
	281	4299	0.00226	0.0138	0	0	0	0	0	0	1	-360	360;
	3616	5150	0.00213	0.02677	0	0	0	0	0	0	1	-360	360;
	3616	4880	0.00154	0.01919	0	0	0	0	0	0	1	-360	360;
	6666	881	0.00687	0.04413	0	0	0	0	0	0	1	-360	360;
	6666	4716	0.0045	0.04725	0	0	0	0	0	0	1	-360	360;
	6666	10759	0.00138	0.00837	0	0	0	0	0	0	1	-360	360;
	4435	9386	0.000419	0.001901	0	0	0	0	0	0	1	-360	360;
	4435	9386	0.00043	0.00186	0	0	0	0	0	0	1	-360	360;
	8929	6929	0.003331	0.015099	0	0	0	0	0	0	1	-360	360;
	4878	9674	0.002099	0.01631	0	0	0	0	0	0	1	-360	360;
	6416	8183	0.001331	0.01082	0	0	0	0	0	0	1	-360	360;
	6416	9787	0.00199	0.00818	0	0	0	0	0	0	1	-360	360;
	6416	9787	0.00169	0.008831	0	0	0	0	0	0	1	-360	360;
	1875	7861	0.001771	0.014281	0	0	0	0	0	0	1	-360	360;
	881	897	0.003099	0.01844	0	0	0	0	0	0	1	-360	360;
	5118	2217	0.003409	0.01819	0	0	0	0	0	0	1	-360	360;
	2743	3181	0.00057	0.00369	0	0	0	0	0	0	1	-360	360;
	11858	8605	0.001081	0.00468	0	0	0	0	0	0	1	-360	360;
	5840	8605	0.00105	0.004771	0	0	0	0	0	0	1	-360	360;
	12862	897	0.001979	0.01549	0	0	0	0	0	0	1	-360	360;
	12862	10759	0.001469	0.015151	0	0	0	0	0	0	1	-360	360;
	8496	7431	0.000711	0.003969	0	0	0	0	0	0	1	-360	360;
	8496	11073	0.00051	0.00274	0	0	0	0	0	0	1	-360	360;
	8496	4621	0.000521	0.00306	0	0	0	0	0	0	1	-360	360;
	8496	5105	0.00188	0.011841	0	0	0	0	0	0	1	-360	360;
	8496	2251	0.002469	0.010539	0	0	0	0	0	0	1	-360	360;
	6101	7730	0.00045	0.00351	0	0	0	0	0	0	1	-360	360;
	7478	13565	0.00037	0.000659	0	0	0	0	0	0	1	-360	360;
	2957	5601	0.000401	0.0007	0	0	0	0	0	0	1	-360	360;
	2957	9097	0.00014	0.000531	0	0	0	0	0	0	1	-360	360;
	2957	2349	0.000289	0.0013	0	0	0	0	0	0	1	-360	360;
	2957	9125	7e-05	0.000289	0	0	0	0	0	0	1	-360	360;
	2957	11146	0.00051	0.002151	0	0	0	0	0	0	1	-360	360;
	2957	3981	0.00037	0.00157	0	0	0	0	0	0	1	-360	360;
	7414	661	0.00022	0.00277	0	0	0	0	0	0	1	-360	360;
	7414	661	0.00025	0.00271	0	0	0	0	0	0	1	-360	360;
	11183	4370	0.000271	0.002081	0	0	0	0	0	0	1	-360	360;
	11183	4370	0.00025	0.00231	0	0	0	0	0	0	1	-360	360;
	4621	3349	0.00095	0.00331	0	0	0	0	0	0	1	-360	360;
	4621	12893	0.00093	0.00336	0	0	0	0	0	0	1	-360	360;
	3583	12913	0.000341	0.002211	0	0	0	0	0	0	1	-360	360;
	3789	5121	0.00032	0.00232	0	0	0	0	0	0	1	-360	360;
	6064	12568	3.9e-05	0.000271	0	0	0	0	0	0	1	-360	360;
	885	227	0.00075	0.0101	0	0	0	0	0	0	1	-360	360;
	885	8070	0.00118	0.01455	0	0	0	0	0	0	1	-360	360;
	5567	13336	0.0017	0.0098	0	0	0	0	0	0	1	-360	360;
	5567	13336	0.001521	0.01111	0	0	0	0	0	0	1	-360	360;
	5567	4620	0.000169	0.001401	0	0	0	0	0	0	1	-360	360;
	5567	1000	0.002591	0.01725	0	0	0	0	0	0	1	-360	360;
	5567	8426	0.000169	0.001159	0	0	0	0	0	0	1	-360	360;
	5567	6929	0.00307	0.01339	0	0	0	0	0	0	1	-360	360;
	3785	2392	0.002591	0.013521	0	0	0	0	0	0	1	-360	360;
	13405	13372	0.000581	0.005031	0	0	0	0	0	0	1	-360	360;
	10182	8105	0.001419	0.00595	0	0	0	0	0	0	1	-360	360;
	970	2124	0.00205	0.02664	0	0	0	0	0	0	1	-360	360;
	970	3917	0.00295	0.03721	0	0	0	0	0	0	1	-360	360;
	970	6675	0.00335	0.03648	0	0	0	0	0	0	1	-360	360;
	970	2294	0.00022	0.00251	0	0	0	0	0	0	1	-360	360;
	970	2961	0.00059	0.00651	0	0	0	0	0	0	1	-360	360;
	5234	3083	0.008539	0.041979	0	0	0	0	0	0	1	-360	360;
	5234	4077	0.001349	0.0052	0	0	0	0	0	0	1	-360	360;
	5312	11722	0.00245	0.013909	0	0	0	0	0	0	1	-360	360;
	5312	3559	0.001469	0.008419	0	0	0	0	0	0	1	-360	360;
	7888	13130	0.00084	0.00836	0	0	0	0	0	0	1	-360	360;
	10411	5004	0.002591	0.017789	0	0	0	0	0	0	1	-360	360;
	10411	1607	0.009651	0.05968	0	0	0	0	0	0	1	-360	360;
	10877	11003	0.00143	0.00799	0	0	0	0	0	0	1	-360	360;
	10877	9849	0.001969	0.014909	0	0	0	0	0	0	1	-360	360;
	1464	4954	0.01039	0.06406	0	0	0	0	0	0	1	-360	360;
	6980	4851	0.000719	0.005331	0	0	0	0	0	0	1	-360	360;
	12476	6599	0.00052	0.00565	0	0	0	0	0	0	1	-360	360;
	12299	10746	0.002409	0.019479	0	0	0	0	0	0	1	-360	360;
	12299	10746	0.00256	0.019151	0	0	0	0	0	0	1	-360	360;
	12299	4108	0.00656	0.02288	0	0	0	0	0	0	1	-360	360;
	5922	10761	0.0027	0.014901	0	0	0	0	0	0	1	-360	360;
	5922	8338	0.003591	0.01362	0	0	0	0	0	0	1	-360	360;
	10737	3559	0.002211	0.01188	0	0	0	0	0	0	1	-360	360;
	10737	11276	0.00068	0.00357	0	0	0	0	0	0	1	-360	360;
	10944	2259	0.000211	0.000969	0	0	0	0	0	0	1	-360	360;
	13234	4953	9.9e-05	0.000539	0	0	0	0	0	0	1	-360	360;
	2581	9334	0.00645	0.0263	0	0	0	0	0	0	1	-360	360;
	6778	8788	0.00063	0.001901	0	0	0	0	0	0	1	-360	360;
	2454	5222	0.000281	0.000669	0	0	0	0	0	0	1	-360	360;
	11278	5677	0.000219	0.00064	0	0	0	0	0	0	1	-360	360;
	13573	4916	0.00014	0.00105	0	0	0	0	0	0	1	-360	360;
	12343	12945	0.003219	0.020729	0	0	0	0	0	0	1	-360	360;
	12343	12495	0.00051	0.003151	0	0	0	0	0	0	1	-360	360;
	12343	12495	0.000409	0.00331	0	0	0	0	0	0	1	-360	360;
	8203	9710	0.002289	0.017479	0	0	0	0	0	0	1	-360	360;
	8203	1694	0.002831	0.01895	0	0	0	0	0	0	1	-360	360;
	8203	9799	0.006289	0.036521	0	0	0	0	0	0	1	-360	360;
	8203	13294	0.005651	0.032169	0	0	0	0	0	0	1	-360	360;
	8203	10645	0.00414	0.02262	0	0	0	0	0	0	1	-360	360;
	5112	9274	0.01095	0.05814	0	0	0	0	0	0	1	-360	360;
	5112	1445	0.009	0.0368	0	0	0	0	0	0	1	-360	360;
	5112	5545	0.000229	0.00181	0	0	0	0	0	0	1	-360	360;
	4684	9611	0.00119	0.00726	0	0	0	0	0	0	1	-360	360;
	12744	1086	0.011771	0.051669	0	0	0	0	0	0	1	-360	360;
	227	3090	0.00209	0.02279	0	0	0	0	0	0	1	-360	360;
	11066	5760	0.002031	0.01443	0	0	0	0	0	0	1	-360	360;
	11066	10062	0.00145	0.008479	0	0	0	0	0	0	1	-360	360;
	3627	9059	0.00124	0.01433	0	0	0	0	0	0	1	-360	360;
	3627	9059	0.00127	0.01402	0	0	0	0	0	0	1	-360	360;
	7129	5386	0.013039	0.052841	0	0	0	0	0	0	1	-360	360;
	3383	9059	0.00235	0.02739	0	0	0	0	0	0	1	-360	360;
	3383	9059	0.0024	0.02681	0	0	0	0	0	0	1	-360	360;
	3383	6081	0.00135	0.02056	0	0	0	0	0	0	1	-360	360;
	3383	1859	0.00087	0.01124	0	0	0	0	0	0	1	-360	360;
	5616	247	0.0012	0.00818	0	0	0	0	0	0	1	-360	360;
	247	9412	0.006961	0.05561	0	0	0	0	0	0	1	-360	360;
	6400	11592	0.00105	0.00763	0	0	0	0	0	0	1	-360	360;
	6400	820	0.000219	0.00182	0	0	0	0	0	0	1	-360	360;
	6400	13097	3.9e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	6400	13097	4.5e-05	0.000195	0	0	0	0	0	0	1	-360	360;
	6400	6228	0.0035	0.022651	0	0	0	0	0	0	1	-360	360;
	5678	4954	0.004281	0.02319	0	0	0	0	0	0	1	-360	360;
	5678	1252	0.003729	0.014021	0	0	0	0	0	0	1	-360	360;
	10292	3083	0.00212	0.01582	0	0	0	0	0	0	1	-360	360;
	12835	13130	0.00056	0.00784	0	0	0	0	0	0	1	-360	360;
	4954	11395	0.004401	0.01813	0	0	0	0	0	0	1	-360	360;
	4954	11395	0.005021	0.02144	0	0	0	0	0	0	1	-360	360;
	9839	4212	0.00018	0.000219	0	0	0	0	0	0	1	-360	360;
	12014	4212	0.000151	0.000211	0	0	0	0	0	0	1	-360	360;
	6508	4212	0.000151	0.000211	0	0	0	0	0	0	1	-360	360;
	2937	8710	0.00282	0.02591	0	0	0	0	0	0	1	-360	360;
	9847	3455	0.00131	0.008539	0	0	0	0	0	0	1	-360	360;
	7687	10989	0.002159	0.015969	0	0	0	0	0	0	1	-360	360;
	10299	13324	0.00019	0.000281	0	0	0	0	0	0	1	-360	360;
	8103	3181	0.0002	0.000281	0	0	0	0	0	0	1	-360	360;
	5004	11774	0.00139	0.00774	0	0	0	0	0	0	1	-360	360;
	934	9289	0.01101	0.06739	0	0	0	0	0	0	1	-360	360;
	934	3940	0.00063	0.00463	0	0	0	0	0	0	1	-360	360;
	934	12691	0.002341	0.009599	0	0	0	0	0	0	1	-360	360;
	11399	6148	0.00375	0.021659	0	0	0	0	0	0	1	-360	360;
	2124	8404	0.00056	0.00749	0	0	0	0	0	0	1	-360	360;
	158	9674	0.00118	0.00926	0	0	0	0	0	0	1	-360	360;
	158	8788	0.002039	0.013081	0	0	0	0	0	0	1	-360	360;
	2492	10989	0.00155	0.010961	0	0	0	0	0	0	1	-360	360;
	4780	11176	3.9e-05	0.00036	0	0	0	0	0	0	1	-360	360;
	261	282	0.00024	0.000409	0	0	0	0	0	0	1	-360	360;
	12568	8498	0.000341	0.001659	0	0	0	0	0	0	1	-360	360;
	12568	7793	0.005461	0.024919	0	0	0	0	0	0	1	-360	360;
	11823	10309	0.00158	0.02413	0	0	0	0	0	0	1	-360	360;
	11823	13067	0.00155	0.02458	0	0	0	0	0	0	1	-360	360;
	11823	10059	0.00021	0.00256	0	0	0	0	0	0	1	-360	360;
	11823	10059	0.00024	0.00243	0	0	0	0	0	0	1	-360	360;
	11823	2849	0.00144	0.02266	0	0	0	0	0	0	1	-360	360;
	11823	2849	0.0014	0.02098	0	0	0	0	0	0	1	-360	360;
	11823	12512	0.00086	0.01074	0	0	0	0	0	0	1	-360	360;
	12945	12495	0.003341	0.02611	0	0	0	0	0	0	1	-360	360;
	320	9913	0.00387	0.02063	0	0	0	0	0	0	1	-360	360;
	320	9913	0.00405	0.018271	0	0	0	0	0	0	1	-360	360;
	320	1686	0.00101	0.00462	0	0	0	0	0	0	1	-360	360;
	320	3702	0.00088	0.00389	0	0	0	0	0	0	1	-360	360;
	320	9799	0.005081	0.039	0	0	0	0	0	0	1	-360	360;
	11722	11276	0.00257	0.0208	0	0	0	0	0	0	1	-360	360;
	11722	11276	0.00251	0.01926	0	0	0	0	0	0	1	-360	360;
	11722	1493	0.002531	0.01314	0	0	0	0	0	0	1	-360	360;
	10407	11515	0.00132	0.00869	0	0	0	0	0	0	1	-360	360;
	12397	10852	0.00082	0.004021	0	0	0	0	0	0	1	-360	360;
	12397	7112	0.00143	0.007479	0	0	0	0	0	0	1	-360	360;
	3995	2508	0.000659	0.005771	0	0	0	0	0	0	1	-360	360;
	3995	2508	0.000591	0.006539	0	0	0	0	0	0	1	-360	360;
	1850	9183	0.00589	0.025599	0	0	0	0	0	0	1	-360	360;
	1850	10204	0.00614	0.02343	0	0	0	0	0	0	1	-360	360;
	3917	2564	0.00155	0.01994	0	0	0	0	0	0	1	-360	360;
	3917	4880	0.00205	0.03218	0	0	0	0	0	0	1	-360	360;
	6675	4880	0.00233	0.03154	0	0	0	0	0	0	1	-360	360;
	6924	8338	0.002341	0.01982	0	0	0	0	0	0	1	-360	360;
	1563	5545	8.1e-05	0.000521	0	0	0	0	0	0	1	-360	360;
	1981	8095	9.9e-05	0.00069	0	0	0	0	0	0	1	-360	360;
	6274	6715	0.00276	0.016419	0	0	0	0	0	0	1	-360	360;
	10303	7195	0.00036	0.001229	0	0	0	0	0	0	1	-360	360;
	10303	5499	0.000229	0.000711	0	0	0	0	0	0	1	-360	360;
	3547	13332	0.00013	0.00099	0	0	0	0	0	0	1	-360	360;
	13302	8040	0.001159	0.00689	0	0	0	0	0	0	1	-360	360;
	10058	10485	0.008531	0.03045	0	0	0	0	0	0	1	-360	360;
	7866	11774	0.000781	0.004531	0	0	0	0	0	0	1	-360	360;
	7866	11774	0.00075	0.00474	0	0	0	0	0	0	1	-360	360;
	8572	5921	0.002159	0.008349	0	0	0	0	0	0	1	-360	360;
	4416	12272	0.00162	0.01651	0	0	0	0	0	0	1	-360	360;
	1314	9296	0.01236	0.044211	0	0	0	0	0	0	1	-360	360;
	1314	9296	0.00825	0.03989	0	0	0	0	0	0	1	-360	360;
	1314	3424	0.02187	0.066961	0	0	0	0	0	0	1	-360	360;
	1314	5844	0.01212	0.06661	0	0	0	0	0	0	1	-360	360;
	5763	2348	0.00036	0.00534	0	0	0	0	0	0	1	-360	360;
	9938	1653	0.000229	0.00063	0	0	0	0	0	0	1	-360	360;
	9938	11114	0.000219	0.000581	0	0	0	0	0	0	1	-360	360;
	5914	5872	0.000281	0.00063	0	0	0	0	0	0	1	-360	360;
	2274	2007	0.005021	0.035659	0	0	0	0	0	0	1	-360	360;
	4504	5395	0.00174	0.007901	0	0	0	0	0	0	1	-360	360;
	7638	3180	0.003919	0.020581	0	0	0	0	0	0	1	-360	360;
	7638	11931	0.003341	0.016349	0	0	0	0	0	0	1	-360	360;
	7638	2104	0.003271	0.016669	0	0	0	0	0	0	1	-360	360;
	4002	12559	0.000151	0.000659	0	0	0	0	0	0	1	-360	360;
	9372	1607	0.00736	0.035659	0	0	0	0	0	0	1	-360	360;
	9372	5722	0.00944	0.042169	0	0	0	0	0	0	1	-360	360;
	6512	6937	0.00068	0.00652	0	0	0	0	0	0	1	-360	360;
	6512	3448	0.00188	0.02281	0	0	0	0	0	0	1	-360	360;
	4974	7761	0.00168	0.00956	0	0	0	0	0	0	1	-360	360;
	4974	7761	0.001669	0.00981	0	0	0	0	0	0	1	-360	360;
	4974	1116	0.002229	0.01599	0	0	0	0	0	0	1	-360	360;
	4974	1116	0.00218	0.01631	0	0	0	0	0	0	1	-360	360;
	1554	2936	0.00238	0.02124	0	0	0	0	0	0	1	-360	360;
	1554	2936	0.00194	0.02023	0	0	0	0	0	0	1	-360	360;
	10795	4581	0.013831	0.05955	0	0	0	0	0	0	1	-360	360;
	10795	627	0.00876	0.04875	0	0	0	0	0	0	1	-360	360;
	4105	2641	0.0002	0.00099	0	0	0	0	0	0	1	-360	360;
	1930	2729	0.00105	0.00826	0	0	0	0	0	0	1	-360	360;
	3083	1086	0.0172	0.06637	0	0	0	0	0	0	1	-360	360;
	5280	4090	0.00305	0.015781	0	0	0	0	0	0	1	-360	360;
	9592	5299	0.00208	0.02105	0	0	0	0	0	0	1	-360	360;
	12495	2804	0.00614	0.043729	0	0	0	0	0	0	1	-360	360;
	12495	8629	0.000229	0.001849	0	0	0	0	0	0	1	-360	360;
	8336	8821	0.00386	0.025789	0	0	0	0	0	0	1	-360	360;
	8336	3350	0.002031	0.012	0	0	0	0	0	0	1	-360	360;
	3770	10989	0.000591	0.004	0	0	0	0	0	0	1	-360	360;
	1633	3444	9.1e-05	0.00051	0	0	0	0	0	0	1	-360	360;
	11074	8460	0.00732	0.039969	0	0	0	0	0	0	1	-360	360;
	11074	4700	0.004581	0.02031	0	0	0	0	0	0	1	-360	360;
	11222	8460	0.00132	0.00537	0	0	0	0	0	0	1	-360	360;
	2508	7048	0.007521	0.0387	0	0	0	0	0	0	1	-360	360;
	10008	7396	0.005159	0.031281	0	0	0	0	0	0	1	-360	360;
	12234	13028	9.9e-05	0.001159	0	0	0	0	0	0	1	-360	360;
	1268	13028	9.9e-05	0.00113	0	0	0	0	0	0	1	-360	360;
	6605	6394	0.00019	0.000909	0	0	0	0	0	0	1	-360	360;
	10463	11424	0.00107	0.00413	0	0	0	0	0	0	1	-360	360;
	10059	2953	9e-05	0.00058	0	0	0	0	0	0	1	-360	360;
	10059	5913	9e-05	0.00059	0	0	0	0	0	0	1	-360	360;
	10059	13403	8e-05	0.00053	0	0	0	0	0	0	1	-360	360;
	10059	965	8e-05	0.00052	0	0	0	0	0	0	1	-360	360;
	9669	13125	0.000271	0.00151	0	0	0	0	0	0	1	-360	360;
	13215	9059	0.00053	0.00585	0	0	0	0	0	0	1	-360	360;
	13215	9059	0.00047	0.00663	0	0	0	0	0	0	1	-360	360;
	13215	9059	0.00046	0.00676	0	0	0	0	0	0	1	-360	360;
	13215	9059	0.00045	0.00689	0	0	0	0	0	0	1	-360	360;
	13215	12531	0.00028	0.00338	0	0	0	0	0	0	1	-360	360;
	13215	12531	0.0003	0.00319	0	0	0	0	0	0	1	-360	360;
	2349	12594	0.00025	0.00101	0	0	0	0	0	0	1	-360	360;
	282	2813	0.000281	0.00119	0	0	0	0	0	0	1	-360	360;
	4165	8105	0.00018	0.00063	0	0	0	0	0	0	1	-360	360;
	2958	6475	0.00382	0.02868	0	0	0	0	0	0	1	-360	360;
	2958	495	0.00361	0.0235	0	0	0	0	0	0	1	-360	360;
	2958	6490	0.00538	0.03569	0	0	0	0	0	0	1	-360	360;
	4927	203	6e-05	0.000409	0	0	0	0	0	0	1	-360	360;
	1465	4719	9e-05	0.00102	0	0	0	0	0	0	1	-360	360;
	1465	10731	0.0014	0.01649	0	0	0	0	0	0	1	-360	360;
	1465	4194	0.00034	0.00427	0	0	0	0	0	0	1	-360	360;
	1229	4406	0.0001	0.00109	0	0	0	0	0	0	1	-360	360;
	1229	9409	0.00137	0.01682	0	0	0	0	0	0	1	-360	360;
	1229	5846	0.00131	0.01262	0	0	0	0	0	0	1	-360	360;
	10468	13541	0.00262	0.019669	0	0	0	0	0	0	1	-360	360;
	10468	3466	0.00082	0.00526	0	0	0	0	0	0	1	-360	360;
	10468	3466	0.000841	0.005151	0	0	0	0	0	0	1	-360	360;
	10468	3466	0.000669	0.005081	0	0	0	0	0	0	1	-360	360;
	10468	13083	0.001919	0.012669	0	0	0	0	0	0	1	-360	360;
	55	301	0.002659	0.02	0	0	0	0	0	0	1	-360	360;
	55	5194	0.003031	0.02343	0	0	0	0	0	0	1	-360	360;
	9096	6906	0.01176	0.059159	0	0	0	0	0	0	1	-360	360;
	9096	12329	0.002469	0.017531	0	0	0	0	0	0	1	-360	360;
	9096	6738	0.000669	0.005729	0	0	0	0	0	0	1	-360	360;
	12207	12492	0.000781	0.00876	0	0	0	0	0	0	1	-360	360;
	12207	7938	0.00062	0.00482	0	0	0	0	0	0	1	-360	360;
	12207	11176	0.001349	0.012599	0	0	0	0	0	0	1	-360	360;
	9992	12492	0.000979	0.0082	0	0	0	0	0	0	1	-360	360;
	9992	1242	0.008091	0.028831	0	0	0	0	0	0	1	-360	360;
	9992	6738	0.000711	0.005969	0	0	0	0	0	0	1	-360	360;
	5836	203	0.00145	0.00857	0	0	0	0	0	0	1	-360	360;
	1571	1621	0.00869	0.03768	0	0	0	0	0	0	1	-360	360;
	1571	4436	0.001719	0.00837	0	0	0	0	0	0	1	-360	360;
	12559	7811	0.01582	0.067349	0	0	0	0	0	0	1	-360	360;
	12559	7811	0.01457	0.06469	0	0	0	0	0	0	1	-360	360;
	12559	6646	0.021711	0.090539	0	0	0	0	0	0	1	-360	360;
	12559	4907	0.00019	0.000729	0	0	0	0	0	0	1	-360	360;
	1317	7838	0.001271	0.00638	0	0	0	0	0	0	1	-360	360;
	1291	9611	3.9e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	9740	6497	0.000281	0.00236	0	0	0	0	0	0	1	-360	360;
	11015	3090	0.00019	0.00329	0	0	0	0	0	0	1	-360	360;
	10996	8070	0.00035	0.00324	0	0	0	0	0	0	1	-360	360;
	9611	13167	0.00049	0.00349	0	0	0	0	0	0	1	-360	360;
	9611	13167	0.000469	0.00331	0	0	0	0	0	0	1	-360	360;
	9611	2375	0.00313	0.01324	0	0	0	0	0	0	1	-360	360;
	9611	13481	0.001031	0.006719	0	0	0	0	0	0	1	-360	360;
	9611	13092	0.00038	0.00401	0	0	0	0	0	0	1	-360	360;
	738	5265	0.00037	0.002849	0	0	0	0	0	0	1	-360	360;
	12492	1494	0.00032	0.00256	0	0	0	0	0	0	1	-360	360;
	12492	206	0.00037	0.00118	0	0	0	0	0	0	1	-360	360;
	12492	6258	0.00068	0.001151	0	0	0	0	0	0	1	-360	360;
	1154	5010	0.00168	0.009841	0	0	0	0	0	0	1	-360	360;
	12662	13111	0.000599	0.002401	0	0	0	0	0	0	1	-360	360;
	3301	6	0.002169	0.00911	0	0	0	0	0	0	1	-360	360;
	3301	7077	0.002849	0.01282	0	0	0	0	0	0	1	-360	360;
	9125	7838	0.001229	0.01207	0	0	0	0	0	0	1	-360	360;
	9125	10989	0.001289	0.01089	0	0	0	0	0	0	1	-360	360;
	2849	2350	0.00093	0.01036	0	0	0	0	0	0	1	-360	360;
	2849	3448	0.00083	0.01174	0	0	0	0	0	0	1	-360	360;
	9710	11515	0.00118	0.009469	0	0	0	0	0	0	1	-360	360;
	9710	11515	0.001151	0.009651	0	0	0	0	0	0	1	-360	360;
	7948	11927	0.00101	0.01327	0	0	0	0	0	0	1	-360	360;
	9183	6148	0.00364	0.02486	0	0	0	0	0	0	1	-360	360;
	12385	6955	0.00056	0.00425	0	0	0	0	0	0	1	-360	360;
	12385	6715	0.00369	0.02681	0	0	0	0	0	0	1	-360	360;
	501	11609	0.000531	0.00276	0	0	0	0	0	0	1	-360	360;
	3541	11609	0.000521	0.00281	0	0	0	0	0	0	1	-360	360;
	7665	11609	0.000419	0.00232	0	0	0	0	0	0	1	-360	360;
	3105	11609	0.000409	0.00237	0	0	0	0	0	0	1	-360	360;
	12472	13408	0.00014	0.00193	0	0	0	0	0	0	1	-360	360;
	3832	8718	0.003211	0.01481	0	0	0	0	0	0	1	-360	360;
	1694	8309	0.00381	0.01175	0	0	0	0	0	0	1	-360	360;
	1694	9544	0.003781	0.02676	0	0	0	0	0	0	1	-360	360;
	3377	5230	0.000919	0.005461	0	0	0	0	0	0	1	-360	360;
	3377	13071	0.000271	0.001211	0	0	0	0	0	0	1	-360	360;
	3377	12739	0.002229	0.017401	0	0	0	0	0	0	1	-360	360;
	3377	12739	0.00243	0.017659	0	0	0	0	0	0	1	-360	360;
	3377	12739	0.002531	0.01705	0	0	0	0	0	0	1	-360	360;
	13045	7450	0.012531	0.05451	0	0	0	0	0	0	1	-360	360;
	8404	6937	0.00087	0.01212	0	0	0	0	0	0	1	-360	360;
	8404	1269	0.0006	0.00802	0	0	0	0	0	0	1	-360	360;
	8404	12272	0.00072	0.00667	0	0	0	0	0	0	1	-360	360;
	8404	2961	0.00234	0.02626	0	0	0	0	0	0	1	-360	360;
	9892	13083	0.00038	0.002969	0	0	0	0	0	0	1	-360	360;
	11582	5753	0.00024	0.00045	0	0	0	0	0	0	1	-360	360;
	9835	5753	0.000229	0.000409	0	0	0	0	0	0	1	-360	360;
	6306	5753	0.000281	0.000401	0	0	0	0	0	0	1	-360	360;
	7807	13541	0.001211	0.005271	0	0	0	0	0	0	1	-360	360;
	7807	521	0.000281	0.00107	0	0	0	0	0	0	1	-360	360;
	7807	11933	0.00025	0.001091	0	0	0	0	0	0	1	-360	360;
	5662	2813	0.00019	0.00081	0	0	0	0	0	0	1	-360	360;
	5662	2813	0.00045	0.00136	0	0	0	0	0	0	1	-360	360;
	5097	812	0.003461	0.023289	0	0	0	0	0	0	1	-360	360;
	7280	12094	0.0013	0.01609	0	0	0	0	0	0	1	-360	360;
	7280	12094	0.00128	0.0164	0	0	0	0	0	0	1	-360	360;
	7280	3891	0.00089	0.01082	0	0	0	0	0	0	1	-360	360;
	7280	3891	0.00087	0.01102	0	0	0	0	0	0	1	-360	360;
	5559	7758	0.003349	0.02514	0	0	0	0	0	0	1	-360	360;
	720	3802	0.0003	0.00061	0	0	0	0	0	0	1	-360	360;
	211	10746	0.000289	0.001211	0	0	0	0	0	0	1	-360	360;
	10630	9358	0.01301	0.044091	0	0	0	0	0	0	1	-360	360;
	10630	2483	0.02162	0.065969	0	0	0	0	0	0	1	-360	360;
	10970	467	0.006461	0.05275	0	0	0	0	0	0	1	-360	360;
	10159	11726	0.0002	0.000271	0	0	0	0	0	0	1	-360	360;
	4849	11726	0.00018	0.00026	0	0	0	0	0	0	1	-360	360;
	8197	2483	0.00032	0.00188	0	0	0	0	0	0	1	-360	360;
	13541	13083	0.00105	0.006591	0	0	0	0	0	0	1	-360	360;
	26	9012	0.00013	0.00055	0	0	0	0	0	0	1	-360	360;
	11446	7811	0.006151	0.04013	0	0	0	0	0	0	1	-360	360;
	11446	10204	0.007159	0.040591	0	0	0	0	0	0	1	-360	360;
	3268	4528	0.00874	0.048461	0	0	0	0	0	0	1	-360	360;
	3268	6943	0.004159	0.01606	0	0	0	0	0	0	1	-360	360;
	3268	5395	0.002919	0.01912	0	0	0	0	0	0	1	-360	360;
	467	6303	0.00094	0.00394	0	0	0	0	0	0	1	-360	360;
	467	8412	0.000901	0.00395	0	0	0	0	0	0	1	-360	360;
	467	3540	0.003159	0.017789	0	0	0	0	0	0	1	-360	360;
	467	3540	0.002419	0.018331	0	0	0	0	0	0	1	-360	360;
	467	7664	0.00375	0.018281	0	0	0	0	0	0	1	-360	360;
	4632	3802	0.00025	0.001031	0	0	0	0	0	0	1	-360	360;
	6859	7112	0.00264	0.01238	0	0	0	0	0	0	1	-360	360;
	6859	7112	0.001289	0.00795	0	0	0	0	0	0	1	-360	360;
	6859	13092	0.00087	0.003081	0	0	0	0	0	0	1	-360	360;
	13492	7606	0.000401	0.00224	0	0	0	0	0	0	1	-360	360;
	82	326	0.0003	0.00385	0	0	0	0	0	0	1	-360	360;
	4479	1206	0.00126	0.008219	0	0	0	0	0	0	1	-360	360;
	213	7450	0.01905	0.055729	0	0	0	0	0	0	1	-360	360;
	213	5222	0.00201	0.00444	0	0	0	0	0	0	1	-360	360;
	10564	4220	0.00101	0.00324	0	0	0	0	0	0	1	-360	360;
	2075	7013	0.00105	0.00677	0	0	0	0	0	0	1	-360	360;
	2075	5623	0.00039	0.00502	0	0	0	0	0	0	1	-360	360;
	2075	8232	0.0004	0.00492	0	0	0	0	0	0	1	-360	360;
	2075	8070	0.00051	0.00617	0	0	0	0	0	0	1	-360	360;
	11026	12417	0.0003	0.00314	0	0	0	0	0	0	1	-360	360;
	11026	5623	0.00051	0.00668	0	0	0	0	0	0	1	-360	360;
	11026	8232	0.0005	0.00678	0	0	0	0	0	0	1	-360	360;
	1783	3472	0.002909	0.017	0	0	0	0	0	0	1	-360	360;
	1783	1657	0.00245	0.011281	0	0	0	0	0	0	1	-360	360;
	1783	7203	0.0027	0.01532	0	0	0	0	0	0	1	-360	360;
	1783	7203	0.00105	0.011151	0	0	0	0	0	0	1	-360	360;
	1783	6632	0.00036	0.002669	0	0	0	0	0	0	1	-360	360;
	1783	816	0.0018	0.011169	0	0	0	0	0	0	1	-360	360;
	1783	4350	0.000711	0.007771	0	0	0	0	0	0	1	-360	360;
	1783	11640	0.00036	0.00257	0	0	0	0	0	0	1	-360	360;
	2217	9004	0.00201	0.01195	0	0	0	0	0	0	1	-360	360;
	2217	7132	0.001091	0.00731	0	0	0	0	0	0	1	-360	360;
	2217	7942	0.00105	0.007591	0	0	0	0	0	0	1	-360	360;
	2217	13324	0.00299	0.02256	0	0	0	0	0	0	1	-360	360;
	4436	9530	0.00036	0.00288	0	0	0	0	0	0	1	-360	360;
	2224	6604	0.00589	0.045599	0	0	0	0	0	0	1	-360	360;
	2224	5251	0.00157	0.0127	0	0	0	0	0	0	1	-360	360;
	2224	5251	0.00168	0.011969	0	0	0	0	0	0	1	-360	360;
	2224	11608	0.00289	0.021031	0	0	0	0	0	0	1	-360	360;
	1744	4776	0.00395	0.025211	0	0	0	0	0	0	1	-360	360;
	1744	13448	0.002771	0.021659	0	0	0	0	0	0	1	-360	360;
	1744	8237	0.002979	0.016901	0	0	0	0	0	0	1	-360	360;
	12498	8605	0.007419	0.034021	0	0	0	0	0	0	1	-360	360;
	4266	6599	0.00056	0.00813	0	0	0	0	0	0	1	-360	360;
	6	2590	0.00361	0.015211	0	0	0	0	0	0	1	-360	360;
	11080	10301	0.000409	0.003331	0	0	0	0	0	0	1	-360	360;
	5968	9694	0.004091	0.019599	0	0	0	0	0	0	1	-360	360;
	5968	9274	0.00294	0.012841	0	0	0	0	0	0	1	-360	360;
	6497	8183	0.020771	0.080479	0	0	0	0	0	0	1	-360	360;
	6497	2804	0.00662	0.0513	0	0	0	0	0	0	1	-360	360;
	6497	2528	0.00607	0.04581	0	0	0	0	0	0	1	-360	360;
	4370	4776	0.000669	0.0042	0	0	0	0	0	0	1	-360	360;
	4370	4564	0.001271	0.007419	0	0	0	0	0	0	1	-360	360;
	4370	8237	0.001599	0.012271	0	0	0	0	0	0	1	-360	360;
	1252	6839	0.009599	0.05024	0	0	0	0	0	0	1	-360	360;
	1252	761	0.00149	0.009219	0	0	0	0	0	0	1	-360	360;
	1252	9955	0.01488	0.06724	0	0	0	0	0	0	1	-360	360;
	1252	8620	0.003599	0.024031	0	0	0	0	0	0	1	-360	360;
	1252	8620	0.00519	0.024	0	0	0	0	0	0	1	-360	360;
	5230	12270	0.003461	0.013349	0	0	0	0	0	0	1	-360	360;
	5011	6228	0.00105	0.01236	0	0	0	0	0	0	1	-360	360;
	5011	6228	0.00199	0.012331	0	0	0	0	0	0	1	-360	360;
	5011	5395	0.00711	0.044711	0	0	0	0	0	0	1	-360	360;
	9065	5428	0.006729	0.02543	0	0	0	0	0	0	1	-360	360;
	13382	5428	0.005771	0.02999	0	0	0	0	0	0	1	-360	360;
	10485	140	0.0048	0.021771	0	0	0	0	0	0	1	-360	360;
	3961	459	0.00286	0.02212	0	0	0	0	0	0	1	-360	360;
	1039	12376	0.00351	0.02175	0	0	0	0	0	0	1	-360	360;
	1039	10204	0.008401	0.0448	0	0	0	0	0	0	1	-360	360;
	9155	13028	0.0018	0.01187	0	0	0	0	0	0	1	-360	360;
	9155	13028	0.001831	0.01163	0	0	0	0	0	0	1	-360	360;
	12683	4315	0.00075	0.00538	0	0	0	0	0	0	1	-360	360;
	9633	1173	2e-05	0.00033	0	0	0	0	0	0	1	-360	360;
	9633	12161	1e-05	0.00034	0	0	0	0	0	0	1	-360	360;
	9059	1345	0.00023	0.00258	0	0	0	0	0	0	1	-360	360;
	9059	2595	0.00023	0.00252	0	0	0	0	0	0	1	-360	360;
	4700	2141	0.002031	0.008531	0	0	0	0	0	0	1	-360	360;
	4700	10585	0.005349	0.020789	0	0	0	0	0	0	1	-360	360;
	4212	1509	0.0002	0.00082	0	0	0	0	0	0	1	-360	360;
	5760	10062	0.002091	0.011219	0	0	0	0	0	0	1	-360	360;
	8821	244	0.00262	0.016979	0	0	0	0	0	0	1	-360	360;
	8821	244	0.002669	0.01663	0	0	0	0	0	0	1	-360	360;
	13071	13203	0.00025	0.00118	0	0	0	0	0	0	1	-360	360;
	2218	1702	0.00056	0.00715	0	0	0	0	0	0	1	-360	360;
	5753	5499	0.0007	0.00455	0	0	0	0	0	0	1	-360	360;
	5753	10193	0.000331	0.001711	0	0	0	0	0	0	1	-360	360;
	5753	10193	0.00011	0.000781	0	0	0	0	0	0	1	-360	360;
	5753	13142	0.00111	0.00664	0	0	0	0	0	0	1	-360	360;
	11004	12094	0.00066	0.00838	0	0	0	0	0	0	1	-360	360;
	11004	12094	0.00064	0.00854	0	0	0	0	0	0	1	-360	360;
	3466	11522	0.003091	0.017091	0	0	0	0	0	0	1	-360	360;
	3466	2007	0.00643	0.03687	0	0	0	0	0	0	1	-360	360;
	13144	11522	0.00275	0.01937	0	0	0	0	0	0	1	-360	360;
	12376	8620	0.00357	0.021289	0	0	0	0	0	0	1	-360	360;
	2735	10204	0.000289	0.002789	0	0	0	0	0	0	1	-360	360;
	1607	9709	0.002531	0.01345	0	0	0	0	0	0	1	-360	360;
	6692	955	0.00162	0.012159	0	0	0	0	0	0	1	-360	360;
	6692	955	0.00262	0.01168	0	0	0	0	0	0	1	-360	360;
	13078	1657	0.00099	0.006081	0	0	0	0	0	0	1	-360	360;
	13078	9310	0.000909	0.005979	0	0	0	0	0	0	1	-360	360;
	4953	601	0.00011	0.0008	0	0	0	0	0	0	1	-360	360;
	4953	10594	0.00011	0.00068	0	0	0	0	0	0	1	-360	360;
	4953	8724	0.00207	0.008341	0	0	0	0	0	0	1	-360	360;
	5757	3662	0.00019	0.001031	0	0	0	0	0	0	1	-360	360;
	5844	10787	0.014039	0.0667	0	0	0	0	0	0	1	-360	360;
	4519	13214	0.004961	0.03043	0	0	0	0	0	0	1	-360	360;
	3850	8360	0.00056	0.003531	0	0	0	0	0	0	1	-360	360;
	1710	12739	0.001461	0.00889	0	0	0	0	0	0	1	-360	360;
	11163	10341	6e-05	0.000419	0	0	0	0	0	0	1	-360	360;
	7062	9999	0.00139	0.005169	0	0	0	0	0	0	1	-360	360;
	6081	1859	0.00069	0.00825	0	0	0	0	0	0	1	-360	360;
	5428	6699	0.004969	0.030781	0	0	0	0	0	0	1	-360	360;
	5428	13138	0.00012	0.000969	0	0	0	0	0	0	1	-360	360;
	11146	7838	0.001091	0.010159	0	0	0	0	0	0	1	-360	360;
	3981	7838	0.000969	0.00887	0	0	0	0	0	0	1	-360	360;
	6394	9677	0.00099	0.006789	0	0	0	0	0	0	1	-360	360;
	12055	3560	0.001031	0.00586	0	0	0	0	0	0	1	-360	360;
	10933	7958	0.00401	0.03345	0	0	0	0	0	0	1	-360	360;
	4969	7793	9.1e-05	0.000349	0	0	0	0	0	0	1	-360	360;
	894	333	0.00125	0.00745	0	0	0	0	0	0	1	-360	360;
	11295	5921	0.003711	0.024219	0	0	0	0	0	0	1	-360	360;
	12094	6599	0.00169	0.02156	0	0	0	0	0	0	1	-360	360;
	12094	6599	0.00171	0.02114	0	0	0	0	0	0	1	-360	360;
	9593	6228	0.00088	0.00562	0	0	0	0	0	0	1	-360	360;
	9593	6228	0.0008	0.00618	0	0	0	0	0	0	1	-360	360;
	4228	9003	0.00161	0.01014	0	0	0	0	0	0	1	-360	360;
	5785	3502	0.00032	0.002031	0	0	0	0	0	0	1	-360	360;
	5785	9544	0.00011	0.000659	0	0	0	0	0	0	1	-360	360;
	2746	5623	0.00042	0.00505	0	0	0	0	0	0	1	-360	360;
	2746	8232	0.00043	0.00494	0	0	0	0	0	0	1	-360	360;
	7408	7203	0.00038	0.000711	0	0	0	0	0	0	1	-360	360;
	7408	7203	0.00037	0.000659	0	0	0	0	0	0	1	-360	360;
	6061	2350	0.00248	0.03096	0	0	0	0	0	0	1	-360	360;
	235	5120	0.00375	0.019729	0	0	0	0	0	0	1	-360	360;
	235	2729	0.00518	0.020021	0	0	0	0	0	0	1	-360	360;
	11675	11931	0.000771	0.003711	0	0	0	0	0	0	1	-360	360;
	11675	2104	0.00076	0.003789	0	0	0	0	0	0	1	-360	360;
	3290	5194	0.000781	0.00574	0	0	0	0	0	0	1	-360	360;
	5185	13237	0.00089	0.00562	0	0	0	0	0	0	1	-360	360;
	2875	13167	0.000331	0.001401	0	0	0	0	0	0	1	-360	360;
	7555	13167	0.000401	0.00137	0	0	0	0	0	0	1	-360	360;
	11678	7606	0.00051	0.003349	0	0	0	0	0	0	1	-360	360;
	2920	12691	0.000461	0.00168	0	0	0	0	0	0	1	-360	360;
	2920	12691	0.000469	0.00164	0	0	0	0	0	0	1	-360	360;
	6839	9955	0.006081	0.03114	0	0	0	0	0	0	1	-360	360;
	11003	1122	0.005831	0.02218	0	0	0	0	0	0	1	-360	360;
	13289	7567	0.00013	0.000461	0	0	0	0	0	0	1	-360	360;
	5282	2876	0.00013	0.000461	0	0	0	0	0	0	1	-360	360;
	1228	5921	0.00262	0.020289	0	0	0	0	0	0	1	-360	360;
	1228	4121	0.00164	0.01194	0	0	0	0	0	0	1	-360	360;
	10787	12978	0.006479	0.02845	0	0	0	0	0	0	1	-360	360;
	11726	2866	0.00294	0.014419	0	0	0	0	0	0	1	-360	360;
	11726	2866	0.00238	0.01463	0	0	0	0	0	0	1	-360	360;
	11726	2497	0.002219	0.00993	0	0	0	0	0	0	1	-360	360;
	1984	3285	0.008651	0.04886	0	0	0	0	0	0	1	-360	360;
	6604	11608	0.00464	0.030979	0	0	0	0	0	0	1	-360	360;
	8497	661	0.00092	0.00889	0	0	0	0	0	0	1	-360	360;
	8497	661	0.00093	0.00869	0	0	0	0	0	0	1	-360	360;
	8497	2595	0.0022	0.03485	0	0	0	0	0	0	1	-360	360;
	8497	10731	0.00074	0.00941	0	0	0	0	0	0	1	-360	360;
	8497	9409	0.00087	0.00884	0	0	0	0	0	0	1	-360	360;
	8497	2467	0.00139	0.01662	0	0	0	0	0	0	1	-360	360;
	13448	872	0.002781	0.022979	0	0	0	0	0	0	1	-360	360;
	9289	7856	0.000521	0.00407	0	0	0	0	0	0	1	-360	360;
	1024	955	0.008479	0.032841	0	0	0	0	0	0	1	-360	360;
	588	10966	0.005669	0.02901	0	0	0	0	0	0	1	-360	360;
	5921	4114	0.001031	0.004099	0	0	0	0	0	0	1	-360	360;
	5921	11837	0.00301	0.018719	0	0	0	0	0	0	1	-360	360;
	2590	4108	0.00257	0.01713	0	0	0	0	0	0	1	-360	360;
	7811	10327	0.010461	0.03836	0	0	0	0	0	0	1	-360	360;
	7811	6031	0.001719	0.011531	0	0	0	0	0	0	1	-360	360;
	7811	7150	9.1e-05	0.00056	0	0	0	0	0	0	1	-360	360;
	5722	5677	0.00476	0.025289	0	0	0	0	0	0	1	-360	360;
	184	2220	0.000219	0.00164	0	0	0	0	0	0	1	-360	360;
	285	11753	0.002159	0.01163	0	0	0	0	0	0	1	-360	360;
	285	2220	0.00295	0.01539	0	0	0	0	0	0	1	-360	360;
	4299	4528	0.007789	0.033039	0	0	0	0	0	0	1	-360	360;
	10724	4437	0.00026	0.0013	0	0	0	0	0	0	1	-360	360;
	7400	13294	0.00162	0.013229	0	0	0	0	0	0	1	-360	360;
	4564	872	0.007919	0.04194	0	0	0	0	0	0	1	-360	360;
	994	12497	0.001229	0.00401	0	0	0	0	0	0	1	-360	360;
	7517	7985	0.008021	0.042831	0	0	0	0	0	0	1	-360	360;
	9709	1122	0.007711	0.03006	0	0	0	0	0	0	1	-360	360;
	9694	1139	0.00261	0.015469	0	0	0	0	0	0	1	-360	360;
	9263	7150	0.00139	0.00536	0	0	0	0	0	0	1	-360	360;
	9263	7150	0.001419	0.00525	0	0	0	0	0	0	1	-360	360;
	2685	955	0.006289	0.032271	0	0	0	0	0	0	1	-360	360;
	2685	3018	0.011031	0.05625	0	0	0	0	0	0	1	-360	360;
	244	3350	0.001841	0.01324	0	0	0	0	0	0	1	-360	360;
	4110	14	0.000599	0.00186	0	0	0	0	0	0	1	-360	360;
	6577	6228	0.00036	0.001591	0	0	0	0	0	0	1	-360	360;
	13203	12739	0.003849	0.027651	0	0	0	0	0	0	1	-360	360;
	1206	6534	0.00186	0.00874	0	0	0	0	0	0	1	-360	360;
	12417	3090	0.00025	0.00285	0	0	0	0	0	0	1	-360	360;
	1657	11640	0.00207	0.008841	0	0	0	0	0	0	1	-360	360;
	1657	9310	6e-05	0.00039	0	0	0	0	0	0	1	-360	360;
	10852	7112	0.002159	0.008909	0	0	0	0	0	0	1	-360	360;
	5937	9310	0.000219	0.00039	0	0	0	0	0	0	1	-360	360;
	7203	3802	0.00011	0.000521	0	0	0	0	0	0	1	-360	360;
	7203	3802	0.00011	0.00049	0	0	0	0	0	0	1	-360	360;
	7203	816	0.000771	0.003789	0	0	0	0	0	0	1	-360	360;
	7203	4350	0.000229	0.002849	0	0	0	0	0	0	1	-360	360;
	11753	7664	0.00445	0.027599	0	0	0	0	0	0	1	-360	360;
	2433	4880	0.00218	0.02	0	0	0	0	0	0	1	-360	360;
	2433	6905	0.00074	0.00865	0	0	0	0	0	0	1	-360	360;
	955	1745	0.002531	0.019419	0	0	0	0	0	0	1	-360	360;
	955	11829	0.003	0.01631	0	0	0	0	0	0	1	-360	360;
	12841	2041	0.008461	0.025081	0	0	0	0	0	0	1	-360	360;
	11276	4994	0.00382	0.013669	0	0	0	0	0	0	1	-360	360;
	4528	12497	0.002581	0.01814	0	0	0	0	0	0	1	-360	360;
	9849	941	0.00201	0.012219	0	0	0	0	0	0	1	-360	360;
	13167	6320	0.00087	0.001599	0	0	0	0	0	0	1	-360	360;
	1113	7631	0.00455	0.02394	0	0	0	0	0	0	1	-360	360;
	2804	8629	0.005651	0.041771	0	0	0	0	0	0	1	-360	360;
	13326	5877	0.00069	0.001271	0	0	0	0	0	0	1	-360	360;
	2740	12386	0.00068	0.00269	0	0	0	0	0	0	1	-360	360;
	2740	2878	0.00036	0.00261	0	0	0	0	0	0	1	-360	360;
	2740	12210	0.001669	0.012031	0	0	0	0	0	0	1	-360	360;
	6943	5395	0.009669	0.03513	0	0	0	0	0	0	1	-360	360;
	6943	12497	0.001099	0.00862	0	0	0	0	0	0	1	-360	360;
	1127	2350	0.0005	0.00711	0	0	0	0	0	0	1	-360	360;
	9274	1445	0.003919	0.017781	0	0	0	0	0	0	1	-360	360;
	9274	13294	0.00161	0.019781	0	0	0	0	0	0	1	-360	360;
	9274	1139	0.0007	0.01256	0	0	0	0	0	0	1	-360	360;
	7321	7856	0.000409	0.0018	0	0	0	0	0	0	1	-360	360;
	2432	7219	2e-05	0.00043	0	0	0	0	0	0	1	-360	360;
	13529	7219	3e-05	0.0004	0	0	0	0	0	0	1	-360	360;
	422	13544	0.00968	0.056659	0	0	0	0	0	0	1	-360	360;
	11837	5395	0.003969	0.02925	0	0	0	0	0	0	1	-360	360;
	8309	13294	0.00219	0.01468	0	0	0	0	0	0	1	-360	360;
	8309	12484	0.000401	0.003159	0	0	0	0	0	0	1	-360	360;
	8309	10048	0.000531	0.003	0	0	0	0	0	0	1	-360	360;
	7606	10645	0.00136	0.00576	0	0	0	0	0	0	1	-360	360;
	13028	13294	0.000849	0.006289	0	0	0	0	0	0	1	-360	360;
	13028	13294	0.000841	0.006409	0	0	0	0	0	0	1	-360	360;
	13028	13294	0.00045	0.00506	0	0	0	0	0	0	1	-360	360;
	321	5028	0.003669	0.02988	0	0	0	0	0	0	1	-360	360;
	761	8620	0.002	0.01568	0	0	0	0	0	0	1	-360	360;
	3802	3009	0.00031	0.0008	0	0	0	0	0	0	1	-360	360;
	1725	7631	0.00043	0.00038	0	0	0	0	0	0	1	-360	360;
	10217	8724	0.00043	0.002229	0	0	0	0	0	0	1	-360	360;
	7604	8724	0.00049	0.00218	0	0	0	0	0	0	1	-360	360;
	5395	9999	0.00282	0.01844	0	0	0	0	0	0	1	-360	360;
	1242	2041	0.00462	0.014849	0	0	0	0	0	0	1	-360	360;
	7515	4880	0.00263	0.04076	0	0	0	0	0	0	1	-360	360;
	5120	3502	0.01062	0.056719	0	0	0	0	0	0	1	-360	360;
	6057	4439	0.0001	0.00105	0	0	0	0	0	0	1	-360	360;
	12691	13125	0.00239	0.017341	0	0	0	0	0	0	1	-360	360;
	5132	3602	0.001341	0.010581	0	0	0	0	0	0	1	-360	360;
	5132	8987	0.00101	0.007599	0	0	0	0	0	0	1	-360	360;
	592	2259	0.00011	0.000289	0	0	0	0	0	0	1	-360	360;
	10294	6599	0.00146	0.01769	0	0	0	0	0	0	1	-360	360;
	10294	6599	0.0013	0.02004	0	0	0	0	0	0	1	-360	360;
	9955	8105	0.007909	0.028349	0	0	0	0	0	0	1	-360	360;
	10842	3527	0.00012	0.00049	0	0	0	0	0	0	1	-360	360;
	10842	3527	0.00012	0.000479	0	0	0	0	0	0	1	-360	360;
	6599	634	0.00076	0.01018	0	0	0	0	0	0	1	-360	360;
	6599	634	0.0009	0.00865	0	0	0	0	0	0	1	-360	360;
	6599	634	0.00096	0.01294	0	0	0	0	0	0	1	-360	360;
	6599	634	0.0004	0.00756	0	0	0	0	0	0	1	-360	360;
	6599	634	0.00039	0.00771	0	0	0	0	0	0	1	-360	360;
	6599	4439	0.00163	0.02291	0	0	0	0	0	0	1	-360	360;
	4246	4511	0.00636	0.03855	0	0	0	0	0	0	1	-360	360;
	4246	4511	0.00639	0.03863	0	0	0	0	0	0	1	-360	360;
	1345	2467	0.00113	0.01362	0	0	0	0	0	0	1	-360	360;
	2528	7861	0.00861	0.05318	0	0	0	0	0	0	1	-360	360;
	6854	2392	0.0012	0.004669	0	0	0	0	0	0	1	-360	360;
	6854	10341	0.00018	0.001091	0	0	0	0	0	0	1	-360	360;
	1745	11829	0.001401	0.00612	0	0	0	0	0	0	1	-360	360;
	169	333	0.00068	0.00387	0	0	0	0	0	0	1	-360	360;
	8318	7758	0.0043	0.033591	0	0	0	0	0	0	1	-360	360;
	5251	7958	0.003039	0.01911	0	0	0	0	0	0	1	-360	360;
	4730	178	6e-05	0.00111	0	0	0	0	0	0	1	-360	360;
	10310	7008	5e-05	0.001091	0	0	0	0	0	0	1	-360	360;
	3428	4822	3.9e-05	0.00089	0	0	0	0	0	0	1	-360	360;
	2710	6572	3.9e-05	0.000901	0	0	0	0	0	0	1	-360	360;
	990	6534	0.000169	0.00111	0	0	0	0	0	0	1	-360	360;
	1422	7112	0.000331	0.0027	0	0	0	0	0	0	1	-360	360;
	9411	2866	0.00026	0.00369	0	0	0	0	0	0	1	-360	360;
	2392	10341	0.000341	0.00274	0	0	0	0	0	0	1	-360	360;
	7787	10966	0.00038	0.00506	0	0	0	0	0	0	1	-360	360;
	3502	9544	0.00012	0.000831	0	0	0	0	0	0	1	-360	360;
	1598	872	0.003031	0.0123	0	0	0	0	0	0	1	-360	360;
	6955	11424	0.001521	0.00774	0	0	0	0	0	0	1	-360	360;
	7219	6187	0.00059	0.00552	0	0	0	0	0	0	1	-360	360;
	7219	1702	0.00053	0.00625	0	0	0	0	0	0	1	-360	360;
	8724	13022	0.00143	0.01076	0	0	0	0	0	0	1	-360	360;
	2564	12512	0.00056	0.00593	0	0	0	0	0	0	1	-360	360;
	3540	9003	0.00205	0.01461	0	0	0	0	0	0	1	-360	360;
	3018	3285	0.01068	0.054521	0	0	0	0	0	0	1	-360	360;
	13294	12484	0.00194	0.011039	0	0	0	0	0	0	1	-360	360;
	13294	10048	0.00131	0.010099	0	0	0	0	0	0	1	-360	360;
	13294	4315	0.001289	0.014229	0	0	0	0	0	0	1	-360	360;
	7229	4672	0.00455	0.027031	0	0	0	0	0	0	1	-360	360;
	7229	4851	0.00239	0.01414	0	0	0	0	0	0	1	-360	360;
	7229	5545	0.004039	0.029669	0	0	0	0	0	0	1	-360	360;
	7229	5028	0.001099	0.007771	0	0	0	0	0	0	1	-360	360;
	7838	2878	0.00082	0.00762	0	0	0	0	0	0	1	-360	360;
	7838	6047	0.000719	0.0033	0	0	0	0	0	0	1	-360	360;
	7838	8030	0.000599	0.00361	0	0	0	0	0	0	1	-360	360;
	7838	12210	0.000521	0.004039	0	0	0	0	0	0	1	-360	360;
	7838	7567	0.000901	0.0088	0	0	0	0	0	0	1	-360	360;
	10989	10214	0.000781	0.00306	0	0	0	0	0	0	1	-360	360;
	10989	3444	0.000419	0.00399	0	0	0	0	0	0	1	-360	360;
	10989	2876	0.000841	0.009841	0	0	0	0	0	0	1	-360	360;
	9484	8987	0.004021	0.03064	0	0	0	0	0	0	1	-360	360;
	7112	10193	0.0023	0.009031	0	0	0	0	0	0	1	-360	360;
	7112	10193	0.002081	0.009469	0	0	0	0	0	0	1	-360	360;
	7112	8007	0.00164	0.008831	0	0	0	0	0	0	1	-360	360;
	7112	13142	0.00094	0.00488	0	0	0	0	0	0	1	-360	360;
	7112	13481	0.000901	0.00551	0	0	0	0	0	0	1	-360	360;
	7112	13092	0.00051	0.005961	0	0	0	0	0	0	1	-360	360;
	2866	2497	0.00113	0.004349	0	0	0	0	0	0	1	-360	360;
	2104	9674	0.002081	0.013031	0	0	0	0	0	0	1	-360	360;
	1158	6736	0.00116	0.01438	0	0	0	0	0	0	1	-360	360;
	12805	6929	0.00013	0.001031	0	0	0	0	0	0	1	-360	360;
	6031	896	0.00036	0.002659	0	0	0	0	0	0	1	-360	360;
	9999	7856	0.000771	0.004539	0	0	0	0	0	0	1	-360	360;
	4808	6001	0.011902	0.04088	0	0	0	0	0	0	1	-360	360;
	252	4808	0.002231	0.010138	0	0	0	0	0	0	1	-360	360;
	1639	237	0.00023	0.0025	0	0	0	0	0	0	1	-360	360;
	1639	237	0.00023	0.00254	0	0	0	0	0	0	1	-360	360;
	7298	1639	0.00179	0.01478	0	0	0	0	0	0	1	-360	360;
	7298	1639	0.00183	0.01446	0	0	0	0	0	0	1	-360	360;
	1639	2194	0.00031	0.00312	0	0	0	0	0	0	1	-360	360;
	1523	1639	0.00031	0.00289	0	0	0	0	0	0	1	-360	360;
	1283	1639	0.00021	0.00194	0	0	0	0	0	0	1	-360	360;
	9626	11157	0.001489	0.006249	0	0	0	0	0	0	1	-360	360;
	11343	11157	0.010209	0.042502	0	0	0	0	0	0	1	-360	360;
	9626	9463	0.001462	0.006369	0	0	0	0	0	0	1	-360	360;
	9463	12239	0.0048	0.01772	0	0	0	0	0	0	1	-360	360;
	9626	8365	0.00328	0.01472	0	0	0	0	0	0	1	-360	360;
	675	9626	0	0.000271	0	0	0	0	0	0	1	-360	360;
	12239	9626	0.002031	0.007591	0	0	0	0	0	0	1	-360	360;
	6077	10086	0.00372	0.014431	0	0	0	0	0	0	1	-360	360;
	5652	6077	0	0.000271	0	0	0	0	0	0	1	-360	360;
	6077	6701	0.00108	0.003698	0	0	0	0	0	0	1	-360	360;
	6077	13319	0.000431	0.001658	0	0	0	0	0	0	1	-360	360;
	10086	4731	0.000782	0.002662	0	0	0	0	0	0	1	-360	360;
	8365	914	0.001729	0.005498	0	0	0	0	0	0	1	-360	360;
	11701	2753	0.00221	0.01891	0	0	0	0	0	0	1	-360	360;
	12028	2753	0.00335	0.03107	0	0	0	0	0	0	1	-360	360;
	2753	642	0.00109	0.01324	0	0	0	0	0	0	1	-360	360;
	3087	4118	0.027218	0.118858	0	0	0	0	0	0	1	-360	360;
	9293	3087	0.002782	0.010778	0	0	0	0	0	0	1	-360	360;
	3087	3703	0.03336	0.110049	0	0	0	0	0	0	1	-360	360;
	3087	3703	0.02972	0.124729	0	0	0	0	0	0	1	-360	360;
	3087	13618	0.00244	0.009351	0	0	0	0	0	0	1	-360	360;
	3087	13618	0.002151	0.009649	0	0	0	0	0	0	1	-360	360;
	3087	13618	0.00244	0.009458	0	0	0	0	0	0	1	-360	360;
	9433	8446	0.017102	0.080911	0	0	0	0	0	0	1	-360	360;
	3512	9433	0.030871	0.129529	0	0	0	0	0	0	1	-360	360;
	9433	2678	0.008778	0.041551	0	0	0	0	0	0	1	-360	360;
	9433	2678	0.008591	0.038471	0	0	0	0	0	0	1	-360	360;
	9433	2678	0.01032	0.038049	0	0	0	0	0	0	1	-360	360;
	9433	1014	0.004151	0.017409	0	0	0	0	0	0	1	-360	360;
	9433	1014	0.004062	0.017751	0	0	0	0	0	0	1	-360	360;
	11444	8932	0.000338	0.00156	0	0	0	0	0	0	1	-360	360;
	12617	8932	0.002689	0.010018	0	0	0	0	0	0	1	-360	360;
	8932	6345	0.020471	0.070289	0	0	0	0	0	0	1	-360	360;
	5873	4232	4e-05	0.000271	0	0	0	0	0	0	1	-360	360;
	7419	5873	0.023289	0.097751	0	0	0	0	0	0	1	-360	360;
	6677	4232	0.005969	0.02292	0	0	0	0	0	0	1	-360	360;
	1312	4232	0.00428	0.019138	0	0	0	0	0	0	1	-360	360;
	3645	12617	0.01588	0.054529	0	0	0	0	0	0	1	-360	360;
	9255	6247	0	0.000271	0	0	0	0	0	0	1	-360	360;
	6247	4487	0.00368	0.007311	0	0	0	0	0	0	1	-360	360;
	10344	6247	0.006369	0.0324	0	0	0	0	0	0	1	-360	360;
	1444	6247	0.000791	0.003569	0	0	0	0	0	0	1	-360	360;
	6126	6247	0.000809	0.003431	0	0	0	0	0	0	1	-360	360;
	4487	8617	0.008871	0.004431	0	0	0	0	0	0	1	-360	360;
	13528	7973	0	0.000271	0	0	0	0	0	0	1	-360	360;
	9798	5047	0	0.000271	0	0	0	0	0	0	1	-360	360;
	11397	5526	0	0.000271	0	0	0	0	0	0	1	-360	360;
	3873	1312	0	0.000271	0	0	0	0	0	0	1	-360	360;
	911	6565	0	0.000271	0	0	0	0	0	0	1	-360	360;
	1374	11610	0	0.000271	0	0	0	0	0	0	1	-360	360;
	12941	6568	0.00096	0.003391	0	0	0	0	0	0	1	-360	360;
	2773	6568	0	0.000271	0	0	0	0	0	0	1	-360	360;
	6568	9371	0.004471	0.021831	0	0	0	0	0	0	1	-360	360;
	6568	5098	0.00624	0.03	0	0	0	0	0	0	1	-360	360;
	4840	7947	0.0008	0.0038	0	0	0	0	0	0	1	-360	360;
	252	4840	0.010058	0.035969	0	0	0	0	0	0	1	-360	360;
	13183	7947	0.000782	0.00352	0	0	0	0	0	0	1	-360	360;
	8650	13183	0.008471	0.040089	0	0	0	0	0	0	1	-360	360;
	11379	8575	0.00038	0.00312	0	0	0	0	0	0	1	-360	360;
	11379	8575	0.00038	0.00305	0	0	0	0	0	0	1	-360	360;
	884	11379	0.00255	0.02375	0	0	0	0	0	0	1	-360	360;
	884	11379	0.0026	0.02326	0	0	0	0	0	0	1	-360	360;
	11379	8853	0.00268	0.03039	0	0	0	0	0	0	1	-360	360;
	11379	8853	0.00262	0.02814	0	0	0	0	0	0	1	-360	360;
	3896	12046	0.000258	0.000929	0	0	0	0	0	0	1	-360	360;
	4579	12046	0.003049	0.013911	0	0	0	0	0	0	1	-360	360;
	3896	8381	0.000258	0.000911	0	0	0	0	0	0	1	-360	360;
	11559	8381	0.000871	0.003991	0	0	0	0	0	0	1	-360	360;
	8049	4824	0.024218	0.1016	0	0	0	0	0	0	1	-360	360;
	4824	10042	0	0.000271	0	0	0	0	0	0	1	-360	360;
	3006	3247	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8049	3006	0.02372	0.103591	0	0	0	0	0	0	1	-360	360;
	3006	10279	0	0.000271	0	0	0	0	0	0	1	-360	360;
	13554	3247	0	0.00028	0	0	0	0	0	0	1	-360	360;
	237	4835	0.00038	0.00396	0	0	0	0	0	0	1	-360	360;
	237	4835	0.00038	0.00404	0	0	0	0	0	0	1	-360	360;
	237	4835	0.00042	0.0036	0	0	0	0	0	0	1	-360	360;
	237	4835	0.00043	0.00352	0	0	0	0	0	0	1	-360	360;
	7298	237	0.00177	0.01845	0	0	0	0	0	0	1	-360	360;
	11610	5564	0.007831	0.035089	0	0	0	0	0	0	1	-360	360;
	11610	10760	0.002089	0.007182	0	0	0	0	0	0	1	-360	360;
	7510	5564	0.008871	0.034391	0	0	0	0	0	0	1	-360	360;
	7510	1991	0.002129	0.007031	0	0	0	0	0	0	1	-360	360;
	2911	7510	0.007222	0.032831	0	0	0	0	0	0	1	-360	360;
	7510	11277	0.006049	0.021431	0	0	0	0	0	0	1	-360	360;
	4073	12941	0.0114	0.005622	0	0	0	0	0	0	1	-360	360;
	4101	3639	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3639	13173	0.00976	0.043991	0	0	0	0	0	0	1	-360	360;
	5398	1951	0.017658	0.0684	0	0	0	0	0	0	1	-360	360;
	5398	1951	0.018	0.067	0	0	0	0	0	0	1	-360	360;
	1951	11375	0.003378	0.009631	0	0	0	0	0	0	1	-360	360;
	1951	7684	0.016409	0.07168	0	0	0	0	0	0	1	-360	360;
	1951	1941	0.0182	0.04612	0	0	0	0	0	0	1	-360	360;
	1951	4496	0.004818	0.017231	0	0	0	0	0	0	1	-360	360;
	10956	7520	0.000209	0.000911	0	0	0	0	0	0	1	-360	360;
	10956	3829	0.000249	0.000831	0	0	0	0	0	0	1	-360	360;
	12767	7285	0.00136	0.006458	0	0	0	0	0	0	1	-360	360;
	7285	7006	0.002782	0.013129	0	0	0	0	0	0	1	-360	360;
	9132	2596	0.008049	0.036031	0	0	0	0	0	0	1	-360	360;
	2596	3818	0.022311	0.0864	0	0	0	0	0	0	1	-360	360;
	974	2170	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9132	2170	0.00912	0.035311	0	0	0	0	0	0	1	-360	360;
	2170	1719	0.001471	0.005489	0	0	0	0	0	0	1	-360	360;
	10587	12770	0.00058	0.00639	0	0	0	0	0	0	1	-360	360;
	10587	12770	0.00057	0.00651	0	0	0	0	0	0	1	-360	360;
	7496	5871	0.002111	0.007542	0	0	0	0	0	0	1	-360	360;
	1713	5871	0.014049	0.037031	0	0	0	0	0	0	1	-360	360;
	393	3896	0	0.000222	0	0	0	0	0	0	1	-360	360;
	393	9903	0.002071	0.005609	0	0	0	0	0	0	1	-360	360;
	2805	393	0.00368	0.016711	0	0	0	0	0	0	1	-360	360;
	2805	393	0.0036	0.017018	0	0	0	0	0	0	1	-360	360;
	393	11929	0.00012	0.000431	0	0	0	0	0	0	1	-360	360;
	3896	10832	0.00012	0.000431	0	0	0	0	0	0	1	-360	360;
	7973	8617	0.01328	0.031071	0	0	0	0	0	0	1	-360	360;
	12426	7973	0.031911	0.071711	0	0	0	0	0	0	1	-360	360;
	9077	5047	0.000382	0.001289	0	0	0	0	0	0	1	-360	360;
	2805	9077	0.002582	0.00988	0	0	0	0	0	0	1	-360	360;
	3733	5526	0.00036	0.001218	0	0	0	0	0	0	1	-360	360;
	3263	3733	0.002458	0.00908	0	0	0	0	0	0	1	-360	360;
	10122	8913	0.000911	0.003271	0	0	0	0	0	0	1	-360	360;
	252	10122	0.01436	0.05344	0	0	0	0	0	0	1	-360	360;
	230	4148	0.000462	0.001631	0	0	0	0	0	0	1	-360	360;
	230	8913	0.000929	0.0032	0	0	0	0	0	0	1	-360	360;
	12244	4148	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9173	4148	0.013391	0.060849	0	0	0	0	0	0	1	-360	360;
	1312	7335	0.010631	0.048342	0	0	0	0	0	0	1	-360	360;
	8446	4829	0.005498	0.021302	0	0	0	0	0	0	1	-360	360;
	4829	13318	0.001449	0.006329	0	0	0	0	0	0	1	-360	360;
	1191	13187	0.007951	0.029329	0	0	0	0	0	0	1	-360	360;
	1191	2678	0.02072	0.086938	0	0	0	0	0	0	1	-360	360;
	347	1191	0.007182	0.02672	0	0	0	0	0	0	1	-360	360;
	10616	3388	0.00424	0.019129	0	0	0	0	0	0	1	-360	360;
	4118	10616	0.010542	0.047938	0	0	0	0	0	0	1	-360	360;
	10616	13618	0.016751	0.06488	0	0	0	0	0	0	1	-360	360;
	6565	5255	0.016671	0.059071	0	0	0	0	0	0	1	-360	360;
	6565	9086	0.00536	0.024151	0	0	0	0	0	0	1	-360	360;
	8212	4118	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4118	7739	0.007089	0.017938	0	0	0	0	0	0	1	-360	360;
	8171	4118	0.020849	0.052831	0	0	0	0	0	0	1	-360	360;
	8362	7739	0.010609	0.005209	0	0	0	0	0	0	1	-360	360;
	3516	7514	0.020898	0.098071	0	0	0	0	0	0	1	-360	360;
	9533	7514	0.002258	0.008409	0	0	0	0	0	0	1	-360	360;
	618	9235	0.00056	0.00236	0	0	0	0	0	0	1	-360	360;
	9235	1444	0.014258	0.064311	0	0	0	0	0	0	1	-360	360;
	12947	2832	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9717	2832	0.040249	0.143902	0	0	0	0	0	0	1	-360	360;
	12933	2832	0.021778	0.071862	0	0	0	0	0	0	1	-360	360;
	2832	6618	0.022742	0.09932	0	0	0	0	0	0	1	-360	360;
	7682	2832	0.021511	0.07092	0	0	0	0	0	0	1	-360	360;
	7871	6243	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6243	12473	0.010818	0.0454	0	0	0	0	0	0	1	-360	360;
	3271	6243	0.013671	0.059698	0	0	0	0	0	0	1	-360	360;
	8748	11441	0.00011	0.00059	0	0	0	0	0	0	1	-360	360;
	2551	11441	0.00093	0.00993	0	0	0	0	0	0	1	-360	360;
	2922	8748	0.00037	0.00305	0	0	0	0	0	0	1	-360	360;
	11252	1428	0.0001	0.00066	0	0	0	0	0	0	1	-360	360;
	2551	11252	0.00105	0.00973	0	0	0	0	0	0	1	-360	360;
	1428	8526	0.00037	0.00298	0	0	0	0	0	0	1	-360	360;
	2922	7298	0.00105	0.01196	0	0	0	0	0	0	1	-360	360;
	8526	237	0.00073	0.00626	0	0	0	0	0	0	1	-360	360;
	5694	11284	0.00112	0.002329	0	0	0	0	0	0	1	-360	360;
	5694	8131	0.0068	0.032178	0	0	0	0	0	0	1	-360	360;
	8083	5694	0.013871	0.06308	0	0	0	0	0	0	1	-360	360;
	9563	6184	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5713	6184	0.00636	0.028689	0	0	0	0	0	0	1	-360	360;
	5290	6184	0.005231	0.020942	0	0	0	0	0	0	1	-360	360;
	6536	6184	0.019129	0.063111	0	0	0	0	0	0	1	-360	360;
	6184	150	0.006649	0.0298	0	0	0	0	0	0	1	-360	360;
	11284	8296	0.0006	0.001609	0	0	0	0	0	0	1	-360	360;
	163	618	0	0.00028	0	0	0	0	0	0	1	-360	360;
	12128	618	0.014849	0.066942	0	0	0	0	0	0	1	-360	360;
	12700	8863	0.00392	0.009182	0	0	0	0	0	0	1	-360	360;
	8863	5451	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8863	450	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8863	8643	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3649	9001	0.00384	0.00936	0	0	0	0	0	0	1	-360	360;
	9001	4541	0	0.00028	0	0	0	0	0	0	1	-360	360;
	9001	9531	0	0.00028	0	0	0	0	0	0	1	-360	360;
	9001	2255	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3649	12700	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5451	10876	0.02008	0.09128	0	0	0	0	0	0	1	-360	360;
	5976	4541	0.023991	0.092942	0	0	0	0	0	0	1	-360	360;
	8339	450	0.01384	0.06552	0	0	0	0	0	0	1	-360	360;
	542	9531	0.00476	0.022538	0	0	0	0	0	0	1	-360	360;
	8643	3629	0.00496	0.016382	0	0	0	0	0	0	1	-360	360;
	2255	11339	0.00244	0.008049	0	0	0	0	0	0	1	-360	360;
	13246	822	0.000671	0.002502	0	0	0	0	0	0	1	-360	360;
	822	7359	0.0066	0.029569	0	0	0	0	0	0	1	-360	360;
	13246	10805	0.000689	0.002449	0	0	0	0	0	0	1	-360	360;
	139	10805	0.007271	0.024969	0	0	0	0	0	0	1	-360	360;
	7762	13246	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5713	4681	0.007649	0.031849	0	0	0	0	0	0	1	-360	360;
	10606	8631	0	0.000222	0	0	0	0	0	0	1	-360	360;
	10606	10364	0	0.000222	0	0	0	0	0	0	1	-360	360;
	419	10606	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5040	8631	0.019471	0.074818	0	0	0	0	0	0	1	-360	360;
	9040	10364	0.009231	0.041622	0	0	0	0	0	0	1	-360	360;
	3931	12565	0	0.000222	0	0	0	0	0	0	1	-360	360;
	11067	9293	0.000258	0.001089	0	0	0	0	0	0	1	-360	360;
	11067	5294	0.019871	0.042889	0	0	0	0	0	0	1	-360	360;
	5294	12565	0.00076	0.001978	0	0	0	0	0	0	1	-360	360;
	8825	12565	0.000751	0.001831	0	0	0	0	0	0	1	-360	360;
	11613	8825	0.039151	0.088	0	0	0	0	0	0	1	-360	360;
	586	8062	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8062	4100	0.032351	0.141289	0	0	0	0	0	0	1	-360	360;
	1357	8062	0.022569	0.074449	0	0	0	0	0	0	1	-360	360;
	8062	7803	5.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	9780	13508	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2161	13508	0.009542	0.031471	0	0	0	0	0	0	1	-360	360;
	10857	13508	0.021671	0.09856	0	0	0	0	0	0	1	-360	360;
	4459	3333	0.005609	0.013978	0	0	0	0	0	0	1	-360	360;
	6186	3333	0.003751	0.01016	0	0	0	0	0	0	1	-360	360;
	13360	5398	0.000529	0.00208	0	0	0	0	0	0	1	-360	360;
	13360	5991	0.008862	0.029231	0	0	0	0	0	0	1	-360	360;
	1341	8940	0.000538	0.00204	0	0	0	0	0	0	1	-360	360;
	1341	13579	0.009569	0.026289	0	0	0	0	0	0	1	-360	360;
	5398	4990	0.00028	0.001209	0	0	0	0	0	0	1	-360	360;
	1713	4990	0.005489	0.014249	0	0	0	0	0	0	1	-360	360;
	8940	4459	0.000271	0.001231	0	0	0	0	0	0	1	-360	360;
	5398	8382	0.000338	0.001311	0	0	0	0	0	0	1	-360	360;
	8382	8368	0.005791	0.02716	0	0	0	0	0	0	1	-360	360;
	6381	10587	0.00288	0.03132	0	0	0	0	0	0	1	-360	360;
	6381	10587	0.00282	0.03191	0	0	0	0	0	0	1	-360	360;
	10587	8575	0.00111	0.01111	0	0	0	0	0	0	1	-360	360;
	10587	9846	0.00108	0.01133	0	0	0	0	0	0	1	-360	360;
	8940	1704	0.008751	0.033911	0	0	0	0	0	0	1	-360	360;
	8940	440	0.002418	0.00936	0	0	0	0	0	0	1	-360	360;
	6095	6950	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6095	413	0.00944	0.044271	0	0	0	0	0	0	1	-360	360;
	6950	11798	0.006542	0.0272	0	0	0	0	0	0	1	-360	360;
	5255	3483	0.001591	0.005658	0	0	0	0	0	0	1	-360	360;
	3703	1595	0.000711	0.001822	0	0	0	0	0	0	1	-360	360;
	1595	6728	0.010369	0.021511	0	0	0	0	0	0	1	-360	360;
	3703	11613	0.001978	0.0068	0	0	0	0	0	0	1	-360	360;
	790	8757	0.009062	0.023902	0	0	0	0	0	0	1	-360	360;
	8757	7635	0.0092	0.025249	0	0	0	0	0	0	1	-360	360;
	12443	1270	0.0022	0.005942	0	0	0	0	0	0	1	-360	360;
	7225	12443	0.016169	0.03352	0	0	0	0	0	0	1	-360	360;
	794	7635	0.006902	0.014898	0	0	0	0	0	0	1	-360	360;
	11291	794	0.018609	0.041849	0	0	0	0	0	0	1	-360	360;
	5218	7202	0.00156	0.00696	0	0	0	0	0	0	1	-360	360;
	7746	5218	0.01592	0.07532	0	0	0	0	0	0	1	-360	360;
	3544	7202	0.00176	0.006822	0	0	0	0	0	0	1	-360	360;
	3544	12677	0.015542	0.069582	0	0	0	0	0	0	1	-360	360;
	10732	2326	0.00176	0.004631	0	0	0	0	0	0	1	-360	360;
	3117	2326	0.02272	0.061458	0	0	0	0	0	0	1	-360	360;
	9132	4036	0.00348	0.015809	0	0	0	0	0	0	1	-360	360;
	12950	4036	0.013449	0.056431	0	0	0	0	0	0	1	-360	360;
	7923	1802	0.002911	0.01272	0	0	0	0	0	0	1	-360	360;
	9132	1802	0.0034	0.016102	0	0	0	0	0	0	1	-360	360;
	9635	7923	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7923	11258	0.030151	0.079538	0	0	0	0	0	0	1	-360	360;
	7746	7923	0.017769	0.063511	0	0	0	0	0	0	1	-360	360;
	7923	6022	0.031249	0.084538	0	0	0	0	0	0	1	-360	360;
	12842	9132	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9132	6648	0.013698	0.047049	0	0	0	0	0	0	1	-360	360;
	9132	6648	0.01396	0.046031	0	0	0	0	0	0	1	-360	360;
	12007	9132	0.007938	0.037569	0	0	0	0	0	0	1	-360	360;
	12950	9132	0.016751	0.070271	0	0	0	0	0	0	1	-360	360;
	10732	9014	0.00172	0.00472	0	0	0	0	0	0	1	-360	360;
	9014	2947	0.007271	0.015689	0	0	0	0	0	0	1	-360	360;
	764	3719	0.001471	0.004	0	0	0	0	0	0	1	-360	360;
	764	4236	0.001991	0.007729	0	0	0	0	0	0	1	-360	360;
	12284	764	0.003151	0.047689	0	0	0	0	0	0	1	-360	360;
	764	529	0.000809	0.002769	0	0	0	0	0	0	1	-360	360;
	6198	764	0.002009	0.020209	0	0	0	0	0	0	1	-360	360;
	6176	9325	0	0.00028	0	0	0	0	0	0	1	-360	360;
	9325	12936	0.001	0.002831	0	0	0	0	0	0	1	-360	360;
	9325	11865	0.001502	0.005169	0	0	0	0	0	0	1	-360	360;
	8670	9325	0.001969	0.0206	0	0	0	0	0	0	1	-360	360;
	11178	5290	0.000151	0.000649	0	0	0	0	0	0	1	-360	360;
	11178	10185	0.000142	0.000662	0	0	0	0	0	0	1	-360	360;
	3402	10185	0.004409	0.01628	0	0	0	0	0	0	1	-360	360;
	9786	9817	0	0.000222	0	0	0	0	0	0	1	-360	360;
	13518	11035	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9817	13076	0.00028	0.000791	0	0	0	0	0	0	1	-360	360;
	8225	11035	0.000622	0.001671	0	0	0	0	0	0	1	-360	360;
	6411	6633	0.000471	0.0016	0	0	0	0	0	0	1	-360	360;
	6323	6633	0.013889	0.065751	0	0	0	0	0	0	1	-360	360;
	6411	4659	0.000471	0.00156	0	0	0	0	0	0	1	-360	360;
	6648	4659	0.018262	0.082978	0	0	0	0	0	0	1	-360	360;
	1719	3818	0.02168	0.077489	0	0	0	0	0	0	1	-360	360;
	11708	8049	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8049	13542	0.000538	0.001902	0	0	0	0	0	0	1	-360	360;
	4579	8049	0.004578	0.017031	0	0	0	0	0	0	1	-360	360;
	4579	8049	0.004662	0.016671	0	0	0	0	0	0	1	-360	360;
	6423	6648	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6648	4588	0.0242	0.093769	0	0	0	0	0	0	1	-360	360;
	12473	2908	0.002289	0.008129	0	0	0	0	0	0	1	-360	360;
	7076	8795	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8795	5410	0.00668	0.02588	0	0	0	0	0	0	1	-360	360;
	1671	5410	0.00308	0.013778	0	0	0	0	0	0	1	-360	360;
	5040	8795	0.032169	0.124618	0	0	0	0	0	0	1	-360	360;
	1891	12284	0.006778	0.02236	0	0	0	0	0	0	1	-360	360;
	11958	1891	0.00308	0.011929	0	0	0	0	0	0	1	-360	360;
	8525	10380	0.016911	0.070391	0	0	0	0	0	0	1	-360	360;
	3703	8525	0.03196	0.149969	0	0	0	0	0	0	1	-360	360;
	8775	11701	0.00016	0.00049	0	0	0	0	0	0	1	-360	360;
	4404	8775	7e-05	0.00049	0	0	0	0	0	0	1	-360	360;
	11957	9164	0.0022	0.00852	0	0	0	0	0	0	1	-360	360;
	9361	9164	0.000431	0.000951	0	0	0	0	0	0	1	-360	360;
	11957	8491	0.002249	0.008351	0	0	0	0	0	0	1	-360	360;
	11338	8491	0.019431	0.053351	0	0	0	0	0	0	1	-360	360;
	7279	11957	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8431	7685	0.001929	0.008111	0	0	0	0	0	0	1	-360	360;
	6198	8431	0.00652	0.021511	0	0	0	0	0	0	1	-360	360;
	9405	7685	0.001889	0.008271	0	0	0	0	0	0	1	-360	360;
	9405	12284	0.012289	0.051551	0	0	0	0	0	0	1	-360	360;
	4300	7685	0.001431	0.005071	0	0	0	0	0	0	1	-360	360;
	1672	7685	0	0.00028	0	0	0	0	0	0	1	-360	360;
	12950	7419	0.007271	0.0344	0	0	0	0	0	0	1	-360	360;
	10600	7419	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2830	12473	0.003289	0.011329	0	0	0	0	0	0	1	-360	360;
	12473	1941	0.012129	0.032831	0	0	0	0	0	0	1	-360	360;
	12473	252	0.006031	0.025289	0	0	0	0	0	0	1	-360	360;
	12473	252	0.005911	0.025778	0	0	0	0	0	0	1	-360	360;
	685	12473	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3091	12128	0.000529	0.001969	0	0	0	0	0	0	1	-360	360;
	12425	12128	0.000538	0.001929	0	0	0	0	0	0	1	-360	360;
	2451	10540	0.001151	0.004222	0	0	0	0	0	0	1	-360	360;
	1718	2451	0.003818	0.017222	0	0	0	0	0	0	1	-360	360;
	3629	2639	0.00416	0.019502	0	0	0	0	0	0	1	-360	360;
	7802	9931	0.001991	0.00372	0	0	0	0	0	0	1	-360	360;
	7802	4600	0.009889	0.026751	0	0	0	0	0	0	1	-360	360;
	5621	9931	0.002018	0.00364	0	0	0	0	0	0	1	-360	360;
	10590	5621	0.01644	0.036942	0	0	0	0	0	0	1	-360	360;
	12895	9931	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3703	8527	0.019071	0.07388	0	0	0	0	0	0	1	-360	360;
	758	8527	0.02804	0.1256	0	0	0	0	0	0	1	-360	360;
	8532	8527	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8527	8362	0.001911	0.000751	0	0	0	0	0	0	1	-360	360;
	10252	6239	0.008098	0.028689	0	0	0	0	0	0	1	-360	360;
	10926	10252	0.017609	0.068191	0	0	0	0	0	0	1	-360	360;
	10252	3263	0.015018	0.058191	0	0	0	0	0	0	1	-360	360;
	9796	9610	0.007031	0.014578	0	0	0	0	0	0	1	-360	360;
	9796	6054	0.003791	0.012529	0	0	0	0	0	0	1	-360	360;
	9796	6748	0.010249	0.035209	0	0	0	0	0	0	1	-360	360;
	11343	1319	0.002489	0.005818	0	0	0	0	0	0	1	-360	360;
	4961	11343	0.002071	0.009751	0	0	0	0	0	0	1	-360	360;
	4961	4744	0.001871	0.006902	0	0	0	0	0	0	1	-360	360;
	4961	6902	0.003658	0.014062	0	0	0	0	0	0	1	-360	360;
	11339	11765	0.002289	0.008178	0	0	0	0	0	0	1	-360	360;
	1724	11765	0.002329	0.008	0	0	0	0	0	0	1	-360	360;
	1724	3629	0.002071	0.009249	0	0	0	0	0	0	1	-360	360;
	2161	8850	0.005951	0.021249	0	0	0	0	0	0	1	-360	360;
	542	8850	0.006058	0.0208	0	0	0	0	0	0	1	-360	360;
	6932	1570	0.000942	0.003129	0	0	0	0	0	0	1	-360	360;
	6932	6054	0.030849	0.10176	0	0	0	0	0	0	1	-360	360;
	3202	1570	0.00084	0.003551	0	0	0	0	0	0	1	-360	360;
	3629	3202	0.002378	0.007831	0	0	0	0	0	0	1	-360	360;
	5636	1943	0.007231	0.01624	0	0	0	0	0	0	1	-360	360;
	2169	5636	0.001311	0.005111	0	0	0	0	0	0	1	-360	360;
	5636	8528	0.001338	0.005009	0	0	0	0	0	0	1	-360	360;
	5340	5636	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1943	830	0.000338	0.00116	0	0	0	0	0	0	1	-360	360;
	830	1824	0.025778	0.069751	0	0	0	0	0	0	1	-360	360;
	3155	830	0.004009	0.01376	0	0	0	0	0	0	1	-360	360;
	7204	830	0.00408	0.013462	0	0	0	0	0	0	1	-360	360;
	5491	830	0.000329	0.001262	0	0	0	0	0	0	1	-360	360;
	7015	830	0	0.00028	0	0	0	0	0	0	1	-360	360;
	12793	2076	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3849	1843	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2378	2076	0.000711	0.002018	0	0	0	0	0	0	1	-360	360;
	1843	7413	0.00068	0.001502	0	0	0	0	0	0	1	-360	360;
	5527	1318	0.00012	0.00056	0	0	0	0	0	0	1	-360	360;
	5239	1318	0.02516	0.086418	0	0	0	0	0	0	1	-360	360;
	5527	9073	0.00012	0.00052	0	0	0	0	0	0	1	-360	360;
	9073	10857	0.032591	0.116511	0	0	0	0	0	0	1	-360	360;
	12037	5527	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3269	3091	0.01752	0.062098	0	0	0	0	0	0	1	-360	360;
	12425	3859	0.00076	0.003551	0	0	0	0	0	0	1	-360	360;
	13657	3483	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	13657	8214	0.000858	0.004458	0	0	0	0	0	0	1	-360	360;
	8214	1933	8e-05	0.000338	0	0	0	0	0	0	1	-360	360;
	8330	9918	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	9918	11992	0.000982	0.004369	0	0	0	0	0	0	1	-360	360;
	3977	8330	0.01704	0.083271	0	0	0	0	0	0	1	-360	360;
	11992	4374	7.1e-05	0.000351	0	0	0	0	0	0	1	-360	360;
	7160	8831	0.000271	0.00068	0	0	0	0	0	0	1	-360	360;
	9985	7160	0.012809	0.05376	0	0	0	0	0	0	1	-360	360;
	7160	10822	0.00364	0.015262	0	0	0	0	0	0	1	-360	360;
	1669	7266	0.00012	0.000431	0	0	0	0	0	0	1	-360	360;
	12370	7266	0.003982	0.015271	0	0	0	0	0	0	1	-360	360;
	1669	5392	0.00012	0.000418	0	0	0	0	0	0	1	-360	360;
	11150	5392	0.000431	0.00172	0	0	0	0	0	0	1	-360	360;
	171	4812	0.00024	0.00032	0	0	0	0	0	0	1	-360	360;
	636	171	0.002418	0.011462	0	0	0	0	0	0	1	-360	360;
	10097	6536	0.00028	0.000311	0	0	0	0	0	0	1	-360	360;
	10097	12370	0.003658	0.01664	0	0	0	0	0	0	1	-360	360;
	12370	4812	0.003582	0.016951	0	0	0	0	0	0	1	-360	360;
	2675	1175	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1175	3358	0.006231	0.029249	0	0	0	0	0	0	1	-360	360;
	1175	11150	0.00144	0.005529	0	0	0	0	0	0	1	-360	360;
	1175	5954	0.011111	0.023978	0	0	0	0	0	0	1	-360	360;
	12932	1669	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3438	8136	0.001182	0.002409	0	0	0	0	0	0	1	-360	360;
	9201	4395	0.006098	0.029822	0	0	0	0	0	0	1	-360	360;
	4395	13445	0.00096	0.003378	0	0	0	0	0	0	1	-360	360;
	6153	9717	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9717	2277	0.01044	0.03444	0	0	0	0	0	0	1	-360	360;
	9717	2277	0.009298	0.03904	0	0	0	0	0	0	1	-360	360;
	9717	7682	0.017311	0.078738	0	0	0	0	0	0	1	-360	360;
	9210	11258	0	0.000222	0	0	0	0	0	0	1	-360	360;
	11258	790	0.012449	0.029129	0	0	0	0	0	0	1	-360	360;
	1052	1711	0.00015	0.00045	0	0	0	0	0	0	1	-360	360;
	679	1711	0.00107	0.01215	0	0	0	0	0	0	1	-360	360;
	4404	1052	6e-05	0.00056	0	0	0	0	0	0	1	-360	360;
	6112	12028	0.00035	0.00284	0	0	0	0	0	0	1	-360	360;
	11185	12028	0.00163	0.01512	0	0	0	0	0	0	1	-360	360;
	7298	12028	0.00115	0.00913	0	0	0	0	0	0	1	-360	360;
	4111	12028	0.00133	0.01513	0	0	0	0	0	0	1	-360	360;
	4111	12028	0.00131	0.01401	0	0	0	0	0	0	1	-360	360;
	679	12028	0.00013	0.00129	0	0	0	0	0	0	1	-360	360;
	10415	12028	0.00077	0.00664	0	0	0	0	0	0	1	-360	360;
	3629	6054	0.028391	0.12908	0	0	0	0	0	0	1	-360	360;
	12725	3629	0.01216	0.05532	0	0	0	0	0	0	1	-360	360;
	11778	8989	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3436	8989	0.023609	0.053071	0	0	0	0	0	0	1	-360	360;
	7269	2521	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2521	10009	0.020471	0.051871	0	0	0	0	0	0	1	-360	360;
	5040	2521	0.01744	0.049862	0	0	0	0	0	0	1	-360	360;
	12677	3818	0.009111	0.039809	0	0	0	0	0	0	1	-360	360;
	7746	10319	0.033022	0.13856	0	0	0	0	0	0	1	-360	360;
	10319	3818	0.009529	0.045089	0	0	0	0	0	0	1	-360	360;
	11266	2378	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2378	1223	0.001889	0.003462	0	0	0	0	0	0	1	-360	360;
	2378	1216	0.006569	0.0142	0	0	0	0	0	0	1	-360	360;
	10297	2378	0.003982	0.018102	0	0	0	0	0	0	1	-360	360;
	2378	923	0.00232	0.007969	0	0	0	0	0	0	1	-360	360;
	2026	3231	0.002502	0.011698	0	0	0	0	0	0	1	-360	360;
	8073	3231	0.009351	0.039209	0	0	0	0	0	0	1	-360	360;
	3231	13543	0.008622	0.0296	0	0	0	0	0	0	1	-360	360;
	2905	3709	0.008951	0.037289	0	0	0	0	0	0	1	-360	360;
	3709	6645	0.005369	0.02144	0	0	0	0	0	0	1	-360	360;
	9832	12087	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8073	12087	0.00608	0.025551	0	0	0	0	0	0	1	-360	360;
	8073	12087	0.00596	0.026049	0	0	0	0	0	0	1	-360	360;
	12087	2905	0.001369	0.004898	0	0	0	0	0	0	1	-360	360;
	12087	3438	0.00144	0.006031	0	0	0	0	0	0	1	-360	360;
	7060	4801	0.0012	0.00328	0	0	0	0	0	0	1	-360	360;
	12370	7060	0.00132	0.006271	0	0	0	0	0	0	1	-360	360;
	7060	12766	0.009089	0.040711	0	0	0	0	0	0	1	-360	360;
	12922	1825	5.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	4579	1825	0.000302	0.00132	0	0	0	0	0	0	1	-360	360;
	12117	4442	0.00368	0.016502	0	0	0	0	0	0	1	-360	360;
	2805	12117	0.000329	0.00124	0	0	0	0	0	0	1	-360	360;
	7405	2761	0.0012	0.002689	0	0	0	0	0	0	1	-360	360;
	4579	2761	0.000338	0.001289	0	0	0	0	0	0	1	-360	360;
	607	12922	0.000942	0.002151	0	0	0	0	0	0	1	-360	360;
	7728	3048	0.00316	0.010862	0	0	0	0	0	0	1	-360	360;
	7728	10104	0.003222	0.010618	0	0	0	0	0	0	1	-360	360;
	5355	3048	0.010169	0.037858	0	0	0	0	0	0	1	-360	360;
	2604	10104	0.00552	0.024129	0	0	0	0	0	0	1	-360	360;
	3040	6222	0.000449	0.001889	0	0	0	0	0	0	1	-360	360;
	6714	8783	0.00044	0.001929	0	0	0	0	0	0	1	-360	360;
	5615	7265	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7265	10893	0.000418	0.00192	0	0	0	0	0	0	1	-360	360;
	7265	10540	0.000409	0.00196	0	0	0	0	0	0	1	-360	360;
	12370	10893	0.00608	0.022431	0	0	0	0	0	0	1	-360	360;
	9171	7894	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6302	7894	0.00016	0.000751	0	0	0	0	0	0	1	-360	360;
	6302	12680	0.046622	0.173511	0	0	0	0	0	0	1	-360	360;
	2151	7894	0.000151	0.000698	0	0	0	0	0	0	1	-360	360;
	5239	2151	0.01464	0.052329	0	0	0	0	0	0	1	-360	360;
	9977	9399	0.006129	0.023751	0	0	0	0	0	0	1	-360	360;
	7234	9399	0.001049	0.003311	0	0	0	0	0	0	1	-360	360;
	12602	7809	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7809	4442	0.003271	0.012089	0	0	0	0	0	0	1	-360	360;
	7809	2680	0.002951	0.013311	0	0	0	0	0	0	1	-360	360;
	7809	7522	0.017489	0.082769	0	0	0	0	0	0	1	-360	360;
	7809	3433	0.006231	0.029498	0	0	0	0	0	0	1	-360	360;
	3987	7657	0	0.00028	0	0	0	0	0	0	1	-360	360;
	9316	12813	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2911	7657	0.00124	0.004458	0	0	0	0	0	0	1	-360	360;
	8551	12813	0.001169	0.001969	0	0	0	0	0	0	1	-360	360;
	3266	13076	0.000791	0.002929	0	0	0	0	0	0	1	-360	360;
	1540	13076	0.000911	0.003138	0	0	0	0	0	0	1	-360	360;
	2805	2680	0.004671	0.01672	0	0	0	0	0	0	1	-360	360;
	2081	3719	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3719	4300	0.00064	0.00188	0	0	0	0	0	0	1	-360	360;
	884	11210	0.00014	0.00061	0	0	0	0	0	0	1	-360	360;
	7007	8597	0.000698	0.0024	0	0	0	0	0	0	1	-360	360;
	7007	3542	0.000711	0.002351	0	0	0	0	0	0	1	-360	360;
	12370	8597	0.001858	0.006391	0	0	0	0	0	0	1	-360	360;
	7199	3542	0.014129	0.06328	0	0	0	0	0	0	1	-360	360;
	3263	7199	0.0002	0.00092	0	0	0	0	0	0	1	-360	360;
	12488	1023	4e-05	0.0004	0	0	0	0	0	0	1	-360	360;
	12488	6112	4e-05	0.00039	0	0	0	0	0	0	1	-360	360;
	7298	1023	0.00066	0.00705	0	0	0	0	0	0	1	-360	360;
	6837	431	0.003622	0.016209	0	0	0	0	0	0	1	-360	360;
	13355	6837	5.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	3716	6837	3.1e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	11771	491	0.00084	0.002031	0	0	0	0	0	0	1	-360	360;
	10460	9398	0.004511	0.01972	0	0	0	0	0	0	1	-360	360;
	10460	12676	0.00064	0.0028	0	0	0	0	0	0	1	-360	360;
	7176	11771	0.004422	0.020102	0	0	0	0	0	0	1	-360	360;
	11771	7002	0	0.00028	0	0	0	0	0	0	1	-360	360;
	10036	1069	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1069	4681	0.000822	0.003622	0	0	0	0	0	0	1	-360	360;
	1069	3358	0.000809	0.003689	0	0	0	0	0	0	1	-360	360;
	5176	9806	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3755	9806	0.001471	0.00528	0	0	0	0	0	0	1	-360	360;
	3755	9806	0.001502	0.005169	0	0	0	0	0	0	1	-360	360;
	3064	9806	0.000769	0.002858	0	0	0	0	0	0	1	-360	360;
	9806	3271	0.003698	0.013209	0	0	0	0	0	0	1	-360	360;
	9806	3040	0.00116	0.005271	0	0	0	0	0	0	1	-360	360;
	9806	7399	0.003378	0.006178	0	0	0	0	0	0	1	-360	360;
	9806	1105	0.000849	0.004009	0	0	0	0	0	0	1	-360	360;
	9806	242	0.000431	0.00112	0	0	0	0	0	0	1	-360	360;
	3883	8131	0.006778	0.002631	0	0	0	0	0	0	1	-360	360;
	2189	150	0.003569	0.00468	0	0	0	0	0	0	1	-360	360;
	2189	150	0.003178	0.005298	0	0	0	0	0	0	1	-360	360;
	8492	6249	0.00388	0.0174	0	0	0	0	0	0	1	-360	360;
	1078	6249	0.003138	0.011689	0	0	0	0	0	0	1	-360	360;
	8492	2426	0.0044	0.017049	0	0	0	0	0	0	1	-360	360;
	2426	11261	0.000791	0.002631	0	0	0	0	0	0	1	-360	360;
	7836	4236	0	0.000222	0	0	0	0	0	0	1	-360	360;
	678	10009	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3269	10009	0.017698	0.0584	0	0	0	0	0	0	1	-360	360;
	3269	10009	0.015769	0.066191	0	0	0	0	0	0	1	-360	360;
	6381	12293	0.00225	0.02425	0	0	0	0	0	0	1	-360	360;
	6381	12293	0.00161	0.02118	0	0	0	0	0	0	1	-360	360;
	1624	12293	0.00312	0.03268	0	0	0	0	0	0	1	-360	360;
	1624	12293	0.00305	0.0333	0	0	0	0	0	0	1	-360	360;
	9094	8073	0.000182	0.000831	0	0	0	0	0	0	1	-360	360;
	11486	9094	0.013649	0.054609	0	0	0	0	0	0	1	-360	360;
	13277	11485	0.007209	0.03252	0	0	0	0	0	0	1	-360	360;
	13063	13277	0	0.00028	0	0	0	0	0	0	1	-360	360;
	11485	8073	0.000169	0.000769	0	0	0	0	0	0	1	-360	360;
	3927	8073	0.011649	0.026778	0	0	0	0	0	0	1	-360	360;
	9398	3927	0.004049	0.01112	0	0	0	0	0	0	1	-360	360;
	9398	8073	0.007898	0.037338	0	0	0	0	0	0	1	-360	360;
	9398	8073	0.00772	0.034578	0	0	0	0	0	0	1	-360	360;
	8073	6891	0.042582	0.193609	0	0	0	0	0	0	1	-360	360;
	8073	12351	0.027591	0.130569	0	0	0	0	0	0	1	-360	360;
	8073	13088	0.018431	0.065898	0	0	0	0	0	0	1	-360	360;
	8073	38	0.02836	0.1134	0	0	0	0	0	0	1	-360	360;
	8073	1357	0.046142	0.15844	0	0	0	0	0	0	1	-360	360;
	11024	8073	0.011409	0.027311	0	0	0	0	0	0	1	-360	360;
	8073	13543	0.019271	0.063569	0	0	0	0	0	0	1	-360	360;
	8073	13543	0.02212	0.082329	0	0	0	0	0	0	1	-360	360;
	4270	7017	0.018249	0.04624	0	0	0	0	0	0	1	-360	360;
	973	7017	0.008929	0.01872	0	0	0	0	0	0	1	-360	360;
	973	7017	0.008742	0.017329	0	0	0	0	0	0	1	-360	360;
	2250	8662	0.001569	0.00412	0	0	0	0	0	0	1	-360	360;
	8662	11489	0.037409	0.106911	0	0	0	0	0	0	1	-360	360;
	2250	6186	0.01836	0.049671	0	0	0	0	0	0	1	-360	360;
	5976	7608	0.007791	0.037502	0	0	0	0	0	0	1	-360	360;
	5976	13318	0.014911	0.051218	0	0	0	0	0	0	1	-360	360;
	3512	7177	0.007071	0.0294	0	0	0	0	0	0	1	-360	360;
	3512	10876	0.010169	0.039378	0	0	0	0	0	0	1	-360	360;
	9398	7176	0	0.00028	0	0	0	0	0	0	1	-360	360;
	9398	9501	0.004449	0.00924	0	0	0	0	0	0	1	-360	360;
	5442	7774	0.002191	0.011142	0	0	0	0	0	0	1	-360	360;
	7774	11730	0.00244	0.011951	0	0	0	0	0	0	1	-360	360;
	9040	10529	0.004662	0.020871	0	0	0	0	0	0	1	-360	360;
	5442	10529	0.00528	0.020449	0	0	0	0	0	0	1	-360	360;
	8796	10529	0.0036	0.0144	0	0	0	0	0	0	1	-360	360;
	4091	5673	0.00212	0.008871	0	0	0	0	0	0	1	-360	360;
	8988	4091	0.001991	0.008271	0	0	0	0	0	0	1	-360	360;
	569	5673	0.002071	0.009049	0	0	0	0	0	0	1	-360	360;
	569	5790	0.002031	0.008098	0	0	0	0	0	0	1	-360	360;
	3188	3858	0.002631	0.00868	0	0	0	0	0	0	1	-360	360;
	4213	3858	0.00048	0.001769	0	0	0	0	0	0	1	-360	360;
	3188	8564	0.002338	0.009831	0	0	0	0	0	0	1	-360	360;
	10166	8564	0.006911	0.025742	0	0	0	0	0	0	1	-360	360;
	8190	3188	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3436	4989	0.005649	0.020209	0	0	0	0	0	0	1	-360	360;
	6728	4989	0.037822	0.075338	0	0	0	0	0	0	1	-360	360;
	913	3436	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3436	10380	0.005609	0.020871	0	0	0	0	0	0	1	-360	360;
	3436	11139	0.042569	0.11516	0	0	0	0	0	0	1	-360	360;
	2200	10380	0	0.000222	0	0	0	0	0	0	1	-360	360;
	12480	8339	0.00396	0.010458	0	0	0	0	0	0	1	-360	360;
	10857	8339	0.019671	0.088151	0	0	0	0	0	0	1	-360	360;
	11575	12674	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2568	7937	0.00912	0.03364	0	0	0	0	0	0	1	-360	360;
	7937	10729	0.010942	0.051369	0	0	0	0	0	0	1	-360	360;
	2826	12674	0.000182	0.000822	0	0	0	0	0	0	1	-360	360;
	2826	12680	0.023209	0.118049	0	0	0	0	0	0	1	-360	360;
	2568	12674	0.0002	0.000809	0	0	0	0	0	0	1	-360	360;
	3014	9100	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2194	11185	0.00076	0.00811	0	0	0	0	0	0	1	-360	360;
	11993	4073	5.8e-05	0.000231	0	0	0	0	0	0	1	-360	360;
	11139	11993	0.013471	0.053849	0	0	0	0	0	0	1	-360	360;
	10708	11139	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2338	2658	0.007489	0.03376	0	0	0	0	0	0	1	-360	360;
	16	2338	0.004271	0.016431	0	0	0	0	0	0	1	-360	360;
	8080	7465	0	0.000222	0	0	0	0	0	0	1	-360	360;
	12933	7465	0.013289	0.04384	0	0	0	0	0	0	1	-360	360;
	12933	7465	0.01184	0.049689	0	0	0	0	0	0	1	-360	360;
	12933	7465	0.011338	0.0508	0	0	0	0	0	0	1	-360	360;
	6618	7465	0.008502	0.035671	0	0	0	0	0	0	1	-360	360;
	7465	5264	0.02216	0.088622	0	0	0	0	0	0	1	-360	360;
	7465	7862	0.019911	0.101262	0	0	0	0	0	0	1	-360	360;
	7465	653	0.014551	0.068258	0	0	0	0	0	0	1	-360	360;
	7465	6493	0.014129	0.050529	0	0	0	0	0	0	1	-360	360;
	7465	10830	0.018022	0.08072	0	0	0	0	0	0	1	-360	360;
	758	11139	0.005382	0.02004	0	0	0	0	0	0	1	-360	360;
	13429	943	0.007769	0.034791	0	0	0	0	0	0	1	-360	360;
	943	10857	0.068351	0.30616	0	0	0	0	0	0	1	-360	360;
	2549	6902	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1979	9985	0	0.00028	0	0	0	0	0	0	1	-360	360;
	356	3269	0	0.00028	0	0	0	0	0	0	1	-360	360;
	105	642	0.00163	0.01819	0	0	0	0	0	0	1	-360	360;
	8575	105	0.00241	0.02622	0	0	0	0	0	0	1	-360	360;
	9846	105	0.00236	0.02672	0	0	0	0	0	0	1	-360	360;
	105	4111	0.00188	0.02029	0	0	0	0	0	0	1	-360	360;
	105	4111	0.00213	0.01988	0	0	0	0	0	0	1	-360	360;
	2110	5239	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1429	5239	0.009822	0.046449	0	0	0	0	0	0	1	-360	360;
	10848	5239	0.00344	0.01628	0	0	0	0	0	0	1	-360	360;
	5239	2466	0.00648	0.025102	0	0	0	0	0	0	1	-360	360;
	6907	1610	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1610	1588	0.02368	0.099338	0	0	0	0	0	0	1	-360	360;
	5040	1610	0.017169	0.061382	0	0	0	0	0	0	1	-360	360;
	2194	10290	0.00209	0.01796	0	0	0	0	0	0	1	-360	360;
	2194	10290	0.00213	0.01758	0	0	0	0	0	0	1	-360	360;
	7872	10290	0.0013	0.01389	0	0	0	0	0	0	1	-360	360;
	6546	10290	0.00147	0.01361	0	0	0	0	0	0	1	-360	360;
	884	6381	0.00052	0.00446	0	0	0	0	0	0	1	-360	360;
	884	6381	0.00053	0.00436	0	0	0	0	0	0	1	-360	360;
	4100	11338	0.014889	0.05544	0	0	0	0	0	0	1	-360	360;
	4100	11338	0.015178	0.054289	0	0	0	0	0	0	1	-360	360;
	3161	4100	0.003111	0.013569	0	0	0	0	0	0	1	-360	360;
	9248	4100	0.000791	0.00376	0	0	0	0	0	0	1	-360	360;
	4100	10950	0.002751	0.009809	0	0	0	0	0	0	1	-360	360;
	4100	13418	0.007129	0.024471	0	0	0	0	0	0	1	-360	360;
	4100	13038	0.017951	0.0668	0	0	0	0	0	0	1	-360	360;
	4100	7546	0.006178	0.02812	0	0	0	0	0	0	1	-360	360;
	4100	2406	0.006049	0.028649	0	0	0	0	0	0	1	-360	360;
	8853	7298	0.00396	0.03136	0	0	0	0	0	0	1	-360	360;
	8853	7298	0.00353	0.03554	0	0	0	0	0	0	1	-360	360;
	8853	7298	0.00387	0.0331	0	0	0	0	0	0	1	-360	360;
	8853	7298	0.00394	0.03239	0	0	0	0	0	0	1	-360	360;
	2812	1780	0.00225	0.01786	0	0	0	0	0	0	1	-360	360;
	1780	4404	0.00124	0.01101	0	0	0	0	0	0	1	-360	360;
	2812	6546	0.00201	0.02024	0	0	0	0	0	0	1	-360	360;
	2812	9129	0.00197	0.02063	0	0	0	0	0	0	1	-360	360;
	9129	4404	0.00126	0.01078	0	0	0	0	0	0	1	-360	360;
	2812	7872	0.00193	0.02103	0	0	0	0	0	0	1	-360	360;
	7208	3755	0.001209	0.004551	0	0	0	0	0	0	1	-360	360;
	3755	1866	0.000449	0.002018	0	0	0	0	0	0	1	-360	360;
	1515	3755	0	0.00028	0	0	0	0	0	0	1	-360	360;
	10166	5552	0.007049	0.0252	0	0	0	0	0	0	1	-360	360;
	11202	10166	0.001191	0.003222	0	0	0	0	0	0	1	-360	360;
	10166	1216	0.000511	0.001978	0	0	0	0	0	0	1	-360	360;
	10166	10297	0.001009	0.004782	0	0	0	0	0	0	1	-360	360;
	10166	4760	0.000991	0.004422	0	0	0	0	0	0	1	-360	360;
	13435	3708	0.000929	0.003098	0	0	0	0	0	0	1	-360	360;
	5544	13435	0.005858	0.0246	0	0	0	0	0	0	1	-360	360;
	2189	5564	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3883	5564	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7532	12942	0	0.000222	0	0	0	0	0	0	1	-360	360;
	12942	2788	0.002031	0.009222	0	0	0	0	0	0	1	-360	360;
	8689	12942	0.01036	0.045231	0	0	0	0	0	0	1	-360	360;
	12942	3626	0.026858	0.062822	0	0	0	0	0	0	1	-360	360;
	1159	2788	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2788	3818	0.021569	0.098089	0	0	0	0	0	0	1	-360	360;
	12471	5564	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6116	7702	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7702	13161	0.000418	0.001769	0	0	0	0	0	0	1	-360	360;
	7702	10729	0.000409	0.001809	0	0	0	0	0	0	1	-360	360;
	4334	3266	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8398	9903	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5953	6748	0.000569	0.00168	0	0	0	0	0	0	1	-360	360;
	5953	12924	0.000582	0.001649	0	0	0	0	0	0	1	-360	360;
	12924	3263	0.001858	0.006911	0	0	0	0	0	0	1	-360	360;
	13637	5953	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7380	413	0.0154	0.006022	0	0	0	0	0	0	1	-360	360;
	5083	8627	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8627	13452	8e-05	0.0004	0	0	0	0	0	0	1	-360	360;
	8627	8977	9.8e-05	0.000329	0	0	0	0	0	0	1	-360	360;
	3703	13452	0.023129	0.076289	0	0	0	0	0	0	1	-360	360;
	758	8977	0.011258	0.047249	0	0	0	0	0	0	1	-360	360;
	3703	9201	0.018822	0.072271	0	0	0	0	0	0	1	-360	360;
	3703	11383	0.002009	0.006649	0	0	0	0	0	0	1	-360	360;
	3703	6551	0.000942	0.00424	0	0	0	0	0	0	1	-360	360;
	3703	13445	0.017711	0.060809	0	0	0	0	0	0	1	-360	360;
	3703	13445	0.01804	0.059489	0	0	0	0	0	0	1	-360	360;
	8575	8853	0.0028	0.03055	0	0	0	0	0	0	1	-360	360;
	8575	8853	0.00274	0.03113	0	0	0	0	0	0	1	-360	360;
	12619	5104	0	0.000222	0	0	0	0	0	0	1	-360	360;
	12680	431	0.0002	0.000942	0	0	0	0	0	0	1	-360	360;
	5752	1704	0.004351	0.014929	0	0	0	0	0	0	1	-360	360;
	5752	12680	0.004831	0.017289	0	0	0	0	0	0	1	-360	360;
	7023	11094	0.014289	0.067609	0	0	0	0	0	0	1	-360	360;
	7023	11094	0.013978	0.0626	0	0	0	0	0	0	1	-360	360;
	5355	5991	0.004431	0.014609	0	0	0	0	0	0	1	-360	360;
	5355	6001	0.00664	0.025711	0	0	0	0	0	0	1	-360	360;
	5355	10389	0.001289	0.004422	0	0	0	0	0	0	1	-360	360;
	5355	2604	0.01448	0.065822	0	0	0	0	0	0	1	-360	360;
	5355	4813	0.004631	0.017938	0	0	0	0	0	0	1	-360	360;
	5355	4813	0.00492	0.01692	0	0	0	0	0	0	1	-360	360;
	10857	5104	0.027511	0.120182	0	0	0	0	0	0	1	-360	360;
	1904	5104	0.004	0.014898	0	0	0	0	0	0	1	-360	360;
	11261	1240	0.000351	0.001591	0	0	0	0	0	0	1	-360	360;
	10857	1240	0.018929	0.084791	0	0	0	0	0	0	1	-360	360;
	2678	1078	0.013502	0.063889	0	0	0	0	0	0	1	-360	360;
	2658	7823	0.01632	0.057871	0	0	0	0	0	0	1	-360	360;
	11476	3054	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3778	3054	0.01176	0.04704	0	0	0	0	0	0	1	-360	360;
	9767	11489	0	0.000222	0	0	0	0	0	0	1	-360	360;
	13355	11489	0.02892	0.057622	0	0	0	0	0	0	1	-360	360;
	3945	6853	0.001711	0.0078	0	0	0	0	0	0	1	-360	360;
	3977	6853	0.01224	0.042022	0	0	0	0	0	0	1	-360	360;
	9485	3117	0.007769	0.061249	0	0	0	0	0	0	1	-360	360;
	13248	9173	0.001391	0.0048	0	0	0	0	0	0	1	-360	360;
	6676	13248	0.004071	0.01988	0	0	0	0	0	0	1	-360	360;
	11511	9173	0.001418	0.004702	0	0	0	0	0	0	1	-360	360;
	11511	4465	0.012711	0.048809	0	0	0	0	0	0	1	-360	360;
	6556	4531	0	0.000222	0	0	0	0	0	0	1	-360	360;
	11981	9173	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4531	9173	0.00448	0.016698	0	0	0	0	0	0	1	-360	360;
	10144	6676	0.014689	0.052049	0	0	0	0	0	0	1	-360	360;
	4531	252	0.025631	0.084538	0	0	0	0	0	0	1	-360	360;
	13063	3769	0.005431	0.026551	0	0	0	0	0	0	1	-360	360;
	3769	6493	0.013151	0.058898	0	0	0	0	0	0	1	-360	360;
	7410	9398	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5345	1588	0.00036	0.001329	0	0	0	0	0	0	1	-360	360;
	8645	12391	0	0.00028	0	0	0	0	0	0	1	-360	360;
	12515	8645	0	0.00028	0	0	0	0	0	0	1	-360	360;
	12391	8405	0.00612	0.014671	0	0	0	0	0	0	1	-360	360;
	10596	12515	0.006942	0.014378	0	0	0	0	0	0	1	-360	360;
	12885	1564	4.9e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	6041	12885	0.002071	0.008289	0	0	0	0	0	0	1	-360	360;
	7651	1564	4.9e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	7651	10129	0.000729	0.003049	0	0	0	0	0	0	1	-360	360;
	3552	1564	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2947	13578	0.000409	0.00184	0	0	0	0	0	0	1	-360	360;
	4270	13578	0.0004	0.00188	0	0	0	0	0	0	1	-360	360;
	659	10054	0	0.000222	0	0	0	0	0	0	1	-360	360;
	10054	9962	0.003982	0.020249	0	0	0	0	0	0	1	-360	360;
	10054	12676	0.006498	0.031791	0	0	0	0	0	0	1	-360	360;
	13084	9371	0.000129	0.000609	0	0	0	0	0	0	1	-360	360;
	13084	895	0.00036	0.000729	0	0	0	0	0	0	1	-360	360;
	13084	8847	0.000129	0.000569	0	0	0	0	0	0	1	-360	360;
	1547	13084	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8847	791	0.006862	0.030942	0	0	0	0	0	0	1	-360	360;
	4032	6469	0.0016	0.00716	0	0	0	0	0	0	1	-360	360;
	4032	11899	0.001849	0.006542	0	0	0	0	0	0	1	-360	360;
	1111	6469	0.001809	0.007018	0	0	0	0	0	0	1	-360	360;
	1111	2851	0.004302	0.01588	0	0	0	0	0	0	1	-360	360;
	4737	6469	0	0.000222	0	0	0	0	0	0	1	-360	360;
	11338	10013	0.008831	0.036751	0	0	0	0	0	0	1	-360	360;
	11338	93	0.011031	0.045942	0	0	0	0	0	0	1	-360	360;
	9600	11338	0	0.000222	0	0	0	0	0	0	1	-360	360;
	38	11338	0.043418	0.124111	0	0	0	0	0	0	1	-360	360;
	5787	13161	0.011698	0.048702	0	0	0	0	0	0	1	-360	360;
	12933	6618	0.005058	0.02572	0	0	0	0	0	0	1	-360	360;
	5563	6618	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1938	8769	0	0.00028	0	0	0	0	0	0	1	-360	360;
	12950	12007	0.008969	0.032062	0	0	0	0	0	0	1	-360	360;
	5780	10263	0.000489	0.002369	0	0	0	0	0	0	1	-360	360;
	4579	10263	0.0028	0.0096	0	0	0	0	0	0	1	-360	360;
	5780	10495	0.00048	0.002191	0	0	0	0	0	0	1	-360	360;
	4579	10495	0.002849	0.0094	0	0	0	0	0	0	1	-360	360;
	6923	5780	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6110	12950	0	0.000222	0	0	0	0	0	0	1	-360	360;
	12022	12950	0.005138	0.023391	0	0	0	0	0	0	1	-360	360;
	9821	12950	0.005031	0.023831	0	0	0	0	0	0	1	-360	360;
	8781	12950	0.00492	0.022062	0	0	0	0	0	0	1	-360	360;
	12950	7225	0.012449	0.03368	0	0	0	0	0	0	1	-360	360;
	6289	7746	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7109	7746	0.005582	0.021622	0	0	0	0	0	0	1	-360	360;
	7746	3645	0.011769	0.052729	0	0	0	0	0	0	1	-360	360;
	2555	12898	0	0.00028	0	0	0	0	0	0	1	-360	360;
	12725	12898	0.005489	0.019618	0	0	0	0	0	0	1	-360	360;
	10926	12898	0.005591	0.0192	0	0	0	0	0	0	1	-360	360;
	10279	2173	0	0.00028	0	0	0	0	0	0	1	-360	360;
	10701	12022	0	0.00028	0	0	0	0	0	0	1	-360	360;
	9821	1446	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2173	1446	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3924	1525	0	0.000222	0	0	0	0	0	0	1	-360	360;
	551	3924	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1525	625	0	0.000222	0	0	0	0	0	0	1	-360	360;
	625	8781	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4562	10042	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7522	726	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4562	551	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5876	1742	0	0.000222	0	0	0	0	0	0	1	-360	360;
	472	5876	0.012969	0.046369	0	0	0	0	0	0	1	-360	360;
	1742	5432	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5432	7109	0	0.000222	0	0	0	0	0	0	1	-360	360;
	726	10701	0	0.000222	0	0	0	0	0	0	1	-360	360;
	11798	10939	0.003151	0.00128	0	0	0	0	0	0	1	-360	360;
	2194	463	0.00154	0.01649	0	0	0	0	0	0	1	-360	360;
	4579	8523	0.001298	0.00568	0	0	0	0	0	0	1	-360	360;
	4579	2613	0.0002	0.000871	0	0	0	0	0	0	1	-360	360;
	12972	11637	0	0.000222	0	0	0	0	0	0	1	-360	360;
	13429	11637	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	4327	11637	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	4327	12680	0.010071	0.039	0	0	0	0	0	0	1	-360	360;
	6345	7520	0.000218	0.000849	0	0	0	0	0	0	1	-360	360;
	2622	7520	0.000231	0.00084	0	0	0	0	0	0	1	-360	360;
	8120	7520	0.000231	0.000822	0	0	0	0	0	0	1	-360	360;
	9021	7520	0.00024	0.0008	0	0	0	0	0	0	1	-360	360;
	13102	7520	0.00024	0.000782	0	0	0	0	0	0	1	-360	360;
	7520	9943	0.000209	0.000889	0	0	0	0	0	0	1	-360	360;
	15	7520	0	0.00028	0	0	0	0	0	0	1	-360	360;
	4912	7520	0.000209	0.000911	0	0	0	0	0	0	1	-360	360;
	7006	2622	0.00784	0.025862	0	0	0	0	0	0	1	-360	360;
	8120	8542	0.001991	0.009391	0	0	0	0	0	0	1	-360	360;
	9021	7335	0.003591	0.01184	0	0	0	0	0	0	1	-360	360;
	11303	790	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2194	2850	0.00013	0.00077	0	0	0	0	0	0	1	-360	360;
	1523	2850	0.00013	0.00076	0	0	0	0	0	0	1	-360	360;
	1283	2850	0.0002	0.00127	0	0	0	0	0	0	1	-360	360;
	6010	1991	0.002449	0.006871	0	0	0	0	0	0	1	-360	360;
	3652	10760	0.00132	0.004929	0	0	0	0	0	0	1	-360	360;
	8628	3426	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3426	3626	0.036111	0.084471	0	0	0	0	0	0	1	-360	360;
	3426	5199	0.043129	0.093129	0	0	0	0	0	0	1	-360	360;
	8769	12458	0.001071	0.004129	0	0	0	0	0	0	1	-360	360;
	1671	12458	0.01016	0.036298	0	0	0	0	0	0	1	-360	360;
	1718	8769	0.010951	0.055689	0	0	0	0	0	0	1	-360	360;
	5736	10359	0.005831	0.022591	0	0	0	0	0	0	1	-360	360;
	139	10359	0.016542	0.075222	0	0	0	0	0	0	1	-360	360;
	10359	7759	0.015938	0.00716	0	0	0	0	0	0	1	-360	360;
	7909	3749	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3749	803	0.009591	0.02332	0	0	0	0	0	0	1	-360	360;
	7909	8405	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3818	3027	0.03828	0.10096	0	0	0	0	0	0	1	-360	360;
	139	3027	0.03036	0.086778	0	0	0	0	0	0	1	-360	360;
	4838	1824	0	0.00028	0	0	0	0	0	0	1	-360	360;
	11552	3161	0.002182	0.010329	0	0	0	0	0	0	1	-360	360;
	11552	6891	0.002129	0.009569	0	0	0	0	0	0	1	-360	360;
	8244	13340	0.0018	0.00668	0	0	0	0	0	0	1	-360	360;
	3727	13340	0.004569	0.01636	0	0	0	0	0	0	1	-360	360;
	8244	4588	0.001831	0.006542	0	0	0	0	0	0	1	-360	360;
	3727	6323	0.006338	0.026609	0	0	0	0	0	0	1	-360	360;
	9136	10857	0	0.000222	0	0	0	0	0	0	1	-360	360;
	10857	8862	0.016231	0.070902	0	0	0	0	0	0	1	-360	360;
	10857	12228	0.010138	0.046102	0	0	0	0	0	0	1	-360	360;
	8599	5040	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5040	13459	0.006431	0.028991	0	0	0	0	0	0	1	-360	360;
	3945	9086	0.01184	0.053022	0	0	0	0	0	0	1	-360	360;
	3945	3977	0.012529	0.044422	0	0	0	0	0	0	1	-360	360;
	11089	8210	0.00436	0.010689	0	0	0	0	0	0	1	-360	360;
	7359	11089	0.019529	0.055831	0	0	0	0	0	0	1	-360	360;
	4600	8210	0.004942	0.01048	0	0	0	0	0	0	1	-360	360;
	3157	8210	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1795	7852	0.006431	0.01764	0	0	0	0	0	0	1	-360	360;
	6775	1795	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	13173	7852	0.003431	0.013218	0	0	0	0	0	0	1	-360	360;
	8014	7852	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7852	8667	0.00708	0.014702	0	0	0	0	0	0	1	-360	360;
	4854	7852	0.065422	0.130311	0	0	0	0	0	0	1	-360	360;
	13173	2992	0.003049	0.013831	0	0	0	0	0	0	1	-360	360;
	9985	13173	0.011209	0.051009	0	0	0	0	0	0	1	-360	360;
	4591	13173	0	0.00028	0	0	0	0	0	0	1	-360	360;
	11769	13173	0.00504	0.024622	0	0	0	0	0	0	1	-360	360;
	3269	2992	0.009049	0.033671	0	0	0	0	0	0	1	-360	360;
	9011	2992	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6775	9716	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	7025	6775	0	0.00028	0	0	0	0	0	0	1	-360	360;
	9985	8667	0.007631	0.015569	0	0	0	0	0	0	1	-360	360;
	9985	3516	0.000822	0.00316	0	0	0	0	0	0	1	-360	360;
	9985	9672	0.017751	0.060929	0	0	0	0	0	0	1	-360	360;
	3269	9716	0.007418	0.016022	0	0	0	0	0	0	1	-360	360;
	3269	13358	0.007542	0.035689	0	0	0	0	0	0	1	-360	360;
	13216	689	0	0.000222	0	0	0	0	0	0	1	-360	360;
	689	1837	0.000338	0.001218	0	0	0	0	0	0	1	-360	360;
	7603	255	0	0.000222	0	0	0	0	0	0	1	-360	360;
	255	11434	0.000351	0.001191	0	0	0	0	0	0	1	-360	360;
	10145	1671	0.004662	0.016009	0	0	0	0	0	0	1	-360	360;
	10145	12370	0.017489	0.084191	0	0	0	0	0	0	1	-360	360;
	1671	37	0.003809	0.013111	0	0	0	0	0	0	1	-360	360;
	1671	8886	0.003489	0.013502	0	0	0	0	0	0	1	-360	360;
	5345	1671	0.004738	0.015658	0	0	0	0	0	0	1	-360	360;
	11548	1671	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1671	3692	0.010351	0.035529	0	0	0	0	0	0	1	-360	360;
	1671	4098	0.00388	0.012822	0	0	0	0	0	0	1	-360	360;
	1837	37	0.003729	0.013218	0	0	0	0	0	0	1	-360	360;
	8886	10370	0.002418	0.008298	0	0	0	0	0	0	1	-360	360;
	11434	10370	0.001538	0.007822	0	0	0	0	0	0	1	-360	360;
	1824	1480	0.005369	0.010702	0	0	0	0	0	0	1	-360	360;
	4835	4568	0.00088	0.0075	0	0	0	0	0	0	1	-360	360;
	4835	4568	0.0009	0.00734	0	0	0	0	0	0	1	-360	360;
	11826	10596	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2239	12284	0	0.00028	0	0	0	0	0	0	1	-360	360;
	11826	12284	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8670	12284	0.018831	0.072951	0	0	0	0	0	0	1	-360	360;
	9111	12284	0.012351	0.042409	0	0	0	0	0	0	1	-360	360;
	7359	5736	0.001938	0.008698	0	0	0	0	0	0	1	-360	360;
	2915	7359	0	0.00028	0	0	0	0	0	0	1	-360	360;
	9423	12804	0.017871	0.047151	0	0	0	0	0	0	1	-360	360;
	12804	973	0.00076	0.002098	0	0	0	0	0	0	1	-360	360;
	7823	7380	0	0.000222	0	0	0	0	0	0	1	-360	360;
	11730	3692	0.005249	0.024631	0	0	0	0	0	0	1	-360	360;
	13364	11730	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2463	5869	0	0.00028	0	0	0	0	0	0	1	-360	360;
	12546	5869	0.00016	0.000502	0	0	0	0	0	0	1	-360	360;
	5570	12546	0.008658	0.037809	0	0	0	0	0	0	1	-360	360;
	8862	5869	0.000142	0.00056	0	0	0	0	0	0	1	-360	360;
	9877	9423	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	9877	8171	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	10950	4561	0.0002	0.00092	0	0	0	0	0	0	1	-360	360;
	4561	1858	0.002542	0.010649	0	0	0	0	0	0	1	-360	360;
	9248	38	0.054	0.121351	0	0	0	0	0	0	1	-360	360;
	9645	10950	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5045	6287	0.001969	0.004271	0	0	0	0	0	0	1	-360	360;
	5045	4854	0.009111	0.025009	0	0	0	0	0	0	1	-360	360;
	2472	5045	0.027502	0.074369	0	0	0	0	0	0	1	-360	360;
	11562	9894	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9894	5264	0.0002	0.000942	0	0	0	0	0	0	1	-360	360;
	9894	9962	0.0002	0.000871	0	0	0	0	0	0	1	-360	360;
	2830	13105	0.00052	0.00136	0	0	0	0	0	0	1	-360	360;
	4592	2830	0.020471	0.07928	0	0	0	0	0	0	1	-360	360;
	7097	1198	3.1e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	9090	1198	0.015609	0.076289	0	0	0	0	0	0	1	-360	360;
	7097	11769	3.1e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	3398	7097	0	0.000222	0	0	0	0	0	0	1	-360	360;
	12426	10685	0.008809	0.041342	0	0	0	0	0	0	1	-360	360;
	12867	10685	0.011698	0.017782	0	0	0	0	0	0	1	-360	360;
	5954	12426	0.029769	0.069649	0	0	0	0	0	0	1	-360	360;
	13642	8302	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8670	8302	0.006738	0.024089	0	0	0	0	0	0	1	-360	360;
	13642	11958	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1947	13642	0	0.000222	0	0	0	0	0	0	1	-360	360;
	12710	2	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9010	3652	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2	3652	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1244	2	0.001062	0.003649	0	0	0	0	0	0	1	-360	360;
	3818	2545	0.00204	0.009129	0	0	0	0	0	0	1	-360	360;
	3818	2545	0.00232	0.008951	0	0	0	0	0	0	1	-360	360;
	3818	11951	0.00152	0.006911	0	0	0	0	0	0	1	-360	360;
	3818	11951	0.001489	0.00704	0	0	0	0	0	0	1	-360	360;
	3818	11951	0.001462	0.00652	0	0	0	0	0	0	1	-360	360;
	3818	12944	0.001391	0.003609	0	0	0	0	0	0	1	-360	360;
	3818	12233	0.001658	0.003311	0	0	0	0	0	0	1	-360	360;
	3818	11995	0.000991	0.004329	0	0	0	0	0	0	1	-360	360;
	3818	3851	0.010369	0.038569	0	0	0	0	0	0	1	-360	360;
	318	2545	0	0.000222	0	0	0	0	0	0	1	-360	360;
	318	8789	0.026809	0.126889	0	0	0	0	0	0	1	-360	360;
	12755	2545	0	0.000222	0	0	0	0	0	0	1	-360	360;
	12755	12479	0.026218	0.117489	0	0	0	0	0	0	1	-360	360;
	11951	2545	0.00064	0.002662	0	0	0	0	0	0	1	-360	360;
	11951	2545	0.000622	0.002711	0	0	0	0	0	0	1	-360	360;
	11951	126	0.000609	0.002271	0	0	0	0	0	0	1	-360	360;
	6818	12944	0.002098	0.005702	0	0	0	0	0	0	1	-360	360;
	5791	12233	0.002582	0.006031	0	0	0	0	0	0	1	-360	360;
	6022	5791	0.019458	0.055618	0	0	0	0	0	0	1	-360	360;
	12956	11995	0.00092	0.003289	0	0	0	0	0	0	1	-360	360;
	12956	126	0.000942	0.003222	0	0	0	0	0	0	1	-360	360;
	12120	12956	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6818	8949	0.03868	0.1062	0	0	0	0	0	0	1	-360	360;
	8465	13561	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3433	13561	0.000689	0.00312	0	0	0	0	0	0	1	-360	360;
	472	13561	0.0008	0.002991	0	0	0	0	0	0	1	-360	360;
	8794	12379	0.000511	0.001191	0	0	0	0	0	0	1	-360	360;
	13094	7635	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3626	3979	0.01216	0.048609	0	0	0	0	0	0	1	-360	360;
	5665	3851	0.000769	0.00348	0	0	0	0	0	0	1	-360	360;
	5665	8689	0.000871	0.003409	0	0	0	0	0	0	1	-360	360;
	9201	13445	0.007071	0.03316	0	0	0	0	0	0	1	-360	360;
	791	12740	0.009609	0.038418	0	0	0	0	0	0	1	-360	360;
	1101	791	0	0.00028	0	0	0	0	0	0	1	-360	360;
	12836	12740	0.00724	0.014951	0	0	0	0	0	0	1	-360	360;
	4720	6323	0	0.00028	0	0	0	0	0	0	1	-360	360;
	11747	6323	0.016071	0.067418	0	0	0	0	0	0	1	-360	360;
	11747	6323	0.015738	0.068751	0	0	0	0	0	0	1	-360	360;
	6323	10133	0.017431	0.037622	0	0	0	0	0	0	1	-360	360;
	5269	2185	0.001031	0.003378	0	0	0	0	0	0	1	-360	360;
	2185	11780	0.000169	0.00068	0	0	0	0	0	0	1	-360	360;
	2406	5269	0.006249	0.023262	0	0	0	0	0	0	1	-360	360;
	13418	9977	8.9e-05	0.000471	0	0	0	0	0	0	1	-360	360;
	4308	3778	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6236	3778	0.01564	0.065089	0	0	0	0	0	0	1	-360	360;
	3778	7102	0.018938	0.085391	0	0	0	0	0	0	1	-360	360;
	3778	2100	0.016511	0.044662	0	0	0	0	0	0	1	-360	360;
	415	758	0	0.000222	0	0	0	0	0	0	1	-360	360;
	758	12836	0.002689	0.012938	0	0	0	0	0	0	1	-360	360;
	7420	4115	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1171	7420	0.007649	0.019631	0	0	0	0	0	0	1	-360	360;
	9127	540	0.008849	0.02096	0	0	0	0	0	0	1	-360	360;
	4115	9127	0	0.000222	0	0	0	0	0	0	1	-360	360;
	540	7091	0	0.00028	0	0	0	0	0	0	1	-360	360;
	16	3402	0.00232	0.00928	0	0	0	0	0	0	1	-360	360;
	3859	2169	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9672	8528	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4957	3155	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8568	4957	0.0088	0.034102	0	0	0	0	0	0	1	-360	360;
	10822	7204	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7445	4731	0	0.00028	0	0	0	0	0	0	1	-360	360;
	12005	6010	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7882	2295	0	0.00028	0	0	0	0	0	0	1	-360	360;
	11531	11277	0	0.00028	0	0	0	0	0	0	1	-360	360;
	11293	7684	0	0.00028	0	0	0	0	0	0	1	-360	360;
	9940	12239	0	0.00028	0	0	0	0	0	0	1	-360	360;
	12239	3590	0.002542	0.00392	0	0	0	0	0	0	1	-360	360;
	2911	6701	0.00296	0.013462	0	0	0	0	0	0	1	-360	360;
	11664	2911	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2911	4489	0.002489	0.003991	0	0	0	0	0	0	1	-360	360;
	1110	7413	0	0.00028	0	0	0	0	0	0	1	-360	360;
	11728	7413	0.000289	0.000742	0	0	0	0	0	0	1	-360	360;
	11038	7469	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3708	7469	0.000302	0.000729	0	0	0	0	0	0	1	-360	360;
	7469	10191	0.000662	0.002111	0	0	0	0	0	0	1	-360	360;
	6468	1715	0	0.00028	0	0	0	0	0	0	1	-360	360;
	9560	2295	0.000822	0.002871	0	0	0	0	0	0	1	-360	360;
	2295	12853	0.006391	0.028809	0	0	0	0	0	0	1	-360	360;
	4046	11277	0.000831	0.002809	0	0	0	0	0	0	1	-360	360;
	9952	9560	0.014791	0.005969	0	0	0	0	0	0	1	-360	360;
	6334	4046	0.01508	0.005849	0	0	0	0	0	0	1	-360	360;
	6256	7684	0.011302	0.050578	0	0	0	0	0	0	1	-360	360;
	8794	7684	0.01704	0.033938	0	0	0	0	0	0	1	-360	360;
	12444	7862	0.001262	0.00532	0	0	0	0	0	0	1	-360	360;
	12444	2851	0.00124	0.005431	0	0	0	0	0	0	1	-360	360;
	9212	529	0	0.00028	0	0	0	0	0	0	1	-360	360;
	529	7366	0.001538	0.004609	0	0	0	0	0	0	1	-360	360;
	1841	12936	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3863	8761	0	0.00028	0	0	0	0	0	0	1	-360	360;
	12214	8761	0.00244	0.004071	0	0	0	0	0	0	1	-360	360;
	3150	7500	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8100	7500	0.002649	0.003631	0	0	0	0	0	0	1	-360	360;
	7277	13318	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3630	1941	0	0.000222	0	0	0	0	0	0	1	-360	360;
	12190	8523	0.000849	0.001769	0	0	0	0	0	0	1	-360	360;
	8523	11559	0.000951	0.003991	0	0	0	0	0	0	1	-360	360;
	6881	7091	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6881	11618	0.0034	0.00872	0	0	0	0	0	0	1	-360	360;
	10016	11747	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1715	11291	0.001978	0.007671	0	0	0	0	0	0	1	-360	360;
	1715	1270	0.002018	0.00752	0	0	0	0	0	0	1	-360	360;
	7130	6236	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6236	1357	0.018351	0.070502	0	0	0	0	0	0	1	-360	360;
	12370	6902	0.0028	0.014209	0	0	0	0	0	0	1	-360	360;
	10801	13319	0	0.000222	0	0	0	0	0	0	1	-360	360;
	10823	2277	0	0.000222	0	0	0	0	0	0	1	-360	360;
	10129	2277	0.02624	0.096858	0	0	0	0	0	0	1	-360	360;
	662	8650	5.8e-05	0.00024	0	0	0	0	0	0	1	-360	360;
	662	10389	5.8e-05	0.000231	0	0	0	0	0	0	1	-360	360;
	9555	6001	0.0012	0.005809	0	0	0	0	0	0	1	-360	360;
	9555	11034	8.9e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	9555	9329	8e-05	0.00032	0	0	0	0	0	0	1	-360	360;
	11034	4813	0.006889	0.032631	0	0	0	0	0	0	1	-360	360;
	9329	4813	0.006738	0.030209	0	0	0	0	0	0	1	-360	360;
	5529	4213	0.001218	0.004031	0	0	0	0	0	0	1	-360	360;
	4213	8794	0.005751	0.012929	0	0	0	0	0	0	1	-360	360;
	232	5552	0.000462	0.001791	0	0	0	0	0	0	1	-360	360;
	232	9100	0.00352	0.007302	0	0	0	0	0	0	1	-360	360;
	6800	7208	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6576	3064	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1871	8225	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8225	1402	0.001289	0.005022	0	0	0	0	0	0	1	-360	360;
	4192	11865	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6198	12853	0.00036	0.0016	0	0	0	0	0	0	1	-360	360;
	6198	7427	0.000409	0.001569	0	0	0	0	0	0	1	-360	360;
	10477	6198	0	0.00028	0	0	0	0	0	0	1	-360	360;
	12026	6198	0.001658	0.004511	0	0	0	0	0	0	1	-360	360;
	8670	3741	0.00176	0.007911	0	0	0	0	0	0	1	-360	360;
	1402	8670	0.00172	0.008058	0	0	0	0	0	0	1	-360	360;
	9095	8670	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2051	12853	0.0022	0.00852	0	0	0	0	0	0	1	-360	360;
	11044	7427	0.000111	0.000489	0	0	0	0	0	0	1	-360	360;
	11044	12876	0.00396	0.013058	0	0	0	0	0	0	1	-360	360;
	3741	13013	0.001138	0.002529	0	0	0	0	0	0	1	-360	360;
	12767	1347	0	0.00028	0	0	0	0	0	0	1	-360	360;
	13284	12767	0	0.000222	0	0	0	0	0	0	1	-360	360;
	12767	3829	0.001329	0.005978	0	0	0	0	0	0	1	-360	360;
	7837	11486	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7980	9562	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9562	11618	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7980	4807	0.010191	0.028329	0	0	0	0	0	0	1	-360	360;
	2614	13102	0.002391	0.00788	0	0	0	0	0	0	1	-360	360;
	1513	2614	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8542	2614	0.002249	0.008351	0	0	0	0	0	0	1	-360	360;
	8927	4374	0.000302	0.000969	0	0	0	0	0	0	1	-360	360;
	3505	1933	0.000271	0.001102	0	0	0	0	0	0	1	-360	360;
	9943	8927	0.000391	0.001738	0	0	0	0	0	0	1	-360	360;
	3505	4912	0.002289	0.008178	0	0	0	0	0	0	1	-360	360;
	10262	8542	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8542	7006	0.00112	0.00504	0	0	0	0	0	0	1	-360	360;
	8542	7006	0.001271	0.004942	0	0	0	0	0	0	1	-360	360;
	4758	7006	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7006	10590	0.000418	0.00192	0	0	0	0	0	0	1	-360	360;
	7006	6571	0.000409	0.001951	0	0	0	0	0	0	1	-360	360;
	7006	139	0.03228	0.106458	0	0	0	0	0	0	1	-360	360;
	7006	139	0.02876	0.120649	0	0	0	0	0	0	1	-360	360;
	6571	139	0.05888	0.117289	0	0	0	0	0	0	1	-360	360;
	10129	11862	0.00324	0.01584	0	0	0	0	0	0	1	-360	360;
	1811	11862	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5491	11301	0.00176	0.004449	0	0	0	0	0	0	1	-360	360;
	12113	1429	0.009929	0.046969	0	0	0	0	0	0	1	-360	360;
	12113	3053	0.009711	0.043489	0	0	0	0	0	0	1	-360	360;
	1904	3053	0.005031	0.02252	0	0	0	0	0	0	1	-360	360;
	1888	7366	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8443	9410	0	0.000222	0	0	0	0	0	0	1	-360	360;
	10832	9410	0.000462	0.00108	0	0	0	0	0	0	1	-360	360;
	11929	7366	0.000471	0.001049	0	0	0	0	0	0	1	-360	360;
	12026	2051	0.00104	0.003729	0	0	0	0	0	0	1	-360	360;
	12351	7620	0.002329	0.008	0	0	0	0	0	0	1	-360	360;
	13038	7620	0.002378	0.007831	0	0	0	0	0	0	1	-360	360;
	2864	7620	0	0.000222	0	0	0	0	0	0	1	-360	360;
	878	1904	0.006711	0.026018	0	0	0	0	0	0	1	-360	360;
	1904	2466	0.009298	0.04232	0	0	0	0	0	0	1	-360	360;
	13543	13088	0.004751	0.01632	0	0	0	0	0	0	1	-360	360;
	3976	2701	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2701	4465	0.000338	0.001151	0	0	0	0	0	0	1	-360	360;
	2701	9090	0.000338	0.00112	0	0	0	0	0	0	1	-360	360;
	8899	11202	0	0.00028	0	0	0	0	0	0	1	-360	360;
	11094	690	0.00484	0.023329	0	0	0	0	0	0	1	-360	360;
	7525	9952	0.004222	0.017751	0	0	0	0	0	0	1	-360	360;
	1650	6334	0.004138	0.018102	0	0	0	0	0	0	1	-360	360;
	12224	4834	0	0.000222	0	0	0	0	0	0	1	-360	360;
	12224	1737	0.000622	0.002418	0	0	0	0	0	0	1	-360	360;
	4834	923	0.000631	0.002369	0	0	0	0	0	0	1	-360	360;
	11383	13114	0.00012	0.00032	0	0	0	0	0	0	1	-360	360;
	11844	7682	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1581	5954	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3218	4489	0	0.00028	0	0	0	0	0	0	1	-360	360;
	12263	3590	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8591	914	0	0.00028	0	0	0	0	0	0	1	-360	360;
	4976	10890	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3241	12680	0.001409	0.00464	0	0	0	0	0	0	1	-360	360;
	2472	3241	0.003098	0.014911	0	0	0	0	0	0	1	-360	360;
	12680	3716	0.004662	0.015369	0	0	0	0	0	0	1	-360	360;
	173	13579	9.8e-05	0.000369	0	0	0	0	0	0	1	-360	360;
	12680	173	0.001031	0.003658	0	0	0	0	0	0	1	-360	360;
	4813	8368	0.001049	0.003591	0	0	0	0	0	0	1	-360	360;
	8988	5436	0.003209	0.012338	0	0	0	0	0	0	1	-360	360;
	4098	8988	0.003311	0.012689	0	0	0	0	0	0	1	-360	360;
	11744	3626	0	0.00028	0	0	0	0	0	0	1	-360	360;
	139	554	0.003871	0.010458	0	0	0	0	0	0	1	-360	360;
	12479	139	0.000218	0.000849	0	0	0	0	0	0	1	-360	360;
	8789	139	0.000231	0.00084	0	0	0	0	0	0	1	-360	360;
	13553	139	0	0.00028	0	0	0	0	0	0	1	-360	360;
	554	4079	0.000231	0.000822	0	0	0	0	0	0	1	-360	360;
	5199	4079	0.00024	0.0008	0	0	0	0	0	0	1	-360	360;
	3373	4079	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8235	1718	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3660	8551	0.000151	0.00036	0	0	0	0	0	0	1	-360	360;
	3660	12214	8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	1540	8100	8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	1971	3660	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3660	1540	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3175	1540	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2722	12881	0.001711	0.003058	0	0	0	0	0	0	1	-360	360;
	11559	2722	0.003369	0.008889	0	0	0	0	0	0	1	-360	360;
	2495	12881	0.002009	0.005311	0	0	0	0	0	0	1	-360	360;
	4250	2495	0	0.000222	0	0	0	0	0	0	1	-360	360;
	11899	1213	0.002551	0.011578	0	0	0	0	0	0	1	-360	360;
	6041	1213	0.002489	0.0118	0	0	0	0	0	0	1	-360	360;
	10623	12370	0	0.00028	0	0	0	0	0	0	1	-360	360;
	12370	5790	0.010009	0.04892	0	0	0	0	0	0	1	-360	360;
	7234	5426	0.000182	0.000671	0	0	0	0	0	0	1	-360	360;
	11318	10860	0.006169	0.02036	0	0	0	0	0	0	1	-360	360;
	10860	2100	0.007369	0.0322	0	0	0	0	0	0	1	-360	360;
	9581	11318	0.005498	0.023071	0	0	0	0	0	0	1	-360	360;
	11780	9581	0.004809	0.021671	0	0	0	0	0	0	1	-360	360;
	10111	11318	0	0.00028	0	0	0	0	0	0	1	-360	360;
	11318	1357	0.016169	0.06784	0	0	0	0	0	0	1	-360	360;
	8889	4854	0	0.000222	0	0	0	0	0	0	1	-360	360;
	13104	38	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7615	38	0.01124	0.039822	0	0	0	0	0	0	1	-360	360;
	10144	252	0.001169	0.00532	0	0	0	0	0	0	1	-360	360;
	690	252	0.001138	0.005418	0	0	0	0	0	0	1	-360	360;
	12628	5098	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9715	7615	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3271	11882	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7027	8674	0.00212	0.008871	0	0	0	0	0	0	1	-360	360;
	3040	7027	0.002289	0.008902	0	0	0	0	0	0	1	-360	360;
	8674	12510	0.00024	0.000782	0	0	0	0	0	0	1	-360	360;
	4592	12510	0.000209	0.000889	0	0	0	0	0	0	1	-360	360;
	8248	11882	0	0.000222	0	0	0	0	0	0	1	-360	360;
	143	11882	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2930	12510	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2491	6668	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6668	7855	0.001209	0.005529	0	0	0	0	0	0	1	-360	360;
	6668	10133	0.001191	0.00564	0	0	0	0	0	0	1	-360	360;
	4096	7855	0.014982	0.0428	0	0	0	0	0	0	1	-360	360;
	1480	5759	0.012849	0.022929	0	0	0	0	0	0	1	-360	360;
	5759	12964	0.000431	0.000929	0	0	0	0	0	0	1	-360	360;
	12867	12964	0.000849	0.003831	0	0	0	0	0	0	1	-360	360;
	11061	12964	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7617	4496	0.000231	0.000769	0	0	0	0	0	0	1	-360	360;
	7617	6256	0.0002	0.000871	0	0	0	0	0	0	1	-360	360;
	10851	7617	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5983	878	0.004049	0.01844	0	0	0	0	0	0	1	-360	360;
	5983	5570	0.003969	0.018791	0	0	0	0	0	0	1	-360	360;
	7224	5983	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8949	578	0.002271	0.004258	0	0	0	0	0	0	1	-360	360;
	4096	5244	0.012449	0.0248	0	0	0	0	0	0	1	-360	360;
	5244	578	0.002311	0.004169	0	0	0	0	0	0	1	-360	360;
	5907	4096	0	0.000222	0	0	0	0	0	0	1	-360	360;
	10543	578	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6961	4407	0	0.000222	0	0	0	0	0	0	1	-360	360;
	12876	4407	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9111	4407	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3347	12228	0.002071	0.009049	0	0	0	0	0	0	1	-360	360;
	3347	347	0.002031	0.009222	0	0	0	0	0	0	1	-360	360;
	13519	3347	0	0.000222	0	0	0	0	0	0	1	-360	360;
	13368	1357	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2100	1357	0.015511	0.06772	0	0	0	0	0	0	1	-360	360;
	5582	13358	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	5582	8568	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	35	5582	0	0.000222	0	0	0	0	0	0	1	-360	360;
	10939	2398	0.0052	0.013898	0	0	0	0	0	0	1	-360	360;
	13459	6126	0.009071	0.032151	0	0	0	0	0	0	1	-360	360;
	9678	13459	0	0.00028	0	0	0	0	0	0	1	-360	360;
	803	11883	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2398	10163	0	0.000222	0	0	0	0	0	0	1	-360	360;
	11883	1171	0	0.000222	0	0	0	0	0	0	1	-360	360;
	11883	4807	0	0.000222	0	0	0	0	0	0	1	-360	360;
	11883	10163	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7803	11024	0.042889	0.088942	0	0	0	0	0	0	1	-360	360;
	1105	8783	0.000609	0.001618	0	0	0	0	0	0	1	-360	360;
	653	11486	0.011289	0.057409	0	0	0	0	0	0	1	-360	360;
	5920	1223	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7546	5544	8.9e-05	0.000449	0	0	0	0	0	0	1	-360	360;
	12111	7002	0.000169	0.000662	0	0	0	0	0	0	1	-360	360;
	3295	7399	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1858	10152	9.8e-05	0.000511	0	0	0	0	0	0	1	-360	360;
	12753	10152	0.006338	0.022662	0	0	0	0	0	0	1	-360	360;
	1281	10297	0.000809	0.001729	0	0	0	0	0	0	1	-360	360;
	4760	1737	0.00204	0.007898	0	0	0	0	0	0	1	-360	360;
	7335	13283	0.00188	0.00516	0	0	0	0	0	0	1	-360	360;
	12753	11728	0.000831	0.003511	0	0	0	0	0	0	1	-360	360;
	1866	242	0.00064	0.002871	0	0	0	0	0	0	1	-360	360;
	10191	5426	0.000182	0.00064	0	0	0	0	0	0	1	-360	360;
	10191	7102	0.000191	0.000631	0	0	0	0	0	0	1	-360	360;
	5182	10191	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2993	463	0.00043	0.00337	0	0	0	0	0	0	1	-360	360;
	2993	10415	0.00038	0.00383	0	0	0	0	0	0	1	-360	360;
	12766	7839	0.000431	0.00112	0	0	0	0	0	0	1	-360	360;
	2613	12766	0.006458	0.021302	0	0	0	0	0	0	1	-360	360;
	11620	3977	0	0.00028	0	0	0	0	0	0	1	-360	360;
	10848	5787	0.000249	0.001138	0	0	0	0	0	0	1	-360	360;
	6489	10848	0	0.000222	0	0	0	0	0	0	1	-360	360;
	10848	2466	0.002449	0.011169	0	0	0	0	0	0	1	-360	360;
	6251	2466	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7661	6022	0	0.000222	0	0	0	0	0	0	1	-360	360;
	11150	12955	0.000271	0.001089	0	0	0	0	0	0	1	-360	360;
	4766	5790	0	0.00028	0	0	0	0	0	0	1	-360	360;
	461	2100	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1968	1244	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4890	12026	0	0.000222	0	0	0	0	0	0	1	-360	360;
	10038	6493	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1132	3645	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5954	636	0.006729	0.031862	0	0	0	0	0	0	1	-360	360;
	5954	8083	0.006591	0.029498	0	0	0	0	0	0	1	-360	360;
	10890	13013	0.000258	0.000711	0	0	0	0	0	0	1	-360	360;
	7759	9485	0.020311	0.045649	0	0	0	0	0	0	1	-360	360;
	12111	10830	0.004102	0.014089	0	0	0	0	0	0	1	-360	360;
	10830	2457	0	0.000222	0	0	0	0	0	0	1	-360	360;
	11985	4283	0.038777	0.100653	0	0	0	0	0	0	1	-360	360;
	12107	11985	0.026711	0.08357	0	0	0	0	0	0	1	-360	360;
	691	12130	0.062719	0.144537	0	0	0	0	0	0	1	-360	360;
	7433	691	0.045033	0.112364	0	0	0	0	0	0	1	-360	360;
	6583	11209	0.011659	0.070599	0	0	0	0	0	0	1	-360	360;
	6583	7347	0.00968	0.04149	0	0	0	0	0	0	1	-360	360;
	6583	7347	0.00862	0.047021	0	0	0	0	0	0	1	-360	360;
	11907	12130	0.077471	0.185529	0	0	0	0	0	0	1	-360	360;
	5481	12130	0.045157	0.126248	0	0	0	0	0	0	1	-360	360;
	4863	12130	0.01119	0.043099	0	0	0	0	0	0	1	-360	360;
	4863	12130	0.01095	0.043909	0	0	0	0	0	0	1	-360	360;
	4016	11907	0.03681	0.090496	0	0	0	0	0	0	1	-360	360;
	4586	9093	0.056174	0.129223	0	0	0	0	0	0	1	-360	360;
	5635	4586	0.029256	0.116322	0	0	0	0	0	0	1	-360	360;
	8013	12639	0.061347	0.097917	0	0	0	0	0	0	1	-360	360;
	12639	13176	0.088719	0.24762	0	0	0	0	0	0	1	-360	360;
	11885	410	0	0.000413	0	0	0	0	0	0	1	-360	360;
	11885	8262	0.004612	0.01738	0	0	0	0	0	0	1	-360	360;
	11885	8355	0.008364	0.02695	0	0	0	0	0	0	1	-360	360;
	11885	8355	0.014992	0.043636	0	0	0	0	0	0	1	-360	360;
	12257	410	0.003033	0.012851	0	0	0	0	0	0	1	-360	360;
	12257	410	0.002967	0.011901	0	0	0	0	0	0	1	-360	360;
	410	9385	0.015256	0.042686	0	0	0	0	0	0	1	-360	360;
	2751	410	0.004397	0.025653	0	0	0	0	0	0	1	-360	360;
	4915	5800	0.0027	0.012961	0	0	0	0	0	0	1	-360	360;
	7347	5800	0.038169	0.16812	0	0	0	0	0	0	1	-360	360;
	5800	8006	0.015479	0.073409	0	0	0	0	0	0	1	-360	360;
	9650	5066	0.027868	0.087843	0	0	0	0	0	0	1	-360	360;
	7363	9650	0.013934	0.040479	0	0	0	0	0	0	1	-360	360;
	2546	5066	0.019231	0.076471	0	0	0	0	0	0	1	-360	360;
	3274	13278	0.013421	0.037256	0	0	0	0	0	0	1	-360	360;
	8064	3274	0.022438	0.046851	0	0	0	0	0	0	1	-360	360;
	12188	723	0.014562	0.045314	0	0	0	0	0	0	1	-360	360;
	723	4016	0.000826	0.002207	0	0	0	0	0	0	1	-360	360;
	2720	13255	0.040628	0.113562	0	0	0	0	0	0	1	-360	360;
	8009	2720	0.008058	0.024438	0	0	0	0	0	0	1	-360	360;
	2720	8553	0.022471	0.041083	0	0	0	0	0	0	1	-360	360;
	6439	12663	0.047322	0.096669	0	0	0	0	0	0	1	-360	360;
	12663	12107	0.01914	0.07614	0	0	0	0	0	0	1	-360	360;
	1067	8828	0.007802	0.01924	0	0	0	0	0	0	1	-360	360;
	1067	8828	0.005397	0.015554	0	0	0	0	0	0	1	-360	360;
	8828	84	0.024719	0.040702	0	0	0	0	0	0	1	-360	360;
	8828	84	0.02419	0.041471	0	0	0	0	0	0	1	-360	360;
	6235	3825	0.039603	0.109529	0	0	0	0	0	0	1	-360	360;
	5042	6235	0.013826	0.039826	0	0	0	0	0	0	1	-360	360;
	10014	6235	0.015174	0.04981	0	0	0	0	0	0	1	-360	360;
	11855	6732	0.003669	0.009008	0	0	0	0	0	0	1	-360	360;
	6732	2855	0.03276	0.106603	0	0	0	0	0	0	1	-360	360;
	6732	2855	0.03019	0.107612	0	0	0	0	0	0	1	-360	360;
	5422	6732	0.019033	0.051653	0	0	0	0	0	0	1	-360	360;
	265	6732	0.037603	0.124017	0	0	0	0	0	0	1	-360	360;
	6732	9134	0.034868	0.097471	0	0	0	0	0	0	1	-360	360;
	6732	9420	0.043099	0.130603	0	0	0	0	0	0	1	-360	360;
	6439	11855	0.06105	0.110339	0	0	0	0	0	0	1	-360	360;
	4035	9154	0.041446	0.108843	0	0	0	0	0	0	1	-360	360;
	9154	350	0.063471	0.123207	0	0	0	0	0	0	1	-360	360;
	4475	8009	0.02343	0.073603	0	0	0	0	0	0	1	-360	360;
	4475	2689	0.022537	0.091587	0	0	0	0	0	0	1	-360	360;
	30	13610	0.036744	0.106934	0	0	0	0	0	0	1	-360	360;
	30	1176	0.006207	0.023959	0	0	0	0	0	0	1	-360	360;
	5042	4984	0.003157	0.011008	0	0	0	0	0	0	1	-360	360;
	5042	4984	0.003223	0.010793	0	0	0	0	0	0	1	-360	360;
	10014	4984	0.005	0.035653	0	0	0	0	0	0	1	-360	360;
	10014	4984	0.004901	0.036347	0	0	0	0	0	0	1	-360	360;
	11857	2244	0.007917	0.029603	0	0	0	0	0	0	1	-360	360;
	1051	2244	0.004876	0.01476	0	0	0	0	0	0	1	-360	360;
	7959	13610	0.007182	0.023579	0	0	0	0	0	0	1	-360	360;
	7959	6720	0.007322	0.023099	0	0	0	0	0	0	1	-360	360;
	8553	560	0.010942	0.040562	0	0	0	0	0	0	1	-360	360;
	560	1279	0.030603	0.103702	0	0	0	0	0	0	1	-360	360;
	13278	3681	0.01176	0.041421	0	0	0	0	0	0	1	-360	360;
	13278	9344	0.011868	0.03581	0	0	0	0	0	0	1	-360	360;
	1130	13278	0.01757	0.064579	0	0	0	0	0	0	1	-360	360;
	10395	10745	0.017099	0.064281	0	0	0	0	0	0	1	-360	360;
	1146	5635	0.015207	0.057397	0	0	0	0	0	0	1	-360	360;
	9344	1146	0.009198	0.036512	0	0	0	0	0	0	1	-360	360;
	1146	10014	0.024281	0.239446	0	0	0	0	0	0	1	-360	360;
	13537	1861	0.052471	0.157347	0	0	0	0	0	0	1	-360	360;
	604	1861	0.009314	0.035909	0	0	0	0	0	0	1	-360	360;
	11857	1861	0.039314	0.122777	0	0	0	0	0	0	1	-360	360;
	12257	664	0.001223	0.003711	0	0	0	0	0	0	1	-360	360;
	7347	10824	0.00518	0.023091	0	0	0	0	0	0	1	-360	360;
	7347	10824	0.005281	0.022591	0	0	0	0	0	0	1	-360	360;
	3780	11841	0.018198	0.037149	0	0	0	0	0	0	1	-360	360;
	11628	8625	0.012893	0.040033	0	0	0	0	0	0	1	-360	360;
	5638	8625	0.007132	0.020331	0	0	0	0	0	0	1	-360	360;
	4642	2024	0.00479	0.05497	0	0	0	0	0	0	1	-360	360;
	6439	2855	0.047322	0.139868	0	0	0	0	0	0	1	-360	360;
	5080	2855	0.005471	0.017471	0	0	0	0	0	0	1	-360	360;
	5080	2855	0.013826	0.040207	0	0	0	0	0	0	1	-360	360;
	4283	2855	0.010678	0.03814	0	0	0	0	0	0	1	-360	360;
	4283	2855	0.010603	0.034793	0	0	0	0	0	0	1	-360	360;
	6826	2855	0.014579	0.043331	0	0	0	0	0	0	1	-360	360;
	2855	9134	0.021413	0.079248	0	0	0	0	0	0	1	-360	360;
	2855	10334	0.031198	0.098298	0	0	0	0	0	0	1	-360	360;
	8015	4915	0.009651	0.045031	0	0	0	0	0	0	1	-360	360;
	13255	8028	0.050207	0.124066	0	0	0	0	0	0	1	-360	360;
	13603	6659	0.034314	0.070521	0	0	0	0	0	0	1	-360	360;
	9712	8064	0.007463	0.022612	0	0	0	0	0	0	1	-360	360;
	9712	8064	0.007603	0.022132	0	0	0	0	0	0	1	-360	360;
	8009	5701	0.01081	0.033603	0	0	0	0	0	0	1	-360	360;
	13537	8009	0.011223	0.036446	0	0	0	0	0	0	1	-360	360;
	8009	11628	0.014182	0.045669	0	0	0	0	0	0	1	-360	360;
	11628	8009	0.014463	0.044744	0	0	0	0	0	0	1	-360	360;
	265	5422	0.02124	0.069099	0	0	0	0	0	0	1	-360	360;
	13571	11706	0.008041	0.030504	0	0	0	0	0	0	1	-360	360;
	13571	3318	0.009107	0.029893	0	0	0	0	0	0	1	-360	360;
	12796	11224	0.024802	0.074537	0	0	0	0	0	0	1	-360	360;
	187	12796	0.038711	0.136876	0	0	0	0	0	0	1	-360	360;
	1051	2689	0.069413	0.194223	0	0	0	0	0	0	1	-360	360;
	13537	5701	0.01276	0.035281	0	0	0	0	0	0	1	-360	360;
	6174	10974	0.008744	0.025802	0	0	0	0	0	0	1	-360	360;
	10974	3318	0.008909	0.025248	0	0	0	0	0	0	1	-360	360;
	4863	11706	0.009289	0.029281	0	0	0	0	0	0	1	-360	360;
	2387	5481	0.021421	0.081298	0	0	0	0	0	0	1	-360	360;
	2387	10395	0.044479	0.139058	0	0	0	0	0	0	1	-360	360;
	13362	11209	0.008479	0.049461	0	0	0	0	0	0	1	-360	360;
	7347	11209	0.00394	0.02236	0	0	0	0	0	0	1	-360	360;
	1067	8238	0.016174	0.044463	0	0	0	0	0	0	1	-360	360;
	1067	8238	0.014413	0.05038	0	0	0	0	0	0	1	-360	360;
	1115	1067	0.020198	0.064678	0	0	0	0	0	0	1	-360	360;
	1115	1067	0.020587	0.063364	0	0	0	0	0	0	1	-360	360;
	1115	1067	0.031711	0.059223	0	0	0	0	0	0	1	-360	360;
	13603	3002	0.056678	0.178372	0	0	0	0	0	0	1	-360	360;
	3002	7433	0.019496	0.078083	0	0	0	0	0	0	1	-360	360;
	8238	455	0.045198	0.111479	0	0	0	0	0	0	1	-360	360;
	7832	8064	0.015041	0.057983	0	0	0	0	0	0	1	-360	360;
	1495	7832	0.03519	0.105653	0	0	0	0	0	0	1	-360	360;
	187	9093	0.044	0.099339	0	0	0	0	0	0	1	-360	360;
	9385	12257	0.013636	0.039512	0	0	0	0	0	0	1	-360	360;
	9385	12257	0.012149	0.044777	0	0	0	0	0	0	1	-360	360;
	8064	13595	0.035281	0.122702	0	0	0	0	0	0	1	-360	360;
	5783	13595	0.031967	0.090248	0	0	0	0	0	0	1	-360	360;
	10014	13595	0.023983	0.066421	0	0	0	0	0	0	1	-360	360;
	13603	7753	0.032901	0.126777	0	0	0	0	0	0	1	-360	360;
	7753	12990	0.015372	0.059083	0	0	0	0	0	0	1	-360	360;
	6174	10395	0.012636	0.024496	0	0	0	0	0	0	1	-360	360;
	6174	10395	0.012893	0.024	0	0	0	0	0	0	1	-360	360;
	6174	10915	0.001917	0.006091	0	0	0	0	0	0	1	-360	360;
	6174	3318	0.018149	0.049413	0	0	0	0	0	0	1	-360	360;
	3084	1219	0.01276	0.037083	0	0	0	0	0	0	1	-360	360;
	13176	3084	0.069636	0.146636	0	0	0	0	0	0	1	-360	360;
	3795	13272	0.00219	0.02213	0	0	0	0	0	0	1	-360	360;
	8344	7347	0.00194	0.0087	0	0	0	0	0	0	1	-360	360;
	8344	7347	0.001979	0.00851	0	0	0	0	0	0	1	-360	360;
	5635	3825	0.028893	0.108496	0	0	0	0	0	0	1	-360	360;
	3681	5635	0.021008	0.05814	0	0	0	0	0	0	1	-360	360;
	8238	5726	0.018661	0.056488	0	0	0	0	0	0	1	-360	360;
	963	1495	0.01381	0.03776	0	0	0	0	0	0	1	-360	360;
	963	5783	0.012298	0.042793	0	0	0	0	0	0	1	-360	360;
	8536	8553	0.009281	0.03038	0	0	0	0	0	0	1	-360	360;
	11224	4035	0.011438	0.035711	0	0	0	0	0	0	1	-360	360;
	9949	11224	0.05095	0.124661	0	0	0	0	0	0	1	-360	360;
	8028	13603	0.058207	0.110223	0	0	0	0	0	0	1	-360	360;
	604	1051	0.061876	0.195174	0	0	0	0	0	0	1	-360	360;
	3833	2929	0.013438	0.032033	0	0	0	0	0	0	1	-360	360;
	6720	3833	0.03619	0.085066	0	0	0	0	0	0	1	-360	360;
	5726	3828	0.05324	0.154826	0	0	0	0	0	0	1	-360	360;
	5638	2929	0.004579	0.026017	0	0	0	0	0	0	1	-360	360;
	2132	13415	0.00795	0.03657	0	0	0	0	0	0	1	-360	360;
	4035	1787	0.055967	0.167983	0	0	0	0	0	0	1	-360	360;
	9949	4035	0.070942	0.168066	0	0	0	0	0	0	1	-360	360;
	3592	3795	0.00389	0.04149	0	0	0	0	0	0	1	-360	360;
	10347	3592	0.003	0.02776	0	0	0	0	0	0	1	-360	360;
	8002	8015	0.002159	0.01261	0	0	0	0	0	0	1	-360	360;
	8002	8015	0.002021	0.016849	0	0	0	0	0	0	1	-360	360;
	10240	8015	0.007151	0.046719	0	0	0	0	0	0	1	-360	360;
	8015	10118	0.000539	0.00368	0	0	0	0	0	0	1	-360	360;
	8015	10118	0.000539	0.003591	0	0	0	0	0	0	1	-360	360;
	11628	8536	0.007446	0.010579	0	0	0	0	0	0	1	-360	360;
	11628	3653	0.011992	0.03276	0	0	0	0	0	0	1	-360	360;
	1115	3780	0.034207	0.085124	0	0	0	0	0	0	1	-360	360;
	7098	1115	0.001248	0.004471	0	0	0	0	0	0	1	-360	360;
	7098	1115	0.001223	0.004562	0	0	0	0	0	0	1	-360	360;
	8006	2132	0.006961	0.04039	0	0	0	0	0	0	1	-360	360;
	8006	2132	0.004521	0.02305	0	0	0	0	0	0	1	-360	360;
	8006	6485	0.00395	0.022789	0	0	0	0	0	0	1	-360	360;
	187	11321	0.031934	0.123661	0	0	0	0	0	0	1	-360	360;
	187	5042	0.019802	0.070901	0	0	0	0	0	0	1	-360	360;
	187	5042	0.022438	0.069479	0	0	0	0	0	0	1	-360	360;
	187	9340	0.014727	0.04305	0	0	0	0	0	0	1	-360	360;
	8355	187	0.01343	0.06824	0	0	0	0	0	0	1	-360	360;
	2751	187	0.022	0.060388	0	0	0	0	0	0	1	-360	360;
	2751	187	0.019603	0.068438	0	0	0	0	0	0	1	-360	360;
	1787	1727	0.048769	0.114711	0	0	0	0	0	0	1	-360	360;
	5080	6439	0.046901	0.148033	0	0	0	0	0	0	1	-360	360;
	10095	6439	0.055744	0.139066	0	0	0	0	0	0	1	-360	360;
	1130	8064	0.010917	0.032901	0	0	0	0	0	0	1	-360	360;
	84	8013	0.026322	0.044388	0	0	0	0	0	0	1	-360	360;
	7433	12188	0.047438	0.143727	0	0	0	0	0	0	1	-360	360;
	11126	12188	0.026876	0.09776	0	0	0	0	0	0	1	-360	360;
	12990	11126	0.01705	0.060802	0	0	0	0	0	0	1	-360	360;
	1279	12990	0.028727	0.075289	0	0	0	0	0	0	1	-360	360;
	7433	8133	0.002388	0.015942	0	0	0	0	0	0	1	-360	360;
	7433	4016	0.088471	0.217198	0	0	0	0	0	0	1	-360	360;
	3828	9733	0.01714	0.056248	0	0	0	0	0	0	1	-360	360;
	9435	5080	0.004479	0.014008	0	0	0	0	0	0	1	-360	360;
	4283	9435	0.009851	0.032314	0	0	0	0	0	0	1	-360	360;
	6826	9435	0.001636	0.005587	0	0	0	0	0	0	1	-360	360;
	6826	12107	0.021603	0.084	0	0	0	0	0	0	1	-360	360;
	2546	6890	0.021727	0.065893	0	0	0	0	0	0	1	-360	360;
	2546	6720	0.025686	0.080256	0	0	0	0	0	0	1	-360	360;
	6720	2546	0.026182	0.078587	0	0	0	0	0	0	1	-360	360;
	10240	8002	0.00536	0.040919	0	0	0	0	0	0	1	-360	360;
	3653	5638	0.007388	0.026628	0	0	0	0	0	0	1	-360	360;
	9385	12404	0.005826	0.017488	0	0	0	0	0	0	1	-360	360;
	9385	12404	0.005826	0.016653	0	0	0	0	0	0	1	-360	360;
	5783	12404	0.026826	0.082273	0	0	0	0	0	0	1	-360	360;
	13202	9385	0.003256	0.010587	0	0	0	0	0	0	1	-360	360;
	13202	1362	0.003331	0.010372	0	0	0	0	0	0	1	-360	360;
	13400	11857	0.025174	0.077661	0	0	0	0	0	0	1	-360	360;
	2795	6890	0.030248	0.085413	0	0	0	0	0	0	1	-360	360;
	7363	6890	0.029471	0.109198	0	0	0	0	0	0	1	-360	360;
	1727	265	0.058942	0.115529	0	0	0	0	0	0	1	-360	360;
	9340	11321	0.028413	0.083	0	0	0	0	0	0	1	-360	360;
	9949	11321	0.005826	0.016	0	0	0	0	0	0	1	-360	360;
	12624	11321	0.005512	0.016612	0	0	0	0	0	0	1	-360	360;
	8293	9385	0.014256	0.033909	0	0	0	0	0	0	1	-360	360;
	1051	9385	0.052826	0.200702	0	0	0	0	0	0	1	-360	360;
	2751	9385	0.01976	0.061058	0	0	0	0	0	0	1	-360	360;
	2751	9385	0.014008	0.053207	0	0	0	0	0	0	1	-360	360;
	2751	9385	0.013711	0.054207	0	0	0	0	0	0	1	-360	360;
	8355	5042	0.009132	0.024843	0	0	0	0	0	0	1	-360	360;
	10014	5042	0.010711	0.040256	0	0	0	0	0	0	1	-360	360;
	9559	5042	0.004562	0.012421	0	0	0	0	0	0	1	-360	360;
	11770	2795	0.025273	0.079777	0	0	0	0	0	0	1	-360	360;
	1362	8293	0.009149	0.028992	0	0	0	0	0	0	1	-360	360;
	4863	3318	0.027149	0.096802	0	0	0	0	0	0	1	-360	360;
	1176	6720	0.063033	0.157099	0	0	0	0	0	0	1	-360	360;
	1176	11770	0.009702	0.035397	0	0	0	0	0	0	1	-360	360;
	1176	11770	0.009504	0.036091	0	0	0	0	0	0	1	-360	360;
	350	10095	0.029893	0.105364	0	0	0	0	0	0	1	-360	360;
	8355	8262	0.005397	0.020298	0	0	0	0	0	0	1	-360	360;
	10915	4117	0.001124	0.002942	0	0	0	0	0	0	1	-360	360;
	7098	4117	0.001207	0.003174	0	0	0	0	0	0	1	-360	360;
	7098	4117	0.00124	0.003099	0	0	0	0	0	0	1	-360	360;
	9559	8355	0.004066	0.014083	0	0	0	0	0	0	1	-360	360;
	3423	10240	0.005099	0.025479	0	0	0	0	0	0	1	-360	360;
	3423	10240	0.0052	0.024961	0	0	0	0	0	0	1	-360	360;
	12624	9949	0.003868	0.013901	0	0	0	0	0	0	1	-360	360;
	10347	2024	0.00126	0.01433	0	0	0	0	0	0	1	-360	360;
	13400	2751	0.011752	0.044314	0	0	0	0	0	0	1	-360	360;
	9134	10334	0.021719	0.046248	0	0	0	0	0	0	1	-360	360;
	9134	9420	0.05324	0.100281	0	0	0	0	0	0	1	-360	360;
	6459	9848	0.00109	0.01155	0	0	0	0	0	0	1	-360	360;
	6459	9848	0.0011	0.01158	0	0	0	0	0	0	1	-360	360;
	6459	4348	0.00096	0.01293	0	0	0	0	0	0	1	-360	360;
	6459	4348	0.00098	0.01264	0	0	0	0	0	0	1	-360	360;
	6459	829	0.00101	0.01059	0	0	0	0	0	0	1	-360	360;
	11848	9066	0.00112	0.010969	0	0	0	0	0	0	1	-360	360;
	435	9066	0.00126	0.01075	0	0	0	0	0	0	1	-360	360;
	5430	12666	0.00197	0.0207	0	0	0	0	0	0	1	-360	360;
	5430	829	0.00303	0.0273	0	0	0	0	0	0	1	-360	360;
	6869	12432	0.00194	0.02339	0	0	0	0	0	0	1	-360	360;
	4348	6869	0.00057	0.00492	0	0	0	0	0	0	1	-360	360;
	9779	4099	0.01105	0.056599	0	0	0	0	0	0	1	-360	360;
	7590	4099	0.00482	0.046521	0	0	0	0	0	0	1	-360	360;
	9848	12573	0.00194	0.01645	0	0	0	0	0	0	1	-360	360;
	12573	10686	0.00134	0.01459	0	0	0	0	0	0	1	-360	360;
	7706	12584	0.000401	0.00318	0	0	0	0	0	0	1	-360	360;
	7706	10323	0.000349	0.00293	0	0	0	0	0	0	1	-360	360;
	7706	1986	0.000401	0.00324	0	0	0	0	0	0	1	-360	360;
	7706	4099	0.01214	0.058401	0	0	0	0	0	0	1	-360	360;
	7706	4099	0.012979	0.053281	0	0	0	0	0	0	1	-360	360;
	7706	1313	0.007919	0.053729	0	0	0	0	0	0	1	-360	360;
	7706	9934	0.014771	0.06813	0	0	0	0	0	0	1	-360	360;
	7706	9934	0.013789	0.067711	0	0	0	0	0	0	1	-360	360;
	2356	9066	0.005349	0.044469	0	0	0	0	0	0	1	-360	360;
	13314	9066	0.005229	0.04118	0	0	0	0	0	0	1	-360	360;
	9066	7201	0.0008	0.00676	0	0	0	0	0	0	1	-360	360;
	9066	3036	0.000781	0.00626	0	0	0	0	0	0	1	-360	360;
	9066	11556	0.00363	0.017409	0	0	0	0	0	0	1	-360	360;
	9066	11556	0.0037	0.01705	0	0	0	0	0	0	1	-360	360;
	9066	3430	0.00449	0.046469	0	0	0	0	0	0	1	-360	360;
	7196	5295	0.00052	0.00727	0	0	0	0	0	0	1	-360	360;
	7196	10686	0.00047	0.0062	0	0	0	0	0	0	1	-360	360;
	7196	10686	0.00046	0.00631	0	0	0	0	0	0	1	-360	360;
	7957	9934	0.003531	0.013669	0	0	0	0	0	0	1	-360	360;
	10922	12432	0.00032	0.00322	0	0	0	0	0	0	1	-360	360;
	10922	7049	0.00156	0.01603	0	0	0	0	0	0	1	-360	360;
	10922	3389	0.0022	0.02389	0	0	0	0	0	0	1	-360	360;
	13262	3430	0.00337	0.032669	0	0	0	0	0	0	1	-360	360;
	13575	7049	0.00199	0.0195	0	0	0	0	0	0	1	-360	360;
	13575	3389	0.00144	0.01354	0	0	0	0	0	0	1	-360	360;
	13575	3389	0.00147	0.01325	0	0	0	0	0	0	1	-360	360;
	13575	7409	0.00211	0.02433	0	0	0	0	0	0	1	-360	360;
	10019	4099	0.013901	0.088849	0	0	0	0	0	0	1	-360	360;
	7049	4348	0.00149	0.01564	0	0	0	0	0	0	1	-360	360;
	7049	11735	0.00294	0.02278	0	0	0	0	0	0	1	-360	360;
	4348	11735	0.00133	0.01873	0	0	0	0	0	0	1	-360	360;
	11556	9934	0.00581	0.034229	0	0	0	0	0	0	1	-360	360;
	11735	9919	0.00145	0.02024	0	0	0	0	0	0	1	-360	360;
	11735	9919	0.00142	0.01874	0	0	0	0	0	0	1	-360	360;
	11735	12666	0.00245	0.02367	0	0	0	0	0	0	1	-360	360;
	11735	7409	0.00178	0.02175	0	0	0	0	0	0	1	-360	360;
	8194	10478	0.000701	0.009299	0	0	0	0	0	0	1	-360	360;
	8194	10478	0.001042	0.007833	0	0	0	0	0	0	1	-360	360;
	8194	10478	0.001403	0.00534	0	0	0	0	0	0	1	-360	360;
	10686	4523	0.00022	0.00258	0	0	0	0	0	0	1	-360	360;
	4099	1313	0.02024	0.099349	0	0	0	0	0	0	1	-360	360;
	4099	13107	0.00044	0.003081	0	0	0	0	0	0	1	-360	360;
	4099	6922	0.00043	0.00314	0	0	0	0	0	0	1	-360	360;
	4099	12919	0.00461	0.027531	0	0	0	0	0	0	1	-360	360;
	13229	1313	0.01212	0.09901	0	0	0	0	0	0	1	-360	360;
	541	3196	0.000289	0.00261	0	0	0	0	0	0	1	-360	360;
	6705	541	0.00043	0.002831	0	0	0	0	0	0	1	-360	360;
	5139	541	0.000289	0.000599	0	0	0	0	0	0	1	-360	360;
	10964	1637	0.000289	0.00224	0	0	0	0	0	0	1	-360	360;
	6071	10964	0.000219	0.001591	0	0	0	0	0	0	1	-360	360;
	3856	13018	0.00045	0.00786	0	0	0	0	0	0	1	-360	360;
	3856	13018	0.00044	0.00728	0	0	0	0	0	0	1	-360	360;
	13018	12148	0.00049	0.00466	0	0	0	0	0	0	1	-360	360;
	13018	11938	0.00125	0.01973	0	0	0	0	0	0	1	-360	360;
	9947	9945	0.001461	0.00714	0	0	0	0	0	0	1	-360	360;
	13472	9947	0.00037	0.00094	0	0	0	0	0	0	1	-360	360;
	9947	2575	0.001039	0.00737	0	0	0	0	0	0	1	-360	360;
	7737	1430	8.1e-05	0.00061	0	0	0	0	0	0	1	-360	360;
	10714	11870	0.00053	0.00852	0	0	0	0	0	0	1	-360	360;
	7968	11870	0.00053	0.00704	0	0	0	0	0	0	1	-360	360;
	7461	4899	0.0001	0.00128	0	0	0	0	0	0	1	-360	360;
	6384	13417	0.0015	0.00705	0	0	0	0	0	0	1	-360	360;
	13417	10325	0.009771	0.05363	0	0	0	0	0	0	1	-360	360;
	6384	13417	0.00088	0.007771	0	0	0	0	0	0	1	-360	360;
	9406	12442	0.00539	0.032979	0	0	0	0	0	0	1	-360	360;
	2092	9406	0.00482	0.03086	0	0	0	0	0	0	1	-360	360;
	5423	9406	0.010539	0.065479	0	0	0	0	0	0	1	-360	360;
	9406	11593	0.01074	0.06405	0	0	0	0	0	0	1	-360	360;
	7554	9406	0.0063	0.048909	0	0	0	0	0	0	1	-360	360;
	2597	1260	0.0002	0.00022	0	0	0	0	0	0	1	-360	360;
	7288	1260	0.0002	0.00022	0	0	0	0	0	0	1	-360	360;
	9966	1260	0.00019	0.00023	0	0	0	0	0	0	1	-360	360;
	1295	1809	0.000289	0.0017	0	0	0	0	0	0	1	-360	360;
	1260	1700	0.00145	0.01937	0	0	0	0	0	0	1	-360	360;
	4515	1260	0.00048	0.00814	0	0	0	0	0	0	1	-360	360;
	4515	1260	0.00057	0.00758	0	0	0	0	0	0	1	-360	360;
	12355	11550	0.0002	0.00187	0	0	0	0	0	0	1	-360	360;
	12355	11550	0.000169	0.00193	0	0	0	0	0	0	1	-360	360;
	8875	10160	0.00216	0.03591	0	0	0	0	0	0	1	-360	360;
	8875	2196	0.00092	0.01287	0	0	0	0	0	0	1	-360	360;
	1782	5296	0.02412	0.10443	0	0	0	0	0	0	1	-360	360;
	1809	12442	0.00063	0.00401	0	0	0	0	0	0	1	-360	360;
	1809	5240	0.00038	0.002409	0	0	0	0	0	0	1	-360	360;
	7147	6024	0.010091	0.05524	0	0	0	0	0	0	1	-360	360;
	7994	7147	0.00668	0.04457	0	0	0	0	0	0	1	-360	360;
	7398	8255	0.00057	0.004961	0	0	0	0	0	0	1	-360	360;
	8255	7977	0.00025	0.002159	0	0	0	0	0	0	1	-360	360;
	4765	6626	0.00443	0.01807	0	0	0	0	0	0	1	-360	360;
	4765	8500	0.00037	0.001659	0	0	0	0	0	0	1	-360	360;
	7398	11495	0.00037	0.003599	0	0	0	0	0	0	1	-360	360;
	7398	10969	0.000659	0.00836	0	0	0	0	0	0	1	-360	360;
	7398	10969	0.00151	0.012289	0	0	0	0	0	0	1	-360	360;
	7398	3422	0.000539	0.003711	0	0	0	0	0	0	1	-360	360;
	12378	1885	0.00036	0.00606	0	0	0	0	0	0	1	-360	360;
	12378	1885	0.00035	0.00618	0	0	0	0	0	0	1	-360	360;
	1885	11460	0.00101	0.01675	0	0	0	0	0	0	1	-360	360;
	1885	7797	4e-05	0.00071	0	0	0	0	0	0	1	-360	360;
	1885	7797	4e-05	0.00081	0	0	0	0	0	0	1	-360	360;
	1885	1395	0.00013	0.00152	0	0	0	0	0	0	1	-360	360;
	1885	1395	0.00014	0.00149	0	0	0	0	0	0	1	-360	360;
	1885	6133	0.00013	0.00164	0	0	0	0	0	0	1	-360	360;
	1885	6133	0.00013	0.00168	0	0	0	0	0	0	1	-360	360;
	8766	7554	0.000669	0.00511	0	0	0	0	0	0	1	-360	360;
	13251	9007	3.9e-05	0.000341	0	0	0	0	0	0	1	-360	360;
	8579	8264	0.004219	0.02145	0	0	0	0	0	0	1	-360	360;
	8579	5926	0.000229	0.002849	0	0	0	0	0	0	1	-360	360;
	1782	70	0.0137	0.10436	0	0	0	0	0	0	1	-360	360;
	3830	1782	0.001521	0.01449	0	0	0	0	0	0	1	-360	360;
	1108	643	8.1e-05	0.00057	0	0	0	0	0	0	1	-360	360;
	9945	643	0.001461	0.00813	0	0	0	0	0	0	1	-360	360;
	2017	643	0.00049	0.00381	0	0	0	0	0	0	1	-360	360;
	13225	3705	0.000401	0.00287	0	0	0	0	0	0	1	-360	360;
	8298	7058	8.3e-05	0.000347	0	0	0	0	0	0	1	-360	360;
	3518	10740	0.00023	0.00285	0	0	0	0	0	0	1	-360	360;
	10740	4628	0.00016	0.00218	0	0	0	0	0	0	1	-360	360;
	4302	10740	0.00057	0.00736	0	0	0	0	0	0	1	-360	360;
	6332	11542	0.000591	0.005789	0	0	0	0	0	0	1	-360	360;
	6332	4868	0.002729	0.011461	0	0	0	0	0	0	1	-360	360;
	6332	8920	0.00051	0.00205	0	0	0	0	0	0	1	-360	360;
	12176	9859	0.00031	0.00374	0	0	0	0	0	0	1	-360	360;
	3876	12176	0.00032	0.00518	0	0	0	0	0	0	1	-360	360;
	2835	10707	0.00136	0.01038	0	0	0	0	0	0	1	-360	360;
	12587	8306	0.00151	0.02083	0	0	0	0	0	0	1	-360	360;
	11223	8306	0.00015	0.00232	0	0	0	0	0	0	1	-360	360;
	7434	11942	1e-05	0.00025	0	0	0	0	0	0	1	-360	360;
	8840	11942	1e-05	0.00023	0	0	0	0	0	0	1	-360	360;
	3470	11942	0.00047	0.00683	0	0	0	0	0	0	1	-360	360;
	1241	11942	0.00067	0.00858	0	0	0	0	0	0	1	-360	360;
	3648	1487	0.003669	0.025331	0	0	0	0	0	0	1	-360	360;
	546	3648	0.00525	0.02994	0	0	0	0	0	0	1	-360	360;
	6852	10025	0.00595	0.03625	0	0	0	0	0	0	1	-360	360;
	6852	1272	0.004969	0.03411	0	0	0	0	0	0	1	-360	360;
	9133	2643	0.005669	0.04293	0	0	0	0	0	0	1	-360	360;
	9133	2643	0.00555	0.043771	0	0	0	0	0	0	1	-360	360;
	2643	5625	0.002831	0.022401	0	0	0	0	0	0	1	-360	360;
	2643	5625	0.00325	0.01938	0	0	0	0	0	0	1	-360	360;
	11845	233	0.00068	0.00082	0	0	0	0	0	0	1	-360	360;
	11845	2898	0.00054	0.00868	0	0	0	0	0	0	1	-360	360;
	2519	5509	0.00061	0.002909	0	0	0	0	0	0	1	-360	360;
	2519	2089	0.002271	0.01014	0	0	0	0	0	0	1	-360	360;
	5725	1785	0.00083	0.00986	0	0	0	0	0	0	1	-360	360;
	1785	11966	0.00031	0.00511	0	0	0	0	0	0	1	-360	360;
	1785	11966	0.0003	0.00521	0	0	0	0	0	0	1	-360	360;
	12715	1785	0.00081	0.01288	0	0	0	0	0	0	1	-360	360;
	1785	13320	0.00039	0.00645	0	0	0	0	0	0	1	-360	360;
	5538	6984	0.00243	0.012211	0	0	0	0	0	0	1	-360	360;
	5232	7735	0.00018	0.00226	0	0	0	0	0	0	1	-360	360;
	6957	7735	0.0002	0.00347	0	0	0	0	0	0	1	-360	360;
	4628	7735	0.00018	0.0029	0	0	0	0	0	0	1	-360	360;
	8003	9859	0.00045	0.00494	0	0	0	0	0	0	1	-360	360;
	9859	3843	0.00042	0.00497	0	0	0	0	0	0	1	-360	360;
	4689	12587	0.00019	0.00234	0	0	0	0	0	0	1	-360	360;
	10160	12587	8e-05	0.00105	0	0	0	0	0	0	1	-360	360;
	10160	12587	8e-05	0.00107	0	0	0	0	0	0	1	-360	360;
	12587	11639	8e-05	0.001	0	0	0	0	0	0	1	-360	360;
	12587	7796	0.00063	0.01015	0	0	0	0	0	0	1	-360	360;
	1239	10646	0.00064	0.005651	0	0	0	0	0	0	1	-360	360;
	1239	10646	0.000581	0.006531	0	0	0	0	0	0	1	-360	360;
	7467	9197	0.007099	0.04432	0	0	0	0	0	0	1	-360	360;
	3196	2575	0.00144	0.01157	0	0	0	0	0	0	1	-360	360;
	3196	860	0.00114	0.00695	0	0	0	0	0	0	1	-360	360;
	5232	2403	0.00044	0.0053	0	0	0	0	0	0	1	-360	360;
	10333	9683	0.00162	0.009039	0	0	0	0	0	0	1	-360	360;
	10333	9683	0.00187	0.0072	0	0	0	0	0	0	1	-360	360;
	11772	9683	0.00101	0.00564	0	0	0	0	0	0	1	-360	360;
	5453	9683	0.0007	0.00445	0	0	0	0	0	0	1	-360	360;
	4177	9683	0.001081	0.00619	0	0	0	0	0	0	1	-360	360;
	10160	12229	0.00022	0.00228	0	0	0	0	0	0	1	-360	360;
	10160	9069	0.00023	0.00223	0	0	0	0	0	0	1	-360	360;
	10160	11671	0.0002	0.00262	0	0	0	0	0	0	1	-360	360;
	10160	2742	0.00021	0.00261	0	0	0	0	0	0	1	-360	360;
	60	10216	0.00663	0.04181	0	0	0	0	0	0	1	-360	360;
	10160	3519	0.00059	0.00921	0	0	0	0	0	0	1	-360	360;
	10160	6109	0.0014	0.02227	0	0	0	0	0	0	1	-360	360;
	5685	4037	0.000401	0.002469	0	0	0	0	0	0	1	-360	360;
	8229	4037	0.00212	0.01218	0	0	0	0	0	0	1	-360	360;
	4037	2882	0.004169	0.02439	0	0	0	0	0	0	1	-360	360;
	12442	13406	0.02336	0.122419	0	0	0	0	0	0	1	-360	360;
	12442	997	0.002729	0.0137	0	0	0	0	0	0	1	-360	360;
	4123	12442	0.000919	0.003831	0	0	0	0	0	0	1	-360	360;
	4123	12442	0.00082	0.004349	0	0	0	0	0	0	1	-360	360;
	6626	12442	0.01325	0.055159	0	0	0	0	0	0	1	-360	360;
	6626	12442	0.0108	0.06074	0	0	0	0	0	0	1	-360	360;
	12442	7044	0.001099	0.00631	0	0	0	0	0	0	1	-360	360;
	11420	1430	0.00495	0.029151	0	0	0	0	0	0	1	-360	360;
	1430	13217	0.003909	0.0227	0	0	0	0	0	0	1	-360	360;
	5423	5219	0.00061	0.00424	0	0	0	0	0	0	1	-360	360;
	4113	1637	0.00068	0.003919	0	0	0	0	0	0	1	-360	360;
	6707	1637	0.00368	0.019901	0	0	0	0	0	0	1	-360	360;
	10873	6957	0.00025	0.00388	0	0	0	0	0	0	1	-360	360;
	10873	557	0.00038	0.00623	0	0	0	0	0	0	1	-360	360;
	9945	1247	0.000909	0.006031	0	0	0	0	0	0	1	-360	360;
	1247	2575	0.001211	0.0062	0	0	0	0	0	0	1	-360	360;
	3470	597	0.00022	0.00322	0	0	0	0	0	0	1	-360	360;
	3470	6619	0.00018	0.00307	0	0	0	0	0	0	1	-360	360;
	3470	7220	0.00068	0.00873	0	0	0	0	0	0	1	-360	360;
	3470	12006	0.00124	0.02016	0	0	0	0	0	0	1	-360	360;
	13516	10714	0.00025	0.00332	0	0	0	0	0	0	1	-360	360;
	13516	10714	0.00025	0.00335	0	0	0	0	0	0	1	-360	360;
	13516	10714	0.00025	0.00325	0	0	0	0	0	0	1	-360	360;
	4421	13516	0.00076	0.01075	0	0	0	0	0	0	1	-360	360;
	10206	13516	3e-05	0.00044	0	0	0	0	0	0	1	-360	360;
	9007	11027	0.001651	0.010849	0	0	0	0	0	0	1	-360	360;
	12256	9007	0.00288	0.018469	0	0	0	0	0	0	1	-360	360;
	10069	2159	0.00072	0.00923	0	0	0	0	0	0	1	-360	360;
	10069	12112	0.00069	0.00855	0	0	0	0	0	0	1	-360	360;
	4599	6710	0.00124	0.01667	0	0	0	0	0	0	1	-360	360;
	4599	7383	0.00034	0.00571	0	0	0	0	0	0	1	-360	360;
	6767	4599	0.0004	0.00638	0	0	0	0	0	0	1	-360	360;
	6767	4599	0.00039	0.0065	0	0	0	0	0	0	1	-360	360;
	11591	4599	0.00027	0.00451	0	0	0	0	0	0	1	-360	360;
	13479	5720	0.012409	0.074961	0	0	0	0	0	0	1	-360	360;
	13479	12694	0.002419	0.01761	0	0	0	0	0	0	1	-360	360;
	13479	2017	0.006901	0.053409	0	0	0	0	0	0	1	-360	360;
	4113	17	0.000599	0.00356	0	0	0	0	0	0	1	-360	360;
	4113	6626	0.011841	0.049341	0	0	0	0	0	0	1	-360	360;
	6705	1267	0.000521	0.004831	0	0	0	0	0	0	1	-360	360;
	128	1267	0.00055	0.00143	0	0	0	0	0	0	1	-360	360;
	11316	1267	0.0005	0.00364	0	0	0	0	0	0	1	-360	360;
	11316	1267	0.00056	0.003469	0	0	0	0	0	0	1	-360	360;
	1267	7570	0.00082	0.00257	0	0	0	0	0	0	1	-360	360;
	6643	7934	0.00249	0.01536	0	0	0	0	0	0	1	-360	360;
	6643	3563	0.00113	0.00651	0	0	0	0	0	0	1	-360	360;
	11910	12598	0.00643	0.045031	0	0	0	0	0	0	1	-360	360;
	12549	11910	0.003581	0.02706	0	0	0	0	0	0	1	-360	360;
	2591	2196	0.00046	0.00774	0	0	0	0	0	0	1	-360	360;
	2591	9458	0.00031	0.00483	0	0	0	0	0	0	1	-360	360;
	13153	5535	0.001099	0.00345	0	0	0	0	0	0	1	-360	360;
	1778	1396	0.00913	0.053211	0	0	0	0	0	0	1	-360	360;
	1396	2476	0.021151	0.095401	0	0	0	0	0	0	1	-360	360;
	9008	6710	0.00034	0.0055	0	0	0	0	0	0	1	-360	360;
	3949	5200	0.00064	0.01945	0	0	0	0	0	0	1	-360	360;
	3949	12823	0.0004	0.01026	0	0	0	0	0	0	1	-360	360;
	9752	3949	0.00071	0.01087	0	0	0	0	0	0	1	-360	360;
	12996	2083	0.00968	0.088169	0	0	0	0	0	0	1	-360	360;
	12996	2083	0.012349	0.086409	0	0	0	0	0	0	1	-360	360;
	2083	6018	0.00288	0.017211	0	0	0	0	0	0	1	-360	360;
	6018	2083	0.00257	0.0195	0	0	0	0	0	0	1	-360	360;
	4450	1487	0.00051	0.00407	0	0	0	0	0	0	1	-360	360;
	157	9558	0.00019	0.00309	0	0	0	0	0	0	1	-360	360;
	157	10392	0.00023	0.00357	0	0	0	0	0	0	1	-360	360;
	4795	10467	0.00088	0.0092	0	0	0	0	0	0	1	-360	360;
	4795	437	0.00023	0.00179	0	0	0	0	0	0	1	-360	360;
	8849	1321	0.00145	0.007919	0	0	0	0	0	0	1	-360	360;
	692	1321	0.00143	0.00812	0	0	0	0	0	0	1	-360	360;
	1321	8264	0.00524	0.02181	0	0	0	0	0	0	1	-360	360;
	2899	5538	0.00261	0.01913	0	0	0	0	0	0	1	-360	360;
	6626	5538	0.00164	0.016591	0	0	0	0	0	0	1	-360	360;
	6574	7660	0.00087	0.012719	0	0	0	0	0	0	1	-360	360;
	3	11181	0.00775	0.03501	0	0	0	0	0	0	1	-360	360;
	3	11505	0.01944	0.093479	0	0	0	0	0	0	1	-360	360;
	3	11505	0.01981	0.091539	0	0	0	0	0	0	1	-360	360;
	9382	8209	0.003771	0.020599	0	0	0	0	0	0	1	-360	360;
	12846	9382	0.001271	0.00811	0	0	0	0	0	0	1	-360	360;
	9207	1541	0.00048	0.01459	0	0	0	0	0	0	1	-360	360;
	5200	9207	0.00047	0.01351	0	0	0	0	0	0	1	-360	360;
	3386	2092	0.021729	0.088281	0	0	0	0	0	0	1	-360	360;
	2092	4123	0.00038	0.001771	0	0	0	0	0	0	1	-360	360;
	2092	1778	0.022901	0.101401	0	0	0	0	0	0	1	-360	360;
	7	7944	0.002909	0.01432	0	0	0	0	0	0	1	-360	360;
	7	3830	0.015729	0.08824	0	0	0	0	0	0	1	-360	360;
	4170	10167	0.00038	0.002169	0	0	0	0	0	0	1	-360	360;
	2575	4170	0.000479	0.00212	0	0	0	0	0	0	1	-360	360;
	4170	590	0.00019	0.00086	0	0	0	0	0	0	1	-360	360;
	1106	201	0.00089	0.004901	0	0	0	0	0	0	1	-360	360;
	5689	1541	0.00109	0.03151	0	0	0	0	0	0	1	-360	360;
	9133	9070	0.003409	0.020271	0	0	0	0	0	0	1	-360	360;
	9133	9070	0.002909	0.02388	0	0	0	0	0	0	1	-360	360;
	9070	5625	0.00545	0.042969	0	0	0	0	0	0	1	-360	360;
	9070	5625	0.005341	0.0438	0	0	0	0	0	0	1	-360	360;
	7977	10969	0.001521	0.00863	0	0	0	0	0	0	1	-360	360;
	4625	7977	0.00057	0.005979	0	0	0	0	0	0	1	-360	360;
	1117	8349	0.00085	0.01449	0	0	0	0	0	0	1	-360	360;
	1241	8349	0.0006	0.00777	0	0	0	0	0	0	1	-360	360;
	7220	8349	0.0003	0.00492	0	0	0	0	0	0	1	-360	360;
	3844	3779	0.003349	0.02686	0	0	0	0	0	0	1	-360	360;
	3172	3779	0.00213	0.016849	0	0	0	0	0	0	1	-360	360;
	5591	13652	0.00077	0.0106	0	0	0	0	0	0	1	-360	360;
	12119	13652	0.00155	0.0148	0	0	0	0	0	0	1	-360	360;
	2300	9133	0.016651	0.10632	0	0	0	0	0	0	1	-360	360;
	2300	9133	0.014841	0.12049	0	0	0	0	0	0	1	-360	360;
	6845	9133	9.9e-05	0.001229	0	0	0	0	0	0	1	-360	360;
	11349	9133	9.9e-05	0.00114	0	0	0	0	0	0	1	-360	360;
	12162	3126	0.00048	0.00781	0	0	0	0	0	0	1	-360	360;
	12162	4877	0.00076	0.01058	0	0	0	0	0	0	1	-360	360;
	12675	12162	0.00073	0.01181	0	0	0	0	0	0	1	-360	360;
	10597	7383	0.00059	0.00929	0	0	0	0	0	0	1	-360	360;
	212	7383	8e-05	0.00092	0	0	0	0	0	0	1	-360	360;
	443	1487	0.00019	0.000849	0	0	0	0	0	0	1	-360	360;
	3413	5807	0.003479	0.01538	0	0	0	0	0	0	1	-360	360;
	3413	5807	0.003539	0.01505	0	0	0	0	0	0	1	-360	360;
	3413	2089	0.01387	0.084099	0	0	0	0	0	0	1	-360	360;
	1573	3413	0.00375	0.021341	0	0	0	0	0	0	1	-360	360;
	12345	7830	0.00023	0.00345	0	0	0	0	0	0	1	-360	360;
	7830	2403	0.00054	0.00662	0	0	0	0	0	0	1	-360	360;
	597	12655	7e-05	0.00099	0	0	0	0	0	0	1	-360	360;
	8016	597	0.00012	0.002	0	0	0	0	0	0	1	-360	360;
	343	10467	0.00075	0.00888	0	0	0	0	0	0	1	-360	360;
	4486	10467	0.00054	0.00705	0	0	0	0	0	0	1	-360	360;
	10025	6078	0.00574	0.0287	0	0	0	0	0	0	1	-360	360;
	10025	10632	0.001331	0.009479	0	0	0	0	0	0	1	-360	360;
	10632	10025	0.00539	0.024159	0	0	0	0	0	0	1	-360	360;
	7316	10025	0.000539	0.0028	0	0	0	0	0	0	1	-360	360;
	11316	11785	0.000969	0.00724	0	0	0	0	0	0	1	-360	360;
	9723	11785	0.000349	0.00339	0	0	0	0	0	0	1	-360	360;
	3843	12345	0.00021	0.00349	0	0	0	0	0	0	1	-360	360;
	11772	60	0.00318	0.020669	0	0	0	0	0	0	1	-360	360;
	11772	60	0.003599	0.02026	0	0	0	0	0	0	1	-360	360;
	60	9311	0.0053	0.035521	0	0	0	0	0	0	1	-360	360;
	60	4007	0.00486	0.024331	0	0	0	0	0	0	1	-360	360;
	9945	6705	0.000659	0.00645	0	0	0	0	0	0	1	-360	360;
	7944	5240	0.00239	0.01531	0	0	0	0	0	0	1	-360	360;
	11901	2241	0.00119	0.01215	0	0	0	0	0	0	1	-360	360;
	2241	10536	0.00012	0.0015	0	0	0	0	0	0	1	-360	360;
	2241	2550	0.00047	0.00738	0	0	0	0	0	0	1	-360	360;
	11542	8678	0.00068	0.00189	0	0	0	0	0	0	1	-360	360;
	7309	12574	0.003669	0.022099	0	0	0	0	0	0	1	-360	360;
	12846	12574	0.004771	0.0218	0	0	0	0	0	0	1	-360	360;
	4177	11181	0.001469	0.006591	0	0	0	0	0	0	1	-360	360;
	4881	747	8.1e-05	0.00045	0	0	0	0	0	0	1	-360	360;
	747	692	0.001021	0.00782	0	0	0	0	0	0	1	-360	360;
	5926	747	0.0015	0.009591	0	0	0	0	0	0	1	-360	360;
	8998	11495	0.00019	0.001581	0	0	0	0	0	0	1	-360	360;
	8998	11316	3.9e-05	0.000479	0	0	0	0	0	0	1	-360	360;
	10714	12703	0.00027	0.00281	0	0	0	0	0	0	1	-360	360;
	7968	10714	0.00096	0.01609	0	0	0	0	0	0	1	-360	360;
	10714	12675	0.00027	0.00339	0	0	0	0	0	0	1	-360	360;
	10714	12675	0.00023	0.00398	0	0	0	0	0	0	1	-360	360;
	1867	11027	0.000901	0.00937	0	0	0	0	0	0	1	-360	360;
	13225	11027	0.001901	0.01549	0	0	0	0	0	0	1	-360	360;
	12256	11027	0.003591	0.02205	0	0	0	0	0	0	1	-360	360;
	11027	8297	0.0013	0.009901	0	0	0	0	0	0	1	-360	360;
	3824	4421	0.00022	0.00299	0	0	0	0	0	0	1	-360	360;
	4421	12762	0.00047	0.00564	0	0	0	0	0	0	1	-360	360;
	8	4421	0.00046	0.0075	0	0	0	0	0	0	1	-360	360;
	9203	1028	0.00214	0.020979	0	0	0	0	0	0	1	-360	360;
	4123	1028	0.004651	0.03332	0	0	0	0	0	0	1	-360	360;
	1028	7554	0.00468	0.021281	0	0	0	0	0	0	1	-360	360;
	1028	7044	0.00581	0.028331	0	0	0	0	0	0	1	-360	360;
	2520	10548	7e-05	0.00086	0	0	0	0	0	0	1	-360	360;
	3777	6523	0.0002	0.00279	0	0	0	0	0	0	1	-360	360;
	3777	6523	0.00017	0.00333	0	0	0	0	0	0	1	-360	360;
	8283	11901	1e-05	0.00036	0	0	0	0	0	0	1	-360	360;
	7570	12746	0.000401	0.004081	0	0	0	0	0	0	1	-360	360;
	12746	10499	0.001581	0.015719	0	0	0	0	0	0	1	-360	360;
	11326	1272	0.00012	0.000229	0	0	0	0	0	0	1	-360	360;
	11326	1272	0.00012	0.00024	0	0	0	0	0	0	1	-360	360;
	1272	7315	0.00132	0.005729	0	0	0	0	0	0	1	-360	360;
	12979	1272	0.00438	0.024901	0	0	0	0	0	0	1	-360	360;
	8209	152	0.005521	0.023711	0	0	0	0	0	0	1	-360	360;
	343	10793	0.0008	0.01014	0	0	0	0	0	0	1	-360	360;
	343	10793	0.00068	0.01229	0	0	0	0	0	0	1	-360	360;
	3844	9307	0.001901	0.01163	0	0	0	0	0	0	1	-360	360;
	2575	804	0.00101	0.00824	0	0	0	0	0	0	1	-360	360;
	12996	8376	0.00787	0.068219	0	0	0	0	0	0	1	-360	360;
	12996	8376	0.007711	0.069539	0	0	0	0	0	0	1	-360	360;
	2300	8730	0.00032	0.002091	0	0	0	0	0	0	1	-360	360;
	4010	6756	0.00012	0.00182	0	0	0	0	0	0	1	-360	360;
	6756	8	0.00058	0.00736	0	0	0	0	0	0	1	-360	360;
	4010	2520	0.00074	0.01143	0	0	0	0	0	0	1	-360	360;
	10597	8107	0.00084	0.01378	0	0	0	0	0	0	1	-360	360;
	10597	10525	0.00037	0.00356	0	0	0	0	0	0	1	-360	360;
	10392	10597	0.00012	0.00206	0	0	0	0	0	0	1	-360	360;
	9718	10565	0.00036	0.00151	0	0	0	0	0	0	1	-360	360;
	1117	2159	0.0012	0.01835	0	0	0	0	0	0	1	-360	360;
	2159	12112	0.00088	0.01503	0	0	0	0	0	0	1	-360	360;
	3055	2159	0.00057	0.00939	0	0	0	0	0	0	1	-360	360;
	2196	5443	0.00157	0.02006	0	0	0	0	0	0	1	-360	360;
	2196	8242	0.00027	0.00349	0	0	0	0	0	0	1	-360	360;
	11223	2196	0.00136	0.02325	0	0	0	0	0	0	1	-360	360;
	2196	5725	0.00038	0.00478	0	0	0	0	0	0	1	-360	360;
	2196	9432	0.00077	0.01171	0	0	0	0	0	0	1	-360	360;
	1117	7968	0.00129	0.02076	0	0	0	0	0	0	1	-360	360;
	1117	2520	0.00037	0.0046	0	0	0	0	0	0	1	-360	360;
	2520	1117	0.00021	0.00375	0	0	0	0	0	0	1	-360	360;
	1117	3055	0.00055	0.00885	0	0	0	0	0	0	1	-360	360;
	9945	2575	0.001659	0.01138	0	0	0	0	0	0	1	-360	360;
	9945	7190	0.001169	0.010021	0	0	0	0	0	0	1	-360	360;
	9945	2017	0.002229	0.00975	0	0	0	0	0	0	1	-360	360;
	9945	8384	0.00049	0.004521	0	0	0	0	0	0	1	-360	360;
	9945	9723	0.0002	0.001969	0	0	0	0	0	0	1	-360	360;
	4302	7092	0.0001	0.00158	0	0	0	0	0	0	1	-360	360;
	4302	7092	0.00011	0.00151	0	0	0	0	0	0	1	-360	360;
	1700	7238	3e-05	0.00043	0	0	0	0	0	0	1	-360	360;
	1700	998	0.00037	0.00589	0	0	0	0	0	0	1	-360	360;
	2898	1700	0.00087	0.01218	0	0	0	0	0	0	1	-360	360;
	11430	8376	0.00362	0.024409	0	0	0	0	0	0	1	-360	360;
	11430	5625	0.00488	0.03119	0	0	0	0	0	0	1	-360	360;
	11772	6940	0.00055	0.00264	0	0	0	0	0	0	1	-360	360;
	8920	6940	0.00031	0.001341	0	0	0	0	0	0	1	-360	360;
	8678	6940	0.00012	0.000669	0	0	0	0	0	0	1	-360	360;
	1188	2407	3.1e-05	0.000469	0	0	0	0	0	0	1	-360	360;
	2837	2407	3.1e-05	0.000461	0	0	0	0	0	0	1	-360	360;
	2899	6600	0.00607	0.03457	0	0	0	0	0	0	1	-360	360;
	6600	6626	0.0087	0.047719	0	0	0	0	0	0	1	-360	360;
	5591	12119	0.00129	0.02159	0	0	0	0	0	0	1	-360	360;
	5591	10874	0.00091	0.00903	0	0	0	0	0	0	1	-360	360;
	12715	5591	0.0004	0.00518	0	0	0	0	0	0	1	-360	360;
	5591	9841	0.00032	0.00557	0	0	0	0	0	0	1	-360	360;
	10969	6917	3.1e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	10969	9046	3.1e-05	0.000211	0	0	0	0	0	0	1	-360	360;
	7934	3691	0.00199	0.011591	0	0	0	0	0	0	1	-360	360;
	5447	7934	0.000211	0.000919	0	0	0	0	0	0	1	-360	360;
	7796	3519	0.00109	0.01525	0	0	0	0	0	0	1	-360	360;
	1084	3691	0.005789	0.02826	0	0	0	0	0	0	1	-360	360;
	9558	4128	0.00041	0.0064	0	0	0	0	0	0	1	-360	360;
	9558	3843	0.00051	0.00603	0	0	0	0	0	0	1	-360	360;
	13023	6282	0.00245	0.02726	0	0	0	0	0	0	1	-360	360;
	13023	10499	0.000271	0.00337	0	0	0	0	0	0	1	-360	360;
	9332	10063	0.000151	0.00064	0	0	0	0	0	0	1	-360	360;
	2769	9332	0.003341	0.022091	0	0	0	0	0	0	1	-360	360;
	9332	9311	0.022	0.125771	0	0	0	0	0	0	1	-360	360;
	9332	13428	0.0153	0.08863	0	0	0	0	0	0	1	-360	360;
	9332	13428	0.015599	0.08682	0	0	0	0	0	0	1	-360	360;
	198	2891	0.001169	0.00844	0	0	0	0	0	0	1	-360	360;
	10565	198	0.00081	0.004271	0	0	0	0	0	0	1	-360	360;
	198	8615	0.000349	0.0015	0	0	0	0	0	0	1	-360	360;
	5443	9432	0.00058	0.00717	0	0	0	0	0	0	1	-360	360;
	5443	8192	0.0007	0.01203	0	0	0	0	0	0	1	-360	360;
	8192	5443	0.00061	0.00957	0	0	0	0	0	0	1	-360	360;
	12823	3722	2e-05	0.00075	0	0	0	0	0	0	1	-360	360;
	9752	12823	0.00041	0.01221	0	0	0	0	0	0	1	-360	360;
	1487	12879	0.00157	0.007969	0	0	0	0	0	0	1	-360	360;
	6232	12879	0.00151	0.00625	0	0	0	0	0	0	1	-360	360;
	6509	6513	2e-05	0.00023	0	0	0	0	0	0	1	-360	360;
	6509	6513	2e-05	0.00022	0	0	0	0	0	0	1	-360	360;
	5689	6509	0.00026	0.00395	0	0	0	0	0	0	1	-360	360;
	5689	6509	0.00025	0.00403	0	0	0	0	0	0	1	-360	360;
	13153	2112	0.00111	0.005099	0	0	0	0	0	0	1	-360	360;
	2112	6913	0.000729	0.0023	0	0	0	0	0	0	1	-360	360;
	6913	11830	0.003831	0.016271	0	0	0	0	0	0	1	-360	360;
	6554	13509	0.001849	0.014719	0	0	0	0	0	0	1	-360	360;
	10853	6554	0.00157	0.010169	0	0	0	0	0	0	1	-360	360;
	11428	9203	0.00149	0.01513	0	0	0	0	0	0	1	-360	360;
	9453	11428	0.003229	0.02918	0	0	0	0	0	0	1	-360	360;
	12119	9524	1e-05	0.00021	0	0	0	0	0	0	1	-360	360;
	3281	3126	0.00054	0.00765	0	0	0	0	0	0	1	-360	360;
	5927	8242	1e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	8242	9432	0.00053	0.00746	0	0	0	0	0	0	1	-360	360;
	4713	5720	0.010469	0.057979	0	0	0	0	0	0	1	-360	360;
	4713	3796	0.00014	0.000781	0	0	0	0	0	0	1	-360	360;
	4713	1253	0.00181	0.013039	0	0	0	0	0	0	1	-360	360;
	6078	12979	0.01107	0.04905	0	0	0	0	0	0	1	-360	360;
	3752	2297	0.00038	0.00468	0	0	0	0	0	0	1	-360	360;
	3752	2297	0.00034	0.0053	0	0	0	0	0	0	1	-360	360;
	4877	3752	0.00012	0.00185	0	0	0	0	0	0	1	-360	360;
	4877	3752	0.00013	0.00181	0	0	0	0	0	0	1	-360	360;
	7718	11372	0.000211	0.0018	0	0	0	0	0	0	1	-360	360;
	2108	2942	0.00037	0.00225	0	0	0	0	0	0	1	-360	360;
	7394	11591	0.00025	0.00317	0	0	0	0	0	0	1	-360	360;
	2403	7394	0.00029	0.00484	0	0	0	0	0	0	1	-360	360;
	142	10618	0.00057	0.00858	0	0	0	0	0	0	1	-360	360;
	142	13320	0.0005	0.00665	0	0	0	0	0	0	1	-360	360;
	860	4356	0.00239	0.012289	0	0	0	0	0	0	1	-360	360;
	9607	4356	0.0008	0.005211	0	0	0	0	0	0	1	-360	360;
	12112	5908	0.00067	0.01081	0	0	0	0	0	0	1	-360	360;
	3837	12112	0.00105	0.01455	0	0	0	0	0	0	1	-360	360;
	11421	10618	0.0002	0.00261	0	0	0	0	0	0	1	-360	360;
	8708	12890	1e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	10874	12890	0.00036	0.00601	0	0	0	0	0	0	1	-360	360;
	13320	12890	0.0001	0.00151	0	0	0	0	0	0	1	-360	360;
	6580	10148	2e-05	0.00021	0	0	0	0	0	0	1	-360	360;
	12119	10148	0.00046	0.00568	0	0	0	0	0	0	1	-360	360;
	11460	10148	0.00041	0.00531	0	0	0	0	0	0	1	-360	360;
	3824	3281	0.0003	0.00482	0	0	0	0	0	0	1	-360	360;
	2898	1934	1e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	11223	1623	1e-05	0.00022	0	0	0	0	0	0	1	-360	360;
	3875	12655	2e-05	0.00025	0	0	0	0	0	0	1	-360	360;
	3875	12655	2e-05	0.00026	0	0	0	0	0	0	1	-360	360;
	1241	7518	0.00032	0.00528	0	0	0	0	0	0	1	-360	360;
	4204	7518	0.0006	0.0084	0	0	0	0	0	0	1	-360	360;
	9458	11966	0.00105	0.01735	0	0	0	0	0	0	1	-360	360;
	6109	9458	0.00135	0.01758	0	0	0	0	0	0	1	-360	360;
	10618	11966	0.0004	0.0068	0	0	0	0	0	0	1	-360	360;
	10618	11966	0.00039	0.0063	0	0	0	0	0	0	1	-360	360;
	11966	13320	0.00067	0.00834	0	0	0	0	0	0	1	-360	360;
	11247	998	0.00084	0.01114	0	0	0	0	0	0	1	-360	360;
	2898	998	0.00081	0.011	0	0	0	0	0	0	1	-360	360;
	4515	4569	0.00059	0.00726	0	0	0	0	0	0	1	-360	360;
	6109	4569	0.001	0.01178	0	0	0	0	0	0	1	-360	360;
	6513	3355	0.00031	0.00197	0	0	0	0	0	0	1	-360	360;
	6513	3355	0.00035	0.00193	0	0	0	0	0	0	1	-360	360;
	9380	3355	2e-05	0.0004	0	0	0	0	0	0	1	-360	360;
	9380	3355	2e-05	0.00041	0	0	0	0	0	0	1	-360	360;
	7136	6695	3.1e-05	0.000271	0	0	0	0	0	0	1	-360	360;
	6695	7136	3.1e-05	0.000281	0	0	0	0	0	0	1	-360	360;
	6695	4688	0.000229	0.000271	0	0	0	0	0	0	1	-360	360;
	6338	8518	0.000289	0.001229	0	0	0	0	0	0	1	-360	360;
	6338	12979	0.002479	0.01106	0	0	0	0	0	0	1	-360	360;
	2476	1084	0.00738	0.04069	0	0	0	0	0	0	1	-360	360;
	6767	3518	0.00058	0.00872	0	0	0	0	0	0	1	-360	360;
	3518	11901	0.00116	0.01483	0	0	0	0	0	0	1	-360	360;
	3518	11901	0.00118	0.01452	0	0	0	0	0	0	1	-360	360;
	10618	4515	0.00139	0.02265	0	0	0	0	0	0	1	-360	360;
	10618	4515	0.00149	0.02349	0	0	0	0	0	0	1	-360	360;
	5938	4515	0.00086	0.01488	0	0	0	0	0	0	1	-360	360;
	5938	4515	0.00095	0.01554	0	0	0	0	0	0	1	-360	360;
	505	1102	0.019651	0.09944	0	0	0	0	0	0	1	-360	360;
	10333	1790	0.0002	0.001831	0	0	0	0	0	0	1	-360	360;
	13115	4038	7e-05	0.00062	0	0	0	0	0	0	1	-360	360;
	4038	17	0.004039	0.023081	0	0	0	0	0	0	1	-360	360;
	6626	4038	0.0132	0.064331	0	0	0	0	0	0	1	-360	360;
	4038	13325	0.001401	0.008729	0	0	0	0	0	0	1	-360	360;
	9432	13503	0.00021	0.00247	0	0	0	0	0	0	1	-360	360;
	9432	8192	0.00138	0.01791	0	0	0	0	0	0	1	-360	360;
	12119	11460	0.00056	0.00795	0	0	0	0	0	0	1	-360	360;
	12119	7772	0.00081	0.00788	0	0	0	0	0	0	1	-360	360;
	11153	6767	3e-05	0.00042	0	0	0	0	0	0	1	-360	360;
	2494	6767	3e-05	0.00042	0	0	0	0	0	0	1	-360	360;
	12643	6767	3e-05	0.00035	0	0	0	0	0	0	1	-360	360;
	9899	6767	3e-05	0.00036	0	0	0	0	0	0	1	-360	360;
	11591	6767	0.00035	0.00423	0	0	0	0	0	0	1	-360	360;
	4486	7461	0.00076	0.00904	0	0	0	0	0	0	1	-360	360;
	7461	437	0.00029	0.00361	0	0	0	0	0	0	1	-360	360;
	10383	8849	0.010349	0.056729	0	0	0	0	0	0	1	-360	360;
	10383	2891	0.001581	0.012771	0	0	0	0	0	0	1	-360	360;
	10383	1487	0.001081	0.01136	0	0	0	0	0	0	1	-360	360;
	10383	6232	0.001669	0.00914	0	0	0	0	0	0	1	-360	360;
	10383	8264	0.01757	0.07575	0	0	0	0	0	0	1	-360	360;
	1166	10525	0.00037	0.00558	0	0	0	0	0	0	1	-360	360;
	1166	12762	0.00041	0.0072	0	0	0	0	0	0	1	-360	360;
	11420	6707	0.002479	0.017961	0	0	0	0	0	0	1	-360	360;
	7061	11420	0.00155	0.0127	0	0	0	0	0	0	1	-360	360;
	201	11505	0.001961	0.010419	0	0	0	0	0	0	1	-360	360;
	3023	2891	0.00138	0.007521	0	0	0	0	0	0	1	-360	360;
	11939	3023	0.00281	0.02543	0	0	0	0	0	0	1	-360	360;
	3023	12757	0.000211	0.000901	0	0	0	0	0	0	1	-360	360;
	13225	5599	0.000789	0.00519	0	0	0	0	0	0	1	-360	360;
	13225	149	0.00056	0.004281	0	0	0	0	0	0	1	-360	360;
	13225	12256	0.00163	0.012031	0	0	0	0	0	0	1	-360	360;
	9453	13225	0.000211	0.00161	0	0	0	0	0	0	1	-360	360;
	13225	8297	0.00018	0.00144	0	0	0	0	0	0	1	-360	360;
	10730	2575	9.1e-05	0.00081	0	0	0	0	0	0	1	-360	360;
	12508	2575	9.1e-05	0.00099	0	0	0	0	0	0	1	-360	360;
	4868	7438	0.001219	0.01461	0	0	0	0	0	0	1	-360	360;
	4868	3129	0.00769	0.035169	0	0	0	0	0	0	1	-360	360;
	11152	4933	5e-05	0.000219	0	0	0	0	0	0	1	-360	360;
	4933	8615	0.000281	0.00112	0	0	0	0	0	0	1	-360	360;
	8348	6282	0.003591	0.01825	0	0	0	0	0	0	1	-360	360;
	8348	6282	0.00125	0.01386	0	0	0	0	0	0	1	-360	360;
	505	6282	0.000961	0.00513	0	0	0	0	0	0	1	-360	360;
	2297	6438	0.00022	0.00295	0	0	0	0	0	0	1	-360	360;
	8254	640	0.00326	0.02562	0	0	0	0	0	0	1	-360	360;
	8254	11100	0.00164	0.013219	0	0	0	0	0	0	1	-360	360;
	6118	1181	8.1e-05	0.00019	0	0	0	0	0	0	1	-360	360;
	6118	1181	3.9e-05	0.00032	0	0	0	0	0	0	1	-360	360;
	6118	1181	0.00011	0.0002	0	0	0	0	0	0	1	-360	360;
	6118	1181	0.000103	0.000171	0	0	0	0	0	0	1	-360	360;
	6118	1181	0.00011	0.00019	0	0	0	0	0	0	1	-360	360;
	6118	5599	8.1e-05	0.001	0	0	0	0	0	0	1	-360	360;
	6118	149	0.000281	0.00213	0	0	0	0	0	0	1	-360	360;
	9453	6118	0.00025	0.000831	0	0	0	0	0	0	1	-360	360;
	12181	6118	0.00013	0.000469	0	0	0	0	0	0	1	-360	360;
	4486	343	0.00088	0.01144	0	0	0	0	0	0	1	-360	360;
	3187	343	0.00076	0.01445	0	0	0	0	0	0	1	-360	360;
	3102	11247	0.00048	0.0076	0	0	0	0	0	0	1	-360	360;
	11247	3380	0.00017	0.00273	0	0	0	0	0	0	1	-360	360;
	13217	12414	0.00014	0.000849	0	0	0	0	0	0	1	-360	360;
	7309	13217	0.0022	0.01289	0	0	0	0	0	0	1	-360	360;
	5217	11550	0.00582	0.03749	0	0	0	0	0	0	1	-360	360;
	5217	11550	0.00593	0.03669	0	0	0	0	0	0	1	-360	360;
	11550	8448	0.000281	0.003419	0	0	0	0	0	0	1	-360	360;
	11550	6018	0.00645	0.03995	0	0	0	0	0	0	1	-360	360;
	11550	6018	0.00657	0.039081	0	0	0	0	0	0	1	-360	360;
	12378	13132	8e-05	0.00125	0	0	0	0	0	0	1	-360	360;
	12378	12938	8e-05	0.00127	0	0	0	0	0	0	1	-360	360;
	12378	6751	7e-05	0.0013	0	0	0	0	0	0	1	-360	360;
	12378	9366	7e-05	0.0012	0	0	0	0	0	0	1	-360	360;
	10553	8107	5e-05	0.0006	0	0	0	0	0	0	1	-360	360;
	8107	8	0.00034	0.00466	0	0	0	0	0	0	1	-360	360;
	10632	10565	0.000591	0.00588	0	0	0	0	0	0	1	-360	360;
	10565	10632	0.00139	0.00818	0	0	0	0	0	0	1	-360	360;
	6232	10632	0.00169	0.013969	0	0	0	0	0	0	1	-360	360;
	5819	502	0.000229	0.00163	0	0	0	0	0	0	1	-360	360;
	6388	7315	0.00275	0.02836	0	0	0	0	0	0	1	-360	360;
	70	7315	0.004909	0.02468	0	0	0	0	0	0	1	-360	360;
	7315	12979	0.001419	0.02212	0	0	0	0	0	0	1	-360	360;
	6688	12378	0.0016	0.02118	0	0	0	0	0	0	1	-360	360;
	12378	1309	0.00116	0.01869	0	0	0	0	0	0	1	-360	360;
	12006	12378	0.00169	0.02149	0	0	0	0	0	0	1	-360	360;
	12378	7772	0.00171	0.02913	0	0	0	0	0	0	1	-360	360;
	12378	9199	0.00138	0.0212	0	0	0	0	0	0	1	-360	360;
	3563	5423	0.00511	0.02775	0	0	0	0	0	0	1	-360	360;
	7796	6109	0.0008	0.01262	0	0	0	0	0	0	1	-360	360;
	7492	557	0.0005	0.00851	0	0	0	0	0	0	1	-360	360;
	7492	3876	0.00032	0.00558	0	0	0	0	0	0	1	-360	360;
	3856	8016	0.00039	0.00601	0	0	0	0	0	0	1	-360	360;
	3856	6619	0.00036	0.00575	0	0	0	0	0	0	1	-360	360;
	3856	11938	0.00117	0.01616	0	0	0	0	0	0	1	-360	360;
	1241	9008	0.0003	0.00471	0	0	0	0	0	0	1	-360	360;
	5217	1080	0.00011	0.00081	0	0	0	0	0	0	1	-360	360;
	5217	2300	0.01087	0.06457	0	0	0	0	0	0	1	-360	360;
	5217	2300	0.00968	0.073169	0	0	0	0	0	0	1	-360	360;
	2891	11152	0.00105	0.005781	0	0	0	0	0	0	1	-360	360;
	3080	4123	0.002961	0.029021	0	0	0	0	0	0	1	-360	360;
	9642	557	4e-05	0.00033	0	0	0	0	0	0	1	-360	360;
	3331	557	4e-05	0.00035	0	0	0	0	0	0	1	-360	360;
	7482	557	4e-05	0.0003	0	0	0	0	0	0	1	-360	360;
	4882	557	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	7190	128	0.000211	0.00045	0	0	0	0	0	0	1	-360	360;
	13596	5475	0.00024	0.001099	0	0	0	0	0	0	1	-360	360;
	6071	13596	0.00544	0.029831	0	0	0	0	0	0	1	-360	360;
	3777	10525	0.00076	0.0101	0	0	0	0	0	0	1	-360	360;
	2899	8500	0.01106	0.06313	0	0	0	0	0	0	1	-360	360;
	8286	8348	0.001909	0.011401	0	0	0	0	0	0	1	-360	360;
	9607	8286	0.00113	0.005479	0	0	0	0	0	0	1	-360	360;
	11772	5453	0.000151	0.000961	0	0	0	0	0	0	1	-360	360;
	11772	13428	0.00156	0.01125	0	0	0	0	0	0	1	-360	360;
	11772	13428	0.00344	0.021599	0	0	0	0	0	0	1	-360	360;
	1855	7438	0.00106	0.01276	0	0	0	0	0	0	1	-360	360;
	6384	8541	0.001349	0.008831	0	0	0	0	0	0	1	-360	360;
	5475	13115	0.001901	0.010031	0	0	0	0	0	0	1	-360	360;
	3172	10853	0.001719	0.012159	0	0	0	0	0	0	1	-360	360;
	4628	155	4e-05	0.00061	0	0	0	0	0	0	1	-360	360;
	11452	1658	0.00126	0.00595	0	0	0	0	0	0	1	-360	360;
	4204	8	0.00018	0.00224	0	0	0	0	0	0	1	-360	360;
	8920	9311	0.000521	0.002039	0	0	0	0	0	0	1	-360	360;
	11593	13115	0.002169	0.016479	0	0	0	0	0	0	1	-360	360;
	6688	8016	0.00073	0.01168	0	0	0	0	0	0	1	-360	360;
	12006	8016	0.00131	0.0162	0	0	0	0	0	0	1	-360	360;
	10969	3422	0.00112	0.007591	0	0	0	0	0	0	1	-360	360;
	8384	10969	0.00162	0.01332	0	0	0	0	0	0	1	-360	360;
	9079	12006	0.00027	0.00344	0	0	0	0	0	0	1	-360	360;
	2758	12006	0.0003	0.00313	0	0	0	0	0	0	1	-360	360;
	2601	2520	0.00014	0.00243	0	0	0	0	0	0	1	-360	360;
	2608	7086	0.00019	0.001781	0	0	0	0	0	0	1	-360	360;
	4613	7086	0.000151	0.001969	0	0	0	0	0	0	1	-360	360;
	2769	9311	0.025461	0.14561	0	0	0	0	0	0	1	-360	360;
	11100	640	0.004599	0.028581	0	0	0	0	0	0	1	-360	360;
	6024	2407	0.00987	0.05706	0	0	0	0	0	0	1	-360	360;
	155	3704	9e-05	0.00103	0	0	0	0	0	0	1	-360	360;
	155	3704	0.0001	0.00101	0	0	0	0	0	0	1	-360	360;
	3876	3704	0.00023	0.00381	0	0	0	0	0	0	1	-360	360;
	183	2891	0.002219	0.01374	0	0	0	0	0	0	1	-360	360;
	12737	4486	0.00092	0.00895	0	0	0	0	0	0	1	-360	360;
	183	10565	0.001151	0.009479	0	0	0	0	0	0	1	-360	360;
	1404	3129	0.00188	0.012151	0	0	0	0	0	0	1	-360	360;
	6559	2297	0.00036	0.00471	0	0	0	0	0	0	1	-360	360;
	6559	1332	0.00085	0.01337	0	0	0	0	0	0	1	-360	360;
	4368	6559	9e-05	0.00105	0	0	0	0	0	0	1	-360	360;
	4877	6559	0.00037	0.00634	0	0	0	0	0	0	1	-360	360;
	6018	4296	0.00056	0.004979	0	0	0	0	0	0	1	-360	360;
	5625	8376	0.00488	0.029039	0	0	0	0	0	0	1	-360	360;
	5625	8376	0.004349	0.032919	0	0	0	0	0	0	1	-360	360;
	3777	4128	0.00046	0.00571	0	0	0	0	0	0	1	-360	360;
	13509	8229	0.010841	0.04989	0	0	0	0	0	0	1	-360	360;
	4706	1404	0.00282	0.01436	0	0	0	0	0	0	1	-360	360;
	4706	1404	0.00282	0.014789	0	0	0	0	0	0	1	-360	360;
	389	3796	0.00012	0.000841	0	0	0	0	0	0	1	-360	360;
	389	1253	0.00175	0.01189	0	0	0	0	0	0	1	-360	360;
	389	2407	0.00851	0.04389	0	0	0	0	0	0	1	-360	360;
	12889	280	0.00063	0.00374	0	0	0	0	0	0	1	-360	360;
	5689	233	0.00037	0.00484	0	0	0	0	0	0	1	-360	360;
	12715	7772	0.00161	0.02178	0	0	0	0	0	0	1	-360	360;
	12694	2017	0.00745	0.03588	0	0	0	0	0	0	1	-360	360;
	2987	7305	0.00051	0.00829	0	0	0	0	0	0	1	-360	360;
	2987	11938	0.00124	0.01751	0	0	0	0	0	0	1	-360	360;
	2987	11591	0.0008	0.01242	0	0	0	0	0	0	1	-360	360;
	2520	7968	0.00118	0.01502	0	0	0	0	0	0	1	-360	360;
	5689	9752	0.00131	0.01723	0	0	0	0	0	0	1	-360	360;
	1790	8678	0.000271	0.001081	0	0	0	0	0	0	1	-360	360;
	10536	4302	0.00014	0.00178	0	0	0	0	0	0	1	-360	360;
	1790	9311	0.00051	0.00201	0	0	0	0	0	0	1	-360	360;
	11591	12029	9e-05	0.00091	0	0	0	0	0	0	1	-360	360;
	2766	692	9.1e-05	0.000591	0	0	0	0	0	0	1	-360	360;
	4688	7336	2.1e-05	0.00025	0	0	0	0	0	0	1	-360	360;
	2898	9380	0.00035	0.00556	0	0	0	0	0	0	1	-360	360;
	2898	9380	0.0004	0.00545	0	0	0	0	0	0	1	-360	360;
	1309	6636	0.0004	0.00517	0	0	0	0	0	0	1	-360	360;
	6636	7772	0.00037	0.00456	0	0	0	0	0	0	1	-360	360;
	7797	1309	0.00133	0.01413	0	0	0	0	0	0	1	-360	360;
	10325	6384	0.01187	0.05524	0	0	0	0	0	0	1	-360	360;
	6384	4613	0.00063	0.0065	0	0	0	0	0	0	1	-360	360;
	2407	6384	0.011841	0.052289	0	0	0	0	0	0	1	-360	360;
	11460	1921	0.00055	0.00517	0	0	0	0	0	0	1	-360	360;
	1921	7797	0.00097	0.01163	0	0	0	0	0	0	1	-360	360;
	2608	6574	0.000969	0.01039	0	0	0	0	0	0	1	-360	360;
	11939	9747	0.00326	0.015581	0	0	0	0	0	0	1	-360	360;
	3876	437	0.00099	0.01348	0	0	0	0	0	0	1	-360	360;
	11901	437	0.00045	0.00368	0	0	0	0	0	0	1	-360	360;
	1487	6232	0.002021	0.011331	0	0	0	0	0	0	1	-360	360;
	13475	1487	0.00076	0.0073	0	0	0	0	0	0	1	-360	360;
	3102	5938	0.00101	0.01505	0	0	0	0	0	0	1	-360	360;
	9303	1008	0	0.000347	0	0	0	0	0	0	1	-360	360;
	3837	5908	0.00065	0.00768	0	0	0	0	0	0	1	-360	360;
	7305	9008	0.00025	0.0035	0	0	0	0	0	0	1	-360	360;
	5599	9170	0.00013	0.000331	0	0	0	0	0	0	1	-360	360;
	149	9170	3.9e-05	0.00036	0	0	0	0	0	0	1	-360	360;
	6626	4123	0.011021	0.061901	0	0	0	0	0	0	1	-360	360;
	4123	8500	0.007901	0.04495	0	0	0	0	0	0	1	-360	360;
	7316	10565	0.00369	0.02186	0	0	0	0	0	0	1	-360	360;
	10565	9608	0.000281	0.001091	0	0	0	0	0	0	1	-360	360;
	10565	8615	0.000461	0.001841	0	0	0	0	0	0	1	-360	360;
	1778	5296	0.006271	0.02999	0	0	0	0	0	0	1	-360	360;
	6626	17	0.011219	0.05245	0	0	0	0	0	0	1	-360	360;
	12549	6388	0.000281	0.004091	0	0	0	0	0	0	1	-360	360;
	3875	5827	0.00041	0.00534	0	0	0	0	0	0	1	-360	360;
	10563	12256	0.002659	0.01557	0	0	0	0	0	0	1	-360	360;
	12979	70	0.00795	0.0483	0	0	0	0	0	0	1	-360	360;
	9723	4625	0.00037	0.00361	0	0	0	0	0	0	1	-360	360;
	2882	7718	0.001211	0.0082	0	0	0	0	0	0	1	-360	360;
	2300	13003	9.9e-05	0.00101	0	0	0	0	0	0	1	-360	360;
	2300	5096	9.9e-05	0.00099	0	0	0	0	0	0	1	-360	360;
	2300	7583	0.000169	0.001539	0	0	0	0	0	0	1	-360	360;
	2300	5643	0.000169	0.00157	0	0	0	0	0	0	1	-360	360;
	1253	2661	0.00011	0.000349	0	0	0	0	0	0	1	-360	360;
	1253	2661	0.00012	0.000341	0	0	0	0	0	0	1	-360	360;
	3837	1233	0.00017	0.00314	0	0	0	0	0	0	1	-360	360;
	590	5191	0.00012	0.00056	0	0	0	0	0	0	1	-360	360;
	2407	2229	0.000219	0.001469	0	0	0	0	0	0	1	-360	360;
	2407	2229	0.0002	0.001781	0	0	0	0	0	0	1	-360	360;
	12675	1332	0.00038	0.00623	0	0	0	0	0	0	1	-360	360;
	10563	10853	0.003669	0.01631	0	0	0	0	0	0	1	-360	360;
	1015	7797	0.00013	0.00117	0	0	0	0	0	0	1	-360	360;
	12395	7797	0.00012	0.0013	0	0	0	0	0	0	1	-360	360;
	10325	4351	0.00082	0.004969	0	0	0	0	0	0	1	-360	360;
	10325	4351	0.0008	0.004599	0	0	0	0	0	0	1	-360	360;
	2407	10325	0.00914	0.060289	0	0	0	0	0	0	1	-360	360;
	6626	11830	0.005219	0.02681	0	0	0	0	0	0	1	-360	360;
	6626	11830	0.005919	0.026271	0	0	0	0	0	0	1	-360	360;
	6626	8500	0.00381	0.02086	0	0	0	0	0	0	1	-360	360;
	9311	11071	0.004539	0.01957	0	0	0	0	0	0	1	-360	360;
	4007	9311	0.00112	0.00632	0	0	0	0	0	0	1	-360	360;
	9841	6482	0.00027	0.00426	0	0	0	0	0	0	1	-360	360;
	13320	6482	0.0002	0.00238	0	0	0	0	0	0	1	-360	360;
	1555	6482	8e-05	0.00077	0	0	0	0	0	0	1	-360	360;
	6626	7061	0.00149	0.012091	0	0	0	0	0	0	1	-360	360;
	13645	6232	0.00064	0.00356	0	0	0	0	0	0	1	-360	360;
	7309	9307	0.005729	0.033	0	0	0	0	0	0	1	-360	360;
	13325	12889	0.0017	0.007599	0	0	0	0	0	0	1	-360	360;
	11938	2550	0.00221	0.02589	0	0	0	0	0	0	1	-360	360;
	5453	1658	0.000331	0.000831	0	0	0	0	0	0	1	-360	360;
	9453	12181	9.9e-05	0.00045	0	0	0	0	0	0	1	-360	360;
	12757	6232	0.00039	0.00163	0	0	0	0	0	0	1	-360	360;
	12006	7772	0.00383	0.04628	0	0	0	0	0	0	1	-360	360;
	9607	10499	0.00106	0.01101	0	0	0	0	0	0	1	-360	360;
	13475	10707	0.00131	0.012659	0	0	0	0	0	0	1	-360	360;
	3837	8192	0.00052	0.00735	0	0	0	0	0	0	1	-360	360;
	8192	3837	0.00058	0.0074	0	0	0	0	0	0	1	-360	360;
	9199	3837	0.0013	0.01837	0	0	0	0	0	0	1	-360	360;
	6845	11349	3.9e-05	0.000599	0	0	0	0	0	0	1	-360	360;
	1778	11071	0.00031	0.001461	0	0	0	0	0	0	1	-360	360;
	4688	10175	0.001151	0.008771	0	0	0	0	0	0	1	-360	360;
	7136	6062	0.00095	0.00457	0	0	0	0	0	0	1	-360	360;
	7136	1847	0.001419	0.00982	0	0	0	0	0	0	1	-360	360;
	12757	9608	0.00037	0.00144	0	0	0	0	0	0	1	-360	360;
	10707	1855	0.00143	0.01225	0	0	0	0	0	0	1	-360	360;
	1573	11071	0.00044	0.00207	0	0	0	0	0	0	1	-360	360;
	3187	2550	0.00076	0.01509	0	0	0	0	0	0	1	-360	360;
	13428	4177	0.0007	0.004039	0	0	0	0	0	0	1	-360	360;
	7554	12846	0.004419	0.01918	0	0	0	0	0	0	1	-360	360;
	9747	5926	0.006099	0.029539	0	0	0	0	0	0	1	-360	360;
	11159	10796	0.00106	0.00513	0	0	0	0	0	0	1	-360	360;
	12995	13647	0.001021	0.004711	0	0	0	0	0	0	1	-360	360;
	2360	12894	0.000211	0.00106	0	0	0	0	0	0	1	-360	360;
	10887	12295	7e-05	0.00087	0	0	0	0	0	0	1	-360	360;
	10887	12894	3.9e-05	0.000651	0	0	0	0	0	0	1	-360	360;
	13647	13352	0.002039	0.008271	0	0	0	0	0	0	1	-360	360;
	13647	13352	0.002	0.00843	0	0	0	0	0	0	1	-360	360;
	5972	10796	0.001219	0.00725	0	0	0	0	0	0	1	-360	360;
	10796	13647	0.001151	0.006789	0	0	0	0	0	0	1	-360	360;
	5972	10796	0.0012	0.006711	0	0	0	0	0	0	1	-360	360;
	10796	13647	0.00112	0.006289	0	0	0	0	0	0	1	-360	360;
	11232	10481	0.002021	0.01101	0	0	0	0	0	0	1	-360	360;
	10481	5972	0.0015	0.008521	0	0	0	0	0	0	1	-360	360;
	10481	5972	0.001469	0.00868	0	0	0	0	0	0	1	-360	360;
	13647	12295	0.000719	0.00443	0	0	0	0	0	0	1	-360	360;
	13647	12894	0.0005	0.00268	0	0	0	0	0	0	1	-360	360;
	9076	2755	0.029149	0.051802	0	0	0	0	0	0	1	-360	360;
	8291	2755	0.035777	0.063587	0	0	0	0	0	0	1	-360	360;
	9076	4471	0.025967	0.058702	0	0	0	0	0	0	1	-360	360;
	4471	10817	0.01943	0.043942	0	0	0	0	0	0	1	-360	360;
	4471	2674	0.05776	0.115802	0	0	0	0	0	0	1	-360	360;
	316	4471	0.027388	0.054917	0	0	0	0	0	0	1	-360	360;
	10817	2674	0.028438	0.104719	0	0	0	0	0	0	1	-360	360;
	7591	7344	0.008091	0.03755	0	0	0	0	0	0	1	-360	360;
	12534	7344	0.00556	0.02913	0	0	0	0	0	0	1	-360	360;
	174	316	0.033512	0.069909	0	0	0	0	0	0	1	-360	360;
	12534	11008	0.00587	0.02949	0	0	0	0	0	0	1	-360	360;
	3901	8358	0.013314	0.040719	0	0	0	0	0	0	1	-360	360;
	3901	8358	0.01357	0.039876	0	0	0	0	0	0	1	-360	360;
	3901	8358	0.014579	0.041157	0	0	0	0	0	0	1	-360	360;
	3901	3174	0.031298	0.112364	0	0	0	0	0	0	1	-360	360;
	3901	3174	0.030653	0.114521	0	0	0	0	0	0	1	-360	360;
	372	12265	0.00537	0.04394	0	0	0	0	0	0	1	-360	360;
	372	12265	0.005461	0.04726	0	0	0	0	0	0	1	-360	360;
	7591	372	0.01312	0.074419	0	0	0	0	0	0	1	-360	360;
	8358	6142	0	0.000413	0	0	0	0	0	0	1	-360	360;
	6142	9503	0.052702	0.121281	0	0	0	0	0	0	1	-360	360;
	8452	4750	0.00103	0.01397	0	0	0	0	0	0	1	-360	360;
	3174	2674	0.002959	0.014851	0	0	0	0	0	0	1	-360	360;
	3174	2674	0.002769	0.016339	0	0	0	0	0	0	1	-360	360;
	8452	11249	0.00167	0.02105	0	0	0	0	0	0	1	-360	360;
	8291	2674	0.067826	0.163702	0	0	0	0	0	0	1	-360	360;
	5329	3458	0.009463	0.038388	0	0	0	0	0	0	1	-360	360;
	5577	3458	0.021289	0.124008	0	0	0	0	0	0	1	-360	360;
	8596	8059	0.035587	0.104182	0	0	0	0	0	0	1	-360	360;
	10532	8059	0.040678	0.119107	0	0	0	0	0	0	1	-360	360;
	10744	5821	0.00156	0.01963	0	0	0	0	0	0	1	-360	360;
	10744	4747	0.00255	0.02314	0	0	0	0	0	0	1	-360	360;
	10278	550	0.01305	0.035686	0	0	0	0	0	0	1	-360	360;
	10278	550	0.013314	0.034959	0	0	0	0	0	0	1	-360	360;
	2909	550	0.005612	0.022636	0	0	0	0	0	0	1	-360	360;
	92	550	0.029826	0.113909	0	0	0	0	0	0	1	-360	360;
	9861	550	0.04824	0.109727	0	0	0	0	0	0	1	-360	360;
	10278	11021	0.010802	0.04124	0	0	0	0	0	0	1	-360	360;
	2909	10278	0.021851	0.050934	0	0	0	0	0	0	1	-360	360;
	5156	10278	0.070223	0.192198	0	0	0	0	0	0	1	-360	360;
	5156	10278	0.071603	0.188281	0	0	0	0	0	0	1	-360	360;
	2909	11021	0.008339	0.021066	0	0	0	0	0	0	1	-360	360;
	3057	2909	0.024744	0.092099	0	0	0	0	0	0	1	-360	360;
	4247	5690	0.029579	0.114661	0	0	0	0	0	0	1	-360	360;
	10532	5690	0.009298	0.035488	0	0	0	0	0	0	1	-360	360;
	6908	5690	0.020149	0.078107	0	0	0	0	0	0	1	-360	360;
	9743	8596	0.017826	0.063777	0	0	0	0	0	0	1	-360	360;
	4747	5821	0.00158	0.02095	0	0	0	0	0	0	1	-360	360;
	494	5821	0.00076	0.00957	0	0	0	0	0	0	1	-360	360;
	6974	4247	0.005678	0.044198	0	0	0	0	0	0	1	-360	360;
	6974	4247	0.006438	0.043314	0	0	0	0	0	0	1	-360	360;
	9510	4247	0.013612	0.043157	0	0	0	0	0	0	1	-360	360;
	1276	4247	0.056248	0.147934	0	0	0	0	0	0	1	-360	360;
	1276	4247	0.057339	0.144851	0	0	0	0	0	0	1	-360	360;
	12482	4247	0.000463	0.003603	0	0	0	0	0	0	1	-360	360;
	4247	11050	0.037107	0.138157	0	0	0	0	0	0	1	-360	360;
	6974	11941	0.004058	0.015107	0	0	0	0	0	0	1	-360	360;
	9615	1817	0.011793	0.034521	0	0	0	0	0	0	1	-360	360;
	2939	1817	0.019636	0.073132	0	0	0	0	0	0	1	-360	360;
	6315	2474	0	0.000413	0	0	0	0	0	0	1	-360	360;
	5329	1286	0.014339	0.041983	0	0	0	0	0	0	1	-360	360;
	7588	1286	0.029314	0.118281	0	0	0	0	0	0	1	-360	360;
	12533	3898	0.016636	0.064488	0	0	0	0	0	0	1	-360	360;
	9355	3898	0.034273	0.127612	0	0	0	0	0	0	1	-360	360;
	3898	5577	0.00886	0.02495	0	0	0	0	0	0	1	-360	360;
	12533	9490	0.026587	0.084364	0	0	0	0	0	0	1	-360	360;
	9355	12446	0.039298	0.115058	0	0	0	0	0	0	1	-360	360;
	9355	9615	0.036066	0.134289	0	0	0	0	0	0	1	-360	360;
	9355	2417	0.000529	0.001554	0	0	0	0	0	0	1	-360	360;
	92	9510	0.03981	0.160628	0	0	0	0	0	0	1	-360	360;
	11941	9510	0.008438	0.025744	0	0	0	0	0	0	1	-360	360;
	9510	5571	0.024223	0.093876	0	0	0	0	0	0	1	-360	360;
	9490	6899	0.019868	0.054017	0	0	0	0	0	0	1	-360	360;
	9861	6899	0.051248	0.121347	0	0	0	0	0	0	1	-360	360;
	8150	8084	0.042793	0.101314	0	0	0	0	0	0	1	-360	360;
	8150	8084	0.030091	0.099372	0	0	0	0	0	0	1	-360	360;
	3256	8084	0.005488	0.020959	0	0	0	0	0	0	1	-360	360;
	2925	7368	0.009198	0.025851	0	0	0	0	0	0	1	-360	360;
	7368	8946	0.019752	0.057843	0	0	0	0	0	0	1	-360	360;
	2939	7368	0.032934	0.096413	0	0	0	0	0	0	1	-360	360;
	9826	12554	0.05586	0.112686	0	0	0	0	0	0	1	-360	360;
	8606	12554	0.029744	0.057636	0	0	0	0	0	0	1	-360	360;
	5247	9826	0.004868	0.013686	0	0	0	0	0	0	1	-360	360;
	612	9826	0.005413	0.015207	0	0	0	0	0	0	1	-360	360;
	9615	9826	0.043669	0.084653	0	0	0	0	0	0	1	-360	360;
	612	5247	0.004818	0.01724	0	0	0	0	0	0	1	-360	360;
	59	5247	0.017719	0.054017	0	0	0	0	0	0	1	-360	360;
	9615	5727	0.004364	0.017719	0	0	0	0	0	0	1	-360	360;
	5727	1776	0.007463	0.041727	0	0	0	0	0	0	1	-360	360;
	2925	12446	0.016289	0.065711	0	0	0	0	0	0	1	-360	360;
	3256	9743	0.02014	0.072066	0	0	0	0	0	0	1	-360	360;
	4341	12460	0.003091	0.011826	0	0	0	0	0	0	1	-360	360;
	5156	12460	0.00386	0.013793	0	0	0	0	0	0	1	-360	360;
	3057	4341	0.036	0.114223	0	0	0	0	0	0	1	-360	360;
	7588	5156	0.027942	0.081793	0	0	0	0	0	0	1	-360	360;
	8150	6884	0.027008	0.061041	0	0	0	0	0	0	1	-360	360;
	10655	6884	0.025504	0.072198	0	0	0	0	0	0	1	-360	360;
	8946	9490	0.01686	0.060331	0	0	0	0	0	0	1	-360	360;
	3774	1776	0.010091	0.038529	0	0	0	0	0	0	1	-360	360;
	10655	3774	0.027099	0.082603	0	0	0	0	0	0	1	-360	360;
	6897	8036	0.014463	0.051719	0	0	0	0	0	0	1	-360	360;
	8036	5577	0.014612	0.041066	0	0	0	0	0	0	1	-360	360;
	2612	92	0.001661	0.006446	0	0	0	0	0	0	1	-360	360;
	10679	92	0.005198	0.019331	0	0	0	0	0	0	1	-360	360;
	9910	2612	0.003512	0.011587	0	0	0	0	0	0	1	-360	360;
	9910	10679	0.004339	0.015512	0	0	0	0	0	0	1	-360	360;
	8150	8606	0.043628	0.081231	0	0	0	0	0	0	1	-360	360;
	3887	10532	0.031661	0.092669	0	0	0	0	0	0	1	-360	360;
	1776	59	0.017562	0.070818	0	0	0	0	0	0	1	-360	360;
	2939	12446	0.008843	0.033802	0	0	0	0	0	0	1	-360	360;
	10542	13118	0.01164	0.05205	0	0	0	0	0	0	1	-360	360;
	1776	9615	0.007934	0.042653	0	0	0	0	0	0	1	-360	360;
	13257	9615	0.00343	0.014488	0	0	0	0	0	0	1	-360	360;
	175	9615	0.004339	0.022421	0	0	0	0	0	0	1	-360	360;
	9615	6315	0.00143	0.007149	0	0	0	0	0	0	1	-360	360;
	9615	6315	0.001446	0.007	0	0	0	0	0	0	1	-360	360;
	9615	6315	0.001504	0.007256	0	0	0	0	0	0	1	-360	360;
	9615	6315	0.001529	0.007099	0	0	0	0	0	0	1	-360	360;
	1776	2939	0.015653	0.051669	0	0	0	0	0	0	1	-360	360;
	6182	4747	0.00046	0.00424	0	0	0	0	0	0	1	-360	360;
	5329	5577	0.049339	0.138727	0	0	0	0	0	0	1	-360	360;
	6908	8150	0.005421	0.017876	0	0	0	0	0	0	1	-360	360;
	1276	3887	0.001678	0.006421	0	0	0	0	0	0	1	-360	360;
	1276	11050	0.024512	0.055752	0	0	0	0	0	0	1	-360	360;
	5700	2417	0.019347	0.072058	0	0	0	0	0	0	1	-360	360;
	6897	5700	0.002364	0.008777	0	0	0	0	0	0	1	-360	360;
	5700	5577	0.032331	0.102537	0	0	0	0	0	0	1	-360	360;
	175	13257	0.00305	0.017711	0	0	0	0	0	0	1	-360	360;
	8004	10655	0.005289	0.009463	0	0	0	0	0	0	1	-360	360;
	8004	6502	0.00181	0.007298	0	0	0	0	0	0	1	-360	360;
	6502	10655	0.000893	0.00338	0	0	0	0	0	0	1	-360	360;
	5571	6502	0.012033	0.046579	0	0	0	0	0	0	1	-360	360;
	8737	2409	0.00017	0.01734	0	0	0	0	0	0	1	-360	360;
	6151	1593	0.00017	0.01767	0	0	0	0	0	0	1	-360	360;
	13039	2887	0.0006	0.040218	0	0	0	0	0	0	1	-360	360;
	13074	5030	0.00076	0.002409	0	0	0	0	0	0	1	-360	360;
	13074	5539	0	0.000538	0	0	0	0	0	0	1	-360	360;
	9360	2232	0.00528	0.027569	0	0	0	0	0	0	1	-360	360;
	9360	2232	0.005382	0.026978	0	0	0	0	0	0	1	-360	360;
	7663	9360	0.00104	0.009649	0	0	0	0	0	0	1	-360	360;
	7663	9360	0.001062	0.009449	0	0	0	0	0	0	1	-360	360;
	12145	7653	0.000289	0.000889	0	0	0	0	0	0	1	-360	360;
	1716	7653	0.003502	0.01524	0	0	0	0	0	0	1	-360	360;
	12145	11700	0.000289	0.00088	0	0	0	0	0	0	1	-360	360;
	1716	11700	0.003431	0.014111	0	0	0	0	0	0	1	-360	360;
	6233	209	0.001502	0.1272	0	0	0	0	0	0	1	-360	360;
	12901	1047	8e-05	0.000729	0	0	0	0	0	0	1	-360	360;
	1716	1047	0.003818	0.012911	0	0	0	0	0	0	1	-360	360;
	13431	1047	0.001618	0.013711	0	0	0	0	0	0	1	-360	360;
	4906	2668	0.001062	0.007791	0	0	0	0	0	0	1	-360	360;
	9568	2668	0.001809	0.008951	0	0	0	0	0	0	1	-360	360;
	8418	2668	0.007142	0.020311	0	0	0	0	0	0	1	-360	360;
	4906	12311	0.000942	0.008831	0	0	0	0	0	0	1	-360	360;
	9568	12311	0.001769	0.008289	0	0	0	0	0	0	1	-360	360;
	8418	12311	0.00728	0.019898	0	0	0	0	0	0	1	-360	360;
	13070	8340	0.003569	0.017951	0	0	0	0	0	0	1	-360	360;
	5022	13070	0.001991	0.007271	0	0	0	0	0	0	1	-360	360;
	11290	13070	0.000391	0.001462	0	0	0	0	0	0	1	-360	360;
	3758	12017	0.003502	0.018289	0	0	0	0	0	0	1	-360	360;
	370	3758	0.001769	0.00824	0	0	0	0	0	0	1	-360	360;
	11290	3758	0.0004	0.001431	0	0	0	0	0	0	1	-360	360;
	2585	6946	0	0.000413	0	0	0	0	0	0	1	-360	360;
	5910	11011	0.001471	0.1296	0	0	0	0	0	0	1	-360	360;
	5030	10815	0.00056	0.015018	0	0	0	0	0	0	1	-360	360;
	236	12520	0.000631	0.01472	0	0	0	0	0	0	1	-360	360;
	236	5030	0	0.001169	0	0	0	0	0	0	1	-360	360;
	236	12436	0.000942	0.003342	0	0	0	0	0	0	1	-360	360;
	236	3874	0.001582	0.002831	0	0	0	0	0	0	1	-360	360;
	4508	7794	0.0008	0.003422	0	0	0	0	0	0	1	-360	360;
	4508	7794	0.000822	0.003351	0	0	0	0	0	0	1	-360	360;
	12436	3874	0.002631	0.004889	0	0	0	0	0	0	1	-360	360;
	12436	3287	0.005418	0.022489	0	0	0	0	0	0	1	-360	360;
	12436	3287	0.00552	0.022018	0	0	0	0	0	0	1	-360	360;
	10839	11072	0.004091	0.020893	0	0	0	0	0	0	1	-360	360;
	10839	11072	0.004	0.019339	0	0	0	0	0	0	1	-360	360;
	10839	6483	0.016479	0.037058	0	0	0	0	0	0	1	-360	360;
	10839	6872	0.016347	0.05124	0	0	0	0	0	0	1	-360	360;
	8406	11072	0.007058	0.030587	0	0	0	0	0	0	1	-360	360;
	8406	11072	0.007149	0.029934	0	0	0	0	0	0	1	-360	360;
	9967	9063	0.002089	0.003449	0	0	0	0	0	0	1	-360	360;
	9967	9063	0.002129	0.003378	0	0	0	0	0	0	1	-360	360;
	3324	1033	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1033	6264	0.007489	0.0516	0	0	0	0	0	0	1	-360	360;
	1033	6264	0.008489	0.050569	0	0	0	0	0	0	1	-360	360;
	1236	4640	0	0.047951	0	0	0	0	0	0	1	-360	360;
	8098	4640	0.001649	0.016591	0	0	0	0	0	0	1	-360	360;
	11519	4640	0.001618	0.016898	0	0	0	0	0	0	1	-360	360;
	5119	9063	0.006871	0.039911	0	0	0	0	0	0	1	-360	360;
	5119	9063	0.007009	0.039089	0	0	0	0	0	0	1	-360	360;
	5119	4601	0.001658	0.011871	0	0	0	0	0	0	1	-360	360;
	5119	4601	0.001631	0.010991	0	0	0	0	0	0	1	-360	360;
	7898	13483	0.00092	0.004689	0	0	0	0	0	0	1	-360	360;
	190	13483	0.002049	0.0038	0	0	0	0	0	0	1	-360	360;
	9249	13483	0.005649	0.026	0	0	0	0	0	0	1	-360	360;
	7898	4391	0.000902	0.004782	0	0	0	0	0	0	1	-360	360;
	417	4391	0.003471	0.014471	0	0	0	0	0	0	1	-360	360;
	190	4391	0.002089	0.00372	0	0	0	0	0	0	1	-360	360;
	6147	2682	0.003569	0.010671	0	0	0	0	0	0	1	-360	360;
	5733	6147	0.019782	0.061658	0	0	0	0	0	0	1	-360	360;
	714	8340	0.000409	0.0014	0	0	0	0	0	0	1	-360	360;
	8340	792	0.003302	0.017311	0	0	0	0	0	0	1	-360	360;
	714	12017	0.000409	0.001369	0	0	0	0	0	0	1	-360	360;
	12017	792	0.003231	0.016031	0	0	0	0	0	0	1	-360	360;
	13626	12481	0.000751	0.005711	0	0	0	0	0	0	1	-360	360;
	13626	13296	0.0018	0.00716	0	0	0	0	0	0	1	-360	360;
	1667	13626	0.001978	0.013462	0	0	0	0	0	0	1	-360	360;
	13626	2763	0.001502	0.0056	0	0	0	0	0	0	1	-360	360;
	4785	9403	0.005991	0.0284	0	0	0	0	0	0	1	-360	360;
	2918	9403	0.000302	0.000858	0	0	0	0	0	0	1	-360	360;
	9157	9403	0.002622	0.013231	0	0	0	0	0	0	1	-360	360;
	10451	8621	0.004271	0.019502	0	0	0	0	0	0	1	-360	360;
	2918	8621	0.000271	0.000969	0	0	0	0	0	0	1	-360	360;
	9157	8621	0.00356	0.014378	0	0	0	0	0	0	1	-360	360;
	10562	9850	0.0022	0.009569	0	0	0	0	0	0	1	-360	360;
	3935	9850	0.000311	0.001169	0	0	0	0	0	0	1	-360	360;
	4817	9850	0.001889	0.006311	0	0	0	0	0	0	1	-360	360;
	3935	4785	0.000271	0.00132	0	0	0	0	0	0	1	-360	360;
	4817	4785	0.001929	0.006178	0	0	0	0	0	0	1	-360	360;
	6977	9653	0.000339	0.001149	0	0	0	0	0	0	1	-360	360;
	694	10631	0.002591	0.014418	0	0	0	0	0	0	1	-360	360;
	694	1120	0	0.00028	0	0	0	0	0	0	1	-360	360;
	9295	5022	0.0014	0.00652	0	0	0	0	0	0	1	-360	360;
	7696	5022	0.000418	0.001338	0	0	0	0	0	0	1	-360	360;
	9295	370	0.001369	0.006649	0	0	0	0	0	0	1	-360	360;
	7696	370	0.000369	0.00152	0	0	0	0	0	0	1	-360	360;
	13039	10305	0.004689	0.018151	0	0	0	0	0	0	1	-360	360;
	13039	10305	0.004591	0.018498	0	0	0	0	0	0	1	-360	360;
	1775	202	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	1775	5843	0.003698	0.0198	0	0	0	0	0	0	1	-360	360;
	2815	1775	0.005471	0.0206	0	0	0	0	0	0	1	-360	360;
	2815	1775	0.0062	0.020191	0	0	0	0	0	0	1	-360	360;
	1775	5048	0.006098	0.030551	0	0	0	0	0	0	1	-360	360;
	1337	8418	0.004809	0.019751	0	0	0	0	0	0	1	-360	360;
	1337	8418	0.004711	0.018289	0	0	0	0	0	0	1	-360	360;
	12145	8811	0	0.00028	0	0	0	0	0	0	1	-360	360;
	10562	10451	0.007929	0.044022	0	0	0	0	0	0	1	-360	360;
	8826	1917	0.009782	0.033938	0	0	0	0	0	0	1	-360	360;
	8826	9596	0.003609	0.016889	0	0	0	0	0	0	1	-360	360;
	952	1974	0.00044	0.027591	0	0	0	0	0	0	1	-360	360;
	9983	5189	0.000419	0.027979	0	0	0	0	0	0	1	-360	360;
	522	5189	0.000409	0.02851	0	0	0	0	0	0	1	-360	360;
	1974	5189	0	0.000461	0	0	0	0	0	0	1	-360	360;
	1974	11313	0.000531	0.005581	0	0	0	0	0	0	1	-360	360;
	1508	1974	0.003841	0.022461	0	0	0	0	0	0	1	-360	360;
	5189	199	0.000539	0.005461	0	0	0	0	0	0	1	-360	360;
	5693	5189	0.00376	0.022901	0	0	0	0	0	0	1	-360	360;
	7318	13121	0.028843	0.060992	0	0	0	0	0	0	1	-360	360;
	7318	13121	0.029397	0.059744	0	0	0	0	0	0	1	-360	360;
	2585	9277	0.001727	0.00924	0	0	0	0	0	0	1	-360	360;
	11741	2585	0.004851	0.02724	0	0	0	0	0	0	1	-360	360;
	1917	9596	0.0138	0.040249	0	0	0	0	0	0	1	-360	360;
	1917	939	0.007098	0.022329	0	0	0	0	0	0	1	-360	360;
	1917	2618	0.001902	0.008391	0	0	0	0	0	0	1	-360	360;
	1917	2618	0.001929	0.008209	0	0	0	0	0	0	1	-360	360;
	2386	1917	0.001649	0.01228	0	0	0	0	0	0	1	-360	360;
	2386	1917	0.001471	0.01392	0	0	0	0	0	0	1	-360	360;
	1917	11165	0.000582	0.040991	0	0	0	0	0	0	1	-360	360;
	1917	464	0.01116	0.033671	0	0	0	0	0	0	1	-360	360;
	10226	3561	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7325	10030	0.00031	0.00312	0	0	0	0	0	0	1	-360	360;
	8845	10030	0.00069	0.00769	0	0	0	0	0	0	1	-360	360;
	8979	7210	0.010471	0.029289	0	0	0	0	0	0	1	-360	360;
	8979	7210	0.010671	0.028671	0	0	0	0	0	0	1	-360	360;
	507	8979	0.00624	0.021871	0	0	0	0	0	0	1	-360	360;
	507	8979	0.007071	0.021431	0	0	0	0	0	0	1	-360	360;
	10591	11365	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	10591	9596	0.001062	0.003351	0	0	0	0	0	0	1	-360	360;
	10591	10306	0.002111	0.02008	0	0	0	0	0	0	1	-360	360;
	10591	9365	0.001009	0.010929	0	0	0	0	0	0	1	-360	360;
	2386	10591	0.004711	0.050609	0	0	0	0	0	0	1	-360	360;
	10591	13431	0.011462	0.02396	0	0	0	0	0	0	1	-360	360;
	5910	3363	0	0.0444	0	0	0	0	0	0	1	-360	360;
	5910	1614	0.001698	0.1128	0	0	0	0	0	0	1	-360	360;
	5910	8589	8.9e-05	0.000742	0	0	0	0	0	0	1	-360	360;
	5910	8589	8.9e-05	0.000729	0	0	0	0	0	0	1	-360	360;
	5910	4019	0.000302	0.000809	0	0	0	0	0	0	1	-360	360;
	5910	11783	0.001502	0.002182	0	0	0	0	0	0	1	-360	360;
	5910	8990	0.001311	0.001871	0	0	0	0	0	0	1	-360	360;
	5910	8990	0.001369	0.00204	0	0	0	0	0	0	1	-360	360;
	5910	10137	0.001049	0.00308	0	0	0	0	0	0	1	-360	360;
	11741	8587	0.014967	0.09105	0	0	0	0	0	0	1	-360	360;
	7725	8587	0.028959	0.12905	0	0	0	0	0	0	1	-360	360;
	3555	8587	0.006438	0.041256	0	0	0	0	0	0	1	-360	360;
	3555	8587	0.006554	0.040364	0	0	0	0	0	0	1	-360	360;
	3363	1781	0	0.012742	0	0	0	0	0	0	1	-360	360;
	1667	12481	0.001018	0.00868	0	0	0	0	0	0	1	-360	360;
	1667	13296	0.001871	0.00548	0	0	0	0	0	0	1	-360	360;
	13296	10948	0.00228	0.004338	0	0	0	0	0	0	1	-360	360;
	13296	10948	0.00224	0.004431	0	0	0	0	0	0	1	-360	360;
	4355	1155	0.00076	0.00747	0	0	0	0	0	0	1	-360	360;
	8261	1155	0.00103	0.01061	0	0	0	0	0	0	1	-360	360;
	3197	13291	0.002689	0.013818	0	0	0	0	0	0	1	-360	360;
	792	3197	0.002111	0.003769	0	0	0	0	0	0	1	-360	360;
	792	3197	0.002151	0.003689	0	0	0	0	0	0	1	-360	360;
	13039	7898	0.00176	0.012702	0	0	0	0	0	0	1	-360	360;
	13039	7898	0.001729	0.012951	0	0	0	0	0	0	1	-360	360;
	8351	1501	0.000151	0.000511	0	0	0	0	0	0	1	-360	360;
	8351	1501	0.000151	0.000502	0	0	0	0	0	0	1	-360	360;
	8351	2618	0.01076	0.05508	0	0	0	0	0	0	1	-360	360;
	8351	2618	0.010542	0.05616	0	0	0	0	0	0	1	-360	360;
	8351	9968	0.003089	0.009	0	0	0	0	0	0	1	-360	360;
	8351	9968	0.003031	0.009169	0	0	0	0	0	0	1	-360	360;
	8351	9157	0.001849	0.00872	0	0	0	0	0	0	1	-360	360;
	8351	9157	0.00188	0.008538	0	0	0	0	0	0	1	-360	360;
	4281	4817	0.004151	0.01556	0	0	0	0	0	0	1	-360	360;
	2386	4281	0.006591	0.026778	0	0	0	0	0	0	1	-360	360;
	10226	13140	0.00056	0.038671	0	0	0	0	0	0	1	-360	360;
	2493	871	0	0.00045	0	0	0	0	0	0	1	-360	360;
	871	1593	0.00162	0.01968	0	0	0	0	0	0	1	-360	360;
	10226	6671	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	10226	9295	0.002649	0.015111	0	0	0	0	0	0	1	-360	360;
	10226	9295	0.002702	0.014809	0	0	0	0	0	0	1	-360	360;
	10226	7362	0.0008	0.005711	0	0	0	0	0	0	1	-360	360;
	10226	7362	0.000782	0.005809	0	0	0	0	0	0	1	-360	360;
	10226	4815	0.001049	0.007738	0	0	0	0	0	0	1	-360	360;
	10226	4815	0.001071	0.007569	0	0	0	0	0	0	1	-360	360;
	10226	4508	0.005138	0.025338	0	0	0	0	0	0	1	-360	360;
	10226	4508	0.005031	0.025831	0	0	0	0	0	0	1	-360	360;
	10226	2074	0.003658	0.029071	0	0	0	0	0	0	1	-360	360;
	10226	2074	0.004151	0.028489	0	0	0	0	0	0	1	-360	360;
	1730	1915	0.001471	0.011471	0	0	0	0	0	0	1	-360	360;
	1730	1915	0.001502	0.011218	0	0	0	0	0	0	1	-360	360;
	3287	1915	0.001871	0.014511	0	0	0	0	0	0	1	-360	360;
	3287	1915	0.001902	0.0142	0	0	0	0	0	0	1	-360	360;
	287	10546	0.000351	0.004662	0	0	0	0	0	0	1	-360	360;
	65	287	0.00772	0.033818	0	0	0	0	0	0	1	-360	360;
	7268	12422	0	0.00051	0	0	0	0	0	0	1	-360	360;
	12422	2719	0.00076	0.0119	0	0	0	0	0	0	1	-360	360;
	12422	4886	0.00025	0.00316	0	0	0	0	0	0	1	-360	360;
	7268	5101	0.00077	0.01164	0	0	0	0	0	0	1	-360	360;
	7268	9879	0.00025	0.00322	0	0	0	0	0	0	1	-360	360;
	6959	10224	0.001369	0.021182	0	0	0	0	0	0	1	-360	360;
	2719	5101	0	0.00052	0	0	0	0	0	0	1	-360	360;
	2719	11660	0.00015	0.00328	0	0	0	0	0	0	1	-360	360;
	583	2719	0.00016	0.00314	0	0	0	0	0	0	1	-360	360;
	5101	8845	0.00017	0.00321	0	0	0	0	0	0	1	-360	360;
	8504	5101	0.00018	0.00308	0	0	0	0	0	0	1	-360	360;
	4653	8478	0.00024	0.01948	0	0	0	0	0	0	1	-360	360;
	2386	11098	0.000631	0.037902	0	0	0	0	0	0	1	-360	360;
	3408	10416	0.00244	0.01532	0	0	0	0	0	0	1	-360	360;
	424	11893	0	0.00053	0	0	0	0	0	0	1	-360	360;
	424	8845	0.00013	0.00282	0	0	0	0	0	0	1	-360	360;
	5998	424	0.00038	0.00708	0	0	0	0	0	0	1	-360	360;
	11893	11660	0.00013	0.00261	0	0	0	0	0	0	1	-360	360;
	8583	11893	0.00038	0.00722	0	0	0	0	0	0	1	-360	360;
	2101	7663	0.00016	0.002769	0	0	0	0	0	0	1	-360	360;
	5822	7663	0.00016	0.00272	0	0	0	0	0	0	1	-360	360;
	11831	7663	0.00016	0.002662	0	0	0	0	0	0	1	-360	360;
	2834	2276	0.024479	0.081727	0	0	0	0	0	0	1	-360	360;
	10113	2276	0.024248	0.064893	0	0	0	0	0	0	1	-360	360;
	988	12888	0.0084	0.04056	0	0	0	0	0	0	1	-360	360;
	12888	295	0.003809	0.0248	0	0	0	0	0	0	1	-360	360;
	12888	5048	0.002542	0.014569	0	0	0	0	0	0	1	-360	360;
	12888	5048	0.002591	0.01428	0	0	0	0	0	0	1	-360	360;
	11082	11340	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	11082	8139	0.003071	0.01636	0	0	0	0	0	0	1	-360	360;
	11082	7810	0.003889	0.027022	0	0	0	0	0	0	1	-360	360;
	6997	11082	0.00624	0.036009	0	0	0	0	0	0	1	-360	360;
	6997	11082	0.00636	0.035258	0	0	0	0	0	0	1	-360	360;
	7663	3996	0.000982	0.001529	0	0	0	0	0	0	1	-360	360;
	7663	3996	0.000991	0.001502	0	0	0	0	0	0	1	-360	360;
	12773	7663	0.00052	0.004991	0	0	0	0	0	0	1	-360	360;
	7663	3868	0	0.00028	0	0	0	0	0	0	1	-360	360;
	947	6035	0.012182	0.034826	0	0	0	0	0	0	1	-360	360;
	2834	947	0.025744	0.059	0	0	0	0	0	0	1	-360	360;
	6067	2336	0	0.00028	0	0	0	0	0	0	1	-360	360;
	9017	6067	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9017	8554	0.003129	0.015031	0	0	0	0	0	0	1	-360	360;
	6067	728	0.00228	0.019871	0	0	0	0	0	0	1	-360	360;
	6067	8554	0.002791	0.017031	0	0	0	0	0	0	1	-360	360;
	9634	5049	0.001431	0.01252	0	0	0	0	0	0	1	-360	360;
	9634	8829	0.001649	0.007631	0	0	0	0	0	0	1	-360	360;
	9634	7652	0.001471	0.008649	0	0	0	0	0	0	1	-360	360;
	728	8554	0.00656	0.009071	0	0	0	0	0	0	1	-360	360;
	8813	11724	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7246	13401	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1145	11839	0.000521	0.001529	0	0	0	0	0	0	1	-360	360;
	113	12513	0.000219	0.00095	0	0	0	0	0	0	1	-360	360;
	48	12513	0.000219	0.00088	0	0	0	0	0	0	1	-360	360;
	13039	13175	0.002391	0.015618	0	0	0	0	0	0	1	-360	360;
	7325	1289	0	0.00054	0	0	0	0	0	0	1	-360	360;
	2553	7325	0.00079	0.01037	0	0	0	0	0	0	1	-360	360;
	11462	7325	0.00022	0.00296	0	0	0	0	0	0	1	-360	360;
	7960	1289	0.00077	0.01056	0	0	0	0	0	0	1	-360	360;
	1667	6295	0.000649	0.03712	0	0	0	0	0	0	1	-360	360;
	1667	12627	0.003302	0.019138	0	0	0	0	0	0	1	-360	360;
	1667	12627	0.003369	0.018742	0	0	0	0	0	0	1	-360	360;
	1667	694	0.001769	0.018222	0	0	0	0	0	0	1	-360	360;
	1667	10948	0.000858	0.006809	0	0	0	0	0	0	1	-360	360;
	1667	10948	0.000969	0.006671	0	0	0	0	0	0	1	-360	360;
	1667	12180	0.005551	0.03908	0	0	0	0	0	0	1	-360	360;
	1667	12180	0.005431	0.039822	0	0	0	0	0	0	1	-360	360;
	1667	12180	0.005311	0.036871	0	0	0	0	0	0	1	-360	360;
	1456	1575	0.002169	0.016271	0	0	0	0	0	0	1	-360	360;
	4601	1575	0.007418	0.07832	0	0	0	0	0	0	1	-360	360;
	7067	4355	0	0.0005	0	0	0	0	0	0	1	-360	360;
	7067	1592	0	0.00049	0	0	0	0	0	0	1	-360	360;
	7067	11934	0.00079	0.00848	0	0	0	0	0	0	1	-360	360;
	7067	4653	0.00183	0.0174	0	0	0	0	0	0	1	-360	360;
	4355	1592	0	0.00048	0	0	0	0	0	0	1	-360	360;
	4355	12977	0.0008	0.0083	0	0	0	0	0	0	1	-360	360;
	6275	106	0.004449	0.02404	0	0	0	0	0	0	1	-360	360;
	6275	106	0.004529	0.023551	0	0	0	0	0	0	1	-360	360;
	6275	9063	0.006791	0.013271	0	0	0	0	0	0	1	-360	360;
	6275	1589	0.00424	0.030378	0	0	0	0	0	0	1	-360	360;
	6275	1589	0.004151	0.028129	0	0	0	0	0	0	1	-360	360;
	6275	10706	0.002311	0.003622	0	0	0	0	0	0	1	-360	360;
	6275	10706	0.002062	0.004102	0	0	0	0	0	0	1	-360	360;
	6275	10209	0.00452	0.029609	0	0	0	0	0	0	1	-360	360;
	6275	10209	0.004422	0.030169	0	0	0	0	0	0	1	-360	360;
	6275	6997	0.00416	0.023058	0	0	0	0	0	0	1	-360	360;
	6275	5334	0.002831	0.020742	0	0	0	0	0	0	1	-360	360;
	4508	2455	0.002338	0.01592	0	0	0	0	0	0	1	-360	360;
	5337	4555	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	1589	5337	0.004969	0.03636	0	0	0	0	0	0	1	-360	360;
	10209	5337	0.003431	0.015698	0	0	0	0	0	0	1	-360	360;
	10661	5337	0.00456	0.027582	0	0	0	0	0	0	1	-360	360;
	11934	12977	0	0.00047	0	0	0	0	0	0	1	-360	360;
	12369	11934	0.00093	0.01185	0	0	0	0	0	0	1	-360	360;
	1112	12977	0.00091	0.01207	0	0	0	0	0	0	1	-360	360;
	988	9425	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	9064	988	0.000529	0.003151	0	0	0	0	0	0	1	-360	360;
	9064	988	0.00052	0.003209	0	0	0	0	0	0	1	-360	360;
	988	295	0.00352	0.020578	0	0	0	0	0	0	1	-360	360;
	13121	10103	0.039182	0.089413	0	0	0	0	0	0	1	-360	360;
	13121	13608	0.033686	0.067124	0	0	0	0	0	0	1	-360	360;
	4955	5962	0.044992	0.093174	0	0	0	0	0	0	1	-360	360;
	8785	4955	0.013678	0.052479	0	0	0	0	0	0	1	-360	360;
	7145	4955	0.005686	0.022893	0	0	0	0	0	0	1	-360	360;
	4955	10449	0.034281	0.094653	0	0	0	0	0	0	1	-360	360;
	3965	8934	0	0.00045	0	0	0	0	0	0	1	-360	360;
	113	3965	0.00032	0.001469	0	0	0	0	0	0	1	-360	360;
	48	8934	0.000289	0.001039	0	0	0	0	0	0	1	-360	360;
	1145	11392	0.00324	0.017628	0	0	0	0	0	0	1	-360	360;
	1145	12571	0.002678	0.019298	0	0	0	0	0	0	1	-360	360;
	1145	7789	0.000529	0.001504	0	0	0	0	0	0	1	-360	360;
	7307	2067	0.00042	0.0261	0	0	0	0	0	0	1	-360	360;
	9980	7024	0.00043	0.02555	0	0	0	0	0	0	1	-360	360;
	9496	7024	0.00044	0.02499	0	0	0	0	0	0	1	-360	360;
	12198	7024	0.00039	0.02833	0	0	0	0	0	0	1	-360	360;
	842	1927	0.00019	0.01594	0	0	0	0	0	0	1	-360	360;
	409	9452	0.000419	0.026099	0	0	0	0	0	0	1	-360	360;
	3900	11947	0.00043	0.02555	0	0	0	0	0	0	1	-360	360;
	12821	12218	0.00038	0.02888	0	0	0	0	0	0	1	-360	360;
	10053	5967	0.00038	0.02944	0	0	0	0	0	0	1	-360	360;
	5478	2067	0.00037	0.02999	0	0	0	0	0	0	1	-360	360;
	2699	2067	0.00036	0.02777	0	0	0	0	0	0	1	-360	360;
	106	11969	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	10694	106	0.003951	0.02048	0	0	0	0	0	0	1	-360	360;
	106	11965	0.003858	0.020871	0	0	0	0	0	0	1	-360	360;
	7024	12218	0	0.00046	0	0	0	0	0	0	1	-360	360;
	7024	5967	0	0.00045	0	0	0	0	0	0	1	-360	360;
	7024	10831	0	0.00051	0	0	0	0	0	0	1	-360	360;
	7024	2853	0	0.00052	0	0	0	0	0	0	1	-360	360;
	3509	11947	0	0.00051	0	0	0	0	0	0	1	-360	360;
	3509	9452	0	0.000521	0	0	0	0	0	0	1	-360	360;
	3509	9165	0	0.000531	0	0	0	0	0	0	1	-360	360;
	3509	6373	0	0.000539	0	0	0	0	0	0	1	-360	360;
	2067	12218	0	0.00053	0	0	0	0	0	0	1	-360	360;
	2067	5967	0	0.00054	0	0	0	0	0	0	1	-360	360;
	2067	10831	0	0.0005	0	0	0	0	0	0	1	-360	360;
	2067	2853	0	0.00049	0	0	0	0	0	0	1	-360	360;
	486	11947	0	0.0005	0	0	0	0	0	0	1	-360	360;
	486	9452	0	0.00049	0	0	0	0	0	0	1	-360	360;
	486	9165	0	0.000479	0	0	0	0	0	0	1	-360	360;
	486	6373	0	0.000469	0	0	0	0	0	0	1	-360	360;
	10205	12218	0.00043	0.00544	0	0	0	0	0	0	1	-360	360;
	5125	11947	0	0.000461	0	0	0	0	0	0	1	-360	360;
	9452	1927	0	0.00045	0	0	0	0	0	0	1	-360	360;
	2792	10831	0.00038	0.00616	0	0	0	0	0	0	1	-360	360;
	7729	2853	0.00041	0.02721	0	0	0	0	0	0	1	-360	360;
	6373	1508	0.002081	0.023	0	0	0	0	0	0	1	-360	360;
	9737	6959	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7545	3892	0	0.00048	0	0	0	0	0	0	1	-360	360;
	7545	2409	0.0011	0.01189	0	0	0	0	0	0	1	-360	360;
	7545	4653	0.00084	0.00906	0	0	0	0	0	0	1	-360	360;
	3892	2409	0.00112	0.01165	0	0	0	0	0	0	1	-360	360;
	3892	1593	0.00115	0.0114	0	0	0	0	0	0	1	-360	360;
	3892	8261	0.00086	0.00887	0	0	0	0	0	0	1	-360	360;
	4906	10546	0.00116	0.00912	0	0	0	0	0	0	1	-360	360;
	4906	10546	0.001129	0.009298	0	0	0	0	0	0	1	-360	360;
	10546	13275	0.000609	0.005058	0	0	0	0	0	0	1	-360	360;
	65	10546	0.008009	0.04	0	0	0	0	0	0	1	-360	360;
	10546	5680	0.002289	0.016218	0	0	0	0	0	0	1	-360	360;
	4906	579	0.003098	0.032729	0	0	0	0	0	0	1	-360	360;
	4906	5180	0.002982	0.013058	0	0	0	0	0	0	1	-360	360;
	4906	5180	0.00304	0.0128	0	0	0	0	0	0	1	-360	360;
	4906	8486	0.008791	0.059622	0	0	0	0	0	0	1	-360	360;
	13275	9568	0.000511	0.004409	0	0	0	0	0	0	1	-360	360;
	11253	9568	0.009391	0.030809	0	0	0	0	0	0	1	-360	360;
	11253	9568	0.009191	0.031391	0	0	0	0	0	0	1	-360	360;
	2423	4153	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	9063	2423	0.00276	0.014782	0	0	0	0	0	0	1	-360	360;
	9063	2423	0.002809	0.014458	0	0	0	0	0	0	1	-360	360;
	1589	2423	0.003138	0.017502	0	0	0	0	0	0	1	-360	360;
	1589	2423	0.0032	0.017138	0	0	0	0	0	0	1	-360	360;
	8618	9277	0.002033	0.005273	0	0	0	0	0	0	1	-360	360;
	6307	9277	0.036066	0.160107	0	0	0	0	0	0	1	-360	360;
	1062	13197	0.027686	0.04614	0	0	0	0	0	0	1	-360	360;
	5855	13197	0.025107	0.051372	0	0	0	0	0	0	1	-360	360;
	11741	805	0.002479	0.01086	0	0	0	0	0	0	1	-360	360;
	805	10113	0.013603	0.051876	0	0	0	0	0	0	1	-360	360;
	11783	12548	0.00176	0.108	0	0	0	0	0	0	1	-360	360;
	2269	8081	0	0.00047	0	0	0	0	0	0	1	-360	360;
	2269	7960	0.00038	0.00366	0	0	0	0	0	0	1	-360	360;
	2269	12160	0.00052	0.00633	0	0	0	0	0	0	1	-360	360;
	11567	12550	0	0.00051	0	0	0	0	0	0	1	-360	360;
	11567	4743	0.001599	0.015599	0	0	0	0	0	0	1	-360	360;
	7752	11567	0.00212	0.022349	0	0	0	0	0	0	1	-360	360;
	8081	2553	0.00034	0.00415	0	0	0	0	0	0	1	-360	360;
	12080	8081	0.00051	0.00646	0	0	0	0	0	0	1	-360	360;
	12550	547	0.00163	0.015271	0	0	0	0	0	0	1	-360	360;
	13242	12550	0.002159	0.021901	0	0	0	0	0	0	1	-360	360;
	8669	1133	0.00224	0.015018	0	0	0	0	0	0	1	-360	360;
	8669	4601	0.004111	0.00652	0	0	0	0	0	0	1	-360	360;
	8669	4601	0.004191	0.006391	0	0	0	0	0	0	1	-360	360;
	10113	7460	0.00238	0.006744	0	0	0	0	0	0	1	-360	360;
	8486	579	0.003329	0.033129	0	0	0	0	0	0	1	-360	360;
	8589	6233	0.005809	0.021182	0	0	0	0	0	0	1	-360	360;
	8589	6233	0.005689	0.021591	0	0	0	0	0	0	1	-360	360;
	6233	3890	0.00132	0.01052	0	0	0	0	0	0	1	-360	360;
	6233	367	0.001129	0.012329	0	0	0	0	0	0	1	-360	360;
	6233	3209	0.001569	0.1224	0	0	0	0	0	0	1	-360	360;
	4508	3693	0.000662	0.036351	0	0	0	0	0	0	1	-360	360;
	6330	10854	0.001777	0.009107	0	0	0	0	0	0	1	-360	360;
	6330	1899	0.002967	0.002901	0	0	0	0	0	0	1	-360	360;
	6330	12203	0.017893	0.057479	0	0	0	0	0	0	1	-360	360;
	10854	13148	0.000554	0.001091	0	0	0	0	0	0	1	-360	360;
	10854	11060	0.006339	0.022603	0	0	0	0	0	0	1	-360	360;
	1899	536	0.004174	0.006702	0	0	0	0	0	0	1	-360	360;
	536	11060	0.009818	0.040099	0	0	0	0	0	0	1	-360	360;
	417	9249	0.003729	0.01876	0	0	0	0	0	0	1	-360	360;
	8554	10592	0.000782	0.059938	0	0	0	0	0	0	1	-360	360;
	12059	8554	0.000769	0.061111	0	0	0	0	0	0	1	-360	360;
	1456	9536	0.000382	0.019378	0	0	0	0	0	0	1	-360	360;
	1456	9518	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1589	9988	1.8e-05	0.00188	0	0	0	0	0	0	1	-360	360;
	12501	8782	0.000372	0.001223	0	0	0	0	0	0	1	-360	360;
	11783	8495	0.000809	0.001751	0	0	0	0	0	0	1	-360	360;
	4920	1236	0	0.043511	0	0	0	0	0	0	1	-360	360;
	10912	4920	0.003529	0.022178	0	0	0	0	0	0	1	-360	360;
	10912	4920	0.0036	0.021711	0	0	0	0	0	0	1	-360	360;
	9333	4265	0	0.000413	0	0	0	0	0	0	1	-360	360;
	3177	9333	0.047479	0.102603	0	0	0	0	0	0	1	-360	360;
	3177	9333	0.053727	0.091281	0	0	0	0	0	0	1	-360	360;
	8948	9333	0.021678	0.045669	0	0	0	0	0	0	1	-360	360;
	3177	12372	0.004397	0.030678	0	0	0	0	0	0	1	-360	360;
	3177	12372	0.004314	0.031256	0	0	0	0	0	0	1	-360	360;
	3177	11392	0.008347	0.055512	0	0	0	0	0	0	1	-360	360;
	3177	3860	0.006223	0.039744	0	0	0	0	0	0	1	-360	360;
	12184	3177	0.006124	0.041686	0	0	0	0	0	0	1	-360	360;
	12184	3177	0.006942	0.040851	0	0	0	0	0	0	1	-360	360;
	5962	8406	0.019248	0.039041	0	0	0	0	0	0	1	-360	360;
	10449	5962	0.015719	0.04943	0	0	0	0	0	0	1	-360	360;
	8068	1716	0.00296	0.011871	0	0	0	0	0	0	1	-360	360;
	8068	13431	0.01328	0.04144	0	0	0	0	0	0	1	-360	360;
	3935	10451	0.00736	0.033222	0	0	0	0	0	0	1	-360	360;
	3935	10451	0.007502	0.032529	0	0	0	0	0	0	1	-360	360;
	2386	3935	0.00192	0.02576	0	0	0	0	0	0	1	-360	360;
	2386	3935	0.002169	0.02524	0	0	0	0	0	0	1	-360	360;
	3935	12527	0.00656	0.017071	0	0	0	0	0	0	1	-360	360;
	3935	3938	0.002662	0.007471	0	0	0	0	0	0	1	-360	360;
	3935	4660	0.003191	0.009062	0	0	0	0	0	0	1	-360	360;
	10103	12831	0.00981	0.057298	0	0	0	0	0	0	1	-360	360;
	4595	9295	0.008342	0.037031	0	0	0	0	0	0	1	-360	360;
	2074	9295	0.0086	0.039791	0	0	0	0	0	0	1	-360	360;
	2409	1593	0	0.00046	0	0	0	0	0	0	1	-360	360;
	8845	2409	0.00055	0.00685	0	0	0	0	0	0	1	-360	360;
	11660	1593	0.00054	0.00699	0	0	0	0	0	0	1	-360	360;
	13039	103	0.0056	0.052889	0	0	0	0	0	0	1	-360	360;
	13039	8664	0.006058	0.023591	0	0	0	0	0	0	1	-360	360;
	13039	8664	0.0054	0.026729	0	0	0	0	0	0	1	-360	360;
	13039	12722	0.00168	0.01188	0	0	0	0	0	0	1	-360	360;
	13039	13317	0.003178	0.028658	0	0	0	0	0	0	1	-360	360;
	13039	7579	0.003178	0.028542	0	0	0	0	0	0	1	-360	360;
	7246	9430	0.00068	0.004449	0	0	0	0	0	0	1	-360	360;
	7246	5049	0.00076	0.004818	0	0	0	0	0	0	1	-360	360;
	4019	2542	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4019	8589	9e-06	0.000222	0	0	0	0	0	0	1	-360	360;
	7891	8589	0	0.000222	0	0	0	0	0	0	1	-360	360;
	10902	8589	1.8e-05	0.004329	0	0	0	0	0	0	1	-360	360;
	8813	7688	0.001831	0.001698	0	0	0	0	0	0	1	-360	360;
	8813	7688	0.002209	0.001471	0	0	0	0	0	0	1	-360	360;
	1848	11394	0.017653	0.070289	0	0	0	0	0	0	1	-360	360;
	6070	11394	0.013231	0.042207	0	0	0	0	0	0	1	-360	360;
	6070	11394	0.013479	0.041331	0	0	0	0	0	0	1	-360	360;
	4008	11394	0.013934	0.055702	0	0	0	0	0	0	1	-360	360;
	6190	8785	0.000537	0.001471	0	0	0	0	0	0	1	-360	360;
	8785	7145	0.009967	0.048959	0	0	0	0	0	0	1	-360	360;
	8785	13524	0.030612	0.055058	0	0	0	0	0	0	1	-360	360;
	8785	13524	0.031207	0.053934	0	0	0	0	0	0	1	-360	360;
	8827	8785	0.01314	0.040579	0	0	0	0	0	0	1	-360	360;
	8827	8785	0.012851	0.03757	0	0	0	0	0	0	1	-360	360;
	8785	10245	0.006901	0.021174	0	0	0	0	0	0	1	-360	360;
	8785	10245	0.007653	0.01924	0	0	0	0	0	0	1	-360	360;
	10694	8155	0.008982	0.038462	0	0	0	0	0	0	1	-360	360;
	8155	11965	0.009151	0.037649	0	0	0	0	0	0	1	-360	360;
	8155	1456	0.004258	0.03952	0	0	0	0	0	0	1	-360	360;
	3753	8155	0.003489	0.0264	0	0	0	0	0	0	1	-360	360;
	8155	3287	0.003849	0.05232	0	0	0	0	0	0	1	-360	360;
	2834	6035	0.032669	0.080273	0	0	0	0	0	0	1	-360	360;
	2697	6035	0.004901	0.014967	0	0	0	0	0	0	1	-360	360;
	465	8406	0.000562	0.001058	0	0	0	0	0	0	1	-360	360;
	12369	1112	0	0.00045	0	0	0	0	0	0	1	-360	360;
	12369	12160	0.001	0.01169	0	0	0	0	0	0	1	-360	360;
	1112	12080	0.00098	0.01192	0	0	0	0	0	0	1	-360	360;
	9653	11060	0.002339	0.006223	0	0	0	0	0	0	1	-360	360;
	9653	11060	0.002388	0.006091	0	0	0	0	0	0	1	-360	360;
	600	9653	0.000504	0.001207	0	0	0	0	0	0	1	-360	360;
	8406	11060	0.00105	0.007959	0	0	0	0	0	0	1	-360	360;
	8406	11060	0.001033	0.007372	0	0	0	0	0	0	1	-360	360;
	8406	3794	0.018752	0.042339	0	0	0	0	0	0	1	-360	360;
	8406	3794	0.018339	0.039231	0	0	0	0	0	0	1	-360	360;
	2834	10113	0.036793	0.081719	0	0	0	0	0	0	1	-360	360;
	2834	12501	0.003702	0.015132	0	0	0	0	0	0	1	-360	360;
	3555	2834	0.00124	0.007314	0	0	0	0	0	0	1	-360	360;
	766	9851	0.002329	0.00356	0	0	0	0	0	0	1	-360	360;
	766	9851	0.002378	0.00348	0	0	0	0	0	0	1	-360	360;
	766	9851	0.00212	0.003951	0	0	0	0	0	0	1	-360	360;
	766	10948	0.001951	0.003791	0	0	0	0	0	0	1	-360	360;
	103	11538	0.002542	0.01472	0	0	0	0	0	0	1	-360	360;
	2386	10306	0.003262	0.024151	0	0	0	0	0	0	1	-360	360;
	13611	7881	0.0036	0.01548	0	0	0	0	0	0	1	-360	360;
	13611	7881	0.00352	0.014329	0	0	0	0	0	0	1	-360	360;
	2682	6437	0.002591	0.014418	0	0	0	0	0	0	1	-360	360;
	5180	2682	0.001071	0.003058	0	0	0	0	0	0	1	-360	360;
	5180	525	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	5180	1887	0.017458	0.052129	0	0	0	0	0	0	1	-360	360;
	10694	12439	0.00136	0.00216	0	0	0	0	0	0	1	-360	360;
	11965	12439	0.001391	0.00212	0	0	0	0	0	0	1	-360	360;
	9365	13431	0.005858	0.016591	0	0	0	0	0	0	1	-360	360;
	1259	12831	0.004562	0.011653	0	0	0	0	0	0	1	-360	360;
	1259	6983	0.030868	0.064174	0	0	0	0	0	0	1	-360	360;
	4743	547	0	0.000521	0	0	0	0	0	0	1	-360	360;
	3408	1456	0.005929	0.053271	0	0	0	0	0	0	1	-360	360;
	3408	445	0.00664	0.040849	0	0	0	0	0	0	1	-360	360;
	3408	4601	0.00968	0.049382	0	0	0	0	0	0	1	-360	360;
	8486	6882	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	8486	103	0.00044	0.004489	0	0	0	0	0	0	1	-360	360;
	8486	5733	0.01064	0.043751	0	0	0	0	0	0	1	-360	360;
	8486	5733	0.0104	0.040511	0	0	0	0	0	0	1	-360	360;
	7579	8486	0.0024	0.025498	0	0	0	0	0	0	1	-360	360;
	6264	1714	0.00264	0.01412	0	0	0	0	0	0	1	-360	360;
	694	8968	0.000609	0.03944	0	0	0	0	0	0	1	-360	360;
	5815	694	0	0.000222	0	0	0	0	0	0	1	-360	360;
	694	9851	0.001182	0.003018	0	0	0	0	0	0	1	-360	360;
	694	9851	0.0012	0.002951	0	0	0	0	0	0	1	-360	360;
	694	9851	0.001071	0.003351	0	0	0	0	0	0	1	-360	360;
	694	10336	0.001031	0.01544	0	0	0	0	0	0	1	-360	360;
	694	520	0.001009	0.015738	0	0	0	0	0	0	1	-360	360;
	694	2388	0.000991	0.01604	0	0	0	0	0	0	1	-360	360;
	9063	10706	0.003898	0.008862	0	0	0	0	0	0	1	-360	360;
	9063	10706	0.003818	0.009031	0	0	0	0	0	0	1	-360	360;
	939	464	0.003831	0.012502	0	0	0	0	0	0	1	-360	360;
	8845	11660	0	0.00051	0	0	0	0	0	0	1	-360	360;
	8768	11660	0.00127	0.0132	0	0	0	0	0	0	1	-360	360;
	10912	9802	0	0.000222	0	0	0	0	0	0	1	-360	360;
	10912	11228	0.001338	0.011209	0	0	0	0	0	0	1	-360	360;
	10912	7432	0.004222	0.030062	0	0	0	0	0	0	1	-360	360;
	13558	10912	0.001182	0.0104	0	0	0	0	0	0	1	-360	360;
	9802	5649	9e-06	0.000222	0	0	0	0	0	0	1	-360	360;
	9802	7432	0.004138	0.03064	0	0	0	0	0	0	1	-360	360;
	5649	7432	0.004049	0.031209	0	0	0	0	0	0	1	-360	360;
	1062	5855	0.022603	0.051959	0	0	0	0	0	0	1	-360	360;
	12289	1062	0.013636	0.032041	0	0	0	0	0	0	1	-360	360;
	12527	4661	0.00228	0.009222	0	0	0	0	0	0	1	-360	360;
	3938	4661	0.002471	0.010302	0	0	0	0	0	0	1	-360	360;
	4660	4661	0.001831	0.007378	0	0	0	0	0	0	1	-360	360;
	12372	6307	0.006843	0.049248	0	0	0	0	0	0	1	-360	360;
	12372	6307	0.006686	0.045603	0	0	0	0	0	0	1	-360	360;
	3954	2815	0.004102	0.011529	0	0	0	0	0	0	1	-360	360;
	3954	2815	0.004182	0.01128	0	0	0	0	0	0	1	-360	360;
	2733	3001	0	0.000222	0	0	0	0	0	0	1	-360	360;
	11997	2733	0.003649	0.0122	0	0	0	0	0	0	1	-360	360;
	11997	2733	0.00372	0.011938	0	0	0	0	0	0	1	-360	360;
	2733	8098	0.005031	0.028102	0	0	0	0	0	0	1	-360	360;
	2733	8098	0.005129	0.02752	0	0	0	0	0	0	1	-360	360;
	2733	11519	0.005218	0.026951	0	0	0	0	0	0	1	-360	360;
	6480	5049	0.00288	0.01992	0	0	0	0	0	0	1	-360	360;
	6480	5049	0.003262	0.01952	0	0	0	0	0	0	1	-360	360;
	9064	1589	0.00956	0.061022	0	0	0	0	0	0	1	-360	360;
	9064	3954	0.005382	0.033911	0	0	0	0	0	0	1	-360	360;
	9064	3954	0.005271	0.034569	0	0	0	0	0	0	1	-360	360;
	9064	9080	0.004102	0.023018	0	0	0	0	0	0	1	-360	360;
	2553	7960	0	0.00052	0	0	0	0	0	0	1	-360	360;
	1456	7099	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1456	4601	0.011631	0.094591	0	0	0	0	0	0	1	-360	360;
	3753	1456	0.0012	0.008769	0	0	0	0	0	0	1	-360	360;
	1456	13414	0.001218	0.008582	0	0	0	0	0	0	1	-360	360;
	10451	12642	0.002689	0.013818	0	0	0	0	0	0	1	-360	360;
	3794	12203	0.005	0.008934	0	0	0	0	0	0	1	-360	360;
	11313	199	0	0.000531	0	0	0	0	0	0	1	-360	360;
	11313	7752	0.001281	0.0158	0	0	0	0	0	0	1	-360	360;
	199	13242	0.00125	0.016099	0	0	0	0	0	0	1	-360	360;
	1848	6983	0.010058	0.048	0	0	0	0	0	0	1	-360	360;
	1848	6983	0.010248	0.047	0	0	0	0	0	0	1	-360	360;
	1848	4008	0.004314	0.024537	0	0	0	0	0	0	1	-360	360;
	2386	4817	0.010791	0.042151	0	0	0	0	0	0	1	-360	360;
	507	8951	0.0006	0.040218	0	0	0	0	0	0	1	-360	360;
	367	12734	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4653	8261	0	0.00053	0	0	0	0	0	0	1	-360	360;
	4653	6366	0	0.00054	0	0	0	0	0	0	1	-360	360;
	2386	13252	0.00196	0.010249	0	0	0	0	0	0	1	-360	360;
	2386	2165	0.002182	0.010498	0	0	0	0	0	0	1	-360	360;
	8664	12958	0.000582	0.040991	0	0	0	0	0	0	1	-360	360;
	8664	8239	0.000569	0.04176	0	0	0	0	0	0	1	-360	360;
	2074	4595	0.003818	0.013209	0	0	0	0	0	0	1	-360	360;
	8664	9249	0.00428	0.023511	0	0	0	0	0	0	1	-360	360;
	8664	9249	0.004182	0.021769	0	0	0	0	0	0	1	-360	360;
	2792	10205	0	0.0005	0	0	0	0	0	0	1	-360	360;
	2792	12160	0.00114	0.01661	0	0	0	0	0	0	1	-360	360;
	6333	6237	0	0.000539	0	0	0	0	0	0	1	-360	360;
	6333	113	0.00119	0.014081	0	0	0	0	0	0	1	-360	360;
	10205	12080	0.00117	0.01626	0	0	0	0	0	0	1	-360	360;
	6237	48	0.00106	0.01595	0	0	0	0	0	0	1	-360	360;
	6307	8804	0.018058	0.065719	0	0	0	0	0	0	1	-360	360;
	6307	8948	0.012826	0.025058	0	0	0	0	0	0	1	-360	360;
	6307	8948	0.013331	0.024	0	0	0	0	0	0	1	-360	360;
	13252	2165	0.001089	0.003151	0	0	0	0	0	0	1	-360	360;
	8418	3479	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	507	1804	0.000631	0.054471	0	0	0	0	0	0	1	-360	360;
	507	2998	0.000622	0.055489	0	0	0	0	0	0	1	-360	360;
	507	13166	0.0006	0.051382	0	0	0	0	0	0	1	-360	360;
	6070	6983	0.017702	0.064331	0	0	0	0	0	0	1	-360	360;
	6070	2552	0.012314	0.04924	0	0	0	0	0	0	1	-360	360;
	7059	12352	0.001223	0.001248	0	0	0	0	0	0	1	-360	360;
	7725	12352	0.001471	0.00481	0	0	0	0	0	0	1	-360	360;
	8804	12352	0.011521	0.03138	0	0	0	0	0	0	1	-360	360;
	11783	11627	0	0.000222	0	0	0	0	0	0	1	-360	360;
	11783	8990	0.000289	0.00056	0	0	0	0	0	0	1	-360	360;
	5835	4914	0.00034	0.01805	0	0	0	0	0	0	1	-360	360;
	10953	3890	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3044	9782	0.00035	0.01766	0	0	0	0	0	0	1	-360	360;
	13244	6183	0.00035	0.01728	0	0	0	0	0	0	1	-360	360;
	7412	12253	0.00031	0.01958	0	0	0	0	0	0	1	-360	360;
	10651	4914	0.00031	0.01997	0	0	0	0	0	0	1	-360	360;
	9404	4914	0.0003	0.02035	0	0	0	0	0	0	1	-360	360;
	8583	8758	0.00016	0.0025	0	0	0	0	0	0	1	-360	360;
	7000	4914	0.00029	0.02074	0	0	0	0	0	0	1	-360	360;
	11827	12253	0.00029	0.0192	0	0	0	0	0	0	1	-360	360;
	12504	12253	0.00033	0.01882	0	0	0	0	0	0	1	-360	360;
	9782	6183	0	0.00049	0	0	0	0	0	0	1	-360	360;
	6934	9782	0	0.00048	0	0	0	0	0	0	1	-360	360;
	11543	6183	0	0.00047	0	0	0	0	0	0	1	-360	360;
	4914	8593	0	0.00046	0	0	0	0	0	0	1	-360	360;
	4914	616	0	0.00045	0	0	0	0	0	0	1	-360	360;
	4914	13170	0	0.00051	0	0	0	0	0	0	1	-360	360;
	12253	11543	0	0.00052	0	0	0	0	0	0	1	-360	360;
	12253	6934	0	0.00053	0	0	0	0	0	0	1	-360	360;
	12253	13170	0	0.00054	0	0	0	0	0	0	1	-360	360;
	8593	8872	0	0.0005	0	0	0	0	0	0	1	-360	360;
	8872	11543	0	0.00049	0	0	0	0	0	0	1	-360	360;
	8872	5998	0.00028	0.00403	0	0	0	0	0	0	1	-360	360;
	9167	8872	0.00021	0.00396	0	0	0	0	0	0	1	-360	360;
	616	4211	0	0.00048	0	0	0	0	0	0	1	-360	360;
	4211	6934	0	0.00047	0	0	0	0	0	0	1	-360	360;
	4211	8583	0.00029	0.00394	0	0	0	0	0	0	1	-360	360;
	6578	4211	0.0002	0.00404	0	0	0	0	0	0	1	-360	360;
	3890	367	0	0.000222	0	0	0	0	0	0	1	-360	360;
	367	818	0	0.000222	0	0	0	0	0	0	1	-360	360;
	12527	3938	0.004471	0.020791	0	0	0	0	0	0	1	-360	360;
	12527	4660	0.002791	0.01132	0	0	0	0	0	0	1	-360	360;
	9851	12368	0.000822	0.001649	0	0	0	0	0	0	1	-360	360;
	9851	12368	0.0008	0.00168	0	0	0	0	0	0	1	-360	360;
	9851	12368	0.000782	0.00156	0	0	0	0	0	0	1	-360	360;
	548	464	0.00432	0.012662	0	0	0	0	0	0	1	-360	360;
	548	464	0.003849	0.014351	0	0	0	0	0	0	1	-360	360;
	1589	3666	0.001418	0.004618	0	0	0	0	0	0	1	-360	360;
	1589	10661	0.001689	0.011631	0	0	0	0	0	0	1	-360	360;
	1589	9080	0.005991	0.03832	0	0	0	0	0	0	1	-360	360;
	13611	2763	0.001169	0.007098	0	0	0	0	0	0	1	-360	360;
	10948	2763	0.0012	0.00696	0	0	0	0	0	0	1	-360	360;
	6483	13524	0.030066	0.046289	0	0	0	0	0	0	1	-360	360;
	6483	13524	0.026793	0.052463	0	0	0	0	0	0	1	-360	360;
	11331	6483	0.023413	0.057397	0	0	0	0	0	0	1	-360	360;
	8139	7810	0.001631	0.00832	0	0	0	0	0	0	1	-360	360;
	8813	6506	0.012569	0.04732	0	0	0	0	0	0	1	-360	360;
	8813	6506	0.012311	0.048231	0	0	0	0	0	0	1	-360	360;
	11790	9541	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7432	11790	0.000982	0.004751	0	0	0	0	0	0	1	-360	360;
	7442	1781	0.00144	0.12	0	0	0	0	0	0	1	-360	360;
	507	7810	0.001249	0.0064	0	0	0	0	0	0	1	-360	360;
	507	7810	0.00112	0.007249	0	0	0	0	0	0	1	-360	360;
	7752	13242	0	0.0005	0	0	0	0	0	0	1	-360	360;
	13608	12831	0.022843	0.098893	0	0	0	0	0	0	1	-360	360;
	12831	6983	0.034909	0.072893	0	0	0	0	0	0	1	-360	360;
	11228	558	0.00188	0.092222	0	0	0	0	0	0	1	-360	360;
	11228	1738	0.00184	0.093991	0	0	0	0	0	0	1	-360	360;
	11228	5890	0.0018	0.09576	0	0	0	0	0	0	1	-360	360;
	8554	11228	0.00132	0.01088	0	0	0	0	0	0	1	-360	360;
	8827	2697	0.01924	0.056603	0	0	0	0	0	0	1	-360	360;
	8827	2697	0.017124	0.064149	0	0	0	0	0	0	1	-360	360;
	6264	5106	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	6264	6959	0.006889	0.03404	0	0	0	0	0	0	1	-360	360;
	6264	6959	0.007018	0.033329	0	0	0	0	0	0	1	-360	360;
	6264	11685	0.006169	0.035689	0	0	0	0	0	0	1	-360	360;
	6264	11685	0.00628	0.034929	0	0	0	0	0	0	1	-360	360;
	6264	12180	0.004271	0.026418	0	0	0	0	0	0	1	-360	360;
	6264	12180	0.00436	0.02588	0	0	0	0	0	0	1	-360	360;
	6264	12180	0.00444	0.025338	0	0	0	0	0	0	1	-360	360;
	5733	1887	0.00852	0.02444	0	0	0	0	0	0	1	-360	360;
	12722	65	0.002671	0.018929	0	0	0	0	0	0	1	-360	360;
	11462	4886	0	0.00046	0	0	0	0	0	0	1	-360	360;
	11462	9879	0	0.00045	0	0	0	0	0	0	1	-360	360;
	8768	4886	0	0.00051	0	0	0	0	0	0	1	-360	360;
	8768	9879	0	0.00052	0	0	0	0	0	0	1	-360	360;
	3666	10661	0.000329	0.000902	0	0	0	0	0	0	1	-360	360;
	12180	5462	0.00068	0.050351	0	0	0	0	0	0	1	-360	360;
	4950	6959	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1927	5125	0	0.00049	0	0	0	0	0	0	1	-360	360;
	1927	1508	0.00212	0.022531	0	0	0	0	0	0	1	-360	360;
	1927	113	0.000919	0.012789	0	0	0	0	0	0	1	-360	360;
	5125	5693	0.002159	0.02206	0	0	0	0	0	0	1	-360	360;
	5125	48	0.00094	0.012531	0	0	0	0	0	0	1	-360	360;
	279	7432	0.00032	0.014102	0	0	0	0	0	0	1	-360	360;
	7103	7432	0.000329	0.0138	0	0	0	0	0	0	1	-360	360;
	5437	7432	0.000329	0.013502	0	0	0	0	0	0	1	-360	360;
	9249	792	0.005818	0.02396	0	0	0	0	0	0	1	-360	360;
	9249	792	0.005942	0.023462	0	0	0	0	0	0	1	-360	360;
	8554	9810	0.000671	0.001871	0	0	0	0	0	0	1	-360	360;
	7128	9810	0	0.001791	0	0	0	0	0	0	1	-360	360;
	8554	5141	0.002249	0.003471	0	0	0	0	0	0	1	-360	360;
	8554	5141	0.001631	0.0042	0	0	0	0	0	0	1	-360	360;
	8554	5549	0.00092	0.002582	0	0	0	0	0	0	1	-360	360;
	7432	5141	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7432	8696	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	13558	7432	0.000822	0.00572	0	0	0	0	0	0	1	-360	360;
	7432	6696	0.001338	0.013622	0	0	0	0	0	0	1	-360	360;
	3192	7432	0.00136	0.004649	0	0	0	0	0	0	1	-360	360;
	5141	7128	0	0.001751	0	0	0	0	0	0	1	-360	360;
	5549	5141	0	0.001711	0	0	0	0	0	0	1	-360	360;
	8829	9430	0.000662	0.004529	0	0	0	0	0	0	1	-360	360;
	7652	9430	0.000649	0.004618	0	0	0	0	0	0	1	-360	360;
	6983	2552	0.005669	0.028992	0	0	0	0	0	0	1	-360	360;
	10137	10637	0.001631	0.1176	0	0	0	0	0	0	1	-360	360;
	1730	7801	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	507	1730	0.006942	0.04516	0	0	0	0	0	0	1	-360	360;
	507	1730	0.0068	0.04604	0	0	0	0	0	0	1	-360	360;
	12571	3860	0.003289	0.017909	0	0	0	0	0	0	1	-360	360;
	4660	12053	0.00056	0.038671	0	0	0	0	0	0	1	-360	360;
	6872	11331	0.02162	0.062612	0	0	0	0	0	0	1	-360	360;
	65	3578	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	65	13317	0.000822	0.00656	0	0	0	0	0	0	1	-360	360;
	65	1300	0.003009	0.004649	0	0	0	0	0	0	1	-360	360;
	65	1300	0.003071	0.004551	0	0	0	0	0	0	1	-360	360;
	10209	599	0.00464	0.029431	0	0	0	0	0	0	1	-360	360;
	10209	599	0.005249	0.02884	0	0	0	0	0	0	1	-360	360;
	4508	8813	0.001902	0.012702	0	0	0	0	0	0	1	-360	360;
	4508	8813	0.001858	0.012938	0	0	0	0	0	0	1	-360	360;
	4508	11218	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8990	7919	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9157	5602	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	507	10492	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	13431	1406	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	13611	4584	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	13611	6959	0.004738	0.012529	0	0	0	0	0	0	1	-360	360;
	13611	6959	0.004831	0.012262	0	0	0	0	0	0	1	-360	360;
	13611	6959	0.002711	0.018431	0	0	0	0	0	0	1	-360	360;
	5855	13410	0.01205	0.03281	0	0	0	0	0	0	1	-360	360;
	4832	6959	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6959	1531	0.000649	0.03712	0	0	0	0	0	0	1	-360	360;
	6959	6569	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9242	6959	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6997	5334	0.00076	0.005329	0	0	0	0	0	0	1	-360	360;
	4601	445	0.005769	0.03528	0	0	0	0	0	0	1	-360	360;
	10948	2072	0.002258	0.004849	0	0	0	0	0	0	1	-360	360;
	10948	2072	0.002209	0.004489	0	0	0	0	0	0	1	-360	360;
	1508	5693	0	0.000479	0	0	0	0	0	0	1	-360	360;
	1508	9483	0.00256	0.01561	0	0	0	0	0	0	1	-360	360;
	5693	7724	0.00269	0.01451	0	0	0	0	0	0	1	-360	360;
	12184	11881	0.003372	0.028231	0	0	0	0	0	0	1	-360	360;
	12184	13495	0.007033	0.047083	0	0	0	0	0	0	1	-360	360;
	113	48	0	0.000469	0	0	0	0	0	0	1	-360	360;
	4601	2827	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	3954	4601	0.00392	0.025191	0	0	0	0	0	0	1	-360	360;
	3954	4601	0.00444	0.024689	0	0	0	0	0	0	1	-360	360;
	13495	11881	0.003529	0.021223	0	0	0	0	0	0	1	-360	360;
	8829	7652	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6578	583	0.00069	0.00134	0	0	0	0	0	0	1	-360	360;
	8504	9167	0.00067	0.00136	0	0	0	0	0	0	1	-360	360;
	3555	12501	0.012992	0.070207	0	0	0	0	0	0	1	-360	360;
	583	8504	0	0.00053	0	0	0	0	0	0	1	-360	360;
	9483	7724	0	0.000461	0	0	0	0	0	0	1	-360	360;
	12160	12080	0	0.00054	0	0	0	0	0	0	1	-360	360;
	10245	12289	0.002314	0.005661	0	0	0	0	0	0	1	-360	360;
	12289	10245	0.002512	0.005909	0	0	0	0	0	0	1	-360	360;
	8827	10245	0.006364	0.018488	0	0	0	0	0	0	1	-360	360;
	8827	10245	0.005959	0.017752	0	0	0	0	0	0	1	-360	360;
	8827	13410	0.018397	0.04176	0	0	0	0	0	0	1	-360	360;
	8827	13410	0.018058	0.042653	0	0	0	0	0	0	1	-360	360;
	8098	11519	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3753	12784	0	0.000529	0	0	0	0	0	0	1	-360	360;
	13414	12784	0	0.00052	0	0	0	0	0	0	1	-360	360;
	3287	13414	0.003369	0.031329	0	0	0	0	0	0	1	-360	360;
	3601	9145	0.07449	0.527603	0	0	0	0	0	0	1	-360	360;
	4373	9145	0.01093	0.074531	0	0	0	0	0	0	1	-360	360;
	7999	9145	0.00887	0.04576	0	0	0	0	0	0	1	-360	360;
	5745	9145	0.01099	0.06689	0	0	0	0	0	0	1	-360	360;
	8031	1506	0.04732	0.42186	0	0	0	0	0	0	1	-360	360;
	8031	1506	0.004469	0.03124	0	0	0	0	0	0	1	-360	360;
	1506	633	0.500289	1.586157	0	0	0	0	0	0	1	-360	360;
	1506	7690	0.01089	0.060719	0	0	0	0	0	0	1	-360	360;
	1506	12514	0.133469	0.882645	0	0	0	0	0	0	1	-360	360;
	8031	1506	0.005531	0.032521	0	0	0	0	0	0	1	-360	360;
	1506	12514	0.005219	0.025659	0	0	0	0	0	0	1	-360	360;
	12798	5651	0.00186	0.012789	0	0	0	0	0	0	1	-360	360;
	5867	5651	0.001	0.234938	0	0	0	0	0	0	1	-360	360;
	7423	9801	0.00081	0.004781	0	0	0	0	0	0	1	-360	360;
	4104	12032	0.00082	0.00468	0	0	0	0	0	0	1	-360	360;
	6716	6737	0.03374	0.614711	0	0	0	0	0	0	1	-360	360;
	12597	6737	0.04612	0.366777	0	0	0	0	0	0	1	-360	360;
	7534	6737	0.001461	0.01239	0	0	0	0	0	0	1	-360	360;
	12597	6716	0.078599	0.394649	0	0	0	0	0	0	1	-360	360;
	1998	6716	0.00143	0.01263	0	0	0	0	0	0	1	-360	360;
	10689	1379	0.007841	0.05613	0	0	0	0	0	0	1	-360	360;
	7749	3514	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7749	3514	0.001	0.328946	0	0	0	0	0	0	1	-360	360;
	5321	3514	0.224855	1.532376	0	0	0	0	0	0	1	-360	360;
	1063	3514	0.136781	1.457355	0	0	0	0	0	0	1	-360	360;
	1063	3514	0.01055	0.074979	0	0	0	0	0	0	1	-360	360;
	2109	3514	0.014909	0.523802	0	0	0	0	0	0	1	-360	360;
	2619	3514	0.001	0.71781	0	0	0	0	0	0	1	-360	360;
	11043	3514	0.134271	1.687004	0	0	0	0	0	0	1	-360	360;
	6113	7749	0.002341	0.01263	0	0	0	0	0	0	1	-360	360;
	5321	7749	0.235227	1.541777	0	0	0	0	0	0	1	-360	360;
	1063	7749	0.16344	1.266136	0	0	0	0	0	0	1	-360	360;
	2109	7749	0.015479	0.562376	0	0	0	0	0	0	1	-360	360;
	2619	7749	0.001	0.767789	0	0	0	0	0	0	1	-360	360;
	11043	7749	0.158281	1.596364	0	0	0	0	0	0	1	-360	360;
	10849	5015	0.006169	0.039289	0	0	0	0	0	0	1	-360	360;
	12209	5015	0.08256	0.638182	0	0	0	0	0	0	1	-360	360;
	3967	5015	0.221653	1.097397	0	0	0	0	0	0	1	-360	360;
	5015	9811	0.0118	0.05445	0	0	0	0	0	0	1	-360	360;
	4325	5015	0.08986	0.756921	0	0	0	0	0	0	1	-360	360;
	8076	933	0.03887	0.335083	0	0	0	0	0	0	1	-360	360;
	933	8076	0.00455	0.036539	0	0	0	0	0	0	1	-360	360;
	246	933	0.002169	0.01688	0	0	0	0	0	0	1	-360	360;
	246	933	0.01045	0.244731	0	0	0	0	0	0	1	-360	360;
	9654	933	0.001	0.259525	0	0	0	0	0	0	1	-360	360;
	5298	933	0.036651	0.378657	0	0	0	0	0	0	1	-360	360;
	9643	9177	0.004669	0.03839	0	0	0	0	0	0	1	-360	360;
	857	9177	0.000781	0.00674	0	0	0	0	0	0	1	-360	360;
	5321	9177	0.001461	0.01201	0	0	0	0	0	0	1	-360	360;
	13650	9177	0.00087	0.00614	0	0	0	0	0	0	1	-360	360;
	13263	6113	0.00464	0.02919	0	0	0	0	0	0	1	-360	360;
	2619	6113	0.004719	0.02719	0	0	0	0	0	0	1	-360	360;
	8240	6113	0.00169	0.01361	0	0	0	0	0	0	1	-360	360;
	3037	6113	0.00751	0.04581	0	0	0	0	0	0	1	-360	360;
	9335	9643	0.00368	0.022469	0	0	0	0	0	0	1	-360	360;
	8240	9643	0.001659	0.012599	0	0	0	0	0	0	1	-360	360;
	11043	9643	0.007341	0.042419	0	0	0	0	0	0	1	-360	360;
	9643	10993	0.00186	0.01456	0	0	0	0	0	0	1	-360	360;
	9104	4911	0.00312	0.03151	0	0	0	0	0	0	1	-360	360;
	13061	9104	0.00073	0.00808	0	0	0	0	0	0	1	-360	360;
	10513	4322	0.013479	0.27064	0	0	0	0	0	0	1	-360	360;
	10513	4322	0.00187	0.020289	0	0	0	0	0	0	1	-360	360;
	13357	4322	0.012169	0.298636	0	0	0	0	0	0	1	-360	360;
	10689	10131	0.00337	0.027831	0	0	0	0	0	0	1	-360	360;
	5745	10689	0.00461	0.03301	0	0	0	0	0	0	1	-360	360;
	10131	10689	0.005781	0.039789	0	0	0	0	0	0	1	-360	360;
	10131	10689	0.035151	0.987397	0	0	0	0	0	0	1	-360	360;
	6712	10689	0.013099	0.631736	0	0	0	0	0	0	1	-360	360;
	4373	3601	0.118409	1.072417	0	0	0	0	0	0	1	-360	360;
	4373	3601	0.006281	0.051961	0	0	0	0	0	0	1	-360	360;
	2926	3601	0.284979	1.491054	0	0	0	0	0	0	1	-360	360;
	8076	4736	0.006409	0.038081	0	0	0	0	0	0	1	-360	360;
	246	8076	0.225806	1.262624	0	0	0	0	0	0	1	-360	360;
	9654	8076	0.282252	1.121529	0	0	0	0	0	0	1	-360	360;
	5298	8076	0.561921	1.55469	0	0	0	0	0	0	1	-360	360;
	10338	5572	0.00322	0.02808	0	0	0	0	0	0	1	-360	360;
	317	5572	0.00196	0.01843	0	0	0	0	0	0	1	-360	360;
	10265	12450	0.552066	1.574607	0	0	0	0	0	0	1	-360	360;
	4850	12450	0.010151	0.087021	0	0	0	0	0	0	1	-360	360;
	10697	12450	0.16999	0.991054	0	0	0	0	0	0	1	-360	360;
	399	12450	0.003831	0.02386	0	0	0	0	0	0	1	-360	360;
	399	12450	0.00863	0.535785	0	0	0	0	0	0	1	-360	360;
	9019	12450	0.04357	0.617335	0	0	0	0	0	0	1	-360	360;
	8333	4051	0.008081	0.03963	0	0	0	0	0	0	1	-360	360;
	4051	2619	0.00213	0.01488	0	0	0	0	0	0	1	-360	360;
	317	6653	0.00219	0.02232	0	0	0	0	0	0	1	-360	360;
	6653	9356	0.00352	0.03947	0	0	0	0	0	0	1	-360	360;
	6653	9162	0.00222	0.02857	0	0	0	0	0	0	1	-360	360;
	537	8472	0.00104	0.00901	0	0	0	0	0	0	1	-360	360;
	3952	537	0.00106	0.00881	0	0	0	0	0	0	1	-360	360;
	9705	5950	0.00301	0.031271	0	0	0	0	0	0	1	-360	360;
	8472	530	0.00133	0.01852	0	0	0	0	0	0	1	-360	360;
	530	7328	0.00253	0.0287	0	0	0	0	0	0	1	-360	360;
	8472	530	0.06395	0.413802	0	0	0	0	0	0	1	-360	360;
	9356	13654	0.00075	0.00981	0	0	0	0	0	0	1	-360	360;
	11590	6644	0.234814	1.148326	0	0	0	0	0	0	1	-360	360;
	13563	6644	0.222417	1.185971	0	0	0	0	0	0	1	-360	360;
	12963	6644	0.08124	0.548574	0	0	0	0	0	0	1	-360	360;
	12963	6644	0.01132	0.074211	0	0	0	0	0	0	1	-360	360;
	1107	6644	0.115849	0.953099	0	0	0	0	0	0	1	-360	360;
	12187	6246	0	0.0002	0	0	0	0	0	0	1	-360	360;
	12187	6246	0.001	0.517348	0	0	0	0	0	0	1	-360	360;
	6246	11045	0.00051	0.00504	0	0	0	0	0	0	1	-360	360;
	6246	7601	0.00258	0.02748	0	0	0	0	0	0	1	-360	360;
	1405	12187	0.00224	0.02852	0	0	0	0	0	0	1	-360	360;
	12187	3952	0.00144	0.01774	0	0	0	0	0	0	1	-360	360;
	8216	12361	0	0.0002	0	0	0	0	0	0	1	-360	360;
	8216	12361	0.06794	0.65593	0	0	0	0	0	0	1	-360	360;
	12361	10513	0.028479	0.136841	0	0	0	0	0	0	1	-360	360;
	8216	5950	0.02019	0.10464	0	0	0	0	0	0	1	-360	360;
	11045	3952	0.00128	0.01333	0	0	0	0	0	0	1	-360	360;
	488	9973	0.667066	1.759628	0	0	0	0	0	0	1	-360	360;
	7101	9973	0.0232	0.385289	0	0	0	0	0	0	1	-360	360;
	7101	9973	0.00686	0.03882	0	0	0	0	0	0	1	-360	360;
	11541	9973	0.04143	0.328037	0	0	0	0	0	0	1	-360	360;
	13004	9973	0.00107	0.00707	0	0	0	0	0	0	1	-360	360;
	7101	488	0.173021	1.092025	0	0	0	0	0	0	1	-360	360;
	7101	488	0.010849	0.074771	0	0	0	0	0	0	1	-360	360;
	843	488	0.18632	0.723574	0	0	0	0	0	0	1	-360	360;
	11297	488	0.077651	0.853368	0	0	0	0	0	0	1	-360	360;
	11297	488	0.011169	0.07225	0	0	0	0	0	0	1	-360	360;
	11541	488	0.14268	0.506488	0	0	0	0	0	0	1	-360	360;
	10697	488	0.352417	1.255207	0	0	0	0	0	0	1	-360	360;
	399	488	0.391322	1.63	0	0	0	0	0	0	1	-360	360;
	4669	3967	0.006271	0.037341	0	0	0	0	0	0	1	-360	360;
	5298	4669	0.005961	0.054719	0	0	0	0	0	0	1	-360	360;
	5153	6887	0	0.0002	0	0	0	0	0	0	1	-360	360;
	5153	6887	0.001	0.333822	0	0	0	0	0	0	1	-360	360;
	6887	5474	0.00587	0.040281	0	0	0	0	0	0	1	-360	360;
	5153	13357	0.003599	0.046841	0	0	0	0	0	0	1	-360	360;
	4243	7601	0.00224	0.02796	0	0	0	0	0	0	1	-360	360;
	3967	12209	0.08693	0.670269	0	0	0	0	0	0	1	-360	360;
	5321	12209	0.301839	1.820372	0	0	0	0	0	0	1	-360	360;
	2109	12209	0.00136	0.01051	0	0	0	0	0	0	1	-360	360;
	2109	12209	0.00445	0.344835	0	0	0	0	0	0	1	-360	360;
	12267	12209	0.197159	1.014752	0	0	0	0	0	0	1	-360	360;
	8333	12209	0.097341	0.78186	0	0	0	0	0	0	1	-360	360;
	4325	12209	0.013039	0.297211	0	0	0	0	0	0	1	-360	360;
	10849	12209	0.035521	0.685351	0	0	0	0	0	0	1	-360	360;
	5867	12209	0.367335	1.746467	0	0	0	0	0	0	1	-360	360;
	4463	12209	0.04186	0.505517	0	0	0	0	0	0	1	-360	360;
	2109	12209	0.001021	0.284401	0	0	0	0	0	0	1	-360	360;
	9811	12209	0.02575	0.548884	0	0	0	0	0	0	1	-360	360;
	10993	12209	0.002211	0.016599	0	0	0	0	0	0	1	-360	360;
	10273	12798	0.0015	0.00876	0	0	0	0	0	0	1	-360	360;
	5867	10273	0.00062	0.00386	0	0	0	0	0	0	1	-360	360;
	9811	13263	0.00188	0.008651	0	0	0	0	0	0	1	-360	360;
	4463	13263	0.00651	0.278678	0	0	0	0	0	0	1	-360	360;
	8240	13263	0.080781	0.659587	0	0	0	0	0	0	1	-360	360;
	7817	10020	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7817	10020	0.001	0.232851	0	0	0	0	0	0	1	-360	360;
	10020	12998	0.01614	0.08307	0	0	0	0	0	0	1	-360	360;
	633	10020	0.190479	0.901095	0	0	0	0	0	0	1	-360	360;
	633	10020	0.00281	0.01663	0	0	0	0	0	0	1	-360	360;
	7690	10020	0.032841	0.593161	0	0	0	0	0	0	1	-360	360;
	12597	10020	0.004599	0.03094	0	0	0	0	0	0	1	-360	360;
	12514	10020	0.100021	0.542231	0	0	0	0	0	0	1	-360	360;
	633	7817	0.1945	0.883884	0	0	0	0	0	0	1	-360	360;
	7690	7817	0.028409	0.617727	0	0	0	0	0	0	1	-360	360;
	7817	7534	0.003281	0.04162	0	0	0	0	0	0	1	-360	360;
	12514	7817	0.09636	0.566219	0	0	0	0	0	0	1	-360	360;
	12514	7817	0.007521	0.044719	0	0	0	0	0	0	1	-360	360;
	13357	10513	0	0.0002	0	0	0	0	0	0	1	-360	360;
	13357	10513	0.001	0.19056	0	0	0	0	0	0	1	-360	360;
	10513	4850	0.01356	0.08876	0	0	0	0	0	0	1	-360	360;
	13357	4850	0.005271	0.068591	0	0	0	0	0	0	1	-360	360;
	4972	12798	0.003591	0.03075	0	0	0	0	0	0	1	-360	360;
	7534	12798	0.01	0.056159	0	0	0	0	0	0	1	-360	360;
	2867	12798	0.011831	0.067969	0	0	0	0	0	0	1	-360	360;
	1998	12798	0.00937	0.05893	0	0	0	0	0	0	1	-360	360;
	12798	10968	0.000909	0.004719	0	0	0	0	0	0	1	-360	360;
	3159	9356	0.00176	0.01759	0	0	0	0	0	0	1	-360	360;
	4318	3159	0.00137	0.01801	0	0	0	0	0	0	1	-360	360;
	3159	11162	0.00184	0.01756	0	0	0	0	0	0	1	-360	360;
	9811	4463	0.00213	0.012539	0	0	0	0	0	0	1	-360	360;
	8240	4463	0.07664	0.597665	0	0	0	0	0	0	1	-360	360;
	2109	4463	0.07132	0.544959	0	0	0	0	0	0	1	-360	360;
	2109	4463	0.00405	0.02507	0	0	0	0	0	0	1	-360	360;
	9811	4463	0.001	0.348058	0	0	0	0	0	0	1	-360	360;
	4325	3967	0.06064	0.52688	0	0	0	0	0	0	1	-360	360;
	3967	4325	0.005271	0.03093	0	0	0	0	0	0	1	-360	360;
	4972	9335	0.002841	0.023409	0	0	0	0	0	0	1	-360	360;
	10187	2867	0.00226	0.015841	0	0	0	0	0	0	1	-360	360;
	9015	2867	0.01237	0.060909	0	0	0	0	0	0	1	-360	360;
	2867	4972	0.01637	0.07345	0	0	0	0	0	0	1	-360	360;
	5874	2867	0.00243	0.018341	0	0	0	0	0	0	1	-360	360;
	5039	1141	0.0026	0.02813	0	0	0	0	0	0	1	-360	360;
	7423	4104	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7423	4104	0.02631	0.515517	0	0	0	0	0	0	1	-360	360;
	4104	10131	0.00394	0.041169	0	0	0	0	0	0	1	-360	360;
	4104	10131	0.343719	1.744855	0	0	0	0	0	0	1	-360	360;
	4104	3037	0.017159	0.080591	0	0	0	0	0	0	1	-360	360;
	7423	8020	0.005	0.066591	0	0	0	0	0	0	1	-360	360;
	7423	10131	0.065229	1.288988	0	0	0	0	0	0	1	-360	360;
	1063	5321	0	0.0002	0	0	0	0	0	0	1	-360	360;
	1063	5321	0.001	0.56	0	0	0	0	0	0	1	-360	360;
	2109	5321	0.063599	0.651508	0	0	0	0	0	0	1	-360	360;
	12267	5321	0.075159	1.271095	0	0	0	0	0	0	1	-360	360;
	8333	5321	0.360475	1.617479	0	0	0	0	0	0	1	-360	360;
	2619	5321	0.001	0.740702	0	0	0	0	0	0	1	-360	360;
	4325	5321	0.363884	2.00314	0	0	0	0	0	0	1	-360	360;
	10849	5321	0.3925	1.858533	0	0	0	0	0	0	1	-360	360;
	2109	1063	0.04769	0.602128	0	0	0	0	0	0	1	-360	360;
	12267	1063	0.050341	1.259855	0	0	0	0	0	0	1	-360	360;
	2619	1063	0.001	0.829566	0	0	0	0	0	0	1	-360	360;
	4850	8031	0.00049	0.00324	0	0	0	0	0	0	1	-360	360;
	8031	5555	0.018659	0.111289	0	0	0	0	0	0	1	-360	360;
	8031	12514	0.509835	1.831384	0	0	0	0	0	0	1	-360	360;
	4850	8031	0.0005	0.00318	0	0	0	0	0	0	1	-360	360;
	9811	2109	0.05936	0.582149	0	0	0	0	0	0	1	-360	360;
	4325	2109	0.003331	0.0203	0	0	0	0	0	0	1	-360	360;
	12267	2109	0.12286	0.888244	0	0	0	0	0	0	1	-360	360;
	8333	2109	0.033599	0.599959	0	0	0	0	0	0	1	-360	360;
	2619	2109	0.100539	0.767066	0	0	0	0	0	0	1	-360	360;
	4325	2109	0.0822	0.730331	0	0	0	0	0	0	1	-360	360;
	10849	2109	0.15512	0.82874	0	0	0	0	0	0	1	-360	360;
	7741	2121	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7741	2121	0.001	0.194781	0	0	0	0	0	0	1	-360	360;
	2121	2754	0.0132	0.071979	0	0	0	0	0	0	1	-360	360;
	2121	561	0.0148	0.073031	0	0	0	0	0	0	1	-360	360;
	7999	2121	0.005781	0.0378	0	0	0	0	0	0	1	-360	360;
	9015	2121	0.153961	1.590021	0	0	0	0	0	0	1	-360	360;
	9015	2121	0.00631	0.03855	0	0	0	0	0	0	1	-360	360;
	13206	2121	0.14124	1.705702	0	0	0	0	0	0	1	-360	360;
	7999	7741	0.00589	0.03699	0	0	0	0	0	0	1	-360	360;
	9015	7741	0.18032	1.449298	0	0	0	0	0	0	1	-360	360;
	13206	7741	0.150651	1.729298	0	0	0	0	0	0	1	-360	360;
	13206	7741	0.00636	0.045409	0	0	0	0	0	0	1	-360	360;
	4071	5525	0.00173	0.01602	0	0	0	0	0	0	1	-360	360;
	6805	4071	0.00207	0.02233	0	0	0	0	0	0	1	-360	360;
	2040	4071	0.00178	0.02178	0	0	0	0	0	0	1	-360	360;
	5874	10187	0.00589	0.214174	0	0	0	0	0	0	1	-360	360;
	9015	10187	0.105789	1.66719	0	0	0	0	0	0	1	-360	360;
	13206	10187	0.108969	1.629731	0	0	0	0	0	0	1	-360	360;
	11541	7101	0.024581	0.231446	0	0	0	0	0	0	1	-360	360;
	11541	7101	0.004599	0.035341	0	0	0	0	0	0	1	-360	360;
	7328	1585	0.00106	0.01023	0	0	0	0	0	0	1	-360	360;
	1597	1585	0.00403	0.04889	0	0	0	0	0	0	1	-360	360;
	13169	8011	0.00123	0.01589	0	0	0	0	0	0	1	-360	360;
	13169	4911	0.00158	0.01592	0	0	0	0	0	0	1	-360	360;
	374	10360	0.00219	0.02123	0	0	0	0	0	0	1	-360	360;
	374	12364	0.00171	0.01708	0	0	0	0	0	0	1	-360	360;
	374	12364	0.00174	0.0167	0	0	0	0	0	0	1	-360	360;
	1597	374	0.00515	0.285942	0	0	0	0	0	0	1	-360	360;
	1597	374	0.00058	0.00775	0	0	0	0	0	0	1	-360	360;
	1597	374	0.00057	0.0079	0	0	0	0	0	0	1	-360	360;
	4972	9811	0.009341	0.6725	0	0	0	0	0	0	1	-360	360;
	5867	9811	0.171229	1.013678	0	0	0	0	0	0	1	-360	360;
	10993	9811	0.00251	0.02049	0	0	0	0	0	0	1	-360	360;
	9811	10968	0.005909	0.036349	0	0	0	0	0	0	1	-360	360;
	11767	11162	0.00091	0.00886	0	0	0	0	0	0	1	-360	360;
	13061	11767	0.00054	0.00532	0	0	0	0	0	0	1	-360	360;
	399	10265	0.009531	0.051211	0	0	0	0	0	0	1	-360	360;
	399	10265	0.785124	1.943678	0	0	0	0	0	0	1	-360	360;
	10265	5560	0.010789	0.05194	0	0	0	0	0	0	1	-360	360;
	9019	10265	0.173281	0.62781	0	0	0	0	0	0	1	-360	360;
	8240	4972	0.00613	0.033659	0	0	0	0	0	0	1	-360	360;
	5867	4972	0.11576	0.840434	0	0	0	0	0	0	1	-360	360;
	11297	843	0	0.0002	0	0	0	0	0	0	1	-360	360;
	11297	843	0.05964	0.314938	0	0	0	0	0	0	1	-360	360;
	843	9320	0.002169	0.017341	0	0	0	0	0	0	1	-360	360;
	611	843	0.454215	1.265475	0	0	0	0	0	0	1	-360	360;
	843	611	0.007091	0.061521	0	0	0	0	0	0	1	-360	360;
	11297	2452	0.013021	0.087711	0	0	0	0	0	0	1	-360	360;
	1363	11297	0.35781	1.954277	0	0	0	0	0	0	1	-360	360;
	611	11297	0.0485	0.815537	0	0	0	0	0	0	1	-360	360;
	2619	8020	0.415537	1.939483	0	0	0	0	0	0	1	-360	360;
	8240	8020	0.097031	0.677913	0	0	0	0	0	0	1	-360	360;
	8020	8240	0.003031	0.033021	0	0	0	0	0	0	1	-360	360;
	8020	8240	0.00637	0.038979	0	0	0	0	0	0	1	-360	360;
	11043	8020	0.024271	0.588967	0	0	0	0	0	0	1	-360	360;
	3037	8020	0.00439	0.372438	0	0	0	0	0	0	1	-360	360;
	314	8020	0.001	0.198331	0	0	0	0	0	0	1	-360	360;
	314	8020	0.00232	0.01226	0	0	0	0	0	0	1	-360	360;
	5560	2452	0.014211	0.079219	0	0	0	0	0	0	1	-360	360;
	246	2452	0.015151	0.07637	0	0	0	0	0	0	1	-360	360;
	4736	2452	0	0.0002	0	0	0	0	0	0	1	-360	360;
	4736	2452	0.001	0.258698	0	0	0	0	0	0	1	-360	360;
	9654	2452	0.01544	0.074781	0	0	0	0	0	0	1	-360	360;
	1363	4736	0.013909	0.073349	0	0	0	0	0	0	1	-360	360;
	7843	7483	0.006919	0.04764	0	0	0	0	0	0	1	-360	360;
	3208	7483	0.224318	1.85469	0	0	0	0	0	0	1	-360	360;
	12963	7483	0.015169	0.1077	0	0	0	0	0	0	1	-360	360;
	561	7483	0.130771	1.06564	0	0	0	0	0	0	1	-360	360;
	7843	7483	0.249669	1.169917	0	0	0	0	0	0	1	-360	360;
	5525	733	0.00127	0.01444	0	0	0	0	0	0	1	-360	360;
	5525	8923	0.00318	0.04235	0	0	0	0	0	0	1	-360	360;
	5525	12852	0.00131	0.01758	0	0	0	0	0	0	1	-360	360;
	6712	4373	0.469607	1.302025	0	0	0	0	0	0	1	-360	360;
	2926	4373	0.001	0.205711	0	0	0	0	0	0	1	-360	360;
	2926	4373	0.002969	0.02201	0	0	0	0	0	0	1	-360	360;
	13004	11541	0.00549	0.044831	0	0	0	0	0	0	1	-360	360;
	2802	3208	0	0.0002	0	0	0	0	0	0	1	-360	360;
	2802	3208	0.08343	1.583099	0	0	0	0	0	0	1	-360	360;
	2754	3208	0.094651	1.363719	0	0	0	0	0	0	1	-360	360;
	3208	12998	0.00112	0.00895	0	0	0	0	0	0	1	-360	360;
	12998	3208	0.00151	0.00944	0	0	0	0	0	0	1	-360	360;
	561	3208	0.018669	0.563368	0	0	0	0	0	0	1	-360	360;
	7843	3208	0.00538	0.652851	0	0	0	0	0	0	1	-360	360;
	2754	2802	0	0.0002	0	0	0	0	0	0	1	-360	360;
	2754	2802	0.001	0.495455	0	0	0	0	0	0	1	-360	360;
	561	2802	0.036849	0.734215	0	0	0	0	0	0	1	-360	360;
	7843	2802	0.001	0.459566	0	0	0	0	0	0	1	-360	360;
	7843	2802	0.00095	0.00638	0	0	0	0	0	0	1	-360	360;
	12998	2754	0.00105	0.00818	0	0	0	0	0	0	1	-360	360;
	561	2754	0.040151	0.657893	0	0	0	0	0	0	1	-360	360;
	561	2754	0.00193	0.01406	0	0	0	0	0	0	1	-360	360;
	7843	2754	0.001	0.49438	0	0	0	0	0	0	1	-360	360;
	8333	12267	0	0.0002	0	0	0	0	0	0	1	-360	360;
	8333	12267	0.08057	0.906178	0	0	0	0	0	0	1	-360	360;
	4325	12267	0.0347	0.658202	0	0	0	0	0	0	1	-360	360;
	10849	12267	0.032719	0.534318	0	0	0	0	0	0	1	-360	360;
	10849	12267	0.004281	0.033461	0	0	0	0	0	0	1	-360	360;
	4325	8333	0.001	0.395785	0	0	0	0	0	0	1	-360	360;
	10849	8333	0.14007	0.838595	0	0	0	0	0	0	1	-360	360;
	12852	733	0.04194	0.384211	0	0	0	0	0	0	1	-360	360;
	733	12852	0.00034	0.00341	0	0	0	0	0	0	1	-360	360;
	1114	11590	0.001969	0.014021	0	0	0	0	0	0	1	-360	360;
	13563	11590	0	0.0002	0	0	0	0	0	0	1	-360	360;
	13563	11590	0.001	0.26219	0	0	0	0	0	0	1	-360	360;
	11590	4316	0.03032	0.14413	0	0	0	0	0	0	1	-360	360;
	13563	1107	0.01325	0.11038	0	0	0	0	0	0	1	-360	360;
	3694	12998	0.011901	0.07319	0	0	0	0	0	0	1	-360	360;
	12998	5555	0.00401	0.03175	0	0	0	0	0	0	1	-360	360;
	12852	4318	0.00368	0.03712	0	0	0	0	0	0	1	-360	360;
	12852	4243	0.00169	0.02158	0	0	0	0	0	0	1	-360	360;
	1141	6805	0.00116	0.01472	0	0	0	0	0	0	1	-360	360;
	317	10360	0.05575	0.602438	0	0	0	0	0	0	1	-360	360;
	10360	317	0.00163	0.01604	0	0	0	0	0	0	1	-360	360;
	10360	4318	0.00269	0.02759	0	0	0	0	0	0	1	-360	360;
	10360	9356	0.00256	0.02877	0	0	0	0	0	0	1	-360	360;
	1107	12963	0	0.0002	0	0	0	0	0	0	1	-360	360;
	1107	12963	0.09549	0.768967	0	0	0	0	0	0	1	-360	360;
	7690	633	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7690	633	0.05899	0.428306	0	0	0	0	0	0	1	-360	360;
	7534	633	0.005591	0.04031	0	0	0	0	0	0	1	-360	360;
	12514	633	0.00714	0.15039	0	0	0	0	0	0	1	-360	360;
	1998	7690	0.005469	0.041099	0	0	0	0	0	0	1	-360	360;
	12514	7690	0.036711	0.380269	0	0	0	0	0	0	1	-360	360;
	7104	317	0.00056	0.124578	0	0	0	0	0	0	1	-360	360;
	7104	317	0.0007	0.00889	0	0	0	0	0	0	1	-360	360;
	4850	3694	0.013479	0.11436	0	0	0	0	0	0	1	-360	360;
	4316	4850	0.012	0.066919	0	0	0	0	0	0	1	-360	360;
	5555	3694	0.3925	1.654463	0	0	0	0	0	0	1	-360	360;
	5874	10131	0.301178	1.948719	0	0	0	0	0	0	1	-360	360;
	5874	10131	0.009169	0.06301	0	0	0	0	0	0	1	-360	360;
	4911	1141	0.00176	0.02342	0	0	0	0	0	0	1	-360	360;
	5421	1141	0.00108	0.01167	0	0	0	0	0	0	1	-360	360;
	5555	561	0.191961	1.234649	0	0	0	0	0	0	1	-360	360;
	7843	561	0.03699	0.374917	0	0	0	0	0	0	1	-360	360;
	1998	12597	0.00361	0.03024	0	0	0	0	0	0	1	-360	360;
	399	10697	0.350806	1.554649	0	0	0	0	0	0	1	-360	360;
	10697	399	0.004831	0.06375	0	0	0	0	0	0	1	-360	360;
	10697	3667	0.007099	0.07076	0	0	0	0	0	0	1	-360	360;
	9019	399	0.0022	0.01318	0	0	0	0	0	0	1	-360	360;
	399	1363	0.01775	0.099419	0	0	0	0	0	0	1	-360	360;
	9019	399	0.186581	1.067335	0	0	0	0	0	0	1	-360	360;
	4318	4243	0.00373	0.05005	0	0	0	0	0	0	1	-360	360;
	5560	1363	0	0.0002	0	0	0	0	0	0	1	-360	360;
	5560	1363	0.001	0.16038	0	0	0	0	0	0	1	-360	360;
	611	1363	0.010531	0.17819	0	0	0	0	0	0	1	-360	360;
	611	1363	0.0025	0.017081	0	0	0	0	0	0	1	-360	360;
	611	5560	0.0777	0.498492	0	0	0	0	0	0	1	-360	360;
	11043	2619	0.01907	1.485992	0	0	0	0	0	0	1	-360	360;
	3037	2619	0.120669	1.890888	0	0	0	0	0	0	1	-360	360;
	13206	7999	0.010271	0.06219	0	0	0	0	0	0	1	-360	360;
	9015	5874	0.298079	1.623781	0	0	0	0	0	0	1	-360	360;
	13206	5874	0.273223	1.612107	0	0	0	0	0	0	1	-360	360;
	4318	13061	0.00277	0.02823	0	0	0	0	0	0	1	-360	360;
	9356	4318	0.00077	0.00747	0	0	0	0	0	0	1	-360	360;
	11162	9162	0.00102	0.01145	0	0	0	0	0	0	1	-360	360;
	11162	9162	0.00112	0.01049	0	0	0	0	0	0	1	-360	360;
	13206	9015	0	0.0002	0	0	0	0	0	0	1	-360	360;
	13206	9015	0.001	0.22343	0	0	0	0	0	0	1	-360	360;
	8011	5421	0	0.0002	0	0	0	0	0	0	1	-360	360;
	8011	5421	0.001	0.26178	0	0	0	0	0	0	1	-360	360;
	4911	5421	0.00153	0.0214	0	0	0	0	0	0	1	-360	360;
	13061	5421	0.00468	0.06143	0	0	0	0	0	0	1	-360	360;
	314	8240	0.271756	1.50345	0	0	0	0	0	0	1	-360	360;
	3037	11043	0	0.0002	0	0	0	0	0	0	1	-360	360;
	3037	11043	0.015271	0.796736	0	0	0	0	0	0	1	-360	360;
	314	11043	0.02776	0.935537	0	0	0	0	0	0	1	-360	360;
	314	11043	0.004219	0.026599	0	0	0	0	0	0	1	-360	360;
	314	3037	0.001	0.620888	0	0	0	0	0	0	1	-360	360;
	7843	5555	0.361364	1.973368	0	0	0	0	0	0	1	-360	360;
	5745	6712	0.001599	0.00825	0	0	0	0	0	0	1	-360	360;
	2926	6712	0.659194	1.990062	0	0	0	0	0	0	1	-360	360;
	9654	246	0	0.0002	0	0	0	0	0	0	1	-360	360;
	9654	246	0.11788	0.822417	0	0	0	0	0	0	1	-360	360;
	5298	246	0.086341	0.745145	0	0	0	0	0	0	1	-360	360;
	5298	9654	0.00569	0.033281	0	0	0	0	0	0	1	-360	360;
	5298	9654	0.03618	0.609897	0	0	0	0	0	0	1	-360	360;
	34	7104	0.00087	0.00898	0	0	0	0	0	0	1	-360	360;
	200	5474	0.001	0.211322	0	0	0	0	0	0	1	-360	360;
	200	5474	0	0.0002	0	0	0	0	0	0	1	-360	360;
	4316	5474	0.13407	1.029876	0	0	0	0	0	0	1	-360	360;
	2617	5474	0.141419	0.956033	0	0	0	0	0	0	1	-360	360;
	4316	200	0.12913	1.151054	0	0	0	0	0	0	1	-360	360;
	2617	200	0.14487	1.008244	0	0	0	0	0	0	1	-360	360;
	200	2617	0.007599	0.05356	0	0	0	0	0	0	1	-360	360;
	10849	4325	0	0.0002	0	0	0	0	0	0	1	-360	360;
	10849	4325	0.024219	0.488657	0	0	0	0	0	0	1	-360	360;
	2617	4316	0.001	0.226901	0	0	0	0	0	0	1	-360	360;
	2617	4316	0	0.0002	0	0	0	0	0	0	1	-360	360;
	10968	5867	0.00063	0.005159	0	0	0	0	0	0	1	-360	360;
	5950	3667	0	0.0002	0	0	0	0	0	0	1	-360	360;
	5950	3667	0.001	0.214277	0	0	0	0	0	0	1	-360	360;
	4873	6248	0.019671	0.07004	0	0	0	0	0	0	1	-360	360;
	4873	6248	0.01812	0.06872	0	0	0	0	0	0	1	-360	360;
	8437	6920	0.001582	0.00724	0	0	0	0	0	0	1	-360	360;
	8437	6920	0.001551	0.007231	0	0	0	0	0	0	1	-360	360;
	8437	6920	0.001262	0.00796	0	0	0	0	0	0	1	-360	360;
	7935	3919	0.0264	0.075129	0	0	0	0	0	0	1	-360	360;
	7935	7693	0.00824	0.044751	0	0	0	0	0	0	1	-360	360;
	11572	6920	0.018271	0.070191	0	0	0	0	0	0	1	-360	360;
	11572	6920	0.021711	0.063129	0	0	0	0	0	0	1	-360	360;
	11572	11064	0.013191	0.0428	0	0	0	0	0	0	1	-360	360;
	13366	3919	0.027631	0.102502	0	0	0	0	0	0	1	-360	360;
	13366	10792	0.005369	0.052622	0	0	0	0	0	0	1	-360	360;
	13366	10792	0.013938	0.0534	0	0	0	0	0	0	1	-360	360;
	13366	10792	0.01364	0.049449	0	0	0	0	0	0	1	-360	360;
	13366	1993	0.007991	0.037738	0	0	0	0	0	0	1	-360	360;
	6349	6920	0.007671	0.04096	0	0	0	0	0	0	1	-360	360;
	6349	6920	0.007511	0.041769	0	0	0	0	0	0	1	-360	360;
	6349	7566	0.004089	0.010978	0	0	0	0	0	0	1	-360	360;
	6349	7566	0.003	0.014769	0	0	0	0	0	0	1	-360	360;
	6349	891	0.001351	0.00784	0	0	0	0	0	0	1	-360	360;
	6349	8438	0.003671	0.022089	0	0	0	0	0	0	1	-360	360;
	6349	4251	0.002378	0.00772	0	0	0	0	0	0	1	-360	360;
	6920	891	0.006911	0.03268	0	0	0	0	0	0	1	-360	360;
	6920	3030	0.00944	0.03284	0	0	0	0	0	0	1	-360	360;
	6920	6378	0.009831	0.02848	0	0	0	0	0	0	1	-360	360;
	6920	8438	0.003671	0.02092	0	0	0	0	0	0	1	-360	360;
	12999	3919	0.019378	0.10932	0	0	0	0	0	0	1	-360	360;
	12999	3919	0.021969	0.107129	0	0	0	0	0	0	1	-360	360;
	12999	7693	0.009151	0.063249	0	0	0	0	0	0	1	-360	360;
	12999	6050	0.029169	0.088102	0	0	0	0	0	0	1	-360	360;
	12999	7135	0.031351	0.115698	0	0	0	0	0	0	1	-360	360;
	12999	10792	0.04936	0.182018	0	0	0	0	0	0	1	-360	360;
	12999	9259	0.001609	0.016889	0	0	0	0	0	0	1	-360	360;
	12999	4033	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1022	13210	0.008698	0.033591	0	0	0	0	0	0	1	-360	360;
	8027	2649	0.01792	0.070169	0	0	0	0	0	0	1	-360	360;
	8027	12751	0.00308	0.008422	0	0	0	0	0	0	1	-360	360;
	8027	810	0.002031	0.012102	0	0	0	0	0	0	1	-360	360;
	8027	810	0.0068	0.03452	0	0	0	0	0	0	1	-360	360;
	8027	5565	0.01176	0.055871	0	0	0	0	0	0	1	-360	360;
	6379	6378	0.04856	0.135111	0	0	0	0	0	0	1	-360	360;
	3919	9044	0.039289	0.146129	0	0	0	0	0	0	1	-360	360;
	3919	9044	0.038471	0.148938	0	0	0	0	0	0	1	-360	360;
	3919	4981	0.010911	0.04116	0	0	0	0	0	0	1	-360	360;
	9044	1993	0.002898	0.017409	0	0	0	0	0	0	1	-360	360;
	7323	1767	0.00544	0.051431	0	0	0	0	0	0	1	-360	360;
	7323	7775	0.003791	0.014342	0	0	0	0	0	0	1	-360	360;
	7323	8630	0.005969	0.026302	0	0	0	0	0	0	1	-360	360;
	7323	1254	0.00608	0.02572	0	0	0	0	0	0	1	-360	360;
	7323	3809	0.010818	0.066769	0	0	0	0	0	0	1	-360	360;
	2649	3445	0.006649	0.024138	0	0	0	0	0	0	1	-360	360;
	2649	3445	0.007542	0.023662	0	0	0	0	0	0	1	-360	360;
	2649	3445	0.007991	0.036622	0	0	0	0	0	0	1	-360	360;
	2649	3445	0.0116	0.046018	0	0	0	0	0	0	1	-360	360;
	6050	3030	0.011631	0.047049	0	0	0	0	0	0	1	-360	360;
	7135	10792	0.018822	0.064009	0	0	0	0	0	0	1	-360	360;
	1767	8630	0.008071	0.049138	0	0	0	0	0	0	1	-360	360;
	1767	1254	0.007889	0.045502	0	0	0	0	0	0	1	-360	360;
	1767	12185	0.009822	0.057222	0	0	0	0	0	0	1	-360	360;
	3445	5461	0.00656	0.023831	0	0	0	0	0	0	1	-360	360;
	3445	5461	0.007431	0.023351	0	0	0	0	0	0	1	-360	360;
	3445	10863	0.007391	0.026978	0	0	0	0	0	0	1	-360	360;
	3445	6269	0.001871	0.012142	0	0	0	0	0	0	1	-360	360;
	3445	1398	0.003969	0.026169	0	0	0	0	0	0	1	-360	360;
	3445	3140	0.008151	0.04664	0	0	0	0	0	0	1	-360	360;
	12062	49	0.00076	0.005622	0	0	0	0	0	0	1	-360	360;
	12062	5461	0.0094	0.035791	0	0	0	0	0	0	1	-360	360;
	12062	11425	0.00844	0.031031	0	0	0	0	0	0	1	-360	360;
	12062	2404	0.00028	0.002062	0	0	0	0	0	0	1	-360	360;
	12062	4146	0.004849	0.030031	0	0	0	0	0	0	1	-360	360;
	12062	6269	0.00268	0.022391	0	0	0	0	0	0	1	-360	360;
	12062	1398	0.002391	0.025382	0	0	0	0	0	0	1	-360	360;
	810	5461	0.017551	0.078191	0	0	0	0	0	0	1	-360	360;
	9519	2208	0.003849	0.02475	0	0	0	0	0	0	1	-360	360;
	9519	10291	0.0042	0.02005	0	0	0	0	0	0	1	-360	360;
	9519	5259	0.00011	0.000581	0	0	0	0	0	0	1	-360	360;
	9519	1129	0.000169	0.000651	0	0	0	0	0	0	1	-360	360;
	9519	3089	0.000169	0.000659	0	0	0	0	0	0	1	-360	360;
	11655	9126	0.00219	0.01125	0	0	0	0	0	0	1	-360	360;
	11655	13245	0.00018	0.000599	0	0	0	0	0	0	1	-360	360;
	11655	13120	0.00018	0.000591	0	0	0	0	0	0	1	-360	360;
	11655	2471	0.00019	0.000581	0	0	0	0	0	0	1	-360	360;
	11655	4759	0.000169	0.000651	0	0	0	0	0	0	1	-360	360;
	3827	2814	0.00231	0.0138	0	0	0	0	0	0	1	-360	360;
	3827	2814	0.00082	0.003841	0	0	0	0	0	0	1	-360	360;
	3827	254	0.00305	0.01762	0	0	0	0	0	0	1	-360	360;
	8247	7778	0.01101	0.055469	0	0	0	0	0	0	1	-360	360;
	8247	1849	0.009169	0.0462	0	0	0	0	0	0	1	-360	360;
	8247	10291	0.00201	0.00905	0	0	0	0	0	0	1	-360	360;
	8247	2859	0.005969	0.038841	0	0	0	0	0	0	1	-360	360;
	8247	11381	0.00474	0.02687	0	0	0	0	0	0	1	-360	360;
	8247	8082	0.005409	0.03438	0	0	0	0	0	0	1	-360	360;
	8247	4654	0.005401	0.03264	0	0	0	0	0	0	1	-360	360;
	8247	13253	0.00311	0.015919	0	0	0	0	0	0	1	-360	360;
	8247	2531	0.00462	0.02613	0	0	0	0	0	0	1	-360	360;
	4609	1596	0.01044	0.05187	0	0	0	0	0	0	1	-360	360;
	4609	1596	0.01001	0.05113	0	0	0	0	0	0	1	-360	360;
	4609	12530	0.008331	0.051729	0	0	0	0	0	0	1	-360	360;
	4609	960	0.000169	0.00089	0	0	0	0	0	0	1	-360	360;
	4609	5860	0.00214	0.01099	0	0	0	0	0	0	1	-360	360;
	2814	8480	0.004669	0.042349	0	0	0	0	0	0	1	-360	360;
	2814	10624	0.00139	0.008789	0	0	0	0	0	0	1	-360	360;
	2814	10624	0.000219	0.001669	0	0	0	0	0	0	1	-360	360;
	2814	254	0.00119	0.01063	0	0	0	0	0	0	1	-360	360;
	5100	2859	0.004169	0.02724	0	0	0	0	0	0	1	-360	360;
	5100	2859	0.004081	0.025219	0	0	0	0	0	0	1	-360	360;
	5100	468	0.00168	0.01039	0	0	0	0	0	0	1	-360	360;
	5100	2118	0.0017	0.01055	0	0	0	0	0	0	1	-360	360;
	10949	10161	0.00139	0.008771	0	0	0	0	0	0	1	-360	360;
	10949	9330	0.000651	0.00261	0	0	0	0	0	0	1	-360	360;
	10949	9330	0.000729	0.00256	0	0	0	0	0	0	1	-360	360;
	10949	1754	0	0.008169	0	0	0	0	0	0	1	-360	360;
	10949	8227	5e-05	0.00025	0	0	0	0	0	0	1	-360	360;
	8259	1849	0.003719	0.022669	0	0	0	0	0	0	1	-360	360;
	8259	9390	0.00732	0.042711	0	0	0	0	0	0	1	-360	360;
	8259	9390	0.003599	0.025331	0	0	0	0	0	0	1	-360	360;
	8259	13301	0.00494	0.027401	0	0	0	0	0	0	1	-360	360;
	7778	9925	0.0062	0.03519	0	0	0	0	0	0	1	-360	360;
	7778	9925	0.00739	0.04574	0	0	0	0	0	0	1	-360	360;
	7778	4654	0.006781	0.034289	0	0	0	0	0	0	1	-360	360;
	9925	4401	0.003669	0.03205	0	0	0	0	0	0	1	-360	360;
	9925	4401	0.002789	0.026289	0	0	0	0	0	0	1	-360	360;
	9925	3679	0.005979	0.031979	0	0	0	0	0	0	1	-360	360;
	4401	9126	0.00425	0.02132	0	0	0	0	0	0	1	-360	360;
	4401	2208	0.002469	0.015099	0	0	0	0	0	0	1	-360	360;
	4401	13245	0.004341	0.02662	0	0	0	0	0	0	1	-360	360;
	4401	4759	0.005219	0.02287	0	0	0	0	0	0	1	-360	360;
	4401	11149	0.0095	0.04693	0	0	0	0	0	0	1	-360	360;
	8480	254	0.01099	0.052909	0	0	0	0	0	0	1	-360	360;
	8480	9390	0.004901	0.02949	0	0	0	0	0	0	1	-360	360;
	8480	6951	0.00832	0.038531	0	0	0	0	0	0	1	-360	360;
	8480	9050	0.003909	0.018091	0	0	0	0	0	0	1	-360	360;
	8480	2846	0.000331	0.00189	0	0	0	0	0	0	1	-360	360;
	1849	9390	0.01101	0.05514	0	0	0	0	0	0	1	-360	360;
	1849	10982	0.00345	0.01925	0	0	0	0	0	0	1	-360	360;
	13464	254	0.00162	0.00889	0	0	0	0	0	0	1	-360	360;
	13464	13301	0.00039	0.00464	0	0	0	0	0	0	1	-360	360;
	13464	7806	0.00064	0.010099	0	0	0	0	0	0	1	-360	360;
	9126	10161	0.003151	0.01613	0	0	0	0	0	0	1	-360	360;
	9126	10161	0.002091	0.012409	0	0	0	0	0	0	1	-360	360;
	9126	13513	0	0.00799	0	0	0	0	0	0	1	-360	360;
	9126	9305	0	0.00906	0	0	0	0	0	0	1	-360	360;
	9126	10284	0.000909	0.010401	0	0	0	0	0	0	1	-360	360;
	9126	3584	0.003211	0.016081	0	0	0	0	0	0	1	-360	360;
	9126	13120	0.00314	0.017711	0	0	0	0	0	0	1	-360	360;
	9126	2471	0.00214	0.010591	0	0	0	0	0	0	1	-360	360;
	5094	4654	0.00056	0.003169	0	0	0	0	0	0	1	-360	360;
	10624	10599	0.001719	0.0108	0	0	0	0	0	0	1	-360	360;
	10161	5813	0.0005	0.00174	0	0	0	0	0	0	1	-360	360;
	10161	8227	0.00126	0.00587	0	0	0	0	0	0	1	-360	360;
	10169	3679	0.007909	0.03607	0	0	0	0	0	0	1	-360	360;
	10169	10284	0.00057	0.00457	0	0	0	0	0	0	1	-360	360;
	10169	8227	0.00132	0.008409	0	0	0	0	0	0	1	-360	360;
	254	3532	0.003581	0.017289	0	0	0	0	0	0	1	-360	360;
	254	6951	0.002711	0.01526	0	0	0	0	0	0	1	-360	360;
	254	9217	0.00039	0.006081	0	0	0	0	0	0	1	-360	360;
	254	7806	0.00037	0.005409	0	0	0	0	0	0	1	-360	360;
	254	10599	0.00244	0.013789	0	0	0	0	0	0	1	-360	360;
	5813	3584	0.000349	0.00149	0	0	0	0	0	0	1	-360	360;
	3532	9390	0.00212	0.010531	0	0	0	0	0	0	1	-360	360;
	563	7785	0.003039	0.031151	0	0	0	0	0	0	1	-360	360;
	563	9050	0.001831	0.0187	0	0	0	0	0	0	1	-360	360;
	2631	7491	0.006599	0.043169	0	0	0	0	0	0	1	-360	360;
	2631	7443	0.006729	0.046599	0	0	0	0	0	0	1	-360	360;
	2631	5339	0.002159	0.010789	0	0	0	0	0	0	1	-360	360;
	2631	1134	0.0022	0.01057	0	0	0	0	0	0	1	-360	360;
	2631	1490	0.003461	0.02049	0	0	0	0	0	0	1	-360	360;
	11795	1016	0.00311	0.019469	0	0	0	0	0	0	1	-360	360;
	11795	915	0.01064	0.065969	0	0	0	0	0	0	1	-360	360;
	1596	468	0.003539	0.020479	0	0	0	0	0	0	1	-360	360;
	1596	2118	0.00357	0.020719	0	0	0	0	0	0	1	-360	360;
	1596	7491	0.00275	0.013831	0	0	0	0	0	0	1	-360	360;
	1596	7443	0.0028	0.013531	0	0	0	0	0	0	1	-360	360;
	9390	9050	0.005151	0.04851	0	0	0	0	0	0	1	-360	360;
	9390	12530	0.005401	0.025281	0	0	0	0	0	0	1	-360	360;
	9390	12530	0.0055	0.02475	0	0	0	0	0	0	1	-360	360;
	10982	8082	0.00093	0.00538	0	0	0	0	0	0	1	-360	360;
	10291	2859	0.010331	0.061849	0	0	0	0	0	0	1	-360	360;
	12487	5339	0.00686	0.032771	0	0	0	0	0	0	1	-360	360;
	12487	1134	0.007	0.032081	0	0	0	0	0	0	1	-360	360;
	7785	6120	0.007729	0.038099	0	0	0	0	0	0	1	-360	360;
	6120	1016	0.006581	0.035281	0	0	0	0	0	0	1	-360	360;
	6120	1016	0.003711	0.02651	0	0	0	0	0	0	1	-360	360;
	6120	6356	0.00482	0.021979	0	0	0	0	0	0	1	-360	360;
	6120	6356	0.003979	0.026099	0	0	0	0	0	0	1	-360	360;
	2859	13253	0.0038	0.017271	0	0	0	0	0	0	1	-360	360;
	2859	11216	0.00626	0.03626	0	0	0	0	0	0	1	-360	360;
	12530	915	0.00538	0.02588	0	0	0	0	0	0	1	-360	360;
	12530	915	0.005479	0.025341	0	0	0	0	0	0	1	-360	360;
	186	13247	0.00134	0.01235	0	0	0	0	0	0	1	-360	360;
	186	789	0.00127	0.01399	0	0	0	0	0	0	1	-360	360;
	186	11488	0.00142	0.01374	0	0	0	0	0	0	1	-360	360;
	11490	11719	0.00039	0.00365	0	0	0	0	0	0	1	-360	360;
	11490	5201	0.00058	0.00507	0	0	0	0	0	0	1	-360	360;
	1463	9196	0.00129	0.01577	0	0	0	0	0	0	1	-360	360;
	1463	4330	0.00033	0.00373	0	0	0	0	0	0	1	-360	360;
	1463	5331	0.00086	0.00799	0	0	0	0	0	0	1	-360	360;
	1900	5331	0.00189	0.02135	0	0	0	0	0	0	1	-360	360;
	1900	5483	0.00106	0.01219	0	0	0	0	0	0	1	-360	360;
	1900	11869	0.00147	0.01622	0	0	0	0	0	0	1	-360	360;
	1900	9534	0.00215	0.02108	0	0	0	0	0	0	1	-360	360;
	647	3465	0.00063	0.0052	0	0	0	0	0	0	1	-360	360;
	647	5331	0.00159	0.01792	0	0	0	0	0	0	1	-360	360;
	9196	5483	0.00157	0.01981	0	0	0	0	0	0	1	-360	360;
	9196	11869	0.00158	0.02014	0	0	0	0	0	0	1	-360	360;
	3465	13247	0.0018	0.02388	0	0	0	0	0	0	1	-360	360;
	3465	801	0.00034	0.00318	0	0	0	0	0	0	1	-360	360;
	3465	10345	0.00015	0.00132	0	0	0	0	0	0	1	-360	360;
	3465	12695	0.00014	0.0015	0	0	0	0	0	0	1	-360	360;
	3465	1655	0.00013	0.00153	0	0	0	0	0	0	1	-360	360;
	3465	4742	0.00013	0.00156	0	0	0	0	0	0	1	-360	360;
	3465	8461	0.0004	0.00449	0	0	0	0	0	0	1	-360	360;
	3465	8461	0.00045	0.0044	0	0	0	0	0	0	1	-360	360;
	13247	2715	0.0023	0.01984	0	0	0	0	0	0	1	-360	360;
	13247	13348	0.0002	0.00235	0	0	0	0	0	0	1	-360	360;
	13247	8743	0.00019	0.00239	0	0	0	0	0	0	1	-360	360;
	13247	11445	0.00169	0.01602	0	0	0	0	0	0	1	-360	360;
	13247	12752	0.00024	0.00253	0	0	0	0	0	0	1	-360	360;
	13247	12044	0.00022	0.00244	0	0	0	0	0	0	1	-360	360;
	13247	1977	0.00025	0.00239	0	0	0	0	0	0	1	-360	360;
	12100	11719	0.00066	0.0061	0	0	0	0	0	0	1	-360	360;
	12100	11719	0.00062	0.00844	0	0	0	0	0	0	1	-360	360;
	12100	11547	0.00124	0.01353	0	0	0	0	0	0	1	-360	360;
	12100	9043	0.00046	0.00452	0	0	0	0	0	0	1	-360	360;
	4330	5331	0.00045	0.00531	0	0	0	0	0	0	1	-360	360;
	4330	801	0.00123	0.01433	0	0	0	0	0	0	1	-360	360;
	4330	11445	0.00174	0.01696	0	0	0	0	0	0	1	-360	360;
	7038	9043	0.00086	0.00877	0	0	0	0	0	0	1	-360	360;
	7038	5201	0.00096	0.00999	0	0	0	0	0	0	1	-360	360;
	7038	7217	0.00082	0.00761	0	0	0	0	0	0	1	-360	360;
	11719	9534	0.00185	0.01803	0	0	0	0	0	0	1	-360	360;
	11719	9511	0.00138	0.01246	0	0	0	0	0	0	1	-360	360;
	11719	11737	0.00062	0.00605	0	0	0	0	0	0	1	-360	360;
	11719	5201	0.00073	0.00817	0	0	0	0	0	0	1	-360	360;
	789	13219	0.00094	0.00999	0	0	0	0	0	0	1	-360	360;
	789	5866	0.00058	0.00671	0	0	0	0	0	0	1	-360	360;
	5483	9534	0.00108	0.01279	0	0	0	0	0	0	1	-360	360;
	5483	11737	0.00225	0.02472	0	0	0	0	0	0	1	-360	360;
	5483	6664	0.00019	0.00244	0	0	0	0	0	0	1	-360	360;
	5483	9272	0.00018	0.00248	0	0	0	0	0	0	1	-360	360;
	5483	397	0.0001	0.0011	0	0	0	0	0	0	1	-360	360;
	5483	2921	0.00011	0.00108	0	0	0	0	0	0	1	-360	360;
	11547	9043	0.00089	0.00815	0	0	0	0	0	0	1	-360	360;
	8865	11869	0.0009	0.00859	0	0	0	0	0	0	1	-360	360;
	8865	13219	0.00167	0.02239	0	0	0	0	0	0	1	-360	360;
	11869	7679	2e-05	0.00024	0	0	0	0	0	0	1	-360	360;
	9534	195	0.00157	0.01326	0	0	0	0	0	0	1	-360	360;
	2715	8085	0.00155	0.0139	0	0	0	0	0	0	1	-360	360;
	195	9511	0.00142	0.01472	0	0	0	0	0	0	1	-360	360;
	6592	9511	0.00187	0.02232	0	0	0	0	0	0	1	-360	360;
	6592	2240	0.00049	0.00572	0	0	0	0	0	0	1	-360	360;
	9511	2848	0	0.00021	0	0	0	0	0	0	1	-360	360;
	13219	5866	0.00104	0.00939	0	0	0	0	0	0	1	-360	360;
	13219	10952	0.00043	0.0049	0	0	0	0	0	0	1	-360	360;
	13219	10952	0.00049	0.0048	0	0	0	0	0	0	1	-360	360;
	49	1681	0.002009	0.01464	0	0	0	0	0	0	1	-360	360;
	1681	5461	0.003302	0.026	0	0	0	0	0	0	1	-360	360;
	574	6248	0.006222	0.018089	0	0	0	0	0	0	1	-360	360;
	5461	10863	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5461	4146	8.9e-05	0.000911	0	0	0	0	0	0	1	-360	360;
	12976	11425	1.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	9451	10863	0.00388	0.014289	0	0	0	0	0	0	1	-360	360;
	8168	891	0.001329	0.007898	0	0	0	0	0	0	1	-360	360;
	8168	6711	0.00048	0.002951	0	0	0	0	0	0	1	-360	360;
	8168	13344	8e-05	0.00032	0	0	0	0	0	0	1	-360	360;
	4093	7775	0.000302	0.001102	0	0	0	0	0	0	1	-360	360;
	6378	11374	0.003742	0.013098	0	0	0	0	0	0	1	-360	360;
	11064	8973	0.008911	0.032298	0	0	0	0	0	0	1	-360	360;
	8630	9569	0.00124	0.005409	0	0	0	0	0	0	1	-360	360;
	1016	8230	0.0025	0.0123	0	0	0	0	0	0	1	-360	360;
	1016	8230	0.001781	0.013729	0	0	0	0	0	0	1	-360	360;
	1016	6356	0.00324	0.016039	0	0	0	0	0	0	1	-360	360;
	13301	9217	0.00037	0.00474	0	0	0	0	0	0	1	-360	360;
	11381	2531	9.1e-05	0.000521	0	0	0	0	0	0	1	-360	360;
	9217	7806	3.1e-05	0.00043	0	0	0	0	0	0	1	-360	360;
	2531	11300	0.001281	0.005909	0	0	0	0	0	0	1	-360	360;
	2118	8414	0.000909	0.00476	0	0	0	0	0	0	1	-360	360;
	7443	7683	0.00011	0.000599	0	0	0	0	0	0	1	-360	360;
	5339	13451	0.000289	0.001719	0	0	0	0	0	0	1	-360	360;
	10599	6029	0.00026	0.001591	0	0	0	0	0	0	1	-360	360;
	8461	5866	0.00108	0.01382	0	0	0	0	0	0	1	-360	360;
	10741	11216	0	0.0002	0	0	0	0	0	0	1	-360	360;
	6972	11516	0.014281	0.033446	0	0	0	0	0	0	1	-360	360;
	12931	11516	0.038273	0.11138	0	0	0	0	0	0	1	-360	360;
	3999	1890	0.039504	0.11795	0	0	0	0	0	0	1	-360	360;
	4188	3999	0.039893	0.095083	0	0	0	0	0	0	1	-360	360;
	6752	13155	0.025107	0.052256	0	0	0	0	0	0	1	-360	360;
	2522	6752	0.010124	0.017818	0	0	0	0	0	0	1	-360	360;
	13564	9982	0.035198	0.083893	0	0	0	0	0	0	1	-360	360;
	9982	9539	0.026149	0.064851	0	0	0	0	0	0	1	-360	360;
	13141	10872	0	0.000793	0	0	0	0	0	0	1	-360	360;
	5513	13141	0.02086	0.051744	0	0	0	0	0	0	1	-360	360;
	736	10872	0.041992	0.117446	0	0	0	0	0	0	1	-360	360;
	8186	7395	0.014031	0.079591	0	0	0	0	0	0	1	-360	360;
	7395	12413	0.0075	0.054099	0	0	0	0	0	0	1	-360	360;
	5020	7395	0.013841	0.0733	0	0	0	0	0	0	1	-360	360;
	7395	2954	0.006469	0.041401	0	0	0	0	0	0	1	-360	360;
	677	3726	0.005628	0.015826	0	0	0	0	0	0	1	-360	360;
	12776	677	0.026364	0.056017	0	0	0	0	0	0	1	-360	360;
	677	5513	0.026099	0.062182	0	0	0	0	0	0	1	-360	360;
	677	7623	0.04481	0.131231	0	0	0	0	0	0	1	-360	360;
	736	3726	0.005512	0.01614	0	0	0	0	0	0	1	-360	360;
	12776	736	0.023488	0.063479	0	0	0	0	0	0	1	-360	360;
	736	9121	0.044198	0.129446	0	0	0	0	0	0	1	-360	360;
	9211	2018	0.006612	0.01657	0	0	0	0	0	0	1	-360	360;
	3933	9211	0.003967	0.007521	0	0	0	0	0	0	1	-360	360;
	8450	9270	0.027496	0.067438	0	0	0	0	0	0	1	-360	360;
	7040	8450	0.044587	0.102702	0	0	0	0	0	0	1	-360	360;
	8450	6579	0.051438	0.143868	0	0	0	0	0	0	1	-360	360;
	8450	11697	0.029256	0.072579	0	0	0	0	0	0	1	-360	360;
	1316	8450	0.049149	0.122653	0	0	0	0	0	0	1	-360	360;
	1316	8450	0.043372	0.125157	0	0	0	0	0	0	1	-360	360;
	12851	8450	0.016471	0.042529	0	0	0	0	0	0	1	-360	360;
	10960	90	0	0.00076	0	0	0	0	0	0	1	-360	360;
	90	3016	0.038017	0.110653	0	0	0	0	0	0	1	-360	360;
	7497	90	0.025174	0.07938	0	0	0	0	0	0	1	-360	360;
	7964	90	0.043017	0.128446	0	0	0	0	0	0	1	-360	360;
	6175	10960	0.051636	0.189413	0	0	0	0	0	0	1	-360	360;
	10960	5175	0.04714	0.113008	0	0	0	0	0	0	1	-360	360;
	4064	10960	0.008149	0.025711	0	0	0	0	0	0	1	-360	360;
	6652	12237	0.039686	0.102843	0	0	0	0	0	0	1	-360	360;
	10393	12237	0.006372	0.01857	0	0	0	0	0	0	1	-360	360;
	12237	6830	0.045711	0.108901	0	0	0	0	0	0	1	-360	360;
	1699	2094	0.04119	0.106331	0	0	0	0	0	0	1	-360	360;
	4913	1699	0.013752	0.03024	0	0	0	0	0	0	1	-360	360;
	8512	6888	0.00595	0.019008	0	0	0	0	0	0	1	-360	360;
	6888	11651	0.00576	0.020331	0	0	0	0	0	0	1	-360	360;
	4097	11791	0.03586	0.082107	0	0	0	0	0	0	1	-360	360;
	630	4097	0.022223	0.055099	0	0	0	0	0	0	1	-360	360;
	12083	9714	0.008802	0.02862	0	0	0	0	0	0	1	-360	360;
	9714	11041	0.032719	0.124942	0	0	0	0	0	0	1	-360	360;
	1103	12115	0.019793	0.055331	0	0	0	0	0	0	1	-360	360;
	12115	12125	0.01395	0.034612	0	0	0	0	0	0	1	-360	360;
	5427	3149	0.004727	0.010107	0	0	0	0	0	0	1	-360	360;
	12909	5427	0.007289	0.014364	0	0	0	0	0	0	1	-360	360;
	8012	5944	0.004818	0.009901	0	0	0	0	0	0	1	-360	360;
	688	8012	0.012223	0.031554	0	0	0	0	0	0	1	-360	360;
	1925	763	0.00108	0.01189	0	0	0	0	0	0	1	-360	360;
	1371	1925	0.0032	0.03265	0	0	0	0	0	0	1	-360	360;
	9813	1925	0.0007	0.00731	0	0	0	0	0	0	1	-360	360;
	3640	8341	0	0.000744	0	0	0	0	0	0	1	-360	360;
	3640	2706	0.009248	0.022041	0	0	0	0	0	0	1	-360	360;
	8341	12672	0.008124	0.02286	0	0	0	0	0	0	1	-360	360;
	6947	11002	0.007702	0.019893	0	0	0	0	0	0	1	-360	360;
	11002	10088	0.015579	0.04914	0	0	0	0	0	0	1	-360	360;
	13051	4510	0.00981	0.068021	0	0	0	0	0	0	1	-360	360;
	13051	2416	0.006341	0.042211	0	0	0	0	0	0	1	-360	360;
	1403	10539	0.060281	0.144512	0	0	0	0	0	0	1	-360	360;
	1403	10539	0.061438	0.141504	0	0	0	0	0	0	1	-360	360;
	10539	3933	0.016579	0.039917	0	0	0	0	0	0	1	-360	360;
	2258	5488	0	0.000843	0	0	0	0	0	0	1	-360	360;
	9557	2258	0.010256	0.022711	0	0	0	0	0	0	1	-360	360;
	2258	7127	0.024983	0.073149	0	0	0	0	0	0	1	-360	360;
	5488	6253	0.008388	0.02557	0	0	0	0	0	0	1	-360	360;
	3685	12183	0.063512	0.200339	0	0	0	0	0	0	1	-360	360;
	6652	3685	0.037562	0.085967	0	0	0	0	0	0	1	-360	360;
	4362	1081	0.002579	0.007752	0	0	0	0	0	0	1	-360	360;
	1945	1081	0.001777	0.00495	0	0	0	0	0	0	1	-360	360;
	9671	10682	0.030248	0.084612	0	0	0	0	0	0	1	-360	360;
	10682	12827	0.013281	0.037124	0	0	0	0	0	0	1	-360	360;
	7961	5590	0.009488	0.023521	0	0	0	0	0	0	1	-360	360;
	953	5590	0.022579	0.058314	0	0	0	0	0	0	1	-360	360;
	6687	7089	0.02743	0.062793	0	0	0	0	0	0	1	-360	360;
	10577	6687	0.001719	0.003777	0	0	0	0	0	0	1	-360	360;
	3972	2833	0.035992	0.085876	0	0	0	0	0	0	1	-360	360;
	3501	3972	0.008777	0.01962	0	0	0	0	0	0	1	-360	360;
	4361	3249	0.001793	0.001521	0	0	0	0	0	0	1	-360	360;
	3249	3250	0.004917	0.007702	0	0	0	0	0	0	1	-360	360;
	10421	4175	0.001521	0.004554	0	0	0	0	0	0	1	-360	360;
	744	4175	0.011198	0.041099	0	0	0	0	0	0	1	-360	360;
	2469	12012	0.00144	0.01155	0	0	0	0	0	0	1	-360	360;
	2469	10698	0.00123	0.01345	0	0	0	0	0	0	1	-360	360;
	9101	9118	0.007669	0.02419	0	0	0	0	0	0	1	-360	360;
	9101	268	0.003661	0.008769	0	0	0	0	0	0	1	-360	360;
	12774	9101	0.00943	0.036008	0	0	0	0	0	0	1	-360	360;
	848	9101	0.012661	0.037802	0	0	0	0	0	0	1	-360	360;
	11800	12242	0.006083	0.024099	0	0	0	0	0	0	1	-360	360;
	11800	1660	0.002248	0.004248	0	0	0	0	0	0	1	-360	360;
	2511	6954	0.019651	0.141771	0	0	0	0	0	0	1	-360	360;
	6954	5372	0.004841	0.02718	0	0	0	0	0	0	1	-360	360;
	6954	5372	0.00493	0.02661	0	0	0	0	0	0	1	-360	360;
	6343	6954	0.008729	0.04569	0	0	0	0	0	0	1	-360	360;
	13098	10326	0.002529	0.005793	0	0	0	0	0	0	1	-360	360;
	10326	8918	0.003579	0.007876	0	0	0	0	0	0	1	-360	360;
	12077	1205	0.014182	0.03119	0	0	0	0	0	0	1	-360	360;
	12077	9941	0.061124	0.192777	0	0	0	0	0	0	1	-360	360;
	7422	1091	0.114603	0.284256	0	0	0	0	0	0	1	-360	360;
	1091	11147	0.00786	0.019488	0	0	0	0	0	0	1	-360	360;
	13098	12685	0.002669	0.006124	0	0	0	0	0	0	1	-360	360;
	5351	7373	0.00895	0.022182	0	0	0	0	0	0	1	-360	360;
	7373	348	0.002281	0.008033	0	0	0	0	0	0	1	-360	360;
	5351	1205	0.005314	0.013719	0	0	0	0	0	0	1	-360	360;
	9131	7927	0.031521	0.069298	0	0	0	0	0	0	1	-360	360;
	7927	12685	0.003719	0.013223	0	0	0	0	0	0	1	-360	360;
	7927	348	0.009719	0.027818	0	0	0	0	0	0	1	-360	360;
	7927	5811	0.042669	0.127397	0	0	0	0	0	0	1	-360	360;
	7927	9816	0.031157	0.077281	0	0	0	0	0	0	1	-360	360;
	8918	1205	0.00319	0.008934	0	0	0	0	0	0	1	-360	360;
	1205	9644	0.01462	0.046124	0	0	0	0	0	0	1	-360	360;
	12147	6492	0.007587	0.018628	0	0	0	0	0	0	1	-360	360;
	10843	12147	0.024529	0.052826	0	0	0	0	0	0	1	-360	360;
	7455	12596	0.041223	0.115314	0	0	0	0	0	0	1	-360	360;
	7455	10412	0.022512	0.051521	0	0	0	0	0	0	1	-360	360;
	10393	7455	0.063537	0.189702	0	0	0	0	0	0	1	-360	360;
	6304	8124	0.01524	0.045496	0	0	0	0	0	0	1	-360	360;
	703	8124	0.035876	0.092612	0	0	0	0	0	0	1	-360	360;
	9118	7744	0.015471	0.046198	0	0	0	0	0	0	1	-360	360;
	12410	10385	0.043364	0.152818	0	0	0	0	0	0	1	-360	360;
	1799	10385	0.03095	0.100686	0	0	0	0	0	0	1	-360	360;
	2995	10385	0.035628	0.107959	0	0	0	0	0	0	1	-360	360;
	10634	9671	0.010901	0.034397	0	0	0	0	0	0	1	-360	360;
	5024	3991	0.026099	0.060099	0	0	0	0	0	0	1	-360	360;
	3944	5024	0.025942	0.057008	0	0	0	0	0	0	1	-360	360;
	13150	2309	0.001901	0.004636	0	0	0	0	0	0	1	-360	360;
	2309	12953	0.00176	0.003124	0	0	0	0	0	0	1	-360	360;
	10024	12183	0.003876	0.010909	0	0	0	0	0	0	1	-360	360;
	12183	1034	0.003802	0.011124	0	0	0	0	0	0	1	-360	360;
	6725	12183	0.013339	0.029298	0	0	0	0	0	0	1	-360	360;
	2632	12183	0.031719	0.11181	0	0	0	0	0	0	1	-360	360;
	4919	5454	0.021521	0.05557	0	0	0	0	0	0	1	-360	360;
	6205	4919	0.037008	0.107752	0	0	0	0	0	0	1	-360	360;
	9535	2800	0.05743	0.142463	0	0	0	0	0	0	1	-360	360;
	9535	2213	0.036091	0.074041	0	0	0	0	0	0	1	-360	360;
	11589	6146	0.001211	0.00545	0	0	0	0	0	0	1	-360	360;
	2986	6146	0.013711	0.078669	0	0	0	0	0	0	1	-360	360;
	1761	6669	0.050521	0.125331	0	0	0	0	0	0	1	-360	360;
	6669	7629	0.085818	0.221579	0	0	0	0	0	0	1	-360	360;
	6669	8180	0.014909	0.034132	0	0	0	0	0	0	1	-360	360;
	6669	6979	0.035826	0.113033	0	0	0	0	0	0	1	-360	360;
	9139	9276	0.008289	0.026017	0	0	0	0	0	0	1	-360	360;
	5859	9276	0.023339	0.089587	0	0	0	0	0	0	1	-360	360;
	5486	3411	0.00661	0.039521	0	0	0	0	0	0	1	-360	360;
	12516	5486	0.00349	0.01574	0	0	0	0	0	0	1	-360	360;
	1103	9632	0.03714	0.088496	0	0	0	0	0	0	1	-360	360;
	1103	12268	0.037256	0.087669	0	0	0	0	0	0	1	-360	360;
	1103	6722	0.03081	0.07957	0	0	0	0	0	0	1	-360	360;
	7358	7781	0.033248	0.096818	0	0	0	0	0	0	1	-360	360;
	13446	7781	0.011479	0.034298	0	0	0	0	0	0	1	-360	360;
	7093	1572	0	0.000893	0	0	0	0	0	0	1	-360	360;
	7093	2671	0.019273	0.042628	0	0	0	0	0	0	1	-360	360;
	7093	5487	0.019628	0.041702	0	0	0	0	0	0	1	-360	360;
	8673	7093	0.029512	0.062719	0	0	0	0	0	0	1	-360	360;
	9939	1572	0.034901	0.105744	0	0	0	0	0	0	1	-360	360;
	10346	1572	0.026289	0.071083	0	0	0	0	0	0	1	-360	360;
	5068	1572	0.041893	0.121942	0	0	0	0	0	0	1	-360	360;
	3051	12596	0.024851	0.071711	0	0	0	0	0	0	1	-360	360;
	10799	2415	0.006512	0.01681	0	0	0	0	0	0	1	-360	360;
	5699	10799	0.00305	0.009099	0	0	0	0	0	0	1	-360	360;
	2046	10571	0.014678	0.04105	0	0	0	0	0	0	1	-360	360;
	9208	2046	0.054149	0.119066	0	0	0	0	0	0	1	-360	360;
	12226	7040	0.08557	0.197909	0	0	0	0	0	0	1	-360	360;
	13089	12226	0.038314	0.075496	0	0	0	0	0	0	1	-360	360;
	9537	12174	0.036554	0.115289	0	0	0	0	0	0	1	-360	360;
	3537	9537	0.047636	0.11019	0	0	0	0	0	0	1	-360	360;
	3195	3970	0	0.000826	0	0	0	0	0	0	1	-360	360;
	12496	3970	0.013397	0.037488	0	0	0	0	0	0	1	-360	360;
	3195	1296	0.03095	0.079917	0	0	0	0	0	0	1	-360	360;
	8101	3195	0.014058	0.032198	0	0	0	0	0	0	1	-360	360;
	6968	13599	0.03086	0.093488	0	0	0	0	0	0	1	-360	360;
	9208	6968	0.023669	0.071066	0	0	0	0	0	0	1	-360	360;
	9208	6968	0.028083	0.078537	0	0	0	0	0	0	1	-360	360;
	1137	9575	0.016661	0.052554	0	0	0	0	0	0	1	-360	360;
	9575	11954	0.022223	0.050868	0	0	0	0	0	0	1	-360	360;
	499	1539	0.017	0.043909	0	0	0	0	0	0	1	-360	360;
	1539	11984	0.017727	0.05595	0	0	0	0	0	0	1	-360	360;
	903	2199	0.022223	0.052967	0	0	0	0	0	0	1	-360	360;
	5374	2199	0.027628	0.071347	0	0	0	0	0	0	1	-360	360;
	8833	6723	0	0.00081	0	0	0	0	0	0	1	-360	360;
	8833	13236	0.002983	0.010149	0	0	0	0	0	0	1	-360	360;
	3688	8833	0.008678	0.023463	0	0	0	0	0	0	1	-360	360;
	6723	1890	0.002917	0.010347	0	0	0	0	0	0	1	-360	360;
	3688	6723	0.008512	0.023917	0	0	0	0	0	0	1	-360	360;
	8325	5829	0.003603	0.011339	0	0	0	0	0	0	1	-360	360;
	8325	6972	0.006347	0.015132	0	0	0	0	0	0	1	-360	360;
	8329	8325	0.027777	0.08295	0	0	0	0	0	0	1	-360	360;
	703	5283	0.045438	0.105124	0	0	0	0	0	0	1	-360	360;
	5283	11357	0.07486	0.187603	0	0	0	0	0	0	1	-360	360;
	2548	7247	0.01919	0.053198	0	0	0	0	0	0	1	-360	360;
	12205	7247	0.008488	0.032421	0	0	0	0	0	0	1	-360	360;
	3476	10871	0.00365	0.03434	0	0	0	0	0	0	1	-360	360;
	2517	10871	0.0049	0.04432	0	0	0	0	0	0	1	-360	360;
	4510	4908	0.000901	0.005539	0	0	0	0	0	0	1	-360	360;
	3047	4510	0.00894	0.06106	0	0	0	0	0	0	1	-360	360;
	4510	2524	0.006919	0.049419	0	0	0	0	0	0	1	-360	360;
	4510	6474	0.006771	0.04576	0	0	0	0	0	0	1	-360	360;
	922	4852	0	0.000793	0	0	0	0	0	0	1	-360	360;
	922	12297	0.019628	0.061364	0	0	0	0	0	0	1	-360	360;
	922	9291	0.011727	0.033826	0	0	0	0	0	0	1	-360	360;
	4852	8684	0.006438	0.01943	0	0	0	0	0	0	1	-360	360;
	4852	8684	0.006562	0.019033	0	0	0	0	0	0	1	-360	360;
	6521	4852	0.006421	0.020893	0	0	0	0	0	0	1	-360	360;
	8366	6941	0.00031	0.00355	0	0	0	0	0	0	1	-360	360;
	6281	2989	0.016752	0.046397	0	0	0	0	0	0	1	-360	360;
	12104	6281	0.009347	0.035702	0	0	0	0	0	0	1	-360	360;
	2989	12301	0.011298	0.040033	0	0	0	0	0	0	1	-360	360;
	12952	1838	0.001298	0.004322	0	0	0	0	0	0	1	-360	360;
	11292	1838	0.000347	0.000876	0	0	0	0	0	0	1	-360	360;
	3476	1193	0.00028	0.00301	0	0	0	0	0	0	1	-360	360;
	12012	3476	0.00297	0.02796	0	0	0	0	0	0	1	-360	360;
	3476	883	0.00274	0.03157	0	0	0	0	0	0	1	-360	360;
	9139	11164	0.00286	0.010554	0	0	0	0	0	0	1	-360	360;
	9139	13615	0.025818	0.091446	0	0	0	0	0	0	1	-360	360;
	2833	9139	0.027066	0.065554	0	0	0	0	0	0	1	-360	360;
	9139	4081	0.086198	0.151504	0	0	0	0	0	0	1	-360	360;
	1095	9139	0.010893	0.02719	0	0	0	0	0	0	1	-360	360;
	8703	9139	0.03162	0.072818	0	0	0	0	0	0	1	-360	360;
	9888	11164	0.002802	0.010752	0	0	0	0	0	0	1	-360	360;
	9888	12886	0.054802	0.152636	0	0	0	0	0	0	1	-360	360;
	9888	10190	0.014223	0.048397	0	0	0	0	0	0	1	-360	360;
	10551	9888	0.01814	0.043521	0	0	0	0	0	0	1	-360	360;
	9888	5828	0.038248	0.07586	0	0	0	0	0	0	1	-360	360;
	5952	10002	0.017331	0.061314	0	0	0	0	0	0	1	-360	360;
	13015	10002	0.009182	0.032339	0	0	0	0	0	0	1	-360	360;
	1193	11513	0.00281	0.02539	0	0	0	0	0	0	1	-360	360;
	9834	11513	0.00158	0.01426	0	0	0	0	0	0	1	-360	360;
	9234	6415	0	0.000211	0	0	0	0	0	0	1	-360	360;
	13627	9234	0.00888	0.061021	0	0	0	0	0	0	1	-360	360;
	9234	2524	0.006961	0.04406	0	0	0	0	0	0	1	-360	360;
	6415	13315	0.00869	0.06218	0	0	0	0	0	0	1	-360	360;
	6415	10654	0.002401	0.027581	0	0	0	0	0	0	1	-360	360;
	6415	6474	0.00682	0.04493	0	0	0	0	0	0	1	-360	360;
	13293	2861	0.00238	0.009446	0	0	0	0	0	0	1	-360	360;
	10843	13293	0.017653	0.060008	0	0	0	0	0	0	1	-360	360;
	10770	4363	0.01838	0.051397	0	0	0	0	0	0	1	-360	360;
	4363	989	0.02995	0.071388	0	0	0	0	0	0	1	-360	360;
	5035	5734	0.023223	0.042223	0	0	0	0	0	0	1	-360	360;
	5734	172	0.018636	0.052124	0	0	0	0	0	0	1	-360	360;
	9834	10593	0.00096	0.01102	0	0	0	0	0	0	1	-360	360;
	10593	7326	0.00364	0.02923	0	0	0	0	0	0	1	-360	360;
	10593	6261	0.00248	0.02649	0	0	0	0	0	0	1	-360	360;
	10256	10593	0.00384	0.04229	0	0	0	0	0	0	1	-360	360;
	8281	3637	0	0.000219	0	0	0	0	0	0	1	-360	360;
	13627	3637	0.00344	0.02181	0	0	0	0	0	0	1	-360	360;
	3637	13315	0.00337	0.02224	0	0	0	0	0	0	1	-360	360;
	3637	293	0.010521	0.05913	0	0	0	0	0	0	1	-360	360;
	3637	802	0.00081	0.005341	0	0	0	0	0	0	1	-360	360;
	4361	3250	0.002579	0.00476	0	0	0	0	0	0	1	-360	360;
	8928	4361	0.004388	0.01595	0	0	0	0	0	0	1	-360	360;
	4361	10300	0.004752	0.017868	0	0	0	0	0	0	1	-360	360;
	1683	6849	0.00338	0.012959	0	0	0	0	0	0	1	-360	360;
	11615	1683	0.007008	0.02338	0	0	0	0	0	0	1	-360	360;
	8700	6520	0.025347	0.057942	0	0	0	0	0	0	1	-360	360;
	5650	8700	0.06314	0.162959	0	0	0	0	0	0	1	-360	360;
	5626	4383	0.007967	0.023802	0	0	0	0	0	0	1	-360	360;
	12981	5626	0.006636	0.016463	0	0	0	0	0	0	1	-360	360;
	7302	11091	0.012091	0.022893	0	0	0	0	0	0	1	-360	360;
	7302	5448	0.012322	0.022388	0	0	0	0	0	0	1	-360	360;
	8985	5584	0.016529	0.049347	0	0	0	0	0	0	1	-360	360;
	12349	12952	0.00119	0.007017	0	0	0	0	0	0	1	-360	360;
	12349	13622	0.002331	0.009636	0	0	0	0	0	0	1	-360	360;
	4365	2563	0.008091	0.02681	0	0	0	0	0	0	1	-360	360;
	2002	2563	0.001347	0.006876	0	0	0	0	0	0	1	-360	360;
	11292	2563	0.002281	0.009818	0	0	0	0	0	0	1	-360	360;
	591	11799	0.001149	0.00243	0	0	0	0	0	0	1	-360	360;
	5224	591	0.002909	0.006909	0	0	0	0	0	0	1	-360	360;
	989	3144	0.000413	0.000959	0	0	0	0	0	0	1	-360	360;
	989	3144	0.000421	0.000942	0	0	0	0	0	0	1	-360	360;
	10571	11113	0.024628	0.073504	0	0	0	0	0	0	1	-360	360;
	11113	9271	0.041496	0.094992	0	0	0	0	0	0	1	-360	360;
	4827	658	0.01	0.021273	0	0	0	0	0	0	1	-360	360;
	658	2632	0.030479	0.08795	0	0	0	0	0	0	1	-360	360;
	658	2632	0.031157	0.094992	0	0	0	0	0	0	1	-360	360;
	7704	9029	0.00962	0.023298	0	0	0	0	0	0	1	-360	360;
	2679	7704	0.018248	0.040397	0	0	0	0	0	0	1	-360	360;
	9921	8188	0.005868	0.01781	0	0	0	0	0	0	1	-360	360;
	9921	13335	0.01	0.02814	0	0	0	0	0	0	1	-360	360;
	9921	47	0.016562	0.052033	0	0	0	0	0	0	1	-360	360;
	9921	47	0.016893	0.050967	0	0	0	0	0	0	1	-360	360;
	8188	5425	0.013281	0.033149	0	0	0	0	0	0	1	-360	360;
	7424	1852	0.005463	0.020008	0	0	0	0	0	0	1	-360	360;
	7424	7656	0.02695	0.066868	0	0	0	0	0	0	1	-360	360;
	5425	7424	0.002752	0.008397	0	0	0	0	0	0	1	-360	360;
	8455	7424	0.002686	0.007777	0	0	0	0	0	0	1	-360	360;
	7424	47	0.019099	0.057876	0	0	0	0	0	0	1	-360	360;
	7424	5744	0.017339	0.043008	0	0	0	0	0	0	1	-360	360;
	1852	47	0.017843	0.068107	0	0	0	0	0	0	1	-360	360;
	5381	9302	0.022289	0.070314	0	0	0	0	0	0	1	-360	360;
	9302	3293	0.035008	0.101917	0	0	0	0	0	0	1	-360	360;
	12524	1403	0.045231	0.147174	0	0	0	0	0	0	1	-360	360;
	12649	1403	0.011413	0.03424	0	0	0	0	0	0	1	-360	360;
	10459	6553	0.001124	0.002579	0	0	0	0	0	0	1	-360	360;
	10459	8917	0.046909	0.135298	0	0	0	0	0	0	1	-360	360;
	10459	7295	0.035446	0.075124	0	0	0	0	0	0	1	-360	360;
	12324	10459	0.011107	0.025793	0	0	0	0	0	0	1	-360	360;
	10459	13422	0.005983	0.023752	0	0	0	0	0	0	1	-360	360;
	7873	3160	0.016132	0.04695	0	0	0	0	0	0	1	-360	360;
	4606	3160	0.008008	0.019083	0	0	0	0	0	0	1	-360	360;
	6318	682	0.034281	0.088488	0	0	0	0	0	0	1	-360	360;
	682	7385	0.009149	0.036372	0	0	0	0	0	0	1	-360	360;
	682	1121	0.008711	0.024124	0	0	0	0	0	0	1	-360	360;
	682	2193	0.03624	0.109818	0	0	0	0	0	0	1	-360	360;
	8917	6553	0.036777	0.116041	0	0	0	0	0	0	1	-360	360;
	11791	12588	0.003851	0.00995	0	0	0	0	0	0	1	-360	360;
	11791	12503	0.025793	0.056702	0	0	0	0	0	0	1	-360	360;
	5102	11791	0.027942	0.061421	0	0	0	0	0	0	1	-360	360;
	12588	13204	0.021488	0.047248	0	0	0	0	0	0	1	-360	360;
	6364	763	0.00015	0.00136	0	0	0	0	0	0	1	-360	360;
	10634	12371	0.016909	0.037174	0	0	0	0	0	0	1	-360	360;
	6449	10634	0.040496	0.088512	0	0	0	0	0	0	1	-360	360;
	7263	6925	0.035231	0.111132	0	0	0	0	0	0	1	-360	360;
	3125	7263	0.031017	0.086752	0	0	0	0	0	0	1	-360	360;
	5829	6972	0.003983	0.014017	0	0	0	0	0	0	1	-360	360;
	9807	11836	0.003521	0.010496	0	0	0	0	0	0	1	-360	360;
	12496	11836	0.002686	0.00805	0	0	0	0	0	0	1	-360	360;
	10049	9131	0.05195	0.145298	0	0	0	0	0	0	1	-360	360;
	13636	10937	0.023207	0.053132	0	0	0	0	0	0	1	-360	360;
	2996	13636	0.049587	0.150256	0	0	0	0	0	0	1	-360	360;
	3677	2094	0.036719	0.111298	0	0	0	0	0	0	1	-360	360;
	744	3677	0.008364	0.023207	0	0	0	0	0	0	1	-360	360;
	2002	8798	0.005256	0.024653	0	0	0	0	0	0	1	-360	360;
	5737	8798	0.001347	0.004397	0	0	0	0	0	0	1	-360	360;
	12952	10108	0.005149	0.022826	0	0	0	0	0	0	1	-360	360;
	8324	10108	0.001207	0.004992	0	0	0	0	0	0	1	-360	360;
	9941	7422	0.003934	0.009744	0	0	0	0	0	0	1	-360	360;
	8115	10571	0.01438	0.04186	0	0	0	0	0	0	1	-360	360;
	2018	2616	0.005744	0.014967	0	0	0	0	0	0	1	-360	360;
	9116	425	0.00176	0.01659	0	0	0	0	0	0	1	-360	360;
	7326	425	0.00151	0.01378	0	0	0	0	0	0	1	-360	360;
	7326	425	0.00128	0.01392	0	0	0	0	0	0	1	-360	360;
	425	5789	0.00052	0.00397	0	0	0	0	0	0	1	-360	360;
	6261	425	0.00221	0.02248	0	0	0	0	0	0	1	-360	360;
	10393	9936	0.007182	0.017107	0	0	0	0	0	0	1	-360	360;
	9936	6033	0.005149	0.011339	0	0	0	0	0	0	1	-360	360;
	12672	2706	0.00943	0.02157	0	0	0	0	0	0	1	-360	360;
	10498	6492	0.040636	0.11257	0	0	0	0	0	0	1	-360	360;
	13239	10498	0.014496	0.042231	0	0	0	0	0	0	1	-360	360;
	2833	11164	0.002744	0.00995	0	0	0	0	0	0	1	-360	360;
	11164	9631	0.003099	0.009752	0	0	0	0	0	0	1	-360	360;
	5424	13040	0.007322	0.016744	0	0	0	0	0	0	1	-360	360;
	12090	13040	0.03024	0.069231	0	0	0	0	0	0	1	-360	360;
	9838	1926	0.008438	0.018669	0	0	0	0	0	0	1	-360	360;
	1926	744	0.006727	0.027942	0	0	0	0	0	0	1	-360	360;
	1926	744	0.006587	0.028488	0	0	0	0	0	0	1	-360	360;
	6175	3612	0.014579	0.040463	0	0	0	0	0	0	1	-360	360;
	1800	3612	0.012826	0.036992	0	0	0	0	0	0	1	-360	360;
	3612	7222	0.006909	0.019917	0	0	0	0	0	0	1	-360	360;
	9112	11615	0.003421	0.011041	0	0	0	0	0	0	1	-360	360;
	9112	8928	0.005364	0.016289	0	0	0	0	0	0	1	-360	360;
	6520	6134	0.012463	0.030909	0	0	0	0	0	0	1	-360	360;
	6520	6134	0.012702	0.030273	0	0	0	0	0	0	1	-360	360;
	10289	6134	0.008331	0.02438	0	0	0	0	0	0	1	-360	360;
	12177	6134	0.008149	0.024843	0	0	0	0	0	0	1	-360	360;
	2385	2599	0.010789	0.071169	0	0	0	0	0	0	1	-360	360;
	5659	2599	0.011901	0.08499	0	0	0	0	0	0	1	-360	360;
	3051	6397	0.035942	0.113397	0	0	0	0	0	0	1	-360	360;
	10412	3051	0.035818	0.078752	0	0	0	0	0	0	1	-360	360;
	3051	6518	0.052587	0.147083	0	0	0	0	0	0	1	-360	360;
	12981	10684	0.007273	0.016661	0	0	0	0	0	0	1	-360	360;
	4571	10684	0.023488	0.072488	0	0	0	0	0	0	1	-360	360;
	13204	4527	0.004207	0.011777	0	0	0	0	0	0	1	-360	360;
	12503	4527	0.005339	0.020388	0	0	0	0	0	0	1	-360	360;
	11751	1846	0.007446	0.01638	0	0	0	0	0	0	1	-360	360;
	4409	1846	0.023033	0.057124	0	0	0	0	0	0	1	-360	360;
	268	3852	0.030198	0.095248	0	0	0	0	0	0	1	-360	360;
	9174	7484	0	0.000744	0	0	0	0	0	0	1	-360	360;
	4365	9174	0.003628	0.012	0	0	0	0	0	0	1	-360	360;
	9174	5737	0.00681	0.028273	0	0	0	0	0	0	1	-360	360;
	7484	8324	0.006669	0.028826	0	0	0	0	0	0	1	-360	360;
	4297	4885	0.04543	0.100521	0	0	0	0	0	0	1	-360	360;
	4297	4885	0.046273	0.098331	0	0	0	0	0	0	1	-360	360;
	3332	4885	0.005041	0.014554	0	0	0	0	0	0	1	-360	360;
	7094	7004	0.007818	0.021653	0	0	0	0	0	0	1	-360	360;
	7004	12667	0.041992	0.154033	0	0	0	0	0	0	1	-360	360;
	10403	9053	0.003711	0.009612	0	0	0	0	0	0	1	-360	360;
	6449	10403	0.009959	0.025719	0	0	0	0	0	0	1	-360	360;
	11041	9053	0.006744	0.014314	0	0	0	0	0	0	1	-360	360;
	1630	9814	0.023909	0.052901	0	0	0	0	0	0	1	-360	360;
	9814	9750	0.024347	0.051752	0	0	0	0	0	0	1	-360	360;
	5035	1336	0.04324	0.088702	0	0	0	0	0	0	1	-360	360;
	5035	785	0.045223	0.125488	0	0	0	0	0	0	1	-360	360;
	7656	1863	0.00305	0.007893	0	0	0	0	0	0	1	-360	360;
	6410	12380	0.01905	0.060107	0	0	0	0	0	0	1	-360	360;
	12380	9954	0.023471	0.055934	0	0	0	0	0	0	1	-360	360;
	11607	6540	0.004893	0.014339	0	0	0	0	0	0	1	-360	360;
	7961	9915	0.022934	0.050397	0	0	0	0	0	0	1	-360	360;
	85	10644	0.037248	0.081901	0	0	0	0	0	0	1	-360	360;
	1870	10644	0.013347	0.039083	0	0	0	0	0	0	1	-360	360;
	7301	10644	0.013058	0.039818	0	0	0	0	0	0	1	-360	360;
	10566	12988	0.04038	0.11757	0	0	0	0	0	0	1	-360	360;
	12988	11041	0.037397	0.111653	0	0	0	0	0	0	1	-360	360;
	58	6216	0.029273	0.069752	0	0	0	0	0	0	1	-360	360;
	58	3747	0.023678	0.070702	0	0	0	0	0	0	1	-360	360;
	1819	9632	0.001868	0.005463	0	0	0	0	0	0	1	-360	360;
	5142	6935	0.00048	0.00423	0	0	0	0	0	0	1	-360	360;
	8512	12982	0.003992	0.010289	0	0	0	0	0	0	1	-360	360;
	8512	4884	0.003777	0.009413	0	0	0	0	0	0	1	-360	360;
	9271	8512	0.015091	0.047174	0	0	0	0	0	0	1	-360	360;
	11651	5583	0.013826	0.039917	0	0	0	0	0	0	1	-360	360;
	11651	899	0.015678	0.039124	0	0	0	0	0	0	1	-360	360;
	11651	8917	0.066851	0.192818	0	0	0	0	0	0	1	-360	360;
	11651	13422	0.039653	0.128983	0	0	0	0	0	0	1	-360	360;
	12982	10233	0.005091	0.009702	0	0	0	0	0	0	1	-360	360;
	10233	1503	0.004066	0.010083	0	0	0	0	0	0	1	-360	360;
	1503	5583	0.007661	0.024058	0	0	0	0	0	0	1	-360	360;
	11512	5583	0.045636	0.13286	0	0	0	0	0	0	1	-360	360;
	4884	1503	0.018421	0.042868	0	0	0	0	0	0	1	-360	360;
	2659	8186	0.00805	0.04211	0	0	0	0	0	0	1	-360	360;
	8186	6424	0.003211	0.02194	0	0	0	0	0	0	1	-360	360;
	8186	6343	0.00974	0.05531	0	0	0	0	0	0	1	-360	360;
	13615	2562	0.019719	0.05495	0	0	0	0	0	0	1	-360	360;
	8115	3442	0.03495	0.086686	0	0	0	0	0	0	1	-360	360;
	9208	3442	0.041149	0.122851	0	0	0	0	0	0	1	-360	360;
	10611	5424	0.01124	0.027893	0	0	0	0	0	0	1	-360	360;
	825	2060	0.02219	0.057281	0	0	0	0	0	0	1	-360	360;
	2060	1214	0.011843	0.030529	0	0	0	0	0	0	1	-360	360;
	1336	172	0.014298	0.039983	0	0	0	0	0	0	1	-360	360;
	13591	6520	0.008083	0.028636	0	0	0	0	0	0	1	-360	360;
	10289	6520	0.012942	0.02962	0	0	0	0	0	0	1	-360	360;
	3798	8463	0.01657	0.050017	0	0	0	0	0	0	1	-360	360;
	2562	3798	0.006331	0.019107	0	0	0	0	0	0	1	-360	360;
	12886	5859	0.003983	0.013537	0	0	0	0	0	0	1	-360	360;
	2216	9116	0.00136	0.01515	0	0	0	0	0	0	1	-360	360;
	8455	13335	0.007876	0.01743	0	0	0	0	0	0	1	-360	360;
	10583	1890	0.018636	0.055653	0	0	0	0	0	0	1	-360	360;
	3852	10583	0.010554	0.031504	0	0	0	0	0	0	1	-360	360;
	8310	2479	0.013488	0.039686	0	0	0	0	0	0	1	-360	360;
	2479	7415	0.021934	0.048198	0	0	0	0	0	0	1	-360	360;
	7205	11872	0	0.00086	0	0	0	0	0	0	1	-360	360;
	7205	12538	0.026463	0.075669	0	0	0	0	0	0	1	-360	360;
	7205	1364	0.029446	0.081364	0	0	0	0	0	0	1	-360	360;
	7926	7205	0.055851	0.127826	0	0	0	0	0	0	1	-360	360;
	5102	11872	0.047818	0.124744	0	0	0	0	0	0	1	-360	360;
	7158	5837	0.02562	0.063256	0	0	0	0	0	0	1	-360	360;
	5837	4312	0.006372	0.012298	0	0	0	0	0	0	1	-360	360;
	9449	3937	0.009603	0.021099	0	0	0	0	0	0	1	-360	360;
	13204	10189	0.014917	0.052579	0	0	0	0	0	0	1	-360	360;
	6739	10189	0.03014	0.083512	0	0	0	0	0	0	1	-360	360;
	10691	8869	0.03319	0.092818	0	0	0	0	0	0	1	-360	360;
	8869	11437	0.024248	0.067826	0	0	0	0	0	0	1	-360	360;
	4388	12247	0.048529	0.153091	0	0	0	0	0	0	1	-360	360;
	5033	4388	0.055397	0.137421	0	0	0	0	0	0	1	-360	360;
	13638	5480	0	0.000219	0	0	0	0	0	0	1	-360	360;
	13638	7984	0.0013	0.008539	0	0	0	0	0	0	1	-360	360;
	6914	13638	0.007169	0.04463	0	0	0	0	0	0	1	-360	360;
	13638	10584	0.01787	0.097419	0	0	0	0	0	0	1	-360	360;
	5480	7984	0.00164	0.007409	0	0	0	0	0	0	1	-360	360;
	5480	12756	0.0097	0.06274	0	0	0	0	0	0	1	-360	360;
	6914	5480	0.007021	0.045469	0	0	0	0	0	0	1	-360	360;
	4070	11936	0.007496	0.022397	0	0	0	0	0	0	1	-360	360;
	6033	11936	0.007959	0.024124	0	0	0	0	0	0	1	-360	360;
	11199	4317	0.00405	0.011686	0	0	0	0	0	0	1	-360	360;
	11199	9454	0.005471	0.01581	0	0	0	0	0	0	1	-360	360;
	6953	10393	0.012727	0.031579	0	0	0	0	0	0	1	-360	360;
	6904	6953	0.009529	0.027496	0	0	0	0	0	0	1	-360	360;
	1293	2924	0	0.000876	0	0	0	0	0	0	1	-360	360;
	11439	2924	0.01243	0.040413	0	0	0	0	0	0	1	-360	360;
	6904	2924	0.002636	0.007967	0	0	0	0	0	0	1	-360	360;
	6904	2924	0.002686	0.007802	0	0	0	0	0	0	1	-360	360;
	1293	6652	0.049421	0.127587	0	0	0	0	0	0	1	-360	360;
	7986	1293	0.006587	0.018248	0	0	0	0	0	0	1	-360	360;
	1293	2800	0.064438	0.166364	0	0	0	0	0	0	1	-360	360;
	10812	6304	0.085322	0.155132	0	0	0	0	0	0	1	-360	360;
	534	7158	0.050488	0.130339	0	0	0	0	0	0	1	-360	360;
	13073	534	0.035471	0.111893	0	0	0	0	0	0	1	-360	360;
	10566	8057	0.013876	0.045132	0	0	0	0	0	0	1	-360	360;
	8057	6449	0.042744	0.16319	0	0	0	0	0	0	1	-360	360;
	7331	6838	0.027512	0.080083	0	0	0	0	0	0	1	-360	360;
	5918	7331	0.013132	0.038223	0	0	0	0	0	0	1	-360	360;
	12827	11000	0.017347	0.051802	0	0	0	0	0	0	1	-360	360;
	5033	11000	0.036446	0.086851	0	0	0	0	0	0	1	-360	360;
	5692	127	0	0.000893	0	0	0	0	0	0	1	-360	360;
	669	127	0.013223	0.056876	0	0	0	0	0	0	1	-360	360;
	127	8673	0.01662	0.035322	0	0	0	0	0	0	1	-360	360;
	9939	5692	0.023752	0.069157	0	0	0	0	0	0	1	-360	360;
	5692	7584	0.029851	0.100967	0	0	0	0	0	0	1	-360	360;
	5692	6094	0.014322	0.031504	0	0	0	0	0	0	1	-360	360;
	5692	11431	0.01481	0.040017	0	0	0	0	0	0	1	-360	360;
	6397	6518	0.038066	0.094413	0	0	0	0	0	0	1	-360	360;
	8215	12637	0.018702	0.058967	0	0	0	0	0	0	1	-360	360;
	11431	8215	0.008554	0.023917	0	0	0	0	0	0	1	-360	360;
	2511	2659	0.009849	0.05651	0	0	0	0	0	0	1	-360	360;
	6947	6739	0.013554	0.032066	0	0	0	0	0	0	1	-360	360;
	6947	3598	0.013719	0.042876	0	0	0	0	0	0	1	-360	360;
	10882	12635	0.032512	0.094636	0	0	0	0	0	0	1	-360	360;
	12635	6205	0.040008	0.116479	0	0	0	0	0	0	1	-360	360;
	11380	630	0.02319	0.055273	0	0	0	0	0	0	1	-360	360;
	1364	11380	0.013248	0.046413	0	0	0	0	0	0	1	-360	360;
	7926	11380	0.073289	0.23843	0	0	0	0	0	0	1	-360	360;
	2578	451	0.008083	0.029661	0	0	0	0	0	0	1	-360	360;
	10531	451	0.014661	0.058273	0	0	0	0	0	0	1	-360	360;
	7089	5231	0.017273	0.044587	0	0	0	0	0	0	1	-360	360;
	5231	4923	0.016413	0.049	0	0	0	0	0	0	1	-360	360;
	2443	11280	0.054587	0.120008	0	0	0	0	0	0	1	-360	360;
	12931	2443	0.049521	0.144149	0	0	0	0	0	0	1	-360	360;
	12776	3726	0.02757	0.063124	0	0	0	0	0	0	1	-360	360;
	10791	825	0.037686	0.089818	0	0	0	0	0	0	1	-360	360;
	8115	899	0.106033	0.321289	0	0	0	0	0	0	1	-360	360;
	13599	8115	0.066959	0.153273	0	0	0	0	0	0	1	-360	360;
	8115	9816	0.074421	0.208132	0	0	0	0	0	0	1	-360	360;
	8115	13623	0.054157	0.123967	0	0	0	0	0	0	1	-360	360;
	989	5553	0.086802	0.24276	0	0	0	0	0	0	1	-360	360;
	5553	3467	0.007628	0.018826	0	0	0	0	0	0	1	-360	360;
	3963	5381	0.021256	0.050669	0	0	0	0	0	0	1	-360	360;
	9577	159	0.015893	0.046223	0	0	0	0	0	0	1	-360	360;
	159	4447	0.01562	0.034331	0	0	0	0	0	0	1	-360	360;
	11449	8756	0.011174	0.023008	0	0	0	0	0	0	1	-360	360;
	703	11449	0.014438	0.036174	0	0	0	0	0	0	1	-360	360;
	8756	4447	0.012636	0.029752	0	0	0	0	0	0	1	-360	360;
	4317	11439	0	0.000826	0	0	0	0	0	0	1	-360	360;
	4317	4070	0.00086	0.003157	0	0	0	0	0	0	1	-360	360;
	4317	9454	0.01005	0.027843	0	0	0	0	0	0	1	-360	360;
	11439	1459	0.000843	0.003223	0	0	0	0	0	0	1	-360	360;
	9029	2833	0.022983	0.064256	0	0	0	0	0	0	1	-360	360;
	2833	8832	0.008603	0.018273	0	0	0	0	0	0	1	-360	360;
	2833	8832	0.007661	0.020702	0	0	0	0	0	0	1	-360	360;
	2833	9631	0.002678	0.007174	0	0	0	0	0	0	1	-360	360;
	2833	9631	0.002388	0.008132	0	0	0	0	0	0	1	-360	360;
	2524	1040	0.00237	0.011781	0	0	0	0	0	0	1	-360	360;
	6474	5186	0.001901	0.012831	0	0	0	0	0	0	1	-360	360;
	3149	5944	0	0.00081	0	0	0	0	0	0	1	-360	360;
	3149	8198	0.017273	0.025289	0	0	0	0	0	0	1	-360	360;
	5944	7332	0.030909	0.042397	0	0	0	0	0	0	1	-360	360;
	5944	11554	0.020967	0.066149	0	0	0	0	0	0	1	-360	360;
	7032	8304	0.049719	0.109298	0	0	0	0	0	0	1	-360	360;
	8304	6972	0.065909	0.144901	0	0	0	0	0	0	1	-360	360;
	1720	9834	0.00614	0.05775	0	0	0	0	0	0	1	-360	360;
	8310	12952	0.020678	0.079256	0	0	0	0	0	0	1	-360	360;
	6216	12952	0.020529	0.051702	0	0	0	0	0	0	1	-360	360;
	2002	13404	0.015066	0.050711	0	0	0	0	0	0	1	-360	360;
	373	12960	0.032364	0.074058	0	0	0	0	0	0	1	-360	360;
	3634	373	0.024893	0.069612	0	0	0	0	0	0	1	-360	360;
	5650	11280	0.051289	0.112769	0	0	0	0	0	0	1	-360	360;
	11508	13204	0.024124	0.069562	0	0	0	0	0	0	1	-360	360;
	13204	1799	0.026769	0.10219	0	0	0	0	0	0	1	-360	360;
	11658	13204	0.025909	0.077347	0	0	0	0	0	0	1	-360	360;
	13204	1118	0.01324	0.046628	0	0	0	0	0	0	1	-360	360;
	12410	11450	0.01676	0.03686	0	0	0	0	0	0	1	-360	360;
	3598	12410	0.017612	0.043678	0	0	0	0	0	0	1	-360	360;
	11751	6694	0.02043	0.057124	0	0	0	0	0	0	1	-360	360;
	9238	10256	0.00084	0.00669	0	0	0	0	0	0	1	-360	360;
	2636	6318	0.019893	0.062752	0	0	0	0	0	0	1	-360	360;
	9373	117	0.058719	0.164223	0	0	0	0	0	0	1	-360	360;
	9373	3634	0.021661	0.049587	0	0	0	0	0	0	1	-360	360;
	876	2874	0.008	0.030686	0	0	0	0	0	0	1	-360	360;
	5454	6655	0.024017	0.057248	0	0	0	0	0	0	1	-360	360;
	6849	1038	0.005636	0.020727	0	0	0	0	0	0	1	-360	360;
	7533	10300	0.004769	0.016826	0	0	0	0	0	0	1	-360	360;
	11994	7533	0.00895	0.031554	0	0	0	0	0	0	1	-360	360;
	11679	5579	0.010919	0.07275	0	0	0	0	0	0	1	-360	360;
	5579	12015	0.004841	0.02861	0	0	0	0	0	0	1	-360	360;
	13155	4260	0.044909	0.092107	0	0	0	0	0	0	1	-360	360;
	4571	4383	0.011901	0.039537	0	0	0	0	0	0	1	-360	360;
	596	13378	0.008421	0.020066	0	0	0	0	0	0	1	-360	360;
	13378	2142	0.07076	0.184612	0	0	0	0	0	0	1	-360	360;
	13378	7300	0.00624	0.018917	0	0	0	0	0	0	1	-360	360;
	13378	2865	0.055157	0.127587	0	0	0	0	0	0	1	-360	360;
	12637	1099	0.03562	0.084876	0	0	0	0	0	0	1	-360	360;
	1459	12463	0.024248	0.073463	0	0	0	0	0	0	1	-360	360;
	2616	9557	0.006372	0.012686	0	0	0	0	0	0	1	-360	360;
	5996	674	0.003397	0.009802	0	0	0	0	0	0	1	-360	360;
	5996	7222	0.002793	0.008397	0	0	0	0	0	0	1	-360	360;
	953	10900	0.035636	0.100248	0	0	0	0	0	0	1	-360	360;
	4708	10900	0.034901	0.102182	0	0	0	0	0	0	1	-360	360;
	1706	8231	0.032388	0.094281	0	0	0	0	0	0	1	-360	360;
	7094	8231	0.031421	0.093802	0	0	0	0	0	0	1	-360	360;
	11679	7281	0.02455	0.175521	0	0	0	0	0	0	1	-360	360;
	11491	7281	0.011281	0.058469	0	0	0	0	0	0	1	-360	360;
	7281	595	0.015979	0.102281	0	0	0	0	0	0	1	-360	360;
	10770	9459	0	0.000777	0	0	0	0	0	0	1	-360	360;
	10770	12382	0.010868	0.037488	0	0	0	0	0	0	1	-360	360;
	10770	11553	0.011967	0.024562	0	0	0	0	0	0	1	-360	360;
	9459	12382	0.010636	0.034711	0	0	0	0	0	0	1	-360	360;
	9459	13116	0.055091	0.164496	0	0	0	0	0	0	1	-360	360;
	9459	11557	0.073091	0.221488	0	0	0	0	0	0	1	-360	360;
	12174	5267	0.01938	0.056421	0	0	0	0	0	0	1	-360	360;
	4524	5267	0.018983	0.057496	0	0	0	0	0	0	1	-360	360;
	7032	10289	0.031413	0.07795	0	0	0	0	0	0	1	-360	360;
	7032	10705	0.046231	0.119372	0	0	0	0	0	0	1	-360	360;
	7744	4188	0.012066	0.029901	0	0	0	0	0	0	1	-360	360;
	1492	499	0.022529	0.071066	0	0	0	0	0	0	1	-360	360;
	6655	1492	0.027471	0.065471	0	0	0	0	0	0	1	-360	360;
	1038	6319	0	0.00076	0	0	0	0	0	0	1	-360	360;
	11994	6319	0.005868	0.020678	0	0	0	0	0	0	1	-360	360;
	2193	903	0.022636	0.049769	0	0	0	0	0	0	1	-360	360;
	9784	7332	0.033909	0.047901	0	0	0	0	0	0	1	-360	360;
	9784	7332	0.034537	0.04686	0	0	0	0	0	0	1	-360	360;
	7332	8198	0.012959	0.028661	0	0	0	0	0	0	1	-360	360;
	4396	11589	0.003919	0.02719	0	0	0	0	0	0	1	-360	360;
	7659	4560	0	0.000744	0	0	0	0	0	0	1	-360	360;
	5768	7659	0.013909	0.038909	0	0	0	0	0	0	1	-360	360;
	1697	4560	0.007579	0.023	0	0	0	0	0	0	1	-360	360;
	2334	6652	0.032587	0.077653	0	0	0	0	0	0	1	-360	360;
	6652	2213	0.035744	0.085851	0	0	0	0	0	0	1	-360	360;
	1705	5719	0.011969	0.0687	0	0	0	0	0	0	1	-360	360;
	5719	1383	0.00431	0.024719	0	0	0	0	0	0	1	-360	360;
	498	7158	0.107463	0.250107	0	0	0	0	0	0	1	-360	360;
	7158	1222	0.043099	0.099678	0	0	0	0	0	0	1	-360	360;
	7158	4923	0.033058	0.075669	0	0	0	0	0	0	1	-360	360;
	12540	7158	0.045273	0.099537	0	0	0	0	0	0	1	-360	360;
	7158	3539	0.071182	0.141157	0	0	0	0	0	0	1	-360	360;
	7158	3539	0.072554	0.138223	0	0	0	0	0	0	1	-360	360;
	1503	899	0.00781	0.02357	0	0	0	0	0	0	1	-360	360;
	899	9539	0.014298	0.042702	0	0	0	0	0	0	1	-360	360;
	3992	11846	0.016793	0.041661	0	0	0	0	0	0	1	-360	360;
	11846	11091	0.011017	0.021711	0	0	0	0	0	0	1	-360	360;
	2986	11679	0.01255	0.08125	0	0	0	0	0	0	1	-360	360;
	13328	11540	0.008041	0.022488	0	0	0	0	0	0	1	-360	360;
	13328	11540	0.009149	0.017438	0	0	0	0	0	0	1	-360	360;
	13328	9721	0.032876	0.06481	0	0	0	0	0	0	1	-360	360;
	11877	13328	0.022512	0.056769	0	0	0	0	0	0	1	-360	360;
	11540	2442	0.01276	0.035702	0	0	0	0	0	0	1	-360	360;
	2527	8868	0.007793	0.02457	0	0	0	0	0	0	1	-360	360;
	5597	8868	0.00762	0.022752	0	0	0	0	0	0	1	-360	360;
	3293	9630	0.010157	0.025198	0	0	0	0	0	0	1	-360	360;
	3293	7390	0.018289	0.054603	0	0	0	0	0	0	1	-360	360;
	3293	8180	0.031876	0.082322	0	0	0	0	0	0	1	-360	360;
	13606	2527	0.011289	0.033917	0	0	0	0	0	0	1	-360	360;
	7926	11508	0.055537	0.203711	0	0	0	0	0	0	1	-360	360;
	9770	12953	0.001488	0.003314	0	0	0	0	0	0	1	-360	360;
	11832	9770	0.009083	0.012124	0	0	0	0	0	0	1	-360	360;
	5431	4641	0	0.000876	0	0	0	0	0	0	1	-360	360;
	9128	5142	0.0011	0.01267	0	0	0	0	0	0	1	-360	360;
	5142	13221	0.00252	0.02751	0	0	0	0	0	0	1	-360	360;
	6838	11554	0.04314	0.128802	0	0	0	0	0	0	1	-360	360;
	5768	1697	0.011562	0.029843	0	0	0	0	0	0	1	-360	360;
	11433	12242	0.011256	0.041174	0	0	0	0	0	0	1	-360	360;
	11433	12242	0.011017	0.04195	0	0	0	0	0	0	1	-360	360;
	10937	3594	0.029496	0.076149	0	0	0	0	0	0	1	-360	360;
	12015	1705	0.009401	0.05757	0	0	0	0	0	0	1	-360	360;
	7094	1706	0.056934	0.185223	0	0	0	0	0	0	1	-360	360;
	12667	1706	0.005479	0.016479	0	0	0	0	0	0	1	-360	360;
	12774	9876	0.011289	0.029157	0	0	0	0	0	0	1	-360	360;
	12774	848	0.000504	0.001446	0	0	0	0	0	0	1	-360	360;
	1250	12774	0.014347	0.035802	0	0	0	0	0	0	1	-360	360;
	1250	12774	0.014628	0.035066	0	0	0	0	0	0	1	-360	360;
	12774	8985	0.019537	0.05462	0	0	0	0	0	0	1	-360	360;
	9454	7986	0.003281	0.009488	0	0	0	0	0	0	1	-360	360;
	9271	3342	0.033339	0.086091	0	0	0	0	0	0	1	-360	360;
	13623	3342	0.032628	0.095	0	0	0	0	0	0	1	-360	360;
	8463	10190	0.008281	0.029331	0	0	0	0	0	0	1	-360	360;
	293	802	0.008711	0.059841	0	0	0	0	0	0	1	-360	360;
	293	10584	0.0127	0.06389	0	0	0	0	0	0	1	-360	360;
	6175	8417	0	0.000826	0	0	0	0	0	0	1	-360	360;
	6175	1800	0.016992	0.051281	0	0	0	0	0	0	1	-360	360;
	6175	674	0.017322	0.050223	0	0	0	0	0	0	1	-360	360;
	6175	5175	0.016901	0.06357	0	0	0	0	0	0	1	-360	360;
	6175	4909	0.037231	0.112777	0	0	0	0	0	0	1	-360	360;
	7415	6175	0.039636	0.115388	0	0	0	0	0	0	1	-360	360;
	3747	6175	0.067339	0.212438	0	0	0	0	0	0	1	-360	360;
	1214	9876	0.061322	0.134818	0	0	0	0	0	0	1	-360	360;
	3400	674	0.04114	0.119752	0	0	0	0	0	0	1	-360	360;
	2010	2364	0.03657	0.109198	0	0	0	0	0	0	1	-360	360;
	1296	2010	0.042488	0.097248	0	0	0	0	0	0	1	-360	360;
	2010	6945	0.039099	0.089512	0	0	0	0	0	0	1	-360	360;
	2727	2450	0.013537	0.032471	0	0	0	0	0	0	1	-360	360;
	7326	6941	0.00158	0.01747	0	0	0	0	0	0	1	-360	360;
	2884	7984	0.00894	0.04824	0	0	0	0	0	0	1	-360	360;
	6914	2884	0.003531	0.018341	0	0	0	0	0	0	1	-360	360;
	12301	10882	0.029818	0.068256	0	0	0	0	0	0	1	-360	360;
	9161	8337	0	0.00081	0	0	0	0	0	0	1	-360	360;
	9161	11292	0.003512	0.004661	0	0	0	0	0	0	1	-360	360;
	6518	4827	0.010959	0.027322	0	0	0	0	0	0	1	-360	360;
	13622	8337	0.003579	0.004562	0	0	0	0	0	0	1	-360	360;
	5374	2636	0.045843	0.136851	0	0	0	0	0	0	1	-360	360;
	5374	5971	0.044678	0.135421	0	0	0	0	0	0	1	-360	360;
	5374	2499	0.015793	0.04786	0	0	0	0	0	0	1	-360	360;
	10296	6205	0.036248	0.10138	0	0	0	0	0	0	1	-360	360;
	2204	12012	0.00336	0.02922	0	0	0	0	0	0	1	-360	360;
	8051	11491	0.01663	0.097229	0	0	0	0	0	0	1	-360	360;
	13236	1890	0	0.000777	0	0	0	0	0	0	1	-360	360;
	1890	9721	0.09357	0.184397	0	0	0	0	0	0	1	-360	360;
	7924	1890	0.090248	0.177719	0	0	0	0	0	0	1	-360	360;
	7924	1890	0.091983	0.174017	0	0	0	0	0	0	1	-360	360;
	1890	1773	0.018256	0.05314	0	0	0	0	0	0	1	-360	360;
	1890	11877	0.112273	0.205446	0	0	0	0	0	0	1	-360	360;
	3411	5020	0.003401	0.02362	0	0	0	0	0	0	1	-360	360;
	3411	10999	0.002719	0.012781	0	0	0	0	0	0	1	-360	360;
	3411	3092	0.00205	0.022719	0	0	0	0	0	0	1	-360	360;
	11779	11354	0.036579	0.090719	0	0	0	0	0	0	1	-360	360;
	11779	11075	0.038802	0.092446	0	0	0	0	0	0	1	-360	360;
	2744	7358	0.029397	0.067298	0	0	0	0	0	0	1	-360	360;
	2744	13446	0.054289	0.124289	0	0	0	0	0	0	1	-360	360;
	7358	11607	0.007421	0.023438	0	0	0	0	0	0	1	-360	360;
	7358	7669	0.006124	0.018554	0	0	0	0	0	0	1	-360	360;
	12104	12301	0.010446	0.034017	0	0	0	0	0	0	1	-360	360;
	1137	4064	0.045702	0.127802	0	0	0	0	0	0	1	-360	360;
	7584	10293	0.057438	0.171496	0	0	0	0	0	0	1	-360	360;
	2126	10293	0.025182	0.057636	0	0	0	0	0	0	1	-360	360;
	6817	4081	0.03224	0.075033	0	0	0	0	0	0	1	-360	360;
	498	1222	0.075653	0.13295	0	0	0	0	0	0	1	-360	360;
	989	11433	0.005058	0.017017	0	0	0	0	0	0	1	-360	360;
	989	11433	0.005157	0.016669	0	0	0	0	0	0	1	-360	360;
	6972	5980	0.046777	0.101124	0	0	0	0	0	0	1	-360	360;
	9270	13564	0.045364	0.135446	0	0	0	0	0	0	1	-360	360;
	10289	9630	0.037471	0.113529	0	0	0	0	0	0	1	-360	360;
	6657	13199	0.058372	0.168446	0	0	0	0	0	0	1	-360	360;
	6657	13199	0.066157	0.165083	0	0	0	0	0	0	1	-360	360;
	4524	13199	0.046628	0.120388	0	0	0	0	0	0	1	-360	360;
	8042	6725	0.029868	0.087488	0	0	0	0	0	0	1	-360	360;
	6979	3963	0.015562	0.047107	0	0	0	0	0	0	1	-360	360;
	11450	2995	0.045967	0.109562	0	0	0	0	0	0	1	-360	360;
	989	11799	0.00705	0.016876	0	0	0	0	0	0	1	-360	360;
	989	5224	0.004	0.009537	0	0	0	0	0	0	1	-360	360;
	989	3564	0.06243	0.161207	0	0	0	0	0	0	1	-360	360;
	1660	11553	0.005942	0.014149	0	0	0	0	0	0	1	-360	360;
	13139	9488	0.004651	0.024159	0	0	0	0	0	0	1	-360	360;
	13139	11682	0.00474	0.02363	0	0	0	0	0	0	1	-360	360;
	11041	12371	0.036669	0.115669	0	0	0	0	0	0	1	-360	360;
	1496	8416	0.043273	0.095132	0	0	0	0	0	0	1	-360	360;
	1496	8257	0.004521	0.011661	0	0	0	0	0	0	1	-360	360;
	883	9128	0.00207	0.01871	0	0	0	0	0	0	1	-360	360;
	1383	2511	0.00419	0.02712	0	0	0	0	0	0	1	-360	360;
	1383	2511	0.002349	0.022039	0	0	0	0	0	0	1	-360	360;
	8917	7295	0.02	0.04314	0	0	0	0	0	0	1	-360	360;
	2231	13404	0.00676	0.016124	0	0	0	0	0	0	1	-360	360;
	10868	13649	0.007876	0.022529	0	0	0	0	0	0	1	-360	360;
	13649	5597	0.025793	0.060298	0	0	0	0	0	0	1	-360	360;
	3435	10296	0.067587	0.204793	0	0	0	0	0	0	1	-360	360;
	8832	10691	0.053008	0.116554	0	0	0	0	0	0	1	-360	360;
	9792	9721	0.01195	0.030868	0	0	0	0	0	0	1	-360	360;
	8257	9792	0.015463	0.04876	0	0	0	0	0	0	1	-360	360;
	744	4409	0.018702	0.071413	0	0	0	0	0	0	1	-360	360;
	256	6941	0.00142	0.01555	0	0	0	0	0	0	1	-360	360;
	256	2204	0.00123	0.01312	0	0	0	0	0	0	1	-360	360;
	256	2204	0.00121	0.01337	0	0	0	0	0	0	1	-360	360;
	2216	256	0.00079	0.01226	0	0	0	0	0	0	1	-360	360;
	4169	256	0.00062	0.00673	0	0	0	0	0	0	1	-360	360;
	8931	7984	0	0.000211	0	0	0	0	0	0	1	-360	360;
	6914	7984	0.00763	0.05451	0	0	0	0	0	0	1	-360	360;
	10393	12463	0	0.00076	0	0	0	0	0	0	1	-360	360;
	13591	8503	0.00657	0.019603	0	0	0	0	0	0	1	-360	360;
	9807	9304	0.011512	0.028562	0	0	0	0	0	0	1	-360	360;
	11658	12538	0.014612	0.05362	0	0	0	0	0	0	1	-360	360;
	13557	596	0.00976	0.02419	0	0	0	0	0	0	1	-360	360;
	2193	4514	0.007876	0.022934	0	0	0	0	0	0	1	-360	360;
	4514	5971	0.01795	0.042769	0	0	0	0	0	0	1	-360	360;
	2578	11105	0.023488	0.046364	0	0	0	0	0	0	1	-360	360;
	2578	2450	0.013802	0.031802	0	0	0	0	0	0	1	-360	360;
	4362	10897	0.036678	0.080636	0	0	0	0	0	0	1	-360	360;
	10897	6387	0.005917	0.017066	0	0	0	0	0	0	1	-360	360;
	2231	2862	0.003364	0.012901	0	0	0	0	0	0	1	-360	360;
	3400	2231	0.038107	0.120198	0	0	0	0	0	0	1	-360	360;
	8310	2862	0.003289	0.011942	0	0	0	0	0	0	1	-360	360;
	1296	6263	0.047818	0.109479	0	0	0	0	0	0	1	-360	360;
	6263	6594	0.023603	0.060744	0	0	0	0	0	0	1	-360	360;
	6263	7306	0.042851	0.125488	0	0	0	0	0	0	1	-360	360;
	6263	7306	0.04376	0.123124	0	0	0	0	0	0	1	-360	360;
	6941	2204	0.00089	0.0071	0	0	0	0	0	0	1	-360	360;
	12301	3435	0	0.000744	0	0	0	0	0	0	1	-360	360;
	12301	11984	0.039554	0.086942	0	0	0	0	0	0	1	-360	360;
	744	4913	0.046364	0.078843	0	0	0	0	0	0	1	-360	360;
	1901	5417	0.006248	0.007702	0	0	0	0	0	0	1	-360	360;
	13150	1901	0.001273	0.002661	0	0	0	0	0	0	1	-360	360;
	12960	6410	0.027711	0.061314	0	0	0	0	0	0	1	-360	360;
	12960	6410	0.028231	0.059967	0	0	0	0	0	0	1	-360	360;
	1502	8416	0.029702	0.056231	0	0	0	0	0	0	1	-360	360;
	4312	9402	0.006256	0.011397	0	0	0	0	0	0	1	-360	360;
	6722	7019	0.010298	0.032512	0	0	0	0	0	0	1	-360	360;
	7019	13527	0.01805	0.039686	0	0	0	0	0	0	1	-360	360;
	12860	10413	0.012849	0.07226	0	0	0	0	0	0	1	-360	360;
	12860	12669	0.013099	0.07075	0	0	0	0	0	0	1	-360	360;
	2548	12205	0.008314	0.033033	0	0	0	0	0	0	1	-360	360;
	9939	13006	0.031843	0.096463	0	0	0	0	0	0	1	-360	360;
	13019	12268	0.002207	0.005273	0	0	0	0	0	0	1	-360	360;
	9517	13019	0.009124	0.027669	0	0	0	0	0	0	1	-360	360;
	2364	8042	0.010777	0.032653	0	0	0	0	0	0	1	-360	360;
	1630	11727	0.00757	0.018901	0	0	0	0	0	0	1	-360	360;
	4272	9750	0.007719	0.018512	0	0	0	0	0	0	1	-360	360;
	5828	6817	0.016198	0.040818	0	0	0	0	0	0	1	-360	360;
	1066	4125	0.054512	0.162752	0	0	0	0	0	0	1	-360	360;
	4125	3944	0.030802	0.067719	0	0	0	0	0	0	1	-360	360;
	6694	10421	0.012298	0.029281	0	0	0	0	0	0	1	-360	360;
	420	12206	0.074322	0.221901	0	0	0	0	0	0	1	-360	360;
	1316	12206	0.009669	0.023033	0	0	0	0	0	0	1	-360	360;
	1250	7924	0	0.000843	0	0	0	0	0	0	1	-360	360;
	57	11404	0.073463	0.189678	0	0	0	0	0	0	1	-360	360;
	11404	11554	0.058653	0.145488	0	0	0	0	0	0	1	-360	360;
	2216	5789	0.00102	0.01773	0	0	0	0	0	0	1	-360	360;
	7584	6241	0.00838	0.02438	0	0	0	0	0	0	1	-360	360;
	7584	10570	0.01219	0.03024	0	0	0	0	0	0	1	-360	360;
	7584	2126	0.048744	0.11614	0	0	0	0	0	0	1	-360	360;
	10570	2671	0.012967	0.030934	0	0	0	0	0	0	1	-360	360;
	6241	5487	0.006983	0.021149	0	0	0	0	0	0	1	-360	360;
	669	6241	0.017347	0.058702	0	0	0	0	0	0	1	-360	360;
	3539	10577	0.02457	0.058562	0	0	0	0	0	0	1	-360	360;
	9545	7912	0.016083	0.044983	0	0	0	0	0	0	1	-360	360;
	7127	7912	0.004843	0.011512	0	0	0	0	0	0	1	-360	360;
	11512	12324	0.005058	0.011339	0	0	0	0	0	0	1	-360	360;
	10231	9000	0.023727	0.074843	0	0	0	0	0	0	1	-360	360;
	13006	9000	0.031157	0.098281	0	0	0	0	0	0	1	-360	360;
	1118	6739	0.014372	0.048661	0	0	0	0	0	0	1	-360	360;
	1371	883	0.00331	0.02754	0	0	0	0	0	0	1	-360	360;
	10999	10664	0.002669	0.01732	0	0	0	0	0	0	1	-360	360;
	10664	398	0.019719	0.103289	0	0	0	0	0	0	1	-360	360;
	10999	5020	0.00261	0.016039	0	0	0	0	0	0	1	-360	360;
	703	9577	0.008587	0.017446	0	0	0	0	0	0	1	-360	360;
	3992	12524	0	0.000876	0	0	0	0	0	0	1	-360	360;
	12297	3992	0.017413	0.05443	0	0	0	0	0	0	1	-360	360;
	3992	9291	0.024744	0.074289	0	0	0	0	0	0	1	-360	360;
	12524	12649	0.036421	0.11381	0	0	0	0	0	0	1	-360	360;
	7385	1121	0.004198	0.012587	0	0	0	0	0	0	1	-360	360;
	11954	12243	0.014686	0.044521	0	0	0	0	0	0	1	-360	360;
	3747	12243	0.015231	0.041256	0	0	0	0	0	0	1	-360	360;
	85	9915	0.01438	0.045364	0	0	0	0	0	0	1	-360	360;
	10771	9254	0.025562	0.076298	0	0	0	0	0	0	1	-360	360;
	9254	12931	0.028083	0.061744	0	0	0	0	0	0	1	-360	360;
	47	7495	0.05019	0.152107	0	0	0	0	0	0	1	-360	360;
	1020	7495	0.027909	0.072033	0	0	0	0	0	0	1	-360	360;
	2800	6830	0.020512	0.061248	0	0	0	0	0	0	1	-360	360;
	1214	8814	0.025364	0.053893	0	0	0	0	0	0	1	-360	360;
	12247	10566	0.054479	0.165099	0	0	0	0	0	0	1	-360	360;
	12247	10566	0.055636	0.161983	0	0	0	0	0	0	1	-360	360;
	12247	12083	0.051579	0.154	0	0	0	0	0	0	1	-360	360;
	11554	785	0.035752	0.106752	0	0	0	0	0	0	1	-360	360;
	172	9784	0.021124	0.054537	0	0	0	0	0	0	1	-360	360;
	172	9784	0.021537	0.05343	0	0	0	0	0	0	1	-360	360;
	9741	802	0.0082	0.04119	0	0	0	0	0	0	1	-360	360;
	6410	9954	0.040281	0.122058	0	0	0	0	0	0	1	-360	360;
	8042	6594	0.031934	0.10086	0	0	0	0	0	0	1	-360	360;
	8042	2334	0.048347	0.145008	0	0	0	0	0	0	1	-360	360;
	8042	6945	0.056636	0.164901	0	0	0	0	0	0	1	-360	360;
	6657	10595	0.048091	0.119281	0	0	0	0	0	0	1	-360	360;
	6657	7188	0.00305	0.008793	0	0	0	0	0	0	1	-360	360;
	6657	7188	0.003463	0.00862	0	0	0	0	0	0	1	-360	360;
	3467	6657	0.004223	0.009339	0	0	0	0	0	0	1	-360	360;
	3467	6657	0.005281	0.009603	0	0	0	0	0	0	1	-360	360;
	9270	6579	0.028033	0.066033	0	0	0	0	0	0	1	-360	360;
	7138	13606	0.012983	0.051587	0	0	0	0	0	0	1	-360	360;
	9813	6453	0.00183	0.01416	0	0	0	0	0	0	1	-360	360;
	6094	10318	0.012504	0.036397	0	0	0	0	0	0	1	-360	360;
	13006	10318	0.024488	0.074198	0	0	0	0	0	0	1	-360	360;
	5699	2861	0.010579	0.045496	0	0	0	0	0	0	1	-360	360;
	10595	13116	0.025372	0.065512	0	0	0	0	0	0	1	-360	360;
	12792	9644	0.015058	0.042124	0	0	0	0	0	0	1	-360	360;
	5235	12792	0.038248	0.084099	0	0	0	0	0	0	1	-360	360;
	12081	1066	0.045719	0.113397	0	0	0	0	0	0	1	-360	360;
	12081	5744	0.046603	0.111041	0	0	0	0	0	0	1	-360	360;
	1099	12201	0.013843	0.041934	0	0	0	0	0	0	1	-360	360;
	13073	12540	0.052083	0.134446	0	0	0	0	0	0	1	-360	360;
	7040	420	0.013579	0.03981	0	0	0	0	0	0	1	-360	360;
	7040	420	0.013868	0.039058	0	0	0	0	0	0	1	-360	360;
	7040	12851	0.035099	0.077182	0	0	0	0	0	0	1	-360	360;
	13089	7040	0.11376	0.27386	0	0	0	0	0	0	1	-360	360;
	420	11871	0.069628	0.153091	0	0	0	0	0	0	1	-360	360;
	13221	6453	0.00296	0.03263	0	0	0	0	0	0	1	-360	360;
	5597	7873	0.048322	0.115149	0	0	0	0	0	0	1	-360	360;
	1819	7436	0.009802	0.028678	0	0	0	0	0	0	1	-360	360;
	11832	5417	0.003132	0.004347	0	0	0	0	0	0	1	-360	360;
	4894	6889	0.00284	0.02937	0	0	0	0	0	0	1	-360	360;
	6889	10256	0.00353	0.03885	0	0	0	0	0	0	1	-360	360;
	1720	6889	0.00193	0.02044	0	0	0	0	0	0	1	-360	360;
	5659	2385	0.00918	0.065591	0	0	0	0	0	0	1	-360	360;
	9703	5659	0	0.000211	0	0	0	0	0	0	1	-360	360;
	5659	13139	0.01776	0.09207	0	0	0	0	0	0	1	-360	360;
	5659	13139	0.018081	0.09007	0	0	0	0	0	0	1	-360	360;
	5659	5758	0.0017	0.0117	0	0	0	0	0	0	1	-360	360;
	5659	1619	0.001669	0.011919	0	0	0	0	0	0	1	-360	360;
	6122	13557	0.009041	0.023331	0	0	0	0	0	0	1	-360	360;
	3534	4898	0.075851	0.15557	0	0	0	0	0	0	1	-360	360;
	2893	3534	0.012471	0.036587	0	0	0	0	0	0	1	-360	360;
	7138	10868	0.003702	0.011537	0	0	0	0	0	0	1	-360	360;
	8673	10346	0.003661	0.009066	0	0	0	0	0	0	1	-360	360;
	8673	11431	0	0.000893	0	0	0	0	0	0	1	-360	360;
	7575	12516	0.00537	0.032901	0	0	0	0	0	0	1	-360	360;
	13461	9517	0.016661	0.041331	0	0	0	0	0	0	1	-360	360;
	13461	10779	0.002628	0.007587	0	0	0	0	0	0	1	-360	360;
	13461	13527	0.009446	0.025388	0	0	0	0	0	0	1	-360	360;
	11597	11354	0.023248	0.060017	0	0	0	0	0	0	1	-360	360;
	1863	1066	0.053711	0.127983	0	0	0	0	0	0	1	-360	360;
	12201	5068	0.009851	0.022537	0	0	0	0	0	0	1	-360	360;
	3975	11376	0.007256	0.021702	0	0	0	0	0	0	1	-360	360;
	3975	12125	0.001058	0.003149	0	0	0	0	0	0	1	-360	360;
	11376	10779	0.014744	0.048893	0	0	0	0	0	0	1	-360	360;
	4909	3978	0.040901	0.119017	0	0	0	0	0	0	1	-360	360;
	10049	11147	0.026653	0.063504	0	0	0	0	0	0	1	-360	360;
	6658	2865	0.005198	0.011909	0	0	0	0	0	0	1	-360	360;
	6658	1380	0.006603	0.020793	0	0	0	0	0	0	1	-360	360;
	1296	1916	0.023909	0.054744	0	0	0	0	0	0	1	-360	360;
	7306	1916	0.01286	0.038959	0	0	0	0	0	0	1	-360	360;
	9545	526	0	0.000826	0	0	0	0	0	0	1	-360	360;
	11091	526	0.005488	0.011256	0	0	0	0	0	0	1	-360	360;
	5448	526	0.018983	0.055579	0	0	0	0	0	0	1	-360	360;
	9545	6521	0.013479	0.038876	0	0	0	0	0	0	1	-360	360;
	10843	9158	0.010041	0.02205	0	0	0	0	0	0	1	-360	360;
	9158	13239	0.01224	0.037099	0	0	0	0	0	0	1	-360	360;
	10843	5952	0	0.00081	0	0	0	0	0	0	1	-360	360;
	8196	10843	0.011	0.030959	0	0	0	0	0	0	1	-360	360;
	4278	5952	0.010777	0.031554	0	0	0	0	0	0	1	-360	360;
	5952	12090	0.030521	0.069868	0	0	0	0	0	0	1	-360	360;
	5952	10106	0.016901	0.051198	0	0	0	0	0	0	1	-360	360;
	2415	10611	0.008868	0.027967	0	0	0	0	0	0	1	-360	360;
	13239	13015	0.006322	0.02319	0	0	0	0	0	0	1	-360	360;
	3991	8196	0.006488	0.019008	0	0	0	0	0	0	1	-360	360;
	1584	4278	0.006347	0.019372	0	0	0	0	0	0	1	-360	360;
	3944	1584	0.018339	0.038967	0	0	0	0	0	0	1	-360	360;
	1761	7629	0.037793	0.093752	0	0	0	0	0	0	1	-360	360;
	1818	1095	0.018603	0.048017	0	0	0	0	0	0	1	-360	360;
	6150	9580	0.04086	0.128893	0	0	0	0	0	0	1	-360	360;
	9580	1945	0.004587	0.010083	0	0	0	0	0	0	1	-360	360;
	3978	7964	0.012529	0.028661	0	0	0	0	0	0	1	-360	360;
	57	5918	0.016339	0.035909	0	0	0	0	0	0	1	-360	360;
	1818	10551	0.021702	0.058653	0	0	0	0	0	0	1	-360	360;
	9121	11727	0.037339	0.109364	0	0	0	0	0	0	1	-360	360;
	7623	4272	0.036554	0.11143	0	0	0	0	0	0	1	-360	360;
	8814	8985	0.027149	0.062149	0	0	0	0	0	0	1	-360	360;
	10039	5650	0.034281	0.103876	0	0	0	0	0	0	1	-360	360;
	10039	7390	0.023479	0.06843	0	0	0	0	0	0	1	-360	360;
	12177	10039	0.053149	0.116868	0	0	0	0	0	0	1	-360	360;
	11105	9838	0.004802	0.013298	0	0	0	0	0	0	1	-360	360;
	9838	744	0.012669	0.038231	0	0	0	0	0	0	1	-360	360;
	7188	3537	0.040339	0.079496	0	0	0	0	0	0	1	-360	360;
	12540	2414	0.054521	0.171992	0	0	0	0	0	0	1	-360	360;
	8207	2414	0.05514	0.122	0	0	0	0	0	0	1	-360	360;
	3332	2414	0.056157	0.119339	0	0	0	0	0	0	1	-360	360;
	11697	6579	0.034248	0.075289	0	0	0	0	0	0	1	-360	360;
	4894	6989	0.0018	0.01865	0	0	0	0	0	0	1	-360	360;
	11041	6449	0	0.000793	0	0	0	0	0	0	1	-360	360;
	11514	10771	0.023231	0.064983	0	0	0	0	0	0	1	-360	360;
	2416	3047	0.00163	0.01087	0	0	0	0	0	0	1	-360	360;
	9381	47	0.020727	0.053512	0	0	0	0	0	0	1	-360	360;
	744	10531	0.022364	0.064488	0	0	0	0	0	0	1	-360	360;
	1296	6434	0.014579	0.04805	0	0	0	0	0	0	1	-360	360;
	1296	6434	0.014281	0.048992	0	0	0	0	0	0	1	-360	360;
	10698	13036	0.00228	0.02412	0	0	0	0	0	0	1	-360	360;
	11075	2142	0.004298	0.009132	0	0	0	0	0	0	1	-360	360;
	4898	11075	0.003826	0.010347	0	0	0	0	0	0	1	-360	360;
	4898	2142	0	0.000777	0	0	0	0	0	0	1	-360	360;
	11597	2142	0.042256	0.092917	0	0	0	0	0	0	1	-360	360;
	5372	7313	0.005021	0.029341	0	0	0	0	0	0	1	-360	360;
	708	1857	0.066248	0.151661	0	0	0	0	0	0	1	-360	360;
	1857	3016	0.007587	0.016702	0	0	0	0	0	0	1	-360	360;
	3937	7300	0.008579	0.019636	0	0	0	0	0	0	1	-360	360;
	12177	8503	0.008942	0.024992	0	0	0	0	0	0	1	-360	360;
	9935	5235	0.049124	0.154983	0	0	0	0	0	0	1	-360	360;
	3125	9935	0.018289	0.04186	0	0	0	0	0	0	1	-360	360;
	6989	6364	0.00064	0.00617	0	0	0	0	0	0	1	-360	360;
	5597	4879	0	0.00076	0	0	0	0	0	0	1	-360	360;
	5597	4606	0.024421	0.073331	0	0	0	0	0	0	1	-360	360;
	4879	12981	0	0.000744	0	0	0	0	0	0	1	-360	360;
	10231	13006	0.025653	0.056388	0	0	0	0	0	0	1	-360	360;
	6343	7782	0.003711	0.02356	0	0	0	0	0	0	1	-360	360;
	4362	6387	0.015777	0.089537	0	0	0	0	0	0	1	-360	360;
	1945	4362	0.001587	0.005612	0	0	0	0	0	0	1	-360	360;
	1380	6122	0.008934	0.027091	0	0	0	0	0	0	1	-360	360;
	2893	2865	0.021298	0.052603	0	0	0	0	0	0	1	-360	360;
	11215	12756	0.01036	0.05943	0	0	0	0	0	0	1	-360	360;
	9400	1316	0.036438	0.114909	0	0	0	0	0	0	1	-360	360;
	6925	9400	0.038876	0.113207	0	0	0	0	0	0	1	-360	360;
	5811	10665	0.045058	0.127901	0	0	0	0	0	0	1	-360	360;
	10665	3125	0.014083	0.042661	0	0	0	0	0	0	1	-360	360;
	1325	708	0.016537	0.052157	0	0	0	0	0	0	1	-360	360;
	11437	8703	0.045711	0.136504	0	0	0	0	0	0	1	-360	360;
	2517	883	0.00579	0.06394	0	0	0	0	0	0	1	-360	360;
	2522	1870	0.012512	0.03381	0	0	0	0	0	0	1	-360	360;
	1870	953	0.053149	0.149512	0	0	0	0	0	0	1	-360	360;
	2522	7301	0.012248	0.034471	0	0	0	0	0	0	1	-360	360;
	7301	4708	0.052041	0.152388	0	0	0	0	0	0	1	-360	360;
	1316	6150	0.024471	0.056033	0	0	0	0	0	0	1	-360	360;
	2679	3501	0.008752	0.016322	0	0	0	0	0	0	1	-360	360;
	2517	10256	0.00211	0.01832	0	0	0	0	0	0	1	-360	360;
	5738	10088	0	0.00086	0	0	0	0	0	0	1	-360	360;
	2636	117	0	0.000876	0	0	0	0	0	0	1	-360	360;
	3594	117	0.007248	0.021107	0	0	0	0	0	0	1	-360	360;
	2499	2636	0.036653	0.106661	0	0	0	0	0	0	1	-360	360;
	12496	5986	0.003107	0.007727	0	0	0	0	0	0	1	-360	360;
	5986	8101	0.013612	0.029917	0	0	0	0	0	0	1	-360	360;
	10705	7629	0.064182	0.141124	0	0	0	0	0	0	1	-360	360;
	10154	7629	0.058331	0.150579	0	0	0	0	0	0	1	-360	360;
	10654	7303	0.00337	0.03668	0	0	0	0	0	0	1	-360	360;
	5980	11514	0.033959	0.080934	0	0	0	0	0	0	1	-360	360;
	10289	12177	0	0.000826	0	0	0	0	0	0	1	-360	360;
	10154	12177	0.07314	0.181438	0	0	0	0	0	0	1	-360	360;
	12177	8329	0.01476	0.04295	0	0	0	0	0	0	1	-360	360;
	1316	11871	0.018488	0.044091	0	0	0	0	0	0	1	-360	360;
	4769	2517	0.00025	0.00227	0	0	0	0	0	0	1	-360	360;
	4769	2517	0.00025	0.00222	0	0	0	0	0	0	1	-360	360;
	2517	1720	0.00179	0.01494	0	0	0	0	0	0	1	-360	360;
	2216	4169	0.00133	0.01403	0	0	0	0	0	0	1	-360	360;
	1502	2442	0.027438	0.076752	0	0	0	0	0	0	1	-360	360;
	13239	5699	0	0.00081	0	0	0	0	0	0	1	-360	360;
	5699	10106	0.014347	0.03705	0	0	0	0	0	0	1	-360	360;
	11215	7575	0.00286	0.015591	0	0	0	0	0	0	1	-360	360;
	9304	8101	0	0.000793	0	0	0	0	0	0	1	-360	360;
	8101	6434	0.008231	0.021273	0	0	0	0	0	0	1	-360	360;
	1773	10791	0.013653	0.031231	0	0	0	0	0	0	1	-360	360;
	4571	7138	0	0.000777	0	0	0	0	0	0	1	-360	360;
	12909	11554	0.028364	0.063479	0	0	0	0	0	0	1	-360	360;
	5918	11554	0.023636	0.051983	0	0	0	0	0	0	1	-360	360;
	688	11554	0.022752	0.066248	0	0	0	0	0	0	1	-360	360;
	9402	8207	0.004512	0.013099	0	0	0	0	0	0	1	-360	360;
	9402	3332	0.006331	0.015124	0	0	0	0	0	0	1	-360	360;
	7497	1325	0.048058	0.143504	0	0	0	0	0	0	1	-360	360;
	11557	3564	0.02676	0.077901	0	0	0	0	0	0	1	-360	360;
	3871	838	0	0.000744	0	0	0	0	0	0	1	-360	360;
	6856	1720	0.00031	0.00329	0	0	0	0	0	0	1	-360	360;
	6856	1720	0.00035	0.00322	0	0	0	0	0	0	1	-360	360;
	12090	1020	0.01276	0.028033	0	0	0	0	0	0	1	-360	360;
	10812	11357	0.049983	0.088198	0	0	0	0	0	0	1	-360	360;
	7127	5448	0	0.000876	0	0	0	0	0	0	1	-360	360;
	6253	5448	0.020157	0.04643	0	0	0	0	0	0	1	-360	360;
	11784	12056	0.050554	0.09781	0	0	0	0	0	0	1	-360	360;
	10519	11784	0.050157	0.093818	0	0	0	0	0	0	1	-360	360;
	11127	5756	0.03338	0.08238	0	0	0	0	0	0	1	-360	360;
	180	5756	0.041554	0.156107	0	0	0	0	0	0	1	-360	360;
	5508	12532	0.04676	0.124372	0	0	0	0	0	0	1	-360	360;
	3041	12532	0.043868	0.116678	0	0	0	0	0	0	1	-360	360;
	10046	8328	0.015529	0.049017	0	0	0	0	0	0	1	-360	360;
	10046	2022	0.019471	0.052802	0	0	0	0	0	0	1	-360	360;
	6978	8078	0.030339	0.062124	0	0	0	0	0	0	1	-360	360;
	12604	6978	0.043488	0.11219	0	0	0	0	0	0	1	-360	360;
	6978	8540	0.025934	0.052198	0	0	0	0	0	0	1	-360	360;
	11385	11022	0.024372	0.057347	0	0	0	0	0	0	1	-360	360;
	11385	9765	0.023868	0.058446	0	0	0	0	0	0	1	-360	360;
	11320	2022	0.032636	0.103934	0	0	0	0	0	0	1	-360	360;
	13260	11320	0.02695	0.081719	0	0	0	0	0	0	1	-360	360;
	8334	6514	0.022521	0.13874	0	0	0	0	0	0	1	-360	360;
	8334	4491	0.0117	0.053091	0	0	0	0	0	0	1	-360	360;
	8334	10454	0.001401	0.006789	0	0	0	0	0	0	1	-360	360;
	8334	3439	0.00139	0.0113	0	0	0	0	0	0	1	-360	360;
	3902	8334	0.01818	0.095349	0	0	0	0	0	0	1	-360	360;
	11733	8334	0.005729	0.043599	0	0	0	0	0	0	1	-360	360;
	8334	13559	0.00095	0.006409	0	0	0	0	0	0	1	-360	360;
	8334	12405	0.00093	0.006531	0	0	0	0	0	0	1	-360	360;
	8334	7184	0.025659	0.11894	0	0	0	0	0	0	1	-360	360;
	8334	1011	0.01036	0.04869	0	0	0	0	0	0	1	-360	360;
	7529	6060	0.022231	0.051521	0	0	0	0	0	0	1	-360	360;
	1568	6060	0.038099	0.092579	0	0	0	0	0	0	1	-360	360;
	2988	1826	0.018041	0.066793	0	0	0	0	0	0	1	-360	360;
	7010	1826	0.019942	0.075554	0	0	0	0	0	0	1	-360	360;
	264	12403	0.028347	0.064727	0	0	0	0	0	0	1	-360	360;
	264	371	0.028124	0.066826	0	0	0	0	0	0	1	-360	360;
	9290	1842	0.017107	0.031438	0	0	0	0	0	0	1	-360	360;
	67	1842	0.028322	0.109331	0	0	0	0	0	0	1	-360	360;
	6552	11093	0	0.000413	0	0	0	0	0	0	1	-360	360;
	6552	2187	0.002248	0.008413	0	0	0	0	0	0	1	-360	360;
	11093	5193	0.003421	0.010397	0	0	0	0	0	0	1	-360	360;
	4291	11093	0.039554	0.071554	0	0	0	0	0	0	1	-360	360;
	2079	11093	0.012612	0.036818	0	0	0	0	0	0	1	-360	360;
	8636	219	0.004471	0.014041	0	0	0	0	0	0	1	-360	360;
	219	5775	0.013628	0.050934	0	0	0	0	0	0	1	-360	360;
	219	4874	0.002438	0.007959	0	0	0	0	0	0	1	-360	360;
	219	4960	0.014579	0.030132	0	0	0	0	0	0	1	-360	360;
	219	12402	0.04186	0.08719	0	0	0	0	0	0	1	-360	360;
	219	11847	0.072653	0.128339	0	0	0	0	0	0	1	-360	360;
	8607	1322	0.013612	0.049421	0	0	0	0	0	0	1	-360	360;
	8607	11720	0.037256	0.107727	0	0	0	0	0	0	1	-360	360;
	3902	4227	0.003349	0.0237	0	0	0	0	0	0	1	-360	360;
	3902	4227	0.003281	0.024159	0	0	0	0	0	0	1	-360	360;
	3902	4227	0.003599	0.023469	0	0	0	0	0	0	1	-360	360;
	3902	4227	0.003669	0.02299	0	0	0	0	0	0	1	-360	360;
	504	10789	0.009124	0.02781	0	0	0	0	0	0	1	-360	360;
	504	10789	0.008661	0.026479	0	0	0	0	0	0	1	-360	360;
	504	10789	0.008488	0.027033	0	0	0	0	0	0	1	-360	360;
	504	2560	0.002248	0.006554	0	0	0	0	0	0	1	-360	360;
	504	7403	0.001612	0.004636	0	0	0	0	0	0	1	-360	360;
	504	2187	0.012223	0.027331	0	0	0	0	0	0	1	-360	360;
	504	7863	0.002099	0.01181	0	0	0	0	0	0	1	-360	360;
	504	7863	0.002058	0.012058	0	0	0	0	0	0	1	-360	360;
	10789	1948	0.019008	0.055876	0	0	0	0	0	0	1	-360	360;
	10789	4106	0.020727	0.05795	0	0	0	0	0	0	1	-360	360;
	4291	11646	0.004471	0.010471	0	0	0	0	0	0	1	-360	360;
	12729	11646	0.010479	0.025967	0	0	0	0	0	0	1	-360	360;
	2560	7403	0.001636	0.004537	0	0	0	0	0	0	1	-360	360;
	5193	8636	0.001603	0.005256	0	0	0	0	0	0	1	-360	360;
	12489	453	0.001039	0.00812	0	0	0	0	0	0	1	-360	360;
	12489	453	0.001169	0.007961	0	0	0	0	0	0	1	-360	360;
	1322	5775	0.001802	0.006628	0	0	0	0	0	0	1	-360	360;
	7771	13329	0.00022	0.00217	0	0	0	0	0	0	1	-360	360;
	2405	13329	0.00067	0.0072	0	0	0	0	0	0	1	-360	360;
	3689	4291	0.019826	0.06962	0	0	0	0	0	0	1	-360	360;
	3902	6514	0.005919	0.02874	0	0	0	0	0	0	1	-360	360;
	3356	6514	0.001271	0.009651	0	0	0	0	0	0	1	-360	360;
	12489	6514	0.00126	0.00955	0	0	0	0	0	0	1	-360	360;
	10798	7771	0.0009	0.01149	0	0	0	0	0	0	1	-360	360;
	7771	12313	0.00115	0.01172	0	0	0	0	0	0	1	-360	360;
	3356	2505	0.00431	0.033729	0	0	0	0	0	0	1	-360	360;
	12489	12928	0.001841	0.0115	0	0	0	0	0	0	1	-360	360;
	4491	11733	0.008401	0.053719	0	0	0	0	0	0	1	-360	360;
	5735	4491	0.0013	0.006021	0	0	0	0	0	0	1	-360	360;
	13353	12490	0.016083	0.036347	0	0	0	0	0	0	1	-360	360;
	12490	12399	0.01576	0.037058	0	0	0	0	0	0	1	-360	360;
	9135	12490	0.039661	0.101099	0	0	0	0	0	0	1	-360	360;
	5925	562	0.015777	0.058099	0	0	0	0	0	0	1	-360	360;
	7033	562	0.044901	0.08681	0	0	0	0	0	0	1	-360	360;
	7065	4095	0.00173	0.01919	0	0	0	0	0	0	1	-360	360;
	4095	12632	0.00213	0.02194	0	0	0	0	0	0	1	-360	360;
	329	1946	0.00395	0.008331	0	0	0	0	0	0	1	-360	360;
	12235	1946	0.157273	0.234413	0	0	0	0	0	0	1	-360	360;
	329	5675	0.003421	0.009099	0	0	0	0	0	0	1	-360	360;
	5675	8991	0.06795	0.151058	0	0	0	0	0	0	1	-360	360;
	329	4574	0.040099	0.154554	0	0	0	0	0	0	1	-360	360;
	329	6858	0.070769	0.175471	0	0	0	0	0	0	1	-360	360;
	329	757	0.008504	0.028322	0	0	0	0	0	0	1	-360	360;
	329	13632	0.02738	0.069876	0	0	0	0	0	0	1	-360	360;
	1400	12992	0.010347	0.023041	0	0	0	0	0	0	1	-360	360;
	13260	12992	0.044281	0.083967	0	0	0	0	0	0	1	-360	360;
	1568	1400	0.055744	0.101992	0	0	0	0	0	0	1	-360	360;
	2420	12403	0.015446	0.026554	0	0	0	0	0	0	1	-360	360;
	12403	4875	0.035388	0.067603	0	0	0	0	0	0	1	-360	360;
	4358	10702	0.007273	0.013752	0	0	0	0	0	0	1	-360	360;
	10702	10212	0.00362	0.007033	0	0	0	0	0	0	1	-360	360;
	6119	10212	0.001182	0.003727	0	0	0	0	0	0	1	-360	360;
	6119	12418	0.013314	0.052752	0	0	0	0	0	0	1	-360	360;
	4358	5639	0.028372	0.075868	0	0	0	0	0	0	1	-360	360;
	11712	4358	0.033917	0.080413	0	0	0	0	0	0	1	-360	360;
	12298	7786	0.004669	0.03695	0	0	0	0	0	0	1	-360	360;
	7786	11733	0.004229	0.025271	0	0	0	0	0	0	1	-360	360;
	11127	9719	0.041909	0.072157	0	0	0	0	0	0	1	-360	360;
	6843	9719	0.031174	0.078554	0	0	0	0	0	0	1	-360	360;
	2941	2988	0.030777	0.114017	0	0	0	0	0	0	1	-360	360;
	2380	9865	0.021479	0.05195	0	0	0	0	0	0	1	-360	360;
	9865	2022	0.04881	0.09786	0	0	0	0	0	0	1	-360	360;
	10608	10500	0.043992	0.107769	0	0	0	0	0	0	1	-360	360;
	10608	2022	0.017347	0.059826	0	0	0	0	0	0	1	-360	360;
	4289	10212	0.006281	0.014802	0	0	0	0	0	0	1	-360	360;
	5639	4289	0.045818	0.086281	0	0	0	0	0	0	1	-360	360;
	9765	4289	0.022521	0.077653	0	0	0	0	0	0	1	-360	360;
	3004	8066	0.014132	0.053744	0	0	0	0	0	0	1	-360	360;
	1151	3004	0.025901	0.082504	0	0	0	0	0	0	1	-360	360;
	3548	3004	0.05905	0.233421	0	0	0	0	0	0	1	-360	360;
	10	6376	0.035174	0.062992	0	0	0	0	0	0	1	-360	360;
	10	9275	0.015901	0.059529	0	0	0	0	0	0	1	-360	360;
	6376	8602	0.005174	0.015322	0	0	0	0	0	0	1	-360	360;
	6376	9572	0.021471	0.063207	0	0	0	0	0	0	1	-360	360;
	8602	2896	0.030033	0.114521	0	0	0	0	0	0	1	-360	360;
	8602	8201	0.050537	0.090521	0	0	0	0	0	0	1	-360	360;
	8602	8201	0.026446	0.070959	0	0	0	0	0	0	1	-360	360;
	9308	6832	0.028157	0.075636	0	0	0	0	0	0	1	-360	360;
	9308	12312	0.021413	0.212132	0	0	0	0	0	0	1	-360	360;
	9308	12312	0.046446	0.096702	0	0	0	0	0	0	1	-360	360;
	12034	6813	0.042562	0.102702	0	0	0	0	0	0	1	-360	360;
	11014	6813	0.034314	0.071603	0	0	0	0	0	0	1	-360	360;
	6926	12729	0.034033	0.077537	0	0	0	0	0	0	1	-360	360;
	12604	6926	0.01919	0.035669	0	0	0	0	0	0	1	-360	360;
	11127	12056	0.046826	0.094587	0	0	0	0	0	0	1	-360	360;
	13290	4241	0.059909	0.111868	0	0	0	0	0	0	1	-360	360;
	11663	13290	0.020909	0.037934	0	0	0	0	0	0	1	-360	360;
	8558	9994	0.009727	0.025347	0	0	0	0	0	0	1	-360	360;
	7046	8558	0.01981	0.040331	0	0	0	0	0	0	1	-360	360;
	3847	8074	1.7e-05	0.000413	0	0	0	0	0	0	1	-360	360;
	868	3847	0.017091	0.04114	0	0	0	0	0	0	1	-360	360;
	3689	4874	0.03014	0.111744	0	0	0	0	0	0	1	-360	360;
	11435	2420	0.012628	0.026339	0	0	0	0	0	0	1	-360	360;
	1762	7686	0.032934	0.078298	0	0	0	0	0	0	1	-360	360;
	2798	1762	0.025198	0.050942	0	0	0	0	0	0	1	-360	360;
	4006	10733	0.025529	0.049421	0	0	0	0	0	0	1	-360	360;
	9135	10733	0.026124	0.06962	0	0	0	0	0	0	1	-360	360;
	580	9455	0.00036	0.00188	0	0	0	0	0	0	1	-360	360;
	580	9455	0.00043	0.00174	0	0	0	0	0	0	1	-360	360;
	12354	1220	0.026091	0.052298	0	0	0	0	0	0	1	-360	360;
	12354	1220	0.015413	0.048636	0	0	0	0	0	0	1	-360	360;
	1220	9691	0.024463	0.08705	0	0	0	0	0	0	1	-360	360;
	9455	5434	0.00638	0.04836	0	0	0	0	0	0	1	-360	360;
	9455	12567	0.002081	0.01214	0	0	0	0	0	0	1	-360	360;
	9455	12567	0.002081	0.01257	0	0	0	0	0	0	1	-360	360;
	13228	4103	0.005314	0.015017	0	0	0	0	0	0	1	-360	360;
	13228	371	0.004331	0.011752	0	0	0	0	0	0	1	-360	360;
	4103	371	0.00386	0.013322	0	0	0	0	0	0	1	-360	360;
	371	6832	0.02181	0.080802	0	0	0	0	0	0	1	-360	360;
	371	2444	0.036992	0.107727	0	0	0	0	0	0	1	-360	360;
	371	7581	0.029273	0.105248	0	0	0	0	0	0	1	-360	360;
	9765	12418	0.016909	0.05057	0	0	0	0	0	0	1	-360	360;
	12493	7065	0.00302	0.03553	0	0	0	0	0	0	1	-360	360;
	6858	11134	0.005983	0.023132	0	0	0	0	0	0	1	-360	360;
	10471	11134	0.003777	0.011421	0	0	0	0	0	0	1	-360	360;
	757	11134	0.050041	0.180397	0	0	0	0	0	0	1	-360	360;
	11134	13632	0.056562	0.179397	0	0	0	0	0	0	1	-360	360;
	7607	3118	0.012818	0.046769	0	0	0	0	0	0	1	-360	360;
	12731	12928	0.00286	0.017581	0	0	0	0	0	0	1	-360	360;
	856	13353	0.012554	0.029083	0	0	0	0	0	0	1	-360	360;
	4394	4295	0.000909	0.002876	0	0	0	0	0	0	1	-360	360;
	9688	856	0.018719	0.033769	0	0	0	0	0	0	1	-360	360;
	4532	9688	0.032149	0.059041	0	0	0	0	0	0	1	-360	360;
	11906	7299	0.005198	0.013488	0	0	0	0	0	0	1	-360	360;
	4241	11906	0.018636	0.044149	0	0	0	0	0	0	1	-360	360;
	4241	7299	0.014554	0.028438	0	0	0	0	0	0	1	-360	360;
	11663	4241	0.032934	0.155826	0	0	0	0	0	0	1	-360	360;
	4241	9666	0.015231	0.035843	0	0	0	0	0	0	1	-360	360;
	4241	13588	0.037248	0.089901	0	0	0	0	0	0	1	-360	360;
	4241	632	0.04643	0.09338	0	0	0	0	0	0	1	-360	360;
	6851	12502	0.075612	0.155504	0	0	0	0	0	0	1	-360	360;
	12502	6377	0.04057	0.08876	0	0	0	0	0	0	1	-360	360;
	12502	489	0.114628	0.20105	0	0	0	0	0	0	1	-360	360;
	6260	13628	0.018521	0.090322	0	0	0	0	0	0	1	-360	360;
	4106	13628	0.013702	0.052983	0	0	0	0	0	0	1	-360	360;
	2468	11533	0.006017	0.019198	0	0	0	0	0	0	1	-360	360;
	2468	9971	0.052686	0.090653	0	0	0	0	0	0	1	-360	360;
	2468	3118	0.065826	0.122818	0	0	0	0	0	0	1	-360	360;
	9765	11022	0.04443	0.105719	0	0	0	0	0	0	1	-360	360;
	13273	11787	0.00838	0.018942	0	0	0	0	0	0	1	-360	360;
	12231	11787	0.058017	0.142983	0	0	0	0	0	0	1	-360	360;
	2195	11494	0.00085	0.01148	0	0	0	0	0	0	1	-360	360;
	12632	11494	0.00172	0.0234	0	0	0	0	0	0	1	-360	360;
	11127	13273	0.04805	0.152256	0	0	0	0	0	0	1	-360	360;
	13273	8487	0.028372	0.069471	0	0	0	0	0	0	1	-360	360;
	9994	13273	0.038364	0.12214	0	0	0	0	0	0	1	-360	360;
	5508	8066	0.013463	0.05405	0	0	0	0	0	0	1	-360	360;
	4960	12402	0.025083	0.060273	0	0	0	0	0	0	1	-360	360;
	3776	13208	0.012438	0.047124	0	0	0	0	0	0	1	-360	360;
	13260	3776	0.022388	0.084818	0	0	0	0	0	0	1	-360	360;
	2941	4336	0.005471	0.020281	0	0	0	0	0	0	1	-360	360;
	7010	4336	0.107174	0.201248	0	0	0	0	0	0	1	-360	360;
	7529	2941	0.035058	0.124769	0	0	0	0	0	0	1	-360	360;
	2896	11533	0.015934	0.05876	0	0	0	0	0	0	1	-360	360;
	868	8074	0.020099	0.03862	0	0	0	0	0	0	1	-360	360;
	3162	5505	0.04176	0.103083	0	0	0	0	0	0	1	-360	360;
	5465	5505	0.066719	0.165537	0	0	0	0	0	0	1	-360	360;
	12567	7311	0.00013	0.00095	0	0	0	0	0	0	1	-360	360;
	4006	9812	0.033777	0.06776	0	0	0	0	0	0	1	-360	360;
	12156	9812	0.057388	0.126289	0	0	0	0	0	0	1	-360	360;
	2195	13376	0.0019	0.0204	0	0	0	0	0	0	1	-360	360;
	12231	12409	0.007529	0.025372	0	0	0	0	0	0	1	-360	360;
	12231	2374	0.023917	0.064636	0	0	0	0	0	0	1	-360	360;
	12231	8540	0.039612	0.099818	0	0	0	0	0	0	1	-360	360;
	12231	10867	0.012388	0.038802	0	0	0	0	0	0	1	-360	360;
	12409	10867	0.004281	0.015678	0	0	0	0	0	0	1	-360	360;
	11103	11712	0.013463	0.033289	0	0	0	0	0	0	1	-360	360;
	12644	11712	0.015893	0.039901	0	0	0	0	0	0	1	-360	360;
	12644	11103	0.005934	0.021686	0	0	0	0	0	0	1	-360	360;
	9436	11103	0.009107	0.02843	0	0	0	0	0	0	1	-360	360;
	12644	9436	0.005802	0.020083	0	0	0	0	0	0	1	-360	360;
	12644	2579	0.10495	0.203661	0	0	0	0	0	0	1	-360	360;
	12644	9038	0.026372	0.04995	0	0	0	0	0	0	1	-360	360;
	967	12298	0.00738	0.044	0	0	0	0	0	0	1	-360	360;
	12298	11733	0.010591	0.05825	0	0	0	0	0	0	1	-360	360;
	7125	967	0.01318	0.05764	0	0	0	0	0	0	1	-360	360;
	967	6450	0.0143	0.06869	0	0	0	0	0	0	1	-360	360;
	11127	3131	0.022099	0.080636	0	0	0	0	0	0	1	-360	360;
	1568	2380	0.039843	0.08557	0	0	0	0	0	0	1	-360	360;
	2380	12427	0.007909	0.022124	0	0	0	0	0	0	1	-360	360;
	12290	180	0.051017	0.111273	0	0	0	0	0	0	1	-360	360;
	9994	2374	0.010331	0.038289	0	0	0	0	0	0	1	-360	360;
	7686	3689	0.02343	0.058603	0	0	0	0	0	0	1	-360	360;
	12632	2818	0.0009	0.00851	0	0	0	0	0	0	1	-360	360;
	2818	8891	0.00157	0.01488	0	0	0	0	0	0	1	-360	360;
	7125	10005	0.00682	0.03224	0	0	0	0	0	0	1	-360	360;
	11663	12034	0.01719	0.033322	0	0	0	0	0	0	1	-360	360;
	12034	11014	0.080446	0.14995	0	0	0	0	0	0	1	-360	360;
	7033	6851	0.055678	0.123917	0	0	0	0	0	0	1	-360	360;
	1739	2444	0.014463	0.040066	0	0	0	0	0	0	1	-360	360;
	1739	12220	0.013107	0.048314	0	0	0	0	0	0	1	-360	360;
	6377	2798	0.123174	0.23414	0	0	0	0	0	0	1	-360	360;
	4574	8991	0.00557	0.011256	0	0	0	0	0	0	1	-360	360;
	4574	8289	0.007967	0.018719	0	0	0	0	0	0	1	-360	360;
	12235	8289	0.076711	0.159554	0	0	0	0	0	0	1	-360	360;
	808	9572	0.012669	0.049364	0	0	0	0	0	0	1	-360	360;
	808	7586	0.015719	0.058826	0	0	0	0	0	0	1	-360	360;
	4816	1994	0.007397	0.027686	0	0	0	0	0	0	1	-360	360;
	4816	3118	0.006967	0.02505	0	0	0	0	0	0	1	-360	360;
	8078	12826	0.037653	0.076876	0	0	0	0	0	0	1	-360	360;
	5925	67	0.011	0.029826	0	0	0	0	0	0	1	-360	360;
	7010	7529	0.042446	0.107058	0	0	0	0	0	0	1	-360	360;
	6858	10471	0.011008	0.025711	0	0	0	0	0	0	1	-360	360;
	10519	11014	0.014818	0.04719	0	0	0	0	0	0	1	-360	360;
	10519	11014	0.011802	0.040711	0	0	0	0	0	0	1	-360	360;
	11014	10519	0.010802	0.039769	0	0	0	0	0	0	1	-360	360;
	10519	2807	0.003331	0.011496	0	0	0	0	0	0	1	-360	360;
	10519	12290	0.041256	0.083702	0	0	0	0	0	0	1	-360	360;
	2807	11014	0.011223	0.030413	0	0	0	0	0	0	1	-360	360;
	11014	4394	0.105562	0.208901	0	0	0	0	0	0	1	-360	360;
	11014	3162	0.021612	0.065496	0	0	0	0	0	0	1	-360	360;
	12632	9582	0.0056	0.03178	0	0	0	0	0	0	1	-360	360;
	4106	1948	0.005364	0.01443	0	0	0	0	0	0	1	-360	360;
	8805	12326	0.007132	0.016347	0	0	0	0	0	0	1	-360	360;
	12326	3118	0.035281	0.111719	0	0	0	0	0	0	1	-360	360;
	2579	8805	0.027669	0.067686	0	0	0	0	0	0	1	-360	360;
	8805	4532	0.086157	0.181149	0	0	0	0	0	0	1	-360	360;
	6681	12354	0.019983	0.054711	0	0	0	0	0	0	1	-360	360;
	6681	6260	0.007041	0.032174	0	0	0	0	0	0	1	-360	360;
	5495	9290	0.050587	0.128397	0	0	0	0	0	0	1	-360	360;
	12354	10602	0.014983	0.031364	0	0	0	0	0	0	1	-360	360;
	10602	9160	0.010653	0.019	0	0	0	0	0	0	1	-360	360;
	8271	7029	0.00189	0.02064	0	0	0	0	0	0	1	-360	360;
	8271	7029	0.00227	0.01922	0	0	0	0	0	0	1	-360	360;
	7029	1255	0.00167	0.02044	0	0	0	0	0	0	1	-360	360;
	3902	9828	0.016341	0.07818	0	0	0	0	0	0	1	-360	360;
	11007	9828	0.005409	0.024729	0	0	0	0	0	0	1	-360	360;
	9828	2095	0.00525	0.04099	0	0	0	0	0	0	1	-360	360;
	12354	4978	0.00714	0.025347	0	0	0	0	0	0	1	-360	360;
	12354	5518	0.00819	0.025802	0	0	0	0	0	0	1	-360	360;
	9160	12354	0.006322	0.014314	0	0	0	0	0	0	1	-360	360;
	7611	1568	0.037678	0.087512	0	0	0	0	0	0	1	-360	360;
	7611	10552	0.010942	0.034529	0	0	0	0	0	0	1	-360	360;
	4978	10552	0.007041	0.02714	0	0	0	0	0	0	1	-360	360;
	10552	5518	0.002826	0.010496	0	0	0	0	0	0	1	-360	360;
	10798	8271	0.0001	0.00124	0	0	0	0	0	0	1	-360	360;
	10798	6525	3e-05	0.00027	0	0	0	0	0	0	1	-360	360;
	10798	9119	2e-05	0.00021	0	0	0	0	0	0	1	-360	360;
	10798	8271	0.00011	0.00133	0	0	0	0	0	0	1	-360	360;
	10798	13376	0.00062	0.00694	0	0	0	0	0	0	1	-360	360;
	3902	11413	0.005281	0.04189	0	0	0	0	0	0	1	-360	360;
	3902	5284	0.00014	0.000651	0	0	0	0	0	0	1	-360	360;
	3902	1635	0.000229	0.00107	0	0	0	0	0	0	1	-360	360;
	3902	4800	0.000229	0.001039	0	0	0	0	0	0	1	-360	360;
	3902	6991	0.00013	0.00086	0	0	0	0	0	0	1	-360	360;
	3902	10515	0.00013	0.00087	0	0	0	0	0	0	1	-360	360;
	3902	8332	0.001539	0.012651	0	0	0	0	0	0	1	-360	360;
	3902	1011	0.00806	0.047091	0	0	0	0	0	0	1	-360	360;
	12427	10500	0.04943	0.099331	0	0	0	0	0	0	1	-360	360;
	2532	489	0.050488	0.104149	0	0	0	0	0	0	1	-360	360;
	9691	10936	0.003702	0.008934	0	0	0	0	0	0	1	-360	360;
	1568	10936	0.074446	0.138033	0	0	0	0	0	0	1	-360	360;
	5508	12402	0.008479	0.021653	0	0	0	0	0	0	1	-360	360;
	5508	12402	0.005314	0.019479	0	0	0	0	0	0	1	-360	360;
	12493	2405	0.00104	0.01121	0	0	0	0	0	0	1	-360	360;
	2095	12731	0.01118	0.07193	0	0	0	0	0	0	1	-360	360;
	8487	6843	0.002777	0.008174	0	0	0	0	0	0	1	-360	360;
	9275	9971	0.006058	0.02362	0	0	0	0	0	0	1	-360	360;
	9582	10718	0.00192	0.01477	0	0	0	0	0	0	1	-360	360;
	11533	9971	0.025331	0.098711	0	0	0	0	0	0	1	-360	360;
	7033	5762	0.05886	0.113157	0	0	0	0	0	0	1	-360	360;
	740	5762	0.046851	0.110231	0	0	0	0	0	0	1	-360	360;
	1151	7010	0.02895	0.10819	0	0	0	0	0	0	1	-360	360;
	1528	12235	0.079438	0.11481	0	0	0	0	0	0	1	-360	360;
	1528	12235	0.077967	0.117314	0	0	0	0	0	0	1	-360	360;
	12235	9994	0.032669	0.091388	0	0	0	0	0	0	1	-360	360;
	1234	4394	0.008421	0.02657	0	0	0	0	0	0	1	-360	360;
	3162	1234	0.082727	0.14876	0	0	0	0	0	0	1	-360	360;
	6450	12567	0.00637	0.029669	0	0	0	0	0	0	1	-360	360;
	6069	1528	0.020223	0.033198	0	0	0	0	0	0	1	-360	360;
	6069	1528	0.017843	0.033876	0	0	0	0	0	0	1	-360	360;
	9765	4006	0.022058	0.05405	0	0	0	0	0	0	1	-360	360;
	9765	13126	0.019826	0.073909	0	0	0	0	0	0	1	-360	360;
	13588	4875	0.063107	0.131711	0	0	0	0	0	0	1	-360	360;
	7485	2444	0.012099	0.029041	0	0	0	0	0	0	1	-360	360;
	7485	2444	0.006273	0.019033	0	0	0	0	0	0	1	-360	360;
	7485	11269	0.007603	0.021273	0	0	0	0	0	0	1	-360	360;
	7485	11269	0.011041	0.027339	0	0	0	0	0	0	1	-360	360;
	12661	9917	0.011636	0.032562	0	0	0	0	0	0	1	-360	360;
	12661	2444	0.007091	0.019058	0	0	0	0	0	0	1	-360	360;
	9917	11269	0.006653	0.020967	0	0	0	0	0	0	1	-360	360;
	13546	12567	0.00212	0.01695	0	0	0	0	0	0	1	-360	360;
	13546	12567	0.002349	0.0155	0	0	0	0	0	0	1	-360	360;
	2444	12312	0.053901	0.092744	0	0	0	0	0	0	1	-360	360;
	1923	8201	0.003273	0.010033	0	0	0	0	0	0	1	-360	360;
	1923	3597	0.005066	0.013769	0	0	0	0	0	0	1	-360	360;
	8201	3597	0.007826	0.028587	0	0	0	0	0	0	1	-360	360;
	3597	7586	0.032769	0.058711	0	0	0	0	0	0	1	-360	360;
	11847	11720	0.053702	0.134017	0	0	0	0	0	0	1	-360	360;
	8271	3715	0.00147	0.01305	0	0	0	0	0	0	1	-360	360;
	6745	8271	0.00019	0.00195	0	0	0	0	0	0	1	-360	360;
	12832	8271	0.00019	0.00191	0	0	0	0	0	0	1	-360	360;
	7047	4346	0.006413	0.019917	0	0	0	0	0	0	1	-360	360;
	6069	4346	0.011562	0.044182	0	0	0	0	0	0	1	-360	360;
	8792	7047	0.002512	0.008421	0	0	0	0	0	0	1	-360	360;
	725	5495	0.041702	0.098711	0	0	0	0	0	0	1	-360	360;
	740	725	0.034347	0.061636	0	0	0	0	0	0	1	-360	360;
	7033	2103	0.004289	0.01243	0	0	0	0	0	0	1	-360	360;
	8727	2103	0.000686	0.001471	0	0	0	0	0	0	1	-360	360;
	7033	8392	0.006612	0.01886	0	0	0	0	0	0	1	-360	360;
	11415	8392	0.008388	0.023339	0	0	0	0	0	0	1	-360	360;
	7033	8727	0.004372	0.012174	0	0	0	0	0	0	1	-360	360;
	7033	11415	0.013413	0.047793	0	0	0	0	0	0	1	-360	360;
	1568	13208	0.08438	0.183769	0	0	0	0	0	0	1	-360	360;
	4006	5904	0.006372	0.01681	0	0	0	0	0	0	1	-360	360;
	12220	7586	0.028579	0.084149	0	0	0	0	0	0	1	-360	360;
	12220	6229	0.007008	0.022388	0	0	0	0	0	0	1	-360	360;
	11847	12638	0.003	0.005058	0	0	0	0	0	0	1	-360	360;
	416	12638	0.002669	0.005727	0	0	0	0	0	0	1	-360	360;
	12638	3273	0.003661	0.008818	0	0	0	0	0	0	1	-360	360;
	12638	6069	0.029843	0.069744	0	0	0	0	0	0	1	-360	360;
	416	5079	0.006438	0.01824	0	0	0	0	0	0	1	-360	360;
	416	5079	0.006322	0.018636	0	0	0	0	0	0	1	-360	360;
	12313	12493	0.00068	0.00797	0	0	0	0	0	0	1	-360	360;
	5079	12826	0.032364	0.088289	0	0	0	0	0	0	1	-360	360;
	5079	6069	0.027066	0.084496	0	0	0	0	0	0	1	-360	360;
	5079	6069	0.026554	0.086256	0	0	0	0	0	0	1	-360	360;
	6069	3273	0.029901	0.071322	0	0	0	0	0	0	1	-360	360;
	868	740	0.011901	0.023818	0	0	0	0	0	0	1	-360	360;
	67	740	0.012132	0.023322	0	0	0	0	0	0	1	-360	360;
	7184	11007	0.0098	0.062349	0	0	0	0	0	0	1	-360	360;
	868	10072	0.004182	0.012678	0	0	0	0	0	0	1	-360	360;
	67	10072	0.003636	0.014636	0	0	0	0	0	0	1	-360	360;
	12826	7046	0.01919	0.045041	0	0	0	0	0	0	1	-360	360;
	8328	5221	0.009793	0.037174	0	0	0	0	0	0	1	-360	360;
	5221	2022	0.004281	0.010339	0	0	0	0	0	0	1	-360	360;
	5221	2022	0.005521	0.021471	0	0	0	0	0	0	1	-360	360;
	1255	387	0.0011	0.01325	0	0	0	0	0	0	1	-360	360;
	1568	12427	0.018421	0.051372	0	0	0	0	0	0	1	-360	360;
	1568	1424	0.011521	0.042529	0	0	0	0	0	0	1	-360	360;
	3689	2079	0.025182	0.071554	0	0	0	0	0	0	1	-360	360;
	13260	4701	0	0.000413	0	0	0	0	0	0	1	-360	360;
	1994	6197	0.014603	0.03386	0	0	0	0	0	0	1	-360	360;
	6441	3689	0.022066	0.054521	0	0	0	0	0	0	1	-360	360;
	2798	6441	0.041471	0.090793	0	0	0	0	0	0	1	-360	360;
	6069	8792	0.008843	0.031901	0	0	0	0	0	0	1	-360	360;
	5511	12567	0.00038	0.003091	0	0	0	0	0	0	1	-360	360;
	5511	12567	0.0005	0.002901	0	0	0	0	0	0	1	-360	360;
	11269	12312	0.032132	0.318198	0	0	0	0	0	0	1	-360	360;
	11269	3118	0.058554	0.146124	0	0	0	0	0	0	1	-360	360;
	8265	11269	0.023917	0.073289	0	0	0	0	0	0	1	-360	360;
	11269	6197	0.040818	0.089273	0	0	0	0	0	0	1	-360	360;
	9582	3212	2e-05	0.00025	0	0	0	0	0	0	1	-360	360;
	9582	3212	2e-05	0.00026	0	0	0	0	0	0	1	-360	360;
	3212	3923	0.00082	0.00822	0	0	0	0	0	0	1	-360	360;
	3131	9038	0.014339	0.031777	0	0	0	0	0	0	1	-360	360;
	7010	9298	0.011364	0.034364	0	0	0	0	0	0	1	-360	360;
	9298	4304	0.000347	0.001066	0	0	0	0	0	0	1	-360	360;
	6229	12312	0.012752	0.036041	0	0	0	0	0	0	1	-360	360;
	7581	12312	0.000669	0.001281	0	0	0	0	0	0	1	-360	360;
	7581	12312	0.000661	0.001314	0	0	0	0	0	0	1	-360	360;
	13126	5904	0.005289	0.014942	0	0	0	0	0	0	1	-360	360;
	2798	12156	0.004256	0.007364	0	0	0	0	0	0	1	-360	360;
	2532	12156	0.003992	0.009107	0	0	0	0	0	0	1	-360	360;
	9666	11435	0.041488	0.08662	0	0	0	0	0	0	1	-360	360;
	11435	13588	0.00486	0.010132	0	0	0	0	0	0	1	-360	360;
	11435	632	0.00495	0.009934	0	0	0	0	0	0	1	-360	360;
	632	13588	0.00505	0.009719	0	0	0	0	0	0	1	-360	360;
	1424	12427	0.00776	0.027521	0	0	0	0	0	0	1	-360	360;
	3041	3548	0.00262	0.007628	0	0	0	0	0	0	1	-360	360;
	8265	6197	0.001793	0.005471	0	0	0	0	0	0	1	-360	360;
	13632	5465	0.000537	0.015661	0	0	0	0	0	0	1	-360	360;
	7010	4304	0.01124	0.036826	0	0	0	0	0	0	1	-360	360;
	10322	7215	0.00156	0.01791	0	0	0	0	0	0	1	-360	360;
	7215	10709	0.00155	0.01686	0	0	0	0	0	0	1	-360	360;
	7215	10709	0.00152	0.01719	0	0	0	0	0	0	1	-360	360;
	12158	7215	0.00068	0.00644	0	0	0	0	0	0	1	-360	360;
	12158	7215	0.00069	0.0063	0	0	0	0	0	0	1	-360	360;
	7215	8477	0.00112	0.01064	0	0	0	0	0	0	1	-360	360;
	3071	7270	0.00151	0.011539	0	0	0	0	0	0	1	-360	360;
	11977	3071	0.00788	0.05236	0	0	0	0	0	0	1	-360	360;
	9388	10709	0.00128	0.0126	0	0	0	0	0	0	1	-360	360;
	1165	9388	0.00037	0.00409	0	0	0	0	0	0	1	-360	360;
	1165	9388	0.00036	0.00417	0	0	0	0	0	0	1	-360	360;
	9388	8477	0.00122	0.0112	0	0	0	0	0	0	1	-360	360;
	11977	12261	0.006591	0.04075	0	0	0	0	0	0	1	-360	360;
	4131	7270	0.007961	0.05756	0	0	0	0	0	0	1	-360	360;
	7441	3791	0.00059	0.0085	0	0	0	0	0	0	1	-360	360;
	3791	8008	0.00084	0.0118	0	0	0	0	0	0	1	-360	360;
	3791	9434	0.00178	0.01804	0	0	0	0	0	0	1	-360	360;
	7264	8432	0.00937	0.073969	0	0	0	0	0	0	1	-360	360;
	7264	4629	0.011081	0.06593	0	0	0	0	0	0	1	-360	360;
	7264	4082	0.009669	0.04251	0	0	0	0	0	0	1	-360	360;
	10124	13346	0.00073	0.00713	0	0	0	0	0	0	1	-360	360;
	7678	8008	0.00034	0.00345	0	0	0	0	0	0	1	-360	360;
	10910	10320	0.00012	0.00128	0	0	0	0	0	0	1	-360	360;
	10910	388	0.00012	0.00123	0	0	0	0	0	0	1	-360	360;
	2456	2913	9e-05	0.00112	0	0	0	0	0	0	1	-360	360;
	1898	1199	0.00062	0.00866	0	0	0	0	0	0	1	-360	360;
	10910	1898	0.00157	0.01614	0	0	0	0	0	0	1	-360	360;
	10910	1898	0.0014	0.01829	0	0	0	0	0	0	1	-360	360;
	430	4014	0.00078	0.00805	0	0	0	0	0	0	1	-360	360;
	4014	6044	0.00094	0.01427	0	0	0	0	0	0	1	-360	360;
	11378	8008	0.00024	0.00283	0	0	0	0	0	0	1	-360	360;
	8008	1199	0.00099	0.00865	0	0	0	0	0	0	1	-360	360;
	10910	8008	0.0011	0.01426	0	0	0	0	0	0	1	-360	360;
	8432	6284	0.006919	0.050521	0	0	0	0	0	0	1	-360	360;
	5959	7459	0.0006	0.00646	0	0	0	0	0	0	1	-360	360;
	5959	8466	0.00127	0.0128	0	0	0	0	0	0	1	-360	360;
	5959	13463	0.00052	0.00496	0	0	0	0	0	0	1	-360	360;
	5959	11565	0.00064	0.00886	0	0	0	0	0	0	1	-360	360;
	5959	11565	0.00072	0.00868	0	0	0	0	0	0	1	-360	360;
	430	12141	0.00232	0.02522	0	0	0	0	0	0	1	-360	360;
	10910	430	0.0003	0.00304	0	0	0	0	0	0	1	-360	360;
	10910	430	0.00032	0.00306	0	0	0	0	0	0	1	-360	360;
	10910	430	0.00028	0.00307	0	0	0	0	0	0	1	-360	360;
	13346	11056	0.00092	0.00969	0	0	0	0	0	0	1	-360	360;
	13346	8466	0.00181	0.01696	0	0	0	0	0	0	1	-360	360;
	13346	9434	0.00132	0.01148	0	0	0	0	0	0	1	-360	360;
	11056	6044	0.00108	0.01087	0	0	0	0	0	0	1	-360	360;
	7459	12141	0.00209	0.01924	0	0	0	0	0	0	1	-360	360;
	13463	7459	0.00036	0.00457	0	0	0	0	0	0	1	-360	360;
	7459	2456	0.0017	0.01986	0	0	0	0	0	0	1	-360	360;
	1199	3272	0.00059	0.00982	0	0	0	0	0	0	1	-360	360;
	6921	4082	0.00494	0.03182	0	0	0	0	0	0	1	-360	360;
	6921	4082	0.004841	0.03245	0	0	0	0	0	0	1	-360	360;
	6044	9434	0.00036	0.00391	0	0	0	0	0	0	1	-360	360;
	11565	2456	0.00071	0.00983	0	0	0	0	0	0	1	-360	360;
	11565	2456	0.00081	0.00963	0	0	0	0	0	0	1	-360	360;
	7850	1144	0.006271	0.03645	0	0	0	0	0	0	1	-360	360;
	11323	12959	0.186281	0.374281	0	0	0	0	0	0	1	-360	360;
	11835	12959	0.089612	0.230339	0	0	0	0	0	0	1	-360	360;
	11561	6149	0.00393	0.04552	0	0	0	0	0	0	1	-360	360;
	5799	8882	0.004169	0.02651	0	0	0	0	0	0	1	-360	360;
	5799	8882	0.003831	0.02438	0	0	0	0	0	0	1	-360	360;
	5799	7850	0.00789	0.04725	0	0	0	0	0	0	1	-360	360;
	5799	7850	0.008039	0.046219	0	0	0	0	0	0	1	-360	360;
	1351	3321	0.040174	0.084397	0	0	0	0	0	0	1	-360	360;
	10826	12633	0.00166	0.01276	0	0	0	0	0	0	1	-360	360;
	13474	8882	0.00094	0.00599	0	0	0	0	0	0	1	-360	360;
	13474	8882	0.00107	0.00587	0	0	0	0	0	0	1	-360	360;
	9599	8560	0.011901	0.028364	0	0	0	0	0	0	1	-360	360;
	8560	3321	0.059091	0.102636	0	0	0	0	0	0	1	-360	360;
	11835	8560	0.090471	0.19919	0	0	0	0	0	0	1	-360	360;
	879	2402	0.00755	0.050919	0	0	0	0	0	0	1	-360	360;
	879	2402	0.0077	0.04988	0	0	0	0	0	0	1	-360	360;
	6231	11323	0.238893	0.453901	0	0	0	0	0	0	1	-360	360;
	657	6231	0.062719	0.104479	0	0	0	0	0	0	1	-360	360;
	7850	2402	0.000979	0.074909	0	0	0	0	0	0	1	-360	360;
	7850	2402	0.000961	0.076349	0	0	0	0	0	0	1	-360	360;
	7397	2402	0.002159	0.079031	0	0	0	0	0	0	1	-360	360;
	7397	2402	0.001961	0.09125	0	0	0	0	0	0	1	-360	360;
	657	11323	0.066636	0.183934	0	0	0	0	0	0	1	-360	360;
	6538	10041	0.00028	0.00282	0	0	0	0	0	0	1	-360	360;
	6538	8623	0.00018	0.00233	0	0	0	0	0	0	1	-360	360;
	1161	7955	4e-05	0.0004	0	0	0	0	0	0	1	-360	360;
	1161	8364	0.0017	0.01741	0	0	0	0	0	0	1	-360	360;
	1161	8623	0.00097	0.0123	0	0	0	0	0	0	1	-360	360;
	1161	2885	0.00057	0.00725	0	0	0	0	0	0	1	-360	360;
	7260	7955	0.00177	0.01811	0	0	0	0	0	0	1	-360	360;
	7260	10123	0.00049	0.00505	0	0	0	0	0	0	1	-360	360;
	7260	9640	0.00199	0.02534	0	0	0	0	0	0	1	-360	360;
	7260	10041	0.00024	0.00242	0	0	0	0	0	0	1	-360	360;
	781	6831	0.01379	0.0443	0	0	0	0	0	0	1	-360	360;
	781	9251	0.00615	0.04848	0	0	0	0	0	0	1	-360	360;
	781	12787	0.02224	0.07142	0	0	0	0	0	0	1	-360	360;
	781	8335	0.0084	0.06621	0	0	0	0	0	0	1	-360	360;
	11	1019	7e-05	0.00068	0	0	0	0	0	0	1	-360	360;
	11	9103	0.00669	0.05272	0	0	0	0	0	0	1	-360	360;
	11	12491	0.003	0.03068	0	0	0	0	0	0	1	-360	360;
	11	12420	0.01129	0.05381	0	0	0	0	0	0	1	-360	360;
	1019	2379	0.02312	0.11021	0	0	0	0	0	0	1	-360	360;
	1019	12491	0.0044	0.03467	0	0	0	0	0	0	1	-360	360;
	2379	851	0.00113	0.00362	0	0	0	0	0	0	1	-360	360;
	2379	851	0.00113	0.00362	0	0	0	0	0	0	1	-360	360;
	2379	3392	0.00603	0.02877	0	0	0	0	0	0	1	-360	360;
	11825	9103	0.01127	0.0691	0	0	0	0	0	0	1	-360	360;
	11825	214	0.01447	0.04647	0	0	0	0	0	0	1	-360	360;
	11825	6831	0.00858	0.06762	0	0	0	0	0	0	1	-360	360;
	11825	10797	0.00026	0.00161	0	0	0	0	0	0	1	-360	360;
	9103	214	0.0037	0.02918	0	0	0	0	0	0	1	-360	360;
	9103	11918	0.00847	0.0404	0	0	0	0	0	0	1	-360	360;
	9103	12974	0.00026	0.00161	0	0	0	0	0	0	1	-360	360;
	214	12491	0.00935	0.03001	0	0	0	0	0	0	1	-360	360;
	214	12491	0.00553	0.02634	0	0	0	0	0	0	1	-360	360;
	214	6019	0.00318	0.0251	0	0	0	0	0	0	1	-360	360;
	214	3392	0.00689	0.05429	0	0	0	0	0	0	1	-360	360;
	10736	3075	0.00617	0.02941	0	0	0	0	0	0	1	-360	360;
	10736	9251	0.00476	0.0375	0	0	0	0	0	0	1	-360	360;
	3075	10550	0.0157	0.07482	0	0	0	0	0	0	1	-360	360;
	4821	11337	0.02731	0.08769	0	0	0	0	0	0	1	-360	360;
	4821	9339	0.03423	0.10993	0	0	0	0	0	0	1	-360	360;
	6831	9565	0.00112	0.01141	0	0	0	0	0	0	1	-360	360;
	6831	6019	0.00895	0.07061	0	0	0	0	0	0	1	-360	360;
	6831	12089	0.01187	0.05656	0	0	0	0	0	0	1	-360	360;
	3446	9339	0.00922	0.04395	0	0	0	0	0	0	1	-360	360;
	3446	8335	0.00995	0.07845	0	0	0	0	0	0	1	-360	360;
	10550	12787	0.01548	0.04972	0	0	0	0	0	0	1	-360	360;
	10550	12089	0.00339	0.01616	0	0	0	0	0	0	1	-360	360;
	1856	11918	0.00851	0.04056	0	0	0	0	0	0	1	-360	360;
	9565	10797	0.0099	0.06074	0	0	0	0	0	0	1	-360	360;
	9339	1654	0.01001	0.07892	0	0	0	0	0	0	1	-360	360;
	9339	8335	0.01647	0.07854	0	0	0	0	0	0	1	-360	360;
	12491	1654	0.00392	0.03091	0	0	0	0	0	0	1	-360	360;
	12420	3392	0.00603	0.02877	0	0	0	0	0	0	1	-360	360;
	9313	8364	0.00043	0.00443	0	0	0	0	0	0	1	-360	360;
	9640	3730	6e-05	0.0008	0	0	0	0	0	0	1	-360	360;
	9640	1187	0.00029	0.00366	0	0	0	0	0	0	1	-360	360;
	9640	940	0.00051	0.00643	0	0	0	0	0	0	1	-360	360;
	9640	9200	0.00033	0.00424	0	0	0	0	0	0	1	-360	360;
	8364	5348	0.00034	0.00432	0	0	0	0	0	0	1	-360	360;
	8364	13005	0.00013	0.00145	0	0	0	0	0	0	1	-360	360;
	8364	1187	0.00018	0.00233	0	0	0	0	0	0	1	-360	360;
	5348	11896	0.00034	0.00432	0	0	0	0	0	0	1	-360	360;
	5348	6535	0.00208	0.02643	0	0	0	0	0	0	1	-360	360;
	5348	10858	0.00209	0.0266	0	0	0	0	0	0	1	-360	360;
	5348	2885	0.00057	0.00725	0	0	0	0	0	0	1	-360	360;
	11707	9704	0.00034	0.00153	0	0	0	0	0	0	1	-360	360;
	11707	9704	0.00034	0.00153	0	0	0	0	0	0	1	-360	360;
	13005	1187	8e-05	0.00089	0	0	0	0	0	0	1	-360	360;
	13005	12365	6e-05	0.00067	0	0	0	0	0	0	1	-360	360;
	9146	7900	0.00181	0.01428	0	0	0	0	0	0	1	-360	360;
	9146	7900	0.00181	0.01428	0	0	0	0	0	0	1	-360	360;
	9146	4389	0.00576	0.03535	0	0	0	0	0	0	1	-360	360;
	9146	13288	0.00576	0.03535	0	0	0	0	0	0	1	-360	360;
	9147	7550	0.00526	0.02505	0	0	0	0	0	0	1	-360	360;
	9147	7550	0.00526	0.02505	0	0	0	0	0	0	1	-360	360;
	9147	12357	0.00159	0.01612	0	0	0	0	0	0	1	-360	360;
	9147	12357	0.00159	0.01612	0	0	0	0	0	0	1	-360	360;
	9147	3941	0.00121	0.01227	0	0	0	0	0	0	1	-360	360;
	9147	3941	0.00121	0.01227	0	0	0	0	0	0	1	-360	360;
	10353	1207	0.00132	0.0063	0	0	0	0	0	0	1	-360	360;
	7293	4263	0.0019	0.00905	0	0	0	0	0	0	1	-360	360;
	7293	4263	0.0019	0.00905	0	0	0	0	0	0	1	-360	360;
	7293	2825	0.00025	0.00264	0	0	0	0	0	0	1	-360	360;
	7293	2825	0.0002	0.00204	0	0	0	0	0	0	1	-360	360;
	7293	1207	0.00031	0.00316	0	0	0	0	0	0	1	-360	360;
	7293	5969	0.00031	0.00316	0	0	0	0	0	0	1	-360	360;
	7900	3237	0.00145	0.01145	0	0	0	0	0	0	1	-360	360;
	7900	13211	0.00141	0.01116	0	0	0	0	0	0	1	-360	360;
	9058	10916	0.00036	0.00359	0	0	0	0	0	0	1	-360	360;
	9058	10916	0.00036	0.00359	0	0	0	0	0	0	1	-360	360;
	3237	13211	0.00066	0.00518	0	0	0	0	0	0	1	-360	360;
	11319	8501	0.00049	0.00496	0	0	0	0	0	0	1	-360	360;
	11319	3397	0.00028	0.00296	0	0	0	0	0	0	1	-360	360;
	11319	9768	0.00028	0.00304	0	0	0	0	0	0	1	-360	360;
	3760	7550	0.00654	0.03119	0	0	0	0	0	0	1	-360	360;
	3760	4971	0.00108	0.00517	0	0	0	0	0	0	1	-360	360;
	3760	4971	0.00108	0.00517	0	0	0	0	0	0	1	-360	360;
	8501	7747	0.00119	0.00941	0	0	0	0	0	0	1	-360	360;
	8501	7747	0.00119	0.00941	0	0	0	0	0	0	1	-360	360;
	4389	4970	0.0061	0.03744	0	0	0	0	0	0	1	-360	360;
	4389	4970	0.0061	0.03744	0	0	0	0	0	0	1	-360	360;
	4389	13288	0.00052	0.00321	0	0	0	0	0	0	1	-360	360;
	4970	13211	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	4970	13211	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	4970	12811	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	4970	12811	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	3399	11944	0.00026	0.00267	0	0	0	0	0	0	1	-360	360;
	3399	12634	0.00021	0.00224	0	0	0	0	0	0	1	-360	360;
	4971	9576	0.00173	0.01365	0	0	0	0	0	0	1	-360	360;
	4971	9576	0.00173	0.01365	0	0	0	0	0	0	1	-360	360;
	3397	7747	0.00037	0.00401	0	0	0	0	0	0	1	-360	360;
	3397	7747	0.00037	0.00401	0	0	0	0	0	0	1	-360	360;
	7747	13511	0.00073	0.00744	0	0	0	0	0	0	1	-360	360;
	7747	13511	0.00073	0.00744	0	0	0	0	0	0	1	-360	360;
	9159	9704	0.00031	0.00329	0	0	0	0	0	0	1	-360	360;
	9159	9704	0.00025	0.00254	0	0	0	0	0	0	1	-360	360;
	9159	9113	0.00031	0.0031	0	0	0	0	0	0	1	-360	360;
	9159	9113	0.00031	0.0031	0	0	0	0	0	0	1	-360	360;
	8950	8037	0.00034	0.00108	0	0	0	0	0	0	1	-360	360;
	13499	8037	0.00042	0.00449	0	0	0	0	0	0	1	-360	360;
	13499	12634	0.00025	0.00248	0	0	0	0	0	0	1	-360	360;
	7703	10667	0.00119	0.01202	0	0	0	0	0	0	1	-360	360;
	7703	10667	0.00119	0.01202	0	0	0	0	0	0	1	-360	360;
	7703	10916	0.00034	0.00341	0	0	0	0	0	0	1	-360	360;
	7703	10916	0.00034	0.00341	0	0	0	0	0	0	1	-360	360;
	7480	7788	0.00034	0.00369	0	0	0	0	0	0	1	-360	360;
	11944	12634	0.00053	0.00569	0	0	0	0	0	0	1	-360	360;
	11944	5969	0.00336	0.016	0	0	0	0	0	0	1	-360	360;
	10667	3276	0.00827	0.03943	0	0	0	0	0	0	1	-360	360;
	10667	3276	0.00827	0.03943	0	0	0	0	0	0	1	-360	360;
	12357	10663	0.00073	0.00785	0	0	0	0	0	0	1	-360	360;
	12357	3941	0.00117	0.00926	0	0	0	0	0	0	1	-360	360;
	10663	12626	0.00037	0.00401	0	0	0	0	0	0	1	-360	360;
	3730	12365	0.00012	0.00134	0	0	0	0	0	0	1	-360	360;
	11896	2609	0.00023	0.00298	0	0	0	0	0	0	1	-360	360;
	2273	5861	0.00087	0.01433	0	0	0	0	0	0	1	-360	360;
	2273	5861	0.00088	0.01455	0	0	0	0	0	0	1	-360	360;
	2273	12845	0.00133	0.01364	0	0	0	0	0	0	1	-360	360;
	2273	7206	0.00064	0.00644	0	0	0	0	0	0	1	-360	360;
	2273	10858	0.00025	0.00324	0	0	0	0	0	0	1	-360	360;
	2273	12897	0.00276	0.02823	0	0	0	0	0	0	1	-360	360;
	2273	2477	0.00266	0.02691	0	0	0	0	0	0	1	-360	360;
	2273	9769	0.00196	0.01981	0	0	0	0	0	0	1	-360	360;
	2273	7769	0.0023	0.03797	0	0	0	0	0	0	1	-360	360;
	5861	2609	0.00027	0.00344	0	0	0	0	0	0	1	-360	360;
	5861	940	0.00032	0.00526	0	0	0	0	0	0	1	-360	360;
	5861	7206	0.001	0.0101	0	0	0	0	0	0	1	-360	360;
	5861	7166	0.00017	0.00279	0	0	0	0	0	0	1	-360	360;
	12845	6535	0.00097	0.0123	0	0	0	0	0	0	1	-360	360;
	12845	7209	0.00228	0.02898	0	0	0	0	0	0	1	-360	360;
	12845	1652	0.00269	0.03425	0	0	0	0	0	0	1	-360	360;
	12845	10819	0.00096	0.00982	0	0	0	0	0	0	1	-360	360;
	2609	10858	0.00105	0.01739	0	0	0	0	0	0	1	-360	360;
	2609	9200	0.00028	0.00351	0	0	0	0	0	0	1	-360	360;
	940	7166	0.00017	0.00279	0	0	0	0	0	0	1	-360	360;
	5268	1828	0.00278	0.01325	0	0	0	0	0	0	1	-360	360;
	5268	6730	0.04068	0.19392	0	0	0	0	0	0	1	-360	360;
	1828	9755	0.01464	0.04701	0	0	0	0	0	0	1	-360	360;
	1828	3756	0.01526	0.07272	0	0	0	0	0	0	1	-360	360;
	1828	8947	0.02017	0.09615	0	0	0	0	0	0	1	-360	360;
	1828	10083	0.01254	0.05979	0	0	0	0	0	0	1	-360	360;
	9336	12328	0.0046	0.03624	0	0	0	0	0	0	1	-360	360;
	9336	2860	0.01993	0.064	0	0	0	0	0	0	1	-360	360;
	686	1733	0.00974	0.03128	0	0	0	0	0	0	1	-360	360;
	686	2367	0.01464	0.04701	0	0	0	0	0	0	1	-360	360;
	686	6730	0.04605	0.14789	0	0	0	0	0	0	1	-360	360;
	1733	9882	0.00287	0.00922	0	0	0	0	0	0	1	-360	360;
	2889	5975	0.00308	0.0315	0	0	0	0	0	0	1	-360	360;
	2889	7107	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	3855	2565	0.0033	0.02605	0	0	0	0	0	0	1	-360	360;
	3855	7001	0.00239	0.02444	0	0	0	0	0	0	1	-360	360;
	3855	11937	0.0003	0.00312	0	0	0	0	0	0	1	-360	360;
	3855	6308	0.00318	0.0251	0	0	0	0	0	0	1	-360	360;
	1258	6292	0.00043	0.00434	0	0	0	0	0	0	1	-360	360;
	1258	1498	0.00259	0.0204	0	0	0	0	0	0	1	-360	360;
	2565	11879	0.00305	0.01454	0	0	0	0	0	0	1	-360	360;
	2565	11879	0.00305	0.01454	0	0	0	0	0	0	1	-360	360;
	2565	7107	0.00119	0.00941	0	0	0	0	0	0	1	-360	360;
	9755	12328	0.02353	0.07557	0	0	0	0	0	0	1	-360	360;
	7798	13453	0.00306	0.02416	0	0	0	0	0	0	1	-360	360;
	7798	1124	0.00398	0.03138	0	0	0	0	0	0	1	-360	360;
	1860	12328	0.01471	0.07013	0	0	0	0	0	0	1	-360	360;
	391	7991	0.00325	0.03326	0	0	0	0	0	0	1	-360	360;
	391	7627	0.0006	0.00607	0	0	0	0	0	0	1	-360	360;
	391	7001	0.00146	0.01493	0	0	0	0	0	0	1	-360	360;
	391	8408	0.00061	0.0062	0	0	0	0	0	0	1	-360	360;
	391	555	0.00075	0.00356	0	0	0	0	0	0	1	-360	360;
	8925	7715	0.00921	0.0305	0	0	0	0	0	0	1	-360	360;
	2367	12328	0.01233	0.0396	0	0	0	0	0	0	1	-360	360;
	12689	5556	0.0117	0.05575	0	0	0	0	0	0	1	-360	360;
	12689	5556	0.0117	0.05575	0	0	0	0	0	0	1	-360	360;
	12689	1760	0.00169	0.00808	0	0	0	0	0	0	1	-360	360;
	12689	4034	0.00235	0.01851	0	0	0	0	0	0	1	-360	360;
	12689	4034	0.00235	0.01851	0	0	0	0	0	0	1	-360	360;
	12689	1124	0.00119	0.00941	0	0	0	0	0	0	1	-360	360;
	12689	1124	0.00119	0.00941	0	0	0	0	0	0	1	-360	360;
	11587	5891	0.00046	0.00497	0	0	0	0	0	0	1	-360	360;
	5975	7991	0.00024	0.00188	0	0	0	0	0	0	1	-360	360;
	12609	3756	0.01288	0.06141	0	0	0	0	0	0	1	-360	360;
	12609	6730	0.02105	0.10035	0	0	0	0	0	0	1	-360	360;
	10066	11937	0.001	0.00785	0	0	0	0	0	0	1	-360	360;
	10066	6308	0.008	0.06307	0	0	0	0	0	0	1	-360	360;
	1674	473	0.00046	0.00497	0	0	0	0	0	0	1	-360	360;
	1674	10641	0.00034	0.0036	0	0	0	0	0	0	1	-360	360;
	6787	6308	0.01068	0.0509	0	0	0	0	0	0	1	-360	360;
	6787	13501	0.00475	0.02262	0	0	0	0	0	0	1	-360	360;
	4538	555	0.00022	0.0024	0	0	0	0	0	0	1	-360	360;
	4538	1741	0.00054	0.00259	0	0	0	0	0	0	1	-360	360;
	473	6013	0.00041	0.00441	0	0	0	0	0	0	1	-360	360;
	473	6013	0.00041	0.00441	0	0	0	0	0	0	1	-360	360;
	473	10641	0.00031	0.0031	0	0	0	0	0	0	1	-360	360;
	473	4583	0.00259	0.0204	0	0	0	0	0	0	1	-360	360;
	13458	11396	0.0044	0.03467	0	0	0	0	0	0	1	-360	360;
	13458	5556	0.01353	0.10669	0	0	0	0	0	0	1	-360	360;
	9720	5556	0.01736	0.08274	0	0	0	0	0	0	1	-360	360;
	9720	5556	0.01736	0.08274	0	0	0	0	0	0	1	-360	360;
	5891	10641	0.00034	0.0036	0	0	0	0	0	0	1	-360	360;
	7627	7466	0.00035	0.00353	0	0	0	0	0	0	1	-360	360;
	7627	7466	0.00035	0.00353	0	0	0	0	0	0	1	-360	360;
	13453	1124	0.00131	0.01036	0	0	0	0	0	0	1	-360	360;
	6535	10858	1.5e-05	0.000199	0	0	0	0	0	0	1	-360	360;
	11396	5556	0.00637	0.05021	0	0	0	0	0	0	1	-360	360;
	5556	7715	0.0105	0.03466	0	0	0	0	0	0	1	-360	360;
	6308	13501	0.01576	0.07514	0	0	0	0	0	0	1	-360	360;
	6308	1498	0.00112	0.01202	0	0	0	0	0	0	1	-360	360;
	5798	4034	0.00028	0.0009	0	0	0	0	0	0	1	-360	360;
	9882	12564	0.0165	0.05297	0	0	0	0	0	0	1	-360	360;
	7206	12897	0.00233	0.02385	0	0	0	0	0	0	1	-360	360;
	4034	3861	0.00547	0.04315	0	0	0	0	0	0	1	-360	360;
	4034	3861	0.00547	0.04315	0	0	0	0	0	0	1	-360	360;
	10858	7769	0.00219	0.03613	0	0	0	0	0	0	1	-360	360;
	12809	12897	0.00032	0.00326	0	0	0	0	0	0	1	-360	360;
	12809	2683	0.00215	0.022	0	0	0	0	0	0	1	-360	360;
	12809	10885	0.00173	0.01767	0	0	0	0	0	0	1	-360	360;
	7910	12801	0.00318	0.0251	0	0	0	0	0	0	1	-360	360;
	7910	6125	0.00497	0.03923	0	0	0	0	0	0	1	-360	360;
	7910	10089	0.00199	0.01569	0	0	0	0	0	0	1	-360	360;
	7218	799	0.00147	0.01161	0	0	0	0	0	0	1	-360	360;
	7218	10277	0.00511	0.04032	0	0	0	0	0	0	1	-360	360;
	13373	12437	0.00103	0.00816	0	0	0	0	0	0	1	-360	360;
	13373	1628	0.00139	0.01098	0	0	0	0	0	0	1	-360	360;
	13373	2042	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	12437	2042	0.00199	0.01569	0	0	0	0	0	0	1	-360	360;
	6783	12292	0.01132	0.05397	0	0	0	0	0	0	1	-360	360;
	6783	9026	0.00358	0.02824	0	0	0	0	0	0	1	-360	360;
	6783	3930	0.00085	0.00404	0	0	0	0	0	0	1	-360	360;
	6783	12801	0.00358	0.02824	0	0	0	0	0	0	1	-360	360;
	6783	12801	0.00358	0.02824	0	0	0	0	0	0	1	-360	360;
	11052	11710	0.00086	0.0053	0	0	0	0	0	0	1	-360	360;
	11052	11710	0.00086	0.0053	0	0	0	0	0	0	1	-360	360;
	11052	10277	0.00219	0.01726	0	0	0	0	0	0	1	-360	360;
	11052	2152	0.00505	0.02408	0	0	0	0	0	0	1	-360	360;
	799	5865	0.00402	0.03169	0	0	0	0	0	0	1	-360	360;
	799	9036	0.00468	0.03687	0	0	0	0	0	0	1	-360	360;
	799	9036	0.00468	0.03687	0	0	0	0	0	0	1	-360	360;
	799	10277	0.00643	0.05068	0	0	0	0	0	0	1	-360	360;
	799	8947	0.04025	0.12927	0	0	0	0	0	0	1	-360	360;
	5865	10277	0.00302	0.02385	0	0	0	0	0	0	1	-360	360;
	160	3936	0.01417	0.06755	0	0	0	0	0	0	1	-360	360;
	160	5942	0.01503	0.04827	0	0	0	0	0	0	1	-360	360;
	160	12292	0.02556	0.12185	0	0	0	0	0	0	1	-360	360;
	3861	9036	0.02618	0.08407	0	0	0	0	0	0	1	-360	360;
	3936	12292	0.01105	0.05268	0	0	0	0	0	0	1	-360	360;
	5942	9711	0.00788	0.06213	0	0	0	0	0	0	1	-360	360;
	5942	9026	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	5942	1628	0.00199	0.01569	0	0	0	0	0	0	1	-360	360;
	6706	12292	0.02098	0.10003	0	0	0	0	0	0	1	-360	360;
	6706	3257	0.00793	0.03781	0	0	0	0	0	0	1	-360	360;
	12292	3930	0.01136	0.05414	0	0	0	0	0	0	1	-360	360;
	12292	3257	0.02597	0.12379	0	0	0	0	0	0	1	-360	360;
	9711	8947	0.02004	0.06436	0	0	0	0	0	0	1	-360	360;
	12897	8508	0.00229	0.02909	0	0	0	0	0	0	1	-360	360;
	9026	1628	0.00338	0.02667	0	0	0	0	0	0	1	-360	360;
	9026	1628	0.00338	0.02667	0	0	0	0	0	0	1	-360	360;
	12801	10089	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	6125	3257	0.01407	0.0452	0	0	0	0	0	0	1	-360	360;
	2683	3973	0.00128	0.01308	0	0	0	0	0	0	1	-360	360;
	5044	2226	0.00438	0.03452	0	0	0	0	0	0	1	-360	360;
	5044	13574	0.01992	0.15706	0	0	0	0	0	0	1	-360	360;
	5044	3257	0.03389	0.10884	0	0	0	0	0	0	1	-360	360;
	5044	1690	0.01313	0.10355	0	0	0	0	0	0	1	-360	360;
	5044	11788	0.00557	0.04393	0	0	0	0	0	0	1	-360	360;
	2226	4867	0.01008	0.03236	0	0	0	0	0	0	1	-360	360;
	2226	12783	0.01493	0.11768	0	0	0	0	0	0	1	-360	360;
	4867	10061	0.03074	0.09872	0	0	0	0	0	0	1	-360	360;
	4867	10061	0.04301	0.13813	0	0	0	0	0	0	1	-360	360;
	6791	10139	0.01407	0.0452	0	0	0	0	0	0	1	-360	360;
	6791	3257	0.02883	0.09257	0	0	0	0	0	0	1	-360	360;
	6791	1690	0.00796	0.06276	0	0	0	0	0	0	1	-360	360;
	6791	10003	0.0045	0.01446	0	0	0	0	0	0	1	-360	360;
	10139	2535	0.01712	0.05496	0	0	0	0	0	0	1	-360	360;
	13085	13574	0.02573	0.08263	0	0	0	0	0	0	1	-360	360;
	13085	2535	0.05152	0.16543	0	0	0	0	0	0	1	-360	360;
	13574	5266	0.01871	0.11474	0	0	0	0	0	0	1	-360	360;
	13574	11932	0.05489	0.17628	0	0	0	0	0	0	1	-360	360;
	13574	1690	0.01186	0.09351	0	0	0	0	0	0	1	-360	360;
	13574	7868	0.00169	0.01334	0	0	0	0	0	0	1	-360	360;
	13574	7868	0.00169	0.01334	0	0	0	0	0	0	1	-360	360;
	5266	263	0.00811	0.02603	0	0	0	0	0	0	1	-360	360;
	5266	10758	0.00034	0.00162	0	0	0	0	0	0	1	-360	360;
	5266	10758	0.00034	0.00162	0	0	0	0	0	0	1	-360	360;
	11932	263	0.00736	0.05805	0	0	0	0	0	0	1	-360	360;
	13396	12783	0.01094	0.0863	0	0	0	0	0	0	1	-360	360;
	13396	263	0.01214	0.09571	0	0	0	0	0	0	1	-360	360;
	12783	3623	0.00895	0.07061	0	0	0	0	0	0	1	-360	360;
	12783	1280	0.03063	0.09836	0	0	0	0	0	0	1	-360	360;
	8508	3973	0.00242	0.03076	0	0	0	0	0	0	1	-360	360;
	7720	10885	0.00083	0.00845	0	0	0	0	0	0	1	-360	360;
	7720	6700	0.0056	0.05722	0	0	0	0	0	0	1	-360	360;
	7720	8641	0.0036	0.03646	0	0	0	0	0	0	1	-360	360;
	7720	13258	0.001	0.0101	0	0	0	0	0	0	1	-360	360;
	3993	7364	0.02308	0.07413	0	0	0	0	0	0	1	-360	360;
	3993	11555	0.01683	0.05406	0	0	0	0	0	0	1	-360	360;
	9833	8481	0.03322	0.10667	0	0	0	0	0	0	1	-360	360;
	9833	8930	0.01094	0.0863	0	0	0	0	0	0	1	-360	360;
	11788	8464	0.01546	0.07369	0	0	0	0	0	0	1	-360	360;
	13430	8947	0.00526	0.02505	0	0	0	0	0	0	1	-360	360;
	13430	6786	0.02898	0.13817	0	0	0	0	0	0	1	-360	360;
	12288	7255	0.01773	0.08452	0	0	0	0	0	0	1	-360	360;
	12288	7172	0.01368	0.04393	0	0	0	0	0	0	1	-360	360;
	12288	1551	0.02365	0.07594	0	0	0	0	0	0	1	-360	360;
	12288	10601	0.00078	0.00372	0	0	0	0	0	0	1	-360	360;
	12288	10601	0.00078	0.00372	0	0	0	0	0	0	1	-360	360;
	7364	5941	0.02759	0.08859	0	0	0	0	0	0	1	-360	360;
	6782	12236	0.01576	0.05062	0	0	0	0	0	0	1	-360	360;
	8464	13630	0.02664	0.12702	0	0	0	0	0	0	1	-360	360;
	12670	13630	0.02162	0.06943	0	0	0	0	0	0	1	-360	360;
	12670	6986	0.03187	0.10233	0	0	0	0	0	0	1	-360	360;
	7907	11555	0.02168	0.06961	0	0	0	0	0	0	1	-360	360;
	11555	6786	0.03827	0.18245	0	0	0	0	0	0	1	-360	360;
	11555	6771	0.01465	0.06981	0	0	0	0	0	0	1	-360	360;
	11555	1551	0.00661	0.05209	0	0	0	0	0	0	1	-360	360;
	11555	1551	0.00782	0.06166	0	0	0	0	0	0	1	-360	360;
	5941	7255	0.0366	0.11752	0	0	0	0	0	0	1	-360	360;
	8481	8638	0.00326	0.01049	0	0	0	0	0	0	1	-360	360;
	8481	8638	0.00326	0.01049	0	0	0	0	0	0	1	-360	360;
	8481	8638	0.00326	0.01049	0	0	0	0	0	0	1	-360	360;
	8481	7255	0.00732	0.03491	0	0	0	0	0	0	1	-360	360;
	8481	7255	0.00732	0.03491	0	0	0	0	0	0	1	-360	360;
	8481	13630	0.03063	0.09836	0	0	0	0	0	0	1	-360	360;
	8481	13630	0.03063	0.09836	0	0	0	0	0	0	1	-360	360;
	8947	7172	0.01768	0.05677	0	0	0	0	0	0	1	-360	360;
	8947	6786	0.03438	0.16386	0	0	0	0	0	0	1	-360	360;
	8947	10083	0.01017	0.04848	0	0	0	0	0	0	1	-360	360;
	10397	6786	0	0.04891	0	0	0	0	0	0	1	-360	360;
	10397	13630	0.00081	0.00388	0	0	0	0	0	0	1	-360	360;
	10397	13630	0.00081	0.00388	0	0	0	0	0	0	1	-360	360;
	6786	1551	0.03824	0.18228	0	0	0	0	0	0	1	-360	360;
	6986	12236	0.02207	0.07087	0	0	0	0	0	0	1	-360	360;
	6986	10244	0.0165	0.05297	0	0	0	0	0	0	1	-360	360;
	8778	6700	0.00094	0.00958	0	0	0	0	0	0	1	-360	360;
	8778	3808	0.0009	0.01147	0	0	0	0	0	0	1	-360	360;
	8778	12251	2e-05	0.00026	0	0	0	0	0	0	1	-360	360;
	8778	12251	2e-05	0.00026	0	0	0	0	0	0	1	-360	360;
	8778	3973	0.00162	0.01656	0	0	0	0	0	0	1	-360	360;
	8778	3841	0.00039	0.00641	0	0	0	0	0	0	1	-360	360;
	6700	3841	0.00027	0.00444	0	0	0	0	0	0	1	-360	360;
	6700	1297	0.00288	0.02946	0	0	0	0	0	0	1	-360	360;
	6700	1297	0.00192	0.02444	0	0	0	0	0	0	1	-360	360;
	3808	12975	0.00094	0.01197	0	0	0	0	0	0	1	-360	360;
	3808	5794	0.00106	0.01087	0	0	0	0	0	0	1	-360	360;
	2776	7815	0.00677	0.05335	0	0	0	0	0	0	1	-360	360;
	2776	7815	0.00677	0.05335	0	0	0	0	0	0	1	-360	360;
	7117	716	0.00051	0.00242	0	0	0	0	0	0	1	-360	360;
	7117	716	0.00051	0.00242	0	0	0	0	0	0	1	-360	360;
	7117	8590	0.00026	0.00204	0	0	0	0	0	0	1	-360	360;
	7117	8590	0.00026	0.00204	0	0	0	0	0	0	1	-360	360;
	9735	716	0.00139	0.00663	0	0	0	0	0	0	1	-360	360;
	9735	716	0.00139	0.00663	0	0	0	0	0	0	1	-360	360;
	9735	8590	0.00075	0.00356	0	0	0	0	0	0	1	-360	360;
	9735	8590	0.00075	0.00356	0	0	0	0	0	0	1	-360	360;
	6152	11207	0.00321	0.03286	0	0	0	0	0	0	1	-360	360;
	6152	11207	0.00321	0.03286	0	0	0	0	0	0	1	-360	360;
	11427	7357	0.01875	0.06021	0	0	0	0	0	0	1	-360	360;
	11427	10988	0.02489	0.07991	0	0	0	0	0	0	1	-360	360;
	13613	3920	0.00115	0.0091	0	0	0	0	0	0	1	-360	360;
	13613	1712	0.00037	0.00164	0	0	0	0	0	0	1	-360	360;
	13613	5870	0.00082	0.00881	0	0	0	0	0	0	1	-360	360;
	13613	497	0.001	0.00785	0	0	0	0	0	0	1	-360	360;
	13613	10316	0.00043	0.00465	0	0	0	0	0	0	1	-360	360;
	7402	8046	0.00045	0.00203	0	0	0	0	0	0	1	-360	360;
	7402	11465	0.00028	0.00304	0	0	0	0	0	0	1	-360	360;
	2628	8854	0.00037	0.00164	0	0	0	0	0	0	1	-360	360;
	2628	13347	0.00033	0.00148	0	0	0	0	0	0	1	-360	360;
	2628	12605	0.0002	0.00216	0	0	0	0	0	0	1	-360	360;
	2628	4320	0.00037	0.00372	0	0	0	0	0	0	1	-360	360;
	9429	12605	0.00065	0.00697	0	0	0	0	0	0	1	-360	360;
	9429	1256	0.00015	0.0016	0	0	0	0	0	0	1	-360	360;
	3370	11207	0.002	0.00953	0	0	0	0	0	0	1	-360	360;
	3370	11207	0.002	0.00953	0	0	0	0	0	0	1	-360	360;
	5390	13347	0.0006	0.00268	0	0	0	0	0	0	1	-360	360;
	5390	11921	0.00239	0.01883	0	0	0	0	0	0	1	-360	360;
	5390	4320	0.00117	0.00926	0	0	0	0	0	0	1	-360	360;
	8590	11740	0.00026	0.00161	0	0	0	0	0	0	1	-360	360;
	8590	11740	0.00026	0.00161	0	0	0	0	0	0	1	-360	360;
	7773	4408	0.00061	0.00291	0	0	0	0	0	0	1	-360	360;
	7773	4408	0.00061	0.00291	0	0	0	0	0	0	1	-360	360;
	11271	3920	0.00062	0.00632	0	0	0	0	0	0	1	-360	360;
	11271	3920	0.00062	0.00632	0	0	0	0	0	0	1	-360	360;
	11271	497	0.001	0.00785	0	0	0	0	0	0	1	-360	360;
	11921	4320	0.00231	0.0182	0	0	0	0	0	0	1	-360	360;
	8046	11207	0.00233	0.01836	0	0	0	0	0	0	1	-360	360;
	8046	13624	0.0043	0.03389	0	0	0	0	0	0	1	-360	360;
	12605	4320	0.00017	0.00174	0	0	0	0	0	0	1	-360	360;
	12605	4320	0.00017	0.00174	0	0	0	0	0	0	1	-360	360;
	12605	11255	0.0005	0.00537	0	0	0	0	0	0	1	-360	360;
	4565	5870	0.006	0.0286	0	0	0	0	0	0	1	-360	360;
	4565	602	0.01187	0.05656	0	0	0	0	0	0	1	-360	360;
	4565	4739	0.00022	0.0024	0	0	0	0	0	0	1	-360	360;
	8511	7357	0.01419	0.04556	0	0	0	0	0	0	1	-360	360;
	8511	4015	0.01712	0.08161	0	0	0	0	0	0	1	-360	360;
	12068	4408	0.00173	0.00824	0	0	0	0	0	0	1	-360	360;
	12068	4408	0.00173	0.00824	0	0	0	0	0	0	1	-360	360;
	11465	11207	0.00199	0.01569	0	0	0	0	0	0	1	-360	360;
	11465	13624	0.00716	0.05648	0	0	0	0	0	0	1	-360	360;
	11451	7815	0.00915	0.07217	0	0	0	0	0	0	1	-360	360;
	11451	4320	0.00793	0.03781	0	0	0	0	0	0	1	-360	360;
	7357	10988	0.00796	0.06276	0	0	0	0	0	0	1	-360	360;
	7815	4320	0.01796	0.05768	0	0	0	0	0	0	1	-360	360;
	7815	13634	0.00013	0.00136	0	0	0	0	0	0	1	-360	360;
	13413	4408	0.00179	0.01412	0	0	0	0	0	0	1	-360	360;
	13413	7421	0.00338	0.01085	0	0	0	0	0	0	1	-360	360;
	5870	13624	0.0009	0.00706	0	0	0	0	0	0	1	-360	360;
	5870	13624	0.0009	0.00706	0	0	0	0	0	0	1	-360	360;
	12660	3606	0.02308	0.07413	0	0	0	0	0	0	1	-360	360;
	12660	6504	0.00669	0.05272	0	0	0	0	0	0	1	-360	360;
	8984	6755	0.02866	0.09203	0	0	0	0	0	0	1	-360	360;
	8984	3821	0.06565	0.21081	0	0	0	0	0	0	1	-360	360;
	8984	1579	0.03479	0.11173	0	0	0	0	0	0	1	-360	360;
	3623	1280	0.03243	0.10414	0	0	0	0	0	0	1	-360	360;
	1280	10061	0.00784	0.06182	0	0	0	0	0	0	1	-360	360;
	1280	7421	0.02196	0.07051	0	0	0	0	0	0	1	-360	360;
	6755	3158	0.01323	0.04249	0	0	0	0	0	0	1	-360	360;
	11373	602	0.00336	0.016	0	0	0	0	0	0	1	-360	360;
	11373	602	0.00325	0.01551	0	0	0	0	0	0	1	-360	360;
	11373	10909	0.02303	0.07395	0	0	0	0	0	0	1	-360	360;
	11373	3606	0.00845	0.02712	0	0	0	0	0	0	1	-360	360;
	602	10316	0.00678	0.03232	0	0	0	0	0	0	1	-360	360;
	3158	10909	0.0322	0.10342	0	0	0	0	0	0	1	-360	360;
	10061	6504	0.0002	0.00157	0	0	0	0	0	0	1	-360	360;
	10244	8181	0.05906	0.18966	0	0	0	0	0	0	1	-360	360;
	10244	9548	0.00676	0.0217	0	0	0	0	0	0	1	-360	360;
	3821	8930	0.02488	0.19613	0	0	0	0	0	0	1	-360	360;
	3821	9548	0.00676	0.0217	0	0	0	0	0	0	1	-360	360;
	10586	1297	0.00094	0.00956	0	0	0	0	0	0	1	-360	360;
	10586	12110	0.00025	0.00258	0	0	0	0	0	0	1	-360	360;
	1297	12110	0.00082	0.00843	0	0	0	0	0	0	1	-360	360;
	1297	10975	0.00156	0.01988	0	0	0	0	0	0	1	-360	360;
	12110	12975	0.00146	0.01855	0	0	0	0	0	0	1	-360	360;
	6313	7276	0.02449	0.07865	0	0	0	0	0	0	1	-360	360;
	6313	7879	0.01055	0.08316	0	0	0	0	0	0	1	-360	360;
	6037	13065	0.02753	0.08841	0	0	0	0	0	0	1	-360	360;
	6037	2039	0.02393	0.07684	0	0	0	0	0	0	1	-360	360;
	12296	484	0.01566	0.07466	0	0	0	0	0	0	1	-360	360;
	4015	10988	0.00519	0.04095	0	0	0	0	0	0	1	-360	360;
	4015	6257	0.02286	0.0734	0	0	0	0	0	0	1	-360	360;
	4015	2824	0.02226	0.08078	0	0	0	0	0	0	1	-360	360;
	8178	484	0.01817	0.08662	0	0	0	0	0	0	1	-360	360;
	8178	2567	0.01576	0.07514	0	0	0	0	0	0	1	-360	360;
	11777	13065	0.02697	0.0866	0	0	0	0	0	0	1	-360	360;
	11777	1722	0.00929	0.02983	0	0	0	0	0	0	1	-360	360;
	11777	7276	0.03508	0.11264	0	0	0	0	0	0	1	-360	360;
	11777	10716	0.01405	0.11077	0	0	0	0	0	0	1	-360	360;
	13065	10199	0.03479	0.11173	0	0	0	0	0	0	1	-360	360;
	10259	2096	0.01937	0.06219	0	0	0	0	0	0	1	-360	360;
	10259	1912	0.03778	0.12132	0	0	0	0	0	0	1	-360	360;
	10259	2037	0.00497	0.03923	0	0	0	0	0	0	1	-360	360;
	1277	1497	0.01864	0.08888	0	0	0	0	0	0	1	-360	360;
	1277	1497	0.01864	0.08888	0	0	0	0	0	0	1	-360	360;
	1722	2039	0.01774	0.05695	0	0	0	0	0	0	1	-360	360;
	1722	7879	0.00292	0.02306	0	0	0	0	0	0	1	-360	360;
	2039	2412	0.004	0.01284	0	0	0	0	0	0	1	-360	360;
	2039	2188	0.00985	0.03164	0	0	0	0	0	0	1	-360	360;
	2039	7879	0.00336	0.02652	0	0	0	0	0	0	1	-360	360;
	2039	7879	0.00336	0.02652	0	0	0	0	0	0	1	-360	360;
	2412	5441	0.02179	0.07184	0	0	0	0	0	0	1	-360	360;
	2096	2037	0.00697	0.05491	0	0	0	0	0	0	1	-360	360;
	2096	10109	0.01119	0.05333	0	0	0	0	0	0	1	-360	360;
	2096	10109	0.01119	0.05333	0	0	0	0	0	0	1	-360	360;
	2096	52	0.04645	0.14916	0	0	0	0	0	0	1	-360	360;
	10988	1098	0.00402	0.04114	0	0	0	0	0	0	1	-360	360;
	10988	8050	0.00279	0.02851	0	0	0	0	0	0	1	-360	360;
	10988	5818	0.02688	0.12815	0	0	0	0	0	0	1	-360	360;
	1579	6257	0.03125	0.10034	0	0	0	0	0	0	1	-360	360;
	1579	2188	0.00619	0.01989	0	0	0	0	0	0	1	-360	360;
	6966	1912	0.02466	0.07919	0	0	0	0	0	0	1	-360	360;
	6966	12149	0.00563	0.01808	0	0	0	0	0	0	1	-360	360;
	12718	2567	0.00693	0.02224	0	0	0	0	0	0	1	-360	360;
	12718	2567	0.00693	0.02224	0	0	0	0	0	0	1	-360	360;
	7616	7276	0.02567	0.08245	0	0	0	0	0	0	1	-360	360;
	7616	9965	0.02944	0.09456	0	0	0	0	0	0	1	-360	360;
	1098	8050	0.0025	0.02552	0	0	0	0	0	0	1	-360	360;
	1497	1912	0.00584	0.05974	0	0	0	0	0	0	1	-360	360;
	1497	5818	0.00956	0.09776	0	0	0	0	0	0	1	-360	360;
	1497	7594	0.00563	0.01808	0	0	0	0	0	0	1	-360	360;
	484	5818	0.01044	0.04977	0	0	0	0	0	0	1	-360	360;
	484	2567	0.00495	0.03907	0	0	0	0	0	0	1	-360	360;
	1912	5818	0.00545	0.04299	0	0	0	0	0	0	1	-360	360;
	6257	12149	0.02252	0.07232	0	0	0	0	0	0	1	-360	360;
	5818	2567	0.01603	0.07644	0	0	0	0	0	0	1	-360	360;
	5818	5441	0.03924	0.12602	0	0	0	0	0	0	1	-360	360;
	5818	2824	0.01798	0.06704	0	0	0	0	0	0	1	-360	360;
	10975	12144	0	-0.0207	0	0	0	0	0	0	1	-360	360;
	215	7953	0.00149	0.01526	0	0	0	0	0	0	1	-360	360;
	215	7953	0.00099	0.01264	0	0	0	0	0	0	1	-360	360;
	215	2748	0.00168	0.01713	0	0	0	0	0	0	1	-360	360;
	215	4804	0.00131	0.01663	0	0	0	0	0	0	1	-360	360;
	7953	2748	0.00131	0.01663	0	0	0	0	0	0	1	-360	360;
	7953	12144	0.00222	0.02826	0	0	0	0	0	0	1	-360	360;
	1897	11963	0.01284	0.04122	0	0	0	0	0	0	1	-360	360;
	1897	8600	0.00563	0.01808	0	0	0	0	0	0	1	-360	360;
	11864	9428	0.01094	0.0863	0	0	0	0	0	0	1	-360	360;
	11864	1830	0.0061	0.02909	0	0	0	0	0	0	1	-360	360;
	11864	886	0.00537	0.04236	0	0	0	0	0	0	1	-360	360;
	13186	2266	0.01255	0.04032	0	0	0	0	0	0	1	-360	360;
	13186	9965	0.01092	0.03508	0	0	0	0	0	0	1	-360	360;
	13186	7457	0.01689	0.05424	0	0	0	0	0	0	1	-360	360;
	13186	11852	0.00011	0.00036	0	0	0	0	0	0	1	-360	360;
	12327	12036	0.02393	0.07684	0	0	0	0	0	0	1	-360	360;
	12327	12036	0.00856	0.06747	0	0	0	0	0	0	1	-360	360;
	12327	12036	0.00856	0.06747	0	0	0	0	0	0	1	-360	360;
	12327	2266	0.00602	0.01935	0	0	0	0	0	0	1	-360	360;
	12327	9965	0.0058	0.02763	0	0	0	0	0	0	1	-360	360;
	12327	9965	0.0058	0.02763	0	0	0	0	0	0	1	-360	360;
	12036	12848	0.02978	0.09564	0	0	0	0	0	0	1	-360	360;
	12036	12848	0.01194	0.09414	0	0	0	0	0	0	1	-360	360;
	12036	12848	0.01194	0.09414	0	0	0	0	0	0	1	-360	360;
	11963	8514	0.00796	0.06276	0	0	0	0	0	0	1	-360	360;
	11963	3762	0.00398	0.03138	0	0	0	0	0	0	1	-360	360;
	11963	11991	0.00458	0.03609	0	0	0	0	0	0	1	-360	360;
	10568	2266	0.00298	0.00958	0	0	0	0	0	0	1	-360	360;
	12848	52	0.03001	0.09637	0	0	0	0	0	0	1	-360	360;
	10274	1830	0.00983	0.04686	0	0	0	0	0	0	1	-360	360;
	10274	1830	0.00796	0.06276	0	0	0	0	0	0	1	-360	360;
	2266	9965	0.00462	0.01483	0	0	0	0	0	0	1	-360	360;
	2266	1717	0.001	0.00785	0	0	0	0	0	0	1	-360	360;
	2266	1717	0.001	0.00785	0	0	0	0	0	0	1	-360	360;
	9428	8514	0.01239	0.03978	0	0	0	0	0	0	1	-360	360;
	9428	8514	0.00749	0.03571	0	0	0	0	0	0	1	-360	360;
	9428	8514	0.00749	0.03571	0	0	0	0	0	0	1	-360	360;
	9428	886	0.00478	0.03766	0	0	0	0	0	0	1	-360	360;
	9428	11991	0.0038	0.02997	0	0	0	0	0	0	1	-360	360;
	9428	519	0.00394	0.01266	0	0	0	0	0	0	1	-360	360;
	8514	1830	0.0199	0.1569	0	0	0	0	0	0	1	-360	360;
	8514	11389	0.00677	0.05335	0	0	0	0	0	0	1	-360	360;
	9965	8600	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	9965	8600	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	8600	3762	0.00458	0.03609	0	0	0	0	0	0	1	-360	360;
	10490	11742	0.00133	0.01349	0	0	0	0	0	0	1	-360	360;
	10490	10018	0.00047	0.00772	0	0	0	0	0	0	1	-360	360;
	10490	6028	0.00698	0.07071	0	0	0	0	0	0	1	-360	360;
	11742	10018	0.00048	0.00615	0	0	0	0	0	0	1	-360	360;
	11742	8624	0.00048	0.00615	0	0	0	0	0	0	1	-360	360;
	11742	10266	0.0014	0.01429	0	0	0	0	0	0	1	-360	360;
	11742	4116	0.00156	0.01982	0	0	0	0	0	0	1	-360	360;
	1952	9639	0.02032	0.06527	0	0	0	0	0	0	1	-360	360;
	1952	9528	0.04842	0.15549	0	0	0	0	0	0	1	-360	360;
	1952	5853	0.00957	0.03074	0	0	0	0	0	0	1	-360	360;
	1952	7439	0.00729	0.03474	0	0	0	0	0	0	1	-360	360;
	9907	9528	0.0198	0.15612	0	0	0	0	0	0	1	-360	360;
	9907	6270	0.07263	0.23323	0	0	0	0	0	0	1	-360	360;
	9907	5313	0.01194	0.09414	0	0	0	0	0	0	1	-360	360;
	9907	5313	0.01194	0.09414	0	0	0	0	0	0	1	-360	360;
	10511	1569	0.00902	0.04299	0	0	0	0	0	0	1	-360	360;
	10511	1569	0.00902	0.04299	0	0	0	0	0	0	1	-360	360;
	10511	3200	0.00203	0.0097	0	0	0	0	0	0	1	-360	360;
	1569	12620	0.01407	0.0452	0	0	0	0	0	0	1	-360	360;
	1569	10179	0.01149	0.05478	0	0	0	0	0	0	1	-360	360;
	1119	2344	0.01396	0.04484	0	0	0	0	0	0	1	-360	360;
	1119	2344	0.01396	0.04484	0	0	0	0	0	0	1	-360	360;
	2589	2765	0.00712	0.03394	0	0	0	0	0	0	1	-360	360;
	2589	2765	0.00712	0.03394	0	0	0	0	0	0	1	-360	360;
	2589	2344	0.02421	0.07774	0	0	0	0	0	0	1	-360	360;
	2589	10807	0.02083	0.0669	0	0	0	0	0	0	1	-360	360;
	2765	13534	0.02034	0.09696	0	0	0	0	0	0	1	-360	360;
	2765	452	0.01526	0.07272	0	0	0	0	0	0	1	-360	360;
	2765	9684	0.02949	0.14059	0	0	0	0	0	0	1	-360	360;
	2765	4017	0.01095	0.0522	0	0	0	0	0	0	1	-360	360;
	2765	4017	0.01095	0.0522	0	0	0	0	0	0	1	-360	360;
	2765	11062	0.01763	0.08403	0	0	0	0	0	0	1	-360	360;
	2765	8165	0.00814	0.03878	0	0	0	0	0	0	1	-360	360;
	10781	10820	0.01898	0.0905	0	0	0	0	0	0	1	-360	360;
	10781	10820	0.01898	0.0905	0	0	0	0	0	0	1	-360	360;
	3024	12103	0.00394	0.01266	0	0	0	0	0	0	1	-360	360;
	3024	9684	0.00576	0.02747	0	0	0	0	0	0	1	-360	360;
	3024	12067	0.01436	0.0461	0	0	0	0	0	0	1	-360	360;
	3024	7829	0.02534	0.08136	0	0	0	0	0	0	1	-360	360;
	12103	11148	0.02646	0.08498	0	0	0	0	0	0	1	-360	360;
	12103	9684	0.0059	0.02812	0	0	0	0	0	0	1	-360	360;
	12103	2344	0.02984	0.09582	0	0	0	0	0	0	1	-360	360;
	12103	8321	0.00796	0.06276	0	0	0	0	0	0	1	-360	360;
	12103	8165	0.01729	0.08242	0	0	0	0	0	0	1	-360	360;
	6528	7827	0.00569	0.02715	0	0	0	0	0	0	1	-360	360;
	6528	11160	0.01543	0.07353	0	0	0	0	0	0	1	-360	360;
	6528	11160	0.01543	0.07353	0	0	0	0	0	0	1	-360	360;
	6528	9684	0.00529	0.02521	0	0	0	0	0	0	1	-360	360;
	6528	9684	0.00529	0.02521	0	0	0	0	0	0	1	-360	360;
	6528	10955	0.00024	0.00113	0	0	0	0	0	0	1	-360	360;
	6528	10955	0.00024	0.00113	0	0	0	0	0	0	1	-360	360;
	6528	12957	0.00959	0.04573	0	0	0	0	0	0	1	-360	360;
	6528	13168	0.02252	0.07232	0	0	0	0	0	0	1	-360	360;
	6528	12935	0.01689	0.05424	0	0	0	0	0	0	1	-360	360;
	10457	9528	0.03125	0.10034	0	0	0	0	0	0	1	-360	360;
	10457	5853	0.0078	0.03717	0	0	0	0	0	0	1	-360	360;
	10457	5853	0.0078	0.03717	0	0	0	0	0	0	1	-360	360;
	10457	10582	0.00678	0.03232	0	0	0	0	0	0	1	-360	360;
	10457	10582	0.00678	0.03232	0	0	0	0	0	0	1	-360	360;
	10457	7439	0.01295	0.06173	0	0	0	0	0	0	1	-360	360;
	6042	10179	0.012	0.05721	0	0	0	0	0	0	1	-360	360;
	6042	10179	0.012	0.05721	0	0	0	0	0	0	1	-360	360;
	6042	2344	0.00332	0.01584	0	0	0	0	0	0	1	-360	360;
	6042	2344	0.00332	0.01584	0	0	0	0	0	0	1	-360	360;
	6042	11884	0.00986	0.04703	0	0	0	0	0	0	1	-360	360;
	6042	11884	0.00986	0.04703	0	0	0	0	0	0	1	-360	360;
	4124	11884	0.0002	0.00204	0	0	0	0	0	0	1	-360	360;
	4124	11884	0.0002	0.00204	0	0	0	0	0	0	1	-360	360;
	13534	7827	0.02081	0.09922	0	0	0	0	0	0	1	-360	360;
	13534	7827	0.02081	0.09922	0	0	0	0	0	0	1	-360	360;
	13534	11884	0.01672	0.0537	0	0	0	0	0	0	1	-360	360;
	13534	12067	0.01171	0.03761	0	0	0	0	0	0	1	-360	360;
	13534	11062	0.00407	0.01939	0	0	0	0	0	0	1	-360	360;
	7827	12957	0.00959	0.04573	0	0	0	0	0	0	1	-360	360;
	7827	3511	0.00113	0.00362	0	0	0	0	0	0	1	-360	360;
	12620	11884	0.00296	0.02338	0	0	0	0	0	0	1	-360	360;
	10179	9676	0.01366	0.06513	0	0	0	0	0	0	1	-360	360;
	10179	11884	0.00093	0.0095	0	0	0	0	0	0	1	-360	360;
	10179	1989	0.00034	0.00162	0	0	0	0	0	0	1	-360	360;
	9676	3200	0.01786	0.08516	0	0	0	0	0	0	1	-360	360;
	7292	9639	0.0045	0.01446	0	0	0	0	0	0	1	-360	360;
	7292	2950	0.0116	0.03725	0	0	0	0	0	0	1	-360	360;
	7292	9456	0.02629	0.08443	0	0	0	0	0	0	1	-360	360;
	9639	2950	0.00709	0.02278	0	0	0	0	0	0	1	-360	360;
	452	11148	0.01051	0.0501	0	0	0	0	0	0	1	-360	360;
	11160	713	0.02353	0.07557	0	0	0	0	0	0	1	-360	360;
	11160	8321	0.00799	0.02567	0	0	0	0	0	0	1	-360	360;
	11160	2950	0.02168	0.06961	0	0	0	0	0	0	1	-360	360;
	11160	8983	0.00608	0.01953	0	0	0	0	0	0	1	-360	360;
	11160	787	0.00169	0.00808	0	0	0	0	0	0	1	-360	360;
	11160	787	0.00169	0.00808	0	0	0	0	0	0	1	-360	360;
	1949	2950	0.15283	0.86416	0	0	0	0	0	0	1	-360	360;
	1949	2762	0.0045	0.01446	0	0	0	0	0	0	1	-360	360;
	1949	8876	0.00336	0.016	0	0	0	0	0	0	1	-360	360;
	10820	2950	0.00125	0.00989	0	0	0	0	0	0	1	-360	360;
	10820	7723	0.00153	0.01208	0	0	0	0	0	0	1	-360	360;
	10820	787	0.01265	0.06028	0	0	0	0	0	0	1	-360	360;
	10820	787	0.01265	0.06028	0	0	0	0	0	0	1	-360	360;
	10820	7439	0.00576	0.02747	0	0	0	0	0	0	1	-360	360;
	10820	7439	0.00576	0.02747	0	0	0	0	0	0	1	-360	360;
	2344	6532	0.00056	0.00181	0	0	0	0	0	0	1	-360	360;
	11884	11999	0.00225	0.00723	0	0	0	0	0	0	1	-360	360;
	11884	1204	0.0006	0.00471	0	0	0	0	0	0	1	-360	360;
	8321	8983	0.01413	0.04538	0	0	0	0	0	0	1	-360	360;
	8321	7829	0.00338	0.01085	0	0	0	0	0	0	1	-360	360;
	4017	8584	0.01065	0.05074	0	0	0	0	0	0	1	-360	360;
	4017	8584	0.01065	0.05074	0	0	0	0	0	0	1	-360	360;
	4017	12098	0.00506	0.03985	0	0	0	0	0	0	1	-360	360;
	2950	8983	0.01368	0.04393	0	0	0	0	0	0	1	-360	360;
	9528	7883	0.00034	0.00108	0	0	0	0	0	0	1	-360	360;
	9528	6270	0.05067	0.16272	0	0	0	0	0	0	1	-360	360;
	9456	3573	0.03125	0.10034	0	0	0	0	0	0	1	-360	360;
	7723	2762	0.00113	0.00362	0	0	0	0	0	0	1	-360	360;
	12067	12957	0.01293	0.10199	0	0	0	0	0	0	1	-360	360;
	12067	12957	0.01293	0.10199	0	0	0	0	0	0	1	-360	360;
	12067	8485	0.00106	0.01086	0	0	0	0	0	0	1	-360	360;
	12067	8485	0.00106	0.01086	0	0	0	0	0	0	1	-360	360;
	8983	11766	0.00139	0.01098	0	0	0	0	0	0	1	-360	360;
	8983	11766	0.00139	0.01098	0	0	0	0	0	0	1	-360	360;
	10018	2752	0.0017	0.02161	0	0	0	0	0	0	1	-360	360;
	3200	8140	0.00895	0.07061	0	0	0	0	0	0	1	-360	360;
	3200	8140	0.00895	0.07061	0	0	0	0	0	0	1	-360	360;
	5853	7439	0.00695	0.03313	0	0	0	0	0	0	1	-360	360;
	5853	7439	0.00746	0.03555	0	0	0	0	0	0	1	-360	360;
	5853	8876	0.00929	0.04428	0	0	0	0	0	0	1	-360	360;
	191	8624	0.00084	0.01064	0	0	0	0	0	0	1	-360	360;
	191	4116	0.00274	0.03491	0	0	0	0	0	0	1	-360	360;
	2477	7209	0.00153	0.01547	0	0	0	0	0	0	1	-360	360;
	9769	3045	0.00328	0.03317	0	0	0	0	0	0	1	-360	360;
	9769	13258	0.00137	0.01387	0	0	0	0	0	0	1	-360	360;
	3045	7209	0.00066	0.00666	0	0	0	0	0	0	1	-360	360;
	3045	12844	0.0053	0.05369	0	0	0	0	0	0	1	-360	360;
	3045	7647	0.0053	0.05369	0	0	0	0	0	0	1	-360	360;
	7209	7769	0.00025	0.00417	0	0	0	0	0	0	1	-360	360;
	7209	3317	0.00037	0.00465	0	0	0	0	0	0	1	-360	360;
	7209	8817	0.00347	0.04417	0	0	0	0	0	0	1	-360	360;
	7209	6223	0.00347	0.04417	0	0	0	0	0	0	1	-360	360;
	7769	10487	0.0017	0.02819	0	0	0	0	0	0	1	-360	360;
	7769	3066	0.0017	0.02819	0	0	0	0	0	0	1	-360	360;
	12424	2928	0.00049	0.00496	0	0	0	0	0	0	1	-360	360;
	12424	9803	0.00139	0.01098	0	0	0	0	0	0	1	-360	360;
	12424	8705	0.00031	0.00329	0	0	0	0	0	0	1	-360	360;
	12424	9529	0.00424	0.0202	0	0	0	0	0	0	1	-360	360;
	12424	9529	0.00424	0.0202	0	0	0	0	0	0	1	-360	360;
	12424	5233	0.01197	0.05705	0	0	0	0	0	0	1	-360	360;
	12424	7763	0.00052	0.00561	0	0	0	0	0	0	1	-360	360;
	5469	9803	0.00199	0.01569	0	0	0	0	0	0	1	-360	360;
	5469	9803	0.00139	0.01098	0	0	0	0	0	0	1	-360	360;
	5469	8808	0.00255	0.02008	0	0	0	0	0	0	1	-360	360;
	5469	8808	0.00777	0.02495	0	0	0	0	0	0	1	-360	360;
	5469	8808	0.00255	0.02008	0	0	0	0	0	0	1	-360	360;
	5469	8808	0.00777	0.02495	0	0	0	0	0	0	1	-360	360;
	2928	9803	0.00475	0.02262	0	0	0	0	0	0	1	-360	360;
	2928	8303	0.00031	0.00329	0	0	0	0	0	0	1	-360	360;
	2928	5233	0.0101	0.04816	0	0	0	0	0	0	1	-360	360;
	3964	9529	0.00929	0.04428	0	0	0	0	0	0	1	-360	360;
	3964	13457	0.01119	0.05333	0	0	0	0	0	0	1	-360	360;
	6244	6781	0.00468	0.03687	0	0	0	0	0	0	1	-360	360;
	6244	12868	0.00225	0.00723	0	0	0	0	0	0	1	-360	360;
	6244	12967	0.05517	0.17718	0	0	0	0	0	0	1	-360	360;
	13044	13	0.01616	0.05189	0	0	0	0	0	0	1	-360	360;
	13044	5233	0.03603	0.11571	0	0	0	0	0	0	1	-360	360;
	13184	3562	0.00263	0.02071	0	0	0	0	0	0	1	-360	360;
	13184	629	0.00508	0.02424	0	0	0	0	0	0	1	-360	360;
	13	6730	0.00545	0.04299	0	0	0	0	0	0	1	-360	360;
	13297	3562	0.01505	0.07175	0	0	0	0	0	0	1	-360	360;
	13297	3562	0.01505	0.07175	0	0	0	0	0	0	1	-360	360;
	11188	1746	0.01431	0.0682	0	0	0	0	0	0	1	-360	360;
	11188	3315	0.00267	0.02103	0	0	0	0	0	0	1	-360	360;
	11188	3315	0.00267	0.02103	0	0	0	0	0	0	1	-360	360;
	3317	7972	0.00346	0.04406	0	0	0	0	0	0	1	-360	360;
	3562	629	0.00332	0.0262	0	0	0	0	0	0	1	-360	360;
	3562	8166	0.00895	0.07061	0	0	0	0	0	0	1	-360	360;
	3562	6647	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	2704	9369	0.00079	0.00253	0	0	0	0	0	0	1	-360	360;
	2704	11356	0.00451	0.02149	0	0	0	0	0	0	1	-360	360;
	2704	10703	0.01458	0.06949	0	0	0	0	0	0	1	-360	360;
	1746	11356	0.02044	0.06563	0	0	0	0	0	0	1	-360	360;
	1746	1174	0.03316	0.10649	0	0	0	0	0	0	1	-360	360;
	11356	13457	0.00441	0.02101	0	0	0	0	0	0	1	-360	360;
	11356	13457	0.00441	0.02101	0	0	0	0	0	0	1	-360	360;
	11356	9180	0.02373	0.11312	0	0	0	0	0	0	1	-360	360;
	9803	7893	0.00251	0.01977	0	0	0	0	0	0	1	-360	360;
	9803	3315	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	9529	8169	0.00607	0.02893	0	0	0	0	0	0	1	-360	360;
	9529	7763	0.0003	0.0032	0	0	0	0	0	0	1	-360	360;
	7893	3315	0.00106	0.00832	0	0	0	0	0	0	1	-360	360;
	6781	6958	0.00762	0.06009	0	0	0	0	0	0	1	-360	360;
	8808	11843	0.00524	0.03214	0	0	0	0	0	0	1	-360	360;
	8808	11843	0.00524	0.03214	0	0	0	0	0	0	1	-360	360;
	8808	5533	0.00068	0.00323	0	0	0	0	0	0	1	-360	360;
	9328	5233	0.00526	0.02505	0	0	0	0	0	0	1	-360	360;
	9328	5233	0.00526	0.02505	0	0	0	0	0	0	1	-360	360;
	9328	6958	0.00345	0.0353	0	0	0	0	0	0	1	-360	360;
	9328	6958	0.00345	0.0353	0	0	0	0	0	0	1	-360	360;
	9328	9495	0.00122	0.01306	0	0	0	0	0	0	1	-360	360;
	9328	9495	0.00122	0.01306	0	0	0	0	0	0	1	-360	360;
	11843	3315	0.00796	0.06276	0	0	0	0	0	0	1	-360	360;
	8166	6647	0.00677	0.05335	0	0	0	0	0	0	1	-360	360;
	3315	9990	0.00239	0.01883	0	0	0	0	0	0	1	-360	360;
	3315	12623	0.00239	0.01883	0	0	0	0	0	0	1	-360	360;
	12868	13450	0.01407	0.0452	0	0	0	0	0	0	1	-360	360;
	13457	8169	0.00763	0.03636	0	0	0	0	0	0	1	-360	360;
	9990	7763	0.00022	0.0024	0	0	0	0	0	0	1	-360	360;
	12623	7763	0.00022	0.0024	0	0	0	0	0	0	1	-360	360;
	12096	1652	0.00282	0.03584	0	0	0	0	0	0	1	-360	360;
	12096	11647	0	-0.0228	0	0	0	0	0	0	1	-360	360;
	7972	11647	0	-0.0228	0	0	0	0	0	0	1	-360	360;
	12230	11647	0	-0.0228	0	0	0	0	0	0	1	-360	360;
	12230	6634	0.00219	0.02793	0	0	0	0	0	0	1	-360	360;
	11647	8026	0.00198	0.02025	0	0	0	0	0	0	1	-360	360;
	11647	2129	0.00248	0.02534	0	0	0	0	0	0	1	-360	360;
	6718	639	0.02365	0.07594	0	0	0	0	0	0	1	-360	360;
	6718	8363	0.0532	0.17086	0	0	0	0	0	0	1	-360	360;
	4271	11833	0.02252	0.07232	0	0	0	0	0	0	1	-360	360;
	4271	9252	0.02252	0.07232	0	0	0	0	0	0	1	-360	360;
	10880	12927	0.02624	0.08425	0	0	0	0	0	0	1	-360	360;
	10880	10777	0.00951	0.03056	0	0	0	0	0	0	1	-360	360;
	10880	20	0.01802	0.05786	0	0	0	0	0	0	1	-360	360;
	13438	12154	0.00845	0.02712	0	0	0	0	0	0	1	-360	360;
	13438	908	0.02736	0.08787	0	0	0	0	0	0	1	-360	360;
	1601	9462	0.01393	0.10983	0	0	0	0	0	0	1	-360	360;
	1601	5985	0.0131	0.08035	0	0	0	0	0	0	1	-360	360;
	1601	9800	0.01847	0.0593	0	0	0	0	0	0	1	-360	360;
	1174	2117	0.06942	0.22293	0	0	0	0	0	0	1	-360	360;
	1174	12608	0.0183	0.05876	0	0	0	0	0	0	1	-360	360;
	1174	11890	0.04065	0.13054	0	0	0	0	0	0	1	-360	360;
	2117	11833	0.01098	0.03526	0	0	0	0	0	0	1	-360	360;
	2117	12154	0.03665	0.1177	0	0	0	0	0	0	1	-360	360;
	2117	908	0.03001	0.09637	0	0	0	0	0	0	1	-360	360;
	2117	10549	0.00478	0.03766	0	0	0	0	0	0	1	-360	360;
	2117	10549	0.00478	0.03766	0	0	0	0	0	0	1	-360	360;
	11833	12154	0.01592	0.12552	0	0	0	0	0	0	1	-360	360;
	12927	8449	0.04189	0.13452	0	0	0	0	0	0	1	-360	360;
	12154	5458	0.02049	0.06581	0	0	0	0	0	0	1	-360	360;
	10777	292	0.06227	0.19997	0	0	0	0	0	0	1	-360	360;
	12608	4500	0.03806	0.12222	0	0	0	0	0	0	1	-360	360;
	12608	2910	0.01393	0.10983	0	0	0	0	0	0	1	-360	360;
	7799	639	0.01671	0.07967	0	0	0	0	0	0	1	-360	360;
	7799	4930	0.02712	0.12928	0	0	0	0	0	0	1	-360	360;
	639	6396	0.03361	0.10794	0	0	0	0	0	0	1	-360	360;
	9462	9252	0.01204	0.05737	0	0	0	0	0	0	1	-360	360;
	20	5458	0.02522	0.081	0	0	0	0	0	0	1	-360	360;
	20	8363	0.00591	0.01898	0	0	0	0	0	0	1	-360	360;
	20	6164	0.04375	0.14048	0	0	0	0	0	0	1	-360	360;
	6479	10609	0.01171	0.03761	0	0	0	0	0	0	1	-360	360;
	6479	8877	0.03389	0.10884	0	0	0	0	0	0	1	-360	360;
	10609	9252	0.01019	0.03272	0	0	0	0	0	0	1	-360	360;
	5985	3033	0.0131	0.08035	0	0	0	0	0	0	1	-360	360;
	6874	10487	0	-0.03116	0	0	0	0	0	0	1	-360	360;
	6874	13337	0.00123	0.02042	0	0	0	0	0	0	1	-360	360;
	13231	3066	0	-0.03116	0	0	0	0	0	0	1	-360	360;
	13231	13337	0.00123	0.02042	0	0	0	0	0	0	1	-360	360;
	2202	12486	0	-0.0228	0	0	0	0	0	0	1	-360	360;
	2202	8817	0	-0.0228	0	0	0	0	0	0	1	-360	360;
	1618	6223	0	-0.023	0	0	0	0	0	0	1	-360	360;
	1618	9905	0	-0.023	0	0	0	0	0	0	1	-360	360;
	12486	1036	0.00262	0.03338	0	0	0	0	0	0	1	-360	360;
	9905	10726	0.00266	0.03382	0	0	0	0	0	0	1	-360	360;
	12844	6395	0	-0.0242	0	0	0	0	0	0	1	-360	360;
	7647	6395	0	-0.0242	0	0	0	0	0	0	1	-360	360;
	6395	1036	0.00287	0.02938	0	0	0	0	0	0	1	-360	360;
	6395	8751	0.00544	0.05508	0	0	0	0	0	0	1	-360	360;
	6395	8751	0.00544	0.05508	0	0	0	0	0	0	1	-360	360;
	13337	1365	0.00139	0.02301	0	0	0	0	0	0	1	-360	360;
	13337	7644	0.00139	0.02301	0	0	0	0	0	0	1	-360	360;
	13337	4417	0.00099	0.01644	0	0	0	0	0	0	1	-360	360;
	13337	10351	0.00294	0.03741	0	0	0	0	0	0	1	-360	360;
	9583	3033	0.02252	0.07232	0	0	0	0	0	0	1	-360	360;
	9583	9800	0.0174	0.05587	0	0	0	0	0	0	1	-360	360;
	9583	6396	0.02509	0.11958	0	0	0	0	0	0	1	-360	360;
	9583	9564	0.00902	0.04299	0	0	0	0	0	0	1	-360	360;
	9583	9564	0.00902	0.04299	0	0	0	0	0	0	1	-360	360;
	9583	598	0.04087	0.13126	0	0	0	0	0	0	1	-360	360;
	9583	10528	0.02196	0.07051	0	0	0	0	0	0	1	-360	360;
	3573	11079	0.01582	0.0508	0	0	0	0	0	0	1	-360	360;
	3573	972	0.04712	0.15133	0	0	0	0	0	0	1	-360	360;
	3033	231	0.02055	0.06599	0	0	0	0	0	0	1	-360	360;
	3033	6901	0.00339	0.01616	0	0	0	0	0	0	1	-360	360;
	3033	6901	0.00339	0.01616	0	0	0	0	0	0	1	-360	360;
	3033	1703	0.01098	0.03526	0	0	0	0	0	0	1	-360	360;
	5684	8877	0.03271	0.10504	0	0	0	0	0	0	1	-360	360;
	5684	5851	0.02145	0.06888	0	0	0	0	0	0	1	-360	360;
	5684	10703	0.00644	0.0307	0	0	0	0	0	0	1	-360	360;
	5684	9180	0.00644	0.0307	0	0	0	0	0	0	1	-360	360;
	5684	2311	0.03029	0.09727	0	0	0	0	0	0	1	-360	360;
	6396	231	0.00255	0.02008	0	0	0	0	0	0	1	-360	360;
	6396	231	0.00434	0.02069	0	0	0	0	0	0	1	-360	360;
	6396	231	0.00434	0.02069	0	0	0	0	0	0	1	-360	360;
	4249	231	0.01633	0.05243	0	0	0	0	0	0	1	-360	360;
	231	6901	0.00729	0.03474	0	0	0	0	0	0	1	-360	360;
	231	6901	0.00729	0.03474	0	0	0	0	0	0	1	-360	360;
	6901	1942	0.00034	0.00162	0	0	0	0	0	0	1	-360	360;
	6901	1942	0.00034	0.00162	0	0	0	0	0	0	1	-360	360;
	12215	10703	0.00085	0.00404	0	0	0	0	0	0	1	-360	360;
	12215	9180	0.00085	0.00404	0	0	0	0	0	0	1	-360	360;
	8877	598	0.00501	0.01609	0	0	0	0	0	0	1	-360	360;
	1169	11079	0.02015	0.06473	0	0	0	0	0	0	1	-360	360;
	1169	10528	0.01407	0.0452	0	0	0	0	0	0	1	-360	360;
	11079	7710	0.00935	0.07374	0	0	0	0	0	0	1	-360	360;
	8507	1703	0.01042	0.03345	0	0	0	0	0	0	1	-360	360;
	8507	7710	0.01478	0.07046	0	0	0	0	0	0	1	-360	360;
	5851	972	0.04572	0.14681	0	0	0	0	0	0	1	-360	360;
	7710	11683	0.01029	0.08112	0	0	0	0	0	0	1	-360	360;
	7710	11683	0.01029	0.08112	0	0	0	0	0	0	1	-360	360;
	7710	10528	0.00836	0.0659	0	0	0	0	0	0	1	-360	360;
	6028	2748	0	-0.0207	0	0	0	0	0	0	1	-360	360;
	2748	2162	0.00193	0.01978	0	0	0	0	0	0	1	-360	360;
	2748	8641	0.00224	0.02269	0	0	0	0	0	0	1	-360	360;
	11389	12569	0.01071	0.08441	0	0	0	0	0	0	1	-360	360;
	3000	2644	0.02733	0.09705	0	0	0	0	0	0	1	-360	360;
	3000	2311	0.03029	0.09727	0	0	0	0	0	0	1	-360	360;
	13299	11006	0.02308	0.07413	0	0	0	0	0	0	1	-360	360;
	13299	2644	0.02305	0.08331	0	0	0	0	0	0	1	-360	360;
	7043	1622	0.03288	0.10559	0	0	0	0	0	0	1	-360	360;
	7043	8728	0.0077	0.04725	0	0	0	0	0	0	1	-360	360;
	7043	11359	0.00796	0.06276	0	0	0	0	0	0	1	-360	360;
	7043	13587	0.00648	0.02079	0	0	0	0	0	0	1	-360	360;
	6270	9629	0.00895	0.07061	0	0	0	0	0	0	1	-360	360;
	6270	13050	0.01407	0.0452	0	0	0	0	0	0	1	-360	360;
	1622	8411	0.00816	0.02622	0	0	0	0	0	0	1	-360	360;
	1622	9629	0.00959	0.07563	0	0	0	0	0	0	1	-360	360;
	1622	10441	0.01017	0.04848	0	0	0	0	0	0	1	-360	360;
	5016	3784	0.01786	0.08516	0	0	0	0	0	0	1	-360	360;
	5016	8728	0.01926	0.11812	0	0	0	0	0	0	1	-360	360;
	4866	1221	0.00239	0.01883	0	0	0	0	0	0	1	-360	360;
	4866	1221	0.00239	0.01883	0	0	0	0	0	0	1	-360	360;
	4866	12569	0.04076	0.1309	0	0	0	0	0	0	1	-360	360;
	4866	8728	0.00814	0.06417	0	0	0	0	0	0	1	-360	360;
	1221	11006	0.00776	0.06119	0	0	0	0	0	0	1	-360	360;
	11006	8728	0.00665	0.04082	0	0	0	0	0	0	1	-360	360;
	11006	8728	0.00665	0.04082	0	0	0	0	0	0	1	-360	360;
	3784	8411	0.04943	0.15874	0	0	0	0	0	0	1	-360	360;
	12569	7650	0.02196	0.07051	0	0	0	0	0	0	1	-360	360;
	12569	13587	0.04155	0.13343	0	0	0	0	0	0	1	-360	360;
	2162	10351	0	-0.01731	0	0	0	0	0	0	1	-360	360;
	8728	3387	0.01127	0.0691	0	0	0	0	0	0	1	-360	360;
	7650	10125	0.04279	0.13741	0	0	0	0	0	0	1	-360	360;
	8375	13587	0.00113	0.00362	0	0	0	0	0	0	1	-360	360;
	8181	8930	0.00108	0.00847	0	0	0	0	0	0	1	-360	360;
	8181	3554	0.05089	0.16344	0	0	0	0	0	0	1	-360	360;
	8930	5580	0.05315	0.17067	0	0	0	0	0	0	1	-360	360;
	8930	10199	0.00995	0.07845	0	0	0	0	0	0	1	-360	360;
	10125	5580	0.03406	0.10938	0	0	0	0	0	0	1	-360	360;
	10125	13515	0.00966	0.04606	0	0	0	0	0	0	1	-360	360;
	10125	11173	0.03429	0.11011	0	0	0	0	0	0	1	-360	360;
	3554	8160	0.01858	0.05966	0	0	0	0	0	0	1	-360	360;
	3554	9675	0.00466	0.03672	0	0	0	0	0	0	1	-360	360;
	2927	81	0.01485	0.07078	0	0	0	0	0	0	1	-360	360;
	2927	4474	0.01509	0.04845	0	0	0	0	0	0	1	-360	360;
	2927	7119	0.01824	0.05858	0	0	0	0	0	0	1	-360	360;
	2927	7457	0.01464	0.04701	0	0	0	0	0	0	1	-360	360;
	81	11876	0.00432	0.03405	0	0	0	0	0	0	1	-360	360;
	81	10716	0.01183	0.0564	0	0	0	0	0	0	1	-360	360;
	5580	12967	0.03367	0.10812	0	0	0	0	0	0	1	-360	360;
	5580	10199	0.04712	0.15133	0	0	0	0	0	0	1	-360	360;
	8160	7119	0.0165	0.05297	0	0	0	0	0	0	1	-360	360;
	8160	2207	0.02204	0.10504	0	0	0	0	0	0	1	-360	360;
	11876	9675	0.00219	0.01726	0	0	0	0	0	0	1	-360	360;
	11876	1698	0.00225	0.00723	0	0	0	0	0	0	1	-360	360;
	4474	8539	0.00715	0.02296	0	0	0	0	0	0	1	-360	360;
	9675	10199	0.01791	0.14121	0	0	0	0	0	0	1	-360	360;
	13515	2207	0.01356	0.06464	0	0	0	0	0	0	1	-360	360;
	9367	3888	0.01768	0.05677	0	0	0	0	0	0	1	-360	360;
	9367	4335	0.0103	0.03309	0	0	0	0	0	0	1	-360	360;
	8026	2129	0.00124	0.01579	0	0	0	0	0	0	1	-360	360;
	2129	4820	0.00037	0.00374	0	0	0	0	0	0	1	-360	360;
	2129	4820	0.00037	0.00374	0	0	0	0	0	0	1	-360	360;
	2129	5088	0.00207	0.03419	0	0	0	0	0	0	1	-360	360;
	9927	3888	0.00946	0.04509	0	0	0	0	0	0	1	-360	360;
	9927	5820	0.00576	0.02747	0	0	0	0	0	0	1	-360	360;
	7248	10369	0.02387	0.07666	0	0	0	0	0	0	1	-360	360;
	7248	7578	0.04842	0.15549	0	0	0	0	0	0	1	-360	360;
	3888	5820	0.00508	0.02424	0	0	0	0	0	0	1	-360	360;
	3888	10369	0.05444	0.17483	0	0	0	0	0	0	1	-360	360;
	9071	12157	0.0326	0.10468	0	0	0	0	0	0	1	-360	360;
	9071	9300	0.02269	0.07286	0	0	0	0	0	0	1	-360	360;
	11535	3624	0.01831	0.08726	0	0	0	0	0	0	1	-360	360;
	11535	6444	0.04183	0.13433	0	0	0	0	0	0	1	-360	360;
	8449	7697	0.00513	0.04048	0	0	0	0	0	0	1	-360	360;
	8449	767	0.00169	0.00542	0	0	0	0	0	0	1	-360	360;
	8449	13178	0.00541	0.04268	0	0	0	0	0	0	1	-360	360;
	8449	3565	0.00396	0.03122	0	0	0	0	0	0	1	-360	360;
	8449	168	0.02288	0.10908	0	0	0	0	0	0	1	-360	360;
	8449	9854	0.02881	0.13736	0	0	0	0	0	0	1	-360	360;
	11964	2910	0.03941	0.12656	0	0	0	0	0	0	1	-360	360;
	11964	9300	0.01396	0.04484	0	0	0	0	0	0	1	-360	360;
	12157	4996	0.03969	0.12746	0	0	0	0	0	0	1	-360	360;
	4	3507	0.00034	0.00162	0	0	0	0	0	0	1	-360	360;
	4	7182	0.00051	0.00242	0	0	0	0	0	0	1	-360	360;
	2860	6382	0.02145	0.06888	0	0	0	0	0	0	1	-360	360;
	2860	1461	0.00518	0.01663	0	0	0	0	0	0	1	-360	360;
	2860	3507	0.00085	0.00404	0	0	0	0	0	0	1	-360	360;
	2860	7182	0.00052	0.00246	0	0	0	0	0	0	1	-360	360;
	2860	5894	0.00555	0.04377	0	0	0	0	0	0	1	-360	360;
	3235	4500	0.02252	0.07232	0	0	0	0	0	0	1	-360	360;
	3235	12564	0.0165	0.05297	0	0	0	0	0	0	1	-360	360;
	4500	12356	0.02398	0.07702	0	0	0	0	0	0	1	-360	360;
	12356	2910	0.00822	0.0264	0	0	0	0	0	0	1	-360	360;
	11612	908	0.04954	0.1591	0	0	0	0	0	0	1	-360	360;
	11612	11890	0.0237	0.07612	0	0	0	0	0	0	1	-360	360;
	5558	3624	0.02117	0.06798	0	0	0	0	0	0	1	-360	360;
	5558	11890	0.02809	0.09022	0	0	0	0	0	0	1	-360	360;
	5558	3459	0.01126	0.03616	0	0	0	0	0	0	1	-360	360;
	4996	3459	0.01278	0.04104	0	0	0	0	0	0	1	-360	360;
	4996	7578	0.0036	0.01157	0	0	0	0	0	0	1	-360	360;
	11960	908	0.02871	0.09221	0	0	0	0	0	0	1	-360	360;
	11960	4335	0.03384	0.10866	0	0	0	0	0	0	1	-360	360;
	908	4335	0.04971	0.15965	0	0	0	0	0	0	1	-360	360;
	908	6444	0.01576	0.05062	0	0	0	0	0	0	1	-360	360;
	4335	7697	0.00306	0.02416	0	0	0	0	0	0	1	-360	360;
	4335	13178	0.00346	0.0273	0	0	0	0	0	0	1	-360	360;
	4335	3565	0.00522	0.02489	0	0	0	0	0	0	1	-360	360;
	7697	13178	0.00048	0.00226	0	0	0	0	0	0	1	-360	360;
	7697	12262	0.00048	0.00226	0	0	0	0	0	0	1	-360	360;
	9300	6382	0.00766	0.02459	0	0	0	0	0	0	1	-360	360;
	9300	5894	0.00557	0.04393	0	0	0	0	0	0	1	-360	360;
	12262	3565	0.00237	0.01131	0	0	0	0	0	0	1	-360	360;
	5088	10067	0.00258	0.04274	0	0	0	0	0	0	1	-360	360;
	125	2725	0.00861	0.04105	0	0	0	0	0	0	1	-360	360;
	125	1200	0.01719	0.08193	0	0	0	0	0	0	1	-360	360;
	125	1200	0.01719	0.08193	0	0	0	0	0	0	1	-360	360;
	125	6426	0.01305	0.06222	0	0	0	0	0	0	1	-360	360;
	3453	13182	0.01326	0.06319	0	0	0	0	0	0	1	-360	360;
	3453	5574	0.03288	0.15675	0	0	0	0	0	0	1	-360	360;
	3453	11614	0.0059	0.02812	0	0	0	0	0	0	1	-360	360;
	3453	13416	0.01898	0.0905	0	0	0	0	0	0	1	-360	360;
	13488	7865	0.02173	0.06979	0	0	0	0	0	0	1	-360	360;
	13488	10572	0.01007	0.07939	0	0	0	0	0	0	1	-360	360;
	13488	8294	0.04673	0.15006	0	0	0	0	0	0	1	-360	360;
	13488	7348	0.001	0.00785	0	0	0	0	0	0	1	-360	360;
	7865	1200	0.02168	0.06961	0	0	0	0	0	0	1	-360	360;
	2725	13182	0.01261	0.06011	0	0	0	0	0	0	1	-360	360;
	7658	5574	0.01464	0.04701	0	0	0	0	0	0	1	-360	360;
	8467	1307	0.01647	0.07854	0	0	0	0	0	0	1	-360	360;
	8467	1307	0.0264	0.08479	0	0	0	0	0	0	1	-360	360;
	8467	5574	0.01441	0.06868	0	0	0	0	0	0	1	-360	360;
	8467	5574	0.02393	0.07684	0	0	0	0	0	0	1	-360	360;
	2243	1307	0.01407	0.06706	0	0	0	0	0	0	1	-360	360;
	2243	11754	0.00508	0.02424	0	0	0	0	0	0	1	-360	360;
	1200	2629	0.00557	0.04393	0	0	0	0	0	0	1	-360	360;
	1200	2629	0.00557	0.04393	0	0	0	0	0	0	1	-360	360;
	1200	12105	0.00393	0.0241	0	0	0	0	0	0	1	-360	360;
	13182	6426	0.00783	0.03733	0	0	0	0	0	0	1	-360	360;
	13182	11614	0.01824	0.08694	0	0	0	0	0	0	1	-360	360;
	5363	5574	0.00951	0.075	0	0	0	0	0	0	1	-360	360;
	5363	3513	0.00438	0.03452	0	0	0	0	0	0	1	-360	360;
	951	7159	0.02314	0.07431	0	0	0	0	0	0	1	-360	360;
	951	9668	0.0199	0.1569	0	0	0	0	0	0	1	-360	360;
	951	13442	0.00818	0.06449	0	0	0	0	0	0	1	-360	360;
	1307	5574	0.03085	0.14706	0	0	0	0	0	0	1	-360	360;
	1307	168	0.01814	0.08646	0	0	0	0	0	0	1	-360	360;
	1307	9854	0.01187	0.05656	0	0	0	0	0	0	1	-360	360;
	1307	11754	0.01356	0.06464	0	0	0	0	0	0	1	-360	360;
	6426	5879	0.05844	0.18767	0	0	0	0	0	0	1	-360	360;
	5574	3513	0.01075	0.08473	0	0	0	0	0	0	1	-360	360;
	5574	11614	0.04729	0.15187	0	0	0	0	0	0	1	-360	360;
	5574	13416	0.01458	0.06949	0	0	0	0	0	0	1	-360	360;
	7159	13442	0.00774	0.06103	0	0	0	0	0	0	1	-360	360;
	3513	11614	0.01094	0.0863	0	0	0	0	0	0	1	-360	360;
	3513	11614	0.01094	0.0863	0	0	0	0	0	0	1	-360	360;
	10067	3088	0.00033	0.00416	0	0	0	0	0	0	1	-360	360;
	2629	10572	0.00318	0.0251	0	0	0	0	0	0	1	-360	360;
	2629	6255	0.00199	0.01569	0	0	0	0	0	0	1	-360	360;
	10572	7348	0.00597	0.04707	0	0	0	0	0	0	1	-360	360;
	10572	6255	0.00199	0.01569	0	0	0	0	0	0	1	-360	360;
	8462	10673	0.00238	0.03034	0	0	0	0	0	0	1	-360	360;
	8462	8874	0.00264	0.03358	0	0	0	0	0	0	1	-360	360;
	8462	3088	0.00228	0.02909	0	0	0	0	0	0	1	-360	360;
	10153	3959	0.05152	0.16543	0	0	0	0	0	0	1	-360	360;
	10153	13100	0.01719	0.13556	0	0	0	0	0	0	1	-360	360;
	10153	3431	0.02074	0.16349	0	0	0	0	0	0	1	-360	360;
	8294	10561	0.04358	0.13994	0	0	0	0	0	0	1	-360	360;
	1143	5879	0.03136	0.10071	0	0	0	0	0	0	1	-360	360;
	1143	8252	0.00782	0.06166	0	0	0	0	0	0	1	-360	360;
	1143	12557	0.01807	0.05804	0	0	0	0	0	0	1	-360	360;
	1143	1659	0.00022	0.00072	0	0	0	0	0	0	1	-360	360;
	5879	7783	0.03158	0.10143	0	0	0	0	0	0	1	-360	360;
	10561	9668	0.02477	0.07955	0	0	0	0	0	0	1	-360	360;
	3959	8280	0.02314	0.07431	0	0	0	0	0	0	1	-360	360;
	4799	3765	0.01464	0.04701	0	0	0	0	0	0	1	-360	360;
	4799	7783	0.02902	0.13833	0	0	0	0	0	0	1	-360	360;
	4799	4402	0.0241	0.07738	0	0	0	0	0	0	1	-360	360;
	4799	292	0.07702	0.24733	0	0	0	0	0	0	1	-360	360;
	10200	7680	0.00619	0.01989	0	0	0	0	0	0	1	-360	360;
	3736	8252	0.00448	0.02748	0	0	0	0	0	0	1	-360	360;
	3736	8252	0.00448	0.02748	0	0	0	0	0	0	1	-360	360;
	3736	605	0.02929	0.13962	0	0	0	0	0	0	1	-360	360;
	3736	13100	0.01572	0.12395	0	0	0	0	0	0	1	-360	360;
	3736	6753	0.01194	0.09414	0	0	0	0	0	0	1	-360	360;
	8252	931	0.00629	0.04958	0	0	0	0	0	0	1	-360	360;
	3765	5043	0.03491	0.1121	0	0	0	0	0	0	1	-360	360;
	605	4050	0.01593	0.07595	0	0	0	0	0	0	1	-360	360;
	13100	9668	0.05405	0.17357	0	0	0	0	0	0	1	-360	360;
	13100	6753	0.01174	0.09257	0	0	0	0	0	0	1	-360	360;
	8280	9668	0.03254	0.19959	0	0	0	0	0	0	1	-360	360;
	7680	292	0.00659	0.02115	0	0	0	0	0	0	1	-360	360;
	7680	292	0.00659	0.02115	0	0	0	0	0	0	1	-360	360;
	4402	12557	0.03356	0.10776	0	0	0	0	0	0	1	-360	360;
	4402	240	0.03277	0.10523	0	0	0	0	0	0	1	-360	360;
	5043	292	0.03783	0.1215	0	0	0	0	0	0	1	-360	360;
	292	7791	0.0181	0.08629	0	0	0	0	0	0	1	-360	360;
	10673	11184	0.00235	0.02992	0	0	0	0	0	0	1	-360	360;
	10266	10726	0.00211	0.02153	0	0	0	0	0	0	1	-360	360;
	474	10726	0.00232	0.02949	0	0	0	0	0	0	1	-360	360;
	474	11407	0.00179	0.02959	0	0	0	0	0	0	1	-360	360;
	474	10892	0.00179	0.02959	0	0	0	0	0	0	1	-360	360;
	474	4116	0.00134	0.01702	0	0	0	0	0	0	1	-360	360;
	474	7800	0.00084	0.01064	0	0	0	0	0	0	1	-360	360;
	474	7448	0.00124	0.01268	0	0	0	0	0	0	1	-360	360;
	474	4624	0.00106	0.01347	0	0	0	0	0	0	1	-360	360;
	1036	10726	0.00023	0.00231	0	0	0	0	0	0	1	-360	360;
	1036	8751	0.00335	0.03421	0	0	0	0	0	0	1	-360	360;
	10726	13612	0.00024	0.00241	0	0	0	0	0	0	1	-360	360;
	10726	13612	0.00024	0.00241	0	0	0	0	0	0	1	-360	360;
	10726	13612	0.00024	0.00241	0	0	0	0	0	0	1	-360	360;
	10726	13612	0.00024	0.00241	0	0	0	0	0	0	1	-360	360;
	10726	4417	0.00075	0.01233	0	0	0	0	0	0	1	-360	360;
	11407	1365	0	-0.03116	0	0	0	0	0	0	1	-360	360;
	10892	7644	0	-0.03116	0	0	0	0	0	0	1	-360	360;
	4116	8356	3e-05	0.00033	0	0	0	0	0	0	1	-360	360;
	5743	8751	0.00182	0.02317	0	0	0	0	0	0	1	-360	360;
	5743	6634	0.00081	0.01031	0	0	0	0	0	0	1	-360	360;
	3198	8751	0.00177	0.01813	0	0	0	0	0	0	1	-360	360;
	3198	8751	0.00175	0.01789	0	0	0	0	0	0	1	-360	360;
	3198	4624	0.00099	0.01257	0	0	0	0	0	0	1	-360	360;
	3198	3669	0.00125	0.01594	0	0	0	0	0	0	1	-360	360;
	3198	3669	0.00125	0.01594	0	0	0	0	0	0	1	-360	360;
	8751	1608	0.00014	0.00139	0	0	0	0	0	0	1	-360	360;
	8751	1608	0.00014	0.00139	0	0	0	0	0	0	1	-360	360;
	8751	1608	0.00014	0.00139	0	0	0	0	0	0	1	-360	360;
	8751	8874	0.00198	0.02527	0	0	0	0	0	0	1	-360	360;
	1500	7614	0.00845	0.02712	0	0	0	0	0	0	1	-360	360;
	1500	25	0.00736	0.05805	0	0	0	0	0	0	1	-360	360;
	1500	1922	0.02438	0.07829	0	0	0	0	0	0	1	-360	360;
	7614	4293	0.02308	0.07413	0	0	0	0	0	0	1	-360	360;
	7614	7569	0.01802	0.05786	0	0	0	0	0	0	1	-360	360;
	7614	12759	0.02252	0.07232	0	0	0	0	0	0	1	-360	360;
	9092	12686	0.01397	0.08565	0	0	0	0	0	0	1	-360	360;
	9092	10350	0.01937	0.06219	0	0	0	0	0	0	1	-360	360;
	9092	12061	0.01205	0.07392	0	0	0	0	0	0	1	-360	360;
	8584	11262	0.00788	0.02531	0	0	0	0	0	0	1	-360	360;
	8584	7090	0.01464	0.04701	0	0	0	0	0	0	1	-360	360;
	25	4293	0.00394	0.01266	0	0	0	0	0	0	1	-360	360;
	25	5801	0.03491	0.1121	0	0	0	0	0	0	1	-360	360;
	25	4410	0.02534	0.08136	0	0	0	0	0	0	1	-360	360;
	11972	1956	0.0469	0.15061	0	0	0	0	0	0	1	-360	360;
	11972	1956	0.0469	0.15061	0	0	0	0	0	0	1	-360	360;
	11972	7906	0.03913	0.12566	0	0	0	0	0	0	1	-360	360;
	11972	7906	0.03913	0.12566	0	0	0	0	0	0	1	-360	360;
	11972	8633	0.05332	0.17309	0	0	0	0	0	0	1	-360	360;
	11972	8633	0.05332	0.17309	0	0	0	0	0	0	1	-360	360;
	5801	1886	0.01015	0.08002	0	0	0	0	0	0	1	-360	360;
	5801	12686	0.0317	0.10179	0	0	0	0	0	0	1	-360	360;
	5801	9507	0.00719	0.03426	0	0	0	0	0	0	1	-360	360;
	5801	9507	0.00719	0.03426	0	0	0	0	0	0	1	-360	360;
	5801	6558	0.00143	0.0113	0	0	0	0	0	0	1	-360	360;
	5801	6558	0.00143	0.0113	0	0	0	0	0	0	1	-360	360;
	2233	8052	0.03463	0.11119	0	0	0	0	0	0	1	-360	360;
	2233	9819	0.01306	0.04195	0	0	0	0	0	0	1	-360	360;
	2233	11943	0.01689	0.05424	0	0	0	0	0	0	1	-360	360;
	8429	1510	0.03074	0.09872	0	0	0	0	0	0	1	-360	360;
	8429	10243	0.03547	0.1139	0	0	0	0	0	0	1	-360	360;
	667	11594	0.04622	0.14844	0	0	0	0	0	0	1	-360	360;
	667	9253	0.04949	0.15892	0	0	0	0	0	0	1	-360	360;
	667	7791	0.05106	0.16399	0	0	0	0	0	0	1	-360	360;
	1510	4930	0.0282	0.13445	0	0	0	0	0	0	1	-360	360;
	1510	8633	0.01391	0.04653	0	0	0	0	0	0	1	-360	360;
	1510	8633	0.01391	0.04653	0	0	0	0	0	0	1	-360	360;
	1510	4768	0.00282	0.00904	0	0	0	0	0	0	1	-360	360;
	1510	4768	0.00282	0.00904	0	0	0	0	0	0	1	-360	360;
	10807	5403	0.03744	0.12023	0	0	0	0	0	0	1	-360	360;
	6638	5403	0.00974	0.03128	0	0	0	0	0	0	1	-360	360;
	6638	5403	0.01357	0.04357	0	0	0	0	0	0	1	-360	360;
	6638	1886	0.00563	0.01808	0	0	0	0	0	0	1	-360	360;
	6638	5648	0.02263	0.07268	0	0	0	0	0	0	1	-360	360;
	10541	4768	0.03265	0.10486	0	0	0	0	0	0	1	-360	360;
	10541	3003	0.03491	0.1121	0	0	0	0	0	0	1	-360	360;
	5403	7050	0.00444	0.02117	0	0	0	0	0	0	1	-360	360;
	5403	7050	0.00444	0.02117	0	0	0	0	0	0	1	-360	360;
	5403	7906	0.01605	0.05153	0	0	0	0	0	0	1	-360	360;
	5403	7906	0.01605	0.05153	0	0	0	0	0	0	1	-360	360;
	5403	6816	0.01407	0.0452	0	0	0	0	0	0	1	-360	360;
	7050	1886	0.01244	0.03996	0	0	0	0	0	0	1	-360	360;
	7050	1886	0.01244	0.03996	0	0	0	0	0	0	1	-360	360;
	1886	11911	0.01271	0.0606	0	0	0	0	0	0	1	-360	360;
	10243	11594	0.02928	0.09402	0	0	0	0	0	0	1	-360	360;
	10243	11594	0.02928	0.09402	0	0	0	0	0	0	1	-360	360;
	10243	8233	0.04909	0.15766	0	0	0	0	0	0	1	-360	360;
	7569	3713	0.04335	0.13922	0	0	0	0	0	0	1	-360	360;
	11594	6164	0.02695	0.12847	0	0	0	0	0	0	1	-360	360;
	11594	2299	0.02695	0.12847	0	0	0	0	0	0	1	-360	360;
	3713	6074	0.0122	0.09618	0	0	0	0	0	0	1	-360	360;
	3713	3357	0.0304	0.09763	0	0	0	0	0	0	1	-360	360;
	9253	10810	0.03536	0.11354	0	0	0	0	0	0	1	-360	360;
	9253	10248	0.00338	0.01085	0	0	0	0	0	0	1	-360	360;
	9253	3357	0.02837	0.09112	0	0	0	0	0	0	1	-360	360;
	9253	729	0.0139	0.06626	0	0	0	0	0	0	1	-360	360;
	12098	10828	0.00325	0.01551	0	0	0	0	0	0	1	-360	360;
	12098	7906	0.01376	0.06561	0	0	0	0	0	0	1	-360	360;
	12098	11731	0.00472	0.03718	0	0	0	0	0	0	1	-360	360;
	8052	13484	0.00756	0.05962	0	0	0	0	0	0	1	-360	360;
	8052	13484	0.00756	0.05962	0	0	0	0	0	0	1	-360	360;
	8052	4965	0.00631	0.02025	0	0	0	0	0	0	1	-360	360;
	8052	402	0.01194	0.09414	0	0	0	0	0	0	1	-360	360;
	429	10828	0.0047	0.03703	0	0	0	0	0	0	1	-360	360;
	429	7906	0.0152	0.04882	0	0	0	0	0	0	1	-360	360;
	429	11731	0.00115	0.0091	0	0	0	0	0	0	1	-360	360;
	8233	10810	0.0179	0.05749	0	0	0	0	0	0	1	-360	360;
	8233	1956	0.01211	0.03887	0	0	0	0	0	0	1	-360	360;
	8233	12759	0.02517	0.08082	0	0	0	0	0	0	1	-360	360;
	10828	3003	0.00619	0.01989	0	0	0	0	0	0	1	-360	360;
	7906	6816	0.01464	0.04701	0	0	0	0	0	0	1	-360	360;
	12366	8580	0.01757	0.05641	0	0	0	0	0	0	1	-360	360;
	12366	8974	0.01878	0.06123	0	0	0	0	0	0	1	-360	360;
	1922	5648	0.02674	0.08588	0	0	0	0	0	0	1	-360	360;
	6164	7791	0.04673	0.15006	0	0	0	0	0	0	1	-360	360;
	6164	2299	0.00271	0.01293	0	0	0	0	0	0	1	-360	360;
	6074	8974	0.02424	0.07877	0	0	0	0	0	0	1	-360	360;
	12686	9507	0.00822	0.0648	0	0	0	0	0	0	1	-360	360;
	12686	1693	0.02871	0.09221	0	0	0	0	0	0	1	-360	360;
	9187	3956	0.00263	0.02071	0	0	0	0	0	0	1	-360	360;
	9187	10350	0.02224	0.07142	0	0	0	0	0	0	1	-360	360;
	9187	8138	0.00261	0.02055	0	0	0	0	0	0	1	-360	360;
	9187	1693	0.00845	0.02712	0	0	0	0	0	0	1	-360	360;
	9819	12061	0.0114	0.0699	0	0	0	0	0	0	1	-360	360;
	3956	5440	0.01407	0.0452	0	0	0	0	0	0	1	-360	360;
	8580	13484	0.00796	0.06276	0	0	0	0	0	0	1	-360	360;
	8580	11943	0.00563	0.01808	0	0	0	0	0	0	1	-360	360;
	8580	7860	0.01126	0.03616	0	0	0	0	0	0	1	-360	360;
	8138	1693	0.00597	0.04707	0	0	0	0	0	0	1	-360	360;
	8138	12061	0.01393	0.10983	0	0	0	0	0	0	1	-360	360;
	8138	7860	0.03828	0.12294	0	0	0	0	0	0	1	-360	360;
	7800	11422	0.00117	0.01496	0	0	0	0	0	0	1	-360	360;
	7448	8356	0.00076	0.00964	0	0	0	0	0	0	1	-360	360;
	13473	8270	0.00085	0.01081	0	0	0	0	0	0	1	-360	360;
	13473	11422	0.00093	0.0118	0	0	0	0	0	0	1	-360	360;
	2227	8974	0.00697	0.05491	0	0	0	0	0	0	1	-360	360;
	2227	8974	0.00697	0.05491	0	0	0	0	0	0	1	-360	360;
	466	3003	0.00113	0.00362	0	0	0	0	0	0	1	-360	360;
	3300	7860	0.00056	0.00181	0	0	0	0	0	0	1	-360	360;
	10248	4364	0.00338	0.01085	0	0	0	0	0	0	1	-360	360;
	3669	12408	0.00122	0.01546	0	0	0	0	0	0	1	-360	360;
	4818	3431	0.03378	0.10848	0	0	0	0	0	0	1	-360	360;
	4818	9950	0.0117	0.09226	0	0	0	0	0	0	1	-360	360;
	4818	4050	0.01891	0.14906	0	0	0	0	0	0	1	-360	360;
	4818	4050	0.01891	0.14906	0	0	0	0	0	0	1	-360	360;
	11743	5237	0.07702	0.24733	0	0	0	0	0	0	1	-360	360;
	11743	11282	0.01526	0.07272	0	0	0	0	0	0	1	-360	360;
	2576	6537	0.0308	0.0989	0	0	0	0	0	0	1	-360	360;
	2576	1006	0.02787	0.0895	0	0	0	0	0	0	1	-360	360;
	2576	4748	0.00038	0.00298	0	0	0	0	0	0	1	-360	360;
	2576	4748	0.00038	0.00298	0	0	0	0	0	0	1	-360	360;
	3203	9950	0.02291	0.18059	0	0	0	0	0	0	1	-360	360;
	3203	3574	0.00593	0.02828	0	0	0	0	0	0	1	-360	360;
	3203	11432	0.00776	0.06119	0	0	0	0	0	0	1	-360	360;
	8787	380	0.04262	0.13687	0	0	0	0	0	0	1	-360	360;
	8787	3600	0.02065	0.09841	0	0	0	0	0	0	1	-360	360;
	8787	7140	0.05591	0.17953	0	0	0	0	0	0	1	-360	360;
	8787	1064	0.01196	0.0943	0	0	0	0	0	0	1	-360	360;
	8787	1064	0.01196	0.0943	0	0	0	0	0	0	1	-360	360;
	741	3443	0.00644	0.0307	0	0	0	0	0	0	1	-360	360;
	741	3443	0.00644	0.0307	0	0	0	0	0	0	1	-360	360;
	11013	1049	0.00241	0.01898	0	0	0	0	0	0	1	-360	360;
	11013	1049	0.00241	0.01898	0	0	0	0	0	0	1	-360	360;
	11013	1049	0.00249	0.01961	0	0	0	0	0	0	1	-360	360;
	11013	9163	0.01208	0.09524	0	0	0	0	0	0	1	-360	360;
	11013	5407	0.07229	0.23215	0	0	0	0	0	0	1	-360	360;
	1049	1387	0.0264	0.08479	0	0	0	0	0	0	1	-360	360;
	1049	3443	0.00762	0.06009	0	0	0	0	0	0	1	-360	360;
	1049	1006	0.01057	0.08331	0	0	0	0	0	0	1	-360	360;
	1049	5522	0.0029	0.02291	0	0	0	0	0	0	1	-360	360;
	1049	5522	0.0029	0.02291	0	0	0	0	0	0	1	-360	360;
	6627	10121	0.01582	0.0508	0	0	0	0	0	0	1	-360	360;
	6627	9950	0.04887	0.15693	0	0	0	0	0	0	1	-360	360;
	3357	380	0.01668	0.07951	0	0	0	0	0	0	1	-360	360;
	3357	380	0.01668	0.07951	0	0	0	0	0	0	1	-360	360;
	3357	6412	0.00749	0.02405	0	0	0	0	0	0	1	-360	360;
	3357	729	0.00617	0.02941	0	0	0	0	0	0	1	-360	360;
	3357	4364	0.02815	0.0904	0	0	0	0	0	0	1	-360	360;
	3357	1056	0.02388	0.0804	0	0	0	0	0	0	1	-360	360;
	5237	9950	0.05461	0.17538	0	0	0	0	0	0	1	-360	360;
	5237	11684	0.02032	0.06527	0	0	0	0	0	0	1	-360	360;
	3431	11684	0.05236	0.16814	0	0	0	0	0	0	1	-360	360;
	1387	3443	0.00555	0.04377	0	0	0	0	0	0	1	-360	360;
	1387	2944	0.00788	0.02531	0	0	0	0	0	0	1	-360	360;
	9163	4965	0.01003	0.07908	0	0	0	0	0	0	1	-360	360;
	380	7185	0.01194	0.09414	0	0	0	0	0	0	1	-360	360;
	6412	2944	0.01982	0.06364	0	0	0	0	0	0	1	-360	360;
	6412	10314	0.00169	0.00542	0	0	0	0	0	0	1	-360	360;
	3600	3574	0.01393	0.10983	0	0	0	0	0	0	1	-360	360;
	4965	5407	0.01172	0.09241	0	0	0	0	0	0	1	-360	360;
	8738	3437	0.00441	0.02101	0	0	0	0	0	0	1	-360	360;
	7140	9950	0.04014	0.12891	0	0	0	0	0	0	1	-360	360;
	8346	7673	0.00597	0.04707	0	0	0	0	0	0	1	-360	360;
	8346	7673	0.00597	0.04707	0	0	0	0	0	0	1	-360	360;
	8346	3253	0.01017	0.04848	0	0	0	0	0	0	1	-360	360;
	8346	3253	0.01017	0.04848	0	0	0	0	0	0	1	-360	360;
	8346	7185	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	8346	1064	0.0002	0.00157	0	0	0	0	0	0	1	-360	360;
	8346	1064	0.0002	0.00157	0	0	0	0	0	0	1	-360	360;
	2175	7642	0.01689	0.05424	0	0	0	0	0	0	1	-360	360;
	2175	1178	4e-05	0.00041	0	0	0	0	0	0	1	-360	360;
	2175	918	0.01848	0.06679	0	0	0	0	0	0	1	-360	360;
	2175	918	0.02061	0.06617	0	0	0	0	0	0	1	-360	360;
	12315	5407	0.01925	0.06183	0	0	0	0	0	0	1	-360	360;
	12315	7642	0.00617	0.04864	0	0	0	0	0	0	1	-360	360;
	5407	402	0.00847	0.0404	0	0	0	0	0	0	1	-360	360;
	10121	6537	0.01064	0.03417	0	0	0	0	0	0	1	-360	360;
	10121	918	0.04286	0.14508	0	0	0	0	0	0	1	-360	360;
	240	3437	0.01729	0.08242	0	0	0	0	0	0	1	-360	360;
	240	3437	0.01729	0.08242	0	0	0	0	0	0	1	-360	360;
	240	1056	0.01149	0.04063	0	0	0	0	0	0	1	-360	360;
	5618	11282	0.01328	0.13577	0	0	0	0	0	0	1	-360	360;
	5618	918	0.02061	0.06617	0	0	0	0	0	0	1	-360	360;
	7642	4748	0.02567	0.2024	0	0	0	0	0	0	1	-360	360;
	12408	11184	0.00268	0.03408	0	0	0	0	0	0	1	-360	360;
	12408	8270	0.00052	0.00665	0	0	0	0	0	0	1	-360	360;
	4748	11432	0.0004	0.00314	0	0	0	0	0	0	1	-360	360;
	3190	4364	0.01126	0.03616	0	0	0	0	0	0	1	-360	360;
	11184	5949	0.00209	0.0266	0	0	0	0	0	0	1	-360	360;
	8270	8119	0.00105	0.0133	0	0	0	0	0	0	1	-360	360;
	4804	2752	0.00189	0.02411	0	0	0	0	0	0	1	-360	360;
	11928	11422	0	0.0002	0	0	0	0	0	0	1	-360	360;
	5949	8119	0.00196	0.02494	0	0	0	0	0	0	1	-360	360;
	8119	8892	0.00034	0.00432	0	0	0	0	0	0	1	-360	360;
	8408	13128	0.00067	0.00682	0	0	0	0	0	0	1	-360	360;
	13128	1741	0.00022	0.0024	0	0	0	0	0	0	1	-360	360;
	11255	1256	0.00636	0.02043	0	0	0	0	0	0	1	-360	360;
	9601	12149	6.3e-05	0.00019	0	0	0	0	0	0	1	-360	360;
	7348	13442	0.00398	0.03138	0	0	0	0	0	0	1	-360	360;
	7348	13442	0.00398	0.03138	0	0	0	0	0	0	1	-360	360;
	5440	4410	0.00394	0.01266	0	0	0	0	0	0	1	-360	360;
	12438	11943	0.00011	0.00036	0	0	0	0	0	0	1	-360	360;
	8859	10123	0.00135	0.01379	0	0	0	0	1.052632	0	1	-360	360;
	4222	7260	0.00097	0.01239	0	0	0	0	1.052632	0	1	-360	360;
	642	6538	0.00084	0.01064	0	0	0	0	1.052632	0	1	-360	360;
	11825	2422	0	2.80858	0	0	0	0	0	0	1	-360	360;
	11825	7731	0	8.84701	0	0	0	0	0	0	1	-360	360;
	11825	1599	0	2.52772	0	0	0	0	0	0	1	-360	360;
	11825	9830	0	2.52772	0	0	0	0	0	0	1	-360	360;
	11825	7962	0	3.53881	0	0	0	0	0	0	1	-360	360;
	11825	11903	0	3.53881	0	0	0	0	0	0	1	-360	360;
	11825	12507	0	8.84701	0	0	0	0	0	0	1	-360	360;
	11825	3894	0	2.52772	0	0	0	0	0	0	1	-360	360;
	11825	8359	0	1.41552	0	0	0	0	0	0	1	-360	360;
	11825	5658	0	8.74925	0	0	0	0	0	0	1	-360	360;
	11825	423	0	8.74925	0	0	0	0	0	0	1	-360	360;
	11825	2192	0	8.74925	0	0	0	0	0	0	1	-360	360;
	11825	4611	0	8.74925	0	0	0	0	0	0	1	-360	360;
	11825	2526	0	8.74925	0	0	0	0	0	0	1	-360	360;
	9103	4746	0	2.11983	0	0	0	0	0	0	1	-360	360;
	9103	10613	0	6.3595	0	0	0	0	0	0	1	-360	360;
	214	1048	0	5.36212	0	0	0	0	0	0	1	-360	360;
	214	5254	0	5.36212	0	0	0	0	0	0	1	-360	360;
	214	8246	0	5.36212	0	0	0	0	0	0	1	-360	360;
	214	3360	0	5.36212	0	0	0	0	0	0	1	-360	360;
	214	6347	0	5.36212	0	0	0	0	0	0	1	-360	360;
	214	11781	0	2.29805	0	0	0	0	0	0	1	-360	360;
	214	3348	0	0.94626	0	0	0	0	0	0	1	-360	360;
	214	13510	0	8.31475	0	0	0	0	0	0	1	-360	360;
	214	9797	0	8.31475	0	0	0	0	0	0	1	-360	360;
	214	5210	0	3.3259	0	0	0	0	0	0	1	-360	360;
	214	9728	0	5.54317	0	0	0	0	0	0	1	-360	360;
	214	635	0	5.54317	0	0	0	0	0	0	1	-360	360;
	214	12541	0	3.3259	0	0	0	0	0	0	1	-360	360;
	214	11393	0	8.31475	0	0	0	0	0	0	1	-360	360;
	214	9056	0	2.07869	0	0	0	0	0	0	1	-360	360;
	214	4009	0	8.31475	0	0	0	0	0	0	1	-360	360;
	214	12221	0	1.66295	0	0	0	0	0	0	1	-360	360;
	214	392	0	2.53082	0	0	0	0	0	0	1	-360	360;
	214	9421	0	2.95262	0	0	0	0	0	0	1	-360	360;
	214	10829	0	2.53082	0	0	0	0	0	0	1	-360	360;
	214	1088	0	8.85787	0	0	0	0	0	0	1	-360	360;
	214	8764	0	0.70863	0	0	0	0	0	0	1	-360	360;
	214	5333	0	2.53082	0	0	0	0	0	0	1	-360	360;
	214	9997	0	8.31475	0	0	0	0	0	0	1	-360	360;
	2042	10452	-0.02733	0.34076	0	0	0	0	0	0	1	-360	360;
	2042	5116	-0.01665	0.37435	0	0	0	0	0	0	1	-360	360;
	2042	12383	0	0.34335	0	0	0	0	0	0	1	-360	360;
	2042	7114	0	0.42788	0	0	0	0	0	0	1	-360	360;
	2042	10891	-0.02145	0.31331	0	0	0	0	0	0	1	-360	360;
	2042	8033	0	0.19668	0	0	0	0	0	0	1	-360	360;
	2042	2712	0	0.37196	0	0	0	0	0	0	1	-360	360;
	7827	5316	0.11123	1.47467	0	0	0	0	0	0	1	-360	360;
	7119	332	0.06731	4.18579	0	0	0	0	0	0	1	-360	360;
	7119	1421	0.06731	4.18579	0	0	0	0	0	0	1	-360	360;
	7119	6717	0.06731	4.18579	0	0	0	0	0	0	1	-360	360;
	4116	6272	0	2.04539	0	0	0	0	0	0	1	-360	360;
	4116	11672	0	4.09078	0	0	0	0	0	0	1	-360	360;
	10180	111	0.00431	0.09324	0	0	0	0	0	0	1	-360	360;
	10180	111	0.0044	0.091339	0	0	0	0	0	0	1	-360	360;
	11863	8494	0.006356	0.142878	0	0	0	0	1.026667	0	1	-360	360;
	3549	10458	0.00523	0.10214	0	0	0	0	0	0	1	-360	360;
	3549	10458	0.00532	0.099917	0	0	0	0	0	0	1	-360	360;
	3549	10458	0.00442	0.105388	0	0	0	0	0	0	1	-360	360;
	12153	8241	0.001002	0.059033	0	0	0	0	1.052632	0	1	-360	360;
	12153	8241	0.000812	0.056028	0	0	0	0	1.052632	0	1	-360	360;
	10717	11653	0.000921	0.058347	0	0	0	0	1.052632	0	1	-360	360;
	11653	10859	0.0016	0.07243	0	0	0	0	0	0	1	-360	360;
	11653	10859	0.00182	0.07098	0	0	0	0	0	0	1	-360	360;
	11653	10859	0.00185	0.06953	0	0	0	0	0	0	1	-360	360;
	414	12283	0.00349	0.07969	0	0	0	0	0	0	1	-360	360;
	414	12283	0.00355	0.07799	0	0	0	0	0	0	1	-360	360;
	2209	414	0.001101	0.048618	0	0	0	0	1.052632	0	1	-360	360;
	2209	414	0.000984	0.055107	0	0	0	0	1.052632	0	1	-360	360;
	2972	13068	0.00175	0.07752	0	0	0	0	0	0	1	-360	360;
	2972	13068	0.00171	0.07901	0	0	0	0	0	0	1	-360	360;
	8698	6348	0.00302	0.091446	0	0	0	0	0	0	1	-360	360;
	8698	6348	0.00295	0.084678	0	0	0	0	0	0	1	-360	360;
	8720	6779	0.0018	0.07815	0	0	0	0	0	0	1	-360	360;
	8720	6779	0.00185	0.07724	0	0	0	0	0	0	1	-360	360;
	8720	6779	0.00193	0.07007	0	0	0	0	0	0	1	-360	360;
	850	3658	0.00205	0.094653	0	0	0	0	1.052632	0	1	-360	360;
	7565	756	0.000454	0.028947	0	0	0	0	1.013237	-0.010791	1	-360	360;
	7565	756	0.0004	0.032809	0	0	0	0	1.013237	-0.010791	1	-360	360;
	3133	11180	0.000342	0.033341	0	0	0	0	0.85879	0	1	-360	360;
	1929	11180	0.000342	0.03398	0	0	0	0	0.859279	0	1	-360	360;
	9701	11180	0.000121	0.016316	0	0	0	0	0.998827	0.048422	1	-360	360;
	9701	5617	0.000121	0.025279	0	0	0	0	0.953802	-0.305146	1	-360	360;
	8760	2611	0.000347	0.031823	0	0	0	0	1.019726	-0.021448	1	-360	360;
	8760	2611	0.000356	0.031177	0	0	0	0	1.019726	-0.021448	1	-360	360;
	6316	971	0	0.046856	0	0	0	0	0	-0.305565	1	-360	360;
	6316	2611	0	0.045844	0	0	0	0	0	0.305565	1	-360	360;
	4778	8152	0.000431	0.03794	0	0	0	0	1.083008	-0.111289	1	-360	360;
	4778	8152	0.00038	0.046489	0	0	0	0	1.098173	-0.129803	1	-360	360;
	5140	8509	0.0004	0.033343	0	0	0	0	1.006865	0	1	-360	360;
	5140	8509	0.0004	0.033981	0	0	0	0	1.006865	0	1	-360	360;
	8134	8509	0.000121	0.015958	0	0	0	0	0.998827	0.048422	1	-360	360;
	8134	5450	0.000121	0.014773	0	0	0	0	0.998827	0.048422	1	-360	360;
	517	446	0.000131	0.014593	0	0	0	0	0.999706	-0.024225	1	-360	360;
	517	9874	0.000141	0.013923	0	0	0	0	0	0	1	-360	360;
	4357	5711	0.000194	0.013059	0	0	0	0	0.981541	0	1	-360	360;
	11750	404	0.000463	0.031001	0	0	0	0	0.966619	0	1	-360	360;
	11750	404	0.000472	0.030331	0	0	0	0	0.967946	0	1	-360	360;
	672	249	0.000163	0.015501	0	0	0	0	0.947992	0	1	-360	360;
	8315	8142	0.001987	0.071722	0	0	0	0	0.960533	0	1	-360	360;
	3604	9284	0.000756	0.051399	0	0	0	0	0.948208	0	1	-360	360;
	3604	9284	0.000746	0.05345	0	0	0	0	0.957785	0	1	-360	360;
	10639	13644	0.00069	0.04333	0	0	0	0	1.052632	0	1	-360	360;
	3925	10834	0.0022	0.065399	0	0	0	0	0.966291	0	1	-360	360;
	3925	10834	0.002235	0.064019	0	0	0	0	0.916317	0	1	-360	360;
	12721	7418	0.00168	0.07424	0	0	0	0	0	0	1	-360	360;
	1360	5067	0.002036	0.070048	0	0	0	0	0.95687	0	1	-360	360;
	1360	5067	0.001997	0.071368	0	0	0	0	0.955577	0	1	-360	360;
	5338	1360	0.00035	0.032859	0	0	0	0	1.052632	0	1	-360	360;
	5338	1360	0.0004	0.0322	0	0	0	0	1.052632	0	1	-360	360;
	3750	5578	0.00205	0.06915	0	0	0	0	0	0	1	-360	360;
	3750	5578	0.00082	0.07342	0	0	0	0	0	0	1	-360	360;
	524	4231	0.002658	0.07732	0	0	0	0	0.985917	0	1	-360	360;
	524	4231	0.002712	0.075649	0	0	0	0	0.989209	0	1	-360	360;
	7701	5663	0.00065	0.04604	0	0	0	0	1.052632	0	1	-360	360;
	10387	3766	0.00062	0.03629	0	0	0	0	1.052632	0	1	-360	360;
	10387	2914	0.00019	0.03263	0	0	0	0	1.052632	0	1	-360	360;
	3840	1375	0.001797	0.07284	0	0	0	0	0.973898	0	1	-360	360;
	4711	12377	0.002712	0.072578	0	0	0	0	0.922299	0	1	-360	360;
	10677	524	0.00047	0.0307	0	0	0	0	1.052632	0	1	-360	360;
	11279	5793	0.002601	0.071481	0	0	0	0	0.969761	0	1	-360	360;
	11279	5793	0.002653	0.06993	0	0	0	0	0.957071	0	1	-360	360;
	8195	11279	0.00037	0.034169	0	0	0	0	1.052632	0	1	-360	360;
	8195	11279	0.00037	0.034831	0	0	0	0	1.052632	0	1	-360	360;
	9783	5452	0.0004	0.03524	0	0	0	0	1.052632	0	1	-360	360;
	9783	5161	0.00069	0.04525	0	0	0	0	1.052632	0	1	-360	360;
	13164	11388	0.00069	0.04231	0	0	0	0	1.052632	0	1	-360	360;
	4485	11018	0.00119	0.073146	0	0	0	0	0.969375	0	1	-360	360;
	4485	11018	0.002764	0.071607	0	0	0	0	0.944216	0	1	-360	360;
	3183	3359	0.000742	0.067219	0	0	0	0	0.96844	0	1	-360	360;
	3183	3359	0.000793	0.06388	0	0	0	0	0.973571	0	1	-360	360;
	8848	13586	0.000609	0.046809	0	0	0	0	0.930147	0	1	-360	360;
	8848	4381	0.00059	0.04521	0	0	0	0	0.878149	0	1	-360	360;
	11493	4381	0.519194	5.108058	0	0	0	0	0	0	1	-360	360;
	7727	3277	0.000559	0.039982	0	0	0	0	0.938747	0	1	-360	360;
	7727	8135	0.000537	0.045312	0	0	0	0	0.962486	0	1	-360	360;
	7221	2593	0.000889	0.061549	0	0	0	0	0.965788	0	1	-360	360;
	2593	1142	1.467822	7.764888	0	0	0	0	0	0	1	-360	360;
	9805	6724	0.000586	0.049031	0	0	0	0	0.920896	0	1	-360	360;
	3500	5137	0.000417	0.047801	0	0	0	0	0.93016	0	1	-360	360;
	1937	9970	0.000755	0.05143	0	0	0	0	0.979149	0	1	-360	360;
	2394	342	0.000733	0.052399	0	0	0	0	0.893108	0	1	-360	360;
	11493	9970	0.00078	0.064311	0	0	0	0	0	0	1	-360	360;
	962	4784	0.00089	0.048632	0	0	0	0	0.950966	0	1	-360	360;
	7602	1142	0.00855	0.08972	0	0	0	0	0	0	1	-360	360;
	2044	1142	0.00223	0.04836	0	0	0	0	0	0	1	-360	360;
	10523	3070	0.000581	0.043111	0	0	0	0	0.912599	0	1	-360	360;
	9981	13254	0.000579	0.043272	0	0	0	0	0.90589	0	1	-360	360;
	5402	10366	0.000586	0.049031	0	0	0	0	0.910127	0	1	-360	360;
	12041	6499	0.000559	0.049192	0	0	0	0	0.922734	0	1	-360	360;
	13341	11058	0.000573	0.051129	0	0	0	0	0.907986	0	1	-360	360;
	3238	7602	0.000417	0.039229	0	0	0	0	0.953969	0	1	-360	360;
	7602	9831	0.230542	1.197733	0	0	0	0	0	0	1	-360	360;
	7602	12273	0.02284	0.2644	0	0	0	0	0	0	1	-360	360;
	7602	3510	0.08537	0.736798	0	0	0	0	0	0	1	-360	360;
	7602	6844	1.402768	3.640289	0	0	0	0	0	0	1	-360	360;
	7602	6590	1.050227	4.450207	0	0	0	0	0	0	1	-360	360;
	12038	9831	0.000716	0.04371	0	0	0	0	0.946441	0	1	-360	360;
	2044	9831	0.518622	2.317689	0	0	0	0	0	0	1	-360	360;
	7626	12273	0.00059	0.045662	0	0	0	0	1.01916	0	1	-360	360;
	7626	2191	0.000603	0.050959	0	0	0	0	0.956199	0	1	-360	360;
	2044	12273	0.047769	0.554178	0	0	0	0	0	0	1	-360	360;
	11382	10883	0.000591	0.045153	0	0	0	0	0.930134	0	1	-360	360;
	11382	412	0.0006	0.047778	0	0	0	0	0.924997	0	1	-360	360;
	11493	8395	0.02436	0.270151	0	0	0	0	0	0	1	-360	360;
	5712	2044	0.000802	0.05727	0	0	0	0	0.928766	0	1	-360	360;
	5712	2044	0.000823	0.056059	0	0	0	0	0.934686	0	1	-360	360;
	2044	3510	0.188099	1.291901	0	0	0	0	0	0	1	-360	360;
	2044	6844	2.466323	7.26343	0	0	0	0	0	0	1	-360	360;
	3239	4174	0.000523	0.050352	0	0	0	0	0.901137	0	1	-360	360;
	13103	8395	0.000682	0.049638	0	0	0	0	0.969199	0	1	-360	360;
	13103	3579	0.00099	0.04736	0	0	0	0	1.014121	0	1	-360	360;
	471	3510	0.000616	0.048759	0	0	0	0	0.897787	0	1	-360	360;
	11179	7780	9e-05	0.008611	0	0	0	0	0	0	1	-360	360;
	8510	11012	9e-05	0.015192	0	0	0	0	0	-0.428189	1	-360	360;
	129	6590	0.000664	0.045609	0	0	0	0	0.910348	0	1	-360	360;
	3618	6844	0.000649	0.044862	0	0	0	0	0.926974	0	1	-360	360;
	8729	6473	0.00067	0.04326	0	0	0	0	0.906404	0	1	-360	360;
	3049	8729	9e-05	0.012629	0	0	0	0	0	0.304578	1	-360	360;
	10227	5892	0.000246	0.018419	0	0	0	0	0.915375	0.063819	1	-360	360;
	5892	8032	0.00131	0.07578	0	0	0	0	1.045454	0	1	-360	360;
	5892	8032	0.00128	0.07077	0	0	0	0	1.045454	0	1	-360	360;
	4589	12971	0.00131	0.062651	0	0	0	0	0	0	1	-360	360;
	4589	12971	0.00133	0.0577	0	0	0	0	0	0	1	-360	360;
	7401	2028	0.001495	0.059475	0	0	0	0	0.974576	0	1	-360	360;
	7401	13653	0.001453	0.061392	0	0	0	0	0.974576	0	1	-360	360;
	3581	7895	0.001475	0.074554	0	0	0	0	0.981267	0	1	-360	360;
	3581	7895	0.001309	0.084508	0	0	0	0	0.981267	0	1	-360	360;
	1248	2267	0.00119	0.084099	0	0	0	0	0	0	1	-360	360;
	6446	1542	0.00117	0.064859	0	0	0	0	0	0	1	-360	360;
	6446	1542	0.00114	0.066091	0	0	0	0	0	0	1	-360	360;
	11068	8672	0.001225	0.083401	0	0	0	0	0.973021	0	1	-360	360;
	11068	8672	0.001394	0.082574	0	0	0	0	0.973021	0	1	-360	360;
	5019	6742	0.001415	0.080475	0	0	0	0	0.973021	0	1	-360	360;
	5019	6742	0.001447	0.078793	0	0	0	0	0.973021	0	1	-360	360;
	13321	13072	0.001444	0.074683	0	0	0	0	0.981267	0	1	-360	360;
	13321	13072	0.001475	0.073051	0	0	0	0	0.981267	0	1	-360	360;
	1292	6703	0.001316	0.070415	0	0	0	0	0.974576	0	1	-360	360;
	11423	5014	0.00128	0.065919	0	0	0	0	0	0	1	-360	360;
	13151	704	0.001056	0.075919	0	0	0	0	1.052632	0	1	-360	360;
	1784	704	0.001451	0.078431	0	0	0	0	0.905512	0	1	-360	360;
	1784	704	0.001415	0.075297	0	0	0	0	0.905512	0	1	-360	360;
	3834	2401	0.001411	0.061844	0	0	0	0	0.974576	0	1	-360	360;
	3834	2401	0.001432	0.058938	0	0	0	0	0.974576	0	1	-360	360;
	3420	2931	0.00229	0.085231	0	0	0	0	0.979339	0	1	-360	360;
	7537	832	0.001474	0.06058	0	0	0	0	0.974576	0	1	-360	360;
	2035	1873	0.00125	0.064651	0	0	0	0	0	0	1	-360	360;
	2035	1873	0.00123	0.065919	0	0	0	0	0	0	1	-360	360;
	2035	1873	0.0012	0.06718	0	0	0	0	0	0	1	-360	360;
	573	10751	0.001221	0.069631	0	0	0	0	0.982906	0	1	-360	360;
	573	10751	0.00119	0.065604	0	0	0	0	0.982906	0	1	-360	360;
	2384	13523	0.001039	0.045925	0	0	0	0	0.981267	0	1	-360	360;
	13646	13522	0.000406	0.027181	0	0	0	0	1.57924	-0.661366	1	-360	360;
	13522	2673	0.001563	0.060356	0	0	0	0	0.982906	0	1	-360	360;
	13522	2673	0.001459	0.059041	0	0	0	0	0.982906	0	1	-360	360;
	8812	1580	0.001531	0.068989	0	0	0	0	0.942623	0	1	-360	360;
	12258	9419	0.000318	0.014328	0	0	0	0	1.002506	0	1	-360	360;
	2900	8749	0.000246	0.018419	0	0	0	0	0.915375	0.063819	1	-360	360;
	8749	5414	0.00121	0.066091	0	0	0	0	0	0	1	-360	360;
	8749	5414	0.00118	0.062289	0	0	0	0	0	0	1	-360	360;
	12618	9130	0.001646	0.071174	0	0	0	0	0.905512	0	1	-360	360;
	12618	9130	0.00229	0.087099	0	0	0	0	1.045454	0	1	-360	360;
	2745	1565	0.001421	0.061107	0	0	0	0	0.974576	0	1	-360	360;
	12830	4536	0.001573	0.080941	0	0	0	0	0.956938	0	1	-360	360;
	12830	4536	0.001594	0.078385	0	0	0	0	0.956938	0	1	-360	360;
	7538	12389	0.001369	0.06454	0	0	0	0	0.974576	0	1	-360	360;
	7538	12389	0.001348	0.065203	0	0	0	0	0.974576	0	1	-360	360;
	1643	4221	0.001295	0.071353	0	0	0	0	0.974576	0	1	-360	360;
	1643	4221	0.001295	0.070826	0	0	0	0	0.974576	0	1	-360	360;
	2145	6309	0.001441	0.084934	0	0	0	0	0.956938	0	1	-360	360;
	2145	6309	0.001463	0.083626	0	0	0	0	0.956938	0	1	-360	360;
	5852	46	0.000448	0.016447	0	0	0	0	1.002506	0	1	-360	360;
	5852	2346	0.001134	0.053193	0	0	0	0	0.998185	0	1	-360	360;
	46	2346	0.00136	0.057539	0	0	0	0	0	0	1	-360	360;
	5205	6143	0.001432	0.06454	0	0	0	0	0.974576	0	1	-360	360;
	969	5917	0.00195	0.114628	0	0	0	0	0.979339	0	1	-360	360;
	2598	10920	0.001295	0.067067	0	0	0	0	0.974576	0	1	-360	360;
	349	13276	0.001242	0.074563	0	0	0	0	0.974576	0	1	-360	360;
	9002	6629	0.00126	0.060099	0	0	0	0	0	0	1	-360	360;
	9002	6629	0.00135	0.06051	0	0	0	0	0	0	1	-360	360;
	10919	5136	0.00036	0.042469	0	0	0	0	1.052174	0	1	-360	360;
	1708	8682	0.001278	0.05335	0	0	0	0	0.973021	0	1	-360	360;
	9895	7914	0.000503	0.049251	0	0	0	0	0.977126	-0.14746	1	-360	360;
	9895	7914	0.000503	0.048177	0	0	0	0	0.977126	-0.14746	1	-360	360;
	2956	6660	0.000333	0.021103	0	0	0	0	0.972707	0	1	-360	360;
	13180	10140	0.00044	0.033083	0	0	0	0	0.977126	-0.043442	1	-360	360;
	5115	3596	0.000164	0.013858	0	0	0	0	0.9919	0	1	-360	360;
	7231	4303	0.000353	0.046754	0	0	0	0	0.937567	-0.262319	1	-360	360;
	2225	5238	0.000159	0.030001	0	0	0	0	0.937567	0.16528	1	-360	360;
	4782	3568	0.000273	0.016349	0	0	0	0	0.956938	0.01672	1	-360	360;
	1939	9543	0.000289	0.027718	0	0	0	0	1.036553	-0.169835	1	-360	360;
	8399	9418	0.00142	0.02765	0	0	0	0	0	0	1	-360	360;
	8399	9418	0.00145	0.02707	0	0	0	0	0	0	1	-360	360;
	2399	565	0.000469	0.030842	0	0	0	0	0.997643	0	1	-360	360;
	6503	565	0.00435	0.076291	0	0	0	0	0	0.432244	1	-360	360;
	6173	10574	0.000352	0.020852	0	0	0	0	1.03275	0	1	-360	360;
	6000	217	0.0003	0.01099	0	0	0	0	0	0	1	-360	360;
	6000	217	0.00029	0.0112	0	0	0	0	0	0	1	-360	360;
	7078	6129	0.000375	0.018768	0	0	0	0	1.028418	0	1	-360	360;
	4399	3110	0.000537	0.028553	0	0	0	0	0.991288	0	1	-360	360;
	4399	3110	0.000537	0.028002	0	0	0	0	0.988586	0	1	-360	360;
	2362	5841	0.000622	0.033296	0	0	0	0	0.956938	0	1	-360	360;
	2362	5841	0.000622	0.029475	0	0	0	0	0.956938	0.00962	1	-360	360;
	1194	566	0.00072	0.026242	0	0	0	0	0.957131	0	1	-360	360;
	1194	566	0.000644	0.029735	0	0	0	0	0.957131	0	1	-360	360;
	1788	1140	0.00314	0.168928	0	0	0	0	0	0.024998	1	-360	360;
	1788	1140	0.00307	0.172182	0	0	0	0	0	0.024998	1	-360	360;
	892	8601	0.000188	0.016506	0	0	0	0	0.973613	0	1	-360	360;
	10332	8762	0.000259	0.027074	0	0	0	0	1.002506	0	1	-360	360;
	4802	436	0.000437	0.02611	0	0	0	0	1.047869	0	1	-360	360;
	12785	4872	0.000446	0.025573	0	0	0	0	1.047869	0	1	-360	360;
	10774	3721	0.00032	0.02243	0	0	0	0	1.034211	0	1	-360	360;
	10838	9099	6.091942	8.676029	0	0	0	0	0	0	1	-360	360;
	8035	9099	6.444008	8.676035	0	0	0	0	0	0	1	-360	360;
	11812	9099	23.9876	40.776864	0	0	0	0	0	0	1	-360	360;
	12474	9099	35.24587	57.202482	0	0	0	0	0	0	1	-360	360;
	91	9099	45.97521	69.597102	0	0	0	0	0	0	1	-360	360;
	10349	9099	0.081469	0.775971	0	0	0	0	0	0	1	-360	360;
	12274	9099	32.03719	50.568181	0	0	0	0	0	0	1	-360	360;
	1789	9099	0.628884	2.485744	0	0	0	0	0	0	1	-360	360;
	10838	11009	12.31426	19.110329	0	0	0	0	0	0	1	-360	360;
	8035	11009	13.03616	19.128307	0	0	0	0	0	0	1	-360	360;
	10349	11009	0.197581	1.393719	0	0	0	0	0	0	1	-360	360;
	1789	11009	1.415165	4.81467	0	0	0	0	0	0	1	-360	360;
	7867	3932	0.13108	0.821607	0	0	0	0	0	0	1	-360	360;
	9172	3932	0.129079	0.839681	0	0	0	0	0	0	1	-360	360;
	9153	3773	0.069591	0.599051	0	0	0	0	0	0	1	-360	360;
	9153	1645	0.04983	0.471267	0	0	0	0	0	0	1	-360	360;
	9153	4860	0.00502	0.310928	0	0	0	0	0	0	1	-360	360;
	131	707	5.68843	10.297518	0	0	0	0	0	0	1	-360	360;
	12474	707	48.32851	56.749998	0	0	0	0	0	0	1	-360	360;
	11732	707	23.27066	28.824377	0	0	0	0	0	0	1	-360	360;
	3895	707	50.96074	55.43596	0	0	0	0	0	0	1	-360	360;
	131	9072	5.967355	9.180988	0	0	0	0	0	0	1	-360	360;
	12474	9072	44.27686	58.522723	0	0	0	0	0	0	1	-360	360;
	11732	9072	20.49132	30.954553	0	0	0	0	0	0	1	-360	360;
	3895	9072	43.12397	61.993787	0	0	0	0	0	0	1	-360	360;
	10335	5899	-0.04093	0.365413	0	0	0	0	0	0	1	-360	360;
	10335	1467	0.140533	1.138574	0	0	0	0	0	0	1	-360	360;
	10335	5005	0.151053	1.206163	0	0	0	0	0	0	1	-360	360;
	10335	5923	3.496607	8.23892	0	0	0	0	0	0	1	-360	360;
	10335	1752	3.21482	7.345567	0	0	0	0	0	0	1	-360	360;
	10335	644	0.136738	1.1482	0	0	0	0	0	0	1	-360	360;
	10335	2726	0.138539	1.094044	0	0	0	0	0	0	1	-360	360;
	10335	5344	0.077251	0.917036	0	0	0	0	0	0	1	-360	360;
	10335	2264	0.954778	3.73151	0	0	0	0	0	0	1	-360	360;
	10335	4342	0.745845	4.144945	0	0	0	0	0	0	1	-360	360;
	10335	1684	0.723892	4.195776	0	0	0	0	0	0	1	-360	360;
	12622	5899	0.47093	3.07252	0	0	0	0	0	0	1	-360	360;
	272	5899	0.45124	2.967768	0	0	0	0	0	0	1	-360	360;
	4883	5899	0.460083	2.907231	0	0	0	0	0	0	1	-360	360;
	7997	5899	2.286777	7.419422	0	0	0	0	0	0	1	-360	360;
	6414	5899	3.438636	9.067975	0	0	0	0	0	0	1	-360	360;
	12406	5899	0.885578	3.747314	0	0	0	0	0	0	1	-360	360;
	8314	5899	0.788967	4.246901	0	0	0	0	0	0	1	-360	360;
	9085	5899	1.804731	7.772521	0	0	0	0	0	0	1	-360	360;
	10280	7699	0.000716	0.032505	0	0	0	0	0.957364	0	1	-360	360;
	10280	7699	0.00056	0.04105	0	0	0	0	0.982537	0	1	-360	360;
	10280	7699	0.00057	0.040211	0	0	0	0	0.987993	0	1	-360	360;
	376	7699	10.24442	15.278311	0	0	0	0	0	0	1	-360	360;
	6032	7699	10.4376	14.953303	0	0	0	0	0	0	1	-360	360;
	3603	7699	31.55372	31.927684	0	0	0	0	0	0	1	-360	360;
	2174	7699	0.02494	0.501983	0	0	0	0	0	0	1	-360	360;
	7131	7699	2.458264	5.693595	0	0	0	0	0	0	1	-360	360;
	13017	7699	21.7686	29.266527	0	0	0	0	0	0	1	-360	360;
	6550	7699	0.260765	1.554236	0	0	0	0	0	0	1	-360	360;
	964	7699	0.41095	2.453926	0	0	0	0	0	0	1	-360	360;
	13282	7699	1.358636	3.804546	0	0	0	0	0	0	1	-360	360;
	12961	12536	0.00445	0.11669	0	0	0	0	0	0	1	-360	360;
	3288	260	0.02812	0.471488	0	0	0	0	0	0	1	-360	360;
	2382	260	0.109781	0.670062	0	0	0	0	0	0	1	-360	360;
	5152	260	0.11731	0.676136	0	0	0	0	0	0	1	-360	360;
	10382	11302	0.000443	0.034282	0	0	0	0	0.914188	0	1	-360	360;
	6823	11302	0.000407	0.0336	0	0	0	0	0.914919	0	1	-360	360;
	131	11302	0.05005	0.961922	0	0	0	0	0	0	1	-360	360;
	10838	11302	20.2438	55.378103	0	0	0	0	0	0	1	-360	360;
	8035	11302	20.42293	51.698341	0	0	0	0	0	0	1	-360	360;
	11812	11302	7.97376	20.509093	0	0	0	0	0	0	1	-360	360;
	12474	11302	1.402252	5.36219	0	0	0	0	0	0	1	-360	360;
	91	11302	4.842562	10.371072	0	0	0	0	0	0	1	-360	360;
	11732	11302	0.593884	2.602686	0	0	0	0	0	0	1	-360	360;
	12274	11302	9.390701	21.464872	0	0	0	0	0	0	1	-360	360;
	3895	11302	1.334504	5.801447	0	0	0	0	0	0	1	-360	360;
	12622	1467	1.754986	5.357133	0	0	0	0	0	0	1	-360	360;
	12622	5005	1.766898	5.923268	0	0	0	0	0	0	1	-360	360;
	12622	5923	1.832964	7.959834	0	0	0	0	0	0	1	-360	360;
	12622	1752	1.611565	6.707825	0	0	0	0	0	0	1	-360	360;
	12622	644	-0.019578	0.602618	0	0	0	0	0	0	1	-360	360;
	12622	2726	-0.018109	0.575138	0	0	0	0	0	0	1	-360	360;
	12622	5344	2.305263	5.580056	0	0	0	0	0	0	1	-360	360;
	12622	2264	0.324709	1.981648	0	0	0	0	0	0	1	-360	360;
	12622	4342	11.49377	18.997923	0	0	0	0	0	0	1	-360	360;
	12622	1684	10.16551	21.393357	0	0	0	0	0	0	1	-360	360;
	12622	5667	0.00083	0.06663	0	0	0	0	1.041579	0	1	-360	360;
	9636	7765	0.220599	1.426839	0	0	0	0	0	0	1	-360	360;
	13354	7765	0.215413	1.319979	0	0	0	0	0	0	1	-360	360;
	9506	7765	0.07932	0.491198	0	0	0	0	0	0	1	-360	360;
	1172	7765	0.05112	0.388202	0	0	0	0	0	0	1	-360	360;
	5569	7765	7.957025	11.110533	0	0	0	0	0	0	1	-360	360;
	8176	7765	8.10723	10.874171	0	0	0	0	0	0	1	-360	360;
	6825	7765	0.19505	0.651797	0	0	0	0	0	0	1	-360	360;
	10957	7765	0.273698	1.122666	0	0	0	0	0	0	1	-360	360;
	9636	9820	-0.016694	0.279194	0	0	0	0	0	0	1	-360	360;
	9636	9820	0.00066	0.04694	0	0	0	0	1.071156	0	1	-360	360;
	9636	7733	0.262431	2.115858	0	0	0	0	0	0	1	-360	360;
	9636	3193	0.251911	2.002563	0	0	0	0	0	0	1	-360	360;
	9636	6386	0.086669	0.634127	0	0	0	0	0	0	1	-360	360;
	13354	9820	-0.018471	0.253657	0	0	0	0	0	0	1	-360	360;
	13354	9820	0.000298	0.012803	0	0	0	0	1.148319	0	1	-360	360;
	13354	7733	0.307389	1.800831	0	0	0	0	0	0	1	-360	360;
	13354	3193	0.307181	1.800693	0	0	0	0	0	0	1	-360	360;
	13354	6386	0.083082	0.659398	0	0	0	0	0	0	1	-360	360;
	9506	9820	2.852479	6.564462	0	0	0	0	0	0	1	-360	360;
	1172	9820	0.610723	2.391322	0	0	0	0	0	0	1	-360	360;
	5569	9820	0.796095	2.828719	0	0	0	0	0	0	1	-360	360;
	8176	9820	0.778781	2.619215	0	0	0	0	0	0	1	-360	360;
	6825	9820	5.253512	8.690912	0	0	0	0	0	0	1	-360	360;
	10957	9820	0.14057	0.924359	0	0	0	0	0	0	1	-360	360;
	622	7915	0.00012	0.01406	0	0	0	0	0	0.41413	1	-360	360;
	9506	3170	0.00012	0.013772	0	0	0	0	0	0.41413	1	-360	360;
	9506	7733	34.83864	37.361504	0	0	0	0	0	0	1	-360	360;
	9506	3193	31.4626	43.434901	0	0	0	0	0	0	1	-360	360;
	9506	6386	9.837258	14.382266	0	0	0	0	0	0	1	-360	360;
	2872	13213	0.124359	1.222541	0	0	0	0	0	0	1	-360	360;
	7278	13213	0.121719	1.24562	0	0	0	0	0	0	1	-360	360;
	2872	10921	0.11651	1.13624	0	0	0	0	0	0	1	-360	360;
	7278	10921	0.13205	1.113513	0	0	0	0	0	0	1	-360	360;
	6362	1167	0.000447	0.03387	0	0	0	0	0.987247	0	1	-360	360;
	6362	3932	0.188878	0.872161	0	0	0	0	0	0	1	-360	360;
	6362	10699	0.000399	0.010683	0	0	0	0	1.015002	0	1	-360	360;
	13454	3746	0.01902	0.177292	0	0	0	0	0	0	1	-360	360;
	13454	12363	0.079771	0.509993	0	0	0	0	0	0	1	-360	360;
	13454	5547	0.00919	0.137159	0	0	0	0	0	0	1	-360	360;
	131	13629	10.8518	37.942518	0	0	0	0	0	0	1	-360	360;
	131	8267	20.41343	54.805408	0	0	0	0	0	0	1	-360	360;
	131	4147	20.30263	51.706374	0	0	0	0	0	0	1	-360	360;
	131	11448	0.787812	4.341828	0	0	0	0	0	0	1	-360	360;
	131	2359	0.590422	4.326316	0	0	0	0	0	0	1	-360	360;
	131	366	0.712119	2.758379	0	0	0	0	0	0	1	-360	360;
	131	1535	0.891274	4.745083	0	0	0	0	0	0	1	-360	360;
	131	5183	27.32618	53.325484	0	0	0	0	0	0	1	-360	360;
	131	7014	0.072382	1.414889	0	0	0	0	0	0	1	-360	360;
	131	12720	0.070083	1.331163	0	0	0	0	0	0	1	-360	360;
	131	9401	0.02163	0.743629	0	0	0	0	0	0	1	-360	360;
	131	2256	3.623961	20.259002	0	0	0	0	0	0	1	-360	360;
	131	1566	0.851039	4.324031	0	0	0	0	0	0	1	-360	360;
	131	5754	1.040443	4.789612	0	0	0	0	0	0	1	-360	360;
	1172	7733	8.633058	14.86384	0	0	0	0	0	0	1	-360	360;
	5569	7733	0.05238	0.529835	0	0	0	0	0	0	1	-360	360;
	8176	7733	0.05337	0.518554	0	0	0	0	0	0	1	-360	360;
	6825	7733	56.80165	51.109504	0	0	0	0	0	0	1	-360	360;
	10957	7733	2.452686	6.967768	0	0	0	0	0	0	1	-360	360;
	1172	3193	8.052273	16.496276	0	0	0	0	0	0	1	-360	360;
	5569	3193	0.04414	0.60938	0	0	0	0	0	0	1	-360	360;
	8176	3193	0.0432	0.620868	0	0	0	0	0	0	1	-360	360;
	6825	3193	47.16736	58.301648	0	0	0	0	0	0	1	-360	360;
	10957	3193	2.566529	6.851652	0	0	0	0	0	0	1	-360	360;
	10838	13629	21.60319	36.731304	0	0	0	0	0	0	1	-360	360;
	10838	8267	38.25416	49.701504	0	0	0	0	0	0	1	-360	360;
	10838	4147	39.65235	49.572723	0	0	0	0	0	0	1	-360	360;
	10838	11448	13.37881	27.533241	0	0	0	0	0	0	1	-360	360;
	10838	2359	11.70014	34.83726	0	0	0	0	0	0	1	-360	360;
	10838	366	0.03578	2.06572	0	0	0	0	0	0	1	-360	360;
	10838	9823	44.92382	57.412053	0	0	0	0	0	0	1	-360	360;
	10838	6896	5.792867	13.966068	0	0	0	0	0	0	1	-360	360;
	10838	10979	5.68705	12.979224	0	0	0	0	0	0	1	-360	360;
	10838	1535	23.49307	56.524238	0	0	0	0	0	0	1	-360	360;
	10838	5183	0.817382	3.615513	0	0	0	0	0	0	1	-360	360;
	10838	7014	1.090651	5.780887	0	0	0	0	0	0	1	-360	360;
	10838	12720	1.04626	5.209279	0	0	0	0	0	0	1	-360	360;
	10838	7916	0.644758	1.803532	0	0	0	0	0	0	1	-360	360;
	10838	3743	0.580852	2.069945	0	0	0	0	0	0	1	-360	360;
	10838	9401	15.32341	41.617723	0	0	0	0	0	0	1	-360	360;
	10838	4717	37.75623	69.158594	0	0	0	0	0	0	1	-360	360;
	10838	2256	4.031925	12.920362	0	0	0	0	0	0	1	-360	360;
	8035	13629	19.41759	39.081025	0	0	0	0	0	0	1	-360	360;
	8035	8267	38.13227	52.87396	0	0	0	0	0	0	1	-360	360;
	8035	4147	39.55609	52.783939	0	0	0	0	0	0	1	-360	360;
	8035	11448	13.35596	29.188359	0	0	0	0	0	0	1	-360	360;
	8035	2359	13.38227	31.89335	0	0	0	0	0	0	1	-360	360;
	8035	366	0.06002	1.808379	0	0	0	0	0	0	1	-360	360;
	8035	9823	48.46122	56.359411	0	0	0	0	0	0	1	-360	360;
	8035	6896	6.294598	13.758307	0	0	0	0	0	0	1	-360	360;
	8035	10979	6.185319	14.074795	0	0	0	0	0	0	1	-360	360;
	8035	1535	21.89474	63.064402	0	0	0	0	0	0	1	-360	360;
	8035	5183	0.750554	3.860318	0	0	0	0	0	0	1	-360	360;
	8035	7014	1.114889	6.088781	0	0	0	0	0	0	1	-360	360;
	8035	12720	1.069114	5.491413	0	0	0	0	0	0	1	-360	360;
	8035	7916	0.653109	1.929778	0	0	0	0	0	0	1	-360	360;
	8035	3743	0.672888	1.912604	0	0	0	0	0	0	1	-360	360;
	8035	9401	18.13504	36.319253	0	0	0	0	0	0	1	-360	360;
	8035	4717	40.86842	68.019379	0	0	0	0	0	0	1	-360	360;
	8035	2256	4.398892	12.729917	0	0	0	0	0	0	1	-360	360;
	8653	3932	0.0384	0.48786	0	0	0	0	0	0	1	-360	360;
	272	1467	0.17575	1.589628	0	0	0	0	0	0	1	-360	360;
	272	5005	0.16178	1.59079	0	0	0	0	0	0	1	-360	360;
	272	5923	22.34695	35.318561	0	0	0	0	0	0	1	-360	360;
	272	1752	20.61288	31.520779	0	0	0	0	0	0	1	-360	360;
	272	644	1.071953	3.96018	0	0	0	0	0	0	1	-360	360;
	272	2726	1.073546	3.773961	0	0	0	0	0	0	1	-360	360;
	272	5344	1.477701	3.981233	0	0	0	0	0	0	1	-360	360;
	272	2264	5.635665	13.348342	0	0	0	0	0	0	1	-360	360;
	272	4342	13.41828	27.828945	0	0	0	0	0	0	1	-360	360;
	272	1684	13.04432	28.183522	0	0	0	0	0	0	1	-360	360;
	4883	1467	0.17575	1.589628	0	0	0	0	0	0	1	-360	360;
	4883	5005	0.161789	1.590785	0	0	0	0	0	0	1	-360	360;
	4883	5923	22.34695	35.318561	0	0	0	0	0	0	1	-360	360;
	4883	1752	20.61288	31.520779	0	0	0	0	0	0	1	-360	360;
	4883	644	1.071953	3.96018	0	0	0	0	0	0	1	-360	360;
	4883	2726	1.073546	3.773961	0	0	0	0	0	0	1	-360	360;
	4883	5344	1.477701	3.981233	0	0	0	0	0	0	1	-360	360;
	4883	2264	5.635665	13.348342	0	0	0	0	0	0	1	-360	360;
	4883	4342	13.41828	27.828945	0	0	0	0	0	0	1	-360	360;
	4883	1684	13.04432	28.183522	0	0	0	0	0	0	1	-360	360;
	7997	1467	7.514876	17.076863	0	0	0	0	0	0	1	-360	360;
	6414	1467	10.18967	19.679343	0	0	0	0	0	0	1	-360	360;
	12406	1467	1.403533	4.891322	0	0	0	0	0	0	1	-360	360;
	8314	1467	1.431054	4.791528	0	0	0	0	0	0	1	-360	360;
	9085	1467	7.830992	14.143185	0	0	0	0	0	0	1	-360	360;
	7997	5005	9.209711	15.822724	0	0	0	0	0	0	1	-360	360;
	6414	5005	13.04525	19.283054	0	0	0	0	0	0	1	-360	360;
	12406	5005	1.367066	5.515496	0	0	0	0	0	0	1	-360	360;
	8314	5005	1.339153	5.62376	0	0	0	0	0	0	1	-360	360;
	9085	5005	7.23719	17.339673	0	0	0	0	0	0	1	-360	360;
	3464	1758	0.00696	0.20656	0	0	0	0	0	0	1	-360	360;
	2872	1758	0.085711	0.903781	0	0	0	0	0	0	1	-360	360;
	7278	1758	0.09713	0.885702	0	0	0	0	0	0	1	-360	360;
	1172	6386	2.771122	4.936011	0	0	0	0	0	0	1	-360	360;
	2872	9140	0.296032	1.447438	0	0	0	0	0	0	1	-360	360;
	2872	141	0.281413	1.320152	0	0	0	0	0	0	1	-360	360;
	7278	9140	0.307202	1.385873	0	0	0	0	0	0	1	-360	360;
	7278	141	0.25536	1.463643	0	0	0	0	0	0	1	-360	360;
	1550	9140	0.00046	0.03669	0	0	0	0	0.902975	0	1	-360	360;
	10000	141	0.000808	0.04562	0	0	0	0	0.895125	0	1	-360	360;
	10000	141	0.000438	0.03862	0	0	0	0	0.879211	0	1	-360	360;
	3288	4180	0.037081	0.480186	0	0	0	0	0	0	1	-360	360;
	2382	4180	0.074531	0.583946	0	0	0	0	0	0	1	-360	360;
	5152	4180	0.08	0.590186	0	0	0	0	0	0	1	-360	360;
	5912	11734	0.000127	0.009835	0	0	0	0	1.034606	0	1	-360	360;
	7454	3986	0.10579	0.434903	0	0	0	0	0	0	1	-360	360;
	7454	12363	0.09937	0.415727	0	0	0	0	0	0	1	-360	360;
	2586	3986	0.095997	0.482181	0	0	0	0	0	0	1	-360	360;
	2586	12363	0.086717	0.480388	0	0	0	0	0	0	1	-360	360;
	5380	4199	0.00013	0.011536	0	0	0	0	1.088419	0	1	-360	360;
	7997	5923	0.873905	3.992149	0	0	0	0	0	0	1	-360	360;
	6414	5923	0.120211	1.173533	0	0	0	0	0	0	1	-360	360;
	12406	5923	7.082025	15.591937	0	0	0	0	0	0	1	-360	360;
	8314	5923	7.220868	15.273755	0	0	0	0	0	0	1	-360	360;
	9085	5923	0.872831	3.28905	0	0	0	0	0	0	1	-360	360;
	7997	1752	0.921508	3.094834	0	0	0	0	0	0	1	-360	360;
	6414	1752	0.129919	0.960764	0	0	0	0	0	0	1	-360	360;
	12406	1752	6.142769	14.777069	0	0	0	0	0	0	1	-360	360;
	8314	1752	6.017355	15.066946	0	0	0	0	0	0	1	-360	360;
	9085	1752	0.694463	3.375206	0	0	0	0	0	0	1	-360	360;
	7997	644	7.254848	17.556091	0	0	0	0	0	0	1	-360	360;
	7997	2726	6.944598	15.819948	0	0	0	0	0	0	1	-360	360;
	7997	5344	0.962188	3.649169	0	0	0	0	0	0	1	-360	360;
	7997	2264	2.525	6.52029	0	0	0	0	0	0	1	-360	360;
	7997	4342	0.00154	0.39572	0	0	0	0	0	0	1	-360	360;
	7997	1684	0.0014	0.384758	0	0	0	0	0	0	1	-360	360;
	5569	6386	0.590782	1.406718	0	0	0	0	0	0	1	-360	360;
	8176	6386	0.526337	1.594252	0	0	0	0	0	0	1	-360	360;
	11812	13629	0.836859	5.618595	0	0	0	0	0	0	1	-360	360;
	12474	13629	1.237665	7.884711	0	0	0	0	0	0	1	-360	360;
	11732	13629	3.449174	20.571071	0	0	0	0	0	0	1	-360	360;
	10349	13629	5.302273	16.162393	0	0	0	0	0	0	1	-360	360;
	12274	13629	0.789525	4.620455	0	0	0	0	0	0	1	-360	360;
	1789	13629	2.372521	10.507644	0	0	0	0	0	0	1	-360	360;
	2123	4147	0.000495	0.033881	0	0	0	0	0.892514	0	1	-360	360;
	11812	8267	2.23905	7.164463	0	0	0	0	0	0	1	-360	360;
	12474	8267	3.327273	9.634503	0	0	0	0	0	0	1	-360	360;
	11732	8267	8.134091	27.90289	0	0	0	0	0	0	1	-360	360;
	10349	8267	10.55806	23.307847	0	0	0	0	0	0	1	-360	360;
	12274	8267	1.718843	7.201859	0	0	0	0	0	0	1	-360	360;
	1789	8267	4.571488	16.921905	0	0	0	0	0	0	1	-360	360;
	11812	4147	1.894649	7.933677	0	0	0	0	0	0	1	-360	360;
	12474	4147	3.133678	10.68802	0	0	0	0	0	0	1	-360	360;
	11732	4147	8.769008	26.754135	0	0	0	0	0	0	1	-360	360;
	10349	4147	11.85517	21.464873	0	0	0	0	0	0	1	-360	360;
	12274	4147	2.005517	6.368182	0	0	0	0	0	0	1	-360	360;
	1789	4147	5.553718	14.366735	0	0	0	0	0	0	1	-360	360;
	7042	359	0.000206	0.020261	0	0	0	0	1.029618	0	1	-360	360;
	4338	1990	9.6e-05	0.010041	0	0	0	0	1.024823	0	1	-360	360;
	6414	644	4.540289	12.081814	0	0	0	0	0	0	1	-360	360;
	12406	644	0.14887	1.547789	0	0	0	0	0	0	1	-360	360;
	8314	644	0.145641	1.433141	0	0	0	0	0	0	1	-360	360;
	9085	644	7.255165	15.148962	0	0	0	0	0	0	1	-360	360;
	6414	2726	4.918802	10.649998	0	0	0	0	0	0	1	-360	360;
	12406	2726	0.170281	1.312108	0	0	0	0	0	0	1	-360	360;
	8314	2726	0.17349	1.284174	0	0	0	0	0	0	1	-360	360;
	9085	2726	7.658058	13.540087	0	0	0	0	0	0	1	-360	360;
	10367	4860	1.803947	6.514197	0	0	0	0	0	0	1	-360	360;
	10367	8960	1.414404	5.42687	0	0	0	0	0	0	1	-360	360;
	10367	1779	0.097091	1.338712	0	0	0	0	0	0	1	-360	360;
	10367	6012	0.069813	0.61338	0	0	0	0	0	0	1	-360	360;
	9178	4860	1.656925	6.386704	0	0	0	0	0	0	1	-360	360;
	9178	8960	1.503047	5.113989	0	0	0	0	0	0	1	-360	360;
	9178	1779	0.107382	1.212396	0	0	0	0	0	0	1	-360	360;
	9178	6012	0.080478	0.53394	0	0	0	0	0	0	1	-360	360;
	6414	5344	1.566157	4.554339	0	0	0	0	0	0	1	-360	360;
	12406	5344	3.21405	6.727273	0	0	0	0	0	0	1	-360	360;
	8314	5344	2.86343	7.62438	0	0	0	0	0	0	1	-360	360;
	9085	5344	0.784091	3.665289	0	0	0	0	0	0	1	-360	360;
	11812	11448	-0.013061	0.68473	0	0	0	0	0	0	1	-360	360;
	11812	2359	0.02627	0.92126	0	0	0	0	0	0	1	-360	360;
	11812	366	5.225831	17.344184	0	0	0	0	0	0	1	-360	360;
	11812	917	37.01108	55.722988	0	0	0	0	0	0	1	-360	360;
	11812	4995	38.53255	55.73961	0	0	0	0	0	0	1	-360	360;
	11812	1535	0.00421	0.609432	0	0	0	0	0	0	1	-360	360;
	11812	5183	3.373338	12.718144	0	0	0	0	0	0	1	-360	360;
	11812	7014	5.841413	14.226459	0	0	0	0	0	0	1	-360	360;
	11812	12720	4.851247	14.847647	0	0	0	0	0	0	1	-360	360;
	11812	7916	15.98753	46.096945	0	0	0	0	0	0	1	-360	360;
	11812	3743	15.83449	47.580335	0	0	0	0	0	0	1	-360	360;
	11812	9401	5.167521	17.612876	0	0	0	0	0	0	1	-360	360;
	11812	2256	0.170402	3.131509	0	0	0	0	0	0	1	-360	360;
	11812	7691	13.13227	24.978533	0	0	0	0	0	0	1	-360	360;
	11812	6065	11.55748	21.738223	0	0	0	0	0	0	1	-360	360;
	12474	11448	0.071828	1.118006	0	0	0	0	0	0	1	-360	360;
	12474	2359	0.01624	0.986219	0	0	0	0	0	0	1	-360	360;
	12474	366	1.180471	4.696329	0	0	0	0	0	0	1	-360	360;
	12474	1535	0.01296	0.834834	0	0	0	0	0	0	1	-360	360;
	12474	5183	4.962535	20.238919	0	0	0	0	0	0	1	-360	360;
	12474	7014	0.872438	4.821676	0	0	0	0	0	0	1	-360	360;
	12474	12720	0.801593	4.526524	0	0	0	0	0	0	1	-360	360;
	12474	7916	20.20014	50.806099	0	0	0	0	0	0	1	-360	360;
	12474	3743	23.16066	50.423823	0	0	0	0	0	0	1	-360	360;
	12474	9401	0.966621	4.164474	0	0	0	0	0	0	1	-360	360;
	12474	2256	0.448781	4.337397	0	0	0	0	0	0	1	-360	360;
	12474	1566	11.58657	23.771465	0	0	0	0	0	0	1	-360	360;
	12474	5754	13.00554	26.333102	0	0	0	0	0	0	1	-360	360;
	12474	7691	17.94044	35.608035	0	0	0	0	0	0	1	-360	360;
	12474	6065	15.18213	32.262467	0	0	0	0	0	0	1	-360	360;
	6414	2264	0.430249	2.466066	0	0	0	0	0	0	1	-360	360;
	6414	4342	0.0683	0.916828	0	0	0	0	0	0	1	-360	360;
	6414	1684	0.06612	0.843352	0	0	0	0	0	0	1	-360	360;
	12406	2264	1.444298	4.757438	0	0	0	0	0	0	1	-360	360;
	8314	2264	1.472603	4.660331	0	0	0	0	0	0	1	-360	360;
	9085	2264	2.284917	6.055578	0	0	0	0	0	0	1	-360	360;
	722	10991	0.000504	0.031781	0	0	0	0	1.041866	0	1	-360	360;
	722	4860	0.086683	1.502285	0	0	0	0	0	0	1	-360	360;
	722	8960	0.05572	1.389266	0	0	0	0	0	0	1	-360	360;
	722	1779	0.40606	1.910249	0	0	0	0	0	0	1	-360	360;
	722	6012	0.0265	0.438137	0	0	0	0	0	0	1	-360	360;
	11650	9023	0.000669	0.032368	0	0	0	0	1.076608	0	1	-360	360;
	12406	4342	16.48554	31.878093	0	0	0	0	0	0	1	-360	360;
	8314	4342	18.68347	31.239671	0	0	0	0	0	0	1	-360	360;
	9085	4342	-0.006984	0.380578	0	0	0	0	0	0	1	-360	360;
	12406	1684	19.27748	29.774787	0	0	0	0	0	0	1	-360	360;
	8314	1684	19.64132	29.142564	0	0	0	0	0	0	1	-360	360;
	9085	1684	-0.0075	0.354442	0	0	0	0	0	0	1	-360	360;
	478	4860	0.979504	4.164463	0	0	0	0	0	0	1	-360	360;
	8529	4860	0.241095	2.041488	0	0	0	0	0	0	1	-360	360;
	5951	4860	0.239525	2.084091	0	0	0	0	0	0	1	-360	360;
	478	8960	0.733492	3.60186	0	0	0	0	0	0	1	-360	360;
	8529	8960	0.176539	1.602479	0	0	0	0	0	0	1	-360	360;
	5951	8960	0.203141	1.572934	0	0	0	0	0	0	1	-360	360;
	91	11448	3.658471	11.068596	0	0	0	0	0	0	1	-360	360;
	11732	11448	0.10706	1.950682	0	0	0	0	0	0	1	-360	360;
	10349	11448	10.14339	24.185954	0	0	0	0	0	0	1	-360	360;
	12274	11448	0.07038	1.132707	0	0	0	0	0	0	1	-360	360;
	3895	11448	3.802686	12.010955	0	0	0	0	0	0	1	-360	360;
	1789	11448	3.583058	17.321278	0	0	0	0	0	0	1	-360	360;
	91	2359	2.938636	13.075416	0	0	0	0	0	0	1	-360	360;
	11732	2359	0.00024	2.372521	0	0	0	0	0	0	1	-360	360;
	10349	2359	9.23781	30.750004	0	0	0	0	0	0	1	-360	360;
	12274	2359	0.041469	1.048347	0	0	0	0	0	0	1	-360	360;
	3895	2359	3.576446	11.947316	0	0	0	0	0	0	1	-360	360;
	1789	2359	4.138017	18.151449	0	0	0	0	0	0	1	-360	360;
	478	3932	0.107147	0.58482	0	0	0	0	0	0	1	-360	360;
	478	1779	7.159972	13.527699	0	0	0	0	0	0	1	-360	360;
	478	6012	1.77874	5.115028	0	0	0	0	0	0	1	-360	360;
	10349	8608	7.195455	16.568181	0	0	0	0	0	0	1	-360	360;
	1789	8608	40.97313	63.811984	0	0	0	0	0	0	1	-360	360;
	10349	4055	7.423554	18.762807	0	0	0	0	0	0	1	-360	360;
	1789	4055	42.39256	65.737609	0	0	0	0	0	0	1	-360	360;
	91	366	-0.016674	0.278698	0	0	0	0	0	0	1	-360	360;
	91	1535	5.349861	16.166202	0	0	0	0	0	0	1	-360	360;
	91	5183	9.134349	25.279081	0	0	0	0	0	0	1	-360	360;
	91	7014	0.197001	1.168768	0	0	0	0	0	0	1	-360	360;
	91	12720	0.18955	1.05277	0	0	0	0	0	0	1	-360	360;
	91	7916	6.030263	16.006928	0	0	0	0	0	0	1	-360	360;
	91	3743	5.975346	16.528394	0	0	0	0	0	0	1	-360	360;
	91	9401	3.107895	9.123269	0	0	0	0	0	0	1	-360	360;
	91	2256	11.81025	37.443907	0	0	0	0	0	0	1	-360	360;
	91	1566	27.31856	49.591408	0	0	0	0	0	0	1	-360	360;
	91	5754	34.37812	55.136426	0	0	0	0	0	0	1	-360	360;
	11732	366	0.231281	1.502872	0	0	0	0	0	0	1	-360	360;
	12274	366	7.048967	19.24607	0	0	0	0	0	0	1	-360	360;
	3895	366	0.229277	1.235537	0	0	0	0	0	0	1	-360	360;
	1789	366	2.544835	7.532231	0	0	0	0	0	0	1	-360	360;
	10349	8156	35.47521	51.481403	0	0	0	0	0	0	1	-360	360;
	10349	6030	34.64256	52.336781	0	0	0	0	0	0	1	-360	360;
	6825	6386	15.75277	20.093496	0	0	0	0	0	0	1	-360	360;
	9567	2772	0.000283	0.012832	0	0	0	0	1.014459	0	1	-360	360;
	12271	2801	0.000292	0.01188	0	0	0	0	0.984333	0	1	-360	360;
	10446	917	0.000398	0.034511	0	0	0	0	0.921465	0	1	-360	360;
	12274	917	33.09298	47.6281	0	0	0	0	0	0	1	-360	360;
	12274	4995	34.44215	47.623971	0	0	0	0	0	0	1	-360	360;
	11732	1535	0.161039	2.815236	0	0	0	0	0	0	1	-360	360;
	11732	5183	9.489612	27.308865	0	0	0	0	0	0	1	-360	360;
	11732	7014	0.246808	1.61939	0	0	0	0	0	0	1	-360	360;
	11732	12720	0.228082	1.521121	0	0	0	0	0	0	1	-360	360;
	11732	7916	18.28047	42.772855	0	0	0	0	0	0	1	-360	360;
	11732	3743	18.10111	44.135045	0	0	0	0	0	0	1	-360	360;
	11732	9401	0.331281	2.194459	0	0	0	0	0	0	1	-360	360;
	11732	2256	0.752078	9.020083	0	0	0	0	0	0	1	-360	360;
	11732	1566	4.985803	12.659968	0	0	0	0	0	0	1	-360	360;
	11732	5754	5.581648	14.030467	0	0	0	0	0	0	1	-360	360;
	9492	6495	0.000476	0.03161	0	0	0	0	0.909467	0	1	-360	360;
	9492	9067	0.000457	0.030901	0	0	0	0	0.908902	0	1	-360	360;
	10349	12043	1.305041	5.872521	0	0	0	0	0	0	1	-360	360;
	1789	12043	12.81343	30.764465	0	0	0	0	0	0	1	-360	360;
	10349	9823	0.657355	3.229132	0	0	0	0	0	0	1	-360	360;
	1789	9823	6.467975	16.903509	0	0	0	0	0	0	1	-360	360;
	10349	6896	-0.017839	0.427271	0	0	0	0	0	0	1	-360	360;
	10349	10979	-0.020298	0.420457	0	0	0	0	0	0	1	-360	360;
	10349	5183	0.05596	1.32043	0	0	0	0	0	0	1	-360	360;
	10349	7916	0.430997	2.378602	0	0	0	0	0	0	1	-360	360;
	10349	3743	0.443573	2.357341	0	0	0	0	0	0	1	-360	360;
	10349	4717	0.073677	2.231025	0	0	0	0	0	0	1	-360	360;
	10349	2256	0.927909	4.907895	0	0	0	0	0	0	1	-360	360;
	10349	7691	42.82895	57.943891	0	0	0	0	0	0	1	-360	360;
	10349	6065	36.44598	52.531845	0	0	0	0	0	0	1	-360	360;
	7108	10979	0.00037	0.03817	0	0	0	0	0.98054	0	1	-360	360;
	7108	6896	0.000309	0.035711	0	0	0	0	0.977048	0	1	-360	360;
	1789	6896	0.529731	3.305372	0	0	0	0	0	0	1	-360	360;
	1789	10979	0.54188	3.249793	0	0	0	0	0	0	1	-360	360;
	12274	1535	-0.008017	0.233843	0	0	0	0	0	0	1	-360	360;
	12274	5183	5.680748	16.603181	0	0	0	0	0	0	1	-360	360;
	12274	7014	6.552285	16.597649	0	0	0	0	0	0	1	-360	360;
	12274	12720	5.444182	17.322713	0	0	0	0	0	0	1	-360	360;
	12274	7916	24.32756	59.210516	0	0	0	0	0	0	1	-360	360;
	12274	3743	24.09903	61.117023	0	0	0	0	0	0	1	-360	360;
	12274	9401	5.625068	20.064407	0	0	0	0	0	0	1	-360	360;
	12274	2256	0.456198	4.138435	0	0	0	0	0	0	1	-360	360;
	12274	7691	11.52632	21.784622	0	0	0	0	0	0	1	-360	360;
	12274	6065	10.14543	18.959839	0	0	0	0	0	0	1	-360	360;
	12274	1535	0.000476	0.031631	0	0	0	0	0.889435	0	1	-360	360;
	3895	1535	5.521694	14.730582	0	0	0	0	0	0	1	-360	360;
	1789	1535	19.34525	41.508264	0	0	0	0	0	0	1	-360	360;
	3895	5183	32.73967	68.105374	0	0	0	0	0	0	1	-360	360;
	1789	5183	-0.083244	0.946591	0	0	0	0	0	0	1	-360	360;
	6656	7014	0.000447	0.035271	0	0	0	0	0.895342	0	1	-360	360;
	3895	7014	-0.015103	0.337521	0	0	0	0	0	0	1	-360	360;
	3895	12720	-0.012686	0.286488	0	0	0	0	0	0	1	-360	360;
	3895	7916	27.97992	45.335871	0	0	0	0	0	0	1	-360	360;
	3895	3743	28.86773	44.978518	0	0	0	0	0	0	1	-360	360;
	3895	9401	0.994044	4.152008	0	0	0	0	0	0	1	-360	360;
	3895	2256	19.05609	41.630194	0	0	0	0	0	0	1	-360	360;
	3895	1566	11.95776	23.687676	0	0	0	0	0	0	1	-360	360;
	3895	5754	11.73892	30.398893	0	0	0	0	0	0	1	-360	360;
	1789	7014	10.39773	25.376037	0	0	0	0	0	0	1	-360	360;
	1789	12720	9.457438	23.791321	0	0	0	0	0	0	1	-360	360;
	1789	7916	0.000409	0.036341	0	0	0	0	0.87261	0	1	-360	360;
	1789	7916	-0.006942	0.494855	0	0	0	0	0	0	1	-360	360;
	1789	3743	-0.008161	0.491033	0	0	0	0	0	0	1	-360	360;
	1789	4717	4.701801	16.907201	0	0	0	0	0	0	1	-360	360;
	1789	2256	0.18313	2.973269	0	0	0	0	0	0	1	-360	360;
	1789	7691	30.6392	44.945983	0	0	0	0	0	0	1	-360	360;
	1789	6065	27.01524	39.081706	0	0	0	0	0	0	1	-360	360;
	9085	12614	0.000757	0.064121	0	0	0	0	1.085043	0	1	-360	360;
	8529	1779	2.158264	7.671693	0	0	0	0	0	0	1	-360	360;
	5951	1779	2.128512	7.829339	0	0	0	0	0	0	1	-360	360;
	8529	6012	0.536593	2.637604	0	0	0	0	0	0	1	-360	360;
	5951	6012	0.529487	2.445638	0	0	0	0	0	0	1	-360	360;
	10957	13306	0.00039	0.016602	0	0	0	0	0.960225	-0.068985	1	-360	360;
	10957	6386	0.856787	2.082133	0	0	0	0	0	0	1	-360	360;
	12058	3119	0.000515	0.03175	0	0	0	0	0.996485	0	1	-360	360;
	12058	3119	0.000408	0.036729	0	0	0	0	0.996624	0	1	-360	360;
	3043	5297	0.00648	0.148407	0	0	0	0	0	0	1	-360	360;
	3635	5297	0.00055	0.038349	0	0	0	0	0.857944	0	1	-360	360;
	2376	4127	0.114331	0.894649	0	0	0	0	0	0	1	-360	360;
	2354	4127	3.931818	5.960743	0	0	0	0	0	0	1	-360	360;
	7451	4127	1.022872	2.477479	0	0	0	0	0	0	1	-360	360;
	2265	4127	1.145744	2.668595	0	0	0	0	0	0	1	-360	360;
	12109	4127	0.115521	0.652645	0	0	0	0	0	0	1	-360	360;
	2376	9057	0.342335	1.036053	0	0	0	0	0	0	1	-360	360;
	9661	11453	0.08144	0.487603	0	0	0	0	0	0	1	-360	360;
	10358	80	0.492025	1.533306	0	0	0	0	0	0	1	-360	360;
	3462	7437	0.00079	0.058401	0	0	0	0	0	0	1	-360	360;
	3462	7437	0.00077	0.05951	0	0	0	0	0	0	1	-360	360;
	2376	5802	0.000209	0.016646	0	0	0	0	0.960405	0	1	-360	360;
	2376	587	-0.020062	0.263393	0	0	0	0	0	0	1	-360	360;
	2376	7275	0.02611	0.389661	0	0	0	0	0	0	1	-360	360;
	2376	1611	0.02953	0.274197	0	0	0	0	0	0	1	-360	360;
	2376	587	0.000249	0.015313	0	0	0	0	0.963743	0	1	-360	360;
	2376	13460	0.000151	0.014993	0	0	0	0	1.037521	0	1	-360	360;
	6804	2393	0.01135	0.389938	0	0	0	0	0	0	1	-360	360;
	6804	5190	0.03467	0.4241	0	0	0	0	0	0	1	-360	360;
	10358	2314	0.000159	0.018964	0	0	0	0	0.954163	0	1	-360	360;
	10358	2314	-0.01126	0.214855	0	0	0	0	0	0	1	-360	360;
	10358	12360	0.20036	1.026385	0	0	0	0	0	0	1	-360	360;
	10358	13555	0.117071	0.511302	0	0	0	0	0	0	1	-360	360;
	5633	3742	0.093303	0.945014	0	0	0	0	0	0	1	-360	360;
	1025	3742	0.01297	0.327971	0	0	0	0	0	0	1	-360	360;
	3417	3742	0.05929	0.601787	0	0	0	0	0	0	1	-360	360;
	9661	6301	0.05013	0.919979	0	0	0	0	0	0	1	-360	360;
	9661	3085	0.0477	0.865888	0	0	0	0	0	0	1	-360	360;
	2354	3879	0.749308	2.45464	0	0	0	0	0	0	1	-360	360;
	2354	8543	2.902839	5.007549	0	0	0	0	0	0	1	-360	360;
	4172	2393	0.03282	0.444566	0	0	0	0	0	0	1	-360	360;
	11636	3742	0.186479	0.852913	0	0	0	0	0	0	1	-360	360;
	2948	3742	0.185979	0.818905	0	0	0	0	0	0	1	-360	360;
	10879	3742	1.207107	3.026033	0	0	0	0	0	0	1	-360	360;
	9048	3742	0.414669	1.672355	0	0	0	0	0	0	1	-360	360;
	7451	3879	0.129612	0.934834	0	0	0	0	0	0	1	-360	360;
	7451	8543	0.796053	2.21392	0	0	0	0	0	0	1	-360	360;
	2265	3879	0.136461	1.06849	0	0	0	0	0	0	1	-360	360;
	2265	8543	0.838366	2.528809	0	0	0	0	0	0	1	-360	360;
	1620	11840	0.000206	0.015885	0	0	0	0	1.065581	0	1	-360	360;
	11151	1910	0.000298	0.029295	0	0	0	0	1.085859	-0.169703	1	-360	360;
	6796	8189	0.01604	0.374051	0	0	0	0	0	0	1	-360	360;
	3026	8189	0.01635	0.36626	0	0	0	0	0	0	1	-360	360;
	4172	8189	0.024141	0.327376	0	0	0	0	0	0	1	-360	360;
	9156	8189	0.02194	0.458409	0	0	0	0	0	0	1	-360	360;
	2843	13196	0.00046	0.03512	0	0	0	0	0.998397	0	1	-360	360;
	2843	2600	0.000328	0.017373	0	0	0	0	1.008118	0	1	-360	360;
	4172	5190	0.00056	0.05575	0	0	0	0	0	0	1	-360	360;
	9661	3817	0.09537	0.774091	0	0	0	0	0	0	1	-360	360;
	9661	7035	0.325393	1.201922	0	0	0	0	0	0	1	-360	360;
	9661	11123	0.367459	1.167149	0	0	0	0	0	0	1	-360	360;
	9156	584	0.0008	0.04856	0	0	0	0	0	0	1	-360	360;
	1894	234	0.000151	0.010302	0	0	0	0	1.042652	0	1	-360	360;
	1894	2319	0.00033	0.038031	0	0	0	0	1.005085	0	1	-360	360;
	1894	2319	0.00034	0.037211	0	0	0	0	1.005257	0	1	-360	360;
	12109	3879	-0.016405	0.249504	0	0	0	0	0	0	1	-360	360;
	12109	8543	0.087209	0.621302	0	0	0	0	0	0	1	-360	360;
	12918	7095	0.15538	0.761901	0	0	0	0	0	0	1	-360	360;
	5993	7095	0.15969	0.749463	0	0	0	0	0	0	1	-360	360;
	10087	7095	1.428926	3.608884	0	0	0	0	0	0	1	-360	360;
	8361	7095	0.861797	2.342768	0	0	0	0	0	0	1	-360	360;
	9091	4168	0.119958	0.657978	0	0	0	0	0	0	1	-360	360;
	5557	9394	0.172789	1.348017	0	0	0	0	0	0	1	-360	360;
	3447	9394	0.00088	0.381756	0	0	0	0	0	0	1	-360	360;
	3260	9394	0.0061	0.418244	0	0	0	0	0	0	1	-360	360;
	6117	10372	0.05995	0.76688	0	0	0	0	0	0	1	-360	360;
	8691	10372	0.0653	0.750992	0	0	0	0	0	0	1	-360	360;
	2728	10372	-0.003347	0.292562	0	0	0	0	0	0	1	-360	360;
	4506	11358	1.83593	3.666115	0	0	0	0	0	0	1	-360	360;
	13547	11358	0.052409	0.497211	0	0	0	0	0	0	1	-360	360;
	1649	3636	0.436797	1.36533	0	0	0	0	0	0	1	-360	360;
	10400	3636	0.094151	0.896736	0	0	0	0	0	0	1	-360	360;
	1983	11136	0.00721	0.237881	0	0	0	0	0	0	1	-360	360;
	1983	7018	0.140762	0.856856	0	0	0	0	0	0	1	-360	360;
	12051	11136	0.05419	0.633158	0	0	0	0	0	0	1	-360	360;
	12051	7018	0.0291	0.36903	0	0	0	0	0	0	1	-360	360;
	4556	4168	0.143641	0.91531	0	0	0	0	0	0	1	-360	360;
	1723	4168	0.025969	0.445888	0	0	0	0	0	0	1	-360	360;
	1771	4168	0.270372	0.785723	0	0	0	0	0	0	1	-360	360;
	7818	4168	0.05175	0.49312	0	0	0	0	0	0	1	-360	360;
	7012	4168	0.08601	0.600496	0	0	0	0	0	0	1	-360	360;
	8784	4168	0.531591	1.712665	0	0	0	0	0	0	1	-360	360;
	10047	4168	0.43593	1.756797	0	0	0	0	0	0	1	-360	360;
	7835	6099	0.01926	0.358289	0	0	0	0	0	0	1	-360	360;
	10846	7216	0.01795	0.317397	0	0	0	0	0	0	1	-360	360;
	5557	228	0.00761	0.279277	0	0	0	0	0	0	1	-360	360;
	3447	228	0.260062	1.120476	0	0	0	0	0	0	1	-360	360;
	3260	228	0.158651	0.983822	0	0	0	0	0	0	1	-360	360;
	5557	6735	0.00395	0.366116	0	0	0	0	0	0	1	-360	360;
	8870	6735	0.083409	0.606219	0	0	0	0	0	0	1	-360	360;
	3447	6735	0.195669	1.120144	0	0	0	0	0	0	1	-360	360;
	3260	6735	0.104969	1.02188	0	0	0	0	0	0	1	-360	360;
	4570	6735	0.042831	0.74374	0	0	0	0	0	0	1	-360	360;
	4506	12086	0.000242	0.029698	0	0	0	0	0.945816	0.090914	1	-360	360;
	4506	10015	-0.003203	0.14994	0	0	0	0	0	0	1	-360	360;
	4506	10015	0.000451	0.034969	0	0	0	0	1.022789	0	1	-360	360;
	4506	10225	0.113213	0.498338	0	0	0	0	0	0	1	-360	360;
	9468	3227	0.00055	0.0343	0	0	0	0	0.917643	0	1	-360	360;
	5993	13243	0.000264	0.023375	0	0	0	0	1.049631	-0.088275	1	-360	360;
	12918	7494	0.01862	0.220201	0	0	0	0	0	0	1	-360	360;
	5993	7494	0.0168	0.250561	0	0	0	0	0	0	1	-360	360;
	1649	7018	0.202902	0.952701	0	0	0	0	0	0	1	-360	360;
	1649	13016	0.949862	3.139543	0	0	0	0	0	0	1	-360	360;
	1649	2365	0.00261	0.215422	0	0	0	0	0	0	1	-360	360;
	10904	6104	-0.02155	0.439897	0	0	0	0	0	0	1	-360	360;
	8870	6104	0.07045	1.003492	0	0	0	0	0	0	1	-360	360;
	10904	12768	0.00807	0.465263	0	0	0	0	0	0	1	-360	360;
	10904	9268	0.194723	1.065305	0	0	0	0	0	0	1	-360	360;
	10904	5095	0.7241	1.639128	0	0	0	0	0	0	1	-360	360;
	10904	7630	0.514127	1.973476	0	0	0	0	0	0	1	-360	360;
	10904	12875	0.06643	0.866482	0	0	0	0	0	0	1	-360	360;
	10904	5550	0.55937	2.088574	0	0	0	0	0	0	1	-360	360;
	10904	5687	0.522202	2.029502	0	0	0	0	0	0	1	-360	360;
	10904	11156	0.30876	1.923338	0	0	0	0	0	0	1	-360	360;
	4685	10914	0.06357	0.598203	0	0	0	0	0	0	1	-360	360;
	12475	10914	0.000445	0.035961	0	0	0	0	0.918248	0	1	-360	360;
	9641	10914	0.01179	0.332211	0	0	0	0	0	0	1	-360	360;
	1149	10914	0.02131	0.372955	0	0	0	0	0	0	1	-360	360;
	6597	10914	0.00274	0.381756	0	0	0	0	0	0	1	-360	360;
	10087	7494	0.293141	1.107872	0	0	0	0	0	0	1	-360	360;
	8361	7494	0.144281	0.82905	0	0	0	0	0	0	1	-360	360;
	3566	3503	0.00458	0.288905	0	0	0	0	0	0	1	-360	360;
	6522	462	0.365888	1.746611	0	0	0	0	0	0	1	-360	360;
	3081	462	1.44438	5.252272	0	0	0	0	0	0	1	-360	360;
	2602	462	0.059581	0.615888	0	0	0	0	0	0	1	-360	360;
	5257	462	3.709298	8.934297	0	0	0	0	0	0	1	-360	360;
	10026	462	0.140919	0.944236	0	0	0	0	0	0	1	-360	360;
	2201	239	0.0005	0.0343	0	0	0	0	0.884937	0	1	-360	360;
	13547	239	0.01962	0.216178	0	0	0	0	0	0	1	-360	360;
	8870	12768	0.398678	1.516364	0	0	0	0	0	0	1	-360	360;
	8870	1246	0.09755	0.839897	0	0	0	0	0	0	1	-360	360;
	4570	1246	0.499525	2.142562	0	0	0	0	0	0	1	-360	360;
	1872	6099	0.01537	0.348347	0	0	0	0	0	0	1	-360	360;
	5394	6099	0.01946	0.372727	0	0	0	0	0	0	1	-360	360;
	9641	9245	0.01679	0.350455	0	0	0	0	0	0	1	-360	360;
	1149	9245	0.0005	0.324174	0	0	0	0	0	0	1	-360	360;
	6597	9245	0.049469	0.557004	0	0	0	0	0	0	1	-360	360;
	1723	8592	0.0005	0.031271	0	0	0	0	0	0	1	-360	360;
	6522	8592	0.422748	2.328099	0	0	0	0	0	0	1	-360	360;
	3081	8592	1.022934	4.279959	0	0	0	0	0	0	1	-360	360;
	2602	8592	0.256612	1.738141	0	0	0	0	0	0	1	-360	360;
	5257	8592	2.300207	9.353099	0	0	0	0	0	0	1	-360	360;
	10026	8592	0.0157	0.997789	0	0	0	0	0	0	1	-360	360;
	8870	10093	0.022021	0.625496	0	0	0	0	0	0	1	-360	360;
	4570	10093	0.223988	1.463946	0	0	0	0	0	0	1	-360	360;
	6522	8936	-0.037934	0.385868	0	0	0	0	0	0	1	-360	360;
	3081	8936	2.334711	7.437397	0	0	0	0	0	0	1	-360	360;
	2602	8936	0.137849	1.103905	0	0	0	0	0	0	1	-360	360;
	5257	8936	5.726446	13.392563	0	0	0	0	0	0	1	-360	360;
	10026	8936	0.183531	1.402418	0	0	0	0	0	0	1	-360	360;
	12854	6522	3.1e-05	0.002643	0	0	0	0	0.993885	-0.110645	1	-360	360;
	6522	7527	2.374861	8.881579	0	0	0	0	0	0	1	-360	360;
	6522	11961	0.158643	1.604501	0	0	0	0	0	0	1	-360	360;
	6522	3731	0.204162	1.92777	0	0	0	0	0	0	1	-360	360;
	6522	12457	0.169439	2.010873	0	0	0	0	0	0	1	-360	360;
	3081	7527	-0.011343	0.25657	0	0	0	0	0	0	1	-360	360;
	3081	7527	0.000423	0.033381	0	0	0	0	0.876622	0	1	-360	360;
	2602	7527	6.533884	12.310123	0	0	0	0	0	0	1	-360	360;
	5257	7527	0.04751	0.576054	0	0	0	0	0	0	1	-360	360;
	10026	7527	0.21343	1.602582	0	0	0	0	0	0	1	-360	360;
	3081	11961	7.076177	15.615651	0	0	0	0	0	0	1	-360	360;
	3081	3731	6.591413	15.880881	0	0	0	0	0	0	1	-360	360;
	3081	12457	0.187507	2.209903	0	0	0	0	0	0	1	-360	360;
	1179	9396	0.0004	0.149419	0	0	0	0	0	0	1	-360	360;
	8361	9396	0.027469	0.320806	0	0	0	0	0	0	1	-360	360;
	10846	9396	0.00659	0.53155	0	0	0	0	0	0	1	-360	360;
	6598	8996	0.45281	1.969855	0	0	0	0	0	0	1	-360	360;
	13441	8996	0.561302	2.344628	0	0	0	0	0	0	1	-360	360;
	12362	8996	0.03438	0.839277	0	0	0	0	0	0	1	-360	360;
	1820	3742	-0.00126	0.238202	0	0	0	0	0	0	1	-360	360;
	1820	3742	0.000426	0.017209	0	0	0	0	1.002506	0	1	-360	360;
	1820	185	0.01628	0.291863	0	0	0	0	0	0	1	-360	360;
	10196	12543	-0.010145	0.227045	0	0	0	0	0	0	1	-360	360;
	6598	4735	0.00079	0.680124	0	0	0	0	0	0	1	-360	360;
	13441	4735	-6.2e-05	0.76438	0	0	0	0	0	0	1	-360	360;
	12362	4735	0.288988	1.328967	0	0	0	0	0	0	1	-360	360;
	6598	444	0.289174	0.947893	0	0	0	0	0	0	1	-360	360;
	13441	444	0.357831	1.127893	0	0	0	0	0	0	1	-360	360;
	12362	444	0.00434	0.222459	0	0	0	0	0	0	1	-360	360;
	12850	11860	0.162281	0.824401	0	0	0	0	0	0	1	-360	360;
	2525	11860	0.510455	1.73717	0	0	0	0	0	0	1	-360	360;
	909	11860	0.218822	1.471838	0	0	0	0	0	0	1	-360	360;
	2114	11860	0.029581	0.641756	0	0	0	0	0	0	1	-360	360;
	6916	11860	0.269194	1.398492	0	0	0	0	0	0	1	-360	360;
	1149	10055	0.01824	0.407147	0	0	0	0	0	0	1	-360	360;
	1149	10493	0.01992	0.388733	0	0	0	0	0	0	1	-360	360;
	10196	2779	0.504669	1.26314	0	0	0	0	0	0	1	-360	360;
	854	8044	0.180042	0.913089	0	0	0	0	0	0	1	-360	360;
	854	3171	0.28705	1.359279	0	0	0	0	0	0	1	-360	360;
	13539	6637	0.02939	0.690041	0	0	0	0	0	0	1	-360	360;
	11674	6637	0.106211	0.857975	0	0	0	0	0	0	1	-360	360;
	13539	2431	0.03057	0.640372	0	0	0	0	0	0	1	-360	360;
	11674	2431	0.11774	0.83438	0	0	0	0	0	0	1	-360	360;
	2602	3731	0.000416	0.031501	0	0	0	0	0.909421	0	1	-360	360;
	2602	11961	-0.02392	0.308892	0	0	0	0	0	0	1	-360	360;
	2602	3731	-0.021129	0.40847	0	0	0	0	0	0	1	-360	360;
	2602	12457	0.801524	3.248477	0	0	0	0	0	0	1	-360	360;
	5257	11961	15.18223	29.541321	0	0	0	0	0	0	1	-360	360;
	10026	11961	0.810744	3.364049	0	0	0	0	0	0	1	-360	360;
	5257	3731	13.91921	30.6467	0	0	0	0	0	0	1	-360	360;
	10026	3731	0.716322	3.154132	0	0	0	0	0	0	1	-360	360;
	10846	793	0.00489	0.406095	0	0	0	0	0	0	1	-360	360;
	12850	8677	0.447909	1.098823	0	0	0	0	0	0	1	-360	360;
	12850	9658	0.01056	0.353927	0	0	0	0	0	0	1	-360	360;
	12850	1626	0.7509	1.841551	0	0	0	0	0	0	1	-360	360;
	2525	8677	0.375789	0.752493	0	0	0	0	0	0	1	-360	360;
	2525	9658	0.190388	1.261149	0	0	0	0	0	0	1	-360	360;
	2525	1626	0.05324	0.584501	0	0	0	0	0	0	1	-360	360;
	5784	8044	0.01773	0.615173	0	0	0	0	0	0	1	-360	360;
	5784	10203	0.01987	0.600173	0	0	0	0	0	0	1	-360	360;
	3229	8044	0.466918	1.296815	0	0	0	0	0	0	1	-360	360;
	3229	3171	0.815443	1.930956	0	0	0	0	0	0	1	-360	360;
	6598	6259	0.02061	0.308595	0	0	0	0	0	0	1	-360	360;
	13441	6259	0.02549	0.367231	0	0	0	0	0	0	1	-360	360;
	12362	6259	0.00824	0.278492	0	0	0	0	0	0	1	-360	360;
	5257	12457	0.71662	3.572645	0	0	0	0	0	0	1	-360	360;
	4685	4749	0.045599	0.50595	0	0	0	0	0	0	1	-360	360;
	13539	996	0.00887	0.706717	0	0	0	0	0	0	1	-360	360;
	13539	10794	0.04089	0.90831	0	0	0	0	0	0	1	-360	360;
	13539	6683	0.04322	0.911288	0	0	0	0	0	0	1	-360	360;
	490	8044	0.032849	0.6625	0	0	0	0	0	0	1	-360	360;
	680	8044	0.03726	0.637562	0	0	0	0	0	0	1	-360	360;
	148	8044	0.074341	0.632975	0	0	0	0	0	0	1	-360	360;
	9278	8044	0.07781	0.62157	0	0	0	0	0	0	1	-360	360;
	909	8677	0.534411	1.302908	0	0	0	0	0	0	1	-360	360;
	909	9658	0.02355	0.61356	0	0	0	0	0	0	1	-360	360;
	909	1626	0.420249	1.970499	0	0	0	0	0	0	1	-360	360;
	2114	8677	0.229855	0.963099	0	0	0	0	0	0	1	-360	360;
	6916	8677	0.074669	0.41343	0	0	0	0	0	0	1	-360	360;
	10196	6981	0.06182	0.403213	0	0	0	0	0	0	1	-360	360;
	10026	12457	-0.029256	0.40562	0	0	0	0	0	0	1	-360	360;
	8870	9268	0.0033	0.455227	0	0	0	0	0	0	1	-360	360;
	4685	9268	0.00124	0.284731	0	0	0	0	0	0	1	-360	360;
	118	10055	0.04584	0.400152	0	0	0	0	0	0	1	-360	360;
	118	10493	0.03961	0.441877	0	0	0	0	0	0	1	-360	360;
	8870	3748	0.282149	1.82343	0	0	0	0	0	0	1	-360	360;
	4570	3748	-0.001467	0.288285	0	0	0	0	0	0	1	-360	360;
	9487	5095	0.000203	0.026879	0	0	0	0	1.049631	-0.088275	1	-360	360;
	8870	5095	3.246901	6.102273	0	0	0	0	0	0	1	-360	360;
	7990	8260	0.000187	0.028583	0	0	0	0	1.062286	0.114968	1	-360	360;
	7990	8260	0.000198	0.028	0	0	0	0	1.062286	0.114968	1	-360	360;
	8616	8260	0.046659	0.29938	0	0	0	0	0	0	1	-360	360;
	6117	12560	0.02508	0.489778	0	0	0	0	0	0	1	-360	360;
	6117	11208	0.0359	0.39268	0	0	0	0	0	0	1	-360	360;
	8691	12560	0.01931	0.550478	0	0	0	0	0	0	1	-360	360;
	8691	11208	0.0365	0.48437	0	0	0	0	0	0	1	-360	360;
	2114	9658	-0.015393	0.242665	0	0	0	0	0	0	1	-360	360;
	2114	9658	0.00024	0.016571	0	0	0	0	0.99092	0	1	-360	360;
	6916	9658	0.778884	2.360744	0	0	0	0	0	0	1	-360	360;
	2114	1626	0.834349	2.163366	0	0	0	0	0	0	1	-360	360;
	2728	12560	0.02768	0.486653	0	0	0	0	0	0	1	-360	360;
	2728	12560	0.0005	0.0343	0	0	0	0	0.925399	0	1	-360	360;
	2728	11208	0.170748	0.871399	0	0	0	0	0	0	1	-360	360;
	8870	7630	3.440496	6.966735	0	0	0	0	0	0	1	-360	360;
	8870	12875	-0.025	0.350744	0	0	0	0	0	0	1	-360	360;
	8870	5550	3.294252	7.269391	0	0	0	0	0	0	1	-360	360;
	8870	5687	3.075277	7.063712	0	0	0	0	0	0	1	-360	360;
	8870	11156	2.188573	6.820568	0	0	0	0	0	0	1	-360	360;
	8616	3171	0.03254	0.295443	0	0	0	0	0	0	1	-360	360;
	4570	4305	0.000216	0.016549	0	0	0	0	0.953239	0	1	-360	360;
	10400	7018	0.00914	0.394959	0	0	0	0	0	0	1	-360	360;
	11674	996	0.11943	0.839463	0	0	0	0	0	0	1	-360	360;
	10846	996	0.02832	0.560682	0	0	0	0	0	0	1	-360	360;
	10846	5661	0.025289	0.560351	0	0	0	0	0	0	1	-360	360;
	10400	13016	0.203901	1.623181	0	0	0	0	0	0	1	-360	360;
	9013	4956	0.000184	0.025546	0	0	0	0	1.049631	-0.088275	1	-360	360;
	6597	10055	0.0086	0.46804	0	0	0	0	0	0	1	-360	360;
	6597	10493	0.00772	0.464148	0	0	0	0	0	0	1	-360	360;
	13547	12769	0.08261	0.429463	0	0	0	0	0	0	1	-360	360;
	13547	8118	0.09332	0.419277	0	0	0	0	0	0	1	-360	360;
	490	10203	0.06593	0.705471	0	0	0	0	0	0	1	-360	360;
	680	10203	0.070298	0.687417	0	0	0	0	0	0	1	-360	360;
	13171	7609	0.000348	0.027366	0	0	0	0	1.085859	0.169702	1	-360	360;
	13547	7766	0.000239	0.02019	0	0	0	0	1.043143	0.07517	1	-360	360;
	11674	10794	0.0275	0.64187	0	0	0	0	0	0	1	-360	360;
	11674	6683	0.02423	0.621711	0	0	0	0	0	0	1	-360	360;
	10400	2365	0.021711	0.411839	0	0	0	0	0	0	1	-360	360;
	10400	2365	0.000216	0.018022	0	0	0	0	1.004737	0	1	-360	360;
	10400	9625	0.000173	0.017112	0	0	0	0	1.006033	0	1	-360	360;
	11436	11156	0.000179	0.024707	0	0	0	0	1.049631	-0.088275	1	-360	360;
	11436	11156	0.000209	0.023899	0	0	0	0	1.049631	-0.088275	1	-360	360;
	6916	1626	-0.024793	0.19232	0	0	0	0	0	0	1	-360	360;
	6598	2128	0.000274	0.023047	0	0	0	0	1.049631	-0.088275	1	-360	360;
	12362	2536	0.000476	0.030991	0	0	0	0	0.924062	0	1	-360	360;
	148	3171	0.11386	1.025723	0	0	0	0	0	0	1	-360	360;
	9278	3171	0.111979	1.046363	0	0	0	0	0	0	1	-360	360;
	9278	568	0.00022	0.026909	0	0	0	0	1.049631	-0.088275	1	-360	360;
	12987	12396	0.383283	2.185803	0	0	0	0	0	0	1	-360	360;
	12987	9924	0.497389	1.835595	0	0	0	0	0	0	1	-360	360;
	6850	10235	0.000615	0.029917	0	0	0	0	0.946179	0	1	-360	360;
	6850	10235	0.000634	0.029307	0	0	0	0	1.168511	0	1	-360	360;
	2716	13230	-0.02062	0.218203	0	0	0	0	0	0	1	-360	360;
	2716	4092	0.20759	1.109002	0	0	0	0	0	0	1	-360	360;
	2716	645	0.275429	1.015166	0	0	0	0	0	0	1	-360	360;
	2716	712	1.042036	2.256095	0	0	0	0	0	0	1	-360	360;
	748	13230	0.10976	0.74593	0	0	0	0	0	0	1	-360	360;
	2684	13230	0.287831	1.163678	0	0	0	0	0	0	1	-360	360;
	795	12063	0.07	0.46124	0	0	0	0	0	0	1	-360	360;
	10162	12063	0.068479	0.427066	0	0	0	0	0	0	1	-360	360;
	7574	9680	0.103857	0.477763	0	0	0	0	0	0	1	-360	360;
	7574	6056	1.274515	2.49626	0	0	0	0	0	0	1	-360	360;
	9460	9680	0.107943	0.458262	0	0	0	0	0	0	1	-360	360;
	9460	6056	1.323546	2.392243	0	0	0	0	0	0	1	-360	360;
	12151	12396	0.209487	1.002909	0	0	0	0	0	0	1	-360	360;
	12151	9924	0.126607	0.574751	0	0	0	0	0	0	1	-360	360;
	3115	2049	0.00044	0.03688	0	0	0	0	1.02179	0	1	-360	360;
	3115	11366	0.01559	0.151898	0	0	0	0	0	0	1	-360	360;
	6458	3921	0.20449	0.679959	0	0	0	0	0	0	1	-360	360;
	9450	4092	0.105249	0.487299	0	0	0	0	0	0	1	-360	360;
	748	4092	0.000246	0.024009	0	0	0	0	0.999313	-0.092965	1	-360	360;
	748	4092	-0.01905	0.247025	0	0	0	0	0	0	1	-360	360;
	748	645	1.418837	2.394528	0	0	0	0	0	0	1	-360	360;
	748	712	0.205332	0.854986	0	0	0	0	0	0	1	-360	360;
	6457	4092	1.852727	2.747521	0	0	0	0	0	0	1	-360	360;
	2684	4092	1.359979	2.972315	0	0	0	0	0	0	1	-360	360;
	2684	645	0.02269	0.199771	0	0	0	0	0	0	1	-360	360;
	6458	5521	0.00688	0.14088	0	0	0	0	0	0	1	-360	360;
	11659	9680	0.00021	0.022916	0	0	0	0	0.976337	-0.045364	1	-360	360;
	11659	9680	0.00021	0.02236	0	0	0	0	0.982187	-0.057056	1	-360	360;
	6973	277	0.021781	0.193349	0	0	0	0	0	0	1	-360	360;
	2684	712	0.037729	0.222665	0	0	0	0	0	0	1	-360	360;
	8715	11366	0.05599	0.311752	0	0	0	0	0	0	1	-360	360;
	6933	8132	0.000687	0.037873	0	0	0	0	1.030749	0	1	-360	360;
	4218	8132	0.000697	0.03705	0	0	0	0	1.0213	0	1	-360	360;
	6973	11046	0.000258	0.015179	0	0	0	0	1.043597	0	1	-360	360;
	6973	11046	-0.008698	0.24938	0	0	0	0	0	0	1	-360	360;
	6973	11046	0.000218	0.015068	0	0	0	0	1.043697	0	1	-360	360;
	795	11046	0.055651	0.515413	0	0	0	0	0	0	1	-360	360;
	10162	11046	0.05444	0.477231	0	0	0	0	0	0	1	-360	360;
	6458	4615	0.153248	0.626697	0	0	0	0	0	0	1	-360	360;
	5273	313	-0.003823	0.150277	0	0	0	0	0	0	1	-360	360;
	5273	313	0.00069	0.03981	0	0	0	0	1.129864	0	1	-360	360;
	284	313	0.056219	0.23343	0	0	0	0	0	0	1	-360	360;
	3291	12396	0.000243	0.01637	0	0	0	0	0.933986	-0.032541	1	-360	360;
	3291	12396	-0.011591	0.366818	0	0	0	0	0	0	1	-360	360;
	3291	9924	0.145519	0.713643	0	0	0	0	0	0	1	-360	360;
	739	12396	0.000231	0.015741	0	0	0	0	0.933986	-0.032541	1	-360	360;
	739	12396	-0.010888	0.388409	0	0	0	0	0	0	1	-360	360;
	739	9924	0.157909	0.659799	0	0	0	0	0	0	1	-360	360;
	2509	10997	0.00297	0.141711	0	0	0	0	0	0	1	-360	360;
	6595	4615	0.000245	0.011573	0	0	0	0	1.063511	0	1	-360	360;
	6595	4615	0.000255	0.011329	0	0	0	0	1.066278	0	1	-360	360;
	284	1224	0.000697	0.027473	0	0	0	0	1.140456	0	1	-360	360;
	284	1224	0.000627	0.031139	0	0	0	0	1.050543	0	1	-360	360;
	11275	6442	0.00203	0.101959	0	0	0	0	0.935144	0	1	-360	360;
	11275	6442	0.00195	0.103917	0	0	0	0	0.934857	0	1	-360	360;
	11518	12899	0.00341	0.14428	0	0	0	0	1.007584	-0.012893	1	-360	360;
	9006	3632	0.001812	0.129205	0	0	0	0	1.091659	-0.183486	1	-360	360;
	9006	3632	0.001982	0.126629	0	0	0	0	1.091659	-0.183486	1	-360	360;
	6702	8295	0.00153	0.094124	0	0	0	0	0.923203	0	1	-360	360;
	6702	8295	0.00156	0.092157	0	0	0	0	0.915583	0	1	-360	360;
	5740	6625	0.00512	0.156934	0	0	0	0	1.010993	0	1	-360	360;
	5740	2262	0.00513	0.153636	0	0	0	0	1.010221	0	1	-360	360;
	1615	8420	0.001473	0.055275	0	0	0	0	1.007611	0	1	-360	360;
	8161	11298	0.000628	0.026981	0	0	0	0	0.98557	0	1	-360	360;
	9727	9494	0.000768	0.027532	0	0	0	0	0.998512	0	1	-360	360;
	2337	6209	0.000778	0.05467	0	0	0	0	1.00533	0	1	-360	360;
	2337	6209	0.000767	0.055473	0	0	0	0	1.002994	0	1	-360	360;
	9493	12719	0.000905	0.055047	0	0	0	0	1.017266	0	1	-360	360;
	10869	2932	0.000902	0.025704	0	0	0	0	1.077405	0	1	-360	360;
	10869	2932	0.000902	0.024552	0	0	0	0	0.979902	0	1	-360	360;
	7387	2932	0.000437	0.013336	0	0	0	0	0.973277	0	1	-360	360;
	428	10739	0.001665	0.048767	0	0	0	0	1.023061	0	1	-360	360;
	6191	3261	0.00083	0.05142	0	0	0	0	1.004796	0	1	-360	360;
	6191	3261	0.000754	0.045701	0	0	0	0	1.028944	0	1	-360	360;
	7978	11053	0.000798	0.052223	0	0	0	0	1.002851	0	1	-360	360;
	9500	2391	0.000748	0.052717	0	0	0	0	0.964338	0	1	-360	360;
	5173	327	0.000435	0.02158	0	0	0	0	0.984122	0	1	-360	360;
	7453	12066	0.000871	0.0262	0	0	0	0	0.990725	0	1	-360	360;
	8871	6744	0.000881	0.047183	0	0	0	0	0.9851	0	1	-360	360;
	6230	3205	0.000214	0.018189	0	0	0	0	0.966505	0	1	-360	360;
	6915	8410	0.000545	0.028951	0	0	0	0	0.991711	0	1	-360	360;
	7896	10888	0.000433	0.016641	0	0	0	0	1.034384	0	1	-360	360;
	7896	10888	0.000271	0.019483	0	0	0	0	1.039631	0	1	-360	360;
	1388	5387	0.000292	0.041885	0	0	0	0	1.006448	0	1	-360	360;
	1388	5387	0.000283	0.025435	0	0	0	0	1.022071	0	1	-360	360;
	10105	7259	0.00019	0.020772	0	0	0	0	0.994689	0	1	-360	360;
	4268	10628	0.000567	0.0487	0	0	0	0	1.005621	0	1	-360	360;
	4268	10628	0.000292	0.024683	0	0	0	0	1.021598	0	1	-360	360;
	6662	10992	0.000238	0.020505	0	0	0	0	0.991229	0	1	-360	360;
	6662	10992	0.000302	0.024253	0	0	0	0	1.005556	0	1	-360	360;
	2214	4789	0.000394	0.025662	0	0	0	0	0.97307	0	1	-360	360;
	2214	4789	0.000403	0.026263	0	0	0	0	0.970495	0	1	-360	360;
	2214	4789	0.000358	0.02845	0	0	0	0	0.985386	0	1	-360	360;
	1709	5206	0.000184	0.020004	0	0	0	0	0.999118	0	1	-360	360;
	9891	7037	0.000287	0.030676	0	0	0	0	0.997019	0	1	-360	360;
	1567	9431	0.000183	0.021447	0	0	0	0	0.987623	0	1	-360	360;
	1567	9431	0.000183	0.019858	0	0	0	0	0.981425	0	1	-360	360;
	13545	9509	0.000196	0.019264	0	0	0	0	0.988502	0	1	-360	360;
	13545	9509	0.000205	0.018521	0	0	0	0	1.01561	0	1	-360	360;
	8272	4929	0.000205	0.018131	0	0	0	0	0.961546	0	1	-360	360;
	1009	10220	0.000224	0.01875	0	0	0	0	1.136953	0	1	-360	360;
	1009	10220	0.000685	0.016451	0	0	0	0	0.960333	0	1	-360	360;
	6340	5358	0.00021	0.024641	0	0	0	0	0.963809	0	1	-360	360;
	6340	5358	0.00021	0.025112	0	0	0	0	0.965125	0	1	-360	360;
	11828	5909	0	0.00554	0	0	0	0	0	0	1	-360	360;
	2829	1460	0	0.004842	0	0	0	0	0.997125	-0.075844	1	-360	360;
	701	9886	0.000205	0.016639	0	0	0	0	1.003672	0	1	-360	360;
	701	9886	0.000214	0.017035	0	0	0	0	1.002044	0	1	-360	360;
	701	9886	0.000214	0.016674	0	0	0	0	0.983466	0	1	-360	360;
	5878	4948	0.000224	0.018397	0	0	0	0	0.99081	0	1	-360	360;
	4412	1018	0.000522	0.027299	0	0	0	0	0.983724	0	1	-360	360;
	4412	1018	0.000646	0.037462	0	0	0	0	0.994076	0	1	-360	360;
	8475	10394	0.000244	0.033945	0	0	0	0	1.063896	0	1	-360	360;
	8475	10394	0.000196	0.023262	0	0	0	0	1.032098	0	1	-360	360;
	2059	10394	0.00091	0.001563	0	0	0	0	0.999969	-0.007859	1	-360	360;
	2775	6324	0.000179	0.016955	0	0	0	0	1.033369	0	1	-360	360;
	10704	631	0.000206	0.01883	0	0	0	0	0.979095	0	1	-360	360;
	8800	11725	0.000233	0.018474	0	0	0	0	0.91598	0	1	-360	360;
	8800	11725	0.000421	0.031827	0	0	0	0	1.185762	0	1	-360	360;
	1524	13279	0.000219	0.016973	0	0	0	0	1.036934	0	1	-360	360;
	11828	2503	0.000213	0.020004	0	0	0	0	0.972687	0	1	-360	360;
	1068	1975	0.000556	0.033053	0	0	0	0	0.991034	0	1	-360	360;
	6365	12553	0.000508	0.028375	0	0	0	0	1.003719	0	1	-360	360;
	5002	2036	0.000179	0.022003	0	0	0	0	0.983999	0	1	-360	360;
	3905	13349	0.000209	0.02019	0	0	0	0	0.95643	0	1	-360	360;
	3905	13349	0.000214	0.020311	0	0	0	0	0.957146	0	1	-360	360;
	10840	7911	0.000616	0.030505	0	0	0	0	0.998594	0	1	-360	360;
	10840	7911	0.000626	0.029369	0	0	0	0	0.998622	0	1	-360	360;
	3492	5566	0.000214	0.019409	0	0	0	0	0.98363	0	1	-360	360;
	3492	5566	0.000192	0.022	0	0	0	0	0.990701	0	1	-360	360;
	7640	3189	0.000531	0.03466	0	0	0	0	1.004979	0	1	-360	360;
	8556	11983	0.000531	0.035228	0	0	0	0	0.996232	0	1	-360	360;
	8556	11983	0.00049	0.031535	0	0	0	0	0.996119	0	1	-360	360;
	8556	11983	0.000209	0.020964	0	0	0	0	0.939959	0	1	-360	360;
	205	1755	0.000393	0.02663	0	0	0	0	1.013061	0	1	-360	360;
	205	1755	0.000402	0.026088	0	0	0	0	1.013764	0	1	-360	360;
	1512	9362	0.000201	0.017787	0	0	0	0	0.981478	0	1	-360	360;
	5568	1147	0.000289	0.018121	0	0	0	0	1.030461	0	1	-360	360;
	2052	6733	0.000643	0.024464	0	0	0	0	1.002263	0	1	-360	360;
	2052	6733	0.000568	0.027771	0	0	0	0	0.988227	0	1	-360	360;
	1203	5642	0.000506	0.031833	0	0	0	0	0.996877	0	1	-360	360;
	1203	5642	0.000196	0.024118	0	0	0	0	0.98502	0	1	-360	360;
	8622	13135	0.000205	0.022593	0	0	0	0	1.106495	0	1	-360	360;
	8622	13135	0.000186	0.01968	0	0	0	0	0.945775	0	1	-360	360;
	8622	13135	0.000205	0.01997	0	0	0	0	0.953095	0	1	-360	360;
	1673	3846	0.000215	0.018787	0	0	0	0	1.003969	0	1	-360	360;
	1673	3846	0.000215	0.018388	0	0	0	0	1.002646	0	1	-360	360;
	4779	9114	0.000224	0.018834	0	0	0	0	0.997384	0	1	-360	360;
	5573	1612	0.000433	0.043546	0	0	0	0	1.057492	0	1	-360	360;
	11083	4466	0.000234	0.018757	0	0	0	0	0.989873	0	1	-360	360;
	11083	4466	0.000197	0.019207	0	0	0	0	0.991518	0	1	-360	360;
	11083	4466	0.000197	0.019495	0	0	0	0	0.992052	0	1	-360	360;
	9773	11466	0.00018	0.021226	0	0	0	0	1.04087	0	1	-360	360;
	9773	11466	0.000812	0.03849	0	0	0	0	1.050815	0	1	-360	360;
	3230	6177	0.000205	0.02035	0	0	0	0	1.048518	0	1	-360	360;
	1060	5766	0.000593	0.031521	0	0	0	0	0.998708	0	1	-360	360;
	1060	5766	0.000612	0.031123	0	0	0	0	0.998743	0	1	-360	360;
	1060	5766	0.000427	0.024994	0	0	0	0	0.988957	0	1	-360	360;
	3958	608	0.000667	0.027207	0	0	0	0	0.981105	0	1	-360	360;
	3958	608	0.000621	0.030836	0	0	0	0	1.02183	0	1	-360	360;
	449	1334	0.000192	0.020791	0	0	0	0	1.000069	0	1	-360	360;
	449	1334	0.000192	0.02119	0	0	0	0	0.991682	0	1	-360	360;
	9878	40	0.000686	0.030987	0	0	0	0	1.00917	0	1	-360	360;
	4332	8256	0.000194	0.020764	0	0	0	0	0.991744	0	1	-360	360;
	257	2633	0.000717	0.038087	0	0	0	0	1.014054	0	1	-360	360;
	12754	4323	0.000206	0.01883	0	0	0	0	1.001238	0	1	-360	360;
	12754	116	0.000206	0.018439	0	0	0	0	1.000839	0	1	-360	360;
	10272	41	0.000219	0.017019	0	0	0	0	0.93708	0	1	-360	360;
	10272	41	0.000224	0.016641	0	0	0	0	1.105076	0	1	-360	360;
	9663	340	0.000676	0.047085	0	0	0	0	1.015319	0	1	-360	360;
	9663	340	0.000667	0.048014	0	0	0	0	1.016183	0	1	-360	360;
	10769	5778	0.000196	0.019215	0	0	0	0	1.001385	0	1	-360	360;
	5833	8585	0.000253	0.025528	0	0	0	0	1.014185	0	1	-360	360;
	12079	10130	0.00021	0.017884	0	0	0	0	0.980239	0	1	-360	360;
	12079	10130	0.00021	0.017518	0	0	0	0	0.979239	0	1	-360	360;
	2449	2731	0.000219	0.017358	0	0	0	0	0.886675	0	1	-360	360;
	2449	2731	0.000219	0.016973	0	0	0	0	1.040658	0	1	-360	360;
	5198	1227	0.000295	0.027995	0	0	0	0	0.997171	0	1	-360	360;
	5198	1227	0.000277	0.027486	0	0	0	0	1.002862	0	1	-360	360;
	957	4876	0.000489	0.032074	0	0	0	0	1.001121	0	1	-360	360;
	8883	4497	0.00018	0.02129	0	0	0	0	1.028583	0	1	-360	360;
	1543	11738	0.000366	0.029147	0	0	0	0	0.99364	0	1	-360	360;
	1543	11738	0.000366	0.028067	0	0	0	0	1.002233	0	1	-360	360;
	1543	11738	0.00042	0.026331	0	0	0	0	0.990472	0	1	-360	360;
	7258	11673	0.000441	0.01916	0	0	0	0	0.964133	0	1	-360	360;
	10884	8972	0.000391	0.024766	0	0	0	0	0.979452	0	1	-360	360;
	10884	8972	0.000401	0.024225	0	0	0	0	0.979476	0	1	-360	360;
	10884	6828	0.000408	0.028237	0	0	0	0	0.981165	0	1	-360	360;
	10183	6790	0.000199	0.018609	0	0	0	0	0.972278	0	1	-360	360;
	3862	13240	0.000196	0.021597	0	0	0	0	1.019951	0	1	-360	360;
	3862	13240	0.000186	0.022003	0	0	0	0	0.967128	0	1	-360	360;
	4089	4803	0.000186	0.019718	0	0	0	0	0.992589	0	1	-360	360;
	6524	2062	0.000205	0.022703	0	0	0	0	1.015338	0	1	-360	360;
	1163	11469	0.000228	0.021908	0	0	0	0	0.964453	0	1	-360	360;
	10894	11166	0.000322	0.028734	0	0	0	0	0.987134	0	1	-360	360;
	10894	13462	0.000319	0.026574	0	0	0	0	1.020058	0	1	-360	360;
	5263	1756	0.000289	0.018298	0	0	0	0	1.082094	0	1	-360	360;
	5263	1756	0.000199	0.02111	0	0	0	0	1.070801	0	1	-360	360;
	11270	12225	0.000196	0.02119	0	0	0	0	0.989832	0	1	-360	360;
	11270	12225	0.000196	0.021597	0	0	0	0	0.990939	0	1	-360	360;
	6614	2140	0.001097	0.082091	0	0	0	0	1.004174	0	1	-360	360;
	6614	2140	0.000406	0.048834	0	0	0	0	1.056479	0	1	-360	360;
	6614	2140	7.2e-05	0.017798	0	0	0	0	1.007737	0	1	-360	360;
	6614	2140	0.000424	0.04684	0	0	0	0	1.056821	0	1	-360	360;
	12155	6004	0.000224	0.01875	0	0	0	0	0.992009	0	1	-360	360;
	12155	6004	0.000224	0.018334	0	0	0	0	0.992184	0	1	-360	360;
	1810	6271	0.000379	0.018493	0	0	0	0	1.043052	0	1	-360	360;
	1810	6271	0.000253	0.019711	0	0	0	0	1.045217	0	1	-360	360;
	570	11776	0.000196	0.016483	0	0	0	0	0.987356	0	1	-360	360;
	570	11776	0.000186	0.022421	0	0	0	0	1.000553	0	1	-360	360;
	9883	6027	0.00046	0.030794	0	0	0	0	1.023789	0	1	-360	360;
	13300	3475	0.000695	0.049098	0	0	0	0	0.982143	0	1	-360	360;
	13300	3475	0.000355	0.024893	0	0	0	0	0.945703	0	1	-360	360;
	924	3475	0.000365	0.024372	0	0	0	0	0.951103	0	1	-360	360;
	4543	11390	0.000411	0.023652	0	0	0	0	0.936202	0	1	-360	360;
	9749	7207	0.000386	0.025673	0	0	0	0	0.941054	0	1	-360	360;
	9749	7207	0.000344	0.029336	0	0	0	0	0.940894	0	1	-360	360;
	1190	6613	0.000653	0.052104	0	0	0	0	0.982143	0	1	-360	360;
	1190	6613	0.000767	0.043531	0	0	0	0	0.982143	0	1	-360	360;
	5026	8307	0.000323	0.028314	0	0	0	0	0.942497	0	1	-360	360;
	5026	8307	0.000323	0.025999	0	0	0	0	0.942876	0	1	-360	360;
	5026	10642	0.000365	0.025478	0	0	0	0	0.942686	0	1	-360	360;
	1251	9209	0.00086	0.039291	0	0	0	0	0.982143	0	1	-360	360;
	1041	1777	0.000355	0.026586	0	0	0	0	0.93516	0	1	-360	360;
	1041	1777	0.000375	0.026313	0	0	0	0	0.935102	0	1	-360	360;
	9068	3391	0.000344	0.023329	0	0	0	0	0.959849	0	1	-360	360;
	9068	3391	0.000302	0.026375	0	0	0	0	0.958344	0	1	-360	360;
	13566	4205	0.000788	0.04323	0	0	0	0	0.982143	0	1	-360	360;
	13566	4205	0.000767	0.04407	0	0	0	0	0.982143	0	1	-360	360;
	1920	8521	0.000767	0.043562	0	0	0	0	0.982143	0	1	-360	360;
	1920	8521	0.000788	0.041281	0	0	0	0	0.982143	0	1	-360	360;
	1920	8521	0.000892	0.040524	0	0	0	0	0.982143	0	1	-360	360;
	4638	12483	0.000975	0.043562	0	0	0	0	0.982143	0	1	-360	360;
	4638	12483	0.000757	0.042649	0	0	0	0	0.982143	0	1	-360	360;
	6488	3378	0.000261	0.030762	0	0	0	0	0.92865	0	1	-360	360;
	6488	3378	0.000261	0.030088	0	0	0	0	0.928516	0	1	-360	360;
	7339	8926	0.000323	0.026447	0	0	0	0	0.941588	0	1	-360	360;
	7339	2058	0.000323	0.026885	0	0	0	0	0.93842	0	1	-360	360;
	10941	806	0.000323	0.02748	0	0	0	0	0.939738	0	1	-360	360;
	10941	806	0.000292	0.028158	0	0	0	0	0.939729	0	1	-360	360;
	806	7319	0	0.014868	0	0	0	0	0	0.19252	1	-360	360;
	7987	3056	0.000736	0.044526	0	0	0	0	0.982143	0	1	-360	360;
	4094	13471	0.00086	0.039488	0	0	0	0	0.982143	0	1	-360	360;
	5236	9561	0.000365	0.024372	0	0	0	0	0.939493	0	1	-360	360;
	5236	9561	0.000375	0.023986	0	0	0	0	0.939428	0	1	-360	360;
	6589	2737	0.000386	0.025673	0	0	0	0	0.943953	0	1	-360	360;
	3805	2538	0.000809	0.042318	0	0	0	0	0.982143	0	1	-360	360;
	5160	8217	0.000705	0.042038	0	0	0	0	0.982143	0	1	-360	360;
	5197	544	0.000334	0.028022	0	0	0	0	0.967208	0	1	-360	360;
	5197	2615	0.000353	0.022403	0	0	0	0	0.982143	0	1	-360	360;
	3395	544	0.000313	0.026354	0	0	0	0	0.965958	0	1	-360	360;
	3395	544	0.000323	0.025478	0	0	0	0	0.966322	0	1	-360	360;
	2677	11371	0.000344	0.024622	0	0	0	0	0.946628	0	1	-360	360;
	167	11371	0.000355	0.024038	0	0	0	0	0.94666	0	1	-360	360;
	3609	10683	0.000881	0.038244	0	0	0	0	0.982143	0	1	-360	360;
	3609	10683	0.000757	0.040959	0	0	0	0	0.982143	0	1	-360	360;
	9	2951	0.000354	0.028765	0	0	0	0	0.936462	0	1	-360	360;
	9087	2951	0.000354	0.027114	0	0	0	0	0.937994	0	1	-360	360;
	11842	2951	0.000695	0.04748	0	0	0	0	0.982143	0	1	-360	360;
	6777	2296	0.000323	0.028314	0	0	0	0	0.937091	0	1	-360	360;
	6777	2296	0.000323	0.026354	0	0	0	0	0.937017	0	1	-360	360;
	6777	2296	0.000323	0.025696	0	0	0	0	0.936983	0	1	-360	360;
	8137	1420	0.000355	0.027305	0	0	0	0	0.933898	0	1	-360	360;
	8137	1420	0.000355	0.026666	0	0	0	0	0.93371	0	1	-360	360;
	8137	1420	0.000923	0.037456	0	0	0	0	0.982143	0	1	-360	360;
	8137	1420	0.000788	0.035217	0	0	0	0	0.982143	0	1	-360	360;
	1707	10943	0.000323	0.026739	0	0	0	0	0.945782	0	1	-360	360;
	1707	10943	0.000323	0.02749	0	0	0	0	0.945551	0	1	-360	360;
	3556	7192	0.000313	0.027324	0	0	0	0	0.941516	0	1	-360	360;
	3556	7192	0.000313	0.028074	0	0	0	0	0.94145	0	1	-360	360;
	1185	8038	0.000601	0.024984	0	0	0	0	0.983506	0	1	-360	360;
	1185	8038	0.000715	0.022797	0	0	0	0	0.986405	0	1	-360	360;
	1185	8038	0.000344	0.024893	0	0	0	0	0.933442	0	1	-360	360;
	7274	11696	0.000365	0.024163	0	0	0	0	0.944221	0	1	-360	360;
	7274	11696	0.000344	0.02385	0	0	0	0	0.94451	0	1	-360	360;
	7324	10802	0.000757	0.040182	0	0	0	0	0.982143	0	1	-360	360;
	1100	5520	0.000323	0.028937	0	0	0	0	0.940338	0	1	-360	360;
	5445	1829	0.000354	0.026958	0	0	0	0	0.935886	0	1	-360	360;
	5445	1829	0.000778	0.045656	0	0	0	0	0.982143	0	1	-360	360;
	5445	1829	0.000695	0.048932	0	0	0	0	0.982143	0	1	-360	360;
	5445	1829	0.000302	0.025926	0	0	0	0	0.936144	0	1	-360	360;
	9843	13380	0.000355	0.027328	0	0	0	0	0.930802	0	1	-360	360;
	9843	13380	0.000365	0.027156	0	0	0	0	0.930683	0	1	-360	360;
	13490	8642	0.000355	0.024299	0	0	0	0	0.94568	0	1	-360	360;
	2580	1153	0.000375	0.026313	0	0	0	0	0.937729	0	1	-360	360;
	2580	1153	0.000375	0.025605	0	0	0	0	0.937682	0	1	-360	360;
	4714	2203	0.000334	0.026291	0	0	0	0	0.945241	0	1	-360	360;
	4714	2203	0.000705	0.046735	0	0	0	0	0.982143	0	1	-360	360;
	2919	6011	0.000313	0.026875	0	0	0	0	0.952228	0	1	-360	360;
	2919	6011	0.000302	0.027376	0	0	0	0	0.952006	0	1	-360	360;
	6741	7462	0.000664	0.044868	0	0	0	0	0.982143	0	1	-360	360;
	3010	7462	0.000809	0.044526	0	0	0	0	0.982143	0	1	-360	360;
	8243	12724	0.00086	0.034253	0	0	0	0	0.98188	0	1	-360	360;
	8243	12724	0.000355	0.024372	0	0	0	0	0.931285	0	1	-360	360;
	4088	2184	0.000365	0.023788	0	0	0	0	0.936712	0	1	-360	360;
	6661	10021	0.000375	0.023204	0	0	0	0	0.948852	0	1	-360	360;
	6661	10021	0.000302	0.026593	0	0	0	0	0.947797	0	1	-360	360;
	2507	5664	0.000323	0.029256	0	0	0	0	0.934986	0	1	-360	360;
	2507	5664	0.000334	0.029986	0	0	0	0	0.93499	0	1	-360	360;
	2507	5664	0.000355	0.030979	0	0	0	0	0.934976	0	1	-360	360;
	3616	6666	0.000302	0.025707	0	0	0	0	0.942193	0	1	-360	360;
	3616	6666	0.000344	0.025405	0	0	0	0	0.942155	0	1	-360	360;
	4435	8929	0.00012	0.010039	0	0	0	0	0	0.108521	1	-360	360;
	7414	11183	0.000365	0.023715	0	0	0	0	0.941104	0	1	-360	360;
	12476	12299	0.000365	0.023517	0	0	0	0	0.946514	0	1	-360	360;
	12476	8034	0.000334	0.026812	0	0	0	0	0.949752	0	1	-360	360;
	12511	13001	0.000695	0.047253	0	0	0	0	0.982143	0	1	-360	360;
	11017	13001	0.000778	0.043759	0	0	0	0	0.982143	0	1	-360	360;
	227	11066	0.000313	0.030636	0	0	0	0	0.930512	0	1	-360	360;
	227	11066	0.000313	0.028286	0	0	0	0	0.929988	0	1	-360	360;
	227	11066	0.000344	0.028115	0	0	0	0	0.929896	0	1	-360	360;
	3627	7129	0.000902	0.039425	0	0	0	0	0.982143	0	1	-360	360;
	3383	5616	0.000355	0.024237	0	0	0	0	0.942651	0	1	-360	360;
	12835	4954	0.000375	0.023986	0	0	0	0	0.945624	0	1	-360	360;
	12835	4954	0.000375	0.023861	0	0	0	0	0.945648	0	1	-360	360;
	2124	158	0.000819	0.04153	0	0	0	0	0.982143	0	1	-360	360;
	2124	158	0.000809	0.042701	0	0	0	0	0.982143	0	1	-360	360;
	13480	3995	0.000334	0.028022	0	0	0	0	0.930785	0	1	-360	360;
	4416	1314	0.000632	0.048237	0	0	0	0	0.982143	0	1	-360	360;
	4416	1314	0.000632	0.045159	0	0	0	0	0.982143	0	1	-360	360;
	4416	1314	0.000715	0.044039	0	0	0	0	0.982143	0	1	-360	360;
	5763	9938	0.00074	0.044479	0	0	0	0	0.986547	0	1	-360	360;
	5763	5914	0.00074	0.043811	0	0	0	0	0.986547	0	1	-360	360;
	6512	4974	0.000375	0.02385	0	0	0	0	0.951331	0	1	-360	360;
	6512	4974	0.000365	0.023267	0	0	0	0	0.951635	0	1	-360	360;
	6512	4974	0.000323	0.026739	0	0	0	0	0.95056	0	1	-360	360;
	1554	10795	0.000684	0.047025	0	0	0	0	0.982143	0	1	-360	360;
	13455	3083	0.000313	0.02748	0	0	0	0	0.944823	0	1	-360	360;
	9592	12495	0.000323	0.028554	0	0	0	0	0.933305	0	1	-360	360;
	9592	12495	0.000302	0.026437	0	0	0	0	0.933048	0	1	-360	360;
	10309	11074	0.000365	0.025551	0	0	0	0	0.934453	0	1	-360	360;
	10309	11074	0.000344	0.024893	0	0	0	0	0.934463	0	1	-360	360;
	10309	13372	0.000375	0.024508	0	0	0	0	0.960551	0	1	-360	360;
	10059	9913	0.000778	0.041157	0	0	0	0	0.982143	0	1	-360	360;
	10059	9913	0.000923	0.03702	0	0	0	0	0.982143	0	1	-360	360;
	10059	9913	0.000302	0.026593	0	0	0	0	0.953169	0	1	-360	360;
	1465	10468	0.000344	0.029667	0	0	0	0	0.928657	0	1	-360	360;
	1465	55	0.000334	0.02748	0	0	0	0	0.933328	0	1	-360	360;
	1229	10468	0.000323	0.030636	0	0	0	0	0.932366	0	1	-360	360;
	1229	55	0.000323	0.028526	0	0	0	0	0.937672	0	1	-360	360;
	13621	9096	0.000344	0.025332	0	0	0	0	0.943426	0	1	-360	360;
	13621	12207	0.000334	0.024674	0	0	0	0	0.941774	0	1	-360	360;
	13621	9992	0.000344	0.02457	0	0	0	0	0.940947	0	1	-360	360;
	5299	9740	0.000365	0.023788	0	0	0	0	0.935351	0	1	-360	360;
	11015	9611	0.000365	0.023601	0	0	0	0	0.955118	0	1	-360	360;
	11015	9611	0.000344	0.02652	0	0	0	0	0.953676	0	1	-360	360;
	10996	9611	0.000344	0.027272	0	0	0	0	0.953967	0	1	-360	360;
	10996	9611	0.000334	0.027563	0	0	0	0	0.953841	0	1	-360	360;
	2849	9710	0.000313	0.030636	0	0	0	0	0.951606	0	1	-360	360;
	2849	9710	0.000313	0.028286	0	0	0	0	0.952467	0	1	-360	360;
	7948	1162	0.000881	0.036803	0	0	0	0	0.992378	0	1	-360	360;
	7948	1162	0.000902	0.036668	0	0	0	0	0.992425	0	1	-360	360;
	7948	1162	0.000881	0.034014	0	0	0	0	0.993191	0	1	-360	360;
	7280	5559	0.000365	0.026027	0	0	0	0	0.937778	0	1	-360	360;
	7280	5559	0.000386	0.025673	0	0	0	0	0.937768	0	1	-360	360;
	10630	6586	0.00012	0.009966	0	0	0	0	0	-0.036236	1	-360	360;
	13583	82	0.00017	0.006789	0	0	0	0	0	-0.086984	1	-360	360;
	13583	4479	0.000664	0.047107	0	0	0	0	0.982143	0	1	-360	360;
	2075	1783	0.000313	0.028314	0	0	0	0	0.93521	0	1	-360	360;
	2075	1783	0.000334	0.025707	0	0	0	0	0.934745	0	1	-360	360;
	2075	2217	0.000344	0.025478	0	0	0	0	0.931132	0	1	-360	360;
	11026	2217	0.000344	0.027236	0	0	0	0	0.931428	0	1	-360	360;
	11026	2217	0.000375	0.026518	0	0	0	0	0.931193	0	1	-360	360;
	10081	2224	0.000386	0.02646	0	0	0	0	0.925602	0	1	-360	360;
	4266	6	0.000788	0.040317	0	0	0	0	0.982143	0	1	-360	360;
	11004	820	0.000323	0.025863	0	0	0	0	0.949117	0	1	-360	360;
	1086	3185	1e-05	0.023516	0	0	0	0	0	-0.146705	1	-360	360;
	6081	5428	0.000334	0.030306	0	0	0	0	0.945073	0	1	-360	360;
	6081	5428	0.000302	0.030636	0	0	0	0	0.945165	0	1	-360	360;
	10933	5386	0.00011	0.011411	0	0	0	0	0	0.085905	1	-360	360;
	12094	9593	0.000344	0.024841	0	0	0	0	0.945982	0	1	-360	360;
	12094	9593	0.000365	0.025029	0	0	0	0	0.945927	0	1	-360	360;
	12094	9593	0.000355	0.024101	0	0	0	0	0.946218	0	1	-360	360;
	6061	235	0.000344	0.02385	0	0	0	0	0.943808	0	1	-360	360;
	6061	235	0.000411	0.02313	0	0	0	0	0.943451	0	1	-360	360;
	6061	235	0.000829	0.042764	0	0	0	0	0.982143	0	1	-360	360;
	8497	13448	0.000705	0.04748	0	0	0	0	0.982143	0	1	-360	360;
	8497	2050	0.000323	0.029986	0	0	0	0	0.93585	0	1	-360	360;
	8497	2050	0.000302	0.029952	0	0	0	0	0.935892	0	1	-360	360;
	13408	7811	0.000746	0.041064	0	0	0	0	0.982143	0	1	-360	360;
	13408	7811	0.000881	0.042214	0	0	0	0	0.982143	0	1	-360	360;
	3376	11774	0.00086	0.041209	0	0	0	0	0.982143	0	1	-360	360;
	3376	11774	0.000912	0.038669	0	0	0	0	0.982143	0	1	-360	360;
	11927	244	0.000375	0.026244	0	0	0	0	0.95253	0	1	-360	360;
	11927	244	0.000375	0.025457	0	0	0	0	0.95286	0	1	-360	360;
	12417	1657	0.000302	0.029096	0	0	0	0	0.934819	0	1	-360	360;
	6937	3180	0.000302	0.029667	0	0	0	0	0.947685	0	1	-360	360;
	2433	955	0.000334	0.027563	0	0	0	0	0.945284	0	1	-360	360;
	2433	955	0.000323	0.028074	0	0	0	0	0.94517	0	1	-360	360;
	1269	11276	0.000736	0.041064	0	0	0	0	0.982143	0	1	-360	360;
	12272	11276	0.001234	0.040109	0	0	0	0	0.982143	0	1	-360	360;
	12272	11276	0.000757	0.043209	0	0	0	0	0.982143	0	1	-360	360;
	1127	9274	0.000323	0.026586	0	0	0	0	0.962929	0	1	-360	360;
	6157	422	0.000788	0.040296	0	0	0	0	0.982143	0	1	-360	360;
	6157	422	0.000912	0.040514	0	0	0	0	0.982143	0	1	-360	360;
	2936	5132	0.000798	0.042183	0	0	0	0	0.982143	0	1	-360	360;
	2936	5132	0.000778	0.043147	0	0	0	0	0.982143	0	1	-360	360;
	10294	12739	0.000323	0.027793	0	0	0	0	0.924286	0	1	-360	360;
	10294	12739	0.000302	0.027918	0	0	0	0	0.924396	0	1	-360	360;
	10294	12739	0.000302	0.025926	0	0	0	0	0.923543	0	1	-360	360;
	6599	4246	0.000344	0.024956	0	0	0	0	0.942973	0	1	-360	360;
	6599	4246	0.000355	0.024435	0	0	0	0	0.942949	0	1	-360	360;
	5623	459	0.000386	0.026392	0	0	0	0	0.931256	0	1	-360	360;
	8232	459	0.000365	0.025525	0	0	0	0	0.928969	0	1	-360	360;
	2467	169	0.000344	0.029986	0	0	0	0	0.938164	0	1	-360	360;
	2467	169	0.000323	0.029838	0	0	0	0	0.938248	0	1	-360	360;
	3891	8318	0.000323	0.02748	0	0	0	0	0.933289	0	1	-360	360;
	3891	8318	0.000313	0.028001	0	0	0	0	0.933329	0	1	-360	360;
	634	4730	0.000179	0.012988	0	0	0	0	0.997782	0	1	-360	360;
	634	10310	0.000198	0.012724	0	0	0	0	0.994148	0	1	-360	360;
	634	3428	0.000207	0.012469	0	0	0	0	0.994117	0	1	-360	360;
	634	2710	0.000207	0.012206	0	0	0	0	0.99408	0	1	-360	360;
	634	4980	0.000217	0.012912	0	0	0	0	0.99109	0	1	-360	360;
	634	8350	0.000217	0.012627	0	0	0	0	0.99109	0	1	-360	360;
	634	4502	0.000198	0.014317	0	0	0	0	0.991478	0	1	-360	360;
	634	4502	0.000189	0.014592	0	0	0	0	0.991479	0	1	-360	360;
	634	11544	0.000354	0.02725	0	0	0	0	0.96645	0	1	-360	360;
	1356	12535	0.000628	0.042428	0	0	0	0	1.009174	0	1	-360	360;
	1356	12535	0.000726	0.044039	0	0	0	0	0.982143	0	1	-360	360;
	5249	8605	0.000365	0.024893	0	0	0	0	0.95138	0	1	-360	360;
	5249	8605	0.000438	0.02457	0	0	0	0	0.951402	0	1	-360	360;
	5249	8605	0.000365	0.023986	0	0	0	0	0.9518	0	1	-360	360;
	13130	8620	0.000375	0.023267	0	0	0	0	0.955216	0	1	-360	360;
	13130	8620	0.000323	0.026375	0	0	0	0	0.954101	0	1	-360	360;
	2564	3540	0.000334	0.029735	0	0	0	0	0.929892	0	1	-360	360;
	2564	3540	0.000313	0.030237	0	0	0	0	0.930119	0	1	-360	360;
	4880	3018	0.000767	0.044049	0	0	0	0	0.982143	0	1	-360	360;
	4880	3018	0.000302	0.026072	0	0	0	0	0.940778	0	1	-360	360;
	4880	3018	0.00085	0.039892	0	0	0	0	0.982143	0	1	-360	360;
	3448	13294	0.000344	0.024956	0	0	0	0	0.969247	0	1	-360	360;
	13152	13294	0.000334	0.026666	0	0	0	0	0.935423	0	1	-360	360;
	2961	7229	0.000365	0.023788	0	0	0	0	0.957187	0	1	-360	360;
	2961	7229	0.000365	0.023465	0	0	0	0	0.957349	0	1	-360	360;
	326	2206	0.000334	0.026447	0	0	0	0	0.942707	0	1	-360	360;
	5846	7838	0.000323	0.027042	0	0	0	0	0.941267	0	1	-360	360;
	5846	7838	0.000313	0.02748	0	0	0	0	0.94124	0	1	-360	360;
	5846	7838	0.000302	0.028314	0	0	0	0	0.94117	0	1	-360	360;
	5846	10989	0.000323	0.025853	0	0	0	0	0.943398	0	1	-360	360;
	5846	10989	0.000365	0.026197	0	0	0	0	0.943251	0	1	-360	360;
	5846	10989	0.000365	0.025165	0	0	0	0	0.943464	0	1	-360	360;
	8070	7112	0.000355	0.024237	0	0	0	0	0.955747	0	1	-360	360;
	8070	2866	0.000375	0.024247	0	0	0	0	0.94273	0	1	-360	360;
	3090	7112	0.000365	0.023267	0	0	0	0	0.956532	0	1	-360	360;
	3090	2866	0.000342	0.023139	0	0	0	0	0.982143	0	1	-360	360;
	2348	8175	0.000344	0.026666	0	0	0	0	0.956163	0	1	-360	360;
	2348	1000	0.000334	0.027021	0	0	0	0	0.946614	0	1	-360	360;
	1158	8175	0.000302	0.027615	0	0	0	0	0.955743	0	1	-360	360;
	1158	1000	0.000313	0.026135	0	0	0	0	0.946825	0	1	-360	360;
	6736	13022	0.000365	0.025905	0	0	0	0	0.942272	0	1	-360	360;
	6736	13022	0.000355	0.025238	0	0	0	0	0.942404	0	1	-360	360;
	4194	6189	0.000334	0.026814	0	0	0	0	0.93534	0	1	-360	360;
	1639	9626	0.001242	0.083148	0	0	0	0	0.945477	0	1	-360	360;
	1639	6077	0.0017	0.090508	0	0	0	0	0.945477	0	1	-360	360;
	2753	3087	0.001648	0.071935	0	0	0	0	1.018675	0	1	-360	360;
	2753	9433	0.001742	0.073131	0	0	0	0	0.999333	0	1	-360	360;
	237	11610	0.000914	0.088676	0	0	0	0	0.986842	0	1	-360	360;
	237	7510	0.000955	0.085404	0	0	0	0	0.986842	0	1	-360	360;
	2551	5694	0.001479	0.088626	0	0	0	0	0.986842	0	1	-360	360;
	2551	6184	0.00153	0.084928	0	0	0	0	0.986842	0	1	-360	360;
	10587	8940	0.00151	0.102577	0	0	0	0	0.945477	0	1	-360	360;
	12028	3629	0.000884	0.108054	0	0	0	0	0.945477	0	1	-360	360;
	12488	2856	0.001413	0.034472	0	0	0	0	1.053539	0	1	-360	360;
	12293	8073	0.001099	0.069265	0	0	0	0	1.018675	0	1	-360	360;
	105	5239	0.000904	0.088121	0	0	0	0	0.986842	0	1	-360	360;
	6381	4100	0.001386	0.094158	0	0	0	0	0.986842	0	1	-360	360;
	6381	4100	0.001355	0.096007	0	0	0	0	0.986842	0	1	-360	360;
	7298	1671	0.001456	0.098925	0	0	0	0	0.962772	0	1	-360	360;
	7298	1671	0.001424	0.100867	0	0	0	0	0.962772	0	1	-360	360;
	2812	2545	0.001442	0.086424	0	0	0	0	0.999333	0	1	-360	360;
	1624	12444	0.001317	0.070127	0	0	0	0	1.074114	0	1	-360	360;
	884	3755	0.000991	0.083602	0	0	0	0	0.999333	0	1	-360	360;
	884	3755	0.001272	0.097228	0	0	0	0	0.999333	0	1	-360	360;
	884	10166	0.000968	0.088836	0	0	0	0	0.974659	0	1	-360	360;
	884	10166	0.001537	0.08896	0	0	0	0	0.974659	0	1	-360	360;
	884	10166	0.0018	0.078579	0	0	0	0	0.974659	0	1	-360	360;
	642	3703	0.001071	0.079233	0	0	0	0	0.999333	0	1	-360	360;
	8575	5752	0.001301	0.088365	0	0	0	0	1.018675	0	1	-360	360;
	8575	7023	0.001436	0.082886	0	0	0	0	1.018675	0	1	-360	360;
	8575	5355	0.001701	0.102782	0	0	0	0	0.923361	0	1	-360	360;
	10290	12950	0.001352	0.091819	0	0	0	0	0.999333	0	1	-360	360;
	2194	2805	0.001737	0.085791	0	0	0	0	0.962772	0	1	-360	360;
	2194	4579	0.001331	0.087657	0	0	0	0	0.945477	0	1	-360	360;
	2194	2805	0.001877	0.07879	0	0	0	0	0.962772	0	1	-360	360;
	2194	3263	0.001672	0.089295	0	0	0	0	0.962772	0	1	-360	360;
	2194	4579	0.001421	0.10862	0	0	0	0	0.945477	0	1	-360	360;
	2194	3263	0.001338	0.09699	0	0	0	0	0.962772	0	1	-360	360;
	4404	7520	0.001224	0.093571	0	0	0	0	1.018675	0	1	-360	360;
	2850	3660	0.001667	0.092522	0	0	0	0	0.945477	0	1	-360	360;
	2850	1540	0.0017	0.090508	0	0	0	0	0.945477	0	1	-360	360;
	4111	5040	0.001177	0.089984	0	0	0	0	1.038781	0	1	-360	360;
	8853	9985	0.000976	0.083762	0	0	0	0	0.986842	0	1	-360	360;
	8853	3269	0.001221	0.074424	0	0	0	0	0.974659	0	1	-360	360;
	7622	12284	0.001205	0.088315	0	0	0	0	1.038781	0	1	-360	360;
	4835	6198	0.000951	0.081618	0	0	0	0	0.999333	0	1	-360	360;
	4835	8670	0.0011	0.085679	0	0	0	0	0.962772	0	1	-360	360;
	4835	8670	0.000895	0.099866	0	0	0	0	0.962772	0	1	-360	360;
	6583	12130	0	0.064	0	0	0	0	0.955796	0	1	-360	360;
	6583	12130	0	0.06267	0	0	0	0	0.95527	0	1	-360	360;
	6583	12130	0	0.06134	0	0	0	0	0.954721	0	1	-360	360;
	2245	6732	0	0.06295	0	0	0	0	0.983015	0	1	-360	360;
	2245	6732	0	0.07135	0	0	0	0	0.982368	0	1	-360	360;
	4642	2855	0	0.04729	0	0	0	0	1.012296	0	1	-360	360;
	4642	2855	0	0.0482	0	0	0	0	1.013058	0	1	-360	360;
	4642	2855	0	0.04911	0	0	0	0	1.013799	0	1	-360	360;
	4915	13255	0	0.06744	0	0	0	0	0.990953	0	1	-360	360;
	11209	1067	0	0.145661	0	0	0	0	1.052174	0	1	-360	360;
	13272	7347	0	0.03024	0	0	0	0	1.052632	0	1	-360	360;
	13272	7347	0	0.029729	0	0	0	0	1.052632	0	1	-360	360;
	7347	10395	0	0.07191	0	0	0	0	0.965461	0	1	-360	360;
	7347	10395	0	0.0815	0	0	0	0	0.967264	0	1	-360	360;
	13415	4035	0	0.07843	0	0	0	0	0.952986	0	1	-360	360;
	3592	8015	0	0.033891	0	0	0	0	1.052632	0	1	-360	360;
	3592	8015	0	0.035419	0	0	0	0	1.052632	0	1	-360	360;
	8015	11628	0	0.07323	0	0	0	0	0.960506	0	1	-360	360;
	8015	11628	0	0.07177	0	0	0	0	0.961229	0	1	-360	360;
	8006	187	0	0.06789	0	0	0	0	0.936347	0	1	-360	360;
	8006	187	0	0.06524	0	0	0	0	0.936691	0	1	-360	360;
	8006	187	0	0.06385	0	0	0	0	0.936788	0	1	-360	360;
	3795	8133	0	0.04033	0	0	0	0	1.052174	0	1	-360	360;
	8002	5638	0	0.07024	0	0	0	0	0.967111	0	1	-360	360;
	8002	5638	0	0.07161	0	0	0	0	0.966242	0	1	-360	360;
	9037	1219	0	0.06784	0	0	0	0	1.03429	0	1	-360	360;
	2132	9949	0	0.071916	0	0	0	0	0.945577	0	1	-360	360;
	10240	6720	0	0.07323	0	0	0	0	0.985392	0	1	-360	360;
	10240	6720	0	0.07177	0	0	0	0	0.985537	0	1	-360	360;
	10347	2751	0	0.0447	0	0	0	0	1.052174	0	1	-360	360;
	10347	2751	0	0.04377	0	0	0	0	1.052174	0	1	-360	360;
	10347	2751	0	0.04123	0	0	0	0	1.052174	0	1	-360	360;
	3423	11770	0	0.06295	0	0	0	0	0.989703	0	1	-360	360;
	3423	11770	0	0.07135	0	0	0	0	0.986826	0	1	-360	360;
	8344	7098	0	0.07014	0	0	0	0	0.911963	0	1	-360	360;
	8344	7098	0	0.07149	0	0	0	0	0.912529	0	1	-360	360;
	2024	6485	0	0.034531	0	0	0	0	0.973598	0	1	-360	360;
	2024	10014	0	0.04656	0	0	0	0	0.980849	0	1	-360	360;
	2024	10014	0	0.04563	0	0	0	0	0.980159	0	1	-360	360;
	2582	6459	0.000229	0.01234	0	0	0	0	0.872915	0	1	-360	360;
	2582	6459	0.000229	0.0122	0	0	0	0	0.872741	0	1	-360	360;
	12573	4076	0.001485	0.065805	0	0	0	0	0.956938	0	1	-360	360;
	12573	4076	0.001583	0.06586	0	0	0	0	0.956938	0	1	-360	360;
	7706	4076	0.00175	0.070701	0	0	0	0	0.952381	0	1	-360	360;
	7706	4076	0.00169	0.0715	0	0	0	0	0.952381	0	1	-360	360;
	9848	9665	0.001228	0.060916	0	0	0	0	0.950231	0	1	-360	360;
	9848	9665	0.00116	0.063798	0	0	0	0	1.02073	0	1	-360	360;
	9848	9665	0.00113	0.059295	0	0	0	0	1.024588	0	1	-360	360;
	7957	9665	0.00177	0.06853	0	0	0	0	0.93473	0	1	-360	360;
	10922	3077	0.001363	0.063835	0	0	0	0	0.990157	0	1	-360	360;
	10922	3077	0.001457	0.067155	0	0	0	0	0.987625	0	1	-360	360;
	13262	3077	0.00193	0.06377	0	0	0	0	0.937593	0	1	-360	360;
	13262	3077	0.00194	0.06188	0	0	0	0	0.937259	0	1	-360	360;
	4348	7201	0.00045	0.02807	0	0	0	0	1.002506	0	1	-360	360;
	4348	3036	0.00042	0.028021	0	0	0	0	1.002506	0	1	-360	360;
	12666	8194	0.001208	0.06631	0	0	0	0	1.006816	0	1	-360	360;
	12666	8194	0.001228	0.066264	0	0	0	0	1.006868	0	1	-360	360;
	10686	4099	0.00047	0.026469	0	0	0	0	1.002506	0	1	-360	360;
	13229	10478	0.00186	0.06826	0	0	0	0	0.892378	0	1	-360	360;
	829	208	0.001431	0.068667	0	0	0	0	0.956938	0	1	-360	360;
	829	208	0.001529	0.068787	0	0	0	0	0.956938	0	1	-360	360;
	1313	208	0.00193	0.06377	0	0	0	0	0.952381	0	1	-360	360;
	1313	208	0.00194	0.06188	0	0	0	0	0.952381	0	1	-360	360;
	4899	7506	0	0.00898	0	0	0	0	0	0	1	-360	360;
	3705	9266	0	0.00433	0	0	0	0	0	0	1	-360	360;
	9266	7058	0.007955	0.151636	0	0	0	0	0.929293	0	1	-360	360;
	10740	6332	0.000271	0.030694	0	0	0	0	1.052632	0	1	-360	360;
	10740	6332	0.000271	0.02842	0	0	0	0	1.052632	0	1	-360	360;
	7434	6145	0.00064	0.0518	0	0	0	0	0	0	1	-360	360;
	8840	8919	0.00072	0.05075	0	0	0	0	0	0	1	-360	360;
	2519	12023	0	0.23657	0	0	0	0	0	0	1	-360	360;
	12587	10646	0	0.020113	0	0	0	0	1.006865	0	1	-360	360;
	12587	9197	0	0.019669	0	0	0	0	1.006865	0	1	-360	360;
	5232	9683	0.000325	0.028508	0	0	0	0	1.006865	0	1	-360	360;
	5232	9683	0.000316	0.031545	0	0	0	0	1.006865	0	1	-360	360;
	13516	9007	0.000355	0.023171	0	0	0	0	1.006865	0	1	-360	360;
	3949	2083	0.000296	0.032897	0	0	0	0	1.006865	0	1	-360	360;
	3949	2083	0.000296	0.030461	0	0	0	0	1.006865	0	1	-360	360;
	6957	3	5.9e-05	0.022767	0	0	0	0	1.006865	0	1	-360	360;
	1541	9070	0.000335	0.029237	0	0	0	0	1.006865	0	1	-360	360;
	8349	13406	0.000385	0.027846	0	0	0	0	1.006865	0	1	-360	360;
	10467	10025	0.000444	0.020064	0	0	0	0	1.006865	0	1	-360	360;
	12345	60	0.00033	0.02558	0	0	0	0	0	0	1	-360	360;
	11181	9338	0	0.026909	0	0	0	0	0	0	1	-360	360;
	10714	11027	0.000385	0.022737	0	0	0	0	1.006865	0	1	-360	360;
	4421	1028	0.000375	0.022441	0	0	0	0	1.006865	0	1	-360	360;
	11639	8959	-0.002742	0.00908	0	0	0	0	0	0	1	-360	360;
	11639	13350	-0.00268	0.00837	0	0	0	0	0	0	1	-360	360;
	11639	10040	-0.00304	0.00824	0	0	0	0	0	0	1	-360	360;
	10548	7864	-0.003927	0.00795	0	0	0	0	0	0	1	-360	360;
	10548	21	-0.00401	0.00794	0	0	0	0	0	0	1	-360	360;
	6756	13304	0	0.00765	0	0	0	0	0	0	1	-360	360;
	6756	2211	0	0.00748	0	0	0	0	0	0	1	-360	360;
	10793	2588	-0.000111	0.00565	0	0	0	0	0	0	1	-360	360;
	10793	5916	-0.000339	0.00645	0	0	0	0	0	0	1	-360	360;
	10793	8502	-0.000332	0.00657	0	0	0	0	0	0	1	-360	360;
	10793	6461	-0.000118	0.00525	0	0	0	0	0	0	1	-360	360;
	7092	9760	0.00593	0.0602	0	0	0	0	0	0	1	-360	360;
	7092	4687	8e-05	0.00811	0	0	0	0	0	0	1	-360	360;
	7092	7308	-0.003927	0.00835	0	0	0	0	0	0	1	-360	360;
	3752	11372	0	0.028951	0	0	0	0	1.006865	0	1	-360	360;
	3752	2942	0	0.02833	0	0	0	0	1.006865	0	1	-360	360;
	142	4356	0.000424	0.020212	0	0	0	0	1.006865	0	1	-360	360;
	6695	2484	0.00128	0.05539	0	0	0	0	0	0	1	-360	360;
	6695	8104	0.00126	0.056469	0	0	0	0	0	0	1	-360	360;
	6695	5995	0.00123	0.05756	0	0	0	0	0	0	1	-360	360;
	2089	434	0	0.046831	0	0	0	0	0	0	1	-360	360;
	4515	1102	0	0.03197	0	0	0	0	1.006865	0	1	-360	360;
	7461	10383	0	0.027151	0	0	0	0	1.044622	0	1	-360	360;
	5447	1299	0	0.0197	0	0	0	0	0	0	1	-360	360;
	10618	6282	0.000335	0.028458	0	0	0	0	1.006865	0	1	-360	360;
	10618	6282	0.000375	0.027846	0	0	0	0	1.006865	0	1	-360	360;
	12414	1676	0.00073	0.01489	0	0	0	0	0	0	1	-360	360;
	10553	12002	0	0.00864	0	0	0	0	0	0	1	-360	360;
	10553	1931	0	0.00881	0	0	0	0	0	0	1	-360	360;
	10525	2899	0.000325	0.028961	0	0	0	0	1.006865	0	1	-360	360;
	4628	11452	0.000296	0.025084	0	0	0	0	1.006865	0	1	-360	360;
	10874	10969	0.000325	0.025765	0	0	0	0	1.006865	0	1	-360	360;
	10874	10969	0.000375	0.020301	0	0	0	0	1.006865	0	1	-360	360;
	10874	10969	0.000345	0.021109	0	0	0	0	1.006865	0	1	-360	360;
	4486	183	0.000385	0.02101	0	0	0	0	1.006865	0	1	-360	360;
	2297	11100	0.000424	0.019472	0	0	0	0	1.006865	0	1	-360	360;
	2297	11100	0.000434	0.019048	0	0	0	0	1.006865	0	1	-360	360;
	1309	6384	0.000325	0.025637	0	0	0	0	1.006865	0	1	-360	360;
	1309	6384	0.000355	0.026564	0	0	0	0	1.006865	0	1	-360	360;
	11460	6574	0.000355	0.021553	0	0	0	0	1.006865	0	1	-360	360;
	11460	6574	0.000128	0.022007	0	0	0	0	1.006865	0	1	-360	360;
	1186	437	6e-05	0.00994	0	0	0	0	0.979973	0.200467	1	-360	360;
	12182	437	7e-05	0.009939	0	0	0	0	0.977987	0.210204	1	-360	360;
	437	1487	0.000395	0.021454	0	0	0	0	1.006865	0	1	-360	360;
	9642	4794	-0.000388	0.0074	0	0	0	0	0	0	1	-360	360;
	3331	7406	-0.000402	0.00724	0	0	0	0	0	0	1	-360	360;
	7482	2310	-0.000589	0.01411	0	0	0	0	0	0	1	-360	360;
	4882	11681	-0.00036	0.00803	0	0	0	0	0	0	1	-360	360;
	12148	9303	0	0.040449	0	0	0	0	0.956938	0	1	-360	360;
	5827	7994	0.000345	0.026071	0	0	0	0	1.006865	0	1	-360	360;
	3777	12047	0	0.00046	0	0	0	0	0	0	1	-360	360;
	3777	10091	0	0.00043	0	0	0	0	0	0	1	-360	360;
	1233	2661	0.000375	0.022007	0	0	0	0	1.006865	0	1	-360	360;
	1332	10853	0.000385	0.020196	0	0	0	0	0.993901	0	1	-360	360;
	7797	10325	0.000375	0.026624	0	0	0	0	0.993901	0	1	-360	360;
	6482	2017	0.000425	0.021208	0	0	0	0	0.993901	0	1	-360	360;
	11938	3830	0.000466	0.020185	0	0	0	0	0.993901	0	1	-360	360;
	5689	2300	0.000324	0.022139	0	0	0	0	0.993901	0	1	-360	360;
	5689	2300	0.000324	0.022706	0	0	0	0	0.993901	0	1	-360	360;
	5689	2300	0.000314	0.023142	0	0	0	0	0.993901	0	1	-360	360;
	13320	10730	0.000344	0.024832	0	0	0	0	0.993901	0	1	-360	360;
	13320	12508	0.000334	0.023071	0	0	0	0	0.993901	0	1	-360	360;
	3876	546	0.000335	0.022767	0	0	0	0	1.006865	0	1	-360	360;
	3837	502	0.00071	0.030845	0	0	0	0	1.006865	0	1	-360	360;
	7796	10175	0.000385	0.019472	0	0	0	0	1.006865	0	1	-360	360;
	7796	6062	0.000395	0.019057	0	0	0	0	1.006865	0	1	-360	360;
	7796	1847	0.000404	0.018643	0	0	0	0	1.006865	0	1	-360	360;
	7796	7467	0	0.027866	0	0	0	0	1.006865	0	1	-360	360;
	2403	11071	0.000325	0.032245	0	0	0	0	1.006865	0	1	-360	360;
	3187	12549	0.000365	0.022599	0	0	0	0	1.006865	0	1	-360	360;
	3187	12549	0.000355	0.023023	0	0	0	0	1.006865	0	1	-360	360;
	2476	5906	0	0.032004	0	0	0	0	1.045454	0	1	-360	360;
	330	5906	0.00141	0.091249	0	0	0	0	0.916326	-0.411983	1	-360	360;
	3843	13428	0.000434	0.019018	0	0	0	0	1.006865	0	1	-360	360;
	2550	12979	0.000404	0.020497	0	0	0	0	1.006865	0	1	-360	360;
	8192	11171	0.00074	0.029553	0	0	0	0	1.006865	0	1	-360	360;
	9199	2407	0.000454	0.019669	0	0	0	0	1.006865	0	1	-360	360;
	11008	3901	0.00206	0.09005	0	0	0	0	0	0	1	-360	360;
	4750	372	0.00039	0.03411	0	0	0	0	1.052632	0	1	-360	360;
	4750	372	0.0004	0.034469	0	0	0	0	1.052632	0	1	-360	360;
	12534	3174	0.00183	0.090471	0	0	0	0	1.026087	0	1	-360	360;
	12534	3174	0.00187	0.07429	0	0	0	0	1.026087	0	1	-360	360;
	11249	2674	0.00077	0.04355	0	0	0	0	0.999313	0	1	-360	360;
	11249	2674	0.00077	0.04266	0	0	0	0	0.999313	0	1	-360	360;
	10744	10278	0.00116	0.04106	0	0	0	0	1.052632	0	1	-360	360;
	10744	10278	0.00118	0.04119	0	0	0	0	1.052632	0	1	-360	360;
	5821	4247	0.0012	0.03931	0	0	0	0	1.052632	0	1	-360	360;
	5821	4247	0.00107	0.04567	0	0	0	0	1.052632	0	1	-360	360;
	10542	9615	0.002282	0.077224	0	0	0	0	0.911217	0	1	-360	360;
	6182	9615	0.00102	0.04862	0	0	0	0	1.052632	0	1	-360	360;
	6182	9615	0.001	0.04954	0	0	0	0	1.052632	0	1	-360	360;
	4747	1776	0.00098	0.04477	0	0	0	0	1.052632	0	1	-360	360;
	4747	1776	0.00111	0.04387	0	0	0	0	1.052632	0	1	-360	360;
	494	8150	0.00113	0.04193	0	0	0	0	1.052632	0	1	-360	360;
	13118	5577	0.002522	0.074933	0	0	0	0	1.014981	0	1	-360	360;
	1974	7318	0.001252	0.073601	0	0	0	0	0.969039	0	1	-360	360;
	5189	7318	0.001271	0.072001	0	0	0	0	0.967402	0	1	-360	360;
	10030	8979	0.000442	0.040802	0	0	0	0	0.88875	0	1	-360	360;
	1155	12901	0.00048	0.041601	0	0	0	0	0.970291	0	1	-360	360;
	2493	10226	0.000543	0.04261	0	0	0	0	1.040113	0	1	-360	360;
	871	10226	0.000534	0.04342	0	0	0	0	1.039423	0	1	-360	360;
	12422	10224	0.003895	0.040564	0	0	0	0	0.993456	0	1	-360	360;
	12422	10224	0.004413	0.039762	0	0	0	0	0.962152	0	1	-360	360;
	12422	10224	0.004498	0.038954	0	0	0	0	0.912809	0	1	-360	360;
	2719	8098	0.000621	0.0376	0	0	0	0	0.989172	0	1	-360	360;
	5101	8098	0.00063	0.0368	0	0	0	0	0.961251	0	1	-360	360;
	5101	11519	0.00064	0.036	0	0	0	0	1.032687	0	1	-360	360;
	424	5822	0.000555	0.040801	0	0	0	0	0.994301	0	1	-360	360;
	11893	2101	0.000546	0.0416	0	0	0	0	0.995285	0	1	-360	360;
	11893	11831	0.000527	0.0424	0	0	0	0	0.995648	0	1	-360	360;
	7325	1667	0.000456	0.043201	0	0	0	0	0.989708	0	1	-360	360;
	7325	1667	0.000447	0.040001	0	0	0	0	0.989326	0	1	-360	360;
	1289	1667	0.000505	0.039202	0	0	0	0	0.992153	0	1	-360	360;
	1289	1667	0.000513	0.038402	0	0	0	0	0.992281	0	1	-360	360;
	7067	6275	0.000513	0.037602	0	0	0	0	0.995172	0	1	-360	360;
	4355	6275	0.00052	0.036802	0	0	0	0	0.99467	0	1	-360	360;
	1592	6275	0.000529	0.036002	0	0	0	0	0.994917	0	1	-360	360;
	11934	9064	0.000473	0.040802	0	0	0	0	0.971823	0	1	-360	360;
	12977	9064	0.000465	0.041601	0	0	0	0	0.971689	0	1	-360	360;
	5967	9165	7e-05	0.021201	0	0	0	0	0.962697	0	1	-360	360;
	2853	6373	7e-05	0.021599	0	0	0	0	0.955482	0	1	-360	360;
	7545	4906	0.00054	0.04	0	0	0	0	0.98421	0	1	-360	360;
	7545	4906	0.000472	0.039202	0	0	0	0	1.053753	0	1	-360	360;
	3892	4906	0.00048	0.038402	0	0	0	0	1.042821	0	1	-360	360;
	3892	4906	0.00064	0.0376	0	0	0	0	0.98421	0	1	-360	360;
	2269	11567	0.000333	0.029441	0	0	0	0	0.982097	0	1	-360	360;
	8081	12550	0.000353	0.028801	0	0	0	0	0.925857	0	1	-360	360;
	8081	12550	0.000312	0.032641	0	0	0	0	0.991744	0	1	-360	360;
	2409	13039	0.000529	0.041601	0	0	0	0	0.955654	0	1	-360	360;
	1593	13039	0.000511	0.042401	0	0	0	0	0.96444	0	1	-360	360;
	12369	11060	0.000633	0.051521	0	0	0	0	0.989866	0	1	-360	360;
	12369	11060	0.000616	0.047701	0	0	0	0	0.969648	0	1	-360	360;
	1112	11060	0.000701	0.046752	0	0	0	0	0.97025	0	1	-360	360;
	4743	8827	0.000811	0.04579	0	0	0	0	0.937412	0	1	-360	360;
	547	8827	0.000831	0.04484	0	0	0	0	0.936876	0	1	-360	360;
	547	8827	0.000841	0.043881	0	0	0	0	0.935984	0	1	-360	360;
	8845	10912	0.000621	0.036001	0	0	0	0	1.032266	0	1	-360	360;
	8845	9802	0.000555	0.040801	0	0	0	0	1.023829	0	1	-360	360;
	11660	5649	0.000563	0.0416	0	0	0	0	1.022087	0	1	-360	360;
	2553	1456	0.000543	0.0424	0	0	0	0	0.946093	0	1	-360	360;
	11313	1848	0.001026	0.086398	0	0	0	0	0.955665	0	1	-360	360;
	199	1848	0.001007	0.080001	0	0	0	0	0.951634	0	1	-360	360;
	4653	2386	0.000522	0.039201	0	0	0	0	1.00279	0	1	-360	360;
	8261	2386	0.00053	0.038401	0	0	0	0	1.006204	0	1	-360	360;
	6366	2386	0.000548	0.037601	0	0	0	0	0.85156	0	1	-360	360;
	7003	2792	3e-05	0.00853	0	0	0	0	0.902577	0.445071	1	-360	360;
	6333	6307	0.000907	0.042929	0	0	0	0	0.978955	0	1	-360	360;
	10205	6237	7.7e-05	0.0204	0	0	0	0	1.114721	0	1	-360	360;
	8646	10205	3e-05	0.00975	0	0	0	0	0.902577	0.445071	1	-360	360;
	6237	6307	0.000773	0.05056	0	0	0	0	0.986156	0	1	-360	360;
	8593	3890	0.000728	0.043197	0	0	0	0	1.041291	0	1	-360	360;
	616	367	0.000447	0.040001	0	0	0	0	1.042918	0	1	-360	360;
	13170	8819	0.00061	0.0392	0	0	0	0	1.01718	0	1	-360	360;
	7752	12831	0.001203	0.076801	0	0	0	0	0.986892	0	1	-360	360;
	13242	12831	0.001223	0.075201	0	0	0	0	0.987006	0	1	-360	360;
	11462	10336	0.004533	0.037356	0	0	0	0	0.99379	0	1	-360	360;
	8768	520	0.004615	0.036549	0	0	0	0	0.99232	0	1	-360	360;
	9879	2388	0.004113	0.041392	0	0	0	0	0.993178	0	1	-360	360;
	5998	10902	0.000627	0.041599	0	0	0	0	1.044268	0	1	-360	360;
	1508	12184	0.001048	0.084802	0	0	0	0	1.000584	0	1	-360	360;
	5693	12184	0.001029	0.086397	0	0	0	0	1.000869	0	1	-360	360;
	113	7789	0.000808	0.047709	0	0	0	0	1.052369	0	1	-360	360;
	6578	4321	0.00061	0.0392	0	0	0	0	0.989376	0	1	-360	360;
	6578	4321	0.00062	0.0384	0	0	0	0	0.989186	0	1	-360	360;
	9167	4321	0.00064	0.0376	0	0	0	0	0.993068	0	1	-360	360;
	583	8829	0.00063	0.0368	0	0	0	0	0.968844	0	1	-360	360;
	8504	7652	0.00064	0.036	0	0	0	0	1.024568	0	1	-360	360;
	8504	7652	0.000573	0.0408	0	0	0	0	0.979515	0	1	-360	360;
	9483	3555	0.001067	0.083199	0	0	0	0	1.003168	0	1	-360	360;
	7724	3555	0.001048	0.084802	0	0	0	0	1.001883	0	1	-360	360;
	2040	9145	6e-05	0.11487	0	0	0	0	0	0	1	-360	360;
	6805	9145	0.20187	1.758347	0	0	0	0	0	0	1	-360	360;
	9356	5651	0.03911	0.342562	0	0	0	0	0	0	1	-360	360;
	4318	6737	0.01564	0.58593	0	0	0	0	0	0	1	-360	360;
	4318	6716	-0.000248	0.227727	0	0	0	0	0	0	1	-360	360;
	13169	1379	0.01414	0.513223	0	0	0	0	0	0	1	-360	360;
	5421	1379	-0.007335	0.347624	0	0	0	0	0	0	1	-360	360;
	8011	1379	-0.006901	0.404112	0	0	0	0	0	0	1	-360	360;
	11162	5015	0.051021	0.452583	0	0	0	0	0	0	1	-360	360;
	5572	933	0.586797	1.843698	0	0	0	0	0	0	1	-360	360;
	317	933	0.487562	1.572644	0	0	0	0	0	0	1	-360	360;
	7104	933	0.12936	0.588864	0	0	0	0	0	0	1	-360	360;
	9104	6113	0.000246	0.025121	0	0	0	0	0.961864	0	1	-360	360;
	6805	10689	0.10005	1.276798	0	0	0	0	0	0	1	-360	360;
	5421	10689	0.21936	1.401591	0	0	0	0	0	0	1	-360	360;
	8011	10689	0.22688	1.394834	0	0	0	0	0	0	1	-360	360;
	2040	3601	0.04986	0.379174	0	0	0	0	0	0	1	-360	360;
	5572	1363	0.02303	0.62419	0	0	0	0	0	0	1	-360	360;
	5572	611	0.03435	0.639633	0	0	0	0	0	0	1	-360	360;
	374	12450	0.01742	0.603616	0	0	0	0	0	0	1	-360	360;
	12364	12450	0.14245	1.97967	0	0	0	0	0	0	1	-360	360;
	1597	12450	-0.000145	0.32938	0	0	0	0	0	0	1	-360	360;
	6653	4669	-0.000298	0.102701	0	0	0	0	0	0	1	-360	360;
	530	9705	0.000708	0.03271	0	0	0	0	1.035157	0	1	-360	360;
	530	11541	0.06719	0.834339	0	0	0	0	0	0	1	-360	360;
	8923	6644	0.050909	0.307686	0	0	0	0	0	0	1	-360	360;
	1405	6644	0.212521	0.943657	0	0	0	0	0	0	1	-360	360;
	6246	12361	-0.002652	0.490201	0	0	0	0	0	0	1	-360	360;
	6246	8216	0.02628	0.438469	0	0	0	0	0	0	1	-360	360;
	12187	12361	-0.008241	0.388947	0	0	0	0	0	0	1	-360	360;
	11045	12361	0.000226	0.024837	0	0	0	0	1.036011	0	1	-360	360;
	3952	12361	0.106109	0.699359	0	0	0	0	0	0	1	-360	360;
	3952	8216	0.02855	0.418988	0	0	0	0	0	0	1	-360	360;
	1597	488	0.292479	1.229545	0	0	0	0	0	0	1	-360	360;
	317	4669	0.757066	1.631798	0	0	0	0	0	0	1	-360	360;
	7104	4669	0.384752	1.021653	0	0	0	0	0	0	1	-360	360;
	7601	6887	-0.014297	0.212851	0	0	0	0	0	0	1	-360	360;
	7601	6887	0.000444	0.031171	0	0	0	0	0.960552	0	1	-360	360;
	7601	5153	-0.014091	0.217459	0	0	0	0	0	0	1	-360	360;
	11162	12209	0.00894	0.300393	0	0	0	0	0	0	1	-360	360;
	13061	13263	0.049771	0.546674	0	0	0	0	0	0	1	-360	360;
	3159	12798	0.000737	0.026925	0	0	0	0	1.044741	0	1	-360	360;
	3159	12798	0.000845	0.026029	0	0	0	0	1.013524	0	1	-360	360;
	13061	4463	0.05157	0.475537	0	0	0	0	0	0	1	-360	360;
	11162	3967	0.054781	0.446446	0	0	0	0	0	0	1	-360	360;
	5039	2867	0.000452	0.023578	0	0	0	0	0.970414	0	1	-360	360;
	4911	4104	-0.010579	0.27814	0	0	0	0	0	0	1	-360	360;
	4911	7423	-0.007004	0.324484	0	0	0	0	0	0	1	-360	360;
	12364	8031	-0.034029	0.423698	0	0	0	0	0	0	1	-360	360;
	4071	2121	0.000384	0.0242	0	0	0	0	1.003202	0	1	-360	360;
	2040	2121	0.299587	1.849794	0	0	0	0	0	0	1	-360	360;
	2040	7741	0.301095	1.740868	0	0	0	0	0	0	1	-360	360;
	1585	7101	0.000759	0.019191	0	0	0	0	1.143233	0	1	-360	360;
	374	10265	0.170173	0.568199	0	0	0	0	0	0	1	-360	360;
	11162	9811	0.00708	0.453843	0	0	0	0	0	0	1	-360	360;
	13061	9811	0.057719	0.644008	0	0	0	0	0	0	1	-360	360;
	10360	10265	0.077609	0.938223	0	0	0	0	0	0	1	-360	360;
	1597	10265	0.200409	0.618946	0	0	0	0	0	0	1	-360	360;
	11162	4972	0.152771	0.863657	0	0	0	0	0	0	1	-360	360;
	13061	4972	-0.001921	0.311818	0	0	0	0	0	0	1	-360	360;
	1597	843	0.305992	1.948119	0	0	0	0	0	0	1	-360	360;
	13061	8020	0.17005	0.993099	0	0	0	0	0	0	1	-360	360;
	10338	2452	0.000257	0.029742	0	0	0	0	0.983674	0	1	-360	360;
	10338	4736	0.000268	0.029134	0	0	0	0	0.977352	0	1	-360	360;
	5525	7483	-0.014277	0.154039	0	0	0	0	0	0	1	-360	360;
	5525	7483	0.000738	0.031616	0	0	0	0	0.971178	0	1	-360	360;
	733	7483	0.251178	0.958099	0	0	0	0	0	0	1	-360	360;
	5525	561	0.08599	0.668788	0	0	0	0	0	0	1	-360	360;
	7328	11541	0.000241	0.02217	0	0	0	0	1.003292	0	1	-360	360;
	733	3208	-0.003678	0.380909	0	0	0	0	0	0	1	-360	360;
	12852	3208	-0.011302	0.627045	0	0	0	0	0	0	1	-360	360;
	733	2802	0.00797	0.426157	0	0	0	0	0	0	1	-360	360;
	12852	2802	0.058159	0.835806	0	0	0	0	0	0	1	-360	360;
	733	2754	0.00907	0.406178	0	0	0	0	0	0	1	-360	360;
	12852	2754	0.059789	0.790103	0	0	0	0	0	0	1	-360	360;
	733	561	0.04738	0.355028	0	0	0	0	0	0	1	-360	360;
	733	7843	0.00166	0.210118	0	0	0	0	0	0	1	-360	360;
	1405	1114	0.000234	0.03334	0	0	0	0	0.954161	0	1	-360	360;
	1405	11590	-0.007978	0.33365	0	0	0	0	0	0	1	-360	360;
	1405	13563	-0.007673	0.339439	0	0	0	0	0	0	1	-360	360;
	12852	12998	0.000213	0.026084	0	0	0	0	0.949952	0	1	-360	360;
	12852	561	0.079162	0.676053	0	0	0	0	0	0	1	-360	360;
	12852	5555	-0.012742	0.556163	0	0	0	0	0	0	1	-360	360;
	12852	7843	0.05625	0.586697	0	0	0	0	0	0	1	-360	360;
	9356	7690	0.11012	1.389876	0	0	0	0	0	0	1	-360	360;
	12364	4850	0.12839	0.634174	0	0	0	0	0	0	1	-360	360;
	12364	4850	0.000265	0.029415	0	0	0	0	0.954132	0	1	-360	360;
	4243	3694	0.01031	0.165651	0	0	0	0	0	0	1	-360	360;
	1141	10131	0.000906	0.032368	0	0	0	0	1.074593	0	1	-360	360;
	4911	10131	0.258574	1.91281	0	0	0	0	0	0	1	-360	360;
	4318	12597	0.114169	0.69781	0	0	0	0	0	0	1	-360	360;
	1597	10697	0.215475	0.921488	0	0	0	0	0	0	1	-360	360;
	1597	399	0.000388	0.026559	0	0	0	0	1.024188	0	1	-360	360;
	1597	399	-0.012727	0.168081	0	0	0	0	0	0	1	-360	360;
	1597	9019	0.11171	0.548317	0	0	0	0	0	0	1	-360	360;
	4243	5555	0.04893	0.453587	0	0	0	0	0	0	1	-360	360;
	4318	7534	0.000427	0.025869	0	0	0	0	1.014006	0	1	-360	360;
	4318	1998	0.000379	0.029319	0	0	0	0	1.029958	0	1	-360	360;
	11162	4325	0.086233	0.611399	0	0	0	0	0	0	1	-360	360;
	11162	5867	0.03295	0.399238	0	0	0	0	0	0	1	-360	360;
	13061	8240	0.034401	0.560661	0	0	0	0	0	0	1	-360	360;
	8472	3667	0.04848	0.4525	0	0	0	0	0	0	1	-360	360;
	8472	5950	0.05627	0.448553	0	0	0	0	0	0	1	-360	360;
	34	246	0.000262	0.020993	0	0	0	0	1.007843	0	1	-360	360;
	7104	246	0.093521	0.726302	0	0	0	0	0	0	1	-360	360;
	7104	9654	0.170831	0.824835	0	0	0	0	0	0	1	-360	360;
	9356	5867	0.05836	0.35759	0	0	0	0	0	0	1	-360	360;
	9162	4325	0.000563	0.031553	0	0	0	0	0.989721	0	1	-360	360;
	7104	5298	0.04447	0.619827	0	0	0	0	0	0	1	-360	360;
	9519	4873	0.001519	0.050701	0	0	0	0	1.013513	0	1	-360	360;
	186	7935	0.000821	0.051524	0	0	0	0	1.052632	0	1	-360	360;
	186	7935	0.000803	0.047869	0	0	0	0	1.052632	0	1	-360	360;
	647	6349	0.000466	0.029318	0	0	0	0	1.025291	0	1	-360	360;
	3465	6920	0.000463	0.028006	0	0	0	0	1.038781	0	1	-360	360;
	3465	6920	0.000463	0.026569	0	0	0	0	1.038781	0	1	-360	360;
	13247	12999	0.000556	0.030088	0	0	0	0	1.08147	0	1	-360	360;
	13247	12999	0.000581	0.029019	0	0	0	0	1.08147	0	1	-360	360;
	8480	1022	0.0017	0.077979	0	0	0	0	0	0	1	-360	360;
	12100	8027	0.000569	0.046371	0	0	0	0	1.052632	0	1	-360	360;
	12100	8027	0.00056	0.047255	0	0	0	0	1.052632	0	1	-360	360;
	789	6379	0.0008	0.050159	0	0	0	0	1.066857	0	1	-360	360;
	8865	7323	0.000379	0.039836	0	0	0	0	1.052632	0	1	-360	360;
	8865	7323	0.000433	0.039042	0	0	0	0	1.052632	0	1	-360	360;
	2715	10792	0.000448	0.032183	0	0	0	0	1.066857	0	1	-360	360;
	12487	1767	0.000498	0.034255	0	0	0	0	1.041667	0	1	-360	360;
	12487	1767	0.000507	0.033509	0	0	0	0	1.041667	0	1	-360	360;
	9043	3445	0.000415	0.040631	0	0	0	0	1.052632	0	1	-360	360;
	9043	3445	0.000406	0.041434	0	0	0	0	1.052632	0	1	-360	360;
	8085	1993	0.000397	0.037986	0	0	0	0	1.052632	0	1	-360	360;
	11490	2814	0.000322	0.035886	0	0	0	0	1.043149	0	1	-360	360;
	1463	10949	0.000408	0.039074	0	0	0	0	0.989653	0	1	-360	360;
	9196	9925	0.000417	0.026115	0	0	0	0	1.062288	0	1	-360	360;
	9196	9925	0.000443	0.02577	0	0	0	0	1.062288	0	1	-360	360;
	4330	9126	0.000396	0.033918	0	0	0	0	1.02924	0	1	-360	360;
	11719	254	0.000406	0.023438	0	0	0	0	1.052632	0	1	-360	360;
	11719	254	0.000352	0.026759	0	0	0	0	1.052632	0	1	-360	360;
	9534	10982	0.000334	0.036651	0	0	0	0	1.052632	0	1	-360	360;
	9534	10982	0.000334	0.037354	0	0	0	0	1.052632	0	1	-360	360;
	6592	6120	0.000305	0.035675	0	0	0	0	1.087225	0	1	-360	360;
	6592	6120	0.000296	0.033035	0	0	0	0	1.087225	0	1	-360	360;
	6592	6120	0.000338	0.032376	0	0	0	0	1.087225	0	1	-360	360;
	9511	12530	0.000285	0.02604	0	0	0	0	1.199891	0	1	-360	360;
	1463	1754	0.000419	0.02814	0	0	0	0	0.989653	0	1	-360	360;
	4330	13513	0.000453	0.025129	0	0	0	0	1.02924	0	1	-360	360;
	4330	9305	0.000453	0.02528	0	0	0	0	1.02924	0	1	-360	360;
	7395	677	0.00174	0.07467	0	0	0	0	1.049587	0	1	-360	360;
	7395	736	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	1925	8341	0.00172	0.099322	0	0	0	0	1.052632	0	1	-360	360;
	1925	8341	0.00168	0.101198	0	0	0	0	1.052632	0	1	-360	360;
	13051	10539	0.00363	0.146413	0	0	0	0	1.049587	0	1	-360	360;
	13051	10539	0.00182	0.07175	0	0	0	0	1.049587	0	1	-360	360;
	2469	9101	0.0019	0.08995	0	0	0	0	1.052632	0	1	-360	360;
	6954	7927	0.00189	0.06882	0	0	0	0	1.049587	0	1	-360	360;
	6954	1205	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	6146	10976	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	5486	1103	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	8051	8833	0.00167	0.0776	0	0	0	0	1.049587	0	1	-360	360;
	10871	4510	0.000646	0.061526	0	0	0	0	1.003288	-0.001353	1	-360	360;
	10871	4510	0.000636	0.056994	0	0	0	0	1.003288	-0.001353	1	-360	360;
	4510	922	0.00182	0.07175	0	0	0	0	1.049587	0	1	-360	360;
	4510	4852	0.00185	0.07028	0	0	0	0	1.049587	0	1	-360	360;
	3476	9139	0.00194	0.088083	0	0	0	0	1.052632	0	1	-360	360;
	3476	9888	0.00198	0.086207	0	0	0	0	1.052632	0	1	-360	360;
	6415	10002	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	11513	9234	0.000697	0.032338	0	0	0	0	1.002506	0	1	-360	360;
	9234	13293	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	10593	3637	0.000667	0.033601	0	0	0	0	1.002506	0	1	-360	360;
	10593	3637	0.000647	0.034238	0	0	0	0	1.002506	0	1	-360	360;
	3637	4361	0.0016	0.07553	0	0	0	0	1.061984	0	1	-360	360;
	3637	1683	0.00182	0.07402	0	0	0	0	1.061984	0	1	-360	360;
	6935	8512	0.0019	0.08995	0	0	0	0	1.052632	0	1	-360	360;
	8186	5583	0.00189	0.06882	0	0	0	0	1.049587	0	1	-360	360;
	8186	899	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	9116	127	0.00201	0.084331	0	0	0	0	1.052632	0	1	-360	360;
	9116	127	0.00179	0.095578	0	0	0	0	1.052632	0	1	-360	360;
	13638	2924	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	5480	1293	0.00167	0.0776	0	0	0	0	1.049587	0	1	-360	360;
	1193	2833	0.00168	0.101198	0	0	0	0	1.052632	0	1	-360	360;
	9834	12952	0.00165	0.093702	0	0	0	0	1.052632	0	1	-360	360;
	9834	2002	0.00187	0.091826	0	0	0	0	1.052632	0	1	-360	360;
	9238	13204	0.0019	0.08995	0	0	0	0	1.052632	0	1	-360	360;
	5579	13378	0.00189	0.06882	0	0	0	0	1.049587	0	1	-360	360;
	5579	13378	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	7281	10770	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	7281	9459	0.00174	0.07467	0	0	0	0	1.049587	0	1	-360	360;
	11589	4560	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	5719	7158	0.00167	0.0776	0	0	0	0	1.049587	0	1	-360	360;
	2986	11540	0.00164	0.07907	0	0	0	0	1.049587	0	1	-360	360;
	13627	9770	0.00045	0.04299	0	0	0	0	1.113821	0	1	-360	360;
	13315	4641	0.00051	0.04213	0	0	0	0	1.113821	0	1	-360	360;
	1705	1706	0.00185	0.07028	0	0	0	0	1.049587	0	1	-360	360;
	1705	1706	0.00189	0.06882	0	0	0	0	1.049587	0	1	-360	360;
	293	6175	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	293	8417	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	7326	5374	0.00179	0.095578	0	0	0	0	1.052632	0	1	-360	360;
	7326	5374	0.00176	0.097446	0	0	0	0	1.052632	0	1	-360	360;
	2884	1799	0.00167	0.0776	0	0	0	0	1.049587	0	1	-360	360;
	2884	1799	0.00164	0.07907	0	0	0	0	1.049587	0	1	-360	360;
	12012	8051	0.000637	0.031701	0	0	0	0	1.002506	0	1	-360	360;
	12012	11679	0.000716	0.031064	0	0	0	0	1.002506	0	1	-360	360;
	11679	13236	0.00185	0.07028	0	0	0	0	1.049587	0	1	-360	360;
	11679	1890	0.00189	0.06882	0	0	0	0	1.049587	0	1	-360	360;
	3411	2744	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	3411	2744	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	9488	12371	0.00174	0.07467	0	0	0	0	1.049587	0	1	-360	360;
	11682	7764	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	9128	2511	0.00084	0.0322	0	0	0	0	1.052632	0	1	-360	360;
	2511	4297	0.00238	0.123248	0	0	0	0	1.049587	0	1	-360	360;
	7984	10393	0.00269	0.109562	0	0	0	0	1.049587	0	1	-360	360;
	8931	12463	0.00275	0.107281	0	0	0	0	1.049587	0	1	-360	360;
	6941	4396	0.000766	0.029164	0	0	0	0	1.002506	0	1	-360	360;
	6941	4396	0.000776	0.028526	0	0	0	0	1.002506	0	1	-360	360;
	4396	12301	0.00174	0.07467	0	0	0	0	1.049587	0	1	-360	360;
	4396	3435	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	12860	2548	0.00167	0.0776	0	0	0	0	1.049587	0	1	-360	360;
	5789	2671	0.00168	0.101198	0	0	0	0	1.052632	0	1	-360	360;
	5789	5487	0.00165	0.093702	0	0	0	0	1.052632	0	1	-360	360;
	1371	10664	0.000716	0.031064	0	0	0	0	1.002506	0	1	-360	360;
	1371	5020	0.000736	0.030437	0	0	0	0	1.002506	0	1	-360	360;
	5020	703	0.00189	0.06882	0	0	0	0	1.049587	0	1	-360	360;
	5020	703	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	9741	6410	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	11491	6657	0.00174	0.07467	0	0	0	0	1.049587	0	1	-360	360;
	2385	9239	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	2385	9239	0.00167	0.0776	0	0	0	0	1.049587	0	1	-360	360;
	6453	7040	0.00168	0.101198	0	0	0	0	1.052632	0	1	-360	360;
	6889	5659	0.00062	0.03068	0	0	0	0	1.015698	0	1	-360	360;
	6889	9703	0.000458	0.027443	0	0	0	0	1.002506	0	1	-360	360;
	6889	9703	0.000468	0.026875	0	0	0	0	1.002506	0	1	-360	360;
	12516	13461	0.001543	0.057853	0	0	0	0	0.892857	0	1	-360	360;
	6261	1121	0.00198	0.086207	0	0	0	0	1.052632	0	1	-360	360;
	10999	4447	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	4908	8684	0.00174	0.07467	0	0	0	0	1.049587	0	1	-360	360;
	2416	9838	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	4894	13139	0.000667	0.033601	0	0	0	0	1.002506	0	1	-360	360;
	4894	11041	0.00168	0.101198	0	0	0	0	1.052632	0	1	-360	360;
	13139	11041	0.0016	0.07321	0	0	0	0	1.049587	0	1	-360	360;
	10698	11075	0.0019	0.08995	0	0	0	0	1.052632	0	1	-360	360;
	13221	5372	0.000746	0.029801	0	0	0	0	1.002506	0	1	-360	360;
	6364	5597	0.00198	0.086207	0	0	0	0	1.052632	0	1	-360	360;
	7782	4362	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	12756	7306	0.00174	0.07467	0	0	0	0	1.049587	0	1	-360	360;
	883	10413	0.000677	0.032965	0	0	0	0	1.002506	0	1	-360	360;
	883	12669	0.000667	0.033601	0	0	0	0	1.002506	0	1	-360	360;
	883	1870	0.00168	0.101198	0	0	0	0	1.052632	0	1	-360	360;
	883	7301	0.00165	0.093702	0	0	0	0	1.052632	0	1	-360	360;
	10256	6914	0.000716	0.031064	0	0	0	0	1.002506	0	1	-360	360;
	10256	6914	0.000736	0.030437	0	0	0	0	1.002506	0	1	-360	360;
	10256	6914	0.000746	0.029801	0	0	0	0	1.002506	0	1	-360	360;
	6914	5738	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	6914	5738	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	2204	2960	0.00179	0.095578	0	0	0	0	1.052632	0	1	-360	360;
	2204	5768	0.00176	0.097446	0	0	0	0	1.052632	0	1	-360	360;
	12015	7629	0.00167	0.0776	0	0	0	0	1.049587	0	1	-360	360;
	7303	47	0.00164	0.07907	0	0	0	0	1.049587	0	1	-360	360;
	3047	744	0.0016	0.07321	0	0	0	0	1.049587	0	1	-360	360;
	3047	744	0.00182	0.07175	0	0	0	0	1.049587	0	1	-360	360;
	13036	12177	0.0019	0.08995	0	0	0	0	1.052632	0	1	-360	360;
	13036	12177	0.00194	0.088083	0	0	0	0	1.052632	0	1	-360	360;
	595	12177	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	10654	13239	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	7575	12496	0.00174	0.07467	0	0	0	0	1.049587	0	1	-360	360;
	6343	6387	0.00387	0.152273	0	0	0	0	1.049587	0	1	-360	360;
	6989	398	0.000667	0.033601	0	0	0	0	1.002506	0	1	-360	360;
	6989	4571	0.00168	0.101198	0	0	0	0	1.052632	0	1	-360	360;
	6989	7138	0.00165	0.093702	0	0	0	0	1.052632	0	1	-360	360;
	6474	11554	0.00182	0.07175	0	0	0	0	1.049587	0	1	-360	360;
	6474	11554	0.00185	0.07028	0	0	0	0	1.049587	0	1	-360	360;
	10584	1325	0.00189	0.06882	0	0	0	0	1.049587	0	1	-360	360;
	10584	1325	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	5758	876	0.00196	0.07015	0	0	0	0	1.07438	0	1	-360	360;
	1619	876	0.00174	0.0795	0	0	0	0	1.07438	0	1	-360	360;
	4169	3871	0.00136	0.05901	0	0	0	0	1.052632	0	1	-360	360;
	4169	838	0.00133	0.06014	0	0	0	0	1.052632	0	1	-360	360;
	1383	8207	0.00164	0.07907	0	0	0	0	1.049587	0	1	-360	360;
	1383	3332	0.0016	0.07321	0	0	0	0	1.049587	0	1	-360	360;
	1720	11215	0.000716	0.031064	0	0	0	0	1.002506	0	1	-360	360;
	11215	6434	0.00185	0.07028	0	0	0	0	1.049587	0	1	-360	360;
	7313	3125	0.00189	0.06882	0	0	0	0	1.049587	0	1	-360	360;
	6514	219	0.00158	0.05564	0	0	0	0	0	0	1	-360	360;
	6514	8607	0.00179	0.0695	0	0	0	0	0	0	1	-360	360;
	4227	504	0.00084	0.04445	0	0	0	0	0	0	1	-360	360;
	4227	504	0.0008	0.04546	0	0	0	0	0	0	1	-360	360;
	4227	504	0.00219	0.08229	0	0	0	0	0	0	1	-360	360;
	4227	504	0.00069	0.04566	0	0	0	0	1.047826	0	1	-360	360;
	13329	5775	0.00072	0.04835	0	0	0	0	0.999313	0	1	-360	360;
	13329	5775	0.00082	0.04738	0	0	0	0	0.999313	0	1	-360	360;
	7771	3356	0.00045	0.03151	0	0	0	0	1.052632	0	1	-360	360;
	7771	12489	0.00048	0.026419	0	0	0	0	1.052632	0	1	-360	360;
	4491	12490	0.0022	0.07023	0	0	0	0	0	0	1	-360	360;
	4095	329	0.00241	0.0727	0	0	0	0	1.052174	0	1	-360	360;
	4095	329	0.00081	0.04644	0	0	0	0	1.052174	0	1	-360	360;
	7786	4358	0.00158	0.08045	0	0	0	0	0	0	1	-360	360;
	11494	13273	0.00069	0.04707	0	0	0	0	1.052174	0	1	-360	360;
	2195	12231	0.00075	0.04866	0	0	0	0	1.052174	0	1	-360	360;
	2195	12231	0.00076	0.04586	0	0	0	0	1.052174	0	1	-360	360;
	12298	12644	0.00189	0.07781	0	0	0	0	0	0	1	-360	360;
	12298	12644	0.00205	0.07804	0	0	0	0	0	0	1	-360	360;
	967	11127	0.00207	0.091083	0	0	0	0	0	0	1	-360	360;
	967	11127	0.00211	0.06802	0	0	0	0	0	0	1	-360	360;
	2818	11663	0.0008	0.03997	0	0	0	0	1.052174	0	1	-360	360;
	10005	11663	0.00159	0.086876	0	0	0	0	0	0	1	-360	360;
	10005	11663	0.0026	0.07998	0	0	0	0	0	0	1	-360	360;
	7125	11014	0.00154	0.086488	0	0	0	0	0	0	1	-360	360;
	12632	7125	0.00052	0.033159	0	0	0	0	1.052632	0	1	-360	360;
	12632	11014	0.00072	0.0451	0	0	0	0	1.052174	0	1	-360	360;
	12632	11014	0.00068	0.04531	0	0	0	0	1.052174	0	1	-360	360;
	7029	9828	0.00042	0.030281	0	0	0	0	1.052632	0	1	-360	360;
	7029	12354	0.00079	0.04086	0	0	0	0	1.052174	0	1	-360	360;
	9828	7611	0.00168	0.07343	0	0	0	0	0	0	1	-360	360;
	10798	3902	0.00038	0.033591	0	0	0	0	1.052632	0	1	-360	360;
	10798	3902	0.00037	0.03405	0	0	0	0	1.052632	0	1	-360	360;
	2405	12731	0.00039	0.034849	0	0	0	0	1.052632	0	1	-360	360;
	2405	5508	0.00063	0.04764	0	0	0	0	1.052174	0	1	-360	360;
	2405	5508	0.00074	0.044626	0	0	0	0	1.006865	0	1	-360	360;
	10718	11533	0.00069	0.04232	0	0	0	0	1.052632	0	1	-360	360;
	11733	9765	0.00228	0.07286	0	0	0	0	0	0	1	-360	360;
	11733	9765	0.00189	0.07059	0	0	0	0	0	0	1	-360	360;
	13546	2444	0.00195	0.07374	0	0	0	0	0	0	1	-360	360;
	13546	2444	0.00199	0.07214	0	0	0	0	0	0	1	-360	360;
	5434	8201	0.00177	0.07931	0	0	0	0	0	0	1	-360	360;
	5434	8201	0.00174	0.08086	0	0	0	0	0	0	1	-360	360;
	11413	7033	0.0015	0.08239	0	0	0	0	0	0	1	-360	360;
	11413	7033	0.00149	0.091992	0	0	0	0	0	0	1	-360	360;
	12313	5079	0.00066	0.04441	0	0	0	0	1.052174	0	1	-360	360;
	2505	5079	0.00171	0.07418	0	0	0	0	0	0	1	-360	360;
	2505	5079	0.00175	0.07267	0	0	0	0	0	0	1	-360	360;
	3715	7184	0.00053	0.03038	0	0	0	0	1.052632	0	1	-360	360;
	7184	868	0.00165	0.07003	0	0	0	0	0	0	1	-360	360;
	7184	67	0.00217	0.07323	0	0	0	0	0	0	1	-360	360;
	387	2022	0.00072	0.0453	0	0	0	0	1.052174	0	1	-360	360;
	11007	1568	0.0021	0.08257	0	0	0	0	0	0	1	-360	360;
	11007	1568	0.00186	0.087719	0	0	0	0	0	0	1	-360	360;
	1255	13260	0.00076	0.04796	0	0	0	0	1.052174	0	1	-360	360;
	1255	4701	0.00075	0.04466	0	0	0	0	1.052174	0	1	-360	360;
	5511	11269	0.00401	0.10005	0	0	0	0	0	0	1	-360	360;
	5511	11269	0.00272	0.07842	0	0	0	0	0	0	1	-360	360;
	5511	11269	0.00278	0.07679	0	0	0	0	0	0	1	-360	360;
	5511	11269	0.00283	0.07515	0	0	0	0	0	0	1	-360	360;
	9582	12567	0.00045	0.029271	0	0	0	0	1.052632	0	1	-360	360;
	9582	12567	0.0004	0.033169	0	0	0	0	1.052632	0	1	-360	360;
	9582	12567	0.00039	0.03382	0	0	0	0	1.052632	0	1	-360	360;
	3923	7581	0.00064	0.04577	0	0	0	0	1.052632	0	1	-360	360;
	1011	2798	0.00147	0.08395	0	0	0	0	0	0	1	-360	360;
	1011	2532	0.00184	0.07633	0	0	0	0	0	0	1	-360	360;
	8891	13588	0.00074	0.04352	0	0	0	0	1.052174	0	1	-360	360;
	2095	7010	0.00176	0.08177	0	0	0	0	0	0	1	-360	360;
	2095	7010	0.00242	0.06613	0	0	0	0	0	0	1	-360	360;
	7215	3071	0.00042	0.030031	0	0	0	0	0.99815	0	1	-360	360;
	7215	3071	0.00043	0.02938	0	0	0	0	0.998666	0	1	-360	360;
	10322	13504	0.00014	0.019637	0	0	0	0	0	0.199334	1	-360	360;
	8477	11977	0.00037	0.03395	0	0	0	0	0.969972	0	1	-360	360;
	6044	5814	1.8e-05	0.052587	0	0	0	0	1.052632	0	1	-360	360;
	4082	5814	5.5e-05	0.082396	0	0	0	0	1.04762	0	1	-360	360;
	1144	12959	0.001416	0.043162	0	0	0	0	0.92437	0	1	-360	360;
	1144	12959	0.001603	0.042296	0	0	0	0	0.92437	0	1	-360	360;
	6149	5799	0.000806	0.025741	0	0	0	0	1.002506	0	1	-360	360;
	6149	5799	0.000826	0.025204	0	0	0	0	1.002506	0	1	-360	360;
	6149	10826	0.000107	0.01357	0	0	0	0	1.059051	0	1	-360	360;
	5799	1351	0.00228	0.07939	0	0	0	0	0	0	1	-360	360;
	10826	5799	0.001087	0.043438	0	0	0	0	1.000679	-0.007974	1	-360	360;
	10826	5799	0.001067	0.077593	0	0	0	0	1.00317	-0.04898	1	-360	360;
	8882	8560	0.001301	0.061931	0	0	0	0	0.948276	0	1	-360	360;
	8882	8560	0.001268	0.063098	0	0	0	0	0.948276	0	1	-360	360;
	13474	9599	0.002358	0.088053	0	0	0	0	0.948276	0	1	-360	360;
	13474	9599	0.00268	0.086295	0	0	0	0	0.948276	0	1	-360	360;
	11561	879	0.000779	0.042361	0	0	0	0	1.061123	0.005681	1	-360	360;
	879	6231	0.001565	0.062128	0	0	0	0	0.932204	0	1	-360	360;
	879	6231	0.001588	0.060806	0	0	0	0	0.932204	0	1	-360	360;
	7850	11835	0.002713	0.092944	0	0	0	0	0.916667	0	1	-360	360;
	7850	11835	0.002464	0.122138	0	0	0	0	0.916667	0	1	-360	360;
	7397	657	0.002739	0.116189	0	0	0	0	0.932204	0	1	-360	360;
	7397	657	0.002681	0.118434	0	0	0	0	0.932204	0	1	-360	360;
	9363	12633	0.000212	0.013811	0	0	0	0	1.067899	0.133383	1	-360	360;
	12894	5804	0.005208	0.208268	0	0	0	0	1.023076	0	1	-360	360;
	12894	7115	0.005208	0.208268	0	0	0	0	1.023076	0	1	-360	360;
	12894	3481	0.005208	0.208268	0	0	0	0	1.023076	0	1	-360	360;
	12894	4910	0.005208	0.208268	0	0	0	0	1.023076	0	1	-360	360;
	12894	6291	0.005208	0.208268	0	0	0	0	1.023076	0	1	-360	360;
	12894	7625	0.005208	0.208268	0	0	0	0	1.023076	0	1	-360	360;
	12894	10785	0.005208	0.208268	0	0	0	0	1.023076	0	1	-360	360;
	12894	7057	0.005208	0.208268	0	0	0	0	1.023076	0	1	-360	360;
	12894	8370	0.005208	0.208268	0	0	0	0	1.023076	0	1	-360	360;
	12894	8159	0.005208	0.208268	0	0	0	0	1.023076	0	1	-360	360;
	7344	4480	0.005208	0.208268	0	0	0	0	1.019827	0	1	-360	360;
	7344	6135	0.005208	0.208268	0	0	0	0	1.019827	0	1	-360	360;
	3901	10983	0.005208	0.208268	0	0	0	0	1.01416	0	1	-360	360;
	3901	564	0.005208	0.208268	0	0	0	0	1.01416	0	1	-360	360;
	3901	3617	0.005208	0.208268	0	0	0	0	1.01416	0	1	-360	360;
	12265	6846	0.005208	0.208268	0	0	0	0	1.02288	0	1	-360	360;
	12265	9976	0.005208	0.208268	0	0	0	0	1.02288	0	1	-360	360;
	12265	1895	0.005208	0.208268	0	0	0	0	1.02288	0	1	-360	360;
	372	4692	0.000531	0.021225	0	0	0	0	1.024035	0	1	-360	360;
	372	5627	0.000531	0.021225	0	0	0	0	1.024035	0	1	-360	360;
	10709	6780	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	3577	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	531	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	7105	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	10352	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	2473	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	13122	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	6049	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	2786	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	8524	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	11699	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	1192	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	4298	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	7447	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	10605	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	12035	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	1516	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	4671	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	10709	6225	0.005208	0.208268	0	0	0	0	1.021194	0	1	-360	360;
	8477	2462	0.000531	0.021225	0	0	0	0	1.015768	0	1	-360	360;
	8477	6317	0.000531	0.021225	0	0	0	0	1.015768	0	1	-360	360;
	8477	2084	0.000531	0.021225	0	0	0	0	1.015768	0	1	-360	360;
	11379	1606	0.002119	0.084719	0	0	0	0	1.086887	0	1	-360	360;
	11379	6885	0.002119	0.084719	0	0	0	0	1.086887	0	1	-360	360;
	11379	9960	0.002119	0.084719	0	0	0	0	1.086887	0	1	-360	360;
	11379	10847	0.002119	0.084719	0	0	0	0	1.086887	0	1	-360	360;
	11379	7930	0.000531	0.021225	0	0	0	0	1.081998	0	1	-360	360;
	11379	9687	0.000531	0.021225	0	0	0	0	1.081998	0	1	-360	360;
	11379	12962	0.000531	0.021225	0	0	0	0	1.081998	0	1	-360	360;
	12770	7173	0.002119	0.084719	0	0	0	0	1.083997	0	1	-360	360;
	12770	3962	0.002119	0.084719	0	0	0	0	1.083997	0	1	-360	360;
	12770	910	0.002119	0.084719	0	0	0	0	1.083997	0	1	-360	360;
	3896	1096	0.005208	0.208268	0	0	0	0	1.007824	0	1	-360	360;
	3896	10865	0.005208	0.208268	0	0	0	0	1.007824	0	1	-360	360;
	3896	7689	0.005208	0.208268	0	0	0	0	1.007824	0	1	-360	360;
	3896	5765	0.005208	0.208268	0	0	0	0	1.007824	0	1	-360	360;
	5694	7356	0.005208	0.208268	0	0	0	0	1.017619	0	1	-360	360;
	5694	11891	0.005208	0.208268	0	0	0	0	1.017619	0	1	-360	360;
	5694	8704	0.005208	0.208268	0	0	0	0	1.017619	0	1	-360	360;
	5694	5739	0.005208	0.208268	0	0	0	0	1.017619	0	1	-360	360;
	6184	8793	0.005208	0.208268	0	0	0	0	1.014202	0	1	-360	360;
	6184	11996	0.005208	0.208268	0	0	0	0	1.014202	0	1	-360	360;
	6184	9236	0.005208	0.208268	0	0	0	0	1.014202	0	1	-360	360;
	6184	6799	0.005208	0.208268	0	0	0	0	1.014202	0	1	-360	360;
	12700	9884	0.005208	0.208268	0	0	0	0	1.012597	0	1	-360	360;
	12700	13133	0.005208	0.208268	0	0	0	0	1.012597	0	1	-360	360;
	12700	2485	0.005208	0.208268	0	0	0	0	1.012597	0	1	-360	360;
	12700	5420	0.005208	0.208268	0	0	0	0	1.012597	0	1	-360	360;
	12700	8490	0.005208	0.208268	0	0	0	0	1.012597	0	1	-360	360;
	12700	7064	0.005208	0.208268	0	0	0	0	1.012597	0	1	-360	360;
	3649	3854	0.005208	0.208268	0	0	0	0	1.012275	0	1	-360	360;
	3649	798	0.005208	0.208268	0	0	0	0	1.012275	0	1	-360	360;
	3649	13409	0.005208	0.208268	0	0	0	0	1.012275	0	1	-360	360;
	3649	8096	0.005208	0.208268	0	0	0	0	1.012275	0	1	-360	360;
	10587	4448	0.000531	0.021225	0	0	0	0	1.080715	0	1	-360	360;
	10587	1308	0.000531	0.021225	0	0	0	0	1.080715	0	1	-360	360;
	11210	13327	0.000531	0.021225	0	0	0	0	1.051932	0	1	-360	360;
	12488	5319	0.000531	0.021225	0	0	0	0	1.071813	0	1	-360	360;
	2551	5581	0.000531	0.021225	0	0	0	0	1.034901	0	1	-360	360;
	2551	11483	0.000531	0.021225	0	0	0	0	1.034901	0	1	-360	360;
	2551	6436	0.000531	0.021225	0	0	0	0	1.034901	0	1	-360	360;
	8575	1323	0.000531	0.021225	0	0	0	0	1.080717	0	1	-360	360;
	8575	4443	0.000531	0.021225	0	0	0	0	1.080717	0	1	-360	360;
	8575	8632	0.000531	0.021225	0	0	0	0	1.080717	0	1	-360	360;
	12933	10535	0.000531	0.021225	0	0	0	0	0.974802	0	1	-360	360;
	12933	2895	0.000531	0.021225	0	0	0	0	0.974802	0	1	-360	360;
	4568	9370	0.000531	0.021225	0	0	0	0	1.02559	0	1	-360	360;
	4568	9281	0.000531	0.021225	0	0	0	0	1.02559	0	1	-360	360;
	7622	8763	0.005208	0.208268	0	0	0	0	1.01765	0	1	-360	360;
	7622	2777	0.005208	0.208268	0	0	0	0	1.01765	0	1	-360	360;
	7622	1455	0.005208	0.208268	0	0	0	0	1.01765	0	1	-360	360;
	7622	6507	0.005208	0.208268	0	0	0	0	1.01765	0	1	-360	360;
	7622	6034	0.005208	0.208268	0	0	0	0	1.01765	0	1	-360	360;
	7622	1749	0.005208	0.208268	0	0	0	0	1.01765	0	1	-360	360;
	7622	5924	0.005208	0.208268	0	0	0	0	1.01765	0	1	-360	360;
	7622	458	0.005208	0.208268	0	0	0	0	1.01765	0	1	-360	360;
	4568	6754	0.000531	0.021225	0	0	0	0	1.025589	0	1	-360	360;
	4568	39	0.000531	0.021225	0	0	0	0	1.025589	0	1	-360	360;
	12284	1125	0.005208	0.208268	0	0	0	0	1.013076	0	1	-360	360;
	12284	7668	0.005208	0.208268	0	0	0	0	1.013076	0	1	-360	360;
	12284	666	0.005208	0.208268	0	0	0	0	1.013076	0	1	-360	360;
	12284	824	0.005208	0.208268	0	0	0	0	1.013076	0	1	-360	360;
	3818	6103	0.000531	0.021225	0	0	0	0	1.007697	0	1	-360	360;
	11951	1072	0.002119	0.084719	0	0	0	0	1.026731	0	1	-360	360;
	11951	5341	0.002119	0.084719	0	0	0	0	1.026731	0	1	-360	360;
	11951	11299	0.002119	0.084719	0	0	0	0	1.026731	0	1	-360	360;
	2678	2970	0.005208	0.208268	0	0	0	0	1.024028	0	1	-360	360;
	2678	5037	0.005208	0.208268	0	0	0	0	1.024028	0	1	-360	360;
	2678	12099	0.005208	0.208268	0	0	0	0	1.024028	0	1	-360	360;
	2678	5548	0.005208	0.208268	0	0	0	0	1.024028	0	1	-360	360;
	2678	12612	0.005208	0.208268	0	0	0	0	1.024028	0	1	-360	360;
	2678	8969	0.005208	0.208268	0	0	0	0	1.024028	0	1	-360	360;
	6551	10408	0.005208	0.208268	0	0	0	0	1.025318	0	1	-360	360;
	6551	6539	0.005208	0.208268	0	0	0	0	1.025318	0	1	-360	360;
	6551	13651	0.005208	0.208268	0	0	0	0	1.025318	0	1	-360	360;
	4813	1278	0.005208	0.208268	0	0	0	0	1.030195	0	1	-360	360;
	4813	2662	0.005208	0.208268	0	0	0	0	1.030195	0	1	-360	360;
	4813	888	0.005208	0.208268	0	0	0	0	1.030195	0	1	-360	360;
	4813	7392	0.005208	0.208268	0	0	0	0	1.030195	0	1	-360	360;
	4813	1324	0.005208	0.208268	0	0	0	0	1.030195	0	1	-360	360;
	4813	7875	0.005208	0.208268	0	0	0	0	1.030195	0	1	-360	360;
	12680	12641	0.000708	0.02832	0	0	0	0	1.067331	0	1	-360	360;
	12680	7111	0.000708	0.02832	0	0	0	0	1.067331	0	1	-360	360;
	679	10640	0.000708	0.02832	0	0	0	0	1.072713	0	1	-360	360;
	679	11154	0.000708	0.02832	0	0	0	0	1.072713	0	1	-360	360;
	2993	4310	0.000531	0.021225	0	0	0	0	1.074993	0	1	-360	360;
	12838	1071	0.005208	0.208268	0	0	0	0	1.009536	0	1	-360	360;
	12838	6948	0.005208	0.208268	0	0	0	0	1.009536	0	1	-360	360;
	4783	394	0.005208	0.208268	0	0	0	0	1.003735	0	1	-360	360;
	4783	10418	0.005208	0.208268	0	0	0	0	1.003735	0	1	-360	360;
	4783	869	0.005208	0.208268	0	0	0	0	1.003735	0	1	-360	360;
	5808	7859	0.005208	0.208268	0	0	0	0	1.017956	0	1	-360	360;
	5808	1305	0.005208	0.208268	0	0	0	0	1.017956	0	1	-360	360;
	5808	9614	0.005208	0.208268	0	0	0	0	1.017956	0	1	-360	360;
	5808	1763	0.005208	0.208268	0	0	0	0	1.017956	0	1	-360	360;
	5808	4988	0.005208	0.208268	0	0	0	0	1.017956	0	1	-360	360;
	1613	4339	0.00547	0.21875	0	0	0	0	1.01239	0	1	-360	360;
	1613	4938	0.00547	0.21875	0	0	0	0	1.01239	0	1	-360	360;
	1613	7039	0.00547	0.21875	0	0	0	0	1.006664	0	1	-360	360;
	1613	8029	0.00547	0.21875	0	0	0	0	1.006664	0	1	-360	360;
	1613	1017	0.00547	0.21875	0	0	0	0	1.006664	0	1	-360	360;
	11260	6670	0.005208	0.208268	0	0	0	0	1.009764	0	1	-360	360;
	11260	13292	0.005208	0.208268	0	0	0	0	1.009764	0	1	-360	360;
	11260	7116	0.005208	0.208268	0	0	0	0	1.009764	0	1	-360	360;
	11260	133	0.005208	0.208268	0	0	0	0	1.009764	0	1	-360	360;
	11260	5723	0.005208	0.208268	0	0	0	0	1.009764	0	1	-360	360;
	11260	9374	0.005208	0.208268	0	0	0	0	1.009764	0	1	-360	360;
	11260	4440	0.005208	0.208268	0	0	0	0	1.009764	0	1	-360	360;
	11260	11487	0.005208	0.208268	0	0	0	0	1.009764	0	1	-360	360;
	11260	3915	0.005208	0.208268	0	0	0	0	1.009764	0	1	-360	360;
	11260	11005	0.005208	0.208268	0	0	0	0	1.009764	0	1	-360	360;
	13548	5439	0.005208	0.208268	0	0	0	0	1.0091	0	1	-360	360;
	13548	6068	0.005208	0.208268	0	0	0	0	1.0091	0	1	-360	360;
	13548	4934	0.005208	0.208268	0	0	0	0	1.0091	0	1	-360	360;
	13548	11988	0.005208	0.208268	0	0	0	0	1.0091	0	1	-360	360;
	13548	2459	0.005208	0.208268	0	0	0	0	1.0091	0	1	-360	360;
	13548	3248	0.005208	0.208268	0	0	0	0	1.0091	0	1	-360	360;
	13548	1304	0.005208	0.208268	0	0	0	0	1.0091	0	1	-360	360;
	13548	7858	0.005208	0.208268	0	0	0	0	1.0091	0	1	-360	360;
	13548	859	0.005208	0.208268	0	0	0	0	1.0091	0	1	-360	360;
	13548	7365	0.005208	0.208268	0	0	0	0	1.0091	0	1	-360	360;
	2873	3292	0.002119	0.084719	0	0	0	0	1.0106	0	1	-360	360;
	2873	11643	0.002119	0.084719	0	0	0	0	1.0106	0	1	-360	360;
	2873	626	0.002119	0.084719	0	0	0	0	1.0106	0	1	-360	360;
	2873	11131	0.002119	0.084719	0	0	0	0	1.0106	0	1	-360	360;
	2873	4063	0.002119	0.084719	0	0	0	0	1.0106	0	1	-360	360;
	7748	9766	0.004587	0.183429	0	0	0	0	1.016154	0	1	-360	360;
	7748	11802	0.004587	0.183429	0	0	0	0	1.010452	0	1	-360	360;
	7748	4377	0.004587	0.183429	0	0	0	0	1.010452	0	1	-360	360;
	7748	5940	0.004587	0.183429	0	0	0	0	1.010452	0	1	-360	360;
	7748	5755	0.004587	0.183429	0	0	0	0	1.010452	0	1	-360	360;
	7748	10833	0.004587	0.183429	0	0	0	0	1.010452	0	1	-360	360;
	7748	4238	0.004587	0.183429	0	0	0	0	1.010452	0	1	-360	360;
	7748	181	0.004587	0.183429	0	0	0	0	1.010452	0	1	-360	360;
	7748	4761	0.004587	0.183429	0	0	0	0	1.010452	0	1	-360	360;
	3764	144	0.005208	0.208268	0	0	0	0	1.009661	0	1	-360	360;
	3764	7139	0.005208	0.208268	0	0	0	0	1.009661	0	1	-360	360;
	3764	12551	0.005208	0.208268	0	0	0	0	1.009661	0	1	-360	360;
	7231	7252	0.000212	0.008494	0	0	0	0	1.02036	0	1	-360	360;
	7231	241	0.000212	0.008494	0	0	0	0	1.02036	0	1	-360	360;
	4303	5979	0.005208	0.208268	0	0	0	0	1.021922	0	1	-360	360;
	4303	12282	0.005208	0.208268	0	0	0	0	1.021922	0	1	-360	360;
	4303	6487	0.005208	0.208268	0	0	0	0	1.021922	0	1	-360	360;
	4303	395	0.005208	0.208268	0	0	0	0	1.021922	0	1	-360	360;
	4303	995	0.005208	0.208268	0	0	0	0	1.021922	0	1	-360	360;
	4303	9317	0.005208	0.208268	0	0	0	0	1.021922	0	1	-360	360;
	4303	3468	0.005208	0.208268	0	0	0	0	1.021922	0	1	-360	360;
	4303	9354	0.005208	0.208268	0	0	0	0	1.021922	0	1	-360	360;
	2116	3404	0.005208	0.208268	0	0	0	0	1.02243	0	1	-360	360;
	2116	8852	0.005208	0.208268	0	0	0	0	1.02243	0	1	-360	360;
	2116	2242	0.005208	0.208268	0	0	0	0	1.02243	0	1	-360	360;
	2116	8367	0.005208	0.208268	0	0	0	0	1.02243	0	1	-360	360;
	2116	1794	0.005208	0.208268	0	0	0	0	1.02243	0	1	-360	360;
	2116	7890	0.005208	0.208268	0	0	0	0	1.02243	0	1	-360	360;
	2116	1330	0.005208	0.208268	0	0	0	0	1.02243	0	1	-360	360;
	2116	11315	0.005208	0.208268	0	0	0	0	1.02243	0	1	-360	360;
	2116	4763	0.005208	0.208268	0	0	0	0	1.02243	0	1	-360	360;
	2116	10001	0.005208	0.208268	0	0	0	0	1.02243	0	1	-360	360;
	8695	3361	0.005208	0.208268	0	0	0	0	1.021706	0	1	-360	360;
	8695	9520	0.005208	0.208268	0	0	0	0	1.021706	0	1	-360	360;
	8695	7812	0.005208	0.208268	0	0	0	0	1.021706	0	1	-360	360;
	8695	10935	0.005208	0.208268	0	0	0	0	1.021706	0	1	-360	360;
	12322	4354	0.005208	0.208268	0	0	0	0	1.016195	0	1	-360	360;
	12322	10469	0.005208	0.208268	0	0	0	0	1.016195	0	1	-360	360;
	12322	3845	0.005208	0.208268	0	0	0	0	1.016195	0	1	-360	360;
	12322	8067	0.005208	0.208268	0	0	0	0	1.016195	0	1	-360	360;
	12322	9149	0.005208	0.208268	0	0	0	0	1.016195	0	1	-360	360;
	12322	13223	0.005208	0.208268	0	0	0	0	1.016195	0	1	-360	360;
	12322	6123	0.005208	0.208268	0	0	0	0	1.016195	0	1	-360	360;
	12322	54	0.005208	0.208268	0	0	0	0	1.016195	0	1	-360	360;
	12322	6602	0.005208	0.208268	0	0	0	0	1.016195	0	1	-360	360;
	12322	12216	0.005208	0.208268	0	0	0	0	1.016195	0	1	-360	360;
	12322	5187	0.005208	0.208268	0	0	0	0	1.016195	0	1	-360	360;
	1378	12728	0.005208	0.208268	0	0	0	0	1.019852	0	1	-360	360;
	1378	5672	0.005208	0.208268	0	0	0	0	1.019852	0	1	-360	360;
	1378	11221	0.005208	0.208268	0	0	0	0	1.019852	0	1	-360	360;
	1378	4133	0.005208	0.208268	0	0	0	0	1.019852	0	1	-360	360;
	1378	3683	0.005208	0.208268	0	0	0	0	1.019852	0	1	-360	360;
	1378	10788	0.005208	0.208268	0	0	0	0	1.019852	0	1	-360	360;
	1378	4191	0.005208	0.208268	0	0	0	0	1.019852	0	1	-360	360;
	1378	11264	0.005208	0.208268	0	0	0	0	1.019852	0	1	-360	360;
	1378	2698	0.005208	0.208268	0	0	0	0	1.019852	0	1	-360	360;
	1378	9808	0.005208	0.208268	0	0	0	0	1.019852	0	1	-360	360;
	1378	3176	0.005208	0.208268	0	0	0	0	1.019852	0	1	-360	360;
	2225	5347	0.002119	0.084719	0	0	0	0	1.01408	0	1	-360	360;
	2225	2080	0.002119	0.084719	0	0	0	0	1.01408	0	1	-360	360;
	2225	8637	0.002119	0.084719	0	0	0	0	1.01408	0	1	-360	360;
	2225	10633	0.002119	0.084719	0	0	0	0	1.01408	0	1	-360	360;
	2225	4004	0.002119	0.084719	0	0	0	0	1.01408	0	1	-360	360;
	2225	11128	0.002119	0.084719	0	0	0	0	1.01408	0	1	-360	360;
	2225	4573	0.002119	0.084719	0	0	0	0	1.01408	0	1	-360	360;
	2225	9662	0.002119	0.084719	0	0	0	0	1.01408	0	1	-360	360;
	2225	3029	0.002119	0.084719	0	0	0	0	1.01408	0	1	-360	360;
	2225	2681	0.002119	0.084719	0	0	0	0	1.01408	0	1	-360	360;
	2225	3535	0.002119	0.084719	0	0	0	0	1.01408	0	1	-360	360;
	1736	12448	0.005208	0.208268	0	0	0	0	1.019924	0	1	-360	360;
	1736	1701	0.005208	0.208268	0	0	0	0	1.019924	0	1	-360	360;
	1736	5188	0.005208	0.208268	0	0	0	0	1.019924	0	1	-360	360;
	1736	12217	0.005208	0.208268	0	0	0	0	1.019924	0	1	-360	360;
	1736	4675	0.005208	0.208268	0	0	0	0	1.019924	0	1	-360	360;
	1736	11729	0.005208	0.208268	0	0	0	0	1.019924	0	1	-360	360;
	1736	4135	0.005208	0.208268	0	0	0	0	1.019924	0	1	-360	360;
	2917	11230	0.113636	4.544034	0	0	0	0	1.018546	0	1	-360	360;
	2917	3650	0.113636	4.544034	0	0	0	0	1.013574	0	1	-360	360;
	8505	7142	0.002119	0.084719	0	0	0	0	1.020717	0	1	-360	360;
	8505	5335	0.002119	0.084719	0	0	0	0	1.020717	0	1	-360	360;
	8505	7289	0.002119	0.084719	0	0	0	0	1.020717	0	1	-360	360;
	8505	3644	0.002119	0.084719	0	0	0	0	1.020717	0	1	-360	360;
	8505	9179	0.002119	0.084719	0	0	0	0	1.020717	0	1	-360	360;
	8505	3139	0.002119	0.084719	0	0	0	0	1.020717	0	1	-360	360;
	8505	9775	0.002119	0.084719	0	0	0	0	1.020717	0	1	-360	360;
	8505	3050	0.002119	0.084719	0	0	0	0	1.020717	0	1	-360	360;
	322	3724	0.000212	0.008494	0	0	0	0	1.032123	0	1	-360	360;
	322	10330	0.000212	0.008494	0	0	0	0	1.032123	0	1	-360	360;
	10861	8573	0.005208	0.208268	0	0	0	0	1.010402	0	1	-360	360;
	10861	2429	0.005208	0.208268	0	0	0	0	1.010402	0	1	-360	360;
	10861	9061	0.005208	0.208268	0	0	0	0	1.010402	0	1	-360	360;
	10861	12528	0.005208	0.208268	0	0	0	0	1.010402	0	1	-360	360;
	10861	5963	0.005208	0.208268	0	0	0	0	1.010402	0	1	-360	360;
	10861	13041	0.005208	0.208268	0	0	0	0	1.010402	0	1	-360	360;
	10861	11467	0.005208	0.208268	0	0	0	0	1.010402	0	1	-360	360;
	10861	9873	0.005208	0.208268	0	0	0	0	1.010402	0	1	-360	360;
	4525	4264	0.005208	0.208268	0	0	0	0	1.016218	0	1	-360	360;
	4525	354	0.005208	0.208268	0	0	0	0	1.016218	0	1	-360	360;
	4525	7317	0.005208	0.208268	0	0	0	0	1.016218	0	1	-360	360;
	4525	10527	0.005208	0.208268	0	0	0	0	1.016218	0	1	-360	360;
	4525	3899	0.005208	0.208268	0	0	0	0	1.016218	0	1	-360	360;
	4525	1972	0.005208	0.208268	0	0	0	0	1.016218	0	1	-360	360;
	4525	11648	0.005208	0.208268	0	0	0	0	1.016218	0	1	-360	360;
	4525	1514	0.005208	0.208268	0	0	0	0	1.016218	0	1	-360	360;
	4525	8063	0.005208	0.208268	0	0	0	0	1.016218	0	1	-360	360;
	4525	2877	0.005208	0.208268	0	0	0	0	1.016218	0	1	-360	360;
	8693	9521	0.005208	0.208268	0	0	0	0	1.008217	0	1	-360	360;
	6185	2410	0.005208	0.208268	0	0	0	0	1.011955	0	1	-360	360;
	4831	9041	0.005208	0.208268	0	0	0	0	1.011309	0	1	-360	360;
	4831	121	0.005208	0.208268	0	0	0	0	1.011309	0	1	-360	360;
	4831	12911	0.005208	0.208268	0	0	0	0	1.011309	0	1	-360	360;
	4831	6210	0.005208	0.208268	0	0	0	0	1.011309	0	1	-360	360;
	4831	13303	0.005208	0.208268	0	0	0	0	1.011309	0	1	-360	360;
	4831	5741	0.005208	0.208268	0	0	0	0	1.011309	0	1	-360	360;
	4831	12822	0.005208	0.208268	0	0	0	0	1.011309	0	1	-360	360;
	4831	7168	0.005208	0.208268	0	0	0	0	1.011309	0	1	-360	360;
	4831	623	0.005208	0.208268	0	0	0	0	1.011309	0	1	-360	360;
	4831	10384	0.005208	0.208268	0	0	0	0	1.011309	0	1	-360	360;
	210	153	0.005208	0.208268	0	0	0	0	1.021101	0	1	-360	360;
	210	8092	0.005208	0.208268	0	0	0	0	1.021101	0	1	-360	360;
	210	1553	0.005208	0.208268	0	0	0	0	1.021101	0	1	-360	360;
	210	9297	0.005208	0.208268	0	0	0	0	1.021101	0	1	-360	360;
	11596	2210	0.005208	0.208268	0	0	0	0	1.021887	0	1	-360	360;
	11596	9809	0.005208	0.208268	0	0	0	0	1.021887	0	1	-360	360;
	11596	2700	0.005208	0.208268	0	0	0	0	1.021887	0	1	-360	360;
	11596	8326	0.005208	0.208268	0	0	0	0	1.021887	0	1	-360	360;
	11596	1288	0.005208	0.208268	0	0	0	0	1.021887	0	1	-360	360;
	11596	8797	0.005208	0.208268	0	0	0	0	1.021887	0	1	-360	360;
	11596	1757	0.005208	0.208268	0	0	0	0	1.021887	0	1	-360	360;
	11596	11265	0.005208	0.208268	0	0	0	0	1.021887	0	1	-360	360;
	3591	4193	0.005208	0.208268	0	0	0	0	1.012762	0	1	-360	360;
	3591	259	0.005208	0.208268	0	0	0	0	1.012762	0	1	-360	360;
	11417	6820	0.005208	0.208268	0	0	0	0	1.017812	0	1	-360	360;
	11417	8851	0.005208	0.208268	0	0	0	0	1.017812	0	1	-360	360;
	11417	6341	0.005208	0.208268	0	0	0	0	1.017812	0	1	-360	360;
	11417	1202	0.005208	0.208268	0	0	0	0	1.017812	0	1	-360	360;
	11417	7740	0.005208	0.208268	0	0	0	0	1.017812	0	1	-360	360;
	11417	724	0.005208	0.208268	0	0	0	0	1.017812	0	1	-360	360;
	7621	7242	0.005208	0.208268	0	0	0	0	1.017708	0	1	-360	360;
	7621	13412	0.005208	0.208268	0	0	0	0	1.017708	0	1	-360	360;
	7621	8686	0.005208	0.208268	0	0	0	0	1.017708	0	1	-360	360;
	7621	10696	0.005208	0.208268	0	0	0	0	1.017708	0	1	-360	360;
	7621	4074	0.005208	0.208268	0	0	0	0	1.017708	0	1	-360	360;
	7621	11155	0.005208	0.208268	0	0	0	0	1.017708	0	1	-360	360;
	7621	4594	0.005208	0.208268	0	0	0	0	1.017708	0	1	-360	360;
	7621	11654	0.005208	0.208268	0	0	0	0	1.017708	0	1	-360	360;
	566	5109	0.005208	0.208268	0	0	0	0	1.021154	0	1	-360	360;
	566	12195	0.005208	0.208268	0	0	0	0	1.021154	0	1	-360	360;
	5209	5641	0.005208	0.208268	0	0	0	0	1.014692	0	1	-360	360;
	5209	12717	0.005208	0.208268	0	0	0	0	1.014692	0	1	-360	360;
	5209	6107	0.005208	0.208268	0	0	0	0	1.014692	0	1	-360	360;
	5209	12227	0.005208	0.208268	0	0	0	0	1.014692	0	1	-360	360;
	5209	13174	0.005208	0.208268	0	0	0	0	1.014692	0	1	-360	360;
	5209	7028	0.005208	0.208268	0	0	0	0	1.014692	0	1	-360	360;
	5209	18	0.005208	0.208268	0	0	0	0	1.014692	0	1	-360	360;
	5209	4384	0.005208	0.208268	0	0	0	0	1.014692	0	1	-360	360;
	5209	477	0.005208	0.208268	0	0	0	0	1.014692	0	1	-360	360;
	11908	7939	0.005208	0.208268	0	0	0	0	1.016774	0	1	-360	360;
	11908	938	0.005208	0.208268	0	0	0	0	1.016774	0	1	-360	360;
	11908	4563	0.005208	0.208268	0	0	0	0	1.016774	0	1	-360	360;
	11908	11115	0.005208	0.208268	0	0	0	0	1.016774	0	1	-360	360;
	11908	3323	0.005208	0.208268	0	0	0	0	1.016774	0	1	-360	360;
	4468	12819	0.005208	0.208268	0	0	0	0	1.015232	0	1	-360	360;
	4468	2796	0.005208	0.208268	0	0	0	0	1.015232	0	1	-360	360;
	4468	9901	0.005208	0.208268	0	0	0	0	1.015232	0	1	-360	360;
	4468	2332	0.005208	0.208268	0	0	0	0	1.015232	0	1	-360	360;
	4468	9427	0.005208	0.208268	0	0	0	0	1.015232	0	1	-360	360;
	4468	1889	0.005208	0.208268	0	0	0	0	1.015232	0	1	-360	360;
	8450	29	0.005208	0.208268	0	0	0	0	1.009474	0	1	-360	360;
	13051	9123	0.002119	0.084719	0	0	0	0	1.034311	0	1	-360	360;
	13051	2066	0.002119	0.084719	0	0	0	0	1.034311	0	1	-360	360;
	9101	4052	0.004836	0.193363	0	0	0	0	1.006755	0	1	-360	360;
	9101	2583	0.004836	0.193363	0	0	0	0	1.001944	0	1	-360	360;
	9101	9192	0.004836	0.193363	0	0	0	0	1.001944	0	1	-360	360;
	10976	8687	0.002119	0.084719	0	0	0	0	0.993441	0	1	-360	360;
	10976	721	0.002119	0.084719	0	0	0	0	0.993441	0	1	-360	360;
	7247	1685	0.005208	0.208268	0	0	0	0	1.024172	0	1	-360	360;
	7247	8253	0.005208	0.208268	0	0	0	0	1.024172	0	1	-360	360;
	7247	1208	0.005208	0.208268	0	0	0	0	1.024172	0	1	-360	360;
	8366	12390	0.000531	0.021225	0	0	0	0	1.036935	0	1	-360	360;
	8366	5354	0.000531	0.021225	0	0	0	0	1.036935	0	1	-360	360;
	11513	5155	0.000708	0.02832	0	0	0	0	1.04316	0	1	-360	360;
	6415	11704	0.000708	0.02832	0	0	0	0	1.04464	0	1	-360	360;
	3637	5356	0.002119	0.084719	0	0	0	0	1.03628	0	1	-360	360;
	3637	3154	0.002119	0.084719	0	0	0	0	1.03628	0	1	-360	360;
	4361	5653	0.005208	0.208268	0	0	0	0	1.013203	0	1	-360	360;
	4361	3508	0.005208	0.208268	0	0	0	0	1.013203	0	1	-360	360;
	1683	6296	0.002119	0.084719	0	0	0	0	1.015199	0	1	-360	360;
	1683	12900	0.002119	0.084719	0	0	0	0	1.015199	0	1	-360	360;
	12349	2789	0.002119	0.084719	0	0	0	0	1.009563	0	1	-360	360;
	12349	9422	0.002119	0.084719	0	0	0	0	1.009563	0	1	-360	360;
	12349	8976	0.002119	0.084719	0	0	0	0	1.009563	0	1	-360	360;
	2563	8961	0.005208	0.208268	0	0	0	0	1.007875	0	1	-360	360;
	2563	2803	0.005208	0.208268	0	0	0	0	1.007875	0	1	-360	360;
	3144	6440	0.005208	0.208268	0	0	0	0	1.009522	0	1	-360	360;
	3144	8726	0.005208	0.208268	0	0	0	0	1.009522	0	1	-360	360;
	658	7951	0.005208	0.208268	0	0	0	0	1.01597	0	1	-360	360;
	1403	1896	0.005208	0.208268	0	0	0	0	1.006727	0	1	-360	360;
	425	12537	0.000212	0.008494	0	0	0	0	1.039709	0	1	-360	360;
	425	9148	0.000212	0.008494	0	0	0	0	1.03971	0	1	-360	360;
	12672	7031	0.005208	0.208268	0	0	0	0	1.005675	0	1	-360	360;
	12672	9288	0.005208	0.208268	0	0	0	0	1.005675	0	1	-360	360;
	10498	6346	0.005208	0.208268	0	0	0	0	1.007155	0	1	-360	360;
	10498	12400	0.005208	0.208268	0	0	0	0	1.007155	0	1	-360	360;
	10498	5842	0.005208	0.208268	0	0	0	0	1.007155	0	1	-360	360;
	11164	266	0.005208	0.208268	0	0	0	0	1.009142	0	1	-360	360;
	11164	7249	0.005208	0.208268	0	0	0	0	1.009142	0	1	-360	360;
	11164	13456	0.005208	0.208268	0	0	0	0	1.009142	0	1	-360	360;
	1926	5592	0.002119	0.084719	0	0	0	0	1.000368	0	1	-360	360;
	1926	496	0.002119	0.084719	0	0	0	0	1.000368	0	1	-360	360;
	3612	4324	0.005208	0.208268	0	0	0	0	1.006795	0	1	-360	360;
	9112	11065	0.005208	0.208268	0	0	0	0	1.016738	0	1	-360	360;
	9112	3990	0.005208	0.208268	0	0	0	0	1.016738	0	1	-360	360;
	9112	10612	0.005208	0.208268	0	0	0	0	1.016738	0	1	-360	360;
	6134	3517	0.005208	0.208268	0	0	0	0	1.015029	0	1	-360	360;
	6134	12091	0.005208	0.208268	0	0	0	0	1.015029	0	1	-360	360;
	6134	5029	0.005208	0.208268	0	0	0	0	1.015029	0	1	-360	360;
	10024	11577	0.005208	0.208268	0	0	0	0	1.009875	0	1	-360	360;
	10024	4513	0.005208	0.208268	0	0	0	0	1.009875	0	1	-360	360;
	1034	13086	0.005208	0.208268	0	0	0	0	1.012243	0	1	-360	360;
	1034	6016	0.005208	0.208268	0	0	0	0	1.012243	0	1	-360	360;
	1034	7933	0.005208	0.208268	0	0	0	0	1.012243	0	1	-360	360;
	2874	1382	0.005208	0.208268	0	0	0	0	1.00515	0	1	-360	360;
	2874	8415	0.005208	0.208268	0	0	0	0	1.00515	0	1	-360	360;
	11508	1834	0.013661	0.546277	0	0	0	0	1.002341	0	1	-360	360;
	12413	8900	0.005208	0.208268	0	0	0	0	1.03175	0	1	-360	360;
	12413	9321	0.005208	0.208268	0	0	0	0	1.03175	0	1	-360	360;
	9876	9389	0.020833	0.833073	0	0	0	0	1.004194	0	1	-360	360;
	2010	2756	0.005208	0.208268	0	0	0	0	1.004507	0	1	-360	360;
	2010	9860	0.005208	0.208268	0	0	0	0	1.004507	0	1	-360	360;
	9161	3234	0.005208	0.208268	0	0	0	0	1.007962	0	1	-360	360;
	9161	9169	0.005208	0.208268	0	0	0	0	1.007962	0	1	-360	360;
	8337	5593	0.005208	0.208268	0	0	0	0	1.008019	0	1	-360	360;
	8337	12681	0.005208	0.208268	0	0	0	0	1.008019	0	1	-360	360;
	11433	6076	0.005208	0.208268	0	0	0	0	1.009669	0	1	-360	360;
	11433	11145	0.005208	0.208268	0	0	0	0	1.009669	0	1	-360	360;
	11433	4590	0.005208	0.208268	0	0	0	0	1.009669	0	1	-360	360;
	12371	11652	0.005208	0.208268	0	0	0	0	1.00586	0	1	-360	360;
	12371	5099	0.005208	0.208268	0	0	0	0	1.00586	0	1	-360	360;
	2511	1031	0.002119	0.084719	0	0	0	0	1.023477	0	1	-360	360;
	2511	7561	0.002119	0.084719	0	0	0	0	1.023477	0	1	-360	360;
	8917	10218	0.005208	0.208268	0	0	0	0	1.007312	0	1	-360	360;
	8917	8205	0.005208	0.208268	0	0	0	0	1.007312	0	1	-360	360;
	8917	9081	0.005208	0.208268	0	0	0	0	1.007312	0	1	-360	360;
	7984	11285	0.002119	0.084719	0	0	0	0	1.020825	0	1	-360	360;
	7984	4752	0.002119	0.084719	0	0	0	0	1.020825	0	1	-360	360;
	7984	7754	0.002119	0.084719	0	0	0	0	1.020825	0	1	-360	360;
	12860	2627	0.000531	0.021225	0	0	0	0	1.054708	0	1	-360	360;
	12860	9265	0.000531	0.021225	0	0	0	0	1.054708	0	1	-360	360;
	6424	3543	0.005208	0.208268	0	0	0	0	1.025153	0	1	-360	360;
	6424	10158	0.005208	0.208268	0	0	0	0	1.025153	0	1	-360	360;
	7584	8109	0.005208	0.208268	0	0	0	0	1.01477	0	1	-360	360;
	7584	11880	0.005208	0.208268	0	0	0	0	1.01477	0	1	-360	360;
	10664	12563	0.000531	0.021225	0	0	0	0	1.037102	0	1	-360	360;
	5020	2670	0.002119	0.084719	0	0	0	0	1.030227	0	1	-360	360;
	5020	9301	0.002119	0.084719	0	0	0	0	1.030227	0	1	-360	360;
	9270	12902	0.006944	0.277691	0	0	0	0	1.012092	0	1	-360	360;
	9270	5357	0.006944	0.277691	0	0	0	0	1.004875	0	1	-360	360;
	12081	12387	0.005208	0.208268	0	0	0	0	1.015132	0	1	-360	360;
	12081	2790	0.005208	0.208268	0	0	0	0	1.015132	0	1	-360	360;
	9239	5166	0.002119	0.084719	0	0	0	0	1.019313	0	1	-360	360;
	9239	7146	0.002119	0.084719	0	0	0	0	1.019313	0	1	-360	360;
	7040	937	0.005208	0.208268	0	0	0	0	1.009191	0	1	-360	360;
	7040	8409	0.005208	0.208268	0	0	0	0	1.009191	0	1	-360	360;
	420	1376	0.005208	0.208268	0	0	0	0	1.00599	0	1	-360	360;
	420	6990	0.005208	0.208268	0	0	0	0	1.00599	0	1	-360	360;
	4827	13639	0.006831	0.273139	0	0	0	0	1.01509	0	1	-360	360;
	4827	7167	0.006831	0.273139	0	0	0	0	1.009134	0	1	-360	360;
	5659	4718	0.000531	0.021225	0	0	0	0	1.033693	0	1	-360	360;
	9703	2624	0.000531	0.021225	0	0	0	0	1.03365	0	1	-360	360;
	8673	12636	0.005208	0.208268	0	0	0	0	1.011827	0	1	-360	360;
	8673	7236	0.005208	0.208268	0	0	0	0	1.011827	0	1	-360	360;
	11431	715	0.005208	0.208268	0	0	0	0	1.011851	0	1	-360	360;
	11431	7700	0.005208	0.208268	0	0	0	0	1.011851	0	1	-360	360;
	13461	11030	0.002119	0.084719	0	0	0	0	1.043444	0	1	-360	360;
	8684	8206	0.005208	0.208268	0	0	0	0	1.007837	0	1	-360	360;
	8684	1641	0.005208	0.208268	0	0	0	0	1.007837	0	1	-360	360;
	2862	11638	0.005208	0.208268	0	0	0	0	1.007064	0	1	-360	360;
	2862	2102	0.005208	0.208268	0	0	0	0	1.007064	0	1	-360	360;
	3092	5416	0.005208	0.208268	0	0	0	0	1.030223	0	1	-360	360;
	3092	12451	0.005208	0.208268	0	0	0	0	1.030223	0	1	-360	360;
	3092	10386	0.005208	0.208268	0	0	0	0	1.030223	0	1	-360	360;
	8985	3270	0.020833	0.833073	0	0	0	0	1.003211	0	1	-360	360;
	6579	9896	0.005208	0.208268	0	0	0	0	1.013992	0	1	-360	360;
	9454	2785	0.005208	0.208268	0	0	0	0	0.999066	0	1	-360	360;
	9454	2767	0.005208	0.208268	0	0	0	0	0.999066	0	1	-360	360;
	9454	4282	0.005208	0.208268	0	0	0	0	0.999066	0	1	-360	360;
	6904	10881	0.005208	0.208268	0	0	0	0	0.999253	0	1	-360	360;
	6904	3771	0.005208	0.208268	0	0	0	0	0.999253	0	1	-360	360;
	6904	10417	0.005208	0.208268	0	0	0	0	0.999253	0	1	-360	360;
	7629	754	0.002119	0.084719	0	0	0	0	1.020652	0	1	-360	360;
	7629	5456	0.002119	0.084719	0	0	0	0	1.020652	0	1	-360	360;
	2954	6992	0.005208	0.208268	0	0	0	0	1.030856	0	1	-360	360;
	2954	13147	0.005208	0.208268	0	0	0	0	1.030856	0	1	-360	360;
	2216	4367	0.000531	0.021225	0	0	0	0	1.043072	0	1	-360	360;
	4769	13322	0.000531	0.021225	0	0	0	0	1.037866	0	1	-360	360;
	4769	4670	0.000531	0.021225	0	0	0	0	1.037866	0	1	-360	360;
	4769	10715	0.000531	0.021225	0	0	0	0	1.037866	0	1	-360	360;
	2450	11076	0.005208	0.208268	0	0	0	0	1.010814	0	1	-360	360;
	6856	12208	0.000531	0.021225	0	0	0	0	1.041055	0	1	-360	360;
	6856	5674	0.000531	0.021225	0	0	0	0	1.041055	0	1	-360	360;
	1380	13467	0.013736	0.549279	0	0	0	0	1.016236	0	1	-360	360;
	6945	1906	0.005208	0.208268	0	0	0	0	1.009007	0	1	-360	360;
	6945	4366	0.005208	0.208268	0	0	0	0	1.009007	0	1	-360	360;
	5744	2351	0.020661	0.826188	0	0	0	0	1.001301	0	1	-360	360;
	9066	8958	0.000212	0.008494	0	0	0	0	1.026768	0	1	-360	360;
	12432	3607	0.000708	0.02832	0	0	0	0	1.023314	0	1	-360	360;
	12432	9738	0.000708	0.02832	0	0	0	0	1.023314	0	1	-360	360;
	12584	9084	0.002119	0.084719	0	0	0	0	1.012083	0	1	-360	360;
	12584	5444	0.002119	0.084719	0	0	0	0	1.012083	0	1	-360	360;
	10323	12494	0.002119	0.084719	0	0	0	0	1.010488	0	1	-360	360;
	10323	9035	0.002119	0.084719	0	0	0	0	1.010488	0	1	-360	360;
	10323	1969	0.002119	0.084719	0	0	0	0	1.010488	0	1	-360	360;
	1986	2592	0.000708	0.02832	0	0	0	0	1.010656	0	1	-360	360;
	11735	7388	0.000212	0.008494	0	0	0	0	1.029522	0	1	-360	360;
	11735	5797	0.000531	0.021225	0	0	0	0	1.030088	0	1	-360	360;
	11735	12874	0.000531	0.021225	0	0	0	0	1.030088	0	1	-360	360;
	11735	6903	0.000531	0.021225	0	0	0	0	1.030104	0	1	-360	360;
	11735	13530	0.000531	0.021225	0	0	0	0	1.030104	0	1	-360	360;
	11735	3280	0.000708	0.02832	0	0	0	0	1.032185	0	1	-360	360;
	11735	11619	0.000708	0.02832	0	0	0	0	1.032185	0	1	-360	360;
	4523	66	0.000708	0.02832	0	0	0	0	1.016289	0	1	-360	360;
	4523	7073	0.000708	0.02832	0	0	0	0	1.016289	0	1	-360	360;
	13107	533	0.000708	0.02832	0	0	0	0	1.004105	0	1	-360	360;
	6922	7511	0.000708	0.02832	0	0	0	0	1.004095	0	1	-360	360;
	10124	7551	0.002133	0.085298	0	0	0	0	1.012741	0	1	-360	360;
	10124	567	0.002133	0.085298	0	0	0	0	1.006257	0	1	-360	360;
	7441	5587	0.000424	0.016973	0	0	0	0	1.019586	0	1	-360	360;
	7678	12665	0.000424	0.016973	0	0	0	0	1.020616	0	1	-360	360;
	10320	6058	0.000424	0.016973	0	0	0	0	1.02349	0	1	-360	360;
	388	13158	0.000424	0.016973	0	0	0	0	1.023498	0	1	-360	360;
	2913	2964	0.005747	0.229813	0	0	0	0	1.020977	0	1	-360	360;
	2913	10100	0.005747	0.229813	0	0	0	0	1.020977	0	1	-360	360;
	2913	3474	0.005747	0.229813	0	0	0	0	1.015977	0	1	-360	360;
	2913	10581	0.005747	0.229813	0	0	0	0	1.015977	0	1	-360	360;
	2913	2056	0.005747	0.229813	0	0	0	0	1.015977	0	1	-360	360;
	2913	9109	0.005747	0.229813	0	0	0	0	1.015977	0	1	-360	360;
	2913	2496	0.005747	0.229813	0	0	0	0	1.015977	0	1	-360	360;
	2913	9604	0.005747	0.229813	0	0	0	0	1.015977	0	1	-360	360;
	1898	9597	0.000531	0.021225	0	0	0	0	1.016787	0	1	-360	360;
	11378	8191	0.005208	0.208268	0	0	0	0	1.017272	0	1	-360	360;
	11378	10249	0.005208	0.208268	0	0	0	0	1.017272	0	1	-360	360;
	11378	6465	0.005208	0.208268	0	0	0	0	1.017272	0	1	-360	360;
	11378	5764	0.005208	0.208268	0	0	0	0	1.017272	0	1	-360	360;
	11378	12321	0.005208	0.208268	0	0	0	0	1.017272	0	1	-360	360;
	11378	5292	0.005208	0.208268	0	0	0	0	1.017272	0	1	-360	360;
	11378	11822	0.005208	0.208268	0	0	0	0	1.017272	0	1	-360	360;
	11378	4777	0.005208	0.208268	0	0	0	0	1.017272	0	1	-360	360;
	11378	11328	0.005208	0.208268	0	0	0	0	1.017272	0	1	-360	360;
	11378	4252	0.005208	0.208268	0	0	0	0	1.017272	0	1	-360	360;
	11378	10845	0.005208	0.208268	0	0	0	0	1.017272	0	1	-360	360;
	12707	5637	0.000708	0.02832	0	0	0	0	1.018396	0	1	-360	360;
	12707	5718	0.000708	0.02832	0	0	0	0	1.018396	0	1	-360	360;
	1282	3823	0.002119	0.084719	0	0	0	0	1.009963	0	1	-360	360;
	1282	10918	0.002119	0.084719	0	0	0	0	1.009963	0	1	-360	360;
	1282	2368	0.002119	0.084719	0	0	0	0	1.009963	0	1	-360	360;
	1282	470	0.002119	0.084719	0	0	0	0	1.009963	0	1	-360	360;
	1282	4203	0.002119	0.084719	0	0	0	0	1.009963	0	1	-360	360;
	2985	6651	0.002119	0.084719	0	0	0	0	1.014438	0	1	-360	360;
	2985	10738	0.002119	0.084719	0	0	0	0	1.014438	0	1	-360	360;
	2985	4120	0.002119	0.084719	0	0	0	0	1.014438	0	1	-360	360;
	2985	2172	0.002119	0.084719	0	0	0	0	1.014438	0	1	-360	360;
	2985	8753	0.002119	0.084719	0	0	0	0	1.014438	0	1	-360	360;
	2985	9627	0.000212	0.008494	0	0	0	0	1.014172	0	1	-360	360;
	1282	8770	0.000708	0.02832	0	0	0	0	1.011503	0	1	-360	360;
	1282	819	0.000708	0.02832	0	0	0	0	1.011503	0	1	-360	360;
	1282	3553	0.000708	0.02832	0	0	0	0	1.011503	0	1	-360	360;
	1282	1262	0.000708	0.02832	0	0	0	0	1.011503	0	1	-360	360;
	6724	11677	0.002119	0.084719	0	0	0	0	1.020245	0	1	-360	360;
	6724	10390	0.002119	0.084719	0	0	0	0	1.020245	0	1	-360	360;
	5402	4509	0.000212	0.008494	0	0	0	0	1.023751	0	1	-360	360;
	5402	6085	0.000212	0.008494	0	0	0	0	1.023751	0	1	-360	360;
	10981	3330	0.01	0.399875	0	0	0	0	1.023355	0	1	-360	360;
	3238	4309	0.002119	0.084719	0	0	0	0	1.02251	0	1	-360	360;
	3238	585	0.002119	0.084719	0	0	0	0	1.02251	0	1	-360	360;
	3238	3792	0.002119	0.084719	0	0	0	0	1.02251	0	1	-360	360;
	3238	10901	0.002119	0.084719	0	0	0	0	1.02251	0	1	-360	360;
	3238	3296	0.002119	0.084719	0	0	0	0	1.02251	0	1	-360	360;
	9831	10406	0.006068	0.242643	0	0	0	0	1.013469	0	1	-360	360;
	13656	9206	0.000708	0.02832	0	0	0	0	1.027689	0	1	-360	360;
	13656	2625	0.000708	0.02832	0	0	0	0	1.027689	0	1	-360	360;
	11702	693	0.000708	0.02832	0	0	0	0	1.027677	0	1	-360	360;
	11702	7233	0.000708	0.02832	0	0	0	0	1.027677	0	1	-360	360;
	9098	11403	0.000531	0.021225	0	0	0	0	1.028413	0	1	-360	360;
	9098	2053	0.000531	0.021225	0	0	0	0	1.028413	0	1	-360	360;
	9098	7674	0.000531	0.021225	0	0	0	0	1.028413	0	1	-360	360;
	61	13192	0.000212	0.008494	0	0	0	0	1.035966	0	1	-360	360;
	10744	1559	0.000531	0.021225	0	0	0	0	1.080738	0	1	-360	360;
	10278	2197	0.000531	0.021225	0	0	0	0	1.079849	0	1	-360	360;
	2474	3922	0.002119	0.084719	0	0	0	0	1.017692	0	1	-360	360;
	2474	11010	0.002119	0.084719	0	0	0	0	1.017692	0	1	-360	360;
	1286	2955	0.005208	0.208268	0	0	0	0	1.017539	0	1	-360	360;
	7368	10092	0.005208	0.208268	0	0	0	0	1.016598	0	1	-360	360;
	7368	3471	0.005208	0.208268	0	0	0	0	1.016598	0	1	-360	360;
	9490	10575	0.005208	0.208268	0	0	0	0	1.019432	0	1	-360	360;
	9490	3947	0.005208	0.208268	0	0	0	0	1.019432	0	1	-360	360;
	9490	11039	0.005208	0.208268	0	0	0	0	1.019432	0	1	-360	360;
	5329	4483	0.005208	0.208268	0	0	0	0	1.016081	0	1	-360	360;
	5329	11551	0.005208	0.208268	0	0	0	0	1.016081	0	1	-360	360;
	11941	4985	0.005208	0.208268	0	0	0	0	1.022769	0	1	-360	360;
	11941	12049	0.005208	0.208268	0	0	0	0	1.022769	0	1	-360	360;
	12482	9408	0.002119	0.084719	0	0	0	0	1.030673	0	1	-360	360;
	12482	7499	0.002119	0.084719	0	0	0	0	1.030673	0	1	-360	360;
	5577	5761	0.005208	0.208268	0	0	0	0	1.004987	0	1	-360	360;
	5577	12310	0.005208	0.208268	0	0	0	0	1.004987	0	1	-360	360;
	5577	7164	0.005208	0.208268	0	0	0	0	1.004987	0	1	-360	360;
	5577	182	0.005208	0.208268	0	0	0	0	1.004987	0	1	-360	360;
	1506	10165	0.000531	0.021225	0	0	0	0	1.028483	0	1	-360	360;
	5015	5036	0.005208	0.208268	0	0	0	0	1.017975	0	1	-360	360;
	5015	13658	0.005208	0.208268	0	0	0	0	1.017975	0	1	-360	360;
	6113	3165	0.000531	0.021225	0	0	0	0	1.019664	0	1	-360	360;
	9643	5517	0.000531	0.021225	0	0	0	0	1.016681	0	1	-360	360;
	10689	6847	0.005208	0.208268	0	0	0	0	1.01291	0	1	-360	360;
	10689	12973	0.005208	0.208268	0	0	0	0	1.01291	0	1	-360	360;
	3601	10987	0.005208	0.208268	0	0	0	0	1.011987	0	1	-360	360;
	3601	753	0.005208	0.208268	0	0	0	0	1.011987	0	1	-360	360;
	5572	7734	0.005208	0.208268	0	0	0	0	1.010487	0	1	-360	360;
	5572	289	0.005208	0.208268	0	0	0	0	1.010493	0	1	-360	360;
	6653	7071	0.000531	0.021225	0	0	0	0	1.025474	0	1	-360	360;
	6653	13235	0.000531	0.021225	0	0	0	0	1.025474	0	1	-360	360;
	6653	6620	0.000531	0.021225	0	0	0	0	1.025474	0	1	-360	360;
	6653	10505	0.000531	0.021225	0	0	0	0	1.025474	0	1	-360	360;
	6653	6729	0.000531	0.021225	0	0	0	0	1.025474	0	1	-360	360;
	6653	11786	0.000531	0.021225	0	0	0	0	1.025474	0	1	-360	360;
	537	9852	0.000531	0.021225	0	0	0	0	1.019566	0	1	-360	360;
	530	3224	0.000531	0.021225	0	0	0	0	1.026211	0	1	-360	360;
	13654	4400	0.000531	0.021225	0	0	0	0	1.035648	0	1	-360	360;
	13654	11982	0.000531	0.021225	0	0	0	0	1.035648	0	1	-360	360;
	13654	4928	0.000531	0.021225	0	0	0	0	1.035648	0	1	-360	360;
	488	476	0.000531	0.021225	0	0	0	0	0.990884	0	1	-360	360;
	4669	4371	0.000531	0.021225	0	0	0	0	1.022655	0	1	-360	360;
	6887	12381	0.005208	0.208268	0	0	0	0	1.008553	0	1	-360	360;
	6887	5350	0.005208	0.208268	0	0	0	0	1.008553	0	1	-360	360;
	5153	12914	0.005208	0.208268	0	0	0	0	1.008542	0	1	-360	360;
	5153	5825	0.005208	0.208268	0	0	0	0	1.008542	0	1	-360	360;
	12209	1394	0.000531	0.021225	0	0	0	0	1.014772	0	1	-360	360;
	13263	6667	0.000531	0.021225	0	0	0	0	1.014886	0	1	-360	360;
	10020	229	0.005208	0.208268	0	0	0	0	1.013241	0	1	-360	360;
	10020	6769	0.005208	0.208268	0	0	0	0	1.013241	0	1	-360	360;
	7817	10388	0.005208	0.208268	0	0	0	0	1.01327	0	1	-360	360;
	7817	6842	0.005208	0.208268	0	0	0	0	1.01327	0	1	-360	360;
	4463	2301	0.000531	0.021225	0	0	0	0	1.016044	0	1	-360	360;
	4104	2561	0.005208	0.208268	0	0	0	0	1.009961	0	1	-360	360;
	4104	8651	0.005208	0.208268	0	0	0	0	1.009961	0	1	-360	360;
	7423	11504	0.003521	0.140801	0	0	0	0	1.010373	0	1	-360	360;
	8031	13012	0.000531	0.021225	0	0	0	0	1.0253	0	1	-360	360;
	8031	1636	0.005208	0.208268	0	0	0	0	1.022618	0	1	-360	360;
	8031	7712	0.005208	0.208268	0	0	0	0	1.022618	0	1	-360	360;
	2109	7005	0.000531	0.021225	0	0	0	0	1.016125	0	1	-360	360;
	2109	492	0.000531	0.021225	0	0	0	0	1.014629	0	1	-360	360;
	2109	7045	0.000531	0.021225	0	0	0	0	1.014629	0	1	-360	360;
	2121	5534	0.000531	0.021225	0	0	0	0	1.019116	0	1	-360	360;
	2121	12606	0.000531	0.021225	0	0	0	0	1.019116	0	1	-360	360;
	7741	5034	0.000531	0.021225	0	0	0	0	1.019552	0	1	-360	360;
	4071	13620	0.000531	0.021225	0	0	0	0	1.018741	0	1	-360	360;
	4071	6500	0.000531	0.021225	0	0	0	0	1.018741	0	1	-360	360;
	4071	13123	0.000531	0.021225	0	0	0	0	1.018741	0	1	-360	360;
	7101	4013	0.000531	0.021225	0	0	0	0	0.98444	0	1	-360	360;
	7101	11096	0.000531	0.021225	0	0	0	0	0.98444	0	1	-360	360;
	1585	3551	0.000531	0.021225	0	0	0	0	1.026268	0	1	-360	360;
	1585	10647	0.000531	0.021225	0	0	0	0	1.026268	0	1	-360	360;
	374	2819	0.005208	0.208268	0	0	0	0	1.010526	0	1	-360	360;
	374	5575	0.005208	0.208268	0	0	0	0	1.010526	0	1	-360	360;
	11767	5023	0.000531	0.021225	0	0	0	0	1.022689	0	1	-360	360;
	11767	11586	0.000531	0.021225	0	0	0	0	1.022689	0	1	-360	360;
	10265	9815	0.005208	0.208268	0	0	0	0	0.998115	0	1	-360	360;
	843	2642	0.003521	0.140801	0	0	0	0	1.004532	0	1	-360	360;
	11297	4344	0.005208	0.208268	0	0	0	0	1.003947	0	1	-360	360;
	11297	778	0.005208	0.208268	0	0	0	0	1.003947	0	1	-360	360;
	8020	6476	0.000531	0.021225	0	0	0	0	1.01097	0	1	-360	360;
	2452	13079	0.000531	0.021225	0	0	0	0	1.018906	0	1	-360	360;
	2452	7654	0.000531	0.021225	0	0	0	0	1.018728	0	1	-360	360;
	4736	12016	0.000531	0.021225	0	0	0	0	1.018853	0	1	-360	360;
	4736	13333	0.000531	0.021225	0	0	0	0	1.018853	0	1	-360	360;
	10338	1561	0.000531	0.021225	0	0	0	0	1.013748	0	1	-360	360;
	7483	2559	0.005208	0.208268	0	0	0	0	1.007021	0	1	-360	360;
	7483	11588	0.005208	0.208268	0	0	0	0	1.007021	0	1	-360	360;
	4373	3031	0.005208	0.208268	0	0	0	0	1.013739	0	1	-360	360;
	11541	648	0.000531	0.021225	0	0	0	0	1.000552	0	1	-360	360;
	3208	328	0.003521	0.140801	0	0	0	0	0.99659	0	1	-360	360;
	2802	6886	0.003521	0.140801	0	0	0	0	0.996518	0	1	-360	360;
	2754	8772	0.003521	0.140801	0	0	0	0	0.996625	0	1	-360	360;
	1405	7474	0.005208	0.208268	0	0	0	0	0.997192	0	1	-360	360;
	11590	248	0.005208	0.208268	0	0	0	0	1.013833	0	1	-360	360;
	13563	1629	0.005208	0.208268	0	0	0	0	1.013822	0	1	-360	360;
	6805	1419	0.005208	0.208268	0	0	0	0	1.021964	0	1	-360	360;
	6805	8447	0.005208	0.208268	0	0	0	0	1.021964	0	1	-360	360;
	10360	956	0.005208	0.208268	0	0	0	0	1.020059	0	1	-360	360;
	10360	7342	0.005208	0.208268	0	0	0	0	1.020059	0	1	-360	360;
	12963	12794	0.000531	0.021225	0	0	0	0	1.014455	0	1	-360	360;
	1107	5714	0.000531	0.021225	0	0	0	0	1.014539	0	1	-360	360;
	317	358	0.000531	0.021225	0	0	0	0	1.020862	0	1	-360	360;
	4850	7334	0.000531	0.021225	0	0	0	0	1.025181	0	1	-360	360;
	4850	13560	0.000531	0.021225	0	0	0	0	1.025181	0	1	-360	360;
	4850	10567	0.000531	0.021225	0	0	0	0	1.025243	0	1	-360	360;
	12364	7929	0.000531	0.021225	0	0	0	0	1.012377	0	1	-360	360;
	12364	10102	0.000531	0.021225	0	0	0	0	1.012377	0	1	-360	360;
	10131	828	0.000531	0.021225	0	0	0	0	1.012069	0	1	-360	360;
	1141	7808	0.000531	0.021225	0	0	0	0	1.02645	0	1	-360	360;
	1141	12065	0.000531	0.021225	0	0	0	0	1.02645	0	1	-360	360;
	1141	5503	0.000531	0.021225	0	0	0	0	1.025452	0	1	-360	360;
	13004	11578	0.005208	0.208268	0	0	0	0	0.990651	0	1	-360	360;
	399	8914	0.005208	0.208268	0	0	0	0	1.006118	0	1	-360	360;
	399	2292	0.005208	0.208268	0	0	0	0	1.006118	0	1	-360	360;
	9019	9387	0.005208	0.208268	0	0	0	0	1.006561	0	1	-360	360;
	9019	932	0.005208	0.208268	0	0	0	0	1.006561	0	1	-360	360;
	5874	7931	0.005208	0.208268	0	0	0	0	1.016527	0	1	-360	360;
	7534	2894	0.000531	0.021225	0	0	0	0	1.020891	0	1	-360	360;
	7534	9527	0.000531	0.021225	0	0	0	0	1.020891	0	1	-360	360;
	1998	1552	0.000531	0.021225	0	0	0	0	1.022077	0	1	-360	360;
	1998	8097	0.000531	0.021225	0	0	0	0	1.022077	0	1	-360	360;
	4318	9875	0.000531	0.021225	0	0	0	0	1.025043	0	1	-360	360;
	4318	2760	0.000531	0.021225	0	0	0	0	1.025043	0	1	-360	360;
	4318	10357	0.000531	0.021225	0	0	0	0	1.025043	0	1	-360	360;
	4318	3255	0.000531	0.021225	0	0	0	0	1.025043	0	1	-360	360;
	4318	13137	0.000531	0.021225	0	0	0	0	1.025221	0	1	-360	360;
	4318	1851	0.000531	0.021225	0	0	0	0	1.025221	0	1	-360	360;
	4318	9392	0.000531	0.021225	0	0	0	0	1.025221	0	1	-360	360;
	4318	2302	0.000531	0.021225	0	0	0	0	1.025221	0	1	-360	360;
	8240	10946	0.000531	0.021225	0	0	0	0	1.016109	0	1	-360	360;
	8472	13145	0.000531	0.021225	0	0	0	0	1.024494	0	1	-360	360;
	5555	13081	0.005208	0.208268	0	0	0	0	1.002029	0	1	-360	360;
	6712	6477	0.005208	0.208268	0	0	0	0	1.012988	0	1	-360	360;
	6712	12570	0.005208	0.208268	0	0	0	0	1.012988	0	1	-360	360;
	13061	4622	0.000531	0.021225	0	0	0	0	1.020316	0	1	-360	360;
	5474	12088	0.005208	0.208268	0	0	0	0	0.998327	0	1	-360	360;
	200	5512	0.005208	0.208268	0	0	0	0	0.998339	0	1	-360	360;
	9356	11579	0.005208	0.208268	0	0	0	0	1.026884	0	1	-360	360;
	9356	5025	0.005208	0.208268	0	0	0	0	1.026884	0	1	-360	360;
	4325	532	0.000531	0.021225	0	0	0	0	1.021429	0	1	-360	360;
	4325	7072	0.000531	0.021225	0	0	0	0	1.021429	0	1	-360	360;
	10849	12373	0.000531	0.021225	0	0	0	0	1.022217	0	1	-360	360;
	10849	5816	0.000531	0.021225	0	0	0	0	1.022217	0	1	-360	360;
	9162	9250	0.000531	0.021225	0	0	0	0	1.017502	0	1	-360	360;
	9162	3463	0.000531	0.021225	0	0	0	0	1.017502	0	1	-360	360;
	4316	6713	0.003521	0.140801	0	0	0	0	1.003702	0	1	-360	360;
	7104	3943	0.000531	0.021225	0	0	0	0	1.0207	0	1	-360	360;
	5867	12477	0.005208	0.208268	0	0	0	0	1.014852	0	1	-360	360;
	5867	5911	0.005208	0.208268	0	0	0	0	1.014852	0	1	-360	360;
	2926	12013	0.001328	0.05309	0	0	0	0	1.015309	0	1	-360	360;
	857	4952	0.000531	0.021225	0	0	0	0	1.019486	0	1	-360	360;
	9320	11037	0.000531	0.021225	0	0	0	0	1.009842	0	1	-360	360;
	12514	4473	0.000531	0.021225	0	0	0	0	1.018644	0	1	-360	360;
	3952	1369	0.000531	0.021225	0	0	0	0	1.0148	0	1	-360	360;
	3952	7925	0.000531	0.021225	0	0	0	0	1.0148	0	1	-360	360;
	3952	925	0.000531	0.021225	0	0	0	0	1.0148	0	1	-360	360;
	3952	3469	0.000531	0.021225	0	0	0	0	1.011666	0	1	-360	360;
	611	6096	0.005208	0.208268	0	0	0	0	0.992718	0	1	-360	360;
	3532	9542	0.005208	0.208268	0	0	0	0	0.988462	0	1	-360	360;
	3532	4256	0.005208	0.208268	0	0	0	0	0.988462	0	1	-360	360;
	9390	10841	0.002119	0.084719	0	0	0	0	0.990449	0	1	-360	360;
	9390	9556	0.002119	0.084719	0	0	0	0	0.990449	0	1	-360	360;
	11547	9690	0.000531	0.021225	0	0	0	0	1.018049	0	1	-360	360;
	11547	2584	0.000531	0.021225	0	0	0	0	1.018049	0	1	-360	360;
	11488	12799	0.002119	0.084719	0	0	0	0	1.01708	0	1	-360	360;
	11488	4686	0.002119	0.084719	0	0	0	0	1.01708	0	1	-360	360;
	11488	11251	0.002119	0.084719	0	0	0	0	1.01708	0	1	-360	360;
	49	4126	0.005208	0.208268	0	0	0	0	1.001032	0	1	-360	360;
	1681	11214	0.005208	0.208268	0	0	0	0	1.000717	0	1	-360	360;
	574	3646	0.005208	0.208268	0	0	0	0	1.005173	0	1	-360	360;
	574	10743	0.005208	0.208268	0	0	0	0	1.005173	0	1	-360	360;
	574	1070	0.005208	0.208268	0	0	0	0	1.005173	0	1	-360	360;
	5461	7613	0.005208	0.208268	0	0	0	0	0.999317	0	1	-360	360;
	5461	1532	0.005208	0.208268	0	0	0	0	0.999317	0	1	-360	360;
	6248	8089	0.005208	0.208268	0	0	0	0	1.002569	0	1	-360	360;
	12976	12042	0.005208	0.208268	0	0	0	0	1.002186	0	1	-360	360;
	13210	8563	0.005208	0.208268	0	0	0	0	0.999674	0	1	-360	360;
	13210	2458	0.005208	0.208268	0	0	0	0	0.999674	0	1	-360	360;
	9451	12419	0.003521	0.140801	0	0	0	0	1.001874	0	1	-360	360;
	9451	9120	0.003521	0.140801	0	0	0	0	1.001874	0	1	-360	360;
	9451	2063	0.003521	0.140801	0	0	0	0	1.001874	0	1	-360	360;
	4093	4167	0.005208	0.208268	0	0	0	0	0.997667	0	1	-360	360;
	4093	11250	0.005208	0.208268	0	0	0	0	0.997667	0	1	-360	360;
	4093	1359	0.005208	0.208268	0	0	0	0	0.997667	0	1	-360	360;
	4093	11773	0.005208	0.208268	0	0	0	0	0.997667	0	1	-360	360;
	2404	3166	0.005208	0.208268	0	0	0	0	1.000999	0	1	-360	360;
	2404	10283	0.005208	0.208268	0	0	0	0	1.000999	0	1	-360	360;
	2404	3680	0.005208	0.208268	0	0	0	0	1.000999	0	1	-360	360;
	9259	4130	0.000708	0.02832	0	0	0	0	0.997419	0	1	-360	360;
	5259	6195	0.005208	0.208268	0	0	0	0	0.999525	0	1	-360	360;
	5259	13298	0.005208	0.208268	0	0	0	0	0.999525	0	1	-360	360;
	1129	8421	0.005208	0.208268	0	0	0	0	0.999538	0	1	-360	360;
	1129	2222	0.005208	0.208268	0	0	0	0	0.999538	0	1	-360	360;
	1016	11377	0.000708	0.02832	0	0	0	0	0.9975	0	1	-360	360;
	8230	1796	0.000708	0.02832	0	0	0	0	0.996803	0	1	-360	360;
	6356	11895	0.003521	0.140801	0	0	0	0	0.996434	0	1	-360	360;
	6356	13232	0.003521	0.140801	0	0	0	0	0.996434	0	1	-360	360;
	6356	1043	0.003521	0.140801	0	0	0	0	0.996434	0	1	-360	360;
	2846	8945	0.003521	0.140801	0	0	0	0	0.993484	0	1	-360	360;
	2846	2787	0.003521	0.140801	0	0	0	0	0.993484	0	1	-360	360;
	2846	9417	0.003521	0.140801	0	0	0	0	0.993484	0	1	-360	360;
	915	9282	0.003521	0.140801	0	0	0	0	0.989587	0	1	-360	360;
	915	2647	0.003521	0.140801	0	0	0	0	0.989587	0	1	-360	360;
	915	9756	0.003521	0.140801	0	0	0	0	0.989587	0	1	-360	360;
	11381	3396	0.005208	0.208268	0	0	0	0	1.003999	0	1	-360	360;
	11381	11473	0.005208	0.208268	0	0	0	0	1.003999	0	1	-360	360;
	11381	4925	0.005208	0.208268	0	0	0	0	1.003999	0	1	-360	360;
	960	10978	0.005208	0.208268	0	0	0	0	1.010194	0	1	-360	360;
	960	4411	0.005208	0.208268	0	0	0	0	1.010194	0	1	-360	360;
	960	10862	0.005208	0.208268	0	0	0	0	1.010194	0	1	-360	360;
	5860	1960	0.005208	0.208268	0	0	0	0	1.010498	0	1	-360	360;
	5860	115	0.005208	0.208268	0	0	0	0	1.010498	0	1	-360	360;
	4401	6066	0.000531	0.021225	0	0	0	0	0.99924	0	1	-360	360;
	2240	11745	0.002119	0.084719	0	0	0	0	1.026659	0	1	-360	360;
	2240	4697	0.002119	0.084719	0	0	0	0	1.026659	0	1	-360	360;
	12752	11226	0.002119	0.084719	0	0	0	0	1.020882	0	1	-360	360;
	12752	3663	0.002119	0.084719	0	0	0	0	1.020882	0	1	-360	360;
	12752	9793	0.002119	0.084719	0	0	0	0	1.020882	0	1	-360	360;
	10741	6156	0.005208	0.208268	0	0	0	0	1.016153	0	1	-360	360;
	10741	11716	0.005208	0.208268	0	0	0	0	1.016153	0	1	-360	360;
	4033	1013	0.005208	0.208268	0	0	0	0	0.993208	0	1	-360	360;
	4033	9138	0.005208	0.208268	0	0	0	0	0.993208	0	1	-360	360;
	8480	10037	0.003521	0.140801	0	0	0	0	0.993172	0	1	-360	360;
	8480	1869	0.003521	0.140801	0	0	0	0	0.993172	0	1	-360	360;
	8480	12388	0.003521	0.140801	0	0	0	0	0.993172	0	1	-360	360;
	5565	10110	0.005208	0.208268	0	0	0	0	0.996909	0	1	-360	360;
	5565	7976	0.005208	0.208268	0	0	0	0	0.996909	0	1	-360	360;
	5565	9314	0.005208	0.208268	0	0	0	0	0.996909	0	1	-360	360;
	3809	10534	0.003521	0.140801	0	0	0	0	1.004248	0	1	-360	360;
	3809	4949	0.003521	0.140801	0	0	0	0	1.004248	0	1	-360	360;
	3809	11986	0.003521	0.140801	0	0	0	0	1.004248	0	1	-360	360;
	1490	10017	0.002119	0.084719	0	0	0	0	1.018767	0	1	-360	360;
	1490	3403	0.002119	0.084719	0	0	0	0	1.018767	0	1	-360	360;
	1490	9508	0.002119	0.084719	0	0	0	0	1.018767	0	1	-360	360;
	12185	4793	0.005208	0.208268	0	0	0	0	1.002358	0	1	-360	360;
	12185	13330	0.005208	0.208268	0	0	0	0	1.002358	0	1	-360	360;
	12185	10223	0.005208	0.208268	0	0	0	0	1.002358	0	1	-360	360;
	3140	8538	0.002119	0.084719	0	0	0	0	1.007799	0	1	-360	360;
	3140	1963	0.002119	0.084719	0	0	0	0	1.007799	0	1	-360	360;
	3140	11975	0.002119	0.084719	0	0	0	0	1.007799	0	1	-360	360;
	9050	7156	0.005208	0.208268	0	0	0	0	0.997082	0	1	-360	360;
	9050	5715	0.005208	0.208268	0	0	0	0	0.997082	0	1	-360	360;
	11216	5896	0.003521	0.140801	0	0	0	0	1.016528	0	1	-360	360;
	11216	13468	0.003521	0.140801	0	0	0	0	1.016528	0	1	-360	360;
	11216	6363	0.003521	0.140801	0	0	0	0	1.016528	0	1	-360	360;
	2848	11976	0.003521	0.140801	0	0	0	0	1.027914	0	1	-360	360;
	2848	4922	0.003521	0.140801	0	0	0	0	1.027914	0	1	-360	360;
	2848	12459	0.003521	0.140801	0	0	0	0	1.027914	0	1	-360	360;
	5799	8655	0.002119	0.084719	0	0	0	0	1.009997	0	1	-360	360;
	5799	8912	0.002119	0.084719	0	0	0	0	1.009997	0	1	-360	360;
	5799	1839	0.002119	0.084719	0	0	0	0	1.009997	0	1	-360	360;
	5799	11558	0.002119	0.084719	0	0	0	0	1.009997	0	1	-360	360;
	5799	4998	0.002119	0.084719	0	0	0	0	1.009997	0	1	-360	360;
	5799	11051	0.002119	0.084719	0	0	0	0	1.009997	0	1	-360	360;
	5799	4499	0.002119	0.084719	0	0	0	0	1.009997	0	1	-360	360;
	5799	10589	0.002119	0.084719	0	0	0	0	1.009997	0	1	-360	360;
	5799	3971	0.002119	0.084719	0	0	0	0	1.009997	0	1	-360	360;
	5799	10101	0.002119	0.084719	0	0	0	0	1.009997	0	1	-360	360;
	10826	3482	0.002119	0.084719	0	0	0	0	1.004834	0	1	-360	360;
	10826	9613	0.002119	0.084719	0	0	0	0	1.004834	0	1	-360	360;
	10826	2976	0.002119	0.084719	0	0	0	0	1.004834	0	1	-360	360;
	10826	8387	0.002119	0.084719	0	0	0	0	1.004834	0	1	-360	360;
	10826	1344	0.002119	0.084719	0	0	0	0	1.004834	0	1	-360	360;
	10826	10617	0.002119	0.084719	0	0	0	0	1.004834	0	1	-360	360;
	10826	901	0.002119	0.084719	0	0	0	0	1.004834	0	1	-360	360;
	10826	9359	0.002119	0.084719	0	0	0	0	1.004834	0	1	-360	360;
	10826	681	0.002119	0.084719	0	0	0	0	1.004834	0	1	-360	360;
	10826	8884	0.002119	0.084719	0	0	0	0	1.004834	0	1	-360	360;
	10180	7849	0.000531	0.021225	0	0	0	0	1.016245	0	1	-360	360;
	9689	11736	0.002119	0.084719	0	0	0	0	1.016673	0	1	-360	360;
	10859	2569	0.000708	0.02832	0	0	0	0	1.011995	0	1	-360	360;
	7106	1627	0.000531	0.021225	0	0	0	0	1.018131	0	1	-360	360;
	4150	10373	0.001059	0.04236	0	0	0	0	1.019146	0	1	-360	360;
	119	5192	0.000708	0.02832	0	0	0	0	1.01486	0	1	-360	360;
	8722	921	0.000708	0.02832	0	0	0	0	1.004646	0	1	-360	360;
	8722	676	0.000708	0.02832	0	0	0	0	1.004646	0	1	-360	360;
	8722	1372	0.000708	0.02832	0	0	0	0	1.004646	0	1	-360	360;
	8722	8894	0.000708	0.02832	0	0	0	0	1.004646	0	1	-360	360;
	8410	5138	0.003521	0.140801	0	0	0	0	1.009277	0	1	-360	360;
	8410	10246	0.003521	0.140801	0	0	0	0	1.009277	0	1	-360	360;
	8410	3633	0.003521	0.140801	0	0	0	0	1.009277	0	1	-360	360;
	8410	9342	0.003521	0.140801	0	0	0	0	1.009277	0	1	-360	360;
	1303	3984	0.000708	0.02832	0	0	0	0	1.014305	0	1	-360	360;
	1303	12415	0.000708	0.02832	0	0	0	0	1.014305	0	1	-360	360;
	1303	4505	0.000708	0.02832	0	0	0	0	1.014305	0	1	-360	360;
	3116	2205	0.025773	1.030606	0	0	0	0	1.003846	0	1	-360	360;
	3116	8791	0.025773	1.030606	0	0	0	0	1.00377	0	1	-360	360;
	7896	10448	0.003521	0.140801	0	0	0	0	1.026048	0	1	-360	360;
	7896	3343	0.003521	0.140801	0	0	0	0	1.026048	0	1	-360	360;
	7896	7133	0.003521	0.140801	0	0	0	0	1.026048	0	1	-360	360;
	7896	134	0.003521	0.140801	0	0	0	0	1.026048	0	1	-360	360;
	10888	197	0.000708	0.02832	0	0	0	0	1.012659	0	1	-360	360;
	10888	6734	0.000708	0.02832	0	0	0	0	1.012659	0	1	-360	360;
	10888	926	0.000708	0.02832	0	0	0	0	1.012659	0	1	-360	360;
	5604	12411	0.002119	0.084719	0	0	0	0	1.005436	0	1	-360	360;
	5604	5850	0.002119	0.084719	0	0	0	0	1.005436	0	1	-360	360;
	5604	12934	0.002119	0.084719	0	0	0	0	1.005436	0	1	-360	360;
	5604	6331	0.002119	0.084719	0	0	0	0	1.005436	0	1	-360	360;
	1388	7354	0.000212	0.008494	0	0	0	0	1.010913	0	1	-360	360;
	1388	11521	0.000212	0.008494	0	0	0	0	1.010913	0	1	-360	360;
	1388	4959	0.000212	0.008494	0	0	0	0	1.010913	0	1	-360	360;
	2306	3409	0.013298	0.531749	0	0	0	0	1.010032	0	1	-360	360;
	2306	10022	0.013298	0.531749	0	0	0	0	1.004611	0	1	-360	360;
	2306	3886	0.013298	0.531749	0	0	0	0	1.004611	0	1	-360	360;
	2306	10516	0.013298	0.531749	0	0	0	0	1.004611	0	1	-360	360;
	6137	12097	0.002119	0.084719	0	0	0	0	1.004259	0	1	-360	360;
	6137	5546	0.002119	0.084719	0	0	0	0	1.004259	0	1	-360	360;
	6137	7539	0.002119	0.084719	0	0	0	0	1.004259	0	1	-360	360;
	6137	6025	0.002119	0.084719	0	0	0	0	1.004259	0	1	-360	360;
	6137	8519	0.002119	0.084719	0	0	0	0	1.004259	0	1	-360	360;
	6137	8389	0.002119	0.084719	0	0	0	0	1.004259	0	1	-360	360;
	6137	2510	0.002119	0.084719	0	0	0	0	1.004259	0	1	-360	360;
	6137	815	0.002119	0.084719	0	0	0	0	1.004259	0	1	-360	360;
	1029	3893	0.012755	0.510045	0	0	0	0	1.022966	0	1	-360	360;
	1029	10986	0.012755	0.510045	0	0	0	0	1.017812	0	1	-360	360;
	1029	3434	0.012755	0.510045	0	0	0	0	1.017812	0	1	-360	360;
	1029	10520	0.012755	0.510045	0	0	0	0	1.017812	0	1	-360	360;
	12004	1123	0.002119	0.084719	0	0	0	0	1.015216	0	1	-360	360;
	12004	7667	0.002119	0.084719	0	0	0	0	1.015216	0	1	-360	360;
	12004	4495	0.002119	0.084719	0	0	0	0	1.015216	0	1	-360	360;
	12004	11049	0.002119	0.084719	0	0	0	0	1.015216	0	1	-360	360;
	12004	10629	0.002119	0.084719	0	0	0	0	1.015216	0	1	-360	360;
	12004	4000	0.002119	0.084719	0	0	0	0	1.015216	0	1	-360	360;
	12004	6912	0.002119	0.084719	0	0	0	0	1.015216	0	1	-360	360;
	290	2464	0.039062	1.562012	0	0	0	0	1.005436	0	1	-360	360;
	2064	5135	0.000531	0.021225	0	0	0	0	1.005105	0	1	-360	360;
	2064	12211	0.000531	0.021225	0	0	0	0	1.005105	0	1	-360	360;
	2064	4616	0.000531	0.021225	0	0	0	0	1.005105	0	1	-360	360;
	2064	11680	0.000531	0.021225	0	0	0	0	1.005105	0	1	-360	360;
	2064	4102	0.000531	0.021225	0	0	0	0	1.005105	0	1	-360	360;
	11501	1104	0.005208	0.208268	0	0	0	0	1.001065	0	1	-360	360;
	11501	1055	0.005208	0.208268	0	0	0	0	1.001065	0	1	-360	360;
	11501	408	0.005208	0.208268	0	0	0	0	1.001065	0	1	-360	360;
	11501	7371	0.005208	0.208268	0	0	0	0	1.001065	0	1	-360	360;
	11501	5391	0.005208	0.208268	0	0	0	0	1.001065	0	1	-360	360;
	11501	8820	0.005208	0.208268	0	0	0	0	1.001065	0	1	-360	360;
	3786	10998	0.000212	0.008494	0	0	0	0	1.020727	0	1	-360	360;
	3786	3265	0.000212	0.008494	0	0	0	0	1.020727	0	1	-360	360;
	2557	13607	0.005208	0.208268	0	0	0	0	1.004215	0	1	-360	360;
	2557	4987	0.005208	0.208268	0	0	0	0	1.004215	0	1	-360	360;
	6841	4326	0.005208	0.208268	0	0	0	0	1.000189	0	1	-360	360;
	6841	12542	0.005208	0.208268	0	0	0	0	1.000189	0	1	-360	360;
	6841	5977	0.005208	0.208268	0	0	0	0	1.000189	0	1	-360	360;
	8978	12045	0.005208	0.208268	0	0	0	0	1.001074	0	1	-360	360;
	1377	5490	0.009653	0.38598	0	0	0	0	0.999185	0	1	-360	360;
	1377	11560	0.009653	0.38598	0	0	0	0	0.993513	0	1	-360	360;
	1377	2809	0.009653	0.38598	0	0	0	0	0.993513	0	1	-360	360;
	1377	12323	0.009653	0.38598	0	0	0	0	0.993513	0	1	-360	360;
	1377	188	0.009653	0.38598	0	0	0	0	0.993513	0	1	-360	360;
	7876	4703	0.003521	0.140801	0	0	0	0	1.004786	0	1	-360	360;
	7876	11749	0.003521	0.140801	0	0	0	0	1.004786	0	1	-360	360;
	7876	9713	0.003521	0.140801	0	0	0	0	1.004786	0	1	-360	360;
	7876	2610	0.003521	0.140801	0	0	0	0	1.004786	0	1	-360	360;
	13307	9144	0.002119	0.084719	0	0	0	0	1.011119	0	1	-360	360;
	13307	1423	0.002119	0.084719	0	0	0	0	1.011119	0	1	-360	360;
	13307	6998	0.002119	0.084719	0	0	0	0	1.011119	0	1	-360	360;
	13307	4347	0.002119	0.084719	0	0	0	0	1.011119	0	1	-360	360;
	13307	10120	0.002119	0.084719	0	0	0	0	1.011119	0	1	-360	360;
	1996	1462	0.002119	0.084719	0	0	0	0	1.007174	0	1	-360	360;
	1996	9512	0.002119	0.084719	0	0	0	0	1.007174	0	1	-360	360;
	1996	2408	0.002119	0.084719	0	0	0	0	1.007174	0	1	-360	360;
	1996	841	0.002119	0.084719	0	0	0	0	1.007174	0	1	-360	360;
	1996	7351	0.002119	0.084719	0	0	0	0	1.007174	0	1	-360	360;
	1996	1274	0.002119	0.084719	0	0	0	0	1.007174	0	1	-360	360;
	1996	6650	0.002119	0.084719	0	0	0	0	1.007174	0	1	-360	360;
	1996	13569	0.002119	0.084719	0	0	0	0	1.007174	0	1	-360	360;
	10312	637	0.005208	0.208268	0	0	0	0	1.016623	0	1	-360	360;
	10312	7634	0.005208	0.208268	0	0	0	0	1.016623	0	1	-360	360;
	3461	12786	0.003521	0.140801	0	0	0	0	1.005028	0	1	-360	360;
	3461	7122	0.003521	0.140801	0	0	0	0	1.005028	0	1	-360	360;
	3461	7304	0.003521	0.140801	0	0	0	0	1.005028	0	1	-360	360;
	3461	9357	0.003521	0.140801	0	0	0	0	1.005028	0	1	-360	360;
	3461	1257	0.003521	0.140801	0	0	0	0	1.005028	0	1	-360	360;
	3461	8774	0.003521	0.140801	0	0	0	0	1.005028	0	1	-360	360;
	3461	1732	0.003521	0.140801	0	0	0	0	1.005028	0	1	-360	360;
	5419	3329	0.002119	0.084719	0	0	0	0	1.001974	0	1	-360	360;
	5419	9953	0.002119	0.084719	0	0	0	0	1.001974	0	1	-360	360;
	5419	2828	0.002119	0.084719	0	0	0	0	1.001974	0	1	-360	360;
	5419	9464	0.002119	0.084719	0	0	0	0	1.001974	0	1	-360	360;
	5419	4301	0.002119	0.084719	0	0	0	0	1.001974	0	1	-360	360;
	5419	10907	0.002119	0.084719	0	0	0	0	1.001974	0	1	-360	360;
	1457	3228	0.005208	0.208268	0	0	0	0	0.997477	0	1	-360	360;
	1457	783	0.005208	0.208268	0	0	0	0	0.997477	0	1	-360	360;
	1457	6861	0.005208	0.208268	0	0	0	0	0.997477	0	1	-360	360;
	6367	304	0.005208	0.208268	0	0	0	0	1.007255	0	1	-360	360;
	13233	4209	0.002119	0.084719	0	0	0	0	1.013579	0	1	-360	360;
	13233	10808	0.002119	0.084719	0	0	0	0	1.013579	0	1	-360	360;
	13233	3710	0.002119	0.084719	0	0	0	0	1.013579	0	1	-360	360;
	13233	7784	0.002119	0.084719	0	0	0	0	1.013579	0	1	-360	360;
	13233	1306	0.002119	0.084719	0	0	0	0	1.013579	0	1	-360	360;
	13233	6870	0.002119	0.084719	0	0	0	0	1.013579	0	1	-360	360;
	13233	179	0.002119	0.084719	0	0	0	0	1.013579	0	1	-360	360;
	13233	11968	0.002119	0.084719	0	0	0	0	1.013579	0	1	-360	360;
	8692	10315	0.003521	0.140801	0	0	0	0	1.004319	0	1	-360	360;
	8692	3206	0.003521	0.140801	0	0	0	0	1.004319	0	1	-360	360;
	8692	11761	0.003521	0.140801	0	0	0	0	1.004319	0	1	-360	360;
	8692	4715	0.003521	0.140801	0	0	0	0	1.004319	0	1	-360	360;
	8692	33	0.003521	0.140801	0	0	0	0	1.004319	0	1	-360	360;
	8692	4217	0.003521	0.140801	0	0	0	0	1.004319	0	1	-360	360;
	7241	5473	0.002119	0.084719	0	0	0	0	1.014928	0	1	-360	360;
	7241	12024	0.002119	0.084719	0	0	0	0	1.014928	0	1	-360	360;
	7241	7297	0.002119	0.084719	0	0	0	0	1.014928	0	1	-360	360;
	7241	5317	0.002119	0.084719	0	0	0	0	1.014928	0	1	-360	360;
	7241	12341	0.002119	0.084719	0	0	0	0	1.014928	0	1	-360	360;
	4003	11797	0.000708	0.02832	0	0	0	0	1.001878	0	1	-360	360;
	4003	4757	0.000708	0.02832	0	0	0	0	1.001878	0	1	-360	360;
	4003	7833	0.000708	0.02832	0	0	0	0	1.001878	0	1	-360	360;
	4003	4226	0.000708	0.02832	0	0	0	0	1.001878	0	1	-360	360;
	10578	5992	0.003521	0.140801	0	0	0	0	1.012864	0	1	-360	360;
	10578	2183	0.003521	0.140801	0	0	0	0	1.012864	0	1	-360	360;
	10578	9273	0.003521	0.140801	0	0	0	0	1.012864	0	1	-360	360;
	10578	3384	0.003521	0.140801	0	0	0	0	1.012864	0	1	-360	360;
	10578	10004	0.003521	0.140801	0	0	0	0	1.012864	0	1	-360	360;
	10578	3877	0.003521	0.140801	0	0	0	0	1.012864	0	1	-360	360;
	10578	8810	0.003521	0.140801	0	0	0	0	1.012864	0	1	-360	360;
	64	3148	0.027778	1.110764	0	0	0	0	1.000772	0	1	-360	360;
	64	4673	0.027778	1.110764	0	0	0	0	0.99517	0	1	-360	360;
	11878	2676	0.005208	0.208268	0	0	0	0	1.004754	0	1	-360	360;
	4566	9312	0.022727	0.908807	0	0	0	0	1.00487	0	1	-360	360;
	4566	6127	0.022727	0.908807	0	0	0	0	0.999253	0	1	-360	360;
	13117	8099	0.003521	0.140801	0	0	0	0	1.005602	0	1	-360	360;
	13117	2635	0.003521	0.140801	0	0	0	0	1.005602	0	1	-360	360;
	13117	9742	0.003521	0.140801	0	0	0	0	1.005602	0	1	-360	360;
	13117	3146	0.003521	0.140801	0	0	0	0	1.005602	0	1	-360	360;
	13117	10264	0.003521	0.140801	0	0	0	0	1.005602	0	1	-360	360;
	13117	1731	0.003521	0.140801	0	0	0	0	1.005602	0	1	-360	360;
	4557	1148	0.022936	0.917144	0	0	0	0	1.003357	0	1	-360	360;
	4557	2793	0.022936	0.917144	0	0	0	0	0.997818	0	1	-360	360;
	5258	698	0.007716	0.308546	0	0	0	0	1.003186	0	1	-360	360;
	5258	10615	0.007716	0.308546	0	0	0	0	0.997349	0	1	-360	360;
	8475	13411	0.002119	0.084719	0	0	0	0	1.020516	0	1	-360	360;
	8475	8342	0.002119	0.084719	0	0	0	0	1.020516	0	1	-360	360;
	8475	1310	0.002119	0.084719	0	0	0	0	1.020516	0	1	-360	360;
	8475	2347	0.002119	0.084719	0	0	0	0	1.020516	0	1	-360	360;
	8475	1774	0.002119	0.084719	0	0	0	0	1.020516	0	1	-360	360;
	2059	12887	0.038462	1.537981	0	0	0	0	1.00422	0	1	-360	360;
	7254	5805	0.005208	0.208268	0	0	0	0	1.00527	0	1	-360	360;
	7254	13370	0.005208	0.208268	0	0	0	0	1.00527	0	1	-360	360;
	7254	6278	0.005208	0.208268	0	0	0	0	1.00527	0	1	-360	360;
	2775	9881	0.043103	1.723599	0	0	0	0	1.012693	0	1	-360	360;
	10704	5260	0.000708	0.02832	0	0	0	0	0.996531	0	1	-360	360;
	10704	2881	0.000708	0.02832	0	0	0	0	0.996531	0	1	-360	360;
	10704	2880	0.000708	0.02832	0	0	0	0	0.996531	0	1	-360	360;
	13279	8530	0.010638	0.425399	0	0	0	0	1.003196	0	1	-360	360;
	13279	8531	0.010638	0.425399	0	0	0	0	0.997489	0	1	-360	360;
	13279	8534	0.010638	0.425399	0	0	0	0	0.997489	0	1	-360	360;
	13279	8535	0.010638	0.425399	0	0	0	0	0.997489	0	1	-360	360;
	13279	8533	0.010638	0.425399	0	0	0	0	0.997489	0	1	-360	360;
	5256	6155	0.006649	0.265874	0	0	0	0	1.00474	0	1	-360	360;
	5256	6159	0.006649	0.265874	0	0	0	0	0.998979	0	1	-360	360;
	5256	6160	0.006649	0.265874	0	0	0	0	0.998979	0	1	-360	360;
	5256	10977	0.006649	0.265874	0	0	0	0	0.998979	0	1	-360	360;
	5256	13261	0.006649	0.265874	0	0	0	0	0.998979	0	1	-360	360;
	1068	6548	0.002119	0.084719	0	0	0	0	1.012774	0	1	-360	360;
	1068	6547	0.002119	0.084719	0	0	0	0	1.012774	0	1	-360	360;
	1068	13069	0.002119	0.084719	0	0	0	0	1.012774	0	1	-360	360;
	1068	12102	0.002119	0.084719	0	0	0	0	1.012774	0	1	-360	360;
	1068	6544	0.002119	0.084719	0	0	0	0	1.012774	0	1	-360	360;
	10340	1957	0.016234	0.649148	0	0	0	0	1.001406	0	1	-360	360;
	10340	1954	0.016234	0.649148	0	0	0	0	0.995859	0	1	-360	360;
	10340	1953	0.016234	0.649148	0	0	0	0	0.995859	0	1	-360	360;
	5566	8053	0.005208	0.208268	0	0	0	0	1.009577	0	1	-360	360;
	5566	7943	0.005208	0.208268	0	0	0	0	1.009577	0	1	-360	360;
	5566	8055	0.005208	0.208268	0	0	0	0	1.009577	0	1	-360	360;
	5566	8056	0.005208	0.208268	0	0	0	0	1.009577	0	1	-360	360;
	5566	7490	0.005208	0.208268	0	0	0	0	1.009577	0	1	-360	360;
	5566	7932	0.005208	0.208268	0	0	0	0	1.009577	0	1	-360	360;
	11468	6641	0.003521	0.140801	0	0	0	0	1.014193	0	1	-360	360;
	11468	6642	0.003521	0.140801	0	0	0	0	1.014193	0	1	-360	360;
	11468	6639	0.003521	0.140801	0	0	0	0	1.014193	0	1	-360	360;
	11468	6640	0.003521	0.140801	0	0	0	0	1.014193	0	1	-360	360;
	11468	100	0.003521	0.140801	0	0	0	0	1.014193	0	1	-360	360;
	11468	99	0.003521	0.140801	0	0	0	0	1.014193	0	1	-360	360;
	2714	102	0.003521	0.140801	0	0	0	0	1.025369	0	1	-360	360;
	2714	101	0.003521	0.140801	0	0	0	0	1.025369	0	1	-360	360;
	2714	97	0.003521	0.140801	0	0	0	0	1.025369	0	1	-360	360;
	2714	96	0.003521	0.140801	0	0	0	0	1.025369	0	1	-360	360;
	2714	98	0.003521	0.140801	0	0	0	0	1.025369	0	1	-360	360;
	9646	1505	0.009728	0.388983	0	0	0	0	0.99539	0	1	-360	360;
	9646	9416	0.009728	0.388983	0	0	0	0	0.989722	0	1	-360	360;
	9646	1511	0.009728	0.388983	0	0	0	0	0.989722	0	1	-360	360;
	9646	8200	0.009728	0.388983	0	0	0	0	0.989722	0	1	-360	360;
	9646	7587	0.009728	0.388983	0	0	0	0	0.989722	0	1	-360	360;
	3450	5212	0.003098	0.123877	0	0	0	0	1.010409	0	1	-360	360;
	3450	5213	0.003098	0.123877	0	0	0	0	1.004436	0	1	-360	360;
	3450	5585	0.003098	0.123877	0	0	0	0	1.004436	0	1	-360	360;
	3450	5215	0.003098	0.123877	0	0	0	0	1.004436	0	1	-360	360;
	3450	5216	0.003098	0.123877	0	0	0	0	1.004436	0	1	-360	360;
	3450	5624	0.003098	0.123877	0	0	0	0	1.004436	0	1	-360	360;
	205	3867	0.000708	0.02832	0	0	0	0	1.009578	0	1	-360	360;
	205	3866	0.000708	0.02832	0	0	0	0	1.009578	0	1	-360	360;
	205	8759	0.000708	0.02832	0	0	0	0	1.009578	0	1	-360	360;
	2383	1044	0.039062	1.562012	0	0	0	0	1.001311	0	1	-360	360;
	719	1046	0.024272	0.97057	0	0	0	0	1.003605	0	1	-360	360;
	719	1045	0.024272	0.97057	0	0	0	0	0.998099	0	1	-360	360;
	900	362	0.002119	0.084719	0	0	0	0	1.000042	0	1	-360	360;
	900	363	0.002119	0.084719	0	0	0	0	1.000042	0	1	-360	360;
	900	360	0.002119	0.084719	0	0	0	0	1.000042	0	1	-360	360;
	900	361	0.002119	0.084719	0	0	0	0	1.000042	0	1	-360	360;
	900	368	0.002119	0.084719	0	0	0	0	1.000042	0	1	-360	360;
	900	369	0.002119	0.084719	0	0	0	0	1.000042	0	1	-360	360;
	900	6931	0.002119	0.084719	0	0	0	0	1.000042	0	1	-360	360;
	900	6930	0.002119	0.084719	0	0	0	0	1.000042	0	1	-360	360;
	2052	5777	0.000531	0.021225	0	0	0	0	1.008331	0	1	-360	360;
	2052	5776	0.000531	0.021225	0	0	0	0	1.008331	0	1	-360	360;
	2052	5782	0.000531	0.021225	0	0	0	0	1.008331	0	1	-360	360;
	2052	5781	0.000531	0.021225	0	0	0	0	1.008331	0	1	-360	360;
	1203	7143	0.02381	0.952083	0	0	0	0	1.012157	0	1	-360	360;
	1203	7144	0.02381	0.952083	0	0	0	0	1.006915	0	1	-360	360;
	5642	2166	0.009615	0.384495	0	0	0	0	1.009944	0	1	-360	360;
	5642	7141	0.009615	0.384495	0	0	0	0	1.004376	0	1	-360	360;
	5642	594	0.009615	0.384495	0	0	0	0	1.004376	0	1	-360	360;
	5642	593	0.009615	0.384495	0	0	0	0	1.004376	0	1	-360	360;
	5642	3614	0.009615	0.384495	0	0	0	0	1.004376	0	1	-360	360;
	7440	8343	0.003521	0.140801	0	0	0	0	1.013087	0	1	-360	360;
	7440	12761	0.003521	0.140801	0	0	0	0	1.013087	0	1	-360	360;
	7440	12760	0.003521	0.140801	0	0	0	0	1.013087	0	1	-360	360;
	7440	12763	0.003521	0.140801	0	0	0	0	1.013087	0	1	-360	360;
	7440	8354	0.003521	0.140801	0	0	0	0	1.013087	0	1	-360	360;
	7440	8357	0.003521	0.140801	0	0	0	0	1.013087	0	1	-360	360;
	10906	589	0.015244	0.609566	0	0	0	0	1.002407	0	1	-360	360;
	10906	10509	0.015244	0.609566	0	0	0	0	0.996654	0	1	-360	360;
	10906	10510	0.015244	0.609566	0	0	0	0	0.996654	0	1	-360	360;
	3191	10507	0.039062	1.562012	0	0	0	0	1.007221	0	1	-360	360;
	1982	10508	0.018797	0.751645	0	0	0	0	1.005485	0	1	-360	360;
	1982	6252	0.018797	0.751645	0	0	0	0	1.000027	0	1	-360	360;
	1982	10512	0.018797	0.751645	0	0	0	0	1.000027	0	1	-360	360;
	10450	4893	0.000708	0.02832	0	0	0	0	0.998716	0	1	-360	360;
	10450	4892	0.000708	0.02832	0	0	0	0	0.998716	0	1	-360	360;
	10450	4897	0.000708	0.02832	0	0	0	0	0.998716	0	1	-360	360;
	10450	4896	0.000708	0.02832	0	0	0	0	0.998716	0	1	-360	360;
	13135	1454	0.003521	0.140801	0	0	0	0	1.008012	0	1	-360	360;
	13135	1453	0.003521	0.140801	0	0	0	0	1.008012	0	1	-360	360;
	13135	9047	0.003521	0.140801	0	0	0	0	1.008012	0	1	-360	360;
	13135	1451	0.003521	0.140801	0	0	0	0	1.008012	0	1	-360	360;
	7100	10361	0.030488	1.219131	0	0	0	0	1.014807	0	1	-360	360;
	7100	3885	0.030488	1.219131	0	0	0	0	1.009649	0	1	-360	360;
	4705	1450	0.003521	0.140801	0	0	0	0	0.997964	0	1	-360	360;
	4705	1449	0.003521	0.140801	0	0	0	0	0.997964	0	1	-360	360;
	4705	1448	0.003521	0.140801	0	0	0	0	0.997964	0	1	-360	360;
	4705	1447	0.003521	0.140801	0	0	0	0	0.997964	0	1	-360	360;
	4705	9246	0.003521	0.140801	0	0	0	0	0.997964	0	1	-360	360;
	4705	1381	0.003521	0.140801	0	0	0	0	0.997964	0	1	-360	360;
	4705	9247	0.003521	0.140801	0	0	0	0	0.997964	0	1	-360	360;
	8484	11705	0.015625	0.624805	0	0	0	0	1.012683	0	1	-360	360;
	8484	11709	0.015625	0.624805	0	0	0	0	1.007095	0	1	-360	360;
	8484	2335	0.015625	0.624805	0	0	0	0	1.007095	0	1	-360	360;
	3846	9244	0.006039	0.24147	0	0	0	0	1.011205	0	1	-360	360;
	3846	9243	0.006039	0.24147	0	0	0	0	1.005549	0	1	-360	360;
	3846	9241	0.006039	0.24147	0	0	0	0	1.005549	0	1	-360	360;
	3846	9240	0.006039	0.24147	0	0	0	0	1.005549	0	1	-360	360;
	3846	2163	0.006039	0.24147	0	0	0	0	1.005549	0	1	-360	360;
	3846	2164	0.006039	0.24147	0	0	0	0	1.005549	0	1	-360	360;
	6501	4644	0.013812	0.552314	0	0	0	0	1.004827	0	1	-360	360;
	6501	13351	0.013812	0.552314	0	0	0	0	0.999231	0	1	-360	360;
	6501	4646	0.013812	0.552314	0	0	0	0	0.999231	0	1	-360	360;
	6501	4647	0.013812	0.552314	0	0	0	0	0.999231	0	1	-360	360;
	4869	4648	0.009579	0.383022	0	0	0	0	1.016353	0	1	-360	360;
	4869	4649	0.009579	0.383022	0	0	0	0	1.011028	0	1	-360	360;
	4869	4650	0.009579	0.383022	0	0	0	0	1.011028	0	1	-360	360;
	4869	4651	0.009579	0.383022	0	0	0	0	1.011028	0	1	-360	360;
	4869	12246	0.009579	0.383022	0	0	0	0	1.011028	0	1	-360	360;
	8880	12245	0.005208	0.208268	0	0	0	0	1.010158	0	1	-360	360;
	8880	7178	0.005208	0.208268	0	0	0	0	1.010158	0	1	-360	360;
	1612	8747	0.003521	0.140801	0	0	0	0	1.007099	0	1	-360	360;
	1612	8746	0.003521	0.140801	0	0	0	0	1.007099	0	1	-360	360;
	1612	8745	0.003521	0.140801	0	0	0	0	1.007099	0	1	-360	360;
	1612	8744	0.003521	0.140801	0	0	0	0	1.007099	0	1	-360	360;
	1612	4206	0.003521	0.140801	0	0	0	0	1.007099	0	1	-360	360;
	1612	8755	0.003521	0.140801	0	0	0	0	1.007099	0	1	-360	360;
	4466	12240	0.004019	0.160721	0	0	0	0	1.002751	0	1	-360	360;
	4466	5170	0.004019	0.160721	0	0	0	0	0.997027	0	1	-360	360;
	4466	5171	0.004019	0.160721	0	0	0	0	0.997027	0	1	-360	360;
	4466	5168	0.004019	0.160721	0	0	0	0	0.997027	0	1	-360	360;
	10044	5169	0.010373	0.414808	0	0	0	0	1.016656	0	1	-360	360;
	10044	12812	0.010373	0.414808	0	0	0	0	1.011392	0	1	-360	360;
	10044	5167	0.010373	0.414808	0	0	0	0	1.011392	0	1	-360	360;
	10044	5164	0.010373	0.414808	0	0	0	0	1.011392	0	1	-360	360;
	10044	5165	0.010373	0.414808	0	0	0	0	1.011392	0	1	-360	360;
	7181	5177	0.005208	0.208268	0	0	0	0	1.005009	0	1	-360	360;
	7181	5178	0.005208	0.208268	0	0	0	0	1.005009	0	1	-360	360;
	9773	12782	0.009615	0.384495	0	0	0	0	1.020687	0	1	-360	360;
	9773	12781	0.009615	0.384495	0	0	0	0	1.015549	0	1	-360	360;
	9773	12780	0.009615	0.384495	0	0	0	0	1.015549	0	1	-360	360;
	9773	12779	0.009615	0.384495	0	0	0	0	1.015549	0	1	-360	360;
	9773	12778	0.009615	0.384495	0	0	0	0	1.015549	0	1	-360	360;
	11466	12777	0.008091	0.323523	0	0	0	0	1.004404	0	1	-360	360;
	11466	8403	0.008091	0.323523	0	0	0	0	0.99875	0	1	-360	360;
	10984	364	0.039062	1.562012	0	0	0	0	1.003108	0	1	-360	360;
	6177	8396	0.048077	1.922476	0	0	0	0	1.003621	0	1	-360	360;
	3958	6601	0.000708	0.02832	0	0	0	0	1.00853	0	1	-360	360;
	3958	250	0.000708	0.02832	0	0	0	0	1.00853	0	1	-360	360;
	3958	45	0.000708	0.02832	0	0	0	0	1.00853	0	1	-360	360;
	456	5702	0.032468	1.298295	0	0	0	0	1.01309	0	1	-360	360;
	456	5703	0.032468	1.298295	0	0	0	0	1.007869	0	1	-360	360;
	449	6628	0.000531	0.021225	0	0	0	0	1.003276	0	1	-360	360;
	449	10117	0.000531	0.021225	0	0	0	0	1.003276	0	1	-360	360;
	449	10116	0.000531	0.021225	0	0	0	0	1.003276	0	1	-360	360;
	449	10115	0.000531	0.021225	0	0	0	0	1.003276	0	1	-360	360;
	40	5697	0.025773	1.030606	0	0	0	0	1.003295	0	1	-360	360;
	40	5698	0.025773	1.030606	0	0	0	0	0.997732	0	1	-360	360;
	8256	13268	0.012315	0.492457	0	0	0	0	1.003225	0	1	-360	360;
	8256	11288	0.012315	0.492457	0	0	0	0	0.997532	0	1	-360	360;
	8256	10317	0.012315	0.492457	0	0	0	0	0.997532	0	1	-360	360;
	8256	13269	0.012315	0.492457	0	0	0	0	0.997532	0	1	-360	360;
	12985	13265	0.020161	0.8062	0	0	0	0	0.994624	0	1	-360	360;
	12985	13264	0.020161	0.8062	0	0	0	0	0.98892	0	1	-360	360;
	12985	13266	0.020161	0.8062	0	0	0	0	0.98892	0	1	-360	360;
	9491	3496	0.000531	0.021225	0	0	0	0	0.997131	0	1	-360	360;
	9491	3494	0.000531	0.021225	0	0	0	0	0.997131	0	1	-360	360;
	9491	3495	0.000531	0.021225	0	0	0	0	0.997131	0	1	-360	360;
	6878	10954	0.000708	0.02832	0	0	0	0	0.995055	0	1	-360	360;
	6878	13209	0.000708	0.02832	0	0	0	0	0.995055	0	1	-360	360;
	6878	12556	0.000708	0.02832	0	0	0	0	0.995055	0	1	-360	360;
	1680	6163	0.04717	1.886203	0	0	0	0	1.002428	0	1	-360	360;
	999	6167	0.01	0.399875	0	0	0	0	1.003934	0	1	-360	360;
	999	6168	0.01	0.399875	0	0	0	0	0.998152	0	1	-360	360;
	999	6165	0.01	0.399875	0	0	0	0	0.998152	0	1	-360	360;
	999	6166	0.01	0.399875	0	0	0	0	0.998152	0	1	-360	360;
	999	6170	0.01	0.399875	0	0	0	0	0.998152	0	1	-360	360;
	257	809	0.006684	0.267296	0	0	0	0	1.003557	0	1	-360	360;
	257	11142	0.006684	0.267296	0	0	0	0	0.997824	0	1	-360	360;
	257	11141	0.006684	0.267296	0	0	0	0	0.997824	0	1	-360	360;
	257	11144	0.006684	0.267296	0	0	0	0	0.997824	0	1	-360	360;
	257	11143	0.006684	0.267296	0	0	0	0	0.997824	0	1	-360	360;
	2633	9778	0.018939	0.757339	0	0	0	0	1.001686	0	1	-360	360;
	2633	9777	0.018939	0.757339	0	0	0	0	0.996106	0	1	-360	360;
	2633	8562	0.018939	0.757339	0	0	0	0	0.996106	0	1	-360	360;
	8327	2107	0.01087	0.434647	0	0	0	0	1.006804	0	1	-360	360;
	8327	9772	0.01087	0.434647	0	0	0	0	1.001171	0	1	-360	360;
	8327	9771	0.01087	0.434647	0	0	0	0	1.001171	0	1	-360	360;
	8327	2664	0.01087	0.434647	0	0	0	0	1.001171	0	1	-360	360;
	8327	2665	0.01087	0.434647	0	0	0	0	1.001171	0	1	-360	360;
	11213	2666	0.005208	0.208268	0	0	0	0	1.0043	0	1	-360	360;
	11213	2667	0.005208	0.208268	0	0	0	0	1.0043	0	1	-360	360;
	11213	1988	0.005208	0.208268	0	0	0	0	1.0043	0	1	-360	360;
	11213	2669	0.005208	0.208268	0	0	0	0	1.0043	0	1	-360	360;
	9731	4066	0.003521	0.140801	0	0	0	0	1.007249	0	1	-360	360;
	9731	4067	0.003521	0.140801	0	0	0	0	1.007249	0	1	-360	360;
	9731	4068	0.003521	0.140801	0	0	0	0	1.007249	0	1	-360	360;
	9731	4069	0.003521	0.140801	0	0	0	0	1.007249	0	1	-360	360;
	9731	12434	0.003521	0.140801	0	0	0	0	1.007249	0	1	-360	360;
	13271	11294	0.000212	0.008494	0	0	0	0	1.021812	0	1	-360	360;
	13271	814	0.000212	0.008494	0	0	0	0	1.021812	0	1	-360	360;
	3807	12433	0.003521	0.140801	0	0	0	0	1.004319	0	1	-360	360;
	3807	12430	0.003521	0.140801	0	0	0	0	1.004319	0	1	-360	360;
	3807	12431	0.003521	0.140801	0	0	0	0	1.004319	0	1	-360	360;
	3807	12428	0.003521	0.140801	0	0	0	0	1.004319	0	1	-360	360;
	3807	12429	0.003521	0.140801	0	0	0	0	1.004319	0	1	-360	360;
	1682	12440	0.003521	0.140801	0	0	0	0	1.013896	0	1	-360	360;
	1682	6654	0.003521	0.140801	0	0	0	0	1.013896	0	1	-360	360;
	1682	7967	0.003521	0.140801	0	0	0	0	1.013896	0	1	-360	360;
	1682	5864	0.003521	0.140801	0	0	0	0	1.013896	0	1	-360	360;
	10272	11510	0.000708	0.02832	0	0	0	0	1.011734	0	1	-360	360;
	10272	10930	0.000708	0.02832	0	0	0	0	1.011734	0	1	-360	360;
	10272	10927	0.000708	0.02832	0	0	0	0	1.011734	0	1	-360	360;
	10272	10928	0.000708	0.02832	0	0	0	0	1.011734	0	1	-360	360;
	41	9327	0.038462	1.537981	0	0	0	0	1.003553	0	1	-360	360;
	4149	1836	0.05814	2.324855	0	0	0	0	1.000207	0	1	-360	360;
	5349	9323	0.005208	0.208268	0	0	0	0	0.999338	0	1	-360	360;
	5349	9322	0.005208	0.208268	0	0	0	0	0.999338	0	1	-360	360;
	5349	2693	0.005208	0.208268	0	0	0	0	0.999338	0	1	-360	360;
	5349	2694	0.005208	0.208268	0	0	0	0	0.999338	0	1	-360	360;
	340	2691	0.039062	1.562012	0	0	0	0	1.006031	0	1	-360	360;
	8019	3233	0.005208	0.208268	0	0	0	0	1.007887	0	1	-360	360;
	8019	3225	0.005208	0.208268	0	0	0	0	1.007887	0	1	-360	360;
	8019	2690	0.005208	0.208268	0	0	0	0	1.007887	0	1	-360	360;
	8019	2687	0.005208	0.208268	0	0	0	0	1.007887	0	1	-360	360;
	5457	2688	0.005208	0.208268	0	0	0	0	1.01053	0	1	-360	360;
	12079	346	0.000531	0.021225	0	0	0	0	0.998487	0	1	-360	360;
	12079	345	0.000531	0.021225	0	0	0	0	0.998487	0	1	-360	360;
	12079	646	0.000531	0.021225	0	0	0	0	0.998487	0	1	-360	360;
	12079	649	0.000531	0.021225	0	0	0	0	0.998487	0	1	-360	360;
	10130	12726	0.000708	0.02832	0	0	0	0	1.006201	0	1	-360	360;
	10130	10461	0.000708	0.02832	0	0	0	0	1.006201	0	1	-360	360;
	10130	10462	0.000708	0.02832	0	0	0	0	1.006201	0	1	-360	360;
	5854	5401	0.005353	0.214066	0	0	0	0	0.998883	0	1	-360	360;
	5854	5400	0.005353	0.214066	0	0	0	0	0.993032	0	1	-360	360;
	5854	5399	0.005353	0.214066	0	0	0	0	0.993032	0	1	-360	360;
	5854	5405	0.005353	0.214066	0	0	0	0	0.993032	0	1	-360	360;
	5854	5204	0.005353	0.214066	0	0	0	0	0.993032	0	1	-360	360;
	5854	13434	0.005353	0.214066	0	0	0	0	0.993032	0	1	-360	360;
	5854	12317	0.005353	0.214066	0	0	0	0	0.993032	0	1	-360	360;
	2130	3164	0.039062	1.562012	0	0	0	0	1.021238	0	1	-360	360;
	5336	13433	0.003521	0.140801	0	0	0	0	1.005684	0	1	-360	360;
	5336	12325	0.003521	0.140801	0	0	0	0	1.005684	0	1	-360	360;
	5336	13439	0.003521	0.140801	0	0	0	0	1.005684	0	1	-360	360;
	5336	12319	0.003521	0.140801	0	0	0	0	1.005684	0	1	-360	360;
	5336	13437	0.003521	0.140801	0	0	0	0	1.005684	0	1	-360	360;
	4575	13440	0.006127	0.245021	0	0	0	0	1.004608	0	1	-360	360;
	4575	12342	0.006127	0.245021	0	0	0	0	0.998741	0	1	-360	360;
	4575	6810	0.006127	0.245021	0	0	0	0	0.998741	0	1	-360	360;
	4575	6809	0.006127	0.245021	0	0	0	0	0.998741	0	1	-360	360;
	4575	6812	0.006127	0.245021	0	0	0	0	0.998741	0	1	-360	360;
	4575	6811	0.006127	0.245021	0	0	0	0	0.998741	0	1	-360	360;
	2959	10281	0.005208	0.208268	0	0	0	0	1.009947	0	1	-360	360;
	6867	6806	0.003521	0.140801	0	0	0	0	1.005617	0	1	-360	360;
	6867	6808	0.003521	0.140801	0	0	0	0	1.005617	0	1	-360	360;
	6867	6807	0.003521	0.140801	0	0	0	0	1.005617	0	1	-360	360;
	6867	6815	0.003521	0.140801	0	0	0	0	1.005617	0	1	-360	360;
	6867	6814	0.003521	0.140801	0	0	0	0	1.005617	0	1	-360	360;
	4503	6709	0.01269	0.507456	0	0	0	0	1.000098	0	1	-360	360;
	4503	8997	0.01269	0.507456	0	0	0	0	0.994564	0	1	-360	360;
	4503	3668	0.01269	0.507456	0	0	0	0	0.994564	0	1	-360	360;
	4503	6698	0.01269	0.507456	0	0	0	0	0.994564	0	1	-360	360;
	2537	3674	0.048077	1.922476	0	0	0	0	0.998226	0	1	-360	360;
	9105	3675	0.039062	1.562012	0	0	0	0	1.010126	0	1	-360	360;
	8469	8999	0.008993	0.3596	0	0	0	0	1.005209	0	1	-360	360;
	8469	3673	0.008993	0.3596	0	0	0	0	0.999399	0	1	-360	360;
	8469	3664	0.008993	0.3596	0	0	0	0	0.999399	0	1	-360	360;
	8469	3665	0.008993	0.3596	0	0	0	0	0.999399	0	1	-360	360;
	8469	10445	0.008993	0.3596	0	0	0	0	0.999399	0	1	-360	360;
	8469	10776	0.008993	0.3596	0	0	0	0	0.999399	0	1	-360	360;
	6519	13371	0.01004	0.401481	0	0	0	0	1.00113	0	1	-360	360;
	6519	10778	0.01004	0.401481	0	0	0	0	0.995531	0	1	-360	360;
	6519	10773	0.01004	0.401481	0	0	0	0	0.995531	0	1	-360	360;
	6519	10772	0.01004	0.401481	0	0	0	0	0.995531	0	1	-360	360;
	6519	13323	0.01004	0.401481	0	0	0	0	0.995531	0	1	-360	360;
	6835	6337	0.003521	0.140801	0	0	0	0	1.01161	0	1	-360	360;
	6835	6336	0.003521	0.140801	0	0	0	0	1.01161	0	1	-360	360;
	6835	6335	0.003521	0.140801	0	0	0	0	1.01161	0	1	-360	360;
	6835	10488	0.003521	0.140801	0	0	0	0	1.01161	0	1	-360	360;
	6835	10489	0.003521	0.140801	0	0	0	0	1.01161	0	1	-360	360;
	1472	6088	0.005208	0.208268	0	0	0	0	1.007725	0	1	-360	360;
	1472	5984	0.005208	0.208268	0	0	0	0	1.007725	0	1	-360	360;
	1472	11875	0.005208	0.208268	0	0	0	0	1.007725	0	1	-360	360;
	1472	4154	0.005208	0.208268	0	0	0	0	1.007725	0	1	-360	360;
	13287	8725	0.009025	0.360898	0	0	0	0	1.002848	0	1	-360	360;
	13287	10486	0.009025	0.360898	0	0	0	0	0.997224	0	1	-360	360;
	13287	10483	0.009025	0.360898	0	0	0	0	0.997224	0	1	-360	360;
	13287	10484	0.009025	0.360898	0	0	0	0	0.997224	0	1	-360	360;
	2418	12692	0.010917	0.436545	0	0	0	0	1.000577	0	1	-360	360;
	2418	12679	0.010917	0.436545	0	0	0	0	0.994983	0	1	-360	360;
	2418	11234	0.010917	0.436545	0	0	0	0	0.994983	0	1	-360	360;
	2418	13374	0.010917	0.436545	0	0	0	0	0.994983	0	1	-360	360;
	2418	12741	0.010917	0.436545	0	0	0	0	0.994983	0	1	-360	360;
	5435	11239	0.040323	1.612399	0	0	0	0	1.008391	0	1	-360	360;
	6898	12723	0.005208	0.208268	0	0	0	0	1.012835	0	1	-360	360;
	6898	11237	0.005208	0.208268	0	0	0	0	1.012835	0	1	-360	360;
	6898	11244	0.005208	0.208268	0	0	0	0	1.012835	0	1	-360	360;
	1285	3857	0.003521	0.140801	0	0	0	0	1.006375	0	1	-360	360;
	1285	11796	0.003521	0.140801	0	0	0	0	1.006375	0	1	-360	360;
	1285	1884	0.003521	0.140801	0	0	0	0	1.006375	0	1	-360	360;
	1285	1883	0.003521	0.140801	0	0	0	0	1.006375	0	1	-360	360;
	1285	1882	0.003521	0.140801	0	0	0	0	1.006375	0	1	-360	360;
	5438	1881	0.003521	0.140801	0	0	0	0	0.998845	0	1	-360	360;
	5438	1880	0.003521	0.140801	0	0	0	0	0.998845	0	1	-360	360;
	5438	1879	0.003521	0.140801	0	0	0	0	0.998845	0	1	-360	360;
	5438	1878	0.003521	0.140801	0	0	0	0	0.998845	0	1	-360	360;
	5438	1877	0.003521	0.140801	0	0	0	0	0.998845	0	1	-360	360;
	5438	1876	0.003521	0.140801	0	0	0	0	0.998845	0	1	-360	360;
	5438	8456	0.003521	0.140801	0	0	0	0	0.998845	0	1	-360	360;
	12743	12027	0.010121	0.404732	0	0	0	0	1.01165	0	1	-360	360;
	12743	1237	0.010121	0.404732	0	0	0	0	1.006209	0	1	-360	360;
	12743	5031	0.010121	0.404732	0	0	0	0	1.006209	0	1	-360	360;
	12743	12159	0.010121	0.404732	0	0	0	0	1.006209	0	1	-360	360;
	12743	4176	0.010121	0.404732	0	0	0	0	1.006209	0	1	-360	360;
	1973	12020	0.005208	0.208268	0	0	0	0	1.01334	0	1	-360	360;
	11168	12021	0.004744	0.189694	0	0	0	0	0.998715	0	1	-360	360;
	11168	12018	0.004744	0.189694	0	0	0	0	0.993096	0	1	-360	360;
	11168	3059	0.004744	0.189694	0	0	0	0	0.993096	0	1	-360	360;
	3615	2695	0.004441	0.177564	0	0	0	0	1.005871	0	1	-360	360;
	3615	8279	0.004441	0.177564	0	0	0	0	0.999436	0	1	-360	360;
	3615	2686	0.004441	0.177564	0	0	0	0	0.999436	0	1	-360	360;
	3615	4962	0.004441	0.177564	0	0	0	0	0.999436	0	1	-360	360;
	3615	4968	0.004441	0.177564	0	0	0	0	0.999436	0	1	-360	360;
	13295	4967	0.005208	0.208268	0	0	0	0	1.005792	0	1	-360	360;
	13295	2702	0.005208	0.208268	0	0	0	0	1.005792	0	1	-360	360;
	4876	2318	0.003521	0.140801	0	0	0	0	1.006969	0	1	-360	360;
	4876	2322	0.003521	0.140801	0	0	0	0	1.006969	0	1	-360	360;
	4876	2321	0.003521	0.140801	0	0	0	0	1.006969	0	1	-360	360;
	4876	8939	0.003521	0.140801	0	0	0	0	1.006969	0	1	-360	360;
	4876	4924	0.003521	0.140801	0	0	0	0	1.006969	0	1	-360	360;
	4876	8937	0.003521	0.140801	0	0	0	0	1.006969	0	1	-360	360;
	4876	8938	0.003521	0.140801	0	0	0	0	1.006969	0	1	-360	360;
	8986	8943	0.003521	0.140801	0	0	0	0	1.011449	0	1	-360	360;
	8986	8944	0.003521	0.140801	0	0	0	0	1.011449	0	1	-360	360;
	8986	8941	0.003521	0.140801	0	0	0	0	1.011449	0	1	-360	360;
	8986	13285	0.003521	0.140801	0	0	0	0	1.011449	0	1	-360	360;
	8883	8717	0.000708	0.02832	0	0	0	0	1.01898	0	1	-360	360;
	8883	8716	0.000708	0.02832	0	0	0	0	1.01898	0	1	-360	360;
	8883	2147	0.000708	0.02832	0	0	0	0	1.01898	0	1	-360	360;
	1235	6431	0.009881	0.395133	0	0	0	0	1.004044	0	1	-360	360;
	1235	6433	0.009881	0.395133	0	0	0	0	0.998264	0	1	-360	360;
	1235	6432	0.009881	0.395133	0	0	0	0	0.998264	0	1	-360	360;
	1235	11350	0.009881	0.395133	0	0	0	0	0.998264	0	1	-360	360;
	1235	11346	0.009881	0.395133	0	0	0	0	0.998264	0	1	-360	360;
	12350	11355	0.005208	0.208268	0	0	0	0	1.006096	0	1	-360	360;
	7258	2146	0.000708	0.02832	0	0	0	0	1.011449	0	1	-360	360;
	7258	2143	0.000708	0.02832	0	0	0	0	1.011449	0	1	-360	360;
	7258	2144	0.000708	0.02832	0	0	0	0	1.011449	0	1	-360	360;
	7258	1263	0.000708	0.02832	0	0	0	0	1.011449	0	1	-360	360;
	5915	3515	0.002119	0.084719	0	0	0	0	1.015759	0	1	-360	360;
	5915	3781	0.002119	0.084719	0	0	0	0	1.015759	0	1	-360	360;
	5915	3782	0.002119	0.084719	0	0	0	0	1.015759	0	1	-360	360;
	5915	3787	0.002119	0.084719	0	0	0	0	1.015759	0	1	-360	360;
	5915	7970	0.002119	0.084719	0	0	0	0	1.015759	0	1	-360	360;
	7790	13029	0.019084	0.76312	0	0	0	0	1.002299	0	1	-360	360;
	7790	13030	0.019084	0.76312	0	0	0	0	0.996763	0	1	-360	360;
	7790	13026	0.019084	0.76312	0	0	0	0	0.996763	0	1	-360	360;
	7552	13027	0.010593	0.423596	0	0	0	0	1.002662	0	1	-360	360;
	7552	5958	0.010593	0.423596	0	0	0	0	0.996851	0	1	-360	360;
	7552	5957	0.010593	0.423596	0	0	0	0	0.996851	0	1	-360	360;
	7552	5956	0.010593	0.423596	0	0	0	0	0.996851	0	1	-360	360;
	7552	5955	0.010593	0.423596	0	0	0	0	0.996851	0	1	-360	360;
	4803	10405	0.002119	0.084719	0	0	0	0	1.002302	0	1	-360	360;
	4803	10404	0.002119	0.084719	0	0	0	0	1.002302	0	1	-360	360;
	4803	8301	0.002119	0.084719	0	0	0	0	1.002302	0	1	-360	360;
	4803	10402	0.002119	0.084719	0	0	0	0	1.002302	0	1	-360	360;
	4803	10401	0.002119	0.084719	0	0	0	0	1.002302	0	1	-360	360;
	6993	5961	0.005208	0.208268	0	0	0	0	1.007029	0	1	-360	360;
	6993	660	0.005208	0.208268	0	0	0	0	1.007029	0	1	-360	360;
	6993	12521	0.005208	0.208268	0	0	0	0	1.007029	0	1	-360	360;
	6993	12522	0.005208	0.208268	0	0	0	0	1.007029	0	1	-360	360;
	6993	12523	0.005208	0.208268	0	0	0	0	1.007029	0	1	-360	360;
	650	7070	0.044643	1.785156	0	0	0	0	0.995676	0	1	-360	360;
	7271	7981	0.003521	0.140801	0	0	0	0	1.008678	0	1	-360	360;
	7271	7979	0.003521	0.140801	0	0	0	0	1.008678	0	1	-360	360;
	7271	7452	0.003521	0.140801	0	0	0	0	1.008678	0	1	-360	360;
	7271	7988	0.003521	0.140801	0	0	0	0	1.008678	0	1	-360	360;
	7271	7989	0.003521	0.140801	0	0	0	0	1.008678	0	1	-360	360;
	7271	69	0.003521	0.140801	0	0	0	0	1.008678	0	1	-360	360;
	7271	68	0.003521	0.140801	0	0	0	0	1.008678	0	1	-360	360;
	9031	7350	0.005208	0.208268	0	0	0	0	1.015263	0	1	-360	360;
	9031	7349	0.005208	0.208268	0	0	0	0	1.015263	0	1	-360	360;
	4865	7352	0.005208	0.208268	0	0	0	0	1.004302	0	1	-360	360;
	4865	6963	0.005208	0.208268	0	0	0	0	1.004302	0	1	-360	360;
	8177	3714	0.005208	0.208268	0	0	0	0	1.005823	0	1	-360	360;
	8177	3712	0.005208	0.208268	0	0	0	0	1.005823	0	1	-360	360;
	8177	7346	0.005208	0.208268	0	0	0	0	1.005823	0	1	-360	360;
	8489	7345	0.055556	2.221528	0	0	0	0	0.996402	0	1	-360	360;
	6995	336	0.012755	0.510045	0	0	0	0	1.007861	0	1	-360	360;
	6995	337	0.012755	0.510045	0	0	0	0	1.00231	0	1	-360	360;
	6995	334	0.012755	0.510045	0	0	0	0	1.00231	0	1	-360	360;
	6995	12562	0.012755	0.510045	0	0	0	0	1.00231	0	1	-360	360;
	2965	6372	0.000708	0.02832	0	0	0	0	1.009877	0	1	-360	360;
	2965	6371	0.000708	0.02832	0	0	0	0	1.009877	0	1	-360	360;
	2965	6370	0.000708	0.02832	0	0	0	0	1.009877	0	1	-360	360;
	2965	6369	0.000708	0.02832	0	0	0	0	1.009877	0	1	-360	360;
	1756	338	0.01269	0.507456	0	0	0	0	1.001925	0	1	-360	360;
	1756	339	0.01269	0.507456	0	0	0	0	0.996225	0	1	-360	360;
	1756	13538	0.01269	0.507456	0	0	0	0	0.996225	0	1	-360	360;
	1756	11887	0.01269	0.507456	0	0	0	0	0.996225	0	1	-360	360;
	3173	12863	0.039062	1.562012	0	0	0	0	1.008878	0	1	-360	360;
	7954	12864	0.011161	0.446289	0	0	0	0	1.001165	0	1	-360	360;
	7954	12858	0.011161	0.446289	0	0	0	0	0.995363	0	1	-360	360;
	7954	13535	0.011161	0.446289	0	0	0	0	0.995363	0	1	-360	360;
	7954	13532	0.011161	0.446289	0	0	0	0	0.995363	0	1	-360	360;
	7954	13533	0.011161	0.446289	0	0	0	0	0.995363	0	1	-360	360;
	1370	12847	0.018797	0.751645	0	0	0	0	1.004207	0	1	-360	360;
	1370	13531	0.018797	0.751645	0	0	0	0	0.998764	0	1	-360	360;
	1370	6892	0.018797	0.751645	0	0	0	0	0.998764	0	1	-360	360;
	12923	10222	0.006234	0.249299	0	0	0	0	1.009643	0	1	-360	360;
	12923	613	0.006234	0.249299	0	0	0	0	1.003323	0	1	-360	360;
	12923	6893	0.006234	0.249299	0	0	0	0	1.003323	0	1	-360	360;
	1156	8843	0.000531	0.021225	0	0	0	0	1.019287	0	1	-360	360;
	1156	8844	0.000531	0.021225	0	0	0	0	1.019287	0	1	-360	360;
	1156	8841	0.000531	0.021225	0	0	0	0	1.019287	0	1	-360	360;
	1156	8842	0.000531	0.021225	0	0	0	0	1.019287	0	1	-360	360;
	1156	12772	0.000531	0.021225	0	0	0	0	1.019287	0	1	-360	360;
	12143	4764	0.003521	0.140801	0	0	0	0	1.014355	0	1	-360	360;
	12143	165	0.003521	0.140801	0	0	0	0	1.014355	0	1	-360	360;
	12143	11805	0.003521	0.140801	0	0	0	0	1.014355	0	1	-360	360;
	12143	135	0.003521	0.140801	0	0	0	0	1.014355	0	1	-360	360;
	12143	2870	0.003521	0.140801	0	0	0	0	1.014355	0	1	-360	360;
	12143	114	0.003521	0.140801	0	0	0	0	1.014355	0	1	-360	360;
	3480	9897	0.005208	0.208268	0	0	0	0	1.004936	0	1	-360	360;
	3480	9900	0.005208	0.208268	0	0	0	0	1.004936	0	1	-360	360;
	5377	13032	0.012626	0.504893	0	0	0	0	1.00684	0	1	-360	360;
	5377	13024	0.012626	0.504893	0	0	0	0	1.00129	0	1	-360	360;
	5377	6179	0.012626	0.504893	0	0	0	0	1.00129	0	1	-360	360;
	5377	6430	0.012626	0.504893	0	0	0	0	1.00129	0	1	-360	360;
	7407	6429	0.015337	0.613305	0	0	0	0	1.018686	0	1	-360	360;
	7407	6428	0.015337	0.613305	0	0	0	0	1.013588	0	1	-360	360;
	7407	6427	0.015337	0.613305	0	0	0	0	1.013588	0	1	-360	360;
	7714	4345	0.003521	0.140801	0	0	0	0	1.008519	0	1	-360	360;
	7714	4343	0.003521	0.140801	0	0	0	0	1.008519	0	1	-360	360;
	7714	5262	0.003521	0.140801	0	0	0	0	1.008519	0	1	-360	360;
	7714	5261	0.003521	0.140801	0	0	0	0	1.008519	0	1	-360	360;
	7714	4314	0.003521	0.140801	0	0	0	0	1.008519	0	1	-360	360;
	7714	5252	0.003521	0.140801	0	0	0	0	1.008519	0	1	-360	360;
	7353	839	0.027174	1.086617	0	0	0	0	1.001829	0	1	-360	360;
	7353	840	0.027174	1.086617	0	0	0	0	0.996237	0	1	-360	360;
	6271	5323	0.002119	0.084719	0	0	0	0	1.009813	0	1	-360	360;
	6271	5324	0.002119	0.084719	0	0	0	0	1.009813	0	1	-360	360;
	6271	5327	0.002119	0.084719	0	0	0	0	1.009813	0	1	-360	360;
	6271	5328	0.002119	0.084719	0	0	0	0	1.009813	0	1	-360	360;
	6271	5325	0.002119	0.084719	0	0	0	0	1.009813	0	1	-360	360;
	6271	5326	0.002119	0.084719	0	0	0	0	1.009813	0	1	-360	360;
	1793	836	0.010684	0.427217	0	0	0	0	0.995406	0	1	-360	360;
	1793	837	0.010684	0.427217	0	0	0	0	0.989647	0	1	-360	360;
	1793	7824	0.010684	0.427217	0	0	0	0	0.989647	0	1	-360	360;
	1793	6665	0.010684	0.427217	0	0	0	0	0.989647	0	1	-360	360;
	1793	6673	0.010684	0.427217	0	0	0	0	0.989647	0	1	-360	360;
	7223	7825	0.010121	0.404732	0	0	0	0	1.011129	0	1	-360	360;
	7223	7820	0.010121	0.404732	0	0	0	0	1.005661	0	1	-360	360;
	7223	7819	0.010121	0.404732	0	0	0	0	1.005661	0	1	-360	360;
	7223	7822	0.010121	0.404732	0	0	0	0	1.005661	0	1	-360	360;
	7223	7821	0.010121	0.404732	0	0	0	0	1.005661	0	1	-360	360;
	11657	7476	0.000708	0.02832	0	0	0	0	1.009608	0	1	-360	360;
	11657	7477	0.000708	0.02832	0	0	0	0	1.009608	0	1	-360	360;
	11657	1401	0.000708	0.02832	0	0	0	0	1.009608	0	1	-360	360;
	9729	3997	0.002119	0.084719	0	0	0	0	1.010514	0	1	-360	360;
	9729	5163	0.002119	0.084719	0	0	0	0	1.010514	0	1	-360	360;
	9729	5159	0.002119	0.084719	0	0	0	0	1.010514	0	1	-360	360;
	9729	2198	0.002119	0.084719	0	0	0	0	1.010514	0	1	-360	360;
	9729	9862	0.002119	0.084719	0	0	0	0	1.010514	0	1	-360	360;
	9729	12915	0.002119	0.084719	0	0	0	0	1.010514	0	1	-360	360;
	9729	5143	0.002119	0.084719	0	0	0	0	1.010514	0	1	-360	360;
	9729	9870	0.002119	0.084719	0	0	0	0	1.010514	0	1	-360	360;
	10227	6141	0.000212	0.008494	0	0	0	0	1.033809	0	1	-360	360;
	10227	11909	0.000212	0.008494	0	0	0	0	1.033809	0	1	-360	360;
	9285	3013	0.000708	0.02832	0	0	0	0	1.023002	0	1	-360	360;
	1248	3015	0.000708	0.02832	0	0	0	0	1.018748	0	1	-360	360;
	9904	5359	0.014535	0.581214	0	0	0	0	1.029865	0	1	-360	360;
	9904	3011	0.014535	0.581214	0	0	0	0	1.023462	0	1	-360	360;
	9845	12986	0.005208	0.208268	0	0	0	0	1.005575	0	1	-360	360;
	9845	12984	0.005208	0.208268	0	0	0	0	1.005575	0	1	-360	360;
	7816	3012	0.000708	0.02832	0	0	0	0	1.017952	0	1	-360	360;
	2626	12993	0.005208	0.208268	0	0	0	0	0.988287	0	1	-360	360;
	2626	12994	0.005208	0.208268	0	0	0	0	0.988287	0	1	-360	360;
	2626	4926	0.005208	0.208268	0	0	0	0	0.988287	0	1	-360	360;
	1054	5060	0.000531	0.021225	0	0	0	0	1.071275	0	1	-360	360;
	1054	5059	0.000531	0.021225	0	0	0	0	1.071275	0	1	-360	360;
	8812	5058	0.000531	0.021225	0	0	0	0	0.995514	0	1	-360	360;
	8812	5065	0.000531	0.021225	0	0	0	0	0.995514	0	1	-360	360;
	9419	5064	0.000531	0.021225	0	0	0	0	1.004959	0	1	-360	360;
	9419	5063	0.000531	0.021225	0	0	0	0	1.004959	0	1	-360	360;
	9419	5062	0.000531	0.021225	0	0	0	0	1.004959	0	1	-360	360;
	929	4932	0.005208	0.208268	0	0	0	0	1.013798	0	1	-360	360;
	929	4931	0.005208	0.208268	0	0	0	0	1.013798	0	1	-360	360;
	929	1079	0.005208	0.208268	0	0	0	0	1.013798	0	1	-360	360;
	7355	12084	0.002119	0.084719	0	0	0	0	1.031633	0	1	-360	360;
	7355	12085	0.002119	0.084719	0	0	0	0	1.031633	0	1	-360	360;
	4754	1090	0.005208	0.208268	0	0	0	0	1.02899	0	1	-360	360;
	4754	11471	0.005208	0.208268	0	0	0	0	1.02899	0	1	-360	360;
	12640	1073	0.005208	0.208268	0	0	0	0	1.005352	0	1	-360	360;
	10889	6865	0.005208	0.208268	0	0	0	0	1.006368	0	1	-360	360;
	10889	11474	0.005208	0.208268	0	0	0	0	1.006368	0	1	-360	360;
	7239	11481	0.005208	0.208268	0	0	0	0	0.996372	0	1	-360	360;
	5893	11482	0.005208	0.208268	0	0	0	0	1.019827	0	1	-360	360;
	5893	8719	0.005208	0.208268	0	0	0	0	1.019827	0	1	-360	360;
	5893	3625	0.005208	0.208268	0	0	0	0	1.019827	0	1	-360	360;
	3390	6162	0.002119	0.084719	0	0	0	0	1.026009	0	1	-360	360;
	3390	6161	0.002119	0.084719	0	0	0	0	1.026009	0	1	-360	360;
	13014	3621	0.006887	0.275396	0	0	0	0	0.990478	0	1	-360	360;
	13014	3622	0.006887	0.275396	0	0	0	0	0.98327	0	1	-360	360;
	11341	3619	0.005208	0.208268	0	0	0	0	1.021717	0	1	-360	360;
	11341	3620	0.005208	0.208268	0	0	0	0	1.021717	0	1	-360	360;
	4230	11333	0.000708	0.02832	0	0	0	0	1.061078	0	1	-360	360;
	4230	11332	0.000708	0.02832	0	0	0	0	1.061078	0	1	-360	360;
	2724	4253	0.000708	0.02832	0	0	0	0	1.061049	0	1	-360	360;
	2724	4254	0.000708	0.02832	0	0	0	0	1.061049	0	1	-360	360;
	1580	12764	0.002119	0.084719	0	0	0	0	1.014526	0	1	-360	360;
	1580	12765	0.002119	0.084719	0	0	0	0	1.014526	0	1	-360	360;
	5506	10720	0.005208	0.208268	0	0	0	0	1.011624	0	1	-360	360;
	5506	10719	0.005208	0.208268	0	0	0	0	1.011624	0	1	-360	360;
	10725	10722	0.005208	0.208268	0	0	0	0	1.017259	0	1	-360	360;
	10725	10721	0.005208	0.208268	0	0	0	0	1.017259	0	1	-360	360;
	10725	10713	0.005208	0.208268	0	0	0	0	1.017259	0	1	-360	360;
	3950	10712	0.005208	0.208268	0	0	0	0	1.000028	0	1	-360	360;
	3950	3107	0.005208	0.208268	0	0	0	0	1.000028	0	1	-360	360;
	3950	3108	0.005208	0.208268	0	0	0	0	1.000028	0	1	-360	360;
	269	3109	0.005208	0.208268	0	0	0	0	0.999904	0	1	-360	360;
	269	5931	0.005208	0.208268	0	0	0	0	0.999904	0	1	-360	360;
	269	3111	0.005208	0.208268	0	0	0	0	0.999904	0	1	-360	360;
	5113	3112	0.005208	0.208268	0	0	0	0	0.997275	0	1	-360	360;
	5917	3113	0.005208	0.208268	0	0	0	0	1.005853	0	1	-360	360;
	3069	5695	0.002119	0.084719	0	0	0	0	1.009328	0	1	-360	360;
	3069	4393	0.002119	0.084719	0	0	0	0	1.009328	0	1	-360	360;
	8790	3114	0.005208	0.208268	0	0	0	0	1.009286	0	1	-360	360;
	8790	7291	0.005208	0.208268	0	0	0	0	1.009286	0	1	-360	360;
	8790	10236	0.005208	0.208268	0	0	0	0	1.009286	0	1	-360	360;
	349	10378	0.000531	0.021225	0	0	0	0	1.018538	0	1	-360	360;
	349	10377	0.000531	0.021225	0	0	0	0	1.018538	0	1	-360	360;
	9516	5322	0.005208	0.208268	0	0	0	0	1.018339	0	1	-360	360;
	9516	10238	0.005208	0.208268	0	0	0	0	1.018339	0	1	-360	360;
	13276	10375	0.000531	0.021225	0	0	0	0	1.019667	0	1	-360	360;
	13276	10374	0.000531	0.021225	0	0	0	0	1.019667	0	1	-360	360;
	355	10242	0.005208	0.208268	0	0	0	0	1.022088	0	1	-360	360;
	355	10241	0.005208	0.208268	0	0	0	0	1.022088	0	1	-360	360;
	7941	4618	0.005208	0.208268	0	0	0	0	1.020489	0	1	-360	360;
	7941	4619	0.005208	0.208268	0	0	0	0	1.020489	0	1	-360	360;
	3032	855	0.005208	0.208268	0	0	0	0	1.014053	0	1	-360	360;
	3060	4617	0.005208	0.208268	0	0	0	0	1.001991	0	1	-360	360;
	5203	5287	0.002119	0.084719	0	0	0	0	1.01887	0	1	-360	360;
	5203	5286	0.002119	0.084719	0	0	0	0	1.01887	0	1	-360	360;
	12735	2389	0.005208	0.208268	0	0	0	0	1.014122	0	1	-360	360;
	12735	405	0.005208	0.208268	0	0	0	0	1.014122	0	1	-360	360;
	12735	406	0.005208	0.208268	0	0	0	0	1.014122	0	1	-360	360;
	12339	6406	0.000531	0.021225	0	0	0	0	1.036689	0	1	-360	360;
	11063	7372	0.005208	0.208268	0	0	0	0	1.014323	0	1	-360	360;
	11063	11529	0.005208	0.208268	0	0	0	0	1.014323	0	1	-360	360;
	11063	7370	0.005208	0.208268	0	0	0	0	1.014323	0	1	-360	360;
	7310	7369	0.005208	0.208268	0	0	0	0	1.016047	0	1	-360	360;
	7310	7376	0.005208	0.208268	0	0	0	0	1.016047	0	1	-360	360;
	7310	7375	0.005208	0.208268	0	0	0	0	1.016047	0	1	-360	360;
	7310	7374	0.005208	0.208268	0	0	0	0	1.016047	0	1	-360	360;
	6124	3838	0.005208	0.208268	0	0	0	0	1.010237	0	1	-360	360;
	8697	7379	0.005208	0.208268	0	0	0	0	1.017221	0	1	-360	360;
	8697	7378	0.005208	0.208268	0	0	0	0	1.017221	0	1	-360	360;
	3659	3934	0.006812	0.272394	0	0	0	0	1.006012	0	1	-360	360;
	3659	867	0.006812	0.272394	0	0	0	0	1.000062	0	1	-360	360;
	7708	865	0.005208	0.208268	0	0	0	0	1.020068	0	1	-360	360;
	7708	5385	0.005208	0.208268	0	0	0	0	1.020068	0	1	-360	360;
	7708	864	0.005208	0.208268	0	0	0	0	1.020068	0	1	-360	360;
	7708	861	0.005208	0.208268	0	0	0	0	1.020068	0	1	-360	360;
	7708	862	0.005208	0.208268	0	0	0	0	1.020068	0	1	-360	360;
	4311	4452	0.000531	0.021225	0	0	0	0	1.036461	0	1	-360	360;
	9193	874	0.005208	0.208268	0	0	0	0	1.014398	0	1	-360	360;
	9193	10363	0.005208	0.208268	0	0	0	0	1.014398	0	1	-360	360;
	9193	10348	0.005208	0.208268	0	0	0	0	1.014398	0	1	-360	360;
	9193	7847	0.005208	0.208268	0	0	0	0	1.014398	0	1	-360	360;
	3759	6774	0.005208	0.208268	0	0	0	0	1.007294	0	1	-360	360;
	3759	6785	0.005208	0.208268	0	0	0	0	1.007294	0	1	-360	360;
	3759	7848	0.005208	0.208268	0	0	0	0	1.007294	0	1	-360	360;
	3759	7851	0.005208	0.208268	0	0	0	0	1.007294	0	1	-360	360;
	3759	6789	0.005208	0.208268	0	0	0	0	1.007294	0	1	-360	360;
	3759	7842	0.005208	0.208268	0	0	0	0	1.007294	0	1	-360	360;
	13194	7841	0.005208	0.208268	0	0	0	0	1.018581	0	1	-360	360;
	13194	10034	0.005208	0.208268	0	0	0	0	1.018581	0	1	-360	360;
	1616	13052	0.005208	0.208268	0	0	0	0	1.013895	0	1	-360	360;
	1616	13053	0.005208	0.208268	0	0	0	0	1.013895	0	1	-360	360;
	1616	13054	0.005208	0.208268	0	0	0	0	1.013895	0	1	-360	360;
	1616	13047	0.005208	0.208268	0	0	0	0	1.013895	0	1	-360	360;
	1616	13048	0.005208	0.208268	0	0	0	0	1.013895	0	1	-360	360;
	1616	13049	0.005208	0.208268	0	0	0	0	1.013895	0	1	-360	360;
	1375	5352	0.005208	0.208268	0	0	0	0	1.012511	0	1	-360	360;
	1375	13057	0.005208	0.208268	0	0	0	0	1.012511	0	1	-360	360;
	1375	13058	0.005208	0.208268	0	0	0	0	1.012511	0	1	-360	360;
	12377	6456	0.005208	0.208268	0	0	0	0	1.017604	0	1	-360	360;
	12377	6455	0.005208	0.208268	0	0	0	0	1.017604	0	1	-360	360;
	4711	4197	0.000531	0.021225	0	0	0	0	1.001876	0	1	-360	360;
	13164	4195	0.000531	0.021225	0	0	0	0	1.03326	0	1	-360	360;
	13267	6452	0.005208	0.208268	0	0	0	0	1.016595	0	1	-360	360;
	303	6451	0.005208	0.208268	0	0	0	0	1.008958	0	1	-360	360;
	10454	2974	0.005208	0.208268	0	0	0	0	1.027567	0	1	-360	360;
	10454	2973	0.005208	0.208268	0	0	0	0	1.027567	0	1	-360	360;
	10454	2969	0.005208	0.208268	0	0	0	0	1.027567	0	1	-360	360;
	10454	2968	0.005208	0.208268	0	0	0	0	1.027567	0	1	-360	360;
	10454	9880	0.005208	0.208268	0	0	0	0	1.027567	0	1	-360	360;
	3439	8373	0.005208	0.208268	0	0	0	0	1.027993	0	1	-360	360;
	3439	9885	0.005208	0.208268	0	0	0	0	1.027993	0	1	-360	360;
	3439	10410	0.005208	0.208268	0	0	0	0	1.027993	0	1	-360	360;
	3439	8378	0.005208	0.208268	0	0	0	0	1.027993	0	1	-360	360;
	3439	9887	0.005208	0.208268	0	0	0	0	1.027993	0	1	-360	360;
	5735	8380	0.005208	0.208268	0	0	0	0	1.020775	0	1	-360	360;
	5735	10096	0.005208	0.208268	0	0	0	0	1.020775	0	1	-360	360;
	5735	8369	0.005208	0.208268	0	0	0	0	1.020775	0	1	-360	360;
	7065	6262	0.000531	0.021225	0	0	0	0	1.036199	0	1	-360	360;
	7065	12859	0.000531	0.021225	0	0	0	0	1.036199	0	1	-360	360;
	7065	8921	0.000531	0.021225	0	0	0	0	1.036199	0	1	-360	360;
	11134	1342	0.005208	0.208268	0	0	0	0	1.021621	0	1	-360	360;
	11134	1343	0.005208	0.208268	0	0	0	0	1.021621	0	1	-360	360;
	11134	8457	0.005208	0.208268	0	0	0	0	1.021621	0	1	-360	360;
	11134	8459	0.005208	0.208268	0	0	0	0	1.021621	0	1	-360	360;
	11134	1338	0.005208	0.208268	0	0	0	0	1.021621	0	1	-360	360;
	11134	1339	0.005208	0.208268	0	0	0	0	1.021621	0	1	-360	360;
	11134	1348	0.005208	0.208268	0	0	0	0	1.021621	0	1	-360	360;
	7607	1349	0.005208	0.208268	0	0	0	0	0.986347	0	1	-360	360;
	4295	7902	0.005208	0.208268	0	0	0	0	0.999582	0	1	-360	360;
	4295	7901	0.005208	0.208268	0	0	0	0	0.999582	0	1	-360	360;
	12399	7056	0.005208	0.208268	0	0	0	0	1.013814	0	1	-360	360;
	9688	7899	0.005208	0.208268	0	0	0	0	1.00918	0	1	-360	360;
	11906	7075	0.005208	0.208268	0	0	0	0	1.021759	0	1	-360	360;
	11906	7905	0.005208	0.208268	0	0	0	0	1.021759	0	1	-360	360;
	12502	7904	0.005208	0.208268	0	0	0	0	1.00786	0	1	-360	360;
	12502	7903	0.005208	0.208268	0	0	0	0	1.00786	0	1	-360	360;
	12502	7908	0.005208	0.208268	0	0	0	0	1.00786	0	1	-360	360;
	13559	2345	0.000708	0.02832	0	0	0	0	1.03506	0	1	-360	360;
	12405	9439	0.000708	0.02832	0	0	0	0	1.035095	0	1	-360	360;
	12493	12346	0.000531	0.021225	0	0	0	0	1.045702	0	1	-360	360;
	12493	485	0.000531	0.021225	0	0	0	0	1.045702	0	1	-360	360;
	6525	12348	0.000531	0.021225	0	0	0	0	1.04939	0	1	-360	360;
	1635	4501	0.005208	0.208268	0	0	0	0	1.015794	0	1	-360	360;
	1635	4512	0.005208	0.208268	0	0	0	0	1.015794	0	1	-360	360;
	1635	904	0.005208	0.208268	0	0	0	0	1.015794	0	1	-360	360;
	1635	5668	0.005208	0.208268	0	0	0	0	1.015794	0	1	-360	360;
	1635	5670	0.005208	0.208268	0	0	0	0	1.015794	0	1	-360	360;
	1635	907	0.005208	0.208268	0	0	0	0	1.015794	0	1	-360	360;
	6991	9350	0.005208	0.208268	0	0	0	0	1.015729	0	1	-360	360;
	6991	9349	0.005208	0.208268	0	0	0	0	1.015729	0	1	-360	360;
	6991	9352	0.005208	0.208268	0	0	0	0	1.015729	0	1	-360	360;
	6991	9351	0.005208	0.208268	0	0	0	0	1.015729	0	1	-360	360;
	6991	9346	0.005208	0.208268	0	0	0	0	1.015729	0	1	-360	360;
	6991	9345	0.005208	0.208268	0	0	0	0	1.015729	0	1	-360	360;
	6991	9348	0.005208	0.208268	0	0	0	0	1.015729	0	1	-360	360;
	6991	9347	0.005208	0.208268	0	0	0	0	1.015729	0	1	-360	360;
	10515	4612	0.000531	0.021225	0	0	0	0	1.021677	0	1	-360	360;
	6745	5311	0.000531	0.021225	0	0	0	0	1.048652	0	1	-360	360;
	6745	11853	0.000531	0.021225	0	0	0	0	1.048652	0	1	-360	360;
	12832	11854	0.000531	0.021225	0	0	0	0	1.049636	0	1	-360	360;
	12832	11851	0.000531	0.021225	0	0	0	0	1.049636	0	1	-360	360;
	13376	3152	0.000531	0.021225	0	0	0	0	1.050924	0	1	-360	360;
	3689	2249	0.005208	0.208268	0	0	0	0	0.99761	0	1	-360	360;
	3689	610	0.005208	0.208268	0	0	0	0	0.99761	0	1	-360	360;
	3689	2247	0.005208	0.208268	0	0	0	0	0.99761	0	1	-360	360;
	3689	2248	0.005208	0.208268	0	0	0	0	0.99761	0	1	-360	360;
	3689	2252	0.005208	0.208268	0	0	0	0	0.99761	0	1	-360	360;
	5511	2253	0.005208	0.208268	0	0	0	0	1.009317	0	1	-360	360;
	5511	12856	0.005208	0.208268	0	0	0	0	1.009317	0	1	-360	360;
	5511	8867	0.005208	0.208268	0	0	0	0	1.009317	0	1	-360	360;
	5511	8866	0.005208	0.208268	0	0	0	0	1.009317	0	1	-360	360;
	5511	9914	0.005208	0.208268	0	0	0	0	1.009317	0	1	-360	360;
	5511	8864	0.005208	0.208268	0	0	0	0	1.009317	0	1	-360	360;
	5511	12843	0.005208	0.208268	0	0	0	0	1.009317	0	1	-360	360;
	5511	12840	0.005208	0.208268	0	0	0	0	1.009317	0	1	-360	360;
	11269	8861	0.005208	0.208268	0	0	0	0	1.000207	0	1	-360	360;
	11269	8860	0.005208	0.208268	0	0	0	0	1.000207	0	1	-360	360;
	9582	8167	0.000531	0.021225	0	0	0	0	1.050391	0	1	-360	360;
	9582	11360	0.000531	0.021225	0	0	0	0	1.050391	0	1	-360	360;
	3212	11361	0.000531	0.021225	0	0	0	0	1.050989	0	1	-360	360;
	3212	11362	0.000531	0.021225	0	0	0	0	1.050989	0	1	-360	360;
	6069	11363	0.000531	0.021225	0	0	0	0	1.018821	0	1	-360	360;
	9994	1805	0.005208	0.208268	0	0	0	0	1.000404	0	1	-360	360;
	9994	11336	0.005208	0.208268	0	0	0	0	1.000404	0	1	-360	360;
	9994	1807	0.005208	0.208268	0	0	0	0	1.000404	0	1	-360	360;
	7863	1808	0.005208	0.208268	0	0	0	0	1.001243	0	1	-360	360;
	7863	1803	0.005208	0.208268	0	0	0	0	1.001243	0	1	-360	360;
	10552	11327	0.005208	0.208268	0	0	0	0	0.999161	0	1	-360	360;
	10552	6542	0.005208	0.208268	0	0	0	0	0.999161	0	1	-360	360;
	10552	6541	0.005208	0.208268	0	0	0	0	0.999161	0	1	-360	360;
	10552	12106	0.005208	0.208268	0	0	0	0	0.999161	0	1	-360	360;
	10552	6543	0.005208	0.208268	0	0	0	0	0.999161	0	1	-360	360;
	9298	12904	0.005208	0.208268	0	0	0	0	0.988425	0	1	-360	360;
	9298	6545	0.005208	0.208268	0	0	0	0	0.988425	0	1	-360	360;
	5046	3325	0.005208	0.208268	0	0	0	0	1.002306	0	1	-360	360;
	5046	3326	0.005208	0.208268	0	0	0	0	1.002306	0	1	-360	360;
	4215	3327	0.005208	0.208268	0	0	0	0	1.002332	0	1	-360	360;
	4215	3328	0.005208	0.208268	0	0	0	0	1.002332	0	1	-360	360;
	2033	8445	0.005208	0.208268	0	0	0	0	1.002303	0	1	-360	360;
	2033	10422	0.005208	0.208268	0	0	0	0	1.002303	0	1	-360	360;
	2634	8441	0.013021	0.520671	0	0	0	0	1.002567	0	1	-360	360;
	2634	10420	0.013021	0.520671	0	0	0	0	0.997547	0	1	-360	360;
	3475	10427	0.005376	0.214987	0	0	0	0	1.008205	0	1	-360	360;
	3475	10426	0.005376	0.214987	0	0	0	0	1.008205	0	1	-360	360;
	3475	10425	0.005376	0.214987	0	0	0	0	1.002391	0	1	-360	360;
	3475	10424	0.005376	0.214987	0	0	0	0	1.002391	0	1	-360	360;
	3475	10430	0.005376	0.214987	0	0	0	0	1.002391	0	1	-360	360;
	3475	10429	0.005376	0.214987	0	0	0	0	1.002391	0	1	-360	360;
	3475	4546	0.005376	0.214987	0	0	0	0	1.002391	0	1	-360	360;
	3475	4545	0.005376	0.214987	0	0	0	0	1.002391	0	1	-360	360;
	606	4548	0.005208	0.208268	0	0	0	0	1.002921	0	1	-360	360;
	606	4547	0.005208	0.208268	0	0	0	0	1.002921	0	1	-360	360;
	606	4550	0.005208	0.208268	0	0	0	0	1.002921	0	1	-360	360;
	606	4549	0.005208	0.208268	0	0	0	0	1.002921	0	1	-360	360;
	606	4552	0.005208	0.208268	0	0	0	0	1.002921	0	1	-360	360;
	606	4551	0.005208	0.208268	0	0	0	0	1.002921	0	1	-360	360;
	606	4554	0.005208	0.208268	0	0	0	0	1.002921	0	1	-360	360;
	4667	4553	0.005208	0.208268	0	0	0	0	0.994464	0	1	-360	360;
	4667	11110	0.005208	0.208268	0	0	0	0	0.994464	0	1	-360	360;
	13517	11111	0.005208	0.208268	0	0	0	0	1.004859	0	1	-360	360;
	13517	11108	0.005208	0.208268	0	0	0	0	1.004859	0	1	-360	360;
	13517	11109	0.005208	0.208268	0	0	0	0	1.004859	0	1	-360	360;
	13517	11106	0.005208	0.208268	0	0	0	0	1.004859	0	1	-360	360;
	9749	11107	0.005208	0.208268	0	0	0	0	0.995889	0	1	-360	360;
	9749	11104	0.005208	0.208268	0	0	0	0	0.995889	0	1	-360	360;
	9749	12747	0.005208	0.208268	0	0	0	0	0.995889	0	1	-360	360;
	9749	11102	0.005208	0.208268	0	0	0	0	0.995889	0	1	-360	360;
	9749	12742	0.005208	0.208268	0	0	0	0	0.995889	0	1	-360	360;
	7207	4025	0.005208	0.208268	0	0	0	0	1.01276	0	1	-360	360;
	7207	4024	0.005208	0.208268	0	0	0	0	1.01276	0	1	-360	360;
	12238	11229	0.005208	0.208268	0	0	0	0	1.010163	0	1	-360	360;
	3251	4022	0.005241	0.209578	0	0	0	0	1.012049	0	1	-360	360;
	3251	4029	0.005241	0.209578	0	0	0	0	1.006682	0	1	-360	360;
	3251	4028	0.005241	0.209578	0	0	0	0	1.006682	0	1	-360	360;
	6613	4027	0.005208	0.208268	0	0	0	0	1.00475	0	1	-360	360;
	6613	4026	0.005208	0.208268	0	0	0	0	1.00475	0	1	-360	360;
	5839	4021	0.005208	0.208268	0	0	0	0	1.00953	0	1	-360	360;
	5839	4020	0.005208	0.208268	0	0	0	0	1.00953	0	1	-360	360;
	5839	10653	0.005208	0.208268	0	0	0	0	1.00953	0	1	-360	360;
	9998	11289	0.005208	0.208268	0	0	0	0	1.00503	0	1	-360	360;
	9998	11309	0.005208	0.208268	0	0	0	0	1.00503	0	1	-360	360;
	9998	10656	0.005208	0.208268	0	0	0	0	1.00503	0	1	-360	360;
	8287	10657	0.004587	0.183429	0	0	0	0	1.001503	0	1	-360	360;
	8287	10658	0.004587	0.183429	0	0	0	0	1.001503	0	1	-360	360;
	8287	10659	0.004587	0.183429	0	0	0	0	0.995758	0	1	-360	360;
	5786	10660	0.005208	0.208268	0	0	0	0	1.01197	0	1	-360	360;
	5786	10648	0.005208	0.208268	0	0	0	0	1.01197	0	1	-360	360;
	5786	10649	0.005208	0.208268	0	0	0	0	1.01197	0	1	-360	360;
	2291	5612	0.005208	0.208268	0	0	0	0	0.999252	0	1	-360	360;
	2291	5611	0.005208	0.208268	0	0	0	0	0.999252	0	1	-360	360;
	4078	5614	0.005208	0.208268	0	0	0	0	0.999206	0	1	-360	360;
	1919	5613	0.005208	0.208268	0	0	0	0	0.99686	0	1	-360	360;
	1919	5608	0.005208	0.208268	0	0	0	0	0.99686	0	1	-360	360;
	3391	5607	0.005981	0.23916	0	0	0	0	1.013491	0	1	-360	360;
	3391	5610	0.005981	0.23916	0	0	0	0	1.013491	0	1	-360	360;
	3391	5609	0.005981	0.23916	0	0	0	0	1.008757	0	1	-360	360;
	3391	5606	0.005981	0.23916	0	0	0	0	1.008757	0	1	-360	360;
	3391	5605	0.005981	0.23916	0	0	0	0	1.008757	0	1	-360	360;
	12003	12168	0.005208	0.208268	0	0	0	0	1.014777	0	1	-360	360;
	12003	12169	0.005208	0.208268	0	0	0	0	1.014777	0	1	-360	360;
	12003	12166	0.005208	0.208268	0	0	0	0	1.014777	0	1	-360	360;
	12003	12167	0.005208	0.208268	0	0	0	0	1.014777	0	1	-360	360;
	12003	12172	0.005208	0.208268	0	0	0	0	1.014777	0	1	-360	360;
	13566	11203	0.005208	0.208268	0	0	0	0	0.980347	0	1	-360	360;
	13566	12170	0.005208	0.208268	0	0	0	0	0.980347	0	1	-360	360;
	13566	12171	0.005208	0.208268	0	0	0	0	0.980347	0	1	-360	360;
	13566	12163	0.005208	0.208268	0	0	0	0	0.980347	0	1	-360	360;
	13566	12164	0.005208	0.208268	0	0	0	0	0.980347	0	1	-360	360;
	13566	5057	0.005208	0.208268	0	0	0	0	0.980347	0	1	-360	360;
	13566	5056	0.005208	0.208268	0	0	0	0	0.980347	0	1	-360	360;
	13566	5055	0.005208	0.208268	0	0	0	0	0.980347	0	1	-360	360;
	13566	5054	0.005208	0.208268	0	0	0	0	0.980347	0	1	-360	360;
	3802	5053	0.005208	0.208268	0	0	0	0	1.009707	0	1	-360	360;
	3802	5052	0.005208	0.208268	0	0	0	0	1.009707	0	1	-360	360;
	3802	5051	0.005208	0.208268	0	0	0	0	1.009707	0	1	-360	360;
	13177	5050	0.012019	0.480619	0	0	0	0	0.987915	0	1	-360	360;
	13177	3316	0.012019	0.480619	0	0	0	0	0.982064	0	1	-360	360;
	13177	3311	0.012019	0.480619	0	0	0	0	0.982064	0	1	-360	360;
	13177	11621	0.012019	0.480619	0	0	0	0	0.982064	0	1	-360	360;
	13177	11622	0.012019	0.480619	0	0	0	0	0.982064	0	1	-360	360;
	11517	11623	0.005208	0.208268	0	0	0	0	0.994434	0	1	-360	360;
	11517	11624	0.005208	0.208268	0	0	0	0	0.994434	0	1	-360	360;
	11517	11617	0.005208	0.208268	0	0	0	0	0.994434	0	1	-360	360;
	11517	10151	0.005208	0.208268	0	0	0	0	0.994434	0	1	-360	360;
	11517	10164	0.005208	0.208268	0	0	0	0	0.994434	0	1	-360	360;
	8047	10178	0.005208	0.208268	0	0	0	0	1.008507	0	1	-360	360;
	8047	11625	0.005208	0.208268	0	0	0	0	1.008507	0	1	-360	360;
	8047	11626	0.005208	0.208268	0	0	0	0	1.008507	0	1	-360	360;
	8047	6567	0.005208	0.208268	0	0	0	0	1.008507	0	1	-360	360;
	8047	6566	0.005208	0.208268	0	0	0	0	1.008507	0	1	-360	360;
	8521	12070	0.005208	0.208268	0	0	0	0	1.004493	0	1	-360	360;
	8521	12069	0.005208	0.208268	0	0	0	0	1.004493	0	1	-360	360;
	8521	4449	0.005208	0.208268	0	0	0	0	1.004493	0	1	-360	360;
	8521	12071	0.005208	0.208268	0	0	0	0	1.004493	0	1	-360	360;
	8521	12074	0.005208	0.208268	0	0	0	0	1.004493	0	1	-360	360;
	8521	12073	0.005208	0.208268	0	0	0	0	1.004493	0	1	-360	360;
	8521	12076	0.005208	0.208268	0	0	0	0	1.004493	0	1	-360	360;
	8521	12075	0.005208	0.208268	0	0	0	0	1.004493	0	1	-360	360;
	8521	5009	0.005208	0.208268	0	0	0	0	1.004493	0	1	-360	360;
	8521	2963	0.005208	0.208268	0	0	0	0	1.004493	0	1	-360	360;
	8521	5007	0.005208	0.208268	0	0	0	0	1.004493	0	1	-360	360;
	8521	5008	0.005208	0.208268	0	0	0	0	1.004493	0	1	-360	360;
	4638	3763	0.003521	0.140801	0	0	0	0	1.023046	0	1	-360	360;
	4638	3297	0.003521	0.140801	0	0	0	0	1.023046	0	1	-360	360;
	4638	3298	0.003521	0.140801	0	0	0	0	1.023046	0	1	-360	360;
	4638	3299	0.003521	0.140801	0	0	0	0	1.023046	0	1	-360	360;
	4638	3306	0.003521	0.140801	0	0	0	0	1.023046	0	1	-360	360;
	4638	3307	0.003521	0.140801	0	0	0	0	1.023046	0	1	-360	360;
	4638	7844	0.003521	0.140801	0	0	0	0	1.023046	0	1	-360	360;
	4638	6773	0.003521	0.140801	0	0	0	0	1.023046	0	1	-360	360;
	4638	7846	0.003521	0.140801	0	0	0	0	1.023046	0	1	-360	360;
	7987	10558	0.000212	0.008494	0	0	0	0	1.013775	0	1	-360	360;
	7987	9152	0.000212	0.008494	0	0	0	0	1.013775	0	1	-360	360;
	7987	10555	0.000212	0.008494	0	0	0	0	1.013775	0	1	-360	360;
	7987	10556	0.000212	0.008494	0	0	0	0	1.013775	0	1	-360	360;
	9538	6398	0.005208	0.208268	0	0	0	0	1.023982	0	1	-360	360;
	9538	11227	0.005208	0.208268	0	0	0	0	1.023982	0	1	-360	360;
	9538	11231	0.005208	0.208268	0	0	0	0	1.023982	0	1	-360	360;
	9538	9974	0.005208	0.208268	0	0	0	0	1.023982	0	1	-360	360;
	9538	13500	0.005208	0.208268	0	0	0	0	1.023982	0	1	-360	360;
	9538	2858	0.005208	0.208268	0	0	0	0	1.023982	0	1	-360	360;
	9538	13502	0.005208	0.208268	0	0	0	0	1.023982	0	1	-360	360;
	11283	13505	0.005208	0.208268	0	0	0	0	1.023979	0	1	-360	360;
	11283	2871	0.005208	0.208268	0	0	0	0	1.023979	0	1	-360	360;
	11283	13507	0.005208	0.208268	0	0	0	0	1.023979	0	1	-360	360;
	11283	13506	0.005208	0.208268	0	0	0	0	1.023979	0	1	-360	360;
	11283	13497	0.005208	0.208268	0	0	0	0	1.023979	0	1	-360	360;
	11283	13496	0.005208	0.208268	0	0	0	0	1.023979	0	1	-360	360;
	11283	5928	0.005208	0.208268	0	0	0	0	1.023979	0	1	-360	360;
	3572	5929	0.005208	0.208268	0	0	0	0	1.011158	0	1	-360	360;
	3572	5930	0.005208	0.208268	0	0	0	0	1.011158	0	1	-360	360;
	3572	10622	0.005208	0.208268	0	0	0	0	1.011158	0	1	-360	360;
	3572	5932	0.005208	0.208268	0	0	0	0	1.011158	0	1	-360	360;
	3572	5933	0.005208	0.208268	0	0	0	0	1.011158	0	1	-360	360;
	3572	5934	0.005208	0.208268	0	0	0	0	1.011158	0	1	-360	360;
	3572	5935	0.005208	0.208268	0	0	0	0	1.010863	0	1	-360	360;
	3572	5936	0.005208	0.208268	0	0	0	0	1.010863	0	1	-360	360;
	3572	8323	0.005208	0.208268	0	0	0	0	1.010863	0	1	-360	360;
	4605	13009	0.005208	0.208268	0	0	0	0	1.008667	0	1	-360	360;
	4605	13008	0.005208	0.208268	0	0	0	0	1.008667	0	1	-360	360;
	4605	13007	0.005208	0.208268	0	0	0	0	1.008667	0	1	-360	360;
	2696	11167	0.002119	0.084719	0	0	0	0	1.014632	0	1	-360	360;
	2696	10627	0.002119	0.084719	0	0	0	0	1.014632	0	1	-360	360;
	2696	11172	0.002119	0.084719	0	0	0	0	1.014632	0	1	-360	360;
	2696	13062	0.002119	0.084719	0	0	0	0	1.014632	0	1	-360	360;
	2696	13066	0.002119	0.084719	0	0	0	0	1.014632	0	1	-360	360;
	2696	11174	0.002119	0.084719	0	0	0	0	1.014632	0	1	-360	360;
	2696	11175	0.002119	0.084719	0	0	0	0	1.014632	0	1	-360	360;
	2696	3104	0.002119	0.084719	0	0	0	0	1.014632	0	1	-360	360;
	2696	3103	0.002119	0.084719	0	0	0	0	1.014632	0	1	-360	360;
	823	7282	0.029412	1.176103	0	0	0	0	0.997898	0	1	-360	360;
	823	7283	0.029412	1.176103	0	0	0	0	0.991389	0	1	-360	360;
	823	7286	0.029412	1.176103	0	0	0	0	0.991389	0	1	-360	360;
	823	7287	0.029412	1.176103	0	0	0	0	0.991389	0	1	-360	360;
	823	7284	0.029412	1.176103	0	0	0	0	0.991389	0	1	-360	360;
	823	3226	0.029412	1.176103	0	0	0	0	0.991389	0	1	-360	360;
	823	3242	0.029412	1.176103	0	0	0	0	0.991389	0	1	-360	360;
	823	3245	0.029412	1.176103	0	0	0	0	0.991389	0	1	-360	360;
	6350	773	0.005208	0.208268	0	0	0	0	0.99459	0	1	-360	360;
	6350	1791	0.005208	0.208268	0	0	0	0	0.99459	0	1	-360	360;
	9018	775	0.00774	0.309501	0	0	0	0	1.011102	0	1	-360	360;
	9018	774	0.00774	0.309501	0	0	0	0	1.005775	0	1	-360	360;
	6895	769	0.004433	0.17725	0	0	0	0	1.007293	0	1	-360	360;
	6895	768	0.004433	0.17725	0	0	0	0	1.007293	0	1	-360	360;
	6895	771	0.004433	0.17725	0	0	0	0	1.001638	0	1	-360	360;
	6895	770	0.004433	0.17725	0	0	0	0	1.001638	0	1	-360	360;
	6895	4781	0.004433	0.17725	0	0	0	0	1.001638	0	1	-360	360;
	6895	776	0.004433	0.17725	0	0	0	0	1.001638	0	1	-360	360;
	7515	10064	0.000212	0.008494	0	0	0	0	1.002116	0	1	-360	360;
	7515	10065	0.000212	0.008494	0	0	0	0	1.002116	0	1	-360	360;
	5150	4083	0.000708	0.02832	0	0	0	0	0.999658	0	1	-360	360;
	5150	11520	0.000708	0.02832	0	0	0	0	0.999658	0	1	-360	360;
	5150	11536	0.000708	0.02832	0	0	0	0	0.999658	0	1	-360	360;
	1499	4084	0.000708	0.02832	0	0	0	0	0.999761	0	1	-360	360;
	1499	4087	0.000708	0.02832	0	0	0	0	0.999761	0	1	-360	360;
	1499	11649	0.000708	0.02832	0	0	0	0	0.999761	0	1	-360	360;
	13477	10060	0.000212	0.008494	0	0	0	0	1.011119	0	1	-360	360;
	13477	6368	0.000212	0.008494	0	0	0	0	1.011119	0	1	-360	360;
	13477	3451	0.000212	0.008494	0	0	0	0	1.011119	0	1	-360	360;
	13477	7776	0.000212	0.008494	0	0	0	0	1.011119	0	1	-360	360;
	8249	312	0.005208	0.208268	0	0	0	0	1.016091	0	1	-360	360;
	8249	311	0.005208	0.208268	0	0	0	0	1.016091	0	1	-360	360;
	8249	310	0.005208	0.208268	0	0	0	0	1.016091	0	1	-360	360;
	8249	309	0.005208	0.208268	0	0	0	0	1.016091	0	1	-360	360;
	6575	308	0.005208	0.208268	0	0	0	0	1.002747	0	1	-360	360;
	6575	307	0.005208	0.208268	0	0	0	0	1.002747	0	1	-360	360;
	13011	306	0.005208	0.208268	0	0	0	0	1.015999	0	1	-360	360;
	13011	305	0.005208	0.208268	0	0	0	0	1.015999	0	1	-360	360;
	13011	4434	0.005208	0.208268	0	0	0	0	1.015999	0	1	-360	360;
	13011	13512	0.005208	0.208268	0	0	0	0	1.015999	0	1	-360	360;
	13011	4432	0.005208	0.208268	0	0	0	0	1.015999	0	1	-360	360;
	13011	4433	0.005208	0.208268	0	0	0	0	1.015999	0	1	-360	360;
	12101	4430	0.005208	0.208268	0	0	0	0	1.00849	0	1	-360	360;
	8287	4431	0.004587	0.183429	0	0	0	0	0.995767	0	1	-360	360;
	8287	4428	0.004587	0.183429	0	0	0	0	0.995767	0	1	-360	360;
	8287	4429	0.004587	0.183429	0	0	0	0	0.995767	0	1	-360	360;
	9324	4426	0.005208	0.208268	0	0	0	0	1.008256	0	1	-360	360;
	9324	4427	0.005208	0.208268	0	0	0	0	1.008256	0	1	-360	360;
	9324	1189	0.005208	0.208268	0	0	0	0	1.008256	0	1	-360	360;
	9324	5330	0.005208	0.208268	0	0	0	0	1.008256	0	1	-360	360;
	9324	11498	0.005208	0.208268	0	0	0	0	1.008256	0	1	-360	360;
	9324	11497	0.005208	0.208268	0	0	0	0	1.008256	0	1	-360	360;
	9324	11500	0.005208	0.208268	0	0	0	0	1.008256	0	1	-360	360;
	9324	11499	0.005208	0.208268	0	0	0	0	1.008256	0	1	-360	360;
	11400	5990	0.000212	0.008494	0	0	0	0	1.059341	0	1	-360	360;
	11400	5989	0.000212	0.008494	0	0	0	0	1.059341	0	1	-360	360;
	11888	5988	0.000212	0.008494	0	0	0	0	1.058922	0	1	-360	360;
	11888	5987	0.000212	0.008494	0	0	0	0	1.058922	0	1	-360	360;
	13198	145	0.000212	0.008494	0	0	0	0	1.06028	0	1	-360	360;
	13198	800	0.000212	0.008494	0	0	0	0	1.06028	0	1	-360	360;
	1373	147	0.000212	0.008494	0	0	0	0	1.060147	0	1	-360	360;
	1373	146	0.000212	0.008494	0	0	0	0	1.060147	0	1	-360	360;
	7251	2711	0.005208	0.208268	0	0	0	0	1.011501	0	1	-360	360;
	7251	9477	0.005208	0.208268	0	0	0	0	1.011501	0	1	-360	360;
	5446	2705	0.005208	0.208268	0	0	0	0	1.006986	0	1	-360	360;
	5446	2703	0.005208	0.208268	0	0	0	0	1.006986	0	1	-360	360;
	4349	9474	0.005208	0.208268	0	0	0	0	0.99541	0	1	-360	360;
	4349	9473	0.005208	0.208268	0	0	0	0	0.99541	0	1	-360	360;
	4349	2369	0.005208	0.208268	0	0	0	0	0.99541	0	1	-360	360;
	4349	2370	0.005208	0.208268	0	0	0	0	0.99541	0	1	-360	360;
	4349	2371	0.005208	0.208268	0	0	0	0	0.99541	0	1	-360	360;
	665	10023	0.003521	0.140801	0	0	0	0	0.997229	0	1	-360	360;
	665	3910	0.003521	0.140801	0	0	0	0	0.997229	0	1	-360	360;
	665	3911	0.003521	0.140801	0	0	0	0	0.997229	0	1	-360	360;
	665	3908	0.003521	0.140801	0	0	0	0	0.997229	0	1	-360	360;
	665	3909	0.003521	0.140801	0	0	0	0	0.997229	0	1	-360	360;
	665	3906	0.003521	0.140801	0	0	0	0	0.997229	0	1	-360	360;
	665	3907	0.003521	0.140801	0	0	0	0	0.997229	0	1	-360	360;
	665	10538	0.003521	0.140801	0	0	0	0	0.997229	0	1	-360	360;
	665	10537	0.003521	0.140801	0	0	0	0	0.997229	0	1	-360	360;
	1251	12884	0.000531	0.021225	0	0	0	0	1.008992	0	1	-360	360;
	1251	9060	0.000531	0.021225	0	0	0	0	1.008992	0	1	-360	360;
	1251	12882	0.000531	0.021225	0	0	0	0	1.008992	0	1	-360	360;
	1251	12883	0.000531	0.021225	0	0	0	0	1.008992	0	1	-360	360;
	1251	12880	0.000531	0.021225	0	0	0	0	1.008992	0	1	-360	360;
	1251	9051	0.000531	0.021225	0	0	0	0	1.008992	0	1	-360	360;
	12241	12878	0.000531	0.021225	0	0	0	0	1.023745	0	1	-360	360;
	12241	9045	0.000531	0.021225	0	0	0	0	1.023745	0	1	-360	360;
	12241	6277	0.000531	0.021225	0	0	0	0	1.023745	0	1	-360	360;
	12241	6276	0.000531	0.021225	0	0	0	0	1.023745	0	1	-360	360;
	12241	6279	0.000531	0.021225	0	0	0	0	1.023745	0	1	-360	360;
	12241	12509	0.000531	0.021225	0	0	0	0	1.023745	0	1	-360	360;
	3506	6201	0.000212	0.008494	0	0	0	0	1.032835	0	1	-360	360;
	3506	6202	0.000212	0.008494	0	0	0	0	1.032835	0	1	-360	360;
	3768	6203	0.000212	0.008494	0	0	0	0	1.032071	0	1	-360	360;
	3768	6204	0.000212	0.008494	0	0	0	0	1.032071	0	1	-360	360;
	7975	6208	0.000212	0.008494	0	0	0	0	1.008136	0	1	-360	360;
	7975	1656	0.000212	0.008494	0	0	0	0	1.008136	0	1	-360	360;
	1892	12818	0.000212	0.008494	0	0	0	0	1.008159	0	1	-360	360;
	1892	12817	0.000212	0.008494	0	0	0	0	1.008159	0	1	-360	360;
	13605	10670	0.005252	0.210018	0	0	0	0	0.997047	0	1	-360	360;
	13605	10456	0.005252	0.210018	0	0	0	0	0.997047	0	1	-360	360;
	13605	10690	0.005252	0.210018	0	0	0	0	0.991031	0	1	-360	360;
	13605	8513	0.005252	0.210018	0	0	0	0	0.991031	0	1	-360	360;
	13605	10518	0.005252	0.210018	0	0	0	0	0.991031	0	1	-360	360;
	13605	10681	0.005252	0.210018	0	0	0	0	0.991031	0	1	-360	360;
	13605	10700	0.005252	0.210018	0	0	0	0	0.991031	0	1	-360	360;
	13605	8516	0.005252	0.210018	0	0	0	0	0.991031	0	1	-360	360;
	3831	1477	0.005208	0.208268	0	0	0	0	1.005883	0	1	-360	360;
	3831	1478	0.005208	0.208268	0	0	0	0	1.005883	0	1	-360	360;
	3831	1479	0.005208	0.208268	0	0	0	0	1.005883	0	1	-360	360;
	13250	9214	0.005285	0.21135	0	0	0	0	1.007908	0	1	-360	360;
	13250	1481	0.005285	0.21135	0	0	0	0	1.002153	0	1	-360	360;
	13250	1482	0.005285	0.21135	0	0	0	0	1.002153	0	1	-360	360;
	13250	1483	0.005285	0.21135	0	0	0	0	1.002153	0	1	-360	360;
	1420	1484	0.005208	0.208268	0	0	0	0	1.007233	0	1	-360	360;
	1420	1485	0.005208	0.208268	0	0	0	0	1.007233	0	1	-360	360;
	1420	1486	0.005208	0.208268	0	0	0	0	1.007233	0	1	-360	360;
	4636	1473	0.002119	0.084719	0	0	0	0	1.008308	0	1	-360	360;
	4636	1474	0.002119	0.084719	0	0	0	0	1.008308	0	1	-360	360;
	4636	1475	0.002119	0.084719	0	0	0	0	1.008308	0	1	-360	360;
	4636	1476	0.002119	0.084719	0	0	0	0	1.008308	0	1	-360	360;
	4636	1470	0.002119	0.084719	0	0	0	0	1.008308	0	1	-360	360;
	4636	1471	0.002119	0.084719	0	0	0	0	1.008308	0	1	-360	360;
	5253	11496	0.003521	0.140801	0	0	0	0	1.008176	0	1	-360	360;
	5253	2236	0.003521	0.140801	0	0	0	0	1.008176	0	1	-360	360;
	5253	11507	0.003521	0.140801	0	0	0	0	1.008176	0	1	-360	360;
	5253	11506	0.003521	0.140801	0	0	0	0	1.008176	0	1	-360	360;
	5253	4424	0.003521	0.140801	0	0	0	0	1.008176	0	1	-360	360;
	5253	4425	0.003521	0.140801	0	0	0	0	1.008176	0	1	-360	360;
	5253	13482	0.003521	0.140801	0	0	0	0	1.008176	0	1	-360	360;
	5253	13486	0.003521	0.140801	0	0	0	0	1.008176	0	1	-360	360;
	5253	13466	0.003521	0.140801	0	0	0	0	1.008176	0	1	-360	360;
	5253	4422	0.003521	0.140801	0	0	0	0	1.008176	0	1	-360	360;
	1126	1928	0.012376	0.494895	0	0	0	0	1.007643	0	1	-360	360;
	1126	3346	0.012376	0.494895	0	0	0	0	1.002354	0	1	-360	360;
	5375	1935	0.005208	0.208268	0	0	0	0	1.01511	0	1	-360	360;
	5375	1936	0.005208	0.208268	0	0	0	0	1.01511	0	1	-360	360;
	5375	7564	0.005208	0.208268	0	0	0	0	1.01511	0	1	-360	360;
	2212	7563	0.005208	0.208268	0	0	0	0	1.010744	0	1	-360	360;
	2212	7562	0.005208	0.208268	0	0	0	0	1.010744	0	1	-360	360;
	7192	11752	0.005208	0.208268	0	0	0	0	1.005182	0	1	-360	360;
	7192	7560	0.005208	0.208268	0	0	0	0	1.005182	0	1	-360	360;
	7192	7559	0.005208	0.208268	0	0	0	0	1.005182	0	1	-360	360;
	7192	7558	0.005208	0.208268	0	0	0	0	1.005182	0	1	-360	360;
	7192	7557	0.005208	0.208268	0	0	0	0	1.005182	0	1	-360	360;
	7192	11892	0.005208	0.208268	0	0	0	0	1.005182	0	1	-360	360;
	7792	1997	0.000212	0.008494	0	0	0	0	1.014224	0	1	-360	360;
	10491	8570	0.000212	0.008494	0	0	0	0	1.01423	0	1	-360	360;
	5404	8571	0.000212	0.008494	0	0	0	0	1.01453	0	1	-360	360;
	7983	8569	0.000212	0.008494	0	0	0	0	1.014502	0	1	-360	360;
	13572	581	0.005208	0.208268	0	0	0	0	1.017389	0	1	-360	360;
	13572	575	0.005208	0.208268	0	0	0	0	1.017389	0	1	-360	360;
	13572	576	0.005208	0.208268	0	0	0	0	1.017389	0	1	-360	360;
	13572	577	0.005208	0.208268	0	0	0	0	1.017389	0	1	-360	360;
	1768	7645	0.000531	0.021225	0	0	0	0	1.00697	0	1	-360	360;
	1768	7649	0.000531	0.021225	0	0	0	0	1.00697	0	1	-360	360;
	1768	7648	0.000531	0.021225	0	0	0	0	1.00697	0	1	-360	360;
	1574	9256	0.000531	0.021225	0	0	0	0	1.006962	0	1	-360	360;
	1574	9257	0.000531	0.021225	0	0	0	0	1.006962	0	1	-360	360;
	1574	9258	0.000531	0.021225	0	0	0	0	1.006962	0	1	-360	360;
	12142	7426	0.000531	0.021225	0	0	0	0	1.00696	0	1	-360	360;
	12142	9260	0.000531	0.021225	0	0	0	0	1.00696	0	1	-360	360;
	12142	9261	0.000531	0.021225	0	0	0	0	1.00696	0	1	-360	360;
	12250	2090	0.000708	0.02832	0	0	0	0	1.008008	0	1	-360	360;
	12250	2091	0.000708	0.02832	0	0	0	0	1.008008	0	1	-360	360;
	12250	2093	0.000708	0.02832	0	0	0	0	1.008008	0	1	-360	360;
	1100	5388	0.005208	0.208268	0	0	0	0	0.999237	0	1	-360	360;
	1100	11913	0.005208	0.208268	0	0	0	0	0.999237	0	1	-360	360;
	1100	11914	0.005208	0.208268	0	0	0	0	0.999237	0	1	-360	360;
	7576	9679	0.000212	0.008494	0	0	0	0	1.021823	0	1	-360	360;
	4529	11916	0.005208	0.208268	0	0	0	0	0.97527	0	1	-360	360;
	4529	13163	0.005208	0.208268	0	0	0	0	0.97527	0	1	-360	360;
	4529	11099	0.005208	0.208268	0	0	0	0	0.97527	0	1	-360	360;
	4529	11919	0.005208	0.208268	0	0	0	0	0.97527	0	1	-360	360;
	9274	10635	0.003521	0.140801	0	0	0	0	1.02372	0	1	-360	360;
	9274	10803	0.003521	0.140801	0	0	0	0	1.02372	0	1	-360	360;
	9274	10804	0.003521	0.140801	0	0	0	0	1.02372	0	1	-360	360;
	9274	3698	0.003521	0.140801	0	0	0	0	1.02372	0	1	-360	360;
	9274	3697	0.003521	0.140801	0	0	0	0	1.02372	0	1	-360	360;
	9274	3696	0.003521	0.140801	0	0	0	0	1.02372	0	1	-360	360;
	9274	3695	0.003521	0.140801	0	0	0	0	1.02372	0	1	-360	360;
	9274	3701	0.003521	0.140801	0	0	0	0	1.02372	0	1	-360	360;
	9274	3700	0.003521	0.140801	0	0	0	0	1.02372	0	1	-360	360;
	9274	3699	0.003521	0.140801	0	0	0	0	1.02372	0	1	-360	360;
	12839	1853	0.005208	0.208268	0	0	0	0	0.996937	0	1	-360	360;
	12839	5849	0.005208	0.208268	0	0	0	0	0.996937	0	1	-360	360;
	8404	1631	0.000708	0.02832	0	0	0	0	1.017715	0	1	-360	360;
	8404	8648	0.000708	0.02832	0	0	0	0	1.017715	0	1	-360	360;
	8404	8647	0.000708	0.02832	0	0	0	0	1.017715	0	1	-360	360;
	8404	4729	0.002119	0.084719	0	0	0	0	1.015542	0	1	-360	360;
	8404	4728	0.002119	0.084719	0	0	0	0	1.015542	0	1	-360	360;
	8404	4727	0.002119	0.084719	0	0	0	0	1.015542	0	1	-360	360;
	8404	4726	0.002119	0.084719	0	0	0	0	1.015542	0	1	-360	360;
	8404	4725	0.002119	0.084719	0	0	0	0	1.015542	0	1	-360	360;
	8404	4724	0.002119	0.084719	0	0	0	0	1.015542	0	1	-360	360;
	8404	4723	0.002119	0.084719	0	0	0	0	1.015542	0	1	-360	360;
	8404	4722	0.002119	0.084719	0	0	0	0	1.015542	0	1	-360	360;
	8404	10811	0.002119	0.084719	0	0	0	0	1.015542	0	1	-360	360;
	13274	6329	0.005208	0.208268	0	0	0	0	1.012806	0	1	-360	360;
	13274	6328	0.005208	0.208268	0	0	0	0	1.012806	0	1	-360	360;
	13274	6327	0.005208	0.208268	0	0	0	0	1.012806	0	1	-360	360;
	13274	6326	0.005208	0.208268	0	0	0	0	1.012806	0	1	-360	360;
	13274	6325	0.005208	0.208268	0	0	0	0	1.012806	0	1	-360	360;
	6011	5229	0.003521	0.140801	0	0	0	0	1.009982	0	1	-360	360;
	6011	5228	0.003521	0.140801	0	0	0	0	1.009982	0	1	-360	360;
	6011	5226	0.003521	0.140801	0	0	0	0	1.009982	0	1	-360	360;
	6011	5225	0.003521	0.140801	0	0	0	0	1.009982	0	1	-360	360;
	6011	11759	0.003521	0.140801	0	0	0	0	1.009982	0	1	-360	360;
	6011	11760	0.003521	0.140801	0	0	0	0	1.009982	0	1	-360	360;
	6011	2646	0.003521	0.140801	0	0	0	0	1.009982	0	1	-360	360;
	6011	11758	0.003521	0.140801	0	0	0	0	1.009982	0	1	-360	360;
	6011	11762	0.003521	0.140801	0	0	0	0	1.009982	0	1	-360	360;
	6011	11763	0.003521	0.140801	0	0	0	0	1.009982	0	1	-360	360;
	10526	8713	0.000212	0.008494	0	0	0	0	1.052995	0	1	-360	360;
	10526	8714	0.000212	0.008494	0	0	0	0	1.052995	0	1	-360	360;
	7834	8711	0.000212	0.008494	0	0	0	0	1.053262	0	1	-360	360;
	7834	8712	0.000212	0.008494	0	0	0	0	1.053262	0	1	-360	360;
	1748	1689	0.000212	0.008494	0	0	0	0	1.055552	0	1	-360	360;
	1748	10530	0.000212	0.008494	0	0	0	0	1.055552	0	1	-360	360;
	9437	6798	0.005208	0.208268	0	0	0	0	0.996988	0	1	-360	360;
	9437	6797	0.005208	0.208268	0	0	0	0	0.996988	0	1	-360	360;
	5960	6793	0.005208	0.208268	0	0	0	0	1.007213	0	1	-360	360;
	5960	6792	0.005208	0.208268	0	0	0	0	1.007213	0	1	-360	360;
	7462	6795	0.005208	0.208268	0	0	0	0	0.996634	0	1	-360	360;
	7462	6794	0.005208	0.208268	0	0	0	0	0.996634	0	1	-360	360;
	4776	6802	0.005208	0.208268	0	0	0	0	1.002795	0	1	-360	360;
	4776	6801	0.005208	0.208268	0	0	0	0	1.002795	0	1	-360	360;
	1902	13419	0.005208	0.208268	0	0	0	0	1.009263	0	1	-360	360;
	1902	6731	0.005208	0.208268	0	0	0	0	1.009263	0	1	-360	360;
	1902	12219	0.005208	0.208268	0	0	0	0	1.009263	0	1	-360	360;
	3998	12222	0.005208	0.208268	0	0	0	0	1.009695	0	1	-360	360;
	3998	13423	0.005208	0.208268	0	0	0	0	1.009695	0	1	-360	360;
	3998	13424	0.005208	0.208268	0	0	0	0	1.009695	0	1	-360	360;
	3998	13421	0.005208	0.208268	0	0	0	0	1.009695	0	1	-360	360;
	3998	10050	0.005208	0.208268	0	0	0	0	1.009695	0	1	-360	360;
	5810	13426	0.003268	0.130678	0	0	0	0	0.997548	0	1	-360	360;
	5810	13427	0.003268	0.130678	0	0	0	0	0.992409	0	1	-360	360;
	5810	3305	0.003268	0.130678	0	0	0	0	0.992409	0	1	-360	360;
	5810	3304	0.003268	0.130678	0	0	0	0	0.992409	0	1	-360	360;
	5810	3303	0.003268	0.130678	0	0	0	0	0.992409	0	1	-360	360;
	5810	3302	0.003268	0.130678	0	0	0	0	0.992409	0	1	-360	360;
	8243	3309	0.005208	0.208268	0	0	0	0	0.978538	0	1	-360	360;
	8243	3308	0.005208	0.208268	0	0	0	0	0.978538	0	1	-360	360;
	8243	6969	0.005208	0.208268	0	0	0	0	0.978538	0	1	-360	360;
	8243	6967	0.005208	0.208268	0	0	0	0	0.978538	0	1	-360	360;
	8243	5463	0.005208	0.208268	0	0	0	0	0.978538	0	1	-360	360;
	8243	6944	0.005208	0.208268	0	0	0	0	0.978538	0	1	-360	360;
	3889	9926	0.005297	0.211798	0	0	0	0	1.010342	0	1	-360	360;
	3889	5528	0.005297	0.211798	0	0	0	0	1.010342	0	1	-360	360;
	3889	9928	0.005297	0.211798	0	0	0	0	1.004656	0	1	-360	360;
	3889	9929	0.005297	0.211798	0	0	0	0	1.004656	0	1	-360	360;
	3889	9930	0.005297	0.211798	0	0	0	0	1.004656	0	1	-360	360;
	8087	11922	0.005208	0.208268	0	0	0	0	1.014926	0	1	-360	360;
	8087	9932	0.005208	0.208268	0	0	0	0	1.014926	0	1	-360	360;
	10282	9933	0.005208	0.208268	0	0	0	0	1.014966	0	1	-360	360;
	10282	9922	0.005208	0.208268	0	0	0	0	1.014966	0	1	-360	360;
	4682	9923	0.005208	0.208268	0	0	0	0	1.010242	0	1	-360	360;
	4682	2153	0.005208	0.208268	0	0	0	0	1.010242	0	1	-360	360;
	1944	2154	0.005208	0.208268	0	0	0	0	1.017484	0	1	-360	360;
	1944	13585	0.005208	0.208268	0	0	0	0	1.017484	0	1	-360	360;
	1944	13589	0.005208	0.208268	0	0	0	0	1.017484	0	1	-360	360;
	1944	13604	0.005208	0.208268	0	0	0	0	1.017484	0	1	-360	360;
	9262	2158	0.005208	0.208268	0	0	0	0	1.014014	0	1	-360	360;
	9262	2155	0.005208	0.208268	0	0	0	0	1.014014	0	1	-360	360;
	2016	2156	0.005208	0.208268	0	0	0	0	0.996597	0	1	-360	360;
	2424	62	0.000212	0.008494	0	0	0	0	1.012282	0	1	-360	360;
	2424	63	0.000212	0.008494	0	0	0	0	1.012282	0	1	-360	360;
	13136	1726	0.000212	0.008494	0	0	0	0	1.012028	0	1	-360	360;
	13136	7063	0.000212	0.008494	0	0	0	0	1.012028	0	1	-360	360;
	2797	9225	0.005208	0.208268	0	0	0	0	1.015876	0	1	-360	360;
	1152	1298	0.005208	0.208268	0	0	0	0	0.996815	0	1	-360	360;
	1152	1287	0.005208	0.208268	0	0	0	0	0.996815	0	1	-360	360;
	1152	9218	0.005208	0.208268	0	0	0	0	0.996815	0	1	-360	360;
	1152	11344	0.005208	0.208268	0	0	0	0	0.996815	0	1	-360	360;
	6052	1290	0.005208	0.208268	0	0	0	0	1.001459	0	1	-360	360;
	6052	9216	0.005208	0.208268	0	0	0	0	1.001459	0	1	-360	360;
	2507	9215	0.005208	0.208268	0	0	0	0	0.997291	0	1	-360	360;
	1215	510	0.000212	0.008494	0	0	0	0	1.013208	0	1	-360	360;
	4360	3068	0.000212	0.008494	0	0	0	0	1.012326	0	1	-360	360;
	7051	508	0.000212	0.008494	0	0	0	0	1.013245	0	1	-360	360;
	2892	509	0.000212	0.008494	0	0	0	0	1.012399	0	1	-360	360;
	5683	514	0.000212	0.008494	0	0	0	0	1.013272	0	1	-360	360;
	5500	3757	0.000531	0.021225	0	0	0	0	1.007977	0	1	-360	360;
	5500	3754	0.000531	0.021225	0	0	0	0	1.007977	0	1	-360	360;
	5500	11950	0.000531	0.021225	0	0	0	0	1.007977	0	1	-360	360;
	5981	1662	0.007886	0.315359	0	0	0	0	0.999304	0	1	-360	360;
	5981	1663	0.007886	0.315359	0	0	0	0	0.993609	0	1	-360	360;
	6416	13520	0.002119	0.084719	0	0	0	0	1.009959	0	1	-360	360;
	6416	4187	0.002119	0.084719	0	0	0	0	1.009959	0	1	-360	360;
	6416	13489	0.002119	0.084719	0	0	0	0	1.009959	0	1	-360	360;
	6416	4185	0.002119	0.084719	0	0	0	0	1.009959	0	1	-360	360;
	6416	4182	0.002119	0.084719	0	0	0	0	1.009959	0	1	-360	360;
	6416	4183	0.002119	0.084719	0	0	0	0	1.009959	0	1	-360	360;
	6416	1074	0.002119	0.084719	0	0	0	0	1.009959	0	1	-360	360;
	12849	8675	0.005208	0.208268	0	0	0	0	1.007062	0	1	-360	360;
	12849	8680	0.005208	0.208268	0	0	0	0	1.007062	0	1	-360	360;
	12849	8679	0.005208	0.208268	0	0	0	0	1.007062	0	1	-360	360;
	7414	2783	0.003521	0.140801	0	0	0	0	0.995036	0	1	-360	360;
	7414	2799	0.003521	0.140801	0	0	0	0	0.995036	0	1	-360	360;
	7414	2810	0.003521	0.140801	0	0	0	0	0.995036	0	1	-360	360;
	7414	12673	0.003521	0.140801	0	0	0	0	0.995036	0	1	-360	360;
	7414	2012	0.003521	0.140801	0	0	0	0	0.995036	0	1	-360	360;
	7414	2013	0.003521	0.140801	0	0	0	0	0.995036	0	1	-360	360;
	7414	2014	0.003521	0.140801	0	0	0	0	0.995036	0	1	-360	360;
	7414	12659	0.003521	0.140801	0	0	0	0	0.995036	0	1	-360	360;
	7414	2020	0.003521	0.140801	0	0	0	0	0.995036	0	1	-360	360;
	7414	12690	0.003521	0.140801	0	0	0	0	0.995036	0	1	-360	360;
	11183	8219	0.005208	0.208268	0	0	0	0	1.004973	0	1	-360	360;
	11183	8218	0.005208	0.208268	0	0	0	0	1.004973	0	1	-360	360;
	11183	8221	0.005208	0.208268	0	0	0	0	1.004973	0	1	-360	360;
	11183	8220	0.005208	0.208268	0	0	0	0	1.004973	0	1	-360	360;
	11183	8223	0.005208	0.208268	0	0	0	0	1.004973	0	1	-360	360;
	11183	8222	0.005208	0.208268	0	0	0	0	1.004973	0	1	-360	360;
	11183	8224	0.005208	0.208268	0	0	0	0	1.004973	0	1	-360	360;
	11183	4691	0.002119	0.084719	0	0	0	0	1.00545	0	1	-360	360;
	11183	4698	0.002119	0.084719	0	0	0	0	1.00545	0	1	-360	360;
	11183	4699	0.002119	0.084719	0	0	0	0	1.00545	0	1	-360	360;
	11183	5523	0.002119	0.084719	0	0	0	0	1.00545	0	1	-360	360;
	11183	12232	0.002119	0.084719	0	0	0	0	1.00545	0	1	-360	360;
	11183	5530	0.002119	0.084719	0	0	0	0	1.00545	0	1	-360	360;
	6064	749	0.005208	0.208268	0	0	0	0	0.999883	0	1	-360	360;
	6064	750	0.005208	0.208268	0	0	0	0	0.999883	0	1	-360	360;
	6064	751	0.005208	0.208268	0	0	0	0	0.999883	0	1	-360	360;
	13405	10132	0.005208	0.208268	0	0	0	0	1.019382	0	1	-360	360;
	13405	6749	0.005208	0.208268	0	0	0	0	1.019382	0	1	-360	360;
	13405	10134	0.005208	0.208268	0	0	0	0	1.019382	0	1	-360	360;
	13405	10135	0.005208	0.208268	0	0	0	0	1.019382	0	1	-360	360;
	13405	3523	0.005208	0.208268	0	0	0	0	1.019382	0	1	-360	360;
	13405	8130	0.005208	0.208268	0	0	0	0	1.019382	0	1	-360	360;
	13405	3521	0.005208	0.208268	0	0	0	0	1.019382	0	1	-360	360;
	13405	3520	0.005208	0.208268	0	0	0	0	1.019382	0	1	-360	360;
	13405	3526	0.005208	0.208268	0	0	0	0	1.019382	0	1	-360	360;
	13405	3525	0.005208	0.208268	0	0	0	0	1.019382	0	1	-360	360;
	13405	3524	0.005208	0.208268	0	0	0	0	1.019382	0	1	-360	360;
	10182	7732	0.005208	0.208268	0	0	0	0	1.013776	0	1	-360	360;
	10182	11401	0.005208	0.208268	0	0	0	0	1.013776	0	1	-360	360;
	5234	5973	0.005208	0.208268	0	0	0	0	1.018757	0	1	-360	360;
	5234	4041	0.005208	0.208268	0	0	0	0	1.018757	0	1	-360	360;
	5312	4042	0.005208	0.208268	0	0	0	0	1.008125	0	1	-360	360;
	5312	4039	0.005208	0.208268	0	0	0	0	1.008125	0	1	-360	360;
	5312	4040	0.005208	0.208268	0	0	0	0	1.008125	0	1	-360	360;
	8652	4045	0.005208	0.208268	0	0	0	0	1.008578	0	1	-360	360;
	12476	13190	0.000708	0.02832	0	0	0	0	0.990397	0	1	-360	360;
	12476	13189	0.000708	0.02832	0	0	0	0	0.990397	0	1	-360	360;
	12476	13188	0.000708	0.02832	0	0	0	0	0.990397	0	1	-360	360;
	12299	4048	0.005208	0.208268	0	0	0	0	1.002469	0	1	-360	360;
	12299	11330	0.005208	0.208268	0	0	0	0	1.002469	0	1	-360	360;
	12511	12825	0.005208	0.208268	0	0	0	0	0.998862	0	1	-360	360;
	12511	11122	0.005208	0.208268	0	0	0	0	0.998862	0	1	-360	360;
	10737	11125	0.005208	0.208268	0	0	0	0	1.010853	0	1	-360	360;
	10737	11124	0.005208	0.208268	0	0	0	0	1.010853	0	1	-360	360;
	10737	11119	0.005208	0.208268	0	0	0	0	1.010853	0	1	-360	360;
	10737	11118	0.005208	0.208268	0	0	0	0	1.010853	0	1	-360	360;
	10944	11121	0.005208	0.208268	0	0	0	0	0.989334	0	1	-360	360;
	10944	11120	0.005208	0.208268	0	0	0	0	0.989334	0	1	-360	360;
	10944	11130	0.005208	0.208268	0	0	0	0	0.989334	0	1	-360	360;
	2581	11129	0.005208	0.208268	0	0	0	0	1.009184	0	1	-360	360;
	2581	7946	0.005208	0.208268	0	0	0	0	1.009184	0	1	-360	360;
	2581	12697	0.005208	0.208268	0	0	0	0	1.009184	0	1	-360	360;
	2581	3878	0.005208	0.208268	0	0	0	0	1.009184	0	1	-360	360;
	2581	12699	0.005208	0.208268	0	0	0	0	1.009184	0	1	-360	360;
	2581	12702	0.005208	0.208268	0	0	0	0	1.009184	0	1	-360	360;
	6778	12701	0.005208	0.208268	0	0	0	0	1.010864	0	1	-360	360;
	6778	12704	0.005208	0.208268	0	0	0	0	1.010864	0	1	-360	360;
	2454	3916	0.005208	0.208268	0	0	0	0	1.019687	0	1	-360	360;
	5112	12706	0.005208	0.208268	0	0	0	0	1.014538	0	1	-360	360;
	5112	12705	0.005208	0.208268	0	0	0	0	1.014538	0	1	-360	360;
	5112	5622	0.005208	0.208268	0	0	0	0	1.014538	0	1	-360	360;
	12744	5476	0.002119	0.084719	0	0	0	0	1.005347	0	1	-360	360;
	12744	7386	0.002119	0.084719	0	0	0	0	1.005347	0	1	-360	360;
	12744	12586	0.002119	0.084719	0	0	0	0	1.005347	0	1	-360	360;
	12744	12592	0.002119	0.084719	0	0	0	0	1.005347	0	1	-360	360;
	12744	12591	0.002119	0.084719	0	0	0	0	1.005347	0	1	-360	360;
	12744	12590	0.002119	0.084719	0	0	0	0	1.005347	0	1	-360	360;
	12744	12589	0.002119	0.084719	0	0	0	0	1.005347	0	1	-360	360;
	8065	5619	0.005208	0.208268	0	0	0	0	1.011409	0	1	-360	360;
	896	6593	0.005208	0.208268	0	0	0	0	1.015669	0	1	-360	360;
	227	6008	0.002119	0.084719	0	0	0	0	0.991144	0	1	-360	360;
	227	8838	0.002119	0.084719	0	0	0	0	0.991144	0	1	-360	360;
	227	6006	0.002119	0.084719	0	0	0	0	0.991144	0	1	-360	360;
	227	6007	0.002119	0.084719	0	0	0	0	0.991144	0	1	-360	360;
	227	8824	0.002119	0.084719	0	0	0	0	0.991144	0	1	-360	360;
	227	6005	0.002119	0.084719	0	0	0	0	0.991144	0	1	-360	360;
	227	6002	0.002119	0.084719	0	0	0	0	0.991144	0	1	-360	360;
	227	6003	0.002119	0.084719	0	0	0	0	0.991144	0	1	-360	360;
	6400	5134	0.005208	0.208268	0	0	0	0	1.004944	0	1	-360	360;
	6400	12175	0.005208	0.208268	0	0	0	0	1.004944	0	1	-360	360;
	5691	5126	0.005208	0.208268	0	0	0	0	1.008895	0	1	-360	360;
	5691	5127	0.005208	0.208268	0	0	0	0	1.008895	0	1	-360	360;
	934	5128	0.005208	0.208268	0	0	0	0	0.991535	0	1	-360	360;
	934	5129	0.005208	0.208268	0	0	0	0	0.991535	0	1	-360	360;
	934	5130	0.005208	0.208268	0	0	0	0	0.991535	0	1	-360	360;
	934	5131	0.005208	0.208268	0	0	0	0	0.991535	0	1	-360	360;
	934	13584	0.005208	0.208268	0	0	0	0	0.991535	0	1	-360	360;
	934	5133	0.005208	0.208268	0	0	0	0	0.991535	0	1	-360	360;
	934	5122	0.005208	0.208268	0	0	0	0	0.991535	0	1	-360	360;
	158	5123	0.005208	0.208268	0	0	0	0	1.011892	0	1	-360	360;
	3455	4405	0.005208	0.208268	0	0	0	0	1.012367	0	1	-360	360;
	3455	77	0.005208	0.208268	0	0	0	0	1.012367	0	1	-360	360;
	12568	454	0.02451	0.980086	0	0	0	0	0.999473	0	1	-360	360;
	897	2167	0.005208	0.208268	0	0	0	0	1.013998	0	1	-360	360;
	897	9377	0.005208	0.208268	0	0	0	0	1.013998	0	1	-360	360;
	897	9376	0.005208	0.208268	0	0	0	0	1.013998	0	1	-360	360;
	897	9379	0.005208	0.208268	0	0	0	0	1.013998	0	1	-360	360;
	897	9378	0.005208	0.208268	0	0	0	0	1.013998	0	1	-360	360;
	897	9384	0.005208	0.208268	0	0	0	0	1.013998	0	1	-360	360;
	897	9383	0.005208	0.208268	0	0	0	0	1.013998	0	1	-360	360;
	897	2284	0.005208	0.208268	0	0	0	0	1.013998	0	1	-360	360;
	897	2285	0.005208	0.208268	0	0	0	0	1.013998	0	1	-360	360;
	897	2282	0.005208	0.208268	0	0	0	0	1.013998	0	1	-360	360;
	897	2283	0.005208	0.208268	0	0	0	0	1.013998	0	1	-360	360;
	897	811	0.005208	0.208268	0	0	0	0	1.013998	0	1	-360	360;
	12397	12664	0.005208	0.208268	0	0	0	0	1.00865	0	1	-360	360;
	12397	11484	0.005208	0.208268	0	0	0	0	1.00865	0	1	-360	360;
	1563	6630	0.005208	0.208268	0	0	0	0	1.014801	0	1	-360	360;
	1563	6623	0.005208	0.208268	0	0	0	0	1.014801	0	1	-360	360;
	1563	6624	0.005208	0.208268	0	0	0	0	1.014801	0	1	-360	360;
	10058	13064	0.005208	0.208268	0	0	0	0	1.010547	0	1	-360	360;
	10058	13060	0.005208	0.208268	0	0	0	0	1.010547	0	1	-360	360;
	10058	13059	0.005208	0.208268	0	0	0	0	1.010547	0	1	-360	360;
	10058	13055	0.005208	0.208268	0	0	0	0	1.010547	0	1	-360	360;
	10058	8906	0.005208	0.208268	0	0	0	0	1.010547	0	1	-360	360;
	10058	13046	0.005208	0.208268	0	0	0	0	1.010547	0	1	-360	360;
	10058	13043	0.005208	0.208268	0	0	0	0	1.010547	0	1	-360	360;
	10058	8903	0.005208	0.208268	0	0	0	0	1.010547	0	1	-360	360;
	10058	8902	0.005208	0.208268	0	0	0	0	1.010547	0	1	-360	360;
	10058	8901	0.005208	0.208268	0	0	0	0	1.010547	0	1	-360	360;
	8572	6090	0.005208	0.208268	0	0	0	0	1.009889	0	1	-360	360;
	8572	6091	0.005208	0.208268	0	0	0	0	1.009889	0	1	-360	360;
	4504	6092	0.005208	0.208268	0	0	0	0	1.00878	0	1	-360	360;
	7638	6093	0.005208	0.208268	0	0	0	0	1.016594	0	1	-360	360;
	7638	10447	0.005208	0.208268	0	0	0	0	1.016594	0	1	-360	360;
	7638	6087	0.005208	0.208268	0	0	0	0	1.016594	0	1	-360	360;
	7638	10470	0.005208	0.208268	0	0	0	0	1.016594	0	1	-360	360;
	4002	6089	0.005208	0.208268	0	0	0	0	1.007312	0	1	-360	360;
	4002	6097	0.005208	0.208268	0	0	0	0	1.007312	0	1	-360	360;
	4002	6098	0.005208	0.208268	0	0	0	0	1.007312	0	1	-360	360;
	4002	975	0.005208	0.208268	0	0	0	0	1.007312	0	1	-360	360;
	13455	7470	0.000212	0.008494	0	0	0	0	1.003058	0	1	-360	360;
	3083	977	0.005208	0.208268	0	0	0	0	1.012007	0	1	-360	360;
	3083	976	0.005208	0.208268	0	0	0	0	1.012007	0	1	-360	360;
	3083	978	0.005208	0.208268	0	0	0	0	1.012007	0	1	-360	360;
	10309	3528	0.002119	0.084719	0	0	0	0	0.998279	0	1	-360	360;
	10309	3529	0.002119	0.084719	0	0	0	0	0.998279	0	1	-360	360;
	10309	3530	0.002119	0.084719	0	0	0	0	0.998279	0	1	-360	360;
	10309	8123	0.002119	0.084719	0	0	0	0	0.998279	0	1	-360	360;
	10309	8126	0.002119	0.084719	0	0	0	0	0.998279	0	1	-360	360;
	10309	10619	0.002119	0.084719	0	0	0	0	0.998279	0	1	-360	360;
	10008	7508	0.005208	0.208268	0	0	0	0	0.984973	0	1	-360	360;
	1268	4707	0.005208	0.208268	0	0	0	0	1.022564	0	1	-360	360;
	1268	10339	0.005208	0.208268	0	0	0	0	1.022564	0	1	-360	360;
	1268	516	0.005208	0.208268	0	0	0	0	1.022564	0	1	-360	360;
	1268	4702	0.005208	0.208268	0	0	0	0	1.022564	0	1	-360	360;
	1268	4690	0.005208	0.208268	0	0	0	0	1.022564	0	1	-360	360;
	1268	4694	0.005208	0.208268	0	0	0	0	1.022564	0	1	-360	360;
	1268	7512	0.005208	0.208268	0	0	0	0	1.022564	0	1	-360	360;
	4165	7513	0.005208	0.208268	0	0	0	0	1.013437	0	1	-360	360;
	4077	9475	0.005208	0.208268	0	0	0	0	1.019745	0	1	-360	360;
	4077	9476	0.005208	0.208268	0	0	0	0	1.019745	0	1	-360	360;
	4077	2708	0.005208	0.208268	0	0	0	0	1.019745	0	1	-360	360;
	12559	9478	0.005208	0.208268	0	0	0	0	1.007384	0	1	-360	360;
	12559	9479	0.005208	0.208268	0	0	0	0	1.007384	0	1	-360	360;
	12559	9480	0.005208	0.208268	0	0	0	0	1.007384	0	1	-360	360;
	12559	9481	0.005208	0.208268	0	0	0	0	1.007384	0	1	-360	360;
	12559	9482	0.005208	0.208268	0	0	0	0	1.007384	0	1	-360	360;
	12492	9471	0.005208	0.208268	0	0	0	0	1.008819	0	1	-360	360;
	12492	9472	0.005208	0.208268	0	0	0	0	1.008819	0	1	-360	360;
	12492	2841	0.005208	0.208268	0	0	0	0	1.008819	0	1	-360	360;
	12492	2840	0.005208	0.208268	0	0	0	0	1.008819	0	1	-360	360;
	12492	12738	0.005208	0.208268	0	0	0	0	1.008819	0	1	-360	360;
	7948	2838	0.005208	0.208268	0	0	0	0	1.005944	0	1	-360	360;
	7948	2845	0.005208	0.208268	0	0	0	0	1.005944	0	1	-360	360;
	1162	2844	0.005208	0.208268	0	0	0	0	1.007417	0	1	-360	360;
	1162	4200	0.005208	0.208268	0	0	0	0	1.007417	0	1	-360	360;
	9183	2842	0.005208	0.208268	0	0	0	0	0.995547	0	1	-360	360;
	9183	4166	0.005208	0.208268	0	0	0	0	0.995547	0	1	-360	360;
	9183	2836	0.005208	0.208268	0	0	0	0	0.995547	0	1	-360	360;
	501	4171	0.005208	0.208268	0	0	0	0	1.007942	0	1	-360	360;
	501	207	0.005208	0.208268	0	0	0	0	1.007942	0	1	-360	360;
	501	9963	0.005208	0.208268	0	0	0	0	1.007942	0	1	-360	360;
	3541	9964	0.005208	0.208268	0	0	0	0	1.007773	0	1	-360	360;
	3541	9961	0.005208	0.208268	0	0	0	0	1.007773	0	1	-360	360;
	12472	5838	0.000212	0.008494	0	0	0	0	1.00669	0	1	-360	360;
	13045	9959	0.005208	0.208268	0	0	0	0	1.012324	0	1	-360	360;
	13045	13491	0.005208	0.208268	0	0	0	0	1.012324	0	1	-360	360;
	13045	9957	0.005208	0.208268	0	0	0	0	1.012324	0	1	-360	360;
	13045	9958	0.005208	0.208268	0	0	0	0	1.012324	0	1	-360	360;
	9009	12921	0.000212	0.008494	0	0	0	0	1.00594	0	1	-360	360;
	9009	12920	0.000212	0.008494	0	0	0	0	1.00594	0	1	-360	360;
	9009	12937	0.000212	0.008494	0	0	0	0	1.00594	0	1	-360	360;
	9009	12925	0.000212	0.008494	0	0	0	0	1.00594	0	1	-360	360;
	3268	3335	0.005208	0.208268	0	0	0	0	1.011375	0	1	-360	360;
	3268	3334	0.005208	0.208268	0	0	0	0	1.011375	0	1	-360	360;
	3268	3337	0.005208	0.208268	0	0	0	0	1.011375	0	1	-360	360;
	3268	3336	0.005208	0.208268	0	0	0	0	1.011375	0	1	-360	360;
	3268	3339	0.005208	0.208268	0	0	0	0	1.011375	0	1	-360	360;
	4479	3338	0.005208	0.208268	0	0	0	0	1.000873	0	1	-360	360;
	4479	10436	0.005208	0.208268	0	0	0	0	1.000873	0	1	-360	360;
	3559	10437	0.005208	0.208268	0	0	0	0	1.010394	0	1	-360	360;
	3559	10438	0.005208	0.208268	0	0	0	0	1.010394	0	1	-360	360;
	4479	10439	0.005208	0.208268	0	0	0	0	1.000811	0	1	-360	360;
	4479	10432	0.005208	0.208268	0	0	0	0	1.000811	0	1	-360	360;
	213	8479	0.005208	0.208268	0	0	0	0	1.019841	0	1	-360	360;
	213	10434	0.005208	0.208268	0	0	0	0	1.019841	0	1	-360	360;
	213	10435	0.005208	0.208268	0	0	0	0	1.019841	0	1	-360	360;
	213	3608	0.005208	0.208268	0	0	0	0	1.019841	0	1	-360	360;
	7396	10443	0.005208	0.208268	0	0	0	0	0.992269	0	1	-360	360;
	7396	9946	0.005208	0.208268	0	0	0	0	0.992269	0	1	-360	360;
	7396	3816	0.005208	0.208268	0	0	0	0	0.992269	0	1	-360	360;
	7396	3815	0.005208	0.208268	0	0	0	0	0.992269	0	1	-360	360;
	10564	3814	0.028409	1.136008	0	0	0	0	1.008318	0	1	-360	360;
	1744	3491	0.005208	0.208268	0	0	0	0	0.998258	0	1	-360	360;
	1744	3812	0.005208	0.208268	0	0	0	0	0.998258	0	1	-360	360;
	1744	3811	0.005208	0.208268	0	0	0	0	0.998258	0	1	-360	360;
	11080	3810	0.005208	0.208268	0	0	0	0	1.013576	0	1	-360	360;
	11080	3820	0.005208	0.208268	0	0	0	0	1.013576	0	1	-360	360;
	8460	13392	0.005208	0.208268	0	0	0	0	1.010244	0	1	-360	360;
	8460	6757	0.005208	0.208268	0	0	0	0	1.010244	0	1	-360	360;
	8460	6758	0.005208	0.208268	0	0	0	0	1.010244	0	1	-360	360;
	8460	13580	0.005208	0.208268	0	0	0	0	1.010244	0	1	-360	360;
	8460	13582	0.005208	0.208268	0	0	0	0	1.010244	0	1	-360	360;
	8460	6761	0.005208	0.208268	0	0	0	0	1.010244	0	1	-360	360;
	8460	6762	0.005208	0.208268	0	0	0	0	1.010244	0	1	-360	360;
	8460	6759	0.005208	0.208268	0	0	0	0	1.010244	0	1	-360	360;
	8460	6760	0.005208	0.208268	0	0	0	0	1.010244	0	1	-360	360;
	8460	6763	0.005208	0.208268	0	0	0	0	1.010244	0	1	-360	360;
	8460	9659	0.005208	0.208268	0	0	0	0	1.010244	0	1	-360	360;
	1252	4386	0.005208	0.208268	0	0	0	0	1.000067	0	1	-360	360;
	1252	4385	0.005208	0.208268	0	0	0	0	1.000067	0	1	-360	360;
	1252	4387	0.005208	0.208268	0	0	0	0	1.000067	0	1	-360	360;
	4031	1675	0.000212	0.008494	0	0	0	0	1.01776	0	1	-360	360;
	4031	276	0.000212	0.008494	0	0	0	0	1.01776	0	1	-360	360;
	2558	275	0.000212	0.008494	0	0	0	0	1.017551	0	1	-360	360;
	2558	274	0.000212	0.008494	0	0	0	0	1.017551	0	1	-360	360;
	11442	273	0.000212	0.008494	0	0	0	0	1.017539	0	1	-360	360;
	11442	1661	0.000212	0.008494	0	0	0	0	1.017539	0	1	-360	360;
	523	271	0.000212	0.008494	0	0	0	0	1.017942	0	1	-360	360;
	523	12968	0.000212	0.008494	0	0	0	0	1.017942	0	1	-360	360;
	5011	3065	0.000531	0.021225	0	0	0	0	1.00666	0	1	-360	360;
	5011	7531	0.000531	0.021225	0	0	0	0	1.00666	0	1	-360	360;
	5011	5657	0.000531	0.021225	0	0	0	0	1.00666	0	1	-360	360;
	1835	11454	0.005208	0.208268	0	0	0	0	0.999473	0	1	-360	360;
	1835	11455	0.005208	0.208268	0	0	0	0	0.999473	0	1	-360	360;
	10485	11456	0.006812	0.272394	0	0	0	0	1.007176	0	1	-360	360;
	10485	11457	0.006812	0.272394	0	0	0	0	1.001423	0	1	-360	360;
	10485	11458	0.006812	0.272394	0	0	0	0	1.001423	0	1	-360	360;
	10485	11459	0.006812	0.272394	0	0	0	0	1.001423	0	1	-360	360;
	10485	4901	0.006812	0.272394	0	0	0	0	1.001423	0	1	-360	360;
	10485	2289	0.006812	0.272394	0	0	0	0	1.001423	0	1	-360	360;
	1173	11419	0.000212	0.008494	0	0	0	0	1.014078	0	1	-360	360;
	1173	11418	0.000212	0.008494	0	0	0	0	1.014078	0	1	-360	360;
	12161	6254	0.000212	0.008494	0	0	0	0	1.013997	0	1	-360	360;
	12161	709	0.000212	0.008494	0	0	0	0	1.013997	0	1	-360	360;
	5760	4903	0.005208	0.208268	0	0	0	0	1.010522	0	1	-360	360;
	5760	4902	0.005208	0.208268	0	0	0	0	1.010522	0	1	-360	360;
	8710	4905	0.005208	0.208268	0	0	0	0	1.00576	0	1	-360	360;
	8710	4904	0.005208	0.208268	0	0	0	0	1.00576	0	1	-360	360;
	11004	6611	0.002119	0.084719	0	0	0	0	0.985768	0	1	-360	360;
	11004	6612	0.002119	0.084719	0	0	0	0	0.985768	0	1	-360	360;
	11004	6609	0.002119	0.084719	0	0	0	0	0.985768	0	1	-360	360;
	11004	6610	0.002119	0.084719	0	0	0	0	0.985768	0	1	-360	360;
	11004	6607	0.002119	0.084719	0	0	0	0	0.985768	0	1	-360	360;
	11004	6608	0.002119	0.084719	0	0	0	0	0.985768	0	1	-360	360;
	11004	12529	0.002119	0.084719	0	0	0	0	0.985768	0	1	-360	360;
	11004	6606	0.002119	0.084719	0	0	0	0	0.985768	0	1	-360	360;
	3466	10202	0.005208	0.208268	0	0	0	0	1.004858	0	1	-360	360;
	3466	10201	0.005208	0.208268	0	0	0	0	1.004858	0	1	-360	360;
	2735	12582	0.005297	0.211798	0	0	0	0	1.001039	0	1	-360	360;
	2735	12583	0.005297	0.211798	0	0	0	0	1.001039	0	1	-360	360;
	2735	7428	0.005297	0.211798	0	0	0	0	1.001039	0	1	-360	360;
	2735	12585	0.005297	0.211798	0	0	0	0	0.995164	0	1	-360	360;
	2735	12578	0.005297	0.211798	0	0	0	0	0.995164	0	1	-360	360;
	2735	7361	0.005297	0.211798	0	0	0	0	0.995164	0	1	-360	360;
	2735	12580	0.005297	0.211798	0	0	0	0	0.995164	0	1	-360	360;
	2735	12581	0.005297	0.211798	0	0	0	0	0.995164	0	1	-360	360;
	2735	12575	0.005297	0.211798	0	0	0	0	0.995164	0	1	-360	360;
	2735	12576	0.005297	0.211798	0	0	0	0	0.995164	0	1	-360	360;
	2735	6481	0.005297	0.211798	0	0	0	0	0.995164	0	1	-360	360;
	2735	13469	0.005297	0.211798	0	0	0	0	0.995164	0	1	-360	360;
	1086	10676	0.005208	0.208268	0	0	0	0	1.014931	0	1	-360	360;
	1086	10675	0.005208	0.208268	0	0	0	0	1.014931	0	1	-360	360;
	1086	10672	0.005208	0.208268	0	0	0	0	1.014931	0	1	-360	360;
	1086	10671	0.005208	0.208268	0	0	0	0	1.014931	0	1	-360	360;
	1086	10674	0.005208	0.208268	0	0	0	0	1.014931	0	1	-360	360;
	1086	898	0.005208	0.208268	0	0	0	0	1.014931	0	1	-360	360;
	4953	10669	0.005208	0.208268	0	0	0	0	1.007499	0	1	-360	360;
	4953	10668	0.005208	0.208268	0	0	0	0	1.007499	0	1	-360	360;
	4953	3571	0.005208	0.208268	0	0	0	0	1.007499	0	1	-360	360;
	7062	8425	0.019841	0.793403	0	0	0	0	1.003966	0	1	-360	360;
	10933	3569	0.005423	0.216852	0	0	0	0	0.996526	0	1	-360	360;
	10933	3570	0.005423	0.216852	0	0	0	0	0.987742	0	1	-360	360;
	10933	3575	0.005423	0.216852	0	0	0	0	0.987742	0	1	-360	360;
	10933	8435	0.005423	0.216852	0	0	0	0	0.987742	0	1	-360	360;
	10933	8427	0.005423	0.216852	0	0	0	0	0.987742	0	1	-360	360;
	10933	8430	0.005423	0.216852	0	0	0	0	0.987742	0	1	-360	360;
	12094	8397	0.005208	0.208268	0	0	0	0	0.980186	0	1	-360	360;
	12094	8402	0.005208	0.208268	0	0	0	0	0.980186	0	1	-360	360;
	2953	7240	0.005208	0.208268	0	0	0	0	1.017742	0	1	-360	360;
	2953	9220	0.005208	0.208268	0	0	0	0	1.017742	0	1	-360	360;
	2953	9219	0.005208	0.208268	0	0	0	0	1.017742	0	1	-360	360;
	2953	10007	0.005208	0.208268	0	0	0	0	1.017742	0	1	-360	360;
	2953	7475	0.005208	0.208268	0	0	0	0	1.017742	0	1	-360	360;
	5913	9224	0.005208	0.208268	0	0	0	0	1.01761	0	1	-360	360;
	5913	10068	0.005208	0.208268	0	0	0	0	1.01761	0	1	-360	360;
	5913	9222	0.005208	0.208268	0	0	0	0	1.01761	0	1	-360	360;
	5913	1273	0.005208	0.208268	0	0	0	0	1.01761	0	1	-360	360;
	5913	9213	0.005208	0.208268	0	0	0	0	1.01761	0	1	-360	360;
	6061	3410	0.005208	0.208268	0	0	0	0	1.005176	0	1	-360	360;
	6061	12200	0.005208	0.208268	0	0	0	0	1.005176	0	1	-360	360;
	9916	13594	0.005208	0.208268	0	0	0	0	1.017597	0	1	-360	360;
	9916	13597	0.005208	0.208268	0	0	0	0	1.017597	0	1	-360	360;
	9916	7547	0.005208	0.208268	0	0	0	0	1.017597	0	1	-360	360;
	6839	13601	0.005208	0.208268	0	0	0	0	1.004984	0	1	-360	360;
	9289	2157	0.005208	0.208268	0	0	0	0	1.003139	0	1	-360	360;
	9289	13609	0.005208	0.208268	0	0	0	0	1.003139	0	1	-360	360;
	13408	2148	0.005208	0.208268	0	0	0	0	1.007086	0	1	-360	360;
	13408	2149	0.005208	0.208268	0	0	0	0	1.007086	0	1	-360	360;
	13408	9696	0.005208	0.208268	0	0	0	0	1.007086	0	1	-360	360;
	13408	9695	0.005208	0.208268	0	0	0	0	1.007086	0	1	-360	360;
	13408	9698	0.005208	0.208268	0	0	0	0	1.007086	0	1	-360	360;
	13408	9697	0.005208	0.208268	0	0	0	0	1.007086	0	1	-360	360;
	7811	9700	0.005208	0.208268	0	0	0	0	1.013653	0	1	-360	360;
	3376	9699	0.005208	0.208268	0	0	0	0	0.999657	0	1	-360	360;
	3376	9702	0.005208	0.208268	0	0	0	0	0.999657	0	1	-360	360;
	13583	7179	0.002201	0.088001	0	0	0	0	1.001304	0	1	-360	360;
	13583	7053	0.002201	0.088001	0	0	0	0	0.994637	0	1	-360	360;
	13583	7054	0.002201	0.088001	0	0	0	0	0.994637	0	1	-360	360;
	13583	515	0.002201	0.088001	0	0	0	0	0.994637	0	1	-360	360;
	13583	3078	0.002201	0.088001	0	0	0	0	0.994637	0	1	-360	360;
	13583	513	0.002201	0.088001	0	0	0	0	0.994637	0	1	-360	360;
	13583	512	0.002201	0.088001	0	0	0	0	0.994637	0	1	-360	360;
	13583	511	0.002201	0.088001	0	0	0	0	0.994637	0	1	-360	360;
	4299	8274	0.003425	0.136943	0	0	0	0	1.017447	0	1	-360	360;
	4299	8275	0.003425	0.136943	0	0	0	0	1.017447	0	1	-360	360;
	4299	8276	0.003425	0.136943	0	0	0	0	1.017447	0	1	-360	360;
	4299	9182	0.003425	0.136943	0	0	0	0	1.015837	0	1	-360	360;
	4299	8278	0.003425	0.136943	0	0	0	0	1.010601	0	1	-360	360;
	4299	1231	0.003425	0.136943	0	0	0	0	1.010601	0	1	-360	360;
	4299	1230	0.003425	0.136943	0	0	0	0	1.010601	0	1	-360	360;
	4299	1232	0.003425	0.136943	0	0	0	0	1.010601	0	1	-360	360;
	1582	12127	0.005208	0.208268	0	0	0	0	1.009832	0	1	-360	360;
	1582	12126	0.005208	0.208268	0	0	0	0	1.009832	0	1	-360	360;
	994	4870	0.005208	0.208268	0	0	0	0	1.013612	0	1	-360	360;
	994	12124	0.005208	0.208268	0	0	0	0	1.013612	0	1	-360	360;
	4114	12123	0.046296	1.851273	0	0	0	0	1.008844	0	1	-360	360;
	4114	12122	0.046296	1.851273	0	0	0	0	1.003788	0	1	-360	360;
	9263	12121	0.005208	0.208268	0	0	0	0	1.014036	0	1	-360	360;
	9263	3351	0.005208	0.208268	0	0	0	0	1.014036	0	1	-360	360;
	9263	5076	0.005208	0.208268	0	0	0	0	1.014036	0	1	-360	360;
	9263	5077	0.005208	0.208268	0	0	0	0	1.014036	0	1	-360	360;
	9263	5078	0.005208	0.208268	0	0	0	0	1.014036	0	1	-360	360;
	1206	5071	0.005208	0.208268	0	0	0	0	0.998104	0	1	-360	360;
	1206	5072	0.005208	0.208268	0	0	0	0	0.998104	0	1	-360	360;
	1206	5073	0.005208	0.208268	0	0	0	0	0.998104	0	1	-360	360;
	1206	5074	0.005208	0.208268	0	0	0	0	0.998104	0	1	-360	360;
	1206	5069	0.005208	0.208268	0	0	0	0	0.998104	0	1	-360	360;
	5709	5070	0.012821	0.51266	0	0	0	0	1.007739	0	1	-360	360;
	5709	7088	0.012821	0.51266	0	0	0	0	1.002444	0	1	-360	360;
	7157	9706	0.005208	0.208268	0	0	0	0	1.009513	0	1	-360	360;
	7157	1815	0.005208	0.208268	0	0	0	0	1.009513	0	1	-360	360;
	7157	7087	0.005208	0.208268	0	0	0	0	1.009513	0	1	-360	360;
	7157	7084	0.005208	0.208268	0	0	0	0	1.009513	0	1	-360	360;
	4528	7085	0.005208	0.208268	0	0	0	0	1.015825	0	1	-360	360;
	4528	7082	0.005208	0.208268	0	0	0	0	1.015825	0	1	-360	360;
	4528	7083	0.005208	0.208268	0	0	0	0	1.015825	0	1	-360	360;
	4528	7080	0.005208	0.208268	0	0	0	0	1.015825	0	1	-360	360;
	4528	7081	0.005208	0.208268	0	0	0	0	1.015825	0	1	-360	360;
	4528	528	0.005208	0.208268	0	0	0	0	1.015825	0	1	-360	360;
	6212	2137	0.000212	0.008494	0	0	0	0	1.001886	0	1	-360	360;
	6212	2136	0.000212	0.008494	0	0	0	0	1.001886	0	1	-360	360;
	6212	2135	0.000212	0.008494	0	0	0	0	1.001886	0	1	-360	360;
	6212	2138	0.000212	0.008494	0	0	0	0	1.001886	0	1	-360	360;
	1127	2540	0.002119	0.084719	0	0	0	0	1.028503	0	1	-360	360;
	1127	2577	0.002119	0.084719	0	0	0	0	1.028503	0	1	-360	360;
	1127	2543	0.002119	0.084719	0	0	0	0	1.028503	0	1	-360	360;
	1127	2571	0.002119	0.084719	0	0	0	0	1.028503	0	1	-360	360;
	1127	2534	0.002119	0.084719	0	0	0	0	1.028503	0	1	-360	360;
	1127	10172	0.002119	0.084719	0	0	0	0	1.028503	0	1	-360	360;
	1127	10171	0.002119	0.084719	0	0	0	0	1.028503	0	1	-360	360;
	1127	10174	0.002119	0.084719	0	0	0	0	1.028503	0	1	-360	360;
	1127	10173	0.002119	0.084719	0	0	0	0	1.028503	0	1	-360	360;
	9274	12656	0.005208	0.208268	0	0	0	0	1.023793	0	1	-360	360;
	9274	10210	0.005208	0.208268	0	0	0	0	1.023793	0	1	-360	360;
	9274	6631	0.005208	0.208268	0	0	0	0	1.023793	0	1	-360	360;
	6699	10232	0.005208	0.208268	0	0	0	0	1.007671	0	1	-360	360;
	6699	6621	0.005208	0.208268	0	0	0	0	1.007671	0	1	-360	360;
	4121	6622	0.003876	0.15499	0	0	0	0	1.007253	0	1	-360	360;
	4121	76	0.003876	0.15499	0	0	0	0	1.001535	0	1	-360	360;
	4121	75	0.003876	0.15499	0	0	0	0	1.001535	0	1	-360	360;
	4121	74	0.003876	0.15499	0	0	0	0	1.001535	0	1	-360	360;
	3662	438	0.005208	0.208268	0	0	0	0	1.007811	0	1	-360	360;
	3662	8039	0.005208	0.208268	0	0	0	0	1.007811	0	1	-360	360;
	3662	79	0.005208	0.208268	0	0	0	0	1.007811	0	1	-360	360;
	7321	78	0.003788	0.151468	0	0	0	0	1.004692	0	1	-360	360;
	7321	447	0.003788	0.151468	0	0	0	0	0.997453	0	1	-360	360;
	7321	72	0.003788	0.151468	0	0	0	0	0.997453	0	1	-360	360;
	7321	71	0.003788	0.151468	0	0	0	0	0.997453	0	1	-360	360;
	2432	9681	0.000212	0.008494	0	0	0	0	1.01356	0	1	-360	360;
	13529	9682	0.000212	0.008494	0	0	0	0	1.013518	0	1	-360	360;
	11837	7995	0.005208	0.208268	0	0	0	0	1.008199	0	1	-360	360;
	11837	8386	0.005208	0.208268	0	0	0	0	1.008199	0	1	-360	360;
	11837	8000	0.005208	0.208268	0	0	0	0	1.008199	0	1	-360	360;
	11837	8001	0.005208	0.208268	0	0	0	0	1.008199	0	1	-360	360;
	4220	7998	0.060976	2.438262	0	0	0	0	1.008285	0	1	-360	360;
	4220	8442	0.060976	2.438262	0	0	0	0	1.003185	0	1	-360	360;
	13028	7992	0.005208	0.208268	0	0	0	0	1.022422	0	1	-360	360;
	13028	7993	0.005208	0.208268	0	0	0	0	1.022422	0	1	-360	360;
	4907	1440	0.004621	0.184785	0	0	0	0	1.007556	0	1	-360	360;
	4907	1439	0.004621	0.184785	0	0	0	0	1.002847	0	1	-360	360;
	4907	1442	0.004621	0.184785	0	0	0	0	1.002847	0	1	-360	360;
	3473	1441	0.005208	0.208268	0	0	0	0	1.018212	0	1	-360	360;
	3473	1436	0.005208	0.208268	0	0	0	0	1.018212	0	1	-360	360;
	3473	1435	0.005208	0.208268	0	0	0	0	1.018212	0	1	-360	360;
	5395	1438	0.005208	0.208268	0	0	0	0	1.008298	0	1	-360	360;
	5395	1437	0.005208	0.208268	0	0	0	0	1.008298	0	1	-360	360;
	5395	1434	0.005208	0.208268	0	0	0	0	1.008298	0	1	-360	360;
	5395	1433	0.005208	0.208268	0	0	0	0	1.008298	0	1	-360	360;
	5395	7505	0.005208	0.208268	0	0	0	0	1.008298	0	1	-360	360;
	5395	4709	0.005208	0.208268	0	0	0	0	1.008298	0	1	-360	360;
	10204	7507	0.005208	0.208268	0	0	0	0	1.000664	0	1	-360	360;
	10204	11723	0.005208	0.208268	0	0	0	0	1.000664	0	1	-360	360;
	10204	7501	0.005208	0.208268	0	0	0	0	1.000664	0	1	-360	360;
	10301	7502	0.005208	0.208268	0	0	0	0	1.01492	0	1	-360	360;
	10301	9028	0.005208	0.208268	0	0	0	0	1.01492	0	1	-360	360;
	10301	7504	0.005208	0.208268	0	0	0	0	1.01492	0	1	-360	360;
	10294	4733	0.005208	0.208268	0	0	0	0	0.972052	0	1	-360	360;
	10294	4738	0.005208	0.208268	0	0	0	0	0.972052	0	1	-360	360;
	9955	987	0.005208	0.208268	0	0	0	0	1.010165	0	1	-360	360;
	9955	986	0.005208	0.208268	0	0	0	0	1.010165	0	1	-360	360;
	9955	985	0.005208	0.208268	0	0	0	0	1.010165	0	1	-360	360;
	9955	984	0.005208	0.208268	0	0	0	0	1.010165	0	1	-360	360;
	3527	983	0.005208	0.208268	0	0	0	0	1.001827	0	1	-360	360;
	3527	982	0.005208	0.208268	0	0	0	0	1.001827	0	1	-360	360;
	3527	981	0.005208	0.208268	0	0	0	0	1.001827	0	1	-360	360;
	1122	980	0.005208	0.208268	0	0	0	0	0.997809	0	1	-360	360;
	178	3558	0.000212	0.008494	0	0	0	0	1.020691	0	1	-360	360;
	7008	3557	0.000212	0.008494	0	0	0	0	1.014465	0	1	-360	360;
	4822	6879	0.000212	0.008494	0	0	0	0	1.014326	0	1	-360	360;
	6572	6880	0.000212	0.008494	0	0	0	0	1.014284	0	1	-360	360;
	7985	8956	0.005208	0.208268	0	0	0	0	0.997657	0	1	-360	360;
	7985	13363	0.005208	0.208268	0	0	0	0	0.997657	0	1	-360	360;
	1422	8954	0.005208	0.208268	0	0	0	0	1.010148	0	1	-360	360;
	1422	8955	0.005208	0.208268	0	0	0	0	1.010148	0	1	-360	360;
	1422	8952	0.005208	0.208268	0	0	0	0	1.010148	0	1	-360	360;
	1422	8953	0.005208	0.208268	0	0	0	0	1.010148	0	1	-360	360;
	1422	8962	0.005208	0.208268	0	0	0	0	1.010148	0	1	-360	360;
	9411	8963	0.005208	0.208268	0	0	0	0	1.007693	0	1	-360	360;
	9411	2329	0.005208	0.208268	0	0	0	0	1.007693	0	1	-360	360;
	9411	2328	0.005208	0.208268	0	0	0	0	1.007693	0	1	-360	360;
	9411	2331	0.005208	0.208268	0	0	0	0	1.007693	0	1	-360	360;
	9411	2330	0.005208	0.208268	0	0	0	0	1.007693	0	1	-360	360;
	4999	323	0.000212	0.008494	0	0	0	0	0.995203	0	1	-360	360;
	4999	1932	0.000212	0.008494	0	0	0	0	0.995203	0	1	-360	360;
	3737	2333	0.005208	0.208268	0	0	0	0	1.008578	0	1	-360	360;
	3737	1077	0.005208	0.208268	0	0	0	0	1.008578	0	1	-360	360;
	3737	2325	0.005208	0.208268	0	0	0	0	1.008578	0	1	-360	360;
	3737	2324	0.005208	0.208268	0	0	0	0	1.008578	0	1	-360	360;
	4880	11604	0.005208	0.208268	0	0	0	0	0.999885	0	1	-360	360;
	4880	11603	0.005208	0.208268	0	0	0	0	0.999885	0	1	-360	360;
	7698	11606	0.005208	0.208268	0	0	0	0	1.007227	0	1	-360	360;
	7698	11605	0.005208	0.208268	0	0	0	0	1.007227	0	1	-360	360;
	7698	11600	0.005208	0.208268	0	0	0	0	1.007227	0	1	-360	360;
	326	13393	0.005208	0.208268	0	0	0	0	1.005811	0	1	-360	360;
	326	13395	0.005208	0.208268	0	0	0	0	1.005811	0	1	-360	360;
	326	13394	0.005208	0.208268	0	0	0	0	1.005811	0	1	-360	360;
	326	13391	0.005208	0.208268	0	0	0	0	1.005811	0	1	-360	360;
	326	13390	0.005208	0.208268	0	0	0	0	1.005811	0	1	-360	360;
	326	10621	0.005208	0.208268	0	0	0	0	1.005811	0	1	-360	360;
	326	10625	0.005208	0.208268	0	0	0	0	1.005811	0	1	-360	360;
	326	6298	0.005208	0.208268	0	0	0	0	1.005811	0	1	-360	360;
	326	6299	0.005208	0.208268	0	0	0	0	1.005811	0	1	-360	360;
	326	8942	0.005208	0.208268	0	0	0	0	1.005811	0	1	-360	360;
	11931	448	0.008224	0.328845	0	0	0	0	1.012901	0	1	-360	360;
	11931	441	0.008224	0.328845	0	0	0	0	1.007766	0	1	-360	360;
	2104	442	0.005208	0.208268	0	0	0	0	1.013881	0	1	-360	360;
	2104	4539	0.005208	0.208268	0	0	0	0	1.013881	0	1	-360	360;
	4848	4540	0.005208	0.208268	0	0	0	0	0.994225	0	1	-360	360;
	4848	11095	0.005208	0.208268	0	0	0	0	0.994225	0	1	-360	360;
	13214	12912	0.005208	0.208268	0	0	0	0	1.009807	0	1	-360	360;
	13214	2411	0.005208	0.208268	0	0	0	0	1.009807	0	1	-360	360;
	13214	12910	0.005208	0.208268	0	0	0	0	1.009807	0	1	-360	360;
	13214	9184	0.005208	0.208268	0	0	0	0	1.009807	0	1	-360	360;
	13214	12908	0.005208	0.208268	0	0	0	0	1.009807	0	1	-360	360;
	13214	12907	0.005208	0.208268	0	0	0	0	1.009807	0	1	-360	360;
	13214	12906	0.005208	0.208268	0	0	0	0	1.009807	0	1	-360	360;
	13214	12917	0.005208	0.208268	0	0	0	0	1.009807	0	1	-360	360;
	13214	12916	0.005208	0.208268	0	0	0	0	1.009807	0	1	-360	360;
	13214	5831	0.005208	0.208268	0	0	0	0	1.009807	0	1	-360	360;
	13214	5832	0.005208	0.208268	0	0	0	0	1.009807	0	1	-360	360;
	2348	4011	0.005208	0.208268	0	0	0	0	1.004739	0	1	-360	360;
	2348	10898	0.005208	0.208268	0	0	0	0	1.004739	0	1	-360	360;
	7867	10643	0.002119	0.084719	0	0	0	0	1.020131	0	1	-360	360;
	9172	7599	0.002119	0.084719	0	0	0	0	1.020177	0	1	-360	360;
	10335	7598	0.002119	0.084719	0	0	0	0	1.028958	0	1	-360	360;
	5899	9781	0.005208	0.208268	0	0	0	0	1.029771	0	1	-360	360;
	5899	4607	0.005208	0.208268	0	0	0	0	1.029771	0	1	-360	360;
	12622	1966	0.002119	0.084719	0	0	0	0	1.031766	0	1	-360	360;
	12622	1967	0.002119	0.084719	0	0	0	0	1.031766	0	1	-360	360;
	12622	6139	0.000212	0.008494	0	0	0	0	1.059942	0	1	-360	360;
	12622	6140	0.000212	0.008494	0	0	0	0	1.059942	0	1	-360	360;
	12622	6138	0.000212	0.008494	0	0	0	0	1.059942	0	1	-360	360;
	7765	3141	0.005208	0.208268	0	0	0	0	1.008128	0	1	-360	360;
	7765	3142	0.005208	0.208268	0	0	0	0	1.008128	0	1	-360	360;
	9820	3143	0.005208	0.208268	0	0	0	0	1.008223	0	1	-360	360;
	622	6663	0.002119	0.084719	0	0	0	0	1.004203	0	1	-360	360;
	622	8439	0.002119	0.084719	0	0	0	0	1.004203	0	1	-360	360;
	622	1412	0.002119	0.084719	0	0	0	0	1.004203	0	1	-360	360;
	622	1413	0.002119	0.084719	0	0	0	0	1.004203	0	1	-360	360;
	622	1410	0.002119	0.084719	0	0	0	0	1.004203	0	1	-360	360;
	622	1411	0.002119	0.084719	0	0	0	0	1.004203	0	1	-360	360;
	622	1416	0.002119	0.084719	0	0	0	0	1.004203	0	1	-360	360;
	9506	1417	0.002119	0.084719	0	0	0	0	1.006993	0	1	-360	360;
	9506	1414	0.002119	0.084719	0	0	0	0	1.006993	0	1	-360	360;
	9506	1415	0.002119	0.084719	0	0	0	0	1.006993	0	1	-360	360;
	9506	1407	0.002119	0.084719	0	0	0	0	1.006993	0	1	-360	360;
	9506	1408	0.002119	0.084719	0	0	0	0	1.006993	0	1	-360	360;
	9506	7333	0.002119	0.084719	0	0	0	0	1.006993	0	1	-360	360;
	9506	7966	0.002119	0.084719	0	0	0	0	1.006993	0	1	-360	360;
	9506	7965	0.002119	0.084719	0	0	0	0	1.006993	0	1	-360	360;
	6362	1521	0.002119	0.084719	0	0	0	0	1.021091	0	1	-360	360;
	7733	1517	0.002119	0.084719	0	0	0	0	1.017918	0	1	-360	360;
	7733	1518	0.002119	0.084719	0	0	0	0	1.017918	0	1	-360	360;
	3193	1519	0.002119	0.084719	0	0	0	0	1.017911	0	1	-360	360;
	3193	1520	0.002119	0.084719	0	0	0	0	1.017911	0	1	-360	360;
	5496	3641	0.005208	0.208268	0	0	0	0	1.01792	0	1	-360	360;
	5496	3642	0.005208	0.208268	0	0	0	0	1.01792	0	1	-360	360;
	8653	10311	0.005208	0.208268	0	0	0	0	1.015224	0	1	-360	360;
	8653	10742	0.005208	0.208268	0	0	0	0	1.015224	0	1	-360	360;
	272	8550	0.002119	0.084719	0	0	0	0	1.025836	0	1	-360	360;
	272	8549	0.002119	0.084719	0	0	0	0	1.025836	0	1	-360	360;
	4883	8548	0.002119	0.084719	0	0	0	0	1.025538	0	1	-360	360;
	4883	8547	0.002119	0.084719	0	0	0	0	1.025538	0	1	-360	360;
	1467	8546	0.002119	0.084719	0	0	0	0	1.028159	0	1	-360	360;
	1467	8545	0.002119	0.084719	0	0	0	0	1.028159	0	1	-360	360;
	5005	8544	0.002119	0.084719	0	0	0	0	1.028094	0	1	-360	360;
	5005	9393	0.002119	0.084719	0	0	0	0	1.028094	0	1	-360	360;
	10000	12790	0.002119	0.084719	0	0	0	0	1.006418	0	1	-360	360;
	10000	12791	0.002119	0.084719	0	0	0	0	1.006418	0	1	-360	360;
	10000	12788	0.002119	0.084719	0	0	0	0	1.006418	0	1	-360	360;
	10000	12789	0.002119	0.084719	0	0	0	0	1.006418	0	1	-360	360;
	5569	11212	0.005208	0.208268	0	0	0	0	1.011493	0	1	-360	360;
	8176	11211	0.005208	0.208268	0	0	0	0	1.025506	0	1	-360	360;
	644	9890	0.002119	0.084719	0	0	0	0	1.025242	0	1	-360	360;
	2726	4144	0.002119	0.084719	0	0	0	0	1.025287	0	1	-360	360;
	2257	4143	0.002119	0.084719	0	0	0	0	1.02407	0	1	-360	360;
	2257	2734	0.002119	0.084719	0	0	0	0	1.02407	0	1	-360	360;
	2257	4141	0.002119	0.084719	0	0	0	0	1.02407	0	1	-360	360;
	2257	4140	0.002119	0.084719	0	0	0	0	1.02407	0	1	-360	360;
	4997	4139	0.002119	0.084719	0	0	0	0	1.023728	0	1	-360	360;
	4997	4138	0.002119	0.084719	0	0	0	0	1.023728	0	1	-360	360;
	4997	4137	0.002119	0.084719	0	0	0	0	1.023728	0	1	-360	360;
	10367	4136	0.002119	0.084719	0	0	0	0	1.013659	0	1	-360	360;
	10367	11803	0.002119	0.084719	0	0	0	0	1.013659	0	1	-360	360;
	9178	10753	0.002119	0.084719	0	0	0	0	1.013627	0	1	-360	360;
	9178	10754	0.002119	0.084719	0	0	0	0	1.013627	0	1	-360	360;
	9415	10755	0.002119	0.084719	0	0	0	0	1.036331	0	1	-360	360;
	9415	10756	0.002119	0.084719	0	0	0	0	1.036331	0	1	-360	360;
	9415	10337	0.002119	0.084719	0	0	0	0	1.036331	0	1	-360	360;
	9415	10750	0.002119	0.084719	0	0	0	0	1.036331	0	1	-360	360;
	5344	10343	0.002119	0.084719	0	0	0	0	1.034204	0	1	-360	360;
	5344	10752	0.002119	0.084719	0	0	0	0	1.034204	0	1	-360	360;
	11902	10328	0.002119	0.084719	0	0	0	0	1.031833	0	1	-360	360;
	11902	10747	0.002119	0.084719	0	0	0	0	1.031833	0	1	-360	360;
	11902	4665	0.002119	0.084719	0	0	0	0	1.031833	0	1	-360	360;
	11902	4664	0.002119	0.084719	0	0	0	0	1.031833	0	1	-360	360;
	722	1057	0.002119	0.084719	0	0	0	0	1.010944	0	1	-360	360;
	722	4666	0.002119	0.084719	0	0	0	0	1.010944	0	1	-360	360;
	8960	4946	0.005208	0.208268	0	0	0	0	1.012933	0	1	-360	360;
	8960	4945	0.005208	0.208268	0	0	0	0	1.012933	0	1	-360	360;
	478	4663	0.002119	0.084719	0	0	0	0	1.025231	0	1	-360	360;
	478	4662	0.002119	0.084719	0	0	0	0	1.025231	0	1	-360	360;
	945	4943	0.005208	0.208268	0	0	0	0	1.013183	0	1	-360	360;
	945	4942	0.005208	0.208268	0	0	0	0	1.013183	0	1	-360	360;
	6825	11219	0.002119	0.084719	0	0	0	0	1.008652	0	1	-360	360;
	12961	8806	0.002119	0.084719	0	0	0	0	1.01068	0	1	-360	360;
	12961	8807	0.002119	0.084719	0	0	0	0	1.01068	0	1	-360	360;
	12961	12519	0.002119	0.084719	0	0	0	0	1.01068	0	1	-360	360;
	12961	8809	0.002119	0.084719	0	0	0	0	1.01068	0	1	-360	360;
	12406	9279	0.002119	0.084719	0	0	0	0	1.030019	0	1	-360	360;
	8314	11220	0.002119	0.084719	0	0	0	0	1.029971	0	1	-360	360;
	9863	9280	0.005208	0.208268	0	0	0	0	1.004413	0	1	-360	360;
	9863	9075	0.005208	0.208268	0	0	0	0	1.004413	0	1	-360	360;
	3932	9088	0.005208	0.208268	0	0	0	0	1.016311	0	1	-360	360;
	3932	11225	0.005208	0.208268	0	0	0	0	1.016311	0	1	-360	360;
	1779	6405	0.005208	0.208268	0	0	0	0	1.011415	0	1	-360	360;
	1779	6408	0.005208	0.208268	0	0	0	0	1.011415	0	1	-360	360;
	8529	6407	0.005208	0.208268	0	0	0	0	1.023825	0	1	-360	360;
	8529	6402	0.005208	0.208268	0	0	0	0	1.023825	0	1	-360	360;
	5951	6401	0.005208	0.208268	0	0	0	0	1.023807	0	1	-360	360;
	5951	6404	0.005208	0.208268	0	0	0	0	1.023807	0	1	-360	360;
	7322	8839	0.002119	0.084719	0	0	0	0	1.02219	0	1	-360	360;
	7322	10748	0.002119	0.084719	0	0	0	0	1.02219	0	1	-360	360;
	7322	3647	0.002119	0.084719	0	0	0	0	1.02219	0	1	-360	360;
	7322	10253	0.002119	0.084719	0	0	0	0	1.02219	0	1	-360	360;
	7322	10254	0.002119	0.084719	0	0	0	0	1.02219	0	1	-360	360;
	6386	9785	0.002119	0.084719	0	0	0	0	1.013786	0	1	-360	360;
	6386	4652	0.002119	0.084719	0	0	0	0	1.013786	0	1	-360	360;
	4756	9751	0.005208	0.208268	0	0	0	0	1.016334	0	1	-360	360;
	1128	9753	0.005208	0.208268	0	0	0	0	1.016322	0	1	-360	360;
	6012	3655	0.002119	0.084719	0	0	0	0	1.012575	0	1	-360	360;
	6012	3186	0.002119	0.084719	0	0	0	0	1.012575	0	1	-360	360;
	6570	3137	0.002119	0.084719	0	0	0	0	1.021682	0	1	-360	360;
	6570	3136	0.002119	0.084719	0	0	0	0	1.021682	0	1	-360	360;
	6570	3135	0.002119	0.084719	0	0	0	0	1.021682	0	1	-360	360;
	6570	3134	0.002119	0.084719	0	0	0	0	1.021682	0	1	-360	360;
	6570	6048	0.002119	0.084719	0	0	0	0	1.021682	0	1	-360	360;
	2113	3656	0.002119	0.084719	0	0	0	0	1.013632	0	1	-360	360;
	2113	3651	0.002119	0.084719	0	0	0	0	1.013632	0	1	-360	360;
	2113	8885	0.002119	0.084719	0	0	0	0	1.013632	0	1	-360	360;
	12866	12500	0.005208	0.208268	0	0	0	0	1.029586	0	1	-360	360;
	12866	6976	0.005208	0.208268	0	0	0	0	1.029586	0	1	-360	360;
	9948	10269	0.002119	0.084719	0	0	0	0	1.029404	0	1	-360	360;
	9948	10270	0.002119	0.084719	0	0	0	0	1.029404	0	1	-360	360;
	10965	10267	0.002119	0.084719	0	0	0	0	1.033807	0	1	-360	360;
	10965	10268	0.002119	0.084719	0	0	0	0	1.033807	0	1	-360	360;
	1691	2179	0.002119	0.084719	0	0	0	0	1.015569	0	1	-360	360;
	1691	2178	0.002119	0.084719	0	0	0	0	1.015569	0	1	-360	360;
	1691	2177	0.002119	0.084719	0	0	0	0	1.015569	0	1	-360	360;
	1691	2176	0.002119	0.084719	0	0	0	0	1.015569	0	1	-360	360;
	12622	3094	0.000531	0.021225	0	0	0	0	1.037731	0	1	-360	360;
	12622	3096	0.000531	0.021225	0	0	0	0	1.037731	0	1	-360	360;
	866	3093	0.000531	0.021225	0	0	0	0	1.027871	0	1	-360	360;
	866	3101	0.000531	0.021225	0	0	0	0	1.027871	0	1	-360	360;
	3429	3100	0.000531	0.021225	0	0	0	0	1.027928	0	1	-360	360;
	3429	13526	0.000531	0.021225	0	0	0	0	1.027928	0	1	-360	360;
	4472	3582	0.005208	0.208268	0	0	0	0	1.015316	0	1	-360	360;
	4472	3580	0.005208	0.208268	0	0	0	0	1.015316	0	1	-360	360;
	8588	12093	0.002119	0.084719	0	0	0	0	1.01369	0	1	-360	360;
	8588	10355	0.002119	0.084719	0	0	0	0	1.01369	0	1	-360	360;
	2750	13397	0.002119	0.084719	0	0	0	0	1.013791	0	1	-360	360;
	2750	13398	0.002119	0.084719	0	0	0	0	1.013791	0	1	-360	360;
	2750	13399	0.002119	0.084719	0	0	0	0	1.013791	0	1	-360	360;
	2750	13388	0.002119	0.084719	0	0	0	0	1.013791	0	1	-360	360;
	2750	13389	0.002119	0.084719	0	0	0	0	1.013791	0	1	-360	360;
	1425	6764	0.002119	0.084719	0	0	0	0	0.996464	0	1	-360	360;
	12051	7526	0.002119	0.084719	0	0	0	0	1.00607	0	1	-360	360;
	12051	7055	0.002119	0.084719	0	0	0	0	1.00607	0	1	-360	360;
	12051	7528	0.002119	0.084719	0	0	0	0	1.00607	0	1	-360	360;
	12051	7523	0.002119	0.084719	0	0	0	0	1.00607	0	1	-360	360;
	12051	7524	0.002119	0.084719	0	0	0	0	1.00607	0	1	-360	360;
	462	6765	0.002119	0.084719	0	0	0	0	0.995797	0	1	-360	360;
	2201	6422	0.000531	0.021225	0	0	0	0	1.028549	0	1	-360	360;
	2201	6417	0.000531	0.021225	0	0	0	0	1.028549	0	1	-360	360;
	2201	6418	0.000531	0.021225	0	0	0	0	1.028549	0	1	-360	360;
	4372	224	0.002119	0.084719	0	0	0	0	1.010236	0	1	-360	360;
	4372	225	0.002119	0.084719	0	0	0	0	1.010236	0	1	-360	360;
	4372	222	0.002119	0.084719	0	0	0	0	1.010236	0	1	-360	360;
	8454	223	0.002119	0.084719	0	0	0	0	1.018164	0	1	-360	360;
	8454	220	0.002119	0.084719	0	0	0	0	1.018164	0	1	-360	360;
	8454	221	0.002119	0.084719	0	0	0	0	1.018164	0	1	-360	360;
	8592	3061	0.005208	0.208268	0	0	0	0	0.997196	0	1	-360	360;
	8592	3062	0.005208	0.208268	0	0	0	0	0.997196	0	1	-360	360;
	8936	3063	0.005208	0.208268	0	0	0	0	1.021523	0	1	-360	360;
	8936	5660	0.005208	0.208268	0	0	0	0	1.021523	0	1	-360	360;
	6522	10195	0.005208	0.208268	0	0	0	0	1.00739	0	1	-360	360;
	6522	10194	0.005208	0.208268	0	0	0	0	1.00739	0	1	-360	360;
	7527	2440	0.005208	0.208268	0	0	0	0	1.029311	0	1	-360	360;
	7527	10192	0.005208	0.208268	0	0	0	0	1.029311	0	1	-360	360;
	3081	10198	0.005208	0.208268	0	0	0	0	1.00963	0	1	-360	360;
	3081	10197	0.005208	0.208268	0	0	0	0	1.00963	0	1	-360	360;
	10271	7628	0.002119	0.084719	0	0	0	0	1.004824	0	1	-360	360;
	10271	8017	0.002119	0.084719	0	0	0	0	1.004824	0	1	-360	360;
	10271	8021	0.002119	0.084719	0	0	0	0	1.004824	0	1	-360	360;
	10271	8022	0.002119	0.084719	0	0	0	0	1.004824	0	1	-360	360;
	10271	104	0.002119	0.084719	0	0	0	0	1.004824	0	1	-360	360;
	2602	4572	0.005208	0.208268	0	0	0	0	1.00443	0	1	-360	360;
	2602	11633	0.005208	0.208268	0	0	0	0	1.00443	0	1	-360	360;
	11961	11632	0.005208	0.208268	0	0	0	0	1.010587	0	1	-360	360;
	11961	11635	0.005208	0.208268	0	0	0	0	1.010587	0	1	-360	360;
	3731	11634	0.005208	0.208268	0	0	0	0	1.010696	0	1	-360	360;
	3731	11631	0.005208	0.208268	0	0	0	0	1.010696	0	1	-360	360;
	10850	7672	0.002119	0.084719	0	0	0	0	1.002271	0	1	-360	360;
	10850	7671	0.002119	0.084719	0	0	0	0	1.002271	0	1	-360	360;
	10850	7670	0.002119	0.084719	0	0	0	0	1.002271	0	1	-360	360;
	3452	7676	0.002119	0.084719	0	0	0	0	0.996445	0	1	-360	360;
	3452	7675	0.002119	0.084719	0	0	0	0	0.996445	0	1	-360	360;
	23	11979	0.002119	0.084719	0	0	0	0	1.006948	0	1	-360	360;
	23	11978	0.002119	0.084719	0	0	0	0	1.006948	0	1	-360	360;
	23	3904	0.002119	0.084719	0	0	0	0	1.006948	0	1	-360	360;
	23	11980	0.002119	0.084719	0	0	0	0	1.006948	0	1	-360	360;
	23	3914	0.002119	0.084719	0	0	0	0	1.006948	0	1	-360	360;
	5257	4047	0.005208	0.208268	0	0	0	0	1.008334	0	1	-360	360;
	8044	11325	0.005208	0.208268	0	0	0	0	1.00304	0	1	-360	360;
	8044	4049	0.005208	0.208268	0	0	0	0	1.00304	0	1	-360	360;
	12457	4043	0.005208	0.208268	0	0	0	0	1.004713	0	1	-360	360;
	12457	4044	0.005208	0.208268	0	0	0	0	1.004713	0	1	-360	360;
	13600	8444	0.002119	0.084719	0	0	0	0	1.011145	0	1	-360	360;
	13600	8151	0.002119	0.084719	0	0	0	0	1.011145	0	1	-360	360;
	13600	8154	0.002119	0.084719	0	0	0	0	1.011145	0	1	-360	360;
	7069	8153	0.002119	0.084719	0	0	0	0	1.010552	0	1	-360	360;
	7069	8148	0.002119	0.084719	0	0	0	0	1.010552	0	1	-360	360;
	7069	8147	0.002119	0.084719	0	0	0	0	1.010552	0	1	-360	360;
	3744	1534	0.002119	0.084719	0	0	0	0	1.010912	0	1	-360	360;
	3744	8149	0.002119	0.084719	0	0	0	0	1.010912	0	1	-360	360;
	3744	8146	0.002119	0.084719	0	0	0	0	1.010912	0	1	-360	360;
	10354	8145	0.002119	0.084719	0	0	0	0	1.022406	0	1	-360	360;
	10354	1586	0.002119	0.084719	0	0	0	0	1.022406	0	1	-360	360;
	10354	1587	0.002119	0.084719	0	0	0	0	1.022406	0	1	-360	360;
	6971	10476	0.002119	0.084719	0	0	0	0	1.018442	0	1	-360	360;
	6971	10475	0.002119	0.084719	0	0	0	0	1.018442	0	1	-360	360;
	6971	10474	0.002119	0.084719	0	0	0	0	1.018442	0	1	-360	360;
	6971	8665	0.002119	0.084719	0	0	0	0	1.018442	0	1	-360	360;
	6971	10472	0.002119	0.084719	0	0	0	0	1.018442	0	1	-360	360;
	6971	8658	0.002119	0.084719	0	0	0	0	1.018442	0	1	-360	360;
	5773	10480	0.002119	0.084719	0	0	0	0	1.023459	0	1	-360	360;
	5773	10479	0.002119	0.084719	0	0	0	0	1.023459	0	1	-360	360;
	5773	3365	0.002119	0.084719	0	0	0	0	1.023459	0	1	-360	360;
	5773	3366	0.002119	0.084719	0	0	0	0	1.023459	0	1	-360	360;
	5773	3367	0.002119	0.084719	0	0	0	0	1.023459	0	1	-360	360;
	5773	3368	0.002119	0.084719	0	0	0	0	1.023459	0	1	-360	360;
	238	8604	0.002119	0.084719	0	0	0	0	1.010933	0	1	-360	360;
	238	8610	0.002119	0.084719	0	0	0	0	1.010933	0	1	-360	360;
	238	8609	0.002119	0.084719	0	0	0	0	1.010933	0	1	-360	360;
	9397	9232	0.002119	0.084719	0	0	0	0	1.033253	0	1	-360	360;
	9397	9233	0.002119	0.084719	0	0	0	0	1.033253	0	1	-360	360;
	9397	9230	0.002119	0.084719	0	0	0	0	1.033253	0	1	-360	360;
	9397	9231	0.002119	0.084719	0	0	0	0	1.033253	0	1	-360	360;
	9397	9228	0.002119	0.084719	0	0	0	0	1.033253	0	1	-360	360;
	9397	9229	0.002119	0.084719	0	0	0	0	1.033253	0	1	-360	360;
	8077	11924	0.002119	0.084719	0	0	0	0	1.023147	0	1	-360	360;
	8077	12758	0.002119	0.084719	0	0	0	0	1.023147	0	1	-360	360;
	8077	9237	0.002119	0.084719	0	0	0	0	1.023147	0	1	-360	360;
	8077	1335	0.002119	0.084719	0	0	0	0	1.023147	0	1	-360	360;
	8077	2651	0.002119	0.084719	0	0	0	0	1.023147	0	1	-360	360;
	8077	2650	0.002119	0.084719	0	0	0	0	1.023147	0	1	-360	360;
	9397	2653	0.002119	0.084719	0	0	0	0	1.033007	0	1	-360	360;
	9397	2652	0.002119	0.084719	0	0	0	0	1.033007	0	1	-360	360;
	9397	2655	0.002119	0.084719	0	0	0	0	1.033007	0	1	-360	360;
	9397	2654	0.002119	0.084719	0	0	0	0	1.033007	0	1	-360	360;
	9397	2657	0.002119	0.084719	0	0	0	0	1.033007	0	1	-360	360;
	9397	2656	0.002119	0.084719	0	0	0	0	1.033007	0	1	-360	360;
	2453	3611	0.000531	0.021225	0	0	0	0	1.026628	0	1	-360	360;
	2453	3610	0.000531	0.021225	0	0	0	0	1.026628	0	1	-360	360;
	2453	10764	0.000531	0.021225	0	0	0	0	1.026628	0	1	-360	360;
	3732	10765	0.000531	0.021225	0	0	0	0	1.021183	0	1	-360	360;
	3732	10762	0.000531	0.021225	0	0	0	0	1.021183	0	1	-360	360;
	3732	10763	0.000531	0.021225	0	0	0	0	1.021183	0	1	-360	360;
	2728	8731	0.002119	0.084719	0	0	0	0	1.016872	0	1	-360	360;
	2728	8732	0.002119	0.084719	0	0	0	0	1.016872	0	1	-360	360;
	2728	8733	0.002119	0.084719	0	0	0	0	1.016872	0	1	-360	360;
	2728	8734	0.002119	0.084719	0	0	0	0	1.016872	0	1	-360	360;
	2728	8735	0.002119	0.084719	0	0	0	0	1.016872	0	1	-360	360;
	10026	7228	0.005208	0.208268	0	0	0	0	1.009461	0	1	-360	360;
	10026	243	0.005208	0.208268	0	0	0	0	1.009461	0	1	-360	360;
	10149	8904	0.002119	0.084719	0	0	0	0	1.006052	0	1	-360	360;
	10149	8905	0.002119	0.084719	0	0	0	0	1.006052	0	1	-360	360;
	10149	3631	0.002119	0.084719	0	0	0	0	1.006052	0	1	-360	360;
	10149	8907	0.002119	0.084719	0	0	0	0	1.006052	0	1	-360	360;
	31	8908	0.002119	0.084719	0	0	0	0	1.009987	0	1	-360	360;
	31	8909	0.002119	0.084719	0	0	0	0	1.009987	0	1	-360	360;
	31	8910	0.002119	0.084719	0	0	0	0	1.009987	0	1	-360	360;
	31	8897	0.002119	0.084719	0	0	0	0	1.009987	0	1	-360	360;
	9736	8898	0.002119	0.084719	0	0	0	0	0.996492	0	1	-360	360;
	7609	3729	0.002119	0.084719	0	0	0	0	1.008807	0	1	-360	360;
	7609	3728	0.002119	0.084719	0	0	0	0	1.008807	0	1	-360	360;
	7869	12461	0.002119	0.084719	0	0	0	0	1.009947	0	1	-360	360;
	7869	12462	0.002119	0.084719	0	0	0	0	1.009947	0	1	-360	360;
	7869	12466	0.002119	0.084719	0	0	0	0	1.009947	0	1	-360	360;
	7869	12467	0.002119	0.084719	0	0	0	0	1.009947	0	1	-360	360;
	7869	12464	0.002119	0.084719	0	0	0	0	1.009947	0	1	-360	360;
	638	12465	0.002119	0.084719	0	0	0	0	1.008334	0	1	-360	360;
	638	12469	0.002119	0.084719	0	0	0	0	1.008334	0	1	-360	360;
	638	12470	0.002119	0.084719	0	0	0	0	1.008334	0	1	-360	360;
	638	5901	0.002119	0.084719	0	0	0	0	1.008334	0	1	-360	360;
	638	5900	0.002119	0.084719	0	0	0	0	1.008334	0	1	-360	360;
	11573	8173	0.002119	0.084719	0	0	0	0	1.00834	0	1	-360	360;
	11573	5902	0.002119	0.084719	0	0	0	0	1.00834	0	1	-360	360;
	11573	5898	0.002119	0.084719	0	0	0	0	1.00834	0	1	-360	360;
	11573	5897	0.002119	0.084719	0	0	0	0	1.00834	0	1	-360	360;
	11573	8162	0.002119	0.084719	0	0	0	0	1.00834	0	1	-360	360;
	7161	10329	0.002119	0.084719	0	0	0	0	1.021478	0	1	-360	360;
	7161	10324	0.002119	0.084719	0	0	0	0	1.021478	0	1	-360	360;
	7161	1801	0.002119	0.084719	0	0	0	0	1.021478	0	1	-360	360;
	7161	9761	0.002119	0.084719	0	0	0	0	1.021581	0	1	-360	360;
	7161	9762	0.002119	0.084719	0	0	0	0	1.021581	0	1	-360	360;
	7161	9763	0.002119	0.084719	0	0	0	0	1.021581	0	1	-360	360;
	7161	9764	0.002119	0.084719	0	0	0	0	1.021581	0	1	-360	360;
	7161	9757	0.002119	0.084719	0	0	0	0	1.021581	0	1	-360	360;
	7161	9758	0.002119	0.084719	0	0	0	0	1.021581	0	1	-360	360;
	5515	9759	0.002119	0.084719	0	0	0	0	1.020404	0	1	-360	360;
	5515	4145	0.002119	0.084719	0	0	0	0	1.020404	0	1	-360	360;
	5515	9754	0.002119	0.084719	0	0	0	0	1.020404	0	1	-360	360;
	5515	4470	0.002119	0.084719	0	0	0	0	1.020404	0	1	-360	360;
	5515	3122	0.002119	0.084719	0	0	0	0	1.020404	0	1	-360	360;
	5515	3121	0.002119	0.084719	0	0	0	0	1.020404	0	1	-360	360;
	5479	9856	0.002119	0.084719	0	0	0	0	1.010153	0	1	-360	360;
	582	4376	0.000531	0.021225	0	0	0	0	1.037212	0	1	-360	360;
	582	4379	0.000531	0.021225	0	0	0	0	1.037212	0	1	-360	360;
	582	4380	0.000531	0.021225	0	0	0	0	1.037212	0	1	-360	360;
	582	4378	0.000531	0.021225	0	0	0	0	1.037212	0	1	-360	360;
	582	3095	0.000531	0.021225	0	0	0	0	1.038292	0	1	-360	360;
	582	3098	0.000531	0.021225	0	0	0	0	1.038292	0	1	-360	360;
	582	3097	0.000531	0.021225	0	0	0	0	1.038292	0	1	-360	360;
	6850	9550	0.002119	0.084719	0	0	0	0	1.037103	0	1	-360	360;
	6850	9551	0.002119	0.084719	0	0	0	0	1.037103	0	1	-360	360;
	6850	3145	0.002119	0.084719	0	0	0	0	1.037103	0	1	-360	360;
	6850	9549	0.002119	0.084719	0	0	0	0	1.037103	0	1	-360	360;
	6850	9547	0.002119	0.084719	0	0	0	0	1.037103	0	1	-360	360;
	2716	4751	0.002119	0.084719	0	0	0	0	1.037889	0	1	-360	360;
	2716	1458	0.002119	0.084719	0	0	0	0	1.037889	0	1	-360	360;
	2716	4753	0.002119	0.084719	0	0	0	0	1.037889	0	1	-360	360;
	2716	1427	0.002119	0.084719	0	0	0	0	1.037889	0	1	-360	360;
	2716	4745	0.002119	0.084719	0	0	0	0	1.035865	0	1	-360	360;
	2716	11307	0.002119	0.084719	0	0	0	0	1.035865	0	1	-360	360;
	13230	11308	0.002119	0.084719	0	0	0	0	1.037898	0	1	-360	360;
	3921	11305	0.002119	0.084719	0	0	0	0	1.018989	0	1	-360	360;
	3921	849	0.002119	0.084719	0	0	0	0	1.018989	0	1	-360	360;
	3921	107	0.002119	0.084719	0	0	0	0	1.018989	0	1	-360	360;
	9450	110	0.002119	0.084719	0	0	0	0	1.032859	0	1	-360	360;
	6840	11032	0.000212	0.008494	0	0	0	0	1.041634	0	1	-360	360;
	6840	12173	0.000212	0.008494	0	0	0	0	1.041634	0	1	-360	360;
	748	11312	0.002119	0.084719	0	0	0	0	1.035018	0	1	-360	360;
	748	136	0.002119	0.084719	0	0	0	0	1.035018	0	1	-360	360;
	4092	10775	0.002119	0.084719	0	0	0	0	1.033659	0	1	-360	360;
	4092	3800	0.002119	0.084719	0	0	0	0	1.033659	0	1	-360	360;
	11254	1183	0.005208	0.208268	0	0	0	0	1.030671	0	1	-360	360;
	11254	1182	0.005208	0.208268	0	0	0	0	1.030671	0	1	-360	360;
	7543	11367	0.002119	0.084719	0	0	0	0	1.036532	0	1	-360	360;
	7543	11368	0.002119	0.084719	0	0	0	0	1.036532	0	1	-360	360;
	7543	11369	0.002119	0.084719	0	0	0	0	1.036532	0	1	-360	360;
	7543	11370	0.002119	0.084719	0	0	0	0	1.036532	0	1	-360	360;
	7543	11364	0.002119	0.084719	0	0	0	0	1.036532	0	1	-360	360;
	7543	5868	0.002119	0.084719	0	0	0	0	1.036532	0	1	-360	360;
	7543	4814	0.002119	0.084719	0	0	0	0	1.036532	0	1	-360	360;
	10219	8994	0.002119	0.084719	0	0	0	0	1.028791	0	1	-360	360;
	10219	10913	0.002119	0.084719	0	0	0	0	1.028791	0	1	-360	360;
	645	3671	0.002119	0.084719	0	0	0	0	1.041775	0	1	-360	360;
	645	3672	0.002119	0.084719	0	0	0	0	1.041775	0	1	-360	360;
	4520	4811	0.002119	0.084719	0	0	0	0	1.037384	0	1	-360	360;
	4520	4810	0.002119	0.084719	0	0	0	0	1.037384	0	1	-360	360;
	4520	4809	0.002119	0.084719	0	0	0	0	1.037384	0	1	-360	360;
	4520	1772	0.002119	0.084719	0	0	0	0	1.037384	0	1	-360	360;
	4520	1770	0.002119	0.084719	0	0	0	0	1.037384	0	1	-360	360;
	4520	11849	0.002119	0.084719	0	0	0	0	1.037384	0	1	-360	360;
	645	11233	0.002119	0.084719	0	0	0	0	1.039417	0	1	-360	360;
	645	11236	0.002119	0.084719	0	0	0	0	1.039417	0	1	-360	360;
	6772	11235	0.002119	0.084719	0	0	0	0	1.044613	0	1	-360	360;
	6772	11238	0.002119	0.084719	0	0	0	0	1.044613	0	1	-360	360;
	2021	13383	0.002119	0.084719	0	0	0	0	1.026432	0	1	-360	360;
	2021	11240	0.002119	0.084719	0	0	0	0	1.026432	0	1	-360	360;
	2021	3217	0.002119	0.084719	0	0	0	0	1.026432	0	1	-360	360;
	2021	11242	0.002119	0.084719	0	0	0	0	1.026432	0	1	-360	360;
	2021	11241	0.002119	0.084719	0	0	0	0	1.026432	0	1	-360	360;
	5521	4163	0.002119	0.084719	0	0	0	0	1.01824	0	1	-360	360;
	5521	4164	0.002119	0.084719	0	0	0	0	1.01824	0	1	-360	360;
	11659	4161	0.002119	0.084719	0	0	0	0	1.029319	0	1	-360	360;
	11659	10442	0.002119	0.084719	0	0	0	0	1.029319	0	1	-360	360;
	11659	4159	0.002119	0.084719	0	0	0	0	1.029319	0	1	-360	360;
	9680	4160	0.002119	0.084719	0	0	0	0	1.021461	0	1	-360	360;
	9680	4157	0.002119	0.084719	0	0	0	0	1.021461	0	1	-360	360;
	9680	4158	0.002119	0.084719	0	0	0	0	1.021461	0	1	-360	360;
	712	4155	0.002119	0.084719	0	0	0	0	1.025918	0	1	-360	360;
	712	4156	0.002119	0.084719	0	0	0	0	1.025918	0	1	-360	360;
	712	2482	0.002119	0.084719	0	0	0	0	1.025918	0	1	-360	360;
	6457	2435	0.002119	0.084719	0	0	0	0	1.040032	0	1	-360	360;
	6457	2436	0.002119	0.084719	0	0	0	0	1.040032	0	1	-360	360;
	6457	2437	0.002119	0.084719	0	0	0	0	1.040032	0	1	-360	360;
	6457	2438	0.002119	0.084719	0	0	0	0	1.040032	0	1	-360	360;
	6457	2439	0.002119	0.084719	0	0	0	0	1.040032	0	1	-360	360;
	1393	9585	0.000212	0.008494	0	0	0	0	1.030552	0	1	-360	360;
	1393	9584	0.000212	0.008494	0	0	0	0	1.030552	0	1	-360	360;
	6128	9579	0.000212	0.008494	0	0	0	0	1.036319	0	1	-360	360;
	6128	154	0.000212	0.008494	0	0	0	0	1.036319	0	1	-360	360;
	8715	2480	0.000212	0.008494	0	0	0	0	1.055969	0	1	-360	360;
	8715	2481	0.000212	0.008494	0	0	0	0	1.055969	0	1	-360	360;
	10099	4679	0.002119	0.084719	0	0	0	0	1.039927	0	1	-360	360;
	10099	4680	0.002119	0.084719	0	0	0	0	1.039927	0	1	-360	360;
	13256	4676	0.002119	0.084719	0	0	0	0	1.037329	0	1	-360	360;
	13256	4677	0.002119	0.084719	0	0	0	0	1.037329	0	1	-360	360;
	13256	4678	0.002119	0.084719	0	0	0	0	1.037329	0	1	-360	360;
	10502	8668	0.005208	0.208268	0	0	0	0	1.0373	0	1	-360	360;
	10502	8671	0.005208	0.208268	0	0	0	0	1.0373	0	1	-360	360;
	9319	4683	0.002119	0.084719	0	0	0	0	1.03731	0	1	-360	360;
	9319	12279	0.002119	0.084719	0	0	0	0	1.03731	0	1	-360	360;
	9319	12278	0.002119	0.084719	0	0	0	0	1.03731	0	1	-360	360;
	9319	12281	0.002119	0.084719	0	0	0	0	1.03731	0	1	-360	360;
	2684	12280	0.002119	0.084719	0	0	0	0	1.037203	0	1	-360	360;
	2684	12276	0.002119	0.084719	0	0	0	0	1.037203	0	1	-360	360;
	2684	12275	0.002119	0.084719	0	0	0	0	1.037203	0	1	-360	360;
	6458	2372	0.002119	0.084719	0	0	0	0	1.037602	0	1	-360	360;
	6458	2373	0.002119	0.084719	0	0	0	0	1.037602	0	1	-360	360;
	6458	9467	0.002119	0.084719	0	0	0	0	1.037602	0	1	-360	360;
	6458	2377	0.002119	0.084719	0	0	0	0	1.037602	0	1	-360	360;
	6458	9486	0.002119	0.084719	0	0	0	0	1.037602	0	1	-360	360;
	853	4284	0.002119	0.084719	0	0	0	0	1.038861	0	1	-360	360;
	853	4286	0.002119	0.084719	0	0	0	0	1.038861	0	1	-360	360;
	853	4285	0.002119	0.084719	0	0	0	0	1.038861	0	1	-360	360;
	853	4288	0.002119	0.084719	0	0	0	0	1.038861	0	1	-360	360;
	853	4287	0.002119	0.084719	0	0	0	0	1.038861	0	1	-360	360;
	853	4290	0.002119	0.084719	0	0	0	0	1.038861	0	1	-360	360;
	6056	5246	0.002119	0.084719	0	0	0	0	0.99378	0	1	-360	360;
	6056	5242	0.002119	0.084719	0	0	0	0	0.99378	0	1	-360	360;
	6056	4242	0.002119	0.084719	0	0	0	0	0.99378	0	1	-360	360;
	1759	8552	0.002119	0.084719	0	0	0	0	1.033959	0	1	-360	360;
	6595	7498	0.002119	0.084719	0	0	0	0	1.036996	0	1	-360	360;
	6595	12803	0.002119	0.084719	0	0	0	0	1.036996	0	1	-360	360;
	4615	12802	0.002119	0.084719	0	0	0	0	1.016167	0	1	-360	360;
	284	12808	0.002119	0.084719	0	0	0	0	1.039078	0	1	-360	360;
	284	12807	0.002119	0.084719	0	0	0	0	1.039078	0	1	-360	360;
	284	12806	0.002119	0.084719	0	0	0	0	1.039078	0	1	-360	360;
	13593	5498	0.005208	0.208268	0	0	0	0	1.015062	0	1	-360	360;
	13593	12930	0.005208	0.208268	0	0	0	0	1.015062	0	1	-360	360;
	2445	12929	0.005208	0.208268	0	0	0	0	1.003218	0	1	-360	360;
	2445	5847	0.005208	0.208268	0	0	0	0	1.003218	0	1	-360	360;
	11518	5728	0.002119	0.084719	0	0	0	0	1.036278	0	1	-360	360;
	11518	5729	0.002119	0.084719	0	0	0	0	1.036278	0	1	-360	360;
	11518	5730	0.002119	0.084719	0	0	0	0	1.036278	0	1	-360	360;
	11518	5731	0.002119	0.084719	0	0	0	0	1.036278	0	1	-360	360;
	12899	5732	0.002119	0.084719	0	0	0	0	1.030166	0	1	-360	360;
	7516	13379	0.000494	0.019737	0	0	0	0	1.031642	0	1	-360	360;
	7516	13377	0.000494	0.019737	0	0	0	0	1.021069	0	1	-360	360;
	7516	11246	0.000494	0.019737	0	0	0	0	1.021069	0	1	-360	360;
	1032	6918	0.000531	0.021225	0	0	0	0	1.00548	0	1	-360	360;
	1032	12833	0.000531	0.021225	0	0	0	0	1.00548	0	1	-360	360;
	8878	8116	0.005208	0.208268	0	0	0	0	1.018667	0	1	-360	360;
	8878	8113	0.005208	0.208268	0	0	0	0	1.018667	0	1	-360	360;
	13384	8114	0.005208	0.208268	0	0	0	0	1.008951	0	1	-360	360;
	13384	8111	0.005208	0.208268	0	0	0	0	1.008951	0	1	-360	360;
	1854	8112	0.005208	0.208268	0	0	0	0	1.008541	0	1	-360	360;
	1854	8110	0.005208	0.208268	0	0	0	0	1.008541	0	1	-360	360;
	8187	8317	0.003521	0.140801	0	0	0	0	1.015278	0	1	-360	360;
	8187	8319	0.003521	0.140801	0	0	0	0	1.015278	0	1	-360	360;
	8187	8320	0.003521	0.140801	0	0	0	0	1.015278	0	1	-360	360;
	5588	2119	0.000212	0.008494	0	0	0	0	1.037334	0	1	-360	360;
	12621	6227	0.000531	0.021225	0	0	0	0	0.988424	0	1	-360	360;
	10332	7814	0.000708	0.02832	0	0	0	0	1.012948	0	1	-360	360;
	10398	1556	0.005208	0.208268	0	0	0	0	1.025893	0	1	-360	360;
	10398	1558	0.005208	0.208268	0	0	0	0	1.025893	0	1	-360	360;
	10398	1557	0.005208	0.208268	0	0	0	0	1.025893	0	1	-360	360;
	5767	6226	0.000531	0.021225	0	0	0	0	1.015251	0	1	-360	360;
	2967	5510	0.002119	0.084719	0	0	0	0	1.012008	0	1	-360	360;
	2967	5514	0.002119	0.084719	0	0	0	0	1.012008	0	1	-360	360;
	1197	1560	0.005208	0.208268	0	0	0	0	1.010751	0	1	-360	360;
	1197	7636	0.005208	0.208268	0	0	0	0	1.010751	0	1	-360	360;
	1197	7637	0.005208	0.208268	0	0	0	0	1.010751	0	1	-360	360;
	5945	12255	0.000531	0.021225	0	0	0	0	1.01432	0	1	-360	360;
	8520	5640	0.000531	0.021225	0	0	0	0	1.009467	0	1	-360	360;
	8520	5084	0.000531	0.021225	0	0	0	0	1.009467	0	1	-360	360;
	7273	4245	0.000531	0.021225	0	0	0	0	1.009457	0	1	-360	360;
	7273	12252	0.000531	0.021225	0	0	0	0	1.009457	0	1	-360	360;
	2221	4274	0.000531	0.021225	0	0	0	0	1.009453	0	1	-360	360;
	2221	5654	0.000531	0.021225	0	0	0	0	1.009453	0	1	-360	360;
	1035	12259	0.000531	0.021225	0	0	0	0	1.009453	0	1	-360	360;
	1035	5708	0.000531	0.021225	0	0	0	0	1.009453	0	1	-360	360;
	13619	5707	0.000531	0.021225	0	0	0	0	1.024548	0	1	-360	360;
	13619	5706	0.000531	0.021225	0	0	0	0	1.024548	0	1	-360	360;
	4951	1089	0.005208	0.208268	0	0	0	0	0.995167	0	1	-360	360;
	4951	1093	0.005208	0.208268	0	0	0	0	0.995167	0	1	-360	360;
	4951	1092	0.005208	0.208268	0	0	0	0	0.995167	0	1	-360	360;
	10569	5705	0.000531	0.021225	0	0	0	0	1.030285	0	1	-360	360;
	7880	5704	0.000531	0.021225	0	0	0	0	1.030279	0	1	-360	360;
	11998	6039	0.002119	0.084719	0	0	0	0	1.026775	0	1	-360	360;
	11998	6040	0.002119	0.084719	0	0	0	0	1.026775	0	1	-360	360;
	1729	2186	0.000708	0.02832	0	0	0	0	1.030363	0	1	-360	360;
	5749	7736	0.000531	0.021225	0	0	0	0	1.037534	0	1	-360	360;
	5997	1914	0.002119	0.084719	0	0	0	0	1.000306	0	1	-360	360;
	5997	1913	0.002119	0.084719	0	0	0	0	1.000306	0	1	-360	360;
	4280	9655	0.000212	0.008494	0	0	0	0	1.059014	0	1	-360	360;
	1634	9656	0.000212	0.008494	0	0	0	0	1.059006	0	1	-360	360;
	2938	9106	0.005208	0.208268	0	0	0	0	0.998143	0	1	-360	360;
	2938	9107	0.005208	0.208268	0	0	0	0	0.998143	0	1	-360	360;
	2938	9102	0.005208	0.208268	0	0	0	0	0.998143	0	1	-360	360;
	6463	6409	0.000531	0.021225	0	0	0	0	1.033803	0	1	-360	360;
	2293	5223	0.000531	0.021225	0	0	0	0	1.033762	0	1	-360	360;
	4802	846	0.003521	0.140801	0	0	0	0	1.026335	0	1	-360	360;
	4802	845	0.003521	0.140801	0	0	0	0	1.026335	0	1	-360	360;
	4802	844	0.003521	0.140801	0	0	0	0	1.026335	0	1	-360	360;
	3374	2489	0.005208	0.208268	0	0	0	0	1.006944	0	1	-360	360;
	3374	2488	0.005208	0.208268	0	0	0	0	1.006944	0	1	-360	360;
	7738	8773	0.000708	0.02832	0	0	0	0	0.991514	0	1	-360	360;
	9318	2490	0.005208	0.208268	0	0	0	0	0.995022	0	1	-360	360;
	9318	2487	0.005208	0.208268	0	0	0	0	0.995022	0	1	-360	360;
	9318	2486	0.005208	0.208268	0	0	0	0	0.995022	0	1	-360	360;
	9318	8576	0.005208	0.208268	0	0	0	0	0.995022	0	1	-360	360;
	5288	8577	0.005208	0.208268	0	0	0	0	0.994181	0	1	-360	360;
	5288	8578	0.005208	0.208268	0	0	0	0	0.994181	0	1	-360	360;
	759	12814	0.002119	0.084719	0	0	0	0	1.005303	0	1	-360	360;
	759	12072	0.002119	0.084719	0	0	0	0	1.005303	0	1	-360	360;
	11688	1735	0.000708	0.02832	0	0	0	0	1.007359	0	1	-360	360;
	11688	1734	0.000708	0.02832	0	0	0	0	1.007359	0	1	-360	360;
	10391	13592	0.000531	0.021225	0	0	0	0	0.993134	0	1	-360	360;
	4306	8581	0.005208	0.208268	0	0	0	0	1.002991	0	1	-360	360;
	4306	8582	0.005208	0.208268	0	0	0	0	1.002991	0	1	-360	360;
	4306	2006	0.005208	0.208268	0	0	0	0	1.002991	0	1	-360	360;
	4872	2005	0.005208	0.208268	0	0	0	0	0.985964	0	1	-360	360;
	4872	2004	0.005208	0.208268	0	0	0	0	0.985964	0	1	-360	360;
	4872	2003	0.005208	0.208268	0	0	0	0	0.985964	0	1	-360	360;
	6154	12855	0.000212	0.008494	0	0	0	0	1.049418	0	1	-360	360;
	3721	2001	0.005208	0.208268	0	0	0	0	0.991543	0	1	-360	360;
	8574	2000	0.005208	0.208268	0	0	0	0	1.026313	0	1	-360	360;
	8574	1999	0.005208	0.208268	0	0	0	0	1.026313	0	1	-360	360;
	8574	2011	0.005208	0.208268	0	0	0	0	1.026313	0	1	-360	360;
	5211	4710	0.000531	0.021225	0	0	0	0	1.000644	0	1	-360	360;
	553	5858	0.000708	0.02832	0	0	0	0	1.027095	0	1	-360	360;
	553	5856	0.000708	0.02832	0	0	0	0	1.027095	0	1	-360	360;
	4921	1050	0.000531	0.021225	0	0	0	0	0.996774	0	1	-360	360;
	8266	9818	0.003521	0.140801	0	0	0	0	1.013246	0	1	-360	360;
	4842	3913	0.002119	0.084719	0	0	0	0	1.012283	0	1	-360	360;
	1201	4240	0.003098	0.123877	0	0	0	0	1.010818	0	1	-360	360;
	1201	4239	0.003098	0.123877	0	0	0	0	1.005111	0	1	-360	360;
	1212	10453	0.003521	0.140801	0	0	0	0	1.003133	0	1	-360	360;
	3364	3903	0.002119	0.084719	0	0	0	0	1.009261	0	1	-360	360;
	6517	4234	0.002119	0.084719	0	0	0	0	1.008214	0	1	-360	360;
	6517	11806	0.002119	0.084719	0	0	0	0	1.008214	0	1	-360	360;
	7530	11886	0.003521	0.140801	0	0	0	0	1.007633	0	1	-360	360;
	6209	6420	0.002119	0.084719	0	0	0	0	1.010568	0	1	-360	360;
	8699	7338	0.000531	0.021225	0	0	0	0	1.024425	0	1	-360	360;
	4567	1218	0.000708	0.02832	0	0	0	0	1.01045	0	1	-360	360;
	4333	6421	0.002119	0.084719	0	0	0	0	1.010834	0	1	-360	360;
	5655	4162	0.003521	0.140801	0	0	0	0	1.004714	0	1	-360	360;
	9664	9074	0.002119	0.084719	0	0	0	0	1.012149	0	1	-360	360;
	11245	4152	0.003521	0.140801	0	0	0	0	1.001089	0	1	-360	360;
	4075	4151	0.003521	0.140801	0	0	0	0	1.014373	0	1	-360	360;
	12719	10362	0.002119	0.084719	0	0	0	0	1.007634	0	1	-360	360;
	8690	5406	0.003521	0.140801	0	0	0	0	1.01045	0	1	-360	360;
	13381	5245	0.002119	0.084719	0	0	0	0	1.009766	0	1	-360	360;
	3156	8740	0.000708	0.02832	0	0	0	0	1.019589	0	1	-360	360;
	6425	8741	0.000708	0.02832	0	0	0	0	1.014377	0	1	-360	360;
	10931	873	0.003521	0.140801	0	0	0	0	1.01008	0	1	-360	360;
	9500	1995	0.000531	0.021225	0	0	0	0	1.014292	0	1	-360	360;
	3813	5411	0.003521	0.140801	0	0	0	0	1.009591	0	1	-360	360;
	327	10287	0.007184	0.287267	0	0	0	0	1.014694	0	1	-360	360;
	8557	5243	0.002119	0.084719	0	0	0	0	1.017021	0	1	-360	360;
	3690	2307	0.005208	0.208268	0	0	0	0	1.007694	0	1	-360	360;
	5502	7377	0.003521	0.140801	0	0	0	0	1.009279	0	1	-360	360;
	10128	3853	0.003521	0.140801	0	0	0	0	1.011716	0	1	-360	360;
	11904	12	0.000531	0.021225	0	0	0	0	1.023676	0	1	-360	360;
	7692	8451	0.002119	0.084719	0	0	0	0	1.00974	0	1	-360	360;
	1962	13655	0.005208	0.208268	0	0	0	0	1.006172	0	1	-360	360;
	8737	11889	0.000708	0.02832	0	0	0	0	1.044306	0	1	-360	360;
	8737	1940	0.000708	0.02832	0	0	0	0	1.044306	0	1	-360	360;
	8737	11897	0.000708	0.02832	0	0	0	0	1.044306	0	1	-360	360;
	6151	11898	0.000708	0.02832	0	0	0	0	1.041635	0	1	-360	360;
	6151	4858	0.000708	0.02832	0	0	0	0	1.041635	0	1	-360	360;
	6151	4857	0.000708	0.02832	0	0	0	0	1.041635	0	1	-360	360;
	2887	4856	0.000708	0.02832	0	0	0	0	1.043948	0	1	-360	360;
	2887	4855	0.000708	0.02832	0	0	0	0	1.043948	0	1	-360	360;
	2493	4862	0.000708	0.02832	0	0	0	0	1.027024	0	1	-360	360;
	2493	4861	0.000708	0.02832	0	0	0	0	1.027024	0	1	-360	360;
	2493	2047	0.000708	0.02832	0	0	0	0	1.027024	0	1	-360	360;
	13140	4859	0.000708	0.02832	0	0	0	0	1.029372	0	1	-360	360;
	13140	4864	0.000708	0.02832	0	0	0	0	1.029372	0	1	-360	360;
	6366	10250	0.000531	0.021225	0	0	0	0	1.014233	0	1	-360	360;
	6366	6017	0.000531	0.021225	0	0	0	0	1.014233	0	1	-360	360;
	6366	3127	0.000531	0.021225	0	0	0	0	1.014233	0	1	-360	360;
	11839	11408	0.000708	0.02832	0	0	0	0	1.015667	0	1	-360	360;
	11839	11409	0.000708	0.02832	0	0	0	0	1.015667	0	1	-360	360;
	12513	11410	0.002495	0.099769	0	0	0	0	1.02086	0	1	-360	360;
	5478	3130	0.000531	0.021225	0	0	0	0	1.025981	0	1	-360	360;
	5478	3128	0.000531	0.021225	0	0	0	0	1.025981	0	1	-360	360;
	5478	6023	0.000531	0.021225	0	0	0	0	1.025981	0	1	-360	360;
	2699	3123	0.000531	0.021225	0	0	0	0	1.018288	0	1	-360	360;
	2699	3124	0.000531	0.021225	0	0	0	0	1.018288	0	1	-360	360;
	2699	11406	0.000531	0.021225	0	0	0	0	1.018288	0	1	-360	360;
	2699	8803	0.000531	0.021225	0	0	0	0	1.018288	0	1	-360	360;
	3693	6358	0.000708	0.02832	0	0	0	0	1.034253	0	1	-360	360;
	3693	6357	0.000708	0.02832	0	0	0	0	1.034253	0	1	-360	360;
	9988	6352	0.000708	0.02832	0	0	0	0	1.019154	0	1	-360	360;
	9988	6351	0.000708	0.02832	0	0	0	0	1.019154	0	1	-360	360;
	8968	6354	0.000708	0.02832	0	0	0	0	1.043867	0	1	-360	360;
	8968	6353	0.000708	0.02832	0	0	0	0	1.043867	0	1	-360	360;
	8968	6361	0.000708	0.02832	0	0	0	0	1.043867	0	1	-360	360;
	2553	6360	0.000708	0.02832	0	0	0	0	1.00932	0	1	-360	360;
	2553	12965	0.000708	0.02832	0	0	0	0	1.00932	0	1	-360	360;
	7099	9522	0.000708	0.02832	0	0	0	0	1.028812	0	1	-360	360;
	7099	9513	0.000708	0.02832	0	0	0	0	1.028812	0	1	-360	360;
	11165	9515	0.000708	0.02832	0	0	0	0	1.043848	0	1	-360	360;
	3044	12969	0.000708	0.02832	0	0	0	0	1.031212	0	1	-360	360;
	3044	12970	0.000708	0.02832	0	0	0	0	1.031212	0	1	-360	360;
	13244	9523	0.000708	0.02832	0	0	0	0	1.031102	0	1	-360	360;
	13244	9525	0.000708	0.02832	0	0	0	0	1.031102	0	1	-360	360;
	7412	1750	0.000531	0.021225	0	0	0	0	1.03605	0	1	-360	360;
	7412	1751	0.000531	0.021225	0	0	0	0	1.03605	0	1	-360	360;
	7412	9294	0.000531	0.021225	0	0	0	0	1.03605	0	1	-360	360;
	10651	5889	0.000708	0.02832	0	0	0	0	1.032862	0	1	-360	360;
	10651	9540	0.000708	0.02832	0	0	0	0	1.032862	0	1	-360	360;
	10651	5887	0.000708	0.02832	0	0	0	0	1.032862	0	1	-360	360;
	8758	8090	0.000708	0.02832	0	0	0	0	1.027152	0	1	-360	360;
	8758	8088	0.000708	0.02832	0	0	0	0	1.027152	0	1	-360	360;
	7000	5884	0.000708	0.02832	0	0	0	0	1.032522	0	1	-360	360;
	7000	5883	0.000708	0.02832	0	0	0	0	1.032522	0	1	-360	360;
	7000	5882	0.000708	0.02832	0	0	0	0	1.032522	0	1	-360	360;
	279	5881	0.000708	0.02832	0	0	0	0	1.02075	0	1	-360	360;
	7103	12452	0.000708	0.02832	0	0	0	0	1.020873	0	1	-360	360;
	5437	12453	0.000708	0.02832	0	0	0	0	1.020808	0	1	-360	360;
	7891	12454	0.000708	0.02832	0	0	0	0	1.015181	0	1	-360	360;
	10637	6726	0.006494	0.259659	0	0	0	0	1.014067	0	1	-360	360;
	7729	1546	0.001384	0.055323	0	0	0	0	1.034666	0	1	-360	360;
	7729	1545	0.001384	0.055323	0	0	0	0	1.027558	0	1	-360	360;
	7729	1544	0.001384	0.055323	0	0	0	0	1.027558	0	1	-360	360;
	7729	13056	0.001384	0.055323	0	0	0	0	1.027558	0	1	-360	360;
	7729	1537	0.001384	0.055323	0	0	0	0	1.027558	0	1	-360	360;
	7729	9979	0.001384	0.055323	0	0	0	0	1.027558	0	1	-360	360;
	7729	13124	0.001384	0.055323	0	0	0	0	1.027558	0	1	-360	360;
	7729	11463	0.001384	0.055323	0	0	0	0	1.027558	0	1	-360	360;
	7729	13099	0.001384	0.055323	0	0	0	0	1.027558	0	1	-360	360;
	7729	11461	0.001384	0.055323	0	0	0	0	1.027558	0	1	-360	360;
	7729	11464	0.001384	0.055323	0	0	0	0	1.027558	0	1	-360	360;
	1531	3344	0.000708	0.02832	0	0	0	0	1.040265	0	1	-360	360;
	1531	7137	0.000708	0.02832	0	0	0	0	1.040265	0	1	-360	360;
	2043	9602	0.002747	0.109856	0	0	0	0	1.029337	0	1	-360	360;
	2043	9606	0.002747	0.109856	0	0	0	0	1.021	0	1	-360	360;
	2043	9605	0.002747	0.109856	0	0	0	0	1.021	0	1	-360	360;
	2043	2500	0.002747	0.109856	0	0	0	0	1.021	0	1	-360	360;
	2043	2501	0.002747	0.109856	0	0	0	0	1.021	0	1	-360	360;
	2043	2502	0.002747	0.109856	0	0	0	0	1.021	0	1	-360	360;
	4109	5366	0.005353	0.214066	0	0	0	0	1.009344	0	1	-360	360;
	4109	4939	0.005353	0.214066	0	0	0	0	1.009344	0	1	-360	360;
	4109	5365	0.005353	0.214066	0	0	0	0	1.009344	0	1	-360	360;
	4109	5362	0.005353	0.214066	0	0	0	0	1.004456	0	1	-360	360;
	4109	13207	0.005353	0.214066	0	0	0	0	1.004456	0	1	-360	360;
	4109	5360	0.005353	0.214066	0	0	0	0	1.004456	0	1	-360	360;
	4109	5361	0.005353	0.214066	0	0	0	0	1.004456	0	1	-360	360;
	4109	11084	0.005353	0.214066	0	0	0	0	1.004456	0	1	-360	360;
	756	11085	0.005208	0.208268	0	0	0	0	1.006703	0	1	-360	360;
	756	11086	0.005208	0.208268	0	0	0	0	1.006703	0	1	-360	360;
	756	11087	0.005208	0.208268	0	0	0	0	1.006703	0	1	-360	360;
	756	11088	0.005208	0.208268	0	0	0	0	1.006703	0	1	-360	360;
	7425	12646	0.005208	0.208268	0	0	0	0	1.026505	0	1	-360	360;
	7425	11090	0.005208	0.208268	0	0	0	0	1.026505	0	1	-360	360;
	12423	652	0.000531	0.021225	0	0	0	0	1.02312	0	1	-360	360;
	12423	651	0.000531	0.021225	0	0	0	0	1.02312	0	1	-360	360;
	1646	5455	0.001873	0.074883	0	0	0	0	1.027046	0	1	-360	360;
	1646	12010	0.001873	0.074883	0	0	0	0	1.020907	0	1	-360	360;
	1646	12011	0.001873	0.074883	0	0	0	0	1.020907	0	1	-360	360;
	1646	12009	0.001873	0.074883	0	0	0	0	1.020907	0	1	-360	360;
	6181	433	0.005208	0.208268	0	0	0	0	1.025949	0	1	-360	360;
	6181	4537	0.005208	0.208268	0	0	0	0	1.025949	0	1	-360	360;
	6181	12709	0.005208	0.208268	0	0	0	0	1.025949	0	1	-360	360;
	6181	4535	0.005208	0.208268	0	0	0	0	1.025949	0	1	-360	360;
	6181	4534	0.005208	0.208268	0	0	0	0	1.025949	0	1	-360	360;
	6181	2286	0.000708	0.02832	0	0	0	0	1.026694	0	1	-360	360;
	6181	2287	0.000708	0.02832	0	0	0	0	1.026694	0	1	-360	360;
	8152	1813	0.005208	0.208268	0	0	0	0	1.005842	0	1	-360	360;
	177	1816	0.005208	0.208268	0	0	0	0	1.025876	0	1	-360	360;
	5140	5332	0.005208	0.208268	0	0	0	0	1.025539	0	1	-360	360;
	5140	5594	0.005208	0.208268	0	0	0	0	1.025539	0	1	-360	360;
	5140	5196	0.005208	0.208268	0	0	0	0	1.025539	0	1	-360	360;
	4559	11602	0.005208	0.208268	0	0	0	0	1.02756	0	1	-360	360;
	4559	11599	0.005208	0.208268	0	0	0	0	1.02756	0	1	-360	360;
	4559	5540	0.005208	0.208268	0	0	0	0	1.02756	0	1	-360	360;
	2461	1831	0.005208	0.208268	0	0	0	0	1.024287	0	1	-360	360;
	2461	11611	0.005208	0.208268	0	0	0	0	1.024287	0	1	-360	360;
	2461	5038	0.005208	0.208268	0	0	0	0	1.024287	0	1	-360	360;
	249	5746	0.006494	0.259659	0	0	0	0	1.03136	0	1	-360	360;
	249	5748	0.006494	0.259659	0	0	0	0	1.025041	0	1	-360	360;
	249	5747	0.006494	0.259659	0	0	0	0	1.025041	0	1	-360	360;
	13278	12711	0.006527	0.261015	0	0	0	0	1.002931	0	1	-360	360;
	13278	12712	0.006527	0.261015	0	0	0	0	0.996669	0	1	-360	360;
	10745	12713	0.005208	0.208268	0	0	0	0	1.019055	0	1	-360	360;
	1146	12714	0.005208	0.208268	0	0	0	0	1.007266	0	1	-360	360;
	1146	12708	0.005208	0.208268	0	0	0	0	1.007266	0	1	-360	360;
	1861	7540	0.005208	0.208268	0	0	0	0	1.009862	0	1	-360	360;
	664	5629	0.005208	0.208268	0	0	0	0	1.002399	0	1	-360	360;
	664	5628	0.005208	0.208268	0	0	0	0	1.002399	0	1	-360	360;
	10824	7262	0.002096	0.083796	0	0	0	0	1.003762	0	1	-360	360;
	11841	5630	0.018939	0.757339	0	0	0	0	1.006491	0	1	-360	360;
	8625	5632	0.005208	0.208268	0	0	0	0	1.014052	0	1	-360	360;
	4915	5631	0.005208	0.208268	0	0	0	0	1.010825	0	1	-360	360;
	4915	5634	0.005208	0.208268	0	0	0	0	1.010825	0	1	-360	360;
	13255	11757	0.003521	0.140801	0	0	0	0	1.020236	0	1	-360	360;
	13255	11756	0.003521	0.140801	0	0	0	0	1.020236	0	1	-360	360;
	13255	11755	0.003521	0.140801	0	0	0	0	1.020236	0	1	-360	360;
	9712	12193	0.005208	0.208268	0	0	0	0	1.003827	0	1	-360	360;
	9712	12194	0.005208	0.208268	0	0	0	0	1.003827	0	1	-360	360;
	8009	12191	0.005208	0.208268	0	0	0	0	1.014614	0	1	-360	360;
	8009	12192	0.005208	0.208268	0	0	0	0	1.014614	0	1	-360	360;
	11209	382	0.002034	0.081342	0	0	0	0	1.006628	0	1	-360	360;
	11209	381	0.002034	0.081342	0	0	0	0	1.000622	0	1	-360	360;
	11209	384	0.002034	0.081342	0	0	0	0	1.000622	0	1	-360	360;
	1067	383	0.002041	0.081607	0	0	0	0	1.007282	0	1	-360	360;
	1067	378	0.002041	0.081607	0	0	0	0	1.003092	0	1	-360	360;
	1067	377	0.002041	0.081607	0	0	0	0	1.003092	0	1	-360	360;
	604	5144	0.005208	0.208268	0	0	0	0	1.010345	0	1	-360	360;
	3795	12601	0.002119	0.084719	0	0	0	0	1.033336	0	1	-360	360;
	3795	12599	0.002119	0.084719	0	0	0	0	1.033336	0	1	-360	360;
	3795	12600	0.002119	0.084719	0	0	0	0	1.033336	0	1	-360	360;
	4283	5148	0.005208	0.208268	0	0	0	0	1.01495	0	1	-360	360;
	9037	5147	0.005208	0.208268	0	0	0	0	1.0178	0	1	-360	360;
	9037	5146	0.005208	0.208268	0	0	0	0	1.0178	0	1	-360	360;
	9037	5145	0.005208	0.208268	0	0	0	0	1.0178	0	1	-360	360;
	1219	5151	0.005208	0.208268	0	0	0	0	1.009281	0	1	-360	360;
	1219	1764	0.005208	0.208268	0	0	0	0	1.009281	0	1	-360	360;
	1219	11686	0.005208	0.208268	0	0	0	0	1.009281	0	1	-360	360;
	13202	11687	0.005208	0.208268	0	0	0	0	1.002838	0	1	-360	360;
	2132	2246	0.005208	0.208268	0	0	0	0	1.002513	0	1	-360	360;
	2132	11689	0.005208	0.208268	0	0	0	0	1.002513	0	1	-360	360;
	2132	11690	0.005208	0.208268	0	0	0	0	1.002513	0	1	-360	360;
	2132	11691	0.005208	0.208268	0	0	0	0	1.002513	0	1	-360	360;
	2132	11692	0.005208	0.208268	0	0	0	0	1.002513	0	1	-360	360;
	2132	11693	0.005208	0.208268	0	0	0	0	1.002513	0	1	-360	360;
	10240	11694	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	11695	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	4634	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	4633	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	930	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	4635	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	4630	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	912	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	920	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	4631	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	4627	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	4626	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	11193	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	11194	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	11191	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	13160	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	11197	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	11198	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	11195	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10240	11196	0.005208	0.208268	0	0	0	0	1.017093	0	1	-360	360;
	10118	2303	0.001598	0.063919	0	0	0	0	1.006298	0	1	-360	360;
	8355	13091	0.002119	0.084719	0	0	0	0	1.007157	0	1	-360	360;
	8355	9574	0.002119	0.084719	0	0	0	0	1.007157	0	1	-360	360;
	8355	9573	0.002119	0.084719	0	0	0	0	1.007157	0	1	-360	360;
	7506	8966	0.002119	0.084719	0	0	0	0	1.012046	0	1	-360	360;
	7506	8967	0.002119	0.084719	0	0	0	0	1.012046	0	1	-360	360;
	7506	8964	0.002119	0.084719	0	0	0	0	1.012046	0	1	-360	360;
	7506	8965	0.002119	0.084719	0	0	0	0	1.012046	0	1	-360	360;
	2597	13577	0.003521	0.140801	0	0	0	0	1.004111	0	1	-360	360;
	2597	156	0.003521	0.140801	0	0	0	0	1.004111	0	1	-360	360;
	2597	162	0.003521	0.140801	0	0	0	0	1.004111	0	1	-360	360;
	2597	161	0.003521	0.140801	0	0	0	0	1.004111	0	1	-360	360;
	2597	6690	0.003521	0.140801	0	0	0	0	1.004111	0	1	-360	360;
	7288	6691	0.003521	0.140801	0	0	0	0	1.004111	0	1	-360	360;
	7288	13080	0.003521	0.140801	0	0	0	0	1.004111	0	1	-360	360;
	7288	6689	0.003521	0.140801	0	0	0	0	1.004111	0	1	-360	360;
	7288	3912	0.003521	0.140801	0	0	0	0	1.004111	0	1	-360	360;
	7288	13101	0.003521	0.140801	0	0	0	0	1.004111	0	1	-360	360;
	9966	7600	0.003521	0.140801	0	0	0	0	1.00414	0	1	-360	360;
	9966	6693	0.003521	0.140801	0	0	0	0	1.00414	0	1	-360	360;
	9966	13108	0.003521	0.140801	0	0	0	0	1.00414	0	1	-360	360;
	9966	6697	0.003521	0.140801	0	0	0	0	1.00414	0	1	-360	360;
	9966	13316	0.003521	0.140801	0	0	0	0	1.00414	0	1	-360	360;
	12355	8470	0.002119	0.084719	0	0	0	0	0.982596	0	1	-360	360;
	12355	8471	0.002119	0.084719	0	0	0	0	0.982596	0	1	-360	360;
	12355	10423	0.002119	0.084719	0	0	0	0	0.982596	0	1	-360	360;
	12355	8473	0.002119	0.084719	0	0	0	0	0.982596	0	1	-360	360;
	12355	8474	0.002119	0.084719	0	0	0	0	0.982596	0	1	-360	360;
	8579	13313	0.003521	0.140801	0	0	0	0	1.010156	0	1	-360	360;
	8579	11580	0.003521	0.140801	0	0	0	0	1.010156	0	1	-360	360;
	8298	13311	0.003521	0.140801	0	0	0	0	1.03642	0	1	-360	360;
	9266	13310	0.003521	0.140801	0	0	0	0	0.999915	0	1	-360	360;
	9266	6219	0.003521	0.140801	0	0	0	0	0.999915	0	1	-360	360;
	8919	6220	0.003125	0.124961	0	0	0	0	1.00568	0	1	-360	360;
	8919	6221	0.003125	0.124961	0	0	0	0	0.999654	0	1	-360	360;
	8919	10155	0.003125	0.124961	0	0	0	0	0.999654	0	1	-360	360;
	6145	6215	0.003079	0.123114	0	0	0	0	1.006057	0	1	-360	360;
	6145	8185	0.003079	0.123114	0	0	0	0	1.00004	0	1	-360	360;
	6145	8199	0.003079	0.123114	0	0	0	0	1.00004	0	1	-360	360;
	6984	6218	0.003521	0.140801	0	0	0	0	1.01016	0	1	-360	360;
	6984	6213	0.003521	0.140801	0	0	0	0	1.01016	0	1	-360	360;
	4689	3352	0.002119	0.084719	0	0	0	0	0.995148	0	1	-360	360;
	4689	9911	0.002119	0.084719	0	0	0	0	0.995148	0	1	-360	360;
	4689	9912	0.002119	0.084719	0	0	0	0	0.995148	0	1	-360	360;
	4689	9909	0.002119	0.084719	0	0	0	0	0.995148	0	1	-360	360;
	12229	11472	0.000708	0.02832	0	0	0	0	0.998196	0	1	-360	360;
	12229	11477	0.000708	0.02832	0	0	0	0	0.998196	0	1	-360	360;
	9069	11478	0.000708	0.02832	0	0	0	0	0.998227	0	1	-360	360;
	9069	11475	0.000708	0.02832	0	0	0	0	0.998227	0	1	-360	360;
	11671	1083	0.000708	0.02832	0	0	0	0	0.998207	0	1	-360	360;
	11671	11479	0.000708	0.02832	0	0	0	0	0.998207	0	1	-360	360;
	2742	11480	0.000708	0.02832	0	0	0	0	0.998251	0	1	-360	360;
	2742	2434	0.000708	0.02832	0	0	0	0	0.998251	0	1	-360	360;
	13479	8873	0.003521	0.140801	0	0	0	0	1.0025	0	1	-360	360;
	13479	5750	0.003521	0.140801	0	0	0	0	1.0025	0	1	-360	360;
	6643	5751	0.003521	0.140801	0	0	0	0	1.011301	0	1	-360	360;
	2591	3282	0.002119	0.084719	0	0	0	0	1.002083	0	1	-360	360;
	2591	6862	0.002119	0.084719	0	0	0	0	1.002083	0	1	-360	360;
	2591	3284	0.002119	0.084719	0	0	0	0	1.002083	0	1	-360	360;
	2591	3283	0.002119	0.084719	0	0	0	0	1.002083	0	1	-360	360;
	4795	2902	0.000531	0.021225	0	0	0	0	1.009643	0	1	-360	360;
	4795	2901	0.000531	0.021225	0	0	0	0	1.009643	0	1	-360	360;
	4450	9440	0.002119	0.084719	0	0	0	0	1.002649	0	1	-360	360;
	4450	9441	0.002119	0.084719	0	0	0	0	1.002649	0	1	-360	360;
	4450	9442	0.002119	0.084719	0	0	0	0	1.002649	0	1	-360	360;
	4450	9443	0.002119	0.084719	0	0	0	0	1.002649	0	1	-360	360;
	4450	9438	0.002119	0.084719	0	0	0	0	1.002649	0	1	-360	360;
	5538	12304	0.003521	0.140801	0	0	0	0	1.008639	0	1	-360	360;
	5538	12303	0.003521	0.140801	0	0	0	0	1.008639	0	1	-360	360;
	5538	12302	0.003521	0.140801	0	0	0	0	1.008639	0	1	-360	360;
	5538	12307	0.003521	0.140801	0	0	0	0	1.008639	0	1	-360	360;
	1106	12306	0.003521	0.140801	0	0	0	0	1.013853	0	1	-360	360;
	12345	5270	0.003521	0.140801	0	0	0	0	1.002404	0	1	-360	360;
	12345	5271	0.003521	0.140801	0	0	0	0	1.002404	0	1	-360	360;
	12345	4403	0.003521	0.140801	0	0	0	0	1.002404	0	1	-360	360;
	12345	9343	0.003521	0.140801	0	0	0	0	1.002404	0	1	-360	360;
	60	5274	0.003521	0.140801	0	0	0	0	1.001752	0	1	-360	360;
	60	5275	0.003521	0.140801	0	0	0	0	1.001752	0	1	-360	360;
	60	5276	0.003521	0.140801	0	0	0	0	1.001752	0	1	-360	360;
	60	5277	0.003521	0.140801	0	0	0	0	1.001752	0	1	-360	360;
	12345	5278	0.003521	0.140801	0	0	0	0	1.002404	0	1	-360	360;
	12345	5279	0.003521	0.140801	0	0	0	0	1.002404	0	1	-360	360;
	12345	11818	0.003521	0.140801	0	0	0	0	1.002404	0	1	-360	360;
	12345	11817	0.003521	0.140801	0	0	0	0	1.002404	0	1	-360	360;
	60	11820	0.003521	0.140801	0	0	0	0	1.001653	0	1	-360	360;
	60	11819	0.003521	0.140801	0	0	0	0	1.001653	0	1	-360	360;
	60	11814	0.003521	0.140801	0	0	0	0	1.001653	0	1	-360	360;
	60	11813	0.003521	0.140801	0	0	0	0	1.001653	0	1	-360	360;
	1108	11816	0.003521	0.140801	0	0	0	0	0.996098	0	1	-360	360;
	1108	11815	0.003521	0.140801	0	0	0	0	0.996098	0	1	-360	360;
	1108	11811	0.003521	0.140801	0	0	0	0	0.996098	0	1	-360	360;
	5240	11810	0.003521	0.140801	0	0	0	0	0.995875	0	1	-360	360;
	5240	4772	0.003521	0.140801	0	0	0	0	0.995875	0	1	-360	360;
	5240	4773	0.003521	0.140801	0	0	0	0	0.995875	0	1	-360	360;
	5240	4770	0.003521	0.140801	0	0	0	0	0.995875	0	1	-360	360;
	11181	4771	0.003521	0.140801	0	0	0	0	0.9995	0	1	-360	360;
	9338	4774	0.003521	0.140801	0	0	0	0	1.001046	0	1	-360	360;
	9338	1577	0.003521	0.140801	0	0	0	0	1.001046	0	1	-360	360;
	8959	481	0.002119	0.084719	0	0	0	0	0.994019	0	1	-360	360;
	8959	480	0.002119	0.084719	0	0	0	0	0.994019	0	1	-360	360;
	8959	483	0.002119	0.084719	0	0	0	0	0.994019	0	1	-360	360;
	8959	482	0.002119	0.084719	0	0	0	0	0.994019	0	1	-360	360;
	13350	12140	0.002119	0.084719	0	0	0	0	0.992995	0	1	-360	360;
	13350	12139	0.002119	0.084719	0	0	0	0	0.992995	0	1	-360	360;
	13350	12138	0.002119	0.084719	0	0	0	0	0.992995	0	1	-360	360;
	13350	12137	0.002119	0.084719	0	0	0	0	0.992995	0	1	-360	360;
	10040	12136	0.002119	0.084719	0	0	0	0	0.997453	0	1	-360	360;
	10040	12135	0.002119	0.084719	0	0	0	0	0.997453	0	1	-360	360;
	10040	12134	0.002119	0.084719	0	0	0	0	0.997453	0	1	-360	360;
	10040	12133	0.002119	0.084719	0	0	0	0	0.997453	0	1	-360	360;
	7864	12132	0.002119	0.084719	0	0	0	0	1.013451	0	1	-360	360;
	7864	12131	0.002119	0.084719	0	0	0	0	1.013451	0	1	-360	360;
	7864	5089	0.002119	0.084719	0	0	0	0	1.013451	0	1	-360	360;
	7864	5090	0.002119	0.084719	0	0	0	0	1.013451	0	1	-360	360;
	21	5091	0.002119	0.084719	0	0	0	0	1.011752	0	1	-360	360;
	21	5092	0.002119	0.084719	0	0	0	0	1.011752	0	1	-360	360;
	21	5085	0.002119	0.084719	0	0	0	0	1.011752	0	1	-360	360;
	21	5086	0.002119	0.084719	0	0	0	0	1.011752	0	1	-360	360;
	6523	4695	0.000825	0.032971	0	0	0	0	1.003138	0	1	-360	360;
	6523	4696	0.000825	0.032971	0	0	0	0	0.995103	0	1	-360	360;
	6523	4693	0.000825	0.032971	0	0	0	0	0.995103	0	1	-360	360;
	13304	5082	0.002119	0.084719	0	0	0	0	1.013258	0	1	-360	360;
	13304	12652	0.002119	0.084719	0	0	0	0	1.013258	0	1	-360	360;
	13304	12651	0.002119	0.084719	0	0	0	0	1.013258	0	1	-360	360;
	13304	12654	0.002119	0.084719	0	0	0	0	1.013258	0	1	-360	360;
	2211	12653	0.002119	0.084719	0	0	0	0	1.013227	0	1	-360	360;
	2211	12648	0.002119	0.084719	0	0	0	0	1.013227	0	1	-360	360;
	2211	12647	0.002119	0.084719	0	0	0	0	1.013227	0	1	-360	360;
	2211	12650	0.002119	0.084719	0	0	0	0	1.013227	0	1	-360	360;
	8283	11259	0.000531	0.021225	0	0	0	0	1.01231	0	1	-360	360;
	8283	11256	0.000531	0.021225	0	0	0	0	1.01231	0	1	-360	360;
	10793	11028	0.003521	0.140801	0	0	0	0	1.014528	0	1	-360	360;
	10793	4455	0.003521	0.140801	0	0	0	0	1.014528	0	1	-360	360;
	2588	4454	0.003521	0.140801	0	0	0	0	1.014699	0	1	-360	360;
	2588	4457	0.003521	0.140801	0	0	0	0	1.014699	0	1	-360	360;
	2588	4456	0.003521	0.140801	0	0	0	0	1.014699	0	1	-360	360;
	6461	4451	0.003521	0.140801	0	0	0	0	1.014941	0	1	-360	360;
	6461	7183	0.003521	0.140801	0	0	0	0	1.014941	0	1	-360	360;
	6461	4453	0.003521	0.140801	0	0	0	0	1.014941	0	1	-360	360;
	6461	7200	0.003521	0.140801	0	0	0	0	1.014941	0	1	-360	360;
	6461	4461	0.003521	0.140801	0	0	0	0	1.014941	0	1	-360	360;
	3844	4460	0.003521	0.140801	0	0	0	0	1.004844	0	1	-360	360;
	3844	9586	0.003521	0.140801	0	0	0	0	1.004844	0	1	-360	360;
	3844	9587	0.003521	0.140801	0	0	0	0	1.004844	0	1	-360	360;
	9760	9588	0.003521	0.140801	0	0	0	0	1.021581	0	1	-360	360;
	9760	9589	0.003521	0.140801	0	0	0	0	1.021581	0	1	-360	360;
	9760	9590	0.003521	0.140801	0	0	0	0	1.021581	0	1	-360	360;
	9760	9591	0.003521	0.140801	0	0	0	0	1.021581	0	1	-360	360;
	4687	11133	0.002119	0.084719	0	0	0	0	1.013516	0	1	-360	360;
	4687	11132	0.002119	0.084719	0	0	0	0	1.013516	0	1	-360	360;
	4687	11137	0.002119	0.084719	0	0	0	0	1.013516	0	1	-360	360;
	4687	12869	0.002119	0.084719	0	0	0	0	1.013516	0	1	-360	360;
	7308	4053	0.002119	0.084719	0	0	0	0	1.016404	0	1	-360	360;
	7308	4054	0.002119	0.084719	0	0	0	0	1.016404	0	1	-360	360;
	7308	11345	0.002119	0.084719	0	0	0	0	1.016404	0	1	-360	360;
	7308	4056	0.002119	0.084719	0	0	0	0	1.016404	0	1	-360	360;
	12703	4057	0.002119	0.084719	0	0	0	0	1.005292	0	1	-360	360;
	12703	4058	0.002119	0.084719	0	0	0	0	1.005292	0	1	-360	360;
	12703	4059	0.002119	0.084719	0	0	0	0	1.005292	0	1	-360	360;
	12703	4060	0.002119	0.084719	0	0	0	0	1.005292	0	1	-360	360;
	12703	4061	0.002119	0.084719	0	0	0	0	1.005292	0	1	-360	360;
	12703	4062	0.002119	0.084719	0	0	0	0	1.005292	0	1	-360	360;
	1867	6454	0.003521	0.140801	0	0	0	0	1.002241	0	1	-360	360;
	1867	10080	0.003521	0.140801	0	0	0	0	1.002241	0	1	-360	360;
	1867	10077	0.003521	0.140801	0	0	0	0	1.002241	0	1	-360	360;
	1867	10078	0.003521	0.140801	0	0	0	0	1.002241	0	1	-360	360;
	1867	10075	0.003521	0.140801	0	0	0	0	1.002241	0	1	-360	360;
	1867	10076	0.003521	0.140801	0	0	0	0	1.002241	0	1	-360	360;
	1188	10073	0.003521	0.140801	0	0	0	0	1.001755	0	1	-360	360;
	2837	10074	0.003521	0.140801	0	0	0	0	1.001755	0	1	-360	360;
	6600	6471	0.003521	0.140801	0	0	0	0	1.008075	0	1	-360	360;
	6600	10084	0.003521	0.140801	0	0	0	0	1.008075	0	1	-360	360;
	5509	3456	0.003521	0.140801	0	0	0	0	1.018347	0	1	-360	360;
	9332	4580	0.002119	0.084719	0	0	0	0	1.027388	0	1	-360	360;
	9332	4577	0.002119	0.084719	0	0	0	0	1.027388	0	1	-360	360;
	9332	4578	0.002119	0.084719	0	0	0	0	1.027388	0	1	-360	360;
	9332	6391	0.002119	0.084719	0	0	0	0	1.027388	0	1	-360	360;
	9332	4576	0.002119	0.084719	0	0	0	0	1.027388	0	1	-360	360;
	5443	5207	0.000531	0.021225	0	0	0	0	1.007133	0	1	-360	360;
	5443	5208	0.000531	0.021225	0	0	0	0	1.007133	0	1	-360	360;
	12823	4585	0.002119	0.084719	0	0	0	0	1.004586	0	1	-360	360;
	12823	654	0.002119	0.084719	0	0	0	0	1.004586	0	1	-360	360;
	12823	10143	0.002119	0.084719	0	0	0	0	1.004586	0	1	-360	360;
	12823	10142	0.002119	0.084719	0	0	0	0	1.004586	0	1	-360	360;
	12823	10141	0.002119	0.084719	0	0	0	0	1.004586	0	1	-360	360;
	9524	86	0.001377	0.055049	0	0	0	0	1.006083	0	1	-360	360;
	9524	88	0.001377	0.055049	0	0	0	0	0.998959	0	1	-360	360;
	9524	87	0.001377	0.055049	0	0	0	0	0.998959	0	1	-360	360;
	1934	11714	0.000708	0.02832	0	0	0	0	1.01208	0	1	-360	360;
	1934	11713	0.000708	0.02832	0	0	0	0	1.01208	0	1	-360	360;
	1934	6996	0.000708	0.02832	0	0	0	0	1.01208	0	1	-360	360;
	1623	11717	0.000708	0.02832	0	0	0	0	0.999681	0	1	-360	360;
	1623	4655	0.000708	0.02832	0	0	0	0	0.999681	0	1	-360	360;
	1623	4656	0.000708	0.02832	0	0	0	0	0.999681	0	1	-360	360;
	3102	4657	0.000708	0.02832	0	0	0	0	1.003271	0	1	-360	360;
	3102	4658	0.000708	0.02832	0	0	0	0	1.003271	0	1	-360	360;
	3102	1002	0.000708	0.02832	0	0	0	0	1.003271	0	1	-360	360;
	13472	12610	0.003521	0.140801	0	0	0	0	0.994842	0	1	-360	360;
	13472	12607	0.003521	0.140801	0	0	0	0	0.994842	0	1	-360	360;
	3722	3017	0.002119	0.084719	0	0	0	0	1.004816	0	1	-360	360;
	3722	3021	0.002119	0.084719	0	0	0	0	1.004816	0	1	-360	360;
	3722	3022	0.002119	0.084719	0	0	0	0	1.004816	0	1	-360	360;
	3722	2533	0.002119	0.084719	0	0	0	0	1.004816	0	1	-360	360;
	3722	2572	0.002119	0.084719	0	0	0	0	1.004816	0	1	-360	360;
	3722	2573	0.002119	0.084719	0	0	0	0	1.004816	0	1	-360	360;
	6695	12334	0.003521	0.140801	0	0	0	0	0.994421	0	1	-360	360;
	6695	12333	0.003521	0.140801	0	0	0	0	0.994421	0	1	-360	360;
	6695	12332	0.003521	0.140801	0	0	0	0	0.994421	0	1	-360	360;
	6695	6055	0.003521	0.140801	0	0	0	0	0.994421	0	1	-360	360;
	6695	12330	0.003521	0.140801	0	0	0	0	0.994421	0	1	-360	360;
	2484	12338	0.003521	0.140801	0	0	0	0	0.998191	0	1	-360	360;
	2484	12337	0.003521	0.140801	0	0	0	0	0.998191	0	1	-360	360;
	8104	5305	0.003521	0.140801	0	0	0	0	0.998244	0	1	-360	360;
	8104	5306	0.003521	0.140801	0	0	0	0	0.998244	0	1	-360	360;
	5995	5307	0.003521	0.140801	0	0	0	0	0.998295	0	1	-360	360;
	5995	5308	0.003521	0.140801	0	0	0	0	0.998295	0	1	-360	360;
	5807	5301	0.003521	0.140801	0	0	0	0	1.011449	0	1	-360	360;
	434	5302	0.003521	0.140801	0	0	0	0	1.018797	0	1	-360	360;
	4038	5303	0.003521	0.140801	0	0	0	0	1.009509	0	1	-360	360;
	4038	5304	0.003521	0.140801	0	0	0	0	1.009509	0	1	-360	360;
	11153	5309	0.003521	0.140801	0	0	0	0	1.008338	0	1	-360	360;
	11153	5310	0.003521	0.140801	0	0	0	0	1.008338	0	1	-360	360;
	2494	1812	0.002119	0.084719	0	0	0	0	1.008688	0	1	-360	360;
	2494	2446	0.002119	0.084719	0	0	0	0	1.008688	0	1	-360	360;
	2494	2447	0.002119	0.084719	0	0	0	0	1.008688	0	1	-360	360;
	2494	2448	0.002119	0.084719	0	0	0	0	1.008688	0	1	-360	360;
	12643	2991	0.002119	0.084719	0	0	0	0	1.008601	0	1	-360	360;
	12643	8537	0.002119	0.084719	0	0	0	0	1.008601	0	1	-360	360;
	12643	5013	0.002119	0.084719	0	0	0	0	1.008601	0	1	-360	360;
	12643	5012	0.002119	0.084719	0	0	0	0	1.008601	0	1	-360	360;
	9899	5018	0.002119	0.084719	0	0	0	0	1.008554	0	1	-360	360;
	9899	5017	0.002119	0.084719	0	0	0	0	1.008554	0	1	-360	360;
	9899	11568	0.002119	0.084719	0	0	0	0	1.008554	0	1	-360	360;
	9899	11569	0.002119	0.084719	0	0	0	0	1.008554	0	1	-360	360;
	1299	13425	0.003521	0.140801	0	0	0	0	1.014838	0	1	-360	360;
	201	5769	0.003521	0.140801	0	0	0	0	1.014043	0	1	-360	360;
	201	5772	0.003521	0.140801	0	0	0	0	1.014043	0	1	-360	360;
	201	13525	0.003521	0.140801	0	0	0	0	1.014043	0	1	-360	360;
	8254	5770	0.003521	0.140801	0	0	0	0	1.007223	0	1	-360	360;
	8254	5771	0.003521	0.140801	0	0	0	0	1.007223	0	1	-360	360;
	8254	13345	0.003521	0.140801	0	0	0	0	1.007223	0	1	-360	360;
	8254	7391	0.003521	0.140801	0	0	0	0	1.007223	0	1	-360	360;
	8254	11352	0.003521	0.140801	0	0	0	0	1.007223	0	1	-360	360;
	8254	11351	0.003521	0.140801	0	0	0	0	1.007223	0	1	-360	360;
	13132	351	0.000531	0.021225	0	0	0	0	1.011871	0	1	-360	360;
	13132	2097	0.000531	0.021225	0	0	0	0	1.011871	0	1	-360	360;
	10632	3167	0.000708	0.02832	0	0	0	0	1.002423	0	1	-360	360;
	10632	3168	0.000708	0.02832	0	0	0	0	1.002423	0	1	-360	360;
	10632	3169	0.000708	0.02832	0	0	0	0	1.002423	0	1	-360	360;
	5819	11353	0.003521	0.140801	0	0	0	0	1.004868	0	1	-360	360;
	5819	11348	0.003521	0.140801	0	0	0	0	1.004868	0	1	-360	360;
	3080	11347	0.003521	0.140801	0	0	0	0	0.994284	0	1	-360	360;
	12002	7020	0.002119	0.084719	0	0	0	0	1.013556	0	1	-360	360;
	12002	8702	0.002119	0.084719	0	0	0	0	1.013556	0	1	-360	360;
	12002	8701	0.002119	0.084719	0	0	0	0	1.013556	0	1	-360	360;
	12002	11946	0.002119	0.084719	0	0	0	0	1.013556	0	1	-360	360;
	1931	8707	0.002119	0.084719	0	0	0	0	1.013589	0	1	-360	360;
	1931	8706	0.002119	0.084719	0	0	0	0	1.013589	0	1	-360	360;
	1931	11962	0.002119	0.084719	0	0	0	0	1.013589	0	1	-360	360;
	1931	11959	0.002119	0.084719	0	0	0	0	1.013589	0	1	-360	360;
	11316	10985	0.002119	0.084719	0	0	0	0	0.992901	0	1	-360	360;
	11316	11894	0.002119	0.084719	0	0	0	0	0.992901	0	1	-360	360;
	11316	3418	0.002119	0.084719	0	0	0	0	0.992901	0	1	-360	360;
	11316	3419	0.002119	0.084719	0	0	0	0	0.992901	0	1	-360	360;
	13596	3058	0.004448	0.17788	0	0	0	0	1.008199	0	1	-360	360;
	13596	11838	0.004448	0.17788	0	0	0	0	1.002232	0	1	-360	360;
	8	3414	0.002119	0.084719	0	0	0	0	1.01052	0	1	-360	360;
	8	3415	0.002119	0.084719	0	0	0	0	1.01052	0	1	-360	360;
	8	3416	0.002119	0.084719	0	0	0	0	1.01052	0	1	-360	360;
	8	7544	0.002119	0.084719	0	0	0	0	1.01052	0	1	-360	360;
	8	3412	0.002119	0.084719	0	0	0	0	1.01052	0	1	-360	360;
	8	8598	0.002119	0.084719	0	0	0	0	1.01052	0	1	-360	360;
	8	10522	0.002119	0.084719	0	0	0	0	1.01052	0	1	-360	360;
	8	10521	0.002119	0.084719	0	0	0	0	1.01052	0	1	-360	360;
	8	9189	0.002119	0.084719	0	0	0	0	1.010533	0	1	-360	360;
	8	9188	0.002119	0.084719	0	0	0	0	1.010533	0	1	-360	360;
	8	9191	0.002119	0.084719	0	0	0	0	1.010533	0	1	-360	360;
	8	9190	0.002119	0.084719	0	0	0	0	1.010533	0	1	-360	360;
	2601	6910	0.000531	0.021225	0	0	0	0	1.013761	0	1	-360	360;
	2601	6909	0.000531	0.021225	0	0	0	0	1.013761	0	1	-360	360;
	2769	4791	0.003521	0.140801	0	0	0	0	1.026719	0	1	-360	360;
	2769	4792	0.003521	0.140801	0	0	0	0	1.026719	0	1	-360	360;
	2769	619	0.003521	0.140801	0	0	0	0	1.026719	0	1	-360	360;
	6024	4602	0.002119	0.084719	0	0	0	0	0.990176	0	1	-360	360;
	6024	4603	0.002119	0.084719	0	0	0	0	0.990176	0	1	-360	360;
	6024	4604	0.002119	0.084719	0	0	0	0	0.990176	0	1	-360	360;
	6024	4598	0.002119	0.084719	0	0	0	0	0.990176	0	1	-360	360;
	6024	765	0.002119	0.084719	0	0	0	0	0.990176	0	1	-360	360;
	1181	6193	0.000708	0.02832	0	0	0	0	1.001706	0	1	-360	360;
	1181	6194	0.000708	0.02832	0	0	0	0	1.001706	0	1	-360	360;
	4706	615	0.003521	0.140801	0	0	0	0	1.007969	0	1	-360	360;
	4706	614	0.003521	0.140801	0	0	0	0	1.007969	0	1	-360	360;
	280	7163	0.003521	0.140801	0	0	0	0	1.012443	0	1	-360	360;
	12715	11670	0.002119	0.084719	0	0	0	0	1.004291	0	1	-360	360;
	12715	11669	0.002119	0.084719	0	0	0	0	1.004291	0	1	-360	360;
	12715	11668	0.002119	0.084719	0	0	0	0	1.004291	0	1	-360	360;
	12715	11667	0.002119	0.084719	0	0	0	0	1.004291	0	1	-360	360;
	12715	11666	0.002119	0.084719	0	0	0	0	1.004291	0	1	-360	360;
	8448	8308	0.002119	0.084719	0	0	0	0	0.983598	0	1	-360	360;
	8448	8125	0.002119	0.084719	0	0	0	0	0.983598	0	1	-360	360;
	8448	8128	0.002119	0.084719	0	0	0	0	0.983598	0	1	-360	360;
	8448	8127	0.002119	0.084719	0	0	0	0	0.983598	0	1	-360	360;
	3796	9141	0.003521	0.140801	0	0	0	0	1.006074	0	1	-360	360;
	3796	9142	0.003521	0.140801	0	0	0	0	1.006074	0	1	-360	360;
	12029	1389	0.000531	0.021225	0	0	0	0	1.009288	0	1	-360	360;
	12029	1390	0.000531	0.021225	0	0	0	0	1.009288	0	1	-360	360;
	12029	1391	0.000531	0.021225	0	0	0	0	1.009288	0	1	-360	360;
	12148	196	0.002119	0.084719	0	0	0	0	1.00721	0	1	-360	360;
	12148	5886	0.002119	0.084719	0	0	0	0	1.00721	0	1	-360	360;
	12148	5885	0.002119	0.084719	0	0	0	0	1.00721	0	1	-360	360;
	12148	5880	0.002119	0.084719	0	0	0	0	1.00721	0	1	-360	360;
	7406	9137	0.003521	0.140801	0	0	0	0	1.004828	0	1	-360	360;
	7406	2513	0.003521	0.140801	0	0	0	0	1.004828	0	1	-360	360;
	7406	2190	0.003521	0.140801	0	0	0	0	1.004828	0	1	-360	360;
	4794	2515	0.003521	0.140801	0	0	0	0	1.004847	0	1	-360	360;
	4794	2514	0.003521	0.140801	0	0	0	0	1.004847	0	1	-360	360;
	4794	2516	0.003521	0.140801	0	0	0	0	1.004847	0	1	-360	360;
	1008	12455	0.002119	0.084719	0	0	0	0	1.02838	0	1	-360	360;
	1008	6719	0.002119	0.084719	0	0	0	0	1.02838	0	1	-360	360;
	1008	6721	0.002119	0.084719	0	0	0	0	1.02838	0	1	-360	360;
	1008	12447	0.002119	0.084719	0	0	0	0	1.02838	0	1	-360	360;
	5296	2518	0.003521	0.140801	0	0	0	0	1.012026	0	1	-360	360;
	5296	8639	0.003521	0.140801	0	0	0	0	1.012026	0	1	-360	360;
	5296	11570	0.003521	0.140801	0	0	0	0	1.012026	0	1	-360	360;
	5827	5371	0.002119	0.084719	0	0	0	0	1.00294	0	1	-360	360;
	5827	5370	0.002119	0.084719	0	0	0	0	1.00294	0	1	-360	360;
	5827	5369	0.002119	0.084719	0	0	0	0	1.00294	0	1	-360	360;
	5827	5368	0.002119	0.084719	0	0	0	0	1.00294	0	1	-360	360;
	3380	4888	0.000531	0.021225	0	0	0	0	1.008763	0	1	-360	360;
	3380	4887	0.000531	0.021225	0	0	0	0	1.008763	0	1	-360	360;
	4204	1840	0.000531	0.021225	0	0	0	0	1.013933	0	1	-360	360;
	4204	11564	0.000531	0.021225	0	0	0	0	1.013933	0	1	-360	360;
	4204	8911	0.000531	0.021225	0	0	0	0	1.013933	0	1	-360	360;
	12047	13095	0.003587	0.143427	0	0	0	0	1.001823	0	1	-360	360;
	12047	2082	0.003587	0.143427	0	0	0	0	1.001823	0	1	-360	360;
	12047	9570	0.003587	0.143427	0	0	0	0	0.995326	0	1	-360	360;
	12047	2088	0.003587	0.143427	0	0	0	0	0.995326	0	1	-360	360;
	12047	2087	0.003587	0.143427	0	0	0	0	0.995326	0	1	-360	360;
	10091	2086	0.003571	0.142812	0	0	0	0	1.00179	0	1	-360	360;
	10091	2085	0.003571	0.142812	0	0	0	0	1.00179	0	1	-360	360;
	10091	2078	0.003571	0.142812	0	0	0	0	0.995275	0	1	-360	360;
	10091	2077	0.003571	0.142812	0	0	0	0	0.995275	0	1	-360	360;
	10091	10107	0.003571	0.142812	0	0	0	0	0.995275	0	1	-360	360;
	13003	6498	0.002119	0.084719	0	0	0	0	1.007585	0	1	-360	360;
	13003	3219	0.002119	0.084719	0	0	0	0	1.007585	0	1	-360	360;
	13003	3220	0.002119	0.084719	0	0	0	0	1.007585	0	1	-360	360;
	13003	3213	0.002119	0.084719	0	0	0	0	1.007585	0	1	-360	360;
	5096	3214	0.002119	0.084719	0	0	0	0	1.007584	0	1	-360	360;
	5096	3215	0.002119	0.084719	0	0	0	0	1.007584	0	1	-360	360;
	5096	3216	0.002119	0.084719	0	0	0	0	1.007584	0	1	-360	360;
	5096	3210	0.002119	0.084719	0	0	0	0	1.007584	0	1	-360	360;
	7583	6557	0.003521	0.140801	0	0	0	0	1.00868	0	1	-360	360;
	7583	13093	0.003521	0.140801	0	0	0	0	1.00868	0	1	-360	360;
	7583	3487	0.003521	0.140801	0	0	0	0	1.00868	0	1	-360	360;
	5643	3490	0.003521	0.140801	0	0	0	0	1.008681	0	1	-360	360;
	5643	3489	0.003521	0.140801	0	0	0	0	1.008681	0	1	-360	360;
	5643	3485	0.003521	0.140801	0	0	0	0	1.008681	0	1	-360	360;
	1233	11416	0.002119	0.084719	0	0	0	0	1.006861	0	1	-360	360;
	1233	684	0.002119	0.084719	0	0	0	0	1.006861	0	1	-360	360;
	1233	11414	0.002119	0.084719	0	0	0	0	1.006861	0	1	-360	360;
	1233	11411	0.002119	0.084719	0	0	0	0	1.006861	0	1	-360	360;
	1233	11412	0.002119	0.084719	0	0	0	0	1.006861	0	1	-360	360;
	6626	7110	0.002119	0.084719	0	0	0	0	1.007481	0	1	-360	360;
	6626	3720	0.002119	0.084719	0	0	0	0	1.007481	0	1	-360	360;
	6626	3717	0.002119	0.084719	0	0	0	0	1.007481	0	1	-360	360;
	6626	3718	0.002119	0.084719	0	0	0	0	1.007481	0	1	-360	360;
	13645	9616	0.003521	0.140801	0	0	0	0	1.000671	0	1	-360	360;
	8229	9617	0.003521	0.140801	0	0	0	0	0.99286	0	1	-360	360;
	8229	9618	0.003521	0.140801	0	0	0	0	0.99286	0	1	-360	360;
	5139	9619	0.003521	0.140801	0	0	0	0	0.992942	0	1	-360	360;
	5139	9622	0.003521	0.140801	0	0	0	0	0.992942	0	1	-360	360;
	11505	9623	0.003521	0.140801	0	0	0	0	1.0116	0	1	-360	360;
	11505	2982	0.003521	0.140801	0	0	0	0	1.0116	0	1	-360	360;
	11505	2981	0.003521	0.140801	0	0	0	0	1.0116	0	1	-360	360;
	9307	2980	0.003521	0.140801	0	0	0	0	1.005249	0	1	-360	360;
	9307	2979	0.003521	0.140801	0	0	0	0	1.005249	0	1	-360	360;
	9307	2978	0.003521	0.140801	0	0	0	0	1.005249	0	1	-360	360;
	9841	12525	0.000708	0.02832	0	0	0	0	1.003642	0	1	-360	360;
	9841	12526	0.000708	0.02832	0	0	0	0	1.003642	0	1	-360	360;
	9841	7113	0.000708	0.02832	0	0	0	0	1.003642	0	1	-360	360;
	13325	2984	0.003521	0.140801	0	0	0	0	1.011161	0	1	-360	360;
	11938	12518	0.000708	0.02832	0	0	0	0	1.011192	0	1	-360	360;
	11938	9283	0.000708	0.02832	0	0	0	0	1.011192	0	1	-360	360;
	11938	7719	0.002119	0.084719	0	0	0	0	1.009361	0	1	-360	360;
	11938	7722	0.002119	0.084719	0	0	0	0	1.009361	0	1	-360	360;
	11938	7721	0.002119	0.084719	0	0	0	0	1.009361	0	1	-360	360;
	11938	7717	0.002119	0.084719	0	0	0	0	1.009361	0	1	-360	360;
	11938	7716	0.002119	0.084719	0	0	0	0	1.009361	0	1	-360	360;
	11938	5978	0.002119	0.084719	0	0	0	0	1.009361	0	1	-360	360;
	1555	4207	0.002119	0.084719	0	0	0	0	1.000657	0	1	-360	360;
	1555	4214	0.002119	0.084719	0	0	0	0	1.000657	0	1	-360	360;
	1555	12152	0.002119	0.084719	0	0	0	0	1.000657	0	1	-360	360;
	1555	10809	0.002119	0.084719	0	0	0	0	1.000657	0	1	-360	360;
	12889	4481	0.003521	0.140801	0	0	0	0	1.012234	0	1	-360	360;
	11830	4484	0.003521	0.140801	0	0	0	0	1.010846	0	1	-360	360;
	13503	9033	0.000531	0.021225	0	0	0	0	1.008933	0	1	-360	360;
	13503	9034	0.000531	0.021225	0	0	0	0	1.008933	0	1	-360	360;
	6845	732	0.002119	0.084719	0	0	0	0	0.991864	0	1	-360	360;
	6845	730	0.002119	0.084719	0	0	0	0	0.991864	0	1	-360	360;
	6845	731	0.002119	0.084719	0	0	0	0	0.991864	0	1	-360	360;
	6845	734	0.002119	0.084719	0	0	0	0	0.991864	0	1	-360	360;
	6845	4982	0.002119	0.084719	0	0	0	0	0.991864	0	1	-360	360;
	11349	4963	0.003521	0.140801	0	0	0	0	0.991519	0	1	-360	360;
	11349	5342	0.003521	0.140801	0	0	0	0	0.991519	0	1	-360	360;
	11349	5382	0.003521	0.140801	0	0	0	0	0.991519	0	1	-360	360;
	11349	4964	0.003521	0.140801	0	0	0	0	0.991519	0	1	-360	360;
	11901	3585	0.000531	0.021225	0	0	0	0	1.011258	0	1	-360	360;
	11901	3586	0.000531	0.021225	0	0	0	0	1.011258	0	1	-360	360;
	4368	3587	0.000531	0.021225	0	0	0	0	1.011551	0	1	-360	360;
	4368	3588	0.000531	0.021225	0	0	0	0	1.011551	0	1	-360	360;
	4368	2934	0.000531	0.021225	0	0	0	0	1.011551	0	1	-360	360;
	8500	4958	0.003521	0.140801	0	0	0	0	1.004896	0	1	-360	360;
	8500	5676	0.003521	0.140801	0	0	0	0	1.004896	0	1	-360	360;
	3876	11530	0.003521	0.140801	0	0	0	0	1.004751	0	1	-360	360;
	3876	11527	0.003521	0.140801	0	0	0	0	1.004751	0	1	-360	360;
	3876	11528	0.003521	0.140801	0	0	0	0	1.004751	0	1	-360	360;
	3876	11525	0.003521	0.140801	0	0	0	0	1.004746	0	1	-360	360;
	3876	11526	0.003521	0.140801	0	0	0	0	1.004746	0	1	-360	360;
	3876	11523	0.003521	0.140801	0	0	0	0	1.004746	0	1	-360	360;
	3876	11524	0.003521	0.140801	0	0	0	0	1.004746	0	1	-360	360;
	3876	1367	0.003521	0.140801	0	0	0	0	1.004746	0	1	-360	360;
	3876	5830	0.002119	0.084719	0	0	0	0	1.004928	0	1	-360	360;
	3876	13386	0.002119	0.084719	0	0	0	0	1.004928	0	1	-360	360;
	3876	13385	0.002119	0.084719	0	0	0	0	1.004928	0	1	-360	360;
	3876	13387	0.002119	0.084719	0	0	0	0	1.004928	0	1	-360	360;
	3876	13633	0.003521	0.140801	0	0	0	0	1.004749	0	1	-360	360;
	3876	13487	0.003521	0.140801	0	0	0	0	1.004749	0	1	-360	360;
	3876	13648	0.003521	0.140801	0	0	0	0	1.004749	0	1	-360	360;
	3876	6014	0.003521	0.140801	0	0	0	0	1.004749	0	1	-360	360;
	3876	5999	0.003521	0.140801	0	0	0	0	1.004749	0	1	-360	360;
	3876	1	0.003521	0.140801	0	0	0	0	1.004729	0	1	-360	360;
	3876	4458	0.003521	0.140801	0	0	0	0	1.004729	0	1	-360	360;
	3876	11023	0.003521	0.140801	0	0	0	0	1.004729	0	1	-360	360;
	3876	12116	0.003521	0.140801	0	0	0	0	1.004729	0	1	-360	360;
	3876	11025	0.003521	0.140801	0	0	0	0	1.004729	0	1	-360	360;
	1015	11204	0.000531	0.021225	0	0	0	0	1.014683	0	1	-360	360;
	1015	11201	0.000531	0.021225	0	0	0	0	1.014683	0	1	-360	360;
	12395	6877	0.002119	0.084719	0	0	0	0	1.013169	0	1	-360	360;
	12395	6876	0.002119	0.084719	0	0	0	0	1.013169	0	1	-360	360;
	12395	6875	0.002119	0.084719	0	0	0	0	1.013169	0	1	-360	360;
	12395	12095	0.002119	0.084719	0	0	0	0	1.013169	0	1	-360	360;
	1395	2811	0.000531	0.021225	0	0	0	0	1.013854	0	1	-360	360;
	1395	6280	0.000531	0.021225	0	0	0	0	1.013854	0	1	-360	360;
	1395	8424	0.000531	0.021225	0	0	0	0	1.013854	0	1	-360	360;
	1395	8423	0.000531	0.021225	0	0	0	0	1.013854	0	1	-360	360;
	6133	858	0.000708	0.02832	0	0	0	0	1.014584	0	1	-360	360;
	6133	10908	0.000708	0.02832	0	0	0	0	1.014584	0	1	-360	360;
	2403	2606	0.000531	0.021225	0	0	0	0	1.00802	0	1	-360	360;
	2403	2607	0.000531	0.021225	0	0	0	0	1.00802	0	1	-360	360;
	2403	12727	0.002119	0.084719	0	0	0	0	1.0066	0	1	-360	360;
	2403	396	0.002119	0.084719	0	0	0	0	1.0066	0	1	-360	360;
	2403	8143	0.002119	0.084719	0	0	0	0	1.0066	0	1	-360	360;
	2403	6121	0.002119	0.084719	0	0	0	0	1.0066	0	1	-360	360;
	8264	7079	0.003521	0.140801	0	0	0	0	1.010644	0	1	-360	360;
	8264	12517	0.003521	0.140801	0	0	0	0	1.010644	0	1	-360	360;
	3187	4843	0.002119	0.084719	0	0	0	0	1.016068	0	1	-360	360;
	3187	4844	0.002119	0.084719	0	0	0	0	1.016068	0	1	-360	360;
	3187	4845	0.002119	0.084719	0	0	0	0	1.016068	0	1	-360	360;
	3187	4846	0.002119	0.084719	0	0	0	0	1.016068	0	1	-360	360;
	3187	4847	0.002119	0.084719	0	0	0	0	1.016068	0	1	-360	360;
	3187	4837	0.002119	0.084719	0	0	0	0	1.016068	0	1	-360	360;
	3187	1955	0.002119	0.084719	0	0	0	0	1.016068	0	1	-360	360;
	3187	9984	0.002119	0.084719	0	0	0	0	1.016068	0	1	-360	360;
	3187	3074	0.000531	0.021225	0	0	0	0	1.018397	0	1	-360	360;
	3187	3072	0.000531	0.021225	0	0	0	0	1.018397	0	1	-360	360;
	10536	6985	0.002119	0.084719	0	0	0	0	1.012036	0	1	-360	360;
	10536	6987	0.002119	0.084719	0	0	0	0	1.012036	0	1	-360	360;
	10536	12505	0.002119	0.084719	0	0	0	0	1.012036	0	1	-360	360;
	10536	12506	0.002119	0.084719	0	0	0	0	1.012036	0	1	-360	360;
	12737	5468	0.003521	0.140801	0	0	0	0	1.007774	0	1	-360	360;
	12737	5467	0.003521	0.140801	0	0	0	0	1.007774	0	1	-360	360;
	12737	5466	0.003521	0.140801	0	0	0	0	1.007774	0	1	-360	360;
	12737	4181	0.003521	0.140801	0	0	0	0	1.007774	0	1	-360	360;
	1404	12025	0.003521	0.140801	0	0	0	0	1.005553	0	1	-360	360;
	6538	781	0	0.04332	0	0	0	0	1.013449	0	1	-360	360;
	6538	781	0	0.04332	0	0	0	0	1.013449	0	1	-360	360;
	1161	13143	0	0.05435	0	0	0	0	1.020422	0	1	-360	360;
	1161	12001	0	0.05435	0	0	0	0	1.020422	0	1	-360	360;
	1161	6747	0	0.05435	0	0	0	0	1.020362	0	1	-360	360;
	7260	6831	0	0.0505	0	0	0	0	0.997208	0	1	-360	360;
	7260	10963	0	0.104	0	0	0	0	1.01095	0	1	-360	360;
	7260	12393	0	0.104	0	0	0	0	1.011333	0	1	-360	360;
	7260	6305	0	0.104	0	0	0	0	1.011333	0	1	-360	360;
	7260	11539	0	0.104	0	0	0	0	1.01095	0	1	-360	360;
	7260	12678	0	0.104	0	0	0	0	1.011333	0	1	-360	360;
	7260	13010	0	0.104	0	0	0	0	1.011333	0	1	-360	360;
	7260	9204	0	0.104	0	0	0	0	1.01095	0	1	-360	360;
	11	12266	0	0.05667	0	0	0	0	1.007055	0	1	-360	360;
	11	3201	0	0.05667	0	0	0	0	1.007055	0	1	-360	360;
	11	8816	0	0.04545	0	0	0	0	1.006593	0	1	-360	360;
	6831	9937	0	0.104	0	0	0	0	1.038514	0	1	-360	360;
	6831	8290	0	0.104	0	0	0	0	1.038514	0	1	-360	360;
	12420	12815	0	0.28571	0	0	0	0	1.0109	0	1	-360	360;
	6019	7473	0	0.6	0	0	0	0	1.044077	0	1	-360	360;
	6019	1903	0	2.85714	0	0	0	0	1.000888	0	1	-360	360;
	6019	5982	0	2.85714	0	0	0	0	1.000888	0	1	-360	360;
	6019	13165	0	2.85714	0	0	0	0	1.000888	0	1	-360	360;
	6019	10186	0	2.85714	0	0	0	0	1.000888	0	1	-360	360;
	6019	5162	0	1.2	0	0	0	0	1.043041	0	1	-360	360;
	6019	10466	0	1.2	0	0	0	0	1.043041	0	1	-360	360;
	6019	8493	0	2.57143	0	0	0	0	1.044066	0	1	-360	360;
	6019	324	0	2.57143	0	0	0	0	1.043911	0	1	-360	360;
	6019	13131	0	0.6	0	0	0	0	1.044077	0	1	-360	360;
	11918	4614	0	0.36	0	0	0	0	1.004596	0	1	-360	360;
	10797	12441	0	0.775	0	0	0	0	1.035164	0	1	-360	360;
	10797	7639	0	0.797	0	0	0	0	1.035183	0	1	-360	360;
	10797	11768	0	0.20567	0	0	0	0	1.001194	0	1	-360	360;
	10797	2906	0	0.80867	0	0	0	0	1.000505	0	1	-360	360;
	12089	9292	0	0.3	0	0	0	0	1.002898	0	1	-360	360;
	12089	10428	0	0.3	0	0	0	0	1.002898	0	1	-360	360;
	12089	2048	0	0.4	0	0	0	0	1.002952	0	1	-360	360;
	9313	9146	0	0.044	0	0	0	0	0.954853	0	1	-360	360;
	9313	9146	0	0.044	0	0	0	0	0.954853	0	1	-360	360;
	9313	6491	0	0.06667	0	0	0	0	0.959366	0	1	-360	360;
	9313	7535	0	0.06667	0	0	0	0	0.959366	0	1	-360	360;
	9313	12865	0	0.06	0	0	0	0	0.958638	0	1	-360	360;
	9640	3760	0	0.047975	0	0	0	0	0.950169	0	1	-360	360;
	9640	3760	0	0.047975	0	0	0	0	0.950169	0	1	-360	360;
	9640	8501	0	0.037831	0	0	0	0	0.953808	0	1	-360	360;
	9640	8501	0	0.037831	0	0	0	0	0.953808	0	1	-360	360;
	8364	7550	0	0.045294	0	0	0	0	0.982613	0	1	-360	360;
	8364	7550	0	0.045294	0	0	0	0	0.982613	0	1	-360	360;
	8364	4263	0	0.04563	0	0	0	0	0.965755	0	1	-360	360;
	8364	4263	0	0.04563	0	0	0	0	0.965755	0	1	-360	360;
	5348	4970	0	0.047975	0	0	0	0	0.954185	0	1	-360	360;
	5348	4970	0	0.047975	0	0	0	0	0.954185	0	1	-360	360;
	5348	9576	0	0.049246	0	0	0	0	0.947234	0	1	-360	360;
	5348	9576	0	0.049246	0	0	0	0	0.947234	0	1	-360	360;
	5348	13191	0	0.04444	0	0	0	0	0.988749	0	1	-360	360;
	5348	7327	0	0.04444	0	0	0	0	0.988749	0	1	-360	360;
	5348	2069	0	0.04138	0	0	0	0	0.988661	0	1	-360	360;
	12365	11707	0	0.049246	0	0	0	0	0.940146	0	1	-360	360;
	12365	11707	0	0.049246	0	0	0	0	0.940146	0	1	-360	360;
	13005	9058	0	0.046721	0	0	0	0	0.955135	0	1	-360	360;
	13005	7788	0	0.047975	0	0	0	0	0.959446	0	1	-360	360;
	9147	6616	0	0.06667	0	0	0	0	1.008237	0	1	-360	360;
	9147	7643	0	0.06667	0	0	0	0	1.04988	0	1	-360	360;
	9147	13002	0	0.06	0	0	0	0	1.049514	0	1	-360	360;
	1187	10353	0	0.047974	0	0	0	0	0.954972	0	1	-360	360;
	1187	10353	0	0.047974	0	0	0	0	0.954972	0	1	-360	360;
	3237	8887	0	0.26718	0	0	0	0	1.005967	0	1	-360	360;
	3237	10784	0	0.26542	0	0	0	0	1.005975	0	1	-360	360;
	3237	5542	0	0.26542	0	0	0	0	1.005975	0	1	-360	360;
	3237	11509	0	0.17674	0	0	0	0	1.00828	0	1	-360	360;
	4389	9395	0	0.16	0	0	0	0	1.007486	0	1	-360	360;
	3730	11944	0	0.049246	0	0	0	0	0.946864	0	1	-360	360;
	3730	11944	0	0.049246	0	0	0	0	0.946864	0	1	-360	360;
	3730	8037	0	0.047975	0	0	0	0	0.950137	0	1	-360	360;
	3730	8037	0	0.047975	0	0	0	0	0.950137	0	1	-360	360;
	10667	6585	0	0.09487	0	0	0	0	1.045129	0	1	-360	360;
	10667	1868	0	0.09487	0	0	0	0	1.045129	0	1	-360	360;
	10667	10239	0	0.10625	0	0	0	0	1.045245	0	1	-360	360;
	10667	10932	0	0.06486	0	0	0	0	1.045356	0	1	-360	360;
	10667	12983	0	0.06486	0	0	0	0	1.045356	0	1	-360	360;
	11896	13511	0	0.047975	0	0	0	0	0.94413	0	1	-360	360;
	11896	13511	0	0.047975	0	0	0	0	0.94413	0	1	-360	360;
	1187	9113	0	0.045482	0	0	0	0	0.959865	0	1	-360	360;
	1187	9113	0	0.045482	0	0	0	0	0.959865	0	1	-360	360;
	2273	1828	0	0.048	0	0	0	0	0.987462	0	1	-360	360;
	2273	1828	0	0.048	0	0	0	0	0.987462	0	1	-360	360;
	2273	5556	0	0.047975	0	0	0	0	0.945067	0	1	-360	360;
	2273	5556	0	0.047975	0	0	0	0	0.945067	0	1	-360	360;
	5861	2565	0	0.049041	0	0	0	0	0.942257	0	1	-360	360;
	5861	2565	0	0.049041	0	0	0	0	0.942257	0	1	-360	360;
	5861	6308	0	0.047975	0	0	0	0	0.947317	0	1	-360	360;
	5861	6308	0	0.047975	0	0	0	0	0.947317	0	1	-360	360;
	5861	5343	0	0.04444	0	0	0	0	1.000571	0	1	-360	360;
	5861	3369	0	0.04444	0	0	0	0	1.000571	0	1	-360	360;
	5861	7593	0	0.04138	0	0	0	0	1.000464	0	1	-360	360;
	12845	12328	0	0.049246	0	0	0	0	0.940793	0	1	-360	360;
	12845	12328	0	0.049246	0	0	0	0	0.940793	0	1	-360	360;
	2609	6292	0	0.047975	0	0	0	0	0.945337	0	1	-360	360;
	2609	6292	0	0.047975	0	0	0	0	0.945337	0	1	-360	360;
	2609	7466	0	0.049246	0	0	0	0	0.952901	0	1	-360	360;
	2609	7466	0	0.049246	0	0	0	0	0.952901	0	1	-360	360;
	940	473	0	0.05533	0	0	0	0	0.953216	0	1	-360	360;
	940	473	0	0.05533	0	0	0	0	0.953216	0	1	-360	360;
	940	473	0	0.05533	0	0	0	0	0.953216	0	1	-360	360;
	940	473	0	0.05533	0	0	0	0	0.953216	0	1	-360	360;
	2889	5314	0	0.0735	0	0	0	0	1.049171	0	1	-360	360;
	2889	365	0	0.07581	0	0	0	0	1.051303	0	1	-360	360;
	7166	11587	0	0.047975	0	0	0	0	0.944506	0	1	-360	360;
	7166	11587	0	0.047975	0	0	0	0	0.944506	0	1	-360	360;
	10066	12178	0	0.2	0	0	0	0	1.010856	0	1	-360	360;
	10066	12401	0	0.096	0	0	0	0	1.003951	0	1	-360	360;
	10066	11257	0	0.19355	0	0	0	0	1.00542	0	1	-360	360;
	13458	9855	0	1.5	0	0	0	0	1.00524	0	1	-360	360;
	13458	11177	0	1.2	0	0	0	0	1.003544	0	1	-360	360;
	13458	10012	0	1.2	0	0	0	0	1.003544	0	1	-360	360;
	13458	12358	0	0.72	0	0	0	0	1.003756	0	1	-360	360;
	13458	8933	0	0.72	0	0	0	0	1.001332	0	1	-360	360;
	13453	786	0	0.33333	0	0	0	0	1.00436	0	1	-360	360;
	13453	2852	0	0.33333	0	0	0	0	1.004412	0	1	-360	360;
	13453	7949	0	0.33333	0	0	0	0	1.002165	0	1	-360	360;
	6535	7449	0	0.04407	0	0	0	0	1.016847	0	1	-360	360;
	6535	6836	0	0.04407	0	0	0	0	1.016847	0	1	-360	360;
	6535	1085	0	0.04127	0	0	0	0	1.016745	0	1	-360	360;
	11937	2260	0	0.07869	0	0	0	0	1.049981	0	1	-360	360;
	11937	5965	0	0.07737	0	0	0	0	1.049567	0	1	-360	360;
	7206	1124	0	0.0505	0	0	0	0	0.926937	0	1	-360	360;
	7206	1124	0	0.0505	0	0	0	0	0.926937	0	1	-360	360;
	7206	1124	0	0.0505	0	0	0	0	0.926937	0	1	-360	360;
	1760	8352	0	0.3175	0	0	0	0	1.004459	0	1	-360	360;
	1760	7257	0	0.3175	0	0	0	0	1.004459	0	1	-360	360;
	1760	1959	0	0.3175	0	0	0	0	1.004543	0	1	-360	360;
	10858	10094	0	0.04407	0	0	0	0	1.016852	0	1	-360	360;
	10858	2759	0	0.04407	0	0	0	0	1.016852	0	1	-360	360;
	10858	8388	0	0.04127	0	0	0	0	1.016749	0	1	-360	360;
	10858	12291	0	0.04407	0	0	0	0	1.016852	0	1	-360	360;
	10858	7662	0	0.04407	0	0	0	0	1.016852	0	1	-360	360;
	10858	12248	0	0.04127	0	0	0	0	1.016749	0	1	-360	360;
	12809	6783	0	0.048	0	0	0	0	0.934758	0	1	-360	360;
	12809	6783	0	0.04563	0	0	0	0	0.944487	0	1	-360	360;
	12809	2042	0	0.049141	0	0	0	0	0.926102	0	1	-360	360;
	12809	2042	0	0.049141	0	0	0	0	0.926102	0	1	-360	360;
	11052	1007	0	0.6	0	0	0	0	0.999671	0	1	-360	360;
	6706	10973	0	0.075	0	0	0	0	1.004981	0	1	-360	360;
	12292	5742	0	0.044	0	0	0	0	1.069583	0	1	-360	360;
	9036	3682	0	0.72	0	0	0	0	0.999922	0	1	-360	360;
	9036	11967	0	2.57143	0	0	0	0	0.999699	0	1	-360	360;
	9036	10052	0	2.57143	0	0	0	0	0.999699	0	1	-360	360;
	9036	4337	0	0.57143	0	0	0	0	0.999727	0	1	-360	360;
	9036	8881	0	1.28571	0	0	0	0	0.999625	0	1	-360	360;
	9036	10757	0	1.28571	0	0	0	0	0.999625	0	1	-360	360;
	9036	9168	0	2.85714	0	0	0	0	0.99887	0	1	-360	360;
	9036	1004	0	2.85714	0	0	0	0	0.998989	0	1	-360	360;
	9036	2784	0	2.4	0	0	0	0	0.999025	0	1	-360	360;
	12897	6527	0	0.03908	0	0	0	0	0.959402	0	1	-360	360;
	12897	1822	0	0.03908	0	0	0	0	0.959402	0	1	-360	360;
	12897	9871	0	0.03908	0	0	0	0	0.959226	0	1	-360	360;
	9026	151	0	0.02908	0	0	0	0	1.006767	0	1	-360	360;
	9026	11856	0	0.02908	0	0	0	0	1.006641	0	1	-360	360;
	11710	4353	0	0.13618	0	0	0	0	1.000365	0	1	-360	360;
	11710	9628	0	0.13047	0	0	0	0	1.000356	0	1	-360	360;
	11710	3593	0	0.13163	0	0	0	0	1.000327	0	1	-360	360;
	10277	9024	0	0.02908	0	0	0	0	1.002347	0	1	-360	360;
	2152	8433	0	0.25714	0	0	0	0	0.998938	0	1	-360	360;
	2152	9552	0	0.45	0	0	0	0	0.999062	0	1	-360	360;
	2152	13075	0	6	0	0	0	0	0.998562	0	1	-360	360;
	2152	315	0	6	0	0	0	0	0.998562	0	1	-360	360;
	2683	5044	0	0.067933	0	0	0	0	0.956303	0	1	-360	360;
	2683	5044	0	0.067933	0	0	0	0	0.956303	0	1	-360	360;
	2683	5044	0	0.047975	0	0	0	0	0.932939	0	1	-360	360;
	2226	7487	0	0.1978	0	0	0	0	1.006462	0	1	-360	360;
	4867	8193	0	0.232	0	0	0	0	1.003527	0	1	-360	360;
	6791	8385	0	0.51429	0	0	0	0	1.000793	0	1	-360	360;
	6791	11281	0	0.9	0	0	0	0	1.000918	0	1	-360	360;
	5266	6132	0	0.58065	0	0	0	0	1.003186	0	1	-360	360;
	7868	9722	0	0.07095	0	0	0	0	1.041895	0	1	-360	360;
	7868	10825	0	0.07095	0	0	0	0	1.041895	0	1	-360	360;
	10758	2912	0	0.18462	0	0	0	0	1.004094	0	1	-360	360;
	10758	11502	0	0.36667	0	0	0	0	1.00425	0	1	-360	360;
	8508	10368	0	0.08	0	0	0	0	0.995908	0	1	-360	360;
	7720	7255	0	0.046721	0	0	0	0	0.940126	0	1	-360	360;
	7720	8846	0	0.06722	0	0	0	0	0.95944	0	1	-360	360;
	7720	9972	0	0.06722	0	0	0	0	0.95944	0	1	-360	360;
	10885	13630	0	0.070733	0	0	0	0	0.929076	0	1	-360	360;
	10885	13630	0	0.07187	0	0	0	0	0.949888	0	1	-360	360;
	10885	2990	0	0.06667	0	0	0	0	0.958813	0	1	-360	360;
	11555	6383	0	0.34286	0	0	0	0	1.003452	0	1	-360	360;
	11555	7250	0	0.72	0	0	0	0	1.003928	0	1	-360	360;
	11555	10409	0	1.904818	0	0	0	0	1.000924	0	1	-360	360;
	11555	6009	0	3	0	0	0	0	1.000311	0	1	-360	360;
	11555	432	0	3	0	0	0	0	1.000311	0	1	-360	360;
	7255	4208	0	0.07217	0	0	0	0	1.00268	0	1	-360	360;
	10397	7340	0	0.06111	0	0	0	0	1.103098	0	1	-360	360;
	6771	6171	0	0.44444	0	0	0	0	1.005159	0	1	-360	360;
	6771	4235	0	0.44444	0	0	0	0	1.005159	0	1	-360	360;
	6771	13447	0	0.44444	0	0	0	0	1.005159	0	1	-360	360;
	13258	1551	0	0.049246	0	0	0	0	0.933179	0	1	-360	360;
	13258	1551	0	0.049246	0	0	0	0	0.933179	0	1	-360	360;
	13258	1551	0	0.049246	0	0	0	0	0.933179	0	1	-360	360;
	10601	5397	0	0.21315	0	0	0	0	1.000795	0	1	-360	360;
	10601	3432	0	0.21315	0	0	0	0	1.000795	0	1	-360	360;
	10601	12687	0	0.21315	0	0	0	0	1.000629	0	1	-360	360;
	8778	9735	0	0.0505	0	0	0	0	0.937597	0	1	-360	360;
	8778	9735	0	0.0505	0	0	0	0	0.937597	0	1	-360	360;
	8778	6152	0	0.046807	0	0	0	0	0.954913	0	1	-360	360;
	8778	6152	0	0.046807	0	0	0	0	0.954913	0	1	-360	360;
	8778	4408	0	0.049246	0	0	0	0	0.944387	0	1	-360	360;
	8778	4408	0	0.049246	0	0	0	0	0.944387	0	1	-360	360;
	6700	3920	0	0.047975	0	0	0	0	0.943562	0	1	-360	360;
	6700	3920	0	0.047975	0	0	0	0	0.943562	0	1	-360	360;
	6700	13624	0	0.0505	0	0	0	0	0.940759	0	1	-360	360;
	6700	13624	0	0.0505	0	0	0	0	0.940759	0	1	-360	360;
	3808	4320	0	0.047975	0	0	0	0	0.946595	0	1	-360	360;
	3808	4320	0	0.047975	0	0	0	0	0.946595	0	1	-360	360;
	3808	4320	0	0.047975	0	0	0	0	0.946595	0	1	-360	360;
	3808	4320	0	0.047975	0	0	0	0	0.946595	0	1	-360	360;
	9735	7857	0	0.5	0	0	0	0	1.003651	0	1	-360	360;
	9735	6784	0	0.5	0	0	0	0	1.003651	0	1	-360	360;
	9735	1469	0	0.848	0	0	0	0	1.003056	0	1	-360	360;
	9735	2397	0	0.848	0	0	0	0	1.003056	0	1	-360	360;
	3370	2645	0	0.3	0	0	0	0	1.000521	0	1	-360	360;
	3370	9466	0	0.58065	0	0	0	0	1.00095	0	1	-360	360;
	3370	9223	0	3.6	0	0	0	0	1.000776	0	1	-360	360;
	4565	12657	0	2.57143	0	0	0	0	1.000314	0	1	-360	360;
	4565	4643	0	1.38462	0	0	0	0	1.000888	0	1	-360	360;
	4565	10010	0	1.38462	0	0	0	0	1.000888	0	1	-360	360;
	4565	11945	0	1.38462	0	0	0	0	1.000888	0	1	-360	360;
	4565	7768	0	1.2	0	0	0	0	1.005163	0	1	-360	360;
	4565	2278	0	2.85714	0	0	0	0	1.000282	0	1	-360	360;
	4565	10693	0	2.57143	0	0	0	0	1.000314	0	1	-360	360;
	4565	6942	0	2.57143	0	0	0	0	1.001212	0	1	-360	360;
	12068	1594	0	0.2	0	0	0	0	1.005297	0	1	-360	360;
	12068	2603	0	0.2	0	0	0	0	1.005297	0	1	-360	360;
	12068	7921	0	0.64	0	0	0	0	1.009116	0	1	-360	360;
	12068	7009	0	0.64	0	0	0	0	1.008576	0	1	-360	360;
	12251	13090	0	0.04407	0	0	0	0	1.014872	0	1	-360	360;
	12251	11492	0	0.04407	0	0	0	0	1.014872	0	1	-360	360;
	12251	3179	0	0.04127	0	0	0	0	1.01484	0	1	-360	360;
	12251	1150	0	0.04407	0	0	0	0	1.014872	0	1	-360	360;
	12251	9364	0	0.04407	0	0	0	0	1.014872	0	1	-360	360;
	12251	7489	0	0.04127	0	0	0	0	1.01484	0	1	-360	360;
	11740	12873	0	0.4375	0	0	0	0	1.001373	0	1	-360	360;
	11740	7213	0	0.4375	0	0	0	0	1.001373	0	1	-360	360;
	11740	11138	0	0.20635	0	0	0	0	1.000822	0	1	-360	360;
	11740	10006	0	0.20635	0	0	0	0	1.000822	0	1	-360	360;
	11740	958	0	0.20635	0	0	0	0	1.000822	0	1	-360	360;
	11740	13550	0	0.20635	0	0	0	0	1.000822	0	1	-360	360;
	11740	8141	0	0.4127	0	0	0	0	1.001174	0	1	-360	360;
	11740	7068	0	0.4127	0	0	0	0	1.001174	0	1	-360	360;
	11740	12375	0	0.05778	0	0	0	0	1.000444	0	1	-360	360;
	3973	10061	0	0.05643	0	0	0	0	0.97594	0	1	-360	360;
	3973	10061	0	0.040627	0	0	0	0	0.996982	0	1	-360	360;
	3973	3178	0	0.0565	0	0	0	0	1.047506	0	1	-360	360;
	3973	8499	0	0.0565	0	0	0	0	1.047506	0	1	-360	360;
	3973	10473	0	0.0565	0	0	0	0	1.047506	0	1	-360	360;
	3973	2106	0	0.0565	0	0	0	0	1.047506	0	1	-360	360;
	12660	12989	0	0.25714	0	0	0	0	1.004944	0	1	-360	360;
	12660	717	0	0.36	0	0	0	0	1.004784	0	1	-360	360;
	12660	2268	0	0.58065	0	0	0	0	1.004857	0	1	-360	360;
	1280	6199	0	0.31579	0	0	0	0	1.003857	0	1	-360	360;
	6755	2223	0	0.32024	0	0	0	0	1.013741	0	1	-360	360;
	6755	8157	0	0.32024	0	0	0	0	1.013741	0	1	-360	360;
	6755	13549	0	0.32024	0	0	0	0	1.013741	0	1	-360	360;
	3841	602	0	0.049246	0	0	0	0	0.949121	0	1	-360	360;
	3841	602	0	0.049246	0	0	0	0	0.949121	0	1	-360	360;
	602	12189	0	0.6	0	0	0	0	1.007356	0	1	-360	360;
	602	3880	0	0.6	0	0	0	0	1.007356	0	1	-360	360;
	602	1907	0	0.6	0	0	0	0	1.007356	0	1	-360	360;
	602	4196	0	2.4	0	0	0	0	1.001354	0	1	-360	360;
	602	6130	0	2.4	0	0	0	0	1.001354	0	1	-360	360;
	6504	1526	0	0.476	0	0	0	0	1.044767	0	1	-360	360;
	6504	4823	0	0.476	0	0	0	0	1.044767	0	1	-360	360;
	10586	6038	0	0.06	0	0	0	0	0.999935	0	1	-360	360;
	10586	3926	0	0.06	0	0	0	0	1.00054	0	1	-360	360;
	10586	12367	0	0.06	0	0	0	0	0.999935	0	1	-360	360;
	1297	5818	0	0.092641	0	0	0	0	0.954139	0	1	-360	360;
	1297	5818	0	0.092641	0	0	0	0	0.954139	0	1	-360	360;
	1297	5818	0	0.05415	0	0	0	0	0.957465	0	1	-360	360;
	1297	12199	0	0.05478	0	0	0	0	1.000751	0	1	-360	360;
	1297	6822	0	0.05478	0	0	0	0	1.000751	0	1	-360	360;
	12110	2567	0	0.044468	0	0	0	0	0.960487	0	1	-360	360;
	12110	2567	0	0.044468	0	0	0	0	0.960487	0	1	-360	360;
	12110	9195	0	0.05439	0	0	0	0	0.999891	0	1	-360	360;
	12110	3939	0	0.05439	0	0	0	0	0.999891	0	1	-360	360;
	6037	5202	0	0.28806	0	0	0	0	1.004211	0	1	-360	360;
	6037	552	0	0.28806	0	0	0	0	1.004211	0	1	-360	360;
	10259	4423	0	4.5	0	0	0	0	0.999295	0	1	-360	360;
	10259	8010	0	0.72	0	0	0	0	1.000103	0	1	-360	360;
	10259	6200	0	1.44	0	0	0	0	0.999879	0	1	-360	360;
	10259	6864	0	4.5	0	0	0	0	0.999295	0	1	-360	360;
	10259	8654	0	4.5	0	0	0	0	0.999295	0	1	-360	360;
	10259	7261	0	4.5	0	0	0	0	0.999295	0	1	-360	360;
	10259	9413	0	4.5	0	0	0	0	0.999295	0	1	-360	360;
	10259	2341	0	4.5	0	0	0	0	0.999295	0	1	-360	360;
	10259	13551	0	4.5	0	0	0	0	0.999295	0	1	-360	360;
	10259	11440	0	4.5	0	0	0	0	0.999295	0	1	-360	360;
	10259	8916	0	4.5	0	0	0	0	0.999295	0	1	-360	360;
	10259	6314	0	4.5	0	0	0	0	0.999295	0	1	-360	360;
	10259	8122	0	4.5	0	0	0	0	0.999295	0	1	-360	360;
	10259	12748	0	4.5	0	0	0	0	0.999295	0	1	-360	360;
	10259	11311	0	3.6	0	0	0	0	0.997684	0	1	-360	360;
	10259	4488	0	3.6	0	0	0	0	0.997684	0	1	-360	360;
	10259	9748	0	3.6	0	0	0	0	0.997684	0	1	-360	360;
	2096	1642	0	0.72	0	0	0	0	1.001132	0	1	-360	360;
	12975	10988	0	0.047975	0	0	0	0	0.939465	0	1	-360	360;
	12975	10988	0	0.047975	0	0	0	0	0.939465	0	1	-360	360;
	12975	10988	0	0.047975	0	0	0	0	0.939465	0	1	-360	360;
	10988	94	0	0.375	0	0	0	0	1.00764	0	1	-360	360;
	1579	9166	0	0.4969	0	0	0	0	1.001278	0	1	-360	360;
	6966	7416	0	0.37826	0	0	0	0	1.001164	0	1	-360	360;
	6966	12816	0	0.37826	0	0	0	0	1.001164	0	1	-360	360;
	6966	270	0	0.37826	0	0	0	0	1.001164	0	1	-360	360;
	484	12499	0	0.72	0	0	0	0	1.007465	0	1	-360	360;
	2188	2721	0	0.69231	0	0	0	0	1.006393	0	1	-360	360;
	10975	7879	0	0.047975	0	0	0	0	0.953084	0	1	-360	360;
	10975	7879	0	0.047975	0	0	0	0	0.953084	0	1	-360	360;
	10975	7879	0	0.047975	0	0	0	0	0.953084	0	1	-360	360;
	10975	4085	0	2.57143	0	0	0	0	0.999195	0	1	-360	360;
	10975	11859	0	2.57143	0	0	0	0	0.999195	0	1	-360	360;
	10975	1329	0	2.57143	0	0	0	0	0.999195	0	1	-360	360;
	10975	9804	0	0.72	0	0	0	0	0.999	0	1	-360	360;
	10975	9638	0	0.72	0	0	0	0	0.999	0	1	-360	360;
	2037	8559	0	0.16	0	0	0	0	0.999995	0	1	-360	360;
	2037	13200	0	0.16	0	0	0	0	0.999995	0	1	-360	360;
	2824	11990	0	0.48	0	0	0	0	1.003518	0	1	-360	360;
	2824	3371	0	0.48	0	0	0	0	1.003518	0	1	-360	360;
	10109	779	0	0.48	0	0	0	0	0.997126	0	1	-360	360;
	10109	8980	0	0.48	0	0	0	0	0.997126	0	1	-360	360;
	215	8514	0	0.06556	0	0	0	0	0.953866	0	1	-360	360;
	215	8514	0	0.0472	0	0	0	0	0.955722	0	1	-360	360;
	215	13218	0	0.07873	0	0	0	0	0.960191	0	1	-360	360;
	215	7751	0	0.07873	0	0	0	0	0.960191	0	1	-360	360;
	215	5964	0	0.07873	0	0	0	0	0.960191	0	1	-360	360;
	215	12050	0	0.07873	0	0	0	0	0.960191	0	1	-360	360;
	7953	9965	0	0.075233	0	0	0	0	0.934284	0	1	-360	360;
	7953	9965	0	0.043977	0	0	0	0	0.941255	0	1	-360	360;
	7953	9965	0	0.043977	0	0	0	0	0.941255	0	1	-360	360;
	7953	9965	0	0.043977	0	0	0	0	0.941255	0	1	-360	360;
	10274	746	0	0.6	0	0	0	0	1.005532	0	1	-360	360;
	10274	13112	0	0.6	0	0	0	0	1.005865	0	1	-360	360;
	10490	10955	0	0.047783	0	0	0	0	0.944157	0	1	-360	360;
	10490	10955	0	0.046807	0	0	0	0	0.944034	0	1	-360	360;
	11742	6042	0	0.071122	0	0	0	0	0.940129	0	1	-360	360;
	11742	6042	0	0.071122	0	0	0	0	0.940129	0	1	-360	360;
	11742	6042	0	0.07796	0	0	0	0	0.938946	0	1	-360	360;
	11742	6042	0	0.07796	0	0	0	0	0.938946	0	1	-360	360;
	1569	8488	0	0.78261	0	0	0	0	1.005671	0	1	-360	360;
	1569	6192	0	0.5	0	0	0	0	1.00581	0	1	-360	360;
	1569	4796	0	0.58065	0	0	0	0	1.005662	0	1	-360	360;
	1119	9504	0	0.17857	0	0	0	0	1.041054	0	1	-360	360;
	1119	11630	0	0.17857	0	0	0	0	1.041054	0	1	-360	360;
	1119	2237	0	0.17857	0	0	0	0	1.041054	0	1	-360	360;
	2589	10783	0	0.13139	0	0	0	0	1.008548	0	1	-360	360;
	2765	10814	0	0.064	0	0	0	0	1.057613	0	1	-360	360;
	2765	5576	0	0.064	0	0	0	0	1.057613	0	1	-360	360;
	2765	7381	0	0.064	0	0	0	0	1.057613	0	1	-360	360;
	10781	2428	0	0.18462	0	0	0	0	1.022911	0	1	-360	360;
	10781	6680	0	0.18462	0	0	0	0	1.022911	0	1	-360	360;
	10781	11935	0	0.18462	0	0	0	0	1.022911	0	1	-360	360;
	8624	4124	0	0.046721	0	0	0	0	0.953764	0	1	-360	360;
	4124	2131	0	0.26714	0	0	0	0	1.032058	0	1	-360	360;
	4124	3968	0	0.26714	0	0	0	0	1.032058	0	1	-360	360;
	4124	12398	0	0.26714	0	0	0	0	1.032058	0	1	-360	360;
	4124	296	0	0.16422	0	0	0	0	1.024502	0	1	-360	360;
	4124	4917	0	0.14687	0	0	0	0	1.024545	0	1	-360	360;
	4124	2933	0	0.29375	0	0	0	0	1.024756	0	1	-360	360;
	7292	12829	0	0.15856	0	0	0	0	1.091237	0	1	-360	360;
	7292	10298	0	0.15856	0	0	0	0	1.091237	0	1	-360	360;
	9639	10376	0	0.14697	0	0	0	0	1.106586	0	1	-360	360;
	4017	7556	0	0.11733	0	0	0	0	1.049659	0	1	-360	360;
	4017	8644	0	0.11733	0	0	0	0	1.049659	0	1	-360	360;
	4017	400	0	0.11733	0	0	0	0	1.049659	0	1	-360	360;
	10018	8485	0	0.047975	0	0	0	0	0.934634	0	1	-360	360;
	10018	278	0	0.02519	0	0	0	0	0.993885	0	1	-360	360;
	10018	12579	0	0.02519	0	0	0	0	0.993885	0	1	-360	360;
	191	3200	0	0.046721	0	0	0	0	0.955214	0	1	-360	360;
	3200	3738	0	0.35294	0	0	0	0	1.005834	0	1	-360	360;
	2762	7705	0	0.18138	0	0	0	0	1.00027	0	1	-360	360;
	2762	5826	0	0.37167	0	0	0	0	1.000451	0	1	-360	360;
	11766	2215	0	0.18369	0	0	0	0	1.002293	0	1	-360	360;
	11766	10295	0	0.18369	0	0	0	0	1.002293	0	1	-360	360;
	11766	12949	0	0.18462	0	0	0	0	1.002308	0	1	-360	360;
	7883	2270	0	0.48	0	0	0	0	1.002552	0	1	-360	360;
	13168	8413	0	0.24	0	0	0	0	1.005389	0	1	-360	360;
	2477	13	0	0.046721	0	0	0	0	0.915966	0	1	-360	360;
	2477	7743	0	0.06667	0	0	0	0	0.998525	0	1	-360	360;
	2477	12287	0	0.06667	0	0	0	0	0.998525	0	1	-360	360;
	2477	539	0	0.06667	0	0	0	0	0.998558	0	1	-360	360;
	2477	8830	0	0.06667	0	0	0	0	0.998558	0	1	-360	360;
	9769	2071	0	0.13491	0	0	0	0	0.972785	0	1	-360	360;
	9769	294	0	0.13491	0	0	0	0	0.972785	0	1	-360	360;
	3045	3562	0	0.08407	0	0	0	0	0.94294	0	1	-360	360;
	3045	3562	0	0.08407	0	0	0	0	0.94294	0	1	-360	360;
	3045	3562	0	0.08407	0	0	0	0	0.94294	0	1	-360	360;
	7209	8808	0	0.050122	0	0	0	0	0.921038	0	1	-360	360;
	7209	8808	0	0.050122	0	0	0	0	0.921038	0	1	-360	360;
	7209	5233	0	0.047176	0	0	0	0	0.93698	0	1	-360	360;
	7209	5233	0	0.047176	0	0	0	0	0.93698	0	1	-360	360;
	7769	8166	0	0.046721	0	0	0	0	0.935115	0	1	-360	360;
	7769	6958	0	0.046721	0	0	0	0	0.949484	0	1	-360	360;
	11188	10433	0	1.28571	0	0	0	0	1.00172	0	1	-360	360;
	11188	10237	0	3.6	0	0	0	0	1.000963	0	1	-360	360;
	11188	5000	0	3.6	0	0	0	0	1.000963	0	1	-360	360;
	3317	3315	0	0.053184	0	0	0	0	0.921665	0	1	-360	360;
	3317	3315	0	0.053184	0	0	0	0	0.921665	0	1	-360	360;
	3317	3315	0	0.053184	0	0	0	0	0.921665	0	1	-360	360;
	2704	11970	0	0.46	0	0	0	0	1.000449	0	1	-360	360;
	2704	11329	0	0.46	0	0	0	0	1.000449	0	1	-360	360;
	9369	7577	0	0.47727	0	0	0	0	1.011462	0	1	-360	360;
	9369	12966	0	0.47727	0	0	0	0	1.011462	0	1	-360	360;
	9369	1245	0	0.47727	0	0	0	0	1.011462	0	1	-360	360;
	6730	12658	0	0.203316	0	0	0	0	1.088237	0	1	-360	360;
	6730	5484	0	0.203316	0	0	0	0	1.088237	0	1	-360	360;
	6730	9975	0	0.230069	0	0	0	0	1.088163	0	1	-360	360;
	6730	12736	0	0.230069	0	0	0	0	1.088163	0	1	-360	360;
	8166	3294	0	0.05072	0	0	0	0	1.044425	0	1	-360	360;
	6958	10211	0	0.05072	0	0	0	0	1.039783	0	1	-360	360;
	6958	12204	0	0.03627	0	0	0	0	1.038345	0	1	-360	360;
	5533	10951	0	0.1975	0	0	0	0	1.035836	0	1	-360	360;
	5533	9836	0	0.61222	0	0	0	0	1.03548	0	1	-360	360;
	1652	12608	0	0.0505	0	0	0	0	0.935308	0	1	-360	360;
	1652	12608	0	0.0505	0	0	0	0	0.935308	0	1	-360	360;
	11647	12154	0	0.119999	0	0	0	0	0.929754	0	1	-360	360;
	11647	12154	0	0.119999	0	0	0	0	0.929754	0	1	-360	360;
	10880	10276	0	0.877958	0	0	0	0	1.044843	0	1	-360	360;
	10880	9151	0	0.877958	0	0	0	0	1.044843	0	1	-360	360;
	10880	3848	0	0.877958	0	0	0	0	1.044843	0	1	-360	360;
	10777	4704	0	0.114122	0	0	0	0	1.05908	0	1	-360	360;
	10777	8060	0	0.114122	0	0	0	0	1.05908	0	1	-360	360;
	639	11020	0	1.366834	0	0	0	0	1.002255	0	1	-360	360;
	639	10114	0	1.366834	0	0	0	0	1.002255	0	1	-360	360;
	639	13617	0	1.366834	0	0	0	0	1.002255	0	1	-360	360;
	5985	4331	0	0.19048	0	0	0	0	1.00432	0	1	-360	360;
	5985	3567	0	0.19048	0	0	0	0	1.00432	0	1	-360	360;
	6395	6396	0	0.074556	0	0	0	0	0.942552	0	1	-360	360;
	6395	6396	0	0.074556	0	0	0	0	0.942552	0	1	-360	360;
	6395	6396	0	0.074556	0	0	0	0	0.942552	0	1	-360	360;
	6395	6396	0	0.078311	0	0	0	0	0.941811	0	1	-360	360;
	13337	7710	0	0.04602	0	0	0	0	0.948826	0	1	-360	360;
	13337	7710	0	0.08441	0	0	0	0	0.953818	0	1	-360	360;
	5684	7641	0	0.2685	0	0	0	0	0.999855	0	1	-360	360;
	5684	1059	0	0.2685	0	0	0	0	0.999855	0	1	-360	360;
	5684	4267	0	0.2685	0	0	0	0	0.999855	0	1	-360	360;
	5684	13195	0	0.2685	0	0	0	0	0.999855	0	1	-360	360;
	12215	8023	0	0.24895	0	0	0	0	1.043206	0	1	-360	360;
	12215	13096	0	0.24895	0	0	0	0	1.043206	0	1	-360	360;
	598	13159	0	1.5	0	0	0	0	1.000288	0	1	-360	360;
	11683	3184	0	0.375	0	0	0	0	1.001588	0	1	-360	360;
	11683	2122	0	0.375	0	0	0	0	1.001588	0	1	-360	360;
	11683	10501	0	0.375	0	0	0	0	1.001588	0	1	-360	360;
	1942	11059	0	0.2	0	0	0	0	1.001002	0	1	-360	360;
	1942	9995	0	0.2	0	0	0	0	1.001002	0	1	-360	360;
	1942	4775	0	0.2	0	0	0	0	1.001002	0	1	-360	360;
	1942	7329	0	0.24	0	0	0	0	1.001052	0	1	-360	360;
	2748	12569	0	0.06561	0	0	0	0	0.952731	0	1	-360	360;
	2748	12569	0	0.07333	0	0	0	0	0.952342	0	1	-360	360;
	2748	5418	0	0.8	0	0	0	0	0.999778	0	1	-360	360;
	6270	10057	0	0.16667	0	0	0	0	0.999878	0	1	-360	360;
	6270	5645	0	0.16667	0	0	0	0	0.999878	0	1	-360	360;
	6270	3686	0	0.16667	0	0	0	0	0.999878	0	1	-360	360;
	2162	8728	0	0.049246	0	0	0	0	0.946587	0	1	-360	360;
	2162	8728	0	0.049246	0	0	0	0	0.946587	0	1	-360	360;
	13050	2566	0	0.2	0	0	0	0	1.000358	0	1	-360	360;
	8641	8930	0	0.047975	0	0	0	0	0.948254	0	1	-360	360;
	8641	8930	0	0.082076	0	0	0	0	0.938336	0	1	-360	360;
	8641	8930	0	0.047975	0	0	0	0	0.948254	0	1	-360	360;
	4474	13087	0	0.56833	0	0	0	0	1.000658	0	1	-360	360;
	4474	3488	0	0.56833	0	0	0	0	1.000658	0	1	-360	360;
	7119	7430	0	0.3125	0	0	0	0	1.000861	0	1	-360	360;
	7119	5494	0	0.3125	0	0	0	0	1.000861	0	1	-360	360;
	8539	8896	0	0.34483	0	0	0	0	1.001408	0	1	-360	360;
	8539	7052	0	0.34276	0	0	0	0	1.001407	0	1	-360	360;
	8026	9927	0	0.076827	0	0	0	0	0.958258	0	1	-360	360;
	8026	9927	0	0.049734	0	0	0	0	0.934397	0	1	-360	360;
	8026	51	0	0.06477	0	0	0	0	0.952478	0	1	-360	360;
	8026	1821	0	0.06477	0	0	0	0	0.952478	0	1	-360	360;
	8026	10177	0	0.06477	0	0	0	0	0.952478	0	1	-360	360;
	2129	8449	0	0.0808	0	0	0	0	0.974141	0	1	-360	360;
	2129	8449	0	0.073292	0	0	0	0	0.933547	0	1	-360	360;
	2129	8449	0	0.08451	0	0	0	0	0.948635	0	1	-360	360;
	2129	8449	0	0.073292	0	0	0	0	0.933547	0	1	-360	360;
	2129	4261	0	0.04444	0	0	0	0	0.987595	0	1	-360	360;
	2129	299	0	0.04444	0	0	0	0	0.987595	0	1	-360	360;
	2129	8561	0	0.04138	0	0	0	0	0.987505	0	1	-360	360;
	10819	2860	0	0.047975	0	0	0	0	0.936384	0	1	-360	360;
	10819	2860	0	0.047975	0	0	0	0	0.936384	0	1	-360	360;
	4820	8179	0	0.1156	0	0	0	0	0.957383	0	1	-360	360;
	4820	2717	0	0.1156	0	0	0	0	0.957327	0	1	-360	360;
	4820	5905	0	0.1156	0	0	0	0	0.957177	0	1	-360	360;
	11964	4983	0	0.96	0	0	0	0	1.000142	0	1	-360	360;
	5820	671	0	0.44643	0	0	0	0	1.006575	0	1	-360	360;
	5820	1180	0	0.44643	0	0	0	0	1.006575	0	1	-360	360;
	5820	12558	0	1.24627	0	0	0	0	1.008042	0	1	-360	360;
	4	8857	0	0.48	0	0	0	0	1.004449	0	1	-360	360;
	4	572	0	0.48	0	0	0	0	1.004449	0	1	-360	360;
	4	2460	0	0.24	0	0	0	0	1.004638	0	1	-360	360;
	4	7777	0	0.24	0	0	0	0	1.007068	0	1	-360	360;
	12356	3385	0	0.60364	0	0	0	0	1.002099	0	1	-360	360;
	12356	8837	0	0.60061	0	0	0	0	1.002095	0	1	-360	360;
	12356	10945	0	0.776	0	0	0	0	1.00235	0	1	-360	360;
	767	9227	0	0.24773	0	0	0	0	1.001528	0	1	-360	360;
	767	8121	0	0.24773	0	0	0	0	1.001528	0	1	-360	360;
	767	2831	0	0.24773	0	0	0	0	1.001082	0	1	-360	360;
	9300	10156	0	0.06667	0	0	0	0	1.042646	0	1	-360	360;
	9300	9054	0	0.06667	0	0	0	0	1.042646	0	1	-360	360;
	3507	331	0	0.24	0	0	0	0	1.006117	0	1	-360	360;
	3507	5716	0	0.24	0	0	0	0	1.006117	0	1	-360	360;
	5088	5574	0	0.070727	0	0	0	0	0.953559	0	1	-360	360;
	5088	5574	0	0.072561	0	0	0	0	0.953467	0	1	-360	360;
	7658	6649	0	0.895631	0	0	0	0	1.002677	0	1	-360	360;
	7658	11850	0	0.895631	0	0	0	0	1.002677	0	1	-360	360;
	7658	258	0	0.895631	0	0	0	0	1.002677	0	1	-360	360;
	7658	6952	0	0.895631	0	0	0	0	1.002677	0	1	-360	360;
	7658	10940	0	0.20778	0	0	0	0	1.010413	0	1	-360	360;
	5363	4918	0	0.25	0	0	0	0	1.0252	0	1	-360	360;
	5363	13222	0	0.25	0	0	0	0	1.0252	0	1	-360	360;
	3513	7707	0	0.18462	0	0	0	0	1.002954	0	1	-360	360;
	3513	559	0	0.18462	0	0	0	0	1.002954	0	1	-360	360;
	10067	10572	0	0.047975	0	0	0	0	0.94873	0	1	-360	360;
	10067	10572	0	0.047975	0	0	0	0	0.94873	0	1	-360	360;
	10067	12603	0	0.06154	0	0	0	0	0.989901	0	1	-360	360;
	10067	11665	0	0.06154	0	0	0	0	0.989901	0	1	-360	360;
	2629	3948	0	0.16	0	0	0	0	1.003023	0	1	-360	360;
	2629	13238	0	0.16	0	0	0	0	1.003023	0	1	-360	360;
	8462	8252	0	0.07575	0	0	0	0	0.93962	0	1	-360	360;
	8462	8252	0	0.049246	0	0	0	0	0.949671	0	1	-360	360;
	8462	8252	0	0.044468	0	0	0	0	0.9877	0	1	-360	360;
	8462	9049	0	0.04444	0	0	0	0	0.987095	0	1	-360	360;
	8462	12775	0	0.04444	0	0	0	0	0.987095	0	1	-360	360;
	8462	8635	0	0.04138	0	0	0	0	0.987064	0	1	-360	360;
	10673	10153	0	0.082076	0	0	0	0	0.930061	0	1	-360	360;
	10673	10153	0	0.046721	0	0	0	0	0.94135	0	1	-360	360;
	7680	1985	0	0.19045	0	0	0	0	1.001671	0	1	-360	360;
	7680	954	0	0.19478	0	0	0	0	1.001623	0	1	-360	360;
	931	10557	0	1	0	0	0	0	1.025593	0	1	-360	360;
	931	5384	0	1	0	0	0	0	1.025593	0	1	-360	360;
	931	7186	0	1	0	0	0	0	1.025593	0	1	-360	360;
	931	9554	0	1	0	0	0	0	1.025425	0	1	-360	360;
	10266	8584	0	0.073292	0	0	0	0	0.934869	0	1	-360	360;
	10266	8584	0	0.07618	0	0	0	0	0.959473	0	1	-360	360;
	474	25	0	0.034784	0	0	0	0	0.957864	0	1	-360	360;
	474	25	0	0.041737	0	0	0	0	0.956293	0	1	-360	360;
	474	1264	0	0.04762	0	0	0	0	0.95327	0	1	-360	360;
	474	10554	0	0.04762	0	0	0	0	0.953224	0	1	-360	360;
	474	8483	0	0.04762	0	0	0	0	0.953224	0	1	-360	360;
	474	12688	0	0.04762	0	0	0	0	0.953224	0	1	-360	360;
	474	4202	0	0.04762	0	0	0	0	0.953224	0	1	-360	360;
	474	2308	0	0.04762	0	0	0	0	0.953239	0	1	-360	360;
	1036	1510	0	0.0842	0	0	0	0	0.953022	0	1	-360	360;
	1036	1510	0	0.0842	0	0	0	0	0.953022	0	1	-360	360;
	1036	8453	0	0.04321	0	0	0	0	0.995668	0	1	-360	360;
	4116	5403	0	0.048	0	0	0	0	0.933396	0	1	-360	360;
	4116	5403	0	0.048	0	0	0	0	0.933396	0	1	-360	360;
	4116	5403	0	0.048	0	0	0	0	0.933396	0	1	-360	360;
	4116	5403	0	0.048	0	0	0	0	0.933396	0	1	-360	360;
	5743	11594	0	0.080704	0	0	0	0	0.954038	0	1	-360	360;
	5743	7750	0	0.07222	0	0	0	0	0.978163	0	1	-360	360;
	5743	2430	0	0.07222	0	0	0	0	0.978163	0	1	-360	360;
	3198	3713	0	0.033909	0	0	0	0	0.967952	0	1	-360	360;
	3198	1138	0	0.04444	0	0	0	0	0.974445	0	1	-360	360;
	3198	11987	0	0.04444	0	0	0	0	0.974445	0	1	-360	360;
	3198	4517	0	0.04444	0	0	0	0	0.974445	0	1	-360	360;
	3198	9889	0	0.04444	0	0	0	0	0.974559	0	1	-360	360;
	8751	9253	0	0.075925	0	0	0	0	0.93144	0	1	-360	360;
	8751	9253	0	0.075925	0	0	0	0	0.93144	0	1	-360	360;
	8751	9253	0	0.08451	0	0	0	0	0.965009	0	1	-360	360;
	10807	1644	0	0.62069	0	0	0	0	1.005972	0	1	-360	360;
	10807	9853	0	0.62069	0	0	0	0	1.005972	0	1	-360	360;
	7569	7677	0	0.40909	0	0	0	0	1.002044	0	1	-360	360;
	11594	916	0	0.07111	0	0	0	0	1.044775	0	1	-360	360;
	12098	1562	0	2.4	0	0	0	0	1.000117	0	1	-360	360;
	12098	13312	0	2.4	0	0	0	0	1.000117	0	1	-360	360;
	12098	10580	0	8	0	0	0	0	0.999937	0	1	-360	360;
	12098	1488	0	8	0	0	0	0	0.999937	0	1	-360	360;
	12098	13193	0	8	0	0	0	0	0.999937	0	1	-360	360;
	12098	2139	0	0.16	0	0	0	0	1.002204	0	1	-360	360;
	13473	8052	0	0.082076	0	0	0	0	0.935732	0	1	-360	360;
	13473	8052	0	0.046721	0	0	0	0	0.951774	0	1	-360	360;
	13473	8052	0	0.046721	0	0	0	0	0.951774	0	1	-360	360;
	10828	13514	0	0.30121	0	0	0	0	1.000845	0	1	-360	360;
	10828	3236	0	0.30121	0	0	0	0	1.000845	0	1	-360	360;
	10828	8555	0	0.30121	0	0	0	0	1.000845	0	1	-360	360;
	10828	7568	0	0.30121	0	0	0	0	1.000845	0	1	-360	360;
	7800	8138	0	0.046417	0	0	0	0	0.948978	0	1	-360	360;
	7800	8138	0	0.046417	0	0	0	0	0.948978	0	1	-360	360;
	9507	4668	0	0.28571	0	0	0	0	0.999929	0	1	-360	360;
	9507	3499	0	0.28571	0	0	0	0	0.999929	0	1	-360	360;
	9507	11920	0	0.28571	0	0	0	0	0.999929	0	1	-360	360;
	9507	12946	0	0.28571	0	0	0	0	0.999929	0	1	-360	360;
	9507	7585	0	0.28571	0	0	0	0	0.999929	0	1	-360	360;
	9507	6563	0	0.28571	0	0	0	0	0.999929	0	1	-360	360;
	7448	6558	0	0.077314	0	0	0	0	0.951427	0	1	-360	360;
	7448	6558	0	0.04639	0	0	0	0	0.952285	0	1	-360	360;
	7448	13590	0	0.09286	0	0	0	0	0.996551	0	1	-360	360;
	7448	1845	0	0.09286	0	0	0	0	0.996551	0	1	-360	360;
	7448	7148	0	0.09286	0	0	0	0	0.996551	0	1	-360	360;
	7448	11170	0	0.09286	0	0	0	0	0.996551	0	1	-360	360;
	7448	8394	0	0.09286	0	0	0	0	0.996551	0	1	-360	360;
	7448	1010	0	0.09286	0	0	0	0	0.996551	0	1	-360	360;
	1608	2180	0	0.06219	0	0	0	0	0.972683	0	1	-360	360;
	1608	5158	0	0.06219	0	0	0	0	0.972715	0	1	-360	360;
	1608	9457	0	0.06219	0	0	0	0	0.972774	0	1	-360	360;
	1608	6871	0	0.06219	0	0	0	0	0.972774	0	1	-360	360;
	1608	5281	0	0.06219	0	0	0	0	0.97225	0	1	-360	360;
	13612	13493	0	0.03226	0	0	0	0	0.99893	0	1	-360	360;
	13612	3797	0	0.03226	0	0	0	0	0.99893	0	1	-360	360;
	4624	2227	0	0.046721	0	0	0	0	0.959854	0	1	-360	360;
	466	11158	0	2.49125	0	0	0	0	1.043188	0	1	-360	360;
	466	12285	0	2.49125	0	0	0	0	1.043188	0	1	-360	360;
	466	3842	0	2.49125	0	0	0	0	1.043188	0	1	-360	360;
	11262	8656	0	0.6	0	0	0	0	1.005315	0	1	-360	360;
	11262	4259	0	0.8	0	0	0	0	1.005104	0	1	-360	360;
	11262	695	0	0.8	0	0	0	0	1.005104	0	1	-360	360;
	11262	10678	0	0.54545	0	0	0	0	1.004859	0	1	-360	360;
	7090	8299	0	0.6	0	0	0	0	1.006265	0	1	-360	360;
	3669	1049	0	0.071508	0	0	0	0	0.951095	0	1	-360	360;
	3669	1049	0	0.071508	0	0	0	0	0.951095	0	1	-360	360;
	3669	1049	0	0.042907	0	0	0	0	0.954937	0	1	-360	360;
	3669	1049	0	0.042907	0	0	0	0	0.954937	0	1	-360	360;
	8874	8346	0	0.079232	0	0	0	0	0.914664	0	1	-360	360;
	3203	9902	0	0.12702	0	0	0	0	1.025478	0	1	-360	360;
	3203	11953	0	0.12702	0	0	0	0	1.025478	0	1	-360	360;
	3203	8091	0	0.12702	0	0	0	0	1.025478	0	1	-360	360;
	3203	3872	0	1.37	0	0	0	0	1.028419	0	1	-360	360;
	741	3082	0	0.15385	0	0	0	0	1.038547	0	1	-360	360;
	741	12286	0	0.15385	0	0	0	0	1.038547	0	1	-360	360;
	3443	10835	0	0.18095	0	0	0	0	1.000216	0	1	-360	360;
	3443	2506	0	0.18095	0	0	0	0	1.000216	0	1	-360	360;
	8738	7243	0	1.27867	0	0	0	0	1.002702	0	1	-360	360;
	8738	2357	0	1.27867	0	0	0	0	1.002702	0	1	-360	360;
	8738	10650	0	1.27867	0	0	0	0	1.002702	0	1	-360	360;
	2175	11789	0	1.176401	0	0	0	0	1.005041	0	1	-360	360;
	2175	4237	0	1.05	0	0	0	0	1.005856	0	1	-360	360;
	2175	7918	0	1.05	0	0	0	0	1.005856	0	1	-360	360;
	2175	7066	0	1.05	0	0	0	0	1.005856	0	1	-360	360;
	2175	8893	0	1.05	0	0	0	0	1.005856	0	1	-360	360;
	2175	13556	0	1.05	0	0	0	0	1.005856	0	1	-360	360;
	2175	12616	0	1.176401	0	0	0	0	1.005041	0	1	-360	360;
	2175	7877	0	1.176401	0	0	0	0	1.005041	0	1	-360	360;
	2175	5408	0	1.176401	0	0	0	0	1.005041	0	1	-360	360;
	2175	132	0	1.176401	0	0	0	0	1.005041	0	1	-360	360;
	2175	7267	0	1.176401	0	0	0	0	1.005041	0	1	-360	360;
	2175	5111	0	1.176401	0	0	0	0	1.005041	0	1	-360	360;
	8892	2175	0	0.045483	0	0	0	0	0.986891	0	1	-360	360;
	11184	11684	0	0.045483	0	0	0	0	0.947104	0	1	-360	360;
	11184	11684	0	0.045483	0	0	0	0	0.947104	0	1	-360	360;
	8119	7642	0	0.045483	0	0	0	0	0.966173	0	1	-360	360;
	402	5679	0	1.18254	0	0	0	0	1.004821	0	1	-360	360;
	402	325	0	1.18571	0	0	0	0	1.004818	0	1	-360	360;
	402	12060	0	1.19048	0	0	0	0	1.004812	0	1	-360	360;
	12408	4748	0	0.047975	0	0	0	0	0.94083	0	1	-360	360;
	12408	4748	0	0.047975	0	0	0	0	0.94083	0	1	-360	360;
	5949	11282	0	0.044262	0	0	0	0	0.968661	0	1	-360	360;
	10314	12421	0	1.125913	0	0	0	0	1.000754	0	1	-360	360;
	10314	13270	0	1.125913	0	0	0	0	1.000754	0	1	-360	360;
	1056	11900	0	0.4	0	0	0	0	1.005458	0	1	-360	360;
	1056	10994	0	0.4	0	0	0	0	1.005458	0	1	-360	360;
	1056	9110	0	0.4	0	0	0	0	1.005458	0	1	-360	360;
	11432	12896	0	0.42067	0	0	0	0	1.00888	0	1	-360	360;
	11432	13643	0	0.42067	0	0	0	0	1.00888	0	1	-360	360;
	11432	3929	0	0.42067	0	0	0	0	1.00888	0	1	-360	360;
	11432	6359	0	0.42067	0	0	0	0	1.00888	0	1	-360	360;
	11432	10971	0	0.631	0	0	0	0	1.008513	0	1	-360	360;
	11432	2816	0	0.631	0	0	0	0	1.008513	0	1	-360	360;
	11432	8163	0	0.631	0	0	0	0	1.008513	0	1	-360	360;
	1178	5061	0	0.85714	0	0	0	0	1.00156	0	1	-360	360;
	1178	6939	0	0.85714	0	0	0	0	1.00156	0	1	-360	360;
	1178	1640	0	0.85714	0	0	0	0	1.00156	0	1	-360	360;
	1178	13212	0	1.5	0	0	0	0	1.001548	0	1	-360	360;
	1178	12277	0	1.5	0	0	0	0	1.001548	0	1	-360	360;
	1178	10285	0	1.5	0	0	0	0	1.001548	0	1	-360	360;
	1178	12359	0	1.5	0	0	0	0	1.001548	0	1	-360	360;
	12974	4392	0	0.20773	0	0	0	0	1.001877	0	1	-360	360;
	12974	10938	0	0.2875	0	0	0	0	1.00111	0	1	-360	360;
	12974	13037	0	0.329628	0	0	0	0	0.960985	0	1	-360	360;
	12974	8393	0	0.346109	0	0	0	0	0.961259	0	1	-360	360;
	10003	7742	0	0.36	0	0	0	0	1.001685	0	1	-360	360;
	4739	927	0	0.2	0	0	0	0	1.00555	0	1	-360	360;
	4739	8436	0	0.2	0	0	0	0	1.00555	0	1	-360	360;
	4739	10302	0	0.22222	0	0	0	0	1.005623	0	1	-360	360;
	519	6686	0	0.23538	0	0	0	0	1.003504	0	1	-360	360;
	519	11581	0	0.23538	0	0	0	0	1.003504	0	1	-360	360;
	1064	4526	0	0.0971	0	0	0	0	1.01076	0	1	-360	360;
	1064	7149	0	0.0971	0	0	0	0	1.01076	0	1	-360	360;
	2885	3392	0	0.047975	0	0	0	0	0.950379	0	1	-360	360;
	2885	3392	0	0.047975	0	0	0	0	0.950379	0	1	-360	360;
	2885	3392	0	0.047975	0	0	0	0	0.950379	0	1	-360	360;
	5794	13634	0	0.046721	0	0	0	0	0.941514	0	1	-360	360;
	5794	13634	0	0.046721	0	0	0	0	0.941514	0	1	-360	360;
	5794	10588	0	0.13846	0	0	0	0	0.98726	0	1	-360	360;
	5794	4786	0	0.13846	0	0	0	0	0.98726	0	1	-360	360;
	5794	5681	0	0.13846	0	0	0	0	0.98726	0	1	-360	360;
	5794	13338	0	0.13846	0	0	0	0	0.98726	0	1	-360	360;
	4804	5313	0	0.047975	0	0	0	0	0.937784	0	1	-360	360;
	4804	5313	0	0.047975	0	0	0	0	0.937784	0	1	-360	360;
	4804	11047	0	0.07143	0	0	0	0	0.987013	0	1	-360	360;
	4804	6584	0	0.07143	0	0	0	0	0.987013	0	1	-360	360;
	2752	7439	0	0.0505	0	0	0	0	0.915592	0	1	-360	360;
	2752	7439	0	0.0505	0	0	0	0	0.915592	0	1	-360	360;
	2752	7439	0	0.0505	0	0	0	0	0.915592	0	1	-360	360;
	4417	12572	0	0.06	0	0	0	0	0.987895	0	1	-360	360;
	4417	7170	0	0.06	0	0	0	0	0.987895	0	1	-360	360;
	3088	7681	0	0.06452	0	0	0	0	0.986513	0	1	-360	360;
	3088	9553	0	0.06452	0	0	0	0	0.986513	0	1	-360	360;
	3088	12625	0	0.06452	0	0	0	0	0.986513	0	1	-360	360;
	11928	12438	0	0.070313	0	0	0	0	0.971656	0	1	-360	360;
	13288	12730	0	0.13333	0	0	0	0	1.006815	0	1	-360	360;
	10089	5464	0	0.08	0	0	0	0	1.001392	0	1	-360	360;
	9548	283	0	1.2	0	0	0	0	1.002586	0	1	-360	360;
	9548	8476	0	1.2	0	0	0	0	1.002586	0	1	-360	360;
	10582	6072	0	0.70588	0	0	0	0	1.003916	0	1	-360	360;
	10582	11584	0	0.70588	0	0	0	0	1.003916	0	1	-360	360;
	10582	13082	0	1.09091	0	0	0	0	1.003824	0	1	-360	360;
	10582	1238	0	1.09091	0	0	0	0	1.003824	0	1	-360	360;
	10582	9693	0	0.8	0	0	0	0	1.003682	0	1	-360	360;
	10582	12555	0	0.8	0	0	0	0	1.003682	0	1	-360	360;
	8876	11549	0	0.08	0	0	0	0	1.002784	0	1	-360	360;
	8165	11641	0	0.8	0	0	0	0	1.004718	0	1	-360	360;
	8165	13226	0	0.8	0	0	0	0	1.004718	0	1	-360	360;
	8165	5289	0	0.8	0	0	0	0	1.004718	0	1	-360	360;
	8165	6994	0	1.2	0	0	0	0	1.00468	0	1	-360	360;
	8165	8595	0	1.2	0	0	0	0	1.00468	0	1	-360	360;
	8165	621	0	1.2	0	0	0	0	1.00468	0	1	-360	360;
	10549	2994	0	0.18462	0	0	0	0	1.00459	0	1	-360	360;
	10549	4190	0	0.18462	0	0	0	0	1.00459	0	1	-360	360;
	10549	9461	0	0.18462	0	0	0	0	1.00459	0	1	-360	360;
	10441	42	0	0.4	0	0	0	0	1.001976	0	1	-360	360;
	12564	10365	0	0.375	0	0	0	0	1.000295	0	1	-360	360;
	12564	1961	0	0.375	0	0	0	0	1.000295	0	1	-360	360;
	12564	10598	0	0.66667	0	0	0	0	1.000528	0	1	-360	360;
	11754	8202	0	0.16	0	0	0	0	1.000721	0	1	-360	360;
	6753	13407	0	0.26667	0	0	0	0	1.004335	0	1	-360	360;
	3253	9906	0	0.06857	0	0	0	0	0.998577	0	1	-360	360;
	4050	10652	0	0.24	0	0	0	0	1.000492	0	1	-360	360;
	4050	11273	0	0.24	0	0	0	0	1.000492	0	1	-360	360;
	4050	9016	0	0.24	0	0	0	0	1.000492	0	1	-360	360;
	4050	8776	0	0.24	0	0	0	0	1.000492	0	1	-360	360;
	7185	8709	0	0.26667	0	0	0	0	1.001292	0	1	-360	360;
];

%%-----  OPF Data  -----%%
%% generator cost data
%	1	startup	shutdown	n	x1	y1	...	xn	yn
%	2	startup	shutdown	n	c(n-1)	...	c0
mpc.gencost = [
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
];
