function mpc = case2869pegase
%CASE2869PEGASE  Power flow data for large part of European system.
%   Please see CASEFORMAT for details on the case file format.
%
%   This case accurately represents the size and complexity of part of the
%   European high voltage transmission network. The network contains 2,869
%   buses, 510 generators, and 4,582 branches and it operates at 380, 220,
%   150, and 110 kV. Please note that the data are fictitious and do not
%   correspond to real world data. They can thus be used to validate
%   methods and tools but should not be used for operation and planning of
%   the European grid.
%
%   The data stems from the Pan European Grid Advanced Simulation and State
%   Estimation (PEGASE) project, part of the 7th Framework Program of the
%   European Union (https://www.fp7-pegase.com/).
%
%   When publishing results based on this data, please cite:
%
%     C. Josz, S. Fliscounakis, J. Maeght, and P. Panciatici, "AC Power Flow
%     Data in MATPOWER and QCQP Format: iTesla, RTE Snapshots, and PEGASE"
%     https://arxiv.org/abs/1603.01533
%
%     S. Fliscounakis, P. Panciatici, F. Capitanescu, and L. Wehenkel,
%     "Contingency ranking with respect to overloads in very large power
%     systems taking into account uncertainty, preventive and corrective
%     actions", Power Systems, IEEE Trans. on, (28)4:4909-4917, 2013.
%     https://doi.org/10.1109/TPWRS.2013.2251015
%
%   Remarks:
%
%   1. Line flow limits are the current flow limits found in PEGASE data.
%
%   2. PEGASE data contains asymmetric shunt conductance and susceptance in
%   the PI transmission line model of branches. Thus total line charging
%   susceptance of branches is set to 0 p.u. and the nodal representation
%   of shunt condutance and susceptance is used. As a result, power flow
%   equations are left unchanged compared with original PEGASE data.
%   However, line flow constraints in the optimal flow problem are
%   modified.
%
%   3. Identical linear costs are used for all generators to form a loss
%   minimizing OPF objective function.
%
%   4. Since some parts of the network are aggregated, some generators
%   (e.g. with negative PMIN) represent aggregations of multiple loads
%   and generators.
%
%   Contacts:
%     Cédric Josz, Stéphane Fliscounakis, Jean Maeght, Patrick Panciatici
%     firstname.lastname@rte-france.com
%     Réseau de Transport d'Electricité (French Transmission System Operator)
%     Département Expertise Système, Immeuble "Le Colbert"
%     9 rue de la Porte de Buc, 78000 Versailles Cedex, France
%
%   March 18th, 2015

%   MATPOWER
%   Copyright (c) 2015, 2016 by Cédric Josz, Stéphane Fliscounakis, Jean Maeght,
%   and Patrick Panciatici
%   Licensed under the Creative Commons Attribution 4.0 International license,
%   https://creativecommons.org/licenses/by/4.0/

%% MATPOWER Case Format : Version 2
mpc.version = '2';

%%-----  Power Flow Data  -----%%
%% system MVA base
mpc.baseMVA = 100;

%% bus data
%	bus_i	type	Pd	Qd	Gs	Bs	area	Vm	Va	baseKV	zone	Vmax	Vmin
mpc.bus = [
	3	1	151	48.8	0	4.69	0	1.015975	-21.797826	220	5	1.1	0.9;
	4	1	171.41	23.4	0	2.1	0	1.025797	-7.009176	220	5	1.1	0.9;
	10	1	134	24.7	0	12.44	0	1.037973	-24.31171	220	5	1.1	0.9;
	15	1	59.67	26.7	0	12.84	0	1.024279	-48.286579	220	4	1.1	0.9;
	21	1	161.2	39.3	0	30.71	0	1.035642	-21.986561	220	5	1.1	0.9;
	22	1	0	-0	0	0.85	0	1.032862	-18.51179	220	5	1.1	0.9;
	26	1	167.3	54.8	0	110.54	0	1.032832	-3.303479	380	5	1.1	0.9;
	29	1	152.9	49.4	0	27.65	0	0.997905	-13.742779	380	5	1.1	0.9;
	32	2	0	0	0	0	0	1.006206	-42.36182	220	4	1.1	0.9;
	33	2	0	0	0	0	0	1.031909	-37.946488	220	4	1.1	0.9;
	38	1	42.75	12.8	0	13.3	0	1.028477	26.320254	220	8	1.1	0.9;
	39	2	0	0	0	0	0	1.005551	-37.656299	150	10	1.1	0.9;
	42	1	20.82	7.9	0	26.89	0	1.047415	-48.480512	380	4	1.1	0.9;
	44	1	-0	0	0	7.27	0	1.045273	-22.20917	220	5	1.1	0.9;
	51	2	0	0	0	0	0	1.01198	-47.438988	220	4	1.1	0.9;
	52	1	127.6	31.8	0	20.51	0	1.013245	0.963784	150	8	1.1	0.9;
	53	1	48.65	16.1	0	7.14	0	1.060523	-9.105713	220	5	1.1	0.9;
	58	1	0	0	0	2.31	0	0.995373	-21.587045	380	5	1.1	0.9;
	59	1	119.2	20.7	0	3.31	0	1.031574	-23.793559	220	5	1.1	0.9;
	69	1	62.87	20.5	0	1.15	0	1.011798	-0.30934	150	8	1.1	0.9;
	71	1	126.3	42.4	0	55.24	0	1.037783	3.197998	150	8	1.1	0.9;
	72	1	6.91	-4.6	0	5.75	0	1.009626	-40.551667	220	4	1.1	0.9;
	75	1	38.74	6.1	0	1.17	0	1.024422	-52.498016	220	4	1.1	0.9;
	76	1	22.13	10.7	0	14.21	0	1.028539	26.329875	220	8	1.1	0.9;
	77	1	-0	-0	0	0.73	0	1.013066	-48.881785	220	4	1.1	0.9;
	89	1	0	0	0.199111	4.530599	0	0.996001	-4.601142	380	2	1.1	0.9;
	90	1	363.25	115.39	0	3.95	0	1.04293	-27.629609	220	5	1.1	0.9;
	91	1	-0	0	0	0.02	0	1.011563	0.577525	150	8	1.1	0.9;
	96	1	-8.4	28.66	0	3.49	0	1.034803	-15.625024	220	5	1.1	0.9;
	113	1	-0	-0	0	6.532711	0	1.030544	-14.260449	380	5	1.1	0.9;
	115	1	0	-0	0	4.21	0	1.065676	-19.810351	220	5	1.1	0.9;
	118	1	704.53	-35.33	0	13.102834	0	1.021915	-11.391408	380	5	1.1	0.9;
	124	2	0	0	0	0	0	1.079291	8.367367	220	5	1.1	0.9;
	128	1	23.73	12.7	0	0.52	0	1.020115	-16.088574	220	5	1.1	0.9;
	130	1	-0	0	0.135375	47.750675	0	1.062972	-46.623005	380	10	1.1	0.9;
	132	1	0	-0	0	21.39	0	1.067846	-43.278308	380	1	1.1	0.9;
	133	1	38.14	12.5	0	6.28	0	1.01895	2.823528	150	8	1.1	0.9;
	138	1	-75.38	7.8	0	29.9	0	1.071168	-41.33849	380	10	1.1	0.9;
	139	1	-764.34	66.91	0	33.47	0	1.002952	17.568852	380	8	1.1	0.9;
	140	1	0	0	0	4.14	0	1.029191	17.813411	220	8	1.1	0.9;
	143	1	0	0	0	0	0	1.013822	4.4155	150	8	1.1	0.9;
	145	1	38.34	5.6	0	7.64	0	1.042169	-8.880162	220	5	1.1	0.9;
	148	1	32.34	12	0	6.1	0	1.041686	-7.998665	220	5	1.1	0.9;
	150	2	0	0	0	0	0	1.062067	2.553662	220	5	1.1	0.9;
	163	1	147.76	32.73	0	8.34	0	1.042505	-10.668912	220	5	1.1	0.9;
	164	1	0	-0	0	5.17	0	1.015861	-0.666839	150	8	1.1	0.9;
	171	1	100.8	13.2	0	5.24	0	1.063782	-25.157599	220	5	1.1	0.9;
	172	1	111.7	10.96	0	4.22	0	1.017044	-46.443671	220	4	1.1	0.9;
	174	1	370.5	-17	0	9.91	0	1.073491	-27.987179	220	5	1.1	0.9;
	177	1	94.9	36.2	0	16.46	0	0.973375	-42.693578	220	10	1.1	0.9;
	179	2	0	0	0	0	0	1.048817	-49.34001	380	4	1.1	0.9;
	184	1	-0	-0	0	2.42	0	1.043679	-11.957994	220	5	1.1	0.9;
	188	1	36.84	13.9	0	3.3	0	1.018271	-19.570046	220	5	1.1	0.9;
	194	2	0	0	0	0	0	1.029584	13.78786	150	8	1.1	0.9;
	195	1	73.14	21.55	0	8.52	0	1.053191	-8.795989	220	5	1.1	0.9;
	196	1	0	-0	0	12.48	0	1.074427	-26.348559	220	5	1.1	0.9;
	198	1	39.74	11.8	0	1.44	0	1.01582	-16.334436	220	5	1.1	0.9;
	199	1	86.8	28.4	0	1.34	0	1.011196	0.998875	150	8	1.1	0.9;
	201	2	0	0	0	0	0	1.024675	-50.722374	220	4	1.1	0.9;
	205	1	46.15	18.2	0	1.29	0	1.00871	4.996181	150	8	1.1	0.9;
	207	1	321.98	84.4	0	1.34	0	1.027593	-13.46248	220	5	1.1	0.9;
	211	1	75.28	19.2	0	1.9	0	1.032527	13.596519	110	8	1.1	0.9;
	216	1	155.6	54.7	0	2.21	0	1.038563	-24.634597	220	5	1.1	0.9;
	217	1	105.2	0.1	0	7.36	0	1.034707	-2.558504	220	5	1.1	0.9;
	218	1	69.83	15.12	0	2.23	0	1.042445	-13.132419	220	5	1.1	0.9;
	221	2	0	0	0	0	0	0.996091	-22.219166	380	5	1.1	0.9;
	225	1	0	0	0	1.15	0	1.062639	-20.201534	220	5	1.1	0.9;
	228	1	-23.43	57.4	0	0	0	1.043885	-8.143399	150	2	1.1	0.9;
	230	1	152.07	22.52	0	7.87	0	1.007001	-46.343334	220	4	1.1	0.9;
	231	1	-0.13	-1	0	0	0	1.019763	23.115587	110	8	1.1	0.9;
	236	1	45.85	7	0	1.05	0	1.02885	-53.54777	220	4	1.1	0.9;
	243	1	0	-0.01	0	13.49	0	1.01715	17.760873	150	8	1.1	0.9;
	244	1	-0	-0	0	0.68	0	1.018463	6.917494	150	8	1.1	0.9;
	249	1	-0	-0	0	26.3	0	1.005395	-42.359771	220	4	1.1	0.9;
	257	1	113.9	24.5	0	6.72	0	1.034093	-45.514467	220	4	1.1	0.9;
	261	1	-0	0	0	1.29	0	1.056876	-45.124465	380	10	1.1	0.9;
	264	1	38.74	9	0	0.57	0	1.01505	-46.406008	220	4	1.1	0.9;
	266	1	0	-0	0	140.96	0	1.059749	-41.880587	380	4	1.1	0.9;
	267	1	103.8	2.7	0	3.33	0	1.016267	-50.549724	220	4	1.1	0.9;
	268	1	-0.13	-1	0	0	0	1.046663	26.481396	220	8	1.1	0.9;
	271	1	96.7	26.8	0	0	0	1.03996	-9.671728	150	2	1.1	0.9;
	275	1	102.1	25.3	0	1.11	0	1.012638	2.38281	150	8	1.1	0.9;
	280	1	119.6	29.6	0	5.34	0	1.033127	-11.186618	220	5	1.1	0.9;
	281	1	-0	-0	0	18.07	0	1.035067	14.007344	380	8	1.1	0.9;
	283	1	27.33	11.9	0	1.17	0	1.036291	-12.101832	220	5	1.1	0.9;
	284	1	191.2	-3.4	0	19.21	0	1.023939	-48.338029	220	4	1.1	0.9;
	292	1	13.82	1.4	0	0.7	0	1.04757	-14.756324	220	5	1.1	0.9;
	297	1	23.73	9.4	0	1.45	0	1.025769	4.498938	150	8	1.1	0.9;
	299	2	0	0	0	0	0	1.046197	-45.832236	380	4	1.1	0.9;
	304	1	2.2	0.5	0	0.21	0	0.994426	-53.724872	150	10	1.1	0.9;
	305	1	103	12	0	0.63	0	1.066061	-32.494764	220	5	1.1	0.9;
	306	1	-0	0	0	4.76	0	1.029806	-18.845783	220	5	1.1	0.9;
	311	1	42.05	17.2	0	1.31	0	1.003155	-2.003262	150	8	1.1	0.9;
	312	1	-0.16	-2.91	0	0.43	0	1.054916	11.672353	110	8	1.1	0.9;
	314	1	163.1	46.3	0	10.6	0	1.000695	-16.068608	220	5	1.1	0.9;
	315	1	-0	-0	0	2.13	0	1.026865	-35.887691	220	10	1.1	0.9;
	317	1	-0	-0	0.045946	-1.82203	0	1.017677	6.394344	380	2	1.1	0.9;
	322	1	303.8	85.4	0	3.63	0	0.963932	-44.221789	220	10	1.1	0.9;
	326	1	0	0	0	0	0	1.046796	26.528223	220	8	1.1	0.9;
	333	1	524.5	43	0	6.85	0	1.012818	-42.225824	220	5	1.1	0.9;
	338	2	0	0	0	0	0	1.049578	-6.144398	220	5	1.1	0.9;
	344	1	372.8	147.5	0	27.73	0	1.022517	-1.388364	150	8	1.1	0.9;
	346	1	197.5	48.3	0	0.71	0	1.039389	-21.333666	220	5	1.1	0.9;
	349	1	-0	0	0	15.15	0	1.047326	9.079132	150	8	1.1	0.9;
	350	1	110.4	26.2	0	5.62	0	1.042975	-24.324408	220	5	1.1	0.9;
	351	1	-0.15	-1.93	0	0	0	1.028054	18.540762	220	8	1.1	0.9;
	352	2	0	0	0	0	0	1.054691	0.247829	380	5	1.1	0.9;
	360	1	35.54	11.5	0	3.65	0	1.047637	11.384137	110	8	1.1	0.9;
	367	1	65.17	-51.5	0	4.85	0	1.071258	-23.626071	220	5	1.1	0.9;
	369	1	-0	0	0	3.78	0	1.045084	13.593865	220	8	1.1	0.9;
	370	1	56.76	23.2	0	0.49	0	1.00045	-2.194776	150	8	1.1	0.9;
	377	1	-0.18	-1	0	0	0	1.021633	11.131046	150	8	1.1	0.9;
	380	1	-95.71	-2.8	0	7.25	0	1.004315	-32.70603	220	10	1.1	0.9;
	385	1	-0	-0	0	105.84	0	1.045045	-49.151666	380	4	1.1	0.9;
	387	1	-0.19	-4.38	0	0	0	1.040472	6.605475	150	8	1.1	0.9;
	389	2	0	0	0	0	0	1.020856	-40.645901	150	10	1.1	0.9;
	392	1	134.1	34.2	0	0.75	0	1.005558	-0.515044	150	8	1.1	0.9;
	395	1	-0	-0	0	7.85	0	1.034154	15.282226	380	8	1.1	0.9;
	401	1	56.5	36.29	0	3.08	0	1.028209	-14.323194	220	5	1.1	0.9;
	408	1	129.1	7.4	0	0.04	0	1.041485	-11.562819	220	5	1.1	0.9;
	410	1	170.89	50.38	0	6.44	0	1.035431	-26.631809	220	5	1.1	0.9;
	413	2	0	0	0	0	0	1.052152	-6.629634	220	5	1.1	0.9;
	415	1	4.71	0.7	0	11.39	0	1.029407	-51.548505	220	4	1.1	0.9;
	416	1	70.68	13.5	0	6.96	0	1.041782	-6.713309	220	5	1.1	0.9;
	420	1	0	0	0	0	0	1.049813	18.568379	380	8	1.1	0.9;
	425	1	-0	0	0	2.72	0	1.008682	6.877656	380	1	1.1	0.9;
	426	1	123.1	6.5	0	3.25	0	1.040559	-18.404022	220	5	1.1	0.9;
	427	1	-0	0	0	1.24	0	1.077855	-35.447133	380	1	1.1	0.9;
	429	1	26.93	-22.4	0	4.95	0	1.018028	-45.166294	220	4	1.1	0.9;
	432	1	-0	-0	0	68.517663	0	1.04617	2.418195	380	5	1.1	0.9;
	440	2	0	0	0	0	0	1.029422	-36.073637	220	4	1.1	0.9;
	441	1	-0	0	0.199768	27.888516	0	1.037169	-51.706225	380	10	1.1	0.9;
	442	2	0	0	0	0	0	1.013689	-46.010399	220	4	1.1	0.9;
	444	1	0	-0	0	2.45	0	1.037192	-12.208318	220	5	1.1	0.9;
	445	1	36.74	10.3	0	8.1	0	1.022265	-44.379049	220	4	1.1	0.9;
	449	1	0	0	0	29.42	0	1.02121	-8.505204	380	5	1.1	0.9;
	453	2	0	0	0	0	0	1.021096	-7.822572	380	5	1.1	0.9;
	455	1	274.32	5.66	0	13.24	0	1.043545	-30.648087	220	5	1.1	0.9;
	472	1	336.2	131.6	0	64.01	0	1.049863	7.888585	150	8	1.1	0.9;
	473	1	-22.95	-3.72	0	14.34	0	1.027789	-45.658214	220	4	1.1	0.9;
	477	1	0	0	0	70.37	0	1.049487	-41.702128	380	4	1.1	0.9;
	481	1	-0	0	0	17.44	0	1.024242	-47.836389	220	4	1.1	0.9;
	483	1	57.26	15.6	0	3.43	0	1.022965	-49.351603	220	4	1.1	0.9;
	486	1	7.01	2	0	0.44	0	1.014633	4.966833	150	8	1.1	0.9;
	489	2	0	0	0	0	0	1.012805	-47.213053	220	4	1.1	0.9;
	490	1	20.52	4	0	2.77	0	1.036343	-23.043793	220	5	1.1	0.9;
	496	1	-0	-0	0	10.04	0	1.018474	11.259593	150	8	1.1	0.9;
	500	1	144.4	35.5	0	7.66	0	1.037969	-15.06839	220	5	1.1	0.9;
	502	1	-0	0	0	1.96	0	1.04397	-6.689582	220	5	1.1	0.9;
	506	1	-0	0	0	0.4	0	1.062897	-9.075959	380	5	1.1	0.9;
	509	1	77.29	7	0	17.6	0	1.015368	-60.258851	220	4	1.1	0.9;
	513	1	57.46	17.1	0	1.61	0	1.038862	-7.455688	220	5	1.1	0.9;
	516	2	0	0	0	0	0	1.074804	-33.515516	220	5	1.1	0.9;
	519	1	207.2	9.4	0	8.77	0	1.026737	-49.849135	220	4	1.1	0.9;
	520	1	118.5	20.4	0	2.57	0	1.048877	-27.213193	220	5	1.1	0.9;
	523	1	189.9	74.7	0	26.2	0	1.043256	7.231269	150	8	1.1	0.9;
	527	1	277.7	45.4	0	0.7	0	1.014388	-46.235817	220	4	1.1	0.9;
	536	1	87.8	-1.4	0.079073	36.047349	0	1.059638	-47.971384	380	10	1.1	0.9;
	539	1	57.96	18.9	0	14.91	0	1.013211	3.762537	150	8	1.1	0.9;
	544	1	79.69	39.6	0	18.83	0	1.039051	-51.441574	380	10	1.1	0.9;
	548	1	30.93	8.9	0	0.45	0	1.03326	16.142358	110	8	1.1	0.9;
	549	1	0	-0	0	5.15	0	1.070777	-10.99741	220	5	1.1	0.9;
	550	1	0	-0	0	10	0	1.040041	-45.192187	220	4	1.1	0.9;
	552	1	180.8	90.8	0	2.42	0	0.969255	-42.920126	150	10	1.1	0.9;
	554	1	0	-0	0	1.97	0	1.021838	-20.831434	220	5	1.1	0.9;
	556	1	16.52	9	0	1.17	0	1.036788	-22.791951	220	5	1.1	0.9;
	558	1	-0.18	-3.57	0	0	0	1.017158	17.760895	150	8	1.1	0.9;
	559	1	42.25	16.6	0	0	0	1.03524	5.466244	150	8	1.1	0.9;
	561	1	0	-0	0	0.32	0	1.051908	-11.066228	220	5	1.1	0.9;
	564	2	0	0	0	0	0	1.069692	-47.162859	220	5	1.1	0.9;
	575	1	0	-0	0	0.67	0	1.045123	-11.170776	220	5	1.1	0.9;
	577	1	-0.13	-1	0	0	0	1.045911	26.397998	220	8	1.1	0.9;
	579	1	17.32	6.6	0	16.52	0	1.021347	-51.682309	380	4	1.1	0.9;
	583	2	0	0	0	0	0	1.043394	-15.595495	220	5	1.1	0.9;
	585	1	21.12	8.3	0	4.26	0	1.032522	3.857451	150	8	1.1	0.9;
	593	1	-0	0	0	29.2	0	1.054621	12.781834	380	8	1.1	0.9;
	594	1	67.57	17	0	5.75	0	1.038322	-8.883436	220	5	1.1	0.9;
	595	1	47.55	5.3	0	7.59	0	1.028698	-49.910829	220	4	1.1	0.9;
	597	1	33.34	15.7	0	64.15	0	1.059114	-47.539487	380	4	1.1	0.9;
	601	1	140.6	42.6	0	5.71	0	1.041109	-14.413923	220	5	1.1	0.9;
	602	1	110.8	-8.8	0	3.15	0	1.023809	-46.846316	220	4	1.1	0.9;
	604	1	33.24	4.4	0	24.54	0	0.996802	-12.954482	380	5	1.1	0.9;
	608	1	150.9	10.4	0	4.77	0	1.067834	-32.673267	220	5	1.1	0.9;
	609	1	-0	-0	0	5.67	0	1.022581	-14.805449	220	5	1.1	0.9;
	610	1	-0	0	0	0.91	0	0.990242	-55.420594	150	10	1.1	0.9;
	615	2	0	0	0	0	0	1.06611	-10.707846	220	5	1.1	0.9;
	616	2	0	0	0	0	0	1.049295	-12.996312	220	5	1.1	0.9;
	618	2	0	0	0	0	0	1.028233	-48.626653	380	4	1.1	0.9;
	619	1	72.88	14.4	0	2.96	0	1.045204	-25.296091	220	5	1.1	0.9;
	621	1	21.42	8.1	0	7.38	0	1.019534	-51.372282	220	4	1.1	0.9;
	628	2	0	0	0	0	0	0.98494	-34.539287	220	10	1.1	0.9;
	639	2	0	0	0	0	0	1.046182	-30.213914	220	5	1.1	0.9;
	641	1	45.32	18.5	0	0.74	0	1.0383	-13.886781	220	5	1.1	0.9;
	642	2	0	0	0	0	0	1.036757	-50.092904	380	4	1.1	0.9;
	643	1	15.52	6.3	0	0.6	0	1.007039	-1.733141	150	8	1.1	0.9;
	649	1	40.94	19.2	0	5.2	0	1.025052	-48.149342	220	4	1.1	0.9;
	650	1	17.22	4.4	0	1.71	0	1.020577	12.404841	110	8	1.1	0.9;
	656	1	-0.15	-1.93	0	0	0	1.028027	18.542452	220	8	1.1	0.9;
	658	1	258.4	66.7	0	5.87	0	1.033358	-17.451459	220	5	1.1	0.9;
	659	1	-0	-0	0.174507	73.977052	0	1.017121	6.345431	380	2	1.1	0.9;
	661	1	-0	-0	0	0.6	0	1.032374	2.535923	380	5	1.1	0.9;
	662	1	0	-0	0	4.14	0	1.024189	11.926507	380	8	1.1	0.9;
	666	1	-5.61	-2.79	0	150.93	0	1.026843	-5.376314	380	5	1.1	0.9;
	676	1	109.3	35.4	0	1.91	0	1.014417	5.816488	150	8	1.1	0.9;
	678	1	232.1	89.8	0	0.09	0	0.992478	-18.417537	220	5	1.1	0.9;
	682	2	0	0	0	0	0	1.065269	-10.065918	220	5	1.1	0.9;
	686	2	0	0	0	0	0	1.018064	-46.344294	220	4	1.1	0.9;
	687	1	-0	0	0	5.6	0	1.061145	-4.109826	220	5	1.1	0.9;
	693	1	-0	0	0	93.75	0	1.061013	-39.903628	380	4	1.1	0.9;
	695	1	96.41	26.5	0	0.08	0	1.001534	-37.751201	220	4	1.1	0.9;
	698	2	0	0	0	0	0	1.001541	-29.340894	220	10	1.1	0.9;
	699	1	36.54	8.4	0	6.58	0	1.015897	-52.533022	220	4	1.1	0.9;
	702	1	-0	-0	0	0.48	0	1.007203	-35.192471	150	10	1.1	0.9;
	706	1	114.8	22.8	0	5.66	0	1.01376	-43.810675	220	4	1.1	0.9;
	707	1	83.99	21.2	0	5.36	0	1.052025	-25.653481	220	5	1.1	0.9;
	709	2	0	0	0	0	0	1.072146	-41.838466	380	4	1.1	0.9;
	710	1	76.38	5.1	0	3.97	0	1.050704	-44.810504	220	4	1.1	0.9;
	712	1	110.4	41.9	0	1.88	0	1.009788	1.2558	150	8	1.1	0.9;
	718	1	135.11	41	0	5.31	0	1.032503	-24.998397	220	5	1.1	0.9;
	720	1	-0	-0	0	23.554121	0	0.976999	-26.946044	380	5	1.1	0.9;
	722	1	0	-0	0	0.78	0	1.011148	0.459046	150	8	1.1	0.9;
	726	1	400.3	189.9	0	0.53	0	1.051828	-4.968545	220	5	1.1	0.9;
	731	1	0	0	0	61.48	0	1.046464	-49.570359	380	4	1.1	0.9;
	733	1	58.36	10.7	0	8.41	0	1.004624	6.681343	110	8	1.1	0.9;
	739	2	0	0	0	0	0	1.040583	-42.166433	380	4	1.1	0.9;
	742	1	0	-0	0	2.83	0	1.061315	-4.111335	220	5	1.1	0.9;
	747	1	69.58	16	0	3.03	0	1.073525	-26.409302	220	5	1.1	0.9;
	749	2	0	0	0	0	0	1.0548	-29.025797	220	5	1.1	0.9;
	750	1	-0	-0	0	18.63	0	1.040344	-24.488621	220	5	1.1	0.9;
	751	2	0	0	0	0	0	1.023645	-43.542115	220	4	1.1	0.9;
	754	1	132.1	32.4	0	58.29	0	1.049392	-44.640438	380	4	1.1	0.9;
	757	2	0	0	0	0	0	1.016623	-13.779113	380	5	1.1	0.9;
	766	1	-0	0	0	7.85	0	1.043257	-19.971137	220	5	1.1	0.9;
	769	1	152.57	40	0	6.12	0	1.02737	-10.890652	220	5	1.1	0.9;
	772	1	343.9	141.9	0	5.55	0	1.03156	-14.189114	220	5	1.1	0.9;
	775	1	-55.14	-27.14	0	0	0	1.049935	7.89661	150	8	1.1	0.9;
	776	2	0	0	0	0	0	1.029038	-13.297102	220	5	1.1	0.9;
	778	2	0	0	0	0	0	1.051035	-6.663814	220	5	1.1	0.9;
	780	2	0	0	0	0	0	1.006139	-44.460594	220	10	1.1	0.9;
	784	1	-26	-14.72	0	0	0	1.026671	4.566419	150	8	1.1	0.9;
	785	1	-0	-0	0	4.03	0	1.032974	-33.396061	220	10	1.1	0.9;
	789	1	-0	-0	0	6.63	0	1.047988	-5.931607	220	5	1.1	0.9;
	790	1	-0.15	-1	0	0.42	0	1.056626	12.502239	110	8	1.1	0.9;
	792	1	295.3	41.9	0	0	0	1.004879	-2.745379	150	2	1.1	0.9;
	795	2	0	0	0	0	0	1.039638	-15.418252	220	5	1.1	0.9;
	797	1	2	114.6	0	16.15	0	1.01981	-39.437643	220	4	1.1	0.9;
	800	1	50.56	18	0	0.11	0	1.035866	-12.135713	220	5	1.1	0.9;
	803	2	0	0	0	0	0	1.04396	-24.894279	220	5	1.1	0.9;
	804	1	-0	-0	0	7.26	0	1.0581	-7.541712	220	5	1.1	0.9;
	805	1	124.5	29.6	0	15.4	0	1.053184	-11.133211	220	5	1.1	0.9;
	806	1	-0	-0	0	32.02	0	0.999621	7.284283	380	8	1.1	0.9;
	807	2	0	0	0	0	0	1.057043	-37.929845	380	4	1.1	0.9;
	808	1	24.53	8.6	0	7.01	0	1.022909	-49.791145	220	4	1.1	0.9;
	809	1	0	-0	0	58.088687	0	1.031022	-0.130758	380	5	1.1	0.9;
	814	1	287.7	16.3	0	12.86	0	1.024011	-50.140745	380	4	1.1	0.9;
	819	1	0	-0	0	0.22	0	1.05327	-0.714338	220	5	1.1	0.9;
	823	2	0	0	0	0	0	1.044347	-2.253517	380	5	1.1	0.9;
	838	1	615.81	286.69	0	101.04	0	1.031933	4.198236	150	8	1.1	0.9;
	839	1	-0	-0	0	8.17514	0	1.027702	-1.443829	380	5	1.1	0.9;
	849	2	0	0	0	0	0	1.041392	-41.391834	380	4	1.1	0.9;
	851	2	0	0	0	0	0	1.006263	-17.209235	220	5	1.1	0.9;
	854	1	49.96	7.5	0	8.86	0	1.029065	-53.698786	220	4	1.1	0.9;
	858	2	0	0	0	0	0	1.035299	1.038476	380	5	1.1	0.9;
	869	1	94.2	5.5	0	6.51	0	1.024451	-46.452848	220	4	1.1	0.9;
	870	1	110	3.4	0	11.95	0	1.039963	-24.646112	220	5	1.1	0.9;
	871	1	108.89	-21.76	0	3.72	0	1.053384	-12.006609	220	5	1.1	0.9;
	878	1	0	-0	0	25.28	0	1.032326	10.347716	150	8	1.1	0.9;
	883	1	192.5	78.3	0	6.65	0	1.019167	-10.099494	220	5	1.1	0.9;
	891	2	0	0	0	0	0	1.063198	-18.995307	380	5	1.1	0.9;
	892	1	-0	-0	0	9.1	0	1.06845	-35.058808	220	5	1.1	0.9;
	898	1	25.63	7.9	0	0.77	0	1.017379	13.778356	110	8	1.1	0.9;
	900	1	0	0	0	0.15	0	1.045753	-15.459465	220	5	1.1	0.9;
	903	1	82.69	20.1	0	1.36	0	1.073487	-26.434473	220	5	1.1	0.9;
	905	1	631.71	-54.16	0	7.534129	0	1.034156	-15.096405	380	5	1.1	0.9;
	907	1	-0	-0	0	0.05	0	1.042523	-9.744132	220	5	1.1	0.9;
	908	1	125.9	-0.5	0	0.21	0	1.062521	-7.960383	220	5	1.1	0.9;
	913	2	0	0	0	0	0	1.032774	-1.888859	380	2	1.1	0.9;
	916	2	0	0	0	0	0	1.024009	-37.476261	220	4	1.1	0.9;
	919	1	-0	0	0	13.89	0	1.015897	12.146494	380	8	1.1	0.9;
	920	1	34.44	8.5	0	1.11	0	1.044238	-20.80386	220	5	1.1	0.9;
	923	1	62.07	14.7	0	4.84	0	1.035473	-23.119527	220	5	1.1	0.9;
	928	1	112.9	19.3	0	17.34	0	1.01151	0.819942	150	8	1.1	0.9;
	929	1	43.75	12.4	0	9.35	0	1.024627	-45.597223	220	4	1.1	0.9;
	930	2	0	0	0	0	0	1.063763	-37.235712	380	4	1.1	0.9;
	932	1	77.59	12.3	0	5.1	0	1.031096	-53.231343	220	4	1.1	0.9;
	933	1	-0	0	0	22.75	0	1.019207	-21.558165	220	5	1.1	0.9;
	935	1	-0	0	0	9.46	0	1.047658	-24.511894	220	5	1.1	0.9;
	947	1	90.7	5.8	0	4.1	0	0.989542	-49.087121	220	4	1.1	0.9;
	948	1	15.82	4	0	0.7	0	0.990808	-55.02646	150	10	1.1	0.9;
	950	1	69.08	22.6	0	0.54	0	1.014233	0.231638	150	8	1.1	0.9;
	953	1	108.7	-12.15	0	37.11	0	1.064103	-13.019958	380	5	1.1	0.9;
	954	1	66.27	13.6	0	0.67	0	1.043381	-21.557176	220	5	1.1	0.9;
	955	1	244.2	152.7	0	0	0	1.010468	-11.004951	150	2	1.1	0.9;
	957	1	65.77	16.1	0	1.66	0	1.014721	-45.935916	220	4	1.1	0.9;
	960	1	-0	-0	0	108.728602	0	1.062896	-9.075954	380	5	1.1	0.9;
	962	1	42.55	17.1	0	1.67	0	1.040501	-45.427736	220	4	1.1	0.9;
	964	1	88.5	16.5	0	146.44	0	1.013923	-11.664656	380	5	1.1	0.9;
	970	1	62.27	-2.1	0	8.01	0	1.019331	-42.032228	220	4	1.1	0.9;
	972	2	0	0	0	0	0	1.068998	5.469581	380	5	1.3	0.7;
	976	2	0	0	0	0	0	1.029847	-49.344807	380	4	1.1	0.9;
	980	1	14.71	4.6	0	7.88	0	1.045627	-25.181981	220	5	1.1	0.9;
	981	1	0	-0	0	0.84	0	1.023028	-43.883268	220	4	1.1	0.9;
	986	1	15.98	0.86	0	5.96	0	1.022036	-52.113357	220	4	1.1	0.9;
	987	1	-0	0	0	1.18	0	1.000938	-38.474493	150	10	1.1	0.9;
	988	1	-0	-0	0	8.62	0	1.028293	-52.276231	220	4	1.1	0.9;
	994	1	-0	-0	0	25	0	1.00959	-0.823416	150	8	1.1	0.9;
	999	1	20.02	6.8	0	0.48	0	1.018279	20.643565	110	8	1.1	0.9;
	1000	1	0	-0	0	1.8	0	1.020326	-32.938159	220	1	1.1	0.9;
	1001	2	0	0	0	0	0	1.073365	-33.684292	220	5	1.1	0.9;
	1002	2	0	0	0	0	0	1.040649	-20.121516	220	5	1.1	0.9;
	1005	1	0	0	0	0.81	0	1.055459	-12.444399	220	5	1.1	0.9;
	1010	1	21.32	7.3	0	0.6	0	1.025244	23.360989	110	8	1.1	0.9;
	1011	1	-0.2	-1	0	0	0	1.022912	-1.384287	150	8	1.1	0.9;
	1013	1	136.06	40.4	0	12.08	0	0.995803	-46.382009	220	4	1.1	0.9;
	1015	1	44.55	0	0	4.78	0	1.052586	-5.374132	220	5	1.1	0.9;
	1020	1	8.21	3.31	0	8.38	0	1.040463	6.605453	150	8	1.1	0.9;
	1021	2	0	0	0	0	0	1.019135	-46.034292	220	4	1.1	0.9;
	1023	1	-0	0	0	0.93	0	1.019064	-43.846601	220	4	1.1	0.9;
	1025	1	241.4	-2.5	0.224624	31.280331	0	1.02861	-51.715097	380	10	1.1	0.9;
	1026	1	201.4	53	0	9.67	0	1.028099	-29.440609	220	5	1.1	0.9;
	1027	1	-5.91	-2.33	0	30.393595	0	0.990137	-18.20472	380	5	1.1	0.9;
	1030	2	0	0	0	0	0	1.00979	-35.817545	220	4	1.1	0.9;
	1033	1	26.33	5.3	0	4.31	0	1.040136	-24.565838	220	5	1.1	0.9;
	1035	1	382.5	57.9	0	8.16	0	1.066688	-32.38237	220	5	1.1	0.9;
	1036	2	0	0	0	0	0	1.038657	-30.920325	220	10	1.1	0.9;
	1037	1	-0	0	0	0.55	0	1.019244	5.129195	380	2	1.1	0.9;
	1039	1	34.64	-0	0	0.91	0	1.033489	-19.577193	220	5	1.1	0.9;
	1040	1	31.94	8.8	0	3.79	0	1.042099	-9.004028	220	5	1.1	0.9;
	1043	2	0	0	0	0	0	1.029789	-3.679585	380	5	1.1	0.9;
	1046	1	47.05	19.3	0	0.04	0	0.996689	-4.648049	150	8	1.1	0.9;
	1051	1	-0	0	0	3.79	0	1.043295	-12.895086	220	5	1.1	0.9;
	1052	1	67.48	32.6	0	13.11	0	1.037394	22.543259	220	8	1.1	0.9;
	1057	1	-0	-0	0	0	0	1.046413	26.479841	220	8	1.1	0.9;
	1060	1	-0	0	0	75.38	0	1.066748	-41.463394	380	4	1.1	0.9;
	1064	1	215.2	-7.6	0	1.24	0	1.023328	-46.764421	220	4	1.1	0.9;
	1065	1	89.6	20	0	9.07	0	1.03126	-53.682126	220	4	1.1	0.9;
	1074	1	403.9	-0.1	0	143.1	0	1.04431	-42.949688	380	4	1.1	0.9;
	1077	1	-0	0	0	0.1	0	1.036293	-12.101853	220	5	1.1	0.9;
	1078	1	-0	-0	0	4.74	0	1.057667	-3.433753	220	5	1.1	0.9;
	1081	1	277.86	58.95	0	182.880644	0	1.038361	-10.516057	380	5	1.1	0.9;
	1083	2	0	0	0	0	0	1.046149	-8.612778	220	5	1.1	0.9;
	1086	1	12.01	5.6	0	39.29	0	1.030324	-46.704132	380	4	1.1	0.9;
	1090	1	56.06	18.7	0	4.24	0	1.009868	-21.716046	220	5	1.1	0.9;
	1093	2	0	0	0	0	0	1.026457	-8.509476	380	5	1.1	0.9;
	1094	1	28.53	11.3	0	3.25	0	1.035847	6.036454	150	8	1.1	0.9;
	1096	1	0	-0	0	0.16	0	1.051471	-1.112508	220	5	1.1	0.9;
	1100	2	0	0	0	0	0	1.045212	-4.342005	220	5	1.1	0.9;
	1101	1	-0	-0	0	2.05	0	1.031004	-14.234122	220	5	1.1	0.9;
	1102	2	0	0	0	0	0	1.051902	-6.664986	220	5	1.1	0.9;
	1105	1	29.83	7.3	0	6.49	0	1.074362	-26.354638	220	5	1.1	0.9;
	1106	1	192.4	63.9	0	10	0	1.039063	4.797809	150	8	1.1	0.9;
	1109	1	-54.06	13.1	0	14.23	0	1.026975	-34.564954	220	10	1.1	0.9;
	1111	1	0	0	0	1.1	0	1.038462	-8.885347	220	5	1.1	0.9;
	1114	1	59.27	5.1	0	26.03	0	1.049971	-38.366222	380	4	1.1	0.9;
	1115	1	-40.54	9.3	0	3.05	0	1.026059	-38.808007	220	4	1.1	0.9;
	1121	2	0	0	0	0	0	1.019674	-38.721714	220	4	1.1	0.9;
	1123	1	-9	-10.19	0	0	0	1.023952	17.512847	150	8	1.1	0.9;
	1124	1	-0	-0	0	8.61	0	1.061545	15.555095	220	8	1.1	0.9;
	1129	1	190.47	55.03	0	6.7	0	1.003091	-22.556815	220	5	1.1	0.9;
	1133	1	115.2	319.2	0	290.91	0	1.060996	-43.314848	380	4	1.1	0.9;
	1136	1	16.22	2	0	0.71	0	1.032589	-18.567697	220	5	1.1	0.9;
	1137	1	-0	-0	0	0.05	0	1.047141	-24.469687	220	5	1.1	0.9;
	1146	1	32.84	-0	0	1.41	0	1.027082	-5.621485	220	5	1.1	0.9;
	1151	1	-0	-0	0	18.82	0	1.042198	-20.641446	220	5	1.1	0.9;
	1153	1	12.41	3.9	0	2.23	0	1.038665	-6.090842	220	5	1.1	0.9;
	1155	1	-0	0	0	1.91	0	1.041573	-51.256265	380	10	1.1	0.9;
	1156	1	65.67	24	0	9.95	0	1.037806	-22.529879	220	5	1.1	0.9;
	1159	1	112.4	13	0	3.99	0	1.05154	-36.310129	220	5	1.1	0.9;
	1163	1	-0	-0	0	16.04	0	0.994226	-1.33508	380	2	1.1	0.9;
	1168	1	0	-0	0	89.06	0	1.044876	-42.073382	380	4	1.1	0.9;
	1172	1	-0	-0	0	2.3	0	1.074074	-25.585411	220	5	1.1	0.9;
	1173	2	0	0	0	0	0	1.026125	-41.202223	220	4	1.1	0.9;
	1174	2	0	0	0	0	0	1.004865	-37.841731	150	10	1.1	0.9;
	1175	2	0	0	0	0	0	1.040018	-39.23186	220	4	1.1	0.9;
	1179	1	86.6	23.4	0	0.39	0	1.032826	-2.760554	220	5	1.1	0.9;
	1183	1	199.4	-19.9	0	6.68	0	1.065333	-33.138248	220	5	1.1	0.9;
	1185	1	150.4	15.3	0	8.09	0	1.014073	-43.85169	220	4	1.1	0.9;
	1187	1	67.57	21.2	0	5.71	0	1.034555	-4.939632	220	5	1.1	0.9;
	1188	1	30.93	6.6	0	9.09	0	1.055512	-14.263576	220	5	1.1	0.9;
	1194	1	129.6	13.3	0	13.98	0	1.038615	-15.144566	380	5	1.1	0.9;
	1198	1	0	0	0	30.637696	0	0.990911	-14.574593	380	5	1.1	0.9;
	1200	1	124.8	7.9	0	48.99	0	1.029044	-41.25002	380	4	1.1	0.9;
	1201	1	114.1	23.5	0	1.08	0	1.044872	-19.749595	220	5	1.1	0.9;
	1205	1	-0.19	-1	0	0	0	1.024412	2.729649	150	8	1.1	0.9;
	1207	1	70.18	28.7	0	0.8	0	1.002426	-0.400479	150	8	1.1	0.9;
	1216	1	0	0	0	4.93	0	1.0433	-10.974434	220	5	1.1	0.9;
	1221	1	230.5	91	0	57.67	0	1.021136	-1.469658	150	8	1.1	0.9;
	1227	1	-0.18	-1	0	0	0	1.02165	11.131243	150	8	1.1	0.9;
	1232	1	0	-0	0	81.52	0	1.050627	-49.560634	380	4	1.1	0.9;
	1233	1	0	-0	0	1.72	0	1.037734	-22.55474	220	5	1.1	0.9;
	1234	1	71.78	17.9	0	1.96	0	1.021063	-16.53158	220	5	1.1	0.9;
	1237	2	0	0	0	0	0	1.105045	-6.257773	380	5	1.3	0.7;
	1239	1	0	-0	0	0	0	1.00029	-54.869777	220	10	1.1	0.9;
	1240	1	23.53	4	0	9.39	0	1.023595	-41.324095	220	4	1.1	0.9;
	1241	2	0	0	0	0	0	1.021438	-37.891799	220	4	1.1	0.9;
	1248	1	0	0	0	1.17	0	1.042662	-10.670454	220	5	1.1	0.9;
	1249	1	348.93	26.47	0	9.36	0	1.019757	-11.249323	380	5	1.1	0.9;
	1250	1	39.44	24.9	0	4.18	0	0.992765	-35.003016	150	10	1.1	0.9;
	1251	2	0	0	0	0	0	1.0265	-8.519028	380	5	1.1	0.9;
	1255	1	20.62	8.2	0	6.62	0	1.013822	4.4155	150	8	1.1	0.9;
	1260	1	50.86	36.6	0	0	0	1.006152	16.289799	150	8	1.1	0.9;
	1262	1	-0	0	0	10.73	0	1.038366	-22.062584	220	5	1.1	0.9;
	1265	1	173	-21.3	0	5.93	0	1.066038	-50.142036	220	5	1.1	0.9;
	1271	2	0	0	0	0	0	1.015573	-46.050489	220	4	1.1	0.9;
	1275	1	52.46	2.9	0	0.35	0	1.029911	-18.88645	220	5	1.1	0.9;
	1277	1	-0.2	-1	0	0	0	1.023051	-1.382743	150	8	1.1	0.9;
	1280	1	0	-0	0	107.39	0	1.074452	-36.539156	380	4	1.1	0.9;
	1282	1	202	34.1	0	1.94	0	1.022795	-46.751133	220	4	1.1	0.9;
	1294	1	6.11	-0	0	2.51	0	1.014068	-48.160177	220	4	1.1	0.9;
	1295	2	0	0	0	0	0	1.038943	-8.62346	220	5	1.1	0.9;
	1301	1	479	99.2	0	10.68	0	1.050333	-28.761692	220	5	1.1	0.9;
	1305	1	-0	-0	0	4.34	0	1.03884	-22.413139	220	5	1.1	0.9;
	1311	1	37.94	15.2	0	35.31	0	1.032916	-21.685441	380	5	1.1	0.9;
	1317	1	149.3	-2	0	20.77	0	1.012625	3.260404	380	2	1.1	0.9;
	1326	1	0	0	0	0.98	0	1.0686	-35.059803	220	5	1.1	0.9;
	1330	1	100.6	15.3	0	1.28	0	0.993108	-1.662148	150	8	1.1	0.9;
	1331	1	-0.17	-1.82	0	0	0	1.050091	18.748368	380	8	1.1	0.9;
	1334	1	0	0	0	0.59	0	1.023695	-16.319863	220	5	1.1	0.9;
	1338	1	29.03	9.4	0	5.11	0	1.046601	11.309824	110	8	1.1	0.9;
	1339	1	108.4	-2.4	0	80.12	0	1.054799	-46.022749	380	10	1.1	0.9;
	1341	2	0	0	0	0	0	1.030396	-14.229973	220	5	1.1	0.9;
	1343	1	-0	-0	0	58.3	0	0.998625	-14.148042	380	5	1.1	0.9;
	1344	1	-0	0	0	2.13	0	1.024638	-46.046982	220	4	1.1	0.9;
	1347	1	0	-0	0	12.2	0	1.027222	-52.983085	220	4	1.1	0.9;
	1350	1	30.53	13.8	0	2.19	0	1.021086	-0.760994	150	8	1.1	0.9;
	1352	1	28.93	11.9	0	3.84	0	1.016527	-1.088293	150	8	1.1	0.9;
	1354	2	0	0	0	0	0	1.04168	-7.390066	220	5	1.1	0.9;
	1355	1	-3.29	-1.32	0	190.704588	0	1.032403	-8.543964	380	5	1.1	0.9;
	1362	1	-0	-0	0	6.93	0	1.045758	26.397012	220	8	1.1	0.9;
	1364	1	40.85	14.5	0	0.59	0	1.035537	-12.095863	220	5	1.1	0.9;
	1367	1	0	-0	0.175687	-1.509401	0	1.008895	-0.195912	380	2	1.1	0.9;
	1380	1	145.2	21.4	0	1.09	0	1.028172	-22.726688	220	5	1.1	0.9;
	1391	2	0	0	0	0	0	1.028915	-37.143274	220	4	1.1	0.9;
	1392	1	-0	-0	0	4.2	0	1.027508	18.540668	220	8	1.1	0.9;
	1393	1	2.4	0.7	0	10.69	0	1.026572	-45.348551	220	4	1.1	0.9;
	1394	1	30.93	7.6	0	4.69	0	1.074262	-26.342945	220	5	1.1	0.9;
	1397	1	0	-0	0	2.14	0	1.039499	-12.885891	220	5	1.1	0.9;
	1398	1	196.7	44.2	0	0.1	0	1.016914	-19.000915	220	5	1.1	0.9;
	1399	2	0	0	0	0	0	1.052722	-48.692525	380	4	1.1	0.9;
	1401	1	11.91	2.9	0	5.27	0	1.035517	-14.762042	220	5	1.1	0.9;
	1407	1	0	-0	0	4.15	0	0.996889	-49.175373	150	10	1.1	0.9;
	1409	2	0	0	0	0	0	1.024907	-48.140766	220	4	1.1	0.9;
	1411	1	0	-0	0	7.52	0	1.01205	-47.31797	220	4	1.1	0.9;
	1414	1	144.7	20.3	0	6.35	0	1.022618	-27.926817	220	5	1.1	0.9;
	1415	1	0	-0	0	0.3	0	1.031083	-11.51315	220	5	1.1	0.9;
	1418	1	40.55	3.6	0	1.88	0	1.021924	-54.400148	220	4	1.1	0.9;
	1421	1	42.95	19.8	0	3.85	0	1.021629	-51.698681	220	4	1.1	0.9;
	1422	2	0	0	0	0	0	1.045292	-21.037138	220	5	1.1	0.9;
	1430	1	20.82	7.9	0	5.62	0	1.015624	-50.375421	220	4	1.1	0.9;
	1435	1	-0	0	0	3.85	0	1.044029	-24.378621	220	5	1.1	0.9;
	1436	2	0	0	0	0	0	1.049261	-12.98242	220	5	1.1	0.9;
	1437	1	204.6	53.2	0	7.4	0	0.9896	-37.927533	220	4	1.1	0.9;
	1439	1	110.1	2	0	13.85	0	1.025342	-48.121825	220	4	1.1	0.9;
	1445	1	238.4	210.9	0	0	0	1.001732	-0.206284	220	2	1.1	0.9;
	1448	1	0	0	0	9.49	0	1.074407	-15.342169	220	5	1.1	0.9;
	1457	1	20.82	36	0	0.97	0	1.016731	-39.321312	220	4	1.1	0.9;
	1458	2	0	0	0	0	0	1.05352	-38.031423	380	4	1.1	0.9;
	1459	1	-0	-0	0	3.38	0	1.041987	-21.849566	220	5	1.1	0.9;
	1461	1	0	0	0	1.44	0	1.038081	-38.816129	220	4	1.1	0.9;
	1462	1	-0	0	0	0.42	0	1.045314	-21.037648	220	5	1.1	0.9;
	1463	1	109.6	36.4	0	13.24	0	1.03605	6.423553	150	8	1.1	0.9;
	1464	1	22.73	6.4	0	2.1	0	1.020741	7.919173	150	8	1.1	0.9;
	1465	1	-11.03	-5.11	0	29.706803	0	0.987783	-15.361156	380	5	1.1	0.9;
	1471	1	95.61	8.6	0	3.48	0	0.981221	-48.994904	220	4	1.1	0.9;
	1475	2	0	0	0	0	0	1.025403	-43.044459	380	4	1.1	0.9;
	1477	1	0	-0	0	5.51	0	1.034673	9.5982	150	8	1.1	0.9;
	1478	2	0	0	0	0	0	1.046165	-15.856607	220	5	1.1	0.9;
	1479	1	82.19	23.4	0	17.91	0	1.029701	-43.025361	380	4	1.1	0.9;
	1483	1	0	0	0	4.86	0	1.063082	-26.916487	220	5	1.1	0.9;
	1486	1	79.59	18.8	0	6.78	0	1.034765	-23.371441	220	5	1.1	0.9;
	1489	1	0	-0	0	2.54	0	1.0268	-35.902682	220	10	1.1	0.9;
	1491	1	117.5	-3	0	5.38	0	1.023943	-47.089173	220	4	1.1	0.9;
	1494	1	-1.85	-1.63	0	31.424588	0	1.032178	-11.529751	380	5	1.1	0.9;
	1498	2	0	0	0	0	0	1.078389	-39.66401	380	4	1.1	0.9;
	1501	1	-0	-0	0	12.57	0	1.030006	-53.597487	220	4	1.1	0.9;
	1502	1	-0	0	0	37.116157	0	1.026558	-26.733601	380	5	1.1	0.9;
	1503	1	142.8	-6.4	0	12.36	0	1.023199	-43.578655	220	4	1.1	0.9;
	1504	1	0	-0	0	1.84	0	1.026035	-14.299045	220	5	1.1	0.9;
	1513	1	-35.74	38.6	0	4.65	0	1.019243	-39.521783	220	4	1.1	0.9;
	1516	1	22.02	9	0	0.24	0	1.020964	-0.403438	150	8	1.1	0.9;
	1520	1	173.6	72.9	0	2.5	0	1.008347	-46.11215	220	4	1.1	0.9;
	1526	1	-10.86	-5.93	0	12.54	0	1.043661	-11.957383	220	5	1.1	0.9;
	1531	1	-0	0	0	0	0	1.00982	-11.496865	150	2	1.1	0.9;
	1538	1	36.34	15.2	0	0.3	0	1.035514	-8.768446	220	5	1.1	0.9;
	1539	1	73.58	3.24	0	52.35	0	1.059767	-15.313651	220	5	1.1	0.9;
	1541	1	125.4	45.8	0	0	0	1.051628	-23.909822	220	5	1.1	0.9;
	1543	1	116.7	64	0	10.41	0	0.993828	-49.272761	220	10	1.1	0.9;
	1545	1	42.55	8.2	0	2.06	0	1.041665	-23.501941	220	5	1.1	0.9;
	1547	1	-0	-0	0	1.02	0	1.038027	-15.068933	220	5	1.1	0.9;
	1549	1	300.2	-3.1	0	29.94	0	1.021156	-52.883236	380	4	1.1	0.9;
	1551	1	-0	-0	0	7.99	0	1.042613	-20.44489	220	5	1.1	0.9;
	1552	1	20.42	8.1	0	35.15	0	1.042022	-16.638586	380	5	1.1	0.9;
	1554	1	0	-0	0	4.96	0	1.020362	-16.068621	220	5	1.1	0.9;
	1556	1	-0	-0	0	0.95	0	1.063865	-11.487753	220	5	1.1	0.9;
	1562	1	77.39	24.1	0	2.67	0	1.040689	-25.900148	220	5	1.1	0.9;
	1564	1	140.4	49.4	0	4.5	0	1.017809	7.583606	150	8	1.1	0.9;
	1566	1	156.9	-4	0	2.48	0	1.035654	-24.787616	220	5	1.1	0.9;
	1568	1	69.08	21.5	0	1.65	0	1.045321	-25.198298	220	5	1.1	0.9;
	1570	2	0	0	0	0	0	1.079057	-35.031052	380	10	1.1	0.9;
	1578	1	0	-0	0	1.38	0	1.041117	-12.03536	220	5	1.1	0.9;
	1579	1	-0	0	0	16.56	0	1.001361	-5.002873	380	2	1.1	0.9;
	1584	1	0	-0	0	3.24	0	1.028531	-14.301081	220	5	1.1	0.9;
	1591	1	-0.16	-11.88	0	16.12	0	1.01399	17.300476	380	8	1.1	0.9;
	1592	1	28.33	44.8	0	0.51	0	1.041196	-2.388101	220	5	1.1	0.9;
	1595	1	71.77	13.57	0	3.84	0	1.015775	-29.314866	220	5	1.1	0.9;
	1597	1	51.96	13.4	0	0.65	0	1.022686	14.250036	110	8	1.1	0.9;
	1599	1	0	0	0	0.48	0	1.014966	-48.210953	220	4	1.1	0.9;
	1600	1	38.74	14.1	0	12.78	0	1.030988	-53.564846	220	4	1.1	0.9;
	1602	1	45.75	4.2	0	4.9	0	1.015546	-59.838921	220	4	1.1	0.9;
	1604	2	0	0	0	0	0	1.044277	-6.654391	220	5	1.1	0.9;
	1605	1	-0	0	0	0.05	0	1.075745	-24.387924	220	5	1.1	0.9;
	1607	1	73.78	-0	0	2.69	0	1.026752	-5.477294	380	5	1.1	0.9;
	1609	1	0	0	0	6.55	0	1.049223	-20.706157	220	5	1.1	0.9;
	1611	1	274.1	166.5	0	0.56	0	1.050807	-9.150756	150	2	1.1	0.9;
	1616	1	0	0	0	14.45	0	1.003428	-3.942054	380	2	1.1	0.9;
	1618	2	0	0	0	0	0	1.079319	-43.562064	380	4	1.1	0.9;
	1625	1	73.18	22.3	0	8.77	0	1.043133	-19.969473	220	5	1.1	0.9;
	1629	1	41.95	8.5	0	1.05	0	1.044497	-11.070857	220	5	1.1	0.9;
	1632	1	20.82	7.9	0	1.07	0	1.024746	-52.071172	220	4	1.1	0.9;
	1633	1	-0	-0	0	45.02	0	1.012918	16.991789	150	8	1.1	0.9;
	1635	1	-0.18	-1.06	0	0	0	1.015425	10.923415	150	8	1.1	0.9;
	1638	1	70.08	11.1	0	9.84	0	1.029115	-53.706686	220	4	1.1	0.9;
	1639	1	51.66	8	0	4.51	0	1.032136	-52.870728	220	4	1.1	0.9;
	1642	2	0	0	0	0	0	1.031925	0.361584	380	5	1.1	0.9;
	1643	1	95.71	25.9	0	11.38	0	1.040729	-25.9337	220	5	1.1	0.9;
	1644	1	449.56	245.72	0	94.55	0	1.049584	3.538596	380	5	1.1	0.9;
	1645	1	60.27	9.4	0	0.62	0	1.028996	-53.918767	220	4	1.1	0.9;
	1662	1	3.67	-1.86	0	3.89	0	1.044929	-20.83595	220	5	1.1	0.9;
	1666	1	34.04	0	0	2.4	0	1.010007	-47.695122	220	4	1.1	0.9;
	1670	1	46.45	19.1	0	6.66	0	1.021732	-0.338232	150	8	1.1	0.9;
	1672	1	-28.13	-0.83	0	0.1	0	1.063888	-11.501012	220	5	1.1	0.9;
	1676	1	-0	0	0.171586	-1.4632	0	0.993902	-1.396138	380	2	1.1	0.9;
	1680	2	0	0	0	0	0	1.044209	-2.249484	380	5	1.1	0.9;
	1685	2	0	0	0	0	0	1.005778	-37.621541	150	10	1.1	0.9;
	1689	1	-0	-0	0	58.96	0	1.033205	11.513144	380	8	1.1	0.9;
	1692	1	83.19	25.8	0	0.33	0	1.047517	-10.481325	220	5	1.1	0.9;
	1697	2	0	0	0	0	0	1.041486	-50.115438	380	4	1.1	0.9;
	1702	1	47.45	18.7	0	1.36	0	1.035777	4.260134	150	8	1.1	0.9;
	1703	1	68.28	39	0.19855	79.84755	0	1.05922	-43.51259	380	10	1.1	0.9;
	1704	1	62.16	8	0	2.99	0	1.043162	-13.087499	220	5	1.1	0.9;
	1708	2	0	0	0	0	0	1.037844	-13.203761	380	5	1.1	0.9;
	1709	1	-15.38	-4.94	0	33.241519	0	1.006049	-14.735655	380	5	1.1	0.9;
	1711	1	25.13	-61.6	0	3.68	0	1.047493	-29.9792	220	5	1.1	0.9;
	1716	1	0	-0	0	16.62	0	1.025999	-40.615588	380	4	1.1	0.9;
	1721	2	0	0	0	0	0	1.061536	2.487523	220	5	1.1	0.9;
	1722	1	-0.2	-1	0	0	0	1.024592	9.445328	150	8	1.1	0.9;
	1728	1	34.44	14.4	0	48.56	0	1.051238	-48.488813	380	4	1.1	0.9;
	1730	1	-17.18	-4.35	0	44.419365	0	1.024629	-14.526553	380	5	1.1	0.9;
	1734	1	-0	0	0	3.65	0	0.989566	-54.079919	220	10	1.1	0.9;
	1741	1	162.5	48.4	0	5.73	0	0.986128	-45.182156	220	4	1.1	0.9;
	1742	1	0	0	0	6.68	0	1.027401	-32.553844	220	5	1.1	0.9;
	1746	1	-0	0	0	1.86	0	1.047833	-24.129867	220	5	1.1	0.9;
	1747	2	0	0	0	0	0	1.055274	12.851962	380	8	1.1	0.9;
	1750	1	58.46	6.6	0	10.13	0	1.035938	-19.217986	220	5	1.1	0.9;
	1752	1	85.8	13.4	0	3.23	0	1.035715	-52.712594	220	4	1.1	0.9;
	1754	2	0	0	0	0	0	1.040804	-17.145993	380	5	1.1	0.9;
	1758	1	0	-0	0	121.29	0	1.034693	-4.444343	380	5	1.1	0.9;
	1763	1	-0	-0	0	0.28	0	1.033831	-30.862865	380	5	1.1	0.9;
	1767	1	333.5	-1.4	0	8.37	0	1.048826	-45.648551	220	5	1.1	0.9;
	1768	1	73.98	22.5	0	1.15	0	1.038366	-11.138188	220	5	1.1	0.9;
	1769	1	35.14	1.8	0	6.12	0	1.03194	-46.566188	220	4	1.1	0.9;
	1770	1	-0	-0	0	2.55	0	1.015596	-41.191024	220	10	1.1	0.9;
	1774	1	-32.04	10.3	0	82.52	0	1.056861	-45.124391	380	10	1.1	0.9;
	1775	1	-0	0	0	1.64	0	1.049957	-26.065865	220	5	1.1	0.9;
	1777	1	0	-0	0	0.98	0	1.033981	16.473721	110	8	1.1	0.9;
	1778	1	-0.18	-3.85	0	0	0	1.012854	16.97545	150	8	1.1	0.9;
	1787	1	32.94	10.1	0	0.41	0	1.017619	12.619151	110	8	1.1	0.9;
	1788	2	0	0	0	0	0	1.018729	15.891067	380	8	1.1	0.9;
	1793	2	0	0	0	0	0	1.023406	-43.929454	220	4	1.1	0.9;
	1794	2	0	0	0	0	0	1.052726	-0.25556	380	5	1.1	0.9;
	1798	1	-0.63	-0.83	0	22.810288	0	0.981441	-7.400061	380	5	1.1	0.9;
	1802	1	8.51	3.2	0	2.93	0	1.004678	-46.035316	220	4	1.1	0.9;
	1803	1	-0	-0	0	0.05	0	1.038641	-22.064619	220	5	1.1	0.9;
	1808	2	0	0	0	0	0	1.047818	-0.83857	220	5	1.1	0.9;
	1810	1	52.46	11.3	0	0.27	0	1.037208	17.408937	110	8	1.1	0.9;
	1813	1	88.92	19.79	0	2.81	0	1.028777	-7.526931	220	5	1.1	0.9;
	1815	1	63.66	9.7	0	0	0	1.025638	0.196834	220	2	1.1	0.9;
	1817	1	108.91	33.7	0	56.24	0	1.067642	-12.571734	380	5	1.1	0.9;
	1833	1	0	0	0	3.33	0	1.036099	-22.50831	220	5	1.1	0.9;
	1836	1	62.37	25.5	0	3.98	0	1.012194	-1.655599	150	8	1.1	0.9;
	1838	1	0	-0	0	3.02	0	1.037099	-22.190305	220	5	1.1	0.9;
	1841	1	56.26	8.5	0	5.32	0	1.032038	-53.00315	220	4	1.1	0.9;
	1844	1	-28.14	-0.76	0	0.1	0	1.063907	-11.500461	220	5	1.1	0.9;
	1848	1	22.02	21.8	0	10.29	0	1.032678	-33.137265	220	10	1.1	0.9;
	1850	2	0	0	0	0	0	1.020156	-51.03363	220	4	1.1	0.9;
	1851	2	0	0	0	0	0	1.0492	-12.99194	220	5	1.1	0.9;
	1852	2	0	0	0	0	0	1.056131	-25.812508	220	5	1.1	0.9;
	1857	1	0	0	0	1.53	0	1.032169	-11.026395	220	5	1.1	0.9;
	1860	1	88.8	-3.2	0	13.63	0	1.061586	-25.618695	220	5	1.1	0.9;
	1862	1	177.4	41.7	0	2.98	0	0.987495	-40.251221	150	10	1.1	0.9;
	1866	1	209.2	23.5	0	6.44	0	1.04107	-28.326716	220	5	1.1	0.9;
	1868	1	11.71	3.8	0	2.96	0	1.039214	-15.807769	220	5	1.1	0.9;
	1869	1	-2.23	-1.27	0	3.64	0	1.035924	-44.095788	220	4	1.1	0.9;
	1870	1	-0	0	0	0.74	0	1.015094	1.339807	150	8	1.1	0.9;
	1872	1	94.3	13.8	0	5.3	0	1.029381	-53.816205	220	4	1.1	0.9;
	1876	1	-0	-0	0	6.273568	0	1.030623	-14.259772	380	5	1.1	0.9;
	1880	1	50.26	16.4	0	0.23	0	1.009514	0.69841	150	8	1.1	0.9;
	1883	1	27.33	5	0	6.77	0	1.028511	-16.427961	220	5	1.1	0.9;
	1888	2	0	0	0	0	0	1.043102	-2.430762	220	5	1.1	0.9;
	1889	1	57.16	0	0	6.58	0	1.024264	10.369446	110	8	1.1	0.9;
	1890	2	0	0	0	0	0	1.050852	54.166807	380	8	1.1	0.9;
	1894	1	22.22	5.9	0	10.28	0	1.02411	-47.579361	220	4	1.1	0.9;
	1895	1	20.82	-0	0	0.14	0	1.067859	7.62397	220	5	1.1	0.9;
	1896	1	-0	-0	0	1.18	0	1.018399	-6.164969	380	5	1.1	0.9;
	1897	1	53.96	4.9	0	4.21	0	1.014095	-60.102849	220	4	1.1	0.9;
	1898	2	0	0	0	0	0	1.077705	-35.592551	380	4	1.1	0.9;
	1899	1	0	-0	0.202076	48.319777	0	1.070541	-46.625448	380	10	1.1	0.9;
	1909	2	0	0	0	0	0	1.021788	-39.436854	220	4	1.1	0.9;
	1910	1	-0	0	0	2.68	0	1.050804	-6.677369	220	5	1.1	0.9;
	1912	1	52.06	14.3	0	1.59	0	0.983155	-44.95221	220	4	1.1	0.9;
	1913	1	91.1	54.8	0	30.59	0	1.023641	9.891407	150	8	1.1	0.9;
	1914	2	0	0	0	0	0	1.038338	-20.75929	220	5	1.1	0.9;
	1917	1	197.5	-31	0	17.34	0	1.048959	-25.176511	220	5	1.1	0.9;
	1923	1	192.11	45.5	0	68.04	0	1.002269	-11.665771	380	5	1.1	0.9;
	1936	1	0	0	0	30.78	0	1.035973	5.613493	150	8	1.1	0.9;
	1940	1	0	0	0	1.19	0	1.035883	-18.056411	220	5	1.1	0.9;
	1942	1	44.75	4.1	0	3.12	0	1.016842	-58.849363	220	4	1.1	0.9;
	1947	2	0	0	0	0	0	1.056511	-39.920881	380	4	1.1	0.9;
	1956	1	0	0	0	51.66	0	1.064864	30.768271	380	8	1.1	0.9;
	1957	1	76.28	2.9	0	7.28	0	1.019244	-48.276935	220	4	1.1	0.9;
	1959	2	0	0	0	0	0	1.066953	-10.612706	220	5	1.1	0.9;
	1965	1	248.7	8.1	0	12.46	0	1.065913	-49.787032	220	5	1.1	0.9;
	1968	1	118.4	137	0	0	0	1.031762	-10.964931	150	2	1.1	0.9;
	1969	1	0	-0	0	0.86	0	1.03098	-53.267997	220	4	1.1	0.9;
	1973	1	210.1	66.2	0	59.36	0	1.074344	-26.328901	220	5	1.1	0.9;
	1974	1	241.9	38.8	0	12.46	0	0.966737	-43.95225	220	10	1.1	0.9;
	1979	1	0	0	0	0.14	0	1.073367	-33.684297	220	5	1.1	0.9;
	1980	1	177.3	36.4	0	0.7	0	1.039678	-21.457316	220	5	1.1	0.9;
	1984	1	0	0	0	0	0	1.024246	4.396981	150	8	1.1	0.9;
	1985	1	147.4	-55.9	0	1.12	0	0.988129	-57.352119	220	4	1.1	0.9;
	1989	1	69.78	20.3	0	10.26	0	1.029228	14.956885	110	8	1.1	0.9;
	1998	1	25.73	7.3	0	2.82	0	1.032648	-20.169174	220	5	1.1	0.9;
	1999	2	0	0	0	0	0	1.080207	-39.478229	380	10	1.1	0.9;
	2002	1	0	-0	0	0	0	1.087207	32.342614	380	8	1.1	0.9;
	2006	1	-40.44	1.5	0	14.2	0	1.022687	-38.139249	220	10	1.1	0.9;
	2009	1	154.8	-47.1	0	4.18	0	1.029527	-50.600329	220	4	1.1	0.9;
	2012	1	41.45	4.9	0	18.43	0	1.037869	-18.701859	220	5	1.1	0.9;
	2019	1	72.08	21.3	0	7.19	0	1.043225	-21.113594	220	5	1.1	0.9;
	2020	1	0	0	0	9.98	0	1.04955	-24.137238	220	5	1.1	0.9;
	2021	1	3.9	1	0	3.81	0	1.048571	-25.008749	220	5	1.1	0.9;
	2030	2	0	0	0	0	0	1.072679	15.936755	150	8	1.1	0.9;
	2031	1	-0	-0	0.104186	55.436651	0	1.080195	-39.49283	380	10	1.1	0.9;
	2035	2	0	0	0	0	0	1.034553	0.621632	380	5	1.1	0.9;
	2042	1	14.02	6.5	0	0.09	0	1.041161	-5.936709	220	5	1.1	0.9;
	2043	1	3.2	1.7	0	0.05	0	1.045278	-21.03811	220	5	1.1	0.9;
	2044	1	0	0	0	0.14	0	1.033509	-12.417674	220	5	1.1	0.9;
	2050	2	0	0	0	0	0	1.008592	-1.119799	380	5	1.1	0.9;
	2055	1	52.66	21.5	0	0.04	0	0.99757	-4.666491	150	8	1.1	0.9;
	2056	1	-0	0	0	7.338808	0	1.034337	-9.107467	380	5	1.1	0.9;
	2057	1	120.2	20.5	0	19.94	0	1.04448	-30.995545	220	5	1.1	0.9;
	2058	1	0	0	0	1.31	0	1.056876	-45.124474	380	10	1.1	0.9;
	2060	1	115.1	23.5	0	3.78	0	1.01599	-58.09796	220	4	1.1	0.9;
	2072	1	101.2	2.1	0	38.43	0	1.061875	-17.377654	380	5	1.1	0.9;
	2077	1	72.18	44.9	0	6.34	0	1.026664	11.93662	150	8	1.1	0.9;
	2078	1	77.19	18.6	0	2.22	0	1.021844	-14.701929	220	5	1.1	0.9;
	2079	1	0	-0	0	13.94	0	1.047339	-12.138161	220	5	1.1	0.9;
	2081	1	30.93	2.8	0	2.11	0	0.981221	-48.708598	220	4	1.1	0.9;
	2083	1	121.4	20.2	0	16.61	0	1.032804	-18.511366	220	5	1.1	0.9;
	2085	2	0	0	0	0	0	1.03273	2.589028	380	5	1.1	0.9;
	2088	1	-0	0	0	6.02	0	1.048725	-12.063245	220	5	1.1	0.9;
	2089	1	199.86	56.93	0	46.83	0	1.041575	-25.815937	220	5	1.1	0.9;
	2091	2	0	0	0	0	0	1.026543	-36.391911	220	4	1.1	0.9;
	2093	1	-5.64	-2.84	0	34.582195	0	1.013241	-8.07029	380	5	1.1	0.9;
	2096	2	0	0	0	0	0	1.050714	-45.901726	380	4	1.1	0.9;
	2101	1	-0	-0	0	38.32	0	1.030385	-5.532	380	5	1.1	0.9;
	2107	2	0	0	0	0	0	1.048234	-0.02966	380	2	1.1	0.9;
	2116	1	53.36	7.1	0	6.02	0	1.027792	-45.963304	220	4	1.1	0.9;
	2117	1	-0.2	-1	0	0	0	1.023061	-1.382627	150	8	1.1	0.9;
	2120	1	-0.19	-3.76	0	0	0	1.023649	9.89143	150	8	1.1	0.9;
	2126	1	92.3	14.6	0	1.14	0	1.022588	-52.26225	220	4	1.1	0.9;
	2128	1	79.09	18.7	0	10.8	0	1.035853	-23.092097	220	5	1.1	0.9;
	2129	1	0	0	0	56.686515	0	1.018774	-8.206837	380	5	1.1	0.9;
	2132	1	107.3	14.2	0	20.23	0	1.038784	-15.777605	220	5	1.1	0.9;
	2140	1	0	0	0	1.2	0	0.991354	-52.688498	150	10	1.1	0.9;
	2142	1	0	-0	0	7.63	0	1.052708	-7.87962	220	5	1.1	0.9;
	2143	1	129.5	38.3	0	5.42	0	0.989853	-43.142584	220	4	1.1	0.9;
	2147	1	6.11	-0	0	0.92	0	1.01158	-47.47921	220	4	1.1	0.9;
	2150	2	0	0	0	0	0	1.064827	23.234609	380	8	1.1	0.9;
	2154	1	0	0	0.29	13.47	0	1.048188	11.193021	380	2	1.1	0.9;
	2155	1	63.67	20.2	0	6.13	0	1.04127	-5.926904	220	5	1.1	0.9;
	2161	1	35.66	-9.39	0	7.89	0	1.048473	-28.734289	220	5	1.1	0.9;
	2166	1	20.92	7.7	0	1.21	0	1.050718	-14.695738	220	5	1.1	0.9;
	2168	1	8.6	-1.48	0	0	0	1.026021	-12.839247	150	2	1.1	0.9;
	2172	1	114.2	34.4	0	10.29	0	1.034538	-49.361418	220	4	1.1	0.9;
	2174	1	0	0	0	61.795782	0	1.009806	-13.208287	380	5	1.1	0.9;
	2177	2	0	0	0	0	0	1.070558	-28.465666	220	5	1.1	0.9;
	2180	1	0	-0	0	2.24	0	1.015768	-0.688874	150	8	1.1	0.9;
	2183	1	0	0	0	32.849821	0	0.994736	-16.202074	380	5	1.1	0.9;
	2188	1	25.23	11.9	0	18.72	0	1.049958	-49.355234	380	4	1.1	0.9;
	2189	1	-0	0	0	0.18	0	1.049117	-6.264179	220	5	1.1	0.9;
	2193	2	0	0	0	0	0	1.014181	-46.990737	220	4	1.1	0.9;
	2197	2	0	0	0	0	0	1.036195	-6.001598	380	5	1.1	0.9;
	2202	1	213	39.6	0	20.63	0	1.027976	-49.093244	380	4	1.1	0.9;
	2206	2	0	0	0	0	0	1.020141	-34.018711	150	10	1.1	0.9;
	2208	1	74.98	26.9	0	2.96	0	1.046913	-13.479325	220	5	1.1	0.9;
	2225	2	0	0	0	0	0	1.02967	-44.119861	220	4	1.1	0.9;
	2228	1	143.1	18.9	0	3.42	0	1.018483	23.24717	110	8	1.1	0.9;
	2229	1	0	-0	0	2.91	0	1.060553	-13.095765	220	5	1.1	0.9;
	2230	1	0	-0	0	1.71	0	1.036241	-22.50954	220	5	1.1	0.9;
	2231	1	0	0	0	0.45	0	1.049238	1.176377	220	5	1.1	0.9;
	2233	1	269.9	12.4	0	4.09	0	0.990149	-55.658191	150	10	1.1	0.9;
	2235	1	193.1	69.9	0	28.91	0	1.014556	-44.573367	380	4	1.1	0.9;
	2236	2	0	0	0	0	0	1.023164	-42.172116	220	4	1.1	0.9;
	2240	1	66.87	21.8	0	12.3	0	1.01274	3.69707	150	8	1.1	0.9;
	2242	1	-0	-0	0	2.76	0	1.024412	-19.040124	220	5	1.1	0.9;
	2248	1	110.3	28.7	0	0.65	0	1.015625	-46.19205	220	4	1.1	0.9;
	2252	1	100.7	-4.2	0	1.57	0	1.067898	-48.0026	220	5	1.1	0.9;
	2261	1	0	-0	0	20.5	0	1.066297	-40.410185	380	4	1.1	0.9;
	2267	2	0	0	0	0	0	1.000857	11.93009	380	2	1.1	0.9;
	2268	1	0	-0	0.183167	22.675025	0	0.997933	-5.134076	380	2	1.1	0.9;
	2271	1	0	-0.01	0	5.48	0	1.021363	11.424715	150	8	1.1	0.9;
	2273	1	141.3	32.4	0	7.29	0	1.073962	-26.368106	220	5	1.1	0.9;
	2276	2	0	0	0	0	0	1.055141	-23.885692	220	5	1.1	0.9;
	2286	1	208.88	60.8	0	6.68	0	1.027113	-8.732478	220	5	1.1	0.9;
	2288	1	170.9	39	0	11.25	0	1.036625	-8.550708	220	5	1.1	0.9;
	2291	2	0	0	0	0	0	1.068687	-10.641009	220	5	1.1	0.9;
	2297	1	93.2	13.6	0	14.9	0	1.029726	-53.825044	220	4	1.1	0.9;
	2299	1	402.1	110.2	0	0	0	1.036701	-11.109317	150	2	1.1	0.9;
	2301	1	82.39	12.3	0	9.02	0	1.029153	-53.664396	220	4	1.1	0.9;
	2303	1	0	-0	0	7.94	0	1.035356	-15.857907	220	5	1.1	0.9;
	2304	1	88.7	-0.1	0	2.83	0	1.053621	-25.856227	220	5	1.1	0.9;
	2305	1	21.22	8.3	0	19.1	0	1.022377	0.249057	150	8	1.1	0.9;
	2308	1	32.04	6.6	0	0.44	0	1.038049	-24.108761	220	5	1.1	0.9;
	2313	1	0	-0	0	0.41	0	1.045713	-21.296439	220	5	1.1	0.9;
	2319	1	39.64	20.9	0	0.65	0	1.058042	-8.99633	220	5	1.1	0.9;
	2327	1	114.5	25.2	0	17.57	0	1.045658	-25.17735	220	5	1.1	0.9;
	2328	1	-0	0	0	0.4	0	1.012206	-14.284271	220	5	1.1	0.9;
	2330	1	-0.2	-1.33	0	0	0	1.009792	1.255832	150	8	1.1	0.9;
	2331	1	-0.15	-1.4	0	0.01	0	1.015977	19.537852	110	8	1.1	0.9;
	2335	2	0	0	0	0	0	1.040978	6.656092	150	8	1.1	0.9;
	2337	1	103.4	27	0	1.41	0	1.044141	-29.826668	220	5	1.1	0.9;
	2340	1	96.91	13.9	0	3.98	0	1.040211	-12.241008	220	5	1.1	0.9;
	2341	1	-0	0	0	3.28	0	1.059334	-26.086007	220	5	1.1	0.9;
	2343	1	33.04	5.6	0	14.55	0	1.023455	-42.321013	220	4	1.1	0.9;
	2350	1	-0	0	0	7.03	0	1.008285	16.440309	150	8	1.1	0.9;
	2355	1	54.46	11.9	0	3.93	0	1.020865	-50.666226	220	4	1.1	0.9;
	2358	1	52.96	8.2	0	5.68	0	1.029163	-53.685458	220	4	1.1	0.9;
	2359	2	0	0	0	0	0	1.025004	-9.912428	380	5	1.1	0.9;
	2360	1	215.5	31.3	0	14.23	0	1.039139	-21.221662	220	5	1.1	0.9;
	2361	1	16.9	3.85	0	25.09	0	1.047053	-29.210131	220	5	1.1	0.9;
	2365	1	37.94	15.4	0	117.386167	0	1.03234	-22.207428	380	5	1.1	0.9;
	2372	1	0	-0	0	14.490904	0	0.990983	-17.504299	380	5	1.1	0.9;
	2377	1	101.7	17.6	0	13.14	0	1.032965	-14.916596	220	5	1.1	0.9;
	2381	1	32.94	13	0	4.04	0	1.032264	5.112304	150	8	1.1	0.9;
	2393	1	20.82	0	0	0.13	0	1.067851	7.737226	220	5	1.1	0.9;
	2403	1	37.24	5.3	0	14.78	0	1.027657	-47.417397	220	4	1.1	0.9;
	2406	1	0	0	0	1.69	0	1.04714	-24.469684	220	5	1.1	0.9;
	2407	1	117.13	52.4	0	7.51	0	0.995534	-39.27632	150	10	1.1	0.9;
	2410	1	3.9	-35.8	0	4.71	0	1.026919	-36.555421	220	10	1.1	0.9;
	2412	2	0	0	0	0	0	1.016849	-42.518502	220	4	1.1	0.9;
	2416	1	14.32	5.6	0	88.82	0	1.05268	-49.057941	380	4	1.1	0.9;
	2421	2	0	0	0	0	0	1.047704	-23.732507	220	5	1.1	0.9;
	2424	1	44.55	16.8	0	2.31	0	1.034737	-22.444225	220	5	1.1	0.9;
	2425	2	0	0	0	0	0	1.043025	-9.904307	220	5	1.1	0.9;
	2426	2	0	0	0	0	0	1.040828	-13.694298	220	5	1.1	0.9;
	2428	1	-0	-0	0	50	0	1.018049	0.563741	150	8	1.1	0.9;
	2430	1	-0	0	0	6.21	0	1.037062	-11.923631	220	5	1.1	0.9;
	2432	1	291.2	35.4	0	21.87	0	1.027519	-32.362347	220	5	1.1	0.9;
	2437	1	-0	-8.6	0	297.94	0	1.063879	-41.501682	380	4	1.1	0.9;
	2438	1	-0	0	0	0.64	0	1.045685	-21.296138	220	5	1.1	0.9;
	2441	1	0	-0	0.242534	1.143584	0	1.001003	-4.783502	380	2	1.1	0.9;
	2446	2	0	0	0	0	0	1.053565	7.758341	380	5	1.1	0.9;
	2448	1	34.24	16.5	0	0	0	1.04676	26.524166	220	8	1.1	0.9;
	2449	1	301.9	-103.2	0	0	0	1.064595	-5.855251	150	2	1.1	0.9;
	2457	1	30.73	9.6	0	2.75	0	1.03926	-24.663306	220	5	1.1	0.9;
	2458	1	132	26	0	27.18	0	1.00611	-13.480214	380	5	1.1	0.9;
	2459	1	-0	0	0	145.98	0	1.05033	-42.244541	380	4	1.1	0.9;
	2460	2	0	0	0	0	0	0.983266	-39.656055	220	10	1.1	0.9;
	2467	1	0	-0	0	11.65	0	1.01269	-14.132965	220	5	1.1	0.9;
	2468	2	0	0	0	0	0	1.049474	-6.197025	220	5	1.1	0.9;
	2471	1	103.5	21.7	0	8.88	0	1.017576	-47.068239	220	4	1.1	0.9;
	2475	1	0	0	0	0.21	0	1.062881	-7.838877	220	5	1.1	0.9;
	2478	1	50.96	6.8	0	9.13	0	1.029979	15.795765	110	8	1.1	0.9;
	2479	1	-2.65	-1.17	0	65.673108	0	1.007738	-2.60887	380	5	1.1	0.9;
	2481	2	0	0	0	0	0	1.041748	-18.217854	220	5	1.1	0.9;
	2482	1	-0.21	-1.98	0	0	0	1.023631	8.63119	150	8	1.1	0.9;
	2489	2	0	0	0	0	0	1.049212	1.176689	220	5	1.1	0.9;
	2494	1	0	0	0	0	0	1.013245	0.963784	150	8	1.1	0.9;
	2496	1	50.56	4.6	0	1.74	0	1.016225	-60.163857	220	4	1.1	0.9;
	2497	1	0	0	0	9.36	0	0.981207	-55.903537	220	10	1.1	0.9;
	2503	1	-0	0	0	2.09	0	1.012625	-21.505114	220	5	1.1	0.9;
	2504	1	119.2	32.1	0	4.14	0	1.009312	-45.199119	220	4	1.1	0.9;
	2507	1	-0	0	0	36.17	0	1.066743	-41.368798	380	4	1.1	0.9;
	2510	1	115.87	29.41	0	7.92	0	1.02733	-12.091763	220	5	1.1	0.9;
	2517	2	0	0	0	0	0	1.017237	-44.572173	220	4	1.1	0.9;
	2518	1	-5.29	-1.91	0	42.622979	0	0.998925	-11.524637	380	5	1.1	0.9;
	2520	1	-0	-0	0.057433	3.80797	0	0.997433	-4.735085	380	2	1.1	0.9;
	2523	1	0	0	0	7.06	0	1.045749	26.39537	220	8	1.1	0.9;
	2526	1	238.45	46.86	0	125.534496	0	1.042721	-19.883704	380	5	1.1	0.9;
	2527	1	59.77	16.7	0	0.64	0	1.039168	-4.615105	220	5	1.1	0.9;
	2535	1	51.56	2.1	0	0.18	0	1.041259	-2.946934	220	5	1.1	0.9;
	2538	1	118.8	39.5	0	2.61	0	1.017121	-43.487032	220	4	1.1	0.9;
	2549	1	0	-0	0	3.86	0	1.049693	-3.110388	220	1	1.1	0.9;
	2550	2	0	0	0	0	0	1.050935	-6.648649	220	5	1.1	0.9;
	2551	1	205.3	27.8	0	8.22	0	1.011995	-60.510986	220	4	1.1	0.9;
	2552	1	28.53	11.3	0	6.81	0	1.036643	4.553272	150	8	1.1	0.9;
	2555	1	-0.15	-1	0	0.38	0	1.056627	12.502131	110	8	1.1	0.9;
	2558	1	159.27	38.31	0	2.18	0	1.024168	-16.655732	220	5	1.1	0.9;
	2562	1	-88.6	-24.2	0	5.66	0	1.003073	-51.388077	220	10	1.1	0.9;
	2563	1	52.86	13	0	6.04	0	1.042342	-4.677762	220	5	1.1	0.9;
	2564	1	0	0	0	0.93	0	1.031089	-53.04439	220	4	1.1	0.9;
	2566	1	0	-0	0	68.93	0	1.040492	-44.635228	380	4	1.1	0.9;
	2572	2	0	0	0	0	0	1.044905	17.443246	150	8	1.1	0.9;
	2575	1	62.97	17.4	0	0.3	0	1.033526	-2.681692	220	5	1.1	0.9;
	2591	1	253.77	84.6	0	20.71	0	1.031532	-15.253656	220	5	1.1	0.9;
	2597	1	14.72	5.4	0	0.72	0	1.035529	-10.83031	220	5	1.1	0.9;
	2598	1	25.63	12.7	0	8.42	0	1.040309	-15.280276	220	5	1.1	0.9;
	2600	2	0	0	0	0	0	1.041924	-13.574935	220	5	1.1	0.9;
	2616	1	-0	0	0	1.46	0	1.035313	5.493092	150	8	1.1	0.9;
	2619	1	0	0	0	1.01	0	1.016786	6.309895	150	8	1.1	0.9;
	2622	1	-0	0	0	6.91	0	1.04898	-49.357142	380	4	1.1	0.9;
	2627	2	0	0	0	0	0	1.052984	7.725147	380	5	1.1	0.9;
	2629	1	203.1	65.2	0	0.64	0	1.043972	-24.533846	220	5	1.1	0.9;
	2641	1	0	0	0	1.72	0	1.043147	-11.261777	220	5	1.1	0.9;
	2642	2	0	0	0	0	0	1.054923	-36.938161	380	4	1.1	0.9;
	2644	1	0	0	0	0.19	0	1.073301	-24.836211	220	5	1.1	0.9;
	2648	1	46.65	15.1	0	9.79	0	1.048915	11.107878	110	8	1.1	0.9;
	2653	2	0	0	0	0	0	1.036409	-23.037279	220	5	1.1	0.9;
	2654	1	207.1	50.4	0	4.93	0	1.01876	-18.768622	220	5	1.1	0.9;
	2656	1	-0.72	0	0	69.582294	0	1.032684	-22.83602	380	5	1.1	0.9;
	2658	2	0	0	0	0	0	1.026597	-48.776737	220	4	1.1	0.9;
	2660	2	0	0	0	0	0	1.013897	-31.381581	150	10	1.1	0.9;
	2667	1	70.68	10.7	0	1.24	0	1.033098	-52.916082	220	4	1.1	0.9;
	2668	1	252.9	95.6	0	1.22	0	1.021145	-43.98353	220	4	1.1	0.9;
	2676	1	123.24	23.65	0	6.44	0	1.0367	-14.939386	220	5	1.1	0.9;
	2686	2	0	0	0	0	0	1.020441	-42.467961	220	4	1.1	0.9;
	2689	1	0	0	0	0.45	0	1.03773	-3.261785	220	5	1.1	0.9;
	2692	1	285.2	65.3	0	11.28	0	0.989518	-54.065312	220	10	1.1	0.9;
	2695	1	16.22	8.9	0	2.2	0	1.037346	-21.040609	220	5	1.1	0.9;
	2698	1	0	-1.2	0	0.41	0	1.03424	25.891078	110	8	1.1	0.9;
	2700	1	-0	0	0	3.08	0	1.035125	-38.951268	380	4	1.1	0.9;
	2702	1	151.7	27.6	0	6.49	0	1.041685	-25.654435	220	5	1.1	0.9;
	2703	1	-0.2	-4.97	0	0	0	1.033424	9.303066	150	8	1.1	0.9;
	2708	1	47.75	18.8	0	4.42	0	1.015434	-0.677624	150	8	1.1	0.9;
	2719	2	0	0	0	0	0	1.039499	-15.429073	220	5	1.1	0.9;
	2720	1	13.01	14.2	0	15.78	0	1.017103	17.754983	150	8	1.1	0.9;
	2721	1	-0	0	0	10.040547	0	0.978984	-22.513554	380	5	1.1	0.9;
	2722	1	0	0	0	56.39	0	1.016299	8.152212	380	8	1.1	0.9;
	2728	1	-0.13	-1	0	0	0	1.04666	26.481466	220	8	1.1	0.9;
	2730	1	0.1	152.7	0	233.41	0	1.045043	-43.518515	380	4	1.1	0.9;
	2732	1	85.59	20.72	0	142.79	0	1.008534	-10.041961	380	5	1.1	0.9;
	2734	1	-61.67	-0.9	0	12.33	0	0.995157	-47.721211	150	10	1.1	0.9;
	2740	1	36.24	5.7	0	15.06	0	1.029789	-53.804959	220	4	1.1	0.9;
	2745	1	0	-0	0	56.27	0	1.007399	-3.385125	150	8	1.1	0.9;
	2748	1	65.47	26.7	0	18.6	0	1.003257	-3.866665	150	8	1.1	0.9;
	2749	1	55.56	26.5	0	1.09	0	1.044722	-30.340412	220	5	1.1	0.9;
	2751	1	34.54	8.6	0	5.05	0	1.041213	-21.928696	220	5	1.1	0.9;
	2754	1	-0	0	0	3.72	0	1.049456	-13.383761	220	5	1.1	0.9;
	2760	1	64.87	25.6	0	14.89	0	1.017866	4.65043	150	8	1.1	0.9;
	2764	2	0	0	0	0	0	1.05096	-48.880173	380	4	1.1	0.9;
	2769	1	98.6	28.3	0	4.76	0	1.027942	26.266559	220	8	1.1	0.9;
	2770	1	85.39	14.3	0	0.35	0	1.049338	-9.678721	220	5	1.1	0.9;
	2782	1	-28.13	-0.81	0	0.1	0	1.063907	-11.500188	220	5	1.1	0.9;
	2785	1	189.4	25.8	0	3.62	0	0.983777	-49.755212	220	4	1.1	0.9;
	2786	2	0	0	0	0	0	1.069288	6.564216	380	5	1.3	0.7;
	2794	1	0	-0	0	0.39	0	1.032808	-18.511406	220	5	1.1	0.9;
	2795	1	-0	-0	0	1.94	0	1.068855	-32.501298	220	5	1.1	0.9;
	2796	1	100.7	17.5	0	6.54	0	1.032852	9.264863	150	8	1.1	0.9;
	2797	2	0	0	0	0	0	1.044662	-4.476602	220	5	1.1	0.9;
	2799	2	0	0	0	0	0	1.055463	-12.420614	220	5	1.1	0.9;
	2800	2	0	0	0	0	0	1.030206	-48.37868	380	4	1.1	0.9;
	2801	1	0	-0	0	12.17	0	1.040225	-24.54678	220	5	1.1	0.9;
	2804	1	31.33	9.6	0	1.86	0	1.026472	12.978183	110	8	1.1	0.9;
	2806	1	-0	-0	0	0.39	0	1.050487	0.178503	220	5	1.1	0.9;
	2812	1	97	18.6	0	1.25	0	1.012846	16.975428	150	8	1.1	0.9;
	2815	1	25.33	14.2	0	4.05	0	1.036427	-23.013761	220	5	1.1	0.9;
	2816	2	0	0	0	0	0	1.054536	-0.415457	380	5	1.1	0.9;
	2817	2	0	0	0	0	0	0.991705	-46.059108	150	10	1.1	0.9;
	2824	1	-0	-0	0	4.17	0	1.029264	17.813781	220	8	1.1	0.9;
	2832	1	48.15	14.8	0	2.16	0	1.019303	12.490936	110	8	1.1	0.9;
	2841	2	0	0	0	0	0	1.079986	-26.265078	220	5	1.1	0.9;
	2842	2	0	0	0	0	0	1.044122	-6.664319	220	5	1.1	0.9;
	2844	1	179	50.5	0	3.08	0	1.011346	-45.900522	220	4	1.1	0.9;
	2847	1	40.14	8.8	0	0.82	0	1.046169	-3.919284	220	5	1.1	0.9;
	2848	1	453.23	139.16	0	25.583218	0	0.998548	-7.849319	380	5	1.1	0.9;
	2849	1	0	0	0	70.11	0	1.040792	-42.513227	380	4	1.1	0.9;
	2850	1	-0	-0	0	4.76	0	1.030335	-18.770897	220	5	1.1	0.9;
	2853	2	0	0	0	0	0	1.000604	-35.710618	150	10	1.1	0.9;
	2854	1	-12.33	-2.65	0	69.323028	0	1.046183	-10.741328	380	5	1.1	0.9;
	2863	1	37.73	13.84	0	0.09	0	1.031964	-11.04501	220	5	1.1	0.9;
	2866	1	-0	0	0	4.21	0	1.038006	-7.282246	220	5	1.1	0.9;
	2870	1	-0	0	0	0.62	0	1.019244	5.129193	380	2	1.1	0.9;
	2872	2	0	0	0	0	0	1.060615	-10.673863	220	5	1.1	0.9;
	2877	1	29.93	-0	0	0.46	0	1.041957	-14.91571	220	5	1.1	0.9;
	2878	2	0	0	0	0	0	1.05418	-10.996458	220	5	1.1	0.9;
	2886	2	0	0	0	0	0	1.052878	-14.148238	220	5	1.1	0.9;
	2888	1	-0	0	0	0.22	0	1.013491	-21.281344	220	5	1.1	0.9;
	2889	1	-0	0	0	3.2	0	1.051468	-1.112479	220	5	1.1	0.9;
	2894	1	-0	0	0	0.64	0	1.001352	-38.531811	150	10	1.1	0.9;
	2896	2	0	0	0	0	0	1.011629	-44.09682	220	4	1.1	0.9;
	2898	1	56.76	13.4	0	1.18	0	1.035736	-23.099879	220	5	1.1	0.9;
	2899	1	-0	-0	0	44.66	0	1.055839	-44.433529	380	10	1.1	0.9;
	2900	1	0	0	0	0	0	1.035777	4.260134	150	8	1.1	0.9;
	2902	2	0	0	0	0	0	1.061022	-23.402903	220	5	1.1	0.9;
	2907	1	241	32.3	0	34.74	0	1.023741	-48.101415	220	4	1.1	0.9;
	2908	1	203.8	61	0	0.6	0	1.026635	-10.516686	150	2	1.1	0.9;
	2910	1	-0	0	0	2.87	0	1.044227	-2.315624	220	5	1.1	0.9;
	2913	1	51.96	9.4	0	42.51	0	1.065675	-41.45215	380	4	1.1	0.9;
	2916	1	63.47	30.2	0	81.36	0	1.049753	-37.87224	380	4	1.1	0.9;
	2917	1	31.64	7.5	0	21.91	0	1.023967	-46.506929	220	4	1.1	0.9;
	2918	1	0	0	0	53.25003	0	1.019273	-28.176889	380	5	1.1	0.9;
	2919	1	-0	0	0	32.049116	0	0.991102	-18.332113	380	5	1.1	0.9;
	2921	1	37.24	7.1	0	1.07	0	1.008526	-40.282483	220	4	1.1	0.9;
	2924	1	19.92	7.4	0	17.28	0	1.040977	-33.52648	220	5	1.1	0.9;
	2927	1	-0	0	0	26.18	0	1.049719	18.392275	380	8	1.1	0.9;
	2928	1	0	-0	0	92.38	0	1.07417	-26.364596	220	5	1.1	0.9;
	2930	2	0	0	0	0	0	1.049271	-12.977156	220	5	1.1	0.9;
	2931	1	230.7	19.3	0	4.56	0	1.042827	-32.310096	220	5	1.1	0.9;
	2934	2	0	0	0	0	0	1.052968	-6.51468	220	5	1.1	0.9;
	2938	1	0	-0	0	137.29	0	1.006614	-7.445421	380	5	1.1	0.9;
	2940	1	103	28.7	0	2.88	0	1.011513	-14.38638	220	5	1.1	0.9;
	2946	2	0	0	0	0	0	1.020023	-45.610776	220	4	1.1	0.9;
	2949	1	-0	-0	0	11.03	0	1.049798	-4.843132	220	5	1.1	0.9;
	2951	1	104.6	40.7	0	1.42	0	1.015286	-53.686825	220	4	1.1	0.9;
	2954	1	4.81	1.2	0	0.13	0	0.98995	-55.678384	150	10	1.1	0.9;
	2957	1	81.19	24.4	0	0.6	0	1.027698	-38.781315	220	4	1.1	0.9;
	2961	1	159.9	53.2	0	7.99	0	1.040418	-24.431519	220	5	1.1	0.9;
	2962	1	93.5	42.5	0	133.56	0	1.067192	-42.399748	380	4	1.1	0.9;
	2965	1	-29.96	-21.31	0	0	0	1.030784	18.259396	150	8	1.1	0.9;
	2967	1	0	-0	0	0.19	0	1.045595	-17.104	220	5	1.1	0.9;
	2968	1	112.51	16.51	0	1.63	0	1.045486	-8.969477	220	5	1.1	0.9;
	2971	1	0	0	0	11.55	0	0.993315	-6.019917	380	1	1.1	0.9;
	2972	1	97.82	20.59	0	6.67	0	1.035382	-15.090981	220	5	1.1	0.9;
	2974	1	65.47	26.7	0	0.6	0	1.008087	-2.426628	150	8	1.1	0.9;
	2978	1	0	0	0	2.47	0	1.024856	-48.16715	220	4	1.1	0.9;
	2980	1	-0	0	0	0.42	0	1.062898	-9.075961	380	5	1.1	0.9;
	2981	1	-0	0	0	1.68	0	1.07476	-15.345009	220	5	1.1	0.9;
	2985	2	0	0	0	0	0	1.05077	-8.481817	220	5	1.1	0.9;
	2995	1	0	-0	0	3.9	0	1.047304	-3.731029	220	5	1.1	0.9;
	3000	1	139.6	68.4	0	0	0	1.017706	15.028535	150	8	1.1	0.9;
	3007	1	108.8	33.2	0.358716	45.62985	0	1.031803	-51.698634	380	10	1.1	0.9;
	3008	1	0	0	0	22.7	0	1.021743	-45.9053	380	4	1.1	0.9;
	3013	1	2.2	1.1	0	0.67	0	1.04929	-5.249851	220	5	1.1	0.9;
	3018	2	0	0	0	0	0	1.043371	-12.865081	220	5	1.1	0.9;
	3019	1	0	-0	0	1.37	0	1.036683	-22.79236	220	5	1.1	0.9;
	3021	1	74.38	8.8	0	6.1	0	1.043451	-11.465032	220	5	1.1	0.9;
	3022	1	149	19.8	0	23.89	0	1.038179	-14.947144	380	5	1.1	0.9;
	3023	1	0	-0	0	27.69	0	0.997435	8.718509	380	8	1.1	0.9;
	3027	1	0	0	0	1.54	0	1.030907	-53.059141	220	4	1.1	0.9;
	3028	2	0	0	0	0	0	1.034318	0.53384	380	5	1.1	0.9;
	3036	1	116.93	33.6	0	3.35	0	1.045271	-25.21954	220	5	1.1	0.9;
	3037	1	221	146.8	0	8.35	0	1.050303	-6.606462	220	5	1.1	0.9;
	3044	1	29.33	5.6	0	6.56	0	1.039563	-21.150198	220	5	1.1	0.9;
	3047	1	0	-0	0	3.84	0	1.01916	2.829744	150	8	1.1	0.9;
	3051	1	0	0	0	0	0	1.032503	-24.998397	220	5	1.1	0.9;
	3054	2	0	0	0	0	0	0.999852	-51.274579	220	10	1.1	0.9;
	3061	1	493.4	53.5	0	58.09	0	1.035223	-47.778687	380	4	1.1	0.9;
	3063	1	71.28	6.9	0	1.93	0	1.015747	-45.233586	220	4	1.1	0.9;
	3065	1	55.86	8.8	0	4.2	0	1.027992	-54.17571	220	4	1.1	0.9;
	3066	1	119.7	3.7	0	0.88	0	1.028937	-53.939129	220	4	1.1	0.9;
	3067	1	272.3	58.5	0	23.57	0	1.027656	-41.914357	380	4	1.1	0.9;
	3069	1	89.6	22.7	0	4.06	0	1.035723	-8.751882	220	5	1.1	0.9;
	3070	1	0	-0	0	4.27	0	1.013485	-21.281298	220	5	1.1	0.9;
	3071	1	0	-0	0	2.83	0	1.04631	-7.264593	220	5	1.1	0.9;
	3072	1	92.9	27.7	0	5.22	0	1.045476	-30.312476	220	5	1.1	0.9;
	3075	1	30.33	10.5	0	0.33	0	1.019161	-16.90632	220	5	1.1	0.9;
	3076	1	164	42.3	0	8.07	0	1.014367	-47.584034	220	4	1.1	0.9;
	3080	1	-0	-0	0.256503	101.229042	0	1.064383	-47.252271	380	10	1.1	0.9;
	3082	1	0	0	0	75.076038	0	0.997398	-15.135632	380	5	1.1	0.9;
	3083	1	567.97	130.73	0	1.09	0	1.067624	-27.680721	220	5	1.1	0.9;
	3085	1	64.67	24.5	0	1.46	0	1.066452	-27.737369	220	5	1.1	0.9;
	3095	1	69.78	15.3	0	1.34	0	1.013045	-53.928694	220	4	1.1	0.9;
	3097	1	128.11	64.9	0	19.1	0	1.0203	4.509568	380	2	1.1	0.9;
	3106	1	30.63	11.6	0	52.28	0	1.053068	-45.074265	380	4	1.1	0.9;
	3109	1	-0	0	0	8.33	0	1.0195	-42.362846	220	4	1.1	0.9;
	3111	1	0	0	0	20.96	0	1.033059	-42.130088	220	4	1.1	0.9;
	3112	1	-0	0	0	0	0	1.05568	2.418195	220	5	1.1	0.9;
	3113	2	0	0	0	0	0	1.072644	-37.775694	380	4	1.1	0.9;
	3114	2	0	0	0	0	0	1.0681	-11.639685	220	5	1.1	0.9;
	3119	1	99.9	20.1	0	17.41	0	1.024432	9.444867	150	8	1.1	0.9;
	3121	1	242.16	73.64	0	12.79	0	1.030975	-7.223321	220	5	1.1	0.9;
	3125	1	-0	0	0	3.25	0	1.01835	-39.759339	220	4	1.1	0.9;
	3126	1	104.78	20.92	0	6.17	0	1.022207	-17.916791	220	5	1.1	0.9;
	3133	2	0	0	0	0	0	1.061763	-9.179778	220	5	1.1	0.9;
	3134	2	0	0	0	0	0	1.02503	-15.762659	220	5	1.1	0.9;
	3137	1	0	-0	0	0.03	0	1.059334	-26.086008	220	5	1.1	0.9;
	3145	1	-0	-0	0	24.909116	0	0.990925	-41.332117	380	5	1.1	0.9;
	3148	1	0	0	0	0	0	1.032522	3.857451	150	8	1.1	0.9;
	3150	2	0	0	0	0	0	1.017094	-47.936998	220	4	1.1	0.9;
	3157	2	0	0	0	0	0	1.023658	-37.534265	220	4	1.1	0.9;
	3163	2	0	0	0	0	0	1.006379	-47.142629	220	4	1.1	0.9;
	3166	1	184.91	32.8	0	8.35	0	1.039311	-17.540223	220	5	1.1	0.9;
	3169	1	0	0	0	75	0	1.033678	4.168232	150	8	1.1	0.9;
	3177	1	39.34	11.2	0	0.85	0	1.018336	7.592426	150	8	1.1	0.9;
	3178	1	180.2	46.4	0	9.7	0	1.041131	-45.048845	220	4	1.1	0.9;
	3182	1	31.94	12.6	0	10.67	0	1.024246	4.396981	150	8	1.1	0.9;
	3183	2	0	0	0	0	0	1.056991	-11.898771	220	5	1.1	0.9;
	3184	1	14.12	7.7	0	1.42	0	1.038993	-21.374036	220	5	1.1	0.9;
	3186	1	-20.92	2.5	0	16.42	0	1.019225	-33.185258	220	10	1.1	0.9;
	3187	1	107	34.7	0	1.54	0	1.039446	-12.967806	220	5	1.1	0.9;
	3190	1	0	-0	0	35.14	0	1.03268	13.864056	380	8	1.1	0.9;
	3191	1	127.8	15.2	0	9.34	0	1.041406	-12.076173	220	5	1.1	0.9;
	3193	1	0	0	0	207.85	0	1.070508	-43.086559	380	4	1.1	0.9;
	3200	1	86.2	19.9	0	5.43	0	1.03606	-23.08282	220	5	1.1	0.9;
	3204	1	8.51	3.2	0	0.96	0	1.01902	-10.232717	220	5	1.1	0.9;
	3205	2	0	0	0	0	0	1.03763	-1.224853	380	5	1.1	0.9;
	3206	1	146.7	20.7	0	62.64	0	1.034144	10.259726	150	8	1.1	0.9;
	3210	1	-0	-0	0	47.99	0	1.009835	6.952975	380	8	1.1	0.9;
	3211	1	0	0	0	5.21	0	1.007781	-44.247841	220	10	1.1	0.9;
	3215	1	63.07	25.8	0	18.63	0	1.002866	-3.944884	150	8	1.1	0.9;
	3216	1	73.58	30.1	0	6.02	0	1.001722	-4.072981	150	8	1.1	0.9;
	3218	2	0	0	0	0	0	1.032146	-11.223745	220	5	1.1	0.9;
	3221	1	76.38	24.8	0	3.24	0	1.044182	-12.158598	220	5	1.1	0.9;
	3224	1	0	-0	0	1.86	0	1.049779	28.300405	110	8	1.1	0.9;
	3231	1	251.3	42.6	0	5.47	0	1.025874	-14.296903	220	5	1.1	0.9;
	3233	2	0	0	0	0	0	1.012423	-45.745678	220	4	1.1	0.9;
	3239	1	-5.06	-1.84	0	187.327025	0	1.055998	-11.619498	380	5	1.1	0.9;
	3240	2	0	0	0	0	0	1.051571	-1.124291	220	5	1.1	0.9;
	3241	1	132	27	0	4.66	0	1.03022	-17.948598	220	5	1.1	0.9;
	3242	1	296.1	82.3	0	0	0	1.059624	-9.310952	150	2	1.1	0.9;
	3243	1	-0	0	0	1.18	0	0.990149	-18.204781	380	5	1.1	0.9;
	3246	1	20.9	-171.5	0	0.21	0	1.078316	7.898222	220	5	1.1	0.9;
	3248	1	84.19	17.3	0	6.2	0	1.05675	-11.335696	220	5	1.1	0.9;
	3255	1	0	0	0	1.44	0	1.038641	-22.064617	220	5	1.1	0.9;
	3256	1	-0	0	0	1.87	0	1.041573	-51.256261	380	10	1.1	0.9;
	3257	1	0	-0	0	3.71	0	1.045223	13.616401	220	8	1.1	0.9;
	3265	1	0	-0	0	61.52	0	1.035215	-51.062654	380	10	1.1	0.9;
	3266	1	-0	0	0	4.62	0	0.989841	-53.987659	220	10	1.1	0.9;
	3268	1	-0	-0	0	29.27	0	1.046344	-39.42253	380	4	1.1	0.9;
	3275	1	-0	0	0	17.75	0	1.046626	-42.734675	380	4	1.1	0.9;
	3276	1	-0	-0	0	0.13	0	1.052405	-7.871112	220	5	1.1	0.9;
	3279	1	0	0	0	0	0	1.004923	-2.703149	150	2	1.1	0.9;
	3280	1	-0	0	0	2.08	0	1.006907	-3.391392	150	8	1.1	0.9;
	3282	1	-0	0	0	9.66	0	1.021422	-56.880436	220	4	1.1	0.9;
	3291	2	0	0	0	0	0	1.027251	-44.134036	220	4	1.1	0.9;
	3293	1	33.14	11.6	0	4.51	0	1.039346	-11.406773	220	5	1.1	0.9;
	3299	1	332.9	99.6	0	0.53	0	1.018099	7.959711	150	8	1.1	0.9;
	3301	1	72.08	29.5	0	1.06	0	1.006897	-3.141399	150	8	1.1	0.9;
	3306	2	0	0	0	0	0	1.069265	8.119243	220	5	1.1	0.9;
	3314	1	-0.2	-1.64	0	0	0	1.031936	4.19826	150	8	1.1	0.9;
	3324	2	0	0	0	0	0	1.016523	-16.112897	220	5	1.1	0.9;
	3325	1	-0	0	0	0.91	0	1.033352	-20.083581	220	5	1.1	0.9;
	3327	1	-0	-0	0	2.27	0	1.034449	-12.168575	220	5	1.1	0.9;
	3331	1	-0	-0	0	4.56	0	1.019496	-10.181699	220	5	1.1	0.9;
	3334	2	0	0	0	0	0	1.03247	-38.167816	220	4	1.1	0.9;
	3335	2	0	0	0	0	0	1.08873	32.451722	380	8	1.1	0.9;
	3341	1	46.85	16.6	0	2.08	0	0.996579	-43.046702	150	10	1.1	0.9;
	3342	1	76.18	13.2	0	7.4	0	1.030961	-39.881222	220	4	1.1	0.9;
	3344	1	83.29	24.1	0	1.23	0	1.037547	-16.626086	220	5	1.1	0.9;
	3346	2	0	0	0	0	0	1.014746	-12.050881	380	5	1.1	0.9;
	3352	1	-0	-0	0	5.22	0	1.02469	11.939007	380	8	1.1	0.9;
	3353	2	0	0	0	0	0	1.047796	-3.038965	220	5	1.1	0.9;
	3363	1	155.8	38.4	0	3.77	0	1.015585	1.898178	150	8	1.1	0.9;
	3364	2	0	0	0	0	0	1.055303	-12.424509	220	5	1.1	0.9;
	3367	1	57.46	15.3	0	4.37	0	1.023439	-52.090148	220	4	1.1	0.9;
	3370	1	-0	-0	0	0	0	1.049776	18.741809	380	8	1.1	0.9;
	3374	1	156.2	24.5	0	2.22	0	1.019323	10.42872	150	8	1.1	0.9;
	3377	1	77.49	21.4	0	0.05	0	1.019707	-10.476462	220	5	1.1	0.9;
	3378	1	-0	-0	0	13.02	0	1.023847	-44.818138	220	4	1.1	0.9;
	3382	1	115	-4.5	0	15.05	0	1.024501	-46.594868	220	4	1.1	0.9;
	3388	1	-0.2	-1.64	0	0	0	1.031936	4.19826	150	8	1.1	0.9;
	3390	2	0	0	0	0	0	1.041678	-11.508701	220	5	1.1	0.9;
	3391	1	302.5	65.1	0	8.87	0	1.058582	-20.966447	220	5	1.1	0.9;
	3397	1	24.12	9.8	0	0.04	0	1.001327	-0.497618	150	8	1.1	0.9;
	3400	1	0	0	0	17.31	0	1.046036	-25.128412	220	5	1.1	0.9;
	3401	1	23.22	6.3	0	0.66	0	1.037884	-14.052358	220	5	1.1	0.9;
	3409	1	-0	-0	0	16.4	0	1.02461	-38.44707	380	1	1.1	0.9;
	3412	1	-0	0	0	0	0	1.05568	2.418195	220	5	1.1	0.9;
	3413	1	19.42	7	0	0.6	0	0.990325	-48.145285	150	10	1.1	0.9;
	3416	2	0	0	0	0	0	0.997085	-36.650296	220	10	1.1	0.9;
	3422	2	0	0	0	0	0	1.065868	-10.780991	220	5	1.1	0.9;
	3425	1	83.59	21.6	0	25.51	0	1.029115	-53.61758	220	4	1.1	0.9;
	3427	1	146.9	26.6	0	11.95	0	1.029707	-53.426186	220	4	1.1	0.9;
	3428	1	-0	-0	0	17.92	0	1.039945	-44.378925	380	4	1.1	0.9;
	3429	1	175.4	28.8	0	2.03	0	0.985017	-56.281675	150	10	1.1	0.9;
	3430	1	125.1	44	0	6.04	0	1.01706	-16.096952	220	5	1.1	0.9;
	3435	1	-0	0	0	10.13	0	1.04207	-25.600219	220	5	1.1	0.9;
	3436	2	0	0	0	0	0	1.056276	-13.341208	220	5	1.1	0.9;
	3442	1	54.96	12.2	0	7.05	0	1.033756	-53.058588	220	4	1.1	0.9;
	3443	1	-0	0	0	0.74	0	1.018501	6.910518	150	8	1.1	0.9;
	3445	1	-7.63	-4.3	0	94.00001	0	0.998626	-5.886772	380	5	1.1	0.9;
	3450	1	78.99	16.2	0	1.97	0	1.045917	-21.516053	220	5	1.1	0.9;
	3456	1	-0	0	0	2.09	0	1.007014	-3.377444	150	8	1.1	0.9;
	3462	1	209.5	95	0	24.39	0	1.024888	11.675038	150	8	1.1	0.9;
	3468	1	50.36	-36.2	0	1.25	0	1.029162	-47.376681	220	4	1.1	0.9;
	3478	1	-3.4	-0.5	0	0.23	0	1.007823	-44.241987	220	10	1.1	0.9;
	3481	1	142.5	23.6	0	4.24	0	1.021369	-20.827525	220	5	1.1	0.9;
	3483	1	157.78	276.03	0	91.192543	0	1.047991	2.353608	380	5	1.1	0.9;
	3484	1	-0	0	0	37.1	0	1.032637	13.843519	380	8	1.1	0.9;
	3485	1	-0	0	0	6.501668	0	0.990338	-16.290207	380	5	1.1	0.9;
	3486	1	101.5	-18.3	0	8.47	0	1.035245	-20.908358	220	5	1.1	0.9;
	3487	1	16.32	7.6	0	3.15	0	1.01071	-39.922801	220	4	1.1	0.9;
	3488	1	55.66	4.5	0	1.33	0	1.039592	-12.379315	220	5	1.1	0.9;
	3489	1	0	-0	0	12.24	0	1.027242	-52.988086	220	4	1.1	0.9;
	3492	2	0	0	0	0	0	1.04519	-8.569839	220	5	1.1	0.9;
	3493	1	226.3	61.1	0	0	0	1.035682	-10.395331	150	2	1.1	0.9;
	3495	1	36.34	3.3	0	2.16	0	1.018242	-57.061111	220	4	1.1	0.9;
	3496	1	99.3	23.3	0	1.88	0	1.03547	-21.365036	220	5	1.1	0.9;
	3497	1	100.5	32	0	1.89	0	1.024686	4.074269	150	8	1.1	0.9;
	3498	1	138.5	19.3	0	5.26	0	1.035615	-10.18023	220	5	1.1	0.9;
	3499	1	180.4	71.6	0	81.042938	0	1.036881	-3.624207	380	5	1.1	0.9;
	3500	1	94.2	21.6	0	2.88	0	1.032595	-52.148697	220	4	1.1	0.9;
	3502	1	98.2	19.1	0	10.92	0	1.022468	-10.518494	220	5	1.1	0.9;
	3503	1	-0	0	0	1.93	0	1.03107	-11.513013	220	5	1.1	0.9;
	3506	1	0	0	0	0.6	0	1.010703	-10.97175	150	2	1.1	0.9;
	3509	1	23.62	16.5	0	8.39	0	1.023232	11.672068	150	8	1.1	0.9;
	3512	1	65.64	-0.98	0	11.55	0	1.011233	-52.616595	220	4	1.1	0.9;
	3513	2	0	0	0	0	0	1.029628	-3.769006	380	5	1.1	0.9;
	3519	1	-0	0	0	68.053762	0	1.032358	-22.871014	380	5	1.1	0.9;
	3520	1	-0	0	0	2.47	0	1.04126	-5.929212	220	5	1.1	0.9;
	3526	1	33.34	13.2	0	0.76	0	1.060598	-4.169707	220	5	1.1	0.9;
	3530	1	118.5	21	0	2.19	0	1.01013	0.854718	150	8	1.1	0.9;
	3535	1	-0	0	0	0.49	0	1.035222	-24.481534	220	5	1.1	0.9;
	3538	1	0	0	0	4.17	0	1.027519	18.538871	220	8	1.1	0.9;
	3541	1	0	-0	0	3.34	0	1.038631	-24.013582	220	5	1.1	0.9;
	3543	1	-0	-0	0	14.079026	0	0.991231	-17.577149	380	5	1.1	0.9;
	3544	1	0	-0	0	83.96	0	1.051823	-43.291508	380	4	1.1	0.9;
	3545	1	20.12	7.6	0	0.89	0	1.051526	-7.986739	220	5	1.1	0.9;
	3549	1	0	0	0	3.91	0	1.017723	-42.242961	220	4	1.1	0.9;
	3557	1	87.2	31.8	0	5.96	0	1.043736	-30.66359	220	5	1.1	0.9;
	3558	1	-0	0	0	2.51	0	1.061451	2.478205	220	5	1.1	0.9;
	3565	2	0	0	0	0	0	1.079121	-26.473248	220	5	1.1	0.9;
	3577	1	-38.7	-12.99	0	13.47383	0	1.032186	-5.20661	380	5	1.1	0.9;
	3579	1	277.72	7.35	0	3.58	0	1.071503	-39.941643	220	5	1.1	0.9;
	3580	2	0	0	0	0	0	1.075232	-33.466802	220	5	1.1	0.9;
	3582	2	0	0	0	0	0	1.045127	-48.989672	380	4	1.1	0.9;
	3584	2	0	0	0	0	0	1.019651	-44.933373	220	4	1.1	0.9;
	3585	2	0	0	0	0	0	1.006097	-44.462006	220	10	1.1	0.9;
	3587	1	0	-0	0	78.22	0	1.069089	-39.84538	380	4	1.1	0.9;
	3589	1	37.54	18.3	0	0.22	0	1.043343	-6.742329	220	5	1.1	0.9;
	3592	1	102.7	77.5	0	4.89	0	1.01555	-48.994509	220	4	1.1	0.9;
	3594	1	31.43	6.6	0	3.32	0	1.031857	-19.837353	220	5	1.1	0.9;
	3601	1	0	0	0	0.23	0	1.039102	-21.39569	220	5	1.1	0.9;
	3602	1	86.6	14.9	0	1.34	0	1.029873	-18.964113	220	5	1.1	0.9;
	3608	1	0	-0	0	58.146813	0	1.00864	-41.723719	380	5	1.1	0.9;
	3609	1	-28.13	-0.82	0	0.1	0	1.063902	-11.501477	220	5	1.1	0.9;
	3610	1	185.8	31.8	0	3.73	0	1.036366	-22.773158	220	5	1.1	0.9;
	3611	2	0	0	0	0	0	1.03827	-18.990532	220	5	1.1	0.9;
	3613	1	152.5	14.4	0	24.61	0	1.029169	-28.094604	380	5	1.1	0.9;
	3615	1	0	-0	0	13.51	0	1.02479	-38.25963	380	1	1.1	0.9;
	3620	1	35.14	16.1	0	0.45	0	1.028924	2.181053	150	8	1.1	0.9;
	3621	2	0	0	0	0	0	1.023915	-51.863704	220	4	1.1	0.9;
	3623	1	0	-0	0	4.08	0	1.032919	-33.405215	220	10	1.1	0.9;
	3629	2	0	0	0	0	0	1.020807	-43.909237	220	4	1.1	0.9;
	3633	1	-0	0	0	15.23	0	1.034754	-54.841716	220	4	1.1	0.9;
	3643	1	24.23	6.4	0	3.9	0	1.032306	-19.396739	220	5	1.1	0.9;
	3644	2	0	0	0	0	0	1.018948	-43.146574	220	4	1.1	0.9;
	3645	1	42.55	17.6	0	17.28	0	1.040651	-33.549688	220	5	1.1	0.9;
	3649	1	-0	-0	0	4.86	0	1.0491	-36.643877	220	5	1.1	0.9;
	3650	1	0	-0	0	4.68	0	1.027803	-43.004237	220	4	1.1	0.9;
	3654	1	20.72	5	0	6.64	0	1.073757	-26.422172	220	5	1.1	0.9;
	3656	2	0	0	0	0	0	1.067132	-11.788937	220	5	1.1	0.9;
	3657	1	131	1.3	0	120.45	0	1.075197	-26.111324	220	5	1.1	0.9;
	3659	2	0	0	0	0	0	1.065964	1.067647	380	2	1.1	0.9;
	3661	2	0	0	0	0	0	1.059844	4.519815	380	5	1.1	0.9;
	3662	1	-0	-0	0	4.74	0	1.025068	-43.262157	380	4	1.1	0.9;
	3664	1	-0.2	-0.96	0	0	0	1.021032	8.869635	150	8	1.1	0.9;
	3665	1	295.4	42	0	0.83	0	1.015517	-42.634054	220	4	1.1	0.9;
	3670	1	28.03	6.5	0	6.32	0	1.035567	-22.00727	220	5	1.1	0.9;
	3672	1	82.99	25.9	0	0.23	0	1.033523	-12.87972	220	5	1.1	0.9;
	3673	2	0	0	0	0	0	1.013624	-38.75114	220	4	1.1	0.9;
	3674	1	155.05	39.56	0	6.27	0	1.042309	-14.830405	220	5	1.1	0.9;
	3680	1	-0	-0	0	12.4	0	1.039486	-15.440764	220	5	1.1	0.9;
	3687	1	0	-0	0	70.71	0	1.048351	-49.738272	380	4	1.1	0.9;
	3689	2	0	0	0	0	0	1.032178	-44.964327	220	4	1.1	0.9;
	3691	2	0	0	0	0	0	1.030169	13.484958	150	8	1.1	0.9;
	3692	2	0	0	0	0	0	1.025431	-42.014259	220	4	1.1	0.9;
	3697	1	-32.71	-13.76	0	135.136611	0	1.047634	-20.602675	380	5	1.1	0.9;
	3698	2	0	0	0	0	0	1.041028	-24.356749	220	5	1.1	0.9;
	3701	1	-0	0	0	0.43	0	1.046495	-25.065923	220	5	1.1	0.9;
	3707	1	77.49	17.3	0	0.1	0	1.046289	-15.379565	220	5	1.1	0.9;
	3711	2	0	0	0	0	0	1.00126	-38.557504	150	10	1.1	0.9;
	3712	1	-0.19	-0.87	0	0	0	1.027985	4.101781	150	8	1.1	0.9;
	3715	1	68.78	13	0	10.45	0	1.024475	-48.261603	220	4	1.1	0.9;
	3718	1	179.4	58.3	0	0.64	0	1.044223	-25.291795	220	5	1.1	0.9;
	3723	2	0	0	0	0	0	1.062798	49.722231	380	8	1.1	0.9;
	3728	1	67.77	10.4	0	7.18	0	1.034117	-52.492502	220	4	1.1	0.9;
	3734	1	51.36	-0.7	0	0.57	0	1.023655	-46.639617	220	4	1.1	0.9;
	3736	1	45.15	4.1	0	2.61	0	1.017883	-57.812403	220	4	1.1	0.9;
	3737	1	42.48	10.36	0	0.37	0	1.040625	-2.984485	220	5	1.1	0.9;
	3740	1	0	-0	0	13.03	0	1.036979	-21.255139	220	5	1.1	0.9;
	3741	2	0	0	0	0	0	1.035379	0.644884	380	5	1.1	0.9;
	3744	1	93.9	10.95	0	4.98	0	1.024137	-46.425744	220	4	1.1	0.9;
	3747	1	39.44	5.9	0	1.92	0	1.02865	-52.883204	220	4	1.1	0.9;
	3749	1	115.1	16.4	0	7.67	0	1.043452	-17.34385	220	5	1.1	0.9;
	3754	1	-24.93	11.8	0	3.51	0	1.038012	-38.799743	220	4	1.1	0.9;
	3758	1	83.49	17.4	0	19.32	0	1.02345	-19.032004	220	5	1.1	0.9;
	3760	1	163.5	44.9	0	1.17	0	1.056951	-20.961308	220	5	1.1	0.9;
	3768	1	19.82	10.9	0	0.05	0	1.041082	-21.937131	220	5	1.1	0.9;
	3771	1	34.14	13.5	0	0	0	1.015175	-0.740046	150	8	1.1	0.9;
	3775	1	55.54	15.06	0	6.28	0	1.000587	-23.540244	220	5	1.1	0.9;
	3776	1	0	-0	0	5.25	0	1.022691	11.561481	150	8	1.1	0.9;
	3786	1	-0	-0	0	26.98	0	1.047822	-19.769094	220	5	1.1	0.9;
	3791	2	0	0	0	0	0	0.997709	-39.697243	150	10	1.1	0.9;
	3792	2	0	0	0	0	0	1.058708	-43.684835	220	4	1.1	0.9;
	3793	1	197.7	67.65	0	13.58	0	1.033718	-20.651557	220	5	1.1	0.9;
	3794	1	165.3	7.4	0	28.32	0	1.029653	-28.186986	380	5	1.1	0.9;
	3798	1	-0	0	0	17.64	0	1.066587	-40.979103	380	4	1.1	0.9;
	3809	2	0	0	0	0	0	1.084761	-34.048929	220	5	1.1	0.9;
	3813	1	226	55.79	0	18.2	0	1.02677	-53.275089	220	4	1.1	0.9;
	3814	1	130.5	24.7	0	0.4	0	1.040225	-22.000602	220	5	1.1	0.9;
	3817	1	0	-0	0	1.02	0	1.084376	-15.14948	220	5	1.1	0.9;
	3818	1	44.45	10.9	0	2.23	0	1.041346	-25.830576	220	5	1.1	0.9;
	3819	1	0	-0	0	35	0	0.997429	-4.848161	150	8	1.1	0.9;
	3821	2	0	0	0	0	0	1.029497	-38.598952	220	4	1.1	0.9;
	3825	2	0	0	0	0	0	1.026381	-8.532195	380	5	1.1	0.9;
	3830	1	-0	0	0	97.326028	0	0.994231	-13.785717	380	5	1.1	0.9;
	3834	1	160	69.2	0	0.19	0	1.036025	-22.276797	220	5	1.1	0.9;
	3840	2	0	0	0	0	0	1.030834	-43.916162	220	4	1.1	0.9;
	3843	1	123.9	32.6	0	1.52	0	1.015649	-43.936354	220	4	1.1	0.9;
	3846	1	-0	0	0	5.36	0	1.021369	11.436684	150	8	1.1	0.9;
	3855	1	-38.73	-15.32	0	97.23	0	1.075113	-26.199377	220	5	1.1	0.9;
	3857	1	-0	-0	0	11.68	0	1.078948	-23.148876	220	5	1.1	0.9;
	3865	1	-0	-0	0	2.26	0	1.028906	-14.828465	220	5	1.1	0.9;
	3866	1	201.06	19.46	0	2.87	0	1.048854	-27.64579	220	5	1.1	0.9;
	3868	1	-8.43	-7.98	0	0	0	1.013932	1.288488	150	8	1.1	0.9;
	3869	2	0	0	0	0	0	1.035477	1.121843	380	5	1.1	0.9;
	3876	2	0	0	0	0	0	1.046886	-9.603363	220	5	1.1	0.9;
	3878	1	-0	-0	0	9.19	0	1.037445	22.511429	220	8	1.1	0.9;
	3880	1	-28.13	-0.82	0	0.09	0	1.063888	-11.500711	220	5	1.1	0.9;
	3891	1	0	0	0	7.89	0	1.04365	16.562241	220	8	1.1	0.9;
	3894	1	0	0	0	4.77	0	1.025966	-15.280447	220	5	1.1	0.9;
	3898	1	121.5	29.6	0	2.49	0	0.997386	-1.503423	150	8	1.1	0.9;
	3903	1	0	0	0	4.36	0	1.04983	-24.556874	220	5	1.1	0.9;
	3906	1	0	0	0	98.2	0	1.037134	-13.399373	380	5	1.1	0.9;
	3912	1	0	-0	0	19.73	0	1.037244	-21.114751	220	5	1.1	0.9;
	3916	2	0	0	0	0	0	1.044945	-14.025037	220	5	1.1	0.9;
	3918	1	-11.8	-5.47	0	88.050735	0	1.027043	-1.921362	380	5	1.1	0.9;
	3919	1	-0	0	0	39.95	0	1.025931	-6.057599	380	5	1.1	0.9;
	3925	1	89.3	25.1	0	2.98	0	1.02786	-13.10527	220	5	1.1	0.9;
	3928	1	105.2	24.2	0	0.25	0	1.041886	-4.72794	220	5	1.1	0.9;
	3929	1	48.95	11.6	0	2	0	1.032023	-20.245503	220	5	1.1	0.9;
	3932	1	-0	0	0	4.86	0	1.029043	-34.223167	220	10	1.1	0.9;
	3937	1	83.19	11.4	0	9.84	0	1.023016	-42.84661	220	4	1.1	0.9;
	3951	2	0	0	0	0	0	1.029014	-12.973188	220	5	1.1	0.9;
	3954	1	310.8	162.6	0	19.47	0	0.980151	-56.048746	220	10	1.1	0.9;
	3956	1	-0	-0	0	1.18	0	1.047817	-24.129675	220	5	1.1	0.9;
	3957	1	-0	-0	0	0.04	0	1.005937	-44.477596	220	10	1.1	0.9;
	3960	1	0	-0	0	5.62	0	1.034669	9.599065	150	8	1.1	0.9;
	3961	1	42.35	7.2	0	1.47	0	1.011587	-47.466292	220	4	1.1	0.9;
	3962	1	19.42	11.8	0	2.6	0	1.040806	-23.72917	220	5	1.1	0.9;
	3966	1	220.7	-14.8	0	125.52	0	1.045921	-42.713529	380	4	1.1	0.9;
	3968	1	-0.18	-2.27	0	0	0	1.050166	18.743096	380	8	1.1	0.9;
	3969	1	-0	-0	0	0.84	0	1.019911	-10.460405	220	5	1.1	0.9;
	3971	2	0	0	0	0	0	1.060173	-8.888833	220	5	1.1	0.9;
	3974	1	-0.18	-1.42	0	0	0	1.014106	4.41456	150	8	1.1	0.9;
	3975	1	148	26.1	0	9.44	0	0.98457	-24.821972	220	5	1.1	0.9;
	3983	2	0	0	0	0	0	1.012943	17.02029	150	8	1.1	0.9;
	3985	1	77.19	26.5	0	6.2	0	1.047185	-0.054112	220	5	1.1	0.9;
	3987	2	0	0	0	0	0	1.021034	-32.925081	220	10	1.1	0.9;
	3990	1	7.01	1.8	0	35.44	0	1.056447	-48.400884	380	10	1.1	0.9;
	3994	1	36.34	10.4	0	2.67	0	1.038572	-2.774266	220	5	1.1	0.9;
	3997	1	26.13	9.7	0	2.37	0	1.043493	-11.965937	220	5	1.1	0.9;
	3999	1	110.8	36.9	0	23.2	0	1.052755	-6.623039	220	5	1.1	0.9;
	4000	1	-0	0	0	24.103274	0	0.990281	-15.541038	380	5	1.1	0.9;
	4005	1	171.18	28.49	0	0	0	1.064294	-31.637164	220	5	1.1	0.9;
	4006	1	-0.18	-1	0	0	0	1.021667	11.13144	150	8	1.1	0.9;
	4011	1	144.8	39.7	0	12.64	0	1.030482	-53.578355	220	4	1.1	0.9;
	4013	1	0	-0	0	0.17	0	0.98986	-55.464877	150	10	1.1	0.9;
	4014	1	0	0	0	0	0	1.030612	-13.14884	150	2	1.1	0.9;
	4016	2	0	0	0	0	0	1.026951	-42.526395	220	4	1.1	0.9;
	4020	1	0	-0	0	14.28	0	1.047428	-49.661223	380	4	1.1	0.9;
	4021	1	-0.18	-1	0	0	0	1.010028	6.95481	380	8	1.1	0.9;
	4024	2	0	0	0	0	0	1.032708	2.56213	380	5	1.1	0.9;
	4025	1	0	0	0	2.67	0	1.038716	-2.763229	220	5	1.1	0.9;
	4026	2	0	0	0	0	0	1.05962	-40.932766	380	4	1.1	0.9;
	4031	1	0	-0	0	6.42	0	1.060541	-10.682858	220	5	1.1	0.9;
	4032	1	265.7	113.5	0	1.85	0	1.001235	-11.661775	220	5	1.1	0.9;
	4039	1	-0	-0	0	1.18	0	1.037738	-22.554888	220	5	1.1	0.9;
	4049	1	-0	-0	0	0.89	0	1.01917	-10.099581	220	5	1.1	0.9;
	4054	1	0	-0	0	2.71	0	1.032827	-18.511552	220	5	1.1	0.9;
	4056	2	0	0	0	0	0	1.042513	-9.76223	220	5	1.1	0.9;
	4057	1	74.78	-31.4	0	1.74	0	1.03204	8.696458	110	8	1.1	0.9;
	4058	1	0	0	0	0.91	0	1.010205	-47.598126	220	4	1.1	0.9;
	4059	1	0	0	0	0	0	1.087376	32.317567	380	8	1.1	0.9;
	4060	1	-0	0	0	2.08	0	1.06252	-14.73886	220	5	1.1	0.9;
	4062	1	6.31	0	0	0.42	0	1.014964	-48.212922	220	4	1.1	0.9;
	4077	1	0	-0	0	23.27	0	1.044199	16.785926	380	8	1.1	0.9;
	4081	1	-0	-0	0	6.9	0	1.033152	-52.522928	220	4	1.1	0.9;
	4082	1	-0	0	0	1.14	0	1.032369	12.590309	150	8	1.1	0.9;
	4084	2	0	0	0	0	0	1.060214	-8.829817	220	5	1.1	0.9;
	4087	1	-0	-0	0	26.36	0	1.049045	18.331888	380	8	1.1	0.9;
	4095	1	65.27	6	0	3.89	0	1.018058	-58.014896	220	4	1.1	0.9;
	4096	1	47.65	7.1	0	7.3	0	1.015565	-48.090558	220	4	1.1	0.9;
	4098	1	61.67	28.3	0	0.1	0	0.965808	-43.736662	220	10	1.1	0.9;
	4100	1	-0	-0	0	1.23	0	1.065816	-10.800151	220	5	1.1	0.9;
	4103	1	81.09	12.5	0	2.4	0	1.036557	-21.727519	220	5	1.1	0.9;
	4104	1	40.34	10.4	0	7.02	0	1.021346	11.229093	110	8	1.1	0.9;
	4110	1	13.01	3.8	0	3.59	0	1.04854	0.638802	220	5	1.1	0.9;
	4112	1	64.77	6.4	0	3.84	0	1.018938	-44.367979	220	4	1.1	0.9;
	4114	1	9.91	5.4	0	4.6	0	1.047098	-24.523765	220	5	1.1	0.9;
	4116	1	2.2	0.5	0	2.31	0	0.988341	-50.469464	150	10	1.1	0.9;
	4118	2	0	0	0	0	0	1.029072	-14.764759	220	5	1.1	0.9;
	4124	1	230.42	-7.87	0	5.13	0	1.02101	-45.251292	220	4	1.1	0.9;
	4125	2	0	0	0	0	0	0.999283	-8.393702	380	5	1.1	0.9;
	4127	1	39.14	7.7	0	11.68	0	1.032589	-20.184781	220	5	1.1	0.9;
	4128	2	0	0	0	0	0	1.023094	-15.634361	220	5	1.1	0.9;
	4130	1	0	-0	0	2.1	0	1.015422	10.923387	150	8	1.1	0.9;
	4132	1	25.43	7.9	0	1.03	0	1.017781	12.498081	110	8	1.1	0.9;
	4134	1	-0	0	0	0.09	0	1.073366	-33.684303	220	5	1.1	0.9;
	4141	1	-0.62	0	0	75.020288	0	1.031921	-21.125525	380	5	1.1	0.9;
	4143	1	11.11	4.1	0	0.8	0	1.051356	-1.147948	220	5	1.1	0.9;
	4144	1	310.2	4.5	0	8.51	0	1.043667	-19.857161	220	5	1.1	0.9;
	4157	1	0	0	0	1.2	0	1.074173	-26.3647	220	5	1.1	0.9;
	4165	1	325.2	38.1	0	19.99	0	1.031761	-43.289414	380	4	1.1	0.9;
	4170	1	-0	0	0	12.88	0	0.993515	-31.913491	220	10	1.1	0.9;
	4177	1	20.22	5.7	0	3.25	0	1.023308	-44.686399	220	4	1.1	0.9;
	4181	1	195.6	-1	0	3.19	0	1.043781	-45.342434	220	4	1.1	0.9;
	4182	2	0	0	0	0	0	1.016453	-43.470268	380	4	1.1	0.9;
	4185	1	87.5	21.3	0	7.52	0	1.041162	-25.851702	220	5	1.1	0.9;
	4186	1	-0	0	0	5.06	0	1.035985	-23.161266	220	5	1.1	0.9;
	4188	1	59.17	19.3	0	1.16	0	1.004131	0.103214	150	8	1.1	0.9;
	4189	1	200.65	30.72	0	9.72	0	1.032545	-26.186137	220	5	1.1	0.9;
	4192	2	0	0	0	0	0	1.077662	18.579115	380	8	1.1	0.9;
	4196	1	-3.2	-1.05	0	48.471563	0	1.026572	-7.529013	380	5	1.1	0.9;
	4197	1	9.01	4.9	0	0.64	0	1.066931	0.703575	220	5	1.1	0.9;
	4204	2	0	0	0	0	0	1.017661	-39.608142	220	4	1.1	0.9;
	4205	1	5.81	1.7	0	0.42	0	1.047904	-12.739022	220	5	1.1	0.9;
	4209	1	-0	-0	0	0	0	1.049923	18.730401	380	8	1.1	0.9;
	4215	1	-0	-0	0	17.29	0	1.04574	-23.060958	220	5	1.1	0.9;
	4216	1	-0.17	-8.06	0	0.83	0	1.0244	8.593062	110	8	1.1	0.9;
	4224	1	-20.56	-3.88	0	1.72	0	1.023033	-37.986866	220	4	1.1	0.9;
	4230	1	57.76	16	0	2.39	0	0.983046	-44.558409	220	4	1.1	0.9;
	4231	3	0	0	0	0	0	1.050918	0	380	5	1.1	0.9;
	4234	1	0	0	0	8.95	0	1.046144	-17.178168	220	5	1.1	0.9;
	4235	1	-0	-0	0	12.68	0	1.048409	-19.236107	220	5	1.1	0.9;
	4238	1	-0	-0	0	8.08	0	1.012199	-47.372938	220	4	1.1	0.9;
	4239	1	190.7	52	0	20.57	0	1.048509	-13.243223	220	5	1.1	0.9;
	4241	1	-0	-0	0	93.61	0	1.026348	-43.342141	380	4	1.1	0.9;
	4245	1	24.93	9.6	0	0.09	0	1.042873	-9.919925	220	5	1.1	0.9;
	4248	1	0	-0	0	3.24	0	1.006959	26.630788	220	8	1.1	0.9;
	4249	1	57.06	15.6	0	2.03	0	1.027416	-54.308862	220	4	1.1	0.9;
	4251	1	47.15	15.7	0	5.41	0	1.026117	-17.723087	220	5	1.1	0.9;
	4253	1	105	18	0	5.41	0	1.02913	-53.844323	220	4	1.1	0.9;
	4259	2	0	0	0	0	0	1.013671	-41.277855	150	10	1.1	0.9;
	4267	1	70.78	19.6	0	7.9	0	1.041691	-8.894803	220	5	1.1	0.9;
	4272	1	184.7	71.5	0	95.42	0	1.024269	2.729626	150	8	1.1	0.9;
	4274	1	35.49	11.53	0	1.31	0	1.030993	-53.273986	220	4	1.1	0.9;
	4277	1	-0	0	0	1.63	0	0.999219	-38.719503	150	10	1.1	0.9;
	4279	2	0	0	0	0	0	1.01711	-33.903407	220	4	1.1	0.9;
	4281	1	-0	-0	0	11.4	0	1.052621	-11.701312	220	5	1.1	0.9;
	4290	1	-1.9	-28.1	0	100.56	0	1.070175	-40.369777	380	4	1.1	0.9;
	4291	2	0	0	0	0	0	1.02141	-35.159361	220	4	1.1	0.9;
	4293	1	80.59	7.3	0	18.92	0	1.015968	-54.317746	220	4	1.1	0.9;
	4295	1	-0.18	-1	0	0	0	1.038977	6.823537	150	8	1.1	0.9;
	4300	1	81.49	25.1	0	0.36	0	0.998948	-16.239304	220	5	1.1	0.9;
	4301	1	-0	-0	0	2.27	0	1.037254	-21.115116	220	5	1.1	0.9;
	4302	1	30.53	9	0	1.9	0	1.028974	23.542154	110	8	1.1	0.9;
	4306	1	0	-0	0	0.3	0	1.033831	-30.862867	380	5	1.1	0.9;
	4313	1	133.4	29.4	0	10.87	0	1.04225	-21.167095	220	5	1.1	0.9;
	4314	1	26.13	2.7	0	1.7	0	1.044902	-14.232728	220	5	1.1	0.9;
	4315	1	-113.33	45.2	0	67.52	0	1.050551	-38.14331	380	4	1.1	0.9;
	4317	1	-0	0	0	3.81	0	1.00999	-40.622522	220	4	1.1	0.9;
	4318	1	51.06	16.5	0	2.37	0	1.046831	11.317068	110	8	1.1	0.9;
	4320	1	50.35	24.3	0	3.64	0	1.007531	26.5698	220	8	1.1	0.9;
	4324	1	-0	0	0	1.68	0	1.056635	-29.843383	220	5	1.1	0.9;
	4331	2	0	0	0	0	0	1.032206	-11.019439	220	5	1.1	0.9;
	4333	2	0	0	0	0	0	0.999193	-30.217701	220	10	1.1	0.9;
	4336	1	55.46	8.4	0	5	0	1.02824	-54.093567	220	4	1.1	0.9;
	4337	2	0	0	0	0	0	1.010847	6.866672	380	8	1.1	0.9;
	4338	2	0	0	0	0	0	1.024208	-47.464582	220	4	1.1	0.9;
	4339	1	-0	-0	0	42.345005	0	0.993094	-15.822682	380	5	1.1	0.9;
	4340	1	-0	-0	0	4.03	0	1.047125	26.552517	220	8	1.1	0.9;
	4344	1	68.48	-10.6	0	2.65	0	1.000951	-50.927954	150	10	1.1	0.9;
	4353	1	165.8	44	0	6.15	0	1.050036	-26.643732	220	5	1.1	0.9;
	4355	1	65.07	15	0	1.37	0	1.073092	-26.464431	220	5	1.1	0.9;
	4357	1	235	58.9	0	14.47	0	1.042363	-16.022891	220	5	1.1	0.9;
	4360	1	20.82	8.2	0	2.96	0	1.019445	-52.008231	220	4	1.1	0.9;
	4362	1	43.75	17.3	0	7.73	0	1.008262	4.432902	150	8	1.1	0.9;
	4363	1	0	-0	0	18.66	0	1.027315	-39.283356	380	1	1.1	0.9;
	4368	1	90.6	15.8	0	7.99	0	1.056502	-25.103292	220	5	1.1	0.9;
	4371	1	72.28	6.6	0	1.42	0	1.016476	-54.57476	220	4	1.1	0.9;
	4380	1	-67.31	-44.04	0	0	0	1.017434	1.176125	150	8	1.1	0.9;
	4390	1	-0	0	0	72.64	0	1.054595	-38.856296	380	4	1.1	0.9;
	4393	2	0	0	0	0	0	1.066547	-44.147015	380	10	1.1	0.9;
	4395	2	0	0	0	0	0	1.039119	-20.695167	220	5	1.1	0.9;
	4397	1	193	17.6	0	4.74	0	1.017881	-47.055799	220	4	1.1	0.9;
	4402	1	-0	0	0	1.02	0	1.084379	-15.149537	220	5	1.1	0.9;
	4407	1	-0	-0	0	51.43	0	1.067036	-40.648917	380	4	1.1	0.9;
	4409	1	-5.15	-16.03	0	0	0	1.035676	17.690796	150	8	1.1	0.9;
	4410	1	45.05	13.2	0	44.93	0	1.045437	-36.944821	220	5	1.1	0.9;
	4413	1	0	0	0	13.44	0	1.026248	11.663183	150	8	1.1	0.9;
	4415	1	30.03	9.7	0	4.1	0	1.023057	8.653225	110	8	1.1	0.9;
	4418	1	-0	0	0	0.05	0	1.075746	-24.387925	220	5	1.1	0.9;
	4419	2	0	0	0	0	0	1.059728	4.502903	380	5	1.1	0.9;
	4423	1	-0	-0	0	0.33	0	1.036754	-11.106241	150	2	1.1	0.9;
	4426	1	485.5	-63.38	0	34.92	0	1.067346	-46.835811	220	5	1.1	0.9;
	4427	1	296.8	77.7	0	0	0	1.035314	-10.898975	150	2	1.1	0.9;
	4428	2	0	0	0	0	0	1.019796	-43.510535	220	4	1.1	0.9;
	4431	1	19.52	5.6	0	1.46	0	1.01887	8.62678	150	8	1.1	0.9;
	4432	1	-0	0	0	0.8	0	1.074172	-26.364665	220	5	1.1	0.9;
	4435	1	-0	-0	0	102.217726	0	1.003844	-8.586482	380	5	1.1	0.9;
	4442	2	0	0	0	0	0	1.036687	-49.906013	380	4	1.1	0.9;
	4446	1	0	-0	0	9.25	0	1.038692	-38.842526	380	4	1.1	0.9;
	4450	1	140.2	11.2	0	9.1	0	1.02892	-53.925388	220	4	1.1	0.9;
	4454	1	-0	-0	0	1	0	0.993879	-18.276696	220	5	1.1	0.9;
	4467	1	-0	0	0	7.22	0	0.997475	-46.089189	220	4	1.1	0.9;
	4477	1	-0.2	-1.64	0	0	0	1.031936	4.19826	150	8	1.1	0.9;
	4480	2	0	0	0	0	0	1.069297	8.010702	220	5	1.1	0.9;
	4481	1	0	0	0	2.7	0	1.019184	-33.230536	220	1	1.1	0.9;
	4482	2	0	0	0	0	0	1.05005	-6.463671	220	5	1.1	0.9;
	4484	1	69.78	25	0	14.24	0	1.047688	-13.383588	220	5	1.1	0.9;
	4485	1	0	0	0	18.32	0	1.042138	15.995386	380	8	1.1	0.9;
	4489	1	0	-0	0	5.9	0	1.02429	-46.767263	220	4	1.1	0.9;
	4491	1	0	-0	0	1.75	0	1.029682	-28.956093	220	5	1.1	0.9;
	4493	1	15.32	18.2	0	27.62	0	1.021108	-40.326904	220	4	1.1	0.9;
	4494	1	-0	0	0	55.514618	0	1.020394	-0.81866	380	5	1.1	0.9;
	4495	1	336.3	104.4	0	0	0	1.033348	-10.869733	150	2	1.1	0.9;
	4504	1	170.1	-9.3	0	5.97	0	1.071798	-42.595944	220	5	1.1	0.9;
	4505	1	95.51	17.1	0	8.5	0	1.042633	-15.989429	220	5	1.1	0.9;
	4506	2	0	0	0	0	0	1.040247	-5.976296	220	5	1.1	0.9;
	4507	1	-0	0	0	43.44	0	1.030641	-41.584691	380	4	1.1	0.9;
	4509	1	77.99	30.8	0	2.41	0	1.022044	2.668853	150	8	1.1	0.9;
	4510	1	75.38	21.8	0	0.63	0	1.027661	-38.789178	220	4	1.1	0.9;
	4511	1	81.89	19.9	0	5.1	0	1.040393	-25.975564	220	5	1.1	0.9;
	4513	1	92.1	29.5	0	3.14	0	1.033259	-9.803992	220	5	1.1	0.9;
	4520	1	-0	-0	0	0.94	0	1.01917	-10.099585	220	5	1.1	0.9;
	4525	1	-0	0	0	6.12	0	1.057452	-8.831755	220	5	1.1	0.9;
	4529	1	48.48	8.7	0	10.14	0	1.056301	-15.154075	220	5	1.1	0.9;
	4541	1	0	-0	0	150.797812	0	1.023808	-9.896599	380	5	1.1	0.9;
	4542	1	-0	0	0	38.49	0	1.048495	-42.936839	380	4	1.1	0.9;
	4543	1	-0	0	0	2.32	0	1.056921	-45.124643	380	10	1.1	0.9;
	4544	1	-0	-0	0	14.64	0	1.047795	-14.689027	220	5	1.1	0.9;
	4549	1	0	-0	0	0.3	0	1.031082	-11.513168	220	5	1.1	0.9;
	4550	1	0	-0	0	124.2	0	1.014914	-10.363845	380	5	1.1	0.9;
	4554	1	119	21.9	0	0.95	0	1.042589	-9.819337	220	5	1.1	0.9;
	4556	1	138.2	21.3	0	2.33	0	0.979907	-48.927394	220	4	1.1	0.9;
	4562	1	130.5	25.7	0	2.27	0	1.019971	-16.826723	220	5	1.1	0.9;
	4565	1	0	0	0	0	0	1.023469	11.73541	150	8	1.1	0.9;
	4566	2	0	0	0	0	0	1.073131	-27.720699	220	5	1.1	0.9;
	4577	1	2.2	0.6	0	0.04	0	0.989833	-55.46852	150	10	1.1	0.9;
	4580	1	11.01	3.9	0	7.98	0	1.047068	-12.008433	220	5	1.1	0.9;
	4586	2	0	0	0	0	0	1.042233	-10.864402	150	2	1.1	0.9;
	4588	1	-0	0	0	75.38	0	1.049447	-49.731549	380	4	1.1	0.9;
	4592	1	183.8	2.2	0	2.98	0	0.981713	-48.59799	220	4	1.1	0.9;
	4594	1	-1.73	-1.71	0	22.062314	0	0.999168	-5.608971	380	5	1.1	0.9;
	4598	1	-0	0	0	52.958443	0	1.025937	-7.972316	380	5	1.1	0.9;
	4600	1	19.12	7.4	0	21.97	0	1.033446	-49.732578	380	4	1.1	0.9;
	4603	1	70.28	10.3	0	13.73	0	1.019644	-51.869124	220	4	1.1	0.9;
	4614	1	74.18	14.6	0	11.71	0	1.021383	8.264351	110	8	1.1	0.9;
	4615	1	42.55	17.6	0	1.2	0	1.061074	-25.400791	220	5	1.1	0.9;
	4616	1	30.83	16.5	0	3.83	0	1.012324	16.969427	150	8	1.1	0.9;
	4623	1	0	-0	0	80.388213	0	0.992881	-16.886734	380	5	1.1	0.9;
	4624	2	0	0	0	0	0	1.039942	-15.714328	220	5	1.1	0.9;
	4632	1	0	-0	0	3.21	0	1.020645	-48.236226	220	4	1.1	0.9;
	4644	1	18.22	8.6	0	34.7	0	1.037671	-43.914953	380	4	1.1	0.9;
	4650	1	-0	0	0	0	0	0.99335	-53.62447	220	10	1.1	0.9;
	4651	1	0	0	0	8.1	0	1.030087	-53.644773	220	4	1.1	0.9;
	4652	1	-0	0	0	165.63	0	1.064983	-44.679234	380	4	1.1	0.9;
	4656	1	93.9	-0.4	0	3.66	0	1.04736	-19.72727	220	5	1.1	0.9;
	4657	2	0	0	0	0	0	1.022686	-38.113595	220	4	1.1	0.9;
	4658	1	51.06	7.4	0	2.49	0	1.012281	-47.208522	220	4	1.1	0.9;
	4661	2	0	0	0	0	0	1.051599	-26.284671	220	5	1.1	0.9;
	4663	2	0	0	0	0	0	1.028053	-49.391084	220	4	1.1	0.9;
	4665	1	390.5	62.5	0	0	0	1.033888	-11.508364	150	2	1.1	0.9;
	4674	1	124	29.2	0	0.53	0	1.033798	-12.437269	220	5	1.1	0.9;
	4679	1	102.7	17.8	0	48.05	0	1.041212	-20.6513	220	5	1.1	0.9;
	4683	1	38.84	8.5	0	7.31	0	1.047572	-3.6912	220	5	1.1	0.9;
	4685	1	46.15	11.2	0	10.16	0	1.041453	-25.836412	220	5	1.1	0.9;
	4686	1	16.02	5.1	0	0.63	0	1.026397	8.915462	110	8	1.1	0.9;
	4687	1	0	0	0	0	0	1.01013	0.854718	150	8	1.1	0.9;
	4689	1	109.7	15	0	30.97	0	1.043624	-31.015452	220	5	1.1	0.9;
	4690	2	0	0	0	0	0	1.033192	-42.772431	380	4	1.1	0.9;
	4692	1	-0	0	0	50	0	1.011635	-0.528165	150	8	1.1	0.9;
	4701	2	0	0	0	0	0	1.044876	-9.959333	380	5	1.1	0.9;
	4704	1	0	-0	0.304554	30.256205	0	1.078317	-35.372047	380	10	1.1	0.9;
	4706	1	0	-0	0	20.81	0	1.060566	-45.898804	380	4	1.1	0.9;
	4710	1	195.4	76.84	0	8.59	0	1.028445	-9.15033	220	5	1.1	0.9;
	4711	1	-0	0	0	16.11	0	1.043852	-11.422203	220	5	1.1	0.9;
	4712	1	-0	0	0.073728	5.986272	0	1.032528	-34.167571	220	10	1.1	0.9;
	4717	1	0	0	0	46.25	0	1.043846	-42.202108	380	4	1.1	0.9;
	4721	1	0	-0	0	6.11	0	0.992962	-54.243678	150	10	1.1	0.9;
	4725	1	4.91	1.6	0	4.15	0	1.015888	-10.567762	220	5	1.1	0.9;
	4728	1	3.1	0.1	0	1.34	0	1.039095	-21.395607	220	5	1.1	0.9;
	4729	1	0	0	0	2.38	0	1.035146	-8.455601	220	5	1.1	0.9;
	4733	1	-0	0	0	0	0	1.049825	18.61991	380	8	1.1	0.9;
	4734	1	0	0	0	25.339116	0	1.012079	-10.191896	380	5	1.1	0.9;
	4735	1	0	0	0	13.59	0	0.998166	-44.27649	380	1	1.1	0.9;
	4738	1	101.89	25.72	0	5.11	0	1.064836	-10.744894	220	5	1.1	0.9;
	4740	1	-0.15	-0.9	0	0	0	1.037211	17.40897	110	8	1.1	0.9;
	4744	1	2.2	0	0	4.81	0	0.978365	-41.024043	220	10	1.1	0.9;
	4747	1	93.3	27.8	0	6.1	0	1.036623	-12.332394	220	5	1.1	0.9;
	4748	1	-0	-0	0	1.49	0	1.033506	-12.417653	220	5	1.1	0.9;
	4750	1	137.7	32.6	0	5.17	0	1.015137	-45.276024	220	4	1.1	0.9;
	4752	1	101.1	19.8	0	3.62	0	1.007473	-47.741639	220	4	1.1	0.9;
	4757	1	147	56.3	0	0.51	0	1.02687	3.635488	150	8	1.1	0.9;
	4765	1	33.04	10.1	0	0.43	0	1.03372	-12.259227	220	5	1.1	0.9;
	4771	1	0	0	0	1.32	0	0.994103	-36.152678	150	10	1.1	0.9;
	4772	2	0	0	0	0	0	1.027623	-47.726342	220	4	1.1	0.9;
	4773	2	0	0	0	0	0	1.020563	-42.19142	220	4	1.1	0.9;
	4774	1	63.27	28.9	0	2.19	0	1.027956	1.772495	150	8	1.1	0.9;
	4776	2	0	0	0	0	0	1.068296	27.287199	380	8	1.1	0.9;
	4779	1	-777.45	39.02	0	0	0	1.061631	33.330378	380	8	1.1	0.9;
	4783	2	0	0	0	0	0	1.069324	8.316401	220	5	1.1	0.9;
	4787	1	71.98	27.1	0	7.2	0	0.98971	-14.726367	380	5	1.1	0.9;
	4795	1	0	0	0	81.29	0	1.072329	-43.593004	380	4	1.1	0.9;
	4797	1	-0.01	0	0	0	0	1.087201	32.32252	380	8	1.1	0.9;
	4799	1	0	0	0	2.62	0	1.0073	5.93443	380	1	1.1	0.9;
	4806	1	210.8	-0.7	0	4.14	0	1.024462	-46.444223	220	4	1.1	0.9;
	4807	1	39.84	6.4	0	34.58	0	1.054594	-41.633661	380	10	1.1	0.9;
	4809	1	151	7.1	0	4.24	0	1.013804	-45.319191	220	4	1.1	0.9;
	4816	2	0	0	0	0	0	1.062911	-23.822779	220	5	1.1	0.9;
	4819	2	0	0	0	0	0	1.035474	1.140999	380	5	1.1	0.9;
	4823	2	0	0	0	0	0	1.06272	-14.712789	220	5	1.1	0.9;
	4825	1	-0	0	0	46.7	0	0.997516	8.732686	380	8	1.1	0.9;
	4826	1	103.6	16	0	11.22	0	1.075745	-24.387923	220	5	1.1	0.9;
	4829	1	171.9	39.1	0	4.23	0	1.020465	-5.667752	220	5	1.1	0.9;
	4831	1	130.4	44.33	0	5.22	0	1.027345	-15.792134	220	5	1.1	0.9;
	4843	2	0	0	0	0	0	1.053038	-42.000594	380	4	1.1	0.9;
	4846	2	0	0	0	0	0	1.030285	13.52358	150	8	1.1	0.9;
	4850	2	0	0	0	0	0	1.052648	-13.908759	220	5	1.1	0.9;
	4852	1	45.55	-59.9	0	30.04	0	1.039682	-20.363097	220	5	1.1	0.9;
	4855	1	127.9	-34	0	19.86	0	1.017659	-59.350334	220	4	1.1	0.9;
	4856	1	45.95	7.5	0	8.39	0	1.02638	-44.841603	220	4	1.1	0.9;
	4858	1	-0	-0	0	53.97	0	1.064819	30.746864	380	8	1.1	0.9;
	4863	1	-0	0	0	0	0	1.022186	11.546445	150	8	1.1	0.9;
	4864	1	-28.21	-6	0	0.05	0	1.039552	-12.879018	220	5	1.1	0.9;
	4867	1	65.77	27.3	0	0.49	0	1.035276	-3.453915	220	5	1.1	0.9;
	4868	1	2.2	0.5	0	2.88	0	0.998971	-47.459828	150	10	1.1	0.9;
	4872	1	15.42	3.9	0	0.79	0	1.016056	7.721617	110	8	1.1	0.9;
	4874	1	-0.58	-53.03	0	6.85	0	1.075519	-39.452902	220	5	1.1	0.9;
	4880	2	0	0	0	0	0	1.03955	-15.833489	220	5	1.1	0.9;
	4885	1	166.4	50.4	0	1.23	0	1.073792	-26.378319	220	5	1.1	0.9;
	4889	1	74.98	19.8	0	6.47	0	1.035733	-12.245951	220	5	1.1	0.9;
	4890	1	185	43.6	0	6.3	0	1.022368	-48.115238	220	4	1.1	0.9;
	4896	1	329.7	32.4	0	4.5	0	1.016398	-58.139663	220	4	1.1	0.9;
	4897	2	0	0	0	0	0	1.018589	-45.870257	220	4	1.1	0.9;
	4906	1	175.8	70.7	0	19.08	0	1.029168	-45.727716	220	4	1.1	0.9;
	4907	1	125.16	29.2	0	2.69	0	1.043664	-13.305602	220	5	1.1	0.9;
	4908	1	50.06	9.9	0	11.56	0	1.036682	-21.309654	220	5	1.1	0.9;
	4910	1	-0	0	0	0.95	0	1.024356	-46.521838	220	4	1.1	0.9;
	4914	1	14.91	-8.73	0	34.66	0	1.036506	-23.010125	220	5	1.1	0.9;
	4918	2	0	0	0	0	0	1.059135	-11.057428	220	5	1.1	0.9;
	4925	1	-0	0	0	0.3	0	1.036252	-22.509638	220	5	1.1	0.9;
	4926	1	3.8	1.1	0	10.69	0	1.055343	-41.653126	380	10	1.1	0.9;
	4929	1	37	8.7	0.070441	50.913893	0	1.019242	5.129207	380	2	1.1	0.9;
	4931	2	0	0	0	0	0	1.040167	-42.992365	380	4	1.1	0.9;
	4936	1	312.2	28	0	48.6	0	1.04373	-30.921402	220	5	1.1	0.9;
	4939	1	56.76	31.1	0	0.37	0	1.039171	-4.5288	220	5	1.1	0.9;
	4942	1	109.9	33.6	0	3.72	0	1.029688	-12.887815	220	5	1.1	0.9;
	4945	2	0	0	0	0	0	1.059702	-40.549906	380	4	1.1	0.9;
	4947	1	88.3	2.8	0	5.21	0	1.018243	-39.334796	220	4	1.1	0.9;
	4948	1	10.21	-3.2	0	5.15	0	1.033168	-39.709053	220	4	1.1	0.9;
	4950	1	218.7	-9.1	0	49	0	1.035276	-40.182771	220	5	1.1	0.9;
	4951	1	71.98	34.1	0	7.65	0	1.044873	-1.598447	220	5	1.1	0.9;
	4952	2	0	0	0	0	0	1.044318	-17.534802	220	5	1.1	0.9;
	4953	1	66.77	11.1	0	2.7	0	1.01977	-51.235898	220	4	1.1	0.9;
	4954	2	0	0	0	0	0	1.019681	-42.948639	220	4	1.1	0.9;
	4957	2	0	0	0	0	0	1.048836	-38.451969	380	4	1.1	0.9;
	4958	1	-0	-0	0	1.94	0	1.032637	-45.30464	220	4	1.1	0.9;
	4964	1	-0	-0	0	29.32	0	1.026451	12.070739	380	8	1.1	0.9;
	4967	2	0	0	0	0	0	1.023151	-41.92128	220	4	1.1	0.9;
	4968	1	-0	-0	0	16.57	0	1.051948	12.030292	380	1	1.1	0.9;
	4970	1	-0	-0	0	171.931977	0	1.05987	-17.639843	380	5	1.1	0.9;
	4974	1	89.08	26.9	0	4.68	0	1.018908	-13.502004	220	5	1.1	0.9;
	4993	1	-0.12	-1.74	0	0	0	1.089148	32.452882	380	8	1.1	0.9;
	5001	1	82.49	24.5	0	0.62	0	1.066984	15.48781	110	8	1.1	0.9;
	5002	1	-0	0	0	4.34	0	1.067682	-8.723226	220	5	1.1	0.9;
	5003	1	17.87	-8.59	0	0.21	0	1.043674	-9.696477	220	5	1.1	0.9;
	5004	2	0	0	0	0	0	1.061776	-15.444827	220	5	1.1	0.9;
	5006	1	-58.77	18.7	0	4.23	0	0.994019	-36.807025	150	10	1.1	0.9;
	5007	1	0	0	0	40.291745	0	1.021403	-15.311577	380	5	1.1	0.9;
	5008	1	0	-0	0	23.76	0	1.016418	12.111698	380	8	1.1	0.9;
	5016	1	-0	0	0	8.347383	0	1.022451	-5.00254	380	5	1.1	0.9;
	5019	2	0	0	0	0	0	1.04927	-12.982732	220	5	1.1	0.9;
	5025	2	0	0	0	0	0	1.006979	-47.649522	220	4	1.1	0.9;
	5031	1	174.7	35	0	2.14	0	1.022015	8.362718	150	8	1.1	0.9;
	5034	1	32.24	4.8	0	0.85	0	1.027075	-51.319318	220	4	1.1	0.9;
	5042	1	24.33	10	0	84.57	0	1.043173	-49.476619	380	4	1.1	0.9;
	5044	1	0	-0	0	12.01	0	1.060018	-39.975233	380	1	1.1	0.9;
	5049	1	136.6	29	0	2.05	0	1.019742	-20.452658	220	5	1.1	0.9;
	5051	2	0	0	0	0	0	1.039098	-16.500428	220	5	1.1	0.9;
	5060	2	0	0	0	0	0	1.067508	-40.432615	380	4	1.1	0.9;
	5065	2	0	0	0	0	0	1.072972	28.809735	380	8	1.1	0.9;
	5067	2	0	0	0	0	0	1.021364	-7.986964	380	5	1.1	0.9;
	5077	1	30.93	5.8	0	4.45	0	1.049884	-26.663944	220	5	1.1	0.9;
	5083	1	80.39	10.8	0	0.35	0	1.040014	-12.198345	220	5	1.1	0.9;
	5091	2	0	0	0	0	0	1.054352	-38.828724	380	4	1.1	0.9;
	5093	1	70.78	28	0	1.49	0	1.005327	-14.902286	220	5	1.1	0.9;
	5094	1	-0	-0	0	5.02	0	1.02894	-34.261029	220	10	1.1	0.9;
	5096	1	-2.27	-1.81	0	42.76	0	1.046399	-38.560768	380	4	1.1	0.9;
	5097	2	0	0	0	0	0	1.059554	-9.307901	150	2	1.1	0.9;
	5099	1	0	0	0	9.34	0	1.058069	1.551956	220	5	1.1	0.9;
	5106	1	39.94	13	0	2.47	0	1.070098	-31.647923	220	5	1.1	0.9;
	5108	1	-55.83	-31.99	0	2.88	0	1.040894	15.511042	110	8	1.1	0.9;
	5110	2	0	0	0	0	0	1.034874	-10.249956	220	5	1.1	0.9;
	5111	1	139.9	1.1	0	5.58	0	1.019649	-52.027452	220	4	1.1	0.9;
	5117	1	46.25	13.1	0	1.29	0	1.021991	-46.476868	220	4	1.1	0.9;
	5120	2	0	0	0	0	0	1.057395	-23.724008	220	5	1.1	0.9;
	5121	1	23.4	8.38	0	1.07	0	1.017828	-43.625753	220	4	1.1	0.9;
	5125	1	0	0	0	16.06	0	1.050162	-37.34517	380	4	1.1	0.9;
	5131	1	-0	0	0	7.84	0	1.074244	-26.36545	220	5	1.1	0.9;
	5137	1	16.92	0	0	0.57	0	1.023524	-16.375719	220	5	1.1	0.9;
	5144	2	0	0	0	0	0	1.035348	-3.111199	380	5	1.1	0.9;
	5146	1	229.9	48.9	0	8.36	0	1.01171	-34.157872	220	5	1.1	0.9;
	5147	1	-0	0	0	17.81	0	1.026032	3.684117	380	1	1.1	0.9;
	5155	1	411.3	168	0	0	0	1.031615	-10.83047	150	2	1.1	0.9;
	5162	1	55.26	0	0	1.05	0	1.016868	-47.850863	220	4	1.1	0.9;
	5163	1	-11.81	-12	0	19.39	0	1.03701	-38.496557	220	4	1.1	0.9;
	5165	1	-0	-0	0	35.06	0	1.016094	8.082006	380	8	1.1	0.9;
	5170	1	0	0	0	19.57	0	1.031754	-53.011941	220	4	1.1	0.9;
	5171	2	0	0	0	0	0	1.01545	-46.857001	220	4	1.1	0.9;
	5174	1	0	-0	0	8.19	0	1.043752	-21.06468	220	5	1.1	0.9;
	5179	1	179.8	5.67	0	12.83	0	1.044243	-30.656519	220	5	1.1	0.9;
	5182	1	0	0	0	0.18	0	1.045136	-7.970779	220	5	1.1	0.9;
	5188	1	107.9	17.6	0	1.03	0	1.015017	0.78079	150	8	1.1	0.9;
	5189	1	24.33	5.8	0	13.48	0	1.023618	-46.612109	220	4	1.1	0.9;
	5191	1	67.98	9.2	0	7.77	0	1.036462	-49.446584	220	4	1.1	0.9;
	5202	1	85.09	13.4	0	0.23	0	1.030983	-53.060445	220	4	1.1	0.9;
	5203	1	221.9	86.3	0	2.18	0	1.027094	5.766989	150	8	1.1	0.9;
	5205	1	-0	-0	0	2.84	0	1.026251	-45.579804	220	1	1.1	0.9;
	5210	1	347.61	74.37	0	0	0	0.999596	0.054117	220	2	1.1	0.9;
	5212	1	-0	0	0	0.2	0	1.044864	-11.010494	220	5	1.1	0.9;
	5213	1	246.8	36.3	0	16.01	0	1.045448	-26.411208	220	5	1.1	0.9;
	5214	1	0	-0	0	0.88	0	1.041	-2.410275	220	1	1.1	0.9;
	5215	1	-0	-0	0	0.89	0	1.024179	-5.83903	220	5	1.1	0.9;
	5217	1	12.11	-0	0	0.6	0	1.016177	-47.894882	220	4	1.1	0.9;
	5233	1	-0	-0	0	14.04	0	1.04328	-10.973219	220	5	1.1	0.9;
	5237	2	0	0	0	0	0	1.050569	-14.773477	220	5	1.1	0.9;
	5239	2	0	0	0	0	0	1.005494	-46.913998	380	4	1.1	0.9;
	5241	1	0	-0	0	70.905048	0	1.028186	-3.618432	380	5	1.1	0.9;
	5246	1	0	-0	0	1.06	0	1.032411	12.598115	150	8	1.1	0.9;
	5247	1	0	-0	0	0.22	0	1.001398	-0.49248	150	8	1.1	0.9;
	5255	1	43.95	8.5	0	29.29	0	1.03341	9.303036	150	8	1.1	0.9;
	5256	1	91	36	0	5.47	0	1.000884	-16.700858	220	5	1.1	0.9;
	5257	1	-0	0	0	0.18	0	1.049117	-6.26418	220	5	1.1	0.9;
	5263	1	0	0	0	0.23	0	1.055839	-44.433532	380	10	1.1	0.9;
	5266	1	2.2	0.5	0	0.06	0	1.028935	-34.261717	220	10	1.1	0.9;
	5270	1	-0	-0	0	7.4	0	1.04013	-21.145434	220	5	1.1	0.9;
	5271	1	69.18	19.7	0	6.99	0	1.020246	8.846019	150	8	1.1	0.9;
	5274	1	2.2	0.5	0	1.84	0	0.995551	-46.965426	150	10	1.1	0.9;
	5276	1	143.2	5.9	0	5.79	0	1.018539	-43.988903	220	4	1.1	0.9;
	5278	2	0	0	0	0	0	1.041244	-24.302931	220	5	1.1	0.9;
	5279	1	158.8	105.3	0	21.76	0	1.012625	16.961421	150	8	1.1	0.9;
	5280	2	0	0	0	0	0	1.021983	-41.283482	220	4	1.1	0.9;
	5286	1	17.22	5.1	0	0.99	0	1.037057	-22.192474	220	5	1.1	0.9;
	5288	1	-0	-0	0	3.06	0	0.997407	-15.135755	380	5	1.1	0.9;
	5289	1	-0	-0	0	165.34	0	1.076031	-37.794005	380	4	1.1	0.9;
	5290	2	0	0	0	0	0	1.011274	-35.534778	220	4	1.1	0.9;
	5295	1	-0	-0	0	2.08	0	1.033108	21.51946	220	8	1.1	0.9;
	5296	1	16.02	4.7	0	0.94	0	1.015725	19.372243	110	8	1.1	0.9;
	5297	1	55.26	11	0	7.68	0	1.040618	-24.406654	220	5	1.1	0.9;
	5299	2	0	0	0	0	0	1.061166	34.421693	380	8	1.1	0.9;
	5300	1	-0	-0	0	1.3	0	1.045905	-10.997713	220	5	1.1	0.9;
	5308	1	0	-0	0	42.19206	0	0.98387	-18.21545	380	5	1.1	0.9;
	5311	1	-0	-0	0	7.85	0	1.043637	16.562698	220	8	1.1	0.9;
	5317	1	293.2	87.4	0	26.36	0	1.028	-44.254256	220	5	1.1	0.9;
	5320	1	18.12	5.9	0	1.52	0	1.02841	3.793243	150	8	1.1	0.9;
	5329	2	0	0	0	0	0	1.018401	-42.886221	220	4	1.1	0.9;
	5331	1	-0	0	0	0.23	0	1.000112	-35.788348	150	10	1.1	0.9;
	5332	1	366.6	57.4	0	60.78	0	1.021425	-50.698742	380	4	1.1	0.9;
	5333	1	83.59	-7.2	0	12.95	0	1.002717	-46.751822	220	10	1.1	0.9;
	5334	1	17.52	6.6	0	3.79	0	1.034725	-12.75309	220	5	1.1	0.9;
	5339	1	-16.62	8.7	0	7.68	0	0.999355	-32.205213	220	10	1.1	0.9;
	5340	2	0	0	0	0	0	1.051438	-11.200258	220	5	1.1	0.9;
	5341	1	64.27	17.5	0	0.43	0	1.01091	-21.676826	220	5	1.1	0.9;
	5342	1	32.14	13.1	0	1.72	0	1.012688	-1.048625	150	8	1.1	0.9;
	5346	1	0	0	0	0.81	0	1.048768	28.252436	110	8	1.1	0.9;
	5348	2	0	0	0	0	0	1.015924	-42.487214	220	4	1.1	0.9;
	5350	1	0	0	0	28.208233	0	0.976782	-24.559425	380	5	1.1	0.9;
	5351	1	0	0	0	3.28	0	1.037736	-14.066675	220	5	1.1	0.9;
	5354	1	70.18	19.5	0	3.59	0	1.046455	-8.661275	220	5	1.1	0.9;
	5361	1	-0	0	0	2.77	0	0.972326	-42.828749	220	10	1.1	0.9;
	5362	1	152.73	-17.2	0	45.85	0	1.043479	-24.26519	220	5	1.1	0.9;
	5364	1	110.6	38.1	0	0.98	0	1.020499	-1.54351	150	8	1.1	0.9;
	5365	2	0	0	0	0	0	1.063822	-11.51529	220	5	1.1	0.9;
	5366	1	66.07	12.5	0	2.08	0	1.02359	-46.644005	220	4	1.1	0.9;
	5372	1	21.02	8	0	13.14	0	1.063524	-40.57254	380	4	1.1	0.9;
	5379	2	0	0	0	0	0	1.105689	-6.79391	380	5	1.3	0.7;
	5383	1	0	-0	0	24.7	0	1.047026	-24.505001	220	5	1.1	0.9;
	5388	1	159.5	42	0	2.3	0	1.007959	-14.803624	220	5	1.1	0.9;
	5391	1	57.06	11.9	0	1.68	0	1.023168	-48.106837	220	4	1.1	0.9;
	5393	1	-0	0	0	9.43	0	1.061367	-15.492076	220	5	1.1	0.9;
	5395	2	0	0	0	0	0	1.049	-27.436125	220	5	1.1	0.9;
	5400	1	241.62	52.78	0	0.82	0	1.026563	-13.995618	220	5	1.1	0.9;
	5409	2	0	0	0	0	0	1.023182	-38.682073	220	4	1.1	0.9;
	5410	1	69.78	26	0	5.18	0	1.044777	-21.406419	220	5	1.1	0.9;
	5413	1	-0	-0	0	5.57	0	1.03575	-18.055374	220	5	1.1	0.9;
	5416	1	0	0	0	25.13	0	1.001202	12.152799	380	2	1.1	0.9;
	5417	1	32.74	5.3	0	3.07	0	1.059856	-12.1436	220	5	1.1	0.9;
	5418	1	-0	0	0	3.96	0	1.036715	-21.310085	220	5	1.1	0.9;
	5419	1	349.3	144.3	0	0.54	0	1.018426	-15.162316	220	5	1.1	0.9;
	5420	1	56.96	1.94	0	26.48	0	1.02044	-22.990881	380	5	1.1	0.9;
	5421	1	-0	-0	0	0.92	0	0.997405	-15.13568	380	5	1.1	0.9;
	5423	1	28.63	7.3	0	4.36	0	1.002482	6.650403	110	8	1.1	0.9;
	5425	2	0	0	0	0	0	1.077356	-35.523318	380	4	1.1	0.9;
	5426	1	152.8	38	0	1.46	0	1.021179	2.993375	150	8	1.1	0.9;
	5435	1	103.3	24.1	0	9.41	0	1.013845	-46.918495	220	4	1.1	0.9;
	5441	1	138.01	34.55	0	8.44	0	1.041494	-2.871036	220	5	1.1	0.9;
	5442	1	-0.17	-0.84	0	0	0	1.012513	16.946551	150	8	1.1	0.9;
	5450	1	-0	-0	0	1.17	0	1.053632	1.072582	150	8	1.1	0.9;
	5452	1	82.09	39.8	0	2.35	0	0.997595	-47.54296	150	10	1.1	0.9;
	5455	1	-0	-0	0	5.27	0	1.04513	-7.970741	220	5	1.1	0.9;
	5458	1	0	0	0	9.09	0	1.038363	-21.451874	220	5	1.1	0.9;
	5460	1	-15.89	-6.42	0	38.657071	0	0.988594	-18.027515	380	5	1.1	0.9;
	5461	2	0	0	0	0	0	1.05743	-18.381726	380	5	1.1	0.9;
	5463	1	88.4	-5.9	0	3.32	0	1.02307	-42.018425	220	4	1.1	0.9;
	5464	1	25.23	3.3	0	35.45	0	1.04942	-38.419571	380	4	1.1	0.9;
	5465	2	0	0	0	0	0	1.001327	-46.65705	220	4	1.1	0.9;
	5469	1	106.6	-18.8	0	4.1	0	1.067045	-38.802465	220	5	1.1	0.9;
	5477	1	-0	-0	0	18.02	0	1.074995	-26.195531	220	5	1.1	0.9;
	5481	2	0	0	0	0	0	1.063327	-8.931038	380	5	1.1	0.9;
	5482	2	0	0	0	0	0	1.046078	-7.297941	220	5	1.1	0.9;
	5486	2	0	0	0	0	0	1.0598	4.502248	380	5	1.1	0.9;
	5488	2	0	0	0	0	0	1.041439	-29.952888	220	5	1.1	0.9;
	5490	2	0	0	0	0	0	1.064044	-15.148619	380	5	1.1	0.9;
	5494	1	0	-0	0	4.69	0	0.999098	-36.908159	220	4	1.1	0.9;
	5500	1	111.5	24.8	0	9.65	0	1.026927	-53.371077	220	4	1.1	0.9;
	5502	1	-0	-0	0	0.58	0	1.037491	-20.54784	220	5	1.1	0.9;
	5509	1	0	0	0.314057	17.296509	0	0.994728	-6.017725	380	2	1.1	0.9;
	5511	1	67.68	6.2	0	6.83	0	1.016789	-59.240772	220	4	1.1	0.9;
	5512	2	0	0	0	0	0	1.020123	-41.318107	220	4	1.1	0.9;
	5517	1	34.24	-19.6	0	3.59	0	0.9775	-42.539832	150	10	1.1	0.9;
	5519	1	0	-0	0	1.98	0	1.047274	-24.565625	220	5	1.1	0.9;
	5522	1	-0	-0	0	1.18	0	1.067053	0.712699	220	5	1.1	0.9;
	5525	1	236.33	59.35	0	18.71	0	1.035462	-14.021081	220	5	1.1	0.9;
	5526	1	-8.71	-19.54	0	0	0	1.040627	17.945588	150	8	1.1	0.9;
	5529	1	0	0	0	0.36	0	1.041514	-2.871216	220	5	1.1	0.9;
	5533	2	0	0	0	0	0	1.041341	-13.595487	220	5	1.1	0.9;
	5546	2	0	0	0	0	0	1.055712	-23.801872	220	5	1.1	0.9;
	5549	1	30.03	12.2	0	0.93	0	1.00229	-0.631077	150	8	1.1	0.9;
	5550	1	0	-0	0	10.83	0	1.044355	-25.941561	220	5	1.1	0.9;
	5558	1	-0	-0	0	15.17	0	1.013694	17.294351	380	8	1.1	0.9;
	5559	1	0	-0	0	2.66	0	1.002365	-42.648238	220	10	1.1	0.9;
	5562	1	-0	0	0.19855	29.549	0	1.069676	-45.57785	380	10	1.1	0.9;
	5564	2	0	0	0	0	0	1.042161	-12.585697	220	5	1.1	0.9;
	5567	1	0	0	0	3.59	0	1.017406	-18.92817	220	5	1.1	0.9;
	5571	1	-0	0	0	1.01	0	1.036243	-22.509656	220	5	1.1	0.9;
	5573	1	0	-0	0	2.94	0	1.060977	-23.409461	220	5	1.1	0.9;
	5574	1	99.7	6	0	2.06	0	1.040549	-11.86988	220	5	1.1	0.9;
	5586	1	-0	0	0	0	0	1.05568	2.418195	220	5	1.1	0.9;
	5587	1	-0	-0	0	0.33	0	1.035808	-10.367302	150	2	1.1	0.9;
	5589	1	0	0	0	173.904854	0	1.03383	-30.862864	380	5	1.1	0.9;
	5591	1	118.6	41.5	0	0.58	0	1.018603	-1.652419	150	8	1.1	0.9;
	5592	1	20.82	7.9	0	18	0	1.014536	-51.043193	220	4	1.1	0.9;
	5595	1	168.5	54.8	0	5.34	0	1.027836	3.426519	150	8	1.1	0.9;
	5608	1	449.4	14.6	0	87.31	0	1.028672	-50.35437	380	4	1.1	0.9;
	5610	1	99.5	9.2	0	27.55	0	1.045642	-24.197425	220	5	1.1	0.9;
	5616	1	225.2	35.2	0	11.99	0	1.075879	-35.292592	220	5	1.1	0.9;
	5617	1	38.04	3.4	0	3.43	0	0.98127	-49.187417	220	4	1.1	0.9;
	5618	1	-0	0	0	1.75	0	0.997607	-4.667747	150	8	1.1	0.9;
	5627	1	-0	-0	0	0.53	0	1.027411	-32.553941	220	5	1.1	0.9;
	5630	1	0	0	0	11.52	0	1.037555	-4.070289	220	5	1.1	0.9;
	5631	1	105.5	31.1	0	9.63	0	0.989076	-43.054381	220	4	1.1	0.9;
	5641	1	-0	-0	0	5.98	0	1.048787	-26.137467	220	5	1.1	0.9;
	5643	1	-0	-0	0	1.1	0	0.989577	-55.845538	220	10	1.1	0.9;
	5648	1	-0	-0	0	115.528858	0	0.995084	-13.471533	380	5	1.1	0.9;
	5653	1	96.59	7.28	0	2.52	0	1.047534	-7.680116	220	5	1.1	0.9;
	5655	1	-0.21	-1.04	0	0	0	1.016772	2.827447	150	8	1.1	0.9;
	5658	2	0	0	0	0	0	1.009808	-23.836753	380	5	1.1	0.9;
	5662	1	185.7	18.4	0	23.68	0	1.010873	-41.932245	220	10	1.1	0.9;
	5664	2	0	0	0	0	0	1.068735	0.968479	220	5	1.1	0.9;
	5666	1	124.8	34.7	0	3.66	0	1.046893	-10.752987	220	5	1.1	0.9;
	5669	2	0	0	0	0	0	1.003946	-34.011961	220	10	1.1	0.9;
	5670	2	0	0	0	0	0	1.018899	-47.108978	220	4	1.1	0.9;
	5672	1	139.7	9.7	0	12.61	0	0.982808	-41.150035	220	10	1.1	0.9;
	5674	1	-536.71	-85.76	0	34.76	0	1.009419	7.023959	380	8	1.1	0.9;
	5681	1	198.1	33.9	0	22.82	0	1.031456	-53.165229	220	4	1.1	0.9;
	5685	1	34.74	16.3	0	33.38	0	1.039032	-42.883507	380	4	1.1	0.9;
	5686	1	52.26	19.6	0	2.64	0	1.033278	-10.192796	220	5	1.1	0.9;
	5688	1	14.12	5.2	0	1.44	0	1.035841	-8.055175	220	5	1.1	0.9;
	5691	1	195.1	-14.2	0	12.89	0	1.060611	-45.305556	220	5	1.1	0.9;
	5695	1	142.8	20.3	0	6.77	0	1.047163	-27.187915	220	5	1.1	0.9;
	5697	1	131.6	33.8	0	6.55	0	1.010447	-47.112562	220	4	1.1	0.9;
	5699	1	-0	0	0	0.59	0	1.049198	-6.529972	220	5	1.1	0.9;
	5709	2	0	0	0	0	0	1.026473	-8.528086	380	5	1.1	0.9;
	5712	1	183.07	44.65	0	2.77	0	1.039647	-4.558258	220	5	1.1	0.9;
	5720	1	337.6	-15.4	0	6.27	0	1.040698	-25.156245	220	5	1.1	0.9;
	5723	1	141.2	45.4	0	7.55	0	1.041768	-9.639222	220	5	1.1	0.9;
	5726	2	0	0	0	0	0	1.020165	-42.201068	220	4	1.1	0.9;
	5735	1	161.3	10.04	0	4.38	0	1.054644	-26.928091	220	5	1.1	0.9;
	5738	1	-0	-0	0	8.13	0	1.015478	-10.510275	220	5	1.1	0.9;
	5740	1	-0	0	0	1.26	0	1.014654	4.973889	150	8	1.1	0.9;
	5742	2	0	0	0	0	0	1.028692	26.344515	220	8	1.1	0.9;
	5743	1	-0	0	0	4.28	0	1.05064	-24.164058	220	5	1.1	0.9;
	5748	1	-0	-0	0	4.27	0	0.996773	-4.642016	150	8	1.1	0.9;
	5753	1	-0	-0	0	1.81	0	1.037714	-3.26164	220	5	1.1	0.9;
	5760	1	0	-0	0	41.75	0	1.047432	7.554087	150	8	1.1	0.9;
	5762	1	0	0	0	2.23	0	1.053432	0.451774	380	2	1.1	0.9;
	5764	1	99.4	4.3	0	3.69	0	1.030004	-25.513064	220	5	1.1	0.9;
	5767	1	-0	0	0	1.87	0	1.024855	-48.167766	220	4	1.1	0.9;
	5771	1	190.5	70.9	0	21.3	0	1.005505	-46.602459	380	4	1.1	0.9;
	5772	1	-38.01	-22.73	0	0	0	1.036506	5.394435	150	8	1.1	0.9;
	5773	1	0	0	0	0.36	0	1.040729	-2.404677	220	1	1.1	0.9;
	5776	1	222.7	80.5	0	0	0	1.063217	-7.210075	150	2	1.1	0.9;
	5780	1	0	-0	0	0.95	0	1.030855	-52.958818	220	4	1.1	0.9;
	5781	2	0	0	0	0	0	1.038035	-11.764738	380	5	1.1	0.9;
	5783	1	69.08	-49.6	0	3.42	0	1.054882	11.673246	110	8	1.1	0.9;
	5787	2	0	0	0	0	0	1.031918	-43.4138	220	4	1.1	0.9;
	5789	1	0	0	0	0.35	0	1.000709	-16.068788	220	5	1.1	0.9;
	5790	1	4.4	1.1	0	0.41	0	1.02673	-35.913747	220	10	1.1	0.9;
	5792	1	40.95	13.4	0	51.25	0	1.017936	0.566548	150	8	1.1	0.9;
	5793	1	6.11	0	0	1.17	0	1.011587	-47.471007	220	4	1.1	0.9;
	5797	1	334.1	103.5	0	14.78	0	0.982903	-55.937579	220	10	1.1	0.9;
	5799	1	-0	-0	0	1.38	0	1.043112	-32.312018	220	5	1.1	0.9;
	5803	1	0	-0	0	0.35	0	1.044961	-14.02515	220	5	1.1	0.9;
	5805	1	167.6	-9	0	0.39	0	0.992355	-54.488935	150	10	1.1	0.9;
	5814	2	0	0	0	0	0	1.048258	-11.867086	220	5	1.1	0.9;
	5815	1	5.21	1.3	0	24.85	0	1.044027	-50.864596	380	10	1.1	0.9;
	5821	2	0	0	0	0	0	1.019461	-43.826963	220	4	1.1	0.9;
	5823	2	0	0	0	0	0	1.063184	-39.213955	380	4	1.1	0.9;
	5825	2	0	0	0	0	0	0.995795	-36.829799	220	10	1.1	0.9;
	5828	2	0	0	0	0	0	1.029274	-36.522798	220	4	1.1	0.9;
	5829	1	155.5	38.3	0	52.84	0	1.011522	-0.525357	150	8	1.1	0.9;
	5831	2	0	0	0	0	0	1.032948	-46.95059	380	4	1.1	0.9;
	5835	1	0	0	0	3.32	0	1.012728	16.946231	150	8	1.1	0.9;
	5836	1	244.09	48.92	0	5.96	0	1.040966	-12.033407	220	5	1.1	0.9;
	5837	1	-0	-0	0	132.04744	0	1.019384	-10.165641	380	5	1.1	0.9;
	5843	1	-0	0	0	8.58	0	1.034072	15.26583	380	8	1.1	0.9;
	5848	1	0	-0	0.29	-0.42	0	1.003622	-4.593673	380	2	1.1	0.9;
	5853	1	122.5	38.9	0	0.84	0	1.044384	-7.444582	220	5	1.1	0.9;
	5856	2	0	0	0	0	0	1.039775	-11.608229	220	5	1.1	0.9;
	5857	1	78.39	21.2	0	0.97	0	1.056548	-21.100497	220	5	1.1	0.9;
	5872	1	177.5	72.5	0	46.35	0	1.020681	11.413214	150	8	1.1	0.9;
	5873	1	0	0	0	132.84	0	1.020973	-45.161164	380	4	1.1	0.9;
	5877	1	-0	0	0	67.98	0	1.048701	-49.785885	380	4	1.1	0.9;
	5881	2	0	0	0	0	0	1.062314	-10.435974	220	5	1.1	0.9;
	5884	1	-0	-0	0	68.73	0	1.039018	-39.486484	380	4	1.1	0.9;
	5887	1	6.91	-11.1	0	0.67	0	1.007987	21.860203	110	8	1.1	0.9;
	5888	1	-0	-0	0	93.07	0	1.052038	17.362903	380	8	1.1	0.9;
	5890	1	0	-0	0	3.66	0	1.030272	-45.767937	220	4	1.1	0.9;
	5891	1	14.02	5.5	0	3.3	0	1.038346	-12.124908	220	5	1.1	0.9;
	5892	1	49.86	14.1	0	11.33	0	1.024992	-48.153536	220	4	1.1	0.9;
	5893	1	26.03	10.6	0	2.14	0	1.026601	16.102265	110	8	1.1	0.9;
	5897	1	0	0	0	0	0	1.049488	18.541477	380	8	1.1	0.9;
	5907	1	186.1	14.8	0	17.28	0	1.047873	-8.494871	220	5	1.1	0.9;
	5916	1	-57.59	-30.52	0	2.45	0	1.043019	18.096338	110	8	1.1	0.9;
	5918	1	73.83	16.41	0	15.77	0	1.053618	-25.89283	220	5	1.1	0.9;
	5919	1	-0	0	0	33.03	0	1.046245	-49.652257	380	4	1.1	0.9;
	5923	1	74.08	10.7	0	13.55	0	1.029055	-53.701552	220	4	1.1	0.9;
	5926	1	0	-0	0	3.98	0	1.071514	-48.003062	220	5	1.1	0.9;
	5927	1	-0	0	0	1.44	0	0.994079	-36.184356	150	10	1.1	0.9;
	5933	1	78.49	18.5	0	6.91	0	1.008443	-48.246125	220	4	1.1	0.9;
	5935	1	-0	0	0	5.75	0	1.066928	-22.224726	220	5	1.1	0.9;
	5940	2	0	0	0	0	0	1.031482	-14.571065	220	5	1.1	0.9;
	5944	1	71.88	21.7	0	9.63	0	1.054767	-25.897507	220	5	1.1	0.9;
	5946	1	0	-0	0	5.96	0	1.023681	-42.208451	220	4	1.1	0.9;
	5948	1	57.06	13.3	0	19.18	0	1.024134	-48.323145	220	4	1.1	0.9;
	5949	1	160.7	40.3	0	3.04	0	1.016343	2.822912	150	8	1.1	0.9;
	5952	1	40.34	15.9	0	1.31	0	1.023957	4.371578	150	8	1.1	0.9;
	5957	1	208.06	36.6	0	3.55	0	1.027417	-16.202424	220	5	1.1	0.9;
	5960	1	-0	0	0	2.78	0	1.019184	-33.227537	220	1	1.1	0.9;
	5961	1	113.5	11.5	0	5.53	0	1.015278	-46.320257	220	4	1.1	0.9;
	5970	2	0	0	0	0	0	1.019864	-40.809127	220	4	1.1	0.9;
	5971	2	0	0	0	0	0	1.067028	5.04202	380	5	1.3	0.7;
	5974	1	0	0	0	0	0	1.023469	11.73541	150	8	1.1	0.9;
	5979	1	-0	-0	0	0	0	1.046415	26.479954	220	8	1.1	0.9;
	5983	2	0	0	0	0	0	1.064637	-14.879564	380	5	1.1	0.9;
	5987	1	24.04	1.68	0	4.02	0	1.03795	-13.934441	220	5	1.1	0.9;
	5990	1	0	-0	0	4.53	0	1.024622	-10.185671	220	5	1.1	0.9;
	5991	2	0	0	0	0	0	1.018313	-49.651632	220	4	1.1	0.9;
	5993	1	0	-0	0	2.75	0	1.044859	-11.010438	220	5	1.1	0.9;
	5994	2	0	0	0	0	0	1.034164	-15.365343	220	5	1.1	0.9;
	5996	1	-0	-0	0.164188	4.023827	0	1.053383	0.452005	380	2	1.1	0.9;
	6000	2	0	0	0	0	0	1.085651	16.9619	380	8	1.3	0.7;
	6004	1	0	0	0	2.37	0	1.064444	-47.252529	380	10	1.1	0.9;
	6009	1	20.82	7.9	0	6.73	0	1.014176	-50.975888	220	4	1.1	0.9;
	6010	1	0	0	0	1.35	0	1.07902	-23.14938	220	5	1.1	0.9;
	6015	1	-0	0	0	25	0	1.022574	-1.389768	150	8	1.1	0.9;
	6017	1	-0	0	0	23.15	0	1.024964	11.939308	380	8	1.1	0.9;
	6022	1	-0	-0	0	5.5	0	1.001596	-37.747037	220	4	1.1	0.9;
	6025	1	-0.16	-1.92	0	0.45	0	1.029944	15.101784	110	8	1.1	0.9;
	6028	1	0	0	0	3.24	0	1.02427	8.595544	110	8	1.1	0.9;
	6031	1	0	0	0	2.22	0	1.039374	-15.780003	220	5	1.1	0.9;
	6036	2	0	0	0	0	0	1.041753	-0.893841	380	5	1.1	0.9;
	6038	1	43.55	-20.5	0	2.53	0	1.047079	-45.177751	220	4	1.1	0.9;
	6039	1	165.9	154.6	0	15.17	0	1.042753	-50.581128	380	4	1.1	0.9;
	6040	1	-57.46	-30.3	0	5.4	0	1.026365	-44.846715	220	4	1.1	0.9;
	6041	1	0	0	0	0.53	0	1.071556	-48.003338	220	5	1.1	0.9;
	6045	1	-0.2	-0.98	0	0	0	1.001329	-0.497586	150	8	1.1	0.9;
	6047	1	79.49	22.7	0	9.26	0	1.021029	8.869605	150	8	1.1	0.9;
	6052	1	-0	-0	0	0	0	1.019191	18.744304	150	8	1.1	0.9;
	6053	1	412.1	79.3	0	7.73	0	1.052661	-26.616993	220	5	1.1	0.9;
	6057	1	47.65	19.4	0	0.79	0	1.006588	-1.618004	150	8	1.1	0.9;
	6058	1	0	-0	0	1.73	0	1.034129	9.64509	110	8	1.1	0.9;
	6060	1	0	0	0	80.66	0	1.076909	-37.809268	380	1	1.1	0.9;
	6068	1	0	0	0	33.08	0	1.032486	13.237978	380	8	1.1	0.9;
	6069	1	-0	-0	0.285753	52.671068	0	0.988415	-6.111052	380	2	1.1	0.9;
	6071	1	163.5	35.6	0	4.59	0	1.033326	-16.372144	220	5	1.1	0.9;
	6080	1	0	0	0.361	33.735	0	1.057151	-42.098594	380	10	1.1	0.9;
	6090	2	0	0	0	0	0	1.032192	-38.570202	220	4	1.1	0.9;
	6092	2	0	0	0	0	0	1.03946	-39.302053	380	4	1.1	0.9;
	6093	1	21.02	12.4	0	2.96	0	1.008226	-40.430643	220	4	1.1	0.9;
	6101	1	103	40.2	0	1.13	0	1.029178	-10.823669	220	5	1.1	0.9;
	6104	1	93.12	4.99	0	0.18	0	1.052842	-12.128098	220	5	1.1	0.9;
	6110	1	78.59	16	0	1.87	0	1.055925	-36.011444	220	5	1.1	0.9;
	6112	1	45.28	15.27	0	7.88	0	1.025898	-9.276213	220	5	1.1	0.9;
	6114	1	0	0	0	8.78	0	1.04901	-10.305588	220	5	1.1	0.9;
	6115	1	-0	-0	0	1.18	0	1.031059	-11.285664	220	5	1.1	0.9;
	6118	1	82.55	30.58	0	4.26	0	1.027942	26.269973	220	8	1.1	0.9;
	6119	1	221.25	72.56	0	8.23	0	1.039138	-15.987796	220	5	1.1	0.9;
	6127	1	-0.1	68.5	0	17.24	0	1.053195	-46.799005	220	4	1.1	0.9;
	6131	2	0	0	0	0	0	1.141159	19.233339	150	8	1.3	0.7;
	6132	1	60.67	5.9	0	5.32	0	1.013934	-47.518736	220	4	1.1	0.9;
	6134	1	42.55	6.5	0	3.7	0	1.037631	-52.750478	220	4	1.1	0.9;
	6135	1	2.2	0.5	0	0.5	0	0.969407	-55.585275	150	10	1.1	0.9;
	6138	2	0	0	0	0	0	1.013527	-43.940748	220	4	1.1	0.9;
	6139	1	76.89	35.1	0	1.9	0	1.053454	1.076127	150	8	1.1	0.9;
	6145	2	0	0	0	0	0	1.021227	-32.669285	220	4	1.1	0.9;
	6146	1	0	-0	0	3.15	0	1.027242	-12.087947	220	5	1.1	0.9;
	6149	1	39.05	24.8	0	7.47	0	1.012511	16.94653	150	8	1.1	0.9;
	6151	1	60.67	33.2	0	0.31	0	1.031952	-24.043559	220	5	1.1	0.9;
	6153	2	0	0	0	0	0	0.996786	-14.746886	380	5	1.1	0.9;
	6163	1	0	-0	0	4.96	0	1.03567	-21.986931	220	5	1.1	0.9;
	6167	1	152.6	61.8	0	2.35	0	1.015341	10.912565	150	8	1.1	0.9;
	6168	2	0	0	0	0	0	1.043938	-13.377168	220	5	1.1	0.9;
	6174	1	289.1	84.2	0	17.34	0	1.010663	-47.870802	220	4	1.1	0.9;
	6178	1	91	28.4	0	10.57	0	1.039445	-24.646301	220	5	1.1	0.9;
	6179	2	0	0	0	0	0	1.038661	-39.189548	220	4	1.1	0.9;
	6183	1	59.37	12.7	0	22.18	0	1.032969	-53.197394	220	4	1.1	0.9;
	6186	1	63.07	13.4	0	15.52	0	1.032387	-53.394752	220	4	1.1	0.9;
	6187	1	-0	0	0.361	25.935	0	1.054472	-41.006336	380	10	1.1	0.9;
	6188	1	209.9	38	0	5.26	0	0.985483	-50.582871	150	10	1.1	0.9;
	6192	1	50.56	5.9	0	10.58	0	1.02478	-42.750534	220	4	1.1	0.9;
	6194	1	0	-0	0	3.14	0	1.024716	-16.334779	220	5	1.1	0.9;
	6199	1	-4.82	-0.88	0	224.85	0	1.070633	-12.60189	380	5	1.1	0.9;
	6202	1	83.79	33.1	0	65.16	0	1.034847	3.977497	150	8	1.1	0.9;
	6203	1	14.02	5.6	0	2.34	0	1.069499	-27.392683	220	5	1.1	0.9;
	6206	1	-7.62	-2.52	0	33.197994	0	1.038423	0.045008	380	5	1.1	0.9;
	6209	1	69.48	19.7	0	22.1	0	1.024252	-48.184075	220	4	1.1	0.9;
	6218	1	88.8	35.1	0	1.24	0	1.01746	4.88991	150	8	1.1	0.9;
	6219	1	-0	-0	0	0.22	0	1.034732	-12.753159	220	5	1.1	0.9;
	6220	1	231.9	27.6	0	0.36	0	1.053799	-22.955261	220	5	1.1	0.9;
	6224	1	0	0	0	58.990785	0	0.998228	-12.912978	380	5	1.1	0.9;
	6231	1	-0	0	0	11.5	0	1.036226	-8.008108	220	5	1.1	0.9;
	6232	1	49.25	12	0	7.28	0	1.041395	-25.840003	220	5	1.1	0.9;
	6233	2	0	0	0	0	0	1.018011	6.454792	380	2	1.1	0.9;
	6239	2	0	0	0	0	0	1.029262	-44.679169	220	4	1.1	0.9;
	6240	1	0	-0	0	6.31	0	1.07487	-34.305726	220	5	1.1	0.9;
	6243	1	88.41	34.42	0	3.45	0	1.037026	-43.881058	220	4	1.1	0.9;
	6244	1	0	-0	0	2.44	0	0.981933	-49.334169	220	4	1.1	0.9;
	6246	1	-0	0	0	86.73	0	1.024401	-37.00382	380	5	1.1	0.9;
	6252	1	164.7	44.5	0	3.67	0	1.01931	-29.984607	220	5	1.1	0.9;
	6253	1	128.8	29.3	0	26.74	0	1.041636	-25.798971	220	5	1.1	0.9;
	6254	1	0	-0	0	9.9	0	0.990568	-54.325821	220	10	1.1	0.9;
	6261	1	75.88	11.2	0	4.58	0	1.028006	-54.171426	220	4	1.1	0.9;
	6263	1	0	-0	0	42.81	0	1.032886	13.276399	380	8	1.1	0.9;
	6267	1	91.3	18.1	0	3.05	0	1.040407	-24.438492	220	5	1.1	0.9;
	6271	1	3.7	2.6	0	4.45	0	1.053813	-20.132206	220	5	1.1	0.9;
	6275	1	122.8	37.2	0	37.13	0	0.997352	-4.846195	150	8	1.1	0.9;
	6280	2	0	0	0	0	0	1.019445	-33.127661	220	10	1.1	0.9;
	6282	1	-0.28	-5.54	0	0	0	1.0472	26.596881	220	8	1.1	0.9;
	6283	1	0	-0	0	18.64	0	1.042145	16.009807	380	8	1.1	0.9;
	6288	1	25.23	10.3	0	2.69	0	1.004251	-3.669685	150	8	1.1	0.9;
	6290	1	47.65	20.4	0	3.63	0	1.058239	-12.880479	220	5	1.1	0.9;
	6291	2	0	0	0	0	0	1.039201	-23.999248	220	5	1.1	0.9;
	6292	1	186.5	49.9	0	13.17	0	1.042112	-48.273587	220	4	1.1	0.9;
	6293	1	-0	0	0	33.66	0	1.004441	-4.400608	380	2	1.1	0.9;
	6295	1	0	-0	0	46.83	0	1.048295	9.044501	150	8	1.1	0.9;
	6298	1	0	-0	0.274712	52.929035	0	1.051189	-47.906417	380	10	1.1	0.9;
	6299	1	-0	-0	0	2.37	0	1.050975	-44.42578	220	4	1.1	0.9;
	6306	1	0	-0	0	2.13	0	1.076004	-24.327992	220	5	1.1	0.9;
	6308	1	0	-0	0	0.05	0	1.043125	-25.590237	220	5	1.1	0.9;
	6311	1	0	-0	0	2.05	0	0.97235	-42.823481	220	10	1.1	0.9;
	6313	1	48.95	14.9	0	4.04	0	1.033328	-23.945702	220	5	1.1	0.9;
	6315	1	95.71	19.6	0	20.2	0	0.988448	-42.71208	220	4	1.1	0.9;
	6320	1	-0.2	-1.64	0	0	0	1.031936	4.19826	150	8	1.1	0.9;
	6323	1	116.1	30.2	0	4.38	0	1.012964	2.806367	150	8	1.1	0.9;
	6329	2	0	0	0	0	0	0.994824	-43.877964	150	10	1.1	0.9;
	6331	2	0	0	0	0	0	1.053365	-0.840778	220	5	1.1	0.9;
	6332	2	0	0	0	0	0	1.064292	-11.355997	220	5	1.1	0.9;
	6333	1	42.75	4	0	5.41	0	1.010821	-40.748261	220	4	1.1	0.9;
	6337	1	0	0	0	8.26	0	1.088467	-9.653232	380	5	1.1	0.9;
	6340	1	-0	0	0	2.27	0	1.056921	-45.124625	380	10	1.1	0.9;
	6342	1	-0	0	0	3.25	0	1.040221	17.9716	110	8	1.1	0.9;
	6344	1	-0	0	0	0.27	0	1.022768	-43.695178	220	4	1.1	0.9;
	6345	1	0	0	0	0	0	0.99741	8.692009	380	8	1.1	0.9;
	6349	1	55.46	14.1	0	0.92	0	1.022378	-48.156496	220	4	1.1	0.9;
	6351	2	0	0	0	0	0	1.059908	-16.273161	220	5	1.1	0.9;
	6354	1	0	-0	0	47.06	0	1.038618	6.820334	150	8	1.1	0.9;
	6356	1	44.95	12.8	0	2.61	0	1.019694	7.425328	150	8	1.1	0.9;
	6357	1	430.8	107.9	0	11.64	0	0.98572	-24.821725	220	5	1.1	0.9;
	6359	1	74.68	10.8	0	5.82	0	1.029132	-53.622036	220	4	1.1	0.9;
	6368	2	0	0	0	0	0	1.046268	-8.318057	220	5	1.1	0.9;
	6371	1	-0	-0	0	0.35	0	1.047834	-0.838701	220	5	1.1	0.9;
	6376	2	0	0	0	0	0	1.047648	-0.304942	220	5	1.1	0.9;
	6381	1	246.2	119.1	0	31.72	0	0.987777	-56.008855	220	10	1.1	0.9;
	6382	1	75.43	10.49	0	7.91	0	1.043547	-9.673367	220	5	1.1	0.9;
	6383	1	15.62	5.3	0	2.17	0	1.019759	23.115557	110	8	1.1	0.9;
	6384	1	159.9	47.4	0	3.65	0	1.044325	-1.646281	220	5	1.1	0.9;
	6390	1	-0	0	0	3.36	0	1.064469	-47.252752	380	10	1.1	0.9;
	6393	1	93.7	29.9	0	2.1	0	1.013488	7.50085	110	8	1.1	0.9;
	6394	1	53.86	-15.3	0	2.87	0	1.021951	-45.870021	220	4	1.1	0.9;
	6395	1	154.7	48.2	0	5.49	0	1.027632	8.567013	150	8	1.1	0.9;
	6397	1	0	0	0	27.4	0	1.027573	-41.568957	220	4	1.1	0.9;
	6399	1	44.05	18	0	0.88	0	1.00669	-1.272997	150	8	1.1	0.9;
	6405	1	76.48	19.8	0	5.41	0	1.018638	-16.011287	220	5	1.1	0.9;
	6415	2	0	0	0	0	0	0.987105	-38.038845	220	10	1.1	0.9;
	6416	1	67.67	16.4	0	3.7	0	1.046481	-9.558553	220	5	1.1	0.9;
	6423	1	-0	0	0	2.26	0	0.997882	-4.642958	150	8	1.1	0.9;
	6424	1	-0.18	-2.28	0	0	0	1.050218	18.743357	380	8	1.1	0.9;
	6426	1	10.21	3.8	0	4.6	0	1.054575	1.539515	220	5	1.1	0.9;
	6427	1	83.19	22.9	0	28.5	0	1.004268	-13.142695	380	5	1.1	0.9;
	6429	2	0	0	0	0	0	1.03915	-18.370028	220	5	1.1	0.9;
	6430	1	0	-0	0	6.03	0	1.05085	-28.971431	220	5	1.1	0.9;
	6436	2	0	0	0	0	0	1.01651	-47.818377	220	4	1.1	0.9;
	6439	1	0	0	0	0	0	1.014417	5.816488	150	8	1.1	0.9;
	6442	1	163.3	70.3	0	5.2	0	1.029153	12.245807	150	8	1.1	0.9;
	6443	1	43.95	15.5	0	6.36	0	1.029433	-49.112793	220	4	1.1	0.9;
	6450	1	5.51	2.1	0	1.92	0	1.01178	-10.898649	220	5	1.1	0.9;
	6454	1	-0	-0	0	11.67	0	1.029382	-53.582851	220	4	1.1	0.9;
	6455	1	20.52	13.3	0	0.17	0	1.038025	-22.084544	220	5	1.1	0.9;
	6456	1	159.4	28.4	0	15.17	0	1.023228	-46.686932	220	4	1.1	0.9;
	6459	2	0	0	0	0	0	1.006942	-37.967795	150	10	1.1	0.9;
	6472	1	0	-0	0	9.93	0	1.042588	-24.033565	220	5	1.1	0.9;
	6474	2	0	0	0	0	0	1.062877	-7.838842	220	5	1.1	0.9;
	6475	1	-0	0	0	149.047157	0	1.0259	-8.704262	380	5	1.1	0.9;
	6477	1	0	0	0	1.99	0	1.03311	21.518539	220	8	1.1	0.9;
	6478	1	141.2	34.8	0	2.37	0	1.010975	-14.403642	220	5	1.1	0.9;
	6480	1	-0.21	-1.04	0	0	0	1.016764	2.82735	150	8	1.1	0.9;
	6484	2	0	0	0	0	0	1.030805	-45.317162	380	4	1.1	0.9;
	6486	1	31.74	17.4	0	0.89	0	1.043874	-3.962885	220	5	1.1	0.9;
	6489	1	-0.12	-1.74	0	0	0	1.087601	32.323601	380	8	1.1	0.9;
	6490	1	84.59	12.9	0	3.66	0	1.026525	-52.273036	220	4	1.1	0.9;
	6494	1	0	-0	0	8.2	0	1.058484	-41.333189	380	1	1.1	0.9;
	6495	1	49.66	12.3	0	2.2	0	1.042112	-21.799784	220	5	1.1	0.9;
	6496	1	0	-0	0	24.38	0	1.061179	-40.943297	380	4	1.1	0.9;
	6499	1	-0	-0	0	13.86	0	1.050329	-44.72942	220	4	1.1	0.9;
	6504	1	49.26	30.9	0.097351	12.97086	0	1.005937	-44.477592	220	10	1.1	0.9;
	6510	1	88.2	26.7	0	8.91	0	1.050373	-26.443611	220	5	1.1	0.9;
	6512	1	-0	0	0	3.4	0	0.994548	-53.708793	150	10	1.1	0.9;
	6514	1	-0.18	-1	0	0	0	1.04065	6.607097	150	8	1.1	0.9;
	6516	2	0	0	0	0	0	1.047972	-3.562112	220	5	1.1	0.9;
	6519	1	-0.19	-1.94	0	0	0	1.024002	11.729825	150	8	1.1	0.9;
	6521	1	0	0	0	2.3	0	1.041998	-4.708978	220	5	1.1	0.9;
	6526	1	8.41	2.6	0	1.77	0	1.022102	13.184764	110	8	1.1	0.9;
	6532	1	120.69	18.1	0	15.46	0	1.071832	-31.431415	220	5	1.1	0.9;
	6539	1	68.48	22.4	0	4.38	0	1.011836	1.010536	150	8	1.1	0.9;
	6540	1	0.9	0.3	0	0.12	0	0.994058	-36.187051	150	10	1.1	0.9;
	6542	1	336.7	-115.3	0	0.78	0	1.040374	-2.397063	220	2	1.1	0.9;
	6543	1	118.4	7.1	0	11.11	0	1.024139	-48.338822	220	4	1.1	0.9;
	6544	1	62.17	0.7	0	5.98	0	1.016797	-42.419225	220	4	1.1	0.9;
	6552	2	0	0	0	0	0	1.048072	-21.173134	220	5	1.1	0.9;
	6555	1	-37.81	-21.64	0	83.362314	0	1.008204	-43.770916	380	5	1.1	0.9;
	6556	1	-0	-0	0	4.61	0	1.047939	-12.734202	220	5	1.1	0.9;
	6559	1	105.5	35.3	0	1.23	0	1.006915	-1.481119	150	8	1.1	0.9;
	6561	1	216	56.2	0	6.98	0	1.006793	-48.872369	220	4	1.1	0.9;
	6563	1	0	-0	0	8.94	0	1.039818	-24.693842	220	5	1.1	0.9;
	6564	1	328.9	-9.3	0	28.37	0	1.049439	-43.12438	380	4	1.1	0.9;
	6565	1	0	-0	0	4.05	0	1.032962	-7.256428	220	5	1.1	0.9;
	6570	1	-0	0	0	32.43	0	1.044265	-20.80039	220	5	1.1	0.9;
	6581	1	153.8	23.4	0	87.73	0	1.043213	-2.471665	380	5	1.1	0.9;
	6582	1	0	-0	0	10.15	0	1.030597	12.371566	150	8	1.1	0.9;
	6596	1	109.1	15	0	8.82	0	1.004701	-54.051001	220	4	1.1	0.9;
	6597	1	52.36	7.4	0	26.39	0	1.025543	-48.149429	220	4	1.1	0.9;
	6600	1	25.83	7.2	0	0.88	0	1.05661	12.502407	110	8	1.1	0.9;
	6604	1	15.22	7.2	0	1.81	0	1.017766	-39.480036	220	4	1.1	0.9;
	6607	1	0	0	0	40.86	0	1.055138	-41.609574	380	4	1.1	0.9;
	6612	1	41.14	15.25	0	3.96	0	1.042708	-31.932488	220	5	1.1	0.9;
	6616	1	-0	-0	0	3.06	0	1.042889	-15.932041	220	5	1.1	0.9;
	6619	1	-0	-0	0	10.67	0	1.045751	-15.459446	220	5	1.1	0.9;
	6624	1	-37.33	-11.88	0	145.18	0	1.058817	4.267441	380	5	1.1	0.9;
	6625	1	18.82	1.7	0	6.24	0	1.025446	-56.568733	220	4	1.1	0.9;
	6629	1	62.63	16.3	0	2.03	0	1.056213	-11.409726	220	5	1.1	0.9;
	6630	1	134.1	24.8	0	13.73	0	1.038952	-24.182883	220	5	1.1	0.9;
	6632	2	0	0	0	0	0	1.062949	-40.452596	380	4	1.1	0.9;
	6636	1	-0	0	0	0.31	0	1.039939	0.565316	220	5	1.1	0.9;
	6638	1	61.65	10.24	0	3.48	0	1.022691	-14.912865	220	5	1.1	0.9;
	6639	1	388.04	52.61	0	2.87	0	1.04321	-2.440722	220	5	1.1	0.9;
	6648	1	-26.03	-0.84	0	0.1	0	1.063897	-11.502112	220	5	1.1	0.9;
	6654	2	0	0	0	0	0	1.02462	-46.240582	220	4	1.1	0.9;
	6656	2	0	0	0	0	0	1.022999	-36.884898	220	4	1.1	0.9;
	6659	1	199.6	27.1	0	33.58	0	1.030902	-44.811571	380	4	1.1	0.9;
	6660	1	435.2	1.3	0	3.25	0	1.012984	-48.913886	220	4	1.1	0.9;
	6662	1	-0.2	-1.64	0	0	0	1.031936	4.19826	150	8	1.1	0.9;
	6663	1	74.58	16.6	0	2.02	0	1.015493	-48.055478	220	4	1.1	0.9;
	6664	1	-0	0	0	10.68	0	1.068966	-46.374278	220	5	1.1	0.9;
	6669	1	132.9	13.6	0	3.44	0	1.014034	-52.545603	220	4	1.1	0.9;
	6675	1	37.94	7.5	0	11.89	0	1.035153	-23.318949	220	5	1.1	0.9;
	6679	1	0	-0	0	2.48	0	1.033934	-53.241623	220	4	1.1	0.9;
	6681	2	0	0	0	0	0	1.07064	-40.401559	380	4	1.1	0.9;
	6684	1	-0	0	0	9.84	0	1.043124	-25.590229	220	5	1.1	0.9;
	6687	1	-0	0	0	0	0	1.050008	18.737299	380	8	1.1	0.9;
	6691	1	42.55	12.5	0	1.26	0	1.083964	-34.13205	220	5	1.1	0.9;
	6692	1	10.33	3.3	0	4.65	0	1.022548	-15.5869	220	5	1.1	0.9;
	6697	1	309	74.5	0	3.69	0	1.045362	-37.168777	220	5	1.1	0.9;
	6704	1	-0	-0	0.211099	0.672832	0	0.995487	-3.368831	380	2	1.1	0.9;
	6707	1	119.5	44.5	0	3.61	0	1.02112	11.417687	150	8	1.1	0.9;
	6714	1	-0	-0	0	0.3	0	1.047832	-0.838669	220	5	1.1	0.9;
	6717	1	-20.82	-8.8	0	9.24	0	1.022189	-45.90348	220	4	1.1	0.9;
	6723	1	0	0	0	1.58	0	1.035057	-9.440724	220	5	1.1	0.9;
	6730	1	-0	0	0	2.18	0	1.037253	-21.115037	220	5	1.1	0.9;
	6734	2	0	0	0	0	0	1.05765	-3.4334	220	5	1.1	0.9;
	6735	1	-0	-0	0	0.4	0	1.037632	-1.22486	380	5	1.1	0.9;
	6737	1	0	-0	0	9.74	0	1.022754	-10.499118	220	5	1.1	0.9;
	6738	1	-4.91	-2.53	0	120.576812	0	1.028082	-24.822962	380	5	1.1	0.9;
	6742	1	137.4	23.9	0	7.26	0	1.037351	-21.529801	220	5	1.1	0.9;
	6744	1	244.12	95.29	0	7.55	0	1.035884	-10.189404	220	5	1.1	0.9;
	6745	1	25.63	51.6	0	2.79	0	1.016392	-39.412948	220	4	1.1	0.9;
	6747	1	0	-0	0	75	0	1.024439	2.725414	150	8	1.1	0.9;
	6749	1	5.91	1.7	0	8.75	0	1.029371	-48.409878	220	4	1.1	0.9;
	6757	1	0	0	0	6.57	0	1.038766	-1.17374	380	5	1.1	0.9;
	6760	1	4.21	1.2	0	30.55	0	1.051203	-42.103673	380	10	1.1	0.9;
	6763	1	0	-0	0	5.46	0	1.043541	-24.265972	220	5	1.1	0.9;
	6766	1	-0	0	0	5.12	0	1.024741	11.926408	380	8	1.1	0.9;
	6768	1	27.53	9.9	0	85.25	0	1.031252	-45.003965	380	4	1.1	0.9;
	6769	1	99	23.2	0	10.16	0	1.045659	-44.124314	220	4	1.1	0.9;
	6771	1	-0	0	0	0.34	0	1.051704	-43.319399	380	4	1.1	0.9;
	6772	1	61.37	23.2	0	0.22	0	1.045221	-24.126683	220	5	1.1	0.9;
	6773	1	0	-0	0	2.18	0	1.041185	-24.317372	220	5	1.1	0.9;
	6783	1	140.1	25.5	0	10.78	0	1.029053	-53.638737	220	4	1.1	0.9;
	6785	1	0	0	0	8.9	0	1.045887	-0.709285	220	5	1.1	0.9;
	6791	1	46.35	9.1	0	18.2	0	1.045382	-25.270988	220	5	1.1	0.9;
	6795	1	-73.98	-29.8	0	17.77	0	0.999791	-48.960893	220	10	1.1	0.9;
	6797	1	42.25	5.9	0	13.07	0	1.027612	-50.167833	220	4	1.1	0.9;
	6798	2	0	0	0	0	0	1.019012	6.75387	380	2	1.1	0.9;
	6802	1	-0	0	0	7.43	0	1.038702	-6.086871	220	5	1.1	0.9;
	6806	1	135.84	15.97	0	4.57	0	1.045649	-21.295765	220	5	1.1	0.9;
	6807	2	0	0	0	0	0	1.012126	-16.436745	220	5	1.1	0.9;
	6816	2	0	0	0	0	0	1.055666	-24.275709	220	5	1.1	0.9;
	6820	2	0	0	0	0	0	1.051187	-6.633991	220	5	1.1	0.9;
	6823	1	36.84	14.4	0	5.09	0	1.034323	3.948895	150	8	1.1	0.9;
	6824	1	68.78	28.2	0	0.26	0	0.994435	-4.797866	150	8	1.1	0.9;
	6826	1	326.9	-82.3	0	0	0	1.067413	-7.63246	150	2	1.1	0.9;
	6828	1	58.16	16.5	0	6.24	0	1.040297	-13.431461	220	5	1.1	0.9;
	6831	2	0	0	0	0	0	1.034052	-9.760419	220	5	1.1	0.9;
	6832	1	-0.12	-1.74	0	0	0	1.08761	32.323642	380	8	1.1	0.9;
	6833	1	0	0	0.246363	4.841277	0	0.979265	-6.81498	380	2	1.1	0.9;
	6835	1	-0.15	-1.93	0	0	0	1.028044	18.540709	220	8	1.1	0.9;
	6836	1	184.88	-21.11	0	39.26	0	1.024919	-48.168635	220	4	1.1	0.9;
	6837	1	47.75	11	0	0.05	0	1.052628	-7.891626	220	5	1.1	0.9;
	6838	2	0	0	0	0	0	1.043313	5.369918	150	8	1.1	0.9;
	6841	1	122.8	20.8	0	0.47	0	1.030565	-53.139202	220	4	1.1	0.9;
	6842	1	98	21.7	0	6	0	1.040109	-8.421644	220	5	1.1	0.9;
	6844	1	-0	-0	0	1.04	0	1.015901	-29.315974	220	5	1.1	0.9;
	6845	2	0	0	0	0	0	1.051715	-11.1809	220	5	1.1	0.9;
	6846	1	44.35	19.3	0	3.25	0	1.049248	-4.58634	220	5	1.1	0.9;
	6852	2	0	0	0	0	0	1.032911	-19.833765	220	5	1.1	0.9;
	6854	1	0	0	0	0	0	1.056677	-11.352791	220	5	1.1	0.9;
	6857	2	0	0	0	0	0	1.014858	-8.546516	380	5	1.1	0.9;
	6869	1	54.9	-96.8	0	16.42	0	1.027632	-52.767366	220	4	1.1	0.9;
	6877	2	0	0	0	0	0	1.066537	-41.619855	380	4	1.1	0.9;
	6879	1	199.6	24.4	0	10.6	0	1.043211	-45.146735	220	4	1.1	0.9;
	6880	1	0	0	0	1.36	0	1.040453	-4.439381	220	5	1.1	0.9;
	6882	1	67.47	6.1	0	1.46	0	1.016215	-53.732109	220	4	1.1	0.9;
	6883	1	0	-0	0	12.4	0	1.059988	-39.977279	380	1	1.1	0.9;
	6884	1	-0.12	-1.31	0	0	0	1.087731	32.319037	380	8	1.1	0.9;
	6887	1	-0	-0	0	3.13	0	1.04725	-3.093328	220	5	1.1	0.9;
	6888	2	0	0	0	0	0	1.042814	-13.607804	220	5	1.1	0.9;
	6889	1	-12.49	-5.83	0	6.235224	0	1.032371	2.535951	380	5	1.1	0.9;
	6891	1	135.61	26.86	0	1.93	0	1.043146	-14.346931	220	5	1.1	0.9;
	6897	1	176.8	39.4	0	28.24	0	1.052772	-16.564834	380	5	1.1	0.9;
	6901	1	-37.84	-19.23	0	28.140258	0	1.010023	-32.207211	380	5	1.1	0.9;
	6904	1	0	0	0	6.97	0	1.030983	-55.55932	220	4	1.1	0.9;
	6908	1	0	0	0	3.01	0	1.03639	-40.192059	220	5	1.1	0.9;
	6909	1	122.5	8.7	0	15.85	0	1.035956	-23.157555	220	5	1.1	0.9;
	6915	1	11.71	3.7	0	0.8	0	1.02472	14.046055	110	8	1.1	0.9;
	6917	1	-0	-0	0	96.89	0	1.06608	-42.268714	380	4	1.1	0.9;
	6921	1	-37.34	-13.33	0	251.529665	0	1.033994	-1.605149	380	5	1.1	0.9;
	6922	1	113.6	44.7	0	1.91	0	1.038591	-8.601875	220	5	1.1	0.9;
	6923	1	33.34	8.6	0	3.93	0	1.037346	15.355591	110	8	1.1	0.9;
	6926	1	124.5	101.4	0	0.53	0	1.029554	-12.429854	220	5	1.1	0.9;
	6932	1	117	20.4	0	22.34	0	1.033352	-53.101088	220	4	1.1	0.9;
	6936	2	0	0	0	0	0	1.030076	-38.53627	220	4	1.1	0.9;
	6940	1	0	-0	0	10.46	0	1.054737	-27.625806	220	5	1.1	0.9;
	6947	2	0	0	0	0	0	1.046726	-3.720941	220	5	1.1	0.9;
	6949	1	35.44	5.3	0	2.24	0	1.029331	-53.544619	220	4	1.1	0.9;
	6952	1	180.9	33.2	0	29.17	0	1.073657	-25.19061	220	5	1.1	0.9;
	6953	1	61.77	10.2	0	4.01	0	1.021083	-49.70633	220	4	1.1	0.9;
	6954	1	18.12	10	0	1.03	0	1.033292	-20.085167	220	5	1.1	0.9;
	6960	1	389.8	-68.1	0	5.03	0	0.991637	-55.881273	220	10	1.1	0.9;
	6961	1	56.56	18.4	0	3.29	0	1.047236	-0.037249	220	5	1.1	0.9;
	6964	1	120.1	2.5	0	6.6	0	1.040388	-44.750394	220	4	1.1	0.9;
	6969	2	0	0	0	0	0	1.061428	-23.28797	220	5	1.1	0.9;
	6970	1	0	0	0	27.8	0	1.028216	-46.031374	380	4	1.1	0.9;
	6982	2	0	0	0	0	0	1.038963	-20.935757	220	5	1.1	0.9;
	6985	1	125.2	1.2	0	15.12	0	1.029081	-53.680846	220	4	1.1	0.9;
	6987	1	112.1	36.3	0	6.4	0	1.028017	3.250246	150	8	1.1	0.9;
	6989	1	0	0	0	1.91	0	1.046097	-24.035778	220	5	1.1	0.9;
	6990	1	-0	0	0	4.54	0	1.047903	-24.132172	220	5	1.1	0.9;
	6994	1	0	0	0	11.7	0	1.020708	-40.858026	220	4	1.1	0.9;
	6998	1	30.03	8.6	0	21.45	0	1.023626	8.631158	150	8	1.1	0.9;
	7002	1	51.16	12	0	6.02	0	1.033942	-41.747854	220	4	1.1	0.9;
	7004	1	0	0	0	0	0	1.049776	18.741809	380	1	1.1	0.9;
	7011	1	84.59	21.7	0	7.42	0	1.023859	8.520642	110	8	1.1	0.9;
	7014	1	0	0	0	0.3	0	1.034719	-2.558654	220	5	1.1	0.9;
	7018	1	53.26	29.5	0	41.35	0	1.023469	11.73541	150	8	1.1	0.9;
	7019	1	52.56	12.8	0	2.87	0	1.041053	-25.864067	220	5	1.1	0.9;
	7021	1	226.8	53.3	0	3.26	0	1.008138	-17.163471	220	5	1.1	0.9;
	7029	2	0	0	0	0	0	1.063951	-36.875689	380	4	1.1	0.9;
	7030	1	0	0	0	0.19	0	1.062881	-7.838873	220	5	1.1	0.9;
	7036	2	0	0	0	0	0	1.042173	-12.578177	220	5	1.1	0.9;
	7037	1	-0	-0	0	2.77	0	0.99153	-55.40565	220	10	1.1	0.9;
	7041	1	-175.8	-19.6	0	14.58	0	1.014317	-41.763015	220	10	1.1	0.9;
	7042	1	220.98	91	0	1.16	0	1.052154	-30.14741	220	5	1.1	0.9;
	7047	1	339.9	70.5	0	28.91	0	1.035936	-21.928794	220	5	1.1	0.9;
	7049	2	0	0	0	0	0	1.04735	-14.367804	220	5	1.1	0.9;
	7050	1	120	26	0	22.01	0	1.036775	-21.275559	220	5	1.1	0.9;
	7051	1	0	0	0	21.19	0	0.995283	-0.23205	380	2	1.1	0.9;
	7052	1	111	36.6	0	2.21	0	1.044853	-12.404558	220	5	1.1	0.9;
	7053	1	-0	0	0	34.13	0	1.067028	-40.860634	380	4	1.1	0.9;
	7056	2	0	0	0	0	0	1.005566	-4.277241	380	5	1.1	0.9;
	7058	2	0	0	0	0	0	1.043279	-36.396436	220	4	1.1	0.9;
	7059	1	28.93	8.3	0	0.42	0	1.024187	9.419858	150	8	1.1	0.9;
	7064	1	-0	0	0	1.02	0	1.014688	-46.559002	220	4	1.1	0.9;
	7069	1	258.79	42.82	0	9.28	0	1.041839	-7.978812	220	5	1.1	0.9;
	7070	1	142.9	25.8	0	24.18	0	1.047318	-13.451419	220	5	1.1	0.9;
	7073	1	-0	-0	0	15.47	0	1.047344	9.041554	150	8	1.1	0.9;
	7075	2	0	0	0	0	0	1.026663	-42.497913	220	4	1.1	0.9;
	7076	1	0	-0	0	0.36	0	1.033945	-12.41346	220	5	1.1	0.9;
	7079	1	-0	0	0	2.2	0	1.041573	-51.256311	380	10	1.1	0.9;
	7090	1	82.89	41.3	0	36.2	0	1.031888	10.922009	150	8	1.1	0.9;
	7092	1	44.65	8.8	0	20.17	0	1.046583	-25.141996	220	5	1.1	0.9;
	7096	1	-0	-0	0	57.66	0	1.070091	-40.307397	380	4	1.1	0.9;
	7098	1	0	-0	0	7.17	0	1.044076	-2.313897	220	5	1.1	0.9;
	7104	2	0	0	0	0	0	1.004388	-28.478299	220	10	1.1	0.9;
	7110	1	-0	0	0	0	0	1.024139	-41.630598	220	4	1.1	0.9;
	7115	2	0	0	0	0	0	1.014496	-16.107551	220	5	1.1	0.9;
	7117	1	0	-0.6	0	0.5	0	1.041082	27.11571	110	8	1.1	0.9;
	7119	1	149.53	34.9	0	8.9	0	1.013198	-16.311742	220	5	1.1	0.9;
	7124	1	-0	0	0	16.73	0	1.07108	-27.224252	220	5	1.1	0.9;
	7129	1	33.34	5	0	0.38	0	1.041363	-2.891577	220	5	1.1	0.9;
	7132	1	222	90.5	0	0.16	0	1.020788	-15.829017	220	5	1.1	0.9;
	7133	1	102.3	19	0	1.84	0	1.038192	-14.196483	220	5	1.1	0.9;
	7136	1	29.23	9.4	0	0.68	0	1.01997	7.817073	110	8	1.1	0.9;
	7137	2	0	0	0	0	0	1.031779	-44.572118	380	4	1.1	0.9;
	7138	1	150.2	51.6	0	2.79	0	0.998265	-4.445181	150	8	1.1	0.9;
	7139	1	99.4	22.2	0	2.16	0	1.032109	-53.496089	220	4	1.1	0.9;
	7144	1	8.48	-1.82	0	0.37	0	1.031135	-13.014295	220	5	1.1	0.9;
	7145	1	46.75	24.4	0	3.66	0	1.017901	-57.933705	220	4	1.1	0.9;
	7146	1	0	0	0	0	0	1.087153	32.322685	380	1	1.1	0.9;
	7148	1	-0	-0	0	10.89	0	1.05068	-23.342154	220	5	1.1	0.9;
	7156	1	-0	-0	0	14.82	0	1.047935	8.994444	150	8	1.1	0.9;
	7159	2	0	0	0	0	0	1.059739	4.507306	380	5	1.1	0.9;
	7162	1	-0	0	0	2.48	0	1.047481	-13.453924	220	5	1.1	0.9;
	7163	1	-0	0	0	2.9	0	1.048464	-24.909243	220	5	1.1	0.9;
	7164	1	-18.6	-6.72	0	152.683271	0	1.025988	-22.171787	380	5	1.1	0.9;
	7165	1	-0	-0	0	9.26	0	1.070226	-34.26549	220	5	1.1	0.9;
	7178	1	-0	-0	0	2.45	0	1.012186	-14.284091	220	5	1.1	0.9;
	7180	1	-0	0	0.234234	4.220723	0	0.981876	-6.902546	380	2	1.1	0.9;
	7183	2	0	0	0	0	0	1.105479	-6.84714	380	5	1.3	0.7;
	7191	1	55.46	15.7	0	13.46	0	1.02383	-48.137341	220	4	1.1	0.9;
	7193	1	-0	-0	0	7.17	0	1.019967	-46.478159	220	4	1.1	0.9;
	7195	1	99.9	19.9	0	6.19	0	1.012466	1.222377	150	8	1.1	0.9;
	7196	1	72.68	1.1	0	6.06	0	1.027609	-54.367035	220	4	1.1	0.9;
	7202	1	18.62	6.1	0	2	0	1.046075	-3.910533	220	5	1.1	0.9;
	7209	2	0	0	0	0	0	1.052921	-6.520132	220	5	1.1	0.9;
	7222	1	0	-0	0	5	0	1.064563	-23.838621	220	5	1.1	0.9;
	7226	1	144	31.3	0	3.18	0	1.022853	-15.545301	220	5	1.1	0.9;
	7230	1	-0.18	-1	0	0	0	1.02127	11.418714	150	8	1.1	0.9;
	7231	1	0	0	0	0.05	0	1.041678	-11.508705	220	5	1.1	0.9;
	7235	1	-893.43	260.67	0	0	0	1.039631	39.664851	380	8	1.1	0.9;
	7236	1	0	0	0	1.99	0	0.966439	-43.680938	220	10	1.1	0.9;
	7249	1	37.94	9.8	0	2.26	0	1.029557	-41.944299	220	4	1.1	0.9;
	7252	1	220.8	68.6	0	1.97	0	0.965436	-44.082482	220	10	1.1	0.9;
	7253	1	7.51	3.1	0	0	0	1.040821	-2.447252	220	5	1.1	0.9;
	7256	1	392.25	92.02	0	9.45	0	1.020196	-35.458798	220	5	1.1	0.9;
	7257	2	0	0	0	0	0	1.030375	-46.213157	380	4	1.1	0.9;
	7259	1	82.39	0	0	0.15	0	1.062678	-7.907408	220	5	1.1	0.9;
	7260	2	0	0	0	0	0	1.016183	19.062716	150	8	1.1	0.9;
	7264	1	0	-0	0	3.42	0	1.043585	-10.98095	220	5	1.1	0.9;
	7266	1	-0	-0	0	2.12	0	1.0272	-5.480785	220	5	1.1	0.9;
	7267	2	0	0	0	0	0	1.012113	-7.716675	380	5	1.1	0.9;
	7270	2	0	0	0	0	0	1.06851	26.565652	380	8	1.1	0.9;
	7273	1	219.6	36.4	0	17.89	0	1.042242	-24.355287	220	5	1.1	0.9;
	7274	1	54.93	16.5	0	6.75	0	1.033096	-3.876706	220	5	1.1	0.9;
	7275	1	79.39	31.3	0	2.23	0	1.036244	4.519004	150	8	1.1	0.9;
	7279	2	0	0	0	0	0	1.030612	-13.14884	150	2	1.1	0.9;
	7282	2	0	0	0	0	0	1.029503	-4.625888	380	5	1.1	0.9;
	7283	1	101.1	6.6	0	5.66	0	1.040242	-45.426953	220	4	1.1	0.9;
	7284	1	0	0	0	0	0	1.089967	-0.81866	220	5	1.1	0.9;
	7287	1	0	-0	0	46.28	0	1.064698	-40.954869	380	4	1.1	0.9;
	7289	1	0	-0	0	26.5	0	1.055132	-6.159267	220	5	1.1	0.9;
	7295	1	-0	0	0	2.63	0	1.033606	3.161377	150	8	1.1	0.9;
	7297	1	-0	0	0	2.67	0	1.033594	3.16114	150	8	1.1	0.9;
	7303	2	0	0	0	0	0	1.033047	-42.739399	380	4	1.1	0.9;
	7305	1	82.89	32.6	0	45.84	0	1.036138	4.015909	150	8	1.1	0.9;
	7309	1	0	-0	0	1.32	0	1.056677	-11.352791	220	5	1.1	0.9;
	7310	1	256	69.9	0	43.56	0	1.029455	-45.744562	380	4	1.1	0.9;
	7312	1	56.66	15.6	0	7.45	0	1.00645	-47.823793	220	4	1.1	0.9;
	7316	1	-0	0	0	0.09	0	1.046311	-7.264606	220	5	1.1	0.9;
	7325	1	0	0	0	0.41	0	1.047946	-12.734265	220	5	1.1	0.9;
	7327	2	0	0	0	0	0	1.042085	-18.10373	220	5	1.1	0.9;
	7328	2	0	0	0	0	0	0.994679	-14.546471	380	5	1.1	0.9;
	7329	2	0	0	0	0	0	1.039675	-33.666038	220	10	1.1	0.9;
	7338	1	0	-0	0	2.98	0	1.045961	-14.736111	220	5	1.1	0.9;
	7341	1	37.44	0	0	0.67	0	1.04381	-2.454778	220	5	1.1	0.9;
	7342	1	169	70.8	0	5.23	0	1.019689	-16.094292	220	5	1.1	0.9;
	7343	1	-0	0	0	61.36	0	1.03114	-42.813676	380	4	1.1	0.9;
	7347	1	101.3	17.6	0	0.12	0	1.014859	-48.229058	220	4	1.1	0.9;
	7349	1	75.56	13.04	0	8.77	0	1.021855	-50.750103	220	4	1.1	0.9;
	7351	1	61.67	16.3	0	0.05	0	1.041265	-2.894937	220	5	1.1	0.9;
	7353	1	127.3	26.7	0	7.02	0	0.983185	-22.469243	220	5	1.1	0.9;
	7358	1	-40.95	-28.4	0	7.29	0	1.001857	-48.125059	150	10	1.1	0.9;
	7361	1	-0	-0	0	14.29	0	1.053884	-15.280698	220	5	1.1	0.9;
	7366	1	70.48	22.5	0	4.82	0	1.015027	-40.142163	220	4	1.1	0.9;
	7367	1	1.5	0.4	0	0.15	0	1.028096	-18.923022	220	5	1.1	0.9;
	7373	1	0	-0	0	1.81	0	1.071572	-23.62773	220	5	1.1	0.9;
	7374	1	-0.2	-0.83	0	0	0	1.015892	-0.687464	150	8	1.1	0.9;
	7377	1	0	0	0	0.63	0	1.016564	-16.113305	220	5	1.1	0.9;
	7380	1	105.67	0.56	0	18.05	0	0.990366	-14.587362	380	5	1.1	0.9;
	7388	1	0	0	0	12.79	0	1.087153	32.322685	380	8	1.1	0.9;
	7392	1	142	27.8	0	17.77	0	1.015493	-54.520223	220	4	1.1	0.9;
	7394	1	55.66	22.4	0	2.11	0	1.0354	9.686825	110	8	1.1	0.9;
	7395	1	0	-0	0	87.69	0	1.025209	-43.665105	380	4	1.1	0.9;
	7396	1	-0	0	0	5.68	0	1.02001	-19.556579	220	5	1.1	0.9;
	7398	1	100.3	17.2	0	6.63	0	1.020787	-50.110773	220	4	1.1	0.9;
	7404	1	19.02	5.4	0	0.63	0	1.048858	11.613006	110	8	1.1	0.9;
	7406	1	94.5	-27.8	0	5.52	0	1.027008	-47.350098	220	4	1.1	0.9;
	7407	1	-0.2	-1	0	0	0	1.003253	-3.940799	150	8	1.1	0.9;
	7411	1	-0	0	0	0.68	0	1.001284	-38.553912	150	10	1.1	0.9;
	7413	1	33.64	13.7	0	0.75	0	1.032769	-53.379462	220	4	1.1	0.9;
	7417	1	106.4	19	0	4.65	0	1.041351	-45.36001	220	4	1.1	0.9;
	7422	1	65.67	25.6	0	2.34	0	1.036095	-10.773128	220	5	1.1	0.9;
	7423	1	-0	0	0	7.3	0	1.020318	-40.092191	220	4	1.1	0.9;
	7425	1	-0.2	-1	0	0	0	1.022104	-0.334418	150	8	1.1	0.9;
	7428	1	0	-0	0	120.17	0	1.048433	-49.801196	380	4	1.1	0.9;
	7431	2	0	0	0	0	0	1.01868	-33.26536	220	4	1.1	0.9;
	7436	1	0	-0	0	19.61	0	1.025488	-49.807558	380	4	1.1	0.9;
	7437	1	18.32	5.3	0	7.32	0	1.042931	-25.681594	220	5	1.1	0.9;
	7438	1	121.3	20.7	0	2.39	0	1.036829	-12.73494	220	5	1.1	0.9;
	7442	1	-0	0	0	36.77	0	1.013545	16.971644	150	8	1.1	0.9;
	7446	2	0	0	0	0	0	1.012514	-45.574723	220	4	1.1	0.9;
	7450	1	108.8	28.1	0	30.9	0	1.021372	11.400623	150	8	1.1	0.9;
	7460	1	-0	-0	0	37.99	0	1.044406	-46.009583	380	4	1.1	0.9;
	7463	1	178.7	31.3	0	5.6	0	1.030888	-53.281727	220	4	1.1	0.9;
	7464	1	197.65	28.23	0	5.87	0	1.041982	-12.603915	220	5	1.1	0.9;
	7466	2	0	0	0	0	0	1.057274	-34.947755	220	5	1.1	0.9;
	7471	1	-18.95	-6.93	0	33.11439	0	1.017806	-6.291292	380	5	1.1	0.9;
	7473	1	113.44	71.42	0	11.08	0	1.03919	-15.952793	220	5	1.1	0.9;
	7474	2	0	0	0	0	0	1.028765	-14.261886	220	5	1.1	0.9;
	7475	1	84.79	12.5	0	2.55	0	1.028694	-53.88004	220	4	1.1	0.9;
	7476	1	59.67	27.1	0	43.03	0	1.037734	6.54651	150	8	1.1	0.9;
	7480	1	-0.18	-3.54	0	0	0	1.017111	17.755005	150	8	1.1	0.9;
	7485	1	44.35	10.8	0	8.72	0	1.041059	-25.872286	220	5	1.1	0.9;
	7491	1	0	-0	0	4.29	0	1.012011	-21.498322	220	5	1.1	0.9;
	7495	2	0	0	0	0	0	1.036344	-15.186951	220	5	1.1	0.9;
	7503	1	-0	-0	0	1.86	0	1.000871	-42.759103	220	10	1.1	0.9;
	7504	2	0	0	0	0	0	1.017989	-42.565789	220	4	1.1	0.9;
	7507	1	4	0.8	0	17.65	0	1.04762	-24.128215	220	5	1.1	0.9;
	7508	2	0	0	0	0	0	1.062735	-39.605713	380	4	1.1	0.9;
	7509	1	88	36.2	0	60	0	1.058758	-47.305611	380	10	1.1	0.9;
	7510	1	46.35	14.9	0	20.62	0	1.025667	-46.498034	220	4	1.1	0.9;
	7513	1	0	-0	0	38.549494	0	0.99089	-15.045436	380	5	1.1	0.9;
	7519	1	97.97	31.8	0	5.02	0	1.039225	-10.766503	220	5	1.1	0.9;
	7520	2	0	0	0	0	0	1.033404	-9.153694	380	5	1.1	0.9;
	7522	2	0	0	0	0	0	1.000638	-10.975301	380	5	1.1	0.9;
	7523	1	-0	0	0	2.11	0	1.053262	-0.714256	220	5	1.1	0.9;
	7525	1	-8.66	-6.01	0	0	0	1.020325	17.578915	150	8	1.1	0.9;
	7526	1	0	-0	0	1.88	0	1.007963	6.743891	380	2	1.1	0.9;
	7530	1	0	0	0	31.242294	0	1.045347	-19.834673	380	5	1.1	0.9;
	7537	1	0	0	0	17.184121	0	0.989467	-15.645872	380	5	1.1	0.9;
	7539	1	40.85	15.7	0	2.08	0	1.04992	-12.426239	220	5	1.1	0.9;
	7541	1	346.42	89.04	0	43.55	0	1.053626	-16.488105	380	5	1.1	0.9;
	7556	1	0	-0	0	3.1	0	1.046355	-32.996066	220	1	1.1	0.9;
	7561	1	0	0	0	20.54	0	0.998171	-4.519251	380	1	1.1	0.9;
	7563	1	531.7	-12.8	0	0	0	1.039801	-13.004235	150	2	1.1	0.9;
	7565	1	-0	0	0	0.81	0	1.056993	12.540707	110	8	1.1	0.9;
	7569	1	0	-0	0	1.18	0	0.96962	-55.542078	150	10	1.1	0.9;
	7571	1	-0	-0	0	13.9	0	1.045785	-14.005381	220	5	1.1	0.9;
	7573	1	126.6	28.7	0	15.02	0	1.032069	-53.465727	220	4	1.1	0.9;
	7575	1	29.03	6.1	0	0.44	0	1.042025	10.4044	110	8	1.1	0.9;
	7576	1	0	0	0	0.5	0	1.04098	-12.033553	220	5	1.1	0.9;
	7577	1	0	0	0	6.69	0	1.048853	-10.476048	220	5	1.1	0.9;
	7579	1	136.9	11.4	0	3.05	0	1.041199	-26.697479	220	5	1.1	0.9;
	7580	1	0	0	0	0.17	0	1.053388	-12.006654	220	5	1.1	0.9;
	7582	1	0	-0	0	3.44	0	1.035193	-24.481285	220	5	1.1	0.9;
	7583	1	5.01	1.8	0	3.68	0	1.048949	-0.937766	220	5	1.1	0.9;
	7584	1	127.6	45.3	0	57.43	0	1.025154	-0.871196	150	8	1.1	0.9;
	7586	1	-18.27	-12.99	0	0	0	1.000547	-4.320401	150	8	1.1	0.9;
	7596	1	59.07	23.4	0	21.88	0	1.016631	-0.587949	150	8	1.1	0.9;
	7599	1	0	-0	0	2.13	0	1.038714	-6.087138	220	5	1.1	0.9;
	7600	1	119	31.7	0	1.94	0	1.039856	-45.491051	220	4	1.1	0.9;
	7611	1	115.4	31	0	28.85	0	1.024474	-46.568986	220	4	1.1	0.9;
	7615	1	70.48	17.6	0	3.92	0	0.985979	-53.121078	220	10	1.1	0.9;
	7624	1	32.74	10.5	0	1.78	0	1.01569	-20.927054	220	5	1.1	0.9;
	7626	1	78.89	45.6	0	0.04	0	1.033715	-12.429714	220	5	1.1	0.9;
	7628	2	0	0	0	0	0	1.057002	4.195982	150	8	1.1	0.9;
	7629	1	0	0	0	1.15	0	1.022311	-55.19756	220	4	1.1	0.9;
	7630	2	0	0	0	0	0	1.019127	-44.28077	220	4	1.1	0.9;
	7632	1	37.94	3.4	0	11.17	0	1.028912	-56.022654	220	4	1.1	0.9;
	7635	1	0	0	0	1.86	0	1.052335	-7.868924	220	5	1.1	0.9;
	7637	1	0	0	0.29	6.18	0	0.998893	18.51968	380	2	1.1	0.9;
	7640	1	68.78	35	0	34.52	0	1.033497	-5.993491	380	5	1.1	0.9;
	7641	2	0	0	0	0	0	1.053056	-6.264318	220	5	1.1	0.9;
	7644	1	124.1	-30.3	0	14.26	0	1.041528	-45.624972	220	4	1.1	0.9;
	7647	1	-0	-0	0	0.1	0	1.062641	-20.201545	220	5	1.1	0.9;
	7653	1	33.94	11.4	0	0	0	1.036646	-22.796085	220	5	1.1	0.9;
	7659	2	0	0	0	0	0	1.035071	-40.258623	220	4	1.1	0.9;
	7661	2	0	0	0	0	0	1.039659	-33.672696	220	10	1.1	0.9;
	7662	1	-0.21	-2.37	0	0	0	1.024437	9.444892	150	8	1.1	0.9;
	7663	1	37.14	2.7	0	2.33	0	1.04595	-11.99307	220	5	1.1	0.9;
	7666	1	93.4	27.8	0	29.52	0	1.020727	11.122959	150	8	1.1	0.9;
	7669	1	67.08	25.2	0	8.11	0	1.023999	-48.354165	220	4	1.1	0.9;
	7673	1	20.82	7.9	0	42.36	0	1.047863	-42.64115	380	4	1.1	0.9;
	7675	1	66.07	21.6	0	0.75	0	1.005518	0.486497	150	8	1.1	0.9;
	7685	1	49.45	6.5	0	3.53	0	1.021717	-43.96567	220	4	1.1	0.9;
	7688	1	12.71	5.1	0	15.31	0	1.046502	-49.522132	380	4	1.1	0.9;
	7691	1	0	-0	0	96.48	0	1.027974	-3.930565	380	5	1.1	0.9;
	7694	1	-14.96	-4.7	0	0.04	0	1.028968	-14.823846	220	5	1.1	0.9;
	7697	2	0	0	0	0	0	1.047945	-3.543018	220	5	1.1	0.9;
	7699	1	-0	0	0	19.62	0	1.041938	11.338235	380	1	1.1	0.9;
	7700	1	66.17	27.4	0	14.12	0	1.01995	-10.494298	220	5	1.1	0.9;
	7701	1	11.91	3.5	0	0.43	0	1.016719	6.309825	150	8	1.1	0.9;
	7702	1	-0	-0	0	11.68	0	1.023659	-16.319508	220	5	1.1	0.9;
	7709	1	2.2	0.5	0	0.58	0	1.015574	-41.198324	220	10	1.1	0.9;
	7726	1	0	0	0	2.03	0	1.019838	-10.459707	220	5	1.1	0.9;
	7729	1	20.82	6.7	0	2.25	0	1.021134	8.462881	110	8	1.1	0.9;
	7736	1	0	-0	0	55	0	1.025312	-0.874285	150	8	1.1	0.9;
	7738	1	31.84	9.2	0	5.23	0	1.030225	-21.980925	220	5	1.1	0.9;
	7751	1	0	0	0	0	0	1.008501	-1.308496	150	8	1.1	0.9;
	7752	1	524.5	126	0	7.45	0	1.030082	-18.432528	220	5	1.1	0.9;
	7755	2	0	0	0	0	0	1.06304	-41.883921	380	4	1.1	0.9;
	7761	2	0	0	0	0	0	1.015679	-41.175535	220	10	1.1	0.9;
	7762	1	-0	0	0.304243	48.453399	0	1.011652	-3.587995	380	2	1.1	0.9;
	7766	1	78.89	25.6	0	1.6	0	1.034191	-12.695906	220	5	1.1	0.9;
	7770	1	47.15	15.5	0	9.69	0	1.030464	-44.236875	220	5	1.1	0.9;
	7772	1	223.6	51.8	0	18.45	0	1.054801	-25.696499	220	5	1.1	0.9;
	7775	1	151.5	7.9	0	5	0	1.052625	-28.722171	220	5	1.1	0.9;
	7776	2	0	0	0	0	0	1.068185	6.31233	380	5	1.3	0.7;
	7778	1	153	21.5	0	28.85	0	1.053947	-16.482217	380	5	1.1	0.9;
	7779	2	0	0	0	0	0	1.083645	10.723074	150	8	1.1	0.9;
	7791	1	-0	0	0	1.33	0	1.031135	-12.994154	220	5	1.1	0.9;
	7792	1	0	-0	0	0.46	0	1.008979	-37.341537	150	10	1.1	0.9;
	7795	1	44.35	5.1	0	4	0	1.024664	-46.191032	220	4	1.1	0.9;
	7797	1	-0	-0	0	2.09	0	1.04956	-24.137393	220	5	1.1	0.9;
	7805	1	-0	0	0	30.46	0	1.030402	-53.502885	220	4	1.1	0.9;
	7808	2	0	0	0	0	0	1.055	-0.211716	380	5	1.1	0.9;
	7809	1	169.5	27.7	0	3.04	0	1.028173	-11.501218	220	5	1.1	0.9;
	7810	1	0	-0	0	43	0	1.051434	-49.397778	380	10	1.1	0.9;
	7819	2	0	0	0	0	0	1.020992	-39.959663	220	4	1.1	0.9;
	7820	1	127.2	-10.7	0	7.8	0	1.01943	-54.187383	220	4	1.1	0.9;
	7821	1	179	22.9	0	25.28	0	1.031901	-50.655816	380	4	1.1	0.9;
	7824	1	9.21	3.4	0	2.63	0	1.049112	-6.264138	220	5	1.1	0.9;
	7827	1	137.2	41.9	0.101088	13.207291	0	1.047944	-42.323354	380	10	1.1	0.9;
	7829	1	-3.2	-7.9	0	1.95	0	1.053991	-4.575479	220	2	1.1	0.9;
	7831	1	-17.17	-7.1	0	57.019803	0	1.007389	-11.404525	380	5	1.1	0.9;
	7833	2	0	0	0	0	0	1.02191	-46.781983	220	4	1.1	0.9;
	7834	1	134.7	17.5	0	4.53	0	1.008157	-52.641682	220	4	1.1	0.9;
	7840	1	-0	0	0	2.06	0	1.041176	-2.414918	220	5	1.1	0.9;
	7842	2	0	0	0	0	0	1.005674	-12.461135	380	5	1.1	0.9;
	7844	1	-0	0	0	27.39	0	1.031944	10.92267	150	8	1.1	0.9;
	7845	1	38.64	6	0	3.26	0	1.032324	-53.26234	220	4	1.1	0.9;
	7847	1	0	-0	0	1.55	0	1.042074	-8.892363	220	5	1.1	0.9;
	7852	1	-3.53	-4.62	0	0	0	1.012558	-0.286848	150	8	1.1	0.9;
	7856	1	0	0	0	0	0	1.049571	18.602181	380	8	1.1	0.9;
	7857	1	0	0	0	2.36	0	1.050468	0.178649	220	5	1.1	0.9;
	7860	2	0	0	0	0	0	1.058289	-40.045371	380	10	1.1	0.9;
	7861	1	178.9	9.1	0	15.71	0	1.022328	-42.36437	220	4	1.1	0.9;
	7862	1	0	0	0	9.74	0	1.040044	-21.14272	220	5	1.1	0.9;
	7865	1	84.99	19.6	0	4.32	0	1.0739	-26.397762	220	5	1.1	0.9;
	7869	1	101.9	20	0	22.64	0	1.024599	-46.279386	220	4	1.1	0.9;
	7873	1	-20.94	-6.16	0	11.99	0	1.023316	-15.192522	220	5	1.1	0.9;
	7875	1	-0	0	0	6.51	0	0.988393	-45.410783	220	4	1.1	0.9;
	7877	1	64.21	5.59	0	2.57	0	1.01583	-45.303092	220	4	1.1	0.9;
	7879	1	73.58	48.6	0	2.87	0	0.969444	-55.475945	150	10	1.1	0.9;
	7881	1	-0	-0	0	7.99	0	1.057819	-15.067961	220	5	1.1	0.9;
	7883	1	47.55	17.8	0	0.32	0	1.049592	-28.831267	220	5	1.1	0.9;
	7884	1	35.84	14.1	0	50.11	0	1.037333	4.124709	150	8	1.1	0.9;
	7885	1	142.4	20	0	2.42	0	1.036839	-21.146665	220	5	1.1	0.9;
	7886	1	162.5	42.7	0	12.79	0	1.01825	-19.302256	220	5	1.1	0.9;
	7892	1	36.94	11	0	0.52	0	1.050308	-11.340814	220	5	1.1	0.9;
	7895	1	338.7	52.5	0	11.36	0	1.058928	-19.588442	220	5	1.1	0.9;
	7901	1	191.7	6.8	0	11.58	0	0.991976	-42.303992	220	10	1.1	0.9;
	7903	1	0	-0	0	7.78	0	1.072292	-42.076237	220	5	1.1	0.9;
	7905	1	347.1	87.2	0	88.35	0	1.046063	-17.916487	380	5	1.1	0.9;
	7913	2	0	0	0	0	0	1.063382	-8.903291	380	5	1.1	0.9;
	7915	1	-0.18	-1.31	0	0	0	1.012672	16.947793	150	8	1.1	0.9;
	7918	1	-0	-0	0	117.47	0	1.03136	-47.278708	380	4	1.1	0.9;
	7923	1	0	-0	0	0.97	0	1.038389	-22.166903	220	5	1.1	0.9;
	7925	1	36.84	14.1	0	33.55	0	1.068086	-42.655468	380	4	1.1	0.9;
	7927	1	2.2	0.5	0	0.13	0	1.032907	-33.407192	220	10	1.1	0.9;
	7929	2	0	0	0	0	0	1.050833	-43.956471	380	4	1.1	0.9;
	7935	1	4.5	1.1	0	12.56	0	1.059045	-47.312721	380	10	1.1	0.9;
	7937	1	56.06	17	0	2.4	0	1.036123	-7.391945	220	5	1.1	0.9;
	7943	1	38.44	2	0	0.56	0	1.04536	-16.204311	220	5	1.1	0.9;
	7945	1	17.92	5.8	0	0.84	0	1.025468	-5.745797	220	5	1.1	0.9;
	7951	1	156	60.7	0	1.52	0	1.008787	1.304576	150	8	1.1	0.9;
	7955	1	111.6	8.9	0	16.19	0	1.064403	-23.358794	220	5	1.1	0.9;
	7958	1	-0	-0	0	0.22	0	1.001402	-0.492115	150	8	1.1	0.9;
	7960	2	0	0	0	0	0	1.018041	6.487865	380	2	1.1	0.9;
	7961	1	50.56	8.1	0	8.02	0	1.036447	-8.590687	220	5	1.1	0.9;
	7965	1	0	0	0	79.5	0	1.076922	-37.80677	380	1	1.1	0.9;
	7967	1	0	-0	0	5.97	0	1.043923	-11.423985	220	5	1.1	0.9;
	7969	1	0	-0	0.2527	79.80215	0	1.051096	-42.100911	380	10	1.1	0.9;
	7972	1	237.9	68.5	0	5.12	0	1.034223	-19.331892	220	5	1.1	0.9;
	7973	1	115.7	10.7	0	19.1	0	1.029695	-53.74855	220	4	1.1	0.9;
	7974	1	234	13.1	0	51.03	0	1.03335	-20.083487	220	5	1.1	0.9;
	7980	1	233.5	151	0	35.31	0	1.045534	-43.261604	380	10	1.1	0.9;
	7982	1	112	19.7	0	38.04	0	1.03592	-23.160805	220	5	1.1	0.9;
	7988	1	-3.55	-1.43	0	11.834121	0	1.03551	-2.130971	380	5	1.1	0.9;
	7989	1	0	0	0	0.27	0	1.047576	-3.691217	220	5	1.1	0.9;
	7994	1	11.91	3.3	0	6.84	0	1.014446	-16.292763	220	5	1.1	0.9;
	7996	1	0	-0	0	55	0	1.021261	-1.472747	150	8	1.1	0.9;
	7998	2	0	0	0	0	0	1.042456	4.346992	380	5	1.1	0.9;
	7999	1	0	-0	0	9.68	0	1.025718	-48.097149	220	4	1.1	0.9;
	8005	1	-0	-0	0	16.77	0	1.02865	-13.702695	220	5	1.1	0.9;
	8007	1	7.91	2.7	0	0.01	0	0.989845	-55.466231	150	10	1.1	0.9;
	8012	1	36.84	14.1	0	17.84	0	1.058815	-40.172861	380	4	1.1	0.9;
	8016	1	23.33	9.1	0	15.36	0	1.016288	4.516916	150	8	1.1	0.9;
	8017	1	-0	-0	0	8.99	0	1.023998	11.729802	150	8	1.1	0.9;
	8021	1	14.22	4.7	0	5.58	0	0.996849	-30.633993	220	10	1.1	0.9;
	8030	1	-0	0	0	122.13	0	1.034462	-4.472294	380	5	1.1	0.9;
	8034	1	83.99	6	0	65.12	0	1.047311	-49.713598	380	4	1.1	0.9;
	8035	1	-0	0	0	4.22	0	1.02206	-16.365944	220	5	1.1	0.9;
	8037	1	0	-0	0	5.73	0	1.034628	9.582266	150	8	1.1	0.9;
	8043	2	0	0	0	0	0	1.063893	-11.286543	220	5	1.1	0.9;
	8044	2	0	0	0	0	0	1.052835	28.542058	110	8	1.1	0.9;
	8057	1	69.18	9.5	0	3.67	0	1.04691	-8.660865	220	5	1.1	0.9;
	8058	1	10.71	17.7	0	10.58	0	1.023025	11.713091	150	8	1.1	0.9;
	8060	1	0	-0	0	9.54	0	1.047468	-13.453767	220	5	1.1	0.9;
	8066	2	0	0	0	0	0	1.01361	-46.688841	220	4	1.1	0.9;
	8068	1	41.74	7.4	0	18.54	0	1.03287	-53.247599	220	4	1.1	0.9;
	8079	1	0	0	0	0.81	0	1.014968	-48.209864	220	4	1.1	0.9;
	8086	1	0	0	0	17.69	0	1.035202	13.999882	380	8	1.1	0.9;
	8087	2	0	0	0	0	0	1.03306	-38.178126	220	4	1.1	0.9;
	8088	1	0	-0	0.379958	52.33965	0	1.041542	-51.256123	380	10	1.1	0.9;
	8098	1	206.4	21	0	8.55	0	0.993253	-49.633965	150	10	1.1	0.9;
	8103	1	39.34	-10.7	0	0	0	0.986395	-1.654563	150	2	1.1	0.9;
	8104	1	127.2	40.8	0	0.2	0	1.070556	-25.20678	220	5	1.1	0.9;
	8107	1	68.18	16.1	0	3.35	0	1.034541	-23.391765	220	5	1.1	0.9;
	8109	2	0	0	0	0	0	1.06101	-21.072941	380	5	1.1	0.9;
	8112	1	87.9	29.3	0	1.21	0	1.035282	-3.58729	220	5	1.1	0.9;
	8114	1	-0	0	0	30.39	0	0.997751	11.674608	380	8	1.1	0.9;
	8124	1	80.49	7.3	0	1.81	0	1.016097	-60.182064	220	4	1.1	0.9;
	8128	1	-0	0	0	8.1	0	1.088468	-9.653161	380	5	1.1	0.9;
	8138	1	-0	0	0	6.66	0	1.053106	-42.708847	220	4	1.1	0.9;
	8141	1	0	0	0	50	0	1.037929	3.19519	150	8	1.1	0.9;
	8147	1	303.5	56.2	0	21.07	0	1.018621	-47.552935	220	4	1.1	0.9;
	8149	1	111.36	27.44	0	2.59	0	1.003847	-45.877241	220	4	1.1	0.9;
	8151	1	0	0	0	14.23	0	1.042949	-25.622195	220	5	1.1	0.9;
	8155	1	125.3	34.6	0	4.36	0	1.019574	9.50676	150	8	1.1	0.9;
	8158	2	0	0	0	0	0	1.05367	-0.553541	220	5	1.1	0.9;
	8159	2	0	0	0	0	0	1.064422	-40.369631	380	4	1.1	0.9;
	8163	1	76.58	16.2	0	21.54	0	1.027285	-53.440227	220	4	1.1	0.9;
	8164	1	1.8	0.6	0	0.43	0	1.01848	6.913157	150	8	1.1	0.9;
	8165	1	0	-0	0	2.42	0	1.057662	-3.433844	220	5	1.1	0.9;
	8167	1	-0	0	0	1.22	0	1.014661	4.964915	150	8	1.1	0.9;
	8179	1	-0	0	0.194464	19.1468	0	1.000011	-5.181422	380	2	1.1	0.9;
	8180	1	291.6	47	0	2.38	0	1.074171	-30.830541	220	5	1.1	0.9;
	8181	1	0	-0	0.210824	6.620599	0	1.001921	-4.815597	380	2	1.1	0.9;
	8183	1	0	0	0	2.97	0	1.064461	-47.252646	380	10	1.1	0.9;
	8186	1	99.6	22.9	0	15.02	0	1.031758	-53.520592	220	4	1.1	0.9;
	8189	1	0	-0	0	9.54	0	1.041138	-18.173764	220	5	1.1	0.9;
	8190	1	0	-0	0	0.32	0	1.049024	-10.305691	220	5	1.1	0.9;
	8191	1	0	0	0	11.38	0	1.047336	-15.336918	220	5	1.1	0.9;
	8194	1	-0.18	-1	0	0	0	1.021281	11.418838	150	8	1.1	0.9;
	8195	1	-101.01	-129.29	0	37.176703	0	1.022946	-17.065718	380	5	1.1	0.9;
	8197	1	-134.35	-6.5	0	3.38	0	1.005264	-37.692173	150	10	1.1	0.9;
	8200	1	0	-0	0	39.58	0	1.018387	-6.164902	380	5	1.1	0.9;
	8204	1	-0	-0	0	0.25	0	1.021186	-0.388948	150	8	1.1	0.9;
	8207	2	0	0	0	0	0	1.030795	-41.862698	220	4	1.1	0.9;
	8209	1	83.39	16.5	0	16.43	0	1.03284	-20.164145	220	5	1.1	0.9;
	8210	2	0	0	0	0	0	1.053137	-37.468708	380	4	1.1	0.9;
	8211	1	-0	0	0	65.93	0	1.019117	20.177021	380	8	1.1	0.9;
	8213	1	207.9	25.4	0	1.75	0	1.040329	-45.735133	220	4	1.1	0.9;
	8214	1	58.97	11.1	0	1.05	0	1.043506	-11.477321	220	5	1.1	0.9;
	8222	2	0	0	0	0	0	1.019521	-10.479346	380	5	1.1	0.9;
	8223	2	0	0	0	0	0	1.058011	17.860107	380	8	1.1	0.9;
	8224	1	92.7	-5.6	0.2527	35.620175	0	1.052137	-41.560675	380	10	1.1	0.9;
	8225	2	0	0	0	0	0	1.067644	-11.747659	220	5	1.1	0.9;
	8229	1	0	0	0	0	0	1.063217	-7.210075	150	2	1.1	0.9;
	8230	1	0	0	0	59.43	0	1.033248	11.611667	380	8	1.1	0.9;
	8236	1	118.1	20.2	0	1.47	0	1.028257	-53.577025	220	4	1.1	0.9;
	8240	2	0	0	0	0	0	1.059307	-38.737835	380	4	1.1	0.9;
	8242	2	0	0	0	0	0	1.01497	-45.885146	220	4	1.1	0.9;
	8247	2	0	0	0	0	0	1.059416	-40.401531	380	4	1.1	0.9;
	8249	1	0	0	0	17.8	0	1.013163	11.791202	380	8	1.1	0.9;
	8250	1	196.1	32.1	0	3.86	0	1.035371	-26.17535	220	5	1.1	0.9;
	8255	1	39.04	13.1	0	3.22	0	1.023269	-13.783725	220	5	1.1	0.9;
	8260	1	0	-0	0	10.21	0	1.047941	-49.577511	380	4	1.1	0.9;
	8264	1	0	-0	0	65.73	0	1.019133	20.170123	380	8	1.1	0.9;
	8265	1	40.44	20.5	0	0.85	0	1.062244	-23.870999	220	5	1.1	0.9;
	8267	2	0	0	0	0	0	1.045102	-10.010182	380	5	1.1	0.9;
	8277	1	57.66	6.1	0	1.77	0	1.007724	22.24887	110	8	1.1	0.9;
	8278	2	0	0	0	0	0	1.006884	-31.874004	150	10	1.1	0.9;
	8284	1	-0.12	-1.31	0	0	0	1.087542	32.323919	380	8	1.1	0.9;
	8285	1	156.6	16.9	0	12.26	0	1.016702	-53.999369	220	4	1.1	0.9;
	8288	1	16.02	5.1	0	8.6	0	1.048417	11.066634	110	8	1.1	0.9;
	8291	1	0	-0	0	6.52	0	1.045972	-7.676921	220	5	1.1	0.9;
	8293	1	112.7	14.5	0	1.24	0	1.046708	-24.849013	220	5	1.1	0.9;
	8294	1	14.22	6.3	0	0.38	0	1.047825	-11.902121	220	5	1.1	0.9;
	8298	1	0	0	0	14.35	0	1.087038	32.309575	380	8	1.1	0.9;
	8301	1	59.67	9.2	0	18.03	0	1.028051	-51.292216	220	4	1.1	0.9;
	8307	1	-0	0	0	0.14	0	1.046458	-3.855147	220	5	1.1	0.9;
	8310	1	-0	0	0	2.46	0	1.046024	-8.637363	220	5	1.1	0.9;
	8311	2	0	0	0	0	0	1.039615	-15.453389	220	5	1.1	0.9;
	8312	2	0	0	0	0	0	1.078014	-12.24119	380	5	1.1	0.9;
	8316	1	0	0	0	0.72	0	1.082642	-26.528331	220	5	1.1	0.9;
	8318	1	-0	-0	0	0	0	1.049758	18.746946	380	8	1.1	0.9;
	8328	1	82.39	14.4	0	13.37	0	1.040546	-24.459456	220	5	1.1	0.9;
	8329	1	0	0	0.149712	2.04987	0	1.010579	5.724362	380	2	1.1	0.9;
	8331	1	-0	-0	0	0.05	0	1.049189	-12.997183	220	5	1.1	0.9;
	8334	1	263.96	-98.64	0	27.848535	0	1.032843	-14.767853	380	5	1.1	0.9;
	8335	1	637.92	139.57	0	0	0	1.009769	-11.521151	150	2	1.1	0.9;
	8337	1	149.9	-0	0	28.38	0	1.0496	-42.597986	380	4	1.1	0.9;
	8338	1	-2.3	-3.34	0	8.15	0	1.02467	-43.798324	220	4	1.1	0.9;
	8347	1	-2.09	-1.43	0	48.358243	0	0.987155	-18.667915	380	5	1.1	0.9;
	8352	1	67.27	15.9	0	2.65	0	1.021243	-39.022936	220	4	1.1	0.9;
	8353	1	155.1	39.9	0	6.37	0	1.029816	-55.720331	220	4	1.1	0.9;
	8356	1	-0	-0	0	0.72	0	1.015127	1.297374	150	8	1.1	0.9;
	8361	1	126.54	29.1	0	12.55	0	1.073255	-26.451713	220	5	1.1	0.9;
	8367	1	77.99	19.3	0	3.04	0	1.036984	-6.681744	220	5	1.1	0.9;
	8369	1	-0	-0	0	1.23	0	1.028676	-9.152628	220	5	1.1	0.9;
	8373	1	216.3	-6.9	0	2.44	0	1.070996	-48.215066	220	5	1.1	0.9;
	8374	1	153.4	86.8	0	53.98	0	1.061288	-41.845858	380	4	1.1	0.9;
	8378	2	0	0	0	0	0	1.022564	-38.108897	220	4	1.1	0.9;
	8380	1	0	-0	0	0.22	0	1.015958	19.538705	110	8	1.1	0.9;
	8381	1	-0	-0	0	17.92	0	1.071454	-44.524455	380	1	1.1	0.9;
	8389	1	72.18	49.6	0	37.52	0	1.016669	17.36147	150	8	1.1	0.9;
	8392	1	100.8	32.3	0	7.16	0	1.047048	7.528918	150	8	1.1	0.9;
	8393	1	-757.23	63.89	0	34.18	0	1.002909	17.544011	380	8	1.1	0.9;
	8396	1	131.7	32.9	0	108.56	0	1.062328	-40.645356	380	4	1.1	0.9;
	8397	1	-0	0	0	4.05	0	1.043616	-25.457032	220	5	1.1	0.9;
	8405	1	40.75	12.5	0	2.61	0	1.037511	-12.364331	220	5	1.1	0.9;
	8406	1	0	0	0	14.38	0	1.049365	-25.182116	220	5	1.1	0.9;
	8411	1	39.64	9.6	0	1.8	0	1.041273	-25.850241	220	5	1.1	0.9;
	8417	1	266	62.8	0	1.95	0	1.01013	-47.59632	220	4	1.1	0.9;
	8419	1	99.1	-98.6	0	6.83	0	1.017228	-59.911374	220	4	1.1	0.9;
	8420	1	-0	0	0	0	0	1.05117	-9.097494	150	2	1.1	0.9;
	8432	1	-0.19	-17.71	0	28.98	0	1.02638	12.015884	380	8	1.1	0.9;
	8438	1	125.5	22.5	0	9.17	0	1.019637	-47.16442	220	4	1.1	0.9;
	8439	1	380.8	62.8	0	7.11	0	0.997692	-23.484518	220	5	1.1	0.9;
	8442	1	31.44	12.4	0	2.89	0	1.033294	3.146647	150	8	1.1	0.9;
	8448	1	286.1	89.3	0	8.89	0	1.017839	-20.618009	220	5	1.1	0.9;
	8458	2	0	0	0	0	0	1.045078	-9.483962	380	5	1.1	0.9;
	8459	1	285.2	83.5	0	5.6	0	0.969593	-43.205859	220	10	1.1	0.9;
	8463	1	24.13	10.8	0	15.2	0	1.037105	6.294399	150	8	1.1	0.9;
	8466	1	0	-0	0	1.97	0	1.040487	-4.452236	220	5	1.1	0.9;
	8467	1	216.2	75.3	0	2.74	0	0.983889	-23.029689	220	5	1.1	0.9;
	8468	1	21.42	11.8	0	28.85	0	1.003777	-14.954113	380	5	1.1	0.9;
	8473	2	0	0	0	0	0	1.042275	-8.85577	220	5	1.1	0.9;
	8475	1	-0	0	0	8.35	0	1.071679	-47.923897	220	5	1.1	0.9;
	8477	1	0	0	0	2.46	0	1.064533	-10.136664	220	5	1.1	0.9;
	8478	1	-0	0	0	3.01	0	1.048947	-8.507678	220	5	1.1	0.9;
	8480	1	0	0	0	1.15	0	1.029915	15.10216	110	8	1.1	0.9;
	8483	1	-0.17	-1.82	0	0	0	1.050084	18.748297	380	8	1.1	0.9;
	8485	1	149	99.8	0	9.27	0	1.04146	-7.387137	220	5	1.1	0.9;
	8486	2	0	0	0	0	0	1.046146	-19.545285	380	5	1.1	0.9;
	8487	1	44.45	7.68	0	17.83	0	0.996752	-3.789756	380	5	1.1	0.9;
	8490	1	-0.19	-1	0	0	0	1.016007	-0.665636	150	8	1.1	0.9;
	8492	1	98.6	27.1	0	42.45	0	1.002793	-4.061923	150	8	1.1	0.9;
	8493	1	-0	-0	0	13.01	0	0.972187	-36.402896	220	10	1.1	0.9;
	8494	1	235.97	34.52	0	7.15	0	1.087369	-9.837706	380	5	1.1	0.9;
	8497	1	38.84	14.2	0	3.5	0	1.046167	-14.558369	220	5	1.1	0.9;
	8504	1	6.41	2.2	0	52.07	0	1.074291	-38.217186	380	4	1.1	0.9;
	8507	1	37.44	9.1	0	7.87	0	1.074019	-26.375849	220	5	1.1	0.9;
	8508	1	-8.02	-15.14	0	0	0	1.028669	17.409525	150	8	1.1	0.9;
	8510	1	0	0	0	8.78	0	1.061792	15.567985	220	8	1.1	0.9;
	8511	1	93.6	-43.5	0	33.51	0	1.040262	-24.549837	220	5	1.1	0.9;
	8512	1	140.7	10.4	0	2.51	0	1.012675	-47.088026	220	4	1.1	0.9;
	8515	2	0	0	0	0	0	1.055375	-12.443359	220	5	1.1	0.9;
	8522	2	0	0	0	0	0	1.023003	-15.643844	220	5	1.1	0.9;
	8525	1	-0	-3.7	0	0.25	0	1.042349	27.051491	110	8	1.1	0.9;
	8529	1	76.99	-0.7	0	1.06	0	1.014093	-46.532827	220	4	1.1	0.9;
	8530	1	-0	0	0	1.31	0	1.01471	-46.176096	220	4	1.1	0.9;
	8534	1	295.2	39.3	0	7.3	0	1.016996	-43.798173	220	4	1.1	0.9;
	8535	1	72.68	24.3	0	3.12	0	1.035755	-21.953277	220	5	1.1	0.9;
	8542	1	87.7	26	0	47.58	0	1.043635	-21.062798	220	5	1.1	0.9;
	8546	1	5.01	1.8	0	0.77	0	1.039564	-11.636283	220	5	1.1	0.9;
	8547	1	21.22	9.9	0	4.64	0	1.015341	-51.627331	220	4	1.1	0.9;
	8560	1	-0	0	0	1.11	0	0.983882	-18.215513	380	5	1.1	0.9;
	8561	1	-0	-0	0	237.54	0	1.065428	-42.770872	380	4	1.1	0.9;
	8564	2	0	0	0	0	0	1.019996	-19.556384	220	5	1.1	0.9;
	8565	1	-0.19	-0.92	0	0	0	1.049865	7.888607	150	8	1.1	0.9;
	8568	1	175.7	56.2	0	0.13	0	1.022255	-6.042108	220	5	1.1	0.9;
	8571	1	126.8	21.4	0	11.43	0	1.029852	-53.46566	220	4	1.1	0.9;
	8574	1	-0	0	0.447839	20.994738	0	0.99818	-3.765849	380	2	1.1	0.9;
	8575	1	89.6	13.1	0	8.14	0	1.029116	-53.698048	220	4	1.1	0.9;
	8578	1	0	-0	0	4.55	0	1.044202	-22.07106	220	5	1.1	0.9;
	8581	1	0	-0	0	1.79	0	1.0043	5.087257	380	2	1.1	0.9;
	8582	1	27.83	-2.6	0	6.07	0	1.026673	-51.284917	220	4	1.1	0.9;
	8592	1	-0	0	0	0.43	0	1.047329	-3.731214	220	5	1.1	0.9;
	8601	1	0	0	0	1.96	0	1.041573	-51.25628	380	10	1.1	0.9;
	8604	1	0	-0	0	20.81	0	1.017915	15.899545	380	8	1.1	0.9;
	8605	2	0	0	0	0	0	1.009283	-0.131172	380	2	1.1	0.9;
	8613	1	101.8	20.2	0	0.91	0	1.020302	2.367885	150	8	1.1	0.9;
	8621	1	-70.45	-35.14	0	0	0	1.017227	17.769534	150	8	1.1	0.9;
	8625	2	0	0	0	0	0	1.055846	-41.065047	380	4	1.1	0.9;
	8626	1	136	36.1	0	6.07	0	0.98923	-17.630415	220	5	1.1	0.9;
	8627	1	-0	0	0	4.26	0	1.047406	-26.434879	220	5	1.1	0.9;
	8628	1	277.1	37	0	21.03	0	1.041444	-23.628029	220	5	1.1	0.9;
	8629	1	98.61	50.3	0	0.17	0	0.969991	-42.948363	150	10	1.1	0.9;
	8630	2	0	0	0	0	0	1.064508	-46.868087	380	10	1.1	0.9;
	8636	1	31.43	10.6	0	1.61	0	1.035848	-21.721595	220	5	1.1	0.9;
	8642	1	163.2	31.9	0	4.9	0	1.031267	8.872582	150	8	1.1	0.9;
	8651	1	162.49	45.11	0	5.38	0	1.028711	-17.231975	220	5	1.1	0.9;
	8653	1	44.55	7.5	0	0.6	0	1.034402	-16.023528	220	5	1.1	0.9;
	8656	1	5.51	1.3	0	0.09	0	1.035127	-8.458553	220	5	1.1	0.9;
	8658	1	38.54	12	0	16.52	0	1.031202	-53.697921	220	4	1.1	0.9;
	8661	1	226.7	39.3	0	0.99	0	1.02646	-47.934753	220	4	1.1	0.9;
	8662	1	-0.12	-1.74	0	0	0	1.087501	32.310934	380	8	1.1	0.9;
	8664	1	34.54	15.4	0	29.82	0	1.061218	-46.141427	380	4	1.1	0.9;
	8668	1	69.88	19.3	0	1.63	0	1.018394	14.026518	110	8	1.1	0.9;
	8669	1	74.98	10.5	0	12.96	0	1.046423	-15.344297	220	5	1.1	0.9;
	8670	2	0	0	0	0	0	1.049183	-6.529837	220	5	1.1	0.9;
	8671	1	62.77	5.7	0	2.41	0	1.018922	-50.527242	220	4	1.1	0.9;
	8672	1	-0	-0	0	28.515005	0	1.023937	-21.027655	380	5	1.1	0.9;
	8676	2	0	0	0	0	0	1.018144	-21.878056	220	5	1.1	0.9;
	8677	1	0	0	0	2.29	0	1.020081	-10.192806	220	5	1.1	0.9;
	8683	2	0	0	0	0	0	1.041179	-24.317296	220	5	1.1	0.9;
	8689	1	42.65	18.2	0	0	0	1.03523	-20.733231	220	5	1.1	0.9;
	8691	1	305.04	77.16	0	17.43	0	1.045574	-33.643601	220	5	1.1	0.9;
	8701	1	82.59	23.4	0	6.65	0	1.023588	-43.556409	220	4	1.1	0.9;
	8702	2	0	0	0	0	0	1.029789	-40.586324	220	4	1.1	0.9;
	8704	1	180.7	71.8	0	2.46	0	1.041183	-7.97841	220	5	1.1	0.9;
	8706	1	40.34	15.9	0	6.02	0	1.008191	4.606463	150	8	1.1	0.9;
	8707	1	101.9	24.3	0	8.53	0	1.029902	-11.704441	220	5	1.1	0.9;
	8710	1	124.5	-6	0	1.91	0	1.024755	-47.559445	220	4	1.1	0.9;
	8711	1	0	0	0	3.9	0	1.041543	-12.078252	220	5	1.1	0.9;
	8719	1	91.8	8.3	0	2.44	0	1.017108	-44.601504	220	4	1.1	0.9;
	8721	2	0	0	0	0	0	1.056768	-15.959174	380	5	1.1	0.9;
	8722	1	-0	0	0	2.13	0	1.071802	-25.073253	220	5	1.1	0.9;
	8729	2	0	0	0	0	0	1.009953	-37.874633	220	4	1.1	0.9;
	8731	1	191.2	16.4	0	17.99	0	1.025891	-39.717321	380	4	1.1	0.9;
	8732	1	66.27	19.3	0	12.94	0	1.03704	-21.186119	220	5	1.1	0.9;
	8738	1	-0	0	0	3.78	0	1.015889	6.018492	150	8	1.1	0.9;
	8743	1	11.61	7.6	0	6.21	0	1.033764	-34.537661	220	5	1.1	0.9;
	8748	1	279.7	4.6	0	3.77	0	1.06653	-49.707057	220	5	1.1	0.9;
	8756	1	54.66	18.6	0	3.04	0	1.016103	-59.644677	220	4	1.1	0.9;
	8763	1	0	-0	0	215.65	0	1.032339	2.72345	380	5	1.1	0.9;
	8765	1	-0	-0	0	11.69	0	1.046637	-15.330191	220	5	1.1	0.9;
	8771	2	0	0	0	0	0	1.008996	-37.338884	150	10	1.1	0.9;
	8772	1	0	0	0	0	0	1.023626	8.631158	150	8	1.1	0.9;
	8773	1	0	-0	0	30.31	0	0.997763	11.671711	380	8	1.1	0.9;
	8776	1	-0	-0	0	6.28	0	1.001752	-37.729943	220	4	1.1	0.9;
	8784	1	100.5	19.3	0	13.51	0	1.022002	11.415331	150	8	1.1	0.9;
	8787	1	22.83	8	0	3.71	0	1.066321	-17.009493	220	5	1.1	0.9;
	8788	1	143.8	46.45	0	3.26	0	1.042393	-5.95552	220	5	1.1	0.9;
	8789	1	-29.73	23.6	0	11.75	0	0.991696	-46.068395	150	10	1.1	0.9;
	8791	1	-0	-0	0	0.36	0	1.056484	-22.378068	220	5	1.1	0.9;
	8792	1	107.9	35.9	0	5.48	0	1.027558	4.096913	150	8	1.1	0.9;
	8795	2	0	0	0	0	0	1.055026	-0.12855	220	5	1.1	0.9;
	8804	1	22.02	4.4	0	4.83	0	1.050456	-9.493537	220	5	1.1	0.9;
	8806	1	-0	0	0	342.15	0	1.069256	-43.457093	380	4	1.1	0.9;
	8807	2	0	0	0	0	0	1.046456	-3.855139	220	5	1.1	0.9;
	8808	1	203.6	19.81	0	15.46	0	1.055503	-25.842277	220	5	1.1	0.9;
	8809	1	101.2	-2.9	0	0.51	0	1.027013	-32.578876	220	5	1.1	0.9;
	8814	1	89.9	29.3	0	6.77	0	1.029202	3.877715	150	8	1.1	0.9;
	8818	2	0	0	0	0	0	1.051013	0.642803	220	5	1.1	0.9;
	8819	1	37.87	11.59	0	1.88	0	1.035296	-52.835216	220	4	1.1	0.9;
	8825	1	0	-0	0	44.95	0	1.008085	-2.700245	380	5	1.1	0.9;
	8828	1	0	0	0	1.05	0	1.016741	6.318703	150	8	1.1	0.9;
	8829	1	91.6	27.1	0	10.95	0	1.039155	-21.21347	220	5	1.1	0.9;
	8832	1	0	-0	0	2.24	0	1.015247	-0.729836	150	8	1.1	0.9;
	8834	1	-0	0	0	1.52	0	1.003348	-22.560352	220	5	1.1	0.9;
	8835	1	113.7	22.3	0	15.55	0	1.036477	-22.934857	220	5	1.1	0.9;
	8843	1	208.7	63	0	3.45	0	1.05134	-11.210293	220	5	1.1	0.9;
	8846	1	20.82	0	0	0.17	0	1.067593	7.848137	220	5	1.1	0.9;
	8847	1	-0	0	0.18076	-0.881251	0	0.993761	-0.54588	380	2	1.1	0.9;
	8853	1	410.3	91.6	0	13.25	0	1.051311	-36.586636	220	5	1.1	0.9;
	8854	1	69.18	35.1	0	12.59	0	1.027085	-44.337295	220	5	1.1	0.9;
	8858	2	0	0	0	0	0	1.027559	-43.116605	220	4	1.1	0.9;
	8860	1	57.16	19.7	0	22.17	0	1.021941	-10.157615	220	5	1.1	0.9;
	8861	1	-0	0	0	3.95	0	0.989557	-54.074535	220	10	1.1	0.9;
	8862	1	50.96	15.7	0	1.52	0	1.042169	14.438159	110	8	1.1	0.9;
	8864	1	63.31	-9.8	0	5.36	0	1.042974	-31.901323	220	5	1.1	0.9;
	8867	1	131.5	12.9	0	18.7	0	1.024036	-48.357834	220	4	1.1	0.9;
	8870	1	-0	0	0	129.47	0	1.054624	-47.749596	380	4	1.1	0.9;
	8871	2	0	0	0	0	0	1.037687	-52.222994	220	4	1.1	0.9;
	8872	2	0	0	0	0	0	1.056528	-15.983313	380	5	1.1	0.9;
	8873	1	0	-0	0	1.87	0	1.070378	-27.323305	220	5	1.1	0.9;
	8874	1	285.2	61.4	0	26.04	0	1.041431	-20.131936	220	5	1.1	0.9;
	8875	2	0	0	0	0	0	1.047011	17.817721	150	8	1.1	0.9;
	8877	1	-0	0	0	1.03	0	1.038573	-20.681367	220	5	1.1	0.9;
	8878	1	58.26	16.9	0	0.95	0	1.020394	23.076892	110	8	1.1	0.9;
	8879	1	0	0	0	1.47	0	1.036661	-22.775474	220	5	1.1	0.9;
	8881	1	0	0	0	3.86	0	1.008501	-1.308496	150	8	1.1	0.9;
	8882	1	-0.16	-1.44	0	0.42	0	1.048879	11.612548	110	8	1.1	0.9;
	8886	1	0	-0	0	-1.392036	0	0.979342	-3.75641	380	5	1.1	0.9;
	8887	1	0	0	0	102.70867	0	1.040073	1.821607	380	5	1.1	0.9;
	8888	1	101.6	4.5	0	3.59	0	1.037601	-44.205902	220	4	1.1	0.9;
	8889	1	-8.15	-5.23	0	0	0	1.050781	7.942576	150	8	1.1	0.9;
	8893	1	278.3	33.6	0	25.9	0	1.046765	-15.235259	220	5	1.1	0.9;
	8895	1	-0.2	-1	0	0	0	1.023021	-1.383064	150	8	1.1	0.9;
	8898	1	0	0	0	0	0	1.049767	18.744304	380	8	1.1	0.9;
	8900	1	3.1	1.6	0	0.09	0	1.047679	-23.734294	220	5	1.1	0.9;
	8903	2	0	0	0	0	0	1.021928	-11.382923	380	5	1.1	0.9;
	8906	1	91	21	0	11.23	0	1.035877	-23.091391	220	5	1.1	0.9;
	8913	1	-0	0	0	1.13	0	1.039275	-20.592607	220	5	1.1	0.9;
	8917	1	59.67	13.1	0	2.26	0	0.992842	5.915957	110	8	1.1	0.9;
	8921	1	0	0	0	16.37	0	0.998492	-2.950962	380	2	1.1	0.9;
	8926	1	67.08	35.5	0	32.32	0	0.989796	-55.840175	220	10	1.1	0.9;
	8927	2	0	0	0	0	0	1.015122	-34.562812	150	10	1.1	0.9;
	8930	1	139.4	21.2	0	6.21	0	0.986412	-18.654346	220	5	1.1	0.9;
	8931	1	-0.56	-0	0	108.21	0	1.088084	-9.652001	380	5	1.1	0.9;
	8936	1	-0	-0	0	3.13	0	1.031095	-53.048263	220	4	1.1	0.9;
	8939	1	186.8	22.9	0	9.94	0	1.015956	-54.590796	220	4	1.1	0.9;
	8946	2	0	0	0	0	0	1.041085	-41.176169	380	4	1.1	0.9;
	8947	1	3.7	0.2	0	0.05	0	1.041674	-11.509748	220	5	1.1	0.9;
	8950	2	0	0	0	0	0	1.045004	-14.86535	220	5	1.1	0.9;
	8953	1	516.7	144.3	0	2.32	0	1.015018	-39.845825	220	4	1.1	0.9;
	8956	2	0	0	0	0	0	1.064772	23.13068	380	8	1.1	0.9;
	8957	1	0	0	0	4.39	0	0.989889	-54.003917	220	10	1.1	0.9;
	8961	2	0	0	0	0	0	1.038824	-14.118012	220	5	1.1	0.9;
	8963	1	-0	0	0	2.78	0	1.015806	-40.105518	220	10	1.1	0.9;
	8964	1	925.91	147.97	0	0	0	1.025866	-12.86993	150	2	1.1	0.9;
	8973	2	0	0	0	0	0	1.050147	-35.915457	380	4	1.1	0.9;
	8975	1	0	-0	0	11.04	0	1.041454	-24.319884	220	5	1.1	0.9;
	8976	2	0	0	0	0	0	1.045593	-17.103981	220	5	1.1	0.9;
	8977	1	81.29	26.5	0	7.04	0	1.011449	1.005067	150	8	1.1	0.9;
	8980	1	63.97	12.8	0	0.17	0	1.044325	-11.082462	220	5	1.1	0.9;
	8981	2	0	0	0	0	0	1.027038	-42.675433	220	4	1.1	0.9;
	8987	2	0	0	0	0	0	1.052269	-41.141403	380	4	1.1	0.9;
	8989	1	88.5	24.5	0	2.23	0	1.041862	-25.620235	220	5	1.1	0.9;
	8991	1	-8.49	-17.09	0	0	0	1.015291	3.736834	150	8	1.1	0.9;
	8992	1	-0	-0	0	14.64	0	1.046455	-3.756643	220	5	1.1	0.9;
	8993	2	0	0	0	0	0	1.022201	-37.558452	220	4	1.1	0.9;
	8994	1	69.28	31.3	0	32.76	0	1.037094	6.46761	150	8	1.1	0.9;
	8997	2	0	0	0	0	0	1.002676	-13.17164	380	5	1.1	0.9;
	8998	1	-0	0	0	4.08	0	0.971706	-42.865881	220	10	1.1	0.9;
	8999	1	41.15	10.3	0	2.41	0	1.043446	-21.589649	220	5	1.1	0.9;
	9001	2	0	0	0	0	0	1.025721	-41.459434	220	4	1.1	0.9;
	9002	1	119.02	12.54	0	23.45	0	1.022623	-10.494423	220	5	1.1	0.9;
	9007	1	73.68	24	0	0.46	0	1.014412	1.30218	150	8	1.1	0.9;
	9011	1	88.4	25.2	0	12.9	0	1.03607	-23.054646	220	5	1.1	0.9;
	9012	1	25.63	6.5	0	0.52	0	1.048604	1.102665	220	5	1.1	0.9;
	9014	1	70.78	19.6	0	8.1	0	1.052162	-6.665777	220	5	1.1	0.9;
	9018	1	55.46	16.9	0	1.92	0	1.022149	-6.603112	220	5	1.1	0.9;
	9019	1	91.4	18.1	0	4.89	0	1.039622	-24.721521	220	5	1.1	0.9;
	9021	1	152.18	21.63	0	1.91	0	1.039071	-6.338232	220	5	1.1	0.9;
	9024	1	0	0	0.141019	25.246	0	1.025444	1.951824	380	2	1.1	0.9;
	9025	1	0	-0	0.205903	3.0768	0	1.005508	-4.49576	380	2	1.1	0.9;
	9028	1	81.99	32.2	0	7.3	0	1.036092	-52.422827	220	4	1.1	0.9;
	9029	1	-0	-0	0	7.92	0	1.01267	16.94777	150	8	1.1	0.9;
	9032	1	0	-0	0	2.42	0	1.064444	-47.252548	380	10	1.1	0.9;
	9033	1	76.88	16.4	0	6.61	0	1.050041	-26.48683	220	5	1.1	0.9;
	9039	2	0	0	0	0	0	1.02063	-43.159076	220	4	1.1	0.9;
	9043	2	0	0	0	0	0	1.02106	-40.012553	220	4	1.1	0.9;
	9045	1	60.57	18.9	0	1.46	0	1.039106	-24.673735	220	5	1.1	0.9;
	9049	1	36.84	14.1	0	3.8	0	1.018489	-40.152369	220	4	1.1	0.9;
	9051	1	0	-0	0	2.15	0	1.039937	0.565336	220	5	1.1	0.9;
	9055	2	0	0	0	0	0	0.982562	-49.677102	220	4	1.1	0.9;
	9057	1	72.38	23	0	0.28	0	1.02341	-48.135629	220	4	1.1	0.9;
	9059	1	0	-0	0	2	0	1.071723	-27.23087	220	5	1.1	0.9;
	9063	1	-41.57	-23.4	0	0	0	1.026728	11.942947	150	8	1.1	0.9;
	9064	1	0	-0	0	15.26	0	1.002528	-3.934171	380	2	1.1	0.9;
	9065	1	-0	-0	0	0.59	0	1.032374	2.535924	380	5	1.1	0.9;
	9066	1	23.26	3.01	0	9.29	0	1.019399	-20.82846	220	5	1.1	0.9;
	9067	2	0	0	0	0	0	1.045155	1.524541	220	5	1.1	0.9;
	9072	1	0	-0	0	1.41	0	1.035251	5.451595	150	8	1.1	0.9;
	9083	1	88.1	36	0	27.89	0	1.009534	-0.822012	150	8	1.1	0.9;
	9091	1	0	0	0	3.65	0	1.042523	-9.744131	220	5	1.1	0.9;
	9094	1	187.4	72.8	0	34.95	0	1.035929	6.023351	150	8	1.1	0.9;
	9097	2	0	0	0	0	0	1.01907	-39.584889	220	4	1.1	0.9;
	9101	2	0	0	0	0	0	1.026751	-21.689348	380	5	1.1	0.9;
	9106	1	-0	0	0	1.12	0	0.988316	-55.413025	220	4	1.1	0.9;
	9108	2	0	0	0	0	0	1.04489	-16.471181	220	5	1.1	0.9;
	9109	1	0	-0	0	0.31	0	1.039939	0.565316	220	5	1.1	0.9;
	9112	1	0	-0	0	19.71	0	1.049188	-12.997177	220	5	1.1	0.9;
	9117	1	36.74	14.1	0	16.98	0	1.023525	-52.097135	380	4	1.1	0.9;
	9118	1	-0.22	-3.34	0	0	0	1.029714	3.875256	150	8	1.1	0.9;
	9119	1	-0	0	0	1.24	0	1.041761	-20.561727	220	5	1.1	0.9;
	9120	1	0	0	0	4.13	0	1.019729	2.890205	150	8	1.1	0.9;
	9127	1	38.84	9.6	0	0.88	0	1.005733	21.671612	110	8	1.1	0.9;
	9128	1	50.56	8.7	0	0.19	0	1.034185	-2.632449	220	5	1.1	0.9;
	9130	1	245.51	9.17	0	24.25	0	1.044738	-30.943014	220	5	1.1	0.9;
	9131	1	-0	-0	0	6.62	0	1.0496	-24.137879	220	5	1.1	0.9;
	9136	1	-0	0	0	0	0	0.9931	-53.684274	220	10	1.1	0.9;
	9137	2	0	0	0	0	0	1.048734	-26.47055	220	5	1.1	0.9;
	9140	2	0	0	0	0	0	1.049339	-5.241785	220	5	1.1	0.9;
	9146	1	35.44	11.4	0	1.15	0	1.019746	8.415269	110	8	1.1	0.9;
	9150	2	0	0	0	0	0	1.037874	-13.218156	380	5	1.1	0.9;
	9155	1	-0	-0	0	5.01	0	1.023228	-20.882747	220	5	1.1	0.9;
	9158	1	54.96	9.42	0	5.92	0	1.023649	-14.568689	220	5	1.1	0.9;
	9159	1	-0	-0	0	81.15	0	1.05214	-43.291737	380	4	1.1	0.9;
	9162	1	94.9	40.3	0	17.2	0	1.01511	-60.285572	220	4	1.1	0.9;
	9164	1	7.71	2.4	0	6.93	0	0.998782	-19.437343	220	5	1.1	0.9;
	9173	1	66.07	15.6	0	5.57	0	1.032259	-20.230037	220	5	1.1	0.9;
	9174	2	0	0	0	0	0	1.020141	-32.801345	380	5	1.1	0.9;
	9176	1	-0	0	0	1.28	0	1.04942	-12.999767	220	5	1.1	0.9;
	9180	2	0	0	0	0	0	1.057766	-8.606877	220	5	1.1	0.9;
	9181	1	2.43	1.35	0	0.1	0	1.035041	-9.441995	220	5	1.1	0.9;
	9185	1	66.97	18.6	0	18.5	0	1.052903	-6.607366	220	5	1.1	0.9;
	9189	1	62.57	-61	0	55.14	0	1.046485	-25.065792	220	5	1.1	0.9;
	9191	1	-0	-0	0	2.34	0	1.073296	-24.836176	220	5	1.1	0.9;
	9192	1	17.92	23.3	0	0	0	1.033027	-10.73637	150	2	1.1	0.9;
	9201	1	137.9	25.6	0	8.24	0	1.022891	-42.801498	220	4	1.1	0.9;
	9203	1	627.36	-73.13	0	27.185221	0	1.005382	-12.53126	380	5	1.1	0.9;
	9205	1	23.43	7.3	0	0.89	0	1.025558	13.881084	110	8	1.1	0.9;
	9206	1	226.1	88.8	0	122.78	0	1.033506	4.172445	150	8	1.1	0.9;
	9213	1	-0	-0	0	10.210952	0	1.024999	-4.638587	380	5	1.1	0.9;
	9216	1	103.6	4.69	0	36.68	0	1.024848	-48.161418	220	4	1.1	0.9;
	9217	1	-0	0	0	6.33	0	1.041729	-8.889947	220	5	1.1	0.9;
	9222	1	-5.15	-2.04	0	154.446402	0	1.055906	-16.165943	380	5	1.1	0.9;
	9223	1	133.5	24.1	0	2.92	0	1.010836	7.611553	110	8	1.1	0.9;
	9231	1	242.08	59.98	0	5.6	0	1.027624	-19.18143	220	5	1.1	0.9;
	9239	2	0	0	0	0	0	1.018061	6.492507	380	2	1.1	0.9;
	9241	1	0	-0	0	0.36	0	1.050472	-9.493678	220	5	1.1	0.9;
];

%% generator data
%	bus	Pg	Qg	Qmax	Qmin	Vg	mBase	status	Pmax	Pmin	Pc1	Pc2	Qc1min	Qc1max	Qc2min	Qc2max	ramp_agc	ramp_10	ramp_30	ramp_q	apf
mpc.gen = [
	32	8	0.06	5.91	-4.56	1.006206	100	1	16.2	0	0	0	0	0	0	0	0	0	0	0	0;
	33	5.3	0.41	2.27	-1.08	1.031909	100	1	5.38	0	0	0	0	0	0	0	0	0	0	0	0;
	39	34.3	4.84	17.76	-7.14	1.005551	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	51	-144.5	0.55	5.48	-4.19	1.01198	100	1	100	-192.67	0	0	0	0	0	0	0	0	0	0	0;
	124	861.3	51.52	440.32	-172.59	1.079291	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	150	67.29	-9.79	77.09	-22.54	1.062067	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	179	-179.2	-33.84	52.64	-51.72	1.048817	100	1	100	-238.93	0	0	0	0	0	0	0	0	0	0	0;
	194	268.6	43.34	130.19	-53.53	1.029584	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	201	5.2	0.39	2.27	-1.07	1.024675	100	1	5.37	0	0	0	0	0	0	0	0	0	0	0	0;
	221	-155.39	91.05	183.92	-65.3	0.996091	100	1	100	-207.18	0	0	0	0	0	0	0	0	0	0	0;
	299	975.9	127.2	721.25	-258.71	1.046197	100	1	1600	533.33	0	0	0	0	0	0	0	0	0	0	0;
	338	83.7	1.48	55.54	-19.16	1.049578	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	352	1250.4	268.17	926.35	-297.68	1.054691	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	389	69	16.81	56.73	-17.84	1.020856	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	413	231	19.25	124.12	-49.84	1.052152	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	440	83.1	-4.05	55.73	-19.01	1.029422	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	442	-40.35	0.43	1.95	-0.93	1.013689	100	1	100	-53.8	0	0	0	0	0	0	0	0	0	0	0;
	453	260.1	54.54	257.23	-74.4	1.021096	100	1	540	0	0	0	0	0	0	0	0	0	0	0	0;
	489	-107.4	0.51	6.28	-4.82	1.012805	100	1	100	-143.2	0	0	0	0	0	0	0	0	0	0	0;
	516	539.4	188.57	261.79	-106.67	1.074804	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	564	-51.67	-0.63	18.47	-19.35	1.069692	100	1	100	-68.89	0	0	0	0	0	0	0	0	0	0	0;
	583	-270	5.38	17.88	-7.03	1.043394	100	1	100	-360	0	0	0	0	0	0	0	0	0	0	0;
	615	38.2	0.1	17.09	-7.79	1.06611	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	616	379.2	41.62	217.9	-81.16	1.049295	100	1	480	160	0	0	0	0	0	0	0	0	0	0	0;
	618	670.3	-35.39	351.18	-139.11	1.028233	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	628	152.4	41.51	78.79	-31.69	0.98494	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	639	-46.2	113.22	135.52	-38.9	1.046182	100	1	100	-61.6	0	0	0	0	0	0	0	0	0	0	0;
	642	63.33	-61.52	90.13	-93.03	1.036757	100	1	276.09	0	0	0	0	0	0	0	0	0	0	0	0;
	682	37.4	7.68	38.77	-11.06	1.065269	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	686	-21.65	-10.17	34.75	-34.26	1.018064	100	1	100	-28.86	0	0	0	0	0	0	0	0	0	0	0;
	698	274.8	1.65	129.94	-53.91	1.001541	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	709	16	-10.2	22.23	-20.83	1.072146	100	1	65.3	0	0	0	0	0	0	0	0	0	0	0	0;
	739	465.7	-64.04	212.97	-93.45	1.040583	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	749	198.88	4.46	105.69	-43.84	1.0548	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	751	65.9	-25.62	136.94	-47.24	1.023645	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	757	116.7	16.76	51.3	-23.41	1.016623	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	776	-111.4	5.74	18.16	-6.73	1.029038	100	1	100	-148.53	0	0	0	0	0	0	0	0	0	0	0;
	778	16.5	-0.5	12.37	-9.36	1.051035	100	1	33.58	0	0	0	0	0	0	0	0	0	0	0	0;
	780	50.8	17.97	37.43	-12.32	1.006139	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	795	58.4	7.49	36.75	-13.04	1.039638	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	803	-89.2	2.65	74.03	-25.54	1.04396	100	1	100	-118.93	0	0	0	0	0	0	0	0	0	0	0;
	807	1597.5	-194.05	864.47	-360.97	1.057043	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	823	1274.2	283.53	926.68	-298.31	1.044347	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	849	8.7	-0.02	6.45	-4.91	1.041392	100	1	17.57	0	0	0	0	0	0	0	0	0	0	0	0;
	851	59.56	49.36	96.64	-27.81	1.006263	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	858	855.6	161.18	440.67	-172.15	1.035299	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	891	1702.8	646.56	1097.41	-373.91	1.063198	100	1	2400	800	0	0	0	0	0	0	0	0	0	0	0;
	913	1249.4	719.82	772.11	-269.94	1.032774	100	1	1700	566.67	0	0	0	0	0	0	0	0	0	0	0;
	916	230.2	-26.12	134.45	-49.76	1.024009	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	930	679.6	-118.4	406.66	-144.75	1.063763	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	972	1221.2	397.48	934.73	-291.52	1.068998	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	976	21.3	-5.88	36.02	-35.35	1.029847	100	1	107.83	0	0	0	0	0	0	0	0	0	0	0	0;
	1001	92.1	29.32	55.02	-19.81	1.073365	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1002	171	15.98	89.55	-35.15	1.040649	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	1021	-179.3	3.66	58.34	-16.23	1.019135	100	1	100	-239.07	0	0	0	0	0	0	0	0	0	0	0;
	1030	164.94	7.29	71.66	-27.89	1.00979	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	1036	202.5	-1.53	137.39	-46.45	1.038657	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	1043	829.5	41.74	387.88	-163.4	1.029789	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	1083	91.8	10.7	54.91	-19.88	1.046149	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1093	837	95.08	550.69	-185.15	1.026457	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	1100	58.2	5.03	36.8	-13.01	1.045212	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1102	-124.15	12.77	38.05	-11.59	1.051902	100	1	100	-165.54	0	0	0	0	0	0	0	0	0	0	0;
	1121	289.79	-18.85	158.2	-62.81	1.019674	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	1173	-124.7	0.31	1.85	-0.88	1.026125	100	1	100	-166.27	0	0	0	0	0	0	0	0	0	0	0;
	1174	26.8	5.72	18.61	-6.28	1.004865	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1175	309.54	-1.37	108.35	-39.12	1.040018	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	1237	1504.8	347.49	906.61	-319.44	1.105045	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	1241	46.13	-14.82	22.29	-20.98	1.021438	100	1	65.63	0	0	0	0	0	0	0	0	0	0	0	0;
	1251	837	89.88	550.69	-185.15	1.0265	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	1271	-100.5	-12.52	32.54	-32.11	1.015573	100	1	100	-134	0	0	0	0	0	0	0	0	0	0	0;
	1295	-124.1	9.31	36.92	-12.88	1.038943	100	1	100	-165.47	0	0	0	0	0	0	0	0	0	0	0;
	1341	67.2	31.22	57.03	-17.62	1.030396	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1354	-246.5	21.44	38.08	-11.73	1.04168	100	1	100	-328.67	0	0	0	0	0	0	0	0	0	0	0;
	1391	26.2	2.19	18.74	-6.17	1.028915	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1399	-174.6	-2.47	11.64	-10.2	1.052722	100	1	100	-232.8	0	0	0	0	0	0	0	0	0	0	0;
	1409	628.19	-89.19	338.66	-151.76	1.024907	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	1422	-101	3.68	18.09	-6.79	1.045292	100	1	100	-134.67	0	0	0	0	0	0	0	0	0	0	0;
	1436	49.4	2.83	37.72	-12.09	1.049261	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1458	1588	-288.24	895.23	-330.78	1.05352	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	1475	783.9	-108.21	394.45	-157	1.025403	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	1478	62.6	7.21	36.4	-13.44	1.046165	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1498	5.2	0.36	2.25	-1.06	1.078389	100	1	5.31	0	0	0	0	0	0	0	0	0	0	0	0;
	1570	121.4	-4.09	93.16	-29.49	1.079057	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	1604	50.6	14.56	37.66	-12.17	1.044277	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1618	-136.6	-10.11	21.27	-20.07	1.079319	100	1	100	-182.13	0	0	0	0	0	0	0	0	0	0	0;
	1642	50.4	13.34	37.74	-12.11	1.031925	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1680	1250.4	271.17	928.89	-296.1	1.044209	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	1685	62.7	16.57	57.41	-17.22	1.005778	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1697	-189	-12.59	31.94	-31.24	1.041486	100	1	100	-252	0	0	0	0	0	0	0	0	0	0	0;
	1708	860.8	132.37	439.37	-173.3	1.037844	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	1721	28.1	1.63	18.52	-6.39	1.061536	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1747	835.5	271.39	387.67	-163.73	1.055274	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	1754	40	5.45	16.99	-7.92	1.040804	100	1	41.06	0	0	0	0	0	0	0	0	0	0	0	0;
	1788	430.2	252.54	272.25	-94.92	1.018729	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	1793	-90.3	-0.67	34.64	-15.2	1.023406	100	1	100	-120.4	0	0	0	0	0	0	0	0	0	0	0;
	1794	1250.4	248.88	926.35	-297.68	1.052726	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	1808	-0.6	-6.42	111.23	-38.16	1.047818	100	1	100	-0.8	0	0	0	0	0	0	0	0	0	0	0;
	1850	2.89	-1.45	11.52	-10.09	1.020156	100	1	32.97	0	0	0	0	0	0	0	0	0	0	0	0;
	1851	15.6	-0.75	11.76	-8.89	1.0492	100	1	31.92	0	0	0	0	0	0	0	0	0	0	0	0;
	1852	-241.8	15.7	58.16	-16.52	1.056131	100	1	100	-322.4	0	0	0	0	0	0	0	0	0	0	0;
	1888	573.68	43.66	396.02	-155.83	1.043102	100	1	912.78	312.78	0	0	0	0	0	0	0	0	0	0	0;
	1890	1526	129.25	686.73	-294.62	1.050852	100	1	1600	533.33	0	0	0	0	0	0	0	0	0	0	0;
	1898	246	-24.92	133.11	-51.25	1.077705	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	1909	-31.18	3.61	38.67	-11.1	1.021788	100	1	100	-41.57	0	0	0	0	0	0	0	0	0	0	0;
	1914	115.5	-35.25	160.36	-150.95	1.038338	100	1	472.16	39.35	0	0	0	0	0	0	0	0	0	0	0;
	1947	-17.02	0.43	2.05	-0.97	1.056511	100	1	100	-22.7	0	0	0	0	0	0	0	0	0	0	0;
	1959	31	1.06	18.14	-6.75	1.066953	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1999	141.6	5.8	80.51	-30.07	1.080207	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	2030	576	167.94	257	-110.85	1.072679	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	2035	855.6	161.48	441.85	-171.27	1.034553	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	2050	836.2	-45.26	962.2	-263.65	1.008592	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	2085	176	36.2	88.39	-36.19	1.03273	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2091	33.9	1.74	17.88	-7.04	1.026543	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2096	1063.5	-194.12	522.74	-212.37	1.050714	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	2107	1269.4	611.1	661.9	-257.62	1.048234	100	1	1500	500	0	0	0	0	0	0	0	0	0	0	0;
	2150	498.4	98.97	266.17	-101.5	1.064827	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	2177	-121.3	15.21	58.35	-16.26	1.070558	100	1	100	-161.73	0	0	0	0	0	0	0	0	0	0	0;
	2193	7.9	1.86	7.47	-2.56	1.014181	100	1	16.12	0	0	0	0	0	0	0	0	0	0	0	0;
	2197	-53.29	2.68	19.51	-14.74	1.036195	100	1	100	-71.05	0	0	0	0	0	0	0	0	0	0	0;
	2206	204.3	24.03	137.29	-46.58	1.020141	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	2225	-83.5	0.41	2.27	-1.04	1.02967	100	1	100	-111.33	0	0	0	0	0	0	0	0	0	0	0;
	2236	-38.45	0.44	2.23	-1.07	1.023164	100	1	100	-51.26	0	0	0	0	0	0	0	0	0	0	0;
	2267	362	147.75	228.16	-78.83	1.000857	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	2276	38.9	-22.76	41.67	-36.5	1.055141	100	1	119.3	0	0	0	0	0	0	0	0	0	0	0	0;
	2291	110.5	12.61	95.69	-28.93	1.068687	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2335	431.2	242.01	273.49	-94.1	1.040978	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	2359	78.2	8.56	34.35	-15.5	1.025004	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2412	67.3	-92.42	134.54	-136.29	1.016849	100	1	409.09	0	0	0	0	0	0	0	0	0	0	0	0;
	2421	-320.4	10.06	37.39	-12.45	1.047704	100	1	100	-427.2	0	0	0	0	0	0	0	0	0	0	0;
	2425	-100.5	14.28	54.3	-20.49	1.043025	100	1	100	-134	0	0	0	0	0	0	0	0	0	0	0;
	2426	-43.2	-33.04	105.34	-99.02	1.040828	100	1	100	-57.6	0	0	0	0	0	0	0	0	0	0	0;
	2446	1367	280.85	921.62	-304.73	1.053565	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	2460	37.79	20.28	38.01	-11.74	0.983266	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2468	46.8	4.68	38.02	-11.81	1.049474	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2481	31.8	11.07	39.12	-10.71	1.041748	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2489	286.5	-17.67	162.4	-61.79	1.049212	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	2517	3.85	-14.82	78.13	-77.07	1.017237	100	1	234.52	0	0	0	0	0	0	0	0	0	0	0	0;
	2550	54.3	3.25	58.23	-16.49	1.050935	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2572	411.2	80.01	276.19	-91.63	1.044905	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	2600	76.48	15.74	76.08	-23.59	1.041924	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	2627	1367	289.47	921.52	-304.79	1.052984	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	2642	1897	-254.15	848.76	-377.67	1.054923	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	2653	41.92	10.03	51.94	-22.8	1.036409	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2658	308.31	-21.53	124.68	-59.39	1.026597	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	2660	144.9	12.3	79.69	-30.77	1.013897	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	2686	-72.9	-46.07	69.01	-69.9	1.020441	100	1	100	-97.2	0	0	0	0	0	0	0	0	0	0	0;
	2719	77.6	-16.82	70.15	-58.1	1.039499	100	1	196.72	0	0	0	0	0	0	0	0	0	0	0	0;
	2764	575.4	-114.34	416.7	-134.5	1.05096	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	2786	1221.2	379.25	934.81	-291.47	1.069288	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	2797	-95.1	-2.82	88.33	-36.33	1.044662	100	1	100	-126.8	0	0	0	0	0	0	0	0	0	0	0;
	2799	66.8	-1.37	77.91	-21.66	1.055463	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	2800	1048.2	-98.17	525.34	-209.84	1.030206	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	2816	1250.4	280.3	926.26	-297.74	1.054536	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	2817	71.8	4.24	34.8	-14.94	0.991705	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2841	80.7	-3.32	55.98	-18.76	1.079986	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2842	31.6	6.06	18.19	-6.74	1.044122	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2853	126	10.21	72.02	-27.47	1.000604	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	2872	4.5	0.35	3.31	-2.54	1.060615	100	1	9.05	0	0	0	0	0	0	0	0	0	0	0	0;
	2878	8.95	-12.23	72.95	-68.71	1.05418	100	1	214.86	0	0	0	0	0	0	0	0	0	0	0	0;
	2886	60.9	10.13	57.69	-17.02	1.052878	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2896	-223.4	0.31	1.52	-0.73	1.011629	100	1	100	-297.87	0	0	0	0	0	0	0	0	0	0	0;
	2902	20.9	1.52	19.19	-5.72	1.061022	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2930	53.2	2.38	37.24	-12.54	1.049271	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2934	3.4	0.17	2.51	-1.91	1.052968	100	1	6.83	0	0	0	0	0	0	0	0	0	0	0	0;
	2946	-222.1	0.29	1.83	-0.87	1.020023	100	1	100	-296.13	0	0	0	0	0	0	0	0	0	0	0;
	2985	46.6	-70.17	184.46	-132.67	1.05077	100	1	493.35	0	0	0	0	0	0	0	0	0	0	0	0;
	3018	162.5	-4.96	89.61	-34.82	1.043371	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	3028	855.7	158.53	441.85	-171.27	1.034318	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	3054	210.4	123.54	167.72	-77.11	0.999852	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	3113	-20.3	-10.92	22.39	-20.99	1.072644	100	1	100	-27.07	0	0	0	0	0	0	0	0	0	0	0;
	3114	38.6	3.91	17.33	-7.61	1.0681	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	3133	85.54	-1.1	113.97	-35.56	1.061763	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	3134	-136.3	-10.75	70.25	-29.33	1.02503	100	1	100	-181.73	0	0	0	0	0	0	0	0	0	0	0;
	3150	-31.34	0.57	6.66	-5.1	1.017094	100	1	100	-41.79	0	0	0	0	0	0	0	0	0	0	0;
	3157	9	-0.09	6.78	-5.13	1.023658	100	1	18.41	0	0	0	0	0	0	0	0	0	0	0	0;
	3163	33.9	-29.77	67.92	-69.02	1.006379	100	1	206.86	0	0	0	0	0	0	0	0	0	0	0	0;
	3183	98.7	-3.82	53.7	-20.98	1.056991	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3205	416.7	108.62	243.61	-88.46	1.03763	100	1	540	0	0	0	0	0	0	0	0	0	0	0	0;
	3218	-107.4	9.03	36.79	-13	1.032146	100	1	100	-143.2	0	0	0	0	0	0	0	0	0	0	0;
	3233	13	-7.58	178.47	-79.11	1.012423	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	3240	-42.3	-24.84	62.23	-51.37	1.051571	100	1	100	-56.4	0	0	0	0	0	0	0	0	0	0	0;
	3291	326.4	-17.26	229	-77.02	1.027251	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	3306	861.3	134.8	439.2	-173.43	1.069265	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	3324	-26.48	16.92	38.13	-11.64	1.016523	100	1	100	-35.3	0	0	0	0	0	0	0	0	0	0	0;
	3334	-45.5	3.24	34.41	-15.34	1.03247	100	1	100	-60.67	0	0	0	0	0	0	0	0	0	0	0;
	3335	-600	2045.7	Inf	-Inf	1.08873	100	1	630	-630	0	0	0	0	0	0	0	0	0	0	0;
	3346	56.8	12.86	37.1	-12.76	1.014746	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3353	19.07	6.07	38.2	-11.61	1.047796	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3364	44.5	-22.17	47.19	-41.44	1.055303	100	1	135.27	0	0	0	0	0	0	0	0	0	0	0	0;
	3390	2.4	21.5	54.59	-20.19	1.041678	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3416	171.3	38.2	76.43	-34.12	0.997085	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	3422	53.7	-2.9	58.19	-16.5	1.065868	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3436	521	207.76	550.97	-184.89	1.056276	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	3492	-370.8	6.56	53.75	-21.08	1.04519	100	1	100	-494.4	0	0	0	0	0	0	0	0	0	0	0;
	3513	829.8	43.52	387.88	-163.4	1.029628	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	3565	15.88	2.99	38.83	-10.97	1.079121	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3580	539	187.63	261.7	-107.36	1.075232	100	1	600	0	0	0	0	0	0	0	0	0	0	0	0;
	3582	-321.8	-36.63	52.83	-52.04	1.045127	100	1	100	-429.07	0	0	0	0	0	0	0	0	0	0	0;
	3584	-88.4	-1.9	19.8	-15.12	1.019651	100	1	100	-117.87	0	0	0	0	0	0	0	0	0	0	0;
	3585	50.8	17.99	37.43	-12.32	1.006097	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3611	-125.4	7.89	54.59	-20.12	1.03827	100	1	100	-167.2	0	0	0	0	0	0	0	0	0	0	0;
	3621	202.03	-32.82	130.01	-54.04	1.023915	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	3629	54.97	1.75	76.85	-22.64	1.020807	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	3644	16.4	-6.77	22.35	-21.22	1.018948	100	1	66.09	0	0	0	0	0	0	0	0	0	0	0	0;
	3656	-26.2	13.18	116.07	-33.32	1.067132	100	1	100	-34.93	0	0	0	0	0	0	0	0	0	0	0;
	3659	1097.4	862.06	943.64	-282.01	1.065964	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	3661	861.2	129.38	441.04	-172.05	1.059844	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	3673	424.5	-84.67	257.61	-110.43	1.013624	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	3689	5.2	0.41	2.18	-1.05	1.032178	100	1	6.2	0	0	0	0	0	0	0	0	0	0	0	0;
	3691	239.6	48.92	133.84	-49.89	1.030169	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	3692	164.4	-53.69	180.11	-77.84	1.025431	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	3698	-12.58	5.71	36.34	-13.47	1.041028	100	1	100	-16.78	0	0	0	0	0	0	0	0	0	0	0;
	3711	40.4	12.69	38.36	-11.38	1.00126	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3723	1146.3	67.92	514.15	-221.73	1.062798	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	3741	854.7	206.11	548.57	-187.16	1.035379	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	3791	90.6	18.06	54.7	-19.97	0.997709	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3792	85.4	-26.48	108.06	-41.25	1.058708	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	3809	257	129.92	313.01	-116.73	1.084761	100	1	700	233.33	0	0	0	0	0	0	0	0	0	0	0;
	3821	292	-39.5	165.76	-79.17	1.029497	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	3825	837.3	64.59	388.17	-163.4	1.026381	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	3840	-22.95	-16.27	36.29	-35.42	1.030834	100	1	100	-30.6	0	0	0	0	0	0	0	0	0	0	0;
	3869	855.6	161.13	440.69	-172.14	1.035477	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	3876	-25.1	16.55	34.27	-15.52	1.046886	100	1	100	-33.47	0	0	0	0	0	0	0	0	0	0	0;
	3916	-21.27	16.17	38.02	-11.77	1.044945	100	1	100	-28.37	0	0	0	0	0	0	0	0	0	0	0;
	3951	-27.1	-74.17	126.16	-127.56	1.029014	100	1	100	-36.13	0	0	0	0	0	0	0	0	0	0	0;
	3971	-134.5	15.75	55.67	-19.09	1.060173	100	1	100	-179.33	0	0	0	0	0	0	0	0	0	0	0;
	3983	229.9	11.5	134.98	-48.75	1.012943	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	3987	44.6	7.31	38.05	-11.72	1.021034	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4016	109.2	-3.77	52.25	-22.44	1.026951	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4024	89.5	44.88	97.07	-27.44	1.032708	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	4026	399.3	-17.62	223.07	-83.7	1.05962	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	4056	-76.9	8.25	38.05	-11.77	1.042513	100	1	100	-102.53	0	0	0	0	0	0	0	0	0	0	0;
	4084	366.1	46.85	264.94	-103.74	1.060214	100	1	600	0	0	0	0	0	0	0	0	0	0	0	0;
	4118	41	4.36	38.37	-11.4	1.029072	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4125	59.6	18.23	36.89	-13	0.999283	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4128	67.5	24.85	57.03	-17.63	1.023094	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4182	710.7	-68.75	344.9	-146.06	1.016453	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	4192	767.4	310.02	396.38	-154.99	1.077662	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	4204	-48.16	0.36	1.83	-0.85	1.017661	100	1	100	-64.21	0	0	0	0	0	0	0	0	0	0	0;
	4231	2641.24	819.49	Inf	-Inf	1.050918	100	1	4188.95	1333.33	0	0	0	0	0	0	0	0	0	0	0;
	4259	28	6.24	18.41	-6.45	1.013671	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	4279	515.39	-71.15	261.6	-106.39	1.01711	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	4291	251	-30.61	196.08	-61.88	1.02141	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	4331	60.2	6.39	36.73	-13.12	1.032206	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4333	152.4	4.96	79.1	-31.46	0.999193	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	4337	1334.3	544.16	576.78	-158.19	1.010847	100	1	2700	800	0	0	0	0	0	0	0	0	0	0	0;
	4338	-52.5	4.46	16.93	-7.97	1.024208	100	1	100	-70	0	0	0	0	0	0	0	0	0	0	0;
	4393	115.36	-20	142	-41.87	1.066547	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	4395	-129.1	12.79	38.48	-11.33	1.039119	100	1	100	-172.13	0	0	0	0	0	0	0	0	0	0	0;
	4419	861.3	131.97	441.04	-172.05	1.059728	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	4428	14.7	-5.61	20.49	-19.24	1.019796	100	1	60.24	0	0	0	0	0	0	0	0	0	0	0	0;
	4442	-11.23	-12.54	34.61	-33.92	1.036687	100	1	100	-14.98	0	0	0	0	0	0	0	0	0	0	0;
	4480	861.3	136.04	439.2	-173.43	1.069297	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	4482	35.2	14.05	38.85	-10.95	1.05005	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4506	-171.8	22.11	37.86	-11.95	1.040247	100	1	100	-229.07	0	0	0	0	0	0	0	0	0	0	0;
	4566	-30.39	8.95	37.94	-11.87	1.073131	100	1	100	-40.51	0	0	0	0	0	0	0	0	0	0	0;
	4586	-545.7	56.41	92.96	-29.81	1.042233	100	1	100	-727.6	0	0	0	0	0	0	0	0	0	0	0;
	4624	68	8.43	35.62	-14.19	1.039942	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4657	101.62	2.77	140.98	-43.31	1.022686	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	4661	-78.9	3.81	19.14	-5.77	1.051599	100	1	100	-105.2	0	0	0	0	0	0	0	0	0	0	0;
	4663	-116	0.12	55.07	-19.63	1.028053	100	1	100	-154.67	0	0	0	0	0	0	0	0	0	0	0;
	4690	541	15.43	417.98	-132.47	1.033192	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	4701	93.2	-37.45	98.4	-86.29	1.044876	100	1	281.85	0	0	0	0	0	0	0	0	0	0	0	0;
	4772	78	-17.24	89.54	-34.98	1.027623	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	4773	16	-5.4	22.15	-20.9	1.020563	100	1	65.3	0	0	0	0	0	0	0	0	0	0	0	0;
	4776	805.5	144.16	389.83	-161.16	1.068296	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	4783	861.3	133.49	439.2	-173.43	1.069324	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	4816	-26.08	13.8	38.37	-11.44	1.062911	100	1	100	-34.77	0	0	0	0	0	0	0	0	0	0	0;
	4819	855.6	162.5	440.69	-172.14	1.035474	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	4823	10.3	1.34	4.44	-2.07	1.06272	100	1	10.46	0	0	0	0	0	0	0	0	0	0	0	0;
	4843	6.8	-0.2	5	-3.81	1.053038	100	1	13.63	0	0	0	0	0	0	0	0	0	0	0	0;
	4846	239.6	48.9	133.84	-49.89	1.030285	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	4850	88.8	11.28	76.52	-23.19	1.052648	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	4880	-40.3	7.65	75.62	-23.96	1.03955	100	1	100	-53.73	0	0	0	0	0	0	0	0	0	0	0;
	4897	18.56	4.81	37.58	-12.23	1.018589	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4918	41.69	10.15	37.27	-12.57	1.059135	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4931	20.2	-17.48	34.39	-33.64	1.040167	100	1	102.79	0	0	0	0	0	0	0	0	0	0	0	0;
	4945	385.4	-33.86	224.4	-82.22	1.059702	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	4952	-12.57	7.48	37.23	-12.61	1.044318	100	1	100	-16.77	0	0	0	0	0	0	0	0	0	0	0;
	4954	40.84	0.35	54.46	-20.26	1.019681	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4957	961	-177.26	533.39	-201.12	1.048836	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	4967	83.2	-17.25	182.2	-75.83	1.023151	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	5004	54.3	3.31	74.73	-70.41	1.061776	100	1	220.12	0	0	0	0	0	0	0	0	0	0	0	0;
	5019	49.4	2.83	37.72	-12.09	1.04927	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5025	-155.1	0.45	5.65	-4.3	1.006979	100	1	100	-206.8	0	0	0	0	0	0	0	0	0	0	0;
	5051	83.48	-0.85	74.55	-24.97	1.039098	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5060	13.3	-4.4	14.22	-12.43	1.067508	100	1	40.68	0	0	0	0	0	0	0	0	0	0	0	0;
	5065	1008.3	199.74	532.96	-203.19	1.072972	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	5067	542	129.85	258.91	-109.58	1.021364	100	1	600	0	0	0	0	0	0	0	0	0	0	0	0;
	5091	276.42	-51.56	151.81	-69.25	1.054352	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	5097	21.1	2.24	8.91	-4.11	1.059554	100	1	21.23	7.08	0	0	0	0	0	0	0	0	0	0	0;
	5110	-52.51	11.13	38.28	-11.5	1.034874	100	1	100	-70.01	0	0	0	0	0	0	0	0	0	0	0;
	5120	75.6	-5.08	56.35	-18.42	1.057395	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5144	1713.6	109.47	882.86	-343.24	1.035348	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	5171	-87.6	-8.56	32.89	-32.29	1.01545	100	1	100	-116.8	0	0	0	0	0	0	0	0	0	0	0;
	5237	106.8	9.87	75.02	-24.74	1.050569	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5239	-1344.4	66.73	Inf	-Inf	1.005494	100	1	100	-1792.53	0	0	0	0	0	0	0	0	0	0	0;
	5278	119.4	1.07	51.3	-23.49	1.041244	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5280	-193.9	0.41	2.29	-1.06	1.021983	100	1	100	-258.53	0	0	0	0	0	0	0	0	0	0	0;
	5290	332.54	-12.08	159.67	-61.42	1.011274	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	5299	152.3	-101.06	542.73	-582.68	1.061166	100	1	1689.14	51.19	0	0	0	0	0	0	0	0	0	0	0;
	5329	-54.78	-11.3	35.25	-34.6	1.018401	100	1	100	-73.05	0	0	0	0	0	0	0	0	0	0	0;
	5340	-159	13.31	36.37	-13.43	1.051438	100	1	100	-212	0	0	0	0	0	0	0	0	0	0	0;
	5348	10.8	-1.39	11.36	-10.08	1.015924	100	1	32.74	0	0	0	0	0	0	0	0	0	0	0	0;
	5365	19.59	1.74	18.14	-6.75	1.063822	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	5379	1250.6	357.14	926.47	-297.63	1.105689	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	5395	-395.8	7.69	58.06	-16.66	1.049	100	1	100	-527.73	0	0	0	0	0	0	0	0	0	0	0;
	5409	219.5	-21.59	102.6	-44.69	1.023182	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	5425	181.1	-12.46	108.27	-39.2	1.077356	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	5461	43.4	12.92	38.26	-11.55	1.05743	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5465	18.5	-9.16	31.09	-30.76	1.001327	100	1	93.47	0	0	0	0	0	0	0	0	0	0	0	0;
	5481	1368.6	154.37	920.31	-305.64	1.063327	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	5482	56.8	-11.92	101.32	-48.12	1.046078	100	1	249.06	0	0	0	0	0	0	0	0	0	0	0	0;
	5486	861.3	130.17	441.04	-172.05	1.0598	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	5488	-96.4	21.75	37.49	-12.33	1.041439	100	1	100	-128.53	0	0	0	0	0	0	0	0	0	0	0;
	5490	3424.8	253.92	1765.36	-686.67	1.064044	100	1	4000	1333.33	0	0	0	0	0	0	0	0	0	0	0;
	5512	75.62	6.26	69.67	-30	1.020123	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5533	173	12.93	88.87	-35.71	1.041341	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	5546	21.5	2.41	19.16	-5.75	1.055712	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	5564	57.6	12.64	36.8	-12.99	1.042161	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5658	229.8	83.69	103.18	-46.25	1.009808	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	5664	128	15.73	72.71	-27.05	1.068735	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5669	294.54	12.08	34.47	-15.34	1.003946	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	5670	16.6	-4.52	22.97	-21.7	1.018899	100	1	67.76	0	0	0	0	0	0	0	0	0	0	0	0;
	5709	837	87.17	550.69	-185.15	1.026473	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	5726	18.5	-10.63	31.98	-31.54	1.020165	100	1	95.98	0	0	0	0	0	0	0	0	0	0	0	0;
	5742	84.5	15.08	35.69	-16.43	1.028692	100	1	85.14	28.38	0	0	0	0	0	0	0	0	0	0	0;
	5781	1102.83	412.72	779.29	-324.79	1.038035	100	1	1800	600	0	0	0	0	0	0	0	0	0	0	0;
	5787	63.55	-12.77	114.67	-32.56	1.031918	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	5814	181.2	64.23	196.98	-77.17	1.048258	100	1	440	146.67	0	0	0	0	0	0	0	0	0	0	0;
	5821	-74.21	-20.45	45.83	-46.3	1.019461	100	1	100	-98.94	0	0	0	0	0	0	0	0	0	0	0;
	5823	404.6	-38.48	220.75	-85.54	1.063184	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	5825	262.6	40.41	76.5	-34.05	0.995795	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	5828	13.2	-3.55	13.93	-12.27	1.029274	100	1	39.98	0	0	0	0	0	0	0	0	0	0	0	0;
	5831	-179.4	0.25	1.58	-0.74	1.032948	100	1	100	-239.2	0	0	0	0	0	0	0	0	0	0	0;
	5856	305.7	32.63	214.13	-85	1.039775	100	1	480	160	0	0	0	0	0	0	0	0	0	0	0;
	5881	60	6.27	36.72	-13.13	1.062314	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5940	-74.9	7.47	17.21	-7.73	1.031482	100	1	100	-99.87	0	0	0	0	0	0	0	0	0	0	0;
	5970	127.57	-18.76	172.47	-48.34	1.019864	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	5971	1221.2	389.96	934.54	-291.64	1.067028	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	5983	-102.5	7.94	37.27	-12.59	1.064637	100	1	100	-136.67	0	0	0	0	0	0	0	0	0	0	0;
	5991	854.6	-173.92	517.55	-217.25	1.018313	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	5994	-87.3	-9.1	60.51	-61.08	1.034164	100	1	100	-116.4	0	0	0	0	0	0	0	0	0	0	0;
	6000	617.7	351.54	411.66	-139.21	1.085651	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6036	820.2	293.16	390.4	-161.16	1.041753	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6090	60.4	5.42	36.58	-13.23	1.032192	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6092	774.9	-110.76	395.59	-155.88	1.03946	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6131	602.4	314.21	415.26	-136.33	1.141159	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6138	-121.9	2.03	16.91	-7.99	1.013527	100	1	100	-162.53	0	0	0	0	0	0	0	0	0	0	0;
	6145	190.58	-9.83	104.76	-42.6	1.021227	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	6153	94.3	-49.64	249.65	-261.82	0.996786	100	1	769.94	32.08	0	0	0	0	0	0	0	0	0	0	0;
	6168	26.3	0.51	19.86	-14.93	1.043938	100	1	53.76	0	0	0	0	0	0	0	0	0	0	0	0;
	6179	65.9	3.14	35.43	-14.41	1.038661	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6233	879.6	231.83	545.68	-189.97	1.018011	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	6239	5.3	0.44	2.24	-1.07	1.029262	100	1	5.32	0	0	0	0	0	0	0	0	0	0	0	0;
	6280	117.9	2.77	51.25	-23.49	1.019445	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6291	55.5	30.37	58.12	-16.6	1.039201	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6329	223.4	0.88	134.08	-49.15	0.994824	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	6331	48.8	4.28	37.76	-12.05	1.053365	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6332	96	-9.22	96.56	-27.95	1.064292	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	6351	22.08	15.6	38.29	-11.56	1.059908	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6368	-145.3	11.28	51.11	-23.46	1.046268	100	1	100	-193.73	0	0	0	0	0	0	0	0	0	0	0;
	6376	-27.8	-15.83	148.65	-50.73	1.047648	100	1	100	-37.07	0	0	0	0	0	0	0	0	0	0	0;
	6415	310.83	46.42	83.58	-38.78	0.987105	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	6429	171	24.15	88.83	-35.69	1.03915	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	6436	-86.5	0.45	2.26	-1.04	1.01651	100	1	100	-115.34	0	0	0	0	0	0	0	0	0	0	0;
	6459	77.4	19.01	86.17	-24.25	1.006942	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	6474	718.5	-5.42	438.72	-174.15	1.062877	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	6484	1168.8	-46.12	508.02	-227.3	1.030805	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	6516	228.2	13.22	126.73	-47.83	1.047972	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	6552	-204.2	13.04	37.58	-12.16	1.048072	100	1	100	-272.27	0	0	0	0	0	0	0	0	0	0	0;
	6632	2013.9	-440.2	1378.19	-459.09	1.062949	100	1	3000	1000	0	0	0	0	0	0	0	0	0	0	0;
	6654	742.76	-60.39	337.04	-153.26	1.02462	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	6656	249	-28.26	131.88	-52.24	1.022999	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	6681	21.3	-23.09	36.22	-35.5	1.07064	100	1	108.38	0	0	0	0	0	0	0	0	0	0	0	0;
	6734	382.7	-63.86	311.52	-132.09	1.05765	100	1	720	0	0	0	0	0	0	0	0	0	0	0	0;
	6798	949.7	162.2	429.61	-183.48	1.019012	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	6807	44.07	30.64	77.14	-22.46	1.012126	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	6816	16.53	-7.48	77.35	-22.26	1.055666	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	6820	15.9	-0.44	11.99	-9.06	1.051187	100	1	32.53	0	0	0	0	0	0	0	0	0	0	0	0;
	6831	-57.5	9.24	35.62	-14.25	1.034052	100	1	100	-76.67	0	0	0	0	0	0	0	0	0	0	0;
	6838	578	233.2	255.23	-112.31	1.043313	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	6845	102.9	17.21	53.24	-21.47	1.051715	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6852	-344.9	3.54	18.56	-6.34	1.032911	100	1	100	-459.87	0	0	0	0	0	0	0	0	0	0	0;
	6857	2463.6	267.78	1865.83	-585.95	1.014858	100	1	4000	1333.33	0	0	0	0	0	0	0	0	0	0	0;
	6877	19.8	-14.33	33.57	-32.95	1.066537	100	1	100.51	0	0	0	0	0	0	0	0	0	0	0	0;
	6888	133	4.46	189.31	-59.87	1.042814	100	1	300	88.67	0	0	0	0	0	0	0	0	0	0	0;
	6936	665.8	-86.19	388.7	-162.64	1.030076	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6947	227.5	5.11	126.09	-48.29	1.046726	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	6969	39.8	-1.88	38.53	-11.27	1.061428	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6982	-12.6	10.73	56.22	-18.44	1.038963	100	1	100	-16.8	0	0	0	0	0	0	0	0	0	0	0;
	7029	1152.7	-198.11	507.73	-226.73	1.063951	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	7036	65.8	11.91	36.16	-13.72	1.042173	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7049	55.5	17.03	58.18	-16.56	1.04735	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7056	448.85	159.6	267.36	-100.98	1.005566	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7058	50.4	7.48	37.59	-12.2	1.043279	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7075	-49.37	-1.16	13.41	-11.89	1.026663	100	1	100	-65.83	0	0	0	0	0	0	0	0	0	0	0;
	7104	234.2	4.4	133.02	-50.29	1.004388	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	7115	65.87	10.39	95.06	-83.34	1.014496	100	1	272.28	0	0	0	0	0	0	0	0	0	0	0	0;
	7137	876.4	-70.99	516.03	-219.17	1.031779	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	7159	861.3	131.58	441.04	-172.05	1.059739	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	7183	1250.6	357.75	926.47	-297.63	1.105479	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	7209	7.3	0.7	3.12	-1.46	1.052921	100	1	7.35	0	0	0	0	0	0	0	0	0	0	0	0;
	7257	21.3	-9.86	35.77	-35.23	1.030375	100	1	107.29	0	0	0	0	0	0	0	0	0	0	0	0;
	7260	32	3.47	13.73	-6.28	1.016183	100	1	32.7	10.9	0	0	0	0	0	0	0	0	0	0	0;
	7267	-39.7	48.68	89.84	-34.84	1.012113	100	1	100	-52.93	0	0	0	0	0	0	0	0	0	0	0;
	7270	766.8	140.62	394.86	-156.12	1.06851	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	7279	-681.7	6.43	14.78	-6.64	1.030612	100	1	100	-908.93	0	0	0	0	0	0	0	0	0	0	0;
	7282	2340.4	17.11	1862.29	-588.16	1.029503	100	1	4000	1333.33	0	0	0	0	0	0	0	0	0	0	0;
	7303	856.4	-40.73	384.34	-167.04	1.033047	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	7327	59.06	18.39	76.11	-23.62	1.042085	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	7328	53	26.39	37.45	-12.39	0.994679	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7329	62.8	8.26	36.26	-13.55	1.039675	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7431	735.6	-100.04	399.88	-151.38	1.01868	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	7446	13.3	-1.22	14.28	-12.55	1.012514	100	1	40.96	0	0	0	0	0	0	0	0	0	0	0	0;
	7466	-202.53	38.79	75.85	-77.06	1.057274	100	1	100	-270.04	0	0	0	0	0	0	0	0	0	0	0;
	7474	76.5	8.16	56.27	-18.41	1.028765	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7495	75.8	68.99	214.23	-59.77	1.036344	100	1	200	50.53	0	0	0	0	0	0	0	0	0	0	0;
	7504	5.2	0.46	2.24	-1.05	1.017989	100	1	5.3	0	0	0	0	0	0	0	0	0	0	0	0;
	7508	715.1	-123.77	402.15	-149.17	1.062735	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	7520	702.4	261.32	350.36	-140.73	1.033404	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	7522	508	245.55	371.11	-119.67	1.000638	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	7628	243	112.06	133.78	-50.04	1.057002	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	7630	33.09	-15.8	45.85	-40.5	1.019127	100	1	131.82	0	0	0	0	0	0	0	0	0	0	0	0;
	7641	769.2	22.1	378.21	-162.35	1.053056	100	1	880	200	0	0	0	0	0	0	0	0	0	0	0;
	7659	-39.5	-2.62	74.7	-24.78	1.035071	100	1	100	-52.67	0	0	0	0	0	0	0	0	0	0	0;
	7661	168.6	6.63	76.75	-33.79	1.039659	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	7697	228.2	13.23	126.73	-47.83	1.047945	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	7755	6.3	0.09	4.78	-3.57	1.06304	100	1	12.9	0	0	0	0	0	0	0	0	0	0	0	0;
	7761	106.8	13.91	52.75	-21.97	1.015679	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7776	1221.2	372.35	934.73	-291.52	1.068185	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	7779	345.2	164.44	281.92	-85.82	1.083645	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7808	1250.4	285.53	926.35	-297.68	1.055	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	7819	25.6	-5.01	18.9	-14.44	1.020992	100	1	51.54	0	0	0	0	0	0	0	0	0	0	0	0;
	7833	15.5	21.22	106.77	-42.81	1.02191	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	7842	165	22.74	90.08	-34.56	1.005674	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	7860	397	-48.07	380.36	-109.91	1.058289	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	7913	1368.6	153.57	920.41	-305.58	1.063382	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	7929	1101.5	-28.09	698.07	-282.17	1.050833	100	1	1600	533.33	0	0	0	0	0	0	0	0	0	0	0;
	7960	419	108.4	274.2	-93.22	1.018041	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7998	811.8	140.4	391.57	-160.95	1.042456	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	8043	-27	8.33	75.67	-23.96	1.063893	100	1	100	-36	0	0	0	0	0	0	0	0	0	0	0;
	8044	337.6	95.16	281.63	-85.71	1.052835	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	8066	-188.9	-0.1	10.73	-9.53	1.01361	100	1	100	-251.87	0	0	0	0	0	0	0	0	0	0	0;
	8087	322.12	-46.45	177.22	-80.27	1.03306	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	8109	992	788.82	Inf	-Inf	1.06101	100	1	1041.6	0	0	0	0	0	0	0	0	0	0	0	0;
	8158	153	-5.59	91.67	-33.01	1.05367	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	8159	698.6	-109.57	291.71	-137.41	1.064422	100	1	709.87	236.62	0	0	0	0	0	0	0	0	0	0	0;
	8207	1.75	-40.56	70.17	-66.45	1.030795	100	1	207.27	0	0	0	0	0	0	0	0	0	0	0	0;
	8210	1321.7	-211.01	709.44	-271.31	1.053137	100	1	1600	533.33	0	0	0	0	0	0	0	0	0	0	0;
	8222	66.8	18.43	35.76	-14.05	1.019521	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8223	401	150.56	276.81	-90.88	1.058011	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	8225	17	6.14	19.49	-5.42	1.067644	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	8240	-15.64	-18.22	36.34	-35.59	1.059307	100	1	100	-20.85	0	0	0	0	0	0	0	0	0	0	0;
	8242	358.36	-36.96	177.81	-79.93	1.01497	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	8247	253.4	-41.35	164.3	-56.85	1.059416	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	8267	203.4	-0.76	106.84	-42.56	1.045102	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	8278	97.2	9.83	53.88	-20.8	1.006884	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	8311	-8.75	5.44	17.8	-7.13	1.039615	100	1	100	-11.66	0	0	0	0	0	0	0	0	0	0	0;
	8312	1702.2	440.14	1096.33	-374.64	1.078014	100	1	2400	800	0	0	0	0	0	0	0	0	0	0	0;
	8378	33.8	1.44	17.78	-7.11	1.022564	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	8458	517.1	-37.85	474.42	-138.49	1.045078	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	8473	-135	3.1	90.21	-34.45	1.042275	100	1	100	-180	0	0	0	0	0	0	0	0	0	0	0;
	8486	78.8	12.34	34.16	-15.67	1.046146	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8515	32.3	-17.43	93.03	-31.43	1.055375	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	8522	-75.69	1.2	3.61	-1.69	1.023003	100	1	100	-100.92	0	0	0	0	0	0	0	0	0	0	0;
	8564	-65	23.28	51.91	-22.86	1.019996	100	1	100	-86.67	0	0	0	0	0	0	0	0	0	0	0;
	8605	427	176.67	273.94	-93.65	1.009283	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	8625	20.3	-5.27	34.48	-33.72	1.055846	100	1	103.03	0	0	0	0	0	0	0	0	0	0	0	0;
	8630	298.2	-25.78	124.56	-59.19	1.064508	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	8670	-138.3	19.67	37.4	-12.46	1.049183	100	1	100	-184.4	0	0	0	0	0	0	0	0	0	0	0;
	8676	-259.1	16.9	34.92	-14.96	1.018144	100	1	100	-345.47	0	0	0	0	0	0	0	0	0	0	0;
	8683	-37.28	4.87	54.32	-20.39	1.041179	100	1	100	-49.7	0	0	0	0	0	0	0	0	0	0	0;
	8702	130.69	-8.62	173.2	-47.67	1.029789	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	8721	1232.6	202.93	932.68	-293.14	1.056768	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	8729	20.89	-7.62	34.95	-30.8	1.009953	100	1	100.36	0	0	0	0	0	0	0	0	0	0	0	0;
	8771	21.7	5.77	19.07	-5.8	1.008996	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	8795	166.2	-9.75	111.7	-37.85	1.055026	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	8807	-304	5.74	35.84	-14.07	1.046456	100	1	100	-405.33	0	0	0	0	0	0	0	0	0	0	0;
	8818	44.66	-3.57	76.87	-22.75	1.051013	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	8858	42.8	-33.21	163.91	-56.98	1.027559	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	8871	-233.8	-26.8	106.62	-40.75	1.037687	100	1	100	-311.73	0	0	0	0	0	0	0	0	0	0	0;
	8872	1232.6	185.36	932.68	-293.14	1.056528	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	8875	443.8	92.44	272.77	-95.02	1.047011	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	8903	17	-0.91	28.99	-28.41	1.021928	100	1	86.73	0	0	0	0	0	0	0	0	0	0	0	0;
	8927	37.2	13.99	38.64	-11.12	1.015122	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8946	756.8	-115.78	339.05	-151.21	1.041085	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	8950	-255.7	10.92	38.23	-11.61	1.045004	100	1	100	-340.94	0	0	0	0	0	0	0	0	0	0	0;
	8956	498.4	100.47	266.17	-101.5	1.064772	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	8961	-61.8	-10.12	53.52	-50.38	1.038824	100	1	100	-82.4	0	0	0	0	0	0	0	0	0	0	0;
	8973	1624.7	-347.16	859	-367.12	1.050147	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	8976	110	12.95	95.65	-28.95	1.045593	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	8981	19.3	-14.48	32.66	-32.18	1.027038	100	1	97.98	0	0	0	0	0	0	0	0	0	0	0	0;
	8987	-21.95	-14.59	31.68	-29.92	1.052269	100	1	100	-29.27	0	0	0	0	0	0	0	0	0	0	0;
	8993	-46.19	2.46	38.98	-10.82	1.022201	100	1	100	-61.58	0	0	0	0	0	0	0	0	0	0	0;
	8997	78.2	-6.48	107.59	-101.48	1.002676	100	1	317.09	0	0	0	0	0	0	0	0	0	0	0	0;
	9001	338.71	-11.52	220.49	-86.14	1.025721	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	9039	-39.85	0.37	2.23	-1.07	1.02063	100	1	100	-53.13	0	0	0	0	0	0	0	0	0	0	0;
	9043	-10.55	0.7	17.82	-7.09	1.02106	100	1	100	-14.07	0	0	0	0	0	0	0	0	0	0	0;
	9055	-90	1.17	11.6	-10.23	0.982562	100	1	100	-120	0	0	0	0	0	0	0	0	0	0	0;
	9067	369.6	67.07	198.07	-76.34	1.045155	100	1	440	146.67	0	0	0	0	0	0	0	0	0	0	0;
	9097	40.1	1.26	17.03	-7.9	1.01907	100	1	41.16	0	0	0	0	0	0	0	0	0	0	0	0;
	9101	698.8	-2.05	442.99	-169.92	1.026751	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	9108	-239.23	0.54	18.65	-14.16	1.04489	100	1	100	-318.97	0	0	0	0	0	0	0	0	0	0	0;
	9137	42.7	24.59	335.96	-93.23	1.048734	100	1	100	28.47	0	0	0	0	0	0	0	0	0	0	0;
	9140	-37.2	35.2	73.19	-26.48	1.049339	100	1	100	-49.6	0	0	0	0	0	0	0	0	0	0	0;
	9150	860.8	128.28	439.38	-173.28	1.037874	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	9174	-331.4	165.94	175	-49.16	1.020141	100	1	100	-441.87	0	0	0	0	0	0	0	0	0	0	0;
	9180	72	6.19	77.69	-21.97	1.057766	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	9239	419	108.18	274.2	-93.22	1.018061	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
];

%% branch data
%	fbus	tbus	r	x	b	rateA	rateB	rateC	ratio	angle	status	angmin	angmax
mpc.branch = [
	5147	3097	0.0006	0.00616	0	823	0	0	0	0	1	-360	360;
	5147	8763	0.0005	0.0074	0	921	0	0	0	0	1	-360	360;
	427	5425	7e-05	0.00057	0	1711	0	0	0	0	1	-360	360;
	427	4704	6e-05	0.00056	0	1514	0	0	0	0	1	-360	360;
	5960	7431	0.003349	0.01656	0	0	0	0	0	0	1	-360	360;
	5960	3186	0.003409	0.016211	0	0	0	0	0	0	1	-360	360;
	4481	7431	0.003479	0.01586	0	0	0	0	0	0	1	-360	360;
	4481	3186	0.003099	0.017969	0	0	0	0	0	0	1	-360	360;
	5205	7833	0.00236	0.01418	0	553	0	0	0	0	1	-360	360;
	5205	7770	0.002169	0.015909	0	514	0	0	0	0	1	-360	360;
	5773	6542	0.00014	0.0012	0	0	0	0	0	0	1	-360	360;
	5773	7840	0.000211	0.001539	0	0	0	0	0	0	1	-360	360;
	5214	6542	0.000401	0.004901	0	0	0	0	0	0	1	-360	360;
	5214	7840	0.000219	0.00151	0	533	0	0	0	0	1	-360	360;
	2971	6069	0.00052	0.00407	0	1481	0	0	0	0	1	-360	360;
	2971	3445	0.00034	0.00502	0	1843	0	0	0	0	1	-360	360;
	7561	6069	0.00065	0.00698	0	1711	0	0	0	0	1	-360	360;
	7561	8825	0.00059	0.00806	0	1678	0	0	0	0	1	-360	360;
	7556	7058	0.00374	0.02095	0	800	0	0	0	0	1	-360	360;
	7556	4324	0.00511	0.019461	0	800	0	0	0	0	1	-360	360;
	132	2962	0.00064	0.00715	0	1481	0	0	0	0	1	-360	360;
	132	4393	0.00089	0.00712	0	1283	0	0	0	0	1	-360	360;
	5044	693	0.00048	0.00555	0	987	0	0	0	0	1	-360	360;
	5044	7860	0.00047	0.00566	0	1086	0	0	0	0	1	-360	360;
	6883	693	0.00046	0.00524	0	1053	0	0	0	0	1	-360	360;
	6883	7860	0.00052	0.00514	0	1053	0	0	0	0	1	-360	360;
	6494	4945	0.00029	0.00298	0	1613	0	0	0	0	1	-360	360;
	6494	6080	0.0003	0.00291	0	1514	0	0	0	0	1	-360	360;
	7146	7388	0.00042	0.02666	0	1185	0	0	0	0	1	-360	360;
	4735	5239	0.00058	0.00534	0	1876	0	0	0	0	1	-360	360;
	4735	3145	0.00054	0.00589	0	1843	0	0	0	0	1	-360	360;
	3615	8731	0.00036	0.00485	0	0	0	0	0	0	1	-360	360;
	3615	6246	0.00042	0.00415	0	0	0	0	0	0	1	-360	360;
	2549	7829	0.003349	0.02445	0	553	0	0	0	0	1	-360	360;
	2549	4951	0.003281	0.02493	0	553	0	0	0	0	1	-360	360;
	7004	3370	0	0.00046	0	757	0	0	0	0	1	-360	360;
	8381	4795	0.00053	0.00587	0	1777	0	0	0	0	1	-360	360;
	8381	5562	0.00055	0.00666	0	1711	0	0	0	0	1	-360	360;
	3409	1716	0.00051	0.00614	0	0	0	0	0	0	1	-360	360;
	3409	6246	0.00043	0.00406	0	0	0	0	0	0	1	-360	360;
	1000	6145	0.002211	0.014419	0	438	0	0	0	0	1	-360	360;
	1000	3186	0.002169	0.013349	0	457	0	0	0	0	1	-360	360;
	6060	5289	0.00033	0.00101	0	823	0	0	0	0	1	-360	360;
	7965	5289	0.00028	0.00104	0	757	0	0	0	0	1	-360	360;
	4363	4507	0.00057	0.0071	0	1613	0	0	0	0	1	-360	360;
	4363	6246	0.00057	0.00697	0	1613	0	0	0	0	1	-360	360;
	425	7526	0.0001	0.00119	0	1810	0	0	0	0	1	-360	360;
	425	5674	0.0001	0.0013	0	1613	0	0	0	0	1	-360	360;
	4799	8581	9e-05	0.00121	0	1514	0	0	0	0	1	-360	360;
	4799	4337	0.0001	0.00134	0	1251	0	0	0	0	1	-360	360;
	4968	2154	0.00039	0.00474	0	1843	0	0	0	0	1	-360	360;
	4968	1747	0.00045	0.00467	0	1448	0	0	0	0	1	-360	360;
	7699	2154	0.00052	0.00567	0	1645	0	0	0	0	1	-360	360;
	7699	1689	0.00054	0.00667	0	1316	0	0	0	0	1	-360	360;
	3097	659	0.00073	0.00905	0	1185	0	0	0	0	1	-360	360;
	9024	4929	0.00192	0.02296	0	889	0	0	0	0	1	-360	360;
	1815	6542	0.01431	0.05899	0	0	0	0	0	0	1	-360	360;
	6542	1445	0.02607	0.07168	0	0	0	0	0	0	1	-360	360;
	6069	6833	0.00054	0.00465	0	1580	0	0	0	0	1	-360	360;
	6069	2268	0.00126	0.01124	0	1613	0	0	0	0	1	-360	360;
	6069	7180	0.00048	0.00527	0	1547	0	0	0	0	1	-360	360;
	6069	6293	0.00082	0.00938	0	1580	0	0	0	0	1	-360	360;
	1968	9192	0.00028	0.003502	0	0	0	0	0	0	1	-360	360;
	1968	3493	0.691467	4.414711	0	0	0	0	0	0	1	-360	360;
	1968	955	0.02004	0.0916	0	0	0	0	0	0	1	-360	360;
	1968	228	0.841911	4.768444	0	0	0	0	0	0	1	-360	360;
	1968	6826	0.026	0.08948	0	0	0	0	0	0	1	-360	360;
	1968	8964	0.002631	0.011889	0	299	0	0	0	0	1	-360	360;
	1968	5776	0.7232	1.944	0	0	0	0	0	0	1	-360	360;
	9192	955	0.156169	0.546756	0	0	0	0	0	0	1	-360	360;
	9192	6826	0.168809	0.580622	0	0	0	0	0	0	1	-360	360;
	9192	8964	0.019222	0.068209	0	0	0	0	0	0	1	-360	360;
	8574	1616	0.00039	0.00386	0	1415	0	0	0	0	1	-360	360;
	8574	1163	0.00096	0.01138	0	1448	0	0	0	0	1	-360	360;
	8574	9064	0.00037	0.00325	0	1415	0	0	0	0	1	-360	360;
	8574	8921	0.00029	0.00339	0	1580	0	0	0	0	1	-360	360;
	3493	2299	0.009458	0.158871	0	0	0	0	0	0	1	-360	360;
	2299	4423	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6826	2299	0.312951	7.130667	0	0	0	0	0	0	1	-360	360;
	7563	2299	0.247649	1.075911	0	0	0	0	0	0	1	-360	360;
	4427	2299	0.06056	0.18612	0	0	0	0	0	0	1	-360	360;
	8335	2299	2.142489	4.347733	0	0	0	0	0	0	1	-360	360;
	5155	2299	0.07132	1.068889	0	0	0	0	0	0	1	-360	360;
	271	2299	0.02268	0.08652	0	0	0	0	0	0	1	-360	360;
	2449	2299	2.007778	4.494222	0	0	0	0	0	0	1	-360	360;
	4665	2299	0.006178	0.028938	0	0	0	0	0	0	1	-360	360;
	4495	2299	0.027711	0.260951	0	0	0	0	0	0	1	-360	360;
	3493	5587	0	0.000222	0	299	0	0	0	0	1	-360	360;
	4586	3493	0.00692	0.337378	0	0	0	0	0	0	1	-360	360;
	228	3493	0.025991	0.10048	0	0	0	0	0	0	1	-360	360;
	6826	3493	0.083711	0.8648	0	0	0	0	0	0	1	-360	360;
	7563	3493	0.447378	1.862089	0	0	0	0	0	0	1	-360	360;
	4427	3493	0.062169	0.153191	0	0	0	0	0	0	1	-360	360;
	8335	3493	2.492533	4.725778	0	0	0	0	0	0	1	-360	360;
	5155	3493	0.020271	0.08244	0	0	0	0	0	0	1	-360	360;
	271	3493	0.11416	1.777556	0	0	0	0	0	0	1	-360	360;
	5776	3493	0.574978	2.992578	0	0	0	0	0	0	1	-360	360;
	2449	3493	1.887067	4.964444	0	0	0	0	0	0	1	-360	360;
	4665	3493	0.250778	0.817422	0	0	0	0	0	0	1	-360	360;
	4495	3493	0.007049	0.029609	0	0	0	0	0	0	1	-360	360;
	5587	4586	0.003049	0.020218	0	0	0	0	0	0	1	-360	360;
	4423	4586	0.002911	0.020262	0	0	0	0	0	0	1	-360	360;
	4929	2870	4e-05	0.00041	0	757	0	0	0	0	1	-360	360;
	4929	659	0.00087	0.01087	0	1283	0	0	0	0	1	-360	360;
	4929	1037	4e-05	0.00046	0	757	0	0	0	0	1	-360	360;
	4929	659	0.00095	0.00924	0	1481	0	0	0	0	1	-360	360;
	1815	5210	0.528017	2.636157	0	0	0	0	0	0	1	-360	360;
	1815	1445	0.032031	0.128659	0	0	0	0	0	0	1	-360	360;
	1616	7762	0.00044	0.0032	0	1645	0	0	0	0	1	-360	360;
	1616	2520	0.00046	0.00489	0	1711	0	0	0	0	1	-360	360;
	9064	7762	0.0003	0.00352	0	1580	0	0	0	0	1	-360	360;
	9064	89	0.00051	0.00454	0	1645	0	0	0	0	1	-360	360;
	1317	659	0.00068	0.00746	0	1613	0	0	0	0	1	-360	360;
	1317	8605	0.00093	0.01022	0	1744	0	0	0	0	1	-360	360;
	8605	1367	4e-05	0.0004	0	1481	0	0	0	0	1	-360	360;
	8605	8921	0.00078	0.00698	0	1744	0	0	0	0	1	-360	360;
	1163	1676	4e-05	0.00043	0	1580	0	0	0	0	1	-360	360;
	1163	7051	0.00025	0.00309	0	1678	0	0	0	0	1	-360	360;
	913	7762	0.00047	0.00538	0	1613	0	0	0	0	1	-360	360;
	2107	7762	0.00046	0.00548	0	1744	0	0	0	0	1	-360	360;
	2107	5996	0.00016	0.00173	0	1843	0	0	0	0	1	-360	360;
	2107	6293	0.00166	0.01471	0	1744	0	0	0	0	1	-360	360;
	913	7762	0.00051	0.00512	0	1777	0	0	0	0	1	-360	360;
	6704	8921	0.00021	0.00248	0	1514	0	0	0	0	1	-360	360;
	228	4586	0.004391	0.021689	0	286	0	0	0	0	1	-360	360;
	2441	6293	0.00024	0.00289	0	1645	0	0	0	0	1	-360	360;
	955	3506	0	0.000222	0	338	0	0	0	0	1	-360	360;
	955	8964	0.01788	0.100471	0	0	0	0	0	0	1	-360	360;
	3506	2908	0.006458	0.039982	0	0	0	0	0	0	1	-360	360;
	228	6826	0.106151	0.993333	0	0	0	0	0	0	1	-360	360;
	228	5776	0.699778	3.232311	0	0	0	0	0	0	1	-360	360;
	8964	6826	0.257058	0.43592	0	0	0	0	0	0	1	-360	360;
	5776	6826	0.002769	0.011449	0	0	0	0	0	0	1	-360	360;
	659	8329	0.00034	0.00305	0	1580	0	0	0	0	1	-360	360;
	659	7051	0.00123	0.01255	0	1843	0	0	0	0	1	-360	360;
	659	9239	6e-05	0.00065	0	1415	0	0	0	0	1	-360	360;
	659	7960	6e-05	0.00063	0	1514	0	0	0	0	1	-360	360;
	659	6233	6e-05	0.00055	0	1744	0	0	0	0	1	-360	360;
	659	5416	0.00079	0.00968	0	1349	0	0	0	0	1	-360	360;
	659	6798	7e-05	0.00079	0	0	0	0	0	0	1	-360	360;
	792	3279	0	0.000222	0	377	0	0	0	0	1	-360	360;
	7563	792	2.064178	7.545333	0	0	0	0	0	0	1	-360	360;
	7279	792	0.153711	0.482489	0	0	0	0	0	0	1	-360	360;
	8335	792	0.015809	0.12368	0	0	0	0	0	0	1	-360	360;
	2449	792	0.057978	0.363422	0	0	0	0	0	0	1	-360	360;
	4665	792	0.6932	2.196889	0	0	0	0	0	0	1	-360	360;
	4495	792	0.560356	2.395867	0	0	0	0	0	0	1	-360	360;
	5416	2267	0.00024	0.00213	0	757	0	0	0	0	1	-360	360;
	5416	7637	0.00093	0.00884	0	1283	0	0	0	0	1	-360	360;
	8964	2168	0	0.000222	0	312	0	0	0	0	1	-360	360;
	3659	5996	0.00018	0.00204	0	1777	0	0	0	0	1	-360	360;
	5762	5996	0.00018	0.00208	0	757	0	0	0	0	1	-360	360;
	3242	5097	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7563	3242	0.02068	0.093391	0	0	0	0	0	0	1	-360	360;
	4427	3242	0.777778	7.408	0	0	0	0	0	0	1	-360	360;
	3242	7279	0.00884	0.07292	0	0	0	0	0	0	1	-360	360;
	271	3242	0.131671	8.295556	0	0	0	0	0	0	1	-360	360;
	3242	1611	0.082462	0.568222	0	0	0	0	0	0	1	-360	360;
	5097	1611	0.0042	0.02624	0	0	0	0	0	0	1	-360	360;
	8181	8179	0.00031	0.00317	0	1251	0	0	0	0	1	-360	360;
	8181	7762	0.00043	0.00498	0	1481	0	0	0	0	1	-360	360;
	9025	7762	0.00042	0.00508	0	1711	0	0	0	0	1	-360	360;
	7563	4427	0.110671	0.773067	0	0	0	0	0	0	1	-360	360;
	7563	7279	0.00912	0.037609	0	0	0	0	0	0	1	-360	360;
	7563	8335	0.098049	0.598711	0	0	0	0	0	0	1	-360	360;
	7563	5155	0.019031	1.160356	0	0	0	0	0	0	1	-360	360;
	7563	271	0.024071	0.915733	0	0	0	0	0	0	1	-360	360;
	2449	7563	1.484267	4.479111	0	0	0	0	0	0	1	-360	360;
	4665	7563	0.011578	0.063271	0	0	0	0	0	0	1	-360	360;
	4495	7563	0.654889	1.839111	0	0	0	0	0	0	1	-360	360;
	7563	1611	0.022049	0.10076	0	0	0	0	0	0	1	-360	360;
	4427	7279	0.459867	2.604711	0	0	0	0	0	0	1	-360	360;
	4427	8335	2.320222	6.088	0	0	0	0	0	0	1	-360	360;
	4427	5155	0.006369	0.048089	0	0	0	0	0	0	1	-360	360;
	4427	271	0.006631	0.0432	0	0	0	0	0	0	1	-360	360;
	2449	4427	2.567244	6.767111	0	0	0	0	0	0	1	-360	360;
	4665	4427	0.309889	1.214178	0	0	0	0	0	0	1	-360	360;
	4495	4427	0.009191	0.043289	0	0	0	0	0	0	1	-360	360;
	4427	1611	0.691022	5.239556	0	0	0	0	0	0	1	-360	360;
	5210	1445	0.002669	0.01718	0	0	0	0	0	0	1	-360	360;
	8179	5509	0.00074	0.00694	0	1251	0	0	0	0	1	-360	360;
	8179	7762	0.00068	0.008	0	1547	0	0	0	0	1	-360	360;
	7279	4014	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8335	7279	0.00524	0.030631	0	0	0	0	0	0	1	-360	360;
	5155	7279	0.33932	3.814667	0	0	0	0	0	0	1	-360	360;
	271	7279	0.22696	2.614933	0	0	0	0	0	0	1	-360	360;
	2449	7279	0.190671	0.719378	0	0	0	0	0	0	1	-360	360;
	4665	7279	1.050222	2.105556	0	0	0	0	0	0	1	-360	360;
	4495	7279	1.431644	4.294578	0	0	0	0	0	0	1	-360	360;
	1611	7279	0.022618	0.161089	0	0	0	0	0	0	1	-360	360;
	5509	1579	0.00088	0.00895	0	1810	0	0	0	0	1	-360	360;
	8335	1531	0	0.000222	0	299	0	0	0	0	1	-360	360;
	5155	8335	1.827778	7.768	0	0	0	0	0	0	1	-360	360;
	2449	8335	0.020849	0.153382	0	0	0	0	0	0	1	-360	360;
	4665	8335	0.156231	0.656267	0	0	0	0	0	0	1	-360	360;
	4495	8335	0.176498	0.891556	0	0	0	0	0	0	1	-360	360;
	7762	2268	0.00067	0.00777	0	1613	0	0	0	0	1	-360	360;
	5155	271	0.004578	0.061889	0	0	0	0	0	0	1	-360	360;
	5155	2908	0.01348	0.05932	0	0	0	0	0	0	1	-360	360;
	4665	5155	0.066391	2.802622	0	0	0	0	0	0	1	-360	360;
	4495	5155	0.194258	2.220444	0	0	0	0	0	0	1	-360	360;
	5155	1611	0.008151	0.072822	0	0	0	0	0	0	1	-360	360;
	271	2908	0.016502	0.055382	0	0	0	0	0	0	1	-360	360;
	4665	271	0.0806	4.633778	0	0	0	0	0	0	1	-360	360;
	4495	271	0.831556	3.729644	0	0	0	0	0	0	1	-360	360;
	271	1611	0.007951	0.074151	0	0	0	0	0	0	1	-360	360;
	1579	5848	0.0004	0.00362	0	1613	0	0	0	0	1	-360	360;
	7051	8847	0.00015	0.00177	0	1580	0	0	0	0	1	-360	360;
	5776	8229	0	0.000222	0	0	0	0	0	0	1	-360	360;
	317	6233	2e-05	0.00022	0	1678	0	0	0	0	1	-360	360;
	4665	2449	0.02864	0.172902	0	0	0	0	0	0	1	-360	360;
	2449	4495	0.017591	0.1404	0	0	0	0	0	0	1	-360	360;
	4665	4495	0.00752	0.031711	0	0	0	0	0	0	1	-360	360;
	1611	8420	0	0.000222	0	546	0	0	0	0	1	-360	360;
	6561	1294	0.003289	0.019091	0	457	0	0	0	0	1	-360	360;
	6561	1666	0.004169	0.025349	0	476	0	0	0	0	1	-360	360;
	6561	4096	0.00439	0.023909	0	457	0	0	0	0	1	-360	360;
	4717	4241	0.00156	0.01956	0	1316	0	0	0	0	1	-360	360;
	4241	7395	0.00029	0.00391	0	1843	0	0	0	0	1	-360	360;
	4241	5873	0.00179	0.02522	0	1645	0	0	0	0	1	-360	360;
	1479	4241	0.00079	0.00756	0	1843	0	0	0	0	1	-360	360;
	4241	1475	0.00027	0.00281	0	1645	0	0	0	0	1	-360	360;
	4241	1475	0.00026	0.00286	0	1547	0	0	0	0	1	-360	360;
	2248	264	0.000281	0.00213	0	0	0	0	0	0	1	-360	360;
	2248	8530	0.000419	0.00299	0	0	0	0	0	0	1	-360	360;
	3744	4806	0.001659	0.009091	0	0	0	0	0	0	1	-360	360;
	3744	6239	0.00439	0.03324	0	381	0	0	0	0	1	-360	360;
	5991	8671	0.000841	0.00612	0	400	0	0	0	0	1	-360	360;
	5991	201	0.003341	0.022841	0	533	0	0	0	0	1	-360	360;
	5991	3512	0.00581	0.04232	0	0	0	0	0	0	1	-360	360;
	5991	1430	0.00495	0.02919	0	438	0	0	0	0	1	-360	360;
	5991	5592	0.01789	0.07795	0	381	0	0	0	0	1	-360	360;
	5991	4953	0.00186	0.014289	0	381	0	0	0	0	1	-360	360;
	5991	2355	0.002281	0.014401	0	0	0	0	0	0	1	-360	360;
	4510	2957	0.00031	0.00132	0	0	0	0	0	0	1	-360	360;
	4510	6936	0.001349	0.0077	0	381	0	0	0	0	1	-360	360;
	5697	5933	0.00586	0.02574	0	476	0	0	0	0	1	-360	360;
	5697	6040	0.00556	0.020211	0	591	0	0	0	0	1	-360	360;
	5189	2917	0.00026	0.001159	0	553	0	0	0	0	1	-360	360;
	2917	6654	0.000599	0.002419	0	591	0	0	0	0	1	-360	360;
	6679	7139	0.0005	0.00386	0	0	0	0	0	0	1	-360	360;
	6679	6134	0.00107	0.008289	0	533	0	0	0	0	1	-360	360;
	6679	6134	0.001091	0.00811	0	0	0	0	0	0	1	-360	360;
	7413	6679	0.00038	0.00255	0	533	0	0	0	0	1	-360	360;
	5189	6456	0.000271	0.000979	0	0	0	0	0	0	1	-360	360;
	2358	1638	7e-05	0.00076	0	0	0	0	0	0	1	-360	360;
	2358	8571	0.000521	0.003901	0	0	0	0	0	0	1	-360	360;
	4717	8247	0.0016	0.01839	0	1251	0	0	0	0	1	-360	360;
	5787	1393	0.007651	0.058781	0	476	0	0	0	0	1	-360	360;
	5787	6243	0.00394	0.02182	0	0	0	0	0	0	1	-360	360;
	5787	8338	0.003461	0.03005	0	553	0	0	0	0	1	-360	360;
	5787	751	0.00337	0.01831	0	438	0	0	0	0	1	-360	360;
	8285	4293	0.000581	0.00194	0	495	0	0	0	0	1	-360	360;
	7392	4293	0.000539	0.001719	0	0	0	0	0	0	1	-360	360;
	6632	930	0.00249	0.03075	0	1580	0	0	0	0	1	-360	360;
	930	4407	0.0014	0.01409	0	1678	0	0	0	0	1	-360	360;
	7834	6596	0.008	0.050031	0	381	0	0	0	0	1	-360	360;
	7834	5592	0.002211	0.01562	0	572	0	0	0	0	1	-360	360;
	1418	3633	0.00281	0.021831	0	476	0	0	0	0	1	-360	360;
	1418	7820	0.000711	0.004919	0	0	0	0	0	0	1	-360	360;
	2921	6093	0.001841	0.00738	0	0	0	0	0	0	1	-360	360;
	2921	3487	0.002849	0.008979	0	0	0	0	0	0	1	-360	360;
	8353	3633	0.00232	0.014099	0	0	0	0	0	0	1	-360	360;
	8353	3633	0.00237	0.0138	0	0	0	0	0	0	1	-360	360;
	8353	6904	0.00081	0.004591	0	476	0	0	0	0	1	-360	360;
	8353	6625	0.002219	0.010469	0	0	0	0	0	0	1	-360	360;
	6261	1872	0.001081	0.00851	0	0	0	0	0	0	1	-360	360;
	1872	7973	0.00087	0.006969	0	0	0	0	0	0	1	-360	360;
	1872	7973	0.00088	0.006831	0	0	0	0	0	0	1	-360	360;
	1872	4651	0.000349	0.00237	0	0	0	0	0	0	1	-360	360;
	710	6499	0.00232	0.00924	0	381	0	0	0	0	1	-360	360;
	710	3792	0.00812	0.03669	0	400	0	0	0	0	1	-360	360;
	2193	6717	0.003521	0.02825	0	476	0	0	0	0	1	-360	360;
	2193	4752	0.00264	0.017651	0	0	0	0	0	0	1	-360	360;
	7644	6499	0.00486	0.02149	0	457	0	0	0	0	1	-360	360;
	7644	6769	0.006831	0.034669	0	419	0	0	0	0	1	-360	360;
	7644	7075	0.017211	0.108151	0	381	0	0	0	0	1	-360	360;
	7644	8213	0.00063	0.00475	0	419	0	0	0	0	1	-360	360;
	7644	8213	0.00062	0.004841	0	0	0	0	0	0	1	-360	360;
	6632	2507	0.0009	0.01092	0	1843	0	0	0	0	1	-360	360;
	2507	6917	0.00124	0.01096	0	1777	0	0	0	0	1	-360	360;
	3095	2951	0.003211	0.00176	0	0	0	0	0	0	1	-360	360;
	3095	7820	0.001521	0.008031	0	0	0	0	0	0	1	-360	360;
	5425	1898	0.0005	0.00441	0	1415	0	0	0	0	1	-360	360;
	5425	1498	0.00247	0.02573	0	1711	0	0	0	0	1	-360	360;
	5425	1280	0.00042	0.00367	0	1283	0	0	0	0	1	-360	360;
	7431	4279	0.0023	0.01644	0	457	0	0	0	0	1	-360	360;
	7431	4279	0.00226	0.01676	0	457	0	0	0	0	1	-360	360;
	7431	4279	0.0023	0.0162	0	476	0	0	0	0	1	-360	360;
	7431	4279	0.00225	0.015	0	419	0	0	0	0	1	-360	360;
	2343	5946	0.007281	0.035729	0	438	0	0	0	0	1	-360	360;
	2236	5946	0.001919	0.00945	0	438	0	0	0	0	1	-360	360;
	3821	1173	0.005211	0.02981	0	514	0	0	0	0	1	-360	360;
	3821	6936	0.000659	0.00586	0	0	0	0	0	0	1	-360	360;
	3821	6936	0.000591	0.00664	0	0	0	0	0	0	1	-360	360;
	3821	2946	0.011969	0.07424	0	533	0	0	0	0	1	-360	360;
	6261	3065	0.000531	0.003919	0	0	0	0	0	0	1	-360	360;
	4124	3378	0.001521	0.00718	0	438	0	0	0	0	1	-360	360;
	4124	3378	0.001211	0.00613	0	476	0	0	0	0	1	-360	360;
	4124	4809	0.00676	0.03725	0	419	0	0	0	0	1	-360	360;
	4124	4750	0.00399	0.018091	0	381	0	0	0	0	1	-360	360;
	6663	5217	0.000341	0.00339	0	0	0	0	0	0	1	-360	360;
	6663	4096	0.001271	0.0102	0	0	0	0	0	0	1	-360	360;
	6663	4096	0.00125	0.010401	0	0	0	0	0	0	1	-360	360;
	4855	509	0.00261	0.016651	0	0	0	0	0	0	1	-360	360;
	9162	509	0.000219	0.002409	0	514	0	0	0	0	1	-360	360;
	6769	6964	0.00601	0.02814	0	438	0	0	0	0	1	-360	360;
	6964	7002	0.008969	0.03787	0	0	0	0	0	0	1	-360	360;
	6964	6879	0.00232	0.01094	0	438	0	0	0	0	1	-360	360;
	6632	6917	0.00196	0.02382	0	1645	0	0	0	0	1	-360	360;
	6632	8870	0.00422	0.03598	0	1218	0	0	0	0	1	-360	360;
	6632	8870	0.0043	0.03522	0	1283	0	0	0	0	1	-360	360;
	8504	6632	0.00208	0.01691	0	1547	0	0	0	0	1	-360	360;
	6632	2730	0.00544	0.05439	0	1251	0	0	0	0	1	-360	360;
	6632	7508	0.00239	0.02598	0	1316	0	0	0	0	1	-360	360;
	6632	385	0.00396	0.03096	0	1711	0	0	0	0	1	-360	360;
	6632	385	0.00353	0.03509	0	1547	0	0	0	0	1	-360	360;
	3650	8858	0.000531	0.003729	0	514	0	0	0	0	1	-360	360;
	3650	8858	0.00038	0.0025	0	0	0	0	0	0	1	-360	360;
	3650	8858	0.00039	0.00244	0	0	0	0	0	0	1	-360	360;
	3650	4856	0.00836	0.044521	0	400	0	0	0	0	1	-360	360;
	1618	1060	0.00264	0.03029	0	1382	0	0	0	0	1	-360	360;
	1618	8806	0.00396	0.04589	0	757	0	0	0	0	1	-360	360;
	8547	6669	0.00411	0.021729	0	400	0	0	0	0	1	-360	360;
	8547	7398	0.00607	0.0285	0	457	0	0	0	0	1	-360	360;
	8124	2496	0.00075	0.0047	0	0	0	0	0	0	1	-360	360;
	8419	2496	0.00114	0.007969	0	553	0	0	0	0	1	-360	360;
	5125	2916	0.00036	0.00324	0	1316	0	0	0	0	1	-360	360;
	5125	8973	0.00094	0.00874	0	1218	0	0	0	0	1	-360	360;
	2471	8888	0.00949	0.05145	0	476	0	0	0	0	1	-360	360;
	2471	4906	0.011719	0.0337	0	400	0	0	0	0	1	-360	360;
	3076	2471	0.0022	0.012331	0	419	0	0	0	0	1	-360	360;
	4182	2235	0.00038	0.00448	0	1382	0	0	0	0	1	-360	360;
	4182	5873	0.00098	0.01126	0	987	0	0	0	0	1	-360	360;
	42	5919	0.00055	0.00621	0	1777	0	0	0	0	1	-360	360;
	42	2096	0.00163	0.01287	0	1711	0	0	0	0	1	-360	360;
	6292	6127	0.0018	0.01886	0	0	0	0	0	0	1	-360	360;
	6292	6127	0.0047	0.024969	0	0	0	0	0	0	1	-360	360;
	6292	2658	0.009711	0.059669	0	400	0	0	0	0	1	-360	360;
	6292	5191	0.005271	0.03451	0	0	0	0	0	0	1	-360	360;
	929	1175	0.016169	0.11936	0	419	0	0	0	0	1	-360	360;
	929	1344	0.00226	0.014789	0	419	0	0	0	0	1	-360	360;
	8671	1421	0.00151	0.01081	0	591	0	0	0	0	1	-360	360;
	709	4652	0.00228	0.0217	0	1810	0	0	0	0	1	-360	360;
	709	3587	0.00137	0.01593	0	1711	0	0	0	0	1	-360	360;
	8159	3798	0.00099	0.0121	0	1086	0	0	0	0	1	-360	360;
	8159	2261	0	0.01881	0	1119	0	0	0	0	1	-360	360;
	8159	7755	0.00143	0.0163	0	1185	0	0	0	0	1	-360	360;
	8159	7287	0.00057	0.00562	0	1119	0	0	0	0	1	-360	360;
	8159	7287	0.00055	0.00572	0	1185	0	0	0	0	1	-360	360;
	8159	6496	0.00057	0.00658	0	1448	0	0	0	0	1	-360	360;
	8159	5372	0.00075	0.0082	0	1251	0	0	0	0	1	-360	360;
	8159	807	0.00253	0.02153	0	1316	0	0	0	0	1	-360	360;
	201	8582	0.00236	0.011219	0	533	0	0	0	0	1	-360	360;
	3193	7053	0.00173	0.0154	0	1415	0	0	0	0	1	-360	360;
	3193	2764	0.00385	0.04172	0	1185	0	0	0	0	1	-360	360;
	3193	2764	0.00377	0.04251	0	1152	0	0	0	0	1	-360	360;
	3193	1498	0.00199	0.02121	0	1580	0	0	0	0	1	-360	360;
	3193	1133	0.00236	0.01933	0	1185	0	0	0	0	1	-360	360;
	3193	1133	0.00241	0.01891	0	1119	0	0	0	0	1	-360	360;
	2301	3427	0.000521	0.004729	0	0	0	0	0	0	1	-360	360;
	2301	8575	0.00094	0.005581	0	0	0	0	0	0	1	-360	360;
	4644	1086	0.00134	0.01324	0	1218	0	0	0	0	1	-360	360;
	4644	2849	0.00073	0.00638	0	1283	0	0	0	0	1	-360	360;
	1458	2459	0.00248	0.02774	0	1810	0	0	0	0	1	-360	360;
	1458	477	0.00152	0.013	0	1580	0	0	0	0	1	-360	360;
	1458	477	0.00155	0.01272	0	1678	0	0	0	0	1	-360	360;
	1458	8210	0.00117	0.01301	0	1349	0	0	0	0	1	-360	360;
	1458	3113	0.00197	0.02054	0	1119	0	0	0	0	1	-360	360;
	1458	8987	0.00157	0.01909	0	1613	0	0	0	0	1	-360	360;
	3065	4450	0.00107	0.008669	0	0	0	0	0	0	1	-360	360;
	4450	1501	0.000469	0.00314	0	0	0	0	0	0	1	-360	360;
	1185	706	0.000409	0.00305	0	0	0	0	0	0	1	-360	360;
	1240	706	0.00569	0.032979	0	0	0	0	0	0	1	-360	360;
	3500	8871	0.00087	0.006771	0	514	0	0	0	0	1	-360	360;
	3500	8582	0.002151	0.014039	0	533	0	0	0	0	1	-360	360;
	6359	6783	9.9e-05	0.00043	0	553	0	0	0	0	1	-360	360;
	6454	6359	0.000151	0.0005	0	0	0	0	0	0	1	-360	360;
	7833	8438	0.001531	0.00812	0	0	0	0	0	0	1	-360	360;
	7833	8438	0.00113	0.00857	0	0	0	0	0	0	1	-360	360;
	4253	7196	0.00226	0.01868	0	514	0	0	0	0	1	-360	360;
	4253	4651	0.000401	0.00239	0	0	0	0	0	0	1	-360	360;
	4230	1912	0.001169	0.008349	0	514	0	0	0	0	1	-360	360;
	4230	6315	0.00324	0.02312	0	533	0	0	0	0	1	-360	360;
	621	4360	0.006031	0.034789	0	495	0	0	0	0	1	-360	360;
	621	808	0.0103	0.05269	0	438	0	0	0	0	1	-360	360;
	2962	6917	0.00129	0.01494	0	1481	0	0	0	0	1	-360	360;
	2962	6917	0.00126	0.01523	0	1349	0	0	0	0	1	-360	360;
	4958	4906	0.001031	0.00601	0	0	0	0	0	0	1	-360	360;
	4958	4906	0.00105	0.00588	0	0	0	0	0	0	1	-360	360;
	2642	5823	0.00197	0.01719	0	1119	0	0	0	0	1	-360	360;
	2642	1947	0.00114	0.01237	0	1382	0	0	0	0	1	-360	360;
	2642	3966	0.00314	0.02868	0	1349	0	0	0	0	1	-360	360;
	2642	1074	0.00294	0.03326	0	1218	0	0	0	0	1	-360	360;
	2642	1074	0.00288	0.0308	0	1448	0	0	0	0	1	-360	360;
	2642	807	0.00088	0.02821	0	1086	0	0	0	0	1	-360	360;
	4165	6659	0.00152	0.01702	0	889	0	0	0	0	1	-360	360;
	4165	7303	0.00022	0.00209	0	1481	0	0	0	0	1	-360	360;
	1793	2668	0.00087	0.005	0	0	0	0	0	0	1	-360	360;
	1793	8338	0.000711	0.00439	0	381	0	0	0	0	1	-360	360;
	1793	8338	0.000729	0.004289	0	419	0	0	0	0	1	-360	360;
	4663	6443	0.001031	0.00549	0	0	0	0	0	0	1	-360	360;
	4663	519	0.003219	0.01512	0	0	0	0	0	0	1	-360	360;
	4663	595	0.007271	0.031419	0	381	0	0	0	0	1	-360	360;
	4663	3291	0.02145	0.085781	0	438	0	0	0	0	1	-360	360;
	6138	7504	0.002841	0.017771	0	0	0	0	0	0	1	-360	360;
	6138	2896	0.003331	0.017289	0	400	0	0	0	0	1	-360	360;
	5280	2343	0.01524	0.064091	0	514	0	0	0	0	1	-360	360;
	8534	5121	0.000831	0.0052	0	514	0	0	0	0	1	-360	360;
	5121	9039	0.00164	0.0103	0	495	0	0	0	0	1	-360	360;
	3961	4238	0.00139	0.00864	0	381	0	0	0	0	1	-360	360;
	3961	1411	0.00143	0.00764	0	438	0	0	0	0	1	-360	360;
	3961	5793	0.000229	0.00126	0	0	0	0	0	0	1	-360	360;
	3961	5793	0.000211	0.00143	0	0	0	0	0	0	1	-360	360;
	976	1549	0.0008	0.01072	0	1547	0	0	0	0	1	-360	360;
	976	4600	0.00067	0.00661	0	1119	0	0	0	0	1	-360	360;
	976	2800	0.0003	0.00427	0	1382	0	0	0	0	1	-360	360;
	976	2800	0.00068	0.00648	0	1349	0	0	0	0	1	-360	360;
	1600	8186	0.0002	0.001289	0	0	0	0	0	0	1	-360	360;
	1600	4011	0.00013	0.000961	0	533	0	0	0	0	1	-360	360;
	5771	2235	0.00118	0.01285	0	1613	0	0	0	0	1	-360	360;
	5771	5239	0.00062	0.00577	0	1086	0	0	0	0	1	-360	360;
	9216	1409	9.1e-05	0.000289	0	0	0	0	0	0	1	-360	360;
	9216	1409	9.1e-05	0.000271	0	0	0	0	0	0	1	-360	360;
	9216	6597	0.0003	0.002719	0	476	0	0	0	0	1	-360	360;
	9216	6836	0.00049	0.002091	0	0	0	0	0	0	1	-360	360;
	9216	15	0.000159	0.001461	0	572	0	0	0	0	1	-360	360;
	429	6132	0.00301	0.02281	0	0	0	0	0	0	1	-360	360;
	6660	6132	0.0027	0.019979	0	0	0	0	0	0	1	-360	360;
	1439	649	0.00019	0.000539	0	0	0	0	0	0	1	-360	360;
	649	6836	0.000169	0.00061	0	514	0	0	0	0	1	-360	360;
	8417	5435	0.00163	0.00939	0	533	0	0	0	0	1	-360	360;
	8417	5435	0.001159	0.009031	0	514	0	0	0	0	1	-360	360;
	8417	4058	9.9e-05	0.00037	0	457	0	0	0	0	1	-360	360;
	5617	9055	0.003789	0.01805	0	0	0	0	0	0	1	-360	360;
	5617	1471	0.00069	0.00389	0	0	0	0	0	0	1	-360	360;
	8396	7925	0.00201	0.01845	0	1283	0	0	0	0	1	-360	360;
	8806	7925	0.00121	0.00902	0	1283	0	0	0	0	1	-360	360;
	4542	1947	0.00155	0.01573	0	1283	0	0	0	0	1	-360	360;
	4542	8337	0.00031	0.00274	0	1316	0	0	0	0	1	-360	360;
	4542	3966	0.00047	0.00563	0	1185	0	0	0	0	1	-360	360;
	4542	3544	0.00066	0.00721	0	1152	0	0	0	0	1	-360	360;
	7510	4806	0.00326	0.016711	0	0	0	0	0	0	1	-360	360;
	7510	2907	0.002479	0.01262	0	629	0	0	0	0	1	-360	360;
	7510	2907	0.00243	0.01286	0	610	0	0	0	0	1	-360	360;
	7510	4338	0.004521	0.02306	0	476	0	0	0	0	1	-360	360;
	7510	1282	0.00211	0.00899	0	0	0	0	0	0	1	-360	360;
	7510	7795	0.001289	0.005651	0	553	0	0	0	0	1	-360	360;
	7510	3382	0.002479	0.009021	0	0	0	0	0	0	1	-360	360;
	7510	6456	0.00151	0.010521	0	419	0	0	0	0	1	-360	360;
	5409	8352	0.00101	0.004151	0	438	0	0	0	0	1	-360	360;
	5409	797	0.015599	0.079	0	381	0	0	0	0	1	-360	360;
	5409	8207	0.013961	0.067591	0	457	0	0	0	0	1	-360	360;
	5409	1241	0.007099	0.043169	0	514	0	0	0	0	1	-360	360;
	2416	5042	0.00317	0.02987	0	987	0	0	0	0	1	-360	360;
	2416	2764	0.00081	0.0088	0	1119	0	0	0	0	1	-360	360;
	2416	2764	0.00098	0.00961	0	987	0	0	0	0	1	-360	360;
	2416	179	0.00194	0.02225	0	921	0	0	0	0	1	-360	360;
	9001	5463	0.001031	0.006349	0	0	0	0	0	0	1	-360	360;
	9001	445	0.00668	0.0297	0	553	0	0	0	0	1	-360	360;
	4809	2538	0.00439	0.02236	0	533	0	0	0	0	1	-360	360;
	2538	6397	0.00249	0.01349	0	591	0	0	0	0	1	-360	360;
	2081	4556	0.001969	0.01205	0	495	0	0	0	0	1	-360	360;
	2081	4592	0.000461	0.003031	0	572	0	0	0	0	1	-360	360;
	3592	267	0.00444	0.023479	0	438	0	0	0	0	1	-360	360;
	3592	4632	0.00175	0.01288	0	0	0	0	0	0	1	-360	360;
	3592	4632	0.001711	0.01312	0	0	0	0	0	0	1	-360	360;
	1549	579	0.0008	0.00784	0	1481	0	0	0	0	1	-360	360;
	1549	9117	0.00051	0.00612	0	1382	0	0	0	0	1	-360	360;
	3282	7145	0.00137	0.00905	0	0	0	0	0	0	1	-360	360;
	3282	4855	0.002771	0.021409	0	381	0	0	0	0	1	-360	360;
	3282	4855	0.00314	0.020979	0	0	0	0	0	0	1	-360	360;
	3282	4095	0.002	0.01118	0	0	0	0	0	0	1	-360	360;
	8285	699	0.00119	0.00875	0	0	0	0	0	0	1	-360	360;
	8285	6882	0.00031	0.00236	0	0	0	0	0	0	1	-360	360;
	8285	7820	0.000831	0.00539	0	0	0	0	0	0	1	-360	360;
	267	7349	0.004349	0.02394	0	419	0	0	0	0	1	-360	360;
	1200	8731	0.0007	0.00782	0	1678	0	0	0	0	1	-360	360;
	1200	1716	0.00046	0.00432	0	1843	0	0	0	0	1	-360	360;
	1200	3067	0.00049	0.00617	0	1678	0	0	0	0	1	-360	360;
	1200	1074	0.00127	0.01514	0	1777	0	0	0	0	1	-360	360;
	3549	6544	0.00232	0.01843	0	0	0	0	0	0	1	-360	360;
	3549	1503	0.00488	0.032151	0	553	0	0	0	0	1	-360	360;
	1409	3715	0.000229	0.00136	0	533	0	0	0	0	1	-360	360;
	1409	6836	0.0003	0.001531	0	476	0	0	0	0	1	-360	360;
	1409	5892	0.00026	0.00126	0	400	0	0	0	0	1	-360	360;
	1409	5948	0.00056	0.002091	0	553	0	0	0	0	1	-360	360;
	4407	2962	0.00103	0.0093	0	1711	0	0	0	0	1	-360	360;
	4407	2437	0.00183	0.01689	0	1316	0	0	0	0	1	-360	360;
	7058	5163	0.00324	0.01095	0	857	0	0	0	0	1	-360	360;
	236	8236	0.001419	0.008539	0	0	0	0	0	0	1	-360	360;
	236	3427	0.00068	0.00438	0	495	0	0	0	0	1	-360	360;
	2916	5884	0.00178	0.01651	0	1283	0	0	0	0	1	-360	360;
	2916	4315	0.00151	0.01827	0	1251	0	0	0	0	1	-360	360;
	2916	4315	0.00148	0.01862	0	1316	0	0	0	0	1	-360	360;
	2962	2437	0.00201	0.02546	0	1580	0	0	0	0	1	-360	360;
	2962	2437	0.00197	0.02357	0	1547	0	0	0	0	1	-360	360;
	3747	3427	0.001091	0.00775	0	0	0	0	0	0	1	-360	360;
	3747	988	0.000841	0.006669	0	0	0	0	0	0	1	-360	360;
	932	3728	0.00151	0.012081	0	514	0	0	0	0	1	-360	360;
	932	7463	0.00032	0.002521	0	438	0	0	0	0	1	-360	360;
	8087	3342	0.006091	0.03019	0	514	0	0	0	0	1	-360	360;
	8087	7659	0.003651	0.016461	0	553	0	0	0	0	1	-360	360;
	8087	3334	0.00076	0.00425	0	381	0	0	0	0	1	-360	360;
	8087	3334	0.00087	0.004031	0	0	0	0	0	0	1	-360	360;
	1639	3728	0.000651	0.004729	0	591	0	0	0	0	1	-360	360;
	1639	8936	0.00045	0.00337	0	514	0	0	0	0	1	-360	360;
	2907	481	0.001539	0.00839	0	438	0	0	0	0	1	-360	360;
	2907	481	0.001919	0.010479	0	514	0	0	0	0	1	-360	360;
	481	9057	0.001169	0.007031	0	438	0	0	0	0	1	-360	360;
	1894	481	0.00739	0.034081	0	0	0	0	0	0	1	-360	360;
	4890	481	0.00164	0.00964	0	419	0	0	0	0	1	-360	360;
	5890	481	0.00206	0.012	0	457	0	0	0	0	1	-360	360;
	5391	481	0.000961	0.00531	0	457	0	0	0	0	1	-360	360;
	3544	6771	6e-05	0.00029	0	987	0	0	0	0	1	-360	360;
	1344	4806	0.001919	0.013151	0	381	0	0	0	0	1	-360	360;
	869	4806	0.00019	0.00107	0	400	0	0	0	0	1	-360	360;
	869	4806	0.00018	0.001091	0	438	0	0	0	0	1	-360	360;
	5685	2459	0.00147	0.01714	0	1514	0	0	0	0	1	-360	360;
	2730	2459	0.00143	0.01632	0	1283	0	0	0	0	1	-360	360;
	2730	2459	0.00142	0.01416	0	1580	0	0	0	0	1	-360	360;
	2730	2459	0.00145	0.01386	0	1744	0	0	0	0	1	-360	360;
	2459	8987	0.00073	0.00732	0	1613	0	0	0	0	1	-360	360;
	1185	1240	0.009039	0.05989	0	476	0	0	0	0	1	-360	360;
	1185	4016	0.007919	0.04975	0	457	0	0	0	0	1	-360	360;
	2668	981	0.000521	0.005841	0	457	0	0	0	0	1	-360	360;
	2668	6344	0.000521	0.0028	0	0	0	0	0	0	1	-360	360;
	6564	1074	0.0007	0.00744	0	889	0	0	0	0	1	-360	360;
	6564	4026	0.00159	0.0148	0	921	0	0	0	0	1	-360	360;
	4336	4249	0.000659	0.005151	0	553	0	0	0	0	1	-360	360;
	7196	4249	0.000729	0.004901	0	419	0	0	0	0	1	-360	360;
	4600	642	0.00064	0.00772	0	1119	0	0	0	0	1	-360	360;
	7573	6186	0.00018	0.00093	0	591	0	0	0	0	1	-360	360;
	7573	8186	0.00019	0.00118	0	495	0	0	0	0	1	-360	360;
	7413	7573	0.000409	0.002341	0	514	0	0	0	0	1	-360	360;
	8719	1023	0.00087	0.006591	0	381	0	0	0	0	1	-360	360;
	8719	3063	0.00162	0.0098	0	0	0	0	0	0	1	-360	360;
	7508	7053	0.00096	0.00849	0	1547	0	0	0	0	1	-360	360;
	6183	6932	0.000341	0.001081	0	495	0	0	0	0	1	-360	360;
	6183	6932	0.0003	0.001219	0	514	0	0	0	0	1	-360	360;
	6183	8068	9.9e-05	0.00076	0	533	0	0	0	0	1	-360	360;
	6183	8068	0.00011	0.00074	0	553	0	0	0	0	1	-360	360;
	1602	4855	0.00151	0.009039	0	0	0	0	0	0	1	-360	360;
	1602	2551	0.00356	0.02331	0	0	0	0	0	0	1	-360	360;
	7999	1439	0.000281	0.00169	0	0	0	0	0	0	1	-360	360;
	7999	6597	0.000169	0.001531	0	495	0	0	0	0	1	-360	360;
	7999	8661	0.00038	0.00326	0	0	0	0	0	0	1	-360	360;
	1086	7257	0.00136	0.01738	0	1481	0	0	0	0	1	-360	360;
	6443	519	0.00186	0.0133	0	0	0	0	0	0	1	-360	360;
	6443	519	0.00182	0.01356	0	0	0	0	0	0	1	-360	360;
	2009	6443	0.002531	0.015919	0	533	0	0	0	0	1	-360	360;
	6443	6749	0.00289	0.015969	0	419	0	0	0	0	1	-360	360;
	1121	4947	0.001771	0.01107	0	400	0	0	0	0	1	-360	360;
	4204	4947	0.007969	0.04238	0	0	0	0	0	0	1	-360	360;
	2907	7191	0.00013	0.00062	0	495	0	0	0	0	1	-360	360;
	2907	284	0.0002	0.00201	0	572	0	0	0	0	1	-360	360;
	477	9159	0.00095	0.01024	0	1349	0	0	0	0	1	-360	360;
	1133	9159	0.00553	0.04909	0	1152	0	0	0	0	1	-360	360;
	3178	6499	0.009229	0.03693	0	438	0	0	0	0	1	-360	360;
	3792	6499	0.00451	0.028469	0	438	0	0	0	0	1	-360	360;
	6499	6038	0.003591	0.01906	0	400	0	0	0	0	1	-360	360;
	6499	7398	0.010159	0.05169	0	495	0	0	0	0	1	-360	360;
	75	2126	0.00039	0.003099	0	0	0	0	0	0	1	-360	360;
	75	6869	0.000729	0.0048	0	0	0	0	0	0	1	-360	360;
	7137	6659	0.00141	0.01296	0	921	0	0	0	0	1	-360	360;
	3673	6745	0.001969	0.019229	0	0	0	0	0	0	1	-360	360;
	3673	8953	0.001081	0.0105	0	0	0	0	0	0	1	-360	360;
	3673	8953	0.00105	0.010711	0	0	0	0	0	0	1	-360	360;
	5685	2849	0.00052	0.00449	0	1415	0	0	0	0	1	-360	360;
	3378	6397	0.01024	0.056349	0	476	0	0	0	0	1	-360	360;
	3378	1520	0.004599	0.023659	0	419	0	0	0	0	1	-360	360;
	3378	3233	0.010021	0.057409	0	438	0	0	0	0	1	-360	360;
	3378	3291	0.00275	0.01525	0	457	0	0	0	0	1	-360	360;
	693	8240	0.00107	0.01122	0	1547	0	0	0	0	1	-360	360;
	693	5884	0.00242	0.02336	0	1218	0	0	0	0	1	-360	360;
	693	3587	0.00165	0.01303	0	823	0	0	0	0	1	-360	360;
	693	3587	0.00155	0.01332	0	856	0	0	0	0	1	-360	360;
	6994	5970	0.00087	0.005469	0	0	0	0	0	0	1	-360	360;
	6994	5970	0.002281	0.011909	0	0	0	0	0	0	1	-360	360;
	6994	970	0.00106	0.01076	0	0	0	0	0	0	1	-360	360;
	6994	5348	0.001409	0.01618	0	381	0	0	0	0	1	-360	360;
	6994	1909	0.00306	0.015521	0	457	0	0	0	0	1	-360	360;
	3843	8534	0.000969	0.006831	0	438	0	0	0	0	1	-360	360;
	3843	9039	0.002531	0.016961	0	533	0	0	0	0	1	-360	360;
	1021	257	0.01207	0.077901	0	0	0	0	0	0	1	-360	360;
	1021	6936	0.014091	0.080581	0	495	0	0	0	0	1	-360	360;
	1021	2946	0.003289	0.01481	0	0	0	0	0	0	1	-360	360;
	1021	172	0.003349	0.0145	0	0	0	0	0	0	1	-360	360;
	8138	3792	0.005651	0.026479	0	419	0	0	0	0	1	-360	360;
	8138	7659	0.01193	0.064669	0	457	0	0	0	0	1	-360	360;
	6669	5111	0.004151	0.017281	0	419	0	0	0	0	1	-360	360;
	5511	7145	0.00213	0.014091	0	0	0	0	0	0	1	-360	360;
	4338	1894	0.00113	0.00651	0	419	0	0	0	0	1	-360	360;
	4338	1894	0.001159	0.00637	0	419	0	0	0	0	1	-360	360;
	4338	6456	0.00611	0.03288	0	457	0	0	0	0	1	-360	360;
	8946	849	0.00012	0.00156	0	1744	0	0	0	0	1	-360	360;
	8946	849	0.00013	0.00153	0	1645	0	0	0	0	1	-360	360;
	8946	6768	0.00285	0.03071	0	1645	0	0	0	0	1	-360	360;
	5970	2896	0.00676	0.03787	0	495	0	0	0	0	1	-360	360;
	5970	9049	0.004461	0.023521	0	438	0	0	0	0	1	-360	360;
	5332	618	0.00064	0.0078	0	1514	0	0	0	0	1	-360	360;
	5332	618	0.00065	0.00764	0	1514	0	0	0	0	1	-360	360;
	5332	579	0.00066	0.00602	0	1481	0	0	0	0	1	-360	360;
	5332	4442	0.00149	0.01224	0	889	0	0	0	0	1	-360	360;
	5332	814	0.00061	0.00582	0	1020	0	0	0	0	1	-360	360;
	3633	6904	0.001461	0.00744	0	0	0	0	0	0	1	-360	360;
	3633	7820	0.00386	0.02406	0	0	0	0	0	0	1	-360	360;
	3633	7632	0.00306	0.01555	0	0	0	0	0	0	1	-360	360;
	3633	7632	0.003	0.01586	0	0	0	0	0	0	1	-360	360;
	3629	4112	0.003271	0.020651	0	0	0	0	0	0	1	-360	360;
	4112	7630	0.000849	0.00657	0	0	0	0	0	0	1	-360	360;
	1645	3066	0.000211	0.001169	0	476	0	0	0	0	1	-360	360;
	1645	7973	0.000469	0.00345	0	495	0	0	0	0	1	-360	360;
	4945	7508	0.00126	0.0145	0	1382	0	0	0	0	1	-360	360;
	1957	5933	0.008219	0.04925	0	0	0	0	0	0	1	-360	360;
	1957	1769	0.006771	0.04132	0	438	0	0	0	0	1	-360	360;
	7417	7600	0.000841	0.004539	0	591	0	0	0	0	1	-360	360;
	7417	3792	0.00701	0.03161	0	476	0	0	0	0	1	-360	360;
	7417	6879	0.000781	0.004271	0	572	0	0	0	0	1	-360	360;
	7417	962	0.00049	0.0032	0	533	0	0	0	0	1	-360	360;
	3489	3813	0.000331	0.00199	0	629	0	0	0	0	1	-360	360;
	1347	3813	0.00032	0.002031	0	629	0	0	0	0	1	-360	360;
	3813	5500	9.1e-05	0.00057	0	476	0	0	0	0	1	-360	360;
	5366	869	0.001651	0.006789	0	533	0	0	0	0	1	-360	360;
	5366	3382	0.001461	0.007229	0	0	0	0	0	0	1	-360	360;
	8352	9049	0.00437	0.023979	0	438	0	0	0	0	1	-360	360;
	1869	473	0.00781	0.04595	0	419	0	0	0	0	1	-360	360;
	1869	6243	0.0015	0.006531	0	457	0	0	0	0	1	-360	360;
	5767	9057	0.001031	0.006469	0	0	0	0	0	0	1	-360	360;
	2800	3061	0.00183	0.01686	0	1119	0	0	0	0	1	-360	360;
	2800	618	0.0002	0.00294	0	1514	0	0	0	0	1	-360	360;
	2800	618	0.00021	0.00288	0	1316	0	0	0	0	1	-360	360;
	2800	5042	0.00171	0.01889	0	1283	0	0	0	0	1	-360	360;
	8870	2764	0.00069	0.00565	0	1185	0	0	0	0	1	-360	360;
	8870	2764	0.0007	0.00553	0	1382	0	0	0	0	1	-360	360;
	7096	4290	0.00011	0.00115	0	1119	0	0	0	0	1	-360	360;
	7096	6681	0.00012	0.00126	0	1316	0	0	0	0	1	-360	360;
	7096	8973	0.00326	0.02919	0	1382	0	0	0	0	1	-360	360;
	7096	5060	0.00092	0.00876	0	823	0	0	0	0	1	-360	360;
	5780	6932	0.00087	0.005409	0	457	0	0	0	0	1	-360	360;
	5780	6869	0.00131	0.00701	0	438	0	0	0	0	1	-360	360;
	4291	6656	0.002531	0.018781	0	0	0	0	0	0	1	-360	360;
	4291	8729	0.00894	0.056909	0	0	0	0	0	0	1	-360	360;
	5512	5463	0.00187	0.010659	0	553	0	0	0	0	1	-360	360;
	7685	5463	0.00286	0.01538	0	610	0	0	0	0	1	-360	360;
	5463	4967	0.0008	0.00444	0	419	0	0	0	0	1	-360	360;
	1013	7875	0.00545	0.03936	0	0	0	0	0	0	1	-360	360;
	1741	7875	0.00119	0.009031	0	0	0	0	0	0	1	-360	360;
	51	5670	0.00295	0.0162	0	419	0	0	0	0	1	-360	360;
	51	6174	0.001581	0.0087	0	514	0	0	0	0	1	-360	360;
	51	2517	0.007729	0.025401	0	591	0	0	0	0	1	-360	360;
	7805	3425	0.00013	0.001461	0	457	0	0	0	0	1	-360	360;
	7805	8658	0.000341	0.02219	0	0	0	0	0	0	1	-360	360;
	7805	4011	0.00018	0.001211	0	591	0	0	0	0	1	-360	360;
	7805	9028	0.003099	0.0147	0	591	0	0	0	0	1	-360	360;
	7805	9028	0.003159	0.014401	0	533	0	0	0	0	1	-360	360;
	6932	4274	0.00281	0.01305	0	0	0	0	0	0	1	-360	360;
	6932	8871	0.002281	0.01075	0	495	0	0	0	0	1	-360	360;
	6932	8871	0.00232	0.010521	0	457	0	0	0	0	1	-360	360;
	6932	3442	0.000151	0.000841	0	0	0	0	0	0	1	-360	360;
	6953	2658	0.002771	0.012091	0	572	0	0	0	0	1	-360	360;
	6953	5592	0.00613	0.03005	0	457	0	0	0	0	1	-360	360;
	3342	7659	0.003219	0.02486	0	0	0	0	0	0	1	-360	360;
	6186	8068	0.000211	0.001331	0	572	0	0	0	0	1	-360	360;
	8504	1898	0.00185	0.01917	0	1613	0	0	0	0	1	-360	360;
	1282	869	0.00082	0.00381	0	572	0	0	0	0	1	-360	360;
	9043	7819	0.00075	0.004531	0	381	0	0	0	0	1	-360	360;
	9043	7819	0.000729	0.00419	0	419	0	0	0	0	1	-360	360;
	9043	32	0.02626	0.15576	0	438	0	0	0	0	1	-360	360;
	9043	7423	0.011469	0.03937	0	0	0	0	0	0	1	-360	360;
	7139	8186	0.00011	0.00061	0	514	0	0	0	0	1	-360	360;
	7139	1065	0.000531	0.00313	0	381	0	0	0	0	1	-360	360;
	7139	6134	0.001211	0.00864	0	0	0	0	0	0	1	-360	360;
	4336	4651	0.00094	0.006289	0	0	0	0	0	0	1	-360	360;
	731	4020	0.001	0.00886	0	823	0	0	0	0	1	-360	360;
	6769	3178	0.001979	0.010469	0	514	0	0	0	0	1	-360	360;
	6769	3178	0.0017	0.00855	0	553	0	0	0	0	1	-360	360;
	3178	2116	0.00757	0.04624	0	419	0	0	0	0	1	-360	360;
	3178	7283	0.001719	0.00794	0	495	0	0	0	0	1	-360	360;
	3178	7283	0.00176	0.007781	0	0	0	0	0	0	1	-360	360;
	3157	916	0.00126	0.0053	0	400	0	0	0	0	1	-360	360;
	3157	1241	0.00444	0.022789	0	419	0	0	0	0	1	-360	360;
	477	2437	0.00163	0.0145	0	1086	0	0	0	0	1	-360	360;
	6769	257	0.003281	0.02124	0	495	0	0	0	0	1	-360	360;
	6769	7283	0.004979	0.025271	0	476	0	0	0	0	1	-360	360;
	8858	2403	0.0072	0.055289	0	495	0	0	0	0	1	-360	360;
	8858	4772	0.00787	0.04582	0	0	0	0	0	0	1	-360	360;
	8858	4016	0.003961	0.017711	0	0	0	0	0	0	1	-360	360;
	2403	6797	0.00756	0.04826	0	457	0	0	0	0	1	-360	360;
	6192	8207	0.003401	0.02086	0	457	0	0	0	0	1	-360	360;
	6192	3937	0.001341	0.00669	0	419	0	0	0	0	1	-360	360;
	7669	8867	6e-05	0.000581	0	438	0	0	0	0	1	-360	360;
	7669	5948	9.1e-05	0.00074	0	514	0	0	0	0	1	-360	360;
	7392	8939	0.000531	0.00175	0	0	0	0	0	0	1	-360	360;
	1114	5096	0.00116	0.01159	0	1382	0	0	0	0	1	-360	360;
	1114	4315	0.00038	0.00457	0	1613	0	0	0	0	1	-360	360;
	3066	7973	0.000729	0.00405	0	553	0	0	0	0	1	-360	360;
	264	7064	0.000331	0.001919	0	0	0	0	0	0	1	-360	360;
	2188	4706	0.00086	0.01029	0	1909	0	0	0	0	1	-360	360;
	7428	2188	0.00027	0.00332	0	1711	0	0	0	0	1	-360	360;
	5170	5681	0.00012	0.00156	0	0	0	0	0	0	1	-360	360;
	5170	5681	0.00012	0.001531	0	0	0	0	0	0	1	-360	360;
	8756	2060	0.00256	0.014211	0	0	0	0	0	0	1	-360	360;
	8756	8419	0.00049	0.00338	0	0	0	0	0	0	1	-360	360;
	6009	1430	0.007531	0.04444	0	400	0	0	0	0	1	-360	360;
	6009	5592	0.007409	0.03226	0	0	0	0	0	0	1	-360	360;
	3150	8147	0.001789	0.012841	0	0	0	0	0	0	1	-360	360;
	3150	4096	0.001831	0.01257	0	0	0	0	0	0	1	-360	360;
	1632	3621	0.000169	0.00143	0	0	0	0	0	0	1	-360	360;
	1632	6869	0.00088	0.005099	0	610	0	0	0	0	1	-360	360;
	3061	7460	0.0004	0.00479	0	1678	0	0	0	0	1	-360	360;
	3061	7460	0.00052	0.00481	0	1744	0	0	0	0	1	-360	360;
	3061	7929	0.00135	0.01431	0	1744	0	0	0	0	1	-360	360;
	3061	814	0.00051	0.00549	0	1481	0	0	0	0	1	-360	360;
	699	6882	0.00105	0.007969	0	476	0	0	0	0	1	-360	360;
	699	947	0.00339	0.020531	0	0	0	0	0	0	1	-360	360;
	699	2951	0.00169	0.009331	0	0	0	0	0	0	1	-360	360;
	3584	4177	0.00374	0.01905	0	438	0	0	0	0	1	-360	360;
	3584	5276	0.00487	0.026341	0	533	0	0	0	0	1	-360	360;
	5034	8301	0.00057	0.004581	0	0	0	0	0	0	1	-360	360;
	5034	8582	0.000409	0.00301	0	0	0	0	0	0	1	-360	360;
	1461	3754	9.1e-05	0.00076	0	419	0	0	0	0	1	-360	360;
	1461	6179	0.003919	0.01682	0	476	0	0	0	0	1	-360	360;
	5823	1060	0.00117	0.0131	0	1514	0	0	0	0	1	-360	360;
	5280	4954	0.008271	0.02787	0	495	0	0	0	0	1	-360	360;
	5280	1121	0.016599	0.0572	0	419	0	0	0	0	1	-360	360;
	5280	7861	0.00999	0.057	0	0	0	0	0	0	1	-360	360;
	5280	7861	0.01132	0.05586	0	0	0	0	0	0	1	-360	360;
	7303	4690	4e-05	0.00045	0	921	0	0	0	0	1	-360	360;
	4954	8534	0.003281	0.01038	0	495	0	0	0	0	1	-360	360;
	4317	72	0.003711	0.01725	0	457	0	0	0	0	1	-360	360;
	4317	6333	0.00069	0.00232	0	457	0	0	0	0	1	-360	360;
	8702	4948	0.004151	0.02436	0	514	0	0	0	0	1	-360	360;
	5163	4948	0.006781	0.028729	0	476	0	0	0	0	1	-360	360;
	3489	6869	0.000331	0.00151	0	438	0	0	0	0	1	-360	360;
	1347	6869	0.000331	0.001479	0	629	0	0	0	0	1	-360	360;
	3687	5877	0.0001	0.00144	0	1514	0	0	0	0	1	-360	360;
	3687	5919	0.00027	0.00221	0	1053	0	0	0	0	1	-360	360;
	1232	4588	0.0001	0.00141	0	1711	0	0	0	0	1	-360	360;
	1232	2764	0.00054	0.00546	0	1448	0	0	0	0	1	-360	360;
	7428	5877	7e-05	0.00053	0	1382	0	0	0	0	1	-360	360;
	4588	7428	6e-05	0.0006	0	1481	0	0	0	0	1	-360	360;
	8939	4371	0.000341	0.00245	0	0	0	0	0	0	1	-360	360;
	8939	7820	0.000841	0.006719	0	0	0	0	0	0	1	-360	360;
	6209	7191	0.000401	0.00164	0	476	0	0	0	0	1	-360	360;
	519	7349	0.004409	0.032531	0	0	0	0	0	0	1	-360	360;
	429	7877	0.001961	0.013581	0	381	0	0	0	0	1	-360	360;
	1894	4890	0.006659	0.03437	0	0	0	0	0	0	1	-360	360;
	4062	1599	9.1e-05	0.0007	0	0	0	0	0	0	1	-360	360;
	4062	8079	0.00051	0.004211	0	0	0	0	0	0	1	-360	360;
	6841	8936	0.00043	0.0027	0	0	0	0	0	0	1	-360	360;
	6841	2564	0.00038	0.00306	0	0	0	0	0	0	1	-360	360;
	8710	1491	0.001461	0.009479	0	0	0	0	0	0	1	-360	360;
	8710	7406	0.001841	0.011081	0	514	0	0	0	0	1	-360	360;
	970	7504	0.00094	0.007169	0	553	0	0	0	0	1	-360	360;
	6039	8034	0.00082	0.00846	0	1415	0	0	0	0	1	-360	360;
	6039	7428	0.00029	0.00325	0	1711	0	0	0	0	1	-360	360;
	7973	7475	0.000781	0.0062	0	0	0	0	0	0	1	-360	360;
	7973	2740	8.1e-05	0.00187	0	495	0	0	0	0	1	-360	360;
	7973	2297	8.1e-05	0.001909	0	419	0	0	0	0	1	-360	360;
	6877	1060	0.00113	0.00956	0	1185	0	0	0	0	1	-360	360;
	6877	1060	0.00115	0.00935	0	1218	0	0	0	0	1	-360	360;
	8981	4428	0.00161	0.02036	0	0	0	0	0	0	1	-360	360;
	8981	4428	0.00157	0.02075	0	0	0	0	0	0	1	-360	360;
	1841	3728	0.00101	0.007401	0	533	0	0	0	0	1	-360	360;
	1841	7463	0.00087	0.00711	0	476	0	0	0	0	1	-360	360;
	3582	5919	0.00135	0.01442	0	1349	0	0	0	0	1	-360	360;
	3582	2096	0.00176	0.01441	0	1777	0	0	0	0	1	-360	360;
	2009	7349	0.00682	0.03187	0	419	0	0	0	0	1	-360	360;
	8034	1399	0.00109	0.00987	0	1185	0	0	0	0	1	-360	360;
	4652	8034	0.00273	0.02829	0	1283	0	0	0	0	1	-360	360;
	1752	7845	0.001581	0.011289	0	400	0	0	0	0	1	-360	360;
	1752	3427	0.001669	0.013099	0	0	0	0	0	0	1	-360	360;
	7347	1599	9.9e-05	0.00069	0	495	0	0	0	0	1	-360	360;
	7347	8079	9.9e-05	0.000669	0	514	0	0	0	0	1	-360	360;
	4896	3736	0.000591	0.00382	0	0	0	0	0	0	1	-360	360;
	4896	7632	0.00524	0.02276	0	0	0	0	0	0	1	-360	360;
	3736	6625	0.002719	0.01143	0	0	0	0	0	0	1	-360	360;
	739	8561	0.00294	0.02744	0	954	0	0	0	0	1	-360	360;
	739	7395	0.0011	0.01339	0	1020	0	0	0	0	1	-360	360;
	739	5873	0.00193	0.01695	0	1349	0	0	0	0	1	-360	360;
	739	4843	0.00134	0.01054	0	1020	0	0	0	0	1	-360	360;
	6092	739	0.00245	0.02064	0	1382	0	0	0	0	1	-360	360;
	1393	6717	0.001841	0.00901	0	457	0	0	0	0	1	-360	360;
	1393	6717	0.00187	0.00882	0	495	0	0	0	0	1	-360	360;
	8337	807	0.00197	0.02312	0	1283	0	0	0	0	1	-360	360;
	2235	7395	0.00078	0.00887	0	1382	0	0	0	0	1	-360	360;
	8512	5435	0.003771	0.0288	0	438	0	0	0	0	1	-360	360;
	8512	4658	0.000909	0.00737	0	457	0	0	0	0	1	-360	360;
	7929	7460	0.00086	0.00783	0	1744	0	0	0	0	1	-360	360;
	7929	7460	0.00095	0.00855	0	1547	0	0	0	0	1	-360	360;
	299	7460	0.00012	0.00157	0	1843	0	0	0	0	1	-360	360;
	299	7460	0.00013	0.00154	0	757	0	0	0	0	1	-360	360;
	5117	6456	0.00089	0.004771	0	533	0	0	0	0	1	-360	360;
	5117	6394	0.002	0.00881	0	553	0	0	0	0	1	-360	360;
	6397	7075	0.007711	0.035331	0	0	0	0	0	0	1	-360	360;
	7075	6936	0.005919	0.030419	0	495	0	0	0	0	1	-360	360;
	7075	2946	0.0053	0.031719	0	591	0	0	0	0	1	-360	360;
	4274	1969	0.00025	0.001289	0	0	0	0	0	0	1	-360	360;
	2946	2116	0.0062	0.040039	0	0	0	0	0	0	1	-360	360;
	5096	5464	0.0008	0.00707	0	1514	0	0	0	0	1	-360	360;
	5096	4957	0.0002	0.00207	0	954	0	0	0	0	1	-360	360;
	5096	4446	0.00036	0.00322	0	1053	0	0	0	0	1	-360	360;
	1115	6090	0.0023	0.0157	0	381	0	0	0	0	1	-360	360;
	1115	4204	0.003479	0.021159	0	419	0	0	0	0	1	-360	360;
	3966	3428	0.00107	0.01013	0	1382	0	0	0	0	1	-360	360;
	1439	6836	0.00025	0.000909	0	514	0	0	0	0	1	-360	360;
	1439	5892	0.00037	0.00188	0	438	0	0	0	0	1	-360	360;
	9201	3937	0.0003	0.00145	0	419	0	0	0	0	1	-360	360;
	9201	3692	0.001419	0.007539	0	476	0	0	0	0	1	-360	360;
	8561	266	0.0012	0.0122	0	1251	0	0	0	0	1	-360	360;
	8561	2566	0.00397	0.02936	0	1251	0	0	0	0	1	-360	360;
	8561	8806	0.00092	0.01112	0	1185	0	0	0	0	1	-360	360;
	8561	7428	0.00387	0.0452	0	1448	0	0	0	0	1	-360	360;
	8561	1280	0.00432	0.03156	0	1086	0	0	0	0	1	-360	360;
	2730	1133	0.00407	0.03937	0	1251	0	0	0	0	1	-360	360;
	2730	1133	0.00415	0.03856	0	1152	0	0	0	0	1	-360	360;
	6127	5111	0.00869	0.05343	0	572	0	0	0	0	1	-360	360;
	6127	2172	0.0068	0.036961	0	457	0	0	0	0	1	-360	360;
	6127	2658	0.01805	0.095469	0	400	0	0	0	0	1	-360	360;
	7600	7283	0.00045	0.00206	0	533	0	0	0	0	1	-360	360;
	473	7193	0.0095	0.037159	0	419	0	0	0	0	1	-360	360;
	473	5025	0.02106	0.0885	0	457	0	0	0	0	1	-360	360;
	7395	5873	0.00157	0.02318	0	1810	0	0	0	0	1	-360	360;
	7395	3662	0.00032	0.00282	0	1711	0	0	0	0	1	-360	360;
	7395	4931	0.00165	0.01605	0	889	0	0	0	0	1	-360	360;
	5435	489	0.00107	0.00714	0	400	0	0	0	0	1	-360	360;
	5435	489	0.000961	0.00555	0	400	0	0	0	0	1	-360	360;
	5435	8529	0.00074	0.00461	0	0	0	0	0	0	1	-360	360;
	5435	7064	0.00076	0.00451	0	0	0	0	0	0	1	-360	360;
	5435	4058	0.001419	0.01093	0	533	0	0	0	0	1	-360	360;
	5435	4658	0.003729	0.024599	0	476	0	0	0	0	1	-360	360;
	2437	7029	0.00233	0.01992	0	1678	0	0	0	0	1	-360	360;
	2437	7029	0.00238	0.01949	0	1645	0	0	0	0	1	-360	360;
	2437	4795	0.00217	0.0253	0	1744	0	0	0	0	1	-360	360;
	2437	4795	0.00212	0.02579	0	1843	0	0	0	0	1	-360	360;
	2437	1133	0.00458	0.04138	0	1086	0	0	0	0	1	-360	360;
	3792	550	0.01555	0.089841	0	381	0	0	0	0	1	-360	360;
	6299	3792	0.00274	0.0168	0	476	0	0	0	0	1	-360	360;
	3792	6879	0.005831	0.032211	0	438	0	0	0	0	1	-360	360;
	3792	4181	0.006669	0.031341	0	476	0	0	0	0	1	-360	360;
	8702	5163	0.01081	0.0538	0	476	0	0	0	0	1	-360	360;
	8702	7861	0.005581	0.02457	0	495	0	0	0	0	1	-360	360;
	8702	7861	0.00569	0.024039	0	533	0	0	0	0	1	-360	360;
	957	7446	0.002081	0.01294	0	457	0	0	0	0	1	-360	360;
	957	4897	0.001599	0.00919	0	0	0	0	0	0	1	-360	360;
	5239	5873	0.00083	0.00768	0	1744	0	0	0	0	1	-360	360;
	8438	8147	0.004979	0.024099	0	0	0	0	0	0	1	-360	360;
	8438	8147	0.005081	0.023591	0	476	0	0	0	0	1	-360	360;
	4290	266	0.00412	0.03917	0	1283	0	0	0	0	1	-360	360;
	266	4843	0.00086	0.00798	0	1119	0	0	0	0	1	-360	360;
	3275	266	0.00087	0.00961	0	1218	0	0	0	0	1	-360	360;
	8247	266	0.00095	0.01034	0	1251	0	0	0	0	1	-360	360;
	8374	266	0.00211	0.0196	0	1119	0	0	0	0	1	-360	360;
	7193	5025	0.00963	0.05281	0	0	0	0	0	0	1	-360	360;
	1064	3734	0.00018	0.001349	0	553	0	0	0	0	1	-360	360;
	1064	6456	0.000961	0.006021	0	381	0	0	0	0	1	-360	360;
	1064	6456	0.000979	0.00589	0	381	0	0	0	0	1	-360	360;
	7877	3063	0.000461	0.002729	0	610	0	0	0	0	1	-360	360;
	986	3367	0.005039	0.01901	0	381	0	0	0	0	1	-360	360;
	986	4603	0.00812	0.028919	0	419	0	0	0	0	1	-360	360;
	7257	6970	0.00106	0.01005	0	954	0	0	0	0	1	-360	360;
	5873	6768	0.00074	0.0081	0	1218	0	0	0	0	1	-360	360;
	8147	6174	0.012961	0.03605	0	0	0	0	0	0	1	-360	360;
	8147	6174	0.00844	0.0378	0	400	0	0	0	0	1	-360	360;
	8147	5162	0.000591	0.004169	0	553	0	0	0	0	1	-360	360;
	6879	6038	0.003091	0.012289	0	457	0	0	0	0	1	-360	360;
	4556	4592	0.000969	0.0053	0	553	0	0	0	0	1	-360	360;
	2412	2686	0.00161	0.012919	0	0	0	0	0	0	1	-360	360;
	2412	2686	0.003219	0.01795	0	0	0	0	0	0	1	-360	360;
	2412	3644	0.004331	0.025	0	0	0	0	0	0	1	-360	360;
	2412	3644	0.004409	0.02445	0	0	0	0	0	0	1	-360	360;
	4855	2551	0.00425	0.02014	0	0	0	0	0	0	1	-360	360;
	4855	6904	0.009531	0.060979	0	495	0	0	0	0	1	-360	360;
	4855	1897	0.002409	0.0135	0	0	0	0	0	0	1	-360	360;
	3268	8210	0.0006	0.00662	0	1678	0	0	0	0	1	-360	360;
	3268	2849	0.00173	0.01701	0	1678	0	0	0	0	1	-360	360;
	6397	1173	0.004289	0.018531	0	438	0	0	0	0	1	-360	360;
	6397	6936	0.006919	0.039789	0	419	0	0	0	0	1	-360	360;
	6397	6936	0.007591	0.035539	0	495	0	0	0	0	1	-360	360;
	6397	4750	0.00961	0.05225	0	419	0	0	0	0	1	-360	360;
	3665	6544	0.000581	0.004599	0	0	0	0	0	0	1	-360	360;
	3665	6544	0.00057	0.00469	0	0	0	0	0	0	1	-360	360;
	3665	6544	0.00087	0.003159	0	591	0	0	0	0	1	-360	360;
	6436	5217	0.00018	0.001419	0	0	0	0	0	0	1	-360	360;
	6436	6717	0.00195	0.018841	0	0	0	0	0	0	1	-360	360;
	72	6093	0.00293	0.01293	0	400	0	0	0	0	1	-360	360;
	1520	3233	0.001969	0.010599	0	438	0	0	0	0	1	-360	360;
	6985	854	3.9e-05	0.000591	0	0	0	0	0	0	1	-360	360;
	6985	415	0.00364	0.02431	0	0	0	0	0	0	1	-360	360;
	6985	1501	0.00114	0.007211	0	0	0	0	0	0	1	-360	360;
	489	4238	0.001271	0.008659	0	381	0	0	0	0	1	-360	360;
	489	1411	0.00105	0.00668	0	0	0	0	0	0	1	-360	360;
	4290	6681	0.00011	0.00122	0	987	0	0	0	0	1	-360	360;
	4290	3587	0.00117	0.0124	0	1349	0	0	0	0	1	-360	360;
	4290	8374	0.00168	0.01891	0	1185	0	0	0	0	1	-360	360;
	797	7819	0.00132	0.008281	0	572	0	0	0	0	1	-360	360;
	797	8378	0.004531	0.01924	0	0	0	0	0	0	1	-360	360;
	797	1241	0.007841	0.041909	0	419	0	0	0	0	1	-360	360;
	7929	7755	0.00113	0.01235	0	1777	0	0	0	0	1	-360	360;
	7929	7755	0.00128	0.0121	0	1744	0	0	0	0	1	-360	360;
	7929	8625	0.0015	0.01792	0	1415	0	0	0	0	1	-360	360;
	4592	1471	0.00111	0.005281	0	0	0	0	0	0	1	-360	360;
	4592	1471	0.00113	0.005169	0	0	0	0	0	0	1	-360	360;
	6040	3111	0.00563	0.03255	0	476	0	0	0	0	1	-360	360;
	1175	3111	0.005289	0.022539	0	0	0	0	0	0	1	-360	360;
	3111	6239	0.007159	0.05086	0	400	0	0	0	0	1	-360	360;
	3111	3689	0.009669	0.05649	0	0	0	0	0	0	1	-360	360;
	3111	6179	0.0092	0.052169	0	553	0	0	0	0	1	-360	360;
	8396	5091	0.00197	0.01594	0	1316	0	0	0	0	1	-360	360;
	8396	8806	0.00343	0.02573	0	1251	0	0	0	0	1	-360	360;
	8396	8247	0.00151	0.01485	0	1185	0	0	0	0	1	-360	360;
	8396	5060	0.00092	0.01042	0	1020	0	0	0	0	1	-360	360;
	8396	8012	0.00029	0.00336	0	1547	0	0	0	0	1	-360	360;
	1491	4489	0.001711	0.01255	0	0	0	0	0	0	1	-360	360;
	1491	1503	0.005	0.03962	0	553	0	0	0	0	1	-360	360;
	849	1168	0.0014	0.01687	0	1251	0	0	0	0	1	-360	360;
	849	1168	0.00067	0.00678	0	1942	0	0	0	0	1	-360	360;
	849	1168	0.00065	0.00692	0	1678	0	0	0	0	1	-360	360;
	6092	849	0.00137	0.01524	0	1218	0	0	0	0	1	-360	360;
	3840	686	0.00424	0.02376	0	0	0	0	0	0	1	-360	360;
	3840	5821	0.005591	0.031711	0	476	0	0	0	0	1	-360	360;
	3840	2225	0.00138	0.00649	0	419	0	0	0	0	1	-360	360;
	3840	2225	0.00107	0.00544	0	514	0	0	0	0	1	-360	360;
	5091	4390	3e-05	0.00041	0	1777	0	0	0	0	1	-360	360;
	5091	4390	3e-05	0.00041	0	1613	0	0	0	0	1	-360	360;
	5091	4957	0.00044	0.00342	0	1316	0	0	0	0	1	-360	360;
	7446	2504	0.003841	0.01737	0	438	0	0	0	0	1	-360	360;
	1741	1912	0.001771	0.01337	0	495	0	0	0	0	1	-360	360;
	483	2658	0.00139	0.007349	0	533	0	0	0	0	1	-360	360;
	483	5592	0.00625	0.03463	0	533	0	0	0	0	1	-360	360;
	5111	4360	0.00055	0.00255	0	0	0	0	0	0	1	-360	360;
	6596	3512	0.00489	0.043331	0	0	0	0	0	0	1	-360	360;
	8701	8207	0.00145	0.013099	0	381	0	0	0	0	1	-360	360;
	8701	1271	0.004039	0.02881	0	0	0	0	0	0	1	-360	360;
	2236	4967	0.002219	0.00988	0	438	0	0	0	0	1	-360	360;
	8575	7845	0.001919	0.0153	0	400	0	0	0	0	1	-360	360;
	8664	1399	0.00189	0.01319	0	1382	0	0	0	0	1	-360	360;
	5202	8936	0.00011	0.000659	0	495	0	0	0	0	1	-360	360;
	5202	2564	0.00011	0.00064	0	457	0	0	0	0	1	-360	360;
	5202	3027	0.000211	0.00111	0	0	0	0	0	0	1	-360	360;
	6836	6209	0.000469	0.001669	0	400	0	0	0	0	1	-360	360;
	5923	854	6e-05	0.000849	0	0	0	0	0	0	1	-360	360;
	7137	7310	0.00051	0.00597	0	1580	0	0	0	0	1	-360	360;
	7137	6970	0.00119	0.01087	0	1218	0	0	0	0	1	-360	360;
	5831	7137	0.00219	0.01922	0	1382	0	0	0	0	1	-360	360;
	5919	2764	0.00051	0.00465	0	1580	0	0	0	0	1	-360	360;
	8871	8819	0.00055	0.0067	0	0	0	0	0	0	1	-360	360;
	2504	5329	0.005099	0.027349	0	553	0	0	0	0	1	-360	360;
	3233	2844	0.000599	0.003961	0	476	0	0	0	0	1	-360	360;
	3233	3291	0.005461	0.032719	0	476	0	0	0	0	1	-360	360;
	8149	2896	0.00349	0.02256	0	476	0	0	0	0	1	-360	360;
	8149	1802	0.00132	0.00861	0	400	0	0	0	0	1	-360	360;
	5828	440	0.0045	0.01838	0	0	0	0	0	0	1	-360	360;
	5828	1391	0.004289	0.019719	0	400	0	0	0	0	1	-360	360;
	8236	3027	0.001409	0.00881	0	495	0	0	0	0	1	-360	360;
	1168	6607	0.00212	0.01857	0	1415	0	0	0	0	1	-360	360;
	1168	7673	0.00049	0.00574	0	1613	0	0	0	0	1	-360	360;
	1168	754	0.00184	0.021	0	1547	0	0	0	0	1	-360	360;
	2686	4773	0.00593	0.037409	0	419	0	0	0	0	1	-360	360;
	2225	4177	0.00514	0.024169	0	438	0	0	0	0	1	-360	360;
	3798	2913	0.00041	0.00492	0	1251	0	0	0	0	1	-360	360;
	3544	2913	0.00296	0.02677	0	1349	0	0	0	0	1	-360	360;
	4906	550	0.0072	0.03406	0	381	0	0	0	0	1	-360	360;
	3427	6949	0.00043	0.00276	0	419	0	0	0	0	1	-360	360;
	3427	9028	0.001789	0.013469	0	533	0	0	0	0	1	-360	360;
	4890	7406	0.00163	0.011091	0	457	0	0	0	0	1	-360	360;
	4890	6349	0.00036	0.002651	0	0	0	0	0	0	1	-360	360;
	8867	15	0.00013	0.00132	0	514	0	0	0	0	1	-360	360;
	8867	284	0.000169	0.001659	0	457	0	0	0	0	1	-360	360;
	4428	8534	0.00136	0.012831	0	0	0	0	0	0	1	-360	360;
	4428	8534	0.00139	0.01256	0	0	0	0	0	0	1	-360	360;
	4428	5276	0.00126	0.009521	0	0	0	0	0	0	1	-360	360;
	4428	5276	0.00143	0.009331	0	0	0	0	0	0	1	-360	360;
	3644	4428	0.00164	0.012961	0	0	0	0	0	0	1	-360	360;
	3644	4428	0.001599	0.012	0	0	0	0	0	0	1	-360	360;
	6936	2957	0.001271	0.00524	0	438	0	0	0	0	1	-360	360;
	8240	8973	0.00226	0.02465	0	1514	0	0	0	0	1	-360	360;
	6543	3715	0.00032	0.0015	0	514	0	0	0	0	1	-360	360;
	6543	5948	0.00032	0.00112	0	419	0	0	0	0	1	-360	360;
	3275	4931	0.00075	0.00817	0	1053	0	0	0	0	1	-360	360;
	8163	3425	0.00037	0.0027	0	553	0	0	0	0	1	-360	360;
	3425	6783	3.9e-05	0.000219	0	0	0	0	0	0	1	-360	360;
	2172	595	0.012781	0.05762	0	0	0	0	0	0	1	-360	360;
	8658	1065	0.00036	0.001289	0	0	0	0	0	0	1	-360	360;
	6090	1175	0.006771	0.039349	0	0	0	0	0	0	1	-360	360;
	3106	7673	0.00292	0.02831	0	1415	0	0	0	0	1	-360	360;
	3106	2096	0.00134	0.01229	0	1514	0	0	0	0	1	-360	360;
	8163	5500	0.00011	0.00063	0	572	0	0	0	0	1	-360	360;
	7249	8207	0.0003	0.003331	0	0	0	0	0	0	1	-360	360;
	7249	3692	0.001289	0.01555	0	0	0	0	0	0	1	-360	360;
	6783	5923	0.00011	0.0017	0	553	0	0	0	0	1	-360	360;
	6783	8575	0.00043	0.00249	0	533	0	0	0	0	1	-360	360;
	1121	4657	0.002651	0.016771	0	0	0	0	0	0	1	-360	360;
	3487	1121	0.0082	0.025081	0	400	0	0	0	0	1	-360	360;
	3442	8819	0.00037	0.002781	0	591	0	0	0	0	1	-360	360;
	4856	4906	0.010719	0.054719	0	400	0	0	0	0	1	-360	360;
	3076	4906	0.006469	0.04074	0	400	0	0	0	0	1	-360	360;
	5884	8973	0.00101	0.01274	0	1283	0	0	0	0	1	-360	360;
	8207	5961	9.1e-05	0.05076	0	419	0	0	0	0	1	-360	360;
	8207	3629	0.00588	0.03938	0	457	0	0	0	0	1	-360	360;
	8207	3692	0.003591	0.022099	0	438	0	0	0	0	1	-360	360;
	8207	7630	0.00555	0.032521	0	0	0	0	0	0	1	-360	360;
	6877	7755	0.00116	0.01347	0	1514	0	0	0	0	1	-360	360;
	6877	7755	0.00113	0.01247	0	1547	0	0	0	0	1	-360	360;
	5042	179	0.00107	0.00871	0	954	0	0	0	0	1	-360	360;
	5042	4442	0.00078	0.00609	0	1382	0	0	0	0	1	-360	360;
	2566	6768	0.00075	0.00604	0	1119	0	0	0	0	1	-360	360;
	4897	6717	0.00224	0.011091	0	0	0	0	0	0	1	-360	360;
	7475	6949	0.001961	0.013969	0	400	0	0	0	0	1	-360	360;
	1294	5162	0.001271	0.00761	0	476	0	0	0	0	1	-360	360;
	6681	8806	0.00374	0.03581	0	1382	0	0	0	0	1	-360	360;
	6681	8806	0.00385	0.03284	0	1316	0	0	0	0	1	-360	360;
	6681	8806	0.00392	0.03216	0	1185	0	0	0	0	1	-360	360;
	6681	8973	0.00307	0.03131	0	1283	0	0	0	0	1	-360	360;
	7819	32	0.027	0.131289	0	438	0	0	0	0	1	-360	360;
	6333	5512	0.005539	0.0195	0	0	0	0	0	0	1	-360	360;
	8534	9039	0.003099	0.017849	0	495	0	0	0	0	1	-360	360;
	8534	5276	0.002831	0.014099	0	533	0	0	0	0	1	-360	360;
	4026	2261	0	0.01368	0	1580	0	0	0	0	1	-360	360;
	4204	4657	0.01082	0.079719	0	0	0	0	0	0	1	-360	360;
	731	2764	0.00102	0.00941	0	1251	0	0	0	0	1	-360	360;
	731	597	0.00183	0.01819	0	1777	0	0	0	0	1	-360	360;
	731	385	0.00028	0.00346	0	1251	0	0	0	0	1	-360	360;
	3728	2564	0.0012	0.00982	0	0	0	0	0	0	1	-360	360;
	2764	2622	0.00027	0.0032	0	1810	0	0	0	0	1	-360	360;
	2764	7688	0.00064	0.00679	0	1152	0	0	0	0	1	-360	360;
	2764	8260	0.00047	0.00442	0	1711	0	0	0	0	1	-360	360;
	415	3367	0.00493	0.017599	0	495	0	0	0	0	1	-360	360;
	415	4081	0.00207	0.0153	0	0	0	0	0	0	1	-360	360;
	7287	7755	0.00094	0.0092	0	1909	0	0	0	0	1	-360	360;
	7287	7755	0.00095	0.00901	0	1711	0	0	0	0	1	-360	360;
	947	2785	0.00101	0.00719	0	0	0	0	0	0	1	-360	360;
	3734	6654	0.000419	0.00331	0	572	0	0	0	0	1	-360	360;
	2551	1897	0.002099	0.015599	0	0	0	0	0	0	1	-360	360;
	8806	8664	0.00163	0.01283	0	1283	0	0	0	0	1	-360	360;
	8806	7428	0.00445	0.03788	0	1514	0	0	0	0	1	-360	360;
	8806	4652	0.00092	0.00703	0	1218	0	0	0	0	1	-360	360;
	249	32	9.1e-05	0.000961	0	0	0	0	0	0	1	-360	360;
	1741	249	0.00387	0.028599	0	0	0	0	0	0	1	-360	360;
	249	5290	0.00949	0.06625	0	0	0	0	0	0	1	-360	360;
	249	4467	0.00414	0.035331	0	0	0	0	0	0	1	-360	360;
	249	5290	0.009781	0.065581	0	0	0	0	0	0	1	-360	360;
	1391	32	0.02818	0.116099	0	381	0	0	0	0	1	-360	360;
	32	2143	0.00324	0.01912	0	381	0	0	0	0	1	-360	360;
	3008	6768	0.00135	0.01609	0	1613	0	0	0	0	1	-360	360;
	6174	5670	0.00294	0.0182	0	438	0	0	0	0	1	-360	360;
	8066	5961	0.000711	0.00649	0	457	0	0	0	0	1	-360	360;
	442	8066	0.002651	0.01375	0	0	0	0	0	0	1	-360	360;
	1271	5961	0.00105	0.008469	0	419	0	0	0	0	1	-360	360;
	869	4489	0.002229	0.016461	0	495	0	0	0	0	1	-360	360;
	869	4489	0.002281	0.01613	0	495	0	0	0	0	1	-360	360;
	602	4489	0.003341	0.012651	0	0	0	0	0	0	1	-360	360;
	602	4489	0.003271	0.012901	0	0	0	0	0	0	1	-360	360;
	6484	7918	0.00101	0.01218	0	1843	0	0	0	0	1	-360	360;
	6484	7436	0.0009	0.0095	0	0	0	0	0	0	1	-360	360;
	7918	4690	0.00098	0.01241	0	0	0	0	0	0	1	-360	360;
	5511	1942	0.001219	0.00951	0	0	0	0	0	0	1	-360	360;
	5511	8419	0.001789	0.01382	0	0	0	0	0	0	1	-360	360;
	5511	8419	0.00175	0.014081	0	0	0	0	0	0	1	-360	360;
	7611	3382	9.1e-05	0.00087	0	0	0	0	0	0	1	-360	360;
	7611	3382	9.1e-05	0.00086	0	0	0	0	0	0	1	-360	360;
	7611	7869	0.000229	0.00236	0	400	0	0	0	0	1	-360	360;
	686	5171	0.00125	0.00557	0	591	0	0	0	0	1	-360	360;
	2172	5191	0.00757	0.04013	0	0	0	0	0	0	1	-360	360;
	1697	179	0.00157	0.01604	0	1382	0	0	0	0	1	-360	360;
	1697	179	0.00154	0.01634	0	1283	0	0	0	0	1	-360	360;
	179	2096	0.00169	0.02383	0	1547	0	0	0	0	1	-360	360;
	1728	179	0.00068	0.0089	0	1547	0	0	0	0	1	-360	360;
	1850	4603	0.010289	0.036651	0	476	0	0	0	0	1	-360	360;
	1850	4603	0.00563	0.03395	0	0	0	0	0	0	1	-360	360;
	3109	5726	0.003591	0.02763	0	0	0	0	0	0	1	-360	360;
	3109	5726	0.00351	0.028159	0	0	0	0	0	0	1	-360	360;
	5726	4773	0.00143	0.00838	0	0	0	0	0	0	1	-360	360;
	7659	6879	0.01213	0.06156	0	476	0	0	0	0	1	-360	360;
	6174	5171	0.004789	0.022849	0	553	0	0	0	0	1	-360	360;
	527	8242	0.00019	0.00211	0	457	0	0	0	0	1	-360	360;
	527	8529	0.000331	0.002349	0	0	0	0	0	0	1	-360	360;
	527	8530	9.1e-05	0.000539	0	381	0	0	0	0	1	-360	360;
	6496	6607	0.00088	0.00776	0	1382	0	0	0	0	1	-360	360;
	4652	4706	0.00039	0.00368	0	0	0	0	0	0	1	-360	360;
	6656	2091	0.00275	0.01268	0	400	0	0	0	0	1	-360	360;
	6656	8993	0.00205	0.00874	0	0	0	0	0	0	1	-360	360;
	6656	1241	0.00082	0.00519	0	591	0	0	0	0	1	-360	360;
	2658	808	0.00519	0.02355	0	457	0	0	0	0	1	-360	360;
	3291	6749	0.016711	0.08874	0	419	0	0	0	0	1	-360	360;
	7406	4397	0.005099	0.022659	0	476	0	0	0	0	1	-360	360;
	7406	1769	0.001469	0.01213	0	591	0	0	0	0	1	-360	360;
	7406	1769	0.001659	0.01189	0	533	0	0	0	0	1	-360	360;
	1013	4467	0.000521	0.00374	0	419	0	0	0	0	1	-360	360;
	1013	5025	0.006461	0.048349	0	0	0	0	0	0	1	-360	360;
	7918	7310	0.00293	0.0257	0	1316	0	0	0	0	1	-360	360;
	7918	5608	0.00178	0.0157	0	1415	0	0	0	0	1	-360	360;
	5831	7918	0.00115	0.01332	0	1086	0	0	0	0	1	-360	360;
	7918	2202	0.00078	0.00752	0	1678	0	0	0	0	1	-360	360;
	77	6660	3.9e-05	0.00037	0	0	0	0	0	0	1	-360	360;
	8124	8419	0.0012	0.006729	0	533	0	0	0	0	1	-360	360;
	33	3334	0.000211	0.001659	0	0	0	0	0	0	1	-360	360;
	6490	8301	0.001659	0.011919	0	495	0	0	0	0	1	-360	360;
	6490	6869	0.002	0.0133	0	438	0	0	0	0	1	-360	360;
	4371	7820	0.00105	0.00837	0	0	0	0	0	0	1	-360	360;
	2896	5348	0.00138	0.014979	0	0	0	0	0	0	1	-360	360;
	5767	6836	0	0.0003	0	0	0	0	0	0	1	-360	360;
	2978	6836	0	0.0003	0	419	0	0	0	0	1	-360	360;
	2978	5391	0.00164	0.00782	0	400	0	0	0	0	1	-360	360;
	3629	230	0.007099	0.04044	0	514	0	0	0	0	1	-360	360;
	9039	7861	0.00181	0.00914	0	572	0	0	0	0	1	-360	360;
	1479	7343	0.00027	0.00247	0	1942	0	0	0	0	1	-360	360;
	7343	4957	0.00266	0.03206	0	1645	0	0	0	0	1	-360	360;
	8338	981	0.000591	0.005081	0	400	0	0	0	0	1	-360	360;
	8338	751	0.002229	0.016099	0	419	0	0	0	0	1	-360	360;
	4026	5372	0.00205	0.02585	0	1086	0	0	0	0	1	-360	360;
	7821	642	0.00101	0.0091	0	1316	0	0	0	0	1	-360	360;
	7821	5608	0.001	0.00871	0	1316	0	0	0	0	1	-360	360;
	9097	1513	0.00112	0.00895	0	0	0	0	0	0	1	-360	360;
	9097	1513	0.001099	0.008289	0	0	0	0	0	0	1	-360	360;
	9097	3125	0.001289	0.0095	0	0	0	0	0	0	1	-360	360;
	9097	3125	0.00126	0.00968	0	0	0	0	0	0	1	-360	360;
	4315	5464	0.00065	0.00723	0	1547	0	0	0	0	1	-360	360;
	1942	4095	0.001781	0.012669	0	0	0	0	0	0	1	-360	360;
	3966	1074	0.00086	0.00956	0	1053	0	0	0	0	1	-360	360;
	3966	807	0.00248	0.02429	0	1283	0	0	0	0	1	-360	360;
	3966	4507	0.00151	0.01699	0	1316	0	0	0	0	1	-360	360;
	5890	3689	0.003159	0.01537	0	0	0	0	0	0	1	-360	360;
	8210	3113	0.00075	0.00683	0	1349	0	0	0	0	1	-360	360;
	8210	2849	0.00223	0.02524	0	1613	0	0	0	0	1	-360	360;
	642	5608	0.00191	0.0208	0	987	0	0	0	0	1	-360	360;
	1697	642	0.00074	0.00784	0	921	0	0	0	0	1	-360	360;
	1697	642	0.00073	0.00799	0	1086	0	0	0	0	1	-360	360;
	1421	3621	0.000789	0.00482	0	514	0	0	0	0	1	-360	360;
	2126	1421	0.000789	0.004409	0	381	0	0	0	0	1	-360	360;
	2844	3291	0.005461	0.03161	0	533	0	0	0	0	1	-360	360;
	1638	8575	6e-05	0.00087	0	0	0	0	0	0	1	-360	360;
	3163	1411	0.00593	0.040961	0	0	0	0	0	0	1	-360	360;
	3163	5025	0.003771	0.030841	0	0	0	0	0	0	1	-360	360;
	6454	8571	0.00038	0.001461	0	533	0	0	0	0	1	-360	360;
	8571	4081	0.00149	0.00938	0	0	0	0	0	0	1	-360	360;
	8571	4081	0.001521	0.00918	0	0	0	0	0	0	1	-360	360;
	2091	440	0.00407	0.015469	0	0	0	0	0	0	1	-360	360;
	4953	1421	0.00082	0.006219	0	0	0	0	0	0	1	-360	360;
	2667	1969	0.00161	0.00888	0	438	0	0	0	0	1	-360	360;
	2667	9028	0.00118	0.006461	0	514	0	0	0	0	1	-360	360;
	8819	6134	0.00082	0.009091	0	0	0	0	0	0	1	-360	360;
	751	4493	0.01893	0.0848	0	457	0	0	0	0	1	-360	360;
	751	4493	0.01313	0.082409	0	438	0	0	0	0	1	-360	360;
	751	6344	0.000281	0.00149	0	0	0	0	0	0	1	-360	360;
	5289	3113	0.00019	0.00199	0	1349	0	0	0	0	1	-360	360;
	3113	5289	0.0002	0.00188	0	1382	0	0	0	0	1	-360	360;
	172	2946	0.003159	0.025331	0	0	0	0	0	0	1	-360	360;
	7685	6394	0.00445	0.01886	0	610	0	0	0	0	1	-360	360;
	5821	5329	0.00461	0.02205	0	514	0	0	0	0	1	-360	360;
	7283	8213	0.00061	0.00494	0	0	0	0	0	0	1	-360	360;
	7283	962	0.00082	0.00462	0	0	0	0	0	0	1	-360	360;
	3468	8661	0.000331	0.003271	0	553	0	0	0	0	1	-360	360;
	3468	1769	0.00056	0.004081	0	591	0	0	0	0	1	-360	360;
	6244	9055	0.00163	0.007521	0	0	0	0	0	0	1	-360	360;
	6244	1471	0.001659	0.00736	0	0	0	0	0	0	1	-360	360;
	6544	1503	0.00719	0.049211	0	0	0	0	0	0	1	-360	360;
	7463	1969	6e-05	0.00039	0	438	0	0	0	0	1	-360	360;
	1030	6022	0.00288	0.018	0	533	0	0	0	0	1	-360	360;
	1030	8776	0.00288	0.0182	0	495	0	0	0	0	1	-360	360;
	1030	5290	0.00111	0.008531	0	438	0	0	0	0	1	-360	360;
	1030	4279	0.00307	0.022289	0	514	0	0	0	0	1	-360	360;
	3067	4507	0.00091	0.00868	0	1547	0	0	0	0	1	-360	360;
	5608	9117	0.00098	0.0118	0	1481	0	0	0	0	1	-360	360;
	5608	7436	0.00053	0.00591	0	1645	0	0	0	0	1	-360	360;
	5608	2202	0.00086	0.00991	0	1349	0	0	0	0	1	-360	360;
	7002	3334	0.00545	0.034099	0	495	0	0	0	0	1	-360	360;
	2060	3495	0.00095	0.006091	0	553	0	0	0	0	1	-360	360;
	3544	807	0.00305	0.0272	0	1316	0	0	0	0	1	-360	360;
	869	4910	0.000281	0.001841	0	438	0	0	0	0	1	-360	360;
	869	6456	0.004531	0.02451	0	0	0	0	0	0	1	-360	360;
	4397	1503	0.00618	0.02961	0	533	0	0	0	0	1	-360	360;
	1513	6745	0.00064	0.004461	0	0	0	0	0	0	1	-360	360;
	916	4224	0.00143	0.00513	0	553	0	0	0	0	1	-360	360;
	916	1241	0.0052	0.021841	0	457	0	0	0	0	1	-360	360;
	602	4910	0.00138	0.008669	0	438	0	0	0	0	1	-360	360;
	6456	602	0.00314	0.013169	0	0	0	0	0	0	1	-360	360;
	4224	1909	0.002479	0.01331	0	476	0	0	0	0	1	-360	360;
	2297	2740	7e-05	0.00194	0	0	0	0	0	0	1	-360	360;
	9162	8419	0.001159	0.00931	0	0	0	0	0	0	1	-360	360;
	5391	6349	0.00062	0.003669	0	0	0	0	0	0	1	-360	360;
	2517	5163	0.015969	0.056711	0	629	0	0	0	0	1	-360	360;
	5163	3754	0.0085	0.03719	0	419	0	0	0	0	1	-360	360;
	2785	9055	0.001531	0.00732	0	0	0	0	0	0	1	-360	360;
	2785	9055	0.00205	0.01062	0	0	0	0	0	0	1	-360	360;
	4957	5464	0.00091	0.01056	0	757	0	0	0	0	1	-360	360;
	807	8625	0.00167	0.02069	0	1547	0	0	0	0	1	-360	360;
	6299	6879	0.002719	0.016219	0	514	0	0	0	0	1	-360	360;
	8247	4390	0.002	0.02376	0	1711	0	0	0	0	1	-360	360;
	8247	4390	0.00226	0.02328	0	1645	0	0	0	0	1	-360	360;
	7630	442	0.004151	0.02119	0	533	0	0	0	0	1	-360	360;
	8993	8378	0.001841	0.010909	0	0	0	0	0	0	1	-360	360;
	6879	4181	0.000669	0.00399	0	476	0	0	0	0	1	-360	360;
	3662	1475	8e-05	0.00098	0	1613	0	0	0	0	1	-360	360;
	8242	8530	0.0002	0.0025	0	0	0	0	0	0	1	-360	360;
	5494	5290	0.006219	0.031599	0	533	0	0	0	0	1	-360	360;
	5494	1437	0.00438	0.023169	0	533	0	0	0	0	1	-360	360;
	6145	4279	0.00581	0.03088	0	514	0	0	0	0	1	-360	360;
	6145	4279	0.0065	0.02637	0	438	0	0	0	0	1	-360	360;
	4652	597	0.00198	0.02208	0	1678	0	0	0	0	1	-360	360;
	3109	5329	0.00093	0.006659	0	0	0	0	0	0	1	-360	360;
	3109	5329	0.000909	0.006789	0	0	0	0	0	0	1	-360	360;
	4026	1074	0.00244	0.01919	0	856	0	0	0	0	1	-360	360;
	2355	8582	0.00194	0.01549	0	0	0	0	0	0	1	-360	360;
	1666	4058	0.0002	0.001479	0	514	0	0	0	0	1	-360	360;
	4493	1513	0.001091	0.009461	0	610	0	0	0	0	1	-360	360;
	1513	6604	0.003081	0.00899	0	495	0	0	0	0	1	-360	360;
	6604	1457	0.00357	0.0138	0	457	0	0	0	0	1	-360	360;
	3495	7629	0.001229	0.00986	0	476	0	0	0	0	1	-360	360;
	1985	9106	0.001211	0.00913	0	572	0	0	0	0	1	-360	360;
	4957	8012	0.00116	0.01053	0	1580	0	0	0	0	1	-360	360;
	4651	5681	0.0017	0.009781	0	0	0	0	0	0	1	-360	360;
	4651	5681	0.001521	0.011091	0	0	0	0	0	0	1	-360	360;
	7795	445	0.0043	0.0233	0	572	0	0	0	0	1	-360	360;
	2700	4446	0.00019	0.0016	0	987	0	0	0	0	1	-360	360;
	4772	8301	0.00838	0.072479	0	0	0	0	0	0	1	-360	360;
	4772	988	0.00787	0.049211	0	0	0	0	0	0	1	-360	360;
	442	230	0.00557	0.03376	0	0	0	0	0	0	1	-360	360;
	4238	7312	0.01213	0.06564	0	0	0	0	0	0	1	-360	360;
	7366	3125	0.003531	0.01086	0	400	0	0	0	0	1	-360	360;
	7366	7423	0.0158	0.057031	0	0	0	0	0	0	1	-360	360;
	754	2096	0.00289	0.0244	0	1283	0	0	0	0	1	-360	360;
	1599	4096	0.000521	0.00413	0	553	0	0	0	0	1	-360	360;
	8079	4096	0.000531	0.004039	0	495	0	0	0	0	1	-360	360;
	230	1802	0.00431	0.02226	0	419	0	0	0	0	1	-360	360;
	3027	4081	0.00163	0.00899	0	553	0	0	0	0	1	-360	360;
	4652	1280	0.00274	0.02428	0	1415	0	0	0	0	1	-360	360;
	6315	6022	0.0093	0.056979	0	514	0	0	0	0	1	-360	360;
	6315	8729	0.011281	0.080039	0	0	0	0	0	0	1	-360	360;
	6315	5631	0.00449	0.02376	0	438	0	0	0	0	1	-360	360;
	2622	8260	0.00014	0.00148	0	1744	0	0	0	0	1	-360	360;
	6797	8301	0.0055	0.033659	0	419	0	0	0	0	1	-360	360;
	4467	5465	0.00249	0.02082	0	457	0	0	0	0	1	-360	360;
	1728	2096	0.00164	0.02252	0	1580	0	0	0	0	1	-360	360;
	5793	2147	0.000961	0.0047	0	0	0	0	0	0	1	-360	360;
	5793	2147	0.000979	0.004599	0	0	0	0	0	0	1	-360	360;
	7688	385	0.00057	0.00515	0	1251	0	0	0	0	1	-360	360;
	8301	6869	0.00143	0.011651	0	381	0	0	0	0	1	-360	360;
	1501	4081	0.00126	0.009211	0	0	0	0	0	0	1	-360	360;
	6022	695	9.1e-05	0.000211	0	381	0	0	0	0	1	-360	360;
	8776	695	0.000211	0.000711	0	0	0	0	0	0	1	-360	360;
	8776	5631	0.01095	0.07062	0	457	0	0	0	0	1	-360	360;
	2143	5631	0.00125	0.007281	0	0	0	0	0	0	1	-360	360;
	2143	5631	0.001	0.0045	0	0	0	0	0	0	1	-360	360;
	5465	5025	0.00282	0.026419	0	0	0	0	0	0	1	-360	360;
	7632	6625	0.003479	0.01456	0	0	0	0	0	0	1	-360	360;
	5290	4279	0.005521	0.03418	0	457	0	0	0	0	1	-360	360;
	5290	4279	0.00563	0.033461	0	476	0	0	0	0	1	-360	360;
	5025	4752	0.00095	0.0072	0	419	0	0	0	0	1	-360	360;
	5025	7312	0.00093	0.007159	0	514	0	0	0	0	1	-360	360;
	1437	4279	0.00988	0.058651	0	572	0	0	0	0	1	-360	360;
	6654	7869	3.9e-05	0.00044	0	591	0	0	0	0	1	-360	360;
	6654	7869	3.9e-05	0.00043	0	533	0	0	0	0	1	-360	360;
	7351	5441	0.00018	0.000781	0	0	0	0	0	0	1	-360	360;
	4314	7571	0.002961	0.01669	0	419	0	0	0	0	1	-360	360;
	5803	3916	0.00056	0.00432	0	0	0	0	0	0	1	-360	360;
	6757	6036	0.0002	0.00246	0	757	0	0	0	0	1	-360	360;
	6757	6921	0.0003	0.00377	0	1415	0	0	0	0	1	-360	360;
	5019	9112	0.00014	0.00057	0	0	0	0	0	0	1	-360	360;
	2930	9112	0.00013	0.000729	0	0	0	0	0	0	1	-360	360;
	1436	9112	0.00012	0.000581	0	0	0	0	0	0	1	-360	360;
	1851	9112	0.00011	0.00064	0	0	0	0	0	0	1	-360	360;
	8331	9112	0.00019	0.00056	0	0	0	0	0	0	1	-360	360;
	6220	8791	0.00062	0.004909	0	629	0	0	0	0	1	-360	360;
	9203	7842	7e-05	0.00076	0	987	0	0	0	0	1	-360	360;
	9203	8997	0.00066	0.00622	0	1711	0	0	0	0	1	-360	360;
	9203	2129	0.00054	0.00752	0	0	0	0	0	0	1	-360	360;
	9203	29	0.00105	0.00877	0	1415	0	0	0	0	1	-360	360;
	118	8903	8e-05	0.00091	0	757	0	0	0	0	1	-360	360;
	118	4598	0.00056	0.00719	0	1053	0	0	0	0	1	-360	360;
	2426	5533	0.000531	0.002039	0	419	0	0	0	0	1	-360	360;
	2426	4831	0.008659	0.03769	0	400	0	0	0	0	1	-360	360;
	2426	8961	0.003849	0.021099	0	0	0	0	0	0	1	-360	360;
	2426	8522	0.00688	0.035969	0	400	0	0	0	0	1	-360	360;
	2426	6888	0.006729	0.03713	0	0	0	0	0	0	1	-360	360;
	2426	3401	0.00111	0.006651	0	514	0	0	0	0	1	-360	360;
	2426	641	0.0023	0.008909	0	0	0	0	0	0	1	-360	360;
	2426	2600	0.000841	0.00313	0	0	0	0	0	0	1	-360	360;
	2426	5351	0.003521	0.017479	0	400	0	0	0	0	1	-360	360;
	3051	718	3.9e-05	0.00025	0	610	0	0	0	0	1	-360	360;
	3036	2327	0.000159	0.000729	0	457	0	0	0	0	1	-360	360;
	8651	7473	0.005281	0.032099	0	495	0	0	0	0	1	-360	360;
	8651	7473	0.005651	0.029419	0	476	0	0	0	0	1	-360	360;
	903	3654	0.000271	0.00036	0	0	0	0	0	0	1	-360	360;
	8732	7050	0.00044	0.00143	0	572	0	0	0	0	1	-360	360;
	8732	3912	0.000159	0.00074	0	553	0	0	0	0	1	-360	360;
	2273	1973	0.00019	0.000591	0	438	0	0	0	0	1	-360	360;
	5131	2928	0.00019	0.00087	0	0	0	0	0	0	1	-360	360;
	6629	4918	0.002591	0.013659	0	419	0	0	0	0	1	-360	360;
	6629	7309	0.00305	0.01168	0	0	0	0	0	0	1	-360	360;
	413	1102	0.000159	0.000599	0	533	0	0	0	0	1	-360	360;
	413	1102	0.000159	0.00062	0	495	0	0	0	0	1	-360	360;
	3248	4918	0.00101	0.004229	0	572	0	0	0	0	1	-360	360;
	3248	4281	0.00319	0.018651	0	553	0	0	0	0	1	-360	360;
	3248	7309	0.000729	0.00337	0	0	0	0	0	0	1	-360	360;
	1146	1187	0.002961	0.012831	0	381	0	0	0	0	1	-360	360;
	1146	7945	0.000591	0.00339	0	0	0	0	0	0	1	-360	360;
	8900	2421	0.00024	0.001219	0	0	0	0	0	0	1	-360	360;
	6071	2972	0.006219	0.036341	0	476	0	0	0	0	1	-360	360;
	6071	6031	0.001841	0.01193	0	572	0	0	0	0	1	-360	360;
	3293	7264	0.001409	0.01113	0	514	0	0	0	0	1	-360	360;
	3293	283	0.00151	0.00768	0	476	0	0	0	0	1	-360	360;
	3293	3069	0.00599	0.03313	0	495	0	0	0	0	1	-360	360;
	3293	3390	0.001771	0.010151	0	457	0	0	0	0	1	-360	360;
	4689	4936	0.00074	0.00311	0	0	0	0	0	0	1	-360	360;
	4689	4936	0.000729	0.003159	0	0	0	0	0	0	1	-360	360;
	3218	7809	0.005271	0.01906	0	381	0	0	0	0	1	-360	360;
	3218	5455	0.00906	0.053151	0	419	0	0	0	0	1	-360	360;
	3218	1857	0.00405	0.0153	0	0	0	0	0	0	1	-360	360;
	6806	5410	0.00055	0.003961	0	495	0	0	0	0	1	-360	360;
	6806	5410	0.00055	0.003729	0	438	0	0	0	0	1	-360	360;
	6806	1609	0.00107	0.007789	0	514	0	0	0	0	1	-360	360;
	6806	1609	0.00112	0.00775	0	572	0	0	0	0	1	-360	360;
	6806	1609	0.001031	0.00775	0	0	0	0	0	0	1	-360	360;
	6806	2438	0.00062	0.003219	0	0	0	0	0	0	1	-360	360;
	6806	954	0.00068	0.002659	0	514	0	0	0	0	1	-360	360;
	9140	502	0.00193	0.012479	0	591	0	0	0	0	1	-360	360;
	9140	3013	0.001159	0.007229	0	0	0	0	0	0	1	-360	360;
	9140	2949	0.00137	0.008729	0	495	0	0	0	0	1	-360	360;
	9140	2949	0.001401	0.00855	0	533	0	0	0	0	1	-360	360;
	9140	2949	0.001331	0.008521	0	533	0	0	0	0	1	-360	360;
	7070	2208	0.00031	0.00081	0	381	0	0	0	0	1	-360	360;
	7070	8060	0.000341	0.00119	0	381	0	0	0	0	1	-360	360;
	7070	4484	0.00031	0.001281	0	476	0	0	0	0	1	-360	360;
	7070	4239	0.00094	0.0035	0	533	0	0	0	0	1	-360	360;
	8535	7047	7e-05	0.000651	0	0	0	0	0	0	1	-360	360;
	5477	3657	0.00019	0.00124	0	591	0	0	0	0	1	-360	360;
	5477	2928	0.000331	0.002539	0	533	0	0	0	0	1	-360	360;
	3145	2918	0.0016	0.02113	0	0	0	0	0	0	1	-360	360;
	7770	5317	0.000909	0.004711	0	457	0	0	0	0	1	-360	360;
	7770	5317	0.00088	0.004969	0	419	0	0	0	0	1	-360	360;
	3	8676	0.001841	0.0088	0	0	0	0	0	0	1	-360	360;
	3	8189	0.01013	0.04238	0	495	0	0	0	0	1	-360	360;
	7267	2938	0.00059	0.00802	0	823	0	0	0	0	1	-360	360;
	7267	2938	0.00054	0.0079	0	757	0	0	0	0	1	-360	360;
	7267	3918	0.00171	0.02128	0	1086	0	0	0	0	1	-360	360;
	7267	3918	0.00174	0.02084	0	954	0	0	0	0	1	-360	360;
	7267	6581	0.00086	0.01433	0	1251	0	0	0	0	1	-360	360;
	7267	6581	0.00087	0.01404	0	823	0	0	0	0	1	-360	360;
	7267	3830	0.00126	0.01447	0	1218	0	0	0	0	1	-360	360;
	7267	5648	0.00122	0.01509	0	1152	0	0	0	0	1	-360	360;
	7267	5648	0.00154	0.019	0	1382	0	0	0	0	1	-360	360;
	4918	4907	0.00474	0.03564	0	572	0	0	0	0	1	-360	360;
	4918	5957	0.006	0.047849	0	533	0	0	0	0	1	-360	360;
	4918	5233	0.008039	0.03375	0	381	0	0	0	0	1	-360	360;
	6854	7309	2.1e-05	0.000211	0	610	0	0	0	0	1	-360	360;
	8311	3680	0.00045	0.002341	0	0	0	0	0	0	1	-360	360;
	2770	8804	0.000669	0.004281	0	533	0	0	0	0	1	-360	360;
	9241	8804	0.00068	0.00419	0	0	0	0	0	0	1	-360	360;
	3083	7124	0.000729	0.00431	0	610	0	0	0	0	1	-360	360;
	3083	7124	0.000591	0.004651	0	553	0	0	0	0	1	-360	360;
	3083	8873	0.00099	0.005081	0	438	0	0	0	0	1	-360	360;
	1566	7582	0.00212	0.013031	0	0	0	0	0	0	1	-360	360;
	1566	3541	0.002159	0.01275	0	0	0	0	0	0	1	-360	360;
	8683	3698	0.00025	0.00095	0	0	0	0	0	0	1	-360	360;
	8683	3698	0.000271	0.00089	0	0	0	0	0	0	1	-360	360;
	8683	2801	0.001961	0.012409	0	381	0	0	0	0	1	-360	360;
	8683	8328	0.00044	0.00207	0	572	0	0	0	0	1	-360	360;
	8683	8511	0.000469	0.00143	0	629	0	0	0	0	1	-360	360;
	8683	6570	0.001271	0.01406	0	629	0	0	0	0	1	-360	360;
	8683	6570	0.00124	0.013021	0	629	0	0	0	0	1	-360	360;
	8683	8975	0.000409	0.00239	0	0	0	0	0	0	1	-360	360;
	8683	216	0.00094	0.005	0	438	0	0	0	0	1	-360	360;
	8683	2961	0.000219	0.0007	0	629	0	0	0	0	1	-360	360;
	8683	6773	6e-05	0.00025	0	0	0	0	0	0	1	-360	360;
	8683	5278	5e-05	0.000229	0	0	0	0	0	0	1	-360	360;
	3768	2751	0.00019	0.000909	0	0	0	0	0	0	1	-360	360;
	2276	5546	0.001711	0.00782	0	0	0	0	0	0	1	-360	360;
	2276	6816	0.002651	0.01176	0	419	0	0	0	0	1	-360	360;
	7766	444	0.001409	0.010599	0	0	0	0	0	0	1	-360	360;
	7766	5334	0.001419	0.010219	0	0	0	0	0	0	1	-360	360;
	839	3918	0.00013	0.00212	0	856	0	0	0	0	1	-360	360;
	839	809	0.00026	0.00313	0	1382	0	0	0	0	1	-360	360;
	4506	2286	0.0047	0.0318	0	476	0	0	0	0	1	-360	360;
	4506	6802	0.00136	0.00768	0	438	0	0	0	0	1	-360	360;
	4506	3520	0.005669	0.027789	0	400	0	0	0	0	1	-360	360;
	7273	6271	0.004409	0.033091	0	381	0	0	0	0	1	-360	360;
	7273	1435	0.01056	0.044469	0	0	0	0	0	0	1	-360	360;
	7273	7148	0.01963	0.07736	0	438	0	0	0	0	1	-360	360;
	7273	7437	0.01157	0.04694	0	457	0	0	0	0	1	-360	360;
	6104	871	0.000479	0.00255	0	457	0	0	0	0	1	-360	360;
	7580	871	0.000469	0.00236	0	0	0	0	0	0	1	-360	360;
	1188	1539	0.01237	0.067039	0	381	0	0	0	0	1	-360	360;
	1188	8515	0.0132	0.056229	0	381	0	0	0	0	1	-360	360;
	3971	4084	0.00026	0.001711	0	0	0	0	0	0	1	-360	360;
	3971	4084	0.000271	0.001669	0	0	0	0	0	0	1	-360	360;
	6845	5340	0.000211	0.000401	0	0	0	0	0	0	1	-360	360;
	8294	5814	0.00105	0.005159	0	0	0	0	0	0	1	-360	360;
	7115	6807	0.00212	0.00931	0	419	0	0	0	0	1	-360	360;
	619	6791	0.000151	0.00068	0	0	0	0	0	0	1	-360	360;
	5418	4908	0.00019	0.00082	0	0	0	0	0	0	1	-360	360;
	5120	6816	0.003219	0.01405	0	400	0	0	0	0	1	-360	360;
	6692	9158	0.007271	0.033159	0	381	0	0	0	0	1	-360	360;
	6692	8035	0.006659	0.03132	0	400	0	0	0	0	1	-360	360;
	3445	2848	0.00057	0.00668	0	1580	0	0	0	0	1	-360	360;
	3445	7056	0.00057	0.00758	0	1514	0	0	0	0	1	-360	360;
	3445	7056	0.00039	0.00607	0	921	0	0	0	0	1	-360	360;
	3445	1642	0.00155	0.0179	0	0	0	0	0	0	1	-360	360;
	3445	4594	0.00036	0.00485	0	1645	0	0	0	0	1	-360	360;
	6114	1692	0.000581	0.004229	0	438	0	0	0	0	1	-360	360;
	6114	8190	0.00056	0.00431	0	0	0	0	0	0	1	-360	360;
	6114	5836	0.003099	0.020531	0	572	0	0	0	0	1	-360	360;
	6114	3191	0.00118	0.016919	0	610	0	0	0	0	1	-360	360;
	6114	5300	0.00188	0.01407	0	0	0	0	0	0	1	-360	360;
	6114	4889	0.00545	0.04082	0	457	0	0	0	0	1	-360	360;
	7264	444	0.00214	0.01639	0	553	0	0	0	0	1	-360	360;
	7264	4747	0.002479	0.01549	0	0	0	0	0	0	1	-360	360;
	7982	9011	0.000159	0.001229	0	572	0	0	0	0	1	-360	360;
	7982	6675	0.00026	0.00162	0	553	0	0	0	0	1	-360	360;
	7982	6909	0.000159	0.00113	0	476	0	0	0	0	1	-360	360;
	7982	8835	0.000461	0.003219	0	572	0	0	0	0	1	-360	360;
	7982	4186	0.000159	0.001229	0	0	0	0	0	0	1	-360	360;
	1604	502	0.000229	0.00139	0	476	0	0	0	0	1	-360	360;
	2842	502	0.0002	0.00156	0	438	0	0	0	0	1	-360	360;
	1354	513	0.00145	0.00824	0	381	0	0	0	0	1	-360	360;
	1354	502	0.00069	0.005	0	591	0	0	0	0	1	-360	360;
	891	3697	0.00036	0.00455	0	1909	0	0	0	0	1	-360	360;
	891	3697	0.00036	0.00453	0	1876	0	0	0	0	1	-360	360;
	8312	7541	0.00066	0.00997	0	1448	0	0	0	0	1	-360	360;
	8312	7541	0.00064	0.01016	0	1086	0	0	0	0	1	-360	360;
	6306	4826	9.9e-05	0.00044	0	629	0	0	0	0	1	-360	360;
	4826	4418	6e-05	0.000469	0	0	0	0	0	0	1	-360	360;
	4826	1605	3.9e-05	0.000341	0	0	0	0	0	0	1	-360	360;
	4826	6952	0.002531	0.01943	0	533	0	0	0	0	1	-360	360;
	4826	7124	0.00945	0.06418	0	514	0	0	0	0	1	-360	360;
	9128	217	0.000591	0.002831	0	457	0	0	0	0	1	-360	360;
	7014	217	0.00087	0.00387	0	0	0	0	0	0	1	-360	360;
	720	5350	0.0015	0.01785	0	1283	0	0	0	0	1	-360	360;
	720	5350	0.00147	0.0182	0	1251	0	0	0	0	1	-360	360;
	1711	639	0.00151	0.0113	0	476	0	0	0	0	1	-360	360;
	1711	639	0.001479	0.011521	0	495	0	0	0	0	1	-360	360;
	1711	9130	0.00174	0.010341	0	514	0	0	0	0	1	-360	360;
	1711	9130	0.001659	0.01206	0	553	0	0	0	0	1	-360	360;
	8209	1998	0.00069	0.000479	0	0	0	0	0	0	1	-360	360;
	8209	7974	0.00044	0.00139	0	0	0	0	0	0	1	-360	360;
	1035	8180	0.002039	0.01601	0	0	0	0	0	0	1	-360	360;
	1035	7466	0.01349	0.05662	0	457	0	0	0	0	1	-360	360;
	1035	305	0.00082	0.00438	0	400	0	0	0	0	1	-360	360;
	1035	305	0.0008	0.004461	0	381	0	0	0	0	1	-360	360;
	1035	5106	0.001091	0.00888	0	553	0	0	0	0	1	-360	360;
	1035	5106	0.00111	0.007039	0	591	0	0	0	0	1	-360	360;
	2527	8466	0.000849	0.00537	0	533	0	0	0	0	1	-360	360;
	8669	8893	0.00061	0.00275	0	457	0	0	0	0	1	-360	360;
	6206	7691	0.00141	0.0188	0	1415	0	0	0	0	1	-360	360;
	6206	6624	0.00121	0.01427	0	1645	0	0	0	0	1	-360	360;
	3240	8158	0.001711	0.007159	0	438	0	0	0	0	1	-360	360;
	3240	8795	0.00545	0.02381	0	400	0	0	0	0	1	-360	360;
	3240	4143	0.000771	0.004331	0	0	0	0	0	0	1	-360	360;
	3240	2797	0.00414	0.03282	0	553	0	0	0	0	1	-360	360;
	3240	6331	0.00264	0.011521	0	438	0	0	0	0	1	-360	360;
	3240	6384	0.00261	0.011	0	476	0	0	0	0	1	-360	360;
	3240	7523	0.00239	0.00919	0	419	0	0	0	0	1	-360	360;
	3240	2889	0.000539	0.00226	0	0	0	0	0	0	1	-360	360;
	3240	6887	0.00325	0.017979	0	572	0	0	0	0	1	-360	360;
	1249	4541	0.00076	0.0071	0	1316	0	0	0	0	1	-360	360;
	819	7523	0.000651	0.00343	0	0	0	0	0	0	1	-360	360;
	2304	4353	0.001219	0.00812	0	438	0	0	0	0	1	-360	360;
	2304	4816	0.00243	0.01439	0	591	0	0	0	0	1	-360	360;
	9174	6246	0.0006	0.00839	0	1251	0	0	0	0	1	-360	360;
	9174	6246	0.00059	0.00856	0	1185	0	0	0	0	1	-360	360;
	9174	5658	0.00087	0.01328	0	1777	0	0	0	0	1	-360	360;
	9174	5658	0.00073	0.01595	0	1349	0	0	0	0	1	-360	360;
	2924	3645	0.00163	0.002039	0	0	0	0	0	0	1	-360	360;
	2924	3645	0.001659	0.002	0	0	0	0	0	0	1	-360	360;
	59	5764	0.00339	0.019669	0	572	0	0	0	0	1	-360	360;
	59	2360	0.00239	0.017281	0	629	0	0	0	0	1	-360	360;
	6510	9033	0.00032	0.00112	0	495	0	0	0	0	1	-360	360;
	6510	5918	0.0018	0.01401	0	495	0	0	0	0	1	-360	360;
	6510	5918	0.00212	0.012469	0	533	0	0	0	0	1	-360	360;
	1275	2850	0.00063	0.00411	0	0	0	0	0	0	1	-360	360;
	6426	4110	0.009729	0.03505	0	0	0	0	0	0	1	-360	360;
	6426	3558	0.01007	0.03056	0	0	0	0	0	0	1	-360	360;
	1758	2816	0.00042	0.00623	0	0	0	0	0	0	1	-360	360;
	1758	352	0.00036	0.00722	0	1843	0	0	0	0	1	-360	360;
	1758	6581	0.00053	0.0078	0	954	0	0	0	0	1	-360	360;
	1758	1923	0.00109	0.0173	0	823	0	0	0	0	1	-360	360;
	1758	5837	0.00071	0.01144	0	0	0	0	0	0	1	-360	360;
	1758	3830	0.00157	0.01884	0	1251	0	0	0	0	1	-360	360;
	1758	3577	0.00032	0.00384	0	856	0	0	0	0	1	-360	360;
	8030	1794	0.00041	0.00649	0	1547	0	0	0	0	1	-360	360;
	8030	7808	0.0004	0.00658	0	0	0	0	0	0	1	-360	360;
	8030	6581	0.00054	0.00763	0	954	0	0	0	0	1	-360	360;
	8030	1923	0.00107	0.01602	0	921	0	0	0	0	1	-360	360;
	8030	5837	0.0007	0.01165	0	0	0	0	0	0	1	-360	360;
	8030	5648	0.0016	0.01844	0	1349	0	0	0	0	1	-360	360;
	8030	3577	0.00032	0.00376	0	954	0	0	0	0	1	-360	360;
	4615	171	0.00212	0.01207	0	495	0	0	0	0	1	-360	360;
	3137	2341	5e-05	0.00037	0	0	0	0	0	0	1	-360	360;
	6271	8787	0.00355	0.024651	0	419	0	0	0	0	1	-360	360;
	6252	9137	0.00675	0.04145	0	533	0	0	0	0	1	-360	360;
	5533	7133	0.00195	0.014341	0	438	0	0	0	0	1	-360	360;
	8361	4355	0.00018	0.00043	0	381	0	0	0	0	1	-360	360;
	8361	2928	0.00037	0.000969	0	457	0	0	0	0	1	-360	360;
	609	1341	0.00257	0.017169	0	476	0	0	0	0	1	-360	360;
	609	5419	0.00076	0.00399	0	495	0	0	0	0	1	-360	360;
	609	5419	0.00075	0.00406	0	495	0	0	0	0	1	-360	360;
	609	2079	0.003901	0.023961	0	572	0	0	0	0	1	-360	360;
	609	772	0.00249	0.019211	0	457	0	0	0	0	1	-360	360;
	3654	2928	0.00025	0.001159	0	476	0	0	0	0	1	-360	360;
	1355	7282	0.00195	0.02924	0	0	0	0	0	0	1	-360	360;
	1355	2197	0.00152	0.01527	0	1481	0	0	0	0	1	-360	360;
	1355	7640	0.00149	0.01558	0	1613	0	0	0	0	1	-360	360;
	1355	1081	0.00253	0.0276	0	1382	0	0	0	0	1	-360	360;
	1355	1081	0.00263	0.02648	0	1580	0	0	0	0	1	-360	360;
	1355	2359	0.00077	0.0092	0	1415	0	0	0	0	1	-360	360;
	1355	4196	0.00112	0.01705	0	757	0	0	0	0	1	-360	360;
	5856	8546	0.00226	0.01101	0	0	0	0	0	0	1	-360	360;
	5856	5987	0.004531	0.029039	0	495	0	0	0	0	1	-360	360;
	5856	8005	0.002081	0.030409	0	0	0	0	0	0	1	-360	360;
	5856	3925	0.0053	0.03257	0	438	0	0	0	0	1	-360	360;
	5856	3324	0.00513	0.041901	0	553	0	0	0	0	1	-360	360;
	5856	4748	0.001461	0.01056	0	457	0	0	0	0	1	-360	360;
	8265	4816	0.0005	0.002591	0	0	0	0	0	0	1	-360	360;
	6151	6313	0.00055	0.0033	0	457	0	0	0	0	1	-360	360;
	4005	6532	0.002289	0.01763	0	476	0	0	0	0	1	-360	360;
	4005	2177	0.0075	0.043531	0	457	0	0	0	0	1	-360	360;
	5723	2079	0.00138	0.019961	0	0	0	0	0	0	1	-360	360;
	5723	789	0.00126	0.018729	0	591	0	0	0	0	1	-360	360;
	2327	9189	0.00038	0.001331	0	495	0	0	0	0	1	-360	360;
	2327	3400	0.00019	0.001461	0	0	0	0	0	0	1	-360	360;
	207	4747	0.002151	0.013401	0	476	0	0	0	0	1	-360	360;
	207	7791	0.000789	0.005669	0	572	0	0	0	0	1	-360	360;
	3205	6735	3e-05	0.00042	0	757	0	0	0	0	1	-360	360;
	3205	666	0.00079	0.00899	0	1382	0	0	0	0	1	-360	360;
	3205	666	0.0008	0.00879	0	1547	0	0	0	0	1	-360	360;
	3205	4435	0.00334	0.03393	0	1382	0	0	0	0	1	-360	360;
	3205	1644	0.00063	0.00789	0	0	0	0	0	0	1	-360	360;
	3205	3483	0.0006	0.00842	0	1316	0	0	0	0	1	-360	360;
	4683	7989	0.00011	0.00125	0	0	0	0	0	0	1	-360	360;
	4683	6516	0.00012	0.001091	0	0	0	0	0	0	1	-360	360;
	4683	7697	9.9e-05	0.00125	0	0	0	0	0	0	1	-360	360;
	4683	416	0.00294	0.02443	0	629	0	0	0	0	1	-360	360;
	4683	2847	0.00144	0.011151	0	495	0	0	0	0	1	-360	360;
	4683	2563	0.001831	0.010849	0	591	0	0	0	0	1	-360	360;
	4683	2563	0.001599	0.01055	0	572	0	0	0	0	1	-360	360;
	4683	2995	0.00144	0.019591	0	381	0	0	0	0	1	-360	360;
	5691	1001	0.0133	0.09551	0	610	0	0	0	0	1	-360	360;
	5691	4426	0.009591	0.073961	0	514	0	0	0	0	1	-360	360;
	5350	2721	0.0007	0.00671	0	1547	0	0	0	0	1	-360	360;
	5350	2721	0.00072	0.00658	0	1580	0	0	0	0	1	-360	360;
	2361	4936	0.001901	0.00964	0	629	0	0	0	0	1	-360	360;
	2361	5179	0.001711	0.01189	0	610	0	0	0	0	1	-360	360;
	2044	4748	0.00025	0.00187	0	0	0	0	0	0	1	-360	360;
	4197	5522	0.00043	0.002211	0	0	0	0	0	0	1	-360	360;
	5699	8670	0.000401	0.00243	0	0	0	0	0	0	1	-360	360;
	7809	5110	0.00336	0.017	0	457	0	0	0	0	1	-360	360;
	7222	4816	0.005521	0.031031	0	0	0	0	0	0	1	-360	360;
	4402	3817	9.9e-05	0.00026	0	0	0	0	0	0	1	-360	360;
	6857	2732	0.00045	0.00718	0	1086	0	0	0	0	1	-360	360;
	6857	4550	0.00044	0.00665	0	921	0	0	0	0	1	-360	360;
	6857	4339	0.00103	0.01516	0	1283	0	0	0	0	1	-360	360;
	6857	7513	0.00101	0.01546	0	954	0	0	0	0	1	-360	360;
	5016	26	0.00045	0.00608	0	1514	0	0	0	0	1	-360	360;
	5016	9213	0.0004	0.00501	0	1349	0	0	0	0	1	-360	360;
	6112	9002	0.001729	0.01337	0	591	0	0	0	0	1	-360	360;
	6112	6231	0.00264	0.01489	0	514	0	0	0	0	1	-360	360;
	6112	883	0.002341	0.018961	0	514	0	0	0	0	1	-360	360;
	6112	5738	0.001841	0.01363	0	476	0	0	0	0	1	-360	360;
	8860	7700	0.001219	0.00949	0	495	0	0	0	0	1	-360	360;
	8860	9002	0.002271	0.01151	0	495	0	0	0	0	1	-360	360;
	8860	5990	0.010781	0.05775	0	0	0	0	0	0	1	-360	360;
	8860	6450	0.00251	0.013919	0	400	0	0	0	0	1	-360	360;
	8860	7726	0.00101	0.007331	0	514	0	0	0	0	1	-360	360;
	2489	4110	0.00061	0.003909	0	629	0	0	0	0	1	-360	360;
	3985	6376	0.004289	0.019331	0	0	0	0	0	0	1	-360	360;
	3985	4110	0.0008	0.004401	0	629	0	0	0	0	1	-360	360;
	3985	4829	0.008099	0.03519	0	610	0	0	0	0	1	-360	360;
	3985	7857	0.00063	0.004591	0	514	0	0	0	0	1	-360	360;
	2806	7857	0.000651	0.00456	0	0	0	0	0	0	1	-360	360;
	2489	9012	0.00114	0.00581	0	0	0	0	0	0	1	-360	360;
	2489	2231	0.001211	0.00538	0	0	0	0	0	0	1	-360	360;
	658	3793	0.00418	0.030159	0	572	0	0	0	0	1	-360	360;
	658	8497	0.00324	0.02413	0	610	0	0	0	0	1	-360	360;
	658	8497	0.003169	0.024591	0	629	0	0	0	0	1	-360	360;
	3502	9002	0.000219	0.00088	0	495	0	0	0	0	1	-360	360;
	3502	9002	0.000289	0.00095	0	514	0	0	0	0	1	-360	360;
	6313	44	0.0035	0.022211	0	553	0	0	0	0	1	-360	360;
	6313	7582	0.00313	0.02256	0	0	0	0	0	0	1	-360	360;
	5944	8808	0.000831	0.00162	0	0	0	0	0	0	1	-360	360;
	3183	8515	0.002099	0.01068	0	400	0	0	0	0	1	-360	360;
	1005	8515	0.00224	0.00976	0	0	0	0	0	0	1	-360	360;
	8787	1448	0.00169	0.01231	0	629	0	0	0	0	1	-360	360;
	6636	9051	0.00011	0.00076	0	0	0	0	0	0	1	-360	360;
	9109	9051	0.00011	0.00068	0	0	0	0	0	0	1	-360	360;
	6961	6376	0.005479	0.021909	0	0	0	0	0	0	1	-360	360;
	6961	8818	0.003331	0.017039	0	438	0	0	0	0	1	-360	360;
	5077	4353	0.000479	0.001271	0	0	0	0	0	0	1	-360	360;
	7471	5241	0.00109	0.01629	0	823	0	0	0	0	1	-360	360;
	7471	8200	4e-05	0.00063	0	1810	0	0	0	0	1	-360	360;
	549	6290	0.0053	0.024531	0	572	0	0	0	0	1	-360	360;
	549	2291	0.004659	0.021919	0	419	0	0	0	0	1	-360	360;
	549	3656	0.001659	0.01043	0	572	0	0	0	0	1	-360	360;
	5002	4144	0.006961	0.053539	0	495	0	0	0	0	1	-360	360;
	1888	5441	0.00036	0.00288	0	629	0	0	0	0	1	-360	360;
	1888	5441	0.00038	0.003	0	629	0	0	0	0	1	-360	360;
	5490	4970	0.00048	0.00643	0	0	0	0	0	0	1	-360	360;
	5490	4970	0.00051	0.0066	0	0	0	0	0	0	1	-360	360;
	5490	4970	0.00046	0.00729	0	1876	0	0	0	0	1	-360	360;
	5490	5461	0.0006	0.00868	0	1810	0	0	0	0	1	-360	360;
	5490	7778	0.00068	0.01076	0	1942	0	0	0	0	1	-360	360;
	5490	6897	0.00066	0.01096	0	1909	0	0	0	0	1	-360	360;
	3075	4562	0.00101	0.005099	0	400	0	0	0	0	1	-360	360;
	564	8475	0.003901	0.02437	0	0	0	0	0	0	1	-360	360;
	564	2252	0.00276	0.016461	0	553	0	0	0	0	1	-360	360;
	564	7903	0.00588	0.047081	0	0	0	0	0	0	1	-360	360;
	4331	1857	3.9e-05	0.000219	0	0	0	0	0	0	1	-360	360;
	2877	3674	0.00131	0.00538	0	0	0	0	0	0	1	-360	360;
	2854	1081	0.00679	0.04679	0	1020	0	0	0	0	1	-360	360;
	2854	8267	0.00138	0.01466	0	1086	0	0	0	0	1	-360	360;
	9112	9176	0.003531	0.01724	0	0	0	0	0	0	1	-360	360;
	9112	1422	0.01324	0.06936	0	476	0	0	0	0	1	-360	360;
	9112	8515	0.00363	0.01743	0	438	0	0	0	0	1	-360	360;
	9112	8515	0.00257	0.01662	0	533	0	0	0	0	1	-360	360;
	9112	5051	0.01399	0.056289	0	419	0	0	0	0	1	-360	360;
	9112	1051	0.00343	0.01751	0	457	0	0	0	0	1	-360	360;
	9112	1397	0.00301	0.015409	0	0	0	0	0	0	1	-360	360;
	9112	1584	0.00857	0.033091	0	476	0	0	0	0	1	-360	360;
	6168	6888	0.003341	0.01689	0	0	0	0	0	0	1	-360	360;
	7653	3019	3.9e-05	0.000219	0	0	0	0	0	0	1	-360	360;
	7341	7098	0.00087	0.00714	0	457	0	0	0	0	1	-360	360;
	4701	8267	0.00011	0.001	0	757	0	0	0	0	1	-360	360;
	1672	5365	0.000219	0.00101	0	0	0	0	0	0	1	-360	360;
	3880	5365	0.000219	0.001031	0	0	0	0	0	0	1	-360	360;
	6648	5365	0.000271	0.00101	0	0	0	0	0	0	1	-360	360;
	3609	5365	0.000271	0.000979	0	0	0	0	0	0	1	-360	360;
	2782	5365	0.000289	0.00107	0	0	0	0	0	0	1	-360	360;
	1844	5365	0.000289	0.00105	0	0	0	0	0	0	1	-360	360;
	4353	6053	0.001591	0.012289	0	381	0	0	0	0	1	-360	360;
	2166	5237	0.000271	0.001901	0	495	0	0	0	0	1	-360	360;
	2166	2886	0.00186	0.010711	0	514	0	0	0	0	1	-360	360;
	1043	3513	1e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	4049	883	0.000169	0.000271	0	0	0	0	0	0	1	-360	360;
	4520	883	0.000169	0.00031	0	0	0	0	0	0	1	-360	360;
	1704	5814	0.00526	0.03787	0	553	0	0	0	0	1	-360	360;
	8107	1486	0.000229	0.000599	0	0	0	0	0	0	1	-360	360;
	5664	5522	0.00089	0.004091	0	438	0	0	0	0	1	-360	360;
	4482	8670	0.00093	0.004	0	400	0	0	0	0	1	-360	360;
	8795	7523	0.00289	0.013099	0	400	0	0	0	0	1	-360	360;
	2337	6430	0.002599	0.016479	0	533	0	0	0	0	1	-360	360;
	455	7775	0.00369	0.02712	0	553	0	0	0	0	1	-360	360;
	455	3857	0.0062	0.04864	0	629	0	0	0	0	1	-360	360;
	455	333	0.008021	0.059031	0	629	0	0	0	0	1	-360	360;
	455	6940	0.004979	0.0295	0	572	0	0	0	0	1	-360	360;
	8225	3656	0.001159	0.005281	0	0	0	0	0	0	1	-360	360;
	5413	7752	0.00361	0.01824	0	0	0	0	0	0	1	-360	360;
	5413	1940	0.001521	0.01075	0	0	0	0	0	0	1	-360	360;
	5413	4234	0.00868	0.041711	0	0	0	0	0	0	1	-360	360;
	2591	6828	0.01118	0.05986	0	0	0	0	0	0	1	-360	360;
	2591	7571	0.003229	0.02543	0	591	0	0	0	0	1	-360	360;
	2591	7571	0.00218	0.01869	0	0	0	0	0	0	1	-360	360;
	2591	7571	0.001901	0.021909	0	572	0	0	0	0	1	-360	360;
	2591	7994	0.00974	0.043969	0	400	0	0	0	0	1	-360	360;
	2591	7226	0.003461	0.01814	0	457	0	0	0	0	1	-360	360;
	2591	1554	0.00487	0.027669	0	419	0	0	0	0	1	-360	360;
	9011	1156	0.000651	0.003961	0	610	0	0	0	0	1	-360	360;
	8689	5502	0.00169	0.00887	0	400	0	0	0	0	1	-360	360;
	3085	8873	0.001711	0.01336	0	514	0	0	0	0	1	-360	360;
	1568	980	0.00031	0.000539	0	400	0	0	0	0	1	-360	360;
	3204	3331	0.0027	0.01157	0	0	0	0	0	0	1	-360	360;
	3557	639	0.000909	0.006901	0	514	0	0	0	0	1	-360	360;
	3557	4936	0.001729	0.01214	0	514	0	0	0	0	1	-360	360;
	416	3492	0.00614	0.03106	0	438	0	0	0	0	1	-360	360;
	416	7937	0.00418	0.02388	0	514	0	0	0	0	1	-360	360;
	3255	1262	0.00238	0.017771	0	0	0	0	0	0	1	-360	360;
	3255	1803	8.1e-05	0.00063	0	0	0	0	0	0	1	-360	360;
	7373	367	0.001599	0.01618	0	0	0	0	0	0	1	-360	360;
	4251	2132	0.00532	0.026961	0	419	0	0	0	0	1	-360	360;
	4251	2132	0.00361	0.02814	0	553	0	0	0	0	1	-360	360;
	4251	5567	0.002469	0.01025	0	533	0	0	0	0	1	-360	360;
	6376	1808	0.005229	0.02675	0	400	0	0	0	0	1	-360	360;
	6376	1808	0.00512	0.02726	0	419	0	0	0	0	1	-360	360;
	6376	7857	0.00355	0.01694	0	400	0	0	0	0	1	-360	360;
	9185	7289	0.00139	0.007659	0	457	0	0	0	0	1	-360	360;
	9185	3999	0.000169	0.000669	0	514	0	0	0	0	1	-360	360;
	8653	2303	0.00111	0.00712	0	476	0	0	0	0	1	-360	360;
	908	6474	0.00031	0.001901	0	533	0	0	0	0	1	-360	360;
	7030	6474	0.000281	0.002091	0	0	0	0	0	0	1	-360	360;
	4205	6556	0.000281	0.001651	0	0	0	0	0	0	1	-360	360;
	7325	6556	0.000271	0.00169	0	0	0	0	0	0	1	-360	360;
	5857	3391	0.001521	0.00843	0	381	0	0	0	0	1	-360	360;
	5857	3760	0.000841	0.0062	0	533	0	0	0	0	1	-360	360;
	3707	8765	0.00019	0.00126	0	514	0	0	0	0	1	-360	360;
	900	6619	0.000219	0.00143	0	0	0	0	0	0	1	-360	360;
	2918	8109	0.00121	0.01796	0	1876	0	0	0	0	1	-360	360;
	2918	8109	0.0014	0.01512	0	0	0	0	0	0	1	-360	360;
	6532	2177	0.008409	0.063651	0	0	0	0	0	0	1	-360	360;
	6532	5317	0.00918	0.07344	0	610	0	0	0	0	1	-360	360;
	6532	608	0.004219	0.0345	0	514	0	0	0	0	1	-360	360;
	6532	608	0.0042	0.031651	0	457	0	0	0	0	1	-360	360;
	4494	8763	0.00121	0.01381	0	1547	0	0	0	0	1	-360	360;
	4494	7056	0.00107	0.01326	0	1580	0	0	0	0	1	-360	360;
	6828	5814	0.003401	0.02607	0	0	0	0	0	0	1	-360	360;
	3670	21	0.000289	0.00138	0	419	0	0	0	0	1	-360	360;
	5137	7702	0.001031	0.006039	0	0	0	0	0	0	1	-360	360;
	1334	7702	0.00105	0.005909	0	0	0	0	0	0	1	-360	360;
	8448	7624	0.001091	0.007901	0	419	0	0	0	0	1	-360	360;
	8448	1129	0.00332	0.024021	0	553	0	0	0	0	1	-360	360;
	8448	3070	0.00344	0.021419	0	400	0	0	0	0	1	-360	360;
	8448	306	0.00095	0.010419	0	514	0	0	0	0	1	-360	360;
	8448	306	0.00093	0.01062	0	476	0	0	0	0	1	-360	360;
	8448	7491	0.00474	0.025409	0	0	0	0	0	0	1	-360	360;
	4685	2089	0.000271	0.000841	0	381	0	0	0	0	1	-360	360;
	7485	6253	0.000271	0.00119	0	381	0	0	0	0	1	-360	360;
	7485	1562	0.000289	0.00076	0	0	0	0	0	0	1	-360	360;
	7776	1644	0.00031	0.00454	0	1942	0	0	0	0	1	-360	360;
	5971	3483	0.00032	0.0044	0	0	0	0	0	0	1	-360	360;
	2786	1644	0.00028	0.00494	0	0	0	0	0	0	1	-360	360;
	972	3483	0.00028	0.00508	0	0	0	0	0	0	1	-360	360;
	8906	923	0.000531	0.00093	0	0	0	0	0	0	1	-360	360;
	8906	4914	0.0003	0.001031	0	495	0	0	0	0	1	-360	360;
	2656	9101	0.00139	0.02163	0	757	0	0	0	0	1	-360	360;
	2656	7164	0.00164	0.02321	0	1876	0	0	0	0	1	-360	360;
	3519	9101	0.00136	0.02205	0	757	0	0	0	0	1	-360	360;
	3519	7164	0.0016	0.02367	0	757	0	0	0	0	1	-360	360;
	1775	9137	0.002091	0.01174	0	381	0	0	0	0	1	-360	360;
	1775	9137	0.001719	0.00961	0	553	0	0	0	0	1	-360	360;
	5641	1026	0.012341	0.076781	0	419	0	0	0	0	1	-360	360;
	5641	9137	0.00176	0.009419	0	591	0	0	0	0	1	-360	360;
	1096	2889	0.00031	0.00182	0	0	0	0	0	0	1	-360	360;
	7380	2732	0.00386	0.02348	0	1283	0	0	0	0	1	-360	360;
	7380	4787	0.00017	0.00104	0	1152	0	0	0	0	1	-360	360;
	3543	5308	0.00061	0.00808	0	1613	0	0	0	0	1	-360	360;
	3543	3830	0.00059	0.0067	0	0	0	0	0	0	1	-360	360;
	2372	1027	0.00059	0.00748	0	1481	0	0	0	0	1	-360	360;
	2372	5648	0.00061	0.0063	0	0	0	0	0	0	1	-360	360;
	367	1172	0.001031	0.008789	0	629	0	0	0	0	1	-360	360;
	367	9191	0.00051	0.006169	0	629	0	0	0	0	1	-360	360;
	367	8722	0.0005	0.00619	0	629	0	0	0	0	1	-360	360;
	1833	3610	0.00049	0.00376	0	572	0	0	0	0	1	-360	360;
	1833	3610	0.000539	0.003229	0	533	0	0	0	0	1	-360	360;
	1833	3962	0.001031	0.00882	0	0	0	0	0	0	1	-360	360;
	1833	1838	0.000711	0.00463	0	457	0	0	0	0	1	-360	360;
	1833	2230	0.000711	0.004539	0	0	0	0	0	0	1	-360	360;
	5388	2467	0.00206	0.015341	0	514	0	0	0	0	1	-360	360;
	5388	2467	0.002021	0.01564	0	514	0	0	0	0	1	-360	360;
	6110	8853	0.00094	0.005781	0	610	0	0	0	0	1	-360	360;
	6110	1001	0.002219	0.016969	0	629	0	0	0	0	1	-360	360;
	7583	1808	0.00238	0.01187	0	381	0	0	0	0	1	-360	360;
	7583	2889	0.00707	0.03095	0	381	0	0	0	0	1	-360	360;
	26	6624	0.00256	0.03401	0	1645	0	0	0	0	1	-360	360;
	26	2479	0.00203	0.02416	0	1152	0	0	0	0	1	-360	360;
	26	4598	0.001	0.01459	0	0	0	0	0	0	1	-360	360;
	26	4231	0.00103	0.00987	0	1613	0	0	0	0	1	-360	360;
	26	5144	0.00038	0.00483	0	1448	0	0	0	0	1	-360	360;
	26	5144	0.00041	0.00398	0	1316	0	0	0	0	1	-360	360;
	7691	2479	0.00167	0.01721	0	1382	0	0	0	0	1	-360	360;
	7691	2129	0.00098	0.01487	0	987	0	0	0	0	1	-360	360;
	7691	9213	0.00042	0.00342	0	1349	0	0	0	0	1	-360	360;
	7691	9213	0.00019	0.00248	0	1580	0	0	0	0	1	-360	360;
	7691	4231	0.00105	0.00966	0	1909	0	0	0	0	1	-360	360;
	7691	3513	0.00043	0.00473	0	1382	0	0	0	0	1	-360	360;
	7691	3513	0.00041	0.00397	0	1316	0	0	0	0	1	-360	360;
	1876	3906	0.00051	0.00514	0	1218	0	0	0	0	1	-360	360;
	113	3906	0.00053	0.00503	0	1415	0	0	0	0	1	-360	360;
	7752	3758	0.00455	0.022771	0	400	0	0	0	0	1	-360	360;
	7752	3758	0.00557	0.02055	0	381	0	0	0	0	1	-360	360;
	7752	3758	0.00339	0.02026	0	0	0	0	0	0	1	-360	360;
	4952	8976	0.00237	0.01068	0	438	0	0	0	0	1	-360	360;
	4952	8189	0.0037	0.019479	0	381	0	0	0	0	1	-360	360;
	6119	7886	0.01407	0.056849	0	0	0	0	0	0	1	-360	360;
	6119	4544	0.001669	0.015409	0	610	0	0	0	0	1	-360	360;
	6119	4544	0.001711	0.015091	0	0	0	0	0	0	1	-360	360;
	7905	3697	0.00172	0.01619	0	1678	0	0	0	0	1	-360	360;
	7905	7541	0.00029	0.00447	0	987	0	0	0	0	1	-360	360;
	7905	7541	0.00033	0.00387	0	1119	0	0	0	0	1	-360	360;
	7905	4141	0.00079	0.01154	0	987	0	0	0	0	1	-360	360;
	7905	4141	0.00077	0.01177	0	987	0	0	0	0	1	-360	360;
	7905	3022	0.00127	0.01498	0	1547	0	0	0	0	1	-360	360;
	1768	163	0.00214	0.012599	0	495	0	0	0	0	1	-360	360;
	4245	2425	0.000159	0.00125	0	457	0	0	0	0	1	-360	360;
	3698	5297	0.000211	0.000669	0	457	0	0	0	0	1	-360	360;
	750	8328	0.000341	0.001169	0	514	0	0	0	0	1	-360	360;
	750	870	0.00024	0.001469	0	572	0	0	0	0	1	-360	360;
	750	2961	0.00018	0.000669	0	533	0	0	0	0	1	-360	360;
	3499	7988	0.00033	0.00683	0	921	0	0	0	0	1	-360	360;
	3499	2197	0.00121	0.01211	0	1415	0	0	0	0	1	-360	360;
	3499	7640	0.00123	0.01185	0	1448	0	0	0	0	1	-360	360;
	3499	4231	0.00058	0.0097	0	1185	0	0	0	0	1	-360	360;
	3499	5144	0.00103	0.01292	0	1251	0	0	0	0	1	-360	360;
	3499	3513	0.00116	0.01266	0	921	0	0	0	0	1	-360	360;
	7289	3999	0.0013	0.006841	0	457	0	0	0	0	1	-360	360;
	3018	1051	6e-05	0.000349	0	438	0	0	0	0	1	-360	360;
	2702	3435	0.000169	0.0007	0	400	0	0	0	0	1	-360	360;
	9131	2020	0.000169	0.000719	0	0	0	0	0	0	1	-360	360;
	6308	6684	0.000271	0.001031	0	0	0	0	0	0	1	-360	360;
	4831	3680	0.008271	0.036841	0	0	0	0	0	0	1	-360	360;
	5410	2751	0.002969	0.02299	0	419	0	0	0	0	1	-360	360;
	3082	4623	0.00124	0.01273	0	1218	0	0	0	0	1	-360	360;
	3082	4339	0.00106	0.00989	0	1481	0	0	0	0	1	-360	360;
	3082	5288	7e-05	0.00029	0	757	0	0	0	0	1	-360	360;
	3082	5421	9e-05	0.00079	0	757	0	0	0	0	1	-360	360;
	3082	6475	0.00196	0.02049	0	1613	0	0	0	0	1	-360	360;
	3082	8468	0.00158	0.01519	0	1283	0	0	0	0	1	-360	360;
	6224	2732	0.00172	0.01811	0	1613	0	0	0	0	1	-360	360;
	6224	7513	0.00096	0.01121	0	1744	0	0	0	0	1	-360	360;
	6224	6475	0.00147	0.01707	0	0	0	0	0	0	1	-360	360;
	2174	4623	0.00129	0.0129	0	1316	0	0	0	0	1	-360	360;
	2174	4550	0.00179	0.01737	0	1645	0	0	0	0	1	-360	360;
	2174	6475	0.00131	0.01935	0	1777	0	0	0	0	1	-360	360;
	2083	9231	0.003901	0.01614	0	419	0	0	0	0	1	-360	360;
	2083	9231	0.002591	0.014771	0	553	0	0	0	0	1	-360	360;
	2083	9231	0.00224	0.0173	0	476	0	0	0	0	1	-360	360;
	2083	4054	0.00012	0.000849	0	0	0	0	0	0	1	-360	360;
	2083	3481	0.00394	0.03039	0	495	0	0	0	0	1	-360	360;
	2083	3602	0.00189	0.01	0	0	0	0	0	0	1	-360	360;
	2083	2794	0.00032	0.00257	0	0	0	0	0	0	1	-360	360;
	2083	2794	0.00038	0.00224	0	0	0	0	0	0	1	-360	360;
	2083	8467	0.00751	0.03955	0	495	0	0	0	0	1	-360	360;
	2083	1136	0.00086	0.006531	0	0	0	0	0	0	1	-360	360;
	2083	22	0.00087	0.006591	0	0	0	0	0	0	1	-360	360;
	8853	1001	0.003031	0.02276	0	629	0	0	0	0	1	-360	360;
	8853	4950	0.006331	0.03911	0	0	0	0	0	0	1	-360	360;
	8853	4410	0.00605	0.02986	0	381	0	0	0	0	1	-360	360;
	8853	6697	0.003919	0.01612	0	438	0	0	0	0	1	-360	360;
	8853	1159	0.002039	0.0108	0	438	0	0	0	0	1	-360	360;
	8853	7165	0.006289	0.026271	0	514	0	0	0	0	1	-360	360;
	2938	2848	0.00042	0.00547	0	856	0	0	0	0	1	-360	360;
	2938	2848	0.00043	0.00536	0	757	0	0	0	0	1	-360	360;
	2938	7522	0.00147	0.01373	0	1777	0	0	0	0	1	-360	360;
	2938	7522	0.00131	0.01555	0	1547	0	0	0	0	1	-360	360;
	2938	1642	0.00221	0.02217	0	1810	0	0	0	0	1	-360	360;
	2938	809	0.00224	0.0217	0	1810	0	0	0	0	1	-360	360;
	9189	6791	0.00095	0.003289	0	400	0	0	0	0	1	-360	360;
	9189	3400	0.000229	0.00082	0	0	0	0	0	0	1	-360	360;
	9189	7092	0.00068	0.00306	0	0	0	0	0	0	1	-360	360;
	9189	3718	0.00049	0.00256	0	419	0	0	0	0	1	-360	360;
	9189	3701	0.000531	0.0022	0	0	0	0	0	0	1	-360	360;
	9189	7507	0.00136	0.00956	0	514	0	0	0	0	1	-360	360;
	9189	7507	0.0012	0.010729	0	476	0	0	0	0	1	-360	360;
	9189	7507	0.00105	0.00861	0	0	0	0	0	0	1	-360	360;
	9189	2406	0.000591	0.006969	0	0	0	0	0	0	1	-360	360;
	2801	8511	0.00025	0.0008	0	0	0	0	0	0	1	-360	360;
	2801	216	0.00113	0.00632	0	400	0	0	0	0	1	-360	360;
	453	2129	0.0003	0.00275	0	954	0	0	0	0	1	-360	360;
	8691	6952	0.00836	0.051659	0	629	0	0	0	0	1	-360	360;
	8691	6952	0.01338	0.073469	0	591	0	0	0	0	1	-360	360;
	8691	1183	0.0027	0.02089	0	0	0	0	0	0	1	-360	360;
	8691	4410	0.003419	0.02636	0	610	0	0	0	0	1	-360	360;
	8691	8743	0.009531	0.033711	0	457	0	0	0	0	1	-360	360;
	2446	8763	0.0004	0.00707	0	0	0	0	0	0	1	-360	360;
	2627	8763	0.00035	0.00701	0	0	0	0	0	0	1	-360	360;
	4951	6785	0.00194	0.01751	0	533	0	0	0	0	1	-360	360;
	4951	6785	0.00156	0.01936	0	514	0	0	0	0	1	-360	360;
	4623	964	0.00078	0.00941	0	1349	0	0	0	0	1	-360	360;
	4623	964	0.00074	0.00969	0	1580	0	0	0	0	1	-360	360;
	4623	8347	0.00042	0.00532	0	1678	0	0	0	0	1	-360	360;
	4623	8347	0.0004	0.00473	0	1843	0	0	0	0	1	-360	360;
	4623	8347	0.0006	0.00499	0	1547	0	0	0	0	1	-360	360;
	1609	6552	0.00086	0.00638	0	610	0	0	0	0	1	-360	360;
	1609	6552	0.000969	0.00625	0	610	0	0	0	0	1	-360	360;
	1609	8999	0.002781	0.01725	0	457	0	0	0	0	1	-360	360;
	4368	4816	0.005031	0.028081	0	553	0	0	0	0	1	-360	360;
	5481	960	1e-05	0.00021	0	0	0	0	0	0	1	-360	360;
	7913	960	1e-05	0.00025	0	0	0	0	0	0	1	-360	360;
	506	960	2e-05	0.00031	0	757	0	0	0	0	1	-360	360;
	2980	960	3e-05	0.0004	0	757	0	0	0	0	1	-360	360;
	4554	6382	0.000841	0.00474	0	533	0	0	0	0	1	-360	360;
	4554	6382	0.00082	0.00487	0	533	0	0	0	0	1	-360	360;
	6844	1595	0.00186	0.011979	0	0	0	0	0	0	1	-360	360;
	3672	5334	0.00056	0.003021	0	381	0	0	0	0	1	-360	360;
	6219	5334	0.00055	0.003081	0	0	0	0	0	0	1	-360	360;
	8997	6153	0.00114	0.01063	0	0	0	0	0	0	1	-360	360;
	1852	5935	0.003719	0.02295	0	629	0	0	0	0	1	-360	360;
	1852	5695	0.006669	0.04288	0	514	0	0	0	0	1	-360	360;
	2931	5799	0.00243	0.0198	0	0	0	0	0	0	1	-360	360;
	2931	7124	0.00845	0.04339	0	533	0	0	0	0	1	-360	360;
	2598	3674	0.00219	0.015169	0	381	0	0	0	0	1	-360	360;
	2598	2132	0.00357	0.0302	0	419	0	0	0	0	1	-360	360;
	2558	7571	0.005849	0.032419	0	419	0	0	0	0	1	-360	360;
	8834	1129	0.00406	0.01688	0	0	0	0	0	0	1	-360	360;
	7377	3324	0.00113	0.0065	0	0	0	0	0	0	1	-360	360;
	1265	4426	0.0068	0.04613	0	514	0	0	0	0	1	-360	360;
	1265	1965	0.00337	0.02011	0	419	0	0	0	0	1	-360	360;
	8060	7162	0.00011	0.0005	0	0	0	0	0	0	1	-360	360;
	6922	5482	0.003711	0.018521	0	438	0	0	0	0	1	-360	360;
	6922	1295	0.000271	0.00145	0	457	0	0	0	0	1	-360	360;
	8369	4710	0.00326	0.01824	0	0	0	0	0	0	1	-360	360;
	8961	7133	0.00118	0.00712	0	419	0	0	0	0	1	-360	360;
	8961	6888	0.00482	0.02232	0	381	0	0	0	0	1	-360	360;
	4185	7019	0.00013	0.000469	0	381	0	0	0	0	1	-360	360;
	4185	6253	0.000229	0.00075	0	419	0	0	0	0	1	-360	360;
	7895	3391	0.004591	0.035021	0	533	0	0	0	0	1	-360	360;
	7895	1448	0.005469	0.03924	0	610	0	0	0	0	1	-360	360;
	7895	1448	0.00536	0.0403	0	0	0	0	0	0	1	-360	360;
	7895	115	0.00424	0.033031	0	0	0	0	0	0	1	-360	360;
	5589	4306	2e-05	0.00024	0	757	0	0	0	0	1	-360	360;
	5589	1763	1e-05	0.00021	0	757	0	0	0	0	1	-360	360;
	5589	3697	0.00286	0.03653	0	1645	0	0	0	0	1	-360	360;
	5589	3794	0.00117	0.01183	0	1744	0	0	0	0	1	-360	360;
	5589	3613	0.00103	0.01336	0	1481	0	0	0	0	1	-360	360;
	5589	1502	0.00156	0.01969	0	1547	0	0	0	0	1	-360	360;
	5589	6555	0.00449	0.05772	0	1777	0	0	0	0	1	-360	360;
	5589	3608	0.00259	0.03469	0	1613	0	0	0	0	1	-360	360;
	1001	4134	0.000219	0.00118	0	0	0	0	0	0	1	-360	360;
	1001	1979	6e-05	0.00095	0	0	0	0	0	0	1	-360	360;
	1001	516	6e-05	0.000651	0	629	0	0	0	0	1	-360	360;
	1001	3580	8.1e-05	0.000841	0	629	0	0	0	0	1	-360	360;
	1001	3809	0.00274	0.020831	0	514	0	0	0	0	1	-360	360;
	1001	5616	0.004781	0.027729	0	553	0	0	0	0	1	-360	360;
	1001	5616	0.002849	0.02093	0	572	0	0	0	0	1	-360	360;
	1001	6697	0.00513	0.03014	0	514	0	0	0	0	1	-360	360;
	1001	7165	0.00113	0.00651	0	0	0	0	0	0	1	-360	360;
	1001	892	0.00125	0.00707	0	572	0	0	0	0	1	-360	360;
	7438	5334	0.003349	0.01463	0	0	0	0	0	0	1	-360	360;
	7438	1526	0.00245	0.0122	0	495	0	0	0	0	1	-360	360;
	1301	6952	0.00587	0.036521	0	400	0	0	0	0	1	-360	360;
	1301	6952	0.00569	0.03774	0	0	0	0	0	0	1	-360	360;
	1301	6952	0.006031	0.03545	0	629	0	0	0	0	1	-360	360;
	1301	7883	0.000531	0.00301	0	0	0	0	0	0	1	-360	360;
	1301	6430	0.00076	0.006099	0	514	0	0	0	0	1	-360	360;
	778	7209	0.008841	0.028919	0	0	0	0	0	0	1	-360	360;
	778	1910	0.00169	0.007091	0	0	0	0	0	0	1	-360	360;
	778	6820	0.001031	0.00361	0	0	0	0	0	0	1	-360	360;
	778	3071	0.00813	0.030961	0	0	0	0	0	0	1	-360	360;
	5297	6267	0.00013	0.00068	0	533	0	0	0	0	1	-360	360;
	3657	3855	0.000211	0.000729	0	629	0	0	0	0	1	-360	360;
	3657	3855	0.000169	0.000669	0	591	0	0	0	0	1	-360	360;
	3657	2928	0.000669	0.00251	0	495	0	0	0	0	1	-360	360;
	3657	2928	0.000711	0.00238	0	476	0	0	0	0	1	-360	360;
	3657	1172	0.000229	0.00238	0	591	0	0	0	0	1	-360	360;
	3657	9191	0.0007	0.0065	0	553	0	0	0	0	1	-360	360;
	3657	8722	0.00061	0.006031	0	514	0	0	0	0	1	-360	360;
	6472	5362	0.000281	0.0018	0	629	0	0	0	0	1	-360	360;
	6472	3962	0.000211	0.002409	0	0	0	0	0	0	1	-360	360;
	1198	1343	0.00056	0.00773	0	757	0	0	0	0	1	-360	360;
	1198	29	0.0023	0.01581	0	1218	0	0	0	0	1	-360	360;
	7473	3680	0.005039	0.032151	0	0	0	0	0	0	1	-360	360;
	7473	96	0.00364	0.01713	0	457	0	0	0	0	1	-360	360;
	7473	1868	0.002169	0.012409	0	0	0	0	0	0	1	-360	360;
	346	3184	0.00113	0.004719	0	0	0	0	0	0	1	-360	360;
	346	6570	0.00113	0.004979	0	476	0	0	0	0	1	-360	360;
	3184	4728	0.00264	0.01174	0	0	0	0	0	0	1	-360	360;
	2654	7886	0.00494	0.03181	0	476	0	0	0	0	1	-360	360;
	2654	2132	0.003711	0.023961	0	610	0	0	0	0	1	-360	360;
	4144	8874	0.001409	0.01101	0	457	0	0	0	0	1	-360	360;
	4144	1625	0.001169	0.00593	0	381	0	0	0	0	1	-360	360;
	4144	1625	0.00118	0.00587	0	0	0	0	0	0	1	-360	360;
	4144	3786	0.00264	0.012469	0	533	0	0	0	0	1	-360	360;
	4144	3786	0.002349	0.01414	0	495	0	0	0	0	1	-360	360;
	3422	4100	0.00014	0.0007	0	0	0	0	0	0	1	-360	360;
	6624	4419	3e-05	0.00054	0	1876	0	0	0	0	1	-360	360;
	6624	7159	3e-05	0.00055	0	1810	0	0	0	0	1	-360	360;
	6624	3661	4e-05	0.00058	0	1777	0	0	0	0	1	-360	360;
	6624	5486	4e-05	0.00054	0	1876	0	0	0	0	1	-360	360;
	6624	4231	0.00171	0.02197	0	1382	0	0	0	0	1	-360	360;
	6624	4231	0.00164	0.02064	0	1645	0	0	0	0	1	-360	360;
	6624	6921	0.00136	0.01756	0	1711	0	0	0	0	1	-360	360;
	6624	432	0.00097	0.00942	0	1481	0	0	0	0	1	-360	360;
	6624	432	0.00086	0.01109	0	1481	0	0	0	0	1	-360	360;
	6624	432	0.00085	0.0113	0	1448	0	0	0	0	1	-360	360;
	3353	6887	0.000979	0.0058	0	400	0	0	0	0	1	-360	360;
	2479	2050	0.00052	0.00593	0	1909	0	0	0	0	1	-360	360;
	2479	2050	0.00046	0.00672	0	1909	0	0	0	0	1	-360	360;
	5482	8704	0.001151	0.00636	0	572	0	0	0	0	1	-360	360;
	5482	8704	0.001031	0.007211	0	591	0	0	0	0	1	-360	360;
	5482	4710	0.002341	0.02286	0	533	0	0	0	0	1	-360	360;
	5482	1295	0.002281	0.02518	0	400	0	0	0	0	1	-360	360;
	5482	1910	0.005479	0.02114	0	381	0	0	0	0	1	-360	360;
	5482	2985	0.011159	0.08856	0	476	0	0	0	0	1	-360	360;
	5482	5853	0.000669	0.00249	0	400	0	0	0	0	1	-360	360;
	5482	7824	0.00193	0.01643	0	572	0	0	0	0	1	-360	360;
	5482	3071	0.00024	0.0017	0	438	0	0	0	0	1	-360	360;
	4525	195	0.010159	0.065849	0	0	0	0	0	0	1	-360	360;
	9066	7367	0.00294	0.02131	0	514	0	0	0	0	1	-360	360;
	9066	9155	0.018841	0.07469	0	0	0	0	0	0	1	-360	360;
	9066	6357	0.015599	0.052169	0	400	0	0	0	0	1	-360	360;
	7886	3126	0.005831	0.03969	0	495	0	0	0	0	1	-360	360;
	7886	8564	0.00087	0.00661	0	457	0	0	0	0	1	-360	360;
	7886	5567	0.00637	0.024659	0	381	0	0	0	0	1	-360	360;
	6416	3069	0.0027	0.020419	0	553	0	0	0	0	1	-360	360;
	6416	3876	0.000409	0.00314	0	438	0	0	0	0	1	-360	360;
	6416	8804	0.001669	0.01301	0	476	0	0	0	0	1	-360	360;
	6416	8804	0.00164	0.01325	0	495	0	0	0	0	1	-360	360;
	7831	4550	0.00089	0.00556	0	1283	0	0	0	0	1	-360	360;
	7831	4125	0.00186	0.01809	0	1613	0	0	0	0	1	-360	360;
	7831	3346	0.00381	0.02853	0	1020	0	0	0	0	1	-360	360;
	5525	2377	0.012419	0.072151	0	400	0	0	0	0	1	-360	360;
	5525	1401	0.0042	0.02556	0	400	0	0	0	0	1	-360	360;
	5525	2850	0.010771	0.0493	0	476	0	0	0	0	1	-360	360;
	5525	2430	0.010159	0.04995	0	400	0	0	0	0	1	-360	360;
	5525	9164	0.013169	0.05113	0	495	0	0	0	0	1	-360	360;
	8511	1033	0.000469	0.001159	0	400	0	0	0	0	1	-360	360;
	8511	6178	0.000229	0.00105	0	457	0	0	0	0	1	-360	360;
	3400	980	0.000271	0.001271	0	533	0	0	0	0	1	-360	360;
	2101	4598	0.00099	0.01194	0	1645	0	0	0	0	1	-360	360;
	2101	4541	0.00077	0.01184	0	0	0	0	0	0	1	-360	360;
	2101	5144	0.00034	0.00426	0	0	0	0	0	0	1	-360	360;
	3919	2129	0.00101	0.01168	0	1744	0	0	0	0	1	-360	360;
	3919	4541	0.00086	0.01046	0	1810	0	0	0	0	1	-360	360;
	3919	3513	0.00035	0.00417	0	0	0	0	0	0	1	-360	360;
	6675	1486	0.000159	0.00069	0	533	0	0	0	0	1	-360	360;
	4711	5836	0.000781	0.00488	0	629	0	0	0	0	1	-360	360;
	4711	575	0.000781	0.00511	0	533	0	0	0	0	1	-360	360;
	4711	8214	0.00064	0.003539	0	457	0	0	0	0	1	-360	360;
	4711	8214	0.00057	0.003781	0	438	0	0	0	0	1	-360	360;
	4711	3021	0.000531	0.00111	0	0	0	0	0	0	1	-360	360;
	4711	7967	0.000521	0.00113	0	0	0	0	0	0	1	-360	360;
	4711	5993	0.00176	0.00843	0	438	0	0	0	0	1	-360	360;
	4711	8804	0.002461	0.01975	0	572	0	0	0	0	1	-360	360;
	4513	6831	0.000591	0.00105	0	0	0	0	0	0	1	-360	360;
	2043	1422	0.00012	0.00064	0	0	0	0	0	0	1	-360	360;
	6714	1808	0.000581	0.004419	0	0	0	0	0	0	1	-360	360;
	6371	1808	0.000651	0.004331	0	0	0	0	0	0	1	-360	360;
	513	2866	0.001289	0.008289	0	495	0	0	0	0	1	-360	360;
	4970	9222	0.00233	0.03502	0	757	0	0	0	0	1	-360	360;
	4970	9222	0.00238	0.03781	0	757	0	0	0	0	1	-360	360;
	4970	2072	0.00052	0.00605	0	1382	0	0	0	0	1	-360	360;
	4970	5461	0.00023	0.00275	0	1514	0	0	0	0	1	-360	360;
	4970	8109	0.00134	0.01343	0	1711	0	0	0	0	1	-360	360;
	4970	8109	0.00116	0.01552	0	1481	0	0	0	0	1	-360	360;
	7955	707	0.004169	0.027919	0	514	0	0	0	0	1	-360	360;
	7955	4816	0.00063	0.008	0	0	0	0	0	0	1	-360	360;
	7955	1866	0.004841	0.028581	0	629	0	0	0	0	1	-360	360;
	7955	5488	0.007419	0.05787	0	591	0	0	0	0	1	-360	360;
	5093	2467	0.002719	0.020331	0	495	0	0	0	0	1	-360	360;
	2328	7178	0.000789	0.005	0	0	0	0	0	0	1	-360	360;
	9180	195	0.00238	0.013219	0	457	0	0	0	0	1	-360	360;
	9180	3133	0.003771	0.02701	0	495	0	0	0	0	1	-360	360;
	292	4544	0.001271	0.00939	0	0	0	0	0	0	1	-360	360;
	9002	6737	0.000841	0.00131	0	0	0	0	0	0	1	-360	360;
	8439	3758	0.005711	0.043039	0	591	0	0	0	0	1	-360	360;
	8439	3758	0.00562	0.041901	0	495	0	0	0	0	1	-360	360;
	4974	2940	0.00356	0.019781	0	533	0	0	0	0	1	-360	360;
	4974	163	0.004841	0.03149	0	533	0	0	0	0	1	-360	360;
	747	8507	0.00057	0.00107	0	0	0	0	0	0	1	-360	360;
	6909	4914	0.000539	0.00213	0	514	0	0	0	0	1	-360	360;
	444	4747	0.0007	0.00425	0	0	0	0	0	0	1	-360	360;
	1459	6495	0.0012	0.009531	0	0	0	0	0	0	1	-360	360;
	1459	2751	0.00206	0.015901	0	0	0	0	0	0	1	-360	360;
	3601	4728	0.00063	0.00312	0	0	0	0	0	0	1	-360	360;
	1251	6475	3e-05	0.00041	0	1481	0	0	0	0	1	-360	360;
	1093	6475	2e-05	0.00043	0	0	0	0	0	0	1	-360	360;
	5709	6475	3e-05	0.00039	0	1283	0	0	0	0	1	-360	360;
	3825	6475	3e-05	0.00038	0	0	0	0	0	0	1	-360	360;
	1415	3503	0.0008	0.004151	0	0	0	0	0	0	1	-360	360;
	4549	3503	0.000901	0.00381	0	0	0	0	0	0	1	-360	360;
	3072	639	0.000591	0.004341	0	476	0	0	0	0	1	-360	360;
	3072	639	0.000591	0.00405	0	495	0	0	0	0	1	-360	360;
	5007	4541	0.00218	0.0275	0	1152	0	0	0	0	1	-360	360;
	5007	8195	0.00184	0.01677	0	1086	0	0	0	0	1	-360	360;
	757	964	0.00058	0.00807	0	1086	0	0	0	0	1	-360	360;
	757	964	0.00056	0.00823	0	921	0	0	0	0	1	-360	360;
	757	3906	0.00123	0.01146	0	1415	0	0	0	0	1	-360	360;
	757	3906	0.0011	0.01298	0	1481	0	0	0	0	1	-360	360;
	757	8347	0.00156	0.01585	0	1777	0	0	0	0	1	-360	360;
	757	2919	0.00226	0.02781	0	1481	0	0	0	0	1	-360	360;
	3758	188	0.00614	0.02831	0	400	0	0	0	0	1	-360	360;
	3758	2242	0.00513	0.03401	0	0	0	0	0	0	1	-360	360;
	3903	4114	0.013219	0.0597	0	0	0	0	0	0	1	-360	360;
	5743	6990	0.012969	0.06087	0	0	0	0	0	0	1	-360	360;
	8626	8930	0.00437	0.032831	0	553	0	0	0	0	1	-360	360;
	8626	2467	0.005651	0.033229	0	591	0	0	0	0	1	-360	360;
	2161	5488	0.008969	0.03625	0	438	0	0	0	0	1	-360	360;
	2161	7772	0.012409	0.05645	0	419	0	0	0	0	1	-360	360;
	5003	6382	0.00038	0.002271	0	400	0	0	0	0	1	-360	360;
	6474	4084	0.00138	0.01089	0	572	0	0	0	0	1	-360	360;
	6474	4084	0.001539	0.012521	0	591	0	0	0	0	1	-360	360;
	6474	53	0.00205	0.014229	0	533	0	0	0	0	1	-360	360;
	6474	7259	0.00026	0.00164	0	495	0	0	0	0	1	-360	360;
	6474	2475	0.000289	0.00169	0	0	0	0	0	0	1	-360	360;
	5334	7577	0.003349	0.051211	0	457	0	0	0	0	1	-360	360;
	7961	2288	0.00038	0.001479	0	0	0	0	0	0	1	-360	360;
	1592	7098	0.00082	0.00625	0	553	0	0	0	0	1	-360	360;
	3187	2079	0.00275	0.015581	0	572	0	0	0	0	1	-360	360;
	6041	5926	0.000909	0.00739	0	0	0	0	0	0	1	-360	360;
	3697	8486	0.00082	0.01038	0	1547	0	0	0	0	1	-360	360;
	3697	7530	0.00093	0.01017	0	1382	0	0	0	0	1	-360	360;
	3697	1502	0.00137	0.01431	0	1876	0	0	0	0	1	-360	360;
	3697	1552	0.00215	0.02021	0	1678	0	0	0	0	1	-360	360;
	6952	3793	0.01189	0.058841	0	457	0	0	0	0	1	-360	360;
	6738	2526	0.00104	0.01241	0	1251	0	0	0	0	1	-360	360;
	6738	2526	0.00092	0.014	0	1251	0	0	0	0	1	-360	360;
	6738	2365	0.00055	0.00845	0	1152	0	0	0	0	1	-360	360;
	6738	2365	0.00054	0.00861	0	987	0	0	0	0	1	-360	360;
	6738	3794	0.00105	0.01057	0	1810	0	0	0	0	1	-360	360;
	6738	3613	0.00089	0.01122	0	1613	0	0	0	0	1	-360	360;
	6738	6901	0.00223	0.02584	0	1777	0	0	0	0	1	-360	360;
	8180	5106	0.00101	0.007841	0	553	0	0	0	0	1	-360	360;
	8180	5106	0.001081	0.008099	0	572	0	0	0	0	1	-360	360;
	6240	7165	0.011081	0.068479	0	0	0	0	0	0	1	-360	360;
	9222	8721	2e-05	0.00033	0	0	0	0	0	0	1	-360	360;
	9222	8872	1e-05	0.00029	0	1777	0	0	0	0	1	-360	360;
	9222	8672	0.00071	0.00892	0	1580	0	0	0	0	1	-360	360;
	9222	7164	0.00136	0.01526	0	1448	0	0	0	0	1	-360	360;
	5935	7772	0.005401	0.036021	0	591	0	0	0	0	1	-360	360;
	3134	3430	0.00437	0.033719	0	457	0	0	0	0	1	-360	360;
	3134	1051	0.005151	0.034919	0	514	0	0	0	0	1	-360	360;
	283	800	0.00038	0.00139	0	0	0	0	0	0	1	-360	360;
	283	1077	0.00037	0.0015	0	0	0	0	0	0	1	-360	360;
	283	4747	0.00069	0.004461	0	533	0	0	0	0	1	-360	360;
	3391	5213	0.00688	0.040591	0	381	0	0	0	0	1	-360	360;
	3391	115	0.001341	0.008349	0	476	0	0	0	0	1	-360	360;
	3391	115	0.001229	0.009669	0	629	0	0	0	0	1	-360	360;
	4867	5753	0.00137	0.006039	0	419	0	0	0	0	1	-360	360;
	7539	6556	0.001919	0.01261	0	476	0	0	0	0	1	-360	360;
	7539	871	0.00163	0.0095	0	533	0	0	0	0	1	-360	360;
	8893	8874	0.00488	0.0497	0	572	0	0	0	0	1	-360	360;
	8893	4505	0.001289	0.0087	0	553	0	0	0	0	1	-360	360;
	8893	3436	0.00211	0.01524	0	572	0	0	0	0	1	-360	360;
	8893	3436	0.00206	0.015539	0	591	0	0	0	0	1	-360	360;
	8893	3436	0.00276	0.01631	0	514	0	0	0	0	1	-360	360;
	8893	6556	0.00305	0.019521	0	610	0	0	0	0	1	-360	360;
	8893	8765	0.000409	0.0023	0	495	0	0	0	0	1	-360	360;
	8893	6619	0.00036	0.00268	0	514	0	0	0	0	1	-360	360;
	8165	6734	0.00032	0.00049	0	0	0	0	0	0	1	-360	360;
	1078	6734	0.00013	0.000349	0	0	0	0	0	0	1	-360	360;
	7988	4231	0.0002	0.00327	0	0	0	0	0	0	1	-360	360;
	804	2142	0.000919	0.00561	0	533	0	0	0	0	1	-360	360;
	804	2142	0.00094	0.00549	0	533	0	0	0	0	1	-360	360;
	804	7519	0.00175	0.01976	0	533	0	0	0	0	1	-360	360;
	804	6744	0.001729	0.019331	0	476	0	0	0	0	1	-360	360;
	2286	6802	0.00406	0.02581	0	572	0	0	0	0	1	-360	360;
	2286	769	0.00526	0.025961	0	514	0	0	0	0	1	-360	360;
	3239	1081	0.00179	0.02098	0	1349	0	0	0	0	1	-360	360;
	3239	5983	0.00162	0.02148	0	1382	0	0	0	0	1	-360	360;
	3239	8267	0.00492	0.04991	0	1053	0	0	0	0	1	-360	360;
	3239	960	0.00109	0.01361	0	1086	0	0	0	0	1	-360	360;
	3239	960	0.00097	0.0153	0	889	0	0	0	0	1	-360	360;
	3239	3906	0.00322	0.03022	0	1481	0	0	0	0	1	-360	360;
	1539	7361	0.032831	0.074151	0	0	0	0	0	0	1	-360	360;
	1539	6852	0.01411	0.058711	0	514	0	0	0	0	1	-360	360;
	1539	8515	0.02725	0.102979	0	438	0	0	0	0	1	-360	360;
	1539	3166	0.012229	0.04049	0	400	0	0	0	0	1	-360	360;
	1539	2676	0.01994	0.066719	0	0	0	0	0	0	1	-360	360;
	1539	3793	0.018591	0.07462	0	438	0	0	0	0	1	-360	360;
	1539	4529	0.008979	0.03251	0	0	0	0	0	0	1	-360	360;
	1539	4234	0.016531	0.086031	0	381	0	0	0	0	1	-360	360;
	8628	44	0.003591	0.02137	0	591	0	0	0	0	1	-360	360;
	8628	6630	0.00124	0.008	0	400	0	0	0	0	1	-360	360;
	8628	6630	0.001271	0.007831	0	438	0	0	0	0	1	-360	360;
	8628	1545	0.000469	0.00294	0	495	0	0	0	0	1	-360	360;
	8628	4215	0.000841	0.003289	0	495	0	0	0	0	1	-360	360;
	7163	1917	0.00219	0.016539	0	400	0	0	0	0	1	-360	360;
	7163	5519	0.002901	0.020901	0	400	0	0	0	0	1	-360	360;
	7647	225	0.0002	0.001211	0	0	0	0	0	0	1	-360	360;
	4864	1397	9.1e-05	0.000479	0	0	0	0	0	0	1	-360	360;
	7021	7119	0.005461	0.029469	0	419	0	0	0	0	1	-360	360;
	7021	3324	0.001979	0.011289	0	591	0	0	0	0	1	-360	360;
	1248	163	0.0023	0.01289	0	0	0	0	0	0	1	-360	360;
	9045	2457	0.000159	0.00037	0	0	0	0	0	0	1	-360	360;
	8874	3786	0.002789	0.01562	0	533	0	0	0	0	1	-360	360;
	5891	5574	0.002531	0.012331	0	381	0	0	0	0	1	-360	360;
	5891	7663	0.00357	0.023289	0	400	0	0	0	0	1	-360	360;
	5891	4889	0.00076	0.006039	0	514	0	0	0	0	1	-360	360;
	3997	1526	0.000469	0.00075	0	0	0	0	0	0	1	-360	360;
	184	1526	0.00044	0.00069	0	0	0	0	0	0	1	-360	360;
	2972	2088	0.00561	0.035521	0	591	0	0	0	0	1	-360	360;
	8676	933	0.004539	0.03468	0	0	0	0	0	0	1	-360	360;
	8676	933	0.00175	0.0067	0	419	0	0	0	0	1	-360	360;
	8676	3643	0.007461	0.031419	0	400	0	0	0	0	1	-360	360;
	8704	6723	0.00224	0.012909	0	610	0	0	0	0	1	-360	360;
	7047	21	0.0003	0.000961	0	419	0	0	0	0	1	-360	360;
	7047	21	0.00032	0.001409	0	0	0	0	0	0	1	-360	360;
	7047	1551	0.002031	0.0157	0	553	0	0	0	0	1	-360	360;
	7047	4103	0.00038	0.00306	0	514	0	0	0	0	1	-360	360;
	7047	7862	0.00118	0.0083	0	553	0	0	0	0	1	-360	360;
	7047	5458	0.001229	0.00787	0	572	0	0	0	0	1	-360	360;
	5574	5666	0.00301	0.01538	0	514	0	0	0	0	1	-360	360;
	6742	5458	0.00063	0.00114	0	381	0	0	0	0	1	-360	360;
	5270	7862	0.00064	0.00111	0	0	0	0	0	0	1	-360	360;
	7775	3857	0.004919	0.03837	0	629	0	0	0	0	1	-360	360;
	6290	3436	0.001831	0.008349	0	438	0	0	0	0	1	-360	360;
	3430	401	0.00536	0.03188	0	533	0	0	0	0	1	-360	360;
	2967	8976	0.00038	0.001461	0	0	0	0	0	0	1	-360	360;
	2967	8976	0.000289	0.001521	0	0	0	0	0	0	1	-360	360;
	8976	2481	0.00305	0.016841	0	476	0	0	0	0	1	-360	360;
	8976	9108	0.001781	0.012729	0	457	0	0	0	0	1	-360	360;
	2056	7520	0.00041	0.00461	0	954	0	0	0	0	1	-360	360;
	2056	7520	0.0004	0.0047	0	1020	0	0	0	0	1	-360	360;
	4084	5340	0.00213	0.01401	0	629	0	0	0	0	1	-360	360;
	4084	2319	0.001271	0.00887	0	438	0	0	0	0	1	-360	360;
	4084	2878	0.00462	0.02743	0	533	0	0	0	0	1	-360	360;
	2597	7422	0.001539	0.00776	0	0	0	0	0	0	1	-360	360;
	7183	8931	0.00025	0.00478	0	0	0	0	0	0	1	-360	360;
	5379	8931	0.00025	0.00487	0	0	0	0	0	0	1	-360	360;
	8128	8931	0.00025	0.00435	0	823	0	0	0	0	1	-360	360;
	6337	8931	0.00026	0.00426	0	757	0	0	0	0	1	-360	360;
	1237	8931	0.00025	0.00479	0	0	0	0	0	0	1	-360	360;
	4127	9173	0.000169	0.00076	0	438	0	0	0	0	1	-360	360;
	4127	7974	0.00039	0.00126	0	495	0	0	0	0	1	-360	360;
	2432	4816	0.011229	0.074659	0	629	0	0	0	0	1	-360	360;
	2432	1866	0.014521	0.058081	0	400	0	0	0	0	1	-360	360;
	2432	8809	0.00064	0.003919	0	533	0	0	0	0	1	-360	360;
	2432	5146	0.014789	0.07768	0	438	0	0	0	0	1	-360	360;
	2432	6940	0.00938	0.05355	0	572	0	0	0	0	1	-360	360;
	2432	1742	0.00124	0.004289	0	419	0	0	0	0	1	-360	360;
	5213	8627	0.00968	0.04388	0	0	0	0	0	0	1	-360	360;
	5213	174	0.008531	0.067151	0	457	0	0	0	0	1	-360	360;
	5213	7437	0.00731	0.04106	0	438	0	0	0	0	1	-360	360;
	7342	5940	0.004349	0.02555	0	533	0	0	0	0	1	-360	360;
	7342	7495	0.00599	0.03462	0	476	0	0	0	0	1	-360	360;
	8808	4816	0.002961	0.021591	0	553	0	0	0	0	1	-360	360;
	8808	4816	0.002901	0.02206	0	572	0	0	0	0	1	-360	360;
	8808	6053	0.002979	0.016969	0	476	0	0	0	0	1	-360	360;
	5764	4189	0.00312	0.020021	0	476	0	0	0	0	1	-360	360;
	8507	1973	0.000219	0.000909	0	0	0	0	0	0	1	-360	360;
	1394	1973	0.000219	0.00093	0	457	0	0	0	0	1	-360	360;
	2898	2128	0.000151	0.000289	0	0	0	0	0	0	1	-360	360;
	4942	8255	0.00563	0.043599	0	0	0	0	0	0	1	-360	360;
	4942	4748	0.000969	0.00612	0	514	0	0	0	0	1	-360	360;
	3486	3594	0.005021	0.020719	0	438	0	0	0	0	1	-360	360;
	3486	4679	0.00988	0.041039	0	0	0	0	0	0	1	-360	360;
	4656	4235	0.003159	0.02011	0	476	0	0	0	0	1	-360	360;
	4656	4235	0.003219	0.01969	0	476	0	0	0	0	1	-360	360;
	7129	5441	0.00049	0.00238	0	0	0	0	0	0	1	-360	360;
	7129	5441	0.000479	0.002419	0	0	0	0	0	0	1	-360	360;
	8250	5720	0.00264	0.019729	0	533	0	0	0	0	1	-360	360;
	8250	5720	0.002581	0.02011	0	495	0	0	0	0	1	-360	360;
	6831	6723	0.00037	0.002831	0	553	0	0	0	0	1	-360	360;
	6831	5686	0.00213	0.012789	0	514	0	0	0	0	1	-360	360;
	2377	805	0.018669	0.092281	0	438	0	0	0	0	1	-360	360;
	907	9091	3.9e-05	0.000401	0	0	0	0	0	0	1	-360	360;
	4056	9091	6e-05	0.00044	0	381	0	0	0	0	1	-360	360;
	6837	2142	6e-05	0.0005	0	514	0	0	0	0	1	-360	360;
	5666	2641	0.001271	0.007031	0	476	0	0	0	0	1	-360	360;
	5666	7577	0.00055	0.003159	0	610	0	0	0	0	1	-360	360;
	5666	7577	0.00055	0.00338	0	591	0	0	0	0	1	-360	360;
	5666	7577	0.00105	0.004771	0	495	0	0	0	0	1	-360	360;
	5666	1629	0.003289	0.015039	0	381	0	0	0	0	1	-360	360;
	726	687	0.00036	0.004349	0	610	0	0	0	0	1	-360	360;
	3526	687	0.00038	0.003669	0	0	0	0	0	0	1	-360	360;
	6382	2425	0.00118	0.00701	0	0	0	0	0	0	1	-360	360;
	6382	2425	0.001229	0.00686	0	533	0	0	0	0	1	-360	360;
	6382	5907	0.00301	0.0217	0	0	0	0	0	0	1	-360	360;
	6382	5907	0.003409	0.01924	0	0	0	0	0	0	1	-360	360;
	6382	5907	0.003409	0.01938	0	572	0	0	0	0	1	-360	360;
	7847	9217	0.002719	0.02131	0	0	0	0	0	0	1	-360	360;
	2981	1448	0.00295	0.01956	0	0	0	0	0	0	1	-360	360;
	5110	5686	0.001091	0.00831	0	476	0	0	0	0	1	-360	360;
	2968	5907	0.00111	0.008211	0	553	0	0	0	0	1	-360	360;
	5441	2155	0.002469	0.015271	0	629	0	0	0	0	1	-360	360;
	5441	8112	0.00212	0.01601	0	0	0	0	0	0	1	-360	360;
	5441	2535	0.00037	0.0028	0	533	0	0	0	0	1	-360	360;
	5441	7098	0.00162	0.01238	0	0	0	0	0	0	1	-360	360;
	5441	7098	0.001909	0.01162	0	0	0	0	0	0	1	-360	360;
	5441	7098	0.001909	0.010729	0	553	0	0	0	0	1	-360	360;
	5441	3737	0.00089	0.00526	0	419	0	0	0	0	1	-360	360;
	5441	5529	0.00087	0.00536	0	0	0	0	0	0	1	-360	360;
	5441	5753	0.001521	0.01176	0	533	0	0	0	0	1	-360	360;
	9019	6563	0.00014	0.000591	0	476	0	0	0	0	1	-360	360;
	1341	1101	0.003531	0.028781	0	0	0	0	0	0	1	-360	360;
	2797	8291	0.004599	0.03657	0	553	0	0	0	0	1	-360	360;
	2797	8367	0.00307	0.019159	0	591	0	0	0	0	1	-360	360;
	2797	1100	0.000581	0.004461	0	514	0	0	0	0	1	-360	360;
	2797	6887	0.00164	0.01161	0	591	0	0	0	0	1	-360	360;
	4939	6880	0.00057	0.003281	0	533	0	0	0	0	1	-360	360;
	7738	6982	0.020419	0.063669	0	0	0	0	0	0	1	-360	360;
	6565	3121	0.01424	0.04751	0	0	0	0	0	0	1	-360	360;
	8707	280	0.00157	0.008849	0	553	0	0	0	0	1	-360	360;
	8707	2510	0.01218	0.07524	0	0	0	0	0	0	1	-360	360;
	1625	766	0.00037	0.00151	0	0	0	0	0	0	1	-360	360;
	8411	6232	0.000219	0.000531	0	0	0	0	0	0	1	-360	360;
	3951	1397	0.002849	0.0137	0	0	0	0	0	0	1	-360	360;
	964	7282	0.00116	0.01692	0	1283	0	0	0	0	1	-360	360;
	964	7282	0.00133	0.01462	0	1415	0	0	0	0	1	-360	360;
	964	6475	0.00063	0.00821	0	987	0	0	0	0	1	-360	360;
	964	6475	0.00064	0.00804	0	1283	0	0	0	0	1	-360	360;
	5658	5420	0.00126	0.01272	0	1382	0	0	0	0	1	-360	360;
	5658	8267	0.00337	0.0494	0	1744	0	0	0	0	1	-360	360;
	5658	7328	0.0014	0.0214	0	0	0	0	0	0	1	-360	360;
	5658	7328	0.00137	0.02182	0	0	0	0	0	0	1	-360	360;
	5658	7164	0.00212	0.02063	0	1218	0	0	0	0	1	-360	360;
	5550	410	0.00186	0.01045	0	610	0	0	0	0	1	-360	360;
	5550	7579	0.00426	0.021771	0	438	0	0	0	0	1	-360	360;
	5550	7579	0.002659	0.020539	0	495	0	0	0	0	1	-360	360;
	5550	4189	0.00606	0.049919	0	457	0	0	0	0	1	-360	360;
	7282	4196	0.00109	0.01192	0	889	0	0	0	0	1	-360	360;
	1526	3916	0.013021	0.049219	0	438	0	0	0	0	1	-360	360;
	1526	4747	0.00236	0.01239	0	438	0	0	0	0	1	-360	360;
	1526	4747	0.0027	0.015169	0	419	0	0	0	0	1	-360	360;
	4541	3346	0.00285	0.03283	0	1514	0	0	0	0	1	-360	360;
	4541	3346	0.00529	0.03969	0	1020	0	0	0	0	1	-360	360;
	1343	3346	0.00254	0.03721	0	1547	0	0	0	0	1	-360	360;
	3680	2719	0.000159	0.00069	0	0	0	0	0	0	1	-360	360;
	3680	2719	0.000229	0.000979	0	0	0	0	0	0	1	-360	360;
	3680	2719	0.00024	0.00082	0	0	0	0	0	0	1	-360	360;
	3680	795	0.000169	0.00075	0	0	0	0	0	0	1	-360	360;
	3680	4880	0.00368	0.01756	0	0	0	0	0	0	1	-360	360;
	3680	1868	0.002599	0.020659	0	0	0	0	0	0	1	-360	360;
	7464	5564	0.00018	0.00064	0	0	0	0	0	0	1	-360	360;
	7464	7036	0.000159	0.000771	0	381	0	0	0	0	1	-360	360;
	7464	5940	0.00357	0.01981	0	476	0	0	0	0	1	-360	360;
	7464	7495	0.006021	0.03205	0	533	0	0	0	0	1	-360	360;
	4850	2886	0.001781	0.01164	0	533	0	0	0	0	1	-360	360;
	4850	7049	0.002781	0.018831	0	514	0	0	0	0	1	-360	360;
	2510	6146	0.000349	0.00281	0	0	0	0	0	0	1	-360	360;
	90	1860	0.00313	0.02444	0	610	0	0	0	0	1	-360	360;
	90	1860	0.00368	0.02074	0	591	0	0	0	0	1	-360	360;
	2142	3276	9.1e-05	0.001539	0	0	0	0	0	0	1	-360	360;
	2142	3545	0.00251	0.01111	0	0	0	0	0	0	1	-360	360;
	2142	6842	0.003081	0.017961	0	419	0	0	0	0	1	-360	360;
	2142	6744	0.005341	0.030419	0	438	0	0	0	0	1	-360	360;
	2142	7635	9.9e-05	0.00126	0	514	0	0	0	0	1	-360	360;
	9091	1040	0.002169	0.017901	0	419	0	0	0	0	1	-360	360;
	2457	6178	0.00011	0.00038	0	0	0	0	0	0	1	-360	360;
	1422	1462	0.00212	0.00494	0	0	0	0	0	0	1	-360	360;
	1422	7772	0.018901	0.077349	0	476	0	0	0	0	1	-360	360;
	7881	5365	0.00957	0.0395	0	457	0	0	0	0	1	-360	360;
	7881	6982	0.02324	0.063469	0	400	0	0	0	0	1	-360	360;
	6926	3327	0.000531	0.004331	0	610	0	0	0	0	1	-360	360;
	1959	615	0.002581	0.006229	0	0	0	0	0	0	1	-360	360;
	6455	1262	0.00031	0.002211	0	419	0	0	0	0	1	-360	360;
	803	6982	0.01087	0.06114	0	438	0	0	0	0	1	-360	360;
	803	5918	0.011581	0.061099	0	419	0	0	0	0	1	-360	360;
	7624	3070	0.002099	0.016271	0	400	0	0	0	0	1	-360	360;
	500	1547	0.00093	0.005461	0	0	0	0	0	0	1	-360	360;
	500	5237	0.009591	0.037039	0	419	0	0	0	0	1	-360	360;
	500	772	0.002591	0.017211	0	553	0	0	0	0	1	-360	360;
	500	8950	0.00164	0.013651	0	495	0	0	0	0	1	-360	360;
	2754	2079	0.003219	0.014719	0	457	0	0	0	0	1	-360	360;
	2754	2229	0.00401	0.022229	0	457	0	0	0	0	1	-360	360;
	2754	7049	0.00126	0.010211	0	591	0	0	0	0	1	-360	360;
	4118	7873	0.009169	0.03993	0	0	0	0	0	0	1	-360	360;
	4118	3865	0.001349	0.006081	0	0	0	0	0	0	1	-360	360;
	1754	8486	0.00156	0.01931	0	1580	0	0	0	0	1	-360	360;
	1754	7530	0.00152	0.01968	0	1481	0	0	0	0	1	-360	360;
	1754	7778	0.00094	0.01433	0	1678	0	0	0	0	1	-360	360;
	1754	6897	0.00092	0.01461	0	1448	0	0	0	0	1	-360	360;
	1754	960	0.00137	0.02012	0	1119	0	0	0	0	1	-360	360;
	1754	960	0.00134	0.02051	0	921	0	0	0	0	1	-360	360;
	3857	4491	0.00314	0.042151	0	572	0	0	0	0	1	-360	360;
	3857	6010	0.000651	0.00494	0	0	0	0	0	0	1	-360	360;
	3857	6940	0.003961	0.02426	0	591	0	0	0	0	1	-360	360;
	858	1642	0.00013	0.0015	0	0	0	0	0	0	1	-360	360;
	3028	809	0.00013	0.00147	0	1876	0	0	0	0	1	-360	360;
	4819	1642	0.00011	0.00172	0	1876	0	0	0	0	1	-360	360;
	2035	809	0.00012	0.00166	0	0	0	0	0	0	1	-360	360;
	3869	1642	0.00012	0.00168	0	1909	0	0	0	0	1	-360	360;
	3741	809	0.00012	0.00172	0	1909	0	0	0	0	1	-360	360;
	742	687	0.00093	0.005669	0	0	0	0	0	0	1	-360	360;
	2910	7098	0.00105	0.00506	0	0	0	0	0	0	1	-360	360;
	5688	6231	0.00113	0.006539	0	0	0	0	0	0	1	-360	360;
	5182	5455	0.00037	0.002919	0	0	0	0	0	0	1	-360	360;
	707	5695	0.005169	0.03738	0	476	0	0	0	0	1	-360	360;
	3814	954	0.00125	0.006669	0	457	0	0	0	0	1	-360	360;
	2313	2438	0.001281	0.00661	0	0	0	0	0	0	1	-360	360;
	195	2985	0.00226	0.0138	0	457	0	0	0	0	1	-360	360;
	2518	3513	0.00213	0.02677	0	1613	0	0	0	0	1	-360	360;
	2518	3346	0.00154	0.01919	0	1349	0	0	0	0	1	-360	360;
	4544	601	0.00687	0.04413	0	0	0	0	0	0	1	-360	360;
	4544	3241	0.0045	0.04725	0	457	0	0	0	0	1	-360	360;
	4544	7338	0.00138	0.00837	0	0	0	0	0	0	1	-360	360;
	3069	1538	0.000419	0.001901	0	0	0	0	0	0	1	-360	360;
	3069	1538	0.00043	0.00186	0	0	0	0	0	0	1	-360	360;
	6115	4729	0.003331	0.015099	0	591	0	0	0	0	1	-360	360;
	3344	6616	0.002099	0.01631	0	533	0	0	0	0	1	-360	360;
	3809	5616	0.001331	0.01082	0	591	0	0	0	0	1	-360	360;
	3809	6691	0.00199	0.00818	0	381	0	0	0	0	1	-360	360;
	3809	6691	0.00169	0.008831	0	0	0	0	0	0	1	-360	360;
	1326	892	0.001771	0.014281	0	0	0	0	0	0	1	-360	360;
	601	616	0.003099	0.01844	0	495	0	0	0	0	1	-360	360;
	3496	1551	0.003409	0.01819	0	438	0	0	0	0	1	-360	360;
	4925	2230	0.00057	0.00369	0	0	0	0	0	0	1	-360	360;
	8057	5907	0.001081	0.00468	0	0	0	0	0	0	1	-360	360;
	5354	5907	0.00105	0.004771	0	0	0	0	0	0	1	-360	360;
	3674	616	0.001979	0.01549	0	591	0	0	0	0	1	-360	360;
	3674	7338	0.001469	0.015151	0	0	0	0	0	0	1	-360	360;
	5836	5083	0.000711	0.003969	0	419	0	0	0	0	1	-360	360;
	5836	7576	0.00051	0.00274	0	0	0	0	0	0	1	-360	360;
	5836	3191	0.000521	0.00306	0	495	0	0	0	0	1	-360	360;
	5836	3488	0.00188	0.011841	0	0	0	0	0	0	1	-360	360;
	5836	1578	0.002469	0.010539	0	0	0	0	0	0	1	-360	360;
	575	5300	0.00045	0.00351	0	514	0	0	0	0	1	-360	360;
	3929	9173	0.00037	0.000659	0	0	0	0	0	0	1	-360	360;
	2089	3818	0.000401	0.0007	0	0	0	0	0	0	1	-360	360;
	2089	6232	0.00014	0.000531	0	0	0	0	0	0	1	-360	360;
	2089	1643	0.000289	0.0013	0	495	0	0	0	0	1	-360	360;
	2089	6253	7e-05	0.000289	0	0	0	0	0	0	1	-360	360;
	2089	6684	0.00051	0.002151	0	381	0	0	0	0	1	-360	360;
	2089	8151	0.00037	0.00157	0	533	0	0	0	0	1	-360	360;
	5067	449	0.00022	0.00277	0	1448	0	0	0	0	1	-360	360;
	5067	449	0.00025	0.00271	0	1547	0	0	0	0	1	-360	360;
	7641	3037	0.000271	0.002081	0	572	0	0	0	0	1	-360	360;
	7641	3037	0.00025	0.00231	0	533	0	0	0	0	1	-360	360;
	3191	2340	0.00095	0.00331	0	0	0	0	0	0	1	-360	360;
	3191	8711	0.00093	0.00336	0	0	0	0	0	0	1	-360	360;
	8104	8722	0.000341	0.002211	0	591	0	0	0	0	1	-360	360;
	2644	9191	0.00032	0.00232	0	0	0	0	0	0	1	-360	360;
	4128	8522	3.9e-05	0.000271	0	514	0	0	0	0	1	-360	360;
	604	7522	0.00075	0.0101	0	1613	0	0	0	0	1	-360	360;
	604	4000	0.00118	0.01455	0	1547	0	0	0	0	1	-360	360;
	2155	9021	0.0017	0.0098	0	514	0	0	0	0	1	-360	360;
	2155	9021	0.001521	0.01111	0	0	0	0	0	0	1	-360	360;
	2155	2042	0.000169	0.001401	0	0	0	0	0	0	1	-360	360;
	2155	687	0.002591	0.01725	0	610	0	0	0	0	1	-360	360;
	2155	3520	0.000169	0.001159	0	438	0	0	0	0	1	-360	360;
	2155	4729	0.00307	0.01339	0	533	0	0	0	0	1	-360	360;
	2641	4889	0.002591	0.013521	0	514	0	0	0	0	1	-360	360;
	9067	9051	0.000581	0.005031	0	553	0	0	0	0	1	-360	360;
	6969	5573	0.001419	0.00595	0	381	0	0	0	0	1	-360	360;
	666	1494	0.00205	0.02664	0	1448	0	0	0	0	1	-360	360;
	666	2732	0.00295	0.03721	0	1514	0	0	0	0	1	-360	360;
	666	4550	0.00335	0.03648	0	1481	0	0	0	0	1	-360	360;
	666	1607	0.00022	0.00251	0	1251	0	0	0	0	1	-360	360;
	666	2093	0.00059	0.00651	0	1645	0	0	0	0	1	-360	360;
	3565	2177	0.008539	0.041979	0	381	0	0	0	0	1	-360	360;
	3565	2841	0.001349	0.0052	0	400	0	0	0	0	1	-360	360;
	3611	7972	0.00245	0.013909	0	514	0	0	0	0	1	-360	360;
	3611	2481	0.001469	0.008419	0	553	0	0	0	0	1	-360	360;
	5420	7164	0.00084	0.00836	0	1349	0	0	0	0	1	-360	360;
	7119	2078	0.002591	0.017789	0	514	0	0	0	0	1	-360	360;
	7119	7873	0.009651	0.05968	0	419	0	0	0	0	1	-360	360;
	7422	7519	0.00143	0.00799	0	381	0	0	0	0	1	-360	360;
	7422	6744	0.001969	0.014909	0	572	0	0	0	0	1	-360	360;
	1026	1860	0.01039	0.06406	0	495	0	0	0	0	1	-360	360;
	4765	3327	0.000719	0.005331	0	438	0	0	0	0	1	-360	360;
	6036	6921	0.00052	0.00565	0	1415	0	0	0	0	1	-360	360;
	8670	7069	0.002409	0.019479	0	495	0	0	0	0	1	-360	360;
	8670	7069	0.00256	0.019151	0	419	0	0	0	0	1	-360	360;
	8670	2866	0.00656	0.02288	0	381	0	0	0	0	1	-360	360;
	4032	6450	0.0027	0.014901	0	400	0	0	0	0	1	-360	360;
	4032	5738	0.003591	0.01362	0	400	0	0	0	0	1	-360	360;
	7327	2481	0.002211	0.01188	0	0	0	0	0	0	1	-360	360;
	7327	8189	0.00068	0.00357	0	400	0	0	0	0	1	-360	360;
	7474	1584	0.000211	0.000969	0	400	0	0	0	0	1	-360	360;
	8947	3390	9.9e-05	0.000539	0	0	0	0	0	0	1	-360	360;
	1808	6384	0.00645	0.0263	0	419	0	0	0	0	1	-360	360;
	4624	6031	0.00063	0.001901	0	0	0	0	0	0	1	-360	360;
	1721	3558	0.000281	0.000669	0	0	0	0	0	0	1	-360	360;
	7694	3865	0.000219	0.00064	0	0	0	0	0	0	1	-360	360;
	9181	6723	0.00014	0.00105	0	0	0	0	0	0	1	-360	360;
	8373	8748	0.003219	0.020729	0	533	0	0	0	0	1	-360	360;
	8373	8475	0.00051	0.003151	0	572	0	0	0	0	1	-360	360;
	8373	8475	0.000409	0.00331	0	514	0	0	0	0	1	-360	360;
	5630	6639	0.002289	0.017479	0	591	0	0	0	0	1	-360	360;
	5630	1187	0.002831	0.01895	0	533	0	0	0	0	1	-360	360;
	5630	7274	0.006289	0.036521	0	0	0	0	0	0	1	-360	360;
	5630	8992	0.005651	0.032169	0	438	0	0	0	0	1	-360	360;
	5630	7266	0.00414	0.02262	0	476	0	0	0	0	1	-360	360;
	3492	6734	0.01095	0.05814	0	457	0	0	0	0	1	-360	360;
	3492	1015	0.009	0.0368	0	495	0	0	0	0	1	-360	360;
	3492	8310	0.000229	0.00181	0	553	0	0	0	0	1	-360	360;
	1980	6570	0.00119	0.00726	0	610	0	0	0	0	1	-360	360;
	9137	749	0.011771	0.051669	0	629	0	0	0	0	1	-360	360;
	7522	2183	0.00209	0.02279	0	1711	0	0	0	0	1	-360	360;
	7571	3916	0.002031	0.01443	0	0	0	0	0	0	1	-360	360;
	7571	6891	0.00145	0.008479	0	553	0	0	0	0	1	-360	360;
	2526	6199	0.00124	0.01433	0	1514	0	0	0	0	1	-360	360;
	2526	6199	0.00127	0.01402	0	1876	0	0	0	0	1	-360	360;
	1483	3649	0.013039	0.052841	0	629	0	0	0	0	1	-360	360;
	2365	6199	0.00235	0.02739	0	954	0	0	0	0	1	-360	360;
	2365	6199	0.0024	0.02681	0	1218	0	0	0	0	1	-360	360;
	2365	4141	0.00135	0.02056	0	757	0	0	0	0	1	-360	360;
	2365	1311	0.00087	0.01124	0	954	0	0	0	0	1	-360	360;
	8316	174	0.0012	0.00818	0	629	0	0	0	0	1	-360	360;
	174	6430	0.006961	0.05561	0	495	0	0	0	0	1	-360	360;
	5340	7892	0.00105	0.00763	0	476	0	0	0	0	1	-360	360;
	5340	561	0.000219	0.00182	0	0	0	0	0	0	1	-360	360;
	5340	8843	3.9e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	5340	8843	4.5e-05	0.000195	0	0	0	0	0	0	1	-360	360;
	5340	4239	0.0035	0.022651	0	533	0	0	0	0	1	-360	360;
	3866	1860	0.004281	0.02319	0	457	0	0	0	0	1	-360	360;
	3866	5395	0.003729	0.014021	0	419	0	0	0	0	1	-360	360;
	7042	2177	0.00212	0.01582	0	572	0	0	0	0	1	-360	360;
	8672	7164	0.00056	0.00784	0	757	0	0	0	0	1	-360	360;
	1860	7772	0.004401	0.01813	0	0	0	0	0	0	1	-360	360;
	1860	7772	0.005021	0.02144	0	0	0	0	0	0	1	-360	360;
	4885	2928	0.00018	0.000219	0	381	0	0	0	0	1	-360	360;
	4157	2928	0.000151	0.000211	0	0	0	0	0	0	1	-360	360;
	4432	2928	0.000151	0.000211	0	0	0	0	0	0	1	-360	360;
	2072	5983	0.00282	0.02591	0	1481	0	0	0	0	1	-360	360;
	3450	6552	0.00131	0.008539	0	533	0	0	0	0	1	-360	360;
	3956	7507	0.002159	0.015969	0	0	0	0	0	0	1	-360	360;
	5286	1838	0.00019	0.000281	0	0	0	0	0	0	1	-360	360;
	5571	2230	0.0002	0.000281	0	0	0	0	0	0	1	-360	360;
	2078	8005	0.00139	0.00774	0	572	0	0	0	0	1	-360	360;
	639	6351	0.01101	0.06739	0	629	0	0	0	0	1	-360	360;
	639	2749	0.00063	0.00463	0	438	0	0	0	0	1	-360	360;
	639	9130	0.002341	0.009599	0	476	0	0	0	0	1	-360	360;
	410	4189	0.00375	0.021659	0	476	0	0	0	0	1	-360	360;
	1494	5781	0.00056	0.00749	0	1283	0	0	0	0	1	-360	360;
	583	6616	0.00118	0.00926	0	533	0	0	0	0	1	-360	360;
	583	6031	0.002039	0.013081	0	514	0	0	0	0	1	-360	360;
	1746	7507	0.00155	0.010961	0	0	0	0	0	0	1	-360	360;
	3276	7635	3.9e-05	0.00036	0	419	0	0	0	0	1	-360	360;
	1105	196	0.00024	0.000409	0	381	0	0	0	0	1	-360	360;
	8522	7132	0.000341	0.001659	0	553	0	0	0	0	1	-360	360;
	8522	5351	0.005461	0.024919	0	457	0	0	0	0	1	-360	360;
	8763	7056	0.00158	0.02413	0	954	0	0	0	0	1	-360	360;
	8763	8825	0.00155	0.02458	0	856	0	0	0	0	1	-360	360;
	8763	6889	0.00021	0.00256	0	1415	0	0	0	0	1	-360	360;
	8763	6889	0.00024	0.00243	0	1514	0	0	0	0	1	-360	360;
	8763	8887	0.00144	0.02266	0	1514	0	0	0	0	1	-360	360;
	8763	8887	0.0014	0.02098	0	1547	0	0	0	0	1	-360	360;
	8763	8487	0.00086	0.01074	0	0	0	0	0	0	1	-360	360;
	8748	8475	0.003341	0.02611	0	553	0	0	0	0	1	-360	360;
	217	6785	0.00387	0.02063	0	514	0	0	0	0	1	-360	360;
	217	6785	0.00405	0.018271	0	514	0	0	0	0	1	-360	360;
	217	1179	0.00101	0.00462	0	400	0	0	0	0	1	-360	360;
	217	2575	0.00088	0.00389	0	419	0	0	0	0	1	-360	360;
	217	7274	0.005081	0.039	0	514	0	0	0	0	1	-360	360;
	7972	8189	0.00257	0.0208	0	553	0	0	0	0	1	-360	360;
	7972	8189	0.00251	0.01926	0	514	0	0	0	0	1	-360	360;
	7972	1039	0.002531	0.01314	0	0	0	0	0	0	1	-360	360;
	7253	7840	0.00132	0.00869	0	0	0	0	0	0	1	-360	360;
	4395	8913	0.00082	0.004021	0	419	0	0	0	0	1	-360	360;
	4395	4852	0.00143	0.007479	0	438	0	0	0	0	1	-360	360;
	2795	1183	0.000659	0.005771	0	0	0	0	0	0	1	-360	360;
	2795	1183	0.000591	0.006539	0	0	0	0	0	0	1	-360	360;
	1305	6291	0.00589	0.025599	0	476	0	0	0	0	1	-360	360;
	1305	6982	0.00614	0.02343	0	438	0	0	0	0	1	-360	360;
	2732	1798	0.00155	0.01994	0	1481	0	0	0	0	1	-360	360;
	2732	3346	0.00205	0.03218	0	1580	0	0	0	0	1	-360	360;
	4550	3346	0.00233	0.03154	0	1580	0	0	0	0	1	-360	360;
	4725	5738	0.002341	0.01982	0	0	0	0	0	0	1	-360	360;
	1083	8310	8.1e-05	0.000521	0	533	0	0	0	0	1	-360	360;
	1398	5567	9.9e-05	0.00069	0	591	0	0	0	0	1	-360	360;
	4281	4580	0.00276	0.016419	0	553	0	0	0	0	1	-360	360;
	7050	4908	0.00036	0.001229	0	457	0	0	0	0	1	-360	360;
	7050	3740	0.000229	0.000711	0	514	0	0	0	0	1	-360	360;
	6563	870	0.00013	0.00099	0	514	0	0	0	0	1	-360	360;
	8999	6495	0.001159	0.00689	0	419	0	0	0	0	1	-360	360;
	6888	5994	0.008531	0.03045	0	381	0	0	0	0	1	-360	360;
	5400	8005	0.000781	0.004531	0	514	0	0	0	0	1	-360	360;
	5400	8005	0.00075	0.00474	0	533	0	0	0	0	1	-360	360;
	5881	4031	0.002159	0.008349	0	0	0	0	0	0	1	-360	360;
	8195	8334	0.00162	0.01651	0	1547	0	0	0	0	1	-360	360;
	933	6357	0.01236	0.044211	0	400	0	0	0	0	1	-360	360;
	933	6357	0.00825	0.03989	0	419	0	0	0	0	1	-360	360;
	933	3775	0.02187	0.066961	0	0	0	0	0	0	1	-360	360;
	933	3975	0.01212	0.06661	0	0	0	0	0	0	1	-360	360;
	3918	1642	0.00036	0.00534	0	889	0	0	0	0	1	-360	360;
	6802	1153	0.000229	0.00063	0	0	0	0	0	0	1	-360	360;
	6802	7599	0.000219	0.000581	0	0	0	0	0	0	1	-360	360;
	4025	3994	0.000281	0.00063	0	0	0	0	0	0	1	-360	360;
	1595	1414	0.005021	0.035659	0	419	0	0	0	0	1	-360	360;
	3114	3656	0.00174	0.007901	0	0	0	0	0	0	1	-360	360;
	5237	2229	0.003919	0.020581	0	495	0	0	0	0	1	-360	360;
	5237	9108	0.003341	0.016349	0	495	0	0	0	0	1	-360	360;
	5237	1478	0.003271	0.016669	0	457	0	0	0	0	1	-360	360;
	2799	8515	0.000151	0.000659	0	419	0	0	0	0	1	-360	360;
	6405	7873	0.00736	0.035659	0	0	0	0	0	0	1	-360	360;
	6405	3894	0.00944	0.042169	0	0	0	0	0	0	1	-360	360;
	4435	4734	0.00068	0.00652	0	856	0	0	0	0	1	-360	360;
	4435	3483	0.00188	0.02281	0	1382	0	0	0	0	1	-360	360;
	2079	7052	0.00168	0.00956	0	533	0	0	0	0	1	-360	360;
	2079	7052	0.001669	0.00981	0	514	0	0	0	0	1	-360	360;
	2079	772	0.002229	0.01599	0	591	0	0	0	0	1	-360	360;
	2079	772	0.00218	0.01631	0	610	0	0	0	0	1	-360	360;
	1081	6475	0.00238	0.02124	0	1053	0	0	0	0	1	-360	360;
	1081	6475	0.00194	0.02023	0	1251	0	0	0	0	1	-360	360;
	7361	3166	0.013831	0.05955	0	0	0	0	0	0	1	-360	360;
	7361	426	0.00876	0.04875	0	419	0	0	0	0	1	-360	360;
	2863	1857	0.0002	0.00099	0	400	0	0	0	0	1	-360	360;
	1364	2430	0.00105	0.00826	0	476	0	0	0	0	1	-360	360;
	2177	749	0.0172	0.06637	0	476	0	0	0	0	1	-360	360;
	3594	2850	0.00305	0.015781	0	476	0	0	0	0	1	-360	360;
	6555	3608	0.00208	0.02105	0	1481	0	0	0	0	1	-360	360;
	8475	1965	0.00614	0.043729	0	476	0	0	0	0	1	-360	360;
	8475	5926	0.000229	0.001849	0	553	0	0	0	0	1	-360	360;
	5735	6053	0.00386	0.025789	0	438	0	0	0	0	1	-360	360;
	5735	2341	0.002031	0.012	0	591	0	0	0	0	1	-360	360;
	2629	7507	0.000591	0.004	0	495	0	0	0	0	1	-360	360;
	1137	2406	9.1e-05	0.00051	0	0	0	0	0	0	1	-360	360;
	7577	5814	0.00732	0.039969	0	419	0	0	0	0	1	-360	360;
	7577	3231	0.004581	0.02031	0	629	0	0	0	0	1	-360	360;
	7663	5814	0.00132	0.00537	0	457	0	0	0	0	1	-360	360;
	1183	1159	0.007521	0.0387	0	476	0	0	0	0	1	-360	360;
	6852	5051	0.005159	0.031281	0	533	0	0	0	0	1	-360	360;
	8307	8807	9.9e-05	0.001159	0	0	0	0	0	0	1	-360	360;
	6947	8807	9.9e-05	0.00113	0	0	0	0	0	0	1	-360	360;
	4505	4357	0.00019	0.000909	0	381	0	0	0	0	1	-360	360;
	7144	7791	0.00107	0.00413	0	0	0	0	0	0	1	-360	360;
	6889	2085	9e-05	0.00058	0	1251	0	0	0	0	1	-360	360;
	6889	4024	9e-05	0.00059	0	1152	0	0	0	0	1	-360	360;
	6889	9065	8e-05	0.00053	0	757	0	0	0	0	1	-360	360;
	6889	661	8e-05	0.00052	0	757	0	0	0	0	1	-360	360;
	6612	8864	0.000271	0.00151	0	495	0	0	0	0	1	-360	360;
	8931	6199	0.00053	0.00585	0	1909	0	0	0	0	1	-360	360;
	8931	6199	0.00047	0.00663	0	1448	0	0	0	0	1	-360	360;
	8931	6199	0.00046	0.00676	0	1251	0	0	0	0	1	-360	360;
	8931	6199	0.00045	0.00689	0	1251	0	0	0	0	1	-360	360;
	8931	8494	0.00028	0.00338	0	1580	0	0	0	0	1	-360	360;
	8931	8494	0.0003	0.00319	0	1349	0	0	0	0	1	-360	360;
	1643	4511	0.00025	0.00101	0	476	0	0	0	0	1	-360	360;
	196	1973	0.000281	0.00119	0	0	0	0	0	0	1	-360	360;
	2902	5573	0.00018	0.00063	0	0	0	0	0	0	1	-360	360;
	4950	4410	0.00382	0.02868	0	0	0	0	0	0	1	-360	360;
	4950	333	0.00361	0.0235	0	572	0	0	0	0	1	-360	360;
	4950	6908	0.00538	0.03569	0	0	0	0	0	0	1	-360	360;
	3377	7726	6e-05	0.000409	0	514	0	0	0	0	1	-360	360;
	1027	3243	9e-05	0.00102	0	757	0	0	0	0	1	-360	360;
	1027	2458	0.0014	0.01649	0	1711	0	0	0	0	1	-360	360;
	1027	2919	0.00034	0.00427	0	1119	0	0	0	0	1	-360	360;
	5308	8560	0.0001	0.00109	0	757	0	0	0	0	1	-360	360;
	5308	6427	0.00137	0.01682	0	1777	0	0	0	0	1	-360	360;
	5308	8347	0.00131	0.01262	0	1514	0	0	0	0	1	-360	360;
	7148	5610	0.00262	0.019669	0	533	0	0	0	0	1	-360	360;
	7148	2421	0.00082	0.00526	0	476	0	0	0	0	1	-360	360;
	7148	2421	0.000841	0.005151	0	457	0	0	0	0	1	-360	360;
	7148	2421	0.000669	0.005081	0	476	0	0	0	0	1	-360	360;
	7148	6989	0.001919	0.012669	0	476	0	0	0	0	1	-360	360;
	44	1545	0.002659	0.02	0	0	0	0	0	0	1	-360	360;
	44	3541	0.003031	0.02343	0	572	0	0	0	0	1	-360	360;
	6231	4710	0.01176	0.059159	0	400	0	0	0	0	1	-360	360;
	6231	8367	0.002469	0.017531	0	495	0	0	0	0	1	-360	360;
	6231	2288	0.000669	0.005729	0	0	0	0	0	0	1	-360	360;
	8291	8473	0.000781	0.00876	0	381	0	0	0	0	1	-360	360;
	8291	5455	0.00062	0.00482	0	553	0	0	0	0	1	-360	360;
	8291	7635	0.001349	0.012599	0	533	0	0	0	0	1	-360	360;
	6842	8473	0.000979	0.0082	0	0	0	0	0	0	1	-360	360;
	6842	883	0.008091	0.028831	0	0	0	0	0	0	1	-360	360;
	6842	2288	0.000711	0.005969	0	514	0	0	0	0	1	-360	360;
	3969	7726	0.00145	0.00857	0	0	0	0	0	0	1	-360	360;
	1090	1129	0.00869	0.03768	0	400	0	0	0	0	1	-360	360;
	1090	3070	0.001719	0.00837	0	457	0	0	0	0	1	-360	360;
	8515	5365	0.01582	0.067349	0	438	0	0	0	0	1	-360	360;
	8515	5365	0.01457	0.06469	0	400	0	0	0	0	1	-360	360;
	8515	4529	0.021711	0.090539	0	381	0	0	0	0	1	-360	360;
	8515	3364	0.00019	0.000729	0	400	0	0	0	0	1	-360	360;
	935	5383	0.001271	0.00638	0	400	0	0	0	0	1	-360	360;
	920	6570	3.9e-05	0.0002	0	400	0	0	0	0	1	-360	360;
	6664	4426	0.000281	0.00236	0	629	0	0	0	0	1	-360	360;
	4339	2183	0.00019	0.00329	0	1744	0	0	0	0	1	-360	360;
	7513	4000	0.00035	0.00324	0	0	0	0	0	0	1	-360	360;
	6570	8542	0.00049	0.00349	0	419	0	0	0	0	1	-360	360;
	6570	8542	0.000469	0.00331	0	419	0	0	0	0	1	-360	360;
	6570	1662	0.00313	0.01324	0	0	0	0	0	0	1	-360	360;
	6570	9119	0.001031	0.006719	0	476	0	0	0	0	1	-360	360;
	6570	1151	0.00038	0.00401	0	0	0	0	0	0	1	-360	360;
	502	3589	0.00037	0.002849	0	495	0	0	0	0	1	-360	360;
	8473	1040	0.00032	0.00256	0	476	0	0	0	0	1	-360	360;
	8473	145	0.00037	0.00118	0	381	0	0	0	0	1	-360	360;
	8473	4267	0.00068	0.001151	0	0	0	0	0	0	1	-360	360;
	4936	5179	0.00168	0.009841	0	495	0	0	0	0	1	-360	360;
	5317	8854	0.000599	0.002401	0	400	0	0	0	0	1	-360	360;
	9018	4	0.002169	0.00911	0	400	0	0	0	0	1	-360	360;
	9018	4829	0.002849	0.01282	0	457	0	0	0	0	1	-360	360;
	6253	5383	0.001229	0.01207	0	553	0	0	0	0	1	-360	360;
	6253	7507	0.001289	0.01089	0	610	0	0	0	0	1	-360	360;
	8887	1644	0.00093	0.01036	0	856	0	0	0	0	1	-360	360;
	8887	3483	0.00083	0.01174	0	823	0	0	0	0	1	-360	360;
	6639	7840	0.00118	0.009469	0	0	0	0	0	0	1	-360	360;
	6639	7840	0.001151	0.009651	0	0	0	0	0	0	1	-360	360;
	5461	8109	0.00101	0.01327	0	1580	0	0	0	0	1	-360	360;
	6291	4189	0.00364	0.02486	0	572	0	0	0	0	1	-360	360;
	8405	4747	0.00056	0.00425	0	533	0	0	0	0	1	-360	360;
	8405	4580	0.00369	0.02681	0	533	0	0	0	0	1	-360	360;
	338	7824	0.000531	0.00276	0	400	0	0	0	0	1	-360	360;
	2468	7824	0.000521	0.00281	0	381	0	0	0	0	1	-360	360;
	5257	7824	0.000419	0.00232	0	0	0	0	0	0	1	-360	360;
	2189	7824	0.000409	0.00237	0	0	0	0	0	0	1	-360	360;
	8458	8267	0.00014	0.00193	0	1514	0	0	0	0	1	-360	360;
	2676	5987	0.003211	0.01481	0	419	0	0	0	0	1	-360	360;
	1187	5712	0.00381	0.01175	0	0	0	0	0	0	1	-360	360;
	1187	6521	0.003781	0.02676	0	419	0	0	0	0	1	-360	360;
	2360	8636	0.000919	0.005461	0	553	0	0	0	0	1	-360	360;
	2360	8829	0.000271	0.001211	0	381	0	0	0	0	1	-360	360;
	2360	4235	0.002229	0.017401	0	572	0	0	0	0	1	-360	360;
	2360	4235	0.00243	0.017659	0	610	0	0	0	0	1	-360	360;
	2360	4235	0.002531	0.01705	0	591	0	0	0	0	1	-360	360;
	8818	5099	0.012531	0.05451	0	0	0	0	0	0	1	-360	360;
	5781	4734	0.00087	0.01212	0	1349	0	0	0	0	1	-360	360;
	5781	905	0.0006	0.00802	0	1711	0	0	0	0	1	-360	360;
	5781	8334	0.00072	0.00667	0	1843	0	0	0	0	1	-360	360;
	5781	2093	0.00234	0.02626	0	1415	0	0	0	0	1	-360	360;
	6772	6989	0.00038	0.002969	0	514	0	0	0	0	1	-360	360;
	7885	3912	0.00024	0.00045	0	0	0	0	0	0	1	-360	360;
	6730	3912	0.000229	0.000409	0	0	0	0	0	0	1	-360	360;
	4301	3912	0.000281	0.000401	0	0	0	0	0	0	1	-360	360;
	5362	5610	0.001211	0.005271	0	0	0	0	0	0	1	-360	360;
	5362	350	0.000281	0.00107	0	438	0	0	0	0	1	-360	360;
	5362	6763	0.00025	0.001091	0	0	0	0	0	0	1	-360	360;
	3855	1973	0.00019	0.00081	0	629	0	0	0	0	1	-360	360;
	3855	1973	0.00045	0.00136	0	591	0	0	0	0	1	-360	360;
	3481	554	0.003461	0.023289	0	0	0	0	0	0	1	-360	360;
	5460	8222	0.0013	0.01609	0	1843	0	0	0	0	1	-360	360;
	5460	8222	0.00128	0.0164	0	1711	0	0	0	0	1	-360	360;
	5460	2721	0.00089	0.01082	0	1580	0	0	0	0	1	-360	360;
	5460	2721	0.00087	0.01102	0	1645	0	0	0	0	1	-360	360;
	3786	8578	0.003349	0.02514	0	533	0	0	0	0	1	-360	360;
	490	2653	0.0003	0.00061	0	0	0	0	0	0	1	-360	360;
	148	7069	0.000289	0.001211	0	514	0	0	0	0	1	-360	360;
	7256	8743	0.01301	0.044091	0	438	0	0	0	0	1	-360	360;
	7256	1742	0.02162	0.065969	0	400	0	0	0	0	1	-360	360;
	8930	314	0.006461	0.05275	0	514	0	0	0	0	1	-360	360;
	6954	7974	0.0002	0.000271	0	0	0	0	0	0	1	-360	360;
	3325	7974	0.00018	0.00026	0	0	0	0	0	0	1	-360	360;
	5627	1742	0.00032	0.00188	0	0	0	0	0	0	1	-360	360;
	5610	6989	0.00105	0.006591	0	476	0	0	0	0	1	-360	360;
	21	6163	0.00013	0.00055	0	0	0	0	0	0	1	-360	360;
	3749	5365	0.006151	0.04013	0	629	0	0	0	0	1	-360	360;
	3749	6982	0.007159	0.040591	0	533	0	0	0	0	1	-360	360;
	2291	3133	0.00874	0.048461	0	400	0	0	0	0	1	-360	360;
	2291	4738	0.004159	0.01606	0	0	0	0	0	0	1	-360	360;
	2291	3656	0.002919	0.01912	0	533	0	0	0	0	1	-360	360;
	314	4300	0.00094	0.00394	0	419	0	0	0	0	1	-360	360;
	314	5789	0.000901	0.00395	0	0	0	0	0	0	1	-360	360;
	314	2467	0.003159	0.017789	0	572	0	0	0	0	1	-360	360;
	314	2467	0.002419	0.018331	0	0	0	0	0	0	1	-360	360;
	314	5256	0.00375	0.018281	0	419	0	0	0	0	1	-360	360;
	3200	2653	0.00025	0.001031	0	419	0	0	0	0	1	-360	360;
	4679	4852	0.00264	0.01238	0	419	0	0	0	0	1	-360	360;
	4679	4852	0.001289	0.00795	0	457	0	0	0	0	1	-360	360;
	4679	1151	0.00087	0.003081	0	476	0	0	0	0	1	-360	360;
	8568	5215	0.000401	0.00224	0	419	0	0	0	0	1	-360	360;
	58	221	0.0003	0.00385	0	0	0	0	0	0	1	-360	360;
	6807	851	0.00126	0.008219	0	533	0	0	0	0	1	-360	360;
	150	5099	0.01905	0.055729	0	0	0	0	0	0	1	-360	360;
	150	3558	0.00201	0.00444	0	0	0	0	0	0	1	-360	360;
	7209	2934	0.00101	0.00324	0	0	0	0	0	0	1	-360	360;
	1465	4787	0.00105	0.00677	0	1218	0	0	0	0	1	-360	360;
	1465	3830	0.00039	0.00502	0	1152	0	0	0	0	1	-360	360;
	1465	5648	0.0004	0.00492	0	954	0	0	0	0	1	-360	360;
	1465	4000	0.00051	0.00617	0	1349	0	0	0	0	1	-360	360;
	7537	3485	0.0003	0.00314	0	1580	0	0	0	0	1	-360	360;
	7537	3830	0.00051	0.00668	0	1645	0	0	0	0	1	-360	360;
	7537	5648	0.0005	0.00678	0	1777	0	0	0	0	1	-360	360;
	1262	2424	0.002909	0.017	0	0	0	0	0	0	1	-360	360;
	1262	1156	0.00245	0.011281	0	476	0	0	0	0	1	-360	360;
	1262	4914	0.0027	0.01532	0	457	0	0	0	0	1	-360	360;
	1262	4914	0.00105	0.011151	0	0	0	0	0	0	1	-360	360;
	1262	3834	0.00036	0.002669	0	0	0	0	0	0	1	-360	360;
	1262	556	0.0018	0.011169	0	457	0	0	0	0	1	-360	360;
	1262	3019	0.000711	0.007771	0	0	0	0	0	0	1	-360	360;
	1262	7923	0.00036	0.00257	0	0	0	0	0	0	1	-360	360;
	1551	4103	0.00201	0.01195	0	610	0	0	0	0	1	-360	360;
	1551	7862	0.001091	0.00731	0	572	0	0	0	0	1	-360	360;
	1551	5458	0.00105	0.007591	0	629	0	0	0	0	1	-360	360;
	1551	1838	0.00299	0.02256	0	476	0	0	0	0	1	-360	360;
	3070	2888	0.00036	0.00288	0	0	0	0	0	0	1	-360	360;
	4874	4504	0.00589	0.045599	0	553	0	0	0	0	1	-360	360;
	4874	3579	0.00157	0.0127	0	476	0	0	0	0	1	-360	360;
	4874	3579	0.00168	0.011969	0	0	0	0	0	0	1	-360	360;
	4874	7903	0.00289	0.021031	0	438	0	0	0	0	1	-360	360;
	6368	1102	0.00395	0.025211	0	572	0	0	0	0	1	-360	360;
	6368	9217	0.002771	0.021659	0	476	0	0	0	0	1	-360	360;
	6368	5653	0.002979	0.016901	0	553	0	0	0	0	1	-360	360;
	8478	5907	0.007419	0.034021	0	0	0	0	0	0	1	-360	360;
	5241	6921	0.00056	0.00813	0	889	0	0	0	0	1	-360	360;
	4	1813	0.00361	0.015211	0	438	0	0	0	0	1	-360	360;
	8950	7049	0.000409	0.003331	0	629	0	0	0	0	1	-360	360;
	6846	8788	0.004091	0.019599	0	476	0	0	0	0	1	-360	360;
	6846	6734	0.00294	0.012841	0	495	0	0	0	0	1	-360	360;
	4426	5616	0.020771	0.080479	0	553	0	0	0	0	1	-360	360;
	4426	1965	0.00662	0.0513	0	553	0	0	0	0	1	-360	360;
	4426	1767	0.00607	0.04581	0	533	0	0	0	0	1	-360	360;
	3037	1102	0.000669	0.0042	0	476	0	0	0	0	1	-360	360;
	3037	8485	0.001271	0.007419	0	476	0	0	0	0	1	-360	360;
	3037	5653	0.001599	0.012271	0	572	0	0	0	0	1	-360	360;
	5395	4661	0.009599	0.05024	0	419	0	0	0	0	1	-360	360;
	5395	520	0.00149	0.009219	0	438	0	0	0	0	1	-360	360;
	5395	6816	0.01488	0.06724	0	400	0	0	0	0	1	-360	360;
	5395	5918	0.003599	0.024031	0	419	0	0	0	0	1	-360	360;
	5395	5918	0.00519	0.024	0	438	0	0	0	0	1	-360	360;
	8636	1380	0.003461	0.013349	0	400	0	0	0	0	1	-360	360;
	3436	4239	0.00105	0.01236	0	0	0	0	0	0	1	-360	360;
	3436	4239	0.00199	0.012331	0	495	0	0	0	0	1	-360	360;
	3436	3656	0.00711	0.044711	0	438	0	0	0	0	1	-360	360;
	6203	7124	0.006729	0.02543	0	0	0	0	0	0	1	-360	360;
	9059	7124	0.005771	0.02999	0	0	0	0	0	0	1	-360	360;
	5994	96	0.0048	0.021771	0	0	0	0	0	0	1	-360	360;
	5049	306	0.00286	0.02212	0	533	0	0	0	0	1	-360	360;
	718	8397	0.00351	0.02175	0	514	0	0	0	0	1	-360	360;
	718	6982	0.008401	0.0448	0	572	0	0	0	0	1	-360	360;
	7202	8807	0.0018	0.01187	0	0	0	0	0	0	1	-360	360;
	7202	8807	0.001831	0.01163	0	0	0	0	0	0	1	-360	360;
	8592	2995	0.00075	0.00538	0	0	0	0	0	0	1	-360	360;
	6581	823	2e-05	0.00033	0	0	0	0	0	0	1	-360	360;
	6581	1680	1e-05	0.00034	0	1810	0	0	0	0	1	-360	360;
	6199	953	0.00023	0.00258	0	823	0	0	0	0	1	-360	360;
	6199	1817	0.00023	0.00252	0	757	0	0	0	0	1	-360	360;
	3231	1504	0.002031	0.008531	0	0	0	0	0	0	1	-360	360;
	3231	7226	0.005349	0.020789	0	419	0	0	0	0	1	-360	360;
	2928	7865	0.0002	0.00082	0	495	0	0	0	0	1	-360	360;
	3916	6891	0.002091	0.011219	0	0	0	0	0	0	1	-360	360;
	6053	171	0.00262	0.016979	0	533	0	0	0	0	1	-360	360;
	6053	171	0.002669	0.01663	0	591	0	0	0	0	1	-360	360;
	8829	3044	0.00025	0.00118	0	514	0	0	0	0	1	-360	360;
	1552	1194	0.00056	0.00715	0	1645	0	0	0	0	1	-360	360;
	3912	3740	0.0007	0.00455	0	438	0	0	0	0	1	-360	360;
	3912	2695	0.000331	0.001711	0	381	0	0	0	0	1	-360	360;
	3912	2695	0.00011	0.000781	0	553	0	0	0	0	1	-360	360;
	3912	8877	0.00111	0.00664	0	438	0	0	0	0	1	-360	360;
	7520	8222	0.00066	0.00838	0	1547	0	0	0	0	1	-360	360;
	7520	8222	0.00064	0.00854	0	1547	0	0	0	0	1	-360	360;
	2421	3610	0.003091	0.017091	0	572	0	0	0	0	1	-360	360;
	2421	1414	0.00643	0.03687	0	629	0	0	0	0	1	-360	360;
	8879	3610	0.00275	0.01937	0	0	0	0	0	0	1	-360	360;
	8397	5918	0.00357	0.021289	0	514	0	0	0	0	1	-360	360;
	1914	6982	0.000289	0.002789	0	0	0	0	0	0	1	-360	360;
	7873	6638	0.002531	0.01345	0	381	0	0	0	0	1	-360	360;
	4562	7702	0.00162	0.012159	0	514	0	0	0	0	1	-360	360;
	4562	7702	0.00262	0.01168	0	419	0	0	0	0	1	-360	360;
	8835	1156	0.00099	0.006081	0	0	0	0	0	0	1	-360	360;
	8835	1233	0.000909	0.005979	0	0	0	0	0	0	1	-360	360;
	3390	408	0.00011	0.0008	0	0	0	0	0	0	1	-360	360;
	3390	7231	0.00011	0.00068	0	0	0	0	0	0	1	-360	360;
	3390	5993	0.00207	0.008341	0	495	0	0	0	0	1	-360	360;
	1910	2550	0.00019	0.001031	0	0	0	0	0	0	1	-360	360;
	3975	7353	0.014039	0.0667	0	0	0	0	0	0	1	-360	360;
	3126	7495	0.004961	0.03043	0	438	0	0	0	0	1	-360	360;
	2689	5753	0.00056	0.003531	0	0	0	0	0	0	1	-360	360;
	1201	4235	0.001461	0.00889	0	514	0	0	0	0	1	-360	360;
	7626	7076	6e-05	0.000419	0	495	0	0	0	0	1	-360	360;
	4823	4060	0.00139	0.005169	0	0	0	0	0	0	1	-360	360;
	4141	1311	0.00069	0.00825	0	757	0	0	0	0	1	-360	360;
	7124	4566	0.004969	0.030781	0	0	0	0	0	0	1	-360	360;
	7124	8873	0.00012	0.000969	0	591	0	0	0	0	1	-360	360;
	6684	5383	0.001091	0.010159	0	572	0	0	0	0	1	-360	360;
	8151	5383	0.000969	0.00887	0	553	0	0	0	0	1	-360	360;
	4357	6619	0.00099	0.006789	0	514	0	0	0	0	1	-360	360;
	8191	8765	0.001031	0.00586	0	381	0	0	0	0	1	-360	360;
	7466	5469	0.00401	0.03345	0	629	0	0	0	0	1	-360	360;
	3401	5351	9.1e-05	0.000349	0	0	0	0	0	0	1	-360	360;
	3760	225	0.00125	0.00745	0	591	0	0	0	0	1	-360	360;
	53	4031	0.003711	0.024219	0	514	0	0	0	0	1	-360	360;
	8222	6921	0.00169	0.02156	0	1678	0	0	0	0	1	-360	360;
	8222	6921	0.00171	0.02114	0	1777	0	0	0	0	1	-360	360;
	6556	4239	0.00088	0.00562	0	533	0	0	0	0	1	-360	360;
	6556	4239	0.0008	0.00618	0	572	0	0	0	0	1	-360	360;
	2940	7178	0.00161	0.01014	0	0	0	0	0	0	1	-360	360;
	3928	2563	0.00032	0.002031	0	533	0	0	0	0	1	-360	360;
	3928	6521	0.00011	0.000659	0	495	0	0	0	0	1	-360	360;
	1923	3830	0.00042	0.00505	0	1086	0	0	0	0	1	-360	360;
	1923	5648	0.00043	0.00494	0	1251	0	0	0	0	1	-360	360;
	2815	4914	0.00038	0.000711	0	0	0	0	0	0	1	-360	360;
	2815	4914	0.00037	0.000659	0	0	0	0	0	0	1	-360	360;
	4125	1644	0.00248	0.03096	0	1843	0	0	0	0	1	-360	360;
	163	3498	0.00375	0.019729	0	495	0	0	0	0	1	-360	360;
	163	2430	0.00518	0.020021	0	438	0	0	0	0	1	-360	360;
	7943	9108	0.000771	0.003711	0	495	0	0	0	0	1	-360	360;
	7943	1478	0.00076	0.003789	0	457	0	0	0	0	1	-360	360;
	2308	3541	0.000781	0.00574	0	0	0	0	0	0	1	-360	360;
	3535	7582	0.00089	0.00562	0	0	0	0	0	0	1	-360	360;
	2019	8542	0.000331	0.001401	0	0	0	0	0	0	1	-360	360;
	5174	8542	0.000401	0.00137	0	0	0	0	0	0	1	-360	360;
	7945	5215	0.00051	0.003349	0	0	0	0	0	0	1	-360	360;
	2057	9130	0.000461	0.00168	0	0	0	0	0	0	1	-360	360;
	2057	9130	0.000469	0.00164	0	0	0	0	0	0	1	-360	360;
	4661	6816	0.006081	0.03114	0	457	0	0	0	0	1	-360	360;
	7519	776	0.005831	0.02218	0	495	0	0	0	0	1	-360	360;
	8989	3435	0.00013	0.000461	0	0	0	0	0	0	1	-360	360;
	7797	2020	0.00013	0.000461	0	0	0	0	0	0	1	-360	360;
	871	4031	0.00262	0.020289	0	514	0	0	0	0	1	-360	360;
	871	2878	0.00164	0.01194	0	553	0	0	0	0	1	-360	360;
	7353	9164	0.006479	0.02845	0	514	0	0	0	0	1	-360	360;
	7974	2012	0.00294	0.014419	0	495	0	0	0	0	1	-360	360;
	7974	2012	0.00238	0.01463	0	495	0	0	0	0	1	-360	360;
	7974	1750	0.002219	0.00993	0	476	0	0	0	0	1	-360	360;
	1401	2303	0.008651	0.04886	0	400	0	0	0	0	1	-360	360;
	4504	7903	0.00464	0.030979	0	0	0	0	0	0	1	-360	360;
	5837	449	0.00092	0.00889	0	1547	0	0	0	0	1	-360	360;
	5837	449	0.00093	0.00869	0	1448	0	0	0	0	1	-360	360;
	5837	1817	0.0022	0.03485	0	1678	0	0	0	0	1	-360	360;
	5837	2458	0.00074	0.00941	0	1843	0	0	0	0	1	-360	360;
	5837	6427	0.00087	0.00884	0	1678	0	0	0	0	1	-360	360;
	5837	1730	0.00139	0.01662	0	0	0	0	0	0	1	-360	360;
	9217	594	0.002781	0.022979	0	400	0	0	0	0	1	-360	360;
	6351	5393	0.000521	0.00407	0	629	0	0	0	0	1	-360	360;
	6194	7702	0.008479	0.032841	0	0	0	0	0	0	1	-360	360;
	2503	7491	0.005669	0.02901	0	0	0	0	0	0	1	-360	360;
	4031	2872	0.001031	0.004099	0	0	0	0	0	0	1	-360	360;
	4031	8043	0.00301	0.018719	0	476	0	0	0	0	1	-360	360;
	1813	2866	0.00257	0.01713	0	438	0	0	0	0	1	-360	360;
	5365	5417	0.010461	0.03836	0	0	0	0	0	0	1	-360	360;
	5365	4100	0.001719	0.011531	0	514	0	0	0	0	1	-360	360;
	5365	1556	9.1e-05	0.00056	0	400	0	0	0	0	1	-360	360;
	3894	3865	0.00476	0.025289	0	0	0	0	0	0	1	-360	360;
	128	1554	0.000219	0.00164	0	419	0	0	0	0	1	-360	360;
	198	7994	0.002159	0.01163	0	0	0	0	0	0	1	-360	360;
	198	1554	0.00295	0.01539	0	400	0	0	0	0	1	-360	360;
	2985	3133	0.007789	0.033039	0	400	0	0	0	0	1	-360	360;
	7316	3071	0.00026	0.0013	0	0	0	0	0	0	1	-360	360;
	6486	8992	0.00162	0.013229	0	438	0	0	0	0	1	-360	360;
	8485	594	0.007919	0.04194	0	438	0	0	0	0	1	-360	360;
	682	8477	0.001229	0.00401	0	400	0	0	0	0	1	-360	360;
	5146	5488	0.008021	0.042831	0	438	0	0	0	0	1	-360	360;
	6638	776	0.007711	0.03006	0	438	0	0	0	0	1	-360	360;
	8788	789	0.00261	0.015469	0	419	0	0	0	0	1	-360	360;
	6332	1556	0.00139	0.00536	0	0	0	0	0	0	1	-360	360;
	6332	1556	0.001419	0.00525	0	0	0	0	0	0	1	-360	360;
	1883	7702	0.006289	0.032271	0	0	0	0	0	0	1	-360	360;
	1883	2132	0.011031	0.05625	0	0	0	0	0	0	1	-360	360;
	171	2341	0.001841	0.01324	0	572	0	0	0	0	1	-360	360;
	6630	10	0.000599	0.00186	0	553	0	0	0	0	1	-360	360;
	4484	4239	0.00036	0.001591	0	533	0	0	0	0	1	-360	360;
	3044	4235	0.003849	0.027651	0	553	0	0	0	0	1	-360	360;
	851	4454	0.00186	0.00874	0	591	0	0	0	0	1	-360	360;
	3485	2183	0.00025	0.00285	0	1580	0	0	0	0	1	-360	360;
	1156	7923	0.00207	0.008841	0	419	0	0	0	0	1	-360	360;
	1156	1233	6e-05	0.00039	0	553	0	0	0	0	1	-360	360;
	8913	4852	0.002159	0.008909	0	457	0	0	0	0	1	-360	360;
	4039	1233	0.000219	0.00039	0	0	0	0	0	0	1	-360	360;
	4914	2653	0.00011	0.000521	0	400	0	0	0	0	1	-360	360;
	4914	2653	0.00011	0.00049	0	400	0	0	0	0	1	-360	360;
	4914	556	0.000771	0.003789	0	438	0	0	0	0	1	-360	360;
	4914	3019	0.000229	0.002849	0	0	0	0	0	0	1	-360	360;
	7994	5256	0.00445	0.027599	0	0	0	0	0	0	1	-360	360;
	1709	3346	0.00218	0.02	0	1481	0	0	0	0	1	-360	360;
	1709	8468	0.00074	0.00865	0	1218	0	0	0	0	1	-360	360;
	7702	1234	0.002531	0.019419	0	400	0	0	0	0	1	-360	360;
	7702	8035	0.003	0.01631	0	0	0	0	0	0	1	-360	360;
	8677	3331	0.008461	0.025081	0	0	0	0	0	0	1	-360	360;
	8189	3643	0.00382	0.013669	0	495	0	0	0	0	1	-360	360;
	3133	8477	0.002581	0.01814	0	514	0	0	0	0	1	-360	360;
	6744	6101	0.00201	0.012219	0	533	0	0	0	0	1	-360	360;
	8542	4313	0.00087	0.001599	0	0	0	0	0	0	1	-360	360;
	769	5233	0.00455	0.02394	0	419	0	0	0	0	1	-360	360;
	1965	5926	0.005651	0.041771	0	457	0	0	0	0	1	-360	360;
	9014	3999	0.00069	0.001271	0	0	0	0	0	0	1	-360	360;
	1917	8406	0.00068	0.00269	0	400	0	0	0	0	1	-360	360;
	1917	2021	0.00036	0.00261	0	476	0	0	0	0	1	-360	360;
	1917	8293	0.001669	0.012031	0	419	0	0	0	0	1	-360	360;
	4738	3656	0.009669	0.03513	0	0	0	0	0	0	1	-360	360;
	4738	8477	0.001099	0.00862	0	0	0	0	0	0	1	-360	360;
	7998	1644	0.0005	0.00711	0	757	0	0	0	0	1	-360	360;
	6734	1015	0.003919	0.017781	0	495	0	0	0	0	1	-360	360;
	6734	8992	0.00161	0.019781	0	0	0	0	0	0	1	-360	360;
	6734	789	0.0007	0.01256	0	610	0	0	0	0	1	-360	360;
	5004	5393	0.000409	0.0018	0	0	0	0	0	0	1	-360	360;
	1708	3906	2e-05	0.00043	0	0	0	0	0	0	1	-360	360;
	9150	3906	3e-05	0.0004	0	1810	0	0	0	0	1	-360	360;
	280	9158	0.00968	0.056659	0	457	0	0	0	0	1	-360	360;
	8043	3656	0.003969	0.02925	0	0	0	0	0	0	1	-360	360;
	5712	8992	0.00219	0.01468	0	514	0	0	0	0	1	-360	360;
	5712	8466	0.000401	0.003159	0	476	0	0	0	0	1	-360	360;
	5712	6880	0.000531	0.003	0	476	0	0	0	0	1	-360	360;
	5215	7266	0.00136	0.00576	0	476	0	0	0	0	1	-360	360;
	8807	8992	0.000849	0.006289	0	457	0	0	0	0	1	-360	360;
	8807	8992	0.000841	0.006409	0	457	0	0	0	0	1	-360	360;
	8807	8992	0.00045	0.00506	0	0	0	0	0	0	1	-360	360;
	218	2088	0.003669	0.02988	0	495	0	0	0	0	1	-360	360;
	520	5918	0.002	0.01568	0	514	0	0	0	0	1	-360	360;
	2653	2128	0.00031	0.0008	0	419	0	0	0	0	1	-360	360;
	1216	5233	0.00043	0.00038	0	0	0	0	0	0	1	-360	360;
	8980	5993	0.00043	0.002229	0	0	0	0	0	0	1	-360	360;
	5212	5993	0.00049	0.00218	0	0	0	0	0	0	1	-360	360;
	3656	4060	0.00282	0.01844	0	629	0	0	0	0	1	-360	360;
	883	3331	0.00462	0.014849	0	0	0	0	0	0	1	-360	360;
	5144	3346	0.00263	0.04076	0	1547	0	0	0	0	1	-360	360;
	3498	2563	0.01062	0.056719	0	457	0	0	0	0	1	-360	360;
	1896	8200	0.0001	0.00105	0	757	0	0	0	0	1	-360	360;
	9130	8864	0.00239	0.017341	0	476	0	0	0	0	1	-360	360;
	3503	2510	0.001341	0.010581	0	533	0	0	0	0	1	-360	360;
	3503	6146	0.00101	0.007599	0	572	0	0	0	0	1	-360	360;
	401	1584	0.00011	0.000289	0	495	0	0	0	0	1	-360	360;
	7328	6921	0.00146	0.01769	0	0	0	0	0	0	1	-360	360;
	7328	6921	0.0013	0.02004	0	1744	0	0	0	0	1	-360	360;
	6816	5573	0.007909	0.028349	0	419	0	0	0	0	1	-360	360;
	7396	8564	0.00012	0.00049	0	0	0	0	0	0	1	-360	360;
	7396	8564	0.00012	0.000479	0	0	0	0	0	0	1	-360	360;
	6921	432	0.00076	0.01018	0	1613	0	0	0	0	1	-360	360;
	6921	432	0.0009	0.00865	0	1777	0	0	0	0	1	-360	360;
	6921	432	0.00096	0.01294	0	1645	0	0	0	0	1	-360	360;
	6921	432	0.0004	0.00756	0	1711	0	0	0	0	1	-360	360;
	6921	432	0.00039	0.00771	0	1547	0	0	0	0	1	-360	360;
	6921	8200	0.00163	0.02291	0	1645	0	0	0	0	1	-360	360;
	2949	3121	0.00636	0.03855	0	495	0	0	0	0	1	-360	360;
	2949	3121	0.00639	0.03863	0	514	0	0	0	0	1	-360	360;
	953	1730	0.00113	0.01362	0	790	0	0	0	0	1	-360	360;
	1767	892	0.00861	0.05318	0	629	0	0	0	0	1	-360	360;
	4674	4889	0.0012	0.004669	0	438	0	0	0	0	1	-360	360;
	4674	7076	0.00018	0.001091	0	514	0	0	0	0	1	-360	360;
	1234	8035	0.001401	0.00612	0	457	0	0	0	0	1	-360	360;
	115	225	0.00068	0.00387	0	591	0	0	0	0	1	-360	360;
	5720	8578	0.0043	0.033591	0	514	0	0	0	0	1	-360	360;
	3579	5469	0.003039	0.01911	0	591	0	0	0	0	1	-360	360;
	3246	124	6e-05	0.00111	0	953	0	0	0	0	1	-360	360;
	8846	4783	5e-05	0.001091	0	953	0	0	0	0	1	-360	360;
	2393	3306	3.9e-05	0.00089	0	953	0	0	0	0	1	-360	360;
	1895	4480	3.9e-05	0.000901	0	953	0	0	0	0	1	-360	360;
	678	4454	0.000169	0.00111	0	610	0	0	0	0	1	-360	360;
	1002	4852	0.000331	0.0027	0	0	0	0	0	0	1	-360	360;
	6429	2012	0.00026	0.00369	0	0	0	0	0	0	1	-360	360;
	4889	7076	0.000341	0.00274	0	0	0	0	0	0	1	-360	360;
	5341	7491	0.00038	0.00506	0	514	0	0	0	0	1	-360	360;
	2563	6521	0.00012	0.000831	0	476	0	0	0	0	1	-360	360;
	1111	594	0.003031	0.0123	0	0	0	0	0	0	1	-360	360;
	4747	7791	0.001521	0.00774	0	495	0	0	0	0	1	-360	360;
	3906	3022	0.00059	0.00552	0	1711	0	0	0	0	1	-360	360;
	3906	1194	0.00053	0.00625	0	1645	0	0	0	0	1	-360	360;
	5993	8804	0.00143	0.01076	0	629	0	0	0	0	1	-360	360;
	1798	8487	0.00056	0.00593	0	0	0	0	0	0	1	-360	360;
	2467	7178	0.00205	0.01461	0	514	0	0	0	0	1	-360	360;
	2132	2303	0.01068	0.054521	0	0	0	0	0	0	1	-360	360;
	8992	8466	0.00194	0.011039	0	533	0	0	0	0	1	-360	360;
	8992	6880	0.00131	0.010099	0	533	0	0	0	0	1	-360	360;
	8992	2995	0.001289	0.014229	0	419	0	0	0	0	1	-360	360;
	805	3221	0.00455	0.027031	0	495	0	0	0	0	1	-360	360;
	805	3327	0.00239	0.01414	0	610	0	0	0	0	1	-360	360;
	805	8310	0.004039	0.029669	0	591	0	0	0	0	1	-360	360;
	805	2088	0.001099	0.007771	0	629	0	0	0	0	1	-360	360;
	5383	2021	0.00082	0.00762	0	514	0	0	0	0	1	-360	360;
	5383	4114	0.000719	0.0033	0	0	0	0	0	0	1	-360	360;
	5383	5519	0.000599	0.00361	0	400	0	0	0	0	1	-360	360;
	5383	8293	0.000521	0.004039	0	476	0	0	0	0	1	-360	360;
	5383	3435	0.000901	0.0088	0	533	0	0	0	0	1	-360	360;
	7507	6990	0.000781	0.00306	0	0	0	0	0	0	1	-360	360;
	7507	2406	0.000419	0.00399	0	0	0	0	0	0	1	-360	360;
	7507	2020	0.000841	0.009841	0	381	0	0	0	0	1	-360	360;
	6478	6146	0.004021	0.03064	0	514	0	0	0	0	1	-360	360;
	4852	2695	0.0023	0.009031	0	438	0	0	0	0	1	-360	360;
	4852	2695	0.002081	0.009469	0	476	0	0	0	0	1	-360	360;
	4852	5502	0.00164	0.008831	0	0	0	0	0	0	1	-360	360;
	4852	8877	0.00094	0.00488	0	419	0	0	0	0	1	-360	360;
	4852	9119	0.000901	0.00551	0	553	0	0	0	0	1	-360	360;
	4852	1151	0.00051	0.005961	0	0	0	0	0	0	1	-360	360;
	2012	1750	0.00113	0.004349	0	533	0	0	0	0	1	-360	360;
	1478	6616	0.002081	0.013031	0	495	0	0	0	0	1	-360	360;
	809	4594	0.00116	0.01438	0	0	0	0	0	0	1	-360	360;
	8656	4729	0.00013	0.001031	0	0	0	0	0	0	1	-360	360;
	4100	615	0.00036	0.002659	0	514	0	0	0	0	1	-360	360;
	4060	5393	0.000771	0.004539	0	629	0	0	0	0	1	-360	360;
	6000	1689	0.00017	0.01734	0	1185	0	0	0	0	1	-360	360;
	4192	8230	0.00017	0.01767	0	1086	0	0	0	0	1	-360	360;
	8814	2030	0.0006	0.040218	0	637	0	0	0	0	1	-360	360;
	8832	2180	0.00076	0.002409	0	0	0	0	0	0	1	-360	360;
	8832	3771	0	0.000538	0	0	0	0	0	0	1	-360	360;
	6395	1564	0.00528	0.027569	0	0	0	0	0	0	1	-360	360;
	6395	1564	0.005382	0.026978	0	0	0	0	0	0	1	-360	360;
	5255	6395	0.00104	0.009649	0	0	0	0	0	0	1	-360	360;
	5255	6395	0.001062	0.009449	0	273	0	0	0	0	1	-360	360;
	3397	5247	0.000289	0.000889	0	0	0	0	0	0	1	-360	360;
	1207	5247	0.003502	0.01524	0	0	0	0	0	0	1	-360	360;
	3397	7958	0.000289	0.00088	0	0	0	0	0	0	1	-360	360;
	1207	7958	0.003431	0.014111	0	0	0	0	0	0	1	-360	360;
	8389	5526	0.001502	0.1272	0	0	0	0	0	0	1	-360	360;
	91	722	8e-05	0.000729	0	312	0	0	0	0	1	-360	360;
	1207	722	0.003818	0.012911	0	0	0	0	0	0	1	-360	360;
	9083	722	0.001618	0.013711	0	0	0	0	0	0	1	-360	360;
	3363	1870	0.001062	0.007791	0	0	0	0	0	0	1	-360	360;
	6539	1870	0.001809	0.008951	0	0	0	0	0	0	1	-360	360;
	5792	1870	0.007142	0.020311	0	0	0	0	0	0	1	-360	360;
	3363	8356	0.000942	0.008831	0	0	0	0	0	0	1	-360	360;
	6539	8356	0.001769	0.008289	0	0	0	0	0	0	1	-360	360;
	5792	8356	0.00728	0.019898	0	0	0	0	0	0	1	-360	360;
	8828	5740	0.003569	0.017951	0	0	0	0	0	0	1	-360	360;
	3443	8828	0.001991	0.007271	0	0	0	0	0	0	1	-360	360;
	7701	8828	0.000391	0.001462	0	0	0	0	0	0	1	-360	360;
	2619	8167	0.003502	0.018289	0	0	0	0	0	0	1	-360	360;
	244	2619	0.001769	0.00824	0	0	0	0	0	0	1	-360	360;
	7701	2619	0.0004	0.001431	0	0	0	0	0	0	1	-360	360;
	1810	4740	0	0.000413	0	0	0	0	0	0	1	-360	360;
	5279	7525	0.001471	0.1296	0	0	0	0	0	0	1	-360	360;
	2180	7374	0.00056	0.015018	0	0	0	0	0	0	1	-360	360;
	164	8490	0.000631	0.01472	0	0	0	0	0	0	1	-360	360;
	164	2180	0	0.001169	0	0	0	0	0	0	1	-360	360;
	164	7596	0.000942	0.003342	0	0	0	0	0	0	1	-360	360;
	164	2708	0.001582	0.002831	0	0	0	0	0	0	1	-360	360;
	3119	7059	0.0008	0.003422	0	0	0	0	0	0	1	-360	360;
	3119	7059	0.000822	0.003351	0	0	0	0	0	0	1	-360	360;
	7596	2708	0.002631	0.004889	0	0	0	0	0	0	1	-360	360;
	7596	2305	0.005418	0.022489	0	0	0	0	0	0	1	-360	360;
	7596	2305	0.00552	0.022018	0	0	0	0	0	0	1	-360	360;
	7394	7575	0.004091	0.020893	0	0	0	0	0	0	1	-360	360;
	7394	7575	0.004	0.019339	0	0	0	0	0	0	1	-360	360;
	7394	4415	0.016479	0.037058	0	0	0	0	0	0	1	-360	360;
	7394	4686	0.016347	0.05124	0	0	0	0	0	0	1	-360	360;
	5783	7575	0.007058	0.030587	0	0	0	0	0	0	1	-360	360;
	5783	7575	0.007149	0.029934	0	0	0	0	0	0	1	-360	360;
	6823	6202	0.002089	0.003449	0	0	0	0	0	0	1	-360	360;
	6823	6202	0.002129	0.003378	0	0	0	0	0	0	1	-360	360;
	2330	712	0	0.00028	0	0	0	0	0	0	1	-360	360;
	712	4272	0.007489	0.0516	0	0	0	0	0	0	1	-360	360;
	712	4272	0.008489	0.050569	0	0	0	0	0	0	1	-360	360;
	878	3206	0	0.047951	0	0	0	0	0	0	1	-360	360;
	7090	3206	0.001649	0.016591	0	0	0	0	0	0	1	-360	360;
	7844	3206	0.001618	0.016898	0	0	0	0	0	0	1	-360	360;
	3497	6202	0.006871	0.039911	0	0	0	0	0	0	1	-360	360;
	3497	6202	0.007009	0.039089	0	0	0	0	0	0	1	-360	360;
	3497	3182	0.001658	0.011871	0	0	0	0	0	0	1	-360	360;
	3497	3182	0.001631	0.010991	0	0	0	0	0	0	1	-360	360;
	5426	9120	0.00092	0.004689	0	0	0	0	0	0	1	-360	360;
	133	9120	0.002049	0.0038	0	0	0	0	0	0	1	-360	360;
	6323	9120	0.005649	0.026	0	0	0	0	0	0	1	-360	360;
	5426	3047	0.000902	0.004782	0	0	0	0	0	0	1	-360	360;
	275	3047	0.003471	0.014471	0	0	0	0	0	0	1	-360	360;
	133	3047	0.002089	0.00372	0	0	0	0	0	0	1	-360	360;
	4188	1880	0.003569	0.010671	0	0	0	0	0	0	1	-360	360;
	3898	4188	0.019782	0.061658	0	0	0	0	0	0	1	-360	360;
	486	5740	0.000409	0.0014	0	0	0	0	0	0	1	-360	360;
	5740	539	0.003302	0.017311	0	0	0	0	0	0	1	-360	360;
	486	8167	0.000409	0.001369	0	0	0	0	0	0	1	-360	360;
	8167	539	0.003231	0.016031	0	0	0	0	0	0	1	-360	360;
	9094	8463	0.000751	0.005711	0	0	0	0	0	0	1	-360	360;
	9094	8994	0.0018	0.00716	0	299	0	0	0	0	1	-360	360;
	6354	9094	0.001978	0.013462	0	0	0	0	0	0	1	-360	360;
	9094	1936	0.001502	0.0056	0	0	0	0	0	0	1	-360	360;
	3280	6423	0.005991	0.0284	0	0	0	0	0	0	1	-360	360;
	2055	6423	0.000302	0.000858	0	0	0	0	0	0	1	-360	360;
	6275	6423	0.002622	0.013231	0	0	0	0	0	0	1	-360	360;
	7138	5618	0.004271	0.019502	0	0	0	0	0	0	1	-360	360;
	2055	5618	0.000271	0.000969	0	0	0	0	0	0	1	-360	360;
	6275	5618	0.00356	0.014378	0	0	0	0	0	0	1	-360	360;
	6288	3456	0.0022	0.009569	0	0	0	0	0	0	1	-360	360;
	2745	3456	0.000311	0.001169	0	0	0	0	0	0	1	-360	360;
	3301	3456	0.001889	0.006311	0	0	0	0	0	0	1	-360	360;
	2745	3280	0.000271	0.00132	0	0	0	0	0	0	1	-360	360;
	3301	3280	0.001929	0.006178	0	260	0	0	0	0	1	-360	360;
	790	6600	0.000339	0.001149	0	0	0	0	0	0	1	-360	360;
	472	8889	0.002591	0.014418	0	0	0	0	0	0	1	-360	360;
	472	775	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6356	3443	0.0014	0.00652	0	0	0	0	0	0	1	-360	360;
	8164	3443	0.000418	0.001338	0	0	0	0	0	0	1	-360	360;
	6356	244	0.001369	0.006649	0	0	0	0	0	0	1	-360	360;
	8164	244	0.000369	0.00152	0	0	0	0	0	0	1	-360	360;
	8814	5320	0.004689	0.018151	0	0	0	0	0	0	1	-360	360;
	8814	5320	0.004591	0.018498	0	0	0	0	0	0	1	-360	360;
	1255	143	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	1255	3974	0.003698	0.0198	0	0	0	0	0	0	1	-360	360;
	8016	1255	0.005471	0.0206	0	0	0	0	0	0	1	-360	360;
	8016	1255	0.0062	0.020191	0	0	0	0	0	0	1	-360	360;
	1255	4362	0.006098	0.030551	0	0	0	0	0	0	1	-360	360;
	950	5792	0.004809	0.019751	0	0	0	0	0	0	1	-360	360;
	950	5792	0.004711	0.018289	0	260	0	0	0	0	1	-360	360;
	3397	6045	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6288	7138	0.007929	0.044022	0	0	0	0	0	0	1	-360	360;
	6057	1352	0.009782	0.033938	0	0	0	0	0	0	1	-360	360;
	6057	6559	0.003609	0.016889	0	0	0	0	0	0	1	-360	360;
	656	1392	0.00044	0.027591	0	0	0	0	0	0	1	-360	360;
	6835	3538	0.000419	0.027979	0	0	0	0	0	0	1	-360	360;
	351	3538	0.000409	0.02851	0	0	0	0	0	0	1	-360	360;
	1392	3538	0	0.000461	0	0	0	0	0	0	1	-360	360;
	1392	2824	0.000531	0.005581	0	0	0	0	0	0	1	-360	360;
	1052	1392	0.003841	0.022461	0	476	0	0	0	0	1	-360	360;
	3538	140	0.000539	0.005461	0	400	0	0	0	0	1	-360	360;
	3878	3538	0.00376	0.022901	0	457	0	0	0	0	1	-360	360;
	5001	8862	0.028843	0.060992	0	0	0	0	0	0	1	-360	360;
	5001	8862	0.029397	0.059744	0	0	0	0	0	0	1	-360	360;
	1810	6342	0.001727	0.00924	0	0	0	0	0	0	1	-360	360;
	1777	1810	0.004851	0.02724	0	0	0	0	0	0	1	-360	360;
	1352	6559	0.0138	0.040249	0	260	0	0	0	0	1	-360	360;
	1352	643	0.007098	0.022329	0	0	0	0	0	0	1	-360	360;
	1352	1836	0.001902	0.008391	0	0	0	0	0	0	1	-360	360;
	1352	1836	0.001929	0.008209	0	0	0	0	0	0	1	-360	360;
	1670	1352	0.001649	0.01228	0	0	0	0	0	0	1	-360	360;
	1670	1352	0.001471	0.01392	0	0	0	0	0	0	1	-360	360;
	1352	7628	0.000582	0.040991	0	299	0	0	0	0	1	-360	360;
	1352	311	0.01116	0.033671	0	286	0	0	0	0	1	-360	360;
	6998	2482	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5008	8249	0.00031	0.00312	0	1251	0	0	0	0	1	-360	360;
	6068	8249	0.00069	0.00769	0	1810	0	0	0	0	1	-360	360;
	6139	5450	0.010471	0.029289	0	0	0	0	0	0	1	-360	360;
	6139	5450	0.010671	0.028671	0	0	0	0	0	0	1	-360	360;
	344	6139	0.00624	0.021871	0	299	0	0	0	0	1	-360	360;
	344	6139	0.007071	0.021431	0	299	0	0	0	0	1	-360	360;
	8881	7751	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	8881	6559	0.001062	0.003351	0	0	0	0	0	0	1	-360	360;
	8881	5342	0.002111	0.02008	0	0	0	0	0	0	1	-360	360;
	8881	6399	0.001009	0.010929	0	0	0	0	0	0	1	-360	360;
	1670	8881	0.004711	0.050609	0	0	0	0	0	0	1	-360	360;
	8881	9083	0.011462	0.02396	0	0	0	0	0	0	1	-360	360;
	5279	2350	0	0.0444	0	0	0	0	0	0	1	-360	360;
	5279	1123	0.001698	0.1128	0	0	0	0	0	0	1	-360	360;
	5279	1633	8.9e-05	0.000742	0	0	0	0	0	0	1	-360	360;
	5279	1633	8.9e-05	0.000729	0	0	0	0	0	0	1	-360	360;
	5279	2812	0.000302	0.000809	0	0	0	0	0	0	1	-360	360;
	5279	9029	0.001502	0.002182	0	0	0	0	0	0	1	-360	360;
	5279	6149	0.001311	0.001871	0	0	0	0	0	0	1	-360	360;
	5279	6149	0.001369	0.00204	0	0	0	0	0	0	1	-360	360;
	5279	4616	0.001049	0.00308	0	273	0	0	0	0	1	-360	360;
	1777	5893	0.014967	0.09105	0	0	0	0	0	0	1	-360	360;
	5296	5893	0.028959	0.12905	0	0	0	0	0	0	1	-360	360;
	2478	5893	0.006438	0.041256	0	0	0	0	0	0	1	-360	360;
	2478	5893	0.006554	0.040364	0	0	0	0	0	0	1	-360	360;
	2350	1260	0	0.012742	0	0	0	0	0	0	1	-360	360;
	6354	8463	0.001018	0.00868	0	0	0	0	0	0	1	-360	360;
	6354	8994	0.001871	0.00548	0	0	0	0	0	0	1	-360	360;
	8994	7476	0.00228	0.004338	0	0	0	0	0	0	1	-360	360;
	8994	7476	0.00224	0.004431	0	0	0	0	0	0	1	-360	360;
	3023	806	0.00076	0.00747	0	1678	0	0	0	0	1	-360	360;
	5674	806	0.00103	0.01061	0	1481	0	0	0	0	1	-360	360;
	2240	8991	0.002689	0.013818	0	0	0	0	0	0	1	-360	360;
	539	2240	0.002111	0.003769	0	0	0	0	0	0	1	-360	360;
	539	2240	0.002151	0.003689	0	273	0	0	0	0	1	-360	360;
	8814	5426	0.00176	0.012702	0	0	0	0	0	0	1	-360	360;
	8814	5426	0.001729	0.012951	0	0	0	0	0	0	1	-360	360;
	5748	1046	0.000151	0.000511	0	0	0	0	0	0	1	-360	360;
	5748	1046	0.000151	0.000502	0	260	0	0	0	0	1	-360	360;
	5748	1836	0.01076	0.05508	0	0	0	0	0	0	1	-360	360;
	5748	1836	0.010542	0.05616	0	0	0	0	0	0	1	-360	360;
	5748	6824	0.003089	0.009	0	0	0	0	0	0	1	-360	360;
	5748	6824	0.003031	0.009169	0	0	0	0	0	0	1	-360	360;
	5748	6275	0.001849	0.00872	0	0	0	0	0	0	1	-360	360;
	5748	6275	0.00188	0.008538	0	0	0	0	0	0	1	-360	360;
	2974	3301	0.004151	0.01556	0	0	0	0	0	0	1	-360	360;
	1670	2974	0.006591	0.026778	0	299	0	0	0	0	1	-360	360;
	6998	8875	0.00056	0.038671	0	507	0	0	0	0	1	-360	360;
	1747	593	0	0.00045	0	757	0	0	0	0	1	-360	360;
	593	8230	0.00162	0.01968	0	1613	0	0	0	0	1	-360	360;
	6998	8772	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	6998	6356	0.002649	0.015111	0	0	0	0	0	0	1	-360	360;
	6998	6356	0.002702	0.014809	0	0	0	0	0	0	1	-360	360;
	6998	5031	0.0008	0.005711	0	0	0	0	0	0	1	-360	360;
	6998	5031	0.000782	0.005809	0	0	0	0	0	0	1	-360	360;
	6998	3299	0.001049	0.007738	0	286	0	0	0	0	1	-360	360;
	6998	3299	0.001071	0.007569	0	299	0	0	0	0	1	-360	360;
	6998	3119	0.005138	0.025338	0	0	0	0	0	0	1	-360	360;
	6998	3119	0.005031	0.025831	0	0	0	0	0	0	1	-360	360;
	6998	1464	0.003658	0.029071	0	0	0	0	0	0	1	-360	360;
	6998	1464	0.004151	0.028489	0	0	0	0	0	0	1	-360	360;
	1221	1350	0.001471	0.011471	0	0	0	0	0	0	1	-360	360;
	1221	1350	0.001502	0.011218	0	0	0	0	0	0	1	-360	360;
	2305	1350	0.001871	0.014511	0	0	0	0	0	0	1	-360	360;
	2305	1350	0.001902	0.0142	0	0	0	0	0	0	1	-360	360;
	199	7195	0.000351	0.004662	0	0	0	0	0	0	1	-360	360;
	52	199	0.00772	0.033818	0	0	0	0	0	0	1	-360	360;
	4964	8432	0	0.00051	0	1909	0	0	0	0	1	-360	360;
	8432	3190	0.00076	0.0119	0	790	0	0	0	0	1	-360	360;
	8432	3352	0.00025	0.00316	0	1580	0	0	0	0	1	-360	360;
	4964	3484	0.00077	0.01164	0	823	0	0	0	0	1	-360	360;
	4964	6766	0.00025	0.00322	0	1481	0	0	0	0	1	-360	360;
	838	6295	0.001369	0.021182	0	494	0	0	0	0	1	-360	360;
	3190	3484	0	0.00052	0	757	0	0	0	0	1	-360	360;
	3190	6263	0.00015	0.00328	0	790	0	0	0	0	1	-360	360;
	395	3190	0.00016	0.00314	0	1349	0	0	0	0	1	-360	360;
	3484	6068	0.00017	0.00321	0	987	0	0	0	0	1	-360	360;
	5843	3484	0.00018	0.00308	0	954	0	0	0	0	1	-360	360;
	3210	4021	0.00024	0.01948	0	757	0	0	0	0	1	-360	360;
	1670	7425	0.000631	0.037902	0	0	0	0	0	0	1	-360	360;
	2381	5772	0.00244	0.01532	0	0	0	0	0	0	1	-360	360;
	281	8086	0	0.00053	0	757	0	0	0	0	1	-360	360;
	281	6068	0.00013	0.00282	0	987	0	0	0	0	1	-360	360;
	4077	281	0.00038	0.00708	0	1218	0	0	0	0	1	-360	360;
	8086	6263	0.00013	0.00261	0	1086	0	0	0	0	1	-360	360;
	5888	8086	0.00038	0.00722	0	1448	0	0	0	0	1	-360	360;
	1477	5255	0.00016	0.002769	0	0	0	0	0	0	1	-360	360;
	3960	5255	0.00016	0.00272	0	299	0	0	0	0	1	-360	360;
	8037	5255	0.00016	0.002662	0	299	0	0	0	0	1	-360	360;
	1989	1597	0.024479	0.081727	0	0	0	0	0	0	1	-360	360;
	6923	1597	0.024248	0.064893	0	0	0	0	0	0	1	-360	360;
	676	8706	0.0084	0.04056	0	0	0	0	0	0	1	-360	360;
	8706	205	0.003809	0.0248	0	0	0	0	0	0	1	-360	360;
	8706	4362	0.002542	0.014569	0	0	0	0	0	0	1	-360	360;
	8706	4362	0.002591	0.01428	0	0	0	0	0	0	1	-360	360;
	7584	7736	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	7584	5591	0.003071	0.01636	0	0	0	0	0	0	1	-360	360;
	7584	5364	0.003889	0.027022	0	0	0	0	0	0	1	-360	360;
	4774	7584	0.00624	0.036009	0	0	0	0	0	0	1	-360	360;
	4774	7584	0.00636	0.035258	0	0	0	0	0	0	1	-360	360;
	5255	2796	0.000982	0.001529	0	0	0	0	0	0	1	-360	360;
	5255	2796	0.000991	0.001502	0	0	0	0	0	0	1	-360	360;
	8642	5255	0.00052	0.004991	0	260	0	0	0	0	1	-360	360;
	5255	2703	0	0.00028	0	0	0	0	0	0	1	-360	360;
	650	4104	0.012182	0.034826	0	0	0	0	0	0	1	-360	360;
	1989	650	0.025744	0.059	0	0	0	0	0	0	1	-360	360;
	4130	1635	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6167	4130	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6167	5872	0.003129	0.015031	0	0	0	0	0	0	1	-360	360;
	4130	496	0.00228	0.019871	0	0	0	0	0	0	1	-360	360;
	4130	5872	0.002791	0.017031	0	0	0	0	0	0	1	-360	360;
	6582	3462	0.001431	0.01252	0	0	0	0	0	0	1	-360	360;
	6582	4082	0.001649	0.007631	0	0	0	0	0	0	1	-360	360;
	6582	5246	0.001471	0.008649	0	0	0	0	0	0	1	-360	360;
	496	5872	0.00656	0.009071	0	260	0	0	0	0	1	-360	360;
	6047	3664	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2077	9063	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3224	8044	0.000521	0.001529	0	391	0	0	0	0	1	-360	360;
	76	5742	0.000219	0.00095	0	438	0	0	0	0	1	-360	360;
	38	5742	0.000219	0.00088	0	533	0	0	0	0	1	-360	360;
	8814	9118	0.002391	0.015618	0	0	0	0	0	0	1	-360	360;
	5008	919	0	0.00054	0	1448	0	0	0	0	1	-360	360;
	1788	5008	0.00079	0.01037	0	1678	0	0	0	0	1	-360	360;
	662	5008	0.00022	0.00296	0	1645	0	0	0	0	1	-360	360;
	8604	919	0.00077	0.01056	0	1810	0	0	0	0	1	-360	360;
	6354	4295	0.000649	0.03712	0	0	0	0	0	0	1	-360	360;
	6354	5203	0.003302	0.019138	0	0	0	0	0	0	1	-360	360;
	6354	5203	0.003369	0.018742	0	0	0	0	0	0	1	-360	360;
	6354	472	0.001769	0.018222	0	0	0	0	0	0	1	-360	360;
	6354	7476	0.000858	0.006809	0	0	0	0	0	0	1	-360	360;
	6354	7476	0.000969	0.006671	0	0	0	0	0	0	1	-360	360;
	6354	8792	0.005551	0.03908	0	0	0	0	0	0	1	-360	360;
	6354	8792	0.005431	0.039822	0	0	0	0	0	0	1	-360	360;
	6354	8792	0.005311	0.036871	0	0	0	0	0	0	1	-360	360;
	1020	1094	0.002169	0.016271	0	0	0	0	0	0	1	-360	360;
	3182	1094	0.007418	0.07832	0	0	0	0	0	0	1	-360	360;
	4825	3023	0	0.0005	0	757	0	0	0	0	1	-360	360;
	4825	6345	0	0.00049	0	757	0	0	0	0	1	-360	360;
	4825	8114	0.00079	0.00848	0	1580	0	0	0	0	1	-360	360;
	4825	3210	0.00183	0.0174	0	1415	0	0	0	0	1	-360	360;
	3023	6345	0	0.00048	0	757	0	0	0	0	1	-360	360;
	3023	8773	0.0008	0.0083	0	1678	0	0	0	0	1	-360	360;
	7884	71	0.004449	0.02404	0	0	0	0	0	0	1	-360	360;
	7884	71	0.004529	0.023551	0	0	0	0	0	0	1	-360	360;
	7884	6202	0.006791	0.013271	0	260	0	0	0	0	1	-360	360;
	7884	1106	0.00424	0.030378	0	0	0	0	0	0	1	-360	360;
	7884	1106	0.004151	0.028129	0	0	0	0	0	0	1	-360	360;
	7884	7305	0.002311	0.003622	0	0	0	0	0	0	1	-360	360;
	7884	7305	0.002062	0.004102	0	0	0	0	0	0	1	-360	360;
	7884	6987	0.00452	0.029609	0	0	0	0	0	0	1	-360	360;
	7884	6987	0.004422	0.030169	0	0	0	0	0	0	1	-360	360;
	7884	4774	0.00416	0.023058	0	0	0	0	0	0	1	-360	360;
	7884	3620	0.002831	0.020742	0	273	0	0	0	0	1	-360	360;
	3119	1722	0.002338	0.01592	0	0	0	0	0	0	1	-360	360;
	585	3148	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	1106	585	0.004969	0.03636	0	0	0	0	0	0	1	-360	360;
	6987	585	0.003431	0.015698	0	0	0	0	0	0	1	-360	360;
	7275	585	0.00456	0.027582	0	0	0	0	0	0	1	-360	360;
	8114	8773	0	0.00047	0	757	0	0	0	0	1	-360	360;
	8393	8114	0.00093	0.01185	0	1810	0	0	0	0	1	-360	360;
	139	8773	0.00091	0.01207	0	1810	0	0	0	0	1	-360	360;
	676	6439	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	8738	676	0.000529	0.003151	0	0	0	0	0	0	1	-360	360;
	8738	676	0.00052	0.003209	0	0	0	0	0	0	1	-360	360;
	676	205	0.00352	0.020578	0	0	0	0	0	0	1	-360	360;
	8862	6915	0.039182	0.089413	0	0	0	0	0	0	1	-360	360;
	8862	9205	0.033686	0.067124	0	0	0	0	0	0	1	-360	360;
	6393	4057	0.044992	0.093174	0	0	0	0	0	0	1	-360	360;
	6028	6393	0.013678	0.052479	0	0	0	0	0	0	1	-360	360;
	4872	6393	0.005686	0.022893	0	0	0	0	0	0	1	-360	360;
	6393	7136	0.034281	0.094653	0	0	0	0	0	0	1	-360	360;
	2769	6118	0	0.00045	0	0	0	0	0	0	1	-360	360;
	76	2769	0.00032	0.001469	0	438	0	0	0	0	1	-360	360;
	38	6118	0.000289	0.001039	0	457	0	0	0	0	1	-360	360;
	3224	7117	0.00324	0.017628	0	0	0	0	0	0	1	-360	360;
	3224	8525	0.002678	0.019298	0	0	0	0	0	0	1	-360	360;
	3224	5346	0.000529	0.001504	0	0	0	0	0	0	1	-360	360;
	4993	3335	0.00042	0.0261	0	757	0	0	0	0	1	-360	360;
	6832	4797	0.00043	0.02555	0	757	0	0	0	0	1	-360	360;
	6489	4797	0.00044	0.02499	0	757	0	0	0	0	1	-360	360;
	8284	4797	0.00039	0.02833	0	757	0	0	0	0	1	-360	360;
	577	1362	0.00019	0.01594	0	0	0	0	0	0	1	-360	360;
	268	1057	0.000419	0.026099	0	0	0	0	0	0	1	-360	360;
	2728	5979	0.00043	0.02555	0	0	0	0	0	0	1	-360	360;
	8662	8298	0.00038	0.02888	0	757	0	0	0	0	1	-360	360;
	6884	4059	0.00038	0.02944	0	757	0	0	0	0	1	-360	360;
	3723	3335	0.00037	0.02999	0	1876	0	0	0	0	1	-360	360;
	1890	3335	0.00036	0.02777	0	0	0	0	0	0	1	-360	360;
	71	8141	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	7295	71	0.003951	0.02048	0	0	0	0	0	0	1	-360	360;
	71	7297	0.003858	0.020871	0	0	0	0	0	0	1	-360	360;
	4797	8298	0	0.00046	0	757	0	0	0	0	1	-360	360;
	4797	4059	0	0.00045	0	757	0	0	0	0	1	-360	360;
	4797	7388	0	0.00051	0	823	0	0	0	0	1	-360	360;
	4797	2002	0	0.00052	0	823	0	0	0	0	1	-360	360;
	2448	5979	0	0.00051	0	0	0	0	0	0	1	-360	360;
	2448	1057	0	0.000521	0	0	0	0	0	0	1	-360	360;
	2448	6282	0	0.000531	0	381	0	0	0	0	1	-360	360;
	2448	4340	0	0.000539	0	0	0	0	0	0	1	-360	360;
	3335	8298	0	0.00053	0	1415	0	0	0	0	1	-360	360;
	3335	4059	0	0.00054	0	954	0	0	0	0	1	-360	360;
	3335	7388	0	0.0005	0	1119	0	0	0	0	1	-360	360;
	3335	2002	0	0.00049	0	921	0	0	0	0	1	-360	360;
	326	5979	0	0.0005	0	0	0	0	0	0	1	-360	360;
	326	1057	0	0.00049	0	0	0	0	0	0	1	-360	360;
	326	6282	0	0.000479	0	0	0	0	0	0	1	-360	360;
	326	4340	0	0.000469	0	0	0	0	0	0	1	-360	360;
	4858	8298	0.00043	0.00544	0	1580	0	0	0	0	1	-360	360;
	2523	5979	0	0.000461	0	476	0	0	0	0	1	-360	360;
	1057	1362	0	0.00045	0	457	0	0	0	0	1	-360	360;
	1956	7388	0.00038	0.00616	0	1448	0	0	0	0	1	-360	360;
	5299	2002	0.00041	0.02721	0	1086	0	0	0	0	1	-360	360;
	4340	1052	0.002081	0.023	0	533	0	0	0	0	1	-360	360;
	6662	838	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5165	2722	0	0.00048	0	1514	0	0	0	0	1	-360	360;
	5165	1689	0.0011	0.01189	0	1810	0	0	0	0	1	-360	360;
	5165	3210	0.00084	0.00906	0	1020	0	0	0	0	1	-360	360;
	2722	1689	0.00112	0.01165	0	1876	0	0	0	0	1	-360	360;
	2722	8230	0.00115	0.0114	0	1744	0	0	0	0	1	-360	360;
	2722	5674	0.00086	0.00887	0	1283	0	0	0	0	1	-360	360;
	3363	7195	0.00116	0.00912	0	260	0	0	0	0	1	-360	360;
	3363	7195	0.001129	0.009298	0	0	0	0	0	0	1	-360	360;
	7195	8977	0.000609	0.005058	0	0	0	0	0	0	1	-360	360;
	52	7195	0.008009	0.04	0	0	0	0	0	0	1	-360	360;
	7195	3868	0.002289	0.016218	0	0	0	0	0	0	1	-360	360;
	3363	392	0.003098	0.032729	0	0	0	0	0	0	1	-360	360;
	3363	3530	0.002982	0.013058	0	0	0	0	0	0	1	-360	360;
	3363	3530	0.00304	0.0128	0	0	0	0	0	0	1	-360	360;
	3363	5829	0.008791	0.059622	0	0	0	0	0	0	1	-360	360;
	8977	6539	0.000511	0.004409	0	0	0	0	0	0	1	-360	360;
	7675	6539	0.009391	0.030809	0	0	0	0	0	0	1	-360	360;
	7675	6539	0.009191	0.031391	0	260	0	0	0	0	1	-360	360;
	1702	2900	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	6202	1702	0.00276	0.014782	0	0	0	0	0	0	1	-360	360;
	6202	1702	0.002809	0.014458	0	0	0	0	0	0	1	-360	360;
	1106	1702	0.003138	0.017502	0	0	0	0	0	0	1	-360	360;
	1106	1702	0.0032	0.017138	0	0	0	0	0	0	1	-360	360;
	5916	6342	0.002033	0.005273	0	0	0	0	0	0	1	-360	360;
	4302	6342	0.036066	0.160107	0	0	0	0	0	0	1	-360	360;
	733	8917	0.027686	0.04614	0	0	0	0	0	0	1	-360	360;
	5423	8917	0.025107	0.051372	0	0	0	0	0	0	1	-360	360;
	1777	548	0.002479	0.01086	0	0	0	0	0	0	1	-360	360;
	548	6923	0.013603	0.051876	0	0	0	0	0	0	1	-360	360;
	9029	8508	0.00176	0.108	0	0	0	0	0	0	1	-360	360;
	1591	5558	0	0.00047	0	987	0	0	0	0	1	-360	360;
	1591	8604	0.00038	0.00366	0	1744	0	0	0	0	1	-360	360;
	1591	8264	0.00052	0.00633	0	1909	0	0	0	0	1	-360	360;
	1124	8510	0	0.00051	0	0	0	0	0	0	1	-360	360;
	1124	3257	0.001599	0.015599	0	0	0	0	0	0	1	-360	360;
	5311	1124	0.00212	0.022349	0	0	0	0	0	0	1	-360	360;
	5558	1788	0.00034	0.00415	0	1645	0	0	0	0	1	-360	360;
	8211	5558	0.00051	0.00646	0	1744	0	0	0	0	1	-360	360;
	8510	369	0.00163	0.015271	0	0	0	0	0	0	1	-360	360;
	3891	8510	0.002159	0.021901	0	0	0	0	0	0	1	-360	360;
	5952	784	0.00224	0.015018	0	0	0	0	0	0	1	-360	360;
	5952	3182	0.004111	0.00652	0	0	0	0	0	0	1	-360	360;
	5952	3182	0.004191	0.006391	0	0	0	0	0	0	1	-360	360;
	6923	5108	0.00238	0.006744	0	0	0	0	0	0	1	-360	360;
	5829	392	0.003329	0.033129	0	0	0	0	0	0	1	-360	360;
	1633	8389	0.005809	0.021182	0	0	0	0	0	0	1	-360	360;
	1633	8389	0.005689	0.021591	0	0	0	0	0	0	1	-360	360;
	8389	2720	0.00132	0.01052	0	0	0	0	0	0	1	-360	360;
	8389	243	0.001129	0.012329	0	0	0	0	0	0	1	-360	360;
	8389	4409	0.001569	0.1224	0	0	0	0	0	0	1	-360	360;
	3119	2572	0.000662	0.036351	0	468	0	0	0	0	1	-360	360;
	4318	7404	0.001777	0.009107	0	0	0	0	0	0	1	-360	360;
	4318	1338	0.002967	0.002901	0	0	0	0	0	0	1	-360	360;
	4318	8288	0.017893	0.057479	0	0	0	0	0	0	1	-360	360;
	7404	8882	0.000554	0.001091	0	0	0	0	0	0	1	-360	360;
	7404	7565	0.006339	0.022603	0	0	0	0	0	0	1	-360	360;
	1338	360	0.004174	0.006702	0	0	0	0	0	0	1	-360	360;
	360	7565	0.009818	0.040099	0	0	0	0	0	0	1	-360	360;
	275	6323	0.003729	0.01876	0	0	0	0	0	0	1	-360	360;
	5872	7230	0.000782	0.059938	0	0	0	0	0	0	1	-360	360;
	8194	5872	0.000769	0.061111	0	0	0	0	0	0	1	-360	360;
	1020	6514	0.000382	0.019378	0	0	0	0	0	0	1	-360	360;
	1020	387	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1106	6838	1.8e-05	0.00188	0	637	0	0	0	0	1	-360	360;
	8480	6025	0.000372	0.001223	0	0	0	0	0	0	1	-360	360;
	9029	5835	0.000809	0.001751	0	0	0	0	0	0	1	-360	360;
	3374	878	0	0.043511	0	0	0	0	0	0	1	-360	360;
	7450	3374	0.003529	0.022178	0	0	0	0	0	0	1	-360	360;
	7450	3374	0.0036	0.021711	0	0	0	0	0	0	1	-360	360;
	6383	231	0	0.000413	0	0	0	0	0	0	1	-360	360;
	2228	6383	0.047479	0.102603	0	0	0	0	0	0	1	-360	360;
	2228	6383	0.053727	0.091281	0	0	0	0	0	0	1	-360	360;
	1010	6383	0.021678	0.045669	0	0	0	0	0	0	1	-360	360;
	2228	8878	0.004397	0.030678	0	0	0	0	0	0	1	-360	360;
	2228	8878	0.004314	0.031256	0	0	0	0	0	0	1	-360	360;
	2228	7117	0.008347	0.055512	0	0	0	0	0	0	1	-360	360;
	2228	2698	0.006223	0.039744	0	0	0	0	0	0	1	-360	360;
	8277	2228	0.006124	0.041686	0	0	0	0	0	0	1	-360	360;
	8277	2228	0.006942	0.040851	0	0	0	0	0	0	1	-360	360;
	4057	5783	0.019248	0.039041	0	191	0	0	0	0	1	-360	360;
	7136	4057	0.015719	0.04943	0	0	0	0	0	0	1	-360	360;
	5549	1207	0.00296	0.011871	0	0	0	0	0	0	1	-360	360;
	5549	9083	0.01328	0.04144	0	0	0	0	0	0	1	-360	360;
	2745	7138	0.00736	0.033222	0	0	0	0	0	0	1	-360	360;
	2745	7138	0.007502	0.032529	0	0	0	0	0	0	1	-360	360;
	1670	2745	0.00192	0.02576	0	299	0	0	0	0	1	-360	360;
	1670	2745	0.002169	0.02524	0	299	0	0	0	0	1	-360	360;
	2745	8492	0.00656	0.017071	0	0	0	0	0	0	1	-360	360;
	2745	2748	0.002662	0.007471	0	0	0	0	0	0	1	-360	360;
	2745	3215	0.003191	0.009062	0	0	0	0	0	0	1	-360	360;
	6915	8668	0.00981	0.057298	0	0	0	0	0	0	1	-360	360;
	3177	6356	0.008342	0.037031	0	0	0	0	0	0	1	-360	360;
	1464	6356	0.0086	0.039791	0	0	0	0	0	0	1	-360	360;
	1689	8230	0	0.00046	0	1843	0	0	0	0	1	-360	360;
	6068	1689	0.00055	0.00685	0	1645	0	0	0	0	1	-360	360;
	6263	8230	0.00054	0.00699	0	1645	0	0	0	0	1	-360	360;
	8814	69	0.0056	0.052889	0	0	0	0	0	0	1	-360	360;
	8814	5949	0.006058	0.023591	0	0	0	0	0	0	1	-360	360;
	8814	5949	0.0054	0.026729	0	0	0	0	0	0	1	-360	360;
	8814	8613	0.00168	0.01188	0	299	0	0	0	0	1	-360	360;
	8814	9007	0.003178	0.028658	0	0	0	0	0	0	1	-360	360;
	8814	5188	0.003178	0.028542	0	299	0	0	0	0	1	-360	360;
	2077	6442	0.00068	0.004449	0	0	0	0	0	0	1	-360	360;
	2077	3462	0.00076	0.004818	0	0	0	0	0	0	1	-360	360;
	2812	1778	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2812	1633	9e-06	0.000222	0	0	0	0	0	0	1	-360	360;
	3983	1633	0	0.000222	0	299	0	0	0	0	1	-360	360;
	7442	1633	1.8e-05	0.004329	0	0	0	0	0	0	1	-360	360;
	6047	5271	0.001831	0.001698	0	0	0	0	0	0	1	-360	360;
	6047	5271	0.002209	0.001471	0	0	0	0	0	0	1	-360	360;
	211	2832	0.017653	0.070289	0	0	0	0	0	0	1	-360	360;
	4132	2832	0.013231	0.042207	0	0	0	0	0	0	1	-360	360;
	4132	2832	0.013479	0.041331	0	0	0	0	0	0	1	-360	360;
	2804	2832	0.013934	0.055702	0	0	0	0	0	0	1	-360	360;
	4216	6028	0.000537	0.001471	0	0	0	0	0	0	1	-360	360;
	6028	4872	0.009967	0.048959	0	0	0	0	0	0	1	-360	360;
	6028	9146	0.030612	0.055058	0	0	0	0	0	0	1	-360	360;
	6028	9146	0.031207	0.053934	0	0	0	0	0	0	1	-360	360;
	6058	6028	0.01314	0.040579	0	0	0	0	0	0	1	-360	360;
	6058	6028	0.012851	0.03757	0	0	0	0	0	0	1	-360	360;
	6028	7011	0.006901	0.021174	0	0	0	0	0	0	1	-360	360;
	6028	7011	0.007653	0.01924	0	0	0	0	0	0	1	-360	360;
	7295	5595	0.008982	0.038462	0	0	0	0	0	0	1	-360	360;
	5595	7297	0.009151	0.037649	0	0	0	0	0	0	1	-360	360;
	5595	1020	0.004258	0.03952	0	0	0	0	0	0	1	-360	360;
	2616	5595	0.003489	0.0264	0	0	0	0	0	0	1	-360	360;
	5595	2305	0.003849	0.05232	0	0	0	0	0	0	1	-360	360;
	1989	4104	0.032669	0.080273	0	0	0	0	0	0	1	-360	360;
	1889	4104	0.004901	0.014967	0	0	0	0	0	0	1	-360	360;
	312	5783	0.000562	0.001058	0	0	0	0	0	0	1	-360	360;
	8393	139	0	0.00045	0	1547	0	0	0	0	1	-360	360;
	8393	8264	0.001	0.01169	0	1481	0	0	0	0	1	-360	360;
	139	8211	0.00098	0.01192	0	1448	0	0	0	0	1	-360	360;
	6600	7565	0.002339	0.006223	0	0	0	0	0	0	1	-360	360;
	6600	7565	0.002388	0.006091	0	0	0	0	0	0	1	-360	360;
	2555	6600	0.000504	0.001207	0	0	0	0	0	0	1	-360	360;
	5783	7565	0.00105	0.007959	0	248	0	0	0	0	1	-360	360;
	5783	7565	0.001033	0.007372	0	267	0	0	0	0	1	-360	360;
	5783	2648	0.018752	0.042339	0	0	0	0	0	0	1	-360	360;
	5783	2648	0.018339	0.039231	0	0	0	0	0	0	1	-360	360;
	1989	6923	0.036793	0.081719	0	0	0	0	0	0	1	-360	360;
	1989	8480	0.003702	0.015132	0	0	0	0	0	0	1	-360	360;
	2478	1989	0.00124	0.007314	0	219	0	0	0	0	1	-360	360;
	523	5760	0.002329	0.00356	0	299	0	0	0	0	1	-360	360;
	523	5760	0.002378	0.00348	0	299	0	0	0	0	1	-360	360;
	523	5760	0.00212	0.003951	0	299	0	0	0	0	1	-360	360;
	523	7476	0.001951	0.003791	0	364	0	0	0	0	1	-360	360;
	69	7852	0.002542	0.01472	0	0	0	0	0	0	1	-360	360;
	1670	5342	0.003262	0.024151	0	0	0	0	0	0	1	-360	360;
	9206	4757	0.0036	0.01548	0	260	0	0	0	0	1	-360	360;
	9206	4757	0.00352	0.014329	0	260	0	0	0	0	1	-360	360;
	1880	4380	0.002591	0.014418	0	0	0	0	0	0	1	-360	360;
	3530	1880	0.001071	0.003058	0	0	0	0	0	0	1	-360	360;
	3530	4687	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	3530	1330	0.017458	0.052129	0	0	0	0	0	0	1	-360	360;
	7295	8442	0.00136	0.00216	0	0	0	0	0	0	1	-360	360;
	7297	8442	0.001391	0.00212	0	0	0	0	0	0	1	-360	360;
	6399	9083	0.005858	0.016591	0	0	0	0	0	0	1	-360	360;
	898	8668	0.004562	0.011653	0	0	0	0	0	0	1	-360	360;
	898	6526	0.030868	0.064174	0	0	0	0	0	0	1	-360	360;
	3257	369	0	0.000521	0	0	0	0	0	0	1	-360	360;
	2381	1020	0.005929	0.053271	0	0	0	0	0	0	1	-360	360;
	2381	297	0.00664	0.040849	0	0	0	0	0	0	1	-360	360;
	2381	3182	0.00968	0.049382	0	0	0	0	0	0	1	-360	360;
	5829	4692	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	5829	69	0.00044	0.004489	0	0	0	0	0	0	1	-360	360;
	5829	3898	0.01064	0.043751	0	0	0	0	0	0	1	-360	360;
	5829	3898	0.0104	0.040511	0	0	0	0	0	0	1	-360	360;
	5188	5829	0.0024	0.025498	0	0	0	0	0	0	1	-360	360;
	4272	1205	0.00264	0.01412	0	0	0	0	0	0	1	-360	360;
	472	6131	0.000609	0.03944	0	663	0	0	0	0	1	-360	360;
	8565	472	0	0.000222	0	0	0	0	0	0	1	-360	360;
	472	5760	0.001182	0.003018	0	299	0	0	0	0	1	-360	360;
	472	5760	0.0012	0.002951	0	299	0	0	0	0	1	-360	360;
	472	5760	0.001071	0.003351	0	286	0	0	0	0	1	-360	360;
	472	7073	0.001031	0.01544	0	0	0	0	0	0	1	-360	360;
	472	349	0.001009	0.015738	0	0	0	0	0	0	1	-360	360;
	472	7156	0.000991	0.01604	0	0	0	0	0	0	1	-360	360;
	6202	7305	0.003898	0.008862	0	0	0	0	0	0	1	-360	360;
	6202	7305	0.003818	0.009031	0	0	0	0	0	0	1	-360	360;
	643	311	0.003831	0.012502	0	0	0	0	0	0	1	-360	360;
	6068	6263	0	0.00051	0	856	0	0	0	0	1	-360	360;
	6017	6263	0.00127	0.0132	0	1481	0	0	0	0	1	-360	360;
	7450	2271	0	0.000222	0	299	0	0	0	0	1	-360	360;
	7450	7666	0.001338	0.011209	0	0	0	0	0	0	1	-360	360;
	7450	7018	0.004222	0.030062	0	0	0	0	0	0	1	-360	360;
	8784	7450	0.001182	0.0104	0	0	0	0	0	0	1	-360	360;
	2271	3846	9e-06	0.000222	0	0	0	0	0	0	1	-360	360;
	2271	7018	0.004138	0.03064	0	0	0	0	0	0	1	-360	360;
	3846	7018	0.004049	0.031209	0	0	0	0	0	0	1	-360	360;
	733	5423	0.022603	0.051959	0	0	0	0	0	0	1	-360	360;
	4614	733	0.013636	0.032041	0	0	0	0	0	0	1	-360	360;
	8492	3216	0.00228	0.009222	0	0	0	0	0	0	1	-360	360;
	2748	3216	0.002471	0.010302	0	273	0	0	0	0	1	-360	360;
	3215	3216	0.001831	0.007378	0	273	0	0	0	0	1	-360	360;
	8878	4302	0.006843	0.049248	0	0	0	0	0	0	1	-360	360;
	8878	4302	0.006686	0.045603	0	0	0	0	0	0	1	-360	360;
	2760	8016	0.004102	0.011529	0	0	0	0	0	0	1	-360	360;
	2760	8016	0.004182	0.01128	0	0	0	0	0	0	1	-360	360;
	1913	2120	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8155	1913	0.003649	0.0122	0	0	0	0	0	0	1	-360	360;
	8155	1913	0.00372	0.011938	0	0	0	0	0	0	1	-360	360;
	1913	7090	0.005031	0.028102	0	0	0	0	0	0	1	-360	360;
	1913	7090	0.005129	0.02752	0	0	0	0	0	0	1	-360	360;
	1913	7844	0.005218	0.026951	0	0	0	0	0	0	1	-360	360;
	4413	3462	0.00288	0.01992	0	0	0	0	0	0	1	-360	360;
	4413	3462	0.003262	0.01952	0	0	0	0	0	0	1	-360	360;
	8738	1106	0.00956	0.061022	0	0	0	0	0	0	1	-360	360;
	8738	2760	0.005382	0.033911	0	0	0	0	0	0	1	-360	360;
	8738	2760	0.005271	0.034569	0	0	0	0	0	0	1	-360	360;
	8738	6218	0.004102	0.023018	0	0	0	0	0	0	1	-360	360;
	1788	8604	0	0.00052	0	757	0	0	0	0	1	-360	360;
	1020	2335	0	0.000222	0	481	0	0	0	0	1	-360	360;
	1020	3182	0.011631	0.094591	0	0	0	0	0	0	1	-360	360;
	2616	1020	0.0012	0.008769	0	299	0	0	0	0	1	-360	360;
	1020	9072	0.001218	0.008582	0	299	0	0	0	0	1	-360	360;
	7138	7586	0.002689	0.013818	0	0	0	0	0	0	1	-360	360;
	2648	8288	0.005	0.008934	0	0	0	0	0	0	1	-360	360;
	2824	140	0	0.000531	0	0	0	0	0	0	1	-360	360;
	2824	5311	0.001281	0.0158	0	0	0	0	0	0	1	-360	360;
	140	3891	0.00125	0.016099	0	0	0	0	0	0	1	-360	360;
	211	6526	0.010058	0.048	0	0	0	0	0	0	1	-360	360;
	211	6526	0.010248	0.047	0	0	0	0	0	0	1	-360	360;
	211	2804	0.004314	0.024537	0	0	0	0	0	0	1	-360	360;
	1670	3301	0.010791	0.042151	0	299	0	0	0	0	1	-360	360;
	344	1011	0.0006	0.040218	0	0	0	0	0	0	1	-360	360;
	243	8621	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3210	5674	0	0.00053	0	1711	0	0	0	0	1	-360	360;
	3210	4337	0	0.00054	0	1251	0	0	0	0	1	-360	360;
	1670	8204	0.00196	0.010249	0	0	0	0	0	0	1	-360	360;
	1670	1516	0.002182	0.010498	0	0	0	0	0	0	1	-360	360;
	5949	6480	0.000582	0.040991	0	0	0	0	0	0	1	-360	360;
	5949	5655	0.000569	0.04176	0	0	0	0	0	0	1	-360	360;
	1464	3177	0.003818	0.013209	0	0	0	0	0	0	1	-360	360;
	5949	6323	0.00428	0.023511	0	0	0	0	0	0	1	-360	360;
	5949	6323	0.004182	0.021769	0	0	0	0	0	0	1	-360	360;
	1956	4858	0	0.0005	0	1086	0	0	0	0	1	-360	360;
	1956	8264	0.00114	0.01661	0	1481	0	0	0	0	1	-360	360;
	4320	4248	0	0.000539	0	0	0	0	0	0	1	-360	360;
	4320	76	0.00119	0.014081	0	0	0	0	0	0	1	-360	360;
	4858	8211	0.00117	0.01626	0	1711	0	0	0	0	1	-360	360;
	4248	38	0.00106	0.01595	0	0	0	0	0	0	1	-360	360;
	4302	999	0.018058	0.065719	0	0	0	0	0	0	1	-360	360;
	4302	1010	0.012826	0.025058	0	0	0	0	0	0	1	-360	360;
	4302	1010	0.013331	0.024	0	0	0	0	0	0	1	-360	360;
	8204	1516	0.001089	0.003151	0	0	0	0	0	0	1	-360	360;
	5792	2428	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	344	1277	0.000631	0.054471	0	0	0	0	0	0	1	-360	360;
	344	2117	0.000622	0.055489	0	0	0	0	0	0	1	-360	360;
	344	8895	0.0006	0.051382	0	0	0	0	0	0	1	-360	360;
	4132	6526	0.017702	0.064331	0	0	0	0	0	0	1	-360	360;
	4132	1787	0.012314	0.04924	0	0	0	0	0	0	1	-360	360;
	2331	8380	0.001223	0.001248	0	0	0	0	0	0	1	-360	360;
	5296	8380	0.001471	0.00481	0	0	0	0	0	0	1	-360	360;
	999	8380	0.011521	0.03138	0	0	0	0	0	0	1	-360	360;
	9029	7915	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9029	6149	0.000289	0.00056	0	0	0	0	0	0	1	-360	360;
	3968	3370	0.00034	0.01805	0	757	0	0	0	0	1	-360	360;
	7480	2720	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2150	6687	0.00035	0.01766	0	1053	0	0	0	0	1	-360	360;
	8956	4209	0.00035	0.01728	0	1053	0	0	0	0	1	-360	360;
	5065	8318	0.00031	0.01958	0	1876	0	0	0	0	1	-360	360;
	7270	3370	0.00031	0.01997	0	856	0	0	0	0	1	-360	360;
	6424	3370	0.0003	0.02035	0	757	0	0	0	0	1	-360	360;
	5888	8223	0.00016	0.0025	0	1020	0	0	0	0	1	-360	360;
	4776	3370	0.00029	0.02074	0	1119	0	0	0	0	1	-360	360;
	1331	8318	0.00029	0.0192	0	757	0	0	0	0	1	-360	360;
	8483	8318	0.00033	0.01882	0	757	0	0	0	0	1	-360	360;
	6687	4209	0	0.00049	0	757	0	0	0	0	1	-360	360;
	4733	6687	0	0.00048	0	1283	0	0	0	0	1	-360	360;
	7856	4209	0	0.00047	0	1382	0	0	0	0	1	-360	360;
	3370	5897	0	0.00046	0	1349	0	0	0	0	1	-360	360;
	3370	420	0	0.00045	0	1086	0	0	0	0	1	-360	360;
	3370	8898	0	0.00051	0	757	0	0	0	0	1	-360	360;
	8318	7856	0	0.00052	0	987	0	0	0	0	1	-360	360;
	8318	4733	0	0.00053	0	921	0	0	0	0	1	-360	360;
	8318	8898	0	0.00054	0	757	0	0	0	0	1	-360	360;
	5897	4087	0	0.0005	0	1448	0	0	0	0	1	-360	360;
	4087	7856	0	0.00049	0	1843	0	0	0	0	1	-360	360;
	4087	4077	0.00028	0.00403	0	1382	0	0	0	0	1	-360	360;
	6283	4087	0.00021	0.00396	0	1777	0	0	0	0	1	-360	360;
	420	2927	0	0.00048	0	1119	0	0	0	0	1	-360	360;
	2927	4733	0	0.00047	0	1251	0	0	0	0	1	-360	360;
	2927	5888	0.00029	0.00394	0	954	0	0	0	0	1	-360	360;
	4485	2927	0.0002	0.00404	0	1744	0	0	0	0	1	-360	360;
	2720	243	0	0.000222	0	0	0	0	0	0	1	-360	360;
	243	558	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8492	2748	0.004471	0.020791	0	0	0	0	0	0	1	-360	360;
	8492	3215	0.002791	0.01132	0	0	0	0	0	0	1	-360	360;
	5760	8392	0.000822	0.001649	0	0	0	0	0	0	1	-360	360;
	5760	8392	0.0008	0.00168	0	0	0	0	0	0	1	-360	360;
	5760	8392	0.000782	0.00156	0	0	0	0	0	0	1	-360	360;
	370	311	0.00432	0.012662	0	0	0	0	0	0	1	-360	360;
	370	311	0.003849	0.014351	0	0	0	0	0	0	1	-360	360;
	1106	2552	0.001418	0.004618	0	0	0	0	0	0	1	-360	360;
	1106	7275	0.001689	0.011631	0	0	0	0	0	0	1	-360	360;
	1106	6218	0.005991	0.03832	0	0	0	0	0	0	1	-360	360;
	9206	1936	0.001169	0.007098	0	403	0	0	0	0	1	-360	360;
	7476	1936	0.0012	0.00696	0	299	0	0	0	0	1	-360	360;
	4415	9146	0.030066	0.046289	0	0	0	0	0	0	1	-360	360;
	4415	9146	0.026793	0.052463	0	0	0	0	0	0	1	-360	360;
	7729	4415	0.023413	0.057397	0	0	0	0	0	0	1	-360	360;
	5591	5364	0.001631	0.00832	0	0	0	0	0	0	1	-360	360;
	6047	4431	0.012569	0.04732	0	0	0	0	0	0	1	-360	360;
	6047	4431	0.012311	0.048231	0	0	0	0	0	0	1	-360	360;
	8017	6519	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7018	8017	0.000982	0.004751	0	0	0	0	0	0	1	-360	360;
	2965	1260	0.00144	0.12	0	0	0	0	0	0	1	-360	360;
	344	5364	0.001249	0.0064	0	0	0	0	0	0	1	-360	360;
	344	5364	0.00112	0.007249	0	0	0	0	0	0	1	-360	360;
	5311	3891	0	0.0005	0	0	0	0	0	0	1	-360	360;
	9205	8668	0.022843	0.098893	0	0	0	0	0	0	1	-360	360;
	8668	6526	0.034909	0.072893	0	0	0	0	0	0	1	-360	360;
	7666	377	0.00188	0.092222	0	0	0	0	0	0	1	-360	360;
	7666	1227	0.00184	0.093991	0	0	0	0	0	0	1	-360	360;
	7666	4006	0.0018	0.09576	0	0	0	0	0	0	1	-360	360;
	5872	7666	0.00132	0.01088	0	0	0	0	0	0	1	-360	360;
	6058	1889	0.01924	0.056603	0	0	0	0	0	0	1	-360	360;
	6058	1889	0.017124	0.064149	0	0	0	0	0	0	1	-360	360;
	4272	6747	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	4272	838	0.006889	0.03404	0	0	0	0	0	0	1	-360	360;
	4272	838	0.007018	0.033329	0	0	0	0	0	0	1	-360	360;
	4272	7951	0.006169	0.035689	0	0	0	0	0	0	1	-360	360;
	4272	7951	0.00628	0.034929	0	0	0	0	0	0	1	-360	360;
	4272	8792	0.004271	0.026418	0	0	0	0	0	0	1	-360	360;
	4272	8792	0.00436	0.02588	0	0	0	0	0	0	1	-360	360;
	4272	8792	0.00444	0.025338	0	0	0	0	0	0	1	-360	360;
	3898	1330	0.00852	0.02444	0	0	0	0	0	0	1	-360	360;
	8613	52	0.002671	0.018929	0	0	0	0	0	0	1	-360	360;
	662	3352	0	0.00046	0	1613	0	0	0	0	1	-360	360;
	662	6766	0	0.00045	0	1481	0	0	0	0	1	-360	360;
	6017	3352	0	0.00051	0	1185	0	0	0	0	1	-360	360;
	6017	6766	0	0.00052	0	1382	0	0	0	0	1	-360	360;
	2552	7275	0.000329	0.000902	0	0	0	0	0	0	1	-360	360;
	8792	3712	0.00068	0.050351	0	0	0	0	0	0	1	-360	360;
	3388	838	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1362	2523	0	0.00049	0	0	0	0	0	0	1	-360	360;
	1362	1052	0.00212	0.022531	0	553	0	0	0	0	1	-360	360;
	1362	76	0.000919	0.012789	0	0	0	0	0	0	1	-360	360;
	2523	3878	0.002159	0.02206	0	438	0	0	0	0	1	-360	360;
	2523	38	0.00094	0.012531	0	0	0	0	0	0	1	-360	360;
	194	7018	0.00032	0.014102	0	299	0	0	0	0	1	-360	360;
	4846	7018	0.000329	0.0138	0	299	0	0	0	0	1	-360	360;
	3691	7018	0.000329	0.013502	0	299	0	0	0	0	1	-360	360;
	6323	539	0.005818	0.02396	0	0	0	0	0	0	1	-360	360;
	6323	539	0.005942	0.023462	0	0	0	0	0	0	1	-360	360;
	5872	6707	0.000671	0.001871	0	260	0	0	0	0	1	-360	360;
	4863	6707	0	0.001791	0	0	0	0	0	0	1	-360	360;
	5872	3509	0.002249	0.003471	0	0	0	0	0	0	1	-360	360;
	5872	3509	0.001631	0.0042	0	0	0	0	0	0	1	-360	360;
	5872	3776	0.00092	0.002582	0	0	0	0	0	0	1	-360	360;
	7018	3509	0	0.000222	0	585	0	0	0	0	1	-360	360;
	7018	5974	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	8784	7018	0.000822	0.00572	0	0	0	0	0	0	1	-360	360;
	7018	4565	0.001338	0.013622	0	0	0	0	0	0	1	-360	360;
	8058	7018	0.00136	0.004649	0	0	0	0	0	0	1	-360	360;
	3509	4863	0	0.001751	0	0	0	0	0	0	1	-360	360;
	3776	3509	0	0.001711	0	0	0	0	0	0	1	-360	360;
	4082	6442	0.000662	0.004529	0	273	0	0	0	0	1	-360	360;
	5246	6442	0.000649	0.004618	0	0	0	0	0	0	1	-360	360;
	6526	1787	0.005669	0.028992	0	0	0	0	0	0	1	-360	360;
	4616	7260	0.001631	0.1176	0	0	0	0	0	0	1	-360	360;
	1221	7996	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	344	1221	0.006942	0.04516	0	0	0	0	0	0	1	-360	360;
	344	1221	0.0068	0.04604	0	0	0	0	0	0	1	-360	360;
	8525	2698	0.003289	0.017909	0	0	0	0	0	0	1	-360	360;
	3215	7407	0.00056	0.038671	0	0	0	0	0	0	1	-360	360;
	4686	7729	0.02162	0.062612	0	0	0	0	0	0	1	-360	360;
	52	2494	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	52	9007	0.000822	0.00656	0	0	0	0	0	0	1	-360	360;
	52	928	0.003009	0.004649	0	273	0	0	0	0	1	-360	360;
	52	928	0.003071	0.004551	0	286	0	0	0	0	1	-360	360;
	6987	4509	0.00464	0.029431	0	0	0	0	0	0	1	-360	360;
	6987	4509	0.005249	0.02884	0	0	0	0	0	0	1	-360	360;
	3119	6047	0.001902	0.012702	0	0	0	0	0	0	1	-360	360;
	3119	6047	0.001858	0.012938	0	0	0	0	0	0	1	-360	360;
	3119	7662	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6149	5442	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6275	3819	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	344	6015	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	9083	994	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	9206	3169	9.8e-05	0.000222	0	260	0	0	0	0	1	-360	360;
	9206	838	0.004738	0.012529	0	0	0	0	0	0	1	-360	360;
	9206	838	0.004831	0.012262	0	0	0	0	0	0	1	-360	360;
	9206	838	0.002711	0.018431	0	0	0	0	0	0	1	-360	360;
	5423	9223	0.01205	0.03281	0	0	0	0	0	0	1	-360	360;
	3314	838	0	0.000222	0	0	0	0	0	0	1	-360	360;
	838	7779	0.000649	0.03712	0	390	0	0	0	0	1	-360	360;
	838	4477	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6320	838	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4774	3620	0.00076	0.005329	0	0	0	0	0	0	1	-360	360;
	3182	297	0.005769	0.03528	0	0	0	0	0	0	1	-360	360;
	7476	1463	0.002258	0.004849	0	0	0	0	0	0	1	-360	360;
	7476	1463	0.002209	0.004489	0	260	0	0	0	0	1	-360	360;
	1052	3878	0	0.000479	0	0	0	0	0	0	1	-360	360;
	1052	6477	0.00256	0.01561	0	533	0	0	0	0	1	-360	360;
	3878	5295	0.00269	0.01451	0	591	0	0	0	0	1	-360	360;
	8277	5887	0.003372	0.028231	0	0	0	0	0	0	1	-360	360;
	8277	9127	0.007033	0.047083	0	0	0	0	0	0	1	-360	360;
	76	38	0	0.000469	0	0	0	0	0	0	1	-360	360;
	3182	1984	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	2760	3182	0.00392	0.025191	0	0	0	0	0	0	1	-360	360;
	2760	3182	0.00444	0.024689	0	0	0	0	0	0	1	-360	360;
	9127	5887	0.003529	0.021223	0	0	0	0	0	0	1	-360	360;
	4082	5246	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4485	395	0.00069	0.00134	0	1481	0	0	0	0	1	-360	360;
	5843	6283	0.00067	0.00136	0	1448	0	0	0	0	1	-360	360;
	2478	8480	0.012992	0.070207	0	0	0	0	0	0	1	-360	360;
	395	5843	0	0.00053	0	823	0	0	0	0	1	-360	360;
	6477	5295	0	0.000461	0	0	0	0	0	0	1	-360	360;
	8264	8211	0	0.00054	0	790	0	0	0	0	1	-360	360;
	7011	4614	0.002314	0.005661	0	0	0	0	0	0	1	-360	360;
	4614	7011	0.002512	0.005909	0	0	0	0	0	0	1	-360	360;
	6058	7011	0.006364	0.018488	0	0	0	0	0	0	1	-360	360;
	6058	7011	0.005959	0.017752	0	0	0	0	0	0	1	-360	360;
	6058	9223	0.018397	0.04176	0	0	0	0	0	0	1	-360	360;
	6058	9223	0.018058	0.042653	0	0	0	0	0	0	1	-360	360;
	7090	7844	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2616	559	0	0.000529	0	0	0	0	0	0	1	-360	360;
	9072	559	0	0.00052	0	0	0	0	0	0	1	-360	360;
	2305	9072	0.003369	0.031329	0	338	0	0	0	0	1	-360	360;
	3341	4259	0.019671	0.07004	0	0	0	0	0	0	1	-360	360;
	3341	4259	0.01812	0.06872	0	0	0	0	0	0	1	-360	360;
	5805	4721	0.001582	0.00724	0	0	0	0	0	0	1	-360	360;
	5805	4721	0.001551	0.007231	0	0	0	0	0	0	1	-360	360;
	5805	4721	0.001262	0.00796	0	0	0	0	0	0	1	-360	360;
	5452	2734	0.0264	0.075129	0	0	0	0	0	0	1	-360	360;
	5452	5274	0.00824	0.044751	0	0	0	0	0	0	1	-360	360;
	7879	4721	0.018271	0.070191	0	0	0	0	0	0	1	-360	360;
	7879	4721	0.021711	0.063129	0	0	0	0	0	0	1	-360	360;
	7879	7569	0.013191	0.0428	0	0	0	0	0	0	1	-360	360;
	8098	2734	0.027631	0.102502	0	0	0	0	0	0	1	-360	360;
	8098	7358	0.005369	0.052622	0	0	0	0	0	0	1	-360	360;
	8098	7358	0.013938	0.0534	0	0	0	0	0	0	1	-360	360;
	8098	7358	0.01364	0.049449	0	0	0	0	0	0	1	-360	360;
	8098	1407	0.007991	0.037738	0	0	0	0	0	0	1	-360	360;
	2233	4721	0.007671	0.04096	0	0	0	0	0	0	1	-360	360;
	2233	4721	0.007511	0.041769	0	0	0	0	0	0	1	-360	360;
	2233	3429	0.004089	0.010978	0	0	0	0	0	0	1	-360	360;
	2233	3429	0.003	0.014769	0	0	0	0	0	0	1	-360	360;
	2233	610	0.001351	0.00784	0	0	0	0	0	0	1	-360	360;
	2233	948	0.003671	0.022089	0	0	0	0	0	0	1	-360	360;
	2233	2954	0.002378	0.00772	0	0	0	0	0	0	1	-360	360;
	4721	610	0.006911	0.03268	0	0	0	0	0	0	1	-360	360;
	4721	2140	0.00944	0.03284	0	0	0	0	0	0	1	-360	360;
	4721	6512	0.009831	0.02848	0	0	0	0	0	0	1	-360	360;
	4721	948	0.003671	0.02092	0	0	0	0	0	0	1	-360	360;
	8789	2734	0.019378	0.10932	0	0	0	0	0	0	1	-360	360;
	8789	2734	0.021969	0.107129	0	0	0	0	0	0	1	-360	360;
	8789	5274	0.009151	0.063249	0	0	0	0	0	0	1	-360	360;
	8789	4116	0.029169	0.088102	0	0	0	0	0	0	1	-360	360;
	8789	4868	0.031351	0.115698	0	0	0	0	0	0	1	-360	360;
	8789	7358	0.04936	0.182018	0	0	0	0	0	0	1	-360	360;
	8789	6329	0.001609	0.016889	0	299	0	0	0	0	1	-360	360;
	8789	2817	0	0.000222	0	0	0	0	0	0	1	-360	360;
	702	8927	0.008698	0.033591	0	0	0	0	0	0	1	-360	360;
	5517	1862	0.01792	0.070169	0	0	0	0	0	0	1	-360	360;
	5517	8629	0.00308	0.008422	0	299	0	0	0	0	1	-360	360;
	5517	552	0.002031	0.012102	0	0	0	0	0	0	1	-360	360;
	5517	552	0.0068	0.03452	0	0	0	0	0	0	1	-360	360;
	5517	3791	0.01176	0.055871	0	0	0	0	0	0	1	-360	360;
	4344	6512	0.04856	0.135111	0	0	0	0	0	0	1	-360	360;
	2734	6188	0.039289	0.146129	0	0	0	0	0	0	1	-360	360;
	2734	6188	0.038471	0.148938	0	0	0	0	0	0	1	-360	360;
	2734	3413	0.010911	0.04116	0	0	0	0	0	0	1	-360	360;
	6188	1407	0.002898	0.017409	0	0	0	0	0	0	1	-360	360;
	5006	1250	0.00544	0.051431	0	0	0	0	0	0	1	-360	360;
	5006	5331	0.003791	0.014342	0	0	0	0	0	0	1	-360	360;
	5006	5927	0.005969	0.026302	0	0	0	0	0	0	1	-360	360;
	5006	4771	0.00608	0.02572	0	0	0	0	0	0	1	-360	360;
	5006	2660	0.010818	0.066769	0	260	0	0	0	0	1	-360	360;
	1862	2407	0.006649	0.024138	0	0	0	0	0	0	1	-360	360;
	1862	2407	0.007542	0.023662	0	0	0	0	0	0	1	-360	360;
	1862	2407	0.007991	0.036622	0	0	0	0	0	0	1	-360	360;
	1862	2407	0.0116	0.046018	0	0	0	0	0	0	1	-360	360;
	4116	2140	0.011631	0.047049	0	0	0	0	0	0	1	-360	360;
	4868	7358	0.018822	0.064009	0	0	0	0	0	0	1	-360	360;
	1250	5927	0.008071	0.049138	0	0	0	0	0	0	1	-360	360;
	1250	4771	0.007889	0.045502	0	0	0	0	0	0	1	-360	360;
	1250	8278	0.009822	0.057222	0	0	0	0	0	0	1	-360	360;
	2407	3711	0.00656	0.023831	0	0	0	0	0	0	1	-360	360;
	2407	3711	0.007431	0.023351	0	0	0	0	0	0	1	-360	360;
	2407	7411	0.007391	0.026978	0	0	0	0	0	0	1	-360	360;
	2407	4277	0.001871	0.012142	0	0	0	0	0	0	1	-360	360;
	2407	987	0.003969	0.026169	0	0	0	0	0	0	1	-360	360;
	2407	2206	0.008151	0.04664	0	299	0	0	0	0	1	-360	360;
	8197	39	0.00076	0.005622	0	0	0	0	0	0	1	-360	360;
	8197	3711	0.0094	0.035791	0	0	0	0	0	0	1	-360	360;
	8197	7792	0.00844	0.031031	0	0	0	0	0	0	1	-360	360;
	8197	1685	0.00028	0.002062	0	0	0	0	0	0	1	-360	360;
	8197	2894	0.004849	0.030031	0	0	0	0	0	0	1	-360	360;
	8197	4277	0.00268	0.022391	0	0	0	0	0	0	1	-360	360;
	8197	987	0.002391	0.025382	0	0	0	0	0	0	1	-360	360;
	552	3711	0.017551	0.078191	0	0	0	0	0	0	1	-360	360;
	6504	1543	0.003849	0.02475	0	591	0	0	0	0	1	-360	360;
	6504	7041	0.0042	0.02005	0	0	0	0	0	0	1	-360	360;
	6504	3585	0.00011	0.000581	0	0	0	0	0	0	1	-360	360;
	6504	780	0.000169	0.000651	0	0	0	0	0	0	1	-360	360;
	6504	3957	0.000169	0.000659	0	0	0	0	0	0	1	-360	360;
	2692	6254	0.00219	0.01125	0	495	0	0	0	0	1	-360	360;
	2692	8957	0.00018	0.000599	0	495	0	0	0	0	1	-360	360;
	2692	8861	0.00018	0.000591	0	400	0	0	0	0	1	-360	360;
	2692	1734	0.00019	0.000581	0	438	0	0	0	0	1	-360	360;
	2692	3266	0.000169	0.000651	0	495	0	0	0	0	1	-360	360;
	322	1974	0.00231	0.0138	0	457	0	0	0	0	1	-360	360;
	322	1974	0.00082	0.003841	0	0	0	0	0	0	1	-360	360;
	322	177	0.00305	0.01762	0	495	0	0	0	0	1	-360	360;
	5662	5333	0.01101	0.055469	0	476	0	0	0	0	1	-360	360;
	5662	7901	0.009169	0.0462	0	533	0	0	0	0	1	-360	360;
	5662	7041	0.00201	0.00905	0	0	0	0	0	0	1	-360	360;
	5662	2006	0.005969	0.038841	0	553	0	0	0	0	1	-360	360;
	5662	7761	0.00474	0.02687	0	438	0	0	0	0	1	-360	360;
	5662	5559	0.005409	0.03438	0	476	0	0	0	0	1	-360	360;
	5662	3211	0.005401	0.03264	0	572	0	0	0	0	1	-360	360;
	5662	8963	0.00311	0.015919	0	610	0	0	0	0	1	-360	360;
	5662	1770	0.00462	0.02613	0	419	0	0	0	0	1	-360	360;
	3186	1109	0.01044	0.05187	0	476	0	0	0	0	1	-360	360;
	3186	1109	0.01001	0.05113	0	476	0	0	0	0	1	-360	360;
	3186	8493	0.008331	0.051729	0	533	0	0	0	0	1	-360	360;
	3186	6280	0.000169	0.00089	0	438	0	0	0	0	1	-360	360;
	3186	3987	0.00214	0.01099	0	495	0	0	0	0	1	-360	360;
	1974	5825	0.004669	0.042349	0	476	0	0	0	0	1	-360	360;
	1974	7252	0.00139	0.008789	0	419	0	0	0	0	1	-360	360;
	1974	7252	0.000219	0.001669	0	495	0	0	0	0	1	-360	360;
	1974	177	0.00119	0.01063	0	0	0	0	0	0	1	-360	360;
	2410	2006	0.004169	0.02724	0	514	0	0	0	0	1	-360	360;
	2410	2006	0.004081	0.025219	0	495	0	0	0	0	1	-360	360;
	2410	315	0.00168	0.01039	0	533	0	0	0	0	1	-360	360;
	2410	1489	0.0017	0.01055	0	495	0	0	0	0	1	-360	360;
	8926	5797	0.00139	0.008771	0	495	0	0	0	0	1	-360	360;
	8926	6381	0.000651	0.00261	0	476	0	0	0	0	1	-360	360;
	8926	6381	0.000729	0.00256	0	476	0	0	0	0	1	-360	360;
	8926	1239	0	0.008169	0	400	0	0	0	0	1	-360	360;
	8926	5643	5e-05	0.00025	0	438	0	0	0	0	1	-360	360;
	5672	7901	0.003719	0.022669	0	514	0	0	0	0	1	-360	360;
	5672	6415	0.00732	0.042711	0	476	0	0	0	0	1	-360	360;
	5672	6415	0.003599	0.025331	0	0	0	0	0	0	1	-360	360;
	5672	8998	0.00494	0.027401	0	533	0	0	0	0	1	-360	360;
	5333	6795	0.0062	0.03519	0	533	0	0	0	0	1	-360	360;
	5333	6795	0.00739	0.04574	0	438	0	0	0	0	1	-360	360;
	5333	3211	0.006781	0.034289	0	572	0	0	0	0	1	-360	360;
	6795	3054	0.003669	0.03205	0	0	0	0	0	0	1	-360	360;
	6795	3054	0.002789	0.026289	0	0	0	0	0	0	1	-360	360;
	6795	2562	0.005979	0.031979	0	533	0	0	0	0	1	-360	360;
	3054	6254	0.00425	0.02132	0	553	0	0	0	0	1	-360	360;
	3054	1543	0.002469	0.015099	0	0	0	0	0	0	1	-360	360;
	3054	8957	0.004341	0.02662	0	514	0	0	0	0	1	-360	360;
	3054	3266	0.005219	0.02287	0	610	0	0	0	0	1	-360	360;
	3054	7615	0.0095	0.04693	0	495	0	0	0	0	1	-360	360;
	5825	177	0.01099	0.052909	0	457	0	0	0	0	1	-360	360;
	5825	6415	0.004901	0.02949	0	533	0	0	0	0	1	-360	360;
	5825	4744	0.00832	0.038531	0	457	0	0	0	0	1	-360	360;
	5825	5669	0.003909	0.018091	0	381	0	0	0	0	1	-360	360;
	5825	3416	0.000331	0.00189	0	476	0	0	0	0	1	-360	360;
	7901	6415	0.01101	0.05514	0	495	0	0	0	0	1	-360	360;
	7901	7503	0.00345	0.01925	0	457	0	0	0	0	1	-360	360;
	8459	177	0.00162	0.00889	0	0	0	0	0	0	1	-360	360;
	8459	8998	0.00039	0.00464	0	0	0	0	0	0	1	-360	360;
	8459	5361	0.00064	0.010099	0	0	0	0	0	0	1	-360	360;
	6254	5797	0.003151	0.01613	0	572	0	0	0	0	1	-360	360;
	6254	5797	0.002091	0.012409	0	629	0	0	0	0	1	-360	360;
	6254	9136	0	0.00799	0	0	0	0	0	0	1	-360	360;
	6254	4650	0	0.00906	0	0	0	0	0	0	1	-360	360;
	6254	7037	0.000909	0.010401	0	0	0	0	0	0	1	-360	360;
	6254	2497	0.003211	0.016081	0	591	0	0	0	0	1	-360	360;
	6254	8861	0.00314	0.017711	0	457	0	0	0	0	1	-360	360;
	6254	1734	0.00214	0.010591	0	476	0	0	0	0	1	-360	360;
	3478	3211	0.00056	0.003169	0	0	0	0	0	0	1	-360	360;
	7252	7236	0.001719	0.0108	0	514	0	0	0	0	1	-360	360;
	5797	3954	0.0005	0.00174	0	514	0	0	0	0	1	-360	360;
	5797	5643	0.00126	0.00587	0	419	0	0	0	0	1	-360	360;
	6960	2562	0.007909	0.03607	0	533	0	0	0	0	1	-360	360;
	6960	7037	0.00057	0.00457	0	0	0	0	0	0	1	-360	360;
	6960	5643	0.00132	0.008409	0	400	0	0	0	0	1	-360	360;
	177	2460	0.003581	0.017289	0	610	0	0	0	0	1	-360	360;
	177	4744	0.002711	0.01526	0	476	0	0	0	0	1	-360	360;
	177	6311	0.00039	0.006081	0	0	0	0	0	0	1	-360	360;
	177	5361	0.00037	0.005409	0	0	0	0	0	0	1	-360	360;
	177	7236	0.00244	0.013789	0	476	0	0	0	0	1	-360	360;
	3954	2497	0.000349	0.00149	0	533	0	0	0	0	1	-360	360;
	2460	6415	0.00212	0.010531	0	591	0	0	0	0	1	-360	360;
	380	5339	0.003039	0.031151	0	0	0	0	0	0	1	-360	360;
	380	5669	0.001831	0.0187	0	0	0	0	0	0	1	-360	360;
	1848	3932	0.006599	0.043169	0	438	0	0	0	0	1	-360	360;
	1848	5094	0.006729	0.046599	0	438	0	0	0	0	1	-360	360;
	1848	3623	0.002159	0.010789	0	476	0	0	0	0	1	-360	360;
	1848	785	0.0022	0.01057	0	476	0	0	0	0	1	-360	360;
	1848	1036	0.003461	0.02049	0	0	0	0	0	0	1	-360	360;
	8021	698	0.00311	0.019469	0	495	0	0	0	0	1	-360	360;
	8021	628	0.01064	0.065969	0	495	0	0	0	0	1	-360	360;
	1109	315	0.003539	0.020479	0	495	0	0	0	0	1	-360	360;
	1109	1489	0.00357	0.020719	0	495	0	0	0	0	1	-360	360;
	1109	3932	0.00275	0.013831	0	457	0	0	0	0	1	-360	360;
	1109	5094	0.0028	0.013531	0	457	0	0	0	0	1	-360	360;
	6415	5669	0.005151	0.04851	0	0	0	0	0	0	1	-360	360;
	6415	8493	0.005401	0.025281	0	533	0	0	0	0	1	-360	360;
	6415	8493	0.0055	0.02475	0	514	0	0	0	0	1	-360	360;
	7503	5559	0.00093	0.00538	0	438	0	0	0	0	1	-360	360;
	7041	2006	0.010331	0.061849	0	553	0	0	0	0	1	-360	360;
	4712	3623	0.00686	0.032771	0	476	0	0	0	0	1	-360	360;
	4712	785	0.007	0.032081	0	476	0	0	0	0	1	-360	360;
	5339	4170	0.007729	0.038099	0	0	0	0	0	0	1	-360	360;
	4170	698	0.006581	0.035281	0	572	0	0	0	0	1	-360	360;
	4170	698	0.003711	0.02651	0	533	0	0	0	0	1	-360	360;
	4170	4333	0.00482	0.021979	0	572	0	0	0	0	1	-360	360;
	4170	4333	0.003979	0.026099	0	533	0	0	0	0	1	-360	360;
	2006	8963	0.0038	0.017271	0	553	0	0	0	0	1	-360	360;
	2006	7661	0.00626	0.03626	0	629	0	0	0	0	1	-360	360;
	8493	628	0.00538	0.02588	0	572	0	0	0	0	1	-360	360;
	8493	628	0.005479	0.025341	0	553	0	0	0	0	1	-360	360;
	130	3080	0.00134	0.01235	0	1218	0	0	0	0	1	-360	360;
	130	536	0.00127	0.01399	0	1349	0	0	0	0	1	-360	360;
	130	4393	0.00142	0.01374	0	1349	0	0	0	0	1	-360	360;
	7827	7969	0.00039	0.00365	0	1218	0	0	0	0	1	-360	360;
	7827	6760	0.00058	0.00507	0	1283	0	0	0	0	1	-360	360;
	1025	6298	0.00129	0.01577	0	1349	0	0	0	0	1	-360	360;
	1025	3007	0.00033	0.00373	0	1316	0	0	0	0	1	-360	360;
	1025	3265	0.00086	0.00799	0	1415	0	0	0	0	1	-360	360;
	1339	3265	0.00189	0.02135	0	1448	0	0	0	0	1	-360	360;
	1339	1774	0.00106	0.01219	0	1218	0	0	0	0	1	-360	360;
	1339	2899	0.00147	0.01622	0	1316	0	0	0	0	1	-360	360;
	1339	1703	0.00215	0.02108	0	1349	0	0	0	0	1	-360	360;
	441	8088	0.00063	0.0052	0	1382	0	0	0	0	1	-360	360;
	441	3265	0.00159	0.01792	0	1119	0	0	0	0	1	-360	360;
	6298	1774	0.00157	0.01981	0	1251	0	0	0	0	1	-360	360;
	6298	2899	0.00158	0.02014	0	1481	0	0	0	0	1	-360	360;
	8088	3080	0.0018	0.02388	0	1316	0	0	0	0	1	-360	360;
	8088	544	0.00034	0.00318	0	1349	0	0	0	0	1	-360	360;
	8088	7079	0.00015	0.00132	0	757	0	0	0	0	1	-360	360;
	8088	8601	0.00014	0.0015	0	757	0	0	0	0	1	-360	360;
	8088	1155	0.00013	0.00153	0	757	0	0	0	0	1	-360	360;
	8088	3256	0.00013	0.00156	0	757	0	0	0	0	1	-360	360;
	8088	5815	0.0004	0.00449	0	1283	0	0	0	0	1	-360	360;
	8088	5815	0.00045	0.0044	0	1152	0	0	0	0	1	-360	360;
	3080	1899	0.0023	0.01984	0	1185	0	0	0	0	1	-360	360;
	3080	9032	0.0002	0.00235	0	757	0	0	0	0	1	-360	360;
	3080	6004	0.00019	0.00239	0	757	0	0	0	0	1	-360	360;
	3080	7810	0.00169	0.01602	0	1316	0	0	0	0	1	-360	360;
	3080	8630	0.00024	0.00253	0	790	0	0	0	0	1	-360	360;
	3080	8183	0.00022	0.00244	0	757	0	0	0	0	1	-360	360;
	3080	6390	0.00025	0.00239	0	757	0	0	0	0	1	-360	360;
	8224	7969	0.00066	0.0061	0	1316	0	0	0	0	1	-360	360;
	8224	7969	0.00062	0.00844	0	1316	0	0	0	0	1	-360	360;
	8224	7860	0.00124	0.01353	0	1580	0	0	0	0	1	-360	360;
	8224	6187	0.00046	0.00452	0	1283	0	0	0	0	1	-360	360;
	3007	3265	0.00045	0.00531	0	1349	0	0	0	0	1	-360	360;
	3007	544	0.00123	0.01433	0	1152	0	0	0	0	1	-360	360;
	3007	7810	0.00174	0.01696	0	1448	0	0	0	0	1	-360	360;
	4807	6187	0.00086	0.00877	0	1382	0	0	0	0	1	-360	360;
	4807	6760	0.00096	0.00999	0	1218	0	0	0	0	1	-360	360;
	4807	4926	0.00082	0.00761	0	757	0	0	0	0	1	-360	360;
	7969	1703	0.00185	0.01803	0	1283	0	0	0	0	1	-360	360;
	7969	2031	0.00138	0.01246	0	1481	0	0	0	0	1	-360	360;
	7969	7980	0.00062	0.00605	0	1514	0	0	0	0	1	-360	360;
	7969	6760	0.00073	0.00817	0	757	0	0	0	0	1	-360	360;
	536	7509	0.00094	0.00999	0	1316	0	0	0	0	1	-360	360;
	536	3990	0.00058	0.00671	0	1218	0	0	0	0	1	-360	360;
	1774	1703	0.00108	0.01279	0	1349	0	0	0	0	1	-360	360;
	1774	7980	0.00225	0.02472	0	1382	0	0	0	0	1	-360	360;
	1774	4543	0.00019	0.00244	0	757	0	0	0	0	1	-360	360;
	1774	6340	0.00018	0.00248	0	757	0	0	0	0	1	-360	360;
	1774	261	0.0001	0.0011	0	757	0	0	0	0	1	-360	360;
	1774	2058	0.00011	0.00108	0	757	0	0	0	0	1	-360	360;
	7860	6187	0.00089	0.00815	0	1283	0	0	0	0	1	-360	360;
	6080	2899	0.0009	0.00859	0	1415	0	0	0	0	1	-360	360;
	6080	7509	0.00167	0.02239	0	1547	0	0	0	0	1	-360	360;
	2899	5263	2e-05	0.00024	0	757	0	0	0	0	1	-360	360;
	1703	138	0.00157	0.01326	0	1349	0	0	0	0	1	-360	360;
	1899	5562	0.00155	0.0139	0	1349	0	0	0	0	1	-360	360;
	138	2031	0.00142	0.01472	0	1349	0	0	0	0	1	-360	360;
	4704	2031	0.00187	0.02232	0	1514	0	0	0	0	1	-360	360;
	4704	1570	0.00049	0.00572	0	1349	0	0	0	0	1	-360	360;
	2031	1999	0	0.00021	0	757	0	0	0	0	1	-360	360;
	7509	3990	0.00104	0.00939	0	1349	0	0	0	0	1	-360	360;
	7509	7935	0.00043	0.0049	0	757	0	0	0	0	1	-360	360;
	7509	7935	0.00049	0.0048	0	757	0	0	0	0	1	-360	360;
	39	1174	0.002009	0.01464	0	0	0	0	0	0	1	-360	360;
	1174	3711	0.003302	0.026	0	0	0	0	0	0	1	-360	360;
	389	4259	0.006222	0.018089	0	0	0	0	0	0	1	-360	360;
	3711	7411	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3711	2894	8.9e-05	0.000911	0	0	0	0	0	0	1	-360	360;
	8771	7792	1.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	6459	7411	0.00388	0.014289	0	286	0	0	0	0	1	-360	360;
	4013	610	0.001329	0.007898	0	0	0	0	0	0	1	-360	360;
	4013	4577	0.00048	0.002951	0	0	0	0	0	0	1	-360	360;
	4013	8007	8e-05	0.00032	0	0	0	0	0	0	1	-360	360;
	2853	5331	0.000302	0.001102	0	0	0	0	0	0	1	-360	360;
	6512	304	0.003742	0.013098	0	0	0	0	0	0	1	-360	360;
	7569	6135	0.008911	0.032298	0	0	0	0	0	0	1	-360	360;
	5927	6540	0.00124	0.005409	0	0	0	0	0	0	1	-360	360;
	698	7104	0.0025	0.0123	0	400	0	0	0	0	1	-360	360;
	698	7104	0.001781	0.013729	0	400	0	0	0	0	1	-360	360;
	698	4333	0.00324	0.016039	0	553	0	0	0	0	1	-360	360;
	8998	6311	0.00037	0.00474	0	0	0	0	0	0	1	-360	360;
	7761	1770	9.1e-05	0.000521	0	419	0	0	0	0	1	-360	360;
	6311	5361	3.1e-05	0.00043	0	0	0	0	0	0	1	-360	360;
	1770	7709	0.001281	0.005909	0	0	0	0	0	0	1	-360	360;
	1489	5790	0.000909	0.00476	0	0	0	0	0	0	1	-360	360;
	5094	5266	0.00011	0.000599	0	0	0	0	0	0	1	-360	360;
	3623	7927	0.000289	0.001719	0	0	0	0	0	0	1	-360	360;
	7236	4098	0.00026	0.001591	0	514	0	0	0	0	1	-360	360;
	5815	3990	0.00108	0.01382	0	1481	0	0	0	0	1	-360	360;
	7329	7661	0	0.0002	0	572	0	0	0	0	1	-360	360;
	9024	6542	0.000742	0.067219	0	0	0	0	0.994359	0	1	-360	360;
	9024	6542	0.000793	0.06388	0	0	0	0	0.998418	0	1	-360	360;
	6069	9192	0.000609	0.046809	0	691	0	0	0.933053	0	1	-360	360;
	6069	1968	0.00059	0.04521	0	691	0	0	0.91223	0	1	-360	360;
	7829	1968	0.519194	5.108058	0	0	0	0	0	0	1	-360	360;
	8574	2299	0.000559	0.039982	0	0	0	0	0.919523	0	1	-360	360;
	8574	5587	0.000537	0.045312	0	0	0	0	0.948262	0	1	-360	360;
	4929	1815	0.000889	0.061549	0	0	0	0	0.996867	0	1	-360	360;
	1815	792	1.467822	7.764888	0	0	0	0	0	0	1	-360	360;
	6704	4586	0.000586	0.049031	0	658	0	0	0.923407	0	1	-360	360;
	2441	3506	0.000417	0.047801	0	658	0	0	0.952413	0	1	-360	360;
	1367	6826	0.000755	0.05143	0	0	0	0	0.924009	0	1	-360	360;
	1676	228	0.000733	0.052399	0	658	0	0	0.934986	0	1	-360	360;
	7829	6826	0.00078	0.064311	0	0	0	0	0	0	1	-360	360;
	659	3279	0.00089	0.048632	0	0	0	0	0.987002	0	1	-360	360;
	5210	792	0.00855	0.08972	0	0	0	0	0	0	1	-360	360;
	1445	792	0.00223	0.04836	0	0	0	0	0	0	1	-360	360;
	7180	2168	0.000581	0.043111	0	658	0	0	0.924038	0	1	-360	360;
	6833	8964	0.000579	0.043272	0	691	0	0	0.901005	0	1	-360	360;
	3659	3242	0.000586	0.049031	0	691	0	0	0.923407	0	1	-360	360;
	8181	4427	0.000559	0.049192	0	0	0	0	0.924009	0	1	-360	360;
	9025	7563	0.000573	0.051129	0	658	0	0	0.934986	0	1	-360	360;
	2267	5210	0.000417	0.039229	0	724	0	0	0.967059	0	1	-360	360;
	5210	7279	0.230542	1.197733	0	0	0	0	0	0	1	-360	360;
	5210	8335	0.02284	0.2644	0	0	0	0	0	0	1	-360	360;
	5210	2449	0.08537	0.736798	0	0	0	0	0	0	1	-360	360;
	5210	4665	1.402768	3.640289	0	0	0	0	0	0	1	-360	360;
	5210	4495	1.050227	4.450207	0	0	0	0	0	0	1	-360	360;
	8179	7279	0.000716	0.04371	0	0	0	0	0.934986	0	1	-360	360;
	1445	7279	0.518622	2.317689	0	0	0	0	0	0	1	-360	360;
	5509	8335	0.00059	0.045662	0	0	0	0	0.955958	0	1	-360	360;
	5509	1531	0.000603	0.050959	0	0	0	0	0.968125	0	1	-360	360;
	1445	8335	0.047769	0.554178	0	0	0	0	0	0	1	-360	360;
	7762	5155	0.000591	0.045153	0	0	0	0	0.918868	0	1	-360	360;
	7762	271	0.0006	0.047778	0	691	0	0	0.948976	0	1	-360	360;
	7829	5776	0.02436	0.270151	0	0	0	0	0	0	1	-360	360;
	8329	1445	0.000802	0.05727	0	0	0	0	0.967021	0	1	-360	360;
	8329	1445	0.000823	0.056059	0	0	0	0	0.967021	0	1	-360	360;
	1445	2449	0.188099	1.291901	0	0	0	0	0	0	1	-360	360;
	1445	4665	2.466323	7.26343	0	0	0	0	0	0	1	-360	360;
	2268	2908	0.000523	0.050352	0	724	0	0	0.934622	0	1	-360	360;
	8847	5776	0.000682	0.049638	0	0	0	0	0.913034	0	1	-360	360;
	8847	8103	0.00099	0.04736	0	0	0	0	1.012146	0	1	-360	360;
	317	2449	0.000616	0.048759	0	658	0	0	0.933053	0	1	-360	360;
	7637	8581	9e-05	0.015499	0	1678	0	0	0	-0.428189	1	-360	360;
	5848	7526	9e-05	0.00963	0	1678	0	0	0	0.178581	1	-360	360;
	89	4495	0.000664	0.045609	0	0	0	0	0.924009	0	1	-360	360;
	2520	4665	0.000649	0.044862	0	691	0	0	0.933053	0	1	-360	360;
	5996	8420	0.00067	0.04326	0	724	0	0	0.923407	0	1	-360	360;
	2154	5996	9e-05	0.009901	0	1678	0	0	0	-0.153178	1	-360	360;
	4241	2248	0.000214	0.018189	0	691	0	0	0.997518	0	1	-360	360;
	4717	5787	0.000545	0.028951	0	0	0	0	1.002304	0	1	-360	360;
	5425	7431	0.000433	0.016641	0	691	0	0	1.052632	0	1	-360	360;
	5425	7431	0.000271	0.019483	0	691	0	0	1.046093	0	1	-360	360;
	6632	3650	0.000292	0.041885	0	0	0	0	1.035986	0	1	-360	360;
	6632	3650	0.000283	0.025435	0	0	0	0	1.035986	0	1	-360	360;
	6917	4958	0.00019	0.020772	0	691	0	0	1.016388	0	1	-360	360;
	2962	8888	0.000567	0.0487	0	0	0	0	1.01961	0	1	-360	360;
	2962	8888	0.000292	0.024683	0	0	0	0	1.026095	0	1	-360	360;
	4542	7510	0.000238	0.020505	0	691	0	0	1.016388	0	1	-360	360;
	4542	7510	0.000302	0.024253	0	0	0	0	1.016399	0	1	-360	360;
	1549	3282	0.000394	0.025662	0	0	0	0	0.997518	0	1	-360	360;
	1549	3282	0.000403	0.026263	0	0	0	0	0.997518	0	1	-360	360;
	1549	3282	0.000358	0.02845	0	0	0	0	0.997518	0	1	-360	360;
	1200	3549	0.000184	0.020004	0	691	0	0	1.016388	0	1	-360	360;
	6771	4806	0.000287	0.030676	0	0	0	0	1.026104	0	1	-360	360;
	1086	6443	0.000183	0.021447	0	691	0	0	0.997518	0	1	-360	360;
	1086	6443	0.000183	0.019858	0	691	0	0	0.997518	0	1	-360	360;
	9159	6499	0.000196	0.019264	0	691	0	0	0.997518	0	1	-360	360;
	9159	6499	0.000205	0.018521	0	691	0	0	0.997518	0	1	-360	360;
	5685	3378	0.000205	0.018131	0	691	0	0	0.997518	0	1	-360	360;
	693	6994	0.000224	0.01875	0	691	0	0	1.035986	0	1	-360	360;
	693	6994	0.000685	0.016451	0	691	0	0	1.032682	0	1	-360	360;
	5332	3633	0.00021	0.024641	0	691	0	0	0.961805	0	1	-360	360;
	5332	3633	0.00021	0.025112	0	691	0	0	0.961805	0	1	-360	360;
	8034	4020	0	0.00554	0	1514	0	0	0	0	1	-360	360;
	1985	1023	0	0.006182	0	895	0	0	0.969385	0.248079	1	-360	360;
	477	6769	0.000205	0.016639	0	691	0	0	0.997518	0	1	-360	360;
	477	6769	0.000214	0.017035	0	691	0	0	0.997518	0	1	-360	360;
	477	6769	0.000214	0.016674	0	691	0	0	0.997518	0	1	-360	360;
	2188	5170	0.000224	0.018397	0	691	0	0	1.016388	0	1	-360	360;
	3061	699	0.000522	0.027299	0	0	0	0	1.006865	0	1	-360	360;
	3061	699	0.000646	0.037462	0	0	0	0	0.997538	0	1	-360	360;
	5823	7110	0.000244	0.033945	0	0	0	0	1.052632	0	1	-360	360;
	5823	7110	0.000196	0.023262	0	691	0	0	1.035986	0	1	-360	360;
	5280	7110	0.00091	0.001761	0	457	0	0	0.99993	-0.01179	1	-360	360;
	1947	4317	0.000179	0.016955	0	691	0	0	1.056356	0	1	-360	360;
	7303	429	0.000206	0.01883	0	691	0	0	1.016388	0	1	-360	360;
	6039	7973	0.000233	0.018474	0	691	0	0	1.006865	0	1	-360	360;
	6039	7973	0.000421	0.031827	0	0	0	0	1.014604	0	1	-360	360;
	1060	8981	0.000219	0.016973	0	691	0	0	1.026095	0	1	-360	360;
	8034	1752	0.000213	0.020004	0	691	0	0	0.997518	0	1	-360	360;
	739	1393	0.000556	0.033053	0	0	0	0	0.998185	0	1	-360	360;
	2235	8512	0.000508	0.028375	0	0	0	0	0.998185	0	1	-360	360;
	3428	1439	0.000179	0.022003	0	691	0	0	1.006865	0	1	-360	360;
	2730	6127	0.000209	0.02019	0	691	0	0	0.970491	0	1	-360	360;
	2730	6127	0.000214	0.020311	0	691	0	0	0.970491	0	1	-360	360;
	7395	5435	0.000616	0.030505	0	0	0	0	0.991996	0	1	-360	360;
	7395	5435	0.000626	0.029369	0	0	0	0	0.991996	0	1	-360	360;
	2437	3792	0.000214	0.019409	0	691	0	0	0.988343	0	1	-360	360;
	2437	3792	0.000192	0.022	0	691	0	0	0.979336	0	1	-360	360;
	5239	8438	0.000531	0.03466	0	0	0	0	0.991996	0	1	-360	360;
	5873	8147	0.000531	0.035228	0	0	0	0	0.99508	0	1	-360	360;
	5873	8147	0.00049	0.031535	0	0	0	0	0.995061	0	1	-360	360;
	5873	8147	0.000209	0.020964	0	691	0	0	0.988343	0	1	-360	360;
	7508	1240	0.000393	0.02663	0	0	0	0	1.035986	0	1	-360	360;
	7508	1240	0.000402	0.026088	0	0	0	0	1.035986	0	1	-360	360;
	3268	6397	0.000201	0.017787	0	691	0	0	1.006865	0	1	-360	360;
	4290	797	0.000289	0.018121	0	691	0	0	1.033257	0	1	-360	360;
	7929	4592	0.000643	0.024464	0	0	0	0	1.066843	0	1	-360	360;
	7929	4592	0.000568	0.027771	0	0	0	0	1.066843	0	1	-360	360;
	849	3840	0.000506	0.031833	0	0	0	0	0.998185	0	1	-360	360;
	849	3840	0.000196	0.024118	0	691	0	0	0.997518	0	1	-360	360;
	5919	8871	0.000205	0.022593	0	691	0	0	0.997518	0	1	-360	360;
	5919	8871	0.000186	0.01968	0	691	0	0	0.997518	0	1	-360	360;
	5919	8871	0.000205	0.01997	0	691	0	0	0.997518	0	1	-360	360;
	1168	2686	0.000215	0.018787	0	691	0	0	1.016388	0	1	-360	360;
	1168	2686	0.000215	0.018388	0	691	0	0	1.016388	0	1	-360	360;
	3275	6243	0.000224	0.018834	0	691	0	0	0.997518	0	1	-360	360;
	3798	1121	0.000433	0.043546	0	0	0	0	1.052632	0	1	-360	360;
	5884	8207	0.000234	0.018757	0	691	0	0	0.996896	0	1	-360	360;
	5884	8207	0.000197	0.019207	0	691	0	0	0.991986	0	1	-360	360;
	5884	8207	0.000197	0.019495	0	691	0	0	0.991986	0	1	-360	360;
	6681	7819	0.00018	0.021226	0	691	0	0	1.046093	0	1	-360	360;
	6681	7819	0.000812	0.03849	0	0	0	0	1.046093	0	1	-360	360;
	2261	4204	0.000205	0.02035	0	691	0	0	1.056356	0	1	-360	360;
	731	3728	0.000593	0.031521	0	0	0	0	1.001309	0	1	-360	360;
	731	3728	0.000612	0.031123	0	0	0	0	1.001309	0	1	-360	360;
	731	3728	0.000427	0.024994	0	0	0	0	1.006865	0	1	-360	360;
	2764	415	0.000667	0.027207	0	0	0	0	1.026095	0	1	-360	360;
	2764	415	0.000621	0.030836	0	0	0	0	1.026095	0	1	-360	360;
	299	947	0.000192	0.020791	0	691	0	0	1.066843	0	1	-360	360;
	299	947	0.000192	0.02119	0	691	0	0	1.066843	0	1	-360	360;
	8806	32	0.000686	0.030987	0	0	0	0	1.026958	0	1	-360	360;
	3008	5670	0.000194	0.020764	0	691	0	0	0.988343	0	1	-360	360;
	179	1850	0.000717	0.038087	0	0	0	0	1.035986	0	1	-360	360;
	7918	6660	0.000206	0.01883	0	691	0	0	1.016388	0	1	-360	360;
	7918	77	0.000206	0.018439	0	691	0	0	1.016388	0	1	-360	360;
	7029	33	0.000219	0.017019	0	691	0	0	1.035986	0	1	-360	360;
	7029	33	0.000224	0.016641	0	691	0	0	1.035986	0	1	-360	360;
	6607	9039	0.000676	0.047085	0	0	0	0	1.026095	0	1	-360	360;
	6607	9039	0.000667	0.048014	0	0	0	0	1.016388	0	1	-360	360;
	7343	8338	0.000196	0.019215	0	691	0	0	0.997518	0	1	-360	360;
	3966	5890	0.000253	0.025528	0	0	0	0	1.011609	0	1	-360	360;
	8210	6936	0.00021	0.017884	0	691	0	0	1.016388	0	1	-360	360;
	8210	6936	0.00021	0.017518	0	691	0	0	1.016388	0	1	-360	360;
	1716	6544	0.000219	0.017358	0	691	0	0	1.006865	0	1	-360	360;
	1716	6544	0.000219	0.016973	0	691	0	0	1.006865	0	1	-360	360;
	3544	869	0.000295	0.027995	0	0	0	0	1.021673	0	1	-360	360;
	3544	869	0.000277	0.027486	0	0	0	0	1.026115	0	1	-360	360;
	3662	8242	0.000489	0.032074	0	0	0	0	0.998185	0	1	-360	360;
	6092	3109	0.00018	0.02129	0	691	0	0	1.020734	0	1	-360	360;
	1074	1769	0.000366	0.029147	0	0	0	0	1.000611	0	1	-360	360;
	1074	1769	0.000366	0.028067	0	0	0	0	1.002235	0	1	-360	360;
	1074	1769	0.00042	0.026331	0	0	0	0	1.000611	0	1	-360	360;
	4957	1513	0.000441	0.01916	0	691	0	0	1.012394	0	1	-360	360;
	7428	6134	0.000391	0.024766	0	0	0	0	0.988343	0	1	-360	360;
	7428	6134	0.000401	0.024225	0	0	0	0	0.997518	0	1	-360	360;
	7428	4651	0.000408	0.028237	0	0	0	0	1.006865	0	1	-360	360;
	6970	4632	0.000199	0.018609	0	691	0	0	0.997518	0	1	-360	360;
	2700	8953	0.000196	0.021597	0	691	0	0	0.997518	0	1	-360	360;
	2700	8953	0.000186	0.022003	0	691	0	0	0.997518	0	1	-360	360;
	2849	3291	0.000186	0.019718	0	691	0	0	1.006865	0	1	-360	360;
	4446	1457	0.000205	0.022703	0	691	0	0	1.016388	0	1	-360	360;
	814	7820	0.000228	0.021908	0	691	0	0	0.988343	0	1	-360	360;
	7436	7629	0.000322	0.028734	0	0	0	0	0.996603	0	1	-360	360;
	7436	9106	0.000319	0.026574	0	0	0	0	1.042861	0	1	-360	360;
	3587	1241	0.000289	0.018298	0	691	0	0	1.046093	0	1	-360	360;
	3587	1241	0.000199	0.02111	0	691	0	0	1.046093	0	1	-360	360;
	7688	8301	0.000196	0.02119	0	691	0	0	1.016388	0	1	-360	360;
	7688	8301	0.000196	0.021597	0	691	0	0	1.016388	0	1	-360	360;
	4507	1503	0.001097	0.082091	0	0	0	0	1.006865	0	1	-360	360;
	4507	1503	0.000406	0.048834	0	0	0	0	1.033257	0	1	-360	360;
	4507	1503	7.2e-05	0.017798	0	691	0	0	0.988343	0	1	-360	360;
	4507	1503	0.000424	0.04684	0	0	0	0	1.033257	0	1	-360	360;
	8260	4081	0.000224	0.01875	0	691	0	0	1.006865	0	1	-360	360;
	8260	4081	0.000224	0.018334	0	691	0	0	1.006865	0	1	-360	360;
	1280	4279	0.000379	0.018493	0	691	0	0	1.046093	0	1	-360	360;
	1280	4279	0.000253	0.019711	0	691	0	0	1.046093	0	1	-360	360;
	385	9028	0.000196	0.016483	0	1053	0	0	0.997518	0	1	-360	360;
	385	9028	0.000186	0.022421	0	691	0	0	0.997518	0	1	-360	360;
	6768	4096	0.00046	0.030794	0	0	0	0	1.002506	0	1	-360	360;
	9203	2426	0.000695	0.049098	0	0	0	0	0.982143	0	1	-360	360;
	9203	2426	0.000355	0.024893	0	691	0	0	0.93617	0	1	-360	360;
	118	2426	0.000365	0.024372	0	691	0	0	0.979227	0	1	-360	360;
	3145	7770	0.000411	0.023652	0	691	0	0	0.93617	0	1	-360	360;
	7267	4918	0.000386	0.025673	0	691	0	0	0.93617	0	1	-360	360;
	7267	4918	0.000344	0.029336	0	691	0	0	0.93617	0	1	-360	360;
	839	4506	0.000653	0.052104	0	0	0	0	0.982143	0	1	-360	360;
	839	4506	0.000767	0.043531	0	0	0	0	0.982143	0	1	-360	360;
	3445	6114	0.000323	0.028314	0	691	0	0	0.93617	0	1	-360	360;
	3445	6114	0.000323	0.025999	0	691	0	0	0.93617	0	1	-360	360;
	3445	7264	0.000365	0.025478	0	691	0	0	0.93617	0	1	-360	360;
	891	6306	0.00086	0.039291	0	0	0	0	0.982143	0	1	-360	360;
	720	1711	0.000355	0.026586	0	691	0	0	0.93617	0	1	-360	360;
	720	1711	0.000375	0.026313	0	691	0	0	0.93617	0	1	-360	360;
	6206	3240	0.000344	0.023329	0	691	0	0	0.979227	0	1	-360	360;
	6206	3240	0.000302	0.026375	0	691	0	0	0.979227	0	1	-360	360;
	9174	2924	0.000788	0.04323	0	0	0	0	0.982143	0	1	-360	360;
	9174	2924	0.000767	0.04407	0	0	0	0	0.982143	0	1	-360	360;
	1355	5856	0.000767	0.043562	0	0	0	0	0.982143	0	1	-360	360;
	1355	5856	0.000788	0.041281	0	0	0	0	0.982143	0	1	-360	360;
	1355	5856	0.000892	0.040524	0	0	0	0	0.982143	0	1	-360	360;
	3205	4683	0.000975	0.043562	0	0	0	0	0.982143	0	1	-360	360;
	3205	4683	0.000757	0.042649	0	0	0	0	0.982143	0	1	-360	360;
	5350	2361	0.000261	0.030762	0	691	0	0	0.93617	0	1	-360	360;
	5350	2361	0.000261	0.030088	0	691	0	0	0.93617	0	1	-360	360;
	5016	6112	0.000323	0.026447	0	691	0	0	0.979227	0	1	-360	360;
	5016	8860	0.000323	0.026885	0	691	0	0	0.979227	0	1	-360	360;
	7471	549	0.000323	0.02748	0	691	0	0	0.93617	0	1	-360	360;
	7471	549	0.000292	0.028158	0	691	0	0	0.93617	0	1	-360	360;
	549	5002	0	0.009197	0	667	0	0	0	0.072386	1	-360	360;
	5490	3817	0.000736	0.044526	0	0	0	0	0.982143	0	1	-360	360;
	2854	9112	0.00086	0.039488	0	0	0	0	0.982143	0	1	-360	360;
	2918	6532	0.000365	0.024372	0	691	0	0	0.93617	0	1	-360	360;
	2918	6532	0.000375	0.023986	0	691	0	0	0.93617	0	1	-360	360;
	4494	7284	0.000386	0.025673	0	691	0	0	0.93617	0	1	-360	360;
	2656	1775	0.000809	0.042318	0	0	0	0	0.982143	0	1	-360	360;
	3519	5641	0.000705	0.042038	0	0	0	0	0.982143	0	1	-360	360;
	3543	367	0.000334	0.028022	0	691	0	0	0.93617	0	1	-360	360;
	3543	1833	0.000353	0.022403	0	691	0	0	0.982143	0	1	-360	360;
	2372	367	0.000313	0.026354	0	691	0	0	0.93617	0	1	-360	360;
	2372	367	0.000323	0.025478	0	691	0	0	0.93617	0	1	-360	360;
	1876	7752	0.000344	0.024622	0	691	0	0	0.979227	0	1	-360	360;
	113	7752	0.000355	0.024038	0	691	0	0	0.979227	0	1	-360	360;
	3499	7289	0.000881	0.038244	0	0	0	0	0.982143	0	1	-360	360;
	3499	7289	0.000757	0.040959	0	0	0	0	0.982143	0	1	-360	360;
	3082	2083	0.000354	0.028765	0	691	0	0	0.93617	0	1	-360	360;
	6224	2083	0.000354	0.027114	0	691	0	0	0.93617	0	1	-360	360;
	2174	2083	0.000695	0.04748	0	0	0	0	0.982143	0	1	-360	360;
	4623	1609	0.000323	0.028314	0	691	0	0	0.93617	0	1	-360	360;
	4623	1609	0.000323	0.026354	0	691	0	0	0.93617	0	1	-360	360;
	4623	1609	0.000323	0.025696	0	691	0	0	0.93617	0	1	-360	360;
	5589	1001	0.000355	0.027305	0	691	0	0	0.979227	0	1	-360	360;
	5589	1001	0.000355	0.026666	0	691	0	0	0.979227	0	1	-360	360;
	5589	1001	0.000923	0.037456	0	0	0	0	0.982143	0	1	-360	360;
	5589	1001	0.000788	0.035217	0	0	0	0	0.982143	0	1	-360	360;
	1198	7473	0.000323	0.026739	0	691	0	0	0.93617	0	1	-360	360;
	1198	7473	0.000323	0.02749	0	691	0	0	0.93617	0	1	-360	360;
	2479	5482	0.000313	0.027324	0	691	0	0	0.93617	0	1	-360	360;
	2479	5482	0.000313	0.028074	0	691	0	0	0.93617	0	1	-360	360;
	7831	5525	0.000601	0.024984	0	0	0	0	0.982143	0	1	-360	360;
	7831	5525	0.000715	0.022797	0	0	0	0	0.982143	0	1	-360	360;
	7831	5525	0.000344	0.024893	0	691	0	0	0.93617	0	1	-360	360;
	4970	7955	0.000365	0.024163	0	691	0	0	0.979227	0	1	-360	360;
	4970	7955	0.000344	0.02385	0	691	0	0	0.979227	0	1	-360	360;
	5007	7367	0.000757	0.040182	0	0	0	0	0.982143	0	1	-360	360;
	757	3758	0.000323	0.028937	0	691	0	0	0.979227	0	1	-360	360;
	3697	6952	0.000354	0.026958	0	691	0	0	0.93617	0	1	-360	360;
	3697	6952	0.000778	0.045656	0	0	0	0	0.982143	0	1	-360	360;
	3697	6952	0.000695	0.048932	0	0	0	0	0.982143	0	1	-360	360;
	3697	6952	0.000302	0.025926	0	691	0	0	0.93617	0	1	-360	360;
	6738	8180	0.000355	0.027328	0	691	0	0	0.93617	0	1	-360	360;
	6738	8180	0.000365	0.027156	0	691	0	0	0.93617	0	1	-360	360;
	9222	5935	0.000355	0.024299	0	691	0	0	0.979227	0	1	-360	360;
	7988	804	0.000375	0.026313	0	691	0	0	0.93617	0	1	-360	360;
	7988	804	0.000375	0.025605	0	691	0	0	0.93617	0	1	-360	360;
	3239	1539	0.000334	0.026291	0	691	0	0	0.979227	0	1	-360	360;
	3239	1539	0.000705	0.046735	0	0	0	0	0.982143	0	1	-360	360;
	2056	4084	0.000313	0.026875	0	691	0	0	0.979227	0	1	-360	360;
	2056	4084	0.000302	0.027376	0	691	0	0	0.979227	0	1	-360	360;
	4598	5110	0.000664	0.044868	0	0	0	0	0.982143	0	1	-360	360;
	2129	5110	0.000809	0.044526	0	0	0	0	0.982143	0	1	-360	360;
	5658	5550	0.00086	0.034253	0	0	0	0	0.982143	0	1	-360	360;
	5658	5550	0.000355	0.024372	0	691	0	0	0.93617	0	1	-360	360;
	2848	1526	0.000365	0.023788	0	691	0	0	0.93617	0	1	-360	360;
	4541	7464	0.000375	0.023204	0	691	0	0	0.979227	0	1	-360	360;
	4541	7464	0.000302	0.026593	0	691	0	0	0.979227	0	1	-360	360;
	1754	3857	0.000323	0.029256	0	691	0	0	0.93617	0	1	-360	360;
	1754	3857	0.000334	0.029986	0	691	0	0	0.93617	0	1	-360	360;
	1754	3857	0.000355	0.030979	0	691	0	0	0.93617	0	1	-360	360;
	2518	4544	0.000302	0.025707	0	691	0	0	0.93617	0	1	-360	360;
	2518	4544	0.000344	0.025405	0	691	0	0	0.93617	0	1	-360	360;
	3069	6115	0.00012	0.008908	0	591	0	0	0	-0.072388	1	-360	360;
	5067	7641	0.000365	0.023715	0	691	0	0	0.93617	0	1	-360	360;
	6036	8670	0.000365	0.023517	0	691	0	0	0.979227	0	1	-360	360;
	6036	5522	0.000334	0.026812	0	691	0	0	0.979227	0	1	-360	360;
	8486	8791	0.000695	0.047253	0	0	0	0	0.982143	0	1	-360	360;
	7530	8791	0.000778	0.043759	0	0	0	0	0.982143	0	1	-360	360;
	7522	7571	0.000313	0.030636	0	691	0	0	0.93617	0	1	-360	360;
	7522	7571	0.000313	0.028286	0	691	0	0	0.93617	0	1	-360	360;
	7522	7571	0.000344	0.028115	0	691	0	0	0.93617	0	1	-360	360;
	2526	1483	0.000902	0.039425	0	0	0	0	0.982143	0	1	-360	360;
	2365	8316	0.000355	0.024237	0	691	0	0	0.93617	0	1	-360	360;
	8672	1860	0.000375	0.023986	0	691	0	0	0.93617	0	1	-360	360;
	8672	1860	0.000375	0.023861	0	691	0	0	0.93617	0	1	-360	360;
	1494	583	0.000819	0.04153	0	0	0	0	0.982143	0	1	-360	360;
	1494	583	0.000809	0.042701	0	0	0	0	0.982143	0	1	-360	360;
	1502	2795	0.000334	0.028022	0	691	0	0	0.93617	0	1	-360	360;
	8195	933	0.000632	0.048237	0	0	0	0	0.982143	0	1	-360	360;
	8195	933	0.000632	0.045159	0	0	0	0	0.982143	0	1	-360	360;
	8195	933	0.000715	0.044039	0	0	0	0	0.982143	0	1	-360	360;
	3918	6802	0.00074	0.044479	0	0	0	0	0.986547	0	1	-360	360;
	3918	4025	0.00074	0.043811	0	0	0	0	0.986547	0	1	-360	360;
	4435	2079	0.000375	0.02385	0	691	0	0	0.93617	0	1	-360	360;
	4435	2079	0.000365	0.023267	0	691	0	0	0.93617	0	1	-360	360;
	4435	2079	0.000323	0.026739	0	691	0	0	0.93617	0	1	-360	360;
	1081	7361	0.000684	0.047025	0	0	0	0	0.982143	0	1	-360	360;
	9101	2177	0.000313	0.02748	0	691	0	0	0.93617	0	1	-360	360;
	6555	8475	0.000323	0.028554	0	691	0	0	0.93617	0	1	-360	360;
	6555	8475	0.000302	0.026437	0	691	0	0	0.93617	0	1	-360	360;
	7056	7577	0.000365	0.025551	0	691	0	0	0.93617	0	1	-360	360;
	7056	7577	0.000344	0.024893	0	691	0	0	0.93617	0	1	-360	360;
	7056	9051	0.000375	0.024508	0	691	0	0	0.979227	0	1	-360	360;
	6889	6785	0.000778	0.041157	0	0	0	0	0.982143	0	1	-360	360;
	6889	6785	0.000923	0.03702	0	0	0	0	0.982143	0	1	-360	360;
	6889	6785	0.000302	0.026593	0	691	0	0	0.979227	0	1	-360	360;
	1027	7148	0.000344	0.029667	0	691	0	0	0.93617	0	1	-360	360;
	1027	44	0.000334	0.02748	0	691	0	0	0.93617	0	1	-360	360;
	5308	7148	0.000323	0.030636	0	691	0	0	0.89674	0	1	-360	360;
	5308	44	0.000323	0.028526	0	691	0	0	0.93617	0	1	-360	360;
	9213	6231	0.000344	0.025332	0	691	0	0	0.979227	0	1	-360	360;
	9213	8291	0.000334	0.024674	0	691	0	0	0.979227	0	1	-360	360;
	9213	6842	0.000344	0.02457	0	691	0	0	0.979227	0	1	-360	360;
	3608	6664	0.000365	0.023788	0	691	0	0	0.93617	0	1	-360	360;
	4339	6570	0.000365	0.023601	0	691	0	0	0.93617	0	1	-360	360;
	4339	6570	0.000344	0.02652	0	691	0	0	0.93617	0	1	-360	360;
	7513	6570	0.000344	0.027272	0	691	0	0	0.93617	0	1	-360	360;
	7513	6570	0.000334	0.027563	0	691	0	0	0.93617	0	1	-360	360;
	8887	6639	0.000313	0.030636	0	691	0	0	0.979227	0	1	-360	360;
	8887	6639	0.000313	0.028286	0	691	0	0	0.979227	0	1	-360	360;
	5461	4816	0.000881	0.036803	0	0	0	0	0.982143	0	1	-360	360;
	5461	4816	0.000902	0.036668	0	0	0	0	0.982143	0	1	-360	360;
	5461	4816	0.000881	0.034014	0	0	0	0	0.982143	0	1	-360	360;
	5460	3786	0.000365	0.026027	0	691	0	0	0.93617	0	1	-360	360;
	5460	3786	0.000386	0.025673	0	691	0	0	0.93617	0	1	-360	360;
	7256	4491	0.00012	0.011186	0	686	0	0	0	0.084507	1	-360	360;
	6153	58	0.00017	0.007091	0	1316	0	0	0	-0.09784	1	-360	360;
	6153	6807	0.000664	0.047107	0	0	0	0	0.982143	0	1	-360	360;
	1465	1262	0.000313	0.028314	0	691	0	0	0.93617	0	1	-360	360;
	1465	1262	0.000334	0.025707	0	691	0	0	0.93617	0	1	-360	360;
	1465	1551	0.000344	0.025478	0	691	0	0	0.93617	0	1	-360	360;
	7537	1551	0.000344	0.027236	0	691	0	0	0.93617	0	1	-360	360;
	7537	1551	0.000375	0.026518	0	691	0	0	0.93617	0	1	-360	360;
	6901	4874	0.000386	0.02646	0	691	0	0	0.93617	0	1	-360	360;
	5241	4	0.000788	0.040317	0	0	0	0	0.982143	0	1	-360	360;
	7520	561	0.000323	0.025863	0	691	0	0	0.979227	0	1	-360	360;
	749	4324	1e-05	0.022863	0	1048	0	0	0	0.04898	1	-360	360;
	4141	7124	0.000334	0.030306	0	691	0	0	0.93617	0	1	-360	360;
	4141	7124	0.000302	0.030636	0	691	0	0	0.93617	0	1	-360	360;
	7466	3649	0.00011	0.010476	0	572	0	0	0	-0.06138	1	-360	360;
	8222	6556	0.000344	0.024841	0	691	0	0	0.979227	0	1	-360	360;
	8222	6556	0.000365	0.025029	0	691	0	0	0.979227	0	1	-360	360;
	8222	6556	0.000355	0.024101	0	691	0	0	0.979227	0	1	-360	360;
	4125	163	0.000344	0.02385	0	691	0	0	0.93617	0	1	-360	360;
	4125	163	0.000411	0.02313	0	691	0	0	0.93617	0	1	-360	360;
	4125	163	0.000829	0.042764	0	0	0	0	0.982143	0	1	-360	360;
	5837	9217	0.000705	0.04748	0	0	0	0	0.982143	0	1	-360	360;
	5837	1448	0.000323	0.029986	0	691	0	0	0.93617	0	1	-360	360;
	5837	1448	0.000302	0.029952	0	691	0	0	0.93617	0	1	-360	360;
	8267	5365	0.000746	0.041064	0	0	0	0	0.982143	0	1	-360	360;
	8267	5365	0.000881	0.042214	0	0	0	0	0.982143	0	1	-360	360;
	2359	8005	0.00086	0.041209	0	0	0	0	0.982143	0	1	-360	360;
	2359	8005	0.000912	0.038669	0	0	0	0	0.982143	0	1	-360	360;
	8109	171	0.000375	0.026244	0	691	0	0	0.979227	0	1	-360	360;
	8109	171	0.000375	0.025457	0	691	0	0	0.979227	0	1	-360	360;
	3485	1156	0.000302	0.029096	0	691	0	0	0.93617	0	1	-360	360;
	4734	2229	0.000302	0.029667	0	691	0	0	0.93617	0	1	-360	360;
	1709	7702	0.000334	0.027563	0	691	0	0	0.979227	0	1	-360	360;
	1709	7702	0.000323	0.028074	0	691	0	0	0.979227	0	1	-360	360;
	905	8189	0.000736	0.041064	0	0	0	0	0.982143	0	1	-360	360;
	8334	8189	0.001234	0.040109	0	0	0	0	0.982143	0	1	-360	360;
	8334	8189	0.000757	0.043209	0	0	0	0	0.982143	0	1	-360	360;
	7998	6734	0.000323	0.026586	0	954	0	0	0.93617	0	1	-360	360;
	4196	280	0.000788	0.040296	0	0	0	0	0.982143	0	1	-360	360;
	4196	280	0.000912	0.040514	0	0	0	0	0.982143	0	1	-360	360;
	6475	3503	0.000798	0.042183	0	0	0	0	0.982143	0	1	-360	360;
	6475	3503	0.000778	0.043147	0	0	0	0	0.982143	0	1	-360	360;
	7328	4235	0.000323	0.027793	0	691	0	0	0.93617	0	1	-360	360;
	7328	4235	0.000302	0.027918	0	691	0	0	0.93617	0	1	-360	360;
	7328	4235	0.000302	0.025926	0	691	0	0	0.93617	0	1	-360	360;
	6921	2949	0.000344	0.024956	0	691	0	0	0.979227	0	1	-360	360;
	6921	2949	0.000355	0.024435	0	691	0	0	0.979227	0	1	-360	360;
	3830	306	0.000386	0.026392	0	691	0	0	0.93617	0	1	-360	360;
	5648	306	0.000365	0.025525	0	691	0	0	0.93617	0	1	-360	360;
	1730	115	0.000344	0.029986	0	691	0	0	0.93617	0	1	-360	360;
	1730	115	0.000323	0.029838	0	691	0	0	0.93617	0	1	-360	360;
	2721	5720	0.000323	0.02748	0	691	0	0	0.93617	0	1	-360	360;
	2721	5720	0.000313	0.028001	0	691	0	0	0.93617	0	1	-360	360;
	432	3246	0.000179	0.012988	0	1251	0	0	0.990991	0	1	-360	360;
	432	8846	0.000198	0.012724	0	1251	0	0	0.990991	0	1	-360	360;
	432	2393	0.000207	0.012469	0	1251	0	0	0.990991	0	1	-360	360;
	432	1895	0.000207	0.012206	0	1251	0	0	0.990991	0	1	-360	360;
	432	3412	0.000217	0.012912	0	1251	0	0	0.990991	0	1	-360	360;
	432	5586	0.000217	0.012627	0	1251	0	0	0.990991	0	1	-360	360;
	432	3112	0.000198	0.014317	0	1251	0	0	0.990991	0	1	-360	360;
	432	3112	0.000189	0.014592	0	1251	0	0	0.990991	0	1	-360	360;
	432	7857	0.000354	0.02725	0	691	0	0	0.979227	0	1	-360	360;
	960	8497	0.000628	0.042428	0	0	0	0	1.009174	0	1	-360	360;
	960	8497	0.000726	0.044039	0	0	0	0	0.982143	0	1	-360	360;
	3577	5907	0.000365	0.024893	0	691	0	0	0.979227	0	1	-360	360;
	3577	5907	0.000438	0.02457	0	691	0	0	0.979227	0	1	-360	360;
	3577	5907	0.000365	0.023986	0	691	0	0	0.979227	0	1	-360	360;
	7164	5918	0.000375	0.023267	0	691	0	0	0.979227	0	1	-360	360;
	7164	5918	0.000323	0.026375	0	691	0	0	0.93617	0	1	-360	360;
	1798	2467	0.000334	0.029735	0	691	0	0	0.93617	0	1	-360	360;
	1798	2467	0.000313	0.030237	0	691	0	0	0.93617	0	1	-360	360;
	3346	2132	0.000767	0.044049	0	0	0	0	0.982143	0	1	-360	360;
	3346	2132	0.000302	0.026072	0	691	0	0	0.93617	0	1	-360	360;
	3346	2132	0.00085	0.039892	0	0	0	0	0.982143	0	1	-360	360;
	3483	8992	0.000344	0.024956	0	691	0	0	0.979227	0	1	-360	360;
	8886	8992	0.000334	0.026666	0	691	0	0	0.93617	0	1	-360	360;
	2093	805	0.000365	0.023788	0	691	0	0	0.93617	0	1	-360	360;
	2093	805	0.000365	0.023465	0	691	0	0	0.93617	0	1	-360	360;
	221	1541	0.000334	0.026447	0	691	0	0	0.93617	0	1	-360	360;
	8347	5383	0.000323	0.027042	0	691	0	0	0.93617	0	1	-360	360;
	8347	5383	0.000313	0.02748	0	691	0	0	0.93617	0	1	-360	360;
	8347	5383	0.000302	0.028314	0	691	0	0	0.93617	0	1	-360	360;
	8347	7507	0.000323	0.025853	0	691	0	0	0.93617	0	1	-360	360;
	8347	7507	0.000365	0.026197	0	691	0	0	0.93617	0	1	-360	360;
	8347	7507	0.000365	0.025165	0	691	0	0	0.93617	0	1	-360	360;
	4000	4852	0.000355	0.024237	0	691	0	0	0.979227	0	1	-360	360;
	4000	2012	0.000375	0.024247	0	691	0	0	0.93617	0	1	-360	360;
	2183	4852	0.000365	0.023267	0	691	0	0	0.979227	0	1	-360	360;
	2183	2012	0.000342	0.023139	0	691	0	0	0.982143	0	1	-360	360;
	1642	7098	0.000344	0.026666	0	691	0	0	0.979227	0	1	-360	360;
	1642	687	0.000334	0.027021	0	691	0	0	0.93617	0	1	-360	360;
	809	7098	0.000302	0.027615	0	691	0	0	0.979227	0	1	-360	360;
	809	687	0.000313	0.026135	0	691	0	0	0.93617	0	1	-360	360;
	4594	8804	0.000365	0.025905	0	691	0	0	0.93617	0	1	-360	360;
	4594	8804	0.000355	0.025238	0	691	0	0	0.93617	0	1	-360	360;
	2919	4215	0.000334	0.026814	0	691	0	0	0.93617	0	1	-360	360;
	1392	5001	0.001252	0.073601	0	0	0	0	0.94	0	1	-360	360;
	3538	5001	0.001271	0.072001	0	0	0	0	0.94	0	1	-360	360;
	8249	6139	0.000442	0.040802	0	921	0	0	0.88	0	1	-360	360;
	806	91	0.00048	0.041601	0	0	0	0	0.97	0	1	-360	360;
	1747	6998	0.000543	0.04261	0	0	0	0	0	0	1	-360	360;
	593	6998	0.000534	0.04342	0	0	0	0	0	0	1	-360	360;
	8432	6295	0.003895	0.040564	0	0	0	0	0.97	0	1	-360	360;
	8432	6295	0.004413	0.039762	0	0	0	0	0.97	0	1	-360	360;
	8432	6295	0.004498	0.038954	0	0	0	0	0.97	0	1	-360	360;
	3190	7090	0.000621	0.0376	0	0	0	0	0	0	1	-360	360;
	3484	7090	0.00063	0.0368	0	0	0	0	0	0	1	-360	360;
	3484	7844	0.00064	0.036	0	0	0	0	0	0	1	-360	360;
	281	3960	0.000555	0.040801	0	0	0	0	0.984999	0	1	-360	360;
	8086	1477	0.000546	0.0416	0	0	0	0	0.984999	0	1	-360	360;
	8086	8037	0.000527	0.0424	0	0	0	0	0.984999	0	1	-360	360;
	5008	6354	0.000456	0.043201	0	0	0	0	0.97	0	1	-360	360;
	5008	6354	0.000447	0.040001	0	0	0	0	0.97	0	1	-360	360;
	919	6354	0.000505	0.039202	0	0	0	0	0.97	0	1	-360	360;
	919	6354	0.000513	0.038402	0	0	0	0	0.97	0	1	-360	360;
	4825	7884	0.000513	0.037602	0	0	0	0	0.954999	0	1	-360	360;
	3023	7884	0.00052	0.036802	0	0	0	0	0.954999	0	1	-360	360;
	6345	7884	0.000529	0.036002	0	0	0	0	0.954999	0	1	-360	360;
	8114	8738	0.000473	0.040802	0	0	0	0	0.984999	0	1	-360	360;
	8773	8738	0.000465	0.041601	0	0	0	0	0.984999	0	1	-360	360;
	4059	6282	7e-05	0.021201	0	790	0	0	0	0	1	-360	360;
	2002	4340	7e-05	0.021599	0	790	0	0	0	0	1	-360	360;
	5165	3363	0.00054	0.04	0	0	0	0	0.98421	0	1	-360	360;
	5165	3363	0.000472	0.039202	0	0	0	0	0.984999	0	1	-360	360;
	2722	3363	0.00048	0.038402	0	0	0	0	0.984999	0	1	-360	360;
	2722	3363	0.00064	0.0376	0	0	0	0	0.98421	0	1	-360	360;
	1591	1124	0.000333	0.029441	0	0	0	0	0.925	0	1	-360	360;
	5558	8510	0.000353	0.028801	0	0	0	0	0.925	0	1	-360	360;
	5558	8510	0.000312	0.032641	0	0	0	0	0.925	0	1	-360	360;
	1689	8814	0.000529	0.041601	0	0	0	0	0.94	0	1	-360	360;
	8230	8814	0.000511	0.042401	0	0	0	0	0.94	0	1	-360	360;
	8393	7565	0.000633	0.051521	0	0	0	0	0.94	0	1	-360	360;
	8393	7565	0.000616	0.047701	0	0	0	0	0.94	0	1	-360	360;
	139	7565	0.000701	0.046752	0	0	0	0	0.94	0	1	-360	360;
	3257	6058	0.000811	0.04579	0	0	0	0	0.98871	0	1	-360	360;
	369	6058	0.000831	0.04484	0	0	0	0	0.98871	0	1	-360	360;
	369	6058	0.000841	0.043881	0	0	0	0	0.98871	0	1	-360	360;
	6068	7450	0.000621	0.036001	0	0	0	0	1.015	0	1	-360	360;
	6068	2271	0.000555	0.040801	0	0	0	0	1.015	0	1	-360	360;
	6263	3846	0.000563	0.0416	0	0	0	0	1.015	0	1	-360	360;
	1788	1020	0.000543	0.0424	0	0	0	0	1.015	0	1	-360	360;
	2824	211	0.001026	0.086398	0	0	0	0	0.97	0	1	-360	360;
	140	211	0.001007	0.080001	0	0	0	0	0.955	0	1	-360	360;
	3210	1670	0.000522	0.039201	0	0	0	0	0.954999	0	1	-360	360;
	5674	1670	0.00053	0.038401	0	0	0	0	0.954999	0	1	-360	360;
	4337	1670	0.000548	0.037601	0	0	0	0	0.94	0	1	-360	360;
	4779	1956	3e-05	0.006499	0	0	0	0	0	0	1	-360	360;
	4320	4302	0.000907	0.042929	0	0	0	0	0.97	0	1	-360	360;
	4858	4248	7.7e-05	0.0204	0	724	0	0	1.105	0	1	-360	360;
	7235	4858	3e-05	0.007522	0	0	0	0	0.995481	-0.09511	1	-360	360;
	4248	4302	0.000773	0.05056	0	0	0	0	0.97	0	1	-360	360;
	5897	2720	0.000728	0.043197	0	0	0	0	1.045	0	1	-360	360;
	420	243	0.000447	0.040001	0	0	0	0	1.045	0	1	-360	360;
	8898	6052	0.00061	0.0392	0	0	0	0	1.03	0	1	-360	360;
	5311	8668	0.001203	0.076801	0	0	0	0	1.03	0	1	-360	360;
	3891	8668	0.001223	0.075201	0	0	0	0	1.03	0	1	-360	360;
	662	7073	0.004533	0.037356	0	0	0	0	0.984999	0	1	-360	360;
	6017	349	0.004615	0.036549	0	0	0	0	0.984999	0	1	-360	360;
	6766	7156	0.004113	0.041392	0	0	0	0	0.984999	0	1	-360	360;
	4077	7442	0.000627	0.041599	0	0	0	0	1.039999	0	1	-360	360;
	1052	8277	0.001048	0.084802	0	0	0	0	1.045	0	1	-360	360;
	3878	8277	0.001029	0.086397	0	0	0	0	1.045	0	1	-360	360;
	76	5346	0.000808	0.047709	0	0	0	0	0	0	1	-360	360;
	4485	3000	0.00061	0.0392	0	0	0	0	1.015	0	1	-360	360;
	4485	3000	0.00062	0.0384	0	0	0	0	1.015	0	1	-360	360;
	6283	3000	0.00064	0.0376	0	0	0	0	1.015	0	1	-360	360;
	395	4082	0.00063	0.0368	0	0	0	0	0.984999	0	1	-360	360;
	5843	5246	0.00064	0.036	0	0	0	0	0.984999	0	1	-360	360;
	5843	5246	0.000573	0.0408	0	0	0	0	0.984999	0	1	-360	360;
	6477	2478	0.001067	0.083199	0	0	0	0	0	0	1	-360	360;
	5295	2478	0.001048	0.084802	0	0	0	0	0	0	1	-360	360;
	6504	3341	0.001519	0.050701	0	0	0	0	1.013513	0	1	-360	360;
	130	5452	0.000821	0.051524	0	0	0	0	1.052632	0	1	-360	360;
	130	5452	0.000803	0.047869	0	0	0	0	1.052632	0	1	-360	360;
	441	2233	0.000466	0.029318	0	0	0	0	1.025291	0	1	-360	360;
	8088	4721	0.000463	0.028006	0	0	0	0	1.038781	0	1	-360	360;
	8088	4721	0.000463	0.026569	0	0	0	0	1.038781	0	1	-360	360;
	3080	8789	0.000556	0.030088	0	0	0	0	1.08147	0	1	-360	360;
	3080	8789	0.000581	0.029019	0	0	0	0	1.08147	0	1	-360	360;
	5825	702	0.0017	0.077979	0	0	0	0	0	0	1	-360	360;
	8224	5517	0.000569	0.046371	0	0	0	0	1.052632	0	1	-360	360;
	8224	5517	0.00056	0.047255	0	0	0	0	1.052632	0	1	-360	360;
	536	4344	0.0008	0.050159	0	0	0	0	1.066857	0	1	-360	360;
	6080	5006	0.000379	0.039836	0	0	0	0	1.052632	0	1	-360	360;
	6080	5006	0.000433	0.039042	0	0	0	0	1.052632	0	1	-360	360;
	1899	7358	0.000448	0.032183	0	0	0	0	1.066857	0	1	-360	360;
	4712	1250	0.000498	0.034255	0	0	0	0	1.041667	0	1	-360	360;
	4712	1250	0.000507	0.033509	0	0	0	0	1.041667	0	1	-360	360;
	6187	2407	0.000415	0.040631	0	0	0	0	1.052632	0	1	-360	360;
	6187	2407	0.000406	0.041434	0	0	0	0	1.052632	0	1	-360	360;
	5562	1407	0.000397	0.037986	0	0	0	0	1.052632	0	1	-360	360;
	7827	1974	0.000322	0.035886	0	0	0	0	1.043149	0	1	-360	360;
	1025	8926	0.000408	0.039074	0	0	0	0	0.989653	0	1	-360	360;
	6298	6795	0.000417	0.026115	0	0	0	0	1.062288	0	1	-360	360;
	6298	6795	0.000443	0.02577	0	0	0	0	1.062288	0	1	-360	360;
	3007	6254	0.000396	0.033918	0	0	0	0	1.02924	0	1	-360	360;
	7969	177	0.000406	0.023438	0	0	0	0	1.052632	0	1	-360	360;
	7969	177	0.000352	0.026759	0	0	0	0	1.052632	0	1	-360	360;
	1703	7503	0.000334	0.036651	0	0	0	0	1.052632	0	1	-360	360;
	1703	7503	0.000334	0.037354	0	0	0	0	1.052632	0	1	-360	360;
	4704	4170	0.000305	0.035675	0	0	0	0	1.087225	0	1	-360	360;
	4704	4170	0.000296	0.033035	0	0	0	0	1.087225	0	1	-360	360;
	4704	4170	0.000338	0.032376	0	0	0	0	1.087225	0	1	-360	360;
	2031	8493	0.000285	0.02604	0	0	0	0	1.199891	0	1	-360	360;
	1025	1239	0.000419	0.02814	0	0	0	0	0.989653	0	1	-360	360;
	3007	9136	0.000453	0.025129	0	0	0	0	1.02924	0	1	-360	360;
	3007	4650	0.000453	0.02528	0	0	0	0	1.02924	0	1	-360	360;
];

%%-----  OPF Data  -----%%
%% generator cost data
%	1	startup	shutdown	n	x1	y1	...	xn	yn
%	2	startup	shutdown	n	c(n-1)	...	c0
mpc.gencost = [
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
];
