function mpc = case9241pegase
%CASE9241PEGASE  Power flow data for European system.
%   Please see CASEFORMAT for details on the case file format.
%
%   This case accurately represents the size and complexity of the European
%   high voltage transmission network. The network contains 9,241 buses,
%   1,445 generators, and 16,049 branches and it operates at 750, 400, 380,
%   330, 220, 154, 150, 120, and 110 kV. Please note that the data are
%   fictitious and do not correspond to real world data. They can thus be
%   used to validate methods and tools but should not be used for operation
%   and planning of the European grid.
%
%   The data stems from the Pan European Grid Advanced Simulation and State
%   Estimation (PEGASE) project, part of the 7th Framework Program of the
%   European Union (https://www.fp7-pegase.com/).
%
%   When publishing results based on this data, please cite:
%
%     C. Josz, S. Fliscounakis, J. Maeght, and P. Panciatici, "AC Power Flow
%     Data in MATPOWER and QCQP Format: iTesla, RTE Snapshots, and PEGASE"
%     https://arxiv.org/abs/1603.01533
%
%     S. Fliscounakis, P. Panciatici, F. Capitanescu, and L. Wehenkel,
%     "Contingency ranking with respect to overloads in very large power
%     systems taking into account uncertainty, preventive and corrective
%     actions", Power Systems, IEEE Trans. on, (28)4:4909-4917, 2013.
%     https://doi.org/10.1109/TPWRS.2013.2251015
%
%   Remarks:
%
%   1. Line flow limits are the current flow limits found in PEGASE data.
%
%   2. PEGASE data contains asymmetric shunt conductance and susceptance in
%   the PI transmission line model of branches. Thus total line charging
%   susceptance of branches is set to 0 p.u. and the nodal representation
%   of shunt condutance and susceptance is used. As a result, power flow
%   equations are left unchanged compared with original PEGASE data.
%   However, line flow constraints in the optimal flow problem are
%   modified.
%
%   3. Identical linear costs are used for all generators to form a loss
%   minimizing OPF objective function.
%
%   4. Since some parts of the network are aggregated, some generators
%   (e.g. with negative PMIN) represent aggregations of multiple loads
%   and generators.
%
%   Contacts:
%     Cédric Josz, Stéphane Fliscounakis, Jean Maeght, Patrick Panciatici
%     firstname.lastname@rte-france.com
%     Réseau de Transport d'Electricité (French Transmission System Operator)
%     Département Expertise Système, Immeuble "Le Colbert"
%     9 rue de la Porte de Buc, 78000 Versailles Cedex, France
%
%   March 18th, 2015

%   MATPOWER
%   Copyright (c) 2015, 2016 by Cédric Josz, Stéphane Fliscounakis, Jean Maeght,
%   and Patrick Panciatici
%   Licensed under the Creative Commons Attribution 4.0 International license,
%   https://creativecommons.org/licenses/by/4.0/

%% MATPOWER Case Format : Version 2
mpc.version = '2';

%%-----  Power Flow Data  -----%%
%% system MVA base
mpc.baseMVA = 100;

%% bus data
%	bus_i	type	Pd	Qd	Gs	Bs	area	Vm	Va	baseKV	zone	Vmax	Vmin
mpc.bus = [
	1	1	378.4	98.5	0	16.61	0	1.007206	-36.856998	220	6	1.1	0.9;
	2	2	0	0	0	0	0	1.031734	-12.028001	154	1	1.1	0.9;
	3	1	151	48.8	0	4.69	0	1.017843	-21.794234	220	5	1.1	0.9;
	4	1	171.41	23.4	0	2.1	0	1.020147	-7.224254	220	5	1.1	0.9;
	5	1	45.95	14	0	13.34	0	0.989775	-49.871783	220	6	1.1	0.9;
	6	2	0	0	0	0	0	1.038476	-42.240491	380	6	1.1	0.9;
	7	1	31.03	10.3	0	0.65	0	0.96914	-36.181195	110	21	1.1	0.9;
	8	2	0	0	0	0	0	1.023005	-20.771666	154	1	1.1	0.9;
	9	1	14.3	2	0	1.91	0	1.032658	-3.651312	154	1	1.1	0.9;
	10	1	134	24.7	0	12.44	0	1.033477	-24.645036	220	5	1.1	0.9;
	11	1	0	0	0	0	0	0.965112	-31.467815	150	9	1.1	0.9;
	12	1	-6.11	2	0	0.47	0	1.034508	-26.383988	150	9	1.1	0.9;
	13	1	-22.74	-8.32	0	6.5	0	1.013859	-48.74904	220	6	1.1	0.9;
	14	1	13.5	14.8	0	132.435168	0	1.14146	-21.558785	380	9	1.3	0.7;
	15	1	59.67	26.7	0	12.84	0	1.02165	-48.932223	220	4	1.1	0.9;
	16	1	30.6	4.3	0	5.15	0	1.032431	6.063256	154	1	1.1	0.9;
	17	1	67.68	12.1	0	2.18	0	1.010381	6.154063	220	3	1.1	0.9;
	18	2	0	0	0	0	0	1.026728	18.107815	380	3	1.1	0.9;
	19	1	25.33	7.5	0	2.39	0	1.005109	-38.460832	110	12	1.1	0.9;
	20	1	1.4	0.2	0	4.93	0	1.024517	6.675769	154	1	1.1	0.9;
	21	1	161.2	39.3	0	30.71	0	1.031479	-22.305911	220	5	1.1	0.9;
	22	1	0	-0	0	0.85	0	1.031644	-18.751971	220	5	1.1	0.9;
	23	1	19.62	7.1	0	1.15	0	1.024755	-25.628654	110	19	1.1	0.9;
	24	1	17.22	5.7	0	1.22	0	0.960802	-32.296073	110	21	1.1	0.9;
	25	2	0	0	0	0	0	1.028526	12.404601	220	3	1.1	0.9;
	26	1	167.3	54.8	0	110.54	0	1.029582	-3.339851	380	5	1.1	0.9;
	27	1	-0	-0	0	14.82	0	1.085434	13.448823	380	13	1.1	0.9;
	28	1	0	0	0	0	0	1.004931	-31.840868	150	9	1.1	0.9;
	29	1	152.9	49.4	0	27.65	0	1.004602	-13.569928	380	5	1.1	0.9;
	30	1	-0	0	0	0.56	0	1.023518	-29.715225	150	9	1.1	0.9;
	31	1	69.68	43.2	0	6.3	0	0.905716	-35.509144	150	9	1.3	0.7;
	32	2	0	0	0	0	0	1.003223	-43.014699	220	4	1.1	0.9;
	33	2	0	0	0	0	0	1.026534	-38.549418	220	4	1.1	0.9;
	34	1	13.52	4.9	0	1.8	0	1.028523	-24.559588	110	19	1.1	0.9;
	35	1	10.51	3.5	0	0.31	0	1.026497	-36.721451	110	12	1.1	0.9;
	36	1	-0	-0	0.12	15.34	0	1.045193	-21.272765	220	19	1.1	0.9;
	37	1	9.21	2.6	0	2	0	0.972198	-29.508765	110	18	1.1	0.9;
	38	1	42.75	12.8	0	13.3	0	1.022868	25.558732	220	8	1.1	0.9;
	39	2	0	0	0	0	0	1.003511	-38.297997	150	10	1.1	0.9;
	40	1	16.02	4.54	0	0.1	0	0.999097	-31.840826	110	23	1.1	0.9;
	41	1	27.9	4	0	4.23	0	1.008625	-8.828964	154	1	1.1	0.9;
	42	1	20.82	7.9	0	26.89	0	1.042404	-49.184531	380	4	1.1	0.9;
	43	1	12.6	1.8	0	2.03	0	1.028739	-15.907171	154	1	1.1	0.9;
	44	1	-0	0	0	7.27	0	1.040868	-22.52467	220	5	1.1	0.9;
	45	1	10.11	3.7	0	0.81	0	1.048003	-25.286846	110	19	1.1	0.9;
	46	1	10.01	2.7	0	1.14	0	0.958698	-29.319698	110	18	1.1	0.9;
	47	1	6.31	0	0	0.67	0	0.985859	-27.824727	110	18	1.1	0.9;
	48	1	2.9	0.9	0	0.6	0	1.045484	-26.609146	110	15	1.1	0.9;
	49	2	0	0	0	0	0	1.020323	-36.088357	220	6	1.1	0.9;
	50	1	48.8	6.9	0	8.71	0	1.025009	-4.460831	154	1	1.1	0.9;
	51	2	0	0	0	0	0	1.006665	-48.120109	220	4	1.1	0.9;
	52	1	127.6	31.8	0	20.51	0	1.017252	0.511391	150	8	1.1	0.9;
	53	1	48.65	16.1	0	7.14	0	1.053406	-9.368997	220	5	1.1	0.9;
	54	1	110.5	-26.8	0	16.37	0	1.023883	-35.660013	220	6	1.1	0.9;
	55	1	0	-0	0	1.23	0	0.99682	9.004928	220	3	1.1	0.9;
	56	1	140	79.7	0	6.5	0	1.010606	-16.919749	220	3	1.1	0.9;
	57	1	21.6	3	0	3.54	0	1.006533	-9.693767	154	1	1.1	0.9;
	58	1	0	0	0	2.31	0	1.027113	-22.181789	380	5	1.1	0.9;
	59	1	119.2	20.7	0	3.31	0	1.026698	-24.169412	220	5	1.1	0.9;
	60	1	0.24	-1.66	0	0.65	0	1.024291	-19.933148	110	18	1.1	0.9;
	61	1	-0	0	0	58.57	0	1.092743	45.900184	380	3	1.1	0.9;
	62	1	-0	0	0	1.34	0	0.970166	-27.705319	110	18	1.1	0.9;
	63	1	187.9	110.2	0	21.25	0	1.020254	5.462635	380	3	1.1	0.9;
	64	1	14.92	5.5	0	1.31	0	1.058218	-24.075639	110	15	1.1	0.9;
	65	1	-18.52	6.1	0	0.79	0	0.920443	-33.979065	150	9	1.1	0.9;
	66	1	4.61	1.7	0	2.32	0	1.07583	-17.389672	110	19	1.1	0.9;
	67	1	0	0	0.099501	82.133595	0	1.076797	-22.668386	380	18	1.1	0.9;
	68	1	10.81	5.1	0	1.09	0	0.932272	-32.555245	110	18	1.1	0.9;
	69	1	62.87	20.5	0	1.15	0	1.016399	-0.749163	150	8	1.1	0.9;
	70	1	0	0	0	2.64	0	1.021842	-7.278252	154	1	1.1	0.9;
	71	1	126.3	42.4	0	55.24	0	1.045653	2.671707	150	8	1.1	0.9;
	72	1	6.91	-4.6	0	5.75	0	1.007342	-41.162683	220	4	1.1	0.9;
	73	1	15.72	5.2	0	0.11	0	1.011319	-38.957605	110	12	1.1	0.9;
	74	1	0	0	0	0	0	1.048486	-29.24882	150	9	1.1	0.9;
	75	1	38.74	6.1	0	1.17	0	1.019038	-53.244715	220	4	1.1	0.9;
	76	1	22.13	10.7	0	14.21	0	1.022914	25.568654	220	8	1.1	0.9;
	77	1	-0	-0	0	0.73	0	1.006664	-49.599091	220	4	1.1	0.9;
	78	1	0	-0	0	0.83	0	0.981839	-20.873081	110	18	1.1	0.9;
	79	1	138.8	136.7	0	8.76	0	1.014386	13.838092	380	3	1.1	0.9;
	80	2	0	0	0	0	0	1.068393	-26.943073	220	23	1.1	0.9;
	81	1	67.76	19.18	0	1.47	0	1.010926	61.408155	150	1	1.1	0.9;
	82	1	22.02	5.1	0	0.69	0	1.03239	-34.385172	110	1	1.1	0.9;
	83	1	0	0	0	6.62	0	0.980327	13.445474	220	3	1.1	0.9;
	84	1	1	0.3	0	0.19	0	1.018715	-30.235601	110	19	1.1	0.9;
	85	1	20.9	3	0	5.74	0	1.022071	16.413645	154	1	1.1	0.9;
	86	1	0	0	0	0.12	0	1.051084	-24.673587	150	9	1.1	0.9;
	87	1	21.52	5.1	0	0.5	0	1.01789	-15.037836	110	18	1.1	0.9;
	88	1	71.98	28.8	0	11	0	1.006747	-37.071358	220	6	1.1	0.9;
	89	1	0	0	0.199111	4.530599	0	0.991009	-5.004163	380	2	1.1	0.9;
	90	1	363.25	115.39	0	3.95	0	1.042035	-27.99494	220	5	1.1	0.9;
	91	1	-0	0	0	0.02	0	1.027005	0.21051	150	8	1.1	0.9;
	92	1	192.9	73.3	0	20.75	0	1.022158	3.828439	380	3	1.1	0.9;
	93	1	17.52	6.3	0	0.26	0	1.059099	-21.280687	110	19	1.1	0.9;
	94	1	33.14	12	0	0.65	0	1.007744	-25.49367	110	19	1.1	0.9;
	95	1	40.45	25	0	9.82	0	0.980011	-30.352301	150	9	1.1	0.9;
	96	1	193.17	53.37	0	4.58	0	1.036787	-15.459067	220	5	1.1	0.9;
	97	1	0	0	0	9.38	0	1.053303	2.669719	380	1	1.1	0.9;
	98	1	90.7	25.5	0	3.11	0	1.048113	-2.631892	220	3	1.1	0.9;
	99	1	176.3	-58.9	0	24.77	0	1.02505	-33.267073	380	6	1.1	0.9;
	100	1	35.94	22.3	0	0	0	1.033873	-29.927627	150	9	1.1	0.9;
	101	1	-0	0	0	0.29	0	0.964847	-31.479281	150	9	1.1	0.9;
	102	1	0	-0	0	3.77	0	1.004007	-50.623198	220	6	1.1	0.9;
	103	1	65.07	40.3	0	5.43	0	0.992015	-34.266143	150	9	1.1	0.9;
	104	1	51.36	-8.8	0	2.15	0	0.965443	-55.943839	220	6	1.1	0.9;
	105	1	-0	-0	0	82.55	0	1.036504	-34.250878	380	6	1.1	0.9;
	106	1	0	-0	0	31.19	0	1.162506	-22.492638	380	23	1.3	0.7;
	107	2	0	0	0	0	0	1.083135	41.615508	380	3	1.3	0.7;
	108	1	3.8	1.3	0	0.4	0	1.005191	-17.950996	110	18	1.1	0.9;
	109	1	137	19.5	0	4.9	0	1.004858	-11.902933	154	1	1.1	0.9;
	110	1	0	0	0	0	0	0.994253	-32.082481	150	9	1.1	0.9;
	111	1	61.07	36	0	3.09	0	1.049848	-22.296332	220	11	1.1	0.9;
	112	1	0	0	0	3.24	0	1.030495	-9.592049	220	3	1.1	0.9;
	113	1	-0	-0	0	6.532711	0	1.027407	-14.491581	380	5	1.1	0.9;
	114	1	55.7	7.9	0	4.21	0	1.026508	9.781802	154	1	1.1	0.9;
	115	1	0	-0	0	4.21	0	1.062005	-20.086677	220	5	1.1	0.9;
	116	1	1.8	0.7	0	0.3	0	1.037818	-21.171048	110	19	1.1	0.9;
	117	1	40.04	16	0	0.52	0	1.002542	-24.312333	110	18	1.1	0.9;
	118	1	-0	0	0	36.562834	0	1.023863	-11.320452	380	5	1.1	0.9;
	119	1	0	0	0	0.09	0	1.066627	-21.228028	150	9	1.1	0.9;
	120	1	140.6	20.1	0	13.81	0	1.017689	-24.512143	154	1	1.1	0.9;
	121	1	45.15	14.1	0	0.19	0	1.061687	-25.751773	110	15	1.1	0.9;
	122	1	-0	0	0	5.75	0	1.002609	8.035022	220	3	1.1	0.9;
	123	2	0	0	0	0	0	1.041503	-19.868767	220	11	1.1	0.9;
	124	2	0	0	0	0	0	1.077112	8.310105	220	5	1.1	0.9;
	125	1	3.42	2.68	0	1.46	0	1.022215	-19.969192	110	18	1.1	0.9;
	126	1	0	-0	0	2.64	0	1.009839	3.639019	154	1	1.1	0.9;
	127	1	-0	-0	0	4.3	0	1.021779	-28.832789	220	6	1.1	0.9;
	128	1	23.73	12.7	0	0.52	0	1.015182	-16.413952	220	5	1.1	0.9;
	129	1	121.9	7.4	0	1.34	0	1.019572	-12.355618	220	3	1.1	0.9;
	130	1	-0	0	0.135375	47.750675	0	1.057946	-47.3148	380	10	1.1	0.9;
	131	1	47.35	16.4	0	4.08	0	1.03239	-40.262984	110	22	1.1	0.9;
	132	1	0	-0	0	21.39	0	1.062408	-43.935749	380	1	1.1	0.9;
	133	1	38.14	12.5	0	6.28	0	1.020783	2.35489	150	8	1.1	0.9;
	134	1	54.2	7.8	0	94.9	0	1.012932	6.186446	400	1	1.1	0.9;
	135	1	130.6	84.9	0	4.22	0	1.021481	-22.444323	220	7	1.1	0.9;
	136	1	7.91	2.6	0	0.74	0	1.006509	-36.386405	110	12	1.1	0.9;
	137	1	42.1	74.61	0	37.58	0	1.003064	61.688917	150	1	1.1	0.9;
	138	1	-75.38	7.8	0	29.9	0	1.067481	-41.992026	380	10	1.1	0.9;
	139	1	-0	-0	0	36.92	0	1.013268	16.764642	380	8	1.1	0.9;
	140	1	0	0	0	4.14	0	1.027453	17.039361	220	8	1.1	0.9;
	141	2	0	0	0	0	0	1.050716	-2.165585	220	13	1.1	0.9;
	142	2	0	0	0	0	0	1.042299	-30.404497	220	6	1.1	0.9;
	143	1	0	0	0	0	0	1.022343	3.87531	150	8	1.1	0.9;
	144	1	0	0	0	98.53	0	1.063798	18.932224	380	14	1.1	0.9;
	145	1	38.34	5.6	0	7.64	0	1.038236	-8.964288	220	5	1.1	0.9;
	146	1	269.6	65.7	0	0	0	1.085135	-25.638373	120	20	1.3	0.7;
	147	2	0	0	0	0	0	1.058041	-13.496004	220	7	1.1	0.9;
	148	1	32.34	12	0	6.1	0	1.034625	-8.217967	220	5	1.1	0.9;
	149	1	100.1	45.1	0	1.11	0	1.021632	-40.902503	380	6	1.1	0.9;
	150	2	0	0	0	0	0	1.062129	2.399597	220	5	1.1	0.9;
	151	2	0	0	0	0	0	1.013748	-24.076149	154	1	1.1	0.9;
	152	1	7	1	0	4.69	0	1.050332	-5.822456	154	1	1.1	0.9;
	153	1	-0	0	0	0.56	0	1.020345	-28.197731	150	7	1.1	0.9;
	154	1	0	0	0	1.29	0	1.016529	-40.246021	110	12	1.1	0.9;
	155	1	-0	-0	0	1.7	0	1.019829	-29.30153	110	21	1.1	0.9;
	156	1	40.34	25	0	1.14	0	0.953098	-31.912298	150	9	1.1	0.9;
	157	1	9.21	3.3	0	1.2	0	1.01451	-13.069704	110	19	1.1	0.9;
	158	1	166	60.3	0	0.57	0	1.01467	6.148661	220	3	1.1	0.9;
	159	1	0	-0	0	0.1	0	1.021326	-29.687582	150	9	1.1	0.9;
	160	1	79.4	11.3	0	4.92	0	1.031014	0.809761	154	1	1.1	0.9;
	161	1	-0	-0	0	0.16	0	1.091016	-27.88035	150	9	1.1	0.9;
	162	1	0	0	0	66.65	0	1.052312	-15.685671	380	6	1.1	0.9;
	163	1	147.76	32.73	0	8.34	0	1.045445	-10.873592	220	5	1.1	0.9;
	164	1	0	-0	0	5.17	0	1.02082	-1.152459	150	8	1.1	0.9;
	165	1	26.8	-3.7	0	51.261678	0	1.073155	-24.141412	380	9	1.1	0.9;
	166	2	0	0	0	0	0	1.040747	11.781552	220	3	1.1	0.9;
	167	1	257.5	131.1	0	1.08	0	1.031121	-4.856472	220	3	1.1	0.9;
	168	1	16.7	2.4	0	6.09	0	1.033177	12.044899	154	1	1.1	0.9;
	169	1	0	0	0	1.91	0	1.045543	-29.415558	150	9	1.1	0.9;
	170	1	-0	-0	0	2.35	0	1.011462	34.762397	220	1	1.1	0.9;
	171	1	100.8	13.2	0	5.24	0	1.061976	-25.502124	220	5	1.1	0.9;
	172	1	111.7	10.96	0	4.22	0	1.010114	-47.141199	220	4	1.1	0.9;
	173	1	90.38	-5.05	0	8.49	0	1.054447	14.634452	220	13	1.1	0.9;
	174	1	370.5	-17	0	9.91	0	1.071451	-28.241418	220	5	1.1	0.9;
	175	1	9.91	3.2	0	0.67	0	1.003654	-41.36559	110	12	1.1	0.9;
	176	1	-0.69	-5.46	0	5.07	0	1.040842	-28.533618	150	9	1.1	0.9;
	177	1	94.9	36.2	0	16.46	0	0.969767	-43.357076	220	10	1.1	0.9;
	178	1	81.09	48	0	0.03	0	1.013199	-30.419221	150	9	1.1	0.9;
	179	2	0	0	0	0	0	1.043661	-50.052119	380	4	1.1	0.9;
	180	1	0	0	0	0.54	0	1.02953	-28.102331	110	21	1.1	0.9;
	181	1	54.96	16.9	0	3.37	0	1.040515	-39.643018	110	22	1.1	0.9;
	182	1	19.32	26	0	0.69	0	1.03553	39.324365	220	3	1.1	0.9;
	183	1	0	0	0	1.83	0	1.031452	-29.237157	150	9	1.1	0.9;
	184	1	-0	-0	0	2.42	0	1.037524	-12.250638	220	5	1.1	0.9;
	185	1	20.9	3	0	4.11	0	1.007995	-7.232267	154	1	1.1	0.9;
	186	1	19.92	6.6	0	0.56	0	1.009472	-30.519212	110	21	1.1	0.9;
	187	1	57.16	2	0	1.48	0	1.002263	-38.456187	110	22	1.1	0.9;
	188	1	36.84	13.9	0	3.3	0	1.015193	-19.834342	220	5	1.1	0.9;
	189	1	-15.04	-5.12	0	0.43	0	1.001228	-27.784349	110	18	1.1	0.9;
	190	2	0	0	0	0	0	1.025544	-21.44389	110	19	1.1	0.9;
	191	1	0	-0	0	14.19	0	1.016435	-36.476155	220	6	1.1	0.9;
	192	2	0	0	0	0	0	1.048355	66.012239	380	3	1.1	0.9;
	193	1	119.9	3.7	0	3.33	0	1.031896	24.418857	220	3	1.1	0.9;
	194	2	0	0	0	0	0	1.02931	13.231568	150	8	1.1	0.9;
	195	1	73.14	21.55	0	8.52	0	1.048915	-9.187839	220	5	1.1	0.9;
	196	1	0	-0	0	12.48	0	1.069098	-26.717314	220	5	1.1	0.9;
	197	2	0	0	0	0	0	1.099985	44.827022	380	3	1.1	0.9;
	198	1	39.74	11.8	0	1.44	0	1.011257	-16.658556	220	5	1.1	0.9;
	199	1	86.8	28.4	0	1.34	0	1.017154	0.540558	150	8	1.1	0.9;
	200	1	22.42	2.1	0	0.19	0	1.010977	61.566699	150	1	1.1	0.9;
	201	2	0	0	0	0	0	1.01914	-51.449761	220	4	1.1	0.9;
	202	1	110.9	32.7	0	15.4	0	1.028758	-0.941811	220	14	1.1	0.9;
	203	1	18.2	2.6	0	12.38	0	1.034855	13.22078	154	1	1.1	0.9;
	204	1	-0	-0	0.181549	14.690889	0	1.054685	-23.538774	220	18	1.1	0.9;
	205	1	46.15	18.2	0	1.29	0	1.017882	4.444704	150	8	1.1	0.9;
	206	1	34.04	12.3	0	0.23	0	1.062901	-21.256948	110	19	1.1	0.9;
	207	1	321.98	84.4	0	1.34	0	1.021352	-13.765901	220	5	1.1	0.9;
	208	1	33.44	10	0	1.36	0	1.041988	-26.978403	110	3	1.1	0.9;
	209	2	0	0	0	0	0	1.026832	-36.641167	110	12	1.1	0.9;
	210	1	0	0	0	0	0	0.959797	-37.411141	150	9	1.1	0.9;
	211	1	75.28	19.2	0	1.9	0	1.031002	12.816647	110	8	1.1	0.9;
	212	1	95.69	24.54	0	2.85	0	1.026845	-5.374895	220	13	1.1	0.9;
	213	1	37.34	10.9	0	0.63	0	1.00854	-35.282646	110	1	1.1	0.9;
	214	2	0	0	0	0	0	1.079765	13.260236	380	13	1.1	0.9;
	215	1	-0	0	0	4.93	0	1.002239	-27.62245	110	18	1.1	0.9;
	216	1	155.6	54.7	0	2.21	0	1.036932	-24.918537	220	5	1.1	0.9;
	217	1	105.2	0.1	0	7.36	0	1.032041	-2.878371	220	5	1.1	0.9;
	218	1	69.83	15.12	0	2.23	0	1.048352	-13.235694	220	5	1.1	0.9;
	219	2	0	0	0	0	0	1.056623	-15.858282	380	7	1.1	0.9;
	220	2	0	0	0	0	0	1.047116	-15.263985	380	6	1.1	0.9;
	221	2	0	0	0	0	0	1.030636	-22.725319	380	5	1.1	0.9;
	222	2	0	0	0	0	0	1.030494	64.590215	150	1	1.1	0.9;
	223	1	27.03	8.8	0	1.84	0	1.02935	-21.253236	110	21	1.1	0.9;
	224	1	2.6	0.5	0	0.14	0	0.999483	-19.779007	110	18	1.1	0.9;
	225	1	0	0	0	1.15	0	1.058955	-20.480408	220	5	1.1	0.9;
	226	1	14.62	4.8	0	1.02	0	1.025663	-34.219329	110	12	1.1	0.9;
	227	1	-0	0	0	2.55	0	1.036575	-23.032139	220	7	1.1	0.9;
	228	1	-23.43	57.4	0	0	0	1.039157	-8.583492	150	2	1.1	0.9;
	229	1	426.6	93.7	0	80.94	0	1.043175	-27.471657	380	6	1.1	0.9;
	230	1	152.07	22.52	0	7.87	0	1.004938	-47.015596	220	4	1.1	0.9;
	231	1	-0.13	-1	0	0	0	1.013236	22.32882	110	8	1.1	0.9;
	232	1	0	0	0	0.7	0	1.012704	-24.275333	150	9	1.1	0.9;
	233	1	2.5	3.8	0	0.15	0	0.951324	-30.509004	110	18	1.1	0.9;
	234	1	16.7	2.4	0	0.87	0	1.011539	-24.669443	154	1	1.1	0.9;
	235	1	34.04	4.4	0	1.25	0	0.988571	-41.615405	110	22	1.1	0.9;
	236	1	45.85	7	0	1.05	0	1.023549	-54.30395	220	4	1.1	0.9;
	237	1	0	-0	0	30.97	0	1.032098	38.584994	380	3	1.1	0.9;
	238	1	0	-0	0	19.36	0	1.042719	-48.730173	220	6	1.1	0.9;
	239	2	0	0	0	0	0	1.018711	-30.200516	110	19	1.1	0.9;
	240	1	0	0	0	0	0	1.024081	-30.555303	150	9	1.1	0.9;
	241	1	-0	0	0	3.95	0	1.035656	58.992416	220	3	1.1	0.9;
	242	1	167.9	18.2	0	5.97	0	1.03642	-2.810922	220	3	1.1	0.9;
	243	1	0	-0.01	0	13.49	0	1.017401	17.219052	150	8	1.1	0.9;
	244	1	-0	-0	0	0.68	0	1.017943	6.456763	150	8	1.1	0.9;
	245	1	16.62	5.5	0	1.34	0	1.011096	-30.247505	110	21	1.1	0.9;
	246	2	0	0	0	0	0	1.06685	-23.368108	220	1	1.1	0.9;
	247	1	0	0	0	0.72	0	0.988152	-23.780097	110	18	1.1	0.9;
	248	2	0	0	0	0	0	1.043991	11.30228	380	13	1.1	0.9;
	249	1	-0	-0	0	26.3	0	1.002426	-43.012777	220	4	1.1	0.9;
	250	1	0	0	0	13.85	0	1.052918	9.637659	380	3	1.1	0.9;
	251	1	118.4	16.9	0	8.02	0	1.014818	9.078327	154	1	1.1	0.9;
	252	1	-0	0	0	2.66	0	1.042194	46.417078	220	3	1.1	0.9;
	253	1	0	-0	0	4.06	0	1.031397	-20.282601	220	1	1.1	0.9;
	254	1	-0	-0	0.135493	12.74062	0	1.067521	-27.715246	380	21	1.1	0.9;
	255	2	0	0	0	0	0	1.059597	-15.459412	380	16	1.1	0.9;
	256	1	0	-0	0	7.09	0	1.036026	-36.417261	220	6	1.1	0.9;
	257	1	113.9	24.5	0	6.72	0	1.027099	-46.196068	220	4	1.1	0.9;
	258	1	88	12.6	0	22.25	0	1.035501	-24.710337	154	1	1.1	0.9;
	259	1	42.67	12.11	0	14.84	0	1.014986	-32.538711	150	9	1.1	0.9;
	260	1	216.8	40.5	0	5.91	0	1.026451	-11.789492	220	3	1.1	0.9;
	261	1	-0	0	0	1.29	0	1.052684	-45.804706	380	10	1.1	0.9;
	262	2	0	0	0	0	0	1.033795	9.17742	220	13	1.1	0.9;
	263	2	0	0	0	0	0	1.028681	0.052818	154	1	1.1	0.9;
	264	1	38.74	9	0	0.57	0	1.010447	-47.085784	220	4	1.1	0.9;
	265	1	350.4	29.3	0	19.31	0	1.051094	-26.879001	220	11	1.1	0.9;
	266	1	0	-0	0	140.96	0	1.05561	-42.525923	380	4	1.1	0.9;
	267	1	103.8	2.7	0	3.33	0	1.008959	-51.29783	220	4	1.1	0.9;
	268	1	-0.13	-1	0	0	0	1.040854	25.721966	220	8	1.1	0.9;
	269	1	0	-0	0	0.66	0	1.016247	-41.093097	110	22	1.1	0.9;
	270	1	-0	0	0	6.34	0	1.038383	66.520284	380	1	1.1	0.9;
	271	1	96.7	26.8	0	0	0	1.034099	-10.123925	150	2	1.1	0.9;
	272	1	0	0	0	19.54	0	1.051	-31.736282	150	9	1.1	0.9;
	273	1	-0	-0	0	14.87	0	1.055045	-27.649778	220	23	1.1	0.9;
	274	1	19.02	6	0	0.25	0	1.016203	-27.399119	110	21	1.1	0.9;
	275	1	102.1	25.3	0	1.11	0	1.014295	1.916687	150	8	1.1	0.9;
	276	1	-0	0	0	2.44	0	1.046151	63.617689	220	3	1.1	0.9;
	277	1	0	0	0	0	0	1.017147	-28.974177	150	9	1.1	0.9;
	278	2	0	0	0	0	0	1.014153	-29.26945	110	18	1.1	0.9;
	279	1	8.31	3	0	0.99	0	1.005219	-28.673064	110	19	1.1	0.9;
	280	1	119.6	29.6	0	5.34	0	1.031588	-11.381836	220	5	1.1	0.9;
	281	1	-0	-0	0	18.07	0	1.034886	13.450704	380	8	1.1	0.9;
	282	2	0	0	0	0	0	1.077605	-12.395364	380	18	1.1	0.9;
	283	1	27.33	11.9	0	1.17	0	1.030144	-12.380018	220	5	1.1	0.9;
	284	1	191.2	-3.4	0	19.21	0	1.021258	-48.983591	220	4	1.1	0.9;
	285	1	0	0	0.029916	18.291787	0	1.033022	66.384978	380	1	1.1	0.9;
	286	1	101.6	14.5	0	1.94	0	1.01692	4.101949	154	1	1.1	0.9;
	287	1	177.1	66.9	0	71.03	0	1.05842	-16.180765	380	16	1.1	0.9;
	288	1	0	0	0	0.3	0	1.064689	-21.139278	150	9	1.1	0.9;
	289	2	0	0	0	0	0	1.046783	-31.796612	220	6	1.1	0.9;
	290	1	53.46	18.7	0	1.64	0	1.026894	-21.57601	220	20	1.1	0.9;
	291	1	0	-0	0	12.73	0	1.037913	3.647073	220	14	1.1	0.9;
	292	1	13.82	1.4	0	0.7	0	1.044729	-14.863031	220	5	1.1	0.9;
	293	1	85.1	67	0	0.18	0	1.071967	-22.394634	150	9	1.1	0.9;
	294	2	0	0	0	0	0	1.003375	-3.704358	400	1	1.1	0.9;
	295	1	0	0	0	0.08	0	1.019298	-29.061846	220	6	1.1	0.9;
	296	1	189.3	76.7	0	3.24	0	0.991132	26.598199	220	3	1.1	0.9;
	297	1	23.73	9.4	0	1.45	0	1.032348	3.967517	150	8	1.1	0.9;
	298	1	10.01	-21	0	8.59	0	1.033157	-31.091365	220	11	1.1	0.9;
	299	2	0	0	0	0	0	1.039248	-46.484496	380	4	1.1	0.9;
	300	1	0	0	0	1.05	0	1.019085	-22.661128	150	9	1.1	0.9;
	301	1	-14.85	-4.95	0	0.37	0	1.016739	-24.050737	110	18	1.1	0.9;
	302	1	20.9	3	0	2.65	0	1.024624	7.719367	154	1	1.1	0.9;
	303	1	431.7	121.6	0	1.14	0	1.008577	-28.714495	220	21	1.1	0.9;
	304	1	2.2	0.5	0	0.21	0	0.989535	-54.483262	150	10	1.1	0.9;
	305	1	103	12	0	0.63	0	1.065982	-32.480519	220	5	1.1	0.9;
	306	1	-0	0	0	4.76	0	1.024949	-19.155495	220	5	1.1	0.9;
	307	1	0	0	0	25.85	0	1.078101	-24.185519	380	1	1.1	0.9;
	308	1	0	0	0	0	0	0.858822	-39.292558	150	9	1.3	0.7;
	309	2	0	0	0	0	0	0.989085	12.894426	220	3	1.1	0.9;
	310	2	0	0	0	0	0	1.155322	-21.430019	380	9	1.3	0.7;
	311	1	42.05	17.2	0	1.31	0	1.018788	-2.280144	150	8	1.1	0.9;
	312	1	-0.16	-2.91	0	0.43	0	1.064463	10.943215	110	8	1.1	0.9;
	313	2	0	0	0	0	0	1.025259	6.096505	154	1	1.1	0.9;
	314	1	163.1	46.3	0	10.6	0	0.998939	-16.361138	220	5	1.1	0.9;
	315	1	-0	-0	0	2.13	0	1.024288	-36.514525	220	10	1.1	0.9;
	316	1	0	-0	0	2.81	0	1.039789	-4.974442	220	3	1.1	0.9;
	317	1	-0	-0	0.045946	-1.82203	0	1.017314	6.026115	380	2	1.1	0.9;
	318	1	-0	-0	0	1.05	0	0.995147	-31.240936	150	9	1.1	0.9;
	319	1	153.3	21.8	0	18.25	0	1.018898	-22.375931	154	1	1.1	0.9;
	320	2	0	0	0	0	0	1.013336	8.731823	400	1	1.1	0.9;
	321	1	67.17	-23.9	0	0.93	0	1.047108	63.391695	220	3	1.1	0.9;
	322	1	303.8	85.4	0	3.63	0	0.960285	-44.896787	220	10	1.1	0.9;
	323	1	-0	-0	0	0	0	1.021536	4.254199	400	1	1.1	0.9;
	324	2	0	0	0	0	0	1.063285	52.392117	380	3	1.1	0.9;
	325	2	0	0	0	0	0	1.009625	-3.505565	154	1	1.1	0.9;
	326	1	0	0	0	0	0	1.040967	25.768951	220	8	1.1	0.9;
	327	1	0	0	0	12.29	0	1.021761	-12.621738	380	3	1.1	0.9;
	328	2	0	0	0	0	0	1.017618	23.471214	220	13	1.1	0.9;
	329	1	24.23	7.1	0	1.64	0	1.019588	-29.061157	110	21	1.1	0.9;
	330	1	108.6	2.6	0	31.8	0	1.05407	10.188241	380	3	1.1	0.9;
	331	1	4	2.4	0	0.03	0	0.97496	-31.995773	150	9	1.1	0.9;
	332	1	30.03	10.9	0	1.45	0	1.03175	-25.481871	110	19	1.1	0.9;
	333	1	524.5	43	0	6.85	0	1.008335	-42.644766	220	5	1.1	0.9;
	334	1	129.9	18.5	0	0.38	0	1.021555	-8.102376	154	1	1.1	0.9;
	335	1	-0	0	0	1.92	0	0.956257	-31.167466	110	18	1.1	0.9;
	336	1	4.4	0	0	0.51	0	1.00793	-20.199223	110	18	1.1	0.9;
	337	1	9.22	1.2	0	0.8	0	1.07406	-19.162915	110	19	1.1	0.9;
	338	2	0	0	0	0	0	1.046957	-6.228603	220	5	1.1	0.9;
	339	1	-0	0	0	0.14	0	1.034126	-34.903603	220	6	1.1	0.9;
	340	1	186.6	56.6	0	30.24	0	1.066043	-22.461372	380	20	1.1	0.9;
	341	1	-0	-0	0	0.81	0	1.017154	-29.141612	110	21	1.1	0.9;
	342	1	35.41	20.3	0	10.07	0	1.008599	-32.559855	220	6	1.1	0.9;
	343	1	3.2	1.2	0	0.69	0	1.044421	-22.619481	110	19	1.1	0.9;
	344	1	372.8	147.5	0	27.73	0	1.02817	-1.864435	150	8	1.1	0.9;
	345	1	0	0	0.080094	-0.240282	0	1.029033	-30.721414	220	11	1.1	0.9;
	346	1	197.5	48.3	0	0.71	0	1.037587	-21.605967	220	5	1.1	0.9;
	347	2	0	0	0	0	0	1.020949	-4.447906	154	1	1.1	0.9;
	348	2	0	0	0	0	0	1.057894	-25.029783	380	6	1.1	0.9;
	349	1	-0	0	0	15.15	0	1.046687	8.496714	150	8	1.1	0.9;
	350	1	110.4	26.2	0	5.62	0	1.038713	-24.658174	220	5	1.1	0.9;
	351	1	-0.15	-1.93	0	0	0	1.025813	17.762925	220	8	1.1	0.9;
	352	2	0	0	0	0	0	1.048835	0.123472	380	5	1.1	0.9;
	353	1	-0	-0	0	14.6	0	1.053404	-29.839213	220	12	1.1	0.9;
	354	1	-0	0	0	8.67	0	1.046845	-16.150037	380	1	1.1	0.9;
	355	1	-4.86	-2.15	0	0.31	0	0.994856	-24.893606	110	18	1.1	0.9;
	356	1	141.1	2.1	0	4.78	0	1.046664	-18.978177	220	7	1.1	0.9;
	357	1	52.96	32.8	0	14.14	0	1.014078	-32.864305	150	9	1.1	0.9;
	358	1	13.9	1.9	0	0.62	0	1.03286	8.494459	154	1	1.1	0.9;
	359	1	0	0	0	2.37	0	1.043241	-13.963134	220	1	1.1	0.9;
	360	1	35.54	11.5	0	3.65	0	1.057579	10.664154	110	8	1.1	0.9;
	361	2	0	0	0	0	0	1.05949	16.125446	380	13	1.3	0.7;
	362	1	48.92	14.52	0	2.85	0	1.012713	61.825927	150	1	1.1	0.9;
	363	1	-0	0	0	26.52	0	1.084924	-36.917309	150	9	1.3	0.7;
	364	1	10.01	4	0	8.11	0	1.009235	-36.457276	220	6	1.1	0.9;
	365	1	0	-0	0	0.82	0	1.011415	-22.836999	150	9	1.1	0.9;
	366	2	0	0	0	0	0	1.008853	-6.356964	400	1	1.1	0.9;
	367	1	65.17	-51.5	0	4.85	0	1.06614	-23.969395	220	5	1.1	0.9;
	368	1	-0	0	0	7.11	0	1.017336	-32.97169	220	6	1.1	0.9;
	369	1	-0	0	0	3.78	0	1.047259	12.873199	220	8	1.1	0.9;
	370	1	56.76	23.2	0	0.49	0	1.016126	-2.465828	150	8	1.1	0.9;
	371	1	0	0	0	11.44	0	1.04061	-20.694247	154	1	1.1	0.9;
	372	2	0	0	0	0	0	1.074991	8.745956	380	14	1.1	0.9;
	373	1	-0	0	0	0.16	0	0.9775	-30.407675	150	9	1.1	0.9;
	374	1	-0	-0	0	2.27	0	1.035025	-24.898686	154	1	1.1	0.9;
	375	1	-0	0	0	1.61	0	1.013524	52.440501	220	3	1.1	0.9;
	376	1	0.01	0	0	35.09	0	1.028991	-31.53336	380	6	1.1	0.9;
	377	1	-0.18	-1	0	0	0	1.021561	10.573296	150	8	1.1	0.9;
	378	1	221.35	51.8	0	8.81	0	1.017306	-8.141276	220	13	1.1	0.9;
	379	1	23.83	7.9	0	0.72	0	0.995571	-31.453181	110	21	1.1	0.9;
	380	1	-95.71	-2.8	0	7.25	0	1.00126	-33.303602	220	10	1.1	0.9;
	381	1	44.05	0	0	0	0	0.935262	-28.688114	150	7	1.1	0.9;
	382	2	0	0	0	0	0	1.060853	-23.782062	220	7	1.1	0.9;
	383	1	19.02	11.5	0	0	0	0.991245	-30.816553	150	9	1.1	0.9;
	384	1	0	-0	0	0.3	0	1.006323	-30.008109	150	9	1.1	0.9;
	385	1	-0	-0	0	105.84	0	1.039766	-49.86378	380	4	1.1	0.9;
	386	1	0	-0	0	28.09	0	1.064436	21.11405	380	3	1.1	0.9;
	387	1	-0.19	-4.38	0	0	0	1.043626	6.058475	150	8	1.1	0.9;
	388	1	0	-0	0	7.15	0	0.988589	6.832917	220	3	1.1	0.9;
	389	2	0	0	0	0	0	1.017731	-41.302632	150	10	1.1	0.9;
	390	1	27.93	10.1	0	0.43	0	1.071096	-17.869204	110	19	1.1	0.9;
	391	1	34.94	21.6	0	0.18	0	0.979313	-27.16062	150	9	1.1	0.9;
	392	1	134.1	34.2	0	0.75	0	1.011308	-0.952791	150	8	1.1	0.9;
	393	1	17.72	5.9	0	0.32	0	1.062143	-27.065705	220	21	1.1	0.9;
	394	2	0	0	0	0	0	1.063147	41.896073	380	3	1.1	0.9;
	395	1	-0	-0	0	7.85	0	1.033317	14.723689	380	8	1.1	0.9;
	396	1	0	0	0	0	0	0.90329	-35.996951	150	9	1.3	0.7;
	397	1	77.79	-40.8	0	12.06	0	1.034394	-13.831084	220	3	1.1	0.9;
	398	1	11.91	3.8	0	0.46	0	0.977257	-21.273325	110	18	1.1	0.9;
	399	1	5.01	2	0	0.13	0	1.013866	-36.125837	220	6	1.1	0.9;
	400	1	-19.66	-10.83	0	0.06	0	1.007695	-30.151396	110	18	1.1	0.9;
	401	1	56.5	36.29	0	3.08	0	1.03497	-14.683133	220	5	1.1	0.9;
	402	1	-0	-0	0	16.38	0	1.008743	-36.11448	220	6	1.1	0.9;
	403	1	-0	-0	0.090774	7.330444	0	1.061673	-28.767097	220	18	1.1	0.9;
	404	1	3.81	0.24	0	0.23	0	1.014378	-26.354271	110	18	1.1	0.9;
	405	1	254.7	79.4	0	9.43	0	1.017423	-46.712396	380	6	1.1	0.9;
	406	2	0	0	0	0	0	1.012399	-36.269005	110	12	1.1	0.9;
	407	1	69.5	9.9	0	98.63	0	1.007216	7.794515	400	1	1.1	0.9;
	408	1	129.1	7.4	0	0.04	0	1.03515	-11.827186	220	5	1.1	0.9;
	409	2	0	0	0	0	0	1.032831	-6.435889	154	1	1.1	0.9;
	410	1	170.89	50.38	0	6.44	0	1.031489	-27.031073	220	5	1.1	0.9;
	411	2	0	0	0	0	0	1.03248	-38.941809	110	22	1.1	0.9;
	412	1	26.4	3.8	0	4.65	0	1.022685	13.654268	154	1	1.1	0.9;
	413	2	0	0	0	0	0	1.048421	-6.81908	220	5	1.1	0.9;
	414	1	20.52	10.3	0	0.04	0	1.017694	-32.178003	150	9	1.1	0.9;
	415	1	4.71	0.7	0	11.39	0	1.024129	-52.284525	220	4	1.1	0.9;
	416	1	70.68	13.5	0	6.96	0	1.04501	-6.926272	220	5	1.1	0.9;
	417	1	187.19	1.8	0	20.47	0	1.021927	-11.809092	220	3	1.1	0.9;
	418	1	24.43	8.7	0	0.17	0	1.030169	-27.397466	110	15	1.1	0.9;
	419	1	8.81	3.2	0	1.12	0	0.992991	-23.021888	110	19	1.1	0.9;
	420	1	0	0	0	0	0	1.047945	18.027026	380	8	1.1	0.9;
	421	1	33.74	20.9	0	1.1	0	0.994253	-32.082481	150	9	1.1	0.9;
	422	1	35.84	12.5	0	0.26	0	0.99831	-31.51274	110	19	1.1	0.9;
	423	2	0	0	0	0	0	1.030299	-21.488115	154	1	1.1	0.9;
	424	2	0	0	0	0	0	1.051241	41.066677	380	3	1.1	0.9;
	425	1	-0	0	0	2.72	0	1.023036	6.329808	380	1	1.1	0.9;
	426	1	123.1	6.5	0	3.25	0	1.038618	-18.666777	220	5	1.1	0.9;
	427	1	-0	0	0	1.24	0	1.074924	-36.064337	380	1	1.1	0.9;
	428	1	0.1	0	0	0.57	0	0.98973	-21.739403	110	18	1.1	0.9;
	429	1	26.93	-22.4	0	4.95	0	1.011621	-45.861365	220	4	1.1	0.9;
	430	1	29.03	9	0	0.56	0	1.045487	-26.661671	110	21	1.1	0.9;
	431	1	188.42	106.93	0	4.89	0	1.031863	6.601988	220	13	1.1	0.9;
	432	1	-0	-0	0	68.517663	0	1.041598	2.324897	380	5	1.1	0.9;
	433	2	0	0	0	0	0	1.024885	20.046114	380	3	1.1	0.9;
	434	2	0	0	0	0	0	1.037999	2.03991	154	1	1.1	0.9;
	435	1	33.14	11.6	0	0.22	0	1.018336	-30.666235	110	19	1.1	0.9;
	436	1	18.3	8.2	0	35.421268	0	1.120535	-21.925748	380	9	1.3	0.7;
	437	1	69.08	24	0	1	0	1.016926	-35.872023	220	6	1.1	0.9;
	438	2	0	0	0	0	0	1.038705	59.611786	220	3	1.1	0.9;
	439	2	0	0	0	0	0	1.072491	59.186166	220	3	1.1	0.9;
	440	2	0	0	0	0	0	1.028093	-36.715425	220	4	1.1	0.9;
	441	1	-0	0	0.199768	27.888516	0	1.032136	-52.444714	380	10	1.1	0.9;
	442	2	0	0	0	0	0	1.011727	-46.67987	220	4	1.1	0.9;
	443	1	-12.61	4.1	0	1.83	0	1.001838	-30.170359	150	9	1.1	0.9;
	444	1	0	-0	0	2.45	0	1.031001	-12.498392	220	5	1.1	0.9;
	445	1	36.74	10.3	0	8.1	0	1.019699	-45.007189	220	4	1.1	0.9;
	446	1	100.53	43.37	0	0	0	1.014581	-26.856538	110	24	1.1	0.9;
	447	2	0	0	0	0	0	1.035215	-9.864665	110	18	1.1	0.9;
	448	1	0	0	0	0	0	0.96795	-32.435222	150	9	1.1	0.9;
	449	1	0	0	0	29.42	0	1.016584	-8.707775	380	5	1.1	0.9;
	450	1	42.04	43	0	0.02	0	1.070059	-26.574071	150	9	1.1	0.9;
	451	1	-0	-0	0	22.71	0	1.063642	22.363908	380	1	1.1	0.9;
	452	2	0	0	0	0	0	1.013382	-41.604993	110	22	1.1	0.9;
	453	2	0	0	0	0	0	1.024712	-7.790435	380	5	1.1	0.9;
	454	1	23.7	3.4	0	8.13	0	1.040257	8.958679	154	1	1.1	0.9;
	455	1	274.32	5.66	0	13.24	0	1.038782	-31.177998	220	5	1.1	0.9;
	456	2	0	0	0	0	0	0.999847	-27.865709	110	18	1.1	0.9;
	457	1	1.8	0.7	0	0.18	0	0.993243	-17.575552	110	19	1.1	0.9;
	458	1	0	0	0.190956	35.262077	0	1.035934	-24.857043	380	11	1.1	0.9;
	459	1	97	40.3	0	16.79	0	1.003397	-8.906619	220	3	1.1	0.9;
	460	1	0	-0	0	0.88	0	0.997299	-26.104521	110	18	1.1	0.9;
	461	1	0	0	0	0	0	1.010475	-33.979657	150	9	1.1	0.9;
	462	1	0	0	0	0	0	1.01029	-30.963324	150	9	1.1	0.9;
	463	1	24.33	7.6	0	1.18	0	0.985095	-20.318385	110	18	1.1	0.9;
	464	1	44.75	0.9	0	6.67	0	1.04845	-14.940147	220	7	1.1	0.9;
	465	1	-0	0	0.11	2.57	0	1.026607	-27.324978	220	21	1.1	0.9;
	466	1	0	0	0	0	0	1.03086	-29.325776	150	9	1.1	0.9;
	467	1	40.5	5.7	0	3.9	0	1.023193	-9.366271	154	1	1.1	0.9;
	468	1	9.01	6.8	0	0.44	0	0.967757	-28.631796	110	18	1.1	0.9;
	469	1	81.09	48	0	0.03	0	1.014501	-30.270983	150	9	1.1	0.9;
	470	1	-0	0	0	0.43	0	1.045753	-28.152812	150	9	1.1	0.9;
	471	1	0	0	0	1.47	0	1.024955	-22.608823	220	6	1.1	0.9;
	472	1	336.2	131.6	0	64.01	0	1.048833	7.3045	150	8	1.1	0.9;
	473	1	-22.95	-3.72	0	14.34	0	1.023798	-46.331002	220	4	1.1	0.9;
	474	1	-0	0	0	0.97	0	1.02207	-39.273576	110	1	1.1	0.9;
	475	1	0	0	0	9.41	0	1.051128	49.661516	380	3	1.1	0.9;
	476	1	67.3	94.8	0	7.96	0	0.992868	5.766625	220	3	1.1	0.9;
	477	1	0	0	0	70.37	0	1.042365	-42.331272	380	4	1.1	0.9;
	478	1	0	-0	0	33.23	0	1.086063	45.381505	380	3	1.1	0.9;
	479	1	16.62	3.1	0	1.14	0	1.006931	-18.033449	110	18	1.1	0.9;
	480	1	5.81	1.1	0	4.02	0	1.04546	-9.596128	110	19	1.1	0.9;
	481	1	-0	0	0	17.44	0	1.021457	-48.478236	220	4	1.1	0.9;
	482	1	0	0	0	1.05	0	0.984082	-25.263297	110	18	1.1	0.9;
	483	1	57.26	15.6	0	3.43	0	1.017211	-50.081883	220	4	1.1	0.9;
	484	2	0	0	0	0	0	1.054091	53.806047	220	3	1.1	0.9;
	485	1	0	0	0	4.9	0	1.024648	-30.843672	380	6	1.1	0.9;
	486	1	7.01	2	0	0.44	0	1.014868	4.502664	150	8	1.1	0.9;
	487	1	87.8	12.5	0	0.38	0	1.014589	-4.325752	154	1	1.1	0.9;
	488	1	6.61	2.92	0	1.17	0	1.036419	2.841859	220	14	1.1	0.9;
	489	2	0	0	0	0	0	1.008032	-47.898933	220	4	1.1	0.9;
	490	1	20.52	4	0	2.77	0	1.032528	-23.365945	220	5	1.1	0.9;
	491	2	0	0	0	0	0	1.033193	34.821441	380	3	1.1	0.9;
	492	1	68.18	22	0	2.42	0	0.9761	-43.380627	110	22	1.1	0.9;
	493	1	23.23	7.6	0	0.76	0	0.977456	-32.668095	110	21	1.1	0.9;
	494	1	0	-0	0	0	0	0.994103	-30.581336	150	9	1.1	0.9;
	495	1	181.7	51	0	73.49	0	1.091084	47.134014	380	3	1.1	0.9;
	496	1	-0	-0	0	10.04	0	1.01836	10.70214	150	8	1.1	0.9;
	497	1	40	5.7	0	2.06	0	1.025576	8.402387	154	1	1.1	0.9;
	498	1	31.23	11.3	0	0.33	0	1.022852	-28.519454	110	19	1.1	0.9;
	499	1	287.31	-5.58	0	19.9	0	1.019259	-5.089216	380	13	1.1	0.9;
	500	1	144.4	35.5	0	7.66	0	1.043162	-15.091023	220	5	1.1	0.9;
	501	1	-0	0	0	0.25	0	1.045916	-27.133616	110	3	1.1	0.9;
	502	1	-0	0	0	1.96	0	1.036985	-6.888217	220	5	1.1	0.9;
	503	1	4.1	-607	0.256125	85.984967	0	1.069204	23.217465	380	3	1.1	0.9;
	504	1	18.62	6.2	0	1.07	0	0.996555	-31.485081	110	21	1.1	0.9;
	505	2	0	0	0	0	0	1.022173	10.007127	154	1	1.1	0.9;
	506	1	-0	0	0	0.4	0	1.057858	-9.331318	380	5	1.1	0.9;
	507	1	23.43	8.5	0	0.33	0	1.047975	-25.997094	110	19	1.1	0.9;
	508	1	23.92	7.75	0	1.91	0	1.004689	-24.87385	110	18	1.1	0.9;
	509	1	77.29	7	0	17.6	0	1.00895	-61.098444	220	4	1.1	0.9;
	510	1	-0	-0	0	1.5	0	1.02625	-22.233434	220	6	1.1	0.9;
	511	2	0	0	0	0	0	1.081589	57.979622	380	3	1.1	0.9;
	512	1	10.11	2.8	0	1.01	0	1.005897	-27.826195	110	18	1.1	0.9;
	513	1	57.46	17.1	0	1.61	0	1.031868	-7.665769	220	5	1.1	0.9;
	514	1	9.61	3.1	0	0.77	0	1.017141	-38.705949	110	12	1.1	0.9;
	515	2	0	0	0	0	0	1.045162	-17.494553	220	11	1.1	0.9;
	516	2	0	0	0	0	0	1.07129	-33.728741	220	5	1.1	0.9;
	517	1	7.71	16.7	0	3.59	0	0.9901	-34.700265	150	9	1.1	0.9;
	518	2	0	0	0	0	0	1.067615	-2.738928	380	14	1.1	0.9;
	519	1	207.2	9.4	0	8.77	0	1.019722	-50.589759	220	4	1.1	0.9;
	520	1	118.5	20.4	0	2.57	0	1.049911	-27.592982	220	5	1.1	0.9;
	521	1	14.72	5.3	0	0.76	0	1.003457	-27.851665	110	19	1.1	0.9;
	522	1	11.61	7.2	0	28.77	0	1.014941	-32.902975	150	9	1.1	0.9;
	523	1	189.9	74.7	0	26.2	0	1.042367	6.642635	150	8	1.1	0.9;
	524	2	0	0	0	0	0	1.046979	10.752972	154	1	1.1	0.9;
	525	1	36.54	27.4	0	0.15	0	0.969544	13.480361	220	3	1.1	0.9;
	526	1	176.7	25.6	0	2.01	0	1.012916	-48.291617	220	6	1.1	0.9;
	527	1	277.7	45.4	0	0.7	0	1.010053	-46.915974	220	4	1.1	0.9;
	528	1	0	-0	0	3.46	0	0.99921	-50.924813	220	6	1.1	0.9;
	529	1	0	-0	0	28.16	0	1.030762	-16.275798	380	1	1.1	0.9;
	530	1	25	3.6	0	4.73	0	1.014401	-22.258328	154	1	1.1	0.9;
	531	1	0	-0	0	47.4	0	1.022093	-18.724236	380	1	1.1	0.9;
	532	1	0	0	0	1.98	0	1.018038	9.201002	220	3	1.1	0.9;
	533	1	16.52	3.8	0	0.98	0	0.959418	-29.583188	110	18	1.1	0.9;
	534	1	118.1	16.8	0	2.94	0	1.009193	1.409034	154	1	1.1	0.9;
	535	1	11.51	3.8	0	1.21	0	1.011587	-36.72714	110	12	1.1	0.9;
	536	1	87.8	-1.4	0.079073	36.047349	0	1.054738	-48.67568	380	10	1.1	0.9;
	537	1	27.53	17	0	0.76	0	0.949809	-29.215766	150	9	1.1	0.9;
	538	1	21.02	13	0	1.17	0	0.972441	-36.764788	150	9	1.1	0.9;
	539	1	57.96	18.9	0	14.91	0	1.013931	3.297995	150	8	1.1	0.9;
	540	1	166.9	86.1	0	4.06	0	0.973758	10.352992	220	3	1.1	0.9;
	541	2	0	0	0	0	0	1.03819	4.193683	154	1	1.1	0.9;
	542	1	2	0.7	0	0.02	0	1.067645	-19.587155	110	19	1.1	0.9;
	543	1	72	10.3	0	6.24	0	0.995488	-22.770476	154	1	1.1	0.9;
	544	1	79.69	39.6	0	18.83	0	1.034016	-52.177549	380	10	1.1	0.9;
	545	1	-0	-0	0	8.48	0	1.065316	-22.090019	220	18	1.1	0.9;
	546	1	-0	-0	0	29.89	0	1.081778	-35.986218	150	9	1.1	0.9;
	547	1	30.03	14	0	0.61	0	1.012961	-36.218409	220	6	1.1	0.9;
	548	1	30.93	8.9	0	0.45	0	1.029939	15.35829	110	8	1.1	0.9;
	549	1	0	-0	0	5.15	0	1.066337	-11.605789	220	5	1.1	0.9;
	550	1	0	-0	0	10	0	1.034222	-45.869511	220	4	1.1	0.9;
	551	1	17.72	5.9	0	0.52	0	0.972889	-36.288819	110	21	1.1	0.9;
	552	1	180.8	90.8	0	2.42	0	0.966453	-43.582462	150	10	1.1	0.9;
	553	1	0	0	0	0.55	0	1.02741	19.17319	154	1	1.1	0.9;
	554	1	0	-0	0	1.97	0	1.020599	-21.077132	220	5	1.1	0.9;
	555	1	0	-0	0	7.35	0	1.036218	-34.48522	220	11	1.1	0.9;
	556	1	16.52	9	0	1.17	0	1.032971	-23.112249	220	5	1.1	0.9;
	557	1	0	0	0	12.14	0	1.042172	-12.78196	220	19	1.1	0.9;
	558	1	-0.18	-3.57	0	0	0	1.017408	17.219074	150	8	1.1	0.9;
	559	1	42.25	16.6	0	0	0	1.038823	4.926964	150	8	1.1	0.9;
	560	1	3.8	1.3	0	0.65	0	1.001206	-16.24915	110	18	1.1	0.9;
	561	1	0	-0	0	0.32	0	1.045371	-11.322037	220	5	1.1	0.9;
	562	1	0	-0	0	0.48	0	1.035608	-4.32039	220	14	1.1	0.9;
	563	1	-0	0	0	0.55	0	0.967338	-31.490411	150	9	1.1	0.9;
	564	2	0	0	0	0	0	1.06702	-47.265609	220	5	1.1	0.9;
	565	1	198.4	32.7	0	31.82	0	1.01065	-21.146836	220	6	1.1	0.9;
	566	1	5.81	3.8	0	0.94	0	0.966543	-30.57403	110	18	1.1	0.9;
	567	1	0	-0	0	14.5	0	1.062189	-37.391032	380	1	1.1	0.9;
	568	1	20.22	7.3	0	0.59	0	1.018032	-30.296337	110	19	1.1	0.9;
	569	1	-0	0	0	35.364539	0	1.076066	-22.926788	380	20	1.1	0.9;
	570	1	47.95	29.7	0	1.59	0	0.993761	-32.641501	150	9	1.1	0.9;
	571	1	120.63	56.2	0	3	0	1.004642	11.297784	220	3	1.1	0.9;
	572	1	27.13	8.9	0	1.37	0	1.055676	-15.620867	110	19	1.1	0.9;
	573	1	10.71	3.5	0	0.82	0	1.008172	-39.537536	110	12	1.1	0.9;
	574	1	82.89	26.8	0	0.65	0	1.003792	-28.302462	110	19	1.1	0.9;
	575	1	0	-0	0	0.67	0	1.038739	-11.457881	220	5	1.1	0.9;
	576	1	108.6	38.7	0	0	0	0.998317	-18.957058	110	18	1.1	0.9;
	577	1	-0.13	-1	0	0	0	1.040133	25.638278	220	8	1.1	0.9;
	578	2	0	0	0	0	0	1.042146	18.799313	220	13	1.1	0.9;
	579	1	17.32	6.6	0	16.52	0	1.015454	-52.416993	380	4	1.1	0.9;
	580	1	322.7	11.9	0	10.28	0	1.015167	-40.787165	380	6	1.1	0.9;
	581	1	48.8	6.9	0	0.12	0	1.018271	-23.574905	154	1	1.1	0.9;
	582	1	51.16	18.2	0	5.23	0	1.047477	-20.544217	220	7	1.1	0.9;
	583	2	0	0	0	0	0	1.048595	-15.609378	220	5	1.1	0.9;
	584	1	151.2	72.4	0	4.57	0	1.02646	4.847775	380	3	1.1	0.9;
	585	1	21.12	8.3	0	4.26	0	1.040818	3.328266	150	8	1.1	0.9;
	586	1	9.51	3.2	0	0.34	0	1.047762	-26.908605	110	21	1.1	0.9;
	587	2	0	0	0	0	0	1.040079	0.128428	220	13	1.1	0.9;
	588	1	1.5	0.8	0	0.55	0	0.987706	-29.52887	110	18	1.1	0.9;
	589	1	27.03	9.8	0	1.01	0	1.051226	-26.821071	110	19	1.1	0.9;
	590	2	0	0	0	0	0	1.065004	54.658978	380	3	1.1	0.9;
	591	1	-0	0	0	0.56	0	1.003047	-31.096767	110	21	1.1	0.9;
	592	1	0	-0	0	0.27	0	1.002326	-0.228479	220	3	1.1	0.9;
	593	1	-0	0	0	29.2	0	1.052806	12.32779	380	8	1.1	0.9;
	594	1	67.57	17	0	5.75	0	1.03434	-9.08879	220	5	1.1	0.9;
	595	1	47.55	5.3	0	7.59	0	1.021721	-50.65158	220	4	1.1	0.9;
	596	1	24.83	7.6	0	0.23	0	0.967588	-29.094022	110	18	1.1	0.9;
	597	1	33.34	15.7	0	64.15	0	1.054276	-48.235591	380	4	1.1	0.9;
	598	1	-0	0	0	0.82	0	0.988306	-27.745722	150	9	1.1	0.9;
	599	1	0	-0	0	37.635138	0	0.984741	-22.697595	220	24	1.1	0.9;
	600	1	-0	0	0	8.72	0	1.020642	2.766315	380	3	1.1	0.9;
	601	1	140.6	42.6	0	5.71	0	1.03709	-14.510851	220	5	1.1	0.9;
	602	1	110.8	-8.8	0	3.15	0	1.02089	-47.482432	220	4	1.1	0.9;
	603	1	48.9	28.6	0	98.806507	0	1.127591	-22.048561	380	9	1.3	0.7;
	604	1	33.24	4.4	0	24.54	0	0.992921	-13.224894	380	5	1.1	0.9;
	605	1	23.4	3.3	0	1.91	0	1.019051	-5.534776	154	1	1.1	0.9;
	606	1	0	-0	0	19.56	0	1.030805	35.094114	380	3	1.1	0.9;
	607	1	-0	0	0	7.06	0	1.030222	-13.978729	330	24	1.1	0.9;
	608	1	150.9	10.4	0	4.77	0	1.063579	-33.089502	220	5	1.1	0.9;
	609	1	-0	-0	0	5.67	0	1.028257	-14.839583	220	5	1.1	0.9;
	610	1	-0	0	0	0.91	0	0.985271	-56.195761	150	10	1.1	0.9;
	611	1	-0	0	0.09381	0.90492	0	1.019174	-11.543717	380	7	1.1	0.9;
	612	1	-0	-0	0	0.52	0	1.035466	-26.190388	110	3	1.1	0.9;
	613	1	181.1	25.8	0	1.1	0	1.020502	3.815144	154	1	1.1	0.9;
	614	1	18.5	3.7	0	0.01	0	0.968407	-37.402489	150	9	1.1	0.9;
	615	2	0	0	0	0	0	1.068592	-11.049558	220	5	1.1	0.9;
	616	2	0	0	0	0	0	1.04484	-13.078255	220	5	1.1	0.9;
	617	1	-0.92	-4.99	0	1.86	0	0.989164	-28.084869	110	18	1.1	0.9;
	618	2	0	0	0	0	0	1.022784	-49.328375	380	4	1.1	0.9;
	619	1	72.88	14.4	0	2.96	0	1.042351	-25.58584	220	5	1.1	0.9;
	620	1	15.42	5.8	0	0.57	0	0.994999	-19.07911	110	18	1.1	0.9;
	621	1	21.42	8.1	0	7.38	0	1.012924	-52.128422	220	4	1.1	0.9;
	622	1	0	-0	0	11.57	0	1.069133	59.203598	380	3	1.1	0.9;
	623	1	36.1	5.2	0	11.12	0	1.029213	1.427637	154	1	1.1	0.9;
	624	1	111.4	36.5	0	70.21	0	1.032738	-10.941054	380	3	1.1	0.9;
	625	1	0	0	0	4.43	0	1.038541	-27.690568	220	1	1.1	0.9;
	626	1	0	0	0	0	0	0.934086	-33.42388	150	9	1.1	0.9;
	627	1	116.8	60.3	0	9.09	0	0.995326	-35.195598	150	9	1.1	0.9;
	628	2	0	0	0	0	0	0.982057	-35.15289	220	10	1.1	0.9;
	629	1	177.2	32.3	0	6.57	0	1.026049	-18.858201	220	3	1.1	0.9;
	630	1	-0	-0	0	17.07	0	1.027499	7.42553	380	3	1.1	0.9;
	631	1	0	0	0	0.71	0	0.996483	-30.702079	150	9	1.1	0.9;
	632	1	-0	-0	0	0.23	0	1.088191	-27.818441	150	9	1.1	0.9;
	633	1	2.5	-0	0	0.96	0	0.958685	-29.55312	110	18	1.1	0.9;
	634	1	0	0	0	104.15	0	1.018563	3.431688	400	1	1.1	0.9;
	635	1	-0	-0	0	4.96	0	0.999819	9.215641	220	3	1.1	0.9;
	636	2	0	0	0	0	0	1.04241	-22.994363	110	19	1.1	0.9;
	637	2	0	0	0	0	0	1.031148	16.262667	154	1	1.1	0.9;
	638	1	181.41	16.52	0	4.07	0	1.052719	14.597739	220	13	1.1	0.9;
	639	2	0	0	0	0	0	1.033532	-30.756579	220	5	1.1	0.9;
	640	1	102	26.6	0	2.31	0	1.015643	36.723091	220	3	1.1	0.9;
	641	1	45.32	18.5	0	0.74	0	1.04342	-13.738863	220	5	1.1	0.9;
	642	2	0	0	0	0	0	1.031368	-50.812889	380	4	1.1	0.9;
	643	1	15.52	6.3	0	0.6	0	1.022609	-2.01817	150	8	1.1	0.9;
	644	1	192.3	27.4	0	30.64	0	1.001275	-19.482697	400	1	1.1	0.9;
	645	1	26.33	9.5	0	1.85	0	1.011105	-26.132773	110	19	1.1	0.9;
	646	1	-0.59	-0.83	0	5.79	0	1.046449	-22.474118	150	9	1.1	0.9;
	647	1	48.05	15.8	0	6.08	0	1.05839	63.82172	150	1	1.1	0.9;
	648	2	0	0	0	0	0	1.02749	49.995344	220	3	1.1	0.9;
	649	1	40.94	19.2	0	5.2	0	1.022389	-48.793894	220	4	1.1	0.9;
	650	1	17.22	4.4	0	1.71	0	1.019857	11.645411	110	8	1.1	0.9;
	651	1	62.46	7.15	0	10.06	0	1.046751	64.301034	150	1	1.1	0.9;
	652	1	0	-0	0	5.12	0	1.022847	-11.260237	220	3	1.1	0.9;
	653	1	10.31	2.3	0	1.11	0	0.989505	-17.821652	110	18	1.1	0.9;
	654	1	205.6	74	0	5.56	0	1.016548	-25.008561	220	7	1.1	0.9;
	655	1	9.7	1.4	0	6.56	0	1.023447	16.49929	154	1	1.1	0.9;
	656	1	-0.15	-1.93	0	0	0	1.025785	17.764591	220	8	1.1	0.9;
	657	1	65.27	30.9	0	3.1	0	1.013809	9.026926	220	3	1.1	0.9;
	658	1	258.4	66.7	0	5.87	0	1.029318	-17.741053	220	5	1.1	0.9;
	659	1	-0	-0	0.174507	73.977052	0	1.016665	5.978239	380	2	1.1	0.9;
	660	1	584.3	166.3	0	48.159681	0	1.019972	-41.903867	380	6	1.1	0.9;
	661	1	-0	-0	0	0.6	0	1.02925	2.262276	380	5	1.1	0.9;
	662	1	0	-0	0	4.14	0	1.024506	11.339798	380	8	1.1	0.9;
	663	1	0	0	0.12	12.89	0	1.010876	-28.859495	220	21	1.1	0.9;
	664	1	38.04	11.7	0	0.18	0	1.056439	-27.160166	110	3	1.1	0.9;
	665	1	0	-0	0	7.2	0	1.034079	-22.397922	220	19	1.1	0.9;
	666	1	-5.61	-2.79	0	150.93	0	1.032739	-5.56264	380	5	1.1	0.9;
	667	1	-0	-0	0	2.76	0	1.000555	-32.932525	220	11	1.1	0.9;
	668	1	48.8	6.9	0	5.14	0	1.000234	-4.191624	154	1	1.1	0.9;
	669	1	-0	-0	0	29.29	0	1.035809	-30.822173	150	9	1.1	0.9;
	670	1	5.51	17.7	0	0.58	0	0.947855	-29.361997	150	9	1.1	0.9;
	671	1	0	0	0	0	0	1.007092	-26.93445	150	9	1.1	0.9;
	672	2	0	0	0	0	0	1.02121	-44.373291	380	6	1.1	0.9;
	673	1	0	0	0	12.75	0	1.041688	12.036915	380	3	1.1	0.9;
	674	1	-0	-0	0	65.82	0	1.093453	-14.521151	380	1	1.3	0.7;
	675	1	5.11	-0.1	0	3.73	0	0.997831	-25.825725	110	18	1.1	0.9;
	676	1	109.3	35.4	0	1.91	0	1.023634	5.250321	150	8	1.1	0.9;
	677	1	30.83	10.8	0	2.59	0	1.006126	-30.215816	110	18	1.1	0.9;
	678	1	232.1	89.8	0	0.09	0	1.005243	-17.794843	220	5	1.1	0.9;
	679	1	-0	-0	0	3.94	0	1.043895	-17.188316	220	1	1.1	0.9;
	680	1	79.59	50.5	0	14.31	0	1.040164	8.217594	380	3	1.1	0.9;
	681	1	0	0	0	14.54	0	1.062258	-37.395143	380	1	1.1	0.9;
	682	2	0	0	0	0	0	1.060787	-10.546299	220	5	1.1	0.9;
	683	1	169.8	55.5	0	3.3	0	0.981733	10.278584	220	3	1.1	0.9;
	684	1	60.07	28	0	4.13	0	0.991967	-50.700151	220	6	1.1	0.9;
	685	1	0	0	0	47.77	0	1.045493	-14.845324	380	6	1.1	0.9;
	686	2	0	0	0	0	0	1.013903	-47.02336	220	4	1.1	0.9;
	687	1	-0	0	0	5.6	0	1.053586	-4.477068	220	5	1.1	0.9;
	688	1	-3.4	-1.3	0	2.27	0	1.024501	-24.443107	220	7	1.1	0.9;
	689	1	124.8	23.5	0	5.84	0	0.982084	-19.04446	220	16	1.1	0.9;
	690	1	22.42	8.2	0	0.53	0	1.056369	-29.885233	110	3	1.1	0.9;
	691	1	83.6	11.9	0	3.53	0	1.021389	2.425783	154	1	1.1	0.9;
	692	2	0	0	0	0	0	1.059006	-37.997965	120	6	1.1	0.9;
	693	1	-0	0	0	93.75	0	1.058173	-40.548699	380	4	1.1	0.9;
	694	1	0	0	0.12	6.65	0	1.044245	-24.486833	220	21	1.1	0.9;
	695	1	96.41	26.5	0	0.08	0	0.999388	-38.384235	220	4	1.1	0.9;
	696	1	12.4	6.89	0	0.63	0	1.055967	-22.074281	150	9	1.1	0.9;
	697	2	0	0	0	0	0	1.031719	-38.113646	380	6	1.1	0.9;
	698	2	0	0	0	0	0	0.999662	-29.934873	220	10	1.1	0.9;
	699	1	36.54	8.4	0	6.58	0	1.009348	-53.274369	220	4	1.1	0.9;
	700	1	12.71	0.8	0	0.51	0	0.966856	-29.467501	110	18	1.1	0.9;
	701	1	20.32	6.7	0	0.32	0	1.00342	-37.061425	110	12	1.1	0.9;
	702	1	-0	-0	0	0.48	0	1.004138	-35.804254	150	10	1.1	0.9;
	703	1	0	0	0	7.84	0	1.126072	-21.917059	380	9	1.3	0.7;
	704	1	41.05	14.3	0	11.53	0	1.01824	-29.261283	220	6	1.1	0.9;
	705	1	8.3	1.2	0	1.6	0	1.021944	8.392464	154	1	1.1	0.9;
	706	1	114.8	22.8	0	5.66	0	1.009578	-44.478532	220	4	1.1	0.9;
	707	1	83.99	21.2	0	5.36	0	1.049779	-25.989404	220	5	1.1	0.9;
	708	1	0	-0	0	9.21	0	0.995859	-50.724378	220	6	1.1	0.9;
	709	2	0	0	0	0	0	1.068875	-42.494144	380	4	1.1	0.9;
	710	1	76.38	5.1	0	3.97	0	1.043988	-45.482724	220	4	1.1	0.9;
	711	2	0	0	0	0	0	1.070147	-1.338091	380	14	1.1	0.9;
	712	1	110.4	41.9	0	1.88	0	1.009004	0.655932	150	8	1.1	0.9;
	713	2	0	0	0	0	0	1.034791	-10.09141	110	18	1.1	0.9;
	714	2	0	0	0	0	0	1.065017	-9.000082	380	14	1.1	0.9;
	715	2	0	0	0	0	0	1.021604	6.902749	400	1	1.1	0.9;
	716	1	29.63	10.8	0	2.39	0	1.033505	-24.40114	110	19	1.1	0.9;
	717	1	4.5	-5.6	0	0.09	0	1.013062	-23.916209	110	18	1.1	0.9;
	718	1	135.11	41	0	5.31	0	1.033036	-25.366125	220	5	1.1	0.9;
	719	1	-0.54	-0.99	0	1.13	0	1.061862	-23.123981	220	18	1.1	0.9;
	720	1	-0	-0	0	23.554121	0	0.96634	-27.364834	380	5	1.1	0.9;
	721	1	-0	-0	0	1.49	0	1.022542	-25.272598	220	7	1.1	0.9;
	722	1	0	-0	0	0.78	0	1.026617	0.095694	150	8	1.1	0.9;
	723	1	104.6	14.9	0	7.35	0	1.026037	4.930819	154	1	1.1	0.9;
	724	1	30.83	6.6	0	3.19	0	1.008661	-41.52775	110	22	1.1	0.9;
	725	1	0	-0	0	34.82	0	1.1296	-23.424903	380	9	1.3	0.7;
	726	1	400.3	189.9	0	0.53	0	1.044199	-5.348266	220	5	1.1	0.9;
	727	2	0	0	0	0	0	1.070143	-13.948739	380	19	1.3	0.7;
	728	2	0	0	0	0	0	1.030375	12.049558	154	1	1.1	0.9;
	729	1	0	-0	0	1.1	0	0.972249	-27.155352	150	9	1.1	0.9;
	730	1	20.62	6.5	0	0.72	0	1.007561	-41.217125	110	12	1.1	0.9;
	731	1	0	0	0	61.48	0	1.041231	-50.286139	380	4	1.1	0.9;
	732	1	-0	-0	0	11.12	0	1.030738	18.834742	380	3	1.1	0.9;
	733	1	58.36	10.7	0	8.41	0	1.007989	6.004639	110	8	1.1	0.9;
	734	1	109.23	66.36	0	5.44	0	1.027653	-1.875847	220	13	1.1	0.9;
	735	2	0	0	0	0	0	1.026593	15.765814	154	1	1.1	0.9;
	736	1	45.56	-1.51	0	19.94	0	1.007191	61.64009	150	1	1.1	0.9;
	737	1	8.11	2	0	1.95	0	0.991402	-28.695305	110	18	1.1	0.9;
	738	2	0	0	0	0	0	1.033376	-38.980898	110	22	1.1	0.9;
	739	2	0	0	0	0	0	1.035442	-42.803605	380	4	1.1	0.9;
	740	1	22.22	13.7	0	0.12	0	0.999884	-29.224674	150	9	1.1	0.9;
	741	1	-0	-0	0	0.98	0	0.98413	-21.448945	110	18	1.1	0.9;
	742	1	0	-0	0	2.83	0	1.053755	-4.478576	220	5	1.1	0.9;
	743	1	5.01	2	0	2.99	0	1.067428	-28.110496	220	23	1.1	0.9;
	744	1	6.51	4.1	0	1.27	0	1.032223	-22.075222	150	9	1.1	0.9;
	745	1	90.4	15.2	0	0.07	0	1.010505	-19.123791	220	6	1.1	0.9;
	746	1	2.5	1.3	0	0.07	0	0.955236	-31.252302	110	18	1.1	0.9;
	747	1	69.58	16	0	3.03	0	1.068191	-26.778671	220	5	1.1	0.9;
	748	1	0	-0	0	5.2	0	1.083131	-48.121049	220	6	1.1	0.9;
	749	2	0	0	0	0	0	1.053546	-29.4801	220	5	1.1	0.9;
	750	1	-0	-0	0	18.63	0	1.038715	-24.772095	220	5	1.1	0.9;
	751	2	0	0	0	0	0	1.020488	-44.20282	220	4	1.1	0.9;
	752	1	-0	-0	0	4.15	0	1.036686	-27.828072	220	7	1.1	0.9;
	753	1	1.9	0.6	0	0.38	0	0.998712	-26.728595	110	18	1.1	0.9;
	754	1	132.1	32.4	0	58.29	0	1.044981	-45.303749	380	4	1.1	0.9;
	755	1	28.9	4.2	0	2.09	0	1.011893	-7.18658	154	1	1.1	0.9;
	756	1	36.74	8.9	0	0.63	0	0.9931	-16.99482	110	18	1.1	0.9;
	757	2	0	0	0	0	0	1.013994	-14.009904	380	5	1.1	0.9;
	758	1	0	0	0	0	0	0.972441	-36.764788	150	9	1.1	0.9;
	759	1	-0	-0	0	9.58	0	1.027334	-24.007216	220	6	1.1	0.9;
	760	1	0	-0	0	1.5	0	1.04271	-16.194539	110	18	1.1	0.9;
	761	1	-0	0	0	0.41	0	1.055841	-19.131647	110	19	1.1	0.9;
	762	1	-0	-0	0	0.08	0	1.044236	-29.495612	150	9	1.1	0.9;
	763	2	0	0	0	0	0	1.042899	-30.310969	220	6	1.1	0.9;
	764	2	0	0	0	0	0	1.048246	-3.309099	220	13	1.1	0.9;
	765	2	0	0	0	0	0	1.017174	-35.819032	220	6	1.1	0.9;
	766	1	-0	0	0	7.85	0	1.037046	-19.904905	220	5	1.1	0.9;
	767	1	0	0	0	0	0	0.915156	-35.96451	150	9	1.1	0.9;
	768	1	0	-0	0	0.45	0	0.986644	-23.658422	150	9	1.1	0.9;
	769	1	152.57	40	0	6.12	0	1.020205	-11.261996	220	5	1.1	0.9;
	770	1	-0	0	0	1.23	0	1.061647	-0.922147	220	13	1.1	0.9;
	771	1	46.15	20.9	0	1.4	0	1.038724	-39.6443	110	22	1.1	0.9;
	772	1	343.9	141.9	0	5.55	0	1.037471	-14.231744	220	5	1.1	0.9;
	773	1	-14.87	-4.17	0	0.58	0	0.992021	-20.861422	110	18	1.1	0.9;
	774	2	0	0	0	0	0	1.038992	8.899765	154	1	1.1	0.9;
	775	1	-55.14	-27.14	0	0	0	1.048905	7.312541	150	8	1.1	0.9;
	776	2	0	0	0	0	0	1.023986	-13.449737	220	5	1.1	0.9;
	777	1	0	0	0	1.44	0	1.029983	-16.451254	154	1	1.1	0.9;
	778	2	0	0	0	0	0	1.047837	-6.746098	220	5	1.1	0.9;
	779	2	0	0	0	0	0	1.039304	17.406318	380	3	1.1	0.9;
	780	2	0	0	0	0	0	1.001975	-45.135804	220	10	1.1	0.9;
	781	1	34.04	11	0	0.49	0	1.000347	-44.894504	110	22	1.1	0.9;
	782	1	-0	-0	0	7.28	0	1.044993	-20.921382	220	3	1.1	0.9;
	783	1	0	0	0	0	0	0.982732	-29.85688	150	9	1.1	0.9;
	784	1	-26	-14.72	0	0	0	1.033901	4.031176	150	8	1.1	0.9;
	785	1	-0	-0	0	4.03	0	1.030881	-34.012074	220	10	1.1	0.9;
	786	1	-0	0	0	0.79	0	1.048812	-20.86338	220	1	1.1	0.9;
	787	1	125.7	17.9	0	0.2	0	1.014445	-4.315582	154	1	1.1	0.9;
	788	1	18.02	6.4	0	0.79	0	0.954883	-29.560185	110	18	1.1	0.9;
	789	1	-0	-0	0	6.63	0	1.052564	-6.129951	220	5	1.1	0.9;
	790	1	-0.15	-1	0	0.42	0	1.066475	11.763458	110	8	1.1	0.9;
	791	2	0	0	0	0	0	1.052091	-6.530683	380	13	1.1	0.9;
	792	1	295.3	41.9	0	0	0	1.003673	-3.167772	150	2	1.1	0.9;
	793	1	9.7	1.4	0	4.22	0	1.031031	13.743552	154	1	1.1	0.9;
	794	1	0	-0	0	1.653516	0	0.984037	-18.179598	220	24	1.1	0.9;
	795	2	0	0	0	0	0	1.041931	-15.251585	220	5	1.1	0.9;
	796	2	0	0	0	0	0	1.01604	-43.685946	110	22	1.1	0.9;
	797	1	2	114.6	0	16.15	0	1.017294	-40.077949	220	4	1.1	0.9;
	798	1	235.2	41.1	0	10.01	0	1.015334	14.085955	380	3	1.1	0.9;
	799	1	26.73	10.8	0	0.43	0	0.999312	-30.615831	110	18	1.1	0.9;
	800	1	50.56	18	0	0.11	0	1.029716	-12.414305	220	5	1.1	0.9;
	801	1	9.71	3	0	0.95	0	0.980529	-33.119841	110	21	1.1	0.9;
	802	1	0	-0	0	12.42	0	1.021496	-41.492157	380	6	1.1	0.9;
	803	2	0	0	0	0	0	1.04662	-25.280535	220	5	1.1	0.9;
	804	1	-0	-0	0	7.26	0	1.052624	-7.620343	220	5	1.1	0.9;
	805	1	124.5	29.6	0	15.4	0	1.059049	-11.263382	220	5	1.1	0.9;
	806	1	-0	-0	0	32.02	0	1.013023	6.715226	380	8	1.1	0.9;
	807	2	0	0	0	0	0	1.053518	-38.514381	380	4	1.1	0.9;
	808	1	24.53	8.6	0	7.01	0	1.016881	-50.527441	220	4	1.1	0.9;
	809	1	0	-0	0	58.088687	0	1.024607	-0.360983	380	5	1.1	0.9;
	810	1	0	0	0	0	0	1.014711	-27.065473	150	9	1.1	0.9;
	811	1	17.83	-9.5	0	2.79	0	1.030392	-22.540776	220	7	1.1	0.9;
	812	1	0	0	0	0	0	1.008755	-31.096175	150	9	1.1	0.9;
	813	2	0	0	0	0	0	1.037003	-19.089875	400	1	1.1	0.9;
	814	1	287.7	16.3	0	12.86	0	1.017789	-50.853491	380	4	1.1	0.9;
	815	1	25.27	5.71	0	0.67	0	1.016118	-28.086312	110	19	1.1	0.9;
	816	1	234.7	-12	0	23.35	0	1.056996	-35.266904	380	17	1.1	0.9;
	817	1	333.1	17.5	0	18.86	0	1.01358	-45.581795	380	6	1.1	0.9;
	818	1	0	-0	0	8.73	0	1.030395	51.296143	220	3	1.1	0.9;
	819	1	0	-0	0	0.22	0	1.05081	-0.785035	220	5	1.1	0.9;
	820	1	12.6	1.8	0	1.89	0	1.025873	-0.511882	154	1	1.1	0.9;
	821	1	-0	0	0	22.93	0	1.082078	-36.000114	150	9	1.1	0.9;
	822	1	121.9	-0.2	0	12.83	0	1.003485	-8.072973	220	3	1.1	0.9;
	823	2	0	0	0	0	0	1.039055	-2.419223	380	5	1.1	0.9;
	824	1	32.1	4.5	0	8.93	0	1.046857	-2.975489	154	1	1.1	0.9;
	825	1	0	0	0	29.76	0	1.007201	-36.665305	220	22	1.1	0.9;
	826	1	31.94	7.7	0	1.11	0	1.035423	-32.770026	110	22	1.1	0.9;
	827	1	167.8	77	0	12.66	0	1.018231	-13.13917	380	3	1.1	0.9;
	828	2	0	0	0	0	0	1.044382	2.077536	154	1	1.1	0.9;
	829	1	113.6	122.7	0	8.74	0	1.018129	15.42687	380	3	1.1	0.9;
	830	1	303.1	75.5	0	22.74	0	1.022107	-40.853901	380	6	1.1	0.9;
	831	2	0	0	0	0	0	1.003271	-26.364611	110	18	1.1	0.9;
	832	2	0	0	0	0	0	1.004583	-50.562152	220	6	1.1	0.9;
	833	1	15.02	10	0	0.55	0	1.046835	-26.285018	110	23	1.1	0.9;
	834	1	52.76	17.2	0	38.38	0	1.008384	-37.145585	220	6	1.1	0.9;
	835	1	-0	0	0	24.4	0	1.034014	-24.956153	380	6	1.1	0.9;
	836	1	-0	-0	0	56.7	0	1.012335	-21.59407	400	1	1.1	0.9;
	837	2	0	0	0	0	0	1.022737	-40.410086	220	6	1.1	0.9;
	838	1	615.81	286.69	0	101.04	0	1.030987	3.606486	150	8	1.1	0.9;
	839	1	-0	-0	0	8.17514	0	1.02122	-1.687736	380	5	1.1	0.9;
	840	1	0	0	0	2.7	0	1.022301	-23.5857	150	9	1.1	0.9;
	841	1	-0	0	0.135375	32.572175	0	1.08276	-21.360294	380	18	1.3	0.7;
	842	1	0	0	0	2.5	0	1.025785	-5.013219	220	3	1.1	0.9;
	843	1	-0	0	0	0.64	0	1.017591	-40.075365	110	1	1.1	0.9;
	844	2	0	0	0	0	0	1.064419	-9.125625	380	14	1.1	0.9;
	845	1	-0	0	0	1.11	0	1.003306	-31.349574	150	9	1.1	0.9;
	846	1	159.2	59.9	0	37.33	0	1.056518	-14.954697	380	16	1.1	0.9;
	847	1	13.9	1.9	0	7.37	0	1.027274	19.174428	154	1	1.1	0.9;
	848	2	0	0	0	0	0	1.011431	61.818191	150	1	1.1	0.9;
	849	2	0	0	0	0	0	1.037264	-42.024951	380	4	1.1	0.9;
	850	1	-0	-0	0	0.47	0	0.956219	-30.232358	110	18	1.1	0.9;
	851	2	0	0	0	0	0	1.018845	-16.616591	220	5	1.1	0.9;
	852	1	-0	0	0	5.46	0	1.007493	-25.256407	154	1	1.1	0.9;
	853	1	190	43.3	0	12.94	0	1.023268	-12.714531	220	3	1.1	0.9;
	854	1	49.96	7.5	0	8.86	0	1.023709	-54.456583	220	4	1.1	0.9;
	855	1	10.01	3.3	0	0.24	0	1.027869	-34.087073	110	12	1.1	0.9;
	856	1	-28.93	9.5	0	0.38	0	0.976644	-23.667349	150	9	1.1	0.9;
	857	1	21.72	4.8	0	1.26	0	0.96854	-30.940544	110	18	1.1	0.9;
	858	2	0	0	0	0	0	1.028813	0.822714	380	5	1.1	0.9;
	859	1	0	0	0	0.28	0	1.093242	-27.533273	150	9	1.1	0.9;
	860	1	32.1	4.5	0	4.34	0	1.033912	1.449836	154	1	1.1	0.9;
	861	2	0	0	0	0	0	1.050818	-30.333474	110	22	1.1	0.9;
	862	1	36.44	12.5	0	0.59	0	1.011149	-30.106886	110	21	1.1	0.9;
	863	1	111.4	15.9	0	2.36	0	1.00914	-7.642598	154	1	1.1	0.9;
	864	1	1.3	0.6	0	1.23	0	0.95494	-31.154654	110	18	1.1	0.9;
	865	1	43.45	26.9	0	3.56	0	1.08382	-28.028856	150	9	1.1	0.9;
	866	1	-0	-0	0	20.07	0	1.055959	43.766007	220	3	1.1	0.9;
	867	1	25.13	14	0	0.04	0	1.020409	-5.272998	220	3	1.1	0.9;
	868	1	13.82	5	0	0.11	0	1.050793	-21.666498	110	19	1.1	0.9;
	869	1	94.2	5.5	0	6.51	0	1.021539	-47.087083	220	4	1.1	0.9;
	870	1	110	3.4	0	11.95	0	1.038333	-24.930069	220	5	1.1	0.9;
	871	1	108.89	-21.76	0	3.72	0	1.047396	-12.310945	220	5	1.1	0.9;
	872	1	185.6	26.4	0	32.78	0	1.018555	-24.221949	154	1	1.1	0.9;
	873	1	143.9	123.2	0	72.64	0	1.029121	5.127443	380	3	1.1	0.9;
	874	1	0	0	0	0	0	0.967758	-31.427129	150	9	1.1	0.9;
	875	2	0	0	0	0	0	1.038959	-35.716579	380	6	1.1	0.9;
	876	1	42.25	13.9	0	0.95	0	1.012126	-26.966924	110	21	1.1	0.9;
	877	1	2	0.6	0	0.53	0	1.019785	-29.346446	110	21	1.1	0.9;
	878	1	0	-0	0	25.28	0	1.032189	9.787383	150	8	1.1	0.9;
	879	1	7.71	3.3	0	0.56	0	1.003179	-44.618548	110	22	1.1	0.9;
	880	1	0	-0	0	2	0	1.026334	-13.567049	220	6	1.1	0.9;
	881	1	-0	0	0	1.47	0	1.01993	-22.777984	150	9	1.1	0.9;
	882	1	394.9	77	0	44.93	0	1.022697	-46.380115	380	6	1.1	0.9;
	883	1	192.5	78.3	0	6.65	0	1.015357	-10.191051	220	5	1.1	0.9;
	884	1	13.82	5	0	0.63	0	1.018802	-30.29798	110	19	1.1	0.9;
	885	1	23.03	14.3	0	6.61	0	0.990763	-30.757298	150	9	1.1	0.9;
	886	1	-2.85	-7.24	0	0.07	0	1.038441	10.699656	220	3	1.1	0.9;
	887	1	0	0	0.3205	28.133148	0	1.068754	-21.629845	380	18	1.1	0.9;
	888	2	0	0	0	0	0	1.037604	-21.604862	220	19	1.1	0.9;
	889	1	3.9	1.4	0	0.78	0	1.014037	-29.373377	110	19	1.1	0.9;
	890	1	8.91	3.8	0	0.38	0	0.992571	-28.524136	110	18	1.1	0.9;
	891	2	0	0	0	0	0	1.062964	-19.250317	380	5	1.1	0.9;
	892	1	-0	-0	0	9.1	0	1.06496	-35.278849	220	5	1.1	0.9;
	893	1	-0	-0	0	1.77	0	1.035537	-36.23179	220	6	1.1	0.9;
	894	2	0	0	0	0	0	1.054199	-24.304132	220	21	1.1	0.9;
	895	1	-0	-0	0.234854	36.333138	0	1.07469	-26.738864	380	21	1.1	0.9;
	896	2	0	0	0	0	0	1.021987	2.509919	154	1	1.1	0.9;
	897	1	153.5	21.9	0	7.79	0	1.034443	-20.896007	154	1	1.1	0.9;
	898	1	25.63	7.9	0	0.77	0	1.016315	12.999579	110	8	1.1	0.9;
	899	1	-0	-0.01	0	59.04	0	1.038888	-19.131057	380	6	1.1	0.9;
	900	1	0	0	0	0.15	0	1.041667	-15.788508	220	5	1.1	0.9;
	901	1	0	0	0	0	0	0.959473	-28.847628	150	9	1.1	0.9;
	902	1	-4.94	-3.67	0	0.11	0	1.015036	-23.101701	110	18	1.1	0.9;
	903	1	82.69	20.1	0	1.36	0	1.068153	-26.804113	220	5	1.1	0.9;
	904	1	146.9	60.6	0	26.49	0	1.007247	-37.013111	220	6	1.1	0.9;
	905	1	0	-0	0	18.004129	0	1.04044	-15.095745	380	5	1.1	0.9;
	906	1	-0	0	0	0.23	0	0.960197	-29.959435	150	9	1.1	0.9;
	907	1	-0	-0	0	0.05	0	1.038463	-9.834075	220	5	1.1	0.9;
	908	1	125.9	-0.5	0	0.21	0	1.054939	-8.182367	220	5	1.1	0.9;
	909	1	12.61	4.5	0	2.63	0	1.05269	-24.666613	110	15	1.1	0.9;
	910	1	6.3	0.9	0	3.82	0	1.013902	-5.883269	154	1	1.1	0.9;
	911	1	13.21	2.2	0	10.05	0	1.021752	-41.209614	110	22	1.1	0.9;
	912	1	22.02	13.2	0	0.03	0	1.091269	-27.621559	150	9	1.1	0.9;
	913	2	0	0	0	0	0	1.026545	-2.255482	380	2	1.1	0.9;
	914	1	-0	0	0	89.51	0	1.077351	-24.737334	380	9	1.1	0.9;
	915	1	24.53	7.9	0	0.63	0	1.037609	-26.120005	110	19	1.1	0.9;
	916	2	0	0	0	0	0	1.022135	-38.115322	220	4	1.1	0.9;
	917	2	0	0	0	0	0	1.061631	-20.525158	110	15	1.1	0.9;
	918	1	110.1	29.4	0	1.11	0	1.006154	-30.037871	110	21	1.1	0.9;
	919	1	-0	0	0	13.89	0	1.016781	11.540968	380	8	1.1	0.9;
	920	1	34.44	8.5	0	1.11	0	1.042446	-21.074337	220	5	1.1	0.9;
	921	1	-0	-0	0.126343	6.455031	0	1.027687	-21.517935	220	19	1.1	0.9;
	922	1	15.22	7.6	0	1.56	0	0.995692	-17.550034	110	18	1.1	0.9;
	923	1	62.07	14.7	0	4.84	0	1.031651	-23.442192	220	5	1.1	0.9;
	924	1	10.01	5	0	0.11	0	0.997399	-50.006382	220	6	1.1	0.9;
	925	1	8.01	4.1	0	2.25	0	0.979496	-16.132225	110	18	1.1	0.9;
	926	2	0	0	0	0	0	1.008066	-0.465497	400	1	1.1	0.9;
	927	2	0	0	0	0	0	1.067901	-48.107874	220	6	1.1	0.9;
	928	1	112.9	19.3	0	17.34	0	1.015527	0.368562	150	8	1.1	0.9;
	929	1	43.75	12.4	0	9.35	0	1.021929	-46.228732	220	4	1.1	0.9;
	930	2	0	0	0	0	0	1.058189	-37.830017	380	4	1.1	0.9;
	931	1	0	0	0	0	0	1.00554	-33.793222	150	9	1.1	0.9;
	932	1	77.59	12.3	0	5.1	0	1.025765	-53.984442	220	4	1.1	0.9;
	933	1	-0	0	0	22.75	0	1.020036	-21.567564	220	5	1.1	0.9;
	934	1	2.5	1.3	0	2.33	0	0.999076	-29.319835	110	18	1.1	0.9;
	935	1	-0	0	0	9.46	0	1.044831	-24.797521	220	5	1.1	0.9;
	936	1	0	-0	0	1.93	0	0.999872	-27.063751	150	9	1.1	0.9;
	937	1	-12.51	4.1	0	0.09	0	0.998116	-32.130798	150	9	1.1	0.9;
	938	1	0	-0	0	31.47	0	1.054487	36.893694	380	3	1.1	0.9;
	939	1	150.1	35	0	3.55	0	1.022797	-23.045862	220	6	1.1	0.9;
	940	1	50.06	16.3	0	0.26	0	1.045541	-28.687115	110	21	1.1	0.9;
	941	1	26.03	9.8	0	0.82	0	0.990209	-24.492637	110	18	1.1	0.9;
	942	1	0	0	0	34.26	0	1.084754	-14.407963	380	1	1.3	0.7;
	943	1	-0	0	0	3.37	0	1.02621	-23.16922	220	7	1.1	0.9;
	944	1	30.33	11	0	1.43	0	1.032883	-20.178508	110	19	1.1	0.9;
	945	1	0.2	0.1	0	0.5	0	1.034208	-32.420002	110	12	1.1	0.9;
	946	1	189.1	-11.9	0	27.462397	0	1.020348	-48.526559	380	6	1.1	0.9;
	947	1	90.7	5.8	0	4.1	0	0.983073	-49.782659	220	4	1.1	0.9;
	948	1	15.82	4	0	0.7	0	0.985851	-55.797711	150	10	1.1	0.9;
	949	1	4.81	2.2	0	0.62	0	0.956304	-29.76252	110	18	1.1	0.9;
	950	1	69.08	22.6	0	0.54	0	1.020914	-0.217378	150	8	1.1	0.9;
	951	1	0.5	-271.9	0	3.12	0	1.047979	-16.169837	380	3	1.1	0.9;
	952	1	0	0	0	0.38	0	1.078738	-22.026301	150	9	1.1	0.9;
	953	1	108.7	-12.15	0	37.11	0	1.062501	-13.250544	380	5	1.1	0.9;
	954	1	66.27	13.6	0	0.67	0	1.040885	-21.822155	220	5	1.1	0.9;
	955	1	244.2	152.7	0	0	0	1.003862	-11.442937	150	2	1.1	0.9;
	956	1	19.42	11.6	0	0	0	1.013505	-13.078642	110	24	1.1	0.9;
	957	1	65.77	16.1	0	1.66	0	1.010333	-46.607879	220	4	1.1	0.9;
	958	1	-0	0	0	51.29	0	1.052882	-31.072147	380	11	1.1	0.9;
	959	1	166.6	8.1	0	10.11	0	1.015348	-14.224034	380	3	1.1	0.9;
	960	1	-0	-0	0	108.728602	0	1.057857	-9.331313	380	5	1.1	0.9;
	961	1	32.94	20.4	0	8.95	0	0.87466	-38.032264	150	9	1.3	0.7;
	962	1	42.55	17.1	0	1.67	0	1.03371	-46.108222	220	4	1.1	0.9;
	963	1	0	0	0	12.74	0	1.036965	-32.220515	220	12	1.1	0.9;
	964	1	88.5	16.5	0	146.44	0	1.011961	-11.884786	380	5	1.1	0.9;
	965	1	45.05	9.9	0	0.24	0	1.008014	-41.589049	110	22	1.1	0.9;
	966	1	190.91	55.46	0	14.77	0	1.031333	13.632899	220	13	1.1	0.9;
	967	1	1.3	0.4	0	0.64	0	0.990771	-21.703522	110	18	1.1	0.9;
	968	1	-0	0	0	49.26	0	1.007664	8.988419	400	1	1.1	0.9;
	969	1	0.4	1.1	0	1.99	0	1.02364	31.655804	380	3	1.1	0.9;
	970	1	62.27	-2.1	0	8.01	0	1.01672	-42.68566	220	4	1.1	0.9;
	971	1	0	0	0.099501	68.573595	0	1.0784	-18.292644	380	18	1.1	0.9;
	972	2	0	0	0	0	0	1.078612	5.133314	380	5	1.3	0.7;
	973	1	0	0	0	0	0	1.006917	-29.683321	150	9	1.1	0.9;
	974	2	0	0	0	0	0	1.033045	-33.938069	110	12	1.1	0.9;
	975	1	33.24	11.1	0	1.56	0	0.983715	-35.520739	110	21	1.1	0.9;
	976	2	0	0	0	0	0	1.024328	-50.054572	380	4	1.1	0.9;
	977	1	95.58	-14	0	4.48	0	1.037866	-11.661106	220	13	1.1	0.9;
	978	2	0	0	0	0	0	1.017319	-26.107578	110	18	1.1	0.9;
	979	1	21.12	7.7	0	1.48	0	0.966427	-22.917522	110	19	1.1	0.9;
	980	1	14.71	4.6	0	7.88	0	1.042778	-25.471141	220	5	1.1	0.9;
	981	1	0	-0	0	0.84	0	1.019705	-44.544584	220	4	1.1	0.9;
	982	1	62.7	9	0	2.94	0	1.019431	7.055187	154	1	1.1	0.9;
	983	2	0	0	0	0	0	1.078039	44.244769	380	3	1.1	0.9;
	984	2	0	0	0	0	0	1.031963	-38.021587	380	6	1.1	0.9;
	985	1	17.52	4.6	0	11.82	0	1.053659	-42.273928	220	6	1.1	0.9;
	986	1	15.98	0.86	0	5.96	0	1.016651	-52.854277	220	4	1.1	0.9;
	987	1	-0	0	0	1.18	0	0.998791	-39.118465	150	10	1.1	0.9;
	988	1	-0	-0	0	8.62	0	1.02316	-53.019946	220	4	1.1	0.9;
	989	1	32.34	10.6	0	0.64	0	0.974921	-32.236537	110	21	1.1	0.9;
	990	1	-0	0	0	8.19	0	0.982837	-34.864842	150	9	1.1	0.9;
	991	2	0	0	0	0	0	1.001634	-23.735457	110	18	1.1	0.9;
	992	2	0	0	0	0	0	1.030255	-25.800054	220	6	1.1	0.9;
	993	2	0	0	0	0	0	1.037248	2.05329	380	13	1.1	0.9;
	994	1	-0	-0	0	25	0	1.025373	-1.14671	150	8	1.1	0.9;
	995	1	0	-0	0	7.66	0	1.045457	-33.794469	220	11	1.1	0.9;
	996	1	0	0	0	0.07	0	0.984075	-16.065522	110	18	1.1	0.9;
	997	1	0	-0	0	0.84	0	1.036677	-29.848168	150	9	1.1	0.9;
	998	1	22.32	13.8	0	2.96	0	1.027792	-29.418146	150	9	1.1	0.9;
	999	1	20.02	6.8	0	0.48	0	1.013381	19.8472	110	8	1.1	0.9;
	1000	1	0	-0	0	1.8	0	1.018423	-33.552172	220	1	1.1	0.9;
	1001	2	0	0	0	0	0	1.069874	-33.89876	220	5	1.1	0.9;
	1002	2	0	0	0	0	0	1.03837	-20.387948	220	5	1.1	0.9;
	1003	1	24.03	8	0	0.33	0	0.974669	-32.950245	110	21	1.1	0.9;
	1004	2	0	0	0	0	0	0.99944	9.540224	220	3	1.1	0.9;
	1005	1	0	0	0	0.81	0	1.059735	-12.798107	220	5	1.1	0.9;
	1006	1	-0	-0	0	45.31	0	1.082777	-21.788125	380	9	1.1	0.9;
	1007	1	-0	-0	0	1.51	0	1.017117	-26.990387	150	9	1.1	0.9;
	1008	1	0	-0	0	0.68	0	0.976559	-32.083934	110	18	1.1	0.9;
	1009	1	0.2	0.4	0	0.73	0	1.021136	-10.635216	220	3	1.1	0.9;
	1010	1	21.32	7.3	0	0.6	0	1.019454	22.573664	110	8	1.1	0.9;
	1011	1	-0.2	-1	0	0	0	1.028562	-1.860402	150	8	1.1	0.9;
	1012	1	-0	-0	0	0.87	0	1.004772	-32.945011	150	9	1.1	0.9;
	1013	1	136.06	40.4	0	12.08	0	0.992433	-47.062443	220	4	1.1	0.9;
	1014	1	-0	0	0	1.07	0	1.009412	-30.80973	150	9	1.1	0.9;
	1015	1	44.55	0	0	4.78	0	1.056717	-5.590237	220	5	1.1	0.9;
	1016	1	-0	-0	0	15.04	0	1.054952	-41.035126	380	1	1.1	0.9;
	1017	1	26.93	9.5	0	7.65	0	1.04868	-23.788662	220	7	1.1	0.9;
	1018	1	21.02	7	0	0.42	0	0.978355	-32.342729	110	21	1.1	0.9;
	1019	1	0	0	0	0.31	0	1.019661	-37.5747	110	1	1.1	0.9;
	1020	1	8.21	3.31	0	8.38	0	1.043617	6.058453	150	8	1.1	0.9;
	1021	2	0	0	0	0	0	1.012345	-46.728053	220	4	1.1	0.9;
	1022	1	0	0	0	0.32	0	0.992408	-17.735225	110	18	1.1	0.9;
	1023	1	-0	0	0	0.93	0	1.012791	-44.581771	220	4	1.1	0.9;
	1024	1	27.9	4	0	0.29	0	1.03027	-12.097523	154	1	1.1	0.9;
	1025	1	241.4	-2.5	0.224624	31.280331	0	1.023682	-52.453183	380	10	1.1	0.9;
	1026	1	201.4	53	0	9.67	0	1.027568	-29.827138	220	5	1.1	0.9;
	1027	1	-5.91	-2.33	0	30.393595	0	0.985955	-18.490902	380	5	1.1	0.9;
	1028	1	13.82	5	0	0.45	0	1.003783	-27.631688	110	19	1.1	0.9;
	1029	2	0	0	0	0	0	1.037886	65.155421	220	3	1.1	0.9;
	1030	2	0	0	0	0	0	1.007808	-36.442619	220	4	1.1	0.9;
	1031	1	84.59	-18.5	0	64.57	0	1.037625	-39.011568	380	6	1.1	0.9;
	1032	1	27.03	13	0	2.32	0	1.02192	-29.453136	220	6	1.1	0.9;
	1033	1	26.33	5.3	0	4.31	0	1.038506	-24.849537	220	5	1.1	0.9;
	1034	1	-0	-0	0	14.23	0	1.019297	-29.061837	220	6	1.1	0.9;
	1035	1	382.5	57.9	0	8.16	0	1.066609	-32.368109	220	5	1.1	0.9;
	1036	2	0	0	0	0	0	1.037074	-31.532599	220	10	1.1	0.9;
	1037	1	-0	0	0	0.55	0	1.018031	4.736977	380	2	1.1	0.9;
	1038	1	3.7	-0	0	0.64	0	1.008041	-20.370492	110	18	1.1	0.9;
	1039	1	34.64	-0	0	0.91	0	1.037717	-19.559946	220	5	1.1	0.9;
	1040	1	31.94	8.8	0	3.79	0	1.038149	-9.088994	220	5	1.1	0.9;
	1041	1	12.11	4.4	0	0.95	0	0.996584	-44.290466	110	22	1.1	0.9;
	1042	1	34.2	4.9	0	9.08	0	1.013159	-5.711376	154	1	1.1	0.9;
	1043	2	0	0	0	0	0	1.027465	-3.716213	380	5	1.1	0.9;
	1044	1	40.5	5.7	0	3.2	0	1.015463	4.567787	154	1	1.1	0.9;
	1045	1	8.91	5.1	0	1.06	0	0.939467	-31.301987	110	18	1.1	0.9;
	1046	1	47.05	19.3	0	0.04	0	1.013503	-4.847615	150	8	1.1	0.9;
	1047	1	1.9	0.8	0	0.65	0	0.989534	-28.010272	110	18	1.1	0.9;
	1048	1	27.8	-2.03	0	0.5	0	0.985687	-28.380151	110	18	1.1	0.9;
	1049	1	-0	-0	0	28.8	0	1.135903	-22.296959	380	1	1.3	0.7;
	1050	2	0	0	0	0	0	1.055375	7.398452	220	13	1.1	0.9;
	1051	1	-0	0	0	3.79	0	1.049425	-13.263849	220	5	1.1	0.9;
	1052	1	67.48	32.6	0	13.11	0	1.032925	21.769886	220	8	1.1	0.9;
	1053	1	39.14	13.2	0	2.92	0	0.973855	-28.223475	110	18	1.1	0.9;
	1054	1	83.6	11.9	0	6.56	0	1.038498	3.951777	154	1	1.1	0.9;
	1055	1	0	0	0	0	0	0.961777	-31.641638	150	9	1.1	0.9;
	1056	1	0	0	0	0	0	1.048731	-29.212214	150	9	1.1	0.9;
	1057	1	-0	-0	0	0	0	1.040602	25.720393	220	8	1.1	0.9;
	1058	1	0	-0	0	11.96	0	1.011927	-4.94966	380	3	1.1	0.9;
	1059	1	-0	0	0	62.76	0	1.077115	-24.805028	380	9	1.1	0.9;
	1060	1	-0	0	0	75.38	0	1.063056	-42.078575	380	4	1.1	0.9;
	1061	1	-0	0.01	0	0	0	0.991146	-30.797621	150	9	1.1	0.9;
	1062	1	84.89	32.4	0	7.77	0	1.034796	-21.679472	220	7	1.1	0.9;
	1063	1	50.06	17	0	1.61	0	1.02324	-26.441023	110	21	1.1	0.9;
	1064	1	215.2	-7.6	0	1.24	0	1.020517	-47.401359	220	4	1.1	0.9;
	1065	1	89.6	20	0	9.07	0	1.025911	-54.439585	220	4	1.1	0.9;
	1066	1	0	-0	0	9.67	0	1.014112	22.983748	380	3	1.1	0.9;
	1067	1	279.4	22.9	0	15.24	0	1.030938	-9.616423	220	3	1.1	0.9;
	1068	1	35.24	11.6	0	1.72	0	1.024401	-34.284548	110	12	1.1	0.9;
	1069	1	-0	0	0	8.72	0	1.037185	-17.757679	220	1	1.1	0.9;
	1070	1	1.6	-0	0	0.44	0	1.008505	-19.977089	110	18	1.1	0.9;
	1071	1	83.3	47.9	0	6.19	0	1.003174	-36.685975	150	9	1.1	0.9;
	1072	1	-0	-0	0	68.81	0	1.039691	-19.16149	380	6	1.1	0.9;
	1073	1	21.92	7.2	0	3.14	0	1.026076	-29.884032	110	19	1.1	0.9;
	1074	1	403.9	-0.1	0	143.1	0	1.04095	-43.559156	380	4	1.1	0.9;
	1075	1	9.08	3.94	0	0.12	0	1.064906	-12.957107	110	24	1.1	0.9;
	1076	1	0	0	0	0	0	0.968504	-37.395062	150	9	1.1	0.9;
	1077	1	-0	0	0	0.1	0	1.030146	-12.380039	220	5	1.1	0.9;
	1078	1	-0	-0	0	4.74	0	1.061677	-3.667956	220	5	1.1	0.9;
	1079	1	0.2	1.3	0.337525	10.582137	0	1.015897	3.296097	380	3	1.1	0.9;
	1080	1	48.8	6.9	0	7.95	0	1.048144	-9.51724	154	1	1.1	0.9;
	1081	1	277.86	58.95	0	182.880644	0	1.035902	-10.739144	380	5	1.1	0.9;
	1082	2	0	0	0	0	0	1.027941	-33.928194	380	6	1.1	0.9;
	1083	2	0	0	0	0	0	1.050507	-8.794066	220	5	1.1	0.9;
	1084	1	42.15	14.2	0	1.71	0	1.028976	-28.282323	110	19	1.1	0.9;
	1085	1	162.1	43.2	0	3.32	0	1.00298	-8.37152	220	3	1.1	0.9;
	1086	1	12.01	5.6	0	39.29	0	1.023093	-47.400901	380	4	1.1	0.9;
	1087	1	-0	0	0	3.75	0	0.984572	-32.181246	110	21	1.1	0.9;
	1088	2	0	0	0	0	0	1.025882	4.447559	154	1	1.1	0.9;
	1089	1	-7.61	2.5	0	0.13	0	1.038058	-23.070431	150	9	1.1	0.9;
	1090	1	56.06	18.7	0	4.24	0	1.004865	-22.053504	220	5	1.1	0.9;
	1091	1	-24.85	-3.56	0	1.31	0	1.018617	-14.616679	110	18	1.1	0.9;
	1092	1	30.93	-46.6	0	1.62	0	1.032023	-38.196172	220	6	1.1	0.9;
	1093	2	0	0	0	0	0	1.024587	-8.719591	380	5	1.1	0.9;
	1094	1	28.53	11.3	0	3.25	0	1.039718	5.490813	150	8	1.1	0.9;
	1095	1	-0	0	0	43.02	0	1.070027	-18.934012	380	1	1.1	0.9;
	1096	1	0	-0	0	0.16	0	1.048892	-1.184862	220	5	1.1	0.9;
	1097	2	0	0	0	0	0	1.013606	-7.316502	154	1	1.1	0.9;
	1098	2	0	0	0	0	0	1.104358	-15.215621	110	19	1.3	0.7;
	1099	1	0	0	0	0	0	0.998984	-30.306116	150	9	1.1	0.9;
	1100	2	0	0	0	0	0	1.042301	-4.415824	220	5	1.1	0.9;
	1101	1	-0	-0	0	2.05	0	1.035848	-14.267458	220	5	1.1	0.9;
	1102	2	0	0	0	0	0	1.048163	-6.85459	220	5	1.1	0.9;
	1103	2	0	0	0	0	0	1.043526	28.251143	380	13	1.1	0.9;
	1104	1	5.11	3.8	0	0.88	0	0.964886	-28.982903	110	18	1.1	0.9;
	1105	1	29.83	7.3	0	6.49	0	1.069031	-26.723445	220	5	1.1	0.9;
	1106	1	192.4	63.9	0	10	0	1.046946	4.257387	150	8	1.1	0.9;
	1107	1	1.8	0.6	0	0.52	0	0.973029	-23.099573	110	19	1.1	0.9;
	1108	1	-0	-0	0	0.4	0	1.052123	-29.359334	150	9	1.1	0.9;
	1109	1	-54.06	13.1	0	14.23	0	1.024606	-35.185395	220	10	1.1	0.9;
	1110	1	227.34	236.13	0	66.84	0	1.035989	11.686414	380	13	1.1	0.9;
	1111	1	0	0	0	1.1	0	1.03448	-9.0907	220	5	1.1	0.9;
	1112	1	10.91	3.6	0	0.4	0	1.033389	-33.610217	110	12	1.1	0.9;
	1113	1	20.9	3	0	5.46	0	1.020325	-0.47106	154	1	1.1	0.9;
	1114	1	59.27	5.1	0	26.03	0	1.047194	-39.000121	380	4	1.1	0.9;
	1115	1	-40.54	9.3	0	3.05	0	1.023643	-39.410745	220	4	1.1	0.9;
	1116	1	70.28	58.9	0	7.33	0	1.018934	14.925452	380	3	1.1	0.9;
	1117	2	0	0	0	0	0	1.009173	11.392662	400	1	1.1	0.9;
	1118	1	254	74.3	0	54.48	0	1.045067	-18.777768	380	16	1.1	0.9;
	1119	2	0	0	0	0	0	1.044725	-17.505101	220	19	1.1	0.9;
	1120	1	244.8	119.3	0	3.78	0	1.01817	-14.430025	220	3	1.1	0.9;
	1121	2	0	0	0	0	0	1.017488	-39.332774	220	4	1.1	0.9;
	1122	2	0	0	0	0	0	1.043868	-23.808841	220	7	1.1	0.9;
	1123	1	-9	-10.19	0	0	0	1.025795	16.953171	150	8	1.1	0.9;
	1124	1	-0	-0	0	8.61	0	1.063307	14.826169	220	8	1.1	0.9;
	1125	2	0	0	0	0	0	1.04319	-30.889362	220	12	1.1	0.9;
	1126	1	0	-0	0	0	0	1.021722	4.402401	400	1	1.1	0.9;
	1127	1	0	-0	0.086633	0.514111	0	1.079314	-12.470313	220	18	1.1	0.9;
	1128	1	-878.1	-312.7	0	31.21	0	1.048934	-7.971599	380	3	1.1	0.9;
	1129	1	190.47	55.03	0	6.7	0	0.99805	-22.902706	220	5	1.1	0.9;
	1130	1	22.5	3.2	0	5.11	0	1.029873	-0.92219	154	1	1.1	0.9;
	1131	2	0	0	0	0	0	1.036981	-19.775059	380	6	1.1	0.9;
	1132	1	7.8	50.9	0	87.765989	0	1.08516	-20.849986	380	9	1.1	0.9;
	1133	1	115.2	319.2	0	290.91	0	1.054793	-43.968979	380	4	1.1	0.9;
	1134	1	103.9	14.8	0	1.91	0	1.018158	-9.960622	154	1	1.1	0.9;
	1135	1	1.4	0.6	0	0.41	0	0.966502	-25.233439	110	18	1.1	0.9;
	1136	1	16.22	2	0	0.71	0	1.031372	-18.80801	220	5	1.1	0.9;
	1137	1	-0	-0	0	0.05	0	1.04431	-24.755071	220	5	1.1	0.9;
	1138	2	0	0	0	0	0	1.088765	7.13698	380	14	1.1	0.9;
	1139	2	0	0	0	0	0	1.054186	-24.325848	220	21	1.1	0.9;
	1140	1	0	0	0	1.78	0	1.009072	-49.276824	220	6	1.1	0.9;
	1141	1	0	0	0	23.36	0	1.031688	-8.571048	380	7	1.1	0.9;
	1142	1	45.9	36.8	0	55.224909	0	1.0762	-24.508279	380	9	1.1	0.9;
	1143	1	0	0	0.210571	8.306088	0	1.030916	-22.356332	220	19	1.1	0.9;
	1144	1	39.04	-5	0	1.31	0	1.009634	36.656414	220	3	1.1	0.9;
	1145	2	0	0	0	0	0	1.040382	-21.673882	220	11	1.1	0.9;
	1146	1	32.84	-0	0	1.41	0	1.028294	-5.941826	220	5	1.1	0.9;
	1147	1	20.02	15	0	1.56	0	0.982943	9.189194	220	3	1.1	0.9;
	1148	1	141.4	123	0	8.52	0	1.015023	2.923453	380	3	1.1	0.9;
	1149	1	0	-0	0	0.32	0	1.029066	-32.126057	150	9	1.1	0.9;
	1150	1	0	0	0	24.6	0	1.032236	26.297269	380	3	1.1	0.9;
	1151	1	-0	-0	0	18.82	0	1.040253	-20.910202	220	5	1.1	0.9;
	1152	1	-0	0	0	46.01	0	0.992231	1.048056	400	1	1.1	0.9;
	1153	1	12.41	3.9	0	2.23	0	1.031173	-6.442651	220	5	1.1	0.9;
	1154	1	55.7	7.9	0	2.62	0	1.011472	-24.564525	154	1	1.1	0.9;
	1155	1	-0	0	0	1.91	0	1.036534	-51.990549	380	10	1.1	0.9;
	1156	1	65.67	24	0	9.95	0	1.03401	-22.848019	220	5	1.1	0.9;
	1157	1	3.7	0.6	0	0.01	0	1.031019	13.742187	154	1	1.1	0.9;
	1158	1	24.13	12.7	0	0.1	0	0.995578	-30.941271	110	18	1.1	0.9;
	1159	1	112.4	13	0	3.99	0	1.047767	-36.539593	220	5	1.1	0.9;
	1160	1	0.1	0.5	0	0.48	0	1.0023	-0.229629	220	3	1.1	0.9;
	1161	1	170.2	60.8	0	16.27	0	1.034371	8.031714	380	3	1.1	0.9;
	1162	1	0	0	0	2.22	0	1.031404	-21.894084	220	1	1.1	0.9;
	1163	1	-0	-0	0	16.04	0	0.991015	-1.71751	380	2	1.1	0.9;
	1164	1	26.43	16.4	0	0.02	0	0.989655	-30.821299	150	9	1.1	0.9;
	1165	1	0	-0	0	14.13	0	1.0392	17.393702	380	3	1.1	0.9;
	1166	1	-39.81	-10.9	0	3.37	0	1.031803	-29.22026	150	9	1.1	0.9;
	1167	1	0	0	0	0	0	1.012986	-32.963293	150	9	1.1	0.9;
	1168	1	0	-0	0	89.06	0	1.040823	-42.711663	380	4	1.1	0.9;
	1169	1	116.2	16.5	0	7.8	0	1.017798	-22.591387	154	1	1.1	0.9;
	1170	1	-9.51	-3.32	0	0	0	0.991448	-50.487819	220	6	1.1	0.9;
	1171	1	79.89	29	0	1.42	0	1.038903	-25.783048	110	19	1.1	0.9;
	1172	1	-0	-0	0	2.3	0	1.068801	-25.946983	220	5	1.1	0.9;
	1173	2	0	0	0	0	0	1.018544	-41.818492	220	4	1.1	0.9;
	1174	2	0	0	0	0	0	1.002801	-38.484166	150	10	1.1	0.9;
	1175	2	0	0	0	0	0	1.038334	-39.844005	220	4	1.1	0.9;
	1176	2	0	0	0	0	0	1.016193	-23.120961	110	18	1.1	0.9;
	1177	1	-2.35	-2.06	0	3.38	0	1.033879	55.816196	220	3	1.1	0.9;
	1178	1	35.84	11.8	0	0.94	0	0.962008	-33.719546	110	21	1.1	0.9;
	1179	1	86.6	23.4	0	0.39	0	1.030155	-3.081468	220	5	1.1	0.9;
	1180	1	43.75	16.8	0	0.42	0	1.070081	-31.060886	110	3	1.1	0.9;
	1181	1	0.1	-19.3	0	0.25	0	1.02568	-22.276087	220	3	1.1	0.9;
	1182	1	10.81	3.9	0	1.39	0	1.04827	-25.28647	110	15	1.1	0.9;
	1183	1	199.4	-19.9	0	6.68	0	1.061746	-33.326648	220	5	1.1	0.9;
	1184	2	0	0	0	0	0	1.037371	-8.714789	380	3	1.1	0.9;
	1185	1	150.4	15.3	0	8.09	0	1.00989	-44.51992	220	4	1.1	0.9;
	1186	1	62.7	9	0	3.18	0	1.002628	1.72414	154	1	1.1	0.9;
	1187	1	67.57	21.2	0	5.71	0	1.036168	-5.245571	220	5	1.1	0.9;
	1188	1	30.93	6.6	0	9.09	0	1.057273	-14.570559	220	5	1.1	0.9;
	1189	1	96.41	18.1	0	2.04	0	1.030652	38.417935	220	3	1.1	0.9;
	1190	1	14.62	5.6	0	0.25	0	0.997527	-20.480441	110	18	1.1	0.9;
	1191	1	37.3	5.4	0	0.12	0	1.002819	-10.814831	154	1	1.1	0.9;
	1192	1	13.21	9.2	0	0.69	0	0.991994	-25.39883	110	18	1.1	0.9;
	1193	1	181.5	21.2	0	64.45	0	1.045756	-16.022	380	6	1.1	0.9;
	1194	1	129.6	13.3	0	13.98	0	1.035804	-15.379297	380	5	1.1	0.9;
	1195	1	3.8	2.5	0	0.74	0	0.999088	-19.554899	110	18	1.1	0.9;
	1196	1	-0	-0	0.181549	8.450889	0	1.039052	-27.709316	220	18	1.1	0.9;
	1197	1	25.23	10.6	0	1.44	0	0.983134	-28.838212	110	18	1.1	0.9;
	1198	1	0	0	0	30.637696	0	0.994534	-14.439072	380	5	1.1	0.9;
	1199	1	-0	-0	0.052811	29.652975	0	1.056094	-20.556173	380	19	1.1	0.9;
	1200	1	124.8	7.9	0	48.99	0	1.024898	-41.828092	380	4	1.1	0.9;
	1201	1	114.1	23.5	0	1.08	0	1.039435	-20.087821	220	5	1.1	0.9;
	1202	1	-0	0	0	46.28	0	1.130115	-23.379521	380	9	1.3	0.7;
	1203	1	41.8	5.9	0	1.75	0	1.022216	-7.938516	154	1	1.1	0.9;
	1204	1	0	0	0	0.72	0	1.055433	-27.399177	150	9	1.1	0.9;
	1205	1	-0.19	-1	0	0	0	1.02364	2.132033	150	8	1.1	0.9;
	1206	1	27.13	16.8	0	0.3	0	0.95719	-29.999812	150	9	1.1	0.9;
	1207	1	70.18	28.7	0	0.8	0	1.018095	-0.737711	150	8	1.1	0.9;
	1208	1	13.9	1.9	0	0.38	0	1.023218	-7.257963	154	1	1.1	0.9;
	1209	1	183.03	15.7	0	1.83	0	1.009976	-49.306917	220	6	1.1	0.9;
	1210	1	-15.82	5.2	0	1.76	0	1.009864	-26.733179	150	9	1.1	0.9;
	1211	1	-0.59	-2.55	0.04975	111.131798	0	1.08863	-12.305489	380	18	1.3	0.7;
	1212	1	-0	-0	0	35.18	0	1.157605	-23.088872	380	1	1.3	0.7;
	1213	1	29.8	4.3	0	2.04	0	1.020277	-6.570021	154	1	1.1	0.9;
	1214	1	105	412	0	37.57	0	1.019583	18.757001	380	3	1.1	0.9;
	1215	1	0	0	0	0.34	0	1.030983	-22.83689	150	9	1.1	0.9;
	1216	1	0	0	0	4.93	0	1.036556	-11.314369	220	5	1.1	0.9;
	1217	1	0	0	0	0	0	1.025557	-24.888109	150	9	1.1	0.9;
	1218	1	30.83	13.2	0	1.12	0	0.983512	-40.888328	110	22	1.1	0.9;
	1219	1	1.4	0.5	0	1.87	0	1.003651	-37.515353	110	12	1.1	0.9;
	1220	2	0	0	0	0	0	1.008655	-45.569202	220	6	1.1	0.9;
	1221	1	230.5	91	0	57.67	0	1.026404	-1.946494	150	8	1.1	0.9;
	1222	1	23.03	14.3	0	16.88	0	1.023094	-32.476655	150	9	1.1	0.9;
	1223	1	26.4	3.8	0	0.7	0	1.023062	-9.030036	154	1	1.1	0.9;
	1224	1	168.3	-57.1	0	16.319522	0	1.025025	-35.945148	380	6	1.1	0.9;
	1225	2	0	0	0	0	0	1.051668	-17.803045	220	7	1.1	0.9;
	1226	1	-0	0	0	0.21	0	1.089904	-27.716794	150	9	1.1	0.9;
	1227	1	-0.18	-1	0	0	0	1.021578	10.573494	150	8	1.1	0.9;
	1228	1	53.16	17.7	0	0.48	0	1.013857	-32.136875	110	21	1.1	0.9;
	1229	1	5.51	2	0	0.66	0	1.018322	-28.501754	110	19	1.1	0.9;
	1230	1	0	-0	0	2.25	0	1.035275	-25.036832	154	1	1.1	0.9;
	1231	1	0	0	0	0	0	0.997535	-29.534575	150	9	1.1	0.9;
	1232	1	0	-0	0	81.52	0	1.045457	-50.275568	380	4	1.1	0.9;
	1233	1	0	-0	0	1.72	0	1.033936	-22.873061	220	5	1.1	0.9;
	1234	1	71.78	17.9	0	1.96	0	1.019897	-16.703598	220	5	1.1	0.9;
	1235	1	4.3	0.6	0	4.43	0	1.028466	-5.275021	154	1	1.1	0.9;
	1236	1	105.5	24.7	0	58.37	0	1.013549	-36.608631	220	6	1.1	0.9;
	1237	2	0	0	0	0	0	1.103909	-6.473548	380	5	1.3	0.7;
	1238	1	60.37	35.4	0	14.88	0	1.02547	56.988426	220	3	1.1	0.9;
	1239	1	0	-0	0	0	0	0.995239	-55.63816	220	10	1.1	0.9;
	1240	1	23.53	4	0	9.39	0	1.019502	-41.971456	220	4	1.1	0.9;
	1241	2	0	0	0	0	0	1.019582	-38.531665	220	4	1.1	0.9;
	1242	1	226.3	51.7	0	3.86	0	1.0386	-3.405307	220	3	1.1	0.9;
	1243	2	0	0	0	0	0	1.080005	57.148478	380	3	1.1	0.9;
	1244	2	0	0	0	0	0	1.026736	-16.807919	154	1	1.1	0.9;
	1245	1	-4.57	-1.31	0	1.12	0	1.018361	-26.287159	110	18	1.1	0.9;
	1246	1	2	1	0	0.61	0	1.020056	-29.074246	110	21	1.1	0.9;
	1247	1	23.03	8.3	0	1.05	0	1.031845	-25.386197	110	19	1.1	0.9;
	1248	1	0	0	0	1.17	0	1.045602	-10.875135	220	5	1.1	0.9;
	1249	1	141.26	30.6	0	18.01	0	1.020424	-11.314118	380	5	1.1	0.9;
	1250	1	39.44	24.9	0	4.18	0	0.99043	-35.622409	150	10	1.1	0.9;
	1251	2	0	0	0	0	0	1.024626	-8.729171	380	5	1.1	0.9;
	1252	1	0	0	0	6.56	0	1.031473	-33.940798	110	12	1.1	0.9;
	1253	1	186.6	155.8	0	28.99	0	1.011465	22.066796	380	3	1.1	0.9;
	1254	1	-0	-0	0	0.4	0	0.989117	-29.091224	110	18	1.1	0.9;
	1255	1	20.62	8.2	0	6.62	0	1.022343	3.87531	150	8	1.1	0.9;
	1256	1	1.9	1.3	0	0.92	0	1.011916	-19.513814	110	18	1.1	0.9;
	1257	1	-0	0	0	13.02	0	1.031007	-38.451986	220	6	1.1	0.9;
	1258	2	0	0	0	0	0	1.032962	38.774006	220	3	1.1	0.9;
	1259	1	-0	-0	0	39.46	0	1.127902	-23.797785	380	9	1.3	0.7;
	1260	1	50.86	36.6	0	0	0	1.008042	15.734617	150	8	1.1	0.9;
	1261	1	279.5	25.8	0	20.3	0	1.020831	-42.985197	220	6	1.1	0.9;
	1262	1	-0	0	0	10.73	0	1.034553	-22.377589	220	5	1.1	0.9;
	1263	1	-0	-0	0.256113	11.74089	0	1.002702	-10.318584	220	19	1.1	0.9;
	1264	1	224.9	-32.5	0	63.28	0	1.029039	-32.320569	380	6	1.1	0.9;
	1265	1	173	-21.3	0	5.93	0	1.063	-50.383849	220	5	1.1	0.9;
	1266	1	15.42	2.2	0	1.46	0	0.998744	-41.577791	110	22	1.1	0.9;
	1267	2	0	0	0	0	0	1.04637	20.205348	220	13	1.1	0.9;
	1268	1	151.2	96.4	0.279148	38.889483	0	1.005562	-3.438926	380	3	1.1	0.9;
	1269	1	112.8	27.6	0	35.18	0	1.012761	-36.678237	220	6	1.1	0.9;
	1270	1	1.8	0.7	0	0.17	0	1.032879	-24.788382	110	19	1.1	0.9;
	1271	2	0	0	0	0	0	1.0139	-46.721298	220	4	1.1	0.9;
	1272	1	0	0	0	0.23	0	1.016359	-30.703637	150	9	1.1	0.9;
	1273	1	7.01	2.3	0	0.69	0	1.014097	-35.942112	110	12	1.1	0.9;
	1274	1	149.9	53.2	0	1.07	0	0.966937	-22.699804	110	18	1.1	0.9;
	1275	1	52.46	2.9	0	0.35	0	1.03039	-19.090541	220	5	1.1	0.9;
	1276	1	-0	0	0	0.48	0	0.986046	-28.634033	150	9	1.1	0.9;
	1277	1	-0.2	-1	0	0	0	1.028701	-1.858875	150	8	1.1	0.9;
	1278	1	0	0	0	28.5	0	1.043988	-16.189968	380	7	1.1	0.9;
	1279	1	-0	-0	0	1.56	0	0.997512	-49.997395	220	6	1.1	0.9;
	1280	1	0	-0	0	107.39	0	1.07141	-37.162106	380	4	1.1	0.9;
	1281	1	0	0	0	0	0	0.969655	-24.039484	150	9	1.1	0.9;
	1282	1	202	34.1	0	1.94	0	1.019897	-47.387129	220	4	1.1	0.9;
	1283	1	42.75	15.2	0	0.54	0	1.052742	-25.899364	110	15	1.1	0.9;
	1284	1	19.02	9.1	0	0.52	0	0.994045	-26.972979	110	18	1.1	0.9;
	1285	1	-0	-0.2	0	0.15	0	1.052292	-15.194946	110	18	1.1	0.9;
	1286	1	7.61	2.5	0	1.67	0	0.963744	-29.059591	110	18	1.1	0.9;
	1287	1	-0.19	-78.58	0	2.31	0	1.052037	15.399435	380	3	1.1	0.9;
	1288	1	9.61	5.9	0	1.11	0	1.009022	-33.106438	150	9	1.1	0.9;
	1289	1	18.82	6.2	0	0.65	0	0.959518	-35.091886	110	21	1.1	0.9;
	1290	1	22.32	7.3	0	0.72	0	1.00796	-36.438787	110	12	1.1	0.9;
	1291	1	124.1	17.6	0	6.03	0	1.021764	-12.699538	154	1	1.1	0.9;
	1292	1	23.4	3.3	0	6.88	0	1.020624	-5.696232	154	1	1.1	0.9;
	1293	1	50.26	19.3	0	2.1	0	1.017813	7.842015	220	3	1.1	0.9;
	1294	1	6.11	-0	0	2.51	0	1.007318	-48.846534	220	4	1.1	0.9;
	1295	2	0	0	0	0	0	1.035594	-8.717281	220	5	1.1	0.9;
	1296	1	0	-0	0	0.3	0	1.015257	-30.22365	150	9	1.1	0.9;
	1297	1	19.02	8	0	2.05	0	1.007222	-26.340503	110	18	1.1	0.9;
	1298	1	-0	0	0	1.49	0	0.981434	-23.900611	110	18	1.1	0.9;
	1299	1	30.53	10	0	0.64	0	0.977286	-33.204156	110	21	1.1	0.9;
	1300	1	42.95	26.6	0	2.95	0	0.91444	-35.99744	150	9	1.1	0.9;
	1301	1	479	99.2	0	10.68	0	1.048311	-29.017601	220	5	1.1	0.9;
	1302	1	-19.83	-6.01	0	0.38	0	1.012848	-24.260593	110	18	1.1	0.9;
	1303	1	-0	0	0	3.47	0	1.02404	-14.144099	220	6	1.1	0.9;
	1304	1	5.41	1.8	0	0.72	0	1.01767	-35.316915	110	12	1.1	0.9;
	1305	1	-0	-0	0	4.34	0	1.038105	-22.77658	220	5	1.1	0.9;
	1306	1	8.51	6.4	0	0.39	0	0.969711	-29.75829	110	18	1.1	0.9;
	1307	2	0	0	0	0	0	1.062062	-10.130363	380	14	1.1	0.9;
	1308	1	71.38	10.3	0	4.56	0	1.000909	-33.917468	220	6	1.1	0.9;
	1309	1	17.72	3.8	0	0.94	0	0.960873	-27.498074	110	18	1.1	0.9;
	1310	1	-0	-0	0	0.2	0	0.934082	-34.726012	150	9	1.1	0.9;
	1311	1	37.94	15.2	0	35.31	0	1.03134	-21.905259	380	5	1.1	0.9;
	1312	1	18.2	2.6	0	1.51	0	1.030628	-11.136572	154	1	1.1	0.9;
	1313	2	0	0	0	0	0	1.031257	-39.135224	110	22	1.1	0.9;
	1314	1	-0	-0	0	2.79	0	1.043789	-20.817719	220	19	1.1	0.9;
	1315	2	0	0	0	0	0	1.091975	11.599735	380	13	1.3	0.7;
	1316	1	13.82	3.6	0	0.72	0	0.976277	-29.618371	110	18	1.1	0.9;
	1317	1	149.3	-2	0	20.77	0	1.010954	2.893831	380	2	1.1	0.9;
	1318	1	0.2	0.8	0	0.05	0	0.986341	14.700503	220	3	1.1	0.9;
	1319	1	-0	-0	0	0.08	0	1.047415	-29.296022	150	9	1.1	0.9;
	1320	2	0	0	0	0	0	1.01339	-48.941694	220	6	1.1	0.9;
	1321	1	11.41	5.1	0	1.04	0	1.0176	-20.137212	110	18	1.1	0.9;
	1322	1	0	0	0	0	0	0.977041	-35.24252	150	9	1.1	0.9;
	1323	1	118.9	129.2	0	10.49	0	1.018452	2.80416	380	3	1.1	0.9;
	1324	1	37.14	11.6	0	1.8	0	1.029565	-25.266254	110	19	1.1	0.9;
	1325	1	37.69	-19.49	0	0.11	0	1.031959	-26.292195	110	19	1.1	0.9;
	1326	1	0	0	0	0.98	0	1.065109	-35.279844	220	5	1.1	0.9;
	1327	1	135.7	31.6	0	26.65	0	1.015199	-36.525423	220	6	1.1	0.9;
	1328	1	0	-0	0	50.16	0	1.031228	-38.438089	380	6	1.1	0.9;
	1329	1	105.8	15.1	0	4.91	0	1.019261	2.862212	154	1	1.1	0.9;
	1330	1	100.6	15.3	0	1.28	0	0.998944	-2.086057	150	8	1.1	0.9;
	1331	1	-0.17	-1.82	0	0	0	1.048144	18.207633	380	8	1.1	0.9;
	1332	1	-0	0	0	2.04	0	1.050721	-24.766219	150	9	1.1	0.9;
	1333	1	-9.61	3.2	0	0.75	0	1.00264	-33.181209	150	9	1.1	0.9;
	1334	1	0	0	0	0.59	0	1.022554	-16.489068	220	5	1.1	0.9;
	1335	1	-0	0	0	0	0	1.085035	-23.705838	150	23	1.1	0.9;
	1336	1	13.9	1.9	0	1.02	0	1.022258	-6.988685	154	1	1.1	0.9;
	1337	2	0	0	0	0	0	1.05919	-17.318411	380	16	1.1	0.9;
	1338	1	29.03	9.4	0	5.11	0	1.056547	10.591015	110	8	1.1	0.9;
	1339	1	108.4	-2.4	0	80.12	0	1.050471	-46.710084	380	10	1.1	0.9;
	1340	1	13.21	6.6	0	1.19	0	0.991895	-42.420238	110	22	1.1	0.9;
	1341	2	0	0	0	0	0	1.035237	-14.263308	220	5	1.1	0.9;
	1342	1	14.31	0.87	0	3.79	0	1.039269	-0.842292	220	13	1.1	0.9;
	1343	1	-0	-0	0	58.3	0	1.001607	-14.057721	380	5	1.1	0.9;
	1344	1	-0	0	0	2.13	0	1.021828	-46.679878	220	4	1.1	0.9;
	1345	1	1.91	3.95	0	0.02	0	1.015627	61.512423	150	1	1.1	0.9;
	1346	1	-0	0	0	1.87	0	1.025101	-22.268256	220	3	1.1	0.9;
	1347	1	0	-0	0	12.2	0	1.021862	-53.734293	220	4	1.1	0.9;
	1348	1	34.9	5	0	6.35	0	1.01292	-5.059564	154	1	1.1	0.9;
	1349	1	-0	-0	0	0.12	0	1.024668	5.632836	154	1	1.1	0.9;
	1350	1	30.53	13.8	0	2.19	0	1.026189	-1.244991	150	8	1.1	0.9;
	1351	1	0	0	0	0.46	0	0.964633	-17.068916	110	18	1.1	0.9;
	1352	1	28.93	11.9	0	3.84	0	1.031938	-1.392462	150	8	1.1	0.9;
	1353	1	6.41	2.3	0	0.85	0	1.022221	-22.0039	110	19	1.1	0.9;
	1354	2	0	0	0	0	0	1.034497	-7.597033	220	5	1.1	0.9;
	1355	1	-3.29	-1.32	0	190.704588	0	1.029327	-8.7153	380	5	1.1	0.9;
	1356	1	538.3	90.8	0	21.34	0	1.018699	-41.899667	380	6	1.1	0.9;
	1357	1	2.1	0.3	0	2.83	0	1.016995	3.722869	154	1	1.1	0.9;
	1358	1	100.3	4.8	0	0.15	0	0.9993	-35.047873	110	21	1.1	0.9;
	1359	1	60.17	30.6	0	0.12	0	1.020143	-5.293872	220	3	1.1	0.9;
	1360	1	-0	0	0.27075	68.09435	0	1.076714	-23.212938	380	18	1.1	0.9;
	1361	2	0	0	0	0	0	1.001499	-25.274364	110	18	1.1	0.9;
	1362	1	-0	-0	0	6.93	0	1.03998	25.637281	220	8	1.1	0.9;
	1363	1	0	0	0.316851	27.808806	0	1.008673	-20.061722	380	11	1.1	0.9;
	1364	1	40.85	14.5	0	0.59	0	1.038263	-12.28852	220	5	1.1	0.9;
	1365	2	0	0	0	0	0	1.045092	-39.45932	380	6	1.1	0.9;
	1366	1	22.02	7.7	0	0.78	0	1.028168	-40.869549	110	22	1.1	0.9;
	1367	1	0	-0	0.175687	-1.509401	0	1.005561	-0.569839	380	2	1.1	0.9;
	1368	1	0	0	0	0	0	1.00266	-30.532747	150	9	1.1	0.9;
	1369	1	0	0	0.258764	32.105228	0	1.033686	-25.315586	380	7	1.1	0.9;
	1370	1	46.55	28.8	0	1.08	0	1.020241	-29.747583	150	9	1.1	0.9;
	1371	2	0	0	0	0	0	1.02639	0.926583	154	1	1.1	0.9;
	1372	1	50.96	16.8	0	1.03	0	0.95916	-33.882239	110	21	1.1	0.9;
	1373	1	-0	0	0	0.29	0	0.993927	-32.619243	150	9	1.1	0.9;
	1374	1	0	-0	0	0.43	0	0.997482	-29.519121	150	9	1.1	0.9;
	1375	1	10.11	8.9	0	0.12	0	0.955338	-31.250238	110	18	1.1	0.9;
	1376	1	0	0	0	0	0	0.993711	-33.877317	150	9	1.1	0.9;
	1377	1	18.02	6	0	0.41	0	0.99069	-31.192636	110	21	1.1	0.9;
	1378	2	0	0	0	0	0	1.015715	1.870619	154	1	1.1	0.9;
	1379	1	2.91	-1.45	0	5.1	0	1.025837	-29.429882	150	9	1.1	0.9;
	1380	1	145.2	21.4	0	1.09	0	1.022758	-23.095174	220	5	1.1	0.9;
	1381	1	93.9	13.4	0	5.87	0	1.001682	1.388839	154	1	1.1	0.9;
	1382	1	10.81	3.5	0	0.98	0	1.010507	-35.86105	110	12	1.1	0.9;
	1383	1	0	0	0	3.97	0	1.045942	-20.988282	220	1	1.1	0.9;
	1384	1	0	-0	0	42.32	0	1.109629	-21.96249	380	1	1.3	0.7;
	1385	1	14.75	-0.07	0	3.07	0	1.019767	61.573327	150	1	1.1	0.9;
	1386	1	0	0	0	0	0	0.990763	-30.757298	150	9	1.1	0.9;
	1387	2	0	0	0	0	0	1.029614	-0.841353	220	14	1.1	0.9;
	1388	1	20.32	6.7	0	1.11	0	1.019538	-34.270193	110	12	1.1	0.9;
	1389	1	0	0	0	0	0	1.0032	-36.682256	150	9	1.1	0.9;
	1390	2	0	0	0	0	0	1.076512	-19.376155	220	21	1.1	0.9;
	1391	2	0	0	0	0	0	1.027192	-37.784565	220	4	1.1	0.9;
	1392	1	-0	-0	0	4.2	0	1.025265	17.7628	220	8	1.1	0.9;
	1393	1	2.4	0.7	0	10.69	0	1.021752	-46.013798	220	4	1.1	0.9;
	1394	1	30.93	7.6	0	4.69	0	1.068933	-26.711668	220	5	1.1	0.9;
	1395	1	52.96	15.2	0	0.56	0	1.053899	-32.549685	110	3	1.1	0.9;
	1396	1	-0	-6	0	0	0	1.027929	-30.435624	150	9	1.1	0.9;
	1397	1	0	-0	0	2.14	0	1.044833	-13.244455	220	5	1.1	0.9;
	1398	1	196.7	44.2	0	0.1	0	1.014695	-19.132196	220	5	1.1	0.9;
	1399	2	0	0	0	0	0	1.04788	-49.397615	380	4	1.1	0.9;
	1400	1	39.25	71.7	0	33.75	0	1.021319	2.297921	380	3	1.1	0.9;
	1401	1	11.91	2.9	0	5.27	0	1.036953	-14.926159	220	5	1.1	0.9;
	1402	2	0	0	0	0	0	1.019959	-20.19317	220	20	1.1	0.9;
	1403	1	-22.07	-1.97	0	13.85	0	1.062912	54.514608	220	3	1.1	0.9;
	1404	1	69.3	9.9	0	0.03	0	1.022946	4.97524	154	1	1.1	0.9;
	1405	1	-0	0	0	1.35	0	0.989062	-25.178694	110	18	1.1	0.9;
	1406	1	157.9	26	0	3.74	0	1.018175	-24.811382	220	7	1.1	0.9;
	1407	1	0	-0	0	4.15	0	0.991705	-49.889604	150	10	1.1	0.9;
	1408	1	15.52	5.2	0	0.29	0	1.00208	-32.664202	110	21	1.1	0.9;
	1409	2	0	0	0	0	0	1.022303	-48.78594	220	4	1.1	0.9;
	1410	1	-0	0	0	0.64	0	0.973115	-32.084772	110	21	1.1	0.9;
	1411	1	0	-0	0	7.52	0	1.007407	-48.005426	220	4	1.1	0.9;
	1412	1	5.11	0.9	0	0.37	0	1.005546	-26.152557	110	18	1.1	0.9;
	1413	1	0	0	0	0	0	0.992915	-27.541743	150	9	1.1	0.9;
	1414	1	144.7	20.3	0	6.35	0	1.01828	-28.289267	220	5	1.1	0.9;
	1415	1	0	-0	0	0.3	0	1.029195	-11.729083	220	5	1.1	0.9;
	1416	1	7.81	2.6	0	0.3	0	1.020594	-23.616181	110	19	1.1	0.9;
	1417	2	0	0	0	0	0	1.009035	-13.438457	110	18	1.1	0.9;
	1418	1	40.55	3.6	0	1.88	0	1.015411	-55.166319	220	4	1.1	0.9;
	1419	1	16.52	3.3	0	0.37	0	1.010045	-44.226167	110	22	1.1	0.9;
	1420	2	0	0	0	0	0	1.001444	-17.527123	400	1	1.1	0.9;
	1421	1	42.95	19.8	0	3.85	0	1.016209	-52.437506	220	4	1.1	0.9;
	1422	2	0	0	0	0	0	1.047835	-21.372401	220	5	1.1	0.9;
	1423	1	-0	-0	0	8.53	0	1.01922	-35.759714	220	6	1.1	0.9;
	1424	1	-4.89	-2.35	0	0.13	0	1.013545	-23.28693	110	18	1.1	0.9;
	1425	1	0.3	0.2	0	4.7	0	1.042357	-8.099866	380	3	1.1	0.9;
	1426	2	0	0	0	0	0	1.072946	26.386805	380	3	1.1	0.9;
	1427	1	0	0	0	1.85	0	0.989682	-31.598298	110	21	1.1	0.9;
	1428	1	22.83	7.5	0	0.36	0	1.006042	-36.807516	110	12	1.1	0.9;
	1429	1	-0	0	0	121.27	0	1.062953	-35.924835	380	22	1.1	0.9;
	1430	1	20.82	7.9	0	5.62	0	1.009939	-51.103497	220	4	1.1	0.9;
	1431	1	-8.31	2.7	0	0.18	0	0.973207	-32.072306	150	9	1.1	0.9;
	1432	1	193.2	121.7	0	11.94	0	1.016823	-9.538924	220	3	1.1	0.9;
	1433	2	0	0	0	0	0	1.01442	-7.183674	154	1	1.1	0.9;
	1434	1	-0	0	0	4.88	0	1.06933	-22.158284	220	1	1.1	0.9;
	1435	1	-0	0	0	3.85	0	1.039832	-24.69425	220	5	1.1	0.9;
	1436	2	0	0	0	0	0	1.054052	-13.334937	220	5	1.1	0.9;
	1437	1	204.6	53.2	0	7.4	0	0.987496	-38.559416	220	4	1.1	0.9;
	1438	1	0	-0	0.31	25.37	0	1.035796	-21.856758	220	19	1.1	0.9;
	1439	1	110.1	2	0	13.85	0	1.022676	-48.766148	220	4	1.1	0.9;
	1440	2	0	0	0	0	0	1.011268	-4.465895	154	1	1.1	0.9;
	1441	1	60.07	20	0	0.17	0	0.986338	-31.835513	110	21	1.1	0.9;
	1442	1	89.3	12.8	0	4.33	0	1.01826	-6.508131	154	1	1.1	0.9;
	1443	1	171.96	40.73	0	23.47	0	1.032827	-7.541273	380	13	1.1	0.9;
	1444	2	0	0	0	0	0	1.023146	-8.694857	154	1	1.1	0.9;
	1445	1	238.4	210.9	0	0	0	1.000392	-0.646604	220	2	1.1	0.9;
	1446	1	10.91	3	0	0.88	0	0.929681	-34.552406	110	18	1.1	0.9;
	1447	1	4.61	1.7	0	0.47	0	1.033304	-20.116521	110	19	1.1	0.9;
	1448	1	0	0	0	9.49	0	1.069974	-15.594879	220	5	1.1	0.9;
	1449	1	-0	0	0	6.65	0	1.023427	-32.472976	150	9	1.1	0.9;
	1450	1	39.54	14.3	0	0.23	0	1.047849	-23.198799	110	19	1.1	0.9;
	1451	1	20.52	6.7	0	0.13	0	1.032216	-34.275921	110	12	1.1	0.9;
	1452	1	-0	-0	0	2.3	0	1.021465	3.898083	380	3	1.1	0.9;
	1453	1	80.09	26.3	0	0.37	0	1.01588	-36.12599	110	12	1.1	0.9;
	1454	1	20.62	10.7	0	0.14	0	1.050287	-24.708967	150	9	1.1	0.9;
	1455	1	18.02	4.8	0	0.43	0	0.964588	-31.478302	110	18	1.1	0.9;
	1456	1	0	0	0	0.37	0	0.998629	-37.756444	110	1	1.1	0.9;
	1457	1	20.82	36	0	0.97	0	1.013965	-39.959399	220	4	1.1	0.9;
	1458	2	0	0	0	0	0	1.045766	-38.603093	380	4	1.1	0.9;
	1459	1	-0	-0	0	3.38	0	1.039481	-22.115895	220	5	1.1	0.9;
	1460	1	67.4	9.6	0	4.68	0	1.010427	-7.350669	154	1	1.1	0.9;
	1461	1	0	0	0	1.44	0	1.035843	-39.417888	220	4	1.1	0.9;
	1462	1	-0	0	0	0.42	0	1.047856	-21.372911	220	5	1.1	0.9;
	1463	1	109.6	36.4	0	13.24	0	1.035408	5.827943	150	8	1.1	0.9;
	1464	1	22.73	6.4	0	2.1	0	1.019811	7.461504	150	8	1.1	0.9;
	1465	1	-11.03	-5.11	0	29.706803	0	0.984287	-15.629121	380	5	1.1	0.9;
	1466	1	42.95	26.6	0	3.11	0	1.084595	-28.010738	150	9	1.1	0.9;
	1467	1	0	0	0	0	0	1.013679	-32.975234	150	9	1.1	0.9;
	1468	1	0	-0	0	17.6	0	1.00405	-18.486305	220	6	1.1	0.9;
	1469	1	0	0	0	6.4	0	1.048685	-33.084852	220	6	1.1	0.9;
	1470	1	18.12	-11.6	0	14.14	0	1.000122	-49.871341	220	6	1.1	0.9;
	1471	1	95.61	8.6	0	3.48	0	0.97509	-49.690385	220	4	1.1	0.9;
	1472	1	0	-0	0	0.98	0	0.997982	-25.220001	110	18	1.1	0.9;
	1473	1	0	0	0.15	6.49	0	1.001303	-29.93877	220	21	1.1	0.9;
	1474	2	0	0	0	0	0	1.010259	-6.19747	154	1	1.1	0.9;
	1475	2	0	0	0	0	0	1.020467	-43.69136	380	4	1.1	0.9;
	1476	1	30.63	19	0	2.33	0	0.858822	-39.292558	150	9	1.3	0.7;
	1477	1	0	-0	0	5.51	0	1.034469	9.039884	150	8	1.1	0.9;
	1478	2	0	0	0	0	0	1.048923	-15.833833	220	5	1.1	0.9;
	1479	1	82.19	23.4	0	17.91	0	1.02528	-43.676962	380	4	1.1	0.9;
	1480	1	100.1	45.6	0	0.23	0	1.007945	-46.699134	220	6	1.1	0.9;
	1481	2	0	0	0	0	0	1.040265	1.809268	220	13	1.1	0.9;
	1482	1	0	0	0	0	0	1.02447	-26.749332	150	9	1.1	0.9;
	1483	1	0	0	0	4.86	0	1.067484	-27.680579	220	5	1.1	0.9;
	1484	2	0	0	0	0	0	1.027385	50.204439	220	3	1.1	0.9;
	1485	1	144.9	87.7	0.214796	36.223148	0	1.010367	-15.445846	380	3	1.1	0.9;
	1486	1	79.59	18.8	0	6.78	0	1.030936	-23.69577	220	5	1.1	0.9;
	1487	2	0	0	0	0	0	1.055167	-12.534271	220	7	1.1	0.9;
	1488	1	-0	-0	0	1.66	0	1.033835	-39.750272	380	6	1.1	0.9;
	1489	1	0	-0	0	2.54	0	1.024224	-36.529588	220	10	1.1	0.9;
	1490	1	23.03	14.2	0	1.24	0	1.029241	-22.185994	150	9	1.1	0.9;
	1491	1	117.5	-3	0	5.38	0	1.020793	-47.723054	220	4	1.1	0.9;
	1492	2	0	0	0	0	0	1.02928	-5.798772	220	13	1.1	0.9;
	1493	1	1.1	2.5	0.229158	23.95807	0	1.010911	-5.35042	380	3	1.1	0.9;
	1494	1	-1.85	-1.63	0	31.424588	0	1.039428	-11.596451	380	5	1.1	0.9;
	1495	1	18.02	6	0	0.98	0	1.023957	-28.318674	110	21	1.1	0.9;
	1496	1	0.7	0.2	0	0.22	0	1.030854	-34.139697	110	12	1.1	0.9;
	1497	1	0	0	0	11.44	0	0.990307	29.776271	220	3	1.1	0.9;
	1498	2	0	0	0	0	0	1.073955	-40.297354	380	4	1.1	0.9;
	1499	1	6.41	2.3	0	1.51	0	1.043221	-20.703725	110	19	1.1	0.9;
	1500	1	13.82	5	0	0.47	0	1.044168	-19.427331	110	19	1.1	0.9;
	1501	1	-0	-0	0	12.57	0	1.024669	-54.354073	220	4	1.1	0.9;
	1502	1	-0	0	0	37.116157	0	1.024544	-26.951039	380	5	1.1	0.9;
	1503	1	142.8	-6.4	0	12.36	0	1.019373	-44.17804	220	4	1.1	0.9;
	1504	1	0	-0	0	1.84	0	1.019788	-14.640739	220	5	1.1	0.9;
	1505	1	6.71	3.3	0	1.31	0	1.0029	-27.795475	110	18	1.1	0.9;
	1506	1	0	0	0	149.14	0	1.079345	-19.716256	380	19	1.1	0.9;
	1507	2	0	0	0	0	0	1.057885	-2.186886	220	13	1.1	0.9;
	1508	2	0	0	0	0	0	1.062401	-20.441537	220	3	1.1	0.9;
	1509	1	16.42	5.9	0	0.25	0	1.029556	-25.986341	110	19	1.1	0.9;
	1510	1	-0	0	0	1.12	0	1.023208	-26.047466	150	9	1.1	0.9;
	1511	1	-0	-0	0	0	0	0.990901	-26.21194	110	18	1.1	0.9;
	1512	1	339.1	89	0	80.71	0	1.028466	-44.106028	380	6	1.1	0.9;
	1513	1	-35.74	38.6	0	4.65	0	1.016487	-40.160602	220	4	1.1	0.9;
	1514	1	-0	-0	0	1.18	0	1.004779	-23.508252	150	9	1.1	0.9;
	1515	1	0	-0	0	28.31	0	1.008712	-1.107611	400	1	1.1	0.9;
	1516	1	22.02	9	0	0.24	0	1.037153	-0.734936	150	8	1.1	0.9;
	1517	1	7.41	2.7	0	0.98	0	1.027855	-25.555887	110	19	1.1	0.9;
	1518	1	0	-0	0	0.1	0	0.997056	-32.275189	150	9	1.1	0.9;
	1519	1	-10.51	3.5	0	0.8	0	1.007092	-26.93445	150	9	1.1	0.9;
	1520	1	173.6	72.9	0	2.5	0	1.002285	-46.811043	220	4	1.1	0.9;
	1521	1	0	-0	0	0	0	0.988659	-30.981824	150	9	1.1	0.9;
	1522	1	72.1	71.8	0	22	0	1.017448	21.68548	380	3	1.1	0.9;
	1523	1	57.86	23.2	0	2.08	0	0.969123	-54.141921	220	6	1.1	0.9;
	1524	1	-0	-0	0	7.22	0	1.014749	17.834529	380	1	1.1	0.9;
	1525	1	0	-0	0	17.24	0	1.055481	17.939827	380	3	1.1	0.9;
	1526	1	-10.86	-5.93	0	12.54	0	1.037506	-12.250028	220	5	1.1	0.9;
	1527	1	39.64	11.9	0	2.86	0	1.011733	-41.237748	110	22	1.1	0.9;
	1528	1	5.81	2	0	0.2	0	1.005085	-30.239128	110	21	1.1	0.9;
	1529	2	0	0	0	0	0	1.01463	-7.159648	154	1	1.1	0.9;
	1530	1	0	8.4	0	15.21	0	1.021998	-28.21235	150	9	1.1	0.9;
	1531	1	-0	0	0	0	0	1.008004	-11.969038	150	2	1.1	0.9;
	1532	1	2.5	1.3	0	0.85	0	0.994007	-19.269916	110	18	1.1	0.9;
	1533	1	43.8	20.7	0	133.678646	0	1.077949	-24.835265	380	9	1.1	0.9;
	1534	1	-0	-0	0.261953	32.381259	0	1.096164	-25.47435	380	21	1.1	0.9;
	1535	1	20.32	8.9	0	1.29	0	0.968445	-29.191236	110	18	1.1	0.9;
	1536	1	0	0	0	0.63	0	1.002229	-18.476273	110	18	1.1	0.9;
	1537	1	0	0	0	0	0	0.973184	-34.115427	150	9	1.1	0.9;
	1538	1	36.34	15.2	0	0.3	0	1.031337	-8.832961	220	5	1.1	0.9;
	1539	1	73.58	3.24	0	52.35	0	1.058457	-15.584872	220	5	1.1	0.9;
	1540	1	0	-0	0.27075	74.70435	0	1.078257	-16.229346	380	18	1.1	0.9;
	1541	1	125.4	45.8	0	0	0	1.088981	-24.30323	220	5	1.1	0.9;
	1542	1	170.4	81.9	0	59.64	0	1.056958	-35.340331	380	17	1.1	0.9;
	1543	1	116.7	64	0	10.41	0	0.989265	-49.989138	220	10	1.1	0.9;
	1544	1	-0	0	0	44.56	0	1.136307	-25.930375	380	23	1.3	0.7;
	1545	1	42.55	8.2	0	2.06	0	1.037202	-23.828385	220	5	1.1	0.9;
	1546	2	0	0	0	0	0	1.042059	-40.600898	380	6	1.1	0.9;
	1547	1	-0	-0	0	1.02	0	1.04322	-15.091566	220	5	1.1	0.9;
	1548	1	7.61	2.5	0	0.72	0	0.984792	-17.582107	110	18	1.1	0.9;
	1549	1	300.2	-3.1	0	29.94	0	1.015193	-53.632592	380	4	1.1	0.9;
	1550	2	0	0	0	0	0	1.082656	-13.662784	380	18	1.1	0.9;
	1551	1	-0	-0	0	7.99	0	1.038512	-20.752173	220	5	1.1	0.9;
	1552	1	20.42	8.1	0	35.15	0	1.039612	-16.879714	380	5	1.1	0.9;
	1553	1	-21.62	-3.9	0	4.01	0	1.04217	-18.228454	220	7	1.1	0.9;
	1554	1	0	-0	0	4.96	0	1.01543	-16.393806	220	5	1.1	0.9;
	1555	2	0	0	0	0	0	1.065019	-8.997774	380	14	1.1	0.9;
	1556	1	-0	-0	0	0.95	0	1.065387	-11.818883	220	5	1.1	0.9;
	1557	2	0	0	0	0	0	1.039811	-17.614282	380	7	1.1	0.9;
	1558	2	0	0	0	0	0	1.01534	-6.08835	154	1	1.1	0.9;
	1559	1	-0	-0	0	2.62	0	1.016899	7.153568	154	1	1.1	0.9;
	1560	1	-0.5	-1	0	0.68	0	1.043962	-20.817307	220	19	1.1	0.9;
	1561	1	175.2	82.4	0	0.21	0	1.021846	-40.493768	220	6	1.1	0.9;
	1562	1	77.39	24.1	0	2.67	0	1.03782	-26.193252	220	5	1.1	0.9;
	1563	1	-0	-0	0.090774	1.980444	0	1.043985	-17.54442	220	18	1.1	0.9;
	1564	1	140.4	49.4	0	4.5	0	1.017601	7.024488	150	8	1.1	0.9;
	1565	1	16.7	2.4	0	3.57	0	1.030949	1.556549	154	1	1.1	0.9;
	1566	1	156.9	-4	0	2.48	0	1.03118	-25.125175	220	5	1.1	0.9;
	1567	1	60.07	27	0	0.48	0	1.037796	-25.98792	110	23	1.1	0.9;
	1568	1	69.08	21.5	0	1.65	0	1.04247	-25.487545	220	5	1.1	0.9;
	1569	1	0	0	0	0	0	1.020459	-32.098619	150	9	1.1	0.9;
	1570	2	0	0	0	0	0	1.076277	-35.646842	380	10	1.1	0.9;
	1571	1	329.4	12.9	0	34.36	0	1.048068	-30.588257	380	6	1.1	0.9;
	1572	1	0	0	0	0	0	1.034403	-30.610144	150	9	1.1	0.9;
	1573	1	12.6	1.8	0	1.96	0	1.024026	12.428625	154	1	1.1	0.9;
	1574	1	34.04	4.4	0	0.22	0	1.008241	-41.563979	110	22	1.1	0.9;
	1575	1	0	0	0	2.18	0	0.999328	-33.899863	220	22	1.1	0.9;
	1576	1	0	0	0	29.73	0	1.061908	11.541998	380	3	1.1	0.9;
	1577	1	32.04	19.8	0	0.72	0	1.053931	-22.761984	150	9	1.1	0.9;
	1578	1	0	-0	0	1.38	0	1.034689	-12.335494	220	5	1.1	0.9;
	1579	1	-0	0	0	16.56	0	1.006977	-5.494028	380	2	1.1	0.9;
	1580	1	-0	-0	0	0.19	0	1.01471	-22.608554	150	9	1.1	0.9;
	1581	1	146.6	3.3	0	18.83	0	1.019111	-14.760439	220	3	1.1	0.9;
	1582	2	0	0	0	0	0	1.042602	66.589497	380	3	1.1	0.9;
	1583	1	-4.14	-1.83	0	0.46	0	1.007978	-23.85957	110	18	1.1	0.9;
	1584	1	0	-0	0	3.24	0	1.035277	-14.661089	220	5	1.1	0.9;
	1585	1	30.53	10.4	0	0.56	0	1.040379	-8.600339	110	19	1.1	0.9;
	1586	1	-0	-0	0	21.01	0	1.025666	-32.943705	380	7	1.1	0.9;
	1587	1	78.5	137.8	0	3.01	0	1.013425	55.443139	220	3	1.1	0.9;
	1588	1	104.5	-56.4	0	23.22	0	1.040313	-8.413966	380	3	1.1	0.9;
	1589	2	0	0	0	0	0	1.023343	-7.228521	154	1	1.1	0.9;
	1590	1	25.73	9.3	0	0.12	0	1.067508	-19.594255	110	19	1.1	0.9;
	1591	1	-0.16	-11.88	0	16.12	0	1.016882	16.592812	380	8	1.1	0.9;
	1592	1	28.33	44.8	0	0.51	0	1.033424	-2.728211	220	5	1.1	0.9;
	1593	1	-0	-0	0	3.01	0	1.07124	59.977227	380	3	1.1	0.9;
	1594	1	0	0	0	288	0	1.005975	-10.923264	400	1	1.1	0.9;
	1595	1	71.77	13.57	0	3.84	0	1.011389	-29.689157	220	5	1.1	0.9;
	1596	1	76.28	45.7	0	10.89	0	1.048407	40.876809	380	3	1.1	0.9;
	1597	1	51.96	13.4	0	0.65	0	1.019934	13.471194	110	8	1.1	0.9;
	1598	1	58.87	36.4	0	3.38	0	0.945608	-29.541812	150	9	1.1	0.9;
	1599	1	0	0	0	0.48	0	1.008649	-48.901836	220	4	1.1	0.9;
	1600	1	38.74	14.1	0	12.78	0	1.025641	-54.321174	220	4	1.1	0.9;
	1601	1	63.5	-381.2	0.52638	207.010914	0	1.065707	25.831984	380	3	1.1	0.9;
	1602	1	45.75	4.2	0	4.9	0	1.009154	-60.6735	220	4	1.1	0.9;
	1603	1	137.2	16.1	0	5.26	0	1.018692	-25.570995	220	7	1.1	0.9;
	1604	2	0	0	0	0	0	1.037285	-6.852465	220	5	1.1	0.9;
	1605	1	-0	0	0	0.05	0	1.074227	-24.63652	220	5	1.1	0.9;
	1606	2	0	0	0	0	0	1.077423	8.098604	380	14	1.1	0.9;
	1607	1	73.78	-0	0	2.69	0	1.03265	-5.662474	380	5	1.1	0.9;
	1608	1	25.13	15.6	0	4.45	0	1.036655	-33.030803	150	9	1.1	0.9;
	1609	1	0	0	0	6.55	0	1.046745	-20.967096	220	5	1.1	0.9;
	1610	1	164.8	42.5	0	32.849681	0	1.031536	-44.292075	380	6	1.1	0.9;
	1611	1	274.1	166.5	0	0.56	0	1.044123	-9.607803	150	2	1.1	0.9;
	1612	1	18.5	2.6	0	3.05	0	1.035831	7.678301	154	1	1.1	0.9;
	1613	1	149.8	63.1	0	34.02	0	1.039635	-15.151261	220	6	1.1	0.9;
	1614	1	-0	0	0	0.04	0	1.040496	9.279654	220	3	1.1	0.9;
	1615	1	50.06	-0	0	0	0	0.963899	-38.948121	150	7	1.1	0.9;
	1616	1	0	0	0	14.45	0	0.998365	-4.338847	380	2	1.1	0.9;
	1617	1	14.72	5.3	0	0.47	0	1.040886	-3.895078	110	19	1.1	0.9;
	1618	2	0	0	0	0	0	1.075715	-44.205199	380	4	1.1	0.9;
	1619	1	20.32	2.5	0	5.2	0	0.927137	-30.621268	110	18	1.1	0.9;
	1620	1	0	0	0	0	0	1.028994	-31.533375	380	6	1.1	0.9;
	1621	1	4	0.6	0	3.36	0	1.022327	-2.896696	154	1	1.1	0.9;
	1622	1	-0.22	-0.4	0	6.67	0	1.003443	-8.907025	220	3	1.1	0.9;
	1623	1	-0	0	0	7.74	0	1.028598	38.481363	220	3	1.1	0.9;
	1624	1	164.2	-10.7	0	11.44	0	1.049136	-17.188943	220	3	1.1	0.9;
	1625	1	73.18	22.3	0	8.77	0	1.036923	-19.90324	220	5	1.1	0.9;
	1626	2	0	0	0	0	0	1.038045	9.540137	154	1	1.1	0.9;
	1627	1	123.5	-49.4	0	2.9	0	1.028749	-11.339861	220	3	1.1	0.9;
	1628	1	3.3	1.1	0	0.28	0	1.039685	-39.992757	110	22	1.1	0.9;
	1629	1	41.95	8.5	0	1.05	0	1.038022	-11.387442	220	5	1.1	0.9;
	1630	1	0	0	0	20.95	0	1.013313	32.823189	380	3	1.1	0.9;
	1631	1	-0	-0	0	0.87	0	0.986199	-31.336551	150	9	1.1	0.9;
	1632	1	20.82	7.9	0	1.07	0	1.01941	-52.813876	220	4	1.1	0.9;
	1633	1	-0	-0	0	45.02	0	1.014774	16.434066	150	8	1.1	0.9;
	1634	1	5.31	2.3	0	1.03	0	0.992605	-15.913829	110	18	1.1	0.9;
	1635	1	-0.18	-1.06	0	0	0	1.01531	10.365883	150	8	1.1	0.9;
	1636	1	-0	-0	0.129305	29.161483	0	1.030825	66.746693	380	1	1.1	0.9;
	1637	1	-8.41	2.7	0	0.83	0	1.036281	-26.892501	150	9	1.1	0.9;
	1638	1	70.08	11.1	0	9.84	0	1.023765	-54.46458	220	4	1.1	0.9;
	1639	1	51.66	8	0	4.51	0	1.026818	-53.620086	220	4	1.1	0.9;
	1640	1	33.6	4.8	0	5.24	0	1.031823	7.264228	154	1	1.1	0.9;
	1641	1	33.24	11.5	0	1.61	0	1.054985	-22.038982	110	19	1.1	0.9;
	1642	2	0	0	0	0	0	1.025423	0.137212	380	5	1.1	0.9;
	1643	1	95.71	25.9	0	11.38	0	1.037859	-26.226981	220	5	1.1	0.9;
	1644	1	-33.1	-14.35	0	94.55	0	1.057845	3.249165	380	5	1.1	0.9;
	1645	1	60.27	9.4	0	0.62	0	1.02376	-54.677814	220	4	1.1	0.9;
	1646	1	15.34	-1.3	0	0.71	0	1.016807	62.601043	150	1	1.1	0.9;
	1647	1	13.31	4.8	0	0.2	0	1.063498	-21.203305	110	19	1.1	0.9;
	1648	1	103.7	14.9	0	3.04	0	1.017786	-23.764488	220	20	1.1	0.9;
	1649	1	33.04	15.7	0	1.5	0	1.047211	-27.566054	110	23	1.1	0.9;
	1650	1	-0	0	0	12.5	0	1.079674	-22.543807	380	9	1.1	0.9;
	1651	1	-0	-0	0	0.73	0	1.033452	-2.878595	220	1	1.1	0.9;
	1652	1	22.02	11.9	0	0.79	0	0.996997	-31.783413	110	23	1.1	0.9;
	1653	1	0	-0	0.207485	26.23319	0	1.012766	-29.645206	380	7	1.1	0.9;
	1654	1	0	-0	0	45.19	0	1.124052	-27.861245	380	1	1.3	0.7;
	1655	1	15.22	7.6	0	0.6	0	0.995368	-14.148154	110	18	1.1	0.9;
	1656	1	-0	-0	0	48.74	0	1.026893	-18.16833	380	1	1.1	0.9;
	1657	1	257.2	-26.8	0	5.2	0	1.01455	-2.583627	220	3	1.1	0.9;
	1658	1	25.83	5.7	0	0.41	0	0.989399	-26.321497	110	18	1.1	0.9;
	1659	1	20.9	3	0	1.49	0	1.02733	-10.20009	154	1	1.1	0.9;
	1660	1	-13.91	4.6	0	0	0	0.963852	-31.39049	150	9	1.1	0.9;
	1661	1	22.02	7.5	0	0.59	0	1.013429	-28.621721	110	21	1.1	0.9;
	1662	1	3.67	-1.86	0	3.89	0	1.043136	-21.106513	220	5	1.1	0.9;
	1663	1	-592.21	-227.4	0.497795	203.015172	0	1.039738	-9.360278	380	3	1.1	0.9;
	1664	1	125.4	17.8	0	3.11	0	1.018662	-23.52823	154	1	1.1	0.9;
	1665	1	20.02	6.6	0	0.74	0	0.972263	-33.025045	110	21	1.1	0.9;
	1666	1	34.04	0	0	2.4	0	1.004945	-48.385165	220	4	1.1	0.9;
	1667	1	-0.75	-5.36	0	10.46	0	1.070751	44.286932	380	3	1.1	0.9;
	1668	1	-0	0	0.259637	9.774901	0	1.066902	-19.957089	380	19	1.1	0.9;
	1669	1	0	0	0.181549	8.990889	0	1.090065	-10.670345	220	18	1.3	0.7;
	1670	1	46.45	19.1	0	6.66	0	1.037909	-0.671739	150	8	1.1	0.9;
	1671	1	3.3	1.1	0	1.12	0	1.0457	-39.672452	110	22	1.1	0.9;
	1672	1	-28.13	-0.83	0	0.1	0	1.065391	-11.831936	220	5	1.1	0.9;
	1673	1	4.68	88.21	0	6.79	0	1.011939	61.912631	150	1	1.1	0.9;
	1674	1	16.92	5.6	0	0.2	0	1.002491	-39.764239	110	12	1.1	0.9;
	1675	1	55.26	-11.1	0	1.45	0	1.022736	-11.560606	220	3	1.1	0.9;
	1676	1	-0	0	0.171586	-1.4632	0	0.990679	-1.778952	380	2	1.1	0.9;
	1677	1	9.21	3.3	0	0.6	0	1.038167	-19.740289	110	19	1.1	0.9;
	1678	1	-0	0	0	15.6	0	1.081736	-35.98421	150	9	1.1	0.9;
	1679	1	0	0	0	21.06	0	1.02663	-33.818605	380	7	1.1	0.9;
	1680	2	0	0	0	0	0	1.038879	-2.415061	380	5	1.1	0.9;
	1681	1	7.41	2.7	0	1.24	0	0.986682	-22.832577	110	19	1.1	0.9;
	1682	1	22.28	5.67	0	1.2	0	1.053188	-21.492028	110	19	1.1	0.9;
	1683	1	-0	0	0	86.56	0	0.994877	-21.000973	220	24	1.1	0.9;
	1684	2	0	0	0	0	0	1.033747	-36.434239	380	6	1.1	0.9;
	1685	2	0	0	0	0	0	1.003785	-38.263404	150	10	1.1	0.9;
	1686	1	0	-0	0.37285	19.231353	0	1.078664	-24.703926	380	21	1.1	0.9;
	1687	1	0	-0	0	0.94	0	0.931358	-34.914023	150	9	1.1	0.9;
	1688	1	72.18	31.9	0	27.25	0	1.022677	-40.421217	220	6	1.1	0.9;
	1689	1	-0	-0	0	58.96	0	1.035394	10.980925	380	8	1.1	0.9;
	1690	1	118.2	-33	0	12.02	0	1.009017	-21.244805	220	3	1.1	0.9;
	1691	1	36.5	5.2	0	1.48	0	1.015921	-6.634268	154	1	1.1	0.9;
	1692	1	83.19	25.8	0	0.33	0	1.04113	-10.778443	220	5	1.1	0.9;
	1693	1	28.74	1.23	0	2.15	0	0.954353	-30.847547	110	18	1.1	0.9;
	1694	1	59.9	8.5	0	7.81	0	1.00883	-7.890285	154	1	1.1	0.9;
	1695	1	0	-0	0.090774	3.510444	0	1.058966	-23.450892	220	18	1.1	0.9;
	1696	1	12.41	4.4	0	0.34	0	1.023756	-25.893009	110	15	1.1	0.9;
	1697	2	0	0	0	0	0	1.036203	-50.835684	380	4	1.1	0.9;
	1698	1	14.72	5.3	0	1.72	0	1.051898	-1.832565	110	19	1.1	0.9;
	1699	1	4.3	1.8	0	0.29	0	1.032709	-27.758645	110	21	1.1	0.9;
	1700	1	5.01	2	0	0.6	0	1.028049	-29.769175	110	23	1.1	0.9;
	1701	1	0	0	0	1.36	0	0.916056	-35.870283	150	9	1.1	0.9;
	1702	1	47.45	18.7	0	1.36	0	1.04402	3.726033	150	8	1.1	0.9;
	1703	1	68.28	39	0.19855	79.84755	0	1.055276	-44.181014	380	10	1.1	0.9;
	1704	1	62.16	8	0	2.99	0	1.035555	-13.41017	220	5	1.1	0.9;
	1705	1	15.22	2.5	0	0.63	0	0.936445	-32.358787	110	18	1.1	0.9;
	1706	1	-0	-0	0	12.18	0	1.021774	19.260146	380	3	1.1	0.9;
	1707	1	92.6	36.3	0	4.51	0	0.971488	10.615995	220	3	1.1	0.9;
	1708	2	0	0	0	0	0	1.034651	-13.427998	380	5	1.1	0.9;
	1709	1	-15.38	-4.94	0	33.241519	0	1.005023	-14.903001	380	5	1.1	0.9;
	1710	1	138.2	40.3	0	1.15	0	1.063034	-25.548982	110	15	1.1	0.9;
	1711	1	25.13	-61.6	0	3.68	0	1.034881	-30.510188	220	5	1.1	0.9;
	1712	1	2.5	0.6	0	0.51	0	0.99131	-27.546404	110	18	1.1	0.9;
	1713	1	4.91	1.4	0	1.32	0	0.974081	-31.959491	110	18	1.1	0.9;
	1714	1	53.56	17.2	0	1.76	0	1.034594	-30.63887	110	21	1.1	0.9;
	1715	2	0	0	0	0	0	1.054322	-29.131045	110	3	1.1	0.9;
	1716	1	0	-0	0	16.62	0	1.021736	-41.185279	380	4	1.1	0.9;
	1717	2	0	0	0	0	0	1.022926	-23.513094	110	18	1.1	0.9;
	1718	1	-14.32	4.7	0	0.35	0	0.98164	-31.464689	150	9	1.1	0.9;
	1719	2	0	0	0	0	0	1.085028	23.495781	220	13	1.1	0.9;
	1720	2	0	0	0	0	0	1.072226	12.409229	380	3	1.1	0.9;
	1721	2	0	0	0	0	0	1.061454	2.336614	220	5	1.1	0.9;
	1722	1	-0.2	-1	0	0	0	1.023597	8.989728	150	8	1.1	0.9;
	1723	1	0	-0	0	57.29	0	1.047966	-18.896383	380	16	1.1	0.9;
	1724	1	8.01	4.9	0	0	0	0.981732	-31.651599	150	9	1.1	0.9;
	1725	1	0	0	0	0	0	1.013967	-32.66926	150	9	1.1	0.9;
	1726	2	0	0	0	0	0	1.067726	-21.121183	220	11	1.1	0.9;
	1727	1	0	0	0	0	0	0.98201	-27.314622	150	9	1.1	0.9;
	1728	1	34.44	14.4	0	48.56	0	1.046184	-49.192379	380	4	1.1	0.9;
	1729	1	68.38	42.4	0	1.39	0	1.010534	-27.286436	150	9	1.1	0.9;
	1730	1	-17.18	-4.35	0	44.419365	0	1.02153	-14.768844	380	5	1.1	0.9;
	1731	1	19.92	6.6	0	1.32	0	1.022153	-31.82525	110	21	1.1	0.9;
	1732	1	-0	0	0.135375	28.222175	0	1.075682	-23.981258	380	18	1.1	0.9;
	1733	1	16.02	5.3	0	0.67	0	1.023671	-36.810876	110	12	1.1	0.9;
	1734	1	-0	0	0	3.65	0	0.98465	-54.840874	220	10	1.1	0.9;
	1735	1	-0	0	0	1.2	0	1.061667	-21.904088	150	9	1.1	0.9;
	1736	2	0	0	0	0	0	1.06645	-25.241726	110	15	1.1	0.9;
	1737	1	9.71	3.5	0	1.44	0	1.058816	-21.952753	110	19	1.1	0.9;
	1738	2	0	0	0	0	0	1.015757	-1.685819	400	1	1.1	0.9;
	1739	1	-0	-0	0	5.09	0	1.030659	-9.889594	220	3	1.1	0.9;
	1740	1	10.11	2.5	0	0.47	0	0.999208	-27.007192	110	18	1.1	0.9;
	1741	1	162.5	48.4	0	5.73	0	0.983117	-45.853217	220	4	1.1	0.9;
	1742	1	0	0	0	6.68	0	1.021131	-32.671209	220	5	1.1	0.9;
	1743	2	0	0	0	0	0	1.033701	-10.940576	220	6	1.1	0.9;
	1744	1	2	0.3	0	1.02	0	0.996895	-31.540434	110	18	1.1	0.9;
	1745	1	0	0	0	0	0	0.962682	-27.37169	150	9	1.1	0.9;
	1746	1	-0	0	0	1.86	0	1.045012	-24.413457	220	5	1.1	0.9;
	1747	2	0	0	0	0	0	1.053359	12.399441	380	8	1.1	0.9;
	1748	2	0	0	0	0	0	1.038042	-36.808013	380	6	1.1	0.9;
	1749	1	39.24	24.3	0	0.07	0	1.008824	-32.768968	150	9	1.1	0.9;
	1750	1	58.46	6.6	0	10.13	0	1.032786	-19.499518	220	5	1.1	0.9;
	1751	1	11.51	2.3	0	0.66	0	1.00013	-18.836576	110	18	1.1	0.9;
	1752	1	85.8	13.4	0	3.23	0	1.030479	-53.459772	220	4	1.1	0.9;
	1753	1	0	-0	0.12	1.51	0	1.022441	-26.752782	220	21	1.1	0.9;
	1754	2	0	0	0	0	0	1.037047	-17.515273	380	5	1.1	0.9;
	1755	1	-9.78	-3.29	0	0.12	0	1.022734	-25.736039	110	18	1.1	0.9;
	1756	2	0	0	0	0	0	1.037533	-24.549182	220	18	1.1	0.9;
	1757	1	80.09	20	0	3.08	0	1.019224	15.628173	220	3	1.1	0.9;
	1758	1	0	-0	0	121.29	0	1.029582	-4.620703	380	5	1.1	0.9;
	1759	1	-0	-0	0	1.15	0	1.049716	-32.849576	220	6	1.1	0.9;
	1760	1	328.9	30.4	0	66.79	0	1.041544	-43.844942	380	6	1.1	0.9;
	1761	1	11.11	6.9	0	1.35	0	1.020634	-29.934806	150	9	1.1	0.9;
	1762	1	45.75	20.1	0	0.41	0	1.009911	-20.721314	110	18	1.1	0.9;
	1763	1	-0	-0	0	0.28	0	1.031495	-31.068516	380	5	1.1	0.9;
	1764	1	25.43	14.7	0	7.79	0	1.061599	-23.124487	220	18	1.1	0.9;
	1765	1	122.2	68.9	0	0.58	0	1.048441	-9.044258	380	3	1.1	0.9;
	1766	1	26.03	8.7	0	0.29	0	0.985394	-26.614731	110	18	1.1	0.9;
	1767	1	333.5	-1.4	0	8.37	0	1.045464	-45.911677	220	5	1.1	0.9;
	1768	1	73.98	22.5	0	1.15	0	1.041318	-11.34023	220	5	1.1	0.9;
	1769	1	35.14	1.8	0	6.12	0	1.02889	-47.198347	220	4	1.1	0.9;
	1770	1	-0	-0	0	2.55	0	1.012291	-41.848384	220	10	1.1	0.9;
	1771	1	22.43	29.5	0	0.08	0	1.014086	-33.092664	150	9	1.1	0.9;
	1772	1	12.6	1.8	0	3.8	0	1.004091	-8.612218	154	1	1.1	0.9;
	1773	1	-0	-0	0	11.49	0	0.99209	27.973102	220	3	1.1	0.9;
	1774	1	-32.04	10.3	0	82.52	0	1.05267	-45.804632	380	10	1.1	0.9;
	1775	1	-0	0	0	1.64	0	1.050004	-26.478626	220	5	1.1	0.9;
	1776	2	0	0	0	0	0	0.967759	68.850675	380	1	1.1	0.9;
	1777	1	0	-0	0	0.98	0	1.030583	15.68913	110	8	1.1	0.9;
	1778	1	-0.18	-3.85	0	0	0	1.014711	16.417794	150	8	1.1	0.9;
	1779	1	5.51	2	0	0.79	0	1.03145	-24.87137	110	19	1.1	0.9;
	1780	1	-1.8	-4.8	0	0.44	0	1.050808	-24.75612	150	9	1.1	0.9;
	1781	1	16.52	5.5	0	1.59	0	1.029294	-34.011025	110	22	1.1	0.9;
	1782	1	28.63	10	0	1.34	0	1.063284	-17.401778	110	19	1.1	0.9;
	1783	1	-0	-0.8	0	0.46	0	1.039054	1.697583	110	18	1.1	0.9;
	1784	1	0	0	0	0	0	0.99228	-31.5665	150	9	1.1	0.9;
	1785	2	0	0	0	0	0	0.998	-17.23253	110	18	1.1	0.9;
	1786	2	0	0	0	0	0	1.083767	-19.098158	380	9	1.1	0.9;
	1787	1	32.94	10.1	0	0.41	0	1.016217	11.837997	110	8	1.1	0.9;
	1788	2	0	0	0	0	0	1.021039	15.216998	380	8	1.1	0.9;
	1789	1	45.75	27.5	0	4.37	0	1.041516	-18.666736	220	7	1.1	0.9;
	1790	1	-0	-0	0	2.19	0	1.040536	-3.329522	220	3	1.1	0.9;
	1791	1	-19.32	6.4	0	0	0	1.041225	-23.215191	150	9	1.1	0.9;
	1792	1	-0	-0	0	6.8	0	1.022722	16.726688	380	3	1.1	0.9;
	1793	2	0	0	0	0	0	1.020092	-44.591103	220	4	1.1	0.9;
	1794	2	0	0	0	0	0	1.04708	-0.386636	380	5	1.1	0.9;
	1795	1	71.18	23.8	0	0.07	0	1.015327	-29.275104	110	21	1.1	0.9;
	1796	1	13.31	7.9	0	0.45	0	0.98894	-27.541031	110	18	1.1	0.9;
	1797	2	0	0	0	0	0	1.005885	-26.295127	110	18	1.1	0.9;
	1798	1	-0.63	-0.83	0	22.810288	0	0.980384	-7.672562	380	5	1.1	0.9;
	1799	1	-0	-0	0	1.46	0	1.039206	-21.602027	154	1	1.1	0.9;
	1800	1	8.3	1.2	0	4.96	0	1.009802	-2.151969	154	1	1.1	0.9;
	1801	1	-0	0	0	0.67	0	1.025453	-24.903774	150	9	1.1	0.9;
	1802	1	8.51	3.2	0	2.93	0	1.002389	-46.706656	220	4	1.1	0.9;
	1803	1	-0	-0	0	0.05	0	1.034827	-22.379623	220	5	1.1	0.9;
	1804	1	23.02	10.5	0	3.96	0	1.014351	-36.087944	220	6	1.1	0.9;
	1805	1	112.6	16.1	0	3.38	0	1.038865	3.663556	154	1	1.1	0.9;
	1806	1	25.43	8.3	0	1.35	0	0.961839	-33.975737	110	21	1.1	0.9;
	1807	1	-0	0	0	24.17	0	1.037678	34.179925	380	3	1.1	0.9;
	1808	2	0	0	0	0	0	1.04521	-0.934819	220	5	1.1	0.9;
	1809	1	-0	0	0	239.642195	0	0.962067	-19.96531	750	20	1.1	0.9;
	1810	1	52.46	11.3	0	0.27	0	1.033529	16.621587	110	8	1.1	0.9;
	1811	1	99.51	1.5	0	32.89	0	1.042994	33.135037	380	3	1.1	0.9;
	1812	2	0	0	0	0	0	1.03994	10.217163	154	1	1.1	0.9;
	1813	1	88.92	19.79	0	2.81	0	1.02257	-7.742655	220	5	1.1	0.9;
	1814	2	0	0	0	0	0	1.039354	-25.863952	380	6	1.1	0.9;
	1815	1	63.66	9.7	0	0	0	1.02229	-0.315036	220	2	1.1	0.9;
	1816	1	335.3	-101.3	0	32.69	0	1.047931	-23.126225	380	11	1.1	0.9;
	1817	1	108.91	33.7	0	56.24	0	1.066116	-12.800258	380	5	1.1	0.9;
	1818	1	-29.73	9.8	0	2.31	0	1.006291	-26.865863	150	9	1.1	0.9;
	1819	2	0	0	0	0	0	1.039397	-19.104141	380	6	1.1	0.9;
	1820	1	-0	0	0	0.06	0	1.05875	-15.920541	220	19	1.1	0.9;
	1821	1	14.62	2	0	10.89	0	1.088441	-11.783927	220	18	1.3	0.7;
	1822	2	0	0	0	0	0	1.043084	-43.003837	380	6	1.1	0.9;
	1823	2	0	0	0	0	0	1.020728	16.810237	380	3	1.1	0.9;
	1824	1	-0.65	-1.81	0	1.52	0	1.075908	-26.411443	150	9	1.1	0.9;
	1825	1	-0	0	0	0.38	0	1.032866	-0.69082	380	3	1.1	0.9;
	1826	1	56.46	22	0	41.96	0	1.023046	-43.116686	220	6	1.1	0.9;
	1827	1	114.2	16.3	0	57.91	0	1.003664	-18.891929	400	1	1.1	0.9;
	1828	1	2.5	0.6	0	1.06	0	1.005228	-18.994194	110	18	1.1	0.9;
	1829	1	5.01	13	0	30.48	0	1.001907	-31.052208	220	11	1.1	0.9;
	1830	1	17.82	6.7	0	0.09	0	1.057162	-24.158999	110	15	1.1	0.9;
	1831	1	0	0	0	0.51	0	1.017814	-32.141888	150	9	1.1	0.9;
	1832	1	34.14	21.1	0	0.35	0	0.961777	-31.641638	150	9	1.1	0.9;
	1833	1	0	0	0	3.33	0	1.031879	-22.829129	220	5	1.1	0.9;
	1834	1	8.21	2.5	0	0.12	0	1.011612	-23.487852	110	18	1.1	0.9;
	1835	1	44.23	22.19	0	4.35	0	1.049216	-0.351418	220	13	1.1	0.9;
	1836	1	62.37	25.5	0	3.98	0	1.027825	-1.944004	150	8	1.1	0.9;
	1837	1	113.27	27.67	0	3.61	0	1.032886	-1.726797	220	13	1.1	0.9;
	1838	1	0	-0	0	3.02	0	1.032898	-22.509058	220	5	1.1	0.9;
	1839	1	0.4	-170.2	0	3.04	0	1.051337	-16.123433	380	3	1.1	0.9;
	1840	1	-0	0	0	0.62	0	0.965003	-31.473703	150	9	1.1	0.9;
	1841	1	56.26	8.5	0	5.32	0	1.026713	-53.753885	220	4	1.1	0.9;
	1842	1	-0	-0	0	25.22	0	1.033955	-3.107877	220	1	1.1	0.9;
	1843	2	0	0	0	0	0	1.080669	-17.696231	110	19	1.1	0.9;
	1844	1	-28.14	-0.76	0	0.1	0	1.065409	-11.831386	220	5	1.1	0.9;
	1845	1	123.9	17.6	0	11.56	0	1.008717	-7.820464	154	1	1.1	0.9;
	1846	2	0	0	0	0	0	1.028122	20.986392	154	1	1.1	0.9;
	1847	2	0	0	0	0	0	1.002793	-26.99395	110	18	1.1	0.9;
	1848	1	22.02	21.8	0	10.29	0	1.030639	-33.753117	220	10	1.1	0.9;
	1849	1	5.11	2.4	0	1.34	0	1.032778	-10.179852	110	18	1.1	0.9;
	1850	2	0	0	0	0	0	1.014744	-51.762708	220	4	1.1	0.9;
	1851	2	0	0	0	0	0	1.053982	-13.344252	220	5	1.1	0.9;
	1852	2	0	0	0	0	0	1.054605	-26.16781	220	5	1.1	0.9;
	1853	1	25.73	8.9	0	1.29	0	0.981852	-20.846506	110	18	1.1	0.9;
	1854	1	0.2	0.5	0	0.42	0	1.034977	58.80321	220	3	1.1	0.9;
	1855	1	0	0	0	1.94	0	0.982802	6.963882	220	3	1.1	0.9;
	1856	1	5.51	2	0	0.95	0	1.058825	-0.684818	110	19	1.1	0.9;
	1857	1	0	0	0	1.53	0	1.029881	-11.083766	220	5	1.1	0.9;
	1858	1	10.51	4.7	0	1.54	0	0.992973	-21.379419	110	18	1.1	0.9;
	1859	1	369.5	16.7	0	10.6	0	0.993828	-26.242893	220	6	1.1	0.9;
	1860	1	88.8	-3.2	0	13.63	0	1.060709	-25.980643	220	5	1.1	0.9;
	1861	1	0	-0	0	19.01	0	1.02821	19.339657	380	3	1.1	0.9;
	1862	1	177.4	41.7	0	2.98	0	0.985152	-40.901374	150	10	1.1	0.9;
	1863	1	-0	0	0	15.04	0	1.084436	25.639895	220	3	1.1	0.9;
	1864	1	-19.62	28.6	0	1.69	0	1.04085	-28.072064	150	9	1.1	0.9;
	1865	1	55.26	23.2	0	8.01	0	1.031858	-26.08876	220	18	1.1	0.9;
	1866	1	209.2	23.5	0	6.44	0	1.037512	-28.611495	220	5	1.1	0.9;
	1867	1	0	0	0	0.06	0	1.035531	-36.227274	220	6	1.1	0.9;
	1868	1	11.71	3.8	0	2.96	0	1.041623	-15.644277	220	5	1.1	0.9;
	1869	1	-2.23	-1.27	0	3.64	0	1.031966	-44.755675	220	4	1.1	0.9;
	1870	1	-0	0	0	0.74	0	1.021635	0.879281	150	8	1.1	0.9;
	1871	2	0	0	0	0	0	1.000876	-50.907624	220	6	1.1	0.9;
	1872	1	94.3	13.8	0	5.3	0	1.024147	-54.574191	220	4	1.1	0.9;
	1873	1	83.29	30.1	0	4.68	0	1.109647	-16.333961	110	19	1.3	0.7;
	1874	1	223.9	72.6	0	2.04	0	1.063601	-32.225677	110	1	1.1	0.9;
	1875	1	0	0	0	0	0	0.990784	-30.586521	150	9	1.1	0.9;
	1876	1	-0	-0	0	6.273568	0	1.027486	-14.490891	380	5	1.1	0.9;
	1877	2	0	0	0	0	0	1.050568	-21.692469	150	9	1.1	0.9;
	1878	1	0	0	0	0.13	0	0.998257	-26.891995	110	18	1.1	0.9;
	1879	1	-0	0	0	0.57	0	1.013756	-31.743428	150	9	1.1	0.9;
	1880	1	50.26	16.4	0	0.23	0	1.015921	0.245944	150	8	1.1	0.9;
	1881	1	-0	-0	0	50.2	0	1.020701	-4.60248	400	1	1.1	0.9;
	1882	1	64.27	20.1	0	5.4	0	1.005906	-37.450211	220	6	1.1	0.9;
	1883	1	27.33	5	0	6.77	0	1.027298	-16.582215	220	5	1.1	0.9;
	1884	1	264.33	86.69	0	91.94	0	1.062468	-10.688115	380	14	1.1	0.9;
	1885	1	-0	0	0	29.88	0	1.081635	-27.637697	380	1	1.1	0.9;
	1886	1	2.2	1.1	0	1.57	0	1.014141	-40.707847	110	22	1.1	0.9;
	1887	1	103.1	14.7	0	4.43	0	1.02578	0.474928	154	1	1.1	0.9;
	1888	2	0	0	0	0	0	1.034904	-2.790789	220	5	1.1	0.9;
	1889	1	57.16	0	0	6.58	0	1.025428	9.639663	110	8	1.1	0.9;
	1890	2	0	0	0	0	0	1.04228	53.799976	380	8	1.1	0.9;
	1891	1	28.63	17.8	0	0.06	0	1.001273	-31.157699	150	9	1.1	0.9;
	1892	2	0	0	0	0	0	1.020022	-4.163834	154	1	1.1	0.9;
	1893	1	0	0	0	11.82	0	1.011913	-4.946377	380	3	1.1	0.9;
	1894	1	22.22	5.9	0	10.28	0	1.021165	-48.218676	220	4	1.1	0.9;
	1895	1	20.82	-0	0	0.14	0	1.063985	7.571984	220	5	1.1	0.9;
	1896	1	-0	-0	0	1.18	0	1.01324	-6.532403	380	5	1.1	0.9;
	1897	1	53.96	4.9	0	4.21	0	1.007691	-60.940761	220	4	1.1	0.9;
	1898	2	0	0	0	0	0	1.074634	-36.208422	380	4	1.1	0.9;
	1899	1	0	-0	0.202076	48.319777	0	1.065172	-47.315755	380	10	1.1	0.9;
	1900	2	0	0	0	0	0	1.077688	59.124475	380	3	1.1	0.9;
	1901	1	0.2	1.4	0	0.49	0	1.01412	8.984855	220	3	1.1	0.9;
	1902	1	26.43	7.7	0	7.88	0	1.034198	-38.764899	110	22	1.1	0.9;
	1903	1	-0	0	0	3.15	0	1.012188	-32.632231	150	9	1.1	0.9;
	1904	1	16.02	5.3	0	0.67	0	1.017157	-38.533775	110	12	1.1	0.9;
	1905	2	0	0	0	0	0	1.049117	-16.463578	380	19	1.1	0.9;
	1906	1	25	3.6	0	2.17	0	1.020081	15.34146	154	1	1.1	0.9;
	1907	2	0	0	0	0	0	1.038735	59.621829	220	3	1.1	0.9;
	1908	1	-11.95	-3.92	0	0.18	0	1.025798	-23.331651	110	18	1.1	0.9;
	1909	2	0	0	0	0	0	1.019677	-40.082212	220	4	1.1	0.9;
	1910	1	-0	0	0	2.68	0	1.047725	-6.761314	220	5	1.1	0.9;
	1911	1	39.54	14.3	0	0.91	0	1.04019	-24.253147	110	19	1.1	0.9;
	1912	1	52.06	14.3	0	1.59	0	0.980263	-45.621958	220	4	1.1	0.9;
	1913	1	91.1	54.8	0	30.59	0	1.023393	9.328465	150	8	1.1	0.9;
	1914	2	0	0	0	0	0	1.038898	-21.122268	220	5	1.1	0.9;
	1915	1	33.64	3.9	0	3.37	0	1.042728	-27.029592	220	7	1.1	0.9;
	1916	1	10.51	3.5	0	0.66	0	1.000045	-39.384612	110	12	1.1	0.9;
	1917	1	197.5	-31	0	17.34	0	1.046127	-25.465701	220	5	1.1	0.9;
	1918	1	-0	0	0	21.97	0	1.032995	-0.176302	380	3	1.1	0.9;
	1919	2	0	0	0	0	0	1.036866	-12.936294	380	6	1.1	0.9;
	1920	1	0	0	0	94.987645	0	1.043475	-40.018739	380	17	1.1	0.9;
	1921	1	49.25	4.9	0	1.06	0	1.006265	-16.469593	110	18	1.1	0.9;
	1922	1	-0	-0	0.126343	10.705031	0	1.044766	-25.117213	220	19	1.1	0.9;
	1923	1	192.11	45.5	0	68.04	0	0.997765	-11.91036	380	5	1.1	0.9;
	1924	1	108.3	62	0	36.65	0	1.02624	-24.378735	220	7	1.1	0.9;
	1925	2	0	0	0	0	0	1.003974	-2.389073	400	1	1.1	0.9;
	1926	1	14.92	4.9	0	0.65	0	1.026522	-34.620402	110	12	1.1	0.9;
	1927	2	0	0	0	0	0	1.0345	-5.60456	220	3	1.1	0.9;
	1928	1	0	0	0	2.16	0	1.018472	-39.815703	110	22	1.1	0.9;
	1929	1	45.6	6.5	0	95.24	0	0.990914	4.410433	400	1	1.1	0.9;
	1930	1	14.8	6.6	0	84.333648	0	1.121369	-22.847507	380	9	1.3	0.7;
	1931	1	24.33	15	0	2.11	0	0.982732	-29.85688	150	9	1.1	0.9;
	1932	1	51.86	13.9	0	0.67	0	1.000503	-36.252297	110	1	1.1	0.9;
	1933	1	-0	0	0	45.35	0	1.05258	-24.478973	380	1	1.1	0.9;
	1934	1	-0	-0	0	0.07	0	1.018297	-32.149517	150	9	1.1	0.9;
	1935	2	0	0	0	0	0	1.014926	1.628294	154	1	1.1	0.9;
	1936	1	0	0	0	30.78	0	1.035286	5.018235	150	8	1.1	0.9;
	1937	1	-0	-0	0	4.76	0	1.050442	30.272156	380	1	1.1	0.9;
	1938	2	0	0	0	0	0	1.038487	11.052176	154	1	1.1	0.9;
	1939	1	-11.44	-3.29	0	0.04	0	1.024984	-22.605301	220	6	1.1	0.9;
	1940	1	0	0	0	1.19	0	1.033019	-18.314754	220	5	1.1	0.9;
	1941	1	0	-0	0	68.01	0	1.009718	-21.601854	400	1	1.1	0.9;
	1942	1	44.75	4.1	0	3.12	0	1.010513	-59.671487	220	4	1.1	0.9;
	1943	1	4.4	99.7	0	40.45	0	1.041345	6.510491	380	3	1.1	0.9;
	1944	1	-0	-0	0	0.27	0	1.026355	56.907195	220	3	1.1	0.9;
	1945	1	0	0	0	0	0	0.97125	-37.361789	150	9	1.1	0.9;
	1946	1	19.42	-0.8	0	0.78	0	1.036261	39.10102	220	3	1.1	0.9;
	1947	2	0	0	0	0	0	1.053959	-40.521126	380	4	1.1	0.9;
	1948	1	32.2	4.6	0	2.54	0	1.013181	-6.395894	154	1	1.1	0.9;
	1949	1	159.8	45	0	5.6	0	1.009215	-9.573534	220	3	1.1	0.9;
	1950	1	-0	0	0	2.1	0	1.019523	-0.457298	380	3	1.1	0.9;
	1951	1	-0	-0	0	0.76	0	1.048768	-20.861454	220	1	1.1	0.9;
	1952	1	115.1	10.4	0	6.12	0	1.035674	-28.381537	220	11	1.1	0.9;
	1953	1	0	-0	0.059832	20.003574	0	1.007258	65.035316	380	1	1.1	0.9;
	1954	1	21.62	13.4	0	1.79	0	1.014711	-27.065473	150	9	1.1	0.9;
	1955	1	0	0	0	10.51	0	1.02487	16.627166	380	3	1.1	0.9;
	1956	1	0	0	0	51.66	0	1.06386	29.988306	380	8	1.1	0.9;
	1957	1	76.28	2.9	0	7.28	0	1.01632	-48.919694	220	4	1.1	0.9;
	1958	1	119.9	17.1	0	8.99	0	1.022554	-7.505209	154	1	1.1	0.9;
	1959	2	0	0	0	0	0	1.069553	-10.957516	220	5	1.1	0.9;
	1960	1	-0	-0	0	1.94	0	1.027615	-28.365664	110	21	1.1	0.9;
	1961	1	3.2	1.5	0	1.81	0	0.988109	-17.936842	110	18	1.1	0.9;
	1962	1	0	0	0	0.27	0	1.026482	56.912845	220	3	1.1	0.9;
	1963	1	29.43	10.7	0	0.2	0	1.039808	-19.664121	110	19	1.1	0.9;
	1964	1	-0	-0	0	30.05	0	1.053304	-24.416889	220	21	1.1	0.9;
	1965	1	248.7	8.1	0	12.46	0	1.062912	-50.015895	220	5	1.1	0.9;
	1966	1	-20.6	-9.4	0	1.12	0	1.023293	-31.684433	150	9	1.1	0.9;
	1967	1	46.05	16.7	0	0.49	0	1.075382	-18.046502	110	19	1.1	0.9;
	1968	1	118.4	137	0	0	0	1.025488	-11.380774	150	2	1.1	0.9;
	1969	1	0	-0	0	0.86	0	1.025648	-54.021474	220	4	1.1	0.9;
	1970	2	0	0	0	0	0	1.026295	-21.249181	110	19	1.1	0.9;
	1971	1	89	-1.8	0	6.15	0	1.027477	-20.372496	220	3	1.1	0.9;
	1972	1	1.8	4.9	0	92.20984	0	1.118085	-24.261644	380	9	1.3	0.7;
	1973	1	210.1	66.2	0	59.36	0	1.069016	-26.697489	220	5	1.1	0.9;
	1974	1	241.9	38.8	0	12.46	0	0.963101	-44.62521	220	10	1.1	0.9;
	1975	1	16.7	2.4	0	2.64	0	1.024667	-6.601257	154	1	1.1	0.9;
	1976	1	0	-0	0	12.41	0	1.032693	7.341385	380	1	1.1	0.9;
	1977	1	-0	0	0	116.64	0	1.034395	-13.50295	380	6	1.1	0.9;
	1978	1	0	-0	0.135493	29.07062	0	1.102201	-24.077627	380	21	1.3	0.7;
	1979	1	0	0	0	0.14	0	1.069875	-33.898765	220	5	1.1	0.9;
	1980	1	177.3	36.4	0	0.7	0	1.037878	-21.73005	220	5	1.1	0.9;
	1981	1	12.41	3.9	0	1.14	0	1.030042	-20.651221	110	19	1.1	0.9;
	1982	2	0	0	0	0	0	1.008845	-5.393338	154	1	1.1	0.9;
	1983	1	188.1	26.8	0	5.85	0	1.007467	-25.426976	154	1	1.1	0.9;
	1984	1	0	0	0	0	0	1.031492	3.864128	150	8	1.1	0.9;
	1985	1	147.4	-55.9	0	1.12	0	0.982099	-58.088778	220	4	1.1	0.9;
	1986	1	0	-0	0	1.83	0	1.031924	-29.387901	150	9	1.1	0.9;
	1987	1	70.18	44.7	0	8.09	0	0.994461	-30.251812	150	9	1.1	0.9;
	1988	1	50.26	5.5	0	1.3	0	1.002741	-26.468764	110	18	1.1	0.9;
	1989	1	69.78	20.3	0	10.26	0	1.026746	14.187408	110	8	1.1	0.9;
	1990	1	61.87	-12.4	0	0.83	0	1.008358	-32.735877	220	6	1.1	0.9;
	1991	2	0	0	0	0	0	1.022737	-40.410279	220	6	1.1	0.9;
	1992	1	0.2	0.2	0	0.3	0	1.015068	-2.593229	220	3	1.1	0.9;
	1993	1	89.5	33.9	0	1.52	0	0.995414	9.386988	220	3	1.1	0.9;
	1994	1	0	0	0.414968	57.524024	0	1.051045	-8.039112	380	3	1.1	0.9;
	1995	1	-0	-0	0	10.44	0	1.009394	-23.403987	154	1	1.1	0.9;
	1996	1	15.22	5.1	0	1.34	0	0.980315	-24.014608	110	18	1.1	0.9;
	1997	1	0.9	0.5	0	0.9	0	1.041222	-31.995411	110	1	1.1	0.9;
	1998	1	25.73	7.3	0	2.82	0	1.029465	-20.456335	220	5	1.1	0.9;
	1999	2	0	0	0	0	0	1.076837	-40.120222	380	10	1.1	0.9;
	2000	1	54	29.1	0	204.291522	0	1.076756	-24.879698	380	9	1.1	0.9;
	2001	1	82.6	11.7	0	0.38	0	1.009625	-24.792399	154	1	1.1	0.9;
	2002	1	0	-0	0	0	0	1.079208	31.615158	380	8	1.1	0.9;
	2003	1	139.1	60.4	0	1.98	0	1.009595	8.946599	220	3	1.1	0.9;
	2004	1	-0	0	0	3.72	0	1.043992	-32.878294	110	22	1.1	0.9;
	2005	1	54.66	-0	0	2.97	0	1.022518	-24.292442	220	7	1.1	0.9;
	2006	1	-40.44	1.5	0	14.2	0	1.019749	-38.778573	220	10	1.1	0.9;
	2007	1	82.1	11.7	0	3.76	0	1.032277	-11.857866	154	1	1.1	0.9;
	2008	1	11.41	4.4	0	0.12	0	0.995998	-30.915861	110	18	1.1	0.9;
	2009	1	154.8	-47.1	0	4.18	0	1.022468	-51.350991	220	4	1.1	0.9;
	2010	1	0	0	0	0	0	0.948815	-33.811513	150	9	1.1	0.9;
	2011	1	-0	-0	0	0.85	0	1.017349	-26.22025	110	18	1.1	0.9;
	2012	1	41.45	4.9	0	18.43	0	1.034735	-18.980386	220	5	1.1	0.9;
	2013	1	0	-0	0	18.99	0	1.03859	-5.132662	220	13	1.1	0.9;
	2014	1	0	-0	0	0.34	0	1.089513	-14.472191	380	1	1.3	0.7;
	2015	1	25.93	8	0	1.24	0	1.019267	-26.382886	110	21	1.1	0.9;
	2016	1	119	-315.3	0	41.89	0	1.03604	7.540805	380	3	1.1	0.9;
	2017	2	0	0	0	0	0	1.042851	-26.988688	220	7	1.1	0.9;
	2018	2	0	0	0	0	0	1.00366	-11.898419	110	18	1.1	0.9;
	2019	1	72.08	21.3	0	7.19	0	1.041426	-21.3851	220	5	1.1	0.9;
	2020	1	0	0	0	9.98	0	1.046724	-24.420828	220	5	1.1	0.9;
	2021	1	3.9	1	0	3.81	0	1.04574	-25.29705	220	5	1.1	0.9;
	2022	1	8.11	2.9	0	0.52	0	1.016664	-23.914148	110	19	1.1	0.9;
	2023	1	36.44	12	0	1.39	0	1.026831	-21.556685	110	21	1.1	0.9;
	2024	1	8.91	1.8	0	1.78	0	0.992581	-28.519503	110	18	1.1	0.9;
	2025	1	89.1	2.3	0	1.68	0	1.006782	-49.35595	220	6	1.1	0.9;
	2026	1	16.02	4	0	5.56	0	1.025828	-33.804307	220	11	1.1	0.9;
	2027	1	0	-0	0.181549	5.390889	0	1.033824	-19.844839	220	18	1.1	0.9;
	2028	1	110.9	15.8	0	30.2	0	1.020802	-19.70646	400	1	1.1	0.9;
	2029	1	90.1	12.7	0	10.61	0	1.014611	9.090578	220	3	1.1	0.9;
	2030	2	0	0	0	0	0	1.067534	15.50451	150	8	1.1	0.9;
	2031	1	-0	-0	0.104186	55.436651	0	1.076819	-40.134915	380	10	1.1	0.9;
	2032	1	56.86	35.2	0	2.29	0	1.010431	-30.336741	150	9	1.1	0.9;
	2033	1	-0	-0	0	0.66	0	1.026288	-4.598874	220	3	1.1	0.9;
	2034	1	153.64	29.2	0	7.65	0	1.017325	-29.183978	220	6	1.1	0.9;
	2035	2	0	0	0	0	0	1.028122	0.401	380	5	1.1	0.9;
	2036	1	6.31	3.8	0	0.36	0	1.014301	-26.609181	110	18	1.1	0.9;
	2037	1	28.83	9.6	0	0.8	0	1.008938	-32.979344	110	21	1.1	0.9;
	2038	1	0	0	0	14.4	0	1.039244	17.342907	380	3	1.1	0.9;
	2039	1	297.7	258.6	0	65.78	0	1.047273	-15.334462	380	6	1.1	0.9;
	2040	1	178.4	69.1	0	14.2	0	1.013224	-46.005069	220	6	1.1	0.9;
	2041	1	-0	-0	0.273444	108.530355	0	1.078629	-16.00504	380	19	1.1	0.9;
	2042	1	14.02	6.5	0	0.09	0	1.033024	-6.443506	220	5	1.1	0.9;
	2043	1	3.2	1.7	0	0.05	0	1.047821	-21.373368	220	5	1.1	0.9;
	2044	1	0	0	0	0.14	0	1.029887	-12.614951	220	5	1.1	0.9;
	2045	1	-0	0	0	0.74	0	0.961784	-32.465814	150	9	1.1	0.9;
	2046	2	0	0	0	0	0	1.042366	66.456328	380	3	1.1	0.9;
	2047	1	-0.65	-1.44	0	0.16	0	1.031001	-29.324914	150	9	1.1	0.9;
	2048	1	19.22	6.9	0	1.16	0	1.087553	-21.787548	110	15	1.1	0.9;
	2049	1	29.3	4.2	0	5.26	0	1.052423	-6.622476	154	1	1.1	0.9;
	2050	2	0	0	0	0	0	1.004886	-1.149277	380	5	1.1	0.9;
	2051	2	0	0	0	0	0	1.047683	-18.853377	380	16	1.1	0.9;
	2052	1	0	0	0	0	0	0.962887	-32.243935	150	9	1.1	0.9;
	2053	1	-0	0	0	2.99	0	1.037116	-28.920314	220	1	1.1	0.9;
	2054	2	0	0	0	0	0	1.055574	-23.781333	220	7	1.1	0.9;
	2055	1	52.66	21.5	0	0.04	0	1.014543	-4.866032	150	8	1.1	0.9;
	2056	1	-0	0	0	7.338808	0	1.027846	-9.315749	380	5	1.1	0.9;
	2057	1	120.2	20.5	0	19.94	0	1.031313	-31.559763	220	5	1.1	0.9;
	2058	1	0	0	0	1.31	0	1.052684	-45.804715	380	10	1.1	0.9;
	2059	1	0	0	0	16.08	0	1.104065	-21.087887	380	9	1.3	0.7;
	2060	1	115.1	23.5	0	3.78	0	1.009599	-58.910148	220	4	1.1	0.9;
	2061	1	-0	0	0	0.31	0	0.995781	-17.545532	110	18	1.1	0.9;
	2062	1	-0.61	-1.33	0	0.43	0	1.05583	-23.414928	110	15	1.1	0.9;
	2063	2	0	0	0	0	0	1.027043	-12.729431	220	13	1.1	0.9;
	2064	1	14.4	2	0	4.17	0	1.013124	-7.475327	154	1	1.1	0.9;
	2065	1	98.6	14	0	12.86	0	1.021047	-7.188533	154	1	1.1	0.9;
	2066	1	-0	-0	0	5.18	0	1.030635	-9.878496	220	3	1.1	0.9;
	2067	1	0	0	0	0	0	1.036075	-29.78924	150	9	1.1	0.9;
	2068	1	50.16	18	0	0.1	0	1.061101	-22.058249	110	19	1.1	0.9;
	2069	1	1.4	-2	0	5.05	0	1.028411	65.082435	150	1	1.1	0.9;
	2070	1	14	2	0	3.32	0	1.01757	-5.507258	154	1	1.1	0.9;
	2071	1	-0	0	0	2.88	0	1.042789	47.058431	220	3	1.1	0.9;
	2072	1	101.2	2.1	0	38.43	0	1.059961	-17.698679	380	5	1.1	0.9;
	2073	1	86.9	30.3	0	1.32	0	1.046226	-25.564376	110	15	1.1	0.9;
	2074	1	-0	-0	0	5.7	0	1.020506	2.243727	380	3	1.1	0.9;
	2075	1	-0	-0	0	20.79	0	1.017699	-1.963235	380	3	1.1	0.9;
	2076	1	7	1	0	1.54	0	1.019808	7.364672	154	1	1.1	0.9;
	2077	1	72.18	44.9	0	6.34	0	1.025776	11.372538	150	8	1.1	0.9;
	2078	1	77.19	18.6	0	2.22	0	1.018119	-14.911286	220	5	1.1	0.9;
	2079	1	0	-0	0	13.94	0	1.053582	-12.209929	220	5	1.1	0.9;
	2080	1	163.7	94.1	0	18.55	0	1.014009	-8.703475	380	3	1.1	0.9;
	2081	1	30.93	2.8	0	2.11	0	0.975157	-49.400861	220	4	1.1	0.9;
	2082	2	0	0	0	0	0	1.014544	1.61216	154	1	1.1	0.9;
	2083	1	121.4	20.2	0	16.61	0	1.031586	-18.751547	220	5	1.1	0.9;
	2084	1	-0	-0	0	6.01	0	1.06401	18.931087	380	14	1.1	0.9;
	2085	2	0	0	0	0	0	1.029587	2.315883	380	5	1.1	0.9;
	2086	2	0	0	0	0	0	1.058532	-19.958565	220	18	1.1	0.9;
	2087	1	23.7	3.4	0	4.27	0	1.013497	-5.459814	154	1	1.1	0.9;
	2088	1	-0	0	0	6.02	0	1.054586	-12.178402	220	5	1.1	0.9;
	2089	1	199.86	56.93	0	46.83	0	1.038708	-26.108581	220	5	1.1	0.9;
	2090	1	11.2	1.6	0	4	0	1.01105	-1.774887	154	1	1.1	0.9;
	2091	2	0	0	0	0	0	1.02498	-37.030652	220	4	1.1	0.9;
	2092	1	110.09	53.32	0	0	0	0.963686	-22.292553	110	18	1.1	0.9;
	2093	1	-5.64	-2.84	0	34.582195	0	1.019357	-8.221102	380	5	1.1	0.9;
	2094	1	65.87	31.4	0	3.46	0	1.060812	42.767804	220	3	1.1	0.9;
	2095	1	-0	-0	0	0	0	1.021064	-32.110852	150	9	1.1	0.9;
	2096	2	0	0	0	0	0	1.046008	-46.580961	380	4	1.1	0.9;
	2097	1	54.26	19.5	0	0.05	0	1.026099	-23.179994	220	7	1.1	0.9;
	2098	2	0	0	0	0	0	1.064541	1.202606	380	14	1.1	0.9;
	2099	1	-3.7	-8	0	19.02	0	1.029049	-27.094361	380	7	1.1	0.9;
	2100	2	0	0	0	0	0	1.022492	-41.032214	380	6	1.1	0.9;
	2101	1	-0	-0	0	38.32	0	1.028647	-5.56639	380	5	1.1	0.9;
	2102	1	-0	0	0	3.95	0	0.971407	27.718442	220	3	1.1	0.9;
	2103	1	0	0	0	2.67	0	1.017917	-29.06054	220	6	1.1	0.9;
	2104	2	0	0	0	0	0	1.005555	-6.627974	154	1	1.1	0.9;
	2105	1	0	-0	0	6.22	0	1.038502	66.587145	380	1	1.1	0.9;
	2106	1	-0	0	0	0.09	0	1.007207	-30.247885	110	21	1.1	0.9;
	2107	2	0	0	0	0	0	1.040878	-0.384592	380	2	1.1	0.9;
	2108	1	0	0	0.090774	10.830444	0	1.052816	-23.539802	220	18	1.1	0.9;
	2109	1	529.9	23.3	0	57.66	0	1.026732	-43.345577	380	6	1.1	0.9;
	2110	1	-0	-0	0.136871	2.413044	0	1.068667	-14.718056	220	19	1.1	0.9;
	2111	1	12.81	4.4	0	0.47	0	1.007865	-19.47523	110	18	1.1	0.9;
	2112	1	152.5	-0.1	0	2.66	0	1.010405	6.151511	220	3	1.1	0.9;
	2113	2	0	0	0	0	0	1.151426	-21.184015	380	9	1.3	0.7;
	2114	1	5.71	2	0	1.03	0	0.962004	-23.15021	110	18	1.1	0.9;
	2115	1	11	3.15	0	0.63	0	0.947883	-23.815387	110	18	1.1	0.9;
	2116	1	53.36	7.1	0	6.02	0	1.020733	-46.651247	220	4	1.1	0.9;
	2117	1	-0.2	-1	0	0	0	1.028711	-1.85876	150	8	1.1	0.9;
	2118	1	-0	0	0	3.61	0	1.009604	-37.776887	220	1	1.1	0.9;
	2119	1	4	0.6	0	3.46	0	1.020215	-4.334775	154	1	1.1	0.9;
	2120	1	-0.19	-3.76	0	0	0	1.023401	9.328488	150	8	1.1	0.9;
	2121	1	133.2	24.8	0	74.43	0	1.037057	-24.297839	380	6	1.1	0.9;
	2122	1	0	0	0	2.33	0	1.025212	6.076546	154	1	1.1	0.9;
	2123	1	19.82	6	0	1.68	0	0.984147	-32.500513	110	21	1.1	0.9;
	2124	1	8.07	1.23	0	3.91	0	1.022575	63.259259	150	1	1.1	0.9;
	2125	1	0	0	0	1.81	0	0.988682	-30.998911	150	9	1.1	0.9;
	2126	1	92.3	14.6	0	1.14	0	1.017186	-53.006723	220	4	1.1	0.9;
	2127	1	30.33	11	0	2.29	0	1.002297	-30.614179	110	19	1.1	0.9;
	2128	1	79.09	18.7	0	10.8	0	1.032035	-23.414596	220	5	1.1	0.9;
	2129	1	0	0	0	56.686515	0	1.021988	-8.169627	380	5	1.1	0.9;
	2130	1	34.34	12.5	0	0.23	0	1.019587	-28.587769	110	19	1.1	0.9;
	2131	1	16.22	4	0	0	0	1.089688	-27.991218	150	9	1.1	0.9;
	2132	1	107.3	14.2	0	20.23	0	1.037463	-15.904235	220	5	1.1	0.9;
	2133	1	-0	0	0	11.63	0	1.062625	52.356991	380	3	1.1	0.9;
	2134	1	-0	-0	0	25.72	0	1.03408	-3.064497	220	1	1.1	0.9;
	2135	1	78.69	56.7	0	10.47	0	1.017173	-29.076524	220	6	1.1	0.9;
	2136	1	23.2	3.3	0	3	0	1.019583	4.657853	154	1	1.1	0.9;
	2137	1	141.5	20.1	0	5.35	0	1.019436	-7.612871	154	1	1.1	0.9;
	2138	1	137.8	64.5	0	18.2	0	1.016838	-11.821955	380	3	1.1	0.9;
	2139	1	0	0	0	9.51	0	1.038846	-49.571529	220	6	1.1	0.9;
	2140	1	0	0	0	1.2	0	0.986532	-53.437129	150	10	1.1	0.9;
	2141	1	29.4	4.2	0	5.52	0	1.02688	0.763636	154	1	1.1	0.9;
	2142	1	0	-0	0	7.63	0	1.047615	-7.960845	220	5	1.1	0.9;
	2143	1	129.5	38.3	0	5.42	0	0.987085	-43.801071	220	4	1.1	0.9;
	2144	1	-0	0	0	0.44	0	1.035938	-21.477095	220	20	1.1	0.9;
	2145	1	130.9	37.8	0	0.27	0	1.017308	-14.54311	220	3	1.1	0.9;
	2146	1	10.81	1.3	0	2.23	0	0.983114	-33.088115	110	18	1.1	0.9;
	2147	1	6.11	-0	0	0.92	0	1.006915	-48.167882	220	4	1.1	0.9;
	2148	1	-0	-0	0	0.3	0	1.042373	-29.571563	150	9	1.1	0.9;
	2149	1	17.62	5.8	0	0.48	0	0.94437	-35.315794	110	21	1.1	0.9;
	2150	2	0	0	0	0	0	1.061614	22.717219	380	8	1.1	0.9;
	2151	1	-0	0	0	116.62	0	1.005592	-0.822806	400	1	1.1	0.9;
	2152	1	-0	0	0.181549	4.450889	0	1.058488	-23.628759	220	18	1.1	0.9;
	2153	2	0	0	0	0	0	1.065758	1.843264	380	14	1.1	0.9;
	2154	1	0	0	0.29	13.47	0	1.045554	10.761812	380	2	1.1	0.9;
	2155	1	63.67	20.2	0	6.13	0	1.033134	-6.433546	220	5	1.1	0.9;
	2156	1	6.01	2.2	0	1.91	0	1.011365	-13.272881	110	18	1.1	0.9;
	2157	1	0	-0	0	3.6	0	1.024695	3.308464	380	3	1.1	0.9;
	2158	1	-0	-0	0	1.12	0	1.000737	-27.441243	150	9	1.1	0.9;
	2159	1	51.96	32.2	0	0.84	0	0.823843	-41.5372	150	9	1.3	0.7;
	2160	1	304.2	114.1	0	26.84	0	1.027602	-45.178171	380	6	1.1	0.9;
	2161	1	35.66	-9.39	0	7.89	0	1.045411	-29.058113	220	5	1.1	0.9;
	2162	1	12.21	4.4	0	1.03	0	1.074568	-22.625366	110	15	1.1	0.9;
	2163	2	0	0	0	0	0	1.042967	-20.417491	110	19	1.1	0.9;
	2164	1	15.82	9.8	0	4.86	0	1.044395	-29.479203	150	9	1.1	0.9;
	2165	1	0	-0	0	10.34	0	1.027513	5.388751	400	1	1.1	0.9;
	2166	1	20.92	7.7	0	1.21	0	1.054425	-14.69511	220	5	1.1	0.9;
	2167	2	0	0	0	0	0	1.027081	-25.732231	110	19	1.1	0.9;
	2168	1	8.6	-1.48	0	0	0	1.019587	-13.276745	150	2	1.1	0.9;
	2169	1	-0	-0	0	1.33	0	1.035649	-24.356152	220	1	1.1	0.9;
	2170	1	0	-0	0.090774	6.319044	0	1.025415	-27.249576	220	18	1.1	0.9;
	2171	1	37.6	5.4	0	2.24	0	0.99897	-24.971014	154	1	1.1	0.9;
	2172	1	114.2	34.4	0	10.29	0	1.027481	-50.091836	220	4	1.1	0.9;
	2173	1	396.7	62.8	0	0	0	1.063497	-27.847436	120	20	1.1	0.9;
	2174	1	0	0	0	61.795782	0	1.00905	-13.428183	380	5	1.1	0.9;
	2175	1	0	0	0.153288	-0.361586	0	1.048436	-20.851948	220	3	1.1	0.9;
	2176	2	0	0	0	0	0	1.0196	29.887392	380	3	1.1	0.9;
	2177	2	0	0	0	0	0	1.068942	-28.881725	220	5	1.1	0.9;
	2178	1	40.65	13.2	0	1.27	0	1.04338	-31.73289	110	22	1.1	0.9;
	2179	1	0	0	0	0.78	0	1.042701	-25.03242	150	9	1.1	0.9;
	2180	1	0	-0	0	2.24	0	1.020728	-1.17428	150	8	1.1	0.9;
	2181	2	0	0	0	0	0	1.00154	23.885802	400	1	1.1	0.9;
	2182	1	25	3.6	0	13	0	1.015192	-3.517904	154	1	1.1	0.9;
	2183	1	0	0	0	32.849821	0	0.9922	-16.467929	380	5	1.1	0.9;
	2184	1	0	-0	0	1.29	0	0.997829	-32.179648	150	9	1.1	0.9;
	2185	2	0	0	0	0	0	1.056295	-15.15523	220	18	1.1	0.9;
	2186	1	16.7	2.4	0	15.54	0	1.022218	2.694959	154	1	1.1	0.9;
	2187	2	0	0	0	0	0	1.036364	-12.86827	380	6	1.1	0.9;
	2188	1	25.23	11.9	0	18.72	0	1.044948	-50.067985	380	4	1.1	0.9;
	2189	1	-0	0	0	0.18	0	1.046449	-6.348503	220	5	1.1	0.9;
	2190	1	-14.27	-6.6	0	0.91	0	1.064499	21.115751	380	3	1.1	0.9;
	2191	1	0	0	0	9.9	0	1.022139	-25.253741	220	7	1.1	0.9;
	2192	1	-19.79	-21.11	0.159201	65.262484	0	1.096122	33.377912	380	3	1.1	0.9;
	2193	2	0	0	0	0	0	1.009622	-47.671613	220	4	1.1	0.9;
	2194	1	0	0	0	23.66	0	0.992066	-32.011375	150	9	1.1	0.9;
	2195	1	55.46	18.5	0	2.23	0	0.997154	-32.953214	110	21	1.1	0.9;
	2196	1	-0	0	0	5.14	0	1.036544	39.19365	220	3	1.1	0.9;
	2197	2	0	0	0	0	0	1.031798	-6.111419	380	5	1.1	0.9;
	2198	1	34.14	11.4	0	0.8	0	0.961565	-30.257046	110	18	1.1	0.9;
	2199	1	297.2	93.2	0	23.74	0	1.00662	-50.639854	380	6	1.1	0.9;
	2200	1	137.3	-2.8	0	16.39	0	1.02289	-27.575099	220	6	1.1	0.9;
	2201	1	29.03	9.6	0	0.53	0	0.976169	-32.476732	110	21	1.1	0.9;
	2202	1	213	39.6	0	20.63	0	1.021647	-49.803988	380	4	1.1	0.9;
	2203	1	0	-0	0.316851	27.238806	0	1.007242	-20.041451	380	11	1.1	0.9;
	2204	1	-0	0	0	-0.46	0	1.021705	-46.279718	380	6	1.1	0.9;
	2205	1	11.01	3.6	0	0.84	0	1.005586	-36.686458	110	12	1.1	0.9;
	2206	2	0	0	0	0	0	1.018916	-34.656516	150	10	1.1	0.9;
	2207	2	0	0	0	0	0	1.006198	-30.213114	110	18	1.1	0.9;
	2208	1	74.98	26.9	0	2.96	0	1.042053	-13.792922	220	5	1.1	0.9;
	2209	1	6.01	4.1	0	0.19	0	0.967888	-28.721386	110	18	1.1	0.9;
	2210	1	-0	-7.5	0	0	0	1.001006	-36.77408	150	9	1.1	0.9;
	2211	1	44.55	16.1	0	1.18	0	0.944699	-22.597549	110	19	1.1	0.9;
	2212	1	60.07	45	0	2.53	0	0.976643	11.936659	220	3	1.1	0.9;
	2213	1	0	-0	0	0.3	0	0.99713	-32.416337	150	9	1.1	0.9;
	2214	1	54.32	7.98	0	13.58	0	1.012833	61.836316	150	1	1.1	0.9;
	2215	1	23.7	3.4	0	2.51	0	1.016427	-7.76953	154	1	1.1	0.9;
	2216	1	-0	-0	0	47.421898	0	1.086968	7.721123	380	13	1.3	0.7;
	2217	1	0	0	0	0.41	0	0.954628	-33.459831	150	9	1.1	0.9;
	2218	1	-0	0	0	1.6	0	1.010599	-27.081656	110	21	1.1	0.9;
	2219	1	20.32	6.7	0	0.27	0	1.01223	-36.67689	110	12	1.1	0.9;
	2220	1	0	-0	0	0.73	0	1.066623	-28.071251	150	9	1.1	0.9;
	2221	1	14.26	-1.13	0	0.65	0	0.997341	-27.133263	110	18	1.1	0.9;
	2222	1	0	0	0	2.67	0	1.075928	43.452293	380	3	1.1	0.9;
	2223	1	142.8	19.5	0	10.53	0	1.026839	-2.340482	220	3	1.1	0.9;
	2224	1	78.79	27.4	0	2.37	0	0.999481	-50.559153	220	6	1.1	0.9;
	2225	2	0	0	0	0	0	1.025903	-44.775248	220	4	1.1	0.9;
	2226	1	145.7	83.6	0.135375	53.822175	0	1.06687	-16.553003	380	18	1.1	0.9;
	2227	1	0	0	0	0	0	1.003174	-36.685975	150	9	1.1	0.9;
	2228	1	143.1	18.9	0	3.42	0	1.011206	22.467426	110	8	1.1	0.9;
	2229	1	0	-0	0	2.91	0	1.065909	-13.135319	220	5	1.1	0.9;
	2230	1	0	-0	0	1.71	0	1.032021	-22.83036	220	5	1.1	0.9;
	2231	1	0	0	0	0.45	0	1.046057	1.068739	220	5	1.1	0.9;
	2232	1	19.52	11	0	1.11	0	1.023898	-33.498353	220	6	1.1	0.9;
	2233	1	269.9	12.4	0	4.09	0	0.985171	-56.435749	150	10	1.1	0.9;
	2234	1	63.67	39.5	0	0.38	0	1.052737	-29.321529	150	9	1.1	0.9;
	2235	1	193.1	69.9	0	28.91	0	1.008042	-45.226775	380	4	1.1	0.9;
	2236	2	0	0	0	0	0	1.020813	-42.798155	220	4	1.1	0.9;
	2237	2	0	0	0	0	0	1.034951	49.087288	220	3	1.1	0.9;
	2238	1	0	-0	0	0.57	0	0.982791	-16.125162	110	18	1.1	0.9;
	2239	1	-3.51	-0.52	0	0.19	0	1.01042	-23.606	110	18	1.1	0.9;
	2240	1	66.87	21.8	0	12.3	0	1.013461	3.23261	150	8	1.1	0.9;
	2241	2	0	0	0	0	0	1.005014	11.382833	400	1	1.1	0.9;
	2242	1	-0	-0	0	2.76	0	1.021353	-19.301242	220	5	1.1	0.9;
	2243	2	0	0	0	0	0	1.029391	4.137431	154	1	1.1	0.9;
	2244	1	0	-0	0	0.24	0	1.037731	-23.018246	150	9	1.1	0.9;
	2245	2	0	0	0	0	0	1.010255	7.628832	154	1	1.1	0.9;
	2246	1	7.61	2.5	0	0.39	0	1.015845	-15.009168	110	18	1.1	0.9;
	2247	1	47.46	11.54	0	14.15	0	1.014007	62.665561	150	1	1.1	0.9;
	2248	1	110.3	28.7	0	0.65	0	1.011098	-46.870246	220	4	1.1	0.9;
	2249	1	1.8	0.7	0	0.22	0	1.037294	-28.117592	110	19	1.1	0.9;
	2250	2	0	0	0	0	0	1.024801	-7.262701	220	13	1.1	0.9;
	2251	1	25.33	12.7	0.181549	6.830889	0	1.043241	-24.893094	220	18	1.1	0.9;
	2252	1	100.7	-4.2	0	1.57	0	1.06522	-48.109558	220	5	1.1	0.9;
	2253	2	0	0	0	0	0	1.147706	-23.576587	380	21	1.3	0.7;
	2254	1	0	-8	0	0	0	0.992038	-34.266143	150	9	1.1	0.9;
	2255	1	76.6	10.9	0	12.63	0	1.012904	10.266561	154	1	1.1	0.9;
	2256	1	23.03	8.3	0	1.14	0	0.977764	-21.60905	110	19	1.1	0.9;
	2257	1	97	18.7	0	21.46	0	1.023136	-25.121398	220	7	1.1	0.9;
	2258	1	-14.83	-3.56	0	0.06	0	1.023856	-15.0105	110	18	1.1	0.9;
	2259	1	5.71	1.6	0	0.48	0	0.986832	-28.686147	110	18	1.1	0.9;
	2260	1	305.6	207.3	0	52.37	0	1.024811	5.411327	380	3	1.1	0.9;
	2261	1	0	-0	0	20.5	0	1.062649	-41.012513	380	4	1.1	0.9;
	2262	1	-0	-0	0	1.55	0	0.973298	-32.13493	150	9	1.1	0.9;
	2263	1	1.9	0.7	0	0.42	0	1.041323	-20.80661	110	19	1.1	0.9;
	2264	1	117.6	16.7	0	10.55	0	1.010553	-23.522211	154	1	1.1	0.9;
	2265	1	27.9	4	0	2.16	0	1.036318	-7.096598	154	1	1.1	0.9;
	2266	2	0	0	0	0	0	1.013425	-21.770998	154	1	1.1	0.9;
	2267	2	0	0	0	0	0	1.003207	11.579452	380	2	1.1	0.9;
	2268	1	0	-0	0.183167	22.675025	0	0.992327	-5.514567	380	2	1.1	0.9;
	2269	1	-0	0	0	2.59	0	1.05231	-20.182035	220	1	1.1	0.9;
	2270	1	0	0	0	0.33	0	1.063847	-21.310503	150	9	1.1	0.9;
	2271	1	0	-0.01	0	5.48	0	1.021336	10.866719	150	8	1.1	0.9;
	2272	1	0	0	0	26.22	0	1.060674	-12.270379	380	1	1.1	0.9;
	2273	1	141.3	32.4	0	7.29	0	1.068632	-26.737078	220	5	1.1	0.9;
	2274	1	268.56	29.72	0	6.5	0	1.004956	-10.486485	220	13	1.1	0.9;
	2275	1	91.9	5.5	0	7.02	0	1.009846	-24.207712	110	18	1.1	0.9;
	2276	2	0	0	0	0	0	1.062473	-24.368144	220	5	1.1	0.9;
	2277	1	-0	0	0	15.57	0	1.043887	46.693183	380	3	1.1	0.9;
	2278	2	0	0	0	0	0	1.030091	17.647967	154	1	1.1	0.9;
	2279	1	73.18	22.2	0	0.32	0	1.014308	-22.801606	110	19	1.1	0.9;
	2280	1	11.61	5.8	0	0.43	0	1.030143	-26.749909	110	15	1.1	0.9;
	2281	1	30.6	4.3	0	7.74	0	1.015438	-8.763494	154	1	1.1	0.9;
	2282	1	0	0	0	0.75	0	1.021275	-38.931447	110	1	1.1	0.9;
	2283	1	1.4	0.5	0	0.63	0	1.0389	-23.599279	110	19	1.1	0.9;
	2284	2	0	0	0	0	0	1.030581	-10.972604	220	13	1.1	0.9;
	2285	1	-0	0	0	3.82	0	1.019024	63.028088	150	1	1.1	0.9;
	2286	1	208.88	60.8	0	6.68	0	1.019611	-9.10866	220	5	1.1	0.9;
	2287	2	0	0	0	0	0	1.034619	-5.128401	154	1	1.1	0.9;
	2288	1	170.9	39	0	11.25	0	1.032863	-8.632598	220	5	1.1	0.9;
	2289	2	0	0	0	0	0	1.036649	-10.586065	154	1	1.1	0.9;
	2290	1	-0	-0	0	15.02	0	1.047189	-14.164277	220	7	1.1	0.9;
	2291	2	0	0	0	0	0	1.063998	-11.16215	220	5	1.1	0.9;
	2292	1	13.04	7.4	0	5.33	0	1.024081	-30.555303	150	9	1.1	0.9;
	2293	1	-0	0	0	1.32	0	1.034016	-29.914798	150	9	1.1	0.9;
	2294	1	71.88	51.3	0	24.75	0	1.056839	-13.846532	380	16	1.1	0.9;
	2295	1	15.02	5	0	0.39	0	1.017204	-27.379016	110	21	1.1	0.9;
	2296	1	30.83	4.4	0	0.46	0	1.002026	-44.832563	110	22	1.1	0.9;
	2297	1	93.2	13.6	0	14.9	0	1.024492	-54.583126	220	4	1.1	0.9;
	2298	1	162.7	23.2	0	1.7	0	1.009365	-25.785602	154	1	1.1	0.9;
	2299	1	402.1	110.2	0	0	0	1.031467	-11.576168	150	2	1.1	0.9;
	2300	1	-0	0	0	13.21	0	1.030534	-22.214833	220	3	1.1	0.9;
	2301	1	82.39	12.3	0	9.02	0	1.023834	-54.421813	220	4	1.1	0.9;
	2302	1	628.5	91.8	0	19.83	0	1.004747	-50.97031	380	6	1.1	0.9;
	2303	1	0	-0	0	7.94	0	1.035492	-16.004196	220	5	1.1	0.9;
	2304	1	88.7	-0.1	0	2.83	0	1.051368	-26.192039	220	5	1.1	0.9;
	2305	1	21.22	8.3	0	19.1	0	1.027261	-0.244746	150	8	1.1	0.9;
	2306	1	-27.25	-9.38	0	18.78	0	1.062445	43.303305	380	3	1.1	0.9;
	2307	1	45.75	3.3	0	0.37	0	1.051906	-28.111468	110	3	1.1	0.9;
	2308	1	32.04	6.6	0	0.44	0	1.03359	-24.440454	220	5	1.1	0.9;
	2309	1	-1.8	-12.01	0.294347	159.376332	0	1.068248	23.198789	380	3	1.1	0.9;
	2310	2	0	0	0	0	0	1.013414	-21.713391	220	20	1.1	0.9;
	2311	1	8.41	3	0	1.2	0	0.992324	-29.260501	110	18	1.1	0.9;
	2312	1	0	0	0	0	0	1.04204	-29.563033	150	9	1.1	0.9;
	2313	1	0	-0	0	0.41	0	1.043224	-21.560168	220	5	1.1	0.9;
	2314	1	110.6	20.4	0	26.9	0	1.012324	-36.615338	220	6	1.1	0.9;
	2315	1	13.9	1.9	0	0.03	0	1.027446	2.280769	154	1	1.1	0.9;
	2316	1	264	40.1	0	9.489841	0	1.025936	-35.138462	380	6	1.1	0.9;
	2317	1	0	0	0	20.53	0	1.017334	-36.081183	220	6	1.1	0.9;
	2318	1	12.51	3.3	0	1.7	0	0.977049	-31.54394	110	18	1.1	0.9;
	2319	1	39.64	20.9	0	0.65	0	1.050919	-9.227171	220	5	1.1	0.9;
	2320	2	0	0	0	0	0	1.025327	-45.259588	220	6	1.1	0.9;
	2321	1	8.41	2.8	0	0.67	0	1.020279	-38.57031	110	12	1.1	0.9;
	2322	1	0	0	0	0	0	1.011753	63.312472	150	1	1.1	0.9;
	2323	1	-0	0	0	3.3	0	0.989214	-22.621975	220	1	1.1	0.9;
	2324	1	0	-0	0	4.16	0	1.029735	-5.984451	154	1	1.1	0.9;
	2325	1	0	-0	0	101.48	0	0.996292	-0.946019	400	1	1.1	0.9;
	2326	1	24.23	6.6	0	1.05	0	1.039189	-39.946279	110	22	1.1	0.9;
	2327	1	114.5	25.2	0	17.57	0	1.042809	-25.466491	220	5	1.1	0.9;
	2328	1	-0	0	0	0.4	0	1.01174	-14.547426	220	5	1.1	0.9;
	2329	1	0	-0	0	28.53	0	1.048968	9.559465	380	1	1.1	0.9;
	2330	1	-0.2	-1.33	0	0	0	1.009007	0.655964	150	8	1.1	0.9;
	2331	1	-0.15	-1.4	0	0.01	0	1.011436	18.739489	110	8	1.1	0.9;
	2332	1	-0	-0	0	0.5	0	1.029158	-31.509433	380	6	1.1	0.9;
	2333	1	27.93	6.4	0	0.89	0	0.968039	-29.31337	110	18	1.1	0.9;
	2334	1	0	0	0	0.35	0	1.074572	-22.21578	150	9	1.1	0.9;
	2335	2	0	0	0	0	0	1.044074	6.10879	150	8	1.1	0.9;
	2336	1	90.5	12.8	0	0.78	0	1.035535	-22.479197	154	1	1.1	0.9;
	2337	1	103.4	27	0	1.41	0	1.042097	-30.086837	220	5	1.1	0.9;
	2338	1	12.71	4.9	0	0.83	0	0.952295	-31.710735	110	18	1.1	0.9;
	2339	1	24.33	15	0	0.31	0	1.009593	-24.521036	150	9	1.1	0.9;
	2340	1	96.91	13.9	0	3.98	0	1.033768	-12.546175	220	5	1.1	0.9;
	2341	1	-0	0	0	3.28	0	1.057467	-26.432568	220	5	1.1	0.9;
	2342	1	-0	-0	0	12.77	0	1.059396	55.280716	380	3	1.1	0.9;
	2343	1	33.04	5.6	0	14.55	0	1.020785	-42.957173	220	4	1.1	0.9;
	2344	1	-0	0	0	426.57	0	1.054222	-15.775316	380	6	1.1	0.9;
	2345	1	-0	0	0	2.92	0	1.027493	-26.383948	220	21	1.1	0.9;
	2346	1	53.3	7.6	0	10.7	0	1.022957	8.75295	154	1	1.1	0.9;
	2347	1	0	0	0	0.5	0	0.998838	-29.376014	150	9	1.1	0.9;
	2348	1	33.24	11.1	0	0.62	0	0.968417	-36.688676	110	21	1.1	0.9;
	2349	1	0	-0	0.12	1.55	0	1.05949	-27.313076	220	21	1.1	0.9;
	2350	1	-0	0	0	7.03	0	1.01017	15.884565	150	8	1.1	0.9;
	2351	2	0	0	0	0	0	1.019563	62.181667	150	1	1.1	0.9;
	2352	1	4.91	1.4	0	0.49	0	0.958373	-29.424716	110	18	1.1	0.9;
	2353	2	0	0	0	0	0	1.010677	-8.11857	154	1	1.1	0.9;
	2354	1	75.78	46.9	0	3.07	0	1.014721	-32.352939	150	9	1.1	0.9;
	2355	1	54.46	11.9	0	3.93	0	1.015354	-51.394188	220	4	1.1	0.9;
	2356	1	0	0	0	0	0	0.977313	-30.412799	150	9	1.1	0.9;
	2357	2	0	0	0	0	0	1.072612	-3.82962	380	14	1.1	0.9;
	2358	1	52.96	8.2	0	5.68	0	1.023813	-54.443138	220	4	1.1	0.9;
	2359	2	0	0	0	0	0	1.021762	-10.091408	380	5	1.1	0.9;
	2360	1	215.5	31.3	0	14.23	0	1.033789	-21.574385	220	5	1.1	0.9;
	2361	1	16.9	3.85	0	25.09	0	1.035677	-29.682934	220	5	1.1	0.9;
	2362	2	0	0	0	0	0	1.038978	-11.117228	380	6	1.1	0.9;
	2363	1	0	-0	0	5.71	0	0.99184	-22.926815	220	1	1.1	0.9;
	2364	1	0	0	0	2.59	0	1.01739	8.973438	220	3	1.1	0.9;
	2365	1	37.94	15.4	0	117.386167	0	1.031058	-22.416671	380	5	1.1	0.9;
	2366	1	-0	0	0	4.24	0	1.003879	-49.924326	220	6	1.1	0.9;
	2367	1	-15.92	5.2	0	0.3	0	1.053354	-28.800199	150	9	1.1	0.9;
	2368	1	178.5	65.8	0	93.42	0	1.059005	-27.635733	380	20	1.1	0.9;
	2369	2	0	0	0	0	0	1.055918	-17.345679	220	19	1.1	0.9;
	2370	1	-0	-0	0	2.53	0	1.034832	-21.563669	154	1	1.1	0.9;
	2371	1	-0	0	0	28.65	0	1.032872	-43.818915	380	1	1.1	0.9;
	2372	1	0	-0	0	14.490904	0	0.986562	-17.792777	380	5	1.1	0.9;
	2373	1	153.3	21.8	0	10	0	1.017902	-24.439685	154	1	1.1	0.9;
	2374	1	0	0	0	0	0	1.003311	-31.350195	150	9	1.1	0.9;
	2375	1	131.1	18.7	0	6.53	0	1.009953	-25.950012	154	1	1.1	0.9;
	2376	1	17.12	3.8	0	1	0	0.993399	-27.237978	110	18	1.1	0.9;
	2377	1	101.7	17.6	0	13.14	0	1.036778	-15.052217	220	5	1.1	0.9;
	2378	1	-41.25	13.5	0	0.48	0	1.03349	-26.129447	150	9	1.1	0.9;
	2379	1	-5.51	1.8	0	0.91	0	0.950151	-29.473578	150	9	1.1	0.9;
	2380	1	-0	-0	0	5.43	0	1.031315	-27.715521	220	1	1.1	0.9;
	2381	1	32.94	13	0	4.04	0	1.038031	4.576831	150	8	1.1	0.9;
	2382	1	0	0	0	0.45	0	1.009187	-34.100259	150	9	1.1	0.9;
	2383	1	52.26	44.2	0.181549	9.190889	0	1.054816	-16.31655	220	18	1.1	0.9;
	2384	1	0	0	0	9.85	0	1.036527	-37.386634	220	6	1.1	0.9;
	2385	1	122.1	16.7	0	27.05	0	1.034739	-8.954845	380	3	1.1	0.9;
	2386	1	24.13	8.7	0.083411	0.134843	0	1.043705	-21.418464	220	19	1.1	0.9;
	2387	1	0.1	1	0	0.06	0	0.983017	9.191644	220	3	1.1	0.9;
	2388	1	57.56	24	0	1.01	0	1.007494	-37.291444	220	6	1.1	0.9;
	2389	1	-0	0	0	4.02	0	1.043727	-30.375841	220	22	1.1	0.9;
	2390	1	-0	0	0	0.79	0	0.995225	9.248959	220	3	1.1	0.9;
	2391	1	20.12	12.5	0	2.83	0	0.97801	-29.767009	150	9	1.1	0.9;
	2392	1	173.7	23	0	18.4	0	1.021881	-25.503771	220	7	1.1	0.9;
	2393	1	20.82	0	0	0.13	0	1.064029	7.685859	220	5	1.1	0.9;
	2394	2	0	0	0	0	0	1.065017	54.664187	380	3	1.1	0.9;
	2395	1	80.29	28.4	0	8.67	0	1.020043	-25.439807	220	20	1.1	0.9;
	2396	1	72.4	10.3	0	8.67	0	1.034233	5.35921	154	1	1.1	0.9;
	2397	1	-0	-0	0	0.52	0	0.996094	-31.3955	150	9	1.1	0.9;
	2398	2	0	0	0	0	0	1.147454	-23.601459	380	21	1.3	0.7;
	2399	1	69.38	43	0	3.22	0	0.971529	-34.186542	150	9	1.1	0.9;
	2400	1	0	-0	0	3.99	0	1.035877	12.455059	154	1	1.1	0.9;
	2401	1	-0	-0	0	0.16	0	0.962213	-32.513726	150	9	1.1	0.9;
	2402	2	0	0	0	0	0	1.076661	-19.339734	220	21	1.1	0.9;
	2403	1	37.24	5.3	0	14.78	0	1.022768	-48.115052	220	4	1.1	0.9;
	2404	1	0	-0	0	0.96	0	0.927417	-34.766043	110	18	1.1	0.9;
	2405	1	9.7	1.4	0	5.11	0	1.026531	-3.69527	154	1	1.1	0.9;
	2406	1	0	0	0	1.69	0	1.04431	-24.755069	220	5	1.1	0.9;
	2407	1	117.13	52.4	0	7.51	0	0.993264	-39.922258	150	10	1.1	0.9;
	2408	1	30.6	4.3	0	2.82	0	1.005765	-26.282357	154	1	1.1	0.9;
	2409	1	64.8	99.1	0	4.28	0	1.055461	10.594296	380	3	1.1	0.9;
	2410	1	3.9	-35.8	0	4.71	0	1.024241	-37.185757	220	10	1.1	0.9;
	2411	1	0	-0	0	9.14	0	1.01792	62.34921	150	1	1.1	0.9;
	2412	2	0	0	0	0	0	1.013238	-43.156083	220	4	1.1	0.9;
	2413	2	0	0	0	0	0	0.996424	9.258591	220	3	1.1	0.9;
	2414	1	30.6	4.3	0	7.3	0	1.022771	14.280372	154	1	1.1	0.9;
	2415	1	5.51	2	0	1.3	0	1.017323	-27.549298	110	19	1.1	0.9;
	2416	1	14.32	5.6	0	88.82	0	1.0474	-49.767877	380	4	1.1	0.9;
	2417	1	-0.54	-1	0	0.81	0	1.054261	-20.876656	110	15	1.1	0.9;
	2418	1	15.9	116.3	0	5.1	0	1.058347	27.795826	220	3	1.1	0.9;
	2419	1	-0	-0	0.258771	12.056323	0	0.994664	-13.92369	220	19	1.1	0.9;
	2420	1	3.4	123.8	0	64.94	0	1.020157	-18.290136	380	3	1.1	0.9;
	2421	2	0	0	0	0	0	1.043537	-24.060562	220	5	1.1	0.9;
	2422	1	39.34	24.3	0	3.06	0	1.012924	-30.696006	150	9	1.1	0.9;
	2423	1	523.3	137.2	0	65.61	0	1.01892	-46.711961	380	6	1.1	0.9;
	2424	1	44.55	16.8	0	2.31	0	1.030907	-22.76203	220	5	1.1	0.9;
	2425	2	0	0	0	0	0	1.037815	-10.136554	220	5	1.1	0.9;
	2426	2	0	0	0	0	0	1.045936	-13.54825	220	5	1.1	0.9;
	2427	1	0	-0	0.27075	102.13435	0	1.082868	-21.362734	380	18	1.3	0.7;
	2428	1	-0	-0	0	50	0	1.024706	0.110381	150	8	1.1	0.9;
	2429	2	0	0	0	0	0	1.063143	12.917367	380	3	1.1	0.9;
	2430	1	-0	0	0	6.21	0	1.039783	-12.117188	220	5	1.1	0.9;
	2431	1	9.61	1.8	0	2.84	0	0.992323	-21.573229	110	18	1.1	0.9;
	2432	1	291.2	35.4	0	21.87	0	1.021679	-32.538904	220	5	1.1	0.9;
	2433	1	0	-0	0	2.31	0	1.015149	6.190111	220	3	1.1	0.9;
	2434	1	0	0	0	0.12	0	1.055535	-27.608504	150	9	1.1	0.9;
	2435	1	-0	-0	0	16.92	0	1.144156	-23.633231	380	1	1.3	0.7;
	2436	1	0	0	0	91.66	0	0.998471	-3.00089	400	1	1.1	0.9;
	2437	1	-0	-8.6	0	297.94	0	1.057795	-42.136814	380	4	1.1	0.9;
	2438	1	-0	0	0	0.64	0	1.043195	-21.559867	220	5	1.1	0.9;
	2439	1	-0	-0	0	8.66	0	1.005857	23.692042	220	3	1.1	0.9;
	2440	1	0	-0	0	10.21	0	1.02046	3.58516	380	3	1.1	0.9;
	2441	1	0	-0	0.242534	1.143584	0	0.994601	-5.150369	380	2	1.1	0.9;
	2442	1	2.43	0.49	0	0.17	0	0.998662	-26.864718	110	18	1.1	0.9;
	2443	1	45.6	142	0	3.17	0	1.009217	8.41599	220	3	1.1	0.9;
	2444	1	0	-0	0	18.25	0	1.026884	7.398524	380	3	1.1	0.9;
	2445	1	-0	-0	0	0.19	0	0.962275	-31.64821	150	9	1.1	0.9;
	2446	2	0	0	0	0	0	1.048971	7.52887	380	5	1.1	0.9;
	2447	2	0	0	0	0	0	1.031874	-11.988534	154	1	1.1	0.9;
	2448	1	34.24	16.5	0	0	0	1.040932	25.764861	220	8	1.1	0.9;
	2449	1	301.9	-103.2	0	0	0	1.062433	-6.280372	150	2	1.1	0.9;
	2450	1	20.3	3.7	0	0.06	0	1.021508	3.940089	154	1	1.1	0.9;
	2451	1	-0	-0	0	3.62	0	1.040158	-21.474336	150	9	1.1	0.9;
	2452	2	0	0	0	0	0	1.026174	16.830485	154	1	1.1	0.9;
	2453	1	98.36	-0.47	0	5.44	0	1.040903	-2.446956	220	13	1.1	0.9;
	2454	1	-0	-0	0	1.55	0	1.02755	-30.467201	150	9	1.1	0.9;
	2455	1	0	0	0	63.59	0	0.999933	-3.616676	400	1	1.1	0.9;
	2456	1	262.5	18.9	0	39.85	0	1.032684	-15.137228	380	6	1.1	0.9;
	2457	1	30.73	9.6	0	2.75	0	1.037629	-24.947308	220	5	1.1	0.9;
	2458	1	132	26	0	27.18	0	1.00176	-13.726748	380	5	1.1	0.9;
	2459	1	-0	0	0	145.98	0	1.04316	-42.879326	380	4	1.1	0.9;
	2460	2	0	0	0	0	0	0.979599	-40.295757	220	10	1.1	0.9;
	2461	1	3.3	1.9	0	0.25	0	1.042742	-8.232623	110	19	1.1	0.9;
	2462	1	-5.22	-3.77	0	0.91	0	1.014526	-26.776272	110	18	1.1	0.9;
	2463	1	0	-0	0	32.55	0	1.086073	45.377841	380	3	1.1	0.9;
	2464	1	1.5	0.4	0	0.89	0	0.978522	-33.017724	110	18	1.1	0.9;
	2465	1	314.65	87.51	0	42.61	0	1.064101	1.055741	380	14	1.1	0.9;
	2466	1	0	0.5	0	1.82	0	0.936778	-32.12342	110	18	1.1	0.9;
	2467	1	0	-0	0	11.65	0	1.011431	-14.414053	220	5	1.1	0.9;
	2468	2	0	0	0	0	0	1.046839	-6.281333	220	5	1.1	0.9;
	2469	1	-0	0	0	1.12	0	1.001957	-30.409265	150	9	1.1	0.9;
	2470	1	0	-0	0	1.31	0	1.020155	-25.904378	150	9	1.1	0.9;
	2471	1	103.5	21.7	0	8.88	0	1.011871	-47.766524	220	4	1.1	0.9;
	2472	1	76.58	14	0	0	0	0.98671	4.854467	220	3	1.1	0.9;
	2473	1	40.44	12.8	0	1.04	0	0.943732	-35.399912	110	21	1.1	0.9;
	2474	1	35.04	16	0	12.14	0	1.060294	-27.122887	220	23	1.1	0.9;
	2475	1	0	0	0	0.21	0	1.055302	-8.059108	220	5	1.1	0.9;
	2476	1	0	0	0	0	0	0.970883	-24.101597	150	9	1.1	0.9;
	2477	1	4.2	0.6	0	4.73	0	1.008651	-9.743001	154	1	1.1	0.9;
	2478	1	50.96	6.8	0	9.13	0	1.027168	15.022854	110	8	1.1	0.9;
	2479	1	-2.65	-1.17	0	65.673108	0	1.004264	-2.650053	380	5	1.1	0.9;
	2480	1	-0	-0	0	1.55	0	1.001341	-36.96814	220	22	1.1	0.9;
	2481	2	0	0	0	0	0	1.045831	-18.206422	220	5	1.1	0.9;
	2482	1	-0.21	-1.98	0	0	0	1.022565	8.174762	150	8	1.1	0.9;
	2483	1	26.6	3.8	0	6.43	0	1.020708	-3.051755	154	1	1.1	0.9;
	2484	1	-0	0	0	2.43	0	1.055668	-48.878125	220	6	1.1	0.9;
	2485	1	33.5	4.8	0	1.52	0	1.040915	8.404641	154	1	1.1	0.9;
	2486	1	79.09	-36.9	0	4.58	0	1.042068	11.909351	380	3	1.1	0.9;
	2487	1	0	-0	0	13.39	0	1.051406	-19.92777	220	7	1.1	0.9;
	2488	1	-0	-0	0	0	0	0.997894	-25.797116	110	18	1.1	0.9;
	2489	2	0	0	0	0	0	1.046032	1.069051	220	5	1.1	0.9;
	2490	1	42.2	6	0	5.22	0	1.012284	-1.786344	154	1	1.1	0.9;
	2491	1	48.8	6.9	0	3.23	0	1.006433	7.066289	154	1	1.1	0.9;
	2492	1	0	0	0	7.67	0	0.972676	-18.887241	220	16	1.1	0.9;
	2493	1	74.18	22.6	0	0.31	0	1.031054	-10.742367	220	3	1.1	0.9;
	2494	1	0	0	0	0	0	1.017252	0.511391	150	8	1.1	0.9;
	2495	1	0	0	0	1.2	0	1.047044	-24.959324	220	1	1.1	0.9;
	2496	1	50.56	4.6	0	1.74	0	1.009835	-61.002304	220	4	1.1	0.9;
	2497	1	0	0	0	9.36	0	0.976143	-56.682465	220	10	1.1	0.9;
	2498	1	87.7	22.2	0	7.35	0	1.039709	-16.75247	220	7	1.1	0.9;
	2499	1	-0	0	0	0.68	0	1.030521	6.209771	154	1	1.1	0.9;
	2500	1	0	-0	0	4.8	0	1.136559	-25.931448	380	23	1.3	0.7;
	2501	2	0	0	0	0	0	1.044679	-23.67977	220	7	1.1	0.9;
	2502	1	0	-0	0	84.7	0	1.065757	-39.403617	380	22	1.1	0.9;
	2503	1	-0	0	0	2.09	0	1.00763	-21.840345	220	5	1.1	0.9;
	2504	1	119.2	32.1	0	4.14	0	1.005424	-45.865158	220	4	1.1	0.9;
	2505	1	48.75	16.2	0	0.77	0	0.996704	-35.118129	110	21	1.1	0.9;
	2506	1	-14.9	-4.81	0	0.45	0	0.974779	-22.08123	110	18	1.1	0.9;
	2507	1	-0	0	0	36.17	0	1.06149	-42.008359	380	4	1.1	0.9;
	2508	1	41.8	5.9	0	4.74	0	1.005889	7.847197	154	1	1.1	0.9;
	2509	2	0	0	0	0	0	1.025789	-14.476871	220	13	1.1	0.9;
	2510	1	115.87	29.41	0	7.92	0	1.025451	-12.308601	220	5	1.1	0.9;
	2511	1	174.1	-519.7	0	54.02	0	1.045056	16.162224	380	3	1.1	0.9;
	2512	1	0	0	0	18.83	0	1.054857	-34.872387	380	12	1.1	0.9;
	2513	1	0	-0	0	29.79	0	1.038107	35.125312	380	3	1.1	0.9;
	2514	1	18.2	2.6	0	1.75	0	1.024359	-6.570675	154	1	1.1	0.9;
	2515	2	0	0	0	0	0	0.998497	-25.884845	110	18	1.1	0.9;
	2516	1	-0	-0	0	6.66	0	1.026768	-12.590765	220	3	1.1	0.9;
	2517	2	0	0	0	0	0	1.013559	-45.235352	220	4	1.1	0.9;
	2518	1	-5.29	-1.91	0	42.622979	0	0.996983	-11.616978	380	5	1.1	0.9;
	2519	2	0	0	0	0	0	1.028396	65.846601	150	1	1.1	0.9;
	2520	1	-0	-0	0.057433	3.80797	0	0.99239	-5.13982	380	2	1.1	0.9;
	2521	1	90.5	12.9	0	2.99	0	1.020279	-7.528828	154	1	1.1	0.9;
	2522	1	59.7	26.3	0	37.680934	0	1.069701	-24.040225	380	9	1.1	0.9;
	2523	1	0	0	0	7.06	0	1.039971	25.635635	220	8	1.1	0.9;
	2524	1	11.2	1.6	0	3.05	0	1.012873	-4.734816	154	1	1.1	0.9;
	2525	1	18.32	14.1	0	3.12	0	0.994561	-28.409429	150	9	1.1	0.9;
	2526	1	238.45	46.86	0	125.534496	0	1.041902	-20.143178	380	5	1.1	0.9;
	2527	1	59.77	16.7	0	0.64	0	1.041963	-4.898556	220	5	1.1	0.9;
	2528	1	-10.41	-3.3	0	4.11	0	1.037117	-22.898476	150	9	1.1	0.9;
	2529	1	0	0	0	0	0	1.020241	-29.747583	150	9	1.1	0.9;
	2530	1	77.39	20.7	0	0.46	0	1.058488	-26.840441	110	3	1.1	0.9;
	2531	1	1.3	-0	0	0.58	0	0.985798	-23.177847	110	18	1.1	0.9;
	2532	1	-0	-0	0.08367	4.954954	0	1.091238	45.829104	380	3	1.1	0.9;
	2533	1	56.7	8	0	4.42	0	1.002974	-7.531116	154	1	1.1	0.9;
	2534	2	0	0	0	0	0	1.066932	-21.855019	220	18	1.1	0.9;
	2535	1	51.56	2.1	0	0.18	0	1.033154	-3.315871	220	5	1.1	0.9;
	2536	1	15.12	9.4	0	0.69	0	1.003452	-31.424496	150	9	1.1	0.9;
	2537	1	70.28	16.2	0	0.12	0	1.001309	-26.217035	110	18	1.1	0.9;
	2538	1	118.8	39.5	0	2.61	0	1.009528	-44.137993	220	4	1.1	0.9;
	2539	2	0	0	0	0	0	1.024702	-7.289116	220	13	1.1	0.9;
	2540	1	278.35	90.4	0	12.1	0	1.000879	-32.900184	220	6	1.1	0.9;
	2541	2	0	0	0	0	0	1.02273	-21.791278	150	9	1.1	0.9;
	2542	1	72.68	45	0	8.25	0	0.991401	-30.700734	150	9	1.1	0.9;
	2543	1	38.44	13.2	0	1.73	0	1.027732	-38.707722	110	22	1.1	0.9;
	2544	1	27.33	5	0	0.15	0	1.017203	61.75354	150	1	1.1	0.9;
	2545	1	0	0	0	10.78	0	1.030444	18.762567	380	3	1.1	0.9;
	2546	1	-0	0	0	0.23	0	1.07694	-25.692549	110	23	1.1	0.9;
	2547	2	0	0	0	0	0	1.014326	-2.727604	154	1	1.1	0.9;
	2548	1	-4.88	5.5	0	1.38	0	1.0032	-36.682256	150	9	1.1	0.9;
	2549	1	0	-0	0	3.86	0	1.046283	-3.447313	220	1	1.1	0.9;
	2550	2	0	0	0	0	0	1.047876	-6.732622	220	5	1.1	0.9;
	2551	1	205.3	27.8	0	8.22	0	1.005574	-61.354086	220	4	1.1	0.9;
	2552	1	28.53	11.3	0	6.81	0	1.044585	4.016008	150	8	1.1	0.9;
	2553	1	25.77	-36	0	9.85	0	1.043674	13.172731	220	13	1.1	0.9;
	2554	1	32.54	10.6	0	1.22	0	0.933736	-36.784298	110	21	1.1	0.9;
	2555	1	-0.15	-1	0	0.38	0	1.066476	11.763351	110	8	1.1	0.9;
	2556	1	0	0	0	2.75	0	1.048219	-19.794039	220	7	1.1	0.9;
	2557	1	-0	-0.2	0	0.69	0	1.018903	-16.139966	110	18	1.1	0.9;
	2558	1	159.27	38.31	0	2.18	0	1.018448	-16.992434	220	5	1.1	0.9;
	2559	1	17.02	5.9	0	0.52	0	1.02844	-24.493083	110	19	1.1	0.9;
	2560	1	6.51	1.4	0	0.54	0	1.00289	-25.000015	110	18	1.1	0.9;
	2561	1	1.8	0.7	0	0.59	0	1.000799	-31.486927	110	19	1.1	0.9;
	2562	1	-88.6	-24.2	0	5.66	0	0.998383	-52.123371	220	10	1.1	0.9;
	2563	1	52.86	13	0	6.04	0	1.044592	-4.927884	220	5	1.1	0.9;
	2564	1	0	0	0	0.93	0	1.025766	-53.795529	220	4	1.1	0.9;
	2565	1	-0	-0	0	1.27	0	1.006421	-14.843498	110	18	1.1	0.9;
	2566	1	0	-0	0	68.93	0	1.034203	-45.289648	380	4	1.1	0.9;
	2567	1	25.43	7.1	0	0.22	0	0.999382	-27.831283	110	18	1.1	0.9;
	2568	2	0	0	0	0	0	1.020248	-29.849094	110	21	1.1	0.9;
	2569	2	0	0	0	0	0	1.027206	62.77383	150	1	1.1	0.9;
	2570	1	-28.96	-8.62	0	3.08	0	1.073931	-48.191881	220	6	1.1	0.9;
	2571	1	-0	-0	0	1.16	0	1.022478	-29.065479	150	9	1.1	0.9;
	2572	2	0	0	0	0	0	1.044176	17.000824	150	8	1.1	0.9;
	2573	1	152.18	23.99	0	14.75	0	1.025584	-4.056179	220	13	1.1	0.9;
	2574	1	100.1	44.5	0	3.28	0	1.003942	-50.629163	220	6	1.1	0.9;
	2575	1	62.97	17.4	0	0.3	0	1.030857	-3.002196	220	5	1.1	0.9;
	2576	1	27.05	22.7	0	14.74	0	1.053356	-29.283394	150	9	1.1	0.9;
	2577	1	0	0	0	87.06	0	1.036793	-34.094563	380	6	1.1	0.9;
	2578	1	-0	0	0	-0.3	0	1.004649	-50.494286	220	6	1.1	0.9;
	2579	1	-0	-0	0	0.31	0	1.047813	-31.69296	110	1	1.1	0.9;
	2580	1	0	-0	0	1.61	0	0.917979	-35.780229	150	9	1.1	0.9;
	2581	1	-15.72	5.2	0	1.02	0	0.958173	-32.107715	150	9	1.1	0.9;
	2582	1	2.3	0.8	0	0.59	0	1.054389	-22.5657	110	19	1.1	0.9;
	2583	1	174.6	48.9	0	7.46	0	1.012051	-14.539673	380	3	1.1	0.9;
	2584	1	13.9	1.9	0	6.37	0	1.013649	9.107989	154	1	1.1	0.9;
	2585	1	-0	0	0.1083	35.537525	0	1.09874	-26.818256	380	21	1.3	0.7;
	2586	1	0	0	0	0.36	0	1.061697	-21.361889	150	9	1.1	0.9;
	2587	1	39.04	24.2	0	13.22	0	1.013679	-32.975234	150	9	1.1	0.9;
	2588	2	0	0	0	0	0	1.028186	-13.772507	220	14	1.1	0.9;
	2589	2	0	0	0	0	0	1.02499	-12.475736	380	6	1.1	0.9;
	2590	1	0	0	0	0.98	0	1.039701	3.293239	220	14	1.1	0.9;
	2591	1	253.77	84.6	0	20.71	0	1.025968	-15.577476	220	5	1.1	0.9;
	2592	1	28.13	14.2	0	0.49	0	0.978821	-24.09493	110	18	1.1	0.9;
	2593	1	-9.51	3.1	0	0.83	0	0.967156	-26.738596	150	9	1.1	0.9;
	2594	2	0	0	0	0	0	1.058099	-28.005721	220	6	1.1	0.9;
	2595	2	0	0	0	0	0	1.073729	13.746806	380	3	1.1	0.9;
	2596	1	0	-0	0	2.08	0	1.040499	-27.864931	150	9	1.1	0.9;
	2597	1	14.72	5.4	0	0.72	0	1.030134	-10.945874	220	5	1.1	0.9;
	2598	1	25.63	12.7	0	8.42	0	1.037585	-15.393641	220	5	1.1	0.9;
	2599	1	48.8	6.9	0	9.44	0	1.029312	14.802272	154	1	1.1	0.9;
	2600	2	0	0	0	0	0	1.047035	-13.430179	220	5	1.1	0.9;
	2601	1	-0	0	0	3.86	0	1.006105	-38.033085	220	1	1.1	0.9;
	2602	1	-23.13	-8.4	0	1.01	0	1.055981	-26.5034	110	19	1.1	0.9;
	2603	1	-0	0	0	5.09	0	0.985189	-34.73886	150	9	1.1	0.9;
	2604	2	0	0	0	0	0	1.031949	13.582886	220	13	1.1	0.9;
	2605	1	115.3	-8.8	0	16.22	0	1.049286	-10.626042	220	3	1.1	0.9;
	2606	2	0	0	0	0	0	1.041601	11.411295	220	3	1.1	0.9;
	2607	1	-0	0	0	11.54	0	1.05799	-19.927793	380	1	1.1	0.9;
	2608	1	226.8	53.2	0	0	0	1.006305	19.873213	220	3	1.1	0.9;
	2609	1	14.92	4.4	0	1.32	0	0.976604	-28.293193	110	18	1.1	0.9;
	2610	1	499.8	110.3	0	4.21	0	1.016134	8.806843	220	3	1.1	0.9;
	2611	1	0	-0	0	9.49	0	1.042426	-15.652309	380	1	1.1	0.9;
	2612	1	-0	0	0	29.89	0	1.064253	-34.598289	150	9	1.1	0.9;
	2613	1	0	-0	0	31.53	0	1.011055	-37.608964	220	12	1.1	0.9;
	2614	1	0	0	0	7.61	0	1.01386	-8.240749	220	3	1.1	0.9;
	2615	1	-0	0	0	23.88	0	1.029475	-45.006416	380	6	1.1	0.9;
	2616	1	-0	0	0	1.46	0	1.0389	4.953667	150	8	1.1	0.9;
	2617	1	14.74	14.7	0	7.86	0	1.048731	-29.212214	150	9	1.1	0.9;
	2618	1	69.7	9.9	0	2.89	0	1.011451	-12.59031	154	1	1.1	0.9;
	2619	1	0	0	0	1.01	0	1.016494	5.847438	150	8	1.1	0.9;
	2620	2	0	0	0	0	0	1.030287	-31.664696	220	12	1.1	0.9;
	2621	1	112.1	16	0	0.89	0	1.014006	-22.877447	154	1	1.1	0.9;
	2622	1	-0	0	0	6.91	0	1.04369	-50.070672	380	4	1.1	0.9;
	2623	1	64.4	9.1	0	1.61	0	1.019356	-7.385053	154	1	1.1	0.9;
	2624	2	0	0	0	0	0	1.037675	-25.413258	220	7	1.1	0.9;
	2625	1	10.61	3.5	0	1.61	0	1.029129	-33.934715	110	12	1.1	0.9;
	2626	1	5.51	2	0	0.8	0	1.062882	-18.416903	110	19	1.1	0.9;
	2627	2	0	0	0	0	0	1.048065	7.49733	380	5	1.1	0.9;
	2628	1	0	-0	0	1.39	0	0.976189	-31.892903	150	9	1.1	0.9;
	2629	1	203.1	65.2	0	0.64	0	1.041141	-24.819636	220	5	1.1	0.9;
	2630	1	166.7	87.6	0	60.34	0	1.017944	-13.180311	380	3	1.1	0.9;
	2631	1	29.93	-4.2	0	1.47	0	1.009262	36.74247	220	3	1.1	0.9;
	2632	1	655.8	72.8	0	10.76	0	1.022404	-10.187631	380	3	1.1	0.9;
	2633	1	3.7	1.3	0	0.91	0	1.015076	-23.917266	110	21	1.1	0.9;
	2634	1	0	-0	0	44.86	0	1.028551	-37.887644	380	6	1.1	0.9;
	2635	1	42.95	26.6	0	3.89	0	0.849692	-39.840933	150	9	1.3	0.7;
	2636	1	98.3	25.9	0	3.37	0	0.999505	-49.363859	220	6	1.1	0.9;
	2637	1	33.04	12.1	0	0.55	0	1.021527	-40.234528	110	22	1.1	0.9;
	2638	1	21.72	7.2	0	0.76	0	0.970876	-32.262237	110	21	1.1	0.9;
	2639	1	45.05	-4.5	0	1.57	0	1.041487	-27.48851	110	23	1.1	0.9;
	2640	1	37.6	5.4	0	4.56	0	1.01507	-5.047736	154	1	1.1	0.9;
	2641	1	0	0	0	1.72	0	1.036589	-11.578452	220	5	1.1	0.9;
	2642	2	0	0	0	0	0	1.052836	-37.523797	380	4	1.1	0.9;
	2643	1	2	0.5	0	1.13	0	1.004558	-30.619224	110	18	1.1	0.9;
	2644	1	0	0	0	0.19	0	1.068086	-25.190835	220	5	1.1	0.9;
	2645	1	-0	-0	0	0.28	0	0.994287	-25.732813	110	18	1.1	0.9;
	2646	1	123.1	3.9	0	59.08	0	1.060015	-11.39007	380	16	1.1	0.9;
	2647	1	70.18	30.5	0	0.48	0	1.001038	-0.332218	220	3	1.1	0.9;
	2648	1	46.65	15.1	0	9.79	0	1.058647	10.388439	110	8	1.1	0.9;
	2649	2	0	0	0	0	0	1.087006	-36.284446	380	22	1.1	0.9;
	2650	2	0	0	0	0	0	1.036234	-36.354105	220	6	1.1	0.9;
	2651	1	0	0	0	0.39	0	0.990147	-27.471409	110	18	1.1	0.9;
	2652	1	93.1	43.7	0	11.87	0	1.013967	-32.66926	150	9	1.1	0.9;
	2653	2	0	0	0	0	0	1.032593	-23.359386	220	5	1.1	0.9;
	2654	1	207.1	50.4	0	4.93	0	1.016656	-18.8984	220	5	1.1	0.9;
	2655	1	-0	-0	0	1.38	0	0.983151	9.190976	220	3	1.1	0.9;
	2656	1	-0.72	0	0	69.582294	0	1.031889	-23.227885	380	5	1.1	0.9;
	2657	1	0	0	0	1.28	0	1.018	51.385645	220	3	1.1	0.9;
	2658	2	0	0	0	0	0	1.020867	-49.501838	220	4	1.1	0.9;
	2659	1	-0	-0	0	64.39	0	1.000418	-4.242926	400	1	1.1	0.9;
	2660	2	0	0	0	0	0	1.011944	-31.986779	150	10	1.1	0.9;
	2661	1	45.45	28.1	0	0.16	0	0.942596	-32.670874	150	9	1.1	0.9;
	2662	2	0	0	0	0	0	1.014043	60.625336	150	1	1.1	0.9;
	2663	1	150	104.3	0	9.21	0	1.016474	-23.509763	220	3	1.1	0.9;
	2664	2	0	0	0	0	0	1.049614	-24.724471	150	9	1.1	0.9;
	2665	1	-0	0	0	2.73	0	0.982173	9.934066	220	3	1.1	0.9;
	2666	1	30.6	4.3	0	8.67	0	1.009676	-10.943069	154	1	1.1	0.9;
	2667	1	70.68	10.7	0	1.24	0	1.027765	-53.665943	220	4	1.1	0.9;
	2668	1	252.9	95.6	0	1.22	0	1.017882	-44.645979	220	4	1.1	0.9;
	2669	1	223.5	48	0	12.74	0	1.00927	-49.478765	380	6	1.1	0.9;
	2670	1	20.92	1.1	0	1.18	0	1.007831	-43.395901	110	22	1.1	0.9;
	2671	1	16.82	5.5	0	0.11	0	1.027314	-34.113967	110	12	1.1	0.9;
	2672	1	7.41	1.8	0	0.3	0	1.02577	-24.719652	110	19	1.1	0.9;
	2673	1	13.21	4.4	0	1.17	0	0.986099	-39.714474	110	22	1.1	0.9;
	2674	1	0	-0	0	0.12	0	0.964839	-31.450337	150	9	1.1	0.9;
	2675	1	0	-0	0	11.28	0	1.02691	-42.54101	220	6	1.1	0.9;
	2676	1	123.24	23.65	0	6.44	0	1.033964	-15.170761	220	5	1.1	0.9;
	2677	1	-0	0	0.252685	32.19909	0	1.031552	-19.193726	220	19	1.1	0.9;
	2678	1	10.61	3.8	0	0.96	0	1.065178	-25.764629	110	19	1.1	0.9;
	2679	1	27.93	10.1	0	0.44	0	1.047069	-22.055493	110	19	1.1	0.9;
	2680	1	-0	0	0	100.092366	0	1.036493	-36.212633	380	6	1.1	0.9;
	2681	1	0	0	0	29.87	0	1.07137	38.529441	380	3	1.1	0.9;
	2682	1	-0	-0	0	0.47	0	1.030385	-32.053978	220	12	1.1	0.9;
	2683	1	-0	-0	0	23.22	0	0.997836	-3.313818	400	1	1.1	0.9;
	2684	1	394.9	45.9	0	19.569841	0	1.019875	-37.557528	380	6	1.1	0.9;
	2685	2	0	0	0	0	0	1.010899	-46.024063	220	6	1.1	0.9;
	2686	2	0	0	0	0	0	1.016926	-43.107811	220	4	1.1	0.9;
	2687	1	2	1	0	0.17	0	0.999412	-31.321582	110	21	1.1	0.9;
	2688	1	0	0	0	0	0	0.91444	-35.99744	150	9	1.1	0.9;
	2689	1	0	0	0	0.45	0	1.029591	-3.635667	220	5	1.1	0.9;
	2690	1	-0	0	0	0.85	0	1.033725	-25.74667	150	9	1.1	0.9;
	2691	1	-5.89	-3.59	0	0.51	0	1.00458	-27.583723	110	18	1.1	0.9;
	2692	1	285.2	65.3	0	11.28	0	0.984603	-54.826143	220	10	1.1	0.9;
	2693	1	149.6	21.3	0	2.09	0	1.035092	-21.709891	154	1	1.1	0.9;
	2694	1	681.5	57.7	0	66.36	0	1.01873	-45.679842	380	6	1.1	0.9;
	2695	1	16.22	8.9	0	2.2	0	1.035198	-21.311735	220	5	1.1	0.9;
	2696	1	0	0	0	0.24	0	1.047114	-29.542662	150	9	1.1	0.9;
	2697	1	-0	0	0	0.06	0	0.997048	-32.273639	150	9	1.1	0.9;
	2698	1	0	-1.2	0	0.41	0	1.025059	25.157021	110	8	1.1	0.9;
	2699	1	23.13	8.8	0	0.95	0	1.022844	-33.862575	110	22	1.1	0.9;
	2700	1	-0	0	0	3.08	0	1.032391	-39.588219	380	4	1.1	0.9;
	2701	1	0	0	0	0	0	0.998514	-36.87879	150	9	1.1	0.9;
	2702	1	151.7	27.6	0	6.49	0	1.038835	-25.946279	220	5	1.1	0.9;
	2703	1	-0.2	-4.97	0	0	0	1.03322	8.744635	150	8	1.1	0.9;
	2704	1	56.26	18.5	0	1.01	0	1.012671	-38.95539	110	12	1.1	0.9;
	2705	1	117.4	12.1	0	1.14	0	1.015261	8.941171	220	3	1.1	0.9;
	2706	1	118	37.8	0	0	0	0.998322	-18.958592	110	18	1.1	0.9;
	2707	1	0	0	0.099501	119.783595	0	1.077199	-18.620358	380	18	1.1	0.9;
	2708	1	47.75	18.8	0	4.42	0	1.020394	-1.163142	150	8	1.1	0.9;
	2709	1	189.7	52	0	30.21	0	1.017832	-46.73738	380	6	1.1	0.9;
	2710	2	0	0	0	0	0	1.035068	-32.913136	380	6	1.1	0.9;
	2711	1	106.6	15.2	0	0.34	0	1.014654	-4.23761	154	1	1.1	0.9;
	2712	1	104.8	22	0	6.99	0	1.0119	5.624279	220	3	1.1	0.9;
	2713	1	11.01	4	0	0.73	0	0.985098	-19.398169	110	19	1.1	0.9;
	2714	1	-0	8.4	0	12.47	0	1.021979	-28.207355	150	9	1.1	0.9;
	2715	1	90.5	27.8	0	7.85	0	1.039521	-17.297222	220	7	1.1	0.9;
	2716	1	-0	0	0.349307	35.121782	0	1.030728	-23.553656	380	7	1.1	0.9;
	2717	1	32.04	11	0	0.57	0	1.051375	-24.797832	110	15	1.1	0.9;
	2718	1	48.8	6.9	0	5.48	0	1.047495	8.441712	154	1	1.1	0.9;
	2719	2	0	0	0	0	0	1.041789	-15.262311	220	5	1.1	0.9;
	2720	1	13.01	14.2	0	15.78	0	1.017355	17.213175	150	8	1.1	0.9;
	2721	1	-0	0	0	10.040547	0	0.970233	-22.79896	380	5	1.1	0.9;
	2722	1	0	0	0	56.39	0	1.024193	7.612357	380	8	1.1	0.9;
	2723	1	3.8	0.6	0	0.73	0	1.004169	-14.384208	110	18	1.1	0.9;
	2724	1	151.2	81.3	0	50.19	0	1.028195	9.110165	380	3	1.1	0.9;
	2725	2	0	0	0	0	0	1.014934	-32.528518	150	9	1.1	0.9;
	2726	1	0	0	0	4.68	0	1.026411	17.398129	380	3	1.1	0.9;
	2727	1	32.54	11.7	0	1.02	0	1.034724	-23.021547	110	15	1.1	0.9;
	2728	1	-0.13	-1	0	0	0	1.040851	25.722036	220	8	1.1	0.9;
	2729	2	0	0	0	0	0	1.050822	-30.778243	110	1	1.1	0.9;
	2730	1	0.1	152.7	0	233.41	0	1.038131	-44.171569	380	4	1.1	0.9;
	2731	1	44.05	14.6	0	0.6	0	0.951201	-34.851087	110	21	1.1	0.9;
	2732	1	85.59	20.72	0	142.79	0	1.009988	-10.274142	380	5	1.1	0.9;
	2733	1	36.19	10.75	0	0.58	0	0.990654	-18.000787	110	19	1.1	0.9;
	2734	1	-61.67	-0.9	0	12.33	0	0.990417	-48.424201	150	10	1.1	0.9;
	2735	1	29.5	4.3	0	21.5	0	1.024055	-7.446564	154	1	1.1	0.9;
	2736	2	0	0	0	0	0	1.032664	48.534354	220	3	1.1	0.9;
	2737	1	-0	-0	0.12635	9.466325	0	1.138972	-24.417364	380	21	1.3	0.7;
	2738	1	0	0	0	4.52	0	0.991581	-33.759111	220	12	1.1	0.9;
	2739	1	-0	-0	0	0.6	0	0.968155	-32.370791	150	9	1.1	0.9;
	2740	1	36.24	5.7	0	15.06	0	1.024555	-54.562835	220	4	1.1	0.9;
	2741	1	148.5	21.2	0	1.26	0	1.013219	-10.342488	154	1	1.1	0.9;
	2742	1	0	0	0	0	0	1.058712	-28.62346	150	9	1.1	0.9;
	2743	2	0	0	0	0	0	1.039467	49.466963	220	3	1.1	0.9;
	2744	1	-9.82	-3.4	0	0.21	0	1.015437	-23.030628	110	18	1.1	0.9;
	2745	1	0	-0	0	56.27	0	1.024272	-3.625612	150	8	1.1	0.9;
	2746	1	27.9	4	0	2.58	0	1.01238	-10.587219	154	1	1.1	0.9;
	2747	1	-11.77	-3.81	0	0.23	0	1.021113	-25.831255	110	18	1.1	0.9;
	2748	1	65.47	26.7	0	18.6	0	1.020283	-4.093043	150	8	1.1	0.9;
	2749	1	55.56	26.5	0	1.09	0	1.032053	-30.886187	220	5	1.1	0.9;
	2750	1	118.4	16.9	0	41.52	0	1.028659	-21.464327	154	1	1.1	0.9;
	2751	1	34.54	8.6	0	5.05	0	1.038706	-22.195411	220	5	1.1	0.9;
	2752	1	12.21	1	0	0.96	0	0.982719	-28.541001	110	18	1.1	0.9;
	2753	1	-0.63	-1.17	0	1.92	0	0.938258	-33.381368	150	9	1.1	0.9;
	2754	1	-0	0	0	3.72	0	1.055002	-13.426452	220	5	1.1	0.9;
	2755	1	0	-0	0	3.22	0	1.026681	-0.863624	380	3	1.1	0.9;
	2756	1	159.3	13.5	0	62.36	0	1.026317	-15.932493	380	6	1.1	0.9;
	2757	1	-0	-0	0	1.56	0	0.977317	-30.407805	150	9	1.1	0.9;
	2758	2	0	0	0	0	0	1.050098	13.34262	380	13	1.1	0.9;
	2759	1	90.4	13.8	0	0.35	0	1.031458	-3.483077	220	1	1.1	0.9;
	2760	1	64.87	25.6	0	14.89	0	1.026109	4.108859	150	8	1.1	0.9;
	2761	1	8.51	2.8	0	1.35	0	1.031993	-32.837218	110	12	1.1	0.9;
	2762	1	59.9	8.5	0	1.27	0	0.999792	1.59613	154	1	1.1	0.9;
	2763	1	16.52	6	0	0.55	0	1.004992	-28.810992	110	19	1.1	0.9;
	2764	2	0	0	0	0	0	1.045688	-49.588961	380	4	1.1	0.9;
	2765	1	40.5	5.7	0	4.13	0	1.018043	5.781934	154	1	1.1	0.9;
	2766	1	0	-0	0	12.96	0	1.034573	-37.389409	380	6	1.1	0.9;
	2767	1	0	0	0	0.47	0	0.98926	-29.34769	110	18	1.1	0.9;
	2768	1	34.9	5	0	2.1	0	1.015763	-7.408629	154	1	1.1	0.9;
	2769	1	98.6	28.3	0	4.76	0	1.022321	25.504546	220	8	1.1	0.9;
	2770	1	85.39	14.3	0	0.35	0	1.043105	-9.922486	220	5	1.1	0.9;
	2771	1	57.21	-7.12	0	6.86	0	1.053762	5.759052	220	13	1.1	0.9;
	2772	1	103.1	3.1	0	61.7	0	1.052122	-31.311203	380	11	1.1	0.9;
	2773	1	4.7	1.8	0	0.17	0	0.982801	-16.125592	110	18	1.1	0.9;
	2774	1	1.5	0.3	0	0.47	0	0.99937	-26.805513	110	18	1.1	0.9;
	2775	2	0	0	0	0	0	1.024728	-2.202138	400	1	1.1	0.9;
	2776	1	16.76	-0.51	0	0.35	0	1.035688	-25.116356	110	19	1.1	0.9;
	2777	1	26.63	7.6	0	0.1	0	1.055463	-17.204768	110	18	1.1	0.9;
	2778	1	0	0	0	0	0	1.001273	-31.157699	150	9	1.1	0.9;
	2779	1	35.34	21.9	0	3.03	0	0.93851	-33.36495	150	9	1.1	0.9;
	2780	1	2.9	0.8	0	0.68	0	0.975698	-27.577467	110	18	1.1	0.9;
	2781	1	-9.41	3.1	0	0.87	0	0.994606	-28.128704	150	9	1.1	0.9;
	2782	1	-28.13	-0.81	0	0.1	0	1.06541	-11.831115	220	5	1.1	0.9;
	2783	1	0	-0	0	4.47	0	0.96428	-31.550888	150	9	1.1	0.9;
	2784	1	0.3	1.3	0	0.28	0	0.998791	9.953837	220	3	1.1	0.9;
	2785	1	189.4	25.8	0	3.62	0	0.977375	-50.459208	220	4	1.1	0.9;
	2786	2	0	0	0	0	0	1.078984	6.21927	380	5	1.3	0.7;
	2787	1	0	-0	0	1.77	0	1.004684	8.971485	220	3	1.1	0.9;
	2788	1	129.5	26	0	0	0	1.011891	20.397835	220	3	1.1	0.9;
	2789	1	0	0	0	0	0	0.991606	-34.386656	150	9	1.1	0.9;
	2790	1	-0	-0	0	0.87	0	1.078102	-18.667474	110	19	1.1	0.9;
	2791	2	0	0	0	0	0	1.015494	-6.230119	154	1	1.1	0.9;
	2792	1	3	0.1	0	0.42	0	0.978902	-28.402875	110	18	1.1	0.9;
	2793	1	16.7	2.4	0	2.21	0	1.029761	-12.0953	154	1	1.1	0.9;
	2794	1	0	-0	0	0.39	0	1.031591	-18.751587	220	5	1.1	0.9;
	2795	1	-0	-0	0	1.94	0	1.065409	-32.692499	220	5	1.1	0.9;
	2796	1	100.7	17.5	0	6.54	0	1.032648	8.706417	150	8	1.1	0.9;
	2797	2	0	0	0	0	0	1.041702	-4.550845	220	5	1.1	0.9;
	2798	1	8.91	2.7	0	1.01	0	0.998181	-28.063947	110	18	1.1	0.9;
	2799	2	0	0	0	0	0	1.059752	-12.774681	220	5	1.1	0.9;
	2800	2	0	0	0	0	0	1.02478	-49.077869	380	4	1.1	0.9;
	2801	1	0	-0	0	12.17	0	1.038595	-24.830424	220	5	1.1	0.9;
	2802	1	24.63	8.1	0	0.93	0	0.996128	-30.484678	110	21	1.1	0.9;
	2803	1	69.08	68	0	2.93	0	1.011048	-36.660202	220	6	1.1	0.9;
	2804	1	31.33	9.6	0	1.86	0	1.024956	12.196809	110	8	1.1	0.9;
	2805	1	390.8	79.5	0	20.79	0	1.03929	-42.532182	380	6	1.1	0.9;
	2806	1	-0	-0	0	0.39	0	1.047179	0.06827	220	5	1.1	0.9;
	2807	1	327.4	125.3	0	34.36	0	1.049225	-15.943237	380	16	1.1	0.9;
	2808	1	72	10.3	0	5.5	0	1.004781	-7.608827	154	1	1.1	0.9;
	2809	1	27.53	11	0	2.49	0	0.976442	-43.345773	110	22	1.1	0.9;
	2810	2	0	0	0	0	0	1.036015	10.886015	154	1	1.1	0.9;
	2811	1	6.91	2.3	0	0.74	0	1.016293	-36.133941	110	12	1.1	0.9;
	2812	1	97	18.6	0	1.25	0	1.014703	16.417772	150	8	1.1	0.9;
	2813	1	32.94	11.8	0	0.83	0	1.030804	-20.424831	110	19	1.1	0.9;
	2814	1	0	0	0	52.89	0	1.083254	-21.619172	380	9	1.1	0.9;
	2815	1	25.33	14.2	0	4.05	0	1.03261	-23.33567	220	5	1.1	0.9;
	2816	2	0	0	0	0	0	1.049577	-0.552935	380	5	1.1	0.9;
	2817	2	0	0	0	0	0	0.987457	-46.749645	150	10	1.1	0.9;
	2818	1	76.6	10.9	0	2.96	0	1.020677	-18.188454	154	1	1.1	0.9;
	2819	1	22.02	2.2	0	2.4	0	1.031706	-40.994508	110	22	1.1	0.9;
	2820	1	0	0	0	1.28	0	0.990161	-28.867672	150	9	1.1	0.9;
	2821	1	81.19	31.4	0	2.04	0	0.996239	-50.859495	220	6	1.1	0.9;
	2822	2	0	0	0	0	0	1.020284	-47.696189	220	6	1.1	0.9;
	2823	1	19.32	6.4	0	3.74	0	1.031745	-34.063986	110	12	1.1	0.9;
	2824	1	-0	-0	0	4.17	0	1.027527	17.039752	220	8	1.1	0.9;
	2825	1	0	-0	0	31.79	0	1.027309	-31.879944	150	9	1.1	0.9;
	2826	2	0	0	0	0	0	1.022551	14.133547	154	1	1.1	0.9;
	2827	1	8.91	4.9	0	4.58	0	1.0311	-0.969133	220	13	1.1	0.9;
	2828	1	166.7	87.6	0	4.22	0	0.991601	-19.963847	220	3	1.1	0.9;
	2829	1	69.28	23.8	0	2.56	0	1.04606	-39.53587	110	22	1.1	0.9;
	2830	1	46.55	9.1	0	0.68	0	0.962464	-28.599476	110	18	1.1	0.9;
	2831	1	7.41	2.7	0	0.56	0	1.030661	-28.268286	110	19	1.1	0.9;
	2832	1	48.15	14.8	0	2.16	0	1.017823	11.708681	110	8	1.1	0.9;
	2833	1	62.7	9	0	8.23	0	1.00055	-8.454266	154	1	1.1	0.9;
	2834	1	3.8	1.3	0	0.11	0	0.991996	-17.769124	110	18	1.1	0.9;
	2835	2	0	0	0	0	0	1.043216	-4.574191	380	13	1.1	0.9;
	2836	1	31.28	5.34	0	8.75	0	1.014779	62.286835	150	1	1.1	0.9;
	2837	1	-0	-0	0	21.27	0	0.979638	-37.172863	150	9	1.1	0.9;
	2838	1	-9.31	3	0	0.02	0	0.975149	-30.416337	150	9	1.1	0.9;
	2839	2	0	0	0	0	0	1.03617	66.168864	150	1	1.1	0.9;
	2840	1	330.9	51.9	0	0	0	1.073314	-23.976137	120	20	1.1	0.9;
	2841	2	0	0	0	0	0	1.080183	-26.698323	220	5	1.1	0.9;
	2842	2	0	0	0	0	0	1.037135	-6.862585	220	5	1.1	0.9;
	2843	1	25.13	15.6	0	0.04	0	0.977313	-30.412799	150	9	1.1	0.9;
	2844	1	179	50.5	0	3.08	0	1.005585	-46.599252	220	4	1.1	0.9;
	2845	1	21.22	5.3	0	1.25	0	0.988307	-27.518793	110	18	1.1	0.9;
	2846	1	42.25	4.4	0	9.58843	0	0.95674	-20.143962	220	16	1.1	0.9;
	2847	1	40.14	8.8	0	0.82	0	1.048491	-4.163013	220	5	1.1	0.9;
	2848	1	453.23	139.16	0	25.583218	0	0.992888	-8.120882	380	5	1.1	0.9;
	2849	1	0	0	0	70.11	0	1.033309	-43.152205	380	4	1.1	0.9;
	2850	1	-0	-0	0	4.76	0	1.030813	-18.975096	220	5	1.1	0.9;
	2851	1	-0	-0	0	0.3	0	1.01491	-29.563895	150	9	1.1	0.9;
	2852	2	0	0	0	0	0	1.063812	55.043975	220	3	1.1	0.9;
	2853	2	0	0	0	0	0	0.998154	-36.333293	150	10	1.1	0.9;
	2854	1	-12.33	-2.65	0	69.323028	0	1.045263	-11.049115	380	5	1.1	0.9;
	2855	1	-0	0	0.234854	45.08542	0	1.091144	-19.374171	380	21	1.1	0.9;
	2856	1	7.11	4.4	0	0.97	0	0.971505	-27.269578	150	9	1.1	0.9;
	2857	1	4.2	1.1	0	0.73	0	0.989308	-21.617076	110	18	1.1	0.9;
	2858	1	192.3	84.6	0	38.41	0	1.012613	-22.000892	220	20	1.1	0.9;
	2859	1	-57	-24.9	0	8.1	0	0.953125	-33.534205	150	9	1.1	0.9;
	2860	1	0	0	0	0	0	1.003452	-31.424496	150	9	1.1	0.9;
	2861	1	17.72	5.9	0	0.17	0	1.007288	-31.886508	110	21	1.1	0.9;
	2862	2	0	0	0	0	0	1.027949	-8.633574	220	13	1.1	0.9;
	2863	1	37.73	13.84	0	0.09	0	1.029675	-11.102464	220	5	1.1	0.9;
	2864	1	30.03	10	0	0.66	0	0.988359	-31.358077	110	21	1.1	0.9;
	2865	1	46.35	16.8	0	8.35	0	1.037073	-21.986951	220	19	1.1	0.9;
	2866	1	-0	0	0	4.21	0	1.031218	-7.49201	220	5	1.1	0.9;
	2867	2	0	0	0	0	0	1.047385	-21.683087	220	11	1.1	0.9;
	2868	1	0	-0	0	1.6	0	1.030278	-10.667649	110	18	1.1	0.9;
	2869	1	18.8	8.5	0	75.218793	0	1.133869	-22.192198	380	9	1.3	0.7;
	2870	1	-0	0	0	0.62	0	1.018031	4.736975	380	2	1.1	0.9;
	2871	1	59.32	-20.32	0	4.86	0	1.01203	-49.019171	220	6	1.1	0.9;
	2872	2	0	0	0	0	0	1.054344	-11.000358	220	5	1.1	0.9;
	2873	1	33.36	20	0	0	0	1.085818	-36.683985	150	9	1.3	0.7;
	2874	2	0	0	0	0	0	1.006921	6.527903	400	1	1.1	0.9;
	2875	1	69.28	23.1	0	4.86	0	1.040036	-39.699982	110	22	1.1	0.9;
	2876	1	24.13	15	0	3.86	0	1.043966	-30.135632	150	9	1.1	0.9;
	2877	1	29.93	-0	0	0.46	0	1.03854	-15.020325	220	5	1.1	0.9;
	2878	2	0	0	0	0	0	1.047923	-11.279115	220	5	1.1	0.9;
	2879	1	13.82	5	0	0.24	0	1.040681	-8.793401	110	19	1.1	0.9;
	2880	1	-0	-0	0	14.2	0	0.999267	-50.029004	220	6	1.1	0.9;
	2881	2	0	0	0	0	0	1.024879	5.732448	154	1	1.1	0.9;
	2882	1	-7.47	-6.45	0	1.08	0	0.990424	-28.476946	110	18	1.1	0.9;
	2883	1	205.9	38.3	0	6.54	0	1.027049	-12.837	220	3	1.1	0.9;
	2884	1	282.5	17.4	0	20.04	0	1.021727	-38.978171	380	6	1.1	0.9;
	2885	1	0	-0	0	24.73	0	1.092743	-27.783251	380	1	1.1	0.9;
	2886	2	0	0	0	0	0	1.057004	-14.159417	220	5	1.1	0.9;
	2887	1	57.66	61	0	6.46	0	1.014384	-42.141546	220	17	1.1	0.9;
	2888	1	-0	0	0	0.22	0	1.008512	-21.614525	220	5	1.1	0.9;
	2889	1	-0	0	0	3.2	0	1.048889	-1.184833	220	5	1.1	0.9;
	2890	1	102.8	30.8	0	2.66	0	1.049854	-8.204924	380	3	1.1	0.9;
	2891	1	254.9	71.4	0	0.7	0	1.00994	0.65866	380	3	1.1	0.9;
	2892	1	10.11	3.7	0	0.52	0	1.107563	-16.256154	110	19	1.3	0.7;
	2893	1	-0	-0	0	2.82	0	0.994768	-33.816686	220	1	1.1	0.9;
	2894	1	-0	0	0	0.64	0	0.999263	-39.177231	150	10	1.1	0.9;
	2895	1	151	45.7	0	3.02	0	1.015901	-14.604545	220	3	1.1	0.9;
	2896	2	0	0	0	0	0	1.009058	-44.759958	220	4	1.1	0.9;
	2897	2	0	0	0	0	0	1.079272	-21.14894	110	23	1.1	0.9;
	2898	1	56.76	13.4	0	1.18	0	1.031918	-23.422434	220	5	1.1	0.9;
	2899	1	-0	-0	0	44.66	0	1.051592	-45.108007	380	10	1.1	0.9;
	2900	1	0	0	0	0	0	1.04402	3.726033	150	8	1.1	0.9;
	2901	1	-0.51	-20.69	0	2.07	0	1.026238	-16.034962	220	3	1.1	0.9;
	2902	2	0	0	0	0	0	1.069	-23.903912	220	5	1.1	0.9;
	2903	1	133.7	59	0	2.48	0	0.984771	7.428385	220	3	1.1	0.9;
	2904	1	0	-0	0.37905	21.97435	0	1.078809	-15.204432	380	18	1.1	0.9;
	2905	1	-0	0	0	0.43	0	1.013927	-36.121069	220	6	1.1	0.9;
	2906	1	255.8	144.8	0	52.33	0	1.022345	-10.578895	380	3	1.1	0.9;
	2907	1	241	32.3	0	34.74	0	1.021017	-48.745448	220	4	1.1	0.9;
	2908	1	203.8	61	0	0.6	0	1.020421	-10.96388	150	2	1.1	0.9;
	2909	1	-0	0	0	5.21	0	1.016841	-36.283483	220	6	1.1	0.9;
	2910	1	-0	0	0	2.87	0	1.036477	-2.654621	220	5	1.1	0.9;
	2911	1	10.91	3.6	0	1.21	0	1.026114	-34.141297	110	12	1.1	0.9;
	2912	1	46.25	32.2	0	1.5	0	1.034581	39.039067	220	3	1.1	0.9;
	2913	1	51.96	9.4	0	42.51	0	1.062326	-42.065477	380	4	1.1	0.9;
	2914	1	-0	-0	0	3.43	0	1.001622	-31.241205	110	19	1.1	0.9;
	2915	1	-0	-0	0	0.66	0	0.999553	-27.968035	110	18	1.1	0.9;
	2916	1	63.47	30.2	0	81.36	0	1.047375	-38.507533	380	4	1.1	0.9;
	2917	1	31.64	7.5	0	21.91	0	1.021158	-47.142434	220	4	1.1	0.9;
	2918	1	0	0	0	53.25003	0	1.015176	-28.547308	380	5	1.1	0.9;
	2919	1	-0	0	0	32.049116	0	0.987081	-18.615623	380	5	1.1	0.9;
	2920	2	0	0	0	0	0	1.036969	-37.703704	380	6	1.1	0.9;
	2921	1	37.24	7.1	0	1.07	0	1.006267	-40.895161	220	4	1.1	0.9;
	2922	1	34.04	12.3	0	0.25	0	1.079186	-24.897557	110	19	1.1	0.9;
	2923	2	0	0	0	0	0	1.040379	-41.959629	380	6	1.1	0.9;
	2924	1	19.92	7.4	0	17.28	0	1.035974	-34.013211	220	5	1.1	0.9;
	2925	1	19.04	7.9	0	4.53	0	1.02447	-26.749332	150	9	1.1	0.9;
	2926	1	61.97	14.3	0	8.38	0	1.044436	-19.423255	220	7	1.1	0.9;
	2927	1	-0	0	0	26.18	0	1.047894	17.850304	380	8	1.1	0.9;
	2928	1	0	-0	0	92.38	0	1.068841	-26.733551	220	5	1.1	0.9;
	2929	1	-0	-0	0	0.34	0	1.045999	-29.589377	150	9	1.1	0.9;
	2930	2	0	0	0	0	0	1.054065	-13.329734	220	5	1.1	0.9;
	2931	1	230.7	19.3	0	4.56	0	1.040596	-32.581614	220	5	1.1	0.9;
	2932	1	17.7	-333.3	0.132526	89.131068	0	1.054877	17.294612	380	3	1.1	0.9;
	2933	1	70.08	10	0	33.04	0	1.054733	-33.13248	380	11	1.1	0.9;
	2934	2	0	0	0	0	0	1.049297	-6.587965	220	5	1.1	0.9;
	2935	1	-0	-0	0	-4.4045	0	1.085479	-16.581927	380	1	1.3	0.7;
	2936	1	-0	-0	0	3.9	0	1.064112	-23.309406	220	1	1.1	0.9;
	2937	1	16.72	6.1	0	0.41	0	0.944097	-22.244519	110	19	1.1	0.9;
	2938	1	0	-0	0	137.29	0	1.001021	-7.705453	380	5	1.1	0.9;
	2939	1	0	0	0.26	6.5	0	1.043126	-25.836237	220	21	1.1	0.9;
	2940	1	103	28.7	0	2.88	0	1.011603	-14.637485	220	5	1.1	0.9;
	2941	1	0	0	0	35.93	0	1.061091	-35.525707	380	1	1.1	0.9;
	2942	2	0	0	0	0	0	1.049098	-16.466845	380	19	1.1	0.9;
	2943	1	0	-0	0	0.75	0	0.956647	-31.595848	150	9	1.1	0.9;
	2944	1	0	0	0	2.32	0	1.0151	6.183297	220	3	1.1	0.9;
	2945	1	41.95	26	0	12.42	0	1.013102	-33.068927	150	9	1.1	0.9;
	2946	2	0	0	0	0	0	1.012911	-46.294974	220	4	1.1	0.9;
	2947	1	-21.38	-13.23	0	2.52	0	1.057374	-25.784371	110	21	1.1	0.9;
	2948	1	215.7	60.28	0	97.42	0	1.03412	-0.509094	380	13	1.1	0.9;
	2949	1	-0	-0	0	11.03	0	1.043611	-5.024372	220	5	1.1	0.9;
	2950	1	16.22	5.7	0	4.68	0	1.07764	-22.720407	110	15	1.1	0.9;
	2951	1	104.6	40.7	0	1.42	0	1.008726	-54.443403	220	4	1.1	0.9;
	2952	1	94.3	13.4	0	0.9	0	1.008174	-1.804058	154	1	1.1	0.9;
	2953	1	0	0	0	0	0	1.008824	-32.768968	150	9	1.1	0.9;
	2954	1	4.81	1.2	0	0.13	0	0.984971	-56.456146	150	10	1.1	0.9;
	2955	1	188.2	23.4	0	17.95	0	1.029208	-20.116623	220	3	1.1	0.9;
	2956	2	0	0	0	0	0	1.032523	-11.250517	380	6	1.1	0.9;
	2957	1	81.19	24.4	0	0.6	0	1.02001	-39.361605	220	4	1.1	0.9;
	2958	1	16.72	5.5	0	0.66	0	1.011738	-39.294638	110	12	1.1	0.9;
	2959	1	-0.45	-0.03	0	0.46	0	1.003446	-21.373326	110	18	1.1	0.9;
	2960	1	-0	0	0	0.4	0	0.991571	-23.778785	150	9	1.1	0.9;
	2961	1	159.9	53.2	0	7.99	0	1.03879	-24.714825	220	5	1.1	0.9;
	2962	1	93.5	42.5	0	133.56	0	1.061514	-43.047139	380	4	1.1	0.9;
	2963	1	-0	-0	0	22.54	0	1.034488	-24.679438	380	7	1.1	0.9;
	2964	1	0	-0	0	0.68	0	1.002355	-30.878774	150	9	1.1	0.9;
	2965	1	-29.96	-21.31	0	0	0	1.032633	17.696999	150	8	1.1	0.9;
	2966	1	4.4	1.5	0	0.58	0	0.965485	-25.299972	110	18	1.1	0.9;
	2967	1	0	-0	0	0.19	0	1.048781	-17.081495	220	5	1.1	0.9;
	2968	1	112.51	16.51	0	1.63	0	1.040217	-9.191608	220	5	1.1	0.9;
	2969	1	9.31	3	0	0.08	0	1.033558	-34.216418	110	12	1.1	0.9;
	2970	1	-0	-0	0	1.66	0	1.025839	-24.861508	150	9	1.1	0.9;
	2971	1	0	0	0	11.55	0	0.987575	-6.33598	380	1	1.1	0.9;
	2972	1	97.82	20.59	0	6.67	0	1.041103	-15.150263	220	5	1.1	0.9;
	2973	2	0	0	0	0	0	1.088757	7.127899	380	14	1.1	0.9;
	2974	1	65.47	26.7	0	0.6	0	1.02475	-2.696229	150	8	1.1	0.9;
	2975	2	0	0	0	0	0	1.03823	-33.552254	110	22	1.1	0.9;
	2976	1	0	0	0	0.93	0	0.987069	-31.650843	110	21	1.1	0.9;
	2977	1	5.81	1.3	0	0.83	0	0.992153	-31.495485	110	21	1.1	0.9;
	2978	1	0	0	0	2.47	0	1.022192	-48.811885	220	4	1.1	0.9;
	2979	1	13.9	1.9	0	1.5	0	1.022074	6.277637	154	1	1.1	0.9;
	2980	1	-0	0	0	0.42	0	1.057858	-9.33132	380	5	1.1	0.9;
	2981	1	-0	0	0	1.68	0	1.070326	-15.59772	220	5	1.1	0.9;
	2982	2	0	0	0	0	0	1.020472	-26.26597	110	21	1.1	0.9;
	2983	1	26.83	13	0	0.36	0	0.996013	-18.679169	220	6	1.1	0.9;
	2984	1	17.35	-0.8	0	1.21	0	0.979502	-28.268969	110	18	1.1	0.9;
	2985	2	0	0	0	0	0	1.046452	-8.853706	220	5	1.1	0.9;
	2986	1	0	0	0	21.88	0	1.049678	-30.353615	380	6	1.1	0.9;
	2987	2	0	0	0	0	0	1.060068	-13.245919	220	7	1.1	0.9;
	2988	1	27.03	9	0	0.2	0	0.985971	-33.277396	110	21	1.1	0.9;
	2989	2	0	0	0	0	0	1.066576	-9.028732	380	14	1.1	0.9;
	2990	1	0	-0	0	12.22	0	1.036355	8.130253	380	3	1.1	0.9;
	2991	1	0	0	0	0	0	0.849692	-39.840933	150	9	1.3	0.7;
	2992	2	0	0	0	0	0	1.104121	-28.617369	380	12	1.3	0.7;
	2993	1	17.72	7.3	0	0.13	0	0.961041	-29.937601	110	18	1.1	0.9;
	2994	1	5.81	1.9	0	3.45	0	1.026577	-34.598751	110	12	1.1	0.9;
	2995	1	0	-0	0	3.9	0	1.05007	-3.99566	220	5	1.1	0.9;
	2996	1	31.84	5.5	0	0.25	0	0.992457	-17.729497	110	18	1.1	0.9;
	2997	2	0	0	0	0	0	1.089588	11.337758	380	13	1.3	0.7;
	2998	1	17.02	6.2	0	0.87	0	1.026711	-18.1185	110	19	1.1	0.9;
	2999	1	48.8	6.9	0	14.55	0	1.009359	-7.569742	154	1	1.1	0.9;
	3000	1	139.6	68.4	0	0	0	1.016633	14.476175	150	8	1.1	0.9;
	3001	1	87.73	23.17	0	2.68	0	1.038947	0.573036	220	13	1.1	0.9;
	3002	2	0	0	0	0	0	1.015981	-4.196438	154	1	1.1	0.9;
	3003	2	0	0	0	0	0	1.048371	4.644305	220	13	1.1	0.9;
	3004	1	-0	0	0	0.77	0	1.049384	-22.134337	150	9	1.1	0.9;
	3005	1	-0	-0	0	2.03	0	1.034768	-8.954997	380	3	1.1	0.9;
	3006	1	87.6	27.4	0	0.99	0	1.02229	-9.808214	110	19	1.1	0.9;
	3007	1	108.8	33.2	0.358716	45.62985	0	1.026831	-52.43662	380	10	1.1	0.9;
	3008	1	0	0	0	22.7	0	1.015578	-46.572417	380	4	1.1	0.9;
	3009	2	0	0	0	0	0	1.016186	60.721526	150	1	1.1	0.9;
	3010	1	0	0	0	0	0	0.997745	-37.019205	150	9	1.1	0.9;
	3011	1	33.34	11	0	1.16	0	0.932629	-36.806773	110	21	1.1	0.9;
	3012	1	-1398.9	-380.6	0	58.1	0	1.045286	-6.147845	380	3	1.1	0.9;
	3013	1	2.2	1.1	0	0.67	0	1.042921	-5.434538	220	5	1.1	0.9;
	3014	1	12.81	3.9	0	0.66	0	1.065098	-23.02092	110	15	1.1	0.9;
	3015	1	-0	-0	0	17.77	0	1.029181	-6.402962	220	13	1.1	0.9;
	3016	1	30.23	10	0	0.31	0	1.029637	-26.416748	110	21	1.1	0.9;
	3017	1	-0	-0	0.27075	22.54135	0	1.073135	-26.089398	380	18	1.1	0.9;
	3018	2	0	0	0	0	0	1.049496	-13.234142	220	5	1.1	0.9;
	3019	1	0	-0	0	1.37	0	1.032865	-23.112666	220	5	1.1	0.9;
	3020	1	0	0	0	0.64	0	1.026507	-40.896191	220	6	1.1	0.9;
	3021	1	74.38	8.8	0	6.1	0	1.037061	-11.750342	220	5	1.1	0.9;
	3022	1	149	19.8	0	23.89	0	1.035342	-15.176209	380	5	1.1	0.9;
	3023	1	0	-0	0	27.69	0	1.009179	8.109483	380	8	1.1	0.9;
	3024	1	-0	0	0	1.57	0	1.00813	-35.11709	220	6	1.1	0.9;
	3025	1	-0	-0	0.169151	40.514944	0	1.037826	-20.011046	380	11	1.1	0.9;
	3026	1	0	-0	0	0.9	0	0.990606	-31.511948	110	21	1.1	0.9;
	3027	1	0	0	0	1.54	0	1.025585	-53.810418	220	4	1.1	0.9;
	3028	2	0	0	0	0	0	1.02789	0.312117	380	5	1.1	0.9;
	3029	2	0	0	0	0	0	1.011669	-24.036628	110	18	1.1	0.9;
	3030	1	78.59	25.6	0	1	0	0.955245	-31.245551	110	18	1.1	0.9;
	3031	1	6.61	2.4	0	0.61	0	1.002742	-30.544865	110	18	1.1	0.9;
	3032	1	-0	0	0	2.37	0	1.03482	8.756423	154	1	1.1	0.9;
	3033	1	41.15	5.7	0	4.11	0	0.995509	-27.246275	110	18	1.1	0.9;
	3034	1	4.61	1.4	0	0.83	0	0.99031	-21.392171	110	18	1.1	0.9;
	3035	2	0	0	0	0	0	1.0405	-44.417652	380	6	1.1	0.9;
	3036	1	116.93	33.6	0	3.35	0	1.042421	-25.50891	220	5	1.1	0.9;
	3037	1	221	146.8	0	8.35	0	1.046509	-6.795112	220	5	1.1	0.9;
	3038	2	0	0	0	0	0	1.041581	11.016318	220	3	1.1	0.9;
	3039	2	0	0	0	0	0	1.027247	-13.246536	220	13	1.1	0.9;
	3040	1	-0	0	0	0.86	0	0.964351	-31.543421	150	9	1.1	0.9;
	3041	1	0	0	0	0.91	0	1.051796	-17.803926	220	7	1.1	0.9;
	3042	1	0	-0	0	50	0	1.033591	-24.386253	380	7	1.1	0.9;
	3043	1	26.93	8	0	0.27	0	0.988359	-26.324046	110	18	1.1	0.9;
	3044	1	29.33	5.6	0	6.56	0	1.034203	-21.502172	220	5	1.1	0.9;
	3045	2	0	0	0	0	0	1.015762	-21.786592	154	1	1.1	0.9;
	3046	1	-0	-0	0	4.21	0	1.001737	-33.665913	220	1	1.1	0.9;
	3047	1	0	-0	0	3.84	0	1.020996	2.361043	150	8	1.1	0.9;
	3048	1	95.91	-15.4	0	1.51	0	1.036371	-29.916545	220	6	1.1	0.9;
	3049	1	-0	0	0	0.5	0	1.055171	-29.298676	150	9	1.1	0.9;
	3050	1	-0.54	-0.77	0.181549	1.270889	0	1.069212	-16.749495	220	18	1.1	0.9;
	3051	1	0	0	0	0	0	1.033036	-25.366125	220	5	1.1	0.9;
	3052	1	0	-0	0	8.21	0	1.026583	2.770976	380	3	1.1	0.9;
	3053	1	10.31	4.6	0	0.64	0	1.026793	-26.97167	110	15	1.1	0.9;
	3054	2	0	0	0	0	0	0.995142	-52.008983	220	10	1.1	0.9;
	3055	1	1.4	0.2	0	5	0	1.032793	12.002819	154	1	1.1	0.9;
	3056	1	22.5	9.5	0	29.37436	0	1.129011	-23.483011	380	9	1.3	0.7;
	3057	1	-17.8	-3.68	0.110438	41.234223	0	1.103456	43.57667	380	3	1.3	0.7;
	3058	1	0	-0	0	0.82	0	1.009579	-5.065324	154	1	1.1	0.9;
	3059	1	-14.85	-4.99	0	0.6	0	1.012119	-24.291729	110	18	1.1	0.9;
	3060	1	0	0	0	0.49	0	0.969256	-26.685103	150	9	1.1	0.9;
	3061	1	493.4	53.5	0	58.09	0	1.028742	-48.459328	380	4	1.1	0.9;
	3062	1	8.71	1.6	0	1.23	0	1.038916	-11.177215	110	19	1.1	0.9;
	3063	1	71.28	6.9	0	1.93	0	1.009389	-45.957702	220	4	1.1	0.9;
	3064	2	0	0	0	0	0	1.022454	6.290456	400	1	1.1	0.9;
	3065	1	55.86	8.8	0	4.2	0	1.022689	-54.937831	220	4	1.1	0.9;
	3066	1	119.7	3.7	0	0.88	0	1.023701	-54.698384	220	4	1.1	0.9;
	3067	1	272.3	58.5	0	23.57	0	1.023558	-42.498692	380	4	1.1	0.9;
	3068	1	566.1	54.2	0	44.559841	0	1.014253	-48.193733	380	6	1.1	0.9;
	3069	1	89.6	22.7	0	4.06	0	1.031548	-8.816262	220	5	1.1	0.9;
	3070	1	0	-0	0	4.27	0	1.008506	-21.61448	220	5	1.1	0.9;
	3071	1	0	-0	0	2.83	0	1.043171	-7.35147	220	5	1.1	0.9;
	3072	1	92.9	27.7	0	5.22	0	1.032815	-30.857548	220	5	1.1	0.9;
	3073	1	100.7	-23.3	0	2.97	0	1.036802	-23.378769	220	7	1.1	0.9;
	3074	1	0	0	0	0.51	0	1.013258	-31.770181	150	9	1.1	0.9;
	3075	1	30.33	10.5	0	0.33	0	1.018015	-17.076837	220	5	1.1	0.9;
	3076	1	164	42.3	0	8.07	0	1.008645	-48.288192	220	4	1.1	0.9;
	3077	1	11.91	3	0	1.22	0	1.002162	-19.016337	110	18	1.1	0.9;
	3078	2	0	0	0	0	0	1.024108	-28.323701	220	6	1.1	0.9;
	3079	1	10.11	3.3	0	0.92	0	0.957868	-22.853426	110	18	1.1	0.9;
	3080	1	-0	-0	0.256503	101.229042	0	1.059423	-47.950493	380	10	1.1	0.9;
	3081	1	0	0	0	0.24	0	1.079373	-22.093837	150	9	1.1	0.9;
	3082	1	0	0	0	75.076038	0	0.995759	-15.36659	380	5	1.1	0.9;
	3083	1	567.97	130.73	0	1.09	0	1.065475	-27.933024	220	5	1.1	0.9;
	3084	2	0	0	0	0	0	1.042602	0.578717	220	13	1.1	0.9;
	3085	1	64.67	24.5	0	1.46	0	1.0643	-27.989896	220	5	1.1	0.9;
	3086	1	0	0	0	0.93	0	1.001534	-31.139027	150	9	1.1	0.9;
	3087	1	0	0	0	3.64	0	1.015735	-38.890869	220	1	1.1	0.9;
	3088	2	0	0	0	0	0	1.043452	-13.261837	220	7	1.1	0.9;
	3089	1	15.22	4.2	0	1.35	0	1.022413	-34.775191	110	1	1.1	0.9;
	3090	2	0	0	0	0	0	1.003861	-8.868521	220	3	1.1	0.9;
	3091	2	0	0	0	0	0	1.021666	-5.352923	154	1	1.1	0.9;
	3092	1	28.63	12.1	0	0.77	0	1.016073	-40.204318	110	22	1.1	0.9;
	3093	1	11.21	3.7	0	0.49	0	1.04636	-31.97867	110	12	1.1	0.9;
	3094	1	4.1	1.5	0	0.52	0	1.061022	-18.293025	110	19	1.1	0.9;
	3095	1	69.78	15.3	0	1.34	0	1.00647	-54.688448	220	4	1.1	0.9;
	3096	2	0	0	0	0	0	1.008864	-5.774957	154	1	1.1	0.9;
	3097	1	128.11	64.9	0	19.1	0	1.01882	4.182018	380	2	1.1	0.9;
	3098	1	-0	-0	0	6.1	0	1.015846	-32.549429	220	12	1.1	0.9;
	3099	1	380.9	85.1	0	57.449841	0	1.038121	-27.474863	380	6	1.1	0.9;
	3100	1	-0	-0	0	4	0	1.009767	-30.858281	150	9	1.1	0.9;
	3101	1	0	8.4	0	15.72	0	1.024102	-28.059188	150	9	1.1	0.9;
	3102	1	125.2	105.4	0	76.73	0	1.030087	11.006241	380	3	1.1	0.9;
	3103	1	-0	-0	0.06	11.38	0	1.058579	-23.210274	220	21	1.1	0.9;
	3104	1	-0	0	0.11	2.05	0	1.034098	-26.275271	220	21	1.1	0.9;
	3105	1	2.3	0.8	0	0.8	0	1.014062	-36.909489	110	12	1.1	0.9;
	3106	1	30.63	11.6	0	52.28	0	1.048498	-45.743612	380	4	1.1	0.9;
	3107	1	-0	-0	0	2.65	0	1.039794	-3.280428	220	1	1.1	0.9;
	3108	1	-0	-0	0	0.39	0	1.015489	-30.06294	150	9	1.1	0.9;
	3109	1	-0	0	0	8.33	0	1.016108	-43.007828	220	4	1.1	0.9;
	3110	1	-0.19	-73.31	0	2.36	0	1.052047	15.39943	380	3	1.1	0.9;
	3111	1	0	0	0	20.96	0	1.030818	-42.746869	220	4	1.1	0.9;
	3112	1	-0	0	0	0	0	1.051067	2.324898	220	5	1.1	0.9;
	3113	2	0	0	0	0	0	1.064513	-38.339894	380	4	1.1	0.9;
	3114	2	0	0	0	0	0	1.063117	-12.151193	220	5	1.1	0.9;
	3115	1	-0	0	0	113.7	0	1.082739	-15.213748	380	18	1.1	0.9;
	3116	1	57.86	70.5	0.495087	218.066972	0	1.026615	-0.863273	380	3	1.1	0.9;
	3117	1	0	0	0	0	0	1.046412	-22.334303	150	9	1.1	0.9;
	3118	1	0	-0	0	4.36	0	1.008613	36.846475	220	3	1.1	0.9;
	3119	1	99.9	20.1	0	17.41	0	1.023437	8.989266	150	8	1.1	0.9;
	3120	1	0	0	0.181549	15.300889	0	1.063948	-22.743998	220	18	1.1	0.9;
	3121	1	242.16	73.64	0	12.79	0	1.024619	-7.433164	220	5	1.1	0.9;
	3122	1	13.21	4.4	0	0.38	0	0.969759	-32.190343	110	21	1.1	0.9;
	3123	1	115.5	71.5	0	150.47	0	1.032457	-21.310663	380	6	1.1	0.9;
	3124	1	0	-0	0	3.13	0	1.036819	-23.850146	220	7	1.1	0.9;
	3125	1	-0	0	0	3.25	0	1.015575	-40.39929	220	4	1.1	0.9;
	3126	1	104.78	20.92	0	6.17	0	1.017916	-18.019007	220	5	1.1	0.9;
	3127	2	0	0	0	0	0	1.097365	54.206958	380	3	1.1	0.9;
	3128	1	2.5	0.5	0	0.8	0	0.989103	-20.887239	110	18	1.1	0.9;
	3129	1	109.5	22.8	0	7.06	0	0.929883	-30.326449	110	18	1.1	0.9;
	3130	2	0	0	0	0	0	1.038694	-20.722363	380	20	1.1	0.9;
	3131	1	3.5	1.4	0	0.83	0	0.969995	-34.069945	110	18	1.1	0.9;
	3132	1	-14.9	-4.6	0	0.14	0	0.995803	-20.666157	110	18	1.1	0.9;
	3133	2	0	0	0	0	0	1.05743	-9.623166	220	5	1.1	0.9;
	3134	2	0	0	0	0	0	1.034426	-16.117435	220	5	1.1	0.9;
	3135	2	0	0	0	0	0	1.03216	-23.565566	154	1	1.1	0.9;
	3136	1	7.61	2.5	0	1.2	0	1.018401	-29.010141	110	21	1.1	0.9;
	3137	1	0	-0	0	0.03	0	1.057467	-26.432569	220	5	1.1	0.9;
	3138	1	0	-0	0	0.86	0	1.013801	-26.606601	150	9	1.1	0.9;
	3139	1	0	0	0	0	0	1.023174	-26.038202	150	9	1.1	0.9;
	3140	1	13.9	1.9	0	2.11	0	1.040626	-21.081939	154	1	1.1	0.9;
	3141	1	20.72	6.3	0	0.37	0	1.048468	-25.314747	110	19	1.1	0.9;
	3142	1	100.5	37	0	14.95	0	1.069389	-22.616678	380	18	1.1	0.9;
	3143	1	91.7	13	0	2.25	0	1.034924	-25.024397	154	1	1.1	0.9;
	3144	1	19.82	7.2	0	0.36	0	1.024183	-25.753432	110	19	1.1	0.9;
	3145	1	-0	-0	0	24.909116	0	0.980241	-41.867779	380	5	1.1	0.9;
	3146	1	-0	0	0	45.39	0	1.125942	-22.906239	380	15	1.3	0.7;
	3147	2	0	0	0	0	0	1.063849	20.694081	380	13	1.3	0.7;
	3148	1	0	0	0	0	0	1.040818	3.328266	150	8	1.1	0.9;
	3149	1	80.09	-1.8	0	4.09	0	1.021578	19.129883	380	3	1.1	0.9;
	3150	2	0	0	0	0	0	1.010317	-48.620794	220	4	1.1	0.9;
	3151	1	8.31	3	0	0.87	0	1.045167	-22.631971	110	19	1.1	0.9;
	3152	2	0	0	0	0	0	1.069675	-20.967243	220	11	1.1	0.9;
	3153	1	68.68	19.1	0	0.1	0	0.997269	-20.527634	110	18	1.1	0.9;
	3154	1	-0	0	0	0.21	0	0.941134	-34.246158	150	9	1.1	0.9;
	3155	1	0	0	0	19.63	0	1.002319	-0.228436	220	3	1.1	0.9;
	3156	2	0	0	0	0	0	1.024848	-6.484438	154	1	1.1	0.9;
	3157	2	0	0	0	0	0	1.021742	-38.172815	220	4	1.1	0.9;
	3158	2	0	0	0	0	0	1.058123	63.8325	150	1	1.1	0.9;
	3159	1	136.7	19.3	0	37.5	0	1.032151	-20.96825	380	6	1.1	0.9;
	3160	1	10.61	51.2	0.203187	8.281355	0	1.046099	12.386083	380	3	1.1	0.9;
	3161	1	59.98	15.54	0	0.49	0	0.994233	-25.746094	110	18	1.1	0.9;
	3162	1	8.31	2.5	0	0.8	0	1.015917	-23.585985	110	21	1.1	0.9;
	3163	2	0	0	0	0	0	1.002654	-47.833185	220	4	1.1	0.9;
	3164	1	83.9	11.9	0	1.14	0	1.015087	-4.216597	154	1	1.1	0.9;
	3165	1	-16.5	3.6	0	1.54	0	1.018334	-32.142815	150	9	1.1	0.9;
	3166	1	184.91	32.8	0	8.35	0	1.037728	-17.809451	220	5	1.1	0.9;
	3167	1	11.11	3.6	0	3.36	0	1.026051	-34.57357	110	12	1.1	0.9;
	3168	1	-0	0	0	0.49	0	1.019	-22.376658	154	1	1.1	0.9;
	3169	1	0	0	0	75	0	1.032824	3.573787	150	8	1.1	0.9;
	3170	1	39.64	9.8	0	1.08	0	0.986731	-45.275791	110	22	1.1	0.9;
	3171	1	13.84	1.85	0	0.2	0	1.012664	-22.795931	110	19	1.1	0.9;
	3172	1	-0	-0	0	1.98	0	0.970959	-26.759528	150	9	1.1	0.9;
	3173	1	0	-0	0.22	21.48	0	1.039184	-21.542575	220	19	1.1	0.9;
	3174	1	292.8	114.2	0	22.01	0	1.025066	-45.160027	380	6	1.1	0.9;
	3175	1	0	0	0	1.58	0	1.036042	-29.785074	150	9	1.1	0.9;
	3176	1	132.9	40.5	0	6.92	0	1.00792	5.571006	220	3	1.1	0.9;
	3177	1	39.34	11.2	0	0.85	0	1.017462	7.133812	150	8	1.1	0.9;
	3178	1	180.2	46.4	0	9.7	0	1.034206	-45.723311	220	4	1.1	0.9;
	3179	1	23.63	7.8	0	0.62	0	1.027153	-34.371397	110	12	1.1	0.9;
	3180	1	0	-0	0	62.76	0	1.032148	-38.739468	380	6	1.1	0.9;
	3181	1	-0	0	0	0.54	0	0.944519	-32.754224	150	9	1.1	0.9;
	3182	1	31.94	12.6	0	10.67	0	1.031492	3.864128	150	8	1.1	0.9;
	3183	2	0	0	0	0	0	1.061538	-12.259951	220	5	1.1	0.9;
	3184	1	14.12	7.7	0	1.42	0	1.037191	-21.646471	220	5	1.1	0.9;
	3185	1	0	0	0	0	0	1.031803	-29.22026	150	9	1.1	0.9;
	3186	1	-20.92	2.5	0	16.42	0	1.017107	-33.798706	220	10	1.1	0.9;
	3187	1	107	34.7	0	1.54	0	1.045742	-13.029721	220	5	1.1	0.9;
	3188	1	75.58	30.3	0	34.66	0	1.022784	-43.114628	220	6	1.1	0.9;
	3189	2	0	0	0	0	0	1.027406	47.583401	220	3	1.1	0.9;
	3190	1	0	-0	0	35.14	0	1.03242	13.30251	380	8	1.1	0.9;
	3191	1	127.8	15.2	0	9.34	0	1.034972	-12.379308	220	5	1.1	0.9;
	3192	1	-0	0	0.370251	146.807945	0	1.084468	-17.766436	380	18	1.3	0.7;
	3193	1	0	0	0	207.85	0	1.065116	-43.742608	380	4	1.1	0.9;
	3194	1	-0	-0	0	54.23	0	1.012434	9.719434	400	1	1.1	0.9;
	3195	1	104.5	31.8	0	0.71	0	1.006576	-37.208653	220	6	1.1	0.9;
	3196	1	398	11.5	0	7.749841	0	1.03574	-34.436191	380	6	1.1	0.9;
	3197	1	15.92	12	0	6.79	0	0.978533	-13.888229	220	16	1.1	0.9;
	3198	2	0	0	0	0	0	1.008016	-37.94449	220	22	1.1	0.9;
	3199	1	23.8	3.4	0	6.09	0	1.038944	1.077782	154	1	1.1	0.9;
	3200	1	86.2	19.9	0	5.43	0	1.032243	-23.40526	220	5	1.1	0.9;
	3201	1	-0	0	0	0.19	0	0.964845	-31.486373	150	9	1.1	0.9;
	3202	1	0.2	0.9	0	0.04	0	1.033191	24.896503	220	3	1.1	0.9;
	3203	1	27.03	9	0	0.98	0	1.028876	-21.014885	110	21	1.1	0.9;
	3204	1	8.51	3.2	0	0.96	0	1.015199	-10.325109	220	5	1.1	0.9;
	3205	2	0	0	0	0	0	1.043987	-1.453156	380	5	1.1	0.9;
	3206	1	146.7	20.7	0	62.64	0	1.033915	9.697252	150	8	1.1	0.9;
	3207	1	98.01	16.6	0	0	0	0.938219	-27.443062	110	18	1.1	0.9;
	3208	1	0	-0	0	104.11	0	1.082871	-28.948201	380	22	1.1	0.9;
	3209	1	2.3	0.8	0	0.42	0	1.004125	-41.22604	110	12	1.1	0.9;
	3210	1	-0	-0	0	47.99	0	1.025177	6.395982	380	8	1.1	0.9;
	3211	1	0	0	0	5.21	0	1.003983	-44.924447	220	10	1.1	0.9;
	3212	1	62.7	9	0	3.22	0	1.022821	-20.853163	154	1	1.1	0.9;
	3213	1	0	0	0.181549	5.810889	0	1.029705	-26.057232	220	18	1.1	0.9;
	3214	1	0	-0	0	1.43	0	0.97262	-27.178609	150	9	1.1	0.9;
	3215	1	63.07	25.8	0	18.63	0	1.019913	-4.168894	150	8	1.1	0.9;
	3216	1	73.58	30.1	0	6.02	0	1.018799	-4.292908	150	8	1.1	0.9;
	3217	1	64	9.1	0	12.23	0	1.024067	11.753871	154	1	1.1	0.9;
	3218	2	0	0	0	0	0	1.02981	-11.281302	220	5	1.1	0.9;
	3219	2	0	0	0	0	0	1.06035	8.785773	220	13	1.1	0.9;
	3220	1	-0	-0	0	24.67	0	1.036268	14.95352	400	1	1.1	0.9;
	3221	1	76.38	24.8	0	3.24	0	1.05011	-12.277426	220	5	1.1	0.9;
	3222	1	9.01	1.1	0	1.26	0	1.012228	-40.450912	110	12	1.1	0.9;
	3223	1	0	-0	0	0.6	0	1.001615	-29.037021	150	9	1.1	0.9;
	3224	1	0	-0	0	1.86	0	1.038864	27.61637	110	8	1.1	0.9;
	3225	1	81.2	89.3	0	15.55	0	1.038338	12.206547	380	3	1.1	0.9;
	3226	1	77.7	11.1	0	4.49	0	1.01262	-8.309205	154	1	1.1	0.9;
	3227	2	0	0	0	0	0	1.051132	-28.487653	380	6	1.1	0.9;
	3228	1	-0	0	0	3.58	0	1.026045	-13.340911	220	6	1.1	0.9;
	3229	2	0	0	0	0	0	1.034116	-13.165635	380	6	1.1	0.9;
	3230	1	18.5	2.7	0	0.11	0	1.024432	5.586276	154	1	1.1	0.9;
	3231	1	251.3	42.6	0	5.47	0	1.019628	-14.638597	220	5	1.1	0.9;
	3232	1	133.9	32.4	0	0	0	1.002743	-30.117477	110	21	1.1	0.9;
	3233	2	0	0	0	0	0	1.006814	-46.444024	220	4	1.1	0.9;
	3234	2	0	0	0	0	0	1.034756	-24.800598	220	6	1.1	0.9;
	3235	1	-6.58	-1.67	0	1.14	0	1.022894	-19.896997	110	18	1.1	0.9;
	3236	2	0	0	0	0	0	1.022886	-31.356422	220	12	1.1	0.9;
	3237	1	16.42	5.9	0	0.17	0	1.028917	-26.001768	110	19	1.1	0.9;
	3238	1	0	-0	0	6.19	0	1.036264	-36.34369	220	6	1.1	0.9;
	3239	1	-5.06	-1.84	0	187.327025	0	1.052583	-11.882218	380	5	1.1	0.9;
	3240	2	0	0	0	0	0	1.048993	-1.195378	220	5	1.1	0.9;
	3241	1	132	27	0	4.66	0	1.027307	-18.073023	220	5	1.1	0.9;
	3242	1	296.1	82.3	0	0	0	1.052842	-9.770395	150	2	1.1	0.9;
	3243	1	-0	0	0	1.18	0	0.985966	-18.490962	380	5	1.1	0.9;
	3244	1	0	0	0	0	0	0.963858	-26.99749	150	9	1.1	0.9;
	3245	1	0	0	0	17.82055	0	1.020645	-24.080506	220	20	1.1	0.9;
	3246	1	20.9	-171.5	0	0.21	0	1.075945	7.839523	220	5	1.1	0.9;
	3247	1	113	56.5	0	4.12	0	0.992233	-34.753986	150	9	1.1	0.9;
	3248	1	84.19	17.3	0	6.2	0	1.050664	-11.63598	220	5	1.1	0.9;
	3249	1	-0	-0	0	17.65	0	1.025749	-0.319615	380	3	1.1	0.9;
	3250	1	26.33	7.6	0	0.32	0	1.006298	-19.586545	110	18	1.1	0.9;
	3251	2	0	0	0	0	0	1.085025	23.509451	220	13	1.1	0.9;
	3252	1	0	0	0	0	0	0.984433	-23.606104	150	9	1.1	0.9;
	3253	2	0	0	0	0	0	1.024777	-6.321564	154	1	1.1	0.9;
	3254	1	-0	-0	0	45.7	0	1.024755	-16.181084	380	3	1.1	0.9;
	3255	1	0	0	0	1.44	0	1.034827	-22.379621	220	5	1.1	0.9;
	3256	1	-0	0	0	1.87	0	1.036534	-51.990546	380	10	1.1	0.9;
	3257	1	0	-0	0	3.71	0	1.047393	12.895607	220	8	1.1	0.9;
	3258	1	-11.71	3.8	0	0.13	0	0.995506	-31.841795	150	9	1.1	0.9;
	3259	1	-0	0	0	53.64	0	1.142893	-23.3897	380	15	1.3	0.7;
	3260	1	83.6	11.9	0	5.03	0	1.039685	3.822592	154	1	1.1	0.9;
	3261	1	87	29.6	0	4.95	0	1.001078	8.535194	220	3	1.1	0.9;
	3262	1	-0	-0	0.2166	10.9704	0	1.126522	-25.245392	380	1	1.3	0.7;
	3263	2	0	0	0	0	0	1.036588	-27.733931	110	19	1.1	0.9;
	3264	2	0	0	0	0	0	1.039352	17.354571	380	3	1.1	0.9;
	3265	1	0	-0	0	61.52	0	1.030307	-51.794533	380	10	1.1	0.9;
	3266	1	-0	0	0	4.62	0	0.984933	-54.747739	220	10	1.1	0.9;
	3267	1	-0	0	0	21.82	0	1.03073	20.597426	380	3	1.1	0.9;
	3268	1	-0	-0	0	29.27	0	1.038384	-40.013191	380	4	1.1	0.9;
	3269	1	115.6	16.4	0	2.41	0	1.024852	-6.895393	154	1	1.1	0.9;
	3270	1	0	0	0	5.12	0	1.012867	3.04441	380	3	1.1	0.9;
	3271	1	-0	-0	0	2.13	0	1.038289	3.97828	154	1	1.1	0.9;
	3272	1	-30.03	9.9	0	0	0	0.970883	-24.101597	150	9	1.1	0.9;
	3273	1	-0	-0	0	0.32	0	0.982449	-28.284052	110	18	1.1	0.9;
	3274	1	12.11	5.5	0	1.19	0	1.020263	-34.990644	110	22	1.1	0.9;
	3275	1	-0	0	0	17.75	0	1.042383	-43.384878	380	4	1.1	0.9;
	3276	1	-0	-0	0	0.13	0	1.047364	-7.951911	220	5	1.1	0.9;
	3277	1	0	0	0.294847	40.776932	0	1.045909	-16.489223	380	7	1.1	0.9;
	3278	2	0	0	0	0	0	1.024557	-28.170285	220	7	1.1	0.9;
	3279	1	0	0	0	0	0	1.00372	-3.125253	150	2	1.1	0.9;
	3280	1	-0	0	0	2.08	0	1.023776	-3.631665	150	8	1.1	0.9;
	3281	1	10.61	3.8	0	0.66	0	1.003133	-15.881363	110	19	1.1	0.9;
	3282	1	-0	0	0	9.66	0	1.015198	-57.678557	220	4	1.1	0.9;
	3283	1	5.5	0.8	0	1.19	0	1.027874	0.721469	154	1	1.1	0.9;
	3284	2	0	0	0	0	0	1.031123	-30.983299	380	6	1.1	0.9;
	3285	2	0	0	0	0	0	1.04544	-27.002752	380	6	1.1	0.9;
	3286	1	0	0	0	5.67	0	1.0048	-16.394162	220	1	1.1	0.9;
	3287	1	27.8	4	0	9.38	0	1.030245	11.41359	154	1	1.1	0.9;
	3288	1	-0.57	-0.91	0	0.09	0	1.053315	-24.416702	220	21	1.1	0.9;
	3289	1	11.01	6	0	0.04	0	0.995682	-31.065574	150	9	1.1	0.9;
	3290	2	0	0	0	0	0	1.076876	7.965584	380	14	1.1	0.9;
	3291	2	0	0	0	0	0	1.020441	-44.798684	220	4	1.1	0.9;
	3292	2	0	0	0	0	0	0.995766	3.120373	400	1	1.1	0.9;
	3293	1	33.14	11.6	0	4.51	0	1.033298	-11.661741	220	5	1.1	0.9;
	3294	1	5.51	2	0	0.39	0	1.035991	-4.951672	110	19	1.1	0.9;
	3295	1	-0.01	0	0	30.55	0	1.082067	-35.999712	150	9	1.1	0.9;
	3296	1	-0	0	0	1.08	0	1.044514	-28.223768	150	9	1.1	0.9;
	3297	1	-0	-0	0	0.8	0	0.996432	-30.704621	150	9	1.1	0.9;
	3298	2	0	0	0	0	0	1.06827	-26.211418	150	9	1.1	0.9;
	3299	1	332.9	99.6	0	0.53	0	1.017027	7.501874	150	8	1.1	0.9;
	3300	1	12.01	3.9	0	0.94	0	1.002796	-30.022842	110	21	1.1	0.9;
	3301	1	72.08	29.5	0	1.06	0	1.023716	-3.389174	150	8	1.1	0.9;
	3302	1	34.9	5	0	11.17	0	1.024008	7.979243	154	1	1.1	0.9;
	3303	1	0	0	0	3.25	0	1.026201	-10.658019	220	3	1.1	0.9;
	3304	2	0	0	0	0	0	1.024949	13.939543	400	1	1.1	0.9;
	3305	1	13.9	1.9	0	3.4	0	1.00183	-27.474238	154	1	1.1	0.9;
	3306	2	0	0	0	0	0	1.065504	8.070469	220	5	1.1	0.9;
	3307	1	-0	-0	0	1.84	0	0.991379	-30.823779	150	9	1.1	0.9;
	3308	1	-0	-0	0	7.75	0	1.049926	-19.383964	380	1	1.1	0.9;
	3309	1	32.64	11.8	0	2.07	0	1.008616	-28.806845	110	19	1.1	0.9;
	3310	2	0	0	0	0	0	1.032355	-10.130263	110	18	1.1	0.9;
	3311	1	19.82	7.3	0	0.25	0	1.023597	-37.693898	110	22	1.1	0.9;
	3312	1	-10.11	3.3	0	0.53	0	1.030866	-22.122041	150	9	1.1	0.9;
	3313	1	0	-0	0	34.49	0	1.04533	-15.776359	380	1	1.1	0.9;
	3314	1	-0.2	-1.64	0	0	0	1.03099	3.60651	150	8	1.1	0.9;
	3315	1	-0	0	0	0.05	0	1.089019	-27.770812	150	9	1.1	0.9;
	3316	1	50.56	16.6	0	2.02	0	1.000662	-36.968905	110	12	1.1	0.9;
	3317	1	89.5	12.8	0	5.08	0	1.017587	1.50211	154	1	1.1	0.9;
	3318	1	0	0	0	0	0	1.009022	-33.106438	150	9	1.1	0.9;
	3319	1	120.4	39.5	0	47.2	0	1.01871	-16.053033	380	3	1.1	0.9;
	3320	1	-0	-0	0	0.83	0	1.054266	-27.699221	150	9	1.1	0.9;
	3321	1	-0	-0	0	2.22	0	1.023353	11.967133	220	3	1.1	0.9;
	3322	2	0	0	0	0	0	1.029763	65.225449	150	1	1.1	0.9;
	3323	2	0	0	0	0	0	1.021399	14.587549	154	1	1.1	0.9;
	3324	2	0	0	0	0	0	1.012342	-16.331122	220	5	1.1	0.9;
	3325	1	-0	0	0	0.91	0	1.030173	-20.370249	220	5	1.1	0.9;
	3326	2	0	0	0	0	0	1.057931	6.3407	220	13	1.1	0.9;
	3327	1	-0	-0	0	2.27	0	1.040431	-12.287145	220	5	1.1	0.9;
	3328	1	14.72	4.6	0	0.33	0	1.002664	-24.296768	110	18	1.1	0.9;
	3329	1	2.3	0.8	0	0.6	0	0.977574	-22.100659	110	19	1.1	0.9;
	3330	2	0	0	0	0	0	1.049026	64.390308	220	3	1.1	0.9;
	3331	1	-0	-0	0	4.56	0	1.015678	-10.273718	220	5	1.1	0.9;
	3332	1	-10.59	-5.25	0	3.59	0	1.026477	38.071701	220	3	1.1	0.9;
	3333	1	34.04	11	0	0.96	0	1.040908	-39.939697	110	22	1.1	0.9;
	3334	2	0	0	0	0	0	1.027102	-38.773472	220	4	1.1	0.9;
	3335	2	0	0	0	0	0	1.08064	31.725381	380	8	1.1	0.9;
	3336	2	0	0	0	0	0	1.010354	-5.899949	154	1	1.1	0.9;
	3337	1	236.2	116.4	0	16.22	0	0.996896	-35.038392	220	6	1.1	0.9;
	3338	1	14.72	6.7	0	0.32	0	0.986716	-27.147128	110	18	1.1	0.9;
	3339	1	0	-0	0	0.38	0	0.988637	-30.966256	150	9	1.1	0.9;
	3340	2	0	0	0	0	0	1.027329	-10.387112	220	14	1.1	0.9;
	3341	1	46.85	16.6	0	2.08	0	0.992929	-43.712078	150	10	1.1	0.9;
	3342	1	76.18	13.2	0	7.4	0	1.025535	-40.505398	220	4	1.1	0.9;
	3343	1	379.9	73	0	59.58	0	1.029529	-45.122723	380	6	1.1	0.9;
	3344	1	83.29	24.1	0	1.23	0	1.041765	-16.618277	220	5	1.1	0.9;
	3345	2	0	0	0	0	0	0.998722	-23.729182	154	1	1.1	0.9;
	3346	2	0	0	0	0	0	1.014446	-12.166453	380	5	1.1	0.9;
	3347	1	38.04	20	0	0.03	0	1.026132	-22.241868	220	6	1.1	0.9;
	3348	1	-0	0	0	0.68	0	1.028993	-31.533375	380	6	1.1	0.9;
	3349	2	0	0	0	0	0	1.047992	65.933266	380	3	1.1	0.9;
	3350	1	1.3	0.4	0	0.24	0	0.985794	-28.367567	110	18	1.1	0.9;
	3351	1	129.5	51.8	0	107.08	0	1.039446	-27.195467	380	6	1.1	0.9;
	3352	1	-0	-0	0	5.22	0	1.024972	11.353365	380	8	1.1	0.9;
	3353	2	0	0	0	0	0	1.045086	-3.111681	220	5	1.1	0.9;
	3354	1	0	0	0	0	0	1.023094	-32.476655	150	9	1.1	0.9;
	3355	1	0	-0	0	5.14	0	1.031224	-4.244743	220	1	1.1	0.9;
	3356	1	-0	-0	0.185125	56.232473	0	1.075263	-18.49023	380	18	1.1	0.9;
	3357	1	36.54	27.4	0	0.16	0	0.976332	13.454371	220	3	1.1	0.9;
	3358	1	0	0	0	0	0	1.024826	-25.076779	110	3	1.1	0.9;
	3359	1	-0	-0	0	0.82	0	1.002922	-27.79104	110	18	1.1	0.9;
	3360	1	-0	-0	0	6.1	0	1.045789	-27.19641	380	6	1.1	0.9;
	3361	1	-0	-0	0	7.68	0	1.023311	4.22218	380	3	1.1	0.9;
	3362	2	0	0	0	0	0	1.060735	-19.68938	220	18	1.1	0.9;
	3363	1	155.8	38.4	0	3.77	0	1.022161	1.431179	150	8	1.1	0.9;
	3364	2	0	0	0	0	0	1.059596	-12.778615	220	5	1.1	0.9;
	3365	1	-0	-0	0.090774	0.080444	0	1.063179	-22.825995	220	18	1.1	0.9;
	3366	1	3.8	1.3	0	0.99	0	0.986109	-26.796119	110	18	1.1	0.9;
	3367	1	57.46	15.3	0	4.37	0	1.018093	-52.831262	220	4	1.1	0.9;
	3368	1	143.79	106.96	0	99.84	0	1.043656	-8.471313	380	13	1.1	0.9;
	3369	1	3.4	1.4	0	0.62	0	0.994634	-25.281268	110	18	1.1	0.9;
	3370	1	-0	-0	0	0	0	1.047845	18.20106	380	8	1.1	0.9;
	3371	2	0	0	0	0	0	1.029674	-35.973456	220	22	1.1	0.9;
	3372	1	0	-0	0.13	10.79	0	1.021707	-27.973729	220	21	1.1	0.9;
	3373	1	0.5	0.5	0	0.74	0	1.010616	-20.118719	110	18	1.1	0.9;
	3374	1	156.2	24.5	0	2.22	0	1.019272	9.870343	150	8	1.1	0.9;
	3375	2	0	0	0	0	0	1.035048	2.179444	220	14	1.1	0.9;
	3376	1	-0	0	0	0.4	0	1.020937	-22.652943	150	9	1.1	0.9;
	3377	1	77.49	21.4	0	0.05	0	1.016074	-10.56574	220	5	1.1	0.9;
	3378	1	-0	-0	0	13.02	0	1.016937	-45.491741	220	4	1.1	0.9;
	3379	1	13.9	1.9	0	5.68	0	1.039815	2.431643	154	1	1.1	0.9;
	3380	1	50.06	19.7	0	0.84	0	1.055018	-27.471518	110	23	1.1	0.9;
	3381	1	8.31	3	0	0.69	0	0.979142	-23.204984	110	19	1.1	0.9;
	3382	1	115	-4.5	0	15.05	0	1.021675	-47.230805	220	4	1.1	0.9;
	3383	1	47.15	15.2	0	0.17	0	0.991029	-36.657183	110	12	1.1	0.9;
	3384	1	0	0	0	0.23	0	1.009327	-46.581744	220	6	1.1	0.9;
	3385	1	-0	0	0	0.52	0	0.920065	-35.658325	150	9	1.1	0.9;
	3386	1	82.6	11.7	0	1.75	0	1.019186	-21.648748	154	1	1.1	0.9;
	3387	1	4.61	1.7	0	0.19	0	1.006574	-28.829212	110	19	1.1	0.9;
	3388	1	-0.2	-1.64	0	0	0	1.03099	3.60651	150	8	1.1	0.9;
	3389	2	0	0	0	0	0	1.037252	-4.035324	220	14	1.1	0.9;
	3390	2	0	0	0	0	0	1.035344	-11.772404	220	5	1.1	0.9;
	3391	1	302.5	65.1	0	8.87	0	1.054876	-21.249752	220	5	1.1	0.9;
	3392	1	-0	0	0	0.67	0	0.997163	-32.412159	150	9	1.1	0.9;
	3393	1	15.57	9.59	0	0.41	0	1.019427	-14.361209	110	18	1.1	0.9;
	3394	1	27.13	9	0	0.42	0	1.015848	-29.529341	110	21	1.1	0.9;
	3395	1	-0	-0	0	0.54	0	0.99554	-31.892503	150	9	1.1	0.9;
	3396	2	0	0	0	0	0	1.095689	53.980552	380	3	1.1	0.9;
	3397	1	24.12	9.8	0	0.04	0	1.017014	-0.831895	150	8	1.1	0.9;
	3398	1	-0.75	-0.93	0	0.26	0	1.036168	-33.730561	110	12	1.1	0.9;
	3399	1	50.06	16	0	0.65	0	1.000784	-30.931974	110	21	1.1	0.9;
	3400	1	0	0	0	17.31	0	1.043188	-25.417287	220	5	1.1	0.9;
	3401	1	23.22	6.3	0	0.66	0	1.042909	-13.901958	220	5	1.1	0.9;
	3402	1	115	16.4	0	3.94	0	1.015139	-21.841008	154	1	1.1	0.9;
	3403	1	234.2	33.3	0	0.88	0	1.013126	-22.985262	154	1	1.1	0.9;
	3404	1	140.93	18.71	0	12.73	0	1.037431	-0.674629	220	13	1.1	0.9;
	3405	1	-0.76	-0.86	0	6.15	0	1.037782	-32.221968	220	12	1.1	0.9;
	3406	1	0	-0	0	0.29	0	1.000105	-30.932302	150	9	1.1	0.9;
	3407	1	108.7	26.3	0	2.06	0	1.025102	-23.44681	220	7	1.1	0.9;
	3408	1	0	0	0	0	0	0.981792	-34.919816	150	9	1.1	0.9;
	3409	1	-0	-0	0	16.4	0	1.020151	-38.993027	380	1	1.1	0.9;
	3410	1	52.06	17	0	0.24	0	1.013843	-29.711149	110	21	1.1	0.9;
	3411	2	0	0	0	0	0	1.024699	-7.29082	220	13	1.1	0.9;
	3412	1	-0	0	0	0	0	1.051067	2.324898	220	5	1.1	0.9;
	3413	1	19.42	7	0	0.6	0	0.985559	-48.852329	150	10	1.1	0.9;
	3414	1	78.09	27	0	0.83	0	1.028494	-40.4823	110	22	1.1	0.9;
	3415	1	14.72	5.3	0	0.58	0	1.017252	-26.562376	110	19	1.1	0.9;
	3416	2	0	0	0	0	0	0.99364	-37.271039	220	10	1.1	0.9;
	3417	1	0	0	0	1.78	0	0.98196	-33.257434	150	9	1.1	0.9;
	3418	1	-0	-0	0	0.24	0	1.055201	-27.39433	150	9	1.1	0.9;
	3419	1	0	-0	0	0.28	0	1.05343	-28.763135	110	3	1.1	0.9;
	3420	1	6.71	2.4	0	0.2	0	1.035139	-24.267368	110	19	1.1	0.9;
	3421	1	11.51	4.2	0	3.64	0	1.063078	-18.313704	110	19	1.1	0.9;
	3422	2	0	0	0	0	0	1.068276	-11.121919	220	5	1.1	0.9;
	3423	1	173.1	4.6	0	2.57	0	1.026069	-18.887132	220	3	1.1	0.9;
	3424	1	11.5	1.7	0	3.11	0	1.007957	-6.106231	154	1	1.1	0.9;
	3425	1	83.59	21.6	0	25.51	0	1.023755	-54.374613	220	4	1.1	0.9;
	3426	2	0	0	0	0	0	1.01782	2.146836	154	1	1.1	0.9;
	3427	1	146.9	26.6	0	11.95	0	1.02442	-54.181061	220	4	1.1	0.9;
	3428	1	-0	-0	0	17.92	0	1.03688	-45.000213	380	4	1.1	0.9;
	3429	1	175.4	28.8	0	2.03	0	0.980011	-57.065563	150	10	1.1	0.9;
	3430	1	125.1	44	0	6.04	0	1.025197	-16.439994	220	5	1.1	0.9;
	3431	2	0	0	0	0	0	1.037857	65.155592	220	3	1.1	0.9;
	3432	1	39.05	-11.82	0	1.12	0	1.042952	-17.524884	220	7	1.1	0.9;
	3433	1	-0	0	0.121066	38.317958	0	1.030669	-22.850681	380	7	1.1	0.9;
	3434	1	4.2	2	0	0.42	0	0.995667	-26.3926	110	18	1.1	0.9;
	3435	1	-0	0	0	10.13	0	1.039221	-25.891769	220	5	1.1	0.9;
	3436	2	0	0	0	0	0	1.052704	-13.695663	220	5	1.1	0.9;
	3437	1	0	-0	0	0.04	0	1.042946	-20.458651	110	19	1.1	0.9;
	3438	2	0	0	0	0	0	1.048328	2.09489	220	13	1.1	0.9;
	3439	1	9.7	1.4	0	4.51	0	1.020047	-5.087122	154	1	1.1	0.9;
	3440	1	0	0	0	80.21	0	1.075387	-16.776668	380	19	1.1	0.9;
	3441	2	0	0	0	0	0	1.059719	-16.905097	220	18	1.1	0.9;
	3442	1	54.96	12.2	0	7.05	0	1.028465	-53.809916	220	4	1.1	0.9;
	3443	1	-0	0	0	0.74	0	1.017983	6.449833	150	8	1.1	0.9;
	3444	1	5.11	1.3	0	0.67	0	0.979988	-28.556612	110	18	1.1	0.9;
	3445	1	-7.63	-4.3	0	94.00001	0	0.992908	-6.161579	380	5	1.1	0.9;
	3446	1	271.43	74.55	0	78.4	0	1.054592	-7.033609	380	14	1.1	0.9;
	3447	1	120.1	30.5	0	5.26	0	1.017418	-41.092897	220	17	1.1	0.9;
	3448	1	-15.8	-2.8	0	7.6	0	1.067993	-27.916992	150	9	1.1	0.9;
	3449	1	0	-0	0	25.28	0	1.03469	-24.748184	380	6	1.1	0.9;
	3450	1	78.99	16.2	0	1.97	0	1.043402	-21.780648	220	5	1.1	0.9;
	3451	1	20.82	9.5	0	1.16	0	0.959432	-22.765455	110	18	1.1	0.9;
	3452	1	15.63	4.8	0	1.21	0	0.95362	-29.988893	110	18	1.1	0.9;
	3453	1	3.9	0.9	0	0.3	0	0.985273	-26.713246	110	18	1.1	0.9;
	3454	1	0	0	0	33.22	0	1.03762	-5.743891	380	13	1.1	0.9;
	3455	1	-18.8	-5.9	0	6.23	0	1.034721	-28.713871	150	9	1.1	0.9;
	3456	1	-0	0	0	2.09	0	1.023885	-3.618127	150	8	1.1	0.9;
	3457	1	23.13	8.8	0	1.42	0	1.027908	-40.515859	110	22	1.1	0.9;
	3458	1	4.2	0.6	0	4.02	0	1.033989	12.131257	154	1	1.1	0.9;
	3459	1	76.6	10.9	0	9.96	0	1.022605	-5.992828	154	1	1.1	0.9;
	3460	1	0	-0	0	1.38	0	1.04269	-24.913869	150	9	1.1	0.9;
	3461	1	119.1	60.5	0	2	0	1.013524	-32.326805	150	9	1.1	0.9;
	3462	1	209.5	95	0	24.39	0	1.023996	11.110517	150	8	1.1	0.9;
	3463	1	0	0	0	0.69	0	0.997198	-30.597829	150	9	1.1	0.9;
	3464	1	-0	-0	0	7.49	0	1.055563	-19.124141	400	1	1.1	0.9;
	3465	1	119.1	22.3	0	2.71	0	1.001856	-50.899311	220	6	1.1	0.9;
	3466	1	7.71	2.2	0	0.81	0	1.020739	-34.936213	110	22	1.1	0.9;
	3467	1	1.5	0.5	0	0.59	0	1.033755	-34.258331	110	12	1.1	0.9;
	3468	1	50.36	-36.2	0	1.25	0	1.026246	-48.014875	220	4	1.1	0.9;
	3469	1	0	-0	0	0.27	0	1.040633	-32.329218	110	1	1.1	0.9;
	3470	1	123.1	51.8	0	1.68	0	1.017239	-27.181267	110	21	1.1	0.9;
	3471	1	18.72	5.5	0	1.15	0	1.037452	-33.497141	110	22	1.1	0.9;
	3472	1	18.02	-15.7	0	1.03	0	1.058466	-24.602689	110	23	1.1	0.9;
	3473	1	0	0	0	0	0	1.004888	-32.9331	150	9	1.1	0.9;
	3474	1	23.93	8.7	0	0.52	0	1.032567	-27.514987	110	19	1.1	0.9;
	3475	1	0	-0	0	64.67	0	1.016081	16.936639	400	1	1.1	0.9;
	3476	1	0	0	0	8.48	0	1.029567	33.682355	380	3	1.1	0.9;
	3477	1	22.12	7.9	0	0.7	0	1.043352	-18.411585	110	19	1.1	0.9;
	3478	1	-3.4	-0.5	0	0.23	0	1.004025	-44.918549	220	10	1.1	0.9;
	3479	1	124.9	104	0	4.31	0	0.995033	10.148104	220	3	1.1	0.9;
	3480	2	0	0	0	0	0	1.039565	-15.023882	220	6	1.1	0.9;
	3481	1	142.5	23.6	0	4.24	0	1.020131	-21.073223	220	5	1.1	0.9;
	3482	1	22.02	13.7	0	0.44	0	0.96568	-37.714763	150	9	1.1	0.9;
	3483	1	-0	-0	0	91.192543	0	1.057283	2.073747	380	5	1.1	0.9;
	3484	1	-0	0	0	37.1	0	1.03238	13.28195	380	8	1.1	0.9;
	3485	1	-0	0	0	6.501668	0	0.987214	-16.561475	380	5	1.1	0.9;
	3486	1	101.5	-18.3	0	8.47	0	1.034572	-21.143387	220	5	1.1	0.9;
	3487	1	16.32	7.6	0	3.15	0	1.008468	-40.53524	220	4	1.1	0.9;
	3488	1	55.66	4.5	0	1.33	0	1.033154	-12.683749	220	5	1.1	0.9;
	3489	1	0	-0	0	12.24	0	1.021883	-53.739345	220	4	1.1	0.9;
	3490	1	12.51	4.1	0	0.71	0	1.016214	-37.064491	110	12	1.1	0.9;
	3491	1	15.92	5.2	0	0.58	0	1.012	-39.093054	110	12	1.1	0.9;
	3492	2	0	0	0	0	0	1.049517	-8.753604	220	5	1.1	0.9;
	3493	1	226.3	61.1	0	0	0	1.030522	-10.855222	150	2	1.1	0.9;
	3494	1	0	0	0.230863	107.524848	0	1.040916	-20.181938	380	7	1.1	0.9;
	3495	1	36.34	3.3	0	2.16	0	1.011873	-57.860179	220	4	1.1	0.9;
	3496	1	99.3	23.3	0	1.88	0	1.031337	-21.679626	220	5	1.1	0.9;
	3497	1	100.5	32	0	1.89	0	1.032235	3.545389	150	8	1.1	0.9;
	3498	1	138.5	19.3	0	5.26	0	1.038411	-10.391747	220	5	1.1	0.9;
	3499	1	180.4	71.6	0	81.042938	0	1.031705	-3.679012	380	5	1.1	0.9;
	3500	1	94.2	21.6	0	2.88	0	1.027326	-52.891329	220	4	1.1	0.9;
	3501	1	33.5	4.8	0	3.47	0	0.997562	-8.529192	154	1	1.1	0.9;
	3502	1	98.2	19.1	0	10.92	0	1.018814	-10.609035	220	5	1.1	0.9;
	3503	1	-0	0	0	1.93	0	1.029182	-11.728946	220	5	1.1	0.9;
	3504	1	8.91	2.5	0	0.31	0	1.001149	-27.034513	110	18	1.1	0.9;
	3505	1	0	0	0	3.49	0	1.004322	-11.424973	220	19	1.1	0.9;
	3506	1	0	0	0	0.6	0	1.004098	-11.409418	150	2	1.1	0.9;
	3507	2	0	0	0	0	0	1.009668	-36.42353	220	6	1.1	0.9;
	3508	2	0	0	0	0	0	1.05417	-24.96694	380	11	1.1	0.9;
	3509	1	23.62	16.5	0	8.39	0	1.023115	11.11478	150	8	1.1	0.9;
	3510	1	0	-0	0	43.87	0	1.077225	-22.203902	380	18	1.1	0.9;
	3511	2	0	0	0	0	0	1.086282	-1.99271	220	18	1.1	0.9;
	3512	1	65.64	-0.98	0	11.55	0	1.00546	-53.369968	220	4	1.1	0.9;
	3513	2	0	0	0	0	0	1.027289	-3.805998	380	5	1.1	0.9;
	3514	1	-0.64	-3.53	0	8.84	0	1.070779	44.26994	380	3	1.1	0.9;
	3515	2	0	0	0	0	0	1.053923	-0.652883	220	13	1.1	0.9;
	3516	1	-0	0	0	3.48	0	1.067373	-22.730945	110	15	1.1	0.9;
	3517	1	-0	0	0	42.71	0	1.074675	5.430037	380	1	1.1	0.9;
	3518	2	0	0	0	0	0	1.006674	-10.001199	330	24	1.1	0.9;
	3519	1	-0	0	0	68.053762	0	1.031544	-23.261886	380	5	1.1	0.9;
	3520	1	-0	0	0	2.47	0	1.033136	-6.430079	220	5	1.1	0.9;
	3521	1	9.7	1.4	0	2.74	0	1.030845	11.444625	154	1	1.1	0.9;
	3522	1	-0	-0	0	13.79	0	1.024607	7.384057	380	3	1.1	0.9;
	3523	1	29.33	9.6	0	0.6	0	0.96895	-32.994795	110	21	1.1	0.9;
	3524	1	29.93	10.8	0	0.46	0	1.035659	-27.412126	110	19	1.1	0.9;
	3525	1	0	-0	0.040985	0.716371	0	1.025525	64.851641	220	1	1.1	0.9;
	3526	1	33.34	13.2	0	0.76	0	1.053035	-4.537809	220	5	1.1	0.9;
	3527	1	-0	0	0	0.9	0	0.989431	-26.529496	110	18	1.1	0.9;
	3528	1	0	0	0	0	0	1.044635	-29.457279	150	9	1.1	0.9;
	3529	1	-0	0	0	2.47	0	1.047492	-20.222834	220	1	1.1	0.9;
	3530	1	118.5	21	0	2.19	0	1.016575	0.400195	150	8	1.1	0.9;
	3531	1	7.81	2.8	0	0.77	0	1.019989	-27.855737	110	19	1.1	0.9;
	3532	1	0	0	0	0	0	0.916596	-35.844928	150	9	1.1	0.9;
	3533	1	159.1	-28.7	0	2.31	0	1.045069	-12.896252	220	3	1.1	0.9;
	3534	1	8.61	3.1	0	0.48	0	1.051225	-25.209197	110	19	1.1	0.9;
	3535	1	-0	0	0	0.49	0	1.030744	-24.81643	220	5	1.1	0.9;
	3536	1	-0.48	0.1	0	0.91	0	1.043359	-24.890743	220	18	1.1	0.9;
	3537	1	96.5	-43	0	0.91	0	1.033796	-10.932693	220	3	1.1	0.9;
	3538	1	0	0	0	4.17	0	1.025276	17.761025	220	8	1.1	0.9;
	3539	1	97	-52.1	0	3.95	0	1.024188	-11.815297	220	3	1.1	0.9;
	3540	1	6.01	3	0	0.05	0	1.013842	-36.127515	220	6	1.1	0.9;
	3541	1	0	-0	0	3.34	0	1.034174	-24.344455	220	5	1.1	0.9;
	3542	1	0	-0	0	0.98	0	0.996699	9.316876	220	3	1.1	0.9;
	3543	1	-0	-0	0	14.079026	0	0.987091	-17.860899	380	5	1.1	0.9;
	3544	1	0	-0	0	83.96	0	1.04866	-43.909021	380	4	1.1	0.9;
	3545	1	20.12	7.6	0	0.89	0	1.046427	-8.068997	220	5	1.1	0.9;
	3546	1	-0	-0	0	48.46	0	1.039138	-17.687503	380	6	1.1	0.9;
	3547	2	0	0	0	0	0	1.05946	-0.972444	110	19	1.1	0.9;
	3548	1	10.01	3	0	0.84	0	0.947688	-35.958443	110	21	1.1	0.9;
	3549	1	0	0	0	3.91	0	1.013673	-42.828833	220	4	1.1	0.9;
	3550	2	0	0	0	0	0	1.042925	-21.756874	220	7	1.1	0.9;
	3551	2	0	0	0	0	0	1.039557	3.294351	220	14	1.1	0.9;
	3552	1	-0	0	0	18.31	0	1.02243	21.250325	380	3	1.1	0.9;
	3553	1	0	0	0	0.25	0	0.98226	-23.625281	150	9	1.1	0.9;
	3554	1	362.2	-10.5	0	19.86	0	1.010725	-19.083646	220	6	1.1	0.9;
	3555	1	0	-0	0	1.33	0	1.016482	-26.249817	150	9	1.1	0.9;
	3556	1	31.23	12.9	0	0.36	0	1.03815	-49.637833	220	6	1.1	0.9;
	3557	1	87.2	31.8	0	5.96	0	1.031344	-31.203578	220	5	1.1	0.9;
	3558	1	-0	0	0	2.51	0	1.061361	2.327475	220	5	1.1	0.9;
	3559	1	4.1	0.9	0	0.09	0	1.006934	-30.186936	110	18	1.1	0.9;
	3560	1	47.95	17.2	0	7.91	0	1.002793	-35.403578	220	6	1.1	0.9;
	3561	1	60.67	37.6	0	0.92	0	1.008119	-31.216752	150	9	1.1	0.9;
	3562	1	6.31	1.3	0	0.43	0	0.930545	-32.781271	110	18	1.1	0.9;
	3563	1	392.8	70.6	0	13.309841	0	1.028862	-36.110427	380	6	1.1	0.9;
	3564	1	0	-0	0	5.34	0	1.034865	-2.725945	154	1	1.1	0.9;
	3565	2	0	0	0	0	0	1.079183	-26.904591	220	5	1.1	0.9;
	3566	1	5.21	1.5	0	1.1	0	0.95235	-30.885389	110	18	1.1	0.9;
	3567	1	0	0	0	0	0	0.95682	-37.631909	150	9	1.1	0.9;
	3568	1	30.6	4.3	0	8.4	0	1.043573	2.632597	154	1	1.1	0.9;
	3569	1	215.3	37.7	0	14.36	0	1.042101	-21.88812	220	7	1.1	0.9;
	3570	2	0	0	0	0	0	1.062522	-22.923792	220	18	1.1	0.9;
	3571	2	0	0	0	0	0	0.997978	-49.825907	220	6	1.1	0.9;
	3572	1	8.31	3	0	0.24	0	0.998955	-16.495251	110	19	1.1	0.9;
	3573	1	-0	-0	0	0	0	0.991174	-30.802996	150	9	1.1	0.9;
	3574	1	-0	0	0	32.63	0	1.031327	-31.908208	220	12	1.1	0.9;
	3575	1	24.43	8.7	0	0.46	0	1.031556	-27.321527	110	15	1.1	0.9;
	3576	1	1.4	0.5	0	0.85	0	1.074439	-16.993961	110	19	1.1	0.9;
	3577	1	-38.7	-12.99	0	13.47383	0	1.027066	-5.390772	380	5	1.1	0.9;
	3578	1	-0	0	0	11.06	0	1.037916	37.328866	380	3	1.1	0.9;
	3579	1	277.72	7.35	0	3.58	0	1.071688	-39.757905	220	5	1.1	0.9;
	3580	2	0	0	0	0	0	1.071709	-33.67965	220	5	1.1	0.9;
	3581	1	-10.61	3.5	0	1.3	0	0.957996	-31.878474	150	9	1.1	0.9;
	3582	2	0	0	0	0	0	1.040375	-49.699237	380	4	1.1	0.9;
	3583	2	0	0	0	0	0	1.041006	32.019238	380	3	1.1	0.9;
	3584	2	0	0	0	0	0	1.016191	-45.589535	220	4	1.1	0.9;
	3585	2	0	0	0	0	0	1.001933	-45.137255	220	10	1.1	0.9;
	3586	1	0	-0	0	0.32	0	1.03216	-26.069193	150	9	1.1	0.9;
	3587	1	0	-0	0	78.22	0	1.066334	-40.491176	380	4	1.1	0.9;
	3588	1	0	-0	0	1.6	0	0.973088	-32.109368	150	9	1.1	0.9;
	3589	1	37.54	18.3	0	0.22	0	1.036354	-6.941677	220	5	1.1	0.9;
	3590	2	0	0	0	0	0	1.014988	-6.451716	154	1	1.1	0.9;
	3591	1	13.9	1.9	0	4.47	0	1.022027	-12.309284	154	1	1.1	0.9;
	3592	1	102.7	77.5	0	4.89	0	1.008158	-49.718813	220	4	1.1	0.9;
	3593	1	84.09	52.1	0	0	0	1.013847	-32.766687	150	9	1.1	0.9;
	3594	1	31.43	6.6	0	3.32	0	1.031842	-20.054986	220	5	1.1	0.9;
	3595	1	14.02	7.6	0	0.33	0	1.00521	-17.829322	110	18	1.1	0.9;
	3596	1	86.7	34.6	0	2.73	0	0.995624	-51.624297	220	6	1.1	0.9;
	3597	1	0	-0	0	0.06	0	1.053304	-24.416894	220	21	1.1	0.9;
	3598	1	4.4	2.2	0	1.79	0	1.020221	-39.571645	110	22	1.1	0.9;
	3599	2	0	0	0	0	0	1.035573	-4.319991	220	14	1.1	0.9;
	3600	1	-11.01	3.6	0	0.55	0	1.024841	-21.89593	150	9	1.1	0.9;
	3601	1	0	0	0	0.23	0	1.037299	-21.668191	220	5	1.1	0.9;
	3602	1	86.6	14.9	0	1.34	0	1.028652	-19.205363	220	5	1.1	0.9;
	3603	1	0	-0	0	0.74	0	1.059036	-28.598785	150	9	1.1	0.9;
	3604	1	124.3	39.4	0	6.43	0	1.031503	-22.253073	380	20	1.1	0.9;
	3605	2	0	0	0	0	0	1.037656	-37.493839	220	6	1.1	0.9;
	3606	1	194.3	27.9	0	3.53	0	1.060971	42.771684	220	3	1.1	0.9;
	3607	1	95.1	-19.36	0	12.46	0	1.057813	11.781879	220	13	1.1	0.9;
	3608	1	0	-0	0	58.146813	0	1.005954	-41.940464	380	5	1.1	0.9;
	3609	1	-28.13	-0.82	0	0.1	0	1.065404	-11.8324	220	5	1.1	0.9;
	3610	1	185.8	31.8	0	3.73	0	1.032147	-23.09589	220	5	1.1	0.9;
	3611	2	0	0	0	0	0	1.0427	-18.977598	220	5	1.1	0.9;
	3612	1	-0	-0	0	4.48	0	1.03493	3.177325	154	1	1.1	0.9;
	3613	1	152.5	14.4	0	24.61	0	1.027552	-28.290396	380	5	1.1	0.9;
	3614	1	9.21	3.3	0	1.13	0	1.051765	-19.459008	110	19	1.1	0.9;
	3615	1	0	-0	0	13.51	0	1.020308	-38.803377	380	1	1.1	0.9;
	3616	1	108.43	31.33	0	0.88	0	1.027726	-1.864399	220	13	1.1	0.9;
	3617	1	81.79	51.1	0	3.06	0	0.981291	29.186876	220	3	1.1	0.9;
	3618	2	0	0	0	0	0	1.059129	-20.439167	110	15	1.1	0.9;
	3619	1	0	0	0	0	0	0.961409	-31.649013	150	9	1.1	0.9;
	3620	1	35.14	16.1	0	0.45	0	1.03722	1.667564	150	8	1.1	0.9;
	3621	2	0	0	0	0	0	1.018584	-52.604451	220	4	1.1	0.9;
	3622	1	-0	-0	0	53.84	0	1.098571	-31.865767	380	12	1.1	0.9;
	3623	1	0	-0	0	4.08	0	1.030825	-34.021264	220	10	1.1	0.9;
	3624	1	0	0	0	0	0	0.997079	-32.31357	150	9	1.1	0.9;
	3625	2	0	0	0	0	0	1.035196	3.481897	154	1	1.1	0.9;
	3626	1	13.62	8.4	0	2.67	0	1.036804	-25.936777	150	9	1.1	0.9;
	3627	1	-0	-0	0	0.39	0	1.031483	-29.235879	150	9	1.1	0.9;
	3628	2	0	0	0	0	0	1.011525	-19.030838	400	1	1.1	0.9;
	3629	2	0	0	0	0	0	1.019084	-44.573053	220	4	1.1	0.9;
	3630	1	21.52	7.6	0	0.18	0	0.952693	-30.456432	110	18	1.1	0.9;
	3631	1	-0	-0	0	11.98	0	1.059298	54.603492	380	3	1.1	0.9;
	3632	1	-12.3	-10.6	0	3.54	0	1.074155	-26.401369	150	9	1.1	0.9;
	3633	1	-0	0	0	15.23	0	1.028353	-55.614104	220	4	1.1	0.9;
	3634	2	0	0	0	0	0	1.058312	-28.162333	220	6	1.1	0.9;
	3635	2	0	0	0	0	0	1.027369	17.315007	154	1	1.1	0.9;
	3636	1	0	0	0	10.82	0	1.021807	16.231169	380	3	1.1	0.9;
	3637	1	23.93	5.8	0	0.64	0	1.044046	-26.789423	110	21	1.1	0.9;
	3638	2	0	0	0	0	0	1.027553	-9.145708	154	1	1.1	0.9;
	3639	1	1.7	0.3	0	0.21	0	0.9699	-29.797995	110	18	1.1	0.9;
	3640	1	0	-0	0	7.1	0	1.038862	-26.224708	220	18	1.1	0.9;
	3641	1	30.33	11	0	0.58	0	1.07845	-24.869934	110	19	1.1	0.9;
	3642	1	1.3	0.4	0	1.7	0	1.011194	-17.661283	110	18	1.1	0.9;
	3643	1	24.23	6.4	0	3.9	0	1.035538	-19.389636	220	5	1.1	0.9;
	3644	2	0	0	0	0	0	1.01566	-43.78835	220	4	1.1	0.9;
	3645	1	42.55	17.6	0	17.28	0	1.035645	-34.036565	220	5	1.1	0.9;
	3646	1	32.24	11.7	0	0.1	0	1.030587	-26.137737	110	19	1.1	0.9;
	3647	1	-668.3	-230.2	0	0.52	0	1.044476	33.28174	380	3	1.1	0.9;
	3648	1	-0	0	0	57.87	0	1.05487	4.822118	380	1	1.1	0.9;
	3649	1	-0	-0	0	4.86	0	1.061718	-38.073141	220	5	1.1	0.9;
	3650	1	0	-0	0	4.68	0	1.023313	-43.662861	220	4	1.1	0.9;
	3651	1	0	-0	0	11.87	0	1.075494	-25.869106	380	1	1.1	0.9;
	3652	1	61.4	8.8	0	3.53	0	1.008222	-7.892933	154	1	1.1	0.9;
	3653	1	11.5	6.16	0	1.79	0	1.06713	-27.981342	150	9	1.1	0.9;
	3654	1	20.72	5	0	6.64	0	1.068425	-26.791691	220	5	1.1	0.9;
	3655	1	114.8	121.5	0	24.25	0	1.018477	2.812602	380	3	1.1	0.9;
	3656	2	0	0	0	0	0	1.062208	-12.302586	220	5	1.1	0.9;
	3657	1	131	1.3	0	120.45	0	1.069887	-26.477929	220	5	1.1	0.9;
	3658	1	0	-0	0	2.8	0	1.055179	-27.394045	150	9	1.1	0.9;
	3659	2	0	0	0	0	0	1.057396	0.724854	380	2	1.1	0.9;
	3660	2	0	0	0	0	0	1.032448	4.126448	154	1	1.1	0.9;
	3661	2	0	0	0	0	0	1.055245	4.477021	380	5	1.1	0.9;
	3662	1	-0	-0	0	4.74	0	1.020061	-43.910876	380	4	1.1	0.9;
	3663	1	50.2	7.1	0	8.14	0	1.027886	0.182084	154	1	1.1	0.9;
	3664	1	-0.2	-0.96	0	0	0	1.020031	8.412929	150	8	1.1	0.9;
	3665	1	295.4	42	0	0.83	0	1.011358	-43.221397	220	4	1.1	0.9;
	3666	1	-0	0	0	0.74	0	1.03051	-29.291022	150	9	1.1	0.9;
	3667	1	-0	-0	0	0.53	0	0.978178	-27.174726	150	9	1.1	0.9;
	3668	1	-0	0	0	13.93	0	1.050201	9.308989	380	3	1.1	0.9;
	3669	1	-0	-0	0	2.04	0	1.04073	-27.873462	110	19	1.1	0.9;
	3670	1	28.03	6.5	0	6.32	0	1.031403	-22.326779	220	5	1.1	0.9;
	3671	1	22.83	9.7	0	5.07	0	0.925563	-30.71087	110	18	1.1	0.9;
	3672	1	82.99	25.9	0	0.23	0	1.027236	-13.187416	220	5	1.1	0.9;
	3673	2	0	0	0	0	0	1.011536	-39.38978	220	4	1.1	0.9;
	3674	1	155.05	39.56	0	6.27	0	1.038893	-14.934461	220	5	1.1	0.9;
	3675	1	-0	0	0	-0.73	0	1.017296	62.359062	150	1	1.1	0.9;
	3676	1	95	10.43	0	89.57	0	1.049019	-8.638156	380	13	1.1	0.9;
	3677	1	5.51	2.2	0	0.59	0	1.011531	-36.657808	110	22	1.1	0.9;
	3678	1	4.69	1.89	0	2.45	0	1.009653	-27.771537	150	9	1.1	0.9;
	3679	1	49.91	23.81	0	0.17	0	0.998124	-12.616636	110	18	1.1	0.9;
	3680	1	-0	-0	0	12.4	0	1.041782	-15.274034	220	5	1.1	0.9;
	3681	1	-0	0	0	0.02	0	0.917096	-35.817405	150	9	1.1	0.9;
	3682	1	2.5	1.3	0	0.12	0	0.968066	-28.685503	110	18	1.1	0.9;
	3683	1	28.93	10.5	0	0.88	0	1.011346	-25.944614	110	19	1.1	0.9;
	3684	1	182.3	65.9	0	62.49	0	1.076869	-24.59787	380	20	1.1	0.9;
	3685	1	52.79	2.64	0	0	0	0.938194	-27.473163	110	18	1.1	0.9;
	3686	1	-0	0.01	0	0	0	0.997509	-29.526848	150	9	1.1	0.9;
	3687	1	0	-0	0	70.71	0	1.04322	-50.454955	380	4	1.1	0.9;
	3688	1	-0	0	0.1	5.73	0	1.03526	-29.588837	220	21	1.1	0.9;
	3689	2	0	0	0	0	0	1.029361	-45.589399	220	4	1.1	0.9;
	3690	1	-17.62	5.8	0	0.23	0	1.018828	-29.278078	150	9	1.1	0.9;
	3691	2	0	0	0	0	0	1.030004	12.928231	150	8	1.1	0.9;
	3692	2	0	0	0	0	0	1.024383	-42.674931	220	4	1.1	0.9;
	3693	1	0	0	0	1	0	1.007682	3.259225	154	1	1.1	0.9;
	3694	1	16.4	2.3	0	3.43	0	1.013472	-5.932551	154	1	1.1	0.9;
	3695	1	-0	-0	0	0.56	0	1.0156	-28.615692	150	9	1.1	0.9;
	3696	2	0	0	0	0	0	1.041264	-30.043549	380	6	1.1	0.9;
	3697	1	-32.71	-13.76	0	135.136611	0	1.0464	-20.856428	380	5	1.1	0.9;
	3698	2	0	0	0	0	0	1.039409	-24.639936	220	5	1.1	0.9;
	3699	1	16.62	-10.3	0	0.1	0	1.068441	-48.37893	220	6	1.1	0.9;
	3700	1	-4.86	-2.12	0	0.64	0	1.001182	-24.476336	110	18	1.1	0.9;
	3701	1	-0	0	0	0.43	0	1.04365	-25.354476	220	5	1.1	0.9;
	3702	1	0	-0	0	13.3	0	1.040283	-21.72381	220	11	1.1	0.9;
	3703	1	0	0	0	1.53	0	1.019123	-22.652098	150	9	1.1	0.9;
	3704	1	72.68	45	0	6.61	0	0.991297	-30.713122	150	9	1.1	0.9;
	3705	1	120.7	24.3	0	4.51	0	1.015587	-36.522251	220	6	1.1	0.9;
	3706	1	2.5	-0	0	0.57	0	1.00948	-20.294045	110	18	1.1	0.9;
	3707	1	77.49	17.3	0	0.1	0	1.042205	-15.70797	220	5	1.1	0.9;
	3708	1	32.1	4.5	0	2.53	0	1.035938	3.727934	154	1	1.1	0.9;
	3709	1	-0	-0	0	1.2	0	1.020456	-5.269925	220	3	1.1	0.9;
	3710	1	0	0	0	7.4	0	1.040313	-26.941454	220	1	1.1	0.9;
	3711	2	0	0	0	0	0	0.999169	-39.203039	150	10	1.1	0.9;
	3712	1	-0.19	-0.87	0	0	0	1.027334	3.504304	150	8	1.1	0.9;
	3713	1	24.03	8	0	0.75	0	1.016072	-22.733185	110	21	1.1	0.9;
	3714	1	-2835.7	-818.9	0	56.48	0	1.04883	-8.994879	380	3	1.1	0.9;
	3715	1	68.78	13	0	10.45	0	1.021862	-48.907312	220	4	1.1	0.9;
	3716	1	85.1	-122.4	0	6.67	0	1.017439	-13.963996	380	3	1.1	0.9;
	3717	1	-0	-0	0	4.67	0	1.058098	-27.28061	220	1	1.1	0.9;
	3718	1	179.4	58.3	0	0.64	0	1.041371	-25.581584	220	5	1.1	0.9;
	3719	1	4.4	2.2	0	0.41	0	1.032431	-38.880183	110	22	1.1	0.9;
	3720	2	0	0	0	0	0	1.050724	-2.16522	220	13	1.1	0.9;
	3721	1	-0	0	0	0.87	0	1.01754	-47.965957	220	6	1.1	0.9;
	3722	1	65.1	26.4	0	0.58	0	1.015245	-30.961669	110	23	1.1	0.9;
	3723	2	0	0	0	0	0	1.048349	49.380362	380	8	1.1	0.9;
	3724	2	0	0	0	0	0	1.020045	-9.45686	220	3	1.1	0.9;
	3725	1	50.96	17	0	0.31	0	1.033935	-30.806474	110	21	1.1	0.9;
	3726	1	16.52	7.1	0	1.1	0	1.040597	-40.004223	110	22	1.1	0.9;
	3727	1	165.3	89.2	0	13.47	0	1.032763	7.097249	380	3	1.1	0.9;
	3728	1	67.77	10.4	0	7.18	0	1.028811	-53.237993	220	4	1.1	0.9;
	3729	1	-0	0	0.27075	43.13435	0	1.08063	-12.865076	380	18	1.1	0.9;
	3730	1	2.3	0.8	0	0.21	0	0.982534	-21.666771	110	19	1.1	0.9;
	3731	1	0	-0	0.090774	3.980444	0	1.045331	-16.015346	220	18	1.1	0.9;
	3732	1	-24.87	-3.7	0	0.35	0	1.02505	-13.957262	110	18	1.1	0.9;
	3733	1	-0	-0	0	0.11	0	1.007979	-23.863667	110	18	1.1	0.9;
	3734	1	51.36	-0.7	0	0.57	0	1.02085	-47.275913	220	4	1.1	0.9;
	3735	1	-0	-0	0	0.09	0	0.993103	-32.678683	150	9	1.1	0.9;
	3736	1	45.15	4.1	0	2.61	0	1.01133	-58.622336	220	4	1.1	0.9;
	3737	1	42.48	10.36	0	0.37	0	1.032515	-3.354012	220	5	1.1	0.9;
	3738	1	18.42	6	0	0.93	0	0.969277	-33.47506	110	21	1.1	0.9;
	3739	2	0	0	0	0	0	1.035011	49.103214	220	3	1.1	0.9;
	3740	1	0	-0	0	13.03	0	1.034825	-21.527104	220	5	1.1	0.9;
	3741	2	0	0	0	0	0	1.029075	0.423962	380	5	1.1	0.9;
	3742	1	122.3	45.9	0	4.03	0	1.029537	-22.781849	220	7	1.1	0.9;
	3743	2	0	0	0	0	0	1.026432	64.598966	150	1	1.1	0.9;
	3744	1	93.9	10.95	0	4.98	0	1.021254	-47.060128	220	4	1.1	0.9;
	3745	1	5.51	2	0	0.46	0	0.998625	-31.326971	110	21	1.1	0.9;
	3746	2	0	0	0	0	0	1.035948	-26.006613	110	19	1.1	0.9;
	3747	1	39.44	5.9	0	1.92	0	1.023444	-53.632832	220	4	1.1	0.9;
	3748	1	194.6	48.3	0	0.85	0	1.022204	-29.051332	110	21	1.1	0.9;
	3749	1	115.1	16.4	0	7.67	0	1.044531	-17.68671	220	5	1.1	0.9;
	3750	2	0	0	0	0	0	1.066155	-26.745554	220	21	1.1	0.9;
	3751	1	2.5	0	0	0.8	0	0.999949	-18.815407	110	18	1.1	0.9;
	3752	1	7.61	0.5	0	0.59	0	0.970786	-24.576145	110	18	1.1	0.9;
	3753	2	0	0	0	0	0	1.055722	9.871976	380	3	1.1	0.9;
	3754	1	-24.93	11.8	0	3.51	0	1.035766	-39.401367	220	4	1.1	0.9;
	3755	2	0	0	0	0	0	0.993935	-23.515599	154	1	1.1	0.9;
	3756	1	35.54	11.7	0	0.19	0	1.018418	-29.345862	110	21	1.1	0.9;
	3757	1	4	1.3	0	0.56	0	1.009695	-36.71181	110	12	1.1	0.9;
	3758	1	83.49	17.4	0	19.32	0	1.020395	-19.293122	220	5	1.1	0.9;
	3759	2	0	0	0	0	0	1.031969	48.203983	220	3	1.1	0.9;
	3760	1	163.5	44.9	0	1.17	0	1.053243	-21.24519	220	5	1.1	0.9;
	3761	1	71.08	16.2	0	11.58	0	1.036413	-17.713414	220	3	1.1	0.9;
	3762	1	338.22	104.11	0	91.11	0	1.020714	-6.273053	380	13	1.1	0.9;
	3763	1	119.1	61.4	0	1.9	0	1.039391	-27.930465	150	9	1.1	0.9;
	3764	1	37.84	23.4	0	0.02	0	0.999708	-27.06631	150	9	1.1	0.9;
	3765	1	18.02	6.5	0	0.64	0	1.037361	-23.418438	110	19	1.1	0.9;
	3766	2	0	0	0	0	0	1.0335	-5.773065	154	1	1.1	0.9;
	3767	1	33.04	21.6	0	26.59	0	1.044123	-48.905275	220	6	1.1	0.9;
	3768	1	19.82	10.9	0	0.05	0	1.038574	-22.203887	220	5	1.1	0.9;
	3769	2	0	0	0	0	0	1.021894	-45.791424	220	6	1.1	0.9;
	3770	1	27.13	8.9	0	0.27	0	0.987648	-31.926626	110	21	1.1	0.9;
	3771	1	34.14	13.5	0	0	0	1.020138	-1.224966	150	8	1.1	0.9;
	3772	1	2.2	-48.8	0	39.69	0	1.047921	6.46881	380	3	1.1	0.9;
	3773	1	0	0	0	30.09	0	1.058639	-33.269095	380	1	1.1	0.9;
	3774	1	-0	0	0	0.46	0	0.935503	-34.665326	150	9	1.1	0.9;
	3775	1	55.54	15.06	0	6.28	0	1.00144	-23.546474	220	5	1.1	0.9;
	3776	1	0	-0	0	5.25	0	1.022576	11.004156	150	8	1.1	0.9;
	3777	1	77.19	25.4	0	4.13	0	0.998564	7.675291	220	3	1.1	0.9;
	3778	1	-0	0	0	6.96	0	1.057766	-25.840464	220	1	1.1	0.9;
	3779	1	0	0	0	0.57	0	1.091216	-27.852933	150	9	1.1	0.9;
	3780	1	-0	0	0	7.6	0	1.028604	38.481234	220	3	1.1	0.9;
	3781	1	52.46	19	0	0.34	0	1.035901	-26.094417	110	19	1.1	0.9;
	3782	2	0	0	0	0	0	1.02915	-5.086551	220	13	1.1	0.9;
	3783	1	-0	-0	0	10.31	0	1.03166	-14.770149	154	1	1.1	0.9;
	3784	1	46.35	86.9	0	10.32	0	1.058319	11.041843	380	3	1.1	0.9;
	3785	1	27.8	4	0	3.34	0	1.011918	-5.392793	154	1	1.1	0.9;
	3786	1	-0	-0	0	26.98	0	1.040861	-19.805642	220	5	1.1	0.9;
	3787	1	80.56	20.85	0	10.33	0	1.031362	13.632437	220	13	1.1	0.9;
	3788	1	23.93	8.7	0	0.48	0	1.050359	-26.33833	110	19	1.1	0.9;
	3789	1	0	-0	0	9.09	0	1.025184	17.266234	380	3	1.1	0.9;
	3790	1	-0	-0	0	1.12	0	1.041201	-28.036796	220	12	1.1	0.9;
	3791	2	0	0	0	0	0	0.995335	-40.34716	150	10	1.1	0.9;
	3792	2	0	0	0	0	0	1.052349	-44.345634	220	4	1.1	0.9;
	3793	1	197.7	67.65	0	13.58	0	1.030761	-20.931618	220	5	1.1	0.9;
	3794	1	165.3	7.4	0	28.32	0	1.028017	-28.383082	380	5	1.1	0.9;
	3795	1	-0	0	0	0.96	0	0.982538	-27.706633	150	9	1.1	0.9;
	3796	1	2.8	0.6	0	0.62	0	1.051732	-25.18782	110	15	1.1	0.9;
	3797	2	0	0	0	0	0	1.060935	18.234389	380	13	1.1	0.9;
	3798	1	-0	0	0	17.64	0	1.063213	-41.590854	380	4	1.1	0.9;
	3799	1	39	5.5	0	18.09	0	1.031586	15.134334	154	1	1.1	0.9;
	3800	2	0	0	0	0	0	1.036528	17.026278	154	1	1.1	0.9;
	3801	1	-0	-0	0	2.97	0	1.014626	-38.774067	110	12	1.1	0.9;
	3802	1	0	0	0.181549	6.920889	0	1.064117	-26.171152	220	18	1.1	0.9;
	3803	1	23.3	3.3	0	9.29	0	1.020288	-8.413808	154	1	1.1	0.9;
	3804	1	10.11	5.6	0	0.89	0	0.986796	-28.076796	110	18	1.1	0.9;
	3805	1	8.91	1.3	0	0.82	0	0.931189	-32.565723	110	18	1.1	0.9;
	3806	1	5.01	2	0	0.84	0	1.036116	-28.47287	110	23	1.1	0.9;
	3807	1	258.3	166	0	16.31	0	1.007444	-16.522529	220	6	1.1	0.9;
	3808	2	0	0	0	0	0	1.073884	10.845509	380	14	1.1	0.9;
	3809	2	0	0	0	0	0	1.081564	-34.264616	220	5	1.1	0.9;
	3810	2	0	0	0	0	0	1.038535	-8.84055	110	18	1.1	0.9;
	3811	1	159.9	10.6	0	69.6	0	1.030166	-36.340808	380	6	1.1	0.9;
	3812	1	-0	0	0	0.05	0	0.964929	-31.486598	150	9	1.1	0.9;
	3813	1	226	55.79	0	18.2	0	1.021403	-54.029162	220	4	1.1	0.9;
	3814	1	130.5	24.7	0	0.4	0	1.037722	-22.267714	220	5	1.1	0.9;
	3815	1	16.32	2.8	0	1.41	0	0.990982	-26.200522	110	18	1.1	0.9;
	3816	1	0	-0	0	6.32	0	1.004119	-50.618066	220	6	1.1	0.9;
	3817	1	0	-0	0	1.02	0	1.082789	-15.473434	220	5	1.1	0.9;
	3818	1	44.45	10.9	0	2.23	0	1.038479	-26.123299	220	5	1.1	0.9;
	3819	1	0	-0	0	35	0	1.014365	-5.042182	150	8	1.1	0.9;
	3820	1	5.31	1.8	0	0.72	0	1.029328	-34.111494	110	12	1.1	0.9;
	3821	2	0	0	0	0	0	1.022257	-39.179203	220	4	1.1	0.9;
	3822	1	-0	-0	0	0.54	0	1.02071	6.692511	154	1	1.1	0.9;
	3823	1	227.6	37.8	0	39.33	0	1.036489	-44.778533	380	6	1.1	0.9;
	3824	1	0	0	0	0	0	0.983081	-31.410789	150	9	1.1	0.9;
	3825	2	0	0	0	0	0	1.024468	-8.742214	380	5	1.1	0.9;
	3826	1	0	0	0	9.47	0	1.017955	-36.045837	220	11	1.1	0.9;
	3827	1	40.5	5.7	0	5.41	0	1.041472	0.343332	154	1	1.1	0.9;
	3828	1	18.82	6.8	0	0.59	0	0.983046	-21.891128	110	19	1.1	0.9;
	3829	1	0	0	0	0.67	0	0.944418	-32.587786	150	9	1.1	0.9;
	3830	1	-0	0	0	97.326028	0	0.990018	-14.047581	380	5	1.1	0.9;
	3831	1	74.98	46.4	0	2.74	0	1.028921	-29.481204	150	9	1.1	0.9;
	3832	1	130.1	19.3	0	19.24	0	0.993255	-26.502145	220	6	1.1	0.9;
	3833	1	14.22	6.1	0	0.18	0	0.988667	-26.331209	110	18	1.1	0.9;
	3834	1	160	69.2	0	0.19	0	1.032203	-22.593387	220	5	1.1	0.9;
	3835	1	180.4	119.2	0	53.22	0	1.009608	-11.135296	380	3	1.1	0.9;
	3836	1	44.15	10.6	0	1.62	0	1.001472	-44.645173	110	22	1.1	0.9;
	3837	1	36.74	22.7	0	0.46	0	0.997079	-32.31357	150	9	1.1	0.9;
	3838	1	89.1	27.5	0	1.01	0	1.028571	-38.295607	110	22	1.1	0.9;
	3839	1	28.83	9.6	0	0.85	0	1.008217	-32.065747	110	21	1.1	0.9;
	3840	2	0	0	0	0	0	1.027073	-44.570683	220	4	1.1	0.9;
	3841	2	0	0	0	0	0	1.03976	-15.031301	220	6	1.1	0.9;
	3842	1	6.41	2.3	0	0.97	0	1.005181	-21.152822	110	19	1.1	0.9;
	3843	1	123.9	32.6	0	1.52	0	1.012416	-44.582275	220	4	1.1	0.9;
	3844	1	12.91	4.7	0	0.39	0	1.046985	-4.119378	110	19	1.1	0.9;
	3845	1	16.7	2.4	0	2.73	0	1.017228	2.936461	154	1	1.1	0.9;
	3846	1	-0	0	0	5.36	0	1.021343	10.878691	150	8	1.1	0.9;
	3847	1	6.81	2.3	0	1.88	0	0.986826	-30.310885	110	18	1.1	0.9;
	3848	1	123.09	9.2	0	0.92	0	1.033356	3.492423	220	13	1.1	0.9;
	3849	1	0	0	0	0	0	1.005911	-33.771512	150	9	1.1	0.9;
	3850	2	0	0	0	0	0	1.039563	-15.021363	220	6	1.1	0.9;
	3851	2	0	0	0	0	0	1.022667	62.987741	150	1	1.1	0.9;
	3852	1	-0	-0	0	6.39	0	1.063393	33.536857	380	1	1.1	0.9;
	3853	2	0	0	0	0	0	1.080197	-12.455971	220	18	1.1	0.9;
	3854	1	85.5	118.1	0	5.09	0	0.977713	11.366736	220	3	1.1	0.9;
	3855	1	-38.73	-15.32	0	97.23	0	1.069795	-26.566752	220	5	1.1	0.9;
	3856	1	28.03	8.8	0	1.05	0	1.02737	-34.077286	110	12	1.1	0.9;
	3857	1	-0	-0	0	11.68	0	1.074004	-23.723781	220	5	1.1	0.9;
	3858	1	0	-0	0	0.13	0	1.03237	-25.842451	150	9	1.1	0.9;
	3859	1	-0	-0	0	3.72	0	1.041425	-2.993357	220	3	1.1	0.9;
	3860	1	0	-0	0	0.23	0	1.036486	59.825516	220	3	1.1	0.9;
	3861	1	-0.58	-2.28	0	0.56	0	1.035472	-24.231604	110	19	1.1	0.9;
	3862	1	0	0	0	8.07	0	1.025684	2.642027	380	3	1.1	0.9;
	3863	1	-18.32	6	0	0.32	0	1.010567	-29.732557	150	9	1.1	0.9;
	3864	1	0	0	0	0.73	0	1.028747	-21.380405	110	21	1.1	0.9;
	3865	1	-0	-0	0	2.26	0	1.025528	-15.028298	220	5	1.1	0.9;
	3866	1	201.06	19.46	0	2.87	0	1.049401	-28.018776	220	5	1.1	0.9;
	3867	1	0	0	0	0	0	1.027292	-29.311327	150	9	1.1	0.9;
	3868	1	-8.43	-7.98	0	0	0	1.020133	0.828151	150	8	1.1	0.9;
	3869	2	0	0	0	0	0	1.028967	0.907249	380	5	1.1	0.9;
	3870	2	0	0	0	0	0	1.023622	-1.201992	154	1	1.1	0.9;
	3871	1	50.06	23	0	0.24	0	0.995472	-50.925622	220	6	1.1	0.9;
	3872	1	115.3	16.4	0	2.45	0	1.028094	-6.486951	154	1	1.1	0.9;
	3873	1	58.97	18.1	0	0.32	0	0.98774	-18.128573	110	18	1.1	0.9;
	3874	1	51.26	-17.2	0	100.22	0	1.048573	-15.840108	380	6	1.1	0.9;
	3875	1	11.61	4.3	0	1	0	1.052577	-24.982026	110	15	1.1	0.9;
	3876	2	0	0	0	0	0	1.04066	-9.801695	220	5	1.1	0.9;
	3877	1	31.74	3.8	0	1.3	0	1.017804	-15.067134	110	18	1.1	0.9;
	3878	1	-0	-0	0	9.19	0	1.03299	21.738093	220	8	1.1	0.9;
	3879	2	0	0	0	0	0	1.042235	-25.413451	150	9	1.1	0.9;
	3880	1	-28.13	-0.82	0	0.09	0	1.065391	-11.831636	220	5	1.1	0.9;
	3881	1	0	0	0	0.44	0	0.988893	-27.264103	110	18	1.1	0.9;
	3882	1	56.26	20.1	0	0.92	0	1.016182	-24.995584	110	15	1.1	0.9;
	3883	1	3.2	1.2	0	0.3	0	1.016025	-22.642882	110	19	1.1	0.9;
	3884	1	293	-45.9	0	33.199841	0	1.041428	-27.961037	380	6	1.1	0.9;
	3885	1	0	-0	0.381919	69.809723	0	1.047858	-16.613454	380	19	1.1	0.9;
	3886	1	94.3	-27.5	0	2.34	0	1.014676	8.853104	220	3	1.1	0.9;
	3887	1	-49.05	1	0	32.5	0	1.048305	-20.831942	220	11	1.1	0.9;
	3888	2	0	0	0	0	0	1.064956	-24.323541	220	1	1.1	0.9;
	3889	1	-10.41	3.4	0	0.99	0	1.017503	-27.418905	150	9	1.1	0.9;
	3890	1	4.61	1.7	0	0.49	0	1.031901	-22.759508	110	19	1.1	0.9;
	3891	1	0	0	0	7.89	0	1.043299	15.804627	220	8	1.1	0.9;
	3892	1	19.62	-3.1	0	10.98	0	1.037667	-36.041159	220	6	1.1	0.9;
	3893	1	8.91	5.1	0	0.42	0	1.05529	-17.093776	110	18	1.1	0.9;
	3894	1	0	0	0	4.77	0	1.022407	-15.481544	220	5	1.1	0.9;
	3895	1	-272.2	6.6	0	29.33	0	1.03666	-28.392836	380	6	1.1	0.9;
	3896	1	33.04	12.1	0	1.19	0	1.004806	-39.948467	110	22	1.1	0.9;
	3897	1	23.43	8.4	0	0.32	0	1.063019	-25.581302	110	15	1.1	0.9;
	3898	1	121.5	29.6	0	2.49	0	1.002855	-1.928796	150	8	1.1	0.9;
	3899	1	5.11	2.3	0	0.48	0	0.955225	-29.860405	110	18	1.1	0.9;
	3900	2	0	0	0	0	0	1.061016	-22.918735	220	21	1.1	0.9;
	3901	1	0	0	0	0.59	0	0.965876	-32.241266	150	9	1.1	0.9;
	3902	1	20.12	8.6	0	0.18	0	0.995358	-27.065993	110	18	1.1	0.9;
	3903	1	0	0	0	4.36	0	1.046996	-24.842582	220	5	1.1	0.9;
	3904	1	7.61	2.5	0	0.25	0	0.973906	-28.572412	110	18	1.1	0.9;
	3905	1	195.3	87.6	0	9.53	0	0.983021	9.191641	220	3	1.1	0.9;
	3906	1	0	0	0	98.2	0	1.033983	-13.62493	380	5	1.1	0.9;
	3907	2	0	0	0	0	0	1.025805	9.632959	400	1	1.1	0.9;
	3908	2	0	0	0	0	0	0.982046	-29.027646	110	18	1.1	0.9;
	3909	1	-0	-0	0	8.16	0	1.030915	-10.701167	220	13	1.1	0.9;
	3910	2	0	0	0	0	0	1.069291	19.435434	380	14	1.1	0.9;
	3911	1	-3.3	-0.7	0	0.7	0	1.081005	-21.199935	150	9	1.1	0.9;
	3912	1	0	-0	0	19.73	0	1.035094	-21.386174	220	5	1.1	0.9;
	3913	1	143.9	85.2	0	3.1	0	1.00287	-8.369051	220	3	1.1	0.9;
	3914	1	16.32	5.3	0	1.04	0	0.971068	-32.261085	110	21	1.1	0.9;
	3915	1	0	0	0.086633	0.534111	0	1.079222	-12.471756	220	18	1.1	0.9;
	3916	2	0	0	0	0	0	1.039177	-14.333423	220	5	1.1	0.9;
	3917	1	20.32	6.7	0	1	0	0.992675	-31.608578	110	21	1.1	0.9;
	3918	1	-11.8	-5.47	0	88.050735	0	1.020637	-2.16455	380	5	1.1	0.9;
	3919	1	-0	0	0	39.95	0	1.025031	-6.086566	380	5	1.1	0.9;
	3920	2	0	0	0	0	0	1.064621	1.351316	380	14	1.1	0.9;
	3921	1	39.34	10.3	0	0.33	0	1.001231	-20.204666	110	18	1.1	0.9;
	3922	2	0	0	0	0	0	1.07194	12.761553	380	3	1.1	0.9;
	3923	1	11.11	4	0	0.15	0	1.03489	-19.865903	110	19	1.1	0.9;
	3924	1	227.4	40.3	0	0.27	0	1.048868	-28.613564	110	21	1.1	0.9;
	3925	1	89.3	25.1	0	2.98	0	1.024217	-13.307404	220	5	1.1	0.9;
	3926	1	29.73	-0	0	1.34	0	1.003405	-43.226976	110	22	1.1	0.9;
	3927	1	35.64	22	0	5.67	0	0.917414	-35.806411	150	9	1.1	0.9;
	3928	1	105.2	24.2	0	0.25	0	1.044126	-4.978831	220	5	1.1	0.9;
	3929	1	48.95	11.6	0	2	0	1.028838	-20.533141	220	5	1.1	0.9;
	3930	1	0	0	0	45.86	0	1.042095	27.099563	380	13	1.1	0.9;
	3931	1	0	-0	0	14.83	0	1.029087	62.729982	150	1	1.1	0.9;
	3932	1	-0	0	0	4.86	0	1.026753	-34.842553	220	10	1.1	0.9;
	3933	1	37.94	23.5	0	0.85	0	1.002413	-30.25303	150	9	1.1	0.9;
	3934	1	236.2	45.1	0	2.88	0	0.998415	5.424314	220	3	1.1	0.9;
	3935	1	0	-0	0	0.89	0	1.041388	-5.736526	220	3	1.1	0.9;
	3936	1	0	0	0	0.81	0	1.03405	-33.85919	110	12	1.1	0.9;
	3937	1	83.19	11.4	0	9.84	0	1.021808	-43.507932	220	4	1.1	0.9;
	3938	2	0	0	0	0	0	1.049226	-0.339481	220	13	1.1	0.9;
	3939	1	27.9	4	0	0.02	0	1.020565	-18.201653	154	1	1.1	0.9;
	3940	2	0	0	0	0	0	1.018221	-10.942619	220	24	1.1	0.9;
	3941	1	4.81	1.3	0	0.85	0	0.958492	-30.478094	110	18	1.1	0.9;
	3942	1	1.1	-0	0	26.91	0	1.027536	-36.381609	220	22	1.1	0.9;
	3943	1	90.5	12.8	0	7.61	0	1.019158	6.298048	154	1	1.1	0.9;
	3944	1	-0	0	0	4.14	0	1.02866	-13.120921	220	3	1.1	0.9;
	3945	1	0	0	0	9.05	0	1.037147	-17.758493	220	1	1.1	0.9;
	3946	1	71.88	44.5	0	0.04	0	1.015015	-30.091097	150	9	1.1	0.9;
	3947	1	20.02	15	0	2.11	0	0.98294	9.18829	220	3	1.1	0.9;
	3948	2	0	0	0	0	0	1.03639	-37.486103	220	6	1.1	0.9;
	3949	2	0	0	0	0	0	1.052164	-12.244761	220	7	1.1	0.9;
	3950	1	0.4	0.9	0	0.38	0	1.035258	34.769482	380	3	1.1	0.9;
	3951	2	0	0	0	0	0	1.03485	-13.335873	220	5	1.1	0.9;
	3952	1	5.11	3.8	0	1.06	0	0.952264	-30.671125	110	18	1.1	0.9;
	3953	1	108.31	14.23	0	8.47	0	1.031948	6.614385	220	13	1.1	0.9;
	3954	1	310.8	162.6	0	19.47	0	0.975077	-56.829144	220	10	1.1	0.9;
	3955	1	0	-0	0	0	0	0.962335	-15.73451	110	16	1.1	0.9;
	3956	1	-0	-0	0	1.18	0	1.044996	-24.413265	220	5	1.1	0.9;
	3957	1	-0	-0	0	0.04	0	1.001778	-45.153027	220	10	1.1	0.9;
	3958	2	0	0	0	0	0	1.034533	-34.836278	220	6	1.1	0.9;
	3959	2	0	0	0	0	0	1.052799	9.681964	154	1	1.1	0.9;
	3960	1	0	-0	0	5.62	0	1.034466	9.04075	150	8	1.1	0.9;
	3961	1	42.35	7.2	0	1.47	0	1.006923	-48.154846	220	4	1.1	0.9;
	3962	1	19.42	11.8	0	2.6	0	1.036556	-24.058669	220	5	1.1	0.9;
	3963	2	0	0	0	0	0	1.020596	-46.298437	380	6	1.1	0.9;
	3964	1	1.3	-0	0	0.57	0	0.995855	-26.915906	110	18	1.1	0.9;
	3965	1	4.2	0.9	0	0.12	0	0.977329	-31.493768	110	18	1.1	0.9;
	3966	1	220.7	-14.8	0	125.52	0	1.042714	-43.322909	380	4	1.1	0.9;
	3967	1	3.9	1.4	0	0.2	0	1.059299	-21.998958	110	19	1.1	0.9;
	3968	1	-0.18	-2.27	0	0	0	1.048237	18.202352	380	8	1.1	0.9;
	3969	1	-0	-0	0	0.84	0	1.016279	-10.549563	220	5	1.1	0.9;
	3970	1	5.11	1.1	0	0.33	0	0.959234	-30.149348	110	18	1.1	0.9;
	3971	2	0	0	0	0	0	1.053061	-9.118186	220	5	1.1	0.9;
	3972	1	419.5	114.1	0	55.629522	0	1.02952	-15.684934	380	6	1.1	0.9;
	3973	1	2.51	-1.64	0	6.26	0	1.038157	-28.281886	220	7	1.1	0.9;
	3974	1	-0.18	-1.42	0	0	0	1.022625	3.874386	150	8	1.1	0.9;
	3975	1	148	26.1	0	9.44	0	0.98597	-24.863841	220	5	1.1	0.9;
	3976	1	-0	0	0	4.39	0	1.024952	3.133455	380	3	1.1	0.9;
	3977	1	29.83	14.4	0	8.89	0	1.033288	-2.962765	220	1	1.1	0.9;
	3978	1	-0	0	0	28.93	0	1.093267	-33.294124	380	1	1.1	0.9;
	3979	2	0	0	0	0	0	1.027309	-1.718871	154	1	1.1	0.9;
	3980	1	13.9	1.9	0	3.72	0	1.014725	-2.51955	154	1	1.1	0.9;
	3981	1	0	-0	0.359411	37.9409	0	1.064123	-19.939188	380	19	1.1	0.9;
	3982	1	130.5	18.6	0	4.56	0	1.018498	1.671347	154	1	1.1	0.9;
	3983	2	0	0	0	0	0	1.014828	16.462462	150	8	1.1	0.9;
	3984	1	33.34	20.7	0	0.02	0	1.023174	-26.038202	150	9	1.1	0.9;
	3985	1	77.19	26.5	0	6.2	0	1.043935	-0.168775	220	5	1.1	0.9;
	3986	1	14.15	3.8	0	0.47	0	0.995145	-27.217275	110	18	1.1	0.9;
	3987	2	0	0	0	0	0	1.019043	-33.538829	220	10	1.1	0.9;
	3988	1	22.73	5.8	0	0.51	0	1.032522	-26.401538	110	3	1.1	0.9;
	3989	1	76.6	10.9	0	1.33	0	0.998086	-22.778757	154	1	1.1	0.9;
	3990	1	7.01	1.8	0	35.44	0	1.051551	-49.108929	380	10	1.1	0.9;
	3991	2	0	0	0	0	0	1.035055	3.238867	220	13	1.1	0.9;
	3992	1	34.64	21.4	0	0.02	0	0.98201	-27.314622	150	9	1.1	0.9;
	3993	1	-0	-0	0	0.66	0	1.055754	-27.405779	150	9	1.1	0.9;
	3994	1	36.34	10.4	0	2.67	0	1.032031	-3.028214	220	5	1.1	0.9;
	3995	1	-7.91	2.6	0	1.77	0	0.956693	-31.59276	150	9	1.1	0.9;
	3996	2	0	0	0	0	0	1.034652	-6.592997	220	13	1.1	0.9;
	3997	1	26.13	9.7	0	2.37	0	1.037337	-12.258676	220	5	1.1	0.9;
	3998	1	0	-0	0	0.99	0	0.997561	-29.542302	150	9	1.1	0.9;
	3999	1	110.8	36.9	0	23.2	0	1.047269	-6.708317	220	5	1.1	0.9;
	4000	1	-0	0	0	24.103274	0	0.987988	-15.802118	380	5	1.1	0.9;
	4001	1	18.2	2.6	0	6.72	0	1.025871	12.586597	154	1	1.1	0.9;
	4002	1	88.93	16.75	0	6.75	0	1.058927	17.854873	220	13	1.1	0.9;
	4003	1	0	0	0	10.43	0	1.043581	-12.669131	380	1	1.1	0.9;
	4004	1	0	0	0	0	0	1.020362	-29.475309	150	9	1.1	0.9;
	4005	1	171.18	28.49	0	0	0	1.060797	-32.050564	220	5	1.1	0.9;
	4006	1	-0.18	-1	0	0	0	1.021595	10.57369	150	8	1.1	0.9;
	4007	1	76.6	10.9	0	7.8	0	1.017324	-22.243506	154	1	1.1	0.9;
	4008	1	-0	-0	0.649603	14.162287	0	1.069936	-25.232127	220	19	1.1	0.9;
	4009	2	0	0	0	0	0	1.043192	-8.175707	110	19	1.1	0.9;
	4010	1	19.32	6	0	1.38	0	1.016756	-40.50939	110	12	1.1	0.9;
	4011	1	144.8	39.7	0	12.64	0	1.025129	-54.334861	220	4	1.1	0.9;
	4012	2	0	0	0	0	0	1.04188	62.600682	220	3	1.1	0.9;
	4013	1	0	-0	0	0.17	0	0.984886	-56.240492	150	10	1.1	0.9;
	4014	1	0	0	0	0	0	1.02664	-13.635302	150	2	1.1	0.9;
	4015	1	-0	0	0	0.2	0	0.993458	-30.724811	110	21	1.1	0.9;
	4016	2	0	0	0	0	0	1.022786	-43.184654	220	4	1.1	0.9;
	4017	1	0	0	0	0	0	1.073977	-49.479938	150	6	1.1	0.9;
	4018	1	0	0	0	0	0	0.971505	-27.269578	150	9	1.1	0.9;
	4019	1	394.9	-17.3	0	30.22	0	1.03341	-39.158514	380	6	1.1	0.9;
	4020	1	0	-0	0	14.28	0	1.042316	-50.37737	380	4	1.1	0.9;
	4021	1	-0.18	-1	0	0	0	1.025368	6.397762	380	8	1.1	0.9;
	4022	1	-0	-0	0	53.04	0	1.095409	-31.632735	380	12	1.1	0.9;
	4023	1	-1270.6	-476.3	0	154.25	0	1.029115	1.738319	380	3	1.1	0.9;
	4024	2	0	0	0	0	0	1.029565	2.288816	380	5	1.1	0.9;
	4025	1	0	0	0	2.67	0	1.032176	-3.017042	220	5	1.1	0.9;
	4026	2	0	0	0	0	0	1.05562	-41.532068	380	4	1.1	0.9;
	4027	1	0	0	0	0	0	1.046487	-26.102165	380	6	1.1	0.9;
	4028	2	0	0	0	0	0	1.022221	-28.502738	110	19	1.1	0.9;
	4029	1	20.92	10.3	0	0.68	0	0.960003	-29.159918	110	18	1.1	0.9;
	4030	1	0	0	0	0	0	1.08382	-28.028856	150	9	1.1	0.9;
	4031	1	0	-0	0	6.42	0	1.054311	-11.010018	220	5	1.1	0.9;
	4032	1	265.7	113.5	0	1.85	0	0.997508	-11.76106	220	5	1.1	0.9;
	4033	1	54.96	31.7	0	8.58	0	1.025548	56.993309	220	3	1.1	0.9;
	4034	1	24.03	7.9	0	1.08	0	1.001466	-26.170849	110	21	1.1	0.9;
	4035	1	58.56	15.9	0	3.75	0	1.030419	-21.593477	220	6	1.1	0.9;
	4036	1	-0	0	0	0.3	0	1.041024	-27.643985	380	6	1.1	0.9;
	4037	1	352.6	193.5	0	39.3	0	1.032069	-19.610125	380	6	1.1	0.9;
	4038	1	15.52	4.1	0	0.63	0	1.018873	-36.406651	110	12	1.1	0.9;
	4039	1	-0	-0	0	1.18	0	1.033941	-22.873209	220	5	1.1	0.9;
	4040	1	37.24	40.5	0	67.07	0	1.032406	-18.580042	380	6	1.1	0.9;
	4041	1	-0	0	0	19.37	0	1.034415	29.478216	380	3	1.1	0.9;
	4042	1	8.9	1.3	0	3.55	0	1.031208	-10.044032	154	1	1.1	0.9;
	4043	1	76.6	10.9	0	3.24	0	1.016992	-10.480195	154	1	1.1	0.9;
	4044	1	1.9	0.7	0	1.34	0	1.04924	-14.04151	110	19	1.1	0.9;
	4045	1	0	-0	0	0.57	0	0.967919	-28.718168	110	18	1.1	0.9;
	4046	2	0	0	0	0	0	1.071125	-1.423326	380	14	1.1	0.9;
	4047	1	-0	-0	0	23.74	0	1.042383	-17.618442	220	3	1.1	0.9;
	4048	1	-0	0	0	17.54	0	1.038673	-3.249222	220	1	1.1	0.9;
	4049	1	-0	-0	0	0.89	0	1.01536	-10.191137	220	5	1.1	0.9;
	4050	1	-0	-0	0	107.36	0	1.05168	3.848506	400	1	1.1	0.9;
	4051	1	27.58	-33.38	0	13.08	0	1.043582	13.183243	220	13	1.1	0.9;
	4052	2	0	0	0	0	0	1.059535	49.902081	380	3	1.1	0.9;
	4053	1	0	-0	0	1.12	0	0.98128	-27.962853	110	18	1.1	0.9;
	4054	1	0	-0	0	2.71	0	1.03161	-18.751733	220	5	1.1	0.9;
	4055	1	347.7	96.6	0	101.73	0	1.043979	-18.088298	380	16	1.1	0.9;
	4056	2	0	0	0	0	0	1.038451	-9.852296	220	5	1.1	0.9;
	4057	1	74.78	-31.4	0	1.74	0	1.039902	8.020393	110	8	1.1	0.9;
	4058	1	0	0	0	0.91	0	1.005208	-48.287338	220	4	1.1	0.9;
	4059	1	0	0	0	0	0	1.079468	31.589738	380	8	1.1	0.9;
	4060	1	-0	0	0	2.08	0	1.055551	-15.2235	220	5	1.1	0.9;
	4061	1	0	0	0	5.66	0	1.008585	-25.37963	154	1	1.1	0.9;
	4062	1	6.31	0	0	0.42	0	1.008647	-48.903829	220	4	1.1	0.9;
	4063	1	56.53	15.4	0	0.12	0	1.060287	-13.44471	110	24	1.1	0.9;
	4064	1	131.35	63.6	0	8.55	0	1.033499	-3.54913	220	1	1.1	0.9;
	4065	1	14.72	5.3	0	0.49	0	1.019819	-29.008702	110	19	1.1	0.9;
	4066	1	111.4	15.9	0	1.04	0	1.025394	-25.953072	154	1	1.1	0.9;
	4067	1	-0	-0	0	3.51	0	1.025119	-22.133999	150	9	1.1	0.9;
	4068	1	18.22	2.3	0	0.97	0	0.973947	-32.243936	110	18	1.1	0.9;
	4069	1	7.29	-0.73	0	0.73	0	1.003753	-31.343984	110	19	1.1	0.9;
	4070	1	35.44	22	0	0.61	0	0.982693	-27.895484	150	9	1.1	0.9;
	4071	2	0	0	0	0	0	1.026598	2.362831	154	1	1.1	0.9;
	4072	1	4.2	2	0	0.21	0	0.984045	-16.08228	110	18	1.1	0.9;
	4073	1	0	-0	0	1.01	0	1.068179	-26.714677	150	9	1.1	0.9;
	4074	1	0	-0	0	0.53	0	1.025354	-24.831739	110	3	1.1	0.9;
	4075	2	0	0	0	0	0	1.033829	-10.847174	220	6	1.1	0.9;
	4076	2	0	0	0	0	0	1.042958	4.479736	220	14	1.1	0.9;
	4077	1	0	-0	0	23.27	0	1.043151	16.235866	380	8	1.1	0.9;
	4078	1	8.3	1.2	0	1.07	0	1.024053	-8.802712	154	1	1.1	0.9;
	4079	1	0	0	0.16	31.35	0	1.021796	-28.526157	380	7	1.1	0.9;
	4080	1	0	-0	0	1.51	0	1.060911	-27.026792	150	9	1.1	0.9;
	4081	1	-0	-0	0	6.9	0	1.027834	-53.268647	220	4	1.1	0.9;
	4082	1	-0	0	0	1.14	0	1.031489	12.027323	150	8	1.1	0.9;
	4083	1	113	57.3	0	12.82	0	1.005675	-32.160727	150	9	1.1	0.9;
	4084	2	0	0	0	0	0	1.053107	-9.058405	220	5	1.1	0.9;
	4085	2	0	0	0	0	0	1.028863	38.500197	220	3	1.1	0.9;
	4086	1	62.7	9	0	8.02	0	1.018756	-22.452168	154	1	1.1	0.9;
	4087	1	-0	-0	0	26.36	0	1.047271	17.789624	380	8	1.1	0.9;
	4088	1	-0	0	0	0.88	0	1.048087	-20.442706	220	1	1.1	0.9;
	4089	1	195.2	92.2	0	11.67	0	0.996796	-18.606163	220	6	1.1	0.9;
	4090	2	0	0	0	0	0	1.010258	-24.552568	154	1	1.1	0.9;
	4091	1	-0	-0	0	28.29	0	1.119258	-23.704462	380	1	1.3	0.7;
	4092	1	0	0	0	0.05	0	0.986344	14.7005	220	3	1.1	0.9;
	4093	1	40.24	24.9	0	1.81	0	1.006283	-27.062298	150	9	1.1	0.9;
	4094	2	0	0	0	0	0	1.026088	-41.974953	220	6	1.1	0.9;
	4095	1	65.27	6	0	3.89	0	1.011773	-58.826771	220	4	1.1	0.9;
	4096	1	47.65	7.1	0	7.3	0	1.009252	-48.779933	220	4	1.1	0.9;
	4097	1	0	0	0	4.05	0	1.021071	-6.030317	400	1	1.1	0.9;
	4098	1	61.67	28.3	0	0.1	0	0.962168	-44.407983	220	10	1.1	0.9;
	4099	1	173	11.3	0	8.41	0	1.02402	-17.261834	220	3	1.1	0.9;
	4100	1	-0	-0	0	1.23	0	1.068198	-11.140706	220	5	1.1	0.9;
	4101	1	128	72.9	0	14.95	0	1.033704	18.896597	380	3	1.1	0.9;
	4102	1	5.31	2	0	0.17	0	0.988737	-18.009915	110	18	1.1	0.9;
	4103	1	81.09	12.5	0	2.4	0	1.032405	-22.044842	220	5	1.1	0.9;
	4104	1	40.34	10.4	0	7.02	0	1.021613	10.483701	110	8	1.1	0.9;
	4105	1	0.5	12.4	0	0.25	0	1.043143	-3.069097	220	3	1.1	0.9;
	4106	2	0	0	0	0	0	1.022601	-5.652252	154	1	1.1	0.9;
	4107	1	-0	-0	0	1.04	0	1.009935	51.392856	220	3	1.1	0.9;
	4108	1	-0	0	0	0.34	0	0.97339	-23.920548	150	9	1.1	0.9;
	4109	1	0	-0	0	5.16	0	1.031638	6.730991	154	1	1.1	0.9;
	4110	1	13.01	3.8	0	3.59	0	1.045421	0.527375	220	5	1.1	0.9;
	4111	1	-0	0	0.009025	43.164575	0	1.048823	-13.156135	380	16	1.1	0.9;
	4112	1	64.77	6.4	0	3.84	0	1.017139	-45.032425	220	4	1.1	0.9;
	4113	1	-0	0	0	84.33	0	1.017304	-16.688031	380	3	1.1	0.9;
	4114	1	9.91	5.4	0	4.6	0	1.044271	-24.809474	220	5	1.1	0.9;
	4115	1	-0	0	0.08	-0.24	0	1.042411	-22.654062	220	11	1.1	0.9;
	4116	1	2.2	0.5	0	2.31	0	0.98368	-51.197164	150	10	1.1	0.9;
	4117	1	-0	0	0	18.49	0	1.010188	30.86743	380	3	1.1	0.9;
	4118	2	0	0	0	0	0	1.025736	-14.964695	220	5	1.1	0.9;
	4119	1	0	0	0	19.2	0	1.062772	43.365819	380	3	1.1	0.9;
	4120	1	-0.54	-1	0	4.8	0	1.044239	-25.077662	150	9	1.1	0.9;
	4121	1	-0	-0	0.316107	31.224257	0	1.021382	14.426094	380	3	1.1	0.9;
	4122	2	0	0	0	0	0	1.016568	36.384533	220	3	1.1	0.9;
	4123	1	0	-0	0	0.04	0	0.989609	-30.814829	150	9	1.1	0.9;
	4124	1	230.42	-7.87	0	5.13	0	1.014009	-45.930495	220	4	1.1	0.9;
	4125	2	0	0	0	0	0	1.002666	-8.596461	380	5	1.1	0.9;
	4126	1	0	0	0	1.48	0	1.024485	-13.787118	220	6	1.1	0.9;
	4127	1	39.14	7.7	0	11.68	0	1.029406	-20.472049	220	5	1.1	0.9;
	4128	2	0	0	0	0	0	1.027647	-15.463095	220	5	1.1	0.9;
	4129	1	-0.02	-0.51	0	1.61	0	1.011842	61.759219	150	1	1.1	0.9;
	4130	1	0	-0	0	2.1	0	1.015307	10.365855	150	8	1.1	0.9;
	4131	2	0	0	0	0	0	1.034065	-26.037778	110	21	1.1	0.9;
	4132	1	25.43	7.9	0	1.03	0	1.016338	11.716163	110	8	1.1	0.9;
	4133	1	145.8	42.4	0	2.34	0	1.006942	-49.469981	220	6	1.1	0.9;
	4134	1	-0	0	0	0.09	0	1.069875	-33.898771	220	5	1.1	0.9;
	4135	1	52.6	7.5	0	3.66	0	1.010695	5.975283	154	1	1.1	0.9;
	4136	1	13.31	4.4	0	1.51	0	1.021921	-31.27007	110	21	1.1	0.9;
	4137	1	13.64	9	0	1.7	0	1.005911	-33.771512	150	9	1.1	0.9;
	4138	1	137.8	22.9	0	6.55	0	1.01547	-32.204194	220	6	1.1	0.9;
	4139	1	-0	0	0	20.86	0	1.025481	5.379476	380	3	1.1	0.9;
	4140	1	-0	-0	0	11.85	0	1.013716	-21.422838	220	3	1.1	0.9;
	4141	1	-0.62	0	0	75.020288	0	1.030132	-21.352318	380	5	1.1	0.9;
	4142	1	-0	0	0	30.64	0	1.062784	-21.951534	380	1	1.1	0.9;
	4143	1	11.11	4.1	0	0.8	0	1.048779	-1.219149	220	5	1.1	0.9;
	4144	1	310.2	4.5	0	8.51	0	1.037466	-19.789647	220	5	1.1	0.9;
	4145	1	128.18	35.03	0	1.76	0	1.010696	61.676903	150	1	1.1	0.9;
	4146	1	246.9	20.3	0	4.52	0	0.982819	9.414059	220	3	1.1	0.9;
	4147	1	-0	-0	0	0.47	0	1.029151	-31.51084	380	6	1.1	0.9;
	4148	1	-13.26	-7.13	0	23.16	0	1.004853	-50.224318	220	6	1.1	0.9;
	4149	1	6.71	1.3	0	0.34	0	1.016767	-15.388627	110	18	1.1	0.9;
	4150	1	1.9	9.2	0	7.57	0	0.993895	-22.508601	220	20	1.1	0.9;
	4151	1	91.4	-88.2	0	0.29	0	1.023997	-3.039664	220	3	1.1	0.9;
	4152	1	71.68	14.8	0	13.66	0	1.004086	-50.610328	220	6	1.1	0.9;
	4153	1	0	0	0	2.69	0	1.030434	48.268758	220	3	1.1	0.9;
	4154	1	110.7	94.3	0	6.43	0	1.003321	8.805053	220	3	1.1	0.9;
	4155	1	15.02	7	0	0.67	0	1.032095	-28.416454	110	23	1.1	0.9;
	4156	1	0.2	109.4	0	0	0	1.071034	59.203115	380	3	1.1	0.9;
	4157	1	0	0	0	1.2	0	1.068844	-26.733655	220	5	1.1	0.9;
	4158	1	121.4	40	0	11.16	0	1.018564	-16.083787	380	3	1.1	0.9;
	4159	1	175.2	70.7	0	109.1	0	1.03101	-19.53333	380	6	1.1	0.9;
	4160	1	0	0	0	0	0	0.990312	-30.214954	150	9	1.1	0.9;
	4161	1	29.03	10.1	0	2.41	0	1.060545	-18.151108	110	19	1.1	0.9;
	4162	1	0	-0	0	4.37	0	1.127149	-21.902131	380	9	1.3	0.7;
	4163	2	0	0	0	0	0	1.047216	-1.442069	220	13	1.1	0.9;
	4164	1	0	0	0	14.62	0	1.017898	-16.185279	380	3	1.1	0.9;
	4165	1	325.2	38.1	0	19.99	0	1.025082	-43.939303	380	4	1.1	0.9;
	4166	1	0	0	0	0	0	1.01376	-26.588907	150	9	1.1	0.9;
	4167	1	50.86	-23	0	16.36	0	1.043491	-20.51172	220	3	1.1	0.9;
	4168	1	200.2	90.6	0	24.07	0	1.004424	-50.572566	220	6	1.1	0.9;
	4169	1	62.07	20	0	0.25	0	0.983291	-31.912299	110	21	1.1	0.9;
	4170	1	-0	0	0	12.88	0	0.99126	-32.514071	220	10	1.1	0.9;
	4171	1	7.91	3.8	0	0.22	0	1.014773	-26.307063	110	18	1.1	0.9;
	4172	2	0	0	0	0	0	1.02637	-36.492048	110	12	1.1	0.9;
	4173	1	34.9	5	0	1.72	0	1.007409	-10.237514	154	1	1.1	0.9;
	4174	1	-0	0	0	0.71	0	1.011231	-30.639196	150	9	1.1	0.9;
	4175	2	0	0	0	0	0	1.078047	44.862058	380	3	1.1	0.9;
	4176	1	144.5	40.9	0	17.35	0	1.025419	-13.727286	220	7	1.1	0.9;
	4177	1	20.22	5.7	0	3.25	0	1.019708	-45.343115	220	4	1.1	0.9;
	4178	2	0	0	0	0	0	1.031771	-38.209109	380	6	1.1	0.9;
	4179	2	0	0	0	0	0	1.010218	-27.997197	110	18	1.1	0.9;
	4180	1	-0	-0	0	4.04	0	1.035036	-17.3396	220	1	1.1	0.9;
	4181	1	195.6	-1	0	3.19	0	1.037185	-46.022811	220	4	1.1	0.9;
	4182	2	0	0	0	0	0	1.011247	-44.11642	380	4	1.1	0.9;
	4183	2	0	0	0	0	0	1.031152	-10.701804	110	18	1.1	0.9;
	4184	1	26.73	9.6	0	2.19	0	1.069781	-19.474091	110	19	1.1	0.9;
	4185	1	87.5	21.3	0	7.52	0	1.038294	-26.144543	220	5	1.1	0.9;
	4186	1	-0	0	0	5.06	0	1.032163	-23.48406	220	5	1.1	0.9;
	4187	1	0	0	0.090774	6.020444	0	1.058969	-18.603344	220	18	1.1	0.9;
	4188	1	59.17	19.3	0	1.16	0	1.010423	-0.341615	150	8	1.1	0.9;
	4189	1	200.65	30.72	0	9.72	0	1.028895	-26.578685	220	5	1.1	0.9;
	4190	1	0	0	0	42.131932	0	1.045226	-11.788173	380	24	1.1	0.9;
	4191	1	12.71	2.5	0	0.83	0	0.982488	-30.116866	110	18	1.1	0.9;
	4192	2	0	0	0	0	0	1.076318	18.042352	380	8	1.1	0.9;
	4193	1	0	0	0	2.06	0	1.014182	-5.477015	154	1	1.1	0.9;
	4194	1	0	0	0	0	0	0.951351	-29.98231	150	9	1.1	0.9;
	4195	2	0	0	0	0	0	1.076653	4.62948	380	14	1.1	0.9;
	4196	1	-3.2	-1.05	0	48.471563	0	1.02511	-7.715077	380	5	1.1	0.9;
	4197	1	9.01	4.9	0	0.64	0	1.061789	0.582825	220	5	1.1	0.9;
	4198	2	0	0	0	0	0	1.01015	1.802303	154	1	1.1	0.9;
	4199	1	50.3	7.2	0	8.06	0	1.035373	7.6217	154	1	1.1	0.9;
	4200	2	0	0	0	0	0	1.016165	66.484445	380	1	1.1	0.9;
	4201	1	318.3	76.8	0.146686	63.265082	0	1.034586	-20.71494	380	7	1.1	0.9;
	4202	1	53.86	18.7	0	3.48	0	1.042806	-39.579659	110	22	1.1	0.9;
	4203	1	0	0	0	0	0	0.962313	-37.255005	150	9	1.1	0.9;
	4204	2	0	0	0	0	0	1.014828	-40.210588	220	4	1.1	0.9;
	4205	1	5.81	1.7	0	0.42	0	1.042725	-13.028579	220	5	1.1	0.9;
	4206	1	0	-0	0	39.36	0	1.034464	-1.395933	380	3	1.1	0.9;
	4207	2	0	0	0	0	0	1.009702	-27.338201	220	11	1.1	0.9;
	4208	1	0	0	0	10.1	0	1.14157	-23.643367	380	15	1.3	0.7;
	4209	1	-0	-0	0	0	0	1.048002	18.18959	380	8	1.1	0.9;
	4210	2	0	0	0	0	0	1.031932	-26.167183	150	9	1.1	0.9;
	4211	2	0	0	0	0	0	1.042563	-24.017297	220	7	1.1	0.9;
	4212	1	0	0	0	61.61	0	1.028935	-21.684767	154	1	1.1	0.9;
	4213	1	32.34	14.1	0	0.24	0	0.998759	-26.288816	110	18	1.1	0.9;
	4214	1	-806.3	-50.1	0	4.19	0	1.074255	61.102292	380	3	1.1	0.9;
	4215	1	-0	-0	0	17.29	0	1.041291	-23.383686	220	5	1.1	0.9;
	4216	1	-0.17	-8.06	0	0.83	0	1.028315	7.907933	110	8	1.1	0.9;
	4217	1	-0	-0	0.117325	64.47707	0	1.031653	66.374308	380	1	1.1	0.9;
	4218	1	0	0	0	27.56	0	1.063506	27.291012	220	3	1.1	0.9;
	4219	1	8.41	2.5	0	1.18	0	0.994825	-31.851557	110	18	1.1	0.9;
	4220	1	27.63	9.2	0	0.58	0	1.013969	-31.986962	110	21	1.1	0.9;
	4221	2	0	0	0	0	0	1.036578	-20.82525	154	1	1.1	0.9;
	4222	1	-0.66	-2.79	0	2.49	0	1.067224	-26.550309	150	9	1.1	0.9;
	4223	1	0	-0	0	1.44	0	1.012498	-19.744499	110	18	1.1	0.9;
	4224	1	-20.56	-3.88	0	1.72	0	1.021092	-38.627211	220	4	1.1	0.9;
	4225	1	46.8	6.7	0	25.55	0	1.013124	10.540111	400	1	1.1	0.9;
	4226	1	0	0	0	122.87	0	1.045038	-16.143617	380	3	1.1	0.9;
	4227	1	-9.87	-3.76	0	0.21	0	1.00828	-24.572415	110	18	1.1	0.9;
	4228	2	0	0	0	0	0	1.060932	64.062878	150	1	1.1	0.9;
	4229	1	0	0	0	0	0	0.870967	-38.524972	150	9	1.3	0.7;
	4230	1	57.76	16	0	2.39	0	0.980235	-45.225947	220	4	1.1	0.9;
	4231	3	0	0	0	0	0	1.042866	0	380	5	1.1	0.9;
	4232	1	11.01	3.1	0	0.14	0	1.021825	-28.614751	110	19	1.1	0.9;
	4233	1	-0	0	0	0.71	0	0.962703	-27.325558	150	9	1.1	0.9;
	4234	1	0	0	0	8.95	0	1.043784	-17.44016	220	5	1.1	0.9;
	4235	1	-0	-0	0	12.68	0	1.042991	-19.568975	220	5	1.1	0.9;
	4236	1	19.52	6.4	0	0.03	0	1.042174	-29.589505	150	9	1.1	0.9;
	4237	1	-0	0	0	42.02	0	1.014748	10.103007	400	1	1.1	0.9;
	4238	1	-0	-0	0	8.08	0	1.007519	-48.060372	220	4	1.1	0.9;
	4239	1	190.7	52	0	20.57	0	1.043664	-13.554722	220	5	1.1	0.9;
	4240	1	44.35	14.8	0	0.36	0	1.015037	-32.004489	110	21	1.1	0.9;
	4241	1	-0	-0	0	93.61	0	1.021345	-43.991555	380	4	1.1	0.9;
	4242	1	269.22	97.24	0	0	0	1.019378	-27.068452	110	24	1.1	0.9;
	4243	1	188.48	128.3	0	18.06	0	1.016657	-29.054582	220	6	1.1	0.9;
	4244	2	0	0	0	0	0	1.039687	11.244957	380	13	1.1	0.9;
	4245	1	24.93	9.6	0	0.09	0	1.037662	-10.152328	220	5	1.1	0.9;
	4246	1	24.23	5.5	0	1.18	0	1.025701	-40.895394	110	22	1.1	0.9;
	4247	1	33.54	20.8	0	2.42	0	0.85042	-39.715801	150	9	1.3	0.7;
	4248	1	0	-0	0	3.24	0	1.00246	25.861349	220	8	1.1	0.9;
	4249	1	57.06	15.6	0	2.03	0	1.022176	-55.071765	220	4	1.1	0.9;
	4250	1	-10.71	3.5	0	0.58	0	1.04599	-25.355234	150	9	1.1	0.9;
	4251	1	47.15	15.7	0	5.41	0	1.02428	-17.852918	220	5	1.1	0.9;
	4252	1	-0	0	0	0.56	0	1.023266	-14.144948	110	18	1.1	0.9;
	4253	1	105	18	0	5.41	0	1.023905	-54.602507	220	4	1.1	0.9;
	4254	1	44.55	27.6	0	1.86	0	1.023046	-30.120615	150	9	1.1	0.9;
	4255	1	2.7	0.4	0	4.05	0	1.026601	-2.601651	154	1	1.1	0.9;
	4256	1	-0	0	0	6.76	0	0.997399	18.8049	220	3	1.1	0.9;
	4257	1	100.42	58.37	0	40.35	0	1.042258	7.921226	380	13	1.1	0.9;
	4258	1	9.51	5.9	0	0.7	0	1.010196	-30.742157	150	9	1.1	0.9;
	4259	2	0	0	0	0	0	1.01039	-41.935972	150	10	1.1	0.9;
	4260	1	0	0	0	0.54	0	1.029271	-22.100951	150	9	1.1	0.9;
	4261	1	0	-0	0	7.91	0	1.019612	2.948367	380	3	1.1	0.9;
	4262	1	0	0	0	0	0	1.010534	-27.286436	150	9	1.1	0.9;
	4263	1	-11.81	-4.02	0	0.37	0	1.007259	-23.98794	110	18	1.1	0.9;
	4264	1	0	0	0	0.74	0	1.028908	21.106834	154	1	1.1	0.9;
	4265	1	-0	-0	0	0.88	0	1.015048	-30.094475	150	9	1.1	0.9;
	4266	1	61.4	8.8	0	4.24	0	1.002825	-7.68797	154	1	1.1	0.9;
	4267	1	70.78	19.6	0	7.9	0	1.037756	-8.97903	220	5	1.1	0.9;
	4268	1	303.7	247.8	0	8.79	0	0.981228	28.310474	220	3	1.1	0.9;
	4269	1	24.93	8.2	0	0.6	0	0.978526	-32.290997	110	21	1.1	0.9;
	4270	1	128.7	25.6	0	1.32	0	1.057033	-32.004083	110	1	1.1	0.9;
	4271	1	66.97	45.1	0	2.04	0	0.946828	-17.562886	110	18	1.1	0.9;
	4272	1	184.7	71.5	0	95.42	0	1.023497	2.13201	150	8	1.1	0.9;
	4273	1	31.94	10	0	0.18	0	0.998443	-19.858555	110	18	1.1	0.9;
	4274	1	35.49	11.53	0	1.31	0	1.025662	-54.027522	220	4	1.1	0.9;
	4275	1	5.71	2.1	0	0.31	0	1.020804	-30.404515	110	19	1.1	0.9;
	4276	1	-0	0	0	60.39	0	1.028246	-15.72612	380	3	1.1	0.9;
	4277	1	-0	0	0	1.63	0	0.997034	-39.364023	150	10	1.1	0.9;
	4278	2	0	0	0	0	0	1.038839	7.739529	154	1	1.1	0.9;
	4279	2	0	0	0	0	0	1.01498	-34.516631	220	4	1.1	0.9;
	4280	1	11.91	4.3	0	0.47	0	1.038835	-7.918348	110	19	1.1	0.9;
	4281	1	-0	-0	0	11.4	0	1.046492	-11.999402	220	5	1.1	0.9;
	4282	1	22.22	6	0	0.43	0	1.004437	-19.772135	110	18	1.1	0.9;
	4283	1	-0	0	0	8.45	0	1.009168	-32.909866	220	6	1.1	0.9;
	4284	1	214.4	52.9	0	2.59	0	1.043054	-17.362805	220	7	1.1	0.9;
	4285	1	-0	-0	0	14.49	0	1.084133	45.095107	380	3	1.1	0.9;
	4286	1	-253.24	-29.59	0	10.61	0	1.022352	67.175363	150	1	1.1	0.9;
	4287	1	0.5	0.3	0	0.07	0	1.042671	-24.914763	150	9	1.1	0.9;
	4288	1	-0	-0	0	1.09	0	1.041887	-5.990756	220	3	1.1	0.9;
	4289	1	0	0	0	0	0	0.998324	-28.72802	150	9	1.1	0.9;
	4290	1	-1.9	-28.1	0	100.56	0	1.067072	-41.01452	380	4	1.1	0.9;
	4291	2	0	0	0	0	0	1.019753	-35.792407	220	4	1.1	0.9;
	4292	1	0	-0	0	7.3	0	1.03438	-20.587427	154	1	1.1	0.9;
	4293	1	80.59	7.3	0	18.92	0	1.009395	-55.082432	220	4	1.1	0.9;
	4294	1	63.87	55.1	0	14.84	0	1.028795	18.803182	380	3	1.1	0.9;
	4295	1	-0.18	-1	0	0	0	1.038508	6.226684	150	8	1.1	0.9;
	4296	2	0	0	0	0	0	1.040271	40.274729	380	3	1.1	0.9;
	4297	1	26.63	9.7	0	0.75	0	1.023326	-28.504442	110	19	1.1	0.9;
	4298	1	0	0	0	6.93	0	1.024646	3.684489	380	3	1.1	0.9;
	4299	1	0	0	0	3.56	0	1.023342	-26.016359	150	9	1.1	0.9;
	4300	1	81.49	25.1	0	0.36	0	0.997189	-16.532436	220	5	1.1	0.9;
	4301	1	-0	-0	0	2.27	0	1.035104	-21.386539	220	5	1.1	0.9;
	4302	1	30.53	9	0	1.9	0	1.023399	22.755352	110	8	1.1	0.9;
	4303	1	90.5	12.8	0	15.77	0	1.042733	-20.562838	154	1	1.1	0.9;
	4304	1	2.4	0.9	0	1.18	0	1.063231	-25.689127	110	19	1.1	0.9;
	4305	1	0	-0	0	138.24	0	1.008749	-0.91416	400	1	1.1	0.9;
	4306	1	0	-0	0	0.3	0	1.031495	-31.068518	380	5	1.1	0.9;
	4307	1	8.21	2.7	0	0.35	0	1.018022	-36.16273	110	12	1.1	0.9;
	4308	1	33.5	4.8	0	3.33	0	1.018993	-7.38201	154	1	1.1	0.9;
	4309	1	60.07	-4.7	0	0.25	0	1.066341	-25.283343	110	15	1.1	0.9;
	4310	1	0	-0	0	0	0	1.000348	-31.938279	220	11	1.1	0.9;
	4311	1	8.91	2.5	0	0.68	0	0.981183	-20.807579	110	18	1.1	0.9;
	4312	1	11.41	-3.7	0	0.1	0	1.013049	-23.93019	110	18	1.1	0.9;
	4313	1	133.4	29.4	0	10.87	0	1.040449	-21.438772	220	5	1.1	0.9;
	4314	1	26.13	2.7	0	1.7	0	1.039315	-14.542727	220	5	1.1	0.9;
	4315	1	-113.33	45.2	0	67.52	0	1.047833	-38.77663	380	4	1.1	0.9;
	4316	1	-9.01	21.8	0	4.58	0	0.963858	-26.99749	150	9	1.1	0.9;
	4317	1	-0	0	0	3.81	0	1.007687	-41.231523	220	4	1.1	0.9;
	4318	1	51.06	16.5	0	2.37	0	1.056768	10.598214	110	8	1.1	0.9;
	4319	1	55.4	7.9	0	0.12	0	1.009299	-5.109314	154	1	1.1	0.9;
	4320	1	50.35	24.3	0	3.64	0	1.002978	25.800176	220	8	1.1	0.9;
	4321	1	-0	0	0	28.03	0	1.029006	-32.125299	150	9	1.1	0.9;
	4322	2	0	0	0	0	0	1.055269	-25.21651	110	23	1.1	0.9;
	4323	2	0	0	0	0	0	1.038456	-23.365393	380	7	1.1	0.9;
	4324	1	-0	0	0	1.68	0	1.054994	-30.340052	220	5	1.1	0.9;
	4325	1	0	0	0	18.33	0	1.079412	-23.930117	380	1	1.1	0.9;
	4326	1	10.81	3.6	0	0.26	0	1.007639	-39.21017	110	12	1.1	0.9;
	4327	1	-0	-0	0.054465	9.927347	0	1.02376	-27.906617	220	18	1.1	0.9;
	4328	1	0	0	0	1.58	0	0.965235	-31.45534	150	9	1.1	0.9;
	4329	1	96.2	37.3	0	0.19	0	1.019083	-30.189885	110	23	1.1	0.9;
	4330	2	0	0	0	0	0	1.049093	-26.995714	380	6	1.1	0.9;
	4331	2	0	0	0	0	0	1.029918	-11.076786	220	5	1.1	0.9;
	4332	1	-0	-0	0	7.13	0	1.037314	34.960373	380	3	1.1	0.9;
	4333	2	0	0	0	0	0	0.997274	-30.814658	220	10	1.1	0.9;
	4334	2	0	0	0	0	0	1.057462	53.972685	380	3	1.1	0.9;
	4335	1	-0	0	0.135375	4.892175	0	1.075192	-22.438681	380	18	1.1	0.9;
	4336	1	55.46	8.4	0	5	0	1.023007	-54.854276	220	4	1.1	0.9;
	4337	2	0	0	0	0	0	1.026934	6.309425	380	8	1.1	0.9;
	4338	2	0	0	0	0	0	1.021243	-48.103156	220	4	1.1	0.9;
	4339	1	-0	-0	0	42.345005	0	0.991594	-16.075924	380	5	1.1	0.9;
	4340	1	-0	-0	0	4.03	0	1.041286	25.793343	220	8	1.1	0.9;
	4341	1	18.62	6.8	0	0.92	0	0.982706	-21.606447	110	19	1.1	0.9;
	4342	1	14.32	7.2	0	3.62	0	1.038927	-20.557735	220	7	1.1	0.9;
	4343	1	43.55	14.3	0	1.63	0	1.011953	-29.854681	110	21	1.1	0.9;
	4344	1	68.48	-10.6	0	2.65	0	0.996295	-51.660806	150	10	1.1	0.9;
	4345	1	0	0	0	18.8	0	1.029059	19.441789	380	3	1.1	0.9;
	4346	1	57	39.9	0	153.693557	0	1.121952	-22.192108	380	9	1.3	0.7;
	4347	1	31.2	4.4	0	1.61	0	1.026364	-10.160124	154	1	1.1	0.9;
	4348	1	0	-0	0	31.35	0	1.021688	-42.801815	220	6	1.1	0.9;
	4349	1	24.13	29.4	0	146.51	0	1.097775	45.321662	380	3	1.1	0.9;
	4350	2	0	0	0	0	0	1.028107	50.105767	220	3	1.1	0.9;
	4351	1	0	0	0	0.57	0	0.961099	-30.713211	110	18	1.1	0.9;
	4352	1	163.63	23	0	5.11	0	1.031777	-30.784781	220	6	1.1	0.9;
	4353	1	165.8	44	0	6.15	0	1.047865	-26.985262	220	5	1.1	0.9;
	4354	1	-0.73	-1.93	0	0.36	0	1.023532	-31.480583	220	12	1.1	0.9;
	4355	1	65.07	15	0	1.37	0	1.067756	-26.834354	220	5	1.1	0.9;
	4356	1	22.12	6.5	0	0.94	0	1.027567	-34.403559	110	12	1.1	0.9;
	4357	1	235	58.9	0	14.47	0	1.038258	-16.356336	220	5	1.1	0.9;
	4358	1	3.7	0.6	0	5.81	0	1.036903	1.417042	154	1	1.1	0.9;
	4359	1	5.11	2.3	0	0.94	0	1.016808	-12.113719	110	18	1.1	0.9;
	4360	1	20.82	8.2	0	2.96	0	1.012505	-52.773741	220	4	1.1	0.9;
	4361	1	0	-0	0.05	4.54	0	1.013688	-28.100746	220	21	1.1	0.9;
	4362	1	43.75	17.3	0	7.73	0	1.017231	3.89138	150	8	1.1	0.9;
	4363	1	0	-0	0	18.66	0	1.023014	-39.840295	380	1	1.1	0.9;
	4364	1	1.3	-0	0	1.48	0	0.993438	-24.391143	110	18	1.1	0.9;
	4365	1	-13.72	4.5	0	0.08	0	0.972435	-27.159885	150	9	1.1	0.9;
	4366	1	156.9	122.5	0	17.71	0	1.035263	-8.915283	380	3	1.1	0.9;
	4367	1	23.23	7.6	0	0.88	0	1.026057	-34.198538	110	12	1.1	0.9;
	4368	1	90.6	15.8	0	7.99	0	1.054091	-25.431523	220	5	1.1	0.9;
	4369	1	136.1	18.8	0	6.55	0	1.02512	-14.187003	220	7	1.1	0.9;
	4370	1	18.9	2.7	0	5.84	0	1.057153	1.871666	154	1	1.1	0.9;
	4371	1	72.28	6.6	0	1.42	0	1.009912	-55.34295	220	4	1.1	0.9;
	4372	1	2.8	1	0	0.56	0	1.073	-16.309536	110	19	1.1	0.9;
	4373	1	28.63	5.6	0	1.28	0	1.031784	-41.15786	110	22	1.1	0.9;
	4374	1	0	0	0	0	0	0.98249	-28.084166	150	9	1.1	0.9;
	4375	2	0	0	0	0	0	1.045136	-25.411977	220	18	1.1	0.9;
	4376	2	0	0	0	0	0	1.018032	63.599951	150	1	1.1	0.9;
	4377	1	37.6	5.4	0	5.37	0	1.020653	14.554382	154	1	1.1	0.9;
	4378	1	-0	-0	0	0.51	0	0.986067	-15.96343	110	18	1.1	0.9;
	4379	1	0	0	0	9.57	0	1.052863	1.908947	380	1	1.1	0.9;
	4380	1	-67.31	-44.04	0	0	0	1.023792	0.717698	150	8	1.1	0.9;
	4381	1	-0	0	0	3.16	0	1.031633	-44.292473	380	6	1.1	0.9;
	4382	1	35.24	12.7	0	3.4	0	1.014988	-35.836229	110	22	1.1	0.9;
	4383	1	17.82	5.9	0	0.64	0	1.018334	-29.585053	110	21	1.1	0.9;
	4384	1	33.74	11.8	0	1.66	0	1.074143	-30.832333	110	3	1.1	0.9;
	4385	1	60.97	19.7	0	2.22	0	0.998388	-37.313561	110	12	1.1	0.9;
	4386	1	16.7	2.4	0	3.18	0	1.02273	-0.206279	154	1	1.1	0.9;
	4387	1	10.91	3.6	0	0.26	0	1.007751	-36.336224	110	12	1.1	0.9;
	4388	1	-0	0	0.26	6.56	0	1.045933	-27.186622	220	19	1.1	0.9;
	4389	1	-0	0	0	20.48	0	1.039327	12.546785	380	3	1.1	0.9;
	4390	1	-0	0	0	72.64	0	1.051477	-39.489487	380	4	1.1	0.9;
	4391	1	-0	-0	0	1.23	0	1.001379	-19.656286	110	18	1.1	0.9;
	4392	1	62.07	20.7	0	6.92	0	1.044445	-27.915096	220	21	1.1	0.9;
	4393	2	0	0	0	0	0	1.061357	-44.814563	380	10	1.1	0.9;
	4394	1	-0	0	0.135375	53.752175	0	1.07967	-23.908777	380	18	1.1	0.9;
	4395	2	0	0	0	0	0	1.036974	-20.964863	220	5	1.1	0.9;
	4396	2	0	0	0	0	0	1.08226	-20.191342	380	18	1.3	0.7;
	4397	1	193	17.6	0	4.74	0	1.014447	-47.685772	220	4	1.1	0.9;
	4398	2	0	0	0	0	0	1.046214	-25.453683	380	6	1.1	0.9;
	4399	2	0	0	0	0	0	1.054017	-26.25452	110	21	1.1	0.9;
	4400	2	0	0	0	0	0	1.077395	8.102273	380	14	1.1	0.9;
	4401	1	-0	0	0	8.33	0	1.041415	-21.104843	220	3	1.1	0.9;
	4402	1	-0	0	0	1.02	0	1.082792	-15.473491	220	5	1.1	0.9;
	4403	1	35.24	12.3	0	0.3	0	1.05117	-28.87403	110	3	1.1	0.9;
	4404	1	0	0	0	0	0	0.95719	-29.999812	150	9	1.1	0.9;
	4405	1	-15.02	4.9	0	0.26	0	0.984433	-23.606104	150	9	1.1	0.9;
	4406	1	29.13	9.6	0	0.76	0	1.020663	-28.695737	110	21	1.1	0.9;
	4407	1	-0	-0	0	51.43	0	1.061264	-41.277281	380	4	1.1	0.9;
	4408	1	11.01	3.3	0	0.67	0	1.010367	-41.467274	110	22	1.1	0.9;
	4409	1	-5.15	-16.03	0	0	0	1.036468	17.146109	150	8	1.1	0.9;
	4410	1	45.05	13.2	0	44.93	0	1.041145	-37.170777	220	5	1.1	0.9;
	4411	1	-0	0	0	6.72	0	1.019039	-9.031289	220	3	1.1	0.9;
	4412	1	19.4	2.8	0	2.53	0	1.007967	-3.533645	154	1	1.1	0.9;
	4413	1	0	0	0	13.44	0	1.025355	11.098662	150	8	1.1	0.9;
	4414	1	303.6	47.5	0	14.96	0	1.027205	-34.077804	380	6	1.1	0.9;
	4415	1	30.03	9.7	0	4.1	0	1.029972	7.953503	110	8	1.1	0.9;
	4416	1	117.4	12.5	0	5.16	0	1.024783	-13.991666	220	7	1.1	0.9;
	4417	1	0	-0	0	5.12	0	1.016455	-37.569258	220	22	1.1	0.9;
	4418	1	-0	0	0	0.05	0	1.074227	-24.636521	220	5	1.1	0.9;
	4419	2	0	0	0	0	0	1.055121	4.46001	380	5	1.1	0.9;
	4420	2	0	0	0	0	0	1.009808	-10.334322	154	1	1.1	0.9;
	4421	1	32.84	9.3	0	0.76	0	0.977645	-28.44939	110	18	1.1	0.9;
	4422	1	77.19	47.8	0	2.08	0	0.980745	-31.583482	150	9	1.1	0.9;
	4423	1	-0	-0	0	0.33	0	1.031521	-11.573048	150	2	1.1	0.9;
	4424	1	0	-0	0	0.51	0	1.028537	-16.531217	220	3	1.1	0.9;
	4425	1	-0	0	0	3.22	0	0.986025	9.380075	220	3	1.1	0.9;
	4426	1	485.5	-63.38	0	34.92	0	1.06424	-47.083696	220	5	1.1	0.9;
	4427	1	296.8	77.7	0	0	0	1.029896	-11.3652	150	2	1.1	0.9;
	4428	2	0	0	0	0	0	1.016576	-44.153521	220	4	1.1	0.9;
	4429	2	0	0	0	0	0	1.007433	-4.171966	154	1	1.1	0.9;
	4430	1	-0	0	0	0.68	0	1.048812	-20.863288	220	1	1.1	0.9;
	4431	1	19.52	5.6	0	1.46	0	1.017867	8.169606	150	8	1.1	0.9;
	4432	1	-0	0	0	0.8	0	1.068843	-26.73362	220	5	1.1	0.9;
	4433	1	142.49	523.51	0	34.96	0	1.051614	-15.841626	380	6	1.1	0.9;
	4434	1	60.07	45	0	0.12	0	1.032779	24.876477	220	3	1.1	0.9;
	4435	1	-0	-0	0	102.217726	0	1.010783	-8.696684	380	5	1.1	0.9;
	4436	1	0.01	-0.01	0	420.1	0	1.052317	-15.83616	380	6	1.1	0.9;
	4437	1	0	-0	0.17	11.68	0	1.026437	-26.750642	220	21	1.1	0.9;
	4438	1	5.51	2	0	0.69	0	1.063656	-26.075553	110	19	1.1	0.9;
	4439	1	13.82	4.5	0	0.25	0	1.024658	-34.268009	110	12	1.1	0.9;
	4440	2	0	0	0	0	0	1.012622	62.214553	150	1	1.1	0.9;
	4441	1	1.3	0.4	0	1.43	0	1.02633	-10.922726	110	18	1.1	0.9;
	4442	2	0	0	0	0	0	1.031243	-50.622294	380	4	1.1	0.9;
	4443	1	5.21	1	0	0.94	0	1.007119	-28.314443	110	18	1.1	0.9;
	4444	1	16.52	4.8	0	0.33	0	0.998693	-24.624089	110	18	1.1	0.9;
	4445	2	0	0	0	0	0	1.031204	-37.688891	380	6	1.1	0.9;
	4446	1	0	-0	0	9.25	0	1.035897	-39.478385	380	4	1.1	0.9;
	4447	2	0	0	0	0	0	1.115565	-23.575874	380	21	1.3	0.7;
	4448	1	182.18	36.4	0	7.79	0	0.976159	-53.121709	220	6	1.1	0.9;
	4449	1	32.8	4.7	0	8.02	0	1.025407	4.559943	154	1	1.1	0.9;
	4450	1	140.2	11.2	0	9.1	0	1.023587	-54.685261	220	4	1.1	0.9;
	4451	1	8.91	8.9	0	0.58	0	0.960231	-27.579716	110	18	1.1	0.9;
	4452	1	0	-0	0	0.03	0	1.031824	7.264218	154	1	1.1	0.9;
	4453	1	52.6	110.6	0	2.72	0	1.041997	-11.076954	220	7	1.1	0.9;
	4454	1	-0	-0	0	1	0	1.006626	-17.657551	220	5	1.1	0.9;
	4455	2	0	0	0	0	0	1.01091	-9.685572	400	1	1.1	0.9;
	4456	1	0	-0	0	2.06	0	1.001715	-29.833816	150	9	1.1	0.9;
	4457	1	55.7	7.9	0	2.29	0	1.015266	0.37257	154	1	1.1	0.9;
	4458	1	413	-353	0	8.88	0	1.060133	-20.243328	400	1	1.1	0.9;
	4459	1	5.5	0.8	0	1.57	0	1.02497	-4.673317	154	1	1.1	0.9;
	4460	1	0	0	0	39.29	0	1.105887	-24.716121	380	9	1.3	0.7;
	4461	1	-0	-0	0	0.31	0	1.022586	-31.689003	150	9	1.1	0.9;
	4462	1	22.42	4.8	0	0.53	0	1.035045	-28.394827	110	3	1.1	0.9;
	4463	1	0	-0	0.083969	-0.251906	0	1.016703	-37.138163	220	11	1.1	0.9;
	4464	2	0	0	0	0	0	1.054883	-29.186919	150	9	1.1	0.9;
	4465	1	15.12	5	0	0.53	0	0.990701	-31.137598	110	21	1.1	0.9;
	4466	1	0	0	0	7.75	0	0.998815	9.954113	220	3	1.1	0.9;
	4467	1	-0	0	0	7.22	0	0.994136	-46.767778	220	4	1.1	0.9;
	4468	2	0	0	0	0	0	1.041888	-6.97892	220	13	1.1	0.9;
	4469	1	-0	-0	0	53.34	0	1.016104	6.117545	400	1	1.1	0.9;
	4470	1	0	-0	0	6.45	0	1.017342	-38.072059	220	1	1.1	0.9;
	4471	1	37.44	-28.3	0	9.28	0	1.014999	6.95466	220	3	1.1	0.9;
	4472	1	0	-0	0	31.6	0	1.022976	2.582263	380	3	1.1	0.9;
	4473	1	0	-0	0	0.05	0	1.089071	-27.765632	150	9	1.1	0.9;
	4474	1	0	0	0	9.72	0	1.020183	18.753546	380	3	1.1	0.9;
	4475	1	49.45	30.6	0	1.57	0	0.934086	-33.42388	150	9	1.1	0.9;
	4476	1	24.93	15.4	0	5.48	0	0.97125	-37.361789	150	9	1.1	0.9;
	4477	1	-0.2	-1.64	0	0	0	1.03099	3.60651	150	8	1.1	0.9;
	4478	2	0	0	0	0	0	1.062733	47.922029	380	3	1.1	0.9;
	4479	1	0	0	0	2.13	0	0.96163	-31.626246	150	9	1.1	0.9;
	4480	2	0	0	0	0	0	1.065482	7.961387	220	5	1.1	0.9;
	4481	1	0	0	0	2.7	0	1.016911	-33.841262	220	1	1.1	0.9;
	4482	2	0	0	0	0	0	1.043069	-6.662274	220	5	1.1	0.9;
	4483	1	0	0	0	0	0	1.044395	-29.479203	150	9	1.1	0.9;
	4484	1	69.78	25	0	14.24	0	1.042834	-13.696329	220	5	1.1	0.9;
	4485	1	0	0	0	18.32	0	1.041056	15.445155	380	8	1.1	0.9;
	4486	2	0	0	0	0	0	1.026359	-26.73494	110	18	1.1	0.9;
	4487	1	-0	0	0	0.24	0	1.021496	-41.49216	380	6	1.1	0.9;
	4488	1	71.88	-2.2	0	0.07	0	1.04134	-21.120784	220	3	1.1	0.9;
	4489	1	0	-0	0	5.9	0	1.021322	-47.40207	220	4	1.1	0.9;
	4490	1	-0	0	0	12.29	0	0.993329	-37.832964	220	22	1.1	0.9;
	4491	1	0	-0	0	1.75	0	1.02361	-30.715686	220	5	1.1	0.9;
	4492	1	-0	0	0	18.07	0	1.045494	38.496563	380	3	1.1	0.9;
	4493	1	15.32	18.2	0	27.62	0	1.018268	-40.970053	220	4	1.1	0.9;
	4494	1	-0	0	0	55.514618	0	1.016149	-1.082662	380	5	1.1	0.9;
	4495	1	336.3	104.4	0	0	0	1.028274	-11.333961	150	2	1.1	0.9;
	4496	1	0	-0	0	2.91	0	1.00363	11.171429	220	3	1.1	0.9;
	4497	1	-0	0	0	0.7	0	0.966015	-16.2589	110	18	1.1	0.9;
	4498	2	0	0	0	0	0	1.085986	45.525024	380	3	1.1	0.9;
	4499	1	80.09	20	0	6.98	0	1.060005	43.379649	220	3	1.1	0.9;
	4500	1	199.3	53.7	0	12.54	0	1.014972	14.443444	380	3	1.1	0.9;
	4501	1	65.6	-382	0.218901	57.448127	0	1.028081	34.889783	380	3	1.1	0.9;
	4502	2	0	0	0	0	0	1.022423	-45.514512	220	6	1.1	0.9;
	4503	1	20.72	6.8	0	0.69	0	1.000822	-37.50556	110	12	1.1	0.9;
	4504	1	170.1	-9.3	0	5.97	0	1.07122	-42.503881	220	5	1.1	0.9;
	4505	1	95.51	17.1	0	8.5	0	1.038528	-16.322612	220	5	1.1	0.9;
	4506	2	0	0	0	0	0	1.032581	-6.338408	220	5	1.1	0.9;
	4507	1	-0	0	0	43.44	0	1.026647	-42.167898	380	4	1.1	0.9;
	4508	1	55.98	18.47	0	9.45	0	1.01112	61.583042	150	1	1.1	0.9;
	4509	1	77.99	30.8	0	2.41	0	1.030757	2.155188	150	8	1.1	0.9;
	4510	1	75.38	21.8	0	0.63	0	1.019973	-39.369587	220	4	1.1	0.9;
	4511	1	81.89	19.9	0	5.1	0	1.037521	-26.269073	220	5	1.1	0.9;
	4512	1	-0	0	0	12.9	0	1.019063	-46.353713	380	6	1.1	0.9;
	4513	1	92.1	29.5	0	3.14	0	1.030937	-9.873276	220	5	1.1	0.9;
	4514	1	60.9	8.7	0	1.34	0	1.022975	4.738024	154	1	1.1	0.9;
	4515	1	-0	-0	0	0	0	1.024584	-25.074941	110	3	1.1	0.9;
	4516	2	0	0	0	0	0	1.018246	-47.782987	220	6	1.1	0.9;
	4517	1	7	1	0	3.58	0	1.016047	1.046488	154	1	1.1	0.9;
	4518	1	12.91	4	0	1.66	0	1.074936	-17.283865	110	19	1.1	0.9;
	4519	1	97.5	13.9	0	29.65	0	1.024115	7.200395	400	1	1.1	0.9;
	4520	1	-0	-0	0	0.94	0	1.01536	-10.191142	220	5	1.1	0.9;
	4521	1	349.6	22	0	10.29	0	1.022583	-42.115886	380	6	1.1	0.9;
	4522	1	91.1	36.1	0	1.01	0	0.97151	10.604814	220	3	1.1	0.9;
	4523	1	114.2	16.3	0	2.85	0	1.020917	2.948453	154	1	1.1	0.9;
	4524	1	19.4	2.8	0	1.61	0	1.005634	-4.312594	154	1	1.1	0.9;
	4525	1	-0	0	0	6.12	0	1.053159	-9.223606	220	5	1.1	0.9;
	4526	2	0	0	0	0	0	1.059432	-48.715162	220	6	1.1	0.9;
	4527	1	34.04	21.1	0	0	0	0.988635	-31.00334	150	9	1.1	0.9;
	4528	1	3.2	8.3	0	0.38	0	0.95621	-32.123967	150	9	1.1	0.9;
	4529	1	48.48	8.7	0	10.14	0	1.056482	-15.446113	220	5	1.1	0.9;
	4530	1	113.6	16.2	0	1.83	0	1.017749	-3.100429	154	1	1.1	0.9;
	4531	1	15.32	9.5	0	5.33	0	0.98249	-28.084166	150	9	1.1	0.9;
	4532	1	5.11	1.9	0	2.31	0	0.992948	-16.6809	110	18	1.1	0.9;
	4533	2	0	0	0	0	0	1.087859	12.675438	380	13	1.3	0.7;
	4534	1	3.2	0.4	0	0.65	0	1.00848	-20.386669	110	18	1.1	0.9;
	4535	1	-0	-0	0.174597	35.618105	0	1.02934	9.534577	380	3	1.1	0.9;
	4536	1	14.92	5.4	0	0.97	0	1.042638	-23.368692	110	19	1.1	0.9;
	4537	1	7.71	1.3	0	2.44	0	0.995045	-31.877583	110	18	1.1	0.9;
	4538	1	-14.82	-5.03	0	0.15	0	1.018205	-26.124455	110	18	1.1	0.9;
	4539	1	-15.88	-4.9	0	0.29	0	1.031929	-38.517056	110	22	1.1	0.9;
	4540	1	0	-0	0	14.72	0	1.025359	-24.523057	220	7	1.1	0.9;
	4541	1	0	-0	0	150.797812	0	1.022812	-9.949699	380	5	1.1	0.9;
	4542	1	-0	0	0	38.49	0	1.045397	-43.550688	380	4	1.1	0.9;
	4543	1	-0	0	0	2.32	0	1.052729	-45.804884	380	10	1.1	0.9;
	4544	1	-0	-0	0	14.64	0	1.044954	-14.795371	220	5	1.1	0.9;
	4545	1	21.12	13.1	0	0.18	0	0.962682	-27.37169	150	9	1.1	0.9;
	4546	1	2.8	0.5	0	2.37	0	1.003212	-27.884307	110	18	1.1	0.9;
	4547	1	33.04	10.9	0	0.97	0	0.978811	-32.268679	110	21	1.1	0.9;
	4548	1	20.02	9	0	0.67	0	1.015228	-31.090387	110	23	1.1	0.9;
	4549	1	0	-0	0	0.3	0	1.029194	-11.729101	220	5	1.1	0.9;
	4550	1	0	-0	0	124.2	0	1.017355	-10.569179	380	5	1.1	0.9;
	4551	1	0	0	0	1.33	0	1.020947	-29.722641	150	9	1.1	0.9;
	4552	2	0	0	0	0	0	1.021862	3.978169	154	1	1.1	0.9;
	4553	1	179.8	76.5	0	8.05	0	1.027689	-14.030999	220	7	1.1	0.9;
	4554	1	119	21.9	0	0.95	0	1.037349	-10.05047	220	5	1.1	0.9;
	4555	1	48.45	16	0	0.47	0	0.98347	-31.80027	110	21	1.1	0.9;
	4556	1	138.2	21.3	0	2.33	0	0.973834	-49.622386	220	4	1.1	0.9;
	4557	1	93	66	0	2.25	0	0.969948	13.501715	220	3	1.1	0.9;
	4558	1	6.41	2.3	0	0.78	0	1.050132	-26.127932	110	19	1.1	0.9;
	4559	1	0	-0	0	1.1	0	1.045792	-25.014876	110	15	1.1	0.9;
	4560	1	13.92	6.3	0	0.37	0	0.923247	-33.083728	110	18	1.1	0.9;
	4561	1	101	20.6	0	54.5	0	1.023444	-44.433238	380	6	1.1	0.9;
	4562	1	130.5	25.7	0	2.27	0	1.018826	-16.997061	220	5	1.1	0.9;
	4563	1	-9.87	-3.77	0	0.42	0	1.010732	-24.420942	110	18	1.1	0.9;
	4564	2	0	0	0	0	0	1.026639	-13.312726	220	6	1.1	0.9;
	4565	1	0	0	0	0	0	1.023351	11.178136	150	8	1.1	0.9;
	4566	2	0	0	0	0	0	1.07048	-27.96894	220	5	1.1	0.9;
	4567	1	92.4	13.1	0	98.17	0	0.999125	-3.513287	400	1	1.1	0.9;
	4568	1	0	0	0	0.31	0	1.0034	-33.847076	150	9	1.1	0.9;
	4569	1	-0	-0	0.158724	-1.020369	0	1.025197	-24.327029	220	3	1.1	0.9;
	4570	1	11.91	4.3	0	1.68	0	1.051768	-21.390504	110	19	1.1	0.9;
	4571	1	73.28	9.4	0	24.65	0	1.051203	32.904719	380	3	1.1	0.9;
	4572	1	47.95	16.3	0	3.46	0	1.008625	-36.67786	220	6	1.1	0.9;
	4573	2	0	0	0	0	0	1.021097	-7.845838	154	1	1.1	0.9;
	4574	1	-31.26	-13.2	0	2.81	0	1.009484	-49.307569	220	6	1.1	0.9;
	4575	1	8.51	2.8	0	1.16	0	1.008687	-36.601691	110	12	1.1	0.9;
	4576	1	311.4	43	0	37.55	0	1.022715	-45.023038	380	6	1.1	0.9;
	4577	1	2.2	0.6	0	0.04	0	0.984859	-56.244171	150	10	1.1	0.9;
	4578	1	50.66	56.5	0	6.86	0	1.05528	10.516173	380	3	1.1	0.9;
	4579	1	19.52	6.4	0	0.03	0	1.043825	-29.528517	150	9	1.1	0.9;
	4580	1	11.01	3.9	0	7.98	0	1.040913	-12.304623	220	5	1.1	0.9;
	4581	1	36.28	-45.18	0	0.84	0	1.053606	10.541803	220	13	1.1	0.9;
	4582	1	13.9	1.9	0	4.25	0	1.034802	2.707797	154	1	1.1	0.9;
	4583	1	-0	0	0	2.07	0	1.045026	-32.815045	110	22	1.1	0.9;
	4584	1	-0	0	0	0.52	0	1.060256	-16.482948	110	18	1.1	0.9;
	4585	1	-0	-0	0	0.16	0	1.001306	-27.676187	110	18	1.1	0.9;
	4586	2	0	0	0	0	0	1.037111	-11.328305	150	2	1.1	0.9;
	4587	1	12.21	7.9	0	0.56	0	1.017057	-11.607643	110	18	1.1	0.9;
	4588	1	-0	0	0	75.38	0	1.044309	-50.448008	380	4	1.1	0.9;
	4589	1	5.77	2.68	0	1.72	0	1.035839	-30.232915	150	9	1.1	0.9;
	4590	1	-6.7	-7.5	0	1.48	0	1.048353	-27.370587	150	9	1.1	0.9;
	4591	1	0	0	0	3.16	0	1.022447	-35.199569	110	22	1.1	0.9;
	4592	1	183.8	2.2	0	2.98	0	0.975653	-49.288883	220	4	1.1	0.9;
	4593	1	20.32	-24.8	0	7.45	0	1.02754	7.892834	220	3	1.1	0.9;
	4594	1	-1.73	-1.71	0	22.062314	0	0.993301	-5.867396	380	5	1.1	0.9;
	4595	1	84.59	42.82	0	0.86	0	1.038273	10.102597	220	13	1.1	0.9;
	4596	1	4.3	1.4	0	1.59	0	1.026034	-35.157759	110	12	1.1	0.9;
	4597	1	-0	-0	0	4.25	0	1.008674	36.796684	220	3	1.1	0.9;
	4598	1	-0	0	0	52.958443	0	1.025605	-7.967697	380	5	1.1	0.9;
	4599	1	0	-0	0	1.27	0	1.067097	-20.860508	110	19	1.1	0.9;
	4600	1	19.12	7.4	0	21.97	0	1.02798	-50.447526	380	4	1.1	0.9;
	4601	1	0	-0	0	37.74	0	1.016828	62.846259	150	1	1.1	0.9;
	4602	2	0	0	0	0	0	1.120537	-23.246353	380	21	1.3	0.7;
	4603	1	70.28	10.3	0	13.73	0	1.014218	-52.607022	220	4	1.1	0.9;
	4604	1	0	-0	0	4.36	0	1.046663	-27.065702	150	9	1.1	0.9;
	4605	1	10.11	3.7	0	2.15	0	1.065277	-25.758953	110	19	1.1	0.9;
	4606	1	-0	-0	0	1.69	0	1.033837	-39.750273	380	6	1.1	0.9;
	4607	1	10.11	0	0	0.32	0	1.0072	-20.965267	110	18	1.1	0.9;
	4608	2	0	0	0	0	0	1.034566	14.064182	154	1	1.1	0.9;
	4609	2	0	0	0	0	0	1.019251	-7.112921	154	1	1.1	0.9;
	4610	1	-0	0	0	16.93	0	1.040483	-41.949678	380	6	1.1	0.9;
	4611	1	0	-0	0	2.72	0	1.025155	-4.184983	154	1	1.1	0.9;
	4612	1	16.52	5.5	0	0.75	0	1.037926	-33.206689	110	22	1.1	0.9;
	4613	2	0	0	0	0	0	1.021632	7.441997	400	1	1.1	0.9;
	4614	1	74.18	14.6	0	11.71	0	1.024787	7.57852	110	8	1.1	0.9;
	4615	1	42.55	17.6	0	1.2	0	1.059263	-25.746141	220	5	1.1	0.9;
	4616	1	30.83	16.5	0	3.83	0	1.014257	16.410389	150	8	1.1	0.9;
	4617	1	64.57	10	0	4.8	0	1.009059	-26.346838	110	18	1.1	0.9;
	4618	1	1.9	0.6	0	1.06	0	0.997564	-39.017154	110	12	1.1	0.9;
	4619	1	74.68	46.3	0	4.98	0	0.997745	-37.019205	150	9	1.1	0.9;
	4620	2	0	0	0	0	0	1.089892	37.788886	380	3	1.3	0.7;
	4621	1	26.03	16.1	0	0	0	1.016409	-30.703143	150	9	1.1	0.9;
	4622	1	40.04	15	0	1.59	0	0.991439	27.885255	220	3	1.1	0.9;
	4623	1	0	-0	0	80.388213	0	0.990686	-17.130144	380	5	1.1	0.9;
	4624	2	0	0	0	0	0	1.045235	-15.735017	220	5	1.1	0.9;
	4625	1	162.1	73	0	0.3	0	1.022752	-30.753846	110	23	1.1	0.9;
	4626	1	44.6	6.4	0	2.31	0	1.025049	-1.48635	154	1	1.1	0.9;
	4627	1	7	1	0	0.87	0	1.002736	-12.227064	154	1	1.1	0.9;
	4628	1	194.1	27.6	0	3.26	0	1.014584	-10.135515	154	1	1.1	0.9;
	4629	1	16.7	2.4	0	14.37	0	1.035957	-10.334885	154	1	1.1	0.9;
	4630	1	18.2	2.6	0	1.59	0	1.040554	-21.071197	154	1	1.1	0.9;
	4631	1	350	129.4	0	33.52	0	1.034671	-20.329864	220	7	1.1	0.9;
	4632	1	0	-0	0	3.21	0	1.013277	-48.949226	220	4	1.1	0.9;
	4633	2	0	0	0	0	0	1.009572	-8.700854	154	1	1.1	0.9;
	4634	1	137.8	64.5	0	5.41	0	1.016909	-11.763164	380	3	1.1	0.9;
	4635	1	16.12	4.3	0	0.49	0	1.050774	-25.185846	110	23	1.1	0.9;
	4636	1	-0	0	0	0.11	0	1.028166	-34.087026	110	12	1.1	0.9;
	4637	1	75.58	-7.5	0	8.08	0	1.019606	-13.307458	380	3	1.1	0.9;
	4638	2	0	0	0	0	0	1.050874	-6.909462	380	13	1.1	0.9;
	4639	1	17.92	7.2	0	0.8	0	1.012845	-27.830536	110	21	1.1	0.9;
	4640	1	-0	-0	0	1.59	0	1.029956	4.087062	154	1	1.1	0.9;
	4641	1	1.7	0.6	0	0.51	0	0.993058	-27.153987	110	18	1.1	0.9;
	4642	1	-0	-0	0	1.45	0	1.044207	-24.904203	150	9	1.1	0.9;
	4643	1	3.7	1.3	0	0.9	0	1.04714	-25.703503	110	19	1.1	0.9;
	4644	1	18.22	8.6	0	34.7	0	1.030242	-44.573358	380	4	1.1	0.9;
	4645	1	-0	0	0.305544	0.218877	0	1.028738	1.67222	380	3	1.1	0.9;
	4646	1	0	0	0	2.07	0	1.043008	-6.663752	220	3	1.1	0.9;
	4647	2	0	0	0	0	0	1.050276	33.029812	380	3	1.1	0.9;
	4648	1	41.75	13.2	0	3.36	0	1.032876	-33.918645	110	22	1.1	0.9;
	4649	1	5.01	2	0	0.77	0	1.031923	-28.555693	110	23	1.1	0.9;
	4650	1	-0	0	0	0	0	0.988431	-54.38092	220	10	1.1	0.9;
	4651	1	0	0	0	8.1	0	1.024869	-54.400941	220	4	1.1	0.9;
	4652	1	-0	0	0	165.63	0	1.060738	-45.351003	380	4	1.1	0.9;
	4653	1	0	0	0	0.84	0	0.988894	-17.901615	110	18	1.1	0.9;
	4654	2	0	0	0	0	0	1.017349	-21.71044	154	1	1.1	0.9;
	4655	1	4.61	1.7	0	0.88	0	1.057375	-25.146177	110	19	1.1	0.9;
	4656	1	93.9	-0.4	0	3.66	0	1.041932	-20.065226	220	5	1.1	0.9;
	4657	2	0	0	0	0	0	1.020179	-38.717477	220	4	1.1	0.9;
	4658	1	51.06	7.4	0	2.49	0	1.006975	-47.89191	220	4	1.1	0.9;
	4659	1	27.13	16.8	0	0.28	0	1.061656	-21.366509	150	9	1.1	0.9;
	4660	1	16.52	3.2	0	0.9	0	0.960681	-29.361569	110	18	1.1	0.9;
	4661	2	0	0	0	0	0	1.056504	-26.712402	220	5	1.1	0.9;
	4662	2	0	0	0	0	0	1.066687	43.668995	380	3	1.1	0.9;
	4663	2	0	0	0	0	0	1.021171	-50.126756	220	4	1.1	0.9;
	4664	1	-0	-0	0	0.33	0	0.995969	-19.109257	110	18	1.1	0.9;
	4665	1	390.5	62.5	0	0	0	1.028821	-11.97837	150	2	1.1	0.9;
	4666	2	0	0	0	0	0	1.011313	-20.889278	220	6	1.1	0.9;
	4667	2	0	0	0	0	0	1.01948	-28.508662	110	18	1.1	0.9;
	4668	1	28.03	10.2	0	0.83	0	1.014022	-30.214599	110	19	1.1	0.9;
	4669	1	11.41	5.1	0	0.15	0	1.014375	-23.471051	110	18	1.1	0.9;
	4670	2	0	0	0	0	0	1.092961	49.659948	380	3	1.1	0.9;
	4671	1	44.35	14.8	0	1.94	0	1.008783	-31.767138	110	21	1.1	0.9;
	4672	1	91.2	56.1	0	8.3	0	1.009819	-32.361289	220	6	1.1	0.9;
	4673	1	0	-0	0	1.06	0	0.93836	-33.379623	150	9	1.1	0.9;
	4674	1	124	29.2	0	0.53	0	1.027034	-12.764531	220	5	1.1	0.9;
	4675	1	0.2	0.8	0	0.04	0	1.033192	24.896508	220	3	1.1	0.9;
	4676	2	0	0	0	0	0	1.088792	-11.699775	380	18	1.3	0.7;
	4677	1	0	0	0	0.66	0	1.054278	-29.798853	220	1	1.1	0.9;
	4678	1	25.13	8.2	0	0.89	0	1.051163	-17.647251	110	21	1.1	0.9;
	4679	1	102.7	17.8	0	48.05	0	1.03925	-20.91864	220	5	1.1	0.9;
	4680	1	344.7	-31.1	0	39.129841	0	1.043133	-27.420476	380	6	1.1	0.9;
	4681	1	84.09	1	0	32.3	0	1.038752	-18.976871	380	11	1.1	0.9;
	4682	1	2.5	1.3	0	1.55	0	0.92094	-33.044671	110	18	1.1	0.9;
	4683	1	38.84	8.5	0	7.31	0	1.049891	-3.935934	220	5	1.1	0.9;
	4684	1	-0	-0	0	0.36	0	1.04512	-20.498566	110	19	1.1	0.9;
	4685	1	46.15	11.2	0	10.16	0	1.038585	-26.129161	220	5	1.1	0.9;
	4686	1	16.02	5.1	0	0.63	0	1.034458	8.213903	110	8	1.1	0.9;
	4687	1	0	0	0	0	0	1.016575	0.400195	150	8	1.1	0.9;
	4688	1	0	0	0	43.7	0	1.018329	-1.583907	400	1	1.1	0.9;
	4689	1	109.7	15	0	30.97	0	1.031744	-31.539915	220	5	1.1	0.9;
	4690	2	0	0	0	0	0	1.026559	-43.415903	380	4	1.1	0.9;
	4691	1	-0	8.3	0	10.19	0	1.084841	-36.918755	150	9	1.3	0.7;
	4692	1	-0	0	0	50	0	1.016442	-0.965415	150	8	1.1	0.9;
	4693	1	29.38	-14.09	0	1.33	0	1.051673	-22.960021	110	19	1.1	0.9;
	4694	1	10.81	3.9	0	0.61	0	1.045223	-25.898854	110	15	1.1	0.9;
	4695	2	0	0	0	0	0	1.053968	-0.654287	220	13	1.1	0.9;
	4696	1	15.82	7.2	0	1.9	0	0.984964	-28.429627	110	18	1.1	0.9;
	4697	1	-0	0	0.147967	92.214944	0	1.089426	-14.435564	380	18	1.3	0.7;
	4698	1	25.33	8.8	0	1.39	0	1.018117	-34.713039	110	22	1.1	0.9;
	4699	1	-0	-0	0	3.4	0	0.956699	-33.348477	150	9	1.1	0.9;
	4700	1	162	83.6	0	5.32	0	1.009378	-26.081154	220	7	1.1	0.9;
	4701	2	0	0	0	0	0	1.042354	-10.260507	380	5	1.1	0.9;
	4702	1	136.8	4.8	0	10	0	1.03984	-14.490377	220	3	1.1	0.9;
	4703	1	19.72	7.1	0	0.28	0	1.016597	-24.867765	110	15	1.1	0.9;
	4704	1	0	-0	0.304554	30.256205	0	1.075406	-35.989052	380	10	1.1	0.9;
	4705	1	19.62	7.4	0	0.76	0	1.062823	-21.501559	110	15	1.1	0.9;
	4706	1	0	-0	0	20.81	0	1.056106	-46.58034	380	4	1.1	0.9;
	4707	1	-0	-0	0	1.74	0	1.015378	-32.311818	150	9	1.1	0.9;
	4708	2	0	0	0	0	0	0.996652	-17.478391	110	18	1.1	0.9;
	4709	1	0	-0	0	27.85	0	1.017401	4.166487	380	3	1.1	0.9;
	4710	1	195.4	76.84	0	8.59	0	1.025123	-9.245807	220	5	1.1	0.9;
	4711	1	-0	0	0	16.11	0	1.037466	-11.707006	220	5	1.1	0.9;
	4712	1	-0	0	0.073728	5.986272	0	1.030273	-34.784151	220	10	1.1	0.9;
	4713	1	14.22	5.1	0	0.39	0	1.050161	-25.455576	110	15	1.1	0.9;
	4714	1	-0	0	0	13.16	0	1.014134	3.843624	380	3	1.1	0.9;
	4715	1	-0	0	0	23.37	0	1.041648	-48.667335	220	6	1.1	0.9;
	4716	1	392	177.7	0.181549	10.880889	0	1.038999	-19.455771	220	18	1.1	0.9;
	4717	1	0	0	0	46.25	0	1.039748	-42.848896	380	4	1.1	0.9;
	4718	1	97	57.4	0	93.82	0	1.029209	-11.803938	380	3	1.1	0.9;
	4719	1	-2.91	-0.6	0	0.02	0	1.010182	-36.984842	220	6	1.1	0.9;
	4720	1	-0	-0	0	4.51	0	1.019139	62.912023	150	1	1.1	0.9;
	4721	1	0	-0	0	6.11	0	0.988034	-55.007165	150	10	1.1	0.9;
	4722	1	19.4	2.8	0	11.38	0	1.036827	12.102438	154	1	1.1	0.9;
	4723	1	-0	-0	0	35.78	0	1.017996	3.131543	380	3	1.1	0.9;
	4724	1	0	0	0	0	0	1.014721	-32.352939	150	9	1.1	0.9;
	4725	1	4.91	1.6	0	4.15	0	1.012202	-10.659528	220	5	1.1	0.9;
	4726	1	47.55	-1.41	0	0.76	0	0.96405	-31.91617	110	18	1.1	0.9;
	4727	1	2	0.6	0	0.53	0	0.976405	-32.063514	110	21	1.1	0.9;
	4728	1	3.1	0.1	0	1.34	0	1.037292	-21.668108	220	5	1.1	0.9;
	4729	1	0	0	0	2.38	0	1.026533	-9.241781	220	5	1.1	0.9;
	4730	1	0	0	0	0.81	0	0.956024	-29.933031	150	9	1.1	0.9;
	4731	1	-0	-0	0	2.42	0	1.038302	-23.13186	150	9	1.1	0.9;
	4732	1	17.7	-334.5	0.132526	55.711068	0	1.055652	17.016439	380	3	1.1	0.9;
	4733	1	-0	0	0	0	0	1.047928	18.07872	380	8	1.1	0.9;
	4734	1	0	0	0	25.339116	0	1.019192	-10.274453	380	5	1.1	0.9;
	4735	1	0	0	0	13.59	0	0.985176	-44.869375	380	1	1.1	0.9;
	4736	1	0	0	0.099501	43.903596	0	1.077459	-15.116153	380	18	1.1	0.9;
	4737	1	277.9	143.1	0	0.69	0	1.009608	0.739976	380	3	1.1	0.9;
	4738	1	101.89	25.72	0	5.11	0	1.06023	-11.244543	220	5	1.1	0.9;
	4739	2	0	0	0	0	0	1.021645	-12.210831	110	18	1.1	0.9;
	4740	1	-0.15	-0.9	0	0	0	1.033532	16.621621	110	8	1.1	0.9;
	4741	1	10.5	7.32	0	0.24	0	0.984405	-16.0631	110	18	1.1	0.9;
	4742	1	-0	0	0	2.03	0	1.019517	-0.458767	380	3	1.1	0.9;
	4743	1	13.77	18.73	0	0.46	0	1.063165	-33.996846	150	9	1.1	0.9;
	4744	1	2.2	0	0	4.81	0	0.974785	-41.675373	220	10	1.1	0.9;
	4745	1	63.67	21.3	0	3.69	0	1.038799	-22.569603	220	7	1.1	0.9;
	4746	1	0	0	0.181549	17.630889	0	1.02826	-27.196577	220	18	1.1	0.9;
	4747	1	93.3	27.8	0	6.1	0	1.030442	-12.622107	220	5	1.1	0.9;
	4748	1	-0	-0	0	1.49	0	1.029885	-12.614931	220	5	1.1	0.9;
	4749	2	0	0	0	0	0	1.030135	1.76446	154	1	1.1	0.9;
	4750	1	137.7	32.6	0	5.17	0	1.007905	-45.954364	220	4	1.1	0.9;
	4751	1	0	0	0	0	0	1.013847	-32.766687	150	9	1.1	0.9;
	4752	1	101.1	19.8	0	3.62	0	1.003304	-48.431638	220	4	1.1	0.9;
	4753	1	-0	-0	0	1.4	0	1.0366	59.160645	220	3	1.1	0.9;
	4754	1	59.17	14	0	4.07	0	0.987901	-23.107942	220	3	1.1	0.9;
	4755	2	0	0	0	0	0	1.0132	-5.020784	154	1	1.1	0.9;
	4756	1	10.97	-6.31	0	1.14	0	0.919635	-36.305721	110	18	1.1	0.9;
	4757	1	147	56.3	0	0.51	0	1.026011	3.040149	150	8	1.1	0.9;
	4758	2	0	0	0	0	0	1.064718	13.012268	380	3	1.1	0.9;
	4759	1	45.05	19.2	0	0.21	0	0.987109	-20.147328	110	18	1.1	0.9;
	4760	1	65.07	30	0	0.99	0	1.071663	-23.040954	110	15	1.1	0.9;
	4761	1	173.4	31	0	1.91	0	1.021014	-25.555438	220	7	1.1	0.9;
	4762	1	0	0	0	0.65	0	1.024473	-26.6387	150	7	1.1	0.9;
	4763	1	31.94	10.5	0	1.02	0	0.962855	-32.204002	110	21	1.1	0.9;
	4764	1	14.92	5.1	0	0.65	0	0.990229	-29.174476	110	18	1.1	0.9;
	4765	1	33.04	10.1	0	0.43	0	1.039707	-12.376759	220	5	1.1	0.9;
	4766	1	-16.84	-7.96	0	53.85	0	1.039299	44.191195	380	3	1.1	0.9;
	4767	1	-0	-0	0	1.04	0	1.016935	-39.034904	110	12	1.1	0.9;
	4768	1	13.9	1.9	0	3.89	0	1.007739	-11.291231	154	1	1.1	0.9;
	4769	1	-0	-0	0	25.04	0	1.019849	-23.59881	154	1	1.1	0.9;
	4770	1	51.86	19.2	0	1.07	0	1.011545	-36.410984	220	6	1.1	0.9;
	4771	1	0	0	0	1.32	0	0.991474	-36.774818	150	10	1.1	0.9;
	4772	2	0	0	0	0	0	1.023126	-48.430245	220	4	1.1	0.9;
	4773	2	0	0	0	0	0	1.017383	-42.835159	220	4	1.1	0.9;
	4774	1	63.27	28.9	0	2.19	0	1.036095	1.264238	150	8	1.1	0.9;
	4775	1	153.2	31.5	0	12.56	0	1.048902	-22.223071	220	11	1.1	0.9;
	4776	2	0	0	0	0	0	1.063105	26.806955	380	8	1.1	0.9;
	4777	1	97.5	13.9	0	1.24	0	1.032992	-24.051017	154	1	1.1	0.9;
	4778	1	0	-0	0	0.01	0	0.975132	-31.984222	150	9	1.1	0.9;
	4779	1	0	0	0	3.52	0	1.062119	32.50466	380	8	1.1	0.9;
	4780	1	2.5	1.5	0	0.83	0	0.974542	-29.644756	110	18	1.1	0.9;
	4781	1	40.14	24.9	0	5.71	0	0.961409	-31.649013	150	9	1.1	0.9;
	4782	1	21.02	14.8	0	0.1	0	1.000106	-30.520244	150	9	1.1	0.9;
	4783	2	0	0	0	0	0	1.06564	8.268721	220	5	1.1	0.9;
	4784	1	69.78	22.4	0	2.72	0	0.981687	-33.345316	110	21	1.1	0.9;
	4785	1	-0	-0	0	6.26	0	1.0433	-3.069758	220	3	1.1	0.9;
	4786	1	89.9	40.2	0	0.37	0	1.024011	19.961509	380	3	1.1	0.9;
	4787	1	71.98	27.1	0	7.2	0	0.987262	-14.984001	380	5	1.1	0.9;
	4788	2	0	0	0	0	0	1.009861	-8.800983	220	19	1.1	0.9;
	4789	1	146.9	-5.1	0	0	0	1.050877	1.718645	220	3	1.1	0.9;
	4790	1	0	0	0	0	0	0.993761	-32.641501	150	9	1.1	0.9;
	4791	1	9.11	3	0	1.24	0	1.028363	-33.549703	110	12	1.1	0.9;
	4792	1	0	0	0	29.91	0	1.032521	-30.871912	150	9	1.1	0.9;
	4793	1	336.9	81	0	19.53	0	1.009198	-4.731499	220	3	1.1	0.9;
	4794	1	11.41	3.8	0	0.36	0	1.067214	-16.568703	110	18	1.1	0.9;
	4795	1	0	0	0	81.29	0	1.066419	-44.250688	380	4	1.1	0.9;
	4796	1	-9.6	-2.5	0	2.15	0	1.097162	-24.424031	150	9	1.3	0.7;
	4797	1	-0.01	0	0	0	0	1.079402	31.594293	380	8	1.1	0.9;
	4798	1	0	0	0	0	0	1.016409	-30.703143	150	9	1.1	0.9;
	4799	1	0	0	0	2.62	0	1.022774	5.3994	380	1	1.1	0.9;
	4800	1	-0	-0	0	0.34	0	1.03935	-29.675036	150	9	1.1	0.9;
	4801	1	-0	0	0.252818	66.30157	0	1.095999	-25.627197	380	21	1.3	0.7;
	4802	1	20.12	4.9	0	1.61	0	0.999385	-28.927844	110	18	1.1	0.9;
	4803	1	0	-0	0	2.03	0	1.007102	-30.594835	110	21	1.1	0.9;
	4804	1	118.6	43.4	0	0.8	0	1.030914	-21.955421	220	3	1.1	0.9;
	4805	2	0	0	0	0	0	1.046224	5.406275	220	13	1.1	0.9;
	4806	1	210.8	-0.7	0	4.14	0	1.021554	-47.07847	220	4	1.1	0.9;
	4807	1	39.84	6.4	0	34.58	0	1.051133	-42.289639	380	10	1.1	0.9;
	4808	1	7.01	1	0	0.92	0	1.020706	-27.864908	150	7	1.1	0.9;
	4809	1	151	7.1	0	4.24	0	1.006384	-45.998249	220	4	1.1	0.9;
	4810	1	-0	0	0.07	3.65	0	1.060127	-29.656805	220	1	1.1	0.9;
	4811	1	-875.4	-172.2	0	40.13	0	1.04153	-6.009141	380	3	1.1	0.9;
	4812	1	11.2	1.6	0	4.75	0	1.024747	-4.028591	154	1	1.1	0.9;
	4813	1	141.36	53.5	0	2.76	0	0.992803	-50.740529	220	6	1.1	0.9;
	4814	1	41.25	13.5	0	0.39	0	0.982604	-30.201957	110	21	1.1	0.9;
	4815	1	0	-0	0	0.15	0	1.031204	-26.292889	110	21	1.1	0.9;
	4816	2	0	0	0	0	0	1.060527	-24.1453	220	5	1.1	0.9;
	4817	1	270.5	77.8	0	81.25	0	1.060926	-24.219659	380	20	1.1	0.9;
	4818	1	110.2	15.7	0	2.3	0	1.025646	3.281034	154	1	1.1	0.9;
	4819	2	0	0	0	0	0	1.028949	0.926703	380	5	1.1	0.9;
	4820	1	0	-0	0	0.06	0	1.08399	-47.775641	120	6	1.1	0.9;
	4821	1	0	-0	0	0.83	0	0.996011	-31.048839	150	9	1.1	0.9;
	4822	1	78.49	48.6	0	1.05	0	0.977947	-31.778345	150	9	1.1	0.9;
	4823	2	0	0	0	0	0	1.055712	-15.19649	220	5	1.1	0.9;
	4824	2	0	0	0	0	0	1.089599	-14.508804	380	21	1.3	0.7;
	4825	1	-0	0	0	46.7	0	1.009259	8.12329	380	8	1.1	0.9;
	4826	1	103.6	16	0	11.22	0	1.074227	-24.636519	220	5	1.1	0.9;
	4827	1	0	-0	0	3.38	0	1.033461	58.438338	220	3	1.1	0.9;
	4828	1	0	0	0	0	0	1.027792	-29.418146	150	9	1.1	0.9;
	4829	1	171.9	39.1	0	4.23	0	1.015623	-5.852548	220	5	1.1	0.9;
	4830	1	-0	-0	0.127127	25.564419	0	1.029413	-11.080903	380	7	1.1	0.9;
	4831	1	130.4	44.33	0	5.22	0	1.031088	-15.625679	220	5	1.1	0.9;
	4832	1	19.22	6.4	0	31.78	0	1.023151	-43.126712	220	6	1.1	0.9;
	4833	1	5.11	2.5	0	0.57	0	0.92644	-32.984156	110	18	1.1	0.9;
	4834	2	0	0	0	0	0	1.038553	10.065557	154	1	1.1	0.9;
	4835	1	34.14	21.1	0	0	0	1.084853	-36.920174	150	9	1.3	0.7;
	4836	1	-0	-0	0.11	4.95	0	1.050935	-30.001995	380	6	1.1	0.9;
	4837	1	-10.28	-2.68	0	0.9	0	1.018654	-14.598693	110	18	1.1	0.9;
	4838	1	19.82	6.5	0	1.47	0	0.97368	-29.714155	110	18	1.1	0.9;
	4839	1	0.01	0	0	0	0	1.063954	-34.589904	150	9	1.1	0.9;
	4840	1	151.1	66.9	0	0.41	0	1.017478	-16.182331	380	3	1.1	0.9;
	4841	1	25.23	15.6	0	0	0	1.003311	-31.350195	150	9	1.1	0.9;
	4842	1	-5.01	-1.81	0	15.16	0	1.036905	-32.653186	220	11	1.1	0.9;
	4843	2	0	0	0	0	0	1.048394	-42.642098	380	4	1.1	0.9;
	4844	2	0	0	0	0	0	1.00703	33.093639	220	13	1.1	0.9;
	4845	2	0	0	0	0	0	1.024167	61.965245	150	1	1.1	0.9;
	4846	2	0	0	0	0	0	1.030121	12.96686	150	8	1.1	0.9;
	4847	2	0	0	0	0	0	1.081656	12.984521	380	13	1.1	0.9;
	4848	2	0	0	0	0	0	1.070363	-26.152532	220	23	1.1	0.9;
	4849	1	0.2	0.5	0.684311	47.304488	0	1.029581	-7.660653	380	3	1.1	0.9;
	4850	2	0	0	0	0	0	1.057137	-13.928912	220	5	1.1	0.9;
	4851	1	2.3	0.3	0	0.29	0	0.984372	-27.718943	110	18	1.1	0.9;
	4852	1	45.55	-59.9	0	30.04	0	1.037552	-20.631579	220	5	1.1	0.9;
	4853	1	-0	-0	0	12.98	0	1.011478	-31.784552	150	9	1.1	0.9;
	4854	1	-0	-0	0	2.47	0	0.98255	6.741383	220	3	1.1	0.9;
	4855	1	127.9	-34	0	19.86	0	1.011289	-60.1788	220	4	1.1	0.9;
	4856	1	45.95	7.5	0	8.39	0	1.021361	-45.516743	220	4	1.1	0.9;
	4857	1	78.69	19.6	0	0.87	0	1.049037	-27.233229	110	3	1.1	0.9;
	4858	1	-0	-0	0	53.97	0	1.064085	29.968955	380	8	1.1	0.9;
	4859	1	261.2	105.6	0	28.53	0	1.06403	55.017036	380	3	1.1	0.9;
	4860	1	0	-0	0	29.32	0	1.014639	-3.738268	380	3	1.1	0.9;
	4861	1	0	-0	0.09	7.22	0	1.032468	-26.038409	220	21	1.1	0.9;
	4862	1	-5.88	-3.76	0	0.39	0	1.001217	-24.471022	110	18	1.1	0.9;
	4863	1	-0	0	0	0	0	1.022069	10.989113	150	8	1.1	0.9;
	4864	1	-28.21	-6	0	0.05	0	1.044886	-13.237652	220	5	1.1	0.9;
	4865	1	0	0	0	0.51	0	1.019121	-48.988888	380	6	1.1	0.9;
	4866	1	169	56.4	0	8.06	0	1.027719	20.158525	380	3	1.1	0.9;
	4867	1	65.77	27.3	0	0.49	0	1.027117	-3.83085	220	5	1.1	0.9;
	4868	1	2.2	0.5	0	2.88	0	0.994275	-48.159668	150	10	1.1	0.9;
	4869	1	-0	0	0	8.1	0	1.026579	-13.337718	220	6	1.1	0.9;
	4870	1	65.5	9.3	0	5.31	0	0.997277	4.262502	154	1	1.1	0.9;
	4871	2	0	0	0	0	0	1.054823	-21.307193	110	3	1.1	0.9;
	4872	1	15.42	3.9	0	0.79	0	1.020922	7.053602	110	8	1.1	0.9;
	4873	1	262.6	31.8	0	8.13	0	0.996169	-48.406794	220	6	1.1	0.9;
	4874	1	-0.58	-53.03	0	6.85	0	1.075415	-39.326121	220	5	1.1	0.9;
	4875	1	-0	-0	0	2.74	0	0.996138	9.405021	220	3	1.1	0.9;
	4876	1	230.9	42.6	0	12.62	0	1.022928	-23.903668	220	7	1.1	0.9;
	4877	1	0.9	0.3	0	1.48	0	1.048444	-24.999745	110	19	1.1	0.9;
	4878	1	5.11	1.7	0	0.48	0	1.028678	-34.027355	110	12	1.1	0.9;
	4879	1	0	-0	0	3.48	0	1.026685	-0.863653	380	3	1.1	0.9;
	4880	2	0	0	0	0	0	1.041763	-15.664109	220	5	1.1	0.9;
	4881	1	5.11	0	0	4.97	0	0.954039	-31.055989	110	18	1.1	0.9;
	4882	2	0	0	0	0	0	1.028068	17.747314	154	1	1.1	0.9;
	4883	1	0	-0	0	1.26	0	1.00371	-32.000976	150	9	1.1	0.9;
	4884	1	0	0	0	9.77	0	1.019233	-0.73514	380	3	1.1	0.9;
	4885	1	166.4	50.4	0	1.23	0	1.06846	-26.747409	220	5	1.1	0.9;
	4886	1	-0	0	0	3.08	0	1.071272	60.019932	380	3	1.1	0.9;
	4887	1	53.56	33.2	0	3.77	0	0.982898	-28.745965	150	9	1.1	0.9;
	4888	1	186.9	26.6	0	11.96	0	1.000821	-18.823128	400	1	1.1	0.9;
	4889	1	74.98	19.8	0	6.47	0	1.028982	-12.570692	220	5	1.1	0.9;
	4890	1	185	43.6	0	6.3	0	1.019498	-48.757938	220	4	1.1	0.9;
	4891	1	25.1	3.6	0	1.74	0	1.01492	-13.803384	154	1	1.1	0.9;
	4892	2	0	0	0	0	0	1.033418	-2.864989	220	1	1.1	0.9;
	4893	1	-0	-0	0	39.21	0	1.039906	4.586896	380	1	1.1	0.9;
	4894	1	21.62	7.8	0	0.68	0	1.102158	-15.536498	110	19	1.3	0.7;
	4895	1	-7.61	2.5	0	0.5	0	1.040119	-21.348536	150	9	1.1	0.9;
	4896	1	329.7	32.4	0	4.5	0	1.009835	-58.953855	220	4	1.1	0.9;
	4897	2	0	0	0	0	0	1.014069	-46.541884	220	4	1.1	0.9;
	4898	1	55.7	7.9	0	0.1	0	1.031771	-11.979093	154	1	1.1	0.9;
	4899	1	0	-0	0.12	12.76	0	1.024222	-27.975557	220	21	1.1	0.9;
	4900	2	0	0	0	0	0	1.023832	14.693276	154	1	1.1	0.9;
	4901	1	-0	0	0	0.88	0	1.004799	-26.239776	110	3	1.1	0.9;
	4902	1	1.9	-0	0	0.5	0	0.994314	-31.942978	110	18	1.1	0.9;
	4903	1	-0	-0	0	220.46	0	1.089802	-13.333108	380	18	1.3	0.7;
	4904	1	96	33.3	0	16.89	0	1.006792	-37.073353	220	6	1.1	0.9;
	4905	1	17.12	5.6	0	1.12	0	1.024715	-34.268369	110	12	1.1	0.9;
	4906	1	175.8	70.7	0	19.08	0	1.023576	-46.41154	220	4	1.1	0.9;
	4907	1	125.16	29.2	0	2.69	0	1.037484	-13.630352	220	5	1.1	0.9;
	4908	1	50.06	9.9	0	11.56	0	1.034526	-21.581828	220	5	1.1	0.9;
	4909	1	220.6	83.4	0	17.73	0	1.03446	-21.776933	380	20	1.1	0.9;
	4910	1	-0	0	0	0.95	0	1.021443	-47.156402	220	4	1.1	0.9;
	4911	1	-0	-0	0	1.09	0	1.047684	-27.412491	150	9	1.1	0.9;
	4912	1	-0	0	0	0.45	0	1.036375	-21.486687	220	20	1.1	0.9;
	4913	1	-24.03	7.9	0	0.25	0	1.018213	-26.024805	150	9	1.1	0.9;
	4914	1	14.91	-8.73	0	34.66	0	1.032689	-23.332011	220	5	1.1	0.9;
	4915	1	0	0	0	0.31	0	0.998633	-32.348994	150	9	1.1	0.9;
	4916	1	0	-0	0	0.94	0	1.021276	-14.380448	110	18	1.1	0.9;
	4917	1	71.6	10.2	0	58.47	0	1.001322	-13.363196	400	1	1.1	0.9;
	4918	2	0	0	0	0	0	1.053068	-11.355832	220	5	1.1	0.9;
	4919	1	15.82	9.8	0	7.61	0	1.048486	-29.24882	150	9	1.1	0.9;
	4920	1	-0	0	0	145.62	0	1.003171	-1.030666	400	1	1.1	0.9;
	4921	1	0	-0	0	18.54	0	1.022562	6.896935	380	3	1.1	0.9;
	4922	1	0	-0	0	2.57	0	1.047775	-16.290741	380	1	1.1	0.9;
	4923	1	0	0	0	26	0	1.032381	-17.722009	380	1	1.1	0.9;
	4924	1	71.76	11.37	0	0	0	1.061023	-19.247849	110	24	1.1	0.9;
	4925	1	-0	0	0	0.3	0	1.032032	-22.830457	220	5	1.1	0.9;
	4926	1	3.8	1.1	0	10.69	0	1.051879	-42.309199	380	10	1.1	0.9;
	4927	1	46.2	6.6	0	1.24	0	0.99652	-22.778897	154	1	1.1	0.9;
	4928	1	40.04	10	0	23.59	0	1.021927	-46.925549	380	6	1.1	0.9;
	4929	1	37	8.7	0.070441	50.913893	0	1.018028	4.736989	380	2	1.1	0.9;
	4930	1	2.5	1.3	0	0.16	0	0.997633	-26.018623	110	18	1.1	0.9;
	4931	2	0	0	0	0	0	1.035713	-43.642893	380	4	1.1	0.9;
	4932	1	0	0	0	0	0	0.982693	-27.895484	150	9	1.1	0.9;
	4933	1	-19.22	6.3	0	1.02	0	0.977104	-29.903365	150	9	1.1	0.9;
	4934	1	22.42	7.3	0	0.77	0	1.013716	-38.904971	110	12	1.1	0.9;
	4935	1	-0	0	0	15.76	0	1.041012	-31.424919	220	12	1.1	0.9;
	4936	1	312.2	28	0	48.6	0	1.031863	-31.443839	220	5	1.1	0.9;
	4937	1	11.61	2.5	0	0.42	0	1.064197	-31.921414	110	3	1.1	0.9;
	4938	2	0	0	0	0	0	1.031692	-7.869883	380	7	1.1	0.9;
	4939	1	56.76	31.1	0	0.37	0	1.041972	-4.812959	220	5	1.1	0.9;
	4940	1	0	0	0	5.5	0	0.917613	-35.788282	150	9	1.1	0.9;
	4941	1	59.67	25.6	0	11.54	0	1.002344	8.412161	220	3	1.1	0.9;
	4942	1	109.9	33.6	0	3.72	0	1.026049	-13.088391	220	5	1.1	0.9;
	4943	1	0	-0	0	0.83	0	1.04576	-16.022015	380	6	1.1	0.9;
	4944	2	0	0	0	0	0	1.0154	-22.809584	110	19	1.1	0.9;
	4945	2	0	0	0	0	0	1.055797	-41.194507	380	4	1.1	0.9;
	4946	1	329.7	220.4	0	46.92	0	1.013935	-13.593736	380	3	1.1	0.9;
	4947	1	88.3	2.8	0	5.21	0	1.015918	-39.946101	220	4	1.1	0.9;
	4948	1	10.21	-3.2	0	5.15	0	1.030299	-40.316404	220	4	1.1	0.9;
	4949	2	0	0	0	0	0	1.046901	2.83582	110	18	1.1	0.9;
	4950	1	218.7	-9.1	0	49	0	1.030927	-40.498397	220	5	1.1	0.9;
	4951	1	71.98	34.1	0	7.65	0	1.041769	-1.912149	220	5	1.1	0.9;
	4952	2	0	0	0	0	0	1.047958	-17.520372	220	5	1.1	0.9;
	4953	1	66.77	11.1	0	2.7	0	1.014288	-51.969966	220	4	1.1	0.9;
	4954	2	0	0	0	0	0	1.016721	-43.596314	220	4	1.1	0.9;
	4955	2	0	0	0	0	0	1.037342	-7.422474	154	1	1.1	0.9;
	4956	1	0	-0	0	0.99	0	1.042247	-3.198061	220	1	1.1	0.9;
	4957	2	0	0	0	0	0	1.045811	-39.083866	380	4	1.1	0.9;
	4958	1	-0	-0	0	1.94	0	1.027072	-45.983946	220	4	1.1	0.9;
	4959	2	0	0	0	0	0	1.051773	-18.155087	400	1	1.1	0.9;
	4960	1	12.91	7.1	0	0.84	0	0.958246	-30.72556	110	18	1.1	0.9;
	4961	1	62.97	68	0	13.59	0	0.953377	-18.616786	220	16	1.1	0.9;
	4962	1	29.13	18	0	0.06	0	0.983081	-31.410789	150	9	1.1	0.9;
	4963	1	-0	-0	0	0.29	0	0.989807	-30.82242	150	9	1.1	0.9;
	4964	1	-0	-0	0	29.32	0	1.026578	11.490083	380	8	1.1	0.9;
	4965	1	0	0	0	0	0	1.020634	-29.934806	150	9	1.1	0.9;
	4966	1	110.7	78.7	0	5.27	0	1.008477	-41.810075	220	17	1.1	0.9;
	4967	2	0	0	0	0	0	1.020942	-42.545101	220	4	1.1	0.9;
	4968	1	-0	-0	0	16.57	0	1.049681	11.588745	380	1	1.1	0.9;
	4969	2	0	0	0	0	0	1.065019	-9.014123	380	14	1.1	0.9;
	4970	1	-0	-0	0	171.931977	0	1.058166	-17.965332	380	5	1.1	0.9;
	4971	1	130	128	0	7.73	0	1.014937	-14.841613	220	3	1.1	0.9;
	4972	1	13.9	1.9	0	4.71	0	1.019416	-7.659192	154	1	1.1	0.9;
	4973	1	0	0	0	0	0	1.029241	-22.185994	150	9	1.1	0.9;
	4974	1	89.08	26.9	0	4.68	0	1.020099	-13.73201	220	5	1.1	0.9;
	4975	1	0	0	0	0	0	0.893435	-35.853165	150	9	1.3	0.7;
	4976	1	23.43	7.5	0	0.1	0	0.984261	-28.464149	110	18	1.1	0.9;
	4977	1	-0	-0	0.181549	24.230889	0	1.059145	-29.297782	220	18	1.1	0.9;
	4978	1	0	0	0	0.31	0	0.963067	-31.477381	150	9	1.1	0.9;
	4979	2	0	0	0	0	0	1.039397	-19.104141	380	6	1.1	0.9;
	4980	1	51.8	7.4	0	3.41	0	1.013058	-8.277201	154	1	1.1	0.9;
	4981	1	-0	0	0	1.2	0	1.00481	-32.06595	150	9	1.1	0.9;
	4982	1	4.81	1.6	0	0.53	0	1.004947	-41.059674	110	12	1.1	0.9;
	4983	1	0	-0	0	16.87	0	1.007916	-25.306148	154	1	1.1	0.9;
	4984	1	10.11	1.6	0	0.66	0	1.000242	-28.038181	110	18	1.1	0.9;
	4985	1	32.5	30.5	0	265.721081	0	1.110771	-22.731958	380	9	1.3	0.7;
	4986	1	-22.8	-16.4	0	0.19	0	1.0658	-25.302839	110	21	1.1	0.9;
	4987	1	-4.66	1.11	0	0.19	0	1.017878	-26.029213	110	18	1.1	0.9;
	4988	1	11.41	3.8	0	1.01	0	1.018071	-20.102983	110	18	1.1	0.9;
	4989	1	-4.89	-2.4	0	0.25	0	0.997604	-24.708005	110	18	1.1	0.9;
	4990	1	0	-0	0.090774	3.730444	0	1.040475	-26.159191	220	18	1.1	0.9;
	4991	1	501.1	10.7	0	18.07	0	1.023645	-46.005472	380	6	1.1	0.9;
	4992	1	58.36	42.7	0	1.29	0	0.95548	-17.778421	110	18	1.1	0.9;
	4993	1	-0.12	-1.74	0	0	0	1.08106	31.726559	380	8	1.1	0.9;
	4994	2	0	0	0	0	0	1.047685	-28.158838	380	6	1.1	0.9;
	4995	1	190.7	13.3	0	4.61	0	0.987735	-50.740816	220	6	1.1	0.9;
	4996	2	0	0	0	0	0	1.040488	-32.985607	110	12	1.1	0.9;
	4997	1	0	0	0	0.06	0	1.066858	-26.629056	220	21	1.1	0.9;
	4998	1	-0	-0	0	1.433772	0	1.000934	-24.434419	220	20	1.1	0.9;
	4999	1	73.7	10.5	0	4.55	0	1.004825	-11.522006	154	1	1.1	0.9;
	5000	1	-0	-0	0	5.98	0	1.066857	-26.629052	220	21	1.1	0.9;
	5001	1	82.49	24.5	0	0.62	0	1.064885	14.706305	110	8	1.1	0.9;
	5002	1	-0	0	0	4.34	0	1.072288	-7.351176	220	5	1.1	0.9;
	5003	1	17.87	-8.59	0	0.21	0	1.03844	-9.926369	220	5	1.1	0.9;
	5004	2	0	0	0	0	0	1.054301	-15.923141	220	5	1.1	0.9;
	5005	2	0	0	0	0	0	1.062291	54.461313	380	3	1.1	0.9;
	5006	1	-58.77	18.7	0	4.23	0	0.991224	-37.430456	150	10	1.1	0.9;
	5007	1	0	0	0	40.291745	0	1.020758	-15.351482	380	5	1.1	0.9;
	5008	1	0	-0	0	23.76	0	1.017261	11.508675	380	8	1.1	0.9;
	5009	1	0	-0	0.27075	87.43435	0	1.079182	-14.241458	380	18	1.1	0.9;
	5010	1	11.2	1.6	0	3.74	0	1.031193	4.637375	154	1	1.1	0.9;
	5011	1	0	-0	0	11.27	0	1.052752	12.478077	380	3	1.1	0.9;
	5012	1	5.11	2.5	0	0.52	0	0.959202	-29.296099	110	18	1.1	0.9;
	5013	1	-0	-0	0	1.15	0	0.960889	-31.648573	150	9	1.1	0.9;
	5014	1	70.08	35	0	1.45	0	1.025949	-13.350064	220	6	1.1	0.9;
	5015	1	27.13	9.8	0	0.17	0	1.051844	-19.601795	110	19	1.1	0.9;
	5016	1	-0	0	0	8.347383	0	1.019196	-5.050368	380	5	1.1	0.9;
	5017	1	0.1	-9.6	0	1.3	0	1.045354	-20.92555	220	3	1.1	0.9;
	5018	1	9.51	2.8	0	0.5	0	0.98564	-26.685436	110	18	1.1	0.9;
	5019	2	0	0	0	0	0	1.05406	-13.335274	220	5	1.1	0.9;
	5020	2	0	0	0	0	0	1.067392	-27.874565	220	1	1.1	0.9;
	5021	1	0	-0	0	1.98	0	1.009321	51.728828	220	3	1.1	0.9;
	5022	1	38.64	23.9	0	0.94	0	0.990784	-30.586521	150	9	1.1	0.9;
	5023	1	-0	0	0	2.81	0	1.028198	18.916671	154	1	1.1	0.9;
	5024	1	9.51	3.1	0	0.52	0	1.054697	-20.712211	110	19	1.1	0.9;
	5025	2	0	0	0	0	0	1.002981	-48.33968	220	4	1.1	0.9;
	5026	1	-0	-0	0	20.11	0	1.056144	-6.90172	380	1	1.1	0.9;
	5027	2	0	0	0	0	0	1.046647	-18.582863	110	19	1.1	0.9;
	5028	1	38.1	5.5	0	3.32	0	1.000587	-8.829005	154	1	1.1	0.9;
	5029	1	-12.92	-1.53	0	0.96	0	1.020712	-15.292549	110	18	1.1	0.9;
	5030	1	30.33	18.8	0	1.35	0	0.962887	-32.243935	150	9	1.1	0.9;
	5031	1	174.7	35	0	2.14	0	1.020947	7.905729	150	8	1.1	0.9;
	5032	1	14.32	4.4	0	0.74	0	1.012239	-35.608603	110	22	1.1	0.9;
	5033	1	18.9	2.7	0	2.81	0	1.028024	-11.574594	154	1	1.1	0.9;
	5034	1	32.24	4.8	0	0.85	0	1.02172	-52.053756	220	4	1.1	0.9;
	5035	1	-0	-0	0	13.25	0	1.031718	-42.193252	380	1	1.1	0.9;
	5036	2	0	0	0	0	0	1.060776	-22.800944	110	15	1.1	0.9;
	5037	1	10.31	4.3	0	0.67	0	0.95097	-30.827123	110	18	1.1	0.9;
	5038	1	-0	0	0	2.02	0	1.019784	-11.434184	380	1	1.1	0.9;
	5039	1	20.32	10.1	0	0.14	0	0.950642	-30.558087	110	18	1.1	0.9;
	5040	1	55.7	7.9	0	11.74	0	1.028521	-7.657229	154	1	1.1	0.9;
	5041	1	0	8.8	0	18.87	0	1.060344	-30.926836	150	9	1.1	0.9;
	5042	1	24.33	10	0	84.57	0	1.037823	-50.188852	380	4	1.1	0.9;
	5043	2	0	0	0	0	0	1.07576	-19.742609	220	21	1.1	0.9;
	5044	1	0	-0	0	12.01	0	1.056908	-40.620016	380	1	1.1	0.9;
	5045	1	11.41	2.5	0	0.22	0	0.985984	-20.249217	110	18	1.1	0.9;
	5046	1	-0	-0	0.07356	1.029149	0	1.012038	65.139289	220	1	1.1	0.9;
	5047	1	4.61	1.3	0	0.53	0	0.990151	-29.55751	110	18	1.1	0.9;
	5048	1	-0	0	0	9.59	0	1.064535	53.239353	380	3	1.1	0.9;
	5049	1	136.6	29	0	2.05	0	1.014828	-20.777774	220	5	1.1	0.9;
	5050	1	0	0	0.181549	18.620889	0	1.047816	-21.271323	220	18	1.1	0.9;
	5051	2	0	0	0	0	0	1.042576	-16.820552	220	5	1.1	0.9;
	5052	1	-0	-0	0	5.173673	0	0.972477	-23.585502	220	24	1.1	0.9;
	5053	1	43.45	26.9	0	6.37	0	1.04204	-29.563033	150	9	1.1	0.9;
	5054	1	55.16	19.8	0	135.46	0	1.074611	37.120818	380	3	1.1	0.9;
	5055	1	0	0	0.263214	3.036617	0	1.024425	-21.542982	220	19	1.1	0.9;
	5056	1	62.7	9	0	4.93	0	1.012688	-8.331742	154	1	1.1	0.9;
	5057	1	0	0	0	0.06	0	1.01624	-29.209206	110	21	1.1	0.9;
	5058	1	9.51	0.5	0	0.04	0	1.018173	-32.163346	150	9	1.1	0.9;
	5059	2	0	0	0	0	0	1.031064	-30.99284	380	6	1.1	0.9;
	5060	2	0	0	0	0	0	1.06427	-41.076251	380	4	1.1	0.9;
	5061	1	205.8	66.8	0	3.65	0	1.051369	9.332018	380	3	1.1	0.9;
	5062	1	160.6	51.2	0	46.39	0	1.059351	-25.889195	380	20	1.1	0.9;
	5063	1	0	0	0	0	0	0.967209	-32.44223	150	9	1.1	0.9;
	5064	1	22.32	7.3	0	0.46	0	1.005521	-36.907508	110	12	1.1	0.9;
	5065	2	0	0	0	0	0	1.06654	28.353276	380	8	1.1	0.9;
	5066	1	35.64	22	0	4.46	0	0.915156	-35.96451	150	9	1.1	0.9;
	5067	2	0	0	0	0	0	1.016664	-8.184508	380	5	1.1	0.9;
	5068	1	3.8	1.3	0	0.78	0	0.99844	-26.718859	110	18	1.1	0.9;
	5069	1	-19.07	-6.06	0	16.68	0	1.039073	-32.600601	220	11	1.1	0.9;
	5070	1	50.06	25	0	1.21	0	1.005401	-36.46178	110	12	1.1	0.9;
	5071	1	0	0	0	0.85	0	1.007605	-13.940011	110	18	1.1	0.9;
	5072	1	0	0	0	22.93	0	1.085787	-36.678116	150	9	1.3	0.7;
	5073	1	49.8	7.1	0	1.4	0	1.005567	-6.023147	154	1	1.1	0.9;
	5074	1	5.11	2.5	0	1.5	0	0.930518	-32.610216	110	18	1.1	0.9;
	5075	2	0	0	0	0	0	1.027964	-8.633841	220	13	1.1	0.9;
	5076	1	19.12	2.8	0	0.97	0	0.970692	-29.74789	110	18	1.1	0.9;
	5077	1	30.93	5.8	0	4.45	0	1.047713	-27.005552	220	5	1.1	0.9;
	5078	1	92.4	6.3	0	4.32	0	1.026431	-32.887984	220	6	1.1	0.9;
	5079	1	-0	0	0	49.26	0	1.007664	8.988419	400	1	1.1	0.9;
	5080	1	0	-0	0	0.04	0	1.022004	-32.45595	150	9	1.1	0.9;
	5081	1	3.7	1.3	0	0.98	0	0.948504	-22.572543	110	19	1.1	0.9;
	5082	1	43.15	11.1	0	0.4	0	1.032319	-26.126865	110	19	1.1	0.9;
	5083	1	80.39	10.8	0	0.35	0	1.033581	-12.500536	220	5	1.1	0.9;
	5084	1	0	-0	0	494.41	0	1.019024	-16.104265	750	1	1.1	0.9;
	5085	1	6.61	3.3	0	2.67	0	1.022169	-39.246409	110	22	1.1	0.9;
	5086	1	7.41	2.7	0	1.08	0	1.046587	-17.262933	110	19	1.1	0.9;
	5087	1	-14.81	-0.52	0	0.13	0	1.053842	-14.978871	110	18	1.1	0.9;
	5088	1	87.38	-90.15	0	16.05	0	1.035486	-22.562506	220	3	1.1	0.9;
	5089	1	155.25	23.3	0	3.74	0	0.996716	-35.041034	220	6	1.1	0.9;
	5090	1	0	-0	0	4.74	0	1.026047	2.440706	154	1	1.1	0.9;
	5091	2	0	0	0	0	0	1.051243	-39.461817	380	4	1.1	0.9;
	5092	2	0	0	0	0	0	1.061075	-9.584041	380	16	1.1	0.9;
	5093	1	70.78	28	0	1.49	0	1.004057	-15.185299	220	5	1.1	0.9;
	5094	1	-0	-0	0	5.02	0	1.026644	-34.880598	220	10	1.1	0.9;
	5095	1	-0	0	0	0.22	0	0.982011	-28.162982	110	18	1.1	0.9;
	5096	1	-2.27	-1.81	0	42.76	0	1.043485	-39.194179	380	4	1.1	0.9;
	5097	2	0	0	0	0	0	1.052774	-9.767291	150	2	1.1	0.9;
	5098	2	0	0	0	0	0	1.006462	8.730842	400	1	1.1	0.9;
	5099	1	0	0	0	9.34	0	1.057548	1.412506	220	5	1.1	0.9;
	5100	1	117.7	115	0	12.29	0	1.040413	-8.41358	380	3	1.1	0.9;
	5101	1	-0	0	0.093329	56.622656	0	1.000338	65.472066	380	1	1.1	0.9;
	5102	1	99.4	-347.1	0	31.86	0	1.043535	33.165934	380	3	1.1	0.9;
	5103	1	10.81	1.9	0	1.6	0	1.000647	-15.004443	110	18	1.1	0.9;
	5104	1	-0	0	0	9.06	0	1.038383	-20.744706	220	3	1.1	0.9;
	5105	1	37	5.3	0	1.74	0	1.013039	-7.922909	154	1	1.1	0.9;
	5106	1	39.94	13	0	2.47	0	1.069889	-31.663003	220	5	1.1	0.9;
	5107	1	122.9	19.3	0	68.84	0	1.046562	-18.547393	380	16	1.1	0.9;
	5108	1	-55.83	-31.99	0	2.88	0	1.037984	14.73529	110	8	1.1	0.9;
	5109	1	213.1	-12.3	0	30.87	0	1.045103	-27.401915	380	6	1.1	0.9;
	5110	2	0	0	0	0	0	1.033935	-10.289662	220	5	1.1	0.9;
	5111	1	139.9	1.1	0	5.58	0	1.012688	-52.793246	220	4	1.1	0.9;
	5112	1	41.35	25.6	0	9.9	0	1.020362	-29.475309	150	9	1.1	0.9;
	5113	1	-0	-0	0	11.9	0	1.039086	-23.556866	154	1	1.1	0.9;
	5114	1	-0	0	0	3.41	0	1.023245	-15.041571	220	6	1.1	0.9;
	5115	1	2.47	-0.62	0	2.54	0	1.022088	64.205077	150	1	1.1	0.9;
	5116	1	124.9	77	0	23.05	0	1.035964	7.244022	380	3	1.1	0.9;
	5117	1	46.25	13.1	0	1.29	0	1.019223	-47.112917	220	4	1.1	0.9;
	5118	1	111.4	15.9	0	19.05	0	1.009127	-7.7018	154	1	1.1	0.9;
	5119	1	1.4	0.5	0	0.35	0	1.0364	-24.671527	110	19	1.1	0.9;
	5120	2	0	0	0	0	0	1.06465	-24.207883	220	5	1.1	0.9;
	5121	1	23.4	8.38	0	1.07	0	1.014598	-44.269584	220	4	1.1	0.9;
	5122	1	163.1	23.2	0	3.36	0	0.998421	-26.410038	154	1	1.1	0.9;
	5123	1	-15.91	-6.11	0	0.21	0	1.002649	-31.677218	110	23	1.1	0.9;
	5124	1	-0	-0	0	0.66	0	1.028994	-31.533375	380	6	1.1	0.9;
	5125	1	0	0	0	16.06	0	1.04785	-37.9788	380	4	1.1	0.9;
	5126	1	17.82	7.3	0	0.14	0	0.992635	-27.392721	110	18	1.1	0.9;
	5127	1	66.47	22.1	0	0.46	0	1.036246	-30.577267	110	21	1.1	0.9;
	5128	1	20.9	3	0	3.48	0	1.016749	-10.162057	154	1	1.1	0.9;
	5129	1	102.3	14.6	0	1.1	0	1.025499	-25.984106	154	1	1.1	0.9;
	5130	1	572.3	159.7	0	19.88	0	1.026774	-33.39654	380	6	1.1	0.9;
	5131	1	-0	0	0	7.84	0	1.068914	-26.734404	220	5	1.1	0.9;
	5132	1	12.31	3	0	0.98	0	0.96521	-29.830591	110	18	1.1	0.9;
	5133	1	0	0	0	0.16	0	1.055766	-27.405972	150	9	1.1	0.9;
	5134	2	0	0	0	0	0	1.022715	-8.773783	154	1	1.1	0.9;
	5135	1	0	0	0	0.93	0	0.977188	-26.554615	110	18	1.1	0.9;
	5136	1	52.96	32.8	0	5.79	0	1.000985	-36.77408	150	9	1.1	0.9;
	5137	1	16.92	0	0	0.57	0	1.022383	-16.545048	220	5	1.1	0.9;
	5138	2	0	0	0	0	0	1.048344	-25.74707	380	6	1.1	0.9;
	5139	1	24.63	2.1	0	1.26	0	1.034744	3.444052	220	13	1.1	0.9;
	5140	1	26.35	30	0	1.82	0	1.016117	-31.378283	150	9	1.1	0.9;
	5141	1	18.92	6.6	0	1.4	0	0.989027	-29.757084	110	18	1.1	0.9;
	5142	1	0	0	0	15	0	1.033675	55.712656	220	3	1.1	0.9;
	5143	1	0	-0	0	1.66	0	1.015791	-31.456773	150	9	1.1	0.9;
	5144	2	0	0	0	0	0	1.032838	-3.148546	380	5	1.1	0.9;
	5145	2	0	0	0	0	0	1.049287	-20.894198	220	3	1.1	0.9;
	5146	1	229.9	48.9	0	8.36	0	1.006633	-34.449474	220	5	1.1	0.9;
	5147	1	-0	0	0	17.81	0	1.023848	3.383433	380	1	1.1	0.9;
	5148	1	139.77	41.14	0	15.57	0	1.023799	-8.049269	220	13	1.1	0.9;
	5149	1	0.2	1.6	0	0.23	0	0.989063	12.894504	220	3	1.1	0.9;
	5150	1	-0	0	0	1.33	0	0.994149	-30.584157	150	9	1.1	0.9;
	5151	1	-0	-0	0	0.32	0	1.040193	-33.295872	150	9	1.1	0.9;
	5152	2	0	0	0	0	0	1.033193	24.896513	220	3	1.1	0.9;
	5153	1	18.52	6.1	0	1.24	0	0.950584	-35.141712	110	21	1.1	0.9;
	5154	2	0	0	0	0	0	1.021968	67.005317	150	1	1.1	0.9;
	5155	1	411.3	168	0	0	0	1.025638	-11.295879	150	2	1.1	0.9;
	5156	1	0	0	0	0	0	1.012996	-27.195344	150	9	1.1	0.9;
	5157	1	-0	-0	0	0.23	0	1.01048	-24.157713	110	18	1.1	0.9;
	5158	2	0	0	0	0	0	1.043194	10.693138	220	13	1.1	0.9;
	5159	1	21.12	7.7	0	0.52	0	1.03892	-8.307439	110	19	1.1	0.9;
	5160	1	-0	-0	0.094757	-0.276956	0	1.041089	-12.912267	220	19	1.1	0.9;
	5161	1	0	-0	0	6.89	0	1.019958	-36.560167	220	22	1.1	0.9;
	5162	1	55.26	0	0	1.05	0	1.009888	-48.531709	220	4	1.1	0.9;
	5163	1	-11.81	-12	0	19.39	0	1.034355	-39.091202	220	4	1.1	0.9;
	5164	2	0	0	0	0	0	1.083362	36.17773	380	3	1.1	0.9;
	5165	1	-0	-0	0	35.06	0	1.024106	7.542973	380	8	1.1	0.9;
	5166	1	0	-0	0	0.48	0	0.935772	-34.646986	150	9	1.1	0.9;
	5167	1	12.01	0.7	0	0.5	0	1.027953	50.072739	220	3	1.1	0.9;
	5168	1	-0	0	0	0.45	0	0.999285	-35.774577	110	12	1.1	0.9;
	5169	1	0	0	0	30.28	0	1.070559	12.625799	380	3	1.1	0.9;
	5170	1	0	0	0	19.57	0	1.026589	-53.761687	220	4	1.1	0.9;
	5171	2	0	0	0	0	0	1.011094	-47.540213	220	4	1.1	0.9;
	5172	1	3	1.7	0	0.53	0	1.0704	-22.127149	110	23	1.1	0.9;
	5173	1	-0	-0	0	1.41	0	1.033809	-34.0408	110	12	1.1	0.9;
	5174	1	0	-0	0	8.19	0	1.041954	-21.336015	220	5	1.1	0.9;
	5175	1	-0	-0	0.087808	3.044992	0	1.031881	-15.733969	220	18	1.1	0.9;
	5176	1	1.9	0.7	0	0.19	0	1.021206	-28.411576	110	19	1.1	0.9;
	5177	1	0	0	0.44806	95.507515	0	1.042099	-18.696816	380	11	1.1	0.9;
	5178	1	90.2	5.8	0	18.69	0	1.04332	35.587159	380	3	1.1	0.9;
	5179	1	179.8	5.67	0	12.83	0	1.032567	-31.167837	220	5	1.1	0.9;
	5180	2	0	0	0	0	0	1.010906	5.415517	154	1	1.1	0.9;
	5181	1	45.05	15.3	0	12.29	0	1.006184	-36.954734	220	6	1.1	0.9;
	5182	1	0	0	0	0.18	0	1.041325	-8.044042	220	5	1.1	0.9;
	5183	1	0	0	0	2.18	0	1.025241	-25.129178	220	1	1.1	0.9;
	5184	1	0	0.2	0	0.12	0	1.020455	-5.269919	220	3	1.1	0.9;
	5185	1	0	0	0.090774	3.340444	0	1.046406	-15.155092	220	18	1.1	0.9;
	5186	1	36.34	22.5	0	2.39	0	0.992915	-27.541743	150	9	1.1	0.9;
	5187	1	10.8	1.6	0	2.88	0	1.007766	-5.873752	154	1	1.1	0.9;
	5188	1	107.9	17.6	0	1.03	0	1.018558	0.328446	150	8	1.1	0.9;
	5189	1	24.33	5.8	0	13.48	0	1.020802	-47.248139	220	4	1.1	0.9;
	5190	1	0	0	0	0.53	0	1.040185	-30.045821	110	21	1.1	0.9;
	5191	1	67.98	9.2	0	7.77	0	1.029465	-50.177641	220	4	1.1	0.9;
	5192	2	0	0	0	0	0	1.039762	-15.033657	220	6	1.1	0.9;
	5193	2	0	0	0	0	0	1.022466	-14.154389	110	18	1.1	0.9;
	5194	1	102.7	39	0	135.53	0	1.074026	36.916958	380	3	1.1	0.9;
	5195	1	33.24	11.1	0	1.14	0	0.984634	-35.249395	110	21	1.1	0.9;
	5196	1	7.01	2.3	0	0.81	0	1.045019	-30.553817	110	12	1.1	0.9;
	5197	1	35.14	12.6	0	0.97	0	1.055397	-22.593679	110	15	1.1	0.9;
	5198	1	7.41	2.7	0	0.8	0	1.040031	-20.975515	110	19	1.1	0.9;
	5199	1	76.08	17.1	0	5.78	0	1.003128	-23.967089	220	20	1.1	0.9;
	5200	1	214.8	30.6	0	3.65	0	1.024102	-23.05819	154	1	1.1	0.9;
	5201	1	35.94	26	0	8.07	0	1.061466	-26.766863	220	1	1.1	0.9;
	5202	1	85.09	13.4	0	0.23	0	1.02566	-53.811747	220	4	1.1	0.9;
	5203	1	221.9	86.3	0	2.18	0	1.02662	5.169172	150	8	1.1	0.9;
	5204	1	18.8	2.7	0	0.31	0	1.011682	-5.893879	154	1	1.1	0.9;
	5205	1	-0	-0	0	2.84	0	1.018645	-46.209653	220	1	1.1	0.9;
	5206	1	-0	-0	0	4.99	0	1.019532	-19.41052	220	1	1.1	0.9;
	5207	1	11.71	3.9	0	0.77	0	1.002522	-39.681438	110	12	1.1	0.9;
	5208	1	41.65	13.6	0	0.31	0	0.98578	-31.882609	110	21	1.1	0.9;
	5209	1	158.79	39.45	0	63.31	0	1.033176	1.355573	380	13	1.1	0.9;
	5210	1	347.61	74.37	0	0	0	0.999109	-0.359974	220	2	1.1	0.9;
	5211	1	0	-0	0	18.22	0	1.039999	-21.26053	380	1	1.1	0.9;
	5212	1	-0	0	0	0.2	0	1.038537	-11.276713	220	5	1.1	0.9;
	5213	1	246.8	36.3	0	16.01	0	1.042076	-26.712515	220	5	1.1	0.9;
	5214	1	0	-0	0	0.88	0	1.035293	-3.021305	220	1	1.1	0.9;
	5215	1	-0	-0	0	0.89	0	1.025196	-6.168163	220	5	1.1	0.9;
	5216	1	0	-0	0	0.38	0	1.032101	-26.069945	150	9	1.1	0.9;
	5217	1	12.11	-0	0	0.6	0	1.010242	-48.583008	220	4	1.1	0.9;
	5218	1	68.68	21	0	0.56	0	1.009924	-30.100704	110	21	1.1	0.9;
	5219	1	23.03	8.3	0	0.09	0	1.024872	-25.973264	110	19	1.1	0.9;
	5220	1	58.5	8.3	0	4	0	1.011786	4.483442	154	1	1.1	0.9;
	5221	1	0	-0	0	0.2	0	0.967916	-28.697244	110	18	1.1	0.9;
	5222	1	13.9	1.9	0	3.05	0	1.022327	-7.463931	154	1	1.1	0.9;
	5223	1	4.61	1.7	0	0.49	0	1.045671	-20.30397	110	19	1.1	0.9;
	5224	1	27.73	17.2	0	0.36	0	0.948815	-33.811513	150	9	1.1	0.9;
	5225	2	0	0	0	0	0	1.050345	-16.499112	220	7	1.1	0.9;
	5226	2	0	0	0	0	0	0.992484	-19.76807	380	9	1.1	0.9;
	5227	1	1.9	1.4	0	1.07	0	0.971517	-24.768318	110	18	1.1	0.9;
	5228	1	44.95	18.7	0	5.36	0	1.034342	-5.624977	220	3	1.1	0.9;
	5229	1	-0	0	0	50.54	0	1.020723	-36.503257	220	22	1.1	0.9;
	5230	1	102.3	14.5	0	22.12	0	1.038229	-23.891029	154	1	1.1	0.9;
	5231	2	0	0	0	0	0	1.026163	-25.303176	110	18	1.1	0.9;
	5232	1	53.46	15.4	0	0.45	0	1.004065	8.888477	220	3	1.1	0.9;
	5233	1	-0	-0	0	14.04	0	1.036537	-11.313155	220	5	1.1	0.9;
	5234	1	264.3	111.6	0	34.93	0	1.049607	30.137973	380	3	1.1	0.9;
	5235	1	14.22	4.7	0	2.05	0	1.020203	-37.140827	110	12	1.1	0.9;
	5236	2	0	0	0	0	0	1.047863	-15.699293	220	7	1.1	0.9;
	5237	2	0	0	0	0	0	1.0542	-14.77106	220	5	1.1	0.9;
	5238	1	30.03	10	0	0.08	0	1.005045	-30.746404	110	21	1.1	0.9;
	5239	2	0	0	0	0	0	0.990479	-47.567504	380	4	1.1	0.9;
	5240	1	27.6	4	0	6.91	0	1.046727	2.053615	154	1	1.1	0.9;
	5241	1	0	-0	0	70.905048	0	1.022883	-3.838129	380	5	1.1	0.9;
	5242	1	0	0	0	2.59	0	1.055573	-23.781828	220	7	1.1	0.9;
	5243	1	-0	-0	0	29.01	0	1.009551	6.826901	400	1	1.1	0.9;
	5244	1	20.9	3	0	3.58	0	1.0215	-6.514417	154	1	1.1	0.9;
	5245	1	-0	-0	0	0.05	0	0.970823	-24.11316	150	9	1.1	0.9;
	5246	1	0	-0	0	1.06	0	1.031531	12.035142	150	8	1.1	0.9;
	5247	1	0	-0	0	0.22	0	1.017084	-0.826915	150	8	1.1	0.9;
	5248	1	4	0.1	0	0.09	0	1.045953	-29.59783	150	9	1.1	0.9;
	5249	1	-4.88	-2.79	0	0.38	0	0.976058	-29.616818	110	18	1.1	0.9;
	5250	1	101.2	48.9	0	7.78	0	0.991606	-34.386656	150	9	1.1	0.9;
	5251	2	0	0	0	0	0	1.038759	-22.367388	110	15	1.1	0.9;
	5252	1	0	-0	0	0.18	0	0.997303	-20.520889	110	18	1.1	0.9;
	5253	1	69.18	38.1	0	28.66	0	1.019086	-43.205705	220	6	1.1	0.9;
	5254	1	0	0	0	0	0	1.006283	-27.062298	150	9	1.1	0.9;
	5255	1	43.95	8.5	0	29.29	0	1.033206	8.744605	150	8	1.1	0.9;
	5256	1	91	36	0	5.47	0	0.998096	-17.006205	220	5	1.1	0.9;
	5257	1	-0	0	0	0.18	0	1.046449	-6.348504	220	5	1.1	0.9;
	5258	1	-0	-0	0	14.79	0	1.06951	26.086158	380	3	1.1	0.9;
	5259	1	-17.8	-3.78	0	0.08	0	1.022475	-15.124227	110	18	1.1	0.9;
	5260	1	0	0	0	2.24	0	1.027358	15.769371	154	1	1.1	0.9;
	5261	1	77	-239.9	0	4.21	0	1.0425	7.235152	220	3	1.1	0.9;
	5262	2	0	0	0	0	0	1.059967	-12.124644	380	16	1.1	0.9;
	5263	1	0	0	0	0.23	0	1.051592	-45.10801	380	10	1.1	0.9;
	5264	2	0	0	0	0	0	1.034663	13.971282	154	1	1.1	0.9;
	5265	1	18.12	11.2	0	2.96	0	0.968795	-30.437259	150	9	1.1	0.9;
	5266	1	2.2	0.5	0	0.06	0	1.026639	-34.881289	220	10	1.1	0.9;
	5267	1	72.18	44.7	0	2.72	0	1.016558	-30.283741	150	9	1.1	0.9;
	5268	1	66.97	41.5	0	0.39	0	1.012986	-32.963293	150	9	1.1	0.9;
	5269	1	40.44	13.3	0	0.94	0	0.997627	-30.728085	110	21	1.1	0.9;
	5270	1	-0	-0	0	7.4	0	1.035997	-21.458109	220	5	1.1	0.9;
	5271	1	69.18	19.7	0	6.99	0	1.019244	8.389274	150	8	1.1	0.9;
	5272	1	-0	-0	0	8.54	0	1.031303	33.539401	380	3	1.1	0.9;
	5273	2	0	0	0	0	0	1.026805	64.035538	150	1	1.1	0.9;
	5274	1	2.2	0.5	0	1.84	0	0.991018	-47.662261	150	10	1.1	0.9;
	5275	1	66.47	14.14	0	0.17	0	1.064437	-19.94623	110	19	1.1	0.9;
	5276	1	143.2	5.9	0	5.79	0	1.015292	-44.635585	220	4	1.1	0.9;
	5277	1	78.5	11.2	0	1.65	0	1.040114	8.289062	154	1	1.1	0.9;
	5278	2	0	0	0	0	0	1.03962	-24.585875	220	5	1.1	0.9;
	5279	1	158.8	105.3	0	21.76	0	1.014488	16.403747	150	8	1.1	0.9;
	5280	2	0	0	0	0	0	1.018895	-41.9271	220	4	1.1	0.9;
	5281	1	19.82	12.3	0	0.06	0	1.01376	-26.588907	150	9	1.1	0.9;
	5282	1	63.5	9.1	0	51.96	0	1.022227	-25.047372	154	1	1.1	0.9;
	5283	1	13.92	6.9	0	0.79	0	0.995772	-24.245834	110	18	1.1	0.9;
	5284	1	0	-0	0	26.46055	0	1.019426	-20.552231	220	20	1.1	0.9;
	5285	2	0	0	0	0	0	1.057359	-28.685756	220	12	1.1	0.9;
	5286	1	17.22	5.1	0	0.99	0	1.032855	-22.511244	220	5	1.1	0.9;
	5287	1	7	1	0	8.71	0	1.036152	1.812718	154	1	1.1	0.9;
	5288	1	-0	-0	0	3.06	0	0.995768	-15.366713	380	5	1.1	0.9;
	5289	1	-0	-0	0	165.34	0	1.067874	-38.358205	380	4	1.1	0.9;
	5290	2	0	0	0	0	0	1.009333	-36.158828	220	4	1.1	0.9;
	5291	1	25	3.6	0	1.15	0	1.026323	-15.478612	154	1	1.1	0.9;
	5292	1	-0	-0	0	0.86	0	1.011478	-48.444663	220	6	1.1	0.9;
	5293	2	0	0	0	0	0	1.008689	-5.494677	400	1	1.1	0.9;
	5294	1	31.2	4.4	0	0.35	0	1.015317	1.679636	154	1	1.1	0.9;
	5295	1	-0	-0	0	2.08	0	1.028907	20.742927	220	8	1.1	0.9;
	5296	1	16.02	4.7	0	0.94	0	1.011237	18.573824	110	8	1.1	0.9;
	5297	1	55.26	11	0	7.68	0	1.038998	-24.689993	220	5	1.1	0.9;
	5298	1	0	-0	0	0.48	0	1.075555	-26.412788	150	9	1.1	0.9;
	5299	2	0	0	0	0	0	1.045101	33.747987	380	8	1.1	0.9;
	5300	1	-0	-0	0	1.3	0	1.039522	-11.286385	220	5	1.1	0.9;
	5301	2	0	0	0	0	0	1.034956	49.088982	220	3	1.1	0.9;
	5302	1	267.1	57.4	0	13.64	0	1.029755	-35.443964	380	6	1.1	0.9;
	5303	1	40.05	20	0	0.05	0	0.994526	-50.407214	220	6	1.1	0.9;
	5304	1	-0	0	0	4.87	0	1.040268	-30.537588	150	9	1.1	0.9;
	5305	2	0	0	0	0	0	1.029341	-5.808129	220	13	1.1	0.9;
	5306	1	-0.47	0.1	0	0.35	0	0.999778	-27.933194	110	18	1.1	0.9;
	5307	1	-0	0	0	6.4	0	0.998324	-28.72802	150	9	1.1	0.9;
	5308	1	0	-0	0	42.19206	0	0.980119	-18.492863	380	5	1.1	0.9;
	5309	2	0	0	0	0	0	1.042975	-17.817254	220	7	1.1	0.9;
	5310	1	97.06	-0.33	0	1.24	0	1.040939	-2.439408	220	13	1.1	0.9;
	5311	1	-0	-0	0	7.85	0	1.043286	15.805059	220	8	1.1	0.9;
	5312	1	59.87	19.9	0	0.32	0	1.029474	-31.041682	110	21	1.1	0.9;
	5313	1	15.22	2.5	0	0.69	0	1.002792	-21.384316	110	18	1.1	0.9;
	5314	1	0	0	0	29.25	0	1.081616	-27.637386	380	1	1.1	0.9;
	5315	1	0	-0	0	3.86	0	1.020828	16.809684	380	3	1.1	0.9;
	5316	1	-0	-0	0	30.04	0	1.071573	38.707798	380	3	1.1	0.9;
	5317	1	293.2	87.4	0	26.36	0	1.019252	-44.845345	220	5	1.1	0.9;
	5318	1	46.95	54.6	0	29.46	0	0.968294	-32.433569	150	9	1.1	0.9;
	5319	1	21.12	7.7	0	0.7	0	1.037618	-26.094347	110	19	1.1	0.9;
	5320	1	18.12	5.9	0	1.52	0	1.030478	3.320751	150	8	1.1	0.9;
	5321	1	185.2	26.4	0	12.41	0	1.020916	-7.445394	154	1	1.1	0.9;
	5322	1	41.2	56.52	0	0	0	0.948009	-19.391946	110	18	1.1	0.9;
	5323	2	0	0	0	0	0	1.031211	47.015813	220	3	1.1	0.9;
	5324	1	0	0	0	15.14	0	1.026268	-4.598739	220	3	1.1	0.9;
	5325	1	46.55	-18.9	0	14.78	0	1.031465	-21.73042	220	3	1.1	0.9;
	5326	1	42.45	-10.2	0	10.02	0	1.068222	-27.526345	220	23	1.1	0.9;
	5327	1	-0	-0	0.225625	26.287525	0	1.086214	-24.82175	380	21	1.1	0.9;
	5328	1	288.7	117.3	0	183.57	0	1.02561	-41.417935	380	6	1.1	0.9;
	5329	2	0	0	0	0	0	1.015053	-43.535832	220	4	1.1	0.9;
	5330	1	-12.61	4.1	0	0.33	0	1.015595	-28.749261	150	9	1.1	0.9;
	5331	1	-0	0	0	0.23	0	0.997637	-36.411037	150	10	1.1	0.9;
	5332	1	366.6	57.4	0	60.78	0	1.015602	-51.421675	380	4	1.1	0.9;
	5333	1	83.59	-7.2	0	12.95	0	0.998616	-47.446935	220	10	1.1	0.9;
	5334	1	17.52	6.6	0	3.79	0	1.028446	-13.059233	220	5	1.1	0.9;
	5335	1	1.6	0.8	0	0.79	0	0.982132	-28.928227	110	18	1.1	0.9;
	5336	1	-0	0	0	0.57	0	1.014454	-15.830309	110	18	1.1	0.9;
	5337	1	0	0	0.090774	1.480444	0	1.017114	-28.690154	220	18	1.1	0.9;
	5338	1	19.4	2.8	0	4.73	0	1.025419	11.439032	154	1	1.1	0.9;
	5339	1	-16.62	8.7	0	7.68	0	0.996653	-32.801882	220	10	1.1	0.9;
	5340	2	0	0	0	0	0	1.044898	-11.45967	220	5	1.1	0.9;
	5341	1	64.27	17.5	0	0.43	0	1.005913	-22.013831	220	5	1.1	0.9;
	5342	1	32.14	13.1	0	1.72	0	1.028756	-1.361457	150	8	1.1	0.9;
	5343	1	-0	-0	0	142.05	0	1.070746	-6.106433	380	14	1.1	0.9;
	5344	1	21.72	7.2	0	0.64	0	0.987054	-31.911414	110	21	1.1	0.9;
	5345	1	0	0	0	3.36	0	0.999728	-26.068443	154	1	1.1	0.9;
	5346	1	0	0	0	0.81	0	1.038003	27.563022	110	8	1.1	0.9;
	5347	1	8.71	2.5	0	0.66	0	0.974897	-31.802144	110	18	1.1	0.9;
	5348	2	0	0	0	0	0	1.013341	-43.142787	220	4	1.1	0.9;
	5349	1	5.5	0.8	0	7.28	0	1.039682	10.492143	154	1	1.1	0.9;
	5350	1	0	0	0	28.208233	0	0.967222	-24.900034	380	5	1.1	0.9;
	5351	1	0	0	0	3.28	0	1.042756	-13.916064	220	5	1.1	0.9;
	5352	2	0	0	0	0	0	1.022029	8.423604	154	1	1.1	0.9;
	5353	1	0	-0	0	0.34	0	1.054254	-27.918373	110	3	1.1	0.9;
	5354	1	70.18	19.5	0	3.59	0	1.041191	-8.880271	220	5	1.1	0.9;
	5355	1	229	77.1	0	36.732397	0	1.030605	-38.511768	380	6	1.1	0.9;
	5356	1	7	1	0	4.5	0	1.039815	1.985184	154	1	1.1	0.9;
	5357	1	0	0	0	0	0	1.010431	-30.336741	150	9	1.1	0.9;
	5358	1	0	0	0	0.43	0	0.942834	-32.683782	150	9	1.1	0.9;
	5359	1	0	-0	0	1.54	0	0.903388	-35.988527	150	9	1.3	0.7;
	5360	1	173.6	105.6	0	21.26	0	1.012933	-15.056043	380	3	1.1	0.9;
	5361	1	-0	0	0	2.77	0	0.968713	-43.493259	220	10	1.1	0.9;
	5362	1	152.73	-17.2	0	45.85	0	1.03922	-24.598476	220	5	1.1	0.9;
	5363	1	79.79	49.4	0	2.24	0	1.006246	-31.812632	150	9	1.1	0.9;
	5364	1	110.6	38.1	0	0.98	0	1.026444	-2.016826	150	8	1.1	0.9;
	5365	2	0	0	0	0	0	1.065324	-11.846174	220	5	1.1	0.9;
	5366	1	66.07	12.5	0	2.08	0	1.020717	-47.279801	220	4	1.1	0.9;
	5367	1	40.5	5.7	0	5.29	0	1.014817	-5.921008	154	1	1.1	0.9;
	5368	2	0	0	0	0	0	1.058977	-0.563499	110	19	1.1	0.9;
	5369	2	0	0	0	0	0	1.037177	5.654189	220	13	1.1	0.9;
	5370	1	13.92	4.1	0	2.27	0	0.995273	-14.1203	110	18	1.1	0.9;
	5371	1	-13.32	20.6	0	1.16	0	1.060178	-30.916528	150	9	1.1	0.9;
	5372	1	21.02	8	0	13.14	0	1.059752	-41.178352	380	4	1.1	0.9;
	5373	1	-8.71	2.9	0	0.42	0	0.956585	-31.461459	150	9	1.1	0.9;
	5374	1	13.2	3.25	0	5.66	0	1.013322	61.623338	150	1	1.1	0.9;
	5375	1	48.8	6.9	0	1.59	0	1.008396	2.085281	154	1	1.1	0.9;
	5376	1	408.3	84.5	0	18.02	0	1.025658	-37.244067	380	6	1.1	0.9;
	5377	1	-0	-0	0	2.96	0	1.033071	65.032272	220	1	1.1	0.9;
	5378	1	13.21	4.4	0	0.86	0	0.998714	-44.729132	110	22	1.1	0.9;
	5379	2	0	0	0	0	0	1.10493	-7.012792	380	5	1.3	0.7;
	5380	1	127.3	90.5	0	6.8	0	1.018969	3.256671	380	3	1.1	0.9;
	5381	1	0	0	0	0	0	1.013102	-33.068927	150	9	1.1	0.9;
	5382	1	0.8	0.5	0	0	0	0.986578	-31.221938	150	9	1.1	0.9;
	5383	1	0	-0	0	24.7	0	1.044201	-24.790628	220	5	1.1	0.9;
	5384	1	48.05	28.8	0	0.02	0	1.004283	-32.089465	150	9	1.1	0.9;
	5385	1	0	0	0	17.83	0	1.040121	-21.319331	380	1	1.1	0.9;
	5386	1	243.45	104.28	0	4.66	0	1.019636	-7.825709	220	13	1.1	0.9;
	5387	1	37.94	51.4	0	0.55	0	1.059001	10.223443	380	3	1.1	0.9;
	5388	1	159.5	42	0	2.3	0	1.006693	-15.086388	220	5	1.1	0.9;
	5389	1	0	-0	0	55.700331	0	0.995955	-16.116371	220	24	1.1	0.9;
	5390	1	0	-0	0	3.62	0	1.03767	-3.28139	220	1	1.1	0.9;
	5391	1	57.06	11.9	0	1.68	0	1.020385	-48.75014	220	4	1.1	0.9;
	5392	1	-0	0	0	0.62	0	0.963665	-27.354454	150	9	1.1	0.9;
	5393	1	-0	0	0	9.43	0	1.053922	-15.971462	220	5	1.1	0.9;
	5394	1	0	0	0	2.77	0	1.027555	2.293855	154	1	1.1	0.9;
	5395	2	0	0	0	0	0	1.05041	-27.819641	220	5	1.1	0.9;
	5396	1	-0	-0	0	1.52	0	0.994262	-24.110122	150	9	1.1	0.9;
	5397	2	0	0	0	0	0	1.016845	-29.585014	110	21	1.1	0.9;
	5398	2	0	0	0	0	0	1.041434	-41.871134	380	6	1.1	0.9;
	5399	1	20.9	3	0	2.4	0	1.025297	18.608931	154	1	1.1	0.9;
	5400	1	241.62	52.78	0	0.82	0	1.022935	-14.201511	220	5	1.1	0.9;
	5401	2	0	0	0	0	0	1.056816	52.771969	380	3	1.1	0.9;
	5402	1	54.76	33.9	0	0.46	0	0.99228	-31.5665	150	9	1.1	0.9;
	5403	2	0	0	0	0	0	1.032334	13.97906	380	3	1.1	0.9;
	5404	1	5.21	1.7	0	0.87	0	1.037044	-31.762642	110	12	1.1	0.9;
	5405	1	0	0	0	0	0	1.023046	-30.120615	150	9	1.1	0.9;
	5406	1	0	-0	0	36.42	0	1.105578	-24.714218	380	9	1.3	0.7;
	5407	1	6.61	2.4	0	0.81	0	0.985403	-26.680829	110	18	1.1	0.9;
	5408	1	11.91	4.3	0	0.66	0	1.025733	-26.104482	110	19	1.1	0.9;
	5409	2	0	0	0	0	0	1.021468	-39.327661	220	4	1.1	0.9;
	5410	1	69.78	26	0	5.18	0	1.042284	-21.670669	220	5	1.1	0.9;
	5411	1	-0	-0	0	3.79	0	1.02277	-5.585233	154	1	1.1	0.9;
	5412	2	0	0	0	0	0	1.07527	8.89403	380	14	1.1	0.9;
	5413	1	-0	-0	0	5.57	0	1.032887	-18.313717	220	5	1.1	0.9;
	5414	1	0	0	0	0	0	1.036655	-33.030803	150	9	1.1	0.9;
	5415	1	-0	0	0	1.81	0	1.008861	-7.879657	154	1	1.1	0.9;
	5416	1	0	0	0	25.13	0	1.003937	11.803413	380	2	1.1	0.9;
	5417	1	32.74	5.3	0	3.07	0	1.061368	-12.472756	220	5	1.1	0.9;
	5418	1	-0	0	0	3.96	0	1.03456	-21.58226	220	5	1.1	0.9;
	5419	1	349.3	144.3	0	0.54	0	1.024125	-15.192507	220	5	1.1	0.9;
	5420	1	56.96	1.94	0	26.48	0	1.017861	-23.368601	380	5	1.1	0.9;
	5421	1	-0	-0	0	0.92	0	0.995766	-15.366638	380	5	1.1	0.9;
	5422	1	104.5	14.9	0	0.67	0	1.026675	-6.579714	154	1	1.1	0.9;
	5423	1	28.63	7.3	0	4.36	0	1.005727	5.973185	110	8	1.1	0.9;
	5424	1	0	0	0	2.04	0	1.082032	-24.688907	110	19	1.1	0.9;
	5425	2	0	0	0	0	0	1.074405	-36.140712	380	4	1.1	0.9;
	5426	1	152.8	38	0	1.46	0	1.023076	2.523931	150	8	1.1	0.9;
	5427	1	233.8	33.4	0	1.22	0	1.01296	-22.278965	154	1	1.1	0.9;
	5428	1	53.36	44.2	0	22.89	0	1.028043	26.098856	380	3	1.1	0.9;
	5429	2	0	0	0	0	0	1.043614	10.777174	220	3	1.1	0.9;
	5430	1	26.4	3.8	0	9.18	0	1.026594	4.059264	154	1	1.1	0.9;
	5431	1	58.5	8.3	0	1.2	0	1.015774	-14.1263	154	1	1.1	0.9;
	5432	1	-0	-0	0	106.24	0	1.021874	7.21054	400	1	1.1	0.9;
	5433	1	121.6	49.3	0	10	0	0.991152	5.727907	220	3	1.1	0.9;
	5434	1	69.5	9.8	0	1.91	0	1.007927	-10.759175	154	1	1.1	0.9;
	5435	1	103.3	24.1	0	9.41	0	1.009014	-47.601522	220	4	1.1	0.9;
	5436	1	-4.86	-2.12	0	0.26	0	1.000228	-24.534422	110	18	1.1	0.9;
	5437	1	-0	0	0	11.59	0	1.035669	-15.838076	380	1	1.1	0.9;
	5438	1	104.7	-12.8	0	2.25	0	1.050658	-19.170239	220	7	1.1	0.9;
	5439	1	-0	0	0	2.72	0	1.038274	38.762763	380	3	1.1	0.9;
	5440	1	116.4	-8.8	0	8.39	0	1.049377	-10.587189	220	3	1.1	0.9;
	5441	1	138.01	34.55	0	8.44	0	1.033391	-3.238779	220	5	1.1	0.9;
	5442	1	-0.17	-0.84	0	0	0	1.014377	16.388914	150	8	1.1	0.9;
	5443	1	2.8	1	0	0.95	0	1.019611	-28.986553	110	19	1.1	0.9;
	5444	2	0	0	0	0	0	1.059208	10.233319	380	3	1.1	0.9;
	5445	1	45.45	7.1	0	1.05	0	1.02477	-25.099863	110	3	1.1	0.9;
	5446	1	-0	0	0	0.57	0	0.960277	-31.774093	150	9	1.1	0.9;
	5447	1	3.33	2.95	0	1.5	0	0.955232	-30.324084	110	18	1.1	0.9;
	5448	1	-0	-0	0	1.06	0	1.021965	-27.426392	150	7	1.1	0.9;
	5449	1	27.9	4	0	4.18	0	1.028401	10.789888	154	1	1.1	0.9;
	5450	1	-0	-0	0	1.17	0	1.058314	0.575404	150	8	1.1	0.9;
	5451	1	-0	-0	0	1.82	0	1.068356	-48.383117	220	6	1.1	0.9;
	5452	1	82.09	39.8	0	2.35	0	0.992868	-48.243558	150	10	1.1	0.9;
	5453	1	45.58	-1.95	0	1.8	0	1.022873	62.697236	150	1	1.1	0.9;
	5454	1	0	-0	0	1.43	0	1.0207	-25.552974	150	9	1.1	0.9;
	5455	1	-0	-0	0	5.27	0	1.041319	-8.044004	220	5	1.1	0.9;
	5456	1	0	-0	0	1.84	0	0.95194	-22.56403	110	19	1.1	0.9;
	5457	2	0	0	0	0	0	1.044659	-9.940212	110	19	1.1	0.9;
	5458	1	0	0	0	9.09	0	1.034223	-21.767033	220	5	1.1	0.9;
	5459	1	211.3	69.3	0	7.96	0	0.986171	-50.721367	220	6	1.1	0.9;
	5460	1	-15.89	-6.42	0	38.657071	0	0.981174	-18.227809	380	5	1.1	0.9;
	5461	2	0	0	0	0	0	1.055685	-18.707132	380	5	1.1	0.9;
	5462	1	114.42	23.58	0	46.99	0	1.042235	7.92914	380	13	1.1	0.9;
	5463	1	88.4	-5.9	0	3.32	0	1.020822	-42.641009	220	4	1.1	0.9;
	5464	1	25.23	3.3	0	35.45	0	1.046547	-39.052752	380	4	1.1	0.9;
	5465	2	0	0	0	0	0	0.997959	-47.341255	220	4	1.1	0.9;
	5466	1	67.5	9.6	0	26.5	0	1.037366	-19.083886	400	1	1.1	0.9;
	5467	1	1	0.3	0	0.46	0	1.002425	-31.412163	110	19	1.1	0.9;
	5468	1	110	6.83	0	13.51	0	1.027297	-9.338361	220	14	1.1	0.9;
	5469	1	106.6	-18.8	0	4.1	0	1.068254	-38.443073	220	5	1.1	0.9;
	5470	1	3.3	1.1	0	0.25	0	1.041272	-33.026205	110	22	1.1	0.9;
	5471	2	0	0	0	0	0	1.026437	-47.947151	380	6	1.1	0.9;
	5472	1	1.3	0	0	0.41	0	1.026232	-19.774007	110	18	1.1	0.9;
	5473	1	-0	0	0	4.42	0	1.050309	30.291614	380	1	1.1	0.9;
	5474	1	8.91	2.5	0	0.7	0	0.973291	-27.712296	110	18	1.1	0.9;
	5475	1	1	1	0	0.27	0	1.052475	-24.929716	110	23	1.1	0.9;
	5476	1	0	0	0	90.33	0	1.031757	-46.789773	380	6	1.1	0.9;
	5477	1	-0	-0	0	18.02	0	1.069678	-26.5629	220	5	1.1	0.9;
	5478	1	-0	0	0	91.78	0	0.996923	13.538273	400	1	1.1	0.9;
	5479	2	0	0	0	0	0	1.05048	29.622719	380	13	1.1	0.9;
	5480	2	0	0	0	0	0	1.043829	-19.748499	220	11	1.1	0.9;
	5481	2	0	0	0	0	0	1.05825	-9.184906	380	5	1.1	0.9;
	5482	2	0	0	0	0	0	1.042942	-7.385046	220	5	1.1	0.9;
	5483	1	144	56.6	0	5.61	0	1.005753	-37.426895	220	6	1.1	0.9;
	5484	1	0	0	0	30.55	0	1.084447	-36.839736	150	9	1.3	0.7;
	5485	1	-0	0	0	5.75	0	1.032409	-3.136568	220	1	1.1	0.9;
	5486	2	0	0	0	0	0	1.055199	4.459301	380	5	1.1	0.9;
	5487	2	0	0	0	0	0	1.050272	-17.936662	220	18	1.1	0.9;
	5488	2	0	0	0	0	0	1.037007	-30.257494	220	5	1.1	0.9;
	5489	1	11.81	2.4	0	0.16	0	0.996826	-17.380713	110	18	1.1	0.9;
	5490	2	0	0	0	0	0	1.062487	-15.472573	380	5	1.1	0.9;
	5491	1	0	-0	0.505613	83.804783	0	1.038788	3.64887	380	3	1.1	0.9;
	5492	1	-0	-0	0	1.23	0	1.011986	5.623815	220	3	1.1	0.9;
	5493	1	0	0	0	3.21	0	1.008785	-47.094951	220	6	1.1	0.9;
	5494	1	0	-0	0	4.69	0	0.99706	-37.536546	220	4	1.1	0.9;
	5495	1	195.74	-32.87	0	0.83	0	1.067943	52.899561	380	3	1.1	0.9;
	5496	1	1.63	8.8	0	0.16	0	0.965112	-31.467815	150	9	1.1	0.9;
	5497	1	0	0	0.252685	6.907589	0	1.043684	-20.144478	220	19	1.1	0.9;
	5498	1	574.3	94.6	0	16.63	0	1.011277	-36.310892	380	6	1.1	0.9;
	5499	1	20.12	7.9	0	0.08	0	1.043636	-26.106859	110	15	1.1	0.9;
	5500	1	111.5	24.8	0	9.65	0	1.021558	-54.126055	220	4	1.1	0.9;
	5501	1	-0	0	0	16.84	0	1.004736	-38.330114	220	22	1.1	0.9;
	5502	1	-0	-0	0	0.58	0	1.035356	-20.817081	220	5	1.1	0.9;
	5503	1	342.7	82.2	0	75.02	0	1.059208	-12.863591	380	16	1.1	0.9;
	5504	2	0	0	0	0	0	1.033273	-38.800825	110	22	1.1	0.9;
	5505	1	92.59	8.37	0	16.58	0	1.049006	-8.701195	380	13	1.1	0.9;
	5506	1	29.53	25.1	0	0.8	0	0.890999	-36.148908	150	9	1.3	0.7;
	5507	1	17.62	7.7	0	10.73	0	1.02587	-39.761737	110	22	1.1	0.9;
	5508	1	0	0	0	0	0	1.012924	-30.696006	150	9	1.1	0.9;
	5509	1	0	0	0.314057	17.296509	0	0.994902	-6.469273	380	2	1.1	0.9;
	5510	1	499	191.1	0	60.73	0	1.017387	-46.391346	380	6	1.1	0.9;
	5511	1	67.68	6.2	0	6.83	0	1.010438	-60.067724	220	4	1.1	0.9;
	5512	2	0	0	0	0	0	1.017916	-41.936012	220	4	1.1	0.9;
	5513	2	0	0	0	0	0	1.030251	-5.016549	220	13	1.1	0.9;
	5514	1	0	-0	0	0.4	0	1.022118	51.364458	220	3	1.1	0.9;
	5515	1	22.12	8	0	0.14	0	1.072081	-17.672798	110	19	1.1	0.9;
	5516	1	18.22	6	0	0.14	0	1.032259	-34.232499	110	12	1.1	0.9;
	5517	1	34.24	-19.6	0	3.59	0	0.974654	-43.199572	150	10	1.1	0.9;
	5518	1	4.4	2.2	0	1.19	0	1.037784	-37.319885	110	22	1.1	0.9;
	5519	1	0	-0	0	1.98	0	1.044448	-24.851567	220	5	1.1	0.9;
	5520	1	55.7	7.9	0	0.76	0	1.007081	-2.487018	154	1	1.1	0.9;
	5521	1	30.43	10.1	0	1.07	0	1.018613	-27.370398	110	19	1.1	0.9;
	5522	1	-0	-0	0	1.18	0	1.061913	0.592035	220	5	1.1	0.9;
	5523	1	150.2	25.5	0	42.49	0	1.02722	1.492171	380	3	1.1	0.9;
	5524	1	14.52	5.3	0	0.49	0	1.025708	-23.614535	110	15	1.1	0.9;
	5525	1	236.33	59.35	0	18.71	0	1.03757	-14.19565	220	5	1.1	0.9;
	5526	1	-8.71	-19.54	0	0	0	1.041416	17.400508	150	8	1.1	0.9;
	5527	1	-0	-0	0	58.79	0	1.041969	-16.353168	380	7	1.1	0.9;
	5528	1	93.6	77.1	0	91.89	0	1.024151	19.97811	380	3	1.1	0.9;
	5529	1	0	0	0	0.36	0	1.033411	-3.238958	220	5	1.1	0.9;
	5530	2	0	0	0	0	0	1.011393	3.278996	220	3	1.1	0.9;
	5531	1	-0	-0	0	0.82	0	1.027817	50.048565	220	3	1.1	0.9;
	5532	1	-0	0	0	53.24	0	1.020753	-42.72507	220	6	1.1	0.9;
	5533	2	0	0	0	0	0	1.046404	-13.450032	220	5	1.1	0.9;
	5534	1	-0	-0	0	0.6	0	1.029683	-33.201948	110	1	1.1	0.9;
	5535	1	68.08	42.2	0	4.4	0	1.009172	-32.442398	150	9	1.1	0.9;
	5536	1	85.8	12.2	0	1.7	0	1.010983	-7.372311	154	1	1.1	0.9;
	5537	1	-0.53	-0.91	0.090774	7.680444	0	1.068354	-24.94168	220	18	1.1	0.9;
	5538	1	91.9	13.1	0	7.35	0	1.035147	1.510321	154	1	1.1	0.9;
	5539	1	12.21	4.4	0	2.77	0	0.960918	-22.964315	110	19	1.1	0.9;
	5540	1	0	-0	0	0.94	0	0.991434	-20.676063	110	18	1.1	0.9;
	5541	1	11.71	2.2	0	0.48	0	1.059687	-32.169418	110	3	1.1	0.9;
	5542	1	6.71	3	0	1.3	0	1.035299	-26.295162	110	15	1.1	0.9;
	5543	1	0	-0	0	7.08	0	1.017345	-13.937245	380	3	1.1	0.9;
	5544	1	47.65	29.5	0	4.19	0	0.90329	-35.996951	150	9	1.3	0.7;
	5545	1	33.04	8.1	0	1.55	0	1.004202	-44.550982	110	22	1.1	0.9;
	5546	2	0	0	0	0	0	1.063123	-24.286461	220	5	1.1	0.9;
	5547	1	10.01	3	0	0.47	0	1.001522	-30.896248	110	21	1.1	0.9;
	5548	1	-0.94	-6.39	0	1.76	0	0.975272	-32.247157	150	9	1.1	0.9;
	5549	1	30.03	12.2	0	0.93	0	1.018012	-0.960924	150	8	1.1	0.9;
	5550	1	0	-0	0	10.83	0	1.040307	-26.337142	220	5	1.1	0.9;
	5551	1	15.12	5	0	0.69	0	1.012651	-38.597824	110	12	1.1	0.9;
	5552	1	23.93	8.7	0	0.42	0	1.023176	-25.397739	110	19	1.1	0.9;
	5553	1	2.22	7.8	0	19.16	0	0.970518	-32.199871	150	9	1.1	0.9;
	5554	1	50.26	16.5	0	0.69	0	0.954288	-32.366016	110	21	1.1	0.9;
	5555	1	0	-0	0	17.79	0	1.034991	8.405677	380	3	1.1	0.9;
	5556	2	0	0	0	0	0	1.067161	12.086246	380	3	1.1	0.9;
	5557	1	21.22	6.7	0	0.2	0	0.999397	-31.323747	110	21	1.1	0.9;
	5558	1	-0	-0	0	15.17	0	1.016579	16.58643	380	8	1.1	0.9;
	5559	1	0	-0	0	2.66	0	0.99875	-43.311421	220	10	1.1	0.9;
	5560	1	0	-0	0	1.56	0	1.01234	-28.707192	150	9	1.1	0.9;
	5561	1	31.13	12	0	1.03	0	1.032807	-26.569915	110	15	1.1	0.9;
	5562	1	-0	0	0.19855	29.549	0	1.064025	-46.256666	380	10	1.1	0.9;
	5563	1	21.62	7.8	0	1.21	0	1.086139	-17.202714	110	19	1.1	0.9;
	5564	2	0	0	0	0	0	1.038385	-12.644533	220	5	1.1	0.9;
	5565	1	5.11	1.8	0	0.62	0	1.039021	-9.508375	110	19	1.1	0.9;
	5566	1	-0	0	0	1.59	0	1.010037	1.120766	380	3	1.1	0.9;
	5567	1	0	0	0	3.59	0	1.015187	-19.059134	220	5	1.1	0.9;
	5568	1	-0	-0	0	6.29	0	1.003095	-36.693826	150	9	1.1	0.9;
	5569	1	-0	0	0	0.45	0	0.983695	-16.09428	110	18	1.1	0.9;
	5570	1	23.53	8.5	0	0.64	0	0.991081	-28.405577	110	19	1.1	0.9;
	5571	1	-0	0	0	1.01	0	1.032023	-22.830475	220	5	1.1	0.9;
	5572	2	0	0	0	0	0	1.033768	-10.89418	220	6	1.1	0.9;
	5573	1	0	-0	0	2.94	0	1.068948	-23.910257	220	5	1.1	0.9;
	5574	1	99.7	6	0	2.06	0	1.033871	-12.192277	220	5	1.1	0.9;
	5575	1	-0	0	0	59.52	0	1.040449	-41.112082	380	6	1.1	0.9;
	5576	1	220	68.3	0	3.09	0	1.034825	-23.248284	220	7	1.1	0.9;
	5577	1	0	-0	0	19.73	0	1.041024	-27.643983	380	6	1.1	0.9;
	5578	1	26.63	8.6	0	4.44	0	0.934329	-33.800641	110	18	1.1	0.9;
	5579	1	16.22	5.3	0	0.25	0	1.010415	-36.389942	110	12	1.1	0.9;
	5580	1	130.7	89.1	0	0.08	0	1.019717	-5.305656	220	3	1.1	0.9;
	5581	1	0	0	0	88.66	0	1.042638	3.534523	400	1	1.1	0.9;
	5582	1	3.2	1	0	0.64	0	1.00513	-18.733291	110	18	1.1	0.9;
	5583	1	1.9	0.7	0	0.48	0	1.069405	-18.465793	110	19	1.1	0.9;
	5584	1	-0.19	-14.8	0	3.08	0	1.026351	16.339817	220	3	1.1	0.9;
	5585	1	0	0	0	0.17	0	1.026637	-38.41479	110	22	1.1	0.9;
	5586	1	-0	0	0	0	0	1.051067	2.324898	220	5	1.1	0.9;
	5587	1	-0	-0	0	0.33	0	1.030648	-10.826911	150	2	1.1	0.9;
	5588	1	0	0	0	0.05	0	0.962214	-32.51376	150	9	1.1	0.9;
	5589	1	0	0	0	173.904854	0	1.031495	-31.068515	380	5	1.1	0.9;
	5590	1	55.4	7.9	0	6.35	0	1.010152	2.8802	154	1	1.1	0.9;
	5591	1	118.6	41.5	0	0.58	0	1.024918	-2.12304	150	8	1.1	0.9;
	5592	1	20.82	7.9	0	18	0	1.008732	-51.786702	220	4	1.1	0.9;
	5593	1	19.52	7	0	1.73	0	1.051054	-25.445268	110	15	1.1	0.9;
	5594	2	0	0	0	0	0	1.048652	-33.520974	220	11	1.1	0.9;
	5595	1	168.5	54.8	0	5.34	0	1.032832	2.906602	150	8	1.1	0.9;
	5596	1	172.8	11.3	0	8.61	0	1.024078	-17.267999	220	3	1.1	0.9;
	5597	1	72	10.3	0	5.95	0	1.023283	12.805407	154	1	1.1	0.9;
	5598	1	-0	0	0	14.53	0	1.030805	-9.685276	220	3	1.1	0.9;
	5599	1	8.3	1.2	0	4.2	0	1.011688	-8.700111	154	1	1.1	0.9;
	5600	1	-0	-0	0.062219	7.992656	0	0.970986	68.621915	380	1	1.1	0.9;
	5601	1	0	0	0	12.61	0	1.023746	-21.503966	220	1	1.1	0.9;
	5602	2	0	0	0	0	0	1.032916	9.689117	154	1	1.1	0.9;
	5603	2	0	0	0	0	0	1.027661	0.058712	154	1	1.1	0.9;
	5604	1	83.4	11.9	0	1.4	0	1.013716	-7.690535	154	1	1.1	0.9;
	5605	1	-0	0	0	20.96	0	1.011071	-33.040519	380	1	1.1	0.9;
	5606	1	-0	0	0	0.74	0	1.036946	-30.614337	150	9	1.1	0.9;
	5607	1	34.14	12.4	0	0.4	0	1.049101	-25.872086	110	19	1.1	0.9;
	5608	1	449.4	14.6	0	87.31	0	1.022612	-51.075631	380	4	1.1	0.9;
	5609	1	23.31	-6.67	0	2.15	0	1.013915	60.756802	150	1	1.1	0.9;
	5610	1	99.5	9.2	0	27.55	0	1.041396	-24.529577	220	5	1.1	0.9;
	5611	2	0	0	0	0	0	1.079526	43.868294	380	3	1.1	0.9;
	5612	1	22.4	3.1	0	3.4	0	1.033844	2.402845	154	1	1.1	0.9;
	5613	1	76.6	10.9	0	3.49	0	0.999749	-4.494898	154	1	1.1	0.9;
	5614	1	0	0	0	0.59	0	0.999853	-28.287278	110	18	1.1	0.9;
	5615	1	48.8	6.9	0	6.29	0	1.026737	-8.119017	154	1	1.1	0.9;
	5616	1	225.2	35.2	0	11.99	0	1.072539	-35.51358	220	5	1.1	0.9;
	5617	1	38.04	3.4	0	3.43	0	0.975099	-49.885167	220	4	1.1	0.9;
	5618	1	-0	0	0	1.75	0	1.014579	-4.867251	150	8	1.1	0.9;
	5619	1	-0	-0	0.181549	16.110889	0	1.037855	-26.053406	220	18	1.1	0.9;
	5620	2	0	0	0	0	0	1.055555	10.797767	380	14	1.1	0.9;
	5621	1	6.51	3.7	0	0.24	0	0.966759	-29.955048	110	18	1.1	0.9;
	5622	1	158.3	-20.5	0	2.07	0	1.025208	-16.586535	220	3	1.1	0.9;
	5623	1	513.4	199.8	0	90.272366	0	1.034284	-33.747114	380	6	1.1	0.9;
	5624	1	152.7	6	0	0.36	0	1.042101	-28.475363	120	20	1.1	0.9;
	5625	1	0	0	0	80	0	1.089754	-30.701634	380	12	1.1	0.9;
	5626	1	13.72	4.7	0	0.22	0	0.978647	-28.322089	110	18	1.1	0.9;
	5627	1	-0	-0	0	0.53	0	1.021141	-32.671306	220	5	1.1	0.9;
	5628	1	-4.43	-3.56	0	0.26	0	0.967096	-28.823328	110	18	1.1	0.9;
	5629	1	0	-0	0	1.36	0	1.004946	-34.428357	110	21	1.1	0.9;
	5630	1	0	0	0	11.52	0	1.037673	-4.438859	220	5	1.1	0.9;
	5631	1	105.5	31.1	0	9.63	0	0.986345	-43.712619	220	4	1.1	0.9;
	5632	1	27.46	5.87	0	0.8	0	0.966943	-29.462839	110	18	1.1	0.9;
	5633	1	29.13	13.5	0	3.62	0	0.965817	-55.158907	220	6	1.1	0.9;
	5634	1	21.12	13.1	0	0.36	0	0.945107	-32.784726	150	9	1.1	0.9;
	5635	1	0	0	0	13.49	0	0.984669	-35.597077	150	9	1.1	0.9;
	5636	1	0	0	0	0	0	1.043966	-30.135632	150	9	1.1	0.9;
	5637	1	1.38	-1.15	0	0.5	0	1.029059	-28.575685	110	19	1.1	0.9;
	5638	1	0	-0	0	4.06	0	0.961838	-31.677351	150	9	1.1	0.9;
	5639	1	-0	0	0	0.35	0	1.011634	-15.392348	110	18	1.1	0.9;
	5640	1	64.33	-57.14	0	8.9	0	1.030182	7.58524	220	13	1.1	0.9;
	5641	1	-0	-0	0	5.98	0	1.04872	-26.545657	220	5	1.1	0.9;
	5642	1	72.78	45.1	0	11.93	0	0.977041	-35.24252	150	9	1.1	0.9;
	5643	1	-0	-0	0	1.1	0	0.984517	-56.623592	220	10	1.1	0.9;
	5644	1	45.05	40	0	1.79	0	0.996385	8.148142	220	3	1.1	0.9;
	5645	2	0	0	0	0	0	1.000282	-50.759087	220	6	1.1	0.9;
	5646	1	20.02	8.5	0	1.68	0	0.960604	-23.244396	110	18	1.1	0.9;
	5647	1	6.31	2.5	0	0.81	0	0.976466	-29.43603	110	18	1.1	0.9;
	5648	1	-0	-0	0	115.528858	0	0.990755	-13.733686	380	5	1.1	0.9;
	5649	1	55.7	7.9	0	5.77	0	1.005758	2.727909	154	1	1.1	0.9;
	5650	1	-0	-0	0	18.66	0	1.044978	14.427814	380	3	1.1	0.9;
	5651	1	0	0	0	0	0	0.977947	-31.778345	150	9	1.1	0.9;
	5652	1	45.05	5	0	33.8	0	1.043285	-25.092566	380	11	1.1	0.9;
	5653	1	96.59	7.28	0	2.52	0	1.043716	-7.876751	220	5	1.1	0.9;
	5654	1	-0.14	-1.51	0	1.85	0	1.023998	-39.099754	110	22	1.1	0.9;
	5655	1	-0.21	-1.04	0	0	0	1.018501	2.360099	150	8	1.1	0.9;
	5656	2	0	0	0	0	0	1.039989	-1.865111	220	13	1.1	0.9;
	5657	1	0	0	0	0.58	0	1.062052	-27.004954	220	23	1.1	0.9;
	5658	2	0	0	0	0	0	1.005722	-24.221396	380	5	1.1	0.9;
	5659	1	-27.43	9	0	0.28	0	0.970077	-26.662415	150	9	1.1	0.9;
	5660	1	30.03	9.9	0	0.22	0	1.015958	-36.15525	110	12	1.1	0.9;
	5661	1	0	0	0	0	0	1.008119	-31.216752	150	9	1.1	0.9;
	5662	1	185.7	18.4	0	23.68	0	1.007382	-42.592885	220	10	1.1	0.9;
	5663	1	0	0	0	0	0	1.033921	-29.923351	150	9	1.1	0.9;
	5664	2	0	0	0	0	0	1.063661	0.849588	220	5	1.1	0.9;
	5665	1	-0	-0	0	6.53	0	1.051109	40.203424	380	1	1.1	0.9;
	5666	1	124.8	34.7	0	3.66	0	1.040436	-11.065628	220	5	1.1	0.9;
	5667	1	75.2	10.7	0	3.89	0	1.029457	15.695778	154	1	1.1	0.9;
	5668	2	0	0	0	0	0	1.028943	-42.875803	220	6	1.1	0.9;
	5669	2	0	0	0	0	0	1.000678	-34.61665	220	10	1.1	0.9;
	5670	2	0	0	0	0	0	1.013301	-47.786839	220	4	1.1	0.9;
	5671	1	25.03	1.4	0	0.26	0	0.984988	-28.800852	110	18	1.1	0.9;
	5672	1	139.7	9.7	0	12.61	0	0.979268	-41.802551	220	10	1.1	0.9;
	5673	1	67.27	1.5	0	16.46	0	1.045466	-2.635573	220	3	1.1	0.9;
	5674	1	0	0	0	41.27	0	1.024555	6.468167	380	8	1.1	0.9;
	5675	1	54.96	17.6	0	0.17	0	1.017851	-41.096757	110	22	1.1	0.9;
	5676	1	6.21	2.2	0	0.46	0	1.025006	-25.07857	110	19	1.1	0.9;
	5677	2	0	0	0	0	0	1.033561	-20.889102	220	6	1.1	0.9;
	5678	1	25.13	8.8	0	0.44	0	1.01515	-31.918978	110	21	1.1	0.9;
	5679	2	0	0	0	0	0	1.057325	63.249417	150	1	1.1	0.9;
	5680	1	153.5	46	0	0	0	1.015818	-14.622302	220	3	1.1	0.9;
	5681	1	198.1	33.9	0	22.82	0	1.026281	-53.916536	220	4	1.1	0.9;
	5682	1	-0	0	0	6.1	0	1.026991	1.798275	380	3	1.1	0.9;
	5683	1	-0	-0	0	64.07	0	1.031443	4.244555	400	1	1.1	0.9;
	5684	1	0	-0	0	71.79	0	1.116329	-23.61188	380	21	1.3	0.7;
	5685	1	34.74	16.3	0	33.38	0	1.031682	-43.528573	380	4	1.1	0.9;
	5686	1	52.26	19.6	0	2.64	0	1.031801	-10.245193	220	5	1.1	0.9;
	5687	1	19.12	11.9	0	1.58	0	1.012743	-30.260645	150	9	1.1	0.9;
	5688	1	14.12	5.2	0	1.44	0	1.032263	-8.133455	220	5	1.1	0.9;
	5689	1	41.8	5.9	0	5.81	0	1.011966	5.765357	154	1	1.1	0.9;
	5690	1	0	0	0	0	0	1.066932	-21.855019	220	18	1.1	0.9;
	5691	1	195.1	-14.2	0	12.89	0	1.057263	-45.566306	220	5	1.1	0.9;
	5692	2	0	0	0	0	0	1.048689	-28.313509	380	6	1.1	0.9;
	5693	1	-0	-0	0	6.13	0	1.044671	-16.526474	380	1	1.1	0.9;
	5694	1	40.24	14.4	0	1.52	0	1.004692	-34.806865	220	6	1.1	0.9;
	5695	1	142.8	20.3	0	6.77	0	1.045234	-27.538048	220	5	1.1	0.9;
	5696	1	45.35	28.1	0	1.9	0	1.012996	-27.195344	150	9	1.1	0.9;
	5697	1	131.6	33.8	0	6.55	0	1.007841	-47.751091	220	4	1.1	0.9;
	5698	1	11.81	4.2	0	1.07	0	1.015975	-35.374905	110	1	1.1	0.9;
	5699	1	-0	0	0	0.59	0	1.042214	-6.729499	220	5	1.1	0.9;
	5700	1	31.94	4.4	0	0.29	0	1.007101	-41.693794	110	22	1.1	0.9;
	5701	1	20.9	3	0	5	0	1.015035	14.156324	154	1	1.1	0.9;
	5702	1	-0	-0	0	0.91	0	1.051762	-27.086268	110	3	1.1	0.9;
	5703	1	29.03	18	0	0.02	0	1.041076	-25.457532	150	9	1.1	0.9;
	5704	1	119.6	10.6	0	0.9	0	1.003745	-50.704049	220	6	1.1	0.9;
	5705	2	0	0	0	0	0	1.084017	-47.770008	120	6	1.1	0.9;
	5706	1	32.24	11.7	0	0.85	0	1.049203	-27.104818	110	19	1.1	0.9;
	5707	1	-0	0	0	0.51	0	0.993703	-33.881672	150	9	1.1	0.9;
	5708	1	30.5	4.3	0	2.99	0	1.020915	-7.243486	154	1	1.1	0.9;
	5709	2	0	0	0	0	0	1.024594	-8.738245	380	5	1.1	0.9;
	5710	1	0.4	0.3	0	1.46	0	0.990126	-30.071523	110	18	1.1	0.9;
	5711	1	0	-0	0	32.63	0	1.050262	10.155329	380	3	1.1	0.9;
	5712	1	183.07	44.65	0	2.77	0	1.042355	-4.844771	220	5	1.1	0.9;
	5713	1	20.72	-2.5	0	0.7	0	1.002389	-27.016013	110	18	1.1	0.9;
	5714	2	0	0	0	0	0	1.020678	-35.27281	220	6	1.1	0.9;
	5715	1	89.51	-12.51	0	7.04	0	1.015021	61.817919	150	1	1.1	0.9;
	5716	1	-0	-0	0	21.85	0	1.019593	21.87074	380	3	1.1	0.9;
	5717	2	0	0	0	0	0	1.070074	61.642171	380	3	1.1	0.9;
	5718	1	-0	0	0	5.05	0	0.99498	-38.82174	220	12	1.1	0.9;
	5719	1	-0	-0	0	2.57	0	1.035793	46.935855	220	3	1.1	0.9;
	5720	1	337.6	-15.4	0	6.27	0	1.031595	-25.4469	220	5	1.1	0.9;
	5721	2	0	0	0	0	0	1.054563	-35.198466	110	22	1.1	0.9;
	5722	2	0	0	0	0	0	1.009245	2.026844	154	1	1.1	0.9;
	5723	1	141.2	45.4	0	7.55	0	1.047237	-9.77142	220	5	1.1	0.9;
	5724	1	20.12	8.6	0	0.17	0	0.996765	-27.051875	110	18	1.1	0.9;
	5725	1	19.42	6.3	0	0.48	0	0.977827	-31.426381	110	18	1.1	0.9;
	5726	2	0	0	0	0	0	1.01699	-42.84553	220	4	1.1	0.9;
	5727	1	42.45	14	0	0.44	0	0.977247	-32.621818	110	21	1.1	0.9;
	5728	1	4.61	0.9	0	0.54	0	0.99603	-29.460103	110	18	1.1	0.9;
	5729	1	-0	-0	0	1.67	0	0.963102	-31.646588	150	9	1.1	0.9;
	5730	1	0	0	0	0.61	0	1.056495	-26.549947	110	19	1.1	0.9;
	5731	1	0	0	0	0.8	0	1.05341	-24.417595	220	21	1.1	0.9;
	5732	1	37.14	4.02	0	3.32	0	1.030213	1.277029	220	13	1.1	0.9;
	5733	1	0	0	0	42.8	0	1.071684	-20.328453	220	21	1.1	0.9;
	5734	1	34.9	5	0	5.14	0	1.008175	-25.211628	154	1	1.1	0.9;
	5735	1	161.3	10.04	0	4.38	0	1.052722	-27.276527	220	5	1.1	0.9;
	5736	1	15.32	3.5	0	0.69	0	1.001805	-16.336173	110	18	1.1	0.9;
	5737	2	0	0	0	0	0	1.008529	-26.402638	110	18	1.1	0.9;
	5738	1	-0	-0	0	8.13	0	1.011797	-10.601661	220	5	1.1	0.9;
	5739	1	0	0	0	2.69	0	1.012967	-23.798598	150	9	1.1	0.9;
	5740	1	-0	0	0	1.26	0	1.014887	4.50971	150	8	1.1	0.9;
	5741	1	38.64	13.8	0	0.11	0	1.001487	-26.183348	110	18	1.1	0.9;
	5742	2	0	0	0	0	0	1.023074	25.583399	220	8	1.1	0.9;
	5743	1	-0	0	0	4.28	0	1.047811	-24.447648	220	5	1.1	0.9;
	5744	1	0	0	0	4.96	0	1.026817	15.773204	154	1	1.1	0.9;
	5745	1	13.72	4.5	0	0.16	0	1.015744	-36.169581	110	12	1.1	0.9;
	5746	1	71.98	28.8	0	4.88	0	1.005852	-33.864557	220	6	1.1	0.9;
	5747	1	371.2	73.5	0	39.869841	0	1.023072	-46.933301	380	6	1.1	0.9;
	5748	1	-0	-0	0	4.27	0	1.013586	-4.84178	150	8	1.1	0.9;
	5749	1	-928	-491.4	0	0.73	0	1.073053	12.833517	380	3	1.1	0.9;
	5750	2	0	0	0	0	0	1.026001	-40.550051	110	22	1.1	0.9;
	5751	1	175.6	25	0	20.78	0	1.006834	6.56926	400	1	1.1	0.9;
	5752	1	52.96	50.3	0	0.71	0	1.043572	-30.853587	110	1	1.1	0.9;
	5753	1	-0	-0	0	1.81	0	1.029575	-3.635523	220	5	1.1	0.9;
	5754	1	222.6	62.6	0	30.19	0	1.051896	11.217526	380	3	1.1	0.9;
	5755	1	-0	-0	0	5.45	0	1.03532	-30.60779	150	9	1.1	0.9;
	5756	1	16.7	2.4	0	3.27	0	1.031823	5.570708	154	1	1.1	0.9;
	5757	1	210.7	30	0	116.98	0	1.0135	-21.167181	400	1	1.1	0.9;
	5758	1	0	0	0	9.34	0	1.000204	-34.877053	150	9	1.1	0.9;
	5759	2	0	0	0	0	0	1.075601	-14.315009	380	18	1.1	0.9;
	5760	1	0	-0	0	41.75	0	1.046466	6.968336	150	8	1.1	0.9;
	5761	1	0	0	0	0.49	0	1.066618	-26.339413	150	9	1.1	0.9;
	5762	1	0	0	0	2.23	0	1.045992	0.091901	380	2	1.1	0.9;
	5763	1	-0	0	0	4.95	0	1.037443	-3.388841	220	1	1.1	0.9;
	5764	1	99.4	4.3	0	3.69	0	1.025723	-25.902611	220	5	1.1	0.9;
	5765	1	9.8	1.4	0	0.06	0	1.024999	-4.203502	154	1	1.1	0.9;
	5766	1	176.4	17.2	0	18.37	0	0.988622	-27.421562	220	6	1.1	0.9;
	5767	1	-0	0	0	1.87	0	1.022193	-48.812496	220	4	1.1	0.9;
	5768	1	0	-0	0	22.54	0	1.014636	-3.801727	380	3	1.1	0.9;
	5769	1	-0	0	0	3.23	0	1.014996	-32.518182	150	9	1.1	0.9;
	5770	1	-0	0	0	0.44	0	1.055745	-27.34578	150	9	1.1	0.9;
	5771	1	190.5	70.9	0	21.3	0	0.993072	-47.263595	380	4	1.1	0.9;
	5772	1	-38.01	-22.73	0	0	0	1.04225	4.855848	150	8	1.1	0.9;
	5773	1	0	0	0	0.36	0	1.034676	-3.028522	220	1	1.1	0.9;
	5774	1	0	0	0	8.92	0	1.058859	33.790167	380	3	1.1	0.9;
	5775	1	30.03	10	0	0.26	0	1.002468	-30.954996	110	21	1.1	0.9;
	5776	1	222.7	80.5	0	0	0	1.059236	-7.623477	150	2	1.1	0.9;
	5777	1	0	0	0	0	0	1.011617	-32.734235	150	9	1.1	0.9;
	5778	1	-0	-0	0	2.16	0	0.996449	-32.437195	220	7	1.1	0.9;
	5779	1	86.2	-30.7	0	8.84	0	1.027385	-5.585061	220	3	1.1	0.9;
	5780	1	0	-0	0	0.95	0	1.025535	-53.709471	220	4	1.1	0.9;
	5781	2	0	0	0	0	0	1.046351	-11.820299	380	5	1.1	0.9;
	5782	1	-0	-0	0	19.08	0	1.063656	-34.581513	150	9	1.1	0.9;
	5783	1	69.08	-49.6	0	3.42	0	1.064429	10.944095	110	8	1.1	0.9;
	5784	1	-0	0	0	0.94	0	1.022502	51.362314	220	3	1.1	0.9;
	5785	1	54.2	7.8	0	21.9	0	1.035848	-24.966777	154	1	1.1	0.9;
	5786	2	0	0	0	0	0	1.006382	-44.432452	110	22	1.1	0.9;
	5787	2	0	0	0	0	0	1.028374	-44.07079	220	4	1.1	0.9;
	5788	1	12.6	1.8	0	3.18	0	1.026933	-1.65936	154	1	1.1	0.9;
	5789	1	0	0	0	0.35	0	0.998953	-16.361319	220	5	1.1	0.9;
	5790	1	4.4	1.1	0	0.41	0	1.024154	-36.540708	220	10	1.1	0.9;
	5791	1	0	-0	0	0.85	0	0.987974	-28.298663	110	18	1.1	0.9;
	5792	1	40.95	13.4	0	51.25	0	1.024592	0.113189	150	8	1.1	0.9;
	5793	1	6.11	0	0	1.17	0	1.006923	-48.159604	220	4	1.1	0.9;
	5794	1	0	0	0	4.64	0	1.002506	63.840851	150	1	1.1	0.9;
	5795	1	124.6	15.9	0	2.29	0	1.020039	-24.528292	220	7	1.1	0.9;
	5796	1	0	0	0	0.49	0	0.933765	-34.74988	150	9	1.1	0.9;
	5797	1	334.1	103.5	0	14.78	0	0.977842	-56.716813	220	10	1.1	0.9;
	5798	1	5.04	6.1	0	8.68	0	1.094064	-27.46693	150	9	1.1	0.9;
	5799	1	-0	-0	0	1.38	0	1.040881	-32.583535	220	5	1.1	0.9;
	5800	1	12.6	1.8	0	3.66	0	1.036238	5.081152	154	1	1.1	0.9;
	5801	1	0	0	0	35.66	0	1.129334	-23.470807	380	9	1.3	0.7;
	5802	1	-0	-0	0	0.65	0	1.015581	-32.80905	150	9	1.1	0.9;
	5803	1	0	-0	0	0.35	0	1.039193	-14.333535	220	5	1.1	0.9;
	5804	1	36.24	12.8	0	2.99	0	0.979141	-21.961167	110	19	1.1	0.9;
	5805	1	167.6	-9	0	0.39	0	0.987424	-55.254875	150	10	1.1	0.9;
	5806	1	0	-0	0	0.61	0	0.977739	11.36644	220	3	1.1	0.9;
	5807	1	0	0	0	18.62	0	1.022417	-24.11271	220	7	1.1	0.9;
	5808	1	-7.91	2.6	0	1.72	0	1.036506	-22.038886	150	9	1.1	0.9;
	5809	2	0	0	0	0	0	1.010105	-15.848444	220	6	1.1	0.9;
	5810	1	-0	-0	0	10.45	0	1.046545	10.545931	154	1	1.1	0.9;
	5811	2	0	0	0	0	0	1.018688	-27.872394	110	18	1.1	0.9;
	5812	2	0	0	0	0	0	1.031099	10.404698	220	3	1.1	0.9;
	5813	1	0	0	0	0.14	0	0.969821	-29.796188	110	18	1.1	0.9;
	5814	2	0	0	0	0	0	1.040708	-12.172003	220	5	1.1	0.9;
	5815	1	5.21	1.3	0	24.85	0	1.039004	-51.595201	380	10	1.1	0.9;
	5816	1	13.01	2.4	0	0.42	0	0.993123	-27.286226	110	18	1.1	0.9;
	5817	1	7	1	0	4.33	0	1.029045	-6.970313	154	1	1.1	0.9;
	5818	1	0	0	0	0	0	0.995012	-31.350024	150	9	1.1	0.9;
	5819	1	343.1	109.5	0	41.33	0	1.037139	-17.774505	380	16	1.1	0.9;
	5820	1	-0	-0	0	2.02	0	1.023371	56.767166	220	3	1.1	0.9;
	5821	2	0	0	0	0	0	1.016327	-44.486059	220	4	1.1	0.9;
	5822	2	0	0	0	0	0	1.063749	17.871794	380	13	1.3	0.7;
	5823	2	0	0	0	0	0	1.060066	-39.802079	380	4	1.1	0.9;
	5824	2	0	0	0	0	0	1.050736	-35.4935	380	17	1.1	0.9;
	5825	2	0	0	0	0	0	0.992342	-37.451762	220	10	1.1	0.9;
	5826	1	50.2	7.1	0	7.26	0	1.03378	-7.835541	154	1	1.1	0.9;
	5827	1	83.5	41.5	0	0	0	0.993711	-33.877317	150	9	1.1	0.9;
	5828	2	0	0	0	0	0	1.027766	-37.164045	220	4	1.1	0.9;
	5829	1	155.5	38.3	0	52.84	0	1.01633	-0.962608	150	8	1.1	0.9;
	5830	1	53.86	17.7	0	0.33	0	0.970359	-32.12441	110	21	1.1	0.9;
	5831	2	0	0	0	0	0	1.026046	-47.638826	380	4	1.1	0.9;
	5832	1	-0	0	0	0.86	0	0.896899	-35.666492	150	9	1.3	0.7;
	5833	1	-46.15	15.1	0	0.64	0	1.02486	-22.148298	150	9	1.1	0.9;
	5834	1	0.9	0.3	0	2.19	0	1.033758	-33.943646	110	12	1.1	0.9;
	5835	1	0	0	0	3.32	0	1.014592	16.388587	150	8	1.1	0.9;
	5836	1	244.09	48.92	0	5.96	0	1.034539	-12.333542	220	5	1.1	0.9;
	5837	1	-0	-0	0	132.04744	0	1.015008	-10.384142	380	5	1.1	0.9;
	5838	2	0	0	0	0	0	1.030725	-0.492637	154	1	1.1	0.9;
	5839	2	0	0	0	0	0	1.013483	-48.180115	220	6	1.1	0.9;
	5840	1	0	-0	0	8.78	0	1.019201	-24.016883	154	1	1.1	0.9;
	5841	1	0	0	0	0	0	1.046487	-26.102165	380	6	1.1	0.9;
	5842	1	-0	-0	0	0.19	0	1.006214	-28.428663	110	18	1.1	0.9;
	5843	1	-0	0	0	8.58	0	1.033237	14.707305	380	8	1.1	0.9;
	5844	2	0	0	0	0	0	1.050341	-17.725294	220	7	1.1	0.9;
	5845	1	23.7	3.4	0	2.1	0	1.029227	0.783771	154	1	1.1	0.9;
	5846	2	0	0	0	0	0	1.02116	-5.089899	400	1	1.1	0.9;
	5847	1	0	-0	0	15.97	0	1.045067	-23.65353	220	11	1.1	0.9;
	5848	1	0	-0	0.29	-0.42	0	1.011427	-5.10517	380	2	1.1	0.9;
	5849	1	20.52	7.6	0	2.15	0	0.985881	-28.357949	110	18	1.1	0.9;
	5850	1	-0	0	0	49.904544	0	1.038252	-21.21357	380	20	1.1	0.9;
	5851	1	0.9	0.1	0	8.74	0	1.025529	-4.052572	154	1	1.1	0.9;
	5852	1	0	0	0	0	0	1.011842	61.759219	150	1	1.1	0.9;
	5853	1	122.5	38.9	0	0.84	0	1.041243	-7.53257	220	5	1.1	0.9;
	5854	1	18.42	1.2	0	0.12	0	1.023833	-33.510557	220	6	1.1	0.9;
	5855	2	0	0	0	0	0	1.065097	-8.984355	380	14	1.1	0.9;
	5856	2	0	0	0	0	0	1.036185	-11.799895	220	5	1.1	0.9;
	5857	1	78.39	21.2	0	0.97	0	1.052836	-21.385103	220	5	1.1	0.9;
	5858	1	-0	-0	0	0.2	0	1.01626	-32.346245	150	9	1.1	0.9;
	5859	1	2.2	-0.48	0	3.47	0	1.000233	-32.592963	150	9	1.1	0.9;
	5860	1	18.52	11.5	0	4.41	0	1.034403	-30.610144	150	9	1.1	0.9;
	5861	1	0	-0	0	0.1	0	0.999379	-32.166617	150	9	1.1	0.9;
	5862	2	0	0	0	0	0	1.059582	49.895507	380	3	1.1	0.9;
	5863	2	0	0	0	0	0	1.048267	4.652501	220	13	1.1	0.9;
	5864	2	0	0	0	0	0	1.024889	-4.770654	154	1	1.1	0.9;
	5865	1	20.72	6.8	0	0.65	0	0.978977	-31.241717	110	21	1.1	0.9;
	5866	1	10.01	-30	0	0.74	0	1.024209	-42.872954	220	6	1.1	0.9;
	5867	1	55.26	34.2	0	0.51	0	0.961286	-31.66657	150	9	1.1	0.9;
	5868	2	0	0	0	0	0	1.048327	-28.435905	380	6	1.1	0.9;
	5869	1	145.7	122.2	0	4.5	0	1.021723	-11.257071	220	3	1.1	0.9;
	5870	1	0	-0	0	3.45	0	1.002783	-36.703624	150	9	1.1	0.9;
	5871	1	28.83	-3.8	0	7.4	0	1.030589	-39.461644	110	22	1.1	0.9;
	5872	1	177.5	72.5	0	46.35	0	1.020568	10.85579	150	8	1.1	0.9;
	5873	1	0	0	0	132.84	0	1.012973	-45.813911	380	4	1.1	0.9;
	5874	2	0	0	0	0	0	1.045045	64.333927	150	1	1.1	0.9;
	5875	1	2	1	0	0.3	0	1.002143	-29.086089	110	21	1.1	0.9;
	5876	1	0	-0	0	0.71	0	1.063836	-28.78544	150	9	1.1	0.9;
	5877	1	-0	0	0	67.98	0	1.043576	-50.502884	380	4	1.1	0.9;
	5878	1	-0	0	0	8.17	0	1.043588	-20.621973	220	3	1.1	0.9;
	5879	1	-0	0	0	0.67	0	1.004879	-31.845761	150	9	1.1	0.9;
	5880	1	159.98	76.4	0	22.64	0	1.03245	24.731469	220	3	1.1	0.9;
	5881	2	0	0	0	0	0	1.056126	-10.760666	220	5	1.1	0.9;
	5882	2	0	0	0	0	0	1.077465	8.100026	380	14	1.1	0.9;
	5883	2	0	0	0	0	0	1.175144	-19.224951	380	9	1.3	0.7;
	5884	1	-0	-0	0	68.73	0	1.036964	-40.132357	380	4	1.1	0.9;
	5885	2	0	0	0	0	0	1.031733	-20.996223	220	6	1.1	0.9;
	5886	1	16.42	5.3	0	0.43	0	1.005533	-24.819539	110	18	1.1	0.9;
	5887	1	6.91	-11.1	0	0.67	0	1.001678	21.061751	110	8	1.1	0.9;
	5888	1	-0	-0	0	93.07	0	1.05046	16.816424	380	8	1.1	0.9;
	5889	1	27.9	4	0	3.43	0	1.037146	9.801924	154	1	1.1	0.9;
	5890	1	0	-0	0	3.66	0	1.027395	-46.396757	220	4	1.1	0.9;
	5891	1	14.02	5.5	0	3.3	0	1.031513	-12.447025	220	5	1.1	0.9;
	5892	1	49.86	14.1	0	11.33	0	1.022362	-48.79846	220	4	1.1	0.9;
	5893	1	26.03	10.6	0	2.14	0	1.023505	15.322267	110	8	1.1	0.9;
	5894	2	0	0	0	0	0	1.035315	-5.512525	220	3	1.1	0.9;
	5895	1	0	0	0	0	0	0.995326	-35.195598	150	9	1.1	0.9;
	5896	2	0	0	0	0	0	0.986344	14.700502	220	3	1.1	0.9;
	5897	1	0	0	0	0	0	1.047643	17.999988	380	8	1.1	0.9;
	5898	1	-85.69	-2	0	21.28	0	1.054015	-14.917486	220	7	1.1	0.9;
	5899	1	83.6	11.9	0	3.51	0	1.013511	-20.047796	154	1	1.1	0.9;
	5900	1	0	0	0	0	0	1.085303	-27.976492	150	9	1.1	0.9;
	5901	1	7.21	3.3	0	0.91	0	1.028065	-26.84935	110	15	1.1	0.9;
	5902	1	-0	0	0	0.19	0	0.998746	-30.611628	150	9	1.1	0.9;
	5903	1	0	-7.2	0	0	0	1.034736	-28.713871	150	9	1.1	0.9;
	5904	1	64.4	9.1	0	2.29	0	1.024114	-7.021107	154	1	1.1	0.9;
	5905	1	196.3	-10.4	0	8.97	0	1.031661	-13.65715	220	7	1.1	0.9;
	5906	1	5.31	2.3	0	0.93	0	0.969449	-29.110513	110	18	1.1	0.9;
	5907	1	186.1	14.8	0	17.28	0	1.042618	-8.712204	220	5	1.1	0.9;
	5908	1	10.51	1.3	0	0.18	0	0.99477	-31.893947	110	18	1.1	0.9;
	5909	1	181.9	92.6	0	4.15	0	0.99022	-20.158549	220	3	1.1	0.9;
	5910	1	42.55	14	0	7.52	0	1.023544	-38.299118	110	12	1.1	0.9;
	5911	1	-0	-0	0	9.84	0	1.03698	-25.21858	220	3	1.1	0.9;
	5912	1	-0	-0	0	0.89	0	1.030072	-20.401195	220	1	1.1	0.9;
	5913	1	80.99	30.1	0	29.8	0	1.018238	-29.270452	220	6	1.1	0.9;
	5914	1	94.1	39	0	27.61	0	1.036238	1.780853	380	3	1.1	0.9;
	5915	1	-0	0	0	4.34	0	1.00769	-30.933315	150	9	1.1	0.9;
	5916	1	-57.59	-30.52	0	2.45	0	1.03926	17.310915	110	8	1.1	0.9;
	5917	1	61.67	17.4	0	0.29	0	1.03057	-26.483209	110	19	1.1	0.9;
	5918	1	73.83	16.41	0	15.77	0	1.054009	-26.268399	220	5	1.1	0.9;
	5919	1	-0	0	0	33.03	0	1.041114	-50.368417	380	4	1.1	0.9;
	5920	1	0	0	0	30.48	0	1.057048	-20.065958	400	1	1.1	0.9;
	5921	1	-0	-0	0	34.98	0	1.009	6.925417	400	1	1.1	0.9;
	5922	2	0	0	0	0	0	1.030869	-38.497457	110	22	1.1	0.9;
	5923	1	74.08	10.7	0	13.55	0	1.023696	-54.459398	220	4	1.1	0.9;
	5924	1	29.03	18	0	0.02	0	1.004888	-32.9331	150	9	1.1	0.9;
	5925	1	0	0	0	0	0	0.97801	-29.767009	150	9	1.1	0.9;
	5926	1	0	-0	0	3.98	0	1.068631	-48.196248	220	5	1.1	0.9;
	5927	1	-0	0	0	1.44	0	0.991444	-36.806588	150	10	1.1	0.9;
	5928	1	-0	0	0	1.39	0	1.017195	-28.972732	150	9	1.1	0.9;
	5929	1	0	0	0	1.74	0	1.045842	63.192896	220	3	1.1	0.9;
	5930	2	0	0	0	0	0	1.042525	-31.111521	220	12	1.1	0.9;
	5931	1	53.06	17.4	0	0.11	0	1.00929	-30.089314	110	21	1.1	0.9;
	5932	1	22.3	3.1	0	0.54	0	1.03346	-7.875888	154	1	1.1	0.9;
	5933	1	78.49	18.5	0	6.91	0	1.005708	-48.889663	220	4	1.1	0.9;
	5934	1	0	0	0	37.02	0	1.003321	-5.390135	400	1	1.1	0.9;
	5935	1	-0	0	0	5.75	0	1.06545	-22.573477	220	5	1.1	0.9;
	5936	1	0	0	0	0.38	0	1.01939	-22.655348	150	9	1.1	0.9;
	5937	1	-0	0	0	0	0	1.044953	-33.486934	150	9	1.1	0.9;
	5938	1	-0	-0	0	15.14	0	1.056395	-34.749164	380	1	1.1	0.9;
	5939	1	0	-0	0	0.65	0	1.069977	-26.579592	150	9	1.1	0.9;
	5940	2	0	0	0	0	0	1.026879	-14.638592	220	5	1.1	0.9;
	5941	2	0	0	0	0	0	1.051545	49.790477	380	3	1.1	0.9;
	5942	1	5.21	1.7	0	1.37	0	1.038025	-33.20709	110	12	1.1	0.9;
	5943	1	15.12	7.5	0	0.33	0	1.006106	11.425005	220	3	1.1	0.9;
	5944	1	71.88	21.7	0	9.63	0	1.052506	-26.232886	220	5	1.1	0.9;
	5945	1	0	-0	0	28.22	0	1.044227	-48.841113	220	6	1.1	0.9;
	5946	1	0	-0	0	5.96	0	1.021262	-42.836601	220	4	1.1	0.9;
	5947	1	0	0	0	1.4	0	1.053714	-22.746881	150	9	1.1	0.9;
	5948	1	57.06	13.3	0	19.18	0	1.021508	-48.968975	220	4	1.1	0.9;
	5949	1	160.7	40.3	0	3.04	0	1.018073	2.35558	150	8	1.1	0.9;
	5950	1	24.93	22.9	0	2.47	0	0.998984	-30.306116	150	9	1.1	0.9;
	5951	1	-0	-0	0	0.55	0	1.020223	-30.568507	150	9	1.1	0.9;
	5952	1	40.34	15.9	0	1.31	0	1.031206	3.839057	150	8	1.1	0.9;
	5953	1	-7.97	6.3	0	2.73	0	0.991065	-34.243743	150	9	1.1	0.9;
	5954	1	106.7	33.6	0	3.28	0	1.048336	-18.628039	110	19	1.1	0.9;
	5955	2	0	0	0	0	0	1.015926	-15.025291	110	18	1.1	0.9;
	5956	1	0	0	0	0.18	0	1.036355	-29.780714	150	9	1.1	0.9;
	5957	1	208.06	36.6	0	3.55	0	1.021081	-16.562596	220	5	1.1	0.9;
	5958	1	209.8	102.6	0	24.31	0	0.993945	-22.313337	220	3	1.1	0.9;
	5959	1	160	50	0	23.19	0	1.01214	-36.645736	220	6	1.1	0.9;
	5960	1	-0	0	0	2.78	0	1.016922	-33.838231	220	1	1.1	0.9;
	5961	1	113.5	11.5	0	5.53	0	1.013483	-46.991416	220	4	1.1	0.9;
	5962	1	64.27	22.9	0	3.06	0	1.043271	-24.414808	110	19	1.1	0.9;
	5963	1	0	-0	0	11.87	0	1.071359	44.333413	380	3	1.1	0.9;
	5964	2	0	0	0	0	0	1.005428	-24.106443	110	18	1.1	0.9;
	5965	1	144.5	67.8	0	2.43	0	0.997579	8.680327	220	3	1.1	0.9;
	5966	1	62.57	22.7	0	0.47	0	1.021894	-28.446813	110	19	1.1	0.9;
	5967	1	-0	-0	0	0.86	0	1.030975	-25.934721	150	9	1.1	0.9;
	5968	1	-0	-0	0	0.15	0	1.047043	-24.829942	150	9	1.1	0.9;
	5969	1	99.5	28.09	0	6.23	0	1.036469	13.002506	380	3	1.1	0.9;
	5970	2	0	0	0	0	0	1.017378	-41.457427	220	4	1.1	0.9;
	5971	2	0	0	0	0	0	1.076888	4.711259	380	5	1.3	0.7;
	5972	1	148.2	65.3	0	10.51	0	1.017198	-16.166595	380	3	1.1	0.9;
	5973	2	0	0	0	0	0	1.055332	-13.19785	220	7	1.1	0.9;
	5974	1	0	0	0	0	0	1.023351	11.178136	150	8	1.1	0.9;
	5975	2	0	0	0	0	0	1.0347	-33.956308	110	12	1.1	0.9;
	5976	1	0	-0	0	6.35	0	1.045419	-27.707756	220	23	1.1	0.9;
	5977	2	0	0	0	0	0	1.030489	67.261675	380	1	1.1	0.9;
	5978	1	3.5	0.9	0	1.12	0	1.000781	-28.914997	110	18	1.1	0.9;
	5979	1	-0	-0	0	0	0	1.040605	25.720506	220	8	1.1	0.9;
	5980	2	0	0	0	0	0	1.05163	-24.648119	150	9	1.1	0.9;
	5981	1	17.12	1.3	0	0.99	0	0.99111	-27.640402	110	18	1.1	0.9;
	5982	1	0	0	0	0.29	0	1.025239	-34.309954	380	6	1.1	0.9;
	5983	2	0	0	0	0	0	1.061878	-15.173411	380	5	1.1	0.9;
	5984	1	27.53	12.1	0	0.88	0	1.018485	-34.915936	110	22	1.1	0.9;
	5985	2	0	0	0	0	0	1.120491	-23.13795	380	21	1.3	0.7;
	5986	2	0	0	0	0	0	1.103376	32.82865	380	3	1.3	0.7;
	5987	1	24.04	1.68	0	4.02	0	1.03492	-14.153007	220	5	1.1	0.9;
	5988	2	0	0	0	0	0	1.069038	6.568109	380	14	1.1	0.9;
	5989	1	83.1	26.1	0	23.95	0	1.050893	-27.991284	220	23	1.1	0.9;
	5990	1	0	-0	0	4.53	0	1.02099	-10.272683	220	5	1.1	0.9;
	5991	2	0	0	0	0	0	1.012716	-50.368894	220	4	1.1	0.9;
	5992	1	408.9	77.3	0	13.58	0	1.019913	-25.506468	220	7	1.1	0.9;
	5993	1	0	-0	0	2.75	0	1.038533	-11.276657	220	5	1.1	0.9;
	5994	2	0	0	0	0	0	1.036793	-15.198385	220	5	1.1	0.9;
	5995	1	18.52	6.1	0	0.09	0	1.005109	-30.74235	110	21	1.1	0.9;
	5996	1	-0	-0	0.164188	4.023827	0	1.045944	0.092131	380	2	1.1	0.9;
	5997	1	25.83	12.5	0	0.2	0	1.039308	-15.177815	220	6	1.1	0.9;
	5998	2	0	0	0	0	0	1.02348	6.16016	154	1	1.1	0.9;
	5999	1	36.04	-8	0	32.73	0	1.028607	-33.173086	220	11	1.1	0.9;
	6000	2	0	0	0	0	0	1.085353	16.420992	380	8	1.3	0.7;
	6001	2	0	0	0	0	0	1.036406	12.673591	154	1	1.1	0.9;
	6002	1	4.61	1.7	0	0.9	0	1.028604	-25.606622	110	19	1.1	0.9;
	6003	1	7.01	2.3	0	0.61	0	1.011541	-36.739082	110	12	1.1	0.9;
	6004	1	0	0	0	2.37	0	1.059483	-47.950752	380	10	1.1	0.9;
	6005	1	0	-0	0	50.27	0	1.102227	-20.658022	380	9	1.3	0.7;
	6006	1	0	0	0.25	20.36	0	1.061051	-25.469497	220	19	1.1	0.9;
	6007	1	191.96	38.57	0	122.89	0	1.064324	-9.14907	380	14	1.1	0.9;
	6008	1	0	-0	0	219.65	0	1.009932	11.192575	400	1	1.1	0.9;
	6009	1	20.82	7.9	0	6.73	0	1.008404	-51.715229	220	4	1.1	0.9;
	6010	1	0	0	0	1.35	0	1.074075	-23.724285	220	5	1.1	0.9;
	6011	1	11.41	0.7	0	0.24	0	1.003207	-18.17065	110	18	1.1	0.9;
	6012	1	-8.81	2.9	0	0.65	0	0.948971	-29.211036	150	9	1.1	0.9;
	6013	1	105.8	51.8	0	5.31	0	0.998514	-36.87879	150	9	1.1	0.9;
	6014	1	0.04	-0.05	0	20.75	0	1.050964	-1.994212	220	14	1.1	0.9;
	6015	1	-0	0	0	25	0	1.028227	-1.865839	150	8	1.1	0.9;
	6016	1	92.1	45	0	0.4	0	0.973172	-53.387565	220	6	1.1	0.9;
	6017	1	-0	0	0	23.15	0	1.025236	11.354177	380	8	1.1	0.9;
	6018	1	27.53	17	0	0.86	0	1.00266	-30.532747	150	9	1.1	0.9;
	6019	1	0	0	0	10.52	0	1.040028	-15.624439	380	1	1.1	0.9;
	6020	1	0.1	-19.3	0	0.27	0	1.025616	-22.276632	220	3	1.1	0.9;
	6021	1	-0	0	0	76.15	0	1.000863	-2.156852	400	1	1.1	0.9;
	6022	1	-0	-0	0	5.5	0	0.99945	-38.380058	220	4	1.1	0.9;
	6023	1	9.01	3	0	0.51	0	1.045064	-27.645735	110	23	1.1	0.9;
	6024	1	-0	-0	0	0.37	0	0.991089	-30.78687	150	9	1.1	0.9;
	6025	1	-0.16	-1.92	0	0.45	0	1.027402	14.331794	110	8	1.1	0.9;
	6026	1	-0	-0	0	0.05	0	1.044025	-29.510191	150	9	1.1	0.9;
	6027	1	61.27	11	0	50.16	0	1.028697	18.809566	380	3	1.1	0.9;
	6028	1	0	0	0	3.24	0	1.028186	7.910399	110	8	1.1	0.9;
	6029	1	23.03	7.6	0	0.78	0	0.99547	-39.22732	110	12	1.1	0.9;
	6030	1	48.8	6.9	0	12.07	0	1.017533	10.839594	154	1	1.1	0.9;
	6031	1	0	0	0	2.22	0	1.044676	-15.800131	220	5	1.1	0.9;
	6032	1	-0	0	0	2.06	0	0.980506	-30.314291	150	9	1.1	0.9;
	6033	2	0	0	0	0	0	1.013332	-25.393443	110	19	1.1	0.9;
	6034	1	0	-0	0	0.19	0	1.03183	-26.239987	110	21	1.1	0.9;
	6035	1	-18.52	6.1	0	0.26	0	1.01559	-28.206303	150	9	1.1	0.9;
	6036	2	0	0	0	0	0	1.036335	-1.030445	380	5	1.1	0.9;
	6037	1	21.62	9.8	0	0.14	0	0.996493	-26.971273	110	18	1.1	0.9;
	6038	1	43.55	-20.5	0	2.53	0	1.040399	-45.85497	220	4	1.1	0.9;
	6039	1	165.9	154.6	0	15.17	0	1.037599	-51.305829	380	4	1.1	0.9;
	6040	1	-57.46	-30.3	0	5.4	0	1.02391	-45.47508	220	4	1.1	0.9;
	6041	1	0	0	0	0.53	0	1.068673	-48.196524	220	5	1.1	0.9;
	6042	1	-0	-0	0	0.48	0	0.935862	-34.589217	150	9	1.1	0.9;
	6043	1	49.76	15	0	1.22	0	0.966175	-33.849002	110	21	1.1	0.9;
	6044	1	117.5	16.7	0	3.35	0	1.033573	-5.804134	154	1	1.1	0.9;
	6045	1	-0.2	-0.98	0	0	0	1.017016	-0.831864	150	8	1.1	0.9;
	6046	2	0	0	0	0	0	1.094629	-11.855334	220	19	1.3	0.7;
	6047	1	79.49	22.7	0	9.26	0	1.020029	8.412898	150	8	1.1	0.9;
	6048	1	6.61	1.9	0	0.67	0	0.976107	-30.206932	110	18	1.1	0.9;
	6049	1	-0	-0	0	3.23	0	1.004227	-8.419176	220	3	1.1	0.9;
	6050	1	0	0	0	42.02	0	1.014473	9.704336	400	1	1.1	0.9;
	6051	1	5.61	2	0	0.96	0	0.955577	-19.048184	110	19	1.1	0.9;
	6052	1	-0	-0	0	0	0	1.017309	18.203559	150	8	1.1	0.9;
	6053	1	412.1	79.3	0	7.73	0	1.050641	-26.961918	220	5	1.1	0.9;
	6054	1	4.1	1.4	0	0.4	0	1.040191	-33.020646	110	12	1.1	0.9;
	6055	1	19.32	7	0	0.39	0	1.032882	-6.143066	110	19	1.1	0.9;
	6056	1	0	0	0	1.43	0	1.058644	-28.636542	150	9	1.1	0.9;
	6057	1	47.65	19.4	0	0.79	0	1.022376	-1.911451	150	8	1.1	0.9;
	6058	1	0	-0	0	1.73	0	1.037048	8.948628	110	8	1.1	0.9;
	6059	1	35.14	14.3	0	7.31	0	1.030759	-39.279516	110	22	1.1	0.9;
	6060	1	0	0	0	80.66	0	1.068745	-38.373468	380	1	1.1	0.9;
	6061	1	15.02	9	0	0.01	0	1.003609	-32.005398	150	9	1.1	0.9;
	6062	1	13.31	2.8	0	0.88	0	1.001227	-26.614531	110	18	1.1	0.9;
	6063	1	61.36	23.66	0	0.16	0	1.001285	-27.675348	110	18	1.1	0.9;
	6064	1	-0	0	0	14.03	0	1.1702	-20.509705	380	1	1.3	0.7;
	6065	1	25.73	9.3	0	0.59	0	1.027429	-20.311759	110	19	1.1	0.9;
	6066	1	0	0	0	30.94	0	1.020695	-12.866151	380	3	1.1	0.9;
	6067	1	-0	-0	0	10.33	0	1.020339	-36.103606	220	6	1.1	0.9;
	6068	1	0	0	0	33.08	0	1.032737	12.679876	380	8	1.1	0.9;
	6069	1	-0	-0	0.285753	52.671068	0	0.982629	-6.460921	380	2	1.1	0.9;
	6070	1	-14.58	-5.96	0	4.76	0	1.02244	-23.638649	220	6	1.1	0.9;
	6071	1	163.5	35.6	0	4.59	0	1.038766	-16.393907	220	5	1.1	0.9;
	6072	1	47.6	41.3	0	295.019105	0	1.139986	-21.685102	380	9	1.3	0.7;
	6073	1	27.9	4	0	1.75	0	1.008579	-7.84577	154	1	1.1	0.9;
	6074	1	2.7	0.9	0	1.49	0	1.017385	-35.961352	110	12	1.1	0.9;
	6075	1	0	-0	0	4.17	0	1.048897	-14.232788	220	1	1.1	0.9;
	6076	1	1.8	0.7	0	0.32	0	1.080309	-17.705586	110	19	1.1	0.9;
	6077	1	-0	0	0	-1.896325	0	1.075401	-17.159039	380	1	1.1	0.9;
	6078	1	0	-0	0	1.25	0	0.982173	-27.299078	150	9	1.1	0.9;
	6079	1	160.2	89.89	0	0	0	1.01909	-22.656195	150	9	1.1	0.9;
	6080	1	0	0	0.361	33.735	0	1.053137	-42.75542	380	10	1.1	0.9;
	6081	1	5.81	1.4	0	0.2	0	0.987882	-26.43586	110	18	1.1	0.9;
	6082	1	6.01	1.5	0	0.73	0	0.985788	-28.136945	110	18	1.1	0.9;
	6083	1	111.5	147.5	0	26.54	0	1.046828	11.879288	380	3	1.1	0.9;
	6084	1	-0	-0	0.072589	52.192656	0	1.01587	66.119603	380	1	1.1	0.9;
	6085	1	-0	0	0	22.6	0	1.004991	14.361891	400	1	1.1	0.9;
	6086	2	0	0	0	0	0	1.017402	2.4054	154	1	1.1	0.9;
	6087	1	27.9	4	0	3.97	0	1.013589	-2.829796	154	1	1.1	0.9;
	6088	2	0	0	0	0	0	1.069764	4.834103	380	14	1.1	0.9;
	6089	1	26.23	8.6	0	6.14	0	1.026807	-34.16069	110	12	1.1	0.9;
	6090	2	0	0	0	0	0	1.030088	-39.174845	220	4	1.1	0.9;
	6091	1	-0	-0	0	17.336728	0	1.013386	-11.921046	220	24	1.1	0.9;
	6092	2	0	0	0	0	0	1.035378	-39.920989	380	4	1.1	0.9;
	6093	1	21.02	12.4	0	2.96	0	1.005956	-41.042922	220	4	1.1	0.9;
	6094	1	0	0	0	0.45	0	1.023585	-29.739277	150	9	1.1	0.9;
	6095	1	0	-0	0	10.05	0	1.050059	49.096409	380	3	1.1	0.9;
	6096	1	0	0	0	0	0	1.036804	-25.936777	150	9	1.1	0.9;
	6097	1	-0	0	0	3.87	0	1.02081	16.80975	380	3	1.1	0.9;
	6098	1	-0	-0	0.135493	28.29062	0	1.112778	-24.13176	380	21	1.3	0.7;
	6099	1	0.3	2	0	10.34	0	1.045983	-1.585831	220	14	1.1	0.9;
	6100	2	0	0	0	0	0	1.041341	0.964908	220	13	1.1	0.9;
	6101	1	103	40.2	0	1.13	0	1.02371	-10.937212	220	5	1.1	0.9;
	6102	1	-0	0	0	0.14	0	1.022422	-29.626211	150	9	1.1	0.9;
	6103	2	0	0	0	0	0	1.013429	-21.718838	220	20	1.1	0.9;
	6104	1	93.12	4.99	0	0.18	0	1.046851	-12.433827	220	5	1.1	0.9;
	6105	2	0	0	0	0	0	1.010683	-27.426254	110	18	1.1	0.9;
	6106	1	1.3	3.8	0	0.09	0	0.95515	-30.275609	110	18	1.1	0.9;
	6107	1	128.6	27.8	0	36.2	0	1.011511	-36.514878	220	6	1.1	0.9;
	6108	1	-0	-0	0	24.64	0	1.038751	34.137872	380	3	1.1	0.9;
	6109	1	176.75	51.7	0	39.79	0	1.007365	-10.436065	380	13	1.1	0.9;
	6110	1	78.59	16	0	1.87	0	1.052198	-36.243599	220	5	1.1	0.9;
	6111	1	13.9	1.9	0	0.54	0	1.024669	-15.702901	154	1	1.1	0.9;
	6112	1	45.28	15.27	0	7.88	0	1.022264	-9.359412	220	5	1.1	0.9;
	6113	1	0	-0	0	0.05	0	0.964579	-31.523448	150	9	1.1	0.9;
	6114	1	0	0	0	8.78	0	1.042632	-10.600548	220	5	1.1	0.9;
	6115	1	-0	-0	0	1.18	0	1.0225	-12.390675	220	5	1.1	0.9;
	6116	1	0	-0	0.118007	-0.299556	0	1.050251	-17.936656	220	18	1.1	0.9;
	6117	1	0	-0	0	1.79	0	0.975165	-30.425695	150	9	1.1	0.9;
	6118	1	82.55	30.58	0	4.26	0	1.022324	25.507998	220	8	1.1	0.9;
	6119	1	221.25	72.56	0	8.23	0	1.036247	-16.100323	220	5	1.1	0.9;
	6120	2	0	0	0	0	0	1.024222	20.450903	220	3	1.1	0.9;
	6121	1	16.12	5.8	0	0.91	0	1.079735	-22.080279	150	9	1.1	0.9;
	6122	1	3.7	1.4	0	0.63	0	1.062638	-21.930056	110	15	1.1	0.9;
	6123	1	55.8	7.9	0	11.5	0	1.014472	-14.458656	154	1	1.1	0.9;
	6124	2	0	0	0	0	0	1.025031	13.594718	400	1	1.1	0.9;
	6125	1	0	-0	0	1.47	0	0.982201	-27.085684	150	9	1.1	0.9;
	6126	1	118.4	16.9	0	0.02	0	1.011893	-25.430103	154	1	1.1	0.9;
	6127	1	-0.1	68.5	0	17.24	0	1.046216	-47.494042	220	4	1.1	0.9;
	6128	1	27.23	8.9	0	0.72	0	0.989931	-36.870947	110	12	1.1	0.9;
	6129	2	0	0	0	0	0	1.033744	-11.132083	380	6	1.1	0.9;
	6130	2	0	0	0	0	0	1.026444	38.067642	220	3	1.1	0.9;
	6131	2	0	0	0	0	0	1.133869	18.740277	150	8	1.3	0.7;
	6132	1	60.67	5.9	0	5.32	0	1.007513	-48.230416	220	4	1.1	0.9;
	6133	1	109	28.3	0	0.55	0	1.0405	-25.928997	110	19	1.1	0.9;
	6134	1	42.55	6.5	0	3.7	0	1.032343	-53.498271	220	4	1.1	0.9;
	6135	1	2.2	0.5	0	0.5	0	0.964329	-56.362142	150	10	1.1	0.9;
	6136	1	58.8	8.4	0	5.06	0	1.016978	5.6322	154	1	1.1	0.9;
	6137	1	-0	-0	0	37.02	0	1.036527	43.984721	380	3	1.1	0.9;
	6138	2	0	0	0	0	0	1.010882	-44.602993	220	4	1.1	0.9;
	6139	1	76.89	35.1	0	1.9	0	1.058135	0.578948	150	8	1.1	0.9;
	6140	1	11.51	3.8	0	7.14	0	1.027477	-38.211715	110	12	1.1	0.9;
	6141	1	44.05	16	0	1.63	0	1.004979	-28.578928	110	21	1.1	0.9;
	6142	1	116.2	16.5	0	2.4	0	1.009131	1.535376	154	1	1.1	0.9;
	6143	1	39	5.5	0	7.15	0	1.010717	-8.822062	154	1	1.1	0.9;
	6144	2	0	0	0	0	0	0.987224	-29.149765	110	18	1.1	0.9;
	6145	2	0	0	0	0	0	1.019556	-33.283906	220	4	1.1	0.9;
	6146	1	0	-0	0	3.15	0	1.025358	-12.3051	220	5	1.1	0.9;
	6147	1	-0	0	0	0.61	0	1.018949	-29.407907	150	9	1.1	0.9;
	6148	1	8.71	5.4	0	0.84	0	0.966862	-34.40203	150	9	1.1	0.9;
	6149	1	39.05	24.8	0	7.47	0	1.014375	16.388893	150	8	1.1	0.9;
	6150	1	15.32	5	0	0.76	0	1.015654	-23.507465	110	21	1.1	0.9;
	6151	1	60.67	33.2	0	0.31	0	1.027467	-24.374708	220	5	1.1	0.9;
	6152	1	-0	-0	0	2.57	0	1.014727	53.934674	220	3	1.1	0.9;
	6153	2	0	0	0	0	0	1.019885	-14.247746	380	5	1.1	0.9;
	6154	1	0	0	0	0.67	0	1.054274	-29.79948	220	1	1.1	0.9;
	6155	1	67.98	54.6	0	3.39	0	0.970396	27.73135	220	3	1.1	0.9;
	6156	1	95.9	26.2	0	16.21	0	1.008626	-37.016487	220	6	1.1	0.9;
	6157	1	160.2	88.7	0	0	0	1.014522	-22.608933	150	9	1.1	0.9;
	6158	1	-0	-0	0.25	2.3	0	1.088023	-14.698335	220	19	1.3	0.7;
	6159	1	0	-0	0.173089	4.478398	0	1.023772	-24.941987	220	3	1.1	0.9;
	6160	1	-0	-0	0	16.34	0	1.034987	8.210955	380	3	1.1	0.9;
	6161	1	383.6	-37.9	0	21.3	0	1.023556	-46.171565	380	6	1.1	0.9;
	6162	1	0	0	0	0	0	0.991401	-30.700734	150	9	1.1	0.9;
	6163	1	0	-0	0	4.96	0	1.031507	-22.306281	220	5	1.1	0.9;
	6164	1	-0	-0	0.215903	26.017449	0	1.063268	12.865238	380	3	1.1	0.9;
	6165	1	-0	-0	0	0.32	0	0.988879	-31.192112	150	9	1.1	0.9;
	6166	1	67.28	-0.33	0	8.4	0	1.02932	-6.96665	220	13	1.1	0.9;
	6167	1	152.6	61.8	0	2.35	0	1.015226	10.355031	150	8	1.1	0.9;
	6168	2	0	0	0	0	0	1.047693	-13.227008	220	5	1.1	0.9;
	6169	2	0	0	0	0	0	1.033616	8.6214	220	13	1.1	0.9;
	6170	1	47.3	6.7	0	2.54	0	0.997407	-9.164303	154	1	1.1	0.9;
	6171	1	28.93	0.2	0	0	0	1.017535	20.66908	220	3	1.1	0.9;
	6172	1	66.97	14.1	0	5.58	0	0.986605	-23.409457	220	3	1.1	0.9;
	6173	1	-6.3	-3.18	0	7.28	0	1.006283	37.079999	220	3	1.1	0.9;
	6174	1	289.1	84.2	0	17.34	0	1.005058	-48.558234	220	4	1.1	0.9;
	6175	2	0	0	0	0	0	1.023886	-8.712734	154	1	1.1	0.9;
	6176	1	16.72	5.5	0	0.65	0	1.021694	-36.758293	110	12	1.1	0.9;
	6177	1	18.42	5.3	0	0.39	0	0.996746	-26.991818	110	18	1.1	0.9;
	6178	1	91	28.4	0	10.57	0	1.037814	-24.930251	220	5	1.1	0.9;
	6179	2	0	0	0	0	0	1.036613	-39.795448	220	4	1.1	0.9;
	6180	1	0	-0	0	20.11	0	1.040132	12.738566	380	3	1.1	0.9;
	6181	2	0	0	0	0	0	1.000322	-0.775253	400	1	1.1	0.9;
	6182	2	0	0	0	0	0	1.054529	-29.782644	220	22	1.1	0.9;
	6183	1	59.37	12.7	0	22.18	0	1.027664	-53.950037	220	4	1.1	0.9;
	6184	1	16.72	5.5	0	0.89	0	1.005908	-39.513914	110	12	1.1	0.9;
	6185	1	-0	-0	0	1.01	0	1.016349	-28.570554	150	9	1.1	0.9;
	6186	1	63.07	13.4	0	15.52	0	1.027064	-54.149347	220	4	1.1	0.9;
	6187	1	-0	0	0.361	25.935	0	1.051127	-41.658601	380	10	1.1	0.9;
	6188	1	209.9	38	0	5.26	0	0.980312	-51.312007	150	10	1.1	0.9;
	6189	1	131.9	50.6	0	79.02	0	1.0306	-41.007869	380	6	1.1	0.9;
	6190	1	141.52	50.48	0	66.28	0	1.044175	-12.838162	380	13	1.1	0.9;
	6191	1	-2.91	-0.62	0	0.01	0	1.010183	-36.984825	220	6	1.1	0.9;
	6192	1	50.56	5.9	0	10.58	0	1.023461	-43.410715	220	4	1.1	0.9;
	6193	1	165.4	130.5	0	28.06	0	1.008205	-11.819576	380	3	1.1	0.9;
	6194	1	0	-0	0	3.14	0	1.023574	-16.503984	220	5	1.1	0.9;
	6195	1	0	0	0	0.28	0	1.015711	-26.359994	150	6	1.1	0.9;
	6196	1	17.52	6.1	0	0.53	0	1.048297	-17.424548	110	19	1.1	0.9;
	6197	1	126.7	46.1	0	5.58	0	1.021877	-14.540572	220	3	1.1	0.9;
	6198	1	154.2	22	0	12.1	0	1.022414	-25.277559	154	1	1.1	0.9;
	6199	1	-4.82	-0.88	0	224.85	0	1.069317	-12.830557	380	5	1.1	0.9;
	6200	1	0	-0	0	0.15	0	0.970957	-24.095442	150	9	1.1	0.9;
	6201	2	0	0	0	0	0	1.049209	-30.350265	380	6	1.1	0.9;
	6202	1	83.79	33.1	0	65.16	0	1.043376	3.446178	150	8	1.1	0.9;
	6203	1	14.02	5.6	0	2.34	0	1.067351	-27.64379	220	5	1.1	0.9;
	6204	2	0	0	0	0	0	1.02934	-21.276092	220	20	1.1	0.9;
	6205	1	-0	0	0	1.77	0	1.013865	-22.844663	110	18	1.1	0.9;
	6206	1	-7.62	-2.52	0	33.197994	0	1.035075	-0.012484	380	5	1.1	0.9;
	6207	2	0	0	0	0	0	1.03672	-13.020576	380	6	1.1	0.9;
	6208	1	18.72	7.9	0	1.67	0	1.006088	-27.425904	110	18	1.1	0.9;
	6209	1	69.48	19.7	0	22.1	0	1.021564	-48.828688	220	4	1.1	0.9;
	6210	1	253.2	78.1	0	10.47	0	1.004535	-22.995135	220	6	1.1	0.9;
	6211	1	48.8	6.9	0	3.06	0	1.012162	-10.026558	154	1	1.1	0.9;
	6212	1	117	19.2	0	5.38	0	1.022578	-5.680793	220	3	1.1	0.9;
	6213	1	-0	-0	0	2.49	0	0.999721	-27.058885	150	9	1.1	0.9;
	6214	1	7.61	2.5	0	0.57	0	1.010006	-35.666762	110	1	1.1	0.9;
	6215	1	-0	0	0	2.18	0	1.014717	-32.253715	150	9	1.1	0.9;
	6216	1	13.72	4.5	0	0.91	0	1.005661	-36.701716	110	12	1.1	0.9;
	6217	1	-0	-0	0	3.88	0	1.033582	-44.088027	380	6	1.1	0.9;
	6218	1	88.8	35.1	0	1.24	0	1.026244	4.343265	150	8	1.1	0.9;
	6219	1	-0	-0	0	0.22	0	1.028453	-13.059302	220	5	1.1	0.9;
	6220	1	231.9	27.6	0	0.36	0	1.051531	-23.264883	220	5	1.1	0.9;
	6221	1	0	0	0	0.7	0	1.035504	58.917069	220	3	1.1	0.9;
	6222	1	-11.79	-22.64	0.10131	0.71027	0	1.068025	52.94959	380	3	1.1	0.9;
	6223	1	0	0	0	1.28	0	0.93559	-33.413173	150	9	1.1	0.9;
	6224	1	0	0	0	58.990785	0	0.997449	-13.147218	380	5	1.1	0.9;
	6225	1	-0	0	0	0.8	0	1.057364	-20.767128	220	1	1.1	0.9;
	6226	1	-0	0	0	1.14	0	1.001311	-31.160904	150	9	1.1	0.9;
	6227	1	88	70.3	0	36.55	0	1.024457	18.973752	380	3	1.1	0.9;
	6228	1	26.9	3.8	0	4.69	0	1.002508	0.870952	154	1	1.1	0.9;
	6229	1	26.43	5.5	0	1.03	0	1.002315	-43.161864	110	22	1.1	0.9;
	6230	1	-0	-0	0	0.98	0	0.970755	-32.184776	150	9	1.1	0.9;
	6231	1	-0	0	0	11.5	0	1.03265	-8.086069	220	5	1.1	0.9;
	6232	1	49.25	12	0	7.28	0	1.038527	-26.132776	220	5	1.1	0.9;
	6233	2	0	0	0	0	0	1.017658	6.086699	380	2	1.1	0.9;
	6234	1	227.6	71	0	8.59	0	1.004663	-18.337725	220	3	1.1	0.9;
	6235	1	16.52	-0	0	0.13	0	1.089698	-27.989026	150	9	1.1	0.9;
	6236	1	2.5	-0	0	0.64	0	1.020252	-14.638927	110	18	1.1	0.9;
	6237	2	0	0	0	0	0	1.01301	-24.427366	154	1	1.1	0.9;
	6238	1	-0	0	0	49.59	0	1.048493	0.180714	380	13	1.1	0.9;
	6239	2	0	0	0	0	0	1.026501	-45.304593	220	4	1.1	0.9;
	6240	1	0	-0	0	6.31	0	1.071299	-34.524554	220	5	1.1	0.9;
	6241	1	-0	0	0	0.94	0	1.013864	-32.762942	150	9	1.1	0.9;
	6242	2	0	0	0	0	0	1.018892	-22.911651	110	18	1.1	0.9;
	6243	1	88.41	34.42	0	3.45	0	1.033075	-44.53921	220	4	1.1	0.9;
	6244	1	0	-0	0	2.44	0	0.975696	-50.033323	220	4	1.1	0.9;
	6245	1	-0	-0	0.27075	34.52435	0	1.081036	-13.340032	380	18	1.1	0.9;
	6246	1	-0	0	0	86.73	0	1.019835	-37.533438	380	5	1.1	0.9;
	6247	1	0	-0	0	0.3	0	0.999875	-27.889716	110	18	1.1	0.9;
	6248	1	-0.6	-1.58	0	0.22	0	1.115298	-23.607239	380	21	1.3	0.7;
	6249	2	0	0	0	0	0	1.040939	-41.872191	380	6	1.1	0.9;
	6250	1	0	-0	0	1.07	0	0.974648	-24.580743	110	18	1.1	0.9;
	6251	1	380.53	113.37	0	115	0	1.06817	-3.174976	380	14	1.1	0.9;
	6252	1	164.7	44.5	0	3.67	0	1.019466	-30.39953	220	5	1.1	0.9;
	6253	1	128.8	29.3	0	26.74	0	1.038769	-26.091528	220	5	1.1	0.9;
	6254	1	0	-0	0	9.9	0	0.985623	-55.089132	220	10	1.1	0.9;
	6255	1	-0	0	0	26.52	0	1.08589	-36.686255	150	9	1.3	0.7;
	6256	1	0	0	0.16245	33.4651	0	1.08226	-21.155278	380	18	1.3	0.7;
	6257	1	0	-0	0	36.48	0	1.128073	-23.762504	380	9	1.3	0.7;
	6258	1	11.61	2.5	0	3.84	0	1.017726	-22.318199	110	19	1.1	0.9;
	6259	1	178	33.8	0	2.73	0	1.022938	-25.672006	220	7	1.1	0.9;
	6260	1	47.55	29.4	0	5.79	0	0.989904	-28.489073	150	9	1.1	0.9;
	6261	1	75.88	11.2	0	4.58	0	1.022722	-54.933363	220	4	1.1	0.9;
	6262	1	12.81	4.7	0	0.64	0	1.018884	-28.554307	110	21	1.1	0.9;
	6263	1	0	-0	0	42.81	0	1.033114	12.718572	380	8	1.1	0.9;
	6264	1	0	0	0	0.17	0	0.99319	-28.514207	110	18	1.1	0.9;
	6265	1	46.56	9.25	0	2.37	0	1.001672	-26.56415	110	18	1.1	0.9;
	6266	1	32.1	4.5	0	2.82	0	1.009675	-22.744437	154	1	1.1	0.9;
	6267	1	91.3	18.1	0	3.05	0	1.038786	-24.721929	220	5	1.1	0.9;
	6268	1	112.17	-11.06	0	13.68	0	1.028758	-10.98801	220	13	1.1	0.9;
	6269	1	-0	0	0	2.28	0	1.017159	-21.800592	154	1	1.1	0.9;
	6270	2	0	0	0	0	0	1.02754	-20.916236	154	1	1.1	0.9;
	6271	1	3.7	2.6	0	4.45	0	1.049467	-20.418725	220	5	1.1	0.9;
	6272	1	125.9	-21.2	0	18.41	0	1.045219	38.225194	380	3	1.1	0.9;
	6273	1	12.11	4.4	0	1.26	0	1.007417	-41.689181	110	22	1.1	0.9;
	6274	1	441.2	200.7	0	42.12	0	1.014048	-14.184814	380	3	1.1	0.9;
	6275	1	122.8	37.2	0	37.13	0	1.014286	-5.040216	150	8	1.1	0.9;
	6276	1	65.07	40.3	0	5.54	0	1.002585	-34.537706	150	9	1.1	0.9;
	6277	1	153.7	21.9	0	17.71	0	1.009854	-23.52965	154	1	1.1	0.9;
	6278	1	15.22	4.9	0	0.18	0	1.010298	-29.886387	110	21	1.1	0.9;
	6279	2	0	0	0	0	0	1.008432	-26.417695	110	18	1.1	0.9;
	6280	2	0	0	0	0	0	1.017343	-33.741034	220	10	1.1	0.9;
	6281	1	52.9	7.6	0	4.15	0	1.02011	0.909328	154	1	1.1	0.9;
	6282	1	-0.28	-5.54	0	0	0	1.041347	25.837851	220	8	1.1	0.9;
	6283	1	0	-0	0	18.64	0	1.041071	15.4593	380	8	1.1	0.9;
	6284	2	0	0	0	0	0	1.057119	52.656319	380	3	1.1	0.9;
	6285	1	51.96	22.8	0	0.6	0	0.992621	-27.394743	110	18	1.1	0.9;
	6286	1	0	0	0	1.51	0	1.034293	38.874729	220	3	1.1	0.9;
	6287	1	-0	-0	0	7.1	0	1.038228	-0.750536	220	13	1.1	0.9;
	6288	1	25.23	10.3	0	2.69	0	1.021175	-3.901	150	8	1.1	0.9;
	6289	1	0	-0	0	3.19	0	1.023817	-1.557372	154	1	1.1	0.9;
	6290	1	47.65	20.4	0	3.63	0	1.054471	-13.30135	220	5	1.1	0.9;
	6291	2	0	0	0	0	0	1.036939	-24.369757	220	5	1.1	0.9;
	6292	1	186.5	49.9	0	13.17	0	1.035234	-48.988617	220	4	1.1	0.9;
	6293	1	-0	0	0	33.66	0	0.998042	-4.763116	380	2	1.1	0.9;
	6294	1	90.5	12.8	0	2.68	0	1.001876	1.535738	154	1	1.1	0.9;
	6295	1	0	-0	0	46.83	0	1.047998	8.46211	150	8	1.1	0.9;
	6296	1	15.02	6	0	6.06	0	1.028373	-24.381207	220	3	1.1	0.9;
	6297	1	20.02	8	0	5.4	0	1.026195	-25.005607	220	3	1.1	0.9;
	6298	1	0	-0	0.274712	52.929035	0	1.046599	-48.609726	380	10	1.1	0.9;
	6299	1	-0	-0	0	2.37	0	1.044484	-45.095059	220	4	1.1	0.9;
	6300	1	139.1	19.8	0	16.14	0	1.005743	-20.134354	400	1	1.1	0.9;
	6301	1	-0.57	-1	0	2.27	0	1.055283	-29.297904	150	9	1.1	0.9;
	6302	1	134.05	13.3	0	25.19	0	1.019087	-29.180313	220	6	1.1	0.9;
	6303	1	49.96	15.6	0	2.66	0	0.996209	-51.087726	220	6	1.1	0.9;
	6304	1	1.7	0.6	0	0.56	0	1.031791	-24.727457	110	19	1.1	0.9;
	6305	1	31.33	6.4	0	1.96	0	0.918517	-36.137209	110	18	1.1	0.9;
	6306	1	0	-0	0	2.13	0	1.0745	-24.576742	220	5	1.1	0.9;
	6307	1	132.4	-10.4	0	3.02	0	1.015778	-17.48363	220	3	1.1	0.9;
	6308	1	0	-0	0	0.05	0	1.040265	-25.881678	220	5	1.1	0.9;
	6309	1	0	-7.5	0	0	0	1.014099	-32.864305	150	9	1.1	0.9;
	6310	1	0	0	0	22.91	0	1.012343	-33.151582	150	9	1.1	0.9;
	6311	1	0	-0	0	2.05	0	0.968737	-43.487953	220	10	1.1	0.9;
	6312	1	-0	-0	0	1.66	0	1.035023	-13.116164	220	1	1.1	0.9;
	6313	1	48.95	14.9	0	4.04	0	1.028849	-24.275998	220	5	1.1	0.9;
	6314	1	30.03	15	0	0.81	0	1.047653	-25.220007	110	23	1.1	0.9;
	6315	1	95.71	19.6	0	20.2	0	0.985882	-43.369761	220	4	1.1	0.9;
	6316	1	474.3	203.4	0	47.469681	0	1.021233	-41.921847	380	6	1.1	0.9;
	6317	1	-0	0	0.090774	7.310444	0	1.062566	-22.918964	220	18	1.1	0.9;
	6318	1	0	0	0	1.05	0	0.994817	-32.037385	150	9	1.1	0.9;
	6319	1	-0	0	0.135375	10.042175	0	1.065832	-17.36782	380	18	1.1	0.9;
	6320	1	-0.2	-1.64	0	0	0	1.03099	3.60651	150	8	1.1	0.9;
	6321	1	128.8	70	0	1.87	0	0.999268	10.486422	220	3	1.1	0.9;
	6322	1	0	0	0	1.99	0	0.952053	-33.603403	150	9	1.1	0.9;
	6323	1	116.1	30.2	0	4.38	0	1.014377	2.339782	150	8	1.1	0.9;
	6324	1	13.9	1.9	0	2.05	0	1.004665	-24.090363	154	1	1.1	0.9;
	6325	1	23	3.2	0	3.05	0	1.010867	-2.924483	154	1	1.1	0.9;
	6326	1	9.7	1.4	0	7.89	0	1.039285	9.023334	154	1	1.1	0.9;
	6327	1	-5.01	-0	0	3.54	0	0.994681	-50.394002	220	6	1.1	0.9;
	6328	1	0	0	0	0	0	1.010196	-30.742157	150	9	1.1	0.9;
	6329	2	0	0	0	0	0	0.991058	-44.553431	150	10	1.1	0.9;
	6330	1	-0	-0	0	0.04	0	1.018101	-32.158935	150	9	1.1	0.9;
	6331	2	0	0	0	0	0	1.050971	-0.912749	220	5	1.1	0.9;
	6332	2	0	0	0	0	0	1.065904	-11.688816	220	5	1.1	0.9;
	6333	1	42.75	4	0	5.41	0	1.008527	-41.358456	220	4	1.1	0.9;
	6334	2	0	0	0	0	0	1.005334	-50.248843	220	6	1.1	0.9;
	6335	1	92.76	-46.04	0	1.24	0	1.021846	-25.530587	220	7	1.1	0.9;
	6336	1	245.51	46.8	0	4.89	0	1.014144	8.984849	220	3	1.1	0.9;
	6337	1	0	0	0	8.26	0	1.087363	-9.876082	380	5	1.1	0.9;
	6338	2	0	0	0	0	0	1.022507	-26.966389	110	18	1.1	0.9;
	6339	1	7.21	2.4	0	1.2	0	0.967776	-30.269237	110	18	1.1	0.9;
	6340	1	-0	0	0	2.27	0	1.052729	-45.804866	380	10	1.1	0.9;
	6341	1	22.73	5.3	0	0.54	0	0.997693	-26.954942	110	18	1.1	0.9;
	6342	1	-0	0	0	3.25	0	1.036453	17.18525	110	8	1.1	0.9;
	6343	1	143.9	123.3	0.203819	60.947564	0	1.029324	5.16207	380	3	1.1	0.9;
	6344	1	-0	0	0	0.27	0	1.019574	-44.356473	220	4	1.1	0.9;
	6345	1	0	0	0	0	0	1.00913	8.083384	380	8	1.1	0.9;
	6346	1	-0	0	0	3.8	0	1.031575	-37.915148	110	12	1.1	0.9;
	6347	1	22.22	13.8	0	3.11	0	1.027292	-29.311327	150	9	1.1	0.9;
	6348	1	32.04	15	0	0.15	0	1.025206	-30.620221	110	23	1.1	0.9;
	6349	1	55.46	14.1	0	0.92	0	1.019543	-48.799658	220	4	1.1	0.9;
	6350	1	35.04	20	0	2.49	0	1.015073	-27.645206	220	21	1.1	0.9;
	6351	2	0	0	0	0	0	1.0521	-16.751466	220	5	1.1	0.9;
	6352	1	-0	-0	0	0.68	0	0.975717	-33.254967	110	21	1.1	0.9;
	6353	1	0.6	-0	0	0.6	0	0.998324	-24.595489	110	18	1.1	0.9;
	6354	1	0	-0	0	47.06	0	1.038149	6.223478	150	8	1.1	0.9;
	6355	1	0	0	0	0.23	0	1.066746	-28.057054	150	9	1.1	0.9;
	6356	1	44.95	12.8	0	2.61	0	1.018984	6.965976	150	8	1.1	0.9;
	6357	1	430.8	107.9	0	11.64	0	0.986348	-24.831394	220	5	1.1	0.9;
	6358	2	0	0	0	0	0	0.986287	-33.260386	110	21	1.1	0.9;
	6359	1	74.68	10.8	0	5.82	0	1.023774	-54.379092	220	4	1.1	0.9;
	6360	1	0	0	0	1	0	1.080589	-21.471577	150	9	1.1	0.9;
	6361	1	8.11	1.4	0	1.47	0	0.974587	-32.126019	110	18	1.1	0.9;
	6362	1	24.03	8	0	0	0	1.059022	-38.046681	120	6	1.1	0.9;
	6363	1	12.94	4.9	0	0.15	0	0.983651	-16.099618	110	18	1.1	0.9;
	6364	1	-0	-0	0	0.54	0	1.025309	-29.308119	150	9	1.1	0.9;
	6365	1	0	-0	0	3.28	0	1.056272	-29.780742	220	1	1.1	0.9;
	6366	2	0	0	0	0	0	1.077817	8.163184	380	14	1.1	0.9;
	6367	1	7.41	2.7	0	1.39	0	0.988789	-15.148376	110	19	1.1	0.9;
	6368	2	0	0	0	0	0	1.042432	-8.519581	220	5	1.1	0.9;
	6369	1	-0	-0	0	14.29	0	1.031063	18.744746	380	3	1.1	0.9;
	6370	2	0	0	0	0	0	1.014839	-20.352422	400	1	1.1	0.9;
	6371	1	-0	-0	0	0.35	0	1.045225	-0.934949	220	5	1.1	0.9;
	6372	1	0	0	0	0	0	0.999992	-36.806485	150	9	1.1	0.9;
	6373	2	0	0	0	0	0	1.04869	-2.774001	220	14	1.1	0.9;
	6374	2	0	0	0	0	0	1.077431	59.335183	380	3	1.1	0.9;
	6375	1	-18.69	-7.9	0	2.09	0	1.022959	-25.302017	220	7	1.1	0.9;
	6376	2	0	0	0	0	0	1.044986	-0.4155	220	5	1.1	0.9;
	6377	1	11.61	7.2	0	16.18	0	0.962313	-37.255005	150	9	1.1	0.9;
	6378	1	11.31	3.7	0	0.98	0	1.00561	-36.784313	110	12	1.1	0.9;
	6379	1	118.4	16.9	0	27.1	0	1.035545	-1.331734	154	1	1.1	0.9;
	6380	1	-0	-0	0	0.52	0	1.060164	-26.992958	150	9	1.1	0.9;
	6381	1	246.2	119.1	0	31.72	0	0.9827	-56.788513	220	10	1.1	0.9;
	6382	1	75.43	10.49	0	7.91	0	1.038313	-9.903026	220	5	1.1	0.9;
	6383	1	15.62	5.3	0	2.17	0	1.013232	22.32879	110	8	1.1	0.9;
	6384	1	159.9	47.4	0	3.65	0	1.041723	-1.727757	220	5	1.1	0.9;
	6385	1	5.41	2.81	0	1.9	0	0.996222	-27.549163	150	9	1.1	0.9;
	6386	1	20.9	3	0	1.96	0	1.033209	-11.212932	154	1	1.1	0.9;
	6387	1	0	0	0	3.86	0	1.042671	-35.832808	220	1	1.1	0.9;
	6388	1	27.9	4	0	6.41	0	1.00708	-26.130031	154	1	1.1	0.9;
	6389	1	-0	0	0	0.88	0	0.961942	-31.67051	150	9	1.1	0.9;
	6390	1	-0	0	0	3.36	0	1.059508	-47.950975	380	10	1.1	0.9;
	6391	1	74.32	11.08	0	2.08	0	1.017321	62.550015	150	1	1.1	0.9;
	6392	1	80.09	19	0	1.88	0	1.024028	-25.903809	110	15	1.1	0.9;
	6393	1	93.7	29.9	0	2.1	0	1.018797	6.838622	110	8	1.1	0.9;
	6394	1	53.86	-15.3	0	2.87	0	1.019286	-46.504574	220	4	1.1	0.9;
	6395	1	154.7	48.2	0	5.49	0	1.027426	8.00829	150	8	1.1	0.9;
	6396	1	4	0.1	0	0.02	0	0.893322	-35.880472	150	9	1.3	0.7;
	6397	1	0	0	0	27.4	0	1.019958	-42.190769	220	4	1.1	0.9;
	6398	1	0	0	0	249.421581	0	1.016121	-13.243746	750	24	1.1	0.9;
	6399	1	44.05	18	0	0.88	0	1.022592	-1.581502	150	8	1.1	0.9;
	6400	1	64	9.1	0	9.94	0	1.03561	1.947556	154	1	1.1	0.9;
	6401	1	62.7	9	0	1.55	0	1.036279	7.07135	154	1	1.1	0.9;
	6402	1	7.91	2.6	0	1.72	0	1.033159	-33.633581	110	12	1.1	0.9;
	6403	2	0	0	0	0	0	1.020309	-4.088178	154	1	1.1	0.9;
	6404	1	0	-0	0	0.33	0	0.96858	-37.394208	150	9	1.1	0.9;
	6405	1	76.48	19.8	0	5.41	0	1.014784	-16.214284	220	5	1.1	0.9;
	6406	1	2.5	1.3	0	1.01	0	0.984677	-22.684523	110	18	1.1	0.9;
	6407	1	0.3	-34.1	0	0.25	0	1.043751	-3.07285	220	3	1.1	0.9;
	6408	1	126.9	80.5	0	46.82	0	1.053672	-15.757992	380	6	1.1	0.9;
	6409	1	10.44	2.73	0	0.26	0	0.968577	-29.817241	110	18	1.1	0.9;
	6410	2	0	0	0	0	0	1.044406	2.065252	154	1	1.1	0.9;
	6411	1	115.2	38.9	0	2.96	0	1.020974	-23.843324	220	7	1.1	0.9;
	6412	1	-0	0	0	5.38	0	0.909668	-36.192665	150	9	1.1	0.9;
	6413	2	0	0	0	0	0	1.044796	1.632047	380	3	1.1	0.9;
	6414	1	8	1.1	0	3.65	0	1.013181	-9.318163	154	1	1.1	0.9;
	6415	2	0	0	0	0	0	0.983673	-38.669395	220	10	1.1	0.9;
	6416	1	67.67	16.4	0	3.7	0	1.04044	-9.7577	220	5	1.1	0.9;
	6417	1	0	0	0	0	0	0.991065	-34.243743	150	9	1.1	0.9;
	6418	1	43.75	2.2	0	1.36	0	1.028356	7.81585	220	3	1.1	0.9;
	6419	2	0	0	0	0	0	1.095905	16.059867	380	3	1.1	0.9;
	6420	1	0	-0	0	5.96	0	0.920126	-35.646845	150	9	1.1	0.9;
	6421	1	174.3	11.8	0	0.51	0	1.056443	-3.058176	220	3	1.1	0.9;
	6422	1	17.25	4.26	0	0.21	0	0.964839	-29.645443	110	18	1.1	0.9;
	6423	1	-0	0	0	2.26	0	1.014849	-4.843274	150	8	1.1	0.9;
	6424	1	-0.18	-2.28	0	0	0	1.048288	18.202614	380	8	1.1	0.9;
	6425	1	0	-0	0	1.08	0	1.014064	-32.843762	150	9	1.1	0.9;
	6426	1	10.21	3.8	0	4.6	0	1.053076	1.410155	220	5	1.1	0.9;
	6427	1	83.19	22.9	0	28.5	0	1.000068	-13.384209	380	5	1.1	0.9;
	6428	1	93.1	43.7	0	2.7	0	1.013807	-32.593868	150	9	1.1	0.9;
	6429	2	0	0	0	0	0	1.035857	-18.645859	220	5	1.1	0.9;
	6430	1	0	-0	0	6.03	0	1.048821	-29.228272	220	5	1.1	0.9;
	6431	1	1.4	0.5	0	0.47	0	1.032381	-33.705099	110	12	1.1	0.9;
	6432	2	0	0	0	0	0	1.049448	65.665542	220	3	1.1	0.9;
	6433	1	0	0	0	0	0	0.987316	-35.155752	150	7	1.1	0.9;
	6434	2	0	0	0	0	0	1.116905	-11.690075	220	19	1.3	0.7;
	6435	1	28.63	11	0	1.27	0	1.007686	-36.096522	110	22	1.1	0.9;
	6436	2	0	0	0	0	0	1.01064	-48.505837	220	4	1.1	0.9;
	6437	1	-0	0	0	0.64	0	1.036921	-30.618148	150	9	1.1	0.9;
	6438	1	0.1	-19.1	0	0.26	0	0.981335	6.382274	220	3	1.1	0.9;
	6439	1	0	0	0	0	0	1.023634	5.250321	150	8	1.1	0.9;
	6440	1	81.79	10.8	0	10.84	0	1.023785	37.6441	220	3	1.1	0.9;
	6441	1	3.9	1.1	0	0.19	0	0.980843	-27.928734	110	18	1.1	0.9;
	6442	1	163.3	70.3	0	5.2	0	1.028268	11.682241	150	8	1.1	0.9;
	6443	1	43.95	15.5	0	6.36	0	1.022415	-49.843209	220	4	1.1	0.9;
	6444	1	193.8	18	0	82.73	0	1.043159	-28.591602	380	6	1.1	0.9;
	6445	1	1	-2.3	0	6.33	0	1.056723	-22.024824	150	9	1.1	0.9;
	6446	1	77.19	23	0	63.4	0	1.051187	-11.794001	380	16	1.1	0.9;
	6447	1	54.96	13.2	0	2.95	0	1.033662	-33.869271	110	22	1.1	0.9;
	6448	2	0	0	0	0	0	1.004995	-32.51982	110	21	1.1	0.9;
	6449	1	0	-0	0.135375	37.092175	0	1.075881	-24.985626	380	18	1.1	0.9;
	6450	1	5.51	2.1	0	1.92	0	1.008104	-10.991611	220	5	1.1	0.9;
	6451	1	139.8	-25.8	0	22.95	0	1.024441	-13.686948	220	3	1.1	0.9;
	6452	1	-0	0	0	5.95	0	1.010648	-6.914115	220	3	1.1	0.9;
	6453	1	8.31	3	0	0.56	0	1.047048	-18.507778	110	19	1.1	0.9;
	6454	1	-0	-0	0	11.67	0	1.024026	-54.339495	220	4	1.1	0.9;
	6455	1	20.52	13.3	0	0.17	0	1.034211	-22.39971	220	5	1.1	0.9;
	6456	1	159.4	28.4	0	15.17	0	1.020407	-47.323325	220	4	1.1	0.9;
	6457	1	-1.5	-41.98	0	36.76	0	1.058488	21.135353	380	3	1.1	0.9;
	6458	2	0	0	0	0	0	1.078056	56.614172	380	3	1.1	0.9;
	6459	2	0	0	0	0	0	1.005175	-38.615882	150	10	1.1	0.9;
	6460	1	169.8	-19.6	0	53.619841	0	1.043918	-31.79712	380	6	1.1	0.9;
	6461	1	22.22	8	0	5.7	0	1.062223	-27.058127	220	21	1.1	0.9;
	6462	1	11.1	1.6	0	3.18	0	1.013876	0.058271	154	1	1.1	0.9;
	6463	1	-4.17	-2.16	0	0.66	0	1.016941	-24.028219	110	18	1.1	0.9;
	6464	1	9.41	13.8	0	65.59	0	1.034962	-25.817629	380	6	1.1	0.9;
	6465	1	3.2	-0.4	0	1.76	0	1.002795	-30.602319	110	18	1.1	0.9;
	6466	1	1.83	-5.69	0	5.49	0	1.085178	45.265218	380	3	1.1	0.9;
	6467	1	25	3.6	0	3.86	0	1.013131	-2.607288	154	1	1.1	0.9;
	6468	1	172.5	48.7	0	54.92	0	1.052987	8.023458	380	14	1.1	0.9;
	6469	1	-0	0	0	2.55	0	1.031686	48.420935	220	3	1.1	0.9;
	6470	1	-0	-0	0.225551	5.082241	0	1.013845	13.946269	380	3	1.1	0.9;
	6471	1	34.14	11.5	0	0.26	0	1.008088	-30.00138	110	21	1.1	0.9;
	6472	1	0	-0	0	9.93	0	1.038332	-24.365225	220	5	1.1	0.9;
	6473	1	76.6	10.9	0	6.82	0	1.02296	4.124025	154	1	1.1	0.9;
	6474	2	0	0	0	0	0	1.055298	-8.059073	220	5	1.1	0.9;
	6475	1	-0	0	0	149.047157	0	1.02402	-8.915069	380	5	1.1	0.9;
	6476	1	59.2	8.4	0	2.26	0	1.026494	0.885319	154	1	1.1	0.9;
	6477	1	0	0	0	1.99	0	1.028909	20.742007	220	8	1.1	0.9;
	6478	1	141.2	34.8	0	2.37	0	1.009054	-14.629451	220	5	1.1	0.9;
	6479	1	0	-0	0	23.58	0	0.985639	-32.451664	150	9	1.1	0.9;
	6480	1	-0.21	-1.04	0	0	0	1.018493	2.360002	150	8	1.1	0.9;
	6481	1	-0	-0	0.090774	1.940444	0	1.058624	-17.551184	220	18	1.1	0.9;
	6482	1	-0	0	0	29.2	0	1.032857	-43.819875	380	1	1.1	0.9;
	6483	2	0	0	0	0	0	1.064876	-20.928891	110	15	1.1	0.9;
	6484	2	0	0	0	0	0	1.024853	-45.979688	380	4	1.1	0.9;
	6485	1	0	-0	0.643825	106.534876	0	1.024272	-9.6072	380	3	1.1	0.9;
	6486	1	31.74	17.4	0	0.89	0	1.046967	-4.241061	220	5	1.1	0.9;
	6487	1	18.62	3.6	0	1.14	0	0.989539	-24.431575	110	18	1.1	0.9;
	6488	1	0	0	0	23.8	0	1.037157	-1.340065	154	1	1.1	0.9;
	6489	1	-0.12	-1.74	0	0	0	1.079806	31.595391	380	8	1.1	0.9;
	6490	1	84.59	12.9	0	3.66	0	1.021192	-53.017221	220	4	1.1	0.9;
	6491	1	0	0	0	17.82	0	1.097271	-27.582868	380	1	1.3	0.7;
	6492	2	0	0	0	0	0	1.004319	65.867968	380	1	1.1	0.9;
	6493	1	9.51	4.6	0	0.16	0	0.966345	-32.483962	150	9	1.1	0.9;
	6494	1	0	-0	0	8.2	0	1.054522	-41.983931	380	1	1.1	0.9;
	6495	1	49.66	12.3	0	2.2	0	1.039608	-22.065882	220	5	1.1	0.9;
	6496	1	0	-0	0	24.38	0	1.057457	-41.559059	380	4	1.1	0.9;
	6497	2	0	0	0	0	0	1.064333	34.261506	380	3	1.1	0.9;
	6498	2	0	0	0	0	0	1.019	7.11935	154	1	1.1	0.9;
	6499	1	-0	-0	0	13.86	0	1.04354	-45.399854	220	4	1.1	0.9;
	6500	1	25.13	9	0	1.46	0	1.068208	-23.335816	110	15	1.1	0.9;
	6501	1	25.83	9.4	0	0.56	0	1.000879	-31.297345	110	19	1.1	0.9;
	6502	2	0	0	0	0	0	1.002332	-31.18928	110	19	1.1	0.9;
	6503	1	13.21	3.5	0	0.32	0	1.067733	-16.878507	110	18	1.1	0.9;
	6504	1	49.26	30.9	0.097351	12.97086	0	1.001777	-45.153023	220	10	1.1	0.9;
	6505	1	0	-0	0	5.19	0	1.055869	-26.396286	220	23	1.1	0.9;
	6506	1	41.31	5.96	0	0.89	0	1.013914	61.440605	150	1	1.1	0.9;
	6507	1	41.8	5.9	0	10.95	0	1.02202	2.511103	154	1	1.1	0.9;
	6508	1	-0	0	0	4.52	0	1.028383	2.026329	380	3	1.1	0.9;
	6509	1	0	0	0	0.36	0	1.014496	-22.612039	150	9	1.1	0.9;
	6510	1	88.2	26.7	0	8.91	0	1.050766	-26.818777	220	5	1.1	0.9;
	6511	1	20.02	9.6	0	0.17	0	1.014563	-31.538546	150	9	1.1	0.9;
	6512	1	-0	0	0	3.4	0	0.989658	-54.467029	150	10	1.1	0.9;
	6513	1	13.92	4.3	0	1.1	0	0.981051	-18.068546	110	18	1.1	0.9;
	6514	1	-0.18	-1	0	0	0	1.043803	6.060087	150	8	1.1	0.9;
	6515	1	20.32	8.6	0	0.94	0	0.959609	-34.477699	110	18	1.1	0.9;
	6516	2	0	0	0	0	0	1.050158	-3.806608	220	5	1.1	0.9;
	6517	1	-14.73	-9.85	0	0.51	0	0.997296	-27.836517	110	18	1.1	0.9;
	6518	2	0	0	0	0	0	0.999167	8.848426	400	1	1.1	0.9;
	6519	1	-0.19	-1.94	0	0	0	1.023884	11.17255	150	8	1.1	0.9;
	6520	1	20.32	-22	0	5.35	0	1.037031	-26.556733	220	7	1.1	0.9;
	6521	1	0	0	0	2.3	0	1.044234	-4.960268	220	5	1.1	0.9;
	6522	1	30.43	12	0	0.42	0	0.994852	-24.893891	110	18	1.1	0.9;
	6523	1	24.83	9	0	0.24	0	1.028745	-8.319493	110	19	1.1	0.9;
	6524	1	113.4	61.91	0	1.51	0	1.035962	13.048122	380	3	1.1	0.9;
	6525	2	0	0	0	0	0	1.007591	-11.130613	154	1	1.1	0.9;
	6526	1	8.41	2.6	0	1.77	0	1.020731	12.405464	110	8	1.1	0.9;
	6527	1	100.1	48.9	0	0.11	0	1.060073	-27.001207	150	9	1.1	0.9;
	6528	1	69.08	24	0	4.79	0	1.008749	-36.396484	220	6	1.1	0.9;
	6529	1	446.5	105.6	0	38.83	0	1.015578	-39.22214	380	6	1.1	0.9;
	6530	1	42.85	14.3	0	0.1	0	1.025164	-40.667466	110	22	1.1	0.9;
	6531	1	-0	-0	0	31.19	0	1.038415	-33.058858	150	9	1.1	0.9;
	6532	1	120.69	18.1	0	15.46	0	1.067601	-31.837785	220	5	1.1	0.9;
	6533	1	37.59	21.67	0	0	0	1.08443	-36.845222	150	9	1.3	0.7;
	6534	1	0	0	0	0	0	1.031932	-26.167183	150	9	1.1	0.9;
	6535	1	27.9	4	0	1.86	0	1.022792	-0.959174	154	1	1.1	0.9;
	6536	1	34.9	5	0	1.74	0	1.015099	-22.497253	154	1	1.1	0.9;
	6537	1	20.72	7.5	0	0.33	0	1.033133	-24.344337	110	19	1.1	0.9;
	6538	1	349.2	61.3	0.225755	37.035147	0	1.0642	56.438431	380	3	1.1	0.9;
	6539	1	68.48	22.4	0	4.38	0	1.01829	0.554314	150	8	1.1	0.9;
	6540	1	0.9	0.3	0	0.12	0	0.991423	-36.809297	150	10	1.1	0.9;
	6541	1	27.43	-36.4	0	2.2	0	1.019739	-9.491578	220	3	1.1	0.9;
	6542	1	208.24	11.9	0	24.11	0	1.033854	-3.037915	220	2	1.1	0.9;
	6543	1	118.4	7.1	0	11.11	0	1.021518	-48.984823	220	4	1.1	0.9;
	6544	1	62.17	0.7	0	5.98	0	1.012644	-43.004801	220	4	1.1	0.9;
	6545	1	92.4	13.1	0	0.66	0	1.015478	-22.400542	154	1	1.1	0.9;
	6546	1	1.1	-0	0	0.28	0	1.006447	-18.016355	110	18	1.1	0.9;
	6547	1	26.33	8.7	0	0.58	0	1.022867	-35.857699	110	12	1.1	0.9;
	6548	1	7.61	1.3	0	0.58	0	0.957126	-31.218742	110	18	1.1	0.9;
	6549	1	-0	-0	0	0.76	0	0.890685	-37.323381	150	9	1.3	0.7;
	6550	1	3.1	-5.2	0	0.86	0	1.013572	-18.93499	110	18	1.1	0.9;
	6551	1	20.1	2.9	0	10.11	0	1.024471	-0.636312	154	1	1.1	0.9;
	6552	2	0	0	0	0	0	1.045563	-21.436085	220	5	1.1	0.9;
	6553	1	14	1.9	0	1.33	0	1.009106	-5.875714	154	1	1.1	0.9;
	6554	1	262.1	13.9	0	10.23	0	1.019608	2.984743	380	3	1.1	0.9;
	6555	1	-37.81	-21.64	0	83.362314	0	1.005583	-43.970734	380	5	1.1	0.9;
	6556	1	-0	-0	0	4.61	0	1.04276	-13.023711	220	5	1.1	0.9;
	6557	1	3.7	1.3	0	0.2	0	1.050799	-25.79533	110	19	1.1	0.9;
	6558	1	-0	0	0	17.75	0	1.022941	4.361002	380	3	1.1	0.9;
	6559	1	105.5	35.3	0	1.23	0	1.022806	-1.780878	150	8	1.1	0.9;
	6560	1	488.7	118.4	0	75.45	0	1.038437	-33.560165	380	6	1.1	0.9;
	6561	1	216	56.2	0	6.98	0	1.000631	-49.571421	220	4	1.1	0.9;
	6562	1	0	0	0	0	0	0.920987	-34.594655	150	9	1.1	0.9;
	6563	1	0	-0	0	8.94	0	1.038187	-24.977946	220	5	1.1	0.9;
	6564	1	328.9	-9.3	0	28.37	0	1.045854	-43.736595	380	4	1.1	0.9;
	6565	1	0	-0	0	4.05	0	1.026594	-7.466272	220	5	1.1	0.9;
	6566	1	43.15	14.4	0	1.73	0	1.005069	-35.144988	220	6	1.1	0.9;
	6567	1	43.15	14.8	0	17.41	0	1.018261	-29.166572	220	6	1.1	0.9;
	6568	1	-0	-0	0	59.04	0	1.092467	-21.688002	380	19	1.3	0.7;
	6569	1	178.62	14.44	0	136.23	0	1.087962	6.895826	380	14	1.1	0.9;
	6570	1	-0	0	0	32.43	0	1.042473	-21.070856	220	5	1.1	0.9;
	6571	1	0	-0	0	0.17	0	0.990822	-26.223119	110	18	1.1	0.9;
	6572	1	23.93	7.9	0	0.88	0	0.970523	-32.930385	110	21	1.1	0.9;
	6573	1	114.8	33.1	0	2.03	0	1.065303	-25.402871	110	15	1.1	0.9;
	6574	1	-0	-0	0	24.57	0	1.087544	-21.290004	380	1	1.1	0.9;
	6575	1	-0	-0	0	26.84	0	1.069716	0.756228	380	1	1.1	0.9;
	6576	1	51.76	21	0	0.57	0	1.053162	-30.773664	110	3	1.1	0.9;
	6577	1	15.14	11.3	0	13.16	0	1.010475	-33.979657	150	9	1.1	0.9;
	6578	1	0	0	0	3.48	0	1.032875	3.567164	154	1	1.1	0.9;
	6579	1	8.91	2.4	0	0.6	0	0.994046	-29.200405	110	18	1.1	0.9;
	6580	1	26.63	9.4	0	0.14	0	1.002402	-26.476301	110	18	1.1	0.9;
	6581	1	153.8	23.4	0	87.73	0	1.037904	-2.639566	380	5	1.1	0.9;
	6582	1	0	-0	0	10.15	0	1.029713	11.808225	150	8	1.1	0.9;
	6583	1	0	0	0	0	0	0.982898	-28.745965	150	9	1.1	0.9;
	6584	1	-22.27	-8.9	0.090706	32.332802	0	1.05456	46.610634	380	3	1.1	0.9;
	6585	1	24.73	6.4	0	0.64	0	1.02546	-24.631882	110	3	1.1	0.9;
	6586	1	0	0	0	3.33	0	1.026459	-10.177086	220	3	1.1	0.9;
	6587	2	0	0	0	0	0	1.017615	2.135914	154	1	1.1	0.9;
	6588	1	0	0	0	34.29	0	1.009398	-21.29386	400	1	1.1	0.9;
	6589	1	124.8	45.6	0	3.12	0	1.025082	16.182362	380	3	1.1	0.9;
	6590	1	-0	0	0	23.92	0	1.029523	-19.121462	380	1	1.1	0.9;
	6591	2	0	0	0	0	0	1.040735	-0.830934	220	13	1.1	0.9;
	6592	1	1.3	0	0	0.37	0	0.953545	-30.563089	110	18	1.1	0.9;
	6593	1	0	0	0	0	0	0.941728	-34.204462	150	9	1.1	0.9;
	6594	2	0	0	0	0	0	1.022988	-28.726987	220	6	1.1	0.9;
	6595	1	8.81	2.9	0	0.53	0	1.028867	-34.156928	110	12	1.1	0.9;
	6596	1	109.1	15	0	8.82	0	0.998835	-54.824288	220	4	1.1	0.9;
	6597	1	52.36	7.4	0	26.39	0	1.022878	-48.793956	220	4	1.1	0.9;
	6598	1	16.52	2.1	0	0.72	0	1.013084	-35.644676	110	22	1.1	0.9;
	6599	1	69.98	23.1	0	2.52	0	1.039568	-8.004553	110	19	1.1	0.9;
	6600	1	25.83	7.2	0	0.88	0	1.066459	11.763623	110	8	1.1	0.9;
	6601	1	77.58	-6.54	0	10.66	0	1.054508	14.632944	220	13	1.1	0.9;
	6602	1	-0	-0	0	16.66	0	1.036954	-32.168548	220	12	1.1	0.9;
	6603	1	174	49.6	0	14.84	0	1.006317	-19.270653	220	3	1.1	0.9;
	6604	1	15.22	7.2	0	1.81	0	1.015004	-40.11876	220	4	1.1	0.9;
	6605	1	137.4	81.7	0	6.1	0	1.02251	-16.453805	380	3	1.1	0.9;
	6606	1	0	0	0	0	0	0.953098	-31.912298	150	9	1.1	0.9;
	6607	1	0	0	0	40.86	0	1.051391	-42.235348	380	4	1.1	0.9;
	6608	2	0	0	0	0	0	1.019233	62.350649	150	1	1.1	0.9;
	6609	1	483.8	105.2	0	0	0	1.035975	-25.972361	120	20	1.1	0.9;
	6610	1	14.02	6	0	0.57	0	1.050877	-26.231261	110	21	1.1	0.9;
	6611	1	0	0	0	11.72	0	1.069181	59.197642	380	3	1.1	0.9;
	6612	1	41.14	15.25	0	3.96	0	1.027317	-32.558678	220	5	1.1	0.9;
	6613	1	0	-0	0	1.04	0	1.033639	-2.952605	220	1	1.1	0.9;
	6614	1	13.31	3.5	0	0.52	0	0.99179	-20.213506	110	18	1.1	0.9;
	6615	1	-0	-0	0	11.08	0	1.048798	40.834804	380	3	1.1	0.9;
	6616	1	-0	-0	0	3.06	0	1.047083	-15.929799	220	5	1.1	0.9;
	6617	1	69.7	9.9	0	4.66	0	1.005011	-10.66523	154	1	1.1	0.9;
	6618	1	60.7	8.7	0	3.23	0	1.019156	3.439669	154	1	1.1	0.9;
	6619	1	-0	-0	0	10.67	0	1.041665	-15.788489	220	5	1.1	0.9;
	6620	1	0	0	0	0	0	1.021698	-29.695588	150	9	1.1	0.9;
	6621	2	0	0	0	0	0	1.077535	43.809315	220	3	1.1	0.9;
	6622	1	0	-0	0	7.88	0	1.015825	-36.496694	220	6	1.1	0.9;
	6623	1	62.7	9	0	5.31	0	1.021842	4.726064	154	1	1.1	0.9;
	6624	1	-37.33	-11.88	0	145.18	0	1.054184	4.222548	380	5	1.1	0.9;
	6625	1	18.82	1.7	0	6.24	0	1.018954	-57.362709	220	4	1.1	0.9;
	6626	2	0	0	0	0	0	1.042091	-27.345725	110	21	1.1	0.9;
	6627	2	0	0	0	0	0	1.064448	-23.669638	110	23	1.1	0.9;
	6628	1	35.04	12.9	0	0.46	0	0.974051	-32.701563	110	21	1.1	0.9;
	6629	1	62.63	16.3	0	2.03	0	1.050125	-11.711561	220	5	1.1	0.9;
	6630	1	134.1	24.8	0	13.73	0	1.034462	-24.515125	220	5	1.1	0.9;
	6631	1	57.56	21.1	0	1.46	0	1.004579	-36.783441	220	6	1.1	0.9;
	6632	2	0	0	0	0	0	1.057948	-41.084592	380	4	1.1	0.9;
	6633	1	-0	0	0	2.48	0	1.02863	-24.975347	220	1	1.1	0.9;
	6634	1	-0	0	0	4.18	0	1.022293	16.82388	220	13	1.1	0.9;
	6635	1	23.43	14.5	0	1.05	0	1.020771	-29.719031	150	9	1.1	0.9;
	6636	1	-0	0	0	0.31	0	1.033858	0.35533	220	5	1.1	0.9;
	6637	1	-0	0	0	26.92	0	1.089775	-28.95671	380	1	1.1	0.9;
	6638	1	61.65	10.24	0	3.48	0	1.018314	-15.09801	220	5	1.1	0.9;
	6639	1	176.7	41.2	0	10.66	0	1.04045	-2.949092	220	5	1.1	0.9;
	6640	1	20.9	3	0	2.59	0	1.016745	-12.7671	154	1	1.1	0.9;
	6641	2	0	0	0	0	0	1.007887	-44.355418	110	22	1.1	0.9;
	6642	1	4.81	1	0	0.69	0	0.991117	-20.719538	110	18	1.1	0.9;
	6643	1	0	0	0	0	0	0.890999	-36.148908	150	9	1.3	0.7;
	6644	1	-0	0	0	0.27	0	1.016514	-30.698964	150	9	1.1	0.9;
	6645	1	27.63	17.1	0	5.89	0	0.951351	-29.98231	150	9	1.1	0.9;
	6646	1	48.8	6.9	0	3.56	0	1.026078	-15.84813	154	1	1.1	0.9;
	6647	1	-5.92	-2.24	0	1.43	0	0.991973	-27.913735	110	18	1.1	0.9;
	6648	1	-26.03	-0.84	0	0.1	0	1.065399	-11.833033	220	5	1.1	0.9;
	6649	1	-0	-0	0	0.69	0	1.00861	-36.870958	220	6	1.1	0.9;
	6650	1	15.32	5	0	1.3	0	1.035108	-33.809859	110	12	1.1	0.9;
	6651	1	4.61	1.7	0	0.36	0	1.065345	-19.869365	110	19	1.1	0.9;
	6652	1	-9.32	-10.58	0	11.38	0	1.04051	-3.235703	220	3	1.1	0.9;
	6653	1	-0	0	0.062219	66.642656	0	1.005353	65.666316	380	1	1.1	0.9;
	6654	2	0	0	0	0	0	1.021829	-46.874798	220	4	1.1	0.9;
	6655	1	117.5	40	0	3.69	0	1.015791	-17.492134	220	3	1.1	0.9;
	6656	2	0	0	0	0	0	1.021248	-37.522583	220	4	1.1	0.9;
	6657	1	-0	0	0	19.57	0	1.013122	25.786647	380	3	1.1	0.9;
	6658	1	29.73	11	0	0.61	0	0.983528	-39.236437	110	22	1.1	0.9;
	6659	1	199.6	27.1	0	33.58	0	1.023823	-45.476259	380	4	1.1	0.9;
	6660	1	435.2	1.3	0	3.25	0	1.006581	-49.631709	220	4	1.1	0.9;
	6661	2	0	0	0	0	0	0.997723	9.398578	220	3	1.1	0.9;
	6662	1	-0.2	-1.64	0	0	0	1.03099	3.60651	150	8	1.1	0.9;
	6663	1	74.58	16.6	0	2.02	0	1.009405	-48.745192	220	4	1.1	0.9;
	6664	1	-0	0	0	10.68	0	1.06588	-46.619327	220	5	1.1	0.9;
	6665	1	0	-0	0.090774	3.610444	0	1.056221	-22.880474	220	18	1.1	0.9;
	6666	1	-0	0	0.214323	70.931612	0	1.031749	-25.329109	380	7	1.1	0.9;
	6667	1	0	0	0	35.98	0	1.018785	-3.636685	380	3	1.1	0.9;
	6668	1	261.7	104.6	0	29.76	0	1.064039	55.017174	380	3	1.1	0.9;
	6669	1	132.9	13.6	0	3.44	0	1.006991	-53.318964	220	4	1.1	0.9;
	6670	1	-0	-0	0	3.87	0	1.026778	-24.94221	220	6	1.1	0.9;
	6671	1	-0	0	0	2	0	1.026272	-13.566694	220	6	1.1	0.9;
	6672	1	0	0	0	0.42	0	0.96488	-25.340903	110	18	1.1	0.9;
	6673	1	20.02	6	0	1.96	0	1.004066	-30.031041	110	21	1.1	0.9;
	6674	1	40.5	5.7	0	2.11	0	1.021631	-12.457879	154	1	1.1	0.9;
	6675	1	37.94	7.5	0	11.89	0	1.031326	-23.642893	220	5	1.1	0.9;
	6676	1	27.93	17.29	0	0	0	1.017147	-28.974177	150	9	1.1	0.9;
	6677	1	3.7	0.9	0	0	0	1.010881	-5.581868	154	1	1.1	0.9;
	6678	1	85.9	12.2	0	2.77	0	1.01717	-24.655234	154	1	1.1	0.9;
	6679	1	0	-0	0	2.48	0	1.028614	-53.994532	220	4	1.1	0.9;
	6680	2	0	0	0	0	0	1.025575	-46.91691	220	6	1.1	0.9;
	6681	2	0	0	0	0	0	1.067534	-41.04651	380	4	1.1	0.9;
	6682	1	-0	0	0	0.51	0	1.060301	-32.015299	110	3	1.1	0.9;
	6683	1	66.67	27.8	0	0.51	0	1.029435	-21.636054	220	7	1.1	0.9;
	6684	1	-0	0	0	9.84	0	1.040265	-25.88167	220	5	1.1	0.9;
	6685	1	70.68	11.9	0	4.34	0	0.999762	-24.218159	220	20	1.1	0.9;
	6686	1	0	0	0	0	0	0.999342	-24.208457	150	9	1.1	0.9;
	6687	1	-0	0	0	0	0	1.048081	18.196526	380	8	1.1	0.9;
	6688	1	-0	-0	0	32.21	0	1.103679	-28.69286	380	12	1.3	0.7;
	6689	1	-7.1	-4.02	0	1.05	0	0.962351	-29.210137	110	18	1.1	0.9;
	6690	1	0	0	0	0	0	0.999008	-36.938285	150	9	1.1	0.9;
	6691	1	42.55	12.5	0	1.26	0	1.080764	-34.348226	220	5	1.1	0.9;
	6692	1	10.33	3.3	0	4.65	0	1.021278	-15.766876	220	5	1.1	0.9;
	6693	1	0	0	0	21.03	0	1.063223	-18.651224	380	1	1.1	0.9;
	6694	1	-0	-0	0	0.82	0	1.017693	-42.675585	220	1	1.1	0.9;
	6695	1	45.95	16.7	0	0.34	0	1.044718	-25.525618	110	19	1.1	0.9;
	6696	1	-0	-0	0	0.35	0	0.989039	-28.255999	110	18	1.1	0.9;
	6697	1	309	74.5	0	3.69	0	1.041614	-37.409045	220	5	1.1	0.9;
	6698	1	-0	-0	0	1.32	0	1.044786	-25.523647	150	9	1.1	0.9;
	6699	1	0	0	0	0	0	1.013524	-32.326805	150	9	1.1	0.9;
	6700	1	12.6	1.8	0	1.98	0	1.021422	-0.756478	154	1	1.1	0.9;
	6701	1	11.91	1.8	0	0.37	0	1.027805	-8.66411	220	13	1.1	0.9;
	6702	1	0	0	0	0	0	1.002413	-30.25303	150	9	1.1	0.9;
	6703	2	0	0	0	0	0	1.089943	-12.791757	380	18	1.3	0.7;
	6704	1	-0	-0	0.211099	0.672832	0	0.991196	-3.76298	380	2	1.1	0.9;
	6705	1	89.2	42.7	0	13.74	0	1.044635	-29.457279	150	9	1.1	0.9;
	6706	1	0.1	4.4	0	0.19	0	0.983638	-16.068502	110	18	1.1	0.9;
	6707	1	119.5	44.5	0	3.61	0	1.021005	10.860309	150	8	1.1	0.9;
	6708	1	46.08	-47.15	0	11.39	0	1.044499	0.506424	220	13	1.1	0.9;
	6709	1	33.24	11.1	0	0.27	0	0.996873	-33.045095	110	21	1.1	0.9;
	6710	1	0	-0	0	37.35	0	1.07958	-23.72775	380	18	1.1	0.9;
	6711	1	12.11	5.4	0	0.63	0	0.962861	-25.447365	110	18	1.1	0.9;
	6712	1	7.61	6.3	0	1.33	0	0.943879	-31.452218	110	18	1.1	0.9;
	6713	1	72.48	44.9	0	1.75	0	0.999008	-36.938285	150	9	1.1	0.9;
	6714	1	-0	-0	0	0.3	0	1.045223	-0.934918	220	5	1.1	0.9;
	6715	2	0	0	0	0	0	1.025113	47.677055	220	3	1.1	0.9;
	6716	1	-0	0	0	0.76	0	1.039972	-5.088363	220	3	1.1	0.9;
	6717	1	-20.82	-8.8	0	9.24	0	1.017316	-46.573289	220	4	1.1	0.9;
	6718	1	139.8	-23.6	0	1.91	0	1.029383	-20.114842	220	3	1.1	0.9;
	6719	1	61.07	11.1	0	4.8	0	1.025617	-14.193295	220	7	1.1	0.9;
	6720	1	-1.6	-0.9	0	0.38	0	1.024902	-29.194864	110	3	1.1	0.9;
	6721	1	32.54	11.7	0	1.17	0	1.031379	-27.285526	110	15	1.1	0.9;
	6722	1	18.02	5.9	0	0.3	0	1.017782	-35.780142	110	12	1.1	0.9;
	6723	1	0	0	0	1.58	0	1.03262	-9.513194	220	5	1.1	0.9;
	6724	1	151.56	-314.2	0	60.91	0	1.047253	15.082934	380	3	1.1	0.9;
	6725	1	487.2	116.7	0	56.29	0	1.018821	-44.649613	380	6	1.1	0.9;
	6726	1	18.12	6	0	0.7	0	0.955424	-30.687827	110	18	1.1	0.9;
	6727	1	0	0	0	0	0	0.962201	-32.513552	150	9	1.1	0.9;
	6728	1	4.2	0.6	0	3.9	0	1.03221	-7.850441	154	1	1.1	0.9;
	6729	1	0	0	0.27075	91.79435	0	1.080808	-19.366298	380	18	1.1	0.9;
	6730	1	-0	0	0	2.18	0	1.035104	-21.386459	220	5	1.1	0.9;
	6731	1	62.7	9	0	1.12	0	1.01328	-22.245563	154	1	1.1	0.9;
	6732	1	-6.71	-2.4	0	0.48	0	1.080593	-17.338718	110	19	1.1	0.9;
	6733	1	24.38	5.91	0	0.41	0	1.004375	-24.946676	110	18	1.1	0.9;
	6734	2	0	0	0	0	0	1.061659	-3.667603	220	5	1.1	0.9;
	6735	1	-0	-0	0	0.4	0	1.043988	-1.453163	380	5	1.1	0.9;
	6736	2	0	0	0	0	0	1.032666	-34.198945	380	6	1.1	0.9;
	6737	1	0	-0	0	9.74	0	1.0191	-10.589491	220	5	1.1	0.9;
	6738	1	-4.91	-2.53	0	120.576812	0	1.02709	-25.011209	380	5	1.1	0.9;
	6739	2	0	0	0	0	0	1.07283	-16.070338	110	19	1.1	0.9;
	6740	1	0	0	0	19.54	0	1.03767	-1.073745	400	1	1.1	0.9;
	6741	1	0	0	0	38.84	0	1.166873	-20.237294	380	9	1.3	0.7;
	6742	1	137.4	23.9	0	7.26	0	1.033206	-21.845565	220	5	1.1	0.9;
	6743	1	-0	-0	0	79.151368	0	1.069321	-21.296261	380	20	1.1	0.9;
	6744	1	244.12	95.29	0	7.55	0	1.030454	-10.296212	220	5	1.1	0.9;
	6745	1	25.63	51.6	0	2.79	0	1.013756	-40.052079	220	4	1.1	0.9;
	6746	1	-0	-0	0	9	0	1.041833	3.40788	400	1	1.1	0.9;
	6747	1	0	-0	0	75	0	1.023667	2.127798	150	8	1.1	0.9;
	6748	1	0	-0	0	1.28	0	1.027034	-32.698811	220	1	1.1	0.9;
	6749	1	5.91	1.7	0	8.75	0	1.022364	-49.130749	220	4	1.1	0.9;
	6750	1	267.1	57.4	0	21.33	0	1.018107	-36.407447	380	6	1.1	0.9;
	6751	1	76.6	10.9	0	2.3	0	0.990989	-0.237283	154	1	1.1	0.9;
	6752	1	6.81	2.5	0	1	0	1.074044	-21.761211	110	15	1.1	0.9;
	6753	2	0	0	0	0	0	1.045824	39.68115	380	3	1.1	0.9;
	6754	1	0	0	0	34.44	0	1.045135	-15.724698	380	1	1.1	0.9;
	6755	1	18.42	6	0	0.72	0	0.973002	-32.848031	110	21	1.1	0.9;
	6756	2	0	0	0	0	0	1.025997	-46.853452	220	6	1.1	0.9;
	6757	1	0	0	0	6.57	0	1.033435	-1.313929	380	5	1.1	0.9;
	6758	1	53.66	19.5	0	1.26	0	1.027453	-28.328891	110	19	1.1	0.9;
	6759	1	-0	0	0	2.94	0	1.043522	-16.247543	380	1	1.1	0.9;
	6760	1	4.21	1.2	0	30.55	0	1.047631	-42.762581	380	10	1.1	0.9;
	6761	1	34.35	3.62	0	0.97	0	1.044313	-23.034488	110	19	1.1	0.9;
	6762	1	207.2	28.6	0	13.85	0	1.027936	-32.99687	220	11	1.1	0.9;
	6763	1	0	-0	0	5.46	0	1.039282	-24.599259	220	5	1.1	0.9;
	6764	2	0	0	0	0	0	0.99015	-28.828647	110	18	1.1	0.9;
	6765	1	12.62	7.8	0	0	0	1.036921	-30.617101	150	9	1.1	0.9;
	6766	1	-0	0	0	5.12	0	1.025019	11.340754	380	8	1.1	0.9;
	6767	1	-0	0	0	5.85	0	0.912651	-36.033898	150	9	1.1	0.9;
	6768	1	27.53	9.9	0	85.25	0	1.024616	-45.660094	380	4	1.1	0.9;
	6769	1	99	23.2	0	10.16	0	1.038685	-44.785861	220	4	1.1	0.9;
	6770	1	60.27	21.3	0	1.87	0	1.002077	-26.419817	110	18	1.1	0.9;
	6771	1	-0	0	0	0.34	0	1.048542	-43.937081	380	4	1.1	0.9;
	6772	1	61.37	23.2	0	0.22	0	1.040983	-24.457946	220	5	1.1	0.9;
	6773	1	0	-0	0	2.18	0	1.03956	-24.600336	220	5	1.1	0.9;
	6774	1	34.14	21.1	0	0.04	0	0.972605	-32.139314	150	9	1.1	0.9;
	6775	1	283.2	101.8	0.230421	43.471474	0	1.030191	-16.698866	380	7	1.1	0.9;
	6776	1	139.9	-26.2	0	5.42	0	0.99181	-28.172115	220	6	1.1	0.9;
	6777	1	-1377.93	-320	0	36.89	0	1.051803	-7.55464	380	3	1.1	0.9;
	6778	2	0	0	0	0	0	1.038013	-36.82811	380	6	1.1	0.9;
	6779	1	74.68	46.3	0	12.92	0	1.011617	-32.734235	150	9	1.1	0.9;
	6780	2	0	0	0	0	0	1.074673	-20.589176	380	19	1.1	0.9;
	6781	1	-0	0	0	20.44	0	1.038616	-1.166785	400	1	1.1	0.9;
	6782	1	18.2	2.6	0	12.22	0	1.033058	3.304265	154	1	1.1	0.9;
	6783	1	140.1	25.5	0	10.78	0	1.023694	-54.395974	220	4	1.1	0.9;
	6784	1	11.91	4.4	0	0.13	0	1.056442	-24.212325	110	15	1.1	0.9;
	6785	1	0	0	0	8.9	0	1.04291	-1.012512	220	5	1.1	0.9;
	6786	1	-6.77	-2.13	0	0.47	0	1.026314	-19.780908	110	18	1.1	0.9;
	6787	1	0	-0	0.090774	7.550444	0	1.048295	-18.194704	220	18	1.1	0.9;
	6788	1	15.22	3.9	0	0.89	0	0.983719	-28.350412	110	18	1.1	0.9;
	6789	1	0	0	0	4.09	0	0.963175	-31.640692	150	9	1.1	0.9;
	6790	1	154.9	25	0	72.89	0	1.073696	-25.028264	380	20	1.1	0.9;
	6791	1	46.35	9.1	0	18.2	0	1.042529	-25.5606	220	5	1.1	0.9;
	6792	1	43.35	16.2	0	0.77	0	0.966238	-29.985221	110	18	1.1	0.9;
	6793	2	0	0	0	0	0	1.049876	-32.799947	220	6	1.1	0.9;
	6794	1	79.49	17	0	29.21	0	1.098897	23.835219	220	3	1.1	0.9;
	6795	1	-73.98	-29.8	0	17.77	0	0.995348	-49.674217	220	10	1.1	0.9;
	6796	1	-0	0	0	1.56	0	1.057472	10.260704	154	1	1.1	0.9;
	6797	1	42.25	5.9	0	13.07	0	1.022448	-50.891154	220	4	1.1	0.9;
	6798	2	0	0	0	0	0	1.018682	6.386371	380	2	1.1	0.9;
	6799	1	253.4	81.7	0	16.85	0	1.000628	-24.084029	220	20	1.1	0.9;
	6800	1	-6.84	-0.61	0	0.49	0	1.051979	-15.235826	110	18	1.1	0.9;
	6801	1	4.2	0.7	0	0.16	0	0.989107	-17.984271	110	18	1.1	0.9;
	6802	1	-0	0	0	7.43	0	1.03121	-6.438625	220	5	1.1	0.9;
	6803	1	55.76	19.9	0	1.14	0	1.007231	-16.058199	110	18	1.1	0.9;
	6804	1	0	0	0	0	0	1.005281	-34.431326	150	9	1.1	0.9;
	6805	1	-0.5	-0.17	0	0.21	0	0.987496	-29.138849	110	18	1.1	0.9;
	6806	1	135.84	15.97	0	4.57	0	1.04316	-21.559494	220	5	1.1	0.9;
	6807	2	0	0	0	0	0	1.025274	-15.869079	220	5	1.1	0.9;
	6808	1	270.14	57.56	0	4.32	0	0.973167	5.367763	220	3	1.1	0.9;
	6809	1	76.79	30.7	0	17.7	0	1.010642	-36.564273	220	6	1.1	0.9;
	6810	1	20.02	10	0	8.08	0	1.013906	-36.128707	220	6	1.1	0.9;
	6811	2	0	0	0	0	0	1.093866	45.741934	380	3	1.1	0.9;
	6812	1	55.7	7.9	0	14.44	0	1.019103	4.226701	154	1	1.1	0.9;
	6813	1	-0	-0	0	15.06	0	1.030676	-28.526923	380	1	1.1	0.9;
	6814	1	0	0	0	27.51	0	1.040134	-33.295098	150	9	1.1	0.9;
	6815	1	3.4	1.1	0	0.55	0	0.979516	-21.235793	110	18	1.1	0.9;
	6816	2	0	0	0	0	0	1.062526	-24.747281	220	5	1.1	0.9;
	6817	1	106	24.6	0	5.9	0	1.034935	-24.867879	220	7	1.1	0.9;
	6818	1	-0	-0	0	1.77	0	1.067428	-28.590472	220	1	1.1	0.9;
	6819	1	-0	-0	0	0.3	0	0.972407	-32.151223	150	9	1.1	0.9;
	6820	2	0	0	0	0	0	1.047984	-6.715994	220	5	1.1	0.9;
	6821	1	162.7	23.2	0	4.81	0	1.028628	-6.505156	154	1	1.1	0.9;
	6822	2	0	0	0	0	0	1.039383	-19.102075	380	6	1.1	0.9;
	6823	1	36.84	14.4	0	5.09	0	1.042858	3.417992	150	8	1.1	0.9;
	6824	1	68.78	28.2	0	0.26	0	1.011287	-4.992497	150	8	1.1	0.9;
	6825	1	211.1	93.1	0	0	0	0.959831	-22.876857	110	18	1.1	0.9;
	6826	1	326.9	-82.3	0	0	0	1.063269	-8.045999	150	2	1.1	0.9;
	6827	1	55.46	18.5	0	1.11	0	0.983853	-34.426794	110	21	1.1	0.9;
	6828	1	58.16	16.5	0	6.24	0	1.033305	-13.74934	220	5	1.1	0.9;
	6829	1	46.2	28.47	0	4.19	0	1.006688	31.027518	220	13	1.1	0.9;
	6830	1	0	-0	0	8.12	0	0.990731	7.463538	220	3	1.1	0.9;
	6831	2	0	0	0	0	0	1.031732	-9.829512	220	5	1.1	0.9;
	6832	1	-0.12	-1.74	0	0	0	1.079815	31.595432	380	8	1.1	0.9;
	6833	1	0	0	0.246363	4.841277	0	0.973456	-7.173915	380	2	1.1	0.9;
	6834	1	3.8	2.8	0	0.77	0	0.999719	-27.892541	110	18	1.1	0.9;
	6835	1	-0.15	-1.93	0	0	0	1.025803	17.762871	220	8	1.1	0.9;
	6836	1	184.88	-21.11	0	39.26	0	1.022259	-48.813369	220	4	1.1	0.9;
	6837	1	47.75	11	0	0.05	0	1.047535	-7.972968	220	5	1.1	0.9;
	6838	2	0	0	0	0	0	1.050849	4.821282	150	8	1.1	0.9;
	6839	1	78.74	17.8	0	0.58	0	1.01181	61.954496	150	1	1.1	0.9;
	6840	1	0	0	0	2.64	0	1.021842	-7.278252	154	1	1.1	0.9;
	6841	1	122.8	20.8	0	0.47	0	1.025239	-53.891327	220	4	1.1	0.9;
	6842	1	98	21.7	0	6	0	1.036174	-8.502674	220	5	1.1	0.9;
	6843	2	0	0	0	0	0	1.013427	-28.349592	110	21	1.1	0.9;
	6844	1	-0	-0	0	1.04	0	1.011515	-29.690265	220	5	1.1	0.9;
	6845	2	0	0	0	0	0	1.045173	-11.439959	220	5	1.1	0.9;
	6846	1	44.35	19.3	0	3.25	0	1.053446	-4.807622	220	5	1.1	0.9;
	6847	2	0	0	0	0	0	1.018848	2.887133	380	3	1.1	0.9;
	6848	1	153.3	27.6	0	23.2	0	1.027262	1.580761	380	3	1.1	0.9;
	6849	1	50.46	36	0	0.57	0	0.987046	-26.764003	110	18	1.1	0.9;
	6850	2	0	0	0	0	0	1.010801	-8.455575	154	1	1.1	0.9;
	6851	1	2.5	-0	0	0.11	0	0.99745	-26.064833	110	18	1.1	0.9;
	6852	2	0	0	0	0	0	1.034908	-20.130283	220	5	1.1	0.9;
	6853	1	71.08	44	0	3.28	0	1.01029	-30.963324	150	9	1.1	0.9;
	6854	1	0	0	0	0	0	1.05059	-11.653422	220	5	1.1	0.9;
	6855	1	-16.82	5.5	0	0.63	0	0.92059	-34.147317	150	9	1.1	0.9;
	6856	1	16.52	5.5	0	3.5	0	1.036628	-39.641147	110	22	1.1	0.9;
	6857	2	0	0	0	0	0	1.018722	-8.798226	380	5	1.1	0.9;
	6858	1	0	0	0	0	0	0.960834	-27.235744	150	9	1.1	0.9;
	6859	1	-0	8.5	0	0.67	0	1.063045	-33.986886	150	9	1.1	0.9;
	6860	1	0	0	0	4.6	0	1.022858	-24.2758	220	15	1.1	0.9;
	6861	2	0	0	0	0	0	1.037118	5.638422	220	13	1.1	0.9;
	6862	1	-42.25	2.8	0	0.7	0	1.043582	-30.847862	110	1	1.1	0.9;
	6863	1	-0	-0	0	11.39	0	1.022704	-29.450197	220	6	1.1	0.9;
	6864	2	0	0	0	0	0	1.020438	24.563168	380	3	1.1	0.9;
	6865	1	-0	0	0	3.37	0	1.025349	-24.32202	220	3	1.1	0.9;
	6866	1	6.41	2.1	0	0.51	0	1.013186	-35.824164	110	12	1.1	0.9;
	6867	1	1.2	0.4	0	0.6	0	1.034972	-28.119193	110	19	1.1	0.9;
	6868	2	0	0	0	0	0	1.042934	8.937058	154	1	1.1	0.9;
	6869	1	54.9	-96.8	0	16.42	0	1.022281	-53.5165	220	4	1.1	0.9;
	6870	1	0	0	0	3.67	0	1.028473	65.164851	220	1	1.1	0.9;
	6871	1	0	0	0	0.86	0	1.028131	-22.162596	220	1	1.1	0.9;
	6872	1	0	0	0	0	0	0.999884	-29.224674	150	9	1.1	0.9;
	6873	1	0	0	0	0	0	0.980745	-31.583482	150	9	1.1	0.9;
	6874	2	0	0	0	0	0	1.03371	-11.209656	380	6	1.1	0.9;
	6875	1	-0	-0	0	0	0	0.991313	-30.820225	150	9	1.1	0.9;
	6876	1	1.8	0.6	0	0.75	0	1.020214	-38.870017	110	12	1.1	0.9;
	6877	2	0	0	0	0	0	1.062542	-42.235295	380	4	1.1	0.9;
	6878	1	25.23	8.3	0	0.61	0	0.97868	-32.463806	110	21	1.1	0.9;
	6879	1	199.6	24.4	0	10.6	0	1.036595	-45.824534	220	4	1.1	0.9;
	6880	1	0	0	0	1.36	0	1.043251	-4.724019	220	5	1.1	0.9;
	6881	1	11.41	5.1	0	1	0	0.922343	-32.798823	110	18	1.1	0.9;
	6882	1	67.47	6.1	0	1.46	0	1.009653	-54.489183	220	4	1.1	0.9;
	6883	1	0	-0	0	12.4	0	1.056872	-40.622181	380	1	1.1	0.9;
	6884	1	-0.12	-1.31	0	0	0	1.079826	31.59123	380	8	1.1	0.9;
	6885	1	25.73	15.9	0	0.74	0	0.96795	-32.435222	150	9	1.1	0.9;
	6886	1	119.3	-2.8	0	0.87	0	0.998542	9.756196	220	3	1.1	0.9;
	6887	1	-0	-0	0	3.13	0	1.044495	-3.165926	220	5	1.1	0.9;
	6888	2	0	0	0	0	0	1.046712	-13.457465	220	5	1.1	0.9;
	6889	1	-12.49	-5.83	0	6.235224	0	1.029246	2.262303	380	5	1.1	0.9;
	6890	1	-0	0	0	5.59	0	1.006955	-4.145691	154	1	1.1	0.9;
	6891	1	135.61	26.86	0	1.93	0	1.037473	-14.658529	220	5	1.1	0.9;
	6892	1	-53.06	-100	0	0.06	0	1.058992	-28.152698	220	6	1.1	0.9;
	6893	2	0	0	0	0	0	1.002405	23.539101	400	1	1.1	0.9;
	6894	1	-1235.46	-43	0	90.6	0	1.067746	54.777918	380	3	1.1	0.9;
	6895	1	182.5	47	0	33.69	0	1.028013	-42.93267	380	6	1.1	0.9;
	6896	1	8.31	2.7	0	0.72	0	1.035677	-33.647621	110	12	1.1	0.9;
	6897	1	176.8	39.4	0	28.24	0	1.050254	-16.909677	380	5	1.1	0.9;
	6898	1	31.03	10	0	0.13	0	1.00096	-31.245697	110	21	1.1	0.9;
	6899	1	8.01	2.6	0	1.11	0	1.030208	-32.873404	110	12	1.1	0.9;
	6900	2	0	0	0	0	0	1.08427	48.195272	380	3	1.1	0.9;
	6901	1	-37.84	-19.23	0	28.140258	0	1.00996	-32.23459	380	5	1.1	0.9;
	6902	1	0.3	-9	0	1.96	0	1.046903	-2.652524	220	3	1.1	0.9;
	6903	1	20	2.9	0	2.33	0	1.016371	-13.962953	154	1	1.1	0.9;
	6904	1	0	0	0	6.97	0	1.024553	-56.340587	220	4	1.1	0.9;
	6905	1	-0	-0	0	4.4	0	0.994534	-34.603144	150	9	1.1	0.9;
	6906	1	115.2	55.1	0	30.49	0	1.056962	12.098847	380	3	1.1	0.9;
	6907	2	0	0	0	0	0	1.059798	-37.191473	380	17	1.1	0.9;
	6908	1	0	0	0	3.01	0	1.032036	-40.507685	220	5	1.1	0.9;
	6909	1	122.5	8.7	0	15.85	0	1.032135	-23.480384	220	5	1.1	0.9;
	6910	2	0	0	0	0	0	1.028742	-37.860449	380	6	1.1	0.9;
	6911	1	18.72	7.7	0	1.09	0	0.988014	-39.679641	110	22	1.1	0.9;
	6912	1	-0	0	0	0.88	0	1.001369	-29.850843	150	9	1.1	0.9;
	6913	1	18.41	41.8	0	0.17	0	1.004998	-36.450118	110	12	1.1	0.9;
	6914	2	0	0	0	0	0	1.083288	55.169444	380	3	1.1	0.9;
	6915	1	11.71	3.7	0	0.8	0	1.023422	13.262415	110	8	1.1	0.9;
	6916	1	-0	0	0	0.66	0	1.075717	-26.413433	150	9	1.1	0.9;
	6917	1	-0	-0	0	96.89	0	1.060595	-42.915887	380	4	1.1	0.9;
	6918	1	6.71	1	0	0.39	0	0.984829	-27.636035	110	18	1.1	0.9;
	6919	1	21.62	9.8	0	0.14	0	0.998163	-26.941463	110	18	1.1	0.9;
	6920	2	0	0	0	0	0	1.008862	-6.332689	154	1	1.1	0.9;
	6921	1	-37.34	-13.33	0	251.529665	0	1.028797	-1.750787	380	5	1.1	0.9;
	6922	1	113.6	44.7	0	1.91	0	1.035254	-8.69566	220	5	1.1	0.9;
	6923	1	33.34	8.6	0	3.93	0	1.034428	14.578941	110	8	1.1	0.9;
	6924	2	0	0	0	0	0	1.02139	-36.374905	220	22	1.1	0.9;
	6925	1	-10.36	-4.66	0	0.58	0	1.071735	-30.969262	110	3	1.1	0.9;
	6926	1	124.5	101.4	0	0.53	0	1.035565	-12.545415	220	5	1.1	0.9;
	6927	1	47.3	6.7	0	4.05	0	1.01541	0.234942	154	1	1.1	0.9;
	6928	1	0	-0	0	1.16	0	1.023523	-29.563055	150	9	1.1	0.9;
	6929	1	-92.2	-244.4	0	7.23	0	1.052834	-17.763654	400	1	1.1	0.9;
	6930	2	0	0	0	0	0	1.047953	-13.982011	380	16	1.1	0.9;
	6931	1	20.9	3	0	7.14	0	1.015354	-9.011275	154	1	1.1	0.9;
	6932	1	117	20.4	0	22.34	0	1.028056	-53.852837	220	4	1.1	0.9;
	6933	1	-0	0	0	0.1	0	1.031105	-26.078897	110	19	1.1	0.9;
	6934	2	0	0	0	0	0	1.059171	62.942185	150	1	1.1	0.9;
	6935	1	0	0	0	2.15	0	0.970025	-24.238119	150	9	1.1	0.9;
	6936	2	0	0	0	0	0	1.022407	-39.112868	220	4	1.1	0.9;
	6937	2	0	0	0	0	0	1.025063	-7.633494	220	13	1.1	0.9;
	6938	1	3.2	1.1	0	0.56	0	1.027581	-33.829709	110	12	1.1	0.9;
	6939	1	135	25.7	0	43.01	0	1.0377	-36.880041	380	6	1.1	0.9;
	6940	1	0	-0	0	10.46	0	1.049805	-28.106742	220	5	1.1	0.9;
	6941	1	43.7	6.2	0	1.73	0	1.002574	-9.256496	154	1	1.1	0.9;
	6942	1	21.62	7.8	0	1.43	0	1.035522	-22.267144	110	19	1.1	0.9;
	6943	1	-0	0	0	1.13	0	1.039166	-28.646488	150	9	1.1	0.9;
	6944	1	-0	-0	0	1.6	0	1.025716	-29.458867	150	9	1.1	0.9;
	6945	1	11.8	1.7	0	3.85	0	1.01086	-9.175729	154	1	1.1	0.9;
	6946	1	0.2	1.8	0	0.43	0	1.033877	47.079271	220	3	1.1	0.9;
	6947	2	0	0	0	0	0	1.049432	-3.998316	220	5	1.1	0.9;
	6948	2	0	0	0	0	0	1.010398	11.762126	220	3	1.1	0.9;
	6949	1	35.44	5.3	0	2.24	0	1.024047	-54.300611	220	4	1.1	0.9;
	6950	2	0	0	0	0	0	1.020998	-5.538575	154	1	1.1	0.9;
	6951	1	22.8	3.2	0	9.98	0	1.056422	9.349947	154	1	1.1	0.9;
	6952	1	180.9	33.2	0	29.17	0	1.07171	-25.432953	220	5	1.1	0.9;
	6953	1	61.77	10.2	0	4.01	0	1.015316	-50.439781	220	4	1.1	0.9;
	6954	1	18.12	10	0	1.03	0	1.030112	-20.371845	220	5	1.1	0.9;
	6955	1	0	0	0	0	0	1.001649	-25.341796	150	9	1.1	0.9;
	6956	1	86.5	29	0	1.48	0	1.069137	26.053819	380	3	1.1	0.9;
	6957	1	27.86	14.22	0	0	0	1.081823	-35.990705	150	9	1.1	0.9;
	6958	1	20.02	9	0	0.15	0	1.013663	-36.143405	220	6	1.1	0.9;
	6959	1	-0	-0	0	4.07	0	1.045083	-14.961356	220	1	1.1	0.9;
	6960	1	389.8	-68.1	0	5.03	0	0.986652	-56.659914	220	10	1.1	0.9;
	6961	1	56.56	18.4	0	3.29	0	1.045446	-0.160283	220	5	1.1	0.9;
	6962	1	-0	-0	0.464351	40.788627	0	1.070029	-24.62348	380	11	1.1	0.9;
	6963	1	-0	0	0	0.69	0	1.023802	-14.065178	220	6	1.1	0.9;
	6964	1	120.1	2.5	0	6.6	0	1.033776	-45.4233	220	4	1.1	0.9;
	6965	1	125	107.2	0.21016	53.165208	0	1.034078	8.639908	380	3	1.1	0.9;
	6966	1	94.4	13.5	0	5.37	0	1.023334	5.03485	154	1	1.1	0.9;
	6967	2	0	0	0	0	0	1.065194	-23.706414	220	23	1.1	0.9;
	6968	1	0	-0	0	0	0	1.001887	-31.055317	220	1	1.1	0.9;
	6969	2	0	0	0	0	0	1.069563	-23.792716	220	5	1.1	0.9;
	6970	1	0	0	0	27.8	0	1.020925	-46.712181	380	4	1.1	0.9;
	6971	1	0	-0	0	7.51	0	1.058891	-15.907161	220	19	1.1	0.9;
	6972	1	83.36	2.6	0	1.06	0	1.036075	-5.904054	220	13	1.1	0.9;
	6973	1	92.1	43.8	0	3.69	0	1.016606	7.865112	220	3	1.1	0.9;
	6974	1	-0	0	0	0.75	0	0.991487	-20.978285	110	18	1.1	0.9;
	6975	1	0.6	0.1	0	0.38	0	0.99645	-26.963968	110	18	1.1	0.9;
	6976	1	36.04	22.3	0	4.15	0	0.916596	-35.844928	150	9	1.1	0.9;
	6977	1	53.06	75.8	0	51.93	0	1.032865	-0.690816	380	3	1.1	0.9;
	6978	1	-0	-0	0	3.08	0	1.065903	45.68187	220	3	1.1	0.9;
	6979	1	19.4	2.8	0	9.78	0	1.019782	-8.698855	154	1	1.1	0.9;
	6980	1	-0	-0	0	0.34	0	1.034733	13.970076	154	1	1.1	0.9;
	6981	1	195.5	67.9	0	6.17	0	1.006038	5.432391	220	3	1.1	0.9;
	6982	2	0	0	0	0	0	1.039641	-21.299201	220	5	1.1	0.9;
	6983	1	-0	-0	0	12.69	0	1.018641	13.955711	380	3	1.1	0.9;
	6984	1	0	0	0	0	0	0.85042	-39.715801	150	9	1.3	0.7;
	6985	1	125.2	1.2	0	15.12	0	1.023726	-54.438443	220	4	1.1	0.9;
	6986	1	-0	-0	0	0.39	0	1.021862	51.360184	220	3	1.1	0.9;
	6987	1	112.1	36.3	0	6.4	0	1.036672	2.72691	150	8	1.1	0.9;
	6988	1	0	0	0	0.12	0	0.984204	-31.843275	110	21	1.1	0.9;
	6989	1	0	0	0	1.91	0	1.041863	-24.366301	220	5	1.1	0.9;
	6990	1	-0	0	0	4.54	0	1.045081	-24.415761	220	5	1.1	0.9;
	6991	1	-20.38	-7.04	0	0.65	0	1.023401	-27.835001	110	18	1.1	0.9;
	6992	1	89.7	17.5	0	4.06	0	1.008232	-38.128195	220	6	1.1	0.9;
	6993	2	0	0	0	0	0	1.069361	34.929165	380	3	1.1	0.9;
	6994	1	0	0	0	11.7	0	1.018182	-41.506416	220	4	1.1	0.9;
	6995	1	-0	0	0	3.42	0	1.036507	2.851584	220	14	1.1	0.9;
	6996	1	0	-0	0	1.37	0	1.041131	-17.086191	220	1	1.1	0.9;
	6997	1	190.3	84.7	0	7.15	0	0.990611	-8.595684	220	3	1.1	0.9;
	6998	1	30.03	8.6	0	21.45	0	1.022559	8.17473	150	8	1.1	0.9;
	6999	2	0	0	0	0	0	1.083106	-14.553923	380	19	1.3	0.7;
	7000	1	2.3	0.8	0	0.06	0	1.06783	-19.57607	110	19	1.1	0.9;
	7001	1	46.35	28.7	0	16.37	0	1.001649	-25.341796	150	9	1.1	0.9;
	7002	1	51.16	12	0	6.02	0	1.027944	-42.390022	220	4	1.1	0.9;
	7003	1	2.9	0	0	0.62	0	1.018511	-16.167688	110	18	1.1	0.9;
	7004	1	0	0	0	0	0	1.047845	18.20106	380	1	1.1	0.9;
	7005	1	10.01	2	0	0.61	0	1.020264	-28.161001	150	7	1.1	0.9;
	7006	1	-0	0	0	5.97	0	1.044873	52.29999	220	3	1.1	0.9;
	7007	1	9.31	0.8	0	0.9	0	1.003454	-26.41149	110	3	1.1	0.9;
	7008	2	0	0	0	0	0	1.04776	-29.710984	220	22	1.1	0.9;
	7009	1	4.2	0.6	0	7.92	0	1.036539	7.192909	154	1	1.1	0.9;
	7010	1	-0	0	0.090774	2.110444	0	1.028432	-27.289128	220	18	1.1	0.9;
	7011	1	84.59	21.7	0	7.42	0	1.02726	7.833295	110	8	1.1	0.9;
	7012	1	-15.12	5	0	0.25	0	1.045141	-25.390576	150	9	1.1	0.9;
	7013	1	18.5	3.8	0	0.38	0	1.03657	8.839728	154	1	1.1	0.9;
	7014	1	0	0	0	0.3	0	1.032053	-2.878521	220	5	1.1	0.9;
	7015	1	-0	-0	0	3.06	0	1.045951	-25.528326	150	9	1.1	0.9;
	7016	1	190	72.7	0	30.78	0	1.007348	-37.328107	220	6	1.1	0.9;
	7017	1	0	-0	0.149251	110.015393	0	1.066885	-17.152077	380	18	1.1	0.9;
	7018	1	53.26	29.5	0	41.35	0	1.023351	11.178136	150	8	1.1	0.9;
	7019	1	52.56	12.8	0	2.87	0	1.038185	-26.156975	220	5	1.1	0.9;
	7020	1	0	0	0	0	0	0.992233	-34.753986	150	9	1.1	0.9;
	7021	1	226.8	53.3	0	3.26	0	1.003991	-17.389789	220	5	1.1	0.9;
	7022	1	14.92	5.4	0	0.62	0	1.033177	-23.873027	110	19	1.1	0.9;
	7023	1	0	0	0	0	0	0.961286	-31.66657	150	9	1.1	0.9;
	7024	1	-0	0	0	1.76	0	1.01541	-32.309824	150	9	1.1	0.9;
	7025	2	0	0	0	0	0	1.027158	8.837805	220	13	1.1	0.9;
	7026	1	0	0	0	51.27	0	1.01494	7.860167	400	1	1.1	0.9;
	7027	2	0	0	0	0	0	1.043049	-17.690411	220	7	1.1	0.9;
	7028	2	0	0	0	0	0	1.031022	15.293375	380	3	1.1	0.9;
	7029	2	0	0	0	0	0	1.058412	-37.465559	380	4	1.1	0.9;
	7030	1	0	0	0	0.19	0	1.055302	-8.059104	220	5	1.1	0.9;
	7031	1	-0	-0	0	42.94	0	1.026374	-33.913457	380	1	1.1	0.9;
	7032	2	0	0	0	0	0	1.004806	-21.757638	154	1	1.1	0.9;
	7033	2	0	0	0	0	0	1.098728	-21.123741	110	15	1.3	0.7;
	7034	1	-0	0	0	0	0	0.988671	-30.990507	150	9	1.1	0.9;
	7035	1	0	-0	0.386679	23.856092	0	1.012093	23.698604	380	3	1.1	0.9;
	7036	2	0	0	0	0	0	1.038394	-12.636984	220	5	1.1	0.9;
	7037	1	-0	-0	0	2.77	0	0.986557	-56.179606	220	10	1.1	0.9;
	7038	1	19.52	6.4	0	0.33	0	1.02997	-34.33488	110	12	1.1	0.9;
	7039	1	324.9	37.3	0	10.15	0	1.025647	-11.475663	380	3	1.1	0.9;
	7040	1	21.82	6.3	0	1.14	0	1.006841	-28.372228	110	18	1.1	0.9;
	7041	1	-175.8	-19.6	0	14.58	0	1.010693	-42.421354	220	10	1.1	0.9;
	7042	1	220.98	91	0	1.16	0	1.050508	-30.568646	220	5	1.1	0.9;
	7043	1	6.91	1.9	0	1.05	0	1.020104	-14.479523	110	18	1.1	0.9;
	7044	1	120.6	35.1	0	20.052	0	1.039381	-33.747307	380	6	1.1	0.9;
	7045	1	1.9	-0	0	1.32	0	1.013374	-19.56497	110	18	1.1	0.9;
	7046	1	0	-0	0	0.41	0	1.09164	-27.608366	150	9	1.1	0.9;
	7047	1	339.9	70.5	0	28.91	0	1.031776	-22.247708	220	5	1.1	0.9;
	7048	1	12.91	5.1	0	0.16	0	1.010337	-24.15527	110	18	1.1	0.9;
	7049	2	0	0	0	0	0	1.052472	-14.395209	220	5	1.1	0.9;
	7050	1	120	26	0	22.01	0	1.034621	-21.547605	220	5	1.1	0.9;
	7051	1	0	0	0	21.19	0	0.992555	-0.609221	380	2	1.1	0.9;
	7052	1	111	36.6	0	2.21	0	1.051113	-12.473183	220	5	1.1	0.9;
	7053	1	-0	0	0	34.13	0	1.062405	-41.501124	380	4	1.1	0.9;
	7054	1	129.7	27.1	0	0.36	0	1.021776	-28.875823	110	19	1.1	0.9;
	7055	1	125.2	8.5	0	0.25	0	1.033958	-13.939042	220	3	1.1	0.9;
	7056	2	0	0	0	0	0	1.000187	-4.542481	380	5	1.1	0.9;
	7057	1	-0	0	0	0.2	0	0.964811	-31.483445	150	9	1.1	0.9;
	7058	2	0	0	0	0	0	1.040913	-36.969182	220	4	1.1	0.9;
	7059	1	28.93	8.3	0	0.42	0	1.023192	8.964208	150	8	1.1	0.9;
	7060	1	0	0	0	4.93	0	1.023557	-7.713513	154	1	1.1	0.9;
	7061	1	0	0	0	0.47	0	1.017417	-15.622114	110	18	1.1	0.9;
	7062	1	-20.12	6.6	0	3.24	0	1.037352	-25.302957	150	9	1.1	0.9;
	7063	2	0	0	0	0	0	1.059581	-15.448599	380	16	1.1	0.9;
	7064	1	-0	0	0	1.02	0	1.010017	-47.239712	220	4	1.1	0.9;
	7065	2	0	0	0	0	0	1.019881	-20.228686	220	20	1.1	0.9;
	7066	1	41.25	48.4	0	16.52	0	1.026491	-40.896042	220	6	1.1	0.9;
	7067	1	6.31	5.1	0	0.08	0	0.954479	-30.312019	110	18	1.1	0.9;
	7068	1	2.8	1	0	0.18	0	1.051283	-21.393887	110	19	1.1	0.9;
	7069	1	258.79	42.82	0	9.28	0	1.034779	-8.197856	220	5	1.1	0.9;
	7070	1	142.9	25.8	0	24.18	0	1.042461	-13.76476	220	5	1.1	0.9;
	7071	1	80.19	23.6	0	19.21	0	1.013438	-36.673949	220	6	1.1	0.9;
	7072	1	62.67	14.6	0	0.73	0	1.015382	-35.580216	110	22	1.1	0.9;
	7073	1	-0	-0	0	15.47	0	1.046708	8.458606	150	8	1.1	0.9;
	7074	1	15.72	9.7	0	1.61	0	1.00388	-31.357499	150	9	1.1	0.9;
	7075	2	0	0	0	0	0	1.019093	-43.133872	220	4	1.1	0.9;
	7076	1	0	-0	0	0.36	0	1.027182	-12.740408	220	5	1.1	0.9;
	7077	1	18.82	6.8	0	2.15	0	1.020593	-28.291209	110	19	1.1	0.9;
	7078	1	-13.82	4.6	0	0.45	0	1.009728	-30.47591	150	9	1.1	0.9;
	7079	1	-0	0	0	2.2	0	1.036534	-51.990596	380	10	1.1	0.9;
	7080	1	-0	-0	0	1.83	0	1.014551	-22.607149	150	9	1.1	0.9;
	7081	1	-0	-0	0	63.75	0	1.064899	-37.690039	380	22	1.1	0.9;
	7082	1	152.2	53.2	0	33.26	0	1.021393	-7.314334	380	3	1.1	0.9;
	7083	1	0	-0	0	27.92	0	1.004805	-4.709685	400	1	1.1	0.9;
	7084	1	236.6	33.7	0	0.14	0	1.006489	-25.146599	154	1	1.1	0.9;
	7085	2	0	0	0	0	0	1.044403	10.521716	220	3	1.1	0.9;
	7086	1	10.81	3.6	0	0.08	0	1.011412	-38.830245	110	12	1.1	0.9;
	7087	1	20.22	7.3	0	0.31	0	1.04466	-23.032906	110	19	1.1	0.9;
	7088	1	-0	-0	0	30.43	0	0.973574	-32.247772	150	9	1.1	0.9;
	7089	2	0	0	0	0	0	1.081026	11.515562	380	13	1.1	0.9;
	7090	1	82.89	41.3	0	36.2	0	1.031643	10.359545	150	8	1.1	0.9;
	7091	1	0	0	0	0.66	0	1.017136	-28.972189	150	9	1.1	0.9;
	7092	1	44.65	8.8	0	20.17	0	1.043734	-25.430923	220	5	1.1	0.9;
	7093	2	0	0	0	0	0	1.03985	35.192342	380	3	1.1	0.9;
	7094	1	22.3	3.1	0	4.33	0	1.015676	0.741492	154	1	1.1	0.9;
	7095	1	-0	-0	0	0.29	0	1.01832	-30.055686	150	9	1.1	0.9;
	7096	1	-0	-0	0	57.66	0	1.066995	-40.951798	380	4	1.1	0.9;
	7097	1	109.5	33.6	0	5.75	0	1.014749	6.93111	220	3	1.1	0.9;
	7098	1	0	-0	0	7.17	0	1.036326	-2.652895	220	5	1.1	0.9;
	7099	1	102	26.6	0	2.27	0	1.015959	36.786282	220	3	1.1	0.9;
	7100	1	35.94	22.3	0	1.63	0	0.973184	-34.115427	150	9	1.1	0.9;
	7101	1	4.61	1.7	0	0.34	0	1.028926	-28.291929	110	19	1.1	0.9;
	7102	1	-0	0	0.185509	0.457193	0	1.02884	1.689902	380	3	1.1	0.9;
	7103	1	0	0	0	16.39	0	1.016158	-28.311302	220	6	1.1	0.9;
	7104	2	0	0	0	0	0	1.002634	-29.070294	220	10	1.1	0.9;
	7105	1	0	0	0	7.46	0	1.094875	-31.536519	380	12	1.1	0.9;
	7106	1	-0	-0	0	0.11	0	1.070254	-26.565172	150	9	1.1	0.9;
	7107	2	0	0	0	0	0	1.040026	-5.420572	220	14	1.1	0.9;
	7108	1	282.5	56.2	0	8.6	0	1.016025	-41.954894	380	6	1.1	0.9;
	7109	1	-0	0	0	1.13	0	0.990039	-17.90541	110	18	1.1	0.9;
	7110	1	-0	0	0	0	0	1.021163	-42.21352	220	4	1.1	0.9;
	7111	1	0	-0	0	1.49	0	1.056462	-38.889606	110	22	1.1	0.9;
	7112	1	0.1	-0	0	0.39	0	1.055047	-26.127795	110	19	1.1	0.9;
	7113	2	0	0	0	0	0	1.033847	-7.032003	154	1	1.1	0.9;
	7114	2	0	0	0	0	0	1.083134	4.617435	380	14	1.1	0.9;
	7115	2	0	0	0	0	0	1.026738	-15.536813	220	5	1.1	0.9;
	7116	1	117.5	91	0.373436	45.743751	0	1.016276	1.615489	380	3	1.1	0.9;
	7117	1	0	-0.6	0	0.5	0	1.03102	26.404933	110	8	1.1	0.9;
	7118	1	0	0	0	39.83	0	1.108858	-21.938093	380	1	1.3	0.7;
	7119	1	149.53	34.9	0	8.9	0	1.009263	-16.529806	220	5	1.1	0.9;
	7120	1	0	-0	0	6.17	0	1.065996	-12.920134	220	18	1.1	0.9;
	7121	1	29.43	10.7	0	0.99	0	1.016626	-28.916482	110	19	1.1	0.9;
	7122	1	0	-0	0	13.03	0	1.14349	-21.530056	380	9	1.3	0.7;
	7123	1	38.34	12.5	0	1.59	0	1.010031	-36.741184	220	6	1.1	0.9;
	7124	1	-0	0	0	16.73	0	1.068939	-27.474719	220	5	1.1	0.9;
	7125	1	0.1	2.2	0	1.22	0	1.005032	11.493321	220	3	1.1	0.9;
	7126	1	5.71	3	0	0.18	0	1.008391	-24.561842	110	18	1.1	0.9;
	7127	1	231.2	103.5	0	10.91	0	1.012402	-17.910043	220	3	1.1	0.9;
	7128	2	0	0	0	0	0	1.007918	17.912282	400	1	1.1	0.9;
	7129	1	33.34	5	0	0.38	0	1.033259	-3.259642	220	5	1.1	0.9;
	7130	1	19.32	7	0	0.65	0	1.022819	-24.428657	110	19	1.1	0.9;
	7131	1	-4.73	4.23	0	0.42	0	1.056852	-23.705839	110	23	1.1	0.9;
	7132	1	222	90.5	0	0.16	0	1.025359	-15.656092	220	5	1.1	0.9;
	7133	1	102.3	19	0	1.84	0	1.043055	-14.04535	220	5	1.1	0.9;
	7134	2	0	0	0	0	0	1.039671	11.116123	154	1	1.1	0.9;
	7135	1	0	0	0.321419	66.618267	0	1.026445	-11.206428	380	3	1.1	0.9;
	7136	1	29.23	9.4	0	0.68	0	1.027007	7.151388	110	8	1.1	0.9;
	7137	2	0	0	0	0	0	1.024472	-45.229123	380	4	1.1	0.9;
	7138	1	150.2	51.6	0	2.79	0	1.015276	-4.651351	150	8	1.1	0.9;
	7139	1	99.4	22.2	0	2.16	0	1.026773	-54.251658	220	4	1.1	0.9;
	7140	1	51	7.3	0	6.06	0	1.01882	1.935449	154	1	1.1	0.9;
	7141	1	64.2	26.4	0	1.05	0	1.044296	-28.134322	110	23	1.1	0.9;
	7142	1	32.34	16.8	0	1.42	0	0.986331	-28.704613	110	18	1.1	0.9;
	7143	1	0	-0	0	0.39	0	0.967584	-32.441251	150	9	1.1	0.9;
	7144	1	8.48	-1.82	0	0.37	0	1.024916	-13.312237	220	5	1.1	0.9;
	7145	1	46.75	24.4	0	3.66	0	1.011612	-58.744613	220	4	1.1	0.9;
	7146	1	0	0	0	0	0	1.079481	31.593816	380	1	1.1	0.9;
	7147	1	41.8	5.9	0	7.64	0	1.033187	3.270759	154	1	1.1	0.9;
	7148	1	-0	-0	0	10.89	0	1.046487	-23.666485	220	5	1.1	0.9;
	7149	1	6.01	2	0	0.17	0	1.053577	-17.345053	110	21	1.1	0.9;
	7150	1	26.03	-14.1	0	0.84	0	1.05817	-24.085398	110	23	1.1	0.9;
	7151	1	-0	0	0	0.36	0	0.970303	-28.709834	110	18	1.1	0.9;
	7152	1	89.8	19.7	0	0.36	0	1.042145	-28.526496	120	20	1.1	0.9;
	7153	2	0	0	0	0	0	1.034539	-24.674893	380	7	1.1	0.9;
	7154	1	73.58	26.8	0	3.76	0	1.035929	-3.51424	220	1	1.1	0.9;
	7155	1	-8.01	-3.24	0	0.32	0	1.00228	-31.656527	110	23	1.1	0.9;
	7156	1	-0	-0	0	14.82	0	1.047271	8.411476	150	8	1.1	0.9;
	7157	1	0	-0	0	10.34	0	1.027513	5.388751	400	1	1.1	0.9;
	7158	1	-0	0	0	40.44	0	0.999862	6.947454	400	1	1.1	0.9;
	7159	2	0	0	0	0	0	1.055132	4.464452	380	5	1.1	0.9;
	7160	1	0	-0	0	9.72	0	1.023736	3.817974	380	3	1.1	0.9;
	7161	1	122.8	20.4	0	0.77	0	0.997387	9.734185	220	3	1.1	0.9;
	7162	1	-0	0	0	2.48	0	1.042623	-13.767265	220	5	1.1	0.9;
	7163	1	-0	0	0	2.9	0	1.045634	-25.197	220	5	1.1	0.9;
	7164	1	-18.6	-6.72	0	152.683271	0	1.024409	-22.545445	380	5	1.1	0.9;
	7165	1	-0	-0	0	9.26	0	1.066671	-34.484319	220	5	1.1	0.9;
	7166	2	0	0	0	0	0	1.002006	-26.80616	110	18	1.1	0.9;
	7167	1	64.47	40	0	5.82	0	1.002622	-35.62321	220	6	1.1	0.9;
	7168	1	-20.82	6.8	0	0.91	0	1.008085	-23.034104	150	9	1.1	0.9;
	7169	2	0	0	0	0	0	1.087809	42.474107	380	3	1.1	0.9;
	7170	1	-1083.9	-433.7	0	2.72	0	1.053025	-6.907386	380	3	1.1	0.9;
	7171	1	0.1	-9.6	0	1.28	0	1.04536	-20.925425	220	3	1.1	0.9;
	7172	1	0	-0	0	0	0	0.991117	-30.792246	150	9	1.1	0.9;
	7173	1	51.7	7.4	0	2.07	0	1.027247	4.050137	154	1	1.1	0.9;
	7174	1	77.91	5.03	0	19.71	0	1.043805	1.379832	220	13	1.1	0.9;
	7175	1	-0	0	0	10.22	0	1.014832	23.048441	380	3	1.1	0.9;
	7176	1	10.21	3.3	0	0.62	0	1.015634	-35.212323	110	12	1.1	0.9;
	7177	1	17.62	7.7	0	0.59	0	1.017537	-38.058658	110	22	1.1	0.9;
	7178	1	-0	-0	0	2.45	0	1.01172	-14.547245	220	5	1.1	0.9;
	7179	1	47	6.7	0	6.12	0	1.020654	-4.830216	154	1	1.1	0.9;
	7180	1	-0	0	0.234234	4.220723	0	0.976049	-7.262397	380	2	1.1	0.9;
	7181	1	1	0.2	0	0.52	0	1.056719	-25.761257	110	19	1.1	0.9;
	7182	1	290.8	30.5	0	29.529841	0	1.016598	-43.031895	380	6	1.1	0.9;
	7183	2	0	0	0	0	0	1.104721	-7.066134	380	5	1.3	0.7;
	7184	1	16.7	2.4	0	3.56	0	1.029441	-0.041577	154	1	1.1	0.9;
	7185	1	-18.52	6.1	0	1.01	0	1.015765	-28.366088	150	9	1.1	0.9;
	7186	1	4.75	-0.16	0	0.61	0	1.011199	-24.534116	110	18	1.1	0.9;
	7187	1	7.21	2.4	0	1.67	0	1.007999	-40.530577	110	12	1.1	0.9;
	7188	1	25.1	3.6	0	6.2	0	1.016588	-6.620753	154	1	1.1	0.9;
	7189	1	-0	0	0	0.5	0	1.028643	-16.495545	220	3	1.1	0.9;
	7190	1	-0	-0	0	0.11	0	0.982642	-31.428797	150	9	1.1	0.9;
	7191	1	55.46	15.7	0	13.46	0	1.021115	-48.781587	220	4	1.1	0.9;
	7192	1	0	0	0	0	0	0.979313	-27.16062	150	9	1.1	0.9;
	7193	1	-0	-0	0	7.17	0	1.015962	-47.157967	220	4	1.1	0.9;
	7194	1	456.6	227.7	0	0.61	0	1.006994	51.272937	220	3	1.1	0.9;
	7195	1	99.9	19.9	0	6.19	0	1.018676	0.762842	150	8	1.1	0.9;
	7196	1	72.68	1.1	0	6.06	0	1.022369	-55.130527	220	4	1.1	0.9;
	7197	1	-0	0	0.08	4.38	0	1.04408	-43.548131	380	6	1.1	0.9;
	7198	1	83.19	21.8	0	3.96	0	1.041752	-3.275387	220	3	1.1	0.9;
	7199	1	6.31	1.3	0	0.52	0	0.997694	-26.738112	110	18	1.1	0.9;
	7200	2	0	0	0	0	0	1.022531	-29.01631	110	21	1.1	0.9;
	7201	1	8.11	3.8	0	0.26	0	0.964669	-24.813208	110	18	1.1	0.9;
	7202	1	18.62	6.1	0	2	0	1.048931	-4.187607	220	5	1.1	0.9;
	7203	1	43.25	15.7	0	0.37	0	1.031691	-26.556416	110	19	1.1	0.9;
	7204	1	-241.32	-48.72	0	16.62	0	1.074559	-25.861826	220	23	1.1	0.9;
	7205	1	48.05	23	0	0.78	0	1.004397	-30.090261	110	23	1.1	0.9;
	7206	1	8.3	1.2	0	3.78	0	1.013173	11.41942	154	1	1.1	0.9;
	7207	2	0	0	0	0	0	1.039936	35.201132	380	3	1.1	0.9;
	7208	1	46.55	15.5	0	0.4	0	0.974903	-35.35308	110	21	1.1	0.9;
	7209	2	0	0	0	0	0	1.049275	-6.593882	220	5	1.1	0.9;
	7210	1	-0	0	0	0.44	0	1.068185	-31.443276	110	3	1.1	0.9;
	7211	1	6.61	1.6	0	2.13	0	0.988078	-20.950879	110	18	1.1	0.9;
	7212	2	0	0	0	0	0	1.021331	-44.432519	380	6	1.1	0.9;
	7213	1	20.9	3	0	0.17	0	1.022468	-7.333554	154	1	1.1	0.9;
	7214	2	0	0	0	0	0	1.0752	8.895766	380	14	1.1	0.9;
	7215	1	19.22	3.8	0	0.31	0	1.019269	-14.406567	110	18	1.1	0.9;
	7216	1	-0	0	0	0.68	0	0.935606	-33.742689	110	18	1.1	0.9;
	7217	1	-0	-0	0	3.98	0	1.02958	21.227948	154	1	1.1	0.9;
	7218	1	0	-0	0	6.22	0	1.024869	-9.306127	220	3	1.1	0.9;
	7219	1	89.7	40.5	0	38.13	0	1.02589	-22.734407	220	7	1.1	0.9;
	7220	1	-0	0.2	0	2.19	0	1.009426	-4.732275	220	3	1.1	0.9;
	7221	1	17.72	6.3	0	0.33	0	0.923556	-33.061189	110	18	1.1	0.9;
	7222	1	0	-0	0	5	0	1.062175	-24.161142	220	5	1.1	0.9;
	7223	2	0	0	0	0	0	1.018777	2.39191	154	1	1.1	0.9;
	7224	1	0	-0	0	0.37	0	1.003808	-27.599126	110	18	1.1	0.9;
	7225	1	-0	0	0.181549	13.990889	0	1.045374	-22.979064	220	18	1.1	0.9;
	7226	1	144	31.3	0	3.18	0	1.016944	-15.886541	220	5	1.1	0.9;
	7227	2	0	0	0	0	0	1.008346	1.023457	400	1	1.1	0.9;
	7228	2	0	0	0	0	0	1.172784	-19.671404	380	9	1.3	0.7;
	7229	1	0	-0	0	0.62	0	0.96078	-31.700326	150	9	1.1	0.9;
	7230	1	-0.18	-1	0	0	0	1.021156	10.861291	150	8	1.1	0.9;
	7231	1	0	0	0	0.05	0	1.035344	-11.772407	220	5	1.1	0.9;
	7232	1	7.61	2.5	0	0.93	0	0.994308	-31.830644	110	18	1.1	0.9;
	7233	1	-0	0	0	21.62	0	1.021669	-32.12307	150	9	1.1	0.9;
	7234	1	388.7	97.7	0	41.94	0	1.018027	-42.307076	380	6	1.1	0.9;
	7235	1	0	0	0	3.6	0	1.051434	38.85912	380	8	1.1	0.9;
	7236	1	0	0	0	1.99	0	0.962801	-44.351836	220	10	1.1	0.9;
	7237	1	0	0	0	0	0	0.974217	-31.060355	150	9	1.1	0.9;
	7238	2	0	0	0	0	0	1.019781	-12.721337	154	1	1.1	0.9;
	7239	1	200.4	115.2	0	6.18	0	1.004347	-30.856234	220	7	1.1	0.9;
	7240	1	0	-0	0	2.77	0	1.033934	47.079127	220	3	1.1	0.9;
	7241	1	11.21	3.7	0	0.92	0	1.010777	-39.215522	110	12	1.1	0.9;
	7242	1	3.4	1.2	0	0.19	0	1.021069	-28.41258	110	19	1.1	0.9;
	7243	1	-0	-0	0	0.76	0	0.981044	-32.367839	110	21	1.1	0.9;
	7244	1	16.7	2.4	0	1.21	0	1.008394	-3.368331	154	1	1.1	0.9;
	7245	1	-0	-0	0	1.33	0	0.982593	9.30718	220	3	1.1	0.9;
	7246	1	7.41	2.7	0	0.61	0	1.024264	-25.998758	110	19	1.1	0.9;
	7247	1	-0	-0	0	2.37	0	1.053601	-28.968049	150	9	1.1	0.9;
	7248	1	233.2	39.4	0	104.569841	0	1.024517	-30.84311	380	6	1.1	0.9;
	7249	1	37.94	9.8	0	2.26	0	1.028006	-42.600996	220	4	1.1	0.9;
	7250	2	0	0	0	0	0	1.041313	-25.406873	110	19	1.1	0.9;
	7251	1	0	0	0	0	0	0.995125	-30.856007	150	9	1.1	0.9;
	7252	1	220.8	68.6	0	1.97	0	0.961795	-44.756426	220	10	1.1	0.9;
	7253	1	7.51	3.1	0	0	0	1.035374	-3.050273	220	5	1.1	0.9;
	7254	1	0	-0	0	0.06	0	1.043589	-20.621975	220	3	1.1	0.9;
	7255	1	4.61	1.5	0	1.23	0	1.009289	-40.079993	110	12	1.1	0.9;
	7256	1	392.25	92.02	0	9.45	0	1.007111	-34.649721	220	5	1.1	0.9;
	7257	2	0	0	0	0	0	1.023182	-46.899294	380	4	1.1	0.9;
	7258	1	10.61	3.5	0	1.3	0	1.006474	-36.041466	110	12	1.1	0.9;
	7259	1	82.39	0	0	0.15	0	1.055097	-8.128627	220	5	1.1	0.9;
	7260	2	0	0	0	0	0	1.020734	18.488306	150	8	1.1	0.9;
	7261	1	96.7	23.7	0	39.1	0	1.050327	30.877519	380	3	1.1	0.9;
	7262	1	-0	-0	0	35.98	0	1.071381	-33.363949	380	12	1.1	0.9;
	7263	1	185	26.3	0	11.78	0	1.017688	-22.517768	154	1	1.1	0.9;
	7264	1	0	-0	0	3.42	0	1.037453	-11.256183	220	5	1.1	0.9;
	7265	1	7.61	3.8	0	0.82	0	1.018104	-20.137	110	18	1.1	0.9;
	7266	1	-0	-0	0	2.12	0	1.02803	-5.817903	220	5	1.1	0.9;
	7267	2	0	0	0	0	0	1.006635	-7.962981	380	5	1.1	0.9;
	7268	1	406.3	92.5	0	66.92	0	1.029606	-15.712052	380	6	1.1	0.9;
	7269	1	0	0	0	0.02	0	1.018141	-32.156086	150	9	1.1	0.9;
	7270	2	0	0	0	0	0	1.063771	26.076966	380	8	1.1	0.9;
	7271	1	-0	0	0.090774	6.600444	0	1.050328	-24.814128	220	18	1.1	0.9;
	7272	1	18.62	6.7	0	0.83	0	1.044203	-26.082613	110	15	1.1	0.9;
	7273	1	219.6	36.4	0	17.89	0	1.038052	-24.670916	220	5	1.1	0.9;
	7274	1	54.93	16.5	0	6.75	0	1.03186	-4.222416	220	5	1.1	0.9;
	7275	1	79.39	31.3	0	2.23	0	1.044197	3.98215	150	8	1.1	0.9;
	7276	1	-0	-0	0	8.38	0	1.042871	-21.734523	220	1	1.1	0.9;
	7277	1	0	0	0	10.8	0	1.02596	-22.59388	154	1	1.1	0.9;
	7278	2	0	0	0	0	0	1.063694	-15.573936	220	18	1.1	0.9;
	7279	2	0	0	0	0	0	1.02664	-13.635302	150	2	1.1	0.9;
	7280	1	7.41	2.7	0	0.56	0	1.035372	-21.48109	110	19	1.1	0.9;
	7281	2	0	0	0	0	0	1.020691	-24.050831	154	1	1.1	0.9;
	7282	2	0	0	0	0	0	1.029197	-4.82313	380	5	1.1	0.9;
	7283	1	101.1	6.6	0	5.66	0	1.03338	-46.106964	220	4	1.1	0.9;
	7284	1	0	0	0	0	0	1.085432	-1.082662	220	5	1.1	0.9;
	7285	1	0	-0	0	20.08	0	1.108938	-29.514945	380	12	1.3	0.7;
	7286	1	11.91	4.4	0	0.16	0	1.056629	-24.199026	110	15	1.1	0.9;
	7287	1	0	-0	0	46.28	0	1.060667	-41.565457	380	4	1.1	0.9;
	7288	1	0	0	0	4.11	0	1.012472	-25.361801	154	1	1.1	0.9;
	7289	1	0	-0	0	26.5	0	1.049679	-6.239897	220	5	1.1	0.9;
	7290	1	0	-0	0	21.32	0	1.016232	-32.6798	150	9	1.1	0.9;
	7291	1	0	0	0	15.49	0	1.0255	-11.196419	380	1	1.1	0.9;
	7292	1	11.71	3.8	0	2.39	0	1.028167	-34.087434	110	12	1.1	0.9;
	7293	2	0	0	0	0	0	1.029739	-9.45721	220	13	1.1	0.9;
	7294	1	2.7	0.8	0	1.09	0	0.991451	-27.535638	110	18	1.1	0.9;
	7295	1	-0	0	0	2.63	0	1.040482	2.636868	150	8	1.1	0.9;
	7296	1	36.84	13.3	0	1.05	0	1.0138	-17.179093	110	19	1.1	0.9;
	7297	1	-0	0	0	2.67	0	1.040466	2.636616	150	8	1.1	0.9;
	7298	1	97.41	-4.48	0	16.55	0	1.03893	9.638331	220	13	1.1	0.9;
	7299	1	0	-0	0	1.42	0	1.015409	52.961176	220	3	1.1	0.9;
	7300	1	0	-0	0	0	0	0.994058	-30.578514	150	9	1.1	0.9;
	7301	1	36.04	12	0	0.11	0	0.984897	-31.81215	110	21	1.1	0.9;
	7302	1	0	-0	0	2.25	0	1.023792	-1.567692	154	1	1.1	0.9;
	7303	2	0	0	0	0	0	1.026431	-43.382912	380	4	1.1	0.9;
	7304	1	0.4	0.1	0	1.39	0	1.01082	-27.278081	110	18	1.1	0.9;
	7305	1	82.89	32.6	0	45.84	0	1.044909	3.479513	150	8	1.1	0.9;
	7306	1	-0	0	0	5.38	0	1.00785	-32.36366	220	6	1.1	0.9;
	7307	1	0	0	0	0	0	0.981231	-35.406383	150	9	1.1	0.9;
	7308	1	-9.86	-2.8	0	0.2	0	0.95998	-22.856198	110	18	1.1	0.9;
	7309	1	0	-0	0	1.32	0	1.05059	-11.653422	220	5	1.1	0.9;
	7310	1	256	69.9	0	43.56	0	1.022253	-46.41738	380	4	1.1	0.9;
	7311	1	102.6	12	0	66.849841	0	1.021381	-48.724733	380	6	1.1	0.9;
	7312	1	56.66	15.6	0	7.45	0	1.002377	-48.515496	220	4	1.1	0.9;
	7313	1	0	0	0	16.07	0	1.143494	-24.568389	380	23	1.3	0.7;
	7314	1	0	0	0.12635	23.056325	0	1.130799	-25.608096	380	21	1.3	0.7;
	7315	1	29.53	2.4	0	0.35	0	1.054837	-32.593249	110	3	1.1	0.9;
	7316	1	-0	0	0	0.09	0	1.043172	-7.351483	220	5	1.1	0.9;
	7317	1	210.9	48.3	0	9.75	0	1.036146	-3.693886	220	1	1.1	0.9;
	7318	1	0	0	0	2.1	0	1.023928	19.805822	380	3	1.1	0.9;
	7319	1	0	0	0	0.81	0	1.014013	-26.553685	150	9	1.1	0.9;
	7320	1	0	0	0	0.15	0	1.014993	-36.020364	220	6	1.1	0.9;
	7321	1	201.4	47.5	0	0	0	1.016304	20.570874	220	3	1.1	0.9;
	7322	1	36.94	62.6	0	0.12	0	0.985729	-31.284095	150	9	1.1	0.9;
	7323	1	18.22	6.3	0	0.48	0	1.002613	-29.942299	110	21	1.1	0.9;
	7324	1	0	0	0	2.33	0	1.025703	-5.011162	220	3	1.1	0.9;
	7325	1	0	0	0	0.41	0	1.042768	-13.023775	220	5	1.1	0.9;
	7326	1	69.7	9.9	0	1.52	0	0.998326	-25.231875	154	1	1.1	0.9;
	7327	2	0	0	0	0	0	1.046207	-18.098445	220	5	1.1	0.9;
	7328	2	0	0	0	0	0	0.98954	-14.833223	380	5	1.1	0.9;
	7329	2	0	0	0	0	0	1.037608	-34.291657	220	10	1.1	0.9;
	7330	2	0	0	0	0	0	1.166291	-21.054036	380	15	1.3	0.7;
	7331	2	0	0	0	0	0	1.064241	-37.878234	110	22	1.1	0.9;
	7332	1	-0	0	0	11.4	0	1.018065	-3.120514	220	3	1.1	0.9;
	7333	1	36.54	22.6	0	0.04	0	1.036075	-29.78924	150	9	1.1	0.9;
	7334	1	26.43	9.9	0	9.57	0	1.027071	-39.633663	110	22	1.1	0.9;
	7335	1	25.13	14	0	0.04	0	1.02041	-5.272859	220	3	1.1	0.9;
	7336	1	45.75	15.8	0	3.42	0	0.992771	-17.685822	110	19	1.1	0.9;
	7337	2	0	0	0	0	0	1.015107	-6.438735	154	1	1.1	0.9;
	7338	1	0	-0	0	2.98	0	1.042914	-14.841123	220	5	1.1	0.9;
	7339	1	0	-0	0	8.41	0	0.997809	-30.283644	150	9	1.1	0.9;
	7340	1	23.83	8.6	0	3.08	0	1.019852	-29.550767	110	19	1.1	0.9;
	7341	1	37.44	0	0	0.67	0	1.036058	-2.795886	220	5	1.1	0.9;
	7342	1	169	70.8	0	5.23	0	1.014786	-16.173977	220	5	1.1	0.9;
	7343	1	-0	0	0	61.36	0	1.026913	-43.465223	380	4	1.1	0.9;
	7344	1	31.03	8.9	0	0.49	0	1.002717	-30.609813	110	18	1.1	0.9;
	7345	1	0	0	0	0.61	0	0.978292	-32.010355	110	18	1.1	0.9;
	7346	1	-0	0	0	17.3	0	1.044003	14.367709	380	3	1.1	0.9;
	7347	1	101.3	17.6	0	0.12	0	1.008541	-48.920169	220	4	1.1	0.9;
	7348	1	16.52	6.3	0	0.23	0	1.084423	-16.467495	110	18	1.3	0.7;
	7349	1	75.56	13.04	0	8.77	0	1.014685	-51.502231	220	4	1.1	0.9;
	7350	2	0	0	0	0	0	1.027675	6.194663	154	1	1.1	0.9;
	7351	1	61.67	16.3	0	0.05	0	1.033161	-3.263056	220	5	1.1	0.9;
	7352	1	0	-0	0	4.49	0	1.011167	0.526638	380	3	1.1	0.9;
	7353	1	127.3	26.7	0	7.02	0	0.985079	-22.561059	220	5	1.1	0.9;
	7354	1	28.53	11.5	0	0.72	0	1.003437	-29.755843	110	21	1.1	0.9;
	7355	1	0	0	0	1.72	0	1.040333	8.324539	154	1	1.1	0.9;
	7356	1	12.71	4.2	0	0.65	0	0.980107	-29.959298	110	18	1.1	0.9;
	7357	1	0	-0	0	3.77	0	1.026055	-25.145654	220	1	1.1	0.9;
	7358	1	-40.95	-28.4	0	7.29	0	0.996928	-48.829455	150	10	1.1	0.9;
	7359	2	0	0	0	0	0	1.046487	-26.102165	380	6	1.1	0.9;
	7360	1	88.2	21	0	2.22	0	0.976728	13.47457	220	3	1.1	0.9;
	7361	1	-0	-0	0	14.29	0	1.051981	-15.532005	220	5	1.1	0.9;
	7362	1	223.31	61.3	0	4.51	0	1.032049	-3.472036	220	1	1.1	0.9;
	7363	2	0	0	0	0	0	1.005537	-25.548958	154	1	1.1	0.9;
	7364	1	-0	-0	0.23465	19.69065	0	1.115295	-23.60728	380	21	1.3	0.7;
	7365	1	6.31	1.3	0	0.12	0	0.985997	-27.51912	110	18	1.1	0.9;
	7366	1	70.48	22.5	0	4.82	0	1.012248	-40.784255	220	4	1.1	0.9;
	7367	1	1.5	0.4	0	0.15	0	1.027876	-18.951852	220	5	1.1	0.9;
	7368	1	-0	-0	0	0.6	0	0.968835	-31.329943	150	9	1.1	0.9;
	7369	1	-0	0	0	7.35	0	1.014772	17.844108	380	1	1.1	0.9;
	7370	2	0	0	0	0	0	1.034338	7.51478	154	1	1.1	0.9;
	7371	1	34.9	5	0	2.95	0	1.010706	4.159172	154	1	1.1	0.9;
	7372	1	10.01	4.3	0	5.64	0	1.004081	-19.703313	110	18	1.1	0.9;
	7373	1	0	-0	0	1.81	0	1.066452	-23.971054	220	5	1.1	0.9;
	7374	1	-0.2	-0.83	0	0	0	1.020851	-1.172884	150	8	1.1	0.9;
	7375	1	53.06	19.2	0	0.41	0	1.047027	-25.519254	110	19	1.1	0.9;
	7376	2	0	0	0	0	0	1.041198	3.572761	220	14	1.1	0.9;
	7377	1	0	0	0	0.63	0	1.012383	-16.331531	220	5	1.1	0.9;
	7378	1	-0	0	0	14.06	0	0.994011	-9.356245	400	1	1.1	0.9;
	7379	1	41.5	5.9	0	8.21	0	1.017015	2.059734	154	1	1.1	0.9;
	7380	1	105.67	0.56	0	18.05	0	0.988082	-14.843414	380	5	1.1	0.9;
	7381	1	-0	-0	0	0.28	0	0.998683	-32.343314	150	9	1.1	0.9;
	7382	1	0	0	0	0	0	0.945608	-29.541812	150	9	1.1	0.9;
	7383	2	0	0	0	0	0	1.011076	-36.03631	220	22	1.1	0.9;
	7384	1	6.11	2.2	0	0.88	0	0.990709	-21.745983	110	19	1.1	0.9;
	7385	1	1.1	0	0	1.53	0	1.016823	-40.623861	110	22	1.1	0.9;
	7386	2	0	0	0	0	0	1.023477	6.232212	154	1	1.1	0.9;
	7387	1	-0	-0	0	1.68	0	0.981743	-31.650542	150	9	1.1	0.9;
	7388	1	0	0	0	12.79	0	1.079481	31.593816	380	8	1.1	0.9;
	7389	1	0	-0	0	4.09	0	1.014668	-32.548395	150	9	1.1	0.9;
	7390	1	16.62	5.5	0	1.68	0	1.047851	-35.726585	110	12	1.1	0.9;
	7391	1	-23.83	7.8	0	2.3	0	1.033694	-21.446299	150	9	1.1	0.9;
	7392	1	142	27.8	0	17.77	0	1.008915	-55.287574	220	4	1.1	0.9;
	7393	2	0	0	0	0	0	1.057049	-24.684949	220	11	1.1	0.9;
	7394	1	55.66	22.4	0	2.11	0	1.043868	8.975598	110	8	1.1	0.9;
	7395	1	0	-0	0	87.69	0	1.019928	-44.315775	380	4	1.1	0.9;
	7396	1	-0	0	0	5.68	0	1.016684	-19.679601	220	5	1.1	0.9;
	7397	1	41.45	17.3	0	0.93	0	0.981241	-27.966079	110	18	1.1	0.9;
	7398	1	100.3	17.2	0	6.63	0	1.013724	-50.85172	220	4	1.1	0.9;
	7399	1	168.9	113	0	123.25	0	1.021627	16.485616	380	3	1.1	0.9;
	7400	1	30.03	18.6	0	0.47	0	1.014538	-27.099328	150	9	1.1	0.9;
	7401	2	0	0	0	0	0	1.018508	9.298494	220	3	1.1	0.9;
	7402	1	0	0	0	0	0	1.015015	-30.091097	150	9	1.1	0.9;
	7403	1	59.67	6	0	3.6	0	1.003818	-50.897179	220	6	1.1	0.9;
	7404	1	19.02	5.4	0	0.63	0	1.058779	10.889206	110	8	1.1	0.9;
	7405	1	0	-0.8	0	0.39	0	1.034205	-23.672695	110	3	1.1	0.9;
	7406	1	94.5	-27.8	0	5.52	0	1.023941	-47.986677	220	4	1.1	0.9;
	7407	1	-0.2	-1	0	0	0	1.020293	-4.164944	150	8	1.1	0.9;
	7408	2	0	0	0	0	0	1.010858	-9.720144	400	1	1.1	0.9;
	7409	1	0	-0	0	1.03	0	0.866113	-38.811746	150	9	1.3	0.7;
	7410	1	107.2	45.6	0	5.28	0	1.00468	-18.331887	220	3	1.1	0.9;
	7411	1	-0	0	0	0.68	0	0.999197	-39.199433	150	10	1.1	0.9;
	7412	1	-0	-0	0	1.49	0	1.047787	63.702085	220	3	1.1	0.9;
	7413	1	33.64	13.7	0	0.75	0	1.027443	-54.133826	220	4	1.1	0.9;
	7414	1	23.43	8.5	0	0.28	0	1.017764	-28.743126	110	21	1.1	0.9;
	7415	1	16.22	5	0	0.18	0	0.992829	-25.884348	110	18	1.1	0.9;
	7416	1	-0	0	0.153313	36.266993	0	1.034259	66.585489	380	1	1.1	0.9;
	7417	1	106.4	19	0	4.65	0	1.034614	-46.039871	220	4	1.1	0.9;
	7418	1	0	0	0.099436	90.374061	0	1.082134	-20.046836	380	18	1.1	0.9;
	7419	1	18.02	12.2	0	0.53	0	0.964639	-24.814322	110	18	1.1	0.9;
	7420	1	415.5	103.4	0	37.13	0	1.026099	-33.962005	380	6	1.1	0.9;
	7421	1	8.11	2.7	0	2.03	0	1.030592	-34.106483	110	12	1.1	0.9;
	7422	1	65.67	25.6	0	2.34	0	1.030703	-10.888099	220	5	1.1	0.9;
	7423	1	-0	0	0	7.3	0	1.017598	-40.734394	220	4	1.1	0.9;
	7424	1	123.1	20.9	0	25.95	0	1.013443	-12.494664	380	3	1.1	0.9;
	7425	1	-0.2	-1	0	0	0	1.038275	-0.668044	150	8	1.1	0.9;
	7426	1	1.8	0.6	0	0.79	0	1.012408	-19.2322	110	18	1.1	0.9;
	7427	2	0	0	0	0	0	1.072596	-22.451974	380	9	1.1	0.9;
	7428	1	0	-0	0	120.17	0	1.043312	-50.518322	380	4	1.1	0.9;
	7429	2	0	0	0	0	0	1.011339	-5.731264	400	1	1.1	0.9;
	7430	1	60.37	54.7	0	0.62	0	0.992466	9.075547	220	3	1.1	0.9;
	7431	2	0	0	0	0	0	1.016271	-33.87332	220	4	1.1	0.9;
	7432	2	0	0	0	0	0	1.040989	-21.158174	220	19	1.1	0.9;
	7433	1	116.8	59	0	1.63	0	0.981792	-34.919816	150	9	1.1	0.9;
	7434	1	0	0	0	13.33	0	1.048635	1.215771	400	1	1.1	0.9;
	7435	1	0	-0	0	0.46	0	0.983989	-31.364669	150	9	1.1	0.9;
	7436	1	0	-0	0	19.61	0	1.019386	-50.516499	380	4	1.1	0.9;
	7437	1	18.32	5.3	0	7.32	0	1.039177	-25.992105	220	5	1.1	0.9;
	7438	1	121.3	20.7	0	2.39	0	1.030599	-13.038703	220	5	1.1	0.9;
	7439	1	12.71	2.5	0	0.56	0	0.959246	-30.912503	110	18	1.1	0.9;
	7440	1	0	-0	0	22.76	0	1.023632	31.656073	380	3	1.1	0.9;
	7441	1	5.44	2.14	0	0.93	0	1.021033	-20.073125	110	18	1.1	0.9;
	7442	1	-0	0	0	36.77	0	1.015133	16.414535	150	8	1.1	0.9;
	7443	1	0	0	0	17.05	0	1.044874	14.986516	380	3	1.1	0.9;
	7444	1	111.9	118.3	0	11.48	0	1.038851	11.46126	380	3	1.1	0.9;
	7445	1	-0	0	0	3.24	0	1.026914	49.225874	220	3	1.1	0.9;
	7446	2	0	0	0	0	0	1.008326	-46.24313	220	4	1.1	0.9;
	7447	1	34.9	5	0	3.05	0	1.018055	-8.026783	154	1	1.1	0.9;
	7448	1	-0	0	0	92.73	0	1.059456	-15.726807	380	16	1.1	0.9;
	7449	1	15.96	2.18	0	3.58	0	1.017627	61.478443	150	1	1.1	0.9;
	7450	1	108.8	28.1	0	30.9	0	1.021344	10.842619	150	8	1.1	0.9;
	7451	1	227.49	61.6	0	13.7	0	0.996166	9.117017	220	3	1.1	0.9;
	7452	1	49.45	16.5	0	4.13	0	1.04041	-39.709614	110	22	1.1	0.9;
	7453	1	-0	0	0	1.35	0	0.999445	-32.159093	150	9	1.1	0.9;
	7454	1	-0	-0	0	3.49	0	1.046271	-8.202077	220	3	1.1	0.9;
	7455	1	-0	0	0.135493	90.112445	0	1.086657	-14.067392	380	19	1.3	0.7;
	7456	1	9.01	3.3	0	0.62	0	1.076625	-17.396277	110	19	1.1	0.9;
	7457	1	96.2	23.3	0	2.05	0	1.042806	-3.068932	220	3	1.1	0.9;
	7458	1	0	-0	0	83.185177	0	1.057103	-22.285226	380	20	1.1	0.9;
	7459	1	-7.41	-1.6	0	2	0	1.037978	-21.221717	220	7	1.1	0.9;
	7460	1	-0	-0	0	37.99	0	1.037667	-46.665179	380	4	1.1	0.9;
	7461	1	10.11	3.7	0	0.76	0	1.015254	-23.621982	110	19	1.1	0.9;
	7462	1	-0	-0	0	1.77	0	1.042191	-23.736143	150	9	1.1	0.9;
	7463	1	178.7	31.3	0	5.6	0	1.025555	-54.035347	220	4	1.1	0.9;
	7464	1	0	-0	0	7.58	0	1.038219	-12.663099	220	5	1.1	0.9;
	7465	2	0	0	0	0	0	1.012708	61.67261	150	1	1.1	0.9;
	7466	2	0	0	0	0	0	1.060565	-34.28863	220	5	1.1	0.9;
	7467	1	-11.14	-3.38	0	2.35	0	1.032778	-23.639814	220	7	1.1	0.9;
	7468	1	15.72	5.2	0	0.76	0	0.973649	-32.737856	110	21	1.1	0.9;
	7469	1	1.5	0.4	0	0.66	0	0.96597	-22.232088	110	18	1.1	0.9;
	7470	1	13.72	8.5	0	19.46	0	1.07691	-35.421507	150	9	1.1	0.9;
	7471	1	-18.95	-6.93	0	33.11439	0	1.012658	-6.664898	380	5	1.1	0.9;
	7472	1	-0	-0	0	20.52	0	1.04852	-16.614246	380	7	1.1	0.9;
	7473	1	251.47	90.86	0	14.54	0	1.041663	-15.791744	220	5	1.1	0.9;
	7474	2	0	0	0	0	0	1.035524	-14.622562	220	5	1.1	0.9;
	7475	1	84.79	12.5	0	2.55	0	1.023441	-54.63894	220	4	1.1	0.9;
	7476	1	59.67	27.1	0	43.03	0	1.037093	5.951041	150	8	1.1	0.9;
	7477	1	106.8	22.2	0	0.97	0	1.036246	-26.543946	110	21	1.1	0.9;
	7478	1	47.95	29.7	0	0.23	0	0.941728	-34.204462	150	9	1.1	0.9;
	7479	1	0	0	0	2.06	0	0.999232	-51.517521	220	6	1.1	0.9;
	7480	1	-0.18	-3.54	0	0	0	1.017363	17.213196	150	8	1.1	0.9;
	7481	1	0	0	0	0.04	0	1.025922	4.612908	154	1	1.1	0.9;
	7482	1	-0	0	0	0.04	0	0.964965	-31.459019	150	9	1.1	0.9;
	7483	1	-1316.59	-129.9	0.234668	22.607741	0	1.049175	49.216593	380	3	1.1	0.9;
	7484	1	0.4	0.1	0	0.93	0	1.02013	-25.824468	110	19	1.1	0.9;
	7485	1	44.35	10.8	0	8.72	0	1.038191	-26.165239	220	5	1.1	0.9;
	7486	1	4.41	2.7	0	1.58	0	0.9702	-27.71633	110	18	1.1	0.9;
	7487	1	38.04	18	0	0.15	0	1.018688	-29.208471	220	6	1.1	0.9;
	7488	1	0.2	0.7	0	0.04	0	0.986342	14.700495	220	3	1.1	0.9;
	7489	1	70.08	35	0	0.33	0	1.007426	-49.42421	220	6	1.1	0.9;
	7490	2	0	0	0	0	0	1.097048	45.389496	380	3	1.1	0.9;
	7491	1	0	-0	0	4.29	0	1.007019	-21.833553	220	5	1.1	0.9;
	7492	1	-0	-0	0	43.34	0	1.05365	18.017409	380	3	1.1	0.9;
	7493	1	0	-0	0	4.14	0	1.035207	3.445303	220	13	1.1	0.9;
	7494	1	-6.79	-1.45	0	6.66	0	1.01018	-36.985158	220	6	1.1	0.9;
	7495	2	0	0	0	0	0	1.031216	-15.255053	220	5	1.1	0.9;
	7496	1	10.11	3.7	0	0.25	0	1.026541	-23.154781	110	19	1.1	0.9;
	7497	1	4.4	2.2	0	0.29	0	1.040579	-39.790195	110	22	1.1	0.9;
	7498	2	0	0	0	0	0	1.000261	-3.434589	400	1	1.1	0.9;
	7499	2	0	0	0	0	0	0.999526	-16.796989	110	18	1.1	0.9;
	7500	1	136.2	4.8	0	7.58	0	1.039826	-14.486951	220	3	1.1	0.9;
	7501	1	39.84	24.6	0	13.99	0	0.994734	-34.15431	150	9	1.1	0.9;
	7502	1	128.9	73.7	0	25.37	0	1.035437	19.496096	380	3	1.1	0.9;
	7503	1	-0	-0	0	1.86	0	0.997238	-43.422688	220	10	1.1	0.9;
	7504	2	0	0	0	0	0	1.015325	-43.221319	220	4	1.1	0.9;
	7505	1	0	0	0	5.56	0	1.033062	-0.176606	380	3	1.1	0.9;
	7506	1	0	-0	0.135375	8.900675	0	1.068978	-22.479995	380	18	1.1	0.9;
	7507	1	4	0.8	0	17.65	0	1.044799	-24.411805	220	5	1.1	0.9;
	7508	2	0	0	0	0	0	1.058554	-40.238969	380	4	1.1	0.9;
	7509	1	88	36.2	0	60	0	1.053987	-48.004141	380	10	1.1	0.9;
	7510	1	46.35	14.9	0	20.62	0	1.022821	-47.132805	220	4	1.1	0.9;
	7511	1	0.1	-0	0	5.01	0	1.04215	0.142791	220	13	1.1	0.9;
	7512	1	-0	-0	0	2.28	0	1.013069	3.182736	380	3	1.1	0.9;
	7513	1	0	-0	0	38.549494	0	0.989662	-15.298658	380	5	1.1	0.9;
	7514	1	1.4	100	0	21.43	0	1.019422	24.07354	220	3	1.1	0.9;
	7515	1	-0	-0	0.090774	3.150444	0	1.058579	-16.388932	220	18	1.1	0.9;
	7516	1	3.8	0.6	0	0.68	0	0.969057	-22.128022	110	18	1.1	0.9;
	7517	1	9.01	4	0	1.39	0	1.057243	-24.998707	110	23	1.1	0.9;
	7518	1	-14.89	-4.46	0	0.3	0	0.971666	-22.117368	110	18	1.1	0.9;
	7519	1	97.97	31.8	0	5.02	0	1.033872	-10.882569	220	5	1.1	0.9;
	7520	2	0	0	0	0	0	1.027003	-9.358633	380	5	1.1	0.9;
	7521	1	0	0	0.15	5.08	0	1.008595	-29.116434	220	21	1.1	0.9;
	7522	2	0	0	0	0	0	0.995659	-11.246091	380	5	1.1	0.9;
	7523	1	-0	0	0	2.11	0	1.050802	-0.784953	220	5	1.1	0.9;
	7524	1	-0	-0	0	6.59	0	1.011914	-8.720586	220	3	1.1	0.9;
	7525	1	-8.66	-6.01	0	0	0	1.022174	17.018992	150	8	1.1	0.9;
	7526	1	0	-0	0	1.88	0	1.021601	6.203426	380	2	1.1	0.9;
	7527	1	125.4	17.8	0	7.64	0	1.024309	4.17589	154	1	1.1	0.9;
	7528	1	211	59.3	0	3.46	0	1.015519	-27.648326	110	21	1.1	0.9;
	7529	1	-0	0	0	13.62	0	1.02609	-25.261522	220	7	1.1	0.9;
	7530	1	0	0	0	31.242294	0	1.043248	-20.130634	380	5	1.1	0.9;
	7531	1	26.23	8.4	0	0.96	0	1.025896	-35.482797	110	12	1.1	0.9;
	7532	1	13.82	5	0	0.43	0	1.069576	-25.272383	220	19	1.1	0.9;
	7533	1	43.15	15.4	0	0.36	0	1.087886	-21.828117	110	15	1.1	0.9;
	7534	1	20.62	6.7	0	0.71	0	0.993725	-30.761815	110	21	1.1	0.9;
	7535	1	0	0	0	1.98	0	0.955662	-33.07479	150	9	1.1	0.9;
	7536	1	21.83	6.3	0	2.01	0	1.00694	-28.363401	110	18	1.1	0.9;
	7537	1	0	0	0	17.184121	0	0.98576	-15.917614	380	5	1.1	0.9;
	7538	1	130.5	17.2	0	5.65	0	1.005756	-50.384652	220	6	1.1	0.9;
	7539	1	40.85	15.7	0	2.08	0	1.044276	-12.725707	220	5	1.1	0.9;
	7540	1	-0	0	0	3.72	0	1.003815	-50.653713	220	6	1.1	0.9;
	7541	1	346.42	89.04	0	43.55	0	1.051631	-16.709801	380	5	1.1	0.9;
	7542	1	-0	0	0	0.53	0	1.060157	-26.994694	150	9	1.1	0.9;
	7543	1	72.48	44.9	0	3.86	0	0.975873	-35.299892	150	9	1.1	0.9;
	7544	1	29.65	7.18	0	4.51	0	1.021696	65.99455	150	1	1.1	0.9;
	7545	1	0	0	0.237048	35.833975	0	1.035259	34.769524	380	3	1.1	0.9;
	7546	1	0	0	0	0	0	0.917414	-35.806411	150	9	1.1	0.9;
	7547	1	-0	-0	0	26.93	0	1.096732	-23.861514	380	9	1.3	0.7;
	7548	1	281.3	40.9	0	33.56	0	1.025915	-35.014202	220	11	1.1	0.9;
	7549	1	20.92	9.35	0	1.56	0	1.00146	-19.644156	110	18	1.1	0.9;
	7550	1	0	-0	0	5.55	0	1.001368	-39.424985	220	1	1.1	0.9;
	7551	1	101.48	19.53	0	4.41	0	1.028534	-4.964919	220	13	1.1	0.9;
	7552	1	-0	0	0	0.31	0	1.021733	-32.467743	150	9	1.1	0.9;
	7553	1	-0	0	0	22.68	0	1.047249	9.172669	380	13	1.1	0.9;
	7554	1	2.9	1	0	0.39	0	1.028509	-25.716891	110	19	1.1	0.9;
	7555	1	2.1	0.8	0	0.42	0	0.992214	-25.988665	110	18	1.1	0.9;
	7556	1	0	-0	0	3.1	0	1.044301	-33.52744	220	1	1.1	0.9;
	7557	1	34.44	12.1	0	0.87	0	1.051797	-24.738071	110	15	1.1	0.9;
	7558	2	0	0	0	0	0	1.005058	-22.194728	154	1	1.1	0.9;
	7559	1	8.98	-3.79	0	2.86	0	1.017685	61.838744	150	1	1.1	0.9;
	7560	1	34.64	12.4	0	2.66	0	1.054373	-19.314276	110	19	1.1	0.9;
	7561	1	0	0	0	20.54	0	0.992829	-4.851657	380	1	1.1	0.9;
	7562	1	218	18.6	0	36.29	0	1.047541	-14.372444	380	16	1.1	0.9;
	7563	1	531.7	-12.8	0	0	0	1.0345	-13.492374	150	2	1.1	0.9;
	7564	1	-0	0	0	-0.17	0	0.99088	-50.568992	220	6	1.1	0.9;
	7565	1	-0	0	0	0.81	0	1.066837	11.80124	110	8	1.1	0.9;
	7566	1	0	0	0	0	0	1.014086	-33.092664	150	9	1.1	0.9;
	7567	1	13.9	1.9	0	2.23	0	1.029761	7.240879	154	1	1.1	0.9;
	7568	2	0	0	0	0	0	1.043388	-31.041557	220	12	1.1	0.9;
	7569	1	0	-0	0	1.18	0	0.964544	-56.318516	150	10	1.1	0.9;
	7570	1	80.8	11.6	0	6.96	0	1.038503	7.330109	154	1	1.1	0.9;
	7571	1	-0	-0	0	13.9	0	1.040207	-14.312963	220	5	1.1	0.9;
	7572	1	0	0	0	48.75	0	1.068923	-15.021027	380	19	1.1	0.9;
	7573	1	126.6	28.7	0	15.02	0	1.026738	-54.221012	220	4	1.1	0.9;
	7574	1	0	-0	0	2.12	0	1.030705	-38.228839	220	6	1.1	0.9;
	7575	1	29.03	6.1	0	0.44	0	1.050929	9.688249	110	8	1.1	0.9;
	7576	1	0	0	0	0.5	0	1.034553	-12.333688	220	5	1.1	0.9;
	7577	1	0	0	0	6.69	0	1.042437	-10.786006	220	5	1.1	0.9;
	7578	1	45.95	15.4	0	0.6	0	1.002902	-27.805666	110	18	1.1	0.9;
	7579	1	136.9	11.4	0	3.05	0	1.037135	-27.098952	220	5	1.1	0.9;
	7580	1	0	0	0	0.17	0	1.0474	-12.31099	220	5	1.1	0.9;
	7581	1	32.1	4.5	0	3.75	0	1.023178	-0.742234	154	1	1.1	0.9;
	7582	1	0	-0	0	3.44	0	1.030716	-24.81618	220	5	1.1	0.9;
	7583	1	5.01	1.8	0	3.68	0	1.046345	-1.027496	220	5	1.1	0.9;
	7584	1	127.6	45.3	0	57.43	0	1.03213	-1.348827	150	8	1.1	0.9;
	7585	1	27.93	8.9	0	0.47	0	0.988991	-27.243981	110	18	1.1	0.9;
	7586	1	-18.27	-12.99	0	0	0	1.017521	-4.530708	150	8	1.1	0.9;
	7587	1	-0	-0	0	0.87	0	0.948139	-32.704324	150	9	1.1	0.9;
	7588	1	10.11	4.4	0	0.29	0	0.99287	-24.994466	110	18	1.1	0.9;
	7589	1	9.71	3.2	0	0.26	0	1.006818	-39.42041	110	12	1.1	0.9;
	7590	1	0	0	0	0.71	0	1.002951	-30.560312	110	21	1.1	0.9;
	7591	1	11.21	3.7	0	0.9	0	1.024724	-34.262786	110	12	1.1	0.9;
	7592	1	159.4	-7.3	0	3.3	0	1.050828	1.68523	220	3	1.1	0.9;
	7593	1	58.77	16.1	0	4.11	0	1.023992	-43.76072	220	6	1.1	0.9;
	7594	1	116.5	33.4	0	0.28	0	1.025746	-8.684743	110	19	1.1	0.9;
	7595	1	30.43	10	0	0.41	0	0.987318	-31.944297	110	21	1.1	0.9;
	7596	1	59.07	23.4	0	21.88	0	1.021583	-1.074276	150	8	1.1	0.9;
	7597	1	18.02	4	0	3.3	0	1.050861	-26.64185	220	11	1.1	0.9;
	7598	1	0.5	-0	0	0.84	0	0.947931	-32.207203	110	18	1.1	0.9;
	7599	1	0	-0	0	2.13	0	1.031223	-6.438893	220	5	1.1	0.9;
	7600	1	119	31.7	0	1.94	0	1.033028	-46.17218	220	4	1.1	0.9;
	7601	1	-0	0	0	0.62	0	1.034496	58.686891	220	3	1.1	0.9;
	7602	1	119.7	43.4	0	0.92	0	1.030678	-21.980192	220	3	1.1	0.9;
	7603	1	5.51	2.2	0	1.43	0	1.003179	-41.85431	110	22	1.1	0.9;
	7604	1	87	30.8	0	2.43	0	0.98606	-31.565215	110	21	1.1	0.9;
	7605	1	0	0	0	0	0	0.988635	-31.00334	150	9	1.1	0.9;
	7606	1	2.8	0.5	0	1.26	0	1.003992	-30.714976	110	18	1.1	0.9;
	7607	2	0	0	0	0	0	1.054831	-17.179333	110	21	1.1	0.9;
	7608	1	8.31	3	0	1.27	0	0.967568	-20.735638	110	19	1.1	0.9;
	7609	1	24.03	-56.8	0	2.69	0	1.021437	-3.015749	220	3	1.1	0.9;
	7610	1	26.82	24.2	0	2.89	0	0.981231	-35.406383	150	9	1.1	0.9;
	7611	1	115.4	31	0	28.85	0	1.021653	-47.204811	220	4	1.1	0.9;
	7612	1	-20.12	6.6	0	0.85	0	1.041242	-23.21805	150	9	1.1	0.9;
	7613	1	6.31	2.5	0	0.44	0	0.920496	-33.041733	110	18	1.1	0.9;
	7614	1	41.8	5.9	0	2.54	0	1.019019	5.953759	154	1	1.1	0.9;
	7615	1	70.48	17.6	0	3.92	0	0.981178	-53.873099	220	10	1.1	0.9;
	7616	2	0	0	0	0	0	1.131608	-22.062538	380	9	1.3	0.7;
	7617	1	-0	0	0.203547	31.507823	0	1.049522	-16.107883	380	3	1.1	0.9;
	7618	1	-0	-0	0	1.69	0	1.017916	-29.278342	220	6	1.1	0.9;
	7619	2	0	0	0	0	0	1.037789	-36.854382	380	6	1.1	0.9;
	7620	1	136.95	87.7	0	7.84	0	1.008205	9.580034	220	3	1.1	0.9;
	7621	1	0	0	0	0.84	0	1.008866	-33.752136	150	9	1.1	0.9;
	7622	1	0	0	0	14.94	0	1.033759	-3.23424	220	1	1.1	0.9;
	7623	1	90.5	12.8	0	6.78	0	1.009701	1.582877	154	1	1.1	0.9;
	7624	1	32.74	10.5	0	1.78	0	1.010731	-21.256786	220	5	1.1	0.9;
	7625	1	164.41	101.14	0	61.4	0	1.072566	4.746322	380	13	1.1	0.9;
	7626	1	78.89	45.6	0	0.04	0	1.02695	-12.756876	220	5	1.1	0.9;
	7627	2	0	0	0	0	0	1.054377	-24.30542	220	21	1.1	0.9;
	7628	2	0	0	0	0	0	1.064136	3.784087	150	8	1.1	0.9;
	7629	1	0	0	0	1.15	0	1.015993	-55.973239	220	4	1.1	0.9;
	7630	2	0	0	0	0	0	1.017306	-44.944646	220	4	1.1	0.9;
	7631	1	-0	0	0	2.82	0	1.032549	38.780071	220	3	1.1	0.9;
	7632	1	37.94	3.4	0	11.17	0	1.022447	-56.809735	220	4	1.1	0.9;
	7633	1	91.6	20.3	0	0	0	1.02035	-35.218871	220	6	1.1	0.9;
	7634	1	5.5	0.8	0	1.9	0	1.014731	-9.370106	154	1	1.1	0.9;
	7635	1	0	0	0	1.86	0	1.047307	-7.949655	220	5	1.1	0.9;
	7636	1	-12.62	11.6	0	0.02	0	1.032102	-26.066906	150	9	1.1	0.9;
	7637	1	0	0	0.29	6.18	0	1.006165	18.142288	380	2	1.1	0.9;
	7638	1	429.5	163.1	0	28.67	0	1.064775	-21.698244	220	11	1.1	0.9;
	7639	2	0	0	0	0	0	1.016695	-36.384211	220	6	1.1	0.9;
	7640	1	68.78	35	0	34.52	0	1.029162	-6.10323	380	5	1.1	0.9;
	7641	2	0	0	0	0	0	1.049265	-6.450367	220	5	1.1	0.9;
	7642	1	0	-0	0	138.75	0	1.038438	5.517843	400	1	1.1	0.9;
	7643	1	90.3	5.7	0	19.04	0	1.043026	35.447616	380	3	1.1	0.9;
	7644	1	124.1	-30.3	0	14.26	0	1.034619	-46.307286	220	4	1.1	0.9;
	7645	2	0	0	0	0	0	1.028371	-6.365534	154	1	1.1	0.9;
	7646	1	0	0	0	33.48	0	1.062816	-11.73385	380	1	1.1	0.9;
	7647	1	-0	-0	0	0.1	0	1.058956	-20.48042	220	5	1.1	0.9;
	7648	1	15.12	5.4	0	0.63	0	1.054998	-25.914188	110	15	1.1	0.9;
	7649	1	29.13	11.3	0	0.16	0	0.993367	-17.616084	110	18	1.1	0.9;
	7650	1	439	254.8	0	148.04	0	1.094733	45.525792	380	3	1.1	0.9;
	7651	1	61.97	38.4	0	7.42	0	1.092516	-27.545772	150	9	1.1	0.9;
	7652	1	255.3	78.7	0	6.5	0	0.980657	6.38938	220	3	1.1	0.9;
	7653	1	33.94	11.4	0	0	0	1.032828	-23.116419	220	5	1.1	0.9;
	7654	1	41.8	5.9	0	3.29	0	1.011955	-7.817271	154	1	1.1	0.9;
	7655	1	117	15.2	0	2.87	0	1.013643	8.708415	220	3	1.1	0.9;
	7656	2	0	0	0	0	0	1.013707	-35.671719	220	22	1.1	0.9;
	7657	2	0	0	0	0	0	1.129049	-21.956352	380	9	1.3	0.7;
	7658	1	-0	-0	0	35.28	0	1.004842	-6.875549	400	1	1.1	0.9;
	7659	2	0	0	0	0	0	1.029603	-40.886975	220	4	1.1	0.9;
	7660	1	-0	0	0.090774	5.169044	0	1.055949	-14.740226	220	18	1.1	0.9;
	7661	2	0	0	0	0	0	1.037591	-34.298341	220	10	1.1	0.9;
	7662	1	-0.21	-2.37	0	0	0	1.023442	8.989291	150	8	1.1	0.9;
	7663	1	37.14	2.7	0	2.33	0	1.038529	-12.302936	220	5	1.1	0.9;
	7664	1	256.8	-61.8	0	36.73	0	1.036686	-19.852744	380	6	1.1	0.9;
	7665	1	16.52	5.5	0	1.15	0	1.029597	-41.155975	110	22	1.1	0.9;
	7666	1	93.4	27.8	0	29.52	0	1.020655	10.565208	150	8	1.1	0.9;
	7667	1	42.66	-22.22	0	0.66	0	1.041165	-3.322997	220	1	1.1	0.9;
	7668	1	27.7	4	0	13.66	0	1.035561	-25.045778	154	1	1.1	0.9;
	7669	1	67.08	25.2	0	8.11	0	1.021363	-49.000092	220	4	1.1	0.9;
	7670	2	0	0	0	0	0	1.005135	67.228591	150	1	1.1	0.9;
	7671	2	0	0	0	0	0	1.080257	-12.457265	220	18	1.1	0.9;
	7672	1	-0	-0	0	35.53	0	1.03982	-12.082133	380	6	1.1	0.9;
	7673	1	20.82	7.9	0	42.36	0	1.043711	-43.284962	380	4	1.1	0.9;
	7674	1	0	0	0.24033	46.858516	0	1.109684	-28.896585	380	1	1.3	0.7;
	7675	1	66.07	21.6	0	0.75	0	1.012014	0.036914	150	8	1.1	0.9;
	7676	2	0	0	0	0	0	1.074322	37.491062	380	3	1.1	0.9;
	7677	1	0	-0	0	5.24	0	1.009405	-7.73948	154	1	1.1	0.9;
	7678	1	0	0	0	2.257591	0	1.013019	-11.992692	220	24	1.1	0.9;
	7679	1	-0	-0	0	9.70055	0	1.024301	-26.976625	220	20	1.1	0.9;
	7680	1	-22.2	-5.79	0	5.28	0	1.08519	45.258983	380	3	1.1	0.9;
	7681	2	0	0	0	0	0	1.052405	-21.118883	220	7	1.1	0.9;
	7682	1	0	-0	0	0.09	0	1.020266	-22.730451	150	9	1.1	0.9;
	7683	2	0	0	0	0	0	1.038382	10.377986	154	1	1.1	0.9;
	7684	1	-0	0	0	1.59	0	1.036834	59.794276	220	3	1.1	0.9;
	7685	1	49.45	6.5	0	3.53	0	1.019277	-44.594553	220	4	1.1	0.9;
	7686	1	-0	0	0	48.77	0	1.051977	-26.907359	380	1	1.1	0.9;
	7687	2	0	0	0	0	0	1.043409	-30.144782	110	21	1.1	0.9;
	7688	1	12.71	5.1	0	15.31	0	1.041195	-50.237731	380	4	1.1	0.9;
	7689	1	95.6	13.6	0	21.47	0	1.023364	-7.753859	154	1	1.1	0.9;
	7690	1	0	0	0	5.28	0	1.021398	3.775893	380	3	1.1	0.9;
	7691	1	0	-0	0	96.48	0	1.02503	-3.969225	380	5	1.1	0.9;
	7692	1	0	-0	0.17252	3.43811	0	1.024576	-26.04578	220	3	1.1	0.9;
	7693	1	25.13	15.6	0	4.51	0	1.025574	-31.851042	150	9	1.1	0.9;
	7694	1	-14.96	-4.7	0	0.04	0	1.02559	-15.023648	220	5	1.1	0.9;
	7695	1	0	0	0	10.69	0	1.023436	-31.481035	220	12	1.1	0.9;
	7696	1	10.91	2.5	0	1.34	0	0.976851	-31.464367	110	18	1.1	0.9;
	7697	2	0	0	0	0	0	1.050111	-3.787728	220	5	1.1	0.9;
	7698	1	0	0	0	0.06	0	1.041522	-25.438548	150	9	1.1	0.9;
	7699	1	-0	0	0	19.62	0	1.041518	10.8605	380	1	1.1	0.9;
	7700	1	66.17	27.4	0	14.12	0	1.016311	-10.583646	220	5	1.1	0.9;
	7701	1	11.91	3.5	0	0.43	0	1.016426	5.847338	150	8	1.1	0.9;
	7702	1	-0	-0	0	11.68	0	1.022518	-16.488713	220	5	1.1	0.9;
	7703	1	51.96	6.5	0	5.14	0	1.007167	-26.363898	110	18	1.1	0.9;
	7704	1	0	0	0	0	0	1.016558	-30.283741	150	9	1.1	0.9;
	7705	1	0	-0	0	0.84	0	1.001405	-31.184725	110	21	1.1	0.9;
	7706	1	-0	0	0	1.57	0	1.045653	-7.902502	220	3	1.1	0.9;
	7707	2	0	0	0	0	0	1.042135	18.811854	220	13	1.1	0.9;
	7708	1	0.2	88.5	0	2.21	0	1.003658	67.616703	150	1	1.1	0.9;
	7709	1	2.2	0.5	0	0.58	0	1.012268	-41.855728	220	10	1.1	0.9;
	7710	1	60.07	36	0	0.06	0	0.990419	-32.797505	150	9	1.1	0.9;
	7711	1	126.2	-12.5	0	10.61	0	1.05815	-2.79987	220	3	1.1	0.9;
	7712	1	0	0	0	0	0	0.949809	-29.215766	150	9	1.1	0.9;
	7713	1	15.12	4.9	0	0.68	0	0.999973	-37.688524	110	12	1.1	0.9;
	7714	1	0	0	0	1.08	0	1.028066	9.772573	220	3	1.1	0.9;
	7715	1	85.9	93.8	0	3.93	0	1.053765	9.778906	380	3	1.1	0.9;
	7716	1	15.6	2.2	0	2.31	0	1.022625	1.228696	154	1	1.1	0.9;
	7717	1	0	0	0	0.47	0	1.030901	-9.649569	220	3	1.1	0.9;
	7718	2	0	0	0	0	0	1.008757	-36.951753	220	6	1.1	0.9;
	7719	1	-0	0	0	2.66	0	1.052232	-20.129634	220	1	1.1	0.9;
	7720	1	70.78	43.8	0	2.12	0	0.870967	-38.524972	150	9	1.3	0.7;
	7721	1	38.94	11	0	0.64	0	1.010822	-34.672501	110	1	1.1	0.9;
	7722	1	17.02	5.5	0	1.02	0	0.992167	-31.176403	110	21	1.1	0.9;
	7723	1	220.2	83.2	0	0.61	0	1.036019	-32.272172	220	12	1.1	0.9;
	7724	1	39.64	4.4	0	0	0	1.002154	-26.93002	110	24	1.1	0.9;
	7725	1	65.97	64.1	0	29.14	0	1.05186	9.366514	380	3	1.1	0.9;
	7726	1	0	0	0	2.03	0	1.016206	-10.548865	220	5	1.1	0.9;
	7727	1	46.05	22	0	2.95	0	1.020582	-33.512439	220	6	1.1	0.9;
	7728	2	0	0	0	0	0	1.008042	0.473824	400	1	1.1	0.9;
	7729	1	20.82	6.7	0	2.25	0	1.028639	7.765834	110	8	1.1	0.9;
	7730	1	0	0	0	0.69	0	1.034831	-28.044352	110	21	1.1	0.9;
	7731	1	71.58	48.2	0.288552	130.762047	0	1.021323	-9.598819	380	3	1.1	0.9;
	7732	1	137	33.2	0	3.82	0	1.019657	-26.015061	220	7	1.1	0.9;
	7733	1	7	1	0	1.51	0	1.000347	-27.833154	154	1	1.1	0.9;
	7734	1	65.57	44.6	0	5.95	0	0.920987	-34.594655	150	9	1.1	0.9;
	7735	1	-0	0	0	0.36	0	1.020146	-22.689264	150	9	1.1	0.9;
	7736	1	0	-0	0	55	0	1.032289	-1.351916	150	8	1.1	0.9;
	7737	2	0	0	0	0	0	1.061861	-21.991295	110	19	1.1	0.9;
	7738	1	31.84	9.2	0	5.23	0	1.030914	-22.34307	220	5	1.1	0.9;
	7739	1	82.29	29.6	0	1.86	0	1.005262	-29.681727	110	21	1.1	0.9;
	7740	1	0	-0	0	0.97	0	0.998038	-32.178415	150	9	1.1	0.9;
	7741	1	-24.99	-4.94	0	6.99	0	1.049273	43.29035	220	3	1.1	0.9;
	7742	2	0	0	0	0	0	1.016237	61.636194	150	1	1.1	0.9;
	7743	2	0	0	0	0	0	1.011249	-20.92784	220	6	1.1	0.9;
	7744	1	8.31	3	0	1.37	0	1.012901	-26.277782	110	19	1.1	0.9;
	7745	1	24.23	4.7	0	0.75	0	0.989318	-29.054312	110	18	1.1	0.9;
	7746	1	58.5	8.3	0	4.93	0	1.01801	-5.451643	154	1	1.1	0.9;
	7747	1	-0	-0	0	5.53	0	1.007146	-19.568569	110	18	1.1	0.9;
	7748	1	122.7	-29.8	0	7.11	0	1.022866	-6.081369	220	3	1.1	0.9;
	7749	1	10.9	1.6	0	1.5	0	1.023861	-4.402046	154	1	1.1	0.9;
	7750	1	0	-0	0	9.41	0	1.063487	-20.192378	220	18	1.1	0.9;
	7751	1	0	0	0	0	0	1.024415	-1.614683	150	8	1.1	0.9;
	7752	1	524.5	126	0	7.45	0	1.027001	-18.689663	220	5	1.1	0.9;
	7753	1	-0	0	0	0.13	0	1.012418	-48.244731	220	6	1.1	0.9;
	7754	1	76.6	10.9	0	2.41	0	1.028685	-6.819632	154	1	1.1	0.9;
	7755	2	0	0	0	0	0	1.058479	-42.500332	380	4	1.1	0.9;
	7756	1	8.03	5.5	0	0.12	0	1.025067	-29.462127	150	9	1.1	0.9;
	7757	1	17.52	5.7	0	0.28	0	1.062925	-17.102366	110	18	1.1	0.9;
	7758	2	0	0	0	0	0	1.059721	-12.30921	380	16	1.1	0.9;
	7759	2	0	0	0	0	0	1.17759	-19.026092	380	9	1.3	0.7;
	7760	1	10.61	1	0	1.7	0	0.989841	-21.852274	110	18	1.1	0.9;
	7761	2	0	0	0	0	0	1.012378	-41.832827	220	10	1.1	0.9;
	7762	1	-0	0	0.304243	48.453399	0	1.006235	-3.979104	380	2	1.1	0.9;
	7763	1	-0	0	0	0.15	0	1.053441	-29.282647	150	9	1.1	0.9;
	7764	1	0	-0	0	25.67	0	1.049594	-20.329998	380	1	1.1	0.9;
	7765	1	11.01	3.6	0	0.5	0	0.996648	-30.560945	110	21	1.1	0.9;
	7766	1	78.89	25.6	0	1.6	0	1.027943	-12.996697	220	5	1.1	0.9;
	7767	1	0	-0	0	1.98	0	1.026684	-11.317489	220	3	1.1	0.9;
	7768	1	11.91	3.9	0	2	0	1.024697	-33.481891	110	12	1.1	0.9;
	7769	1	3.4	0.5	0	3.28	0	1.027986	-4.269849	154	1	1.1	0.9;
	7770	1	47.15	15.5	0	9.69	0	1.02163	-44.828126	220	5	1.1	0.9;
	7771	1	-0	0	0	13.4	0	1.071376	44.31689	380	3	1.1	0.9;
	7772	1	223.6	51.8	0	18.45	0	1.053847	-26.048725	220	5	1.1	0.9;
	7773	1	0	0	0	0	0	1.039391	-27.930465	150	9	1.1	0.9;
	7774	1	27.23	10.1	0	0	0	1.001867	-31.058425	220	3	1.1	0.9;
	7775	1	151.5	7.9	0	5	0	1.047794	-29.282448	220	5	1.1	0.9;
	7776	2	0	0	0	0	0	1.077614	5.972286	380	5	1.3	0.7;
	7777	1	-0	-0	0	49.86	0	1.074558	-20.609018	380	19	1.1	0.9;
	7778	1	153	21.5	0	28.85	0	1.051431	-16.82666	380	5	1.1	0.9;
	7779	2	0	0	0	0	0	1.0806	10.157985	150	8	1.1	0.9;
	7780	1	0	0	0	13.33	0	1.048635	1.215771	400	1	1.1	0.9;
	7781	1	108.5	35.3	0	49.63	0	1.046792	9.207136	380	3	1.1	0.9;
	7782	1	1.4	0.5	0	0.89	0	1.029133	-28.802489	110	19	1.1	0.9;
	7783	1	26.13	8.5	0	0.37	0	1.001462	-31.048447	110	21	1.1	0.9;
	7784	1	11.01	5.5	0	1.12	0	1.027093	-39.679574	110	22	1.1	0.9;
	7785	2	0	0	0	0	0	1.048533	-23.635241	220	7	1.1	0.9;
	7786	1	113.4	11.9	0	5.4	0	1.022574	-5.684416	220	3	1.1	0.9;
	7787	1	30.03	9.6	0	4.38	0	1.006266	-49.628428	220	6	1.1	0.9;
	7788	1	7	1	0	1.3	0	1.008294	1.280908	154	1	1.1	0.9;
	7789	1	-0	0	0	55.79	0	1.021551	4.549434	400	1	1.1	0.9;
	7790	1	0	0	0.1	-0.29	0	1.043792	-20.814489	220	19	1.1	0.9;
	7791	1	-0	0	0	1.33	0	1.024916	-13.291854	220	5	1.1	0.9;
	7792	1	0	-0	0	0.46	0	1.006966	-37.982122	150	10	1.1	0.9;
	7793	1	7.81	2.8	0	0.4	0	1.044057	-20.54901	110	19	1.1	0.9;
	7794	1	97.5	13.9	0	2.42	0	0.991709	-9.553785	154	1	1.1	0.9;
	7795	1	44.35	5.1	0	4	0	1.02187	-46.825102	220	4	1.1	0.9;
	7796	1	12.91	4.7	0	1.09	0	1.059894	-25.15852	110	19	1.1	0.9;
	7797	1	-0	-0	0	2.09	0	1.046734	-24.420983	220	5	1.1	0.9;
	7798	2	0	0	0	0	0	1.015906	-15.017134	110	18	1.1	0.9;
	7799	1	86.5	29	0	1.45	0	1.069146	26.047995	380	3	1.1	0.9;
	7800	2	0	0	0	0	0	1.005948	-30.077471	110	18	1.1	0.9;
	7801	1	-0	-0	0	0.14	0	1.013956	-30.373281	150	9	1.1	0.9;
	7802	1	223.1	31.8	0	23.58	0	1.022063	-25.266142	154	1	1.1	0.9;
	7803	1	-0	0	0.378102	20.394045	0	1.031127	13.410495	380	3	1.1	0.9;
	7804	1	91.1	33.5	0	1.8	0	1.006207	-36.150902	220	6	1.1	0.9;
	7805	1	-0	0	0	30.46	0	1.025046	-54.258656	220	4	1.1	0.9;
	7806	1	-0	-0	0	0.22	0	0.992491	-17.727224	110	18	1.1	0.9;
	7807	1	-0	0	0	56.53	0	1.101652	-20.569701	380	9	1.3	0.7;
	7808	2	0	0	0	0	0	1.050047	-0.347653	380	5	1.1	0.9;
	7809	1	169.5	27.7	0	3.04	0	1.026576	-11.552809	220	5	1.1	0.9;
	7810	1	0	-0	0	43	0	1.046439	-50.114824	380	10	1.1	0.9;
	7811	1	-0	0	0	3.21	0	1.026871	49.246356	220	3	1.1	0.9;
	7812	1	22.83	3.8	0	0.26	0	1.002777	-18.289885	110	18	1.1	0.9;
	7813	1	1.8	0.5	0	0.79	0	0.967541	-22.704345	110	18	1.1	0.9;
	7814	1	34.9	5	0	2.53	0	1.009504	-7.482186	154	1	1.1	0.9;
	7815	1	149.1	-18.3	0	10.76	0	1.02006	-23.451919	220	3	1.1	0.9;
	7816	1	40.24	16	0	17.25	0	1.036197	29.195301	380	3	1.1	0.9;
	7817	1	14.82	-0	0	4.07	0	1.061023	-23.840125	220	7	1.1	0.9;
	7818	1	-41.65	13.7	0	0.3	0	1.036237	-22.435799	150	9	1.1	0.9;
	7819	2	0	0	0	0	0	1.018324	-40.60173	220	4	1.1	0.9;
	7820	1	127.2	-10.7	0	7.8	0	1.012896	-54.950671	220	4	1.1	0.9;
	7821	1	179	22.9	0	25.28	0	1.026152	-51.381177	380	4	1.1	0.9;
	7822	1	0	0	0	0	0	0.823843	-41.5372	150	9	1.3	0.7;
	7823	1	34.34	12.2	0	0.34	0	1.048501	-21.383375	110	19	1.1	0.9;
	7824	1	9.21	3.4	0	2.63	0	1.046444	-6.348461	220	5	1.1	0.9;
	7825	1	0	0	0	0.52	0	0.970598	-32.19603	150	9	1.1	0.9;
	7826	1	32.44	20.1	0	2.37	0	1.046412	-22.334303	150	9	1.1	0.9;
	7827	1	137.2	41.9	0.101088	13.207291	0	1.044339	-42.983742	380	10	1.1	0.9;
	7828	1	-0	-0	0.090774	13.090445	0	1.057001	-28.937439	220	18	1.1	0.9;
	7829	1	-3.2	-7.9	0	1.95	0	1.050303	-4.935711	220	2	1.1	0.9;
	7830	1	0	0	0.135493	30.26062	0	1.092455	-25.444573	380	21	1.1	0.9;
	7831	1	-17.17	-7.1	0	57.019803	0	1.009604	-11.59063	380	5	1.1	0.9;
	7832	1	-0	0	0	22.74	0	1.014229	-33.100575	150	9	1.1	0.9;
	7833	2	0	0	0	0	0	1.015416	-47.445689	220	4	1.1	0.9;
	7834	1	134.7	17.5	0	4.53	0	1.002304	-53.402463	220	4	1.1	0.9;
	7835	2	0	0	0	0	0	1.037848	-31.476302	220	6	1.1	0.9;
	7836	2	0	0	0	0	0	0.997593	-20.466901	110	18	1.1	0.9;
	7837	1	0	-0	0	1.01	0	1.011735	-30.33652	150	9	1.1	0.9;
	7838	1	10.11	4.3	0	0.63	0	0.983228	-28.775608	110	18	1.1	0.9;
	7839	1	-9.78	-3.66	0	0.72	0	0.979413	-31.941082	110	18	1.1	0.9;
	7840	1	-0	0	0	2.06	0	1.03573	-3.017599	220	5	1.1	0.9;
	7841	1	26.53	7.1	0	11.02	0	1.020631	-33.510096	220	6	1.1	0.9;
	7842	2	0	0	0	0	0	1.014075	-12.280157	380	5	1.1	0.9;
	7843	2	0	0	0	0	0	1.056109	-19.444876	220	3	1.1	0.9;
	7844	1	-0	0	0	27.39	0	1.0317	10.360208	150	8	1.1	0.9;
	7845	1	38.64	6	0	3.26	0	1.027035	-54.015313	220	4	1.1	0.9;
	7846	1	0	0	0	0	0	1.025574	-31.851042	150	9	1.1	0.9;
	7847	1	0	-0	0	1.55	0	1.038055	-9.097914	220	5	1.1	0.9;
	7848	1	-0	-0	0	0.86	0	1.032951	-30.886537	110	21	1.1	0.9;
	7849	1	-0	0	0	15.21	0	1.023318	30.245947	380	3	1.1	0.9;
	7850	1	12.6	1.8	0	4.19	0	1.013769	-3.348148	154	1	1.1	0.9;
	7851	1	-0	0	0	3.98	0	1.047047	-20.77235	220	1	1.1	0.9;
	7852	1	-3.53	-4.62	0	0	0	1.017155	-0.726874	150	8	1.1	0.9;
	7853	1	40.65	16	0	0.36	0	0.99252	-27.279983	110	18	1.1	0.9;
	7854	2	0	0	0	0	0	1.018633	29.625154	220	13	1.1	0.9;
	7855	1	-0	-0	0	13.08	0	1.011896	-36.505361	220	6	1.1	0.9;
	7856	1	0	0	0	0	0	1.047691	18.060892	380	8	1.1	0.9;
	7857	1	0	0	0	2.36	0	1.04716	0.068415	220	5	1.1	0.9;
	7858	1	-0.01	0	0	0	0	0.988647	-30.973141	150	9	1.1	0.9;
	7859	1	0	0	0	5.59	0	1.003656	8.317584	220	3	1.1	0.9;
	7860	2	0	0	0	0	0	1.054905	-40.689948	380	10	1.1	0.9;
	7861	1	178.9	9.1	0	15.71	0	1.019154	-42.999214	220	4	1.1	0.9;
	7862	1	0	0	0	9.74	0	1.035912	-21.455395	220	5	1.1	0.9;
	7863	1	-0	-0	0	0.33	0	0.955712	-33.401903	150	9	1.1	0.9;
	7864	1	11.41	0.3	0	0.2	0	0.997601	-30.875063	110	18	1.1	0.9;
	7865	1	84.99	19.6	0	4.32	0	1.068569	-26.767043	220	5	1.1	0.9;
	7866	1	2.2	0.8	0	1.31	0	0.968758	-22.304933	110	18	1.1	0.9;
	7867	1	88.5	10.8	0	9.54	0	1.021717	-22.153453	220	20	1.1	0.9;
	7868	1	0	0	0	0.26	0	1.015082	-32.471155	150	9	1.1	0.9;
	7869	1	101.9	20	0	22.64	0	1.021806	-46.913796	220	4	1.1	0.9;
	7870	1	-0	0	0	77.584854	0	1.051969	-20.062222	380	24	1.1	0.9;
	7871	1	0	0	0	0	0	0.972605	-32.139314	150	9	1.1	0.9;
	7872	1	17.02	2	0	3.65	0	1.04054	-19.358814	220	7	1.1	0.9;
	7873	1	-20.94	-6.16	0	11.99	0	1.01927	-15.386634	220	5	1.1	0.9;
	7874	1	-0	0	0	12.28	0	1.0194	16.370413	380	3	1.1	0.9;
	7875	1	-0	0	0	6.51	0	0.985313	-46.083564	220	4	1.1	0.9;
	7876	1	0	0	0.039888	30.741787	0	0.985245	66.968754	380	1	1.1	0.9;
	7877	1	64.21	5.59	0	2.57	0	1.009463	-46.023329	220	4	1.1	0.9;
	7878	1	0.3	-97.2	0	0.07	0	1.057954	-2.784924	220	3	1.1	0.9;
	7879	1	73.58	48.6	0	2.87	0	0.964375	-56.251818	150	10	1.1	0.9;
	7880	2	0	0	0	0	0	1.024905	20.048493	380	3	1.1	0.9;
	7881	1	-0	-0	0	7.99	0	1.059008	-15.410365	220	5	1.1	0.9;
	7882	1	26.63	9.7	0	0.37	0	1.049105	-18.501848	110	19	1.1	0.9;
	7883	1	47.55	17.8	0	0.32	0	1.047568	-29.087445	220	5	1.1	0.9;
	7884	1	35.84	14.1	0	50.11	0	1.046188	3.58426	150	8	1.1	0.9;
	7885	1	142.4	20	0	2.42	0	1.034688	-21.418219	220	5	1.1	0.9;
	7886	1	162.5	42.7	0	12.79	0	1.015163	-19.425489	220	5	1.1	0.9;
	7887	1	41.88	3.02	0	5.65	0	1.003818	63.534546	150	1	1.1	0.9;
	7888	1	-0	-0	0	4.52	0	1.017363	-21.654827	154	1	1.1	0.9;
	7889	1	-0	-0	0.090774	2.080444	0	1.059932	-18.401721	220	18	1.1	0.9;
	7890	2	0	0	0	0	0	1.05789	-2.175715	220	13	1.1	0.9;
	7891	1	427.9	145.8	0	40.72	0	1.034632	-38.013968	380	6	1.1	0.9;
	7892	1	36.94	11	0	0.52	0	1.04376	-11.60199	220	5	1.1	0.9;
	7893	1	0	-0	0	8.77	0	1.018258	-48.246771	220	6	1.1	0.9;
	7894	2	0	0	0	0	0	1.044899	9.503577	154	1	1.1	0.9;
	7895	1	338.7	52.5	0	11.36	0	1.054848	-19.868363	220	5	1.1	0.9;
	7896	2	0	0	0	0	0	1.054626	-12.568941	220	7	1.1	0.9;
	7897	1	3.2	0.6	0	0.82	0	0.993804	-28.622596	110	18	1.1	0.9;
	7898	1	-0	0	0	25.73	0	1.063653	22.361822	380	1	1.1	0.9;
	7899	1	31.74	6.3	0	0.5	0	0.994613	-26.228978	110	18	1.1	0.9;
	7900	1	26.93	9.6	0	1.41	0	1.023414	-27.227398	110	21	1.1	0.9;
	7901	1	191.7	6.8	0	11.58	0	0.988403	-42.964966	220	10	1.1	0.9;
	7902	1	0	0	0	0.15	0	1.022641	-15.117164	110	18	1.1	0.9;
	7903	1	0	-0	0	7.78	0	1.071405	-42.005492	220	5	1.1	0.9;
	7904	1	30.35	30.5	0	0.76	0	1.006917	-29.683321	150	9	1.1	0.9;
	7905	1	347.1	87.2	0	88.35	0	1.044096	-18.14373	380	5	1.1	0.9;
	7906	1	7	1	0	4.05	0	1.027039	-2.259422	154	1	1.1	0.9;
	7907	1	-0	-0	0	1.55	0	1.053712	-29.238447	150	9	1.1	0.9;
	7908	1	77.2	11	0	9.22	0	1.022759	14.628752	154	1	1.1	0.9;
	7909	1	0	0	0	0.33	0	1.018041	-22.978925	110	18	1.1	0.9;
	7910	1	-0	0	0	7.73	0	1.015669	0.040541	380	3	1.1	0.9;
	7911	1	0	8.3	0	10.19	0	1.084458	-36.846643	150	9	1.3	0.7;
	7912	1	0	0	0	0	0	0.93851	-33.36495	150	9	1.1	0.9;
	7913	2	0	0	0	0	0	1.0583	-9.156897	380	5	1.1	0.9;
	7914	1	-596.4	-284.9	0.302939	3.153549	0	1.057921	-7.786969	380	3	1.1	0.9;
	7915	1	-0.18	-1.31	0	0	0	1.014536	16.390148	150	8	1.1	0.9;
	7916	1	35.14	7.6	0	6.49	0	1.034073	-38.613553	110	22	1.1	0.9;
	7917	1	-0	0	0	2.74	0	1.0458	-16.592799	380	1	1.1	0.9;
	7918	1	-0	-0	0	117.47	0	1.024876	-47.970232	380	4	1.1	0.9;
	7919	1	-0	0	0	14.34	0	1.084213	45.203319	380	3	1.1	0.9;
	7920	1	160.3	89.7	0	47.28	0	1.01435	-8.809807	380	3	1.1	0.9;
	7921	1	38.14	23.6	0	0	0	1.04929	-22.14211	150	9	1.1	0.9;
	7922	1	0	0	0.12	20.1	0	1.035891	-12.993029	380	6	1.1	0.9;
	7923	1	0	-0	0	0.97	0	1.03458	-22.482591	220	5	1.1	0.9;
	7924	1	106.1	-6.8	0	22.23	0	1.021353	4.333441	380	3	1.1	0.9;
	7925	1	36.84	14.1	0	33.55	0	1.064269	-43.311179	380	4	1.1	0.9;
	7926	2	0	0	0	0	0	1.033819	-11.987518	154	1	1.1	0.9;
	7927	1	2.2	0.5	0	0.13	0	1.030813	-34.023248	220	10	1.1	0.9;
	7928	1	0	-0	0	65.24	0	1.019492	6.24274	400	1	1.1	0.9;
	7929	2	0	0	0	0	0	1.044801	-44.589679	380	4	1.1	0.9;
	7930	1	17.52	9.1	0	2.26	0	0.986061	-28.33566	110	18	1.1	0.9;
	7931	1	-0.67	-2.78	0	11.64	0	1.061809	-24.814153	220	23	1.1	0.9;
	7932	1	-0	0	0	0.77	0	1.039209	-21.588693	110	19	1.1	0.9;
	7933	1	-14.91	-5	0	0.57	0	1.00233	-20.357998	110	18	1.1	0.9;
	7934	2	0	0	0	0	0	1.090866	46.646981	380	3	1.1	0.9;
	7935	1	4.5	1.1	0	12.56	0	1.054273	-48.011301	380	10	1.1	0.9;
	7936	1	115.8	29.3	0	4.31	0	1.022726	-14.260765	220	7	1.1	0.9;
	7937	1	56.06	17	0	2.4	0	1.039373	-7.600731	220	5	1.1	0.9;
	7938	1	31.03	10.3	0	0.89	0	1.007696	-32.735194	110	21	1.1	0.9;
	7939	1	0	0	0	0	0	0.994734	-34.15431	150	9	1.1	0.9;
	7940	2	0	0	0	0	0	1.036846	-12.93378	380	6	1.1	0.9;
	7941	1	0	-0	0	4.31	0	1.035735	48.901735	220	3	1.1	0.9;
	7942	1	100.9	26.7	0	118.53	0	1.024216	18.003081	380	3	1.1	0.9;
	7943	1	38.44	2	0	0.56	0	1.047452	-16.169101	220	5	1.1	0.9;
	7944	1	0	0	0	2.12	0	1.015065	-2.592812	220	3	1.1	0.9;
	7945	1	17.92	5.8	0	0.84	0	1.02658	-6.07046	220	5	1.1	0.9;
	7946	1	-0	-0	0.181549	21.720889	0	1.064431	-25.178077	220	18	1.1	0.9;
	7947	1	0	0	0	0	0	1.028993	-31.533375	380	6	1.1	0.9;
	7948	1	56.56	3.5	0	0.63	0	1.000152	-27.243358	110	18	1.1	0.9;
	7949	2	0	0	0	0	0	1.033042	3.499382	154	1	1.1	0.9;
	7950	1	125.4	17.8	0	4.02	0	1.056077	3.782498	154	1	1.1	0.9;
	7951	1	156	60.7	0	1.52	0	1.008002	0.704779	150	8	1.1	0.9;
	7952	1	60.07	-0	0	0.04	0	1.008851	-32.506705	150	9	1.1	0.9;
	7953	2	0	0	0	0	0	1.067896	-2.73102	380	14	1.1	0.9;
	7954	1	0	-0	0	27.42	0	1.01999	3.773328	380	3	1.1	0.9;
	7955	1	111.6	8.9	0	16.19	0	1.061959	-23.678465	220	5	1.1	0.9;
	7956	1	-0	0	0	5.78	0	1.002622	-10.707448	220	1	1.1	0.9;
	7957	1	-0	-0	0	8.63	0	1.021124	-10.634569	220	3	1.1	0.9;
	7958	1	-0	-0	0	0.22	0	1.017087	-0.826561	150	8	1.1	0.9;
	7959	1	-0	-0	0	59.88	0	1.129512	-23.45798	380	9	1.3	0.7;
	7960	2	0	0	0	0	0	1.017582	6.120823	380	2	1.1	0.9;
	7961	1	50.56	8.1	0	8.02	0	1.032683	-8.672856	220	5	1.1	0.9;
	7962	1	0	-0	0	3.28	0	1.011183	-23.470952	154	1	1.1	0.9;
	7963	1	0	0	0	0	0	1.009172	-32.442398	150	9	1.1	0.9;
	7964	2	0	0	0	0	0	1.005985	-21.980811	154	1	1.1	0.9;
	7965	1	0	0	0	79.5	0	1.068758	-38.37097	380	1	1.1	0.9;
	7966	1	124.5	59.2	0	77.23	0	1.041514	-6.836651	380	3	1.1	0.9;
	7967	1	0	-0	0	5.97	0	1.037536	-11.708789	220	5	1.1	0.9;
	7968	1	4.81	1.6	0	0.19	0	1.028136	-34.08474	110	12	1.1	0.9;
	7969	1	0	-0	0.2527	79.80215	0	1.047495	-42.759729	380	10	1.1	0.9;
	7970	1	30.03	10	0	1.11	0	1.014231	-28.76162	110	21	1.1	0.9;
	7971	1	-0	0	0	21.6	0	1.030624	6.450427	380	3	1.1	0.9;
	7972	1	237.9	68.5	0	5.12	0	1.038447	-19.316628	220	5	1.1	0.9;
	7973	1	115.7	10.7	0	19.1	0	1.024463	-54.505855	220	4	1.1	0.9;
	7974	1	234	13.1	0	51.03	0	1.030171	-20.370156	220	5	1.1	0.9;
	7975	1	-0	0	0	0.59	0	0.967878	-25.140355	110	18	1.1	0.9;
	7976	1	0	0	0	1.57	0	0.915748	-35.933019	150	9	1.1	0.9;
	7977	1	110.2	15.7	0	1.11	0	1.016023	4.069925	154	1	1.1	0.9;
	7978	1	-0	0	0.12	15.18	0	1.027815	-26.568135	220	21	1.1	0.9;
	7979	2	0	0	0	0	0	1.064388	-21.628838	380	20	1.1	0.9;
	7980	1	233.5	151	0	35.31	0	1.041778	-43.928298	380	10	1.1	0.9;
	7981	1	3	1.5	0	0.4	0	1.031763	-28.453814	110	23	1.1	0.9;
	7982	1	112	19.7	0	38.04	0	1.032099	-23.4836	220	5	1.1	0.9;
	7983	2	0	0	0	0	0	1.014534	-4.149497	154	1	1.1	0.9;
	7984	1	126.98	12.83	0	9.13	0	1.02858	-4.971328	220	13	1.1	0.9;
	7985	1	-0	0	0	107.2	0	1.006287	7.522666	400	1	1.1	0.9;
	7986	1	62.7	9	0	5.83	0	1.044085	0.20693	154	1	1.1	0.9;
	7987	1	-0	-8.1	0	0	0	0.994584	-28.409429	150	9	1.1	0.9;
	7988	1	-3.55	-1.43	0	11.834121	0	1.028689	-2.162063	380	5	1.1	0.9;
	7989	1	0	0	0	0.27	0	1.049894	-3.935951	220	5	1.1	0.9;
	7990	1	12.61	7.8	0	2.42	0	0.960834	-27.235744	150	9	1.1	0.9;
	7991	1	-0	0	0	0.27	0	1.03465	-24.182375	110	19	1.1	0.9;
	7992	1	400.4	23.6	0.590405	133.314256	0	1.044474	-26.640277	380	11	1.1	0.9;
	7993	1	18.22	6.3	0	0.35	0	1.010579	-24.407164	110	18	1.1	0.9;
	7994	1	11.91	3.3	0	6.84	0	1.010173	-16.613735	220	5	1.1	0.9;
	7995	2	0	0	0	0	0	1.024627	12.839869	154	1	1.1	0.9;
	7996	1	0	-0	0	55	0	1.026529	-1.949583	150	8	1.1	0.9;
	7997	1	25.23	47.6	0	4.21	0	1.030629	9.964589	220	3	1.1	0.9;
	7998	2	0	0	0	0	0	1.046131	4.063717	380	5	1.1	0.9;
	7999	1	0	-0	0	9.68	0	1.023023	-48.741108	220	4	1.1	0.9;
	8000	1	-0	0	0	0.35	0	1.025636	-22.76283	150	9	1.1	0.9;
	8001	1	0	0	0	0	0	1.011753	63.312472	150	1	1.1	0.9;
	8002	1	0	0	0	0.36	0	1.003367	-31.347071	150	9	1.1	0.9;
	8003	1	-0	-0	0	0.76	0	1.039083	-32.196713	110	22	1.1	0.9;
	8004	1	-0	0	0	0.37	0	0.975044	-31.992071	150	9	1.1	0.9;
	8005	1	-0	-0	0	16.77	0	1.02503	-13.906514	220	5	1.1	0.9;
	8006	1	0	-0	0	21.18	0	1.02459	31.438507	380	3	1.1	0.9;
	8007	1	7.91	2.7	0	0.01	0	0.984871	-56.241859	150	10	1.1	0.9;
	8008	1	13.9	1.9	0	6.58	0	1.017556	-7.401892	154	1	1.1	0.9;
	8009	1	-0	-0	0	4.29	0	1.033583	-44.088164	380	6	1.1	0.9;
	8010	1	3.2	1.5	0	0.95	0	0.995026	-28.509398	110	18	1.1	0.9;
	8011	1	0	-0	0	0.35	0	0.909072	-36.225682	150	9	1.1	0.9;
	8012	1	36.84	14.1	0	17.84	0	1.055496	-40.813552	380	4	1.1	0.9;
	8013	1	40.04	15.1	0	1.05	0	0.974865	-30.795403	110	21	1.1	0.9;
	8014	1	149.63	59	0	4.99	0	1.007037	-37.083594	220	6	1.1	0.9;
	8015	1	28.83	10.4	0	0.26	0	1.005973	-28.439152	110	21	1.1	0.9;
	8016	1	23.33	9.1	0	15.36	0	1.024642	3.976217	150	8	1.1	0.9;
	8017	1	-0	-0	0	8.99	0	1.02388	11.172528	150	8	1.1	0.9;
	8018	1	9.41	2.3	0	1.18	0	0.990633	-21.238054	110	18	1.1	0.9;
	8019	1	0	0	0	2.59	0	0.971819	27.711925	220	3	1.1	0.9;
	8020	1	1.8	0.7	0	0.27	0	1.016375	-22.556812	110	19	1.1	0.9;
	8021	1	14.22	4.7	0	5.58	0	0.994734	-31.231931	220	10	1.1	0.9;
	8022	1	209.1	1.7	0	18.49	0	1.026477	-11.783692	220	3	1.1	0.9;
	8023	1	0	-0	0	4.32	0	1.078199	-35.260564	150	9	1.1	0.9;
	8024	1	-0	-0	0	0.11	0	1.007491	-30.159237	110	18	1.1	0.9;
	8025	1	-0	0	0	0.26	0	0.994462	-26.541378	110	18	1.1	0.9;
	8026	1	52.48	6.43	0	1.47	0	1.018061	61.87766	150	1	1.1	0.9;
	8027	1	0	-0	0	21.6	0	1.019202	-0.749794	380	3	1.1	0.9;
	8028	1	-0	-0	0	4.28	0	1.032406	-32.982123	220	1	1.1	0.9;
	8029	1	8.31	3	0	0.3	0	1.029909	-8.604911	110	19	1.1	0.9;
	8030	1	-0	0	0	122.13	0	1.029369	-4.649346	380	5	1.1	0.9;
	8031	1	12.91	4.2	0	0.35	0	1.00175	-39.804622	110	12	1.1	0.9;
	8032	1	-0	0	0	6.54	0	1.046606	64.452646	150	1	1.1	0.9;
	8033	1	-0	-0	0	4.79	0	1.030915	9.617072	220	3	1.1	0.9;
	8034	1	83.99	6	0	65.12	0	1.042279	-50.429478	380	4	1.1	0.9;
	8035	1	-0	0	0	4.22	0	1.020888	-16.538295	220	5	1.1	0.9;
	8036	2	0	0	0	0	0	1.029729	-48.20706	220	6	1.1	0.9;
	8037	1	0	-0	0	5.73	0	1.034424	9.023944	150	8	1.1	0.9;
	8038	1	12.11	5.1	0	6.68	0	0.926382	-30.635991	110	18	1.1	0.9;
	8039	1	27.9	4	0	4.85	0	1.03193	0.87619	154	1	1.1	0.9;
	8040	1	70.38	27.5	0	0.77	0	0.988535	-27.421635	110	3	1.1	0.9;
	8041	1	93.8	15.9	0	0	0	1.067309	-17.919785	110	24	1.1	0.9;
	8042	1	-0	0	0	1.01	0	1.075334	-26.411362	150	9	1.1	0.9;
	8043	2	0	0	0	0	0	1.058234	-11.690288	220	5	1.1	0.9;
	8044	2	0	0	0	0	0	1.041629	27.869278	110	8	1.1	0.9;
	8045	2	0	0	0	0	0	1.023558	-39.91457	110	22	1.1	0.9;
	8046	1	106.4	24.2	0	0.11	0	1.023854	-26.107145	110	19	1.1	0.9;
	8047	1	25	3.6	0	2.96	0	1.032226	-6.011158	154	1	1.1	0.9;
	8048	1	0	0	0	0	0	0.968795	-30.437259	150	9	1.1	0.9;
	8049	1	-0	-0	0	66.29	0	1.052765	-15.657994	380	6	1.1	0.9;
	8050	1	42.35	13.5	0	0.32	0	0.986637	-25.363262	110	18	1.1	0.9;
	8051	1	19.02	6	0	1.45	0	1.015284	-27.640966	110	21	1.1	0.9;
	8052	1	53.56	18.8	0	1.45	0	1.026897	-21.584056	220	20	1.1	0.9;
	8053	1	-0	-0	0	0.01	0	1.021077	-7.356538	154	1	1.1	0.9;
	8054	1	0	0	0	0	0	0.945107	-32.784726	150	9	1.1	0.9;
	8055	1	28.63	8.8	0	0.67	0	1.020361	-35.363588	110	22	1.1	0.9;
	8056	1	14.32	2.2	0	1.26	0	1.01142	-41.090225	110	22	1.1	0.9;
	8057	1	69.18	9.5	0	3.67	0	1.041648	-8.879856	220	5	1.1	0.9;
	8058	1	10.71	17.7	0	10.58	0	1.022907	11.155814	150	8	1.1	0.9;
	8059	1	170.8	70.4	0	7.23	0	0.990069	-21.418625	220	3	1.1	0.9;
	8060	1	0	-0	0	9.54	0	1.04261	-13.767109	220	5	1.1	0.9;
	8061	1	-19.32	6.4	0	0.23	0	0.969655	-24.039484	150	9	1.1	0.9;
	8062	1	25.33	8.9	0	5.26	0	1.01077	-24.081637	110	18	1.1	0.9;
	8063	1	36.9	5.3	0	3.7	0	1.019128	-5.785482	154	1	1.1	0.9;
	8064	1	135.2	70.6	0	9.37	0	0.95682	-37.631909	150	9	1.1	0.9;
	8065	2	0	0	0	0	0	1.059118	-21.729369	380	20	1.1	0.9;
	8066	2	0	0	0	0	0	1.011708	-47.36097	220	4	1.1	0.9;
	8067	1	66.77	19.8	0	0	0	1.008887	-14.633934	110	24	1.1	0.9;
	8068	1	41.74	7.4	0	18.54	0	1.02756	-54.000739	220	4	1.1	0.9;
	8069	1	174.2	54.3	0	23.93	0	1.025232	-48.020182	380	6	1.1	0.9;
	8070	1	3.8	1.3	0	1.11	0	1.001789	-29.37689	110	18	1.1	0.9;
	8071	1	0	0	0	0.48	0	0.992962	-21.379846	110	18	1.1	0.9;
	8072	1	-0	-0	0	0.76	0	1.032637	-3.253164	220	1	1.1	0.9;
	8073	1	68.48	42.7	0	6.37	0	1.033677	-5.641487	220	3	1.1	0.9;
	8074	1	33.14	20.5	0	0.54	0	1.015722	-28.98154	150	9	1.1	0.9;
	8075	1	14.2	2	0	1.34	0	1.00404	-10.66813	154	1	1.1	0.9;
	8076	1	6.21	1.5	0	1.42	0	0.992767	-27.64354	110	18	1.1	0.9;
	8077	1	-0	0	0	0.71	0	1.056006	-27.347221	150	9	1.1	0.9;
	8078	1	130	18.5	0	0.62	0	1.036003	-22.090523	154	1	1.1	0.9;
	8079	1	0	0	0	0.81	0	1.008651	-48.900733	220	4	1.1	0.9;
	8080	1	0	0	0	0	0	1.028921	-29.481204	150	9	1.1	0.9;
	8081	1	35.95	22.3	0	0	0	1.033921	-29.923351	150	9	1.1	0.9;
	8082	1	-0.65	-3.26	0	0	0	1.081938	-35.99417	150	9	1.1	0.9;
	8083	1	92.9	13.2	0	4.15	0	1.024659	5.632991	154	1	1.1	0.9;
	8084	1	35.14	11	0	0.48	0	1.016384	-41.084893	110	22	1.1	0.9;
	8085	2	0	0	0	0	0	1.020835	9.051289	154	1	1.1	0.9;
	8086	1	0	0	0	17.69	0	1.035004	13.443262	380	8	1.1	0.9;
	8087	2	0	0	0	0	0	1.027752	-38.784819	220	4	1.1	0.9;
	8088	1	0	-0	0.379958	52.33965	0	1.036504	-51.990407	380	10	1.1	0.9;
	8089	1	316.1	45	0	15.2	0	1.005402	-19.565573	400	1	1.1	0.9;
	8090	1	39.34	24.4	0	0	0	1.004931	-31.840868	150	9	1.1	0.9;
	8091	1	73.18	23.4	0	0.17	0	1.010271	-22.903972	110	19	1.1	0.9;
	8092	1	-0	0	0	0.32	0	0.980217	-23.636295	150	9	1.1	0.9;
	8093	1	41.85	15.2	0	0.21	0	1.031294	-7.30569	110	19	1.1	0.9;
	8094	1	0	0	0	0	0	1.015722	-28.98154	150	9	1.1	0.9;
	8095	1	21.88	4.73	0	-8.06	0	1.016801	62.369068	150	1	1.1	0.9;
	8096	2	0	0	0	0	0	1.065973	-25.177832	110	21	1.1	0.9;
	8097	1	7.61	0	0	6.6	0	0.925909	-30.7488	110	18	1.1	0.9;
	8098	1	206.4	21	0	8.55	0	0.988227	-50.353409	150	10	1.1	0.9;
	8099	1	0	0	0	13.33	0	1.021862	4.491551	380	3	1.1	0.9;
	8100	1	25	3.6	0	1.31	0	1.01468	2.038728	154	1	1.1	0.9;
	8101	2	0	0	0	0	0	1.059601	-20.726175	220	11	1.1	0.9;
	8102	1	82.1	29.1	0	0.4	0	1.020085	-30.999154	110	23	1.1	0.9;
	8103	1	39.34	-10.7	0	0	0	0.983672	-2.039135	150	2	1.1	0.9;
	8104	1	127.2	40.8	0	0.2	0	1.065323	-25.565009	220	5	1.1	0.9;
	8105	2	0	0	0	0	0	1.013571	-24.29809	154	1	1.1	0.9;
	8106	1	59.9	8.5	0	0.88	0	1.007384	-8.03601	154	1	1.1	0.9;
	8107	1	68.18	16.1	0	3.35	0	1.030711	-23.716241	220	5	1.1	0.9;
	8108	1	35.84	13	0	2.26	0	1.03596	-24.818001	110	19	1.1	0.9;
	8109	2	0	0	0	0	0	1.059636	-21.41092	380	5	1.1	0.9;
	8110	1	0	-0	0	0.03	0	1.02145	4.543192	400	1	1.1	0.9;
	8111	1	-0	-0	0	4.13	0	1.014784	-32.577103	150	9	1.1	0.9;
	8112	1	87.9	29.3	0	1.21	0	1.027125	-3.966359	220	5	1.1	0.9;
	8113	1	107	43	0	8.34	0	1.000699	-7.014879	220	3	1.1	0.9;
	8114	1	-0	0	0	30.39	0	1.008795	10.997705	380	8	1.1	0.9;
	8115	1	19.02	6.3	0	0.2	0	0.988958	-18.000178	110	18	1.1	0.9;
	8116	2	0	0	0	0	0	1.035167	-21.323561	154	1	1.1	0.9;
	8117	2	0	0	0	0	0	1.032328	-40.259667	380	6	1.1	0.9;
	8118	1	111.1	35.6	0	0.53	0	1.019467	-29.57919	110	3	1.1	0.9;
	8119	2	0	0	0	0	0	1.073486	-22.807799	110	15	1.1	0.9;
	8120	1	0.01	-0	0	26.46	0	1.021706	-46.288927	380	6	1.1	0.9;
	8121	1	0	0	0	1.25	0	1.041828	2.490809	154	1	1.1	0.9;
	8122	1	164.4	23.4	0	10.01	0	1.010402	-25.765276	154	1	1.1	0.9;
	8123	2	0	0	0	0	0	1.009241	13.00467	220	3	1.1	0.9;
	8124	1	80.49	7.3	0	1.81	0	1.009707	-61.020742	220	4	1.1	0.9;
	8125	1	-0	-0	0	7.21	0	1.033398	-2.906942	220	1	1.1	0.9;
	8126	1	-0	0	0	23.91	0	1.050536	-13.828695	380	1	1.1	0.9;
	8127	1	8.91	2.5	0	0.47	0	0.963758	-29.170502	110	18	1.1	0.9;
	8128	1	-0	0	0	8.1	0	1.087364	-9.876011	380	5	1.1	0.9;
	8129	1	-0	0	0	1.4	0	1.015661	61.512735	150	1	1.1	0.9;
	8130	1	34.44	27	0	0.34	0	0.893435	-35.853165	150	9	1.3	0.7;
	8131	1	-0	0	0	0.23	0	0.993837	-33.859399	150	9	1.1	0.9;
	8132	1	12.61	6.3	0	14.98	0	1.032999	-3.210999	220	1	1.1	0.9;
	8133	1	19.4	2.8	0	3.46	0	1.029563	3.621096	154	1	1.1	0.9;
	8134	2	0	0	0	0	0	1.008958	13.021747	220	3	1.1	0.9;
	8135	1	47.4	6.7	0	3.82	0	1.020133	-6.60554	154	1	1.1	0.9;
	8136	2	0	0	0	0	0	1.028389	-9.568889	154	1	1.1	0.9;
	8137	1	0	0	0.04975	68.521798	0	1.066186	-22.897807	380	18	1.1	0.9;
	8138	1	-0	0	0	6.66	0	1.046996	-43.360058	220	4	1.1	0.9;
	8139	1	242.4	46.7	0	75.67	0	1.0271	-31.739536	380	6	1.1	0.9;
	8140	1	18.72	6.6	0	0.51	0	1.025442	-36.544956	110	22	1.1	0.9;
	8141	1	0	0	0	50	0	1.045799	2.668899	150	8	1.1	0.9;
	8142	1	10.01	2	0	0.73	0	1.015771	-26.476212	150	7	1.1	0.9;
	8143	1	22.32	4	0	0.45	0	1.066559	-31.551132	110	3	1.1	0.9;
	8144	1	0	0	0	0.27	0	1.032076	-6.969711	110	19	1.1	0.9;
	8145	1	169.93	42.88	0	9.84	0	1.037076	-38.14014	220	17	1.1	0.9;
	8146	1	0	0	0	0	0	1.012743	-30.260645	150	9	1.1	0.9;
	8147	1	303.5	56.2	0	21.07	0	1.011517	-48.229112	220	4	1.1	0.9;
	8148	1	3.8	1.9	0	0.72	0	1.008977	-19.643517	110	18	1.1	0.9;
	8149	1	111.36	27.44	0	2.59	0	1.001472	-46.548205	220	4	1.1	0.9;
	8150	1	35.1	5	0	1.57	0	1.019025	-5.948255	154	1	1.1	0.9;
	8151	1	0	0	0	14.23	0	1.040089	-25.913804	220	5	1.1	0.9;
	8152	1	0	0	0	1.28	0	0.979646	-37.165348	150	9	1.1	0.9;
	8153	1	21.52	5.1	0	0.25	0	1.010673	-24.161118	110	18	1.1	0.9;
	8154	1	-0	0	0	0.15	0	1.049994	-24.715315	150	9	1.1	0.9;
	8155	1	125.3	34.6	0	4.36	0	1.019325	8.943633	150	8	1.1	0.9;
	8156	2	0	0	0	0	0	1.075225	8.81697	380	14	1.1	0.9;
	8157	2	0	0	0	0	0	1.045038	-39.487415	380	6	1.1	0.9;
	8158	2	0	0	0	0	0	1.051259	-0.62403	220	5	1.1	0.9;
	8159	2	0	0	0	0	0	1.060727	-40.97699	380	4	1.1	0.9;
	8160	1	0	0	0	0	0	1.005675	-32.160727	150	9	1.1	0.9;
	8161	1	-34.04	-2	0	197.23	0	1.033444	-44.087427	380	6	1.1	0.9;
	8162	1	-0	0	0	1.28	0	0.990709	-29.416452	150	9	1.1	0.9;
	8163	1	76.58	16.2	0	21.54	0	1.021917	-54.195819	220	4	1.1	0.9;
	8164	1	1.8	0.6	0	0.43	0	1.017961	6.452451	150	8	1.1	0.9;
	8165	1	0	-0	0	2.42	0	1.061672	-3.668047	220	5	1.1	0.9;
	8166	1	-0	-0	0.090774	5.709044	0	1.061057	-26.449392	220	18	1.1	0.9;
	8167	1	-0	0	0	1.22	0	1.014897	4.500754	150	8	1.1	0.9;
	8168	1	20.62	7.5	0	0.79	0	1.056007	-20.363996	110	19	1.1	0.9;
	8169	1	-0	0	0	0.39	0	0.994001	-30.574954	150	9	1.1	0.9;
	8170	1	-12.15	-0.84	0	0	0	1.05125	-31.357019	220	6	1.1	0.9;
	8171	1	2.8	14.6	0	113.059828	0	1.128618	-22.125742	380	9	1.3	0.7;
	8172	1	0	-0	0	5.39	0	1.045343	-28.928507	220	12	1.1	0.9;
	8173	1	22.12	8	0	0.53	0	1.012811	-29.203679	110	19	1.1	0.9;
	8174	1	0	-0	0	0.79	0	1.06297	-26.062623	110	19	1.1	0.9;
	8175	1	13.11	3.6	0	0.38	0	1.027698	-8.665439	220	13	1.1	0.9;
	8176	1	10.81	3.6	0	0.38	0	1.004168	-39.342733	110	12	1.1	0.9;
	8177	1	41.8	5.9	0	10.67	0	1.019944	-8.428419	154	1	1.1	0.9;
	8178	1	0	0	0	0	0	0.999708	-27.06631	150	9	1.1	0.9;
	8179	1	-0	0	0.194464	19.1468	0	0.996578	-5.601511	380	2	1.1	0.9;
	8180	1	291.6	47	0	2.38	0	1.073823	-30.874994	220	5	1.1	0.9;
	8181	1	0	-0	0.210824	6.620599	0	0.997673	-5.226676	380	2	1.1	0.9;
	8182	1	73.58	-13.1	0	7.58	0	1.02935	-20.100872	220	3	1.1	0.9;
	8183	1	0	0	0	2.97	0	1.0595	-47.950868	380	10	1.1	0.9;
	8184	1	-0	0	0	3.33	0	1.0152	-32.505094	150	9	1.1	0.9;
	8185	2	0	0	0	0	0	1.028753	-37.858554	380	6	1.1	0.9;
	8186	1	99.6	22.9	0	15.02	0	1.026419	-54.27643	220	4	1.1	0.9;
	8187	2	0	0	0	0	0	1.0215	2.359673	380	3	1.1	0.9;
	8188	1	120.2	30.7	0	9.98	0	1.04758	42.736603	220	3	1.1	0.9;
	8189	1	0	-0	0	9.54	0	1.045168	-18.168399	220	5	1.1	0.9;
	8190	1	0	-0	0	0.32	0	1.042647	-10.60065	220	5	1.1	0.9;
	8191	1	0	0	0	11.38	0	1.04325	-15.664936	220	5	1.1	0.9;
	8192	1	2	1	0	1	0	0.979985	-31.250322	110	21	1.1	0.9;
	8193	1	0	0	0.564887	82.937717	0	1.050524	41.986713	380	3	1.1	0.9;
	8194	1	-0.18	-1	0	0	0	1.021167	10.861415	150	8	1.1	0.9;
	8195	1	-15.25	-6.42	0	43.956703	0	1.022307	-17.099911	380	5	1.1	0.9;
	8196	1	20.2	2.9	0	5.89	0	1.012285	1.194766	154	1	1.1	0.9;
	8197	1	-134.35	-6.5	0	3.38	0	1.003237	-38.334059	150	10	1.1	0.9;
	8198	1	165.9	49.6	0	6.13	0	1.026529	24.114895	220	3	1.1	0.9;
	8199	1	169.6	98.3	0	27.8	0	1.013013	-15.052128	380	3	1.1	0.9;
	8200	1	0	-0	0	39.58	0	1.013227	-6.532335	380	5	1.1	0.9;
	8201	1	0.18	87.5	0	3.94	0	1.011753	63.312472	150	1	1.1	0.9;
	8202	1	11.8	1.7	0	6.98	0	1.029197	5.912943	154	1	1.1	0.9;
	8203	2	0	0	0	0	0	1.008878	-5.25837	154	1	1.1	0.9;
	8204	1	-0	-0	0	0.25	0	1.037372	-0.720895	150	8	1.1	0.9;
	8205	1	4	1.3	0	0.62	0	0.998408	-38.372067	110	12	1.1	0.9;
	8206	1	0	0	0.062219	26.452656	0	1.018881	65.830577	380	1	1.1	0.9;
	8207	2	0	0	0	0	0	1.029137	-42.518288	220	4	1.1	0.9;
	8208	1	3.8	3.8	0	0.95	0	0.950649	-30.58154	110	18	1.1	0.9;
	8209	1	83.39	16.5	0	16.43	0	1.029657	-20.451282	220	5	1.1	0.9;
	8210	2	0	0	0	0	0	1.044939	-38.028359	380	4	1.1	0.9;
	8211	1	-0	0	0	65.93	0	1.023364	19.415183	380	8	1.1	0.9;
	8212	2	0	0	0	0	0	1.009223	-27.52837	110	18	1.1	0.9;
	8213	1	207.9	25.4	0	1.75	0	1.033429	-46.419035	220	4	1.1	0.9;
	8214	1	58.97	11.1	0	1.05	0	1.037117	-11.762803	220	5	1.1	0.9;
	8215	2	0	0	0	0	0	1.031215	-34.760827	220	6	1.1	0.9;
	8216	1	0	-0	0	10.28	0	1.001625	-8.149634	220	3	1.1	0.9;
	8217	1	63.77	39.5	0	1.12	0	0.962201	-32.513552	150	9	1.1	0.9;
	8218	2	0	0	0	0	0	1.013588	-21.939058	154	1	1.1	0.9;
	8219	1	0	-0	0	0.6	0	1.071897	-18.921301	110	19	1.1	0.9;
	8220	1	-0	-0	0	0.35	0	1.066808	-21.238107	150	9	1.1	0.9;
	8221	1	-0	-0	0	0	0	1.04837	-20.841946	220	1	1.1	0.9;
	8222	2	0	0	0	0	0	1.013304	-10.689171	380	5	1.1	0.9;
	8223	2	0	0	0	0	0	1.056121	17.316391	380	8	1.1	0.9;
	8224	1	92.7	-5.6	0.2527	35.620175	0	1.048689	-42.216026	380	10	1.1	0.9;
	8225	2	0	0	0	0	0	1.062685	-12.260471	220	5	1.1	0.9;
	8226	1	65.5	9.3	0	7.21	0	1.018717	4.840908	154	1	1.1	0.9;
	8227	2	0	0	0	0	0	1.022197	12.147234	400	1	1.1	0.9;
	8228	1	-0	0	0.500102	38.553095	0	1.049799	-35.627345	380	11	1.1	0.9;
	8229	1	0	0	0	0	0	1.059236	-7.623477	150	2	1.1	0.9;
	8230	1	0	0	0	59.43	0	1.035365	11.07861	380	8	1.1	0.9;
	8231	1	0	0	0	0.32	0	0.975419	-31.968769	150	9	1.1	0.9;
	8232	1	1.8	0.7	0	0.85	0	1.030544	-20.518157	110	19	1.1	0.9;
	8233	1	0	0	0	1.5	0	1.008979	-27.229246	150	9	1.1	0.9;
	8234	1	0	0	0	0	0	1.022901	-32.41943	150	9	1.1	0.9;
	8235	1	1.9	1.3	0	0.43	0	1.049964	-16.811641	110	18	1.1	0.9;
	8236	1	118.1	20.2	0	1.47	0	1.022936	-54.333594	220	4	1.1	0.9;
	8237	1	370.2	107.7	0	81.59	0	1.02248	-41.049727	380	6	1.1	0.9;
	8238	1	0	0	0	0	0	1.050287	-24.708967	150	9	1.1	0.9;
	8239	1	16.52	6.6	0	0.65	0	1.013181	-40.509773	110	22	1.1	0.9;
	8240	2	0	0	0	0	0	1.056841	-39.378648	380	4	1.1	0.9;
	8241	1	5.11	2.5	0	0.19	0	1.05487	-17.180585	110	18	1.1	0.9;
	8242	2	0	0	0	0	0	1.010793	-46.562916	220	4	1.1	0.9;
	8243	1	44.75	27.7	0	1.14	0	0.996999	-32.232554	150	9	1.1	0.9;
	8244	1	0	0	0	8.13	0	1.020146	0.682307	380	3	1.1	0.9;
	8245	1	216.8	30.7	0	58.2	0	1.052629	-18.762539	380	16	1.1	0.9;
	8246	1	50.76	17.5	0	8.97	0	1.017932	-29.281953	220	6	1.1	0.9;
	8247	2	0	0	0	0	0	1.055867	-41.040462	380	4	1.1	0.9;
	8248	1	0	0	0	8.88	0	1.004094	-4.515905	400	1	1.1	0.9;
	8249	1	0	0	0	17.8	0	1.014077	11.203194	380	8	1.1	0.9;
	8250	1	196.1	32.1	0	3.86	0	1.026211	-26.484125	220	5	1.1	0.9;
	8251	1	9.61	2.5	0	0.37	0	0.977611	-28.394971	110	18	1.1	0.9;
	8252	2	0	0	0	0	0	1.02226	-43.190789	380	6	1.1	0.9;
	8253	1	-0.57	-1	0	3.29	0	1.055141	-29.298438	150	9	1.1	0.9;
	8254	1	-0	0	0	1.56	0	1.010885	-5.581506	154	1	1.1	0.9;
	8255	1	39.04	13.1	0	3.22	0	1.019596	-13.990643	220	5	1.1	0.9;
	8256	1	-0.24	-23.15	0	26.49	0	1.058137	18.767735	380	3	1.1	0.9;
	8257	1	70.08	43.4	0	1.83	0	0.967209	-32.44223	150	9	1.1	0.9;
	8258	1	0	-0	0	69.23	0	1.146803	-24.522183	380	23	1.3	0.7;
	8259	1	19.4	2.8	0	6.62	0	1.047063	2.954121	154	1	1.1	0.9;
	8260	1	0	-0	0	10.21	0	1.042642	-50.293244	380	4	1.1	0.9;
	8261	1	69.38	22.5	0	0.67	0	1.02553	-28.425864	110	21	1.1	0.9;
	8262	1	7	1	0	4	0	1.010184	-8.519846	154	1	1.1	0.9;
	8263	1	255.9	128.3	0	54.58	0	1.037171	-41.654202	380	17	1.1	0.9;
	8264	1	0	-0	0	65.73	0	1.02338	19.408295	380	8	1.1	0.9;
	8265	1	40.44	20.5	0	0.85	0	1.059858	-24.193736	220	5	1.1	0.9;
	8266	1	316.7	46.1	0	46.23	0	1.01568	-49.022963	380	6	1.1	0.9;
	8267	2	0	0	0	0	0	1.042576	-10.311578	380	5	1.1	0.9;
	8268	1	0	-0	0	24.67	0	1.129332	-26.334883	380	1	1.3	0.7;
	8269	2	0	0	0	0	0	1.039628	-19.617502	110	19	1.1	0.9;
	8270	1	-632.1	-447	0	6.4	0	1.031891	17.801282	380	3	1.1	0.9;
	8271	1	15.62	2.8	0	0.71	0	0.960551	-34.800379	110	18	1.1	0.9;
	8272	1	7.41	2.7	0	1.28	0	1.073949	-25.312276	110	19	1.1	0.9;
	8273	1	120.6	35.7	0	15.36	0	1.024472	-35.07305	380	6	1.1	0.9;
	8274	1	15.62	5.7	0	0.17	0	1.031672	-24.817077	110	19	1.1	0.9;
	8275	1	40.05	20	0	7.45	0	1.00419	-50.596527	220	6	1.1	0.9;
	8276	1	73.68	15.4	0	3.73	0	1.019363	-24.62562	220	7	1.1	0.9;
	8277	1	57.66	6.1	0	1.77	0	1.001417	21.455311	110	8	1.1	0.9;
	8278	2	0	0	0	0	0	1.005256	-32.487725	150	10	1.1	0.9;
	8279	1	-0	-0	0	0.62	0	1.064328	-9.14908	380	14	1.1	0.9;
	8280	1	14.32	5.5	0	3.15	0	0.989209	-42.430607	110	22	1.1	0.9;
	8281	1	9.41	5.6	0	0.03	0	1.016085	-32.352858	150	9	1.1	0.9;
	8282	1	-0	0	0	36.43	0	1.085413	-13.321452	380	1	1.1	0.9;
	8283	1	-0	-0	0	0.49	0	0.96539	-34.440456	110	18	1.1	0.9;
	8284	1	-0.12	-1.31	0	0	0	1.079746	31.595713	380	8	1.1	0.9;
	8285	1	156.6	16.9	0	12.26	0	1.010139	-54.759962	220	4	1.1	0.9;
	8286	1	0	-0	0	0.3	0	0.998055	-16.505296	110	18	1.1	0.9;
	8287	1	0	-0	0	2.56	0	1.031132	-23.670977	220	3	1.1	0.9;
	8288	1	16.02	5.1	0	8.6	0	1.058196	10.347966	110	8	1.1	0.9;
	8289	1	-29.84	-2.96	0	0.28	0	1.044611	2.524063	110	18	1.1	0.9;
	8290	1	0	0	0	1.06	0	1.000755	-29.325722	110	18	1.1	0.9;
	8291	1	0	-0	0	6.52	0	1.04203	-7.751434	220	5	1.1	0.9;
	8292	1	141.5	70.8	0	4.92	0	1.013491	-16.872272	380	3	1.1	0.9;
	8293	1	112.7	14.5	0	1.24	0	1.043878	-25.136485	220	5	1.1	0.9;
	8294	1	14.22	6.3	0	0.38	0	1.040272	-12.207544	220	5	1.1	0.9;
	8295	1	5.01	2	0	0.6	0	1.061757	-22.866511	110	23	1.1	0.9;
	8296	1	0	-0	0	5.81	0	1.003	-36.6973	150	9	1.1	0.9;
	8297	2	0	0	0	0	0	1.024007	-1.464406	154	1	1.1	0.9;
	8298	1	0	0	0	14.35	0	1.079416	31.580844	380	8	1.1	0.9;
	8299	1	0	0	0	0	0	1.052737	-29.321529	150	9	1.1	0.9;
	8300	1	0	0	0	15	0	1.083791	25.660356	220	3	1.1	0.9;
	8301	1	59.67	9.2	0	18.03	0	1.022753	-52.026401	220	4	1.1	0.9;
	8302	1	13.92	3.8	0	1.26	0	1.009267	-29.522495	110	18	1.1	0.9;
	8303	1	-0	0	0	0.93	0	1.008328	-24.655408	150	9	1.1	0.9;
	8304	2	0	0	0	0	0	1.036714	-13.009231	380	6	1.1	0.9;
	8305	1	-0	0	0	8.18	0	1.023509	2.100177	400	1	1.1	0.9;
	8306	1	99.3	23.6	0	1.77	0	1.022295	-28.392614	110	21	1.1	0.9;
	8307	1	-0	0	0	0.14	0	1.049312	-4.132517	220	5	1.1	0.9;
	8308	1	33.44	11	0	4	0	1.009947	-26.481593	110	21	1.1	0.9;
	8309	1	11.2	1.6	0	2.09	0	1.003905	-12.026506	154	1	1.1	0.9;
	8310	1	-0	0	0	2.46	0	1.050396	-8.818559	220	5	1.1	0.9;
	8311	2	0	0	0	0	0	1.041866	-15.286135	220	5	1.1	0.9;
	8312	2	0	0	0	0	0	1.075986	-12.446539	380	5	1.1	0.9;
	8313	1	-9.75	-2.74	0	0.41	0	1.026214	-19.752056	110	18	1.1	0.9;
	8314	1	13.92	8.9	0	2.65	0	0.978742	-21.555476	110	18	1.1	0.9;
	8315	1	19.32	7	0	0.44	0	0.991615	-28.318119	110	19	1.1	0.9;
	8316	1	0	0	0	0.72	0	1.080774	-26.771111	220	5	1.1	0.9;
	8317	2	0	0	0	0	0	1.000913	-2.045998	400	1	1.1	0.9;
	8318	1	-0	-0	0	0	0	1.04781	18.206206	380	8	1.1	0.9;
	8319	1	18.02	6.5	0	2.19	0	1.033139	-28.365301	110	19	1.1	0.9;
	8320	1	313.75	96.8	0	5.67	0	0.995898	-51.607781	220	6	1.1	0.9;
	8321	1	74.78	26.4	0	2.42	0	1.013389	-41.474694	110	22	1.1	0.9;
	8322	1	0	0	0.278602	31.723346	0	1.070507	-15.352172	380	19	1.1	0.9;
	8323	1	8.3	1.2	0	2.86	0	1.031263	12.95326	154	1	1.1	0.9;
	8324	1	0	-0	0	3.03	0	1.025361	-24.322068	220	3	1.1	0.9;
	8325	1	41.45	56.2	0	6.55	0	1.015882	-38.437584	220	17	1.1	0.9;
	8326	1	0	0	0	0.29	0	1.040253	8.308249	154	1	1.1	0.9;
	8327	1	-0	-8	0	0	0	1.002607	-34.537706	150	9	1.1	0.9;
	8328	1	82.39	14.4	0	13.37	0	1.038918	-24.742845	220	5	1.1	0.9;
	8329	1	0	0	0.149712	2.04987	0	1.010013	5.350987	380	2	1.1	0.9;
	8330	1	89.44	63.66	0	2.6	0	1.030157	1.289604	220	13	1.1	0.9;
	8331	1	-0	-0	0	0.05	0	1.053965	-13.3494	220	5	1.1	0.9;
	8332	1	0	0	0	9.41	0	1.064955	53.263257	380	3	1.1	0.9;
	8333	1	174.7	31	0.216349	27.621586	0	1.064258	56.475531	380	3	1.1	0.9;
	8334	1	0	0	0	33.298535	0	1.037243	-14.775985	380	5	1.1	0.9;
	8335	1	637.92	139.57	0	0	0	1.007944	-11.993469	150	2	1.1	0.9;
	8336	1	257.6	-303.12	0.244607	64.309042	0	1.019599	-0.404353	380	3	1.1	0.9;
	8337	1	149.9	-0	0	28.38	0	1.046451	-43.209981	380	4	1.1	0.9;
	8338	1	-2.3	-3.34	0	8.15	0	1.021297	-44.458579	220	4	1.1	0.9;
	8339	1	0	0	0	0.3	0	1.029002	-30.391257	110	23	1.1	0.9;
	8340	2	0	0	0	0	0	1.020459	-32.098619	150	9	1.1	0.9;
	8341	1	133.7	19	0	4.05	0	1.019672	-12.702071	154	1	1.1	0.9;
	8342	1	97	-539.3	0	50.89	0	1.045055	16.16968	380	3	1.1	0.9;
	8343	1	44.05	14.4	0	0.92	0	0.967673	-31.207462	110	21	1.1	0.9;
	8344	1	0	-0	0	58.09	0	1.140155	-22.513812	380	15	1.3	0.7;
	8345	1	19.2	9.3	0	128.079082	0	1.102477	-22.522128	380	9	1.3	0.7;
	8346	1	80.29	-4.1	0	0.43	0	1.02704	7.750185	220	3	1.1	0.9;
	8347	1	-2.09	-1.43	0	48.358243	0	0.984674	-18.922895	380	5	1.1	0.9;
	8348	1	2.5	1.3	0	0.62	0	0.994914	-24.815343	110	18	1.1	0.9;
	8349	1	0.1	-0.9	0	0.91	0	1.030715	38.433991	220	3	1.1	0.9;
	8350	1	92.3	13.1	0	2.09	0	1.024609	-11.590033	154	1	1.1	0.9;
	8351	1	14.86	2.99	0	1.24	0	1.013482	-18.953127	110	18	1.1	0.9;
	8352	1	67.27	15.9	0	2.65	0	1.019463	-39.669039	220	4	1.1	0.9;
	8353	1	155.1	39.9	0	6.37	0	1.023371	-56.50364	220	4	1.1	0.9;
	8354	1	32.84	20.3	0	7.23	0	1.046637	-27.138917	150	9	1.1	0.9;
	8355	1	0	0	0	8.82	0	1.019471	2.927282	380	3	1.1	0.9;
	8356	1	-0	-0	0	0.72	0	1.021663	0.83732	150	8	1.1	0.9;
	8357	1	50.96	17	0	2.66	0	1.039159	-30.106392	110	21	1.1	0.9;
	8358	1	0	0	0.135493	21.11062	0	1.088958	-23.293723	380	21	1.1	0.9;
	8359	1	27.9	4	0	2.23	0	1.037988	0.809889	154	1	1.1	0.9;
	8360	1	2.8	1	0	0.74	0	1.05831	-18.522173	110	19	1.1	0.9;
	8361	1	126.54	29.1	0	12.55	0	1.06792	-26.821509	220	5	1.1	0.9;
	8362	2	0	0	0	0	0	1.047632	-10.195555	220	7	1.1	0.9;
	8363	1	6.31	2.5	0	0.16	0	0.983771	-28.778868	110	18	1.1	0.9;
	8364	1	12.01	4	0	0.72	0	0.969193	-33.842232	110	21	1.1	0.9;
	8365	1	69.7	9.9	0	6.07	0	1.023365	-7.444922	154	1	1.1	0.9;
	8366	1	66.4	9.4	0	4.35	0	1.033976	-10.458179	154	1	1.1	0.9;
	8367	1	77.99	19.3	0	3.04	0	1.033693	-6.759533	220	5	1.1	0.9;
	8368	1	-0	0	0	4.11	0	1.013549	-32.972603	150	9	1.1	0.9;
	8369	1	-0	-0	0	1.23	0	1.025353	-9.248105	220	5	1.1	0.9;
	8370	1	81.2	23.4	0	52.6	0	1.051458	32.794035	380	3	1.1	0.9;
	8371	2	0	0	0	0	0	1.107597	-29.894132	380	12	1.3	0.7;
	8372	1	14.72	5.3	0	0.91	0	1.003475	-27.753131	110	19	1.1	0.9;
	8373	1	216.3	-6.9	0	2.44	0	1.068117	-48.408259	220	5	1.1	0.9;
	8374	1	153.4	86.8	0	53.98	0	1.057625	-42.495639	380	4	1.1	0.9;
	8375	2	0	0	0	0	0	1.024751	-36.913054	380	6	1.1	0.9;
	8376	1	0	0	0	33.61	0	1.084755	-14.407486	380	1	1.3	0.7;
	8377	2	0	0	0	0	0	1.011199	-25.896242	110	18	1.1	0.9;
	8378	2	0	0	0	0	0	1.020496	-38.748447	220	4	1.1	0.9;
	8379	1	-0	0	0	2.28	0	0.950343	-33.723168	150	9	1.1	0.9;
	8380	1	0	-0	0	0.22	0	1.011417	18.74035	110	8	1.1	0.9;
	8381	1	-0	-0	0	17.92	0	1.06566	-45.192169	380	1	1.1	0.9;
	8382	1	-0	-0	0	1.32	0	1.013369	-29.690839	110	21	1.1	0.9;
	8383	2	0	0	0	0	0	1.008495	-16.34939	220	6	1.1	0.9;
	8384	2	0	0	0	0	0	1.051363	-6.810401	154	1	1.1	0.9;
	8385	1	0	-0	0	0.79	0	1.071133	-22.308506	150	9	1.1	0.9;
	8386	1	6.73	14.7	0	11.51	0	1.022901	-32.41943	150	9	1.1	0.9;
	8387	1	115.03	2.21	0	10.42	0	1.030064	7.578606	220	13	1.1	0.9;
	8388	1	96.8	76.1	0.23521	14.338731	0	1.004349	31.472688	380	3	1.1	0.9;
	8389	1	72.18	49.6	0	37.52	0	1.017477	16.817296	150	8	1.1	0.9;
	8390	1	107.9	15.3	0	99.52	0	1.011409	-21.678886	400	1	1.1	0.9;
	8391	1	4.2	0.6	0	2.04	0	1.025477	3.865363	154	1	1.1	0.9;
	8392	1	100.8	32.3	0	7.16	0	1.046081	6.943122	150	8	1.1	0.9;
	8393	1	0	0	0	37.57	0	1.013216	16.740162	380	8	1.1	0.9;
	8394	1	70.88	43.9	0	3.65	0	0.995125	-30.856007	150	9	1.1	0.9;
	8395	1	117.13	33.86	0	6.04	0	1.027319	-22.979397	220	7	1.1	0.9;
	8396	1	131.7	32.9	0	108.56	0	1.058922	-41.288699	380	4	1.1	0.9;
	8397	1	-0	0	0	4.05	0	1.044077	-25.82874	220	5	1.1	0.9;
	8398	2	0	0	0	0	0	1.034241	-33.913063	110	12	1.1	0.9;
	8399	1	79.39	24.7	0	198.67	0	1.033814	-39.750205	380	6	1.1	0.9;
	8400	1	15.82	9.5	0	0.02	0	1.03649	-29.85563	150	9	1.1	0.9;
	8401	2	0	0	0	0	0	1.074893	4.701019	380	13	1.1	0.9;
	8402	1	-0	-0	0	0.54	0	0.9876	-24.925906	110	18	1.1	0.9;
	8403	1	0	-0	0	0.28	0	1.002808	-30.606777	110	18	1.1	0.9;
	8404	1	-0	0	0	12.64	0	1.059469	55.335907	380	3	1.1	0.9;
	8405	1	40.75	12.5	0	2.61	0	1.031329	-12.655848	220	5	1.1	0.9;
	8406	1	0	0	0	14.38	0	1.046532	-25.471306	220	5	1.1	0.9;
	8407	1	39.94	13.3	0	0.85	0	0.976226	-36.046729	110	21	1.1	0.9;
	8408	1	64.57	23.1	0	1.88	0	1.051004	-22.578573	110	19	1.1	0.9;
	8409	1	-0	-0	0	19.08	0	1.045438	-33.498156	150	9	1.1	0.9;
	8410	1	-0	0	0	0.6	0	1.038778	-23.326859	110	3	1.1	0.9;
	8411	1	39.64	9.6	0	1.8	0	1.038405	-26.143069	220	5	1.1	0.9;
	8412	1	33.84	11.1	0	1.57	0	1.022914	-35.028767	110	12	1.1	0.9;
	8413	2	0	0	0	0	0	1.031312	-38.24745	380	6	1.1	0.9;
	8414	1	-0.24	-25.61	0	11.32	0	1.116729	23.686346	220	3	1.3	0.7;
	8415	2	0	0	0	0	0	1.082292	53.782849	380	3	1.1	0.9;
	8416	1	0	-0	0	91.64	0	1.086385	-21.442648	380	20	1.3	0.7;
	8417	1	266	62.8	0	1.95	0	1.005143	-48.285615	220	4	1.1	0.9;
	8418	1	18.52	6.1	0	0.84	0	1.020354	-29.212066	110	21	1.1	0.9;
	8419	1	99.1	-98.6	0	6.83	0	1.010846	-60.746633	220	4	1.1	0.9;
	8420	1	-0	0	0	0	0	1.044478	-9.554053	150	2	1.1	0.9;
	8421	1	223.3	-7.6	0	19.97	0	1.013008	25.327957	380	3	1.1	0.9;
	8422	1	68.6	9.8	0	117.06	0	1.024934	4.963525	400	1	1.1	0.9;
	8423	1	19.72	6.5	0	0.2	0	1.018322	-28.704092	110	21	1.1	0.9;
	8424	1	-0	-0	0	1.15	0	0.969739	-22.513967	110	18	1.1	0.9;
	8425	1	11.21	4.4	0	0.25	0	1.044233	-26.058998	110	15	1.1	0.9;
	8426	1	176.2	-46.9	0	5.47	0	1.020758	-43.187108	220	6	1.1	0.9;
	8427	2	0	0	0	0	0	1.012295	21.310711	380	3	1.1	0.9;
	8428	1	0	0	0	1.67	0	1.035015	-13.126798	220	1	1.1	0.9;
	8429	1	0	0	0	14.69	0	1.031411	18.850336	380	3	1.1	0.9;
	8430	1	88.5	27.4	0	0.04	0	1.002869	-30.564646	110	21	1.1	0.9;
	8431	1	-0	-0	0	2.64	0	1.011512	34.728159	220	1	1.1	0.9;
	8432	1	-0.19	-17.71	0	28.98	0	1.0265	11.435168	380	8	1.1	0.9;
	8433	2	0	0	0	0	0	1.055181	-34.290765	380	11	1.1	0.9;
	8434	1	133	18.9	0	18.79	0	1.020874	-7.36671	154	1	1.1	0.9;
	8435	1	0	-0	0	0.1	0	0.997008	-32.24721	150	9	1.1	0.9;
	8436	1	8.81	5.5	0	2.83	0	1.005642	-31.843939	150	9	1.1	0.9;
	8437	1	35.04	12	0	1.07	0	0.972298	-33.087967	110	21	1.1	0.9;
	8438	1	125.5	22.5	0	9.17	0	1.01228	-47.827544	220	4	1.1	0.9;
	8439	1	380.8	62.8	0	7.11	0	0.994526	-23.773173	220	5	1.1	0.9;
	8440	1	139.3	19.9	0	0.56	0	1.017254	-9.964329	154	1	1.1	0.9;
	8441	1	-0	-0	0	0.01	0	1.042054	2.498815	154	1	1.1	0.9;
	8442	1	31.44	12.4	0	2.89	0	1.040171	2.622307	150	8	1.1	0.9;
	8443	1	332.8	92.2	0	29.46	0	0.996955	-50.261227	220	6	1.1	0.9;
	8444	1	0	0	0	0.66	0	0.96267	-29.940055	150	9	1.1	0.9;
	8445	1	-7.6	28.6	0	0.19	0	0.993246	-23.815094	150	9	1.1	0.9;
	8446	1	33.64	4.5	0	0.4	0	1.019171	-30.654209	110	19	1.1	0.9;
	8447	1	17.02	5.6	0	0.19	0	1.032791	-34.255223	110	12	1.1	0.9;
	8448	1	286.1	89.3	0	8.89	0	1.012896	-20.944744	220	5	1.1	0.9;
	8449	1	111.08	-44.35	0	7.94	0	1.044106	7.845761	220	13	1.1	0.9;
	8450	1	19.82	5.8	0	4.99	0	1.031128	-38.926529	110	22	1.1	0.9;
	8451	2	0	0	0	0	0	0.998015	18.852253	220	3	1.1	0.9;
	8452	1	-0	0	0	0.85	0	1.005589	-30.393461	150	9	1.1	0.9;
	8453	1	35.84	10.9	0	0.11	0	1.064895	-22.992552	110	15	1.1	0.9;
	8454	1	143.3	71.4	0	11.16	0	1.013448	-16.868276	380	3	1.1	0.9;
	8455	1	13.92	10.1	0	0.31	0	0.990052	-17.909625	110	18	1.1	0.9;
	8456	1	24.22	-1.06	0	2.92	0	1.011946	61.701308	150	1	1.1	0.9;
	8457	1	0	0	0	0	0	1.021959	-28.230262	150	9	1.1	0.9;
	8458	2	0	0	0	0	0	1.042019	-9.780413	380	5	1.1	0.9;
	8459	1	285.2	83.5	0	5.6	0	0.96597	-43.873196	220	10	1.1	0.9;
	8460	1	-0	-0	0	1.62	0	1.009914	3.722562	154	1	1.1	0.9;
	8461	1	-0	-0	0	2.02	0	0.980526	-30.31025	150	9	1.1	0.9;
	8462	1	-0.59	-0.86	0	0.14	0	1.013093	-23.797372	150	9	1.1	0.9;
	8463	1	24.13	10.8	0	15.2	0	1.036561	5.69769	150	8	1.1	0.9;
	8464	2	0	0	0	0	0	1.080213	-22.387307	110	15	1.1	0.9;
	8465	1	0	0	0	23.94	0	1.048179	65.975127	380	3	1.1	0.9;
	8466	1	0	-0	0	1.97	0	1.043279	-4.736557	220	5	1.1	0.9;
	8467	1	216.2	75.3	0	2.74	0	0.982596	-23.281146	220	5	1.1	0.9;
	8468	1	21.42	11.8	0	28.85	0	1.002528	-15.14463	380	5	1.1	0.9;
	8469	1	10.81	3.9	0	1.64	0	1.088634	-21.665499	110	15	1.1	0.9;
	8470	1	0	-0	0	2.78	0	1.015878	-27.78116	220	21	1.1	0.9;
	8471	1	-0	-0	0	0.73	0	1.053611	-26.038452	110	21	1.1	0.9;
	8472	1	19.4	2.8	0	3.63	0	1.017467	-23.010388	154	1	1.1	0.9;
	8473	2	0	0	0	0	0	1.038343	-8.939723	220	5	1.1	0.9;
	8474	2	0	0	0	0	0	1.093709	-21.648428	380	21	1.1	0.9;
	8475	1	-0	0	0	8.35	0	1.068803	-48.115522	220	5	1.1	0.9;
	8476	1	17.12	6.3	0	0.51	0	0.974696	-21.991786	110	18	1.1	0.9;
	8477	1	0	0	0	2.46	0	1.060032	-10.617375	220	5	1.1	0.9;
	8478	1	-0	0	0	3.01	0	1.043687	-8.725012	220	5	1.1	0.9;
	8479	1	161.28	54	0	2.12	0	0.995898	9.360812	220	3	1.1	0.9;
	8480	1	0	0	0	1.15	0	1.027373	14.332171	110	8	1.1	0.9;
	8481	2	0	0	0	0	0	1.035652	-28.583007	110	21	1.1	0.9;
	8482	1	0	-0	0	0.42	0	0.994876	-20.820169	110	18	1.1	0.9;
	8483	1	-0.17	-1.82	0	0	0	1.048138	18.207562	380	8	1.1	0.9;
	8484	1	0	0	0	0.13	0	1.015152	-35.999149	220	6	1.1	0.9;
	8485	1	149	99.8	0	9.27	0	1.037606	-7.581446	220	5	1.1	0.9;
	8486	2	0	0	0	0	0	1.043845	-19.840695	380	5	1.1	0.9;
	8487	1	44.45	7.68	0	17.83	0	0.994986	-4.06639	380	5	1.1	0.9;
	8488	1	66.67	-0	0	0.11	0	1.022527	-24.16707	220	7	1.1	0.9;
	8489	1	0	0	0	21.62	0	1.044468	-33.475701	150	9	1.1	0.9;
	8490	1	-0.19	-1	0	0	0	1.020965	-1.151268	150	8	1.1	0.9;
	8491	1	20.9	3	0	2.5	0	1.018354	-1.418547	154	1	1.1	0.9;
	8492	1	98.6	27.1	0	42.45	0	1.019872	-4.282529	150	8	1.1	0.9;
	8493	1	-0	-0	0	13.01	0	0.969171	-37.027093	220	10	1.1	0.9;
	8494	1	235.97	34.52	0	7.15	0	1.086265	-10.060933	380	5	1.1	0.9;
	8495	1	9.51	3.4	0	0.73	0	1.04515	-22.254829	110	15	1.1	0.9;
	8496	1	0	0	0.123474	4.29472	0	1.067205	-29.196427	220	1	1.1	0.9;
	8497	1	38.84	14.2	0	3.5	0	1.04174	-14.836957	220	5	1.1	0.9;
	8498	1	236.44	-98.55	0	1.31	0	1.044703	46.218731	220	3	1.1	0.9;
	8499	1	-14.89	-4.43	0	0.63	0	1.001762	-20.542659	110	18	1.1	0.9;
	8500	1	0	0	0	15.91	0	1.027198	-35.522122	380	1	1.1	0.9;
	8501	1	10.11	-0	0	1.92	0	0.943472	-31.974863	110	18	1.1	0.9;
	8502	1	-0	0	0	0.34	0	0.967291	-32.43496	150	9	1.1	0.9;
	8503	1	97.6	33.9	0	6.47	0	1.012117	-6.296154	220	3	1.1	0.9;
	8504	1	6.41	2.2	0	52.07	0	1.070173	-38.840203	380	4	1.1	0.9;
	8505	1	-0.59	-2.63	0	0.14	0	1.052274	-19.616218	110	19	1.1	0.9;
	8506	1	0	-0	0	0.67	0	0.982012	-27.319954	150	9	1.1	0.9;
	8507	1	37.44	9.1	0	7.87	0	1.068688	-26.744893	220	5	1.1	0.9;
	8508	1	-8.02	-15.14	0	0	0	1.030505	16.850212	150	8	1.1	0.9;
	8509	1	-0	0	0	3.76	0	1.034288	-30.033529	220	6	1.1	0.9;
	8510	1	0	0	0	8.78	0	1.063564	14.839237	220	8	1.1	0.9;
	8511	1	93.6	-43.5	0	33.51	0	1.038633	-24.833489	220	5	1.1	0.9;
	8512	1	140.7	10.4	0	2.51	0	1.007232	-47.769593	220	4	1.1	0.9;
	8513	1	13.51	5.6	0	0.86	0	1.027518	-27.107897	110	15	1.1	0.9;
	8514	1	30.33	11	0	0.56	0	1.067535	-19.55661	110	19	1.1	0.9;
	8515	2	0	0	0	0	0	1.059651	-12.797067	220	5	1.1	0.9;
	8516	1	104.3	12.3	0	8.29	0	1.018486	7.90663	220	3	1.1	0.9;
	8517	1	268.2	-257.9	0	42.56	0	1.037198	8.190569	380	3	1.1	0.9;
	8518	2	0	0	0	0	0	1.031547	-7.26949	154	1	1.1	0.9;
	8519	1	13.71	8.5	0	0.06	0	1.058712	-28.62346	150	9	1.1	0.9;
	8520	1	-0	-0	0	4.5	0	1.027418	-10.670989	220	3	1.1	0.9;
	8521	1	53.86	33.4	0	12.95	0	1.022682	-33.542758	220	6	1.1	0.9;
	8522	2	0	0	0	0	0	1.027564	-15.472563	220	5	1.1	0.9;
	8523	1	20.9	3	0	7.77	0	1.029809	-4.439668	154	1	1.1	0.9;
	8524	1	-0	0	0	11.14	0	1.093985	-47.983991	220	6	1.1	0.9;
	8525	1	-0	-3.7	0	0.25	0	1.032332	26.341746	110	8	1.1	0.9;
	8526	1	0	0	0	45.57454	0	1.049684	-21.785872	380	20	1.1	0.9;
	8527	1	28.13	10.2	0	1.73	0	0.985975	-22.366185	110	19	1.1	0.9;
	8528	1	0	0	0	0.84	0	1.022215	51.357588	220	3	1.1	0.9;
	8529	1	76.99	-0.7	0	1.06	0	1.00959	-47.214593	220	4	1.1	0.9;
	8530	1	-0	0	0	1.31	0	1.010375	-46.855704	220	4	1.1	0.9;
	8531	2	0	0	0	0	0	1.019991	-20.209592	220	20	1.1	0.9;
	8532	2	0	0	0	0	0	1.032486	-13.566695	380	6	1.1	0.9;
	8533	1	12.31	4.1	0	0.32	0	0.990619	-21.190201	110	18	1.1	0.9;
	8534	1	295.2	39.3	0	7.3	0	1.013794	-44.443797	220	4	1.1	0.9;
	8535	1	72.68	24.3	0	3.12	0	1.031595	-22.272387	220	5	1.1	0.9;
	8536	1	36.44	22.6	0	2.32	0	0.974217	-31.060355	150	9	1.1	0.9;
	8537	1	187.64	31.94	0	4.75	0	1.035589	7.167784	220	13	1.1	0.9;
	8538	1	0	0	0	0	0	1.013807	-32.593868	150	9	1.1	0.9;
	8539	1	0	-0	0	15.15	0	1.056399	-34.762887	380	1	1.1	0.9;
	8540	1	0	0	0	0	0	1.006246	-31.812632	150	9	1.1	0.9;
	8541	1	0	-0	0	1.23	0	1.040496	9.279655	220	3	1.1	0.9;
	8542	1	87.7	26	0	47.58	0	1.041837	-21.334133	220	5	1.1	0.9;
	8543	1	10	1.4	0	5.74	0	1.073534	-1.689652	154	1	1.1	0.9;
	8544	1	-0	-0	0	2.74	0	0.982538	6.790327	220	3	1.1	0.9;
	8545	1	9.91	3.3	0	6.72	0	1.028466	-40.123318	110	22	1.1	0.9;
	8546	1	5.01	1.8	0	0.77	0	1.035972	-11.828136	220	5	1.1	0.9;
	8547	1	21.22	9.9	0	4.64	0	1.008267	-52.388483	220	4	1.1	0.9;
	8548	1	0	0	0	0	0	1.014538	-27.099328	150	9	1.1	0.9;
	8549	1	-0	-0	0	0.63	0	1.015773	-24.245852	110	19	1.1	0.9;
	8550	1	0	0	0	1.42	0	0.976417	-30.337941	150	9	1.1	0.9;
	8551	1	0	0	0.176855	3.419803	0	0.994046	-21.126665	220	19	1.1	0.9;
	8552	1	0	0	0	5.51	0	1.016198	8.888573	220	3	1.1	0.9;
	8553	2	0	0	0	0	0	1.077556	60.968069	380	3	1.1	0.9;
	8554	1	0	-0	0	0.23	0	1.092619	-27.557647	150	9	1.1	0.9;
	8555	1	22.02	7.7	0	0.17	0	1.032085	-41.190379	110	22	1.1	0.9;
	8556	1	128.1	42.8	0	25.57	0	1.040779	-43.284256	380	17	1.1	0.9;
	8557	1	0	-0	0	0.92	0	1.012552	-39.089667	110	12	1.1	0.9;
	8558	1	0	0	0	0	0	0.96568	-37.714763	150	9	1.1	0.9;
	8559	1	-0	-7.5	0	0	0	1.013215	-30.419221	150	9	1.1	0.9;
	8560	1	-0	0	0	1.11	0	0.980131	-18.492926	380	5	1.1	0.9;
	8561	1	-0	-0	0	237.54	0	1.060954	-43.422544	380	4	1.1	0.9;
	8562	1	0	0	0.379262	96.912184	0	1.096728	-23.569307	380	21	1.3	0.7;
	8563	1	172.6	24.6	0	11.39	0	1.010131	-25.941552	154	1	1.1	0.9;
	8564	2	0	0	0	0	0	1.01667	-19.679406	220	5	1.1	0.9;
	8565	1	-0.19	-0.92	0	0	0	1.048835	7.304522	150	8	1.1	0.9;
	8566	1	38.04	15	0	0.47	0	1.016108	-27.483837	110	21	1.1	0.9;
	8567	1	15.42	4.4	0	10.84	0	1.030261	-38.98912	110	22	1.1	0.9;
	8568	1	175.7	56.2	0	0.13	0	1.023273	-6.370838	220	5	1.1	0.9;
	8569	2	0	0	0	0	0	1.038006	-36.82583	380	6	1.1	0.9;
	8570	1	0	-0	0	1.71	0	0.991237	-31.624371	110	21	1.1	0.9;
	8571	1	126.8	21.4	0	11.43	0	1.024502	-54.221095	220	4	1.1	0.9;
	8572	1	0.39	-0.24	0	0.15	0	0.997313	-26.959645	110	18	1.1	0.9;
	8573	1	0	-0	0	22.04	0	1.017811	-46.734623	380	6	1.1	0.9;
	8574	1	-0	0	0.447839	20.994738	0	0.993535	-4.162453	380	2	1.1	0.9;
	8575	1	89.6	13.1	0	8.14	0	1.023769	-54.455841	220	4	1.1	0.9;
	8576	2	0	0	0	0	0	1.017525	-5.150807	154	1	1.1	0.9;
	8577	1	10.91	3.6	0	0.55	0	1.050327	-31.670632	110	12	1.1	0.9;
	8578	1	0	-0	0	4.55	0	1.036215	-22.212869	220	5	1.1	0.9;
	8579	1	9.91	3.3	0	0.82	0	1.016275	-35.57685	110	22	1.1	0.9;
	8580	1	-0	-0	0	72.611776	0	1.076159	-25.123339	380	20	1.1	0.9;
	8581	1	0	-0	0	1.79	0	1.019212	4.571542	380	2	1.1	0.9;
	8582	1	27.83	-2.6	0	6.07	0	1.02128	-52.018951	220	4	1.1	0.9;
	8583	1	-0	-0	0	6.29	0	1.065295	-12.788581	220	18	1.1	0.9;
	8584	1	9.7	1.4	0	2.96	0	1.022184	-8.982181	154	1	1.1	0.9;
	8585	1	0	-0	0	8.89	0	1.028916	13.84417	380	3	1.1	0.9;
	8586	2	0	0	0	0	0	1.002287	-26.032897	110	18	1.1	0.9;
	8587	1	27.53	17	0	0.03	0	1.025557	-24.888109	150	9	1.1	0.9;
	8588	1	317.9	55.4	0	37.61	0	1.020087	-48.936381	380	6	1.1	0.9;
	8589	1	-0	0	0	0.51	0	0.967627	-32.440968	150	9	1.1	0.9;
	8590	1	-33.74	11.1	0	1.87	0	1.038424	-24.662973	150	9	1.1	0.9;
	8591	2	0	0	0	0	0	1.06486	-21.126008	150	9	1.1	0.9;
	8592	1	-0	0	0	0.43	0	1.050094	-3.995844	220	5	1.1	0.9;
	8593	1	171.3	13.9	0	59.09	0	1.095968	45.426526	380	3	1.1	0.9;
	8594	1	3.8	3.8	0	0.08	0	0.954489	-30.331753	110	18	1.1	0.9;
	8595	1	72.4	10.3	0	6.82	0	1.006996	3.372222	154	1	1.1	0.9;
	8596	1	0	-0	0	0.19	0	0.893342	-35.877764	150	9	1.3	0.7;
	8597	1	65.2	9.3	0	3.89	0	1.026523	-16.879696	154	1	1.1	0.9;
	8598	1	-0	-0	0	6.07	0	1.031185	-22.053606	220	7	1.1	0.9;
	8599	1	26.93	9.6	0	1.09	0	1.04312	-21.80475	110	19	1.1	0.9;
	8600	1	-0	-0	0	4.82	0	1.028118	-35.099662	220	6	1.1	0.9;
	8601	1	0	0	0	1.96	0	1.036534	-51.990564	380	10	1.1	0.9;
	8602	1	-0	0	0	2.05	0	0.982449	9.7613	220	3	1.1	0.9;
	8603	2	0	0	0	0	0	1.022771	-21.785917	150	9	1.1	0.9;
	8604	1	0	-0	0	20.81	0	1.020243	15.224814	380	8	1.1	0.9;
	8605	2	0	0	0	0	0	1.005953	-0.504817	380	2	1.1	0.9;
	8606	1	0	0	0	0	0	0.991297	-30.713122	150	9	1.1	0.9;
	8607	1	0	0	0	0	0	0.9901	-34.700265	150	9	1.1	0.9;
	8608	2	0	0	0	0	0	1.033413	-35.620556	380	6	1.1	0.9;
	8609	1	382.4	86.3	0	22.29	0	1.024958	-22.744767	220	11	1.1	0.9;
	8610	2	0	0	0	0	0	1.056933	6.177576	220	13	1.1	0.9;
	8611	2	0	0	0	0	0	1.024775	63.476826	150	1	1.1	0.9;
	8612	1	-0	0	0	12.69	0	1.019124	-34.958062	220	12	1.1	0.9;
	8613	1	101.8	20.2	0	0.91	0	1.023127	1.903711	150	8	1.1	0.9;
	8614	1	0	0	0	36.99	0	1.035176	44.215941	380	3	1.1	0.9;
	8615	1	-0	-0	0	1.35	0	1.040415	-23.134266	150	9	1.1	0.9;
	8616	1	-0	0	0	0.17	0	0.964784	-31.482874	150	9	1.1	0.9;
	8617	1	0	-0	0	0.45	0	0.983977	-31.897049	110	21	1.1	0.9;
	8618	1	-0	-0	0	5.72	0	1.017905	-27.123566	220	21	1.1	0.9;
	8619	1	-0	0	0	13.3	0	1.026622	-16.900837	380	1	1.1	0.9;
	8620	1	76.68	23.3	0	0.3	0	1.025816	-10.251952	220	3	1.1	0.9;
	8621	1	-70.45	-35.14	0	0	0	1.017477	17.227709	150	8	1.1	0.9;
	8622	2	0	0	0	0	0	1.028478	-36.114605	110	12	1.1	0.9;
	8623	2	0	0	0	0	0	1.034977	-24.552494	380	6	1.1	0.9;
	8624	1	25.03	18.8	0	1.46	0	0.982939	9.190133	220	3	1.1	0.9;
	8625	2	0	0	0	0	0	1.051079	-41.671715	380	4	1.1	0.9;
	8626	1	136	36.1	0	6.07	0	0.987786	-17.922777	220	5	1.1	0.9;
	8627	1	-0	0	0	4.26	0	1.044027	-26.736186	220	5	1.1	0.9;
	8628	1	277.1	37	0	21.03	0	1.036975	-23.955541	220	5	1.1	0.9;
	8629	1	98.61	50.3	0	0.17	0	0.967122	-43.610509	150	10	1.1	0.9;
	8630	2	0	0	0	0	0	1.059594	-47.562957	380	10	1.1	0.9;
	8631	1	-0	-0	0	0.54	0	0.918045	-35.76819	150	9	1.1	0.9;
	8632	1	-0	-0	0	2	0	1.050716	-24.767052	150	9	1.1	0.9;
	8633	1	0	-0	0.090774	8.200444	0	1.039381	-16.713316	220	18	1.1	0.9;
	8634	1	74.18	27.3	0	25.25	0	1.017282	-29.113408	220	6	1.1	0.9;
	8635	1	282	21.5	0	21.3	0	1.013338	-47.542885	380	6	1.1	0.9;
	8636	1	31.43	10.6	0	1.61	0	1.030477	-22.079517	220	5	1.1	0.9;
	8637	1	-20.32	6.7	0	0.2	0	0.963865	-31.392893	150	9	1.1	0.9;
	8638	1	62.27	23.8	0	0.13	0	1.027699	9.746278	220	3	1.1	0.9;
	8639	1	130.2	88.8	0	0.08	0	1.019721	-5.304798	220	3	1.1	0.9;
	8640	2	0	0	0	0	0	1.174902	-19.739552	380	9	1.3	0.7;
	8641	1	-0	-0	0	15.67	0	1.11078	-29.279417	380	1	1.3	0.7;
	8642	1	163.2	31.9	0	4.9	0	1.031062	8.31398	150	8	1.1	0.9;
	8643	1	29.13	9.5	0	0.98	0	0.992929	-28.53906	110	18	1.1	0.9;
	8644	1	-0	0	0	32.28	0	1.165566	-22.013417	380	1	1.3	0.7;
	8645	1	25.18	3.37	0	1.01	0	0.977499	-24.296922	110	18	1.1	0.9;
	8646	1	-2.3	-0.8	0	0.33	0	1.062285	-25.159125	110	19	1.1	0.9;
	8647	1	0	0	0	4.63	0	1.020365	1.923148	380	3	1.1	0.9;
	8648	1	20.9	3	0	2.09	0	1.009446	-22.883661	154	1	1.1	0.9;
	8649	1	0	0	0	0	0	1.084595	-28.010738	150	9	1.1	0.9;
	8650	1	4.4	2.2	0	1.09	0	1.029588	-40.945637	110	22	1.1	0.9;
	8651	1	162.49	45.11	0	5.38	0	1.031215	-17.06483	220	5	1.1	0.9;
	8652	1	0	0	0	21.22	0	1.034612	4.025941	220	13	1.1	0.9;
	8653	1	44.55	7.5	0	0.6	0	1.034538	-16.169774	220	5	1.1	0.9;
	8654	1	125.2	17.8	0	2.5	0	1.009473	-10.803889	154	1	1.1	0.9;
	8655	1	-16.42	5.4	0	0.67	0	1.035878	-30.794051	150	9	1.1	0.9;
	8656	1	5.51	1.3	0	0.09	0	1.026513	-9.244782	220	5	1.1	0.9;
	8657	1	0	-0	0	61.44	0	1.011958	-7.088597	400	1	1.1	0.9;
	8658	1	38.54	12	0	16.52	0	1.02585	-54.455522	220	4	1.1	0.9;
	8659	1	0	-0	0	0.41	0	1.041316	-20.872613	110	19	1.1	0.9;
	8660	1	129.8	67.2	0	2.57	0	0.999992	-36.806485	150	9	1.1	0.9;
	8661	1	226.7	39.3	0	0.99	0	1.023652	-48.576903	220	4	1.1	0.9;
	8662	1	-0.12	-1.74	0	0	0	1.079881	31.582222	380	8	1.1	0.9;
	8663	2	0	0	0	0	0	1.025281	-12.708169	380	6	1.1	0.9;
	8664	1	34.54	15.4	0	29.82	0	1.056778	-46.823851	380	4	1.1	0.9;
	8665	1	45.45	14.9	0	1.12	0	0.976784	-30.555308	110	21	1.1	0.9;
	8666	1	1.3	0.5	0	0.39	0	0.974899	-21.620786	110	18	1.1	0.9;
	8667	1	9.91	3.3	0	0.98	0	1.00122	-39.501095	110	12	1.1	0.9;
	8668	1	69.88	19.3	0	1.63	0	1.017386	13.248317	110	8	1.1	0.9;
	8669	1	74.98	10.5	0	12.96	0	1.042341	-15.672433	220	5	1.1	0.9;
	8670	2	0	0	0	0	0	1.042199	-6.729364	220	5	1.1	0.9;
	8671	1	62.77	5.7	0	2.41	0	1.013382	-51.253863	220	4	1.1	0.9;
	8672	1	-0	-0	0	28.515005	0	1.022513	-21.387883	380	5	1.1	0.9;
	8673	1	-0	-0	0	13	0	0.989519	-35.131931	150	9	1.1	0.9;
	8674	1	-0	0	0	1.31	0	1.046981	-24.95506	220	1	1.1	0.9;
	8675	2	0	0	0	0	0	1.037158	-23.941358	220	7	1.1	0.9;
	8676	2	0	0	0	0	0	1.019558	-21.879637	220	5	1.1	0.9;
	8677	1	0	0	0	2.29	0	1.016261	-10.284826	220	5	1.1	0.9;
	8678	1	0	0	0	0	0	0.989904	-28.489073	150	9	1.1	0.9;
	8679	1	-0	-0	0	29.01	0	1.009551	6.826901	400	1	1.1	0.9;
	8680	1	0	-0	0	90.93	0	1.001193	-8.273624	400	1	1.1	0.9;
	8681	1	0	0	0	4.26	0	0.972909	-53.475502	220	6	1.1	0.9;
	8682	1	34.9	5	0	7.79	0	1.016275	-9.010491	154	1	1.1	0.9;
	8683	2	0	0	0	0	0	1.039554	-24.600261	220	5	1.1	0.9;
	8684	1	0	-0	0	0.5	0	0.968578	-37.38368	150	9	1.1	0.9;
	8685	1	168.13	36.31	0	88.12	0	1.019225	-4.135965	380	13	1.1	0.9;
	8686	1	-0	-0	0	0.65	0	1.023602	-32.455421	150	9	1.1	0.9;
	8687	1	-0.15	-152.82	0.390048	24.599897	0	1.016095	18.588008	380	3	1.1	0.9;
	8688	1	0	0	0	0	0	1.092516	-27.545772	150	9	1.1	0.9;
	8689	1	42.65	18.2	0	0	0	1.03309	-21.003238	220	5	1.1	0.9;
	8690	1	-0	-0	0	2.81	0	1.045485	-16.317647	380	1	1.1	0.9;
	8691	1	305.04	77.16	0	17.43	0	1.041127	-33.775172	220	5	1.1	0.9;
	8692	1	0	0	0.090774	21.420444	0	1.059128	-17.181332	220	18	1.1	0.9;
	8693	2	0	0	0	0	0	1.073334	60.748503	380	3	1.1	0.9;
	8694	1	-0	0	0	20.75	0	1.014733	-33.035812	150	9	1.1	0.9;
	8695	1	5.5	0.8	0	0.9	0	1.026323	-2.650757	154	1	1.1	0.9;
	8696	1	-0	-0	0	6.52	0	1.026382	-12.469966	220	3	1.1	0.9;
	8697	1	-0	-0	0	0.55	0	1.015734	-26.417686	150	1	1.1	0.9;
	8698	1	20.02	15	0	0.14	0	0.982379	9.295612	220	3	1.1	0.9;
	8699	1	89.3	10	0	9.62	0	1.019275	9.277342	220	3	1.1	0.9;
	8700	1	0	0	0	0.3	0	1.013945	-32.760027	150	9	1.1	0.9;
	8701	1	82.59	23.4	0	6.65	0	1.021917	-44.217953	220	4	1.1	0.9;
	8702	2	0	0	0	0	0	1.026759	-41.204592	220	4	1.1	0.9;
	8703	1	0	-0	0	3.1	0	1.010883	-32.679218	150	9	1.1	0.9;
	8704	1	180.7	71.8	0	2.46	0	1.038186	-8.064046	220	5	1.1	0.9;
	8705	1	-0	0	0	1.01	0	0.995757	-27.153575	110	18	1.1	0.9;
	8706	1	40.34	15.9	0	6.02	0	1.017244	4.061846	150	8	1.1	0.9;
	8707	1	101.9	24.3	0	8.53	0	1.028321	-11.903148	220	5	1.1	0.9;
	8708	1	-0	-0	0	11.22	0	1.025238	-34.309952	380	6	1.1	0.9;
	8709	1	3.7	1.3	0	0.48	0	1.033387	-23.424365	110	19	1.1	0.9;
	8710	1	124.5	-6	0	1.91	0	1.02164	-48.196695	220	4	1.1	0.9;
	8711	1	0	0	0	3.9	0	1.035108	-12.381387	220	5	1.1	0.9;
	8712	1	3.04	-13.76	0	4.17	0	0.975055	-32.24574	150	9	1.1	0.9;
	8713	1	0	-0	0	1.13	0	1.02082	16.809698	380	3	1.1	0.9;
	8714	2	0	0	0	0	0	1.015463	-7.074502	400	1	1.1	0.9;
	8715	1	0	0	0	90.194994	0	1.069467	-22.161838	380	20	1.1	0.9;
	8716	1	0	0	0	14.27	0	1.069514	26.080416	380	3	1.1	0.9;
	8717	1	23.53	8.5	0	1.28	0	1.083125	-17.411194	110	19	1.1	0.9;
	8718	1	15.32	5	0	0.49	0	1.026636	-34.506198	110	12	1.1	0.9;
	8719	1	91.8	8.3	0	2.44	0	1.010794	-45.334737	220	4	1.1	0.9;
	8720	1	3.5	2.7	0	0.39	0	0.968716	-28.617584	110	18	1.1	0.9;
	8721	2	0	0	0	0	0	1.055054	-16.297809	380	5	1.1	0.9;
	8722	1	-0	0	0	2.13	0	1.066576	-25.430169	220	5	1.1	0.9;
	8723	1	-0	0	0	3.41	0	1.026073	-14.019533	220	6	1.1	0.9;
	8724	1	127.9	3.7	0	41.47	0	1.041758	-18.876378	380	6	1.1	0.9;
	8725	1	126.3	109.2	0	54.97	0	1.036813	8.212135	380	3	1.1	0.9;
	8726	1	100.4	1.4	0	5.71	0	1.003185	-23.363326	220	20	1.1	0.9;
	8727	1	0	0	0	0	0	0.966862	-34.40203	150	9	1.1	0.9;
	8728	1	443.8	172.1	0	52.54	0	1.069096	3.169893	380	14	1.1	0.9;
	8729	2	0	0	0	0	0	1.008386	-38.518465	220	4	1.1	0.9;
	8730	1	0	-0	0	3.79	0	1.046974	-27.18651	150	9	1.1	0.9;
	8731	1	191.2	16.4	0	17.99	0	1.021526	-40.277052	380	4	1.1	0.9;
	8732	1	66.27	19.3	0	12.94	0	1.034889	-21.457823	220	5	1.1	0.9;
	8733	1	25	3.6	0	3.77	0	1.037546	9.238636	154	1	1.1	0.9;
	8734	1	-0	-0	0.12	8.51	0	1.008034	-28.929654	220	21	1.1	0.9;
	8735	1	13.41	3.4	0	0.72	0	0.986546	-29.724015	110	18	1.1	0.9;
	8736	1	0	0	0	0	0	0.989655	-30.821299	150	9	1.1	0.9;
	8737	2	0	0	0	0	0	1.058573	-28.372231	150	9	1.1	0.9;
	8738	1	-0	0	0	3.78	0	1.025108	5.448735	150	8	1.1	0.9;
	8739	1	105.8	53.6	0	6.27	0	0.959797	-37.411141	150	9	1.1	0.9;
	8740	1	2.5	0.6	0	0.61	0	0.985151	-28.761493	110	18	1.1	0.9;
	8741	1	-0	-0	0	1.72	0	1.033837	-39.750284	380	6	1.1	0.9;
	8742	1	-0	0	0	31.05	0	1.031843	-9.377241	380	7	1.1	0.9;
	8743	1	11.61	7.6	0	6.21	0	1.025606	-34.269151	220	5	1.1	0.9;
	8744	1	0	0	0	25.67	0	0.972972	-37.275725	150	9	1.1	0.9;
	8745	1	-0	0	0	0.15	0	0.989405	-30.788811	150	9	1.1	0.9;
	8746	1	-0	0	0	0.47	0	1.048871	-20.066491	110	19	1.1	0.9;
	8747	1	0	0	0	0.12	0	1.047501	-24.793142	150	9	1.1	0.9;
	8748	1	279.7	4.6	0	3.77	0	1.063635	-49.908328	220	5	1.1	0.9;
	8749	1	5.11	2.5	0	0.95	0	0.980914	-30.105334	110	18	1.1	0.9;
	8750	1	0	0	0	0	0	0.994461	-30.251812	150	9	1.1	0.9;
	8751	1	0	-0	0	4.15	0	1.043288	-17.358459	220	1	1.1	0.9;
	8752	1	61.98	39.2	0	5.1	0	0.990312	-30.214954	150	9	1.1	0.9;
	8753	1	-23.09	-46.44	0	0	0	1.029088	-11.334255	220	3	1.1	0.9;
	8754	1	30.03	15	0	2.49	0	1.007818	-26.185595	110	18	1.1	0.9;
	8755	1	19.02	3.8	0	4.55	0	0.928442	-30.498908	110	18	1.1	0.9;
	8756	1	54.66	18.6	0	3.04	0	1.009712	-60.476539	220	4	1.1	0.9;
	8757	1	43.05	25	0	0.02	0	0.989012	-30.812297	150	9	1.1	0.9;
	8758	2	0	0	0	0	0	1.047294	10.896261	220	13	1.1	0.9;
	8759	1	-0	0	0.206089	25.776518	0	1.026786	14.2136	380	3	1.1	0.9;
	8760	1	92.1	38	0	4.93	0	0.948567	-21.02637	220	16	1.1	0.9;
	8761	1	18.36	-0.95	0	3.83	0	1.017931	62.866743	150	1	1.1	0.9;
	8762	2	0	0	0	0	0	1.046022	46.235938	380	3	1.1	0.9;
	8763	1	0	-0	0	215.65	0	1.02928	2.451775	380	5	1.1	0.9;
	8764	2	0	0	0	0	0	1.048209	-3.324162	220	13	1.1	0.9;
	8765	1	-0	-0	0	11.69	0	1.042555	-15.65821	220	5	1.1	0.9;
	8766	1	9.7	1.4	0	4.91	0	1.021881	-6.603587	154	1	1.1	0.9;
	8767	1	17.52	6.5	0	0.74	0	1.03213	-26.669148	110	15	1.1	0.9;
	8768	1	25.73	9.3	0	1.49	0	1.035222	-24.061987	110	19	1.1	0.9;
	8769	1	0	0	0	0	0	1.04929	-22.14211	150	9	1.1	0.9;
	8770	2	0	0	0	0	0	1.012975	-24.464313	154	1	1.1	0.9;
	8771	2	0	0	0	0	0	1.006983	-37.979458	150	10	1.1	0.9;
	8772	1	0	0	0	0	0	1.022559	8.17473	150	8	1.1	0.9;
	8773	1	0	-0	0	30.31	0	1.008807	10.994888	380	8	1.1	0.9;
	8774	1	0	-0	0	15.04	0	1.026003	-33.284232	220	6	1.1	0.9;
	8775	1	45.25	19.1	0	1.89	0	1.006926	5.67316	220	3	1.1	0.9;
	8776	1	-0	-0	0	6.28	0	0.999607	-38.362871	220	4	1.1	0.9;
	8777	1	7.61	6.3	0	0.1	0	0.984627	-28.428042	110	18	1.1	0.9;
	8778	1	35.34	37.8	0	9.28	0	1.008213	-2.607965	220	3	1.1	0.9;
	8779	1	292.4	-228.7	0	106.08	0	1.05095	15.405182	380	3	1.1	0.9;
	8780	1	1.8	0.8	0	0.98	0	0.994674	-20.33854	110	18	1.1	0.9;
	8781	1	8.81	2.2	0	4.78	0	1.013383	-41.325087	110	22	1.1	0.9;
	8782	1	26.23	9.5	0	0.65	0	1.025664	-28.456413	110	19	1.1	0.9;
	8783	1	11.51	3.8	0	0.53	0	0.97866	-31.872838	110	21	1.1	0.9;
	8784	1	100.5	19.3	0	13.51	0	1.021916	10.857723	150	8	1.1	0.9;
	8785	1	0	0	0	15.18	0	1.03393	-3.107838	220	1	1.1	0.9;
	8786	1	128.8	54.5	0	24.74	0	0.982627	-24.865359	220	6	1.1	0.9;
	8787	1	22.83	8	0	3.71	0	1.061905	-17.274053	220	5	1.1	0.9;
	8788	1	143.8	46.45	0	3.26	0	1.046837	-6.159102	220	5	1.1	0.9;
	8789	1	-29.73	23.6	0	11.75	0	0.987443	-46.759012	150	10	1.1	0.9;
	8790	1	0	0	0	0.93	0	1.025925	4.550938	154	1	1.1	0.9;
	8791	1	-0	-0	0	0.36	0	1.054222	-22.685203	220	5	1.1	0.9;
	8792	1	107.9	35.9	0	5.48	0	1.026906	3.499429	150	8	1.1	0.9;
	8793	1	179.6	25.6	0	149.28	0	1.012192	-21.626493	400	1	1.1	0.9;
	8794	1	3.8	-12.7	0	1.42	0	1.018534	-16.087801	110	18	1.1	0.9;
	8795	2	0	0	0	0	0	1.052731	-0.198514	220	5	1.1	0.9;
	8796	1	0	-0	0	6.59	0	0.982515	-34.879291	150	9	1.1	0.9;
	8797	2	0	0	0	0	0	1.072681	-17.583553	110	19	1.1	0.9;
	8798	1	80.59	5.2	0	7.77	0	1.013086	-6.857034	220	3	1.1	0.9;
	8799	1	24.03	5.6	0	0.23	0	1.044314	-27.26816	110	3	1.1	0.9;
	8800	1	9.91	3.3	0	0.66	0	0.998919	-43.895136	110	22	1.1	0.9;
	8801	1	31.03	8.9	0	0.14	0	1.056476	-17.127916	110	18	1.1	0.9;
	8802	1	0	0	0	23.68	0	1.047815	65.89492	380	3	1.1	0.9;
	8803	1	31.21	-13.87	0	0.32	0	1.026988	-28.497266	110	19	1.1	0.9;
	8804	1	22.02	4.4	0	4.83	0	1.044229	-9.735086	220	5	1.1	0.9;
	8805	1	35.24	11.4	0	1.84	0	1.008439	-35.641062	110	12	1.1	0.9;
	8806	1	-0	0	0	342.15	0	1.065258	-44.117934	380	4	1.1	0.9;
	8807	2	0	0	0	0	0	1.049311	-4.132509	220	5	1.1	0.9;
	8808	1	203.6	19.81	0	15.46	0	1.053244	-26.177439	220	5	1.1	0.9;
	8809	1	101.2	-2.9	0	0.51	0	1.021169	-32.757915	220	5	1.1	0.9;
	8810	1	0.3	1.5	0	0.1	0	1.057257	-2.779316	220	3	1.1	0.9;
	8811	1	47.35	15.5	0	1.41	0	1.034197	-41.067687	110	22	1.1	0.9;
	8812	1	32.24	10.1	0	0.81	0	0.980979	-25.408064	110	18	1.1	0.9;
	8813	1	-0	0	0	1.55	0	0.949387	-33.784726	150	9	1.1	0.9;
	8814	1	89.9	29.3	0	6.77	0	1.031268	3.404893	150	8	1.1	0.9;
	8815	1	12.02	3.6	0	0.47	0	0.966775	-29.185099	110	18	1.1	0.9;
	8816	1	82.19	15.6	0	0.43	0	1.026497	7.754492	220	3	1.1	0.9;
	8817	1	15.7	2.2	0	2.88	0	1.031901	-3.671816	154	1	1.1	0.9;
	8818	2	0	0	0	0	0	1.049908	0.513122	220	5	1.1	0.9;
	8819	1	37.87	11.59	0	1.88	0	1.030028	-53.584288	220	4	1.1	0.9;
	8820	2	0	0	0	0	0	1.063983	-22.01868	220	18	1.1	0.9;
	8821	1	8.31	3	0	0.52	0	1.030073	-28.186147	110	19	1.1	0.9;
	8822	2	0	0	0	0	0	1.064182	2.842782	380	13	1.1	0.9;
	8823	1	-0	0	0	0.39	0	0.973835	-32.030631	150	9	1.1	0.9;
	8824	1	21.9	3.1	0	4.94	0	1.018238	-7.533838	154	1	1.1	0.9;
	8825	1	0	-0	0	44.95	0	1.003275	-3.014246	380	5	1.1	0.9;
	8826	1	75.3	7.1	0	0.59	0	1.047082	-27.54358	110	23	1.1	0.9;
	8827	1	-0	0	0	1.77	0	0.98658	-31.221806	150	9	1.1	0.9;
	8828	1	0	0	0	1.05	0	1.016447	5.856189	150	8	1.1	0.9;
	8829	1	91.6	27.1	0	10.95	0	1.033798	-21.566094	220	5	1.1	0.9;
	8830	1	21.32	7.2	0	1.2	0	1.066758	-24.688851	110	19	1.1	0.9;
	8831	1	26.33	10.3	0	1.11	0	0.924772	-33.215113	110	18	1.1	0.9;
	8832	1	0	-0	0	2.24	0	1.020209	-1.214855	150	8	1.1	0.9;
	8833	1	0	0	0	12.42	0	0.999553	-36.912321	150	9	1.1	0.9;
	8834	1	-0	0	0	1.52	0	0.998306	-22.906243	220	5	1.1	0.9;
	8835	1	113.7	22.3	0	15.55	0	1.032665	-23.255961	220	5	1.1	0.9;
	8836	1	54.56	16.7	0	0.91	0	0.989016	-25.195238	110	18	1.1	0.9;
	8837	1	13.31	4.4	0	0.33	0	1.034416	-36.423723	110	12	1.1	0.9;
	8838	1	20.02	12.4	0	0.03	0	0.968504	-37.395062	150	9	1.1	0.9;
	8839	1	86.7	41.5	0	18.57	0	1.005281	-34.431326	150	9	1.1	0.9;
	8840	1	1.5	0.3	0	1.54	0	1.01206	-29.373767	110	18	1.1	0.9;
	8841	1	8.31	2.7	0	0.69	0	1.005244	-36.825291	110	12	1.1	0.9;
	8842	1	0	0	0	0	0	0.947855	-29.361997	150	9	1.1	0.9;
	8843	1	208.7	63	0	3.45	0	1.0448	-11.469831	220	5	1.1	0.9;
	8844	1	1.3	0	0	0.06	0	0.954467	-30.324095	110	18	1.1	0.9;
	8845	1	13	8.8	0	78.85984	0	1.091749	-24.992145	380	9	1.1	0.9;
	8846	1	20.82	0	0	0.17	0	1.06383	7.797356	220	5	1.1	0.9;
	8847	1	-0	0	0.18076	-0.881251	0	0.990993	-0.924277	380	2	1.1	0.9;
	8848	1	0	0	0	0	0	0.905716	-35.509144	150	9	1.3	0.7;
	8849	1	4.4	2.2	0	0.28	0	1.03897	-40.00065	110	22	1.1	0.9;
	8850	1	7.41	2.4	0	2.23	0	1.040505	-32.488756	110	12	1.1	0.9;
	8851	2	0	0	0	0	0	1.063062	63.161074	380	3	1.1	0.9;
	8852	1	6.01	2.2	0	0.82	0	0.999293	-31.516305	110	19	1.1	0.9;
	8853	1	410.3	91.6	0	13.25	0	1.047513	-36.823466	220	5	1.1	0.9;
	8854	1	69.18	35.1	0	12.59	0	1.018324	-44.929743	220	5	1.1	0.9;
	8855	1	88.4	51.1	0.161949	133.035662	0	1.040582	-13.396994	380	3	1.1	0.9;
	8856	1	-0.57	-1	0	0	0	1.053444	-29.282615	150	9	1.1	0.9;
	8857	1	-0	-0	0	2.29	0	1.020096	-47.730647	220	6	1.1	0.9;
	8858	2	0	0	0	0	0	1.023098	-43.776563	220	4	1.1	0.9;
	8859	1	-0	-0	0	4.6	0	1.038171	-23.570752	220	15	1.1	0.9;
	8860	1	57.16	19.7	0	22.17	0	1.018319	-10.244627	220	5	1.1	0.9;
	8861	1	-0	0	0	3.95	0	0.984642	-54.835443	220	10	1.1	0.9;
	8862	1	50.96	15.7	0	1.52	0	1.040369	13.653554	110	8	1.1	0.9;
	8863	1	55.56	34.4	0	2.57	0	1.03086	-29.325776	150	9	1.1	0.9;
	8864	1	0	0	0	6.26	0	1.027588	-32.526591	220	5	1.1	0.9;
	8865	1	50.96	16.7	0	0.42	0	0.991171	-30.939935	110	21	1.1	0.9;
	8866	1	-0	0	0	0.48	0	1.0309	-9.648344	220	3	1.1	0.9;
	8867	1	131.5	12.9	0	18.7	0	1.021392	-49.003739	220	4	1.1	0.9;
	8868	1	0	-0	0	5.58	0	1.010425	1.428844	380	3	1.1	0.9;
	8869	2	0	0	0	0	0	1.034597	-39.528177	380	6	1.1	0.9;
	8870	1	-0	0	0	129.47	0	1.049355	-48.447617	380	4	1.1	0.9;
	8871	2	0	0	0	0	0	1.032491	-52.966137	220	4	1.1	0.9;
	8872	2	0	0	0	0	0	1.054777	-16.321937	380	5	1.1	0.9;
	8873	1	0	-0	0	1.87	0	1.068235	-27.574168	220	5	1.1	0.9;
	8874	1	285.2	61.4	0	26.04	0	1.035192	-20.158196	220	5	1.1	0.9;
	8875	2	0	0	0	0	0	1.045119	17.388436	150	8	1.1	0.9;
	8876	1	0	-0	0	31.34	0	1.041399	-40.915836	380	6	1.1	0.9;
	8877	1	-0	0	0	1.03	0	1.036434	-20.951083	220	5	1.1	0.9;
	8878	1	58.26	16.9	0	0.95	0	1.013774	22.290758	110	8	1.1	0.9;
	8879	1	0	0	0	1.47	0	1.032441	-23.098207	220	5	1.1	0.9;
	8880	1	9.21	3.4	0	1.48	0	1.041681	-19.352535	110	19	1.1	0.9;
	8881	1	0	0	0	3.86	0	1.024415	-1.614683	150	8	1.1	0.9;
	8882	1	-0.16	-1.44	0	0.42	0	1.058799	10.888754	110	8	1.1	0.9;
	8883	1	0	0	0	49.35	0	1.065815	-32.388754	380	1	1.1	0.9;
	8884	1	35.64	10.7	0	5.94	0	0.925993	-30.739938	110	18	1.1	0.9;
	8885	1	0	0	0.1444	24.6772	0	1.08736	-14.047593	380	19	1.3	0.7;
	8886	1	0	-0	0	-1.392036	0	0.982229	-4.035795	380	5	1.1	0.9;
	8887	1	0	0	0	102.70867	0	1.043215	1.507901	380	5	1.1	0.9;
	8888	1	101.6	4.5	0	3.59	0	1.032012	-44.872754	220	4	1.1	0.9;
	8889	1	-8.15	-5.23	0	0	0	1.049752	7.358598	150	8	1.1	0.9;
	8890	1	22.12	8	0	1.51	0	1.034323	-21.843664	110	19	1.1	0.9;
	8891	1	18.32	6.6	0	0.45	0	1.06486	-22.276124	110	23	1.1	0.9;
	8892	1	132.1	13.2	0	6.9	0	1.027002	-32.300734	220	6	1.1	0.9;
	8893	1	278.3	33.6	0	25.9	0	1.042688	-15.562576	220	5	1.1	0.9;
	8894	2	0	0	0	0	0	1.093076	-29.068311	380	12	1.1	0.9;
	8895	1	-0.2	-1	0	0	0	1.028671	-1.859193	150	8	1.1	0.9;
	8896	2	0	0	0	0	0	1.032282	6.151305	154	1	1.1	0.9;
	8897	2	0	0	0	0	0	1.053267	-25.681833	110	19	1.1	0.9;
	8898	1	0	0	0	0	0	1.047828	18.203559	380	8	1.1	0.9;
	8899	1	0	-0	0.245843	35.267139	0	1.0431	-16.571185	380	3	1.1	0.9;
	8900	1	3.1	1.6	0	0.09	0	1.043513	-24.062364	220	5	1.1	0.9;
	8901	1	36.04	22.3	0	2.37	0	1.008755	-31.096175	150	9	1.1	0.9;
	8902	1	5.51	2.2	0	3.41	0	1.032502	-34.761287	110	22	1.1	0.9;
	8903	2	0	0	0	0	0	1.023893	-11.312079	380	5	1.1	0.9;
	8904	1	24.63	15.31	0	1.28	0	1.021959	-28.230262	150	9	1.1	0.9;
	8905	1	-0	-0	0.11521	58.265737	0	1.031297	-16.055238	380	7	1.1	0.9;
	8906	1	91	21	0	11.23	0	1.032057	-23.413858	220	5	1.1	0.9;
	8907	1	29.23	-49.9	0	5.49	0	1.02557	-26.70978	220	6	1.1	0.9;
	8908	1	-0	-0	0	0.7	0	1.056729	-27.525127	150	9	1.1	0.9;
	8909	1	98.7	14	0	1.01	0	1.015225	-4.187786	154	1	1.1	0.9;
	8910	1	33.54	12.2	0	2.41	0	1.033582	-7.348785	110	19	1.1	0.9;
	8911	1	32.1	4.5	0	2.23	0	1.021077	-7.356537	154	1	1.1	0.9;
	8912	1	-6.81	2.2	0	0.57	0	1.022368	-23.469255	150	9	1.1	0.9;
	8913	1	-0	0	0	1.13	0	1.037134	-20.861918	220	5	1.1	0.9;
	8914	2	0	0	0	0	0	1.034467	-34.15771	110	12	1.1	0.9;
	8915	1	-0	-0	0	4.12	0	0.99658	9.112436	220	3	1.1	0.9;
	8916	1	0	0	0	4.8	0	1.031753	-10.685374	220	3	1.1	0.9;
	8917	1	59.67	13.1	0	2.26	0	0.996189	5.244024	110	8	1.1	0.9;
	8918	1	10.11	2.5	0	0.18	0	1.014492	-23.519461	110	18	1.1	0.9;
	8919	1	93.1	-15.12	0	1.04	0	1.01895	62.05253	150	1	1.1	0.9;
	8920	2	0	0	0	0	0	1.01084	-41.307091	110	22	1.1	0.9;
	8921	1	0	0	0	16.37	0	0.994236	-3.341618	380	2	1.1	0.9;
	8922	1	3.4	1.1	0	2.14	0	0.995301	-20.66783	110	18	1.1	0.9;
	8923	1	7.41	2.7	0	3.34	0	1.009078	-30.763999	110	19	1.1	0.9;
	8924	1	66.58	0.56	0	7.63	0	1.029314	-6.971888	220	13	1.1	0.9;
	8925	1	6.41	2.3	0	0.93	0	1.052395	-19.630593	110	19	1.1	0.9;
	8926	1	67.08	35.5	0	32.32	0	0.984734	-56.618154	220	10	1.1	0.9;
	8927	2	0	0	0	0	0	1.012236	-35.17317	150	10	1.1	0.9;
	8928	1	122.3	17.5	0	1.51	0	1.012849	-22.093516	154	1	1.1	0.9;
	8929	1	93.9	21.6	0	2.05	0	1.039322	-3.419259	220	3	1.1	0.9;
	8930	1	139.4	21.2	0	6.21	0	0.984831	-18.95199	220	5	1.1	0.9;
	8931	1	-0.56	-0	0	108.21	0	1.08698	-9.874851	380	5	1.1	0.9;
	8932	1	0	-7.5	0	0	0	1.014517	-30.270984	150	9	1.1	0.9;
	8933	1	0	-0	0	2.47	0	0.99084	-50.573447	220	6	1.1	0.9;
	8934	1	-10.07	-3.91	0	14.08	0	1.00634	37.082907	220	3	1.1	0.9;
	8935	1	0	-0	0	26.15	0	1.050895	-13.828639	380	1	1.1	0.9;
	8936	1	-0	-0	0	3.13	0	1.025772	-53.799441	220	4	1.1	0.9;
	8937	1	62.7	9	0	4.6	0	1.01815	-5.782371	154	1	1.1	0.9;
	8938	1	-2.35	-2.06	0	18.45	0	1.033893	55.812115	220	3	1.1	0.9;
	8939	1	186.8	22.9	0	9.94	0	1.009385	-55.359157	220	4	1.1	0.9;
	8940	1	15.82	3.7	0	0.22	0	1.00563	-28.479486	110	18	1.1	0.9;
	8941	1	4.2	1.4	0	0.98	0	1.037713	-33.524636	110	12	1.1	0.9;
	8942	1	19.92	6.6	0	0.18	0	1.007225	-31.884836	110	21	1.1	0.9;
	8943	1	-0	0	0	6.315866	0	0.989314	-26.764082	220	20	1.1	0.9;
	8944	2	0	0	0	0	0	1.06934	57.258502	220	3	1.1	0.9;
	8945	1	0	0	0	43.7	0	1.018329	-1.583907	400	1	1.1	0.9;
	8946	2	0	0	0	0	0	1.036925	-41.807358	380	4	1.1	0.9;
	8947	1	3.7	0.2	0	0.05	0	1.03534	-11.773464	220	5	1.1	0.9;
	8948	1	-0	0	0	3.03	0	1.023879	3.74107	380	3	1.1	0.9;
	8949	1	20.32	6.3	0	0.05	0	1.002255	-27.609858	110	18	1.1	0.9;
	8950	2	0	0	0	0	0	1.050158	-14.888414	220	5	1.1	0.9;
	8951	1	0	-0	0	3.32	0	0.980982	-52.115488	220	6	1.1	0.9;
	8952	1	142.2	33.7	0	6.58	0	1.013006	23.407664	220	3	1.1	0.9;
	8953	1	516.7	144.3	0	2.32	0	1.012701	-40.487754	220	4	1.1	0.9;
	8954	1	51.56	18.7	0	0.18	0	1.02953	-26.425363	110	19	1.1	0.9;
	8955	1	0	0	0	16.23	0	1.0077	6.146306	220	3	1.1	0.9;
	8956	2	0	0	0	0	0	1.061596	22.612591	380	8	1.1	0.9;
	8957	1	0	0	0	4.39	0	0.98498	-54.764172	220	10	1.1	0.9;
	8958	1	28.63	17.7	0	0.1	0	0.967758	-31.427129	150	9	1.1	0.9;
	8959	1	-0	-0	0	0.4	0	1.014491	-30.159914	150	9	1.1	0.9;
	8960	1	-0	-0	0	9.64	0	1.023767	19.630412	380	3	1.1	0.9;
	8961	2	0	0	0	0	0	1.043576	-13.967427	220	5	1.1	0.9;
	8962	1	44.05	21	0	0.03	0	1.001767	-50.907419	220	6	1.1	0.9;
	8963	1	-0	0	0	2.78	0	1.012573	-40.755147	220	10	1.1	0.9;
	8964	1	925.91	147.97	0	0	0	1.019433	-13.307838	150	2	1.1	0.9;
	8965	1	45.05	12.7	0	1.73	0	1.018416	-35.400316	110	22	1.1	0.9;
	8966	2	0	0	0	0	0	1.087897	42.466296	380	3	1.1	0.9;
	8967	1	23.43	8.4	0	0.17	0	1.062579	-25.645512	110	15	1.1	0.9;
	8968	1	261.5	95	0	49.05	0	1.026387	-11.3284	380	3	1.1	0.9;
	8969	1	0	0	0	1.29	0	1.00273	-30.117913	110	21	1.1	0.9;
	8970	1	-17.32	5.7	0	1.52	0	0.959473	-28.847628	150	9	1.1	0.9;
	8971	2	0	0	0	0	0	1.043573	-0.237282	220	13	1.1	0.9;
	8972	2	0	0	0	0	0	1.038393	-33.53821	110	12	1.1	0.9;
	8973	2	0	0	0	0	0	1.048016	-36.544908	380	4	1.1	0.9;
	8974	1	-0	-0	0	44.96	0	1.08148	-35.19762	380	22	1.1	0.9;
	8975	1	0	-0	0	11.04	0	1.039828	-24.602849	220	5	1.1	0.9;
	8976	2	0	0	0	0	0	1.04878	-17.081476	220	5	1.1	0.9;
	8977	1	81.29	26.5	0	7.04	0	1.017795	0.548554	150	8	1.1	0.9;
	8978	2	0	0	0	0	0	1.038763	-7.38624	110	19	1.1	0.9;
	8979	1	0	-0	0	0.51	0	0.973815	-32.032315	150	9	1.1	0.9;
	8980	1	63.97	12.8	0	0.17	0	1.037996	-11.349561	220	5	1.1	0.9;
	8981	2	0	0	0	0	0	1.023844	-43.307726	220	4	1.1	0.9;
	8982	1	0	0	0	1.49	0	1.009058	13.020866	220	3	1.1	0.9;
	8983	1	4.1	1.5	0	1.01	0	1.042963	-22.908146	110	19	1.1	0.9;
	8984	1	18.52	11.1	0	0.07	0	0.959932	-31.695048	150	9	1.1	0.9;
	8985	1	0	0	0	0	0	0.963865	-31.392893	150	9	1.1	0.9;
	8986	1	92.9	48.4	0	5.76	0	1.031926	-21.431005	220	7	1.1	0.9;
	8987	2	0	0	0	0	0	1.045054	-41.760561	380	4	1.1	0.9;
	8988	2	0	0	0	0	0	1.016047	-21.671966	154	1	1.1	0.9;
	8989	1	88.5	24.5	0	2.23	0	1.039012	-25.911894	220	5	1.1	0.9;
	8990	1	2	1	0	0.8	0	1.041302	-26.379277	110	21	1.1	0.9;
	8991	1	-8.49	-17.09	0	0	0	1.01601	3.272318	150	8	1.1	0.9;
	8992	1	-0	-0	0	14.64	0	1.04954	-4.036028	220	5	1.1	0.9;
	8993	2	0	0	0	0	0	1.020391	-38.19801	220	4	1.1	0.9;
	8994	1	69.28	31.3	0	32.76	0	1.036502	5.871334	150	8	1.1	0.9;
	8995	1	11.5	1.7	0	3.08	0	1.020381	-3.284953	154	1	1.1	0.9;
	8996	1	8.7	1.2	0	2.82	0	1.017579	-5.851805	154	1	1.1	0.9;
	8997	2	0	0	0	0	0	1.01701	-12.887035	380	5	1.1	0.9;
	8998	1	-0	0	0	4.08	0	0.968092	-43.530682	220	10	1.1	0.9;
	8999	1	41.15	10.3	0	2.41	0	1.040946	-21.854751	220	5	1.1	0.9;
	9000	2	0	0	0	0	0	1.042081	-40.570202	380	6	1.1	0.9;
	9001	2	0	0	0	0	0	1.023475	-42.079218	220	4	1.1	0.9;
	9002	1	119.02	12.54	0	23.45	0	1.01897	-10.584797	220	5	1.1	0.9;
	9003	1	-0	0	0	31.57	0	1.031835	38.611947	380	3	1.1	0.9;
	9004	1	102.6	14.7	0	2.97	0	1.018564	-23.55241	220	20	1.1	0.9;
	9005	1	-0	0	0.048364	5.54206	0	1.058241	-23.362413	220	18	1.1	0.9;
	9006	1	-0	0	0	-0.44	0	1.021706	-46.280543	380	6	1.1	0.9;
	9007	1	73.68	24	0	0.46	0	1.018066	0.846812	150	8	1.1	0.9;
	9008	1	27.13	16.8	0	1.34	0	1.00554	-33.793222	150	9	1.1	0.9;
	9009	1	145.3	3.8	0	41.054794	0	1.025291	-33.898513	380	6	1.1	0.9;
	9010	1	-0	0	0.270022	107.094537	0	1.086436	-20.936454	380	19	1.1	0.9;
	9011	1	88.4	25.2	0	12.9	0	1.032253	-23.376641	220	5	1.1	0.9;
	9012	1	25.63	6.5	0	0.52	0	1.045421	0.994578	220	5	1.1	0.9;
	9013	2	0	0	0	0	0	1.023407	-47.210196	220	6	1.1	0.9;
	9014	1	70.78	19.6	0	8.1	0	1.046672	-6.75147	220	5	1.1	0.9;
	9015	1	15.63	3.27	0	0.77	0	0.991796	-27.462791	110	18	1.1	0.9;
	9016	1	0	0	0.280385	9.908955	0	1.081763	-25.026346	380	21	1.1	0.9;
	9017	1	5.11	1.3	0	0.8	0	1.021183	-27.412183	110	18	1.1	0.9;
	9018	1	55.46	16.9	0	1.92	0	1.016825	-6.807104	220	5	1.1	0.9;
	9019	1	91.4	18.1	0	4.89	0	1.037991	-25.005711	220	5	1.1	0.9;
	9020	1	6.41	4.3	0	0.23	0	0.967608	-29.896714	110	18	1.1	0.9;
	9021	1	152.18	21.63	0	1.91	0	1.030914	-6.85138	220	5	1.1	0.9;
	9022	1	-0	-0	0	249.03	0	1.018529	3.046714	400	1	1.1	0.9;
	9023	1	80.49	30	0	5.89	0	0.997367	-20.381122	220	3	1.1	0.9;
	9024	1	0	0	0.141019	25.246	0	1.021826	1.47047	380	2	1.1	0.9;
	9025	1	0	-0	0.205903	3.0768	0	1.000113	-4.896569	380	2	1.1	0.9;
	9026	1	0.2	0.1	0	0.28	0	1.068959	-25.032073	110	19	1.1	0.9;
	9027	2	0	0	0	0	0	1.070556	4.312968	380	13	1.1	0.9;
	9028	1	81.99	32.2	0	7.3	0	1.030767	-53.167609	220	4	1.1	0.9;
	9029	1	-0	-0	0	7.92	0	1.014533	16.390126	150	8	1.1	0.9;
	9030	1	132.2	49.1	0	69.43	0	1.046647	-14.344531	380	16	1.1	0.9;
	9031	1	54.3	7.8	0	4.44	0	1.008395	-7.875015	154	1	1.1	0.9;
	9032	1	0	-0	0	2.42	0	1.059483	-47.950771	380	10	1.1	0.9;
	9033	1	76.88	16.4	0	6.61	0	1.050435	-26.861964	220	5	1.1	0.9;
	9034	2	0	0	0	0	0	1.005697	-25.507825	154	1	1.1	0.9;
	9035	2	0	0	0	0	0	1.009337	-18.778819	110	18	1.1	0.9;
	9036	1	-19.83	-13.51	0	0.31	0	1.053739	-26.40549	110	21	1.1	0.9;
	9037	1	-22.37	-8.9	0.162285	11.818574	0	1.099889	45.549831	380	3	1.1	0.9;
	9038	1	0	-0	0	1.03	0	1.061613	-21.371557	150	9	1.1	0.9;
	9039	2	0	0	0	0	0	1.017349	-43.798544	220	4	1.1	0.9;
	9040	1	78.01	4.25	0	12.32	0	1.043881	1.377245	220	13	1.1	0.9;
	9041	1	0	0	0	0.57	0	1.005066	-31.830279	150	9	1.1	0.9;
	9042	1	0	0	0	6.27	0	1.034241	-23.702133	380	1	1.1	0.9;
	9043	2	0	0	0	0	0	1.018389	-40.654877	220	4	1.1	0.9;
	9044	1	0	0	0	4.74	0	1.027751	-33.642346	220	22	1.1	0.9;
	9045	1	60.57	18.9	0	1.46	0	1.037474	-24.957769	220	5	1.1	0.9;
	9046	2	0	0	0	0	0	1.070778	44.490271	380	3	1.1	0.9;
	9047	2	0	0	0	0	0	1.03154	66.593064	380	1	1.1	0.9;
	9048	2	0	0	0	0	0	1.050428	-12.580966	220	7	1.1	0.9;
	9049	1	36.84	14.1	0	3.8	0	1.016347	-40.800283	220	4	1.1	0.9;
	9050	1	0	0	0	0	0	0.87466	-38.032264	150	9	1.3	0.7;
	9051	1	0	-0	0	2.15	0	1.033855	0.35535	220	5	1.1	0.9;
	9052	1	121.8	17.4	0	1.01	0	1.017289	-10.01002	154	1	1.1	0.9;
	9053	1	-0	-0	0	0.44	0	1.030453	48.268546	220	3	1.1	0.9;
	9054	2	0	0	0	0	0	1.116695	-23.366531	380	21	1.3	0.7;
	9055	2	0	0	0	0	0	0.97622	-50.38012	220	4	1.1	0.9;
	9056	1	11.35	3.27	0	1.47	0	1.015088	-26.281333	110	18	1.1	0.9;
	9057	1	72.38	23	0	0.28	0	1.020685	-48.779675	220	4	1.1	0.9;
	9058	2	0	0	0	0	0	1.013989	63.346088	150	1	1.1	0.9;
	9059	1	0	-0	0	2	0	1.06958	-27.481336	220	5	1.1	0.9;
	9060	2	0	0	0	0	0	1.066622	0.64924	380	14	1.1	0.9;
	9061	1	27.9	4	0	4.34	0	1.002505	-8.594522	154	1	1.1	0.9;
	9062	1	86.1	40.6	0	7.39	0	1.085303	-27.976492	150	9	1.1	0.9;
	9063	1	-41.57	-23.4	0	0	0	1.025839	11.378876	150	8	1.1	0.9;
	9064	1	0	-0	0	15.26	0	0.997517	-4.330374	380	2	1.1	0.9;
	9065	1	-0	-0	0	0.59	0	1.02925	2.262276	380	5	1.1	0.9;
	9066	1	23.26	3.01	0	9.29	0	1.019416	-20.851529	220	5	1.1	0.9;
	9067	2	0	0	0	0	0	1.03898	1.326718	220	5	1.1	0.9;
	9068	1	-0	-0	0	11.35	0	1.016522	-47.563374	220	6	1.1	0.9;
	9069	1	13.9	1.9	0	3.36	0	1.029345	3.868055	154	1	1.1	0.9;
	9070	1	0	-0	0	0.85	0	1.014526	-27.088277	150	9	1.1	0.9;
	9071	1	8.3	1.2	0	0.53	0	1.007888	-5.635688	154	1	1.1	0.9;
	9072	1	0	-0	0	1.41	0	1.038832	4.912377	150	8	1.1	0.9;
	9073	1	-0	0	0	5.05	0	0.999787	-38.482125	220	22	1.1	0.9;
	9074	1	46.9	6.7	0	3.45	0	1.022605	13.774034	154	1	1.1	0.9;
	9075	1	242.4	5.1	0	4.93	0	1.020531	-43.139819	220	6	1.1	0.9;
	9076	1	0	0	0	0.55	0	0.936148	-34.567675	150	9	1.1	0.9;
	9077	1	213.61	50.1	0	0	0	1.00627	19.875811	220	3	1.1	0.9;
	9078	1	24.33	9.9	0	0.76	0	0.983071	-28.908274	110	18	1.1	0.9;
	9079	1	0	0	0	0	0	1.020771	-29.719031	150	9	1.1	0.9;
	9080	1	188.6	164.9	0	21.03	0	1.016757	-13.428436	380	3	1.1	0.9;
	9081	1	0	0	0	0.59	0	1.049272	-22.144817	150	9	1.1	0.9;
	9082	1	50.2	7.1	0	3.52	0	1.014442	-14.397143	154	1	1.1	0.9;
	9083	1	88.1	36	0	27.89	0	1.025316	-1.145306	150	8	1.1	0.9;
	9084	1	0	-0	0	5.88	0	1.031473	20.252563	380	3	1.1	0.9;
	9085	1	41.6	5.9	0	2.5	0	1.023432	-7.241486	154	1	1.1	0.9;
	9086	1	125	9.6	0	0.25	0	1.033977	-13.93054	220	3	1.1	0.9;
	9087	1	33.5	4.8	0	1.98	0	1.040157	2.241184	154	1	1.1	0.9;
	9088	1	55.9	102.9	0	23.9	0	1.027431	34.922619	380	3	1.1	0.9;
	9089	1	-0	-0	0	4.5	0	1.02776	18.322319	154	1	1.1	0.9;
	9090	1	0	0	0	0.58	0	1.009716	51.520897	220	3	1.1	0.9;
	9091	1	0	0	0	3.65	0	1.038463	-9.834074	220	5	1.1	0.9;
	9092	1	13.94	9.5	0	3.53	0	1.027912	-30.435624	150	9	1.1	0.9;
	9093	1	3.8	2.5	0	0.84	0	1.012092	-16.019132	110	18	1.1	0.9;
	9094	1	187.4	72.8	0	34.95	0	1.035337	5.426812	150	8	1.1	0.9;
	9095	1	119	97.8	0	0.14	0	0.994238	9.203678	220	3	1.1	0.9;
	9096	1	7	1	0	0.78	0	1.025278	-2.821658	154	1	1.1	0.9;
	9097	2	0	0	0	0	0	1.016295	-40.223897	220	4	1.1	0.9;
	9098	1	0	-0	0	1.77	0	1.005097	-32.912516	150	9	1.1	0.9;
	9099	2	0	0	0	0	0	1.028779	-17.085682	154	1	1.1	0.9;
	9100	2	0	0	0	0	0	1.07025	61.624988	380	3	1.1	0.9;
	9101	2	0	0	0	0	0	1.026276	-22.084157	380	5	1.1	0.9;
	9102	1	-0	0	0	2.83	0	1.016926	-6.147136	154	1	1.1	0.9;
	9103	2	0	0	0	0	0	1.027331	-16.457891	154	1	1.1	0.9;
	9104	1	11.31	13.8	0	1.11	0	1.021698	-29.695588	150	9	1.1	0.9;
	9105	2	0	0	0	0	0	1.088711	-16.324228	110	18	1.3	0.7;
	9106	1	-0	0	0	1.12	0	0.982322	-56.142555	220	4	1.1	0.9;
	9107	1	87.4	35.5	0	16.41	0	1.044678	-18.435797	380	16	1.1	0.9;
	9108	2	0	0	0	0	0	1.046328	-16.424177	220	5	1.1	0.9;
	9109	1	0	-0	0	0.31	0	1.033857	0.35533	220	5	1.1	0.9;
	9110	2	0	0	0	0	0	1.126918	-21.871563	380	9	1.3	0.7;
	9111	1	35.14	12.7	0	1.42	0	1.070321	-17.970841	110	19	1.1	0.9;
	9112	1	0	-0	0	19.71	0	1.053965	-13.349394	220	5	1.1	0.9;
	9113	2	0	0	0	0	0	1.014547	-36.06836	220	6	1.1	0.9;
	9114	1	0	0	0	47.1	0	1.026014	3.694409	400	1	1.1	0.9;
	9115	1	25.03	12	0	2.76	0	1.014174	-30.322741	220	6	1.1	0.9;
	9116	1	152.7	81	0	17.54	0	1.014534	-13.764914	380	3	1.1	0.9;
	9117	1	36.74	14.1	0	16.98	0	1.017517	-52.837809	380	4	1.1	0.9;
	9118	1	-0.22	-3.34	0	0	0	1.031779	3.402444	150	8	1.1	0.9;
	9119	1	-0	0	0	1.24	0	1.039783	-20.831188	220	5	1.1	0.9;
	9120	1	0	0	0	4.13	0	1.021557	2.421393	150	8	1.1	0.9;
	9121	1	349.6	96.3	0	105.72	0	1.035973	-44.757751	380	6	1.1	0.9;
	9122	1	50.46	-10.8	0	42.99	0	1.126477	-26.62396	380	1	1.3	0.7;
	9123	1	13.9	1.9	0	5.87	0	1.028128	18.829626	154	1	1.1	0.9;
	9124	1	22.02	7.1	0	1.38	0	1.016906	-22.219757	110	21	1.1	0.9;
	9125	2	0	0	0	0	0	1.042649	-20.934031	220	19	1.1	0.9;
	9126	1	0	-0	0	32.64	0	1.016113	-33.749547	380	1	1.1	0.9;
	9127	1	38.84	9.6	0	0.88	0	0.999409	20.870776	110	8	1.1	0.9;
	9128	1	50.56	8.7	0	0.19	0	1.031517	-2.952699	220	5	1.1	0.9;
	9129	1	39.5	5.6	0	8.26	0	1.010953	-25.770716	154	1	1.1	0.9;
	9130	1	245.51	9.17	0	24.25	0	1.031578	-31.505951	220	5	1.1	0.9;
	9131	1	-0	-0	0	6.62	0	1.046774	-24.421469	220	5	1.1	0.9;
	9132	2	0	0	0	0	0	1.04609	-27.739751	380	6	1.1	0.9;
	9133	1	0	0	0	7.12	0	1.033538	-40.135467	380	17	1.1	0.9;
	9134	1	21.72	13.4	0	2.38	0	0.999342	-24.208457	150	9	1.1	0.9;
	9135	1	169.04	33.9	0	5.48	0	0.992973	-52.429746	220	6	1.1	0.9;
	9136	1	-0	0	0	0	0	0.988179	-54.441306	220	10	1.1	0.9;
	9137	2	0	0	0	0	0	1.048884	-26.886512	220	5	1.1	0.9;
	9138	1	16.7	2.4	0	2.38	0	1.013519	-9.389245	154	1	1.1	0.9;
	9139	1	431.5	111.5	0	36.709841	0	1.019118	-48.988879	380	6	1.1	0.9;
	9140	2	0	0	0	0	0	1.042971	-5.426378	220	5	1.1	0.9;
	9141	1	0	0	0	0	0	0.975873	-35.299892	150	9	1.1	0.9;
	9142	1	0	0	0	0	0	1.009593	-24.521036	150	9	1.1	0.9;
	9143	1	89.6	96.2	0	48.28	0	1.051592	9.362104	380	3	1.1	0.9;
	9144	1	0	0	0.24842	10.168602	0	1.144394	-11.902874	220	19	1.3	0.7;
	9145	1	66.57	23.6	0	0.14	0	1.063676	-21.93976	110	19	1.1	0.9;
	9146	1	35.44	11.4	0	1.15	0	1.025255	7.725131	110	8	1.1	0.9;
	9147	1	0	0	0	0	0	0.99121	-30.809775	150	9	1.1	0.9;
	9148	1	15.22	5.1	0	0.36	0	1.078994	-16.53734	110	18	1.1	0.9;
	9149	1	0	0	0	0	0	1.00388	-31.357499	150	9	1.1	0.9;
	9150	2	0	0	0	0	0	1.034687	-13.442439	380	5	1.1	0.9;
	9151	1	5.5	0.8	0	8.36	0	1.028823	6.500564	154	1	1.1	0.9;
	9152	1	-0	0	0	17.63	0	1.044039	14.370587	380	3	1.1	0.9;
	9153	1	2.2	1.1	0	1.31	0	1.032631	-38.891322	110	22	1.1	0.9;
	9154	1	83.19	37.3	0	28.16	0	1.027702	17.543704	380	3	1.1	0.9;
	9155	1	-0	-0	0	5.01	0	1.023245	-20.905815	220	5	1.1	0.9;
	9156	1	0	8.5	0	12.99	0	1.031942	-28.003569	150	9	1.1	0.9;
	9157	1	0	0	0	0	0	0.942596	-32.670874	150	9	1.1	0.9;
	9158	1	54.96	9.42	0	5.92	0	1.022273	-14.756084	220	5	1.1	0.9;
	9159	1	-0	-0	0	81.15	0	1.045228	-43.943091	380	4	1.1	0.9;
	9160	1	311.4	310.5	0.199001	104.051798	0	1.036025	0.526834	380	3	1.1	0.9;
	9161	2	0	0	0	0	0	1.033176	-21.938164	220	7	1.1	0.9;
	9162	1	94.9	40.3	0	17.2	0	1.008691	-61.125492	220	4	1.1	0.9;
	9163	1	0	0	0	0	0	0.980011	-30.352301	150	9	1.1	0.9;
	9164	1	7.71	2.4	0	6.93	0	1.000808	-19.55897	220	5	1.1	0.9;
	9165	1	244.6	96.3	0	7.45	0	1.012348	-17.920039	220	3	1.1	0.9;
	9166	1	9.91	2	0	0.23	0	1.013916	-26.3912	110	18	1.1	0.9;
	9167	2	0	0	0	0	0	1.046867	-25.749667	380	6	1.1	0.9;
	9168	2	0	0	0	0	0	1.076561	-19.394973	220	21	1.1	0.9;
	9169	1	126.21	-7.73	0	6.2	0	1.043761	-18.939825	220	7	1.1	0.9;
	9170	1	43.15	26.7	0	0.11	0	0.995012	-31.350024	150	9	1.1	0.9;
	9171	1	6.01	3	0	0.43	0	1.048212	-25.50859	110	23	1.1	0.9;
	9172	1	0	0	0	0	0	0.971529	-34.186542	150	9	1.1	0.9;
	9173	1	66.07	15.6	0	5.57	0	1.029075	-20.517581	220	5	1.1	0.9;
	9174	2	0	0	0	0	0	1.015292	-33.281168	380	5	1.1	0.9;
	9175	1	20.32	6.7	0	1.27	0	1.025176	-34.235555	110	12	1.1	0.9;
	9176	1	-0	0	0	1.28	0	1.054197	-13.351984	220	5	1.1	0.9;
	9177	1	11.41	3.7	0	0.63	0	1.009783	-36.09656	110	12	1.1	0.9;
	9178	1	4.61	0.8	0	2.46	0	1.036985	-20.124603	110	19	1.1	0.9;
	9179	1	-0.68	-2.44	0	0.02	0	1.031959	-29.388091	150	9	1.1	0.9;
	9180	2	0	0	0	0	0	1.053552	-9.013967	220	5	1.1	0.9;
	9181	1	2.43	1.35	0	0.1	0	1.032604	-9.514471	220	5	1.1	0.9;
	9182	1	29.7	4.3	0	13.67	0	1.025252	-2.823766	154	1	1.1	0.9;
	9183	1	156.3	34.9	0	125.56	0	1.060831	-28.67579	380	20	1.1	0.9;
	9184	1	50.06	44	0	52.57	0	1.032579	-20.429918	220	11	1.1	0.9;
	9185	1	66.97	18.6	0	18.5	0	1.047418	-6.692485	220	5	1.1	0.9;
	9186	1	-28.33	9.3	0	0.06	0	1.001193	-30.889935	150	9	1.1	0.9;
	9187	2	0	0	0	0	0	1.017206	-28.96425	220	6	1.1	0.9;
	9188	1	78	11.1	0	1.33	0	1.026633	-17.377747	154	1	1.1	0.9;
	9189	1	62.57	-61	0	55.14	0	1.04364	-25.354346	220	5	1.1	0.9;
	9190	1	0	0	0	0.98	0	1.038421	10.701718	220	3	1.1	0.9;
	9191	1	-0	-0	0	2.34	0	1.068081	-25.1908	220	5	1.1	0.9;
	9192	1	17.92	23.3	0	0	0	1.026755	-11.14904	150	2	1.1	0.9;
	9193	1	0	-0	0	1.15	0	1.047872	-26.503737	110	21	1.1	0.9;
	9194	1	0	0	0.262086	66.559018	0	1.054181	-20.330877	380	19	1.1	0.9;
	9195	2	0	0	0	0	0	1.025329	-24.847964	110	3	1.1	0.9;
	9196	1	18.71	0.64	0	1.58	0	1.015066	-42.242142	110	22	1.1	0.9;
	9197	2	0	0	0	0	0	1.017237	-47.991389	220	6	1.1	0.9;
	9198	1	282.2	268.3	0	0.28	0	0.992287	9.626195	220	3	1.1	0.9;
	9199	1	14.42	3.8	0	1.24	0	0.918836	-35.875884	110	18	1.1	0.9;
	9200	2	0	0	0	0	0	1.040977	9.324287	220	3	1.1	0.9;
	9201	1	137.9	25.6	0	8.24	0	1.021708	-43.462923	220	4	1.1	0.9;
	9202	1	15.42	6.6	0	2.46	0	1.023129	-39.027798	110	22	1.1	0.9;
	9203	1	-0	-0	0	50.185221	0	1.013724	-12.348813	380	5	1.1	0.9;
	9204	1	11.71	3.8	0	0.41	0	0.987777	-26.396232	110	18	1.1	0.9;
	9205	1	23.43	7.3	0	0.89	0	1.024091	13.095331	110	8	1.1	0.9;
	9206	1	226.1	88.8	0	122.78	0	1.032652	3.577999	150	8	1.1	0.9;
	9207	1	160.2	22.8	0	14.42	0	1.022433	-7.901761	154	1	1.1	0.9;
	9208	1	-0	0	0	2.29	0	1.057265	-2.779327	220	3	1.1	0.9;
	9209	1	11.44	2.38	0	0.78	0	0.992157	-27.396182	110	18	1.1	0.9;
	9210	1	12.11	4.4	0	2.35	0	0.958193	-22.277418	110	19	1.1	0.9;
	9211	1	217.7	164.6	0	11.88	0	1.021484	5.206879	380	3	1.1	0.9;
	9212	1	14.81	0.6	0	0.68	0	1.034219	-23.665647	110	3	1.1	0.9;
	9213	1	-0	-0	0	10.210952	0	1.021887	-4.683055	380	5	1.1	0.9;
	9214	1	1.3	-0	0	1.1	0	0.945446	-32.670959	110	18	1.1	0.9;
	9215	2	0	0	0	0	0	1.008696	30.844349	220	13	1.1	0.9;
	9216	1	103.6	4.69	0	36.68	0	1.022236	-48.806553	220	4	1.1	0.9;
	9217	1	-0	0	0	6.33	0	1.037712	-9.095499	220	5	1.1	0.9;
	9218	1	-0	-0	0.048364	5.66206	0	1.053757	-23.95137	220	18	1.1	0.9;
	9219	1	39.34	13.3	0	0.67	0	0.976857	-32.419649	110	21	1.1	0.9;
	9220	1	26.4	3.8	0	7.48	0	1.034016	-7.070462	154	1	1.1	0.9;
	9221	1	0	0	0	6.58	0	1.04807	12.002102	380	3	1.1	0.9;
	9222	1	-5.15	-2.04	0	154.446402	0	1.05418	-16.505219	380	5	1.1	0.9;
	9223	1	133.5	24.1	0	2.92	0	1.013922	6.927286	110	8	1.1	0.9;
	9224	1	-0	-0	0	0.04	0	1.015032	-5.799018	154	1	1.1	0.9;
	9225	1	-0	0	0	5.98	0	1.015668	-6.175679	220	3	1.1	0.9;
	9226	1	0	0	0	0	0	1.046637	-27.138917	150	9	1.1	0.9;
	9227	1	-0	0	0.090774	11.940444	0	1.048335	-18.202764	220	18	1.1	0.9;
	9228	1	-0	-0	0	6.95	0	1.004437	65.850708	380	1	1.1	0.9;
	9229	1	38.34	12.6	0	0.58	0	1.029484	-34.223347	110	12	1.1	0.9;
	9230	2	0	0	0	0	0	1.027519	-7.771709	220	14	1.1	0.9;
	9231	1	242.08	59.98	0	5.6	0	1.0264	-19.423195	220	5	1.1	0.9;
	9232	1	31.94	8.8	0	0.73	0	1.032761	-33.158924	110	22	1.1	0.9;
	9233	1	157	44.2	0	22.63	0	1.032731	-3.177461	220	1	1.1	0.9;
	9234	1	-121.89	-52.61	0	0.48	0	1.042353	-0.37878	220	13	1.1	0.9;
	9235	2	0	0	0	0	0	1.04114	-22.50441	380	7	1.1	0.9;
	9236	1	246.5	-1.9	0	31.37	0	1.027495	-39.134578	380	6	1.1	0.9;
	9237	1	-0	-0	0	1.27	0	1.047699	-23.228502	110	19	1.1	0.9;
	9238	2	0	0	0	0	0	1.097633	15.484001	380	13	1.3	0.7;
	9239	2	0	0	0	0	0	1.017601	6.12547	380	2	1.1	0.9;
	9240	1	0	-0	0	3.66	0	0.960374	-31.768249	150	9	1.1	0.9;
	9241	1	0	-0	0	0.36	0	1.044245	-9.735227	220	5	1.1	0.9;
];

%% generator data
%	bus	Pg	Qg	Qmax	Qmin	Vg	mBase	status	Pmax	Pmin	Pc1	Pc2	Qc1min	Qc1max	Qc2min	Qc2max	ramp_agc	ramp_10	ramp_30	ramp_q	apf
mpc.gen = [
	2	-81.4	6.64	42.19	-20.45	1.031734	100	1	100	-108.53	0	0	0	0	0	0	0	0	0	0	0;
	6	701	183.79	524.53	-211.57	1.038476	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	8	499.7	-0.95	251.73	-246.25	1.023005	100	1	524.68	0	0	0	0	0	0	0	0	0	0	0	0;
	18	467.5	201.91	212.97	-93.48	1.026728	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	25	319.6	164.04	179.43	-66.21	1.028526	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	32	8	-1.02	5.91	-4.56	1.003223	100	1	16.2	0	0	0	0	0	0	0	0	0	0	0	0;
	33	5.3	-0.56	2.27	-1.08	1.026534	100	1	5.38	0	0	0	0	0	0	0	0	0	0	0	0;
	39	34.3	4.77	17.76	-7.14	1.003511	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	49	309.2	62.33	218.52	-75.97	1.020323	100	1	480	0	0	0	0	0	0	0	0	0	0	0	0;
	51	-144.5	-0.31	5.48	-4.19	1.006665	100	1	100	-192.67	0	0	0	0	0	0	0	0	0	0	0;
	80	43.2	2.23	139.94	-43.78	1.068393	100	1	100	28.8	0	0	0	0	0	0	0	0	0	0	0;
	107	570	-75.96	253.85	-113.79	1.083135	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	123	24.9	8.45	18.84	-6.06	1.041503	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	124	861.3	69.97	440.32	-172.59	1.077112	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	141	-73.2	5.11	18.59	-6.33	1.050716	100	1	100	-97.6	0	0	0	0	0	0	0	0	0	0	0;
	142	158.7	2.2	78.15	-32.4	1.042299	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	147	-16	9.74	70.12	-29.5	1.058041	100	1	100	-21.33	0	0	0	0	0	0	0	0	0	0	0;
	150	67.29	-5.08	77.09	-22.54	1.062129	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	151	-110.35	-11.36	37.97	-58.18	1.013748	100	1	100	-147.13	0	0	0	0	0	0	0	0	0	0	0;
	166	292.2	107.13	127.67	-56.53	1.040747	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	179	-179.2	-30.19	52.64	-51.72	1.043661	100	1	100	-238.93	0	0	0	0	0	0	0	0	0	0	0;
	190	90.69	31	54.24	-20.63	1.025544	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	192	135.8	12.15	91.61	-30.98	1.048355	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	194	268.6	42.2	130.19	-53.53	1.02931	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	197	292.2	-57.05	123.89	-59.54	1.099985	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	201	5.2	-0.6	2.27	-1.07	1.01914	100	1	5.37	0	0	0	0	0	0	0	0	0	0	0	0;
	209	25.4	9.44	18.7	-6.17	1.026832	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	214	-175.2	1.82	192	-52.89	1.079765	100	1	100	-233.6	0	0	0	0	0	0	0	0	0	0	0;
	219	1171	704.29	937.43	-288.16	1.056623	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	220	668.7	-115.27	402.92	-146.95	1.047116	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	221	296	43.03	183.92	-65.3	1.030636	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	222	-147.3	2.36	12.51	-5.62	1.030494	100	1	100	-196.4	0	0	0	0	0	0	0	0	0	0	0;
	239	-7.05	7.21	38.52	-11.24	1.018711	100	1	100	-9.4	0	0	0	0	0	0	0	0	0	0	0;
	246	307.1	-22.47	378.08	-112.59	1.06685	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	248	-172	11.84	37.47	-12.35	1.043991	100	1	100	-229.33	0	0	0	0	0	0	0	0	0	0	0;
	255	423.2	-30.86	220.66	-85.89	1.059597	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	262	-115.4	12.33	38.1	-11.72	1.033795	100	1	100	-153.87	0	0	0	0	0	0	0	0	0	0	0;
	263	-34.4	-3.16	5.05	-6.95	1.028681	100	1	100	-45.87	0	0	0	0	0	0	0	0	0	0	0;
	278	-64.1	7.04	37.7	-12.09	1.014153	100	1	100	-85.47	0	0	0	0	0	0	0	0	0	0	0;
	282	1245.7	-279.39	924.64	-300.35	1.077605	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	289	52.7	-3.66	26.02	-10.81	1.046783	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	294	250	10.09	45.1	-21.32	1.003375	100	1	480	160	0	0	0	0	0	0	0	0	0	0	0;
	299	975.9	97.97	721.25	-258.71	1.039248	100	1	1600	533.33	0	0	0	0	0	0	0	0	0	0	0;
	309	95.1	36.53	42.62	-18.72	0.989085	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	310	346.18	632.57	Inf	-Inf	1.155322	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	313	9.6	-1.71	4.83	-7.17	1.025259	100	1	10.08	0	0	0	0	0	0	0	0	0	0	0	0;
	320	1760	-355.74	320.04	-843.96	1.013336	100	1	1812	604	0	0	0	0	0	0	0	0	0	0	0;
	324	94	30.47	95.32	-27.29	1.063285	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	325	-21.4	0.06	1.93	-1.01	1.009625	100	1	100	-28.53	0	0	0	0	0	0	0	0	0	0	0;
	328	-77.9	0.12	188.91	-56.14	1.017618	100	1	100	-103.87	0	0	0	0	0	0	0	0	0	0	0;
	338	83.7	3.24	55.54	-19.16	1.046957	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	347	50.8	-13.47	33.33	-42.67	1.020949	100	1	53.42	0	0	0	0	0	0	0	0	0	0	0	0;
	348	160.4	3.38	110.57	-36.9	1.057894	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	352	1250.4	256.45	926.35	-297.68	1.048835	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	361	238.3	27.11	186.39	-58.5	1.05949	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	366	185.6	-33.91	61.97	-120.27	1.008853	100	1	196	65.33	0	0	0	0	0	0	0	0	0	0	0;
	372	-54.9	-13.56	282.95	-84.3	1.074991	100	1	100	-73.2	0	0	0	0	0	0	0	0	0	0	0;
	382	25.68	11.15	38.4	-11.44	1.060853	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	389	69	17.57	56.73	-17.84	1.017731	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	394	2404.4	-96.56	1224.26	-490.17	1.063147	100	1	2800	933.33	0	0	0	0	0	0	0	0	0	0	0;
	406	31	10.74	22.22	-17.35	1.012399	100	1	61.23	0	0	0	0	0	0	0	0	0	0	0	0;
	409	-62	-7.35	24.32	-29.67	1.032831	100	1	100	-82.67	0	0	0	0	0	0	0	0	0	0	0;
	411	35.9	4.64	17.66	-7.27	1.03248	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	413	231	21.57	124.12	-49.84	1.048421	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	423	-19.2	0.02	5.22	-2.6	1.030299	100	1	100	-25.6	0	0	0	0	0	0	0	0	0	0	0;
	424	636.3	269.64	304.38	-125.48	1.051241	100	1	700	233.33	0	0	0	0	0	0	0	0	0	0	0;
	433	436	206.71	217.18	-89.22	1.024885	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	434	-2.8	-1.11	2.75	-3.25	1.037999	100	1	100	-3.73	0	0	0	0	0	0	0	0	0	0	0;
	438	74	9.04	45.1	-16.18	1.038705	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	439	308.6	-59.31	175.16	-69.21	1.072491	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	440	83.1	-1.51	55.73	-19.01	1.028093	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	442	-40.35	-0.32	1.95	-0.93	1.011727	100	1	100	-53.8	0	0	0	0	0	0	0	0	0	0	0;
	447	37.7	2.42	16.84	-8.04	1.035215	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	452	-56.8	16.67	37.86	-11.94	1.013382	100	1	100	-75.73	0	0	0	0	0	0	0	0	0	0	0;
	453	260.1	69.83	257.23	-74.4	1.024712	100	1	540	0	0	0	0	0	0	0	0	0	0	0	0;
	456	-28.08	-11.78	44.8	-39.81	0.999847	100	1	100	-37.44	0	0	0	0	0	0	0	0	0	0	0;
	484	229.5	-43.46	132.41	-50.85	1.054091	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	489	-107.4	-0.53	6.28	-4.82	1.008032	100	1	100	-143.2	0	0	0	0	0	0	0	0	0	0	0;
	491	105.4	83.32	94.56	-28.1	1.033193	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	505	110.6	-14.73	37.34	-58.33	1.022173	100	1	116.03	0	0	0	0	0	0	0	0	0	0	0	0;
	511	101.4	-14.26	94.38	-28.04	1.081589	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	515	324.82	28.39	72.5	-27.24	1.045162	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	516	539.4	184.01	261.79	-106.67	1.07129	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	518	-410.3	1.69	85.63	-37.21	1.067615	100	1	100	-547.07	0	0	0	0	0	0	0	0	0	0	0;
	524	69	-13.09	30.15	-43.07	1.046979	100	1	72.45	0	0	0	0	0	0	0	0	0	0	0	0;
	541	118.2	-15	7.81	-27.19	1.03819	100	1	124.19	0	0	0	0	0	0	0	0	0	0	0	0;
	564	-51.67	-3.19	18.47	-19.35	1.06702	100	1	100	-68.89	0	0	0	0	0	0	0	0	0	0	0;
	578	-110.3	6.59	25.88	-11.01	1.042146	100	1	100	-147.07	0	0	0	0	0	0	0	0	0	0	0;
	583	-270	4.89	17.88	-7.03	1.048595	100	1	100	-360	0	0	0	0	0	0	0	0	0	0	0;
	587	246	1.47	185.56	-59.39	1.040079	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	590	100.7	135.79	380.31	-110.08	1.065004	100	1	300	67.13	0	0	0	0	0	0	0	0	0	0	0;
	615	38.2	2.13	17.09	-7.79	1.068592	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	616	379.2	31.85	217.9	-81.16	1.04484	100	1	480	160	0	0	0	0	0	0	0	0	0	0	0;
	618	670.3	-26.77	351.18	-139.11	1.022784	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	628	152.4	41.23	78.79	-31.69	0.982057	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	636	101.89	2.48	52.99	-21.87	1.04241	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	637	21.3	-0.33	11.2	-8.44	1.031148	100	1	22.32	0	0	0	0	0	0	0	0	0	0	0	0;
	639	-46.2	106.37	135.52	-38.9	1.033532	100	1	100	-61.6	0	0	0	0	0	0	0	0	0	0	0;
	642	63.33	-55.21	90.13	-93.03	1.031368	100	1	276.09	0	0	0	0	0	0	0	0	0	0	0	0;
	648	47.2	2.22	37.95	-11.87	1.02749	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	672	67	-0.69	71.32	-62.92	1.02121	100	1	205.81	0	0	0	0	0	0	0	0	0	0	0	0;
	682	37.4	8.11	38.77	-11.06	1.060787	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	686	-21.65	-8.24	34.75	-34.26	1.013903	100	1	100	-28.86	0	0	0	0	0	0	0	0	0	0	0;
	692	274.8	-4.69	182.17	-62.79	1.059006	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	697	634.8	27.98	357.4	-132.75	1.031719	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	698	274.8	3.65	129.94	-53.91	0.999662	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	709	16	-8.56	22.23	-20.83	1.068875	100	1	65.3	0	0	0	0	0	0	0	0	0	0	0	0;
	711	27.8	52.62	377.01	-113.63	1.070147	100	1	100	18.53	0	0	0	0	0	0	0	0	0	0	0;
	713	99.3	8.47	53.98	-20.79	1.034791	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	714	453.4	62.78	376.36	-114.08	1.065017	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	715	323.1	-67.84	66.42	-179.95	1.021604	100	1	366	122	0	0	0	0	0	0	0	0	0	0	0;
	727	384.26	-63.12	379.61	-111.09	1.070143	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	728	70.8	-18.79	35.73	-44.31	1.030375	100	1	74.28	0	0	0	0	0	0	0	0	0	0	0	0;
	735	170	-27.79	89.94	-122.06	1.026593	100	1	178.5	0	0	0	0	0	0	0	0	0	0	0	0;
	738	90.79	47.3	107.86	-95.15	1.033376	100	1	311.39	34.6	0	0	0	0	0	0	0	0	0	0	0;
	739	465.7	-58.22	212.97	-93.45	1.035442	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	749	198.88	4.93	105.69	-43.84	1.053546	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	751	65.9	-20.17	136.94	-47.24	1.020488	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	757	116.7	15.15	51.3	-23.41	1.013994	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	763	37.3	5.58	27.88	-8.97	1.042899	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	764	138.3	8.35	174.17	-71.12	1.048246	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	765	170.7	20.09	76.76	-33.84	1.017174	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	774	134.4	-8.34	81.57	-60.74	1.038992	100	1	141.12	0	0	0	0	0	0	0	0	0	0	0	0;
	776	-111.4	5.14	18.16	-6.73	1.023986	100	1	100	-148.53	0	0	0	0	0	0	0	0	0	0	0;
	778	16.5	-0.64	12.37	-9.36	1.047837	100	1	33.58	0	0	0	0	0	0	0	0	0	0	0	0;
	779	37.8	12.06	17.34	-7.58	1.039304	100	1	45	0	0	0	0	0	0	0	0	0	0	0	0;
	780	50.8	17.14	37.43	-12.32	1.001975	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	791	569.74	-78	567.63	-167.48	1.052091	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	795	58.4	7.1	36.75	-13.04	1.041931	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	796	31.6	19.62	39	-10.74	1.01604	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	803	-89.2	9.91	74.03	-25.54	1.04662	100	1	100	-118.93	0	0	0	0	0	0	0	0	0	0	0;
	807	1597.5	-195.51	864.47	-360.97	1.053518	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	813	505.6	0.16	309.08	-290.93	1.037003	100	1	602	200.67	0	0	0	0	0	0	0	0	0	0	0;
	823	1274.2	287.09	926.68	-298.31	1.039055	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	828	55.9	-11.69	20.87	-22.43	1.044382	100	1	58.7	0	0	0	0	0	0	0	0	0	0	0	0;
	831	85.2	16.49	55.59	-19.18	1.003271	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	832	507	137.14	264.71	-102.83	1.004583	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	837	44	7.99	27.19	-9.66	1.022737	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	844	97.5	14.12	54.35	-20.47	1.064419	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	848	4.03	-7.12	49.63	-38.12	1.011431	100	1	100	2.69	0	0	0	0	0	0	0	0	0	0	0;
	849	8.7	-1.18	6.45	-4.91	1.037264	100	1	17.57	0	0	0	0	0	0	0	0	0	0	0	0;
	851	59.56	41.16	96.64	-27.81	1.018845	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	858	855.6	160.89	440.67	-172.15	1.028813	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	861	55.8	10.72	57.98	-16.68	1.050818	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	875	343.5	45.89	228.16	-78.19	1.038959	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	888	181.98	-7.29	138.6	-45.27	1.037604	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	891	1702.8	696.58	1097.41	-373.91	1.062964	100	1	2400	800	0	0	0	0	0	0	0	0	0	0	0;
	894	264.67	69.9	141.93	-57.74	1.054199	100	1	320	0	0	0	0	0	0	0	0	0	0	0	0;
	896	-2.7	-2.25	8.6	-11.4	1.021987	100	1	100	-3.6	0	0	0	0	0	0	0	0	0	0	0;
	913	1249.4	683.52	772.11	-269.94	1.026545	100	1	1700	566.67	0	0	0	0	0	0	0	0	0	0	0;
	916	230.2	-24.09	134.45	-49.76	1.022135	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	917	32	3.93	18.22	-6.73	1.061631	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	926	420	11.01	201.12	-87.96	1.008066	100	1	472	157.33	0	0	0	0	0	0	0	0	0	0	0;
	927	27.4	-2.86	28.69	-8.16	1.067901	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	930	679.6	-118	406.66	-144.75	1.058189	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	972	1221.2	407.52	934.73	-291.52	1.078612	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	974	83.89	18.08	54.36	-20.21	1.033045	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	976	21.3	-5.46	36.02	-35.35	1.024328	100	1	107.83	0	0	0	0	0	0	0	0	0	0	0	0;
	978	14.9	0.14	6.43	-2.99	1.017319	100	1	15.15	0	0	0	0	0	0	0	0	0	0	0	0;
	983	1034.8	-256.65	948.36	-277.27	1.078039	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	984	1029.2	8.1	738.93	-240.84	1.031963	100	1	1600	533.33	0	0	0	0	0	0	0	0	0	0	0;
	991	21.68	2.06	16.83	-8.05	1.001634	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	992	14.96	6.97	25.97	-10.87	1.030255	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	993	-136.9	5.75	17.85	-7.07	1.037248	100	1	100	-182.53	0	0	0	0	0	0	0	0	0	0	0;
	1001	92.1	23.68	55.02	-19.81	1.069874	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1002	171	10.18	89.55	-35.15	1.03837	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	1004	86.4	35.09	43.99	-17.4	0.99944	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	1021	-179.3	8.75	58.34	-16.23	1.012345	100	1	100	-239.07	0	0	0	0	0	0	0	0	0	0	0;
	1029	138.6	5.85	90.64	-31.73	1.037886	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	1030	164.94	9.31	71.66	-27.89	1.007808	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	1036	202.5	0.73	137.39	-46.45	1.037074	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	1043	829.5	49.52	387.88	-163.4	1.027465	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	1050	337.91	1.63	173.5	-71.79	1.055375	100	1	518.72	252.05	0	0	0	0	0	0	0	0	0	0	0;
	1082	364.4	56.84	171.62	-73.5	1.027941	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	1083	91.8	8.04	54.91	-19.88	1.050507	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1088	4.6	1.01	8.06	-3.66	1.025882	100	1	45.99	0	0	0	0	0	0	0	0	0	0	0	0;
	1093	837	97.07	550.69	-185.15	1.024587	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	1097	17.5	-0.91	6.17	-6.64	1.013606	100	1	18.38	0	0	0	0	0	0	0	0	0	0	0	0;
	1098	93.35	-19.73	91.42	-31.24	1.104358	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	1100	58.2	6.13	36.8	-13.01	1.042301	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1102	-124.15	12.14	38.05	-11.59	1.048163	100	1	100	-165.54	0	0	0	0	0	0	0	0	0	0	0;
	1103	405.69	-35.07	376.66	-112.66	1.043526	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	1117	744.1	-245.23	307.47	-487.11	1.009173	100	1	864	288	0	0	0	0	0	0	0	0	0	0	0;
	1119	298.7	-10.62	126.49	-57.33	1.044725	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	1121	289.79	-11.59	158.2	-62.81	1.017488	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	1122	75.2	2.31	68	-56.08	1.043868	100	1	190.28	0	0	0	0	0	0	0	0	0	0	0	0;
	1125	129.6	13.26	114.57	-34.74	1.04319	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	1131	667.5	108.89	407.66	-143.61	1.036981	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	1138	930.3	93.88	432.44	-180.7	1.088765	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	1139	175.48	49.36	107.89	-41.5	1.054186	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	1145	173.47	-1.29	153.07	-144.24	1.040382	100	1	452.53	37.71	0	0	0	0	0	0	0	0	0	0	0;
	1173	-124.7	-0.42	1.85	-0.88	1.018544	100	1	100	-166.27	0	0	0	0	0	0	0	0	0	0	0;
	1174	26.8	5.65	18.61	-6.28	1.002801	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1175	309.54	3.13	108.35	-39.12	1.038334	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	1176	6.7	13.93	94.13	-28.09	1.016193	100	1	100	4.46	0	0	0	0	0	0	0	0	0	0	0;
	1184	329.2	159.58	178.02	-67.55	1.037371	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	1220	107.4	57.33	84.07	-26.48	1.008655	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	1225	282.8	18.73	118.84	-55.64	1.051668	100	1	290.43	0	0	0	0	0	0	0	0	0	0	0	0;
	1237	1504.8	346.47	906.61	-319.44	1.103909	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	1241	46.13	-7.09	22.29	-20.98	1.019582	100	1	65.63	0	0	0	0	0	0	0	0	0	0	0	0;
	1243	81.7	-2.62	44.3	-17.02	1.080005	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	1244	16.2	-0.85	23.86	-28.85	1.026736	100	1	63.79	0	0	0	0	0	0	0	0	0	0	0	0;
	1251	837	90.81	550.69	-185.15	1.024626	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	1258	71	20.62	34.91	-14.83	1.032962	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1267	155.3	6.62	191.38	-52.9	1.04637	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	1271	-100.5	-11	32.54	-32.11	1.0139	100	1	100	-134	0	0	0	0	0	0	0	0	0	0	0;
	1295	-124.1	7.57	36.92	-12.88	1.035594	100	1	100	-165.47	0	0	0	0	0	0	0	0	0	0	0;
	1307	-410.1	8.83	35.48	-14.34	1.062062	100	1	100	-546.8	0	0	0	0	0	0	0	0	0	0	0;
	1313	18.18	4.52	17.51	-7.42	1.031257	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1315	-234.9	12.44	38.26	-11.55	1.091975	100	1	100	-313.2	0	0	0	0	0	0	0	0	0	0	0;
	1320	279.6	57.98	160.62	-60.3	1.01339	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	1337	184.1	2.66	190.67	-54.2	1.05919	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	1341	67.2	26.36	57.03	-17.62	1.035237	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1354	-246.5	15.08	38.08	-11.73	1.034497	100	1	100	-328.67	0	0	0	0	0	0	0	0	0	0	0;
	1361	-83.6	-11.72	91.51	-30.76	1.001499	100	1	100	-111.47	0	0	0	0	0	0	0	0	0	0	0;
	1365	314	47.8	179.34	-66.1	1.045092	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	1371	87.1	-31.41	47.16	-88.84	1.02639	100	1	197.4	0	0	0	0	0	0	0	0	0	0	0	0;
	1378	-30.21	-1.88	4.86	-7.37	1.015715	100	1	100	-40.28	0	0	0	0	0	0	0	0	0	0	0;
	1387	-198.4	-13.94	174.11	-71.24	1.029614	100	1	100	-264.53	0	0	0	0	0	0	0	0	0	0	0;
	1390	299.2	38.81	126.8	-57.1	1.076512	100	1	301.01	100.34	0	0	0	0	0	0	0	0	0	0	0;
	1391	26.2	2.2	18.74	-6.17	1.027192	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1399	-174.6	-3.16	11.64	-10.2	1.04788	100	1	100	-232.8	0	0	0	0	0	0	0	0	0	0	0;
	1402	200.2	-7.59	136.48	-46.67	1.019959	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	1409	628.19	-75.91	338.66	-151.76	1.022303	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	1417	39.99	12.96	38.08	-11.7	1.009035	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1420	399.5	101.83	251.08	-126.77	1.001444	100	1	790	263.33	0	0	0	0	0	0	0	0	0	0	0;
	1422	-101	4.92	18.09	-6.79	1.047835	100	1	100	-134.67	0	0	0	0	0	0	0	0	0	0	0;
	1426	416.5	190.14	220.32	-86.22	1.072946	100	1	500	0	0	0	0	0	0	0	0	0	0	0	0;
	1433	57.2	-11.62	35.74	-50.26	1.01442	100	1	60.06	0	0	0	0	0	0	0	0	0	0	0	0;
	1436	49.4	5.52	37.72	-12.09	1.054052	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1440	115	-7.3	41.52	-57.86	1.011268	100	1	120.75	0	0	0	0	0	0	0	0	0	0	0	0;
	1444	130.32	-74.8	142.87	-273.1	1.023146	100	1	357.63	-86.36	0	0	0	0	0	0	0	0	0	0	0;
	1458	1588	-294.83	895.23	-330.78	1.045766	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	1474	-27	0.2	1.9	-1.04	1.010259	100	1	100	-36	0	0	0	0	0	0	0	0	0	0	0;
	1475	783.9	-95.73	394.45	-157	1.020467	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	1478	62.6	8.63	36.4	-13.44	1.048923	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1481	297.8	-56.44	539.99	-195.85	1.040265	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	1484	288.3	-36.25	127.72	-56.36	1.027385	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	1487	389.03	-5.18	176.62	-72.75	1.055167	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	1492	261.6	3.44	376.31	-114.18	1.02928	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	1498	5.2	-0.57	2.25	-1.06	1.073955	100	1	5.31	0	0	0	0	0	0	0	0	0	0	0	0;
	1507	-99.2	5.83	18.77	-6.16	1.057885	100	1	100	-132.26	0	0	0	0	0	0	0	0	0	0	0;
	1508	149.2	9.16	90.09	-32.49	1.062401	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	1529	-22.7	-0.03	1.88	-1.06	1.01463	100	1	100	-30.27	0	0	0	0	0	0	0	0	0	0	0;
	1546	341.2	25.97	175.3	-69.96	1.042059	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	1550	350	-54.69	174.21	-70.75	1.082656	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	1555	453.4	62.5	376.36	-114.08	1.065019	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	1557	993.3	344.07	478.45	-196.84	1.039811	100	1	1100	366.67	0	0	0	0	0	0	0	0	0	0	0;
	1558	24.1	1.36	7.91	-3.82	1.01534	100	1	47.25	0	0	0	0	0	0	0	0	0	0	0	0;
	1570	121.4	-1.61	93.16	-29.49	1.076277	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	1582	312	0.62	177.98	-67.01	1.042602	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	1589	-31.54	-7.08	5.55	-17	1.023343	100	1	100	-42.05	0	0	0	0	0	0	0	0	0	0	0;
	1604	50.6	13.86	37.66	-12.17	1.037285	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1606	358.2	24.61	174.6	-70.71	1.077423	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	1618	-136.6	-8.43	21.27	-20.07	1.075715	100	1	100	-182.13	0	0	0	0	0	0	0	0	0	0	0;
	1626	22.8	-0.41	17.31	-14.75	1.038045	100	1	28.35	0	0	0	0	0	0	0	0	0	0	0	0;
	1642	50.4	13.84	37.74	-12.11	1.025423	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1680	1250.4	263.29	928.89	-296.1	1.038879	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	1684	599.3	132.44	530.3	-205.67	1.033747	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	1685	62.7	18.2	57.41	-17.22	1.003785	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1697	-189	-10.95	31.94	-31.24	1.036203	100	1	100	-252	0	0	0	0	0	0	0	0	0	0	0;
	1708	860.8	121.45	439.37	-173.3	1.034651	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	1715	25.35	0.82	35.11	-14.76	1.054322	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1717	19.9	1.13	19.25	-5.65	1.022926	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1719	269.17	92.54	378.7	-111.85	1.085028	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	1720	994.5	504.86	534.24	-201.82	1.072226	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	1721	28.1	2.89	18.52	-6.39	1.061454	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1726	48.66	10.82	55.28	-19.44	1.067726	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1736	200	28.4	84.74	-38.13	1.06645	100	1	202.58	67.53	0	0	0	0	0	0	0	0	0	0	0;
	1738	623.6	42.99	267.73	-215.51	1.015757	100	1	720	240	0	0	0	0	0	0	0	0	0	0	0;
	1743	79.6	12.98	55.21	-18.48	1.033701	100	1	125	0	0	0	0	0	0	0	0	0	0	0	0;
	1747	835.5	254.03	387.67	-163.73	1.053359	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	1748	326.8	60.91	176.69	-68.45	1.038042	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	1754	40	3.39	16.99	-7.92	1.037047	100	1	41.06	0	0	0	0	0	0	0	0	0	0	0	0;
	1756	178.8	74.32	87.48	-35.37	1.037533	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	1776	595	-773.19	Inf	-Inf	0.967759	100	1	624.75	0	0	0	0	0	0	0	0	0	0	0	0;
	1785	108.3	23.49	51.66	-22.85	0.998	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1786	1124	-115.84	515.09	-220.89	1.083767	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	1788	430.2	245.8	272.25	-94.92	1.021039	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	1793	-90.3	1.11	34.64	-15.2	1.020092	100	1	100	-120.4	0	0	0	0	0	0	0	0	0	0	0;
	1794	1250.4	239.03	926.35	-297.68	1.04708	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	1797	42.29	11.75	36.93	-12.79	1.005885	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1808	-0.6	-6	111.23	-38.16	1.04521	100	1	100	-0.8	0	0	0	0	0	0	0	0	0	0	0;
	1812	53.6	2.97	12.89	-5.66	1.03994	100	1	70.88	0	0	0	0	0	0	0	0	0	0	0	0;
	1814	311.46	58.05	174.22	-70.91	1.039354	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	1819	246	16.81	132.46	-51.71	1.039397	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	1822	658.2	29.36	356.05	-134.47	1.043084	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	1823	64.8	27.99	36.04	-13.78	1.020728	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1843	29.17	-7.52	58.01	-16.8	1.080669	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1846	117.6	-22.86	40.94	-62.74	1.028122	100	1	123.48	0	0	0	0	0	0	0	0	0	0	0	0;
	1847	40.6	13.24	38.42	-11.36	1.002793	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1850	2.89	-2.53	11.52	-10.09	1.014744	100	1	32.97	0	0	0	0	0	0	0	0	0	0	0	0;
	1851	15.6	0.08	11.76	-8.89	1.053982	100	1	31.92	0	0	0	0	0	0	0	0	0	0	0	0;
	1852	-241.8	16.53	58.16	-16.52	1.054605	100	1	100	-322.4	0	0	0	0	0	0	0	0	0	0	0;
	1871	-51.7	9.04	27.31	-9.56	1.000876	100	1	100	-68.93	0	0	0	0	0	0	0	0	0	0	0;
	1877	99	-12.35	117.14	-32.37	1.050568	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	1888	573.68	36.11	396.02	-155.83	1.034904	100	1	912.78	312.78	0	0	0	0	0	0	0	0	0	0	0;
	1890	1526	133.56	686.73	-294.62	1.04228	100	1	1600	533.33	0	0	0	0	0	0	0	0	0	0	0;
	1892	-12.6	1.19	11.03	-8.79	1.020022	100	1	100	-16.8	0	0	0	0	0	0	0	0	0	0	0;
	1898	246	-21.87	133.11	-51.25	1.074634	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	1900	478.2	-90.11	263.49	-103.77	1.077688	100	1	600	0	0	0	0	0	0	0	0	0	0	0	0;
	1905	313.1	108.79	283.42	-84.45	1.049117	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	1907	76.2	8.8	44.86	-16.43	1.038735	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	1909	-31.18	5.07	38.67	-11.1	1.019677	100	1	100	-41.57	0	0	0	0	0	0	0	0	0	0	0;
	1914	115.5	-39.66	160.36	-150.95	1.038898	100	1	472.16	39.35	0	0	0	0	0	0	0	0	0	0	0;
	1919	411.8	63.9	274.12	-93.01	1.036866	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	1925	-184	-0.69	5.13	-4.7	1.003974	100	1	100	-245.33	0	0	0	0	0	0	0	0	0	0	0;
	1927	349.13	200.69	221.35	-85.36	1.0345	100	1	500	0	0	0	0	0	0	0	0	0	0	0	0;
	1935	-55.4	0.52	31.24	-26.98	1.014926	100	1	100	-73.87	0	0	0	0	0	0	0	0	0	0	0;
	1938	506.1	-36.31	314.85	-480.33	1.038487	100	1	531.56	0	0	0	0	0	0	0	0	0	0	0	0;
	1947	-17.02	-0.22	2.05	-0.97	1.053959	100	1	100	-22.7	0	0	0	0	0	0	0	0	0	0	0;
	1959	31	3.11	18.14	-6.75	1.069553	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1970	90.69	31.2	54.24	-20.63	1.026295	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1982	-2.2	-5.28	20.66	-31.34	1.008845	100	1	100	-2.93	0	0	0	0	0	0	0	0	0	0	0;
	1991	44	7.99	27.19	-9.66	1.022737	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	1999	141.6	8.84	80.51	-30.07	1.076837	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	2017	216	92.42	239.75	-66.86	1.042851	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	2018	42.8	7.03	38.1	-11.62	1.00366	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2030	576	148.48	257	-110.85	1.067534	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	2035	855.6	159.27	441.85	-171.27	1.028122	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	2046	234	6.2	133.49	-50.26	1.042366	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	2050	836.2	-52.53	962.2	-263.65	1.004886	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	2051	71	-33.53	98.71	-92.7	1.047683	100	1	290.29	0	0	0	0	0	0	0	0	0	0	0	0;
	2054	1.8	-0.15	1.37	-1.02	1.055574	100	1	3.7	0	0	0	0	0	0	0	0	0	0	0	0;
	2063	47.6	-8.08	67.03	-30.83	1.027043	100	1	100	31.73	0	0	0	0	0	0	0	0	0	0	0;
	2082	-142.47	-2.35	4.66	-7.7	1.014544	100	1	100	-189.95	0	0	0	0	0	0	0	0	0	0	0;
	2085	176	32.54	88.39	-36.19	1.029587	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2086	64.2	14.61	36.34	-13.54	1.058532	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2091	33.9	1.68	17.88	-7.04	1.02498	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2096	1063.5	-188.42	522.74	-212.37	1.046008	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	2098	178	5.19	86.24	-36.29	1.064541	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	2100	152.1	-1.71	160.69	-140.97	1.022492	100	1	462.73	51.41	0	0	0	0	0	0	0	0	0	0	0;
	2104	-20.9	0.58	4.45	-2.4	1.005555	100	1	100	-27.87	0	0	0	0	0	0	0	0	0	0	0;
	2107	1269.4	568.11	661.9	-257.62	1.040878	100	1	1500	500	0	0	0	0	0	0	0	0	0	0	0;
	2113	379.2	52.49	172.07	-73.29	1.151426	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	2150	498.4	91.07	266.17	-101.5	1.061614	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	2153	-496.6	16.51	137.41	-46.47	1.065758	100	1	100	-662.13	0	0	0	0	0	0	0	0	0	0	0;
	2163	33.09	0.38	17.24	-7.72	1.042967	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2167	95.07	12.28	92.35	-30.11	1.027081	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	2176	1443.55	27.75	38.2	-11.68	1.0196	100	1	2900	900	0	0	0	0	0	0	0	0	0	0	0;
	2177	-121.3	15.15	58.35	-16.26	1.068942	100	1	100	-161.73	0	0	0	0	0	0	0	0	0	0	0;
	2181	502.2	-132.48	198.93	-274.78	1.00154	100	1	534	178	0	0	0	0	0	0	0	0	0	0	0;
	2185	99.3	-5.75	53.94	-20.83	1.056295	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2187	511.9	-56.25	405.25	-145.32	1.036364	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	2193	7.9	0.77	7.47	-2.56	1.009622	100	1	16.12	0	0	0	0	0	0	0	0	0	0	0	0;
	2197	-53.29	1.19	19.51	-14.74	1.031798	100	1	100	-71.05	0	0	0	0	0	0	0	0	0	0	0;
	2206	204.3	26.26	137.29	-46.58	1.018916	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	2207	14.08	9.17	38.96	-10.76	1.006198	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2225	-83.5	-0.44	2.27	-1.04	1.025903	100	1	100	-111.33	0	0	0	0	0	0	0	0	0	0	0;
	2236	-38.45	-0.25	2.23	-1.07	1.020813	100	1	100	-51.26	0	0	0	0	0	0	0	0	0	0	0;
	2237	100	-9.31	94.84	-27.76	1.034951	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2241	1170	-365.53	216.04	-567.96	1.005014	100	1	1208	402.67	0	0	0	0	0	0	0	0	0	0	0;
	2243	-30.6	0.11	4.72	-2.11	1.029391	100	1	100	-40.8	0	0	0	0	0	0	0	0	0	0	0;
	2245	165.1	-63.63	85.19	-130.81	1.010255	100	1	208.47	0	0	0	0	0	0	0	0	0	0	0	0;
	2250	-108.5	9.33	26.07	-10.82	1.024801	100	1	100	-144.67	0	0	0	0	0	0	0	0	0	0	0;
	2253	706.8	277.06	351.3	-139.52	1.147706	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	2266	190	-8.49	101.16	-129.47	1.013425	100	1	199.5	0	0	0	0	0	0	0	0	0	0	0	0;
	2267	362	138.59	228.16	-78.83	1.003207	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	2276	38.9	-19.68	41.67	-36.5	1.062473	100	1	119.3	0	0	0	0	0	0	0	0	0	0	0	0;
	2278	140	-15.33	53.35	-86.65	1.030091	100	1	147	0	0	0	0	0	0	0	0	0	0	0	0;
	2284	-59.5	6.01	38.48	-11.35	1.030581	100	1	100	-79.33	0	0	0	0	0	0	0	0	0	0	0;
	2287	148.8	-3.23	56.21	-95.79	1.034619	100	1	168	0	0	0	0	0	0	0	0	0	0	0	0;
	2289	36.9	-4.69	15.59	-20.41	1.036649	100	1	38.74	0	0	0	0	0	0	0	0	0	0	0	0;
	2291	110.5	11.33	95.69	-28.93	1.063998	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2310	164.5	2.75	141.39	-42.37	1.013414	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	2320	82.2	11.83	54.86	-18.8	1.025327	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2335	431.2	215.21	273.49	-94.1	1.044074	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	2351	-35.89	6.16	47.28	-13.91	1.019563	100	1	100	-47.86	0	0	0	0	0	0	0	0	0	0	0;
	2353	-113.5	-3.43	30.35	-35.67	1.010677	100	1	100	-151.33	0	0	0	0	0	0	0	0	0	0	0;
	2357	-184.1	8.93	38.31	-11.47	1.072612	100	1	100	-245.47	0	0	0	0	0	0	0	0	0	0	0;
	2359	78.2	8.57	34.35	-15.5	1.021762	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2362	662.2	-71.97	351.42	-138.02	1.038978	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	2369	129.4	-0.75	92.72	-30.04	1.055918	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	2394	200.3	150.64	380.31	-110.08	1.065017	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	2398	473.42	293.64	368.84	-121.98	1.147454	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	2402	179.5	26.25	88.7	-36.06	1.076661	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2412	67.3	-95.99	134.54	-136.29	1.013238	100	1	409.09	0	0	0	0	0	0	0	0	0	0	0	0;
	2413	51.3	80.69	90.68	-31.93	0.996424	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2421	-320.4	12.79	37.39	-12.45	1.043537	100	1	100	-427.2	0	0	0	0	0	0	0	0	0	0	0;
	2425	-100.5	15.1	54.3	-20.49	1.037815	100	1	100	-134	0	0	0	0	0	0	0	0	0	0	0;
	2426	-43.2	-29	105.34	-99.02	1.045936	100	1	100	-57.6	0	0	0	0	0	0	0	0	0	0	0;
	2429	370.6	193.25	214.1	-92.65	1.063143	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	2446	1367	257.39	921.62	-304.73	1.048971	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	2447	-14.5	1.59	49.04	-32.31	1.031874	100	1	100	-19.33	0	0	0	0	0	0	0	0	0	0	0;
	2452	103	-19.92	47.33	-61.69	1.026174	100	1	108.15	0	0	0	0	0	0	0	0	0	0	0	0;
	2460	37.79	17.77	38.01	-11.74	0.979599	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2468	46.8	5.87	38.02	-11.81	1.046839	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2481	31.8	12.06	39.12	-10.71	1.045831	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2489	286.5	-19.35	162.4	-61.79	1.046032	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	2501	76.8	7.72	34.56	-15.29	1.044679	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2509	-92.5	5.49	38.46	-11.36	1.025789	100	1	100	-123.33	0	0	0	0	0	0	0	0	0	0	0;
	2515	27.7	6.44	18.43	-6.44	0.998497	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2517	3.85	-9.83	78.13	-77.07	1.013559	100	1	234.52	0	0	0	0	0	0	0	0	0	0	0	0;
	2519	-48	5.16	27.24	-9.55	1.028396	100	1	100	-64	0	0	0	0	0	0	0	0	0	0	0;
	2534	179	27.06	86.23	-36.33	1.066932	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	2539	-122.4	9.3	25.99	-10.9	1.024702	100	1	100	-163.2	0	0	0	0	0	0	0	0	0	0	0;
	2541	123.2	3.39	73.34	-26.44	1.02273	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	2547	-16.7	-0.07	2.6	-2.32	1.014326	100	1	100	-22.27	0	0	0	0	0	0	0	0	0	0	0;
	2550	54.3	5.24	58.23	-16.49	1.047876	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2568	80.91	61.89	91.25	-33.52	1.020248	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2569	-17.85	3.12	18.86	-6.03	1.027206	100	1	100	-23.8	0	0	0	0	0	0	0	0	0	0	0;
	2572	411.2	80.78	276.19	-91.63	1.044176	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	2588	-196.1	3	17.06	-7.84	1.028186	100	1	100	-261.47	0	0	0	0	0	0	0	0	0	0	0;
	2589	567	-54.77	250.66	-116.19	1.02499	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	2594	127.2	2.04	81.79	-28.68	1.058099	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	2595	994.2	511.43	533.87	-202.07	1.073729	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	2600	76.48	16.03	76.08	-23.59	1.047035	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	2604	-196.5	6.1	17.14	-7.73	1.031949	100	1	100	-262	0	0	0	0	0	0	0	0	0	0	0;
	2606	291.9	110.24	128.02	-56.23	1.041601	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	2620	201	38.52	107.62	-41.92	1.030287	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	2624	53.4	21.86	58.27	-16.44	1.037675	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2627	1367	260.89	921.52	-304.79	1.048065	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	2642	1897	-225.17	848.76	-377.67	1.052836	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	2649	156.9	17.66	141.85	-42.12	1.087006	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	2650	116.6	5.92	50.85	-22.91	1.036234	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2653	41.92	11.13	51.94	-22.8	1.032593	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2658	308.31	-16.39	124.68	-59.39	1.020867	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	2660	144.9	13.53	79.69	-30.77	1.011944	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	2662	-27.87	5.55	27.99	-8.81	1.014043	100	1	100	-37.16	0	0	0	0	0	0	0	0	0	0	0;
	2664	371.59	162.21	171.72	-73.55	1.049614	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	2685	157.2	48.72	78.09	-32.38	1.010899	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	2686	-72.9	-40.28	69.01	-69.9	1.016926	100	1	100	-97.2	0	0	0	0	0	0	0	0	0	0	0;
	2710	768.7	279.5	665.77	-243	1.035068	100	1	1480	133.33	0	0	0	0	0	0	0	0	0	0	0;
	2719	77.6	-19.13	70.15	-58.1	1.041789	100	1	196.72	0	0	0	0	0	0	0	0	0	0	0	0;
	2725	105.45	-10.6	69.22	-30.35	1.014934	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	2729	38.47	14.26	57.41	-17.32	1.050822	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2736	204.3	-30.6	137.02	-46.75	1.032664	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	2743	69.2	5.17	35.58	-14.26	1.039467	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2758	608.9	-54.79	748.51	-231.56	1.050098	100	1	1600	466.67	0	0	0	0	0	0	0	0	0	0	0;
	2764	575.4	-114.16	416.7	-134.5	1.045688	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	2775	660	37.02	313.82	-111.37	1.024728	100	1	744	248	0	0	0	0	0	0	0	0	0	0	0;
	2786	1221.2	413.65	934.81	-291.47	1.078984	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	2791	-4.7	-0.61	11.28	-11.47	1.015494	100	1	100	-6.27	0	0	0	0	0	0	0	0	0	0	0;
	2797	-95.1	-2.35	88.33	-36.33	1.041702	100	1	100	-126.8	0	0	0	0	0	0	0	0	0	0	0;
	2799	66.8	0.9	77.91	-21.66	1.059752	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	2800	1048.2	-86.22	525.34	-209.84	1.02478	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	2810	283.5	-18.98	90.6	-146.96	1.036015	100	1	297.68	0	0	0	0	0	0	0	0	0	0	0	0;
	2816	1250.4	281.83	926.26	-297.74	1.049577	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	2817	71.8	6.19	34.8	-14.94	0.987457	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2822	86.4	8.74	54.62	-19.11	1.020284	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2826	160	-24.38	57.74	-94.26	1.022551	100	1	168	0	0	0	0	0	0	0	0	0	0	0	0;
	2835	756.52	-21.62	551.02	-184.74	1.043216	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	2839	29.2	4.72	28.38	-8.38	1.03617	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	2841	80.7	-0.53	55.98	-18.76	1.080183	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2842	31.6	5.81	18.19	-6.74	1.037135	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2852	138.4	-21.04	91.2	-31.35	1.063812	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2853	126	12.34	72.02	-27.47	0.998154	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	2862	-109.6	7.54	37.94	-11.89	1.027949	100	1	100	-146.13	0	0	0	0	0	0	0	0	0	0	0;
	2867	-120.1	-19.34	110.29	-93.13	1.047385	100	1	100	-160.13	0	0	0	0	0	0	0	0	0	0	0;
	2872	4.5	-0.7	3.31	-2.54	1.054344	100	1	9.05	0	0	0	0	0	0	0	0	0	0	0	0;
	2874	-144.07	-1.31	21.01	-19.61	1.006921	100	1	100	-192.1	0	0	0	0	0	0	0	0	0	0	0;
	2878	8.95	-11.02	72.95	-68.71	1.047923	100	1	214.86	0	0	0	0	0	0	0	0	0	0	0	0;
	2881	131.9	-17.39	134.18	-78.85	1.024879	100	1	186.98	0	0	0	0	0	0	0	0	0	0	0	0;
	2886	60.9	11.04	57.69	-17.02	1.057004	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2896	-223.4	-0.12	1.52	-0.73	1.009058	100	1	100	-297.87	0	0	0	0	0	0	0	0	0	0	0;
	2897	99.6	-13.74	94.68	-27.87	1.079272	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	2902	20.9	2.79	19.19	-5.72	1.069	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2920	658.5	201.52	563.93	-170.68	1.036969	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	2923	962.1	207.61	536.84	-198.88	1.040379	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	2930	53.2	5.05	37.24	-12.54	1.054065	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2934	3.4	-0.64	2.51	-1.91	1.049297	100	1	6.83	0	0	0	0	0	0	0	0	0	0	0	0;
	2942	313.1	108.89	283.42	-84.45	1.049098	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	2946	-222.1	-0.41	1.83	-0.87	1.012911	100	1	100	-296.13	0	0	0	0	0	0	0	0	0	0	0;
	2956	362.8	80.52	172.53	-72.75	1.032523	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	2973	930.1	95.59	432.47	-180.67	1.088757	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	2975	-23.66	5.96	18.35	-6.58	1.03823	100	1	100	-31.55	0	0	0	0	0	0	0	0	0	0	0;
	2982	79.4	17.7	34.04	-15.78	1.020472	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2985	46.6	-71.83	184.46	-132.67	1.046452	100	1	493.35	0	0	0	0	0	0	0	0	0	0	0	0;
	2987	296.8	0.48	139.64	-59.75	1.060068	100	1	320	0	0	0	0	0	0	0	0	0	0	0	0;
	2989	-542	13.11	55.22	-19.48	1.066576	100	1	100	-722.67	0	0	0	0	0	0	0	0	0	0	0;
	2992	200.1	33.77	189.58	-55.33	1.104121	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	2997	2334.5	109.9	1434.05	-527.42	1.089588	100	1	3200	1066.67	0	0	0	0	0	0	0	0	0	0	0;
	3002	-18	-3.78	21.56	-26.44	1.015981	100	1	100	-24	0	0	0	0	0	0	0	0	0	0	0;
	3003	176.9	-24.23	370.58	-119.77	1.048371	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	3009	-103	9.04	47.72	-13.5	1.016186	100	1	100	-137.33	0	0	0	0	0	0	0	0	0	0	0;
	3018	162.5	-6.58	89.61	-34.82	1.049496	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	3028	855.7	156.32	441.85	-171.27	1.02789	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	3029	20.87	8.87	37.49	-12.24	1.011669	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3035	676.2	339.33	564.62	-170.92	1.0405	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	3038	292.2	105.65	127.91	-56.34	1.041581	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	3039	-59.31	2.38	17.43	-7.49	1.027247	100	1	100	-79.08	0	0	0	0	0	0	0	0	0	0	0;
	3045	-58.4	-2.78	23.05	-26.17	1.015762	100	1	100	-77.87	0	0	0	0	0	0	0	0	0	0	0;
	3054	210.4	123.15	167.72	-77.11	0.995142	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	3064	224	-37.31	137.76	-138.96	1.022454	100	1	384	128	0	0	0	0	0	0	0	0	0	0	0;
	3078	340.5	79.13	228.19	-78.06	1.024108	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	3084	-82.33	0.43	188.93	-56.32	1.042602	100	1	100	-109.77	0	0	0	0	0	0	0	0	0	0	0;
	3088	203.4	18.79	106.96	-42.49	1.043452	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	3090	47	27.75	38.06	-11.8	1.003861	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3091	31.4	-5.7	15.99	-26.25	1.021666	100	1	33.03	0	0	0	0	0	0	0	0	0	0	0	0;
	3096	-59.3	-7.67	21.04	-36.55	1.008864	100	1	100	-79.07	0	0	0	0	0	0	0	0	0	0	0;
	3113	-20.3	-8.58	22.39	-20.99	1.064513	100	1	100	-27.07	0	0	0	0	0	0	0	0	0	0	0;
	3114	38.6	3.07	17.33	-7.61	1.063117	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	3127	580.7	-110.71	250.27	-116.97	1.097365	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	3130	357.4	-13.9	278.16	-88.47	1.038694	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	3133	85.54	0.61	113.97	-35.56	1.05743	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	3134	-136.3	2.44	70.25	-29.33	1.034426	100	1	100	-181.73	0	0	0	0	0	0	0	0	0	0	0;
	3135	54.6	6.61	161.69	-103.11	1.03216	100	1	198.08	0	0	0	0	0	0	0	0	0	0	0	0;
	3147	84.2	8.22	187.35	-57.76	1.063849	100	1	200	56.13	0	0	0	0	0	0	0	0	0	0	0;
	3150	-31.34	-0.52	6.66	-5.1	1.010317	100	1	100	-41.79	0	0	0	0	0	0	0	0	0	0	0;
	3152	94.5	12.28	54.68	-20.13	1.069675	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3156	-63.2	-7.96	21.53	-32.38	1.024848	100	1	100	-84.27	0	0	0	0	0	0	0	0	0	0	0;
	3157	9	-1.18	6.78	-5.13	1.021742	100	1	18.41	0	0	0	0	0	0	0	0	0	0	0	0;
	3158	-23.7	-25.2	138.81	-44.5	1.058123	100	1	100	-31.6	0	0	0	0	0	0	0	0	0	0	0;
	3163	33.9	-26.5	67.92	-69.02	1.002654	100	1	206.86	0	0	0	0	0	0	0	0	0	0	0	0;
	3183	98.7	-1.06	53.7	-20.98	1.061538	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3189	77.1	-3.53	44.68	-16.58	1.027406	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	3198	144	59.93	113.55	-35.86	1.008016	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	3205	416.7	74.13	243.61	-88.46	1.043987	100	1	540	0	0	0	0	0	0	0	0	0	0	0	0;
	3218	-107.4	7.74	36.79	-13	1.02981	100	1	100	-143.2	0	0	0	0	0	0	0	0	0	0	0;
	3219	-36.5	-22.62	180.55	-64.63	1.06035	100	1	100	-48.67	0	0	0	0	0	0	0	0	0	0	0;
	3227	360	3.76	173.96	-71.56	1.051132	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	3229	323.2	44.1	177.61	-67.66	1.034116	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	3233	13	6.69	178.47	-79.11	1.006814	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	3234	108.24	19.49	56.14	-17.53	1.034756	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3236	359.3	64.06	173.85	-71.31	1.022886	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	3240	-42.3	-21.97	62.23	-51.37	1.048993	100	1	100	-56.4	0	0	0	0	0	0	0	0	0	0	0;
	3251	285.6	104.22	378.64	-111.91	1.085025	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	3253	123.1	-16.44	64.47	-72.73	1.024777	100	1	129.25	0	0	0	0	0	0	0	0	0	0	0	0;
	3263	30.39	2.91	38.65	-11.2	1.036588	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3264	37.8	12.1	17.36	-7.56	1.039352	100	1	45	0	0	0	0	0	0	0	0	0	0	0	0;
	3278	46.73	21.66	52.31	-22.52	1.024557	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3284	130.2	34.9	81.97	-28.66	1.031123	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	3285	654.8	92.94	356.44	-134.09	1.04544	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	3290	176.7	15.21	76.54	-34.2	1.076876	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	3291	326.4	-20.22	229	-77.02	1.020441	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	3292	309	-101.33	115.17	-182.61	0.995766	100	1	320	106.67	0	0	0	0	0	0	0	0	0	0	0;
	3298	206.48	-21.12	103.58	-45.89	1.06827	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	3304	334.5	-47.86	83.58	-174.42	1.024949	100	1	414	138	0	0	0	0	0	0	0	0	0	0	0;
	3306	861.3	141.54	439.2	-173.43	1.065504	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	3310	29.8	-1.53	22.58	-16.94	1.032355	100	1	61.06	0	0	0	0	0	0	0	0	0	0	0	0;
	3322	-25.59	9.63	46.86	-14.26	1.029763	100	1	100	-34.12	0	0	0	0	0	0	0	0	0	0	0;
	3323	-12.51	-2.29	4.45	-7.55	1.021399	100	1	100	-16.68	0	0	0	0	0	0	0	0	0	0	0;
	3324	-26.48	15.21	38.13	-11.64	1.012342	100	1	100	-35.3	0	0	0	0	0	0	0	0	0	0	0;
	3326	441.18	1.6	564.44	-170.47	1.057931	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	3330	44.12	6.53	93.36	-29.11	1.049026	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	3334	-45.5	2.75	34.41	-15.34	1.027102	100	1	100	-60.67	0	0	0	0	0	0	0	0	0	0	0;
	3335	-600	1825.38	Inf	-Inf	1.08064	100	1	630	-630	0	0	0	0	0	0	0	0	0	0	0;
	3336	-42.3	-0.32	4.85	-4.13	1.010354	100	1	100	-56.4	0	0	0	0	0	0	0	0	0	0	0;
	3340	-287.7	16.95	55	-19.81	1.027329	100	1	100	-383.6	0	0	0	0	0	0	0	0	0	0	0;
	3345	-111.9	-1.83	39.1	-47.42	0.998722	100	1	100	-149.2	0	0	0	0	0	0	0	0	0	0	0;
	3346	56.8	11.85	37.1	-12.76	1.014446	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3349	136.8	11.71	91.48	-31.11	1.047992	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	3353	19.07	6.86	38.2	-11.61	1.045086	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3362	59.2	13.88	36.83	-13.03	1.060735	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3364	44.5	-19.63	47.19	-41.44	1.059596	100	1	135.27	0	0	0	0	0	0	0	0	0	0	0	0;
	3371	61.8	8.32	36.31	-13.47	1.029674	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3375	25.6	-9.24	175.66	-69.57	1.035048	100	1	100	17.07	0	0	0	0	0	0	0	0	0	0	0;
	3389	78.6	6.14	56.35	-18.44	1.037252	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3390	2.4	18.47	54.59	-20.19	1.035344	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3396	249.1	-44.22	131.75	-52.04	1.095689	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	3411	-123.7	8.93	25.57	-11.26	1.024699	100	1	100	-164.93	0	0	0	0	0	0	0	0	0	0	0;
	3416	171.3	38.36	76.43	-34.12	0.99364	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	3422	53.7	1.18	58.19	-16.5	1.068276	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3426	70	-8.13	44.78	-57.03	1.01782	100	1	73.42	0	0	0	0	0	0	0	0	0	0	0	0;
	3431	127.6	7.58	92.26	-30.26	1.037857	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	3436	521	242.21	550.97	-184.89	1.052704	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	3438	-87.5	7.68	37.4	-12.42	1.048328	100	1	100	-116.67	0	0	0	0	0	0	0	0	0	0	0;
	3441	159.2	-12.64	89.45	-33.28	1.059719	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	3480	239.6	-31.21	185.19	-59.55	1.039565	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	3492	-370.8	7.94	53.75	-21.08	1.049517	100	1	100	-494.4	0	0	0	0	0	0	0	0	0	0	0;
	3507	109.8	17.58	51.58	-22.13	1.009668	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3508	96.9	3.27	54.09	-20.64	1.05417	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3511	522.8	-5.35	366.74	-122.45	1.086282	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	3513	829.8	50.36	387.88	-163.4	1.027289	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	3515	-85.4	9.97	38.54	-11.22	1.053923	100	1	100	-113.87	0	0	0	0	0	0	0	0	0	0	0;
	3518	808.66	-125.32	433.83	-177.9	1.006674	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	3547	140	10.96	91.06	-31.5	1.05946	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	3550	341.1	59.49	157.15	-67.4	1.042925	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	3551	-37.1	-18.62	175.51	-69.76	1.039557	100	1	100	-49.47	0	0	0	0	0	0	0	0	0	0	0;
	3565	15.88	4.48	38.83	-10.97	1.079183	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3570	298.9	81.81	124.15	-59.21	1.062522	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	3571	233.6	56.61	100.91	-46.43	0.997978	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	3580	539	182.86	261.7	-107.36	1.071709	100	1	600	0	0	0	0	0	0	0	0	0	0	0	0;
	3582	-321.8	-33.51	52.83	-52.04	1.040375	100	1	100	-429.07	0	0	0	0	0	0	0	0	0	0	0;
	3583	36.3	8.71	17.7	-7.25	1.041006	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	3584	-88.4	-1.89	19.8	-15.12	1.016191	100	1	100	-117.87	0	0	0	0	0	0	0	0	0	0	0;
	3585	50.8	17.16	37.43	-12.32	1.001933	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3590	-31	0.35	2.66	-1.25	1.014988	100	1	100	-41.33	0	0	0	0	0	0	0	0	0	0	0;
	3599	-85.6	7.19	38.12	-11.64	1.035573	100	1	100	-114.13	0	0	0	0	0	0	0	0	0	0	0;
	3605	158.7	-3.06	78.6	-32.04	1.037656	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	3611	-125.4	13.66	54.59	-20.12	1.0427	100	1	100	-167.2	0	0	0	0	0	0	0	0	0	0	0;
	3618	54.49	4.99	36.22	-13.68	1.059129	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3621	202.03	-30.52	130.01	-54.04	1.018584	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	3625	150	-21.2	58.77	-75.06	1.035196	100	1	157.5	0	0	0	0	0	0	0	0	0	0	0	0;
	3628	490	81.97	253.06	-125.28	1.011525	100	1	790	263.33	0	0	0	0	0	0	0	0	0	0	0;
	3629	54.97	2.84	76.85	-22.64	1.019084	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	3634	412	-42.4	217.44	-88.2	1.058312	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	3635	78.2	-13.76	22.56	-41.44	1.027369	100	1	82.11	0	0	0	0	0	0	0	0	0	0	0	0;
	3638	214.1	-16.86	153.73	-109.03	1.027553	100	1	335.58	0	0	0	0	0	0	0	0	0	0	0	0;
	3644	16.4	-5.19	22.35	-21.22	1.01566	100	1	66.09	0	0	0	0	0	0	0	0	0	0	0	0;
	3656	-26.2	18.12	116.07	-33.32	1.062208	100	1	100	-34.93	0	0	0	0	0	0	0	0	0	0	0;
	3659	1097.4	790.96	943.64	-282.01	1.057396	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	3660	-128.54	-15.82	35.56	-48.53	1.032448	100	1	100	-171.38	0	0	0	0	0	0	0	0	0	0	0;
	3661	861.2	134.9	441.04	-172.05	1.055245	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	3673	424.5	-77.32	257.61	-110.43	1.011536	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	3689	5.2	-0.19	2.18	-1.05	1.029361	100	1	6.2	0	0	0	0	0	0	0	0	0	0	0	0;
	3691	239.6	48.56	133.84	-49.89	1.030004	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	3692	164.4	-45.72	180.11	-77.84	1.024383	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	3696	662.8	86.36	354.58	-135.7	1.041264	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	3698	-12.58	7.25	36.34	-13.47	1.039409	100	1	100	-16.78	0	0	0	0	0	0	0	0	0	0	0;
	3711	40.4	13.02	38.36	-11.38	0.999169	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3720	-75.4	4.94	18.4	-6.49	1.050724	100	1	100	-100.53	0	0	0	0	0	0	0	0	0	0	0;
	3723	1146.3	50.9	514.15	-221.73	1.048349	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	3724	94.9	28.33	42.94	-18.47	1.020045	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	3739	34.1	2.85	38.9	-10.9	1.035011	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3741	854.7	211.29	548.57	-187.16	1.029075	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	3743	-41.48	5.47	28.38	-8.4	1.026432	100	1	100	-55.31	0	0	0	0	0	0	0	0	0	0	0;
	3746	10.66	4.33	38.06	-11.72	1.035948	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3750	169.92	90.72	148.02	-51.7	1.066155	100	1	320	0	0	0	0	0	0	0	0	0	0	0	0;
	3753	572.4	226.8	255.79	-112.33	1.055722	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	3755	-141.7	-0.77	32.42	-42.82	0.993935	100	1	100	-188.93	0	0	0	0	0	0	0	0	0	0	0;
	3759	147.6	-17.08	90.31	-32.27	1.031969	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	3766	16.7	-5.93	32.06	-37.75	1.0335	100	1	66.78	0	0	0	0	0	0	0	0	0	0	0	0;
	3769	162.4	22.12	110.34	-37.12	1.021894	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	3782	-126.9	6.37	17.76	-7.18	1.02915	100	1	100	-169.2	0	0	0	0	0	0	0	0	0	0	0;
	3791	90.6	18.83	54.7	-19.97	0.995335	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3792	85.4	-23.53	108.06	-41.25	1.052349	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	3797	-204.2	11.12	35.41	-14.37	1.060935	100	1	100	-272.27	0	0	0	0	0	0	0	0	0	0	0;
	3800	71.9	-11.13	35.4	-48.46	1.036528	100	1	76.97	0	0	0	0	0	0	0	0	0	0	0	0;
	3808	615	-57.36	467.02	-146.12	1.073884	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	3809	257	132.48	313.01	-116.73	1.081564	100	1	700	233.33	0	0	0	0	0	0	0	0	0	0	0;
	3810	89.6	-0.62	94.81	-27.36	1.038535	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	3821	292	-24.15	165.76	-79.17	1.022257	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	3825	837.3	55.38	388.17	-163.4	1.024468	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	3840	-22.95	-13.55	36.29	-35.42	1.027073	100	1	100	-30.6	0	0	0	0	0	0	0	0	0	0	0;
	3841	143.4	-7.13	79.52	-30.85	1.03976	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	3850	239.6	-30.79	185.19	-59.55	1.039563	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	3851	-41.48	4.66	28.23	-8.55	1.022667	100	1	100	-55.31	0	0	0	0	0	0	0	0	0	0	0;
	3853	287	-20.4	182.73	-62.54	1.080197	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	3869	855.6	159.36	440.69	-172.14	1.028967	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	3870	134.4	-19.43	85.5	-135.09	1.023622	100	1	154.56	0	0	0	0	0	0	0	0	0	0	0	0;
	3876	-25.1	10.33	34.27	-15.52	1.04066	100	1	100	-33.47	0	0	0	0	0	0	0	0	0	0	0;
	3879	140.69	-5.19	71.02	-28.68	1.042235	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	3888	70.2	11.45	56.84	-17.84	1.064956	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3900	99.6	21.01	54.19	-20.65	1.061016	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3907	319.6	9.5	152.36	-110.07	1.025805	100	1	360	120	0	0	0	0	0	0	0	0	0	0	0;
	3908	4.69	0.16	4.25	-2.02	0.982046	100	1	10.08	0	0	0	0	0	0	0	0	0	0	0	0;
	3910	362	149.08	173.84	-71.4	1.069291	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	3916	-21.27	14.73	38.02	-11.77	1.039177	100	1	100	-28.37	0	0	0	0	0	0	0	0	0	0	0;
	3920	349.3	-8.65	175.53	-69.73	1.064621	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	3922	571.2	227.31	256.11	-112.06	1.07194	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	3938	-43.8	4.41	25.39	-11.45	1.049226	100	1	100	-58.4	0	0	0	0	0	0	0	0	0	0	0;
	3940	660.6	-113.14	450.57	-161.53	1.018221	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	3948	-23.68	0.49	26.31	-10.52	1.03639	100	1	100	-31.58	0	0	0	0	0	0	0	0	0	0	0;
	3949	124	19.51	94.12	-30.35	1.052164	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	3951	-27.1	-70.83	126.16	-127.56	1.03485	100	1	100	-36.13	0	0	0	0	0	0	0	0	0	0	0;
	3958	79.2	14.63	55.3	-18.4	1.034533	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3959	56.4	-6.65	20	-28.58	1.052799	100	1	59.25	0	0	0	0	0	0	0	0	0	0	0	0;
	3963	207	51.71	177.09	-68.41	1.020596	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	3971	-134.5	15.26	55.67	-19.09	1.053061	100	1	100	-179.33	0	0	0	0	0	0	0	0	0	0	0;
	3979	304	-4.57	120.7	-206.8	1.027309	100	1	319.2	0	0	0	0	0	0	0	0	0	0	0	0;
	3983	229.9	24.94	134.98	-48.75	1.014828	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	3987	44.6	8.46	38.05	-11.72	1.019043	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3991	-210.89	11.91	37.63	-12.22	1.035055	100	1	100	-281.19	0	0	0	0	0	0	0	0	0	0	0;
	3996	-121.2	4.16	18.62	-6.31	1.034652	100	1	100	-161.6	0	0	0	0	0	0	0	0	0	0	0;
	4009	66.47	10.91	54.88	-19.8	1.043192	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4012	46	9.06	37.87	-11.87	1.04188	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4016	109.2	-2.07	52.25	-22.44	1.022786	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4024	89.5	41.31	97.07	-27.44	1.029565	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	4026	399.3	-27.46	223.07	-83.7	1.05562	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	4028	-63.7	4.77	18.44	-6.51	1.022221	100	1	100	-84.94	0	0	0	0	0	0	0	0	0	0	0;
	4046	323.1	-15.89	178.68	-66.57	1.071125	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	4052	68.2	12.09	35.73	-14.12	1.059535	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4056	-76.9	7.64	38.05	-11.77	1.038451	100	1	100	-102.53	0	0	0	0	0	0	0	0	0	0	0;
	4071	100	-10.38	38.72	-50.47	1.026598	100	1	105	0	0	0	0	0	0	0	0	0	0	0	0;
	4075	79.6	12.92	55.21	-18.48	1.033829	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4076	162.8	2.33	88.51	-34.08	1.042958	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	4084	366.1	44.03	264.94	-103.74	1.053107	100	1	600	0	0	0	0	0	0	0	0	0	0	0	0;
	4085	188	67	85.83	-36.92	1.028863	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	4090	-107.2	-11	38.67	-58.85	1.010258	100	1	100	-142.93	0	0	0	0	0	0	0	0	0	0	0;
	4094	165.2	33.4	223.91	-82.75	1.026088	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	4106	57	-9.11	30.6	-36.62	1.022601	100	1	64.26	0	0	0	0	0	0	0	0	0	0	0	0;
	4118	41	6.87	38.37	-11.4	1.025736	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4122	216	-30.79	134.14	-49.12	1.016568	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	4125	59.6	15.79	36.89	-13	1.002666	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4128	67.5	21.6	57.03	-17.63	1.027647	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4131	229.26	80.13	184.19	-60.96	1.034065	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	4163	191.97	-9.59	188.07	-56.82	1.047216	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	4172	26.2	9.34	18.58	-6.28	1.02637	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	4175	1530.2	-301.1	904.67	-321.61	1.078047	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	4178	514.6	24.43	264	-103.61	1.031771	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	4179	95.1	12.7	54.59	-20.21	1.010218	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4182	710.7	-13.98	344.9	-146.06	1.011247	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	4183	-2.55	9.11	38.27	-11.43	1.031152	100	1	100	-3.4	0	0	0	0	0	0	0	0	0	0	0;
	4192	767.4	288.6	396.38	-154.99	1.076318	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	4195	-488.9	177.35	465.42	-147.31	1.076653	100	1	100	-651.87	0	0	0	0	0	0	0	0	0	0	0;
	4198	87.5	7.35	64.15	-56.21	1.01015	100	1	91.83	0	0	0	0	0	0	0	0	0	0	0	0;
	4200	260.1	-13.03	131.25	-52.45	1.016165	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	4204	-48.16	-0.17	1.83	-0.85	1.014828	100	1	100	-64.21	0	0	0	0	0	0	0	0	0	0	0;
	4207	423.93	23.68	373.46	-118.66	1.009702	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	4210	141.89	-13.69	71.61	-28.24	1.031932	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	4211	16	5.84	19.55	-5.36	1.042563	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	4221	110.4	-6.82	72.12	-76.5	1.036578	100	1	203.75	0	0	0	0	0	0	0	0	0	0	0	0;
	4228	-73.1	1.66	47.62	-13.63	1.060932	100	1	100	-97.47	0	0	0	0	0	0	0	0	0	0	0;
	4231	2641.24	604.18	Inf	-Inf	1.042866	100	1	4188.95	1333.33	0	0	0	0	0	0	0	0	0	0	0;
	4244	354.7	-13.59	354.48	-135.04	1.039687	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	4259	28	6.43	18.41	-6.45	1.01039	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	4278	153.8	-16.21	75.89	-136.31	1.038839	100	1	167.36	0	0	0	0	0	0	0	0	0	0	0	0;
	4279	515.39	-60.67	261.6	-106.39	1.01498	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	4291	251	-30.24	196.08	-61.88	1.019753	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	4296	33.5	6.93	18.02	-6.93	1.040271	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	4322	108.74	-12.84	141.75	-41.97	1.055269	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	4323	212.8	74.9	149.98	-49.47	1.038456	100	1	320	0	0	0	0	0	0	0	0	0	0	0	0;
	4330	1176	113.62	723.46	-256.99	1.049093	100	1	1600	533.33	0	0	0	0	0	0	0	0	0	0	0;
	4331	60.2	6.72	36.73	-13.12	1.029918	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4333	152.4	7.48	79.1	-31.46	0.997274	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	4334	73.3	20.25	45.09	-16.17	1.057462	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	4337	490.5	542.31	576.78	-158.19	1.026934	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	4338	-52.5	2.88	16.93	-7.97	1.021243	100	1	100	-70	0	0	0	0	0	0	0	0	0	0	0;
	4350	167	-16.07	87.37	-35.06	1.028107	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	4375	59.2	24.66	36.93	-12.95	1.045136	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4376	273.6	-12.75	129.29	-54.38	1.018032	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	4393	115.36	-17.2	142	-41.87	1.061357	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	4395	-129.1	12.5	38.48	-11.33	1.036974	100	1	100	-172.13	0	0	0	0	0	0	0	0	0	0	0;
	4396	147.6	15.97	142.82	-41.02	1.08226	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	4398	236	1.19	134.39	-49.99	1.046214	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	4399	-9.15	4.89	18.1	-6.81	1.054017	100	1	100	-12.19	0	0	0	0	0	0	0	0	0	0	0;
	4400	358.2	24.67	174.6	-70.71	1.077395	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	4419	861.3	136.32	441.04	-172.05	1.055121	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	4420	84	2.27	51.76	-57.39	1.009808	100	1	88.2	0	0	0	0	0	0	0	0	0	0	0	0;
	4428	14.7	-4.43	20.49	-19.24	1.016576	100	1	60.24	0	0	0	0	0	0	0	0	0	0	0	0;
	4429	5	9.69	24.74	-4.75	1.007433	100	1	46.2	0	0	0	0	0	0	0	0	0	0	0	0;
	4440	-50.6	7.72	94.73	-27.43	1.012622	100	1	100	-67.47	0	0	0	0	0	0	0	0	0	0	0;
	4442	-11.23	-10.66	34.61	-33.92	1.031243	100	1	100	-14.98	0	0	0	0	0	0	0	0	0	0	0;
	4445	252.6	155.85	269.01	-234.65	1.031204	100	1	772.54	85.84	0	0	0	0	0	0	0	0	0	0	0;
	4447	261.77	82.45	182.87	-62.52	1.115565	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	4455	742.6	68.51	204.12	-173.63	1.01091	100	1	902	300.67	0	0	0	0	0	0	0	0	0	0	0;
	4464	50.1	26.81	58.41	-16.26	1.054883	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4468	-76.7	14.09	38.01	-11.83	1.041888	100	1	100	-102.27	0	0	0	0	0	0	0	0	0	0	0;
	4478	410.5	109.19	220.62	-85.81	1.062733	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	4480	861.3	142.46	439.2	-173.43	1.065482	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	4482	35.2	13.98	38.85	-10.95	1.043069	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4486	38.5	1.4	16.9	-7.98	1.026359	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	4498	156.8	-17.45	89.58	-33.11	1.085986	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	4502	115	10.66	51.15	-22.62	1.022423	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4506	-171.8	15.64	37.86	-11.95	1.032581	100	1	100	-229.07	0	0	0	0	0	0	0	0	0	0	0;
	4516	364.8	11.93	173.07	-72.39	1.018246	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	4526	6.45	-3.38	26.83	-10.04	1.059432	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	4533	1530.1	40.53	1074.34	-397.66	1.087859	100	1	2400	800	0	0	0	0	0	0	0	0	0	0	0;
	4552	-117.74	-1.44	3.57	-4.42	1.021862	100	1	100	-156.99	0	0	0	0	0	0	0	0	0	0	0;
	4564	318	12.17	122.64	-61.66	1.026639	100	1	325	0	0	0	0	0	0	0	0	0	0	0	0;
	4566	-30.39	7.17	37.94	-11.87	1.07048	100	1	100	-40.51	0	0	0	0	0	0	0	0	0	0	0;
	4573	61.1	-1.78	51.69	-46.86	1.021097	100	1	93.49	0	0	0	0	0	0	0	0	0	0	0	0;
	4586	-545.7	54.63	92.96	-29.81	1.037111	100	1	100	-727.6	0	0	0	0	0	0	0	0	0	0	0;
	4602	428.86	198.47	375.7	-114.93	1.120537	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	4608	25.2	-3.94	17.31	-17.23	1.034566	100	1	26.46	0	0	0	0	0	0	0	0	0	0	0	0;
	4609	60.7	-4.43	45.93	-39.15	1.019251	100	1	72.58	0	0	0	0	0	0	0	0	0	0	0	0;
	4613	689.2	-122.64	254.86	-368.11	1.021632	100	1	772	257.33	0	0	0	0	0	0	0	0	0	0	0;
	4620	115.2	21.17	93.85	-28.85	1.089892	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	4624	68	8.12	35.62	-14.19	1.045235	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4633	-66.1	-4.03	-0.72	-6.6	1.009572	100	1	100	-88.13	0	0	0	0	0	0	0	0	0	0	0;
	4638	-236.3	9.42	38.28	-11.57	1.050874	100	1	100	-315.07	0	0	0	0	0	0	0	0	0	0	0;
	4647	78.4	22.32	45.01	-16.4	1.050276	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	4654	200	-15.84	133.34	-115.59	1.017349	100	1	210	0	0	0	0	0	0	0	0	0	0	0	0;
	4657	101.62	1.21	140.98	-43.31	1.020179	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	4661	-78.9	4.32	19.14	-5.77	1.056504	100	1	100	-105.2	0	0	0	0	0	0	0	0	0	0	0;
	4662	1509.3	-223.5	897.58	-328.8	1.066687	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	4663	-116	4.15	55.07	-19.63	1.021171	100	1	100	-154.67	0	0	0	0	0	0	0	0	0	0	0;
	4666	491	28.31	198.84	-105.44	1.011313	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	4667	-20.78	6.39	37.78	-12.04	1.01948	100	1	100	-27.7	0	0	0	0	0	0	0	0	0	0	0;
	4670	450.2	-86.23	214.71	-91.58	1.092961	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	4676	763.6	-70.38	343.25	-147.43	1.088792	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	4690	541	10.35	417.98	-132.47	1.026559	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	4695	-85.9	9.98	38.53	-11.22	1.053968	100	1	100	-114.53	0	0	0	0	0	0	0	0	0	0	0;
	4701	93.2	-37.03	98.4	-86.29	1.042354	100	1	281.85	0	0	0	0	0	0	0	0	0	0	0	0;
	4708	79.5	27.06	56.03	-18.68	0.996652	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4739	49.6	10.6	37.44	-12.28	1.021645	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4749	527.3	-66.96	280	-279.32	1.030135	100	1	553.66	0	0	0	0	0	0	0	0	0	0	0	0;
	4755	55.2	-5.3	30.48	-41.52	1.0132	100	1	57.96	0	0	0	0	0	0	0	0	0	0	0	0;
	4758	570.6	237.26	257.49	-110.95	1.064718	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	4772	78	-14.48	89.54	-34.98	1.023126	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	4773	16	-4.51	22.15	-20.9	1.017383	100	1	65.3	0	0	0	0	0	0	0	0	0	0	0	0;
	4776	805.5	127.43	389.83	-161.16	1.063105	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	4783	861.3	140.63	439.2	-173.43	1.06564	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	4788	387.47	62.3	379.51	-111.07	1.009861	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	4805	-183.5	32.22	190.29	-54.52	1.046224	100	1	100	-244.67	0	0	0	0	0	0	0	0	0	0	0;
	4816	-26.08	11.74	38.37	-11.44	1.060527	100	1	100	-34.77	0	0	0	0	0	0	0	0	0	0	0;
	4819	855.6	159.85	440.69	-172.14	1.028949	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	4823	10.3	0.51	4.44	-2.07	1.055712	100	1	10.46	0	0	0	0	0	0	0	0	0	0	0	0;
	4824	728.4	-75.11	558.52	-176.39	1.089599	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	4834	10.5	-1.01	7	-6.84	1.038553	100	1	11.03	0	0	0	0	0	0	0	0	0	0	0	0;
	4843	6.8	-1.73	5	-3.81	1.048394	100	1	13.63	0	0	0	0	0	0	0	0	0	0	0	0;
	4844	226.6	-78.28	377.48	-111.42	1.00703	100	1	500	100	0	0	0	0	0	0	0	0	0	0	0;
	4845	-6.25	3.32	27.54	-9.24	1.024167	100	1	100	-8.33	0	0	0	0	0	0	0	0	0	0	0;
	4846	239.6	48.55	133.84	-49.89	1.030121	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	4847	-173.7	-13.76	192.1	-53.15	1.081656	100	1	100	-231.6	0	0	0	0	0	0	0	0	0	0	0;
	4848	160.94	15.95	187.31	-57.24	1.070363	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	4850	88.8	11.12	76.52	-23.19	1.057137	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	4871	58.27	10	43.87	-17.53	1.054823	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	4880	-40.3	7.15	75.62	-23.96	1.041763	100	1	100	-53.73	0	0	0	0	0	0	0	0	0	0	0;
	4882	13	-2.75	6.22	-9.78	1.028068	100	1	13.65	0	0	0	0	0	0	0	0	0	0	0	0;
	4892	357	-28.65	176.3	-72.86	1.033418	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	4897	18.56	6.55	37.58	-12.23	1.014069	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4900	23.8	-3.06	14.79	-19.21	1.023832	100	1	24.99	0	0	0	0	0	0	0	0	0	0	0	0;
	4918	41.69	11.87	37.27	-12.57	1.053068	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4931	20.2	-14.61	34.39	-33.64	1.035713	100	1	102.79	0	0	0	0	0	0	0	0	0	0	0	0;
	4938	1201	128.03	933.87	-291.32	1.031692	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	4944	-8.84	8.01	18.64	-6.3	1.0154	100	1	100	-11.79	0	0	0	0	0	0	0	0	0	0	0;
	4945	385.4	-29.74	224.4	-82.22	1.055797	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	4949	60	3.02	57.63	-17	1.046901	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4952	-12.57	9.87	37.23	-12.61	1.047958	100	1	100	-16.77	0	0	0	0	0	0	0	0	0	0	0;
	4954	40.84	3.16	54.46	-20.26	1.016721	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4955	123.9	-20.46	73.86	-66.08	1.037342	100	1	159.4	0	0	0	0	0	0	0	0	0	0	0	0;
	4957	961	-175.55	533.39	-201.12	1.045811	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	4959	853.9	192.65	460.83	-440.49	1.051773	100	1	876	292	0	0	0	0	0	0	0	0	0	0	0;
	4967	83.2	-14.91	182.2	-75.83	1.020942	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	4969	453.4	62.07	375.73	-114.45	1.065019	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	4979	246	16.81	132.46	-51.71	1.039397	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	4994	360.4	135.52	173.76	-71.72	1.047685	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	4996	130.8	16.37	71.89	-27.72	1.040488	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5004	54.3	1.52	74.73	-70.41	1.054301	100	1	220.12	0	0	0	0	0	0	0	0	0	0	0	0;
	5005	406	151.95	220.88	-85.45	1.062291	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	5019	49.4	5.52	37.72	-12.09	1.05406	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5020	35.2	7.78	38.87	-10.93	1.067392	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5025	-155.1	-0.48	5.65	-4.3	1.002981	100	1	100	-206.8	0	0	0	0	0	0	0	0	0	0	0;
	5027	-10.3	-14.47	93.58	-29.18	1.046647	100	1	100	-13.73	0	0	0	0	0	0	0	0	0	0	0;
	5036	75.2	7.15	35.06	-14.85	1.060776	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5043	179.5	27.83	88.65	-36.1	1.07576	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	5051	83.48	1.58	74.55	-24.97	1.042576	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5059	130.2	34.96	81.97	-28.66	1.031064	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	5060	13.3	-4.32	14.22	-12.43	1.06427	100	1	40.68	0	0	0	0	0	0	0	0	0	0	0	0;
	5065	1008.3	175.33	532.96	-203.19	1.06654	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	5067	542	117.94	258.91	-109.58	1.016664	100	1	600	0	0	0	0	0	0	0	0	0	0	0	0;
	5075	-92.6	4.26	25.91	-10.99	1.027964	100	1	100	-123.47	0	0	0	0	0	0	0	0	0	0	0;
	5091	276.42	-47.94	151.81	-69.25	1.051243	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	5092	416.4	-21.8	221.62	-84.96	1.061075	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	5097	21.1	2.69	8.91	-4.11	1.052774	100	1	21.23	7.08	0	0	0	0	0	0	0	0	0	0	0;
	5098	517.2	-103.83	278.62	-562.09	1.006462	100	1	804	268	0	0	0	0	0	0	0	0	0	0	0;
	5110	-52.51	10.15	38.28	-11.5	1.033935	100	1	100	-70.01	0	0	0	0	0	0	0	0	0	0	0;
	5120	75.6	-2.03	56.35	-18.42	1.06465	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5134	197.3	26.91	100.9	-35.13	1.022715	100	1	220.45	0	0	0	0	0	0	0	0	0	0	0	0;
	5138	308.8	33.72	179.86	-65.56	1.048344	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	5144	1713.6	141.65	882.86	-343.24	1.032838	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	5145	422	219.01	450.4	-391.11	1.049287	100	1	1290.52	143.39	0	0	0	0	0	0	0	0	0	0	0;
	5152	50.59	28.52	37.75	-12.11	1.033193	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5154	-54.09	8.3	28.42	-8.32	1.021968	100	1	100	-72.12	0	0	0	0	0	0	0	0	0	0	0;
	5158	690.05	79.89	345.71	-144.76	1.043194	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	5164	629.3	204.21	301.89	-127.19	1.083362	100	1	700	233.33	0	0	0	0	0	0	0	0	0	0	0;
	5171	-87.6	-6.41	32.89	-32.29	1.011094	100	1	100	-116.8	0	0	0	0	0	0	0	0	0	0	0;
	5180	-17.5	0.2	4.03	-1.82	1.010906	100	1	100	-23.33	0	0	0	0	0	0	0	0	0	0	0;
	5192	143.4	-7.13	79.52	-30.85	1.039762	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	5193	38.49	-1.19	37.47	-12.26	1.022466	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5225	178.82	26.14	145.28	-54.15	1.050345	100	1	320	0	0	0	0	0	0	0	0	0	0	0	0;
	5226	235.1	-47.98	145.53	-53.81	0.992484	100	1	320	0	0	0	0	0	0	0	0	0	0	0	0;
	5231	120.79	23.81	91.91	-30.54	1.026163	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	5236	403.7	14.5	193.56	-80.83	1.047863	100	1	440	0	0	0	0	0	0	0	0	0	0	0	0;
	5237	106.8	11.87	75.02	-24.74	1.0542	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5239	-1344.4	-125.88	Inf	-Inf	0.990479	100	1	100	-1792.53	0	0	0	0	0	0	0	0	0	0	0;
	5251	137.3	24.03	71.44	-28.36	1.038759	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5262	418.2	-22.48	221.33	-85.22	1.059967	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	5264	120.6	-40.08	38.05	-73.95	1.034663	100	1	126.63	0	0	0	0	0	0	0	0	0	0	0	0;
	5273	-5.35	5.9	47.91	-13.29	1.026805	100	1	100	-7.13	0	0	0	0	0	0	0	0	0	0	0;
	5278	119.4	1.98	51.3	-23.49	1.03962	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5280	-193.9	-0.28	2.29	-1.06	1.018895	100	1	100	-258.53	0	0	0	0	0	0	0	0	0	0	0;
	5285	151	0.45	91.72	-32.91	1.057359	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	5290	332.54	-6.89	159.67	-61.42	1.009333	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	5293	306	-17.74	151.05	-97.5	1.008689	100	1	344	114.67	0	0	0	0	0	0	0	0	0	0	0;
	5299	152.3	-130.43	542.73	-582.68	1.045101	100	1	1689.14	51.19	0	0	0	0	0	0	0	0	0	0	0;
	5301	102.6	-9.51	94.66	-27.95	1.034956	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	5305	75.8	17.27	185.68	-59.56	1.029341	100	1	200	50.53	0	0	0	0	0	0	0	0	0	0	0;
	5309	44.9	-15.99	132.58	-139.97	1.042975	100	1	409.09	0	0	0	0	0	0	0	0	0	0	0	0;
	5323	41	3.48	38.42	-11.38	1.031211	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5329	-54.78	-8.89	35.25	-34.6	1.015053	100	1	100	-73.05	0	0	0	0	0	0	0	0	0	0	0;
	5340	-159	11.36	36.37	-13.43	1.044898	100	1	100	-212	0	0	0	0	0	0	0	0	0	0	0;
	5348	10.8	-1.79	11.36	-10.08	1.013341	100	1	32.74	0	0	0	0	0	0	0	0	0	0	0	0;
	5352	10	-1.61	4.29	-5.71	1.022029	100	1	10.5	0	0	0	0	0	0	0	0	0	0	0	0;
	5365	19.59	3.68	18.14	-6.75	1.065324	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	5368	200	5.66	137.65	-45.91	1.058977	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	5369	-76.5	14.66	38.34	-11.45	1.037177	100	1	100	-101.99	0	0	0	0	0	0	0	0	0	0	0;
	5379	1250.6	364.73	926.47	-297.63	1.10493	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	5395	-395.8	11.26	58.06	-16.66	1.05041	100	1	100	-527.73	0	0	0	0	0	0	0	0	0	0	0;
	5397	-131	17.95	36.81	-12.89	1.016845	100	1	100	-174.67	0	0	0	0	0	0	0	0	0	0	0;
	5398	333.2	135.77	177.29	-68.21	1.041434	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	5401	91.8	18.12	42.68	-18.57	1.056816	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	5403	439	208.3	219.43	-87.6	1.032334	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	5409	219.5	-18.7	102.6	-44.69	1.021468	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	5412	359.1	-8.27	174.75	-70.61	1.07527	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	5425	181.1	-8.83	108.27	-39.2	1.074405	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	5429	292.5	104.21	127.12	-56.96	1.043614	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	5444	865.8	351.49	382.87	-169.37	1.059208	100	1	900	200	0	0	0	0	0	0	0	0	0	0	0;
	5457	33.99	11.2	38.49	-11.3	1.044659	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5461	43.4	11.36	38.26	-11.55	1.055685	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5465	18.5	-6.87	31.09	-30.76	0.997959	100	1	93.47	0	0	0	0	0	0	0	0	0	0	0	0;
	5471	519.6	134.28	263.39	-104.86	1.026437	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	5479	179.07	131.32	188.35	-56.11	1.05048	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	5480	47.8	19.88	37.89	-11.93	1.043829	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5481	1368.6	134.39	920.31	-305.64	1.05825	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	5482	56.8	-10.13	101.32	-48.12	1.042942	100	1	249.06	0	0	0	0	0	0	0	0	0	0	0	0;
	5486	861.3	135.67	441.04	-172.05	1.055199	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	5487	265.8	74.35	129.3	-54.43	1.050272	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	5488	-96.4	15.72	37.49	-12.33	1.037007	100	1	100	-128.53	0	0	0	0	0	0	0	0	0	0	0;
	5490	3424.8	281.52	1765.36	-686.67	1.062487	100	1	4000	1333.33	0	0	0	0	0	0	0	0	0	0	0;
	5504	29.58	10.59	37.98	-11.86	1.033273	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5512	75.62	7.15	69.67	-30	1.017916	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5513	-196	-2.87	185.06	-60.09	1.030251	100	1	100	-261.33	0	0	0	0	0	0	0	0	0	0	0;
	5530	-95.9	27.28	36.94	-12.94	1.011393	100	1	100	-127.87	0	0	0	0	0	0	0	0	0	0	0;
	5533	173	12.32	88.87	-35.71	1.046404	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	5546	21.5	3.53	19.16	-5.75	1.063123	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	5556	571.8	229.47	255.93	-112.2	1.067161	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	5564	57.6	10.39	36.8	-12.99	1.038385	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5572	79.6	12.95	55.21	-18.48	1.033768	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5594	-238.2	0.6	17.52	-7.42	1.048652	100	1	100	-317.6	0	0	0	0	0	0	0	0	0	0	0;
	5602	42	-3.27	25.84	-34.16	1.032916	100	1	44.1	0	0	0	0	0	0	0	0	0	0	0	0;
	5603	228.6	-8.28	141.02	-122.88	1.027661	100	1	240.03	0	0	0	0	0	0	0	0	0	0	0	0;
	5611	33.7	0.38	17.99	-6.95	1.079526	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	5620	-165.1	62.85	80.73	-29.74	1.055555	100	1	100	-220.13	0	0	0	0	0	0	0	0	0	0	0;
	5645	78.17	27.37	52.51	-21.24	1.000282	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5656	35.69	-2.39	187.06	-57.96	1.039989	100	1	100	23.79	0	0	0	0	0	0	0	0	0	0	0;
	5658	229.8	81.02	103.18	-46.25	1.005722	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	5664	128	17.27	72.71	-27.05	1.063661	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5668	152.1	50.4	164.82	-56.07	1.028943	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	5669	294.54	12.23	34.47	-15.34	1.000678	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	5670	16.6	-3.12	22.97	-21.7	1.013301	100	1	67.76	0	0	0	0	0	0	0	0	0	0	0	0;
	5677	64.35	15.03	52.14	-21.63	1.033561	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5679	-36.09	1.41	27.25	-9.55	1.057325	100	1	100	-48.12	0	0	0	0	0	0	0	0	0	0	0;
	5692	654.8	78.31	356.53	-134.01	1.048689	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	5705	33.4	0.38	28.17	-8.65	1.084017	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	5709	837	87.13	550.69	-185.15	1.024594	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	5714	74.8	15.83	55.48	-18.13	1.020678	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5717	84.5	10.79	44	-17.33	1.070074	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	5721	114.99	6.44	82.65	-27.94	1.054563	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	5722	-42.07	-11.46	31.42	-63.03	1.009245	100	1	100	-56.09	0	0	0	0	0	0	0	0	0	0	0;
	5726	18.5	-8.95	31.98	-31.54	1.01699	100	1	95.98	0	0	0	0	0	0	0	0	0	0	0	0;
	5737	13.86	10.54	38.01	-11.78	1.008529	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5742	84.5	14.45	35.69	-16.43	1.023074	100	1	85.14	28.38	0	0	0	0	0	0	0	0	0	0	0;
	5750	128.1	49.62	123.76	-59.95	1.026001	100	1	306	102	0	0	0	0	0	0	0	0	0	0	0;
	5759	448.4	-95.85	376.65	-113.74	1.075601	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	5781	1652.7	590.72	779.29	-324.79	1.046351	100	1	1800	600	0	0	0	0	0	0	0	0	0	0	0;
	5786	31.7	8.04	23.3	-17.98	1.006382	100	1	63.84	0	0	0	0	0	0	0	0	0	0	0	0;
	5787	63.55	-9.4	114.67	-32.56	1.028374	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	5809	354.4	51.07	169.19	-75.03	1.010105	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	5811	2.66	2.33	17.42	-7.41	1.018688	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	5812	292.5	110.27	126.98	-57.08	1.031099	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	5814	181.2	56.15	196.98	-77.17	1.040708	100	1	440	146.67	0	0	0	0	0	0	0	0	0	0	0;
	5821	-74.21	-17.31	45.83	-46.3	1.016327	100	1	100	-98.94	0	0	0	0	0	0	0	0	0	0	0;
	5822	111.7	16.67	185.23	-59.9	1.063749	100	1	300	74.47	0	0	0	0	0	0	0	0	0	0	0;
	5823	404.6	-40.2	220.75	-85.54	1.060066	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	5824	935.4	54.24	539.22	-196.32	1.050736	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	5825	262.6	39.74	76.5	-34.05	0.992342	100	1	600	100	0	0	0	0	0	0	0	0	0	0	0;
	5828	13.2	-3.42	13.93	-12.27	1.027766	100	1	39.98	0	0	0	0	0	0	0	0	0	0	0	0;
	5831	-179.4	-0.29	1.58	-0.74	1.026046	100	1	100	-239.2	0	0	0	0	0	0	0	0	0	0	0;
	5838	16.1	-2.61	13.03	-16.97	1.030725	100	1	16.91	0	0	0	0	0	0	0	0	0	0	0	0;
	5839	101.4	13.53	53.01	-20.76	1.013483	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5844	696.8	45.98	351.25	-139.88	1.050341	100	1	800	0	0	0	0	0	0	0	0	0	0	0	0;
	5846	-46.5	-7.47	51.99	-57.17	1.02116	100	1	100	-62	0	0	0	0	0	0	0	0	0	0	0;
	5855	453.4	62.66	376.39	-114.06	1.065097	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	5856	305.7	29.18	214.13	-85	1.036185	100	1	480	160	0	0	0	0	0	0	0	0	0	0	0;
	5862	68.2	12.09	35.73	-14.12	1.059582	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5863	438.2	-35.2	358.2	-131.84	1.048267	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	5864	46.4	-0.25	28.39	-26.47	1.024889	100	1	48.72	0	0	0	0	0	0	0	0	0	0	0	0;
	5868	514.8	96.08	369.72	-120.26	1.048327	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	5874	16.16	3.36	46.53	-14.59	1.045045	100	1	100	10.77	0	0	0	0	0	0	0	0	0	0	0;
	5881	60	6.68	36.72	-13.13	1.056126	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5882	106.8	12.2	52.96	-21.81	1.077465	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5883	1031.4	513.54	526.77	-209.31	1.175144	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	5885	107.8	15.27	52.11	-21.65	1.031733	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5894	182.6	67.91	86.3	-36.39	1.035315	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	5896	55.6	31.81	37.32	-12.57	0.986344	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5922	-10.23	6.73	19.49	-5.39	1.030869	100	1	100	-13.64	0	0	0	0	0	0	0	0	0	0	0;
	5930	155.19	8	68.33	-31.42	1.042525	100	1	164.92	0	0	0	0	0	0	0	0	0	0	0	0;
	5940	-74.9	4.4	17.21	-7.73	1.026879	100	1	100	-99.87	0	0	0	0	0	0	0	0	0	0	0;
	5941	51.2	17.66	37.53	-12.28	1.051545	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5955	-32.29	-0.25	37.78	-11.99	1.015926	100	1	100	-43.05	0	0	0	0	0	0	0	0	0	0	0;
	5964	9.35	6.13	37.05	-12.63	1.005428	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5970	127.57	-18.86	172.47	-48.34	1.017378	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	5971	1221.2	407.37	934.54	-291.64	1.076888	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	5973	102.8	14.2	75.26	-24.44	1.055332	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5975	60.4	13.58	35.9	-13.72	1.0347	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5977	310.7	-35.81	179.91	-65.27	1.030489	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	5980	269.1	119.8	129.72	-54.22	1.05163	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	5983	-102.5	8.05	37.27	-12.59	1.061878	100	1	100	-136.67	0	0	0	0	0	0	0	0	0	0	0;
	5985	560.66	187.71	362.21	-128.13	1.120491	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	5986	1499.6	799.32	884.39	-341.63	1.103376	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	5988	492	-44.29	373.58	-117.06	1.069038	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	5991	854.6	-175.04	517.55	-217.25	1.012716	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	5994	-87.3	-10.42	60.51	-61.08	1.036793	100	1	100	-116.4	0	0	0	0	0	0	0	0	0	0	0;
	5998	-68.1	-3.53	27.48	-36.06	1.02348	100	1	100	-90.8	0	0	0	0	0	0	0	0	0	0	0;
	6000	617.7	335.83	411.66	-139.21	1.085353	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6001	19.2	-1.86	7.11	-7.7	1.036406	100	1	20.1	0	0	0	0	0	0	0	0	0	0	0	0;
	6033	49.87	10.06	56.86	-17.93	1.013332	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6036	820.2	289.17	390.4	-161.16	1.036335	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6046	636.63	-101.82	351.66	-138.95	1.094629	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	6086	88.2	-11.17	42.01	-57.99	1.017402	100	1	92.61	0	0	0	0	0	0	0	0	0	0	0	0;
	6088	-278.2	6.05	34.67	-15.14	1.069764	100	1	100	-370.93	0	0	0	0	0	0	0	0	0	0	0;
	6090	60.4	6.35	36.58	-13.23	1.030088	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6092	774.9	-107.4	395.59	-155.88	1.035378	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6100	-139.51	5.48	185.63	-59.55	1.041341	100	1	100	-186.01	0	0	0	0	0	0	0	0	0	0	0;
	6103	164.5	2.74	141.39	-42.37	1.013429	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	6105	84.9	42.02	55.73	-19.07	1.010683	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6120	58.6	28.58	36.73	-13.07	1.024222	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6124	490	49.36	254.71	-124.05	1.025031	100	1	790	263.33	0	0	0	0	0	0	0	0	0	0	0;
	6129	356	89.29	173.49	-71.8	1.033744	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	6130	49.4	21.24	37.68	-12.12	1.026444	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6131	602.4	295.03	415.26	-136.33	1.133869	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6138	-121.9	1.77	16.91	-7.99	1.010882	100	1	100	-162.53	0	0	0	0	0	0	0	0	0	0	0;
	6144	9.9	0.42	4.21	-2.03	0.987224	100	1	10.04	0	0	0	0	0	0	0	0	0	0	0	0;
	6145	190.58	-4.95	104.76	-42.6	1.019556	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	6153	94.3	-49.05	249.65	-261.82	1.019885	100	1	769.94	32.08	0	0	0	0	0	0	0	0	0	0	0;
	6168	26.3	-0.36	19.86	-14.93	1.047693	100	1	53.76	0	0	0	0	0	0	0	0	0	0	0	0;
	6169	-85.8	12.05	38.48	-11.27	1.033616	100	1	100	-114.4	0	0	0	0	0	0	0	0	0	0	0;
	6175	405.8	58.75	258.8	-147.69	1.023886	100	1	426.09	0	0	0	0	0	0	0	0	0	0	0	0;
	6179	65.9	4.7	35.43	-14.41	1.036613	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6181	540	-67.57	194.93	-133.98	1.000322	100	1	576	192	0	0	0	0	0	0	0	0	0	0	0;
	6182	55.8	9.34	57.93	-16.7	1.054529	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6201	310.4	21.09	178.97	-66.25	1.049209	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	6204	767.5	-119.16	447.13	-164.56	1.02934	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	6207	411.8	63.77	274.12	-93.01	1.03672	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	6233	879.6	255.34	545.68	-189.97	1.017658	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	6237	-83.33	-1.84	2.42	-3.65	1.01301	100	1	100	-111.11	0	0	0	0	0	0	0	0	0	0	0;
	6239	5.3	-0.18	2.24	-1.07	1.026501	100	1	5.32	0	0	0	0	0	0	0	0	0	0	0	0;
	6242	81.3	8.95	55.52	-19.08	1.018892	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6249	333.2	132.94	177.29	-68.21	1.040939	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	6270	450.4	-124.7	265.02	-187.97	1.02754	100	1	472.92	0	0	0	0	0	0	0	0	0	0	0	0;
	6279	-32.89	10.48	38.09	-11.69	1.008432	100	1	100	-43.85	0	0	0	0	0	0	0	0	0	0	0;
	6280	117.9	4.51	51.25	-23.49	1.017343	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6284	91.9	18.24	42.68	-18.57	1.057119	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	6291	55.5	30.11	58.12	-16.6	1.036939	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6329	223.4	3.69	134.08	-49.15	0.991058	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	6331	48.8	5.91	37.76	-12.05	1.050971	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6332	96	-5.55	96.56	-27.95	1.065904	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	6334	66.6	17.6	56.24	-17.36	1.005334	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6338	29.5	0.45	21.92	-16.86	1.022507	100	1	59.98	0	0	0	0	0	0	0	0	0	0	0	0;
	6351	22.08	13.57	38.29	-11.56	1.0521	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6358	40.79	28.9	37.69	-12.12	0.986287	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6366	188.9	27.37	139.07	-44.62	1.077817	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	6368	-145.3	11.76	51.11	-23.46	1.042432	100	1	100	-193.73	0	0	0	0	0	0	0	0	0	0	0;
	6370	450.4	48.2	279.2	-177.81	1.014839	100	1	538	179.33	0	0	0	0	0	0	0	0	0	0	0;
	6373	-201	1.6	72.5	-27.06	1.04869	100	1	100	-268	0	0	0	0	0	0	0	0	0	0	0;
	6374	318.4	-54.24	176.7	-68.15	1.077431	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	6376	-27.8	-13.24	148.65	-50.73	1.044986	100	1	100	-37.07	0	0	0	0	0	0	0	0	0	0	0;
	6403	52.8	-4.9	32.13	-45.87	1.020309	100	1	55.44	0	0	0	0	0	0	0	0	0	0	0	0;
	6410	-57	-18.17	23.72	-29.99	1.044406	100	1	100	-76	0	0	0	0	0	0	0	0	0	0	0;
	6413	958.2	465.98	537.03	-198.63	1.044796	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	6415	310.83	46	83.58	-38.78	0.983673	100	1	700	200	0	0	0	0	0	0	0	0	0	0	0;
	6419	1156.8	483.73	507.58	-228.2	1.095905	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	6429	171	19.61	88.83	-35.69	1.035857	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	6432	291.2	-18.62	180.93	-64.2	1.049448	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	6434	767.52	155.73	551.33	-184.73	1.116905	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	6436	-86.5	-0.2	2.26	-1.04	1.01064	100	1	100	-115.34	0	0	0	0	0	0	0	0	0	0	0;
	6448	19.39	11.37	18.36	-6.63	1.004995	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	6458	91.2	-2.12	43.09	-18.24	1.078056	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	6459	77.4	21.19	86.17	-24.25	1.005175	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	6474	718.5	-17.68	438.72	-174.15	1.055298	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	6483	86.3	5.42	55.34	-19.51	1.064876	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6484	1168.8	-39.57	508.02	-227.3	1.024853	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	6492	304.8	-9.69	180.02	-65	1.004319	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	6497	701.6	89.4	348.83	-141.85	1.064333	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	6498	187.8	-11.92	52.65	-91.35	1.019	100	1	197.19	0	0	0	0	0	0	0	0	0	0	0	0;
	6502	44.99	12.84	37.56	-12.19	1.002332	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6516	228.2	0.52	126.73	-47.83	1.050158	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	6518	1254	-198.78	388.92	-586.54	0.999167	100	1	1408	469.33	0	0	0	0	0	0	0	0	0	0	0;
	6525	16	-2.44	6.12	-9.88	1.007591	100	1	16.8	0	0	0	0	0	0	0	0	0	0	0	0;
	6552	-204.2	12.17	37.58	-12.16	1.045563	100	1	100	-272.27	0	0	0	0	0	0	0	0	0	0	0;
	6587	56.1	-6.68	27.23	-38.3	1.017615	100	1	58.91	0	0	0	0	0	0	0	0	0	0	0	0;
	6591	209.39	1.22	188.45	-56.61	1.040735	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	6594	488.87	91.52	414.32	-137.06	1.022988	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6608	-22.07	8.66	47.56	-13.56	1.019233	100	1	100	-29.43	0	0	0	0	0	0	0	0	0	0	0;
	6621	290.3	-54.73	125.6	-58.12	1.077535	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	6626	31.9	7.47	18.32	-6.66	1.042091	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	6627	99.5	5.37	42.14	-19.23	1.064448	100	1	100	33.33	0	0	0	0	0	0	0	0	0	0	0;
	6632	2013.9	-432.25	1378.19	-459.09	1.057948	100	1	3000	1000	0	0	0	0	0	0	0	0	0	0	0;
	6641	37.8	20.26	38.53	-11.19	1.007887	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6654	742.76	-57.91	337.04	-153.26	1.021829	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	6656	249	-27.52	131.88	-52.24	1.021248	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	6661	79.9	35.8	44.49	-16.82	0.997723	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	6680	41.41	5.89	28.15	-8.66	1.025575	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	6681	21.3	-20.42	36.22	-35.5	1.067534	100	1	108.38	0	0	0	0	0	0	0	0	0	0	0	0;
	6703	989.7	-97.77	533.98	-201.87	1.089943	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	6715	32.8	1.23	18.03	-6.9	1.025113	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	6734	382.7	-66.68	311.52	-132.09	1.061659	100	1	720	0	0	0	0	0	0	0	0	0	0	0	0;
	6736	670.8	138.75	405.88	-144.96	1.032666	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6739	59.4	-5.89	36.93	-12.97	1.07283	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6753	68.2	20.16	35.87	-14.02	1.045824	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6756	32.2	5.92	28.24	-8.57	1.025997	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	6764	9.9	0.34	4.21	-2	0.99015	100	1	9.99	0	0	0	0	0	0	0	0	0	0	0	0;
	6778	276.8	66.94	183.4	-62.02	1.038013	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	6780	14.5	-0.39	10.68	-8.06	1.074673	100	1	29.19	4.86	0	0	0	0	0	0	0	0	0	0	0;
	6793	32.6	-1.29	28.21	-8.61	1.049876	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	6798	949.7	178.3	429.61	-183.48	1.018682	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	6807	44.07	25.54	77.14	-22.46	1.025274	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	6811	113.9	-24.22	93.15	-29.23	1.093866	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	6816	16.53	-4.05	77.35	-22.26	1.062526	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	6820	15.9	-0.62	11.99	-9.06	1.047984	100	1	32.53	0	0	0	0	0	0	0	0	0	0	0	0;
	6822	251	16.22	131.81	-52.36	1.039383	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	6831	-57.5	6.93	35.62	-14.25	1.031732	100	1	100	-76.67	0	0	0	0	0	0	0	0	0	0	0;
	6838	578	215.46	255.23	-112.31	1.050849	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	6843	60.67	34.01	54.68	-19.96	1.013427	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6845	102.9	16.22	53.24	-21.47	1.045173	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6847	297.7	171.1	181.63	-63.96	1.018848	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	6850	30	-0.94	1.74	-3.18	1.010801	100	1	31.5	0	0	0	0	0	0	0	0	0	0	0	0;
	6852	-344.9	4.29	18.56	-6.34	1.034908	100	1	100	-459.87	0	0	0	0	0	0	0	0	0	0	0;
	6857	2463.6	394.19	1865.83	-585.95	1.018722	100	1	4000	1333.33	0	0	0	0	0	0	0	0	0	0	0;
	6861	-76.4	14.66	38.34	-11.45	1.037118	100	1	100	-101.87	0	0	0	0	0	0	0	0	0	0	0;
	6864	64.6	27.73	36.08	-13.75	1.020438	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6868	48.5	-5.97	23.34	-36.66	1.042934	100	1	50.92	0	0	0	0	0	0	0	0	0	0	0	0;
	6874	362.8	112.78	172.42	-72.85	1.03371	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	6877	19.8	-11.51	33.57	-32.95	1.062542	100	1	100.51	0	0	0	0	0	0	0	0	0	0	0	0;
	6888	133	2.32	189.31	-59.87	1.046712	100	1	300	88.67	0	0	0	0	0	0	0	0	0	0	0;
	6893	306.6	-51.67	122.24	-154.22	1.002405	100	1	364	121.33	0	0	0	0	0	0	0	0	0	0	0;
	6900	420.5	-72.58	218.81	-87.48	1.08427	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	6907	520.5	21.03	327.04	-146.46	1.059798	100	1	760	0	0	0	0	0	0	0	0	0	0	0	0;
	6910	116.8	45.51	104.8	-87.09	1.028742	100	1	295.83	0	0	0	0	0	0	0	0	0	0	0	0;
	6914	173.4	-21.81	87.28	-35.35	1.083288	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	6920	-9.7	-5.58	14.77	-25.23	1.008862	100	1	100	-12.93	0	0	0	0	0	0	0	0	0	0	0;
	6924	131.2	16.98	56.01	-25.39	1.02139	100	1	132.84	44.28	0	0	0	0	0	0	0	0	0	0	0;
	6930	98.3	-1.8	72.88	-55.01	1.047953	100	1	199.23	33.21	0	0	0	0	0	0	0	0	0	0	0;
	6934	-42.68	2.17	28.12	-8.69	1.059171	100	1	100	-56.91	0	0	0	0	0	0	0	0	0	0	0;
	6936	665.8	-96.62	388.7	-162.64	1.022407	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6937	156.62	-9.6	188.15	-56.96	1.025063	100	1	400	100	0	0	0	0	0	0	0	0	0	0	0;
	6947	227.5	-8.57	126.09	-48.29	1.049432	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	6948	319.2	167	179.39	-66.23	1.010398	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	6950	10.6	-3.02	11.7	-14	1.020998	100	1	31.5	0	0	0	0	0	0	0	0	0	0	0	0;
	6967	269	0.93	184.53	-60.73	1.065194	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	6969	39.8	1.1	38.53	-11.27	1.069563	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6982	-12.6	16.76	56.22	-18.44	1.039641	100	1	100	-16.8	0	0	0	0	0	0	0	0	0	0	0;
	6993	152.19	49	89.69	-33.05	1.069361	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	6999	1780.9	-177.13	862.05	-363.97	1.083106	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	7008	862	77.66	401.3	-186.61	1.04776	100	1	960	0	0	0	0	0	0	0	0	0	0	0	0;
	7025	-199.6	5.74	17.59	-7.34	1.027158	100	1	100	-266.13	0	0	0	0	0	0	0	0	0	0	0;
	7027	77.3	13.79	175.51	-73.7	1.043049	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	7028	438.5	206.5	219.32	-87.68	1.031022	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	7029	1152.7	-190.53	507.73	-226.73	1.058412	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	7032	204	38.47	131.74	-71.57	1.004806	100	1	214.2	0	0	0	0	0	0	0	0	0	0	0	0;
	7033	182.88	30.29	190.44	-54.85	1.098728	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	7036	65.8	9.66	36.16	-13.72	1.038394	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7049	55.5	15.24	58.18	-16.56	1.052472	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7056	448.85	159.76	267.36	-100.98	1.000187	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7058	50.4	8.58	37.59	-12.2	1.040913	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7063	423.2	-30.87	220.66	-85.89	1.059581	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	7065	199.5	-7.58	136.87	-46.68	1.019881	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	7075	-49.37	-2.12	13.41	-11.89	1.019093	100	1	100	-65.83	0	0	0	0	0	0	0	0	0	0	0;
	7085	292.8	106.83	126.93	-57.12	1.044403	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	7089	-178.8	9.96	38.35	-11.44	1.081026	100	1	100	-238.4	0	0	0	0	0	0	0	0	0	0	0;
	7093	119.2	80.4	93.14	-29.42	1.03985	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	7104	234.2	6.25	133.02	-50.29	1.002634	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	7107	9.5	-8.39	177.44	-67.55	1.040026	100	1	100	6.33	0	0	0	0	0	0	0	0	0	0	0;
	7113	159.8	-17.04	71.21	-63.88	1.033847	100	1	167.79	0	0	0	0	0	0	0	0	0	0	0	0;
	7114	-207.1	15.82	54.38	-20.35	1.083134	100	1	100	-276.13	0	0	0	0	0	0	0	0	0	0	0;
	7115	65.87	0.7	95.06	-83.34	1.026738	100	1	272.28	0	0	0	0	0	0	0	0	0	0	0	0;
	7128	490	-59.58	251.35	-126.57	1.007918	100	1	790	263.33	0	0	0	0	0	0	0	0	0	0	0;
	7134	73.3	-17.37	37.42	-50.58	1.039671	100	1	94.5	0	0	0	0	0	0	0	0	0	0	0	0;
	7137	876.4	-76.31	516.03	-219.17	1.024472	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	7153	13.48	5.86	17.61	-7.31	1.034539	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	7159	861.3	135.91	441.04	-172.05	1.055132	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	7166	54.66	32.86	55.11	-19.65	1.002006	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7169	56.4	0.64	37.06	-12.77	1.087809	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7183	1250.6	365.49	926.47	-297.63	1.104721	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	7200	177.08	60.6	86.18	-38.61	1.022531	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	7207	119.2	80.35	93.14	-29.42	1.039936	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	7209	7.3	-0.12	3.12	-1.46	1.049275	100	1	7.35	0	0	0	0	0	0	0	0	0	0	0	0;
	7212	66.3	-0.62	70.22	-62.02	1.021331	100	1	202.7	0	0	0	0	0	0	0	0	0	0	0	0;
	7214	359.1	-8.29	174.75	-70.61	1.0752	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	7223	49	-11.8	41.59	-54.41	1.018777	100	1	51.62	0	0	0	0	0	0	0	0	0	0	0	0;
	7227	650	-60.82	279.66	-234.09	1.008346	100	1	730	243.33	0	0	0	0	0	0	0	0	0	0	0;
	7228	524.9	358.08	370.35	-120.32	1.172784	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	7238	-16.8	9.36	92.99	-82.06	1.019781	100	1	100	-22.4	0	0	0	0	0	0	0	0	0	0	0;
	7250	19.58	1.9	17.26	-7.7	1.041313	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	7257	21.3	-8.32	35.77	-35.23	1.023182	100	1	107.29	0	0	0	0	0	0	0	0	0	0	0	0;
	7260	32	5.76	13.73	-6.28	1.020734	100	1	32.7	10.9	0	0	0	0	0	0	0	0	0	0	0;
	7267	-39.7	40.06	89.84	-34.84	1.006635	100	1	100	-52.93	0	0	0	0	0	0	0	0	0	0	0;
	7270	766.8	125.58	394.86	-156.12	1.063771	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	7278	354.7	-48.14	174.01	-71.05	1.063694	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	7279	-681.7	7.51	14.78	-6.64	1.02664	100	1	100	-908.93	0	0	0	0	0	0	0	0	0	0	0;
	7281	460.5	-70.11	318.74	-349.16	1.020691	100	1	673.72	0	0	0	0	0	0	0	0	0	0	0	0;
	7282	2340.4	58.73	1862.29	-588.16	1.029197	100	1	4000	1333.33	0	0	0	0	0	0	0	0	0	0	0;
	7293	-64.7	6.49	37.71	-12.12	1.029739	100	1	100	-86.27	0	0	0	0	0	0	0	0	0	0	0;
	7303	856.4	-33.97	384.34	-167.04	1.026431	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	7327	59.06	21.53	76.11	-23.62	1.046207	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	7328	53	24.52	37.45	-12.39	0.98954	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7329	62.8	8.94	36.26	-13.55	1.037608	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7330	399	147.35	167.11	-77.81	1.166291	100	1	401.01	133.67	0	0	0	0	0	0	0	0	0	0	0;
	7331	32.1	4.28	18.12	-6.81	1.064241	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	7337	30.6	8.42	32.44	-14.31	1.015107	100	1	32.13	0	0	0	0	0	0	0	0	0	0	0	0;
	7350	165.6	-8.63	75.31	-122.69	1.027675	100	1	174.04	0	0	0	0	0	0	0	0	0	0	0	0;
	7359	116.8	5.3	51	-22.79	1.046487	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7363	-134.8	-2.96	33.96	-38.99	1.005537	100	1	100	-179.73	0	0	0	0	0	0	0	0	0	0	0;
	7370	-22.3	-0.2	3.56	-2.33	1.034338	100	1	100	-29.73	0	0	0	0	0	0	0	0	0	0	0;
	7376	80.4	8.84	56.18	-18.62	1.041198	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7383	100.6	7.5	42.54	-19.54	1.011076	100	1	101.4	33.8	0	0	0	0	0	0	0	0	0	0	0;
	7386	124	-8.83	42.61	-69.39	1.023477	100	1	130.2	0	0	0	0	0	0	0	0	0	0	0	0;
	7393	-8.75	-10.7	34.08	-29.87	1.057049	100	1	100	-11.66	0	0	0	0	0	0	0	0	0	0	0;
	7401	242.4	117.46	134.01	-50.15	1.018508	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	7408	1485.2	99.7	408.24	-347.26	1.010858	100	1	1804	601.33	0	0	0	0	0	0	0	0	0	0	0;
	7427	856.4	-30.15	752.61	-226.42	1.072596	100	1	1600	533.33	0	0	0	0	0	0	0	0	0	0	0;
	7429	159.8	-4.54	77	-59.37	1.011339	100	1	180	60	0	0	0	0	0	0	0	0	0	0	0;
	7431	735.6	-104.15	399.88	-151.38	1.016271	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	7432	49.6	3.49	37.65	-12.14	1.040989	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7446	13.3	-1.36	14.28	-12.55	1.008326	100	1	40.96	0	0	0	0	0	0	0	0	0	0	0	0;
	7465	-7.55	6.29	27.69	-9.11	1.012708	100	1	100	-10.07	0	0	0	0	0	0	0	0	0	0	0;
	7466	-202.53	35.04	75.85	-77.06	1.060565	100	1	100	-270.04	0	0	0	0	0	0	0	0	0	0	0;
	7474	76.5	9.57	56.27	-18.41	1.035524	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7490	-2.2	-34.65	83	-39.42	1.097048	100	1	100	-2.93	0	0	0	0	0	0	0	0	0	0	0;
	7495	75.8	60.86	214.23	-59.77	1.031216	100	1	200	50.53	0	0	0	0	0	0	0	0	0	0	0;
	7498	-108.9	-8.62	17.85	-28.26	1.000261	100	1	100	-145.2	0	0	0	0	0	0	0	0	0	0	0;
	7499	47.8	-14.94	90.31	-32.29	0.999526	100	1	100	31.87	0	0	0	0	0	0	0	0	0	0	0;
	7504	5.2	-0.4	2.24	-1.05	1.015325	100	1	5.3	0	0	0	0	0	0	0	0	0	0	0	0;
	7508	715.1	-115.84	402.15	-149.17	1.058554	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	7520	702.4	259.25	350.36	-140.73	1.027003	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	7522	508	234.68	371.11	-119.67	0.995659	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	7558	-75.5	-1.14	6.35	-7.49	1.005058	100	1	100	-100.67	0	0	0	0	0	0	0	0	0	0	0;
	7568	100.8	11.35	53.4	-21.26	1.043388	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7607	166.09	18.53	131.89	-42.62	1.054831	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	7616	375.8	226.75	276.99	-89.88	1.131608	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7619	115.8	13.16	51.09	-22.69	1.037789	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7627	261.97	91.09	180.61	-64.06	1.054377	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	7628	243	91.06	133.78	-50.04	1.064136	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	7630	33.09	-15.93	45.85	-40.5	1.017306	100	1	131.82	0	0	0	0	0	0	0	0	0	0	0	0;
	7639	-72.7	7.55	27.79	-9.03	1.016695	100	1	100	-96.93	0	0	0	0	0	0	0	0	0	0	0;
	7641	769.2	27.55	378.21	-162.35	1.049265	100	1	880	200	0	0	0	0	0	0	0	0	0	0	0;
	7645	-95.8	-1.08	5.32	-6.47	1.028371	100	1	100	-127.73	0	0	0	0	0	0	0	0	0	0	0;
	7656	101.6	-32.92	108.47	-95.18	1.013707	100	1	312.34	34.7	0	0	0	0	0	0	0	0	0	0	0;
	7657	378	165.57	172.41	-72.99	1.129049	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	7659	-39.5	-0.51	74.7	-24.78	1.029603	100	1	100	-52.67	0	0	0	0	0	0	0	0	0	0	0;
	7661	168.6	8.37	76.75	-33.79	1.037591	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	7670	-30.87	10.15	28.05	-8.67	1.005135	100	1	100	-41.16	0	0	0	0	0	0	0	0	0	0	0;
	7671	285.1	-20.5	182.95	-62.33	1.080257	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	7676	46.6	12.69	38.03	-11.81	1.074322	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7681	330.06	13.88	176.18	-73.31	1.052405	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	7683	40.5	-7.8	24.04	-31.96	1.038382	100	1	42.5	0	0	0	0	0	0	0	0	0	0	0	0;
	7687	-40.2	17.73	38	-11.94	1.043409	100	1	100	-53.6	0	0	0	0	0	0	0	0	0	0	0;
	7697	228.2	0.49	126.73	-47.83	1.050111	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	7707	-93.9	11.6	38.67	-11.11	1.042135	100	1	100	-125.2	0	0	0	0	0	0	0	0	0	0	0;
	7718	348	87.15	173.91	-71.22	1.008757	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	7728	420	-22.62	193.14	-153.19	1.008042	100	1	472	157.33	0	0	0	0	0	0	0	0	0	0	0;
	7737	24.26	3.69	36.9	-13	1.061861	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7742	-18.64	4.17	19.1	-5.74	1.016237	100	1	100	-24.86	0	0	0	0	0	0	0	0	0	0	0;
	7743	229.6	20.17	97.38	-49.1	1.011249	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	7755	6.3	-0.96	4.78	-3.57	1.058479	100	1	12.9	0	0	0	0	0	0	0	0	0	0	0	0;
	7758	382.8	-18.73	195.3	-78.83	1.059721	100	1	440	0	0	0	0	0	0	0	0	0	0	0	0;
	7759	1425.1	688.92	700.16	-281.59	1.17759	100	1	1600	400	0	0	0	0	0	0	0	0	0	0	0;
	7761	106.8	15.27	52.75	-21.97	1.012378	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7776	1221.2	403.17	934.73	-291.52	1.077614	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	7779	345.2	157.99	281.92	-85.82	1.0806	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7785	212.4	26.57	105.85	-43.63	1.048533	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	7798	47.6	-0.3	37.78	-11.99	1.015906	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7800	76.49	8.33	55.02	-19.59	1.005948	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7808	1250.4	286.78	926.35	-297.68	1.050047	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	7819	25.6	-4.18	18.9	-14.44	1.018324	100	1	51.54	0	0	0	0	0	0	0	0	0	0	0	0;
	7833	15.5	50.12	106.77	-42.81	1.015416	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	7835	23.5	8.52	81.41	-29.17	1.037848	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	7836	14.9	0.18	6.39	-3.07	0.997593	100	1	15.22	0	0	0	0	0	0	0	0	0	0	0	0;
	7842	165	30.82	90.08	-34.56	1.014075	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	7843	390.4	37.76	170.14	-75.46	1.056109	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	7854	4.8	-33.27	175.29	-68.95	1.018633	100	1	100	3.2	0	0	0	0	0	0	0	0	0	0	0;
	7860	397	-58.45	380.36	-109.91	1.054905	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	7880	436	206.76	217.18	-89.22	1.024905	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	7890	-99	5.82	18.76	-6.17	1.05789	100	1	100	-132	0	0	0	0	0	0	0	0	0	0	0;
	7894	150.1	-24.46	73.03	-57.42	1.044899	100	1	167.79	0	0	0	0	0	0	0	0	0	0	0	0;
	7896	216.8	9.94	149.3	-50.09	1.054626	100	1	320	0	0	0	0	0	0	0	0	0	0	0	0;
	7913	1368.6	134.37	920.41	-305.58	1.0583	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	7926	-157.2	-16.55	35.82	-48.35	1.033819	100	1	100	-209.6	0	0	0	0	0	0	0	0	0	0	0;
	7929	1101.5	-34.68	698.07	-282.17	1.044801	100	1	1600	533.33	0	0	0	0	0	0	0	0	0	0	0;
	7934	207.9	-41.49	136.03	-47.55	1.090866	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	7940	411.8	64.52	274.12	-93.01	1.036846	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7949	75.6	-20.49	43.13	-54.75	1.033042	100	1	79.38	0	0	0	0	0	0	0	0	0	0	0	0;
	7953	-175.7	51.03	282.78	-84.17	1.067896	100	1	100	-234.27	0	0	0	0	0	0	0	0	0	0	0;
	7960	419	107.78	274.2	-93.22	1.017582	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7964	144.4	11.79	74.88	-53.81	1.005985	100	1	151.6	0	0	0	0	0	0	0	0	0	0	0	0;
	7979	1637.1	-390.58	1487.06	-470.65	1.064388	100	1	3200	1066.67	0	0	0	0	0	0	0	0	0	0	0;
	7983	189.3	-45.62	186.32	-199.72	1.014534	100	1	324.87	0	0	0	0	0	0	0	0	0	0	0	0;
	7995	60	-10.75	23	-36.99	1.024627	100	1	63	0	0	0	0	0	0	0	0	0	0	0	0;
	7998	811.8	72.83	391.57	-160.95	1.046131	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	8036	-4.64	3.9	28.2	-8.63	1.029729	100	1	100	-6.19	0	0	0	0	0	0	0	0	0	0	0;
	8043	-27	9.24	75.67	-23.96	1.058234	100	1	100	-36	0	0	0	0	0	0	0	0	0	0	0;
	8044	337.6	73.11	281.63	-85.71	1.041629	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	8045	7.4	2.88	4.64	-2.19	1.023558	100	1	10.98	0	0	0	0	0	0	0	0	0	0	0	0;
	8065	399.4	-33.09	276.94	-90.75	1.059118	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	8066	-188.9	-1.29	10.73	-9.53	1.011708	100	1	100	-251.87	0	0	0	0	0	0	0	0	0	0	0;
	8085	94.6	-14.36	30.24	-49.76	1.020835	100	1	99.22	0	0	0	0	0	0	0	0	0	0	0	0;
	8087	322.12	-42.11	177.22	-80.27	1.027752	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	8096	46.9	2.83	37.66	-12.14	1.065973	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8101	76.3	40.77	151.6	-153.82	1.059601	100	1	461.31	0	0	0	0	0	0	0	0	0	0	0	0;
	8105	22.5	0.03	4.1	-1.77	1.013571	100	1	23.62	0	0	0	0	0	0	0	0	0	0	0	0;
	8109	992	834.95	Inf	-Inf	1.059636	100	1	1041.6	0	0	0	0	0	0	0	0	0	0	0	0;
	8116	110.5	-22.04	155	-154.69	1.035167	100	1	326.65	0	0	0	0	0	0	0	0	0	0	0	0;
	8117	962.42	143.88	525.09	-210.37	1.032328	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	8119	75.8	12.03	34.53	-15.28	1.073486	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8123	57.4	29.59	36.83	-12.96	1.009241	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8134	56.8	29.63	36.87	-12.91	1.008958	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8136	-60.6	-0.67	4.32	-4.6	1.028389	100	1	100	-80.8	0	0	0	0	0	0	0	0	0	0	0;
	8156	178	7.63	87.43	-35.39	1.075225	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	8157	314	47.92	179.34	-66.1	1.045038	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	8158	153	-3.21	91.67	-33.01	1.051259	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	8159	698.6	-93.08	291.71	-137.41	1.060727	100	1	709.87	236.62	0	0	0	0	0	0	0	0	0	0	0;
	8185	116.8	45.07	104.19	-86.74	1.028753	100	1	294.35	0	0	0	0	0	0	0	0	0	0	0	0;
	8187	386.04	207.25	219.1	-87.91	1.0215	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	8203	-83.2	0.09	105.01	-96.06	1.008878	100	1	100	-110.93	0	0	0	0	0	0	0	0	0	0	0;
	8207	1.75	-40.48	70.17	-66.45	1.029137	100	1	207.27	0	0	0	0	0	0	0	0	0	0	0	0;
	8210	1321.7	-224.83	709.44	-271.31	1.044939	100	1	1600	533.33	0	0	0	0	0	0	0	0	0	0	0;
	8212	49.8	19.31	37.48	-12.26	1.009223	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8215	103.8	7.37	52.21	-21.45	1.031215	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	8218	82	-0.38	32.04	-29.68	1.013588	100	1	86.1	0	0	0	0	0	0	0	0	0	0	0	0;
	8222	66.8	15.48	35.76	-14.05	1.013304	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8223	401	137.22	276.81	-90.88	1.056121	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	8225	17	5.38	19.49	-5.42	1.062685	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	8227	483	-58.9	192.45	-305.55	1.022197	100	1	558	186	0	0	0	0	0	0	0	0	0	0	0;
	8240	-15.64	-15.94	36.34	-35.59	1.056841	100	1	100	-20.85	0	0	0	0	0	0	0	0	0	0	0;
	8242	358.36	-20.32	177.81	-79.93	1.010793	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	8247	253.4	-36.5	164.3	-56.85	1.055867	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	8252	320.6	50.95	178.09	-67.32	1.02226	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	8267	203.4	1.68	106.84	-42.56	1.042576	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	8269	12.89	2.61	18.94	-6	1.039628	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	8278	97.2	11.03	53.88	-20.8	1.005256	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	8297	87.2	-8.2	41.33	-61.53	1.024007	100	1	91.56	0	0	0	0	0	0	0	0	0	0	0	0;
	8304	411.8	64.09	274.12	-93.01	1.036714	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	8311	-8.75	3.46	17.8	-7.13	1.041866	100	1	100	-11.66	0	0	0	0	0	0	0	0	0	0	0;
	8312	1702.2	438.81	1096.33	-374.64	1.075986	100	1	2400	800	0	0	0	0	0	0	0	0	0	0	0;
	8317	1485.2	-75.45	402.46	-352.3	1.000913	100	1	1804	601.33	0	0	0	0	0	0	0	0	0	0	0;
	8340	95.05	-20.86	72.47	-27.11	1.020459	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	8362	407	44.27	192.7	-81.59	1.047632	100	1	440	0	0	0	0	0	0	0	0	0	0	0	0;
	8371	242	-6.52	185.66	-59.08	1.107597	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	8375	459.2	95.88	203.22	-91.52	1.024751	100	1	480	0	0	0	0	0	0	0	0	0	0	0	0;
	8377	163.23	37.86	136.04	-48.1	1.011199	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	8378	33.8	1.38	17.78	-7.11	1.020496	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	8383	331	79.17	225.36	-79.49	1.008495	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	8384	-35.8	6.59	24.27	-4.24	1.051363	100	1	100	-47.73	0	0	0	0	0	0	0	0	0	0	0;
	8398	61.78	11.94	34.16	-15.71	1.034241	100	1	82.32	0	0	0	0	0	0	0	0	0	0	0	0;
	8401	414.97	61.1	377.63	-112.99	1.074893	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	8413	379.2	20.12	171.01	-74.43	1.031312	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	8415	425	-73.81	218.07	-88.18	1.082292	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	8427	50	27.87	37.61	-12.19	1.012295	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8433	1.2	-51.84	380.55	-109.05	1.055181	100	1	100	0.8	0	0	0	0	0	0	0	0	0	0	0;
	8451	57.2	30.71	37.15	-12.73	0.998015	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8458	517.1	-66.57	474.42	-138.49	1.042019	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	8464	194.6	52.62	84.61	-38.08	1.080213	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	8473	-135	4.31	90.21	-34.45	1.038343	100	1	100	-180	0	0	0	0	0	0	0	0	0	0	0;
	8474	592.44	150.22	356.66	-133.72	1.093709	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	8481	61.48	20.86	55.64	-18.93	1.035652	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	8486	78.8	8.89	34.16	-15.67	1.043845	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8515	32.3	-7.39	93.03	-31.43	1.059651	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	8518	44.8	-9.87	28.59	-37.41	1.031547	100	1	47.04	0	0	0	0	0	0	0	0	0	0	0	0;
	8522	-75.69	0.41	3.61	-1.69	1.027564	100	1	100	-100.92	0	0	0	0	0	0	0	0	0	0	0;
	8531	194.8	-6.11	83.35	-39.08	1.019991	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	8532	-697.25	-741.74	Inf	-Inf	1.032486	100	1	100	-929.67	0	0	0	0	0	0	0	0	0	0	0;
	8553	2486.5	568.35	1851.27	-596.86	1.077556	100	1	4000	1266.67	0	0	0	0	0	0	0	0	0	0	0;
	8564	-65	19.6	51.91	-22.86	1.01667	100	1	100	-86.67	0	0	0	0	0	0	0	0	0	0	0;
	8569	296.8	65.13	181.24	-64.19	1.038006	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	8576	-1	2.92	12.68	-5.9	1.017525	100	1	100	-1.33	0	0	0	0	0	0	0	0	0	0	0;
	8586	34.78	5.45	37.75	-12.04	1.002287	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8591	342.49	-75.59	278.84	-87.97	1.06486	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	8603	205.2	-7.69	106.87	-42.62	1.022771	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	8605	427	174.48	273.94	-93.65	1.005953	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	8608	285	78.06	233.65	-72.59	1.033413	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	8610	-182.09	-0.22	224.7	-70.31	1.056933	100	1	100	-242.78	0	0	0	0	0	0	0	0	0	0	0;
	8611	-22.19	5.31	46.84	-14.41	1.024775	100	1	100	-29.58	0	0	0	0	0	0	0	0	0	0	0;
	8622	109.2	32.5	52.66	-21.81	1.028478	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	8623	183.23	24.91	110.42	-36.88	1.034977	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	8625	20.3	-3.77	34.48	-33.72	1.051079	100	1	103.03	0	0	0	0	0	0	0	0	0	0	0	0;
	8630	298.2	-23.82	124.56	-59.19	1.059594	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	8640	264.3	117.08	131.33	-52.84	1.174902	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	8663	461	-38.92	207.95	-97.15	1.025281	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	8670	-138.3	13.27	37.4	-12.46	1.042199	100	1	100	-184.4	0	0	0	0	0	0	0	0	0	0	0;
	8675	47	12.59	38.02	-11.83	1.037158	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8676	-259.1	16.18	34.92	-14.96	1.019558	100	1	100	-345.47	0	0	0	0	0	0	0	0	0	0	0;
	8683	-37.28	5.99	54.32	-20.39	1.039554	100	1	100	-49.7	0	0	0	0	0	0	0	0	0	0	0;
	8693	62.2	10.25	36.39	-13.44	1.073334	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8702	130.69	-8.78	173.2	-47.67	1.026759	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	8714	609.8	65.01	376.94	-228.97	1.015463	100	1	804	268	0	0	0	0	0	0	0	0	0	0	0;
	8721	1232.6	206.3	932.68	-293.14	1.055054	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	8729	20.89	-6.14	34.95	-30.8	1.008386	100	1	100.36	0	0	0	0	0	0	0	0	0	0	0	0;
	8737	355.83	378.51	379.68	-111.01	1.058573	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	8758	675.23	59.97	346.49	-144.15	1.047294	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	8762	327.23	140.11	168.92	-76.3	1.046022	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	8764	128.6	13.58	178.04	-67.25	1.048209	100	1	300	85.74	0	0	0	0	0	0	0	0	0	0	0;
	8770	-40.4	4.29	61.15	-47.58	1.012975	100	1	100	-53.87	0	0	0	0	0	0	0	0	0	0	0;
	8771	21.7	5.79	19.07	-5.8	1.006983	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	8795	166.2	-7.21	111.7	-37.85	1.052731	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	8797	29.7	-9.16	22.35	-16.77	1.072681	100	1	60.43	0	0	0	0	0	0	0	0	0	0	0	0;
	8807	-304	6.99	35.84	-14.07	1.049311	100	1	100	-405.33	0	0	0	0	0	0	0	0	0	0	0;
	8818	44.66	-0.48	76.87	-22.75	1.049908	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	8820	118.8	-7.91	92.95	-29.51	1.063983	100	1	200	66.67	0	0	0	0	0	0	0	0	0	0	0;
	8822	-59.72	46.21	192.72	-52.5	1.064182	100	1	100	-79.63	0	0	0	0	0	0	0	0	0	0	0;
	8851	76.9	10.72	44.89	-16.43	1.063062	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	8858	42.8	-30.35	163.91	-56.98	1.023098	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	8869	334.4	42.17	384.3	-105.83	1.034597	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	8871	-233.8	-23.38	106.62	-40.75	1.032491	100	1	100	-311.73	0	0	0	0	0	0	0	0	0	0	0;
	8872	1232.6	176.03	932.68	-293.14	1.054777	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	8875	443.8	90.2	272.77	-95.02	1.045119	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	8894	299.8	22.2	179.73	-65.04	1.093076	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	8896	40	-3.41	15.33	-19.26	1.032282	100	1	42	0	0	0	0	0	0	0	0	0	0	0	0;
	8897	26.05	3.47	56.5	-18.21	1.053267	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	8903	17	0.9	28.99	-28.41	1.023893	100	1	86.73	0	0	0	0	0	0	0	0	0	0	0	0;
	8914	50.4	14.68	37.19	-12.46	1.034467	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8920	38.8	6.33	17.27	-7.67	1.01084	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	8927	37.2	14.46	38.64	-11.12	1.012236	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8944	88.4	-9.35	42.9	-18.29	1.06934	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	8946	756.8	-111.17	339.05	-151.21	1.036925	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	8950	-255.7	11.51	38.23	-11.61	1.050158	100	1	100	-340.94	0	0	0	0	0	0	0	0	0	0	0;
	8956	498.4	92.59	266.17	-101.5	1.061596	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	8961	-61.8	-9.66	53.52	-50.38	1.043576	100	1	100	-82.4	0	0	0	0	0	0	0	0	0	0	0;
	8966	203.1	-28.42	136.96	-46.75	1.087897	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	8971	-95.4	-5.46	189.2	-55.62	1.043573	100	1	100	-127.2	0	0	0	0	0	0	0	0	0	0	0;
	8972	25.3	7.2	18.7	-6.17	1.038393	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	8973	1624.7	-336.51	859	-367.12	1.048016	100	1	2000	666.67	0	0	0	0	0	0	0	0	0	0	0;
	8976	110	17.41	95.65	-28.95	1.04878	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	8978	567.23	27.21	359.96	-130.66	1.038763	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	8981	19.3	-11.77	32.66	-32.18	1.023844	100	1	97.98	0	0	0	0	0	0	0	0	0	0	0	0;
	8987	-21.95	-11.67	31.68	-29.92	1.045054	100	1	100	-29.27	0	0	0	0	0	0	0	0	0	0	0;
	8988	70	-6.34	27.09	-37.49	1.016047	100	1	73.5	0	0	0	0	0	0	0	0	0	0	0	0;
	8993	-46.19	4.21	38.98	-10.82	1.020391	100	1	100	-61.58	0	0	0	0	0	0	0	0	0	0	0;
	8997	78.2	6.36	107.59	-101.48	1.01701	100	1	317.09	0	0	0	0	0	0	0	0	0	0	0	0;
	9000	341.2	25.87	175.3	-69.96	1.042081	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	9001	338.71	-10.47	220.49	-86.14	1.023475	100	1	500	166.67	0	0	0	0	0	0	0	0	0	0	0;
	9013	31.2	6.42	28.36	-8.47	1.023407	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	9027	458	43.78	376.25	-114.31	1.070556	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	9034	-138.3	-8.08	30.47	-45.98	1.005697	100	1	100	-184.4	0	0	0	0	0	0	0	0	0	0	0;
	9035	25.86	13.24	36.35	-13.36	1.009337	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	9039	-39.85	-0.41	2.23	-1.07	1.017349	100	1	100	-53.13	0	0	0	0	0	0	0	0	0	0	0;
	9043	-10.55	0.92	17.82	-7.09	1.018389	100	1	100	-14.07	0	0	0	0	0	0	0	0	0	0	0;
	9046	340.4	-38.99	176.14	-69.3	1.070778	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	9047	238	-20.85	186.96	-58.14	1.03154	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	9048	144	16.2	92.45	-32.16	1.050428	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	9054	306.37	73.2	177.87	-67.51	1.116695	100	1	400	133.33	0	0	0	0	0	0	0	0	0	0	0;
	9055	-90	0.38	11.6	-10.23	0.97622	100	1	100	-120	0	0	0	0	0	0	0	0	0	0	0;
	9058	4.25	10.64	47.34	-13.78	1.013989	100	1	100	2.84	0	0	0	0	0	0	0	0	0	0	0;
	9060	-418.8	5.84	34.79	-15.01	1.066622	100	1	100	-558.4	0	0	0	0	0	0	0	0	0	0	0;
	9067	369.6	64.56	198.07	-76.34	1.03898	100	1	440	146.67	0	0	0	0	0	0	0	0	0	0	0;
	9097	40.1	0.82	17.03	-7.9	1.016295	100	1	41.16	0	0	0	0	0	0	0	0	0	0	0	0;
	9099	-89.8	8.34	63.31	-68.57	1.028779	100	1	100	-119.73	0	0	0	0	0	0	0	0	0	0	0;
	9100	84.5	10.89	44	-17.33	1.07025	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	9101	698.8	5.65	442.99	-169.92	1.026276	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	9103	-58.6	-3.89	25.09	-33.2	1.027331	100	1	100	-78.13	0	0	0	0	0	0	0	0	0	0	0;
	9105	99.4	-13.56	40.58	-20.48	1.088711	100	1	100	33.33	0	0	0	0	0	0	0	0	0	0	0;
	9108	-34.07	1.26	18.65	-14.16	1.046328	100	1	100	-45.42	0	0	0	0	0	0	0	0	0	0	0;
	9110	438.2	126.21	172.53	-72.85	1.126918	100	1	900	200	0	0	0	0	0	0	0	0	0	0	0;
	9113	123.4	13.71	49.65	-24.07	1.014547	100	1	125.78	0	0	0	0	0	0	0	0	0	0	0	0;
	9125	39.7	1.38	17.28	-7.69	1.042649	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	9132	663	67.31	355.14	-135.29	1.04609	100	1	800	266.67	0	0	0	0	0	0	0	0	0	0	0;
	9137	42.7	31.91	335.96	-93.23	1.048884	100	1	100	28.47	0	0	0	0	0	0	0	0	0	0	0;
	9140	-37.2	33.7	73.19	-26.48	1.042971	100	1	100	-49.6	0	0	0	0	0	0	0	0	0	0	0;
	9150	860.8	118.28	439.38	-173.28	1.034687	100	1	1000	333.33	0	0	0	0	0	0	0	0	0	0	0;
	9161	366.4	77.61	175.96	-73.51	1.033176	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	9167	119.1	9.75	82.76	-27.75	1.046867	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	9168	299.2	39.33	126.8	-57.1	1.076561	100	1	301.01	100.34	0	0	0	0	0	0	0	0	0	0	0;
	9174	-331.4	152.65	175	-49.16	1.015292	100	1	100	-441.87	0	0	0	0	0	0	0	0	0	0	0;
	9180	72	7.21	77.69	-21.97	1.053552	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	9187	47.1	6.95	26.85	-10.01	1.017206	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	9195	30.16	17.78	36.15	-13.73	1.025329	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	9197	46.4	-0.98	34.65	-26.42	1.017237	100	1	95.02	0	0	0	0	0	0	0	0	0	0	0	0;
	9200	292.5	104.97	126.9	-57.14	1.040977	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	9215	4.77	3.71	18.19	-6.61	1.008696	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	9230	-126.7	47.4	127.96	-55.88	1.027519	100	1	100	-168.93	0	0	0	0	0	0	0	0	0	0	0;
	9235	353.39	124.55	176.33	-72.95	1.04114	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	9238	829.8	16.82	552.08	-183.99	1.097633	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	9239	419	107.54	274.2	-93.22	1.017601	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
];

%% branch data
%	fbus	tbus	r	x	b	rateA	rateB	rateC	ratio	angle	status	angmin	angmax
mpc.branch = [
	5147	3097	0.0006	0.00616	0	823	0	0	0	0	1	-360	360;
	5147	8763	0.0005	0.0074	0	921	0	0	0	0	1	-360	360;
	2169	3124	0.00107	0.00626	0	476	0	0	0	0	1	-360	360;
	2169	3234	0.000841	0.0055	0	476	0	0	0	0	1	-360	360;
	3517	711	0.00143	0.01563	0	1645	0	0	0	0	1	-360	360;
	3517	4533	0.00146	0.01692	0	1711	0	0	0	0	1	-360	360;
	427	5425	7e-05	0.00057	0	1711	0	0	0	0	1	-360	360;
	427	4704	6e-05	0.00056	0	1514	0	0	0	0	1	-360	360;
	5960	7431	0.003349	0.01656	0	0	0	0	0	0	1	-360	360;
	5960	3186	0.003409	0.016211	0	0	0	0	0	0	1	-360	360;
	4481	7431	0.003479	0.01586	0	0	0	0	0	0	1	-360	360;
	4481	3186	0.003099	0.017969	0	0	0	0	0	0	1	-360	360;
	531	9203	0.00127	0.01842	0	823	0	0	0	0	1	-360	360;
	531	835	0.00126	0.01847	0	1547	0	0	0	0	1	-360	360;
	1656	118	0.0013	0.01808	0	1218	0	0	0	0	1	-360	360;
	1656	3449	0.00131	0.01765	0	1744	0	0	0	0	1	-360	360;
	5084	1809	0.00065	0.01311	0	2793	0	0	0	0	1	-360	360;
	5084	6398	0.00067	0.00974	0	2598	0	0	0	0	1	-360	360;
	5205	7833	0.00236	0.01418	0	553	0	0	0	0	1	-360	360;
	5205	7770	0.002169	0.015909	0	514	0	0	0	0	1	-360	360;
	354	1278	0.00027	0.00254	0	1349	0	0	0	0	1	-360	360;
	354	7617	0.00027	0.00259	0	1119	0	0	0	0	1	-360	360;
	2611	1278	0.00033	0.00302	0	1448	0	0	0	0	1	-360	360;
	2611	905	0.00029	0.00312	0	1613	0	0	0	0	1	-360	360;
	1842	6542	0.000289	0.00137	0	533	0	0	0	0	1	-360	360;
	1842	7622	0.000219	0.00263	0	0	0	0	0	0	1	-360	360;
	2134	6542	0.000289	0.001341	0	457	0	0	0	0	1	-360	360;
	2134	8785	0.000229	0.00257	0	0	0	0	0	0	1	-360	360;
	6613	6542	0.000841	0.00689	0	514	0	0	0	0	1	-360	360;
	6613	1651	0.000979	0.00586	0	495	0	0	0	0	1	-360	360;
	5773	6542	0.00014	0.0012	0	0	0	0	0	0	1	-360	360;
	5773	7840	0.000211	0.001539	0	0	0	0	0	0	1	-360	360;
	5214	6542	0.000401	0.004901	0	0	0	0	0	0	1	-360	360;
	5214	7840	0.000219	0.00151	0	533	0	0	0	0	1	-360	360;
	8072	6542	0.000729	0.004539	0	495	0	0	0	0	1	-360	360;
	8072	2759	0.00062	0.00482	0	495	0	0	0	0	1	-360	360;
	2971	6069	0.00052	0.00407	0	1481	0	0	0	0	1	-360	360;
	2971	3445	0.00034	0.00502	0	1843	0	0	0	0	1	-360	360;
	7561	6069	0.00065	0.00698	0	1711	0	0	0	0	1	-360	360;
	7561	8825	0.00059	0.00806	0	1678	0	0	0	0	1	-360	360;
	6748	8864	0.00063	0.00464	0	495	0	0	0	0	1	-360	360;
	6748	5078	0.000599	0.005099	0	533	0	0	0	0	1	-360	360;
	4956	4785	0.00064	0.00307	0	0	0	0	0	0	1	-360	360;
	4956	7667	0.000659	0.003	0	0	0	0	0	0	1	-360	360;
	8619	6775	0.00053	0.00528	0	1086	0	0	0	0	1	-360	360;
	8619	8195	0.00056	0.0053	0	1283	0	0	0	0	1	-360	360;
	3107	4785	0.0018	0.007919	0	0	0	0	0	0	1	-360	360;
	3107	7154	0.001651	0.00862	0	0	0	0	0	0	1	-360	360;
	5437	6775	0.00045	0.00545	0	1382	0	0	0	0	1	-360	360;
	5437	905	0.00052	0.00476	0	1547	0	0	0	0	1	-360	360;
	2282	6876	0.01286	0.037421	0	0	0	0	0	0	1	-360	360;
	2282	5654	0.032397	0.123	0	0	0	0	0	0	1	-360	360;
	474	6876	0.032851	0.069727	0	0	0	0	0	0	1	-360	360;
	474	8045	0.059942	0.112802	0	0	0	0	0	0	1	-360	360;
	7917	7472	0.00011	0.00109	0	1086	0	0	0	0	1	-360	360;
	7917	8899	0.00011	0.00107	0	1185	0	0	0	0	1	-360	360;
	6225	7681	0.000461	0.00395	0	629	0	0	0	0	1	-360	360;
	6225	1508	0.000599	0.003789	0	591	0	0	0	0	1	-360	360;
	7118	6568	0.00173	0.01542	0	987	0	0	0	0	1	-360	360;
	7118	4346	0.00166	0.02083	0	1119	0	0	0	0	1	-360	360;
	3469	8850	0.006314	0.018488	0	0	0	0	0	0	1	-360	360;
	3469	1997	0.015727	0.037669	0	0	0	0	0	0	1	-360	360;
	7556	7058	0.00374	0.02095	0	800	0	0	0	0	1	-360	360;
	7556	4324	0.00511	0.019461	0	800	0	0	0	0	1	-360	360;
	1069	2867	0.011281	0.050771	0	457	0	0	0	0	1	-360	360;
	1069	2588	0.0115	0.049719	0	438	0	0	0	0	1	-360	360;
	3945	2867	0.011719	0.048659	0	438	0	0	0	0	1	-360	360;
	3945	2588	0.01193	0.047599	0	438	0	0	0	0	1	-360	360;
	132	2962	0.00064	0.00715	0	1481	0	0	0	0	1	-360	360;
	132	4393	0.00089	0.00712	0	1283	0	0	0	0	1	-360	360;
	5044	693	0.00048	0.00555	0	987	0	0	0	0	1	-360	360;
	5044	7860	0.00047	0.00566	0	1086	0	0	0	0	1	-360	360;
	6883	693	0.00046	0.00524	0	1053	0	0	0	0	1	-360	360;
	6883	7860	0.00052	0.00514	0	1053	0	0	0	0	1	-360	360;
	1456	4503	0.00438	0.010686	0	0	0	0	0	0	1	-360	360;
	1456	6658	0.031347	0.062653	0	0	0	0	0	0	1	-360	360;
	6494	4945	0.00029	0.00298	0	1613	0	0	0	0	1	-360	360;
	6494	6080	0.0003	0.00291	0	1514	0	0	0	0	1	-360	360;
	1384	9010	0.00155	0.01529	0	1185	0	0	0	0	1	-360	360;
	1384	3146	0.00148	0.0152	0	1283	0	0	0	0	1	-360	360;
	8697	8142	0.000169	0.00096	0	312	0	0	0	0	1	-360	360;
	8697	6195	0.000182	0.000942	0	312	0	0	0	0	1	-360	360;
	5665	424	0.00016	0.00184	0	0	0	0	0	0	1	-360	360;
	5665	7235	0.00022	0.00287	0	0	0	0	0	0	1	-360	360;
	3852	6497	0.00015	0.00187	0	0	0	0	0	0	1	-360	360;
	3852	4779	0.00022	0.00266	0	0	0	0	0	0	1	-360	360;
	674	8885	0.00138	0.0174	0	1316	0	0	0	0	1	-360	360;
	674	3115	0.00272	0.0288	0	1086	0	0	0	0	1	-360	360;
	8126	5177	0.00065	0.01024	0	1843	0	0	0	0	1	-360	360;
	8126	6007	0.00073	0.01004	0	1876	0	0	0	0	1	-360	360;
	8935	5177	0.00075	0.00983	0	1909	0	0	0	0	1	-360	360;
	8935	6007	0.00076	0.00963	0	1909	0	0	0	0	1	-360	360;
	7146	7388	0.00042	0.02666	0	1185	0	0	0	0	1	-360	360;
	4003	8855	0.00029	0.0026	0	0	0	0	0	0	1	-360	360;
	4003	5781	0.00022	0.00303	0	1185	0	0	0	0	1	-360	360;
	6996	3761	0.001	0.006159	0	0	0	0	0	0	1	-360	360;
	6996	9108	0.00113	0.00657	0	533	0	0	0	0	1	-360	360;
	1212	8258	0.0013	0.01277	0	1316	0	0	0	0	1	-360	360;
	1212	7330	0.00163	0.01808	0	1382	0	0	0	0	1	-360	360;
	8644	106	0.00047	0.00406	0	1316	0	0	0	0	1	-360	360;
	8644	5883	0.00235	0.02343	0	1316	0	0	0	0	1	-360	360;
	97	6413	0.00036	0.00445	0	1547	0	0	0	0	1	-360	360;
	97	1644	0.00017	0.0025	0	1909	0	0	0	0	1	-360	360;
	4379	6413	0.00036	0.00436	0	1580	0	0	0	0	1	-360	360;
	4379	3483	0.00017	0.00255	0	1810	0	0	0	0	1	-360	360;
	1885	3208	0.00096	0.00995	0	1349	0	0	0	0	1	-360	360;
	1885	6790	0.00152	0.01949	0	1382	0	0	0	0	1	-360	360;
	5314	3208	0.00094	0.01014	0	1251	0	0	0	0	1	-360	360;
	5314	6790	0.00149	0.01986	0	1415	0	0	0	0	1	-360	360;
	2885	3208	0.00107	0.01243	0	1152	0	0	0	0	1	-360	360;
	2885	2585	0.00102	0.01034	0	1218	0	0	0	0	1	-360	360;
	3717	4848	0.0047	0.02061	0	438	0	0	0	0	1	-360	360;
	3717	3688	0.00739	0.040531	0	400	0	0	0	0	1	-360	360;
	6064	8640	0.00049	0.00611	0	1942	0	0	0	0	1	-360	360;
	6064	7330	0.00035	0.00431	0	1547	0	0	0	0	1	-360	360;
	5377	1029	0.00261	0.01344	0	419	0	0	0	0	1	-360	360;
	5377	3525	0.003521	0.019591	0	419	0	0	0	0	1	-360	360;
	6870	3431	0.0025	0.01419	0	0	0	0	0	0	1	-360	360;
	6870	5046	0.00287	0.0237	0	381	0	0	0	0	1	-360	360;
	6693	6743	0.00088	0.00956	0	1547	0	0	0	0	1	-360	360;
	6693	287	0.00095	0.00877	0	1382	0	0	0	0	1	-360	360;
	2053	752	0.00332	0.020651	0	457	0	0	0	0	1	-360	360;
	2053	3048	0.003969	0.018781	0	457	0	0	0	0	1	-360	360;
	2893	2738	0.002849	0.014909	0	438	0	0	0	0	1	-360	360;
	2893	1575	0.00439	0.02451	0	438	0	0	0	0	1	-360	360;
	1524	8687	0.0002	0.0018	0	0	0	0	0	0	1	-360	360;
	1524	8393	0.00025	0.00261	0	0	0	0	0	0	1	-360	360;
	7369	8687	0.0002	0.00176	0	0	0	0	0	0	1	-360	360;
	7369	139	0.00025	0.00255	0	1909	0	0	0	0	1	-360	360;
	1019	6176	0.012488	0.047413	0	0	0	0	0	0	1	-360	360;
	1019	7177	0.010479	0.027711	0	0	0	0	0	0	1	-360	360;
	2607	7458	0.00044	0.00428	0	1810	0	0	0	0	1	-360	360;
	2607	1337	0.00036	0.00475	0	0	0	0	0	0	1	-360	360;
	4735	5239	0.00058	0.00534	0	1876	0	0	0	0	1	-360	360;
	4735	3145	0.00054	0.00589	0	1843	0	0	0	0	1	-360	360;
	2941	1429	0.00206	0.02057	0	1152	0	0	0	0	1	-360	360;
	2941	1542	0.00077	0.00872	0	954	0	0	0	0	1	-360	360;
	3773	9183	0.00161	0.01963	0	1349	0	0	0	0	1	-360	360;
	3773	1542	0.00075	0.00888	0	1382	0	0	0	0	1	-360	360;
	1937	7261	0.00016	0.00152	0	1645	0	0	0	0	1	-360	360;
	1937	5479	0.00019	0.00163	0	1711	0	0	0	0	1	-360	360;
	5473	7261	0.00015	0.0016	0	1547	0	0	0	0	1	-360	360;
	5473	5479	0.00015	0.00183	0	1645	0	0	0	0	1	-360	360;
	3648	3290	0.00121	0.01426	0	1415	0	0	0	0	1	-360	360;
	3648	5523	0.00116	0.01456	0	1448	0	0	0	0	1	-360	360;
	451	144	0.00068	0.00727	0	1777	0	0	0	0	1	-360	360;
	451	1601	0.0007	0.00734	0	1876	0	0	0	0	1	-360	360;
	7898	144	0.00076	0.00712	0	1580	0	0	0	0	1	-360	360;
	7898	1601	0.0008	0.00719	0	1909	0	0	0	0	1	-360	360;
	2105	1582	0.0002	0.0022	0	1218	0	0	0	0	1	-360	360;
	2105	7416	0.0002	0.0022	0	1251	0	0	0	0	1	-360	360;
	270	2046	0.000205	0.00215	0	1415	0	0	0	0	1	-360	360;
	270	7416	0.000205	0.00215	0	1448	0	0	0	0	1	-360	360;
	5938	8433	0.00053	0.00589	0	1316	0	0	0	0	1	-360	360;
	5938	816	0.00046	0.00675	0	1119	0	0	0	0	1	-360	360;
	8539	8433	0.00054	0.00577	0	1316	0	0	0	0	1	-360	360;
	8539	816	0.00045	0.00625	0	1086	0	0	0	0	1	-360	360;
	1654	1544	0.00275	0.02752	0	1382	0	0	0	0	1	-360	360;
	1654	7674	0.00184	0.01515	0	1316	0	0	0	0	1	-360	360;
	2323	4150	0.00438	0.02432	0	0	0	0	0	0	1	-360	360;
	2323	599	0.0043	0.01949	0	0	0	0	0	0	1	-360	360;
	8376	6999	0.00033	0.00269	0	1218	0	0	0	0	1	-360	360;
	8376	4903	0.00243	0.01904	0	1119	0	0	0	0	1	-360	360;
	942	6999	0.00034	0.00263	0	1119	0	0	0	0	1	-360	360;
	942	4903	0.00248	0.01864	0	1086	0	0	0	0	1	-360	360;
	6754	6190	0.00122	0.01249	0	987	0	0	0	0	1	-360	360;
	6754	4055	0.00106	0.01031	0	1053	0	0	0	0	1	-360	360;
	3313	6190	0.00119	0.01273	0	987	0	0	0	0	1	-360	360;
	3313	4055	0.00108	0.01009	0	1086	0	0	0	0	1	-360	360;
	1383	4401	0.001039	0.013711	0	0	0	0	0	0	1	-360	360;
	1383	5145	0.00114	0.011409	0	0	0	0	0	0	1	-360	360;
	7851	5878	0.001021	0.013969	0	0	0	0	0	0	1	-360	360;
	7851	5145	0.001169	0.011159	0	0	0	0	0	0	1	-360	360;
	8001	8201	0	0.000271	0	0	0	0	0	0	1	-360	360;
	2322	8201	0	0.000271	0	0	0	0	0	0	1	-360	360;
	843	154	0.009372	0.035702	0	0	0	0	0	0	1	-360	360;
	843	3598	0.029157	0.103686	0	0	0	0	0	0	1	-360	360;
	529	5527	0.00065	0.00673	0	1514	0	0	0	0	1	-360	360;
	529	4164	0.00066	0.00658	0	1678	0	0	0	0	1	-360	360;
	3529	2487	0.000969	0.006021	0	553	0	0	0	0	1	-360	360;
	3529	4167	0.00099	0.005901	0	381	0	0	0	0	1	-360	360;
	1162	8598	0.00088	0.006599	0	0	0	0	0	0	1	-360	360;
	1162	5325	0.00086	0.006729	0	495	0	0	0	0	1	-360	360;
	8690	3277	9e-05	0.00077	0	1711	0	0	0	0	1	-360	360;
	8690	4226	9e-05	0.00078	0	1580	0	0	0	0	1	-360	360;
	4922	3277	8e-05	0.00082	0	1481	0	0	0	0	1	-360	360;
	4922	7617	8e-05	0.00076	0	1251	0	0	0	0	1	-360	360;
	6759	5527	9e-05	0.00074	0	1613	0	0	0	0	1	-360	360;
	6759	4226	9e-05	0.00073	0	1251	0	0	0	0	1	-360	360;
	4088	2487	0.00062	0.003781	0	0	0	0	0	0	1	-360	360;
	4088	782	0.000599	0.0035	0	553	0	0	0	0	1	-360	360;
	6871	8598	0.00064	0.00274	0	0	0	0	0	0	1	-360	360;
	6871	1346	0.000651	0.00268	0	610	0	0	0	0	1	-360	360;
	6813	1369	0.00055	0.00685	0	1777	0	0	0	0	1	-360	360;
	6813	376	0.00061	0.00639	0	1843	0	0	0	0	1	-360	360;
	6019	5527	0.00042	0.00523	0	1514	0	0	0	0	1	-360	360;
	6019	8334	0.00045	0.00605	0	1481	0	0	0	0	1	-360	360;
	5693	3277	0.0003	0.00228	0	1448	0	0	0	0	1	-360	360;
	5693	8899	0.00026	0.00264	0	1218	0	0	0	0	1	-360	360;
	3355	9230	0.00536	0.03568	0	553	0	0	0	0	1	-360	360;
	3355	6287	0.006091	0.03549	0	572	0	0	0	0	1	-360	360;
	5485	9230	0.006419	0.033271	0	591	0	0	0	0	1	-360	360;
	5485	1481	0.005789	0.035901	0	610	0	0	0	0	1	-360	360;
	6968	1829	0	0.0002	0	0	0	0	0	0	1	-360	360;
	6968	7774	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7956	9230	0.008401	0.04011	0	419	0	0	0	0	1	-360	360;
	7956	3197	0.007831	0.041531	0	419	0	0	0	0	1	-360	360;
	7550	667	0.00575	0.046461	0	572	0	0	0	0	1	-360	360;
	7550	1220	0.006711	0.04426	0	591	0	0	0	0	1	-360	360;
	3615	8731	0.00036	0.00485	0	0	0	0	0	0	1	-360	360;
	3615	6246	0.00042	0.00415	0	0	0	0	0	0	1	-360	360;
	7031	1679	9e-05	0.00069	0	1316	0	0	0	0	1	-360	360;
	7031	7420	4e-05	0.00036	0	1415	0	0	0	0	1	-360	360;
	1016	2502	0.00055	0.00569	0	1316	0	0	0	0	1	-360	360;
	1016	8556	0.0007	0.00767	0	1185	0	0	0	0	1	-360	360;
	6077	3885	0.0013	0.01299	0	1185	0	0	0	0	1	-360	360;
	2935	8322	0.00092	0.01031	0	1645	0	0	0	0	1	-360	360;
	1095	3885	0.00158	0.01496	0	1448	0	0	0	0	1	-360	360;
	1095	1132	0.00123	0.01309	0	1909	0	0	0	0	1	-360	360;
	4048	6639	0.000599	0.00274	0	533	0	0	0	0	1	-360	360;
	4048	7317	0.0005	0.00405	0	0	0	0	0	0	1	-360	360;
	2549	7829	0.003349	0.02445	0	553	0	0	0	0	1	-360	360;
	2549	4951	0.003281	0.02493	0	553	0	0	0	0	1	-360	360;
	3978	3622	0.00087	0.01097	0	1251	0	0	0	0	1	-360	360;
	3978	8974	0.00161	0.01464	0	1349	0	0	0	0	1	-360	360;
	7004	3370	0	0.00046	0	757	0	0	0	0	1	-360	360;
	4325	3684	0.00118	0.01223	0	987	0	0	0	0	1	-360	360;
	4325	6710	0.00045	0.00364	0	1251	0	0	0	0	1	-360	360;
	2380	1952	0.00714	0.03393	0	0	0	0	0	0	1	-360	360;
	2380	7679	0.00655	0.036849	0	0	0	0	0	0	1	-360	360;
	5026	7953	0.00088	0.00722	0	1810	0	0	0	0	1	-360	360;
	5026	6446	0.00071	0.00835	0	1909	0	0	0	0	1	-360	360;
	6575	7953	0.00084	0.00705	0	1843	0	0	0	0	1	-360	360;
	6575	8401	0.00102	0.00799	0	1613	0	0	0	0	1	-360	360;
	6387	5594	0.003781	0.029969	0	533	0	0	0	0	1	-360	360;
	6387	8145	0.00386	0.02969	0	533	0	0	0	0	1	-360	360;
	6694	8426	0.00076	0.00463	0	572	0	0	0	0	1	-360	360;
	6694	2887	0.000711	0.004841	0	572	0	0	0	0	1	-360	360;
	5206	9184	0.00631	0.034469	0	400	0	0	0	0	1	-360	360;
	5206	7410	0.006169	0.03512	0	419	0	0	0	0	1	-360	360;
	3087	5161	0.002961	0.017651	0	610	0	0	0	0	1	-360	360;
	3087	2887	0.00299	0.02462	0	591	0	0	0	0	1	-360	360;
	253	9184	0.002789	0.011659	0	495	0	0	0	0	1	-360	360;
	253	6718	0.00249	0.013211	0	476	0	0	0	0	1	-360	360;
	4180	9184	0.002669	0.016031	0	629	0	0	0	0	1	-360	360;
	4180	4702	0.00261	0.014841	0	629	0	0	0	0	1	-360	360;
	5912	9184	0.00074	0.004599	0	381	0	0	0	0	1	-360	360;
	5912	1971	0.000719	0.00468	0	438	0	0	0	0	1	-360	360;
	2014	4697	3.5e-05	0.000355	0	1251	0	0	0	0	1	-360	360;
	2014	4824	3.5e-05	0.000355	0	1251	0	0	0	0	1	-360	360;
	1434	246	0.00511	0.02949	0	419	0	0	0	0	1	-360	360;
	1434	5733	0.007719	0.044271	0	400	0	0	0	0	1	-360	360;
	2936	246	0.004281	0.022469	0	381	0	0	0	0	1	-360	360;
	2936	3103	0.005961	0.03382	0	457	0	0	0	0	1	-360	360;
	4893	6468	0.00073	0.01034	0	1678	0	0	0	0	1	-360	360;
	4893	6848	0.00076	0.00889	0	1843	0	0	0	0	1	-360	360;
	2601	2613	0.005159	0.025789	0	400	0	0	0	0	1	-360	360;
	2601	9073	0.00549	0.027711	0	0	0	0	0	0	1	-360	360;
	2118	2613	0.00614	0.01764	0	400	0	0	0	0	1	-360	360;
	2118	3198	0.003229	0.01788	0	0	0	0	0	0	1	-360	360;
	8381	4795	0.00053	0.00587	0	1777	0	0	0	0	1	-360	360;
	8381	5562	0.00055	0.00666	0	1711	0	0	0	0	1	-360	360;
	5763	6652	0.00326	0.01388	0	0	0	0	0	0	1	-360	360;
	5763	4064	0.003081	0.01488	0	0	0	0	0	0	1	-360	360;
	3409	1716	0.00051	0.00614	0	0	0	0	0	0	1	-360	360;
	3409	6246	0.00043	0.00406	0	0	0	0	0	0	1	-360	360;
	8028	5930	0.00761	0.046031	0	438	0	0	0	0	1	-360	360;
	8028	9044	0.00351	0.01632	0	419	0	0	0	0	1	-360	360;
	5035	1610	0.00031	0.00495	0	1744	0	0	0	0	1	-360	360;
	5035	9133	0.00043	0.00485	0	1942	0	0	0	0	1	-360	360;
	8751	147	0.00426	0.027099	0	629	0	0	0	0	1	-360	360;
	8751	5885	0.00437	0.024969	0	591	0	0	0	0	1	-360	360;
	6075	7896	0.005	0.03932	0	419	0	0	0	0	1	-360	360;
	6075	7473	0.006039	0.036789	0	457	0	0	0	0	1	-360	360;
	8268	9122	0.0009	0.00893	0	1218	0	0	0	0	1	-360	360;
	8268	7314	0.00286	0.02156	0	1053	0	0	0	0	1	-360	360;
	679	147	0.00418	0.022729	0	629	0	0	0	0	1	-360	360;
	679	5677	0.00405	0.02231	0	629	0	0	0	0	1	-360	360;
	7291	4830	0.00013	0.00675	0	1645	0	0	0	0	1	-360	360;
	7291	1249	0.00013	0.00688	0	1678	0	0	0	0	1	-360	360;
	9126	4079	0.00078	0.01201	0	0	0	0	0	0	1	-360	360;
	9126	6529	0.00075	0.01254	0	1777	0	0	0	0	1	-360	360;
	2329	5169	0.00071	0.00722	0	0	0	0	0	0	1	-360	360;
	2329	4337	0.00084	0.0074	0	1876	0	0	0	0	1	-360	360;
	4142	8137	0.00105	0.0081	0	1053	0	0	0	0	1	-360	360;
	4142	7870	0.00203	0.01565	0	1119	0	0	0	0	1	-360	360;
	8500	1586	0.00035	0.00581	0	1876	0	0	0	0	1	-360	360;
	8500	2634	0.00045	0.00534	0	0	0	0	0	0	1	-360	360;
	7357	7529	8.1e-05	0.001331	0	0	0	0	0	0	1	-360	360;
	7357	6297	0.00025	0.001591	0	0	0	0	0	0	1	-360	360;
	6959	2290	0.00401	0.026219	0	419	0	0	0	0	1	-360	360;
	6959	7473	0.004669	0.027521	0	419	0	0	0	0	1	-360	360;
	307	8580	0.00115	0.01265	0	1086	0	0	0	0	1	-360	360;
	307	1360	0.00112	0.0129	0	1086	0	0	0	0	1	-360	360;
	7764	5850	0.00123	0.01589	0	757	0	0	0	0	1	-360	360;
	7764	7870	0.00054	0.00484	0	1185	0	0	0	0	1	-360	360;
	3778	8172	0.00401	0.03961	0	0	0	0	0	0	1	-360	360;
	3778	3888	0.002909	0.01981	0	514	0	0	0	0	1	-360	360;
	1000	6145	0.002211	0.014419	0	438	0	0	0	0	1	-360	360;
	1000	3186	0.002169	0.013349	0	457	0	0	0	0	1	-360	360;
	3651	8580	0.00055	0.00508	0	1218	0	0	0	0	1	-360	360;
	3651	895	0.00046	0.00595	0	1448	0	0	0	0	1	-360	360;
	6633	7467	0.00201	0.00988	0	553	0	0	0	0	1	-360	360;
	6633	2200	0.003401	0.019099	0	572	0	0	0	0	1	-360	360;
	1976	680	0.00028	0.00308	0	1678	0	0	0	0	1	-360	360;
	1976	5674	0.00038	0.00307	0	1843	0	0	0	0	1	-360	360;
	2435	4208	0.00044	0.00577	0	757	0	0	0	0	1	-360	360;
	2435	2398	0.00102	0.01295	0	757	0	0	0	0	1	-360	360;
	5605	1653	0.00085	0.01005	0	1711	0	0	0	0	1	-360	360;
	5605	5498	0.00073	0.0097	0	1843	0	0	0	0	1	-360	360;
	7646	1884	0.00139	0.0126	0	1349	0	0	0	0	1	-360	360;
	7646	5503	0.00117	0.01389	0	1218	0	0	0	0	1	-360	360;
	6574	1506	0.00076	0.00908	0	1382	0	0	0	0	1	-360	360;
	6574	8562	0.00137	0.01338	0	1283	0	0	0	0	1	-360	360;
	3286	2588	0.006	0.0402	0	438	0	0	0	0	1	-360	360;
	3286	689	0.005719	0.03906	0	438	0	0	0	0	1	-360	360;
	2272	1884	0.0007	0.01065	0	856	0	0	0	0	1	-360	360;
	2272	2294	0.00066	0.01062	0	856	0	0	0	0	1	-360	360;
	4091	3146	0.00153	0.01695	0	1152	0	0	0	0	1	-360	360;
	4091	6098	0.00083	0.00925	0	1316	0	0	0	0	1	-360	360;
	6060	5289	0.00033	0.00101	0	823	0	0	0	0	1	-360	360;
	7965	5289	0.00028	0.00104	0	757	0	0	0	0	1	-360	360;
	1049	603	0.00139	0.0125	0	1283	0	0	0	0	1	-360	360;
	1049	8344	0.00119	0.01236	0	1119	0	0	0	0	1	-360	360;
	2363	8726	0.010099	0.05513	0	381	0	0	0	0	1	-360	360;
	2363	599	0.00464	0.027021	0	0	0	0	0	0	1	-360	360;
	5390	3859	0.002211	0.012031	0	591	0	0	0	0	1	-360	360;
	5390	4064	0.002479	0.0115	0	572	0	0	0	0	1	-360	360;
	5534	8577	0.030281	0.058314	0	0	0	0	0	0	1	-360	360;
	5534	7721	0.028174	0.054256	0	0	0	0	0	0	1	-360	360;
	2579	8577	0.01114	0.026231	0	0	0	0	0	0	1	-360	360;
	2579	2178	0.018157	0.045248	0	0	0	0	0	0	1	-360	360;
	6365	353	0.00369	0.017159	0	0	0	0	0	0	1	-360	360;
	6365	4810	0.004909	0.03018	0	0	0	0	0	0	1	-360	360;
	8641	7285	0.00408	0.0056	0	1251	0	0	0	0	1	-360	360;
	8641	7674	0.00301	0.00925	0	1283	0	0	0	0	1	-360	360;
	4677	353	0.001599	0.00489	0	0	0	0	0	0	1	-360	360;
	4677	6182	0.00031	0.00175	0	400	0	0	0	0	1	-360	360;
	6154	353	0.00163	0.004789	0	495	0	0	0	0	1	-360	360;
	6154	6182	0.00031	0.001789	0	400	0	0	0	0	1	-360	360;
	3046	3236	0.00988	0.063229	0	400	0	0	0	0	1	-360	360;
	3046	1575	0.00114	0.00638	0	400	0	0	0	0	1	-360	360;
	681	7081	0.00065	0.00755	0	1086	0	0	0	0	1	-360	360;
	681	6907	0.00052	0.00503	0	987	0	0	0	0	1	-360	360;
	567	7081	0.00063	0.0077	0	1086	0	0	0	0	1	-360	360;
	567	6907	0.00053	0.00493	0	921	0	0	0	0	1	-360	360;
	2371	3343	0.00055	0.00683	0	1645	0	0	0	0	1	-360	360;
	2371	8263	0.00118	0.01134	0	1382	0	0	0	0	1	-360	360;
	6482	3343	0.00056	0.00668	0	1942	0	0	0	0	1	-360	360;
	6482	8263	0.0012	0.0111	0	1349	0	0	0	0	1	-360	360;
	6637	8894	0.0098	0.01047	0	1020	0	0	0	0	1	-360	360;
	6637	3208	0.00134	0.01215	0	1086	0	0	0	0	1	-360	360;
	6491	8894	0.00092	0.01198	0	1283	0	0	0	0	1	-360	360;
	6491	2585	0.00057	0.00616	0	1283	0	0	0	0	1	-360	360;
	8282	7455	0.00108	0.01773	0	1547	0	0	0	0	1	-360	360;
	8282	3729	0.00064	0.01047	0	757	0	0	0	0	1	-360	360;
	359	9048	0.002211	0.013341	0	572	0	0	0	0	1	-360	360;
	359	96	0.001901	0.014229	0	591	0	0	0	0	1	-360	360;
	6818	743	0.00193	0.011771	0	400	0	0	0	0	1	-360	360;
	6818	8496	0.00251	0.014909	0	438	0	0	0	0	1	-360	360;
	5038	611	6e-05	0.00084	0	1448	0	0	0	0	1	-360	360;
	5038	1249	6e-05	0.00092	0	1448	0	0	0	0	1	-360	360;
	8428	5905	0.00138	0.010349	0	533	0	0	0	0	1	-360	360;
	8428	7464	0.00156	0.009151	0	533	0	0	0	0	1	-360	360;
	6312	5905	0.001349	0.010539	0	553	0	0	0	0	1	-360	360;
	6312	7464	0.001591	0.00895	0	514	0	0	0	0	1	-360	360;
	5852	4129	0	0.000271	0	0	0	0	0	0	1	-360	360;
	4363	4507	0.00057	0.0071	0	1613	0	0	0	0	1	-360	360;
	4363	6246	0.00057	0.00697	0	1613	0	0	0	0	1	-360	360;
	8431	4122	0.000969	0.00743	0	629	0	0	0	0	1	-360	360;
	8431	4844	0.00099	0.007271	0	629	0	0	0	0	1	-360	360;
	170	4122	0.00088	0.00824	0	629	0	0	0	0	1	-360	360;
	170	4844	0.00086	0.008401	0	629	0	0	0	0	1	-360	360;
	625	3790	0.00293	0.01664	0	400	0	0	0	0	1	-360	360;
	625	7978	0.008719	0.05249	0	381	0	0	0	0	1	-360	360;
	3308	1723	0.00057	0.00627	0	856	0	0	0	0	1	-360	360;
	3308	7870	0.00088	0.00878	0	790	0	0	0	0	1	-360	360;
	9042	221	0.00033	0.00421	0	1909	0	0	0	0	1	-360	360;
	9042	8623	0.00084	0.00353	0	0	0	0	0	0	1	-360	360;
	425	7526	0.0001	0.00119	0	1810	0	0	0	0	1	-360	360;
	425	5674	0.0001	0.0013	0	1613	0	0	0	0	1	-360	360;
	4799	8581	9e-05	0.00121	0	1514	0	0	0	0	1	-360	360;
	4799	4337	0.0001	0.00134	0	1251	0	0	0	0	1	-360	360;
	6590	3025	0.00063	0.01067	0	1514	0	0	0	0	1	-360	360;
	6590	2420	0.00073	0.00959	0	1514	0	0	0	0	1	-360	360;
	7276	3887	0.002021	0.030271	0	0	0	0	0	0	1	-360	360;
	7276	5088	0.00218	0.02786	0	0	0	0	0	0	1	-360	360;
	4923	3025	0.00086	0.00781	0	1580	0	0	0	0	1	-360	360;
	4923	4276	0.00092	0.00666	0	1514	0	0	0	0	1	-360	360;
	5601	3887	0.003969	0.0272	0	0	0	0	0	0	1	-360	360;
	5601	5958	0.00311	0.029729	0	0	0	0	0	0	1	-360	360;
	5211	3025	0.00055	0.00837	0	1283	0	0	0	0	1	-360	360;
	5211	9235	0.00053	0.0084	0	1415	0	0	0	0	1	-360	360;
	5385	3025	0.00053	0.00853	0	1316	0	0	0	0	1	-360	360;
	5385	9235	0.00052	0.00778	0	1415	0	0	0	0	1	-360	360;
	7719	3887	0.0017	0.008159	0	553	0	0	0	0	1	-360	360;
	7719	7843	0.001729	0.00799	0	0	0	0	0	0	1	-360	360;
	2269	3887	0.00176	0.00781	0	591	0	0	0	0	1	-360	360;
	2269	7843	0.00157	0.008849	0	0	0	0	0	0	1	-360	360;
	786	3887	0.000219	0.002229	0	0	0	0	0	0	1	-360	360;
	786	5145	0.000229	0.00218	0	0	0	0	0	0	1	-360	360;
	1951	3887	0.000229	0.00213	0	0	0	0	0	0	1	-360	360;
	1951	5145	0.000211	0.002419	0	0	0	0	0	0	1	-360	360;
	4430	3887	0.00019	0.00255	0	0	0	0	0	0	1	-360	360;
	4430	5145	0.000211	0.0025	0	0	0	0	0	0	1	-360	360;
	8221	3887	0	0.0002	0	495	0	0	0	0	1	-360	360;
	8221	2175	0	0.0002	0	495	0	0	0	0	1	-360	360;
	1933	7992	0.00067	0.01058	0	1316	0	0	0	0	1	-360	360;
	1933	7458	0.00065	0.01078	0	1185	0	0	0	0	1	-360	360;
	3710	265	0.009281	0.05299	0	0	0	0	0	0	1	-360	360;
	3710	7679	0.009419	0.057219	0	0	0	0	0	0	1	-360	360;
	5183	2392	0.00055	0.00505	0	0	0	0	0	0	1	-360	360;
	5183	6297	9.9e-05	0.00164	0	0	0	0	0	0	1	-360	360;
	7686	7992	0.00064	0.00998	0	1152	0	0	0	0	1	-360	360;
	7686	2368	0.00216	0.02943	0	1020	0	0	0	0	1	-360	360;
	4968	2154	0.00039	0.00474	0	1843	0	0	0	0	1	-360	360;
	4968	1747	0.00045	0.00467	0	1448	0	0	0	0	1	-360	360;
	7699	2154	0.00052	0.00567	0	1645	0	0	0	0	1	-360	360;
	7699	1689	0.00054	0.00667	0	1316	0	0	0	0	1	-360	360;
	4470	4417	0.008909	0.042349	0	0	0	0	0	0	1	-360	360;
	4470	8325	0.00713	0.03293	0	0	0	0	0	0	1	-360	360;
	8883	1429	0.0021	0.02015	0	1448	0	0	0	0	1	-360	360;
	8883	9183	0.00177	0.02084	0	1448	0	0	0	0	1	-360	360;
	2495	7393	0.001461	0.00781	0	0	0	0	0	0	1	-360	360;
	2495	5911	0.001651	0.007651	0	0	0	0	0	0	1	-360	360;
	8674	7393	0.00168	0.0075	0	0	0	0	0	0	1	-360	360;
	8674	5911	0.001719	0.007341	0	0	0	0	0	0	1	-360	360;
	7931	6967	0.00626	0.020771	0	381	0	0	0	0	1	-360	360;
	7931	6967	0.00638	0.02032	0	400	0	0	0	0	1	-360	360;
	4635	6314	0.047983	0.076364	0	0	0	0	0	0	1	-360	360;
	4635	1567	0.010868	0.039083	0	0	0	0	0	0	1	-360	360;
	5475	3472	0.021851	0.032364	0	0	0	0	0	0	1	-360	360;
	4322	3472	0.075008	0.121636	0	0	0	0	0	0	1	-360	360;
	3472	5172	0.049273	0.084942	0	0	0	0	0	0	1	-360	360;
	7150	7131	0.058521	0.110017	0	0	0	0	0	0	1	-360	360;
	7517	6627	0.012992	0.052719	0	0	0	0	0	0	1	-360	360;
	6627	7517	0.012727	0.053752	0	0	0	0	0	0	1	-360	360;
	5326	2474	0.013341	0.054169	0	0	0	0	0	0	1	-360	360;
	4848	5326	0.014961	0.083521	0	419	0	0	0	0	1	-360	360;
	7141	1649	0.023653	0.033603	0	0	0	0	0	0	1	-360	360;
	7141	1649	0.024099	0.032876	0	0	0	0	0	0	1	-360	360;
	1649	7517	0.024934	0.109579	0	0	0	0	0	0	1	-360	360;
	1649	7517	0.024397	0.111653	0	0	0	0	0	0	1	-360	360;
	1649	6023	0.03414	0.051058	0	0	0	0	0	0	1	-360	360;
	6314	5475	0.018893	0.029083	0	0	0	0	0	0	1	-360	360;
	2474	5657	0.000581	0.003081	0	476	0	0	0	0	1	-360	360;
	2474	5657	0.0005	0.002789	0	495	0	0	0	0	1	-360	360;
	2474	6505	0.009281	0.03489	0	381	0	0	0	0	1	-360	360;
	5989	2474	0.006081	0.034	0	438	0	0	0	0	1	-360	360;
	5989	2474	0.00595	0.034651	0	419	0	0	0	0	1	-360	360;
	1700	7141	0.05019	0.076339	0	0	0	0	0	0	1	-360	360;
	7141	3806	0.05724	0.101868	0	0	0	0	0	0	1	-360	360;
	8258	7313	0.0013	0.01496	0	921	0	0	0	0	1	-360	360;
	1544	8258	0.00108	0.01115	0	1415	0	0	0	0	1	-360	360;
	8258	106	0.00223	0.02267	0	1349	0	0	0	0	1	-360	360;
	7931	6505	0.008229	0.03325	0	400	0	0	0	0	1	-360	360;
	7931	5976	0.018159	0.06526	0	419	0	0	0	0	1	-360	360;
	4322	9171	0.02257	0.034777	0	0	0	0	0	0	1	-360	360;
	4322	2639	0.032339	0.103017	0	0	0	0	0	0	1	-360	360;
	4322	2639	0.031678	0.105041	0	0	0	0	0	0	1	-360	360;
	4322	833	0.028331	0.050182	0	0	0	0	0	0	1	-360	360;
	4322	6314	0.047314	0.063678	0	0	0	0	0	0	1	-360	360;
	4848	7204	0.0045	0.02618	0	419	0	0	0	0	1	-360	360;
	4848	5989	0.01913	0.10681	0	381	0	0	0	0	1	-360	360;
	8295	7150	0.059273	0.08086	0	0	0	0	0	0	1	-360	360;
	7981	4155	0.008041	0.014298	0	0	0	0	0	0	1	-360	360;
	4649	7981	0.047802	0.073562	0	0	0	0	0	0	1	-360	360;
	3722	1700	0.049636	0.058818	0	0	0	0	0	0	1	-360	360;
	9171	1567	0.04738	0.075992	0	0	0	0	0	0	1	-360	360;
	6023	8826	0.035603	0.045331	0	0	0	0	0	0	1	-360	360;
	2897	5172	0.024529	0.031678	0	0	0	0	0	0	1	-360	360;
	8891	2897	0.036851	0.047521	0	0	0	0	0	0	1	-360	360;
	6348	8339	0.004058	0.011041	0	0	0	0	0	0	1	-360	360;
	4625	6348	0.012818	0.022554	0	0	0	0	0	0	1	-360	360;
	4548	1652	0.079256	0.118479	0	0	0	0	0	0	1	-360	360;
	4625	4548	0.012554	0.022967	0	0	0	0	0	0	1	-360	360;
	8339	8102	0.007917	0.029628	0	0	0	0	0	0	1	-360	360;
	8339	8102	0.007744	0.027438	0	0	0	0	0	0	1	-360	360;
	1544	2500	0.00039	0.00461	0	757	0	0	0	0	1	-360	360;
	7204	273	0.014039	0.071669	0	438	0	0	0	0	1	-360	360;
	7204	273	0.01251	0.081219	0	476	0	0	0	0	1	-360	360;
	273	5976	0.0045	0.016831	0	0	0	0	0	0	1	-360	360;
	5989	273	0.00261	0.009401	0	381	0	0	0	0	1	-360	360;
	7205	2639	0.08024	0.135628	0	0	0	0	0	0	1	-360	360;
	4329	7205	0.043587	0.057314	0	0	0	0	0	0	1	-360	360;
	7204	80	0.00326	0.01895	0	476	0	0	0	0	1	-360	360;
	743	80	0.005409	0.02675	0	400	0	0	0	0	1	-360	360;
	4155	3380	0.083537	0.116554	0	0	0	0	0	0	1	-360	360;
	2546	3380	0.012463	0.054793	0	0	0	0	0	0	1	-360	360;
	833	8826	0.044992	0.07586	0	0	0	0	0	0	1	-360	360;
	7155	1652	0.032777	0.050628	0	0	0	0	0	0	1	-360	360;
	3806	4649	0.060314	0.092818	0	0	0	0	0	0	1	-360	360;
	8891	8295	0.02462	0.034967	0	0	0	0	0	0	1	-360	360;
	8102	3722	0.026603	0.03038	0	0	0	0	0	0	1	-360	360;
	40	5123	0.01595	0.022281	0	0	0	0	0	0	1	-360	360;
	7155	5123	0.018083	0.021843	0	0	0	0	0	0	1	-360	360;
	80	5989	0.0132	0.06293	0	400	0	0	0	0	1	-360	360;
	80	5989	0.01345	0.06162	0	400	0	0	0	0	1	-360	360;
	4625	3722	0.010612	0.023017	0	0	0	0	0	0	1	-360	360;
	995	7548	0.01601	0.06706	0	419	0	0	0	0	1	-360	360;
	995	5594	0.00362	0.01457	0	457	0	0	0	0	1	-360	360;
	8101	4775	0.010281	0.04007	0	419	0	0	0	0	1	-360	360;
	8101	9184	0.01549	0.06756	0	381	0	0	0	0	1	-360	360;
	1816	5177	0.00073	0.01141	0	1876	0	0	0	0	1	-360	360;
	1816	7992	0.00108	0.01698	0	1481	0	0	0	0	1	-360	360;
	2867	515	0.0022	0.027409	0	381	0	0	0	0	1	-360	360;
	2867	515	0.002151	0.02794	0	381	0	0	0	0	1	-360	360;
	2867	7597	0.00445	0.02582	0	629	0	0	0	0	1	-360	360;
	2867	265	0.005099	0.02426	0	629	0	0	0	0	1	-360	360;
	2867	1726	0.019539	0.100021	0	0	0	0	0	0	1	-360	360;
	2867	1726	0.01914	0.101979	0	0	0	0	0	0	1	-360	360;
	5177	2203	0.00158	0.02551	0	1514	0	0	0	0	1	-360	360;
	5177	4681	0.00021	0.00296	0	1382	0	0	0	0	1	-360	360;
	5177	5652	0.00157	0.02091	0	1580	0	0	0	0	1	-360	360;
	1363	4681	0.00162	0.02163	0	1481	0	0	0	0	1	-360	360;
	7638	4775	0.008521	0.03463	0	0	0	0	0	0	1	-360	360;
	7638	9184	0.00463	0.078531	0	0	0	0	0	0	1	-360	360;
	7638	9184	0.004531	0.072711	0	0	0	0	0	0	1	-360	360;
	7638	111	0.00575	0.02876	0	381	0	0	0	0	1	-360	360;
	7638	3152	0.00411	0.01826	0	438	0	0	0	0	1	-360	360;
	7638	1726	0.008091	0.03594	0	381	0	0	0	0	1	-360	360;
	3826	2026	0.011031	0.060581	0	419	0	0	0	0	1	-360	360;
	3826	5999	0.01326	0.07163	0	419	0	0	0	0	1	-360	360;
	5069	1829	0.007031	0.069341	0	0	0	0	0	0	1	-360	360;
	5069	5594	0.00238	0.031081	0	0	0	0	0	0	1	-360	360;
	4775	9184	0.010831	0.057479	0	457	0	0	0	0	1	-360	360;
	4775	111	0.001849	0.01024	0	381	0	0	0	0	1	-360	360;
	7548	6762	0.017581	0.07664	0	495	0	0	0	0	1	-360	360;
	7548	6762	0.017909	0.07501	0	495	0	0	0	0	1	-360	360;
	7548	5999	0.01312	0.05724	0	381	0	0	0	0	1	-360	360;
	7548	5999	0.01337	0.056021	0	400	0	0	0	0	1	-360	360;
	7548	555	0.007961	0.0347	0	419	0	0	0	0	1	-360	360;
	5480	9184	0.01181	0.044729	0	381	0	0	0	0	1	-360	360;
	5480	123	0.00457	0.018729	0	0	0	0	0	0	1	-360	360;
	8433	8228	0.0012	0.0141	0	1283	0	0	0	0	1	-360	360;
	8433	8228	0.00107	0.01598	0	1283	0	0	0	0	1	-360	360;
	8433	2933	0.00046	0.00666	0	889	0	0	0	0	1	-360	360;
	8433	2772	0.001	0.0181	0	757	0	0	0	0	1	-360	360;
	1145	3702	9.1e-05	0.00139	0	0	0	0	0	0	1	-360	360;
	1145	3702	0.00013	0.00089	0	0	0	0	0	0	1	-360	360;
	7597	265	0.000281	0.001289	0	572	0	0	0	0	1	-360	360;
	6962	3508	0.00119	0.01964	0	1119	0	0	0	0	1	-360	360;
	6962	3508	0.00117	0.02001	0	1185	0	0	0	0	1	-360	360;
	1829	4207	0.00238	0.032341	0	0	0	0	0	0	1	-360	360;
	1829	4207	0.002331	0.02994	0	0	0	0	0	0	1	-360	360;
	1829	4842	0.005479	0.07155	0	0	0	0	0	0	1	-360	360;
	1952	265	0.008091	0.036841	0	381	0	0	0	0	1	-360	360;
	5594	4842	0.003081	0.03707	0	0	0	0	0	0	1	-360	360;
	5594	555	0.01086	0.05936	0	438	0	0	0	0	1	-360	360;
	2933	958	0.00063	0.00975	0	823	0	0	0	0	1	-360	360;
	9184	8609	0.00143	0.025771	0	0	0	0	0	0	1	-360	360;
	9184	8609	0.001401	0.026271	0	0	0	0	0	0	1	-360	360;
	9184	123	0.00631	0.03037	0	400	0	0	0	0	1	-360	360;
	2026	5999	0.00325	0.01388	0	400	0	0	0	0	1	-360	360;
	8609	3702	0.015461	0.0828	0	419	0	0	0	0	1	-360	360;
	8609	3702	0.017521	0.08114	0	419	0	0	0	0	1	-360	360;
	5652	7992	0.00036	0.00554	0	1514	0	0	0	0	1	-360	360;
	2772	7992	0.0018	0.02089	0	1645	0	0	0	0	1	-360	360;
	958	7992	0.00183	0.02045	0	1613	0	0	0	0	1	-360	360;
	3508	458	0.00052	0.01489	0	921	0	0	0	0	1	-360	360;
	3508	458	0.00051	0.01518	0	757	0	0	0	0	1	-360	360;
	5847	5999	0.020669	0.110081	0	0	0	0	0	0	1	-360	360;
	5847	5999	0.02024	0.1122	0	0	0	0	0	0	1	-360	360;
	298	265	0.01089	0.05145	0	591	0	0	0	0	1	-360	360;
	298	265	0.01111	0.050401	0	553	0	0	0	0	1	-360	360;
	3152	1726	0.00493	0.026719	0	0	0	0	0	0	1	-360	360;
	3887	7393	0.005169	0.067461	0	0	0	0	0	0	1	-360	360;
	3887	7393	0.00506	0.068729	0	0	0	0	0	0	1	-360	360;
	1904	6876	0.011397	0.04438	0	0	0	0	0	0	1	-360	360;
	1926	8412	0.022504	0.082851	0	0	0	0	0	0	1	-360	360;
	1926	3936	0.015793	0.061587	0	0	0	0	0	0	1	-360	360;
	5718	2613	0.012659	0.05405	0	381	0	0	0	0	1	-360	360;
	730	4010	0.016149	0.057248	0	0	0	0	0	0	1	-360	360;
	154	4010	0.038612	0.122471	0	0	0	0	0	0	1	-360	360;
	4010	7187	0.044273	0.108959	0	0	0	0	0	0	1	-360	360;
	8850	5942	0.038876	0.139587	0	0	0	0	0	0	1	-360	360;
	8850	6402	0.10805	0.191992	0	0	0	0	0	0	1	-360	360;
	3093	8850	0.035802	0.084231	0	0	0	0	0	0	1	-360	360;
	3383	5168	0.00614	0.02205	0	0	0	0	0	0	1	-360	360;
	3383	6128	0.003587	0.01395	0	0	0	0	0	0	1	-360	360;
	5168	7768	0.019554	0.05981	0	0	0	0	0	0	1	-360	360;
	6128	3316	0.038066	0.142281	0	0	0	0	0	0	1	-360	360;
	2704	4326	0.007421	0.02181	0	0	0	0	0	0	1	-360	360;
	4326	8176	0.007562	0.021339	0	0	0	0	0	0	1	-360	360;
	2704	7589	0.007388	0.026537	0	0	0	0	0	0	1	-360	360;
	7589	1674	0.00724	0.02705	0	0	0	0	0	0	1	-360	360;
	73	7086	0.004893	0.018273	0	0	0	0	0	0	1	-360	360;
	7086	5551	0.004777	0.018612	0	0	0	0	0	0	1	-360	360;
	2704	6346	0.015289	0.041463	0	0	0	0	0	0	1	-360	360;
	2704	6346	0.01143	0.042091	0	0	0	0	0	0	1	-360	360;
	73	2704	0.006793	0.014182	0	0	0	0	0	0	1	-360	360;
	5551	2704	0.012967	0.048471	0	0	0	0	0	0	1	-360	360;
	6140	6346	0.009281	0.029554	0	0	0	0	0	0	1	-360	360;
	6184	6346	0.041248	0.116479	0	0	0	0	0	0	1	-360	360;
	3801	6346	0.071479	0.137678	0	0	0	0	0	0	1	-360	360;
	2321	6346	0.020612	0.075893	0	0	0	0	0	0	1	-360	360;
	6346	514	0.05214	0.100421	0	0	0	0	0	0	1	-360	360;
	4767	6346	0.047942	0.179157	0	0	0	0	0	0	1	-360	360;
	7262	2512	0.0016	0.01402	0	1119	0	0	0	0	1	-360	360;
	5910	6140	0.009149	0.027992	0	0	0	0	0	0	1	-360	360;
	8622	5910	0.031107	0.116281	0	0	0	0	0	0	1	-360	360;
	8622	5551	0.026463	0.091273	0	0	0	0	0	0	1	-360	360;
	406	8622	0.023562	0.075017	0	0	0	0	0	0	1	-360	360;
	4038	8622	0.015702	0.050017	0	0	0	0	0	0	1	-360	360;
	4385	3316	0.004802	0.017231	0	0	0	0	0	0	1	-360	360;
	4385	1219	0.065579	0.208826	0	0	0	0	0	0	1	-360	360;
	4385	3490	0.049917	0.112843	0	0	0	0	0	0	1	-360	360;
	4385	136	0.047793	0.095603	0	0	0	0	0	0	1	-360	360;
	7390	3316	0.043421	0.168942	0	0	0	0	0	0	1	-360	360;
	7768	3316	0.029364	0.093017	0	0	0	0	0	0	1	-360	360;
	2911	3820	0.023521	0.086562	0	0	0	0	0	0	1	-360	360;
	974	3820	0.017909	0.058463	0	0	0	0	0	0	1	-360	360;
	3757	535	0.008198	0.030248	0	0	0	0	0	0	1	-360	360;
	535	3105	0.015579	0.053744	0	0	0	0	0	0	1	-360	360;
	535	6003	0.030471	0.09738	0	0	0	0	0	0	1	-360	360;
	2811	535	0.017157	0.066	0	0	0	0	0	0	1	-360	360;
	1273	4503	0.017628	0.065868	0	0	0	0	0	0	1	-360	360;
	1273	5173	0.022686	0.072273	0	0	0	0	0	0	1	-360	360;
	2219	4307	0.010322	0.03719	0	0	0	0	0	0	1	-360	360;
	4307	7531	0.010802	0.038281	0	0	0	0	0	0	1	-360	360;
	4982	7187	0.019298	0.073041	0	0	0	0	0	0	1	-360	360;
	3209	4982	0.011678	0.035273	0	0	0	0	0	0	1	-360	360;
	5064	4503	0.019719	0.057256	0	0	0	0	0	0	1	-360	360;
	175	730	0.026421	0.086471	0	0	0	0	0	0	1	-360	360;
	8176	1916	0.014281	0.045471	0	0	0	0	0	0	1	-360	360;
	974	7038	0.004529	0.013686	0	0	0	0	0	0	1	-360	360;
	7038	7531	0.014603	0.058463	0	0	0	0	0	0	1	-360	360;
	6722	5579	0.007702	0.020876	0	0	0	0	0	0	1	-360	360;
	6722	3467	0.010281	0.03957	0	0	0	0	0	0	1	-360	360;
	5579	5064	0.00686	0.023661	0	0	0	0	0	0	1	-360	360;
	3757	5070	0.02457	0.09557	0	0	0	0	0	0	1	-360	360;
	19	1219	0.03024	0.085372	0	0	0	0	0	0	1	-360	360;
	19	7390	0.04157	0.14924	0	0	0	0	0	0	1	-360	360;
	6029	19	0.028132	0.097008	0	0	0	0	0	0	1	-360	360;
	19	8667	0.038661	0.123091	0	0	0	0	0	0	1	-360	360;
	1304	8718	0.019364	0.077413	0	0	0	0	0	0	1	-360	360;
	8718	3936	0.009562	0.036182	0	0	0	0	0	0	1	-360	360;
	7262	5625	0.00147	0.01643	0	1415	0	0	0	0	1	-360	360;
	701	6216	0.005719	0.016826	0	0	0	0	0	0	1	-360	360;
	6216	9177	0.007843	0.027041	0	0	0	0	0	0	1	-360	360;
	6216	1290	0.016826	0.045628	0	0	0	0	0	0	1	-360	360;
	6216	2205	0.02238	0.062777	0	0	0	0	0	0	1	-360	360;
	1219	701	0.012959	0.041248	0	0	0	0	0	0	1	-360	360;
	9177	4387	0.004702	0.017562	0	0	0	0	0	0	1	-360	360;
	7713	9177	0.023331	0.080446	0	0	0	0	0	0	1	-360	360;
	1496	974	0.00505	0.018653	0	0	0	0	0	0	1	-360	360;
	4356	1496	0.008562	0.025686	0	0	0	0	0	0	1	-360	360;
	3491	2958	0.026124	0.083149	0	0	0	0	0	0	1	-360	360;
	2958	4767	0.02038	0.04605	0	0	0	0	0	0	1	-360	360;
	7723	6602	0.000169	0.00106	0	0	0	0	0	0	1	-360	360;
	963	7723	0.00043	0.003901	0	0	0	0	0	0	1	-360	360;
	7968	4636	0.005298	0.021	0	0	0	0	0	0	1	-360	360;
	4636	7292	0.001322	0.004289	0	0	0	0	0	0	1	-360	360;
	6431	1112	0.005471	0.016554	0	0	0	0	0	0	1	-360	360;
	6431	4878	0.022198	0.064653	0	0	0	0	0	0	1	-360	360;
	4791	2761	0.027653	0.099256	0	0	0	0	0	0	1	-360	360;
	2625	4791	0.044099	0.152107	0	0	0	0	0	0	1	-360	360;
	2994	3167	0.017603	0.047711	0	0	0	0	0	0	1	-360	360;
	2994	4356	0.024198	0.06562	0	0	0	0	0	0	1	-360	360;
	8371	3622	0.00165	0.01428	0	1283	0	0	0	0	1	-360	360;
	7285	8371	0.00131	0.01467	0	1152	0	0	0	0	1	-360	360;
	5942	8941	0.040802	0.129909	0	0	0	0	0	0	1	-360	360;
	945	2761	0.005058	0.019686	0	0	0	0	0	0	1	-360	360;
	2761	1388	0.075521	0.145463	0	0	0	0	0	0	1	-360	360;
	945	5196	0.024033	0.088488	0	0	0	0	0	0	1	-360	360;
	7568	4935	0.001599	0.007401	0	438	0	0	0	0	1	-360	360;
	1290	8412	0.025289	0.082934	0	0	0	0	0	0	1	-360	360;
	8837	7390	0.012421	0.035058	0	0	0	0	0	0	1	-360	360;
	3236	2738	0.00855	0.03774	0	457	0	0	0	0	1	-360	360;
	6184	5207	0.010686	0.03019	0	0	0	0	0	0	1	-360	360;
	4172	6176	0.010331	0.03438	0	0	0	0	0	0	1	-360	360;
	6176	6547	0.01581	0.050339	0	0	0	0	0	0	1	-360	360;
	7968	4878	0.005661	0.022744	0	0	0	0	0	0	1	-360	360;
	5834	8398	0.000636	0.002512	0	0	0	0	0	0	1	-360	360;
	3936	5834	0.005281	0.017521	0	0	0	0	0	0	1	-360	360;
	5834	2823	0.002826	0.00781	0	0	0	0	0	0	1	-360	360;
	1112	4996	0.018066	0.056876	0	0	0	0	0	0	1	-360	360;
	4996	5173	0.038537	0.144033	0	0	0	0	0	0	1	-360	360;
	4996	1252	0.049669	0.141083	0	0	0	0	0	0	1	-360	360;
	4996	7292	0.058628	0.17286	0	0	0	0	0	0	1	-360	360;
	4996	6054	0.00043	0.001273	0	0	0	0	0	0	1	-360	360;
	4996	6896	0.025917	0.093058	0	0	0	0	0	0	1	-360	360;
	1733	4172	0.02295	0.065	0	0	0	0	0	0	1	-360	360;
	5975	5173	0.001504	0.00486	0	0	0	0	0	0	1	-360	360;
	5975	5173	0.002182	0.006678	0	0	0	0	0	0	1	-360	360;
	3405	963	0.002979	0.02218	0	0	0	0	0	0	1	-360	360;
	3405	963	0.003021	0.022531	0	0	0	0	0	0	1	-360	360;
	406	4038	0.020843	0.058843	0	0	0	0	0	0	1	-360	360;
	5070	406	0.009331	0.027413	0	0	0	0	0	0	1	-360	360;
	406	6913	0.009182	0.024769	0	0	0	0	0	0	1	-360	360;
	5285	353	0.00275	0.01488	0	476	0	0	0	0	1	-360	360;
	6688	2992	0.0001	0.00082	0	1152	0	0	0	0	1	-360	360;
	5930	963	0.01476	0.08713	0	0	0	0	0	0	1	-360	360;
	5930	4935	0.00212	0.011469	0	400	0	0	0	0	1	-360	360;
	7292	6595	0.013521	0.049959	0	0	0	0	0	0	1	-360	360;
	6595	974	0.017793	0.07124	0	0	0	0	0	0	1	-360	360;
	8612	2613	0.01481	0.086021	0	419	0	0	0	0	1	-360	360;
	4935	8612	0.01686	0.07207	0	457	0	0	0	0	1	-360	360;
	7531	5660	0.012331	0.034413	0	0	0	0	0	0	1	-360	360;
	5745	5660	0.000777	0.00276	0	0	0	0	0	0	1	-360	360;
	3222	154	0.028421	0.08657	0	0	0	0	0	0	1	-360	360;
	1916	4618	0.023521	0.081132	0	0	0	0	0	0	1	-360	360;
	3574	2620	0.000901	0.005	0	438	0	0	0	0	1	-360	360;
	2620	3098	0.00263	0.01551	0	419	0	0	0	0	1	-360	360;
	7176	3179	0.017463	0.071669	0	0	0	0	0	0	1	-360	360;
	7176	8805	0.016826	0.065512	0	0	0	0	0	0	1	-360	360;
	3167	3856	0.019587	0.076612	0	0	0	0	0	0	1	-360	360;
	1382	1304	0.022537	0.071769	0	0	0	0	0	0	1	-360	360;
	3209	175	0.013256	0.04005	0	0	0	0	0	0	1	-360	360;
	7255	3222	0.046397	0.13586	0	0	0	0	0	0	1	-360	360;
	7255	573	0.026471	0.105893	0	0	0	0	0	0	1	-360	360;
	8557	7241	0.025504	0.08119	0	0	0	0	0	0	1	-360	360;
	4934	8557	0.027603	0.10738	0	0	0	0	0	0	1	-360	360;
	7187	3801	0.084843	0.156942	0	0	0	0	0	0	1	-360	360;
	5235	3105	0.030521	0.115653	0	0	0	0	0	0	1	-360	360;
	1674	8031	0.004	0.010851	0	0	0	0	0	0	1	-360	360;
	8031	5207	0.016017	0.043421	0	0	0	0	0	0	1	-360	360;
	5235	1904	0.027091	0.087488	0	0	0	0	0	0	1	-360	360;
	7421	6938	0.02243	0.047793	0	0	0	0	0	0	1	-360	360;
	7768	6938	0.016851	0.051537	0	0	0	0	0	0	1	-360	360;
	8398	8412	0.01638	0.053901	0	0	0	0	0	0	1	-360	360;
	8398	8412	0.012298	0.045521	0	0	0	0	0	0	1	-360	360;
	8412	6866	0.025017	0.048653	0	0	0	0	0	0	1	-360	360;
	3467	8914	0.00186	0.007174	0	0	0	0	0	0	1	-360	360;
	2969	8914	0.001397	0.005372	0	0	0	0	0	0	1	-360	360;
	2205	4387	0.010777	0.043479	0	0	0	0	0	0	1	-360	360;
	6378	2205	0.016413	0.047628	0	0	0	0	0	0	1	-360	360;
	1451	5516	0.004	0.011281	0	0	0	0	0	0	1	-360	360;
	5173	5516	0.004041	0.015099	0	0	0	0	0	0	1	-360	360;
	3467	8447	0.004636	0.017579	0	0	0	0	0	0	1	-360	360;
	8447	5173	0.006529	0.025091	0	0	0	0	0	0	1	-360	360;
	6547	3467	0.01781	0.052339	0	0	0	0	0	0	1	-360	360;
	1451	2969	0.003744	0.010149	0	0	0	0	0	0	1	-360	360;
	6650	5173	0.006248	0.023702	0	0	0	0	0	0	1	-360	360;
	6650	6402	0.035107	0.106207	0	0	0	0	0	0	1	-360	360;
	6650	8972	0.014769	0.045182	0	0	0	0	0	0	1	-360	360;
	6650	3398	0.004322	0.012198	0	0	0	0	0	0	1	-360	360;
	6650	6896	0.015678	0.054083	0	0	0	0	0	0	1	-360	360;
	35	8837	0.011331	0.023636	0	0	0	0	0	0	1	-360	360;
	963	6602	0.000591	0.002969	0	400	0	0	0	0	1	-360	360;
	963	6602	0.000281	0.0025	0	0	0	0	0	0	1	-360	360;
	4935	6602	0.00781	0.039159	0	0	0	0	0	0	1	-360	360;
	6602	1125	0.00419	0.02431	0	419	0	0	0	0	1	-360	360;
	353	6602	0.014479	0.06989	0	381	0	0	0	0	1	-360	360;
	353	6602	0.013419	0.07594	0	400	0	0	0	0	1	-360	360;
	3622	4022	0.00208	0.02183	0	1053	0	0	0	0	1	-360	360;
	7241	3801	0.028083	0.103397	0	0	0	0	0	0	1	-360	360;
	6003	2219	0.005793	0.017802	0	0	0	0	0	0	1	-360	360;
	8941	8972	0.016124	0.049322	0	0	0	0	0	0	1	-360	360;
	35	3490	0.018471	0.041802	0	0	0	0	0	0	1	-360	360;
	1388	1068	0.031719	0.066198	0	0	0	0	0	0	1	-360	360;
	209	1733	0.017554	0.056124	0	0	0	0	0	0	1	-360	360;
	6089	1252	0.011207	0.037917	0	0	0	0	0	0	1	-360	360;
	2613	3574	0.02261	0.105521	0	0	0	0	0	0	1	-360	360;
	3801	4934	0.002397	0.013603	0	0	0	0	0	0	1	-360	360;
	4934	514	0.021901	0.055818	0	0	0	0	0	0	1	-360	360;
	3491	3801	0.007719	0.030066	0	0	0	0	0	0	1	-360	360;
	3801	2321	0.017711	0.05	0	0	0	0	0	0	1	-360	360;
	573	3801	0.035322	0.070802	0	0	0	0	0	0	1	-360	360;
	4618	6029	0.017281	0.050777	0	0	0	0	0	0	1	-360	360;
	4596	5235	0.04324	0.168223	0	0	0	0	0	0	1	-360	360;
	6054	4596	0.022678	0.082017	0	0	0	0	0	0	1	-360	360;
	4596	2811	0.026	0.079537	0	0	0	0	0	0	1	-360	360;
	5404	6899	0.027347	0.080372	0	0	0	0	0	0	1	-360	360;
	5196	5404	0.023364	0.074372	0	0	0	0	0	0	1	-360	360;
	8577	3093	0.011421	0.02757	0	0	0	0	0	0	1	-360	360;
	4935	3574	0.011039	0.047169	0	0	0	0	0	0	1	-360	360;
	1125	3574	0.015159	0.0732	0	381	0	0	0	0	1	-360	360;
	2682	3574	0.00119	0.005729	0	381	0	0	0	0	1	-360	360;
	7695	3574	0.00182	0.060349	0	0	0	0	0	0	1	-360	360;
	7292	3856	0.00795	0.032132	0	0	0	0	0	0	1	-360	360;
	855	7292	0.006372	0.021934	0	0	0	0	0	0	1	-360	360;
	2625	7292	0.015992	0.036041	0	0	0	0	0	0	1	-360	360;
	7292	4356	0.022843	0.087157	0	0	0	0	0	0	1	-360	360;
	7591	2911	0.00495	0.014942	0	0	0	0	0	0	1	-360	360;
	1068	7591	0.000901	0.00238	0	0	0	0	0	0	1	-360	360;
	3856	7591	0.008942	0.027339	0	0	0	0	0	0	1	-360	360;
	3856	2911	0.00476	0.017587	0	0	0	0	0	0	1	-360	360;
	6899	2911	0.037339	0.128802	0	0	0	0	0	0	1	-360	360;
	1068	4905	0.00081	0.00219	0	0	0	0	0	0	1	-360	360;
	1068	4439	0.005132	0.018438	0	0	0	0	0	0	1	-360	360;
	4367	3856	0.005934	0.023107	0	0	0	0	0	0	1	-360	360;
	4367	3856	0.005802	0.021397	0	0	0	0	0	0	1	-360	360;
	4367	226	0.000463	0.001421	0	0	0	0	0	0	1	-360	360;
	4367	6089	0.000909	0.002521	0	0	0	0	0	0	1	-360	360;
	3856	4439	0.009347	0.025364	0	0	0	0	0	0	1	-360	360;
	9175	4905	0.00062	0.002017	0	0	0	0	0	0	1	-360	360;
	226	9175	0.000917	0.002364	0	0	0	0	0	0	1	-360	360;
	9175	2625	0.009248	0.016446	0	0	0	0	0	0	1	-360	360;
	6089	2671	0.00295	0.008	0	0	0	0	0	0	1	-360	360;
	2625	6089	0.003339	0.012612	0	0	0	0	0	0	1	-360	360;
	2625	2671	0.00319	0.012504	0	0	0	0	0	0	1	-360	360;
	855	2625	0.007826	0.030521	0	0	0	0	0	0	1	-360	360;
	7105	4022	0.00063	0.00702	0	987	0	0	0	0	1	-360	360;
	4022	5625	0.00184	0.02018	0	1086	0	0	0	0	1	-360	360;
	1453	6074	0.003818	0.010661	0	0	0	0	0	0	1	-360	360;
	7531	1453	0.008678	0.025198	0	0	0	0	0	0	1	-360	360;
	5745	1453	0.008843	0.024661	0	0	0	0	0	0	1	-360	360;
	136	6866	0.023347	0.041909	0	0	0	0	0	0	1	-360	360;
	8667	5207	0.018091	0.055347	0	0	0	0	0	0	1	-360	360;
	7258	1382	0.03443	0.093347	0	0	0	0	0	0	1	-360	360;
	7258	8805	0.041033	0.111248	0	0	0	0	0	0	1	-360	360;
	8205	7713	0.015843	0.05686	0	0	0	0	0	0	1	-360	360;
	4618	8205	0.020397	0.064959	0	0	0	0	0	0	1	-360	360;
	209	5235	0.017893	0.054959	0	0	0	0	0	0	1	-360	360;
	974	6074	0.029496	0.09295	0	0	0	0	0	0	1	-360	360;
	9229	8398	0.011579	0.046397	0	0	0	0	0	0	1	-360	360;
	7695	3236	0.00099	0.0042	0	381	0	0	0	0	1	-360	360;
	7695	3236	0.00076	0.00449	0	381	0	0	0	0	1	-360	360;
	8841	1428	0.006587	0.016868	0	0	0	0	0	0	1	-360	360;
	1428	6074	0.016876	0.064909	0	0	0	0	0	0	1	-360	360;
	8841	5070	0.036612	0.115372	0	0	0	0	0	0	1	-360	360;
	3179	3936	0.009421	0.030826	0	0	0	0	0	0	1	-360	360;
	9229	3936	0.008041	0.030438	0	0	0	0	0	0	1	-360	360;
	5070	6913	0.000744	0.002182	0	0	0	0	0	0	1	-360	360;
	7695	4354	0.000901	0.003979	0	0	0	0	0	0	1	-360	360;
	7695	3098	0.01212	0.05176	0	381	0	0	0	0	1	-360	360;
	6688	5625	0.00256	0.02765	0	1316	0	0	0	0	1	-360	360;
	2823	7421	0.00224	0.007256	0	0	0	0	0	0	1	-360	360;
	8894	5625	0.0009	0.00943	0	1316	0	0	0	0	1	-360	360;
	3179	9229	0.007744	0.024273	0	0	0	0	0	0	1	-360	360;
	7421	9229	0.006372	0.017992	0	0	0	0	0	0	1	-360	360;
	8805	7768	0.048554	0.142669	0	0	0	0	0	0	1	-360	360;
	3398	8972	0.010802	0.030488	0	0	0	0	0	0	1	-360	360;
	6378	4575	0.015496	0.047397	0	0	0	0	0	0	1	-360	360;
	7531	6074	0.010207	0.030826	0	0	0	0	0	0	1	-360	360;
	6074	4575	0.026504	0.077876	0	0	0	0	0	0	1	-360	360;
	3097	659	0.00073	0.00905	0	1185	0	0	0	0	1	-360	360;
	9024	4929	0.00192	0.02296	0	889	0	0	0	0	1	-360	360;
	1815	6542	0.01431	0.05899	0	0	0	0	0	0	1	-360	360;
	6542	1445	0.02607	0.07168	0	0	0	0	0	0	1	-360	360;
	6069	6833	0.00054	0.00465	0	1580	0	0	0	0	1	-360	360;
	6069	2268	0.00126	0.01124	0	1613	0	0	0	0	1	-360	360;
	6069	7180	0.00048	0.00527	0	1547	0	0	0	0	1	-360	360;
	6069	6293	0.00082	0.00938	0	1580	0	0	0	0	1	-360	360;
	1968	9192	0.00028	0.003502	0	0	0	0	0	0	1	-360	360;
	1968	3493	0.691467	4.414711	0	0	0	0	0	0	1	-360	360;
	1968	955	0.02004	0.0916	0	0	0	0	0	0	1	-360	360;
	1968	228	0.841911	4.768444	0	0	0	0	0	0	1	-360	360;
	1968	6826	0.026	0.08948	0	0	0	0	0	0	1	-360	360;
	1968	8964	0.002631	0.011889	0	299	0	0	0	0	1	-360	360;
	1968	5776	0.7232	1.944	0	0	0	0	0	0	1	-360	360;
	9192	955	0.156169	0.546756	0	0	0	0	0	0	1	-360	360;
	9192	6826	0.168809	0.580622	0	0	0	0	0	0	1	-360	360;
	9192	8964	0.019222	0.068209	0	0	0	0	0	0	1	-360	360;
	8574	1616	0.00039	0.00386	0	1415	0	0	0	0	1	-360	360;
	8574	1163	0.00096	0.01138	0	1448	0	0	0	0	1	-360	360;
	8574	9064	0.00037	0.00325	0	1415	0	0	0	0	1	-360	360;
	8574	8921	0.00029	0.00339	0	1580	0	0	0	0	1	-360	360;
	3493	2299	0.009458	0.158871	0	0	0	0	0	0	1	-360	360;
	2299	4423	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6826	2299	0.312951	7.130667	0	0	0	0	0	0	1	-360	360;
	7563	2299	0.247649	1.075911	0	0	0	0	0	0	1	-360	360;
	4427	2299	0.06056	0.18612	0	0	0	0	0	0	1	-360	360;
	8335	2299	2.142489	4.347733	0	0	0	0	0	0	1	-360	360;
	5155	2299	0.07132	1.068889	0	0	0	0	0	0	1	-360	360;
	271	2299	0.02268	0.08652	0	0	0	0	0	0	1	-360	360;
	2449	2299	2.007778	4.494222	0	0	0	0	0	0	1	-360	360;
	4665	2299	0.006178	0.028938	0	0	0	0	0	0	1	-360	360;
	4495	2299	0.027711	0.260951	0	0	0	0	0	0	1	-360	360;
	3493	5587	0	0.000222	0	299	0	0	0	0	1	-360	360;
	4586	3493	0.00692	0.337378	0	0	0	0	0	0	1	-360	360;
	228	3493	0.025991	0.10048	0	0	0	0	0	0	1	-360	360;
	6826	3493	0.083711	0.8648	0	0	0	0	0	0	1	-360	360;
	7563	3493	0.447378	1.862089	0	0	0	0	0	0	1	-360	360;
	4427	3493	0.062169	0.153191	0	0	0	0	0	0	1	-360	360;
	8335	3493	2.492533	4.725778	0	0	0	0	0	0	1	-360	360;
	5155	3493	0.020271	0.08244	0	0	0	0	0	0	1	-360	360;
	271	3493	0.11416	1.777556	0	0	0	0	0	0	1	-360	360;
	5776	3493	0.574978	2.992578	0	0	0	0	0	0	1	-360	360;
	2449	3493	1.887067	4.964444	0	0	0	0	0	0	1	-360	360;
	4665	3493	0.250778	0.817422	0	0	0	0	0	0	1	-360	360;
	4495	3493	0.007049	0.029609	0	0	0	0	0	0	1	-360	360;
	5587	4586	0.003049	0.020218	0	0	0	0	0	0	1	-360	360;
	4423	4586	0.002911	0.020262	0	0	0	0	0	0	1	-360	360;
	4929	2870	4e-05	0.00041	0	757	0	0	0	0	1	-360	360;
	4929	659	0.00087	0.01087	0	1283	0	0	0	0	1	-360	360;
	4929	1037	4e-05	0.00046	0	757	0	0	0	0	1	-360	360;
	4929	659	0.00095	0.00924	0	1481	0	0	0	0	1	-360	360;
	1815	5210	0.528017	2.636157	0	0	0	0	0	0	1	-360	360;
	1815	1445	0.032031	0.128659	0	0	0	0	0	0	1	-360	360;
	1616	7762	0.00044	0.0032	0	1645	0	0	0	0	1	-360	360;
	1616	2520	0.00046	0.00489	0	1711	0	0	0	0	1	-360	360;
	9064	7762	0.0003	0.00352	0	1580	0	0	0	0	1	-360	360;
	9064	89	0.00051	0.00454	0	1645	0	0	0	0	1	-360	360;
	1317	659	0.00068	0.00746	0	1613	0	0	0	0	1	-360	360;
	1317	8605	0.00093	0.01022	0	1744	0	0	0	0	1	-360	360;
	8605	1367	4e-05	0.0004	0	1481	0	0	0	0	1	-360	360;
	8605	8921	0.00078	0.00698	0	1744	0	0	0	0	1	-360	360;
	1163	1676	4e-05	0.00043	0	1580	0	0	0	0	1	-360	360;
	1163	7051	0.00025	0.00309	0	1678	0	0	0	0	1	-360	360;
	913	7762	0.00047	0.00538	0	1613	0	0	0	0	1	-360	360;
	2107	7762	0.00046	0.00548	0	1744	0	0	0	0	1	-360	360;
	2107	5996	0.00016	0.00173	0	1843	0	0	0	0	1	-360	360;
	2107	6293	0.00166	0.01471	0	1744	0	0	0	0	1	-360	360;
	913	7762	0.00051	0.00512	0	1777	0	0	0	0	1	-360	360;
	6704	8921	0.00021	0.00248	0	1514	0	0	0	0	1	-360	360;
	228	4586	0.004391	0.021689	0	286	0	0	0	0	1	-360	360;
	2441	6293	0.00024	0.00289	0	1645	0	0	0	0	1	-360	360;
	955	3506	0	0.000222	0	338	0	0	0	0	1	-360	360;
	955	8964	0.01788	0.100471	0	0	0	0	0	0	1	-360	360;
	3506	2908	0.006458	0.039982	0	0	0	0	0	0	1	-360	360;
	228	6826	0.106151	0.993333	0	0	0	0	0	0	1	-360	360;
	228	5776	0.699778	3.232311	0	0	0	0	0	0	1	-360	360;
	8964	6826	0.257058	0.43592	0	0	0	0	0	0	1	-360	360;
	5776	6826	0.002769	0.011449	0	0	0	0	0	0	1	-360	360;
	659	8329	0.00034	0.00305	0	1580	0	0	0	0	1	-360	360;
	659	7051	0.00123	0.01255	0	1843	0	0	0	0	1	-360	360;
	659	9239	6e-05	0.00065	0	1415	0	0	0	0	1	-360	360;
	659	7960	6e-05	0.00063	0	1514	0	0	0	0	1	-360	360;
	659	6233	6e-05	0.00055	0	1744	0	0	0	0	1	-360	360;
	659	5416	0.00079	0.00968	0	1349	0	0	0	0	1	-360	360;
	659	6798	7e-05	0.00079	0	0	0	0	0	0	1	-360	360;
	792	3279	0	0.000222	0	377	0	0	0	0	1	-360	360;
	7563	792	2.064178	7.545333	0	0	0	0	0	0	1	-360	360;
	7279	792	0.153711	0.482489	0	0	0	0	0	0	1	-360	360;
	8335	792	0.015809	0.12368	0	0	0	0	0	0	1	-360	360;
	2449	792	0.057978	0.363422	0	0	0	0	0	0	1	-360	360;
	4665	792	0.6932	2.196889	0	0	0	0	0	0	1	-360	360;
	4495	792	0.560356	2.395867	0	0	0	0	0	0	1	-360	360;
	5416	2267	0.00024	0.00213	0	757	0	0	0	0	1	-360	360;
	5416	7637	0.00093	0.00884	0	1283	0	0	0	0	1	-360	360;
	8964	2168	0	0.000222	0	312	0	0	0	0	1	-360	360;
	3659	5996	0.00018	0.00204	0	1777	0	0	0	0	1	-360	360;
	5762	5996	0.00018	0.00208	0	757	0	0	0	0	1	-360	360;
	3242	5097	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7563	3242	0.02068	0.093391	0	0	0	0	0	0	1	-360	360;
	4427	3242	0.777778	7.408	0	0	0	0	0	0	1	-360	360;
	3242	7279	0.00884	0.07292	0	0	0	0	0	0	1	-360	360;
	271	3242	0.131671	8.295556	0	0	0	0	0	0	1	-360	360;
	3242	1611	0.082462	0.568222	0	0	0	0	0	0	1	-360	360;
	5097	1611	0.0042	0.02624	0	0	0	0	0	0	1	-360	360;
	8181	8179	0.00031	0.00317	0	1251	0	0	0	0	1	-360	360;
	8181	7762	0.00043	0.00498	0	1481	0	0	0	0	1	-360	360;
	9025	7762	0.00042	0.00508	0	1711	0	0	0	0	1	-360	360;
	7563	4427	0.110671	0.773067	0	0	0	0	0	0	1	-360	360;
	7563	7279	0.00912	0.037609	0	0	0	0	0	0	1	-360	360;
	7563	8335	0.098049	0.598711	0	0	0	0	0	0	1	-360	360;
	7563	5155	0.019031	1.160356	0	0	0	0	0	0	1	-360	360;
	7563	271	0.024071	0.915733	0	0	0	0	0	0	1	-360	360;
	2449	7563	1.484267	4.479111	0	0	0	0	0	0	1	-360	360;
	4665	7563	0.011578	0.063271	0	0	0	0	0	0	1	-360	360;
	4495	7563	0.654889	1.839111	0	0	0	0	0	0	1	-360	360;
	7563	1611	0.022049	0.10076	0	0	0	0	0	0	1	-360	360;
	4427	7279	0.459867	2.604711	0	0	0	0	0	0	1	-360	360;
	4427	8335	2.320222	6.088	0	0	0	0	0	0	1	-360	360;
	4427	5155	0.006369	0.048089	0	0	0	0	0	0	1	-360	360;
	4427	271	0.006631	0.0432	0	0	0	0	0	0	1	-360	360;
	2449	4427	2.567244	6.767111	0	0	0	0	0	0	1	-360	360;
	4665	4427	0.309889	1.214178	0	0	0	0	0	0	1	-360	360;
	4495	4427	0.009191	0.043289	0	0	0	0	0	0	1	-360	360;
	4427	1611	0.691022	5.239556	0	0	0	0	0	0	1	-360	360;
	5210	1445	0.002669	0.01718	0	0	0	0	0	0	1	-360	360;
	8179	5509	0.00074	0.00694	0	1251	0	0	0	0	1	-360	360;
	8179	7762	0.00068	0.008	0	1547	0	0	0	0	1	-360	360;
	7279	4014	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8335	7279	0.00524	0.030631	0	0	0	0	0	0	1	-360	360;
	5155	7279	0.33932	3.814667	0	0	0	0	0	0	1	-360	360;
	271	7279	0.22696	2.614933	0	0	0	0	0	0	1	-360	360;
	2449	7279	0.190671	0.719378	0	0	0	0	0	0	1	-360	360;
	4665	7279	1.050222	2.105556	0	0	0	0	0	0	1	-360	360;
	4495	7279	1.431644	4.294578	0	0	0	0	0	0	1	-360	360;
	1611	7279	0.022618	0.161089	0	0	0	0	0	0	1	-360	360;
	5509	1579	0.00088	0.00895	0	1810	0	0	0	0	1	-360	360;
	8335	1531	0	0.000222	0	299	0	0	0	0	1	-360	360;
	5155	8335	1.827778	7.768	0	0	0	0	0	0	1	-360	360;
	2449	8335	0.020849	0.153382	0	0	0	0	0	0	1	-360	360;
	4665	8335	0.156231	0.656267	0	0	0	0	0	0	1	-360	360;
	4495	8335	0.176498	0.891556	0	0	0	0	0	0	1	-360	360;
	7762	2268	0.00067	0.00777	0	1613	0	0	0	0	1	-360	360;
	5155	271	0.004578	0.061889	0	0	0	0	0	0	1	-360	360;
	5155	2908	0.01348	0.05932	0	0	0	0	0	0	1	-360	360;
	4665	5155	0.066391	2.802622	0	0	0	0	0	0	1	-360	360;
	4495	5155	0.194258	2.220444	0	0	0	0	0	0	1	-360	360;
	5155	1611	0.008151	0.072822	0	0	0	0	0	0	1	-360	360;
	271	2908	0.016502	0.055382	0	0	0	0	0	0	1	-360	360;
	4665	271	0.0806	4.633778	0	0	0	0	0	0	1	-360	360;
	4495	271	0.831556	3.729644	0	0	0	0	0	0	1	-360	360;
	271	1611	0.007951	0.074151	0	0	0	0	0	0	1	-360	360;
	1579	5848	0.0004	0.00362	0	1613	0	0	0	0	1	-360	360;
	7051	8847	0.00015	0.00177	0	1580	0	0	0	0	1	-360	360;
	5776	8229	0	0.000222	0	0	0	0	0	0	1	-360	360;
	317	6233	2e-05	0.00022	0	1678	0	0	0	0	1	-360	360;
	4665	2449	0.02864	0.172902	0	0	0	0	0	0	1	-360	360;
	2449	4495	0.017591	0.1404	0	0	0	0	0	0	1	-360	360;
	4665	4495	0.00752	0.031711	0	0	0	0	0	0	1	-360	360;
	1611	8420	0	0.000222	0	546	0	0	0	0	1	-360	360;
	8768	2283	0.032587	0.058388	0	0	0	0	0	0	1	-360	360;
	636	2283	0.029752	0.070686	0	0	0	0	0	0	1	-360	360;
	4388	4008	0.00699	0.04244	0	476	0	0	0	0	1	-360	360;
	4008	7532	0.0008	0.006081	0	0	0	0	0	0	1	-360	360;
	4008	6006	0.004669	0.041919	0	0	0	0	0	0	1	-360	360;
	4008	6006	0.00476	0.04106	0	0	0	0	0	0	1	-360	360;
	6999	2041	0.00146	0.0163	0	1316	0	0	0	0	1	-360	360;
	6999	2041	0.00098	0.01485	0	1218	0	0	0	0	1	-360	360;
	6999	2041	0.00143	0.01661	0	1119	0	0	0	0	1	-360	360;
	1506	6999	0.00209	0.0241	0	1909	0	0	0	0	1	-360	360;
	1506	6999	0.00273	0.02677	0	1448	0	0	0	0	1	-360	360;
	1506	6999	0.00279	0.02623	0	1251	0	0	0	0	1	-360	360;
	5521	815	0.007479	0.032207	0	0	0	0	0	0	1	-360	360;
	5521	2127	0.04319	0.16186	0	0	0	0	0	0	1	-360	360;
	1617	3294	0.007256	0.029702	0	0	0	0	0	0	1	-360	360;
	3294	6055	0.006488	0.036041	0	0	0	0	0	0	1	-360	360;
	3788	4558	0.028653	0.066364	0	0	0	0	0	0	1	-360	360;
	7112	3788	0.010463	0.027446	0	0	0	0	0	0	1	-360	360;
	1438	3173	0.012409	0.071919	0	400	0	0	0	0	1	-360	360;
	1438	3173	0.012659	0.07045	0	438	0	0	0	0	1	-360	360;
	3173	9125	0.003419	0.02344	0	419	0	0	0	0	1	-360	360;
	3173	7432	0.002469	0.01486	0	0	0	0	0	0	1	-360	360;
	3173	5497	0.00682	0.03368	0	457	0	0	0	0	1	-360	360;
	3173	921	0.02036	0.09743	0	0	0	0	0	0	1	-360	360;
	8768	3765	0.048	0.080331	0	0	0	0	0	0	1	-360	360;
	8768	3420	0.00262	0.008083	0	0	0	0	0	0	1	-360	360;
	8768	6537	0.004107	0.023083	0	0	0	0	0	0	1	-360	360;
	8768	7991	0.003207	0.015719	0	0	0	0	0	0	1	-360	360;
	8768	716	0.004727	0.02143	0	0	0	0	0	0	1	-360	360;
	8768	3861	0.004041	0.012702	0	0	0	0	0	0	1	-360	360;
	8768	2559	0.017802	0.037777	0	0	0	0	0	0	1	-360	360;
	2922	5424	0.002934	0.014273	0	0	0	0	0	0	1	-360	360;
	3641	2922	0.006231	0.032174	0	0	0	0	0	0	1	-360	360;
	6051	2937	0.026876	0.085248	0	0	0	0	0	0	1	-360	360;
	6051	6367	0.050636	0.099223	0	0	0	0	0	0	1	-360	360;
	979	3329	0.045347	0.092	0	0	0	0	0	0	1	-360	360;
	5804	3329	0.005231	0.013504	0	0	0	0	0	0	1	-360	360;
	1499	8659	0.024752	0.057446	0	0	0	0	0	0	1	-360	360;
	1499	7932	0.015083	0.065182	0	0	0	0	0	0	1	-360	360;
	1499	8925	0.03986	0.070686	0	0	0	0	0	0	1	-360	360;
	5198	1499	0.036231	0.064248	0	0	0	0	0	0	1	-360	360;
	1499	4684	0.004537	0.022041	0	0	0	0	0	0	1	-360	360;
	9111	7560	0.052603	0.157802	0	0	0	0	0	0	1	-360	360;
	8925	7560	0.005562	0.024083	0	0	0	0	0	0	1	-360	360;
	8925	7560	0.005446	0.024554	0	0	0	0	0	0	1	-360	360;
	3614	7560	0.009107	0.051661	0	0	0	0	0	0	1	-360	360;
	7560	761	0.00419	0.009868	0	0	0	0	0	0	1	-360	360;
	4518	7560	0.024281	0.13338	0	0	0	0	0	0	1	-360	360;
	4518	7560	0.02376	0.123496	0	0	0	0	0	0	1	-360	360;
	5015	7560	0.005331	0.020752	0	0	0	0	0	0	1	-360	360;
	4570	7068	0.001107	0.004603	0	0	0	0	0	0	1	-360	360;
	7068	7823	0.005628	0.030702	0	0	0	0	0	0	1	-360	360;
	716	8108	0.024628	0.130893	0	0	0	0	0	0	1	-360	360;
	716	8108	0.024132	0.133463	0	0	0	0	0	0	1	-360	360;
	5962	8108	0.01419	0.07062	0	0	0	0	0	0	1	-360	360;
	5962	8108	0.014471	0.069182	0	0	0	0	0	0	1	-360	360;
	6761	8408	0.008372	0.045231	0	0	0	0	0	0	1	-360	360;
	6942	6761	0.024628	0.097983	0	0	0	0	0	0	1	-360	360;
	6761	7087	0.002471	0.009331	0	0	0	0	0	0	1	-360	360;
	337	1873	0.025702	0.144587	0	0	0	0	0	0	1	-360	360;
	4184	337	0.003421	0.019752	0	0	0	0	0	0	1	-360	360;
	3534	45	0.004562	0.017174	0	0	0	0	0	0	1	-360	360;
	45	915	0.018868	0.083661	0	0	0	0	0	0	1	-360	360;
	45	4877	0.01019	0.027917	0	0	0	0	0	0	1	-360	360;
	8830	3534	0.019471	0.064058	0	0	0	0	0	0	1	-360	360;
	6258	1681	0.053446	0.118207	0	0	0	0	0	0	1	-360	360;
	1681	5539	0.049521	0.123438	0	0	0	0	0	0	1	-360	360;
	2677	5055	0.00914	0.042039	0	0	0	0	0	0	1	-360	360;
	5055	921	0.004711	0.021591	0	0	0	0	0	0	1	-360	360;
	1450	9237	0	0.000893	0	0	0	0	0	0	1	-360	360;
	4693	1450	0.00405	0.021132	0	0	0	0	0	0	1	-360	360;
	4693	1450	0.004587	0.020711	0	0	0	0	0	0	1	-360	360;
	9237	3683	0.044182	0.091041	0	0	0	0	0	0	1	-360	360;
	9237	4536	0.022653	0.047421	0	0	0	0	0	0	1	-360	360;
	9237	7484	0.058876	0.108463	0	0	0	0	0	0	1	-360	360;
	457	3572	0.013174	0.026992	0	0	0	0	0	0	1	-360	360;
	7336	457	0.001273	0.002818	0	0	0	0	0	0	1	-360	360;
	1698	1617	0.009471	0.048769	0	0	0	0	0	0	1	-360	360;
	5966	3474	0.017388	0.04138	0	0	0	0	0	0	1	-360	360;
	3474	8897	0.017339	0.054661	0	0	0	0	0	0	1	-360	360;
	5424	4605	0.025033	0.137942	0	0	0	0	0	0	1	-360	360;
	5424	4605	0.024521	0.140653	0	0	0	0	0	0	1	-360	360;
	4605	7181	0.021198	0.056041	0	0	0	0	0	0	1	-360	360;
	4605	8174	0.007174	0.035	0	0	0	0	0	0	1	-360	360;
	2678	4605	0.010826	0.029992	0	0	0	0	0	0	1	-360	360;
	2678	4605	0.012876	0.028058	0	0	0	0	0	0	1	-360	360;
	2415	7744	0.034248	0.143124	0	0	0	0	0	0	1	-360	360;
	6033	7744	0.015421	0.066364	0	0	0	0	0	0	1	-360	360;
	5198	8659	0.014512	0.035901	0	0	0	0	0	0	1	-360	360;
	1119	1438	0.00863	0.054419	0	533	0	0	0	0	1	-360	360;
	1119	1438	0.008461	0.055479	0	514	0	0	0	0	1	-360	360;
	480	4280	0.013256	0.073248	0	0	0	0	0	0	1	-360	360;
	8978	4280	0.004033	0.018174	0	0	0	0	0	0	1	-360	360;
	636	3765	0.016008	0.027066	0	0	0	0	0	0	1	-360	360;
	5223	3614	0.013537	0.070083	0	0	0	0	0	0	1	-360	360;
	7793	5223	0.004744	0.024876	0	0	0	0	0	0	1	-360	360;
	3420	2776	0.014364	0.037438	0	0	0	0	0	0	1	-360	360;
	8319	2249	0.010364	0.020314	0	0	0	0	0	0	1	-360	360;
	2249	3669	0.007289	0.018207	0	0	0	0	0	0	1	-360	360;
	8360	1500	0.011744	0.052446	0	0	0	0	0	0	1	-360	360;
	7456	8360	0.015438	0.062876	0	0	0	0	0	0	1	-360	360;
	9010	6568	0.00157	0.01612	0	1020	0	0	0	0	1	-360	360;
	9010	6568	0.00154	0.01643	0	1020	0	0	0	0	1	-360	360;
	3641	5424	0.001967	0.008438	0	0	0	0	0	0	1	-360	360;
	5424	8272	0.01162	0.033653	0	0	0	0	0	0	1	-360	360;
	1143	888	0.006901	0.04661	0	419	0	0	0	0	1	-360	360;
	1143	888	0.00782	0.04568	0	438	0	0	0	0	1	-360	360;
	2865	888	0.005401	0.0343	0	381	0	0	0	0	1	-360	360;
	665	888	0.011419	0.07275	0	381	0	0	0	0	1	-360	360;
	4388	6006	0.01044	0.05919	0	476	0	0	0	0	1	-360	360;
	8923	1073	0.050314	0.112132	0	0	0	0	0	0	1	-360	360;
	8923	1073	0.057017	0.109893	0	0	0	0	0	0	1	-360	360;
	1073	4275	0.007	0.028231	0	0	0	0	0	0	1	-360	360;
	1073	8446	0.010719	0.040868	0	0	0	0	0	0	1	-360	360;
	1073	4668	0.04443	0.10976	0	0	0	0	0	0	1	-360	360;
	1073	568	0.021372	0.121033	0	0	0	0	0	0	1	-360	360;
	1073	884	0.020579	0.095331	0	0	0	0	0	0	1	-360	360;
	2914	8923	0.02386	0.06986	0	0	0	0	0	0	1	-360	360;
	8852	2914	0.062901	0.105628	0	0	0	0	0	0	1	-360	360;
	2914	2127	0.035686	0.149314	0	0	0	0	0	0	1	-360	360;
	6502	2914	0.00181	0.006876	0	0	0	0	0	0	1	-360	360;
	6501	2914	0.002628	0.01462	0	0	0	0	0	0	1	-360	360;
	6501	2914	0.002983	0.014331	0	0	0	0	0	0	1	-360	360;
	422	2914	0.006157	0.03114	0	0	0	0	0	0	1	-360	360;
	2022	7461	0.022	0.043066	0	0	0	0	0	0	1	-360	360;
	1416	2022	0.013438	0.033479	0	0	0	0	0	0	1	-360	360;
	2022	8549	0.013488	0.028893	0	0	0	0	0	0	1	-360	360;
	3421	9178	0.114058	0.230033	0	0	0	0	0	0	1	-360	360;
	9178	944	0.062901	0.133777	0	0	0	0	0	0	1	-360	360;
	9178	1981	0.041868	0.106554	0	0	0	0	0	0	1	-360	360;
	3415	8315	0.052347	0.07786	0	0	0	0	0	0	1	-360	360;
	8315	5570	0.003099	0.006174	0	0	0	0	0	0	1	-360	360;
	889	7340	0.04357	0.108678	0	0	0	0	0	0	1	-360	360;
	889	8173	0.020893	0.043289	0	0	0	0	0	0	1	-360	360;
	4558	8897	0.035289	0.072058	0	0	0	0	0	0	1	-360	360;
	2127	8923	0.050579	0.074298	0	0	0	0	0	0	1	-360	360;
	4069	8923	0.042934	0.087917	0	0	0	0	0	0	1	-360	360;
	815	7121	0.036711	0.086397	0	0	0	0	0	0	1	-360	360;
	7121	8319	0.053124	0.100066	0	0	0	0	0	0	1	-360	360;
	7121	4065	0.022959	0.041727	0	0	0	0	0	0	1	-360	360;
	3967	1641	0.00476	0.023868	0	0	0	0	0	0	1	-360	360;
	7737	3967	0.005529	0.011182	0	0	0	0	0	0	1	-360	360;
	7572	727	0.00183	0.02821	0	1678	0	0	0	0	1	-360	360;
	7455	7572	0.00198	0.02074	0	1218	0	0	0	0	1	-360	360;
	8880	5954	0.022917	0.094893	0	0	0	0	0	0	1	-360	360;
	8880	5954	0.023347	0.092868	0	0	0	0	0	0	1	-360	360;
	5954	2998	0.030364	0.073256	0	0	0	0	0	0	1	-360	360;
	5954	6453	0.004562	0.020132	0	0	0	0	0	0	1	-360	360;
	5954	6453	0.010562	0.020686	0	0	0	0	0	0	1	-360	360;
	944	5954	0.01838	0.07314	0	0	0	0	0	0	1	-360	360;
	5954	5027	0.005628	0.029702	0	0	0	0	0	0	1	-360	360;
	5954	5027	0.005777	0.029273	0	0	0	0	0	0	1	-360	360;
	2813	5954	0.014529	0.236802	0	0	0	0	0	0	1	-360	360;
	2582	1737	0.005818	0.032182	0	0	0	0	0	0	1	-360	360;
	2582	4693	0.003636	0.021843	0	0	0	0	0	0	1	-360	360;
	3440	2041	0.00318	0.02624	0	1119	0	0	0	0	1	-360	360;
	3440	7455	0.00442	0.03649	0	1251	0	0	0	0	1	-360	360;
	4599	1647	0.005099	0.027868	0	0	0	0	0	0	1	-360	360;
	4599	1737	0.00995	0.049397	0	0	0	0	0	0	1	-360	360;
	4599	1737	0.01014	0.048397	0	0	0	0	0	0	1	-360	360;
	4599	1682	0.007917	0.04757	0	0	0	0	0	0	1	-360	360;
	4599	93	0.004008	0.022182	0	0	0	0	0	0	1	-360	360;
	4599	206	0.006603	0.039124	0	0	0	0	0	0	1	-360	360;
	8319	5637	0.00657	0.028281	0	0	0	0	0	0	1	-360	360;
	7340	8319	0.071471	0.124661	0	0	0	0	0	0	1	-360	360;
	7340	8319	0.069983	0.127058	0	0	0	0	0	0	1	-360	360;
	8319	3669	0.00743	0.018182	0	0	0	0	0	0	1	-360	360;
	8830	7796	0.004653	0.024339	0	0	0	0	0	0	1	-360	360;
	8897	8830	0.007157	0.03462	0	0	0	0	0	0	1	-360	360;
	8897	8830	0.007298	0.033909	0	0	0	0	0	0	1	-360	360;
	6780	7777	0.00024	0.00281	0	1218	0	0	0	0	1	-360	360;
	5408	7246	0.010711	0.056603	0	0	0	0	0	0	1	-360	360;
	3746	5408	0.064273	0.068058	0	0	0	0	0	0	1	-360	360;
	8168	5583	0.013917	0.06676	0	0	0	0	0	0	1	-360	360;
	8168	8599	0.034298	0.076091	0	0	0	0	0	0	1	-360	360;
	4570	3151	0.00676	0.044942	0	0	0	0	0	0	1	-360	360;
	4570	4184	0.025579	0.109438	0	0	0	0	0	0	1	-360	360;
	4570	4184	0.026091	0.107198	0	0	0	0	0	0	1	-360	360;
	3309	3387	0.00319	0.01543	0	0	0	0	0	0	1	-360	360;
	3387	2763	0.003669	0.014628	0	0	0	0	0	0	1	-360	360;
	2879	1585	0.001653	0.007331	0	0	0	0	0	0	1	-360	360;
	480	2879	0.009661	0.04095	0	0	0	0	0	0	1	-360	360;
	1506	9010	0.00089	0.00981	0	1152	0	0	0	0	1	-360	360;
	1506	9010	0.00087	0.01	0	1316	0	0	0	0	1	-360	360;
	9010	7777	0.00114	0.01867	0	1448	0	0	0	0	1	-360	360;
	4894	2892	0.034661	0.089231	0	0	0	0	0	0	1	-360	360;
	1098	4894	0.009463	0.018603	0	0	0	0	0	0	1	-360	360;
	2831	7101	0.032777	0.061248	0	0	0	0	0	0	1	-360	360;
	6867	2831	0.029182	0.043207	0	0	0	0	0	0	1	-360	360;
	6758	5966	0.012826	0.022587	0	0	0	0	0	0	1	-360	360;
	1967	9111	0.01519	0.065017	0	0	0	0	0	0	1	-360	360;
	390	9111	0.002107	0.009	0	0	0	0	0	0	1	-360	360;
	9111	4518	0.010099	0.046587	0	0	0	0	0	0	1	-360	360;
	206	1647	0.001934	0.009008	0	0	0	0	0	0	1	-360	360;
	6258	7461	0.041537	0.116843	0	0	0	0	0	0	1	-360	360;
	6942	7932	0.013091	0.050041	0	0	0	0	0	0	1	-360	360;
	1084	7101	0.003058	0.005579	0	0	0	0	0	0	1	-360	360;
	1677	8880	0.008554	0.035397	0	0	0	0	0	0	1	-360	360;
	9210	5456	0.020314	0.102727	0	0	0	0	0	0	1	-360	360;
	5081	5456	0.005008	0.025364	0	0	0	0	0	0	1	-360	360;
	5539	5456	0.042182	0.161132	0	0	0	0	0	0	1	-360	360;
	7790	1314	0.00014	0.000581	0	0	0	0	0	0	1	-360	360;
	3437	2163	0.001339	0.007438	0	0	0	0	0	0	1	-360	360;
	2461	1585	0.004223	0.009347	0	0	0	0	0	0	1	-360	360;
	1585	5565	0.016314	0.034562	0	0	0	0	0	0	1	-360	360;
	8978	1585	0.009331	0.045339	0	0	0	0	0	0	1	-360	360;
	2461	4009	0.00038	0.002769	0	0	0	0	0	0	1	-360	360;
	8978	2461	0.007066	0.043107	0	0	0	0	0	0	1	-360	360;
	1843	1967	0.003124	0.017231	0	0	0	0	0	0	1	-360	360;
	2561	5467	0.020512	0.048562	0	0	0	0	0	0	1	-360	360;
	4069	5467	0.016934	0.03314	0	0	0	0	0	0	1	-360	360;
	9026	3641	0.011083	0.030769	0	0	0	0	0	0	1	-360	360;
	480	6599	0.024314	0.099174	0	0	0	0	0	0	1	-360	360;
	480	6599	0.024769	0.097058	0	0	0	0	0	0	1	-360	360;
	480	6599	0.025231	0.09495	0	0	0	0	0	0	1	-360	360;
	480	5159	0.012917	0.049612	0	0	0	0	0	0	1	-360	360;
	480	3281	0.048058	0.135157	0	0	0	0	0	0	1	-360	360;
	480	4009	0.009372	0.048876	0	0	0	0	0	0	1	-360	360;
	480	5457	0.00538	0.031752	0	0	0	0	0	0	1	-360	360;
	480	5457	0.005273	0.029397	0	0	0	0	0	0	1	-360	360;
	480	8910	0.044562	0.09762	0	0	0	0	0	0	1	-360	360;
	6599	3844	0.017876	0.097091	0	0	0	0	0	0	1	-360	360;
	6599	3547	0.037207	0.166008	0	0	0	0	0	0	1	-360	360;
	7455	8885	0.00018	0.00154	0	987	0	0	0	0	1	-360	360;
	7455	8885	0.00023	0.00194	0	889	0	0	0	0	1	-360	360;
	1263	2419	0.005669	0.040521	0	514	0	0	0	0	1	-360	360;
	1263	4788	0.001581	0.012159	0	381	0	0	0	0	1	-360	360;
	1263	4788	0.001409	0.013781	0	400	0	0	0	0	1	-360	360;
	1263	3505	0.00111	0.0075	0	438	0	0	0	0	1	-360	360;
	1263	557	0.0115	0.07662	0	476	0	0	0	0	1	-360	360;
	2211	2937	0.002661	0.01286	0	0	0	0	0	0	1	-360	360;
	1737	4693	0.010793	0.046942	0	0	0	0	0	0	1	-360	360;
	979	9210	0.023826	0.100992	0	0	0	0	0	0	1	-360	360;
	9210	7608	0.014603	0.07986	0	0	0	0	0	0	1	-360	360;
	9210	5081	0.026132	0.099818	0	0	0	0	0	0	1	-360	360;
	8020	3883	0.002289	0.012554	0	0	0	0	0	0	1	-360	360;
	8020	6258	0.00743	0.030719	0	0	0	0	0	0	1	-360	360;
	7554	3415	0.014653	0.029174	0	0	0	0	0	0	1	-360	360;
	5962	7554	0.016388	0.04286	0	0	0	0	0	0	1	-360	360;
	4536	8983	0.056248	0.122041	0	0	0	0	0	0	1	-360	360;
	343	8983	0.024901	0.049099	0	0	0	0	0	0	1	-360	360;
	4643	7181	0.026554	0.061851	0	0	0	0	0	0	1	-360	360;
	3421	6739	0.073678	0.305793	0	0	0	0	0	0	1	-360	360;
	3094	3421	0.029083	0.060017	0	0	0	0	0	0	1	-360	360;
	3421	66	0.021893	0.134521	0	0	0	0	0	0	1	-360	360;
	1324	5552	0.018983	0.03914	0	0	0	0	0	0	1	-360	360;
	5552	5676	0.024529	0.041612	0	0	0	0	0	0	1	-360	360;
	7077	5443	0.039421	0.085686	0	0	0	0	0	0	1	-360	360;
	7340	5443	0.039421	0.085769	0	0	0	0	0	0	1	-360	360;
	7242	7077	0.00876	0.016901	0	0	0	0	0	0	1	-360	360;
	4028	7242	0.006967	0.018248	0	0	0	0	0	0	1	-360	360;
	8372	2763	0.025099	0.107901	0	0	0	0	0	0	1	-360	360;
	3683	8372	0.02605	0.103438	0	0	0	0	0	0	1	-360	360;
	5198	2263	0.012719	0.033719	0	0	0	0	0	0	1	-360	360;
	2263	7793	0.022273	0.043661	0	0	0	0	0	0	1	-360	360;
	8046	5219	0.00124	0.006653	0	0	0	0	0	0	1	-360	360;
	2167	5219	0.001587	0.007636	0	0	0	0	0	0	1	-360	360;
	6133	3781	0.006207	0.032331	0	0	0	0	0	0	1	-360	360;
	6133	3781	0.006083	0.032942	0	0	0	0	0	0	1	-360	360;
	3781	3746	0.001314	0.005446	0	0	0	0	0	0	1	-360	360;
	4275	435	0.003851	0.01662	0	0	0	0	0	0	1	-360	360;
	3006	8029	0.005124	0.031934	0	0	0	0	0	0	1	-360	360;
	8978	8029	0.005008	0.029562	0	0	0	0	0	0	1	-360	360;
	589	8174	0.048331	0.094372	0	0	0	0	0	0	1	-360	360;
	589	2602	0.015678	0.073826	0	0	0	0	0	0	1	-360	360;
	2679	868	0.005579	0.019174	0	0	0	0	0	0	1	-360	360;
	343	2679	0.029397	0.077132	0	0	0	0	0	0	1	-360	360;
	1682	868	0.002901	0.006802	0	0	0	0	0	0	1	-360	360;
	1682	93	0.004769	0.020826	0	0	0	0	0	0	1	-360	360;
	1682	8709	0.031388	0.06305	0	0	0	0	0	0	1	-360	360;
	1682	3890	0.032587	0.068752	0	0	0	0	0	0	1	-360	360;
	1682	1353	0.042802	0.083983	0	0	0	0	0	0	1	-360	360;
	3844	3547	0.01214	0.069124	0	0	0	0	0	0	1	-360	360;
	2892	1873	0.008388	0.017033	0	0	0	0	0	0	1	-360	360;
	8925	8746	0.008521	0.046653	0	0	0	0	0	0	1	-360	360;
	8925	8505	0.003331	0.013818	0	0	0	0	0	0	1	-360	360;
	2677	2419	0.016729	0.090401	0	495	0	0	0	0	1	-360	360;
	2677	6046	0.01526	0.081479	0	495	0	0	0	0	1	-360	360;
	2677	8551	0.00905	0.068219	0	457	0	0	0	0	1	-360	360;
	2677	1922	0.01205	0.08575	0	476	0	0	0	0	1	-360	360;
	2677	6971	0.007841	0.04113	0	495	0	0	0	0	1	-360	360;
	84	239	0.002413	0.010686	0	0	0	0	0	0	1	-360	360;
	884	84	0.005041	0.022769	0	0	0	0	0	0	1	-360	360;
	2386	36	0.00126	0.007531	0	419	0	0	0	0	1	-360	360;
	7737	2068	0.00176	0.007603	0	0	0	0	0	0	1	-360	360;
	7737	2068	0.00157	0.00862	0	0	0	0	0	0	1	-360	360;
	8852	2561	0.02743	0.059157	0	0	0	0	0	0	1	-360	360;
	1677	1447	0.028992	0.060893	0	0	0	0	0	0	1	-360	360;
	3531	2415	0.012942	0.051851	0	0	0	0	0	0	1	-360	360;
	1084	3531	0.071157	0.104149	0	0	0	0	0	0	1	-360	360;
	1171	6695	0.002017	0.010686	0	0	0	0	0	0	1	-360	360;
	3141	6695	0.00195	0.009364	0	0	0	0	0	0	1	-360	360;
	3141	6695	0.001992	0.009174	0	0	0	0	0	0	1	-360	360;
	1324	574	0.027017	0.074521	0	0	0	0	0	0	1	-360	360;
	574	7077	0.025107	0.072339	0	0	0	0	0	0	1	-360	360;
	4341	3730	0.002942	0.006248	0	0	0	0	0	0	1	-360	360;
	3730	3828	0.006983	0.028678	0	0	0	0	0	0	1	-360	360;
	1873	5563	0.049636	0.077364	0	0	0	0	0	0	1	-360	360;
	8797	5563	0.071752	0.121198	0	0	0	0	0	0	1	-360	360;
	390	5515	0.00219	0.009174	0	0	0	0	0	0	1	-360	360;
	6076	390	0.018719	0.067983	0	0	0	0	0	0	1	-360	360;
	8978	5159	0.006339	0.025397	0	0	0	0	0	0	1	-360	360;
	7246	23	0.011876	0.064967	0	0	0	0	0	0	1	-360	360;
	6739	4372	0.004603	0.024603	0	0	0	0	0	0	1	-360	360;
	1782	6739	0.037934	0.102868	0	0	0	0	0	0	1	-360	360;
	2998	7296	0.020818	0.067711	0	0	0	0	0	0	1	-360	360;
	2776	1171	0.020223	0.048471	0	0	0	0	0	0	1	-360	360;
	1171	3746	0.00705	0.034314	0	0	0	0	0	0	1	-360	360;
	1171	7250	0.015388	0.074008	0	0	0	0	0	0	1	-360	360;
	1171	7250	0.015066	0.075463	0	0	0	0	0	0	1	-360	360;
	1324	34	0.060331	0.128752	0	0	0	0	0	0	1	-360	360;
	34	7022	0.037752	0.085843	0	0	0	0	0	0	1	-360	360;
	34	636	0.060471	0.109992	0	0	0	0	0	0	1	-360	360;
	5676	34	0.021959	0.050058	0	0	0	0	0	0	1	-360	360;
	7077	2130	0.005347	0.026198	0	0	0	0	0	0	1	-360	360;
	4028	2130	0.005058	0.015769	0	0	0	0	0	0	1	-360	360;
	979	1107	0.010256	0.053223	0	0	0	0	0	0	1	-360	360;
	3576	4372	0.01476	0.082157	0	0	0	0	0	0	1	-360	360;
	4643	1247	0.047843	0.112033	0	0	0	0	0	0	1	-360	360;
	7882	4161	0.013909	0.036793	0	0	0	0	0	0	1	-360	360;
	4161	3094	0.030289	0.064818	0	0	0	0	0	0	1	-360	360;
	4161	1098	0.026132	0.123132	0	0	0	0	0	0	1	-360	360;
	4161	1098	0.026628	0.120562	0	0	0	0	0	0	1	-360	360;
	8269	4161	0.016727	0.063959	0	0	0	0	0	0	1	-360	360;
	6651	5275	0.002413	0.010248	0	0	0	0	0	0	1	-360	360;
	6651	4184	0.009	0.040843	0	0	0	0	0	0	1	-360	360;
	435	8446	0.008421	0.018488	0	0	0	0	0	0	1	-360	360;
	8514	8219	0.014917	0.034322	0	0	0	0	0	0	1	-360	360;
	8514	2790	0.00743	0.035942	0	0	0	0	0	0	1	-360	360;
	5275	8514	0.003851	0.016	0	0	0	0	0	0	1	-360	360;
	1590	8514	0.00162	0.008008	0	0	0	0	0	0	1	-360	360;
	3477	7882	0.014372	0.045793	0	0	0	0	0	0	1	-360	360;
	5086	3477	0.038678	0.089967	0	0	0	0	0	0	1	-360	360;
	7796	4655	0.009504	0.021727	0	0	0	0	0	0	1	-360	360;
	4655	4877	0.018364	0.096521	0	0	0	0	0	0	1	-360	360;
	5027	6453	0.00195	0.010669	0	0	0	0	0	0	1	-360	360;
	6453	6196	0.021182	0.046826	0	0	0	0	0	0	1	-360	360;
	7336	7608	0.031752	0.131826	0	0	0	0	0	0	1	-360	360;
	5515	4518	0.002917	0.01238	0	0	0	0	0	0	1	-360	360;
	23	1779	0.016157	0.035826	0	0	0	0	0	0	1	-360	360;
	23	3144	0.003727	0.020529	0	0	0	0	0	0	1	-360	360;
	23	6002	0.019992	0.079231	0	0	0	0	0	0	1	-360	360;
	3151	1779	0.023653	0.094587	0	0	0	0	0	0	1	-360	360;
	3144	1517	0.009322	0.049058	0	0	0	0	0	0	1	-360	360;
	5962	1517	0.026727	0.146752	0	0	0	0	0	0	1	-360	360;
	5160	557	0.000531	0.002919	0	0	0	0	0	0	1	-360	360;
	572	6196	0.021992	0.059397	0	0	0	0	0	0	1	-360	360;
	572	5086	0.044413	0.10176	0	0	0	0	0	0	1	-360	360;
	4044	572	0.043223	0.088562	0	0	0	0	0	0	1	-360	360;
	3572	3281	0.006992	0.014066	0	0	0	0	0	0	1	-360	360;
	1698	5368	0.007256	0.028802	0	0	0	0	0	0	1	-360	360;
	1698	5368	0.006471	0.032636	0	0	0	0	0	0	1	-360	360;
	1698	8910	0.043198	0.156397	0	0	0	0	0	0	1	-360	360;
	7823	5024	0.016248	0.04014	0	0	0	0	0	0	1	-360	360;
	1438	1143	0.00206	0.01393	0	457	0	0	0	0	1	-360	360;
	1438	1143	0.002021	0.01419	0	495	0	0	0	0	1	-360	360;
	1438	36	0.00324	0.02076	0	438	0	0	0	0	1	-360	360;
	1438	36	0.00318	0.021159	0	457	0	0	0	0	1	-360	360;
	1324	7077	0.031603	0.070331	0	0	0	0	0	0	1	-360	360;
	2167	1324	0.003132	0.014554	0	0	0	0	0	0	1	-360	360;
	1509	1324	0.006669	0.032091	0	0	0	0	0	0	1	-360	360;
	4684	8746	0.008934	0.046298	0	0	0	0	0	0	1	-360	360;
	915	5319	0.002942	0.012008	0	0	0	0	0	0	1	-360	360;
	5637	4065	0.010942	0.062207	0	0	0	0	0	0	1	-360	360;
	7203	5917	0.002529	0.011777	0	0	0	0	0	0	1	-360	360;
	7203	507	0.015107	0.042769	0	0	0	0	0	0	1	-360	360;
	8646	9026	0.00705	0.022099	0	0	0	0	0	0	1	-360	360;
	6065	3842	0.038231	0.105893	0	0	0	0	0	0	1	-360	360;
	3923	6065	0.005901	0.021289	0	0	0	0	0	0	1	-360	360;
	116	5198	0.013636	0.022686	0	0	0	0	0	0	1	-360	360;
	7022	636	0.023826	0.059364	0	0	0	0	0	0	1	-360	360;
	7077	3524	0.016091	0.045967	0	0	0	0	0	0	1	-360	360;
	3524	4304	0.017033	0.053008	0	0	0	0	0	0	1	-360	360;
	7384	3842	0.017744	0.09686	0	0	0	0	0	0	1	-360	360;
	5917	5082	0.003636	0.018818	0	0	0	0	0	0	1	-360	360;
	5917	5082	0.003562	0.019174	0	0	0	0	0	0	1	-360	360;
	6557	5607	0.008256	0.021909	0	0	0	0	0	0	1	-360	360;
	5607	7796	0.026917	0.061636	0	0	0	0	0	0	1	-360	360;
	5730	4438	0.022702	0.057149	0	0	0	0	0	0	1	-360	360;
	4438	8272	0.023983	0.070124	0	0	0	0	0	0	1	-360	360;
	3614	2626	0.035231	0.078314	0	0	0	0	0	0	1	-360	360;
	2626	8797	0.02462	0.053603	0	0	0	0	0	0	1	-360	360;
	3883	4944	0.008091	0.03214	0	0	0	0	0	0	1	-360	360;
	419	3381	0.016793	0.081388	0	0	0	0	0	0	1	-360	360;
	3381	1107	0.008636	0.047388	0	0	0	0	0	0	1	-360	360;
	5176	7077	0.009364	0.018207	0	0	0	0	0	0	1	-360	360;
	7077	8173	0.023331	0.063727	0	0	0	0	0	0	1	-360	360;
	7077	498	0.014983	0.033579	0	0	0	0	0	0	1	-360	360;
	3141	6133	0.004636	0.019678	0	0	0	0	0	0	1	-360	360;
	3141	6133	0.004727	0.019273	0	0	0	0	0	0	1	-360	360;
	3006	157	0.018446	0.089686	0	0	0	0	0	0	1	-360	360;
	8978	3006	0.011529	0.056587	0	0	0	0	0	0	1	-360	360;
	8091	6258	0.00862	0.021587	0	0	0	0	0	0	1	-360	360;
	8091	3171	0.005702	0.013017	0	0	0	0	0	0	1	-360	360;
	7130	645	0.048256	0.078099	0	0	0	0	0	0	1	-360	360;
	645	4297	0.065529	0.159802	0	0	0	0	0	0	1	-360	360;
	6033	645	0.02681	0.110702	0	0	0	0	0	0	1	-360	360;
	66	1782	0.04538	0.142983	0	0	0	0	0	0	1	-360	360;
	1084	3309	0.024992	0.097777	0	0	0	0	0	0	1	-360	360;
	8782	3309	0.020256	0.109372	0	0	0	0	0	0	1	-360	360;
	3309	521	0.024636	0.095653	0	0	0	0	0	0	1	-360	360;
	2211	5081	0.003364	0.017669	0	0	0	0	0	0	1	-360	360;
	3669	8821	0.018463	0.039124	0	0	0	0	0	0	1	-360	360;
	8821	1229	0.021636	0.060793	0	0	0	0	0	0	1	-360	360;
	7280	116	0.01786	0.039711	0	0	0	0	0	0	1	-360	360;
	2419	3505	0.00588	0.03186	0	0	0	0	0	0	1	-360	360;
	5804	7336	0.045058	0.177512	0	0	0	0	0	0	1	-360	360;
	7336	2733	0.003686	0.017711	0	0	0	0	0	0	1	-360	360;
	7336	2733	0.003612	0.018058	0	0	0	0	0	0	1	-360	360;
	7336	2733	0.007876	0.021066	0	0	0	0	0	0	1	-360	360;
	2256	7336	0.03186	0.159347	0	0	0	0	0	0	1	-360	360;
	8797	3576	0.034901	0.071347	0	0	0	0	0	0	1	-360	360;
	1856	5368	0.006339	0.033281	0	0	0	0	0	0	1	-360	360;
	1856	5368	0.006207	0.033917	0	0	0	0	0	0	1	-360	360;
	3547	5368	0.006207	0.034983	0	0	0	0	0	0	1	-360	360;
	6523	8978	0.006289	0.024661	0	0	0	0	0	0	1	-360	360;
	6523	7594	0.002554	0.01376	0	0	0	0	0	0	1	-360	360;
	944	1447	0.003579	0.01486	0	0	0	0	0	0	1	-360	360;
	5583	1843	0.01495	0.02843	0	0	0	0	0	0	1	-360	360;
	4877	7375	0.011727	0.028934	0	0	0	0	0	0	1	-360	360;
	7375	8646	0.013843	0.048909	0	0	0	0	0	0	1	-360	360;
	1873	1843	0.032388	0.080711	0	0	0	0	0	0	1	-360	360;
	1873	1843	0.011719	0.074793	0	0	0	0	0	0	1	-360	360;
	1843	6076	0.00057	0.002612	0	0	0	0	0	0	1	-360	360;
	5119	1270	0.012554	0.026314	0	0	0	0	0	0	1	-360	360;
	5962	5119	0.022554	0.049397	0	0	0	0	0	0	1	-360	360;
	944	2813	0.005438	0.021669	0	0	0	0	0	0	1	-360	360;
	1668	1506	0.00079	0.00721	0	1119	0	0	0	0	1	-360	360;
	1668	3981	0.00023	0.00228	0	1119	0	0	0	0	1	-360	360;
	7737	9145	0.002248	0.012083	0	0	0	0	0	0	1	-360	360;
	7737	9145	0.002198	0.012314	0	0	0	0	0	0	1	-360	360;
	7340	7782	0.014727	0.064099	0	0	0	0	0	0	1	-360	360;
	239	7340	0.016364	0.040901	0	0	0	0	0	0	1	-360	360;
	4668	7340	0.024471	0.063686	0	0	0	0	0	0	1	-360	360;
	8322	3885	0.00037	0.00429	0	1185	0	0	0	0	1	-360	360;
	9194	3885	0.00201	0.02375	0	1283	0	0	0	0	1	-360	360;
	3885	2942	9e-05	0.00093	0	1415	0	0	0	0	1	-360	360;
	3885	1905	9e-05	0.00095	0	1415	0	0	0	0	1	-360	360;
	6046	9144	0.004969	0.034979	0	438	0	0	0	0	1	-360	360;
	6046	9144	0.00468	0.033031	0	419	0	0	0	0	1	-360	360;
	9144	6434	0.00076	0.007909	0	629	0	0	0	0	1	-360	360;
	9144	5497	0.020539	0.080461	0	476	0	0	0	0	1	-360	360;
	1873	2790	0.024562	0.099231	0	0	0	0	0	0	1	-360	360;
	1873	8797	0.103397	0.231174	0	0	0	0	0	0	1	-360	360;
	1873	66	0.063843	0.141661	0	0	0	0	0	0	1	-360	360;
	8322	727	0.00059	0.00906	0	1810	0	0	0	0	1	-360	360;
	6046	6434	0.001979	0.02175	0	0	0	0	0	0	1	-360	360;
	6046	2110	0.005281	0.040039	0	495	0	0	0	0	1	-360	360;
	6046	557	0.012281	0.069219	0	495	0	0	0	0	1	-360	360;
	6046	6971	0.00668	0.04144	0	533	0	0	0	0	1	-360	360;
	6434	6158	0.00725	0.03801	0	553	0	0	0	0	1	-360	360;
	3981	2041	0.00225	0.02188	0	1415	0	0	0	0	1	-360	360;
	6006	1922	0.00525	0.03749	0	514	0	0	0	0	1	-360	360;
	6006	1922	0.00605	0.03069	0	514	0	0	0	0	1	-360	360;
	3669	3263	0.054793	0.13405	0	0	0	0	0	0	1	-360	360;
	7782	3669	0.019099	0.076488	0	0	0	0	0	0	1	-360	360;
	3669	6867	0.035686	0.057678	0	0	0	0	0	0	1	-360	360;
	6758	7054	0.00481	0.020579	0	0	0	0	0	0	1	-360	360;
	6758	7054	0.004909	0.020149	0	0	0	0	0	0	1	-360	360;
	6758	4304	0.010843	0.055512	0	0	0	0	0	0	1	-360	360;
	6758	4304	0.012289	0.054397	0	0	0	0	0	0	1	-360	360;
	498	6758	0.003562	0.018124	0	0	0	0	0	0	1	-360	360;
	8717	8219	0.038017	0.081934	0	0	0	0	0	0	1	-360	360;
	8717	1098	0.020149	0.083124	0	0	0	0	0	0	1	-360	360;
	8717	6732	0.009554	0.040959	0	0	0	0	0	0	1	-360	360;
	2813	8232	0.008331	0.043322	0	0	0	0	0	0	1	-360	360;
	8232	1981	0.02114	0.114479	0	0	0	0	0	0	1	-360	360;
	1084	3263	0.017488	0.03838	0	0	0	0	0	0	1	-360	360;
	7054	4232	0.00381	0.019256	0	0	0	0	0	0	1	-360	360;
	4028	4232	0.001397	0.005802	0	0	0	0	0	0	1	-360	360;
	2279	6258	0.004719	0.024603	0	0	0	0	0	0	1	-360	360;
	2279	6258	0.00462	0.025058	0	0	0	0	0	0	1	-360	360;
	3171	6258	0.005413	0.027876	0	0	0	0	0	0	1	-360	360;
	6258	5804	0.032041	0.196413	0	0	0	0	0	0	1	-360	360;
	1970	6258	0.002198	0.022248	0	0	0	0	0	0	1	-360	360;
	6258	190	0.00262	0.018463	0	0	0	0	0	0	1	-360	360;
	6258	419	0.030347	0.112149	0	0	0	0	0	0	1	-360	360;
	3828	7384	0.014512	0.05981	0	0	0	0	0	0	1	-360	360;
	1270	8274	0.005438	0.009124	0	0	0	0	0	0	1	-360	360;
	9125	1314	0.005169	0.022219	0	0	0	0	0	0	1	-360	360;
	716	6537	0.007793	0.029769	0	0	0	0	0	0	1	-360	360;
	716	7991	0.006091	0.028653	0	0	0	0	0	0	1	-360	360;
	4184	5024	0.02876	0.060388	0	0	0	0	0	0	1	-360	360;
	1641	1911	0.02443	0.09576	0	0	0	0	0	0	1	-360	360;
	5962	1911	0.009083	0.03857	0	0	0	0	0	0	1	-360	360;
	8797	761	0.048157	0.08462	0	0	0	0	0	0	1	-360	360;
	4944	2279	0.002512	0.013826	0	0	0	0	0	0	1	-360	360;
	8274	6304	0.012298	0.026826	0	0	0	0	0	0	1	-360	360;
	8408	8599	0.022744	0.133521	0	0	0	0	0	0	1	-360	360;
	5319	4877	0.022554	0.088876	0	0	0	0	0	0	1	-360	360;
	7130	8709	0.019231	0.033802	0	0	0	0	0	0	1	-360	360;
	8897	6557	0.010017	0.022711	0	0	0	0	0	0	1	-360	360;
	8897	507	0.00219	0.011843	0	0	0	0	0	0	1	-360	360;
	5730	5706	0.019281	0.064488	0	0	0	0	0	0	1	-360	360;
	2602	5706	0.02119	0.091727	0	0	0	0	0	0	1	-360	360;
	4297	8803	0.003099	0.013893	0	0	0	0	0	0	1	-360	360;
	1084	8803	0.004132	0.021959	0	0	0	0	0	0	1	-360	360;
	1084	8803	0.004041	0.02238	0	0	0	0	0	0	1	-360	360;
	1084	8782	0.002471	0.010901	0	0	0	0	0	0	1	-360	360;
	7777	9194	0.002	0.017	0	1283	0	0	0	0	1	-360	360;
	1199	9194	0.0016	0.01758	0	1119	0	0	0	0	1	-360	360;
	8408	332	0.027397	0.111917	0	0	0	0	0	0	1	-360	360;
	8408	7087	0.008752	0.044802	0	0	0	0	0	0	1	-360	360;
	716	6304	0.045727	0.07386	0	0	0	0	0	0	1	-360	360;
	5804	8527	0.044281	0.119579	0	0	0	0	0	0	1	-360	360;
	1353	8527	0.057612	0.11224	0	0	0	0	0	0	1	-360	360;
	4341	8527	0.063314	0.107273	0	0	0	0	0	0	1	-360	360;
	4028	5176	0.006769	0.017909	0	0	0	0	0	0	1	-360	360;
	7000	4184	0.003273	0.01157	0	0	0	0	0	0	1	-360	360;
	7000	542	0.000248	0.001331	0	0	0	0	0	0	1	-360	360;
	2733	2713	0.010603	0.045727	0	0	0	0	0	0	1	-360	360;
	2713	4341	0.016273	0.089678	0	0	0	0	0	0	1	-360	360;
	3890	7496	0.008397	0.022777	0	0	0	0	0	0	1	-360	360;
	7496	1416	0.013421	0.036314	0	0	0	0	0	0	1	-360	360;
	8978	4009	0.009628	0.050752	0	0	0	0	0	0	1	-360	360;
	6002	332	0.011777	0.046843	0	0	0	0	0	0	1	-360	360;
	3746	332	0.014446	0.059967	0	0	0	0	0	0	1	-360	360;
	5804	2256	0.005174	0.029562	0	0	0	0	0	0	1	-360	360;
	8272	2678	0.069256	0.110017	0	0	0	0	0	0	1	-360	360;
	1500	1963	0.00524	0.022124	0	0	0	0	0	0	1	-360	360;
	5082	8954	0.005083	0.019421	0	0	0	0	0	0	1	-360	360;
	8954	1325	0.004099	0.013868	0	0	0	0	0	0	1	-360	360;
	8890	7280	0.04181	0.081397	0	0	0	0	0	0	1	-360	360;
	8890	6942	0.041446	0.067174	0	0	0	0	0	0	1	-360	360;
	36	2369	0.008841	0.0587	0	476	0	0	0	0	1	-360	360;
	1856	3547	0.011876	0.064	0	0	0	0	0	0	1	-360	360;
	5539	2211	0.037579	0.182612	0	0	0	0	0	0	1	-360	360;
	157	7296	0.025752	0.129421	0	0	0	0	0	0	1	-360	360;
	1963	8269	0.00543	0.011421	0	0	0	0	0	0	1	-360	360;
	8269	3923	0.003727	0.010066	0	0	0	0	0	0	1	-360	360;
	4304	7112	0.015868	0.053702	0	0	0	0	0	0	1	-360	360;
	2163	8890	0.026388	0.120603	0	0	0	0	0	0	1	-360	360;
	8046	2167	0.002372	0.010669	0	0	0	0	0	0	1	-360	360;
	5082	1325	0.001289	0.005041	0	0	0	0	0	0	1	-360	360;
	5082	3646	0.001438	0.006421	0	0	0	0	0	0	1	-360	360;
	5082	6933	0.001471	0.006289	0	0	0	0	0	0	1	-360	360;
	3981	1199	0.00105	0.01056	0	1086	0	0	0	0	1	-360	360;
	2865	36	0.01137	0.056789	0	381	0	0	0	0	1	-360	360;
	665	36	0.00519	0.028151	0	476	0	0	0	0	1	-360	360;
	1641	5962	0.030091	0.150033	0	0	0	0	0	0	1	-360	360;
	6033	94	0.011207	0.023851	0	0	0	0	0	0	1	-360	360;
	94	8549	0.073347	0.103488	0	0	0	0	0	0	1	-360	360;
	5457	6367	0.064149	0.131041	0	0	0	0	0	0	1	-360	360;
	1560	1314	0.002341	0.009719	0	0	0	0	0	0	1	-360	360;
	5570	279	0.051231	0.126554	0	0	0	0	0	0	1	-360	360;
	1229	279	0.031967	0.079413	0	0	0	0	0	0	1	-360	360;
	1590	542	0.00024	0.001	0	0	0	0	0	0	1	-360	360;
	5565	3062	0.034008	0.069388	0	0	0	0	0	0	1	-360	360;
	521	1028	0.003	0.01243	0	0	0	0	0	0	1	-360	360;
	1028	7484	0.036281	0.073678	0	0	0	0	0	0	1	-360	360;
	2167	3237	0.003587	0.017231	0	0	0	0	0	0	1	-360	360;
	1820	6971	0.00012	0.00062	0	0	0	0	0	0	1	-360	360;
	7456	6732	0.006909	0.039231	0	0	0	0	0	0	1	-360	360;
	6501	6502	0.003314	0.007017	0	0	0	0	0	0	1	-360	360;
	8910	8978	0.011289	0.063661	0	0	0	0	0	0	1	-360	360;
	8910	8978	0.01105	0.058942	0	0	0	0	0	0	1	-360	360;
	7594	8978	0.007711	0.039413	0	0	0	0	0	0	1	-360	360;
	568	239	0.004017	0.013669	0	0	0	0	0	0	1	-360	360;
	3861	1247	0.040521	0.08386	0	0	0	0	0	0	1	-360	360;
	2559	2672	0.024579	0.061231	0	0	0	0	0	0	1	-360	360;
	3062	4044	0.067992	0.151463	0	0	0	0	0	0	1	-360	360;
	5015	8505	0.003579	0.011942	0	0	0	0	0	0	1	-360	360;
	6501	422	0.00776	0.021587	0	0	0	0	0	0	1	-360	360;
	8910	8093	0.003298	0.009463	0	0	0	0	0	0	1	-360	360;
	8144	8093	0.005157	0.01376	0	0	0	0	0	0	1	-360	360;
	6055	8144	0.009934	0.034479	0	0	0	0	0	0	1	-360	360;
	3646	1509	0.002008	0.010802	0	0	0	0	0	0	1	-360	360;
	6933	3237	0.001967	0.011008	0	0	0	0	0	0	1	-360	360;
	654	5807	0.003219	0.01439	0	0	0	0	0	0	1	-360	360;
	654	1924	0.00239	0.01426	0	0	0	0	0	0	1	-360	360;
	8675	6520	0.00143	0.013591	0	591	0	0	0	0	1	-360	360;
	8675	7219	0.004669	0.05795	0	0	0	0	0	0	1	-360	360;
	8675	1553	0.00412	0.043151	0	610	0	0	0	0	1	-360	360;
	8675	3124	0.000521	0.00612	0	0	0	0	0	0	1	-360	360;
	5448	4808	0.005849	0.02768	0	273	0	0	0	0	1	-360	360;
	4762	5448	0.012742	0.049311	0	273	0	0	0	0	1	-360	360;
	6411	7219	0.00206	0.01061	0	0	0	0	0	0	1	-360	360;
	6411	2005	0.001591	0.01	0	0	0	0	0	0	1	-360	360;
	1278	1557	0.00098	0.00874	0	1547	0	0	0	0	1	-360	360;
	1278	5527	0.00045	0.00548	0	856	0	0	0	0	1	-360	360;
	4761	5992	0.000719	0.003349	0	0	0	0	0	0	1	-360	360;
	4761	2257	0.000919	0.00687	0	0	0	0	0	0	1	-360	360;
	4761	6335	0.000219	0.0015	0	0	0	0	0	0	1	-360	360;
	3278	2624	0.00355	0.019669	0	591	0	0	0	0	1	-360	360;
	3278	752	0.00389	0.02201	0	495	0	0	0	0	1	-360	360;
	3278	7239	0.00251	0.01455	0	629	0	0	0	0	1	-360	360;
	4416	4553	0.00825	0.036781	0	0	0	0	0	0	1	-360	360;
	4416	4176	0.000581	0.00412	0	591	0	0	0	0	1	-360	360;
	6775	8905	0.00103	0.01236	0	1481	0	0	0	0	1	-360	360;
	6775	5527	0.0014	0.01768	0	1349	0	0	0	0	1	-360	360;
	5438	356	0.00088	0.01169	0	0	0	0	0	0	1	-360	360;
	3949	147	0.001711	0.01164	0	629	0	0	0	0	1	-360	360;
	3949	4453	0.00268	0.015031	0	553	0	0	0	0	1	-360	360;
	3949	9048	0.0018	0.01132	0	572	0	0	0	0	1	-360	360;
	1122	4211	0.00051	0.00268	0	476	0	0	0	0	1	-360	360;
	1122	2501	0.000651	0.003289	0	419	0	0	0	0	1	-360	360;
	3433	4201	0.0012	0.01137	0	1744	0	0	0	0	1	-360	360;
	3433	6666	0.00168	0.02275	0	1744	0	0	0	0	1	-360	360;
	4540	5992	0.002331	0.0225	0	0	0	0	0	0	1	-360	360;
	4540	5992	0.002081	0.0255	0	0	0	0	0	0	1	-360	360;
	4540	688	0.000781	0.004979	0	0	0	0	0	0	1	-360	360;
	4540	2191	0.00514	0.06699	0	0	0	0	0	0	1	-360	360;
	7472	3494	0.00088	0.00804	0	1942	0	0	0	0	1	-360	360;
	7472	3277	0.00032	0.00311	0	1678	0	0	0	0	1	-360	360;
	7472	219	0.00015	0.00197	0	1020	0	0	0	0	1	-360	360;
	7681	5807	0.004719	0.023831	0	438	0	0	0	0	1	-360	360;
	7681	2257	0.00669	0.029979	0	457	0	0	0	0	1	-360	360;
	9161	8395	0.001031	0.01225	0	0	0	0	0	0	1	-360	360;
	9161	8598	0.00105	0.00937	0	553	0	0	0	0	1	-360	360;
	9161	943	0.00201	0.01293	0	400	0	0	0	0	1	-360	360;
	2017	752	0.002	0.012229	0	553	0	0	0	0	1	-360	360;
	2017	1915	8.1e-05	0.000521	0	400	0	0	0	0	1	-360	360;
	2017	6520	0.00113	0.010969	0	0	0	0	0	0	1	-360	360;
	8905	8742	0.00195	0.02538	0	1843	0	0	0	0	1	-360	360;
	8905	3277	0.00164	0.01684	0	1514	0	0	0	0	1	-360	360;
	4631	5898	0.006289	0.06895	0	0	0	0	0	0	1	-360	360;
	4631	8986	0.001169	0.010229	0	0	0	0	0	0	1	-360	360;
	4631	5844	0.00805	0.060409	0	0	0	0	0	0	1	-360	360;
	4631	5844	0.00789	0.061581	0	0	0	0	0	0	1	-360	360;
	4631	7219	0.007159	0.04164	0	533	0	0	0	0	1	-360	360;
	4631	7219	0.007289	0.040729	0	533	0	0	0	0	1	-360	360;
	4631	7027	0.004531	0.02731	0	0	0	0	0	0	1	-360	360;
	4631	1062	0.004521	0.033581	0	533	0	0	0	0	1	-360	360;
	8395	3407	0.0005	0.005081	0	0	0	0	0	0	1	-360	360;
	8395	811	0.00132	0.006521	0	0	0	0	0	0	1	-360	360;
	5309	5898	0.004401	0.030531	0	0	0	0	0	0	1	-360	360;
	5309	1553	0.000469	0.003409	0	629	0	0	0	0	1	-360	360;
	3042	3494	0.00239	0.02364	0	1777	0	0	0	0	1	-360	360;
	3042	4323	0.00034	0.0041	0	1777	0	0	0	0	1	-360	360;
	3042	6666	0.00027	0.00327	0	1645	0	0	0	0	1	-360	360;
	3042	6666	0.00026	0.00333	0	1777	0	0	0	0	1	-360	360;
	3042	2716	0.00059	0.00574	0	1547	0	0	0	0	1	-360	360;
	1017	7785	0.000789	0.003711	0	0	0	0	0	0	1	-360	360;
	1017	7529	0.00419	0.04988	0	0	0	0	0	0	1	-360	360;
	1017	5242	0.001651	0.010919	0	0	0	0	0	0	1	-360	360;
	2498	7027	0.007281	0.03162	0	0	0	0	0	0	1	-360	360;
	2498	2290	0.006831	0.03505	0	572	0	0	0	0	1	-360	360;
	3494	5527	0.00108	0.01258	0	1645	0	0	0	0	1	-360	360;
	3494	4323	0.00174	0.0229	0	1547	0	0	0	0	1	-360	360;
	3494	2716	0.00223	0.02125	0	1678	0	0	0	0	1	-360	360;
	5807	5992	0.0023	0.014081	0	0	0	0	0	0	1	-360	360;
	5807	1924	0.00213	0.026831	0	0	0	0	0	0	1	-360	360;
	5807	5795	0.001039	0.00549	0	0	0	0	0	0	1	-360	360;
	5807	8276	0.00301	0.016711	0	0	0	0	0	0	1	-360	360;
	5807	1406	0.000849	0.005229	0	0	0	0	0	0	1	-360	360;
	5807	1603	0.002781	0.020281	0	591	0	0	0	0	1	-360	360;
	5807	943	0.002159	0.01389	0	0	0	0	0	0	1	-360	360;
	153	8142	0.003502	0.031631	0	325	0	0	0	0	1	-360	360;
	4808	7005	0.003929	0.024729	0	260	0	0	0	0	1	-360	360;
	7005	153	0.00076	0.005609	0	260	0	0	0	0	1	-360	360;
	2624	4211	0.002039	0.01249	0	533	0	0	0	0	1	-360	360;
	4938	8742	0.00036	0.00572	0	1942	0	0	0	0	1	-360	360;
	4938	1141	0.00028	0.00448	0	1876	0	0	0	0	1	-360	360;
	2987	5898	0.001091	0.01663	0	457	0	0	0	0	1	-360	360;
	2987	7027	0.011909	0.080909	0	476	0	0	0	0	1	-360	360;
	2987	147	0.000401	0.00232	0	381	0	0	0	0	1	-360	360;
	2987	147	0.000349	0.00263	0	0	0	0	0	0	1	-360	360;
	1487	5973	0.000531	0.0063	0	0	0	0	0	0	1	-360	360;
	1487	464	0.005531	0.02938	0	533	0	0	0	0	1	-360	360;
	1487	7896	0.00019	0.002521	0	0	0	0	0	0	1	-360	360;
	5973	5898	0.00075	0.010909	0	495	0	0	0	0	1	-360	360;
	8362	4176	0.007	0.03088	0	629	0	0	0	0	1	-360	360;
	8362	4453	0.002349	0.009281	0	553	0	0	0	0	1	-360	360;
	5898	5225	0.003219	0.02125	0	0	0	0	0	0	1	-360	360;
	5898	5225	0.004461	0.027961	0	514	0	0	0	0	1	-360	360;
	5236	4284	0.00105	0.007901	0	629	0	0	0	0	1	-360	360;
	6719	4553	0.002461	0.01019	0	0	0	0	0	0	1	-360	360;
	6719	4176	0.00605	0.028409	0	533	0	0	0	0	1	-360	360;
	2097	943	7e-05	0.00039	0	0	0	0	0	0	1	-360	360;
	5992	1924	0.0047	0.02175	0	495	0	0	0	0	1	-360	360;
	7459	4342	0.00118	0.00719	0	591	0	0	0	0	1	-360	360;
	7459	227	0.002349	0.01887	0	0	0	0	0	0	1	-360	360;
	2099	1586	0.00104	0.01314	0	1876	0	0	0	0	1	-360	360;
	2099	6666	0.0004	0.004	0	0	0	0	0	0	1	-360	360;
	8986	6683	0.001289	0.006229	0	476	0	0	0	0	1	-360	360;
	8986	3569	0.00555	0.035419	0	438	0	0	0	0	1	-360	360;
	8986	3569	0.00381	0.0328	0	0	0	0	0	0	1	-360	360;
	4553	5905	0.00093	0.006841	0	0	0	0	0	0	1	-360	360;
	4553	5905	0.000909	0.006979	0	0	0	0	0	0	1	-360	360;
	3073	5576	0.001091	0.0082	0	514	0	0	0	0	1	-360	360;
	3073	4745	0.002081	0.020419	0	495	0	0	0	0	1	-360	360;
	4284	3432	0.00082	0.00276	0	0	0	0	0	0	1	-360	360;
	4284	7027	0.00162	0.007021	0	0	0	0	0	0	1	-360	360;
	7872	5844	0.002831	0.015021	0	572	0	0	0	0	1	-360	360;
	7872	135	0.006781	0.03026	0	514	0	0	0	0	1	-360	360;
	1557	3277	0.00041	0.00523	0	1777	0	0	0	0	1	-360	360;
	1557	4201	0.00062	0.0098	0	1810	0	0	0	0	1	-360	360;
	1557	4201	0.00083	0.00975	0	1843	0	0	0	0	1	-360	360;
	3569	8598	0.00131	0.01564	0	0	0	0	0	0	1	-360	360;
	3569	7219	0.002039	0.026831	0	0	0	0	0	0	1	-360	360;
	3569	4745	0.001031	0.00945	0	0	0	0	0	0	1	-360	360;
	3569	3742	0.001789	0.015211	0	0	0	0	0	0	1	-360	360;
	1924	7219	0.003021	0.03739	0	0	0	0	0	0	1	-360	360;
	1924	382	0.004211	0.04031	0	0	0	0	0	0	1	-360	360;
	1924	3550	0.004979	0.029409	0	610	0	0	0	0	1	-360	360;
	1924	3550	0.00488	0.029979	0	0	0	0	0	0	1	-360	360;
	1924	2392	0.002211	0.02399	0	0	0	0	0	0	1	-360	360;
	1924	2392	0.00231	0.023021	0	0	0	0	0	0	1	-360	360;
	464	7027	0.00719	0.046841	0	553	0	0	0	0	1	-360	360;
	1225	5844	0.00155	0.00968	0	476	0	0	0	0	1	-360	360;
	1225	5844	0.0015	0.01051	0	0	0	0	0	0	1	-360	360;
	1225	3041	0.00169	0.013401	0	0	0	0	0	0	1	-360	360;
	1225	2556	0.001659	0.012409	0	629	0	0	0	0	1	-360	360;
	2715	7027	0.005091	0.0167	0	0	0	0	0	0	1	-360	360;
	2715	2290	0.009919	0.04649	0	553	0	0	0	0	1	-360	360;
	2054	5242	9.1e-05	0.000531	0	0	0	0	0	0	1	-360	360;
	5844	7219	0.00982	0.045331	0	553	0	0	0	0	1	-360	360;
	5844	9169	0.003039	0.02964	0	0	0	0	0	0	1	-360	360;
	1915	3973	0.004159	0.02051	0	381	0	0	0	0	1	-360	360;
	3432	7027	0.00087	0.00469	0	0	0	0	0	0	1	-360	360;
	5576	582	0.0017	0.021039	0	400	0	0	0	0	1	-360	360;
	3277	219	0.00019	0.0032	0	889	0	0	0	0	1	-360	360;
	2487	2926	0.006979	0.03975	0	495	0	0	0	0	1	-360	360;
	2487	582	0.000521	0.004081	0	0	0	0	0	0	1	-360	360;
	2487	1062	0.00382	0.02214	0	553	0	0	0	0	1	-360	360;
	1369	4201	0.00204	0.02487	0	1678	0	0	0	0	1	-360	360;
	6520	4211	0.00312	0.018469	0	572	0	0	0	0	1	-360	360;
	2926	7027	0.005479	0.038841	0	553	0	0	0	0	1	-360	360;
	135	7219	0.000789	0.007729	0	0	0	0	0	0	1	-360	360;
	5778	7239	0.001789	0.01618	0	572	0	0	0	0	1	-360	360;
	5778	7239	0.00175	0.016479	0	553	0	0	0	0	1	-360	360;
	7239	3973	0.00375	0.025099	0	0	0	0	0	0	1	-360	360;
	1615	6433	0.010778	0.035609	0	0	0	0	0	0	1	-360	360;
	6433	1615	0.0096	0.04036	0	0	0	0	0	0	1	-360	360;
	7936	4176	0.001031	0.016919	0	0	0	0	0	0	1	-360	360;
	7936	4176	0.00101	0.01725	0	0	0	0	0	0	1	-360	360;
	7219	4876	0.00124	0.013169	0	0	0	0	0	0	1	-360	360;
	7219	4876	0.00119	0.01356	0	591	0	0	0	0	1	-360	360;
	7219	3742	0.00119	0.013281	0	0	0	0	0	0	1	-360	360;
	4342	7467	0.00306	0.02405	0	0	0	0	0	0	1	-360	360;
	4342	5225	0.002469	0.018081	0	629	0	0	0	0	1	-360	360;
	7732	6259	0.0018	0.014651	0	0	0	0	0	0	1	-360	360;
	7732	2392	0.001081	0.010219	0	0	0	0	0	0	1	-360	360;
	6259	7529	0.000669	0.003521	0	0	0	0	0	0	1	-360	360;
	7027	356	0.001039	0.0142	0	0	0	0	0	0	1	-360	360;
	7027	1789	0.002771	0.019919	0	0	0	0	0	0	1	-360	360;
	3407	4876	0.00295	0.013401	0	0	0	0	0	0	1	-360	360;
	4876	2257	0.00411	0.02488	0	0	0	0	0	0	1	-360	360;
	4876	8488	0.00061	0.007211	0	0	0	0	0	0	1	-360	360;
	1062	811	0.002219	0.011081	0	0	0	0	0	0	1	-360	360;
	7153	2963	6e-05	0.00066	0	757	0	0	0	0	1	-360	360;
	4211	2556	0.00411	0.026081	0	629	0	0	0	0	1	-360	360;
	9235	6666	0.00168	0.02222	0	1547	0	0	0	0	1	-360	360;
	9235	2963	0.00049	0.0061	0	1843	0	0	0	0	1	-360	360;
	2257	6335	0.00082	0.00575	0	0	0	0	0	0	1	-360	360;
	7896	2290	0.001409	0.01918	0	0	0	0	0	0	1	-360	360;
	5795	8276	0.001841	0.011159	0	0	0	0	0	0	1	-360	360;
	4079	2963	0.00109	0.01044	0	0	0	0	0	0	1	-360	360;
	4830	1141	0.00088	0.01608	0	1909	0	0	0	0	1	-360	360;
	4176	4369	0.000591	0.007781	0	0	0	0	0	0	1	-360	360;
	4176	3088	0.001901	0.0247	0	0	0	0	0	0	1	-360	360;
	7785	6817	0.00382	0.018401	0	0	0	0	0	0	1	-360	360;
	7529	6817	0.004169	0.02775	0	0	0	0	0	0	1	-360	360;
	7529	2392	0.001849	0.01889	0	0	0	0	0	0	1	-360	360;
	2290	3088	0.00076	0.01036	0	0	0	0	0	0	1	-360	360;
	2005	688	0.00207	0.009849	0	0	0	0	0	0	1	-360	360;
	7817	382	0.001849	0.013841	0	0	0	0	0	0	1	-360	360;
	7817	382	0.00243	0.015771	0	0	0	0	0	0	1	-360	360;
	6666	1653	0.00128	0.01149	0	1678	0	0	0	0	1	-360	360;
	2191	721	0.00057	0.005961	0	0	0	0	0	0	1	-360	360;
	2191	6375	0.00218	0.015099	0	0	0	0	0	0	1	-360	360;
	382	5242	0.00131	0.008771	0	553	0	0	0	0	1	-360	360;
	721	6375	0.00163	0.008729	0	0	0	0	0	0	1	-360	360;
	1406	4700	0.004831	0.02588	0	533	0	0	0	0	1	-360	360;
	227	3124	0.00062	0.00857	0	0	0	0	0	0	1	-360	360;
	4369	5905	0.002281	0.0362	0	0	0	0	0	0	1	-360	360;
	1789	9169	0.00132	0.01036	0	0	0	0	0	0	1	-360	360;
	4700	2392	0.00281	0.017099	0	0	0	0	0	0	1	-360	360;
	2392	1603	0.00424	0.0258	0	0	0	0	0	0	1	-360	360;
	711	7953	0.00036	0.00313	0	1744	0	0	0	0	1	-360	360;
	6088	9060	0.00112	0.0114	0	1678	0	0	0	0	1	-360	360;
	6088	5988	0.00046	0.007	0	1053	0	0	0	0	1	-360	360;
	6088	372	0.00113	0.01516	0	1448	0	0	0	0	1	-360	360;
	9060	2153	0.00138	0.01478	0	1185	0	0	0	0	1	-360	360;
	9060	518	0.00107	0.01447	0	1448	0	0	0	0	1	-360	360;
	6099	1387	0.012651	0.08939	0	0	0	0	0	0	1	-360	360;
	6099	6014	0.00957	0.05424	0	0	0	0	0	0	1	-360	360;
	1307	3446	0.0017	0.01823	0	987	0	0	0	0	1	-360	360;
	1307	6007	0.00092	0.00892	0	1283	0	0	0	0	1	-360	360;
	1307	1884	0.00063	0.00829	0	1251	0	0	0	0	1	-360	360;
	2153	8728	0.00052	0.00757	0	921	0	0	0	0	1	-360	360;
	8728	6569	0.00189	0.02546	0	987	0	0	0	0	1	-360	360;
	8728	4195	0.00044	0.00602	0	987	0	0	0	0	1	-360	360;
	144	5620	0.00138	0.01477	0	1777	0	0	0	0	1	-360	360;
	5620	6468	0.00055	0.00597	0	1645	0	0	0	0	1	-360	360;
	372	3808	0.00047	0.00684	0	1613	0	0	0	0	1	-360	360;
	1387	3375	0.0038	0.026331	0	476	0	0	0	0	1	-360	360;
	2153	2465	0.001	0.00937	0	823	0	0	0	0	1	-360	360;
	2153	372	0.00203	0.01819	0	1613	0	0	0	0	1	-360	360;
	6014	6373	0.002419	0.015021	0	419	0	0	0	0	1	-360	360;
	6014	6373	0.00274	0.014719	0	0	0	0	0	0	1	-360	360;
	6014	3599	0.014599	0.077461	0	0	0	0	0	0	1	-360	360;
	6014	7107	0.012461	0.05613	0	0	0	0	0	0	1	-360	360;
	6014	3375	0.014581	0.083849	0	0	0	0	0	0	1	-360	360;
	7114	6569	0.0006	0.00665	0	1547	0	0	0	0	1	-360	360;
	7114	6569	0.00057	0.00856	0	1316	0	0	0	0	1	-360	360;
	7114	6007	0.00234	0.02648	0	0	0	0	0	0	1	-360	360;
	2465	3920	0.00019	0.00167	0	889	0	0	0	0	1	-360	360;
	2465	2098	0.00019	0.00164	0	823	0	0	0	0	1	-360	360;
	6007	8279	3e-05	0.00047	0	757	0	0	0	0	1	-360	360;
	6007	844	3e-05	0.00048	0	757	0	0	0	0	1	-360	360;
	6251	4046	0.00097	0.01084	0	954	0	0	0	0	1	-360	360;
	6007	5855	8e-05	0.00073	0	1086	0	0	0	0	1	-360	360;
	6007	4969	8e-05	0.0006	0	1020	0	0	0	0	1	-360	360;
	6007	1555	7e-05	0.00067	0	1020	0	0	0	0	1	-360	360;
	6007	714	7e-05	0.00066	0	1053	0	0	0	0	1	-360	360;
	3599	562	0.00145	0.00707	0	0	0	0	0	0	1	-360	360;
	3599	3389	0.00164	0.00693	0	0	0	0	0	0	1	-360	360;
	372	7214	8e-05	0.00084	0	1119	0	0	0	0	1	-360	360;
	372	5412	0.0001	0.00083	0	1152	0	0	0	0	1	-360	360;
	372	8156	0.0001	0.00081	0	1119	0	0	0	0	1	-360	360;
	3290	6366	0.00019	0.00215	0	757	0	0	0	0	1	-360	360;
	144	3910	0.00037	0.00292	0	1119	0	0	0	0	1	-360	360;
	144	2084	0.00033	0.00332	0	757	0	0	0	0	1	-360	360;
	3551	4076	0.002031	0.01382	0	457	0	0	0	0	1	-360	360;
	3551	2590	0.001979	0.014091	0	0	0	0	0	0	1	-360	360;
	6569	2973	4e-05	0.00052	0	1843	0	0	0	0	1	-360	360;
	6569	1138	4e-05	0.00054	0	1678	0	0	0	0	1	-360	360;
	3551	7376	0.00136	0.0067	0	0	0	0	0	0	1	-360	360;
	3290	4400	0.0001	0.00078	0	1119	0	0	0	0	1	-360	360;
	3290	1606	0.00011	0.00076	0	1086	0	0	0	0	1	-360	360;
	3446	3290	0.00376	0.04872	0	1448	0	0	0	0	1	-360	360;
	3290	5882	0.0002	0.00258	0	1185	0	0	0	0	1	-360	360;
	3290	372	0.00089	0.0087	0	1086	0	0	0	0	1	-360	360;
	291	3551	0.01282	0.053159	0	0	0	0	0	0	1	-360	360;
	291	6995	0.002219	0.01038	0	457	0	0	0	0	1	-360	360;
	291	3551	0.01307	0.05205	0	0	0	0	0	0	1	-360	360;
	291	3375	0.00449	0.01863	0	457	0	0	0	0	1	-360	360;
	144	4195	0.00215	0.0248	0	1744	0	0	0	0	1	-360	360;
	2357	6251	0.00192	0.02113	0	1053	0	0	0	0	1	-360	360;
	2357	7953	0.00165	0.01704	0	757	0	0	0	0	1	-360	360;
	6569	6468	0.00168	0.02469	0	1053	0	0	0	0	1	-360	360;
	6569	4195	0.00222	0.02324	0	1283	0	0	0	0	1	-360	360;
	6251	518	0.00173	0.01682	0	1119	0	0	0	0	1	-360	360;
	6251	5343	0.00174	0.01872	0	1349	0	0	0	0	1	-360	360;
	6251	2465	0.00194	0.02194	0	1218	0	0	0	0	1	-360	360;
	9230	3340	0.013831	0.056841	0	0	0	0	0	0	1	-360	360;
	9230	3340	0.014091	0.055599	0	0	0	0	0	0	1	-360	360;
	3599	202	0.009081	0.05263	0	400	0	0	0	0	1	-360	360;
	3599	5468	0.007401	0.05661	0	419	0	0	0	0	1	-360	360;
	7953	5343	0.00164	0.01627	0	1382	0	0	0	0	1	-360	360;
	7107	2588	0.022909	0.124591	0	381	0	0	0	0	1	-360	360;
	2989	5343	0.00074	0.00791	0	1448	0	0	0	0	1	-360	360;
	2989	1884	0.00136	0.01726	0	1152	0	0	0	0	1	-360	360;
	202	3551	0.011211	0.07026	0	381	0	0	0	0	1	-360	360;
	202	3551	0.01143	0.06882	0	0	0	0	0	0	1	-360	360;
	5343	4195	0.00544	0.06455	0	889	0	0	0	0	1	-360	360;
	3340	2588	0.016289	0.066719	0	381	0	0	0	0	1	-360	360;
	3340	2588	0.016591	0.065271	0	381	0	0	0	0	1	-360	360;
	6007	1884	0.00077	0.0096	0	1349	0	0	0	0	1	-360	360;
	6007	1884	0.00093	0.00905	0	1316	0	0	0	0	1	-360	360;
	2588	5468	0.021729	0.15268	0	0	0	0	0	0	1	-360	360;
	488	6995	0.00064	0.002919	0	0	0	0	0	0	1	-360	360;
	6995	3375	0.002211	0.009169	0	457	0	0	0	0	1	-360	360;
	4107	9090	0.00014	0.001729	0	0	0	0	0	0	1	-360	360;
	5021	9090	0.0002	0.00281	0	0	0	0	0	0	1	-360	360;
	2657	5514	0.000711	0.004289	0	629	0	0	0	0	1	-360	360;
	2657	6986	0.0007	0.00438	0	610	0	0	0	0	1	-360	360;
	4107	2657	0.00032	0.00437	0	629	0	0	0	0	1	-360	360;
	5514	5784	7e-05	0.000401	0	610	0	0	0	0	1	-360	360;
	6986	8528	7e-05	0.000401	0	572	0	0	0	0	1	-360	360;
	818	5784	0.00189	0.0083	0	629	0	0	0	0	1	-360	360;
	818	8528	0.00168	0.009401	0	572	0	0	0	0	1	-360	360;
	4107	7194	0.000151	0.00169	0	629	0	0	0	0	1	-360	360;
	6234	7410	0	0.0002	0	0	0	0	0	0	1	-360	360;
	6234	7410	0.001	0.120531	0	0	0	0	0	0	1	-360	360;
	6451	6234	4.583678	10.16446	0	0	0	0	0	0	1	-360	360;
	6234	5680	8.430371	11.9126	0	0	0	0	0	0	1	-360	360;
	6234	2895	8.911157	11.64008	0	0	0	0	0	0	1	-360	360;
	8022	6234	7.455785	17.24773	0	0	0	0	0	0	1	-360	360;
	260	6234	9.746074	17.22748	0	0	0	0	0	0	1	-360	360;
	5909	6234	0.796777	2.654959	0	0	0	0	0	0	1	-360	360;
	2828	6234	0.960248	2.559091	0	0	0	0	0	0	1	-360	360;
	5596	6234	5.215909	9.009917	0	0	0	0	0	0	1	-360	360;
	4099	6234	5.097107	9.164256	0	0	0	0	0	0	1	-360	360;
	629	6234	19.22748	20.93388	0	0	0	0	0	0	1	-360	360;
	3423	6234	17.18905	24.75	0	0	0	0	0	0	1	-360	360;
	6234	2955	17.2064	22.05785	0	0	0	0	0	0	1	-360	360;
	8182	6234	0.227459	1.377252	0	0	0	0	0	0	1	-360	360;
	6718	6234	0.13382	0.643926	0	0	0	0	0	0	1	-360	360;
	4702	6234	0.211219	1.206674	0	0	0	0	0	0	1	-360	360;
	7500	6234	0.230165	1.120413	0	0	0	0	0	0	1	-360	360;
	1067	6234	29.19835	48.54959	0	0	0	0	0	0	1	-360	360;
	3533	6234	0.13231	0.939504	0	0	0	0	0	0	1	-360	360;
	5440	6234	0.842851	2.338843	0	0	0	0	0	0	1	-360	360;
	2605	6234	0.820764	2.471488	0	0	0	0	0	0	1	-360	360;
	1971	6234	0.553244	1.973719	0	0	0	0	0	0	1	-360	360;
	1971	6234	0.01405	0.0548	0	419	0	0	0	0	1	-360	360;
	6234	1581	1.071157	2.739256	0	0	0	0	0	0	1	-360	360;
	6234	1581	0.00774	0.04075	0	438	0	0	0	0	1	-360	360;
	6655	6234	16.64566	21.19835	0	0	0	0	0	0	1	-360	360;
	6307	6234	16.22913	16.64318	0	0	0	0	0	0	1	-360	360;
	6451	7410	10.87128	19.07149	0	0	0	0	0	0	1	-360	360;
	7410	5680	15.62583	25.67355	0	0	0	0	0	0	1	-360	360;
	7410	2895	16.95331	27.19835	0	0	0	0	0	0	1	-360	360;
	8022	7410	16.6593	34.37603	0	0	0	0	0	0	1	-360	360;
	260	7410	19.6814	37.93595	0	0	0	0	0	0	1	-360	360;
	5909	7410	1.881012	4.967975	0	0	0	0	0	0	1	-360	360;
	2828	7410	1.902851	5.765909	0	0	0	0	0	0	1	-360	360;
	5596	7410	10.14587	20.68595	0	0	0	0	0	0	1	-360	360;
	4099	7410	10.93058	18.98595	0	0	0	0	0	0	1	-360	360;
	629	7410	36.07851	50.20248	0	0	0	0	0	0	1	-360	360;
	3423	7410	40.72727	46.32438	0	0	0	0	0	0	1	-360	360;
	7410	2955	36.24587	46.68595	0	0	0	0	0	0	1	-360	360;
	8182	7410	0.444855	2.8	0	0	0	0	0	0	1	-360	360;
	6718	7410	0.273946	1.391054	0	0	0	0	0	0	1	-360	360;
	4702	7410	0.514545	2.160744	0	0	0	0	0	0	1	-360	360;
	7500	7410	0.450992	2.507851	0	0	0	0	0	0	1	-360	360;
	3533	7410	0.292665	1.871012	0	0	0	0	0	0	1	-360	360;
	5440	7410	1.795062	4.826033	0	0	0	0	0	0	1	-360	360;
	2605	7410	1.551219	5.755991	0	0	0	0	0	0	1	-360	360;
	1971	7410	1.089959	4.426653	0	0	0	0	0	0	1	-360	360;
	7410	1581	0.00757	0.037729	0	495	0	0	0	0	1	-360	360;
	7410	1581	2.158264	6.029752	0	0	0	0	0	0	1	-360	360;
	6655	7410	35.04132	44.83471	0	0	0	0	0	0	1	-360	360;
	6307	7410	30.36983	39.79132	0	0	0	0	0	0	1	-360	360;
	5401	4334	0.00058	0.00554	0	1316	0	0	0	0	1	-360	360;
	5401	5941	0.00116	0.01169	0	1382	0	0	0	0	1	-360	360;
	6284	4334	0.00052	0.00627	0	1448	0	0	0	0	1	-360	360;
	6284	5941	0.00118	0.01144	0	1448	0	0	0	0	1	-360	360;
	5401	6284	0.001	0.223012	0	757	0	0	0	0	1	-360	360;
	5401	4334	0.05072	0.617493	0	757	0	0	0	0	1	-360	360;
	5401	5941	0.05373	0.600208	0	757	0	0	0	0	1	-360	360;
	324	5401	0.381143	1.437258	0	757	0	0	0	0	1	-360	360;
	6284	4334	0.04991	0.630229	0	757	0	0	0	0	1	-360	360;
	6284	5941	0.05131	0.624993	0	757	0	0	0	0	1	-360	360;
	324	6284	0.389938	1.410526	0	757	0	0	0	0	1	-360	360;
	4492	5178	0.00059	0.00463	0	0	0	0	0	0	1	-360	360;
	4492	1596	0.00047	0.0038	0	0	0	0	0	0	1	-360	360;
	6272	7643	0.0006	0.00453	0	0	0	0	0	0	1	-360	360;
	6272	5178	0.03442	0.298892	0	757	0	0	0	0	1	-360	360;
	6272	7643	0.03377	0.305229	0	757	0	0	0	0	1	-360	360;
	6272	6615	0.00047	0.00372	0	0	0	0	0	0	1	-360	360;
	4472	600	0.00036	0.00341	0	1185	0	0	0	0	1	-360	360;
	4023	4472	0.00164	0.01639	0	1185	0	0	0	0	1	-360	360;
	7786	6212	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7786	6212	0.001	0.13669	0	0	0	0	0	0	1	-360	360;
	7786	7711	2.103099	5.086777	0	0	0	0	0	0	1	-360	360;
	7786	7198	0.004781	0.02089	0	495	0	0	0	0	1	-360	360;
	7786	4789	46.67975	63.95455	0	0	0	0	0	0	1	-360	360;
	7786	7592	45.98967	55.45455	0	0	0	0	0	0	1	-360	360;
	7786	6421	1.337128	3.255992	0	0	0	0	0	0	1	-360	360;
	7786	1085	0.09056	0.365041	0	0	0	0	0	0	1	-360	360;
	7786	3913	0.085081	0.477665	0	0	0	0	0	0	1	-360	360;
	7786	3913	0.008	0.040669	0	419	0	0	0	0	1	-360	360;
	6212	7711	2.446281	4.115289	0	0	0	0	0	0	1	-360	360;
	6212	98	0.00488	0.020469	0	591	0	0	0	0	1	-360	360;
	6212	4789	44.43595	63.26653	0	0	0	0	0	0	1	-360	360;
	6212	7592	47.34298	50.52893	0	0	0	0	0	0	1	-360	360;
	6212	6421	1.382355	2.978719	0	0	0	0	0	0	1	-360	360;
	6212	1085	0.07226	0.393864	0	0	0	0	0	0	1	-360	360;
	6212	1085	0.007831	0.04144	0	419	0	0	0	0	1	-360	360;
	6212	3913	0.08989	0.428161	0	0	0	0	0	0	1	-360	360;
	3727	7971	0.00055	0.00435	0	1810	0	0	0	0	1	-360	360;
	2260	7971	0.00128	0.00706	0	1909	0	0	0	0	1	-360	360;
	6160	5555	0.00053	0.00407	0	1283	0	0	0	0	1	-360	360;
	6965	5555	0.00056	0.00478	0	1711	0	0	0	0	1	-360	360;
	8342	3727	0.208393	1.089612	0	757	0	0	0	0	1	-360	360;
	2511	3727	0.199571	1.1759	0	757	0	0	0	0	1	-360	360;
	1161	3727	0.02232	0.346918	0	757	0	0	0	0	1	-360	360;
	3102	3727	0.858172	1.866898	0	757	0	0	0	0	1	-360	360;
	2016	3727	0.00031	0.00216	0	1547	0	0	0	0	1	-360	360;
	8517	3727	0.001	0.274529	0	757	0	0	0	0	1	-360	360;
	6724	3727	4.431302	8.136427	0	757	0	0	0	0	1	-360	360;
	8851	8553	0.36115	2.404917	0	757	0	0	0	0	1	-360	360;
	8851	192	0.319377	2.230125	0	757	0	0	0	0	1	-360	360;
	8851	3349	0.325519	2.183657	0	757	0	0	0	0	1	-360	360;
	5495	8851	1.726316	5.458795	0	757	0	0	0	0	1	-360	360;
	8851	6894	2.428047	7.275623	0	757	0	0	0	0	1	-360	360;
	8851	9100	0.00056	0.00716	0	1349	0	0	0	0	1	-360	360;
	8851	9100	0.527043	3.451801	0	757	0	0	0	0	1	-360	360;
	8851	5717	0.515582	3.196122	0	757	0	0	0	0	1	-360	360;
	8851	6222	1.334072	5.856163	0	757	0	0	0	0	1	-360	360;
	8851	8465	0.0011	0.01318	0	1547	0	0	0	0	1	-360	360;
	8851	8802	0.00108	0.01344	0	1580	0	0	0	0	1	-360	360;
	8851	5717	0.00063	0.00702	0	1613	0	0	0	0	1	-360	360;
	4012	1029	0.246612	1.335723	0	0	0	0	0	0	1	-360	360;
	4012	1029	0.008229	0.04962	0	476	0	0	0	0	1	-360	360;
	4012	3431	0.219628	1.572521	0	0	0	0	0	0	1	-360	360;
	4012	4033	4.089256	11.14855	0	0	0	0	0	0	1	-360	360;
	4012	1238	0.00943	0.04726	0	514	0	0	0	0	1	-360	360;
	4012	1238	3.689876	10.15083	0	0	0	0	0	0	1	-360	360;
	4012	438	0.202581	1.497417	0	0	0	0	0	0	1	-360	360;
	4012	1907	0.1973	1.490124	0	0	0	0	0	0	1	-360	360;
	4012	276	0.0028	0.013271	0	457	0	0	0	0	1	-360	360;
	4012	3330	0.12369	1.125744	0	0	0	0	0	0	1	-360	360;
	4012	1587	1.335496	4.761364	0	0	0	0	0	0	1	-360	360;
	4012	8938	1.15	4.888636	0	0	0	0	0	0	1	-360	360;
	4012	1177	1.052913	5.261983	0	0	0	0	0	0	1	-360	360;
	4012	3431	0.00932	0.04863	0	476	0	0	0	0	1	-360	360;
	4012	4033	0.00961	0.046281	0	514	0	0	0	0	1	-360	360;
	4012	5929	0.0025	0.015031	0	419	0	0	0	0	1	-360	360;
	4827	7601	0.0008	0.00519	0	438	0	0	0	0	1	-360	360;
	4827	1238	0.00612	0.03036	0	495	0	0	0	0	1	-360	360;
	7601	1854	0.00038	0.00243	0	457	0	0	0	0	1	-360	360;
	4753	6221	0.000909	0.005091	0	438	0	0	0	0	1	-360	360;
	4753	438	0.00187	0.00944	0	438	0	0	0	0	1	-360	360;
	6221	1854	0.00043	0.00238	0	457	0	0	0	0	1	-360	360;
	8693	4214	0.00013	0.00176	0	1415	0	0	0	0	1	-360	360;
	8693	4214	0.00013	0.00179	0	1580	0	0	0	0	1	-360	360;
	7035	1066	0.00018	0.00189	0	0	0	0	0	0	1	-360	360;
	7035	7175	0.0002	0.00185	0	0	0	0	0	0	1	-360	360;
	1066	1522	0.00039	0.00344	0	1876	0	0	0	0	1	-360	360;
	7175	5716	0.0004	0.00337	0	1942	0	0	0	0	1	-360	360;
	7035	8421	0.00027	0.00274	0	0	0	0	0	0	1	-360	360;
	7035	1253	0.00056	0.00437	0	0	0	0	0	0	1	-360	360;
	7035	6657	0.00026	0.00279	0	0	0	0	0	0	1	-360	360;
	6171	2788	0.001349	0.00763	0	0	0	0	0	0	1	-360	360;
	6171	2608	0.00374	0.02218	0	0	0	0	0	0	1	-360	360;
	6171	7321	0.000349	0.00186	0	0	0	0	0	0	1	-360	360;
	6171	2608	0.004651	0.02025	0	0	0	0	0	0	1	-360	360;
	6171	7321	0.00037	0.00182	0	0	0	0	0	0	1	-360	360;
	7502	4101	0.001	0.209093	0	757	0	0	0	0	1	-360	360;
	7502	8342	0.00062	0.00513	0	0	0	0	0	0	1	-360	360;
	7502	2511	0.04427	0.493158	0	757	0	0	0	0	1	-360	360;
	7502	1522	2.319321	5.378532	0	757	0	0	0	0	1	-360	360;
	4866	7502	2.762188	4.809072	0	757	0	0	0	0	1	-360	360;
	6724	7502	24.69598	23.43698	0	757	0	0	0	0	1	-360	360;
	7502	8429	1.383864	4.675693	0	757	0	0	0	0	1	-360	360;
	7816	7502	0.470817	2.500277	0	757	0	0	0	0	1	-360	360;
	1150	7502	0.00099	0.00946	0	0	0	0	0	0	1	-360	360;
	5428	7502	0.05862	0.781925	0	757	0	0	0	0	1	-360	360;
	4101	2511	0.04335	0.502652	0	757	0	0	0	0	1	-360	360;
	4101	2511	0.00063	0.00502	0	1678	0	0	0	0	1	-360	360;
	4101	1522	2.364751	5.268767	0	757	0	0	0	0	1	-360	360;
	4866	4101	2.814266	4.706787	0	757	0	0	0	0	1	-360	360;
	6724	4101	25.15305	22.92729	0	757	0	0	0	0	1	-360	360;
	4101	9084	0.00032	0.00223	0	0	0	0	0	0	1	-360	360;
	4101	8429	1.353809	4.329363	0	757	0	0	0	0	1	-360	360;
	7816	4101	0.460582	2.315097	0	757	0	0	0	0	1	-360	360;
	5428	4101	0.0574	0.796953	0	757	0	0	0	0	1	-360	360;
	8019	2102	0.00439	0.016341	0	0	0	0	0	0	1	-360	360;
	252	5719	0.004039	0.019031	0	533	0	0	0	0	1	-360	360;
	252	8498	0.001331	0.007669	0	514	0	0	0	0	1	-360	360;
	5719	7240	0.001091	0.005159	0	533	0	0	0	0	1	-360	360;
	8498	2071	0.00131	0.00782	0	457	0	0	0	0	1	-360	360;
	2071	2743	0.00449	0.022021	0	457	0	0	0	0	1	-360	360;
	8342	2511	0	0.0002	0	1448	0	0	0	0	1	-360	360;
	8342	2511	0.001	0.229342	0	757	0	0	0	0	1	-360	360;
	1161	8342	0.159093	0.880956	0	757	0	0	0	0	1	-360	360;
	3102	8342	0.11009	0.785942	0	757	0	0	0	0	1	-360	360;
	2016	8342	0.00133	0.01275	0	0	0	0	0	0	1	-360	360;
	8517	8342	0.08106	0.897645	0	757	0	0	0	0	1	-360	360;
	6724	8342	0.0009	0.00658	0	1514	0	0	0	0	1	-360	360;
	6724	8342	1.066967	3.005956	0	757	0	0	0	0	1	-360	360;
	2038	8342	0.00037	0.00335	0	1678	0	0	0	0	1	-360	360;
	1161	2511	0.14919	0.968975	0	757	0	0	0	0	1	-360	360;
	3102	2511	0.107472	0.83054	0	757	0	0	0	0	1	-360	360;
	2511	1522	6.591066	12.31994	0	757	0	0	0	0	1	-360	360;
	8517	2511	0.089861	0.931163	0	757	0	0	0	0	1	-360	360;
	8517	2511	0.00151	0.0125	0	0	0	0	0	0	1	-360	360;
	4866	2511	7.50831	11.99515	0	757	0	0	0	0	1	-360	360;
	6724	2511	57.05332	52.93698	0	757	0	0	0	0	1	-360	360;
	6724	2511	1.218213	2.697091	0	757	0	0	0	0	1	-360	360;
	6724	2511	0.00091	0.00644	0	1481	0	0	0	0	1	-360	360;
	8429	2511	5.391066	9.229224	0	757	0	0	0	0	1	-360	360;
	7816	2511	1.899723	5.672438	0	757	0	0	0	0	1	-360	360;
	1165	2511	0.00034	0.00341	0	1810	0	0	0	0	1	-360	360;
	5428	2511	0.372853	1.916066	0	757	0	0	0	0	1	-360	360;
	182	2912	0.001	0.398657	0	0	0	0	0	0	1	-360	360;
	182	2196	0.00157	0.00875	0	400	0	0	0	0	1	-360	360;
	1946	2912	0.013419	0.18632	0	0	0	0	0	0	1	-360	360;
	1946	2196	0.001771	0.00857	0	0	0	0	0	0	1	-360	360;
	4023	7102	3e-05	0.00035	0	1316	0	0	0	0	1	-360	360;
	4023	4645	4e-05	0.00047	0	1283	0	0	0	0	1	-360	360;
	8022	7711	6.406818	19.24731	0	0	0	0	0	0	1	-360	360;
	260	7711	5.408471	18.64401	0	0	0	0	0	0	1	-360	360;
	242	7711	4.777893	12.85021	0	0	0	0	0	0	1	-360	360;
	7711	1067	7.316529	22.04545	0	0	0	0	0	0	1	-360	360;
	7711	1067	0.00518	0.053229	0	400	0	0	0	0	1	-360	360;
	4789	7711	1.28	5.819421	0	0	0	0	0	0	1	-360	360;
	7592	7711	1.306178	5.47314	0	0	0	0	0	0	1	-360	360;
	7711	6421	0.001	0.244421	0	0	0	0	0	0	1	-360	360;
	1085	7711	0.326384	2.031736	0	0	0	0	0	0	1	-360	360;
	3913	7711	0.34876	2.295661	0	0	0	0	0	0	1	-360	360;
	7711	6421	0.000401	0.00324	0	419	0	0	0	0	1	-360	360;
	192	8553	1.103116	4.165166	0	757	0	0	0	0	1	-360	360;
	3349	8553	0.982756	4.720568	0	757	0	0	0	0	1	-360	360;
	5495	8553	17.77701	32.87327	0	757	0	0	0	0	1	-360	360;
	8553	6894	11.94321	24.78324	0	757	0	0	0	0	1	-360	360;
	9100	8553	0.13187	1.326939	0	757	0	0	0	0	1	-360	360;
	5717	8553	0.129169	1.352909	0	757	0	0	0	0	1	-360	360;
	6222	8553	15.47438	32.5367	0	757	0	0	0	0	1	-360	360;
	8693	8553	0.00019	0.00211	0	1251	0	0	0	0	1	-360	360;
	8693	8553	0.0002	0.00206	0	1218	0	0	0	0	1	-360	360;
	7684	3860	0.00013	0.001771	0	0	0	0	0	0	1	-360	360;
	6715	5323	0.008081	0.041651	0	400	0	0	0	0	1	-360	360;
	6715	5323	0.168091	0.83657	0	0	0	0	0	0	1	-360	360;
	6715	5323	0.008229	0.040771	0	438	0	0	0	0	1	-360	360;
	5301	5323	2.293802	5.98719	0	0	0	0	0	0	1	-360	360;
	2237	5323	2.825	5.516736	0	0	0	0	0	0	1	-360	360;
	5323	4350	0.731178	1.985165	0	0	0	0	0	0	1	-360	360;
	7811	6715	0.001349	0.01345	0	0	0	0	0	0	1	-360	360;
	648	7811	0.00051	0.00643	0	400	0	0	0	0	1	-360	360;
	7445	6715	0.00137	0.013159	0	0	0	0	0	0	1	-360	360;
	648	7445	0.0005	0.00655	0	381	0	0	0	0	1	-360	360;
	9037	6584	0.001	0.425381	0	757	0	0	0	0	1	-360	360;
	6497	6584	4.659141	9.407202	0	757	0	0	0	0	1	-360	360;
	4766	6584	1.022853	3.286357	0	757	0	0	0	0	1	-360	360;
	6584	4296	1.18795	4.384626	0	757	0	0	0	0	1	-360	360;
	6584	5611	1.213781	4.303324	0	757	0	0	0	0	1	-360	360;
	4647	6584	8.581024	12.40582	0	757	0	0	0	0	1	-360	360;
	7483	6584	0.203857	1.400623	0	757	0	0	0	0	1	-360	360;
	6584	4296	0.00082	0.0099	0	0	0	0	0	0	1	-360	360;
	7483	6584	0.00056	0.0066	0	1777	0	0	0	0	1	-360	360;
	4478	6584	0.00031	0.00634	0	1020	0	0	0	0	1	-360	360;
	6497	9037	4.02687	9.790858	0	757	0	0	0	0	1	-360	360;
	4766	9037	1.000554	3.350416	0	757	0	0	0	0	1	-360	360;
	9037	5611	0.00081	0.0101	0	1514	0	0	0	0	1	-360	360;
	9037	4296	1.418213	3.650623	0	757	0	0	0	0	1	-360	360;
	9037	5611	1.263504	4.137327	0	757	0	0	0	0	1	-360	360;
	4647	9037	8.415512	12.67105	0	757	0	0	0	0	1	-360	360;
	7483	9037	0.172188	1.486496	0	757	0	0	0	0	1	-360	360;
	5301	6715	0.008531	0.04201	0	457	0	0	0	0	1	-360	360;
	5301	6715	0.19088	1.317025	0	0	0	0	0	0	1	-360	360;
	2237	6715	0.20351	1.242748	0	0	0	0	0	0	1	-360	360;
	6715	7240	0.003419	0.02226	0	514	0	0	0	0	1	-360	360;
	6715	4350	0.05611	0.44343	0	0	0	0	0	0	1	-360	360;
	2237	6715	0.00869	0.041091	0	457	0	0	0	0	1	-360	360;
	6715	4153	0.003349	0.02268	0	495	0	0	0	0	1	-360	360;
	6978	6715	0.00176	0.018651	0	800	0	0	0	0	1	-360	360;
	7505	1918	0.00019	0.00245	0	757	0	0	0	0	1	-360	360;
	7505	1918	0.00019	0.00227	0	757	0	0	0	0	1	-360	360;
	5439	4296	0.00018	0.00209	0	0	0	0	0	0	1	-360	360;
	2222	5611	0.00017	0.00213	0	1744	0	0	0	0	1	-360	360;
	6753	424	0.00024	0.0025	0	1382	0	0	0	0	1	-360	360;
	6497	4766	0.897576	2.903878	0	757	0	0	0	0	1	-360	360;
	4647	6497	0.01432	0.301302	0	757	0	0	0	0	1	-360	360;
	5774	6497	0.00025	0.00245	0	1382	0	0	0	0	1	-360	360;
	6497	7483	3.381094	7.270776	0	757	0	0	0	0	1	-360	360;
	4709	3102	0.00142	0.01435	0	1777	0	0	0	0	1	-360	360;
	4709	1079	0.00023	0.00181	0	1810	0	0	0	0	1	-360	360;
	7954	3102	0.00139	0.01463	0	1843	0	0	0	0	1	-360	360;
	7954	6847	0.00023	0.00178	0	1843	0	0	0	0	1	-360	360;
	1161	6160	0.00047	0.00384	0	1218	0	0	0	0	1	-360	360;
	1161	2990	0.0005	0.00403	0	1645	0	0	0	0	1	-360	360;
	8517	2990	0.00025	0.00236	0	1251	0	0	0	0	1	-360	360;
	1161	3102	0.838573	2.37223	0	757	0	0	0	0	1	-360	360;
	8517	1161	0.078193	0.509841	0	757	0	0	0	0	1	-360	360;
	6724	1161	5.654848	7.590028	0	757	0	0	0	0	1	-360	360;
	7346	3102	0.0009	0.0069	0	1810	0	0	0	0	1	-360	360;
	6724	7346	0.0002	0.00149	0	0	0	0	0	0	1	-360	360;
	9152	3102	0.00092	0.00675	0	1876	0	0	0	0	1	-360	360;
	6724	9152	0.0002	0.00145	0	0	0	0	0	0	1	-360	360;
	8517	3102	0.413608	1.666759	0	757	0	0	0	0	1	-360	360;
	6724	3102	0.03562	0.272507	0	757	0	0	0	0	1	-360	360;
	4886	622	0.0001	0.00144	0	1744	0	0	0	0	1	-360	360;
	8693	4886	9e-05	0.00129	0	0	0	0	0	0	1	-360	360;
	1593	6611	0.0001	0.00133	0	0	0	0	0	0	1	-360	360;
	8693	1593	9e-05	0.00132	0	0	0	0	0	0	1	-360	360;
	622	8404	0.00063	0.00676	0	1942	0	0	0	0	1	-360	360;
	6611	2342	0.00065	0.00662	0	1777	0	0	0	0	1	-360	360;
	4334	8404	0.00017	0.00236	0	1645	0	0	0	0	1	-360	360;
	4334	2342	0.00017	0.00219	0	1744	0	0	0	0	1	-360	360;
	1242	8929	0.0017	0.007469	0	0	0	0	0	0	1	-360	360;
	8929	7198	0.001039	0.005599	0	514	0	0	0	0	1	-360	360;
	1242	7457	0.001729	0.00731	0	495	0	0	0	0	1	-360	360;
	7457	98	0.001021	0.0057	0	495	0	0	0	0	1	-360	360;
	1242	1790	0.00168	0.007159	0	419	0	0	0	0	1	-360	360;
	1790	7198	0.00112	0.00495	0	419	0	0	0	0	1	-360	360;
	8929	7457	0.001	0.364917	0	0	0	0	0	0	1	-360	360;
	1242	8929	0.106669	0.821178	0	0	0	0	0	0	1	-360	360;
	8929	7198	0.0977	0.935248	0	0	0	0	0	0	1	-360	360;
	8929	98	0.10481	0.756446	0	0	0	0	0	0	1	-360	360;
	1242	7457	0.094539	0.895083	0	0	0	0	0	0	1	-360	360;
	7457	7198	0.09563	0.921384	0	0	0	0	0	0	1	-360	360;
	7457	98	0.085479	0.810062	0	0	0	0	0	0	1	-360	360;
	4334	3631	0.0002	0.0021	0	1481	0	0	0	0	1	-360	360;
	3631	6538	0.00051	0.00613	0	1678	0	0	0	0	1	-360	360;
	7299	6152	0.00062	0.006331	0	0	0	0	0	0	1	-360	360;
	6152	1587	0.00075	0.00981	0	0	0	0	0	0	1	-360	360;
	4334	8614	0.00105	0.01369	0	0	0	0	0	0	1	-360	360;
	8614	491	0.00132	0.01291	0	1645	0	0	0	0	1	-360	360;
	4334	6137	0.00103	0.01395	0	1777	0	0	0	0	1	-360	360;
	6137	7545	0.00134	0.01263	0	0	0	0	0	0	1	-360	360;
	4334	5941	0.269647	1.156579	0	757	0	0	0	0	1	-360	360;
	324	4334	0.06211	0.718075	0	757	0	0	0	0	1	-360	360;
	4334	590	0.00062	0.005	0	1580	0	0	0	0	1	-360	360;
	5005	4334	0.00031	0.00248	0	1744	0	0	0	0	1	-360	360;
	4334	2394	0.0006	0.0051	0	1678	0	0	0	0	1	-360	360;
	818	2743	0.007331	0.044461	0	457	0	0	0	0	1	-360	360;
	818	2743	0.00718	0.0453	0	457	0	0	0	0	1	-360	360;
	7299	375	0.0002	0.00211	0	629	0	0	0	0	1	-360	360;
	5021	375	0.000659	0.002909	0	629	0	0	0	0	1	-360	360;
	6657	7440	0.00088	0.0079	0	0	0	0	0	0	1	-360	360;
	8421	8006	0.00089	0.00774	0	0	0	0	0	0	1	-360	360;
	5716	4345	0.00089	0.00706	0	0	0	0	0	0	1	-360	360;
	1522	1861	0.0009	0.0069	0	0	0	0	0	0	1	-360	360;
	4866	1522	0.527957	1.890097	0	757	0	0	0	0	1	-360	360;
	6724	1522	6.434972	8.855956	0	757	0	0	0	0	1	-360	360;
	8429	1522	0.02292	0.469398	0	757	0	0	0	0	1	-360	360;
	7816	1522	0.03034	0.81482	0	757	0	0	0	0	1	-360	360;
	5428	1522	0.225173	1.343006	0	757	0	0	0	0	1	-360	360;
	4298	3361	0.0001	0.00122	0	1349	0	0	0	0	1	-360	360;
	9211	3361	0.00025	0.00221	0	1876	0	0	0	0	1	-360	360;
	92	6848	2.696676	13.13573	0	757	0	0	0	0	1	-360	360;
	92	5523	3.125	11.9903	0	757	0	0	0	0	1	-360	360;
	4023	92	0.00054	0.00664	0	1678	0	0	0	0	1	-360	360;
	2891	92	0.899169	5.002355	0	757	0	0	0	0	1	-360	360;
	4737	92	0.06172	1.437119	0	757	0	0	0	0	1	-360	360;
	63	92	0.521247	2.927216	0	757	0	0	0	0	1	-360	360;
	9211	92	0.00397	0.614771	0	757	0	0	0	0	1	-360	360;
	8336	92	0.956579	5.569321	0	757	0	0	0	0	1	-360	360;
	2724	92	0.071011	1.246122	0	757	0	0	0	0	1	-360	360;
	1268	92	22.47576	40.81994	0	757	0	0	0	0	1	-360	360;
	630	92	0.00141	0.00778	0	0	0	0	0	0	1	-360	360;
	6369	1861	0.00021	0.00179	0	0	0	0	0	0	1	-360	360;
	8429	4345	0.00021	0.00175	0	0	0	0	0	0	1	-360	360;
	5102	5178	0.00051	0.00411	0	0	0	0	0	0	1	-360	360;
	1811	7643	0.00052	0.00402	0	0	0	0	0	0	1	-360	360;
	5178	7643	0.00439	0.248677	0	757	0	0	0	0	1	-360	360;
	5301	2237	0	0.0002	0	0	0	0	0	0	1	-360	360;
	5301	2237	0.001	0.289959	0	0	0	0	0	0	1	-360	360;
	5301	3739	9.9e-05	0.000789	0	0	0	0	0	0	1	-360	360;
	5301	4350	0.13343	0.727438	0	0	0	0	0	0	1	-360	360;
	2237	4350	0.11368	0.773802	0	0	0	0	0	0	1	-360	360;
	4714	7512	0.0001	0.0011	0	0	0	0	0	0	1	-360	360;
	4921	4714	0.00069	0.00512	0	0	0	0	0	0	1	-360	360;
	8099	1452	0.0001	0.00102	0	0	0	0	0	0	1	-360	360;
	2444	8099	0.0007	0.005	0	0	0	0	0	0	1	-360	360;
	5682	3862	0.00031	0.00399	0	1580	0	0	0	0	1	-360	360;
	6848	5682	9e-05	0.00103	0	1547	0	0	0	0	1	-360	360;
	3862	3976	0.00021	0.00231	0	1547	0	0	0	0	1	-360	360;
	3976	2157	8e-05	0.00082	0	1415	0	0	0	0	1	-360	360;
	2157	8948	0.00016	0.00203	0	1448	0	0	0	0	1	-360	360;
	8948	7160	3e-05	0.00036	0	1382	0	0	0	0	1	-360	360;
	63	7160	0.00061	0.00766	0	1547	0	0	0	0	1	-360	360;
	6848	5523	0	0.0002	0	1909	0	0	0	0	1	-360	360;
	6848	5523	0.01257	0.43678	0	757	0	0	0	0	1	-360	360;
	2891	6848	22.73892	49	0	757	0	0	0	0	1	-360	360;
	4737	6848	4.844044	19.18698	0	757	0	0	0	0	1	-360	360;
	63	6848	0.001	1.752701	0	757	0	0	0	0	1	-360	360;
	9211	6848	0.908172	6.096537	0	757	0	0	0	0	1	-360	360;
	7082	6848	4.358657	9.756232	0	757	0	0	0	0	1	-360	360;
	8336	6848	25.29224	66.18213	0	757	0	0	0	0	1	-360	360;
	2724	6848	1.071399	5.397922	0	757	0	0	0	0	1	-360	360;
	1268	6848	0.155789	0.985526	0	757	0	0	0	0	1	-360	360;
	2891	5523	23.51524	49.62604	0	757	0	0	0	0	1	-360	360;
	4737	5523	5.179155	18.97368	0	757	0	0	0	0	1	-360	360;
	63	5523	0.0144	1.921053	0	757	0	0	0	0	1	-360	360;
	9211	5523	0.997091	6.027978	0	757	0	0	0	0	1	-360	360;
	7082	5523	4.636704	9.781162	0	757	0	0	0	0	1	-360	360;
	8336	5523	29.33518	59.12049	0	757	0	0	0	0	1	-360	360;
	2724	5523	1.098269	5.695983	0	757	0	0	0	0	1	-360	360;
	1268	5523	0.00085	0.00853	0	1777	0	0	0	0	1	-360	360;
	1268	5523	0.15653	1.070014	0	757	0	0	0	0	1	-360	360;
	6095	5941	0.00015	0.00148	0	0	0	0	0	0	1	-360	360;
	475	5941	0.00014	0.00167	0	1349	0	0	0	0	1	-360	360;
	324	5941	0.525672	1.551039	0	757	0	0	0	0	1	-360	360;
	6469	7941	0.00411	0.01861	0	495	0	0	0	0	1	-360	360;
	4153	6469	0.000969	0.005841	0	476	0	0	0	0	1	-360	360;
	7941	2743	0.004581	0.02156	0	495	0	0	0	0	1	-360	360;
	2632	6485	0.00026	0.00323	0	954	0	0	0	0	1	-360	360;
	2632	6485	0.00026	0.00329	0	987	0	0	0	0	1	-360	360;
	192	8465	6e-05	0.00065	0	1185	0	0	0	0	1	-360	360;
	8465	1582	0.00038	0.00377	0	1316	0	0	0	0	1	-360	360;
	3349	8802	6e-05	0.00066	0	1349	0	0	0	0	1	-360	360;
	8802	2046	0.00038	0.00369	0	1481	0	0	0	0	1	-360	360;
	192	3349	0.01122	0.71669	0	757	0	0	0	0	1	-360	360;
	5495	192	25.10803	39.16205	0	757	0	0	0	0	1	-360	360;
	192	6894	32.07964	43.87258	0	757	0	0	0	0	1	-360	360;
	192	9100	0.427673	3.90554	0	757	0	0	0	0	1	-360	360;
	192	5717	0.418573	3.979224	0	757	0	0	0	0	1	-360	360;
	6222	192	23.26662	36.54363	0	757	0	0	0	0	1	-360	360;
	5495	3349	25.59141	38.34626	0	757	0	0	0	0	1	-360	360;
	3349	6894	32.6849	42.93906	0	757	0	0	0	0	1	-360	360;
	3349	9100	0.409481	3.684488	0	757	0	0	0	0	1	-360	360;
	3349	5717	0.464072	3.610803	0	757	0	0	0	0	1	-360	360;
	6222	3349	23.70568	35.76593	0	757	0	0	0	0	1	-360	360;
	1029	3431	0	0.0002	0	400	0	0	0	0	1	-360	360;
	1029	3431	0.001	0.19151	0	0	0	0	0	0	1	-360	360;
	1029	4033	17.56446	20.66736	0	0	0	0	0	0	1	-360	360;
	1029	1238	13.87645	21.76446	0	0	0	0	0	0	1	-360	360;
	1029	438	1.145248	3.637603	0	0	0	0	0	0	1	-360	360;
	1029	1907	1.372004	3.186983	0	0	0	0	0	0	1	-360	360;
	3330	1029	0.119159	0.998905	0	0	0	0	0	0	1	-360	360;
	1029	1587	6.052479	9.575826	0	0	0	0	0	0	1	-360	360;
	8938	1029	4.474793	10.74483	0	0	0	0	0	0	1	-360	360;
	1177	1029	4.534711	10.47087	0	0	0	0	0	0	1	-360	360;
	3431	4033	18.49236	22.54545	0	0	0	0	0	0	1	-360	360;
	3431	1238	13.69401	22.82645	0	0	0	0	0	0	1	-360	360;
	3431	438	1.333636	3.866942	0	0	0	0	0	0	1	-360	360;
	3431	1907	1.179442	4.148554	0	0	0	0	0	0	1	-360	360;
	3330	3431	0.114281	1.035289	0	0	0	0	0	0	1	-360	360;
	3431	1587	5.617769	11.75372	0	0	0	0	0	0	1	-360	360;
	8938	3431	4.939669	11.00083	0	0	0	0	0	0	1	-360	360;
	1177	3431	4.903306	10.93058	0	0	0	0	0	0	1	-360	360;
	7690	1452	2e-05	0.00021	0	0	0	0	0	0	1	-360	360;
	7690	2074	0.00028	0.00263	0	0	0	0	0	0	1	-360	360;
	3270	7512	2e-05	0.00023	0	0	0	0	0	0	1	-360	360;
	3270	8868	0.00027	0.00268	0	0	0	0	0	0	1	-360	360;
	1242	7198	0.02601	0.665517	0	0	0	0	0	0	1	-360	360;
	1242	98	0.028591	0.528326	0	0	0	0	0	0	1	-360	360;
	7924	6554	0.01088	0.212832	0	757	0	0	0	0	1	-360	360;
	2440	6554	0.00032	0.00331	0	1580	0	0	0	0	1	-360	360;
	4261	6554	0.00031	0.00306	0	1218	0	0	0	0	1	-360	360;
	4766	237	0.00057	0.01002	0	0	0	0	0	0	1	-360	360;
	4766	2277	0.00053	0.00558	0	1942	0	0	0	0	1	-360	360;
	4766	4296	22.58241	30.56787	0	757	0	0	0	0	1	-360	360;
	4766	5611	22.12119	31.16759	0	757	0	0	0	0	1	-360	360;
	4647	4766	1.627632	4.119529	0	757	0	0	0	0	1	-360	360;
	8762	4766	0.0003	0.00319	0	1580	0	0	0	0	1	-360	360;
	7483	4766	0.00232	0.519501	0	757	0	0	0	0	1	-360	360;
	4766	9003	0.00058	0.00981	0	0	0	0	0	0	1	-360	360;
	9088	9003	0.00065	0.00643	0	0	0	0	0	0	1	-360	360;
	4501	237	0.00063	0.00655	0	0	0	0	0	0	1	-360	360;
	2016	8517	0.001	0.215229	0	757	0	0	0	0	1	-360	360;
	2016	7924	0.00102	0.00911	0	1481	0	0	0	0	1	-360	360;
	8517	6558	0.00094	0.00901	0	1744	0	0	0	0	1	-360	360;
	6724	8517	3.359072	5.253809	0	757	0	0	0	0	1	-360	360;
	1079	4723	0.00019	0.00221	0	1481	0	0	0	0	1	-360	360;
	7924	1079	0.00039	0.00355	0	1547	0	0	0	0	1	-360	360;
	1079	1148	0.0001	0.00073	0	0	0	0	0	0	1	-360	360;
	6847	600	0.00018	0.00226	0	1547	0	0	0	0	1	-360	360;
	8355	6847	0.00038	0.00368	0	987	0	0	0	0	1	-360	360;
	6847	7116	0.00047	0.00386	0	1843	0	0	0	0	1	-360	360;
	7198	98	0.001	0.309525	0	0	0	0	0	0	1	-360	360;
	2912	2196	0.000711	0.00393	0	553	0	0	0	0	1	-360	360;
	2912	7631	0.001219	0.00569	0	0	0	0	0	0	1	-360	360;
	8349	6286	0.00125	0.005581	0	457	0	0	0	0	1	-360	360;
	2196	6286	0.0007	0.00401	0	0	0	0	0	0	1	-360	360;
	6724	4866	0.0599	0.313109	0	757	0	0	0	0	1	-360	360;
	4866	3267	0.00051	0.00517	0	1744	0	0	0	0	1	-360	360;
	4866	8429	0.92403	2.090097	0	757	0	0	0	0	1	-360	360;
	4866	7816	0.05633	0.587403	0	757	0	0	0	0	1	-360	360;
	4866	5428	0.333109	1.178601	0	757	0	0	0	0	1	-360	360;
	5021	7194	0.0002	0.002599	0	381	0	0	0	0	1	-360	360;
	6508	4023	6e-05	0.00066	0	1843	0	0	0	0	1	-360	360;
	4023	4723	0.00203	0.01921	0	1415	0	0	0	0	1	-360	360;
	4023	8336	0.00132	0.01562	0	1645	0	0	0	0	1	-360	360;
	4023	1918	0.00039	0.00541	0	0	0	0	0	0	1	-360	360;
	4023	8336	0.0013	0.01592	0	1678	0	0	0	0	1	-360	360;
	4023	6667	0.00084	0.00852	0	0	0	0	0	0	1	-360	360;
	6508	3052	0.00016	0.0017	0	1547	0	0	0	0	1	-360	360;
	1807	5102	0.00092	0.00761	0	1678	0	0	0	0	1	-360	360;
	1807	491	0.00062	0.00468	0	1711	0	0	0	0	1	-360	360;
	6108	1811	0.00094	0.00746	0	1580	0	0	0	0	1	-360	360;
	6108	7545	0.00063	0.00458	0	1645	0	0	0	0	1	-360	360;
	5102	1811	0	0.0002	0	1547	0	0	0	0	1	-360	360;
	5102	1811	0.001	0.227327	0	757	0	0	0	0	1	-360	360;
	5102	3647	3e-05	0.00034	0	0	0	0	0	0	1	-360	360;
	7816	5102	0.00052	0.00501	0	0	0	0	0	0	1	-360	360;
	4041	1811	0.00059	0.00491	0	0	0	0	0	0	1	-360	360;
	241	1238	0.006229	0.029719	0	495	0	0	0	0	1	-360	360;
	241	438	0.00194	0.00924	0	476	0	0	0	0	1	-360	360;
	3118	2631	0.00189	0.00769	0	0	0	0	0	0	1	-360	360;
	3118	8934	0.0037	0.017021	0	476	0	0	0	0	1	-360	360;
	4597	1144	0.00168	0.008719	0	438	0	0	0	0	1	-360	360;
	4597	6173	0.003771	0.016669	0	514	0	0	0	0	1	-360	360;
	6440	7631	0.004669	0.024729	0	438	0	0	0	0	1	-360	360;
	1189	8349	0	0.0002	0	0	0	0	0	0	1	-360	360;
	1189	8349	0.00099	0.00607	0	0	0	0	0	0	1	-360	360;
	6440	1189	0.00476	0.02624	0	381	0	0	0	0	1	-360	360;
	1238	5820	0.00051	0.00286	0	457	0	0	0	0	1	-360	360;
	1238	5820	0.0005	0.002909	0	476	0	0	0	0	1	-360	360;
	4033	1238	0	0.0002	0	0	0	0	0	0	1	-360	360;
	4033	1238	0.001	0.18874	0	0	0	0	0	0	1	-360	360;
	4033	438	0.704421	2.986983	0	0	0	0	0	0	1	-360	360;
	4033	1907	0.680207	2.970661	0	0	0	0	0	0	1	-360	360;
	3330	4033	4.389256	10.44091	0	0	0	0	0	0	1	-360	360;
	4033	1587	0.00861	0.432252	0	0	0	0	0	0	1	-360	360;
	8938	4033	0.006521	0.039849	0	457	0	0	0	0	1	-360	360;
	8938	4033	0.450599	2.525826	0	0	0	0	0	0	1	-360	360;
	1177	4033	0.51345	2.178099	0	0	0	0	0	0	1	-360	360;
	4033	1962	0.000159	0.002091	0	0	0	0	0	0	1	-360	360;
	1238	438	0.547603	3.122934	0	0	0	0	0	0	1	-360	360;
	1238	1907	0.63281	2.590496	0	0	0	0	0	0	1	-360	360;
	3330	1238	4.541736	8.227686	0	0	0	0	0	0	1	-360	360;
	1238	1587	0.00687	0.424525	0	0	0	0	0	0	1	-360	360;
	8938	1238	0.394752	2.341736	0	0	0	0	0	0	1	-360	360;
	1177	1238	0.00638	0.04063	0	457	0	0	0	0	1	-360	360;
	1177	1238	0.459607	1.981798	0	0	0	0	0	0	1	-360	360;
	1238	1944	0.000159	0.00213	0	0	0	0	0	0	1	-360	360;
	2631	1144	0.001	0.319917	0	0	0	0	0	0	1	-360	360;
	3332	2631	0.074539	0.567996	0	0	0	0	0	0	1	-360	360;
	3332	1144	0.054219	0.438326	0	0	0	0	0	0	1	-360	360;
	1667	4119	0.00049	0.0046	0	1744	0	0	0	0	1	-360	360;
	8193	4119	0.00063	0.00676	0	1547	0	0	0	0	1	-360	360;
	1667	2306	0.02777	0.440042	0	757	0	0	0	0	1	-360	360;
	3514	2306	0.02704	0.444668	0	757	0	0	0	0	1	-360	360;
	3514	2306	0.00048	0.00469	0	1777	0	0	0	0	1	-360	360;
	8193	2306	0.00062	0.00626	0	1514	0	0	0	0	1	-360	360;
	1667	3514	0	0.0002	0	1415	0	0	0	0	1	-360	360;
	1667	3514	0.001	0.247292	0	757	0	0	0	0	1	-360	360;
	1667	5963	3e-05	0.00035	0	1514	0	0	0	0	1	-360	360;
	9046	1667	0.00014	0.00118	0	1251	0	0	0	0	1	-360	360;
	3514	7771	3e-05	0.00034	0	1514	0	0	0	0	1	-360	360;
	4498	5963	0.00072	0.00903	0	1514	0	0	0	0	1	-360	360;
	4498	7771	0.00082	0.00885	0	1678	0	0	0	0	1	-360	360;
	7440	969	0.00012	0.00093	0	757	0	0	0	0	1	-360	360;
	6946	7240	0.000831	0.004229	0	0	0	0	0	0	1	-360	360;
	9053	4153	0.000841	0.00414	0	0	0	0	0	0	1	-360	360;
	8762	6095	0.00049	0.00607	0	1876	0	0	0	0	1	-360	360;
	7483	2277	0.00048	0.00567	0	1547	0	0	0	0	1	-360	360;
	7483	475	0.00047	0.00578	0	1349	0	0	0	0	1	-360	360;
	5495	6222	6e-05	0.00069	0	1613	0	0	0	0	1	-360	360;
	5495	6894	0.198269	1.245429	0	757	0	0	0	0	1	-360	360;
	5495	9100	28.23615	48.46122	0	757	0	0	0	0	1	-360	360;
	5495	5717	32.00069	47.49169	0	757	0	0	0	0	1	-360	360;
	5495	6222	0.001	0.339079	0	757	0	0	0	0	1	-360	360;
	4296	5611	0.001	0.23876	0	757	0	0	0	0	1	-360	360;
	7483	4296	7.242382	14.44598	0	757	0	0	0	0	1	-360	360;
	7483	5611	7.08795	14.71814	0	757	0	0	0	0	1	-360	360;
	6451	5598	0.00493	0.059419	0	0	0	0	0	0	1	-360	360;
	6451	5598	0.01006	0.07863	0	514	0	0	0	0	1	-360	360;
	6451	5680	0.001	0.434587	0	0	0	0	0	0	1	-360	360;
	6451	2895	0.001	0.434132	0	0	0	0	0	0	1	-360	360;
	6451	4488	9.525206	20.12748	0	0	0	0	0	0	1	-360	360;
	6451	8022	0.490393	3.121901	0	0	0	0	0	0	1	-360	360;
	6451	260	0.324483	3.440909	0	0	0	0	0	0	1	-360	360;
	6451	5909	7.110331	13.27211	0	0	0	0	0	0	1	-360	360;
	6451	2828	8.147314	13.64236	0	0	0	0	0	0	1	-360	360;
	6451	5596	1.873657	4.730165	0	0	0	0	0	0	1	-360	360;
	6451	4099	1.953223	4.544628	0	0	0	0	0	0	1	-360	360;
	6451	629	0.893533	3.267562	0	0	0	0	0	0	1	-360	360;
	6451	3423	0.947417	3.197521	0	0	0	0	0	0	1	-360	360;
	6451	2955	1.735227	5.478926	0	0	0	0	0	0	1	-360	360;
	8182	6451	49.21281	47.44215	0	0	0	0	0	0	1	-360	360;
	6718	6451	21.70248	30.12603	0	0	0	0	0	0	1	-360	360;
	6451	4702	9.99835	24.78719	0	0	0	0	0	0	1	-360	360;
	6451	7500	10.26198	26.88017	0	0	0	0	0	0	1	-360	360;
	6451	1067	0.625785	4.196281	0	0	0	0	0	0	1	-360	360;
	3533	6451	0.531426	3.034091	0	0	0	0	0	0	1	-360	360;
	5440	6451	3.245868	10.84669	0	0	0	0	0	0	1	-360	360;
	2605	6451	3.282025	10.98388	0	0	0	0	0	0	1	-360	360;
	6451	1581	0.002919	0.01451	0	553	0	0	0	0	1	-360	360;
	6451	1581	0.001	0.67564	0	0	0	0	0	0	1	-360	360;
	6451	6655	0.227169	1.409215	0	0	0	0	0	0	1	-360	360;
	6451	6307	0.192229	1.194917	0	0	0	0	0	0	1	-360	360;
	6451	1581	0.002979	0.0142	0	553	0	0	0	0	1	-360	360;
	5167	5531	0.000151	0.0015	0	0	0	0	0	0	1	-360	360;
	5167	4350	0.000151	0.001461	0	0	0	0	0	0	1	-360	360;
	6485	1493	0.00136	0.0167	0	1415	0	0	0	0	1	-360	360;
	5680	2895	0	0.0002	0	0	0	0	0	0	1	-360	360;
	5680	2895	0.001	0.256632	0	0	0	0	0	0	1	-360	360;
	4488	5680	6.78595	11.8655	0	0	0	0	0	0	1	-360	360;
	8022	5680	1.097169	5.085744	0	0	0	0	0	0	1	-360	360;
	260	5680	0.904029	5.181818	0	0	0	0	0	0	1	-360	360;
	5909	5680	4.170248	8.645247	0	0	0	0	0	0	1	-360	360;
	2828	5680	4.773967	8.884503	0	0	0	0	0	0	1	-360	360;
	5596	5680	0.874421	3.172727	0	0	0	0	0	0	1	-360	360;
	4099	5680	1.05436	2.929339	0	0	0	0	0	0	1	-360	360;
	629	5680	0.513368	2.495455	0	0	0	0	0	0	1	-360	360;
	3423	5680	0.604525	2.218802	0	0	0	0	0	0	1	-360	360;
	2955	5680	1.308616	3.033264	0	0	0	0	0	0	1	-360	360;
	8182	5680	39.69008	38.26033	0	0	0	0	0	0	1	-360	360;
	6718	5680	17.11942	22.4938	0	0	0	0	0	0	1	-360	360;
	4702	5680	20.47789	32.19835	0	0	0	0	0	0	1	-360	360;
	7500	5680	21.35331	30.98554	0	0	0	0	0	0	1	-360	360;
	1067	5680	1.604091	6.167149	0	0	0	0	0	0	1	-360	360;
	3533	5680	1.137851	5.037397	0	0	0	0	0	0	1	-360	360;
	5440	5680	6.625413	15.09483	0	0	0	0	0	0	1	-360	360;
	2605	5680	6.196074	16.5874	0	0	0	0	0	0	1	-360	360;
	1971	5680	59.24587	56.3595	0	0	0	0	0	0	1	-360	360;
	5680	1581	0.086409	1.077231	0	0	0	0	0	0	1	-360	360;
	6655	5680	0.080841	0.916612	0	0	0	0	0	0	1	-360	360;
	6307	5680	0.07195	0.701488	0	0	0	0	0	0	1	-360	360;
	4488	2895	5.952479	12.55207	0	0	0	0	0	0	1	-360	360;
	8022	2895	1.208636	4.790289	0	0	0	0	0	0	1	-360	360;
	260	2895	0.897293	4.88781	0	0	0	0	0	0	1	-360	360;
	5909	2895	4.399793	8.436157	0	0	0	0	0	0	1	-360	360;
	2828	2895	4.466736	9.775413	0	0	0	0	0	0	1	-360	360;
	5596	2895	0.924855	3.415083	0	0	0	0	0	0	1	-360	360;
	4099	2895	1.008368	3.168595	0	0	0	0	0	0	1	-360	360;
	629	2895	0.519401	2.533471	0	0	0	0	0	0	1	-360	360;
	3423	2895	0.611653	2.252893	0	0	0	0	0	0	1	-360	360;
	2955	2895	1.180868	3.490496	0	0	0	0	0	0	1	-360	360;
	8182	2895	38.08471	41.4876	0	0	0	0	0	0	1	-360	360;
	6718	2895	19.69132	22.41116	0	0	0	0	0	0	1	-360	360;
	4702	2895	21.63017	31.43802	0	0	0	0	0	0	1	-360	360;
	7500	2895	20.91529	32.94215	0	0	0	0	0	0	1	-360	360;
	1067	2895	1.695537	6.03657	0	0	0	0	0	0	1	-360	360;
	3533	2895	1.302851	4.551033	0	0	0	0	0	0	1	-360	360;
	5440	2895	5.718182	16.36116	0	0	0	0	0	0	1	-360	360;
	2605	2895	6.683678	15.9062	0	0	0	0	0	0	1	-360	360;
	2895	1581	0.10661	0.987583	0	0	0	0	0	0	1	-360	360;
	6655	2895	0.079729	0.948512	0	0	0	0	0	0	1	-360	360;
	6307	2895	0.06449	0.807169	0	0	0	0	0	0	1	-360	360;
	6307	2895	0.002151	0.02656	0	0	0	0	0	0	1	-360	360;
	6485	4811	0.00094	0.01246	0	1711	0	0	0	0	1	-360	360;
	4811	4849	0.00109	0.00971	0	1942	0	0	0	0	1	-360	360;
	6485	3012	0.00099	0.01086	0	1678	0	0	0	0	1	-360	360;
	6485	3012	0.00092	0.0127	0	1613	0	0	0	0	1	-360	360;
	3012	4849	0.00097	0.01101	0	1909	0	0	0	0	1	-360	360;
	438	1907	0	0.0002	0	0	0	0	0	0	1	-360	360;
	438	1907	0.001	0.236901	0	0	0	0	0	0	1	-360	360;
	3330	438	1.033202	3.18595	0	0	0	0	0	0	1	-360	360;
	1587	438	0.07336	0.952996	0	0	0	0	0	0	1	-360	360;
	8938	438	4.823554	9.755785	0	0	0	0	0	0	1	-360	360;
	1177	438	4.344835	10.74711	0	0	0	0	0	0	1	-360	360;
	7684	438	0.000789	0.01039	0	0	0	0	0	0	1	-360	360;
	3330	1907	0.877293	3.230992	0	0	0	0	0	0	1	-360	360;
	1587	1907	0.08095	0.839008	0	0	0	0	0	0	1	-360	360;
	8938	1907	4.635744	9.702067	0	0	0	0	0	0	1	-360	360;
	1177	1907	4.775207	9.229752	0	0	0	0	0	0	1	-360	360;
	6558	2440	0.00017	0.0018	0	1942	0	0	0	0	1	-360	360;
	8355	4261	0.00017	0.00184	0	1251	0	0	0	0	1	-360	360;
	8647	2074	6e-05	0.00055	0	0	0	0	0	0	1	-360	360;
	5566	8868	5e-05	0.00051	0	0	0	0	0	0	1	-360	360;
	7093	7207	4.9e-05	0.000194	0	1645	0	0	0	0	1	-360	360;
	7093	7207	0.001	0.502417	0	757	0	0	0	0	1	-360	360;
	7093	491	0.11597	1.150069	0	757	0	0	0	0	1	-360	360;
	7093	4332	0.00027	0.00201	0	1645	0	0	0	0	1	-360	360;
	7093	324	12.69044	26.69945	0	757	0	0	0	0	1	-360	360;
	5862	7093	5.039543	12.54224	0	757	0	0	0	0	1	-360	360;
	4052	7093	4.861565	13.04917	0	757	0	0	0	0	1	-360	360;
	7207	491	0.131427	1.127078	0	757	0	0	0	0	1	-360	360;
	7207	2513	0.00034	0.00324	0	1777	0	0	0	0	1	-360	360;
	7207	324	14.34765	23.73407	0	757	0	0	0	0	1	-360	360;
	5862	7207	4.937535	12.78878	0	757	0	0	0	0	1	-360	360;
	4052	7207	4.758934	13.29571	0	757	0	0	0	0	1	-360	360;
	3330	276	0.00189	0.01011	0	495	0	0	0	0	1	-360	360;
	3330	7412	0.00193	0.009901	0	495	0	0	0	0	1	-360	360;
	321	7412	0.00095	0.004461	0	495	0	0	0	0	1	-360	360;
	321	5929	0.000841	0.00505	0	457	0	0	0	0	1	-360	360;
	6432	3330	0.00074	0.00837	0	419	0	0	0	0	1	-360	360;
	3330	1587	3.807645	6.613017	0	0	0	0	0	0	1	-360	360;
	8938	3330	0.512211	2.208058	0	0	0	0	0	0	1	-360	360;
	1177	3330	0.432727	2.575207	0	0	0	0	0	0	1	-360	360;
	3330	321	0.00551	0.13675	0	0	0	0	0	0	1	-360	360;
	8193	6615	0.00018	0.00165	0	0	0	0	0	0	1	-360	360;
	8193	1596	0.00017	0.00168	0	0	0	0	0	0	1	-360	360;
	3052	4298	0.00022	0.00207	0	1415	0	0	0	0	1	-360	360;
	7254	5878	6e-05	0.000531	0	0	0	0	0	0	1	-360	360;
	4488	4401	6e-05	0.000521	0	495	0	0	0	0	1	-360	360;
	4488	5909	2.538636	7.69814	0	0	0	0	0	0	1	-360	360;
	4488	2828	2.881818	7.882025	0	0	0	0	0	0	1	-360	360;
	4488	5596	0.648822	3.687603	0	0	0	0	0	0	1	-360	360;
	4488	4099	0.662004	3.615496	0	0	0	0	0	0	1	-360	360;
	4488	629	0.347417	1.771488	0	0	0	0	0	0	1	-360	360;
	4488	3423	0.302727	2.122521	0	0	0	0	0	0	1	-360	360;
	4488	2955	0.13874	1.055682	0	0	0	0	0	0	1	-360	360;
	8182	4488	32.21488	46.94215	0	0	0	0	0	0	1	-360	360;
	6718	4488	16.64897	25.36157	0	0	0	0	0	0	1	-360	360;
	4488	1581	31.40496	45.72727	0	0	0	0	0	0	1	-360	360;
	4488	6655	1.96531	5.971488	0	0	0	0	0	0	1	-360	360;
	4488	6307	1.628843	5.349793	0	0	0	0	0	0	1	-360	360;
	5878	2955	0.003831	0.04362	0	0	0	0	0	0	1	-360	360;
	4401	2955	0.003901	0.042711	0	0	0	0	0	0	1	-360	360;
	2891	4737	0	0.0002	0	1448	0	0	0	0	1	-360	360;
	2891	4737	0.03143	0.812673	0	757	0	0	0	0	1	-360	360;
	7352	2891	6e-05	0.00062	0	1613	0	0	0	0	1	-360	360;
	63	2891	5.540582	14.44598	0	757	0	0	0	0	1	-360	360;
	9211	2891	0.171863	2.799169	0	757	0	0	0	0	1	-360	360;
	7082	2891	31.73753	62.63988	0	757	0	0	0	0	1	-360	360;
	8336	2891	0.0037	0.670762	0	757	0	0	0	0	1	-360	360;
	2724	2891	4.69723	16.67659	0	757	0	0	0	0	1	-360	360;
	1268	2891	13.8608	37.26524	0	757	0	0	0	0	1	-360	360;
	5566	4737	6e-05	0.00063	0	0	0	0	0	0	1	-360	360;
	63	4737	0.842036	5.10651	0	757	0	0	0	0	1	-360	360;
	9211	4737	0.001	0.767382	0	757	0	0	0	0	1	-360	360;
	8336	4737	0.0122	0.920499	0	757	0	0	0	0	1	-360	360;
	2724	4737	0.831094	4.707341	0	757	0	0	0	0	1	-360	360;
	1268	4737	13.1392	39.18767	0	757	0	0	0	0	1	-360	360;
	7910	7352	0.00022	0.0023	0	1777	0	0	0	0	1	-360	360;
	8244	8647	0.00022	0.00213	0	0	0	0	0	0	1	-360	360;
	8332	6894	0.00034	0.003	0	0	0	0	0	0	1	-360	360;
	324	8332	0.00022	0.00172	0	0	0	0	0	0	1	-360	360;
	5048	6894	0.00034	0.00294	0	0	0	0	0	0	1	-360	360;
	2133	5048	0.00022	0.00168	0	0	0	0	0	0	1	-360	360;
	9100	6894	25.47853	44.51731	0	757	0	0	0	0	1	-360	360;
	5717	6894	24.93629	45.35734	0	757	0	0	0	0	1	-360	360;
	6222	6894	0.138622	1.351247	0	757	0	0	0	0	1	-360	360;
	8938	1587	1.420021	4.04938	0	0	0	0	0	0	1	-360	360;
	1177	1587	1.226446	4.635537	0	0	0	0	0	0	1	-360	360;
	1587	5820	0.000771	0.0083	0	0	0	0	0	0	1	-360	360;
	2788	9077	0.00449	0.023349	0	0	0	0	0	0	1	-360	360;
	640	7099	0.001	0.313037	0	0	0	0	0	0	1	-360	360;
	640	6440	0.025331	0.364938	0	0	0	0	0	0	1	-360	360;
	640	6440	0.004219	0.018211	0	0	0	0	0	0	1	-360	360;
	7099	6440	0.02581	0.357169	0	0	0	0	0	0	1	-360	360;
	7099	6440	0.00413	0.01686	0	553	0	0	0	0	1	-360	360;
	491	3476	0.00018	0.00157	0	0	0	0	0	0	1	-360	360;
	324	491	4.622922	11.33102	0	757	0	0	0	0	1	-360	360;
	5862	491	1.337465	6.217936	0	757	0	0	0	0	1	-360	360;
	4052	491	1.319252	5.765305	0	757	0	0	0	0	1	-360	360;
	7545	5272	0.00017	0.0016	0	0	0	0	0	0	1	-360	360;
	7545	4332	0.00017	0.00163	0	1777	0	0	0	0	1	-360	360;
	7545	3950	2e-05	0.00022	0	757	0	0	0	0	1	-360	360;
	8934	6173	0	0.0002	0	533	0	0	0	0	1	-360	360;
	8934	6173	0.01018	0.46657	0	0	0	0	0	0	1	-360	360;
	6130	8934	0.00887	0.046021	0	0	0	0	0	0	1	-360	360;
	3780	8934	0.007479	0.02551	0	553	0	0	0	0	1	-360	360;
	1623	6173	0.00762	0.024979	0	572	0	0	0	0	1	-360	360;
	7440	3476	0.00034	0.00277	0	0	0	0	0	0	1	-360	360;
	8006	5272	0.00035	0.00271	0	1777	0	0	0	0	1	-360	360;
	4501	2513	0.00146	0.01134	0	1547	0	0	0	0	1	-360	360;
	8938	1177	0	0.0002	0	0	0	0	0	0	1	-360	360;
	8938	1177	0.001	0.250041	0	0	0	0	0	0	1	-360	360;
	8938	5142	0.00031	0.00232	0	438	0	0	0	0	1	-360	360;
	3332	6130	0	0.0002	0	0	0	0	0	0	1	-360	360;
	3332	6130	0.001	0.319463	0	0	0	0	0	0	1	-360	360;
	3332	6440	0.002479	0.011479	0	438	0	0	0	0	1	-360	360;
	3332	1258	1.778616	5.164256	0	0	0	0	0	0	1	-360	360;
	2196	3332	0.00644	0.026229	0	419	0	0	0	0	1	-360	360;
	4085	3332	0.342769	1.509876	0	0	0	0	0	0	1	-360	360;
	6130	6440	0.002521	0.01124	0	438	0	0	0	0	1	-360	360;
	6130	1258	1.29219	4.678099	0	0	0	0	0	0	1	-360	360;
	4085	6130	0.236529	1.420661	0	0	0	0	0	0	1	-360	360;
	8022	260	0	0.0002	0	0	0	0	0	0	1	-360	360;
	8022	260	0.05505	0.67124	0	0	0	0	0	0	1	-360	360;
	242	8022	4.467975	11.21446	0	0	0	0	0	0	1	-360	360;
	8022	5596	34.72521	52.45248	0	0	0	0	0	0	1	-360	360;
	8022	4099	36.34711	50.63223	0	0	0	0	0	0	1	-360	360;
	8022	629	22.76033	33.1281	0	0	0	0	0	0	1	-360	360;
	8022	3423	21.48554	36.70661	0	0	0	0	0	0	1	-360	360;
	8022	2955	39.50826	49.23967	0	0	0	0	0	0	1	-360	360;
	4702	8022	17.04917	42.27893	0	0	0	0	0	0	1	-360	360;
	7500	8022	16.04855	40.82025	0	0	0	0	0	0	1	-360	360;
	8022	1067	0.053771	1.208946	0	0	0	0	0	0	1	-360	360;
	8022	1067	0.003281	0.016341	0	514	0	0	0	0	1	-360	360;
	3533	8022	0.825227	5.014463	0	0	0	0	0	0	1	-360	360;
	8022	4789	4.278926	11.33285	0	0	0	0	0	0	1	-360	360;
	8022	7592	3.46157	11.98905	0	0	0	0	0	0	1	-360	360;
	5440	8022	4.532025	17.77417	0	0	0	0	0	0	1	-360	360;
	5440	8022	0.009169	0.03831	0	514	0	0	0	0	1	-360	360;
	2605	8022	4.582645	17.99917	0	0	0	0	0	0	1	-360	360;
	8022	6421	4.507231	15.28285	0	0	0	0	0	0	1	-360	360;
	8022	1581	0.001	1.211157	0	0	0	0	0	0	1	-360	360;
	8022	6655	7.098967	16.7219	0	0	0	0	0	0	1	-360	360;
	8022	6307	6.249587	14.57107	0	0	0	0	0	0	1	-360	360;
	242	260	3.945041	12.27665	0	0	0	0	0	0	1	-360	360;
	260	5596	34.19215	51.58471	0	0	0	0	0	0	1	-360	360;
	260	4099	31.52686	50.73347	0	0	0	0	0	0	1	-360	360;
	260	629	18.00289	35.73554	0	0	0	0	0	0	1	-360	360;
	260	3423	21.65083	34.30165	0	0	0	0	0	0	1	-360	360;
	260	2955	35.85331	52.29959	0	0	0	0	0	0	1	-360	360;
	4702	260	18.05847	51.41322	0	0	0	0	0	0	1	-360	360;
	7500	260	21.28512	43.97521	0	0	0	0	0	0	1	-360	360;
	260	1067	0.036771	0.970744	0	0	0	0	0	0	1	-360	360;
	3533	260	0.78843	6.091736	0	0	0	0	0	0	1	-360	360;
	260	4789	3.228512	12.74483	0	0	0	0	0	0	1	-360	360;
	260	7592	3.551653	11.03347	0	0	0	0	0	0	1	-360	360;
	5440	260	5.941736	18.63595	0	0	0	0	0	0	1	-360	360;
	2605	260	4.915083	20.96901	0	0	0	0	0	0	1	-360	360;
	2605	260	0.01039	0.037539	0	495	0	0	0	0	1	-360	360;
	260	6421	4.096281	15.0905	0	0	0	0	0	0	1	-360	360;
	260	1581	0.001	1.352211	0	0	0	0	0	0	1	-360	360;
	260	6655	6.635537	16.54979	0	0	0	0	0	0	1	-360	360;
	260	6307	5.82438	14.40847	0	0	0	0	0	0	1	-360	360;
	4052	2133	0.00046	0.00466	0	0	0	0	0	0	1	-360	360;
	5862	324	0.03625	0.46376	0	757	0	0	0	0	1	-360	360;
	5862	324	0.00045	0.00475	0	0	0	0	0	0	1	-360	360;
	4052	324	0.04062	0.4091	0	757	0	0	0	0	1	-360	360;
	8022	1581	0.006271	0.0373	0	0	0	0	0	0	1	-360	360;
	648	5531	0.000331	0.00331	0	0	0	0	0	0	1	-360	360;
	648	4350	0.000341	0.00324	0	0	0	0	0	0	1	-360	360;
	648	1484	0.00013	0.00132	0	381	0	0	0	0	1	-360	360;
	5909	2828	0.001	0.1802	0	0	0	0	0	0	1	-360	360;
	5909	5596	0.04812	0.726715	0	0	0	0	0	0	1	-360	360;
	5909	5596	0.005169	0.036401	0	495	0	0	0	0	1	-360	360;
	5909	4099	0.04006	0.78624	0	0	0	0	0	0	1	-360	360;
	629	5909	0.559318	2.376033	0	0	0	0	0	0	1	-360	360;
	3423	5909	0.574607	2.433264	0	0	0	0	0	0	1	-360	360;
	5909	2955	0.449463	2.310744	0	0	0	0	0	0	1	-360	360;
	8182	5909	0.18062	1.392789	0	0	0	0	0	0	1	-360	360;
	6718	5909	0.092211	0.750041	0	0	0	0	0	0	1	-360	360;
	4702	5909	6.405372	12.55372	0	0	0	0	0	0	1	-360	360;
	7500	5909	5.734504	14.28161	0	0	0	0	0	0	1	-360	360;
	3533	5909	15.87066	21.82231	0	0	0	0	0	0	1	-360	360;
	5440	5909	45.78719	60.88223	0	0	0	0	0	0	1	-360	360;
	2605	5909	52.13017	54.55165	0	0	0	0	0	0	1	-360	360;
	1971	5909	0.615	2.63781	0	0	0	0	0	0	1	-360	360;
	5909	1581	16.26798	23.5124	0	0	0	0	0	0	1	-360	360;
	5909	6655	0.818636	3.652686	0	0	0	0	0	0	1	-360	360;
	5909	6307	0.670434	2.968182	0	0	0	0	0	0	1	-360	360;
	2828	5596	0.041711	0.912376	0	0	0	0	0	0	1	-360	360;
	2828	4099	0.04611	0.822479	0	0	0	0	0	0	1	-360	360;
	2828	4099	0.005271	0.03561	0	533	0	0	0	0	1	-360	360;
	629	2828	0.605434	2.534091	0	0	0	0	0	0	1	-360	360;
	3423	2828	0.542851	2.99876	0	0	0	0	0	0	1	-360	360;
	2828	2955	0.520909	2.569628	0	0	0	0	0	0	1	-360	360;
	8182	2828	0.177789	1.458264	0	0	0	0	0	0	1	-360	360;
	6718	2828	0.09881	0.801281	0	0	0	0	0	0	1	-360	360;
	4702	2828	6.124793	15.47397	0	0	0	0	0	0	1	-360	360;
	7500	2828	5.91157	16.17149	0	0	0	0	0	0	1	-360	360;
	3533	2828	14.54876	25.34091	0	0	0	0	0	0	1	-360	360;
	5440	2828	54.67355	60.11777	0	0	0	0	0	0	1	-360	360;
	1971	2828	0.802355	2.58657	0	0	0	0	0	0	1	-360	360;
	2828	1581	19.94897	25.18388	0	0	0	0	0	0	1	-360	360;
	2828	6655	1.027479	3.393802	0	0	0	0	0	0	1	-360	360;
	2828	6307	0.843471	3.042975	0	0	0	0	0	0	1	-360	360;
	6724	7492	0.00078	0.00782	0	1909	0	0	0	0	1	-360	360;
	6724	8429	7.575485	11.77562	0	757	0	0	0	0	1	-360	360;
	6724	7816	1.23705	3.353947	0	757	0	0	0	0	1	-360	360;
	6724	5428	3.429501	6.941136	0	757	0	0	0	0	1	-360	360;
	6724	1525	0.00076	0.00797	0	1810	0	0	0	0	1	-360	360;
	8256	1525	0.00025	0.00233	0	1942	0	0	0	0	1	-360	360;
	2309	7492	0.00179	0.01397	0	1613	0	0	0	0	1	-360	360;
	9211	63	0.173643	1.472438	0	757	0	0	0	0	1	-360	360;
	63	8336	5.182133	18.94529	0	757	0	0	0	0	1	-360	360;
	63	2724	0.188989	1.114889	0	757	0	0	0	0	1	-360	360;
	1268	63	1.860596	8.193906	0	757	0	0	0	0	1	-360	360;
	242	316	0.00424	0.029211	0	572	0	0	0	0	1	-360	360;
	242	1067	6.176859	17.47149	0	0	0	0	0	0	1	-360	360;
	3533	242	10.52665	27.33264	0	0	0	0	0	0	1	-360	360;
	242	4789	0.202	1.377087	0	0	0	0	0	0	1	-360	360;
	242	7592	0.19876	1.343988	0	0	0	0	0	0	1	-360	360;
	242	7592	0.00545	0.042341	0	553	0	0	0	0	1	-360	360;
	5440	242	7.300207	16.26798	0	0	0	0	0	0	1	-360	360;
	2605	242	6.816736	17.84773	0	0	0	0	0	0	1	-360	360;
	242	6421	3.997727	8.729958	0	0	0	0	0	0	1	-360	360;
	242	1581	15.84711	34.44008	0	0	0	0	0	0	1	-360	360;
	242	1085	30.61777	53.27893	0	0	0	0	0	0	1	-360	360;
	242	3913	36.28512	62.88017	0	0	0	0	0	0	1	-360	360;
	63	2724	0.00082	0.01007	0	1580	0	0	0	0	1	-360	360;
	7454	7706	0.0005	0.004169	0	572	0	0	0	0	1	-360	360;
	5440	7454	0.00525	0.033151	0	495	0	0	0	0	1	-360	360;
	7706	4646	0.001969	0.01718	0	495	0	0	0	0	1	-360	360;
	4646	4288	0.001289	0.009229	0	553	0	0	0	0	1	-360	360;
	4288	3935	0.00043	0.00349	0	553	0	0	0	0	1	-360	360;
	3935	6716	0.000979	0.008901	0	495	0	0	0	0	1	-360	360;
	6716	316	0.000219	0.00155	0	514	0	0	0	0	1	-360	360;
	7039	7135	0.00025	0.00319	0	889	0	0	0	0	1	-360	360;
	7039	7135	0.00024	0.00295	0	856	0	0	0	0	1	-360	360;
	8429	732	0.00022	0.00171	0	1448	0	0	0	0	1	-360	360;
	4294	732	0.0005	0.00382	0	1514	0	0	0	0	1	-360	360;
	6369	2545	0.0002	0.00192	0	1251	0	0	0	0	1	-360	360;
	4294	2545	0.0005	0.00383	0	1580	0	0	0	0	1	-360	360;
	5596	4099	0	0.0002	0	0	0	0	0	0	1	-360	360;
	5596	4099	0.001	0.268678	0	0	0	0	0	0	1	-360	360;
	629	5596	0.042159	0.956198	0	0	0	0	0	0	1	-360	360;
	629	5596	0.002341	0.022341	0	0	0	0	0	0	1	-360	360;
	3423	5596	0.04555	0.93969	0	0	0	0	0	0	1	-360	360;
	5596	2955	0.077909	1.08686	0	0	0	0	0	0	1	-360	360;
	4424	5596	0.0003	0.003289	0	629	0	0	0	0	1	-360	360;
	8182	5596	1.527831	6.098554	0	0	0	0	0	0	1	-360	360;
	6718	5596	0.958926	2.68595	0	0	0	0	0	0	1	-360	360;
	4702	5596	28.84298	48.1157	0	0	0	0	0	0	1	-360	360;
	7500	5596	33.55372	46.46901	0	0	0	0	0	0	1	-360	360;
	1067	5596	47.90702	63.74587	0	0	0	0	0	0	1	-360	360;
	3533	5596	28.68802	30.77066	0	0	0	0	0	0	1	-360	360;
	1971	5596	4.526653	11.10372	0	0	0	0	0	0	1	-360	360;
	5596	1581	6.754959	11.13905	0	0	0	0	0	0	1	-360	360;
	5596	6655	0.105229	1.236012	0	0	0	0	0	0	1	-360	360;
	5596	6307	0.085781	1.020186	0	0	0	0	0	0	1	-360	360;
	629	4099	0.044289	0.919483	0	0	0	0	0	0	1	-360	360;
	3423	4099	0.047781	0.901963	0	0	0	0	0	0	1	-360	360;
	4099	2955	0.076341	1.006446	0	0	0	0	0	0	1	-360	360;
	7189	4099	0.0003	0.00336	0	610	0	0	0	0	1	-360	360;
	8182	4099	1.621322	5.733678	0	0	0	0	0	0	1	-360	360;
	6718	4099	0.903202	2.852273	0	0	0	0	0	0	1	-360	360;
	4702	4099	30.65289	49.86157	0	0	0	0	0	0	1	-360	360;
	7500	4099	32.78719	47.26653	0	0	0	0	0	0	1	-360	360;
	1067	4099	44.17355	62.70041	0	0	0	0	0	0	1	-360	360;
	3533	4099	25.07438	35.58471	0	0	0	0	0	0	1	-360	360;
	1971	4099	5.060744	9.753098	0	0	0	0	0	0	1	-360	360;
	4099	1581	6.015083	12.60103	0	0	0	0	0	0	1	-360	360;
	4099	6655	0.099409	1.193017	0	0	0	0	0	0	1	-360	360;
	4099	6655	0.002021	0.01812	0	0	0	0	0	0	1	-360	360;
	4099	6307	0.09157	0.961405	0	0	0	0	0	0	1	-360	360;
	4647	5774	0.00032	0.00387	0	1349	0	0	0	0	1	-360	360;
	4647	3583	0.00038	0.00378	0	1448	0	0	0	0	1	-360	360;
	3578	6753	0.00032	0.00394	0	1481	0	0	0	0	1	-360	360;
	3578	606	0.00038	0.0037	0	1382	0	0	0	0	1	-360	360;
	4647	7483	6.036911	10.78324	0	757	0	0	0	0	1	-360	360;
	6485	7135	0.0008	0.00945	0	1481	0	0	0	0	1	-360	360;
	7135	624	0.00079	0.00828	0	1349	0	0	0	0	1	-360	360;
	6485	7135	0.00078	0.00964	0	1415	0	0	0	0	1	-360	360;
	7135	624	0.00081	0.00811	0	1316	0	0	0	0	1	-360	360;
	629	3423	0	0.0002	0	419	0	0	0	0	1	-360	360;
	629	3423	0.001	0.26281	0	0	0	0	0	0	1	-360	360;
	629	2955	0.001	0.461798	0	0	0	0	0	0	1	-360	360;
	629	2955	0.00164	0.020729	0	533	0	0	0	0	1	-360	360;
	8182	629	7.774793	17.76901	0	0	0	0	0	0	1	-360	360;
	6718	629	3.756405	9.209917	0	0	0	0	0	0	1	-360	360;
	629	1067	27.82645	45.00826	0	0	0	0	0	0	1	-360	360;
	3533	629	20.96901	31.09711	0	0	0	0	0	0	1	-360	360;
	1971	629	18.12293	35.08471	0	0	0	0	0	0	1	-360	360;
	629	1581	3.360124	8.657025	0	0	0	0	0	0	1	-360	360;
	629	6655	0.14569	1.294938	0	0	0	0	0	0	1	-360	360;
	629	6307	0.125919	1.049607	0	0	0	0	0	0	1	-360	360;
	3423	2955	0.001	0.471508	0	0	0	0	0	0	1	-360	360;
	3423	2955	0.001599	0.02112	0	0	0	0	0	0	1	-360	360;
	8182	3423	8.624174	16.76653	0	0	0	0	0	0	1	-360	360;
	6718	3423	4.365909	9.241321	0	0	0	0	0	0	1	-360	360;
	3423	1067	29.57231	44.15909	0	0	0	0	0	0	1	-360	360;
	3533	3423	22.67562	29.81818	0	0	0	0	0	0	1	-360	360;
	1971	3423	22.21281	29.92975	0	0	0	0	0	0	1	-360	360;
	3423	1581	3.288017	9.179959	0	0	0	0	0	0	1	-360	360;
	3423	6655	0.15644	1.276653	0	0	0	0	0	0	1	-360	360;
	3423	6307	0.124669	1.119298	0	0	0	0	0	0	1	-360	360;
	8182	2955	6.748554	16.31901	0	0	0	0	0	0	1	-360	360;
	6718	2955	3.853099	7.976446	0	0	0	0	0	0	1	-360	360;
	1067	2955	53.92562	59.04132	0	0	0	0	0	0	1	-360	360;
	3533	2955	37.32851	39.04752	0	0	0	0	0	0	1	-360	360;
	1971	2955	16.76198	30.60331	0	0	0	0	0	0	1	-360	360;
	2955	1581	6.070868	12.11942	0	0	0	0	0	0	1	-360	360;
	6655	2955	0.29624	1.60312	0	0	0	0	0	0	1	-360	360;
	6307	2955	0.293884	1.293946	0	0	0	0	0	0	1	-360	360;
	9211	8336	0.15633	3.578947	0	757	0	0	0	0	1	-360	360;
	9211	2724	0.242957	1.518213	0	757	0	0	0	0	1	-360	360;
	9211	1268	8.069251	24.53047	0	757	0	0	0	0	1	-360	360;
	9211	3522	0.00029	0.00357	0	0	0	0	0	0	1	-360	360;
	2440	5380	0.00012	0.00131	0	1777	0	0	0	0	1	-360	360;
	8182	6718	0	0.0002	0	0	0	0	0	0	1	-360	360;
	8182	6718	0.001	0.168909	0	0	0	0	0	0	1	-360	360;
	8182	4702	0.837231	3.56405	0	0	0	0	0	0	1	-360	360;
	8182	7500	0.690682	3.987397	0	0	0	0	0	0	1	-360	360;
	8182	7500	0.00988	0.05469	0	572	0	0	0	0	1	-360	360;
	3533	8182	5.606818	11.56529	0	0	0	0	0	0	1	-360	360;
	5440	8182	15.6624	27.83058	0	0	0	0	0	0	1	-360	360;
	2605	8182	16.76116	26.42355	0	0	0	0	0	0	1	-360	360;
	8182	1971	0.039521	0.57	0	0	0	0	0	0	1	-360	360;
	8182	1581	25.64463	32.3843	0	0	0	0	0	0	1	-360	360;
	8182	6655	12.49401	21.31612	0	0	0	0	0	0	1	-360	360;
	8182	6307	10.90826	18.38058	0	0	0	0	0	0	1	-360	360;
	6718	4702	0.367355	2.263636	0	0	0	0	0	0	1	-360	360;
	6718	7500	0.394566	2.152479	0	0	0	0	0	0	1	-360	360;
	3533	6718	2.95062	6.117354	0	0	0	0	0	0	1	-360	360;
	5440	6718	8.41281	14.42025	0	0	0	0	0	0	1	-360	360;
	2605	6718	8.672314	14.29339	0	0	0	0	0	0	1	-360	360;
	6718	1971	0.0192	0.321384	0	0	0	0	0	0	1	-360	360;
	6718	1581	12.78037	18.27831	0	0	0	0	0	0	1	-360	360;
	6718	6655	6.830371	10.80145	0	0	0	0	0	0	1	-360	360;
	6718	6307	5.092149	10.9593	0	0	0	0	0	0	1	-360	360;
	4849	7082	0.00043	0.00522	0	1218	0	0	0	0	1	-360	360;
	1268	7082	0.00184	0.01667	0	1580	0	0	0	0	1	-360	360;
	1268	7082	0.248089	2.761427	0	757	0	0	0	0	1	-360	360;
	2075	4884	0.00037	0.00464	0	1481	0	0	0	0	1	-360	360;
	2075	5768	0.00059	0.00694	0	1547	0	0	0	0	1	-360	360;
	9100	8693	0.00027	0.00348	0	1481	0	0	0	0	1	-360	360;
	5717	8693	0.00027	0.00354	0	1547	0	0	0	0	1	-360	360;
	9100	5717	0.02114	0.839889	0	757	0	0	0	0	1	-360	360;
	6222	9100	30.32756	43.46053	0	757	0	0	0	0	1	-360	360;
	6222	5717	30.89958	42.53601	0	757	0	0	0	0	1	-360	360;
	4702	7500	0	0.0002	0	0	0	0	0	0	1	-360	360;
	4702	7500	0.001	0.238595	0	0	0	0	0	0	1	-360	360;
	3533	4702	0.256632	1.964752	0	0	0	0	0	0	1	-360	360;
	5440	4702	0.63343	3.574793	0	0	0	0	0	0	1	-360	360;
	2605	4702	0.709752	3.284711	0	0	0	0	0	0	1	-360	360;
	2605	4702	0.01324	0.090909	0	495	0	0	0	0	1	-360	360;
	1971	4702	0.000591	1.201426	0	0	0	0	0	0	1	-360	360;
	4702	1581	1.821963	7.541736	0	0	0	0	0	0	1	-360	360;
	3533	7500	0.210826	2.191942	0	0	0	0	0	0	1	-360	360;
	5440	7500	0.746074	3.056198	0	0	0	0	0	0	1	-360	360;
	2605	7500	0.740041	3.16219	0	0	0	0	0	0	1	-360	360;
	1971	7500	0.000841	1.29095	0	0	0	0	0	0	1	-360	360;
	7500	1581	2.154545	7.118182	0	0	0	0	0	0	1	-360	360;
	4884	1950	8e-05	0.00105	0	1481	0	0	0	0	1	-360	360;
	8244	8336	0.00025	0.00186	0	0	0	0	0	0	1	-360	360;
	1067	5598	9.1e-05	0.000581	0	572	0	0	0	0	1	-360	360;
	7910	8336	0.00023	0.00211	0	1876	0	0	0	0	1	-360	360;
	4742	8027	8e-05	0.00107	0	1613	0	0	0	0	1	-360	360;
	4742	8336	2e-05	0.0002	0	1415	0	0	0	0	1	-360	360;
	7717	1739	0.00063	0.00406	0	553	0	0	0	0	1	-360	360;
	7717	1067	9.1e-05	0.00056	0	533	0	0	0	0	1	-360	360;
	8866	2066	0.00064	0.003969	0	553	0	0	0	0	1	-360	360;
	1067	8866	9.1e-05	0.00055	0	476	0	0	0	0	1	-360	360;
	8336	1950	2e-05	0.0002	0	1580	0	0	0	0	1	-360	360;
	8027	4860	0.00091	0.01097	0	1448	0	0	0	0	1	-360	360;
	1739	2516	0.00711	0.045841	0	514	0	0	0	0	1	-360	360;
	2066	8696	0.00724	0.044849	0	553	0	0	0	0	1	-360	360;
	2724	8336	5.952838	17.29294	0	757	0	0	0	0	1	-360	360;
	1268	8336	20.08726	50.82479	0	757	0	0	0	0	1	-360	360;
	3533	1067	4.890702	15.41012	0	0	0	0	0	0	1	-360	360;
	4789	1067	5.639256	15.40868	0	0	0	0	0	0	1	-360	360;
	7592	1067	4.76281	15.7	0	0	0	0	0	0	1	-360	360;
	5440	1067	21.2624	53.66529	0	0	0	0	0	0	1	-360	360;
	2605	1067	25.16942	46.14463	0	0	0	0	0	0	1	-360	360;
	1067	6421	5.023554	17.83409	0	0	0	0	0	0	1	-360	360;
	1067	1581	0.483492	3.46095	0	0	0	0	0	0	1	-360	360;
	112	1067	0.000211	0.0012	0	419	0	0	0	0	1	-360	360;
	1067	6655	9.888017	20.17066	0	0	0	0	0	0	1	-360	360;
	1067	6307	9.581818	15.8126	0	0	0	0	0	0	1	-360	360;
	1493	1058	0.00015	0.00151	0	1580	0	0	0	0	1	-360	360;
	4860	1893	0.00044	0.00443	0	1547	0	0	0	0	1	-360	360;
	2516	6451	0.00301	0.01868	0	495	0	0	0	0	1	-360	360;
	8696	6451	0.00268	0.02118	0	476	0	0	0	0	1	-360	360;
	5768	1058	0.00044	0.00433	0	1481	0	0	0	0	1	-360	360;
	1893	1493	0.00015	0.00148	0	1448	0	0	0	0	1	-360	360;
	3533	4789	36.67975	64.23967	0	0	0	0	0	0	1	-360	360;
	3533	7592	34.68595	57.82851	0	0	0	0	0	0	1	-360	360;
	5440	3533	0.001	0.997107	0	0	0	0	0	0	1	-360	360;
	2605	3533	0.001	0.990558	0	0	0	0	0	0	1	-360	360;
	2605	3533	0.00389	0.03206	0	476	0	0	0	0	1	-360	360;
	3533	1971	2.394421	10.2281	0	0	0	0	0	0	1	-360	360;
	3533	1581	0.001	0.942996	0	0	0	0	0	0	1	-360	360;
	3533	6655	7.162397	16.07149	0	0	0	0	0	0	1	-360	360;
	3533	6307	6.970248	12.65785	0	0	0	0	0	0	1	-360	360;
	4535	6180	0.00041	0.00489	0	0	0	0	0	0	1	-360	360;
	8779	6180	0.00035	0.00415	0	0	0	0	0	0	1	-360	360;
	2724	4389	0.00041	0.00478	0	0	0	0	0	0	1	-360	360;
	8779	4389	0.00036	0.00406	0	1711	0	0	0	0	1	-360	360;
	4535	7443	0.00069	0.00736	0	1810	0	0	0	0	1	-360	360;
	3522	4535	0.00035	0.00355	0	0	0	0	0	0	1	-360	360;
	2444	4535	0.0003	0.0037	0	0	0	0	0	0	1	-360	360;
	2724	5650	0.00079	0.00639	0	0	0	0	0	0	1	-360	360;
	1268	2724	10.65028	23.35249	0	757	0	0	0	0	1	-360	360;
	4921	2724	0.00035	0.00375	0	0	0	0	0	0	1	-360	360;
	630	2724	0.00035	0.00368	0	0	0	0	0	0	1	-360	360;
	4789	7592	0	0.0002	0	476	0	0	0	0	1	-360	360;
	4789	7592	0.001	0.307397	0	0	0	0	0	0	1	-360	360;
	5440	4789	30.07851	44.4938	0	0	0	0	0	0	1	-360	360;
	2605	4789	28.13843	48.89876	0	0	0	0	0	0	1	-360	360;
	4789	6421	1.021095	4.339256	0	0	0	0	0	0	1	-360	360;
	4789	1581	19.04773	44.46694	0	0	0	0	0	0	1	-360	360;
	4789	1085	11.71198	25.83058	0	0	0	0	0	0	1	-360	360;
	4789	3913	13.4062	28.03306	0	0	0	0	0	0	1	-360	360;
	5440	7592	25.77686	44.3843	0	0	0	0	0	0	1	-360	360;
	2605	7592	29.27686	39.66322	0	0	0	0	0	0	1	-360	360;
	7592	6421	0.90343	4.251446	0	0	0	0	0	0	1	-360	360;
	7592	1581	15.67913	46.04752	0	0	0	0	0	0	1	-360	360;
	7592	1085	10.06777	25.78512	0	0	0	0	0	0	1	-360	360;
	7592	3913	11.7936	27.51653	0	0	0	0	0	0	1	-360	360;
	4732	5650	0.00037	0.00319	0	1744	0	0	0	0	1	-360	360;
	2932	7443	0.00037	0.0032	0	1678	0	0	0	0	1	-360	360;
	3267	9084	6e-05	0.00057	0	0	0	0	0	0	1	-360	360;
	5428	3267	0.00081	0.00793	0	0	0	0	0	0	1	-360	360;
	5440	2605	0	0.0002	0	533	0	0	0	0	1	-360	360;
	5440	2605	0.001	0.218595	0	0	0	0	0	0	1	-360	360;
	5440	1971	5.508884	16.92169	0	0	0	0	0	0	1	-360	360;
	5440	1581	0.417335	3.096901	0	0	0	0	0	0	1	-360	360;
	5440	6655	35.56405	45.20868	0	0	0	0	0	0	1	-360	360;
	5440	6307	26.29545	47.22107	0	0	0	0	0	0	1	-360	360;
	2605	1971	5.134091	18.53492	0	0	0	0	0	0	1	-360	360;
	2605	1581	0.373512	3.536364	0	0	0	0	0	0	1	-360	360;
	2605	6655	31.99174	51.7624	0	0	0	0	0	0	1	-360	360;
	2605	6307	26.61157	47.82645	0	0	0	0	0	0	1	-360	360;
	6421	1581	22.02479	61.94007	0	0	0	0	0	0	1	-360	360;
	1085	6421	0.20395	1.135248	0	0	0	0	0	0	1	-360	360;
	3913	6421	0.207314	1.339256	0	0	0	0	0	0	1	-360	360;
	1971	1581	15.82955	28.76446	0	0	0	0	0	0	1	-360	360;
	1971	6655	27.39876	40.44215	0	0	0	0	0	0	1	-360	360;
	1971	6307	23.59711	35.69835	0	0	0	0	0	0	1	-360	360;
	6655	1581	1.210248	3.369008	0	0	0	0	0	0	1	-360	360;
	6307	1581	0.914421	3.375826	0	0	0	0	0	0	1	-360	360;
	3780	1258	0.005729	0.02032	0	400	0	0	0	0	1	-360	360;
	1623	1258	0.005841	0.019909	0	438	0	0	0	0	1	-360	360;
	4085	1258	0.084781	0.493471	0	0	0	0	0	0	1	-360	360;
	5862	4052	0	0.0002	0	1349	0	0	0	0	1	-360	360;
	5862	4052	0.001	0.19356	0	757	0	0	0	0	1	-360	360;
	5862	8193	0.00192	0.01356	0	0	0	0	0	0	1	-360	360;
	4052	8193	0.00172	0.01537	0	0	0	0	0	0	1	-360	360;
	1918	4206	0.00035	0.00346	0	1909	0	0	0	0	1	-360	360;
	1085	3913	0	0.0002	0	533	0	0	0	0	1	-360	360;
	1085	3913	0.001	0.159719	0	0	0	0	0	0	1	-360	360;
	1085	6049	0.000341	0.00199	0	591	0	0	0	0	1	-360	360;
	6049	7524	0.00219	0.012901	0	0	0	0	0	0	1	-360	360;
	7524	4411	0.002531	0.013539	0	0	0	0	0	0	1	-360	360;
	4411	7218	0.002281	0.01263	0	0	0	0	0	0	1	-360	360;
	7218	112	0.00232	0.01393	0	0	0	0	0	0	1	-360	360;
	6369	2038	0.00043	0.00418	0	1810	0	0	0	0	1	-360	360;
	7816	8429	0.113102	0.936981	0	757	0	0	0	0	1	-360	360;
	8429	1165	0.00043	0.00419	0	0	0	0	0	0	1	-360	360;
	5428	8429	0.124889	1.006302	0	757	0	0	0	0	1	-360	360;
	2608	9077	0	0.0002	0	0	0	0	0	0	1	-360	360;
	2608	9077	0.003651	0.01288	0	0	0	0	0	0	1	-360	360;
	9077	7321	0.004021	0.01951	0	0	0	0	0	0	1	-360	360;
	7816	1150	0.00057	0.00402	0	0	0	0	0	0	1	-360	360;
	4041	5428	0.00064	0.00448	0	0	0	0	0	0	1	-360	360;
	7816	5428	0.00646	0.473407	0	757	0	0	0	0	1	-360	360;
	3264	2038	8e-05	0.00061	0	889	0	0	0	0	1	-360	360;
	779	1165	6e-05	0.00065	0	1086	0	0	0	0	1	-360	360;
	4085	3780	0.00012	0.000461	0	400	0	0	0	0	1	-360	360;
	4085	1623	0.00013	0.00045	0	0	0	0	0	0	1	-360	360;
	6655	6307	0	0.0002	0	0	0	0	0	0	1	-360	360;
	6655	6307	0.001	0.249814	0	0	0	0	0	0	1	-360	360;
	8593	7650	0.0001	0.00075	0	1218	0	0	0	0	1	-360	360;
	8593	7490	0.0038	0.11597	0	757	0	0	0	0	1	-360	360;
	8593	7490	0.0001	0.00081	0	1283	0	0	0	0	1	-360	360;
	6811	7650	0.0001	0.00094	0	1283	0	0	0	0	1	-360	360;
	6811	7650	9e-05	0.00107	0	1152	0	0	0	0	1	-360	360;
	61	6811	7e-05	0.00085	0	1119	0	0	0	0	1	-360	360;
	61	6811	8e-05	0.00081	0	1283	0	0	0	0	1	-360	360;
	7650	7490	0.00018	0.00171	0	1086	0	0	0	0	1	-360	360;
	61	2532	0.00017	0.00191	0	1415	0	0	0	0	1	-360	360;
	61	2532	0.00016	0.00194	0	1349	0	0	0	0	1	-360	360;
	61	7934	0.0005	0.00317	0	1514	0	0	0	0	1	-360	360;
	61	7934	0.00048	0.00322	0	1547	0	0	0	0	1	-360	360;
	7490	4349	0.00427	0.19732	0	757	0	0	0	0	1	-360	360;
	7490	4349	0.0001	0.00122	0	1152	0	0	0	0	1	-360	360;
	7490	4349	0.0001	0.00119	0	1251	0	0	0	0	1	-360	360;
	4349	4498	0.0024	0.02048	0	1547	0	0	0	0	1	-360	360;
	4349	197	0.00042	0.00474	0	1876	0	0	0	0	1	-360	360;
	3606	2094	0	0.0002	0	0	0	0	0	0	1	-360	360;
	3606	2094	0.003409	0.129831	0	0	0	0	0	0	1	-360	360;
	3606	7741	0.00605	0.024031	0	457	0	0	0	0	1	-360	360;
	2094	7741	0.005919	0.0245	0	419	0	0	0	0	1	-360	360;
	2883	6197	0.406364	1.478988	0	0	0	0	0	0	1	-360	360;
	2883	397	0.12912	0.95312	0	0	0	0	0	0	1	-360	360;
	2883	4971	0.23062	0.90281	0	0	0	0	0	0	1	-360	360;
	1120	2883	0.05107	0.366901	0	0	0	0	0	0	1	-360	360;
	2883	397	0.00207	0.0168	0	0	0	0	0	0	1	-360	360;
	2883	1627	0.00119	0.009659	0	514	0	0	0	0	1	-360	360;
	3537	2883	0.120151	0.700041	0	0	0	0	0	0	1	-360	360;
	2883	5869	0.032581	0.326756	0	0	0	0	0	0	1	-360	360;
	3716	5543	2e-05	0.0002	0	757	0	0	0	0	1	-360	360;
	3716	959	0.00025	0.00344	0	757	0	0	0	0	1	-360	360;
	3716	959	0.01671	0.288767	0	757	0	0	0	0	1	-360	360;
	3716	2583	0.01954	0.281191	0	757	0	0	0	0	1	-360	360;
	6274	3716	0.00356	0.222701	0	757	0	0	0	0	1	-360	360;
	6274	3716	0.00271	0.199571	0	757	0	0	0	0	1	-360	360;
	4637	5543	0.00016	0.00253	0	921	0	0	0	0	1	-360	360;
	6274	5543	0.00018	0.00214	0	1251	0	0	0	0	1	-360	360;
	6197	397	0.002091	0.0138	0	0	0	0	0	0	1	-360	360;
	6197	397	0.291467	1.202521	0	0	0	0	0	0	1	-360	360;
	4971	6197	0.204719	0.794277	0	0	0	0	0	0	1	-360	360;
	1120	6197	0.026401	0.214814	0	0	0	0	0	0	1	-360	360;
	1120	6197	0.00295	0.01256	0	438	0	0	0	0	1	-360	360;
	5325	7815	0.359525	1.423657	0	0	0	0	0	0	1	-360	360;
	4167	7815	0.32907	1.339607	0	0	0	0	0	0	1	-360	360;
	4167	7815	0.00469	0.03913	0	0	0	0	0	0	1	-360	360;
	2663	7815	0.000841	0.00544	0	0	0	0	0	0	1	-360	360;
	2663	7815	0.014581	0.16557	0	0	0	0	0	0	1	-360	360;
	4804	7815	1.038884	1.650227	0	0	0	0	0	0	1	-360	360;
	7602	7815	0.863926	1.96155	0	0	0	0	0	0	1	-360	360;
	56	397	0.005081	0.02426	0	419	0	0	0	0	1	-360	360;
	1624	56	0.470062	1.363636	0	0	0	0	0	0	1	-360	360;
	56	7127	0.035969	0.35219	0	0	0	0	0	0	1	-360	360;
	56	9165	0.0033	0.01864	0	476	0	0	0	0	1	-360	360;
	56	9165	0.018021	0.1825	0	0	0	0	0	0	1	-360	360;
	5088	6296	0.002651	0.027659	0	0	0	0	0	0	1	-360	360;
	5104	5088	0.00226	0.04061	0	514	0	0	0	0	1	-360	360;
	5088	8287	0.000711	0.0098	0	0	0	0	0	0	1	-360	360;
	2300	5088	0.002211	0.037599	0	0	0	0	0	0	1	-360	360;
	4113	2420	0.0016	0.01356	0	1316	0	0	0	0	1	-360	360;
	4113	2420	0.00165	0.01328	0	1283	0	0	0	0	1	-360	360;
	6274	959	0.01336	0.25865	0	757	0	0	0	0	1	-360	360;
	6274	959	0.00015	0.00253	0	757	0	0	0	0	1	-360	360;
	8855	1663	0.00225	0.02106	0	1613	0	0	0	0	1	-360	360;
	4226	1663	0.00426	0.03506	0	1678	0	0	0	0	1	-360	360;
	6777	1663	0.00043	0.00581	0	0	0	0	0	0	1	-360	360;
	4971	3944	0.001151	0.009909	0	419	0	0	0	0	1	-360	360;
	1120	3944	0.000961	0.00751	0	610	0	0	0	0	1	-360	360;
	853	3944	0.00032	0.00324	0	438	0	0	0	0	1	-360	360;
	8855	1663	0.00245	0.02191	0	1053	0	0	0	0	1	-360	360;
	1128	1663	0.00037	0.00606	0	921	0	0	0	0	1	-360	360;
	6977	1663	0.0029	0.0289	0	1744	0	0	0	0	1	-360	360;
	7055	397	0.000151	0.00156	0	0	0	0	0	0	1	-360	360;
	9086	397	0.00014	0.00144	0	0	0	0	0	0	1	-360	360;
	4971	397	0.026841	0.406942	0	0	0	0	0	0	1	-360	360;
	1120	397	0.025909	0.304649	0	0	0	0	0	0	1	-360	360;
	2630	827	0	0.0002	0	0	0	0	0	0	1	-360	360;
	2630	827	0.001	0.18678	0	757	0	0	0	0	1	-360	360;
	8454	2630	0.989958	2.675	0	757	0	0	0	0	1	-360	360;
	8292	2630	0.994529	2.704017	0	757	0	0	0	0	1	-360	360;
	4158	2630	0.154938	0.902078	0	757	0	0	0	0	1	-360	360;
	3319	2630	0.141247	1.002008	0	757	0	0	0	0	1	-360	360;
	9116	2630	0.01709	0.303539	0	757	0	0	0	0	1	-360	360;
	2630	7424	0.00081	0.00826	0	1547	0	0	0	0	1	-360	360;
	2630	4113	0.00174	0.01199	0	1481	0	0	0	0	1	-360	360;
	2630	6193	0.00136	0.012	0	1020	0	0	0	0	1	-360	360;
	8454	827	1.135249	2.306648	0	757	0	0	0	0	1	-360	360;
	8292	827	1.077147	2.483726	0	757	0	0	0	0	1	-360	360;
	4158	827	0.140319	0.988089	0	757	0	0	0	0	1	-360	360;
	3319	827	0.15108	0.940582	0	757	0	0	0	0	1	-360	360;
	9116	827	0.00045	0.00442	0	1382	0	0	0	0	1	-360	360;
	9116	827	0.01578	0.332528	0	757	0	0	0	0	1	-360	360;
	4718	827	0.00032	0.00308	0	0	0	0	0	0	1	-360	360;
	327	4637	0.00018	0.0025	0	1942	0	0	0	0	1	-360	360;
	2906	4637	0.03152	0.366053	0	757	0	0	0	0	1	-360	360;
	1624	3761	0.00251	0.13469	0	0	0	0	0	0	1	-360	360;
	1624	5325	0.00943	0.04787	0	419	0	0	0	0	1	-360	360;
	1624	7127	0.002781	0.02494	0	0	0	0	0	0	1	-360	360;
	1624	7127	0.22876	0.857376	0	0	0	0	0	0	1	-360	360;
	1624	9165	0.113521	0.454752	0	0	0	0	0	0	1	-360	360;
	8855	7617	0.00159	0.01603	0	1415	0	0	0	0	1	-360	360;
	3761	4167	0.00924	0.04881	0	495	0	0	0	0	1	-360	360;
	3761	9165	0.00643	0.027271	0	533	0	0	0	0	1	-360	360;
	8855	4226	0.00183	0.01368	0	1810	0	0	0	0	1	-360	360;
	8199	5360	0	0.0002	0	954	0	0	0	0	1	-360	360;
	8199	5360	0.001	0.191392	0	757	0	0	0	0	1	-360	360;
	3254	8199	0.00081	0.00981	0	1547	0	0	0	0	1	-360	360;
	8199	4946	0.090907	0.571413	0	757	0	0	0	0	1	-360	360;
	8199	4946	0.00058	0.00647	0	1547	0	0	0	0	1	-360	360;
	8199	8968	0.441572	1.595429	0	757	0	0	0	0	1	-360	360;
	8199	4840	0.0463	0.493158	0	757	0	0	0	0	1	-360	360;
	8199	5972	0.053	0.456427	0	757	0	0	0	0	1	-360	360;
	5972	5360	0.00048	0.00707	0	1514	0	0	0	0	1	-360	360;
	5360	4946	0.0007	0.00626	0	1777	0	0	0	0	1	-360	360;
	5360	4946	0.099273	0.491122	0	757	0	0	0	0	1	-360	360;
	5360	8968	0.411558	1.466205	0	757	0	0	0	0	1	-360	360;
	5360	4840	0.05226	0.428179	0	757	0	0	0	0	1	-360	360;
	5360	5972	0.04979	0.473137	0	757	0	0	0	0	1	-360	360;
	7170	1425	0.00018	0.00218	0	0	0	0	0	0	1	-360	360;
	3835	4366	0.344619	1.353116	0	757	0	0	0	0	1	-360	360;
	3835	2385	0.581787	1.764958	0	757	0	0	0	0	1	-360	360;
	3835	7920	0.075831	0.835526	0	757	0	0	0	0	1	-360	360;
	3835	7920	0.00095	0.0105	0	1481	0	0	0	0	1	-360	360;
	3835	2080	0.081281	0.782202	0	757	0	0	0	0	1	-360	360;
	3835	7424	0.06719	0.554377	0	757	0	0	0	0	1	-360	360;
	3835	6193	0.00409	0.271641	0	757	0	0	0	0	1	-360	360;
	3835	6193	0.00041	0.00351	0	1448	0	0	0	0	1	-360	360;
	7957	652	0.00138	0.008789	0	533	0	0	0	0	1	-360	360;
	7957	3303	0.002331	0.01832	0	0	0	0	0	0	1	-360	360;
	4366	2385	0	0.0002	0	1876	0	0	0	0	1	-360	360;
	4366	2385	0.001	0.311717	0	757	0	0	0	0	1	-360	360;
	1588	4366	0.00043	0.00408	0	1843	0	0	0	0	1	-360	360;
	4366	7920	0.083899	0.525478	0	757	0	0	0	0	1	-360	360;
	4366	2080	0.07124	0.605921	0	757	0	0	0	0	1	-360	360;
	4366	1425	0.00013	0.00147	0	0	0	0	0	0	1	-360	360;
	4366	7424	2.055194	4.798753	0	757	0	0	0	0	1	-360	360;
	8968	4366	0.00048	0.00587	0	1744	0	0	0	0	1	-360	360;
	6193	4366	0.81115	2.560526	0	757	0	0	0	0	1	-360	360;
	1009	7957	0.00095	0.004169	0	0	0	0	0	0	1	-360	360;
	8113	7957	0.01012	0.04631	0	495	0	0	0	0	1	-360	360;
	3005	2385	0.00013	0.0014	0	757	0	0	0	0	1	-360	360;
	1184	2385	0.00013	0.00143	0	1415	0	0	0	0	1	-360	360;
	2385	7920	0.091447	0.726385	0	757	0	0	0	0	1	-360	360;
	2385	7920	0.00093	0.00869	0	987	0	0	0	0	1	-360	360;
	2385	2080	0.089307	0.727078	0	757	0	0	0	0	1	-360	360;
	2385	7424	3.590651	5.504778	0	757	0	0	0	0	1	-360	360;
	2906	2385	0.00047	0.00544	0	1580	0	0	0	0	1	-360	360;
	6193	2385	1.2741	3.747645	0	757	0	0	0	0	1	-360	360;
	7966	4206	0.00142	0.01566	0	1415	0	0	0	0	1	-360	360;
	2080	6667	0.00082	0.00794	0	0	0	0	0	0	1	-360	360;
	8454	8292	0	0.0002	0	1415	0	0	0	0	1	-360	360;
	8454	8292	0.001	0.149882	0	757	0	0	0	0	1	-360	360;
	4158	8454	0.088033	0.626579	0	757	0	0	0	0	1	-360	360;
	4158	8454	0.001	0.00866	0	1283	0	0	0	0	1	-360	360;
	3319	8454	0.090852	0.619418	0	757	0	0	0	0	1	-360	360;
	9116	8454	0.87133	2.390028	0	757	0	0	0	0	1	-360	360;
	4158	8292	0.087632	0.633012	0	757	0	0	0	0	1	-360	360;
	3319	8292	0.097832	0.57669	0	757	0	0	0	0	1	-360	360;
	9116	8292	0.857064	2.463089	0	757	0	0	0	0	1	-360	360;
	8292	4113	0.0004	0.0039	0	1086	0	0	0	0	1	-360	360;
	5325	4167	0.001	0.486508	0	0	0	0	0	0	1	-360	360;
	2663	5325	0.003909	0.034711	0	591	0	0	0	0	1	-360	360;
	2663	5325	0.497789	2.055971	0	0	0	0	0	0	1	-360	360;
	5325	4804	0.357087	1.677376	0	0	0	0	0	0	1	-360	360;
	5325	7602	0.349194	1.694876	0	0	0	0	0	0	1	-360	360;
	2663	4167	0.438409	2.013802	0	0	0	0	0	0	1	-360	360;
	4167	4804	0.307107	1.517831	0	0	0	0	0	0	1	-360	360;
	4167	7602	0.354174	1.443988	0	0	0	0	0	0	1	-360	360;
	8620	2493	0.00274	0.144159	0	0	0	0	0	0	1	-360	360;
	8620	6586	0.00039	0.00175	0	0	0	0	0	0	1	-360	360;
	2493	8916	0.000401	0.001711	0	457	0	0	0	0	1	-360	360;
	7966	5100	0.06537	0.54912	0	757	0	0	0	0	1	-360	360;
	7966	1588	0.00098	0.00852	0	1086	0	0	0	0	1	-360	360;
	7966	1588	0.076039	0.582729	0	757	0	0	0	0	1	-360	360;
	7966	1994	0.00185	0.01418	0	1053	0	0	0	0	1	-360	360;
	2890	7966	0.790374	2.170014	0	757	0	0	0	0	1	-360	360;
	1675	652	0.00188	0.0087	0	495	0	0	0	0	1	-360	360;
	1675	3539	0.042461	0.437066	0	0	0	0	0	0	1	-360	360;
	417	3539	0.00068	0.008531	0	476	0	0	0	0	1	-360	360;
	3835	7920	0.00107	0.01108	0	1942	0	0	0	0	1	-360	360;
	3835	7424	0.00092	0.00947	0	1711	0	0	0	0	1	-360	360;
	4158	3319	0	0.0002	0	1678	0	0	0	0	1	-360	360;
	4158	3319	0.001	0.196378	0	757	0	0	0	0	1	-360	360;
	4158	9116	0.102652	0.915512	0	757	0	0	0	0	1	-360	360;
	3319	9116	0.110602	0.871537	0	757	0	0	0	0	1	-360	360;
	6066	3319	0.00153	0.01371	0	1316	0	0	0	0	1	-360	360;
	8899	3319	0.00296	0.02271	0	1053	0	0	0	0	1	-360	360;
	5100	1588	0	0.0002	0	757	0	0	0	0	1	-360	360;
	5100	1588	0.001	0.229661	0	757	0	0	0	0	1	-360	360;
	5100	1994	0.00076	0.0066	0	1086	0	0	0	0	1	-360	360;
	2890	5100	0.134051	0.802355	0	757	0	0	0	0	1	-360	360;
	2890	1588	0.136413	0.960942	0	757	0	0	0	0	1	-360	360;
	8520	8916	0.0027	0.02195	0	495	0	0	0	0	1	-360	360;
	8520	3303	0.001211	0.004961	0	0	0	0	0	0	1	-360	360;
	2145	7055	0.010091	0.211612	0	0	0	0	0	0	1	-360	360;
	2145	9086	0.008349	0.215682	0	0	0	0	0	0	1	-360	360;
	1120	2145	0.000401	0.001781	0	0	0	0	0	0	1	-360	360;
	7055	9086	0.00076	0.205461	0	0	0	0	0	0	1	-360	360;
	8968	6777	0.00086	0.0138	0	1711	0	0	0	0	1	-360	360;
	2906	1128	0.00101	0.01174	0	1613	0	0	0	0	1	-360	360;
	7920	2080	0	0.0002	0	1481	0	0	0	0	1	-360	360;
	7920	2080	0.001	0.17286	0	757	0	0	0	0	1	-360	360;
	7920	7424	0.917036	2.770706	0	757	0	0	0	0	1	-360	360;
	6193	7920	0.299162	1.508587	0	757	0	0	0	0	1	-360	360;
	2080	7424	0.973546	2.445291	0	757	0	0	0	0	1	-360	360;
	6193	2080	0.335817	1.383795	0	757	0	0	0	0	1	-360	360;
	782	4804	0.00114	0.00625	0	553	0	0	0	0	1	-360	360;
	5017	782	0.00069	0.00343	0	0	0	0	0	0	1	-360	360;
	7171	782	0.000669	0.00349	0	0	0	0	0	0	1	-360	360;
	1508	782	0.00264	0.01612	0	533	0	0	0	0	1	-360	360;
	3254	4226	0.001	0.01303	0	1744	0	0	0	0	1	-360	360;
	951	4226	0.00018	0.00112	0	823	0	0	0	0	1	-360	360;
	6020	1346	0.000771	0.0027	0	419	0	0	0	0	1	-360	360;
	1181	1346	0.000719	0.003039	0	0	0	0	0	0	1	-360	360;
	1346	7602	0.00138	0.005409	0	495	0	0	0	0	1	-360	360;
	1839	7617	0.00017	0.0011	0	1744	0	0	0	0	1	-360	360;
	1120	4971	0.01424	0.207211	0	0	0	0	0	0	1	-360	360;
	7127	4971	0.00288	0.023031	0	0	0	0	0	0	1	-360	360;
	9116	6274	0.00059	0.00738	0	1218	0	0	0	0	1	-360	360;
	4164	4840	2e-05	0.0002	0	1251	0	0	0	0	1	-360	360;
	3254	6605	0.00034	0.00243	0	757	0	0	0	0	1	-360	360;
	6274	2583	0.00939	0.239481	0	757	0	0	0	0	1	-360	360;
	6274	2583	0.0003	0.00385	0	757	0	0	0	0	1	-360	360;
	6297	8287	0.00082	0.0117	0	0	0	0	0	0	1	-360	360;
	6297	6296	0.000831	0.010711	0	0	0	0	0	0	1	-360	360;
	4634	2138	0	0.0002	0	1152	0	0	0	0	1	-360	360;
	4634	2138	0.001	0.424958	0	757	0	0	0	0	1	-360	360;
	4634	2906	0.00035	0.00315	0	0	0	0	0	0	1	-360	360;
	4634	2906	0.01362	0.396288	0	757	0	0	0	0	1	-360	360;
	4634	6274	0.088213	0.739543	0	757	0	0	0	0	1	-360	360;
	2138	2906	0.01388	0.388199	0	757	0	0	0	0	1	-360	360;
	2138	6274	0.0012	0.01148	0	1843	0	0	0	0	1	-360	360;
	2138	6274	0.086413	0.754086	0	757	0	0	0	0	1	-360	360;
	327	8968	0.00049	0.00475	0	1547	0	0	0	0	1	-360	360;
	2901	5622	0.002669	0.01443	0	419	0	0	0	0	1	-360	360;
	3539	652	0.00338	0.01488	0	572	0	0	0	0	1	-360	360;
	3537	8916	0.00101	0.005841	0	0	0	0	0	0	1	-360	360;
	6777	1994	0.00034	0.00381	0	1744	0	0	0	0	1	-360	360;
	2890	1994	0.00017	0.00223	0	1678	0	0	0	0	1	-360	360;
	5869	6586	0.00161	0.013789	0	553	0	0	0	0	1	-360	360;
	1994	7914	0.00032	0.00234	0	1777	0	0	0	0	1	-360	360;
	6193	7424	0.01246	0.249882	0	757	0	0	0	0	1	-360	360;
	4946	2906	0.00073	0.00838	0	1514	0	0	0	0	1	-360	360;
	4946	8968	0.01019	0.13124	0	757	0	0	0	0	1	-360	360;
	4946	8968	0.00078	0.00873	0	1580	0	0	0	0	1	-360	360;
	4946	4840	0.463179	1.173823	0	757	0	0	0	0	1	-360	360;
	4946	5972	0.428629	1.321537	0	757	0	0	0	0	1	-360	360;
	6274	2906	0.099591	0.647687	0	757	0	0	0	0	1	-360	360;
	7731	2906	0.00019	0.00245	0	1547	0	0	0	0	1	-360	360;
	8968	4840	1.591828	3.383587	0	757	0	0	0	0	1	-360	360;
	8968	5972	1.558587	3.254017	0	757	0	0	0	0	1	-360	360;
	2663	4804	1.410971	2.609298	0	0	0	0	0	0	1	-360	360;
	2663	7602	1.456178	2.474793	0	0	0	0	0	0	1	-360	360;
	4804	7602	0	0.0002	0	495	0	0	0	0	1	-360	360;
	4804	7602	0.001	0.139341	0	0	0	0	0	0	1	-360	360;
	4840	5972	0	0.0002	0	1119	0	0	0	0	1	-360	360;
	4840	5972	0.001	0.196801	0	757	0	0	0	0	1	-360	360;
	7767	1627	0.00025	0.00232	0	0	0	0	0	0	1	-360	360;
	5869	7767	0.000719	0.005461	0	0	0	0	0	0	1	-360	360;
	1627	8753	0.00012	0.00069	0	0	0	0	0	0	1	-360	360;
	6541	1627	0.003099	0.11999	0	0	0	0	0	0	1	-360	360;
	7127	9165	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7127	9165	0.001	0.13145	0	0	0	0	0	0	1	-360	360;
	3537	5869	0.10799	0.554256	0	0	0	0	0	0	1	-360	360;
	8333	8553	0.00077	0.00927	0	1744	0	0	0	0	1	-360	360;
	6538	8553	0.00078	0.00908	0	0	0	0	0	0	1	-360	360;
	4859	6894	0.00038	0.0045	0	790	0	0	0	0	1	-360	360;
	6894	6458	0.00147	0.0125	0	1777	0	0	0	0	1	-360	360;
	6668	6894	0.00039	0.00441	0	757	0	0	0	0	1	-360	360;
	6894	1243	0.00202	0.01654	0	1613	0	0	0	0	1	-360	360;
	6538	8333	0	0.0002	0	1448	0	0	0	0	1	-360	360;
	6538	8333	0.00147	0.112341	0	757	0	0	0	0	1	-360	360;
	6538	4859	0.00069	0.00564	0	1613	0	0	0	0	1	-360	360;
	8333	6668	0.00068	0.00574	0	1580	0	0	0	0	1	-360	360;
	4859	6668	0	0.0002	0	757	0	0	0	0	1	-360	360;
	4859	6668	0.00088	0.07777	0	757	0	0	0	0	1	-360	360;
	590	4859	0.00062	0.00515	0	1185	0	0	0	0	1	-360	360;
	2394	6668	0.0007	0.00505	0	1185	0	0	0	0	1	-360	360;
	590	2394	0	0.0002	0	1086	0	0	0	0	1	-360	360;
	590	2394	0.00076	0.06481	0	757	0	0	0	0	1	-360	360;
	3090	459	0.00011	0.00149	0	0	0	0	0	0	1	-360	360;
	5894	1927	9.9e-05	0.000979	0	0	0	0	0	0	1	-360	360;
	5433	3934	0.01788	0.227831	0	0	0	0	0	0	1	-360	360;
	5433	3934	0.004669	0.015969	0	419	0	0	0	0	1	-360	360;
	6830	5433	0.011229	0.03868	0	381	0	0	0	0	1	-360	360;
	388	5433	0.01001	0.043841	0	381	0	0	0	0	1	-360	360;
	5433	476	0.002151	0.01032	0	0	0	0	0	0	1	-360	360;
	4256	5896	0.003281	0.017901	0	629	0	0	0	0	1	-360	360;
	2439	4256	0.005271	0.026909	0	457	0	0	0	0	1	-360	360;
	4256	8451	0.00044	0.002169	0	495	0	0	0	0	1	-360	360;
	4256	8451	0.00045	0.00213	0	495	0	0	0	0	1	-360	360;
	3149	6227	0.452251	2.097438	0	757	0	0	0	0	1	-360	360;
	6227	1214	0.247659	1.380679	0	757	0	0	0	0	1	-360	360;
	1253	6227	0.436378	1.290097	0	757	0	0	0	0	1	-360	360;
	5528	6227	0.218089	1.220914	0	757	0	0	0	0	1	-360	360;
	4786	6227	0.0612	1.024169	0	757	0	0	0	0	1	-360	360;
	6027	6227	0.06168	0.959765	0	757	0	0	0	0	1	-360	360;
	6027	6227	0.0011	0.00838	0	1613	0	0	0	0	1	-360	360;
	4294	6227	0.05459	0.962258	0	757	0	0	0	0	1	-360	360;
	5528	6227	0.00126	0.01278	0	1514	0	0	0	0	1	-360	360;
	8541	1614	5e-05	0.000229	0	0	0	0	0	0	1	-360	360;
	4593	8541	0.001331	0.009331	0	629	0	0	0	0	1	-360	360;
	8541	9200	6e-05	0.00031	0	457	0	0	0	0	1	-360	360;
	4593	6418	0.00155	0.00805	0	0	0	0	0	0	1	-360	360;
	6418	158	0.351674	1.414483	0	0	0	0	0	0	1	-360	360;
	4593	6418	0.239545	1.416095	0	0	0	0	0	0	1	-360	360;
	5228	1927	0.00011	0.00088	0	381	0	0	0	0	1	-360	360;
	1622	459	0.00012	0.00064	0	0	0	0	0	0	1	-360	360;
	8216	459	0.002349	0.01088	0	438	0	0	0	0	1	-360	360;
	8216	459	0.0023	0.011099	0	419	0	0	0	0	1	-360	360;
	459	853	0.004771	0.03713	0	0	0	0	0	0	1	-360	360;
	8073	1927	0.00056	0.0013	0	0	0	0	0	0	1	-360	360;
	5779	1927	0.001091	0.007409	0	438	0	0	0	0	1	-360	360;
	5779	1927	0.00168	0.007479	0	0	0	0	0	0	1	-360	360;
	1432	1927	0.003151	0.01876	0	514	0	0	0	0	1	-360	360;
	1160	2647	0.005719	0.255764	0	0	0	0	0	0	1	-360	360;
	3155	1160	0.000599	0.00294	0	0	0	0	0	0	1	-360	360;
	3155	2647	0.00062	0.00288	0	553	0	0	0	0	1	-360	360;
	5673	4785	0.001169	0.012591	0	0	0	0	0	0	1	-360	360;
	5673	4785	0.00131	0.01119	0	0	0	0	0	0	1	-360	360;
	6407	4785	0.000159	0.00137	0	0	0	0	0	0	1	-360	360;
	4105	4785	0.000159	0.001341	0	0	0	0	0	0	1	-360	360;
	8516	1293	0.00051	0.0025	0	0	0	0	0	0	1	-360	360;
	4471	7097	0.00031	0.001289	0	457	0	0	0	0	1	-360	360;
	8516	7097	0.05163	0.55124	0	0	0	0	0	0	1	-360	360;
	7655	7097	0.201229	1.043864	0	0	0	0	0	0	1	-360	360;
	7748	5779	1.027128	2.385331	0	0	0	0	0	0	1	-360	360;
	7748	167	0.288306	1.382335	0	0	0	0	0	0	1	-360	360;
	6997	7748	0.05949	0.340661	0	0	0	0	0	0	1	-360	360;
	7748	2614	0.007211	0.030409	0	0	0	0	0	0	1	-360	360;
	8639	7748	0.899153	2.660744	0	0	0	0	0	0	1	-360	360;
	5580	7748	0.809979	2.92686	0	0	0	0	0	0	1	-360	360;
	7748	5324	0.00251	0.011211	0	419	0	0	0	0	1	-360	360;
	8113	4793	0.08695	0.512231	0	0	0	0	0	0	1	-360	360;
	8113	8798	0.006289	0.693946	0	0	0	0	0	0	1	-360	360;
	1657	8113	0.140669	0.495269	0	0	0	0	0	0	1	-360	360;
	8778	8113	0.00882	0.041831	0	476	0	0	0	0	1	-360	360;
	6977	1825	2e-05	0.00025	0	757	0	0	0	0	1	-360	360;
	5152	4434	0.000159	0.00074	0	0	0	0	0	0	1	-360	360;
	1004	6661	0.004469	0.139591	0	0	0	0	0	0	1	-360	360;
	1004	2413	0.001039	0.006211	0	438	0	0	0	0	1	-360	360;
	7430	9095	0.00813	0.186659	0	0	0	0	0	0	1	-360	360;
	7430	2413	0.00101	0.00625	0	400	0	0	0	0	1	-360	360;
	5943	571	8.1e-05	0.000781	0	438	0	0	0	0	1	-360	360;
	5943	6948	0.0003	0.002031	0	476	0	0	0	0	1	-360	360;
	7125	571	9.1e-05	0.000719	0	629	0	0	0	0	1	-360	360;
	7125	8123	0.00086	0.005581	0	629	0	0	0	0	1	-360	360;
	4474	1214	0.00062	0.00605	0	856	0	0	0	0	1	-360	360;
	4466	571	0.00181	0.009599	0	0	0	0	0	0	1	-360	360;
	6321	571	0.00099	0.00481	0	629	0	0	0	0	1	-360	360;
	571	4496	0.00038	0.00218	0	0	0	0	0	0	1	-360	360;
	3655	1400	0.00031	0.00342	0	1777	0	0	0	0	1	-360	360;
	7116	1400	0.00083	0.00929	0	1415	0	0	0	0	1	-360	360;
	3249	1400	0.00074	0.00793	0	0	0	0	0	0	1	-360	360;
	3116	3249	0.00016	0.00166	0	1843	0	0	0	0	1	-360	360;
	1400	8187	0	0.0002	0	1119	0	0	0	0	1	-360	360;
	1400	8187	0.01471	0.443012	0	757	0	0	0	0	1	-360	360;
	1323	8187	0.00037	0.00337	0	1843	0	0	0	0	1	-360	360;
	6830	2903	0.001461	0.006711	0	0	0	0	0	0	1	-360	360;
	1855	2903	0.001479	0.00656	0	591	0	0	0	0	1	-360	360;
	7220	4793	0.00068	0.0113	0	0	0	0	0	0	1	-360	360;
	4879	3116	0.00019	0.00196	0	757	0	0	0	0	1	-360	360;
	2755	3116	0.00019	0.00198	0	757	0	0	0	0	1	-360	360;
	1323	5380	0.00886	0.294079	0	757	0	0	0	0	1	-360	360;
	1323	5380	0.00028	0.00275	0	1711	0	0	0	0	1	-360	360;
	3655	5380	0.01321	0.296717	0	757	0	0	0	0	1	-360	360;
	8346	8816	0.001	0.169901	0	0	0	0	0	0	1	-360	360;
	8346	158	0.03111	0.255	0	0	0	0	0	0	1	-360	360;
	4593	8346	0.00057	0.002849	0	0	0	0	0	0	1	-360	360;
	8816	158	0.03143	0.234731	0	0	0	0	0	0	1	-360	360;
	4593	8816	0.000581	0.002789	0	553	0	0	0	0	1	-360	360;
	8602	3905	0.00157	0.007	0	572	0	0	0	0	1	-360	360;
	8602	683	0.00131	0.00636	0	591	0	0	0	0	1	-360	360;
	7451	3905	0.001669	0.00724	0	514	0	0	0	0	1	-360	360;
	3905	4146	0.00663	0.249339	0	0	0	0	0	0	1	-360	360;
	3905	1147	0.000169	0.00032	0	0	0	0	0	0	1	-360	360;
	3905	3947	0.000151	0.00038	0	0	0	0	0	0	1	-360	360;
	3905	2655	0.000841	0.009581	0	0	0	0	0	0	1	-360	360;
	3905	540	0.0355	0.360558	0	0	0	0	0	0	1	-360	360;
	3905	683	0.027539	0.325145	0	0	0	0	0	0	1	-360	360;
	3905	3854	0.025271	0.339442	0	0	0	0	0	0	1	-360	360;
	3905	8624	0.000169	0.000219	0	0	0	0	0	0	1	-360	360;
	2433	158	0.000401	0.0018	0	0	0	0	0	0	1	-360	360;
	4593	2433	0.00205	0.01162	0	0	0	0	0	0	1	-360	360;
	2712	2433	0.000651	0.0045	0	0	0	0	0	0	1	-360	360;
	2944	158	0.00038	0.00187	0	0	0	0	0	0	1	-360	360;
	4593	2944	0.002091	0.01139	0	438	0	0	0	0	1	-360	360;
	2712	2944	0.00064	0.004169	0	0	0	0	0	0	1	-360	360;
	4593	158	0.03551	0.421116	0	0	0	0	0	0	1	-360	360;
	4593	886	0.090159	0.624236	0	0	0	0	0	0	1	-360	360;
	4593	7997	0.025919	0.457603	0	0	0	0	0	0	1	-360	360;
	4593	2610	0.140581	0.992087	0	0	0	0	0	0	1	-360	360;
	4593	8033	0.00082	0.007331	0	629	0	0	0	0	1	-360	360;
	8216	822	0.008469	0.038599	0	0	0	0	0	0	1	-360	360;
	8216	6452	0.003729	0.012081	0	457	0	0	0	0	1	-360	360;
	6997	5779	0.071281	0.33186	0	0	0	0	0	0	1	-360	360;
	6997	5779	0.011919	0.055651	0	419	0	0	0	0	1	-360	360;
	5779	9225	0.0038	0.01182	0	457	0	0	0	0	1	-360	360;
	3116	7731	0.00233	0.0191	0	0	0	0	0	0	1	-360	360;
	3116	6343	0.00258	0.02125	0	1810	0	0	0	0	1	-360	360;
	3116	7731	0.00246	0.01795	0	0	0	0	0	0	1	-360	360;
	3116	9160	0.00231	0.02199	0	1645	0	0	0	0	1	-360	360;
	7116	3116	0.00105	0.0114	0	1909	0	0	0	0	1	-360	360;
	3116	873	0.00268	0.02035	0	1843	0	0	0	0	1	-360	360;
	4466	1993	0.00124	0.005711	0	0	0	0	0	0	1	-360	360;
	4466	9198	0.0002	0.0022	0	572	0	0	0	0	1	-360	360;
	4466	2784	0.00019	0.00225	0	0	0	0	0	0	1	-360	360;
	4466	6886	0.001169	0.005159	0	553	0	0	0	0	1	-360	360;
	4466	8479	0.001271	0.005591	0	0	0	0	0	0	1	-360	360;
	4466	3321	0.00182	0.01218	0	476	0	0	0	0	1	-360	360;
	4466	7161	0.00107	0.005789	0	0	0	0	0	0	1	-360	360;
	6470	79	0.00026	0.00275	0	1613	0	0	0	0	1	-360	360;
	6470	798	0.0001	0.00093	0	1909	0	0	0	0	1	-360	360;
	9154	8270	0.00019	0.00166	0	1876	0	0	0	0	1	-360	360;
	9154	8270	0.00019	0.00163	0	1909	0	0	0	0	1	-360	360;
	3772	5491	0.00126	0.00953	0	1810	0	0	0	0	1	-360	360;
	3772	7725	0.00131	0.00994	0	1942	0	0	0	0	1	-360	360;
	1943	5491	0.00128	0.00933	0	1777	0	0	0	0	1	-360	360;
	1943	9143	0.00134	0.00972	0	1777	0	0	0	0	1	-360	360;
	3772	1943	0.001	0.180201	0	757	0	0	0	0	1	-360	360;
	3772	7725	1.365928	3.472299	0	757	0	0	0	0	1	-360	360;
	3772	9143	0.186177	1.029224	0	757	0	0	0	0	1	-360	360;
	680	3772	0.98608	2.33608	0	757	0	0	0	0	1	-360	360;
	3772	7715	0.487888	1.618975	0	757	0	0	0	0	1	-360	360;
	3772	5387	0.850277	2.290235	0	757	0	0	0	0	1	-360	360;
	1943	7725	1.29162	3.486565	0	757	0	0	0	0	1	-360	360;
	1943	9143	0.201801	0.8991	0	757	0	0	0	0	1	-360	360;
	680	1943	0.842729	2.35651	0	757	0	0	0	0	1	-360	360;
	1943	7715	0.415547	1.630194	0	757	0	0	0	0	1	-360	360;
	1943	5387	0.830956	1.995914	0	757	0	0	0	0	1	-360	360;
	476	8955	0.005729	0.021281	0	400	0	0	0	0	1	-360	360;
	476	8955	0.005081	0.029331	0	0	0	0	0	0	1	-360	360;
	6981	8955	0.00576	0.035781	0	0	0	0	0	0	1	-360	360;
	3934	8955	0.002021	0.00976	0	533	0	0	0	0	1	-360	360;
	3176	8955	0.00563	0.03313	0	457	0	0	0	0	1	-360	360;
	8725	6965	0.001	0.260568	0	757	0	0	0	0	1	-360	360;
	6906	8725	0.127098	1.298546	0	757	0	0	0	0	1	-360	360;
	5754	8725	0.04182	0.986219	0	757	0	0	0	0	1	-360	360;
	5754	8725	0.00129	0.01393	0	1909	0	0	0	0	1	-360	360;
	2260	8725	0.00082	0.00915	0	1876	0	0	0	0	1	-360	360;
	330	8725	0.00119	0.01287	0	1777	0	0	0	0	1	-360	360;
	6906	6965	0.118677	1.418975	0	757	0	0	0	0	1	-360	360;
	6906	6965	0.00148	0.01585	0	1909	0	0	0	0	1	-360	360;
	5754	6965	0.04574	0.913504	0	757	0	0	0	0	1	-360	360;
	7781	6965	0.00116	0.01192	0	1448	0	0	0	0	1	-360	360;
	7116	1148	0.06698	0.909972	0	757	0	0	0	0	1	-360	360;
	7116	1148	0.00042	0.00331	0	0	0	0	0	0	1	-360	360;
	1690	4140	0.002591	0.01299	0	0	0	0	0	0	1	-360	360;
	4140	2300	0.01169	0.067849	0	0	0	0	0	0	1	-360	360;
	7324	3709	0.00074	0.00368	0	438	0	0	0	0	1	-360	360;
	7324	167	0.00064	0.0042	0	0	0	0	0	0	1	-360	360;
	7324	5324	0.0018	0.01139	0	591	0	0	0	0	1	-360	360;
	842	3709	0.00082	0.003401	0	476	0	0	0	0	1	-360	360;
	842	167	0.00069	0.00381	0	0	0	0	0	0	1	-360	360;
	842	5324	0.00189	0.01094	0	591	0	0	0	0	1	-360	360;
	3709	5184	0.000151	0.000711	0	0	0	0	0	0	1	-360	360;
	3709	1359	0.00013	0.000789	0	0	0	0	0	0	1	-360	360;
	3709	7335	5e-05	0.00024	0	0	0	0	0	0	1	-360	360;
	3709	867	5e-05	0.00025	0	0	0	0	0	0	1	-360	360;
	8639	3709	0.00012	0.000711	0	0	0	0	0	0	1	-360	360;
	5580	3709	0.00011	0.000719	0	0	0	0	0	0	1	-360	360;
	657	4154	0.003091	0.015719	0	0	0	0	0	0	1	-360	360;
	4154	8638	0.018159	0.39124	0	0	0	0	0	0	1	-360	360;
	4154	6336	0.169419	0.883946	0	0	0	0	0	0	1	-360	360;
	4154	3479	0.656054	1.65126	0	0	0	0	0	0	1	-360	360;
	4154	5232	0.452913	1.939091	0	0	0	0	0	0	1	-360	360;
	4154	8699	0.055969	0.74907	0	0	0	0	0	0	1	-360	360;
	3777	4154	0.549483	1.897438	0	0	0	0	0	0	1	-360	360;
	5261	4154	0.483574	1.959731	0	0	0	0	0	0	1	-360	360;
	4154	7620	0.279731	1.818967	0	0	0	0	0	0	1	-360	360;
	4154	7620	0.006521	0.02749	0	381	0	0	0	0	1	-360	360;
	7444	6083	0.07687	1.255956	0	757	0	0	0	0	1	-360	360;
	7444	6083	0.00068	0.00577	0	1678	0	0	0	0	1	-360	360;
	4471	6973	0.00812	0.12314	0	0	0	0	0	0	1	-360	360;
	17	4471	0.00157	0.007031	0	514	0	0	0	0	1	-360	360;
	2112	4471	0.001599	0.00688	0	495	0	0	0	0	1	-360	360;
	4471	8516	0.00113	0.00724	0	629	0	0	0	0	1	-360	360;
	4471	8516	0.001271	0.007091	0	591	0	0	0	0	1	-360	360;
	657	8699	0.000919	0.005211	0	572	0	0	0	0	1	-360	360;
	7451	6321	0.00199	0.010219	0	514	0	0	0	0	1	-360	360;
	7451	8544	0.005409	0.02475	0	514	0	0	0	0	1	-360	360;
	7451	4854	0.00482	0.02805	0	572	0	0	0	0	1	-360	360;
	7451	3261	0.028479	0.302273	0	0	0	0	0	0	1	-360	360;
	7451	6336	0.11506	0.606198	0	0	0	0	0	0	1	-360	360;
	7451	6321	0.025271	0.321983	0	0	0	0	0	0	1	-360	360;
	7451	8915	0.00194	0.010091	0	0	0	0	0	0	1	-360	360;
	7451	3261	0.00314	0.019719	0	0	0	0	0	0	1	-360	360;
	4121	7028	0.00042	0.00384	0	1909	0	0	0	0	1	-360	360;
	4121	798	0.00031	0.00229	0	1744	0	0	0	0	1	-360	360;
	4121	1116	0.00024	0.00235	0	1645	0	0	0	0	1	-360	360;
	4121	2429	0.00103	0.01072	0	1678	0	0	0	0	1	-360	360;
	2443	8516	0.00275	0.01311	0	495	0	0	0	0	1	-360	360;
	2443	8552	0.001771	0.01225	0	476	0	0	0	0	1	-360	360;
	309	5896	0.001	0.385207	0	0	0	0	0	0	1	-360	360;
	309	5896	0.003349	0.022711	0	438	0	0	0	0	1	-360	360;
	309	5149	0.00032	0.001531	0	0	0	0	0	0	1	-360	360;
	309	6120	0.359525	1.472438	0	0	0	0	0	0	1	-360	360;
	309	5152	1.632355	4.693388	0	0	0	0	0	0	1	-360	360;
	309	8134	0.08494	0.651364	0	0	0	0	0	0	1	-360	360;
	309	8123	0.052581	0.671529	0	0	0	0	0	0	1	-360	360;
	8451	309	0.12794	1.157789	0	0	0	0	0	0	1	-360	360;
	683	309	0.002461	0.014531	0	629	0	0	0	0	1	-360	360;
	7652	6808	0.002659	0.013271	0	591	0	0	0	0	1	-360	360;
	7652	6808	0.002711	0.01299	0	0	0	0	0	0	1	-360	360;
	4675	5152	3.9e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	3202	5152	5e-05	0.00031	0	0	0	0	0	0	1	-360	360;
	3552	6864	0.00089	0.00698	0	0	0	0	0	0	1	-360	360;
	3552	5528	0.00028	0.00271	0	0	0	0	0	0	1	-360	360;
	1706	3149	0.00025	0.00273	0	1349	0	0	0	0	1	-360	360;
	1706	1214	0.00027	0.00238	0	1613	0	0	0	0	1	-360	360;
	1706	5528	0.00027	0.00277	0	1942	0	0	0	0	1	-360	360;
	3149	1214	0.00808	0.60428	0	757	0	0	0	0	1	-360	360;
	1253	3149	0.01086	0.386371	0	757	0	0	0	0	1	-360	360;
	3149	5528	0.03494	0.767659	0	757	0	0	0	0	1	-360	360;
	3149	4786	0.0265	0.810665	0	757	0	0	0	0	1	-360	360;
	6027	3149	0.524993	2.841898	0	757	0	0	0	0	1	-360	360;
	4294	3149	0.475568	2.62313	0	757	0	0	0	0	1	-360	360;
	8639	167	0.02113	0.408595	0	0	0	0	0	0	1	-360	360;
	5580	167	0.022469	0.384442	0	0	0	0	0	0	1	-360	360;
	9160	6413	0.00045	0.0038	0	0	0	0	0	0	1	-360	360;
	9160	6413	0.00047	0.00374	0	1843	0	0	0	0	1	-360	360;
	8638	7714	0.00019	0.00086	0	400	0	0	0	0	1	-360	360;
	6336	8638	0.542397	1.40281	0	0	0	0	0	0	1	-360	360;
	3479	8638	1.299236	2.713017	0	0	0	0	0	0	1	-360	360;
	8638	5232	1.066591	3.481612	0	0	0	0	0	0	1	-360	360;
	8638	8699	0.276529	1.159153	0	0	0	0	0	0	1	-360	360;
	3777	8638	1.192211	3.154339	0	0	0	0	0	0	1	-360	360;
	5261	8638	1.210785	2.840289	0	0	0	0	0	0	1	-360	360;
	7620	8638	0.804979	2.837397	0	0	0	0	0	0	1	-360	360;
	886	9190	9.1e-05	0.00055	0	0	0	0	0	0	1	-360	360;
	7997	886	0.009909	0.377748	0	0	0	0	0	0	1	-360	360;
	886	2610	0.914793	3.247727	0	0	0	0	0	0	1	-360	360;
	3038	9190	0.000349	0.002159	0	629	0	0	0	0	1	-360	360;
	7997	7714	0.00132	0.00624	0	533	0	0	0	0	1	-360	360;
	7997	9190	0.00088	0.005159	0	514	0	0	0	0	1	-360	360;
	5812	2610	0.001349	0.01051	0	476	0	0	0	0	1	-360	360;
	532	2029	0.000539	0.003099	0	610	0	0	0	0	1	-360	360;
	532	8033	0.00207	0.01182	0	572	0	0	0	0	1	-360	360;
	6973	8516	0.000539	0.003159	0	0	0	0	0	0	1	-360	360;
	6973	8552	0.004081	0.0247	0	0	0	0	0	0	1	-360	360;
	7609	4151	0.000531	0.003521	0	495	0	0	0	0	1	-360	360;
	7944	1992	0.000539	0.00345	0	0	0	0	0	0	1	-360	360;
	1657	7944	0.005169	0.02325	0	0	0	0	0	0	1	-360	360;
	7652	4854	0.000669	0.004099	0	572	0	0	0	0	1	-360	360;
	7652	6438	0.000599	0.00344	0	0	0	0	0	0	1	-360	360;
	388	7652	0.00139	0.008919	0	0	0	0	0	0	1	-360	360;
	7652	8544	0.00069	0.00419	0	572	0	0	0	0	1	-360	360;
	1855	7652	0.001539	0.008099	0	0	0	0	0	0	1	-360	360;
	6336	2705	0.00062	0.00356	0	495	0	0	0	0	1	-360	360;
	2705	8699	0.001091	0.006479	0	438	0	0	0	0	1	-360	360;
	8960	7318	6e-05	0.00052	0	1942	0	0	0	0	1	-360	360;
	8960	7942	0.00023	0.00485	0	1876	0	0	0	0	1	-360	360;
	5528	7318	7e-05	0.00051	0	1909	0	0	0	0	1	-360	360;
	7488	5896	5e-05	0.000229	0	0	0	0	0	0	1	-360	360;
	1318	5896	8.1e-05	0.0003	0	0	0	0	0	0	1	-360	360;
	4092	5896	6e-05	0.000341	0	0	0	0	0	0	1	-360	360;
	6120	5896	0.339112	2.39938	0	0	0	0	0	0	1	-360	360;
	5152	5896	0.982252	4.492975	0	0	0	0	0	0	1	-360	360;
	5896	8134	0.299979	2.042583	0	0	0	0	0	0	1	-360	360;
	5896	8123	0.216508	1.7125	0	0	0	0	0	0	1	-360	360;
	83	5896	0.002841	0.01551	0	0	0	0	0	0	1	-360	360;
	8451	5896	0.00326	1.128223	0	0	0	0	0	0	1	-360	360;
	4557	5896	0.00232	0.016961	0	0	0	0	0	0	1	-360	360;
	2212	5896	0.003979	0.02513	0	495	0	0	0	0	1	-360	360;
	7360	5896	0.002281	0.017289	0	591	0	0	0	0	1	-360	360;
	1253	1214	0.111143	0.652313	0	757	0	0	0	0	1	-360	360;
	18	1214	0.00045	0.00399	0	1711	0	0	0	0	1	-360	360;
	5528	1214	0.01009	0.588213	0	757	0	0	0	0	1	-360	360;
	4786	1214	0.0029	0.60196	0	757	0	0	0	0	1	-360	360;
	6027	1214	0.468573	2.303532	0	757	0	0	0	0	1	-360	360;
	4294	1214	0.401032	2.262812	0	757	0	0	0	0	1	-360	360;
	5528	1214	0.00054	0.00451	0	1777	0	0	0	0	1	-360	360;
	18	1214	0.0004	0.00439	0	1645	0	0	0	0	1	-360	360;
	2364	2610	0.00132	0.0077	0	0	0	0	0	0	1	-360	360;
	2364	8699	0.002289	0.013909	0	457	0	0	0	0	1	-360	360;
	5011	3160	0.00015	0.00153	0	1448	0	0	0	0	1	-360	360;
	5011	6164	0.00022	0.00174	0	1777	0	0	0	0	1	-360	360;
	5011	9221	0.00031	0.00276	0	1514	0	0	0	0	1	-360	360;
	7997	2610	0.498802	2.231818	0	0	0	0	0	0	1	-360	360;
	7997	8699	0.00243	0.01461	0	533	0	0	0	0	1	-360	360;
	7997	8033	0.0013	0.005919	0	553	0	0	0	0	1	-360	360;
	5152	6120	2.422314	7.164876	0	0	0	0	0	0	1	-360	360;
	5152	6120	0.005271	0.03874	0	572	0	0	0	0	1	-360	360;
	6120	8134	0.157781	1.324793	0	0	0	0	0	0	1	-360	360;
	6120	8134	0.00695	0.055591	0	572	0	0	0	0	1	-360	360;
	6120	8123	0.211777	1.52969	0	0	0	0	0	0	1	-360	360;
	8451	6120	0.1823	1.686529	0	0	0	0	0	0	1	-360	360;
	2176	8427	2.525346	8.326177	0	757	0	0	0	0	1	-360	360;
	2176	8427	0.00086	0.00871	0	0	0	0	0	0	1	-360	360;
	8427	1823	0.13018	1.288158	0	757	0	0	0	0	1	-360	360;
	8427	3583	5.25734	15.96122	0	757	0	0	0	0	1	-360	360;
	8427	6864	0.196787	1.606648	0	757	0	0	0	0	1	-360	360;
	8687	1823	0.00124	0.01282	0	1645	0	0	0	0	1	-360	360;
	17	2112	0	0.0002	0	0	0	0	0	0	1	-360	360;
	17	2112	0.227149	1.244607	0	0	0	0	0	0	1	-360	360;
	17	2029	0.024581	0.402459	0	0	0	0	0	0	1	-360	360;
	2712	17	0.04388	0.657934	0	0	0	0	0	0	1	-360	360;
	6981	17	0.12525	0.756384	0	0	0	0	0	0	1	-360	360;
	3176	17	0.18987	0.937128	0	0	0	0	0	0	1	-360	360;
	17	8775	0.00163	0.007031	0	0	0	0	0	0	1	-360	360;
	2112	2029	0.456736	1.546157	0	0	0	0	0	0	1	-360	360;
	2712	2112	0.03062	0.528388	0	0	0	0	0	0	1	-360	360;
	6981	2112	0.192099	0.773678	0	0	0	0	0	0	1	-360	360;
	3176	2112	0.027081	0.45345	0	0	0	0	0	0	1	-360	360;
	3176	2112	0.002211	0.01299	0	0	0	0	0	0	1	-360	360;
	2102	6155	0.00344	0.015909	0	0	0	0	0	0	1	-360	360;
	1253	5528	0.123539	0.685277	0	757	0	0	0	0	1	-360	360;
	1253	4786	0.096593	0.575748	0	757	0	0	0	0	1	-360	360;
	1253	6027	0.225512	1.104709	0	757	0	0	0	0	1	-360	360;
	1253	4294	0.17135	1.111427	0	757	0	0	0	0	1	-360	360;
	1253	5528	0.00117	0.00907	0	1810	0	0	0	0	1	-360	360;
	3714	1765	3e-05	0.00039	0	1777	0	0	0	0	1	-360	360;
	624	3714	0.00079	0.0065	0	1744	0	0	0	0	1	-360	360;
	4718	3714	0.00091	0.00755	0	0	0	0	0	0	1	-360	360;
	624	3714	0.00081	0.00636	0	1843	0	0	0	0	1	-360	360;
	4718	3714	0.00086	0.00818	0	1843	0	0	0	0	1	-360	360;
	2439	5152	0.003979	0.02364	0	591	0	0	0	0	1	-360	360;
	5152	8134	6.941322	16.0093	0	0	0	0	0	0	1	-360	360;
	5152	8123	6.73657	15.66095	0	0	0	0	0	0	1	-360	360;
	8451	5152	0.185229	2.220661	0	0	0	0	0	0	1	-360	360;
	2176	1823	5.786496	13.8054	0	757	0	0	0	0	1	-360	360;
	2176	3583	0.04375	0.648878	0	757	0	0	0	0	1	-360	360;
	2176	606	0.0008	0.00847	0	1711	0	0	0	0	1	-360	360;
	2176	6864	0.19205	1.80741	0	757	0	0	0	0	1	-360	360;
	2176	6864	0.00157	0.01084	0	0	0	0	0	0	1	-360	360;
	4501	2176	0.00155	0.01203	0	0	0	0	0	0	1	-360	360;
	2176	4117	0.00062	0.00656	0	1777	0	0	0	0	1	-360	360;
	2176	1823	0.00203	0.0219	0	0	0	0	0	0	1	-360	360;
	2176	7849	0.00017	0.00135	0	1876	0	0	0	0	1	-360	360;
	4875	8479	0.000341	0.001729	0	419	0	0	0	0	1	-360	360;
	3479	4875	0.00101	0.00926	0	553	0	0	0	0	1	-360	360;
	2413	4875	0.00055	0.00262	0	0	0	0	0	0	1	-360	360;
	1993	8479	0.018781	0.265517	0	0	0	0	0	0	1	-360	360;
	4941	1993	0.0298	0.392211	0	0	0	0	0	0	1	-360	360;
	1993	2390	0.00057	0.002841	0	533	0	0	0	0	1	-360	360;
	2413	1993	0.00943	0.27814	0	0	0	0	0	0	1	-360	360;
	4941	8479	0.027961	0.338967	0	0	0	0	0	0	1	-360	360;
	8479	3542	0.000581	0.002781	0	0	0	0	0	0	1	-360	360;
	2413	8479	0.010831	0.239876	0	0	0	0	0	0	1	-360	360;
	4941	5644	0.00261	0.01243	0	381	0	0	0	0	1	-360	360;
	2665	4146	0.00214	0.01187	0	533	0	0	0	0	1	-360	360;
	540	4146	0.04931	0.648058	0	0	0	0	0	0	1	-360	360;
	540	4146	0.001711	0.010781	0	572	0	0	0	0	1	-360	360;
	4146	4425	0.00068	0.003271	0	0	0	0	0	0	1	-360	360;
	683	4146	0.07599	0.788884	0	0	0	0	0	0	1	-360	360;
	3854	4146	0.05989	0.967314	0	0	0	0	0	0	1	-360	360;
	7245	4146	0.00036	0.002719	0	0	0	0	0	0	1	-360	360;
	829	1116	0.00022	0.00201	0	1711	0	0	0	0	1	-360	360;
	5754	829	0.04617	0.493712	0	757	0	0	0	0	1	-360	360;
	829	7399	0.00031	0.00302	0	1810	0	0	0	0	1	-360	360;
	6906	5754	0.001	0.254681	0	757	0	0	0	0	1	-360	360;
	6906	2429	0.00034	0.00288	0	1777	0	0	0	0	1	-360	360;
	5061	9143	8e-05	0.00085	0	1283	0	0	0	0	1	-360	360;
	5061	9143	8e-05	0.00087	0	1316	0	0	0	0	1	-360	360;
	5061	9143	7e-05	0.00089	0	1316	0	0	0	0	1	-360	360;
	3617	6155	0.808079	2.998967	0	0	0	0	0	0	1	-360	360;
	296	6155	0.11606	0.766157	0	0	0	0	0	0	1	-360	360;
	4268	6155	0.044669	0.42969	0	0	0	0	0	0	1	-360	360;
	4268	6155	0.0045	0.023841	0	438	0	0	0	0	1	-360	360;
	4047	417	0.009969	0.07739	0	0	0	0	0	0	1	-360	360;
	1508	4047	0.01431	0.10638	0	381	0	0	0	0	1	-360	360;
	4047	5104	0.00774	0.06957	0	495	0	0	0	0	1	-360	360;
	129	417	0.016669	0.26188	0	0	0	0	0	0	1	-360	360;
	8798	417	0.009969	0.061409	0	591	0	0	0	0	1	-360	360;
	129	417	0.001729	0.00995	0	400	0	0	0	0	1	-360	360;
	7731	6066	0.0015	0.01398	0	1251	0	0	0	0	1	-360	360;
	4718	7731	0.00243	0.01942	0	1415	0	0	0	0	1	-360	360;
	8503	822	0.00455	0.02705	0	495	0	0	0	0	1	-360	360;
	5324	8503	0.004331	0.01676	0	591	0	0	0	0	1	-360	360;
	8503	1949	0.24907	0.891963	0	0	0	0	0	0	1	-360	360;
	8503	822	0.033831	0.228099	0	0	0	0	0	0	1	-360	360;
	296	3617	0.344318	0.959132	0	0	0	0	0	0	1	-360	360;
	3617	4268	0.087211	0.561364	0	0	0	0	0	0	1	-360	360;
	3617	1497	0.003909	0.019531	0	0	0	0	0	0	1	-360	360;
	6589	798	0.01509	0.413102	0	757	0	0	0	0	1	-360	360;
	6589	1955	9e-05	0.001	0	0	0	0	0	0	1	-360	360;
	6589	6524	0.00094	0.00943	0	1711	0	0	0	0	1	-360	360;
	6589	4500	0.03361	0.58162	0	757	0	0	0	0	1	-360	360;
	4139	2260	0.0001	0.00107	0	1580	0	0	0	0	1	-360	360;
	4139	873	0.00084	0.00882	0	1580	0	0	0	0	1	-360	360;
	2439	296	0.00501	0.021599	0	572	0	0	0	0	1	-360	360;
	296	4268	0.04149	0.219607	0	0	0	0	0	0	1	-360	360;
	296	1773	0.000781	0.00744	0	629	0	0	0	0	1	-360	360;
	5958	8059	0.011789	0.277645	0	0	0	0	0	0	1	-360	360;
	5958	8059	0.0052	0.02574	0	0	0	0	0	0	1	-360	360;
	8059	6603	0.031669	0.507603	0	0	0	0	0	0	1	-360	360;
	8059	6603	0.00438	0.025461	0	0	0	0	0	0	1	-360	360;
	1690	8059	0.27186	0.993388	0	0	0	0	0	0	1	-360	360;
	79	798	0.0037	0.327618	0	757	0	0	0	0	1	-360	360;
	798	4500	0.01026	0.269931	0	757	0	0	0	0	1	-360	360;
	3155	592	0.000281	0.0023	0	0	0	0	0	0	1	-360	360;
	7609	7332	0.000539	0.005021	0	0	0	0	0	0	1	-360	360;
	7609	4793	0.251054	1.054483	0	0	0	0	0	0	1	-360	360;
	8778	7609	0.00539	0.022409	0	381	0	0	0	0	1	-360	360;
	1949	853	0.18813	0.545888	0	0	0	0	0	0	1	-360	360;
	1949	1432	0.02382	0.200979	0	0	0	0	0	0	1	-360	360;
	1949	1432	0.00375	0.019289	0	476	0	0	0	0	1	-360	360;
	1949	2614	0.003151	0.01874	0	0	0	0	0	0	1	-360	360;
	1949	822	0.02419	0.20719	0	0	0	0	0	0	1	-360	360;
	4754	5958	0.005479	0.022531	0	553	0	0	0	0	1	-360	360;
	4754	6172	0.008581	0.133581	0	0	0	0	0	0	1	-360	360;
	5806	3854	0.000849	0.00436	0	0	0	0	0	0	1	-360	360;
	3784	5169	0.0004	0.00418	0	1843	0	0	0	0	1	-360	360;
	3784	330	0.00023	0.00253	0	1909	0	0	0	0	1	-360	360;
	5492	2712	0.000659	0.006909	0	0	0	0	0	0	1	-360	360;
	3784	2409	0.14356	0.922161	0	757	0	0	0	0	1	-360	360;
	3784	4578	0.05061	0.949861	0	757	0	0	0	0	1	-360	360;
	2472	8479	0.001669	0.19101	0	0	0	0	0	0	1	-360	360;
	3886	2472	0.008901	0.19594	0	0	0	0	0	0	1	-360	360;
	3886	8479	0.008849	0.207376	0	0	0	0	0	0	1	-360	360;
	584	2260	0.04068	0.314813	0	757	0	0	0	0	1	-360	360;
	584	873	0.001	0.358151	0	757	0	0	0	0	1	-360	360;
	584	873	0.00029	0.00339	0	1481	0	0	0	0	1	-360	360;
	584	6343	0.001	0.381247	0	757	0	0	0	0	1	-360	360;
	3655	1323	0	0.0002	0	757	0	0	0	0	1	-360	360;
	3655	1323	0.001	0.261032	0	757	0	0	0	0	1	-360	360;
	3905	2387	5e-05	0.000409	0	0	0	0	0	0	1	-360	360;
	7245	3905	0.00036	0.00411	0	553	0	0	0	0	1	-360	360;
	4522	1707	0	0.0002	0	0	0	0	0	0	1	-360	360;
	4522	1707	0.06963	0.647459	0	0	0	0	0	0	1	-360	360;
	540	4522	0.00138	0.00794	0	553	0	0	0	0	1	-360	360;
	683	4522	0.04905	0.618285	0	0	0	0	0	0	1	-360	360;
	7360	4522	0.01481	0.65845	0	0	0	0	0	0	1	-360	360;
	4557	4522	0.057229	0.834525	0	0	0	0	0	0	1	-360	360;
	83	1707	0.00276	0.034271	0	0	0	0	0	0	1	-360	360;
	683	1707	0.045901	0.629938	0	0	0	0	0	0	1	-360	360;
	7360	1707	0.05236	0.804298	0	0	0	0	0	0	1	-360	360;
	4557	1707	0.010979	0.705971	0	0	0	0	0	0	1	-360	360;
	3655	2260	0.00111	0.00985	0	1942	0	0	0	0	1	-360	360;
	3583	1823	12.66967	25.61427	0	757	0	0	0	0	1	-360	360;
	1823	6864	0.7509	2.64626	0	757	0	0	0	0	1	-360	360;
	8134	8123	0.001	0.522603	0	0	0	0	0	0	1	-360	360;
	8134	8123	0	0.0002	0	0	0	0	0	0	1	-360	360;
	8451	8134	1.015517	3.518182	0	0	0	0	0	0	1	-360	360;
	3854	8134	0.00274	0.01674	0	610	0	0	0	0	1	-360	360;
	8982	8134	0.001031	0.006669	0	0	0	0	0	0	1	-360	360;
	8451	8123	0.827397	3.941942	0	0	0	0	0	0	1	-360	360;
	3636	1823	0.00018	0.00147	0	0	0	0	0	0	1	-360	360;
	5315	1823	0.00025	0.00253	0	757	0	0	0	0	1	-360	360;
	3789	1823	0.00045	0.00372	0	1613	0	0	0	0	1	-360	360;
	7874	1823	0.00019	0.00142	0	1744	0	0	0	0	1	-360	360;
	6097	1823	0.00017	0.00161	0	757	0	0	0	0	1	-360	360;
	4622	1773	0.00031	0.00387	0	0	0	0	0	0	1	-360	360;
	6336	7401	0.00057	0.00261	0	0	0	0	0	0	1	-360	360;
	6336	1901	0.000591	0.002591	0	0	0	0	0	0	1	-360	360;
	2665	683	0.001599	0.00775	0	0	0	0	0	0	1	-360	360;
	540	683	0.014031	0.614215	0	0	0	0	0	0	1	-360	360;
	540	3854	0.012341	0.556384	0	0	0	0	0	0	1	-360	360;
	540	3854	0.001599	0.00738	0	0	0	0	0	0	1	-360	360;
	9080	1765	0.077687	1.036288	0	757	0	0	0	0	1	-360	360;
	9080	624	0.00057	0.00542	0	0	0	0	0	0	1	-360	360;
	9080	624	0.001	0.513781	0	757	0	0	0	0	1	-360	360;
	9080	1485	0.01036	0.396828	0	757	0	0	0	0	1	-360	360;
	9080	1485	0.00067	0.00594	0	0	0	0	0	0	1	-360	360;
	9080	4718	0.897853	2.161981	0	757	0	0	0	0	1	-360	360;
	1485	4276	0.0015	0.01442	0	1514	0	0	0	0	1	-360	360;
	4718	4276	0.00139	0.01387	0	1843	0	0	0	0	1	-360	360;
	624	1765	0.001	0.795637	0	757	0	0	0	0	1	-360	360;
	1485	1765	0.230298	1.27597	0	757	0	0	0	0	1	-360	360;
	4718	1765	0.05565	0.673331	0	757	0	0	0	0	1	-360	360;
	9154	18	0.00019	0.00214	0	1744	0	0	0	0	1	-360	360;
	9154	18	0.00019	0.00198	0	1744	0	0	0	0	1	-360	360;
	5528	4786	0.07358	0.966343	0	757	0	0	0	0	1	-360	360;
	6027	5528	0.06384	1.43054	0	757	0	0	0	0	1	-360	360;
	4294	5528	0.076281	1.166136	0	757	0	0	0	0	1	-360	360;
	6027	4786	0.210457	1.788158	0	757	0	0	0	0	1	-360	360;
	4294	4786	0.185533	1.718213	0	757	0	0	0	0	1	-360	360;
	5528	4786	2e-05	0.00024	0	1547	0	0	0	0	1	-360	360;
	330	5116	0.01408	0.37133	0	757	0	0	0	0	1	-360	360;
	7781	5116	0.01647	0.363712	0	757	0	0	0	0	1	-360	360;
	7781	5116	0.00068	0.00621	0	1810	0	0	0	0	1	-360	360;
	5116	873	0.00079	0.00807	0	1810	0	0	0	0	1	-360	360;
	2429	5754	0.00052	0.00514	0	1843	0	0	0	0	1	-360	360;
	873	2260	0.111198	0.57367	0	757	0	0	0	0	1	-360	360;
	6343	2260	0.126101	0.495623	0	757	0	0	0	0	1	-360	360;
	4268	1497	0.0032	0.016281	0	476	0	0	0	0	1	-360	360;
	4268	1497	0.003659	0.01407	0	514	0	0	0	0	1	-360	360;
	4268	1773	0.002281	0.010039	0	0	0	0	0	0	1	-360	360;
	4268	1773	0.001919	0.011669	0	0	0	0	0	0	1	-360	360;
	7332	4793	0.00236	0.02263	0	0	0	0	0	0	1	-360	360;
	3155	7332	0.009849	0.054531	0	0	0	0	0	0	1	-360	360;
	1432	853	0.02949	0.13405	0	0	0	0	0	0	1	-360	360;
	1432	853	0.005021	0.030341	0	0	0	0	0	0	1	-360	360;
	3583	7849	0.00076	0.00675	0	0	0	0	0	0	1	-360	360;
	3583	6864	0.568047	3.572576	0	757	0	0	0	0	1	-360	360;
	635	4425	0.003419	0.014531	0	0	0	0	0	0	1	-360	360;
	2029	635	0.002979	0.016461	0	0	0	0	0	0	1	-360	360;
	6336	3261	0.09382	0.44438	0	0	0	0	0	0	1	-360	360;
	7655	3261	0.003789	0.02306	0	381	0	0	0	0	1	-360	360;
	2726	3789	0.00014	0.00108	0	1810	0	0	0	0	1	-360	360;
	2726	9154	0.00015	0.00119	0	1810	0	0	0	0	1	-360	360;
	1955	9154	0.00031	0.00232	0	0	0	0	0	0	1	-360	360;
	1792	1955	0.00025	0.00193	0	1415	0	0	0	0	1	-360	360;
	9154	7942	0.137791	1.05367	0	757	0	0	0	0	1	-360	360;
	9154	7399	0.00054	0.00395	0	1876	0	0	0	0	1	-360	360;
	5530	2712	0.00113	0.012581	0	0	0	0	0	0	1	-360	360;
	5530	3155	0.00238	0.02086	0	0	0	0	0	0	1	-360	360;
	5530	6981	0.127331	1.369545	0	0	0	0	0	0	1	-360	360;
	5530	2223	0.699669	2.452893	0	0	0	0	0	0	1	-360	360;
	9023	6603	0.00331	0.025159	0	533	0	0	0	0	1	-360	360;
	624	1485	0.01975	0.825554	0	757	0	0	0	0	1	-360	360;
	4718	624	0.507639	2.270152	0	757	0	0	0	0	1	-360	360;
	5958	6603	0.20336	0.874897	0	0	0	0	0	0	1	-360	360;
	5958	6172	0.00732	0.030659	0	0	0	0	0	0	1	-360	360;
	1690	5958	0.075401	0.292748	0	0	0	0	0	0	1	-360	360;
	1690	5958	0.00314	0.02136	0	0	0	0	0	0	1	-360	360;
	9160	6977	0.00044	0.00325	0	1876	0	0	0	0	1	-360	360;
	55	2413	0.00076	0.00364	0	0	0	0	0	0	1	-360	360;
	55	5965	0.00075	0.00474	0	0	0	0	0	0	1	-360	360;
	6336	3479	3.557851	5.439876	0	0	0	0	0	0	1	-360	360;
	6336	5232	4.011157	6.327273	0	0	0	0	0	0	1	-360	360;
	6336	8699	0.006	0.327025	0	0	0	0	0	0	1	-360	360;
	3777	6336	3.78905	6.419628	0	0	0	0	0	0	1	-360	360;
	5261	6336	3.833471	5.760331	0	0	0	0	0	0	1	-360	360;
	3886	6336	0.000841	0.00362	0	0	0	0	0	0	1	-360	360;
	6336	2003	0.03	0.279855	0	0	0	0	0	0	1	-360	360;
	6336	2003	0.00107	0.004849	0	0	0	0	0	0	1	-360	360;
	6336	7620	3.092975	5.176033	0	0	0	0	0	0	1	-360	360;
	6336	7620	0.02688	0.291116	0	0	0	0	0	0	1	-360	360;
	6336	7620	0.00188	0.0132	0	553	0	0	0	0	1	-360	360;
	6541	1432	0.001469	0.00724	0	0	0	0	0	0	1	-360	360;
	6541	1432	0.0015	0.007081	0	0	0	0	0	0	1	-360	360;
	8033	5429	0.0017	0.00806	0	591	0	0	0	0	1	-360	360;
	2029	2606	0.00369	0.016091	0	533	0	0	0	0	1	-360	360;
	8516	7085	0.00301	0.0177	0	591	0	0	0	0	1	-360	360;
	6164	4758	5e-05	0.00053	0	1415	0	0	0	0	1	-360	360;
	5169	3922	6e-05	0.0005	0	1481	0	0	0	0	1	-360	360;
	5169	5749	4e-05	0.00047	0	0	0	0	0	0	1	-360	360;
	8699	166	0.00174	0.016521	0	400	0	0	0	0	1	-360	360;
	1576	5556	0.00019	0.00196	0	1810	0	0	0	0	1	-360	360;
	2429	6419	0.00074	0.00584	0	0	0	0	0	0	1	-360	360;
	1576	1720	0.00019	0.00183	0	0	0	0	0	0	1	-360	360;
	6164	2595	0.00018	0.00186	0	0	0	0	0	0	1	-360	360;
	6027	4294	0	0.0002	0	757	0	0	0	0	1	-360	360;
	6027	4294	0.001	0.413968	0	757	0	0	0	0	1	-360	360;
	6027	5528	0.00215	0.02177	0	1514	0	0	0	0	1	-360	360;
	6997	822	0.008659	0.0422	0	381	0	0	0	0	1	-360	360;
	79	6983	0.00029	0.00252	0	1415	0	0	0	0	1	-360	360;
	79	4500	0.0343	0.612763	0	757	0	0	0	0	1	-360	360;
	2610	8552	0.00036	0.00195	0	0	0	0	0	0	1	-360	360;
	2610	8552	0.000409	0.00213	0	0	0	0	0	0	1	-360	360;
	4941	2787	0.00118	0.008719	0	476	0	0	0	0	1	-360	360;
	3777	4941	0.0053	0.023409	0	0	0	0	0	0	1	-360	360;
	5261	4941	0.005401	0.022909	0	610	0	0	0	0	1	-360	360;
	5965	4941	0.000901	0.00412	0	610	0	0	0	0	1	-360	360;
	4941	2413	0.05139	0.435785	0	0	0	0	0	0	1	-360	360;
	3479	5232	0.01661	0.339194	0	0	0	0	0	0	1	-360	360;
	3479	8699	2.441736	4.517975	0	0	0	0	0	0	1	-360	360;
	3777	3479	0.03336	0.374174	0	0	0	0	0	0	1	-360	360;
	5261	3479	0.033219	0.37814	0	0	0	0	0	0	1	-360	360;
	3479	4496	0.003151	0.017599	0	0	0	0	0	0	1	-360	360;
	7620	3479	0.02231	0.393079	0	0	0	0	0	0	1	-360	360;
	8759	3636	0.0005	0.00519	0	0	0	0	0	0	1	-360	360;
	8759	8585	0.00024	0.00227	0	1777	0	0	0	0	1	-360	360;
	8759	6983	0.00049	0.00529	0	1448	0	0	0	0	1	-360	360;
	8759	5403	0.00036	0.00305	0	1448	0	0	0	0	1	-360	360;
	3859	5673	0.00264	0.014781	0	0	0	0	0	0	1	-360	360;
	6902	5673	0.002539	0.01357	0	0	0	0	0	0	1	-360	360;
	2223	5673	0.0053	0.021599	0	553	0	0	0	0	1	-360	360;
	6652	5673	0.003091	0.01568	0	0	0	0	0	0	1	-360	360;
	9208	5673	0.00274	0.0127	0	0	0	0	0	0	1	-360	360;
	2223	5673	0.04224	0.3325	0	0	0	0	0	0	1	-360	360;
	5914	5491	0.00086	0.00873	0	1777	0	0	0	0	1	-360	360;
	9160	5491	0.00206	0.01561	0	1678	0	0	0	0	1	-360	360;
	3668	9143	0.00032	0.00271	0	1514	0	0	0	0	1	-360	360;
	7781	3668	0.00051	0.00524	0	1744	0	0	0	0	1	-360	360;
	250	7725	0.00032	0.00265	0	1810	0	0	0	0	1	-360	360;
	330	250	0.0005	0.00534	0	1645	0	0	0	0	1	-360	360;
	7725	9143	0	0.0002	0	1580	0	0	0	0	1	-360	360;
	7725	9143	0.163629	1.021814	0	757	0	0	0	0	1	-360	360;
	680	7725	0.001	0.350312	0	757	0	0	0	0	1	-360	360;
	7715	7725	0.0571	0.699515	0	757	0	0	0	0	1	-360	360;
	3753	7725	0.00022	0.00214	0	1415	0	0	0	0	1	-360	360;
	5387	7725	0.05584	0.597071	0	757	0	0	0	0	1	-360	360;
	4578	9143	0.00044	0.00329	0	1942	0	0	0	0	1	-360	360;
	680	9143	0.0684	0.765166	0	757	0	0	0	0	1	-360	360;
	680	9143	0.0005	0.00377	0	1744	0	0	0	0	1	-360	360;
	7715	9143	0.02916	0.635783	0	757	0	0	0	0	1	-360	360;
	5444	9143	0.00021	0.00218	0	1744	0	0	0	0	1	-360	360;
	5387	9143	0.076648	0.893421	0	757	0	0	0	0	1	-360	360;
	6830	7859	0.003581	0.01713	0	0	0	0	0	0	1	-360	360;
	388	122	0.003651	0.01676	0	0	0	0	0	0	1	-360	360;
	6524	5969	0	0.0002	0	889	0	0	0	0	1	-360	360;
	6524	5969	0.001	0.363179	0	757	0	0	0	0	1	-360	360;
	6524	2429	0.03201	0.610942	0	757	0	0	0	0	1	-360	360;
	5969	2429	0.03493	0.703809	0	757	0	0	0	0	1	-360	360;
	5969	3160	0.00038	0.00378	0	1711	0	0	0	0	1	-360	360;
	1690	6603	0.662562	1.966736	0	0	0	0	0	0	1	-360	360;
	1690	6603	0.007461	0.03539	0	553	0	0	0	0	1	-360	360;
	4718	1485	1.057548	2.384834	0	757	0	0	0	0	1	-360	360;
	2029	7859	0.00313	0.015591	0	0	0	0	0	0	1	-360	360;
	2029	122	0.00325	0.01495	0	0	0	0	0	0	1	-360	360;
	2712	2029	0.174211	0.885269	0	0	0	0	0	0	1	-360	360;
	6981	2029	0.05432	0.507293	0	0	0	0	0	0	1	-360	360;
	3176	2029	0.315744	1.143306	0	0	0	0	0	0	1	-360	360;
	7655	8516	0.106531	0.728988	0	0	0	0	0	0	1	-360	360;
	5169	2429	0.00046	0.00355	0	1909	0	0	0	0	1	-360	360;
	6164	2429	0.00046	0.00356	0	1481	0	0	0	0	1	-360	360;
	2787	5232	0.00056	0.00351	0	457	0	0	0	0	1	-360	360;
	7620	2787	0.001271	0.00699	0	476	0	0	0	0	1	-360	360;
	5232	8699	2.229752	5.528926	0	0	0	0	0	0	1	-360	360;
	3777	5232	0.186169	1.134959	0	0	0	0	0	0	1	-360	360;
	5261	5232	0.191419	0.99374	0	0	0	0	0	0	1	-360	360;
	7620	5232	0.001	0.342025	0	0	0	0	0	0	1	-360	360;
	9221	6083	8e-05	0.00071	0	1843	0	0	0	0	1	-360	360;
	1576	6083	0.00047	0.00487	0	1283	0	0	0	0	1	-360	360;
	3777	8699	2.630992	4.595455	0	0	0	0	0	0	1	-360	360;
	5261	8699	2.602893	4.194835	0	0	0	0	0	0	1	-360	360;
	3886	8699	0.002849	0.0122	0	400	0	0	0	0	1	-360	360;
	7620	8699	1.630455	4.689463	0	0	0	0	0	0	1	-360	360;
	3160	6083	0.136641	1.051662	0	757	0	0	0	0	1	-360	360;
	673	6083	0.00035	0.00282	0	1514	0	0	0	0	1	-360	360;
	5914	9160	0.00092	0.0092	0	1810	0	0	0	0	1	-360	360;
	2409	4578	0	0.0002	0	1020	0	0	0	0	1	-360	360;
	2409	4578	0.02074	0.459238	0	757	0	0	0	0	1	-360	360;
	2409	1576	0.00025	0.00224	0	1876	0	0	0	0	1	-360	360;
	7655	8552	0.000651	0.0037	0	0	0	0	0	0	1	-360	360;
	3777	5261	0.001	0.238182	0	0	0	0	0	0	1	-360	360;
	3777	7620	0.168469	0.943719	0	0	0	0	0	0	1	-360	360;
	5261	7620	0.1577	0.919711	0	0	0	0	0	0	1	-360	360;
	6452	4793	0.0033	0.023531	0	0	0	0	0	0	1	-360	360;
	9225	4793	0.00336	0.023031	0	553	0	0	0	0	1	-360	360;
	8798	4793	0.050961	0.353843	0	0	0	0	0	0	1	-360	360;
	1657	4793	0.07906	0.386591	0	0	0	0	0	0	1	-360	360;
	1657	4793	0.00224	0.02005	0	0	0	0	0	0	1	-360	360;
	8778	4793	0.001979	0.018841	0	0	0	0	0	0	1	-360	360;
	3321	25	0.0003	0.002669	0	533	0	0	0	0	1	-360	360;
	1823	1792	0.00017	0.00166	0	1415	0	0	0	0	1	-360	360;
	7874	4500	0.00057	0.00619	0	1744	0	0	0	0	1	-360	360;
	6097	8713	8e-05	0.00082	0	757	0	0	0	0	1	-360	360;
	683	3854	0.001	0.452087	0	0	0	0	0	0	1	-360	360;
	683	7360	0.028289	0.776901	0	0	0	0	0	0	1	-360	360;
	683	4557	0.027469	0.75064	0	0	0	0	0	0	1	-360	360;
	3854	2212	0.001409	0.00714	0	476	0	0	0	0	1	-360	360;
	5324	6652	0.00776	0.04488	0	0	0	0	0	0	1	-360	360;
	673	2486	0.00028	0.0023	0	1448	0	0	0	0	1	-360	360;
	3225	673	0.00015	0.00155	0	1580	0	0	0	0	1	-360	360;
	4557	525	0.000229	0.00113	0	0	0	0	0	0	1	-360	360;
	7360	3357	0.00024	0.001099	0	514	0	0	0	0	1	-360	360;
	6452	8798	0.000651	0.005031	0	0	0	0	0	0	1	-360	360;
	9225	8798	0.000581	0.0057	0	553	0	0	0	0	1	-360	360;
	1657	8798	0.395351	1.062603	0	0	0	0	0	0	1	-360	360;
	7781	1576	0.00073	0.00574	0	1942	0	0	0	0	1	-360	360;
	5711	6164	0.00074	0.00562	0	0	0	0	0	0	1	-360	360;
	798	4500	0.00021	0.00217	0	1580	0	0	0	0	1	-360	360;
	8639	5580	0.001	0.253636	0	0	0	0	0	0	1	-360	360;
	7803	3225	0.00078	0.00691	0	1645	0	0	0	0	1	-360	360;
	5711	6343	0.00135	0.01007	0	0	0	0	0	0	1	-360	360;
	2712	6981	0.048271	0.517335	0	0	0	0	0	0	1	-360	360;
	2712	6981	0.001901	0.01187	0	0	0	0	0	0	1	-360	360;
	2712	3176	0.02499	0.413099	0	0	0	0	0	0	1	-360	360;
	2712	3176	0.00186	0.01099	0	533	0	0	0	0	1	-360	360;
	330	7781	0.00351	0.296032	0	757	0	0	0	0	1	-360	360;
	2390	9095	0.000169	0.000789	0	0	0	0	0	0	1	-360	360;
	2413	2390	0.000211	0.001031	0	0	0	0	0	0	1	-360	360;
	3542	6661	0.00038	0.001909	0	0	0	0	0	0	1	-360	360;
	2413	3542	0.000219	0.001	0	0	0	0	0	0	1	-360	360;
	680	7715	0.0114	0.491198	0	757	0	0	0	0	1	-360	360;
	680	5387	0.0111	0.44741	0	757	0	0	0	0	1	-360	360;
	6981	3176	0.065849	0.525558	0	0	0	0	0	0	1	-360	360;
	6981	8775	0.00095	0.0053	0	514	0	0	0	0	1	-360	360;
	5965	2413	0.00139	0.008281	0	0	0	0	0	0	1	-360	360;
	6886	7161	0.001	0.26688	0	0	0	0	0	0	1	-360	360;
	5324	2033	0.00036	0.00286	0	0	0	0	0	0	1	-360	360;
	7620	2003	0.02231	0.26281	0	0	0	0	0	0	1	-360	360;
	7620	2003	0.001669	0.01064	0	572	0	0	0	0	1	-360	360;
	5528	7880	3e-05	0.00031	0	0	0	0	0	0	1	-360	360;
	5528	433	3e-05	0.0003	0	1942	0	0	0	0	1	-360	360;
	7942	7399	0.0004	0.00637	0	1645	0	0	0	0	1	-360	360;
	7803	5403	0.00018	0.0019	0	1711	0	0	0	0	1	-360	360;
	7803	8585	0.00032	0.00267	0	1711	0	0	0	0	1	-360	360;
	3753	7715	0.00023	0.00202	0	1613	0	0	0	0	1	-360	360;
	7715	5387	0.001	0.382043	0	757	0	0	0	0	1	-360	360;
	7878	9208	0.00011	0.00075	0	533	0	0	0	0	1	-360	360;
	8810	9208	0.00013	0.000531	0	0	0	0	0	0	1	-360	360;
	7360	4557	0.010909	0.62314	0	0	0	0	0	0	1	-360	360;
	1497	1773	0.005969	0.037081	0	0	0	0	0	0	1	-360	360;
	4501	9088	0	0.0002	0	1514	0	0	0	0	1	-360	360;
	4501	9088	0.001	0.39696	0	757	0	0	0	0	1	-360	360;
	4501	8388	0.192888	1.242729	0	757	0	0	0	0	1	-360	360;
	9088	1630	0.00078	0.00629	0	1876	0	0	0	0	1	-360	360;
	9088	8388	0.215693	1.63338	0	757	0	0	0	0	1	-360	360;
	1630	8388	0.00048	0.00397	0	0	0	0	0	0	1	-360	360;
	4117	8388	0.00049	0.00388	0	0	0	0	0	0	1	-360	360;
	7245	8698	0.000211	0.00113	0	0	0	0	0	0	1	-360	360;
	5444	5387	3e-05	0.00035	0	1316	0	0	0	0	1	-360	360;
	3155	2223	0.007409	0.042281	0	0	0	0	0	0	1	-360	360;
	3724	6541	0.00012	0.0007	0	0	0	0	0	0	1	-360	360;
	873	6343	0	0.0002	0	856	0	0	0	0	1	-360	360;
	873	6343	0.001	0.284017	0	757	0	0	0	0	1	-360	360;
	8779	8256	1.997645	4.178116	0	757	0	0	0	0	1	-360	360;
	8779	3110	0.00015	0.00152	0	1349	0	0	0	0	1	-360	360;
	8779	1287	0.00014	0.00141	0	1316	0	0	0	0	1	-360	360;
	8779	4732	0.00103	0.00914	0	1645	0	0	0	0	1	-360	360;
	8779	2932	0.00105	0.00896	0	1514	0	0	0	0	1	-360	360;
	2309	8779	0.334702	1.492936	0	757	0	0	0	0	1	-360	360;
	2309	8779	0.00142	0.01428	0	1711	0	0	0	0	1	-360	360;
	503	8779	0.340893	1.460526	0	757	0	0	0	0	1	-360	360;
	503	8779	0.00139	0.01455	0	0	0	0	0	0	1	-360	360;
	394	4662	0.0004	0.00376	0	0	0	0	0	0	1	-360	360;
	394	5316	0.00101	0.01129	0	1711	0	0	0	0	1	-360	360;
	394	4571	0.00076	0.0089	0	0	0	0	0	0	1	-360	360;
	394	4662	0.0004	0.00368	0	0	0	0	0	0	1	-360	360;
	394	2681	0.00099	0.01151	0	1810	0	0	0	0	1	-360	360;
	394	938	0.00075	0.00824	0	0	0	0	0	0	1	-360	360;
	7934	4670	0.00156	0.01568	0	1744	0	0	0	0	1	-360	360;
	7934	4670	0.00152	0.01599	0	1711	0	0	0	0	1	-360	360;
	484	7006	0.00539	0.01612	0	591	0	0	0	0	1	-360	360;
	484	7006	0.0048	0.018271	0	572	0	0	0	0	1	-360	360;
	8188	866	0.008729	0.034159	0	419	0	0	0	0	1	-360	360;
	8188	866	0.00855	0.03482	0	419	0	0	0	0	1	-360	360;
	4156	1900	0.00146	0.00651	0	1152	0	0	0	0	1	-360	360;
	511	1900	0.134252	0.771745	0	757	0	0	0	0	1	-360	360;
	511	1900	0.00072	0.00587	0	1810	0	0	0	0	1	-360	360;
	511	1900	0.00071	0.006	0	1876	0	0	0	0	1	-360	360;
	6374	1900	0.00013	0.00132	0	1810	0	0	0	0	1	-360	360;
	3396	1900	0.322853	1.362742	0	757	0	0	0	0	1	-360	360;
	2852	8944	0.161841	1.112975	0	0	0	0	0	0	1	-360	360;
	2852	8944	0.00356	0.022409	0	591	0	0	0	0	1	-360	360;
	439	8944	0.211508	1.021529	0	0	0	0	0	0	1	-360	360;
	439	8944	0.00838	0.038479	0	0	0	0	0	0	1	-360	360;
	484	8944	0.912748	1.922025	0	0	0	0	0	0	1	-360	360;
	5880	8198	0.005211	0.016901	0	457	0	0	0	0	1	-360	360;
	5880	8198	0.219566	0.646591	0	0	0	0	0	0	1	-360	360;
	5880	8198	0.00531	0.016539	0	476	0	0	0	0	1	-360	360;
	5234	7261	0.140519	0.391918	0	757	0	0	0	0	1	-360	360;
	1601	5234	0.00131	0.01484	0	1810	0	0	0	0	1	-360	360;
	5234	4571	0.05123	0.434093	0	757	0	0	0	0	1	-360	360;
	5234	4571	0.0006	0.00673	0	1876	0	0	0	0	1	-360	360;
	5234	8370	0.05191	0.445222	0	757	0	0	0	0	1	-360	360;
	5194	5234	0.736704	3.325	0	757	0	0	0	0	1	-360	360;
	5054	5234	2.622299	8.981301	0	757	0	0	0	0	1	-360	360;
	7680	6466	0	0.0002	0	757	0	0	0	0	1	-360	360;
	7680	6466	0.001	0.291627	0	757	0	0	0	0	1	-360	360;
	2463	7680	0.00016	0.00138	0	1415	0	0	0	0	1	-360	360;
	4285	7680	0.00017	0.00135	0	1448	0	0	0	0	1	-360	360;
	478	6466	0.00017	0.00133	0	1251	0	0	0	0	1	-360	360;
	7919	6466	0.00017	0.0013	0	1316	0	0	0	0	1	-360	360;
	7934	2463	0.00189	0.01446	0	1316	0	0	0	0	1	-360	360;
	4285	983	0.00074	0.00698	0	1382	0	0	0	0	1	-360	360;
	7934	478	0.00193	0.01416	0	1349	0	0	0	0	1	-360	360;
	7919	4175	0.00073	0.00711	0	1349	0	0	0	0	1	-360	360;
	2309	8256	0.136731	0.910873	0	757	0	0	0	0	1	-360	360;
	2309	8256	0.00128	0.0134	0	1876	0	0	0	0	1	-360	360;
	503	8256	0.13394	0.92874	0	757	0	0	0	0	1	-360	360;
	5986	2192	0.271641	0.707964	0	757	0	0	0	0	1	-360	360;
	5986	2192	0.00037	0.0037	0	1645	0	0	0	0	1	-360	360;
	5986	2192	0.00038	0.00363	0	1876	0	0	0	0	1	-360	360;
	5194	2192	0.00156	0.01473	0	1810	0	0	0	0	1	-360	360;
	5054	2192	0.00159	0.01442	0	1678	0	0	0	0	1	-360	360;
	8300	4218	0.01049	0.04336	0	438	0	0	0	0	1	-360	360;
	1863	4218	0.010271	0.04418	0	457	0	0	0	0	1	-360	360;
	5880	4218	0.00881	0.037021	0	457	0	0	0	0	1	-360	360;
	2418	4218	0.004409	0.01682	0	476	0	0	0	0	1	-360	360;
	2418	4218	0.00432	0.017151	0	533	0	0	0	0	1	-360	360;
	7514	4218	0.01025	0.042341	0	553	0	0	0	0	1	-360	360;
	3759	2736	0.05386	0.203789	0	0	0	0	0	0	1	-360	360;
	3759	2736	0.005	0.017849	0	0	0	0	0	0	1	-360	360;
	3759	2736	0.004909	0.01824	0	0	0	0	0	0	1	-360	360;
	2736	3189	0.442645	0.98938	0	0	0	0	0	0	1	-360	360;
	2736	3189	0.00789	0.025831	0	514	0	0	0	0	1	-360	360;
	2736	3189	0.00774	0.02638	0	476	0	0	0	0	1	-360	360;
	6458	511	0.00141	0.01209	0	1744	0	0	0	0	1	-360	360;
	6458	8415	0.104079	0.42973	0	757	0	0	0	0	1	-360	360;
	983	4662	0.00139	0.01016	0	1349	0	0	0	0	1	-360	360;
	4175	4662	0.00142	0.00994	0	1415	0	0	0	0	1	-360	360;
	2190	386	9e-05	0.00064	0	757	0	0	0	0	1	-360	360;
	2190	4732	0.195942	0.820914	0	757	0	0	0	0	1	-360	360;
	2190	2932	0.199778	0.804155	0	757	0	0	0	0	1	-360	360;
	503	2190	0.02836	0.37876	0	757	0	0	0	0	1	-360	360;
	386	4732	0.00109	0.00814	0	0	0	0	0	0	1	-360	360;
	503	386	0.00054	0.00424	0	0	0	0	0	0	1	-360	360;
	3127	511	0.00263	0.02613	0	1613	0	0	0	0	1	-360	360;
	511	6914	0.00166	0.01885	0	1547	0	0	0	0	1	-360	360;
	511	3396	0.847576	2.035042	0	757	0	0	0	0	1	-360	360;
	1243	511	0.00094	0.00794	0	1645	0	0	0	0	1	-360	360;
	2852	1403	0.014081	0.055831	0	457	0	0	0	0	1	-360	360;
	2852	439	2.006198	3.572314	0	0	0	0	0	0	1	-360	360;
	2852	439	0.0122	0.05982	0	0	0	0	0	0	1	-360	360;
	2852	484	0.30564	1.254566	0	0	0	0	0	0	1	-360	360;
	2852	484	0.0235	0.076229	0	553	0	0	0	0	1	-360	360;
	2852	3189	0.029969	0.12889	0	476	0	0	0	0	1	-360	360;
	2852	3189	0.02932	0.119341	0	495	0	0	0	0	1	-360	360;
	6621	4499	0.00699	0.02886	0	629	0	0	0	0	1	-360	360;
	4499	866	0.004909	0.014719	0	553	0	0	0	0	1	-360	360;
	8952	7514	0.005591	0.01814	0	457	0	0	0	0	1	-360	360;
	8952	7514	0.00569	0.01776	0	514	0	0	0	0	1	-360	360;
	7261	4571	0.03624	0.415741	0	757	0	0	0	0	1	-360	360;
	7261	8370	0.03359	0.482777	0	757	0	0	0	0	1	-360	360;
	7261	8370	0.00088	0.01001	0	1711	0	0	0	0	1	-360	360;
	7261	938	0.001	0.00981	0	0	0	0	0	0	1	-360	360;
	5194	7261	0.706579	3.541136	0	757	0	0	0	0	1	-360	360;
	5054	7261	2.102562	10.3982	0	757	0	0	0	0	1	-360	360;
	484	1403	0.00962	0.03813	0	533	0	0	0	0	1	-360	360;
	7676	5164	0.04235	0.255499	0	757	0	0	0	0	1	-360	360;
	7676	6993	0.03604	0.287562	0	757	0	0	0	0	1	-360	360;
	7676	6993	0.00069	0.00676	0	1942	0	0	0	0	1	-360	360;
	7676	4620	0.02303	0.242479	0	757	0	0	0	0	1	-360	360;
	4498	7676	0.002	0.02189	0	0	0	0	0	0	1	-360	360;
	5316	5054	0.00063	0.00569	0	1843	0	0	0	0	1	-360	360;
	2681	5194	0.00064	0.00557	0	1909	0	0	0	0	1	-360	360;
	5164	6993	0.001	0.193123	0	757	0	0	0	0	1	-360	360;
	5164	4620	0.01778	0.156988	0	757	0	0	0	0	1	-360	360;
	5164	4620	0.00104	0.00843	0	1613	0	0	0	0	1	-360	360;
	2309	5164	0.00096	0.01465	0	0	0	0	0	0	1	-360	360;
	4498	5164	0.003	0.02517	0	1942	0	0	0	0	1	-360	360;
	6993	4620	0.462929	1.374931	0	757	0	0	0	0	1	-360	360;
	107	6993	0.00244	0.01729	0	0	0	0	0	0	1	-360	360;
	503	6993	0.00094	0.01356	0	0	0	0	0	0	1	-360	360;
	8300	6794	0.01526	0.05118	0	400	0	0	0	0	1	-360	360;
	1863	6794	0.01555	0.05011	0	419	0	0	0	0	1	-360	360;
	439	484	0.26188	0.583802	0	0	0	0	0	0	1	-360	360;
	3127	7934	0.0036	0.04	0	1547	0	0	0	0	1	-360	360;
	3127	495	0.00363	0.03446	0	1645	0	0	0	0	1	-360	360;
	3127	3396	0.00099	0.01049	0	1218	0	0	0	0	1	-360	360;
	8966	4620	0.00332	0.03036	0	1645	0	0	0	0	1	-360	360;
	8716	1426	0.00027	0.00309	0	1744	0	0	0	0	1	-360	360;
	5258	1426	0.00031	0.00303	0	1415	0	0	0	0	1	-360	360;
	3759	3189	0.027099	0.21312	0	0	0	0	0	0	1	-360	360;
	3759	3189	0.00357	0.011021	0	553	0	0	0	0	1	-360	360;
	3759	3189	0.00351	0.01126	0	495	0	0	0	0	1	-360	360;
	7934	495	0.00037	0.00418	0	1251	0	0	0	0	1	-360	360;
	4122	6621	0.04124	0.192349	0	0	0	0	0	0	1	-360	360;
	4122	6621	0.015669	0.064711	0	629	0	0	0	0	1	-360	360;
	4122	6621	0.016	0.06349	0	629	0	0	0	0	1	-360	360;
	6621	866	0.012281	0.03676	0	610	0	0	0	0	1	-360	360;
	5880	193	0.02043	0.19226	0	0	0	0	0	0	1	-360	360;
	5880	193	0.00287	0.01095	0	495	0	0	0	0	1	-360	360;
	5880	193	0.00281	0.011169	0	495	0	0	0	0	1	-360	360;
	8415	6914	0.00054	0.00596	0	1580	0	0	0	0	1	-360	360;
	6900	8415	0.00134	0.01244	0	1876	0	0	0	0	1	-360	360;
	983	4175	0.001	0.265693	0	757	0	0	0	0	1	-360	360;
	8966	983	0.15358	0.524508	0	757	0	0	0	0	1	-360	360;
	7169	983	0.613823	2.118352	0	757	0	0	0	0	1	-360	360;
	7169	983	0.00081	0.0077	0	1349	0	0	0	0	1	-360	360;
	983	5194	0.00228	0.02083	0	1744	0	0	0	0	1	-360	360;
	8966	4175	0.168047	0.592382	0	757	0	0	0	0	1	-360	360;
	8966	4175	0.0008	0.00785	0	1547	0	0	0	0	1	-360	360;
	7169	4175	0.698199	2.371468	0	757	0	0	0	0	1	-360	360;
	4175	5054	0.00232	0.02039	0	1876	0	0	0	0	1	-360	360;
	2309	5986	0.00177	0.01939	0	0	0	0	0	0	1	-360	360;
	503	5986	0.00173	0.01976	0	0	0	0	0	0	1	-360	360;
	8966	7169	0	0.0002	0	1613	0	0	0	0	1	-360	360;
	8966	7169	0.001	0.213047	0	757	0	0	0	0	1	-360	360;
	8966	5054	0.00194	0.01469	0	1547	0	0	0	0	1	-360	360;
	107	8966	0.00123	0.00838	0	1514	0	0	0	0	1	-360	360;
	8966	3057	0.00121	0.0116	0	1678	0	0	0	0	1	-360	360;
	8966	4498	0.00346	0.03846	0	1711	0	0	0	0	1	-360	360;
	7169	5194	0.00198	0.01438	0	1678	0	0	0	0	1	-360	360;
	7169	4498	0.00338	0.0392	0	1678	0	0	0	0	1	-360	360;
	7169	197	0.00248	0.02446	0	889	0	0	0	0	1	-360	360;
	4732	2932	0.001	0.26367	0	757	0	0	0	0	1	-360	360;
	503	4732	0.580512	1.382895	0	757	0	0	0	0	1	-360	360;
	6457	4732	0.00094	0.00807	0	0	0	0	0	0	1	-360	360;
	1601	2932	0.00232	0.01638	0	0	0	0	0	0	1	-360	360;
	2309	2932	0.00217	0.0131	0	1876	0	0	0	0	1	-360	360;
	503	2932	0.568421	1.409488	0	757	0	0	0	0	1	-360	360;
	1757	5584	0.005331	0.017341	0	0	0	0	0	0	1	-360	360;
	1601	8370	0.00199	0.02218	0	1744	0	0	0	0	1	-360	360;
	1601	5194	0.00163	0.01513	0	0	0	0	0	0	1	-360	360;
	1601	5054	0.00159	0.01541	0	0	0	0	0	0	1	-360	360;
	1601	6457	0.00136	0.00959	0	0	0	0	0	0	1	-360	360;
	1601	8716	0.00042	0.00496	0	1580	0	0	0	0	1	-360	360;
	1601	5258	0.00041	0.00505	0	954	0	0	0	0	1	-360	360;
	5880	7514	0.016591	0.06199	0	0	0	0	0	0	1	-360	360;
	4571	8370	0	0.0002	0	0	0	0	0	0	1	-360	360;
	4571	8370	0.001	0.404868	0	757	0	0	0	0	1	-360	360;
	5194	4571	0.130547	2.317729	0	757	0	0	0	0	1	-360	360;
	5054	4571	0.431627	6.284488	0	757	0	0	0	0	1	-360	360;
	5194	8370	0.127403	2.421884	0	757	0	0	0	0	1	-360	360;
	5054	8370	0.397583	6.979917	0	757	0	0	0	0	1	-360	360;
	3396	6900	0.00356	0.03374	0	1645	0	0	0	0	1	-360	360;
	5194	5054	0.00164	0.14108	0	757	0	0	0	0	1	-360	360;
	4498	6900	0.00058	0.00658	0	1909	0	0	0	0	1	-360	360;
	4498	6900	0.00065	0.00645	0	1909	0	0	0	0	1	-360	360;
	3057	197	0.079079	0.405817	0	757	0	0	0	0	1	-360	360;
	3057	197	0.00139	0.01268	0	1481	0	0	0	0	1	-360	360;
	2309	503	0	0.0002	0	987	0	0	0	0	1	-360	360;
	2309	503	0.001	0.270111	0	757	0	0	0	0	1	-360	360;
	8414	6794	0.185091	0.839215	0	0	0	0	0	0	1	-360	360;
	8414	6794	0.004651	0.039669	0	495	0	0	0	0	1	-360	360;
	8414	6794	0.00476	0.042841	0	495	0	0	0	0	1	-360	360;
	4498	197	0.0018	0.02033	0	1185	0	0	0	0	1	-360	360;
	7799	6956	0.001	0.335748	0	757	0	0	0	0	1	-360	360;
	8716	7799	0.0001	0.00068	0	1580	0	0	0	0	1	-360	360;
	5258	6956	0.0001	0.00067	0	1580	0	0	0	0	1	-360	360;
	4901	9195	0.026107	0.07605	0	0	0	0	0	0	1	-360	360;
	8040	4901	0.019727	0.062124	0	0	0	0	0	0	1	-360	360;
	7007	9195	0.026603	0.074438	0	0	0	0	0	0	1	-360	360;
	8040	7007	0.020107	0.060851	0	0	0	0	0	0	1	-360	360;
	8143	6682	0.014017	0.040471	0	0	0	0	0	0	1	-360	360;
	6682	1395	0.012488	0.045289	0	0	0	0	0	0	1	-360	360;
	7210	5541	0.014281	0.039612	0	0	0	0	0	0	1	-360	360;
	5541	7315	0.010826	0.032686	0	0	0	0	0	0	1	-360	360;
	4462	6720	0.012554	0.049893	0	0	0	0	0	0	1	-360	360;
	6720	8118	0.00705	0.023174	0	0	0	0	0	0	1	-360	360;
	4384	4937	0.018818	0.054132	0	0	0	0	0	0	1	-360	360;
	1395	4937	0.032769	0.043661	0	0	0	0	0	0	1	-360	360;
	5702	4462	0.015322	0.049653	0	0	0	0	0	0	1	-360	360;
	2530	664	0.002413	0.006917	0	0	0	0	0	0	1	-360	360;
	664	5353	0.010537	0.025917	0	0	0	0	0	0	1	-360	360;
	4384	8143	0.009124	0.032314	0	0	0	0	0	0	1	-360	360;
	4384	7210	0.008934	0.032934	0	0	0	0	0	0	1	-360	360;
	9195	4074	0	0.000413	0	0	0	0	0	0	1	-360	360;
	8410	4074	0.015562	0.044669	0	0	0	0	0	0	1	-360	360;
	6585	4074	0.013	0.049983	0	0	0	0	0	0	1	-360	360;
	9195	5445	0.014347	0.041182	0	0	0	0	0	0	1	-360	360;
	9195	5445	0.014628	0.040322	0	0	0	0	0	0	1	-360	360;
	1715	690	0.017446	0.029752	0	0	0	0	0	0	1	-360	360;
	1715	6576	0.031579	0.070992	0	0	0	0	0	0	1	-360	360;
	3419	1715	0.007843	0.018793	0	0	0	0	0	0	1	-360	360;
	4403	1715	0.007992	0.018413	0	0	0	0	0	0	1	-360	360;
	5911	7692	0.001271	0.016781	0	553	0	0	0	0	1	-360	360;
	5911	7692	0.00124	0.015539	0	610	0	0	0	0	1	-360	360;
	690	4384	0.033777	0.06419	0	0	0	0	0	0	1	-360	360;
	4384	1180	0.014347	0.028397	0	0	0	0	0	0	1	-360	360;
	6576	4384	0.023331	0.052182	0	0	0	0	0	0	1	-360	360;
	4384	6925	0.009446	0.022636	0	0	0	0	0	0	1	-360	360;
	7315	1395	0.006992	0.025488	0	0	0	0	0	0	1	-360	360;
	6925	1180	0.003339	0.008851	0	0	0	0	0	0	1	-360	360;
	8410	4871	0.016471	0.06081	0	0	0	0	0	0	1	-360	360;
	7405	6585	0.020149	0.058727	0	0	0	0	0	0	1	-360	360;
	5445	612	0.013992	0.054	0	0	0	0	0	0	1	-360	360;
	5445	3988	0.013686	0.05	0	0	0	0	0	0	1	-360	360;
	5445	4515	0.000223	0.000826	0	0	0	0	0	0	1	-360	360;
	5445	3358	0.000248	0.00081	0	0	0	0	0	0	1	-360	360;
	9212	4871	0.036512	0.098843	0	0	0	0	0	0	1	-360	360;
	612	208	0.012876	0.037438	0	0	0	0	0	0	1	-360	360;
	3988	208	0.013124	0.036636	0	0	0	0	0	0	1	-360	360;
	2307	3419	0.013909	0.033041	0	0	0	0	0	0	1	-360	360;
	5353	4403	0.014182	0.032372	0	0	0	0	0	0	1	-360	360;
	5145	8324	0.00474	0.027521	0	0	0	0	0	0	1	-360	360;
	5145	6865	0.00538	0.026969	0	0	0	0	0	0	1	-360	360;
	4857	8118	0.020372	0.06276	0	0	0	0	0	0	1	-360	360;
	9212	7405	0	0.000413	0	0	0	0	0	0	1	-360	360;
	6159	5911	0.0053	0.02831	0	419	0	0	0	0	1	-360	360;
	6159	5911	0.00601	0.02774	0	438	0	0	0	0	1	-360	360;
	8324	6159	0.00113	0.006229	0	0	0	0	0	0	1	-360	360;
	6865	6159	0.001151	0.00611	0	0	0	0	0	0	1	-360	360;
	2530	2307	0.014364	0.029529	0	0	0	0	0	0	1	-360	360;
	4857	2530	0.009017	0.019727	0	0	0	0	0	0	1	-360	360;
	4857	2530	0.00919	0.019314	0	0	0	0	0	0	1	-360	360;
	6865	4569	9.1e-05	0.000219	0	0	0	0	0	0	1	-360	360;
	8324	4569	9.1e-05	0.000229	0	0	0	0	0	0	1	-360	360;
	5702	4857	0.00205	0.00562	0	295	0	0	0	0	1	-360	360;
	208	5702	0.016	0.049818	0	0	0	0	0	0	1	-360	360;
	208	5702	0.014331	0.054347	0	0	0	0	0	0	1	-360	360;
	8799	4857	0.00738	0.021149	0	0	0	0	0	0	1	-360	360;
	208	8799	0.011256	0.027769	0	0	0	0	0	0	1	-360	360;
	501	4857	0.007207	0.019587	0	0	0	0	0	0	1	-360	360;
	208	501	0.01195	0.024826	0	0	0	0	0	0	1	-360	360;
	5679	3158	0.00796	0.04504	0	0	0	0	0	0	1	-360	360;
	5679	6934	0.00844	0.068649	0	0	0	0	0	0	1	-360	360;
	5679	5874	0.021551	0.0842	0	260	0	0	0	0	1	-360	360;
	2214	8761	0.002809	0.01632	0	299	0	0	0	0	1	-360	360;
	2214	362	0.00028	0.00084	0	0	0	0	0	0	1	-360	360;
	2214	362	0.00032	0.000822	0	0	0	0	0	0	1	-360	360;
	2214	2836	0.007711	0.054271	0	0	0	0	0	0	1	-360	360;
	2214	4508	0.002298	0.017129	0	299	0	0	0	0	1	-360	360;
	8026	8919	0.004711	0.026049	0	0	0	0	0	0	1	-360	360;
	8026	5453	0.008462	0.03792	0	0	0	0	0	0	1	-360	360;
	7876	1776	0.00095	0.01117	0	1086	0	0	0	0	1	-360	360;
	7876	5101	0.00142	0.01429	0	1152	0	0	0	0	1	-360	360;
	5794	7887	0.000569	0.005382	0	260	0	0	0	0	1	-360	360;
	3322	2069	0.00984	0.04424	0	0	0	0	0	0	1	-360	360;
	3322	4286	0.0166	0.105449	0	0	0	0	0	0	1	-360	360;
	848	7465	0.007209	0.034969	0	0	0	0	0	0	1	-360	360;
	848	1673	0.023951	0.07004	0	0	0	0	0	0	1	-360	360;
	8919	2285	0.002258	0.01644	0	0	0	0	0	0	1	-360	360;
	4845	1385	0.00948	0.02892	0	0	0	0	0	0	1	-360	360;
	4845	2569	0.01224	0.04384	0	0	0	0	0	0	1	-360	360;
	2351	2569	0.016	0.062369	0	0	0	0	0	0	1	-360	360;
	2351	2247	0.020422	0.082729	0	0	0	0	0	0	1	-360	360;
	2351	2247	0.020009	0.08436	0	0	0	0	0	0	1	-360	360;
	4440	2247	0.010702	0.030338	0	0	0	0	0	0	1	-360	360;
	4440	2247	0.010902	0.02968	0	0	0	0	0	0	1	-360	360;
	5609	2662	0.00848	0.040809	0	0	0	0	0	0	1	-360	360;
	5609	4129	0.012862	0.05772	0	0	0	0	0	0	1	-360	360;
	5600	1776	0.00047	0.00546	0	889	0	0	0	0	1	-360	360;
	7708	7670	0.002871	0.02108	0	0	0	0	0	0	1	-360	360;
	7708	4286	0.004849	0.02916	0	0	0	0	0	0	1	-360	360;
	7708	4286	0.004751	0.029711	0	0	0	0	0	0	1	-360	360;
	5154	4286	0.001569	0.005489	0	0	0	0	0	0	1	-360	360;
	6653	6492	0.0003	0.00316	0	1218	0	0	0	0	1	-360	360;
	6653	5101	0.00114	0.01207	0	1152	0	0	0	0	1	-360	360;
	6653	6084	0.00147	0.01724	0	1086	0	0	0	0	1	-360	360;
	6653	9228	0.00039	0.00552	0	1251	0	0	0	0	1	-360	360;
	8761	2836	0.004991	0.036511	0	0	0	0	0	0	1	-360	360;
	8761	8095	0.007302	0.041071	0	0	0	0	0	0	1	-360	360;
	1636	5977	0.00022	0.00305	0	889	0	0	0	0	1	-360	360;
	1636	285	0.00074	0.00834	0	954	0	0	0	0	1	-360	360;
	1636	4217	0.00126	0.0174	0	856	0	0	0	0	1	-360	360;
	4228	3158	0.00216	0.012578	0	0	0	0	0	0	1	-360	360;
	4228	647	0.001329	0.0106	0	0	0	0	0	0	1	-360	360;
	4228	6934	0.006049	0.063742	0	0	0	0	0	0	1	-360	360;
	5374	8456	0.006769	0.02648	0	0	0	0	0	0	1	-360	360;
	5374	8129	0.016022	0.047791	0	0	0	0	0	0	1	-360	360;
	5374	2836	0.02884	0.096209	0	260	0	0	0	0	1	-360	360;
	7449	8129	0.00656	0.024578	0	0	0	0	0	0	1	-360	360;
	7449	1385	0.00388	0.017342	0	0	0	0	0	0	1	-360	360;
	7449	7742	0.01772	0.064978	0	0	0	0	0	0	1	-360	360;
	3158	651	0.01312	0.052809	0	299	0	0	0	0	1	-360	360;
	3158	647	0.001738	0.002889	0	0	0	0	0	0	1	-360	360;
	651	8032	0.001169	0.003489	0	299	0	0	0	0	1	-360	360;
	4145	7559	0.00368	0.035022	0	0	0	0	0	0	1	-360	360;
	4145	6839	0.001391	0.010338	0	0	0	0	0	0	1	-360	360;
	4145	4601	0.004222	0.030831	0	0	0	0	0	0	1	-360	360;
	3009	5715	0.004271	0.036089	0	0	0	0	0	0	1	-360	360;
	3009	7559	0.00596	0.04024	0	0	0	0	0	0	1	-360	360;
	3851	5115	0.011529	0.046698	0	0	0	0	0	0	1	-360	360;
	3851	2569	0.02252	0.087338	0	0	0	0	0	0	1	-360	360;
	6608	3675	0.004071	0.035591	0	0	0	0	0	0	1	-360	360;
	6608	4601	0.004138	0.0386	0	0	0	0	0	0	1	-360	360;
	2839	2519	0.010911	0.027231	0	0	0	0	0	0	1	-360	360;
	81	6506	0.00256	0.024622	0	0	0	0	0	0	1	-360	360;
	81	4601	0.005129	0.040671	0	0	0	0	0	0	1	-360	360;
	8456	7465	0.005462	0.032671	0	0	0	0	0	0	1	-360	360;
	8456	1673	0.002871	0.014209	0	0	0	0	0	0	1	-360	360;
	8456	1673	0.002929	0.013911	0	0	0	0	0	0	1	-360	360;
	8456	200	0.003422	0.011009	0	0	0	0	0	0	1	-360	360;
	6506	7559	0.00212	0.016769	0	0	0	0	0	0	1	-360	360;
	1953	5101	0.00125	0.01387	0	1711	0	0	0	0	1	-360	360;
	8611	2569	0.004289	0.017071	0	260	0	0	0	0	1	-360	360;
	8611	5273	0.0084	0.024738	0	0	0	0	0	0	1	-360	360;
	2519	4286	0.01044	0.039671	0	0	0	0	0	0	1	-360	360;
	2519	4286	0.010218	0.040418	0	0	0	0	0	0	1	-360	360;
	2519	3743	0.008351	0.02432	0	0	0	0	0	0	1	-360	360;
	3675	2411	0.00188	0.006711	0	0	0	0	0	0	1	-360	360;
	3675	8095	0.00132	0.00352	0	0	0	0	0	0	1	-360	360;
	7416	4217	0.00113	0.01322	0	1415	0	0	0	0	1	-360	360;
	7416	1776	0.00098	0.01301	0	1053	0	0	0	0	1	-360	360;
	2069	5453	0.005271	0.05104	0	0	0	0	0	0	1	-360	360;
	2069	7544	0.017458	0.109631	0	0	0	0	0	0	1	-360	360;
	2069	222	0.002831	0.01016	0	0	0	0	0	0	1	-360	360;
	2069	222	0.001978	0.010902	0	0	0	0	0	0	1	-360	360;
	285	4217	0.00059	0.00764	0	790	0	0	0	0	1	-360	360;
	1345	8129	0.000231	0.000769	0	0	0	0	0	0	1	-360	360;
	5715	2285	0.002809	0.02468	0	0	0	0	0	0	1	-360	360;
	5715	7887	0.009942	0.050222	0	0	0	0	0	0	1	-360	360;
	4217	8206	0.00117	0.01756	0	1547	0	0	0	0	1	-360	360;
	4217	9047	9e-05	0.0017	0	856	0	0	0	0	1	-360	360;
	2285	6391	0.002569	0.02344	0	0	0	0	0	0	1	-360	360;
	2285	4720	0.00144	0.011369	0	0	0	0	0	0	1	-360	360;
	9058	8201	0.002129	0.019689	0	0	0	0	0	0	1	-360	360;
	4286	7544	0.00756	0.048378	0	0	0	0	0	0	1	-360	360;
	4286	5115	0.031138	0.104969	0	0	0	0	0	0	1	-360	360;
	8206	6084	0.00043	0.00623	0	1185	0	0	0	0	1	-360	360;
	7559	2544	0.000769	0.005769	0	0	0	0	0	0	1	-360	360;
	736	4601	0.01464	0.043018	0	0	0	0	0	0	1	-360	360;
	736	137	0.006609	0.014329	0	0	0	0	0	0	1	-360	360;
	4200	6084	0.00021	0.00252	0	856	0	0	0	0	1	-360	360;
	4376	4601	0.002689	0.019791	0	0	0	0	0	0	1	-360	360;
	4376	4601	0.002738	0.01936	0	0	0	0	0	0	1	-360	360;
	4376	4601	0.001338	0.010018	0	0	0	0	0	0	1	-360	360;
	6839	4601	0.001698	0.012871	0	0	0	0	0	0	1	-360	360;
	1646	4601	0.00384	0.028342	0	0	0	0	0	0	1	-360	360;
	2836	8201	0.00616	0.033618	0	0	0	0	0	0	1	-360	360;
	2836	7742	0.01488	0.045698	0	0	0	0	0	0	1	-360	360;
	6492	9228	4e-05	0.00051	0	1218	0	0	0	0	1	-360	360;
	1673	2662	0.028751	0.090409	0	0	0	0	0	0	1	-360	360;
	1673	4129	0.00092	0.004609	0	0	0	0	0	0	1	-360	360;
	3931	2569	0.005151	0.012329	0	0	0	0	0	0	1	-360	360;
	6391	4720	0.002129	0.018	0	0	0	0	0	0	1	-360	360;
	222	5874	0.008929	0.041471	0	0	0	0	0	0	1	-360	360;
	222	2124	0.0222	0.091822	0	0	0	0	0	0	1	-360	360;
	3743	5273	0.005671	0.021169	0	0	0	0	0	0	1	-360	360;
	4720	2124	0.00956	0.035551	0	273	0	0	0	0	1	-360	360;
	2247	8201	0.003169	0.021911	0	0	0	0	0	0	1	-360	360;
	2247	8201	0.00324	0.021471	0	0	0	0	0	0	1	-360	360;
	2247	4601	0.01168	0.055391	0	0	0	0	0	0	1	-360	360;
	2247	4601	0.011902	0.054209	0	0	0	0	0	0	1	-360	360;
	4601	137	0.01908	0.062769	0	0	0	0	0	0	1	-360	360;
	4129	4508	0.001591	0.011098	0	299	0	0	0	0	1	-360	360;
	6561	1294	0.003289	0.019091	0	457	0	0	0	0	1	-360	360;
	6561	1666	0.004169	0.025349	0	476	0	0	0	0	1	-360	360;
	6561	4096	0.00439	0.023909	0	457	0	0	0	0	1	-360	360;
	4717	4241	0.00156	0.01956	0	1316	0	0	0	0	1	-360	360;
	4241	7395	0.00029	0.00391	0	1843	0	0	0	0	1	-360	360;
	4241	5873	0.00179	0.02522	0	1645	0	0	0	0	1	-360	360;
	1479	4241	0.00079	0.00756	0	1843	0	0	0	0	1	-360	360;
	4241	1475	0.00027	0.00281	0	1645	0	0	0	0	1	-360	360;
	4241	1475	0.00026	0.00286	0	1547	0	0	0	0	1	-360	360;
	2248	264	0.000281	0.00213	0	0	0	0	0	0	1	-360	360;
	2248	8530	0.000419	0.00299	0	0	0	0	0	0	1	-360	360;
	3744	4806	0.001659	0.009091	0	0	0	0	0	0	1	-360	360;
	3744	6239	0.00439	0.03324	0	381	0	0	0	0	1	-360	360;
	5991	8671	0.000841	0.00612	0	400	0	0	0	0	1	-360	360;
	5991	201	0.003341	0.022841	0	533	0	0	0	0	1	-360	360;
	5991	3512	0.00581	0.04232	0	0	0	0	0	0	1	-360	360;
	5991	1430	0.00495	0.02919	0	438	0	0	0	0	1	-360	360;
	5991	5592	0.01789	0.07795	0	381	0	0	0	0	1	-360	360;
	5991	4953	0.00186	0.014289	0	381	0	0	0	0	1	-360	360;
	5991	2355	0.002281	0.014401	0	0	0	0	0	0	1	-360	360;
	4510	2957	0.00031	0.00132	0	0	0	0	0	0	1	-360	360;
	4510	6936	0.001349	0.0077	0	381	0	0	0	0	1	-360	360;
	5697	5933	0.00586	0.02574	0	476	0	0	0	0	1	-360	360;
	5697	6040	0.00556	0.020211	0	591	0	0	0	0	1	-360	360;
	5189	2917	0.00026	0.001159	0	553	0	0	0	0	1	-360	360;
	2917	6654	0.000599	0.002419	0	591	0	0	0	0	1	-360	360;
	6679	7139	0.0005	0.00386	0	0	0	0	0	0	1	-360	360;
	6679	6134	0.00107	0.008289	0	533	0	0	0	0	1	-360	360;
	6679	6134	0.001091	0.00811	0	0	0	0	0	0	1	-360	360;
	7413	6679	0.00038	0.00255	0	533	0	0	0	0	1	-360	360;
	5189	6456	0.000271	0.000979	0	0	0	0	0	0	1	-360	360;
	2358	1638	7e-05	0.00076	0	0	0	0	0	0	1	-360	360;
	2358	8571	0.000521	0.003901	0	0	0	0	0	0	1	-360	360;
	4717	8247	0.0016	0.01839	0	1251	0	0	0	0	1	-360	360;
	5787	1393	0.007651	0.058781	0	476	0	0	0	0	1	-360	360;
	5787	6243	0.00394	0.02182	0	0	0	0	0	0	1	-360	360;
	5787	8338	0.003461	0.03005	0	553	0	0	0	0	1	-360	360;
	5787	751	0.00337	0.01831	0	438	0	0	0	0	1	-360	360;
	8285	4293	0.000581	0.00194	0	495	0	0	0	0	1	-360	360;
	7392	4293	0.000539	0.001719	0	0	0	0	0	0	1	-360	360;
	6632	930	0.00249	0.03075	0	1580	0	0	0	0	1	-360	360;
	930	4407	0.0014	0.01409	0	1678	0	0	0	0	1	-360	360;
	7834	6596	0.008	0.050031	0	381	0	0	0	0	1	-360	360;
	7834	5592	0.002211	0.01562	0	572	0	0	0	0	1	-360	360;
	1418	3633	0.00281	0.021831	0	476	0	0	0	0	1	-360	360;
	1418	7820	0.000711	0.004919	0	0	0	0	0	0	1	-360	360;
	2921	6093	0.001841	0.00738	0	0	0	0	0	0	1	-360	360;
	2921	3487	0.002849	0.008979	0	0	0	0	0	0	1	-360	360;
	8353	3633	0.00232	0.014099	0	0	0	0	0	0	1	-360	360;
	8353	3633	0.00237	0.0138	0	0	0	0	0	0	1	-360	360;
	8353	6904	0.00081	0.004591	0	476	0	0	0	0	1	-360	360;
	8353	6625	0.002219	0.010469	0	0	0	0	0	0	1	-360	360;
	6261	1872	0.001081	0.00851	0	0	0	0	0	0	1	-360	360;
	1872	7973	0.00087	0.006969	0	0	0	0	0	0	1	-360	360;
	1872	7973	0.00088	0.006831	0	0	0	0	0	0	1	-360	360;
	1872	4651	0.000349	0.00237	0	0	0	0	0	0	1	-360	360;
	710	6499	0.00232	0.00924	0	381	0	0	0	0	1	-360	360;
	710	3792	0.00812	0.03669	0	400	0	0	0	0	1	-360	360;
	2193	6717	0.003521	0.02825	0	476	0	0	0	0	1	-360	360;
	2193	4752	0.00264	0.017651	0	0	0	0	0	0	1	-360	360;
	7644	6499	0.00486	0.02149	0	457	0	0	0	0	1	-360	360;
	7644	6769	0.006831	0.034669	0	419	0	0	0	0	1	-360	360;
	7644	7075	0.017211	0.108151	0	381	0	0	0	0	1	-360	360;
	7644	8213	0.00063	0.00475	0	419	0	0	0	0	1	-360	360;
	7644	8213	0.00062	0.004841	0	0	0	0	0	0	1	-360	360;
	6632	2507	0.0009	0.01092	0	1843	0	0	0	0	1	-360	360;
	2507	6917	0.00124	0.01096	0	1777	0	0	0	0	1	-360	360;
	3095	2951	0.003211	0.00176	0	0	0	0	0	0	1	-360	360;
	3095	7820	0.001521	0.008031	0	0	0	0	0	0	1	-360	360;
	5425	1898	0.0005	0.00441	0	1415	0	0	0	0	1	-360	360;
	5425	1498	0.00247	0.02573	0	1711	0	0	0	0	1	-360	360;
	5425	1280	0.00042	0.00367	0	1283	0	0	0	0	1	-360	360;
	7431	4279	0.0023	0.01644	0	457	0	0	0	0	1	-360	360;
	7431	4279	0.00226	0.01676	0	457	0	0	0	0	1	-360	360;
	7431	4279	0.0023	0.0162	0	476	0	0	0	0	1	-360	360;
	7431	4279	0.00225	0.015	0	419	0	0	0	0	1	-360	360;
	2343	5946	0.007281	0.035729	0	438	0	0	0	0	1	-360	360;
	2236	5946	0.001919	0.00945	0	438	0	0	0	0	1	-360	360;
	3821	1173	0.005211	0.02981	0	514	0	0	0	0	1	-360	360;
	3821	6936	0.000659	0.00586	0	0	0	0	0	0	1	-360	360;
	3821	6936	0.000591	0.00664	0	0	0	0	0	0	1	-360	360;
	3821	2946	0.011969	0.07424	0	533	0	0	0	0	1	-360	360;
	6261	3065	0.000531	0.003919	0	0	0	0	0	0	1	-360	360;
	4124	3378	0.001521	0.00718	0	438	0	0	0	0	1	-360	360;
	4124	3378	0.001211	0.00613	0	476	0	0	0	0	1	-360	360;
	4124	4809	0.00676	0.03725	0	419	0	0	0	0	1	-360	360;
	4124	4750	0.00399	0.018091	0	381	0	0	0	0	1	-360	360;
	6663	5217	0.000341	0.00339	0	0	0	0	0	0	1	-360	360;
	6663	4096	0.001271	0.0102	0	0	0	0	0	0	1	-360	360;
	6663	4096	0.00125	0.010401	0	0	0	0	0	0	1	-360	360;
	4855	509	0.00261	0.016651	0	0	0	0	0	0	1	-360	360;
	9162	509	0.000219	0.002409	0	514	0	0	0	0	1	-360	360;
	6769	6964	0.00601	0.02814	0	438	0	0	0	0	1	-360	360;
	6964	7002	0.008969	0.03787	0	0	0	0	0	0	1	-360	360;
	6964	6879	0.00232	0.01094	0	438	0	0	0	0	1	-360	360;
	6632	6917	0.00196	0.02382	0	1645	0	0	0	0	1	-360	360;
	6632	8870	0.00422	0.03598	0	1218	0	0	0	0	1	-360	360;
	6632	8870	0.0043	0.03522	0	1283	0	0	0	0	1	-360	360;
	8504	6632	0.00208	0.01691	0	1547	0	0	0	0	1	-360	360;
	6632	2730	0.00544	0.05439	0	1251	0	0	0	0	1	-360	360;
	6632	7508	0.00239	0.02598	0	1316	0	0	0	0	1	-360	360;
	6632	385	0.00396	0.03096	0	1711	0	0	0	0	1	-360	360;
	6632	385	0.00353	0.03509	0	1547	0	0	0	0	1	-360	360;
	3650	8858	0.000531	0.003729	0	514	0	0	0	0	1	-360	360;
	3650	8858	0.00038	0.0025	0	0	0	0	0	0	1	-360	360;
	3650	8858	0.00039	0.00244	0	0	0	0	0	0	1	-360	360;
	3650	4856	0.00836	0.044521	0	400	0	0	0	0	1	-360	360;
	1618	1060	0.00264	0.03029	0	1382	0	0	0	0	1	-360	360;
	1618	8806	0.00396	0.04589	0	757	0	0	0	0	1	-360	360;
	8547	6669	0.00411	0.021729	0	400	0	0	0	0	1	-360	360;
	8547	7398	0.00607	0.0285	0	457	0	0	0	0	1	-360	360;
	8124	2496	0.00075	0.0047	0	0	0	0	0	0	1	-360	360;
	8419	2496	0.00114	0.007969	0	553	0	0	0	0	1	-360	360;
	5125	2916	0.00036	0.00324	0	1316	0	0	0	0	1	-360	360;
	5125	8973	0.00094	0.00874	0	1218	0	0	0	0	1	-360	360;
	2471	8888	0.00949	0.05145	0	476	0	0	0	0	1	-360	360;
	2471	4906	0.011719	0.0337	0	400	0	0	0	0	1	-360	360;
	3076	2471	0.0022	0.012331	0	419	0	0	0	0	1	-360	360;
	4182	2235	0.00038	0.00448	0	1382	0	0	0	0	1	-360	360;
	4182	5873	0.00098	0.01126	0	987	0	0	0	0	1	-360	360;
	42	5919	0.00055	0.00621	0	1777	0	0	0	0	1	-360	360;
	42	2096	0.00163	0.01287	0	1711	0	0	0	0	1	-360	360;
	6292	6127	0.0018	0.01886	0	0	0	0	0	0	1	-360	360;
	6292	6127	0.0047	0.024969	0	0	0	0	0	0	1	-360	360;
	6292	2658	0.009711	0.059669	0	400	0	0	0	0	1	-360	360;
	6292	5191	0.005271	0.03451	0	0	0	0	0	0	1	-360	360;
	929	1175	0.016169	0.11936	0	419	0	0	0	0	1	-360	360;
	929	1344	0.00226	0.014789	0	419	0	0	0	0	1	-360	360;
	8671	1421	0.00151	0.01081	0	591	0	0	0	0	1	-360	360;
	709	4652	0.00228	0.0217	0	1810	0	0	0	0	1	-360	360;
	709	3587	0.00137	0.01593	0	1711	0	0	0	0	1	-360	360;
	8159	3798	0.00099	0.0121	0	1086	0	0	0	0	1	-360	360;
	8159	2261	0	0.01881	0	1119	0	0	0	0	1	-360	360;
	8159	7755	0.00143	0.0163	0	1185	0	0	0	0	1	-360	360;
	8159	7287	0.00057	0.00562	0	1119	0	0	0	0	1	-360	360;
	8159	7287	0.00055	0.00572	0	1185	0	0	0	0	1	-360	360;
	8159	6496	0.00057	0.00658	0	1448	0	0	0	0	1	-360	360;
	8159	5372	0.00075	0.0082	0	1251	0	0	0	0	1	-360	360;
	8159	807	0.00253	0.02153	0	1316	0	0	0	0	1	-360	360;
	201	8582	0.00236	0.011219	0	533	0	0	0	0	1	-360	360;
	3193	7053	0.00173	0.0154	0	1415	0	0	0	0	1	-360	360;
	3193	2764	0.00385	0.04172	0	1185	0	0	0	0	1	-360	360;
	3193	2764	0.00377	0.04251	0	1152	0	0	0	0	1	-360	360;
	3193	1498	0.00199	0.02121	0	1580	0	0	0	0	1	-360	360;
	3193	1133	0.00236	0.01933	0	1185	0	0	0	0	1	-360	360;
	3193	1133	0.00241	0.01891	0	1119	0	0	0	0	1	-360	360;
	2301	3427	0.000521	0.004729	0	0	0	0	0	0	1	-360	360;
	2301	8575	0.00094	0.005581	0	0	0	0	0	0	1	-360	360;
	4644	1086	0.00134	0.01324	0	1218	0	0	0	0	1	-360	360;
	4644	2849	0.00073	0.00638	0	1283	0	0	0	0	1	-360	360;
	1458	2459	0.00248	0.02774	0	1810	0	0	0	0	1	-360	360;
	1458	477	0.00152	0.013	0	1580	0	0	0	0	1	-360	360;
	1458	477	0.00155	0.01272	0	1678	0	0	0	0	1	-360	360;
	1458	8210	0.00117	0.01301	0	1349	0	0	0	0	1	-360	360;
	1458	3113	0.00197	0.02054	0	1119	0	0	0	0	1	-360	360;
	1458	8987	0.00157	0.01909	0	1613	0	0	0	0	1	-360	360;
	3065	4450	0.00107	0.008669	0	0	0	0	0	0	1	-360	360;
	4450	1501	0.000469	0.00314	0	0	0	0	0	0	1	-360	360;
	1185	706	0.000409	0.00305	0	0	0	0	0	0	1	-360	360;
	1240	706	0.00569	0.032979	0	0	0	0	0	0	1	-360	360;
	3500	8871	0.00087	0.006771	0	514	0	0	0	0	1	-360	360;
	3500	8582	0.002151	0.014039	0	533	0	0	0	0	1	-360	360;
	6359	6783	9.9e-05	0.00043	0	553	0	0	0	0	1	-360	360;
	6454	6359	0.000151	0.0005	0	0	0	0	0	0	1	-360	360;
	7833	8438	0.001531	0.00812	0	0	0	0	0	0	1	-360	360;
	7833	8438	0.00113	0.00857	0	0	0	0	0	0	1	-360	360;
	4253	7196	0.00226	0.01868	0	514	0	0	0	0	1	-360	360;
	4253	4651	0.000401	0.00239	0	0	0	0	0	0	1	-360	360;
	4230	1912	0.001169	0.008349	0	514	0	0	0	0	1	-360	360;
	4230	6315	0.00324	0.02312	0	533	0	0	0	0	1	-360	360;
	621	4360	0.006031	0.034789	0	495	0	0	0	0	1	-360	360;
	621	808	0.0103	0.05269	0	438	0	0	0	0	1	-360	360;
	2962	6917	0.00129	0.01494	0	1481	0	0	0	0	1	-360	360;
	2962	6917	0.00126	0.01523	0	1349	0	0	0	0	1	-360	360;
	4958	4906	0.001031	0.00601	0	0	0	0	0	0	1	-360	360;
	4958	4906	0.00105	0.00588	0	0	0	0	0	0	1	-360	360;
	2642	5823	0.00197	0.01719	0	1119	0	0	0	0	1	-360	360;
	2642	1947	0.00114	0.01237	0	1382	0	0	0	0	1	-360	360;
	2642	3966	0.00314	0.02868	0	1349	0	0	0	0	1	-360	360;
	2642	1074	0.00294	0.03326	0	1218	0	0	0	0	1	-360	360;
	2642	1074	0.00288	0.0308	0	1448	0	0	0	0	1	-360	360;
	2642	807	0.00088	0.02821	0	1086	0	0	0	0	1	-360	360;
	4165	6659	0.00152	0.01702	0	889	0	0	0	0	1	-360	360;
	4165	7303	0.00022	0.00209	0	1481	0	0	0	0	1	-360	360;
	1793	2668	0.00087	0.005	0	0	0	0	0	0	1	-360	360;
	1793	8338	0.000711	0.00439	0	381	0	0	0	0	1	-360	360;
	1793	8338	0.000729	0.004289	0	419	0	0	0	0	1	-360	360;
	4663	6443	0.001031	0.00549	0	0	0	0	0	0	1	-360	360;
	4663	519	0.003219	0.01512	0	0	0	0	0	0	1	-360	360;
	4663	595	0.007271	0.031419	0	381	0	0	0	0	1	-360	360;
	4663	3291	0.02145	0.085781	0	438	0	0	0	0	1	-360	360;
	6138	7504	0.002841	0.017771	0	0	0	0	0	0	1	-360	360;
	6138	2896	0.003331	0.017289	0	400	0	0	0	0	1	-360	360;
	5280	2343	0.01524	0.064091	0	514	0	0	0	0	1	-360	360;
	8534	5121	0.000831	0.0052	0	514	0	0	0	0	1	-360	360;
	5121	9039	0.00164	0.0103	0	495	0	0	0	0	1	-360	360;
	3961	4238	0.00139	0.00864	0	381	0	0	0	0	1	-360	360;
	3961	1411	0.00143	0.00764	0	438	0	0	0	0	1	-360	360;
	3961	5793	0.000229	0.00126	0	0	0	0	0	0	1	-360	360;
	3961	5793	0.000211	0.00143	0	0	0	0	0	0	1	-360	360;
	976	1549	0.0008	0.01072	0	1547	0	0	0	0	1	-360	360;
	976	4600	0.00067	0.00661	0	1119	0	0	0	0	1	-360	360;
	976	2800	0.0003	0.00427	0	1382	0	0	0	0	1	-360	360;
	976	2800	0.00068	0.00648	0	1349	0	0	0	0	1	-360	360;
	1600	8186	0.0002	0.001289	0	0	0	0	0	0	1	-360	360;
	1600	4011	0.00013	0.000961	0	533	0	0	0	0	1	-360	360;
	5771	2235	0.00118	0.01285	0	1613	0	0	0	0	1	-360	360;
	5771	5239	0.00062	0.00577	0	1086	0	0	0	0	1	-360	360;
	9216	1409	9.1e-05	0.000289	0	0	0	0	0	0	1	-360	360;
	9216	1409	9.1e-05	0.000271	0	0	0	0	0	0	1	-360	360;
	9216	6597	0.0003	0.002719	0	476	0	0	0	0	1	-360	360;
	9216	6836	0.00049	0.002091	0	0	0	0	0	0	1	-360	360;
	9216	15	0.000159	0.001461	0	572	0	0	0	0	1	-360	360;
	429	6132	0.00301	0.02281	0	0	0	0	0	0	1	-360	360;
	6660	6132	0.0027	0.019979	0	0	0	0	0	0	1	-360	360;
	1439	649	0.00019	0.000539	0	0	0	0	0	0	1	-360	360;
	649	6836	0.000169	0.00061	0	514	0	0	0	0	1	-360	360;
	8417	5435	0.00163	0.00939	0	533	0	0	0	0	1	-360	360;
	8417	5435	0.001159	0.009031	0	514	0	0	0	0	1	-360	360;
	8417	4058	9.9e-05	0.00037	0	457	0	0	0	0	1	-360	360;
	5617	9055	0.003789	0.01805	0	0	0	0	0	0	1	-360	360;
	5617	1471	0.00069	0.00389	0	0	0	0	0	0	1	-360	360;
	8396	7925	0.00201	0.01845	0	1283	0	0	0	0	1	-360	360;
	8806	7925	0.00121	0.00902	0	1283	0	0	0	0	1	-360	360;
	4542	1947	0.00155	0.01573	0	1283	0	0	0	0	1	-360	360;
	4542	8337	0.00031	0.00274	0	1316	0	0	0	0	1	-360	360;
	4542	3966	0.00047	0.00563	0	1185	0	0	0	0	1	-360	360;
	4542	3544	0.00066	0.00721	0	1152	0	0	0	0	1	-360	360;
	7510	4806	0.00326	0.016711	0	0	0	0	0	0	1	-360	360;
	7510	2907	0.002479	0.01262	0	629	0	0	0	0	1	-360	360;
	7510	2907	0.00243	0.01286	0	610	0	0	0	0	1	-360	360;
	7510	4338	0.004521	0.02306	0	476	0	0	0	0	1	-360	360;
	7510	1282	0.00211	0.00899	0	0	0	0	0	0	1	-360	360;
	7510	7795	0.001289	0.005651	0	553	0	0	0	0	1	-360	360;
	7510	3382	0.002479	0.009021	0	0	0	0	0	0	1	-360	360;
	7510	6456	0.00151	0.010521	0	419	0	0	0	0	1	-360	360;
	5409	8352	0.00101	0.004151	0	438	0	0	0	0	1	-360	360;
	5409	797	0.015599	0.079	0	381	0	0	0	0	1	-360	360;
	5409	8207	0.013961	0.067591	0	457	0	0	0	0	1	-360	360;
	5409	1241	0.007099	0.043169	0	514	0	0	0	0	1	-360	360;
	2416	5042	0.00317	0.02987	0	987	0	0	0	0	1	-360	360;
	2416	2764	0.00081	0.0088	0	1119	0	0	0	0	1	-360	360;
	2416	2764	0.00098	0.00961	0	987	0	0	0	0	1	-360	360;
	2416	179	0.00194	0.02225	0	921	0	0	0	0	1	-360	360;
	9001	5463	0.001031	0.006349	0	0	0	0	0	0	1	-360	360;
	9001	445	0.00668	0.0297	0	553	0	0	0	0	1	-360	360;
	4809	2538	0.00439	0.02236	0	533	0	0	0	0	1	-360	360;
	2538	6397	0.00249	0.01349	0	591	0	0	0	0	1	-360	360;
	2081	4556	0.001969	0.01205	0	495	0	0	0	0	1	-360	360;
	2081	4592	0.000461	0.003031	0	572	0	0	0	0	1	-360	360;
	3592	267	0.00444	0.023479	0	438	0	0	0	0	1	-360	360;
	3592	4632	0.00175	0.01288	0	0	0	0	0	0	1	-360	360;
	3592	4632	0.001711	0.01312	0	0	0	0	0	0	1	-360	360;
	1549	579	0.0008	0.00784	0	1481	0	0	0	0	1	-360	360;
	1549	9117	0.00051	0.00612	0	1382	0	0	0	0	1	-360	360;
	3282	7145	0.00137	0.00905	0	0	0	0	0	0	1	-360	360;
	3282	4855	0.002771	0.021409	0	381	0	0	0	0	1	-360	360;
	3282	4855	0.00314	0.020979	0	0	0	0	0	0	1	-360	360;
	3282	4095	0.002	0.01118	0	0	0	0	0	0	1	-360	360;
	8285	699	0.00119	0.00875	0	0	0	0	0	0	1	-360	360;
	8285	6882	0.00031	0.00236	0	0	0	0	0	0	1	-360	360;
	8285	7820	0.000831	0.00539	0	0	0	0	0	0	1	-360	360;
	267	7349	0.004349	0.02394	0	419	0	0	0	0	1	-360	360;
	1200	8731	0.0007	0.00782	0	1678	0	0	0	0	1	-360	360;
	1200	1716	0.00046	0.00432	0	1843	0	0	0	0	1	-360	360;
	1200	3067	0.00049	0.00617	0	1678	0	0	0	0	1	-360	360;
	1200	1074	0.00127	0.01514	0	1777	0	0	0	0	1	-360	360;
	3549	6544	0.00232	0.01843	0	0	0	0	0	0	1	-360	360;
	3549	1503	0.00488	0.032151	0	553	0	0	0	0	1	-360	360;
	1409	3715	0.000229	0.00136	0	533	0	0	0	0	1	-360	360;
	1409	6836	0.0003	0.001531	0	476	0	0	0	0	1	-360	360;
	1409	5892	0.00026	0.00126	0	400	0	0	0	0	1	-360	360;
	1409	5948	0.00056	0.002091	0	553	0	0	0	0	1	-360	360;
	4407	2962	0.00103	0.0093	0	1711	0	0	0	0	1	-360	360;
	4407	2437	0.00183	0.01689	0	1316	0	0	0	0	1	-360	360;
	7058	5163	0.00324	0.01095	0	857	0	0	0	0	1	-360	360;
	236	8236	0.001419	0.008539	0	0	0	0	0	0	1	-360	360;
	236	3427	0.00068	0.00438	0	495	0	0	0	0	1	-360	360;
	2916	5884	0.00178	0.01651	0	1283	0	0	0	0	1	-360	360;
	2916	4315	0.00151	0.01827	0	1251	0	0	0	0	1	-360	360;
	2916	4315	0.00148	0.01862	0	1316	0	0	0	0	1	-360	360;
	2962	2437	0.00201	0.02546	0	1580	0	0	0	0	1	-360	360;
	2962	2437	0.00197	0.02357	0	1547	0	0	0	0	1	-360	360;
	3747	3427	0.001091	0.00775	0	0	0	0	0	0	1	-360	360;
	3747	988	0.000841	0.006669	0	0	0	0	0	0	1	-360	360;
	932	3728	0.00151	0.012081	0	514	0	0	0	0	1	-360	360;
	932	7463	0.00032	0.002521	0	438	0	0	0	0	1	-360	360;
	8087	3342	0.006091	0.03019	0	514	0	0	0	0	1	-360	360;
	8087	7659	0.003651	0.016461	0	553	0	0	0	0	1	-360	360;
	8087	3334	0.00076	0.00425	0	381	0	0	0	0	1	-360	360;
	8087	3334	0.00087	0.004031	0	0	0	0	0	0	1	-360	360;
	1639	3728	0.000651	0.004729	0	591	0	0	0	0	1	-360	360;
	1639	8936	0.00045	0.00337	0	514	0	0	0	0	1	-360	360;
	2907	481	0.001539	0.00839	0	438	0	0	0	0	1	-360	360;
	2907	481	0.001919	0.010479	0	514	0	0	0	0	1	-360	360;
	481	9057	0.001169	0.007031	0	438	0	0	0	0	1	-360	360;
	1894	481	0.00739	0.034081	0	0	0	0	0	0	1	-360	360;
	4890	481	0.00164	0.00964	0	419	0	0	0	0	1	-360	360;
	5890	481	0.00206	0.012	0	457	0	0	0	0	1	-360	360;
	5391	481	0.000961	0.00531	0	457	0	0	0	0	1	-360	360;
	3544	6771	6e-05	0.00029	0	987	0	0	0	0	1	-360	360;
	1344	4806	0.001919	0.013151	0	381	0	0	0	0	1	-360	360;
	869	4806	0.00019	0.00107	0	400	0	0	0	0	1	-360	360;
	869	4806	0.00018	0.001091	0	438	0	0	0	0	1	-360	360;
	5685	2459	0.00147	0.01714	0	1514	0	0	0	0	1	-360	360;
	2730	2459	0.00143	0.01632	0	1283	0	0	0	0	1	-360	360;
	2730	2459	0.00142	0.01416	0	1580	0	0	0	0	1	-360	360;
	2730	2459	0.00145	0.01386	0	1744	0	0	0	0	1	-360	360;
	2459	8987	0.00073	0.00732	0	1613	0	0	0	0	1	-360	360;
	1185	1240	0.009039	0.05989	0	476	0	0	0	0	1	-360	360;
	1185	4016	0.007919	0.04975	0	457	0	0	0	0	1	-360	360;
	2668	981	0.000521	0.005841	0	457	0	0	0	0	1	-360	360;
	2668	6344	0.000521	0.0028	0	0	0	0	0	0	1	-360	360;
	6564	1074	0.0007	0.00744	0	889	0	0	0	0	1	-360	360;
	6564	4026	0.00159	0.0148	0	921	0	0	0	0	1	-360	360;
	4336	4249	0.000659	0.005151	0	553	0	0	0	0	1	-360	360;
	7196	4249	0.000729	0.004901	0	419	0	0	0	0	1	-360	360;
	4600	642	0.00064	0.00772	0	1119	0	0	0	0	1	-360	360;
	7573	6186	0.00018	0.00093	0	591	0	0	0	0	1	-360	360;
	7573	8186	0.00019	0.00118	0	495	0	0	0	0	1	-360	360;
	7413	7573	0.000409	0.002341	0	514	0	0	0	0	1	-360	360;
	8719	1023	0.00087	0.006591	0	381	0	0	0	0	1	-360	360;
	8719	3063	0.00162	0.0098	0	0	0	0	0	0	1	-360	360;
	7508	7053	0.00096	0.00849	0	1547	0	0	0	0	1	-360	360;
	6183	6932	0.000341	0.001081	0	495	0	0	0	0	1	-360	360;
	6183	6932	0.0003	0.001219	0	514	0	0	0	0	1	-360	360;
	6183	8068	9.9e-05	0.00076	0	533	0	0	0	0	1	-360	360;
	6183	8068	0.00011	0.00074	0	553	0	0	0	0	1	-360	360;
	1602	4855	0.00151	0.009039	0	0	0	0	0	0	1	-360	360;
	1602	2551	0.00356	0.02331	0	0	0	0	0	0	1	-360	360;
	7999	1439	0.000281	0.00169	0	0	0	0	0	0	1	-360	360;
	7999	6597	0.000169	0.001531	0	495	0	0	0	0	1	-360	360;
	7999	8661	0.00038	0.00326	0	0	0	0	0	0	1	-360	360;
	1086	7257	0.00136	0.01738	0	1481	0	0	0	0	1	-360	360;
	6443	519	0.00186	0.0133	0	0	0	0	0	0	1	-360	360;
	6443	519	0.00182	0.01356	0	0	0	0	0	0	1	-360	360;
	2009	6443	0.002531	0.015919	0	533	0	0	0	0	1	-360	360;
	6443	6749	0.00289	0.015969	0	419	0	0	0	0	1	-360	360;
	1121	4947	0.001771	0.01107	0	400	0	0	0	0	1	-360	360;
	4204	4947	0.007969	0.04238	0	0	0	0	0	0	1	-360	360;
	2907	7191	0.00013	0.00062	0	495	0	0	0	0	1	-360	360;
	2907	284	0.0002	0.00201	0	572	0	0	0	0	1	-360	360;
	477	9159	0.00095	0.01024	0	1349	0	0	0	0	1	-360	360;
	1133	9159	0.00553	0.04909	0	1152	0	0	0	0	1	-360	360;
	3178	6499	0.009229	0.03693	0	438	0	0	0	0	1	-360	360;
	3792	6499	0.00451	0.028469	0	438	0	0	0	0	1	-360	360;
	6499	6038	0.003591	0.01906	0	400	0	0	0	0	1	-360	360;
	6499	7398	0.010159	0.05169	0	495	0	0	0	0	1	-360	360;
	75	2126	0.00039	0.003099	0	0	0	0	0	0	1	-360	360;
	75	6869	0.000729	0.0048	0	0	0	0	0	0	1	-360	360;
	7137	6659	0.00141	0.01296	0	921	0	0	0	0	1	-360	360;
	3673	6745	0.001969	0.019229	0	0	0	0	0	0	1	-360	360;
	3673	8953	0.001081	0.0105	0	0	0	0	0	0	1	-360	360;
	3673	8953	0.00105	0.010711	0	0	0	0	0	0	1	-360	360;
	5685	2849	0.00052	0.00449	0	1415	0	0	0	0	1	-360	360;
	3378	6397	0.01024	0.056349	0	476	0	0	0	0	1	-360	360;
	3378	1520	0.004599	0.023659	0	419	0	0	0	0	1	-360	360;
	3378	3233	0.010021	0.057409	0	438	0	0	0	0	1	-360	360;
	3378	3291	0.00275	0.01525	0	457	0	0	0	0	1	-360	360;
	693	8240	0.00107	0.01122	0	1547	0	0	0	0	1	-360	360;
	693	5884	0.00242	0.02336	0	1218	0	0	0	0	1	-360	360;
	693	3587	0.00165	0.01303	0	823	0	0	0	0	1	-360	360;
	693	3587	0.00155	0.01332	0	856	0	0	0	0	1	-360	360;
	6994	5970	0.00087	0.005469	0	0	0	0	0	0	1	-360	360;
	6994	5970	0.002281	0.011909	0	0	0	0	0	0	1	-360	360;
	6994	970	0.00106	0.01076	0	0	0	0	0	0	1	-360	360;
	6994	5348	0.001409	0.01618	0	381	0	0	0	0	1	-360	360;
	6994	1909	0.00306	0.015521	0	457	0	0	0	0	1	-360	360;
	3843	8534	0.000969	0.006831	0	438	0	0	0	0	1	-360	360;
	3843	9039	0.002531	0.016961	0	533	0	0	0	0	1	-360	360;
	1021	257	0.01207	0.077901	0	0	0	0	0	0	1	-360	360;
	1021	6936	0.014091	0.080581	0	495	0	0	0	0	1	-360	360;
	1021	2946	0.003289	0.01481	0	0	0	0	0	0	1	-360	360;
	1021	172	0.003349	0.0145	0	0	0	0	0	0	1	-360	360;
	8138	3792	0.005651	0.026479	0	419	0	0	0	0	1	-360	360;
	8138	7659	0.01193	0.064669	0	457	0	0	0	0	1	-360	360;
	6669	5111	0.004151	0.017281	0	419	0	0	0	0	1	-360	360;
	5511	7145	0.00213	0.014091	0	0	0	0	0	0	1	-360	360;
	4338	1894	0.00113	0.00651	0	419	0	0	0	0	1	-360	360;
	4338	1894	0.001159	0.00637	0	419	0	0	0	0	1	-360	360;
	4338	6456	0.00611	0.03288	0	457	0	0	0	0	1	-360	360;
	8946	849	0.00012	0.00156	0	1744	0	0	0	0	1	-360	360;
	8946	849	0.00013	0.00153	0	1645	0	0	0	0	1	-360	360;
	8946	6768	0.00285	0.03071	0	1645	0	0	0	0	1	-360	360;
	5970	2896	0.00676	0.03787	0	495	0	0	0	0	1	-360	360;
	5970	9049	0.004461	0.023521	0	438	0	0	0	0	1	-360	360;
	5332	618	0.00064	0.0078	0	1514	0	0	0	0	1	-360	360;
	5332	618	0.00065	0.00764	0	1514	0	0	0	0	1	-360	360;
	5332	579	0.00066	0.00602	0	1481	0	0	0	0	1	-360	360;
	5332	4442	0.00149	0.01224	0	889	0	0	0	0	1	-360	360;
	5332	814	0.00061	0.00582	0	1020	0	0	0	0	1	-360	360;
	3633	6904	0.001461	0.00744	0	0	0	0	0	0	1	-360	360;
	3633	7820	0.00386	0.02406	0	0	0	0	0	0	1	-360	360;
	3633	7632	0.00306	0.01555	0	0	0	0	0	0	1	-360	360;
	3633	7632	0.003	0.01586	0	0	0	0	0	0	1	-360	360;
	3629	4112	0.003271	0.020651	0	0	0	0	0	0	1	-360	360;
	4112	7630	0.000849	0.00657	0	0	0	0	0	0	1	-360	360;
	1645	3066	0.000211	0.001169	0	476	0	0	0	0	1	-360	360;
	1645	7973	0.000469	0.00345	0	495	0	0	0	0	1	-360	360;
	4945	7508	0.00126	0.0145	0	1382	0	0	0	0	1	-360	360;
	1957	5933	0.008219	0.04925	0	0	0	0	0	0	1	-360	360;
	1957	1769	0.006771	0.04132	0	438	0	0	0	0	1	-360	360;
	7417	7600	0.000841	0.004539	0	591	0	0	0	0	1	-360	360;
	7417	3792	0.00701	0.03161	0	476	0	0	0	0	1	-360	360;
	7417	6879	0.000781	0.004271	0	572	0	0	0	0	1	-360	360;
	7417	962	0.00049	0.0032	0	533	0	0	0	0	1	-360	360;
	3489	3813	0.000331	0.00199	0	629	0	0	0	0	1	-360	360;
	1347	3813	0.00032	0.002031	0	629	0	0	0	0	1	-360	360;
	3813	5500	9.1e-05	0.00057	0	476	0	0	0	0	1	-360	360;
	5366	869	0.001651	0.006789	0	533	0	0	0	0	1	-360	360;
	5366	3382	0.001461	0.007229	0	0	0	0	0	0	1	-360	360;
	8352	9049	0.00437	0.023979	0	438	0	0	0	0	1	-360	360;
	1869	473	0.00781	0.04595	0	419	0	0	0	0	1	-360	360;
	1869	6243	0.0015	0.006531	0	457	0	0	0	0	1	-360	360;
	5767	9057	0.001031	0.006469	0	0	0	0	0	0	1	-360	360;
	2800	3061	0.00183	0.01686	0	1119	0	0	0	0	1	-360	360;
	2800	618	0.0002	0.00294	0	1514	0	0	0	0	1	-360	360;
	2800	618	0.00021	0.00288	0	1316	0	0	0	0	1	-360	360;
	2800	5042	0.00171	0.01889	0	1283	0	0	0	0	1	-360	360;
	8870	2764	0.00069	0.00565	0	1185	0	0	0	0	1	-360	360;
	8870	2764	0.0007	0.00553	0	1382	0	0	0	0	1	-360	360;
	7096	4290	0.00011	0.00115	0	1119	0	0	0	0	1	-360	360;
	7096	6681	0.00012	0.00126	0	1316	0	0	0	0	1	-360	360;
	7096	8973	0.00326	0.02919	0	1382	0	0	0	0	1	-360	360;
	7096	5060	0.00092	0.00876	0	823	0	0	0	0	1	-360	360;
	5780	6932	0.00087	0.005409	0	457	0	0	0	0	1	-360	360;
	5780	6869	0.00131	0.00701	0	438	0	0	0	0	1	-360	360;
	4291	6656	0.002531	0.018781	0	0	0	0	0	0	1	-360	360;
	4291	8729	0.00894	0.056909	0	0	0	0	0	0	1	-360	360;
	5512	5463	0.00187	0.010659	0	553	0	0	0	0	1	-360	360;
	7685	5463	0.00286	0.01538	0	610	0	0	0	0	1	-360	360;
	5463	4967	0.0008	0.00444	0	419	0	0	0	0	1	-360	360;
	1013	7875	0.00545	0.03936	0	0	0	0	0	0	1	-360	360;
	1741	7875	0.00119	0.009031	0	0	0	0	0	0	1	-360	360;
	51	5670	0.00295	0.0162	0	419	0	0	0	0	1	-360	360;
	51	6174	0.001581	0.0087	0	514	0	0	0	0	1	-360	360;
	51	2517	0.007729	0.025401	0	591	0	0	0	0	1	-360	360;
	7805	3425	0.00013	0.001461	0	457	0	0	0	0	1	-360	360;
	7805	8658	0.000341	0.02219	0	0	0	0	0	0	1	-360	360;
	7805	4011	0.00018	0.001211	0	591	0	0	0	0	1	-360	360;
	7805	9028	0.003099	0.0147	0	591	0	0	0	0	1	-360	360;
	7805	9028	0.003159	0.014401	0	533	0	0	0	0	1	-360	360;
	6932	4274	0.00281	0.01305	0	0	0	0	0	0	1	-360	360;
	6932	8871	0.002281	0.01075	0	495	0	0	0	0	1	-360	360;
	6932	8871	0.00232	0.010521	0	457	0	0	0	0	1	-360	360;
	6932	3442	0.000151	0.000841	0	0	0	0	0	0	1	-360	360;
	6953	2658	0.002771	0.012091	0	572	0	0	0	0	1	-360	360;
	6953	5592	0.00613	0.03005	0	457	0	0	0	0	1	-360	360;
	3342	7659	0.003219	0.02486	0	0	0	0	0	0	1	-360	360;
	6186	8068	0.000211	0.001331	0	572	0	0	0	0	1	-360	360;
	8504	1898	0.00185	0.01917	0	1613	0	0	0	0	1	-360	360;
	1282	869	0.00082	0.00381	0	572	0	0	0	0	1	-360	360;
	9043	7819	0.00075	0.004531	0	381	0	0	0	0	1	-360	360;
	9043	7819	0.000729	0.00419	0	419	0	0	0	0	1	-360	360;
	9043	32	0.02626	0.15576	0	438	0	0	0	0	1	-360	360;
	9043	7423	0.011469	0.03937	0	0	0	0	0	0	1	-360	360;
	7139	8186	0.00011	0.00061	0	514	0	0	0	0	1	-360	360;
	7139	1065	0.000531	0.00313	0	381	0	0	0	0	1	-360	360;
	7139	6134	0.001211	0.00864	0	0	0	0	0	0	1	-360	360;
	4336	4651	0.00094	0.006289	0	0	0	0	0	0	1	-360	360;
	731	4020	0.001	0.00886	0	823	0	0	0	0	1	-360	360;
	6769	3178	0.001979	0.010469	0	514	0	0	0	0	1	-360	360;
	6769	3178	0.0017	0.00855	0	553	0	0	0	0	1	-360	360;
	3178	2116	0.00757	0.04624	0	419	0	0	0	0	1	-360	360;
	3178	7283	0.001719	0.00794	0	495	0	0	0	0	1	-360	360;
	3178	7283	0.00176	0.007781	0	0	0	0	0	0	1	-360	360;
	3157	916	0.00126	0.0053	0	400	0	0	0	0	1	-360	360;
	3157	1241	0.00444	0.022789	0	419	0	0	0	0	1	-360	360;
	477	2437	0.00163	0.0145	0	1086	0	0	0	0	1	-360	360;
	6769	257	0.003281	0.02124	0	495	0	0	0	0	1	-360	360;
	6769	7283	0.004979	0.025271	0	476	0	0	0	0	1	-360	360;
	8858	2403	0.0072	0.055289	0	495	0	0	0	0	1	-360	360;
	8858	4772	0.00787	0.04582	0	0	0	0	0	0	1	-360	360;
	8858	4016	0.003961	0.017711	0	0	0	0	0	0	1	-360	360;
	2403	6797	0.00756	0.04826	0	457	0	0	0	0	1	-360	360;
	6192	8207	0.003401	0.02086	0	457	0	0	0	0	1	-360	360;
	6192	3937	0.001341	0.00669	0	419	0	0	0	0	1	-360	360;
	7669	8867	6e-05	0.000581	0	438	0	0	0	0	1	-360	360;
	7669	5948	9.1e-05	0.00074	0	514	0	0	0	0	1	-360	360;
	7392	8939	0.000531	0.00175	0	0	0	0	0	0	1	-360	360;
	1114	5096	0.00116	0.01159	0	1382	0	0	0	0	1	-360	360;
	1114	4315	0.00038	0.00457	0	1613	0	0	0	0	1	-360	360;
	3066	7973	0.000729	0.00405	0	553	0	0	0	0	1	-360	360;
	264	7064	0.000331	0.001919	0	0	0	0	0	0	1	-360	360;
	2188	4706	0.00086	0.01029	0	1909	0	0	0	0	1	-360	360;
	7428	2188	0.00027	0.00332	0	1711	0	0	0	0	1	-360	360;
	5170	5681	0.00012	0.00156	0	0	0	0	0	0	1	-360	360;
	5170	5681	0.00012	0.001531	0	0	0	0	0	0	1	-360	360;
	8756	2060	0.00256	0.014211	0	0	0	0	0	0	1	-360	360;
	8756	8419	0.00049	0.00338	0	0	0	0	0	0	1	-360	360;
	6009	1430	0.007531	0.04444	0	400	0	0	0	0	1	-360	360;
	6009	5592	0.007409	0.03226	0	0	0	0	0	0	1	-360	360;
	3150	8147	0.001789	0.012841	0	0	0	0	0	0	1	-360	360;
	3150	4096	0.001831	0.01257	0	0	0	0	0	0	1	-360	360;
	1632	3621	0.000169	0.00143	0	0	0	0	0	0	1	-360	360;
	1632	6869	0.00088	0.005099	0	610	0	0	0	0	1	-360	360;
	3061	7460	0.0004	0.00479	0	1678	0	0	0	0	1	-360	360;
	3061	7460	0.00052	0.00481	0	1744	0	0	0	0	1	-360	360;
	3061	7929	0.00135	0.01431	0	1744	0	0	0	0	1	-360	360;
	3061	814	0.00051	0.00549	0	1481	0	0	0	0	1	-360	360;
	699	6882	0.00105	0.007969	0	476	0	0	0	0	1	-360	360;
	699	947	0.00339	0.020531	0	0	0	0	0	0	1	-360	360;
	699	2951	0.00169	0.009331	0	0	0	0	0	0	1	-360	360;
	3584	4177	0.00374	0.01905	0	438	0	0	0	0	1	-360	360;
	3584	5276	0.00487	0.026341	0	533	0	0	0	0	1	-360	360;
	5034	8301	0.00057	0.004581	0	0	0	0	0	0	1	-360	360;
	5034	8582	0.000409	0.00301	0	0	0	0	0	0	1	-360	360;
	1461	3754	9.1e-05	0.00076	0	419	0	0	0	0	1	-360	360;
	1461	6179	0.003919	0.01682	0	476	0	0	0	0	1	-360	360;
	5823	1060	0.00117	0.0131	0	1514	0	0	0	0	1	-360	360;
	5280	4954	0.008271	0.02787	0	495	0	0	0	0	1	-360	360;
	5280	1121	0.016599	0.0572	0	419	0	0	0	0	1	-360	360;
	5280	7861	0.00999	0.057	0	0	0	0	0	0	1	-360	360;
	5280	7861	0.01132	0.05586	0	0	0	0	0	0	1	-360	360;
	7303	4690	4e-05	0.00045	0	921	0	0	0	0	1	-360	360;
	4954	8534	0.003281	0.01038	0	495	0	0	0	0	1	-360	360;
	4317	72	0.003711	0.01725	0	457	0	0	0	0	1	-360	360;
	4317	6333	0.00069	0.00232	0	457	0	0	0	0	1	-360	360;
	8702	4948	0.004151	0.02436	0	514	0	0	0	0	1	-360	360;
	5163	4948	0.006781	0.028729	0	476	0	0	0	0	1	-360	360;
	3489	6869	0.000331	0.00151	0	438	0	0	0	0	1	-360	360;
	1347	6869	0.000331	0.001479	0	629	0	0	0	0	1	-360	360;
	3687	5877	0.0001	0.00144	0	1514	0	0	0	0	1	-360	360;
	3687	5919	0.00027	0.00221	0	1053	0	0	0	0	1	-360	360;
	1232	4588	0.0001	0.00141	0	1711	0	0	0	0	1	-360	360;
	1232	2764	0.00054	0.00546	0	1448	0	0	0	0	1	-360	360;
	7428	5877	7e-05	0.00053	0	1382	0	0	0	0	1	-360	360;
	4588	7428	6e-05	0.0006	0	1481	0	0	0	0	1	-360	360;
	8939	4371	0.000341	0.00245	0	0	0	0	0	0	1	-360	360;
	8939	7820	0.000841	0.006719	0	0	0	0	0	0	1	-360	360;
	6209	7191	0.000401	0.00164	0	476	0	0	0	0	1	-360	360;
	519	7349	0.004409	0.032531	0	0	0	0	0	0	1	-360	360;
	429	7877	0.001961	0.013581	0	381	0	0	0	0	1	-360	360;
	1894	4890	0.006659	0.03437	0	0	0	0	0	0	1	-360	360;
	4062	1599	9.1e-05	0.0007	0	0	0	0	0	0	1	-360	360;
	4062	8079	0.00051	0.004211	0	0	0	0	0	0	1	-360	360;
	6841	8936	0.00043	0.0027	0	0	0	0	0	0	1	-360	360;
	6841	2564	0.00038	0.00306	0	0	0	0	0	0	1	-360	360;
	8710	1491	0.001461	0.009479	0	0	0	0	0	0	1	-360	360;
	8710	7406	0.001841	0.011081	0	514	0	0	0	0	1	-360	360;
	970	7504	0.00094	0.007169	0	553	0	0	0	0	1	-360	360;
	6039	8034	0.00082	0.00846	0	1415	0	0	0	0	1	-360	360;
	6039	7428	0.00029	0.00325	0	1711	0	0	0	0	1	-360	360;
	7973	7475	0.000781	0.0062	0	0	0	0	0	0	1	-360	360;
	7973	2740	8.1e-05	0.00187	0	495	0	0	0	0	1	-360	360;
	7973	2297	8.1e-05	0.001909	0	419	0	0	0	0	1	-360	360;
	6877	1060	0.00113	0.00956	0	1185	0	0	0	0	1	-360	360;
	6877	1060	0.00115	0.00935	0	1218	0	0	0	0	1	-360	360;
	8981	4428	0.00161	0.02036	0	0	0	0	0	0	1	-360	360;
	8981	4428	0.00157	0.02075	0	0	0	0	0	0	1	-360	360;
	1841	3728	0.00101	0.007401	0	533	0	0	0	0	1	-360	360;
	1841	7463	0.00087	0.00711	0	476	0	0	0	0	1	-360	360;
	3582	5919	0.00135	0.01442	0	1349	0	0	0	0	1	-360	360;
	3582	2096	0.00176	0.01441	0	1777	0	0	0	0	1	-360	360;
	2009	7349	0.00682	0.03187	0	419	0	0	0	0	1	-360	360;
	8034	1399	0.00109	0.00987	0	1185	0	0	0	0	1	-360	360;
	4652	8034	0.00273	0.02829	0	1283	0	0	0	0	1	-360	360;
	1752	7845	0.001581	0.011289	0	400	0	0	0	0	1	-360	360;
	1752	3427	0.001669	0.013099	0	0	0	0	0	0	1	-360	360;
	7347	1599	9.9e-05	0.00069	0	495	0	0	0	0	1	-360	360;
	7347	8079	9.9e-05	0.000669	0	514	0	0	0	0	1	-360	360;
	4896	3736	0.000591	0.00382	0	0	0	0	0	0	1	-360	360;
	4896	7632	0.00524	0.02276	0	0	0	0	0	0	1	-360	360;
	3736	6625	0.002719	0.01143	0	0	0	0	0	0	1	-360	360;
	739	8561	0.00294	0.02744	0	954	0	0	0	0	1	-360	360;
	739	7395	0.0011	0.01339	0	1020	0	0	0	0	1	-360	360;
	739	5873	0.00193	0.01695	0	1349	0	0	0	0	1	-360	360;
	739	4843	0.00134	0.01054	0	1020	0	0	0	0	1	-360	360;
	6092	739	0.00245	0.02064	0	1382	0	0	0	0	1	-360	360;
	1393	6717	0.001841	0.00901	0	457	0	0	0	0	1	-360	360;
	1393	6717	0.00187	0.00882	0	495	0	0	0	0	1	-360	360;
	8337	807	0.00197	0.02312	0	1283	0	0	0	0	1	-360	360;
	2235	7395	0.00078	0.00887	0	1382	0	0	0	0	1	-360	360;
	8512	5435	0.003771	0.0288	0	438	0	0	0	0	1	-360	360;
	8512	4658	0.000909	0.00737	0	457	0	0	0	0	1	-360	360;
	7929	7460	0.00086	0.00783	0	1744	0	0	0	0	1	-360	360;
	7929	7460	0.00095	0.00855	0	1547	0	0	0	0	1	-360	360;
	299	7460	0.00012	0.00157	0	1843	0	0	0	0	1	-360	360;
	299	7460	0.00013	0.00154	0	757	0	0	0	0	1	-360	360;
	5117	6456	0.00089	0.004771	0	533	0	0	0	0	1	-360	360;
	5117	6394	0.002	0.00881	0	553	0	0	0	0	1	-360	360;
	6397	7075	0.007711	0.035331	0	0	0	0	0	0	1	-360	360;
	7075	6936	0.005919	0.030419	0	495	0	0	0	0	1	-360	360;
	7075	2946	0.0053	0.031719	0	591	0	0	0	0	1	-360	360;
	4274	1969	0.00025	0.001289	0	0	0	0	0	0	1	-360	360;
	2946	2116	0.0062	0.040039	0	0	0	0	0	0	1	-360	360;
	5096	5464	0.0008	0.00707	0	1514	0	0	0	0	1	-360	360;
	5096	4957	0.0002	0.00207	0	954	0	0	0	0	1	-360	360;
	5096	4446	0.00036	0.00322	0	1053	0	0	0	0	1	-360	360;
	1115	6090	0.0023	0.0157	0	381	0	0	0	0	1	-360	360;
	1115	4204	0.003479	0.021159	0	419	0	0	0	0	1	-360	360;
	3966	3428	0.00107	0.01013	0	1382	0	0	0	0	1	-360	360;
	1439	6836	0.00025	0.000909	0	514	0	0	0	0	1	-360	360;
	1439	5892	0.00037	0.00188	0	438	0	0	0	0	1	-360	360;
	9201	3937	0.0003	0.00145	0	419	0	0	0	0	1	-360	360;
	9201	3692	0.001419	0.007539	0	476	0	0	0	0	1	-360	360;
	8561	266	0.0012	0.0122	0	1251	0	0	0	0	1	-360	360;
	8561	2566	0.00397	0.02936	0	1251	0	0	0	0	1	-360	360;
	8561	8806	0.00092	0.01112	0	1185	0	0	0	0	1	-360	360;
	8561	7428	0.00387	0.0452	0	1448	0	0	0	0	1	-360	360;
	8561	1280	0.00432	0.03156	0	1086	0	0	0	0	1	-360	360;
	2730	1133	0.00407	0.03937	0	1251	0	0	0	0	1	-360	360;
	2730	1133	0.00415	0.03856	0	1152	0	0	0	0	1	-360	360;
	6127	5111	0.00869	0.05343	0	572	0	0	0	0	1	-360	360;
	6127	2172	0.0068	0.036961	0	457	0	0	0	0	1	-360	360;
	6127	2658	0.01805	0.095469	0	400	0	0	0	0	1	-360	360;
	7600	7283	0.00045	0.00206	0	533	0	0	0	0	1	-360	360;
	473	7193	0.0095	0.037159	0	419	0	0	0	0	1	-360	360;
	473	5025	0.02106	0.0885	0	457	0	0	0	0	1	-360	360;
	7395	5873	0.00157	0.02318	0	1810	0	0	0	0	1	-360	360;
	7395	3662	0.00032	0.00282	0	1711	0	0	0	0	1	-360	360;
	7395	4931	0.00165	0.01605	0	889	0	0	0	0	1	-360	360;
	5435	489	0.00107	0.00714	0	400	0	0	0	0	1	-360	360;
	5435	489	0.000961	0.00555	0	400	0	0	0	0	1	-360	360;
	5435	8529	0.00074	0.00461	0	0	0	0	0	0	1	-360	360;
	5435	7064	0.00076	0.00451	0	0	0	0	0	0	1	-360	360;
	5435	4058	0.001419	0.01093	0	533	0	0	0	0	1	-360	360;
	5435	4658	0.003729	0.024599	0	476	0	0	0	0	1	-360	360;
	2437	7029	0.00233	0.01992	0	1678	0	0	0	0	1	-360	360;
	2437	7029	0.00238	0.01949	0	1645	0	0	0	0	1	-360	360;
	2437	4795	0.00217	0.0253	0	1744	0	0	0	0	1	-360	360;
	2437	4795	0.00212	0.02579	0	1843	0	0	0	0	1	-360	360;
	2437	1133	0.00458	0.04138	0	1086	0	0	0	0	1	-360	360;
	3792	550	0.01555	0.089841	0	381	0	0	0	0	1	-360	360;
	6299	3792	0.00274	0.0168	0	476	0	0	0	0	1	-360	360;
	3792	6879	0.005831	0.032211	0	438	0	0	0	0	1	-360	360;
	3792	4181	0.006669	0.031341	0	476	0	0	0	0	1	-360	360;
	8702	5163	0.01081	0.0538	0	476	0	0	0	0	1	-360	360;
	8702	7861	0.005581	0.02457	0	495	0	0	0	0	1	-360	360;
	8702	7861	0.00569	0.024039	0	533	0	0	0	0	1	-360	360;
	957	7446	0.002081	0.01294	0	457	0	0	0	0	1	-360	360;
	957	4897	0.001599	0.00919	0	0	0	0	0	0	1	-360	360;
	5239	5873	0.00083	0.00768	0	1744	0	0	0	0	1	-360	360;
	8438	8147	0.004979	0.024099	0	0	0	0	0	0	1	-360	360;
	8438	8147	0.005081	0.023591	0	476	0	0	0	0	1	-360	360;
	4290	266	0.00412	0.03917	0	1283	0	0	0	0	1	-360	360;
	266	4843	0.00086	0.00798	0	1119	0	0	0	0	1	-360	360;
	3275	266	0.00087	0.00961	0	1218	0	0	0	0	1	-360	360;
	8247	266	0.00095	0.01034	0	1251	0	0	0	0	1	-360	360;
	8374	266	0.00211	0.0196	0	1119	0	0	0	0	1	-360	360;
	7193	5025	0.00963	0.05281	0	0	0	0	0	0	1	-360	360;
	1064	3734	0.00018	0.001349	0	553	0	0	0	0	1	-360	360;
	1064	6456	0.000961	0.006021	0	381	0	0	0	0	1	-360	360;
	1064	6456	0.000979	0.00589	0	381	0	0	0	0	1	-360	360;
	7877	3063	0.000461	0.002729	0	610	0	0	0	0	1	-360	360;
	986	3367	0.005039	0.01901	0	381	0	0	0	0	1	-360	360;
	986	4603	0.00812	0.028919	0	419	0	0	0	0	1	-360	360;
	7257	6970	0.00106	0.01005	0	954	0	0	0	0	1	-360	360;
	5873	6768	0.00074	0.0081	0	1218	0	0	0	0	1	-360	360;
	8147	6174	0.012961	0.03605	0	0	0	0	0	0	1	-360	360;
	8147	6174	0.00844	0.0378	0	400	0	0	0	0	1	-360	360;
	8147	5162	0.000591	0.004169	0	553	0	0	0	0	1	-360	360;
	6879	6038	0.003091	0.012289	0	457	0	0	0	0	1	-360	360;
	4556	4592	0.000969	0.0053	0	553	0	0	0	0	1	-360	360;
	2412	2686	0.00161	0.012919	0	0	0	0	0	0	1	-360	360;
	2412	2686	0.003219	0.01795	0	0	0	0	0	0	1	-360	360;
	2412	3644	0.004331	0.025	0	0	0	0	0	0	1	-360	360;
	2412	3644	0.004409	0.02445	0	0	0	0	0	0	1	-360	360;
	4855	2551	0.00425	0.02014	0	0	0	0	0	0	1	-360	360;
	4855	6904	0.009531	0.060979	0	495	0	0	0	0	1	-360	360;
	4855	1897	0.002409	0.0135	0	0	0	0	0	0	1	-360	360;
	3268	8210	0.0006	0.00662	0	1678	0	0	0	0	1	-360	360;
	3268	2849	0.00173	0.01701	0	1678	0	0	0	0	1	-360	360;
	6397	1173	0.004289	0.018531	0	438	0	0	0	0	1	-360	360;
	6397	6936	0.006919	0.039789	0	419	0	0	0	0	1	-360	360;
	6397	6936	0.007591	0.035539	0	495	0	0	0	0	1	-360	360;
	6397	4750	0.00961	0.05225	0	419	0	0	0	0	1	-360	360;
	3665	6544	0.000581	0.004599	0	0	0	0	0	0	1	-360	360;
	3665	6544	0.00057	0.00469	0	0	0	0	0	0	1	-360	360;
	3665	6544	0.00087	0.003159	0	591	0	0	0	0	1	-360	360;
	6436	5217	0.00018	0.001419	0	0	0	0	0	0	1	-360	360;
	6436	6717	0.00195	0.018841	0	0	0	0	0	0	1	-360	360;
	72	6093	0.00293	0.01293	0	400	0	0	0	0	1	-360	360;
	1520	3233	0.001969	0.010599	0	438	0	0	0	0	1	-360	360;
	6985	854	3.9e-05	0.000591	0	0	0	0	0	0	1	-360	360;
	6985	415	0.00364	0.02431	0	0	0	0	0	0	1	-360	360;
	6985	1501	0.00114	0.007211	0	0	0	0	0	0	1	-360	360;
	489	4238	0.001271	0.008659	0	381	0	0	0	0	1	-360	360;
	489	1411	0.00105	0.00668	0	0	0	0	0	0	1	-360	360;
	4290	6681	0.00011	0.00122	0	987	0	0	0	0	1	-360	360;
	4290	3587	0.00117	0.0124	0	1349	0	0	0	0	1	-360	360;
	4290	8374	0.00168	0.01891	0	1185	0	0	0	0	1	-360	360;
	797	7819	0.00132	0.008281	0	572	0	0	0	0	1	-360	360;
	797	8378	0.004531	0.01924	0	0	0	0	0	0	1	-360	360;
	797	1241	0.007841	0.041909	0	419	0	0	0	0	1	-360	360;
	7929	7755	0.00113	0.01235	0	1777	0	0	0	0	1	-360	360;
	7929	7755	0.00128	0.0121	0	1744	0	0	0	0	1	-360	360;
	7929	8625	0.0015	0.01792	0	1415	0	0	0	0	1	-360	360;
	4592	1471	0.00111	0.005281	0	0	0	0	0	0	1	-360	360;
	4592	1471	0.00113	0.005169	0	0	0	0	0	0	1	-360	360;
	6040	3111	0.00563	0.03255	0	476	0	0	0	0	1	-360	360;
	1175	3111	0.005289	0.022539	0	0	0	0	0	0	1	-360	360;
	3111	6239	0.007159	0.05086	0	400	0	0	0	0	1	-360	360;
	3111	3689	0.009669	0.05649	0	0	0	0	0	0	1	-360	360;
	3111	6179	0.0092	0.052169	0	553	0	0	0	0	1	-360	360;
	8396	5091	0.00197	0.01594	0	1316	0	0	0	0	1	-360	360;
	8396	8806	0.00343	0.02573	0	1251	0	0	0	0	1	-360	360;
	8396	8247	0.00151	0.01485	0	1185	0	0	0	0	1	-360	360;
	8396	5060	0.00092	0.01042	0	1020	0	0	0	0	1	-360	360;
	8396	8012	0.00029	0.00336	0	1547	0	0	0	0	1	-360	360;
	1491	4489	0.001711	0.01255	0	0	0	0	0	0	1	-360	360;
	1491	1503	0.005	0.03962	0	553	0	0	0	0	1	-360	360;
	849	1168	0.0014	0.01687	0	1251	0	0	0	0	1	-360	360;
	849	1168	0.00067	0.00678	0	1942	0	0	0	0	1	-360	360;
	849	1168	0.00065	0.00692	0	1678	0	0	0	0	1	-360	360;
	6092	849	0.00137	0.01524	0	1218	0	0	0	0	1	-360	360;
	3840	686	0.00424	0.02376	0	0	0	0	0	0	1	-360	360;
	3840	5821	0.005591	0.031711	0	476	0	0	0	0	1	-360	360;
	3840	2225	0.00138	0.00649	0	419	0	0	0	0	1	-360	360;
	3840	2225	0.00107	0.00544	0	514	0	0	0	0	1	-360	360;
	5091	4390	3e-05	0.00041	0	1777	0	0	0	0	1	-360	360;
	5091	4390	3e-05	0.00041	0	1613	0	0	0	0	1	-360	360;
	5091	4957	0.00044	0.00342	0	1316	0	0	0	0	1	-360	360;
	7446	2504	0.003841	0.01737	0	438	0	0	0	0	1	-360	360;
	1741	1912	0.001771	0.01337	0	495	0	0	0	0	1	-360	360;
	483	2658	0.00139	0.007349	0	533	0	0	0	0	1	-360	360;
	483	5592	0.00625	0.03463	0	533	0	0	0	0	1	-360	360;
	5111	4360	0.00055	0.00255	0	0	0	0	0	0	1	-360	360;
	6596	3512	0.00489	0.043331	0	0	0	0	0	0	1	-360	360;
	8701	8207	0.00145	0.013099	0	381	0	0	0	0	1	-360	360;
	8701	1271	0.004039	0.02881	0	0	0	0	0	0	1	-360	360;
	2236	4967	0.002219	0.00988	0	438	0	0	0	0	1	-360	360;
	8575	7845	0.001919	0.0153	0	400	0	0	0	0	1	-360	360;
	8664	1399	0.00189	0.01319	0	1382	0	0	0	0	1	-360	360;
	5202	8936	0.00011	0.000659	0	495	0	0	0	0	1	-360	360;
	5202	2564	0.00011	0.00064	0	457	0	0	0	0	1	-360	360;
	5202	3027	0.000211	0.00111	0	0	0	0	0	0	1	-360	360;
	6836	6209	0.000469	0.001669	0	400	0	0	0	0	1	-360	360;
	5923	854	6e-05	0.000849	0	0	0	0	0	0	1	-360	360;
	7137	7310	0.00051	0.00597	0	1580	0	0	0	0	1	-360	360;
	7137	6970	0.00119	0.01087	0	1218	0	0	0	0	1	-360	360;
	5831	7137	0.00219	0.01922	0	1382	0	0	0	0	1	-360	360;
	5919	2764	0.00051	0.00465	0	1580	0	0	0	0	1	-360	360;
	8871	8819	0.00055	0.0067	0	0	0	0	0	0	1	-360	360;
	2504	5329	0.005099	0.027349	0	553	0	0	0	0	1	-360	360;
	3233	2844	0.000599	0.003961	0	476	0	0	0	0	1	-360	360;
	3233	3291	0.005461	0.032719	0	476	0	0	0	0	1	-360	360;
	8149	2896	0.00349	0.02256	0	476	0	0	0	0	1	-360	360;
	8149	1802	0.00132	0.00861	0	400	0	0	0	0	1	-360	360;
	5828	440	0.0045	0.01838	0	0	0	0	0	0	1	-360	360;
	5828	1391	0.004289	0.019719	0	400	0	0	0	0	1	-360	360;
	8236	3027	0.001409	0.00881	0	495	0	0	0	0	1	-360	360;
	1168	6607	0.00212	0.01857	0	1415	0	0	0	0	1	-360	360;
	1168	7673	0.00049	0.00574	0	1613	0	0	0	0	1	-360	360;
	1168	754	0.00184	0.021	0	1547	0	0	0	0	1	-360	360;
	2686	4773	0.00593	0.037409	0	419	0	0	0	0	1	-360	360;
	2225	4177	0.00514	0.024169	0	438	0	0	0	0	1	-360	360;
	3798	2913	0.00041	0.00492	0	1251	0	0	0	0	1	-360	360;
	3544	2913	0.00296	0.02677	0	1349	0	0	0	0	1	-360	360;
	4906	550	0.0072	0.03406	0	381	0	0	0	0	1	-360	360;
	3427	6949	0.00043	0.00276	0	419	0	0	0	0	1	-360	360;
	3427	9028	0.001789	0.013469	0	533	0	0	0	0	1	-360	360;
	4890	7406	0.00163	0.011091	0	457	0	0	0	0	1	-360	360;
	4890	6349	0.00036	0.002651	0	0	0	0	0	0	1	-360	360;
	8867	15	0.00013	0.00132	0	514	0	0	0	0	1	-360	360;
	8867	284	0.000169	0.001659	0	457	0	0	0	0	1	-360	360;
	4428	8534	0.00136	0.012831	0	0	0	0	0	0	1	-360	360;
	4428	8534	0.00139	0.01256	0	0	0	0	0	0	1	-360	360;
	4428	5276	0.00126	0.009521	0	0	0	0	0	0	1	-360	360;
	4428	5276	0.00143	0.009331	0	0	0	0	0	0	1	-360	360;
	3644	4428	0.00164	0.012961	0	0	0	0	0	0	1	-360	360;
	3644	4428	0.001599	0.012	0	0	0	0	0	0	1	-360	360;
	6936	2957	0.001271	0.00524	0	438	0	0	0	0	1	-360	360;
	8240	8973	0.00226	0.02465	0	1514	0	0	0	0	1	-360	360;
	6543	3715	0.00032	0.0015	0	514	0	0	0	0	1	-360	360;
	6543	5948	0.00032	0.00112	0	419	0	0	0	0	1	-360	360;
	3275	4931	0.00075	0.00817	0	1053	0	0	0	0	1	-360	360;
	8163	3425	0.00037	0.0027	0	553	0	0	0	0	1	-360	360;
	3425	6783	3.9e-05	0.000219	0	0	0	0	0	0	1	-360	360;
	2172	595	0.012781	0.05762	0	0	0	0	0	0	1	-360	360;
	8658	1065	0.00036	0.001289	0	0	0	0	0	0	1	-360	360;
	6090	1175	0.006771	0.039349	0	0	0	0	0	0	1	-360	360;
	3106	7673	0.00292	0.02831	0	1415	0	0	0	0	1	-360	360;
	3106	2096	0.00134	0.01229	0	1514	0	0	0	0	1	-360	360;
	8163	5500	0.00011	0.00063	0	572	0	0	0	0	1	-360	360;
	7249	8207	0.0003	0.003331	0	0	0	0	0	0	1	-360	360;
	7249	3692	0.001289	0.01555	0	0	0	0	0	0	1	-360	360;
	6783	5923	0.00011	0.0017	0	553	0	0	0	0	1	-360	360;
	6783	8575	0.00043	0.00249	0	533	0	0	0	0	1	-360	360;
	1121	4657	0.002651	0.016771	0	0	0	0	0	0	1	-360	360;
	3487	1121	0.0082	0.025081	0	400	0	0	0	0	1	-360	360;
	3442	8819	0.00037	0.002781	0	591	0	0	0	0	1	-360	360;
	4856	4906	0.010719	0.054719	0	400	0	0	0	0	1	-360	360;
	3076	4906	0.006469	0.04074	0	400	0	0	0	0	1	-360	360;
	5884	8973	0.00101	0.01274	0	1283	0	0	0	0	1	-360	360;
	8207	5961	9.1e-05	0.05076	0	419	0	0	0	0	1	-360	360;
	8207	3629	0.00588	0.03938	0	457	0	0	0	0	1	-360	360;
	8207	3692	0.003591	0.022099	0	438	0	0	0	0	1	-360	360;
	8207	7630	0.00555	0.032521	0	0	0	0	0	0	1	-360	360;
	6877	7755	0.00116	0.01347	0	1514	0	0	0	0	1	-360	360;
	6877	7755	0.00113	0.01247	0	1547	0	0	0	0	1	-360	360;
	5042	179	0.00107	0.00871	0	954	0	0	0	0	1	-360	360;
	5042	4442	0.00078	0.00609	0	1382	0	0	0	0	1	-360	360;
	2566	6768	0.00075	0.00604	0	1119	0	0	0	0	1	-360	360;
	4897	6717	0.00224	0.011091	0	0	0	0	0	0	1	-360	360;
	7475	6949	0.001961	0.013969	0	400	0	0	0	0	1	-360	360;
	1294	5162	0.001271	0.00761	0	476	0	0	0	0	1	-360	360;
	6681	8806	0.00374	0.03581	0	1382	0	0	0	0	1	-360	360;
	6681	8806	0.00385	0.03284	0	1316	0	0	0	0	1	-360	360;
	6681	8806	0.00392	0.03216	0	1185	0	0	0	0	1	-360	360;
	6681	8973	0.00307	0.03131	0	1283	0	0	0	0	1	-360	360;
	7819	32	0.027	0.131289	0	438	0	0	0	0	1	-360	360;
	6333	5512	0.005539	0.0195	0	0	0	0	0	0	1	-360	360;
	8534	9039	0.003099	0.017849	0	495	0	0	0	0	1	-360	360;
	8534	5276	0.002831	0.014099	0	533	0	0	0	0	1	-360	360;
	4026	2261	0	0.01368	0	1580	0	0	0	0	1	-360	360;
	4204	4657	0.01082	0.079719	0	0	0	0	0	0	1	-360	360;
	731	2764	0.00102	0.00941	0	1251	0	0	0	0	1	-360	360;
	731	597	0.00183	0.01819	0	1777	0	0	0	0	1	-360	360;
	731	385	0.00028	0.00346	0	1251	0	0	0	0	1	-360	360;
	3728	2564	0.0012	0.00982	0	0	0	0	0	0	1	-360	360;
	2764	2622	0.00027	0.0032	0	1810	0	0	0	0	1	-360	360;
	2764	7688	0.00064	0.00679	0	1152	0	0	0	0	1	-360	360;
	2764	8260	0.00047	0.00442	0	1711	0	0	0	0	1	-360	360;
	415	3367	0.00493	0.017599	0	495	0	0	0	0	1	-360	360;
	415	4081	0.00207	0.0153	0	0	0	0	0	0	1	-360	360;
	7287	7755	0.00094	0.0092	0	1909	0	0	0	0	1	-360	360;
	7287	7755	0.00095	0.00901	0	1711	0	0	0	0	1	-360	360;
	947	2785	0.00101	0.00719	0	0	0	0	0	0	1	-360	360;
	3734	6654	0.000419	0.00331	0	572	0	0	0	0	1	-360	360;
	2551	1897	0.002099	0.015599	0	0	0	0	0	0	1	-360	360;
	8806	8664	0.00163	0.01283	0	1283	0	0	0	0	1	-360	360;
	8806	7428	0.00445	0.03788	0	1514	0	0	0	0	1	-360	360;
	8806	4652	0.00092	0.00703	0	1218	0	0	0	0	1	-360	360;
	249	32	9.1e-05	0.000961	0	0	0	0	0	0	1	-360	360;
	1741	249	0.00387	0.028599	0	0	0	0	0	0	1	-360	360;
	249	5290	0.00949	0.06625	0	0	0	0	0	0	1	-360	360;
	249	4467	0.00414	0.035331	0	0	0	0	0	0	1	-360	360;
	249	5290	0.009781	0.065581	0	0	0	0	0	0	1	-360	360;
	1391	32	0.02818	0.116099	0	381	0	0	0	0	1	-360	360;
	32	2143	0.00324	0.01912	0	381	0	0	0	0	1	-360	360;
	3008	6768	0.00135	0.01609	0	1613	0	0	0	0	1	-360	360;
	6174	5670	0.00294	0.0182	0	438	0	0	0	0	1	-360	360;
	8066	5961	0.000711	0.00649	0	457	0	0	0	0	1	-360	360;
	442	8066	0.002651	0.01375	0	0	0	0	0	0	1	-360	360;
	1271	5961	0.00105	0.008469	0	419	0	0	0	0	1	-360	360;
	869	4489	0.002229	0.016461	0	495	0	0	0	0	1	-360	360;
	869	4489	0.002281	0.01613	0	495	0	0	0	0	1	-360	360;
	602	4489	0.003341	0.012651	0	0	0	0	0	0	1	-360	360;
	602	4489	0.003271	0.012901	0	0	0	0	0	0	1	-360	360;
	6484	7918	0.00101	0.01218	0	1843	0	0	0	0	1	-360	360;
	6484	7436	0.0009	0.0095	0	0	0	0	0	0	1	-360	360;
	7918	4690	0.00098	0.01241	0	0	0	0	0	0	1	-360	360;
	5511	1942	0.001219	0.00951	0	0	0	0	0	0	1	-360	360;
	5511	8419	0.001789	0.01382	0	0	0	0	0	0	1	-360	360;
	5511	8419	0.00175	0.014081	0	0	0	0	0	0	1	-360	360;
	7611	3382	9.1e-05	0.00087	0	0	0	0	0	0	1	-360	360;
	7611	3382	9.1e-05	0.00086	0	0	0	0	0	0	1	-360	360;
	7611	7869	0.000229	0.00236	0	400	0	0	0	0	1	-360	360;
	686	5171	0.00125	0.00557	0	591	0	0	0	0	1	-360	360;
	2172	5191	0.00757	0.04013	0	0	0	0	0	0	1	-360	360;
	1697	179	0.00157	0.01604	0	1382	0	0	0	0	1	-360	360;
	1697	179	0.00154	0.01634	0	1283	0	0	0	0	1	-360	360;
	179	2096	0.00169	0.02383	0	1547	0	0	0	0	1	-360	360;
	1728	179	0.00068	0.0089	0	1547	0	0	0	0	1	-360	360;
	1850	4603	0.010289	0.036651	0	476	0	0	0	0	1	-360	360;
	1850	4603	0.00563	0.03395	0	0	0	0	0	0	1	-360	360;
	3109	5726	0.003591	0.02763	0	0	0	0	0	0	1	-360	360;
	3109	5726	0.00351	0.028159	0	0	0	0	0	0	1	-360	360;
	5726	4773	0.00143	0.00838	0	0	0	0	0	0	1	-360	360;
	7659	6879	0.01213	0.06156	0	476	0	0	0	0	1	-360	360;
	6174	5171	0.004789	0.022849	0	553	0	0	0	0	1	-360	360;
	527	8242	0.00019	0.00211	0	457	0	0	0	0	1	-360	360;
	527	8529	0.000331	0.002349	0	0	0	0	0	0	1	-360	360;
	527	8530	9.1e-05	0.000539	0	381	0	0	0	0	1	-360	360;
	6496	6607	0.00088	0.00776	0	1382	0	0	0	0	1	-360	360;
	4652	4706	0.00039	0.00368	0	0	0	0	0	0	1	-360	360;
	6656	2091	0.00275	0.01268	0	400	0	0	0	0	1	-360	360;
	6656	8993	0.00205	0.00874	0	0	0	0	0	0	1	-360	360;
	6656	1241	0.00082	0.00519	0	591	0	0	0	0	1	-360	360;
	2658	808	0.00519	0.02355	0	457	0	0	0	0	1	-360	360;
	3291	6749	0.016711	0.08874	0	419	0	0	0	0	1	-360	360;
	7406	4397	0.005099	0.022659	0	476	0	0	0	0	1	-360	360;
	7406	1769	0.001469	0.01213	0	591	0	0	0	0	1	-360	360;
	7406	1769	0.001659	0.01189	0	533	0	0	0	0	1	-360	360;
	1013	4467	0.000521	0.00374	0	419	0	0	0	0	1	-360	360;
	1013	5025	0.006461	0.048349	0	0	0	0	0	0	1	-360	360;
	7918	7310	0.00293	0.0257	0	1316	0	0	0	0	1	-360	360;
	7918	5608	0.00178	0.0157	0	1415	0	0	0	0	1	-360	360;
	5831	7918	0.00115	0.01332	0	1086	0	0	0	0	1	-360	360;
	7918	2202	0.00078	0.00752	0	1678	0	0	0	0	1	-360	360;
	77	6660	3.9e-05	0.00037	0	0	0	0	0	0	1	-360	360;
	8124	8419	0.0012	0.006729	0	533	0	0	0	0	1	-360	360;
	33	3334	0.000211	0.001659	0	0	0	0	0	0	1	-360	360;
	6490	8301	0.001659	0.011919	0	495	0	0	0	0	1	-360	360;
	6490	6869	0.002	0.0133	0	438	0	0	0	0	1	-360	360;
	4371	7820	0.00105	0.00837	0	0	0	0	0	0	1	-360	360;
	2896	5348	0.00138	0.014979	0	0	0	0	0	0	1	-360	360;
	5767	6836	0	0.0003	0	0	0	0	0	0	1	-360	360;
	2978	6836	0	0.0003	0	419	0	0	0	0	1	-360	360;
	2978	5391	0.00164	0.00782	0	400	0	0	0	0	1	-360	360;
	3629	230	0.007099	0.04044	0	514	0	0	0	0	1	-360	360;
	9039	7861	0.00181	0.00914	0	572	0	0	0	0	1	-360	360;
	1479	7343	0.00027	0.00247	0	1942	0	0	0	0	1	-360	360;
	7343	4957	0.00266	0.03206	0	1645	0	0	0	0	1	-360	360;
	8338	981	0.000591	0.005081	0	400	0	0	0	0	1	-360	360;
	8338	751	0.002229	0.016099	0	419	0	0	0	0	1	-360	360;
	4026	5372	0.00205	0.02585	0	1086	0	0	0	0	1	-360	360;
	7821	642	0.00101	0.0091	0	1316	0	0	0	0	1	-360	360;
	7821	5608	0.001	0.00871	0	1316	0	0	0	0	1	-360	360;
	9097	1513	0.00112	0.00895	0	0	0	0	0	0	1	-360	360;
	9097	1513	0.001099	0.008289	0	0	0	0	0	0	1	-360	360;
	9097	3125	0.001289	0.0095	0	0	0	0	0	0	1	-360	360;
	9097	3125	0.00126	0.00968	0	0	0	0	0	0	1	-360	360;
	4315	5464	0.00065	0.00723	0	1547	0	0	0	0	1	-360	360;
	1942	4095	0.001781	0.012669	0	0	0	0	0	0	1	-360	360;
	3966	1074	0.00086	0.00956	0	1053	0	0	0	0	1	-360	360;
	3966	807	0.00248	0.02429	0	1283	0	0	0	0	1	-360	360;
	3966	4507	0.00151	0.01699	0	1316	0	0	0	0	1	-360	360;
	5890	3689	0.003159	0.01537	0	0	0	0	0	0	1	-360	360;
	8210	3113	0.00075	0.00683	0	1349	0	0	0	0	1	-360	360;
	8210	2849	0.00223	0.02524	0	1613	0	0	0	0	1	-360	360;
	642	5608	0.00191	0.0208	0	987	0	0	0	0	1	-360	360;
	1697	642	0.00074	0.00784	0	921	0	0	0	0	1	-360	360;
	1697	642	0.00073	0.00799	0	1086	0	0	0	0	1	-360	360;
	1421	3621	0.000789	0.00482	0	514	0	0	0	0	1	-360	360;
	2126	1421	0.000789	0.004409	0	381	0	0	0	0	1	-360	360;
	2844	3291	0.005461	0.03161	0	533	0	0	0	0	1	-360	360;
	1638	8575	6e-05	0.00087	0	0	0	0	0	0	1	-360	360;
	3163	1411	0.00593	0.040961	0	0	0	0	0	0	1	-360	360;
	3163	5025	0.003771	0.030841	0	0	0	0	0	0	1	-360	360;
	6454	8571	0.00038	0.001461	0	533	0	0	0	0	1	-360	360;
	8571	4081	0.00149	0.00938	0	0	0	0	0	0	1	-360	360;
	8571	4081	0.001521	0.00918	0	0	0	0	0	0	1	-360	360;
	2091	440	0.00407	0.015469	0	0	0	0	0	0	1	-360	360;
	4953	1421	0.00082	0.006219	0	0	0	0	0	0	1	-360	360;
	2667	1969	0.00161	0.00888	0	438	0	0	0	0	1	-360	360;
	2667	9028	0.00118	0.006461	0	514	0	0	0	0	1	-360	360;
	8819	6134	0.00082	0.009091	0	0	0	0	0	0	1	-360	360;
	751	4493	0.01893	0.0848	0	457	0	0	0	0	1	-360	360;
	751	4493	0.01313	0.082409	0	438	0	0	0	0	1	-360	360;
	751	6344	0.000281	0.00149	0	0	0	0	0	0	1	-360	360;
	5289	3113	0.00019	0.00199	0	1349	0	0	0	0	1	-360	360;
	3113	5289	0.0002	0.00188	0	1382	0	0	0	0	1	-360	360;
	172	2946	0.003159	0.025331	0	0	0	0	0	0	1	-360	360;
	7685	6394	0.00445	0.01886	0	610	0	0	0	0	1	-360	360;
	5821	5329	0.00461	0.02205	0	514	0	0	0	0	1	-360	360;
	7283	8213	0.00061	0.00494	0	0	0	0	0	0	1	-360	360;
	7283	962	0.00082	0.00462	0	0	0	0	0	0	1	-360	360;
	3468	8661	0.000331	0.003271	0	553	0	0	0	0	1	-360	360;
	3468	1769	0.00056	0.004081	0	591	0	0	0	0	1	-360	360;
	6244	9055	0.00163	0.007521	0	0	0	0	0	0	1	-360	360;
	6244	1471	0.001659	0.00736	0	0	0	0	0	0	1	-360	360;
	6544	1503	0.00719	0.049211	0	0	0	0	0	0	1	-360	360;
	7463	1969	6e-05	0.00039	0	438	0	0	0	0	1	-360	360;
	1030	6022	0.00288	0.018	0	533	0	0	0	0	1	-360	360;
	1030	8776	0.00288	0.0182	0	495	0	0	0	0	1	-360	360;
	1030	5290	0.00111	0.008531	0	438	0	0	0	0	1	-360	360;
	1030	4279	0.00307	0.022289	0	514	0	0	0	0	1	-360	360;
	3067	4507	0.00091	0.00868	0	1547	0	0	0	0	1	-360	360;
	5608	9117	0.00098	0.0118	0	1481	0	0	0	0	1	-360	360;
	5608	7436	0.00053	0.00591	0	1645	0	0	0	0	1	-360	360;
	5608	2202	0.00086	0.00991	0	1349	0	0	0	0	1	-360	360;
	7002	3334	0.00545	0.034099	0	495	0	0	0	0	1	-360	360;
	2060	3495	0.00095	0.006091	0	553	0	0	0	0	1	-360	360;
	3544	807	0.00305	0.0272	0	1316	0	0	0	0	1	-360	360;
	869	4910	0.000281	0.001841	0	438	0	0	0	0	1	-360	360;
	869	6456	0.004531	0.02451	0	0	0	0	0	0	1	-360	360;
	4397	1503	0.00618	0.02961	0	533	0	0	0	0	1	-360	360;
	1513	6745	0.00064	0.004461	0	0	0	0	0	0	1	-360	360;
	916	4224	0.00143	0.00513	0	553	0	0	0	0	1	-360	360;
	916	1241	0.0052	0.021841	0	457	0	0	0	0	1	-360	360;
	602	4910	0.00138	0.008669	0	438	0	0	0	0	1	-360	360;
	6456	602	0.00314	0.013169	0	0	0	0	0	0	1	-360	360;
	4224	1909	0.002479	0.01331	0	476	0	0	0	0	1	-360	360;
	2297	2740	7e-05	0.00194	0	0	0	0	0	0	1	-360	360;
	9162	8419	0.001159	0.00931	0	0	0	0	0	0	1	-360	360;
	5391	6349	0.00062	0.003669	0	0	0	0	0	0	1	-360	360;
	2517	5163	0.015969	0.056711	0	629	0	0	0	0	1	-360	360;
	5163	3754	0.0085	0.03719	0	419	0	0	0	0	1	-360	360;
	2785	9055	0.001531	0.00732	0	0	0	0	0	0	1	-360	360;
	2785	9055	0.00205	0.01062	0	0	0	0	0	0	1	-360	360;
	4957	5464	0.00091	0.01056	0	757	0	0	0	0	1	-360	360;
	807	8625	0.00167	0.02069	0	1547	0	0	0	0	1	-360	360;
	6299	6879	0.002719	0.016219	0	514	0	0	0	0	1	-360	360;
	8247	4390	0.002	0.02376	0	1711	0	0	0	0	1	-360	360;
	8247	4390	0.00226	0.02328	0	1645	0	0	0	0	1	-360	360;
	7630	442	0.004151	0.02119	0	533	0	0	0	0	1	-360	360;
	8993	8378	0.001841	0.010909	0	0	0	0	0	0	1	-360	360;
	6879	4181	0.000669	0.00399	0	476	0	0	0	0	1	-360	360;
	3662	1475	8e-05	0.00098	0	1613	0	0	0	0	1	-360	360;
	8242	8530	0.0002	0.0025	0	0	0	0	0	0	1	-360	360;
	5494	5290	0.006219	0.031599	0	533	0	0	0	0	1	-360	360;
	5494	1437	0.00438	0.023169	0	533	0	0	0	0	1	-360	360;
	6145	4279	0.00581	0.03088	0	514	0	0	0	0	1	-360	360;
	6145	4279	0.0065	0.02637	0	438	0	0	0	0	1	-360	360;
	4652	597	0.00198	0.02208	0	1678	0	0	0	0	1	-360	360;
	3109	5329	0.00093	0.006659	0	0	0	0	0	0	1	-360	360;
	3109	5329	0.000909	0.006789	0	0	0	0	0	0	1	-360	360;
	4026	1074	0.00244	0.01919	0	856	0	0	0	0	1	-360	360;
	2355	8582	0.00194	0.01549	0	0	0	0	0	0	1	-360	360;
	1666	4058	0.0002	0.001479	0	514	0	0	0	0	1	-360	360;
	4493	1513	0.001091	0.009461	0	610	0	0	0	0	1	-360	360;
	1513	6604	0.003081	0.00899	0	495	0	0	0	0	1	-360	360;
	6604	1457	0.00357	0.0138	0	457	0	0	0	0	1	-360	360;
	3495	7629	0.001229	0.00986	0	476	0	0	0	0	1	-360	360;
	1985	9106	0.001211	0.00913	0	572	0	0	0	0	1	-360	360;
	4957	8012	0.00116	0.01053	0	1580	0	0	0	0	1	-360	360;
	4651	5681	0.0017	0.009781	0	0	0	0	0	0	1	-360	360;
	4651	5681	0.001521	0.011091	0	0	0	0	0	0	1	-360	360;
	7795	445	0.0043	0.0233	0	572	0	0	0	0	1	-360	360;
	2700	4446	0.00019	0.0016	0	987	0	0	0	0	1	-360	360;
	4772	8301	0.00838	0.072479	0	0	0	0	0	0	1	-360	360;
	4772	988	0.00787	0.049211	0	0	0	0	0	0	1	-360	360;
	442	230	0.00557	0.03376	0	0	0	0	0	0	1	-360	360;
	4238	7312	0.01213	0.06564	0	0	0	0	0	0	1	-360	360;
	7366	3125	0.003531	0.01086	0	400	0	0	0	0	1	-360	360;
	7366	7423	0.0158	0.057031	0	0	0	0	0	0	1	-360	360;
	754	2096	0.00289	0.0244	0	1283	0	0	0	0	1	-360	360;
	1599	4096	0.000521	0.00413	0	553	0	0	0	0	1	-360	360;
	8079	4096	0.000531	0.004039	0	495	0	0	0	0	1	-360	360;
	230	1802	0.00431	0.02226	0	419	0	0	0	0	1	-360	360;
	3027	4081	0.00163	0.00899	0	553	0	0	0	0	1	-360	360;
	4652	1280	0.00274	0.02428	0	1415	0	0	0	0	1	-360	360;
	6315	6022	0.0093	0.056979	0	514	0	0	0	0	1	-360	360;
	6315	8729	0.011281	0.080039	0	0	0	0	0	0	1	-360	360;
	6315	5631	0.00449	0.02376	0	438	0	0	0	0	1	-360	360;
	2622	8260	0.00014	0.00148	0	1744	0	0	0	0	1	-360	360;
	6797	8301	0.0055	0.033659	0	419	0	0	0	0	1	-360	360;
	4467	5465	0.00249	0.02082	0	457	0	0	0	0	1	-360	360;
	1728	2096	0.00164	0.02252	0	1580	0	0	0	0	1	-360	360;
	5793	2147	0.000961	0.0047	0	0	0	0	0	0	1	-360	360;
	5793	2147	0.000979	0.004599	0	0	0	0	0	0	1	-360	360;
	7688	385	0.00057	0.00515	0	1251	0	0	0	0	1	-360	360;
	8301	6869	0.00143	0.011651	0	381	0	0	0	0	1	-360	360;
	1501	4081	0.00126	0.009211	0	0	0	0	0	0	1	-360	360;
	6022	695	9.1e-05	0.000211	0	381	0	0	0	0	1	-360	360;
	8776	695	0.000211	0.000711	0	0	0	0	0	0	1	-360	360;
	8776	5631	0.01095	0.07062	0	457	0	0	0	0	1	-360	360;
	2143	5631	0.00125	0.007281	0	0	0	0	0	0	1	-360	360;
	2143	5631	0.001	0.0045	0	0	0	0	0	0	1	-360	360;
	5465	5025	0.00282	0.026419	0	0	0	0	0	0	1	-360	360;
	7632	6625	0.003479	0.01456	0	0	0	0	0	0	1	-360	360;
	5290	4279	0.005521	0.03418	0	457	0	0	0	0	1	-360	360;
	5290	4279	0.00563	0.033461	0	476	0	0	0	0	1	-360	360;
	5025	4752	0.00095	0.0072	0	419	0	0	0	0	1	-360	360;
	5025	7312	0.00093	0.007159	0	514	0	0	0	0	1	-360	360;
	1437	4279	0.00988	0.058651	0	572	0	0	0	0	1	-360	360;
	6654	7869	3.9e-05	0.00044	0	591	0	0	0	0	1	-360	360;
	6654	7869	3.9e-05	0.00043	0	533	0	0	0	0	1	-360	360;
	7351	5441	0.00018	0.000781	0	0	0	0	0	0	1	-360	360;
	4314	7571	0.002961	0.01669	0	419	0	0	0	0	1	-360	360;
	5803	3916	0.00056	0.00432	0	0	0	0	0	0	1	-360	360;
	6757	6036	0.0002	0.00246	0	757	0	0	0	0	1	-360	360;
	6757	6921	0.0003	0.00377	0	1415	0	0	0	0	1	-360	360;
	5019	9112	0.00014	0.00057	0	0	0	0	0	0	1	-360	360;
	2930	9112	0.00013	0.000729	0	0	0	0	0	0	1	-360	360;
	1436	9112	0.00012	0.000581	0	0	0	0	0	0	1	-360	360;
	1851	9112	0.00011	0.00064	0	0	0	0	0	0	1	-360	360;
	8331	9112	0.00019	0.00056	0	0	0	0	0	0	1	-360	360;
	6220	8791	0.00062	0.004909	0	629	0	0	0	0	1	-360	360;
	9203	7842	7e-05	0.00076	0	987	0	0	0	0	1	-360	360;
	9203	8997	0.00066	0.00622	0	1711	0	0	0	0	1	-360	360;
	9203	2129	0.00054	0.00752	0	0	0	0	0	0	1	-360	360;
	9203	29	0.00105	0.00877	0	1415	0	0	0	0	1	-360	360;
	118	8903	8e-05	0.00091	0	757	0	0	0	0	1	-360	360;
	118	4598	0.00056	0.00719	0	1053	0	0	0	0	1	-360	360;
	2426	5533	0.000531	0.002039	0	419	0	0	0	0	1	-360	360;
	2426	4831	0.008659	0.03769	0	400	0	0	0	0	1	-360	360;
	2426	8961	0.003849	0.021099	0	0	0	0	0	0	1	-360	360;
	2426	8522	0.00688	0.035969	0	400	0	0	0	0	1	-360	360;
	2426	6888	0.006729	0.03713	0	0	0	0	0	0	1	-360	360;
	2426	3401	0.00111	0.006651	0	514	0	0	0	0	1	-360	360;
	2426	641	0.0023	0.008909	0	0	0	0	0	0	1	-360	360;
	2426	2600	0.000841	0.00313	0	0	0	0	0	0	1	-360	360;
	2426	5351	0.003521	0.017479	0	400	0	0	0	0	1	-360	360;
	3051	718	3.9e-05	0.00025	0	610	0	0	0	0	1	-360	360;
	3036	2327	0.000159	0.000729	0	457	0	0	0	0	1	-360	360;
	8651	7473	0.005281	0.032099	0	495	0	0	0	0	1	-360	360;
	8651	7473	0.005651	0.029419	0	476	0	0	0	0	1	-360	360;
	903	3654	0.000271	0.00036	0	0	0	0	0	0	1	-360	360;
	8732	7050	0.00044	0.00143	0	572	0	0	0	0	1	-360	360;
	8732	3912	0.000159	0.00074	0	553	0	0	0	0	1	-360	360;
	2273	1973	0.00019	0.000591	0	438	0	0	0	0	1	-360	360;
	5131	2928	0.00019	0.00087	0	0	0	0	0	0	1	-360	360;
	6629	4918	0.002591	0.013659	0	419	0	0	0	0	1	-360	360;
	6629	7309	0.00305	0.01168	0	0	0	0	0	0	1	-360	360;
	413	1102	0.000159	0.000599	0	533	0	0	0	0	1	-360	360;
	413	1102	0.000159	0.00062	0	495	0	0	0	0	1	-360	360;
	3248	4918	0.00101	0.004229	0	572	0	0	0	0	1	-360	360;
	3248	4281	0.00319	0.018651	0	553	0	0	0	0	1	-360	360;
	3248	7309	0.000729	0.00337	0	0	0	0	0	0	1	-360	360;
	1146	1187	0.002961	0.012831	0	381	0	0	0	0	1	-360	360;
	1146	7945	0.000591	0.00339	0	0	0	0	0	0	1	-360	360;
	8900	2421	0.00024	0.001219	0	0	0	0	0	0	1	-360	360;
	6071	2972	0.006219	0.036341	0	476	0	0	0	0	1	-360	360;
	6071	6031	0.001841	0.01193	0	572	0	0	0	0	1	-360	360;
	3293	7264	0.001409	0.01113	0	514	0	0	0	0	1	-360	360;
	3293	283	0.00151	0.00768	0	476	0	0	0	0	1	-360	360;
	3293	3069	0.00599	0.03313	0	495	0	0	0	0	1	-360	360;
	3293	3390	0.001771	0.010151	0	457	0	0	0	0	1	-360	360;
	4689	4936	0.00074	0.00311	0	0	0	0	0	0	1	-360	360;
	4689	4936	0.000729	0.003159	0	0	0	0	0	0	1	-360	360;
	3218	7809	0.005271	0.01906	0	381	0	0	0	0	1	-360	360;
	3218	5455	0.00906	0.053151	0	419	0	0	0	0	1	-360	360;
	3218	1857	0.00405	0.0153	0	0	0	0	0	0	1	-360	360;
	6806	5410	0.00055	0.003961	0	495	0	0	0	0	1	-360	360;
	6806	5410	0.00055	0.003729	0	438	0	0	0	0	1	-360	360;
	6806	1609	0.00107	0.007789	0	514	0	0	0	0	1	-360	360;
	6806	1609	0.00112	0.00775	0	572	0	0	0	0	1	-360	360;
	6806	1609	0.001031	0.00775	0	0	0	0	0	0	1	-360	360;
	6806	2438	0.00062	0.003219	0	0	0	0	0	0	1	-360	360;
	6806	954	0.00068	0.002659	0	514	0	0	0	0	1	-360	360;
	9140	502	0.00193	0.012479	0	591	0	0	0	0	1	-360	360;
	9140	3013	0.001159	0.007229	0	0	0	0	0	0	1	-360	360;
	9140	2949	0.00137	0.008729	0	495	0	0	0	0	1	-360	360;
	9140	2949	0.001401	0.00855	0	533	0	0	0	0	1	-360	360;
	9140	2949	0.001331	0.008521	0	533	0	0	0	0	1	-360	360;
	7070	2208	0.00031	0.00081	0	381	0	0	0	0	1	-360	360;
	7070	8060	0.000341	0.00119	0	381	0	0	0	0	1	-360	360;
	7070	4484	0.00031	0.001281	0	476	0	0	0	0	1	-360	360;
	7070	4239	0.00094	0.0035	0	533	0	0	0	0	1	-360	360;
	8535	7047	7e-05	0.000651	0	0	0	0	0	0	1	-360	360;
	5477	3657	0.00019	0.00124	0	591	0	0	0	0	1	-360	360;
	5477	2928	0.000331	0.002539	0	533	0	0	0	0	1	-360	360;
	3145	2918	0.0016	0.02113	0	0	0	0	0	0	1	-360	360;
	7770	5317	0.000909	0.004711	0	457	0	0	0	0	1	-360	360;
	7770	5317	0.00088	0.004969	0	419	0	0	0	0	1	-360	360;
	3	8676	0.001841	0.0088	0	0	0	0	0	0	1	-360	360;
	3	8189	0.01013	0.04238	0	495	0	0	0	0	1	-360	360;
	7267	2938	0.00059	0.00802	0	823	0	0	0	0	1	-360	360;
	7267	2938	0.00054	0.0079	0	757	0	0	0	0	1	-360	360;
	7267	3918	0.00171	0.02128	0	1086	0	0	0	0	1	-360	360;
	7267	3918	0.00174	0.02084	0	954	0	0	0	0	1	-360	360;
	7267	6581	0.00086	0.01433	0	1251	0	0	0	0	1	-360	360;
	7267	6581	0.00087	0.01404	0	823	0	0	0	0	1	-360	360;
	7267	3830	0.00126	0.01447	0	1218	0	0	0	0	1	-360	360;
	7267	5648	0.00122	0.01509	0	1152	0	0	0	0	1	-360	360;
	7267	5648	0.00154	0.019	0	1382	0	0	0	0	1	-360	360;
	4918	4907	0.00474	0.03564	0	572	0	0	0	0	1	-360	360;
	4918	5957	0.006	0.047849	0	533	0	0	0	0	1	-360	360;
	4918	5233	0.008039	0.03375	0	381	0	0	0	0	1	-360	360;
	6854	7309	2.1e-05	0.000211	0	610	0	0	0	0	1	-360	360;
	8311	3680	0.00045	0.002341	0	0	0	0	0	0	1	-360	360;
	2770	8804	0.000669	0.004281	0	533	0	0	0	0	1	-360	360;
	9241	8804	0.00068	0.00419	0	0	0	0	0	0	1	-360	360;
	3083	7124	0.000729	0.00431	0	610	0	0	0	0	1	-360	360;
	3083	7124	0.000591	0.004651	0	553	0	0	0	0	1	-360	360;
	3083	8873	0.00099	0.005081	0	438	0	0	0	0	1	-360	360;
	1566	7582	0.00212	0.013031	0	0	0	0	0	0	1	-360	360;
	1566	3541	0.002159	0.01275	0	0	0	0	0	0	1	-360	360;
	8683	3698	0.00025	0.00095	0	0	0	0	0	0	1	-360	360;
	8683	3698	0.000271	0.00089	0	0	0	0	0	0	1	-360	360;
	8683	2801	0.001961	0.012409	0	381	0	0	0	0	1	-360	360;
	8683	8328	0.00044	0.00207	0	572	0	0	0	0	1	-360	360;
	8683	8511	0.000469	0.00143	0	629	0	0	0	0	1	-360	360;
	8683	6570	0.001271	0.01406	0	629	0	0	0	0	1	-360	360;
	8683	6570	0.00124	0.013021	0	629	0	0	0	0	1	-360	360;
	8683	8975	0.000409	0.00239	0	0	0	0	0	0	1	-360	360;
	8683	216	0.00094	0.005	0	438	0	0	0	0	1	-360	360;
	8683	2961	0.000219	0.0007	0	629	0	0	0	0	1	-360	360;
	8683	6773	6e-05	0.00025	0	0	0	0	0	0	1	-360	360;
	8683	5278	5e-05	0.000229	0	0	0	0	0	0	1	-360	360;
	3768	2751	0.00019	0.000909	0	0	0	0	0	0	1	-360	360;
	2276	5546	0.001711	0.00782	0	0	0	0	0	0	1	-360	360;
	2276	6816	0.002651	0.01176	0	419	0	0	0	0	1	-360	360;
	7766	444	0.001409	0.010599	0	0	0	0	0	0	1	-360	360;
	7766	5334	0.001419	0.010219	0	0	0	0	0	0	1	-360	360;
	839	3918	0.00013	0.00212	0	856	0	0	0	0	1	-360	360;
	839	809	0.00026	0.00313	0	1382	0	0	0	0	1	-360	360;
	4506	2286	0.0047	0.0318	0	476	0	0	0	0	1	-360	360;
	4506	6802	0.00136	0.00768	0	438	0	0	0	0	1	-360	360;
	4506	3520	0.005669	0.027789	0	400	0	0	0	0	1	-360	360;
	7273	6271	0.004409	0.033091	0	381	0	0	0	0	1	-360	360;
	7273	1435	0.01056	0.044469	0	0	0	0	0	0	1	-360	360;
	7273	7148	0.01963	0.07736	0	438	0	0	0	0	1	-360	360;
	7273	7437	0.01157	0.04694	0	457	0	0	0	0	1	-360	360;
	6104	871	0.000479	0.00255	0	457	0	0	0	0	1	-360	360;
	7580	871	0.000469	0.00236	0	0	0	0	0	0	1	-360	360;
	1188	1539	0.01237	0.067039	0	381	0	0	0	0	1	-360	360;
	1188	8515	0.0132	0.056229	0	381	0	0	0	0	1	-360	360;
	3971	4084	0.00026	0.001711	0	0	0	0	0	0	1	-360	360;
	3971	4084	0.000271	0.001669	0	0	0	0	0	0	1	-360	360;
	6845	5340	0.000211	0.000401	0	0	0	0	0	0	1	-360	360;
	8294	5814	0.00105	0.005159	0	0	0	0	0	0	1	-360	360;
	7115	6807	0.00212	0.00931	0	419	0	0	0	0	1	-360	360;
	619	6791	0.000151	0.00068	0	0	0	0	0	0	1	-360	360;
	5418	4908	0.00019	0.00082	0	0	0	0	0	0	1	-360	360;
	5120	6816	0.003219	0.01405	0	400	0	0	0	0	1	-360	360;
	6692	9158	0.007271	0.033159	0	381	0	0	0	0	1	-360	360;
	6692	8035	0.006659	0.03132	0	400	0	0	0	0	1	-360	360;
	3445	2848	0.00057	0.00668	0	1580	0	0	0	0	1	-360	360;
	3445	7056	0.00057	0.00758	0	1514	0	0	0	0	1	-360	360;
	3445	7056	0.00039	0.00607	0	921	0	0	0	0	1	-360	360;
	3445	1642	0.00155	0.0179	0	0	0	0	0	0	1	-360	360;
	3445	4594	0.00036	0.00485	0	1645	0	0	0	0	1	-360	360;
	6114	1692	0.000581	0.004229	0	438	0	0	0	0	1	-360	360;
	6114	8190	0.00056	0.00431	0	0	0	0	0	0	1	-360	360;
	6114	5836	0.003099	0.020531	0	572	0	0	0	0	1	-360	360;
	6114	3191	0.00118	0.016919	0	610	0	0	0	0	1	-360	360;
	6114	5300	0.00188	0.01407	0	0	0	0	0	0	1	-360	360;
	6114	4889	0.00545	0.04082	0	457	0	0	0	0	1	-360	360;
	7264	444	0.00214	0.01639	0	553	0	0	0	0	1	-360	360;
	7264	4747	0.002479	0.01549	0	0	0	0	0	0	1	-360	360;
	7982	9011	0.000159	0.001229	0	572	0	0	0	0	1	-360	360;
	7982	6675	0.00026	0.00162	0	553	0	0	0	0	1	-360	360;
	7982	6909	0.000159	0.00113	0	476	0	0	0	0	1	-360	360;
	7982	8835	0.000461	0.003219	0	572	0	0	0	0	1	-360	360;
	7982	4186	0.000159	0.001229	0	0	0	0	0	0	1	-360	360;
	1604	502	0.000229	0.00139	0	476	0	0	0	0	1	-360	360;
	2842	502	0.0002	0.00156	0	438	0	0	0	0	1	-360	360;
	1354	513	0.00145	0.00824	0	381	0	0	0	0	1	-360	360;
	1354	502	0.00069	0.005	0	591	0	0	0	0	1	-360	360;
	891	3697	0.00036	0.00455	0	1909	0	0	0	0	1	-360	360;
	891	3697	0.00036	0.00453	0	1876	0	0	0	0	1	-360	360;
	8312	7541	0.00066	0.00997	0	1448	0	0	0	0	1	-360	360;
	8312	7541	0.00064	0.01016	0	1086	0	0	0	0	1	-360	360;
	6306	4826	9.9e-05	0.00044	0	629	0	0	0	0	1	-360	360;
	4826	4418	6e-05	0.000469	0	0	0	0	0	0	1	-360	360;
	4826	1605	3.9e-05	0.000341	0	0	0	0	0	0	1	-360	360;
	4826	6952	0.002531	0.01943	0	533	0	0	0	0	1	-360	360;
	4826	7124	0.00945	0.06418	0	514	0	0	0	0	1	-360	360;
	9128	217	0.000591	0.002831	0	457	0	0	0	0	1	-360	360;
	7014	217	0.00087	0.00387	0	0	0	0	0	0	1	-360	360;
	720	5350	0.0015	0.01785	0	1283	0	0	0	0	1	-360	360;
	720	5350	0.00147	0.0182	0	1251	0	0	0	0	1	-360	360;
	1711	639	0.00151	0.0113	0	476	0	0	0	0	1	-360	360;
	1711	639	0.001479	0.011521	0	495	0	0	0	0	1	-360	360;
	1711	9130	0.00174	0.010341	0	514	0	0	0	0	1	-360	360;
	1711	9130	0.001659	0.01206	0	553	0	0	0	0	1	-360	360;
	8209	1998	0.00069	0.000479	0	0	0	0	0	0	1	-360	360;
	8209	7974	0.00044	0.00139	0	0	0	0	0	0	1	-360	360;
	1035	8180	0.002039	0.01601	0	0	0	0	0	0	1	-360	360;
	1035	7466	0.01349	0.05662	0	457	0	0	0	0	1	-360	360;
	1035	305	0.00082	0.00438	0	400	0	0	0	0	1	-360	360;
	1035	305	0.0008	0.004461	0	381	0	0	0	0	1	-360	360;
	1035	5106	0.001091	0.00888	0	553	0	0	0	0	1	-360	360;
	1035	5106	0.00111	0.007039	0	591	0	0	0	0	1	-360	360;
	2527	8466	0.000849	0.00537	0	533	0	0	0	0	1	-360	360;
	8669	8893	0.00061	0.00275	0	457	0	0	0	0	1	-360	360;
	6206	7691	0.00141	0.0188	0	1415	0	0	0	0	1	-360	360;
	6206	6624	0.00121	0.01427	0	1645	0	0	0	0	1	-360	360;
	3240	8158	0.001711	0.007159	0	438	0	0	0	0	1	-360	360;
	3240	8795	0.00545	0.02381	0	400	0	0	0	0	1	-360	360;
	3240	4143	0.000771	0.004331	0	0	0	0	0	0	1	-360	360;
	3240	2797	0.00414	0.03282	0	553	0	0	0	0	1	-360	360;
	3240	6331	0.00264	0.011521	0	438	0	0	0	0	1	-360	360;
	3240	6384	0.00261	0.011	0	476	0	0	0	0	1	-360	360;
	3240	7523	0.00239	0.00919	0	419	0	0	0	0	1	-360	360;
	3240	2889	0.000539	0.00226	0	0	0	0	0	0	1	-360	360;
	3240	6887	0.00325	0.017979	0	572	0	0	0	0	1	-360	360;
	1249	4541	0.00076	0.0071	0	1316	0	0	0	0	1	-360	360;
	819	7523	0.000651	0.00343	0	0	0	0	0	0	1	-360	360;
	2304	4353	0.001219	0.00812	0	438	0	0	0	0	1	-360	360;
	2304	4816	0.00243	0.01439	0	591	0	0	0	0	1	-360	360;
	9174	6246	0.0006	0.00839	0	1251	0	0	0	0	1	-360	360;
	9174	6246	0.00059	0.00856	0	1185	0	0	0	0	1	-360	360;
	9174	5658	0.00087	0.01328	0	1777	0	0	0	0	1	-360	360;
	9174	5658	0.00073	0.01595	0	1349	0	0	0	0	1	-360	360;
	2924	3645	0.00163	0.002039	0	0	0	0	0	0	1	-360	360;
	2924	3645	0.001659	0.002	0	0	0	0	0	0	1	-360	360;
	59	5764	0.00339	0.019669	0	572	0	0	0	0	1	-360	360;
	59	2360	0.00239	0.017281	0	629	0	0	0	0	1	-360	360;
	6510	9033	0.00032	0.00112	0	495	0	0	0	0	1	-360	360;
	6510	5918	0.0018	0.01401	0	495	0	0	0	0	1	-360	360;
	6510	5918	0.00212	0.012469	0	533	0	0	0	0	1	-360	360;
	1275	2850	0.00063	0.00411	0	0	0	0	0	0	1	-360	360;
	6426	4110	0.009729	0.03505	0	0	0	0	0	0	1	-360	360;
	6426	3558	0.01007	0.03056	0	0	0	0	0	0	1	-360	360;
	1758	2816	0.00042	0.00623	0	0	0	0	0	0	1	-360	360;
	1758	352	0.00036	0.00722	0	1843	0	0	0	0	1	-360	360;
	1758	6581	0.00053	0.0078	0	954	0	0	0	0	1	-360	360;
	1758	1923	0.00109	0.0173	0	823	0	0	0	0	1	-360	360;
	1758	5837	0.00071	0.01144	0	0	0	0	0	0	1	-360	360;
	1758	3830	0.00157	0.01884	0	1251	0	0	0	0	1	-360	360;
	1758	3577	0.00032	0.00384	0	856	0	0	0	0	1	-360	360;
	8030	1794	0.00041	0.00649	0	1547	0	0	0	0	1	-360	360;
	8030	7808	0.0004	0.00658	0	0	0	0	0	0	1	-360	360;
	8030	6581	0.00054	0.00763	0	954	0	0	0	0	1	-360	360;
	8030	1923	0.00107	0.01602	0	921	0	0	0	0	1	-360	360;
	8030	5837	0.0007	0.01165	0	0	0	0	0	0	1	-360	360;
	8030	5648	0.0016	0.01844	0	1349	0	0	0	0	1	-360	360;
	8030	3577	0.00032	0.00376	0	954	0	0	0	0	1	-360	360;
	4615	171	0.00212	0.01207	0	495	0	0	0	0	1	-360	360;
	3137	2341	5e-05	0.00037	0	0	0	0	0	0	1	-360	360;
	6271	8787	0.00355	0.024651	0	419	0	0	0	0	1	-360	360;
	6252	9137	0.00675	0.04145	0	533	0	0	0	0	1	-360	360;
	5533	7133	0.00195	0.014341	0	438	0	0	0	0	1	-360	360;
	8361	4355	0.00018	0.00043	0	381	0	0	0	0	1	-360	360;
	8361	2928	0.00037	0.000969	0	457	0	0	0	0	1	-360	360;
	609	1341	0.00257	0.017169	0	476	0	0	0	0	1	-360	360;
	609	5419	0.00076	0.00399	0	495	0	0	0	0	1	-360	360;
	609	5419	0.00075	0.00406	0	495	0	0	0	0	1	-360	360;
	609	2079	0.003901	0.023961	0	572	0	0	0	0	1	-360	360;
	609	772	0.00249	0.019211	0	457	0	0	0	0	1	-360	360;
	3654	2928	0.00025	0.001159	0	476	0	0	0	0	1	-360	360;
	1355	7282	0.00195	0.02924	0	0	0	0	0	0	1	-360	360;
	1355	2197	0.00152	0.01527	0	1481	0	0	0	0	1	-360	360;
	1355	7640	0.00149	0.01558	0	1613	0	0	0	0	1	-360	360;
	1355	1081	0.00253	0.0276	0	1382	0	0	0	0	1	-360	360;
	1355	1081	0.00263	0.02648	0	1580	0	0	0	0	1	-360	360;
	1355	2359	0.00077	0.0092	0	1415	0	0	0	0	1	-360	360;
	1355	4196	0.00112	0.01705	0	757	0	0	0	0	1	-360	360;
	5856	8546	0.00226	0.01101	0	0	0	0	0	0	1	-360	360;
	5856	5987	0.004531	0.029039	0	495	0	0	0	0	1	-360	360;
	5856	8005	0.002081	0.030409	0	0	0	0	0	0	1	-360	360;
	5856	3925	0.0053	0.03257	0	438	0	0	0	0	1	-360	360;
	5856	3324	0.00513	0.041901	0	553	0	0	0	0	1	-360	360;
	5856	4748	0.001461	0.01056	0	457	0	0	0	0	1	-360	360;
	8265	4816	0.0005	0.002591	0	0	0	0	0	0	1	-360	360;
	6151	6313	0.00055	0.0033	0	457	0	0	0	0	1	-360	360;
	4005	6532	0.002289	0.01763	0	476	0	0	0	0	1	-360	360;
	4005	2177	0.0075	0.043531	0	457	0	0	0	0	1	-360	360;
	5723	2079	0.00138	0.019961	0	0	0	0	0	0	1	-360	360;
	5723	789	0.00126	0.018729	0	591	0	0	0	0	1	-360	360;
	2327	9189	0.00038	0.001331	0	495	0	0	0	0	1	-360	360;
	2327	3400	0.00019	0.001461	0	0	0	0	0	0	1	-360	360;
	207	4747	0.002151	0.013401	0	476	0	0	0	0	1	-360	360;
	207	7791	0.000789	0.005669	0	572	0	0	0	0	1	-360	360;
	3205	6735	3e-05	0.00042	0	757	0	0	0	0	1	-360	360;
	3205	666	0.00079	0.00899	0	1382	0	0	0	0	1	-360	360;
	3205	666	0.0008	0.00879	0	1547	0	0	0	0	1	-360	360;
	3205	4435	0.00334	0.03393	0	1382	0	0	0	0	1	-360	360;
	3205	1644	0.00063	0.00789	0	0	0	0	0	0	1	-360	360;
	3205	3483	0.0006	0.00842	0	1316	0	0	0	0	1	-360	360;
	4683	7989	0.00011	0.00125	0	0	0	0	0	0	1	-360	360;
	4683	6516	0.00012	0.001091	0	0	0	0	0	0	1	-360	360;
	4683	7697	9.9e-05	0.00125	0	0	0	0	0	0	1	-360	360;
	4683	416	0.00294	0.02443	0	629	0	0	0	0	1	-360	360;
	4683	2847	0.00144	0.011151	0	495	0	0	0	0	1	-360	360;
	4683	2563	0.001831	0.010849	0	591	0	0	0	0	1	-360	360;
	4683	2563	0.001599	0.01055	0	572	0	0	0	0	1	-360	360;
	4683	2995	0.00144	0.019591	0	381	0	0	0	0	1	-360	360;
	5691	1001	0.0133	0.09551	0	610	0	0	0	0	1	-360	360;
	5691	4426	0.009591	0.073961	0	514	0	0	0	0	1	-360	360;
	5350	2721	0.0007	0.00671	0	1547	0	0	0	0	1	-360	360;
	5350	2721	0.00072	0.00658	0	1580	0	0	0	0	1	-360	360;
	2361	4936	0.001901	0.00964	0	629	0	0	0	0	1	-360	360;
	2361	5179	0.001711	0.01189	0	610	0	0	0	0	1	-360	360;
	2044	4748	0.00025	0.00187	0	0	0	0	0	0	1	-360	360;
	4197	5522	0.00043	0.002211	0	0	0	0	0	0	1	-360	360;
	5699	8670	0.000401	0.00243	0	0	0	0	0	0	1	-360	360;
	7809	5110	0.00336	0.017	0	457	0	0	0	0	1	-360	360;
	7222	4816	0.005521	0.031031	0	0	0	0	0	0	1	-360	360;
	4402	3817	9.9e-05	0.00026	0	0	0	0	0	0	1	-360	360;
	6857	2732	0.00045	0.00718	0	1086	0	0	0	0	1	-360	360;
	6857	4550	0.00044	0.00665	0	921	0	0	0	0	1	-360	360;
	6857	4339	0.00103	0.01516	0	1283	0	0	0	0	1	-360	360;
	6857	7513	0.00101	0.01546	0	954	0	0	0	0	1	-360	360;
	5016	26	0.00045	0.00608	0	1514	0	0	0	0	1	-360	360;
	5016	9213	0.0004	0.00501	0	1349	0	0	0	0	1	-360	360;
	6112	9002	0.001729	0.01337	0	591	0	0	0	0	1	-360	360;
	6112	6231	0.00264	0.01489	0	514	0	0	0	0	1	-360	360;
	6112	883	0.002341	0.018961	0	514	0	0	0	0	1	-360	360;
	6112	5738	0.001841	0.01363	0	476	0	0	0	0	1	-360	360;
	8860	7700	0.001219	0.00949	0	495	0	0	0	0	1	-360	360;
	8860	9002	0.002271	0.01151	0	495	0	0	0	0	1	-360	360;
	8860	5990	0.010781	0.05775	0	0	0	0	0	0	1	-360	360;
	8860	6450	0.00251	0.013919	0	400	0	0	0	0	1	-360	360;
	8860	7726	0.00101	0.007331	0	514	0	0	0	0	1	-360	360;
	2489	4110	0.00061	0.003909	0	629	0	0	0	0	1	-360	360;
	3985	6376	0.004289	0.019331	0	0	0	0	0	0	1	-360	360;
	3985	4110	0.0008	0.004401	0	629	0	0	0	0	1	-360	360;
	3985	4829	0.008099	0.03519	0	610	0	0	0	0	1	-360	360;
	3985	7857	0.00063	0.004591	0	514	0	0	0	0	1	-360	360;
	2806	7857	0.000651	0.00456	0	0	0	0	0	0	1	-360	360;
	2489	9012	0.00114	0.00581	0	0	0	0	0	0	1	-360	360;
	2489	2231	0.001211	0.00538	0	0	0	0	0	0	1	-360	360;
	658	3793	0.00418	0.030159	0	572	0	0	0	0	1	-360	360;
	658	8497	0.00324	0.02413	0	610	0	0	0	0	1	-360	360;
	658	8497	0.003169	0.024591	0	629	0	0	0	0	1	-360	360;
	3502	9002	0.000219	0.00088	0	495	0	0	0	0	1	-360	360;
	3502	9002	0.000289	0.00095	0	514	0	0	0	0	1	-360	360;
	6313	44	0.0035	0.022211	0	553	0	0	0	0	1	-360	360;
	6313	7582	0.00313	0.02256	0	0	0	0	0	0	1	-360	360;
	5944	8808	0.000831	0.00162	0	0	0	0	0	0	1	-360	360;
	3183	8515	0.002099	0.01068	0	400	0	0	0	0	1	-360	360;
	1005	8515	0.00224	0.00976	0	0	0	0	0	0	1	-360	360;
	8787	1448	0.00169	0.01231	0	629	0	0	0	0	1	-360	360;
	6636	9051	0.00011	0.00076	0	0	0	0	0	0	1	-360	360;
	9109	9051	0.00011	0.00068	0	0	0	0	0	0	1	-360	360;
	6961	6376	0.005479	0.021909	0	0	0	0	0	0	1	-360	360;
	6961	8818	0.003331	0.017039	0	438	0	0	0	0	1	-360	360;
	5077	4353	0.000479	0.001271	0	0	0	0	0	0	1	-360	360;
	7471	5241	0.00109	0.01629	0	823	0	0	0	0	1	-360	360;
	7471	8200	4e-05	0.00063	0	1810	0	0	0	0	1	-360	360;
	549	6290	0.0053	0.024531	0	572	0	0	0	0	1	-360	360;
	549	2291	0.004659	0.021919	0	419	0	0	0	0	1	-360	360;
	549	3656	0.001659	0.01043	0	572	0	0	0	0	1	-360	360;
	5002	4144	0.006961	0.053539	0	495	0	0	0	0	1	-360	360;
	1888	5441	0.00036	0.00288	0	629	0	0	0	0	1	-360	360;
	1888	5441	0.00038	0.003	0	629	0	0	0	0	1	-360	360;
	5490	4970	0.00048	0.00643	0	0	0	0	0	0	1	-360	360;
	5490	4970	0.00051	0.0066	0	0	0	0	0	0	1	-360	360;
	5490	4970	0.00046	0.00729	0	1876	0	0	0	0	1	-360	360;
	5490	5461	0.0006	0.00868	0	1810	0	0	0	0	1	-360	360;
	5490	7778	0.00068	0.01076	0	1942	0	0	0	0	1	-360	360;
	5490	6897	0.00066	0.01096	0	1909	0	0	0	0	1	-360	360;
	3075	4562	0.00101	0.005099	0	400	0	0	0	0	1	-360	360;
	564	8475	0.003901	0.02437	0	0	0	0	0	0	1	-360	360;
	564	2252	0.00276	0.016461	0	553	0	0	0	0	1	-360	360;
	564	7903	0.00588	0.047081	0	0	0	0	0	0	1	-360	360;
	4331	1857	3.9e-05	0.000219	0	0	0	0	0	0	1	-360	360;
	2877	3674	0.00131	0.00538	0	0	0	0	0	0	1	-360	360;
	2854	1081	0.00679	0.04679	0	1020	0	0	0	0	1	-360	360;
	2854	8267	0.00138	0.01466	0	1086	0	0	0	0	1	-360	360;
	9112	9176	0.003531	0.01724	0	0	0	0	0	0	1	-360	360;
	9112	1422	0.01324	0.06936	0	476	0	0	0	0	1	-360	360;
	9112	8515	0.00363	0.01743	0	438	0	0	0	0	1	-360	360;
	9112	8515	0.00257	0.01662	0	533	0	0	0	0	1	-360	360;
	9112	5051	0.01399	0.056289	0	419	0	0	0	0	1	-360	360;
	9112	1051	0.00343	0.01751	0	457	0	0	0	0	1	-360	360;
	9112	1397	0.00301	0.015409	0	0	0	0	0	0	1	-360	360;
	9112	1584	0.00857	0.033091	0	476	0	0	0	0	1	-360	360;
	6168	6888	0.003341	0.01689	0	0	0	0	0	0	1	-360	360;
	7653	3019	3.9e-05	0.000219	0	0	0	0	0	0	1	-360	360;
	7341	7098	0.00087	0.00714	0	457	0	0	0	0	1	-360	360;
	4701	8267	0.00011	0.001	0	757	0	0	0	0	1	-360	360;
	1672	5365	0.000219	0.00101	0	0	0	0	0	0	1	-360	360;
	3880	5365	0.000219	0.001031	0	0	0	0	0	0	1	-360	360;
	6648	5365	0.000271	0.00101	0	0	0	0	0	0	1	-360	360;
	3609	5365	0.000271	0.000979	0	0	0	0	0	0	1	-360	360;
	2782	5365	0.000289	0.00107	0	0	0	0	0	0	1	-360	360;
	1844	5365	0.000289	0.00105	0	0	0	0	0	0	1	-360	360;
	4353	6053	0.001591	0.012289	0	381	0	0	0	0	1	-360	360;
	2166	5237	0.000271	0.001901	0	495	0	0	0	0	1	-360	360;
	2166	2886	0.00186	0.010711	0	514	0	0	0	0	1	-360	360;
	1043	3513	1e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	4049	883	0.000169	0.000271	0	0	0	0	0	0	1	-360	360;
	4520	883	0.000169	0.00031	0	0	0	0	0	0	1	-360	360;
	1704	5814	0.00526	0.03787	0	553	0	0	0	0	1	-360	360;
	8107	1486	0.000229	0.000599	0	0	0	0	0	0	1	-360	360;
	5664	5522	0.00089	0.004091	0	438	0	0	0	0	1	-360	360;
	4482	8670	0.00093	0.004	0	400	0	0	0	0	1	-360	360;
	8795	7523	0.00289	0.013099	0	400	0	0	0	0	1	-360	360;
	2337	6430	0.002599	0.016479	0	533	0	0	0	0	1	-360	360;
	455	7775	0.00369	0.02712	0	553	0	0	0	0	1	-360	360;
	455	3857	0.0062	0.04864	0	629	0	0	0	0	1	-360	360;
	455	333	0.008021	0.059031	0	629	0	0	0	0	1	-360	360;
	455	6940	0.004979	0.0295	0	572	0	0	0	0	1	-360	360;
	8225	3656	0.001159	0.005281	0	0	0	0	0	0	1	-360	360;
	5413	7752	0.00361	0.01824	0	0	0	0	0	0	1	-360	360;
	5413	1940	0.001521	0.01075	0	0	0	0	0	0	1	-360	360;
	5413	4234	0.00868	0.041711	0	0	0	0	0	0	1	-360	360;
	2591	6828	0.01118	0.05986	0	0	0	0	0	0	1	-360	360;
	2591	7571	0.003229	0.02543	0	591	0	0	0	0	1	-360	360;
	2591	7571	0.00218	0.01869	0	0	0	0	0	0	1	-360	360;
	2591	7571	0.001901	0.021909	0	572	0	0	0	0	1	-360	360;
	2591	7994	0.00974	0.043969	0	400	0	0	0	0	1	-360	360;
	2591	7226	0.003461	0.01814	0	457	0	0	0	0	1	-360	360;
	2591	1554	0.00487	0.027669	0	419	0	0	0	0	1	-360	360;
	9011	1156	0.000651	0.003961	0	610	0	0	0	0	1	-360	360;
	8689	5502	0.00169	0.00887	0	400	0	0	0	0	1	-360	360;
	3085	8873	0.001711	0.01336	0	514	0	0	0	0	1	-360	360;
	1568	980	0.00031	0.000539	0	400	0	0	0	0	1	-360	360;
	3204	3331	0.0027	0.01157	0	0	0	0	0	0	1	-360	360;
	3557	639	0.000909	0.006901	0	514	0	0	0	0	1	-360	360;
	3557	4936	0.001729	0.01214	0	514	0	0	0	0	1	-360	360;
	416	3492	0.00614	0.03106	0	438	0	0	0	0	1	-360	360;
	416	7937	0.00418	0.02388	0	514	0	0	0	0	1	-360	360;
	3255	1262	0.00238	0.017771	0	0	0	0	0	0	1	-360	360;
	3255	1803	8.1e-05	0.00063	0	0	0	0	0	0	1	-360	360;
	7373	367	0.001599	0.01618	0	0	0	0	0	0	1	-360	360;
	4251	2132	0.00532	0.026961	0	419	0	0	0	0	1	-360	360;
	4251	2132	0.00361	0.02814	0	553	0	0	0	0	1	-360	360;
	4251	5567	0.002469	0.01025	0	533	0	0	0	0	1	-360	360;
	6376	1808	0.005229	0.02675	0	400	0	0	0	0	1	-360	360;
	6376	1808	0.00512	0.02726	0	419	0	0	0	0	1	-360	360;
	6376	7857	0.00355	0.01694	0	400	0	0	0	0	1	-360	360;
	9185	7289	0.00139	0.007659	0	457	0	0	0	0	1	-360	360;
	9185	3999	0.000169	0.000669	0	514	0	0	0	0	1	-360	360;
	8653	2303	0.00111	0.00712	0	476	0	0	0	0	1	-360	360;
	908	6474	0.00031	0.001901	0	533	0	0	0	0	1	-360	360;
	7030	6474	0.000281	0.002091	0	0	0	0	0	0	1	-360	360;
	4205	6556	0.000281	0.001651	0	0	0	0	0	0	1	-360	360;
	7325	6556	0.000271	0.00169	0	0	0	0	0	0	1	-360	360;
	5857	3391	0.001521	0.00843	0	381	0	0	0	0	1	-360	360;
	5857	3760	0.000841	0.0062	0	533	0	0	0	0	1	-360	360;
	3707	8765	0.00019	0.00126	0	514	0	0	0	0	1	-360	360;
	900	6619	0.000219	0.00143	0	0	0	0	0	0	1	-360	360;
	2918	8109	0.00121	0.01796	0	1876	0	0	0	0	1	-360	360;
	2918	8109	0.0014	0.01512	0	0	0	0	0	0	1	-360	360;
	6532	2177	0.008409	0.063651	0	0	0	0	0	0	1	-360	360;
	6532	5317	0.00918	0.07344	0	610	0	0	0	0	1	-360	360;
	6532	608	0.004219	0.0345	0	514	0	0	0	0	1	-360	360;
	6532	608	0.0042	0.031651	0	457	0	0	0	0	1	-360	360;
	4494	8763	0.00121	0.01381	0	1547	0	0	0	0	1	-360	360;
	4494	7056	0.00107	0.01326	0	1580	0	0	0	0	1	-360	360;
	6828	5814	0.003401	0.02607	0	0	0	0	0	0	1	-360	360;
	3670	21	0.000289	0.00138	0	419	0	0	0	0	1	-360	360;
	5137	7702	0.001031	0.006039	0	0	0	0	0	0	1	-360	360;
	1334	7702	0.00105	0.005909	0	0	0	0	0	0	1	-360	360;
	8448	7624	0.001091	0.007901	0	419	0	0	0	0	1	-360	360;
	8448	1129	0.00332	0.024021	0	553	0	0	0	0	1	-360	360;
	8448	3070	0.00344	0.021419	0	400	0	0	0	0	1	-360	360;
	8448	306	0.00095	0.010419	0	514	0	0	0	0	1	-360	360;
	8448	306	0.00093	0.01062	0	476	0	0	0	0	1	-360	360;
	8448	7491	0.00474	0.025409	0	0	0	0	0	0	1	-360	360;
	4685	2089	0.000271	0.000841	0	381	0	0	0	0	1	-360	360;
	7485	6253	0.000271	0.00119	0	381	0	0	0	0	1	-360	360;
	7485	1562	0.000289	0.00076	0	0	0	0	0	0	1	-360	360;
	7776	1644	0.00031	0.00454	0	1942	0	0	0	0	1	-360	360;
	5971	3483	0.00032	0.0044	0	0	0	0	0	0	1	-360	360;
	2786	1644	0.00028	0.00494	0	0	0	0	0	0	1	-360	360;
	972	3483	0.00028	0.00508	0	0	0	0	0	0	1	-360	360;
	8906	923	0.000531	0.00093	0	0	0	0	0	0	1	-360	360;
	8906	4914	0.0003	0.001031	0	495	0	0	0	0	1	-360	360;
	2656	9101	0.00139	0.02163	0	757	0	0	0	0	1	-360	360;
	2656	7164	0.00164	0.02321	0	1876	0	0	0	0	1	-360	360;
	3519	9101	0.00136	0.02205	0	757	0	0	0	0	1	-360	360;
	3519	7164	0.0016	0.02367	0	757	0	0	0	0	1	-360	360;
	1775	9137	0.002091	0.01174	0	381	0	0	0	0	1	-360	360;
	1775	9137	0.001719	0.00961	0	553	0	0	0	0	1	-360	360;
	5641	1026	0.012341	0.076781	0	419	0	0	0	0	1	-360	360;
	5641	9137	0.00176	0.009419	0	591	0	0	0	0	1	-360	360;
	1096	2889	0.00031	0.00182	0	0	0	0	0	0	1	-360	360;
	7380	2732	0.00386	0.02348	0	1283	0	0	0	0	1	-360	360;
	7380	4787	0.00017	0.00104	0	1152	0	0	0	0	1	-360	360;
	3543	5308	0.00061	0.00808	0	1613	0	0	0	0	1	-360	360;
	3543	3830	0.00059	0.0067	0	0	0	0	0	0	1	-360	360;
	2372	1027	0.00059	0.00748	0	1481	0	0	0	0	1	-360	360;
	2372	5648	0.00061	0.0063	0	0	0	0	0	0	1	-360	360;
	367	1172	0.001031	0.008789	0	629	0	0	0	0	1	-360	360;
	367	9191	0.00051	0.006169	0	629	0	0	0	0	1	-360	360;
	367	8722	0.0005	0.00619	0	629	0	0	0	0	1	-360	360;
	1833	3610	0.00049	0.00376	0	572	0	0	0	0	1	-360	360;
	1833	3610	0.000539	0.003229	0	533	0	0	0	0	1	-360	360;
	1833	3962	0.001031	0.00882	0	0	0	0	0	0	1	-360	360;
	1833	1838	0.000711	0.00463	0	457	0	0	0	0	1	-360	360;
	1833	2230	0.000711	0.004539	0	0	0	0	0	0	1	-360	360;
	5388	2467	0.00206	0.015341	0	514	0	0	0	0	1	-360	360;
	5388	2467	0.002021	0.01564	0	514	0	0	0	0	1	-360	360;
	6110	8853	0.00094	0.005781	0	610	0	0	0	0	1	-360	360;
	6110	1001	0.002219	0.016969	0	629	0	0	0	0	1	-360	360;
	7583	1808	0.00238	0.01187	0	381	0	0	0	0	1	-360	360;
	7583	2889	0.00707	0.03095	0	381	0	0	0	0	1	-360	360;
	26	6624	0.00256	0.03401	0	1645	0	0	0	0	1	-360	360;
	26	2479	0.00203	0.02416	0	1152	0	0	0	0	1	-360	360;
	26	4598	0.001	0.01459	0	0	0	0	0	0	1	-360	360;
	26	4231	0.00103	0.00987	0	1613	0	0	0	0	1	-360	360;
	26	5144	0.00038	0.00483	0	1448	0	0	0	0	1	-360	360;
	26	5144	0.00041	0.00398	0	1316	0	0	0	0	1	-360	360;
	7691	2479	0.00167	0.01721	0	1382	0	0	0	0	1	-360	360;
	7691	2129	0.00098	0.01487	0	987	0	0	0	0	1	-360	360;
	7691	9213	0.00042	0.00342	0	1349	0	0	0	0	1	-360	360;
	7691	9213	0.00019	0.00248	0	1580	0	0	0	0	1	-360	360;
	7691	4231	0.00105	0.00966	0	1909	0	0	0	0	1	-360	360;
	7691	3513	0.00043	0.00473	0	1382	0	0	0	0	1	-360	360;
	7691	3513	0.00041	0.00397	0	1316	0	0	0	0	1	-360	360;
	1876	3906	0.00051	0.00514	0	1218	0	0	0	0	1	-360	360;
	113	3906	0.00053	0.00503	0	1415	0	0	0	0	1	-360	360;
	7752	3758	0.00455	0.022771	0	400	0	0	0	0	1	-360	360;
	7752	3758	0.00557	0.02055	0	381	0	0	0	0	1	-360	360;
	7752	3758	0.00339	0.02026	0	0	0	0	0	0	1	-360	360;
	4952	8976	0.00237	0.01068	0	438	0	0	0	0	1	-360	360;
	4952	8189	0.0037	0.019479	0	381	0	0	0	0	1	-360	360;
	6119	7886	0.01407	0.056849	0	0	0	0	0	0	1	-360	360;
	6119	4544	0.001669	0.015409	0	610	0	0	0	0	1	-360	360;
	6119	4544	0.001711	0.015091	0	0	0	0	0	0	1	-360	360;
	7905	3697	0.00172	0.01619	0	1678	0	0	0	0	1	-360	360;
	7905	7541	0.00029	0.00447	0	987	0	0	0	0	1	-360	360;
	7905	7541	0.00033	0.00387	0	1119	0	0	0	0	1	-360	360;
	7905	4141	0.00079	0.01154	0	987	0	0	0	0	1	-360	360;
	7905	4141	0.00077	0.01177	0	987	0	0	0	0	1	-360	360;
	7905	3022	0.00127	0.01498	0	1547	0	0	0	0	1	-360	360;
	1768	163	0.00214	0.012599	0	495	0	0	0	0	1	-360	360;
	4245	2425	0.000159	0.00125	0	457	0	0	0	0	1	-360	360;
	3698	5297	0.000211	0.000669	0	457	0	0	0	0	1	-360	360;
	750	8328	0.000341	0.001169	0	514	0	0	0	0	1	-360	360;
	750	870	0.00024	0.001469	0	572	0	0	0	0	1	-360	360;
	750	2961	0.00018	0.000669	0	533	0	0	0	0	1	-360	360;
	3499	7988	0.00033	0.00683	0	921	0	0	0	0	1	-360	360;
	3499	2197	0.00121	0.01211	0	1415	0	0	0	0	1	-360	360;
	3499	7640	0.00123	0.01185	0	1448	0	0	0	0	1	-360	360;
	3499	4231	0.00058	0.0097	0	1185	0	0	0	0	1	-360	360;
	3499	5144	0.00103	0.01292	0	1251	0	0	0	0	1	-360	360;
	3499	3513	0.00116	0.01266	0	921	0	0	0	0	1	-360	360;
	7289	3999	0.0013	0.006841	0	457	0	0	0	0	1	-360	360;
	3018	1051	6e-05	0.000349	0	438	0	0	0	0	1	-360	360;
	2702	3435	0.000169	0.0007	0	400	0	0	0	0	1	-360	360;
	9131	2020	0.000169	0.000719	0	0	0	0	0	0	1	-360	360;
	6308	6684	0.000271	0.001031	0	0	0	0	0	0	1	-360	360;
	4831	3680	0.008271	0.036841	0	0	0	0	0	0	1	-360	360;
	5410	2751	0.002969	0.02299	0	419	0	0	0	0	1	-360	360;
	3082	4623	0.00124	0.01273	0	1218	0	0	0	0	1	-360	360;
	3082	4339	0.00106	0.00989	0	1481	0	0	0	0	1	-360	360;
	3082	5288	7e-05	0.00029	0	757	0	0	0	0	1	-360	360;
	3082	5421	9e-05	0.00079	0	757	0	0	0	0	1	-360	360;
	3082	6475	0.00196	0.02049	0	1613	0	0	0	0	1	-360	360;
	3082	8468	0.00158	0.01519	0	1283	0	0	0	0	1	-360	360;
	6224	2732	0.00172	0.01811	0	1613	0	0	0	0	1	-360	360;
	6224	7513	0.00096	0.01121	0	1744	0	0	0	0	1	-360	360;
	6224	6475	0.00147	0.01707	0	0	0	0	0	0	1	-360	360;
	2174	4623	0.00129	0.0129	0	1316	0	0	0	0	1	-360	360;
	2174	4550	0.00179	0.01737	0	1645	0	0	0	0	1	-360	360;
	2174	6475	0.00131	0.01935	0	1777	0	0	0	0	1	-360	360;
	2083	9231	0.003901	0.01614	0	419	0	0	0	0	1	-360	360;
	2083	9231	0.002591	0.014771	0	553	0	0	0	0	1	-360	360;
	2083	9231	0.00224	0.0173	0	476	0	0	0	0	1	-360	360;
	2083	4054	0.00012	0.000849	0	0	0	0	0	0	1	-360	360;
	2083	3481	0.00394	0.03039	0	495	0	0	0	0	1	-360	360;
	2083	3602	0.00189	0.01	0	0	0	0	0	0	1	-360	360;
	2083	2794	0.00032	0.00257	0	0	0	0	0	0	1	-360	360;
	2083	2794	0.00038	0.00224	0	0	0	0	0	0	1	-360	360;
	2083	8467	0.00751	0.03955	0	495	0	0	0	0	1	-360	360;
	2083	1136	0.00086	0.006531	0	0	0	0	0	0	1	-360	360;
	2083	22	0.00087	0.006591	0	0	0	0	0	0	1	-360	360;
	8853	1001	0.003031	0.02276	0	629	0	0	0	0	1	-360	360;
	8853	4950	0.006331	0.03911	0	0	0	0	0	0	1	-360	360;
	8853	4410	0.00605	0.02986	0	381	0	0	0	0	1	-360	360;
	8853	6697	0.003919	0.01612	0	438	0	0	0	0	1	-360	360;
	8853	1159	0.002039	0.0108	0	438	0	0	0	0	1	-360	360;
	8853	7165	0.006289	0.026271	0	514	0	0	0	0	1	-360	360;
	2938	2848	0.00042	0.00547	0	856	0	0	0	0	1	-360	360;
	2938	2848	0.00043	0.00536	0	757	0	0	0	0	1	-360	360;
	2938	7522	0.00147	0.01373	0	1777	0	0	0	0	1	-360	360;
	2938	7522	0.00131	0.01555	0	1547	0	0	0	0	1	-360	360;
	2938	1642	0.00221	0.02217	0	1810	0	0	0	0	1	-360	360;
	2938	809	0.00224	0.0217	0	1810	0	0	0	0	1	-360	360;
	9189	6791	0.00095	0.003289	0	400	0	0	0	0	1	-360	360;
	9189	3400	0.000229	0.00082	0	0	0	0	0	0	1	-360	360;
	9189	7092	0.00068	0.00306	0	0	0	0	0	0	1	-360	360;
	9189	3718	0.00049	0.00256	0	419	0	0	0	0	1	-360	360;
	9189	3701	0.000531	0.0022	0	0	0	0	0	0	1	-360	360;
	9189	7507	0.00136	0.00956	0	514	0	0	0	0	1	-360	360;
	9189	7507	0.0012	0.010729	0	476	0	0	0	0	1	-360	360;
	9189	7507	0.00105	0.00861	0	0	0	0	0	0	1	-360	360;
	9189	2406	0.000591	0.006969	0	0	0	0	0	0	1	-360	360;
	2801	8511	0.00025	0.0008	0	0	0	0	0	0	1	-360	360;
	2801	216	0.00113	0.00632	0	400	0	0	0	0	1	-360	360;
	453	2129	0.0003	0.00275	0	954	0	0	0	0	1	-360	360;
	8691	6952	0.00836	0.051659	0	629	0	0	0	0	1	-360	360;
	8691	6952	0.01338	0.073469	0	591	0	0	0	0	1	-360	360;
	8691	1183	0.0027	0.02089	0	0	0	0	0	0	1	-360	360;
	8691	4410	0.003419	0.02636	0	610	0	0	0	0	1	-360	360;
	8691	8743	0.009531	0.033711	0	457	0	0	0	0	1	-360	360;
	2446	8763	0.0004	0.00707	0	0	0	0	0	0	1	-360	360;
	2627	8763	0.00035	0.00701	0	0	0	0	0	0	1	-360	360;
	4951	6785	0.00194	0.01751	0	533	0	0	0	0	1	-360	360;
	4951	6785	0.00156	0.01936	0	514	0	0	0	0	1	-360	360;
	4623	964	0.00078	0.00941	0	1349	0	0	0	0	1	-360	360;
	4623	964	0.00074	0.00969	0	1580	0	0	0	0	1	-360	360;
	4623	8347	0.00042	0.00532	0	1678	0	0	0	0	1	-360	360;
	4623	8347	0.0004	0.00473	0	1843	0	0	0	0	1	-360	360;
	4623	8347	0.0006	0.00499	0	1547	0	0	0	0	1	-360	360;
	1609	6552	0.00086	0.00638	0	610	0	0	0	0	1	-360	360;
	1609	6552	0.000969	0.00625	0	610	0	0	0	0	1	-360	360;
	1609	8999	0.002781	0.01725	0	457	0	0	0	0	1	-360	360;
	4368	4816	0.005031	0.028081	0	553	0	0	0	0	1	-360	360;
	5481	960	1e-05	0.00021	0	0	0	0	0	0	1	-360	360;
	7913	960	1e-05	0.00025	0	0	0	0	0	0	1	-360	360;
	506	960	2e-05	0.00031	0	757	0	0	0	0	1	-360	360;
	2980	960	3e-05	0.0004	0	757	0	0	0	0	1	-360	360;
	4554	6382	0.000841	0.00474	0	533	0	0	0	0	1	-360	360;
	4554	6382	0.00082	0.00487	0	533	0	0	0	0	1	-360	360;
	6844	1595	0.00186	0.011979	0	0	0	0	0	0	1	-360	360;
	3672	5334	0.00056	0.003021	0	381	0	0	0	0	1	-360	360;
	6219	5334	0.00055	0.003081	0	0	0	0	0	0	1	-360	360;
	8997	6153	0.00114	0.01063	0	0	0	0	0	0	1	-360	360;
	1852	5935	0.003719	0.02295	0	629	0	0	0	0	1	-360	360;
	1852	5695	0.006669	0.04288	0	514	0	0	0	0	1	-360	360;
	2931	5799	0.00243	0.0198	0	0	0	0	0	0	1	-360	360;
	2931	7124	0.00845	0.04339	0	533	0	0	0	0	1	-360	360;
	2598	3674	0.00219	0.015169	0	381	0	0	0	0	1	-360	360;
	2598	2132	0.00357	0.0302	0	419	0	0	0	0	1	-360	360;
	2558	7571	0.005849	0.032419	0	419	0	0	0	0	1	-360	360;
	8834	1129	0.00406	0.01688	0	0	0	0	0	0	1	-360	360;
	7377	3324	0.00113	0.0065	0	0	0	0	0	0	1	-360	360;
	1265	4426	0.0068	0.04613	0	514	0	0	0	0	1	-360	360;
	1265	1965	0.00337	0.02011	0	419	0	0	0	0	1	-360	360;
	8060	7162	0.00011	0.0005	0	0	0	0	0	0	1	-360	360;
	6922	5482	0.003711	0.018521	0	438	0	0	0	0	1	-360	360;
	6922	1295	0.000271	0.00145	0	457	0	0	0	0	1	-360	360;
	8369	4710	0.00326	0.01824	0	0	0	0	0	0	1	-360	360;
	8961	7133	0.00118	0.00712	0	419	0	0	0	0	1	-360	360;
	8961	6888	0.00482	0.02232	0	381	0	0	0	0	1	-360	360;
	4185	7019	0.00013	0.000469	0	381	0	0	0	0	1	-360	360;
	4185	6253	0.000229	0.00075	0	419	0	0	0	0	1	-360	360;
	7895	3391	0.004591	0.035021	0	533	0	0	0	0	1	-360	360;
	7895	1448	0.005469	0.03924	0	610	0	0	0	0	1	-360	360;
	7895	1448	0.00536	0.0403	0	0	0	0	0	0	1	-360	360;
	7895	115	0.00424	0.033031	0	0	0	0	0	0	1	-360	360;
	5589	4306	2e-05	0.00024	0	757	0	0	0	0	1	-360	360;
	5589	1763	1e-05	0.00021	0	757	0	0	0	0	1	-360	360;
	5589	3697	0.00286	0.03653	0	1645	0	0	0	0	1	-360	360;
	5589	3794	0.00117	0.01183	0	1744	0	0	0	0	1	-360	360;
	5589	3613	0.00103	0.01336	0	1481	0	0	0	0	1	-360	360;
	5589	1502	0.00156	0.01969	0	1547	0	0	0	0	1	-360	360;
	5589	6555	0.00449	0.05772	0	1777	0	0	0	0	1	-360	360;
	5589	3608	0.00259	0.03469	0	1613	0	0	0	0	1	-360	360;
	1001	4134	0.000219	0.00118	0	0	0	0	0	0	1	-360	360;
	1001	1979	6e-05	0.00095	0	0	0	0	0	0	1	-360	360;
	1001	516	6e-05	0.000651	0	629	0	0	0	0	1	-360	360;
	1001	3580	8.1e-05	0.000841	0	629	0	0	0	0	1	-360	360;
	1001	3809	0.00274	0.020831	0	514	0	0	0	0	1	-360	360;
	1001	5616	0.004781	0.027729	0	553	0	0	0	0	1	-360	360;
	1001	5616	0.002849	0.02093	0	572	0	0	0	0	1	-360	360;
	1001	6697	0.00513	0.03014	0	514	0	0	0	0	1	-360	360;
	1001	7165	0.00113	0.00651	0	0	0	0	0	0	1	-360	360;
	1001	892	0.00125	0.00707	0	572	0	0	0	0	1	-360	360;
	7438	5334	0.003349	0.01463	0	0	0	0	0	0	1	-360	360;
	7438	1526	0.00245	0.0122	0	495	0	0	0	0	1	-360	360;
	1301	6952	0.00587	0.036521	0	400	0	0	0	0	1	-360	360;
	1301	6952	0.00569	0.03774	0	0	0	0	0	0	1	-360	360;
	1301	6952	0.006031	0.03545	0	629	0	0	0	0	1	-360	360;
	1301	7883	0.000531	0.00301	0	0	0	0	0	0	1	-360	360;
	1301	6430	0.00076	0.006099	0	514	0	0	0	0	1	-360	360;
	778	7209	0.008841	0.028919	0	0	0	0	0	0	1	-360	360;
	778	1910	0.00169	0.007091	0	0	0	0	0	0	1	-360	360;
	778	6820	0.001031	0.00361	0	0	0	0	0	0	1	-360	360;
	778	3071	0.00813	0.030961	0	0	0	0	0	0	1	-360	360;
	5297	6267	0.00013	0.00068	0	533	0	0	0	0	1	-360	360;
	3657	3855	0.000211	0.000729	0	629	0	0	0	0	1	-360	360;
	3657	3855	0.000169	0.000669	0	591	0	0	0	0	1	-360	360;
	3657	2928	0.000669	0.00251	0	495	0	0	0	0	1	-360	360;
	3657	2928	0.000711	0.00238	0	476	0	0	0	0	1	-360	360;
	3657	1172	0.000229	0.00238	0	591	0	0	0	0	1	-360	360;
	3657	9191	0.0007	0.0065	0	553	0	0	0	0	1	-360	360;
	3657	8722	0.00061	0.006031	0	514	0	0	0	0	1	-360	360;
	6472	5362	0.000281	0.0018	0	629	0	0	0	0	1	-360	360;
	6472	3962	0.000211	0.002409	0	0	0	0	0	0	1	-360	360;
	1198	1343	0.00056	0.00773	0	757	0	0	0	0	1	-360	360;
	1198	29	0.0023	0.01581	0	1218	0	0	0	0	1	-360	360;
	7473	3680	0.005039	0.032151	0	0	0	0	0	0	1	-360	360;
	7473	96	0.00364	0.01713	0	457	0	0	0	0	1	-360	360;
	7473	1868	0.002169	0.012409	0	0	0	0	0	0	1	-360	360;
	346	3184	0.00113	0.004719	0	0	0	0	0	0	1	-360	360;
	346	6570	0.00113	0.004979	0	476	0	0	0	0	1	-360	360;
	3184	4728	0.00264	0.01174	0	0	0	0	0	0	1	-360	360;
	2654	7886	0.00494	0.03181	0	476	0	0	0	0	1	-360	360;
	2654	2132	0.003711	0.023961	0	610	0	0	0	0	1	-360	360;
	4144	8874	0.001409	0.01101	0	457	0	0	0	0	1	-360	360;
	4144	1625	0.001169	0.00593	0	381	0	0	0	0	1	-360	360;
	4144	1625	0.00118	0.00587	0	0	0	0	0	0	1	-360	360;
	4144	3786	0.00264	0.012469	0	533	0	0	0	0	1	-360	360;
	4144	3786	0.002349	0.01414	0	495	0	0	0	0	1	-360	360;
	3422	4100	0.00014	0.0007	0	0	0	0	0	0	1	-360	360;
	6624	4419	3e-05	0.00054	0	1876	0	0	0	0	1	-360	360;
	6624	7159	3e-05	0.00055	0	1810	0	0	0	0	1	-360	360;
	6624	3661	4e-05	0.00058	0	1777	0	0	0	0	1	-360	360;
	6624	5486	4e-05	0.00054	0	1876	0	0	0	0	1	-360	360;
	6624	4231	0.00171	0.02197	0	1382	0	0	0	0	1	-360	360;
	6624	4231	0.00164	0.02064	0	1645	0	0	0	0	1	-360	360;
	6624	6921	0.00136	0.01756	0	1711	0	0	0	0	1	-360	360;
	6624	432	0.00097	0.00942	0	1481	0	0	0	0	1	-360	360;
	6624	432	0.00086	0.01109	0	1481	0	0	0	0	1	-360	360;
	6624	432	0.00085	0.0113	0	1448	0	0	0	0	1	-360	360;
	3353	6887	0.000979	0.0058	0	400	0	0	0	0	1	-360	360;
	2479	2050	0.00052	0.00593	0	1909	0	0	0	0	1	-360	360;
	2479	2050	0.00046	0.00672	0	1909	0	0	0	0	1	-360	360;
	5482	8704	0.001151	0.00636	0	572	0	0	0	0	1	-360	360;
	5482	8704	0.001031	0.007211	0	591	0	0	0	0	1	-360	360;
	5482	4710	0.002341	0.02286	0	533	0	0	0	0	1	-360	360;
	5482	1295	0.002281	0.02518	0	400	0	0	0	0	1	-360	360;
	5482	1910	0.005479	0.02114	0	381	0	0	0	0	1	-360	360;
	5482	2985	0.011159	0.08856	0	476	0	0	0	0	1	-360	360;
	5482	5853	0.000669	0.00249	0	400	0	0	0	0	1	-360	360;
	5482	7824	0.00193	0.01643	0	572	0	0	0	0	1	-360	360;
	5482	3071	0.00024	0.0017	0	438	0	0	0	0	1	-360	360;
	4525	195	0.010159	0.065849	0	0	0	0	0	0	1	-360	360;
	9066	7367	0.00294	0.02131	0	514	0	0	0	0	1	-360	360;
	9066	9155	0.018841	0.07469	0	0	0	0	0	0	1	-360	360;
	9066	6357	0.015599	0.052169	0	400	0	0	0	0	1	-360	360;
	7886	3126	0.005831	0.03969	0	495	0	0	0	0	1	-360	360;
	7886	8564	0.00087	0.00661	0	457	0	0	0	0	1	-360	360;
	7886	5567	0.00637	0.024659	0	381	0	0	0	0	1	-360	360;
	6416	3069	0.0027	0.020419	0	553	0	0	0	0	1	-360	360;
	6416	3876	0.000409	0.00314	0	438	0	0	0	0	1	-360	360;
	6416	8804	0.001669	0.01301	0	476	0	0	0	0	1	-360	360;
	6416	8804	0.00164	0.01325	0	495	0	0	0	0	1	-360	360;
	7831	4550	0.00089	0.00556	0	1283	0	0	0	0	1	-360	360;
	7831	4125	0.00186	0.01809	0	1613	0	0	0	0	1	-360	360;
	7831	3346	0.00381	0.02853	0	1020	0	0	0	0	1	-360	360;
	5525	2377	0.012419	0.072151	0	400	0	0	0	0	1	-360	360;
	5525	1401	0.0042	0.02556	0	400	0	0	0	0	1	-360	360;
	5525	2850	0.010771	0.0493	0	476	0	0	0	0	1	-360	360;
	5525	2430	0.010159	0.04995	0	400	0	0	0	0	1	-360	360;
	5525	9164	0.013169	0.05113	0	495	0	0	0	0	1	-360	360;
	8511	1033	0.000469	0.001159	0	400	0	0	0	0	1	-360	360;
	8511	6178	0.000229	0.00105	0	457	0	0	0	0	1	-360	360;
	3400	980	0.000271	0.001271	0	533	0	0	0	0	1	-360	360;
	2101	4598	0.00099	0.01194	0	1645	0	0	0	0	1	-360	360;
	2101	4541	0.00077	0.01184	0	0	0	0	0	0	1	-360	360;
	2101	5144	0.00034	0.00426	0	0	0	0	0	0	1	-360	360;
	3919	2129	0.00101	0.01168	0	1744	0	0	0	0	1	-360	360;
	3919	4541	0.00086	0.01046	0	1810	0	0	0	0	1	-360	360;
	3919	3513	0.00035	0.00417	0	0	0	0	0	0	1	-360	360;
	6675	1486	0.000159	0.00069	0	533	0	0	0	0	1	-360	360;
	4711	5836	0.000781	0.00488	0	629	0	0	0	0	1	-360	360;
	4711	575	0.000781	0.00511	0	533	0	0	0	0	1	-360	360;
	4711	8214	0.00064	0.003539	0	457	0	0	0	0	1	-360	360;
	4711	8214	0.00057	0.003781	0	438	0	0	0	0	1	-360	360;
	4711	3021	0.000531	0.00111	0	0	0	0	0	0	1	-360	360;
	4711	7967	0.000521	0.00113	0	0	0	0	0	0	1	-360	360;
	4711	5993	0.00176	0.00843	0	438	0	0	0	0	1	-360	360;
	4711	8804	0.002461	0.01975	0	572	0	0	0	0	1	-360	360;
	4513	6831	0.000591	0.00105	0	0	0	0	0	0	1	-360	360;
	2043	1422	0.00012	0.00064	0	0	0	0	0	0	1	-360	360;
	6714	1808	0.000581	0.004419	0	0	0	0	0	0	1	-360	360;
	6371	1808	0.000651	0.004331	0	0	0	0	0	0	1	-360	360;
	513	2866	0.001289	0.008289	0	495	0	0	0	0	1	-360	360;
	4970	9222	0.00233	0.03502	0	757	0	0	0	0	1	-360	360;
	4970	9222	0.00238	0.03781	0	757	0	0	0	0	1	-360	360;
	4970	2072	0.00052	0.00605	0	1382	0	0	0	0	1	-360	360;
	4970	5461	0.00023	0.00275	0	1514	0	0	0	0	1	-360	360;
	4970	8109	0.00134	0.01343	0	1711	0	0	0	0	1	-360	360;
	4970	8109	0.00116	0.01552	0	1481	0	0	0	0	1	-360	360;
	7955	707	0.004169	0.027919	0	514	0	0	0	0	1	-360	360;
	7955	4816	0.00063	0.008	0	0	0	0	0	0	1	-360	360;
	7955	1866	0.004841	0.028581	0	629	0	0	0	0	1	-360	360;
	7955	5488	0.007419	0.05787	0	591	0	0	0	0	1	-360	360;
	5093	2467	0.002719	0.020331	0	495	0	0	0	0	1	-360	360;
	2328	7178	0.000789	0.005	0	0	0	0	0	0	1	-360	360;
	9180	195	0.00238	0.013219	0	457	0	0	0	0	1	-360	360;
	9180	3133	0.003771	0.02701	0	495	0	0	0	0	1	-360	360;
	292	4544	0.001271	0.00939	0	0	0	0	0	0	1	-360	360;
	9002	6737	0.000841	0.00131	0	0	0	0	0	0	1	-360	360;
	8439	3758	0.005711	0.043039	0	591	0	0	0	0	1	-360	360;
	8439	3758	0.00562	0.041901	0	495	0	0	0	0	1	-360	360;
	4974	2940	0.00356	0.019781	0	533	0	0	0	0	1	-360	360;
	4974	163	0.004841	0.03149	0	533	0	0	0	0	1	-360	360;
	747	8507	0.00057	0.00107	0	0	0	0	0	0	1	-360	360;
	6909	4914	0.000539	0.00213	0	514	0	0	0	0	1	-360	360;
	444	4747	0.0007	0.00425	0	0	0	0	0	0	1	-360	360;
	1459	6495	0.0012	0.009531	0	0	0	0	0	0	1	-360	360;
	1459	2751	0.00206	0.015901	0	0	0	0	0	0	1	-360	360;
	3601	4728	0.00063	0.00312	0	0	0	0	0	0	1	-360	360;
	1251	6475	3e-05	0.00041	0	1481	0	0	0	0	1	-360	360;
	1093	6475	2e-05	0.00043	0	0	0	0	0	0	1	-360	360;
	5709	6475	3e-05	0.00039	0	1283	0	0	0	0	1	-360	360;
	3825	6475	3e-05	0.00038	0	0	0	0	0	0	1	-360	360;
	1415	3503	0.0008	0.004151	0	0	0	0	0	0	1	-360	360;
	4549	3503	0.000901	0.00381	0	0	0	0	0	0	1	-360	360;
	3072	639	0.000591	0.004341	0	476	0	0	0	0	1	-360	360;
	3072	639	0.000591	0.00405	0	495	0	0	0	0	1	-360	360;
	5007	4541	0.00218	0.0275	0	1152	0	0	0	0	1	-360	360;
	5007	8195	0.00184	0.01677	0	1086	0	0	0	0	1	-360	360;
	757	964	0.00058	0.00807	0	1086	0	0	0	0	1	-360	360;
	757	964	0.00056	0.00823	0	921	0	0	0	0	1	-360	360;
	757	3906	0.00123	0.01146	0	1415	0	0	0	0	1	-360	360;
	757	3906	0.0011	0.01298	0	1481	0	0	0	0	1	-360	360;
	757	8347	0.00156	0.01585	0	1777	0	0	0	0	1	-360	360;
	757	2919	0.00226	0.02781	0	1481	0	0	0	0	1	-360	360;
	3758	188	0.00614	0.02831	0	400	0	0	0	0	1	-360	360;
	3758	2242	0.00513	0.03401	0	0	0	0	0	0	1	-360	360;
	3903	4114	0.013219	0.0597	0	0	0	0	0	0	1	-360	360;
	5743	6990	0.012969	0.06087	0	0	0	0	0	0	1	-360	360;
	8626	8930	0.00437	0.032831	0	553	0	0	0	0	1	-360	360;
	8626	2467	0.005651	0.033229	0	591	0	0	0	0	1	-360	360;
	2161	5488	0.008969	0.03625	0	438	0	0	0	0	1	-360	360;
	2161	7772	0.012409	0.05645	0	419	0	0	0	0	1	-360	360;
	5003	6382	0.00038	0.002271	0	400	0	0	0	0	1	-360	360;
	6474	4084	0.00138	0.01089	0	572	0	0	0	0	1	-360	360;
	6474	4084	0.001539	0.012521	0	591	0	0	0	0	1	-360	360;
	6474	53	0.00205	0.014229	0	533	0	0	0	0	1	-360	360;
	6474	7259	0.00026	0.00164	0	495	0	0	0	0	1	-360	360;
	6474	2475	0.000289	0.00169	0	0	0	0	0	0	1	-360	360;
	5334	7577	0.003349	0.051211	0	457	0	0	0	0	1	-360	360;
	7961	2288	0.00038	0.001479	0	0	0	0	0	0	1	-360	360;
	1592	7098	0.00082	0.00625	0	553	0	0	0	0	1	-360	360;
	3187	2079	0.00275	0.015581	0	572	0	0	0	0	1	-360	360;
	6041	5926	0.000909	0.00739	0	0	0	0	0	0	1	-360	360;
	3697	8486	0.00082	0.01038	0	1547	0	0	0	0	1	-360	360;
	3697	7530	0.00093	0.01017	0	1382	0	0	0	0	1	-360	360;
	3697	1502	0.00137	0.01431	0	1876	0	0	0	0	1	-360	360;
	3697	1552	0.00215	0.02021	0	1678	0	0	0	0	1	-360	360;
	6952	3793	0.01189	0.058841	0	457	0	0	0	0	1	-360	360;
	6738	2526	0.00104	0.01241	0	1251	0	0	0	0	1	-360	360;
	6738	2526	0.00092	0.014	0	1251	0	0	0	0	1	-360	360;
	6738	2365	0.00055	0.00845	0	1152	0	0	0	0	1	-360	360;
	6738	2365	0.00054	0.00861	0	987	0	0	0	0	1	-360	360;
	6738	3794	0.00105	0.01057	0	1810	0	0	0	0	1	-360	360;
	6738	3613	0.00089	0.01122	0	1613	0	0	0	0	1	-360	360;
	6738	6901	0.00223	0.02584	0	1777	0	0	0	0	1	-360	360;
	8180	5106	0.00101	0.007841	0	553	0	0	0	0	1	-360	360;
	8180	5106	0.001081	0.008099	0	572	0	0	0	0	1	-360	360;
	6240	7165	0.011081	0.068479	0	0	0	0	0	0	1	-360	360;
	9222	8721	2e-05	0.00033	0	0	0	0	0	0	1	-360	360;
	9222	8872	1e-05	0.00029	0	1777	0	0	0	0	1	-360	360;
	9222	8672	0.00071	0.00892	0	1580	0	0	0	0	1	-360	360;
	9222	7164	0.00136	0.01526	0	1448	0	0	0	0	1	-360	360;
	5935	7772	0.005401	0.036021	0	591	0	0	0	0	1	-360	360;
	3134	3430	0.00437	0.033719	0	457	0	0	0	0	1	-360	360;
	3134	1051	0.005151	0.034919	0	514	0	0	0	0	1	-360	360;
	283	800	0.00038	0.00139	0	0	0	0	0	0	1	-360	360;
	283	1077	0.00037	0.0015	0	0	0	0	0	0	1	-360	360;
	283	4747	0.00069	0.004461	0	533	0	0	0	0	1	-360	360;
	3391	5213	0.00688	0.040591	0	381	0	0	0	0	1	-360	360;
	3391	115	0.001341	0.008349	0	476	0	0	0	0	1	-360	360;
	3391	115	0.001229	0.009669	0	629	0	0	0	0	1	-360	360;
	4867	5753	0.00137	0.006039	0	419	0	0	0	0	1	-360	360;
	7539	6556	0.001919	0.01261	0	476	0	0	0	0	1	-360	360;
	7539	871	0.00163	0.0095	0	533	0	0	0	0	1	-360	360;
	8893	8874	0.00488	0.0497	0	572	0	0	0	0	1	-360	360;
	8893	4505	0.001289	0.0087	0	553	0	0	0	0	1	-360	360;
	8893	3436	0.00211	0.01524	0	572	0	0	0	0	1	-360	360;
	8893	3436	0.00206	0.015539	0	591	0	0	0	0	1	-360	360;
	8893	3436	0.00276	0.01631	0	514	0	0	0	0	1	-360	360;
	8893	6556	0.00305	0.019521	0	610	0	0	0	0	1	-360	360;
	8893	8765	0.000409	0.0023	0	495	0	0	0	0	1	-360	360;
	8893	6619	0.00036	0.00268	0	514	0	0	0	0	1	-360	360;
	8165	6734	0.00032	0.00049	0	0	0	0	0	0	1	-360	360;
	1078	6734	0.00013	0.000349	0	0	0	0	0	0	1	-360	360;
	7988	4231	0.0002	0.00327	0	0	0	0	0	0	1	-360	360;
	804	2142	0.000919	0.00561	0	533	0	0	0	0	1	-360	360;
	804	2142	0.00094	0.00549	0	533	0	0	0	0	1	-360	360;
	804	7519	0.00175	0.01976	0	533	0	0	0	0	1	-360	360;
	804	6744	0.001729	0.019331	0	476	0	0	0	0	1	-360	360;
	2286	6802	0.00406	0.02581	0	572	0	0	0	0	1	-360	360;
	2286	769	0.00526	0.025961	0	514	0	0	0	0	1	-360	360;
	3239	1081	0.00179	0.02098	0	1349	0	0	0	0	1	-360	360;
	3239	5983	0.00162	0.02148	0	1382	0	0	0	0	1	-360	360;
	3239	8267	0.00492	0.04991	0	1053	0	0	0	0	1	-360	360;
	3239	960	0.00109	0.01361	0	1086	0	0	0	0	1	-360	360;
	3239	960	0.00097	0.0153	0	889	0	0	0	0	1	-360	360;
	3239	3906	0.00322	0.03022	0	1481	0	0	0	0	1	-360	360;
	1539	7361	0.032831	0.074151	0	0	0	0	0	0	1	-360	360;
	1539	6852	0.01411	0.058711	0	514	0	0	0	0	1	-360	360;
	1539	8515	0.02725	0.102979	0	438	0	0	0	0	1	-360	360;
	1539	3166	0.012229	0.04049	0	400	0	0	0	0	1	-360	360;
	1539	2676	0.01994	0.066719	0	0	0	0	0	0	1	-360	360;
	1539	3793	0.018591	0.07462	0	438	0	0	0	0	1	-360	360;
	1539	4529	0.008979	0.03251	0	0	0	0	0	0	1	-360	360;
	1539	4234	0.016531	0.086031	0	381	0	0	0	0	1	-360	360;
	8628	44	0.003591	0.02137	0	591	0	0	0	0	1	-360	360;
	8628	6630	0.00124	0.008	0	400	0	0	0	0	1	-360	360;
	8628	6630	0.001271	0.007831	0	438	0	0	0	0	1	-360	360;
	8628	1545	0.000469	0.00294	0	495	0	0	0	0	1	-360	360;
	8628	4215	0.000841	0.003289	0	495	0	0	0	0	1	-360	360;
	7163	1917	0.00219	0.016539	0	400	0	0	0	0	1	-360	360;
	7163	5519	0.002901	0.020901	0	400	0	0	0	0	1	-360	360;
	7647	225	0.0002	0.001211	0	0	0	0	0	0	1	-360	360;
	4864	1397	9.1e-05	0.000479	0	0	0	0	0	0	1	-360	360;
	7021	7119	0.005461	0.029469	0	419	0	0	0	0	1	-360	360;
	7021	3324	0.001979	0.011289	0	591	0	0	0	0	1	-360	360;
	1248	163	0.0023	0.01289	0	0	0	0	0	0	1	-360	360;
	9045	2457	0.000159	0.00037	0	0	0	0	0	0	1	-360	360;
	8874	3786	0.002789	0.01562	0	533	0	0	0	0	1	-360	360;
	5891	5574	0.002531	0.012331	0	381	0	0	0	0	1	-360	360;
	5891	7663	0.00357	0.023289	0	400	0	0	0	0	1	-360	360;
	5891	4889	0.00076	0.006039	0	514	0	0	0	0	1	-360	360;
	3997	1526	0.000469	0.00075	0	0	0	0	0	0	1	-360	360;
	184	1526	0.00044	0.00069	0	0	0	0	0	0	1	-360	360;
	2972	2088	0.00561	0.035521	0	591	0	0	0	0	1	-360	360;
	8676	933	0.004539	0.03468	0	0	0	0	0	0	1	-360	360;
	8676	933	0.00175	0.0067	0	419	0	0	0	0	1	-360	360;
	8676	3643	0.007461	0.031419	0	400	0	0	0	0	1	-360	360;
	8704	6723	0.00224	0.012909	0	610	0	0	0	0	1	-360	360;
	7047	21	0.0003	0.000961	0	419	0	0	0	0	1	-360	360;
	7047	21	0.00032	0.001409	0	0	0	0	0	0	1	-360	360;
	7047	1551	0.002031	0.0157	0	553	0	0	0	0	1	-360	360;
	7047	4103	0.00038	0.00306	0	514	0	0	0	0	1	-360	360;
	7047	7862	0.00118	0.0083	0	553	0	0	0	0	1	-360	360;
	7047	5458	0.001229	0.00787	0	572	0	0	0	0	1	-360	360;
	5574	5666	0.00301	0.01538	0	514	0	0	0	0	1	-360	360;
	6742	5458	0.00063	0.00114	0	381	0	0	0	0	1	-360	360;
	5270	7862	0.00064	0.00111	0	0	0	0	0	0	1	-360	360;
	7775	3857	0.004919	0.03837	0	629	0	0	0	0	1	-360	360;
	6290	3436	0.001831	0.008349	0	438	0	0	0	0	1	-360	360;
	3430	401	0.00536	0.03188	0	533	0	0	0	0	1	-360	360;
	2967	8976	0.00038	0.001461	0	0	0	0	0	0	1	-360	360;
	2967	8976	0.000289	0.001521	0	0	0	0	0	0	1	-360	360;
	8976	2481	0.00305	0.016841	0	476	0	0	0	0	1	-360	360;
	8976	9108	0.001781	0.012729	0	457	0	0	0	0	1	-360	360;
	2056	7520	0.00041	0.00461	0	954	0	0	0	0	1	-360	360;
	2056	7520	0.0004	0.0047	0	1020	0	0	0	0	1	-360	360;
	4084	5340	0.00213	0.01401	0	629	0	0	0	0	1	-360	360;
	4084	2319	0.001271	0.00887	0	438	0	0	0	0	1	-360	360;
	4084	2878	0.00462	0.02743	0	533	0	0	0	0	1	-360	360;
	2597	7422	0.001539	0.00776	0	0	0	0	0	0	1	-360	360;
	7183	8931	0.00025	0.00478	0	0	0	0	0	0	1	-360	360;
	5379	8931	0.00025	0.00487	0	0	0	0	0	0	1	-360	360;
	8128	8931	0.00025	0.00435	0	823	0	0	0	0	1	-360	360;
	6337	8931	0.00026	0.00426	0	757	0	0	0	0	1	-360	360;
	1237	8931	0.00025	0.00479	0	0	0	0	0	0	1	-360	360;
	4127	9173	0.000169	0.00076	0	438	0	0	0	0	1	-360	360;
	4127	7974	0.00039	0.00126	0	495	0	0	0	0	1	-360	360;
	2432	4816	0.011229	0.074659	0	629	0	0	0	0	1	-360	360;
	2432	1866	0.014521	0.058081	0	400	0	0	0	0	1	-360	360;
	2432	8809	0.00064	0.003919	0	533	0	0	0	0	1	-360	360;
	2432	5146	0.014789	0.07768	0	438	0	0	0	0	1	-360	360;
	2432	6940	0.00938	0.05355	0	572	0	0	0	0	1	-360	360;
	2432	1742	0.00124	0.004289	0	419	0	0	0	0	1	-360	360;
	5213	8627	0.00968	0.04388	0	0	0	0	0	0	1	-360	360;
	5213	174	0.008531	0.067151	0	457	0	0	0	0	1	-360	360;
	5213	7437	0.00731	0.04106	0	438	0	0	0	0	1	-360	360;
	7342	5940	0.004349	0.02555	0	533	0	0	0	0	1	-360	360;
	7342	7495	0.00599	0.03462	0	476	0	0	0	0	1	-360	360;
	8808	4816	0.002961	0.021591	0	553	0	0	0	0	1	-360	360;
	8808	4816	0.002901	0.02206	0	572	0	0	0	0	1	-360	360;
	8808	6053	0.002979	0.016969	0	476	0	0	0	0	1	-360	360;
	5764	4189	0.00312	0.020021	0	476	0	0	0	0	1	-360	360;
	8507	1973	0.000219	0.000909	0	0	0	0	0	0	1	-360	360;
	1394	1973	0.000219	0.00093	0	457	0	0	0	0	1	-360	360;
	2898	2128	0.000151	0.000289	0	0	0	0	0	0	1	-360	360;
	4942	8255	0.00563	0.043599	0	0	0	0	0	0	1	-360	360;
	4942	4748	0.000969	0.00612	0	514	0	0	0	0	1	-360	360;
	3486	3594	0.005021	0.020719	0	438	0	0	0	0	1	-360	360;
	3486	4679	0.00988	0.041039	0	0	0	0	0	0	1	-360	360;
	4656	4235	0.003159	0.02011	0	476	0	0	0	0	1	-360	360;
	4656	4235	0.003219	0.01969	0	476	0	0	0	0	1	-360	360;
	7129	5441	0.00049	0.00238	0	0	0	0	0	0	1	-360	360;
	7129	5441	0.000479	0.002419	0	0	0	0	0	0	1	-360	360;
	8250	5720	0.00264	0.019729	0	533	0	0	0	0	1	-360	360;
	8250	5720	0.002581	0.02011	0	495	0	0	0	0	1	-360	360;
	6831	6723	0.00037	0.002831	0	553	0	0	0	0	1	-360	360;
	6831	5686	0.00213	0.012789	0	514	0	0	0	0	1	-360	360;
	2377	805	0.018669	0.092281	0	438	0	0	0	0	1	-360	360;
	907	9091	3.9e-05	0.000401	0	0	0	0	0	0	1	-360	360;
	4056	9091	6e-05	0.00044	0	381	0	0	0	0	1	-360	360;
	6837	2142	6e-05	0.0005	0	514	0	0	0	0	1	-360	360;
	5666	2641	0.001271	0.007031	0	476	0	0	0	0	1	-360	360;
	5666	7577	0.00055	0.003159	0	610	0	0	0	0	1	-360	360;
	5666	7577	0.00055	0.00338	0	591	0	0	0	0	1	-360	360;
	5666	7577	0.00105	0.004771	0	495	0	0	0	0	1	-360	360;
	5666	1629	0.003289	0.015039	0	381	0	0	0	0	1	-360	360;
	726	687	0.00036	0.004349	0	610	0	0	0	0	1	-360	360;
	3526	687	0.00038	0.003669	0	0	0	0	0	0	1	-360	360;
	6382	2425	0.00118	0.00701	0	0	0	0	0	0	1	-360	360;
	6382	2425	0.001229	0.00686	0	533	0	0	0	0	1	-360	360;
	6382	5907	0.00301	0.0217	0	0	0	0	0	0	1	-360	360;
	6382	5907	0.003409	0.01924	0	0	0	0	0	0	1	-360	360;
	6382	5907	0.003409	0.01938	0	572	0	0	0	0	1	-360	360;
	7847	9217	0.002719	0.02131	0	0	0	0	0	0	1	-360	360;
	2981	1448	0.00295	0.01956	0	0	0	0	0	0	1	-360	360;
	5110	5686	0.001091	0.00831	0	476	0	0	0	0	1	-360	360;
	2968	5907	0.00111	0.008211	0	553	0	0	0	0	1	-360	360;
	5441	2155	0.002469	0.015271	0	629	0	0	0	0	1	-360	360;
	5441	8112	0.00212	0.01601	0	0	0	0	0	0	1	-360	360;
	5441	2535	0.00037	0.0028	0	533	0	0	0	0	1	-360	360;
	5441	7098	0.00162	0.01238	0	0	0	0	0	0	1	-360	360;
	5441	7098	0.001909	0.01162	0	0	0	0	0	0	1	-360	360;
	5441	7098	0.001909	0.010729	0	553	0	0	0	0	1	-360	360;
	5441	3737	0.00089	0.00526	0	419	0	0	0	0	1	-360	360;
	5441	5529	0.00087	0.00536	0	0	0	0	0	0	1	-360	360;
	5441	5753	0.001521	0.01176	0	533	0	0	0	0	1	-360	360;
	9019	6563	0.00014	0.000591	0	476	0	0	0	0	1	-360	360;
	1341	1101	0.003531	0.028781	0	0	0	0	0	0	1	-360	360;
	2797	8291	0.004599	0.03657	0	553	0	0	0	0	1	-360	360;
	2797	8367	0.00307	0.019159	0	591	0	0	0	0	1	-360	360;
	2797	1100	0.000581	0.004461	0	514	0	0	0	0	1	-360	360;
	2797	6887	0.00164	0.01161	0	591	0	0	0	0	1	-360	360;
	4939	6880	0.00057	0.003281	0	533	0	0	0	0	1	-360	360;
	7738	6982	0.020419	0.063669	0	0	0	0	0	0	1	-360	360;
	6565	3121	0.01424	0.04751	0	0	0	0	0	0	1	-360	360;
	8707	280	0.00157	0.008849	0	553	0	0	0	0	1	-360	360;
	8707	2510	0.01218	0.07524	0	0	0	0	0	0	1	-360	360;
	1625	766	0.00037	0.00151	0	0	0	0	0	0	1	-360	360;
	8411	6232	0.000219	0.000531	0	0	0	0	0	0	1	-360	360;
	3951	1397	0.002849	0.0137	0	0	0	0	0	0	1	-360	360;
	964	7282	0.00116	0.01692	0	1283	0	0	0	0	1	-360	360;
	964	7282	0.00133	0.01462	0	1415	0	0	0	0	1	-360	360;
	964	6475	0.00063	0.00821	0	987	0	0	0	0	1	-360	360;
	964	6475	0.00064	0.00804	0	1283	0	0	0	0	1	-360	360;
	5658	5420	0.00126	0.01272	0	1382	0	0	0	0	1	-360	360;
	5658	8267	0.00337	0.0494	0	1744	0	0	0	0	1	-360	360;
	5658	7328	0.0014	0.0214	0	0	0	0	0	0	1	-360	360;
	5658	7328	0.00137	0.02182	0	0	0	0	0	0	1	-360	360;
	5658	7164	0.00212	0.02063	0	1218	0	0	0	0	1	-360	360;
	5550	410	0.00186	0.01045	0	610	0	0	0	0	1	-360	360;
	5550	7579	0.00426	0.021771	0	438	0	0	0	0	1	-360	360;
	5550	7579	0.002659	0.020539	0	495	0	0	0	0	1	-360	360;
	5550	4189	0.00606	0.049919	0	457	0	0	0	0	1	-360	360;
	7282	4196	0.00109	0.01192	0	889	0	0	0	0	1	-360	360;
	1526	3916	0.013021	0.049219	0	438	0	0	0	0	1	-360	360;
	1526	4747	0.00236	0.01239	0	438	0	0	0	0	1	-360	360;
	1526	4747	0.0027	0.015169	0	419	0	0	0	0	1	-360	360;
	4541	3346	0.00285	0.03283	0	1514	0	0	0	0	1	-360	360;
	4541	3346	0.00529	0.03969	0	1020	0	0	0	0	1	-360	360;
	1343	3346	0.00254	0.03721	0	1547	0	0	0	0	1	-360	360;
	3680	2719	0.000159	0.00069	0	0	0	0	0	0	1	-360	360;
	3680	2719	0.000229	0.000979	0	0	0	0	0	0	1	-360	360;
	3680	2719	0.00024	0.00082	0	0	0	0	0	0	1	-360	360;
	3680	795	0.000169	0.00075	0	0	0	0	0	0	1	-360	360;
	3680	4880	0.00368	0.01756	0	0	0	0	0	0	1	-360	360;
	3680	1868	0.002599	0.020659	0	0	0	0	0	0	1	-360	360;
	7464	5564	0.00018	0.00064	0	0	0	0	0	0	1	-360	360;
	7464	7036	0.000159	0.000771	0	381	0	0	0	0	1	-360	360;
	7464	5940	0.00357	0.01981	0	476	0	0	0	0	1	-360	360;
	7464	7495	0.006021	0.03205	0	533	0	0	0	0	1	-360	360;
	4850	2886	0.001781	0.01164	0	533	0	0	0	0	1	-360	360;
	4850	7049	0.002781	0.018831	0	514	0	0	0	0	1	-360	360;
	2510	6146	0.000349	0.00281	0	0	0	0	0	0	1	-360	360;
	90	1860	0.00313	0.02444	0	610	0	0	0	0	1	-360	360;
	90	1860	0.00368	0.02074	0	591	0	0	0	0	1	-360	360;
	2142	3276	9.1e-05	0.001539	0	0	0	0	0	0	1	-360	360;
	2142	3545	0.00251	0.01111	0	0	0	0	0	0	1	-360	360;
	2142	6842	0.003081	0.017961	0	419	0	0	0	0	1	-360	360;
	2142	6744	0.005341	0.030419	0	438	0	0	0	0	1	-360	360;
	2142	7635	9.9e-05	0.00126	0	514	0	0	0	0	1	-360	360;
	9091	1040	0.002169	0.017901	0	419	0	0	0	0	1	-360	360;
	2457	6178	0.00011	0.00038	0	0	0	0	0	0	1	-360	360;
	1422	1462	0.00212	0.00494	0	0	0	0	0	0	1	-360	360;
	1422	7772	0.018901	0.077349	0	476	0	0	0	0	1	-360	360;
	7881	5365	0.00957	0.0395	0	457	0	0	0	0	1	-360	360;
	7881	6982	0.02324	0.063469	0	400	0	0	0	0	1	-360	360;
	6926	3327	0.000531	0.004331	0	610	0	0	0	0	1	-360	360;
	1959	615	0.002581	0.006229	0	0	0	0	0	0	1	-360	360;
	6455	1262	0.00031	0.002211	0	419	0	0	0	0	1	-360	360;
	803	6982	0.01087	0.06114	0	438	0	0	0	0	1	-360	360;
	803	5918	0.011581	0.061099	0	419	0	0	0	0	1	-360	360;
	7624	3070	0.002099	0.016271	0	400	0	0	0	0	1	-360	360;
	500	1547	0.00093	0.005461	0	0	0	0	0	0	1	-360	360;
	500	5237	0.009591	0.037039	0	419	0	0	0	0	1	-360	360;
	500	772	0.002591	0.017211	0	553	0	0	0	0	1	-360	360;
	500	8950	0.00164	0.013651	0	495	0	0	0	0	1	-360	360;
	2754	2079	0.003219	0.014719	0	457	0	0	0	0	1	-360	360;
	2754	2229	0.00401	0.022229	0	457	0	0	0	0	1	-360	360;
	2754	7049	0.00126	0.010211	0	591	0	0	0	0	1	-360	360;
	4118	7873	0.009169	0.03993	0	0	0	0	0	0	1	-360	360;
	4118	3865	0.001349	0.006081	0	0	0	0	0	0	1	-360	360;
	1754	8486	0.00156	0.01931	0	1580	0	0	0	0	1	-360	360;
	1754	7530	0.00152	0.01968	0	1481	0	0	0	0	1	-360	360;
	1754	7778	0.00094	0.01433	0	1678	0	0	0	0	1	-360	360;
	1754	6897	0.00092	0.01461	0	1448	0	0	0	0	1	-360	360;
	1754	960	0.00137	0.02012	0	1119	0	0	0	0	1	-360	360;
	1754	960	0.00134	0.02051	0	921	0	0	0	0	1	-360	360;
	3857	4491	0.00314	0.042151	0	572	0	0	0	0	1	-360	360;
	3857	6010	0.000651	0.00494	0	0	0	0	0	0	1	-360	360;
	3857	6940	0.003961	0.02426	0	591	0	0	0	0	1	-360	360;
	858	1642	0.00013	0.0015	0	0	0	0	0	0	1	-360	360;
	3028	809	0.00013	0.00147	0	1876	0	0	0	0	1	-360	360;
	4819	1642	0.00011	0.00172	0	1876	0	0	0	0	1	-360	360;
	2035	809	0.00012	0.00166	0	0	0	0	0	0	1	-360	360;
	3869	1642	0.00012	0.00168	0	1909	0	0	0	0	1	-360	360;
	3741	809	0.00012	0.00172	0	1909	0	0	0	0	1	-360	360;
	742	687	0.00093	0.005669	0	0	0	0	0	0	1	-360	360;
	2910	7098	0.00105	0.00506	0	0	0	0	0	0	1	-360	360;
	5688	6231	0.00113	0.006539	0	0	0	0	0	0	1	-360	360;
	5182	5455	0.00037	0.002919	0	0	0	0	0	0	1	-360	360;
	707	5695	0.005169	0.03738	0	476	0	0	0	0	1	-360	360;
	3814	954	0.00125	0.006669	0	457	0	0	0	0	1	-360	360;
	2313	2438	0.001281	0.00661	0	0	0	0	0	0	1	-360	360;
	195	2985	0.00226	0.0138	0	457	0	0	0	0	1	-360	360;
	2518	3513	0.00213	0.02677	0	1613	0	0	0	0	1	-360	360;
	2518	3346	0.00154	0.01919	0	1349	0	0	0	0	1	-360	360;
	4544	601	0.00687	0.04413	0	0	0	0	0	0	1	-360	360;
	4544	3241	0.0045	0.04725	0	457	0	0	0	0	1	-360	360;
	4544	7338	0.00138	0.00837	0	0	0	0	0	0	1	-360	360;
	3069	1538	0.000419	0.001901	0	0	0	0	0	0	1	-360	360;
	3069	1538	0.00043	0.00186	0	0	0	0	0	0	1	-360	360;
	6115	4729	0.003331	0.015099	0	591	0	0	0	0	1	-360	360;
	3344	6616	0.002099	0.01631	0	533	0	0	0	0	1	-360	360;
	3809	5616	0.001331	0.01082	0	591	0	0	0	0	1	-360	360;
	3809	6691	0.00199	0.00818	0	381	0	0	0	0	1	-360	360;
	3809	6691	0.00169	0.008831	0	0	0	0	0	0	1	-360	360;
	1326	892	0.001771	0.014281	0	0	0	0	0	0	1	-360	360;
	601	616	0.003099	0.01844	0	495	0	0	0	0	1	-360	360;
	3496	1551	0.003409	0.01819	0	438	0	0	0	0	1	-360	360;
	4925	2230	0.00057	0.00369	0	0	0	0	0	0	1	-360	360;
	8057	5907	0.001081	0.00468	0	0	0	0	0	0	1	-360	360;
	5354	5907	0.00105	0.004771	0	0	0	0	0	0	1	-360	360;
	3674	616	0.001979	0.01549	0	591	0	0	0	0	1	-360	360;
	3674	7338	0.001469	0.015151	0	0	0	0	0	0	1	-360	360;
	5836	5083	0.000711	0.003969	0	419	0	0	0	0	1	-360	360;
	5836	7576	0.00051	0.00274	0	0	0	0	0	0	1	-360	360;
	5836	3191	0.000521	0.00306	0	495	0	0	0	0	1	-360	360;
	5836	3488	0.00188	0.011841	0	0	0	0	0	0	1	-360	360;
	5836	1578	0.002469	0.010539	0	0	0	0	0	0	1	-360	360;
	575	5300	0.00045	0.00351	0	514	0	0	0	0	1	-360	360;
	3929	9173	0.00037	0.000659	0	0	0	0	0	0	1	-360	360;
	2089	3818	0.000401	0.0007	0	0	0	0	0	0	1	-360	360;
	2089	6232	0.00014	0.000531	0	0	0	0	0	0	1	-360	360;
	2089	1643	0.000289	0.0013	0	495	0	0	0	0	1	-360	360;
	2089	6253	7e-05	0.000289	0	0	0	0	0	0	1	-360	360;
	2089	6684	0.00051	0.002151	0	381	0	0	0	0	1	-360	360;
	2089	8151	0.00037	0.00157	0	533	0	0	0	0	1	-360	360;
	5067	449	0.00022	0.00277	0	1448	0	0	0	0	1	-360	360;
	5067	449	0.00025	0.00271	0	1547	0	0	0	0	1	-360	360;
	7641	3037	0.000271	0.002081	0	572	0	0	0	0	1	-360	360;
	7641	3037	0.00025	0.00231	0	533	0	0	0	0	1	-360	360;
	3191	2340	0.00095	0.00331	0	0	0	0	0	0	1	-360	360;
	3191	8711	0.00093	0.00336	0	0	0	0	0	0	1	-360	360;
	8104	8722	0.000341	0.002211	0	591	0	0	0	0	1	-360	360;
	2644	9191	0.00032	0.00232	0	0	0	0	0	0	1	-360	360;
	4128	8522	3.9e-05	0.000271	0	514	0	0	0	0	1	-360	360;
	604	7522	0.00075	0.0101	0	1613	0	0	0	0	1	-360	360;
	604	4000	0.00118	0.01455	0	1547	0	0	0	0	1	-360	360;
	2155	9021	0.0017	0.0098	0	514	0	0	0	0	1	-360	360;
	2155	9021	0.001521	0.01111	0	0	0	0	0	0	1	-360	360;
	2155	2042	0.000169	0.001401	0	0	0	0	0	0	1	-360	360;
	2155	687	0.002591	0.01725	0	610	0	0	0	0	1	-360	360;
	2155	3520	0.000169	0.001159	0	438	0	0	0	0	1	-360	360;
	2155	4729	0.00307	0.01339	0	533	0	0	0	0	1	-360	360;
	2641	4889	0.002591	0.013521	0	514	0	0	0	0	1	-360	360;
	9067	9051	0.000581	0.005031	0	553	0	0	0	0	1	-360	360;
	6969	5573	0.001419	0.00595	0	381	0	0	0	0	1	-360	360;
	666	1494	0.00205	0.02664	0	1448	0	0	0	0	1	-360	360;
	666	2732	0.00295	0.03721	0	1514	0	0	0	0	1	-360	360;
	666	4550	0.00335	0.03648	0	1481	0	0	0	0	1	-360	360;
	666	1607	0.00022	0.00251	0	1251	0	0	0	0	1	-360	360;
	666	2093	0.00059	0.00651	0	1645	0	0	0	0	1	-360	360;
	3565	2177	0.008539	0.041979	0	381	0	0	0	0	1	-360	360;
	3565	2841	0.001349	0.0052	0	400	0	0	0	0	1	-360	360;
	3611	7972	0.00245	0.013909	0	514	0	0	0	0	1	-360	360;
	3611	2481	0.001469	0.008419	0	553	0	0	0	0	1	-360	360;
	5420	7164	0.00084	0.00836	0	1349	0	0	0	0	1	-360	360;
	7119	2078	0.002591	0.017789	0	514	0	0	0	0	1	-360	360;
	7119	7873	0.009651	0.05968	0	419	0	0	0	0	1	-360	360;
	7422	7519	0.00143	0.00799	0	381	0	0	0	0	1	-360	360;
	7422	6744	0.001969	0.014909	0	572	0	0	0	0	1	-360	360;
	1026	1860	0.01039	0.06406	0	495	0	0	0	0	1	-360	360;
	4765	3327	0.000719	0.005331	0	438	0	0	0	0	1	-360	360;
	6036	6921	0.00052	0.00565	0	1415	0	0	0	0	1	-360	360;
	8670	7069	0.002409	0.019479	0	495	0	0	0	0	1	-360	360;
	8670	7069	0.00256	0.019151	0	419	0	0	0	0	1	-360	360;
	8670	2866	0.00656	0.02288	0	381	0	0	0	0	1	-360	360;
	4032	6450	0.0027	0.014901	0	400	0	0	0	0	1	-360	360;
	4032	5738	0.003591	0.01362	0	400	0	0	0	0	1	-360	360;
	7327	2481	0.002211	0.01188	0	0	0	0	0	0	1	-360	360;
	7327	8189	0.00068	0.00357	0	400	0	0	0	0	1	-360	360;
	7474	1584	0.000211	0.000969	0	400	0	0	0	0	1	-360	360;
	8947	3390	9.9e-05	0.000539	0	0	0	0	0	0	1	-360	360;
	1808	6384	0.00645	0.0263	0	419	0	0	0	0	1	-360	360;
	4624	6031	0.00063	0.001901	0	0	0	0	0	0	1	-360	360;
	1721	3558	0.000281	0.000669	0	0	0	0	0	0	1	-360	360;
	7694	3865	0.000219	0.00064	0	0	0	0	0	0	1	-360	360;
	9181	6723	0.00014	0.00105	0	0	0	0	0	0	1	-360	360;
	8373	8748	0.003219	0.020729	0	533	0	0	0	0	1	-360	360;
	8373	8475	0.00051	0.003151	0	572	0	0	0	0	1	-360	360;
	8373	8475	0.000409	0.00331	0	514	0	0	0	0	1	-360	360;
	5630	6639	0.002289	0.017479	0	591	0	0	0	0	1	-360	360;
	5630	1187	0.002831	0.01895	0	533	0	0	0	0	1	-360	360;
	5630	7274	0.006289	0.036521	0	0	0	0	0	0	1	-360	360;
	5630	8992	0.005651	0.032169	0	438	0	0	0	0	1	-360	360;
	5630	7266	0.00414	0.02262	0	476	0	0	0	0	1	-360	360;
	3492	6734	0.01095	0.05814	0	457	0	0	0	0	1	-360	360;
	3492	1015	0.009	0.0368	0	495	0	0	0	0	1	-360	360;
	3492	8310	0.000229	0.00181	0	553	0	0	0	0	1	-360	360;
	1980	6570	0.00119	0.00726	0	610	0	0	0	0	1	-360	360;
	9137	749	0.011771	0.051669	0	629	0	0	0	0	1	-360	360;
	7522	2183	0.00209	0.02279	0	1711	0	0	0	0	1	-360	360;
	7571	3916	0.002031	0.01443	0	0	0	0	0	0	1	-360	360;
	7571	6891	0.00145	0.008479	0	553	0	0	0	0	1	-360	360;
	2526	6199	0.00124	0.01433	0	1514	0	0	0	0	1	-360	360;
	2526	6199	0.00127	0.01402	0	1876	0	0	0	0	1	-360	360;
	1483	3649	0.013039	0.052841	0	629	0	0	0	0	1	-360	360;
	2365	6199	0.00235	0.02739	0	954	0	0	0	0	1	-360	360;
	2365	6199	0.0024	0.02681	0	1218	0	0	0	0	1	-360	360;
	2365	4141	0.00135	0.02056	0	757	0	0	0	0	1	-360	360;
	2365	1311	0.00087	0.01124	0	954	0	0	0	0	1	-360	360;
	8316	174	0.0012	0.00818	0	629	0	0	0	0	1	-360	360;
	174	6430	0.006961	0.05561	0	495	0	0	0	0	1	-360	360;
	5340	7892	0.00105	0.00763	0	476	0	0	0	0	1	-360	360;
	5340	561	0.000219	0.00182	0	0	0	0	0	0	1	-360	360;
	5340	8843	3.9e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	5340	8843	4.5e-05	0.000195	0	0	0	0	0	0	1	-360	360;
	5340	4239	0.0035	0.022651	0	533	0	0	0	0	1	-360	360;
	3866	1860	0.004281	0.02319	0	457	0	0	0	0	1	-360	360;
	3866	5395	0.003729	0.014021	0	419	0	0	0	0	1	-360	360;
	7042	2177	0.00212	0.01582	0	572	0	0	0	0	1	-360	360;
	8672	7164	0.00056	0.00784	0	757	0	0	0	0	1	-360	360;
	1860	7772	0.004401	0.01813	0	0	0	0	0	0	1	-360	360;
	1860	7772	0.005021	0.02144	0	0	0	0	0	0	1	-360	360;
	4885	2928	0.00018	0.000219	0	381	0	0	0	0	1	-360	360;
	4157	2928	0.000151	0.000211	0	0	0	0	0	0	1	-360	360;
	4432	2928	0.000151	0.000211	0	0	0	0	0	0	1	-360	360;
	2072	5983	0.00282	0.02591	0	1481	0	0	0	0	1	-360	360;
	3450	6552	0.00131	0.008539	0	533	0	0	0	0	1	-360	360;
	3956	7507	0.002159	0.015969	0	0	0	0	0	0	1	-360	360;
	5286	1838	0.00019	0.000281	0	0	0	0	0	0	1	-360	360;
	5571	2230	0.0002	0.000281	0	0	0	0	0	0	1	-360	360;
	2078	8005	0.00139	0.00774	0	572	0	0	0	0	1	-360	360;
	639	6351	0.01101	0.06739	0	629	0	0	0	0	1	-360	360;
	639	2749	0.00063	0.00463	0	438	0	0	0	0	1	-360	360;
	639	9130	0.002341	0.009599	0	476	0	0	0	0	1	-360	360;
	410	4189	0.00375	0.021659	0	476	0	0	0	0	1	-360	360;
	1494	5781	0.00056	0.00749	0	1283	0	0	0	0	1	-360	360;
	583	6616	0.00118	0.00926	0	533	0	0	0	0	1	-360	360;
	583	6031	0.002039	0.013081	0	514	0	0	0	0	1	-360	360;
	1746	7507	0.00155	0.010961	0	0	0	0	0	0	1	-360	360;
	3276	7635	3.9e-05	0.00036	0	419	0	0	0	0	1	-360	360;
	1105	196	0.00024	0.000409	0	381	0	0	0	0	1	-360	360;
	8522	7132	0.000341	0.001659	0	553	0	0	0	0	1	-360	360;
	8522	5351	0.005461	0.024919	0	457	0	0	0	0	1	-360	360;
	8763	7056	0.00158	0.02413	0	954	0	0	0	0	1	-360	360;
	8763	8825	0.00155	0.02458	0	856	0	0	0	0	1	-360	360;
	8763	6889	0.00021	0.00256	0	1415	0	0	0	0	1	-360	360;
	8763	6889	0.00024	0.00243	0	1514	0	0	0	0	1	-360	360;
	8763	8887	0.00144	0.02266	0	1514	0	0	0	0	1	-360	360;
	8763	8887	0.0014	0.02098	0	1547	0	0	0	0	1	-360	360;
	8763	8487	0.00086	0.01074	0	0	0	0	0	0	1	-360	360;
	8748	8475	0.003341	0.02611	0	553	0	0	0	0	1	-360	360;
	217	6785	0.00387	0.02063	0	514	0	0	0	0	1	-360	360;
	217	6785	0.00405	0.018271	0	514	0	0	0	0	1	-360	360;
	217	1179	0.00101	0.00462	0	400	0	0	0	0	1	-360	360;
	217	2575	0.00088	0.00389	0	419	0	0	0	0	1	-360	360;
	217	7274	0.005081	0.039	0	514	0	0	0	0	1	-360	360;
	7972	8189	0.00257	0.0208	0	553	0	0	0	0	1	-360	360;
	7972	8189	0.00251	0.01926	0	514	0	0	0	0	1	-360	360;
	7972	1039	0.002531	0.01314	0	0	0	0	0	0	1	-360	360;
	7253	7840	0.00132	0.00869	0	0	0	0	0	0	1	-360	360;
	4395	8913	0.00082	0.004021	0	419	0	0	0	0	1	-360	360;
	4395	4852	0.00143	0.007479	0	438	0	0	0	0	1	-360	360;
	2795	1183	0.000659	0.005771	0	0	0	0	0	0	1	-360	360;
	2795	1183	0.000591	0.006539	0	0	0	0	0	0	1	-360	360;
	1305	6291	0.00589	0.025599	0	476	0	0	0	0	1	-360	360;
	1305	6982	0.00614	0.02343	0	438	0	0	0	0	1	-360	360;
	2732	1798	0.00155	0.01994	0	1481	0	0	0	0	1	-360	360;
	2732	3346	0.00205	0.03218	0	1580	0	0	0	0	1	-360	360;
	4550	3346	0.00233	0.03154	0	1580	0	0	0	0	1	-360	360;
	4725	5738	0.002341	0.01982	0	0	0	0	0	0	1	-360	360;
	1083	8310	8.1e-05	0.000521	0	533	0	0	0	0	1	-360	360;
	1398	5567	9.9e-05	0.00069	0	591	0	0	0	0	1	-360	360;
	4281	4580	0.00276	0.016419	0	553	0	0	0	0	1	-360	360;
	7050	4908	0.00036	0.001229	0	457	0	0	0	0	1	-360	360;
	7050	3740	0.000229	0.000711	0	514	0	0	0	0	1	-360	360;
	6563	870	0.00013	0.00099	0	514	0	0	0	0	1	-360	360;
	8999	6495	0.001159	0.00689	0	419	0	0	0	0	1	-360	360;
	6888	5994	0.008531	0.03045	0	381	0	0	0	0	1	-360	360;
	5400	8005	0.000781	0.004531	0	514	0	0	0	0	1	-360	360;
	5400	8005	0.00075	0.00474	0	533	0	0	0	0	1	-360	360;
	5881	4031	0.002159	0.008349	0	0	0	0	0	0	1	-360	360;
	8195	8334	0.00162	0.01651	0	1547	0	0	0	0	1	-360	360;
	933	6357	0.01236	0.044211	0	400	0	0	0	0	1	-360	360;
	933	6357	0.00825	0.03989	0	419	0	0	0	0	1	-360	360;
	933	3775	0.02187	0.066961	0	0	0	0	0	0	1	-360	360;
	933	3975	0.01212	0.06661	0	0	0	0	0	0	1	-360	360;
	3918	1642	0.00036	0.00534	0	889	0	0	0	0	1	-360	360;
	6802	1153	0.000229	0.00063	0	0	0	0	0	0	1	-360	360;
	6802	7599	0.000219	0.000581	0	0	0	0	0	0	1	-360	360;
	4025	3994	0.000281	0.00063	0	0	0	0	0	0	1	-360	360;
	1595	1414	0.005021	0.035659	0	419	0	0	0	0	1	-360	360;
	3114	3656	0.00174	0.007901	0	0	0	0	0	0	1	-360	360;
	5237	2229	0.003919	0.020581	0	495	0	0	0	0	1	-360	360;
	5237	9108	0.003341	0.016349	0	495	0	0	0	0	1	-360	360;
	5237	1478	0.003271	0.016669	0	457	0	0	0	0	1	-360	360;
	2799	8515	0.000151	0.000659	0	419	0	0	0	0	1	-360	360;
	6405	7873	0.00736	0.035659	0	0	0	0	0	0	1	-360	360;
	6405	3894	0.00944	0.042169	0	0	0	0	0	0	1	-360	360;
	4435	4734	0.00068	0.00652	0	856	0	0	0	0	1	-360	360;
	4435	3483	0.00188	0.02281	0	1382	0	0	0	0	1	-360	360;
	2079	7052	0.00168	0.00956	0	533	0	0	0	0	1	-360	360;
	2079	7052	0.001669	0.00981	0	514	0	0	0	0	1	-360	360;
	2079	772	0.002229	0.01599	0	591	0	0	0	0	1	-360	360;
	2079	772	0.00218	0.01631	0	610	0	0	0	0	1	-360	360;
	1081	6475	0.00238	0.02124	0	1053	0	0	0	0	1	-360	360;
	1081	6475	0.00194	0.02023	0	1251	0	0	0	0	1	-360	360;
	7361	3166	0.013831	0.05955	0	0	0	0	0	0	1	-360	360;
	7361	426	0.00876	0.04875	0	419	0	0	0	0	1	-360	360;
	2863	1857	0.0002	0.00099	0	400	0	0	0	0	1	-360	360;
	1364	2430	0.00105	0.00826	0	476	0	0	0	0	1	-360	360;
	2177	749	0.0172	0.06637	0	476	0	0	0	0	1	-360	360;
	3594	2850	0.00305	0.015781	0	476	0	0	0	0	1	-360	360;
	6555	3608	0.00208	0.02105	0	1481	0	0	0	0	1	-360	360;
	8475	1965	0.00614	0.043729	0	476	0	0	0	0	1	-360	360;
	8475	5926	0.000229	0.001849	0	553	0	0	0	0	1	-360	360;
	5735	6053	0.00386	0.025789	0	438	0	0	0	0	1	-360	360;
	5735	2341	0.002031	0.012	0	591	0	0	0	0	1	-360	360;
	2629	7507	0.000591	0.004	0	495	0	0	0	0	1	-360	360;
	1137	2406	9.1e-05	0.00051	0	0	0	0	0	0	1	-360	360;
	7577	5814	0.00732	0.039969	0	419	0	0	0	0	1	-360	360;
	7577	3231	0.004581	0.02031	0	629	0	0	0	0	1	-360	360;
	7663	5814	0.00132	0.00537	0	457	0	0	0	0	1	-360	360;
	1183	1159	0.007521	0.0387	0	476	0	0	0	0	1	-360	360;
	6852	5051	0.005159	0.031281	0	533	0	0	0	0	1	-360	360;
	8307	8807	9.9e-05	0.001159	0	0	0	0	0	0	1	-360	360;
	6947	8807	9.9e-05	0.00113	0	0	0	0	0	0	1	-360	360;
	4505	4357	0.00019	0.000909	0	381	0	0	0	0	1	-360	360;
	7144	7791	0.00107	0.00413	0	0	0	0	0	0	1	-360	360;
	6889	2085	9e-05	0.00058	0	1251	0	0	0	0	1	-360	360;
	6889	4024	9e-05	0.00059	0	1152	0	0	0	0	1	-360	360;
	6889	9065	8e-05	0.00053	0	757	0	0	0	0	1	-360	360;
	6889	661	8e-05	0.00052	0	757	0	0	0	0	1	-360	360;
	6612	8864	0.000271	0.00151	0	495	0	0	0	0	1	-360	360;
	8931	6199	0.00053	0.00585	0	1909	0	0	0	0	1	-360	360;
	8931	6199	0.00047	0.00663	0	1448	0	0	0	0	1	-360	360;
	8931	6199	0.00046	0.00676	0	1251	0	0	0	0	1	-360	360;
	8931	6199	0.00045	0.00689	0	1251	0	0	0	0	1	-360	360;
	8931	8494	0.00028	0.00338	0	1580	0	0	0	0	1	-360	360;
	8931	8494	0.0003	0.00319	0	1349	0	0	0	0	1	-360	360;
	1643	4511	0.00025	0.00101	0	476	0	0	0	0	1	-360	360;
	196	1973	0.000281	0.00119	0	0	0	0	0	0	1	-360	360;
	2902	5573	0.00018	0.00063	0	0	0	0	0	0	1	-360	360;
	4950	4410	0.00382	0.02868	0	0	0	0	0	0	1	-360	360;
	4950	333	0.00361	0.0235	0	572	0	0	0	0	1	-360	360;
	4950	6908	0.00538	0.03569	0	0	0	0	0	0	1	-360	360;
	3377	7726	6e-05	0.000409	0	514	0	0	0	0	1	-360	360;
	1027	3243	9e-05	0.00102	0	757	0	0	0	0	1	-360	360;
	1027	2458	0.0014	0.01649	0	1711	0	0	0	0	1	-360	360;
	1027	2919	0.00034	0.00427	0	1119	0	0	0	0	1	-360	360;
	5308	8560	0.0001	0.00109	0	757	0	0	0	0	1	-360	360;
	5308	6427	0.00137	0.01682	0	1777	0	0	0	0	1	-360	360;
	5308	8347	0.00131	0.01262	0	1514	0	0	0	0	1	-360	360;
	7148	5610	0.00262	0.019669	0	533	0	0	0	0	1	-360	360;
	7148	2421	0.00082	0.00526	0	476	0	0	0	0	1	-360	360;
	7148	2421	0.000841	0.005151	0	457	0	0	0	0	1	-360	360;
	7148	2421	0.000669	0.005081	0	476	0	0	0	0	1	-360	360;
	7148	6989	0.001919	0.012669	0	476	0	0	0	0	1	-360	360;
	44	1545	0.002659	0.02	0	0	0	0	0	0	1	-360	360;
	44	3541	0.003031	0.02343	0	572	0	0	0	0	1	-360	360;
	6231	4710	0.01176	0.059159	0	400	0	0	0	0	1	-360	360;
	6231	8367	0.002469	0.017531	0	495	0	0	0	0	1	-360	360;
	6231	2288	0.000669	0.005729	0	0	0	0	0	0	1	-360	360;
	8291	8473	0.000781	0.00876	0	381	0	0	0	0	1	-360	360;
	8291	5455	0.00062	0.00482	0	553	0	0	0	0	1	-360	360;
	8291	7635	0.001349	0.012599	0	533	0	0	0	0	1	-360	360;
	6842	8473	0.000979	0.0082	0	0	0	0	0	0	1	-360	360;
	6842	883	0.008091	0.028831	0	0	0	0	0	0	1	-360	360;
	6842	2288	0.000711	0.005969	0	514	0	0	0	0	1	-360	360;
	3969	7726	0.00145	0.00857	0	0	0	0	0	0	1	-360	360;
	1090	1129	0.00869	0.03768	0	400	0	0	0	0	1	-360	360;
	1090	3070	0.001719	0.00837	0	457	0	0	0	0	1	-360	360;
	8515	5365	0.01582	0.067349	0	438	0	0	0	0	1	-360	360;
	8515	5365	0.01457	0.06469	0	400	0	0	0	0	1	-360	360;
	8515	4529	0.021711	0.090539	0	381	0	0	0	0	1	-360	360;
	8515	3364	0.00019	0.000729	0	400	0	0	0	0	1	-360	360;
	935	5383	0.001271	0.00638	0	400	0	0	0	0	1	-360	360;
	920	6570	3.9e-05	0.0002	0	400	0	0	0	0	1	-360	360;
	6664	4426	0.000281	0.00236	0	629	0	0	0	0	1	-360	360;
	4339	2183	0.00019	0.00329	0	1744	0	0	0	0	1	-360	360;
	7513	4000	0.00035	0.00324	0	0	0	0	0	0	1	-360	360;
	6570	8542	0.00049	0.00349	0	419	0	0	0	0	1	-360	360;
	6570	8542	0.000469	0.00331	0	419	0	0	0	0	1	-360	360;
	6570	1662	0.00313	0.01324	0	0	0	0	0	0	1	-360	360;
	6570	9119	0.001031	0.006719	0	476	0	0	0	0	1	-360	360;
	6570	1151	0.00038	0.00401	0	0	0	0	0	0	1	-360	360;
	502	3589	0.00037	0.002849	0	495	0	0	0	0	1	-360	360;
	8473	1040	0.00032	0.00256	0	476	0	0	0	0	1	-360	360;
	8473	145	0.00037	0.00118	0	381	0	0	0	0	1	-360	360;
	8473	4267	0.00068	0.001151	0	0	0	0	0	0	1	-360	360;
	4936	5179	0.00168	0.009841	0	495	0	0	0	0	1	-360	360;
	5317	8854	0.000599	0.002401	0	400	0	0	0	0	1	-360	360;
	9018	4	0.002169	0.00911	0	400	0	0	0	0	1	-360	360;
	9018	4829	0.002849	0.01282	0	457	0	0	0	0	1	-360	360;
	6253	5383	0.001229	0.01207	0	553	0	0	0	0	1	-360	360;
	6253	7507	0.001289	0.01089	0	610	0	0	0	0	1	-360	360;
	8887	1644	0.00093	0.01036	0	856	0	0	0	0	1	-360	360;
	8887	3483	0.00083	0.01174	0	823	0	0	0	0	1	-360	360;
	6639	7840	0.00118	0.009469	0	0	0	0	0	0	1	-360	360;
	6639	7840	0.001151	0.009651	0	0	0	0	0	0	1	-360	360;
	5461	8109	0.00101	0.01327	0	1580	0	0	0	0	1	-360	360;
	6291	4189	0.00364	0.02486	0	572	0	0	0	0	1	-360	360;
	8405	4747	0.00056	0.00425	0	533	0	0	0	0	1	-360	360;
	8405	4580	0.00369	0.02681	0	533	0	0	0	0	1	-360	360;
	338	7824	0.000531	0.00276	0	400	0	0	0	0	1	-360	360;
	2468	7824	0.000521	0.00281	0	381	0	0	0	0	1	-360	360;
	5257	7824	0.000419	0.00232	0	0	0	0	0	0	1	-360	360;
	2189	7824	0.000409	0.00237	0	0	0	0	0	0	1	-360	360;
	8458	8267	0.00014	0.00193	0	1514	0	0	0	0	1	-360	360;
	2676	5987	0.003211	0.01481	0	419	0	0	0	0	1	-360	360;
	1187	5712	0.00381	0.01175	0	0	0	0	0	0	1	-360	360;
	1187	6521	0.003781	0.02676	0	419	0	0	0	0	1	-360	360;
	2360	8636	0.000919	0.005461	0	553	0	0	0	0	1	-360	360;
	2360	8829	0.000271	0.001211	0	381	0	0	0	0	1	-360	360;
	2360	4235	0.002229	0.017401	0	572	0	0	0	0	1	-360	360;
	2360	4235	0.00243	0.017659	0	610	0	0	0	0	1	-360	360;
	2360	4235	0.002531	0.01705	0	591	0	0	0	0	1	-360	360;
	8818	5099	0.012531	0.05451	0	0	0	0	0	0	1	-360	360;
	5781	4734	0.00087	0.01212	0	1349	0	0	0	0	1	-360	360;
	5781	905	0.0006	0.00802	0	1711	0	0	0	0	1	-360	360;
	5781	8334	0.00072	0.00667	0	1843	0	0	0	0	1	-360	360;
	5781	2093	0.00234	0.02626	0	1415	0	0	0	0	1	-360	360;
	6772	6989	0.00038	0.002969	0	514	0	0	0	0	1	-360	360;
	7885	3912	0.00024	0.00045	0	0	0	0	0	0	1	-360	360;
	6730	3912	0.000229	0.000409	0	0	0	0	0	0	1	-360	360;
	4301	3912	0.000281	0.000401	0	0	0	0	0	0	1	-360	360;
	5362	5610	0.001211	0.005271	0	0	0	0	0	0	1	-360	360;
	5362	350	0.000281	0.00107	0	438	0	0	0	0	1	-360	360;
	5362	6763	0.00025	0.001091	0	0	0	0	0	0	1	-360	360;
	3855	1973	0.00019	0.00081	0	629	0	0	0	0	1	-360	360;
	3855	1973	0.00045	0.00136	0	591	0	0	0	0	1	-360	360;
	3481	554	0.003461	0.023289	0	0	0	0	0	0	1	-360	360;
	5460	8222	0.0013	0.01609	0	1843	0	0	0	0	1	-360	360;
	5460	8222	0.00128	0.0164	0	1711	0	0	0	0	1	-360	360;
	5460	2721	0.00089	0.01082	0	1580	0	0	0	0	1	-360	360;
	5460	2721	0.00087	0.01102	0	1645	0	0	0	0	1	-360	360;
	3786	8578	0.003349	0.02514	0	533	0	0	0	0	1	-360	360;
	490	2653	0.0003	0.00061	0	0	0	0	0	0	1	-360	360;
	148	7069	0.000289	0.001211	0	514	0	0	0	0	1	-360	360;
	7256	8743	0.01301	0.044091	0	438	0	0	0	0	1	-360	360;
	7256	1742	0.02162	0.065969	0	400	0	0	0	0	1	-360	360;
	8930	314	0.006461	0.05275	0	514	0	0	0	0	1	-360	360;
	6954	7974	0.0002	0.000271	0	0	0	0	0	0	1	-360	360;
	3325	7974	0.00018	0.00026	0	0	0	0	0	0	1	-360	360;
	5627	1742	0.00032	0.00188	0	0	0	0	0	0	1	-360	360;
	5610	6989	0.00105	0.006591	0	476	0	0	0	0	1	-360	360;
	21	6163	0.00013	0.00055	0	0	0	0	0	0	1	-360	360;
	3749	5365	0.006151	0.04013	0	629	0	0	0	0	1	-360	360;
	3749	6982	0.007159	0.040591	0	533	0	0	0	0	1	-360	360;
	2291	3133	0.00874	0.048461	0	400	0	0	0	0	1	-360	360;
	2291	4738	0.004159	0.01606	0	0	0	0	0	0	1	-360	360;
	2291	3656	0.002919	0.01912	0	533	0	0	0	0	1	-360	360;
	314	4300	0.00094	0.00394	0	419	0	0	0	0	1	-360	360;
	314	5789	0.000901	0.00395	0	0	0	0	0	0	1	-360	360;
	314	2467	0.003159	0.017789	0	572	0	0	0	0	1	-360	360;
	314	2467	0.002419	0.018331	0	0	0	0	0	0	1	-360	360;
	314	5256	0.00375	0.018281	0	419	0	0	0	0	1	-360	360;
	3200	2653	0.00025	0.001031	0	419	0	0	0	0	1	-360	360;
	4679	4852	0.00264	0.01238	0	419	0	0	0	0	1	-360	360;
	4679	4852	0.001289	0.00795	0	457	0	0	0	0	1	-360	360;
	4679	1151	0.00087	0.003081	0	476	0	0	0	0	1	-360	360;
	8568	5215	0.000401	0.00224	0	419	0	0	0	0	1	-360	360;
	58	221	0.0003	0.00385	0	0	0	0	0	0	1	-360	360;
	6807	851	0.00126	0.008219	0	533	0	0	0	0	1	-360	360;
	150	5099	0.01905	0.055729	0	0	0	0	0	0	1	-360	360;
	150	3558	0.00201	0.00444	0	0	0	0	0	0	1	-360	360;
	7209	2934	0.00101	0.00324	0	0	0	0	0	0	1	-360	360;
	1465	4787	0.00105	0.00677	0	1218	0	0	0	0	1	-360	360;
	1465	3830	0.00039	0.00502	0	1152	0	0	0	0	1	-360	360;
	1465	5648	0.0004	0.00492	0	954	0	0	0	0	1	-360	360;
	1465	4000	0.00051	0.00617	0	1349	0	0	0	0	1	-360	360;
	7537	3485	0.0003	0.00314	0	1580	0	0	0	0	1	-360	360;
	7537	3830	0.00051	0.00668	0	1645	0	0	0	0	1	-360	360;
	7537	5648	0.0005	0.00678	0	1777	0	0	0	0	1	-360	360;
	1262	2424	0.002909	0.017	0	0	0	0	0	0	1	-360	360;
	1262	1156	0.00245	0.011281	0	476	0	0	0	0	1	-360	360;
	1262	4914	0.0027	0.01532	0	457	0	0	0	0	1	-360	360;
	1262	4914	0.00105	0.011151	0	0	0	0	0	0	1	-360	360;
	1262	3834	0.00036	0.002669	0	0	0	0	0	0	1	-360	360;
	1262	556	0.0018	0.011169	0	457	0	0	0	0	1	-360	360;
	1262	3019	0.000711	0.007771	0	0	0	0	0	0	1	-360	360;
	1262	7923	0.00036	0.00257	0	0	0	0	0	0	1	-360	360;
	1551	4103	0.00201	0.01195	0	610	0	0	0	0	1	-360	360;
	1551	7862	0.001091	0.00731	0	572	0	0	0	0	1	-360	360;
	1551	5458	0.00105	0.007591	0	629	0	0	0	0	1	-360	360;
	1551	1838	0.00299	0.02256	0	476	0	0	0	0	1	-360	360;
	3070	2888	0.00036	0.00288	0	0	0	0	0	0	1	-360	360;
	4874	4504	0.00589	0.045599	0	553	0	0	0	0	1	-360	360;
	4874	3579	0.00157	0.0127	0	476	0	0	0	0	1	-360	360;
	4874	3579	0.00168	0.011969	0	0	0	0	0	0	1	-360	360;
	4874	7903	0.00289	0.021031	0	438	0	0	0	0	1	-360	360;
	6368	1102	0.00395	0.025211	0	572	0	0	0	0	1	-360	360;
	6368	9217	0.002771	0.021659	0	476	0	0	0	0	1	-360	360;
	6368	5653	0.002979	0.016901	0	553	0	0	0	0	1	-360	360;
	8478	5907	0.007419	0.034021	0	0	0	0	0	0	1	-360	360;
	5241	6921	0.00056	0.00813	0	889	0	0	0	0	1	-360	360;
	4	1813	0.00361	0.015211	0	438	0	0	0	0	1	-360	360;
	8950	7049	0.000409	0.003331	0	629	0	0	0	0	1	-360	360;
	6846	8788	0.004091	0.019599	0	476	0	0	0	0	1	-360	360;
	6846	6734	0.00294	0.012841	0	495	0	0	0	0	1	-360	360;
	4426	5616	0.020771	0.080479	0	553	0	0	0	0	1	-360	360;
	4426	1965	0.00662	0.0513	0	553	0	0	0	0	1	-360	360;
	4426	1767	0.00607	0.04581	0	533	0	0	0	0	1	-360	360;
	3037	1102	0.000669	0.0042	0	476	0	0	0	0	1	-360	360;
	3037	8485	0.001271	0.007419	0	476	0	0	0	0	1	-360	360;
	3037	5653	0.001599	0.012271	0	572	0	0	0	0	1	-360	360;
	5395	4661	0.009599	0.05024	0	419	0	0	0	0	1	-360	360;
	5395	520	0.00149	0.009219	0	438	0	0	0	0	1	-360	360;
	5395	6816	0.01488	0.06724	0	400	0	0	0	0	1	-360	360;
	5395	5918	0.003599	0.024031	0	419	0	0	0	0	1	-360	360;
	5395	5918	0.00519	0.024	0	438	0	0	0	0	1	-360	360;
	8636	1380	0.003461	0.013349	0	400	0	0	0	0	1	-360	360;
	3436	4239	0.00105	0.01236	0	0	0	0	0	0	1	-360	360;
	3436	4239	0.00199	0.012331	0	495	0	0	0	0	1	-360	360;
	3436	3656	0.00711	0.044711	0	438	0	0	0	0	1	-360	360;
	6203	7124	0.006729	0.02543	0	0	0	0	0	0	1	-360	360;
	9059	7124	0.005771	0.02999	0	0	0	0	0	0	1	-360	360;
	5994	96	0.0048	0.021771	0	0	0	0	0	0	1	-360	360;
	5049	306	0.00286	0.02212	0	533	0	0	0	0	1	-360	360;
	718	8397	0.00351	0.02175	0	514	0	0	0	0	1	-360	360;
	718	6982	0.008401	0.0448	0	572	0	0	0	0	1	-360	360;
	7202	8807	0.0018	0.01187	0	0	0	0	0	0	1	-360	360;
	7202	8807	0.001831	0.01163	0	0	0	0	0	0	1	-360	360;
	8592	2995	0.00075	0.00538	0	0	0	0	0	0	1	-360	360;
	6581	823	2e-05	0.00033	0	0	0	0	0	0	1	-360	360;
	6581	1680	1e-05	0.00034	0	1810	0	0	0	0	1	-360	360;
	6199	953	0.00023	0.00258	0	823	0	0	0	0	1	-360	360;
	6199	1817	0.00023	0.00252	0	757	0	0	0	0	1	-360	360;
	3231	1504	0.002031	0.008531	0	0	0	0	0	0	1	-360	360;
	3231	7226	0.005349	0.020789	0	419	0	0	0	0	1	-360	360;
	2928	7865	0.0002	0.00082	0	495	0	0	0	0	1	-360	360;
	3916	6891	0.002091	0.011219	0	0	0	0	0	0	1	-360	360;
	6053	171	0.00262	0.016979	0	533	0	0	0	0	1	-360	360;
	6053	171	0.002669	0.01663	0	591	0	0	0	0	1	-360	360;
	8829	3044	0.00025	0.00118	0	514	0	0	0	0	1	-360	360;
	1552	1194	0.00056	0.00715	0	1645	0	0	0	0	1	-360	360;
	3912	3740	0.0007	0.00455	0	438	0	0	0	0	1	-360	360;
	3912	2695	0.000331	0.001711	0	381	0	0	0	0	1	-360	360;
	3912	2695	0.00011	0.000781	0	553	0	0	0	0	1	-360	360;
	3912	8877	0.00111	0.00664	0	438	0	0	0	0	1	-360	360;
	7520	8222	0.00066	0.00838	0	1547	0	0	0	0	1	-360	360;
	7520	8222	0.00064	0.00854	0	1547	0	0	0	0	1	-360	360;
	2421	3610	0.003091	0.017091	0	572	0	0	0	0	1	-360	360;
	2421	1414	0.00643	0.03687	0	629	0	0	0	0	1	-360	360;
	8879	3610	0.00275	0.01937	0	0	0	0	0	0	1	-360	360;
	8397	5918	0.00357	0.021289	0	514	0	0	0	0	1	-360	360;
	1914	6982	0.000289	0.002789	0	0	0	0	0	0	1	-360	360;
	7873	6638	0.002531	0.01345	0	381	0	0	0	0	1	-360	360;
	4562	7702	0.00162	0.012159	0	514	0	0	0	0	1	-360	360;
	4562	7702	0.00262	0.01168	0	419	0	0	0	0	1	-360	360;
	8835	1156	0.00099	0.006081	0	0	0	0	0	0	1	-360	360;
	8835	1233	0.000909	0.005979	0	0	0	0	0	0	1	-360	360;
	3390	408	0.00011	0.0008	0	0	0	0	0	0	1	-360	360;
	3390	7231	0.00011	0.00068	0	0	0	0	0	0	1	-360	360;
	3390	5993	0.00207	0.008341	0	495	0	0	0	0	1	-360	360;
	1910	2550	0.00019	0.001031	0	0	0	0	0	0	1	-360	360;
	3975	7353	0.014039	0.0667	0	0	0	0	0	0	1	-360	360;
	3126	7495	0.004961	0.03043	0	438	0	0	0	0	1	-360	360;
	2689	5753	0.00056	0.003531	0	0	0	0	0	0	1	-360	360;
	1201	4235	0.001461	0.00889	0	514	0	0	0	0	1	-360	360;
	7626	7076	6e-05	0.000419	0	495	0	0	0	0	1	-360	360;
	4823	4060	0.00139	0.005169	0	0	0	0	0	0	1	-360	360;
	4141	1311	0.00069	0.00825	0	757	0	0	0	0	1	-360	360;
	7124	4566	0.004969	0.030781	0	0	0	0	0	0	1	-360	360;
	7124	8873	0.00012	0.000969	0	591	0	0	0	0	1	-360	360;
	6684	5383	0.001091	0.010159	0	572	0	0	0	0	1	-360	360;
	8151	5383	0.000969	0.00887	0	553	0	0	0	0	1	-360	360;
	4357	6619	0.00099	0.006789	0	514	0	0	0	0	1	-360	360;
	8191	8765	0.001031	0.00586	0	381	0	0	0	0	1	-360	360;
	7466	5469	0.00401	0.03345	0	629	0	0	0	0	1	-360	360;
	3401	5351	9.1e-05	0.000349	0	0	0	0	0	0	1	-360	360;
	3760	225	0.00125	0.00745	0	591	0	0	0	0	1	-360	360;
	53	4031	0.003711	0.024219	0	514	0	0	0	0	1	-360	360;
	8222	6921	0.00169	0.02156	0	1678	0	0	0	0	1	-360	360;
	8222	6921	0.00171	0.02114	0	1777	0	0	0	0	1	-360	360;
	6556	4239	0.00088	0.00562	0	533	0	0	0	0	1	-360	360;
	6556	4239	0.0008	0.00618	0	572	0	0	0	0	1	-360	360;
	2940	7178	0.00161	0.01014	0	0	0	0	0	0	1	-360	360;
	3928	2563	0.00032	0.002031	0	533	0	0	0	0	1	-360	360;
	3928	6521	0.00011	0.000659	0	495	0	0	0	0	1	-360	360;
	1923	3830	0.00042	0.00505	0	1086	0	0	0	0	1	-360	360;
	1923	5648	0.00043	0.00494	0	1251	0	0	0	0	1	-360	360;
	2815	4914	0.00038	0.000711	0	0	0	0	0	0	1	-360	360;
	2815	4914	0.00037	0.000659	0	0	0	0	0	0	1	-360	360;
	4125	1644	0.00248	0.03096	0	1843	0	0	0	0	1	-360	360;
	163	3498	0.00375	0.019729	0	495	0	0	0	0	1	-360	360;
	163	2430	0.00518	0.020021	0	438	0	0	0	0	1	-360	360;
	7943	9108	0.000771	0.003711	0	495	0	0	0	0	1	-360	360;
	7943	1478	0.00076	0.003789	0	457	0	0	0	0	1	-360	360;
	2308	3541	0.000781	0.00574	0	0	0	0	0	0	1	-360	360;
	3535	7582	0.00089	0.00562	0	0	0	0	0	0	1	-360	360;
	2019	8542	0.000331	0.001401	0	0	0	0	0	0	1	-360	360;
	5174	8542	0.000401	0.00137	0	0	0	0	0	0	1	-360	360;
	7945	5215	0.00051	0.003349	0	0	0	0	0	0	1	-360	360;
	2057	9130	0.000461	0.00168	0	0	0	0	0	0	1	-360	360;
	2057	9130	0.000469	0.00164	0	0	0	0	0	0	1	-360	360;
	4661	6816	0.006081	0.03114	0	457	0	0	0	0	1	-360	360;
	7519	776	0.005831	0.02218	0	495	0	0	0	0	1	-360	360;
	8989	3435	0.00013	0.000461	0	0	0	0	0	0	1	-360	360;
	7797	2020	0.00013	0.000461	0	0	0	0	0	0	1	-360	360;
	871	4031	0.00262	0.020289	0	514	0	0	0	0	1	-360	360;
	871	2878	0.00164	0.01194	0	553	0	0	0	0	1	-360	360;
	7353	9164	0.006479	0.02845	0	514	0	0	0	0	1	-360	360;
	7974	2012	0.00294	0.014419	0	495	0	0	0	0	1	-360	360;
	7974	2012	0.00238	0.01463	0	495	0	0	0	0	1	-360	360;
	7974	1750	0.002219	0.00993	0	476	0	0	0	0	1	-360	360;
	1401	2303	0.008651	0.04886	0	400	0	0	0	0	1	-360	360;
	4504	7903	0.00464	0.030979	0	0	0	0	0	0	1	-360	360;
	5837	449	0.00092	0.00889	0	1547	0	0	0	0	1	-360	360;
	5837	449	0.00093	0.00869	0	1448	0	0	0	0	1	-360	360;
	5837	1817	0.0022	0.03485	0	1678	0	0	0	0	1	-360	360;
	5837	2458	0.00074	0.00941	0	1843	0	0	0	0	1	-360	360;
	5837	6427	0.00087	0.00884	0	1678	0	0	0	0	1	-360	360;
	5837	1730	0.00139	0.01662	0	0	0	0	0	0	1	-360	360;
	9217	594	0.002781	0.022979	0	400	0	0	0	0	1	-360	360;
	6351	5393	0.000521	0.00407	0	629	0	0	0	0	1	-360	360;
	6194	7702	0.008479	0.032841	0	0	0	0	0	0	1	-360	360;
	2503	7491	0.005669	0.02901	0	0	0	0	0	0	1	-360	360;
	4031	2872	0.001031	0.004099	0	0	0	0	0	0	1	-360	360;
	4031	8043	0.00301	0.018719	0	476	0	0	0	0	1	-360	360;
	1813	2866	0.00257	0.01713	0	438	0	0	0	0	1	-360	360;
	5365	5417	0.010461	0.03836	0	0	0	0	0	0	1	-360	360;
	5365	4100	0.001719	0.011531	0	514	0	0	0	0	1	-360	360;
	5365	1556	9.1e-05	0.00056	0	400	0	0	0	0	1	-360	360;
	3894	3865	0.00476	0.025289	0	0	0	0	0	0	1	-360	360;
	128	1554	0.000219	0.00164	0	419	0	0	0	0	1	-360	360;
	198	7994	0.002159	0.01163	0	0	0	0	0	0	1	-360	360;
	198	1554	0.00295	0.01539	0	400	0	0	0	0	1	-360	360;
	2985	3133	0.007789	0.033039	0	400	0	0	0	0	1	-360	360;
	7316	3071	0.00026	0.0013	0	0	0	0	0	0	1	-360	360;
	6486	8992	0.00162	0.013229	0	438	0	0	0	0	1	-360	360;
	8485	594	0.007919	0.04194	0	438	0	0	0	0	1	-360	360;
	682	8477	0.001229	0.00401	0	400	0	0	0	0	1	-360	360;
	5146	5488	0.008021	0.042831	0	438	0	0	0	0	1	-360	360;
	6638	776	0.007711	0.03006	0	438	0	0	0	0	1	-360	360;
	8788	789	0.00261	0.015469	0	419	0	0	0	0	1	-360	360;
	6332	1556	0.00139	0.00536	0	0	0	0	0	0	1	-360	360;
	6332	1556	0.001419	0.00525	0	0	0	0	0	0	1	-360	360;
	1883	7702	0.006289	0.032271	0	0	0	0	0	0	1	-360	360;
	1883	2132	0.011031	0.05625	0	0	0	0	0	0	1	-360	360;
	171	2341	0.001841	0.01324	0	572	0	0	0	0	1	-360	360;
	6630	10	0.000599	0.00186	0	553	0	0	0	0	1	-360	360;
	4484	4239	0.00036	0.001591	0	533	0	0	0	0	1	-360	360;
	3044	4235	0.003849	0.027651	0	553	0	0	0	0	1	-360	360;
	851	4454	0.00186	0.00874	0	591	0	0	0	0	1	-360	360;
	3485	2183	0.00025	0.00285	0	1580	0	0	0	0	1	-360	360;
	1156	7923	0.00207	0.008841	0	419	0	0	0	0	1	-360	360;
	1156	1233	6e-05	0.00039	0	553	0	0	0	0	1	-360	360;
	8913	4852	0.002159	0.008909	0	457	0	0	0	0	1	-360	360;
	4039	1233	0.000219	0.00039	0	0	0	0	0	0	1	-360	360;
	4914	2653	0.00011	0.000521	0	400	0	0	0	0	1	-360	360;
	4914	2653	0.00011	0.00049	0	400	0	0	0	0	1	-360	360;
	4914	556	0.000771	0.003789	0	438	0	0	0	0	1	-360	360;
	4914	3019	0.000229	0.002849	0	0	0	0	0	0	1	-360	360;
	7994	5256	0.00445	0.027599	0	0	0	0	0	0	1	-360	360;
	1709	3346	0.00218	0.02	0	1481	0	0	0	0	1	-360	360;
	1709	8468	0.00074	0.00865	0	1218	0	0	0	0	1	-360	360;
	7702	1234	0.002531	0.019419	0	400	0	0	0	0	1	-360	360;
	7702	8035	0.003	0.01631	0	0	0	0	0	0	1	-360	360;
	8677	3331	0.008461	0.025081	0	0	0	0	0	0	1	-360	360;
	8189	3643	0.00382	0.013669	0	495	0	0	0	0	1	-360	360;
	3133	8477	0.002581	0.01814	0	514	0	0	0	0	1	-360	360;
	6744	6101	0.00201	0.012219	0	533	0	0	0	0	1	-360	360;
	8542	4313	0.00087	0.001599	0	0	0	0	0	0	1	-360	360;
	769	5233	0.00455	0.02394	0	419	0	0	0	0	1	-360	360;
	1965	5926	0.005651	0.041771	0	457	0	0	0	0	1	-360	360;
	9014	3999	0.00069	0.001271	0	0	0	0	0	0	1	-360	360;
	1917	8406	0.00068	0.00269	0	400	0	0	0	0	1	-360	360;
	1917	2021	0.00036	0.00261	0	476	0	0	0	0	1	-360	360;
	1917	8293	0.001669	0.012031	0	419	0	0	0	0	1	-360	360;
	4738	3656	0.009669	0.03513	0	0	0	0	0	0	1	-360	360;
	4738	8477	0.001099	0.00862	0	0	0	0	0	0	1	-360	360;
	7998	1644	0.0005	0.00711	0	757	0	0	0	0	1	-360	360;
	6734	1015	0.003919	0.017781	0	495	0	0	0	0	1	-360	360;
	6734	8992	0.00161	0.019781	0	0	0	0	0	0	1	-360	360;
	6734	789	0.0007	0.01256	0	610	0	0	0	0	1	-360	360;
	5004	5393	0.000409	0.0018	0	0	0	0	0	0	1	-360	360;
	1708	3906	2e-05	0.00043	0	0	0	0	0	0	1	-360	360;
	9150	3906	3e-05	0.0004	0	1810	0	0	0	0	1	-360	360;
	280	9158	0.00968	0.056659	0	457	0	0	0	0	1	-360	360;
	8043	3656	0.003969	0.02925	0	0	0	0	0	0	1	-360	360;
	5712	8992	0.00219	0.01468	0	514	0	0	0	0	1	-360	360;
	5712	8466	0.000401	0.003159	0	476	0	0	0	0	1	-360	360;
	5712	6880	0.000531	0.003	0	476	0	0	0	0	1	-360	360;
	5215	7266	0.00136	0.00576	0	476	0	0	0	0	1	-360	360;
	8807	8992	0.000849	0.006289	0	457	0	0	0	0	1	-360	360;
	8807	8992	0.000841	0.006409	0	457	0	0	0	0	1	-360	360;
	8807	8992	0.00045	0.00506	0	0	0	0	0	0	1	-360	360;
	218	2088	0.003669	0.02988	0	495	0	0	0	0	1	-360	360;
	520	5918	0.002	0.01568	0	514	0	0	0	0	1	-360	360;
	2653	2128	0.00031	0.0008	0	419	0	0	0	0	1	-360	360;
	1216	5233	0.00043	0.00038	0	0	0	0	0	0	1	-360	360;
	8980	5993	0.00043	0.002229	0	0	0	0	0	0	1	-360	360;
	5212	5993	0.00049	0.00218	0	0	0	0	0	0	1	-360	360;
	3656	4060	0.00282	0.01844	0	629	0	0	0	0	1	-360	360;
	883	3331	0.00462	0.014849	0	0	0	0	0	0	1	-360	360;
	5144	3346	0.00263	0.04076	0	1547	0	0	0	0	1	-360	360;
	3498	2563	0.01062	0.056719	0	457	0	0	0	0	1	-360	360;
	1896	8200	0.0001	0.00105	0	757	0	0	0	0	1	-360	360;
	9130	8864	0.00239	0.017341	0	476	0	0	0	0	1	-360	360;
	3503	2510	0.001341	0.010581	0	533	0	0	0	0	1	-360	360;
	3503	6146	0.00101	0.007599	0	572	0	0	0	0	1	-360	360;
	401	1584	0.00011	0.000289	0	495	0	0	0	0	1	-360	360;
	7328	6921	0.00146	0.01769	0	0	0	0	0	0	1	-360	360;
	7328	6921	0.0013	0.02004	0	1744	0	0	0	0	1	-360	360;
	6816	5573	0.007909	0.028349	0	419	0	0	0	0	1	-360	360;
	7396	8564	0.00012	0.00049	0	0	0	0	0	0	1	-360	360;
	7396	8564	0.00012	0.000479	0	0	0	0	0	0	1	-360	360;
	6921	432	0.00076	0.01018	0	1613	0	0	0	0	1	-360	360;
	6921	432	0.0009	0.00865	0	1777	0	0	0	0	1	-360	360;
	6921	432	0.00096	0.01294	0	1645	0	0	0	0	1	-360	360;
	6921	432	0.0004	0.00756	0	1711	0	0	0	0	1	-360	360;
	6921	432	0.00039	0.00771	0	1547	0	0	0	0	1	-360	360;
	6921	8200	0.00163	0.02291	0	1645	0	0	0	0	1	-360	360;
	2949	3121	0.00636	0.03855	0	495	0	0	0	0	1	-360	360;
	2949	3121	0.00639	0.03863	0	514	0	0	0	0	1	-360	360;
	953	1730	0.00113	0.01362	0	790	0	0	0	0	1	-360	360;
	1767	892	0.00861	0.05318	0	629	0	0	0	0	1	-360	360;
	4674	4889	0.0012	0.004669	0	438	0	0	0	0	1	-360	360;
	4674	7076	0.00018	0.001091	0	514	0	0	0	0	1	-360	360;
	1234	8035	0.001401	0.00612	0	457	0	0	0	0	1	-360	360;
	115	225	0.00068	0.00387	0	591	0	0	0	0	1	-360	360;
	5720	8578	0.0043	0.033591	0	514	0	0	0	0	1	-360	360;
	3579	5469	0.003039	0.01911	0	591	0	0	0	0	1	-360	360;
	3246	124	6e-05	0.00111	0	953	0	0	0	0	1	-360	360;
	8846	4783	5e-05	0.001091	0	953	0	0	0	0	1	-360	360;
	2393	3306	3.9e-05	0.00089	0	953	0	0	0	0	1	-360	360;
	1895	4480	3.9e-05	0.000901	0	953	0	0	0	0	1	-360	360;
	678	4454	0.000169	0.00111	0	610	0	0	0	0	1	-360	360;
	1002	4852	0.000331	0.0027	0	0	0	0	0	0	1	-360	360;
	6429	2012	0.00026	0.00369	0	0	0	0	0	0	1	-360	360;
	4889	7076	0.000341	0.00274	0	0	0	0	0	0	1	-360	360;
	5341	7491	0.00038	0.00506	0	514	0	0	0	0	1	-360	360;
	2563	6521	0.00012	0.000831	0	476	0	0	0	0	1	-360	360;
	1111	594	0.003031	0.0123	0	0	0	0	0	0	1	-360	360;
	4747	7791	0.001521	0.00774	0	495	0	0	0	0	1	-360	360;
	3906	3022	0.00059	0.00552	0	1711	0	0	0	0	1	-360	360;
	3906	1194	0.00053	0.00625	0	1645	0	0	0	0	1	-360	360;
	5993	8804	0.00143	0.01076	0	629	0	0	0	0	1	-360	360;
	1798	8487	0.00056	0.00593	0	0	0	0	0	0	1	-360	360;
	2467	7178	0.00205	0.01461	0	514	0	0	0	0	1	-360	360;
	2132	2303	0.01068	0.054521	0	0	0	0	0	0	1	-360	360;
	8992	8466	0.00194	0.011039	0	533	0	0	0	0	1	-360	360;
	8992	6880	0.00131	0.010099	0	533	0	0	0	0	1	-360	360;
	8992	2995	0.001289	0.014229	0	419	0	0	0	0	1	-360	360;
	805	3221	0.00455	0.027031	0	495	0	0	0	0	1	-360	360;
	805	3327	0.00239	0.01414	0	610	0	0	0	0	1	-360	360;
	805	8310	0.004039	0.029669	0	591	0	0	0	0	1	-360	360;
	805	2088	0.001099	0.007771	0	629	0	0	0	0	1	-360	360;
	5383	2021	0.00082	0.00762	0	514	0	0	0	0	1	-360	360;
	5383	4114	0.000719	0.0033	0	0	0	0	0	0	1	-360	360;
	5383	5519	0.000599	0.00361	0	400	0	0	0	0	1	-360	360;
	5383	8293	0.000521	0.004039	0	476	0	0	0	0	1	-360	360;
	5383	3435	0.000901	0.0088	0	533	0	0	0	0	1	-360	360;
	7507	6990	0.000781	0.00306	0	0	0	0	0	0	1	-360	360;
	7507	2406	0.000419	0.00399	0	0	0	0	0	0	1	-360	360;
	7507	2020	0.000841	0.009841	0	381	0	0	0	0	1	-360	360;
	6478	6146	0.004021	0.03064	0	514	0	0	0	0	1	-360	360;
	4852	2695	0.0023	0.009031	0	438	0	0	0	0	1	-360	360;
	4852	2695	0.002081	0.009469	0	476	0	0	0	0	1	-360	360;
	4852	5502	0.00164	0.008831	0	0	0	0	0	0	1	-360	360;
	4852	8877	0.00094	0.00488	0	419	0	0	0	0	1	-360	360;
	4852	9119	0.000901	0.00551	0	553	0	0	0	0	1	-360	360;
	4852	1151	0.00051	0.005961	0	0	0	0	0	0	1	-360	360;
	2012	1750	0.00113	0.004349	0	533	0	0	0	0	1	-360	360;
	1478	6616	0.002081	0.013031	0	495	0	0	0	0	1	-360	360;
	809	4594	0.00116	0.01438	0	0	0	0	0	0	1	-360	360;
	8656	4729	0.00013	0.001031	0	0	0	0	0	0	1	-360	360;
	4100	615	0.00036	0.002659	0	514	0	0	0	0	1	-360	360;
	4060	5393	0.000771	0.004539	0	629	0	0	0	0	1	-360	360;
	3296	4080	0.011902	0.04088	0	0	0	0	0	0	1	-360	360;
	176	3296	0.002231	0.010138	0	0	0	0	0	0	1	-360	360;
	1142	165	0.00023	0.0025	0	1448	0	0	0	0	1	-360	360;
	1142	165	0.00023	0.00254	0	1283	0	0	0	0	1	-360	360;
	4985	1142	0.00179	0.01478	0	1415	0	0	0	0	1	-360	360;
	4985	1142	0.00183	0.01446	0	1316	0	0	0	0	1	-360	360;
	1142	1533	0.00031	0.00312	0	1185	0	0	0	0	1	-360	360;
	1059	1142	0.00031	0.00289	0	1415	0	0	0	0	1	-360	360;
	914	1142	0.00021	0.00194	0	1218	0	0	0	0	1	-360	360;
	6577	7621	0.001489	0.006249	0	0	0	0	0	0	1	-360	360;
	7740	7621	0.010209	0.042502	0	0	0	0	0	0	1	-360	360;
	6577	2382	0.001462	0.006369	0	0	0	0	0	0	1	-360	360;
	2382	8839	0.0048	0.01772	0	0	0	0	0	0	1	-360	360;
	6577	5758	0.00328	0.01472	0	0	0	0	0	0	1	-360	360;
	461	6577	0	0.000271	0	0	0	0	0	0	1	-360	360;
	8839	6577	0.002031	0.007591	0	0	0	0	0	0	1	-360	360;
	4137	6905	0.00372	0.014431	0	0	0	0	0	0	1	-360	360;
	3849	4137	0	0.000271	0	0	0	0	0	0	1	-360	360;
	4137	4568	0.00108	0.003698	0	0	0	0	0	0	1	-360	360;
	4137	9008	0.000431	0.001658	0	0	0	0	0	0	1	-360	360;
	6905	3247	0.000782	0.002662	0	0	0	0	0	0	1	-360	360;
	5758	627	0.001729	0.005498	0	0	0	0	0	0	1	-360	360;
	7959	1930	0.00221	0.01891	0	1119	0	0	0	0	1	-360	360;
	8171	1930	0.00335	0.03107	0	1086	0	0	0	0	1	-360	360;
	1930	436	0.00109	0.01324	0	1777	0	0	0	0	1	-360	360;
	3448	2876	0.027218	0.118858	0	0	0	0	0	0	1	-360	360;
	6355	3448	0.002782	0.010778	0	0	0	0	0	0	1	-360	360;
	3448	2576	0.03336	0.110049	0	0	0	0	0	0	1	-360	360;
	3448	2576	0.02972	0.124729	0	0	0	0	0	0	1	-360	360;
	3448	3653	0.00244	0.009351	0	0	0	0	0	0	1	-360	360;
	3448	3653	0.002151	0.009649	0	0	0	0	0	0	1	-360	360;
	3448	3653	0.00244	0.009458	0	0	0	0	0	0	1	-360	360;
	6445	5808	0.017102	0.080911	0	0	0	0	0	0	1	-360	360;
	2451	6445	0.030871	0.129529	0	0	0	0	0	0	1	-360	360;
	6445	1877	0.008778	0.041551	0	0	0	0	0	0	1	-360	360;
	6445	1877	0.008591	0.038471	0	0	0	0	0	0	1	-360	360;
	6445	1877	0.01032	0.038049	0	0	0	0	0	0	1	-360	360;
	6445	696	0.004151	0.017409	0	0	0	0	0	0	1	-360	360;
	6445	696	0.004062	0.017751	0	0	0	0	0	0	1	-360	360;
	2838	6117	0.000338	0.00156	0	0	0	0	0	0	1	-360	360;
	8550	6117	0.002689	0.010018	0	0	0	0	0	0	1	-360	360;
	6117	4328	0.020471	0.070289	0	0	0	0	0	0	1	-360	360;
	3995	2943	4e-05	0.000271	0	0	0	0	0	0	1	-360	360;
	8536	3995	0.023289	0.097751	0	0	0	0	0	0	1	-360	360;
	5373	2943	0.005969	0.02292	0	0	0	0	0	0	1	-360	360;
	156	2943	0.00428	0.019138	0	0	0	0	0	0	1	-360	360;
	1931	8550	0.01588	0.054529	0	0	0	0	0	0	1	-360	360;
	6328	4258	0	0.000271	0	0	0	0	0	0	1	-360	360;
	4258	3100	0.00368	0.007311	0	0	0	0	0	0	1	-360	360;
	7078	4258	0.006369	0.0324	0	0	0	0	0	0	1	-360	360;
	1014	4258	0.000791	0.003569	0	0	0	0	0	0	1	-360	360;
	4174	4258	0.000809	0.003431	0	0	0	0	0	0	1	-360	360;
	3100	5915	0.008871	0.004431	0	0	0	0	0	0	1	-360	360;
	9149	7074	0	0.000271	0	0	0	0	0	0	1	-360	360;
	6699	3461	0	0.000271	0	0	0	0	0	0	1	-360	360;
	7773	3763	0	0.000271	0	0	0	0	0	0	1	-360	360;
	6606	156	0	0.000271	0	0	0	0	0	0	1	-360	360;
	626	4475	0	0.000271	0	0	0	0	0	0	1	-360	360;
	973	7904	0	0.000271	0	0	0	0	0	0	1	-360	360;
	8744	4476	0.00096	0.003391	0	0	0	0	0	0	1	-360	360;
	1945	4476	0	0.000271	0	0	0	0	0	0	1	-360	360;
	4476	6404	0.004471	0.021831	0	0	0	0	0	0	1	-360	360;
	4476	3482	0.00624	0.03	0	0	0	0	0	0	1	-360	360;
	3320	2434	0.0008	0.0038	0	0	0	0	0	0	1	-360	360;
	176	3320	0.010058	0.035969	0	0	0	0	0	0	1	-360	360;
	8908	2434	0.000782	0.00352	0	0	0	0	0	0	1	-360	360;
	5939	8908	0.008471	0.040089	0	0	0	0	0	0	1	-360	360;
	7759	5883	0.00038	0.00312	0	1448	0	0	0	0	1	-360	360;
	7759	5883	0.00038	0.00305	0	1185	0	0	0	0	1	-360	360;
	603	7759	0.00255	0.02375	0	1349	0	0	0	0	1	-360	360;
	603	7759	0.0026	0.02326	0	1514	0	0	0	0	1	-360	360;
	7759	6072	0.00268	0.03039	0	1349	0	0	0	0	1	-360	360;
	7759	6072	0.00262	0.02814	0	1185	0	0	0	0	1	-360	360;
	2725	8184	0.000258	0.000929	0	0	0	0	0	0	1	-360	360;
	3165	8184	0.003049	0.013911	0	0	0	0	0	0	1	-360	360;
	2725	5769	0.000258	0.000911	0	0	0	0	0	0	1	-360	360;
	7868	5769	0.000871	0.003991	0	0	0	0	0	0	1	-360	360;
	5535	3307	0.024218	0.1016	0	0	0	0	0	0	1	-360	360;
	3307	6875	0	0.000271	0	0	0	0	0	0	1	-360	360;
	2125	4527	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5535	2125	0.02372	0.103591	0	0	0	0	0	0	1	-360	360;
	2125	7034	0	0.000271	0	0	0	0	0	0	1	-360	360;
	7605	4527	0	0.00028	0	0	0	0	0	0	1	-360	360;
	165	2522	0.00038	0.00396	0	889	0	0	0	0	1	-360	360;
	165	2522	0.00038	0.00404	0	921	0	0	0	0	1	-360	360;
	165	2522	0.00042	0.0036	0	987	0	0	0	0	1	-360	360;
	165	2522	0.00043	0.00352	0	954	0	0	0	0	1	-360	360;
	4985	165	0.00177	0.01845	0	1382	0	0	0	0	1	-360	360;
	7904	8904	0.007831	0.035089	0	0	0	0	0	0	1	-360	360;
	7904	7339	0.002089	0.007182	0	260	0	0	0	0	1	-360	360;
	5140	8904	0.008871	0.034391	0	0	0	0	0	0	1	-360	360;
	5140	4853	0.002129	0.007031	0	0	0	0	0	0	1	-360	360;
	7501	5140	0.007222	0.032831	0	0	0	0	0	0	1	-360	360;
	5140	7693	0.006049	0.021431	0	0	0	0	0	0	1	-360	360;
	2837	8744	0.0114	0.005622	0	0	0	0	0	0	1	-360	360;
	2860	2536	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2536	8901	0.00976	0.043991	0	0	0	0	0	0	1	-360	360;
	3658	1379	0.017658	0.0684	0	0	0	0	0	0	1	-360	360;
	3658	1379	0.018	0.067	0	0	0	0	0	0	1	-360	360;
	1379	7756	0.003378	0.009631	0	0	0	0	0	0	1	-360	360;
	1379	5267	0.016409	0.07168	0	0	0	0	0	0	1	-360	360;
	1379	1370	0.0182	0.04612	0	0	0	0	0	0	1	-360	360;
	1379	3108	0.004818	0.017231	0	0	0	0	0	0	1	-360	360;
	7482	5496	0.000209	0.000911	0	0	0	0	0	0	1	-360	360;
	7482	2674	0.000249	0.000831	0	0	0	0	0	0	1	-360	360;
	8637	4978	0.00136	0.006458	0	0	0	0	0	0	1	-360	360;
	4978	4781	0.002782	0.013129	0	0	0	0	0	0	1	-360	360;
	6260	1818	0.008049	0.036031	0	0	0	0	0	0	1	-360	360;
	1818	2664	0.022311	0.0864	0	0	0	0	0	0	1	-360	360;
	671	1519	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6260	1519	0.00912	0.035311	0	0	0	0	0	0	1	-360	360;
	1519	1210	0.001471	0.005489	0	0	0	0	0	0	1	-360	360;
	7228	8640	0.00058	0.00639	0	1053	0	0	0	0	1	-360	360;
	7228	8640	0.00057	0.00651	0	987	0	0	0	0	1	-360	360;
	5133	3993	0.002111	0.007542	0	0	0	0	0	0	1	-360	360;
	1204	3993	0.014049	0.037031	0	0	0	0	0	0	1	-360	360;
	259	2725	0	0.000222	0	0	0	0	0	0	1	-360	360;
	259	6779	0.002071	0.005609	0	0	0	0	0	0	1	-360	360;
	1966	259	0.00368	0.016711	0	0	0	0	0	0	1	-360	360;
	1966	259	0.0036	0.017018	0	0	0	0	0	0	1	-360	360;
	259	8111	0.00012	0.000431	0	0	0	0	0	0	1	-360	360;
	2725	7389	0.00012	0.000431	0	0	0	0	0	0	1	-360	360;
	7074	5915	0.01328	0.031071	0	0	0	0	0	0	1	-360	360;
	8436	7074	0.031911	0.071711	0	0	0	0	0	0	1	-360	360;
	6215	3461	0.000382	0.001289	0	0	0	0	0	0	1	-360	360;
	1966	6215	0.002582	0.00988	0	0	0	0	0	0	1	-360	360;
	2596	3763	0.00036	0.001218	0	0	0	0	0	0	1	-360	360;
	4590	2596	0.002458	0.00908	0	0	0	0	0	0	1	-360	360;
	6928	6102	0.000911	0.003271	0	0	0	0	0	0	1	-360	360;
	176	6928	0.01436	0.05344	0	0	0	0	0	0	1	-360	360;
	159	6635	0.000462	0.001631	0	0	0	0	0	0	1	-360	360;
	159	6102	0.000929	0.0032	0	0	0	0	0	0	1	-360	360;
	9079	6635	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5687	6635	0.013391	0.060849	0	0	0	0	0	0	1	-360	360;
	156	5013	0.010631	0.048342	0	0	0	0	0	0	1	-360	360;
	5808	3312	0.005498	0.021302	0	0	0	0	0	0	1	-360	360;
	3312	1490	0.001449	0.006329	0	0	0	0	0	0	1	-360	360;
	840	8912	0.007951	0.029329	0	0	0	0	0	0	1	-360	360;
	840	1877	0.02072	0.086938	0	0	0	0	0	0	1	-360	360;
	232	840	0.007182	0.02672	0	0	0	0	0	0	1	-360	360;
	7247	2367	0.00424	0.019129	0	0	0	0	0	0	1	-360	360;
	2876	7247	0.010542	0.047938	0	0	0	0	0	0	1	-360	360;
	7247	3653	0.016751	0.06488	0	0	0	0	0	0	1	-360	360;
	4475	3581	0.016671	0.059071	0	0	0	0	0	0	1	-360	360;
	4475	6223	0.00536	0.024151	0	0	0	0	0	0	1	-360	360;
	5636	2876	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2876	5304	0.007089	0.017938	0	260	0	0	0	0	1	-360	360;
	5606	2876	0.020849	0.052831	0	0	0	0	0	0	1	-360	360;
	5755	5304	0.010609	0.005209	0	0	0	0	0	0	1	-360	360;
	2454	5143	0.020898	0.098071	0	0	0	0	0	0	1	-360	360;
	6511	5143	0.002258	0.008409	0	0	0	0	0	0	1	-360	360;
	421	6318	0.00056	0.00236	0	0	0	0	0	0	1	-360	360;
	6318	1014	0.014258	0.064311	0	0	0	0	0	0	1	-360	360;
	8750	1987	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6645	1987	0.040249	0.143902	0	0	0	0	0	0	1	-360	360;
	8737	1987	0.021778	0.071862	0	0	0	0	0	0	1	-360	360;
	1987	3831	0.022742	0.09932	0	0	0	0	0	0	1	-360	360;
	5265	1987	0.021511	0.07092	0	0	0	0	0	0	1	-360	360;
	5405	4254	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4254	8863	0.010818	0.0454	0	0	0	0	0	0	1	-360	360;
	2293	4254	0.013671	0.059698	0	0	0	0	0	0	1	-360	360;
	6005	7807	0.00011	0.00059	0	987	0	0	0	0	1	-360	360;
	1786	7807	0.00093	0.00993	0	1415	0	0	0	0	1	-360	360;
	2059	6005	0.00037	0.00305	0	1481	0	0	0	0	1	-360	360;
	2814	1006	0.0001	0.00066	0	1152	0	0	0	0	1	-360	360;
	1786	2814	0.00105	0.00973	0	1382	0	0	0	0	1	-360	360;
	1006	1650	0.00037	0.00298	0	1613	0	0	0	0	1	-360	360;
	2059	4985	0.00105	0.01196	0	1481	0	0	0	0	1	-360	360;
	1650	165	0.00073	0.00626	0	1514	0	0	0	0	1	-360	360;
	3879	7698	0.00112	0.002329	0	0	0	0	0	0	1	-360	360;
	3879	9156	0.0068	0.032178	0	0	0	0	0	0	1	-360	360;
	5560	3879	0.013871	0.06308	0	0	0	0	0	0	1	-360	360;
	6534	4210	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3889	4210	0.00636	0.028689	0	0	0	0	0	0	1	-360	360;
	5216	4210	0.005231	0.020942	0	0	0	0	0	0	1	-360	360;
	4456	4210	0.019129	0.063111	0	0	0	0	0	0	1	-360	360;
	4210	3101	0.006649	0.0298	0	0	0	0	0	0	1	-360	360;
	7698	5703	0.0006	0.001609	0	260	0	0	0	0	1	-360	360;
	110	421	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8243	421	0.014849	0.066942	0	0	0	0	0	0	1	-360	360;
	8603	6079	0.00392	0.009182	0	0	0	0	0	0	1	-360	360;
	6079	3703	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6079	300	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6079	5936	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2541	6157	0.00384	0.00936	0	0	0	0	0	0	1	-360	360;
	6157	7080	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6157	6509	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6157	1580	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2541	8603	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3703	7391	0.02008	0.09128	0	0	0	0	0	0	1	-360	360;
	4067	7080	0.023991	0.092942	0	0	0	0	0	0	1	-360	360;
	5739	300	0.01384	0.06552	0	0	0	0	0	0	1	-360	360;
	365	6509	0.00476	0.022538	0	0	0	0	0	0	1	-360	360;
	5936	2528	0.00496	0.016382	0	0	0	0	0	0	1	-360	360;
	1580	7735	0.00244	0.008049	0	0	0	0	0	0	1	-360	360;
	8958	563	0.000671	0.002502	0	0	0	0	0	0	1	-360	360;
	563	5030	0.0066	0.029569	0	0	0	0	0	0	1	-360	360;
	8958	7368	0.000689	0.002449	0	0	0	0	0	0	1	-360	360;
	95	7368	0.007271	0.024969	0	0	0	0	0	0	1	-360	360;
	874	8958	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3889	3223	0.007649	0.031849	0	0	0	0	0	0	1	-360	360;
	6676	5928	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6676	7091	0	0.000222	0	0	0	0	0	0	1	-360	360;
	277	6676	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3455	5928	0.019471	0.074818	0	0	0	0	0	0	1	-360	360;
	6185	7091	0.009231	0.041622	0	0	0	0	0	0	1	-360	360;
	2742	8519	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2220	6355	0.000258	0.001089	0	0	0	0	0	0	1	-360	360;
	2220	3603	0.019871	0.042889	0	0	0	0	0	0	1	-360	360;
	3603	8519	0.00076	0.001978	0	260	0	0	0	0	1	-360	360;
	6056	8519	0.000751	0.001831	0	0	0	0	0	0	1	-360	360;
	7907	6056	0.039151	0.088	0	0	0	0	0	0	1	-360	360;
	396	5544	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5544	2859	0.032351	0.141289	0	0	0	0	0	0	1	-360	360;
	961	5544	0.022569	0.074449	0	0	0	0	0	0	1	-360	360;
	5544	5359	5.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	6686	9134	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1514	9134	0.009542	0.031471	0	0	0	0	0	0	1	-360	360;
	7001	9134	0.021671	0.09856	0	0	0	0	0	0	1	-360	360;
	3081	2334	0.005609	0.013978	0	273	0	0	0	0	1	-360	360;
	8385	2334	0.003751	0.01016	0	0	0	0	0	0	1	-360	360;
	8077	3658	0.000529	0.00208	0	0	0	0	0	0	1	-360	360;
	8077	4073	0.008862	0.029231	0	0	0	0	0	0	1	-360	360;
	952	6121	0.000538	0.00204	0	0	0	0	0	0	1	-360	360;
	952	8220	0.009569	0.026289	0	260	0	0	0	0	1	-360	360;
	3658	3418	0.00028	0.001209	0	0	0	0	0	0	1	-360	360;
	1204	3418	0.005489	0.014249	0	0	0	0	0	0	1	-360	360;
	6121	3081	0.000271	0.001231	0	0	0	0	0	0	1	-360	360;
	3658	5770	0.000338	0.001311	0	0	0	0	0	0	1	-360	360;
	5770	5761	0.005791	0.02716	0	0	0	0	0	0	1	-360	360;
	4346	7228	0.00288	0.03132	0	1251	0	0	0	0	1	-360	360;
	4346	7228	0.00282	0.03191	0	1218	0	0	0	0	1	-360	360;
	7228	5883	0.00111	0.01111	0	1316	0	0	0	0	1	-360	360;
	7228	6741	0.00108	0.01133	0	1316	0	0	0	0	1	-360	360;
	6121	6360	0.008751	0.033911	0	0	0	0	0	0	1	-360	360;
	6121	293	0.002418	0.00936	0	0	0	0	0	0	1	-360	360;
	6859	4743	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6859	272	0.00944	0.044271	0	0	0	0	0	0	1	-360	360;
	4743	8023	0.006542	0.0272	0	0	0	0	0	0	1	-360	360;
	3581	5446	0.001591	0.005658	0	0	0	0	0	0	1	-360	360;
	2576	1108	0.000711	0.001822	0	299	0	0	0	0	1	-360	360;
	1108	4589	0.010369	0.021511	0	299	0	0	0	0	1	-360	360;
	2576	7907	0.001978	0.0068	0	0	0	0	0	0	1	-360	360;
	537	6012	0.009062	0.023902	0	0	0	0	0	0	1	-360	360;
	6012	670	0.0092	0.025249	0	0	0	0	0	0	1	-360	360;
	8444	906	0.0022	0.005942	0	0	0	0	0	0	1	-360	360;
	4933	8444	0.016169	0.03352	0	0	0	0	0	0	1	-360	360;
	2379	670	0.006902	0.014898	0	0	0	0	0	0	1	-360	360;
	4730	2379	0.018609	0.041849	0	0	0	0	0	0	1	-360	360;
	3555	4913	0.00156	0.00696	0	0	0	0	0	0	1	-360	360;
	5307	3555	0.01592	0.07532	0	0	0	0	0	0	1	-360	360;
	2470	4913	0.00176	0.006822	0	0	0	0	0	0	1	-360	360;
	2470	8590	0.015542	0.069582	0	0	0	0	0	0	1	-360	360;
	7322	1631	0.00176	0.004631	0	0	0	0	0	0	1	-360	360;
	2194	1631	0.02272	0.061458	0	273	0	0	0	0	1	-360	360;
	6260	2820	0.00348	0.015809	0	0	0	0	0	0	1	-360	360;
	8752	2820	0.013449	0.056431	0	0	0	0	0	0	1	-360	360;
	4887	1276	0.002911	0.01272	0	0	0	0	0	0	1	-360	360;
	6260	1276	0.0034	0.016102	0	0	0	0	0	0	1	-360	360;
	6583	4887	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4887	8970	0.030151	0.079538	0	0	0	0	0	0	1	-360	360;
	5307	4887	0.017769	0.063511	0	0	0	0	0	0	1	-360	360;
	4887	4093	0.031249	0.084538	0	0	0	0	0	0	1	-360	360;
	8678	6260	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6260	4531	0.013698	0.047049	0	0	0	0	0	0	1	-360	360;
	6260	4531	0.01396	0.046031	0	0	0	0	0	0	1	-360	360;
	8162	6260	0.007938	0.037569	0	0	0	0	0	0	1	-360	360;
	8752	6260	0.016751	0.070271	0	0	0	0	0	0	1	-360	360;
	7322	6165	0.00172	0.00472	0	0	0	0	0	0	1	-360	360;
	6165	3406	0.007271	0.015689	0	0	0	0	0	0	1	-360	360;
	522	2587	0.001471	0.004	0	0	0	0	0	0	1	-360	360;
	522	2945	0.001991	0.007729	0	0	0	0	0	0	1	-360	360;
	8340	522	0.003151	0.047689	0	0	0	0	0	0	1	-360	360;
	522	357	0.000809	0.002769	0	0	0	0	0	0	1	-360	360;
	8386	522	0.002009	0.020209	0	0	0	0	0	0	1	-360	360;
	4203	6377	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6377	8739	0.001	0.002831	0	0	0	0	0	0	1	-360	360;
	6377	8064	0.001502	0.005169	0	0	0	0	0	0	1	-360	360;
	5953	6377	0.001969	0.0206	0	299	0	0	0	0	1	-360	360;
	7636	5216	0.000151	0.000649	0	0	0	0	0	0	1	-360	360;
	7636	3586	0.000142	0.000662	0	0	0	0	0	0	1	-360	360;
	2378	3586	0.004409	0.01628	0	0	0	0	0	0	1	-360	360;
	6690	6713	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9141	7543	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6713	8833	0.00028	0.000791	0	0	0	0	0	0	1	-360	360;
	5642	7543	0.000622	0.001671	0	0	0	0	0	0	1	-360	360;
	4365	729	0.000471	0.0016	0	0	0	0	0	0	1	-360	360;
	4316	729	0.013889	0.065751	0	0	0	0	0	0	1	-360	360;
	4365	3214	0.000471	0.00156	0	0	0	0	0	0	1	-360	360;
	4531	3214	0.018262	0.082978	0	0	0	0	0	0	1	-360	360;
	1210	2664	0.02168	0.077489	0	0	0	0	0	0	1	-360	360;
	7963	5535	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5535	7952	0.000538	0.001902	0	0	0	0	0	0	1	-360	360;
	3165	5535	0.004578	0.017031	0	0	0	0	0	0	1	-360	360;
	3165	5535	0.004662	0.016671	0	0	0	0	0	0	1	-360	360;
	4374	4531	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4531	3172	0.0242	0.093769	0	0	0	0	0	0	1	-360	360;
	8863	2047	0.002289	0.008129	0	0	0	0	0	0	1	-360	360;
	4828	998	0	0.000222	0	0	0	0	0	0	1	-360	360;
	998	3666	0.00668	0.02588	0	0	0	0	0	0	1	-360	360;
	1166	3666	0.00308	0.013778	0	0	0	0	0	0	1	-360	360;
	3455	998	0.032169	0.124618	0	0	0	0	0	0	1	-360	360;
	1333	8340	0.006778	0.02236	0	0	0	0	0	0	1	-360	360;
	8131	1333	0.00308	0.011929	0	0	0	0	0	0	1	-360	360;
	5859	7100	0.016911	0.070391	0	0	0	0	0	0	1	-360	360;
	2576	5859	0.03196	0.149969	0	0	0	0	0	0	1	-360	360;
	5801	7959	0.00016	0.00049	0	757	0	0	0	0	1	-360	360;
	3056	5801	7e-05	0.00049	0	1086	0	0	0	0	1	-360	360;
	8130	8596	0.0022	0.00852	0	0	0	0	0	0	1	-360	360;
	6396	8596	0.000431	0.000951	0	0	0	0	0	0	1	-360	360;
	8130	5832	0.002249	0.008351	0	0	0	0	0	0	1	-360	360;
	7734	5832	0.019431	0.053351	0	260	0	0	0	0	1	-360	360;
	4975	8130	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5802	5268	0.001929	0.008111	0	0	0	0	0	0	1	-360	360;
	8386	5802	0.00652	0.021511	0	0	0	0	0	0	1	-360	360;
	6425	5268	0.001889	0.008271	0	0	0	0	0	0	1	-360	360;
	6425	8340	0.012289	0.051551	0	0	0	0	0	0	1	-360	360;
	8368	5268	0.001431	0.005071	0	0	0	0	0	0	1	-360	360;
	1167	5268	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8752	8536	0.007271	0.0344	0	0	0	0	0	0	1	-360	360;
	7237	8536	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1986	8863	0.003289	0.011329	0	0	0	0	0	0	1	-360	360;
	8863	1370	0.012129	0.032831	0	0	0	0	0	0	1	-360	360;
	8863	176	0.006031	0.025289	0	0	0	0	0	0	1	-360	360;
	8863	176	0.005911	0.025778	0	0	0	0	0	0	1	-360	360;
	466	8863	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2184	8243	0.000529	0.001969	0	0	0	0	0	0	1	-360	360;
	8435	8243	0.000538	0.001929	0	0	0	0	0	0	1	-360	360;
	1718	7190	0.001151	0.004222	0	0	0	0	0	0	1	-360	360;
	4822	1718	0.003818	0.017222	0	0	0	0	0	0	1	-360	360;
	2528	7818	0.00416	0.019502	0	0	0	0	0	0	1	-360	360;
	5358	2661	0.001991	0.00372	0	0	0	0	0	0	1	-360	360;
	5358	3181	0.009889	0.026751	0	0	0	0	0	0	1	-360	360;
	3829	2661	0.002018	0.00364	0	260	0	0	0	0	1	-360	360;
	7229	3829	0.01644	0.036942	0	0	0	0	0	0	1	-360	360;
	9157	2661	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2576	5860	0.019071	0.07388	0	0	0	0	0	0	1	-360	360;
	517	5860	0.02804	0.1256	0	0	0	0	0	0	1	-360	360;
	1572	5860	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5860	5755	0.001911	0.000751	0	260	0	0	0	0	1	-360	360;
	7015	4250	0.008098	0.028689	0	0	0	0	0	0	1	-360	360;
	7462	7015	0.017609	0.068191	0	0	0	0	0	0	1	-360	360;
	7015	4590	0.015018	0.058191	0	0	0	0	0	0	1	-360	360;
	6698	7012	0.007031	0.014578	0	0	0	0	0	0	1	-360	360;
	6698	4120	0.003791	0.012529	0	0	0	0	0	0	1	-360	360;
	6698	4604	0.010249	0.035209	0	0	0	0	0	0	1	-360	360;
	7740	937	0.002489	0.005818	0	0	0	0	0	0	1	-360	360;
	3395	7740	0.002071	0.009751	0	0	0	0	0	0	1	-360	360;
	3395	3258	0.001871	0.006902	0	0	0	0	0	0	1	-360	360;
	3395	5402	0.003658	0.014062	0	0	0	0	0	0	1	-360	360;
	7735	8000	0.002289	0.008178	0	0	0	0	0	0	1	-360	360;
	1215	8000	0.002329	0.008	0	0	0	0	0	0	1	-360	360;
	1215	2528	0.002071	0.009249	0	0	0	0	0	0	1	-360	360;
	1514	7168	0.005951	0.021249	0	0	0	0	0	0	1	-360	360;
	365	7168	0.006058	0.0208	0	0	0	0	0	0	1	-360	360;
	4731	1089	0.000942	0.003129	0	0	0	0	0	0	1	-360	360;
	4731	4120	0.030849	0.10176	0	0	0	0	0	0	1	-360	360;
	2244	1089	0.00084	0.003551	0	0	0	0	0	0	1	-360	360;
	2528	2244	0.002378	0.007831	0	0	0	0	0	0	1	-360	360;
	3837	1373	0.007231	0.01624	0	0	0	0	0	0	1	-360	360;
	1518	3837	0.001311	0.005111	0	0	0	0	0	0	1	-360	360;
	3837	5861	0.001338	0.005009	0	0	0	0	0	0	1	-360	360;
	3624	3837	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1373	570	0.000338	0.00116	0	0	0	0	0	0	1	-360	360;
	570	1288	0.025778	0.069751	0	0	0	0	0	0	1	-360	360;
	2213	570	0.004009	0.01376	0	0	0	0	0	0	1	-360	360;
	4915	570	0.00408	0.013462	0	0	0	0	0	0	1	-360	360;
	3735	570	0.000329	0.001262	0	0	0	0	0	0	1	-360	360;
	4790	570	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8649	1466	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2688	1300	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9062	1466	0.000711	0.002018	0	0	0	0	0	0	1	-360	360;
	1300	5066	0.00068	0.001502	0	0	0	0	0	0	1	-360	360;
	3764	936	0.00012	0.00056	0	0	0	0	0	0	1	-360	360;
	2925	936	0.02516	0.086418	0	0	0	0	0	0	1	-360	360;
	3764	6213	0.00012	0.00052	0	0	0	0	0	0	1	-360	360;
	6213	7001	0.032591	0.116511	0	0	0	0	0	0	1	-360	360;
	8178	3764	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2292	2184	0.01752	0.062098	0	0	0	0	0	0	1	-360	360;
	8435	2697	0.00076	0.003551	0	0	0	0	0	0	1	-360	360;
	9240	5446	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	9240	5638	0.000858	0.004458	0	0	0	0	0	0	1	-360	360;
	5638	6389	8e-05	0.000338	0	0	0	0	0	0	1	-360	360;
	5729	6789	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	6789	2783	0.000982	0.004369	0	0	0	0	0	0	1	-360	360;
	2779	5729	0.01704	0.083271	0	0	0	0	0	0	1	-360	360;
	2783	3040	7.1e-05	0.000351	0	0	0	0	0	0	1	-360	360;
	4883	6061	0.000271	0.00068	0	0	0	0	0	0	1	-360	360;
	9092	4883	0.012809	0.05376	0	0	0	0	0	0	1	-360	360;
	4883	7381	0.00364	0.015262	0	0	0	0	0	0	1	-360	360;
	1164	4963	0.00012	0.000431	0	0	0	0	0	0	1	-360	360;
	8394	4963	0.003982	0.015271	0	0	0	0	0	0	1	-360	360;
	1164	4123	0.00012	0.000418	0	0	0	0	0	0	1	-360	360;
	8745	4123	0.000431	0.00172	0	0	0	0	0	0	1	-360	360;
	631	3297	0.00024	0.00032	0	0	0	0	0	0	1	-360	360;
	3463	631	0.002418	0.011462	0	0	0	0	0	0	1	-360	360;
	6912	4456	0.00028	0.000311	0	0	0	0	0	0	1	-360	360;
	6912	8394	0.003658	0.01664	0	0	0	0	0	0	1	-360	360;
	8394	3297	0.003582	0.016951	0	0	0	0	0	0	1	-360	360;
	1875	5022	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5022	2347	0.006231	0.029249	0	0	0	0	0	0	1	-360	360;
	5022	8745	0.00144	0.005529	0	0	0	0	0	0	1	-360	360;
	5022	5950	0.011111	0.023978	0	0	0	0	0	0	1	-360	360;
	8736	1164	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2401	5588	0.001182	0.002409	0	0	0	0	0	0	1	-360	360;
	6301	3049	0.006098	0.029822	0	0	0	0	0	0	1	-360	360;
	3049	8253	0.00096	0.003378	0	0	0	0	0	0	1	-360	360;
	4194	6645	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6645	1598	0.01044	0.03444	0	0	0	0	0	0	1	-360	360;
	6645	1598	0.009298	0.03904	0	0	0	0	0	0	1	-360	360;
	6645	5265	0.017311	0.078738	0	0	0	0	0	0	1	-360	360;
	901	8970	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8970	537	0.012449	0.029129	0	0	0	0	0	0	1	-360	360;
	725	1202	0.00015	0.00045	0	790	0	0	0	0	1	-360	360;
	7616	1202	0.00107	0.01215	0	1316	0	0	0	0	1	-360	360;
	3056	725	6e-05	0.00056	0	1283	0	0	0	0	1	-360	360;
	4162	8171	0.00035	0.00284	0	1349	0	0	0	0	1	-360	360;
	7547	8171	0.00163	0.01512	0	1448	0	0	0	0	1	-360	360;
	4985	8171	0.00115	0.00913	0	1415	0	0	0	0	1	-360	360;
	2869	8171	0.00133	0.01513	0	889	0	0	0	0	1	-360	360;
	2869	8171	0.00131	0.01401	0	987	0	0	0	0	1	-360	360;
	7616	8171	0.00013	0.00129	0	1251	0	0	0	0	1	-360	360;
	7122	8171	0.00077	0.00664	0	1382	0	0	0	0	1	-360	360;
	2528	4120	0.028391	0.12908	0	0	0	0	0	0	1	-360	360;
	8615	2528	0.01216	0.05532	0	0	0	0	0	0	1	-360	360;
	8727	6148	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2399	6148	0.023609	0.053071	0	0	0	0	0	0	1	-360	360;
	4965	1761	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1761	6853	0.020471	0.051871	0	260	0	0	0	0	1	-360	360;
	3455	1761	0.01744	0.049862	0	0	0	0	0	0	1	-360	360;
	8590	2664	0.009111	0.039809	0	0	0	0	0	0	1	-360	360;
	5307	7062	0.033022	0.13856	0	0	0	0	0	0	1	-360	360;
	7062	2664	0.009529	0.045089	0	0	0	0	0	0	1	-360	360;
	5900	9062	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9062	865	0.001889	0.003462	0	260	0	0	0	0	1	-360	360;
	9062	859	0.006569	0.0142	0	286	0	0	0	0	1	-360	360;
	7046	9062	0.003982	0.018102	0	0	0	0	0	0	1	-360	360;
	9062	632	0.00232	0.007969	0	0	0	0	0	0	1	-360	360;
	1431	2262	0.002502	0.011698	0	0	0	0	0	0	1	-360	360;
	5553	2262	0.009351	0.039209	0	0	0	0	0	0	1	-360	360;
	2262	8712	0.008622	0.0296	0	0	0	0	0	0	1	-360	360;
	2045	2581	0.008951	0.037289	0	0	0	0	0	0	1	-360	360;
	2581	4528	0.005369	0.02144	0	0	0	0	0	0	1	-360	360;
	6727	8217	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5553	8217	0.00608	0.025551	0	0	0	0	0	0	1	-360	360;
	5553	8217	0.00596	0.026049	0	0	0	0	0	0	1	-360	360;
	8217	2045	0.001369	0.004898	0	0	0	0	0	0	1	-360	360;
	8217	2401	0.00144	0.006031	0	0	0	0	0	0	1	-360	360;
	4821	3289	0.0012	0.00328	0	0	0	0	0	0	1	-360	360;
	8394	4821	0.00132	0.006271	0	0	0	0	0	0	1	-360	360;
	4821	4981	0.009089	0.040711	0	0	0	0	0	0	1	-360	360;
	6330	7269	5.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	3165	7269	0.000302	0.00132	0	0	0	0	0	0	1	-360	360;
	4461	3074	0.00368	0.016502	0	0	0	0	0	0	1	-360	360;
	1966	4461	0.000329	0.00124	0	0	0	0	0	0	1	-360	360;
	5058	1934	0.0012	0.002689	0	0	0	0	0	0	1	-360	360;
	3165	1934	0.000338	0.001289	0	0	0	0	0	0	1	-360	360;
	414	6330	0.000942	0.002151	0	0	0	0	0	0	1	-360	360;
	5298	8042	0.00316	0.010862	0	0	0	0	0	0	1	-360	360;
	5298	6916	0.003222	0.010618	0	0	0	0	0	0	1	-360	360;
	3632	8042	0.010169	0.037858	0	0	0	0	0	0	1	-360	360;
	1824	6916	0.00552	0.024129	0	0	0	0	0	0	1	-360	360;
	2148	4236	0.000449	0.001889	0	0	0	0	0	0	1	-360	360;
	4579	6026	0.00044	0.001929	0	0	0	0	0	0	1	-360	360;
	3824	4962	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4962	7435	0.000418	0.00192	0	0	0	0	0	0	1	-360	360;
	4962	7190	0.000409	0.00196	0	0	0	0	0	0	1	-360	360;
	8394	7435	0.00608	0.022431	0	0	0	0	0	0	1	-360	360;
	3139	3984	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4299	3984	0.00016	0.000751	0	0	0	0	0	0	1	-360	360;
	4299	8591	0.046622	0.173511	0	0	0	0	0	0	1	-360	360;
	1510	3984	0.000151	0.000698	0	0	0	0	0	0	1	-360	360;
	2925	1510	0.01464	0.052329	0	0	0	0	0	0	1	-360	360;
	6042	6420	0.006129	0.023751	0	0	0	0	0	0	1	-360	360;
	4940	6420	0.001049	0.003311	0	0	0	0	0	0	1	-360	360;
	8540	5363	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5363	3074	0.003271	0.012089	0	0	0	0	0	0	1	-360	360;
	5363	1879	0.002951	0.013311	0	0	0	0	0	0	1	-360	360;
	5363	5150	0.017489	0.082769	0	0	0	0	0	0	1	-360	360;
	5363	2397	0.006231	0.029498	0	0	0	0	0	0	1	-360	360;
	2789	5250	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6372	8660	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7501	5250	0.00124	0.004458	0	0	0	0	0	0	1	-360	360;
	5870	8660	0.001169	0.001969	0	0	0	0	0	0	1	-360	360;
	4619	8833	0.000791	0.002929	0	0	0	0	0	0	1	-360	360;
	1071	8833	0.000911	0.003138	0	0	0	0	0	0	1	-360	360;
	1966	1879	0.004671	0.01672	0	0	0	0	0	0	1	-360	360;
	1467	2587	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2587	8368	0.00064	0.00188	0	0	0	0	0	0	1	-360	360;
	603	7657	0.00014	0.00061	0	889	0	0	0	0	1	-360	360;
	4782	5902	0.000698	0.0024	0	0	0	0	0	0	1	-360	360;
	4782	2469	0.000711	0.002351	0	0	0	0	0	0	1	-360	360;
	8394	5902	0.001858	0.006391	0	0	0	0	0	0	1	-360	360;
	4911	2469	0.014129	0.06328	0	0	0	0	0	0	1	-360	360;
	4590	4911	0.0002	0.00092	0	0	0	0	0	0	1	-360	360;
	9110	703	4e-05	0.0004	0	1382	0	0	0	0	1	-360	360;
	9110	4162	4e-05	0.00039	0	1251	0	0	0	0	1	-360	360;
	4985	703	0.00066	0.00705	0	1448	0	0	0	0	1	-360	360;
	4659	288	0.003622	0.016209	0	0	0	0	0	0	1	-360	360;
	9038	4659	5.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	2586	4659	3.1e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	8004	331	0.00084	0.002031	0	0	0	0	0	0	1	-360	360;
	7143	8257	0.004511	0.01972	0	0	0	0	0	0	1	-360	360;
	7143	8589	0.00064	0.0028	0	0	0	0	0	0	1	-360	360;
	8502	8004	0.004422	0.020102	0	0	0	0	0	0	1	-360	360;
	8004	4778	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6872	740	0	0.000222	0	0	0	0	0	0	1	-360	360;
	740	3223	0.000822	0.003622	0	0	0	0	0	0	1	-360	360;
	740	2347	0.000809	0.003689	0	0	0	0	0	0	1	-360	360;
	3528	6705	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2617	6705	0.001471	0.00528	0	0	0	0	0	0	1	-360	360;
	2617	6705	0.001502	0.005169	0	0	0	0	0	0	1	-360	360;
	2164	6705	0.000769	0.002858	0	0	0	0	0	0	1	-360	360;
	6705	2293	0.003698	0.013209	0	0	0	0	0	0	1	-360	360;
	6705	2148	0.00116	0.005271	0	0	0	0	0	0	1	-360	360;
	6705	5053	0.003378	0.006178	0	260	0	0	0	0	1	-360	360;
	6705	762	0.000849	0.004009	0	0	0	0	0	0	1	-360	360;
	6705	169	0.000431	0.00112	0	0	0	0	0	0	1	-360	360;
	2714	9156	0.006778	0.002631	0	0	0	0	0	0	1	-360	360;
	1530	3101	0.003569	0.00468	0	273	0	0	0	0	1	-360	360;
	1530	3101	0.003178	0.005298	0	260	0	0	0	0	1	-360	360;
	5833	4260	0.00388	0.0174	0	0	0	0	0	0	1	-360	360;
	744	4260	0.003138	0.011689	0	0	0	0	0	0	1	-360	360;
	5833	3376	0.0044	0.017049	0	0	0	0	0	0	1	-360	360;
	3376	7682	0.000791	0.002631	0	0	0	0	0	0	1	-360	360;
	5381	2945	0	0.000222	0	0	0	0	0	0	1	-360	360;
	462	6853	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2292	6853	0.017698	0.0584	0	0	0	0	0	0	1	-360	360;
	2292	6853	0.015769	0.066191	0	0	0	0	0	0	1	-360	360;
	4346	8345	0.00225	0.02425	0	823	0	0	0	0	1	-360	360;
	4346	8345	0.00161	0.02118	0	1152	0	0	0	0	1	-360	360;
	1132	8345	0.00312	0.03268	0	1251	0	0	0	0	1	-360	360;
	1132	8345	0.00305	0.0333	0	1382	0	0	0	0	1	-360	360;
	6230	5553	0.000182	0.000831	0	0	0	0	0	0	1	-360	360;
	8827	6230	0.013649	0.054609	0	0	0	0	0	0	1	-360	360;
	8979	7825	0.007209	0.03252	0	0	0	0	0	0	1	-360	360;
	8823	8979	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7825	5553	0.000169	0.000769	0	0	0	0	0	0	1	-360	360;
	2739	5553	0.011649	0.026778	0	0	0	0	0	0	1	-360	360;
	8257	2739	0.004049	0.01112	0	0	0	0	0	0	1	-360	360;
	8257	5553	0.007898	0.037338	0	0	0	0	0	0	1	-360	360;
	8257	5553	0.00772	0.034578	0	0	0	0	0	0	1	-360	360;
	5553	4699	0.042582	0.193609	0	0	0	0	0	0	1	-360	360;
	5553	8379	0.027591	0.130569	0	0	0	0	0	0	1	-360	360;
	5553	5548	0.018431	0.065898	0	0	0	0	0	0	1	-360	360;
	5553	31	0.02836	0.1134	0	0	0	0	0	0	1	-360	360;
	5553	961	0.046142	0.15844	0	0	0	0	0	0	1	-360	360;
	7535	5553	0.011409	0.027311	0	273	0	0	0	0	1	-360	360;
	5553	8712	0.019271	0.063569	0	0	0	0	0	0	1	-360	360;
	5553	8712	0.02212	0.082329	0	0	0	0	0	0	1	-360	360;
	2964	4792	0.018249	0.04624	0	0	0	0	0	0	1	-360	360;
	669	4792	0.008929	0.01872	0	0	0	0	0	0	1	-360	360;
	669	4792	0.008742	0.017329	0	0	0	0	0	0	1	-360	360;
	1577	5947	0.001569	0.00412	0	0	0	0	0	0	1	-360	360;
	5947	7826	0.037409	0.106911	0	0	0	0	0	0	1	-360	360;
	1577	8385	0.01836	0.049671	0	0	0	0	0	0	1	-360	360;
	4067	3600	0.007791	0.037502	0	0	0	0	0	0	1	-360	360;
	4067	1490	0.014911	0.051218	0	0	0	0	0	0	1	-360	360;
	2451	4895	0.007071	0.0294	0	0	0	0	0	0	1	-360	360;
	2451	7391	0.010169	0.039378	0	0	0	0	0	0	1	-360	360;
	8257	8502	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8257	6493	0.004449	0.00924	0	0	0	0	0	0	1	-360	360;
	3695	5330	0.002191	0.011142	0	0	0	0	0	0	1	-360	360;
	5330	8074	0.00244	0.011951	0	0	0	0	0	0	1	-360	360;
	6185	7185	0.004662	0.020871	0	0	0	0	0	0	1	-360	360;
	3695	7185	0.00528	0.020449	0	0	0	0	0	0	1	-360	360;
	6035	7185	0.0036	0.0144	0	0	0	0	0	0	1	-360	360;
	2851	3863	0.00212	0.008871	0	0	0	0	0	0	1	-360	360;
	6147	2851	0.001991	0.008271	0	0	0	0	0	0	1	-360	360;
	384	3863	0.002071	0.009049	0	0	0	0	0	0	1	-360	360;
	384	3933	0.002031	0.008098	0	0	0	0	0	0	1	-360	360;
	2234	2696	0.002631	0.00868	0	0	0	0	0	0	1	-360	360;
	2929	2696	0.00048	0.001769	0	0	0	0	0	0	1	-360	360;
	2234	5876	0.002338	0.009831	0	0	0	0	0	0	1	-360	360;
	5798	5876	0.006911	0.025742	0	0	0	0	0	0	1	-360	360;
	8299	2234	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2399	3417	0.005649	0.020209	0	0	0	0	0	0	1	-360	360;
	4589	3417	0.037822	0.075338	0	0	0	0	0	0	1	-360	360;
	9172	2399	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2399	7100	0.005609	0.020871	0	0	0	0	0	0	1	-360	360;
	2399	7610	0.042569	0.11516	0	0	0	0	0	0	1	-360	360;
	1537	7100	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8462	5739	0.00396	0.010458	0	0	0	0	0	0	1	-360	360;
	7001	5739	0.019671	0.088151	0	0	0	0	0	0	1	-360	360;
	1217	8587	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1801	5454	0.00912	0.03364	0	0	0	0	0	0	1	-360	360;
	5454	7319	0.010942	0.051369	0	0	0	0	0	0	1	-360	360;
	2970	8587	0.000182	0.000822	0	0	0	0	0	0	1	-360	360;
	2970	8591	0.023209	0.118049	0	0	0	0	0	0	1	-360	360;
	1801	8587	0.0002	0.000809	0	0	0	0	0	0	1	-360	360;
	2131	6235	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1533	7547	0.00076	0.00811	0	1448	0	0	0	0	1	-360	360;
	8152	2837	5.8e-05	0.000231	0	0	0	0	0	0	1	-360	360;
	7610	8152	0.013471	0.053849	0	0	0	0	0	0	1	-360	360;
	7307	7610	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1637	1864	0.007489	0.03376	0	0	0	0	0	0	1	-360	360;
	12	1637	0.004271	0.016431	0	0	0	0	0	0	1	-360	360;
	4004	5112	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8737	5112	0.013289	0.04384	0	0	0	0	0	0	1	-360	360;
	8737	5112	0.01184	0.049689	0	0	0	0	0	0	1	-360	360;
	8737	5112	0.011338	0.0508	0	0	0	0	0	0	1	-360	360;
	3831	5112	0.008502	0.035671	0	0	0	0	0	0	1	-360	360;
	5112	3588	0.02216	0.088622	0	0	0	0	0	0	1	-360	360;
	5112	5396	0.019911	0.101262	0	0	0	0	0	0	1	-360	360;
	5112	443	0.014551	0.068258	0	0	0	0	0	0	1	-360	360;
	5112	4422	0.014129	0.050529	0	0	0	0	0	0	1	-360	360;
	5112	7387	0.018022	0.08072	0	0	0	0	0	0	1	-360	360;
	517	7610	0.005382	0.02004	0	0	0	0	0	0	1	-360	360;
	9081	646	0.007769	0.034791	0	0	0	0	0	0	1	-360	360;
	646	7001	0.068351	0.30616	0	0	0	0	0	0	1	-360	360;
	1784	5402	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1396	9092	0	0.00028	0	0	0	0	0	0	1	-360	360;
	240	2292	0	0.00028	0	0	0	0	0	0	1	-360	360;
	14	436	0.00163	0.01819	0	1580	0	0	0	0	1	-360	360;
	5883	14	0.00241	0.02622	0	1251	0	0	0	0	1	-360	360;
	6741	14	0.00236	0.02672	0	1349	0	0	0	0	1	-360	360;
	14	2869	0.00188	0.02029	0	1020	0	0	0	0	1	-360	360;
	14	2869	0.00213	0.01988	0	1020	0	0	0	0	1	-360	360;
	1482	2925	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1007	2925	0.009822	0.046449	0	0	0	0	0	0	1	-360	360;
	7400	2925	0.00344	0.01628	0	0	0	0	0	0	1	-360	360;
	2925	1729	0.00648	0.025102	0	0	0	0	0	0	1	-360	360;
	3867	6347	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6347	183	0.02368	0.099338	0	0	0	0	0	0	1	-360	360;
	3455	6347	0.017169	0.061382	0	0	0	0	0	0	1	-360	360;
	1533	8845	0.00209	0.01796	0	1218	0	0	0	0	1	-360	360;
	1533	8845	0.00213	0.01758	0	1218	0	0	0	0	1	-360	360;
	5406	8845	0.0013	0.01389	0	1185	0	0	0	0	1	-360	360;
	4460	8845	0.00147	0.01361	0	1283	0	0	0	0	1	-360	360;
	603	4346	0.00052	0.00446	0	1251	0	0	0	0	1	-360	360;
	603	4346	0.00053	0.00436	0	1185	0	0	0	0	1	-360	360;
	2859	7734	0.014889	0.05544	0	0	0	0	0	0	1	-360	360;
	2859	7734	0.015178	0.054289	0	0	0	0	0	0	1	-360	360;
	2217	2859	0.003111	0.013569	0	0	0	0	0	0	1	-360	360;
	6322	2859	0.000791	0.00376	0	0	0	0	0	0	1	-360	360;
	2859	7478	0.002751	0.009809	0	0	0	0	0	0	1	-360	360;
	2859	9076	0.007129	0.024471	0	0	0	0	0	0	1	-360	360;
	2859	8813	0.017951	0.0668	0	0	0	0	0	0	1	-360	360;
	2859	5166	0.006178	0.02812	0	0	0	0	0	0	1	-360	360;
	2859	1687	0.006049	0.028649	0	0	0	0	0	0	1	-360	360;
	6072	4985	0.00396	0.03136	0	1251	0	0	0	0	1	-360	360;
	6072	4985	0.00353	0.03554	0	1119	0	0	0	0	1	-360	360;
	6072	4985	0.00387	0.0331	0	1283	0	0	0	0	1	-360	360;
	6072	4985	0.00394	0.03239	0	1251	0	0	0	0	1	-360	360;
	1972	1259	0.00225	0.01786	0	1086	0	0	0	0	1	-360	360;
	1259	3056	0.00124	0.01101	0	1152	0	0	0	0	1	-360	360;
	1972	4460	0.00201	0.02024	0	1185	0	0	0	0	1	-360	360;
	1972	6257	0.00197	0.02063	0	1086	0	0	0	0	1	-360	360;
	6257	3056	0.00126	0.01078	0	1185	0	0	0	0	1	-360	360;
	1972	5406	0.00193	0.02103	0	1251	0	0	0	0	1	-360	360;
	4919	2617	0.001209	0.004551	0	0	0	0	0	0	1	-360	360;
	2617	1319	0.000449	0.002018	0	0	0	0	0	0	1	-360	360;
	1056	2617	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5798	3779	0.007049	0.0252	0	0	0	0	0	0	1	-360	360;
	7651	5798	0.001191	0.003222	0	0	0	0	0	0	1	-360	360;
	5798	859	0.000511	0.001978	0	0	0	0	0	0	1	-360	360;
	5798	7046	0.001009	0.004782	0	0	0	0	0	0	1	-360	360;
	5798	8554	0.000991	0.004422	0	0	0	0	0	0	1	-360	360;
	3385	2580	0.000929	0.003098	0	0	0	0	0	0	1	-360	360;
	3774	3385	0.005858	0.0246	0	0	0	0	0	0	1	-360	360;
	1530	8904	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2714	8904	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5156	5696	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5696	1954	0.002031	0.009222	0	0	0	0	0	0	1	-360	360;
	5967	5696	0.01036	0.045231	0	0	0	0	0	0	1	-360	360;
	5696	2525	0.026858	0.062822	0	0	0	0	0	0	1	-360	360;
	810	1954	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1954	2664	0.021569	0.098089	0	0	0	0	0	0	1	-360	360;
	8457	8904	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4166	5281	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5281	3138	0.000418	0.001769	0	0	0	0	0	0	1	-360	360;
	5281	7319	0.000409	0.001809	0	0	0	0	0	0	1	-360	360;
	3010	4619	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5777	6779	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8354	4604	0.000569	0.00168	0	0	0	0	0	0	1	-360	360;
	8354	8730	0.000582	0.001649	0	0	0	0	0	0	1	-360	360;
	8730	4590	0.001858	0.006911	0	0	0	0	0	0	1	-360	360;
	9226	8354	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5041	272	0.0154	0.006022	0	286	0	0	0	0	1	-360	360;
	3473	5924	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5924	9098	8e-05	0.0004	0	0	0	0	0	0	1	-360	360;
	5924	1012	9.8e-05	0.000329	0	0	0	0	0	0	1	-360	360;
	2576	9098	0.023129	0.076289	0	0	0	0	0	0	1	-360	360;
	517	1012	0.011258	0.047249	0	0	0	0	0	0	1	-360	360;
	2576	6301	0.018822	0.072271	0	0	0	0	0	0	1	-360	360;
	2576	7763	0.002009	0.006649	0	0	0	0	0	0	1	-360	360;
	2576	4464	0.000942	0.00424	0	0	0	0	0	0	1	-360	360;
	2576	8253	0.017711	0.060809	0	0	0	0	0	0	1	-360	360;
	2576	8253	0.01804	0.059489	0	0	0	0	0	0	1	-360	360;
	5883	6072	0.0028	0.03055	0	1251	0	0	0	0	1	-360	360;
	5883	6072	0.00274	0.03113	0	1382	0	0	0	0	1	-360	360;
	1413	5186	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8591	288	0.0002	0.000942	0	0	0	0	0	0	1	-360	360;
	3911	6360	0.004351	0.014929	0	0	0	0	0	0	1	-360	360;
	3911	8591	0.004831	0.017289	0	0	0	0	0	0	1	-360	360;
	4796	4222	0.014289	0.067609	0	0	0	0	0	0	1	-360	360;
	4796	4222	0.013978	0.0626	0	0	0	0	0	0	1	-360	360;
	3632	4073	0.004431	0.014609	0	0	0	0	0	0	1	-360	360;
	3632	4080	0.00664	0.025711	0	0	0	0	0	0	1	-360	360;
	3632	7106	0.001289	0.004422	0	0	0	0	0	0	1	-360	360;
	3632	1824	0.01448	0.065822	0	0	0	0	0	0	1	-360	360;
	3632	3298	0.004631	0.017938	0	0	0	0	0	0	1	-360	360;
	3632	3298	0.00492	0.01692	0	0	0	0	0	0	1	-360	360;
	7001	5186	0.027511	0.120182	0	0	0	0	0	0	1	-360	360;
	6385	5186	0.004	0.014898	0	0	0	0	0	0	1	-360	360;
	7682	881	0.000351	0.001591	0	0	0	0	0	0	1	-360	360;
	7001	881	0.018929	0.084791	0	0	0	0	0	0	1	-360	360;
	1877	744	0.013502	0.063889	0	0	0	0	0	0	1	-360	360;
	1864	5371	0.01632	0.057871	0	0	0	0	0	0	1	-360	360;
	7822	2159	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2635	2159	0.01176	0.04704	0	0	0	0	0	0	1	-360	360;
	3117	7826	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9038	7826	0.02892	0.057622	0	260	0	0	0	0	1	-360	360;
	2753	4673	0.001711	0.0078	0	0	0	0	0	0	1	-360	360;
	2779	4673	0.01224	0.042022	0	0	0	0	0	0	1	-360	360;
	6479	2194	0.007769	0.061249	0	260	0	0	0	0	1	-360	360;
	8959	5687	0.001391	0.0048	0	0	0	0	0	0	1	-360	360;
	4551	8959	0.004071	0.01988	0	0	0	0	0	0	1	-360	360;
	7837	5687	0.001418	0.004702	0	0	0	0	0	0	1	-360	360;
	7837	3086	0.012711	0.048809	0	0	0	0	0	0	1	-360	360;
	5357	2032	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8146	5687	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2032	5687	0.00448	0.016698	0	0	0	0	0	0	1	-360	360;
	6943	4551	0.014689	0.052049	0	0	0	0	0	0	1	-360	360;
	2032	176	0.025631	0.084538	0	0	0	0	0	0	1	-360	360;
	8823	2628	0.005431	0.026551	0	0	0	0	0	0	1	-360	360;
	2628	4422	0.013151	0.058898	0	0	0	0	0	0	1	-360	360;
	5063	8257	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3627	183	0.00036	0.001329	0	0	0	0	0	0	1	-360	360;
	5937	8409	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8489	5937	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8409	5782	0.00612	0.014671	0	0	0	0	0	0	1	-360	360;
	7233	8489	0.006942	0.014378	0	286	0	0	0	0	1	-360	360;
	6200	3272	4.9e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	4108	6200	0.002071	0.008289	0	0	0	0	0	0	1	-360	360;
	5245	3272	4.9e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	5245	6935	0.000729	0.003049	0	0	0	0	0	0	1	-360	360;
	2476	3272	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3406	9186	0.000409	0.00184	0	0	0	0	0	0	1	-360	360;
	2964	9186	0.0004	0.00188	0	0	0	0	0	0	1	-360	360;
	448	6885	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6885	6819	0.003982	0.020249	0	0	0	0	0	0	1	-360	360;
	6885	8589	0.006498	0.031791	0	0	0	0	0	0	1	-360	360;
	8838	6404	0.000129	0.000609	0	0	0	0	0	0	1	-360	360;
	8838	614	0.00036	0.000729	0	0	0	0	0	0	1	-360	360;
	8838	8684	0.000129	0.000569	0	0	0	0	0	0	1	-360	360;
	1076	8838	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8684	538	0.006862	0.030942	0	0	0	0	0	0	1	-360	360;
	3553	4405	0.0016	0.00716	0	0	0	0	0	0	1	-360	360;
	3553	8092	0.001849	0.006542	0	0	0	0	0	0	1	-360	360;
	768	4405	0.001809	0.007018	0	0	0	0	0	0	1	-360	360;
	768	2960	0.004302	0.01588	0	0	0	0	0	0	1	-360	360;
	3252	4405	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7734	6855	0.008831	0.036751	0	0	0	0	0	0	1	-360	360;
	7734	65	0.011031	0.045942	0	0	0	0	0	0	1	-360	360;
	6562	7734	0	0.000222	0	0	0	0	0	0	1	-360	360;
	31	7734	0.043418	0.124111	0	0	0	0	0	0	1	-360	360;
	9070	3138	0.011698	0.048702	0	0	0	0	0	0	1	-360	360;
	8737	3831	0.005058	0.02572	0	0	0	0	0	0	1	-360	360;
	8080	3831	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1368	6018	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8752	8162	0.008969	0.032062	0	0	0	0	0	0	1	-360	360;
	2354	7024	0.000489	0.002369	0	0	0	0	0	0	1	-360	360;
	3165	7024	0.0028	0.0096	0	0	0	0	0	0	1	-360	360;
	2354	4707	0.00048	0.002191	0	0	0	0	0	0	1	-360	360;
	3165	4707	0.002849	0.0094	0	0	0	0	0	0	1	-360	360;
	4724	2354	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4160	8752	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8169	8752	0.005138	0.023391	0	0	0	0	0	0	1	-360	360;
	3339	8752	0.005031	0.023831	0	0	0	0	0	0	1	-360	360;
	6024	8752	0.00492	0.022062	0	0	0	0	0	0	1	-360	360;
	8752	4933	0.012449	0.03368	0	0	0	0	0	0	1	-360	360;
	4289	5307	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1374	5307	0.005582	0.021622	0	0	0	0	0	0	1	-360	360;
	5307	1931	0.011769	0.052729	0	0	0	0	0	0	1	-360	360;
	1791	7612	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8615	7612	0.005489	0.019618	0	0	0	0	0	0	1	-360	360;
	7462	7612	0.005591	0.0192	0	0	0	0	0	0	1	-360	360;
	7034	1521	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7300	8169	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3339	7858	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1521	7858	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3573	1061	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9147	3573	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1061	7172	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7172	6024	0	0.000222	0	0	0	0	0	0	1	-360	360;
	383	6875	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5150	494	0	0.000222	0	0	0	0	0	0	1	-360	360;
	383	9147	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3998	1231	0	0.000222	0	0	0	0	0	0	1	-360	360;
	318	3998	0.012969	0.046369	0	0	0	0	0	0	1	-360	360;
	1231	3686	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3686	1374	0	0.000222	0	0	0	0	0	0	1	-360	360;
	494	7300	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8023	7470	0.003151	0.00128	0	0	0	0	0	0	1	-360	360;
	1533	310	0.00154	0.01649	0	1415	0	0	0	0	1	-360	360;
	3165	5858	0.001298	0.00568	0	0	0	0	0	0	1	-360	360;
	3165	1831	0.0002	0.000871	0	0	0	0	0	0	1	-360	360;
	8769	7921	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9081	7921	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	3004	7921	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	3004	8591	0.010071	0.039	0	0	0	0	0	0	1	-360	360;
	4328	5496	0.000218	0.000849	0	0	0	0	0	0	1	-360	360;
	1840	5496	0.000231	0.00084	0	0	0	0	0	0	1	-360	360;
	8616	5496	0.000231	0.000822	0	0	0	0	0	0	1	-360	360;
	101	5496	0.00024	0.0008	0	0	0	0	0	0	1	-360	360;
	7057	5496	0.00024	0.000782	0	0	0	0	0	0	1	-360	360;
	5496	3812	0.000209	0.000889	0	0	0	0	0	0	1	-360	360;
	11	5496	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3201	5496	0.000209	0.000911	0	0	0	0	0	0	1	-360	360;
	4781	1840	0.00784	0.025862	0	0	0	0	0	0	1	-360	360;
	8616	5867	0.001991	0.009391	0	0	0	0	0	0	1	-360	360;
	101	5013	0.003591	0.01184	0	0	0	0	0	0	1	-360	360;
	7712	537	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1533	2000	0.00013	0.00077	0	954	0	0	0	0	1	-360	360;
	1059	2000	0.00013	0.00076	0	987	0	0	0	0	1	-360	360;
	914	2000	0.0002	0.00127	0	987	0	0	0	0	1	-360	360;
	4083	4853	0.002449	0.006871	0	0	0	0	0	0	1	-360	360;
	2542	7339	0.00132	0.004929	0	260	0	0	0	0	1	-360	360;
	5925	2391	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2391	2525	0.036111	0.084471	0	0	0	0	0	0	1	-360	360;
	2391	2757	0.043129	0.093129	0	0	0	0	0	0	1	-360	360;
	6018	8452	0.001071	0.004129	0	0	0	0	0	0	1	-360	360;
	1166	8452	0.01016	0.036298	0	0	0	0	0	0	1	-360	360;
	4822	6018	0.010951	0.055689	0	0	0	0	0	0	1	-360	360;
	3901	7088	0.005831	0.022591	0	0	0	0	0	0	1	-360	360;
	95	7088	0.016542	0.075222	0	0	0	0	0	0	1	-360	360;
	7088	5318	0.015938	0.00716	0	0	0	0	0	0	1	-360	360;
	4839	2612	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2612	546	0.009591	0.02332	0	0	0	0	0	0	1	-360	360;
	4839	5782	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2664	3678	0.03828	0.10096	0	0	0	0	0	0	1	-360	360;
	95	3678	0.03036	0.086778	0	260	0	0	0	0	1	-360	360;
	3318	1288	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7863	2217	0.002182	0.010329	0	0	0	0	0	0	1	-360	360;
	7863	4699	0.002129	0.009569	0	0	0	0	0	0	1	-360	360;
	5659	3060	0.0018	0.00668	0	0	0	0	0	0	1	-360	360;
	2593	3060	0.004569	0.01636	0	0	0	0	0	0	1	-360	360;
	5659	3172	0.001831	0.006542	0	0	0	0	0	0	1	-360	360;
	2593	4316	0.006338	0.026609	0	0	0	0	0	0	1	-360	360;
	6955	7001	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7001	6078	0.016231	0.070902	0	0	0	0	0	0	1	-360	360;
	7001	8303	0.010138	0.046102	0	0	0	0	0	0	1	-360	360;
	5903	3455	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3455	9104	0.006431	0.028991	0	0	0	0	0	0	1	-360	360;
	2753	6223	0.01184	0.053022	0	0	0	0	0	0	1	-360	360;
	2753	2779	0.012529	0.044422	0	0	0	0	0	0	1	-360	360;
	7587	5634	0.00436	0.010689	0	0	0	0	0	0	1	-360	360;
	5030	7587	0.019529	0.055831	0	0	0	0	0	0	1	-360	360;
	3181	5634	0.004942	0.01048	0	0	0	0	0	0	1	-360	360;
	8054	5634	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1272	2422	0.006431	0.01764	0	0	0	0	0	0	1	-360	360;
	4621	1272	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	8901	2422	0.003431	0.013218	0	0	0	0	0	0	1	-360	360;
	5508	2422	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2422	5951	0.00708	0.014702	0	260	0	0	0	0	1	-360	360;
	3626	2422	0.065422	0.130311	0	0	0	0	0	0	1	-360	360;
	8901	3561	0.003049	0.013831	0	0	0	0	0	0	1	-360	360;
	9092	8901	0.011209	0.051009	0	0	0	0	0	0	1	-360	360;
	812	8901	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8002	8901	0.00504	0.024622	0	0	0	0	0	0	1	-360	360;
	2292	3561	0.009049	0.033671	0	0	0	0	0	0	1	-360	360;
	5661	3561	0	0.00028	0	0	0	0	0	0	1	-360	360;
	4621	6644	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	4798	4621	0	0.00028	0	0	0	0	0	0	1	-360	360;
	9092	5951	0.007631	0.015569	0	0	0	0	0	0	1	-360	360;
	9092	2454	0.000822	0.00316	0	0	0	0	0	0	1	-360	360;
	9092	7453	0.017751	0.060929	0	0	0	0	0	0	1	-360	360;
	2292	6644	0.007418	0.016022	0	260	0	0	0	0	1	-360	360;
	2292	9041	0.007542	0.035689	0	0	0	0	0	0	1	-360	360;
	8932	469	0	0.000222	0	0	0	0	0	0	1	-360	360;
	469	1296	0.000338	0.001218	0	0	0	0	0	0	1	-360	360;
	8559	178	0	0.000222	0	0	0	0	0	0	1	-360	360;
	178	7801	0.000351	0.001191	0	0	0	0	0	0	1	-360	360;
	6944	1166	0.004662	0.016009	0	0	0	0	0	0	1	-360	360;
	6944	8394	0.017489	0.084191	0	0	0	0	0	0	1	-360	360;
	1166	30	0.003809	0.013111	0	0	0	0	0	0	1	-360	360;
	1166	6094	0.003489	0.013502	0	0	0	0	0	0	1	-360	360;
	3627	1166	0.004738	0.015658	0	0	0	0	0	0	1	-360	360;
	3185	1166	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1166	2571	0.010351	0.035529	0	0	0	0	0	0	1	-360	360;
	1166	6364	0.00388	0.012822	0	0	0	0	0	0	1	-360	360;
	1296	30	0.003729	0.013218	0	0	0	0	0	0	1	-360	360;
	6094	7095	0.002418	0.008298	0	0	0	0	0	0	1	-360	360;
	7801	7095	0.001538	0.007822	0	0	0	0	0	0	1	-360	360;
	1288	6310	0.005369	0.010702	0	260	0	0	0	0	1	-360	360;
	2522	7427	0.00088	0.0075	0	1481	0	0	0	0	1	-360	360;
	2522	7427	0.0009	0.00734	0	1415	0	0	0	0	1	-360	360;
	2095	7233	0	0.00028	0	286	0	0	0	0	1	-360	360;
	1569	8340	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2095	8340	0	0.00028	0	286	0	0	0	0	1	-360	360;
	5953	8340	0.018831	0.072951	0	0	0	0	0	0	1	-360	360;
	6241	8340	0.012351	0.042409	0	0	0	0	0	0	1	-360	360;
	5030	3901	0.001938	0.008698	0	0	0	0	0	0	1	-360	360;
	2052	5030	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6437	8655	0.017871	0.047151	0	0	0	0	0	0	1	-360	360;
	8655	669	0.00076	0.002098	0	0	0	0	0	0	1	-360	360;
	5371	5041	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8074	2571	0.005249	0.024631	0	0	0	0	0	0	1	-360	360;
	8094	8074	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1727	3992	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8506	3992	0.00016	0.000502	0	0	0	0	0	0	1	-360	360;
	3795	8506	0.008658	0.037809	0	0	0	0	0	0	1	-360	360;
	6078	3992	0.000142	0.00056	0	0	0	0	0	0	1	-360	360;
	6765	6437	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	6765	5606	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	7478	3154	0.0002	0.00092	0	0	0	0	0	0	1	-360	360;
	3154	1310	0.002542	0.010649	0	0	0	0	0	0	1	-360	360;
	6322	31	0.054	0.121351	0	0	0	0	0	0	1	-360	360;
	6593	7478	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3460	4287	0.001969	0.004271	0	0	0	0	0	0	1	-360	360;
	3460	3626	0.009111	0.025009	0	286	0	0	0	0	1	-360	360;
	1735	3460	0.027502	0.074369	0	0	0	0	0	0	1	-360	360;
	7871	6774	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6774	3588	0.0002	0.000942	0	0	0	0	0	0	1	-360	360;
	6774	6819	0.0002	0.000871	0	0	0	0	0	0	1	-360	360;
	1986	9179	0.00052	0.00136	0	0	0	0	0	0	1	-360	360;
	3175	1986	0.020471	0.07928	0	0	0	0	0	0	1	-360	360;
	4841	845	3.1e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	6226	845	0.015609	0.076289	0	0	0	0	0	0	1	-360	360;
	4841	8002	3.1e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	2374	4841	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8436	7290	0.008809	0.041342	0	0	0	0	0	0	1	-360	360;
	8694	7290	0.011698	0.017782	0	0	0	0	0	0	1	-360	360;
	5950	8436	0.029769	0.069649	0	260	0	0	0	0	1	-360	360;
	5827	5707	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5953	5707	0.006738	0.024089	0	0	0	0	0	0	1	-360	360;
	5827	8131	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1376	5827	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8606	3704	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6162	2542	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3704	2542	0	0.000222	0	0	0	0	0	0	1	-360	360;
	885	3704	0.001062	0.003649	0	0	0	0	0	0	1	-360	360;
	2664	1780	0.00204	0.009129	0	0	0	0	0	0	1	-360	360;
	2664	1780	0.00232	0.008951	0	0	0	0	0	0	1	-360	360;
	2664	5980	0.00152	0.006911	0	0	0	0	0	0	1	-360	360;
	2664	5980	0.001489	0.00704	0	0	0	0	0	0	1	-360	360;
	2664	5980	0.001462	0.00652	0	0	0	0	0	0	1	-360	360;
	2664	8747	0.001391	0.003609	0	260	0	0	0	0	1	-360	360;
	2664	5968	0.001658	0.003311	0	0	0	0	0	0	1	-360	360;
	2664	8154	0.000991	0.004329	0	0	0	0	0	0	1	-360	360;
	2664	2690	0.010369	0.038569	0	0	0	0	0	0	1	-360	360;
	1332	1780	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1332	6032	0.026809	0.126889	0	0	0	0	0	0	1	-360	360;
	8632	1780	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8632	8461	0.026218	0.117489	0	0	0	0	0	0	1	-360	360;
	5980	1780	0.00064	0.002662	0	0	0	0	0	0	1	-360	360;
	5980	1780	0.000622	0.002711	0	0	0	0	0	0	1	-360	360;
	5980	86	0.000609	0.002271	0	0	0	0	0	0	1	-360	360;
	4642	8747	0.002098	0.005702	0	260	0	0	0	0	1	-360	360;
	2179	5968	0.002582	0.006031	0	0	0	0	0	0	1	-360	360;
	4093	2179	0.019458	0.055618	0	0	0	0	0	0	1	-360	360;
	1454	8154	0.00092	0.003289	0	0	0	0	0	0	1	-360	360;
	1454	86	0.000942	0.003222	0	0	0	0	0	0	1	-360	360;
	8238	1454	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4642	6125	0.03868	0.1062	0	260	0	0	0	0	1	-360	360;
	5818	9170	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2397	9170	0.000689	0.00312	0	0	0	0	0	0	1	-360	360;
	318	9170	0.0008	0.002991	0	0	0	0	0	0	1	-360	360;
	997	8400	0.000511	0.001191	0	0	0	0	0	0	1	-360	360;
	8842	670	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2525	2781	0.01216	0.048609	0	0	0	0	0	0	1	-360	360;
	3858	2690	0.000769	0.00348	0	0	0	0	0	0	1	-360	360;
	3858	5967	0.000871	0.003409	0	0	0	0	0	0	1	-360	360;
	6301	8253	0.007071	0.03316	0	0	0	0	0	0	1	-360	360;
	538	5635	0.009609	0.038418	0	0	0	0	0	0	1	-360	360;
	758	538	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8673	5635	0.00724	0.014951	0	0	0	0	0	0	1	-360	360;
	3244	4316	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7990	4316	0.016071	0.067418	0	0	0	0	0	0	1	-360	360;
	7990	4316	0.015738	0.068751	0	0	0	0	0	0	1	-360	360;
	4316	4233	0.017431	0.037622	0	0	0	0	0	0	1	-360	360;
	6767	6412	0.001031	0.003378	0	0	0	0	0	0	1	-360	360;
	6412	8011	0.000169	0.00068	0	0	0	0	0	0	1	-360	360;
	1687	6767	0.006249	0.023262	0	0	0	0	0	0	1	-360	360;
	9076	6042	8.9e-05	0.000471	0	0	0	0	0	0	1	-360	360;
	2991	2635	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4247	2635	0.01564	0.065089	0	0	0	0	0	0	1	-360	360;
	2635	1701	0.018938	0.085391	0	0	0	0	0	0	1	-360	360;
	2635	1476	0.016511	0.044662	0	0	0	0	0	0	1	-360	360;
	8607	517	0	0.000222	0	0	0	0	0	0	1	-360	360;
	517	8673	0.002689	0.012938	0	0	0	0	0	0	1	-360	360;
	5072	2873	0	0.000222	0	0	0	0	0	0	1	-360	360;
	821	5072	0.007649	0.019631	0	0	0	0	0	0	1	-360	360;
	6255	363	0.008849	0.02096	0	0	0	0	0	0	1	-360	360;
	2873	6255	0	0.000222	0	0	0	0	0	0	1	-360	360;
	363	4835	0	0.00028	0	0	0	0	0	0	1	-360	360;
	12	2378	0.00232	0.00928	0	0	0	0	0	0	1	-360	360;
	2697	1518	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7453	5861	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3392	2213	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5879	3392	0.0088	0.034102	0	0	0	0	0	0	1	-360	360;
	7381	4915	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7020	3247	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8160	4083	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5414	1608	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7846	7693	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7704	5267	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6804	8839	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8839	6276	0.002542	0.00392	0	273	0	0	0	0	1	-360	360;
	7501	4568	0.00296	0.013462	0	0	0	0	0	0	1	-360	360;
	7939	7501	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7501	103	0.002489	0.003991	0	260	0	0	0	0	1	-360	360;
	767	5066	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7976	5066	0.000289	0.000742	0	0	0	0	0	0	1	-360	360;
	7546	3927	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2580	3927	0.000302	0.000729	0	0	0	0	0	0	1	-360	360;
	3927	6976	0.000662	0.002111	0	0	0	0	0	0	1	-360	360;
	4404	1206	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6531	1608	0.000822	0.002871	0	0	0	0	0	0	1	-360	360;
	1608	8686	0.006391	0.028809	0	0	0	0	0	0	1	-360	360;
	2825	7693	0.000831	0.002809	0	0	0	0	0	0	1	-360	360;
	6814	6531	0.014791	0.005969	0	0	0	0	0	0	1	-360	360;
	4321	2825	0.01508	0.005849	0	0	0	0	0	0	1	-360	360;
	4265	5267	0.011302	0.050578	0	0	0	0	0	0	1	-360	360;
	997	5267	0.01704	0.033938	0	0	0	0	0	0	1	-360	360;
	8445	5396	0.001262	0.00532	0	0	0	0	0	0	1	-360	360;
	8445	2960	0.00124	0.005431	0	0	0	0	0	0	1	-360	360;
	6309	357	0	0.00028	0	0	0	0	0	0	1	-360	360;
	357	2652	0.001538	0.004609	0	0	0	0	0	0	1	-360	360;
	210	8739	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2701	6013	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8296	6013	0.00244	0.004071	0	299	0	0	0	0	1	-360	360;
	2210	5136	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5568	5136	0.002649	0.003631	0	273	0	0	0	0	1	-360	360;
	4973	1490	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2529	1370	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8281	5858	0.000849	0.001769	0	0	0	0	0	0	1	-360	360;
	5858	7868	0.000951	0.003991	0	0	0	0	0	0	1	-360	360;
	4691	4835	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4691	7911	0.0034	0.00872	0	0	0	0	0	0	1	-360	360;
	6858	7990	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1206	4730	0.001978	0.007671	0	0	0	0	0	0	1	-360	360;
	1206	906	0.002018	0.00752	0	0	0	0	0	0	1	-360	360;
	6984	4247	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4247	961	0.018351	0.070502	0	0	0	0	0	0	1	-360	360;
	8394	5402	0.0028	0.014209	0	0	0	0	0	0	1	-360	360;
	931	9008	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7382	1598	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6935	1598	0.02624	0.096858	0	0	0	0	0	0	1	-360	360;
	450	5939	5.8e-05	0.00024	0	0	0	0	0	0	1	-360	360;
	450	7106	5.8e-05	0.000231	0	0	0	0	0	0	1	-360	360;
	6527	4080	0.0012	0.005809	0	0	0	0	0	0	1	-360	360;
	6527	7542	8.9e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	6527	6380	8e-05	0.00032	0	0	0	0	0	0	1	-360	360;
	7542	3298	0.006889	0.032631	0	0	0	0	0	0	1	-360	360;
	6380	3298	0.006738	0.030209	0	0	0	0	0	0	1	-360	360;
	5248	2929	0.001218	0.004031	0	0	0	0	0	0	1	-360	360;
	2929	997	0.005751	0.012929	0	286	0	0	0	0	1	-360	360;
	161	3779	0.000462	0.001791	0	0	0	0	0	0	1	-360	360;
	161	6235	0.00352	0.007302	0	0	0	0	0	0	1	-360	360;
	74	4919	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4483	2164	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1322	5642	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5642	990	0.001289	0.005022	0	0	0	0	0	0	1	-360	360;
	3567	8064	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8386	8686	0.00036	0.0016	0	0	0	0	0	0	1	-360	360;
	8386	5080	0.000409	0.001569	0	0	0	0	0	0	1	-360	360;
	8234	8386	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1222	8386	0.001658	0.004511	0	0	0	0	0	0	1	-360	360;
	5953	2603	0.00176	0.007911	0	0	0	0	0	0	1	-360	360;
	990	5953	0.00172	0.008058	0	0	0	0	0	0	1	-360	360;
	6417	5953	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1449	8686	0.0022	0.00852	0	0	0	0	0	0	1	-360	360;
	7552	5080	0.000111	0.000489	0	0	0	0	0	0	1	-360	360;
	7552	8700	0.00396	0.013058	0	0	0	0	0	0	1	-360	360;
	2603	8796	0.001138	0.002529	0	0	0	0	0	0	1	-360	360;
	8637	1660	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8985	8637	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8637	2674	0.001329	0.005978	0	0	0	0	0	0	1	-360	360;
	5382	8827	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5484	6533	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6533	7911	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5484	3295	0.010191	0.028329	0	0	0	0	0	0	1	-360	360;
	1832	7057	0.002391	0.00788	0	0	0	0	0	0	1	-360	360;
	1055	1832	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5867	1832	0.002249	0.008351	0	0	0	0	0	0	1	-360	360;
	6113	3040	0.000302	0.000969	0	0	0	0	0	0	1	-360	360;
	2445	6389	0.000271	0.001102	0	0	0	0	0	0	1	-360	360;
	3812	6113	0.000391	0.001738	0	0	0	0	0	0	1	-360	360;
	2445	3201	0.002289	0.008178	0	0	0	0	0	0	1	-360	360;
	7023	5867	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5867	4781	0.00112	0.00504	0	0	0	0	0	0	1	-360	360;
	5867	4781	0.001271	0.004942	0	0	0	0	0	0	1	-360	360;
	3619	4781	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4781	7229	0.000418	0.00192	0	0	0	0	0	0	1	-360	360;
	4781	4479	0.000409	0.001951	0	0	0	0	0	0	1	-360	360;
	4781	95	0.03228	0.106458	0	0	0	0	0	0	1	-360	360;
	4781	95	0.02876	0.120649	0	0	0	0	0	0	1	-360	360;
	4479	95	0.05888	0.117289	0	0	0	0	0	0	1	-360	360;
	6935	8061	0.00324	0.01584	0	0	0	0	0	0	1	-360	360;
	1281	8061	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3735	7710	0.00176	0.004449	0	273	0	0	0	0	1	-360	360;
	8233	1007	0.009929	0.046969	0	0	0	0	0	0	1	-360	360;
	8233	2158	0.009711	0.043489	0	0	0	0	0	0	1	-360	360;
	6385	2158	0.005031	0.02252	0	0	0	0	0	0	1	-360	360;
	1725	2652	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8538	6428	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7389	6428	0.000462	0.00108	0	0	0	0	0	0	1	-360	360;
	8111	2652	0.000471	0.001049	0	273	0	0	0	0	1	-360	360;
	1222	1449	0.00104	0.003729	0	0	0	0	0	0	1	-360	360;
	8379	5224	0.002329	0.008	0	0	0	0	0	0	1	-360	360;
	8813	5224	0.002378	0.007831	0	0	0	0	0	0	1	-360	360;
	2010	5224	0	0.000222	0	0	0	0	0	0	1	-360	360;
	598	6385	0.006711	0.026018	0	0	0	0	0	0	1	-360	360;
	6385	1729	0.009298	0.04232	0	0	0	0	0	0	1	-360	360;
	8712	5548	0.004751	0.01632	0	0	0	0	0	0	1	-360	360;
	2778	1891	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1891	3086	0.000338	0.001151	0	0	0	0	0	0	1	-360	360;
	1891	6226	0.000338	0.00112	0	0	0	0	0	0	1	-360	360;
	8688	7651	0	0.00028	0	0	0	0	0	0	1	-360	360;
	4222	470	0.00484	0.023329	0	0	0	0	0	0	1	-360	360;
	5151	6814	0.004222	0.017751	0	0	0	0	0	0	1	-360	360;
	1149	4321	0.004138	0.018102	0	0	0	0	0	0	1	-360	360;
	4473	3315	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4473	1226	0.000622	0.002418	0	0	0	0	0	0	1	-360	360;
	3315	632	0.000631	0.002369	0	0	0	0	0	0	1	-360	360;
	7763	8856	0.00012	0.00032	0	0	0	0	0	0	1	-360	360;
	8048	5265	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1099	5950	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2254	103	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8327	6276	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5895	627	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3408	7433	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2270	8591	0.001409	0.00464	0	0	0	0	0	0	1	-360	360;
	1735	2270	0.003098	0.014911	0	0	0	0	0	0	1	-360	360;
	8591	2586	0.004662	0.015369	0	0	0	0	0	0	1	-360	360;
	119	8220	9.8e-05	0.000369	0	0	0	0	0	0	1	-360	360;
	8591	119	0.001031	0.003658	0	0	0	0	0	0	1	-360	360;
	3298	5761	0.001049	0.003591	0	0	0	0	0	0	1	-360	360;
	6147	3690	0.003209	0.012338	0	0	0	0	0	0	1	-360	360;
	6364	6147	0.003311	0.012689	0	0	0	0	0	0	1	-360	360;
	7987	2525	0	0.00028	0	0	0	0	0	0	1	-360	360;
	95	373	0.003871	0.010458	0	0	0	0	0	0	1	-360	360;
	8461	95	0.000218	0.000849	0	0	0	0	0	0	1	-360	360;
	6032	95	0.000231	0.00084	0	0	0	0	0	0	1	-360	360;
	9163	95	0	0.00028	0	0	0	0	0	0	1	-360	360;
	373	2843	0.000231	0.000822	0	0	0	0	0	0	1	-360	360;
	2757	2843	0.00024	0.0008	0	0	0	0	0	0	1	-360	360;
	2356	2843	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5651	4822	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2548	5870	0.000151	0.00036	0	0	0	0	0	0	1	-360	360;
	2548	8296	8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	1071	5568	8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	1389	2548	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2548	1071	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2227	1071	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1903	8703	0.001711	0.003058	0	0	0	0	0	0	1	-360	360;
	7868	1903	0.003369	0.008889	0	260	0	0	0	0	1	-360	360;
	1749	8703	0.002009	0.005311	0	260	0	0	0	0	1	-360	360;
	2953	1749	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8092	856	0.002551	0.011578	0	0	0	0	0	0	1	-360	360;
	4108	856	0.002489	0.0118	0	0	0	0	0	0	1	-360	360;
	7251	8394	0	0.00028	0	0	0	0	0	0	1	-360	360;
	8394	3933	0.010009	0.04892	0	0	0	0	0	0	1	-360	360;
	4940	3681	0.000182	0.000671	0	0	0	0	0	0	1	-360	360;
	7720	7409	0.006169	0.02036	0	0	0	0	0	0	1	-360	360;
	7409	1476	0.007369	0.0322	0	0	0	0	0	0	1	-360	360;
	6549	7720	0.005498	0.023071	0	0	0	0	0	0	1	-360	360;
	8011	6549	0.004809	0.021671	0	0	0	0	0	0	1	-360	360;
	4229	7720	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7720	961	0.016169	0.06784	0	0	0	0	0	0	1	-360	360;
	6096	3626	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8848	31	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5506	31	0.01124	0.039822	0	0	0	0	0	0	1	-360	360;
	6943	176	0.001169	0.00532	0	0	0	0	0	0	1	-360	360;
	470	176	0.001138	0.005418	0	0	0	0	0	0	1	-360	360;
	8558	3482	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6643	5506	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2293	8081	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4800	5956	0.00212	0.008871	0	0	0	0	0	0	1	-360	360;
	2148	4800	0.002289	0.008902	0	0	0	0	0	0	1	-360	360;
	5956	7333	0.00024	0.000782	0	0	0	0	0	0	1	-360	360;
	3175	7333	0.000209	0.000889	0	0	0	0	0	0	1	-360	360;
	5663	8081	0	0.000222	0	0	0	0	0	0	1	-360	360;
	100	8081	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2067	7333	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1745	4545	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4545	5392	0.001209	0.005529	0	0	0	0	0	0	1	-360	360;
	4545	4233	0.001191	0.00564	0	0	0	0	0	0	1	-360	360;
	2856	5392	0.014982	0.0428	0	0	0	0	0	0	1	-360	360;
	6310	7832	0.012849	0.022929	0	0	0	0	0	0	1	-360	360;
	7832	1771	0.000431	0.000929	0	0	0	0	0	0	1	-360	360;
	8694	1771	0.000849	0.003831	0	0	0	0	0	0	1	-360	360;
	7566	1771	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3946	3108	0.000231	0.000769	0	0	0	0	0	0	1	-360	360;
	3946	4265	0.0002	0.000871	0	0	0	0	0	0	1	-360	360;
	7402	3946	0	0.00028	0	0	0	0	0	0	1	-360	360;
	4070	598	0.004049	0.01844	0	0	0	0	0	0	1	-360	360;
	4070	3795	0.003969	0.018791	0	0	0	0	0	0	1	-360	360;
	4932	4070	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6125	391	0.002271	0.004258	0	260	0	0	0	0	1	-360	360;
	2856	3667	0.012449	0.0248	0	0	0	0	0	0	1	-360	360;
	3667	391	0.002311	0.004169	0	0	0	0	0	0	1	-360	360;
	4018	2856	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7192	391	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4751	3593	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8700	3593	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6241	3593	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2339	8303	0.002071	0.009049	0	0	0	0	0	0	1	-360	360;
	2339	232	0.002031	0.009222	0	0	0	0	0	0	1	-360	360;
	9142	2339	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9050	961	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1476	961	0.015511	0.06772	0	0	0	0	0	0	1	-360	360;
	8090	9041	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	8090	5879	4e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	28	8090	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7470	1678	0.0052	0.013898	0	0	0	0	0	0	1	-360	360;
	9104	4174	0.009071	0.032151	0	0	0	0	0	0	1	-360	360;
	6620	9104	0	0.00028	0	0	0	0	0	0	1	-360	360;
	546	8082	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1678	6957	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8082	821	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8082	3295	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8082	6957	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5359	7535	0.042889	0.088942	0	0	0	0	0	0	1	-360	360;
	762	6026	0.000609	0.001618	0	0	0	0	0	0	1	-360	360;
	443	8827	0.011289	0.057409	0	0	0	0	0	0	1	-360	360;
	4030	865	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5166	3774	8.9e-05	0.000449	0	0	0	0	0	0	1	-360	360;
	8231	4778	0.000169	0.000662	0	0	0	0	0	0	1	-360	360;
	2312	5053	0	0.00028	0	0	0	0	0	0	1	-360	360;
	1310	5796	9.8e-05	0.000511	0	0	0	0	0	0	1	-360	360;
	8631	5796	0.006338	0.022662	0	0	0	0	0	0	1	-360	360;
	912	7046	0.000809	0.001729	0	0	0	0	0	0	1	-360	360;
	8554	1226	0.00204	0.007898	0	0	0	0	0	0	1	-360	360;
	5013	8984	0.00188	0.00516	0	0	0	0	0	0	1	-360	360;
	8631	7976	0.000831	0.003511	0	0	0	0	0	0	1	-360	360;
	1319	169	0.00064	0.002871	0	0	0	0	0	0	1	-360	360;
	6976	3681	0.000182	0.00064	0	0	0	0	0	0	1	-360	360;
	6976	1701	0.000191	0.000631	0	0	0	0	0	0	1	-360	360;
	3532	6976	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2113	310	0.00043	0.00337	0	1152	0	0	0	0	1	-360	360;
	2113	7122	0.00038	0.00383	0	1481	0	0	0	0	1	-360	360;
	4981	5384	0.000431	0.00112	0	273	0	0	0	0	1	-360	360;
	1831	4981	0.006458	0.021302	0	0	0	0	0	0	1	-360	360;
	7912	2779	0	0.00028	0	0	0	0	0	0	1	-360	360;
	7400	9070	0.000249	0.001138	0	0	0	0	0	0	1	-360	360;
	8548	7400	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7400	1729	0.002449	0.011169	0	0	0	0	0	0	1	-360	360;
	4262	1729	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5254	4093	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8745	8757	0.000271	0.001089	0	0	0	0	0	0	1	-360	360;
	6702	3933	0	0.00028	0	0	0	0	0	0	1	-360	360;
	308	1476	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1386	885	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3354	1222	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6873	4422	0	0.000222	0	0	0	0	0	0	1	-360	360;
	783	1931	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5950	3463	0.006729	0.031862	0	0	0	0	0	0	1	-360	360;
	5950	5560	0.006591	0.029498	0	0	0	0	0	0	1	-360	360;
	7433	8796	0.000258	0.000711	0	0	0	0	0	0	1	-360	360;
	5318	6479	0.020311	0.045649	0	0	0	0	0	0	1	-360	360;
	8231	7387	0.004102	0.014089	0	0	0	0	0	0	1	-360	360;
	7387	1724	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5984	2975	0.038777	0.100653	0	0	0	0	0	0	1	-360	360;
	3274	5984	0.026711	0.08357	0	0	0	0	0	0	1	-360	360;
	7784	2829	0.062719	0.144537	0	0	0	0	0	0	1	-360	360;
	5085	7784	0.045033	0.112364	0	0	0	0	0	0	1	-360	360;
	4490	7656	0.011659	0.070599	0	381	0	0	0	0	1	-360	360;
	4490	825	0.00968	0.04149	0	0	0	0	0	0	1	-360	360;
	4490	825	0.00862	0.047021	0	381	0	0	0	0	1	-360	360;
	1340	2829	0.077471	0.185529	0	0	0	0	0	0	1	-360	360;
	3726	2829	0.045157	0.126248	0	0	0	0	0	0	1	-360	360;
	3333	2829	0.01119	0.043099	0	0	0	0	0	0	1	-360	360;
	3333	2829	0.01095	0.043909	0	0	0	0	0	0	1	-360	360;
	2809	1340	0.03681	0.090496	0	0	0	0	0	0	1	-360	360;
	3170	6229	0.056174	0.129223	0	0	0	0	0	0	1	-360	360;
	3836	3170	0.029256	0.116322	0	0	0	0	0	0	1	-360	360;
	5507	8567	0.061347	0.097917	0	0	0	0	0	0	1	-360	360;
	8567	8902	0.088719	0.24762	0	0	0	0	0	0	1	-360	360;
	8084	269	0	0.000413	0	0	0	0	0	0	1	-360	360;
	8084	5675	0.004612	0.01738	0	0	0	0	0	0	1	-360	360;
	8084	5750	0.008364	0.02695	0	0	0	0	0	0	1	-360	360;
	8084	5750	0.014992	0.043636	0	0	0	0	0	0	1	-360	360;
	8321	269	0.003033	0.012851	0	0	0	0	0	0	1	-360	360;
	8321	269	0.002967	0.011901	0	0	0	0	0	0	1	-360	360;
	269	1527	0.015256	0.042686	0	0	0	0	0	0	1	-360	360;
	1928	269	0.004397	0.025653	0	0	0	0	0	0	1	-360	360;
	3371	3942	0.0027	0.012961	0	381	0	0	0	0	1	-360	360;
	825	3942	0.038169	0.16812	0	0	0	0	0	0	1	-360	360;
	3942	5501	0.015479	0.073409	0	514	0	0	0	0	1	-360	360;
	6598	3466	0.027868	0.087843	0	0	0	0	0	0	1	-360	360;
	5032	6598	0.013934	0.040479	0	0	0	0	0	0	1	-360	360;
	1781	3466	0.019231	0.076471	0	0	0	0	0	0	1	-360	360;
	2296	5786	0.013421	0.037256	0	0	0	0	0	0	1	-360	360;
	5545	2296	0.022438	0.046851	0	0	0	0	0	0	1	-360	360;
	8280	492	0.014562	0.045314	0	0	0	0	0	0	1	-360	360;
	492	2809	0.000826	0.002207	0	0	0	0	0	0	1	-360	360;
	1902	5721	0.040628	0.113562	0	0	0	0	0	0	1	-360	360;
	5504	1902	0.008058	0.024438	0	0	0	0	0	0	1	-360	360;
	1902	5871	0.022471	0.041083	0	0	0	0	0	0	1	-360	360;
	4382	8579	0.047322	0.096669	0	0	0	0	0	0	1	-360	360;
	8579	3274	0.01914	0.07614	0	0	0	0	0	0	1	-360	360;
	738	6059	0.007802	0.01924	0	0	0	0	0	0	1	-360	360;
	738	6059	0.005397	0.015554	0	0	0	0	0	0	1	-360	360;
	6059	7334	0.024719	0.040702	0	0	0	0	0	0	1	-360	360;
	6059	7334	0.02419	0.041471	0	0	0	0	0	0	1	-360	360;
	4246	2670	0.039603	0.109529	0	0	0	0	0	0	1	-360	360;
	3457	4246	0.013826	0.039826	0	0	0	0	0	0	1	-360	360;
	6856	4246	0.015174	0.04981	0	0	0	0	0	0	1	-360	360;
	8055	4591	0.003669	0.009008	0	0	0	0	0	0	1	-360	360;
	4591	2004	0.03276	0.106603	0	0	0	0	0	0	1	-360	360;
	4591	2004	0.03019	0.107612	0	0	0	0	0	0	1	-360	360;
	3677	4591	0.019033	0.051653	0	0	0	0	0	0	1	-360	360;
	187	4591	0.037603	0.124017	0	0	0	0	0	0	1	-360	360;
	4591	8965	0.034868	0.097471	0	0	0	0	0	0	1	-360	360;
	4591	6435	0.043099	0.130603	0	0	0	0	0	0	1	-360	360;
	4382	8055	0.06105	0.110339	0	0	0	0	0	0	1	-360	360;
	2819	6273	0.041446	0.108843	0	0	0	0	0	0	1	-360	360;
	6273	235	0.063471	0.123207	0	0	0	0	0	0	1	-360	360;
	3092	5504	0.02343	0.073603	0	0	0	0	0	0	1	-360	360;
	3092	1886	0.022537	0.091587	0	0	0	0	0	0	1	-360	360;
	9232	4612	0.036744	0.106934	0	0	0	0	0	0	1	-360	360;
	9232	826	0.006207	0.023959	0	0	0	0	0	0	1	-360	360;
	3457	3414	0.003157	0.011008	0	0	0	0	0	0	1	-360	360;
	3457	3414	0.003223	0.010793	0	0	0	0	0	0	1	-360	360;
	6856	3414	0.005	0.035653	0	0	0	0	0	0	1	-360	360;
	6856	3414	0.004901	0.036347	0	0	0	0	0	0	1	-360	360;
	8056	1574	0.007917	0.029603	0	0	0	0	0	0	1	-360	360;
	724	1574	0.004876	0.01476	0	0	0	0	0	0	1	-360	360;
	5470	4612	0.007182	0.023579	0	0	0	0	0	0	1	-360	360;
	5470	4583	0.007322	0.023099	0	0	0	0	0	0	1	-360	360;
	5871	8545	0.010942	0.040562	0	0	0	0	0	0	1	-360	360;
	8545	911	0.030603	0.103702	0	0	0	0	0	0	1	-360	360;
	5786	879	0.01176	0.041421	0	0	0	0	0	0	1	-360	360;
	5786	1419	0.011868	0.03581	0	0	0	0	0	0	1	-360	360;
	781	5786	0.01757	0.064579	0	0	0	0	0	0	1	-360	360;
	7111	7331	0.017099	0.064281	0	0	0	0	0	0	1	-360	360;
	796	3836	0.015207	0.057397	0	0	0	0	0	0	1	-360	360;
	1419	796	0.009198	0.036512	0	0	0	0	0	0	1	-360	360;
	796	6856	0.024281	0.239446	0	0	0	0	0	0	1	-360	360;
	9153	1313	0.052471	0.157347	0	0	0	0	0	0	1	-360	360;
	411	1313	0.009314	0.035909	0	0	0	0	0	0	1	-360	360;
	8056	1313	0.039314	0.122777	0	0	0	0	0	0	1	-360	360;
	8321	452	0.001223	0.003711	0	0	0	0	0	0	1	-360	360;
	825	7383	0.00518	0.023091	0	400	0	0	0	0	1	-360	360;
	825	7383	0.005281	0.022591	0	381	0	0	0	0	1	-360	360;
	2637	8045	0.018198	0.037149	0	0	0	0	0	0	1	-360	360;
	7916	5922	0.012893	0.040033	0	0	0	0	0	0	1	-360	360;
	3838	5922	0.007132	0.020331	0	0	0	0	0	0	1	-360	360;
	3208	1429	0.00479	0.05497	0	1185	0	0	0	0	1	-360	360;
	4382	2004	0.047322	0.139868	0	0	0	0	0	0	1	-360	360;
	3471	2004	0.005471	0.017471	0	0	0	0	0	0	1	-360	360;
	3471	2004	0.013826	0.040207	0	0	0	0	0	0	1	-360	360;
	2975	2004	0.010678	0.03814	0	0	0	0	0	0	1	-360	360;
	2975	2004	0.010603	0.034793	0	0	0	0	0	0	1	-360	360;
	4648	2004	0.014579	0.043331	0	0	0	0	0	0	1	-360	360;
	2004	8965	0.021413	0.079248	0	0	0	0	0	0	1	-360	360;
	2004	7072	0.031198	0.098298	0	0	0	0	0	0	1	-360	360;
	5229	3371	0.009651	0.045031	0	381	0	0	0	0	1	-360	360;
	5721	5518	0.050207	0.124066	0	0	0	0	0	0	1	-360	360;
	9202	4539	0.034314	0.070521	0	0	0	0	0	0	1	-360	360;
	6641	5545	0.007463	0.022612	0	0	0	0	0	0	1	-360	360;
	6641	5545	0.007603	0.022132	0	0	0	0	0	0	1	-360	360;
	5504	3719	0.01081	0.033603	0	0	0	0	0	0	1	-360	360;
	9153	5504	0.011223	0.036446	0	0	0	0	0	0	1	-360	360;
	5504	7916	0.014182	0.045669	0	0	0	0	0	0	1	-360	360;
	7916	5504	0.014463	0.044744	0	0	0	0	0	0	1	-360	360;
	187	3677	0.02124	0.069099	0	0	0	0	0	0	1	-360	360;
	8849	1628	0.008041	0.030504	0	0	0	0	0	0	1	-360	360;
	8849	2326	0.009107	0.029893	0	0	0	0	0	0	1	-360	360;
	8650	7665	0.024802	0.074537	0	0	0	0	0	0	1	-360	360;
	131	8650	0.038711	0.136876	0	0	0	0	0	0	1	-360	360;
	724	1886	0.069413	0.194223	0	0	0	0	0	0	1	-360	360;
	9153	3719	0.01276	0.035281	0	0	0	0	0	0	1	-360	360;
	4202	7497	0.008744	0.025802	0	0	0	0	0	0	1	-360	360;
	7497	2326	0.008909	0.025248	0	0	0	0	0	0	1	-360	360;
	3333	1628	0.009289	0.029281	0	0	0	0	0	0	1	-360	360;
	1671	3726	0.021421	0.081298	0	0	0	0	0	0	1	-360	360;
	1671	7111	0.044479	0.139058	0	0	0	0	0	0	1	-360	360;
	9044	7656	0.008479	0.049461	0	400	0	0	0	0	1	-360	360;
	825	7656	0.00394	0.02236	0	438	0	0	0	0	1	-360	360;
	738	5654	0.016174	0.044463	0	0	0	0	0	0	1	-360	360;
	738	5654	0.014413	0.05038	0	0	0	0	0	0	1	-360	360;
	771	738	0.020198	0.064678	0	0	0	0	0	0	1	-360	360;
	771	738	0.020587	0.063364	0	0	0	0	0	0	1	-360	360;
	771	738	0.031711	0.059223	0	0	0	0	0	0	1	-360	360;
	9202	3598	0.056678	0.178372	0	0	0	0	0	0	1	-360	360;
	3598	5085	0.019496	0.078083	0	0	0	0	0	0	1	-360	360;
	5654	7177	0.045198	0.111479	0	0	0	0	0	0	1	-360	360;
	5378	5545	0.015041	0.057983	0	0	0	0	0	0	1	-360	360;
	1041	5378	0.03519	0.105653	0	0	0	0	0	0	1	-360	360;
	131	6229	0.044	0.099339	0	0	0	0	0	0	1	-360	360;
	1527	8321	0.013636	0.039512	0	0	0	0	0	0	1	-360	360;
	1527	8321	0.012149	0.044777	0	0	0	0	0	0	1	-360	360;
	5545	9196	0.035281	0.122702	0	0	0	0	0	0	1	-360	360;
	3926	9196	0.031967	0.090248	0	0	0	0	0	0	1	-360	360;
	6856	9196	0.023983	0.066421	0	0	0	0	0	0	1	-360	360;
	9202	7385	0.032901	0.126777	0	0	0	0	0	0	1	-360	360;
	7385	8781	0.015372	0.059083	0	0	0	0	0	0	1	-360	360;
	4202	7111	0.012636	0.024496	0	0	0	0	0	0	1	-360	360;
	4202	7111	0.012893	0.024	0	0	0	0	0	0	1	-360	360;
	4202	7452	0.001917	0.006091	0	0	0	0	0	0	1	-360	360;
	4202	2326	0.018149	0.049413	0	0	0	0	0	0	1	-360	360;
	2178	861	0.01276	0.037083	0	0	0	0	0	0	1	-360	360;
	8902	2178	0.069636	0.146636	0	0	0	0	0	0	1	-360	360;
	2649	8974	0.00219	0.02213	0	1152	0	0	0	0	1	-360	360;
	2480	825	0.00194	0.0087	0	438	0	0	0	0	1	-360	360;
	2480	825	0.001979	0.00851	0	419	0	0	0	0	1	-360	360;
	3836	2670	0.028893	0.108496	0	0	0	0	0	0	1	-360	360;
	879	3836	0.021008	0.05814	0	0	0	0	0	0	1	-360	360;
	5654	3896	0.018661	0.056488	0	0	0	0	0	0	1	-360	360;
	8800	1041	0.01381	0.03776	0	0	0	0	0	0	1	-360	360;
	8800	3926	0.012298	0.042793	0	0	0	0	0	0	1	-360	360;
	8450	5871	0.009281	0.03038	0	0	0	0	0	0	1	-360	360;
	7665	2819	0.011438	0.035711	0	0	0	0	0	0	1	-360	360;
	8811	7665	0.05095	0.124661	0	0	0	0	0	0	1	-360	360;
	5518	9202	0.058207	0.110223	0	0	0	0	0	0	1	-360	360;
	411	724	0.061876	0.195174	0	0	0	0	0	0	1	-360	360;
	8140	3311	0.013438	0.032033	0	0	0	0	0	0	1	-360	360;
	4583	8140	0.03619	0.085066	0	0	0	0	0	0	1	-360	360;
	3896	2673	0.05324	0.154826	0	0	0	0	0	0	1	-360	360;
	3838	3311	0.004579	0.026017	0	0	0	0	0	0	1	-360	360;
	3198	9073	0.00795	0.03657	0	0	0	0	0	0	1	-360	360;
	2819	1266	0.055967	0.167983	0	0	0	0	0	0	1	-360	360;
	8811	2819	0.070942	0.168066	0	0	0	0	0	0	1	-360	360;
	2502	2649	0.00389	0.04149	0	1251	0	0	0	0	1	-360	360;
	7081	2502	0.003	0.02776	0	1119	0	0	0	0	1	-360	360;
	5161	5229	0.002159	0.01261	0	381	0	0	0	0	1	-360	360;
	5161	5229	0.002021	0.016849	0	400	0	0	0	0	1	-360	360;
	7008	5229	0.007151	0.046719	0	610	0	0	0	0	1	-360	360;
	5229	6924	0.000539	0.00368	0	0	0	0	0	0	1	-360	360;
	5229	6924	0.000539	0.003591	0	0	0	0	0	0	1	-360	360;
	7916	8450	0.007446	0.010579	0	0	0	0	0	0	1	-360	360;
	7916	2543	0.011992	0.03276	0	0	0	0	0	0	1	-360	360;
	771	2637	0.034207	0.085124	0	0	0	0	0	0	1	-360	360;
	181	771	0.001248	0.004471	0	0	0	0	0	0	1	-360	360;
	181	771	0.001223	0.004562	0	0	0	0	0	0	1	-360	360;
	5501	3198	0.006961	0.04039	0	381	0	0	0	0	1	-360	360;
	5501	3198	0.004521	0.02305	0	381	0	0	0	0	1	-360	360;
	5501	4417	0.00395	0.022789	0	457	0	0	0	0	1	-360	360;
	131	4373	0.031934	0.123661	0	0	0	0	0	0	1	-360	360;
	131	3457	0.019802	0.070901	0	0	0	0	0	0	1	-360	360;
	131	3457	0.022438	0.069479	0	0	0	0	0	0	1	-360	360;
	131	1366	0.014727	0.04305	0	0	0	0	0	0	1	-360	360;
	5750	131	0.01343	0.06824	0	0	0	0	0	0	1	-360	360;
	1928	131	0.022	0.060388	0	0	0	0	0	0	1	-360	360;
	1928	131	0.019603	0.068438	0	0	0	0	0	0	1	-360	360;
	1266	1218	0.048769	0.114711	0	0	0	0	0	0	1	-360	360;
	3471	4382	0.046901	0.148033	0	0	0	0	0	0	1	-360	360;
	6911	4382	0.055744	0.139066	0	0	0	0	0	0	1	-360	360;
	781	5545	0.010917	0.032901	0	0	0	0	0	0	1	-360	360;
	7334	5507	0.026322	0.044388	0	0	0	0	0	0	1	-360	360;
	5085	8280	0.047438	0.143727	0	0	0	0	0	0	1	-360	360;
	7603	8280	0.026876	0.09776	0	0	0	0	0	0	1	-360	360;
	8781	7603	0.01705	0.060802	0	0	0	0	0	0	1	-360	360;
	911	8781	0.028727	0.075289	0	0	0	0	0	0	1	-360	360;
	5085	5585	0.002388	0.015942	0	0	0	0	0	0	1	-360	360;
	5085	2809	0.088471	0.217198	0	0	0	0	0	0	1	-360	360;
	2673	6658	0.01714	0.056248	0	0	0	0	0	0	1	-360	360;
	6447	3471	0.004479	0.014008	0	0	0	0	0	0	1	-360	360;
	2975	6447	0.009851	0.032314	0	0	0	0	0	0	1	-360	360;
	4648	6447	0.001636	0.005587	0	0	0	0	0	0	1	-360	360;
	4648	3274	0.021603	0.084	0	0	0	0	0	0	1	-360	360;
	1781	4698	0.021727	0.065893	0	0	0	0	0	0	1	-360	360;
	1781	4583	0.025686	0.080256	0	0	0	0	0	0	1	-360	360;
	4583	1781	0.026182	0.078587	0	0	0	0	0	0	1	-360	360;
	7008	5161	0.00536	0.040919	0	629	0	0	0	0	1	-360	360;
	2543	3838	0.007388	0.026628	0	0	0	0	0	0	1	-360	360;
	1527	4408	0.005826	0.017488	0	0	0	0	0	0	1	-360	360;
	1527	4408	0.005826	0.016653	0	0	0	0	0	0	1	-360	360;
	3926	4408	0.026826	0.082273	0	0	0	0	0	0	1	-360	360;
	8920	1527	0.003256	0.010587	0	0	0	0	0	0	1	-360	360;
	8920	965	0.003331	0.010372	0	0	0	0	0	0	1	-360	360;
	8239	8056	0.025174	0.077661	0	0	0	0	0	0	1	-360	360;
	2699	4698	0.030248	0.085413	0	0	0	0	0	0	1	-360	360;
	5032	4698	0.029471	0.109198	0	0	0	0	0	0	1	-360	360;
	1218	187	0.058942	0.115529	0	0	0	0	0	0	1	-360	360;
	1366	4373	0.028413	0.083	0	0	0	0	0	0	1	-360	360;
	8811	4373	0.005826	0.016	0	0	0	0	0	0	1	-360	360;
	8555	4373	0.005512	0.016612	0	0	0	0	0	0	1	-360	360;
	5700	1527	0.014256	0.033909	0	0	0	0	0	0	1	-360	360;
	724	1527	0.052826	0.200702	0	0	0	0	0	0	1	-360	360;
	1928	1527	0.01976	0.061058	0	0	0	0	0	0	1	-360	360;
	1928	1527	0.014008	0.053207	0	0	0	0	0	0	1	-360	360;
	1928	1527	0.013711	0.054207	0	0	0	0	0	0	1	-360	360;
	5750	3457	0.009132	0.024843	0	0	0	0	0	0	1	-360	360;
	6856	3457	0.010711	0.040256	0	0	0	0	0	0	1	-360	360;
	6530	3457	0.004562	0.012421	0	0	0	0	0	0	1	-360	360;
	8003	2699	0.025273	0.079777	0	0	0	0	0	0	1	-360	360;
	965	5700	0.009149	0.028992	0	0	0	0	0	0	1	-360	360;
	3333	2326	0.027149	0.096802	0	0	0	0	0	0	1	-360	360;
	826	4583	0.063033	0.157099	0	0	0	0	0	0	1	-360	360;
	826	8003	0.009702	0.035397	0	0	0	0	0	0	1	-360	360;
	826	8003	0.009504	0.036091	0	0	0	0	0	0	1	-360	360;
	235	6911	0.029893	0.105364	0	0	0	0	0	0	1	-360	360;
	5750	5675	0.005397	0.020298	0	0	0	0	0	0	1	-360	360;
	7452	2875	0.001124	0.002942	0	0	0	0	0	0	1	-360	360;
	181	2875	0.001207	0.003174	0	0	0	0	0	0	1	-360	360;
	181	2875	0.00124	0.003099	0	0	0	0	0	0	1	-360	360;
	6530	5750	0.004066	0.014083	0	0	0	0	0	0	1	-360	360;
	2389	7008	0.005099	0.025479	0	381	0	0	0	0	1	-360	360;
	2389	7008	0.0052	0.024961	0	0	0	0	0	0	1	-360	360;
	8555	8811	0.003868	0.013901	0	0	0	0	0	0	1	-360	360;
	7081	1429	0.00126	0.01433	0	1251	0	0	0	0	1	-360	360;
	8239	1928	0.011752	0.044314	0	0	0	0	0	0	1	-360	360;
	8965	7072	0.021719	0.046248	0	0	0	0	0	0	1	-360	360;
	8965	6435	0.05324	0.100281	0	0	0	0	0	0	1	-360	360;
	8416	6743	0.00109	0.01155	0	1086	0	0	0	0	1	-360	360;
	8416	6743	0.0011	0.01158	0	1020	0	0	0	0	1	-360	360;
	8416	8715	0.00096	0.01293	0	1218	0	0	0	0	1	-360	360;
	8416	8715	0.00098	0.01264	0	1185	0	0	0	0	1	-360	360;
	8416	569	0.00101	0.01059	0	1382	0	0	0	0	1	-360	360;
	8052	6204	0.00112	0.010969	0	533	0	0	0	0	1	-360	360;
	290	6204	0.00126	0.01075	0	533	0	0	0	0	1	-360	360;
	3684	8580	0.00197	0.0207	0	954	0	0	0	0	1	-360	360;
	3684	569	0.00303	0.0273	0	1316	0	0	0	0	1	-360	360;
	340	8065	0.00194	0.02339	0	889	0	0	0	0	1	-360	360;
	8715	340	0.00057	0.00492	0	987	0	0	0	0	1	-360	360;
	6685	2858	0.01105	0.056599	0	0	0	0	0	0	1	-360	360;
	5199	2858	0.00482	0.046521	0	381	0	0	0	0	1	-360	360;
	6743	8526	0.00194	0.01645	0	954	0	0	0	0	1	-360	360;
	8526	5850	0.00134	0.01459	0	1020	0	0	0	0	1	-360	360;
	5284	8531	0.000401	0.00318	0	610	0	0	0	0	1	-360	360;
	5284	7065	0.000349	0.00293	0	572	0	0	0	0	1	-360	360;
	5284	1402	0.000401	0.00324	0	610	0	0	0	0	1	-360	360;
	5284	2858	0.01214	0.058401	0	0	0	0	0	0	1	-360	360;
	5284	2858	0.012979	0.053281	0	0	0	0	0	0	1	-360	360;
	5284	3245	0.007919	0.053729	0	533	0	0	0	0	1	-360	360;
	5284	6799	0.014771	0.06813	0	438	0	0	0	0	1	-360	360;
	5284	6799	0.013789	0.067711	0	438	0	0	0	0	1	-360	360;
	1648	6204	0.005349	0.044469	0	533	0	0	0	0	1	-360	360;
	9004	6204	0.005229	0.04118	0	533	0	0	0	0	1	-360	360;
	6204	4912	0.0008	0.00676	0	553	0	0	0	0	1	-360	360;
	6204	2144	0.000781	0.00626	0	0	0	0	0	0	1	-360	360;
	6204	7867	0.00363	0.017409	0	0	0	0	0	0	1	-360	360;
	6204	7867	0.0037	0.01705	0	0	0	0	0	0	1	-360	360;
	6204	2395	0.00449	0.046469	0	495	0	0	0	0	1	-360	360;
	4909	3604	0.00052	0.00727	0	757	0	0	0	0	1	-360	360;
	4909	5850	0.00047	0.0062	0	757	0	0	0	0	1	-360	360;
	4909	5850	0.00046	0.00631	0	757	0	0	0	0	1	-360	360;
	4998	6799	0.003531	0.013669	0	495	0	0	0	0	1	-360	360;
	7458	8065	0.00032	0.00322	0	1119	0	0	0	0	1	-360	360;
	7458	4817	0.00156	0.01603	0	987	0	0	0	0	1	-360	360;
	7458	2368	0.0022	0.02389	0	1415	0	0	0	0	1	-360	360;
	7679	2395	0.00337	0.032669	0	400	0	0	0	0	1	-360	360;
	9183	4817	0.00199	0.0195	0	1152	0	0	0	0	1	-360	360;
	9183	2368	0.00144	0.01354	0	1020	0	0	0	0	1	-360	360;
	9183	2368	0.00147	0.01325	0	1119	0	0	0	0	1	-360	360;
	9183	5062	0.00211	0.02433	0	1152	0	0	0	0	1	-360	360;
	4150	2858	0.013901	0.088849	0	0	0	0	0	0	1	-360	360;
	4817	8715	0.00149	0.01564	0	1053	0	0	0	0	1	-360	360;
	4817	7979	0.00294	0.02278	0	1349	0	0	0	0	1	-360	360;
	8715	7979	0.00133	0.01873	0	1152	0	0	0	0	1	-360	360;
	7867	6799	0.00581	0.034229	0	572	0	0	0	0	1	-360	360;
	7979	6790	0.00145	0.02024	0	1613	0	0	0	0	1	-360	360;
	7979	6790	0.00142	0.01874	0	1744	0	0	0	0	1	-360	360;
	7979	8580	0.00245	0.02367	0	1152	0	0	0	0	1	-360	360;
	7979	5062	0.00178	0.02175	0	1119	0	0	0	0	1	-360	360;
	5624	7152	0.000701	0.009299	0	0	0	0	0	0	1	-360	360;
	5624	7152	0.001042	0.007833	0	0	0	0	0	0	1	-360	360;
	5624	7152	0.001403	0.00534	0	0	0	0	0	0	1	-360	360;
	5850	3130	0.00022	0.00258	0	823	0	0	0	0	1	-360	360;
	2858	3245	0.02024	0.099349	0	0	0	0	0	0	1	-360	360;
	2858	6103	0.00044	0.003081	0	553	0	0	0	0	1	-360	360;
	2858	2310	0.00043	0.00314	0	553	0	0	0	0	1	-360	360;
	2858	8726	0.00461	0.027531	0	0	0	0	0	0	1	-360	360;
	8943	3245	0.01212	0.09901	0	381	0	0	0	0	1	-360	360;
	364	6528	0.000289	0.00261	0	381	0	0	0	0	1	-360	360;
	4572	364	0.00043	0.002831	0	438	0	0	0	0	1	-360	360;
	3507	364	0.000289	0.000599	0	419	0	0	0	0	1	-360	360;
	7489	1140	0.000289	0.00224	0	533	0	0	0	0	1	-360	360;
	4133	7489	0.000219	0.001591	0	400	0	0	0	0	1	-360	360;
	2694	6725	0.00045	0.00786	0	1810	0	0	0	0	1	-360	360;
	2694	6725	0.00044	0.00728	0	1645	0	0	0	0	1	-360	360;
	6725	8252	0.00049	0.00466	0	1514	0	0	0	0	1	-360	360;
	6725	8117	0.00125	0.01973	0	0	0	0	0	0	1	-360	360;
	6810	6809	0.001461	0.00714	0	419	0	0	0	0	1	-360	360;
	9113	6810	0.00037	0.00094	0	419	0	0	0	0	1	-360	360;
	6810	1804	0.001039	0.00737	0	0	0	0	0	0	1	-360	360;
	5303	6327	8.1e-05	0.00061	0	381	0	0	0	0	1	-360	360;
	7311	8069	0.00053	0.00852	0	1613	0	0	0	0	1	-360	360;
	5476	8069	0.00053	0.00704	0	1744	0	0	0	0	1	-360	360;
	5109	3360	0.0001	0.00128	0	823	0	0	0	0	1	-360	360;
	4348	9075	0.0015	0.00705	0	419	0	0	0	0	1	-360	360;
	9075	7066	0.009771	0.05363	0	400	0	0	0	0	1	-360	360;
	4348	9075	0.00088	0.007771	0	438	0	0	0	0	1	-360	360;
	4148	8443	0.00539	0.032979	0	400	0	0	0	0	1	-360	360;
	1470	4148	0.00482	0.03086	0	0	0	0	0	0	1	-360	360;
	2139	4148	0.010539	0.065479	0	457	0	0	0	0	1	-360	360;
	4148	7893	0.01074	0.06405	0	476	0	0	0	0	1	-360	360;
	4448	4148	0.0063	0.048909	0	533	0	0	0	0	1	-360	360;
	1819	899	0.0002	0.00022	0	1415	0	0	0	0	1	-360	360;
	4979	899	0.0002	0.00022	0	1448	0	0	0	0	1	-360	360;
	6822	899	0.00019	0.00023	0	1481	0	0	0	0	1	-360	360;
	924	1279	0.000289	0.0017	0	0	0	0	0	0	1	-360	360;
	899	1193	0.00145	0.01937	0	0	0	0	0	0	1	-360	360;
	3123	899	0.00048	0.00814	0	1777	0	0	0	0	1	-360	360;
	3123	899	0.00057	0.00758	0	1613	0	0	0	0	1	-360	360;
	8383	3807	0.0002	0.00187	0	0	0	0	0	0	1	-360	360;
	8383	3807	0.000169	0.00193	0	0	0	0	0	0	1	-360	360;
	2121	1977	0.00216	0.03591	0	987	0	0	0	0	1	-360	360;
	2121	3351	0.00092	0.01287	0	1810	0	0	0	0	1	-360	360;
	1261	3605	0.02412	0.10443	0	476	0	0	0	0	1	-360	360;
	1279	8443	0.00063	0.00401	0	0	0	0	0	0	1	-360	360;
	1279	3571	0.00038	0.002409	0	0	0	0	0	0	1	-360	360;
	4873	4094	0.010091	0.05524	0	514	0	0	0	0	1	-360	360;
	5493	4873	0.00668	0.04457	0	438	0	0	0	0	1	-360	360;
	7016	1882	0.00057	0.004961	0	419	0	0	0	0	1	-360	360;
	1882	5483	0.00025	0.002159	0	0	0	0	0	0	1	-360	360;
	526	4516	0.00443	0.01807	0	438	0	0	0	0	1	-360	360;
	526	5839	0.00037	0.001659	0	495	0	0	0	0	1	-360	360;
	7016	834	0.00037	0.003599	0	476	0	0	0	0	1	-360	360;
	7016	7494	0.000659	0.00836	0	438	0	0	0	0	1	-360	360;
	7016	7494	0.00151	0.012289	0	381	0	0	0	0	1	-360	360;
	7016	2388	0.000539	0.003711	0	381	0	0	0	0	1	-360	360;
	8399	1328	0.00036	0.00606	0	1086	0	0	0	0	1	-360	360;
	8399	1328	0.00035	0.00618	0	921	0	0	0	0	1	-360	360;
	1328	660	0.00101	0.01675	0	0	0	0	0	0	1	-360	360;
	1328	5355	4e-05	0.00071	0	1613	0	0	0	0	1	-360	360;
	1328	5355	4e-05	0.00081	0	1777	0	0	0	0	1	-360	360;
	1328	984	0.00013	0.00152	0	0	0	0	0	0	1	-360	360;
	1328	984	0.00014	0.00149	0	987	0	0	0	0	1	-360	360;
	1328	4178	0.00013	0.00164	0	757	0	0	0	0	1	-360	360;
	1328	4178	0.00013	0.00168	0	889	0	0	0	0	1	-360	360;
	6016	4448	0.000669	0.00511	0	476	0	0	0	0	1	-360	360;
	8962	3465	3.9e-05	0.000341	0	0	0	0	0	0	1	-360	360;
	5885	5677	0.004219	0.02145	0	0	0	0	0	0	1	-360	360;
	5885	4035	0.000229	0.002849	0	629	0	0	0	0	1	-360	360;
	1261	54	0.0137	0.10436	0	457	0	0	0	0	1	-360	360;
	2675	1261	0.001521	0.01449	0	514	0	0	0	0	1	-360	360;
	765	437	8.1e-05	0.00057	0	533	0	0	0	0	1	-360	360;
	6809	437	0.001461	0.00813	0	476	0	0	0	0	1	-360	360;
	1423	437	0.00049	0.00381	0	457	0	0	0	0	1	-360	360;
	7540	2578	0.000401	0.00287	0	533	0	0	0	0	1	-360	360;
	5705	4820	8.3e-05	0.000347	0	0	0	0	0	0	1	-360	360;
	6560	7044	0.00023	0.00285	0	1316	0	0	0	0	1	-360	360;
	7044	3196	0.00016	0.00218	0	0	0	0	0	0	1	-360	360;
	2986	7044	0.00057	0.00736	0	1613	0	0	0	0	1	-360	360;
	402	7855	0.000591	0.005789	0	495	0	0	0	0	1	-360	360;
	402	3337	0.002729	0.011461	0	495	0	0	0	0	1	-360	360;
	402	6107	0.00051	0.00205	0	629	0	0	0	0	1	-360	360;
	8273	6750	0.00031	0.00374	0	1876	0	0	0	0	1	-360	360;
	2710	8273	0.00032	0.00518	0	0	0	0	0	0	1	-360	360;
	1990	7306	0.00136	0.01038	0	476	0	0	0	0	1	-360	360;
	8532	4037	0.00151	0.02083	0	1810	0	0	0	0	1	-360	360;
	7664	4037	0.00015	0.00232	0	1678	0	0	0	0	1	-360	360;
	2204	8120	1e-05	0.00025	0	987	0	0	0	0	1	-360	360;
	9006	8120	1e-05	0.00023	0	987	0	0	0	0	1	-360	360;
	2423	8120	0.00047	0.00683	0	1185	0	0	0	0	1	-360	360;
	882	8120	0.00067	0.00858	0	1283	0	0	0	0	1	-360	360;
	2540	1034	0.003669	0.025331	0	514	0	0	0	0	1	-360	360;
	368	2540	0.00525	0.02994	0	0	0	0	0	0	1	-360	360;
	4672	6863	0.00595	0.03625	0	495	0	0	0	0	1	-360	360;
	4672	7841	0.004969	0.03411	0	476	0	0	0	0	1	-360	360;
	565	1859	0.005669	0.04293	0	553	0	0	0	0	1	-360	360;
	565	1859	0.00555	0.043771	0	514	0	0	0	0	1	-360	360;
	1859	3832	0.002831	0.022401	0	0	0	0	0	0	1	-360	360;
	1859	3832	0.00325	0.01938	0	400	0	0	0	0	1	-360	360;
	8049	162	0.00068	0.00082	0	1020	0	0	0	0	1	-360	360;
	8049	2039	0.00054	0.00868	0	1349	0	0	0	0	1	-360	360;
	1759	6793	0.00061	0.002909	0	0	0	0	0	0	1	-360	360;
	1759	1469	0.002271	0.01014	0	0	0	0	0	0	1	-360	360;
	3895	1264	0.00083	0.00986	0	0	0	0	0	0	1	-360	360;
	1264	8139	0.00031	0.00511	0	1613	0	0	0	0	1	-360	360;
	1264	8139	0.0003	0.00521	0	1645	0	0	0	0	1	-360	360;
	8608	1264	0.00081	0.01288	0	0	0	0	0	0	1	-360	360;
	1264	9009	0.00039	0.00645	0	0	0	0	0	0	1	-360	360;
	3769	2320	0.00243	0.012211	0	0	0	0	0	0	1	-360	360;
	3563	5302	0.00018	0.00226	0	0	0	0	0	0	1	-360	360;
	2316	5302	0.0002	0.00347	0	1744	0	0	0	0	1	-360	360;
	3196	5302	0.00018	0.0029	0	0	0	0	0	0	1	-360	360;
	5498	6750	0.00045	0.00494	0	1382	0	0	0	0	1	-360	360;
	6750	2684	0.00042	0.00497	0	1613	0	0	0	0	1	-360	360;
	3229	8532	0.00019	0.00234	0	1251	0	0	0	0	1	-360	360;
	1977	8532	8e-05	0.00105	0	1678	0	0	0	0	1	-360	360;
	1977	8532	8e-05	0.00107	0	1613	0	0	0	0	1	-360	360;
	8532	7922	8e-05	0.001	0	0	0	0	0	0	1	-360	360;
	8532	3972	0.00063	0.01015	0	1843	0	0	0	0	1	-360	360;
	880	6671	0.00064	0.005651	0	0	0	0	0	0	1	-360	360;
	880	6671	0.000581	0.006531	0	0	0	0	0	0	1	-360	360;
	5114	8723	0.007099	0.04432	0	400	0	0	0	0	1	-360	360;
	6528	1804	0.00144	0.01157	0	438	0	0	0	0	1	-360	360;
	6528	7804	0.00114	0.00695	0	419	0	0	0	0	1	-360	360;
	3563	1684	0.00044	0.0053	0	1613	0	0	0	0	1	-360	360;
	7071	6622	0.00162	0.009039	0	400	0	0	0	0	1	-360	360;
	7071	6622	0.00187	0.0072	0	400	0	0	0	0	1	-360	360;
	1327	6622	0.00101	0.00564	0	0	0	0	0	0	1	-360	360;
	3705	6622	0.0007	0.00445	0	0	0	0	0	0	1	-360	360;
	2909	6622	0.001081	0.00619	0	553	0	0	0	0	1	-360	360;
	1977	8304	0.00022	0.00228	0	1448	0	0	0	0	1	-360	360;
	1977	6207	0.00023	0.00223	0	1415	0	0	0	0	1	-360	360;
	1977	7940	0.0002	0.00262	0	1448	0	0	0	0	1	-360	360;
	1977	1919	0.00021	0.00261	0	1283	0	0	0	0	1	-360	360;
	49	6992	0.00663	0.04181	0	457	0	0	0	0	1	-360	360;
	1977	2456	0.00059	0.00921	0	1678	0	0	0	0	1	-360	360;
	1977	4159	0.0014	0.02227	0	1777	0	0	0	0	1	-360	360;
	3871	2821	0.000401	0.002469	0	438	0	0	0	0	1	-360	360;
	5645	2821	0.00212	0.01218	0	381	0	0	0	0	1	-360	360;
	2821	2025	0.004169	0.02439	0	438	0	0	0	0	1	-360	360;
	8443	9068	0.02336	0.122419	0	0	0	0	0	0	1	-360	360;
	8443	684	0.002729	0.0137	0	0	0	0	0	0	1	-360	360;
	2880	8443	0.000919	0.003831	0	0	0	0	0	0	1	-360	360;
	2880	8443	0.00082	0.004349	0	0	0	0	0	0	1	-360	360;
	4516	8443	0.01325	0.055159	0	438	0	0	0	0	1	-360	360;
	4516	8443	0.0108	0.06074	0	438	0	0	0	0	1	-360	360;
	8443	4813	0.001099	0.00631	0	419	0	0	0	0	1	-360	360;
	7787	6327	0.00495	0.029151	0	419	0	0	0	0	1	-360	360;
	6327	8933	0.003909	0.0227	0	400	0	0	0	0	1	-360	360;
	2139	3556	0.00061	0.00424	0	0	0	0	0	0	1	-360	360;
	2871	1140	0.00068	0.003919	0	514	0	0	0	0	1	-360	360;
	4574	1140	0.00368	0.019901	0	0	0	0	0	0	1	-360	360;
	7420	2316	0.00025	0.00388	0	1448	0	0	0	0	1	-360	360;
	7420	376	0.00038	0.00623	0	1218	0	0	0	0	1	-360	360;
	6809	4770	0.000909	0.006031	0	438	0	0	0	0	1	-360	360;
	4770	1804	0.001211	0.0062	0	514	0	0	0	0	1	-360	360;
	2423	405	0.00022	0.00322	0	1119	0	0	0	0	1	-360	360;
	2423	4512	0.00018	0.00307	0	1777	0	0	0	0	1	-360	360;
	2423	4928	0.00068	0.00873	0	1119	0	0	0	0	1	-360	360;
	2423	8161	0.00124	0.02016	0	1316	0	0	0	0	1	-360	360;
	9139	7311	0.00025	0.00332	0	1547	0	0	0	0	1	-360	360;
	9139	7311	0.00025	0.00335	0	1744	0	0	0	0	1	-360	360;
	9139	7311	0.00025	0.00325	0	1711	0	0	0	0	1	-360	360;
	3068	9139	0.00076	0.01075	0	1777	0	0	0	0	1	-360	360;
	4865	9139	3e-05	0.00044	0	757	0	0	0	0	1	-360	360;
	3465	7538	0.001651	0.010849	0	533	0	0	0	0	1	-360	360;
	8320	3465	0.00288	0.018469	0	476	0	0	0	0	1	-360	360;
	6895	1512	0.00072	0.00923	0	1349	0	0	0	0	1	-360	360;
	6895	6189	0.00069	0.00855	0	1744	0	0	0	0	1	-360	360;
	3180	4576	0.00124	0.01667	0	0	0	0	0	0	1	-360	360;
	3180	830	0.00034	0.00571	0	1810	0	0	0	0	1	-360	360;
	6939	3180	0.0004	0.00638	0	1909	0	0	0	0	1	-360	360;
	6939	3180	0.00039	0.0065	0	1876	0	0	0	0	1	-360	360;
	7891	3180	0.00027	0.00451	0	1678	0	0	0	0	1	-360	360;
	8215	3892	0.012409	0.074961	0	381	0	0	0	0	1	-360	360;
	8215	8600	0.002419	0.01761	0	438	0	0	0	0	1	-360	360;
	8215	1423	0.006901	0.053409	0	457	0	0	0	0	1	-360	360;
	2871	13	0.000599	0.00356	0	495	0	0	0	0	1	-360	360;
	2871	4516	0.011841	0.049341	0	419	0	0	0	0	1	-360	360;
	4572	904	0.000521	0.004831	0	533	0	0	0	0	1	-360	360;
	88	904	0.00055	0.00143	0	381	0	0	0	0	1	-360	360;
	7718	904	0.0005	0.00364	0	0	0	0	0	0	1	-360	360;
	7718	904	0.00056	0.003469	0	0	0	0	0	0	1	-360	360;
	904	5181	0.00082	0.00257	0	457	0	0	0	0	1	-360	360;
	4526	5451	0.00249	0.01536	0	457	0	0	0	0	1	-360	360;
	4526	2484	0.00113	0.00651	0	438	0	0	0	0	1	-360	360;
	8892	5078	0.00643	0.045031	0	514	0	0	0	0	1	-360	360;
	8509	8892	0.003581	0.02706	0	514	0	0	0	0	1	-360	360;
	1814	3351	0.00046	0.00774	0	1810	0	0	0	0	1	-360	360;
	1814	6464	0.00031	0.00483	0	1086	0	0	0	0	1	-360	360;
	5945	3767	0.001099	0.00345	0	457	0	0	0	0	1	-360	360;
	1257	985	0.00913	0.053211	0	0	0	0	0	0	1	-360	360;
	985	8524	0.021151	0.095401	0	0	0	0	0	0	1	-360	360;
	6161	4576	0.00034	0.0055	0	1810	0	0	0	0	1	-360	360;
	2756	3546	0.00064	0.01945	0	1810	0	0	0	0	1	-360	360;
	2756	8663	0.0004	0.01026	0	1448	0	0	0	0	1	-360	360;
	7268	2756	0.00071	0.01087	0	1251	0	0	0	0	1	-360	360;
	8786	1468	0.00968	0.088169	0	533	0	0	0	0	1	-360	360;
	8786	1468	0.012349	0.086409	0	457	0	0	0	0	1	-360	360;
	1468	4089	0.00288	0.017211	0	533	0	0	0	0	1	-360	360;
	4089	1468	0.00257	0.0195	0	495	0	0	0	0	1	-360	360;
	3078	1034	0.00051	0.00407	0	553	0	0	0	0	1	-360	360;
	580	6529	0.00019	0.00309	0	1514	0	0	0	0	1	-360	360;
	580	7108	0.00023	0.00357	0	1613	0	0	0	0	1	-360	360;
	3285	3884	0.00088	0.0092	0	1185	0	0	0	0	1	-360	360;
	3285	4680	0.00023	0.00179	0	1580	0	0	0	0	1	-360	360;
	6070	939	0.00145	0.007919	0	495	0	0	0	0	1	-360	360;
	471	939	0.00143	0.00812	0	495	0	0	0	0	1	-360	360;
	939	5677	0.00524	0.02181	0	514	0	0	0	0	1	-360	360;
	2040	3769	0.00261	0.01913	0	438	0	0	0	0	1	-360	360;
	4516	3769	0.00164	0.016591	0	572	0	0	0	0	1	-360	360;
	5532	5253	0.00087	0.012719	0	438	0	0	0	0	1	-360	360;
	1	7639	0.00775	0.03501	0	400	0	0	0	0	1	-360	360;
	1	7835	0.01944	0.093479	0	457	0	0	0	0	1	-360	360;
	1	7835	0.01981	0.091539	0	476	0	0	0	0	1	-360	360;
	1523	5633	0.003771	0.020599	0	419	0	0	0	0	1	-360	360;
	8681	1523	0.001271	0.00811	0	495	0	0	0	0	1	-360	360;
	8724	1072	0.00048	0.01459	0	1152	0	0	0	0	1	-360	360;
	3546	8724	0.00047	0.01351	0	1711	0	0	0	0	1	-360	360;
	2366	1470	0.021729	0.088281	0	0	0	0	0	0	1	-360	360;
	1470	2880	0.00038	0.001771	0	495	0	0	0	0	1	-360	360;
	1470	1257	0.022901	0.101401	0	476	0	0	0	0	1	-360	360;
	5	5459	0.002909	0.01432	0	476	0	0	0	0	1	-360	360;
	5	2675	0.015729	0.08824	0	419	0	0	0	0	1	-360	360;
	2905	6958	0.00038	0.002169	0	0	0	0	0	0	1	-360	360;
	1804	2905	0.000479	0.00212	0	0	0	0	0	0	1	-360	360;
	2905	399	0.00019	0.00086	0	0	0	0	0	0	1	-360	360;
	763	142	0.00089	0.004901	0	0	0	0	0	0	1	-360	360;
	3874	1072	0.00109	0.03151	0	1810	0	0	0	0	1	-360	360;
	565	6210	0.003409	0.020271	0	533	0	0	0	0	1	-360	360;
	565	6210	0.002909	0.02388	0	476	0	0	0	0	1	-360	360;
	6210	3832	0.00545	0.042969	0	495	0	0	0	0	1	-360	360;
	6210	3832	0.005341	0.0438	0	476	0	0	0	0	1	-360	360;
	5483	7494	0.001521	0.00863	0	0	0	0	0	0	1	-360	360;
	3195	5483	0.00057	0.005979	0	381	0	0	0	0	1	-360	360;
	9121	5747	0.00085	0.01449	0	1744	0	0	0	0	1	-360	360;
	882	5747	0.0006	0.00777	0	1547	0	0	0	0	1	-360	360;
	4928	5747	0.0003	0.00492	0	1119	0	0	0	0	1	-360	360;
	2685	2636	0.003349	0.02686	0	629	0	0	0	0	1	-360	360;
	2224	2636	0.00213	0.016849	0	591	0	0	0	0	1	-360	360;
	3811	9236	0.00077	0.0106	0	1415	0	0	0	0	1	-360	360;
	8237	9236	0.00155	0.0148	0	1316	0	0	0	0	1	-360	360;
	1613	565	0.016651	0.10632	0	495	0	0	0	0	1	-360	360;
	1613	565	0.014841	0.12049	0	457	0	0	0	0	1	-360	360;
	4666	565	9.9e-05	0.001229	0	533	0	0	0	0	1	-360	360;
	7743	565	9.9e-05	0.00114	0	533	0	0	0	0	1	-360	360;
	8266	2199	0.00048	0.00781	0	1810	0	0	0	0	1	-360	360;
	8266	3343	0.00076	0.01058	0	0	0	0	0	0	1	-360	360;
	8588	8266	0.00073	0.01181	0	954	0	0	0	0	1	-360	360;
	7234	830	0.00059	0.00929	0	1711	0	0	0	0	1	-360	360;
	149	830	8e-05	0.00092	0	1119	0	0	0	0	1	-360	360;
	295	1034	0.00019	0.000849	0	0	0	0	0	0	1	-360	360;
	2384	3948	0.003479	0.01538	0	0	0	0	0	0	1	-360	360;
	2384	3948	0.003539	0.01505	0	0	0	0	0	0	1	-360	360;
	2384	1469	0.01387	0.084099	0	381	0	0	0	0	1	-360	360;
	1092	2384	0.00375	0.021341	0	381	0	0	0	0	1	-360	360;
	8375	5376	0.00023	0.00345	0	1777	0	0	0	0	1	-360	360;
	5376	1684	0.00054	0.00662	0	1613	0	0	0	0	1	-360	360;
	405	8573	7e-05	0.00099	0	1218	0	0	0	0	1	-360	360;
	5510	405	0.00012	0.002	0	1744	0	0	0	0	1	-360	360;
	229	3884	0.00075	0.00888	0	1185	0	0	0	0	1	-360	360;
	3099	3884	0.00054	0.00705	0	1349	0	0	0	0	1	-360	360;
	6863	4138	0.00574	0.0287	0	495	0	0	0	0	1	-360	360;
	6863	6594	0.001331	0.009479	0	0	0	0	0	0	1	-360	360;
	6594	6863	0.00539	0.024159	0	400	0	0	0	0	1	-360	360;
	1032	6863	0.000539	0.0028	0	438	0	0	0	0	1	-360	360;
	7718	8014	0.000969	0.00724	0	381	0	0	0	0	1	-360	360;
	6649	8014	0.000349	0.00339	0	457	0	0	0	0	1	-360	360;
	2684	8375	0.00021	0.00349	0	1942	0	0	0	0	1	-360	360;
	1327	49	0.00318	0.020669	0	400	0	0	0	0	1	-360	360;
	1327	49	0.003599	0.02026	0	419	0	0	0	0	1	-360	360;
	49	1236	0.0053	0.035521	0	381	0	0	0	0	1	-360	360;
	49	2803	0.00486	0.024331	0	419	0	0	0	0	1	-360	360;
	6809	4572	0.000659	0.00645	0	438	0	0	0	0	1	-360	360;
	5459	3571	0.00239	0.01531	0	0	0	0	0	0	1	-360	360;
	5868	1571	0.00119	0.01215	0	1678	0	0	0	0	1	-360	360;
	1571	6201	0.00012	0.0015	0	1909	0	0	0	0	1	-360	360;
	1571	6460	0.00047	0.00738	0	1415	0	0	0	0	1	-360	360;
	7855	5959	0.00068	0.00189	0	514	0	0	0	0	1	-360	360;
	4995	8951	0.003669	0.022099	0	476	0	0	0	0	1	-360	360;
	8681	8951	0.004771	0.0218	0	457	0	0	0	0	1	-360	360;
	2909	7639	0.001469	0.006591	0	533	0	0	0	0	1	-360	360;
	3347	510	8.1e-05	0.00045	0	0	0	0	0	0	1	-360	360;
	510	471	0.001021	0.00782	0	476	0	0	0	0	1	-360	360;
	4035	510	0.0015	0.009591	0	514	0	0	0	0	1	-360	360;
	6156	834	0.00019	0.001581	0	514	0	0	0	0	1	-360	360;
	6156	7718	3.9e-05	0.000479	0	610	0	0	0	0	1	-360	360;
	7311	5471	0.00027	0.00281	0	1283	0	0	0	0	1	-360	360;
	5476	7311	0.00096	0.01609	0	1777	0	0	0	0	1	-360	360;
	7311	8588	0.00027	0.00339	0	954	0	0	0	0	1	-360	360;
	7311	8588	0.00023	0.00398	0	1547	0	0	0	0	1	-360	360;
	1320	7538	0.000901	0.00937	0	572	0	0	0	0	1	-360	360;
	7540	7538	0.001901	0.01549	0	438	0	0	0	0	1	-360	360;
	8320	7538	0.003591	0.02205	0	476	0	0	0	0	1	-360	360;
	7538	5704	0.0013	0.009901	0	514	0	0	0	0	1	-360	360;
	2669	3068	0.00022	0.00299	0	0	0	0	0	0	1	-360	360;
	3068	8635	0.00047	0.00564	0	1843	0	0	0	0	1	-360	360;
	6	3068	0.00046	0.0075	0	0	0	0	0	0	1	-360	360;
	6303	708	0.00214	0.020979	0	438	0	0	0	0	1	-360	360;
	2880	708	0.004651	0.03332	0	457	0	0	0	0	1	-360	360;
	708	4448	0.00468	0.021281	0	495	0	0	0	0	1	-360	360;
	708	4813	0.00581	0.028331	0	0	0	0	0	0	1	-360	360;
	1760	7197	7e-05	0.00086	0	1613	0	0	0	0	1	-360	360;
	2634	4445	0.0002	0.00279	0	1349	0	0	0	0	1	-360	360;
	2634	4445	0.00017	0.00333	0	1349	0	0	0	0	1	-360	360;
	5692	5868	1e-05	0.00036	0	1448	0	0	0	0	1	-360	360;
	5181	6631	0.000401	0.004081	0	457	0	0	0	0	1	-360	360;
	6631	7167	0.001581	0.015719	0	476	0	0	0	0	1	-360	360;
	7727	7841	0.00012	0.000229	0	0	0	0	0	0	1	-360	360;
	7727	7841	0.00012	0.00024	0	0	0	0	0	0	1	-360	360;
	7841	8521	0.00132	0.005729	0	438	0	0	0	0	1	-360	360;
	8774	7841	0.00438	0.024901	0	419	0	0	0	0	1	-360	360;
	5633	104	0.005521	0.023711	0	419	0	0	0	0	1	-360	360;
	229	7359	0.0008	0.01014	0	1283	0	0	0	0	1	-360	360;
	229	7359	0.00068	0.01229	0	1349	0	0	0	0	1	-360	360;
	2685	1220	0.001901	0.01163	0	495	0	0	0	0	1	-360	360;
	1804	547	0.00101	0.00824	0	0	0	0	0	0	1	-360	360;
	8786	5766	0.00787	0.068219	0	457	0	0	0	0	1	-360	360;
	8786	5766	0.007711	0.069539	0	476	0	0	0	0	1	-360	360;
	1613	5997	0.00032	0.002091	0	0	0	0	0	0	1	-360	360;
	2805	4610	0.00012	0.00182	0	0	0	0	0	0	1	-360	360;
	4610	6	0.00058	0.00736	0	1415	0	0	0	0	1	-360	360;
	2805	1760	0.00074	0.01143	0	1481	0	0	0	0	1	-360	360;
	7234	5575	0.00084	0.01378	0	1843	0	0	0	0	1	-360	360;
	7234	7182	0.00037	0.00356	0	1580	0	0	0	0	1	-360	360;
	7108	7234	0.00012	0.00206	0	1481	0	0	0	0	1	-360	360;
	7487	6302	0.00036	0.00151	0	0	0	0	0	0	1	-360	360;
	9121	1512	0.0012	0.01835	0	954	0	0	0	0	1	-360	360;
	1512	6189	0.00088	0.01503	0	1645	0	0	0	0	1	-360	360;
	2160	1512	0.00057	0.00939	0	1744	0	0	0	0	1	-360	360;
	3351	3696	0.00157	0.02006	0	1744	0	0	0	0	1	-360	360;
	3351	5577	0.00027	0.00349	0	1810	0	0	0	0	1	-360	360;
	7664	3351	0.00136	0.02325	0	0	0	0	0	0	1	-360	360;
	3351	3895	0.00038	0.00478	0	1744	0	0	0	0	1	-360	360;
	3351	6444	0.00077	0.01171	0	1613	0	0	0	0	1	-360	360;
	9121	5476	0.00129	0.02076	0	1185	0	0	0	0	1	-360	360;
	9121	1760	0.00037	0.0046	0	1283	0	0	0	0	1	-360	360;
	1760	9121	0.00021	0.00375	0	1481	0	0	0	0	1	-360	360;
	9121	2160	0.00055	0.00885	0	1382	0	0	0	0	1	-360	360;
	6809	1804	0.001659	0.01138	0	457	0	0	0	0	1	-360	360;
	6809	4904	0.001169	0.010021	0	457	0	0	0	0	1	-360	360;
	6809	1423	0.002229	0.00975	0	533	0	0	0	0	1	-360	360;
	6809	7123	0.00049	0.004521	0	438	0	0	0	0	1	-360	360;
	6809	6649	0.0002	0.001969	0	591	0	0	0	0	1	-360	360;
	2986	4836	0.0001	0.00158	0	1678	0	0	0	0	1	-360	360;
	2986	4836	0.00011	0.00151	0	1547	0	0	0	0	1	-360	360;
	1193	4943	3e-05	0.00043	0	757	0	0	0	0	1	-360	360;
	1193	685	0.00037	0.00589	0	1843	0	0	0	0	1	-360	360;
	2039	1193	0.00087	0.01218	0	1448	0	0	0	0	1	-360	360;
	6776	5766	0.00362	0.024409	0	457	0	0	0	0	1	-360	360;
	6776	3832	0.00488	0.03119	0	438	0	0	0	0	1	-360	360;
	1327	2314	0.00055	0.00264	0	533	0	0	0	0	1	-360	360;
	6107	2314	0.00031	0.001341	0	533	0	0	0	0	1	-360	360;
	5959	2314	0.00012	0.000669	0	495	0	0	0	0	1	-360	360;
	837	1688	3.1e-05	0.000469	0	0	0	0	0	0	1	-360	360;
	1991	1688	3.1e-05	0.000461	0	0	0	0	0	0	1	-360	360;
	2040	4502	0.00607	0.03457	0	381	0	0	0	0	1	-360	360;
	4502	4516	0.0087	0.047719	0	457	0	0	0	0	1	-360	360;
	3811	8237	0.00129	0.02159	0	1382	0	0	0	0	1	-360	360;
	3811	1224	0.00091	0.00903	0	1645	0	0	0	0	1	-360	360;
	8608	3811	0.0004	0.00518	0	1744	0	0	0	0	1	-360	360;
	3811	6736	0.00032	0.00557	0	0	0	0	0	0	1	-360	360;
	7494	4719	3.1e-05	0.0002	0	0	0	0	0	0	1	-360	360;
	7494	6191	3.1e-05	0.000211	0	0	0	0	0	0	1	-360	360;
	5451	2570	0.00199	0.011591	0	457	0	0	0	0	1	-360	360;
	3699	5451	0.000211	0.000919	0	0	0	0	0	0	1	-360	360;
	3972	2456	0.00109	0.01525	0	1547	0	0	0	0	1	-360	360;
	748	2570	0.005789	0.02826	0	438	0	0	0	0	1	-360	360;
	6529	2884	0.00041	0.0064	0	1843	0	0	0	0	1	-360	360;
	6529	2684	0.00051	0.00603	0	1678	0	0	0	0	1	-360	360;
	3560	4283	0.00245	0.02726	0	514	0	0	0	0	1	-360	360;
	3560	7167	0.000271	0.00337	0	438	0	0	0	0	1	-360	360;
	3634	6892	0.000151	0.00064	0	0	0	0	0	0	1	-360	360;
	2594	3634	0.003341	0.022091	0	0	0	0	0	0	1	-360	360;
	3634	1236	0.022	0.125771	0	457	0	0	0	0	1	-360	360;
	3634	2317	0.0153	0.08863	0	457	0	0	0	0	1	-360	360;
	3634	2317	0.015599	0.08682	0	476	0	0	0	0	1	-360	360;
	704	2034	0.001169	0.00844	0	0	0	0	0	0	1	-360	360;
	6302	704	0.00081	0.004271	0	0	0	0	0	0	1	-360	360;
	704	5913	0.000349	0.0015	0	0	0	0	0	0	1	-360	360;
	3696	6444	0.00058	0.00717	0	1942	0	0	0	0	1	-360	360;
	3696	5623	0.0007	0.01203	0	1810	0	0	0	0	1	-360	360;
	5623	3696	0.00061	0.00957	0	1909	0	0	0	0	1	-360	360;
	8663	2589	2e-05	0.00075	0	1514	0	0	0	0	1	-360	360;
	7268	8663	0.00041	0.01221	0	0	0	0	0	0	1	-360	360;
	1034	2103	0.00157	0.007969	0	0	0	0	0	0	1	-360	360;
	4243	2103	0.00151	0.00625	0	0	0	0	0	0	1	-360	360;
	4433	4436	2e-05	0.00023	0	1218	0	0	0	0	1	-360	360;
	4433	4436	2e-05	0.00022	0	1185	0	0	0	0	1	-360	360;
	3874	4433	0.00026	0.00395	0	1613	0	0	0	0	1	-360	360;
	3874	4433	0.00025	0.00403	0	1711	0	0	0	0	1	-360	360;
	5945	238	0.00111	0.005099	0	457	0	0	0	0	1	-360	360;
	238	4715	0.000729	0.0023	0	476	0	0	0	0	1	-360	360;
	4715	8036	0.003831	0.016271	0	438	0	0	0	0	1	-360	360;
	7479	9135	0.001849	0.014719	0	514	0	0	0	0	1	-360	360;
	7403	7479	0.00157	0.010169	0	457	0	0	0	0	1	-360	360;
	528	6303	0.00149	0.01513	0	0	0	0	0	0	1	-360	360;
	4152	528	0.003229	0.02918	0	0	0	0	0	0	1	-360	360;
	8237	2100	1e-05	0.00021	0	1711	0	0	0	0	1	-360	360;
	2302	2199	0.00054	0.00765	0	1547	0	0	0	0	1	-360	360;
	4036	5577	1e-05	0.0002	0	757	0	0	0	0	1	-360	360;
	5577	6444	0.00053	0.00746	0	1645	0	0	0	0	1	-360	360;
	3238	3892	0.010469	0.057979	0	0	0	0	0	0	1	-360	360;
	3238	2650	0.00014	0.000781	0	0	0	0	0	0	1	-360	360;
	3238	893	0.00181	0.013039	0	0	0	0	0	0	1	-360	360;
	4138	8774	0.01107	0.04905	0	419	0	0	0	0	1	-360	360;
	2615	1610	0.00038	0.00468	0	1613	0	0	0	0	1	-360	360;
	2615	1610	0.00034	0.0053	0	1415	0	0	0	0	1	-360	360;
	3343	2615	0.00012	0.00185	0	757	0	0	0	0	1	-360	360;
	3343	2615	0.00013	0.00181	0	757	0	0	0	0	1	-360	360;
	5292	7753	0.000211	0.0018	0	553	0	0	0	0	1	-360	360;
	1480	3384	0.00037	0.00225	0	572	0	0	0	0	1	-360	360;
	2766	7891	0.00025	0.00317	0	1909	0	0	0	0	1	-360	360;
	1684	2766	0.00029	0.00484	0	1876	0	0	0	0	1	-360	360;
	99	7248	0.00057	0.00858	0	0	0	0	0	0	1	-360	360;
	99	9009	0.0005	0.00665	0	1909	0	0	0	0	1	-360	360;
	7804	3024	0.00239	0.012289	0	457	0	0	0	0	1	-360	360;
	6566	3024	0.0008	0.005211	0	381	0	0	0	0	1	-360	360;
	6189	4019	0.00067	0.01081	0	1580	0	0	0	0	1	-360	360;
	2680	6189	0.00105	0.01455	0	1744	0	0	0	0	1	-360	360;
	485	7248	0.0002	0.00261	0	790	0	0	0	0	1	-360	360;
	5982	8708	1e-05	0.0002	0	757	0	0	0	0	1	-360	360;
	1224	8708	0.00036	0.00601	0	0	0	0	0	0	1	-360	360;
	9009	8708	0.0001	0.00151	0	0	0	0	0	0	1	-360	360;
	4487	802	2e-05	0.00021	0	757	0	0	0	0	1	-360	360;
	8237	802	0.00046	0.00568	0	1251	0	0	0	0	1	-360	360;
	660	802	0.00041	0.00531	0	1283	0	0	0	0	1	-360	360;
	2669	2302	0.0003	0.00482	0	1876	0	0	0	0	1	-360	360;
	2039	220	1e-05	0.0002	0	1744	0	0	0	0	1	-360	360;
	7664	1131	1e-05	0.00022	0	1843	0	0	0	0	1	-360	360;
	2709	8573	2e-05	0.00025	0	790	0	0	0	0	1	-360	360;
	2709	8573	2e-05	0.00026	0	757	0	0	0	0	1	-360	360;
	882	3174	0.00032	0.00528	0	1711	0	0	0	0	1	-360	360;
	2923	3174	0.0006	0.0084	0	1876	0	0	0	0	1	-360	360;
	6464	8139	0.00105	0.01735	0	1514	0	0	0	0	1	-360	360;
	4159	6464	0.00135	0.01758	0	1481	0	0	0	0	1	-360	360;
	7248	8139	0.0004	0.0068	0	1810	0	0	0	0	1	-360	360;
	7248	8139	0.00039	0.0063	0	1711	0	0	0	0	1	-360	360;
	8139	9009	0.00067	0.00834	0	1415	0	0	0	0	1	-360	360;
	7672	685	0.00084	0.01114	0	1876	0	0	0	0	1	-360	360;
	2039	685	0.00081	0.011	0	1613	0	0	0	0	1	-360	360;
	3123	3159	0.00059	0.00726	0	1448	0	0	0	0	1	-360	360;
	4159	3159	0.001	0.01178	0	1744	0	0	0	0	1	-360	360;
	4436	2344	0.00031	0.00197	0	1218	0	0	0	0	1	-360	360;
	4436	2344	0.00035	0.00193	0	1251	0	0	0	0	1	-360	360;
	6408	2344	2e-05	0.0004	0	954	0	0	0	0	1	-360	360;
	6408	2344	2e-05	0.00041	0	1020	0	0	0	0	1	-360	360;
	4869	4564	3.1e-05	0.000271	0	457	0	0	0	0	1	-360	360;
	4564	4869	3.1e-05	0.000281	0	457	0	0	0	0	1	-360	360;
	4564	3228	0.000229	0.000271	0	514	0	0	0	0	1	-360	360;
	2232	5854	0.000289	0.001229	0	0	0	0	0	0	1	-360	360;
	2232	8774	0.002479	0.01106	0	0	0	0	0	0	1	-360	360;
	8524	748	0.00738	0.04069	0	419	0	0	0	0	1	-360	360;
	6939	6560	0.00058	0.00872	0	1744	0	0	0	0	1	-360	360;
	6560	5868	0.00116	0.01483	0	1909	0	0	0	0	1	-360	360;
	6560	5868	0.00118	0.01452	0	1942	0	0	0	0	1	-360	360;
	7248	3123	0.00139	0.02265	0	1942	0	0	0	0	1	-360	360;
	7248	3123	0.00149	0.02349	0	1942	0	0	0	0	1	-360	360;
	4040	3123	0.00086	0.01488	0	1876	0	0	0	0	1	-360	360;
	4040	3123	0.00095	0.01554	0	1942	0	0	0	0	1	-360	360;
	342	759	0.019651	0.09944	0	457	0	0	0	0	1	-360	360;
	7071	1269	0.0002	0.001831	0	476	0	0	0	0	1	-360	360;
	8857	2822	7e-05	0.00062	0	457	0	0	0	0	1	-360	360;
	2822	13	0.004039	0.023081	0	457	0	0	0	0	1	-360	360;
	4516	2822	0.0132	0.064331	0	0	0	0	0	0	1	-360	360;
	2822	9013	0.001401	0.008729	0	0	0	0	0	0	1	-360	360;
	6444	9132	0.00021	0.00247	0	1119	0	0	0	0	1	-360	360;
	6444	5623	0.00138	0.01791	0	1810	0	0	0	0	1	-360	360;
	8237	660	0.00056	0.00795	0	1744	0	0	0	0	1	-360	360;
	8237	5328	0.00081	0.00788	0	1152	0	0	0	0	1	-360	360;
	7619	6939	3e-05	0.00042	0	757	0	0	0	0	1	-360	360;
	1748	6939	3e-05	0.00042	0	856	0	0	0	0	1	-360	360;
	8569	6939	3e-05	0.00035	0	823	0	0	0	0	1	-360	360;
	6778	6939	3e-05	0.00036	0	757	0	0	0	0	1	-360	360;
	7891	6939	0.00035	0.00423	0	954	0	0	0	0	1	-360	360;
	3099	5109	0.00076	0.00904	0	1316	0	0	0	0	1	-360	360;
	5109	4680	0.00029	0.00361	0	1382	0	0	0	0	1	-360	360;
	7103	6070	0.010349	0.056729	0	476	0	0	0	0	1	-360	360;
	7103	2034	0.001581	0.012771	0	533	0	0	0	0	1	-360	360;
	7103	1034	0.001081	0.01136	0	0	0	0	0	0	1	-360	360;
	7103	4243	0.001669	0.00914	0	438	0	0	0	0	1	-360	360;
	7103	5677	0.01757	0.07575	0	514	0	0	0	0	1	-360	360;
	817	7182	0.00037	0.00558	0	0	0	0	0	0	1	-360	360;
	817	8635	0.00041	0.0072	0	1415	0	0	0	0	1	-360	360;
	7787	4574	0.002479	0.017961	0	419	0	0	0	0	1	-360	360;
	1209	7787	0.00155	0.0127	0	438	0	0	0	0	1	-360	360;
	142	7835	0.001961	0.010419	0	457	0	0	0	0	1	-360	360;
	2135	2034	0.00138	0.007521	0	0	0	0	0	0	1	-360	360;
	8907	2135	0.00281	0.02543	0	0	0	0	0	0	1	-360	360;
	2135	8634	0.000211	0.000901	0	495	0	0	0	0	1	-360	360;
	7540	3816	0.000789	0.00519	0	0	0	0	0	0	1	-360	360;
	7540	102	0.00056	0.004281	0	0	0	0	0	0	1	-360	360;
	7540	8320	0.00163	0.012031	0	591	0	0	0	0	1	-360	360;
	4152	7540	0.000211	0.00161	0	495	0	0	0	0	1	-360	360;
	7540	5704	0.00018	0.00144	0	476	0	0	0	0	1	-360	360;
	7320	1804	9.1e-05	0.00081	0	0	0	0	0	0	1	-360	360;
	8484	1804	9.1e-05	0.00099	0	0	0	0	0	0	1	-360	360;
	3337	5089	0.001219	0.01461	0	0	0	0	0	0	1	-360	360;
	3337	2200	0.00769	0.035169	0	629	0	0	0	0	1	-360	360;
	7618	8246	5e-05	0.000219	0	438	0	0	0	0	1	-360	360;
	8246	5913	0.000281	0.00112	0	438	0	0	0	0	1	-360	360;
	5746	4283	0.003591	0.01825	0	438	0	0	0	0	1	-360	360;
	5746	4283	0.00125	0.01386	0	457	0	0	0	0	1	-360	360;
	342	4283	0.000961	0.00513	0	514	0	0	0	0	1	-360	360;
	1610	4381	0.00022	0.00295	0	757	0	0	0	0	1	-360	360;
	5668	8426	0.00326	0.02562	0	419	0	0	0	0	1	-360	360;
	5668	7593	0.00164	0.013219	0	572	0	0	0	0	1	-360	360;
	4168	832	8.1e-05	0.00019	0	0	0	0	0	0	1	-360	360;
	4168	832	3.9e-05	0.00032	0	381	0	0	0	0	1	-360	360;
	4168	832	0.00011	0.0002	0	0	0	0	0	0	1	-360	360;
	4168	832	0.000103	0.000171	0	0	0	0	0	0	1	-360	360;
	4168	832	0.00011	0.00019	0	0	0	0	0	0	1	-360	360;
	4168	3816	8.1e-05	0.001	0	438	0	0	0	0	1	-360	360;
	4168	102	0.000281	0.00213	0	419	0	0	0	0	1	-360	360;
	4152	4168	0.00025	0.000831	0	457	0	0	0	0	1	-360	360;
	8275	4168	0.00013	0.000469	0	419	0	0	0	0	1	-360	360;
	3099	229	0.00088	0.01144	0	1283	0	0	0	0	1	-360	360;
	4330	229	0.00076	0.01445	0	1020	0	0	0	0	1	-360	360;
	2187	7672	0.00048	0.0076	0	1777	0	0	0	0	1	-360	360;
	7672	2362	0.00017	0.00273	0	1876	0	0	0	0	1	-360	360;
	8933	7564	0.00014	0.000849	0	0	0	0	0	0	1	-360	360;
	4995	8933	0.0022	0.01289	0	419	0	0	0	0	1	-360	360;
	3554	3807	0.00582	0.03749	0	476	0	0	0	0	1	-360	360;
	3554	3807	0.00593	0.03669	0	476	0	0	0	0	1	-360	360;
	3807	5809	0.000281	0.003419	0	629	0	0	0	0	1	-360	360;
	3807	4089	0.00645	0.03995	0	476	0	0	0	0	1	-360	360;
	3807	4089	0.00657	0.039081	0	457	0	0	0	0	1	-360	360;
	8399	8869	8e-05	0.00125	0	1744	0	0	0	0	1	-360	360;
	8399	8741	8e-05	0.00127	0	757	0	0	0	0	1	-360	360;
	8399	4606	7e-05	0.0013	0	757	0	0	0	0	1	-360	360;
	8399	1488	7e-05	0.0012	0	757	0	0	0	0	1	-360	360;
	8876	5575	5e-05	0.0006	0	1514	0	0	0	0	1	-360	360;
	5575	6	0.00034	0.00466	0	0	0	0	0	0	1	-360	360;
	6594	6302	0.000591	0.00588	0	495	0	0	0	0	1	-360	360;
	6302	6594	0.00139	0.00818	0	495	0	0	0	0	1	-360	360;
	4243	6594	0.00169	0.013969	0	572	0	0	0	0	1	-360	360;
	3958	339	0.000229	0.00163	0	438	0	0	0	0	1	-360	360;
	4352	8521	0.00275	0.02836	0	610	0	0	0	0	1	-360	360;
	54	8521	0.004909	0.02468	0	438	0	0	0	0	1	-360	360;
	8521	8774	0.001419	0.02212	0	419	0	0	0	0	1	-360	360;
	4561	8399	0.0016	0.02118	0	1909	0	0	0	0	1	-360	360;
	8399	6316	0.00116	0.01869	0	1547	0	0	0	0	1	-360	360;
	8161	8399	0.00169	0.02149	0	1415	0	0	0	0	1	-360	360;
	8399	5328	0.00171	0.02913	0	856	0	0	0	0	1	-360	360;
	8399	1031	0.00138	0.0212	0	1645	0	0	0	0	1	-360	360;
	2484	2139	0.00511	0.02775	0	495	0	0	0	0	1	-360	360;
	3972	4159	0.0008	0.01262	0	0	0	0	0	0	1	-360	360;
	5130	376	0.0005	0.00851	0	1744	0	0	0	0	1	-360	360;
	5130	2710	0.00032	0.00558	0	1843	0	0	0	0	1	-360	360;
	2694	5510	0.00039	0.00601	0	1876	0	0	0	0	1	-360	360;
	2694	4512	0.00036	0.00575	0	1514	0	0	0	0	1	-360	360;
	2694	8117	0.00117	0.01616	0	1810	0	0	0	0	1	-360	360;
	882	6161	0.0003	0.00471	0	1448	0	0	0	0	1	-360	360;
	3554	745	0.00011	0.00081	0	476	0	0	0	0	1	-360	360;
	3554	1613	0.01087	0.06457	0	457	0	0	0	0	1	-360	360;
	3554	1613	0.00968	0.073169	0	457	0	0	0	0	1	-360	360;
	2034	7618	0.00105	0.005781	0	0	0	0	0	0	1	-360	360;
	1871	2880	0.002961	0.029021	0	438	0	0	0	0	1	-360	360;
	4147	376	4e-05	0.00033	0	1349	0	0	0	0	1	-360	360;
	2332	376	4e-05	0.00035	0	1415	0	0	0	0	1	-360	360;
	5124	376	4e-05	0.0003	0	757	0	0	0	0	1	-360	360;
	3348	376	4e-05	0.00028	0	757	0	0	0	0	1	-360	360;
	4904	88	0.000211	0.00045	0	0	0	0	0	0	1	-360	360;
	9197	3721	0.00024	0.001099	0	381	0	0	0	0	1	-360	360;
	4133	9197	0.00544	0.029831	0	381	0	0	0	0	1	-360	360;
	2634	7182	0.00076	0.0101	0	1942	0	0	0	0	1	-360	360;
	2040	5839	0.01106	0.06313	0	381	0	0	0	0	1	-360	360;
	5694	5746	0.001909	0.011401	0	476	0	0	0	0	1	-360	360;
	6566	5694	0.00113	0.005479	0	419	0	0	0	0	1	-360	360;
	1327	3705	0.000151	0.000961	0	495	0	0	0	0	1	-360	360;
	1327	2317	0.00156	0.01125	0	400	0	0	0	0	1	-360	360;
	1327	2317	0.00344	0.021599	0	0	0	0	0	0	1	-360	360;
	1308	5089	0.00106	0.01276	0	0	0	0	0	0	1	-360	360;
	4348	5866	0.001349	0.008831	0	0	0	0	0	0	1	-360	360;
	3721	8857	0.001901	0.010031	0	381	0	0	0	0	1	-360	360;
	2224	7403	0.001719	0.012159	0	476	0	0	0	0	1	-360	360;
	3196	105	4e-05	0.00061	0	0	0	0	0	0	1	-360	360;
	6067	191	0.00126	0.00595	0	0	0	0	0	0	1	-360	360;
	2923	6	0.00018	0.00224	0	1942	0	0	0	0	1	-360	360;
	6107	1236	0.000521	0.002039	0	533	0	0	0	0	1	-360	360;
	7893	8857	0.002169	0.016479	0	419	0	0	0	0	1	-360	360;
	4561	5510	0.00073	0.01168	0	1876	0	0	0	0	1	-360	360;
	8161	5510	0.00131	0.0162	0	1481	0	0	0	0	1	-360	360;
	7494	2388	0.00112	0.007591	0	438	0	0	0	0	1	-360	360;
	7123	7494	0.00162	0.01332	0	457	0	0	0	0	1	-360	360;
	6217	8161	0.00027	0.00344	0	757	0	0	0	0	1	-360	360;
	8009	8161	0.0003	0.00313	0	757	0	0	0	0	1	-360	360;
	1822	1760	0.00014	0.00243	0	1744	0	0	0	0	1	-360	360;
	1826	4832	0.00019	0.001781	0	419	0	0	0	0	1	-360	360;
	3188	4832	0.000151	0.001969	0	457	0	0	0	0	1	-360	360;
	2594	1236	0.025461	0.14561	0	438	0	0	0	0	1	-360	360;
	7593	8426	0.004599	0.028581	0	476	0	0	0	0	1	-360	360;
	4094	1688	0.00987	0.05706	0	457	0	0	0	0	1	-360	360;
	105	2577	9e-05	0.00103	0	1218	0	0	0	0	1	-360	360;
	105	2577	0.0001	0.00101	0	1086	0	0	0	0	1	-360	360;
	2710	2577	0.00023	0.00381	0	0	0	0	0	0	1	-360	360;
	127	2034	0.002219	0.01374	0	553	0	0	0	0	1	-360	360;
	8623	3099	0.00092	0.00895	0	0	0	0	0	0	1	-360	360;
	127	6302	0.001151	0.009479	0	553	0	0	0	0	1	-360	360;
	992	2200	0.00188	0.012151	0	610	0	0	0	0	1	-360	360;
	3823	1610	0.00036	0.00471	0	1645	0	0	0	0	1	-360	360;
	3823	946	0.00085	0.01337	0	0	0	0	0	0	1	-360	360;
	3035	3823	9e-05	0.00105	0	1481	0	0	0	0	1	-360	360;
	3343	3823	0.00037	0.00634	0	1514	0	0	0	0	1	-360	360;
	4089	2983	0.00056	0.004979	0	381	0	0	0	0	1	-360	360;
	3832	5766	0.00488	0.029039	0	438	0	0	0	0	1	-360	360;
	3832	5766	0.004349	0.032919	0	438	0	0	0	0	1	-360	360;
	2634	2884	0.00046	0.00571	0	1942	0	0	0	0	1	-360	360;
	9135	5645	0.010841	0.04989	0	438	0	0	0	0	1	-360	360;
	3234	992	0.00282	0.01436	0	400	0	0	0	0	1	-360	360;
	3234	992	0.00282	0.014789	0	476	0	0	0	0	1	-360	360;
	256	2650	0.00012	0.000841	0	419	0	0	0	0	1	-360	360;
	256	893	0.00175	0.01189	0	381	0	0	0	0	1	-360	360;
	256	1688	0.00851	0.04389	0	495	0	0	0	0	1	-360	360;
	6680	6756	0.00063	0.00374	0	0	0	0	0	0	1	-360	360;
	3874	162	0.00037	0.00484	0	987	0	0	0	0	1	-360	360;
	8608	5328	0.00161	0.02178	0	1777	0	0	0	0	1	-360	360;
	8600	1423	0.00745	0.03588	0	419	0	0	0	0	1	-360	360;
	2109	4991	0.00051	0.00829	0	1678	0	0	0	0	1	-360	360;
	2109	8117	0.00124	0.01751	0	1382	0	0	0	0	1	-360	360;
	2109	7891	0.0008	0.01242	0	987	0	0	0	0	1	-360	360;
	1760	5476	0.00118	0.01502	0	1843	0	0	0	0	1	-360	360;
	3874	7268	0.00131	0.01723	0	1152	0	0	0	0	1	-360	360;
	1269	5959	0.000271	0.001081	0	457	0	0	0	0	1	-360	360;
	6201	2986	0.00014	0.00178	0	1218	0	0	0	0	1	-360	360;
	1269	1236	0.00051	0.00201	0	514	0	0	0	0	1	-360	360;
	7891	2920	9e-05	0.00091	0	1744	0	0	0	0	1	-360	360;
	1939	471	9.1e-05	0.000591	0	0	0	0	0	0	1	-360	360;
	3228	5014	2.1e-05	0.00025	0	419	0	0	0	0	1	-360	360;
	2039	6408	0.00035	0.00556	0	1711	0	0	0	0	1	-360	360;
	2039	6408	0.0004	0.00545	0	1777	0	0	0	0	1	-360	360;
	6316	4521	0.0004	0.00517	0	1020	0	0	0	0	1	-360	360;
	4521	5328	0.00037	0.00456	0	1218	0	0	0	0	1	-360	360;
	5355	6316	0.00133	0.01413	0	1415	0	0	0	0	1	-360	360;
	7066	4348	0.01187	0.05524	0	0	0	0	0	0	1	-360	360;
	4348	3188	0.00063	0.0065	0	438	0	0	0	0	1	-360	360;
	1688	4348	0.011841	0.052289	0	400	0	0	0	0	1	-360	360;
	660	1356	0.00055	0.00517	0	921	0	0	0	0	1	-360	360;
	1356	5355	0.00097	0.01163	0	1382	0	0	0	0	1	-360	360;
	1826	5532	0.000969	0.01039	0	495	0	0	0	0	1	-360	360;
	8907	6670	0.00326	0.015581	0	457	0	0	0	0	1	-360	360;
	2710	4680	0.00099	0.01348	0	0	0	0	0	0	1	-360	360;
	5868	4680	0.00045	0.00368	0	0	0	0	0	0	1	-360	360;
	1034	4243	0.002021	0.011331	0	0	0	0	0	0	1	-360	360;
	9115	1034	0.00076	0.0073	0	553	0	0	0	0	1	-360	360;
	2187	4040	0.00101	0.01505	0	0	0	0	0	0	1	-360	360;
	6362	692	0	0.000347	0	301	0	0	0	0	1	-360	360;
	2680	4019	0.00065	0.00768	0	0	0	0	0	0	1	-360	360;
	4991	6161	0.00025	0.0035	0	1415	0	0	0	0	1	-360	360;
	3816	2574	0.00013	0.000331	0	381	0	0	0	0	1	-360	360;
	102	2574	3.9e-05	0.00036	0	495	0	0	0	0	1	-360	360;
	4516	2880	0.011021	0.061901	0	457	0	0	0	0	1	-360	360;
	2880	5839	0.007901	0.04495	0	419	0	0	0	0	1	-360	360;
	1032	6302	0.00369	0.02186	0	438	0	0	0	0	1	-360	360;
	6302	6567	0.000281	0.001091	0	457	0	0	0	0	1	-360	360;
	6302	5913	0.000461	0.001841	0	476	0	0	0	0	1	-360	360;
	1257	3605	0.006271	0.02999	0	400	0	0	0	0	1	-360	360;
	4516	13	0.011219	0.05245	0	381	0	0	0	0	1	-360	360;
	8509	4352	0.000281	0.004091	0	572	0	0	0	0	1	-360	360;
	2709	3963	0.00041	0.00534	0	1678	0	0	0	0	1	-360	360;
	3596	8320	0.002659	0.01557	0	0	0	0	0	0	1	-360	360;
	8774	54	0.00795	0.0483	0	0	0	0	0	0	1	-360	360;
	6649	3195	0.00037	0.00361	0	457	0	0	0	0	1	-360	360;
	2025	5292	0.001211	0.0082	0	591	0	0	0	0	1	-360	360;
	1613	3850	9.9e-05	0.00101	0	400	0	0	0	0	1	-360	360;
	1613	3480	9.9e-05	0.00099	0	0	0	0	0	0	1	-360	360;
	1613	5192	0.000169	0.001539	0	495	0	0	0	0	1	-360	360;
	1613	3841	0.000169	0.00157	0	476	0	0	0	0	1	-360	360;
	893	1867	0.00011	0.000349	0	381	0	0	0	0	1	-360	360;
	893	1867	0.00012	0.000341	0	400	0	0	0	0	1	-360	360;
	2680	875	0.00017	0.00314	0	790	0	0	0	0	1	-360	360;
	399	3540	0.00012	0.00056	0	0	0	0	0	0	1	-360	360;
	1688	1561	0.000219	0.001469	0	476	0	0	0	0	1	-360	360;
	1688	1561	0.0002	0.001781	0	438	0	0	0	0	1	-360	360;
	8588	946	0.00038	0.00623	0	1645	0	0	0	0	1	-360	360;
	3596	7403	0.003669	0.01631	0	438	0	0	0	0	1	-360	360;
	697	5355	0.00013	0.00117	0	1744	0	0	0	0	1	-360	360;
	8413	5355	0.00012	0.0013	0	1415	0	0	0	0	1	-360	360;
	7066	3020	0.00082	0.004969	0	0	0	0	0	0	1	-360	360;
	7066	3020	0.0008	0.004599	0	0	0	0	0	0	1	-360	360;
	1688	7066	0.00914	0.060289	0	495	0	0	0	0	1	-360	360;
	4516	8036	0.005219	0.02681	0	0	0	0	0	0	1	-360	360;
	4516	8036	0.005919	0.026271	0	0	0	0	0	0	1	-360	360;
	4516	5839	0.00381	0.02086	0	419	0	0	0	0	1	-360	360;
	1236	7574	0.004539	0.01957	0	419	0	0	0	0	1	-360	360;
	2803	1236	0.00112	0.00632	0	400	0	0	0	0	1	-360	360;
	6736	4414	0.00027	0.00426	0	1777	0	0	0	0	1	-360	360;
	9009	4414	0.0002	0.00238	0	987	0	0	0	0	1	-360	360;
	1082	4414	8e-05	0.00077	0	1645	0	0	0	0	1	-360	360;
	4516	1209	0.00149	0.012091	0	591	0	0	0	0	1	-360	360;
	9187	4243	0.00064	0.00356	0	0	0	0	0	0	1	-360	360;
	4995	1220	0.005729	0.033	0	572	0	0	0	0	1	-360	360;
	9013	6680	0.0017	0.007599	0	0	0	0	0	0	1	-360	360;
	8117	6460	0.00221	0.02589	0	1547	0	0	0	0	1	-360	360;
	3705	191	0.000331	0.000831	0	0	0	0	0	0	1	-360	360;
	4152	8275	9.9e-05	0.00045	0	0	0	0	0	0	1	-360	360;
	8634	4243	0.00039	0.00163	0	476	0	0	0	0	1	-360	360;
	8161	5328	0.00383	0.04628	0	1448	0	0	0	0	1	-360	360;
	6566	7167	0.00106	0.01101	0	438	0	0	0	0	1	-360	360;
	9115	7306	0.00131	0.012659	0	495	0	0	0	0	1	-360	360;
	2680	5623	0.00052	0.00735	0	1415	0	0	0	0	1	-360	360;
	5623	2680	0.00058	0.0074	0	1086	0	0	0	0	1	-360	360;
	1031	2680	0.0013	0.01837	0	1251	0	0	0	0	1	-360	360;
	4666	7743	3.9e-05	0.000599	0	0	0	0	0	0	1	-360	360;
	1257	7574	0.00031	0.001461	0	495	0	0	0	0	1	-360	360;
	3228	6963	0.001151	0.008771	0	476	0	0	0	0	1	-360	360;
	4869	4126	0.00095	0.00457	0	572	0	0	0	0	1	-360	360;
	4869	1303	0.001419	0.00982	0	438	0	0	0	0	1	-360	360;
	8634	6567	0.00037	0.00144	0	476	0	0	0	0	1	-360	360;
	7306	1308	0.00143	0.01225	0	629	0	0	0	0	1	-360	360;
	1092	7574	0.00044	0.00207	0	381	0	0	0	0	1	-360	360;
	4330	6460	0.00076	0.01509	0	1613	0	0	0	0	1	-360	360;
	2317	2909	0.0007	0.004039	0	0	0	0	0	0	1	-360	360;
	4448	8681	0.004419	0.01918	0	400	0	0	0	0	1	-360	360;
	6670	4035	0.006099	0.029539	0	514	0	0	0	0	1	-360	360;
	7622	7362	0.00106	0.00513	0	0	0	0	0	0	1	-360	360;
	8785	9233	0.001021	0.004711	0	572	0	0	0	0	1	-360	360;
	1651	4892	0.000211	0.00106	0	533	0	0	0	0	1	-360	360;
	8125	3977	7e-05	0.00087	0	0	0	0	0	0	1	-360	360;
	8125	4892	3.9e-05	0.000651	0	0	0	0	0	0	1	-360	360;
	9233	8132	0.002039	0.008271	0	0	0	0	0	0	1	-360	360;
	9233	8132	0.002	0.00843	0	0	0	0	0	0	1	-360	360;
	4064	7362	0.001219	0.00725	0	591	0	0	0	0	1	-360	360;
	7362	9233	0.001151	0.006789	0	514	0	0	0	0	1	-360	360;
	4064	7362	0.0012	0.006711	0	591	0	0	0	0	1	-360	360;
	7362	9233	0.00112	0.006289	0	553	0	0	0	0	1	-360	360;
	7667	7154	0.002021	0.01101	0	0	0	0	0	0	1	-360	360;
	7154	4064	0.0015	0.008521	0	0	0	0	0	0	1	-360	360;
	7154	4064	0.001469	0.00868	0	0	0	0	0	0	1	-360	360;
	9233	3977	0.000719	0.00443	0	0	0	0	0	0	1	-360	360;
	9233	4892	0.0005	0.00268	0	591	0	0	0	0	1	-360	360;
	6214	1932	0.029149	0.051802	0	0	0	0	0	0	1	-360	360;
	5698	1932	0.035777	0.063587	0	0	0	0	0	0	1	-360	360;
	6214	3089	0.025967	0.058702	0	0	0	0	0	0	1	-360	360;
	3089	82	0.01943	0.043942	0	0	0	0	0	0	1	-360	360;
	3089	1874	0.05776	0.115802	0	0	0	0	0	0	1	-360	360;
	213	3089	0.027388	0.054917	0	0	0	0	0	0	1	-360	360;
	82	1874	0.028438	0.104719	0	0	0	0	0	0	1	-360	360;
	5201	5020	0.008091	0.03755	0	419	0	0	0	0	1	-360	360;
	8496	5020	0.00556	0.02913	0	438	0	0	0	0	1	-360	360;
	7721	213	0.033512	0.069909	0	0	0	0	0	0	1	-360	360;
	8496	4810	0.00587	0.02949	0	400	0	0	0	0	1	-360	360;
	2729	5752	0.013314	0.040719	0	0	0	0	0	0	1	-360	360;
	2729	5752	0.01357	0.039876	0	0	0	0	0	0	1	-360	360;
	2729	5752	0.014579	0.041157	0	0	0	0	0	0	1	-360	360;
	2729	4270	0.031298	0.112364	0	0	0	0	0	0	1	-360	360;
	2729	4270	0.030653	0.114521	0	0	0	0	0	0	1	-360	360;
	246	3888	0.00537	0.04394	0	438	0	0	0	0	1	-360	360;
	246	3888	0.005461	0.04726	0	438	0	0	0	0	1	-360	360;
	5201	246	0.01312	0.074419	0	419	0	0	0	0	1	-360	360;
	5752	6862	0	0.000413	0	0	0	0	0	0	1	-360	360;
	6862	1997	0.052702	0.121281	0	0	0	0	0	0	1	-360	360;
	9122	3262	0.00103	0.01397	0	1152	0	0	0	0	1	-360	360;
	4270	1874	0.002959	0.014851	0	0	0	0	0	0	1	-360	360;
	4270	1874	0.002769	0.016339	0	0	0	0	0	0	1	-360	360;
	9122	7674	0.00167	0.02105	0	1218	0	0	0	0	1	-360	360;
	5698	1874	0.067826	0.163702	0	0	0	0	0	0	1	-360	360;
	3618	2417	0.009463	0.038388	0	0	0	0	0	0	1	-360	360;
	5251	2417	0.021289	0.124008	0	0	0	0	0	0	1	-360	360;
	5901	5542	0.035587	0.104182	0	0	0	0	0	0	1	-360	360;
	1182	5542	0.040678	0.119107	0	0	0	0	0	0	1	-360	360;
	7330	8344	0.00156	0.01963	0	1283	0	0	0	0	1	-360	360;
	7330	3259	0.00255	0.02314	0	1415	0	0	0	0	1	-360	360;
	7033	8469	0.01305	0.035686	0	0	0	0	0	0	1	-360	360;
	7033	8469	0.013314	0.034959	0	0	0	0	0	0	1	-360	360;
	2048	8469	0.005612	0.022636	0	0	0	0	0	0	1	-360	360;
	64	8469	0.029826	0.113909	0	0	0	0	0	0	1	-360	360;
	6752	8469	0.04824	0.109727	0	0	0	0	0	0	1	-360	360;
	7033	7533	0.010802	0.04124	0	0	0	0	0	0	1	-360	360;
	2048	7033	0.021851	0.050934	0	0	0	0	0	0	1	-360	360;
	3516	7033	0.070223	0.192198	0	0	0	0	0	0	1	-360	360;
	3516	7033	0.071603	0.188281	0	0	0	0	0	0	1	-360	360;
	2048	7533	0.008339	0.021066	0	0	0	0	0	0	1	-360	360;
	2162	2048	0.024744	0.092099	0	0	0	0	0	0	1	-360	360;
	2950	3875	0.029579	0.114661	0	0	0	0	0	0	1	-360	360;
	1182	3875	0.009298	0.035488	0	0	0	0	0	0	1	-360	360;
	4713	3875	0.020149	0.078107	0	0	0	0	0	0	1	-360	360;
	3053	5901	0.017826	0.063777	0	0	0	0	0	0	1	-360	360;
	3259	8344	0.00158	0.02095	0	1316	0	0	0	0	1	-360	360;
	3146	8344	0.00076	0.00957	0	1251	0	0	0	0	1	-360	360;
	4760	2950	0.005678	0.044198	0	0	0	0	0	0	1	-360	360;
	4760	2950	0.006438	0.043314	0	0	0	0	0	0	1	-360	360;
	6500	2950	0.013612	0.043157	0	0	0	0	0	0	1	-360	360;
	909	2950	0.056248	0.147934	0	0	0	0	0	0	1	-360	360;
	909	2950	0.057339	0.144851	0	0	0	0	0	0	1	-360	360;
	8464	2950	0.000463	0.003603	0	238	0	0	0	0	1	-360	360;
	2950	7557	0.037107	0.138157	0	0	0	0	0	0	1	-360	360;
	4760	8119	0.004058	0.015107	0	0	0	0	0	0	1	-360	360;
	6573	1283	0.011793	0.034521	0	0	0	0	0	0	1	-360	360;
	2073	1283	0.019636	0.073132	0	0	0	0	0	0	1	-360	360;
	4309	1736	0	0.000413	0	248	0	0	0	0	1	-360	360;
	3618	917	0.014339	0.041983	0	0	0	0	0	0	1	-360	360;
	5197	917	0.029314	0.118281	0	0	0	0	0	0	1	-360	360;
	8495	2727	0.016636	0.064488	0	0	0	0	0	0	1	-360	360;
	6392	2727	0.034273	0.127612	0	0	0	0	0	0	1	-360	360;
	2727	5251	0.00886	0.02495	0	0	0	0	0	0	1	-360	360;
	8495	6483	0.026587	0.084364	0	0	0	0	0	0	1	-360	360;
	6392	4559	0.039298	0.115058	0	0	0	0	0	0	1	-360	360;
	6392	6573	0.036066	0.134289	0	0	0	0	0	0	1	-360	360;
	6392	1696	0.000529	0.001554	0	0	0	0	0	0	1	-360	360;
	64	6500	0.03981	0.160628	0	0	0	0	0	0	1	-360	360;
	8119	6500	0.008438	0.025744	0	0	0	0	0	0	1	-360	360;
	6500	3796	0.024223	0.093876	0	0	0	0	0	0	1	-360	360;
	6483	4705	0.019868	0.054017	0	0	0	0	0	0	1	-360	360;
	6752	4705	0.051248	0.121347	0	0	0	0	0	0	1	-360	360;
	5593	5561	0.042793	0.101314	0	0	0	0	0	0	1	-360	360;
	5593	5561	0.030091	0.099372	0	0	0	0	0	0	1	-360	360;
	2280	5561	0.005488	0.020959	0	0	0	0	0	0	1	-360	360;
	2062	5036	0.009198	0.025851	0	0	0	0	0	0	1	-360	360;
	5036	6122	0.019752	0.057843	0	0	0	0	0	0	1	-360	360;
	2073	5036	0.032934	0.096413	0	0	0	0	0	0	1	-360	360;
	6721	8513	0.05586	0.112686	0	0	0	0	0	0	1	-360	360;
	8767	8513	0.029744	0.057636	0	0	0	0	0	0	1	-360	360;
	3575	6721	0.004868	0.013686	0	0	0	0	0	0	1	-360	360;
	418	6721	0.005413	0.015207	0	0	0	0	0	0	1	-360	360;
	6573	6721	0.043669	0.084653	0	0	0	0	0	0	1	-360	360;
	418	3575	0.004818	0.01724	0	0	0	0	0	0	1	-360	360;
	48	3575	0.017719	0.054017	0	0	0	0	0	0	1	-360	360;
	6573	3897	0.004364	0.017719	0	0	0	0	0	0	1	-360	360;
	3897	1710	0.007463	0.041727	0	0	0	0	0	0	1	-360	360;
	2062	4559	0.016289	0.065711	0	0	0	0	0	0	1	-360	360;
	2280	3053	0.02014	0.072066	0	0	0	0	0	0	1	-360	360;
	3014	8453	0.003091	0.011826	0	0	0	0	0	0	1	-360	360;
	3516	8453	0.00386	0.013793	0	0	0	0	0	0	1	-360	360;
	2162	3014	0.036	0.114223	0	0	0	0	0	0	1	-360	360;
	5197	3516	0.027942	0.081793	0	0	0	0	0	0	1	-360	360;
	5593	4694	0.027008	0.061041	0	0	0	0	0	0	1	-360	360;
	7272	4694	0.025504	0.072198	0	0	0	0	0	0	1	-360	360;
	6122	6483	0.01686	0.060331	0	0	0	0	0	0	1	-360	360;
	7648	1710	0.010091	0.038529	0	0	0	0	0	0	1	-360	360;
	7272	7648	0.027099	0.082603	0	0	0	0	0	0	1	-360	360;
	4703	5524	0.014463	0.051719	0	0	0	0	0	0	1	-360	360;
	5524	5251	0.014612	0.041066	0	0	0	0	0	0	1	-360	360;
	1830	64	0.001661	0.006446	0	0	0	0	0	0	1	-360	360;
	7286	64	0.005198	0.019331	0	0	0	0	0	0	1	-360	360;
	6784	1830	0.003512	0.011587	0	0	0	0	0	0	1	-360	360;
	6784	7286	0.004339	0.015512	0	0	0	0	0	0	1	-360	360;
	5593	8767	0.043628	0.081231	0	0	0	0	0	0	1	-360	360;
	2717	1182	0.031661	0.092669	0	0	0	0	0	0	1	-360	360;
	1710	48	0.017562	0.070818	0	0	0	0	0	0	1	-360	360;
	2073	4559	0.008843	0.033802	0	0	0	0	0	0	1	-360	360;
	6860	8859	0.01164	0.05205	0	400	0	0	0	0	1	-360	360;
	1710	6573	0.007934	0.042653	0	0	0	0	0	0	1	-360	360;
	8967	6573	0.00343	0.014488	0	0	0	0	0	0	1	-360	360;
	121	6573	0.004339	0.022421	0	0	0	0	0	0	1	-360	360;
	6573	4309	0.00143	0.007149	0	0	0	0	0	0	1	-360	360;
	6573	4309	0.001446	0.007	0	0	0	0	0	0	1	-360	360;
	6573	4309	0.001504	0.007256	0	0	0	0	0	0	1	-360	360;
	6573	4309	0.001529	0.007099	0	0	0	0	0	0	1	-360	360;
	1710	2073	0.015653	0.051669	0	0	0	0	0	0	1	-360	360;
	4208	3259	0.00046	0.00424	0	1251	0	0	0	0	1	-360	360;
	3618	5251	0.049339	0.138727	0	0	0	0	0	0	1	-360	360;
	4713	5593	0.005421	0.017876	0	0	0	0	0	0	1	-360	360;
	909	2717	0.001678	0.006421	0	0	0	0	0	0	1	-360	360;
	909	7557	0.024512	0.055752	0	0	0	0	0	0	1	-360	360;
	3882	1696	0.019347	0.072058	0	0	0	0	0	0	1	-360	360;
	4703	3882	0.002364	0.008777	0	0	0	0	0	0	1	-360	360;
	3882	5251	0.032331	0.102537	0	0	0	0	0	0	1	-360	360;
	121	8967	0.00305	0.017711	0	0	0	0	0	0	1	-360	360;
	5499	7272	0.005289	0.009463	0	0	0	0	0	0	1	-360	360;
	5499	8425	0.00181	0.007298	0	0	0	0	0	0	1	-360	360;
	8425	7272	0.000893	0.00338	0	0	0	0	0	0	1	-360	360;
	3796	8425	0.012033	0.046579	0	0	0	0	0	0	1	-360	360;
	6000	1689	0.00017	0.01734	0	1185	0	0	0	0	1	-360	360;
	4192	8230	0.00017	0.01767	0	1086	0	0	0	0	1	-360	360;
	8814	2030	0.0006	0.040218	0	637	0	0	0	0	1	-360	360;
	8832	2180	0.00076	0.002409	0	0	0	0	0	0	1	-360	360;
	8832	3771	0	0.000538	0	0	0	0	0	0	1	-360	360;
	6395	1564	0.00528	0.027569	0	0	0	0	0	0	1	-360	360;
	6395	1564	0.005382	0.026978	0	0	0	0	0	0	1	-360	360;
	5255	6395	0.00104	0.009649	0	0	0	0	0	0	1	-360	360;
	5255	6395	0.001062	0.009449	0	273	0	0	0	0	1	-360	360;
	3397	5247	0.000289	0.000889	0	0	0	0	0	0	1	-360	360;
	1207	5247	0.003502	0.01524	0	0	0	0	0	0	1	-360	360;
	3397	7958	0.000289	0.00088	0	0	0	0	0	0	1	-360	360;
	1207	7958	0.003431	0.014111	0	0	0	0	0	0	1	-360	360;
	8389	5526	0.001502	0.1272	0	0	0	0	0	0	1	-360	360;
	91	722	8e-05	0.000729	0	312	0	0	0	0	1	-360	360;
	1207	722	0.003818	0.012911	0	0	0	0	0	0	1	-360	360;
	9083	722	0.001618	0.013711	0	0	0	0	0	0	1	-360	360;
	3363	1870	0.001062	0.007791	0	0	0	0	0	0	1	-360	360;
	6539	1870	0.001809	0.008951	0	0	0	0	0	0	1	-360	360;
	5792	1870	0.007142	0.020311	0	0	0	0	0	0	1	-360	360;
	3363	8356	0.000942	0.008831	0	0	0	0	0	0	1	-360	360;
	6539	8356	0.001769	0.008289	0	0	0	0	0	0	1	-360	360;
	5792	8356	0.00728	0.019898	0	0	0	0	0	0	1	-360	360;
	8828	5740	0.003569	0.017951	0	0	0	0	0	0	1	-360	360;
	3443	8828	0.001991	0.007271	0	0	0	0	0	0	1	-360	360;
	7701	8828	0.000391	0.001462	0	0	0	0	0	0	1	-360	360;
	2619	8167	0.003502	0.018289	0	0	0	0	0	0	1	-360	360;
	244	2619	0.001769	0.00824	0	0	0	0	0	0	1	-360	360;
	7701	2619	0.0004	0.001431	0	0	0	0	0	0	1	-360	360;
	1810	4740	0	0.000413	0	0	0	0	0	0	1	-360	360;
	5279	7525	0.001471	0.1296	0	0	0	0	0	0	1	-360	360;
	2180	7374	0.00056	0.015018	0	0	0	0	0	0	1	-360	360;
	164	8490	0.000631	0.01472	0	0	0	0	0	0	1	-360	360;
	164	2180	0	0.001169	0	0	0	0	0	0	1	-360	360;
	164	7596	0.000942	0.003342	0	0	0	0	0	0	1	-360	360;
	164	2708	0.001582	0.002831	0	0	0	0	0	0	1	-360	360;
	3119	7059	0.0008	0.003422	0	0	0	0	0	0	1	-360	360;
	3119	7059	0.000822	0.003351	0	0	0	0	0	0	1	-360	360;
	7596	2708	0.002631	0.004889	0	0	0	0	0	0	1	-360	360;
	7596	2305	0.005418	0.022489	0	0	0	0	0	0	1	-360	360;
	7596	2305	0.00552	0.022018	0	0	0	0	0	0	1	-360	360;
	7394	7575	0.004091	0.020893	0	0	0	0	0	0	1	-360	360;
	7394	7575	0.004	0.019339	0	0	0	0	0	0	1	-360	360;
	7394	4415	0.016479	0.037058	0	0	0	0	0	0	1	-360	360;
	7394	4686	0.016347	0.05124	0	0	0	0	0	0	1	-360	360;
	5783	7575	0.007058	0.030587	0	0	0	0	0	0	1	-360	360;
	5783	7575	0.007149	0.029934	0	0	0	0	0	0	1	-360	360;
	6823	6202	0.002089	0.003449	0	0	0	0	0	0	1	-360	360;
	6823	6202	0.002129	0.003378	0	0	0	0	0	0	1	-360	360;
	2330	712	0	0.00028	0	0	0	0	0	0	1	-360	360;
	712	4272	0.007489	0.0516	0	0	0	0	0	0	1	-360	360;
	712	4272	0.008489	0.050569	0	0	0	0	0	0	1	-360	360;
	878	3206	0	0.047951	0	0	0	0	0	0	1	-360	360;
	7090	3206	0.001649	0.016591	0	0	0	0	0	0	1	-360	360;
	7844	3206	0.001618	0.016898	0	0	0	0	0	0	1	-360	360;
	3497	6202	0.006871	0.039911	0	0	0	0	0	0	1	-360	360;
	3497	6202	0.007009	0.039089	0	0	0	0	0	0	1	-360	360;
	3497	3182	0.001658	0.011871	0	0	0	0	0	0	1	-360	360;
	3497	3182	0.001631	0.010991	0	0	0	0	0	0	1	-360	360;
	5426	9120	0.00092	0.004689	0	0	0	0	0	0	1	-360	360;
	133	9120	0.002049	0.0038	0	0	0	0	0	0	1	-360	360;
	6323	9120	0.005649	0.026	0	0	0	0	0	0	1	-360	360;
	5426	3047	0.000902	0.004782	0	0	0	0	0	0	1	-360	360;
	275	3047	0.003471	0.014471	0	0	0	0	0	0	1	-360	360;
	133	3047	0.002089	0.00372	0	0	0	0	0	0	1	-360	360;
	4188	1880	0.003569	0.010671	0	0	0	0	0	0	1	-360	360;
	3898	4188	0.019782	0.061658	0	0	0	0	0	0	1	-360	360;
	486	5740	0.000409	0.0014	0	0	0	0	0	0	1	-360	360;
	5740	539	0.003302	0.017311	0	0	0	0	0	0	1	-360	360;
	486	8167	0.000409	0.001369	0	0	0	0	0	0	1	-360	360;
	8167	539	0.003231	0.016031	0	0	0	0	0	0	1	-360	360;
	9094	8463	0.000751	0.005711	0	0	0	0	0	0	1	-360	360;
	9094	8994	0.0018	0.00716	0	299	0	0	0	0	1	-360	360;
	6354	9094	0.001978	0.013462	0	0	0	0	0	0	1	-360	360;
	9094	1936	0.001502	0.0056	0	0	0	0	0	0	1	-360	360;
	3280	6423	0.005991	0.0284	0	0	0	0	0	0	1	-360	360;
	2055	6423	0.000302	0.000858	0	0	0	0	0	0	1	-360	360;
	6275	6423	0.002622	0.013231	0	0	0	0	0	0	1	-360	360;
	7138	5618	0.004271	0.019502	0	0	0	0	0	0	1	-360	360;
	2055	5618	0.000271	0.000969	0	0	0	0	0	0	1	-360	360;
	6275	5618	0.00356	0.014378	0	0	0	0	0	0	1	-360	360;
	6288	3456	0.0022	0.009569	0	0	0	0	0	0	1	-360	360;
	2745	3456	0.000311	0.001169	0	0	0	0	0	0	1	-360	360;
	3301	3456	0.001889	0.006311	0	0	0	0	0	0	1	-360	360;
	2745	3280	0.000271	0.00132	0	0	0	0	0	0	1	-360	360;
	3301	3280	0.001929	0.006178	0	260	0	0	0	0	1	-360	360;
	790	6600	0.000339	0.001149	0	0	0	0	0	0	1	-360	360;
	472	8889	0.002591	0.014418	0	0	0	0	0	0	1	-360	360;
	472	775	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6356	3443	0.0014	0.00652	0	0	0	0	0	0	1	-360	360;
	8164	3443	0.000418	0.001338	0	0	0	0	0	0	1	-360	360;
	6356	244	0.001369	0.006649	0	0	0	0	0	0	1	-360	360;
	8164	244	0.000369	0.00152	0	0	0	0	0	0	1	-360	360;
	8814	5320	0.004689	0.018151	0	0	0	0	0	0	1	-360	360;
	8814	5320	0.004591	0.018498	0	0	0	0	0	0	1	-360	360;
	1255	143	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	1255	3974	0.003698	0.0198	0	0	0	0	0	0	1	-360	360;
	8016	1255	0.005471	0.0206	0	0	0	0	0	0	1	-360	360;
	8016	1255	0.0062	0.020191	0	0	0	0	0	0	1	-360	360;
	1255	4362	0.006098	0.030551	0	0	0	0	0	0	1	-360	360;
	950	5792	0.004809	0.019751	0	0	0	0	0	0	1	-360	360;
	950	5792	0.004711	0.018289	0	260	0	0	0	0	1	-360	360;
	3397	6045	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6288	7138	0.007929	0.044022	0	0	0	0	0	0	1	-360	360;
	6057	1352	0.009782	0.033938	0	0	0	0	0	0	1	-360	360;
	6057	6559	0.003609	0.016889	0	0	0	0	0	0	1	-360	360;
	656	1392	0.00044	0.027591	0	0	0	0	0	0	1	-360	360;
	6835	3538	0.000419	0.027979	0	0	0	0	0	0	1	-360	360;
	351	3538	0.000409	0.02851	0	0	0	0	0	0	1	-360	360;
	1392	3538	0	0.000461	0	0	0	0	0	0	1	-360	360;
	1392	2824	0.000531	0.005581	0	0	0	0	0	0	1	-360	360;
	1052	1392	0.003841	0.022461	0	476	0	0	0	0	1	-360	360;
	3538	140	0.000539	0.005461	0	400	0	0	0	0	1	-360	360;
	3878	3538	0.00376	0.022901	0	457	0	0	0	0	1	-360	360;
	5001	8862	0.028843	0.060992	0	0	0	0	0	0	1	-360	360;
	5001	8862	0.029397	0.059744	0	0	0	0	0	0	1	-360	360;
	1810	6342	0.001727	0.00924	0	0	0	0	0	0	1	-360	360;
	1777	1810	0.004851	0.02724	0	0	0	0	0	0	1	-360	360;
	1352	6559	0.0138	0.040249	0	260	0	0	0	0	1	-360	360;
	1352	643	0.007098	0.022329	0	0	0	0	0	0	1	-360	360;
	1352	1836	0.001902	0.008391	0	0	0	0	0	0	1	-360	360;
	1352	1836	0.001929	0.008209	0	0	0	0	0	0	1	-360	360;
	1670	1352	0.001649	0.01228	0	0	0	0	0	0	1	-360	360;
	1670	1352	0.001471	0.01392	0	0	0	0	0	0	1	-360	360;
	1352	7628	0.000582	0.040991	0	299	0	0	0	0	1	-360	360;
	1352	311	0.01116	0.033671	0	286	0	0	0	0	1	-360	360;
	6998	2482	0	0.00028	0	0	0	0	0	0	1	-360	360;
	5008	8249	0.00031	0.00312	0	1251	0	0	0	0	1	-360	360;
	6068	8249	0.00069	0.00769	0	1810	0	0	0	0	1	-360	360;
	6139	5450	0.010471	0.029289	0	0	0	0	0	0	1	-360	360;
	6139	5450	0.010671	0.028671	0	0	0	0	0	0	1	-360	360;
	344	6139	0.00624	0.021871	0	299	0	0	0	0	1	-360	360;
	344	6139	0.007071	0.021431	0	299	0	0	0	0	1	-360	360;
	8881	7751	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	8881	6559	0.001062	0.003351	0	0	0	0	0	0	1	-360	360;
	8881	5342	0.002111	0.02008	0	0	0	0	0	0	1	-360	360;
	8881	6399	0.001009	0.010929	0	0	0	0	0	0	1	-360	360;
	1670	8881	0.004711	0.050609	0	0	0	0	0	0	1	-360	360;
	8881	9083	0.011462	0.02396	0	0	0	0	0	0	1	-360	360;
	5279	2350	0	0.0444	0	0	0	0	0	0	1	-360	360;
	5279	1123	0.001698	0.1128	0	0	0	0	0	0	1	-360	360;
	5279	1633	8.9e-05	0.000742	0	0	0	0	0	0	1	-360	360;
	5279	1633	8.9e-05	0.000729	0	0	0	0	0	0	1	-360	360;
	5279	2812	0.000302	0.000809	0	0	0	0	0	0	1	-360	360;
	5279	9029	0.001502	0.002182	0	0	0	0	0	0	1	-360	360;
	5279	6149	0.001311	0.001871	0	0	0	0	0	0	1	-360	360;
	5279	6149	0.001369	0.00204	0	0	0	0	0	0	1	-360	360;
	5279	4616	0.001049	0.00308	0	273	0	0	0	0	1	-360	360;
	1777	5893	0.014967	0.09105	0	0	0	0	0	0	1	-360	360;
	5296	5893	0.028959	0.12905	0	0	0	0	0	0	1	-360	360;
	2478	5893	0.006438	0.041256	0	0	0	0	0	0	1	-360	360;
	2478	5893	0.006554	0.040364	0	0	0	0	0	0	1	-360	360;
	2350	1260	0	0.012742	0	0	0	0	0	0	1	-360	360;
	6354	8463	0.001018	0.00868	0	0	0	0	0	0	1	-360	360;
	6354	8994	0.001871	0.00548	0	0	0	0	0	0	1	-360	360;
	8994	7476	0.00228	0.004338	0	0	0	0	0	0	1	-360	360;
	8994	7476	0.00224	0.004431	0	0	0	0	0	0	1	-360	360;
	3023	806	0.00076	0.00747	0	1678	0	0	0	0	1	-360	360;
	5674	806	0.00103	0.01061	0	1481	0	0	0	0	1	-360	360;
	2240	8991	0.002689	0.013818	0	0	0	0	0	0	1	-360	360;
	539	2240	0.002111	0.003769	0	0	0	0	0	0	1	-360	360;
	539	2240	0.002151	0.003689	0	273	0	0	0	0	1	-360	360;
	8814	5426	0.00176	0.012702	0	0	0	0	0	0	1	-360	360;
	8814	5426	0.001729	0.012951	0	0	0	0	0	0	1	-360	360;
	5748	1046	0.000151	0.000511	0	0	0	0	0	0	1	-360	360;
	5748	1046	0.000151	0.000502	0	260	0	0	0	0	1	-360	360;
	5748	1836	0.01076	0.05508	0	0	0	0	0	0	1	-360	360;
	5748	1836	0.010542	0.05616	0	0	0	0	0	0	1	-360	360;
	5748	6824	0.003089	0.009	0	0	0	0	0	0	1	-360	360;
	5748	6824	0.003031	0.009169	0	0	0	0	0	0	1	-360	360;
	5748	6275	0.001849	0.00872	0	0	0	0	0	0	1	-360	360;
	5748	6275	0.00188	0.008538	0	0	0	0	0	0	1	-360	360;
	2974	3301	0.004151	0.01556	0	0	0	0	0	0	1	-360	360;
	1670	2974	0.006591	0.026778	0	299	0	0	0	0	1	-360	360;
	6998	8875	0.00056	0.038671	0	507	0	0	0	0	1	-360	360;
	1747	593	0	0.00045	0	757	0	0	0	0	1	-360	360;
	593	8230	0.00162	0.01968	0	1613	0	0	0	0	1	-360	360;
	6998	8772	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	6998	6356	0.002649	0.015111	0	0	0	0	0	0	1	-360	360;
	6998	6356	0.002702	0.014809	0	0	0	0	0	0	1	-360	360;
	6998	5031	0.0008	0.005711	0	0	0	0	0	0	1	-360	360;
	6998	5031	0.000782	0.005809	0	0	0	0	0	0	1	-360	360;
	6998	3299	0.001049	0.007738	0	286	0	0	0	0	1	-360	360;
	6998	3299	0.001071	0.007569	0	299	0	0	0	0	1	-360	360;
	6998	3119	0.005138	0.025338	0	0	0	0	0	0	1	-360	360;
	6998	3119	0.005031	0.025831	0	0	0	0	0	0	1	-360	360;
	6998	1464	0.003658	0.029071	0	0	0	0	0	0	1	-360	360;
	6998	1464	0.004151	0.028489	0	0	0	0	0	0	1	-360	360;
	1221	1350	0.001471	0.011471	0	0	0	0	0	0	1	-360	360;
	1221	1350	0.001502	0.011218	0	0	0	0	0	0	1	-360	360;
	2305	1350	0.001871	0.014511	0	0	0	0	0	0	1	-360	360;
	2305	1350	0.001902	0.0142	0	0	0	0	0	0	1	-360	360;
	199	7195	0.000351	0.004662	0	0	0	0	0	0	1	-360	360;
	52	199	0.00772	0.033818	0	0	0	0	0	0	1	-360	360;
	4964	8432	0	0.00051	0	1909	0	0	0	0	1	-360	360;
	8432	3190	0.00076	0.0119	0	790	0	0	0	0	1	-360	360;
	8432	3352	0.00025	0.00316	0	1580	0	0	0	0	1	-360	360;
	4964	3484	0.00077	0.01164	0	823	0	0	0	0	1	-360	360;
	4964	6766	0.00025	0.00322	0	1481	0	0	0	0	1	-360	360;
	838	6295	0.001369	0.021182	0	494	0	0	0	0	1	-360	360;
	3190	3484	0	0.00052	0	757	0	0	0	0	1	-360	360;
	3190	6263	0.00015	0.00328	0	790	0	0	0	0	1	-360	360;
	395	3190	0.00016	0.00314	0	1349	0	0	0	0	1	-360	360;
	3484	6068	0.00017	0.00321	0	987	0	0	0	0	1	-360	360;
	5843	3484	0.00018	0.00308	0	954	0	0	0	0	1	-360	360;
	3210	4021	0.00024	0.01948	0	757	0	0	0	0	1	-360	360;
	1670	7425	0.000631	0.037902	0	0	0	0	0	0	1	-360	360;
	2381	5772	0.00244	0.01532	0	0	0	0	0	0	1	-360	360;
	281	8086	0	0.00053	0	757	0	0	0	0	1	-360	360;
	281	6068	0.00013	0.00282	0	987	0	0	0	0	1	-360	360;
	4077	281	0.00038	0.00708	0	1218	0	0	0	0	1	-360	360;
	8086	6263	0.00013	0.00261	0	1086	0	0	0	0	1	-360	360;
	5888	8086	0.00038	0.00722	0	1448	0	0	0	0	1	-360	360;
	1477	5255	0.00016	0.002769	0	0	0	0	0	0	1	-360	360;
	3960	5255	0.00016	0.00272	0	299	0	0	0	0	1	-360	360;
	8037	5255	0.00016	0.002662	0	299	0	0	0	0	1	-360	360;
	1989	1597	0.024479	0.081727	0	0	0	0	0	0	1	-360	360;
	6923	1597	0.024248	0.064893	0	0	0	0	0	0	1	-360	360;
	676	8706	0.0084	0.04056	0	0	0	0	0	0	1	-360	360;
	8706	205	0.003809	0.0248	0	0	0	0	0	0	1	-360	360;
	8706	4362	0.002542	0.014569	0	0	0	0	0	0	1	-360	360;
	8706	4362	0.002591	0.01428	0	0	0	0	0	0	1	-360	360;
	7584	7736	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	7584	5591	0.003071	0.01636	0	0	0	0	0	0	1	-360	360;
	7584	5364	0.003889	0.027022	0	0	0	0	0	0	1	-360	360;
	4774	7584	0.00624	0.036009	0	0	0	0	0	0	1	-360	360;
	4774	7584	0.00636	0.035258	0	0	0	0	0	0	1	-360	360;
	5255	2796	0.000982	0.001529	0	0	0	0	0	0	1	-360	360;
	5255	2796	0.000991	0.001502	0	0	0	0	0	0	1	-360	360;
	8642	5255	0.00052	0.004991	0	260	0	0	0	0	1	-360	360;
	5255	2703	0	0.00028	0	0	0	0	0	0	1	-360	360;
	650	4104	0.012182	0.034826	0	0	0	0	0	0	1	-360	360;
	1989	650	0.025744	0.059	0	0	0	0	0	0	1	-360	360;
	4130	1635	0	0.00028	0	0	0	0	0	0	1	-360	360;
	6167	4130	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6167	5872	0.003129	0.015031	0	0	0	0	0	0	1	-360	360;
	4130	496	0.00228	0.019871	0	0	0	0	0	0	1	-360	360;
	4130	5872	0.002791	0.017031	0	0	0	0	0	0	1	-360	360;
	6582	3462	0.001431	0.01252	0	0	0	0	0	0	1	-360	360;
	6582	4082	0.001649	0.007631	0	0	0	0	0	0	1	-360	360;
	6582	5246	0.001471	0.008649	0	0	0	0	0	0	1	-360	360;
	496	5872	0.00656	0.009071	0	260	0	0	0	0	1	-360	360;
	6047	3664	0	0.00028	0	0	0	0	0	0	1	-360	360;
	2077	9063	0	0.00028	0	0	0	0	0	0	1	-360	360;
	3224	8044	0.000521	0.001529	0	391	0	0	0	0	1	-360	360;
	76	5742	0.000219	0.00095	0	438	0	0	0	0	1	-360	360;
	38	5742	0.000219	0.00088	0	533	0	0	0	0	1	-360	360;
	8814	9118	0.002391	0.015618	0	0	0	0	0	0	1	-360	360;
	5008	919	0	0.00054	0	1448	0	0	0	0	1	-360	360;
	1788	5008	0.00079	0.01037	0	1678	0	0	0	0	1	-360	360;
	662	5008	0.00022	0.00296	0	1645	0	0	0	0	1	-360	360;
	8604	919	0.00077	0.01056	0	1810	0	0	0	0	1	-360	360;
	6354	4295	0.000649	0.03712	0	0	0	0	0	0	1	-360	360;
	6354	5203	0.003302	0.019138	0	0	0	0	0	0	1	-360	360;
	6354	5203	0.003369	0.018742	0	0	0	0	0	0	1	-360	360;
	6354	472	0.001769	0.018222	0	0	0	0	0	0	1	-360	360;
	6354	7476	0.000858	0.006809	0	0	0	0	0	0	1	-360	360;
	6354	7476	0.000969	0.006671	0	0	0	0	0	0	1	-360	360;
	6354	8792	0.005551	0.03908	0	0	0	0	0	0	1	-360	360;
	6354	8792	0.005431	0.039822	0	0	0	0	0	0	1	-360	360;
	6354	8792	0.005311	0.036871	0	0	0	0	0	0	1	-360	360;
	1020	1094	0.002169	0.016271	0	0	0	0	0	0	1	-360	360;
	3182	1094	0.007418	0.07832	0	0	0	0	0	0	1	-360	360;
	4825	3023	0	0.0005	0	757	0	0	0	0	1	-360	360;
	4825	6345	0	0.00049	0	757	0	0	0	0	1	-360	360;
	4825	8114	0.00079	0.00848	0	1580	0	0	0	0	1	-360	360;
	4825	3210	0.00183	0.0174	0	1415	0	0	0	0	1	-360	360;
	3023	6345	0	0.00048	0	757	0	0	0	0	1	-360	360;
	3023	8773	0.0008	0.0083	0	1678	0	0	0	0	1	-360	360;
	7884	71	0.004449	0.02404	0	0	0	0	0	0	1	-360	360;
	7884	71	0.004529	0.023551	0	0	0	0	0	0	1	-360	360;
	7884	6202	0.006791	0.013271	0	260	0	0	0	0	1	-360	360;
	7884	1106	0.00424	0.030378	0	0	0	0	0	0	1	-360	360;
	7884	1106	0.004151	0.028129	0	0	0	0	0	0	1	-360	360;
	7884	7305	0.002311	0.003622	0	0	0	0	0	0	1	-360	360;
	7884	7305	0.002062	0.004102	0	0	0	0	0	0	1	-360	360;
	7884	6987	0.00452	0.029609	0	0	0	0	0	0	1	-360	360;
	7884	6987	0.004422	0.030169	0	0	0	0	0	0	1	-360	360;
	7884	4774	0.00416	0.023058	0	0	0	0	0	0	1	-360	360;
	7884	3620	0.002831	0.020742	0	273	0	0	0	0	1	-360	360;
	3119	1722	0.002338	0.01592	0	0	0	0	0	0	1	-360	360;
	585	3148	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	1106	585	0.004969	0.03636	0	0	0	0	0	0	1	-360	360;
	6987	585	0.003431	0.015698	0	0	0	0	0	0	1	-360	360;
	7275	585	0.00456	0.027582	0	0	0	0	0	0	1	-360	360;
	8114	8773	0	0.00047	0	757	0	0	0	0	1	-360	360;
	8393	8114	0.00093	0.01185	0	1810	0	0	0	0	1	-360	360;
	139	8773	0.00091	0.01207	0	1810	0	0	0	0	1	-360	360;
	676	6439	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	8738	676	0.000529	0.003151	0	0	0	0	0	0	1	-360	360;
	8738	676	0.00052	0.003209	0	0	0	0	0	0	1	-360	360;
	676	205	0.00352	0.020578	0	0	0	0	0	0	1	-360	360;
	8862	6915	0.039182	0.089413	0	0	0	0	0	0	1	-360	360;
	8862	9205	0.033686	0.067124	0	0	0	0	0	0	1	-360	360;
	6393	4057	0.044992	0.093174	0	0	0	0	0	0	1	-360	360;
	6028	6393	0.013678	0.052479	0	0	0	0	0	0	1	-360	360;
	4872	6393	0.005686	0.022893	0	0	0	0	0	0	1	-360	360;
	6393	7136	0.034281	0.094653	0	0	0	0	0	0	1	-360	360;
	2769	6118	0	0.00045	0	0	0	0	0	0	1	-360	360;
	76	2769	0.00032	0.001469	0	438	0	0	0	0	1	-360	360;
	38	6118	0.000289	0.001039	0	457	0	0	0	0	1	-360	360;
	3224	7117	0.00324	0.017628	0	0	0	0	0	0	1	-360	360;
	3224	8525	0.002678	0.019298	0	0	0	0	0	0	1	-360	360;
	3224	5346	0.000529	0.001504	0	0	0	0	0	0	1	-360	360;
	4993	3335	0.00042	0.0261	0	757	0	0	0	0	1	-360	360;
	6832	4797	0.00043	0.02555	0	757	0	0	0	0	1	-360	360;
	6489	4797	0.00044	0.02499	0	757	0	0	0	0	1	-360	360;
	8284	4797	0.00039	0.02833	0	757	0	0	0	0	1	-360	360;
	577	1362	0.00019	0.01594	0	0	0	0	0	0	1	-360	360;
	268	1057	0.000419	0.026099	0	0	0	0	0	0	1	-360	360;
	2728	5979	0.00043	0.02555	0	0	0	0	0	0	1	-360	360;
	8662	8298	0.00038	0.02888	0	757	0	0	0	0	1	-360	360;
	6884	4059	0.00038	0.02944	0	757	0	0	0	0	1	-360	360;
	3723	3335	0.00037	0.02999	0	1876	0	0	0	0	1	-360	360;
	1890	3335	0.00036	0.02777	0	0	0	0	0	0	1	-360	360;
	71	8141	9.8e-05	0.00028	0	0	0	0	0	0	1	-360	360;
	7295	71	0.003951	0.02048	0	0	0	0	0	0	1	-360	360;
	71	7297	0.003858	0.020871	0	0	0	0	0	0	1	-360	360;
	4797	8298	0	0.00046	0	757	0	0	0	0	1	-360	360;
	4797	4059	0	0.00045	0	757	0	0	0	0	1	-360	360;
	4797	7388	0	0.00051	0	823	0	0	0	0	1	-360	360;
	4797	2002	0	0.00052	0	823	0	0	0	0	1	-360	360;
	2448	5979	0	0.00051	0	0	0	0	0	0	1	-360	360;
	2448	1057	0	0.000521	0	0	0	0	0	0	1	-360	360;
	2448	6282	0	0.000531	0	381	0	0	0	0	1	-360	360;
	2448	4340	0	0.000539	0	0	0	0	0	0	1	-360	360;
	3335	8298	0	0.00053	0	1415	0	0	0	0	1	-360	360;
	3335	4059	0	0.00054	0	954	0	0	0	0	1	-360	360;
	3335	7388	0	0.0005	0	1119	0	0	0	0	1	-360	360;
	3335	2002	0	0.00049	0	921	0	0	0	0	1	-360	360;
	326	5979	0	0.0005	0	0	0	0	0	0	1	-360	360;
	326	1057	0	0.00049	0	0	0	0	0	0	1	-360	360;
	326	6282	0	0.000479	0	0	0	0	0	0	1	-360	360;
	326	4340	0	0.000469	0	0	0	0	0	0	1	-360	360;
	4858	8298	0.00043	0.00544	0	1580	0	0	0	0	1	-360	360;
	2523	5979	0	0.000461	0	476	0	0	0	0	1	-360	360;
	1057	1362	0	0.00045	0	457	0	0	0	0	1	-360	360;
	1956	7388	0.00038	0.00616	0	1448	0	0	0	0	1	-360	360;
	5299	2002	0.00041	0.02721	0	1086	0	0	0	0	1	-360	360;
	4340	1052	0.002081	0.023	0	533	0	0	0	0	1	-360	360;
	6662	838	0	0.000222	0	0	0	0	0	0	1	-360	360;
	5165	2722	0	0.00048	0	1514	0	0	0	0	1	-360	360;
	5165	1689	0.0011	0.01189	0	1810	0	0	0	0	1	-360	360;
	5165	3210	0.00084	0.00906	0	1020	0	0	0	0	1	-360	360;
	2722	1689	0.00112	0.01165	0	1876	0	0	0	0	1	-360	360;
	2722	8230	0.00115	0.0114	0	1744	0	0	0	0	1	-360	360;
	2722	5674	0.00086	0.00887	0	1283	0	0	0	0	1	-360	360;
	3363	7195	0.00116	0.00912	0	260	0	0	0	0	1	-360	360;
	3363	7195	0.001129	0.009298	0	0	0	0	0	0	1	-360	360;
	7195	8977	0.000609	0.005058	0	0	0	0	0	0	1	-360	360;
	52	7195	0.008009	0.04	0	0	0	0	0	0	1	-360	360;
	7195	3868	0.002289	0.016218	0	0	0	0	0	0	1	-360	360;
	3363	392	0.003098	0.032729	0	0	0	0	0	0	1	-360	360;
	3363	3530	0.002982	0.013058	0	0	0	0	0	0	1	-360	360;
	3363	3530	0.00304	0.0128	0	0	0	0	0	0	1	-360	360;
	3363	5829	0.008791	0.059622	0	0	0	0	0	0	1	-360	360;
	8977	6539	0.000511	0.004409	0	0	0	0	0	0	1	-360	360;
	7675	6539	0.009391	0.030809	0	0	0	0	0	0	1	-360	360;
	7675	6539	0.009191	0.031391	0	260	0	0	0	0	1	-360	360;
	1702	2900	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	6202	1702	0.00276	0.014782	0	0	0	0	0	0	1	-360	360;
	6202	1702	0.002809	0.014458	0	0	0	0	0	0	1	-360	360;
	1106	1702	0.003138	0.017502	0	0	0	0	0	0	1	-360	360;
	1106	1702	0.0032	0.017138	0	0	0	0	0	0	1	-360	360;
	5916	6342	0.002033	0.005273	0	0	0	0	0	0	1	-360	360;
	4302	6342	0.036066	0.160107	0	0	0	0	0	0	1	-360	360;
	733	8917	0.027686	0.04614	0	0	0	0	0	0	1	-360	360;
	5423	8917	0.025107	0.051372	0	0	0	0	0	0	1	-360	360;
	1777	548	0.002479	0.01086	0	0	0	0	0	0	1	-360	360;
	548	6923	0.013603	0.051876	0	0	0	0	0	0	1	-360	360;
	9029	8508	0.00176	0.108	0	0	0	0	0	0	1	-360	360;
	1591	5558	0	0.00047	0	987	0	0	0	0	1	-360	360;
	1591	8604	0.00038	0.00366	0	1744	0	0	0	0	1	-360	360;
	1591	8264	0.00052	0.00633	0	1909	0	0	0	0	1	-360	360;
	1124	8510	0	0.00051	0	0	0	0	0	0	1	-360	360;
	1124	3257	0.001599	0.015599	0	0	0	0	0	0	1	-360	360;
	5311	1124	0.00212	0.022349	0	0	0	0	0	0	1	-360	360;
	5558	1788	0.00034	0.00415	0	1645	0	0	0	0	1	-360	360;
	8211	5558	0.00051	0.00646	0	1744	0	0	0	0	1	-360	360;
	8510	369	0.00163	0.015271	0	0	0	0	0	0	1	-360	360;
	3891	8510	0.002159	0.021901	0	0	0	0	0	0	1	-360	360;
	5952	784	0.00224	0.015018	0	0	0	0	0	0	1	-360	360;
	5952	3182	0.004111	0.00652	0	0	0	0	0	0	1	-360	360;
	5952	3182	0.004191	0.006391	0	0	0	0	0	0	1	-360	360;
	6923	5108	0.00238	0.006744	0	0	0	0	0	0	1	-360	360;
	5829	392	0.003329	0.033129	0	0	0	0	0	0	1	-360	360;
	1633	8389	0.005809	0.021182	0	0	0	0	0	0	1	-360	360;
	1633	8389	0.005689	0.021591	0	0	0	0	0	0	1	-360	360;
	8389	2720	0.00132	0.01052	0	0	0	0	0	0	1	-360	360;
	8389	243	0.001129	0.012329	0	0	0	0	0	0	1	-360	360;
	8389	4409	0.001569	0.1224	0	0	0	0	0	0	1	-360	360;
	3119	2572	0.000662	0.036351	0	468	0	0	0	0	1	-360	360;
	4318	7404	0.001777	0.009107	0	0	0	0	0	0	1	-360	360;
	4318	1338	0.002967	0.002901	0	0	0	0	0	0	1	-360	360;
	4318	8288	0.017893	0.057479	0	0	0	0	0	0	1	-360	360;
	7404	8882	0.000554	0.001091	0	0	0	0	0	0	1	-360	360;
	7404	7565	0.006339	0.022603	0	0	0	0	0	0	1	-360	360;
	1338	360	0.004174	0.006702	0	0	0	0	0	0	1	-360	360;
	360	7565	0.009818	0.040099	0	0	0	0	0	0	1	-360	360;
	275	6323	0.003729	0.01876	0	0	0	0	0	0	1	-360	360;
	5872	7230	0.000782	0.059938	0	0	0	0	0	0	1	-360	360;
	8194	5872	0.000769	0.061111	0	0	0	0	0	0	1	-360	360;
	1020	6514	0.000382	0.019378	0	0	0	0	0	0	1	-360	360;
	1020	387	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1106	6838	1.8e-05	0.00188	0	637	0	0	0	0	1	-360	360;
	8480	6025	0.000372	0.001223	0	0	0	0	0	0	1	-360	360;
	9029	5835	0.000809	0.001751	0	0	0	0	0	0	1	-360	360;
	3374	878	0	0.043511	0	0	0	0	0	0	1	-360	360;
	7450	3374	0.003529	0.022178	0	0	0	0	0	0	1	-360	360;
	7450	3374	0.0036	0.021711	0	0	0	0	0	0	1	-360	360;
	6383	231	0	0.000413	0	0	0	0	0	0	1	-360	360;
	2228	6383	0.047479	0.102603	0	0	0	0	0	0	1	-360	360;
	2228	6383	0.053727	0.091281	0	0	0	0	0	0	1	-360	360;
	1010	6383	0.021678	0.045669	0	0	0	0	0	0	1	-360	360;
	2228	8878	0.004397	0.030678	0	0	0	0	0	0	1	-360	360;
	2228	8878	0.004314	0.031256	0	0	0	0	0	0	1	-360	360;
	2228	7117	0.008347	0.055512	0	0	0	0	0	0	1	-360	360;
	2228	2698	0.006223	0.039744	0	0	0	0	0	0	1	-360	360;
	8277	2228	0.006124	0.041686	0	0	0	0	0	0	1	-360	360;
	8277	2228	0.006942	0.040851	0	0	0	0	0	0	1	-360	360;
	4057	5783	0.019248	0.039041	0	191	0	0	0	0	1	-360	360;
	7136	4057	0.015719	0.04943	0	0	0	0	0	0	1	-360	360;
	5549	1207	0.00296	0.011871	0	0	0	0	0	0	1	-360	360;
	5549	9083	0.01328	0.04144	0	0	0	0	0	0	1	-360	360;
	2745	7138	0.00736	0.033222	0	0	0	0	0	0	1	-360	360;
	2745	7138	0.007502	0.032529	0	0	0	0	0	0	1	-360	360;
	1670	2745	0.00192	0.02576	0	299	0	0	0	0	1	-360	360;
	1670	2745	0.002169	0.02524	0	299	0	0	0	0	1	-360	360;
	2745	8492	0.00656	0.017071	0	0	0	0	0	0	1	-360	360;
	2745	2748	0.002662	0.007471	0	0	0	0	0	0	1	-360	360;
	2745	3215	0.003191	0.009062	0	0	0	0	0	0	1	-360	360;
	6915	8668	0.00981	0.057298	0	0	0	0	0	0	1	-360	360;
	3177	6356	0.008342	0.037031	0	0	0	0	0	0	1	-360	360;
	1464	6356	0.0086	0.039791	0	0	0	0	0	0	1	-360	360;
	1689	8230	0	0.00046	0	1843	0	0	0	0	1	-360	360;
	6068	1689	0.00055	0.00685	0	1645	0	0	0	0	1	-360	360;
	6263	8230	0.00054	0.00699	0	1645	0	0	0	0	1	-360	360;
	8814	69	0.0056	0.052889	0	0	0	0	0	0	1	-360	360;
	8814	5949	0.006058	0.023591	0	0	0	0	0	0	1	-360	360;
	8814	5949	0.0054	0.026729	0	0	0	0	0	0	1	-360	360;
	8814	8613	0.00168	0.01188	0	299	0	0	0	0	1	-360	360;
	8814	9007	0.003178	0.028658	0	0	0	0	0	0	1	-360	360;
	8814	5188	0.003178	0.028542	0	299	0	0	0	0	1	-360	360;
	2077	6442	0.00068	0.004449	0	0	0	0	0	0	1	-360	360;
	2077	3462	0.00076	0.004818	0	0	0	0	0	0	1	-360	360;
	2812	1778	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2812	1633	9e-06	0.000222	0	0	0	0	0	0	1	-360	360;
	3983	1633	0	0.000222	0	299	0	0	0	0	1	-360	360;
	7442	1633	1.8e-05	0.004329	0	0	0	0	0	0	1	-360	360;
	6047	5271	0.001831	0.001698	0	0	0	0	0	0	1	-360	360;
	6047	5271	0.002209	0.001471	0	0	0	0	0	0	1	-360	360;
	211	2832	0.017653	0.070289	0	0	0	0	0	0	1	-360	360;
	4132	2832	0.013231	0.042207	0	0	0	0	0	0	1	-360	360;
	4132	2832	0.013479	0.041331	0	0	0	0	0	0	1	-360	360;
	2804	2832	0.013934	0.055702	0	0	0	0	0	0	1	-360	360;
	4216	6028	0.000537	0.001471	0	0	0	0	0	0	1	-360	360;
	6028	4872	0.009967	0.048959	0	0	0	0	0	0	1	-360	360;
	6028	9146	0.030612	0.055058	0	0	0	0	0	0	1	-360	360;
	6028	9146	0.031207	0.053934	0	0	0	0	0	0	1	-360	360;
	6058	6028	0.01314	0.040579	0	0	0	0	0	0	1	-360	360;
	6058	6028	0.012851	0.03757	0	0	0	0	0	0	1	-360	360;
	6028	7011	0.006901	0.021174	0	0	0	0	0	0	1	-360	360;
	6028	7011	0.007653	0.01924	0	0	0	0	0	0	1	-360	360;
	7295	5595	0.008982	0.038462	0	0	0	0	0	0	1	-360	360;
	5595	7297	0.009151	0.037649	0	0	0	0	0	0	1	-360	360;
	5595	1020	0.004258	0.03952	0	0	0	0	0	0	1	-360	360;
	2616	5595	0.003489	0.0264	0	0	0	0	0	0	1	-360	360;
	5595	2305	0.003849	0.05232	0	0	0	0	0	0	1	-360	360;
	1989	4104	0.032669	0.080273	0	0	0	0	0	0	1	-360	360;
	1889	4104	0.004901	0.014967	0	0	0	0	0	0	1	-360	360;
	312	5783	0.000562	0.001058	0	0	0	0	0	0	1	-360	360;
	8393	139	0	0.00045	0	1547	0	0	0	0	1	-360	360;
	8393	8264	0.001	0.01169	0	1481	0	0	0	0	1	-360	360;
	139	8211	0.00098	0.01192	0	1448	0	0	0	0	1	-360	360;
	6600	7565	0.002339	0.006223	0	0	0	0	0	0	1	-360	360;
	6600	7565	0.002388	0.006091	0	0	0	0	0	0	1	-360	360;
	2555	6600	0.000504	0.001207	0	0	0	0	0	0	1	-360	360;
	5783	7565	0.00105	0.007959	0	248	0	0	0	0	1	-360	360;
	5783	7565	0.001033	0.007372	0	267	0	0	0	0	1	-360	360;
	5783	2648	0.018752	0.042339	0	0	0	0	0	0	1	-360	360;
	5783	2648	0.018339	0.039231	0	0	0	0	0	0	1	-360	360;
	1989	6923	0.036793	0.081719	0	0	0	0	0	0	1	-360	360;
	1989	8480	0.003702	0.015132	0	0	0	0	0	0	1	-360	360;
	2478	1989	0.00124	0.007314	0	219	0	0	0	0	1	-360	360;
	523	5760	0.002329	0.00356	0	299	0	0	0	0	1	-360	360;
	523	5760	0.002378	0.00348	0	299	0	0	0	0	1	-360	360;
	523	5760	0.00212	0.003951	0	299	0	0	0	0	1	-360	360;
	523	7476	0.001951	0.003791	0	364	0	0	0	0	1	-360	360;
	69	7852	0.002542	0.01472	0	0	0	0	0	0	1	-360	360;
	1670	5342	0.003262	0.024151	0	0	0	0	0	0	1	-360	360;
	9206	4757	0.0036	0.01548	0	260	0	0	0	0	1	-360	360;
	9206	4757	0.00352	0.014329	0	260	0	0	0	0	1	-360	360;
	1880	4380	0.002591	0.014418	0	0	0	0	0	0	1	-360	360;
	3530	1880	0.001071	0.003058	0	0	0	0	0	0	1	-360	360;
	3530	4687	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	3530	1330	0.017458	0.052129	0	0	0	0	0	0	1	-360	360;
	7295	8442	0.00136	0.00216	0	0	0	0	0	0	1	-360	360;
	7297	8442	0.001391	0.00212	0	0	0	0	0	0	1	-360	360;
	6399	9083	0.005858	0.016591	0	0	0	0	0	0	1	-360	360;
	898	8668	0.004562	0.011653	0	0	0	0	0	0	1	-360	360;
	898	6526	0.030868	0.064174	0	0	0	0	0	0	1	-360	360;
	3257	369	0	0.000521	0	0	0	0	0	0	1	-360	360;
	2381	1020	0.005929	0.053271	0	0	0	0	0	0	1	-360	360;
	2381	297	0.00664	0.040849	0	0	0	0	0	0	1	-360	360;
	2381	3182	0.00968	0.049382	0	0	0	0	0	0	1	-360	360;
	5829	4692	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	5829	69	0.00044	0.004489	0	0	0	0	0	0	1	-360	360;
	5829	3898	0.01064	0.043751	0	0	0	0	0	0	1	-360	360;
	5829	3898	0.0104	0.040511	0	0	0	0	0	0	1	-360	360;
	5188	5829	0.0024	0.025498	0	0	0	0	0	0	1	-360	360;
	4272	1205	0.00264	0.01412	0	0	0	0	0	0	1	-360	360;
	472	6131	0.000609	0.03944	0	663	0	0	0	0	1	-360	360;
	8565	472	0	0.000222	0	0	0	0	0	0	1	-360	360;
	472	5760	0.001182	0.003018	0	299	0	0	0	0	1	-360	360;
	472	5760	0.0012	0.002951	0	299	0	0	0	0	1	-360	360;
	472	5760	0.001071	0.003351	0	286	0	0	0	0	1	-360	360;
	472	7073	0.001031	0.01544	0	0	0	0	0	0	1	-360	360;
	472	349	0.001009	0.015738	0	0	0	0	0	0	1	-360	360;
	472	7156	0.000991	0.01604	0	0	0	0	0	0	1	-360	360;
	6202	7305	0.003898	0.008862	0	0	0	0	0	0	1	-360	360;
	6202	7305	0.003818	0.009031	0	0	0	0	0	0	1	-360	360;
	643	311	0.003831	0.012502	0	0	0	0	0	0	1	-360	360;
	6068	6263	0	0.00051	0	856	0	0	0	0	1	-360	360;
	6017	6263	0.00127	0.0132	0	1481	0	0	0	0	1	-360	360;
	7450	2271	0	0.000222	0	299	0	0	0	0	1	-360	360;
	7450	7666	0.001338	0.011209	0	0	0	0	0	0	1	-360	360;
	7450	7018	0.004222	0.030062	0	0	0	0	0	0	1	-360	360;
	8784	7450	0.001182	0.0104	0	0	0	0	0	0	1	-360	360;
	2271	3846	9e-06	0.000222	0	0	0	0	0	0	1	-360	360;
	2271	7018	0.004138	0.03064	0	0	0	0	0	0	1	-360	360;
	3846	7018	0.004049	0.031209	0	0	0	0	0	0	1	-360	360;
	733	5423	0.022603	0.051959	0	0	0	0	0	0	1	-360	360;
	4614	733	0.013636	0.032041	0	0	0	0	0	0	1	-360	360;
	8492	3216	0.00228	0.009222	0	0	0	0	0	0	1	-360	360;
	2748	3216	0.002471	0.010302	0	273	0	0	0	0	1	-360	360;
	3215	3216	0.001831	0.007378	0	273	0	0	0	0	1	-360	360;
	8878	4302	0.006843	0.049248	0	0	0	0	0	0	1	-360	360;
	8878	4302	0.006686	0.045603	0	0	0	0	0	0	1	-360	360;
	2760	8016	0.004102	0.011529	0	0	0	0	0	0	1	-360	360;
	2760	8016	0.004182	0.01128	0	0	0	0	0	0	1	-360	360;
	1913	2120	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8155	1913	0.003649	0.0122	0	0	0	0	0	0	1	-360	360;
	8155	1913	0.00372	0.011938	0	0	0	0	0	0	1	-360	360;
	1913	7090	0.005031	0.028102	0	0	0	0	0	0	1	-360	360;
	1913	7090	0.005129	0.02752	0	0	0	0	0	0	1	-360	360;
	1913	7844	0.005218	0.026951	0	0	0	0	0	0	1	-360	360;
	4413	3462	0.00288	0.01992	0	0	0	0	0	0	1	-360	360;
	4413	3462	0.003262	0.01952	0	0	0	0	0	0	1	-360	360;
	8738	1106	0.00956	0.061022	0	0	0	0	0	0	1	-360	360;
	8738	2760	0.005382	0.033911	0	0	0	0	0	0	1	-360	360;
	8738	2760	0.005271	0.034569	0	0	0	0	0	0	1	-360	360;
	8738	6218	0.004102	0.023018	0	0	0	0	0	0	1	-360	360;
	1788	8604	0	0.00052	0	757	0	0	0	0	1	-360	360;
	1020	2335	0	0.000222	0	481	0	0	0	0	1	-360	360;
	1020	3182	0.011631	0.094591	0	0	0	0	0	0	1	-360	360;
	2616	1020	0.0012	0.008769	0	299	0	0	0	0	1	-360	360;
	1020	9072	0.001218	0.008582	0	299	0	0	0	0	1	-360	360;
	7138	7586	0.002689	0.013818	0	0	0	0	0	0	1	-360	360;
	2648	8288	0.005	0.008934	0	0	0	0	0	0	1	-360	360;
	2824	140	0	0.000531	0	0	0	0	0	0	1	-360	360;
	2824	5311	0.001281	0.0158	0	0	0	0	0	0	1	-360	360;
	140	3891	0.00125	0.016099	0	0	0	0	0	0	1	-360	360;
	211	6526	0.010058	0.048	0	0	0	0	0	0	1	-360	360;
	211	6526	0.010248	0.047	0	0	0	0	0	0	1	-360	360;
	211	2804	0.004314	0.024537	0	0	0	0	0	0	1	-360	360;
	1670	3301	0.010791	0.042151	0	299	0	0	0	0	1	-360	360;
	344	1011	0.0006	0.040218	0	0	0	0	0	0	1	-360	360;
	243	8621	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3210	5674	0	0.00053	0	1711	0	0	0	0	1	-360	360;
	3210	4337	0	0.00054	0	1251	0	0	0	0	1	-360	360;
	1670	8204	0.00196	0.010249	0	0	0	0	0	0	1	-360	360;
	1670	1516	0.002182	0.010498	0	0	0	0	0	0	1	-360	360;
	5949	6480	0.000582	0.040991	0	0	0	0	0	0	1	-360	360;
	5949	5655	0.000569	0.04176	0	0	0	0	0	0	1	-360	360;
	1464	3177	0.003818	0.013209	0	0	0	0	0	0	1	-360	360;
	5949	6323	0.00428	0.023511	0	0	0	0	0	0	1	-360	360;
	5949	6323	0.004182	0.021769	0	0	0	0	0	0	1	-360	360;
	1956	4858	0	0.0005	0	1086	0	0	0	0	1	-360	360;
	1956	8264	0.00114	0.01661	0	1481	0	0	0	0	1	-360	360;
	4320	4248	0	0.000539	0	0	0	0	0	0	1	-360	360;
	4320	76	0.00119	0.014081	0	0	0	0	0	0	1	-360	360;
	4858	8211	0.00117	0.01626	0	1711	0	0	0	0	1	-360	360;
	4248	38	0.00106	0.01595	0	0	0	0	0	0	1	-360	360;
	4302	999	0.018058	0.065719	0	0	0	0	0	0	1	-360	360;
	4302	1010	0.012826	0.025058	0	0	0	0	0	0	1	-360	360;
	4302	1010	0.013331	0.024	0	0	0	0	0	0	1	-360	360;
	8204	1516	0.001089	0.003151	0	0	0	0	0	0	1	-360	360;
	5792	2428	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	344	1277	0.000631	0.054471	0	0	0	0	0	0	1	-360	360;
	344	2117	0.000622	0.055489	0	0	0	0	0	0	1	-360	360;
	344	8895	0.0006	0.051382	0	0	0	0	0	0	1	-360	360;
	4132	6526	0.017702	0.064331	0	0	0	0	0	0	1	-360	360;
	4132	1787	0.012314	0.04924	0	0	0	0	0	0	1	-360	360;
	2331	8380	0.001223	0.001248	0	0	0	0	0	0	1	-360	360;
	5296	8380	0.001471	0.00481	0	0	0	0	0	0	1	-360	360;
	999	8380	0.011521	0.03138	0	0	0	0	0	0	1	-360	360;
	9029	7915	0	0.000222	0	0	0	0	0	0	1	-360	360;
	9029	6149	0.000289	0.00056	0	0	0	0	0	0	1	-360	360;
	3968	3370	0.00034	0.01805	0	757	0	0	0	0	1	-360	360;
	7480	2720	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2150	6687	0.00035	0.01766	0	1053	0	0	0	0	1	-360	360;
	8956	4209	0.00035	0.01728	0	1053	0	0	0	0	1	-360	360;
	5065	8318	0.00031	0.01958	0	1876	0	0	0	0	1	-360	360;
	7270	3370	0.00031	0.01997	0	856	0	0	0	0	1	-360	360;
	6424	3370	0.0003	0.02035	0	757	0	0	0	0	1	-360	360;
	5888	8223	0.00016	0.0025	0	1020	0	0	0	0	1	-360	360;
	4776	3370	0.00029	0.02074	0	1119	0	0	0	0	1	-360	360;
	1331	8318	0.00029	0.0192	0	757	0	0	0	0	1	-360	360;
	8483	8318	0.00033	0.01882	0	757	0	0	0	0	1	-360	360;
	6687	4209	0	0.00049	0	757	0	0	0	0	1	-360	360;
	4733	6687	0	0.00048	0	1283	0	0	0	0	1	-360	360;
	7856	4209	0	0.00047	0	1382	0	0	0	0	1	-360	360;
	3370	5897	0	0.00046	0	1349	0	0	0	0	1	-360	360;
	3370	420	0	0.00045	0	1086	0	0	0	0	1	-360	360;
	3370	8898	0	0.00051	0	757	0	0	0	0	1	-360	360;
	8318	7856	0	0.00052	0	987	0	0	0	0	1	-360	360;
	8318	4733	0	0.00053	0	921	0	0	0	0	1	-360	360;
	8318	8898	0	0.00054	0	757	0	0	0	0	1	-360	360;
	5897	4087	0	0.0005	0	1448	0	0	0	0	1	-360	360;
	4087	7856	0	0.00049	0	1843	0	0	0	0	1	-360	360;
	4087	4077	0.00028	0.00403	0	1382	0	0	0	0	1	-360	360;
	6283	4087	0.00021	0.00396	0	1777	0	0	0	0	1	-360	360;
	420	2927	0	0.00048	0	1119	0	0	0	0	1	-360	360;
	2927	4733	0	0.00047	0	1251	0	0	0	0	1	-360	360;
	2927	5888	0.00029	0.00394	0	954	0	0	0	0	1	-360	360;
	4485	2927	0.0002	0.00404	0	1744	0	0	0	0	1	-360	360;
	2720	243	0	0.000222	0	0	0	0	0	0	1	-360	360;
	243	558	0	0.000222	0	0	0	0	0	0	1	-360	360;
	8492	2748	0.004471	0.020791	0	0	0	0	0	0	1	-360	360;
	8492	3215	0.002791	0.01132	0	0	0	0	0	0	1	-360	360;
	5760	8392	0.000822	0.001649	0	0	0	0	0	0	1	-360	360;
	5760	8392	0.0008	0.00168	0	0	0	0	0	0	1	-360	360;
	5760	8392	0.000782	0.00156	0	0	0	0	0	0	1	-360	360;
	370	311	0.00432	0.012662	0	0	0	0	0	0	1	-360	360;
	370	311	0.003849	0.014351	0	0	0	0	0	0	1	-360	360;
	1106	2552	0.001418	0.004618	0	0	0	0	0	0	1	-360	360;
	1106	7275	0.001689	0.011631	0	0	0	0	0	0	1	-360	360;
	1106	6218	0.005991	0.03832	0	0	0	0	0	0	1	-360	360;
	9206	1936	0.001169	0.007098	0	403	0	0	0	0	1	-360	360;
	7476	1936	0.0012	0.00696	0	299	0	0	0	0	1	-360	360;
	4415	9146	0.030066	0.046289	0	0	0	0	0	0	1	-360	360;
	4415	9146	0.026793	0.052463	0	0	0	0	0	0	1	-360	360;
	7729	4415	0.023413	0.057397	0	0	0	0	0	0	1	-360	360;
	5591	5364	0.001631	0.00832	0	0	0	0	0	0	1	-360	360;
	6047	4431	0.012569	0.04732	0	0	0	0	0	0	1	-360	360;
	6047	4431	0.012311	0.048231	0	0	0	0	0	0	1	-360	360;
	8017	6519	0	0.000222	0	0	0	0	0	0	1	-360	360;
	7018	8017	0.000982	0.004751	0	0	0	0	0	0	1	-360	360;
	2965	1260	0.00144	0.12	0	0	0	0	0	0	1	-360	360;
	344	5364	0.001249	0.0064	0	0	0	0	0	0	1	-360	360;
	344	5364	0.00112	0.007249	0	0	0	0	0	0	1	-360	360;
	5311	3891	0	0.0005	0	0	0	0	0	0	1	-360	360;
	9205	8668	0.022843	0.098893	0	0	0	0	0	0	1	-360	360;
	8668	6526	0.034909	0.072893	0	0	0	0	0	0	1	-360	360;
	7666	377	0.00188	0.092222	0	0	0	0	0	0	1	-360	360;
	7666	1227	0.00184	0.093991	0	0	0	0	0	0	1	-360	360;
	7666	4006	0.0018	0.09576	0	0	0	0	0	0	1	-360	360;
	5872	7666	0.00132	0.01088	0	0	0	0	0	0	1	-360	360;
	6058	1889	0.01924	0.056603	0	0	0	0	0	0	1	-360	360;
	6058	1889	0.017124	0.064149	0	0	0	0	0	0	1	-360	360;
	4272	6747	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	4272	838	0.006889	0.03404	0	0	0	0	0	0	1	-360	360;
	4272	838	0.007018	0.033329	0	0	0	0	0	0	1	-360	360;
	4272	7951	0.006169	0.035689	0	0	0	0	0	0	1	-360	360;
	4272	7951	0.00628	0.034929	0	0	0	0	0	0	1	-360	360;
	4272	8792	0.004271	0.026418	0	0	0	0	0	0	1	-360	360;
	4272	8792	0.00436	0.02588	0	0	0	0	0	0	1	-360	360;
	4272	8792	0.00444	0.025338	0	0	0	0	0	0	1	-360	360;
	3898	1330	0.00852	0.02444	0	0	0	0	0	0	1	-360	360;
	8613	52	0.002671	0.018929	0	0	0	0	0	0	1	-360	360;
	662	3352	0	0.00046	0	1613	0	0	0	0	1	-360	360;
	662	6766	0	0.00045	0	1481	0	0	0	0	1	-360	360;
	6017	3352	0	0.00051	0	1185	0	0	0	0	1	-360	360;
	6017	6766	0	0.00052	0	1382	0	0	0	0	1	-360	360;
	2552	7275	0.000329	0.000902	0	0	0	0	0	0	1	-360	360;
	8792	3712	0.00068	0.050351	0	0	0	0	0	0	1	-360	360;
	3388	838	0	0.000222	0	0	0	0	0	0	1	-360	360;
	1362	2523	0	0.00049	0	0	0	0	0	0	1	-360	360;
	1362	1052	0.00212	0.022531	0	553	0	0	0	0	1	-360	360;
	1362	76	0.000919	0.012789	0	0	0	0	0	0	1	-360	360;
	2523	3878	0.002159	0.02206	0	438	0	0	0	0	1	-360	360;
	2523	38	0.00094	0.012531	0	0	0	0	0	0	1	-360	360;
	194	7018	0.00032	0.014102	0	299	0	0	0	0	1	-360	360;
	4846	7018	0.000329	0.0138	0	299	0	0	0	0	1	-360	360;
	3691	7018	0.000329	0.013502	0	299	0	0	0	0	1	-360	360;
	6323	539	0.005818	0.02396	0	0	0	0	0	0	1	-360	360;
	6323	539	0.005942	0.023462	0	0	0	0	0	0	1	-360	360;
	5872	6707	0.000671	0.001871	0	260	0	0	0	0	1	-360	360;
	4863	6707	0	0.001791	0	0	0	0	0	0	1	-360	360;
	5872	3509	0.002249	0.003471	0	0	0	0	0	0	1	-360	360;
	5872	3509	0.001631	0.0042	0	0	0	0	0	0	1	-360	360;
	5872	3776	0.00092	0.002582	0	0	0	0	0	0	1	-360	360;
	7018	3509	0	0.000222	0	585	0	0	0	0	1	-360	360;
	7018	5974	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	8784	7018	0.000822	0.00572	0	0	0	0	0	0	1	-360	360;
	7018	4565	0.001338	0.013622	0	0	0	0	0	0	1	-360	360;
	8058	7018	0.00136	0.004649	0	0	0	0	0	0	1	-360	360;
	3509	4863	0	0.001751	0	0	0	0	0	0	1	-360	360;
	3776	3509	0	0.001711	0	0	0	0	0	0	1	-360	360;
	4082	6442	0.000662	0.004529	0	273	0	0	0	0	1	-360	360;
	5246	6442	0.000649	0.004618	0	0	0	0	0	0	1	-360	360;
	6526	1787	0.005669	0.028992	0	0	0	0	0	0	1	-360	360;
	4616	7260	0.001631	0.1176	0	0	0	0	0	0	1	-360	360;
	1221	7996	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	344	1221	0.006942	0.04516	0	0	0	0	0	0	1	-360	360;
	344	1221	0.0068	0.04604	0	0	0	0	0	0	1	-360	360;
	8525	2698	0.003289	0.017909	0	0	0	0	0	0	1	-360	360;
	3215	7407	0.00056	0.038671	0	0	0	0	0	0	1	-360	360;
	4686	7729	0.02162	0.062612	0	0	0	0	0	0	1	-360	360;
	52	2494	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	52	9007	0.000822	0.00656	0	0	0	0	0	0	1	-360	360;
	52	928	0.003009	0.004649	0	273	0	0	0	0	1	-360	360;
	52	928	0.003071	0.004551	0	286	0	0	0	0	1	-360	360;
	6987	4509	0.00464	0.029431	0	0	0	0	0	0	1	-360	360;
	6987	4509	0.005249	0.02884	0	0	0	0	0	0	1	-360	360;
	3119	6047	0.001902	0.012702	0	0	0	0	0	0	1	-360	360;
	3119	6047	0.001858	0.012938	0	0	0	0	0	0	1	-360	360;
	3119	7662	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6149	5442	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6275	3819	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	344	6015	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	9083	994	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	9206	3169	9.8e-05	0.000222	0	260	0	0	0	0	1	-360	360;
	9206	838	0.004738	0.012529	0	0	0	0	0	0	1	-360	360;
	9206	838	0.004831	0.012262	0	0	0	0	0	0	1	-360	360;
	9206	838	0.002711	0.018431	0	0	0	0	0	0	1	-360	360;
	5423	9223	0.01205	0.03281	0	0	0	0	0	0	1	-360	360;
	3314	838	0	0.000222	0	0	0	0	0	0	1	-360	360;
	838	7779	0.000649	0.03712	0	390	0	0	0	0	1	-360	360;
	838	4477	0	0.000222	0	0	0	0	0	0	1	-360	360;
	6320	838	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4774	3620	0.00076	0.005329	0	0	0	0	0	0	1	-360	360;
	3182	297	0.005769	0.03528	0	0	0	0	0	0	1	-360	360;
	7476	1463	0.002258	0.004849	0	0	0	0	0	0	1	-360	360;
	7476	1463	0.002209	0.004489	0	260	0	0	0	0	1	-360	360;
	1052	3878	0	0.000479	0	0	0	0	0	0	1	-360	360;
	1052	6477	0.00256	0.01561	0	533	0	0	0	0	1	-360	360;
	3878	5295	0.00269	0.01451	0	591	0	0	0	0	1	-360	360;
	8277	5887	0.003372	0.028231	0	0	0	0	0	0	1	-360	360;
	8277	9127	0.007033	0.047083	0	0	0	0	0	0	1	-360	360;
	76	38	0	0.000469	0	0	0	0	0	0	1	-360	360;
	3182	1984	9.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	2760	3182	0.00392	0.025191	0	0	0	0	0	0	1	-360	360;
	2760	3182	0.00444	0.024689	0	0	0	0	0	0	1	-360	360;
	9127	5887	0.003529	0.021223	0	0	0	0	0	0	1	-360	360;
	4082	5246	0	0.000222	0	0	0	0	0	0	1	-360	360;
	4485	395	0.00069	0.00134	0	1481	0	0	0	0	1	-360	360;
	5843	6283	0.00067	0.00136	0	1448	0	0	0	0	1	-360	360;
	2478	8480	0.012992	0.070207	0	0	0	0	0	0	1	-360	360;
	395	5843	0	0.00053	0	823	0	0	0	0	1	-360	360;
	6477	5295	0	0.000461	0	0	0	0	0	0	1	-360	360;
	8264	8211	0	0.00054	0	790	0	0	0	0	1	-360	360;
	7011	4614	0.002314	0.005661	0	0	0	0	0	0	1	-360	360;
	4614	7011	0.002512	0.005909	0	0	0	0	0	0	1	-360	360;
	6058	7011	0.006364	0.018488	0	0	0	0	0	0	1	-360	360;
	6058	7011	0.005959	0.017752	0	0	0	0	0	0	1	-360	360;
	6058	9223	0.018397	0.04176	0	0	0	0	0	0	1	-360	360;
	6058	9223	0.018058	0.042653	0	0	0	0	0	0	1	-360	360;
	7090	7844	0	0.000222	0	0	0	0	0	0	1	-360	360;
	2616	559	0	0.000529	0	0	0	0	0	0	1	-360	360;
	9072	559	0	0.00052	0	0	0	0	0	0	1	-360	360;
	2305	9072	0.003369	0.031329	0	338	0	0	0	0	1	-360	360;
	2509	6268	0.07449	0.527603	0	0	0	0	0	0	1	-360	360;
	3039	6268	0.01093	0.074531	0	438	0	0	0	0	1	-360	360;
	5148	6268	0.00887	0.04576	0	553	0	0	0	0	1	-360	360;
	3909	6268	0.01099	0.06689	0	0	0	0	0	0	1	-360	360;
	8610	1050	0.04732	0.42186	0	0	0	0	0	0	1	-360	360;
	8610	1050	0.004469	0.03124	0	0	0	0	0	0	1	-360	360;
	1050	431	0.500289	1.586157	0	0	0	0	0	0	1	-360	360;
	1050	3953	0.01089	0.060719	0	457	0	0	0	0	1	-360	360;
	1050	4805	0.133469	0.882645	0	0	0	0	0	0	1	-360	360;
	8610	1050	0.005531	0.032521	0	476	0	0	0	0	1	-360	360;
	1050	4805	0.005219	0.025659	0	495	0	0	0	0	1	-360	360;
	8652	3848	0.00186	0.012789	0	400	0	0	0	0	1	-360	360;
	3991	3848	0.001	0.234938	0	0	0	0	0	0	1	-360	360;
	5075	6701	0.00081	0.004781	0	0	0	0	0	0	1	-360	360;
	2862	8175	0.00082	0.00468	0	0	0	0	0	0	1	-360	360;
	4581	4595	0.03374	0.614711	0	0	0	0	0	0	1	-360	360;
	8537	4595	0.04612	0.366777	0	0	0	0	0	0	1	-360	360;
	5158	4595	0.001461	0.01239	0	0	0	0	0	0	1	-360	360;
	8537	4581	0.078599	0.394649	0	381	0	0	0	0	1	-360	360;
	8758	4581	0.00143	0.01263	0	0	0	0	0	0	1	-360	360;
	7293	977	0.007841	0.05613	0	553	0	0	0	0	1	-360	360;
	5310	2453	0	0.0002	0	0	0	0	0	0	1	-360	360;
	5310	2453	0.001	0.328946	0	572	0	0	0	0	1	-360	360;
	3616	2453	0.224855	1.532376	0	0	0	0	0	0	1	-360	360;
	734	2453	0.136781	1.457355	0	0	0	0	0	0	1	-360	360;
	734	2453	0.01055	0.074979	0	0	0	0	0	0	1	-360	360;
	1481	2453	0.014909	0.523802	0	0	0	0	0	0	1	-360	360;
	1837	2453	0.001	0.71781	0	0	0	0	0	0	1	-360	360;
	7551	2453	0.134271	1.687004	0	0	0	0	0	0	1	-360	360;
	4163	5310	0.002341	0.01263	0	457	0	0	0	0	1	-360	360;
	3616	5310	0.235227	1.541777	0	0	0	0	0	0	1	-360	360;
	734	5310	0.16344	1.266136	0	0	0	0	0	0	1	-360	360;
	1481	5310	0.015479	0.562376	0	0	0	0	0	0	1	-360	360;
	1837	5310	0.001	0.767789	0	0	0	0	0	0	1	-360	360;
	7551	5310	0.158281	1.596364	0	0	0	0	0	0	1	-360	360;
	5863	3438	0.006169	0.039289	0	553	0	0	0	0	1	-360	360;
	6100	3438	0.08256	0.638182	0	0	0	0	0	0	1	-360	360;
	2771	3438	0.221653	1.097397	0	0	0	0	0	0	1	-360	360;
	3438	6708	0.0118	0.05445	0	381	0	0	0	0	1	-360	360;
	3003	3438	0.08986	0.756921	0	0	0	0	0	0	1	-360	360;
	4002	638	0.03887	0.335083	0	0	0	0	0	0	1	-360	360;
	638	4002	0.00455	0.036539	0	457	0	0	0	0	1	-360	360;
	173	638	0.002169	0.01688	0	0	0	0	0	0	1	-360	360;
	173	638	0.01045	0.244731	0	0	0	0	0	0	1	-360	360;
	6601	638	0.001	0.259525	0	0	0	0	0	0	1	-360	360;
	3607	638	0.036651	0.378657	0	0	0	0	0	0	1	-360	360;
	6591	6287	0.004669	0.03839	0	0	0	0	0	0	1	-360	360;
	587	6287	0.000781	0.00674	0	629	0	0	0	0	1	-360	360;
	3616	6287	0.001461	0.01201	0	0	0	0	0	0	1	-360	360;
	9234	6287	0.00087	0.00614	0	591	0	0	0	0	1	-360	360;
	8971	4163	0.00464	0.02919	0	419	0	0	0	0	1	-360	360;
	1837	4163	0.004719	0.02719	0	476	0	0	0	0	1	-360	360;
	5656	4163	0.00169	0.01361	0	476	0	0	0	0	1	-360	360;
	7984	4163	0.00751	0.04581	0	553	0	0	0	0	1	-360	360;
	1342	6591	0.00368	0.022469	0	0	0	0	0	0	1	-360	360;
	5656	6591	0.001659	0.012599	0	495	0	0	0	0	1	-360	360;
	7551	6591	0.007341	0.042419	0	533	0	0	0	0	1	-360	360;
	6591	7511	0.00186	0.01456	0	514	0	0	0	0	1	-360	360;
	6238	3368	0.00312	0.03151	0	1152	0	0	0	0	1	-360	360;
	8822	6238	0.00073	0.00808	0	1283	0	0	0	0	1	-360	360;
	7174	3001	0.013479	0.27064	0	0	0	0	0	0	1	-360	360;
	7174	3001	0.00187	0.020289	0	0	0	0	0	0	1	-360	360;
	9040	3001	0.012169	0.298636	0	0	0	0	0	0	1	-360	360;
	7293	6937	0.00337	0.027831	0	591	0	0	0	0	1	-360	360;
	3909	7293	0.00461	0.03301	0	533	0	0	0	0	1	-360	360;
	6937	7293	0.005781	0.039789	0	533	0	0	0	0	1	-360	360;
	6937	7293	0.035151	0.987397	0	0	0	0	0	0	1	-360	360;
	2284	7293	0.013099	0.631736	0	0	0	0	0	0	1	-360	360;
	3039	2509	0.118409	1.072417	0	0	0	0	0	0	1	-360	360;
	3039	2509	0.006281	0.051961	0	457	0	0	0	0	1	-360	360;
	2063	2509	0.284979	1.491054	0	0	0	0	0	0	1	-360	360;
	4002	3251	0.006409	0.038081	0	629	0	0	0	0	1	-360	360;
	173	4002	0.225806	1.262624	0	0	0	0	0	0	1	-360	360;
	6601	4002	0.282252	1.121529	0	0	0	0	0	0	1	-360	360;
	3607	4002	0.561921	1.55469	0	0	0	0	0	0	1	-360	360;
	5479	3797	0.00322	0.02808	0	1547	0	0	0	0	1	-360	360;
	214	3797	0.00196	0.01843	0	1283	0	0	0	0	1	-360	360;
	7025	8449	0.552066	1.574607	0	0	0	0	0	0	1	-360	360;
	3326	8449	0.010151	0.087021	0	514	0	0	0	0	1	-360	360;
	7298	8449	0.16999	0.991054	0	0	0	0	0	0	1	-360	360;
	262	8449	0.003831	0.02386	0	553	0	0	0	0	1	-360	360;
	262	8449	0.00863	0.535785	0	0	0	0	0	0	1	-360	360;
	6169	8449	0.04357	0.617335	0	0	0	0	0	0	1	-360	360;
	5732	2827	0.008081	0.03963	0	438	0	0	0	0	1	-360	360;
	2827	1837	0.00213	0.01488	0	438	0	0	0	0	1	-360	360;
	214	4533	0.00219	0.02232	0	1218	0	0	0	0	1	-360	360;
	4533	1315	0.00352	0.03947	0	987	0	0	0	0	1	-360	360;
	4533	8401	0.00222	0.02857	0	1481	0	0	0	0	1	-360	360;
	361	5822	0.00104	0.00901	0	1283	0	0	0	0	1	-360	360;
	2758	361	0.00106	0.00881	0	1448	0	0	0	0	1	-360	360;
	6634	4051	0.00301	0.031271	0	0	0	0	0	0	1	-360	360;
	5822	3147	0.00133	0.01852	0	1349	0	0	0	0	1	-360	360;
	3147	3930	0.00253	0.0287	0	954	0	0	0	0	1	-360	360;
	5822	3147	0.06395	0.413802	0	757	0	0	0	0	1	-360	360;
	1315	9238	0.00075	0.00981	0	1909	0	0	0	0	1	-360	360;
	7890	2274	0.234814	1.148326	0	381	0	0	0	0	1	-360	360;
	1507	2274	0.222417	1.185971	0	0	0	0	0	0	1	-360	360;
	8764	2274	0.08124	0.548574	0	419	0	0	0	0	1	-360	360;
	8764	2274	0.01132	0.074211	0	438	0	0	0	0	1	-360	360;
	764	2274	0.115849	0.953099	0	0	0	0	0	0	1	-360	360;
	5462	4257	0	0.0002	0	1283	0	0	0	0	1	-360	360;
	5462	4257	0.001	0.517348	0	987	0	0	0	0	1	-360	360;
	4257	7553	0.00051	0.00504	0	1382	0	0	0	0	1	-360	360;
	4257	5209	0.00258	0.02748	0	1481	0	0	0	0	1	-360	360;
	993	5462	0.00224	0.02852	0	1481	0	0	0	0	1	-360	360;
	5462	2758	0.00144	0.01774	0	1415	0	0	0	0	1	-360	360;
	5640	8387	0	0.0002	0	0	0	0	0	0	1	-360	360;
	5640	8387	0.06794	0.65593	0	0	0	0	0	0	1	-360	360;
	8387	7174	0.028479	0.136841	0	476	0	0	0	0	1	-360	360;
	5640	4051	0.02019	0.10464	0	438	0	0	0	0	1	-360	360;
	7553	2758	0.00128	0.01333	0	1613	0	0	0	0	1	-360	360;
	328	6829	0.667066	1.759628	0	0	0	0	0	0	1	-360	360;
	4844	6829	0.0232	0.385289	0	0	0	0	0	0	1	-360	360;
	4844	6829	0.00686	0.03882	0	553	0	0	0	0	1	-360	360;
	7854	6829	0.04143	0.328037	0	0	0	0	0	0	1	-360	360;
	9215	6829	0.00107	0.00707	0	457	0	0	0	0	1	-360	360;
	4844	328	0.173021	1.092025	0	0	0	0	0	0	1	-360	360;
	4844	328	0.010849	0.074771	0	629	0	0	0	0	1	-360	360;
	578	328	0.18632	0.723574	0	0	0	0	0	0	1	-360	360;
	7707	328	0.077651	0.853368	0	0	0	0	0	0	1	-360	360;
	7707	328	0.011169	0.07225	0	476	0	0	0	0	1	-360	360;
	7854	328	0.14268	0.506488	0	0	0	0	0	0	1	-360	360;
	7298	328	0.352417	1.255207	0	438	0	0	0	0	1	-360	360;
	262	328	0.391322	1.63	0	0	0	0	0	0	1	-360	360;
	3219	2771	0.006271	0.037341	0	553	0	0	0	0	1	-360	360;
	3607	3219	0.005961	0.054719	0	572	0	0	0	0	1	-360	360;
	3515	4695	0	0.0002	0	0	0	0	0	0	1	-360	360;
	3515	4695	0.001	0.333822	0	0	0	0	0	0	1	-360	360;
	4695	3720	0.00587	0.040281	0	419	0	0	0	0	1	-360	360;
	3515	9040	0.003599	0.046841	0	0	0	0	0	0	1	-360	360;
	2948	5209	0.00224	0.02796	0	1415	0	0	0	0	1	-360	360;
	2771	6100	0.08693	0.670269	0	0	0	0	0	0	1	-360	360;
	3616	6100	0.301839	1.820372	0	0	0	0	0	0	1	-360	360;
	1481	6100	0.00136	0.01051	0	495	0	0	0	0	1	-360	360;
	1481	6100	0.00445	0.344835	0	0	0	0	0	0	1	-360	360;
	8330	6100	0.197159	1.014752	0	0	0	0	0	0	1	-360	360;
	5732	6100	0.097341	0.78186	0	0	0	0	0	0	1	-360	360;
	3003	6100	0.013039	0.297211	0	438	0	0	0	0	1	-360	360;
	5863	6100	0.035521	0.685351	0	0	0	0	0	0	1	-360	360;
	3991	6100	0.367335	1.746467	0	0	0	0	0	0	1	-360	360;
	3084	6100	0.04186	0.505517	0	0	0	0	0	0	1	-360	360;
	1481	6100	0.001021	0.284401	0	0	0	0	0	0	1	-360	360;
	6708	6100	0.02575	0.548884	0	0	0	0	0	0	1	-360	360;
	7511	6100	0.002211	0.016599	0	495	0	0	0	0	1	-360	360;
	5139	8652	0.0015	0.00876	0	0	0	0	0	0	1	-360	360;
	3991	5139	0.00062	0.00386	0	514	0	0	0	0	1	-360	360;
	6708	8971	0.00188	0.008651	0	457	0	0	0	0	1	-360	360;
	3084	8971	0.00651	0.278678	0	0	0	0	0	0	1	-360	360;
	5656	8971	0.080781	0.659587	0	0	0	0	0	0	1	-360	360;
	5369	6861	0	0.0002	0	0	0	0	0	0	1	-360	360;
	5369	6861	0.001	0.232851	0	0	0	0	0	0	1	-360	360;
	6861	3015	0.01614	0.08307	0	438	0	0	0	0	1	-360	360;
	431	6861	0.190479	0.901095	0	0	0	0	0	0	1	-360	360;
	431	6861	0.00281	0.01663	0	514	0	0	0	0	1	-360	360;
	3953	6861	0.032841	0.593161	0	0	0	0	0	0	1	-360	360;
	8537	6861	0.004599	0.03094	0	476	0	0	0	0	1	-360	360;
	4805	6861	0.100021	0.542231	0	0	0	0	0	0	1	-360	360;
	431	5369	0.1945	0.883884	0	0	0	0	0	0	1	-360	360;
	3953	5369	0.028409	0.617727	0	0	0	0	0	0	1	-360	360;
	5369	5158	0.003281	0.04162	0	0	0	0	0	0	1	-360	360;
	4805	5369	0.09636	0.566219	0	0	0	0	0	0	1	-360	360;
	4805	5369	0.007521	0.044719	0	0	0	0	0	0	1	-360	360;
	9040	7174	0	0.0002	0	0	0	0	0	0	1	-360	360;
	9040	7174	0.001	0.19056	0	591	0	0	0	0	1	-360	360;
	7174	3326	0.01356	0.08876	0	553	0	0	0	0	1	-360	360;
	9040	3326	0.005271	0.068591	0	0	0	0	0	0	1	-360	360;
	3404	8652	0.003591	0.03075	0	629	0	0	0	0	1	-360	360;
	5158	8652	0.01	0.056159	0	400	0	0	0	0	1	-360	360;
	2013	8652	0.011831	0.067969	0	591	0	0	0	0	1	-360	360;
	8758	8652	0.00937	0.05893	0	629	0	0	0	0	1	-360	360;
	8652	7493	0.000909	0.004719	0	495	0	0	0	0	1	-360	360;
	2216	1315	0.00176	0.01759	0	1448	0	0	0	0	1	-360	360;
	2997	2216	0.00137	0.01801	0	1514	0	0	0	0	1	-360	360;
	2216	7625	0.00184	0.01756	0	1283	0	0	0	0	1	-360	360;
	6708	3084	0.00213	0.012539	0	457	0	0	0	0	1	-360	360;
	5656	3084	0.07664	0.597665	0	0	0	0	0	0	1	-360	360;
	1481	3084	0.07132	0.544959	0	0	0	0	0	0	1	-360	360;
	1481	3084	0.00405	0.02507	0	495	0	0	0	0	1	-360	360;
	6708	3084	0.001	0.348058	0	0	0	0	0	0	1	-360	360;
	3003	2771	0.06064	0.52688	0	0	0	0	0	0	1	-360	360;
	2771	3003	0.005271	0.03093	0	514	0	0	0	0	1	-360	360;
	3404	1342	0.002841	0.023409	0	400	0	0	0	0	1	-360	360;
	6972	2013	0.00226	0.015841	0	457	0	0	0	0	1	-360	360;
	6166	2013	0.01237	0.060909	0	419	0	0	0	0	1	-360	360;
	2013	3404	0.01637	0.07345	0	495	0	0	0	0	1	-360	360;
	3996	2013	0.00243	0.018341	0	553	0	0	0	0	1	-360	360;
	3454	791	0.0026	0.02813	0	1185	0	0	0	0	1	-360	360;
	5075	2862	0	0.0002	0	0	0	0	0	0	1	-360	360;
	5075	2862	0.02631	0.515517	0	572	0	0	0	0	1	-360	360;
	2862	6937	0.00394	0.041169	0	0	0	0	0	0	1	-360	360;
	2862	6937	0.343719	1.744855	0	0	0	0	0	0	1	-360	360;
	2862	7984	0.017159	0.080591	0	457	0	0	0	0	1	-360	360;
	5075	5513	0.005	0.066591	0	438	0	0	0	0	1	-360	360;
	5075	6937	0.065229	1.288988	0	0	0	0	0	0	1	-360	360;
	734	3616	0	0.0002	0	0	0	0	0	0	1	-360	360;
	734	3616	0.001	0.56	0	610	0	0	0	0	1	-360	360;
	1481	3616	0.063599	0.651508	0	0	0	0	0	0	1	-360	360;
	8330	3616	0.075159	1.271095	0	0	0	0	0	0	1	-360	360;
	5732	3616	0.360475	1.617479	0	0	0	0	0	0	1	-360	360;
	1837	3616	0.001	0.740702	0	0	0	0	0	0	1	-360	360;
	3003	3616	0.363884	2.00314	0	0	0	0	0	0	1	-360	360;
	5863	3616	0.3925	1.858533	0	0	0	0	0	0	1	-360	360;
	1481	734	0.04769	0.602128	0	0	0	0	0	0	1	-360	360;
	8330	734	0.050341	1.259855	0	0	0	0	0	0	1	-360	360;
	1837	734	0.001	0.829566	0	0	0	0	0	0	1	-360	360;
	3326	8610	0.00049	0.00324	0	533	0	0	0	0	1	-360	360;
	8610	3782	0.018659	0.111289	0	514	0	0	0	0	1	-360	360;
	8610	4805	0.509835	1.831384	0	0	0	0	0	0	1	-360	360;
	3326	8610	0.0005	0.00318	0	572	0	0	0	0	1	-360	360;
	6708	1481	0.05936	0.582149	0	0	0	0	0	0	1	-360	360;
	3003	1481	0.003331	0.0203	0	629	0	0	0	0	1	-360	360;
	8330	1481	0.12286	0.888244	0	0	0	0	0	0	1	-360	360;
	5732	1481	0.033599	0.599959	0	0	0	0	0	0	1	-360	360;
	1837	1481	0.100539	0.767066	0	0	0	0	0	0	1	-360	360;
	3003	1481	0.0822	0.730331	0	0	0	0	0	0	1	-360	360;
	5863	1481	0.15512	0.82874	0	0	0	0	0	0	1	-360	360;
	5305	1492	0	0.0002	0	0	0	0	0	0	1	-360	360;
	5305	1492	0.001	0.194781	0	0	0	0	0	0	1	-360	360;
	1492	2539	0.0132	0.071979	0	476	0	0	0	0	1	-360	360;
	1492	378	0.0148	0.073031	0	495	0	0	0	0	1	-360	360;
	5148	1492	0.005781	0.0378	0	533	0	0	0	0	1	-360	360;
	6166	1492	0.153961	1.590021	0	0	0	0	0	0	1	-360	360;
	6166	1492	0.00631	0.03855	0	495	0	0	0	0	1	-360	360;
	8924	1492	0.14124	1.705702	0	0	0	0	0	0	1	-360	360;
	5148	5305	0.00589	0.03699	0	533	0	0	0	0	1	-360	360;
	6166	5305	0.18032	1.449298	0	0	0	0	0	0	1	-360	360;
	8924	5305	0.150651	1.729298	0	0	0	0	0	0	1	-360	360;
	8924	5305	0.00636	0.045409	0	438	0	0	0	0	1	-360	360;
	2835	3762	0.00173	0.01602	0	1316	0	0	0	0	1	-360	360;
	4638	2835	0.00207	0.02233	0	1218	0	0	0	0	1	-360	360;
	1443	2835	0.00178	0.02178	0	1283	0	0	0	0	1	-360	360;
	3996	6972	0.00589	0.214174	0	0	0	0	0	0	1	-360	360;
	6166	6972	0.105789	1.66719	0	0	0	0	0	0	1	-360	360;
	8924	6972	0.108969	1.629731	0	0	0	0	0	0	1	-360	360;
	7854	4844	0.024581	0.231446	0	0	0	0	0	0	1	-360	360;
	7854	4844	0.004599	0.035341	0	572	0	0	0	0	1	-360	360;
	3930	1103	0.00106	0.01023	0	757	0	0	0	0	1	-360	360;
	1110	1103	0.00403	0.04889	0	1810	0	0	0	0	1	-360	360;
	6190	5505	0.00123	0.01589	0	1415	0	0	0	0	1	-360	360;
	6190	3368	0.00158	0.01592	0	1909	0	0	0	0	1	-360	360;
	248	7089	0.00219	0.02123	0	1580	0	0	0	0	1	-360	360;
	248	4244	0.00171	0.01708	0	1547	0	0	0	0	1	-360	360;
	248	4244	0.00174	0.0167	0	1580	0	0	0	0	1	-360	360;
	1110	248	0.00515	0.285942	0	757	0	0	0	0	1	-360	360;
	1110	248	0.00058	0.00775	0	1678	0	0	0	0	1	-360	360;
	1110	248	0.00057	0.0079	0	1415	0	0	0	0	1	-360	360;
	3404	6708	0.009341	0.6725	0	0	0	0	0	0	1	-360	360;
	3991	6708	0.171229	1.013678	0	0	0	0	0	0	1	-360	360;
	7511	6708	0.00251	0.02049	0	400	0	0	0	0	1	-360	360;
	6708	7493	0.005909	0.036349	0	591	0	0	0	0	1	-360	360;
	9027	7625	0.00091	0.00886	0	1251	0	0	0	0	1	-360	360;
	8822	9027	0.00054	0.00532	0	1547	0	0	0	0	1	-360	360;
	262	7025	0.009531	0.051211	0	476	0	0	0	0	1	-360	360;
	262	7025	0.785124	1.943678	0	0	0	0	0	0	1	-360	360;
	7025	3787	0.010789	0.05194	0	514	0	0	0	0	1	-360	360;
	6169	7025	0.173281	0.62781	0	0	0	0	0	0	1	-360	360;
	5656	3404	0.00613	0.033659	0	400	0	0	0	0	1	-360	360;
	3991	3404	0.11576	0.840434	0	0	0	0	0	0	1	-360	360;
	7707	578	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7707	578	0.05964	0.314938	0	0	0	0	0	0	1	-360	360;
	578	1267	0.002169	0.017341	0	438	0	0	0	0	1	-360	360;
	2604	578	0.454215	1.265475	0	0	0	0	0	0	1	-360	360;
	578	2604	0.007091	0.061521	0	533	0	0	0	0	1	-360	360;
	7707	1719	0.013021	0.087711	0	591	0	0	0	0	1	-360	360;
	966	7707	0.35781	1.954277	0	0	0	0	0	0	1	-360	360;
	2604	7707	0.0485	0.815537	0	0	0	0	0	0	1	-360	360;
	1837	5513	0.415537	1.939483	0	0	0	0	0	0	1	-360	360;
	5656	5513	0.097031	0.677913	0	0	0	0	0	0	1	-360	360;
	5513	5656	0.003031	0.033021	0	457	0	0	0	0	1	-360	360;
	5513	5656	0.00637	0.038979	0	572	0	0	0	0	1	-360	360;
	7551	5513	0.024271	0.588967	0	0	0	0	0	0	1	-360	360;
	7984	5513	0.00439	0.372438	0	0	0	0	0	0	1	-360	360;
	212	5513	0.001	0.198331	0	0	0	0	0	0	1	-360	360;
	212	5513	0.00232	0.01226	0	381	0	0	0	0	1	-360	360;
	3787	1719	0.014211	0.079219	0	629	0	0	0	0	1	-360	360;
	173	1719	0.015151	0.07637	0	533	0	0	0	0	1	-360	360;
	3251	1719	0	0.0002	0	0	0	0	0	0	1	-360	360;
	3251	1719	0.001	0.258698	0	0	0	0	0	0	1	-360	360;
	6601	1719	0.01544	0.074781	0	553	0	0	0	0	1	-360	360;
	966	3251	0.013909	0.073349	0	629	0	0	0	0	1	-360	360;
	5386	4468	0.006919	0.04764	0	438	0	0	0	0	1	-360	360;
	2250	4468	0.224318	1.85469	0	0	0	0	0	0	1	-360	360;
	8764	4468	0.015169	0.1077	0	476	0	0	0	0	1	-360	360;
	378	4468	0.130771	1.06564	0	0	0	0	0	0	1	-360	360;
	5386	4468	0.249669	1.169917	0	0	0	0	0	0	1	-360	360;
	3762	499	0.00127	0.01444	0	1382	0	0	0	0	1	-360	360;
	3762	6109	0.00318	0.04235	0	1283	0	0	0	0	1	-360	360;
	3762	8685	0.00131	0.01758	0	1283	0	0	0	0	1	-360	360;
	2284	3039	0.469607	1.302025	0	0	0	0	0	0	1	-360	360;
	2063	3039	0.001	0.205711	0	0	0	0	0	0	1	-360	360;
	2063	3039	0.002969	0.02201	0	514	0	0	0	0	1	-360	360;
	9215	7854	0.00549	0.044831	0	495	0	0	0	0	1	-360	360;
	3411	2250	0	0.0002	0	0	0	0	0	0	1	-360	360;
	3411	2250	0.08343	1.583099	0	591	0	0	0	0	1	-360	360;
	2539	2250	0.094651	1.363719	0	0	0	0	0	0	1	-360	360;
	2250	3015	0.00112	0.00895	0	553	0	0	0	0	1	-360	360;
	3015	2250	0.00151	0.00944	0	553	0	0	0	0	1	-360	360;
	378	2250	0.018669	0.563368	0	0	0	0	0	0	1	-360	360;
	5386	2250	0.00538	0.652851	0	0	0	0	0	0	1	-360	360;
	2539	3411	0	0.0002	0	0	0	0	0	0	1	-360	360;
	2539	3411	0.001	0.495455	0	591	0	0	0	0	1	-360	360;
	378	3411	0.036849	0.734215	0	0	0	0	0	0	1	-360	360;
	5386	3411	0.001	0.459566	0	0	0	0	0	0	1	-360	360;
	5386	3411	0.00095	0.00638	0	533	0	0	0	0	1	-360	360;
	3015	2539	0.00105	0.00818	0	514	0	0	0	0	1	-360	360;
	378	2539	0.040151	0.657893	0	0	0	0	0	0	1	-360	360;
	378	2539	0.00193	0.01406	0	514	0	0	0	0	1	-360	360;
	5386	2539	0.001	0.49438	0	0	0	0	0	0	1	-360	360;
	5732	8330	0	0.0002	0	0	0	0	0	0	1	-360	360;
	5732	8330	0.08057	0.906178	0	572	0	0	0	0	1	-360	360;
	3003	8330	0.0347	0.658202	0	0	0	0	0	0	1	-360	360;
	5863	8330	0.032719	0.534318	0	0	0	0	0	0	1	-360	360;
	5863	8330	0.004281	0.033461	0	553	0	0	0	0	1	-360	360;
	3003	5732	0.001	0.395785	0	381	0	0	0	0	1	-360	360;
	5863	5732	0.14007	0.838595	0	0	0	0	0	0	1	-360	360;
	8685	499	0.04194	0.384211	0	757	0	0	0	0	1	-360	360;
	499	8685	0.00034	0.00341	0	1415	0	0	0	0	1	-360	360;
	770	7890	0.001969	0.014021	0	0	0	0	0	0	1	-360	360;
	1507	7890	0	0.0002	0	0	0	0	0	0	1	-360	360;
	1507	7890	0.001	0.26219	0	0	0	0	0	0	1	-360	360;
	7890	3938	0.03032	0.14413	0	400	0	0	0	0	1	-360	360;
	1507	764	0.01325	0.11038	0	476	0	0	0	0	1	-360	360;
	2573	3015	0.011901	0.07319	0	476	0	0	0	0	1	-360	360;
	3015	3782	0.00401	0.03175	0	495	0	0	0	0	1	-360	360;
	8685	2997	0.00368	0.03712	0	1580	0	0	0	0	1	-360	360;
	8685	2948	0.00169	0.02158	0	1283	0	0	0	0	1	-360	360;
	791	4638	0.00116	0.01472	0	1283	0	0	0	0	1	-360	360;
	214	7089	0.05575	0.602438	0	757	0	0	0	0	1	-360	360;
	7089	214	0.00163	0.01604	0	1382	0	0	0	0	1	-360	360;
	7089	2997	0.00269	0.02759	0	1251	0	0	0	0	1	-360	360;
	7089	1315	0.00256	0.02877	0	1316	0	0	0	0	1	-360	360;
	764	8764	0	0.0002	0	0	0	0	0	0	1	-360	360;
	764	8764	0.09549	0.768967	0	610	0	0	0	0	1	-360	360;
	3953	431	0	0.0002	0	0	0	0	0	0	1	-360	360;
	3953	431	0.05899	0.428306	0	0	0	0	0	0	1	-360	360;
	5158	431	0.005591	0.04031	0	610	0	0	0	0	1	-360	360;
	4805	431	0.00714	0.15039	0	0	0	0	0	0	1	-360	360;
	8758	3953	0.005469	0.041099	0	591	0	0	0	0	1	-360	360;
	4805	3953	0.036711	0.380269	0	0	0	0	0	0	1	-360	360;
	4847	214	0.00056	0.124578	0	757	0	0	0	0	1	-360	360;
	4847	214	0.0007	0.00889	0	1645	0	0	0	0	1	-360	360;
	3326	2573	0.013479	0.11436	0	533	0	0	0	0	1	-360	360;
	3938	3326	0.012	0.066919	0	553	0	0	0	0	1	-360	360;
	3782	2573	0.3925	1.654463	0	0	0	0	0	0	1	-360	360;
	3996	6937	0.301178	1.948719	0	0	0	0	0	0	1	-360	360;
	3996	6937	0.009169	0.06301	0	514	0	0	0	0	1	-360	360;
	3368	791	0.00176	0.02342	0	856	0	0	0	0	1	-360	360;
	3676	791	0.00108	0.01167	0	1481	0	0	0	0	1	-360	360;
	3782	378	0.191961	1.234649	0	0	0	0	0	0	1	-360	360;
	5386	378	0.03699	0.374917	0	0	0	0	0	0	1	-360	360;
	8758	8537	0.00361	0.03024	0	610	0	0	0	0	1	-360	360;
	262	7298	0.350806	1.554649	0	0	0	0	0	0	1	-360	360;
	7298	262	0.004831	0.06375	0	0	0	0	0	0	1	-360	360;
	7298	2553	0.007099	0.07076	0	0	0	0	0	0	1	-360	360;
	6169	262	0.0022	0.01318	0	381	0	0	0	0	1	-360	360;
	262	966	0.01775	0.099419	0	438	0	0	0	0	1	-360	360;
	6169	262	0.186581	1.067335	0	0	0	0	0	0	1	-360	360;
	2997	2948	0.00373	0.05005	0	1316	0	0	0	0	1	-360	360;
	3787	966	0	0.0002	0	0	0	0	0	0	1	-360	360;
	3787	966	0.001	0.16038	0	0	0	0	0	0	1	-360	360;
	2604	966	0.010531	0.17819	0	0	0	0	0	0	1	-360	360;
	2604	966	0.0025	0.017081	0	495	0	0	0	0	1	-360	360;
	2604	3787	0.0777	0.498492	0	0	0	0	0	0	1	-360	360;
	7551	1837	0.01907	1.485992	0	0	0	0	0	0	1	-360	360;
	7984	1837	0.120669	1.890888	0	0	0	0	0	0	1	-360	360;
	8924	5148	0.010271	0.06219	0	400	0	0	0	0	1	-360	360;
	6166	3996	0.298079	1.623781	0	0	0	0	0	0	1	-360	360;
	8924	3996	0.273223	1.612107	0	0	0	0	0	0	1	-360	360;
	2997	8822	0.00277	0.02823	0	1448	0	0	0	0	1	-360	360;
	1315	2997	0.00077	0.00747	0	1251	0	0	0	0	1	-360	360;
	7625	8401	0.00102	0.01145	0	757	0	0	0	0	1	-360	360;
	7625	8401	0.00112	0.01049	0	757	0	0	0	0	1	-360	360;
	8924	6166	0	0.0002	0	0	0	0	0	0	1	-360	360;
	8924	6166	0.001	0.22343	0	0	0	0	0	0	1	-360	360;
	5505	3676	0	0.0002	0	1810	0	0	0	0	1	-360	360;
	5505	3676	0.001	0.26178	0	757	0	0	0	0	1	-360	360;
	3368	3676	0.00153	0.0214	0	823	0	0	0	0	1	-360	360;
	8822	3676	0.00468	0.06143	0	987	0	0	0	0	1	-360	360;
	212	5656	0.271756	1.50345	0	0	0	0	0	0	1	-360	360;
	7984	7551	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7984	7551	0.015271	0.796736	0	572	0	0	0	0	1	-360	360;
	212	7551	0.02776	0.935537	0	0	0	0	0	0	1	-360	360;
	212	7551	0.004219	0.026599	0	400	0	0	0	0	1	-360	360;
	212	7984	0.001	0.620888	0	0	0	0	0	0	1	-360	360;
	5386	3782	0.361364	1.973368	0	0	0	0	0	0	1	-360	360;
	3909	2284	0.001599	0.00825	0	476	0	0	0	0	1	-360	360;
	2063	2284	0.659194	1.990062	0	0	0	0	0	0	1	-360	360;
	6601	173	0	0.0002	0	0	0	0	0	0	1	-360	360;
	6601	173	0.11788	0.822417	0	0	0	0	0	0	1	-360	360;
	3607	173	0.086341	0.745145	0	0	0	0	0	0	1	-360	360;
	3607	6601	0.00569	0.033281	0	533	0	0	0	0	1	-360	360;
	3607	6601	0.03618	0.609897	0	0	0	0	0	0	1	-360	360;
	27	4847	0.00087	0.00898	0	1711	0	0	0	0	1	-360	360;
	141	3720	0.001	0.211322	0	629	0	0	0	0	1	-360	360;
	141	3720	0	0.0002	0	0	0	0	0	0	1	-360	360;
	3938	3720	0.13407	1.029876	0	0	0	0	0	0	1	-360	360;
	1835	3720	0.141419	0.956033	0	0	0	0	0	0	1	-360	360;
	3938	141	0.12913	1.151054	0	0	0	0	0	0	1	-360	360;
	1835	141	0.14487	1.008244	0	0	0	0	0	0	1	-360	360;
	141	1835	0.007599	0.05356	0	0	0	0	0	0	1	-360	360;
	5863	3003	0	0.0002	0	0	0	0	0	0	1	-360	360;
	5863	3003	0.024219	0.488657	0	572	0	0	0	0	1	-360	360;
	1835	3938	0.001	0.226901	0	0	0	0	0	0	1	-360	360;
	1835	3938	0	0.0002	0	0	0	0	0	0	1	-360	360;
	7493	3991	0.00063	0.005159	0	495	0	0	0	0	1	-360	360;
	4051	2553	0	0.0002	0	0	0	0	0	0	1	-360	360;
	4051	2553	0.001	0.214277	0	0	0	0	0	0	1	-360	360;
	3341	4259	0.019671	0.07004	0	0	0	0	0	0	1	-360	360;
	3341	4259	0.01812	0.06872	0	0	0	0	0	0	1	-360	360;
	5805	4721	0.001582	0.00724	0	0	0	0	0	0	1	-360	360;
	5805	4721	0.001551	0.007231	0	0	0	0	0	0	1	-360	360;
	5805	4721	0.001262	0.00796	0	0	0	0	0	0	1	-360	360;
	5452	2734	0.0264	0.075129	0	0	0	0	0	0	1	-360	360;
	5452	5274	0.00824	0.044751	0	0	0	0	0	0	1	-360	360;
	7879	4721	0.018271	0.070191	0	0	0	0	0	0	1	-360	360;
	7879	4721	0.021711	0.063129	0	0	0	0	0	0	1	-360	360;
	7879	7569	0.013191	0.0428	0	0	0	0	0	0	1	-360	360;
	8098	2734	0.027631	0.102502	0	0	0	0	0	0	1	-360	360;
	8098	7358	0.005369	0.052622	0	0	0	0	0	0	1	-360	360;
	8098	7358	0.013938	0.0534	0	0	0	0	0	0	1	-360	360;
	8098	7358	0.01364	0.049449	0	0	0	0	0	0	1	-360	360;
	8098	1407	0.007991	0.037738	0	0	0	0	0	0	1	-360	360;
	2233	4721	0.007671	0.04096	0	0	0	0	0	0	1	-360	360;
	2233	4721	0.007511	0.041769	0	0	0	0	0	0	1	-360	360;
	2233	3429	0.004089	0.010978	0	0	0	0	0	0	1	-360	360;
	2233	3429	0.003	0.014769	0	0	0	0	0	0	1	-360	360;
	2233	610	0.001351	0.00784	0	0	0	0	0	0	1	-360	360;
	2233	948	0.003671	0.022089	0	0	0	0	0	0	1	-360	360;
	2233	2954	0.002378	0.00772	0	0	0	0	0	0	1	-360	360;
	4721	610	0.006911	0.03268	0	0	0	0	0	0	1	-360	360;
	4721	2140	0.00944	0.03284	0	0	0	0	0	0	1	-360	360;
	4721	6512	0.009831	0.02848	0	0	0	0	0	0	1	-360	360;
	4721	948	0.003671	0.02092	0	0	0	0	0	0	1	-360	360;
	8789	2734	0.019378	0.10932	0	0	0	0	0	0	1	-360	360;
	8789	2734	0.021969	0.107129	0	0	0	0	0	0	1	-360	360;
	8789	5274	0.009151	0.063249	0	0	0	0	0	0	1	-360	360;
	8789	4116	0.029169	0.088102	0	0	0	0	0	0	1	-360	360;
	8789	4868	0.031351	0.115698	0	0	0	0	0	0	1	-360	360;
	8789	7358	0.04936	0.182018	0	0	0	0	0	0	1	-360	360;
	8789	6329	0.001609	0.016889	0	299	0	0	0	0	1	-360	360;
	8789	2817	0	0.000222	0	0	0	0	0	0	1	-360	360;
	702	8927	0.008698	0.033591	0	0	0	0	0	0	1	-360	360;
	5517	1862	0.01792	0.070169	0	0	0	0	0	0	1	-360	360;
	5517	8629	0.00308	0.008422	0	299	0	0	0	0	1	-360	360;
	5517	552	0.002031	0.012102	0	0	0	0	0	0	1	-360	360;
	5517	552	0.0068	0.03452	0	0	0	0	0	0	1	-360	360;
	5517	3791	0.01176	0.055871	0	0	0	0	0	0	1	-360	360;
	4344	6512	0.04856	0.135111	0	0	0	0	0	0	1	-360	360;
	2734	6188	0.039289	0.146129	0	0	0	0	0	0	1	-360	360;
	2734	6188	0.038471	0.148938	0	0	0	0	0	0	1	-360	360;
	2734	3413	0.010911	0.04116	0	0	0	0	0	0	1	-360	360;
	6188	1407	0.002898	0.017409	0	0	0	0	0	0	1	-360	360;
	5006	1250	0.00544	0.051431	0	0	0	0	0	0	1	-360	360;
	5006	5331	0.003791	0.014342	0	0	0	0	0	0	1	-360	360;
	5006	5927	0.005969	0.026302	0	0	0	0	0	0	1	-360	360;
	5006	4771	0.00608	0.02572	0	0	0	0	0	0	1	-360	360;
	5006	2660	0.010818	0.066769	0	260	0	0	0	0	1	-360	360;
	1862	2407	0.006649	0.024138	0	0	0	0	0	0	1	-360	360;
	1862	2407	0.007542	0.023662	0	0	0	0	0	0	1	-360	360;
	1862	2407	0.007991	0.036622	0	0	0	0	0	0	1	-360	360;
	1862	2407	0.0116	0.046018	0	0	0	0	0	0	1	-360	360;
	4116	2140	0.011631	0.047049	0	0	0	0	0	0	1	-360	360;
	4868	7358	0.018822	0.064009	0	0	0	0	0	0	1	-360	360;
	1250	5927	0.008071	0.049138	0	0	0	0	0	0	1	-360	360;
	1250	4771	0.007889	0.045502	0	0	0	0	0	0	1	-360	360;
	1250	8278	0.009822	0.057222	0	0	0	0	0	0	1	-360	360;
	2407	3711	0.00656	0.023831	0	0	0	0	0	0	1	-360	360;
	2407	3711	0.007431	0.023351	0	0	0	0	0	0	1	-360	360;
	2407	7411	0.007391	0.026978	0	0	0	0	0	0	1	-360	360;
	2407	4277	0.001871	0.012142	0	0	0	0	0	0	1	-360	360;
	2407	987	0.003969	0.026169	0	0	0	0	0	0	1	-360	360;
	2407	2206	0.008151	0.04664	0	299	0	0	0	0	1	-360	360;
	8197	39	0.00076	0.005622	0	0	0	0	0	0	1	-360	360;
	8197	3711	0.0094	0.035791	0	0	0	0	0	0	1	-360	360;
	8197	7792	0.00844	0.031031	0	0	0	0	0	0	1	-360	360;
	8197	1685	0.00028	0.002062	0	0	0	0	0	0	1	-360	360;
	8197	2894	0.004849	0.030031	0	0	0	0	0	0	1	-360	360;
	8197	4277	0.00268	0.022391	0	0	0	0	0	0	1	-360	360;
	8197	987	0.002391	0.025382	0	0	0	0	0	0	1	-360	360;
	552	3711	0.017551	0.078191	0	0	0	0	0	0	1	-360	360;
	6504	1543	0.003849	0.02475	0	591	0	0	0	0	1	-360	360;
	6504	7041	0.0042	0.02005	0	0	0	0	0	0	1	-360	360;
	6504	3585	0.00011	0.000581	0	0	0	0	0	0	1	-360	360;
	6504	780	0.000169	0.000651	0	0	0	0	0	0	1	-360	360;
	6504	3957	0.000169	0.000659	0	0	0	0	0	0	1	-360	360;
	2692	6254	0.00219	0.01125	0	495	0	0	0	0	1	-360	360;
	2692	8957	0.00018	0.000599	0	495	0	0	0	0	1	-360	360;
	2692	8861	0.00018	0.000591	0	400	0	0	0	0	1	-360	360;
	2692	1734	0.00019	0.000581	0	438	0	0	0	0	1	-360	360;
	2692	3266	0.000169	0.000651	0	495	0	0	0	0	1	-360	360;
	322	1974	0.00231	0.0138	0	457	0	0	0	0	1	-360	360;
	322	1974	0.00082	0.003841	0	0	0	0	0	0	1	-360	360;
	322	177	0.00305	0.01762	0	495	0	0	0	0	1	-360	360;
	5662	5333	0.01101	0.055469	0	476	0	0	0	0	1	-360	360;
	5662	7901	0.009169	0.0462	0	533	0	0	0	0	1	-360	360;
	5662	7041	0.00201	0.00905	0	0	0	0	0	0	1	-360	360;
	5662	2006	0.005969	0.038841	0	553	0	0	0	0	1	-360	360;
	5662	7761	0.00474	0.02687	0	438	0	0	0	0	1	-360	360;
	5662	5559	0.005409	0.03438	0	476	0	0	0	0	1	-360	360;
	5662	3211	0.005401	0.03264	0	572	0	0	0	0	1	-360	360;
	5662	8963	0.00311	0.015919	0	610	0	0	0	0	1	-360	360;
	5662	1770	0.00462	0.02613	0	419	0	0	0	0	1	-360	360;
	3186	1109	0.01044	0.05187	0	476	0	0	0	0	1	-360	360;
	3186	1109	0.01001	0.05113	0	476	0	0	0	0	1	-360	360;
	3186	8493	0.008331	0.051729	0	533	0	0	0	0	1	-360	360;
	3186	6280	0.000169	0.00089	0	438	0	0	0	0	1	-360	360;
	3186	3987	0.00214	0.01099	0	495	0	0	0	0	1	-360	360;
	1974	5825	0.004669	0.042349	0	476	0	0	0	0	1	-360	360;
	1974	7252	0.00139	0.008789	0	419	0	0	0	0	1	-360	360;
	1974	7252	0.000219	0.001669	0	495	0	0	0	0	1	-360	360;
	1974	177	0.00119	0.01063	0	0	0	0	0	0	1	-360	360;
	2410	2006	0.004169	0.02724	0	514	0	0	0	0	1	-360	360;
	2410	2006	0.004081	0.025219	0	495	0	0	0	0	1	-360	360;
	2410	315	0.00168	0.01039	0	533	0	0	0	0	1	-360	360;
	2410	1489	0.0017	0.01055	0	495	0	0	0	0	1	-360	360;
	8926	5797	0.00139	0.008771	0	495	0	0	0	0	1	-360	360;
	8926	6381	0.000651	0.00261	0	476	0	0	0	0	1	-360	360;
	8926	6381	0.000729	0.00256	0	476	0	0	0	0	1	-360	360;
	8926	1239	0	0.008169	0	400	0	0	0	0	1	-360	360;
	8926	5643	5e-05	0.00025	0	438	0	0	0	0	1	-360	360;
	5672	7901	0.003719	0.022669	0	514	0	0	0	0	1	-360	360;
	5672	6415	0.00732	0.042711	0	476	0	0	0	0	1	-360	360;
	5672	6415	0.003599	0.025331	0	0	0	0	0	0	1	-360	360;
	5672	8998	0.00494	0.027401	0	533	0	0	0	0	1	-360	360;
	5333	6795	0.0062	0.03519	0	533	0	0	0	0	1	-360	360;
	5333	6795	0.00739	0.04574	0	438	0	0	0	0	1	-360	360;
	5333	3211	0.006781	0.034289	0	572	0	0	0	0	1	-360	360;
	6795	3054	0.003669	0.03205	0	0	0	0	0	0	1	-360	360;
	6795	3054	0.002789	0.026289	0	0	0	0	0	0	1	-360	360;
	6795	2562	0.005979	0.031979	0	533	0	0	0	0	1	-360	360;
	3054	6254	0.00425	0.02132	0	553	0	0	0	0	1	-360	360;
	3054	1543	0.002469	0.015099	0	0	0	0	0	0	1	-360	360;
	3054	8957	0.004341	0.02662	0	514	0	0	0	0	1	-360	360;
	3054	3266	0.005219	0.02287	0	610	0	0	0	0	1	-360	360;
	3054	7615	0.0095	0.04693	0	495	0	0	0	0	1	-360	360;
	5825	177	0.01099	0.052909	0	457	0	0	0	0	1	-360	360;
	5825	6415	0.004901	0.02949	0	533	0	0	0	0	1	-360	360;
	5825	4744	0.00832	0.038531	0	457	0	0	0	0	1	-360	360;
	5825	5669	0.003909	0.018091	0	381	0	0	0	0	1	-360	360;
	5825	3416	0.000331	0.00189	0	476	0	0	0	0	1	-360	360;
	7901	6415	0.01101	0.05514	0	495	0	0	0	0	1	-360	360;
	7901	7503	0.00345	0.01925	0	457	0	0	0	0	1	-360	360;
	8459	177	0.00162	0.00889	0	0	0	0	0	0	1	-360	360;
	8459	8998	0.00039	0.00464	0	0	0	0	0	0	1	-360	360;
	8459	5361	0.00064	0.010099	0	0	0	0	0	0	1	-360	360;
	6254	5797	0.003151	0.01613	0	572	0	0	0	0	1	-360	360;
	6254	5797	0.002091	0.012409	0	629	0	0	0	0	1	-360	360;
	6254	9136	0	0.00799	0	0	0	0	0	0	1	-360	360;
	6254	4650	0	0.00906	0	0	0	0	0	0	1	-360	360;
	6254	7037	0.000909	0.010401	0	0	0	0	0	0	1	-360	360;
	6254	2497	0.003211	0.016081	0	591	0	0	0	0	1	-360	360;
	6254	8861	0.00314	0.017711	0	457	0	0	0	0	1	-360	360;
	6254	1734	0.00214	0.010591	0	476	0	0	0	0	1	-360	360;
	3478	3211	0.00056	0.003169	0	0	0	0	0	0	1	-360	360;
	7252	7236	0.001719	0.0108	0	514	0	0	0	0	1	-360	360;
	5797	3954	0.0005	0.00174	0	514	0	0	0	0	1	-360	360;
	5797	5643	0.00126	0.00587	0	419	0	0	0	0	1	-360	360;
	6960	2562	0.007909	0.03607	0	533	0	0	0	0	1	-360	360;
	6960	7037	0.00057	0.00457	0	0	0	0	0	0	1	-360	360;
	6960	5643	0.00132	0.008409	0	400	0	0	0	0	1	-360	360;
	177	2460	0.003581	0.017289	0	610	0	0	0	0	1	-360	360;
	177	4744	0.002711	0.01526	0	476	0	0	0	0	1	-360	360;
	177	6311	0.00039	0.006081	0	0	0	0	0	0	1	-360	360;
	177	5361	0.00037	0.005409	0	0	0	0	0	0	1	-360	360;
	177	7236	0.00244	0.013789	0	476	0	0	0	0	1	-360	360;
	3954	2497	0.000349	0.00149	0	533	0	0	0	0	1	-360	360;
	2460	6415	0.00212	0.010531	0	591	0	0	0	0	1	-360	360;
	380	5339	0.003039	0.031151	0	0	0	0	0	0	1	-360	360;
	380	5669	0.001831	0.0187	0	0	0	0	0	0	1	-360	360;
	1848	3932	0.006599	0.043169	0	438	0	0	0	0	1	-360	360;
	1848	5094	0.006729	0.046599	0	438	0	0	0	0	1	-360	360;
	1848	3623	0.002159	0.010789	0	476	0	0	0	0	1	-360	360;
	1848	785	0.0022	0.01057	0	476	0	0	0	0	1	-360	360;
	1848	1036	0.003461	0.02049	0	0	0	0	0	0	1	-360	360;
	8021	698	0.00311	0.019469	0	495	0	0	0	0	1	-360	360;
	8021	628	0.01064	0.065969	0	495	0	0	0	0	1	-360	360;
	1109	315	0.003539	0.020479	0	495	0	0	0	0	1	-360	360;
	1109	1489	0.00357	0.020719	0	495	0	0	0	0	1	-360	360;
	1109	3932	0.00275	0.013831	0	457	0	0	0	0	1	-360	360;
	1109	5094	0.0028	0.013531	0	457	0	0	0	0	1	-360	360;
	6415	5669	0.005151	0.04851	0	0	0	0	0	0	1	-360	360;
	6415	8493	0.005401	0.025281	0	533	0	0	0	0	1	-360	360;
	6415	8493	0.0055	0.02475	0	514	0	0	0	0	1	-360	360;
	7503	5559	0.00093	0.00538	0	438	0	0	0	0	1	-360	360;
	7041	2006	0.010331	0.061849	0	553	0	0	0	0	1	-360	360;
	4712	3623	0.00686	0.032771	0	476	0	0	0	0	1	-360	360;
	4712	785	0.007	0.032081	0	476	0	0	0	0	1	-360	360;
	5339	4170	0.007729	0.038099	0	0	0	0	0	0	1	-360	360;
	4170	698	0.006581	0.035281	0	572	0	0	0	0	1	-360	360;
	4170	698	0.003711	0.02651	0	533	0	0	0	0	1	-360	360;
	4170	4333	0.00482	0.021979	0	572	0	0	0	0	1	-360	360;
	4170	4333	0.003979	0.026099	0	533	0	0	0	0	1	-360	360;
	2006	8963	0.0038	0.017271	0	553	0	0	0	0	1	-360	360;
	2006	7661	0.00626	0.03626	0	629	0	0	0	0	1	-360	360;
	8493	628	0.00538	0.02588	0	572	0	0	0	0	1	-360	360;
	8493	628	0.005479	0.025341	0	553	0	0	0	0	1	-360	360;
	130	3080	0.00134	0.01235	0	1218	0	0	0	0	1	-360	360;
	130	536	0.00127	0.01399	0	1349	0	0	0	0	1	-360	360;
	130	4393	0.00142	0.01374	0	1349	0	0	0	0	1	-360	360;
	7827	7969	0.00039	0.00365	0	1218	0	0	0	0	1	-360	360;
	7827	6760	0.00058	0.00507	0	1283	0	0	0	0	1	-360	360;
	1025	6298	0.00129	0.01577	0	1349	0	0	0	0	1	-360	360;
	1025	3007	0.00033	0.00373	0	1316	0	0	0	0	1	-360	360;
	1025	3265	0.00086	0.00799	0	1415	0	0	0	0	1	-360	360;
	1339	3265	0.00189	0.02135	0	1448	0	0	0	0	1	-360	360;
	1339	1774	0.00106	0.01219	0	1218	0	0	0	0	1	-360	360;
	1339	2899	0.00147	0.01622	0	1316	0	0	0	0	1	-360	360;
	1339	1703	0.00215	0.02108	0	1349	0	0	0	0	1	-360	360;
	441	8088	0.00063	0.0052	0	1382	0	0	0	0	1	-360	360;
	441	3265	0.00159	0.01792	0	1119	0	0	0	0	1	-360	360;
	6298	1774	0.00157	0.01981	0	1251	0	0	0	0	1	-360	360;
	6298	2899	0.00158	0.02014	0	1481	0	0	0	0	1	-360	360;
	8088	3080	0.0018	0.02388	0	1316	0	0	0	0	1	-360	360;
	8088	544	0.00034	0.00318	0	1349	0	0	0	0	1	-360	360;
	8088	7079	0.00015	0.00132	0	757	0	0	0	0	1	-360	360;
	8088	8601	0.00014	0.0015	0	757	0	0	0	0	1	-360	360;
	8088	1155	0.00013	0.00153	0	757	0	0	0	0	1	-360	360;
	8088	3256	0.00013	0.00156	0	757	0	0	0	0	1	-360	360;
	8088	5815	0.0004	0.00449	0	1283	0	0	0	0	1	-360	360;
	8088	5815	0.00045	0.0044	0	1152	0	0	0	0	1	-360	360;
	3080	1899	0.0023	0.01984	0	1185	0	0	0	0	1	-360	360;
	3080	9032	0.0002	0.00235	0	757	0	0	0	0	1	-360	360;
	3080	6004	0.00019	0.00239	0	757	0	0	0	0	1	-360	360;
	3080	7810	0.00169	0.01602	0	1316	0	0	0	0	1	-360	360;
	3080	8630	0.00024	0.00253	0	790	0	0	0	0	1	-360	360;
	3080	8183	0.00022	0.00244	0	757	0	0	0	0	1	-360	360;
	3080	6390	0.00025	0.00239	0	757	0	0	0	0	1	-360	360;
	8224	7969	0.00066	0.0061	0	1316	0	0	0	0	1	-360	360;
	8224	7969	0.00062	0.00844	0	1316	0	0	0	0	1	-360	360;
	8224	7860	0.00124	0.01353	0	1580	0	0	0	0	1	-360	360;
	8224	6187	0.00046	0.00452	0	1283	0	0	0	0	1	-360	360;
	3007	3265	0.00045	0.00531	0	1349	0	0	0	0	1	-360	360;
	3007	544	0.00123	0.01433	0	1152	0	0	0	0	1	-360	360;
	3007	7810	0.00174	0.01696	0	1448	0	0	0	0	1	-360	360;
	4807	6187	0.00086	0.00877	0	1382	0	0	0	0	1	-360	360;
	4807	6760	0.00096	0.00999	0	1218	0	0	0	0	1	-360	360;
	4807	4926	0.00082	0.00761	0	757	0	0	0	0	1	-360	360;
	7969	1703	0.00185	0.01803	0	1283	0	0	0	0	1	-360	360;
	7969	2031	0.00138	0.01246	0	1481	0	0	0	0	1	-360	360;
	7969	7980	0.00062	0.00605	0	1514	0	0	0	0	1	-360	360;
	7969	6760	0.00073	0.00817	0	757	0	0	0	0	1	-360	360;
	536	7509	0.00094	0.00999	0	1316	0	0	0	0	1	-360	360;
	536	3990	0.00058	0.00671	0	1218	0	0	0	0	1	-360	360;
	1774	1703	0.00108	0.01279	0	1349	0	0	0	0	1	-360	360;
	1774	7980	0.00225	0.02472	0	1382	0	0	0	0	1	-360	360;
	1774	4543	0.00019	0.00244	0	757	0	0	0	0	1	-360	360;
	1774	6340	0.00018	0.00248	0	757	0	0	0	0	1	-360	360;
	1774	261	0.0001	0.0011	0	757	0	0	0	0	1	-360	360;
	1774	2058	0.00011	0.00108	0	757	0	0	0	0	1	-360	360;
	7860	6187	0.00089	0.00815	0	1283	0	0	0	0	1	-360	360;
	6080	2899	0.0009	0.00859	0	1415	0	0	0	0	1	-360	360;
	6080	7509	0.00167	0.02239	0	1547	0	0	0	0	1	-360	360;
	2899	5263	2e-05	0.00024	0	757	0	0	0	0	1	-360	360;
	1703	138	0.00157	0.01326	0	1349	0	0	0	0	1	-360	360;
	1899	5562	0.00155	0.0139	0	1349	0	0	0	0	1	-360	360;
	138	2031	0.00142	0.01472	0	1349	0	0	0	0	1	-360	360;
	4704	2031	0.00187	0.02232	0	1514	0	0	0	0	1	-360	360;
	4704	1570	0.00049	0.00572	0	1349	0	0	0	0	1	-360	360;
	2031	1999	0	0.00021	0	757	0	0	0	0	1	-360	360;
	7509	3990	0.00104	0.00939	0	1349	0	0	0	0	1	-360	360;
	7509	7935	0.00043	0.0049	0	757	0	0	0	0	1	-360	360;
	7509	7935	0.00049	0.0048	0	757	0	0	0	0	1	-360	360;
	39	1174	0.002009	0.01464	0	0	0	0	0	0	1	-360	360;
	1174	3711	0.003302	0.026	0	0	0	0	0	0	1	-360	360;
	389	4259	0.006222	0.018089	0	0	0	0	0	0	1	-360	360;
	3711	7411	0	0.000222	0	0	0	0	0	0	1	-360	360;
	3711	2894	8.9e-05	0.000911	0	0	0	0	0	0	1	-360	360;
	8771	7792	1.8e-05	0.000222	0	0	0	0	0	0	1	-360	360;
	6459	7411	0.00388	0.014289	0	286	0	0	0	0	1	-360	360;
	4013	610	0.001329	0.007898	0	0	0	0	0	0	1	-360	360;
	4013	4577	0.00048	0.002951	0	0	0	0	0	0	1	-360	360;
	4013	8007	8e-05	0.00032	0	0	0	0	0	0	1	-360	360;
	2853	5331	0.000302	0.001102	0	0	0	0	0	0	1	-360	360;
	6512	304	0.003742	0.013098	0	0	0	0	0	0	1	-360	360;
	7569	6135	0.008911	0.032298	0	0	0	0	0	0	1	-360	360;
	5927	6540	0.00124	0.005409	0	0	0	0	0	0	1	-360	360;
	698	7104	0.0025	0.0123	0	400	0	0	0	0	1	-360	360;
	698	7104	0.001781	0.013729	0	400	0	0	0	0	1	-360	360;
	698	4333	0.00324	0.016039	0	553	0	0	0	0	1	-360	360;
	8998	6311	0.00037	0.00474	0	0	0	0	0	0	1	-360	360;
	7761	1770	9.1e-05	0.000521	0	419	0	0	0	0	1	-360	360;
	6311	5361	3.1e-05	0.00043	0	0	0	0	0	0	1	-360	360;
	1770	7709	0.001281	0.005909	0	0	0	0	0	0	1	-360	360;
	1489	5790	0.000909	0.00476	0	0	0	0	0	0	1	-360	360;
	5094	5266	0.00011	0.000599	0	0	0	0	0	0	1	-360	360;
	3623	7927	0.000289	0.001719	0	0	0	0	0	0	1	-360	360;
	7236	4098	0.00026	0.001591	0	514	0	0	0	0	1	-360	360;
	5815	3990	0.00108	0.01382	0	1481	0	0	0	0	1	-360	360;
	7329	7661	0	0.0002	0	572	0	0	0	0	1	-360	360;
	2318	5347	0.014281	0.033446	0	0	0	0	0	0	1	-360	360;
	6361	5347	0.038273	0.11138	0	0	0	0	0	0	1	-360	360;
	2798	215	0.039504	0.11795	0	0	0	0	0	0	1	-360	360;
	2915	2798	0.039893	0.095083	0	0	0	0	0	0	1	-360	360;
	4607	5313	0.025107	0.052256	0	0	0	0	0	0	1	-360	360;
	1762	4607	0.010124	0.017818	0	0	0	0	0	0	1	-360	360;
	512	6834	0.035198	0.083893	0	0	0	0	0	0	1	-360	360;
	6834	6517	0.026149	0.064851	0	0	0	0	0	0	1	-360	360;
	7201	7419	0	0.000793	0	0	0	0	0	0	1	-360	360;
	3752	7201	0.02086	0.051744	0	0	0	0	0	0	1	-360	360;
	1298	7419	0.041992	0.117446	0	0	0	0	0	0	1	-360	360;
	5619	5050	0.014031	0.079591	0	457	0	0	0	0	1	-360	360;
	5050	3362	0.0075	0.054099	0	400	0	0	0	0	1	-360	360;
	3441	5050	0.013841	0.0733	0	438	0	0	0	0	1	-360	360;
	5050	2086	0.006469	0.041401	0	438	0	0	0	0	1	-360	360;
	1996	2592	0.005628	0.015826	0	0	0	0	0	0	1	-360	360;
	8645	1996	0.026364	0.056017	0	0	0	0	0	0	1	-360	360;
	1996	3752	0.026099	0.062182	0	0	0	0	0	0	1	-360	360;
	1996	5227	0.04481	0.131231	0	0	0	0	0	0	1	-360	360;
	1298	2592	0.005512	0.01614	0	0	0	0	0	0	1	-360	360;
	8645	1298	0.023488	0.063479	0	0	0	0	0	0	1	-360	360;
	1298	6250	0.044198	0.129446	0	0	0	0	0	0	1	-360	360;
	902	1424	0.006612	0.01657	0	0	0	0	0	0	1	-360	360;
	2744	902	0.003967	0.007521	0	0	0	0	0	0	1	-360	360;
	5811	6338	0.027496	0.067438	0	0	0	0	0	0	1	-360	360;
	4667	5811	0.044587	0.102702	0	0	0	0	0	0	1	-360	360;
	5811	4486	0.051438	0.143868	0	0	0	0	0	0	1	-360	360;
	5811	9017	0.029256	0.072579	0	0	0	0	0	0	1	-360	360;
	934	5811	0.049149	0.122653	0	0	0	0	0	0	1	-360	360;
	934	5811	0.043372	0.125157	0	0	0	0	0	0	1	-360	360;
	6991	5811	0.016471	0.042529	0	0	0	0	0	0	1	-360	360;
	7486	62	0	0.00076	0	0	0	0	0	0	1	-360	360;
	62	4451	0.038017	0.110653	0	0	0	0	0	0	1	-360	360;
	5135	62	0.025174	0.07938	0	0	0	0	0	0	1	-360	360;
	5474	62	0.043017	0.128446	0	0	0	0	0	0	1	-360	360;
	675	7486	0.051636	0.189413	0	0	0	0	0	0	1	-360	360;
	7486	3527	0.04714	0.113008	0	0	0	0	0	0	1	-360	360;
	2830	7486	0.008149	0.025711	0	0	0	0	0	0	1	-360	360;
	4532	653	0.039686	0.102843	0	0	0	0	0	0	1	-360	360;
	7109	653	0.006372	0.01857	0	0	0	0	0	0	1	-360	360;
	653	4653	0.045711	0.108901	0	0	0	0	0	0	1	-360	360;
	1192	1472	0.04119	0.106331	0	0	0	0	0	0	1	-360	360;
	3369	1192	0.013752	0.03024	0	0	0	0	0	0	1	-360	360;
	5849	4696	0.00595	0.019008	0	0	0	0	0	0	1	-360	360;
	4696	7930	0.00576	0.020331	0	0	0	0	0	0	1	-360	360;
	2857	8018	0.03586	0.082107	0	0	0	0	0	0	1	-360	360;
	428	2857	0.022223	0.055099	0	0	0	0	0	0	1	-360	360;
	3128	6642	0.008802	0.02862	0	0	0	0	0	0	1	-360	360;
	6642	7549	0.032719	0.124942	0	0	0	0	0	0	1	-360	360;
	760	8235	0.019793	0.055331	0	0	0	0	0	0	1	-360	360;
	8235	8241	0.01395	0.034612	0	0	0	0	0	0	1	-360	360;
	3682	2209	0.004727	0.010107	0	0	0	0	0	0	1	-360	360;
	8720	3682	0.007289	0.014364	0	0	0	0	0	0	1	-360	360;
	5221	4045	0.004818	0.009901	0	0	0	0	0	0	1	-360	360;
	468	5221	0.012223	0.031554	0	0	0	0	0	0	1	-360	360;
	1360	3142	0.00108	0.01189	0	1349	0	0	0	0	1	-360	360;
	971	1360	0.0032	0.03265	0	1316	0	0	0	0	1	-360	360;
	6710	1360	0.0007	0.00731	0	1415	0	0	0	0	1	-360	360;
	2537	5741	0	0.000744	0	0	0	0	0	0	1	-360	360;
	2537	4213	0.009248	0.022041	0	0	0	0	0	0	1	-360	360;
	5741	8586	0.008124	0.02286	0	0	0	0	0	0	1	-360	360;
	8476	7518	0.007702	0.019893	0	0	0	0	0	0	1	-360	360;
	7518	7308	0.015579	0.04914	0	0	0	0	0	0	1	-360	360;
	8820	3120	0.00981	0.068021	0	381	0	0	0	0	1	-360	360;
	8820	1695	0.006341	0.042211	0	400	0	0	0	0	1	-360	360;
	991	6205	0.060281	0.144512	0	0	0	0	0	0	1	-360	360;
	991	6205	0.061438	0.141504	0	0	0	0	0	0	1	-360	360;
	6205	2744	0.016579	0.039917	0	0	0	0	0	0	1	-360	360;
	1583	3733	0	0.000843	0	0	0	0	0	0	1	-360	360;
	2239	1583	0.010256	0.022711	0	0	0	0	0	0	1	-360	360;
	1583	4862	0.024983	0.073149	0	0	0	0	0	0	1	-360	360;
	3733	4263	0.008388	0.02557	0	0	0	0	0	0	1	-360	360;
	2565	2868	0.063512	0.200339	0	0	0	0	0	0	1	-360	360;
	4532	2565	0.037562	0.085967	0	0	0	0	0	0	1	-360	360;
	3030	746	0.002579	0.007752	0	0	0	0	0	0	1	-360	360;
	1375	746	0.001777	0.00495	0	0	0	0	0	0	1	-360	360;
	6614	6815	0.030248	0.084612	0	0	0	0	0	0	1	-360	360;
	6815	8666	0.013281	0.037124	0	0	0	0	0	0	1	-360	360;
	5472	8313	0.009488	0.023521	0	0	0	0	0	0	1	-360	360;
	125	8313	0.022579	0.058314	0	0	0	0	0	0	1	-360	360;
	4560	4833	0.02743	0.062793	0	0	0	0	0	0	1	-360	360;
	7221	4560	0.001719	0.003777	0	0	0	0	0	0	1	-360	360;
	2774	1988	0.035992	0.085876	0	0	0	0	0	0	1	-360	360;
	2442	2774	0.008777	0.01962	0	0	0	0	0	0	1	-360	360;
	3029	8062	0.001793	0.001521	0	0	0	0	0	0	1	-360	360;
	8062	2275	0.004917	0.007702	0	0	0	0	0	0	1	-360	360;
	7126	4227	0.001521	0.004554	0	0	0	0	0	0	1	-360	360;
	508	4227	0.011198	0.041099	0	0	0	0	0	0	1	-360	360;
	1732	67	0.00144	0.01155	0	1218	0	0	0	0	1	-360	360;
	1732	6449	0.00123	0.01345	0	1053	0	0	0	0	1	-360	360;
	456	6247	0.007669	0.02419	0	0	0	0	0	0	1	-360	360;
	456	189	0.003661	0.008769	0	0	0	0	0	0	1	-360	360;
	8643	456	0.00943	0.036008	0	0	0	0	0	0	1	-360	360;
	6264	456	0.012661	0.037802	0	0	0	0	0	0	1	-360	360;
	2008	799	0.006083	0.024099	0	0	0	0	0	0	1	-360	360;
	2008	1158	0.002248	0.004248	0	0	0	0	0	0	1	-360	360;
	1756	4746	0.019651	0.141771	0	381	0	0	0	0	1	-360	360;
	4746	3640	0.004841	0.02718	0	400	0	0	0	0	1	-360	360;
	4746	3640	0.00493	0.02661	0	419	0	0	0	0	1	-360	360;
	4327	4746	0.008729	0.04569	0	400	0	0	0	0	1	-360	360;
	8844	7067	0.002529	0.005793	0	0	0	0	0	0	1	-360	360;
	7067	6106	0.003579	0.007876	0	0	0	0	0	0	1	-360	360;
	8208	850	0.014182	0.03119	0	0	0	0	0	0	1	-360	360;
	8208	3805	0.061124	0.192777	0	0	0	0	0	0	1	-360	360;
	5074	4682	0.114603	0.284256	0	0	0	0	0	0	1	-360	360;
	4682	7613	0.00786	0.019488	0	0	0	0	0	0	1	-360	360;
	8844	8594	0.002669	0.006124	0	0	0	0	0	0	1	-360	360;
	3630	5039	0.00895	0.022182	0	0	0	0	0	0	1	-360	360;
	5039	233	0.002281	0.008033	0	0	0	0	0	0	1	-360	360;
	3630	850	0.005314	0.013719	0	0	0	0	0	0	1	-360	360;
	1045	5447	0.031521	0.069298	0	0	0	0	0	0	1	-360	360;
	5447	8594	0.003719	0.013223	0	0	0	0	0	0	1	-360	360;
	5447	233	0.009719	0.027818	0	0	0	0	0	0	1	-360	360;
	5447	3952	0.042669	0.127397	0	0	0	0	0	0	1	-360	360;
	5447	6712	0.031157	0.077281	0	0	0	0	0	0	1	-360	360;
	6106	850	0.00319	0.008934	0	0	0	0	0	0	1	-360	360;
	850	6592	0.01462	0.046124	0	0	0	0	0	0	1	-360	360;
	8251	4421	0.007587	0.018628	0	0	0	0	0	0	1	-360	360;
	7397	8251	0.024529	0.052826	0	0	0	0	0	0	1	-360	360;
	5103	5071	0.041223	0.115314	0	0	0	0	0	0	1	-360	360;
	5103	2723	0.022512	0.051521	0	0	0	0	0	0	1	-360	360;
	7109	5103	0.063537	0.189702	0	0	0	0	0	0	1	-360	360;
	1828	5582	0.01524	0.045496	0	0	0	0	0	0	1	-360	360;
	479	5582	0.035876	0.092612	0	0	0	0	0	0	1	-360	360;
	6247	5306	0.015471	0.046198	0	0	0	0	0	0	1	-360	360;
	8424	5646	0.043364	0.152818	0	0	0	0	0	0	1	-360	360;
	1274	5646	0.03095	0.100686	0	0	0	0	0	0	1	-360	360;
	2114	5646	0.035628	0.107959	0	0	0	0	0	0	1	-360	360;
	1195	6614	0.010901	0.034397	0	0	0	0	0	0	1	-360	360;
	3444	2792	0.026099	0.060099	0	0	0	0	0	0	1	-360	360;
	2752	3444	0.025942	0.057008	0	0	0	0	0	0	1	-360	360;
	8884	1619	0.001901	0.004636	0	0	0	0	0	0	1	-360	360;
	1619	8755	0.00176	0.003124	0	0	0	0	0	0	1	-360	360;
	4183	2868	0.003876	0.010909	0	0	0	0	0	0	1	-360	360;
	2868	713	0.003802	0.011124	0	0	0	0	0	0	1	-360	360;
	4587	2868	0.013339	0.029298	0	0	0	0	0	0	1	-360	360;
	1849	2868	0.031719	0.11181	0	0	0	0	0	0	1	-360	360;
	3373	3706	0.021521	0.05557	0	0	0	0	0	0	1	-360	360;
	4223	3373	0.037008	0.107752	0	0	0	0	0	0	1	-360	360;
	6513	1961	0.05743	0.142463	0	0	0	0	0	0	1	-360	360;
	6513	1548	0.036091	0.074041	0	0	0	0	0	0	1	-360	360;
	7889	4187	0.001211	0.00545	0	514	0	0	0	0	1	-360	360;
	2108	4187	0.013711	0.078669	0	0	0	0	0	0	1	-360	360;
	1245	4546	0.050521	0.125331	0	0	0	0	0	0	1	-360	360;
	4546	5231	0.085818	0.221579	0	0	0	0	0	0	1	-360	360;
	4546	5614	0.014909	0.034132	0	0	0	0	0	0	1	-360	360;
	4546	4764	0.035826	0.113033	0	0	0	0	0	0	1	-360	360;
	6265	6341	0.008289	0.026017	0	0	0	0	0	0	1	-360	360;
	3986	6341	0.023339	0.089587	0	0	0	0	0	0	1	-360	360;
	3731	2383	0.00661	0.039521	0	419	0	0	0	0	1	-360	360;
	5175	3731	0.00349	0.01574	0	419	0	0	0	0	1	-360	360;
	760	6800	0.03714	0.088496	0	0	0	0	0	0	1	-360	360;
	760	3893	0.037256	0.087669	0	0	0	0	0	0	1	-360	360;
	760	4584	0.03081	0.07957	0	0	0	0	0	0	1	-360	360;
	5029	5336	0.033248	0.096818	0	0	0	0	0	0	1	-360	360;
	9093	5336	0.011479	0.034298	0	0	0	0	0	0	1	-360	360;
	4837	1091	0	0.000893	0	0	0	0	0	0	1	-360	360;
	4837	3393	0.019273	0.042628	0	0	0	0	0	0	1	-360	360;
	4837	3732	0.019628	0.041702	0	0	0	0	0	0	1	-360	360;
	5955	4837	0.029512	0.062719	0	0	0	0	0	0	1	-360	360;
	6803	1091	0.034901	0.105744	0	0	0	0	0	0	1	-360	360;
	2246	1091	0.026289	0.071083	0	0	0	0	0	0	1	-360	360;
	560	1091	0.041893	0.121942	0	0	0	0	0	0	1	-360	360;
	2156	5071	0.024851	0.071711	0	0	0	0	0	0	1	-360	360;
	7365	6441	0.006512	0.01681	0	0	0	0	0	0	1	-360	360;
	3881	7365	0.00305	0.009099	0	0	0	0	0	0	1	-360	360;
	1446	7216	0.014678	0.04105	0	0	0	0	0	0	1	-360	360;
	6305	1446	0.054149	0.119066	0	0	0	0	0	0	1	-360	360;
	8302	4667	0.08557	0.197909	0	0	0	0	0	0	1	-360	360;
	8840	8302	0.038314	0.075496	0	0	0	0	0	0	1	-360	360;
	6515	8271	0.036554	0.115289	0	0	0	0	0	0	1	-360	360;
	2464	6515	0.047636	0.11019	0	0	0	0	0	0	1	-360	360;
	2238	2773	0	0.000826	0	0	0	0	0	0	1	-360	360;
	4741	2773	0.013397	0.037488	0	0	0	0	0	0	1	-360	360;
	2238	925	0.03095	0.079917	0	0	0	0	0	0	1	-360	360;
	5569	2238	0.014058	0.032198	0	0	0	0	0	0	1	-360	360;
	4756	9199	0.03086	0.093488	0	0	0	0	0	0	1	-360	360;
	6305	4756	0.023669	0.071066	0	0	0	0	0	0	1	-360	360;
	6305	4756	0.028083	0.078537	0	0	0	0	0	0	1	-360	360;
	788	2352	0.016661	0.052554	0	0	0	0	0	0	1	-360	360;
	2352	8127	0.022223	0.050868	0	0	0	0	0	0	1	-360	360;
	336	1070	0.017	0.043909	0	0	0	0	0	0	1	-360	360;
	1070	8148	0.017727	0.05595	0	0	0	0	0	0	1	-360	360;
	620	1536	0.022223	0.052967	0	0	0	0	0	0	1	-360	360;
	3642	1536	0.027628	0.071347	0	0	0	0	0	0	1	-360	360;
	6063	4585	0	0.00081	0	0	0	0	0	0	1	-360	360;
	6063	8949	0.002983	0.010149	0	0	0	0	0	0	1	-360	360;
	2567	6063	0.008678	0.023463	0	0	0	0	0	0	1	-360	360;
	4585	215	0.002917	0.010347	0	0	0	0	0	0	1	-360	360;
	2567	4585	0.008512	0.023917	0	0	0	0	0	0	1	-360	360;
	5725	3965	0.003603	0.011339	0	0	0	0	0	0	1	-360	360;
	5725	2318	0.006347	0.015132	0	0	0	0	0	0	1	-360	360;
	5728	5725	0.027777	0.08295	0	0	0	0	0	0	1	-360	360;
	479	3077	0.045438	0.105124	0	0	0	0	0	0	1	-360	360;
	3077	7747	0.07486	0.187603	0	0	0	0	0	0	1	-360	360;
	1783	4949	0.01919	0.053198	0	0	0	0	0	0	1	-360	360;
	8289	4949	0.008488	0.032421	0	0	0	0	0	0	1	-360	360;
	2427	7418	0.00365	0.03434	0	1053	0	0	0	0	1	-360	360;
	4903	7418	0.0049	0.04432	0	1119	0	0	0	0	1	-360	360;
	3120	3365	0.000901	0.005539	0	0	0	0	0	0	1	-360	360;
	2152	3120	0.00894	0.06106	0	438	0	0	0	0	1	-360	360;
	3120	1764	0.006919	0.049419	0	381	0	0	0	0	1	-360	360;
	3120	2251	0.006771	0.04576	0	476	0	0	0	0	1	-360	360;
	117	3328	0	0.000793	0	0	0	0	0	0	1	-360	360;
	117	8348	0.019628	0.061364	0	0	0	0	0	0	1	-360	360;
	117	6353	0.011727	0.033826	0	0	0	0	0	0	1	-360	360;
	3328	5964	0.006438	0.01943	0	0	0	0	0	0	1	-360	360;
	3328	5964	0.006562	0.019033	0	0	0	0	0	0	1	-360	360;
	4444	3328	0.006421	0.020893	0	0	0	0	0	0	1	-360	360;
	5759	4736	0.00031	0.00355	0	1251	0	0	0	0	1	-360	360;
	4282	2111	0.016752	0.046397	0	0	0	0	0	0	1	-360	360;
	3250	4282	0.009347	0.035702	0	0	0	0	0	0	1	-360	360;
	2111	8351	0.011298	0.040033	0	0	0	0	0	0	1	-360	360;
	8754	1297	0.001298	0.004322	0	0	0	0	0	0	1	-360	360;
	7703	1297	0.000347	0.000876	0	0	0	0	0	0	1	-360	360;
	2427	841	0.00028	0.00301	0	757	0	0	0	0	1	-360	360;
	67	2427	0.00297	0.02796	0	1086	0	0	0	0	1	-360	360;
	2427	3192	0.00274	0.03157	0	1020	0	0	0	0	1	-360	360;
	6265	831	0.00286	0.010554	0	0	0	0	0	0	1	-360	360;
	6265	9209	0.025818	0.091446	0	0	0	0	0	0	1	-360	360;
	1988	6265	0.027066	0.065554	0	0	0	0	0	0	1	-360	360;
	6265	2845	0.086198	0.151504	0	0	0	0	0	0	1	-360	360;
	753	6265	0.010893	0.02719	0	0	0	0	0	0	1	-360	360;
	5981	6265	0.03162	0.072818	0	0	0	0	0	0	1	-360	360;
	6770	831	0.002802	0.010752	0	0	0	0	0	0	1	-360	360;
	6770	8705	0.054802	0.152636	0	0	0	0	0	0	1	-360	360;
	6770	6975	0.014223	0.048397	0	0	0	0	0	0	1	-360	360;
	7199	6770	0.01814	0.043521	0	0	0	0	0	0	1	-360	360;
	6770	3964	0.038248	0.07586	0	0	0	0	0	0	1	-360	360;
	4053	6849	0.017331	0.061314	0	0	0	0	0	0	1	-360	360;
	3338	6849	0.009182	0.032339	0	0	0	0	0	0	1	-360	360;
	841	4396	0.00281	0.02539	0	1053	0	0	0	0	1	-360	360;
	6729	4396	0.00158	0.01426	0	1053	0	0	0	0	1	-360	360;
	6317	3570	0	0.000211	0	476	0	0	0	0	1	-360	360;
	9218	6317	0.00888	0.061021	0	419	0	0	0	0	1	-360	360;
	6317	1764	0.006961	0.04406	0	0	0	0	0	0	1	-360	360;
	3570	9005	0.00869	0.06218	0	0	0	0	0	0	1	-360	360;
	3570	7271	0.002401	0.027581	0	438	0	0	0	0	1	-360	360;
	3570	2251	0.00682	0.04493	0	438	0	0	0	0	1	-360	360;
	3434	8025	0.00238	0.009446	0	0	0	0	0	0	1	-360	360;
	7397	3434	0.017653	0.060008	0	0	0	0	0	0	1	-360	360;
	7344	3031	0.01838	0.051397	0	0	0	0	0	0	1	-360	360;
	3031	677	0.02995	0.071388	0	0	0	0	0	0	1	-360	360;
	3452	3899	0.023223	0.042223	0	0	0	0	0	0	1	-360	360;
	3899	633	0.018636	0.052124	0	0	0	0	0	0	1	-360	360;
	6729	2707	0.00096	0.01102	0	1218	0	0	0	0	1	-360	360;
	2707	5009	0.00364	0.02923	0	1185	0	0	0	0	1	-360	360;
	2707	2226	0.00248	0.02649	0	1020	0	0	0	0	1	-360	360;
	7017	2707	0.00384	0.04229	0	1086	0	0	0	0	1	-360	360;
	5690	2534	0	0.000219	0	0	0	0	0	0	1	-360	360;
	9218	2534	0.00344	0.02181	0	514	0	0	0	0	1	-360	360;
	2534	9005	0.00337	0.02224	0	476	0	0	0	0	1	-360	360;
	2534	204	0.010521	0.05913	0	400	0	0	0	0	1	-360	360;
	2534	545	0.00081	0.005341	0	438	0	0	0	0	1	-360	360;
	3029	2275	0.002579	0.00476	0	0	0	0	0	0	1	-360	360;
	8918	3029	0.004388	0.01595	0	0	0	0	0	0	1	-360	360;
	3029	7048	0.004752	0.017868	0	0	0	0	0	0	1	-360	360;
	1176	4669	0.00338	0.012959	0	0	0	0	0	0	1	-360	360;
	7909	1176	0.007008	0.02338	0	0	0	0	0	0	1	-360	360;
	5978	4443	0.025347	0.057942	0	0	0	0	0	0	1	-360	360;
	3847	5978	0.06314	0.162959	0	0	0	0	0	0	1	-360	360;
	3833	3043	0.007967	0.023802	0	0	0	0	0	0	1	-360	360;
	6571	3833	0.006636	0.016463	0	0	0	0	0	0	1	-360	360;
	4989	7588	0.012091	0.022893	0	0	0	0	0	0	1	-360	360;
	4989	3700	0.012322	0.022388	0	0	0	0	0	0	1	-360	360;
	6144	6805	0.016529	0.049347	0	0	0	0	0	0	1	-360	360;
	8377	8754	0.00119	0.007017	0	0	0	0	0	0	1	-360	360;
	8377	4617	0.002331	0.009636	0	0	0	0	0	0	1	-360	360;
	3033	1797	0.008091	0.02681	0	0	0	0	0	0	1	-360	360;
	1412	1797	0.001347	0.006876	0	0	0	0	0	0	1	-360	360;
	7703	1797	0.002281	0.009818	0	0	0	0	0	0	1	-360	360;
	400	8024	0.001149	0.00243	0	0	0	0	0	0	1	-360	360;
	3559	400	0.002909	0.006909	0	0	0	0	0	0	1	-360	360;
	677	2207	0.000413	0.000959	0	0	0	0	0	0	1	-360	360;
	677	2207	0.000421	0.000942	0	0	0	0	0	0	1	-360	360;
	7216	7598	0.024628	0.073504	0	0	0	0	0	0	1	-360	360;
	7598	6339	0.041496	0.094992	0	0	0	0	0	0	1	-360	360;
	3310	447	0.01	0.021273	0	0	0	0	0	0	1	-360	360;
	447	1849	0.030479	0.08795	0	0	0	0	0	0	1	-360	360;
	447	1849	0.031157	0.094992	0	0	0	0	0	0	1	-360	360;
	8572	6177	0.00962	0.023298	0	0	0	0	0	0	1	-360	360;
	1878	8572	0.018248	0.040397	0	0	0	0	0	0	1	-360	360;
	6792	5621	0.005868	0.01781	0	0	0	0	0	0	1	-360	360;
	6792	9020	0.01	0.02814	0	0	0	0	0	0	1	-360	360;
	6792	37	0.016562	0.052033	0	0	0	0	0	0	1	-360	360;
	6792	37	0.016893	0.050967	0	0	0	0	0	0	1	-360	360;
	5621	3639	0.013281	0.033149	0	0	0	0	0	0	1	-360	360;
	5076	1306	0.005463	0.020008	0	0	0	0	0	0	1	-360	360;
	5076	5249	0.02695	0.066868	0	0	0	0	0	0	1	-360	360;
	3639	5076	0.002752	0.008397	0	0	0	0	0	0	1	-360	360;
	5813	5076	0.002686	0.007777	0	0	0	0	0	0	1	-360	360;
	5076	37	0.019099	0.057876	0	0	0	0	0	0	1	-360	360;
	5076	3908	0.017339	0.043008	0	0	0	0	0	0	1	-360	360;
	1306	37	0.017843	0.068107	0	0	0	0	0	0	1	-360	360;
	588	8735	0.022289	0.070314	0	0	0	0	0	0	1	-360	360;
	8735	2311	0.035008	0.101917	0	0	0	0	0	0	1	-360	360;
	1405	991	0.045231	0.147174	0	0	0	0	0	0	1	-360	360;
	5283	991	0.011413	0.03424	0	0	0	0	0	0	1	-360	360;
	7142	2259	0.001124	0.002579	0	0	0	0	0	0	1	-360	360;
	7142	6105	0.046909	0.135298	0	0	0	0	0	0	1	-360	360;
	7142	4984	0.035446	0.075124	0	0	0	0	0	0	1	-360	360;
	8363	7142	0.011107	0.025793	0	0	0	0	0	0	1	-360	360;
	7142	9078	0.005983	0.023752	0	0	0	0	0	0	1	-360	360;
	5407	3453	0.016132	0.04695	0	0	0	0	0	0	1	-360	360;
	5018	3453	0.008008	0.019083	0	0	0	0	0	0	1	-360	360;
	4311	463	0.034281	0.088488	0	0	0	0	0	0	1	-360	360;
	463	5045	0.009149	0.036372	0	0	0	0	0	0	1	-360	360;
	463	4759	0.008711	0.024124	0	0	0	0	0	0	1	-360	360;
	463	1532	0.03624	0.109818	0	0	0	0	0	0	1	-360	360;
	6105	2259	0.036777	0.116041	0	0	0	0	0	0	1	-360	360;
	8018	8533	0.003851	0.00995	0	0	0	0	0	0	1	-360	360;
	8018	8482	0.025793	0.056702	0	0	0	0	0	0	1	-360	360;
	3034	8018	0.027942	0.061421	0	0	0	0	0	0	1	-360	360;
	8533	8922	0.021488	0.047248	0	0	0	0	0	0	1	-360	360;
	7506	3142	0.00015	0.00136	0	1251	0	0	0	0	1	-360	360;
	1195	9035	0.016909	0.037174	0	0	0	0	0	0	1	-360	360;
	4391	1195	0.040496	0.088512	0	0	0	0	0	0	1	-360	360;
	4960	566	0.035231	0.111132	0	0	0	0	0	0	1	-360	360;
	2198	4960	0.031017	0.086752	0	0	0	0	0	0	1	-360	360;
	3965	2318	0.003983	0.014017	0	0	0	0	0	0	1	-360	360;
	6706	996	0.003521	0.010496	0	0	0	0	0	0	1	-360	360;
	4741	996	0.002686	0.00805	0	0	0	0	0	0	1	-360	360;
	6881	1045	0.05195	0.145298	0	0	0	0	0	0	1	-360	360;
	3079	7469	0.023207	0.053132	0	0	0	0	0	0	1	-360	360;
	2115	3079	0.049587	0.150256	0	0	0	0	0	0	1	-360	360;
	2560	1472	0.036719	0.111298	0	0	0	0	0	0	1	-360	360;
	508	2560	0.008364	0.023207	0	0	0	0	0	0	1	-360	360;
	1412	6037	0.005256	0.024653	0	0	0	0	0	0	1	-360	360;
	3902	6037	0.001347	0.004397	0	0	0	0	0	0	1	-360	360;
	8754	6919	0.005149	0.022826	0	0	0	0	0	0	1	-360	360;
	5724	6919	0.001207	0.004992	0	0	0	0	0	0	1	-360	360;
	3805	5074	0.003934	0.009744	0	0	0	0	0	0	1	-360	360;
	5578	7216	0.01438	0.04186	0	0	0	0	0	0	1	-360	360;
	1424	1834	0.005744	0.014967	0	0	0	0	0	0	1	-360	360;
	6245	282	0.00176	0.01659	0	1086	0	0	0	0	1	-360	360;
	5009	282	0.00151	0.01378	0	1185	0	0	0	0	1	-360	360;
	5009	282	0.00128	0.01392	0	1152	0	0	0	0	1	-360	360;
	282	3729	0.00052	0.00397	0	1185	0	0	0	0	1	-360	360;
	2226	282	0.00221	0.02248	0	1152	0	0	0	0	1	-360	360;
	7109	6801	0.007182	0.017107	0	0	0	0	0	0	1	-360	360;
	6801	4102	0.005149	0.011339	0	0	0	0	0	0	1	-360	360;
	8586	4213	0.00943	0.02157	0	0	0	0	0	0	1	-360	360;
	7166	4421	0.040636	0.11257	0	0	0	0	0	0	1	-360	360;
	7585	7166	0.014496	0.042231	0	0	0	0	0	0	1	-360	360;
	1988	831	0.002744	0.00995	0	0	0	0	0	0	1	-360	360;
	831	6580	0.003099	0.009752	0	0	0	0	0	0	1	-360	360;
	596	8815	0.007322	0.016744	0	0	0	0	0	0	1	-360	360;
	5906	8815	0.03024	0.069231	0	0	0	0	0	0	1	-360	360;
	6733	1361	0.008438	0.018669	0	0	0	0	0	0	1	-360	360;
	1361	508	0.006727	0.027942	0	0	0	0	0	0	1	-360	360;
	1361	508	0.006587	0.028488	0	0	0	0	0	0	1	-360	360;
	675	2515	0.014579	0.040463	0	0	0	0	0	0	1	-360	360;
	7899	2515	0.012826	0.036992	0	0	0	0	0	0	1	-360	360;
	2515	4930	0.006909	0.019917	0	0	0	0	0	0	1	-360	360;
	6242	7909	0.003421	0.011041	0	0	0	0	0	0	1	-360	360;
	6242	8918	0.005364	0.016289	0	0	0	0	0	0	1	-360	360;
	4443	4179	0.012463	0.030909	0	0	0	0	0	0	1	-360	360;
	4443	4179	0.012702	0.030273	0	0	0	0	0	0	1	-360	360;
	7040	4179	0.008331	0.02438	0	0	0	0	0	0	1	-360	360;
	7536	4179	0.008149	0.024843	0	0	0	0	0	0	1	-360	360;
	1669	1821	0.010789	0.071169	0	400	0	0	0	0	1	-360	360;
	3853	1821	0.011901	0.08499	0	0	0	0	0	0	1	-360	360;
	2156	4359	0.035942	0.113397	0	0	0	0	0	0	1	-360	360;
	2723	2156	0.035818	0.078752	0	0	0	0	0	0	1	-360	360;
	2156	4441	0.052587	0.147083	0	0	0	0	0	0	1	-360	360;
	6571	1658	0.007273	0.016661	0	0	0	0	0	0	1	-360	360;
	3161	1658	0.023488	0.072488	0	0	0	0	0	0	1	-360	360;
	8922	3132	0.004207	0.011777	0	0	0	0	0	0	1	-360	360;
	8482	3132	0.005339	0.020388	0	0	0	0	0	0	1	-360	360;
	7993	1302	0.007446	0.01638	0	0	0	0	0	0	1	-360	360;
	3059	1302	0.023033	0.057124	0	0	0	0	0	0	1	-360	360;
	189	2691	0.030198	0.095248	0	0	0	0	0	0	1	-360	360;
	6285	5126	0	0.000744	0	0	0	0	0	0	1	-360	360;
	3033	6285	0.003628	0.012	0	0	0	0	0	0	1	-360	360;
	6285	3902	0.00681	0.028273	0	0	0	0	0	0	1	-360	360;
	5126	5724	0.006669	0.028826	0	0	0	0	0	0	1	-360	360;
	2984	1535	0.04543	0.100521	0	0	0	0	0	0	1	-360	360;
	2984	1535	0.046273	0.098331	0	0	0	0	0	0	1	-360	360;
	2333	1535	0.005041	0.014554	0	0	0	0	0	0	1	-360	360;
	4838	4780	0.007818	0.021653	0	0	0	0	0	0	1	-360	360;
	4780	5335	0.041992	0.154033	0	0	0	0	0	0	1	-360	360;
	4273	224	0.003711	0.009612	0	0	0	0	0	0	1	-360	360;
	4391	4273	0.009959	0.025719	0	0	0	0	0	0	1	-360	360;
	7549	224	0.006744	0.014314	0	0	0	0	0	0	1	-360	360;
	1135	6711	0.023909	0.052901	0	0	0	0	0	0	1	-360	360;
	6711	6672	0.024347	0.051752	0	0	0	0	0	0	1	-360	360;
	3452	949	0.04324	0.088702	0	0	0	0	0	0	1	-360	360;
	3452	533	0.045223	0.125488	0	0	0	0	0	0	1	-360	360;
	5249	1316	0.00305	0.007893	0	0	0	0	0	0	1	-360	360;
	4364	8402	0.01905	0.060107	0	0	0	0	0	0	1	-360	360;
	8402	8812	0.023471	0.055934	0	0	0	0	0	0	1	-360	360;
	7902	2258	0.004893	0.014339	0	0	0	0	0	0	1	-360	360;
	5472	6786	0.022934	0.050397	0	0	0	0	0	0	1	-360	360;
	60	7265	0.037248	0.081901	0	0	0	0	0	0	1	-360	360;
	1321	7265	0.013347	0.039083	0	0	0	0	0	0	1	-360	360;
	4988	7265	0.013058	0.039818	0	0	0	0	0	0	1	-360	360;
	7211	8780	0.04038	0.11757	0	0	0	0	0	0	1	-360	360;
	8780	7549	0.037397	0.111653	0	0	0	0	0	0	1	-360	360;
	47	2221	0.029273	0.069752	0	0	0	0	0	0	1	-360	360;
	47	2609	0.023678	0.070702	0	0	0	0	0	0	1	-360	360;
	1285	6800	0.001868	0.005463	0	0	0	0	0	0	1	-360	360;
	3510	4335	0.00048	0.00423	0	1152	0	0	0	0	1	-360	360;
	5849	8777	0.003992	0.010289	0	0	0	0	0	0	1	-360	360;
	5849	3350	0.003777	0.009413	0	0	0	0	0	0	1	-360	360;
	6339	5849	0.015091	0.047174	0	0	0	0	0	0	1	-360	360;
	7930	3804	0.013826	0.039917	0	0	0	0	0	0	1	-360	360;
	7930	617	0.015678	0.039124	0	0	0	0	0	0	1	-360	360;
	7930	6105	0.066851	0.192818	0	0	0	0	0	0	1	-360	360;
	7930	9078	0.039653	0.128983	0	0	0	0	0	0	1	-360	360;
	8777	4976	0.005091	0.009702	0	0	0	0	0	0	1	-360	360;
	4976	1048	0.004066	0.010083	0	0	0	0	0	0	1	-360	360;
	1048	3804	0.007661	0.024058	0	0	0	0	0	0	1	-360	360;
	7838	3804	0.045636	0.13286	0	0	0	0	0	0	1	-360	360;
	3350	1048	0.018421	0.042868	0	0	0	0	0	0	1	-360	360;
	1865	5619	0.00805	0.04211	0	0	0	0	0	0	1	-360	360;
	5619	4375	0.003211	0.02194	0	381	0	0	0	0	1	-360	360;
	5619	4327	0.00974	0.05531	0	438	0	0	0	0	1	-360	360;
	9209	1796	0.019719	0.05495	0	0	0	0	0	0	1	-360	360;
	5578	2404	0.03495	0.086686	0	0	0	0	0	0	1	-360	360;
	6305	2404	0.041149	0.122851	0	0	0	0	0	0	1	-360	360;
	3904	596	0.01124	0.027893	0	0	0	0	0	0	1	-360	360;
	4726	1455	0.02219	0.057281	0	0	0	0	0	0	1	-360	360;
	1455	857	0.011843	0.030529	0	0	0	0	0	0	1	-360	360;
	949	633	0.014298	0.039983	0	0	0	0	0	0	1	-360	360;
	8940	4443	0.008083	0.028636	0	0	0	0	0	0	1	-360	360;
	7040	4443	0.012942	0.02962	0	0	0	0	0	0	1	-360	360;
	2651	5816	0.01657	0.050017	0	0	0	0	0	0	1	-360	360;
	1796	2651	0.006331	0.019107	0	0	0	0	0	0	1	-360	360;
	8705	3986	0.003983	0.013537	0	0	0	0	0	0	1	-360	360;
	1550	6245	0.00136	0.01515	0	954	0	0	0	0	1	-360	360;
	5813	9020	0.007876	0.01743	0	0	0	0	0	0	1	-360	360;
	7224	215	0.018636	0.055653	0	0	0	0	0	0	1	-360	360;
	2691	7224	0.010554	0.031504	0	0	0	0	0	0	1	-360	360;
	5713	1740	0.013488	0.039686	0	0	0	0	0	0	1	-360	360;
	1740	5068	0.021934	0.048198	0	0	0	0	0	0	1	-360	360;
	1858	8071	0	0.00086	0	0	0	0	0	0	1	-360	360;
	1858	8499	0.026463	0.075669	0	0	0	0	0	0	1	-360	360;
	1858	967	0.029446	0.081364	0	0	0	0	0	0	1	-360	360;
	2431	1858	0.055851	0.127826	0	0	0	0	0	0	1	-360	360;
	3034	8071	0.047818	0.124744	0	0	0	0	0	0	1	-360	360;
	4881	3970	0.02562	0.063256	0	0	0	0	0	0	1	-360	360;
	3970	2993	0.006372	0.012298	0	0	0	0	0	0	1	-360	360;
	1755	2747	0.009603	0.021099	0	0	0	0	0	0	1	-360	360;
	8922	6974	0.014917	0.052579	0	0	0	0	0	0	1	-360	360;
	741	6974	0.03014	0.083512	0	0	0	0	0	0	1	-360	360;
	7294	6082	0.03319	0.092818	0	0	0	0	0	0	1	-360	360;
	6082	6788	0.024248	0.067826	0	0	0	0	0	0	1	-360	360;
	7866	8314	0.048529	0.153091	0	0	0	0	0	0	1	-360	360;
	3451	7866	0.055397	0.137421	0	0	0	0	0	0	1	-360	360;
	9227	6787	0	0.000219	0	400	0	0	0	0	1	-360	360;
	9227	5487	0.0013	0.008539	0	400	0	0	0	0	1	-360	360;
	4716	9227	0.007169	0.04463	0	400	0	0	0	0	1	-360	360;
	9227	7225	0.01787	0.097419	0	457	0	0	0	0	1	-360	360;
	6787	5487	0.00164	0.007409	0	381	0	0	0	0	1	-360	360;
	6787	8633	0.0097	0.06274	0	400	0	0	0	0	1	-360	360;
	4716	6787	0.007021	0.045469	0	419	0	0	0	0	1	-360	360;
	2834	8115	0.007496	0.022397	0	0	0	0	0	0	1	-360	360;
	4102	8115	0.007959	0.024124	0	0	0	0	0	0	1	-360	360;
	7649	2996	0.00405	0.011686	0	0	0	0	0	0	1	-360	360;
	7649	1785	0.005471	0.01581	0	0	0	0	0	0	1	-360	360;
	3873	7109	0.012727	0.031579	0	0	0	0	0	0	1	-360	360;
	4708	3873	0.009529	0.027496	0	0	0	0	0	0	1	-360	360;
	922	2061	0	0.000876	0	0	0	0	0	0	1	-360	360;
	7806	2061	0.01243	0.040413	0	0	0	0	0	0	1	-360	360;
	4708	2061	0.002636	0.007967	0	0	0	0	0	0	1	-360	360;
	4708	2061	0.002686	0.007802	0	0	0	0	0	0	1	-360	360;
	922	4532	0.049421	0.127587	0	0	0	0	0	0	1	-360	360;
	5489	922	0.006587	0.018248	0	0	0	0	0	0	1	-360	360;
	922	1961	0.064438	0.166364	0	0	0	0	0	0	1	-360	360;
	7372	1828	0.085322	0.155132	0	0	0	0	0	0	1	-360	360;
	68	4881	0.050488	0.130339	0	0	0	0	0	0	1	-360	360;
	8831	68	0.035471	0.111893	0	0	0	0	0	0	1	-360	360;
	7211	5540	0.013876	0.045132	0	0	0	0	0	0	1	-360	360;
	5540	4391	0.042744	0.16319	0	0	0	0	0	0	1	-360	360;
	5012	4660	0.027512	0.080083	0	0	0	0	0	0	1	-360	360;
	4029	5012	0.013132	0.038223	0	0	0	0	0	0	1	-360	360;
	8666	7516	0.017347	0.051802	0	0	0	0	0	0	1	-360	360;
	3451	7516	0.036446	0.086851	0	0	0	0	0	0	1	-360	360;
	3877	87	0	0.000893	0	0	0	0	0	0	1	-360	360;
	6236	87	0.013223	0.056876	0	0	0	0	0	0	1	-360	360;
	87	5955	0.01662	0.035322	0	0	0	0	0	0	1	-360	360;
	6803	3877	0.023752	0.069157	0	0	0	0	0	0	1	-360	360;
	3877	5193	0.029851	0.100967	0	0	0	0	0	0	1	-360	360;
	3877	4149	0.014322	0.031504	0	0	0	0	0	0	1	-360	360;
	3877	7798	0.01481	0.040017	0	0	0	0	0	0	1	-360	360;
	4359	4441	0.038066	0.094413	0	0	0	0	0	0	1	-360	360;
	5639	5736	0.018702	0.058967	0	0	0	0	0	0	1	-360	360;
	7798	5639	0.008554	0.023917	0	0	0	0	0	0	1	-360	360;
	1756	1865	0.009849	0.05651	0	400	0	0	0	0	1	-360	360;
	8476	741	0.013554	0.032066	0	0	0	0	0	0	1	-360	360;
	8476	2506	0.013719	0.042876	0	0	0	0	0	0	1	-360	360;
	7426	1256	0.032512	0.094636	0	0	0	0	0	0	1	-360	360;
	1256	4223	0.040008	0.116479	0	0	0	0	0	0	1	-360	360;
	7760	428	0.02319	0.055273	0	0	0	0	0	0	1	-360	360;
	967	7760	0.013248	0.046413	0	0	0	0	0	0	1	-360	360;
	2431	7760	0.073289	0.23843	0	0	0	0	0	0	1	-360	360;
	6463	301	0.008083	0.029661	0	0	0	0	0	0	1	-360	360;
	7186	301	0.014661	0.058273	0	0	0	0	0	0	1	-360	360;
	4833	3562	0.017273	0.044587	0	0	0	0	0	0	1	-360	360;
	3562	1705	0.016413	0.049	0	0	0	0	0	0	1	-360	360;
	1713	7696	0.054587	0.120008	0	0	0	0	0	0	1	-360	360;
	6361	1713	0.049521	0.144149	0	0	0	0	0	0	1	-360	360;
	8645	2592	0.02757	0.063124	0	0	0	0	0	0	1	-360	360;
	7356	4726	0.037686	0.089818	0	0	0	0	0	0	1	-360	360;
	5578	617	0.106033	0.321289	0	0	0	0	0	0	1	-360	360;
	9199	5578	0.066959	0.153273	0	0	0	0	0	0	1	-360	360;
	5578	6712	0.074421	0.208132	0	0	0	0	0	0	1	-360	360;
	5578	9214	0.054157	0.123967	0	0	0	0	0	0	1	-360	360;
	677	4219	0.086802	0.24276	0	0	0	0	0	0	1	-360	360;
	4219	5908	0.007628	0.018826	0	0	0	0	0	0	1	-360	360;
	2767	588	0.021256	0.050669	0	0	0	0	0	0	1	-360	360;
	6546	108	0.015893	0.046223	0	0	0	0	0	0	1	-360	360;
	108	3595	0.01562	0.034331	0	0	0	0	0	0	1	-360	360;
	7812	6011	0.011174	0.023008	0	0	0	0	0	0	1	-360	360;
	479	7812	0.014438	0.036174	0	0	0	0	0	0	1	-360	360;
	6011	3595	0.012636	0.029752	0	0	0	0	0	0	1	-360	360;
	2996	7806	0	0.000826	0	0	0	0	0	0	1	-360	360;
	2996	2834	0.00086	0.003157	0	0	0	0	0	0	1	-360	360;
	2996	1785	0.01005	0.027843	0	0	0	0	0	0	1	-360	360;
	7806	1022	0.000843	0.003223	0	0	0	0	0	0	1	-360	360;
	6177	1988	0.022983	0.064256	0	0	0	0	0	0	1	-360	360;
	1988	6062	0.008603	0.018273	0	0	0	0	0	0	1	-360	360;
	1988	6062	0.007661	0.020702	0	0	0	0	0	0	1	-360	360;
	1988	6580	0.002678	0.007174	0	0	0	0	0	0	1	-360	360;
	1988	6580	0.002388	0.008132	0	0	0	0	0	0	1	-360	360;
	1764	719	0.00237	0.011781	0	0	0	0	0	0	1	-360	360;
	2251	3536	0.001901	0.012831	0	0	0	0	0	0	1	-360	360;
	2209	4045	0	0.00081	0	0	0	0	0	0	1	-360	360;
	2209	5628	0.017273	0.025289	0	0	0	0	0	0	1	-360	360;
	4045	1104	0.030909	0.042397	0	0	0	0	0	0	1	-360	360;
	4045	1053	0.020967	0.066149	0	0	0	0	0	0	1	-360	360;
	4802	5710	0.049719	0.109298	0	0	0	0	0	0	1	-360	360;
	5710	2318	0.065909	0.144901	0	0	0	0	0	0	1	-360	360;
	1211	6729	0.00614	0.05775	0	1152	0	0	0	0	1	-360	360;
	5713	8754	0.020678	0.079256	0	0	0	0	0	0	1	-360	360;
	2221	8754	0.020529	0.051702	0	0	0	0	0	0	1	-360	360;
	1412	3504	0.015066	0.050711	0	0	0	0	0	0	1	-360	360;
	247	6487	0.032364	0.074058	0	0	0	0	0	0	1	-360	360;
	2531	247	0.024893	0.069612	0	0	0	0	0	0	1	-360	360;
	3847	7696	0.051289	0.112769	0	0	0	0	0	0	1	-360	360;
	7836	8922	0.024124	0.069562	0	0	0	0	0	0	1	-360	360;
	8922	1274	0.026769	0.10219	0	0	0	0	0	0	1	-360	360;
	7933	8922	0.025909	0.077347	0	0	0	0	0	0	1	-360	360;
	8922	773	0.01324	0.046628	0	0	0	0	0	0	1	-360	360;
	8424	7813	0.01676	0.03686	0	0	0	0	0	0	1	-360	360;
	2506	8424	0.017612	0.043678	0	0	0	0	0	0	1	-360	360;
	7993	4563	0.02043	0.057124	0	0	0	0	0	0	1	-360	360;
	6319	7017	0.00084	0.00669	0	987	0	0	0	0	1	-360	360;
	1853	4311	0.019893	0.062752	0	0	0	0	0	0	1	-360	360;
	6406	78	0.058719	0.164223	0	0	0	0	0	0	1	-360	360;
	6406	2531	0.021661	0.049587	0	0	0	0	0	0	1	-360	360;
	3679	2018	0.008	0.030686	0	0	0	0	0	0	1	-360	360;
	3706	4534	0.024017	0.057248	0	0	0	0	0	0	1	-360	360;
	4669	717	0.005636	0.020727	0	0	0	0	0	0	1	-360	360;
	5157	7048	0.004769	0.016826	0	0	0	0	0	0	1	-360	360;
	8153	5157	0.00895	0.031554	0	0	0	0	0	0	1	-360	360;
	7946	3802	0.010919	0.07275	0	0	0	0	0	0	1	-360	360;
	3802	8166	0.004841	0.02861	0	419	0	0	0	0	1	-360	360;
	5313	2959	0.044909	0.092107	0	0	0	0	0	0	1	-360	360;
	3161	3043	0.011901	0.039537	0	0	0	0	0	0	1	-360	360;
	404	9056	0.008421	0.020066	0	0	0	0	0	0	1	-360	360;
	9056	1505	0.07076	0.184612	0	0	0	0	0	0	1	-360	360;
	9056	4987	0.00624	0.018917	0	0	0	0	0	0	1	-360	360;
	9056	2011	0.055157	0.127587	0	0	0	0	0	0	1	-360	360;
	5736	756	0.03562	0.084876	0	0	0	0	0	0	1	-360	360;
	1022	8455	0.024248	0.073463	0	0	0	0	0	0	1	-360	360;
	1834	2239	0.006372	0.012686	0	0	0	0	0	0	1	-360	360;
	6851	460	0.003397	0.009802	0	0	0	0	0	0	1	-360	360;
	6851	4930	0.002793	0.008397	0	0	0	0	0	0	1	-360	360;
	125	7441	0.035636	0.100248	0	0	0	0	0	0	1	-360	360;
	3235	7441	0.034901	0.102182	0	0	0	0	0	0	1	-360	360;
	1197	5647	0.032388	0.094281	0	0	0	0	0	0	1	-360	360;
	4838	5647	0.031421	0.093802	0	0	0	0	0	0	1	-360	360;
	7946	4977	0.02455	0.175521	0	419	0	0	0	0	1	-360	360;
	7828	4977	0.011281	0.058469	0	0	0	0	0	0	1	-360	360;
	4977	403	0.015979	0.102281	0	0	0	0	0	0	1	-360	360;
	7344	6465	0	0.000777	0	0	0	0	0	0	1	-360	360;
	7344	8403	0.010868	0.037488	0	0	0	0	0	0	1	-360	360;
	7344	7864	0.011967	0.024562	0	0	0	0	0	0	1	-360	360;
	6465	8403	0.010636	0.034711	0	0	0	0	0	0	1	-360	360;
	6465	1744	0.055091	0.164496	0	0	0	0	0	0	1	-360	360;
	6465	7606	0.073091	0.221488	0	0	0	0	0	0	1	-360	360;
	8271	8283	0.01938	0.056421	0	0	0	0	0	0	1	-360	360;
	3131	8283	0.018983	0.057496	0	0	0	0	0	0	1	-360	360;
	4802	7040	0.031413	0.07795	0	0	0	0	0	0	1	-360	360;
	4802	7304	0.046231	0.119372	0	0	0	0	0	0	1	-360	360;
	5306	2915	0.012066	0.029901	0	0	0	0	0	0	1	-360	360;
	1038	336	0.022529	0.071066	0	0	0	0	0	0	1	-360	360;
	4534	1038	0.027471	0.065471	0	0	0	0	0	0	1	-360	360;
	717	4312	0	0.00076	0	0	0	0	0	0	1	-360	360;
	8153	4312	0.005868	0.020678	0	0	0	0	0	0	1	-360	360;
	1532	620	0.022636	0.049769	0	0	0	0	0	0	1	-360	360;
	6689	1104	0.033909	0.047901	0	0	0	0	0	0	1	-360	360;
	6689	1104	0.034537	0.04686	0	0	0	0	0	0	1	-360	360;
	1104	5628	0.012959	0.028661	0	0	0	0	0	0	1	-360	360;
	3050	7889	0.003919	0.02719	0	591	0	0	0	0	1	-360	360;
	5252	3153	0	0.000744	0	0	0	0	0	0	1	-360	360;
	3921	5252	0.013909	0.038909	0	0	0	0	0	0	1	-360	360;
	1190	3153	0.007579	0.023	0	0	0	0	0	0	1	-360	360;
	1634	4532	0.032587	0.077653	0	0	0	0	0	0	1	-360	360;
	4532	1548	0.035744	0.085851	0	0	0	0	0	0	1	-360	360;
	1196	2170	0.011969	0.0687	0	400	0	0	0	0	1	-360	360;
	2170	3213	0.00431	0.024719	0	438	0	0	0	0	1	-360	360;
	335	4881	0.107463	0.250107	0	0	0	0	0	0	1	-360	360;
	4881	864	0.043099	0.099678	0	0	0	0	0	0	1	-360	360;
	4881	1705	0.033058	0.075669	0	0	0	0	0	0	1	-360	360;
	8501	4881	0.045273	0.099537	0	0	0	0	0	0	1	-360	360;
	4881	2466	0.071182	0.141157	0	0	0	0	0	0	1	-360	360;
	4881	2466	0.072554	0.138223	0	0	0	0	0	0	1	-360	360;
	1048	617	0.00781	0.02357	0	0	0	0	0	0	1	-360	360;
	617	6517	0.014298	0.042702	0	0	0	0	0	0	1	-360	360;
	8836	8050	0.016793	0.041661	0	0	0	0	0	0	1	-360	360;
	8050	7588	0.011017	0.021711	0	0	0	0	0	0	1	-360	360;
	2108	7946	0.01255	0.08125	0	495	0	0	0	0	1	-360	360;
	9015	7853	0.008041	0.022488	0	0	0	0	0	0	1	-360	360;
	9015	7853	0.009149	0.017438	0	0	0	0	0	0	1	-360	360;
	9015	6647	0.032876	0.06481	0	0	0	0	0	0	1	-360	360;
	8076	9015	0.022512	0.056769	0	0	0	0	0	0	1	-360	360;
	7853	1712	0.01276	0.035702	0	0	0	0	0	0	1	-360	360;
	1766	6081	0.007793	0.02457	0	0	0	0	0	0	1	-360	360;
	3815	6081	0.00762	0.022752	0	0	0	0	0	0	1	-360	360;
	2311	6579	0.010157	0.025198	0	0	0	0	0	0	1	-360	360;
	2311	5047	0.018289	0.054603	0	0	0	0	0	0	1	-360	360;
	2311	5614	0.031876	0.082322	0	0	0	0	0	0	1	-360	360;
	9204	1766	0.011289	0.033917	0	0	0	0	0	0	1	-360	360;
	2431	7836	0.055537	0.203711	0	0	0	0	0	0	1	-360	360;
	3129	8755	0.001488	0.003314	0	0	0	0	0	0	1	-360	360;
	8038	3129	0.009083	0.012124	0	0	0	0	0	0	1	-360	360;
	3685	3207	0	0.000876	0	0	0	0	0	0	1	-360	360;
	6256	3510	0.0011	0.01267	0	1053	0	0	0	0	1	-360	360;
	3510	8137	0.00252	0.02751	0	1119	0	0	0	0	1	-360	360;
	4660	1053	0.04314	0.128802	0	0	0	0	0	0	1	-360	360;
	3921	1190	0.011562	0.029843	0	0	0	0	0	0	1	-360	360;
	7800	799	0.011256	0.041174	0	0	0	0	0	0	1	-360	360;
	7800	799	0.011017	0.04195	0	0	0	0	0	0	1	-360	360;
	7469	398	0.029496	0.076149	0	0	0	0	0	0	1	-360	360;
	8166	1196	0.009401	0.05757	0	381	0	0	0	0	1	-360	360;
	4838	1197	0.056934	0.185223	0	0	0	0	0	0	1	-360	360;
	5335	1197	0.005479	0.016479	0	0	0	0	0	0	1	-360	360;
	8643	6764	0.011289	0.029157	0	0	0	0	0	0	1	-360	360;
	8643	6264	0.000504	0.001446	0	0	0	0	0	0	1	-360	360;
	890	8643	0.014347	0.035802	0	0	0	0	0	0	1	-360	360;
	890	8643	0.014628	0.035066	0	0	0	0	0	0	1	-360	360;
	8643	6144	0.019537	0.05462	0	0	0	0	0	0	1	-360	360;
	1785	5489	0.003281	0.009488	0	0	0	0	0	0	1	-360	360;
	6339	2338	0.033339	0.086091	0	0	0	0	0	0	1	-360	360;
	9214	2338	0.032628	0.095	0	0	0	0	0	0	1	-360	360;
	5816	6975	0.008281	0.029331	0	0	0	0	0	0	1	-360	360;
	204	545	0.008711	0.059841	0	400	0	0	0	0	1	-360	360;
	204	7225	0.0127	0.06389	0	400	0	0	0	0	1	-360	360;
	675	2488	0	0.000826	0	0	0	0	0	0	1	-360	360;
	675	7899	0.016992	0.051281	0	0	0	0	0	0	1	-360	360;
	675	460	0.017322	0.050223	0	0	0	0	0	0	1	-360	360;
	675	3527	0.016901	0.06357	0	0	0	0	0	0	1	-360	360;
	675	3366	0.037231	0.112777	0	0	0	0	0	0	1	-360	360;
	5068	675	0.039636	0.115388	0	0	0	0	0	0	1	-360	360;
	2609	675	0.067339	0.212438	0	0	0	0	0	0	1	-360	360;
	857	6764	0.061322	0.134818	0	0	0	0	0	0	1	-360	360;
	2376	460	0.04114	0.119752	0	0	0	0	0	0	1	-360	360;
	1417	1655	0.03657	0.109198	0	0	0	0	0	0	1	-360	360;
	925	1417	0.042488	0.097248	0	0	0	0	0	0	1	-360	360;
	1417	4739	0.039099	0.089512	0	0	0	0	0	0	1	-360	360;
	1908	1717	0.013537	0.032471	0	0	0	0	0	0	1	-360	360;
	5009	4736	0.00158	0.01747	0	987	0	0	0	0	1	-360	360;
	2027	5487	0.00894	0.04824	0	419	0	0	0	0	1	-360	360;
	4716	2027	0.003531	0.018341	0	400	0	0	0	0	1	-360	360;
	8351	7426	0.029818	0.068256	0	0	0	0	0	0	1	-360	360;
	6279	5737	0	0.00081	0	0	0	0	0	0	1	-360	360;
	6279	7703	0.003512	0.004661	0	0	0	0	0	0	1	-360	360;
	4441	3310	0.010959	0.027322	0	0	0	0	0	0	1	-360	360;
	4617	5737	0.003579	0.004562	0	0	0	0	0	0	1	-360	360;
	3642	1853	0.045843	0.136851	0	0	0	0	0	0	1	-360	360;
	3642	3751	0.044678	0.135421	0	0	0	0	0	0	1	-360	360;
	3642	1751	0.015793	0.04786	0	0	0	0	0	0	1	-360	360;
	7045	4223	0.036248	0.10138	0	0	0	0	0	0	1	-360	360;
	1540	67	0.00336	0.02922	0	1251	0	0	0	0	1	-360	360;
	5537	7828	0.01663	0.097229	0	419	0	0	0	0	1	-360	360;
	8949	215	0	0.000777	0	0	0	0	0	0	1	-360	360;
	215	6647	0.09357	0.184397	0	0	0	0	0	0	1	-360	360;
	2024	215	0.090248	0.177719	0	0	0	0	0	0	1	-360	360;
	2024	215	0.091983	0.174017	0	0	0	0	0	0	1	-360	360;
	215	1254	0.018256	0.05314	0	0	0	0	0	0	1	-360	360;
	215	8076	0.112273	0.205446	0	0	0	0	0	0	1	-360	360;
	2383	3441	0.003401	0.02362	0	438	0	0	0	0	1	-360	360;
	2383	7515	0.002719	0.012781	0	400	0	0	0	0	1	-360	360;
	2383	2185	0.00205	0.022719	0	438	0	0	0	0	1	-360	360;
	8010	7745	0.036579	0.090719	0	0	0	0	0	0	1	-360	360;
	8010	7578	0.038802	0.092446	0	0	0	0	0	0	1	-360	360;
	1921	5029	0.029397	0.067298	0	0	0	0	0	0	1	-360	360;
	1921	9093	0.054289	0.124289	0	0	0	0	0	0	1	-360	360;
	5029	7902	0.007421	0.023438	0	0	0	0	0	0	1	-360	360;
	5029	5259	0.006124	0.018554	0	0	0	0	0	0	1	-360	360;
	3250	8351	0.010446	0.034017	0	0	0	0	0	0	1	-360	360;
	788	2830	0.045702	0.127802	0	0	0	0	0	0	1	-360	360;
	5193	7043	0.057438	0.171496	0	0	0	0	0	0	1	-360	360;
	4916	7043	0.025182	0.057636	0	0	0	0	0	0	1	-360	360;
	4641	2845	0.03224	0.075033	0	0	0	0	0	0	1	-360	360;
	335	864	0.075653	0.13295	0	0	0	0	0	0	1	-360	360;
	677	7800	0.005058	0.017017	0	0	0	0	0	0	1	-360	360;
	677	7800	0.005157	0.016669	0	0	0	0	0	0	1	-360	360;
	2318	4068	0.046777	0.101124	0	0	0	0	0	0	1	-360	360;
	6338	512	0.045364	0.135446	0	0	0	0	0	0	1	-360	360;
	7040	6579	0.037471	0.113529	0	0	0	0	0	0	1	-360	360;
	4537	2146	0.058372	0.168446	0	0	0	0	0	0	1	-360	360;
	4537	2146	0.066157	0.165083	0	0	0	0	0	0	1	-360	360;
	3131	2146	0.046628	0.120388	0	0	0	0	0	0	1	-360	360;
	5370	4587	0.029868	0.087488	0	0	0	0	0	0	1	-360	360;
	4764	2767	0.015562	0.047107	0	0	0	0	0	0	1	-360	360;
	7813	2114	0.045967	0.109562	0	0	0	0	0	0	1	-360	360;
	677	8024	0.00705	0.016876	0	0	0	0	0	0	1	-360	360;
	677	3559	0.004	0.009537	0	0	0	0	0	0	1	-360	360;
	677	2643	0.06243	0.161207	0	0	0	0	0	0	1	-360	360;
	1158	7864	0.005942	0.014149	0	0	0	0	0	0	1	-360	360;
	8692	6481	0.004651	0.024159	0	419	0	0	0	0	1	-360	360;
	8692	1563	0.00474	0.02363	0	457	0	0	0	0	1	-360	360;
	7549	9035	0.036669	0.115669	0	0	0	0	0	0	1	-360	360;
	8740	5791	0.043273	0.095132	0	0	0	0	0	0	1	-360	360;
	8740	5671	0.004521	0.011661	0	0	0	0	0	0	1	-360	360;
	3192	6256	0.00207	0.01871	0	1283	0	0	0	0	1	-360	360;
	3213	1756	0.00419	0.02712	0	381	0	0	0	0	1	-360	360;
	3213	1756	0.002349	0.022039	0	419	0	0	0	0	1	-360	360;
	6105	4984	0.02	0.04314	0	0	0	0	0	0	1	-360	360;
	7948	3504	0.00676	0.016124	0	0	0	0	0	0	1	-360	360;
	7415	7555	0.007876	0.022529	0	0	0	0	0	0	1	-360	360;
	7555	3815	0.025793	0.060298	0	0	0	0	0	0	1	-360	360;
	6550	7045	0.067587	0.204793	0	0	0	0	0	0	1	-360	360;
	6062	7294	0.053008	0.116554	0	0	0	0	0	0	1	-360	360;
	6696	6647	0.01195	0.030868	0	0	0	0	0	0	1	-360	360;
	5671	6696	0.015463	0.04876	0	0	0	0	0	0	1	-360	360;
	508	3059	0.018702	0.071413	0	0	0	0	0	0	1	-360	360;
	3115	4736	0.00142	0.01555	0	757	0	0	0	0	1	-360	360;
	3115	1540	0.00123	0.01312	0	1086	0	0	0	0	1	-360	360;
	3115	1540	0.00121	0.01337	0	1185	0	0	0	0	1	-360	360;
	1550	3115	0.00079	0.01226	0	1086	0	0	0	0	1	-360	360;
	2904	3115	0.00062	0.00673	0	757	0	0	0	0	1	-360	360;
	6116	5487	0	0.000211	0	0	0	0	0	0	1	-360	360;
	4716	5487	0.00763	0.05451	0	381	0	0	0	0	1	-360	360;
	7109	8455	0	0.00076	0	0	0	0	0	0	1	-360	360;
	8940	5842	0.00657	0.019603	0	0	0	0	0	0	1	-360	360;
	6706	6363	0.011512	0.028562	0	0	0	0	0	0	1	-360	360;
	7933	8499	0.014612	0.05362	0	0	0	0	0	0	1	-360	360;
	9166	404	0.00976	0.02419	0	0	0	0	0	0	1	-360	360;
	1532	4664	0.007876	0.022934	0	0	0	0	0	0	1	-360	360;
	4664	3751	0.01795	0.042769	0	0	0	0	0	0	1	-360	360;
	6463	5886	0.023488	0.046364	0	0	0	0	0	0	1	-360	360;
	6463	1717	0.013802	0.031802	0	0	0	0	0	0	1	-360	360;
	3030	7439	0.036678	0.080636	0	0	0	0	0	0	1	-360	360;
	7439	4351	0.005917	0.017066	0	0	0	0	0	0	1	-360	360;
	7948	1847	0.003364	0.012901	0	0	0	0	0	0	1	-360	360;
	2376	7948	0.038107	0.120198	0	0	0	0	0	0	1	-360	360;
	5713	1847	0.003289	0.011942	0	0	0	0	0	0	1	-360	360;
	925	4271	0.047818	0.109479	0	0	0	0	0	0	1	-360	360;
	4271	4497	0.023603	0.060744	0	0	0	0	0	0	1	-360	360;
	4271	4992	0.042851	0.125488	0	0	0	0	0	0	1	-360	360;
	4271	4992	0.04376	0.123124	0	0	0	0	0	0	1	-360	360;
	4736	1540	0.00089	0.0071	0	1185	0	0	0	0	1	-360	360;
	8351	6550	0	0.000744	0	0	0	0	0	0	1	-360	360;
	8351	8148	0.039554	0.086942	0	0	0	0	0	0	1	-360	360;
	508	3369	0.046364	0.078843	0	0	0	0	0	0	1	-360	360;
	8097	3671	0.006248	0.007702	0	0	0	0	0	0	1	-360	360;
	8884	8097	0.001273	0.002661	0	0	0	0	0	0	1	-360	360;
	6487	4364	0.027711	0.061314	0	0	0	0	0	0	1	-360	360;
	6487	4364	0.028231	0.059967	0	0	0	0	0	0	1	-360	360;
	1047	5791	0.029702	0.056231	0	0	0	0	0	0	1	-360	360;
	2993	6422	0.006256	0.011397	0	0	0	0	0	0	1	-360	360;
	4584	4794	0.010298	0.032512	0	0	0	0	0	0	1	-360	360;
	4794	9148	0.01805	0.039686	0	0	0	0	0	0	1	-360	360;
	3511	7120	0.012849	0.07226	0	591	0	0	0	0	1	-360	360;
	3511	8583	0.013099	0.07075	0	629	0	0	0	0	1	-360	360;
	1783	8289	0.008314	0.033033	0	0	0	0	0	0	1	-360	360;
	6803	8794	0.031843	0.096463	0	0	0	0	0	0	1	-360	360;
	8801	3893	0.002207	0.005273	0	0	0	0	0	0	1	-360	360;
	6503	8801	0.009124	0.027669	0	0	0	0	0	0	1	-360	360;
	1655	5370	0.010777	0.032653	0	0	0	0	0	0	1	-360	360;
	1135	7975	0.00757	0.018901	0	0	0	0	0	0	1	-360	360;
	2966	6672	0.007719	0.018512	0	0	0	0	0	0	1	-360	360;
	3964	4641	0.016198	0.040818	0	0	0	0	0	0	1	-360	360;
	737	2882	0.054512	0.162752	0	0	0	0	0	0	1	-360	360;
	2882	2752	0.030802	0.067719	0	0	0	0	0	0	1	-360	360;
	4563	7126	0.012298	0.029281	0	0	0	0	0	0	1	-360	360;
	278	8290	0.074322	0.221901	0	0	0	0	0	0	1	-360	360;
	934	8290	0.009669	0.023033	0	0	0	0	0	0	1	-360	360;
	890	2024	0	0.000843	0	0	0	0	0	0	1	-360	360;
	46	1286	0.073463	0.189678	0	0	0	0	0	0	1	-360	360;
	1286	1053	0.058653	0.145488	0	0	0	0	0	0	1	-360	360;
	1550	3729	0.00102	0.01773	0	757	0	0	0	0	1	-360	360;
	5193	4252	0.00838	0.02438	0	0	0	0	0	0	1	-360	360;
	5193	7215	0.01219	0.03024	0	0	0	0	0	0	1	-360	360;
	5193	4916	0.048744	0.11614	0	0	0	0	0	0	1	-360	360;
	7215	3393	0.012967	0.030934	0	0	0	0	0	0	1	-360	360;
	4252	3732	0.006983	0.021149	0	0	0	0	0	0	1	-360	360;
	6236	4252	0.017347	0.058702	0	0	0	0	0	0	1	-360	360;
	2466	7221	0.02457	0.058562	0	0	0	0	0	0	1	-360	360;
	6522	5436	0.016083	0.044983	0	0	0	0	0	0	1	-360	360;
	4862	5436	0.004843	0.011512	0	0	0	0	0	0	1	-360	360;
	7838	8363	0.005058	0.011339	0	0	0	0	0	0	1	-360	360;
	7003	2557	0.023727	0.074843	0	0	0	0	0	0	1	-360	360;
	8794	2557	0.031157	0.098281	0	0	0	0	0	0	1	-360	360;
	773	741	0.014372	0.048661	0	0	0	0	0	0	1	-360	360;
	971	3192	0.00331	0.02754	0	1053	0	0	0	0	1	-360	360;
	7515	7278	0.002669	0.01732	0	457	0	0	0	0	1	-360	360;
	7278	7750	0.019719	0.103289	0	438	0	0	0	0	1	-360	360;
	7515	3441	0.00261	0.016039	0	400	0	0	0	0	1	-360	360;
	479	6546	0.008587	0.017446	0	0	0	0	0	0	1	-360	360;
	8836	1405	0	0.000876	0	0	0	0	0	0	1	-360	360;
	8348	8836	0.017413	0.05443	0	0	0	0	0	0	1	-360	360;
	8836	6353	0.024744	0.074289	0	0	0	0	0	0	1	-360	360;
	1405	5283	0.036421	0.11381	0	0	0	0	0	0	1	-360	360;
	5045	4759	0.004198	0.012587	0	0	0	0	0	0	1	-360	360;
	8127	7151	0.014686	0.044521	0	0	0	0	0	0	1	-360	360;
	2609	7151	0.015231	0.041256	0	0	0	0	0	0	1	-360	360;
	60	6786	0.01438	0.045364	0	0	0	0	0	0	1	-360	360;
	7345	1008	0.025562	0.076298	0	0	0	0	0	0	1	-360	360;
	1008	6361	0.028083	0.061744	0	0	0	0	0	0	1	-360	360;
	37	5132	0.05019	0.152107	0	0	0	0	0	0	1	-360	360;
	700	5132	0.027909	0.072033	0	0	0	0	0	0	1	-360	360;
	1961	4653	0.020512	0.061248	0	0	0	0	0	0	1	-360	360;
	857	6048	0.025364	0.053893	0	0	0	0	0	0	1	-360	360;
	8314	7211	0.054479	0.165099	0	0	0	0	0	0	1	-360	360;
	8314	7211	0.055636	0.161983	0	0	0	0	0	0	1	-360	360;
	8314	3128	0.051579	0.154	0	0	0	0	0	0	1	-360	360;
	1053	533	0.035752	0.106752	0	0	0	0	0	0	1	-360	360;
	633	6689	0.021124	0.054537	0	0	0	0	0	0	1	-360	360;
	633	6689	0.021537	0.05343	0	0	0	0	0	0	1	-360	360;
	6665	545	0.0082	0.04119	0	381	0	0	0	0	1	-360	360;
	4364	8812	0.040281	0.122058	0	0	0	0	0	0	1	-360	360;
	5370	4497	0.031934	0.10086	0	0	0	0	0	0	1	-360	360;
	5370	1634	0.048347	0.145008	0	0	0	0	0	0	1	-360	360;
	5370	4739	0.056636	0.164901	0	0	0	0	0	0	1	-360	360;
	4537	7232	0.048091	0.119281	0	0	0	0	0	0	1	-360	360;
	4537	4902	0.00305	0.008793	0	0	0	0	0	0	1	-360	360;
	4537	4902	0.003463	0.00862	0	0	0	0	0	0	1	-360	360;
	5908	4537	0.004223	0.009339	0	0	0	0	0	0	1	-360	360;
	5908	4537	0.005281	0.009603	0	0	0	0	0	0	1	-360	360;
	6338	4486	0.028033	0.066033	0	0	0	0	0	0	1	-360	360;
	2645	9204	0.012983	0.051587	0	0	0	0	0	0	1	-360	360;
	6710	4394	0.00183	0.01416	0	987	0	0	0	0	1	-360	360;
	4149	7061	0.012504	0.036397	0	0	0	0	0	0	1	-360	360;
	8794	7061	0.024488	0.074198	0	0	0	0	0	0	1	-360	360;
	3881	8025	0.010579	0.045496	0	0	0	0	0	0	1	-360	360;
	7232	1744	0.025372	0.065512	0	0	0	0	0	0	1	-360	360;
	5037	6592	0.015058	0.042124	0	0	0	0	0	0	1	-360	360;
	3566	5037	0.038248	0.084099	0	0	0	0	0	0	1	-360	360;
	8212	737	0.045719	0.113397	0	0	0	0	0	0	1	-360	360;
	8212	3908	0.046603	0.111041	0	0	0	0	0	0	1	-360	360;
	756	8286	0.013843	0.041934	0	0	0	0	0	0	1	-360	360;
	8831	8501	0.052083	0.134446	0	0	0	0	0	0	1	-360	360;
	4667	278	0.013579	0.03981	0	0	0	0	0	0	1	-360	360;
	4667	278	0.013868	0.039058	0	0	0	0	0	0	1	-360	360;
	4667	6991	0.035099	0.077182	0	0	0	0	0	0	1	-360	360;
	8840	4667	0.11376	0.27386	0	0	0	0	0	0	1	-360	360;
	278	8070	0.069628	0.153091	0	0	0	0	0	0	1	-360	360;
	8137	4394	0.00296	0.03263	0	987	0	0	0	0	1	-360	360;
	3815	5407	0.048322	0.115149	0	0	0	0	0	0	1	-360	360;
	1285	5087	0.009802	0.028678	0	0	0	0	0	0	1	-360	360;
	8038	3671	0.003132	0.004347	0	0	0	0	0	0	1	-360	360;
	3356	4697	0.00284	0.02937	0	1415	0	0	0	0	1	-360	360;
	4697	7017	0.00353	0.03885	0	1020	0	0	0	0	1	-360	360;
	1211	4697	0.00193	0.02044	0	1185	0	0	0	0	1	-360	360;
	3853	1669	0.00918	0.065591	0	457	0	0	0	0	1	-360	360;
	7671	3853	0	0.000211	0	419	0	0	0	0	1	-360	360;
	3853	8692	0.01776	0.09207	0	419	0	0	0	0	1	-360	360;
	3853	8692	0.018081	0.09007	0	419	0	0	0	0	1	-360	360;
	3853	3915	0.0017	0.0117	0	0	0	0	0	0	1	-360	360;
	3853	1127	0.001669	0.011919	0	0	0	0	0	0	1	-360	360;
	4171	9166	0.009041	0.023331	0	0	0	0	0	0	1	-360	360;
	2462	3359	0.075851	0.15557	0	0	0	0	0	0	1	-360	360;
	2036	2462	0.012471	0.036587	0	0	0	0	0	0	1	-360	360;
	2645	7415	0.003702	0.011537	0	0	0	0	0	0	1	-360	360;
	5955	2246	0.003661	0.009066	0	0	0	0	0	0	1	-360	360;
	5955	7798	0	0.000893	0	0	0	0	0	0	1	-360	360;
	5185	5175	0.00537	0.032901	0	0	0	0	0	0	1	-360	360;
	9105	6503	0.016661	0.041331	0	0	0	0	0	0	1	-360	360;
	9105	7348	0.002628	0.007587	0	0	0	0	0	0	1	-360	360;
	9105	9148	0.009446	0.025388	0	0	0	0	0	0	1	-360	360;
	7897	7745	0.023248	0.060017	0	0	0	0	0	0	1	-360	360;
	1316	737	0.053711	0.127983	0	0	0	0	0	0	1	-360	360;
	8286	560	0.009851	0.022537	0	0	0	0	0	0	1	-360	360;
	2777	7757	0.007256	0.021702	0	0	0	0	0	0	1	-360	360;
	2777	8241	0.001058	0.003149	0	0	0	0	0	0	1	-360	360;
	7757	7348	0.014744	0.048893	0	0	0	0	0	0	1	-360	360;
	3366	2780	0.040901	0.119017	0	0	0	0	0	0	1	-360	360;
	6881	7613	0.026653	0.063504	0	0	0	0	0	0	1	-360	360;
	4538	2011	0.005198	0.011909	0	0	0	0	0	0	1	-360	360;
	4538	978	0.006603	0.020793	0	0	0	0	0	0	1	-360	360;
	925	1351	0.023909	0.054744	0	0	0	0	0	0	1	-360	360;
	4992	1351	0.01286	0.038959	0	0	0	0	0	0	1	-360	360;
	6522	355	0	0.000826	0	0	0	0	0	0	1	-360	360;
	7588	355	0.005488	0.011256	0	0	0	0	0	0	1	-360	360;
	3700	355	0.018983	0.055579	0	0	0	0	0	0	1	-360	360;
	6522	4444	0.013479	0.038876	0	0	0	0	0	0	1	-360	360;
	7397	4851	0.010041	0.02205	0	0	0	0	0	0	1	-360	360;
	4851	7585	0.01224	0.037099	0	0	0	0	0	0	1	-360	360;
	7397	4053	0	0.00081	0	0	0	0	0	0	1	-360	360;
	5626	7397	0.011	0.030959	0	0	0	0	0	0	1	-360	360;
	5095	4053	0.010777	0.031554	0	0	0	0	0	0	1	-360	360;
	4053	5906	0.030521	0.069868	0	0	0	0	0	0	1	-360	360;
	4053	6918	0.016901	0.051198	0	0	0	0	0	0	1	-360	360;
	6441	3904	0.008868	0.027967	0	0	0	0	0	0	1	-360	360;
	7585	3338	0.006322	0.02319	0	0	0	0	0	0	1	-360	360;
	2792	5626	0.006488	0.019008	0	0	0	0	0	0	1	-360	360;
	3273	5095	0.006347	0.019372	0	0	0	0	0	0	1	-360	360;
	2752	3273	0.018339	0.038967	0	0	0	0	0	0	1	-360	360;
	1245	5231	0.037793	0.093752	0	0	0	0	0	0	1	-360	360;
	1284	753	0.018603	0.048017	0	0	0	0	0	0	1	-360	360;
	4191	6548	0.04086	0.128893	0	0	0	0	0	0	1	-360	360;
	6548	1375	0.004587	0.010083	0	0	0	0	0	0	1	-360	360;
	2780	5474	0.012529	0.028661	0	0	0	0	0	0	1	-360	360;
	46	4029	0.016339	0.035909	0	0	0	0	0	0	1	-360	360;
	1284	7199	0.021702	0.058653	0	0	0	0	0	0	1	-360	360;
	6250	7975	0.037339	0.109364	0	0	0	0	0	0	1	-360	360;
	5227	2966	0.036554	0.11143	0	0	0	0	0	0	1	-360	360;
	6048	6144	0.027149	0.062149	0	0	0	0	0	0	1	-360	360;
	5141	3847	0.034281	0.103876	0	0	0	0	0	0	1	-360	360;
	5141	5047	0.023479	0.06843	0	0	0	0	0	0	1	-360	360;
	7536	5141	0.053149	0.116868	0	0	0	0	0	0	1	-360	360;
	5886	6733	0.004802	0.013298	0	0	0	0	0	0	1	-360	360;
	6733	508	0.012669	0.038231	0	0	0	0	0	0	1	-360	360;
	4902	2464	0.040339	0.079496	0	0	0	0	0	0	1	-360	360;
	8501	1693	0.054521	0.171992	0	0	0	0	0	0	1	-360	360;
	5632	1693	0.05514	0.122	0	0	0	0	0	0	1	-360	360;
	2333	1693	0.056157	0.119339	0	0	0	0	0	0	1	-360	360;
	9017	4486	0.034248	0.075289	0	0	0	0	0	0	1	-360	360;
	3356	887	0.0018	0.01865	0	1448	0	0	0	0	1	-360	360;
	7549	4391	0	0.000793	0	0	0	0	0	0	1	-360	360;
	7839	7345	0.023231	0.064983	0	0	0	0	0	0	1	-360	360;
	1695	2152	0.00163	0.01087	0	400	0	0	0	0	1	-360	360;
	6409	37	0.020727	0.053512	0	0	0	0	0	0	1	-360	360;
	508	7186	0.022364	0.064488	0	0	0	0	0	0	1	-360	360;
	925	4378	0.014579	0.04805	0	0	0	0	0	0	1	-360	360;
	925	4378	0.014281	0.048992	0	0	0	0	0	0	1	-360	360;
	6449	3017	0.00228	0.02412	0	1086	0	0	0	0	1	-360	360;
	7578	1505	0.004298	0.009132	0	0	0	0	0	0	1	-360	360;
	3359	7578	0.003826	0.010347	0	0	0	0	0	0	1	-360	360;
	3359	1505	0	0.000777	0	0	0	0	0	0	1	-360	360;
	7897	1505	0.042256	0.092917	0	0	0	0	0	0	1	-360	360;
	3640	7010	0.005021	0.029341	0	419	0	0	0	0	1	-360	360;
	482	1309	0.066248	0.151661	0	0	0	0	0	0	1	-360	360;
	1309	4451	0.007587	0.016702	0	0	0	0	0	0	1	-360	360;
	2747	4987	0.008579	0.019636	0	0	0	0	0	0	1	-360	360;
	7536	5842	0.008942	0.024992	0	0	0	0	0	0	1	-360	360;
	3941	3566	0.049124	0.154983	0	0	0	0	0	0	1	-360	360;
	2198	3941	0.018289	0.04186	0	0	0	0	0	0	1	-360	360;
	887	7506	0.00064	0.00617	0	1415	0	0	0	0	1	-360	360;
	3815	1511	0	0.00076	0	0	0	0	0	0	1	-360	360;
	3815	5018	0.024421	0.073331	0	0	0	0	0	0	1	-360	360;
	1511	6571	0	0.000744	0	0	0	0	0	0	1	-360	360;
	7003	8794	0.025653	0.056388	0	0	0	0	0	0	1	-360	360;
	4327	5337	0.003711	0.02356	0	419	0	0	0	0	1	-360	360;
	3030	4351	0.015777	0.089537	0	0	0	0	0	0	1	-360	360;
	1375	3030	0.001587	0.005612	0	0	0	0	0	0	1	-360	360;
	978	4171	0.008934	0.027091	0	0	0	0	0	0	1	-360	360;
	2036	2011	0.021298	0.052603	0	0	0	0	0	0	1	-360	360;
	7660	8633	0.01036	0.05943	0	400	0	0	0	0	1	-360	360;
	8749	934	0.036438	0.114909	0	0	0	0	0	0	1	-360	360;
	566	8749	0.038876	0.113207	0	0	0	0	0	0	1	-360	360;
	3952	6726	0.045058	0.127901	0	0	0	0	0	0	1	-360	360;
	6726	2198	0.014083	0.042661	0	0	0	0	0	0	1	-360	360;
	941	482	0.016537	0.052157	0	0	0	0	0	0	1	-360	360;
	6788	5981	0.045711	0.136504	0	0	0	0	0	0	1	-360	360;
	4903	3192	0.00579	0.06394	0	1086	0	0	0	0	1	-360	360;
	1762	1321	0.012512	0.03381	0	0	0	0	0	0	1	-360	360;
	1321	125	0.053149	0.149512	0	0	0	0	0	0	1	-360	360;
	1762	4988	0.012248	0.034471	0	0	0	0	0	0	1	-360	360;
	4988	3235	0.052041	0.152388	0	0	0	0	0	0	1	-360	360;
	934	4191	0.024471	0.056033	0	0	0	0	0	0	1	-360	360;
	1878	2442	0.008752	0.016322	0	0	0	0	0	0	1	-360	360;
	4903	7017	0.00211	0.01832	0	1316	0	0	0	0	1	-360	360;
	6825	7308	0	0.00086	0	0	0	0	0	0	1	-360	360;
	1853	78	0	0.000876	0	0	0	0	0	0	1	-360	360;
	398	78	0.007248	0.021107	0	0	0	0	0	0	1	-360	360;
	1751	1853	0.036653	0.106661	0	0	0	0	0	0	1	-360	360;
	4741	4072	0.003107	0.007727	0	0	0	0	0	0	1	-360	360;
	4072	5569	0.013612	0.029917	0	0	0	0	0	0	1	-360	360;
	7304	5231	0.064182	0.141124	0	0	0	0	0	0	1	-360	360;
	6208	5231	0.058331	0.150579	0	0	0	0	0	0	1	-360	360;
	7271	4990	0.00337	0.03668	0	400	0	0	0	0	1	-360	360;
	4068	7839	0.033959	0.080934	0	0	0	0	0	0	1	-360	360;
	7040	7536	0	0.000826	0	0	0	0	0	0	1	-360	360;
	6208	7536	0.07314	0.181438	0	0	0	0	0	0	1	-360	360;
	7536	5728	0.01476	0.04295	0	0	0	0	0	0	1	-360	360;
	934	8070	0.018488	0.044091	0	0	0	0	0	0	1	-360	360;
	6703	4903	0.00025	0.00227	0	1283	0	0	0	0	1	-360	360;
	6703	4903	0.00025	0.00222	0	1415	0	0	0	0	1	-360	360;
	4903	1211	0.00179	0.01494	0	1020	0	0	0	0	1	-360	360;
	1550	2904	0.00133	0.01403	0	1218	0	0	0	0	1	-360	360;
	1047	1712	0.027438	0.076752	0	0	0	0	0	0	1	-360	360;
	7585	3881	0	0.00081	0	0	0	0	0	0	1	-360	360;
	3881	6918	0.014347	0.03705	0	0	0	0	0	0	1	-360	360;
	7660	5185	0.00286	0.015591	0	419	0	0	0	0	1	-360	360;
	6363	5569	0	0.000793	0	0	0	0	0	0	1	-360	360;
	5569	4378	0.008231	0.021273	0	0	0	0	0	0	1	-360	360;
	1254	7356	0.013653	0.031231	0	0	0	0	0	0	1	-360	360;
	3161	2645	0	0.000777	0	0	0	0	0	0	1	-360	360;
	8720	1053	0.028364	0.063479	0	0	0	0	0	0	1	-360	360;
	4029	1053	0.023636	0.051983	0	0	0	0	0	0	1	-360	360;
	468	1053	0.022752	0.066248	0	0	0	0	0	0	1	-360	360;
	6422	5632	0.004512	0.013099	0	0	0	0	0	0	1	-360	360;
	6422	2333	0.006331	0.015124	0	0	0	0	0	0	1	-360	360;
	5135	941	0.048058	0.143504	0	0	0	0	0	0	1	-360	360;
	7606	2643	0.02676	0.077901	0	0	0	0	0	0	1	-360	360;
	2706	576	0	0.000744	0	0	0	0	0	0	1	-360	360;
	4676	1211	0.00031	0.00329	0	1251	0	0	0	0	1	-360	360;
	4676	1211	0.00035	0.00322	0	1218	0	0	0	0	1	-360	360;
	5906	700	0.01276	0.028033	0	0	0	0	0	0	1	-360	360;
	7372	7747	0.049983	0.088198	0	0	0	0	0	0	1	-360	360;
	4862	3700	0	0.000876	0	0	0	0	0	0	1	-360	360;
	4263	3700	0.020157	0.04643	0	0	0	0	0	0	1	-360	360;
	8013	8192	0.050554	0.09781	0	0	0	0	0	0	1	-360	360;
	6141	8013	0.050157	0.093818	0	0	0	0	0	0	1	-360	360;
	7604	3914	0.03338	0.08238	0	0	0	0	0	0	1	-360	360;
	24	3914	0.041554	0.156107	0	0	0	0	0	0	1	-360	360;
	3748	6572	0.04676	0.124372	0	0	0	0	0	0	1	-360	360;
	2149	6572	0.043868	0.116678	0	0	0	0	0	0	1	-360	360;
	6878	5727	0.015529	0.049017	0	0	0	0	0	0	1	-360	360;
	6878	1427	0.019471	0.052802	0	0	0	0	0	0	1	-360	360;
	4763	5554	0.030339	0.062124	0	0	0	0	0	0	1	-360	360;
	1410	4763	0.043488	0.11219	0	0	0	0	0	0	1	-360	360;
	4763	5865	0.025934	0.052198	0	0	0	0	0	0	1	-360	360;
	7765	7534	0.024372	0.057347	0	0	0	0	0	0	1	-360	360;
	7765	6673	0.023868	0.058446	0	0	0	0	0	0	1	-360	360;
	7722	1427	0.032636	0.103934	0	0	0	0	0	0	1	-360	360;
	8969	7722	0.02695	0.081719	0	0	0	0	0	0	1	-360	360;
	5733	4437	0.022521	0.13874	0	457	0	0	0	0	1	-360	360;
	5733	3103	0.0117	0.053091	0	438	0	0	0	0	1	-360	360;
	5733	5043	0.001401	0.006789	0	476	0	0	0	0	1	-360	360;
	5733	2402	0.00139	0.0113	0	457	0	0	0	0	1	-360	360;
	1964	5733	0.01818	0.095349	0	457	0	0	0	0	1	-360	360;
	7978	5733	0.005729	0.043599	0	629	0	0	0	0	1	-360	360;
	5733	9168	0.00095	0.006409	0	0	0	0	0	0	1	-360	360;
	5733	1390	0.00093	0.006531	0	629	0	0	0	0	1	-360	360;
	5733	4899	0.025659	0.11894	0	438	0	0	0	0	1	-360	360;
	5733	694	0.01036	0.04869	0	533	0	0	0	0	1	-360	360;
	5153	2731	0.022231	0.051521	0	0	0	0	0	0	1	-360	360;
	1087	2731	0.038099	0.092579	0	0	0	0	0	0	1	-360	360;
	3548	1289	0.018041	0.066793	0	0	0	0	0	0	1	-360	360;
	4784	1289	0.019942	0.075554	0	0	0	0	0	0	1	-360	360;
	186	4406	0.028347	0.064727	0	0	0	0	0	0	1	-360	360;
	186	4671	0.028124	0.066826	0	0	0	0	0	0	1	-360	360;
	6352	1299	0.017107	0.031438	0	0	0	0	0	0	1	-360	360;
	7705	1299	0.028322	0.109331	0	0	0	0	0	0	1	-360	360;
	8430	7590	0	0.000413	0	0	0	0	0	0	1	-360	360;
	8430	1528	0.002248	0.008413	0	0	0	0	0	0	1	-360	360;
	7590	2106	0.003421	0.010397	0	0	0	0	0	0	1	-360	360;
	2977	7590	0.039554	0.071554	0	0	0	0	0	0	1	-360	360;
	3399	7590	0.012612	0.036818	0	0	0	0	0	0	1	-360	360;
	5931	155	0.004471	0.014041	0	0	0	0	0	0	1	-360	360;
	155	3924	0.013628	0.050934	0	0	0	0	0	0	1	-360	360;
	155	877	0.002438	0.007959	0	0	0	0	0	0	1	-360	360;
	155	3394	0.014579	0.030132	0	0	0	0	0	0	1	-360	360;
	155	8418	0.04186	0.08719	0	0	0	0	0	0	1	-360	360;
	155	8051	0.072653	0.128339	0	0	0	0	0	0	1	-360	360;
	7730	940	0.013612	0.049421	0	0	0	0	0	0	1	-360	360;
	7730	7970	0.037256	0.107727	0	0	0	0	0	0	1	-360	360;
	1964	2939	0.003349	0.0237	0	0	0	0	0	0	1	-360	360;
	1964	2939	0.003281	0.024159	0	0	0	0	0	0	1	-360	360;
	1964	2939	0.003599	0.023469	0	0	0	0	0	0	1	-360	360;
	1964	2939	0.003669	0.02299	0	572	0	0	0	0	1	-360	360;
	341	918	0.009124	0.02781	0	0	0	0	0	0	1	-360	360;
	341	918	0.008661	0.026479	0	0	0	0	0	0	1	-360	360;
	341	918	0.008488	0.027033	0	0	0	0	0	0	1	-360	360;
	341	1795	0.002248	0.006554	0	0	0	0	0	0	1	-360	360;
	341	5057	0.001612	0.004636	0	0	0	0	0	0	1	-360	360;
	341	1528	0.012223	0.027331	0	0	0	0	0	0	1	-360	360;
	341	5397	0.002099	0.01181	0	0	0	0	0	0	1	-360	360;
	341	5397	0.002058	0.012058	0	0	0	0	0	0	1	-360	360;
	918	1377	0.019008	0.055876	0	0	0	0	0	0	1	-360	360;
	918	2864	0.020727	0.05795	0	0	0	0	0	0	1	-360	360;
	2977	1441	0.004471	0.010471	0	0	0	0	0	0	1	-360	360;
	8617	1441	0.010479	0.025967	0	0	0	0	0	0	1	-360	360;
	1795	5057	0.001636	0.004537	0	0	0	0	0	0	1	-360	360;
	2106	5931	0.001603	0.005256	0	0	0	0	0	0	1	-360	360;
	8470	303	0.001039	0.00812	0	591	0	0	0	0	1	-360	360;
	8470	303	0.001169	0.007961	0	629	0	0	0	0	1	-360	360;
	940	3924	0.001802	0.006628	0	0	0	0	0	0	1	-360	360;
	5327	9016	0.00022	0.00217	0	1283	0	0	0	0	1	-360	360;
	1686	9016	0.00067	0.0072	0	1152	0	0	0	0	1	-360	360;
	2568	2977	0.019826	0.06962	0	0	0	0	0	0	1	-360	360;
	1964	4437	0.005919	0.02874	0	533	0	0	0	0	1	-360	360;
	2345	4437	0.001271	0.009651	0	514	0	0	0	0	1	-360	360;
	8470	4437	0.00126	0.00955	0	572	0	0	0	0	1	-360	360;
	7364	5327	0.0009	0.01149	0	1251	0	0	0	0	1	-360	360;
	5327	8358	0.00115	0.01172	0	1185	0	0	0	0	1	-360	360;
	2345	1753	0.00431	0.033729	0	457	0	0	0	0	1	-360	360;
	8470	6350	0.001841	0.0115	0	0	0	0	0	0	1	-360	360;
	3103	7978	0.008401	0.053719	0	572	0	0	0	0	1	-360	360;
	3900	3103	0.0013	0.006021	0	419	0	0	0	0	1	-360	360;
	9036	8471	0.016083	0.036347	0	0	0	0	0	0	1	-360	360;
	8471	4399	0.01576	0.037058	0	0	0	0	0	0	1	-360	360;
	6262	8471	0.039661	0.101099	0	0	0	0	0	0	1	-360	360;
	3745	379	0.015777	0.058099	0	0	0	0	0	0	1	-360	360;
	4803	379	0.044901	0.08681	0	0	0	0	0	0	1	-360	360;
	4824	2855	0.00173	0.01919	0	1251	0	0	0	0	1	-360	360;
	2855	8562	0.00213	0.02194	0	1481	0	0	0	0	1	-360	360;
	223	2023	0.00395	0.008331	0	0	0	0	0	0	1	-360	360;
	8308	2023	0.157273	0.234413	0	0	0	0	0	0	1	-360	360;
	223	3864	0.003421	0.009099	0	0	0	0	0	0	1	-360	360;
	3864	6150	0.06795	0.151058	0	0	0	0	0	0	1	-360	360;
	223	3162	0.040099	0.154554	0	0	0	0	0	0	1	-360	360;
	223	4678	0.070769	0.175471	0	0	0	0	0	0	1	-360	360;
	223	3203	0.008504	0.028322	0	0	0	0	0	0	1	-360	360;
	223	9124	0.02738	0.069876	0	0	0	0	0	0	1	-360	360;
	989	8783	0.010347	0.023041	0	0	0	0	0	0	1	-360	360;
	8969	8783	0.044281	0.083967	0	0	0	0	0	0	1	-360	360;
	1087	989	0.055744	0.101992	0	0	0	0	0	0	1	-360	360;
	1699	4406	0.015446	0.026554	0	0	0	0	0	0	1	-360	360;
	4406	1495	0.035388	0.067603	0	0	0	0	0	0	1	-360	360;
	3026	7301	0.007273	0.013752	0	0	0	0	0	0	1	-360	360;
	7301	6988	0.00362	0.007033	0	0	0	0	0	0	1	-360	360;
	4169	6988	0.001182	0.003727	0	0	0	0	0	0	1	-360	360;
	4169	4465	0.013314	0.052752	0	0	0	0	0	0	1	-360	360;
	3026	4547	0.028372	0.075868	0	0	0	0	0	0	1	-360	360;
	5344	3026	0.033917	0.080413	0	0	0	0	0	0	1	-360	360;
	8734	4361	0.004669	0.03695	0	0	0	0	0	0	1	-360	360;
	4361	7978	0.004229	0.025271	0	0	0	0	0	0	1	-360	360;
	7604	2638	0.041909	0.072157	0	0	0	0	0	0	1	-360	360;
	3122	2638	0.031174	0.078554	0	0	0	0	0	0	1	-360	360;
	2554	3548	0.030777	0.114017	0	0	0	0	0	0	1	-360	360;
	1665	6755	0.021479	0.05195	0	0	0	0	0	0	1	-360	360;
	6755	1427	0.04881	0.09786	0	0	0	0	0	0	1	-360	360;
	7243	1178	0.043992	0.107769	0	0	0	0	0	0	1	-360	360;
	7243	1427	0.017347	0.059826	0	0	0	0	0	0	1	-360	360;
	2976	6988	0.006281	0.014802	0	0	0	0	0	0	1	-360	360;
	4547	2976	0.045818	0.086281	0	0	0	0	0	0	1	-360	360;
	6673	2976	0.022521	0.077653	0	0	0	0	0	0	1	-360	360;
	2123	5547	0.014132	0.053744	0	0	0	0	0	0	1	-360	360;
	801	2123	0.025901	0.082504	0	0	0	0	0	0	1	-360	360;
	2473	2123	0.05905	0.233421	0	0	0	0	0	0	1	-360	360;
	7	8407	0.035174	0.062992	0	0	0	0	0	0	1	-360	360;
	7	7208	0.015901	0.059529	0	0	0	0	0	0	1	-360	360;
	8407	975	0.005174	0.015322	0	0	0	0	0	0	1	-360	360;
	8407	2348	0.021471	0.063207	0	0	0	0	0	0	1	-360	360;
	975	2037	0.030033	0.114521	0	0	0	0	0	0	1	-360	360;
	975	5629	0.050537	0.090521	0	0	0	0	0	0	1	-360	360;
	975	5629	0.026446	0.070959	0	0	0	0	0	0	1	-360	360;
	4136	3839	0.028157	0.075636	0	0	0	0	0	0	1	-360	360;
	4136	8357	0.021413	0.212132	0	0	0	0	0	0	1	-360	360;
	4136	8357	0.046446	0.096702	0	0	0	0	0	0	1	-360	360;
	7900	4639	0.042562	0.102702	0	0	0	0	0	0	1	-360	360;
	7528	4639	0.034314	0.071603	0	0	0	0	0	0	1	-360	360;
	4727	8617	0.034033	0.077537	0	0	0	0	0	0	1	-360	360;
	1410	4727	0.01919	0.035669	0	0	0	0	0	0	1	-360	360;
	7604	8192	0.046826	0.094587	0	0	0	0	0	0	1	-360	360;
	8990	2947	0.059909	0.111868	0	0	0	0	0	0	1	-360	360;
	7477	8990	0.020909	0.037934	0	0	0	0	0	0	1	-360	360;
	5875	6843	0.009727	0.025347	0	0	0	0	0	0	1	-360	360;
	4814	5875	0.01981	0.040331	0	0	0	0	0	0	1	-360	360;
	2687	5557	1.7e-05	0.000413	0	0	0	0	0	0	1	-360	360;
	591	2687	0.017091	0.04114	0	0	0	0	0	0	1	-360	360;
	2568	877	0.03014	0.111744	0	0	0	0	0	0	1	-360	360;
	3637	1699	0.012628	0.026339	0	0	0	0	0	0	1	-360	360;
	1246	862	0.032934	0.078298	0	0	0	0	0	0	1	-360	360;
	1960	1246	0.025198	0.050942	0	0	0	0	0	0	1	-360	360;
	2802	7323	0.025529	0.049421	0	0	0	0	0	0	1	-360	360;
	6262	7323	0.026124	0.06962	0	0	0	0	0	0	1	-360	360;
	393	6461	0.00036	0.00188	0	0	0	0	0	0	1	-360	360;
	393	6461	0.00043	0.00174	0	0	0	0	0	0	1	-360	360;
	8382	5269	0.026091	0.052298	0	0	0	0	0	0	1	-360	360;
	8382	5269	0.015413	0.048636	0	0	0	0	0	0	1	-360	360;
	5269	6628	0.024463	0.08705	0	0	0	0	0	0	1	-360	360;
	6461	3688	0.00638	0.04836	0	0	0	0	0	0	1	-360	360;
	6461	5000	0.002081	0.01214	0	533	0	0	0	0	1	-360	360;
	6461	5000	0.002081	0.01257	0	514	0	0	0	0	1	-360	360;
	8942	2861	0.005314	0.015017	0	0	0	0	0	0	1	-360	360;
	8942	4671	0.004331	0.011752	0	0	0	0	0	0	1	-360	360;
	2861	4671	0.00386	0.013322	0	0	0	0	0	0	1	-360	360;
	4671	3839	0.02181	0.080802	0	0	0	0	0	0	1	-360	360;
	4671	1714	0.036992	0.107727	0	0	0	0	0	0	1	-360	360;
	4671	5190	0.029273	0.105248	0	0	0	0	0	0	1	-360	360;
	6673	4465	0.016909	0.05057	0	0	0	0	0	0	1	-360	360;
	8474	4824	0.00302	0.03553	0	1251	0	0	0	0	1	-360	360;
	4678	7607	0.005983	0.023132	0	0	0	0	0	0	1	-360	360;
	7149	7607	0.003777	0.011421	0	0	0	0	0	0	1	-360	360;
	3203	7607	0.050041	0.180397	0	0	0	0	0	0	1	-360	360;
	7607	9124	0.056562	0.179397	0	0	0	0	0	0	1	-360	360;
	6448	2195	0.012818	0.046769	0	0	0	0	0	0	1	-360	360;
	8618	6350	0.00286	0.017581	0	514	0	0	0	0	1	-360	360;
	586	9036	0.012554	0.029083	0	0	0	0	0	0	1	-360	360;
	2015	2982	0.000909	0.002876	0	0	0	0	0	0	1	-360	360;
	6626	586	0.018719	0.033769	0	0	0	0	0	0	1	-360	360;
	3136	6626	0.032149	0.059041	0	0	0	0	0	0	1	-360	360;
	8096	4986	0.005198	0.013488	0	0	0	0	0	0	1	-360	360;
	2947	8096	0.018636	0.044149	0	0	0	0	0	0	1	-360	360;
	2947	4986	0.014554	0.028438	0	0	0	0	0	0	1	-360	360;
	7477	2947	0.032934	0.155826	0	0	0	0	0	0	1	-360	360;
	2947	6610	0.015231	0.035843	0	0	0	0	0	0	1	-360	360;
	2947	9193	0.037248	0.089901	0	0	0	0	0	0	1	-360	360;
	2947	430	0.04643	0.09338	0	0	0	0	0	0	1	-360	360;
	245	8481	0.075612	0.155504	0	0	0	0	0	0	1	-360	360;
	8481	4343	0.04057	0.08876	0	0	0	0	0	0	1	-360	360;
	8481	329	0.114628	0.20105	0	0	0	0	0	0	1	-360	360;
	4269	9219	0.018521	0.090322	0	0	0	0	0	0	1	-360	360;
	2864	9219	0.013702	0.052983	0	0	0	0	0	0	1	-360	360;
	1731	7848	0.006017	0.019198	0	0	0	0	0	0	1	-360	360;
	1731	6827	0.052686	0.090653	0	0	0	0	0	0	1	-360	360;
	1731	2195	0.065826	0.122818	0	0	0	0	0	0	1	-360	360;
	6673	7534	0.04443	0.105719	0	0	0	0	0	0	1	-360	360;
	7739	7354	0.00838	0.018942	0	0	0	0	0	0	1	-360	360;
	8306	7354	0.058017	0.142983	0	0	0	0	0	0	1	-360	360;
	1534	7830	0.00085	0.01148	0	856	0	0	0	0	1	-360	360;
	8562	7830	0.00172	0.0234	0	1283	0	0	0	0	1	-360	360;
	7604	7739	0.04805	0.152256	0	0	0	0	0	0	1	-360	360;
	7739	5830	0.028372	0.069471	0	0	0	0	0	0	1	-360	360;
	6843	7739	0.038364	0.12214	0	0	0	0	0	0	1	-360	360;
	3748	5547	0.013463	0.05405	0	0	0	0	0	0	1	-360	360;
	3394	8418	0.025083	0.060273	0	0	0	0	0	0	1	-360	360;
	3523	1372	0.012438	0.047124	0	0	0	0	0	0	1	-360	360;
	8969	3523	0.022388	0.084818	0	0	0	0	0	0	1	-360	360;
	2554	3011	0.005471	0.020281	0	0	0	0	0	0	1	-360	360;
	4784	3011	0.107174	0.201248	0	0	0	0	0	0	1	-360	360;
	5153	2554	0.035058	0.124769	0	0	0	0	0	0	1	-360	360;
	2037	7848	0.015934	0.05876	0	0	0	0	0	0	1	-360	360;
	591	5557	0.020099	0.03862	0	0	0	0	0	0	1	-360	360;
	2218	4034	0.04176	0.103083	0	0	0	0	0	0	1	-360	360;
	3713	4034	0.066719	0.165537	0	0	0	0	0	0	1	-360	360;
	5000	4997	0.00013	0.00095	0	0	0	0	0	0	1	-360	360;
	2802	3300	0.033777	0.06776	0	0	0	0	0	0	1	-360	360;
	8261	3300	0.057388	0.126289	0	0	0	0	0	0	1	-360	360;
	1534	9054	0.0019	0.0204	0	1218	0	0	0	0	1	-360	360;
	8306	8423	0.007529	0.025372	0	0	0	0	0	0	1	-360	360;
	8306	1661	0.023917	0.064636	0	0	0	0	0	0	1	-360	360;
	8306	5865	0.039612	0.099818	0	0	0	0	0	0	1	-360	360;
	8306	7414	0.012388	0.038802	0	0	0	0	0	0	1	-360	360;
	8423	7414	0.004281	0.015678	0	0	0	0	0	0	1	-360	360;
	7595	5344	0.013463	0.033289	0	0	0	0	0	0	1	-360	360;
	8570	5344	0.015893	0.039901	0	0	0	0	0	0	1	-360	360;
	8570	7595	0.005934	0.021686	0	0	0	0	0	0	1	-360	360;
	3770	7595	0.009107	0.02843	0	0	0	0	0	0	1	-360	360;
	8570	3770	0.005802	0.020083	0	0	0	0	0	0	1	-360	360;
	8570	1806	0.10495	0.203661	0	0	0	0	0	0	1	-360	360;
	8570	1018	0.026372	0.04995	0	0	0	0	0	0	1	-360	360;
	663	8734	0.00738	0.044	0	0	0	0	0	0	1	-360	360;
	8734	7978	0.010591	0.05825	0	514	0	0	0	0	1	-360	360;
	4861	663	0.01318	0.05764	0	514	0	0	0	0	1	-360	360;
	663	4392	0.0143	0.06869	0	476	0	0	0	0	1	-360	360;
	7604	2201	0.022099	0.080636	0	0	0	0	0	0	1	-360	360;
	1087	1665	0.039843	0.08557	0	0	0	0	0	0	1	-360	360;
	1665	8437	0.007909	0.022124	0	0	0	0	0	0	1	-360	360;
	8343	24	0.051017	0.111273	0	0	0	0	0	0	1	-360	360;
	6843	1661	0.010331	0.038289	0	0	0	0	0	0	1	-360	360;
	862	2568	0.02343	0.058603	0	0	0	0	0	0	1	-360	360;
	8562	1978	0.0009	0.00851	0	1251	0	0	0	0	1	-360	360;
	1978	6098	0.00157	0.01488	0	1086	0	0	0	0	1	-360	360;
	4861	3104	0.00682	0.03224	0	0	0	0	0	0	1	-360	360;
	7477	7900	0.01719	0.033322	0	0	0	0	0	0	1	-360	360;
	7900	7528	0.080446	0.14995	0	0	0	0	0	0	1	-360	360;
	4803	245	0.055678	0.123917	0	0	0	0	0	0	1	-360	360;
	1228	1714	0.014463	0.040066	0	0	0	0	0	0	1	-360	360;
	1228	7938	0.013107	0.048314	0	0	0	0	0	0	1	-360	360;
	4343	1960	0.123174	0.23414	0	0	0	0	0	0	1	-360	360;
	3162	6150	0.00557	0.011256	0	0	0	0	0	0	1	-360	360;
	3162	2633	0.007967	0.018719	0	0	0	0	0	0	1	-360	360;
	8308	2633	0.076711	0.159554	0	0	0	0	0	0	1	-360	360;
	551	2348	0.012669	0.049364	0	0	0	0	0	0	1	-360	360;
	551	5195	0.015719	0.058826	0	0	0	0	0	0	1	-360	360;
	6709	1408	0.007397	0.027686	0	0	0	0	0	0	1	-360	360;
	6709	2195	0.006967	0.02505	0	0	0	0	0	0	1	-360	360;
	5554	8665	0.037653	0.076876	0	0	0	0	0	0	1	-360	360;
	3745	7705	0.011	0.029826	0	0	0	0	0	0	1	-360	360;
	4784	5153	0.042446	0.107058	0	0	0	0	0	0	1	-360	360;
	4678	7149	0.011008	0.025711	0	0	0	0	0	0	1	-360	360;
	6141	7528	0.014818	0.04719	0	0	0	0	0	0	1	-360	360;
	6141	7528	0.011802	0.040711	0	0	0	0	0	0	1	-360	360;
	7528	6141	0.010802	0.039769	0	0	0	0	0	0	1	-360	360;
	6141	8015	0.003331	0.011496	0	0	0	0	0	0	1	-360	360;
	6141	8343	0.041256	0.083702	0	0	0	0	0	0	1	-360	360;
	8015	7528	0.011223	0.030413	0	0	0	0	0	0	1	-360	360;
	7528	2015	0.105562	0.208901	0	0	0	0	0	0	1	-360	360;
	7528	2218	0.021612	0.065496	0	0	0	0	0	0	1	-360	360;
	8562	2398	0.0056	0.03178	0	1053	0	0	0	0	1	-360	360;
	2864	1377	0.005364	0.01443	0	0	0	0	0	0	1	-360	360;
	6043	8364	0.007132	0.016347	0	0	0	0	0	0	1	-360	360;
	8364	2195	0.035281	0.111719	0	0	0	0	0	0	1	-360	360;
	1806	6043	0.027669	0.067686	0	0	0	0	0	0	1	-360	360;
	6043	3136	0.086157	0.181149	0	0	0	0	0	0	1	-360	360;
	4555	8382	0.019983	0.054711	0	0	0	0	0	0	1	-360	360;
	4555	4269	0.007041	0.032174	0	0	0	0	0	0	1	-360	360;
	3738	6352	0.050587	0.128397	0	0	0	0	0	0	1	-360	360;
	8382	6471	0.014983	0.031364	0	0	0	0	0	0	1	-360	360;
	6471	6278	0.010653	0.019	0	0	0	0	0	0	1	-360	360;
	5684	4801	0.00189	0.02064	0	1349	0	0	0	0	1	-360	360;
	5684	4801	0.00227	0.01922	0	1251	0	0	0	0	1	-360	360;
	4801	895	0.00167	0.02044	0	1283	0	0	0	0	1	-360	360;
	1964	3372	0.016341	0.07818	0	495	0	0	0	0	1	-360	360;
	7521	3372	0.005409	0.024729	0	533	0	0	0	0	1	-360	360;
	3372	1473	0.00525	0.04099	0	553	0	0	0	0	1	-360	360;
	8382	3410	0.00714	0.025347	0	0	0	0	0	0	1	-360	360;
	8382	3756	0.00819	0.025802	0	0	0	0	0	0	1	-360	360;
	6278	8382	0.006322	0.014314	0	0	0	0	0	0	1	-360	360;
	5218	1087	0.037678	0.087512	0	0	0	0	0	0	1	-360	360;
	5218	7200	0.010942	0.034529	0	0	0	0	0	0	1	-360	360;
	3410	7200	0.007041	0.02714	0	0	0	0	0	0	1	-360	360;
	7200	3756	0.002826	0.010496	0	0	0	0	0	0	1	-360	360;
	7364	5684	0.0001	0.00124	0	1349	0	0	0	0	1	-360	360;
	7364	4447	3e-05	0.00027	0	1382	0	0	0	0	1	-360	360;
	7364	6248	2e-05	0.00021	0	757	0	0	0	0	1	-360	360;
	7364	5684	0.00011	0.00133	0	1251	0	0	0	0	1	-360	360;
	7364	9054	0.00062	0.00694	0	1251	0	0	0	0	1	-360	360;
	1964	465	0.005281	0.04189	0	610	0	0	0	0	1	-360	360;
	1964	3597	0.00014	0.000651	0	0	0	0	0	0	1	-360	360;
	1964	1139	0.000229	0.00107	0	0	0	0	0	0	1	-360	360;
	1964	3288	0.000229	0.001039	0	0	0	0	0	0	1	-360	360;
	1964	894	0.00013	0.00086	0	629	0	0	0	0	1	-360	360;
	1964	7627	0.00013	0.00087	0	400	0	0	0	0	1	-360	360;
	1964	5731	0.001539	0.012651	0	0	0	0	0	0	1	-360	360;
	1964	694	0.00806	0.047091	0	0	0	0	0	0	1	-360	360;
	8437	1178	0.04943	0.099331	0	0	0	0	0	0	1	-360	360;
	180	329	0.050488	0.104149	0	0	0	0	0	0	1	-360	360;
	6628	7468	0.003702	0.008934	0	0	0	0	0	0	1	-360	360;
	1087	7468	0.074446	0.138033	0	0	0	0	0	0	1	-360	360;
	3748	8418	0.008479	0.021653	0	0	0	0	0	0	1	-360	360;
	3748	8418	0.005314	0.019479	0	0	0	0	0	0	1	-360	360;
	8474	1686	0.00104	0.01121	0	1448	0	0	0	0	1	-360	360;
	1473	8618	0.01118	0.07193	0	438	0	0	0	0	1	-360	360;
	5830	3122	0.002777	0.008174	0	0	0	0	0	0	1	-360	360;
	7208	6827	0.006058	0.02362	0	0	0	0	0	0	1	-360	360;
	2398	7314	0.00192	0.01477	0	1086	0	0	0	0	1	-360	360;
	7848	6827	0.025331	0.098711	0	0	0	0	0	0	1	-360	360;
	4803	3917	0.05886	0.113157	0	0	0	0	0	0	1	-360	360;
	504	3917	0.046851	0.110231	0	0	0	0	0	0	1	-360	360;
	801	4784	0.02895	0.10819	0	0	0	0	0	0	1	-360	360;
	1063	8308	0.079438	0.11481	0	0	0	0	0	0	1	-360	360;
	1063	8308	0.077967	0.117314	0	0	0	0	0	0	1	-360	360;
	8308	6843	0.032669	0.091388	0	0	0	0	0	0	1	-360	360;
	876	2015	0.008421	0.02657	0	0	0	0	0	0	1	-360	360;
	2218	876	0.082727	0.14876	0	0	0	0	0	0	1	-360	360;
	4392	5000	0.00637	0.029669	0	514	0	0	0	0	1	-360	360;
	4131	1063	0.020223	0.033198	0	0	0	0	0	0	1	-360	360;
	4131	1063	0.017843	0.033876	0	0	0	0	0	0	1	-360	360;
	6673	2802	0.022058	0.05405	0	0	0	0	0	0	1	-360	360;
	6673	8865	0.019826	0.073909	0	0	0	0	0	0	1	-360	360;
	9193	1495	0.063107	0.131711	0	0	0	0	0	0	1	-360	360;
	5127	1714	0.012099	0.029041	0	0	0	0	0	0	1	-360	360;
	5127	1714	0.006273	0.019033	0	0	0	0	0	0	1	-360	360;
	5127	7687	0.007603	0.021273	0	0	0	0	0	0	1	-360	360;
	5127	7687	0.011041	0.027339	0	0	0	0	0	0	1	-360	360;
	5312	3725	0.011636	0.032562	0	0	0	0	0	0	1	-360	360;
	5312	1714	0.007091	0.019058	0	0	0	0	0	0	1	-360	360;
	3725	7687	0.006653	0.020967	0	0	0	0	0	0	1	-360	360;
	2349	5000	0.00212	0.01695	0	514	0	0	0	0	1	-360	360;
	2349	5000	0.002349	0.0155	0	0	0	0	0	0	1	-360	360;
	1714	8357	0.053901	0.092744	0	0	0	0	0	0	1	-360	360;
	1358	5629	0.003273	0.010033	0	0	0	0	0	0	1	-360	360;
	1358	2505	0.005066	0.013769	0	0	0	0	0	0	1	-360	360;
	5629	2505	0.007826	0.028587	0	0	0	0	0	0	1	-360	360;
	2505	5195	0.032769	0.058711	0	0	0	0	0	0	1	-360	360;
	8051	7970	0.053702	0.134017	0	0	0	0	0	0	1	-360	360;
	5684	2585	0.00147	0.01305	0	1218	0	0	0	0	1	-360	360;
	4602	5684	0.00019	0.00195	0	1415	0	0	0	0	1	-360	360;
	5985	5684	0.00019	0.00191	0	1448	0	0	0	0	1	-360	360;
	4815	3016	0.006413	0.019917	0	0	0	0	0	0	1	-360	360;
	4131	3016	0.011562	0.044182	0	0	0	0	0	0	1	-360	360;
	6034	4815	0.002512	0.008421	0	0	0	0	0	0	1	-360	360;
	493	3738	0.041702	0.098711	0	0	0	0	0	0	1	-360	360;
	504	493	0.034347	0.061636	0	0	0	0	0	0	1	-360	360;
	4803	5238	0.004289	0.01243	0	0	0	0	0	0	1	-360	360;
	5995	5238	0.000686	0.001471	0	0	0	0	0	0	1	-360	360;
	4803	5775	0.006612	0.01886	0	0	0	0	0	0	1	-360	360;
	7783	5775	0.008388	0.023339	0	0	0	0	0	0	1	-360	360;
	4803	5995	0.004372	0.012174	0	0	0	0	0	0	1	-360	360;
	4803	7783	0.013413	0.047793	0	0	0	0	0	0	1	-360	360;
	1087	1372	0.08438	0.183769	0	0	0	0	0	0	1	-360	360;
	2802	4015	0.006372	0.01681	0	0	0	0	0	0	1	-360	360;
	7938	5195	0.028579	0.084149	0	0	0	0	0	0	1	-360	360;
	7938	4240	0.007008	0.022388	0	0	0	0	0	0	1	-360	360;
	8051	8566	0.003	0.005058	0	0	0	0	0	0	1	-360	360;
	274	8566	0.002669	0.005727	0	0	0	0	0	0	1	-360	360;
	8566	2295	0.003661	0.008818	0	0	0	0	0	0	1	-360	360;
	8566	4131	0.029843	0.069744	0	0	0	0	0	0	1	-360	360;
	274	3470	0.006438	0.01824	0	0	0	0	0	0	1	-360	360;
	274	3470	0.006322	0.018636	0	0	0	0	0	0	1	-360	360;
	8358	8474	0.00068	0.00797	0	1218	0	0	0	0	1	-360	360;
	3470	8665	0.032364	0.088289	0	0	0	0	0	0	1	-360	360;
	3470	4131	0.027066	0.084496	0	0	0	0	0	0	1	-360	360;
	3470	4131	0.026554	0.086256	0	0	0	0	0	0	1	-360	360;
	4131	2295	0.029901	0.071322	0	0	0	0	0	0	1	-360	360;
	591	504	0.011901	0.023818	0	0	0	0	0	0	1	-360	360;
	7705	504	0.012132	0.023322	0	0	0	0	0	0	1	-360	360;
	4899	7521	0.0098	0.062349	0	400	0	0	0	0	1	-360	360;
	591	6898	0.004182	0.012678	0	0	0	0	0	0	1	-360	360;
	7705	6898	0.003636	0.014636	0	0	0	0	0	0	1	-360	360;
	8665	4814	0.01919	0.045041	0	0	0	0	0	0	1	-360	360;
	5727	5208	0.009793	0.037174	0	0	0	0	0	0	1	-360	360;
	5208	1427	0.004281	0.010339	0	0	0	0	0	0	1	-360	360;
	5208	1427	0.005521	0.021471	0	0	0	0	0	0	1	-360	360;
	895	254	0.0011	0.01325	0	1349	0	0	0	0	1	-360	360;
	1087	8437	0.018421	0.051372	0	0	0	0	0	0	1	-360	360;
	1087	1003	0.011521	0.042529	0	0	0	0	0	0	1	-360	360;
	2568	3399	0.025182	0.071554	0	0	0	0	0	0	1	-360	360;
	8969	3232	0	0.000413	0	0	0	0	0	0	1	-360	360;
	1408	4220	0.014603	0.03386	0	0	0	0	0	0	1	-360	360;
	4383	2568	0.022066	0.054521	0	0	0	0	0	0	1	-360	360;
	1960	4383	0.041471	0.090793	0	0	0	0	0	0	1	-360	360;
	4131	6034	0.008843	0.031901	0	0	0	0	0	0	1	-360	360;
	3750	5000	0.00038	0.003091	0	533	0	0	0	0	1	-360	360;
	3750	5000	0.0005	0.002901	0	514	0	0	0	0	1	-360	360;
	7687	8357	0.032132	0.318198	0	0	0	0	0	0	1	-360	360;
	7687	2195	0.058554	0.146124	0	0	0	0	0	0	1	-360	360;
	5678	7687	0.023917	0.073289	0	0	0	0	0	0	1	-360	360;
	7687	4220	0.040818	0.089273	0	0	0	0	0	0	1	-360	360;
	2398	2253	2e-05	0.00025	0	1185	0	0	0	0	1	-360	360;
	2398	2253	2e-05	0.00026	0	1086	0	0	0	0	1	-360	360;
	2253	2737	0.00082	0.00822	0	1086	0	0	0	0	1	-360	360;
	2201	1018	0.014339	0.031777	0	0	0	0	0	0	1	-360	360;
	4784	6358	0.011364	0.034364	0	0	0	0	0	0	1	-360	360;
	6358	2988	0.000347	0.001066	0	0	0	0	0	0	1	-360	360;
	4240	8357	0.012752	0.036041	0	0	0	0	0	0	1	-360	360;
	5190	8357	0.000669	0.001281	0	0	0	0	0	0	1	-360	360;
	5190	8357	0.000661	0.001314	0	0	0	0	0	0	1	-360	360;
	8865	4015	0.005289	0.014942	0	0	0	0	0	0	1	-360	360;
	1960	8261	0.004256	0.007364	0	0	0	0	0	0	1	-360	360;
	180	8261	0.003992	0.009107	0	0	0	0	0	0	1	-360	360;
	6610	3637	0.041488	0.08662	0	0	0	0	0	0	1	-360	360;
	3637	9193	0.00486	0.010132	0	0	0	0	0	0	1	-360	360;
	3637	430	0.00495	0.009934	0	0	0	0	0	0	1	-360	360;
	430	9193	0.00505	0.009719	0	0	0	0	0	0	1	-360	360;
	1003	8437	0.00776	0.027521	0	0	0	0	0	0	1	-360	360;
	2149	2473	0.00262	0.007628	0	0	0	0	0	0	1	-360	360;
	5678	4220	0.001793	0.005471	0	0	0	0	0	0	1	-360	360;
	9124	3713	0.000537	0.015661	0	0	0	0	0	0	1	-360	360;
	4784	2988	0.01124	0.036826	0	0	0	0	0	0	1	-360	360;
	8556	1920	0.00156	0.01791	0	1152	0	0	0	0	1	-360	360;
	1920	6907	0.00155	0.01686	0	1283	0	0	0	0	1	-360	360;
	1920	6907	0.00152	0.01719	0	1218	0	0	0	0	1	-360	360;
	8263	1920	0.00068	0.00644	0	1415	0	0	0	0	1	-360	360;
	8263	1920	0.00069	0.0063	0	1448	0	0	0	0	1	-360	360;
	1920	5824	0.00112	0.01064	0	1711	0	0	0	0	1	-360	360;
	3447	4966	0.00151	0.011539	0	476	0	0	0	0	1	-360	360;
	8145	3447	0.00788	0.05236	0	476	0	0	0	0	1	-360	360;
	1542	6907	0.00128	0.0126	0	1251	0	0	0	0	1	-360	360;
	816	1542	0.00037	0.00409	0	1053	0	0	0	0	1	-360	360;
	816	1542	0.00036	0.00417	0	1152	0	0	0	0	1	-360	360;
	1542	5824	0.00122	0.0112	0	1020	0	0	0	0	1	-360	360;
	8145	8325	0.006591	0.04075	0	419	0	0	0	0	1	-360	360;
	2887	4966	0.007961	0.05756	0	419	0	0	0	0	1	-360	360;
	5092	2646	0.00059	0.0085	0	1053	0	0	0	0	1	-360	360;
	2646	5503	0.00084	0.0118	0	954	0	0	0	0	1	-360	360;
	2646	6446	0.00178	0.01804	0	790	0	0	0	0	1	-360	360;
	4961	2492	0.00937	0.073969	0	419	0	0	0	0	1	-360	360;
	4961	3197	0.011081	0.06593	0	400	0	0	0	0	1	-360	360;
	4961	2846	0.009669	0.04251	0	381	0	0	0	0	1	-360	360;
	6930	9030	0.00073	0.00713	0	790	0	0	0	0	1	-360	360;
	5262	5503	0.00034	0.00345	0	1053	0	0	0	0	1	-360	360;
	7448	7063	0.00012	0.00128	0	987	0	0	0	0	1	-360	360;
	7448	255	0.00012	0.00123	0	1020	0	0	0	0	1	-360	360;
	1723	2051	9e-05	0.00112	0	921	0	0	0	0	1	-360	360;
	1337	846	0.00062	0.00866	0	1448	0	0	0	0	1	-360	360;
	7448	1337	0.00157	0.01614	0	1448	0	0	0	0	1	-360	360;
	7448	1337	0.0014	0.01829	0	1415	0	0	0	0	1	-360	360;
	287	2807	0.00078	0.00805	0	1349	0	0	0	0	1	-360	360;
	2807	4111	0.00094	0.01427	0	1481	0	0	0	0	1	-360	360;
	7758	5503	0.00024	0.00283	0	1349	0	0	0	0	1	-360	360;
	5503	846	0.00099	0.00865	0	1251	0	0	0	0	1	-360	360;
	7448	5503	0.0011	0.01426	0	1316	0	0	0	0	1	-360	360;
	2492	689	0.006919	0.050521	0	419	0	0	0	0	1	-360	360;
	4055	5107	0.0006	0.00646	0	1349	0	0	0	0	1	-360	360;
	4055	5819	0.00127	0.0128	0	823	0	0	0	0	1	-360	360;
	4055	9107	0.00052	0.00496	0	1349	0	0	0	0	1	-360	360;
	4055	1118	0.00064	0.00886	0	1448	0	0	0	0	1	-360	360;
	4055	1118	0.00072	0.00868	0	1547	0	0	0	0	1	-360	360;
	287	8245	0.00232	0.02522	0	823	0	0	0	0	1	-360	360;
	7448	287	0.0003	0.00304	0	1448	0	0	0	0	1	-360	360;
	7448	287	0.00032	0.00306	0	1316	0	0	0	0	1	-360	360;
	7448	287	0.00028	0.00307	0	1448	0	0	0	0	1	-360	360;
	9030	7562	0.00092	0.00969	0	921	0	0	0	0	1	-360	360;
	9030	5819	0.00181	0.01696	0	1020	0	0	0	0	1	-360	360;
	9030	6446	0.00132	0.01148	0	1283	0	0	0	0	1	-360	360;
	7562	4111	0.00108	0.01087	0	1349	0	0	0	0	1	-360	360;
	5107	8245	0.00209	0.01924	0	823	0	0	0	0	1	-360	360;
	9107	5107	0.00036	0.00457	0	1053	0	0	0	0	1	-360	360;
	5107	1723	0.0017	0.01986	0	757	0	0	0	0	1	-360	360;
	846	2294	0.00059	0.00982	0	954	0	0	0	0	1	-360	360;
	8760	2846	0.00494	0.03182	0	0	0	0	0	0	1	-360	360;
	8760	2846	0.004841	0.03245	0	0	0	0	0	0	1	-360	360;
	4111	6446	0.00036	0.00391	0	1613	0	0	0	0	1	-360	360;
	1118	1723	0.00071	0.00983	0	921	0	0	0	0	1	-360	360;
	1118	1723	0.00081	0.00963	0	1020	0	0	0	0	1	-360	360;
	5389	794	0.006271	0.03645	0	0	0	0	0	0	1	-360	360;
	7724	4924	0.186281	0.374281	0	0	0	0	0	0	1	-360	360;
	8041	4924	0.089612	0.230339	0	0	0	0	0	0	1	-360	360;
	7870	4190	0.00393	0.04552	0	1316	0	0	0	0	1	-360	360;
	3940	6091	0.004169	0.02651	0	438	0	0	0	0	1	-360	360;
	3940	6091	0.003831	0.02438	0	457	0	0	0	0	1	-360	360;
	3940	5389	0.00789	0.04725	0	591	0	0	0	0	1	-360	360;
	3940	5389	0.008039	0.046219	0	381	0	0	0	0	1	-360	360;
	956	8067	0.040174	0.084397	0	0	0	0	0	0	1	-360	360;
	3518	607	0.00166	0.01276	0	1429	0	0	0	0	1	-360	360;
	7678	6091	0.00094	0.00599	0	381	0	0	0	0	1	-360	360;
	7678	6091	0.00107	0.00587	0	400	0	0	0	0	1	-360	360;
	1075	4063	0.011901	0.028364	0	0	0	0	0	0	1	-360	360;
	4063	8067	0.059091	0.102636	0	0	0	0	0	0	1	-360	360;
	8041	4063	0.090471	0.19919	0	0	0	0	0	0	1	-360	360;
	599	1683	0.00755	0.050919	0	438	0	0	0	0	1	-360	360;
	599	1683	0.0077	0.04988	0	514	0	0	0	0	1	-360	360;
	4242	7724	0.238893	0.453901	0	0	0	0	0	0	1	-360	360;
	446	4242	0.062719	0.104479	0	0	0	0	0	0	1	-360	360;
	5389	1683	0.000979	0.074909	0	476	0	0	0	0	1	-360	360;
	5389	1683	0.000961	0.076349	0	495	0	0	0	0	1	-360	360;
	5052	1683	0.002159	0.079031	0	419	0	0	0	0	1	-360	360;
	5052	1683	0.001961	0.09125	0	457	0	0	0	0	1	-360	360;
	446	7724	0.066636	0.183934	0	0	0	0	0	0	1	-360	360;
	4458	3464	0.00028	0.00282	0	1593	0	0	0	0	1	-360	360;
	4458	5920	0.00018	0.00233	0	1905	0	0	0	0	1	-360	360;
	813	5466	4e-05	0.0004	0	1212	0	0	0	0	1	-360	360;
	813	5757	0.0017	0.01741	0	1316	0	0	0	0	1	-360	360;
	813	5920	0.00097	0.0123	0	1940	0	0	0	0	1	-360	360;
	813	2028	0.00057	0.00725	0	1871	0	0	0	0	1	-360	360;
	4959	5466	0.00177	0.01811	0	1247	0	0	0	0	1	-360	360;
	4959	6929	0.00049	0.00505	0	1247	0	0	0	0	1	-360	360;
	4959	6588	0.00199	0.02534	0	1905	0	0	0	0	1	-360	360;
	4959	3464	0.00024	0.00242	0	1593	0	0	0	0	1	-360	360;
	530	4654	0.01379	0.0443	0	0	0	0	0	0	1	-360	360;
	530	6324	0.00615	0.04848	0	0	0	0	0	0	1	-360	360;
	530	8648	0.02224	0.07142	0	0	0	0	0	0	1	-360	360;
	530	5734	0.0084	0.06621	0	0	0	0	0	0	1	-360	360;
	8	3212	7e-05	0.00068	0	307	0	0	0	0	1	-360	360;
	8	6237	0.00669	0.05272	0	0	0	0	0	0	1	-360	360;
	8	8472	0.003	0.03068	0	0	0	0	0	0	1	-360	360;
	8	423	0.01129	0.05381	0	0	0	0	0	0	1	-360	360;
	3212	1664	0.02312	0.11021	0	0	0	0	0	0	1	-360	360;
	3212	8472	0.0044	0.03467	0	0	0	0	0	0	1	-360	360;
	1664	581	0.00113	0.00362	0	0	0	0	0	0	1	-360	360;
	1664	581	0.00113	0.00362	0	0	0	0	0	0	1	-360	360;
	1664	2370	0.00603	0.02877	0	280	0	0	0	0	1	-360	360;
	9034	6237	0.01127	0.0691	0	0	0	0	0	0	1	-360	360;
	9034	151	0.01447	0.04647	0	0	0	0	0	0	1	-360	360;
	9034	4654	0.00858	0.06762	0	0	0	0	0	0	1	-360	360;
	9034	7363	0.00026	0.00161	0	0	0	0	0	0	1	-360	360;
	6237	151	0.0037	0.02918	0	0	0	0	0	0	1	-360	360;
	6237	8105	0.00847	0.0404	0	0	0	0	0	0	1	-360	360;
	6237	8770	0.00026	0.00161	0	0	0	0	0	0	1	-360	360;
	151	8472	0.00935	0.03001	0	0	0	0	0	0	1	-360	360;
	151	8472	0.00553	0.02634	0	0	0	0	0	0	1	-360	360;
	151	4090	0.00318	0.0251	0	0	0	0	0	0	1	-360	360;
	151	2370	0.00689	0.05429	0	0	0	0	0	0	1	-360	360;
	7326	2171	0.00617	0.02941	0	0	0	0	0	0	1	-360	360;
	7326	6324	0.00476	0.0375	0	0	0	0	0	0	1	-360	360;
	2171	6266	0.0157	0.07482	0	0	0	0	0	0	1	-360	360;
	3305	7733	0.02731	0.08769	0	0	0	0	0	0	1	-360	360;
	3305	6388	0.03423	0.10993	0	0	0	0	0	0	1	-360	360;
	4654	6536	0.00112	0.01141	0	0	0	0	0	0	1	-360	360;
	4654	4090	0.00895	0.07061	0	0	0	0	0	0	1	-360	360;
	4654	8218	0.01187	0.05656	0	0	0	0	0	0	1	-360	360;
	2408	6388	0.00922	0.04395	0	0	0	0	0	0	1	-360	360;
	2408	5734	0.00995	0.07845	0	0	0	0	0	0	1	-360	360;
	6266	8648	0.01548	0.04972	0	0	0	0	0	0	1	-360	360;
	6266	8218	0.00339	0.01616	0	0	0	0	0	0	1	-360	360;
	234	8105	0.00851	0.04056	0	0	0	0	0	0	1	-360	360;
	6536	7363	0.0099	0.06074	0	280	0	0	0	0	1	-360	360;
	6388	1154	0.01001	0.07892	0	0	0	0	0	0	1	-360	360;
	6388	5734	0.01647	0.07854	0	0	0	0	0	0	1	-360	360;
	8472	1154	0.00392	0.03091	0	0	0	0	0	0	1	-360	360;
	423	2370	0.00603	0.02877	0	0	0	0	0	0	1	-360	360;
	6370	5757	0.00043	0.00443	0	1351	0	0	0	0	1	-360	360;
	6588	1941	6e-05	0.0008	0	0	0	0	0	0	1	-360	360;
	6588	836	0.00029	0.00366	0	1836	0	0	0	0	1	-360	360;
	6588	644	0.00051	0.00643	0	2044	0	0	0	0	1	-360	360;
	6588	6300	0.00033	0.00424	0	0	0	0	0	0	1	-360	360;
	5757	3628	0.00034	0.00432	0	0	0	0	0	0	1	-360	360;
	5757	8793	0.00013	0.00145	0	1247	0	0	0	0	1	-360	360;
	5757	836	0.00018	0.00233	0	1905	0	0	0	0	1	-360	360;
	3628	8089	0.00034	0.00432	0	1871	0	0	0	0	1	-360	360;
	3628	4455	0.00208	0.02643	0	0	0	0	0	0	1	-360	360;
	3628	7408	0.00209	0.0266	0	0	0	0	0	0	1	-360	360;
	3628	2028	0.00057	0.00725	0	1905	0	0	0	0	1	-360	360;
	7962	2264	0.00034	0.00153	0	0	0	0	0	0	1	-360	360;
	7962	2264	0.00034	0.00153	0	0	0	0	0	0	1	-360	360;
	8793	836	8e-05	0.00089	0	935	0	0	0	0	1	-360	360;
	8793	8390	6e-05	0.00067	0	1039	0	0	0	0	1	-360	360;
	6269	5427	0.00181	0.01428	0	0	0	0	0	0	1	-360	360;
	6269	5427	0.00181	0.01428	0	0	0	0	0	0	1	-360	360;
	6269	3045	0.00576	0.03535	0	0	0	0	0	0	1	-360	360;
	6269	8988	0.00576	0.03535	0	0	0	0	0	0	1	-360	360;
	6270	3386	0.00526	0.02505	0	0	0	0	0	0	1	-360	360;
	6270	3386	0.00526	0.02505	0	0	0	0	0	0	1	-360	360;
	6270	4212	0.00159	0.01612	0	0	0	0	0	0	1	-360	360;
	6270	4212	0.00159	0.01612	0	0	0	0	0	0	1	-360	360;
	6270	2750	0.00121	0.01227	0	0	0	0	0	0	1	-360	360;
	6270	2750	0.00121	0.01227	0	0	0	0	0	0	1	-360	360;
	7084	852	0.00132	0.0063	0	0	0	0	0	0	1	-360	360;
	4983	2001	0.0019	0.00905	0	0	0	0	0	0	1	-360	360;
	4983	2001	0.0019	0.00905	0	0	0	0	0	0	1	-360	360;
	4983	1983	0.00025	0.00264	0	0	0	0	0	0	1	-360	360;
	4983	1983	0.0002	0.00204	0	0	0	0	0	0	1	-360	360;
	4983	852	0.00031	0.00316	0	0	0	0	0	0	1	-360	360;
	4983	4061	0.00031	0.00316	0	0	0	0	0	0	1	-360	360;
	5427	2266	0.00145	0.01145	0	0	0	0	0	0	1	-360	360;
	5427	8928	0.00141	0.01116	0	0	0	0	0	0	1	-360	360;
	6198	7802	0.00036	0.00359	0	0	0	0	0	0	1	-360	360;
	6198	7802	0.00036	0.00359	0	0	0	0	0	0	1	-360	360;
	2266	8928	0.00066	0.00518	0	0	0	0	0	0	1	-360	360;
	120	5840	0.00049	0.00496	0	267	0	0	0	0	1	-360	360;
	120	2373	0.00028	0.00296	0	0	0	0	0	0	1	-360	360;
	120	6678	0.00028	0.00304	0	0	0	0	0	0	1	-360	360;
	2621	3386	0.00654	0.03119	0	0	0	0	0	0	1	-360	360;
	2621	3403	0.00108	0.00517	0	0	0	0	0	0	1	-360	360;
	2621	3403	0.00108	0.00517	0	0	0	0	0	0	1	-360	360;
	5840	872	0.00119	0.00941	0	0	0	0	0	0	1	-360	360;
	5840	872	0.00119	0.00941	0	0	0	0	0	0	1	-360	360;
	3045	3402	0.0061	0.03744	0	0	0	0	0	0	1	-360	360;
	3045	3402	0.0061	0.03744	0	0	0	0	0	0	1	-360	360;
	3045	8988	0.00052	0.00321	0	267	0	0	0	0	1	-360	360;
	3402	8928	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	3402	8928	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	3402	6731	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	3402	6731	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	2375	8122	0.00026	0.00267	0	0	0	0	0	0	1	-360	360;
	2375	8563	0.00021	0.00224	0	0	0	0	0	0	1	-360	360;
	3403	6545	0.00173	0.01365	0	0	0	0	0	0	1	-360	360;
	3403	6545	0.00173	0.01365	0	0	0	0	0	0	1	-360	360;
	2373	872	0.00037	0.00401	0	0	0	0	0	0	1	-360	360;
	2373	872	0.00037	0.00401	0	0	0	0	0	0	1	-360	360;
	872	4769	0.00073	0.00744	0	0	0	0	0	0	1	-360	360;
	872	4769	0.00073	0.00744	0	0	0	0	0	0	1	-360	360;
	6277	2264	0.00031	0.00329	0	0	0	0	0	0	1	-360	360;
	6277	2264	0.00025	0.00254	0	0	0	0	0	0	1	-360	360;
	6277	1995	0.00031	0.0031	0	0	0	0	0	0	1	-360	360;
	6277	1995	0.00031	0.0031	0	0	0	0	0	0	1	-360	360;
	6126	7288	0.00034	0.00108	0	0	0	0	0	0	1	-360	360;
	9129	7288	0.00042	0.00449	0	307	0	0	0	0	1	-360	360;
	9129	8563	0.00025	0.00248	0	0	0	0	0	0	1	-360	360;
	5282	7281	0.00119	0.01202	0	0	0	0	0	0	1	-360	360;
	5282	7281	0.00119	0.01202	0	0	0	0	0	0	1	-360	360;
	5282	7802	0.00034	0.00341	0	0	0	0	0	0	1	-360	360;
	5282	7802	0.00034	0.00341	0	0	0	0	0	0	1	-360	360;
	5122	5345	0.00034	0.00369	0	307	0	0	0	0	1	-360	360;
	8122	8563	0.00053	0.00569	0	0	0	0	0	0	1	-360	360;
	8122	4061	0.00336	0.016	0	0	0	0	0	0	1	-360	360;
	7281	2298	0.00827	0.03943	0	0	0	0	0	0	1	-360	360;
	7281	2298	0.00827	0.03943	0	0	0	0	0	0	1	-360	360;
	4212	7277	0.00073	0.00785	0	293	0	0	0	0	1	-360	360;
	4212	2750	0.00117	0.00926	0	0	0	0	0	0	1	-360	360;
	7277	5200	0.00037	0.00401	0	293	0	0	0	0	1	-360	360;
	1941	8390	0.00012	0.00134	0	1039	0	0	0	0	1	-360	360;
	8089	1827	0.00023	0.00298	0	2009	0	0	0	0	1	-360	360;
	1594	1420	0.00087	0.01433	0	1316	0	0	0	0	1	-360	360;
	1594	1420	0.00088	0.01455	0	901	0	0	0	0	1	-360	360;
	1594	8680	0.00133	0.01364	0	1351	0	0	0	0	1	-360	360;
	1594	4917	0.00064	0.00644	0	1524	0	0	0	0	1	-360	360;
	1594	7408	0.00025	0.00324	0	0	0	0	0	0	1	-360	360;
	1594	8714	0.00276	0.02823	0	1282	0	0	0	0	1	-360	360;
	1594	1738	0.00266	0.02691	0	1490	0	0	0	0	1	-360	360;
	1594	366	0.00196	0.01981	0	1386	0	0	0	0	1	-360	360;
	1594	4305	0.0023	0.03797	0	935	0	0	0	0	1	-360	360;
	1420	1827	0.00027	0.00344	0	0	0	0	0	0	1	-360	360;
	1420	644	0.00032	0.00526	0	1074	0	0	0	0	1	-360	360;
	1420	4917	0.001	0.0101	0	1559	0	0	0	0	1	-360	360;
	1420	4888	0.00017	0.00279	0	1351	0	0	0	0	1	-360	360;
	8680	4455	0.00097	0.0123	0	1905	0	0	0	0	1	-360	360;
	8680	4920	0.00228	0.02898	0	2009	0	0	0	0	1	-360	360;
	8680	1152	0.00269	0.03425	0	2044	0	0	0	0	1	-360	360;
	8680	7378	0.00096	0.00982	0	1247	0	0	0	0	1	-360	360;
	1827	7408	0.00105	0.01739	0	1351	0	0	0	0	1	-360	360;
	1827	6300	0.00028	0.00351	0	0	0	0	0	0	1	-360	360;
	644	4888	0.00017	0.00279	0	866	0	0	0	0	1	-360	360;
	3591	1291	0.00278	0.01325	0	0	0	0	0	0	1	-360	360;
	3591	2287	0.04068	0.19392	0	0	0	0	0	0	1	-360	360;
	1291	6674	0.01464	0.04701	0	0	0	0	0	0	1	-360	360;
	1291	2618	0.01526	0.07272	0	0	0	0	0	0	1	-360	360;
	1291	6123	0.02017	0.09615	0	0	0	0	0	0	1	-360	360;
	1291	6903	0.01254	0.05979	0	0	0	0	0	0	1	-360	360;
	6386	8366	0.0046	0.03624	0	0	0	0	0	0	1	-360	360;
	6386	2007	0.01993	0.064	0	0	0	0	0	0	1	-360	360;
	467	1223	0.00974	0.03128	0	0	0	0	0	0	1	-360	360;
	467	1659	0.01464	0.04701	0	0	0	0	0	0	1	-360	360;
	467	2287	0.04605	0.14789	0	0	0	0	0	0	1	-360	360;
	1223	4078	0.00287	0.00922	0	0	0	0	0	0	1	-360	360;
	3135	4066	0.00308	0.0315	0	0	0	0	0	0	1	-360	360;
	3135	2336	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	2693	1799	0.0033	0.02605	0	0	0	0	0	0	1	-360	360;
	2693	4777	0.00239	0.02444	0	0	0	0	0	0	1	-360	360;
	2693	8116	0.0003	0.00312	0	333	0	0	0	0	1	-360	360;
	2693	4303	0.00318	0.0251	0	0	0	0	0	0	1	-360	360;
	897	4292	0.00043	0.00434	0	0	0	0	0	0	1	-360	360;
	897	371	0.00259	0.0204	0	0	0	0	0	0	1	-360	360;
	1799	8078	0.00305	0.01454	0	0	0	0	0	0	1	-360	360;
	1799	8078	0.00305	0.01454	0	0	0	0	0	0	1	-360	360;
	1799	2336	0.00119	0.00941	0	307	0	0	0	0	1	-360	360;
	6674	8366	0.02353	0.07557	0	0	0	0	0	0	1	-360	360;
	9188	9099	0.00306	0.02416	0	0	0	0	0	0	1	-360	360;
	9188	777	0.00398	0.03138	0	0	0	0	0	0	1	-360	360;
	1312	8366	0.01471	0.07013	0	0	0	0	0	0	1	-360	360;
	258	5129	0.00325	0.03326	0	0	0	0	0	0	1	-360	360;
	258	5230	0.0006	0.00607	0	307	0	0	0	0	1	-360	360;
	258	4777	0.00146	0.01493	0	0	0	0	0	0	1	-360	360;
	258	5785	0.00061	0.0062	0	0	0	0	0	0	1	-360	360;
	258	374	0.00075	0.00356	0	0	0	0	0	0	1	-360	360;
	6111	5291	0.00921	0.0305	0	0	0	0	0	0	1	-360	360;
	1659	8366	0.01233	0.0396	0	0	0	0	0	0	1	-360	360;
	8597	3783	0.0117	0.05575	0	0	0	0	0	0	1	-360	360;
	8597	3783	0.0117	0.05575	0	0	0	0	0	0	1	-360	360;
	8597	1244	0.00169	0.00808	0	0	0	0	0	0	1	-360	360;
	8597	2818	0.00235	0.01851	0	0	0	0	0	0	1	-360	360;
	8597	2818	0.00235	0.01851	0	0	0	0	0	0	1	-360	360;
	8597	777	0.00119	0.00941	0	0	0	0	0	0	1	-360	360;
	8597	777	0.00119	0.00941	0	0	0	0	0	0	1	-360	360;
	7888	4007	0.00046	0.00497	0	293	0	0	0	0	1	-360	360;
	4066	5129	0.00024	0.00188	0	0	0	0	0	0	1	-360	360;
	5128	2618	0.01288	0.06141	0	0	0	0	0	0	1	-360	360;
	5128	2287	0.02105	0.10035	0	0	0	0	0	0	1	-360	360;
	4221	8116	0.001	0.00785	0	0	0	0	0	0	1	-360	360;
	4221	4303	0.008	0.06307	0	0	0	0	0	0	1	-360	360;
	1169	319	0.00046	0.00497	0	0	0	0	0	0	1	-360	360;
	1169	7263	0.00034	0.0036	0	0	0	0	0	0	1	-360	360;
	4630	4303	0.01068	0.0509	0	0	0	0	0	0	1	-360	360;
	4630	3140	0.00475	0.02262	0	0	0	0	0	0	1	-360	360;
	3143	374	0.00022	0.0024	0	0	0	0	0	0	1	-360	360;
	3143	1230	0.00054	0.00259	0	0	0	0	0	0	1	-360	360;
	319	4086	0.00041	0.00441	0	0	0	0	0	0	1	-360	360;
	319	4086	0.00041	0.00441	0	0	0	0	0	0	1	-360	360;
	319	7263	0.00031	0.0031	0	0	0	0	0	0	1	-360	360;
	319	3168	0.00259	0.0204	0	0	0	0	0	0	1	-360	360;
	9103	43	0.0044	0.03467	0	0	0	0	0	0	1	-360	360;
	9103	3783	0.01353	0.10669	0	0	0	0	0	0	1	-360	360;
	6646	3783	0.01736	0.08274	0	0	0	0	0	0	1	-360	360;
	6646	3783	0.01736	0.08274	0	0	0	0	0	0	1	-360	360;
	4007	7263	0.00034	0.0036	0	293	0	0	0	0	1	-360	360;
	5230	5113	0.00035	0.00353	0	280	0	0	0	0	1	-360	360;
	5230	5113	0.00035	0.00353	0	307	0	0	0	0	1	-360	360;
	9099	777	0.00131	0.01036	0	0	0	0	0	0	1	-360	360;
	4455	7408	1.5e-05	0.000199	0	1905	0	0	0	0	1	-360	360;
	43	3783	0.00637	0.05021	0	0	0	0	0	0	1	-360	360;
	3783	5291	0.0105	0.03466	0	0	0	0	0	0	1	-360	360;
	4303	3140	0.01576	0.07514	0	0	0	0	0	0	1	-360	360;
	4303	371	0.00112	0.01202	0	0	0	0	0	0	1	-360	360;
	3939	2818	0.00028	0.0009	0	0	0	0	0	0	1	-360	360;
	4078	8518	0.0165	0.05297	0	0	0	0	0	0	1	-360	360;
	4917	8714	0.00233	0.02385	0	1386	0	0	0	0	1	-360	360;
	2818	5899	0.00547	0.04315	0	0	0	0	0	0	1	-360	360;
	2818	5899	0.00547	0.04315	0	0	0	0	0	0	1	-360	360;
	7408	4305	0.00219	0.03613	0	970	0	0	0	0	1	-360	360;
	8657	8714	0.00032	0.00326	0	1212	0	0	0	0	1	-360	360;
	8657	1881	0.00215	0.022	0	1247	0	0	0	0	1	-360	360;
	8657	7429	0.00173	0.01767	0	1247	0	0	0	0	1	-360	360;
	5434	8654	0.00318	0.0251	0	0	0	0	0	0	1	-360	360;
	5434	4173	0.00497	0.03923	0	0	0	0	0	0	1	-360	360;
	5434	4420	0.00199	0.01569	0	0	0	0	0	0	1	-360	360;
	4927	543	0.00147	0.01161	0	0	0	0	0	0	1	-360	360;
	4927	7032	0.00511	0.04032	0	0	0	0	0	0	1	-360	360;
	9052	8440	0.00103	0.00816	0	0	0	0	0	0	1	-360	360;
	9052	1134	0.00139	0.01098	0	0	0	0	0	0	1	-360	360;
	9052	1444	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	8440	1444	0.00199	0.01569	0	0	0	0	0	0	1	-360	360;
	4628	5134	0.01132	0.05397	0	0	0	0	0	0	1	-360	360;
	4628	6175	0.00358	0.02824	0	0	0	0	0	0	1	-360	360;
	4628	2741	0.00085	0.00404	0	0	0	0	0	0	1	-360	360;
	4628	8654	0.00358	0.02824	0	0	0	0	0	0	1	-360	360;
	4628	8654	0.00358	0.02824	0	0	0	0	0	0	1	-360	360;
	7558	7964	0.00086	0.0053	0	267	0	0	0	0	1	-360	360;
	7558	7964	0.00086	0.0053	0	267	0	0	0	0	1	-360	360;
	7558	7032	0.00219	0.01726	0	0	0	0	0	0	1	-360	360;
	7558	3345	0.00505	0.02408	0	267	0	0	0	0	1	-360	360;
	543	3989	0.00402	0.03169	0	0	0	0	0	0	1	-360	360;
	543	3755	0.00468	0.03687	0	0	0	0	0	0	1	-360	360;
	543	3755	0.00468	0.03687	0	0	0	0	0	0	1	-360	360;
	543	7032	0.00643	0.05068	0	0	0	0	0	0	1	-360	360;
	543	6123	0.04025	0.12927	0	0	0	0	0	0	1	-360	360;
	3989	7032	0.00302	0.02385	0	0	0	0	0	0	1	-360	360;
	109	2746	0.01417	0.06755	0	0	0	0	0	0	1	-360	360;
	109	4043	0.01503	0.04827	0	0	0	0	0	0	1	-360	360;
	109	5134	0.02556	0.12185	0	0	0	0	0	0	1	-360	360;
	5899	3755	0.02618	0.08407	0	0	0	0	0	0	1	-360	360;
	2746	5134	0.01105	0.05268	0	0	0	0	0	0	1	-360	360;
	4043	6640	0.00788	0.06213	0	0	0	0	0	0	1	-360	360;
	4043	6175	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	4043	1134	0.00199	0.01569	0	0	0	0	0	0	1	-360	360;
	4573	5134	0.02098	0.10003	0	0	0	0	0	0	1	-360	360;
	4573	2281	0.00793	0.03781	0	0	0	0	0	0	1	-360	360;
	5134	2741	0.01136	0.05414	0	0	0	0	0	0	1	-360	360;
	5134	2281	0.02597	0.12379	0	0	0	0	0	0	1	-360	360;
	6640	6123	0.02004	0.06436	0	0	0	0	0	0	1	-360	360;
	8714	5846	0.00229	0.02909	0	1801	0	0	0	0	1	-360	360;
	6175	1134	0.00338	0.02667	0	0	0	0	0	0	1	-360	360;
	6175	1134	0.00338	0.02667	0	0	0	0	0	0	1	-360	360;
	8654	4420	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	4173	2281	0.01407	0.0452	0	0	0	0	0	0	1	-360	360;
	1881	2775	0.00128	0.01308	0	1351	0	0	0	0	1	-360	360;
	3459	1558	0.00438	0.03452	0	0	0	0	0	0	1	-360	360;
	3459	9182	0.01992	0.15706	0	0	0	0	0	0	1	-360	360;
	3459	2281	0.03389	0.10884	0	0	0	0	0	0	1	-360	360;
	3459	7188	0.01313	0.10355	0	0	0	0	0	0	1	-360	360;
	3459	1975	0.00557	0.04393	0	0	0	0	0	0	1	-360	360;
	1558	3336	0.01008	0.03236	0	0	0	0	0	0	1	-360	360;
	1558	2833	0.01493	0.11768	0	0	0	0	0	0	1	-360	360;
	3336	6890	0.03074	0.09872	0	0	0	0	0	0	1	-360	360;
	3336	6890	0.04301	0.13813	0	0	0	0	0	0	1	-360	360;
	4633	6941	0.01407	0.0452	0	0	0	0	0	0	1	-360	360;
	4633	2281	0.02883	0.09257	0	0	0	0	0	0	1	-360	360;
	4633	7188	0.00796	0.06276	0	0	0	0	0	0	1	-360	360;
	4633	6850	0.0045	0.01446	0	0	0	0	0	0	1	-360	360;
	6941	1772	0.01712	0.05496	0	0	0	0	0	0	1	-360	360;
	2087	9182	0.02573	0.08263	0	0	0	0	0	0	1	-360	360;
	2087	1772	0.05152	0.16543	0	0	0	0	0	0	1	-360	360;
	9182	3590	0.01871	0.11474	0	0	0	0	0	0	1	-360	360;
	9182	2533	0.05489	0.17628	0	0	0	0	0	0	1	-360	360;
	9182	7188	0.01186	0.09351	0	0	0	0	0	0	1	-360	360;
	9182	3979	0.00169	0.01334	0	0	0	0	0	0	1	-360	360;
	9182	3979	0.00169	0.01334	0	0	0	0	0	0	1	-360	360;
	3590	185	0.00811	0.02603	0	0	0	0	0	0	1	-360	360;
	3590	7337	0.00034	0.00162	0	0	0	0	0	0	1	-360	360;
	3590	7337	0.00034	0.00162	0	0	0	0	0	0	1	-360	360;
	2533	185	0.00736	0.05805	0	0	0	0	0	0	1	-360	360;
	9061	2833	0.01094	0.0863	0	0	0	0	0	0	1	-360	360;
	9061	185	0.01214	0.09571	0	0	0	0	0	0	1	-360	360;
	2833	3501	0.00895	0.07061	0	0	0	0	0	0	1	-360	360;
	2833	2104	0.03063	0.09836	0	0	0	0	0	0	1	-360	360;
	5846	2775	0.00242	0.03076	0	1836	0	0	0	0	1	-360	360;
	5293	7429	0.00083	0.00845	0	1039	0	0	0	0	1	-360	360;
	5293	4567	0.0056	0.05722	0	1143	0	0	0	0	1	-360	360;
	5293	5934	0.0036	0.03646	0	797	0	0	0	0	1	-360	360;
	5293	7658	0.001	0.0101	0	1282	0	0	0	0	1	-360	360;
	2793	5033	0.02308	0.07413	0	0	0	0	0	0	1	-360	360;
	2793	7926	0.01683	0.05406	0	0	0	0	0	0	1	-360	360;
	6728	5826	0.03322	0.10667	0	0	0	0	0	0	1	-360	360;
	6728	5040	0.01094	0.0863	0	0	0	0	0	0	1	-360	360;
	1975	5817	0.01546	0.07369	0	0	0	0	0	0	1	-360	360;
	9082	6123	0.00526	0.02505	0	0	0	0	0	0	1	-360	360;
	9082	4629	0.02898	0.13817	0	0	0	0	0	0	1	-360	360;
	8341	4955	0.01773	0.08452	0	267	0	0	0	0	1	-360	360;
	8341	4891	0.01368	0.04393	0	0	0	0	0	0	1	-360	360;
	8341	1080	0.02365	0.07594	0	0	0	0	0	0	1	-360	360;
	8341	7238	0.00078	0.00372	0	0	0	0	0	0	1	-360	360;
	8341	7238	0.00078	0.00372	0	0	0	0	0	0	1	-360	360;
	5033	4042	0.02759	0.08859	0	0	0	0	0	0	1	-360	360;
	4627	8309	0.01576	0.05062	0	0	0	0	0	0	1	-360	360;
	5817	9220	0.02664	0.12702	0	0	0	0	0	0	1	-360	360;
	8584	9220	0.02162	0.06943	0	0	0	0	0	0	1	-360	360;
	8584	4768	0.03187	0.10233	0	0	0	0	0	0	1	-360	360;
	5431	7926	0.02168	0.06961	0	0	0	0	0	0	1	-360	360;
	7926	4629	0.03827	0.18245	0	0	0	0	0	0	1	-360	360;
	7926	2289	0.01465	0.06981	0	0	0	0	0	0	1	-360	360;
	7926	1080	0.00661	0.05209	0	0	0	0	0	0	1	-360	360;
	7926	1080	0.00782	0.06166	0	0	0	0	0	0	1	-360	360;
	4042	4955	0.0366	0.11752	0	0	0	0	0	0	1	-360	360;
	5826	5932	0.00326	0.01049	0	0	0	0	0	0	1	-360	360;
	5826	5932	0.00326	0.01049	0	0	0	0	0	0	1	-360	360;
	5826	5932	0.00326	0.01049	0	0	0	0	0	0	1	-360	360;
	5826	4955	0.00732	0.03491	0	0	0	0	0	0	1	-360	360;
	5826	4955	0.00732	0.03491	0	0	0	0	0	0	1	-360	360;
	5826	9220	0.03063	0.09836	0	0	0	0	0	0	1	-360	360;
	5826	9220	0.03063	0.09836	0	0	0	0	0	0	1	-360	360;
	6123	4891	0.01768	0.05677	0	0	0	0	0	0	1	-360	360;
	6123	4629	0.03438	0.16386	0	0	0	0	0	0	1	-360	360;
	6123	6903	0.01017	0.04848	0	0	0	0	0	0	1	-360	360;
	7113	4629	0	0.04891	0	0	0	0	0	0	1	-360	360;
	7113	9220	0.00081	0.00388	0	0	0	0	0	0	1	-360	360;
	7113	9220	0.00081	0.00388	0	0	0	0	0	0	1	-360	360;
	4629	1080	0.03824	0.18228	0	0	0	0	0	0	1	-360	360;
	4768	8309	0.02207	0.07087	0	0	0	0	0	0	1	-360	360;
	4768	4999	0.0165	0.05297	0	0	0	0	0	0	1	-360	360;
	6021	4567	0.00094	0.00958	0	1282	0	0	0	0	1	-360	360;
	6021	2659	0.0009	0.01147	0	1905	0	0	0	0	1	-360	360;
	6021	8317	2e-05	0.00026	0	1282	0	0	0	0	1	-360	360;
	6021	8317	2e-05	0.00026	0	1143	0	0	0	0	1	-360	360;
	6021	2775	0.00162	0.01656	0	1247	0	0	0	0	1	-360	360;
	6021	2683	0.00039	0.00641	0	1039	0	0	0	0	1	-360	360;
	4567	2683	0.00027	0.00444	0	901	0	0	0	0	1	-360	360;
	4567	926	0.00288	0.02946	0	1247	0	0	0	0	1	-360	360;
	4567	926	0.00192	0.02444	0	1940	0	0	0	0	1	-360	360;
	2659	2455	0.00094	0.01197	0	1663	0	0	0	0	1	-360	360;
	2659	294	0.00106	0.01087	0	1178	0	0	0	0	1	-360	360;
	1948	5367	0.00677	0.05335	0	0	0	0	0	0	1	-360	360;
	1948	5367	0.00677	0.05335	0	0	0	0	0	0	1	-360	360;
	787	487	0.00051	0.00242	0	0	0	0	0	0	1	-360	360;
	787	487	0.00051	0.00242	0	0	0	0	0	0	1	-360	360;
	787	2711	0.00026	0.00204	0	0	0	0	0	0	1	-360	360;
	787	2711	0.00026	0.00204	0	0	0	0	0	0	1	-360	360;
	3002	487	0.00139	0.00663	0	0	0	0	0	0	1	-360	360;
	3002	487	0.00139	0.00663	0	0	0	0	0	0	1	-360	360;
	3002	2711	0.00075	0.00356	0	0	0	0	0	0	1	-360	360;
	3002	2711	0.00075	0.00356	0	0	0	0	0	0	1	-360	360;
	4193	7654	0.00321	0.03286	0	0	0	0	0	0	1	-360	360;
	4193	7654	0.00321	0.03286	0	0	0	0	0	0	1	-360	360;
	7794	5028	0.01875	0.06021	0	0	0	0	0	0	1	-360	360;
	7794	2791	0.02489	0.07991	0	0	0	0	0	0	1	-360	360;
	9207	2735	0.00115	0.0091	0	0	0	0	0	0	1	-360	360;
	9207	1203	0.00037	0.00164	0	0	0	0	0	0	1	-360	360;
	9207	1958	0.00082	0.00881	0	0	0	0	0	0	1	-360	360;
	9207	334	0.001	0.00785	0	0	0	0	0	0	1	-360	360;
	9207	7060	0.00043	0.00465	0	0	0	0	0	0	1	-360	360;
	5056	4980	0.00045	0.00203	0	0	0	0	0	0	1	-360	360;
	5056	3226	0.00028	0.00304	0	0	0	0	0	0	1	-360	360;
	1845	6073	0.00037	0.00164	0	0	0	0	0	0	1	-360	360;
	1845	9031	0.00033	0.00148	0	0	0	0	0	0	1	-360	360;
	1845	5118	0.0002	0.00216	0	0	0	0	0	0	1	-360	360;
	1845	2999	0.00037	0.00372	0	0	0	0	0	0	1	-360	360;
	1694	5118	0.00065	0.00697	0	0	0	0	0	0	1	-360	360;
	1694	5415	0.00015	0.0016	0	0	0	0	0	0	1	-360	360;
	2353	7654	0.002	0.00953	0	0	0	0	0	0	1	-360	360;
	2353	7654	0.002	0.00953	0	0	0	0	0	0	1	-360	360;
	3652	9031	0.0006	0.00268	0	0	0	0	0	0	1	-360	360;
	3652	8106	0.00239	0.01883	0	0	0	0	0	0	1	-360	360;
	3652	2999	0.00117	0.00926	0	0	0	0	0	0	1	-360	360;
	2711	7983	0.00026	0.00161	0	0	0	0	0	0	1	-360	360;
	2711	7983	0.00026	0.00161	0	0	0	0	0	0	1	-360	360;
	4319	3058	0.00061	0.00291	0	0	0	0	0	0	1	-360	360;
	4319	3058	0.00061	0.00291	0	0	0	0	0	0	1	-360	360;
	7689	2735	0.00062	0.00632	0	0	0	0	0	0	1	-360	360;
	7689	2735	0.00062	0.00632	0	0	0	0	0	0	1	-360	360;
	7689	334	0.001	0.00785	0	0	0	0	0	0	1	-360	360;
	8106	2999	0.00231	0.0182	0	0	0	0	0	0	1	-360	360;
	4980	7654	0.00233	0.01836	0	0	0	0	0	0	1	-360	360;
	4980	9085	0.0043	0.03389	0	0	0	0	0	0	1	-360	360;
	5118	2999	0.00017	0.00174	0	0	0	0	0	0	1	-360	360;
	5118	2999	0.00017	0.00174	0	0	0	0	0	0	1	-360	360;
	5118	7677	0.0005	0.00537	0	0	0	0	0	0	1	-360	360;
	3156	1958	0.006	0.0286	0	0	0	0	0	0	1	-360	360;
	3156	409	0.01187	0.05656	0	0	0	0	0	0	1	-360	360;
	3156	3253	0.00022	0.0024	0	267	0	0	0	0	1	-360	360;
	6170	5028	0.01419	0.04556	0	0	0	0	0	0	1	-360	360;
	6170	2808	0.01712	0.08161	0	0	0	0	0	0	1	-360	360;
	8203	3058	0.00173	0.00824	0	0	0	0	0	0	1	-360	360;
	8203	3058	0.00173	0.00824	0	0	0	0	0	0	1	-360	360;
	3226	7654	0.00199	0.01569	0	0	0	0	0	0	1	-360	360;
	3226	9085	0.00716	0.05648	0	0	0	0	0	0	1	-360	360;
	7814	5367	0.00915	0.07217	0	0	0	0	0	0	1	-360	360;
	7814	2999	0.00793	0.03781	0	0	0	0	0	0	1	-360	360;
	5028	2791	0.00796	0.06276	0	0	0	0	0	0	1	-360	360;
	5367	2999	0.01796	0.05768	0	0	0	0	0	0	1	-360	360;
	5367	9224	0.00013	0.00136	0	0	0	0	0	0	1	-360	360;
	9071	3058	0.00179	0.01412	0	0	0	0	0	0	1	-360	360;
	9071	5073	0.00338	0.01085	0	0	0	0	0	0	1	-360	360;
	1958	9085	0.0009	0.00706	0	0	0	0	0	0	1	-360	360;
	1958	9085	0.0009	0.00706	0	0	0	0	0	0	1	-360	360;
	8576	2514	0.02308	0.07413	0	0	0	0	0	0	1	-360	360;
	8576	4429	0.00669	0.05272	0	0	0	0	0	0	1	-360	360;
	6143	4609	0.02866	0.09203	0	0	0	0	0	0	1	-360	360;
	6143	2666	0.06565	0.21081	0	0	0	0	0	0	1	-360	360;
	6143	1097	0.03479	0.11173	0	0	0	0	0	0	1	-360	360;
	3501	2104	0.03243	0.10414	0	0	0	0	0	0	1	-360	360;
	2104	6890	0.00784	0.06182	0	0	0	0	0	0	1	-360	360;
	2104	5073	0.02196	0.07051	0	0	0	0	0	0	1	-360	360;
	4609	2215	0.01323	0.04249	0	0	0	0	0	0	1	-360	360;
	7754	409	0.00336	0.016	0	0	0	0	0	0	1	-360	360;
	7754	409	0.00325	0.01551	0	0	0	0	0	0	1	-360	360;
	7754	7447	0.02303	0.07395	0	0	0	0	0	0	1	-360	360;
	7754	2514	0.00845	0.02712	0	0	0	0	0	0	1	-360	360;
	409	7060	0.00678	0.03232	0	0	0	0	0	0	1	-360	360;
	2215	7447	0.0322	0.10342	0	0	0	0	0	0	1	-360	360;
	6890	4429	0.0002	0.00157	0	0	0	0	0	0	1	-360	360;
	4999	5615	0.05906	0.18966	0	0	0	0	0	0	1	-360	360;
	4999	6525	0.00676	0.0217	0	0	0	0	0	0	1	-360	360;
	2666	5040	0.02488	0.19613	0	0	0	0	0	0	1	-360	360;
	2666	6525	0.00676	0.0217	0	0	0	0	0	0	1	-360	360;
	7227	926	0.00094	0.00956	0	1316	0	0	0	0	1	-360	360;
	7227	7728	0.00025	0.00258	0	1351	0	0	0	0	1	-360	360;
	926	7728	0.00082	0.00843	0	1247	0	0	0	0	1	-360	360;
	926	7498	0.00156	0.01988	0	1871	0	0	0	0	1	-360	360;
	7728	2455	0.00146	0.01855	0	1975	0	0	0	0	1	-360	360;
	4308	4972	0.02449	0.07865	0	0	0	0	0	0	1	-360	360;
	4308	5411	0.01055	0.08316	0	0	0	0	0	0	1	-360	360;
	4106	8824	0.02753	0.08841	0	0	0	0	0	0	1	-360	360;
	4106	1442	0.02393	0.07684	0	0	0	0	0	0	1	-360	360;
	4524	325	0.01566	0.07466	0	0	0	0	0	0	1	-360	360;
	2808	2791	0.00519	0.04095	0	0	0	0	0	0	1	-360	360;
	2808	4266	0.02286	0.0734	0	0	0	0	0	0	1	-360	360;
	2808	1982	0.02226	0.08078	0	0	0	0	0	0	1	-360	360;
	5613	325	0.01817	0.08662	0	0	0	0	0	0	1	-360	360;
	5613	1800	0.01576	0.07514	0	0	0	0	0	0	1	-360	360;
	8008	8824	0.02697	0.0866	0	0	0	0	0	0	1	-360	360;
	8008	1213	0.00929	0.02983	0	0	0	0	0	0	1	-360	360;
	8008	4972	0.03508	0.11264	0	0	0	0	0	0	1	-360	360;
	8008	6945	0.01405	0.11077	0	0	0	0	0	0	1	-360	360;
	8824	6979	0.03479	0.11173	0	0	0	0	0	0	1	-360	360;
	3096	1474	0.01937	0.06219	0	0	0	0	0	0	1	-360	360;
	3096	1348	0.03778	0.12132	0	0	0	0	0	0	1	-360	360;
	3096	1440	0.00497	0.03923	0	0	0	0	0	0	1	-360	360;
	910	1042	0.01864	0.08888	0	0	0	0	0	0	1	-360	360;
	910	1042	0.01864	0.08888	0	0	0	0	0	0	1	-360	360;
	1213	1442	0.01774	0.05695	0	0	0	0	0	0	1	-360	360;
	1213	5411	0.00292	0.02306	0	0	0	0	0	0	1	-360	360;
	1442	1691	0.004	0.01284	0	0	0	0	0	0	1	-360	360;
	1442	1529	0.00985	0.03164	0	0	0	0	0	0	1	-360	360;
	1442	5411	0.00336	0.02652	0	0	0	0	0	0	1	-360	360;
	1442	5411	0.00336	0.02652	0	0	0	0	0	0	1	-360	360;
	1691	3694	0.02179	0.07184	0	0	0	0	0	0	1	-360	360;
	1474	1440	0.00697	0.05491	0	0	0	0	0	0	1	-360	360;
	1474	6920	0.01119	0.05333	0	0	0	0	0	0	1	-360	360;
	1474	6920	0.01119	0.05333	0	0	0	0	0	0	1	-360	360;
	1474	41	0.04645	0.14916	0	0	0	0	0	0	1	-360	360;
	2791	755	0.00402	0.04114	0	0	0	0	0	0	1	-360	360;
	2791	5536	0.00279	0.02851	0	0	0	0	0	0	1	-360	360;
	2791	2182	0.02688	0.12815	0	0	0	0	0	0	1	-360	360;
	1097	4266	0.03125	0.10034	0	0	0	0	0	0	1	-360	360;
	1097	1529	0.00619	0.01989	0	0	0	0	0	0	1	-360	360;
	4755	1348	0.02466	0.07919	0	0	0	0	0	0	1	-360	360;
	4755	8254	0.00563	0.01808	0	0	0	0	0	0	1	-360	360;
	5520	1800	0.00693	0.02224	0	0	0	0	0	0	1	-360	360;
	5520	1800	0.00693	0.02224	0	0	0	0	0	0	1	-360	360;
	5222	4972	0.02567	0.08245	0	0	0	0	0	0	1	-360	360;
	5222	6821	0.02944	0.09456	0	0	0	0	0	0	1	-360	360;
	755	5536	0.0025	0.02552	0	0	0	0	0	0	1	-360	360;
	1042	1348	0.00584	0.05974	0	0	0	0	0	0	1	-360	360;
	1042	2182	0.00956	0.09776	0	0	0	0	0	0	1	-360	360;
	1042	5204	0.00563	0.01808	0	0	0	0	0	0	1	-360	360;
	325	2182	0.01044	0.04977	0	0	0	0	0	0	1	-360	360;
	325	1800	0.00495	0.03907	0	0	0	0	0	0	1	-360	360;
	1348	2182	0.00545	0.04299	0	0	0	0	0	0	1	-360	360;
	4266	8254	0.02252	0.07232	0	0	0	0	0	0	1	-360	360;
	2182	1800	0.01603	0.07644	0	0	0	0	0	0	1	-360	360;
	2182	3694	0.03924	0.12602	0	0	0	0	0	0	1	-360	360;
	2182	1982	0.01798	0.06704	0	0	0	0	0	0	1	-360	360;
	7498	8248	0	-0.0207	0	831	0	0	0	0	1	-360	360;
	6181	2436	0.00149	0.01526	0	1282	0	0	0	0	1	-360	360;
	6181	2436	0.00099	0.01264	0	1905	0	0	0	0	1	-360	360;
	6181	1925	0.00168	0.01713	0	1247	0	0	0	0	1	-360	360;
	6181	3292	0.00131	0.01663	0	2009	0	0	0	0	1	-360	360;
	2436	1925	0.00131	0.01663	0	1524	0	0	0	0	1	-360	360;
	2436	8248	0.00222	0.02826	0	1593	0	0	0	0	1	-360	360;
	1336	8135	0.01284	0.04122	0	0	0	0	0	0	1	-360	360;
	1336	5904	0.00563	0.01808	0	0	0	0	0	0	1	-360	360;
	8063	7179	0.01094	0.0863	0	0	0	0	0	0	1	-360	360;
	8063	1292	0.0061	0.02909	0	0	0	0	0	0	1	-360	360;
	8063	605	0.00537	0.04236	0	0	0	0	0	0	1	-360	360;
	8911	1589	0.01255	0.04032	0	0	0	0	0	0	1	-360	360;
	8911	6821	0.01092	0.03508	0	0	0	0	0	0	1	-360	360;
	8911	5105	0.01689	0.05424	0	0	0	0	0	0	1	-360	360;
	8911	8053	0.00011	0.00036	0	0	0	0	0	0	1	-360	360;
	8365	8177	0.02393	0.07684	0	0	0	0	0	0	1	-360	360;
	8365	8177	0.00856	0.06747	0	0	0	0	0	0	1	-360	360;
	8365	8177	0.00856	0.06747	0	0	0	0	0	0	1	-360	360;
	8365	1589	0.00602	0.01935	0	0	0	0	0	0	1	-360	360;
	8365	6821	0.0058	0.02763	0	0	0	0	0	0	1	-360	360;
	8365	6821	0.0058	0.02763	0	0	0	0	0	0	1	-360	360;
	8177	8682	0.02978	0.09564	0	0	0	0	0	0	1	-360	360;
	8177	8682	0.01194	0.09414	0	0	0	0	0	0	1	-360	360;
	8177	8682	0.01194	0.09414	0	0	0	0	0	0	1	-360	360;
	8135	5851	0.00796	0.06276	0	0	0	0	0	0	1	-360	360;
	8135	2623	0.00398	0.03138	0	0	0	0	0	0	1	-360	360;
	8135	8150	0.00458	0.03609	0	0	0	0	0	0	1	-360	360;
	7213	1589	0.00298	0.00958	0	0	0	0	0	0	1	-360	360;
	8682	41	0.03001	0.09637	0	0	0	0	0	0	1	-360	360;
	6950	1292	0.00983	0.04686	0	0	0	0	0	0	1	-360	360;
	6950	1292	0.00796	0.06276	0	0	0	0	0	0	1	-360	360;
	1589	6821	0.00462	0.01483	0	0	0	0	0	0	1	-360	360;
	1589	1208	0.001	0.00785	0	0	0	0	0	0	1	-360	360;
	1589	1208	0.001	0.00785	0	0	0	0	0	0	1	-360	360;
	7179	5851	0.01239	0.03978	0	0	0	0	0	0	1	-360	360;
	7179	5851	0.00749	0.03571	0	0	0	0	0	0	1	-360	360;
	7179	5851	0.00749	0.03571	0	0	0	0	0	0	1	-360	360;
	7179	605	0.00478	0.03766	0	0	0	0	0	0	1	-360	360;
	7179	8150	0.0038	0.02997	0	0	0	0	0	0	1	-360	360;
	7179	347	0.00394	0.01266	0	0	0	0	0	0	1	-360	360;
	5851	1292	0.0199	0.1569	0	0	0	0	0	0	1	-360	360;
	5851	7769	0.00677	0.05335	0	0	0	0	0	0	1	-360	360;
	6821	5904	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	6821	5904	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	5904	2623	0.00458	0.03609	0	0	0	0	0	0	1	-360	360;
	7158	7985	0.00133	0.01349	0	1143	0	0	0	0	1	-360	360;
	7158	6518	0.00047	0.00772	0	1143	0	0	0	0	1	-360	360;
	7158	4097	0.00698	0.07071	0	1351	0	0	0	0	1	-360	360;
	7985	6518	0.00048	0.00615	0	1940	0	0	0	0	1	-360	360;
	7985	5921	0.00048	0.00615	0	1871	0	0	0	0	1	-360	360;
	7985	7026	0.0014	0.01429	0	1109	0	0	0	0	1	-360	360;
	7985	2874	0.00156	0.01982	0	1697	0	0	0	0	1	-360	360;
	3317	6587	0.02032	0.06527	0	0	0	0	0	0	1	-360	360;
	3317	6507	0.04842	0.15549	0	0	0	0	0	0	1	-360	360;
	3317	3982	0.00957	0.03074	0	0	0	0	0	0	1	-360	360;
	3317	5090	0.00729	0.03474	0	0	0	0	0	0	1	-360	360;
	6782	6507	0.0198	0.15612	0	0	0	0	0	0	1	-360	360;
	6782	4278	0.07263	0.23323	0	0	0	0	0	0	1	-360	360;
	6782	3612	0.01194	0.09414	0	0	0	0	0	0	1	-360	360;
	6782	3612	0.01194	0.09414	0	0	0	0	0	0	1	-360	360;
	7173	1088	0.00902	0.04299	0	0	0	0	0	0	1	-360	360;
	7173	1088	0.00902	0.04299	0	0	0	0	0	0	1	-360	360;
	7173	2243	0.00203	0.0097	0	0	0	0	0	0	1	-360	360;
	1088	4514	0.01407	0.0452	0	0	0	0	0	0	1	-360	360;
	1088	6966	0.01149	0.05478	0	0	0	0	0	0	1	-360	360;
	774	1640	0.01396	0.04484	0	0	0	0	0	0	1	-360	360;
	774	1640	0.01396	0.04484	0	0	0	0	0	0	1	-360	360;
	1812	1938	0.00712	0.03394	0	0	0	0	0	0	1	-360	360;
	1812	1938	0.00712	0.03394	0	0	0	0	0	0	1	-360	360;
	1812	1640	0.02421	0.07774	0	0	0	0	0	0	1	-360	360;
	1812	7370	0.02083	0.0669	0	0	0	0	0	0	1	-360	360;
	1938	9151	0.02034	0.09696	0	0	0	0	0	0	1	-360	360;
	1938	302	0.01526	0.07272	0	0	0	0	0	0	1	-360	360;
	1938	6623	0.02949	0.14059	0	0	0	0	0	0	1	-360	360;
	1938	2810	0.01095	0.0522	0	0	0	0	0	0	1	-360	360;
	1938	2810	0.01095	0.0522	0	0	0	0	0	0	1	-360	360;
	1938	7567	0.01763	0.08403	0	0	0	0	0	0	1	-360	360;
	1938	5602	0.00814	0.03878	0	0	0	0	0	0	1	-360	360;
	7350	7379	0.01898	0.0905	0	0	0	0	0	0	1	-360	360;
	7350	7379	0.01898	0.0905	0	0	0	0	0	0	1	-360	360;
	2136	8226	0.00394	0.01266	0	0	0	0	0	0	1	-360	360;
	2136	6623	0.00576	0.02747	0	0	0	0	0	0	1	-360	360;
	2136	8202	0.01436	0.0461	0	0	0	0	0	0	1	-360	360;
	2136	5375	0.02534	0.08136	0	0	0	0	0	0	1	-360	360;
	8226	7614	0.02646	0.08498	0	0	0	0	0	0	1	-360	360;
	8226	6623	0.0059	0.02812	0	0	0	0	0	0	1	-360	360;
	8226	1640	0.02984	0.09582	0	0	0	0	0	0	1	-360	360;
	8226	5722	0.00796	0.06276	0	0	0	0	0	0	1	-360	360;
	8226	5602	0.01729	0.08242	0	0	0	0	0	0	1	-360	360;
	4449	4552	0.00569	0.02715	0	0	0	0	0	0	1	-360	360;
	4449	7623	0.01543	0.07353	0	0	0	0	0	0	1	-360	360;
	4449	7623	0.01543	0.07353	0	0	0	0	0	0	1	-360	360;
	4449	6623	0.00529	0.02521	0	0	0	0	0	0	1	-360	360;
	4449	6623	0.00529	0.02521	0	0	0	0	0	0	1	-360	360;
	4449	7481	0.00024	0.00113	0	0	0	0	0	0	1	-360	360;
	4449	7481	0.00024	0.00113	0	0	0	0	0	0	1	-360	360;
	4449	6473	0.00959	0.04573	0	0	0	0	0	0	1	-360	360;
	4449	8896	0.02252	0.07232	0	0	0	0	0	0	1	-360	360;
	4449	8790	0.01689	0.05424	0	0	0	0	0	0	1	-360	360;
	7140	6507	0.03125	0.10034	0	0	0	0	0	0	1	-360	360;
	7140	3982	0.0078	0.03717	0	0	0	0	0	0	1	-360	360;
	7140	3982	0.0078	0.03717	0	0	0	0	0	0	1	-360	360;
	7140	7223	0.00678	0.03232	0	0	0	0	0	0	1	-360	360;
	7140	7223	0.00678	0.03232	0	0	0	0	0	0	1	-360	360;
	7140	5090	0.01295	0.06173	0	0	0	0	0	0	1	-360	360;
	4109	6966	0.012	0.05721	0	0	0	0	0	0	1	-360	360;
	4109	6966	0.012	0.05721	0	0	0	0	0	0	1	-360	360;
	4109	1640	0.00332	0.01584	0	0	0	0	0	0	1	-360	360;
	4109	1640	0.00332	0.01584	0	0	0	0	0	0	1	-360	360;
	4109	8083	0.00986	0.04703	0	0	0	0	0	0	1	-360	360;
	4109	8083	0.00986	0.04703	0	0	0	0	0	0	1	-360	360;
	2881	8083	0.0002	0.00204	0	0	0	0	0	0	1	-360	360;
	2881	8083	0.0002	0.00204	0	0	0	0	0	0	1	-360	360;
	9151	4552	0.02081	0.09922	0	0	0	0	0	0	1	-360	360;
	9151	4552	0.02081	0.09922	0	0	0	0	0	0	1	-360	360;
	9151	8083	0.01672	0.0537	0	0	0	0	0	0	1	-360	360;
	9151	8202	0.01171	0.03761	0	0	0	0	0	0	1	-360	360;
	9151	7567	0.00407	0.01939	0	0	0	0	0	0	1	-360	360;
	4552	6473	0.00959	0.04573	0	0	0	0	0	0	1	-360	360;
	4552	2450	0.00113	0.00362	0	0	0	0	0	0	1	-360	360;
	4514	8083	0.00296	0.02338	0	0	0	0	0	0	1	-360	360;
	6966	6618	0.01366	0.06513	0	0	0	0	0	0	1	-360	360;
	6966	8083	0.00093	0.0095	0	0	0	0	0	0	1	-360	360;
	6966	1404	0.00034	0.00162	0	0	0	0	0	0	1	-360	360;
	6618	2243	0.01786	0.08516	0	0	0	0	0	0	1	-360	360;
	3426	6587	0.0045	0.01446	0	0	0	0	0	0	1	-360	360;
	3426	2082	0.0116	0.03725	0	0	0	0	0	0	1	-360	360;
	3426	6462	0.02629	0.08443	0	0	0	0	0	0	1	-360	360;
	6587	2082	0.00709	0.02278	0	0	0	0	0	0	1	-360	360;
	302	7614	0.01051	0.0501	0	0	0	0	0	0	1	-360	360;
	7623	7788	0.02353	0.07557	0	0	0	0	0	0	1	-360	360;
	7623	5722	0.00799	0.02567	0	0	0	0	0	0	1	-360	360;
	7623	2082	0.02168	0.06961	0	0	0	0	0	0	1	-360	360;
	7623	6142	0.00608	0.01953	0	0	0	0	0	0	1	-360	360;
	7623	534	0.00169	0.00808	0	0	0	0	0	0	1	-360	360;
	7623	534	0.00169	0.00808	0	0	0	0	0	0	1	-360	360;
	1378	2082	0.15283	0.86416	0	0	0	0	0	0	1	-360	360;
	1378	1935	0.0045	0.01446	0	0	0	0	0	0	1	-360	360;
	1378	6086	0.00336	0.016	0	0	0	0	0	0	1	-360	360;
	7379	2082	0.00125	0.00989	0	0	0	0	0	0	1	-360	360;
	7379	5294	0.00153	0.01208	0	0	0	0	0	0	1	-360	360;
	7379	534	0.01265	0.06028	0	0	0	0	0	0	1	-360	360;
	7379	534	0.01265	0.06028	0	0	0	0	0	0	1	-360	360;
	7379	5090	0.00576	0.02747	0	0	0	0	0	0	1	-360	360;
	7379	5090	0.00576	0.02747	0	0	0	0	0	0	1	-360	360;
	1640	4452	0.00056	0.00181	0	0	0	0	0	0	1	-360	360;
	8083	1349	0.00225	0.00723	0	0	0	0	0	0	1	-360	360;
	8083	3230	0.0006	0.00471	0	0	0	0	0	0	1	-360	360;
	5722	6142	0.01413	0.04538	0	0	0	0	0	0	1	-360	360;
	5722	5375	0.00338	0.01085	0	0	0	0	0	0	1	-360	360;
	2810	5889	0.01065	0.05074	0	0	0	0	0	0	1	-360	360;
	2810	5889	0.01065	0.05074	0	0	0	0	0	0	1	-360	360;
	2810	5998	0.00506	0.03985	0	307	0	0	0	0	1	-360	360;
	2082	6142	0.01368	0.04393	0	0	0	0	0	0	1	-360	360;
	6507	896	0.00034	0.00108	0	0	0	0	0	0	1	-360	360;
	6507	4278	0.05067	0.16272	0	0	0	0	0	0	1	-360	360;
	6462	2490	0.03125	0.10034	0	0	0	0	0	0	1	-360	360;
	5294	1935	0.00113	0.00362	0	0	0	0	0	0	1	-360	360;
	8202	6473	0.01293	0.10199	0	0	0	0	0	0	1	-360	360;
	8202	6473	0.01293	0.10199	0	0	0	0	0	0	1	-360	360;
	8202	2499	0.00106	0.01086	0	0	0	0	0	0	1	-360	360;
	8202	2499	0.00106	0.01086	0	0	0	0	0	0	1	-360	360;
	6142	4198	0.00139	0.01098	0	0	0	0	0	0	1	-360	360;
	6142	4198	0.00139	0.01098	0	0	0	0	0	0	1	-360	360;
	6518	1929	0.0017	0.02161	0	1940	0	0	0	0	1	-360	360;
	2243	9069	0.00895	0.07061	0	0	0	0	0	0	1	-360	360;
	2243	9069	0.00895	0.07061	0	0	0	0	0	0	1	-360	360;
	3982	5090	0.00695	0.03313	0	0	0	0	0	0	1	-360	360;
	3982	5090	0.00746	0.03555	0	0	0	0	0	0	1	-360	360;
	3982	6086	0.00929	0.04428	0	0	0	0	0	0	1	-360	360;
	134	5921	0.00084	0.01064	0	1801	0	0	0	0	1	-360	360;
	134	2874	0.00274	0.03491	0	1697	0	0	0	0	1	-360	360;
	1738	4920	0.00153	0.01547	0	1282	0	0	0	0	1	-360	360;
	366	2151	0.00328	0.03317	0	1351	0	0	0	0	1	-360	360;
	366	7658	0.00137	0.01387	0	1143	0	0	0	0	1	-360	360;
	2151	4920	0.00066	0.00666	0	1005	0	0	0	0	1	-360	360;
	2151	8679	0.0053	0.05369	0	1282	0	0	0	0	1	-360	360;
	2151	5243	0.0053	0.05369	0	1282	0	0	0	0	1	-360	360;
	4920	4305	0.00025	0.00417	0	797	0	0	0	0	1	-360	360;
	4920	2325	0.00037	0.00465	0	1940	0	0	0	0	1	-360	360;
	4920	6050	0.00347	0.04417	0	2009	0	0	0	0	1	-360	360;
	4920	4237	0.00347	0.04417	0	2044	0	0	0	0	1	-360	360;
	4305	7157	0.0017	0.02819	0	866	0	0	0	0	1	-360	360;
	4305	2165	0.0017	0.02819	0	866	0	0	0	0	1	-360	360;
	8434	2065	0.00049	0.00496	0	0	0	0	0	0	1	-360	360;
	8434	3269	0.00139	0.01098	0	0	0	0	0	0	1	-360	360;
	8434	2521	0.00031	0.00329	0	0	0	0	0	0	1	-360	360;
	8434	2137	0.00424	0.0202	0	0	0	0	0	0	1	-360	360;
	8434	2137	0.00424	0.0202	0	0	0	0	0	0	1	-360	360;
	8434	3564	0.01197	0.05705	0	293	0	0	0	0	1	-360	360;
	8434	5321	0.00052	0.00561	0	0	0	0	0	0	1	-360	360;
	3872	3269	0.00199	0.01569	0	0	0	0	0	0	1	-360	360;
	3872	3269	0.00139	0.01098	0	0	0	0	0	0	1	-360	360;
	3872	6044	0.00255	0.02008	0	0	0	0	0	0	1	-360	360;
	3872	6044	0.00777	0.02495	0	0	0	0	0	0	1	-360	360;
	3872	6044	0.00255	0.02008	0	0	0	0	0	0	1	-360	360;
	3872	6044	0.00777	0.02495	0	0	0	0	0	0	1	-360	360;
	2065	3269	0.00475	0.02262	0	0	0	0	0	0	1	-360	360;
	2065	5708	0.00031	0.00329	0	0	0	0	0	0	1	-360	360;
	2065	3564	0.0101	0.04816	0	307	0	0	0	0	1	-360	360;
	2768	2137	0.00929	0.04428	0	0	0	0	0	0	1	-360	360;
	2768	9102	0.01119	0.05333	0	0	0	0	0	0	1	-360	360;
	4255	4626	0.00468	0.03687	0	0	0	0	0	0	1	-360	360;
	4255	8695	0.00225	0.00723	0	0	0	0	0	0	1	-360	360;
	4255	8766	0.05517	0.17718	0	0	0	0	0	0	1	-360	360;
	8817	9	0.01616	0.05189	0	0	0	0	0	0	1	-360	360;
	8817	3564	0.03603	0.11571	0	0	0	0	0	0	1	-360	360;
	8909	2483	0.00263	0.02071	0	0	0	0	0	0	1	-360	360;
	8909	3164	0.00508	0.02424	0	0	0	0	0	0	1	-360	360;
	9	2287	0.00545	0.04299	0	0	0	0	0	0	1	-360	360;
	8995	2483	0.01505	0.07175	0	0	0	0	0	0	1	-360	360;
	8995	2483	0.01505	0.07175	0	0	0	0	0	0	1	-360	360;
	7645	1235	0.01431	0.0682	0	0	0	0	0	0	1	-360	360;
	7645	2324	0.00267	0.02103	0	0	0	0	0	0	1	-360	360;
	7645	2324	0.00267	0.02103	0	0	0	0	0	0	1	-360	360;
	2325	5478	0.00346	0.04406	0	0	0	0	0	0	1	-360	360;
	2483	3164	0.00332	0.0262	0	0	0	0	0	0	1	-360	360;
	2483	5603	0.00895	0.07061	0	0	0	0	0	0	1	-360	360;
	2483	4530	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	1892	6403	0.00079	0.00253	0	0	0	0	0	0	1	-360	360;
	1892	7746	0.00451	0.02149	0	0	0	0	0	0	1	-360	360;
	1892	7302	0.01458	0.06949	0	0	0	0	0	0	1	-360	360;
	1235	7746	0.02044	0.06563	0	0	0	0	0	0	1	-360	360;
	1235	824	0.03316	0.10649	0	0	0	0	0	0	1	-360	360;
	7746	9102	0.00441	0.02101	0	0	0	0	0	0	1	-360	360;
	7746	9102	0.00441	0.02101	0	0	0	0	0	0	1	-360	360;
	7746	6289	0.02373	0.11312	0	0	0	0	0	0	1	-360	360;
	3269	5422	0.00251	0.01977	0	0	0	0	0	0	1	-360	360;
	3269	2324	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	2137	5604	0.00607	0.02893	0	0	0	0	0	0	1	-360	360;
	2137	5321	0.0003	0.0032	0	0	0	0	0	0	1	-360	360;
	5422	2324	0.00106	0.00832	0	0	0	0	0	0	1	-360	360;
	4626	4749	0.00762	0.06009	0	0	0	0	0	0	1	-360	360;
	6044	8047	0.00524	0.03214	0	0	0	0	0	0	1	-360	360;
	6044	8047	0.00524	0.03214	0	0	0	0	0	0	1	-360	360;
	6044	3766	0.00068	0.00323	0	0	0	0	0	0	1	-360	360;
	6379	3564	0.00526	0.02505	0	0	0	0	0	0	1	-360	360;
	6379	3564	0.00526	0.02505	0	0	0	0	0	0	1	-360	360;
	6379	4749	0.00345	0.0353	0	280	0	0	0	0	1	-360	360;
	6379	4749	0.00345	0.0353	0	0	0	0	0	0	1	-360	360;
	6379	6488	0.00122	0.01306	0	0	0	0	0	0	1	-360	360;
	6379	6488	0.00122	0.01306	0	0	0	0	0	0	1	-360	360;
	8047	2324	0.00796	0.06276	0	0	0	0	0	0	1	-360	360;
	5603	4530	0.00677	0.05335	0	0	0	0	0	0	1	-360	360;
	2324	6840	0.00239	0.01883	0	0	0	0	0	0	1	-360	360;
	2324	70	0.00239	0.01883	0	0	0	0	0	0	1	-360	360;
	8695	9096	0.01407	0.0452	0	0	0	0	0	0	1	-360	360;
	9102	5604	0.00763	0.03636	0	0	0	0	0	0	1	-360	360;
	6840	5321	0.00022	0.0024	0	280	0	0	0	0	1	-360	360;
	70	5321	0.00022	0.0024	0	280	0	0	0	0	1	-360	360;
	3220	1152	0.00282	0.03584	0	0	0	0	0	0	1	-360	360;
	3220	7928	0	-0.0228	0	1143	0	0	0	0	1	-360	360;
	5478	7928	0	-0.0228	0	1039	0	0	0	0	1	-360	360;
	8305	7928	0	-0.0228	0	866	0	0	0	0	1	-360	360;
	8305	4519	0.00219	0.02793	0	1905	0	0	0	0	1	-360	360;
	7928	8227	0.00198	0.02025	0	1455	0	0	0	0	1	-360	360;
	7928	6124	0.00248	0.02534	0	1455	0	0	0	0	1	-360	360;
	4582	434	0.02365	0.07594	0	0	0	0	0	0	1	-360	360;
	4582	5756	0.0532	0.17086	0	0	0	0	0	0	1	-360	360;
	8491	8039	0.02252	0.07232	0	0	0	0	0	0	1	-360	360;
	8491	6325	0.02252	0.07232	0	0	0	0	0	0	1	-360	360;
	1626	8733	0.02624	0.08425	0	0	0	0	0	0	1	-360	360;
	1626	7134	0.00951	0.03056	0	0	0	0	0	0	1	-360	360;
	1626	16	0.01802	0.05786	0	0	0	0	0	0	1	-360	360;
	9087	8259	0.00845	0.02712	0	0	0	0	0	0	1	-360	360;
	9087	623	0.02736	0.08787	0	0	0	0	0	0	1	-360	360;
	1113	6467	0.01393	0.10983	0	0	0	0	0	0	1	-360	360;
	1113	4071	0.0131	0.08035	0	267	0	0	0	0	1	-360	360;
	1113	6700	0.01847	0.0593	0	0	0	0	0	0	1	-360	360;
	824	6400	0.06942	0.22293	0	0	0	0	0	0	1	-360	360;
	824	8543	0.0183	0.05876	0	0	0	0	0	0	1	-360	360;
	824	2405	0.04065	0.13054	0	0	0	0	0	0	1	-360	360;
	6400	8039	0.01098	0.03526	0	0	0	0	0	0	1	-360	360;
	6400	8259	0.03665	0.1177	0	0	0	0	0	0	1	-360	360;
	6400	623	0.03001	0.09637	0	0	0	0	0	0	1	-360	360;
	6400	3625	0.00478	0.03766	0	0	0	0	0	0	1	-360	360;
	6400	3625	0.00478	0.03766	0	0	0	0	0	0	1	-360	360;
	8039	8259	0.01592	0.12552	0	0	0	0	0	0	1	-360	360;
	8733	5810	0.04189	0.13452	0	0	0	0	0	0	1	-360	360;
	8259	3708	0.02049	0.06581	0	0	0	0	0	0	1	-360	360;
	7134	203	0.06227	0.19997	0	0	0	0	0	0	1	-360	360;
	8543	152	0.03806	0.12222	0	0	0	0	0	0	1	-360	360;
	8543	2049	0.01393	0.10983	0	0	0	0	0	0	1	-360	360;
	5356	434	0.01671	0.07967	0	0	0	0	0	0	1	-360	360;
	5356	3379	0.02712	0.12928	0	0	0	0	0	0	1	-360	360;
	434	4358	0.03361	0.10794	0	0	0	0	0	0	1	-360	360;
	6467	6325	0.01204	0.05737	0	0	0	0	0	0	1	-360	360;
	16	3708	0.02522	0.081	0	0	0	0	0	0	1	-360	360;
	16	5756	0.00591	0.01898	0	0	0	0	0	0	1	-360	360;
	16	4199	0.04375	0.14048	0	0	0	0	0	0	1	-360	360;
	4412	7244	0.01171	0.03761	0	0	0	0	0	0	1	-360	360;
	4412	6087	0.03389	0.10884	0	0	0	0	0	0	1	-360	360;
	7244	6325	0.01019	0.03272	0	0	0	0	0	0	1	-360	360;
	4071	2141	0.0131	0.08035	0	0	0	0	0	0	1	-360	360;
	4688	7157	0	-0.03116	0	866	0	0	0	0	1	-360	360;
	4688	9022	0.00123	0.02042	0	1005	0	0	0	0	1	-360	360;
	8945	2165	0	-0.03116	0	866	0	0	0	0	1	-360	360;
	8945	9022	0.00123	0.02042	0	901	0	0	0	0	1	-360	360;
	323	6740	0	-0.0228	0	901	0	0	0	0	1	-360	360;
	323	6050	0	-0.0228	0	1212	0	0	0	0	1	-360	360;
	1126	4237	0	-0.023	0	935	0	0	0	0	1	-360	360;
	1126	6781	0	-0.023	0	935	0	0	0	0	1	-360	360;
	6740	715	0.00262	0.03338	0	2009	0	0	0	0	1	-360	360;
	6781	5432	0.00266	0.03382	0	2044	0	0	0	0	1	-360	360;
	8679	634	0	-0.0242	0	797	0	0	0	0	1	-360	360;
	5243	634	0	-0.0242	0	797	0	0	0	0	1	-360	360;
	634	715	0.00287	0.02938	0	1247	0	0	0	0	1	-360	360;
	634	6008	0.00544	0.05508	0	1282	0	0	0	0	1	-360	360;
	634	6008	0.00544	0.05508	0	1282	0	0	0	0	1	-360	360;
	9022	968	0.00139	0.02301	0	935	0	0	0	0	1	-360	360;
	9022	5079	0.00139	0.02301	0	935	0	0	0	0	1	-360	360;
	9022	3064	0.00099	0.01644	0	866	0	0	0	0	1	-360	360;
	9022	7083	0.00294	0.03741	0	1975	0	0	0	0	1	-360	360;
	6551	2141	0.02252	0.07232	0	0	0	0	0	0	1	-360	360;
	6551	6700	0.0174	0.05587	0	0	0	0	0	0	1	-360	360;
	6551	4358	0.02509	0.11958	0	0	0	0	0	0	1	-360	360;
	6551	6535	0.00902	0.04299	0	0	0	0	0	0	1	-360	360;
	6551	6535	0.00902	0.04299	0	0	0	0	0	0	1	-360	360;
	6551	2547	0.04087	0.13126	0	0	0	0	0	0	1	-360	360;
	6551	7184	0.02196	0.07051	0	0	0	0	0	0	1	-360	360;
	2490	7581	0.01582	0.0508	0	0	0	0	0	0	1	-360	360;
	2490	668	0.04712	0.15133	0	0	0	0	0	0	1	-360	360;
	2141	160	0.02055	0.06599	0	0	0	0	0	0	1	-360	360;
	2141	6476	0.00339	0.01616	0	0	0	0	0	0	1	-360	360;
	2141	6476	0.00339	0.01616	0	0	0	0	0	0	1	-360	360;
	2141	3283	0.01098	0.03526	0	0	0	0	0	0	1	-360	360;
	3870	6087	0.03271	0.10504	0	0	0	0	0	0	1	-360	360;
	3870	3980	0.02145	0.06888	0	0	0	0	0	0	1	-360	360;
	3870	7302	0.00644	0.0307	0	0	0	0	0	0	1	-360	360;
	3870	6289	0.00644	0.0307	0	0	0	0	0	0	1	-360	360;
	3870	1621	0.03029	0.09727	0	0	0	0	0	0	1	-360	360;
	4358	160	0.00255	0.02008	0	0	0	0	0	0	1	-360	360;
	4358	160	0.00434	0.02069	0	0	0	0	0	0	1	-360	360;
	4358	160	0.00434	0.02069	0	0	0	0	0	0	1	-360	360;
	2952	160	0.01633	0.05243	0	0	0	0	0	0	1	-360	360;
	160	6476	0.00729	0.03474	0	0	0	0	0	0	1	-360	360;
	160	6476	0.00729	0.03474	0	0	0	0	0	0	1	-360	360;
	6476	1371	0.00034	0.00162	0	0	0	0	0	0	1	-360	360;
	6476	1371	0.00034	0.00162	0	0	0	0	0	0	1	-360	360;
	8297	7302	0.00085	0.00404	0	0	0	0	0	0	1	-360	360;
	8297	6289	0.00085	0.00404	0	0	0	0	0	0	1	-360	360;
	6087	2547	0.00501	0.01609	0	0	0	0	0	0	1	-360	360;
	820	7581	0.02015	0.06473	0	0	0	0	0	0	1	-360	360;
	820	7184	0.01407	0.0452	0	0	0	0	0	0	1	-360	360;
	7581	5287	0.00935	0.07374	0	0	0	0	0	0	1	-360	360;
	5845	3283	0.01042	0.03345	0	0	0	0	0	0	1	-360	360;
	5845	5287	0.01478	0.07046	0	0	0	0	0	0	1	-360	360;
	3980	668	0.04572	0.14681	0	0	0	0	0	0	1	-360	360;
	5287	7949	0.01029	0.08112	0	0	0	0	0	0	1	-360	360;
	5287	7949	0.01029	0.08112	0	0	0	0	0	0	1	-360	360;
	5287	7184	0.00836	0.0659	0	0	0	0	0	0	1	-360	360;
	4097	1925	0	-0.0207	0	831	0	0	0	0	1	-360	360;
	1925	1515	0.00193	0.01978	0	1178	0	0	0	0	1	-360	360;
	1925	5934	0.00224	0.02269	0	1282	0	0	0	0	1	-360	360;
	7769	8523	0.01071	0.08441	0	0	0	0	0	0	1	-360	360;
	2119	2070	0.02733	0.09705	0	0	0	0	0	0	1	-360	360;
	2119	1621	0.03029	0.09727	0	0	0	0	0	0	1	-360	360;
	8996	8937	0.02308	0.07413	0	0	0	0	0	0	1	-360	360;
	8996	2070	0.02305	0.08331	0	0	0	0	0	0	1	-360	360;
	4812	1130	0.03288	0.10559	0	0	0	0	0	0	1	-360	360;
	4812	50	0.0077	0.04725	0	0	0	0	0	0	1	-360	360;
	4812	7749	0.00796	0.06276	0	0	0	0	0	0	1	-360	360;
	4812	4611	0.00648	0.02079	0	0	0	0	0	0	1	-360	360;
	4278	6578	0.00895	0.07061	0	0	0	0	0	0	1	-360	360;
	4278	6868	0.01407	0.0452	0	0	0	0	0	0	1	-360	360;
	1130	5788	0.00816	0.02622	0	0	0	0	0	0	1	-360	360;
	1130	6578	0.00959	0.07563	0	0	0	0	0	0	1	-360	360;
	1130	5838	0.01017	0.04848	0	0	0	0	0	0	1	-360	360;
	3439	2640	0.01786	0.08516	0	0	0	0	0	0	1	-360	360;
	3439	50	0.01926	0.11812	0	0	0	0	0	0	1	-360	360;
	1460	863	0.00239	0.01883	0	0	0	0	0	0	1	-360	360;
	1460	863	0.00239	0.01883	0	0	0	0	0	0	1	-360	360;
	1460	8523	0.04076	0.1309	0	0	0	0	0	0	1	-360	360;
	1460	50	0.00814	0.06417	0	0	0	0	0	0	1	-360	360;
	863	8937	0.00776	0.06119	0	0	0	0	0	0	1	-360	360;
	8937	50	0.00665	0.04082	0	0	0	0	0	0	1	-360	360;
	8937	50	0.00665	0.04082	0	0	0	0	0	0	1	-360	360;
	2640	5788	0.04943	0.15874	0	0	0	0	0	0	1	-360	360;
	8523	5244	0.02196	0.07051	0	0	0	0	0	0	1	-360	360;
	8523	4611	0.04155	0.13343	0	0	0	0	0	0	1	-360	360;
	1515	7083	0	-0.01731	0	866	0	0	0	0	1	-360	360;
	50	4459	0.01127	0.0691	0	0	0	0	0	0	1	-360	360;
	5244	6931	0.04279	0.13741	0	0	0	0	0	0	1	-360	360;
	5765	4611	0.00113	0.00362	0	0	0	0	0	0	1	-360	360;
	5615	5040	0.00108	0.00847	0	0	0	0	0	0	1	-360	360;
	5615	2477	0.05089	0.16344	0	0	0	0	0	0	1	-360	360;
	5040	3803	0.05315	0.17067	0	0	0	0	0	0	1	-360	360;
	5040	6979	0.00995	0.07845	0	0	0	0	0	0	1	-360	360;
	6931	3803	0.03406	0.10938	0	0	0	0	0	0	1	-360	360;
	6931	9138	0.00966	0.04606	0	0	0	0	0	0	1	-360	360;
	6931	7634	0.03429	0.11011	0	0	0	0	0	0	1	-360	360;
	2477	5599	0.01858	0.05966	0	0	0	0	0	0	1	-360	360;
	2477	6617	0.00466	0.03672	0	0	0	0	0	0	1	-360	360;
	2064	57	0.01485	0.07078	0	0	0	0	0	0	1	-360	360;
	2064	3091	0.01509	0.04845	0	0	0	0	0	0	1	-360	360;
	2064	1433	0.01824	0.05858	0	0	0	0	0	0	1	-360	360;
	2064	5105	0.01464	0.04701	0	0	0	0	0	0	1	-360	360;
	57	8075	0.00432	0.03405	0	0	0	0	0	0	1	-360	360;
	57	6945	0.01183	0.0564	0	0	0	0	0	0	1	-360	360;
	3803	8766	0.03367	0.10812	0	0	0	0	0	0	1	-360	360;
	3803	6979	0.04712	0.15133	0	0	0	0	0	0	1	-360	360;
	5599	1433	0.0165	0.05297	0	0	0	0	0	0	1	-360	360;
	5599	6414	0.02204	0.10504	0	0	0	0	0	0	1	-360	360;
	8075	6617	0.00219	0.01726	0	0	0	0	0	0	1	-360	360;
	8075	1191	0.00225	0.00723	0	0	0	0	0	0	1	-360	360;
	3091	5864	0.00715	0.02296	0	0	0	0	0	0	1	-360	360;
	6617	6979	0.01791	0.14121	0	0	0	0	0	0	1	-360	360;
	9138	6414	0.01356	0.06464	0	0	0	0	0	0	1	-360	360;
	6401	2718	0.01768	0.05677	0	0	0	0	0	0	1	-360	360;
	6401	7570	0.0103	0.03309	0	0	0	0	0	0	1	-360	360;
	8227	6124	0.00124	0.01579	0	1871	0	0	0	0	1	-360	360;
	6124	3304	0.00037	0.00374	0	1247	0	0	0	0	1	-360	360;
	6124	3304	0.00037	0.00374	0	1247	0	0	0	0	1	-360	360;
	6124	3475	0.00207	0.03419	0	797	0	0	0	0	1	-360	360;
	6796	2718	0.00946	0.04509	0	0	0	0	0	0	1	-360	360;
	6796	3959	0.00576	0.02747	0	0	0	0	0	0	1	-360	360;
	2090	7094	0.02387	0.07666	0	0	0	0	0	0	1	-360	360;
	2090	5187	0.04842	0.15549	0	0	0	0	0	0	1	-360	360;
	2718	3959	0.00508	0.02424	0	0	0	0	0	0	1	-360	360;
	2718	7094	0.05444	0.17483	0	0	0	0	0	0	1	-360	360;
	6211	8262	0.0326	0.10468	0	0	0	0	0	0	1	-360	360;
	6211	3638	0.02269	0.07286	0	0	0	0	0	0	1	-360	360;
	7850	2524	0.01831	0.08726	0	0	0	0	0	0	1	-360	360;
	7850	4386	0.04183	0.13433	0	0	0	0	0	0	1	-360	360;
	5810	5277	0.00513	0.04048	0	0	0	0	0	0	1	-360	360;
	5810	524	0.00169	0.00542	0	0	0	0	0	0	1	-360	360;
	5810	7355	0.00541	0.04268	0	0	0	0	0	0	1	-360	360;
	5810	2485	0.00396	0.03122	0	0	0	0	0	0	1	-360	360;
	5810	114	0.02288	0.10908	0	0	0	0	0	0	1	-360	360;
	5810	5449	0.02881	0.13736	0	0	0	0	0	0	1	-360	360;
	8136	2049	0.03941	0.12656	0	0	0	0	0	0	1	-360	360;
	8136	3638	0.01396	0.04484	0	0	0	0	0	0	1	-360	360;
	8262	3424	0.03969	0.12746	0	0	0	0	0	0	1	-360	360;
	2	2447	0.00034	0.00162	0	0	0	0	0	0	1	-360	360;
	2	4898	0.00051	0.00242	0	0	0	0	0	0	1	-360	360;
	2007	4347	0.02145	0.06888	0	0	0	0	0	0	1	-360	360;
	2007	1024	0.00518	0.01663	0	0	0	0	0	0	1	-360	360;
	2007	2447	0.00085	0.00404	0	0	0	0	0	0	1	-360	360;
	2007	4898	0.00052	0.00246	0	0	0	0	0	0	1	-360	360;
	2007	8350	0.00555	0.04377	0	0	0	0	0	0	1	-360	360;
	2265	152	0.02252	0.07232	0	0	0	0	0	0	1	-360	360;
	2265	8518	0.0165	0.05297	0	0	0	0	0	0	1	-360	360;
	152	8384	0.02398	0.07702	0	0	0	0	0	0	1	-360	360;
	8384	2049	0.00822	0.0264	0	0	0	0	0	0	1	-360	360;
	7906	623	0.04954	0.1591	0	0	0	0	0	0	1	-360	360;
	7906	2405	0.0237	0.07612	0	0	0	0	0	0	1	-360	360;
	3785	2524	0.02117	0.06798	0	0	0	0	0	0	1	-360	360;
	3785	2405	0.02809	0.09022	0	0	0	0	0	0	1	-360	360;
	3785	6553	0.01126	0.03616	0	0	0	0	0	0	1	-360	360;
	3424	6553	0.01278	0.04104	0	0	0	0	0	0	1	-360	360;
	3424	5187	0.0036	0.01157	0	0	0	0	0	0	1	-360	360;
	8133	623	0.02871	0.09221	0	0	0	0	0	0	1	-360	360;
	8133	7570	0.03384	0.10866	0	0	0	0	0	0	1	-360	360;
	623	7570	0.04971	0.15965	0	0	0	0	0	0	1	-360	360;
	623	4386	0.01576	0.05062	0	0	0	0	0	0	1	-360	360;
	7570	5277	0.00306	0.02416	0	0	0	0	0	0	1	-360	360;
	7570	7355	0.00346	0.0273	0	0	0	0	0	0	1	-360	360;
	7570	2485	0.00522	0.02489	0	0	0	0	0	0	1	-360	360;
	5277	7355	0.00048	0.00226	0	0	0	0	0	0	1	-360	360;
	5277	8326	0.00048	0.00226	0	0	0	0	0	0	1	-360	360;
	3638	4347	0.00766	0.02459	0	0	0	0	0	0	1	-360	360;
	3638	8350	0.00557	0.04393	0	0	0	0	0	0	1	-360	360;
	8326	2485	0.00237	0.01131	0	0	0	0	0	0	1	-360	360;
	3475	6893	0.00258	0.04274	0	797	0	0	0	0	1	-360	360;
	85	1906	0.00861	0.04105	0	0	0	0	0	0	1	-360	360;
	85	847	0.01719	0.08193	0	0	0	0	0	0	1	-360	360;
	85	847	0.01719	0.08193	0	0	0	0	0	0	1	-360	360;
	85	4377	0.01305	0.06222	0	0	0	0	0	0	1	-360	360;
	2414	3323	0.01326	0.06319	0	0	0	0	0	0	1	-360	360;
	2414	3799	0.03288	0.15675	0	0	0	0	0	0	1	-360	360;
	2414	7908	0.0059	0.02812	0	0	0	0	0	0	1	-360	360;
	2414	9074	0.01898	0.0905	0	0	0	0	0	0	1	-360	360;
	9123	5399	0.02173	0.06979	0	0	0	0	0	0	1	-360	360;
	9123	7217	0.01007	0.07939	0	0	0	0	0	0	1	-360	360;
	9123	5701	0.04673	0.15006	0	0	0	0	0	0	1	-360	360;
	9123	5023	0.001	0.00785	0	0	0	0	0	0	1	-360	360;
	5399	847	0.02168	0.06961	0	0	0	0	0	0	1	-360	360;
	1906	3323	0.01261	0.06011	0	0	0	0	0	0	1	-360	360;
	3800	3799	0.01464	0.04701	0	0	0	0	0	0	1	-360	360;
	5597	3217	0.01647	0.07854	0	0	0	0	0	0	1	-360	360;
	5597	3217	0.0264	0.08479	0	0	0	0	0	0	1	-360	360;
	5597	3799	0.01441	0.06868	0	0	0	0	0	0	1	-360	360;
	5597	3799	0.02393	0.07684	0	0	0	0	0	0	1	-360	360;
	1573	3217	0.01407	0.06706	0	0	0	0	0	0	1	-360	360;
	1573	7995	0.00508	0.02424	0	0	0	0	0	0	1	-360	360;
	847	1846	0.00557	0.04393	0	0	0	0	0	0	1	-360	360;
	847	1846	0.00557	0.04393	0	0	0	0	0	0	1	-360	360;
	847	553	0.00393	0.0241	0	0	0	0	0	0	1	-360	360;
	3323	4377	0.00783	0.03733	0	0	0	0	0	0	1	-360	360;
	3323	7908	0.01824	0.08694	0	0	0	0	0	0	1	-360	360;
	3635	3799	0.00951	0.075	0	0	0	0	0	0	1	-360	360;
	3635	2452	0.00438	0.03452	0	0	0	0	0	0	1	-360	360;
	655	4882	0.02314	0.07431	0	0	0	0	0	0	1	-360	360;
	655	2255	0.0199	0.1569	0	0	0	0	0	0	1	-360	360;
	655	9089	0.00818	0.06449	0	0	0	0	0	0	1	-360	360;
	3217	3799	0.03085	0.14706	0	0	0	0	0	0	1	-360	360;
	3217	114	0.01814	0.08646	0	0	0	0	0	0	1	-360	360;
	3217	5449	0.01187	0.05656	0	0	0	0	0	0	1	-360	360;
	3217	7995	0.01356	0.06464	0	0	0	0	0	0	1	-360	360;
	4377	4001	0.05844	0.18767	0	0	0	0	0	0	1	-360	360;
	3799	2452	0.01075	0.08473	0	0	0	0	0	0	1	-360	360;
	3799	7908	0.04729	0.15187	0	0	0	0	0	0	1	-360	360;
	3799	9074	0.01458	0.06949	0	0	0	0	0	0	1	-360	360;
	4882	9089	0.00774	0.06103	0	0	0	0	0	0	1	-360	360;
	2452	7908	0.01094	0.0863	0	0	0	0	0	0	1	-360	360;
	2452	7908	0.01094	0.0863	0	0	0	0	0	0	1	-360	360;
	6893	2181	0.00033	0.00416	0	1801	0	0	0	0	1	-360	360;
	1846	7217	0.00318	0.0251	0	0	0	0	0	0	1	-360	360;
	1846	4264	0.00199	0.01569	0	0	0	0	0	0	1	-360	360;
	7217	5023	0.00597	0.04707	0	0	0	0	0	0	1	-360	360;
	7217	4264	0.00199	0.01569	0	0	0	0	0	0	1	-360	360;
	7128	4225	0.00238	0.03034	0	2009	0	0	0	0	1	-360	360;
	7128	6085	0.00264	0.03358	0	1871	0	0	0	0	1	-360	360;
	7128	2181	0.00228	0.02909	0	1940	0	0	0	0	1	-360	360;
	6951	2765	0.05152	0.16543	0	0	0	0	0	0	1	-360	360;
	6951	4722	0.01719	0.13556	0	0	0	0	0	0	1	-360	360;
	6951	2396	0.02074	0.16349	0	0	0	0	0	0	1	-360	360;
	5701	7206	0.04358	0.13994	0	0	0	0	0	0	1	-360	360;
	793	4001	0.03136	0.10071	0	0	0	0	0	0	1	-360	360;
	793	5667	0.00782	0.06166	0	0	0	0	0	0	1	-360	360;
	793	8323	0.01807	0.05804	0	0	0	0	0	0	1	-360	360;
	793	1157	0.00022	0.00072	0	0	0	0	0	0	1	-360	360;
	4001	5338	0.03158	0.10143	0	0	0	0	0	0	1	-360	360;
	7206	2255	0.02477	0.07955	0	0	0	0	0	0	1	-360	360;
	2765	5689	0.02314	0.07431	0	0	0	0	0	0	1	-360	360;
	3287	3521	0.01464	0.04701	0	0	0	0	0	0	1	-360	360;
	3287	5338	0.02902	0.13833	0	0	0	0	0	0	1	-360	360;
	3287	3055	0.0241	0.07738	0	0	0	0	0	0	1	-360	360;
	3287	203	0.07702	0.24733	0	0	0	0	0	0	1	-360	360;
	6980	5264	0.00619	0.01989	0	0	0	0	0	0	1	-360	360;
	2599	5667	0.00448	0.02748	0	0	0	0	0	0	1	-360	360;
	2599	5667	0.00448	0.02748	0	0	0	0	0	0	1	-360	360;
	2599	412	0.02929	0.13962	0	0	0	0	0	0	1	-360	360;
	2599	4722	0.01572	0.12395	0	0	0	0	0	0	1	-360	360;
	2599	4608	0.01194	0.09414	0	0	0	0	0	0	1	-360	360;
	5667	637	0.00629	0.04958	0	0	0	0	0	0	1	-360	360;
	3521	3458	0.03491	0.1121	0	0	0	0	0	0	1	-360	360;
	412	2826	0.01593	0.07595	0	0	0	0	0	0	1	-360	360;
	4722	2255	0.05405	0.17357	0	0	0	0	0	0	1	-360	360;
	4722	4608	0.01174	0.09257	0	0	0	0	0	0	1	-360	360;
	5689	2255	0.03254	0.19959	0	0	0	0	0	0	1	-360	360;
	5264	203	0.00659	0.02115	0	0	0	0	0	0	1	-360	360;
	5264	203	0.00659	0.02115	0	0	0	0	0	0	1	-360	360;
	3055	8323	0.03356	0.10776	0	0	0	0	0	0	1	-360	360;
	3055	168	0.03277	0.10523	0	0	0	0	0	0	1	-360	360;
	3458	203	0.03783	0.1215	0	0	0	0	0	0	1	-360	360;
	203	5349	0.0181	0.08629	0	0	0	0	0	0	1	-360	360;
	4225	7642	0.00235	0.02992	0	1905	0	0	0	0	1	-360	360;
	7026	5432	0.00211	0.02153	0	1039	0	0	0	0	1	-360	360;
	320	5432	0.00232	0.02949	0	1593	0	0	0	0	1	-360	360;
	320	7780	0.00179	0.02959	0	970	0	0	0	0	1	-360	360;
	320	7434	0.00179	0.02959	0	1005	0	0	0	0	1	-360	360;
	320	2874	0.00134	0.01702	0	1905	0	0	0	0	1	-360	360;
	320	4469	0.00084	0.01064	0	2044	0	0	0	0	1	-360	360;
	320	5098	0.00124	0.01268	0	866	0	0	0	0	1	-360	360;
	320	3194	0.00106	0.01347	0	1732	0	0	0	0	1	-360	360;
	715	5432	0.00023	0.00231	0	1282	0	0	0	0	1	-360	360;
	715	6008	0.00335	0.03421	0	1282	0	0	0	0	1	-360	360;
	5432	4613	0.00024	0.00241	0	1247	0	0	0	0	1	-360	360;
	5432	4613	0.00024	0.00241	0	1247	0	0	0	0	1	-360	360;
	5432	4613	0.00024	0.00241	0	1247	0	0	0	0	1	-360	360;
	5432	4613	0.00024	0.00241	0	1247	0	0	0	0	1	-360	360;
	5432	3064	0.00075	0.01233	0	797	0	0	0	0	1	-360	360;
	7780	968	0	-0.03116	0	970	0	0	0	0	1	-360	360;
	7434	5079	0	-0.03116	0	970	0	0	0	0	1	-360	360;
	2874	5751	3e-05	0.00033	0	1940	0	0	0	0	1	-360	360;
	3907	6008	0.00182	0.02317	0	1767	0	0	0	0	1	-360	360;
	3907	4519	0.00081	0.01031	0	2009	0	0	0	0	1	-360	360;
	2241	6008	0.00177	0.01813	0	831	0	0	0	0	1	-360	360;
	2241	6008	0.00175	0.01789	0	831	0	0	0	0	1	-360	360;
	2241	3194	0.00099	0.01257	0	1905	0	0	0	0	1	-360	360;
	2241	407	0.00125	0.01594	0	2009	0	0	0	0	1	-360	360;
	2241	407	0.00125	0.01594	0	2009	0	0	0	0	1	-360	360;
	6008	1117	0.00014	0.00139	0	1282	0	0	0	0	1	-360	360;
	6008	1117	0.00014	0.00139	0	1282	0	0	0	0	1	-360	360;
	6008	1117	0.00014	0.00139	0	1282	0	0	0	0	1	-360	360;
	6008	6085	0.00198	0.02527	0	1940	0	0	0	0	1	-360	360;
	1044	5220	0.00845	0.02712	0	0	0	0	0	0	1	-360	360;
	1044	20	0.00736	0.05805	0	0	0	0	0	0	1	-360	360;
	1044	1357	0.02438	0.07829	0	0	0	0	0	0	1	-360	360;
	5220	2979	0.02308	0.07413	0	0	0	0	0	0	1	-360	360;
	5220	5180	0.01802	0.05786	0	0	0	0	0	0	1	-360	360;
	5220	126	0.02252	0.07232	0	0	0	0	0	0	1	-360	360;
	6228	8595	0.01397	0.08565	0	0	0	0	0	0	1	-360	360;
	6228	6751	0.01937	0.06219	0	0	0	0	0	0	1	-360	360;
	6228	8196	0.01205	0.07392	0	0	0	0	0	0	1	-360	360;
	5889	7683	0.00788	0.02531	0	0	0	0	0	0	1	-360	360;
	5889	4834	0.01464	0.04701	0	0	0	0	0	0	1	-360	360;
	20	2979	0.00394	0.01266	0	0	0	0	0	0	1	-360	360;
	20	3943	0.03491	0.1121	0	0	0	0	0	0	1	-360	360;
	20	8460	0.02534	0.08136	0	0	0	0	0	0	1	-360	360;
	2186	1381	0.0469	0.15061	0	0	0	0	0	0	1	-360	360;
	2186	1381	0.0469	0.15061	0	0	0	0	0	0	1	-360	360;
	2186	5430	0.03913	0.12566	0	0	0	0	0	0	1	-360	360;
	2186	5430	0.03913	0.12566	0	0	0	0	0	0	1	-360	360;
	2186	7147	0.05332	0.17309	0	0	0	0	0	0	1	-360	360;
	2186	7147	0.05332	0.17309	0	0	0	0	0	0	1	-360	360;
	3943	1329	0.01015	0.08002	0	0	0	0	0	0	1	-360	360;
	3943	8595	0.0317	0.10179	0	0	0	0	0	0	1	-360	360;
	3943	6498	0.00719	0.03426	0	0	0	0	0	0	1	-360	360;
	3943	6498	0.00719	0.03426	0	0	0	0	0	0	1	-360	360;
	3943	3822	0.00143	0.0113	0	0	0	0	0	0	1	-360	360;
	3943	3822	0.00143	0.0113	0	0	0	0	0	0	1	-360	360;
	1565	5538	0.03463	0.11119	0	0	0	0	0	0	1	-360	360;
	1565	7716	0.01306	0.04195	0	0	0	0	0	0	1	-360	360;
	1565	8121	0.01689	0.05424	0	0	0	0	0	0	1	-360	360;
	5800	1054	0.03074	0.09872	0	0	0	0	0	0	1	-360	360;
	5800	7009	0.03547	0.1139	0	0	0	0	0	0	1	-360	360;
	454	7894	0.04622	0.14844	0	0	0	0	0	0	1	-360	360;
	454	6326	0.04949	0.15892	0	0	0	0	0	0	1	-360	360;
	454	5349	0.05106	0.16399	0	0	0	0	0	0	1	-360	360;
	1054	3379	0.0282	0.13445	0	0	0	0	0	0	1	-360	360;
	1054	7147	0.01391	0.04653	0	0	0	0	0	0	1	-360	360;
	1054	7147	0.01391	0.04653	0	0	0	0	0	0	1	-360	360;
	1054	3271	0.00282	0.00904	0	0	0	0	0	0	1	-360	360;
	1054	3271	0.00282	0.00904	0	0	0	0	0	0	1	-360	360;
	7370	3660	0.03744	0.12023	0	0	0	0	0	0	1	-360	360;
	4523	3660	0.00974	0.03128	0	0	0	0	0	0	1	-360	360;
	4523	3660	0.01357	0.04357	0	0	0	0	0	0	1	-360	360;
	4523	1329	0.00563	0.01808	0	0	0	0	0	0	1	-360	360;
	4523	3845	0.02263	0.07268	0	0	0	0	0	0	1	-360	360;
	5010	3271	0.03265	0.10486	0	0	0	0	0	0	1	-360	360;
	5010	2122	0.03491	0.1121	0	0	0	0	0	0	1	-360	360;
	3660	4818	0.00444	0.02117	0	0	0	0	0	0	1	-360	360;
	3660	4818	0.00444	0.02117	0	0	0	0	0	0	1	-360	360;
	3660	5430	0.01605	0.05153	0	0	0	0	0	0	1	-360	360;
	3660	5430	0.01605	0.05153	0	0	0	0	0	0	1	-360	360;
	3660	4640	0.01407	0.0452	0	0	0	0	0	0	1	-360	360;
	4818	1329	0.01244	0.03996	0	0	0	0	0	0	1	-360	360;
	4818	1329	0.01244	0.03996	0	0	0	0	0	0	1	-360	360;
	1329	8100	0.01271	0.0606	0	0	0	0	0	0	1	-360	360;
	7009	7894	0.02928	0.09402	0	0	0	0	0	0	1	-360	360;
	7009	7894	0.02928	0.09402	0	0	0	0	0	0	1	-360	360;
	7009	5649	0.04909	0.15766	0	0	0	0	0	0	1	-360	360;
	5180	2584	0.04335	0.13922	0	0	0	0	0	0	1	-360	360;
	7894	4199	0.02695	0.12847	0	0	0	0	0	0	1	-360	360;
	7894	1612	0.02695	0.12847	0	0	0	0	0	0	1	-360	360;
	2584	4135	0.0122	0.09618	0	0	0	0	0	0	1	-360	360;
	2584	2346	0.0304	0.09763	0	0	0	0	0	0	1	-360	360;
	6326	7371	0.03536	0.11354	0	0	0	0	0	0	1	-360	360;
	6326	7013	0.00338	0.01085	0	0	0	0	0	0	1	-360	360;
	6326	2346	0.02837	0.09112	0	0	0	0	0	0	1	-360	360;
	6326	497	0.0139	0.06626	0	0	0	0	0	0	1	-360	360;
	5998	7386	0.00325	0.01551	0	0	0	0	0	0	1	-360	360;
	5998	5430	0.01376	0.06561	0	0	0	0	0	0	1	-360	360;
	5998	7977	0.00472	0.03718	0	0	0	0	0	0	1	-360	360;
	5538	860	0.00756	0.05962	0	0	0	0	0	0	1	-360	360;
	5538	860	0.00756	0.05962	0	0	0	0	0	0	1	-360	360;
	5538	1887	0.00631	0.02025	0	0	0	0	0	0	1	-360	360;
	5538	263	0.01194	0.09414	0	0	0	0	0	0	1	-360	360;
	286	7386	0.0047	0.03703	0	0	0	0	0	0	1	-360	360;
	286	5430	0.0152	0.04882	0	0	0	0	0	0	1	-360	360;
	286	7977	0.00115	0.0091	0	0	0	0	0	0	1	-360	360;
	5649	7371	0.0179	0.05749	0	0	0	0	0	0	1	-360	360;
	5649	1381	0.01211	0.03887	0	0	0	0	0	0	1	-360	360;
	5649	126	0.02517	0.08082	0	0	0	0	0	0	1	-360	360;
	7386	2122	0.00619	0.01989	0	0	0	0	0	0	1	-360	360;
	5430	4640	0.01464	0.04701	0	0	0	0	0	0	1	-360	360;
	8391	5612	0.01757	0.05641	0	0	0	0	0	0	1	-360	360;
	8391	6136	0.01878	0.06123	0	0	0	0	0	0	1	-360	360;
	1357	3845	0.02674	0.08588	0	0	0	0	0	0	1	-360	360;
	4199	5349	0.04673	0.15006	0	0	0	0	0	0	1	-360	360;
	4199	1612	0.00271	0.01293	0	0	0	0	0	0	1	-360	360;
	4135	6136	0.02424	0.07877	0	0	0	0	0	0	1	-360	360;
	8595	6498	0.00822	0.0648	0	0	0	0	0	0	1	-360	360;
	8595	1186	0.02871	0.09221	0	0	0	0	0	0	1	-360	360;
	6294	2762	0.00263	0.02071	0	0	0	0	0	0	1	-360	360;
	6294	6751	0.02224	0.07142	0	0	0	0	0	0	1	-360	360;
	6294	5590	0.00261	0.02055	0	0	0	0	0	0	1	-360	360;
	6294	1186	0.00845	0.02712	0	0	0	0	0	0	1	-360	360;
	7716	8196	0.0114	0.0699	0	0	0	0	0	0	1	-360	360;
	2762	3693	0.01407	0.0452	0	0	0	0	0	0	1	-360	360;
	5612	860	0.00796	0.06276	0	0	0	0	0	0	1	-360	360;
	5612	8121	0.00563	0.01808	0	0	0	0	0	0	1	-360	360;
	5612	5394	0.01126	0.03616	0	0	0	0	0	0	1	-360	360;
	5590	1186	0.00597	0.04707	0	0	0	0	0	0	1	-360	360;
	5590	8196	0.01393	0.10983	0	0	0	0	0	0	1	-360	360;
	5590	5394	0.03828	0.12294	0	0	0	0	0	0	1	-360	360;
	4469	7789	0.00117	0.01496	0	1940	0	0	0	0	1	-360	360;
	5098	5751	0.00076	0.00964	0	2009	0	0	0	0	1	-360	360;
	9114	5683	0.00085	0.01081	0	1767	0	0	0	0	1	-360	360;
	9114	7789	0.00093	0.0118	0	1767	0	0	0	0	1	-360	360;
	1559	6136	0.00697	0.05491	0	0	0	0	0	0	1	-360	360;
	1559	6136	0.00697	0.05491	0	0	0	0	0	0	1	-360	360;
	313	2122	0.00113	0.00362	0	0	0	0	0	0	1	-360	360;
	2315	5394	0.00056	0.00181	0	0	0	0	0	0	1	-360	360;
	7013	3032	0.00338	0.01085	0	0	0	0	0	0	1	-360	360;
	407	8422	0.00122	0.01546	0	1940	0	0	0	0	1	-360	360;
	3302	2396	0.03378	0.10848	0	0	0	0	0	0	1	-360	360;
	3302	6812	0.0117	0.09226	0	0	0	0	0	0	1	-360	360;
	3302	2826	0.01891	0.14906	0	0	0	0	0	0	1	-360	360;
	3302	2826	0.01891	0.14906	0	0	0	0	0	0	1	-360	360;
	7986	3568	0.07702	0.24733	0	0	0	0	0	0	1	-360	360;
	7986	4370	0.01526	0.07272	0	0	0	0	0	0	1	-360	360;
	1805	4457	0.0308	0.0989	0	0	0	0	0	0	1	-360	360;
	1805	691	0.02787	0.0895	0	0	0	0	0	0	1	-360	360;
	1805	3260	0.00038	0.00298	0	0	0	0	0	0	1	-360	360;
	1805	3260	0.00038	0.00298	0	0	0	0	0	0	1	-360	360;
	2245	6812	0.02291	0.18059	0	0	0	0	0	0	1	-360	360;
	2245	2491	0.00593	0.02828	0	0	0	0	0	0	1	-360	360;
	2245	541	0.00776	0.06119	0	0	0	0	0	0	1	-360	360;
	6030	251	0.04262	0.13687	0	0	0	0	0	0	1	-360	360;
	6030	2508	0.02065	0.09841	0	0	0	0	0	0	1	-360	360;
	6030	4870	0.05591	0.17953	0	0	0	0	0	0	1	-360	360;
	6030	735	0.01196	0.0943	0	0	0	0	0	0	1	-360	360;
	6030	735	0.01196	0.0943	0	0	0	0	0	0	1	-360	360;
	505	8085	0.00644	0.0307	0	0	0	0	0	0	1	-360	360;
	505	8085	0.00644	0.0307	0	0	0	0	0	0	1	-360	360;
	7527	723	0.00241	0.01898	0	0	0	0	0	0	1	-360	360;
	7527	723	0.00241	0.01898	0	0	0	0	0	0	1	-360	360;
	7527	723	0.00249	0.01961	0	0	0	0	0	0	1	-360	360;
	7527	6281	0.01208	0.09524	0	0	0	0	0	0	1	-360	360;
	7527	3663	0.07229	0.23215	0	0	0	0	0	0	1	-360	360;
	723	982	0.0264	0.08479	0	0	0	0	0	0	1	-360	360;
	723	8085	0.00762	0.06009	0	0	0	0	0	0	1	-360	360;
	723	691	0.01057	0.08331	0	0	0	0	0	0	1	-360	360;
	723	613	0.0029	0.02291	0	0	0	0	0	0	1	-360	360;
	723	613	0.0029	0.02291	0	0	0	0	0	0	1	-360	360;
	4517	6927	0.01582	0.0508	0	0	0	0	0	0	1	-360	360;
	4517	6812	0.04887	0.15693	0	0	0	0	0	0	1	-360	360;
	2346	251	0.01668	0.07951	0	0	0	0	0	0	1	-360	360;
	2346	251	0.01668	0.07951	0	0	0	0	0	0	1	-360	360;
	2346	705	0.00749	0.02405	0	0	0	0	0	0	1	-360	360;
	2346	497	0.00617	0.02941	0	0	0	0	0	0	1	-360	360;
	2346	3032	0.02815	0.0904	0	0	0	0	0	0	1	-360	360;
	2346	728	0.02388	0.0804	0	0	0	0	0	0	1	-360	360;
	3568	6812	0.05461	0.17538	0	0	0	0	0	0	1	-360	360;
	3568	7950	0.02032	0.06527	0	0	0	0	0	0	1	-360	360;
	2396	7950	0.05236	0.16814	0	0	0	0	0	0	1	-360	360;
	982	8085	0.00555	0.04377	0	0	0	0	0	0	1	-360	360;
	982	2076	0.00788	0.02531	0	0	0	0	0	0	1	-360	360;
	6281	1887	0.01003	0.07908	0	0	0	0	0	0	1	-360	360;
	251	4900	0.01194	0.09414	0	0	0	0	0	0	1	-360	360;
	705	2076	0.01982	0.06364	0	0	0	0	0	0	1	-360	360;
	705	5352	0.00169	0.00542	0	0	0	0	0	0	1	-360	360;
	2508	2491	0.01393	0.10983	0	0	0	0	0	0	1	-360	360;
	1887	3663	0.01172	0.09241	0	0	0	0	0	0	1	-360	360;
	6001	2400	0.00441	0.02101	0	0	0	0	0	0	1	-360	360;
	4870	6812	0.04014	0.12891	0	0	0	0	0	0	1	-360	360;
	5744	5260	0.00597	0.04707	0	0	0	0	0	0	1	-360	360;
	5744	5260	0.00597	0.04707	0	0	0	0	0	0	1	-360	360;
	5744	2278	0.01017	0.04848	0	0	0	0	0	0	1	-360	360;
	5744	2278	0.01017	0.04848	0	0	0	0	0	0	1	-360	360;
	5744	4900	0.00298	0.02354	0	0	0	0	0	0	1	-360	360;
	5744	735	0.0002	0.00157	0	0	0	0	0	0	1	-360	360;
	5744	735	0.0002	0.00157	0	0	0	0	0	0	1	-360	360;
	6410	5240	0.01689	0.05424	0	0	0	0	0	0	1	-360	360;
	6410	828	4e-05	0.00041	0	0	0	0	0	0	1	-360	360;
	6410	3199	0.01848	0.06679	0	0	0	0	0	0	1	-360	360;
	6410	3199	0.02061	0.06617	0	0	0	0	0	0	1	-360	360;
	8359	3663	0.01925	0.06183	0	0	0	0	0	0	1	-360	360;
	8359	5240	0.00617	0.04864	0	0	0	0	0	0	1	-360	360;
	3663	263	0.00847	0.0404	0	0	0	0	0	0	1	-360	360;
	6927	4457	0.01064	0.03417	0	0	0	0	0	0	1	-360	360;
	6927	3199	0.04286	0.14508	0	0	0	0	0	0	1	-360	360;
	168	2400	0.01729	0.08242	0	0	0	0	0	0	1	-360	360;
	168	2400	0.01729	0.08242	0	0	0	0	0	0	1	-360	360;
	168	728	0.01149	0.04063	0	0	0	0	0	0	1	-360	360;
	3827	4370	0.01328	0.13577	0	0	0	0	0	0	1	-360	360;
	3827	3199	0.02061	0.06617	0	0	0	0	0	0	1	-360	360;
	5240	3260	0.02567	0.2024	0	0	0	0	0	0	1	-360	360;
	8422	7642	0.00268	0.03408	0	1767	0	0	0	0	1	-360	360;
	8422	5683	0.00052	0.00665	0	1940	0	0	0	0	1	-360	360;
	3260	541	0.0004	0.00314	0	307	0	0	0	0	1	-360	360;
	358	3032	0.01126	0.03616	0	0	0	0	0	0	1	-360	360;
	7642	4050	0.00209	0.0266	0	1767	0	0	0	0	1	-360	360;
	5683	5581	0.00105	0.0133	0	1801	0	0	0	0	1	-360	360;
	3292	1929	0.00189	0.02411	0	1767	0	0	0	0	1	-360	360;
	8110	7789	0	0.0002	0	1351	0	0	0	0	1	-360	360;
	4050	5581	0.00196	0.02494	0	1593	0	0	0	0	1	-360	360;
	5581	6746	0.00034	0.00432	0	1351	0	0	0	0	1	-360	360;
	5785	7668	0.00067	0.00682	0	0	0	0	0	0	1	-360	360;
	7668	1230	0.00022	0.0024	0	0	0	0	0	0	1	-360	360;
	7677	5415	0.00636	0.02043	0	0	0	0	0	0	1	-360	360;
	6677	8254	6.3e-05	0.00019	0	0	0	0	0	0	1	-360	360;
	5023	9089	0.00398	0.03138	0	0	0	0	0	0	1	-360	360;
	5023	9089	0.00398	0.03138	0	0	0	0	0	0	1	-360	360;
	3693	8460	0.00394	0.01266	0	0	0	0	0	0	1	-360	360;
	8441	8121	0.00011	0.00036	0	0	0	0	0	0	1	-360	360;
	6077	6929	0.00135	0.01379	0	1711	0	0	1.052632	0	1	-360	360;
	2935	4959	0.00097	0.01239	0	0	0	0	1.052632	0	1	-360	360;
	436	4458	0.00084	0.01064	0	0	0	0	1.052632	0	1	-360	360;
	6967	4635	0.00431	0.09324	0	0	0	0	0	0	1	-360	360;
	6967	4635	0.0044	0.091339	0	0	0	0	0	0	1	-360	360;
	1335	7131	0.006356	0.142878	0	0	0	0	1.026667	0	1	-360	360;
	2474	7141	0.00523	0.10214	0	0	0	0	0	0	1	-360	360;
	2474	7141	0.00532	0.099917	0	0	0	0	0	0	1	-360	360;
	2474	7141	0.00442	0.105388	0	0	0	0	0	0	1	-360	360;
	8258	5657	0.001002	0.059033	0	0	0	0	1.052632	0	1	-360	360;
	8258	5657	0.000812	0.056028	0	0	0	0	1.052632	0	1	-360	360;
	7313	7931	0.000921	0.058347	0	0	0	0	1.052632	0	1	-360	360;
	7931	4322	0.0016	0.07243	0	0	0	0	0	0	1	-360	360;
	7931	4322	0.00182	0.07098	0	0	0	0	0	0	1	-360	360;
	7931	4322	0.00185	0.06953	0	0	0	0	0	0	1	-360	360;
	273	8339	0.00349	0.07969	0	0	0	0	0	0	1	-360	360;
	273	8339	0.00355	0.07799	0	0	0	0	0	0	1	-360	360;
	1544	273	0.001101	0.048618	0	0	0	0	1.052632	0	1	-360	360;
	1544	273	0.000984	0.055107	0	0	0	0	1.052632	0	1	-360	360;
	6505	8826	0.00175	0.07752	0	0	0	0	0	0	1	-360	360;
	6505	8826	0.00171	0.07901	0	0	0	0	0	0	1	-360	360;
	5976	4329	0.00302	0.091446	0	0	0	0	0	0	1	-360	360;
	5976	4329	0.00295	0.084678	0	0	0	0	0	0	1	-360	360;
	5989	4625	0.0018	0.07815	0	0	0	0	0	0	1	-360	360;
	5989	4625	0.00185	0.07724	0	0	0	0	0	0	1	-360	360;
	5989	4625	0.00193	0.07007	0	0	0	0	0	0	1	-360	360;
	106	2546	0.00205	0.094653	0	0	0	0	1.052632	0	1	-360	360;
	5177	515	0.000454	0.031593	0	691	0	0	0.976831	0.055998	1	-360	360;
	5177	515	0.0004	0.035809	0	691	0	0	0.976831	0.055998	1	-360	360;
	2203	7638	0.000342	0.033341	0	724	0	0	0.905856	0	1	-360	360;
	1363	7638	0.000342	0.03398	0	724	0	0	0.905856	0	1	-360	360;
	4463	7638	0.000121	0.028823	0	705	0	0	0.953802	0.305147	1	-360	360;
	4463	3826	0.000121	0.014044	0	705	0	0	0	0	1	-360	360;
	6962	1829	0.000347	0.044669	0	691	0	0	1.098173	-0.129803	1	-360	360;
	6962	1829	0.000356	0.043761	0	691	0	0	1.098173	-0.129803	1	-360	360;
	4310	667	0	0.026535	0	0	0	0	0	0.047371	1	-360	360;
	4310	1829	0	0.025962	0	0	0	0	0	-0.047371	1	-360	360;
	8228	5594	0.000431	0.041394	0	724	0	0	0.950048	0.115349	1	-360	360;
	8228	5594	0.00038	0.043903	0	724	0	0	0.955131	0.103215	1	-360	360;
	3508	5847	0.0004	0.03651	0	691	0	0	0.976831	0.055998	1	-360	360;
	3508	5847	0.0004	0.037209	0	691	0	0	0.976831	0.055998	1	-360	360;
	4115	5847	0.000121	0.015661	0	838	0	0	0	0	1	-360	360;
	4115	3702	0.000121	0.014499	0	838	0	0	0	0	1	-360	360;
	345	298	0.000131	0.014799	0	800	0	0	0.998827	-0.048422	1	-360	360;
	345	6762	0.000141	0.013923	0	800	0	0	0	0	1	-360	360;
	3025	3887	0.000194	0.013059	0	1185	0	0	1.002506	0	1	-360	360;
	7992	265	0.000463	0.031001	0	757	0	0	0.956361	0	1	-360	360;
	7992	265	0.000472	0.030331	0	757	0	0	0.956361	0	1	-360	360;
	458	7393	0.000163	0.015501	0	1415	0	0	0.943535	0	1	-360	360;
	5718	4010	0.001987	0.071722	0	0	0	0	0.9625	0	1	-360	360;
	2512	6346	0.000756	0.051399	0	0	0	0	0.995296	0	1	-360	360;
	2512	6346	0.000746	0.05345	0	0	0	0	0.995296	0	1	-360	360;
	7262	9177	0.00069	0.04333	0	0	0	0	1.052632	0	1	-360	360;
	2738	7390	0.0022	0.065399	0	0	0	0	0.925	0	1	-360	360;
	2738	7390	0.002235	0.064019	0	0	0	0	0.925	0	1	-360	360;
	8612	5070	0.00168	0.07424	0	0	0	0	0	0	1	-360	360;
	963	3467	0.002036	0.070048	0	0	0	0	0.992391	0	1	-360	360;
	963	3467	0.001997	0.071368	0	0	0	0	0.992391	0	1	-360	360;
	3622	963	0.00035	0.032859	0	0	0	0	1.052632	0	1	-360	360;
	3622	963	0.0004	0.0322	0	0	0	0	1.052632	0	1	-360	360;
	2613	3801	0.00205	0.06915	0	0	0	0	0	0	1	-360	360;
	2613	3801	0.00082	0.07342	0	0	0	0	0	0	1	-360	360;
	353	8577	0.002658	0.07732	0	0	0	0	0.992391	0	1	-360	360;
	353	8577	0.002712	0.075649	0	0	0	0	0.992391	0	1	-360	360;
	4022	3856	0.00065	0.04604	0	0	0	0	1.052632	0	1	-360	360;
	7105	2625	0.00062	0.03629	0	0	0	0	1.052632	0	1	-360	360;
	7105	8172	0.00019	0.03263	0	0	0	0	1.052632	0	1	-360	360;
	2682	974	0.001797	0.07284	0	0	0	0	0.992248	0	1	-360	360;
	3236	8398	0.002712	0.072578	0	0	0	0	0.966477	0	1	-360	360;
	7285	353	0.00047	0.0307	0	0	0	0	1.052632	0	1	-360	360;
	7695	3936	0.002601	0.071481	0	0	0	0	0.975	0	1	-360	360;
	7695	3936	0.002653	0.06993	0	0	0	0	0.975	0	1	-360	360;
	5625	7695	0.00037	0.034169	0	0	0	0	1.052632	0	1	-360	360;
	5625	7695	0.00037	0.034831	0	0	0	0	1.052632	0	1	-360	360;
	6688	3790	0.0004	0.03524	0	0	0	0	1.052632	0	1	-360	360;
	6688	5196	0.00069	0.04525	0	0	0	0	1.052632	0	1	-360	360;
	8894	7768	0.00069	0.04231	0	0	0	0	1.052632	0	1	-360	360;
	3098	7531	0.00119	0.073146	0	0	0	0	0.9625	0	1	-360	360;
	3098	7531	0.002764	0.071607	0	0	0	0	0.9625	0	1	-360	360;
	9024	6542	0.000742	0.067219	0	0	0	0	0.994359	0	1	-360	360;
	9024	6542	0.000793	0.06388	0	0	0	0	0.998418	0	1	-360	360;
	6069	9192	0.000609	0.046809	0	691	0	0	0.933053	0	1	-360	360;
	6069	1968	0.00059	0.04521	0	691	0	0	0.91223	0	1	-360	360;
	7829	1968	0.519194	5.108058	0	0	0	0	0	0	1	-360	360;
	8574	2299	0.000559	0.039982	0	0	0	0	0.919523	0	1	-360	360;
	8574	5587	0.000537	0.045312	0	0	0	0	0.948262	0	1	-360	360;
	4929	1815	0.000889	0.061549	0	0	0	0	0.996867	0	1	-360	360;
	1815	792	1.467822	7.764888	0	0	0	0	0	0	1	-360	360;
	6704	4586	0.000586	0.049031	0	658	0	0	0.923407	0	1	-360	360;
	2441	3506	0.000417	0.047801	0	658	0	0	0.952413	0	1	-360	360;
	1367	6826	0.000755	0.05143	0	0	0	0	0.924009	0	1	-360	360;
	1676	228	0.000733	0.052399	0	658	0	0	0.934986	0	1	-360	360;
	7829	6826	0.00078	0.064311	0	0	0	0	0	0	1	-360	360;
	659	3279	0.00089	0.048632	0	0	0	0	0.987002	0	1	-360	360;
	5210	792	0.00855	0.08972	0	0	0	0	0	0	1	-360	360;
	1445	792	0.00223	0.04836	0	0	0	0	0	0	1	-360	360;
	7180	2168	0.000581	0.043111	0	658	0	0	0.924038	0	1	-360	360;
	6833	8964	0.000579	0.043272	0	691	0	0	0.901005	0	1	-360	360;
	3659	3242	0.000586	0.049031	0	691	0	0	0.923407	0	1	-360	360;
	8181	4427	0.000559	0.049192	0	0	0	0	0.924009	0	1	-360	360;
	9025	7563	0.000573	0.051129	0	658	0	0	0.934986	0	1	-360	360;
	2267	5210	0.000417	0.039229	0	724	0	0	0.967059	0	1	-360	360;
	5210	7279	0.230542	1.197733	0	0	0	0	0	0	1	-360	360;
	5210	8335	0.02284	0.2644	0	0	0	0	0	0	1	-360	360;
	5210	2449	0.08537	0.736798	0	0	0	0	0	0	1	-360	360;
	5210	4665	1.402768	3.640289	0	0	0	0	0	0	1	-360	360;
	5210	4495	1.050227	4.450207	0	0	0	0	0	0	1	-360	360;
	8179	7279	0.000716	0.04371	0	0	0	0	0.934986	0	1	-360	360;
	1445	7279	0.518622	2.317689	0	0	0	0	0	0	1	-360	360;
	5509	8335	0.00059	0.045662	0	0	0	0	0.955958	0	1	-360	360;
	5509	1531	0.000603	0.050959	0	0	0	0	0.968125	0	1	-360	360;
	1445	8335	0.047769	0.554178	0	0	0	0	0	0	1	-360	360;
	7762	5155	0.000591	0.045153	0	0	0	0	0.918868	0	1	-360	360;
	7762	271	0.0006	0.047778	0	691	0	0	0.948976	0	1	-360	360;
	7829	5776	0.02436	0.270151	0	0	0	0	0	0	1	-360	360;
	8329	1445	0.000802	0.05727	0	0	0	0	0.967021	0	1	-360	360;
	8329	1445	0.000823	0.056059	0	0	0	0	0.967021	0	1	-360	360;
	1445	2449	0.188099	1.291901	0	0	0	0	0	0	1	-360	360;
	1445	4665	2.466323	7.26343	0	0	0	0	0	0	1	-360	360;
	2268	2908	0.000523	0.050352	0	724	0	0	0.934622	0	1	-360	360;
	8847	5776	0.000682	0.049638	0	0	0	0	0.913034	0	1	-360	360;
	8847	8103	0.00099	0.04736	0	0	0	0	1.012146	0	1	-360	360;
	317	2449	0.000616	0.048759	0	658	0	0	0.933053	0	1	-360	360;
	7637	8581	9e-05	0.015499	0	1678	0	0	0	-0.428189	1	-360	360;
	5848	7526	9e-05	0.00963	0	1678	0	0	0	0.178581	1	-360	360;
	89	4495	0.000664	0.045609	0	0	0	0	0.924009	0	1	-360	360;
	2520	4665	0.000649	0.044862	0	691	0	0	0.933053	0	1	-360	360;
	5996	8420	0.00067	0.04326	0	724	0	0	0.923407	0	1	-360	360;
	2154	5996	9e-05	0.009901	0	1678	0	0	0	-0.153178	1	-360	360;
	6999	4008	0.000246	0.02678	0	1382	0	0	1.015668	-0.130098	1	-360	360;
	4008	5521	0.00131	0.07578	0	0	0	0	1.045454	0	1	-360	360;
	4008	5521	0.00128	0.07077	0	0	0	0	1.045454	0	1	-360	360;
	3173	8768	0.00131	0.062651	0	0	0	0	0	0	1	-360	360;
	3173	8768	0.00133	0.0577	0	0	0	0	0	0	1	-360	360;
	5055	1450	0.001495	0.059475	0	0	0	0	0.974576	0	1	-360	360;
	5055	9237	0.001453	0.061392	0	0	0	0	0.974576	0	1	-360	360;
	6568	5424	0.001475	0.074554	0	0	0	0	0.981267	0	1	-360	360;
	6568	5424	0.001309	0.084508	0	0	0	0	0.981267	0	1	-360	360;
	888	4877	0.00119	0.084099	0	0	0	0	0	0	1	-360	360;
	4388	1073	0.00117	0.064859	0	0	0	0	0	0	1	-360	360;
	4388	1073	0.00114	0.066091	0	0	0	0	0	0	1	-360	360;
	7572	5954	0.001225	0.083401	0	0	0	0	0.973021	0	1	-360	360;
	7572	5954	0.001394	0.082574	0	0	0	0	0.973021	0	1	-360	360;
	3440	4599	0.001415	0.080475	0	0	0	0	0.973021	0	1	-360	360;
	3440	4599	0.001447	0.078793	0	0	0	0	0.973021	0	1	-360	360;
	9010	8830	0.001444	0.074683	0	0	0	0	0.981267	0	1	-360	360;
	9010	8830	0.001475	0.073051	0	0	0	0	0.981267	0	1	-360	360;
	921	4570	0.001316	0.070415	0	0	0	0	0.974576	0	1	-360	360;
	7790	3437	0.00128	0.065919	0	0	0	0	0	0	1	-360	360;
	8885	480	0.001056	0.075919	0	0	0	0	1.052632	0	1	-360	360;
	1263	480	0.001451	0.078431	0	0	0	0	0.905512	0	1	-360	360;
	1263	480	0.001415	0.075297	0	0	0	0	0.905512	0	1	-360	360;
	2677	1682	0.001411	0.061844	0	0	0	0	0.974576	0	1	-360	360;
	2677	1682	0.001432	0.058938	0	0	0	0	0.974576	0	1	-360	360;
	2386	2068	0.00229	0.085231	0	0	0	0	0.979339	0	1	-360	360;
	5160	572	0.001474	0.06058	0	0	0	0	0.974576	0	1	-360	360;
	1438	1324	0.00125	0.064651	0	0	0	0	0	0	1	-360	360;
	1438	1324	0.00123	0.065919	0	0	0	0	0	0	1	-360	360;
	1438	1324	0.0012	0.06718	0	0	0	0	0	0	1	-360	360;
	2419	7336	0.001221	0.069631	0	0	0	0	0.982906	0	1	-360	360;
	2419	7336	0.00119	0.065604	0	0	0	0	0.982906	0	1	-360	360;
	1668	9145	0.001039	0.045925	0	0	0	0	0.981267	0	1	-360	360;
	3885	9144	0.000406	0.019067	0	1316	0	0	0.79272	0.154768	1	-360	360;
	9144	1873	0.001563	0.060356	0	0	0	0	0.982906	0	1	-360	360;
	9144	1873	0.001459	0.059041	0	0	0	0	0.982906	0	1	-360	360;
	6046	1098	0.001531	0.068989	0	0	0	0	0.942623	0	1	-360	360;
	8322	6434	0.000318	0.014328	0	954	0	0	1.002506	0	1	-360	360;
	2041	6006	0.000246	0.02678	0	954	0	0	1.015668	-0.130098	1	-360	360;
	6006	3669	0.00121	0.066091	0	0	0	0	0	0	1	-360	360;
	6006	3669	0.00118	0.062289	0	0	0	0	0	0	1	-360	360;
	8551	6258	0.001646	0.071174	0	0	0	0	0.905512	0	1	-360	360;
	8551	6258	0.00229	0.087099	0	0	0	0	1.045454	0	1	-360	360;
	1922	1084	0.001421	0.061107	0	0	0	0	0.974576	0	1	-360	360;
	9194	3141	0.001573	0.080941	0	0	0	0	0.956938	0	1	-360	360;
	9194	3141	0.001594	0.078385	0	0	0	0	0.956938	0	1	-360	360;
	5497	8408	0.001369	0.06454	0	0	0	0	0.974576	0	1	-360	360;
	5497	8408	0.001348	0.065203	0	0	0	0	0.974576	0	1	-360	360;
	1143	5082	0.001295	0.071353	0	0	0	0	0.974576	0	1	-360	360;
	1143	5082	0.001295	0.070826	0	0	0	0	0.974576	0	1	-360	360;
	1506	4304	0.001441	0.084934	0	0	0	0	0.956938	0	1	-360	360;
	1506	4304	0.001463	0.083626	0	0	0	0	0.956938	0	1	-360	360;
	3981	36	0.000448	0.016447	0	724	0	0	1.002506	0	1	-360	360;
	3981	1641	0.001134	0.053193	0	0	0	0	0.998185	0	1	-360	360;
	36	1641	0.00136	0.057539	0	0	0	0	0	0	1	-360	360;
	2110	4184	0.001432	0.06454	0	0	0	0	0.974576	0	1	-360	360;
	665	4028	0.00195	0.114628	0	0	0	0	0.979339	0	1	-360	360;
	1820	7456	0.001295	0.067067	0	0	0	0	0.974576	0	1	-360	360;
	4788	8978	0.001242	0.074563	0	0	0	0	0.974576	0	1	-360	360;
	6158	4518	0.00126	0.060099	0	0	0	0	0	0	1	-360	360;
	6158	4518	0.00135	0.06051	0	0	0	0	0	0	1	-360	360;
	7455	3505	0.00036	0.042469	0	691	0	0	1.052174	0	1	-360	360;
	1199	5962	0.001278	0.05335	0	0	0	0	0.973021	0	1	-360	360;
	6775	5438	0.000503	0.028047	0	0	0	0	0.977126	0.026068	1	-360	360;
	6775	5438	0.000503	0.039314	0	0	0	0	0.977126	-0.112848	1	-360	360;
	3433	4540	0.000333	0.021103	0	691	0	0	0.99559	0	1	-360	360;
	8905	4631	0.00044	0.031706	0	0	0	0	0.977126	0.026068	1	-360	360;
	3494	5807	0.000164	0.013858	0	1185	0	0	1.019598	0	1	-360	360;
	4938	2987	0.000353	0.025975	0	921	0	0	0.937567	0	1	-360	360;
	1557	3569	0.000159	0.016667	0	1185	0	0	0.937567	0	1	-360	360;
	3277	2487	0.000273	0.017971	0	987	0	0	0.956938	0.066854	1	-360	360;
	1369	6520	0.000289	0.02558	0	691	0	0	1.030079	-0.153295	1	-360	360;
	5778	6433	0.00142	0.02765	0	476	0	0	0	0	1	-360	360;
	5778	6433	0.00145	0.02707	0	476	0	0	0	0	1	-360	360;
	1679	381	0.000469	0.030842	0	0	0	0	1.097063	0	1	-360	360;
	1615	381	0.00435	0.076291	0	468	0	0	0	0.432244	1	-360	360;
	4201	7219	0.000352	0.020852	0	691	0	0	1.011235	0	1	-360	360;
	4079	153	0.0003	0.01099	0	0	0	0	0	0	1	-360	360;
	4079	153	0.00029	0.0112	0	0	0	0	0	0	1	-360	360;
	4830	4176	0.000375	0.018768	0	691	0	0	1.011235	0	1	-360	360;
	6666	2191	0.000537	0.028553	0	691	0	0	1.013158	0	1	-360	360;
	6666	2191	0.000537	0.028002	0	691	0	0	1.013158	0	1	-360	360;
	1653	3973	0.000622	0.033296	0	0	0	0	0.956938	0	1	-360	360;
	1653	3973	0.000622	0.034627	0	0	0	0	0.956938	0.076921	1	-360	360;
	2716	382	0.00072	0.026242	0	0	0	0	0.957131	0	1	-360	360;
	2716	382	0.000644	0.029735	0	0	0	0	0.957131	0	1	-360	360;
	6375	4762	0.00314	0.168628	0	0	0	0	0	0	1	-360	360;
	6375	4762	0.00307	0.171877	0	0	0	0	0	0	1	-360	360;
	611	5905	0.000188	0.016506	0	954	0	0	0.979483	0	1	-360	360;
	2153	6014	0.000259	0.027074	0	0	0	0	1.002506	0	1	-360	360;
	3290	291	0.000437	0.02611	0	0	0	0	1.047869	0	1	-360	360;
	5343	3340	0.000446	0.025573	0	0	0	0	1.047869	0	1	-360	360;
	1884	2588	0.00032	0.02243	0	0	0	0	1.034211	0	1	-360	360;
	6848	6234	6.091942	8.676029	0	0	0	0	0	0	1	-360	360;
	5523	6234	6.444008	8.676035	0	0	0	0	0	0	1	-360	360;
	2891	6234	23.9876	40.776864	0	0	0	0	0	0	1	-360	360;
	4737	6234	35.24587	57.202482	0	0	0	0	0	0	1	-360	360;
	63	6234	45.97521	69.597102	0	0	0	0	0	0	1	-360	360;
	7082	6234	0.081469	0.775971	0	0	0	0	0	0	1	-360	360;
	8336	6234	32.03719	50.568181	0	0	0	0	0	0	1	-360	360;
	1268	6234	0.628884	2.485744	0	0	0	0	0	0	1	-360	360;
	6848	7410	12.31426	19.110329	0	0	0	0	0	0	1	-360	360;
	5523	7410	13.03616	19.128307	0	0	0	0	0	0	1	-360	360;
	7082	7410	0.197581	1.393719	0	0	0	0	0	0	1	-360	360;
	1268	7410	1.415165	4.81467	0	0	0	0	0	0	1	-360	360;
	5401	2743	0.13108	0.821607	0	0	0	0	0	0	1	-360	360;
	6284	2743	0.129079	0.839681	0	0	0	0	0	0	1	-360	360;
	6272	2631	0.069591	0.599051	0	0	0	0	0	0	1	-360	360;
	6272	1144	0.04983	0.471267	0	0	0	0	0	0	1	-360	360;
	6272	3332	0.00502	0.310928	0	0	0	0	0	0	1	-360	360;
	92	7786	5.68843	10.297518	0	0	0	0	0	0	1	-360	360;
	4737	7786	48.32851	56.749998	0	0	0	0	0	0	1	-360	360;
	9211	7786	23.27066	28.824377	0	0	0	0	0	0	1	-360	360;
	2724	7786	50.96074	55.43596	0	0	0	0	0	0	1	-360	360;
	92	6212	5.967355	9.180988	0	0	0	0	0	0	1	-360	360;
	4737	6212	44.27686	58.522723	0	0	0	0	0	0	1	-360	360;
	9211	6212	20.49132	30.954553	0	0	0	0	0	0	1	-360	360;
	2724	6212	43.12397	61.993787	0	0	0	0	0	0	1	-360	360;
	8851	4012	-0.04093	0.365413	0	0	0	0	0	0	1	-360	360;
	8851	1029	0.140533	1.138574	0	0	0	0	0	0	1	-360	360;
	8851	3431	0.151053	1.206163	0	0	0	0	0	0	1	-360	360;
	8851	4033	3.496607	8.23892	0	0	0	0	0	0	1	-360	360;
	8851	1238	3.21482	7.345567	0	0	0	0	0	0	1	-360	360;
	8851	438	0.136738	1.1482	0	0	0	0	0	0	1	-360	360;
	8851	1907	0.138539	1.094044	0	0	0	0	0	0	1	-360	360;
	8851	3330	0.077251	0.917036	0	0	0	0	0	0	1	-360	360;
	8851	1587	0.954778	3.73151	0	0	0	0	0	0	1	-360	360;
	8851	8938	0.745845	4.144945	0	0	0	0	0	0	1	-360	360;
	8851	1177	0.723892	4.195776	0	0	0	0	0	0	1	-360	360;
	8553	4012	0.47093	3.07252	0	0	0	0	0	0	1	-360	360;
	192	4012	0.45124	2.967768	0	0	0	0	0	0	1	-360	360;
	3349	4012	0.460083	2.907231	0	0	0	0	0	0	1	-360	360;
	5495	4012	2.286777	7.419422	0	0	0	0	0	0	1	-360	360;
	6894	4012	3.438636	9.067975	0	0	0	0	0	0	1	-360	360;
	9100	4012	0.885578	3.747314	0	0	0	0	0	0	1	-360	360;
	5717	4012	0.788967	4.246901	0	0	0	0	0	0	1	-360	360;
	6222	4012	1.804731	7.772521	0	0	0	0	0	0	1	-360	360;
	7035	6171	0.000716	0.032505	0	0	0	0	0.963948	0	1	-360	360;
	7035	6171	0.00056	0.04105	0	0	0	0	0.965484	0	1	-360	360;
	7035	6171	0.00057	0.040211	0	0	0	0	0.965484	0	1	-360	360;
	7502	6171	10.24442	15.278311	0	0	0	0	0	0	1	-360	360;
	4101	6171	10.4376	14.953303	0	0	0	0	0	0	1	-360	360;
	2511	6171	31.55372	31.927684	0	0	0	0	0	0	1	-360	360;
	1522	6171	0.02494	0.501983	0	0	0	0	0	0	1	-360	360;
	4866	6171	2.458264	5.693595	0	0	0	0	0	0	1	-360	360;
	6724	6171	21.7686	29.266527	0	0	0	0	0	0	1	-360	360;
	8429	6171	0.260765	1.554236	0	0	0	0	0	0	1	-360	360;
	7816	6171	0.41095	2.453926	0	0	0	0	0	0	1	-360	360;
	5428	6171	1.358636	3.804546	0	0	0	0	0	0	1	-360	360;
	8762	8498	0.00445	0.11669	0	0	0	0	0	0	1	-360	360;
	2306	182	0.02812	0.471488	0	0	0	0	0	0	1	-360	360;
	1667	182	0.109781	0.670062	0	0	0	0	0	0	1	-360	360;
	3514	182	0.11731	0.676136	0	0	0	0	0	0	1	-360	360;
	7102	7711	0.000443	0.034282	0	724	0	0	0.957287	0	1	-360	360;
	4645	7711	0.000407	0.0336	0	724	0	0	0.957287	0	1	-360	360;
	92	7711	0.05005	0.961922	0	0	0	0	0	0	1	-360	360;
	6848	7711	20.2438	55.378103	0	0	0	0	0	0	1	-360	360;
	5523	7711	20.42293	51.698341	0	0	0	0	0	0	1	-360	360;
	2891	7711	7.97376	20.509093	0	0	0	0	0	0	1	-360	360;
	4737	7711	1.402252	5.36219	0	0	0	0	0	0	1	-360	360;
	63	7711	4.842562	10.371072	0	0	0	0	0	0	1	-360	360;
	9211	7711	0.593884	2.602686	0	0	0	0	0	0	1	-360	360;
	8336	7711	9.390701	21.464872	0	0	0	0	0	0	1	-360	360;
	2724	7711	1.334504	5.801447	0	0	0	0	0	0	1	-360	360;
	8553	1029	1.754986	5.357133	0	0	0	0	0	0	1	-360	360;
	8553	3431	1.766898	5.923268	0	0	0	0	0	0	1	-360	360;
	8553	4033	1.832964	7.959834	0	0	0	0	0	0	1	-360	360;
	8553	1238	1.611565	6.707825	0	0	0	0	0	0	1	-360	360;
	8553	438	-0.019578	0.602618	0	0	0	0	0	0	1	-360	360;
	8553	1907	-0.018109	0.575138	0	0	0	0	0	0	1	-360	360;
	8553	3330	2.305263	5.580056	0	0	0	0	0	0	1	-360	360;
	8553	1587	0.324709	1.981648	0	0	0	0	0	0	1	-360	360;
	8553	8938	11.49377	18.997923	0	0	0	0	0	0	1	-360	360;
	8553	1177	10.16551	21.393357	0	0	0	0	0	0	1	-360	360;
	8553	3860	0.00083	0.06663	0	0	0	0	1.054155	0	1	-360	360;
	6584	5323	0.220599	1.426839	0	0	0	0	0	0	1	-360	360;
	9037	5323	0.215413	1.319979	0	0	0	0	0	0	1	-360	360;
	6497	5323	0.07932	0.491198	0	0	0	0	0	0	1	-360	360;
	4766	5323	0.05112	0.388202	0	0	0	0	0	0	1	-360	360;
	4296	5323	7.957025	11.110533	0	0	0	0	0	0	1	-360	360;
	5611	5323	8.10723	10.874171	0	0	0	0	0	0	1	-360	360;
	4647	5323	0.19505	0.651797	0	0	0	0	0	0	1	-360	360;
	7483	5323	0.273698	1.122666	0	0	0	0	0	0	1	-360	360;
	6584	6715	-0.016694	0.279194	0	0	0	0	0	0	1	-360	360;
	6584	6715	0.00066	0.04694	0	0	0	0	1.101233	0	1	-360	360;
	6584	5301	0.262431	2.115858	0	0	0	0	0	0	1	-360	360;
	6584	2237	0.251911	2.002563	0	0	0	0	0	0	1	-360	360;
	6584	4350	0.086669	0.634127	0	0	0	0	0	0	1	-360	360;
	9037	6715	-0.018471	0.253657	0	0	0	0	0	0	1	-360	360;
	9037	6715	0.000298	0.012803	0	1547	0	0	1.110134	0	1	-360	360;
	9037	5301	0.307389	1.800831	0	0	0	0	0	0	1	-360	360;
	9037	2237	0.307181	1.800693	0	0	0	0	0	0	1	-360	360;
	9037	4350	0.083082	0.659398	0	0	0	0	0	0	1	-360	360;
	6497	6715	2.852479	6.564462	0	0	0	0	0	0	1	-360	360;
	4766	6715	0.610723	2.391322	0	0	0	0	0	0	1	-360	360;
	4296	6715	0.796095	2.828719	0	0	0	0	0	0	1	-360	360;
	5611	6715	0.778781	2.619215	0	0	0	0	0	0	1	-360	360;
	4647	6715	5.253512	8.690912	0	0	0	0	0	0	1	-360	360;
	7483	6715	0.14057	0.924359	0	0	0	0	0	0	1	-360	360;
	424	5439	0.00012	0.008723	0	0	0	0	0	-0.157232	1	-360	360;
	6497	2222	0.00012	0.008273	0	0	0	0	0	0.13111	1	-360	360;
	6497	5301	34.83864	37.361504	0	0	0	0	0	0	1	-360	360;
	6497	2237	31.4626	43.434901	0	0	0	0	0	0	1	-360	360;
	6497	4350	9.837258	14.382266	0	0	0	0	0	0	1	-360	360;
	2016	8929	0.124359	1.222541	0	0	0	0	0	0	1	-360	360;
	8517	8929	0.121719	1.24562	0	0	0	0	0	0	1	-360	360;
	2016	7457	0.11651	1.13624	0	0	0	0	0	0	1	-360	360;
	8517	7457	0.13205	1.113513	0	0	0	0	0	0	1	-360	360;
	4334	818	0.000447	0.03387	0	954	0	0	0.988622	0	1	-360	360;
	4334	2743	0.188878	0.872161	0	0	0	0	0	0	1	-360	360;
	4334	7299	0.000399	0.010683	0	790	0	0	1.031067	0	1	-360	360;
	8421	2608	0.01902	0.177292	0	0	0	0	0	0	1	-360	360;
	8421	9077	0.079771	0.509993	0	0	0	0	0	0	1	-360	360;
	8421	7321	0.00919	0.137159	0	0	0	0	0	0	1	-360	360;
	92	6451	10.8518	37.942518	0	0	0	0	0	0	1	-360	360;
	92	5680	20.41343	54.805408	0	0	0	0	0	0	1	-360	360;
	92	2895	20.30263	51.706374	0	0	0	0	0	0	1	-360	360;
	92	8022	0.787812	4.341828	0	0	0	0	0	0	1	-360	360;
	92	260	0.590422	4.326316	0	0	0	0	0	0	1	-360	360;
	92	242	0.712119	2.758379	0	0	0	0	0	0	1	-360	360;
	92	1067	0.891274	4.745083	0	0	0	0	0	0	1	-360	360;
	92	3533	27.32618	53.325484	0	0	0	0	0	0	1	-360	360;
	92	4789	0.072382	1.414889	0	0	0	0	0	0	1	-360	360;
	92	7592	0.070083	1.331163	0	0	0	0	0	0	1	-360	360;
	92	6421	0.02163	0.743629	0	0	0	0	0	0	1	-360	360;
	92	1581	3.623961	20.259002	0	0	0	0	0	0	1	-360	360;
	92	1085	0.851039	4.324031	0	0	0	0	0	0	1	-360	360;
	92	3913	1.040443	4.789612	0	0	0	0	0	0	1	-360	360;
	4766	5301	8.633058	14.86384	0	0	0	0	0	0	1	-360	360;
	4296	5301	0.05238	0.529835	0	0	0	0	0	0	1	-360	360;
	5611	5301	0.05337	0.518554	0	0	0	0	0	0	1	-360	360;
	4647	5301	56.80165	51.109504	0	0	0	0	0	0	1	-360	360;
	7483	5301	2.452686	6.967768	0	0	0	0	0	0	1	-360	360;
	4766	2237	8.052273	16.496276	0	0	0	0	0	0	1	-360	360;
	4296	2237	0.04414	0.60938	0	0	0	0	0	0	1	-360	360;
	5611	2237	0.0432	0.620868	0	0	0	0	0	0	1	-360	360;
	4647	2237	47.16736	58.301648	0	0	0	0	0	0	1	-360	360;
	7483	2237	2.566529	6.851652	0	0	0	0	0	0	1	-360	360;
	6848	6451	21.60319	36.731304	0	0	0	0	0	0	1	-360	360;
	6848	5680	38.25416	49.701504	0	0	0	0	0	0	1	-360	360;
	6848	2895	39.65235	49.572723	0	0	0	0	0	0	1	-360	360;
	6848	8022	13.37881	27.533241	0	0	0	0	0	0	1	-360	360;
	6848	260	11.70014	34.83726	0	0	0	0	0	0	1	-360	360;
	6848	242	0.03578	2.06572	0	0	0	0	0	0	1	-360	360;
	6848	6718	44.92382	57.412053	0	0	0	0	0	0	1	-360	360;
	6848	4702	5.792867	13.966068	0	0	0	0	0	0	1	-360	360;
	6848	7500	5.68705	12.979224	0	0	0	0	0	0	1	-360	360;
	6848	1067	23.49307	56.524238	0	0	0	0	0	0	1	-360	360;
	6848	3533	0.817382	3.615513	0	0	0	0	0	0	1	-360	360;
	6848	4789	1.090651	5.780887	0	0	0	0	0	0	1	-360	360;
	6848	7592	1.04626	5.209279	0	0	0	0	0	0	1	-360	360;
	6848	5440	0.644758	1.803532	0	0	0	0	0	0	1	-360	360;
	6848	2605	0.580852	2.069945	0	0	0	0	0	0	1	-360	360;
	6848	6421	15.32341	41.617723	0	0	0	0	0	0	1	-360	360;
	6848	1971	37.75623	69.158594	0	0	0	0	0	0	1	-360	360;
	6848	1581	4.031925	12.920362	0	0	0	0	0	0	1	-360	360;
	5523	6451	19.41759	39.081025	0	0	0	0	0	0	1	-360	360;
	5523	5680	38.13227	52.87396	0	0	0	0	0	0	1	-360	360;
	5523	2895	39.55609	52.783939	0	0	0	0	0	0	1	-360	360;
	5523	8022	13.35596	29.188359	0	0	0	0	0	0	1	-360	360;
	5523	260	13.38227	31.89335	0	0	0	0	0	0	1	-360	360;
	5523	242	0.06002	1.808379	0	0	0	0	0	0	1	-360	360;
	5523	6718	48.46122	56.359411	0	0	0	0	0	0	1	-360	360;
	5523	4702	6.294598	13.758307	0	0	0	0	0	0	1	-360	360;
	5523	7500	6.185319	14.074795	0	0	0	0	0	0	1	-360	360;
	5523	1067	21.89474	63.064402	0	0	0	0	0	0	1	-360	360;
	5523	3533	0.750554	3.860318	0	0	0	0	0	0	1	-360	360;
	5523	4789	1.114889	6.088781	0	0	0	0	0	0	1	-360	360;
	5523	7592	1.069114	5.491413	0	0	0	0	0	0	1	-360	360;
	5523	5440	0.653109	1.929778	0	0	0	0	0	0	1	-360	360;
	5523	2605	0.672888	1.912604	0	0	0	0	0	0	1	-360	360;
	5523	6421	18.13504	36.319253	0	0	0	0	0	0	1	-360	360;
	5523	1971	40.86842	68.019379	0	0	0	0	0	0	1	-360	360;
	5523	1581	4.398892	12.729917	0	0	0	0	0	0	1	-360	360;
	5941	2743	0.0384	0.48786	0	0	0	0	0	0	1	-360	360;
	192	1029	0.17575	1.589628	0	0	0	0	0	0	1	-360	360;
	192	3431	0.16178	1.59079	0	0	0	0	0	0	1	-360	360;
	192	4033	22.34695	35.318561	0	0	0	0	0	0	1	-360	360;
	192	1238	20.61288	31.520779	0	0	0	0	0	0	1	-360	360;
	192	438	1.071953	3.96018	0	0	0	0	0	0	1	-360	360;
	192	1907	1.073546	3.773961	0	0	0	0	0	0	1	-360	360;
	192	3330	1.477701	3.981233	0	0	0	0	0	0	1	-360	360;
	192	1587	5.635665	13.348342	0	0	0	0	0	0	1	-360	360;
	192	8938	13.41828	27.828945	0	0	0	0	0	0	1	-360	360;
	192	1177	13.04432	28.183522	0	0	0	0	0	0	1	-360	360;
	3349	1029	0.17575	1.589628	0	0	0	0	0	0	1	-360	360;
	3349	3431	0.161789	1.590785	0	0	0	0	0	0	1	-360	360;
	3349	4033	22.34695	35.318561	0	0	0	0	0	0	1	-360	360;
	3349	1238	20.61288	31.520779	0	0	0	0	0	0	1	-360	360;
	3349	438	1.071953	3.96018	0	0	0	0	0	0	1	-360	360;
	3349	1907	1.073546	3.773961	0	0	0	0	0	0	1	-360	360;
	3349	3330	1.477701	3.981233	0	0	0	0	0	0	1	-360	360;
	3349	1587	5.635665	13.348342	0	0	0	0	0	0	1	-360	360;
	3349	8938	13.41828	27.828945	0	0	0	0	0	0	1	-360	360;
	3349	1177	13.04432	28.183522	0	0	0	0	0	0	1	-360	360;
	5495	1029	7.514876	17.076863	0	0	0	0	0	0	1	-360	360;
	6894	1029	10.18967	19.679343	0	0	0	0	0	0	1	-360	360;
	9100	1029	1.403533	4.891322	0	0	0	0	0	0	1	-360	360;
	5717	1029	1.431054	4.791528	0	0	0	0	0	0	1	-360	360;
	6222	1029	7.830992	14.143185	0	0	0	0	0	0	1	-360	360;
	5495	3431	9.209711	15.822724	0	0	0	0	0	0	1	-360	360;
	6894	3431	13.04525	19.283054	0	0	0	0	0	0	1	-360	360;
	9100	3431	1.367066	5.515496	0	0	0	0	0	0	1	-360	360;
	5717	3431	1.339153	5.62376	0	0	0	0	0	0	1	-360	360;
	6222	3431	7.23719	17.339673	0	0	0	0	0	0	1	-360	360;
	6554	1242	0.00696	0.20656	0	0	0	0	0	0	1	-360	360;
	2016	1242	0.085711	0.903781	0	0	0	0	0	0	1	-360	360;
	8517	1242	0.09713	0.885702	0	0	0	0	0	0	1	-360	360;
	4766	4350	2.771122	4.936011	0	0	0	0	0	0	1	-360	360;
	2016	7198	0.296032	1.447438	0	0	0	0	0	0	1	-360	360;
	2016	98	0.281413	1.320152	0	0	0	0	0	0	1	-360	360;
	8517	7198	0.307202	1.385873	0	0	0	0	0	0	1	-360	360;
	8517	98	0.25536	1.463643	0	0	0	0	0	0	1	-360	360;
	1079	7198	0.00046	0.03669	0	724	0	0	0.926928	0	1	-360	360;
	6847	98	0.000808	0.04562	0	0	0	0	0.937579	0	1	-360	360;
	6847	98	0.000438	0.03862	0	724	0	0	0.933083	0	1	-360	360;
	2306	2912	0.037081	0.480186	0	0	0	0	0	0	1	-360	360;
	1667	2912	0.074531	0.583946	0	0	0	0	0	0	1	-360	360;
	3514	2912	0.08	0.590186	0	0	0	0	0	0	1	-360	360;
	5102	2788	0.10579	0.434903	0	0	0	0	0	0	1	-360	360;
	5102	9077	0.09937	0.415727	0	0	0	0	0	0	1	-360	360;
	1811	2788	0.095997	0.482181	0	0	0	0	0	0	1	-360	360;
	1811	9077	0.086717	0.480388	0	0	0	0	0	0	1	-360	360;
	5495	4033	0.873905	3.992149	0	0	0	0	0	0	1	-360	360;
	6894	4033	0.120211	1.173533	0	0	0	0	0	0	1	-360	360;
	9100	4033	7.082025	15.591937	0	0	0	0	0	0	1	-360	360;
	5717	4033	7.220868	15.273755	0	0	0	0	0	0	1	-360	360;
	6222	4033	0.872831	3.28905	0	0	0	0	0	0	1	-360	360;
	5495	1238	0.921508	3.094834	0	0	0	0	0	0	1	-360	360;
	6894	1238	0.129919	0.960764	0	0	0	0	0	0	1	-360	360;
	9100	1238	6.142769	14.777069	0	0	0	0	0	0	1	-360	360;
	5717	1238	6.017355	15.066946	0	0	0	0	0	0	1	-360	360;
	6222	1238	0.694463	3.375206	0	0	0	0	0	0	1	-360	360;
	5495	438	7.254848	17.556091	0	0	0	0	0	0	1	-360	360;
	5495	1907	6.944598	15.819948	0	0	0	0	0	0	1	-360	360;
	5495	3330	0.962188	3.649169	0	0	0	0	0	0	1	-360	360;
	5495	1587	2.525	6.52029	0	0	0	0	0	0	1	-360	360;
	5495	8938	0.00154	0.39572	0	0	0	0	0	0	1	-360	360;
	5495	1177	0.0014	0.384758	0	0	0	0	0	0	1	-360	360;
	4296	4350	0.590782	1.406718	0	0	0	0	0	0	1	-360	360;
	5611	4350	0.526337	1.594252	0	0	0	0	0	0	1	-360	360;
	2891	6451	0.836859	5.618595	0	0	0	0	0	0	1	-360	360;
	4737	6451	1.237665	7.884711	0	0	0	0	0	0	1	-360	360;
	9211	6451	3.449174	20.571071	0	0	0	0	0	0	1	-360	360;
	7082	6451	5.302273	16.162393	0	0	0	0	0	0	1	-360	360;
	8336	6451	0.789525	4.620455	0	0	0	0	0	0	1	-360	360;
	1268	6451	2.372521	10.507644	0	0	0	0	0	0	1	-360	360;
	1493	2895	0.000495	0.033881	0	724	0	0	0.957287	0	1	-360	360;
	2891	5680	2.23905	7.164463	0	0	0	0	0	0	1	-360	360;
	4737	5680	3.327273	9.634503	0	0	0	0	0	0	1	-360	360;
	9211	5680	8.134091	27.90289	0	0	0	0	0	0	1	-360	360;
	7082	5680	10.55806	23.307847	0	0	0	0	0	0	1	-360	360;
	8336	5680	1.718843	7.201859	0	0	0	0	0	0	1	-360	360;
	1268	5680	4.571488	16.921905	0	0	0	0	0	0	1	-360	360;
	2891	2895	1.894649	7.933677	0	0	0	0	0	0	1	-360	360;
	4737	2895	3.133678	10.68802	0	0	0	0	0	0	1	-360	360;
	9211	2895	8.769008	26.754135	0	0	0	0	0	0	1	-360	360;
	7082	2895	11.85517	21.464873	0	0	0	0	0	0	1	-360	360;
	8336	2895	2.005517	6.368182	0	0	0	0	0	0	1	-360	360;
	1268	2895	5.553718	14.366735	0	0	0	0	0	0	1	-360	360;
	6894	438	4.540289	12.081814	0	0	0	0	0	0	1	-360	360;
	9100	438	0.14887	1.547789	0	0	0	0	0	0	1	-360	360;
	5717	438	0.145641	1.433141	0	0	0	0	0	0	1	-360	360;
	6222	438	7.255165	15.148962	0	0	0	0	0	0	1	-360	360;
	6894	1907	4.918802	10.649998	0	0	0	0	0	0	1	-360	360;
	9100	1907	0.170281	1.312108	0	0	0	0	0	0	1	-360	360;
	5717	1907	0.17349	1.284174	0	0	0	0	0	0	1	-360	360;
	6222	1907	7.658058	13.540087	0	0	0	0	0	0	1	-360	360;
	7093	3332	1.803947	6.514197	0	0	0	0	0	0	1	-360	360;
	7093	6130	1.414404	5.42687	0	0	0	0	0	0	1	-360	360;
	7093	1258	0.097091	1.338712	0	0	0	0	0	0	1	-360	360;
	7093	4085	0.069813	0.61338	0	0	0	0	0	0	1	-360	360;
	7207	3332	1.656925	6.386704	0	0	0	0	0	0	1	-360	360;
	7207	6130	1.503047	5.113989	0	0	0	0	0	0	1	-360	360;
	7207	1258	0.107382	1.212396	0	0	0	0	0	0	1	-360	360;
	7207	4085	0.080478	0.53394	0	0	0	0	0	0	1	-360	360;
	6894	3330	1.566157	4.554339	0	0	0	0	0	0	1	-360	360;
	9100	3330	3.21405	6.727273	0	0	0	0	0	0	1	-360	360;
	5717	3330	2.86343	7.62438	0	0	0	0	0	0	1	-360	360;
	6222	3330	0.784091	3.665289	0	0	0	0	0	0	1	-360	360;
	2891	8022	-0.013061	0.68473	0	0	0	0	0	0	1	-360	360;
	2891	260	0.02627	0.92126	0	0	0	0	0	0	1	-360	360;
	2891	242	5.225831	17.344184	0	0	0	0	0	0	1	-360	360;
	2891	629	37.01108	55.722988	0	0	0	0	0	0	1	-360	360;
	2891	3423	38.53255	55.73961	0	0	0	0	0	0	1	-360	360;
	2891	1067	0.00421	0.609432	0	0	0	0	0	0	1	-360	360;
	2891	3533	3.373338	12.718144	0	0	0	0	0	0	1	-360	360;
	2891	4789	5.841413	14.226459	0	0	0	0	0	0	1	-360	360;
	2891	7592	4.851247	14.847647	0	0	0	0	0	0	1	-360	360;
	2891	5440	15.98753	46.096945	0	0	0	0	0	0	1	-360	360;
	2891	2605	15.83449	47.580335	0	0	0	0	0	0	1	-360	360;
	2891	6421	5.167521	17.612876	0	0	0	0	0	0	1	-360	360;
	2891	1581	0.170402	3.131509	0	0	0	0	0	0	1	-360	360;
	2891	6655	13.13227	24.978533	0	0	0	0	0	0	1	-360	360;
	2891	6307	11.55748	21.738223	0	0	0	0	0	0	1	-360	360;
	4737	8022	0.071828	1.118006	0	0	0	0	0	0	1	-360	360;
	4737	260	0.01624	0.986219	0	0	0	0	0	0	1	-360	360;
	4737	242	1.180471	4.696329	0	0	0	0	0	0	1	-360	360;
	4737	1067	0.01296	0.834834	0	0	0	0	0	0	1	-360	360;
	4737	3533	4.962535	20.238919	0	0	0	0	0	0	1	-360	360;
	4737	4789	0.872438	4.821676	0	0	0	0	0	0	1	-360	360;
	4737	7592	0.801593	4.526524	0	0	0	0	0	0	1	-360	360;
	4737	5440	20.20014	50.806099	0	0	0	0	0	0	1	-360	360;
	4737	2605	23.16066	50.423823	0	0	0	0	0	0	1	-360	360;
	4737	6421	0.966621	4.164474	0	0	0	0	0	0	1	-360	360;
	4737	1581	0.448781	4.337397	0	0	0	0	0	0	1	-360	360;
	4737	1085	11.58657	23.771465	0	0	0	0	0	0	1	-360	360;
	4737	3913	13.00554	26.333102	0	0	0	0	0	0	1	-360	360;
	4737	6655	17.94044	35.608035	0	0	0	0	0	0	1	-360	360;
	4737	6307	15.18213	32.262467	0	0	0	0	0	0	1	-360	360;
	6894	1587	0.430249	2.466066	0	0	0	0	0	0	1	-360	360;
	6894	8938	0.0683	0.916828	0	0	0	0	0	0	1	-360	360;
	6894	1177	0.06612	0.843352	0	0	0	0	0	0	1	-360	360;
	9100	1587	1.444298	4.757438	0	0	0	0	0	0	1	-360	360;
	5717	1587	1.472603	4.660331	0	0	0	0	0	0	1	-360	360;
	6222	1587	2.284917	6.055578	0	0	0	0	0	0	1	-360	360;
	491	8934	0.000504	0.031781	0	0	0	0	1.058045	0	1	-360	360;
	491	3332	0.086683	1.502285	0	0	0	0	0	0	1	-360	360;
	491	6130	0.05572	1.389266	0	0	0	0	0	0	1	-360	360;
	491	1258	0.40606	1.910249	0	0	0	0	0	0	1	-360	360;
	491	4085	0.0265	0.438137	0	0	0	0	0	0	1	-360	360;
	7545	6173	0.000669	0.032368	0	0	0	0	1.067245	0	1	-360	360;
	9100	8938	16.48554	31.878093	0	0	0	0	0	0	1	-360	360;
	5717	8938	18.68347	31.239671	0	0	0	0	0	0	1	-360	360;
	6222	8938	-0.006984	0.380578	0	0	0	0	0	0	1	-360	360;
	9100	1177	19.27748	29.774787	0	0	0	0	0	0	1	-360	360;
	5717	1177	19.64132	29.142564	0	0	0	0	0	0	1	-360	360;
	6222	1177	-0.0075	0.354442	0	0	0	0	0	0	1	-360	360;
	324	3332	0.979504	4.164463	0	0	0	0	0	0	1	-360	360;
	5862	3332	0.241095	2.041488	0	0	0	0	0	0	1	-360	360;
	4052	3332	0.239525	2.084091	0	0	0	0	0	0	1	-360	360;
	324	6130	0.733492	3.60186	0	0	0	0	0	0	1	-360	360;
	5862	6130	0.176539	1.602479	0	0	0	0	0	0	1	-360	360;
	4052	6130	0.203141	1.572934	0	0	0	0	0	0	1	-360	360;
	63	8022	3.658471	11.068596	0	0	0	0	0	0	1	-360	360;
	9211	8022	0.10706	1.950682	0	0	0	0	0	0	1	-360	360;
	7082	8022	10.14339	24.185954	0	0	0	0	0	0	1	-360	360;
	8336	8022	0.07038	1.132707	0	0	0	0	0	0	1	-360	360;
	2724	8022	3.802686	12.010955	0	0	0	0	0	0	1	-360	360;
	1268	8022	3.583058	17.321278	0	0	0	0	0	0	1	-360	360;
	63	260	2.938636	13.075416	0	0	0	0	0	0	1	-360	360;
	9211	260	0.00024	2.372521	0	0	0	0	0	0	1	-360	360;
	7082	260	9.23781	30.750004	0	0	0	0	0	0	1	-360	360;
	8336	260	0.041469	1.048347	0	0	0	0	0	0	1	-360	360;
	2724	260	3.576446	11.947316	0	0	0	0	0	0	1	-360	360;
	1268	260	4.138017	18.151449	0	0	0	0	0	0	1	-360	360;
	324	2743	0.107147	0.58482	0	0	0	0	0	0	1	-360	360;
	324	1258	7.159972	13.527699	0	0	0	0	0	0	1	-360	360;
	324	4085	1.77874	5.115028	0	0	0	0	0	0	1	-360	360;
	7082	5909	7.195455	16.568181	0	0	0	0	0	0	1	-360	360;
	1268	5909	40.97313	63.811984	0	0	0	0	0	0	1	-360	360;
	7082	2828	7.423554	18.762807	0	0	0	0	0	0	1	-360	360;
	1268	2828	42.39256	65.737609	0	0	0	0	0	0	1	-360	360;
	63	242	-0.016674	0.278698	0	0	0	0	0	0	1	-360	360;
	63	1067	5.349861	16.166202	0	0	0	0	0	0	1	-360	360;
	63	3533	9.134349	25.279081	0	0	0	0	0	0	1	-360	360;
	63	4789	0.197001	1.168768	0	0	0	0	0	0	1	-360	360;
	63	7592	0.18955	1.05277	0	0	0	0	0	0	1	-360	360;
	63	5440	6.030263	16.006928	0	0	0	0	0	0	1	-360	360;
	63	2605	5.975346	16.528394	0	0	0	0	0	0	1	-360	360;
	63	6421	3.107895	9.123269	0	0	0	0	0	0	1	-360	360;
	63	1581	11.81025	37.443907	0	0	0	0	0	0	1	-360	360;
	63	1085	27.31856	49.591408	0	0	0	0	0	0	1	-360	360;
	63	3913	34.37812	55.136426	0	0	0	0	0	0	1	-360	360;
	9211	242	0.231281	1.502872	0	0	0	0	0	0	1	-360	360;
	8336	242	7.048967	19.24607	0	0	0	0	0	0	1	-360	360;
	2724	242	0.229277	1.235537	0	0	0	0	0	0	1	-360	360;
	1268	242	2.544835	7.532231	0	0	0	0	0	0	1	-360	360;
	7082	5596	35.47521	51.481403	0	0	0	0	0	0	1	-360	360;
	7082	4099	34.64256	52.336781	0	0	0	0	0	0	1	-360	360;
	4647	4350	15.75277	20.093496	0	0	0	0	0	0	1	-360	360;
	6538	1944	0.000283	0.012832	0	790	0	0	1.031067	0	1	-360	360;
	8333	1962	0.000292	0.01188	0	790	0	0	1.031067	0	1	-360	360;
	7135	629	0.000398	0.034511	0	691	0	0	0.982076	0	1	-360	360;
	8336	629	33.09298	47.6281	0	0	0	0	0	0	1	-360	360;
	8336	3423	34.44215	47.623971	0	0	0	0	0	0	1	-360	360;
	9211	1067	0.161039	2.815236	0	0	0	0	0	0	1	-360	360;
	9211	3533	9.489612	27.308865	0	0	0	0	0	0	1	-360	360;
	9211	4789	0.246808	1.61939	0	0	0	0	0	0	1	-360	360;
	9211	7592	0.228082	1.521121	0	0	0	0	0	0	1	-360	360;
	9211	5440	18.28047	42.772855	0	0	0	0	0	0	1	-360	360;
	9211	2605	18.10111	44.135045	0	0	0	0	0	0	1	-360	360;
	9211	6421	0.331281	2.194459	0	0	0	0	0	0	1	-360	360;
	9211	1581	0.752078	9.020083	0	0	0	0	0	0	1	-360	360;
	9211	1085	4.985803	12.659968	0	0	0	0	0	0	1	-360	360;
	9211	3913	5.581648	14.030467	0	0	0	0	0	0	1	-360	360;
	6485	4424	0.000476	0.03161	0	724	0	0	0.957287	0	1	-360	360;
	6485	7189	0.000457	0.030901	0	724	0	0	0.957287	0	1	-360	360;
	7082	8182	1.305041	5.872521	0	0	0	0	0	0	1	-360	360;
	1268	8182	12.81343	30.764465	0	0	0	0	0	0	1	-360	360;
	7082	6718	0.657355	3.229132	0	0	0	0	0	0	1	-360	360;
	1268	6718	6.467975	16.903509	0	0	0	0	0	0	1	-360	360;
	7082	4702	-0.017839	0.427271	0	0	0	0	0	0	1	-360	360;
	7082	7500	-0.020298	0.420457	0	0	0	0	0	0	1	-360	360;
	7082	3533	0.05596	1.32043	0	0	0	0	0	0	1	-360	360;
	7082	5440	0.430997	2.378602	0	0	0	0	0	0	1	-360	360;
	7082	2605	0.443573	2.357341	0	0	0	0	0	0	1	-360	360;
	7082	1971	0.073677	2.231025	0	0	0	0	0	0	1	-360	360;
	7082	1581	0.927909	4.907895	0	0	0	0	0	0	1	-360	360;
	7082	6655	42.82895	57.943891	0	0	0	0	0	0	1	-360	360;
	7082	6307	36.44598	52.531845	0	0	0	0	0	0	1	-360	360;
	4849	7500	0.00037	0.03817	0	724	0	0	0.982076	0	1	-360	360;
	4849	4702	0.000309	0.035711	0	724	0	0	0.982076	0	1	-360	360;
	1268	4702	0.529731	3.305372	0	0	0	0	0	0	1	-360	360;
	1268	7500	0.54188	3.249793	0	0	0	0	0	0	1	-360	360;
	8336	1067	-0.008017	0.233843	0	0	0	0	0	0	1	-360	360;
	8336	3533	5.680748	16.603181	0	0	0	0	0	0	1	-360	360;
	8336	4789	6.552285	16.597649	0	0	0	0	0	0	1	-360	360;
	8336	7592	5.444182	17.322713	0	0	0	0	0	0	1	-360	360;
	8336	5440	24.32756	59.210516	0	0	0	0	0	0	1	-360	360;
	8336	2605	24.09903	61.117023	0	0	0	0	0	0	1	-360	360;
	8336	6421	5.625068	20.064407	0	0	0	0	0	0	1	-360	360;
	8336	1581	0.456198	4.138435	0	0	0	0	0	0	1	-360	360;
	8336	6655	11.52632	21.784622	0	0	0	0	0	0	1	-360	360;
	8336	6307	10.14543	18.959839	0	0	0	0	0	0	1	-360	360;
	8336	1067	0.000476	0.031631	0	889	0	0	0.969682	0	1	-360	360;
	2724	1067	5.521694	14.730582	0	0	0	0	0	0	1	-360	360;
	1268	1067	19.34525	41.508264	0	0	0	0	0	0	1	-360	360;
	2724	3533	32.73967	68.105374	0	0	0	0	0	0	1	-360	360;
	1268	3533	-0.083244	0.946591	0	0	0	0	0	0	1	-360	360;
	4535	4789	0.000447	0.035271	0	724	0	0	0.957287	0	1	-360	360;
	2724	4789	-0.015103	0.337521	0	0	0	0	0	0	1	-360	360;
	2724	7592	-0.012686	0.286488	0	0	0	0	0	0	1	-360	360;
	2724	5440	27.97992	45.335871	0	0	0	0	0	0	1	-360	360;
	2724	2605	28.86773	44.978518	0	0	0	0	0	0	1	-360	360;
	2724	6421	0.994044	4.152008	0	0	0	0	0	0	1	-360	360;
	2724	1581	19.05609	41.630194	0	0	0	0	0	0	1	-360	360;
	2724	1085	11.95776	23.687676	0	0	0	0	0	0	1	-360	360;
	2724	3913	11.73892	30.398893	0	0	0	0	0	0	1	-360	360;
	1268	4789	10.39773	25.376037	0	0	0	0	0	0	1	-360	360;
	1268	7592	9.457438	23.791321	0	0	0	0	0	0	1	-360	360;
	1268	5440	0.000409	0.036341	0	724	0	0	0.907709	0	1	-360	360;
	1268	5440	-0.006942	0.494855	0	0	0	0	0	0	1	-360	360;
	1268	2605	-0.008161	0.491033	0	0	0	0	0	0	1	-360	360;
	1268	1971	4.701801	16.907201	0	0	0	0	0	0	1	-360	360;
	1268	1581	0.18313	2.973269	0	0	0	0	0	0	1	-360	360;
	1268	6655	30.6392	44.945983	0	0	0	0	0	0	1	-360	360;
	1268	6307	27.01524	39.081706	0	0	0	0	0	0	1	-360	360;
	6222	5142	0.000757	0.064121	0	0	0	0	1.042007	0	1	-360	360;
	5862	1258	2.158264	7.671693	0	0	0	0	0	0	1	-360	360;
	4052	1258	2.128512	7.829339	0	0	0	0	0	0	1	-360	360;
	5862	4085	0.536593	2.637604	0	0	0	0	0	0	1	-360	360;
	4052	4085	0.529487	2.445638	0	0	0	0	0	0	1	-360	360;
	7483	6978	0.00039	0.018483	0	856	0	0	0.945981	-0.090668	1	-360	360;
	7483	4350	0.856787	2.082133	0	0	0	0	0	0	1	-360	360;
	8193	2196	0.000515	0.03175	0	691	0	0	1.002506	0	1	-360	360;
	8193	2196	0.000408	0.036729	0	691	0	0	0.988471	0	1	-360	360;
	7650	3606	0.00648	0.148407	0	0	0	0	0	0	1	-360	360;
	2532	3606	0.00055	0.038349	0	0	0	0	0.977825	0	1	-360	360;
	1663	2883	0.114331	0.894649	0	0	0	0	0	0	1	-360	360;
	7966	2883	3.931818	5.960743	0	0	0	0	0	0	1	-360	360;
	5100	2883	1.022872	2.477479	0	0	0	0	0	0	1	-360	360;
	1588	2883	1.145744	2.668595	0	0	0	0	0	0	1	-360	360;
	2890	2883	0.115521	0.652645	0	0	0	0	0	0	1	-360	360;
	1663	6197	0.342335	1.036053	0	0	0	0	0	0	1	-360	360;
	6605	7815	0.08144	0.487603	0	0	0	0	0	0	1	-360	360;
	8855	56	0.492025	1.533306	0	0	0	0	0	0	1	-360	360;
	2420	5088	0.00079	0.058401	0	0	0	0	0	0	1	-360	360;
	2420	5088	0.00077	0.05951	0	0	0	0	0	0	1	-360	360;
	1663	3944	0.000209	0.016646	0	1185	0	0	0.947763	0	1	-360	360;
	1663	397	-0.020062	0.263393	0	0	0	0	0	0	1	-360	360;
	1663	4971	0.02611	0.389661	0	0	0	0	0	0	1	-360	360;
	1663	1120	0.02953	0.274197	0	0	0	0	0	0	1	-360	360;
	1663	397	0.000249	0.015313	0	1185	0	0	0.975409	0	1	-360	360;
	4637	1675	0.01135	0.389938	0	0	0	0	0	0	1	-360	360;
	4637	3539	0.03467	0.4241	0	0	0	0	0	0	1	-360	360;
	8855	1624	0.000159	0.018964	0	1185	0	0	0.962403	0	1	-360	360;
	8855	1624	-0.01126	0.214855	0	0	0	0	0	0	1	-360	360;
	8855	7127	0.20036	1.026385	0	0	0	0	0	0	1	-360	360;
	8855	9165	0.117071	0.511302	0	0	0	0	0	0	1	-360	360;
	3835	417	0.093303	0.945014	0	0	0	0	0	0	1	-360	360;
	4366	417	0.01297	0.327971	0	0	0	0	0	0	1	-360	360;
	2385	417	0.05929	0.601787	0	0	0	0	0	0	1	-360	360;
	6605	5325	0.05013	0.919979	0	0	0	0	0	0	1	-360	360;
	6605	4167	0.0477	0.865888	0	0	0	0	0	0	1	-360	360;
	7966	3537	0.749308	2.45464	0	0	0	0	0	0	1	-360	360;
	7966	5869	2.902839	5.007549	0	0	0	0	0	0	1	-360	360;
	2906	1675	0.03282	0.444566	0	0	0	0	0	0	1	-360	360;
	7920	417	0.186479	0.852913	0	0	0	0	0	0	1	-360	360;
	2080	417	0.185979	0.818905	0	0	0	0	0	0	1	-360	360;
	7424	417	1.207107	3.026033	0	0	0	0	0	0	1	-360	360;
	6193	417	0.414669	1.672355	0	0	0	0	0	0	1	-360	360;
	5100	3537	0.129612	0.934834	0	0	0	0	0	0	1	-360	360;
	5100	5869	0.796053	2.21392	0	0	0	0	0	0	1	-360	360;
	1588	3537	0.136461	1.06849	0	0	0	0	0	0	1	-360	360;
	1588	5869	0.838366	2.528809	0	0	0	0	0	0	1	-360	360;
	7617	1346	0.000298	0.029295	0	790	0	0	1.085859	-0.169703	1	-360	360;
	4634	5622	0.01604	0.374051	0	0	0	0	0	0	1	-360	360;
	2138	5622	0.01635	0.36626	0	0	0	0	0	0	1	-360	360;
	2906	5622	0.024141	0.327376	0	0	0	0	0	0	1	-360	360;
	6274	5622	0.02194	0.458409	0	0	0	0	0	0	1	-360	360;
	1994	8916	0.00046	0.03512	0	724	0	0	1.01926	0	1	-360	360;
	1994	6586	0.000328	0.017373	0	790	0	0	1.016586	0	1	-360	360;
	2906	3539	0.00056	0.05575	0	0	0	0	0	0	1	-360	360;
	6605	2663	0.09537	0.774091	0	0	0	0	0	0	1	-360	360;
	6605	4804	0.325393	1.201922	0	0	0	0	0	0	1	-360	360;
	6605	7602	0.367459	1.167149	0	0	0	0	0	0	1	-360	360;
	6274	2901	0.0008	0.04856	0	0	0	0	0	0	1	-360	360;
	7914	1627	0.00033	0.038031	0	724	0	0	1.027774	0	1	-360	360;
	7914	1627	0.00034	0.037211	0	724	0	0	1.027774	0	1	-360	360;
	2890	3537	-0.016405	0.249504	0	0	0	0	0	0	1	-360	360;
	2890	5869	0.087209	0.621302	0	0	0	0	0	0	1	-360	360;
	8725	5433	0.15538	0.761901	0	0	0	0	0	0	1	-360	360;
	6965	5433	0.15969	0.749463	0	0	0	0	0	0	1	-360	360;
	6906	5433	1.428926	3.608884	0	0	0	0	0	0	1	-360	360;
	5754	5433	0.861797	2.342768	0	0	0	0	0	0	1	-360	360;
	6227	2903	0.119958	0.657978	0	0	0	0	0	0	1	-360	360;
	3784	6418	0.172789	1.348017	0	0	0	0	0	0	1	-360	360;
	2409	6418	0.00088	0.381756	0	0	0	0	0	0	1	-360	360;
	4578	6418	0.0061	0.418244	0	0	0	0	0	0	1	-360	360;
	6524	7097	0.05995	0.76688	0	0	0	0	0	0	1	-360	360;
	5969	7097	0.0653	0.750992	0	0	0	0	0	0	1	-360	360;
	2429	7097	-0.003347	0.292562	0	0	0	0	0	0	1	-360	360;
	3116	7748	1.83593	3.666115	0	0	0	0	0	0	1	-360	360;
	9160	7748	0.052409	0.497211	0	0	0	0	0	0	1	-360	360;
	1148	8113	0.436797	1.36533	0	0	0	0	0	0	1	-360	360;
	7116	8113	0.094151	0.896736	0	0	0	0	0	0	1	-360	360;
	1400	7609	0.00721	0.237881	0	0	0	0	0	0	1	-360	360;
	1400	4793	0.140762	0.856856	0	0	0	0	0	0	1	-360	360;
	8187	7609	0.05419	0.633158	0	0	0	0	0	0	1	-360	360;
	8187	4793	0.0291	0.36903	0	0	0	0	0	0	1	-360	360;
	3149	2903	0.143641	0.91531	0	0	0	0	0	0	1	-360	360;
	1214	2903	0.025969	0.445888	0	0	0	0	0	0	1	-360	360;
	1253	2903	0.270372	0.785723	0	0	0	0	0	0	1	-360	360;
	5528	2903	0.05175	0.49312	0	0	0	0	0	0	1	-360	360;
	4786	2903	0.08601	0.600496	0	0	0	0	0	0	1	-360	360;
	6027	2903	0.531591	1.712665	0	0	0	0	0	0	1	-360	360;
	4294	2903	0.43593	1.756797	0	0	0	0	0	0	1	-360	360;
	5380	4151	0.01926	0.358289	0	0	0	0	0	0	1	-360	360;
	7399	3905	0.01795	0.317397	0	0	0	0	0	0	1	-360	360;
	3784	158	0.00761	0.279277	0	0	0	0	0	0	1	-360	360;
	2409	158	0.260062	1.120476	0	0	0	0	0	0	1	-360	360;
	4578	158	0.158651	0.983822	0	0	0	0	0	0	1	-360	360;
	3784	4593	0.00395	0.366116	0	0	0	0	0	0	1	-360	360;
	6083	4593	0.083409	0.606219	0	0	0	0	0	0	1	-360	360;
	2409	4593	0.195669	1.120144	0	0	0	0	0	0	1	-360	360;
	4578	4593	0.104969	1.02188	0	0	0	0	0	0	1	-360	360;
	3160	4593	0.042831	0.74374	0	0	0	0	0	0	1	-360	360;
	3116	8216	0.000242	0.032753	0	1185	0	0	1.062286	-0.114968	1	-360	360;
	3116	5779	-0.003203	0.14994	0	0	0	0	0	0	1	-360	360;
	3116	5779	0.000451	0.034969	0	724	0	0	1.031654	0	1	-360	360;
	3116	6997	0.113213	0.498338	0	0	0	0	0	0	1	-360	360;
	6470	4466	0.00055	0.0343	0	724	0	0	0.964912	0	1	-360	360;
	6965	8955	0.000264	0.015752	0	1185	0	0	1.002506	0	1	-360	360;
	8725	3934	0.01862	0.220201	0	0	0	0	0	0	1	-360	360;
	6965	3934	0.0168	0.250561	0	0	0	0	0	0	1	-360	360;
	1148	4793	0.202902	0.952701	0	0	0	0	0	0	1	-360	360;
	1148	8798	0.949862	3.139543	0	0	0	0	0	0	1	-360	360;
	1148	1657	0.00261	0.215422	0	0	0	0	0	0	1	-360	360;
	7444	4154	-0.02155	0.439897	0	0	0	0	0	0	1	-360	360;
	6083	4154	0.07045	1.003492	0	0	0	0	0	0	1	-360	360;
	7444	8638	0.00807	0.465263	0	0	0	0	0	0	1	-360	360;
	7444	6336	0.194723	1.065305	0	0	0	0	0	0	1	-360	360;
	7444	3479	0.7241	1.639128	0	0	0	0	0	0	1	-360	360;
	7444	5232	0.514127	1.973476	0	0	0	0	0	0	1	-360	360;
	7444	8699	0.06643	0.866482	0	0	0	0	0	0	1	-360	360;
	7444	3777	0.55937	2.088574	0	0	0	0	0	0	1	-360	360;
	7444	5261	0.522202	2.029502	0	0	0	0	0	0	1	-360	360;
	7444	7620	0.30876	1.923338	0	0	0	0	0	0	1	-360	360;
	3225	7451	0.06357	0.598203	0	0	0	0	0	0	1	-360	360;
	4121	7451	0.000445	0.035961	0	724	0	0	0.941156	0	1	-360	360;
	6589	7451	0.01179	0.332211	0	0	0	0	0	0	1	-360	360;
	798	7451	0.02131	0.372955	0	0	0	0	0	0	1	-360	360;
	4500	7451	0.00274	0.381756	0	0	0	0	0	0	1	-360	360;
	6906	3934	0.293141	1.107872	0	0	0	0	0	0	1	-360	360;
	5754	3934	0.144281	0.82905	0	0	0	0	0	0	1	-360	360;
	2486	2443	0.00458	0.288905	0	0	0	0	0	0	1	-360	360;
	8427	309	0.365888	1.746611	0	0	0	0	0	0	1	-360	360;
	2176	309	1.44438	5.252272	0	0	0	0	0	0	1	-360	360;
	1823	309	0.059581	0.615888	0	0	0	0	0	0	1	-360	360;
	3583	309	3.709298	8.934297	0	0	0	0	0	0	1	-360	360;
	6864	309	0.140919	0.944236	0	0	0	0	0	0	1	-360	360;
	6413	167	0.0005	0.0343	0	724	0	0	0.893644	0	1	-360	360;
	9160	167	0.01962	0.216178	0	0	0	0	0	0	1	-360	360;
	6083	8638	0.398678	1.516364	0	0	0	0	0	0	1	-360	360;
	6083	886	0.09755	0.839897	0	0	0	0	0	0	1	-360	360;
	3160	886	0.499525	2.142562	0	0	0	0	0	0	1	-360	360;
	1323	4151	0.01537	0.348347	0	0	0	0	0	0	1	-360	360;
	3655	4151	0.01946	0.372727	0	0	0	0	0	0	1	-360	360;
	6589	6321	0.01679	0.350455	0	0	0	0	0	0	1	-360	360;
	798	6321	0.0005	0.324174	0	0	0	0	0	0	1	-360	360;
	4500	6321	0.049469	0.557004	0	0	0	0	0	0	1	-360	360;
	1214	5896	0.0005	0.031271	0	0	0	0	0	0	1	-360	360;
	8427	5896	0.422748	2.328099	0	0	0	0	0	0	1	-360	360;
	2176	5896	1.022934	4.279959	0	0	0	0	0	0	1	-360	360;
	1823	5896	0.256612	1.738141	0	0	0	0	0	0	1	-360	360;
	3583	5896	2.300207	9.353099	0	0	0	0	0	0	1	-360	360;
	6864	5896	0.0157	0.997789	0	0	0	0	0	0	1	-360	360;
	6083	7997	0.022021	0.625496	0	0	0	0	0	0	1	-360	360;
	3160	7997	0.223988	1.463946	0	0	0	0	0	0	1	-360	360;
	8427	6120	-0.037934	0.385868	0	0	0	0	0	0	1	-360	360;
	2176	6120	2.334711	7.437397	0	0	0	0	0	0	1	-360	360;
	1823	6120	0.137849	1.103905	0	0	0	0	0	0	1	-360	360;
	3583	6120	5.726446	13.392563	0	0	0	0	0	0	1	-360	360;
	6864	6120	0.183531	1.402418	0	0	0	0	0	0	1	-360	360;
	8687	8427	3.1e-05	0.002125	0	0	0	0	0.999938	0.011109	1	-360	360;
	8427	5152	2.374861	8.881579	0	0	0	0	0	0	1	-360	360;
	8427	8134	0.158643	1.604501	0	0	0	0	0	0	1	-360	360;
	8427	8123	0.204162	1.92777	0	0	0	0	0	0	1	-360	360;
	8427	8451	0.169439	2.010873	0	0	0	0	0	0	1	-360	360;
	2176	5152	-0.011343	0.25657	0	0	0	0	0	0	1	-360	360;
	2176	5152	0.000423	0.033381	0	757	0	0	0.944892	0	1	-360	360;
	1823	5152	6.533884	12.310123	0	0	0	0	0	0	1	-360	360;
	3583	5152	0.04751	0.576054	0	0	0	0	0	0	1	-360	360;
	6864	5152	0.21343	1.602582	0	0	0	0	0	0	1	-360	360;
	2176	8134	7.076177	15.615651	0	0	0	0	0	0	1	-360	360;
	2176	8123	6.591413	15.880881	0	0	0	0	0	0	1	-360	360;
	2176	8451	0.187507	2.209903	0	0	0	0	0	0	1	-360	360;
	829	4146	0.0004	0.149419	0	0	0	0	0	0	1	-360	360;
	5754	4146	0.027469	0.320806	0	0	0	0	0	0	1	-360	360;
	7399	4146	0.00659	0.53155	0	0	0	0	0	0	1	-360	360;
	4501	6155	0.45281	1.969855	0	0	0	0	0	0	1	-360	360;
	9088	6155	0.561302	2.344628	0	0	0	0	0	0	1	-360	360;
	8388	6155	0.03438	0.839277	0	0	0	0	0	0	1	-360	360;
	7731	417	-0.00126	0.238202	0	0	0	0	0	0	1	-360	360;
	7731	417	0.000426	0.017209	0	921	0	0	1.002506	0	1	-360	360;
	7731	129	0.01628	0.291863	0	0	0	0	0	0	1	-360	360;
	6977	8503	-0.010145	0.227045	0	0	0	0	0	0	1	-360	360;
	4501	3617	0.00079	0.680124	0	0	0	0	0	0	1	-360	360;
	9088	3617	-6.2e-05	0.76438	0	0	0	0	0	0	1	-360	360;
	8388	3617	0.288988	1.328967	0	0	0	0	0	0	1	-360	360;
	4501	296	0.289174	0.947893	0	0	0	0	0	0	1	-360	360;
	9088	296	0.357831	1.127893	0	0	0	0	0	0	1	-360	360;
	8388	296	0.00434	0.222459	0	0	0	0	0	0	1	-360	360;
	9080	8059	0.162281	0.824401	0	0	0	0	0	0	1	-360	360;
	1765	8059	0.510455	1.73717	0	0	0	0	0	0	1	-360	360;
	624	8059	0.218822	1.471838	0	0	0	0	0	0	1	-360	360;
	1485	8059	0.029581	0.641756	0	0	0	0	0	0	1	-360	360;
	4718	8059	0.269194	1.398492	0	0	0	0	0	0	1	-360	360;
	798	6886	0.01824	0.407147	0	0	0	0	0	0	1	-360	360;
	798	7161	0.01992	0.388733	0	0	0	0	0	0	1	-360	360;
	6977	1949	0.504669	1.26314	0	0	0	0	0	0	1	-360	360;
	584	5530	0.180042	0.913089	0	0	0	0	0	0	1	-360	360;
	584	2223	0.28705	1.359279	0	0	0	0	0	0	1	-360	360;
	9154	4522	0.02939	0.690041	0	0	0	0	0	0	1	-360	360;
	7942	4522	0.106211	0.857975	0	0	0	0	0	0	1	-360	360;
	9154	1707	0.03057	0.640372	0	0	0	0	0	0	1	-360	360;
	7942	1707	0.11774	0.83438	0	0	0	0	0	0	1	-360	360;
	1823	8123	0.000416	0.031501	0	724	0	0	0.969682	0	1	-360	360;
	1823	8134	-0.02392	0.308892	0	0	0	0	0	0	1	-360	360;
	1823	8123	-0.021129	0.40847	0	0	0	0	0	0	1	-360	360;
	1823	8451	0.801524	3.248477	0	0	0	0	0	0	1	-360	360;
	3583	8134	15.18223	29.541321	0	0	0	0	0	0	1	-360	360;
	6864	8134	0.810744	3.364049	0	0	0	0	0	0	1	-360	360;
	3583	8123	13.91921	30.6467	0	0	0	0	0	0	1	-360	360;
	6864	8123	0.716322	3.154132	0	0	0	0	0	0	1	-360	360;
	7399	540	0.00489	0.406095	0	0	0	0	0	0	1	-360	360;
	9080	5958	0.447909	1.098823	0	0	0	0	0	0	1	-360	360;
	9080	6603	0.01056	0.353927	0	0	0	0	0	0	1	-360	360;
	9080	1690	0.7509	1.841551	0	0	0	0	0	0	1	-360	360;
	1765	5958	0.375789	0.752493	0	0	0	0	0	0	1	-360	360;
	1765	6603	0.190388	1.261149	0	0	0	0	0	0	1	-360	360;
	1765	1690	0.05324	0.584501	0	0	0	0	0	0	1	-360	360;
	5116	5530	0.01773	0.615173	0	0	0	0	0	0	1	-360	360;
	5116	6981	0.01987	0.600173	0	0	0	0	0	0	1	-360	360;
	2260	5530	0.466918	1.296815	0	0	0	0	0	0	1	-360	360;
	2260	2223	0.815443	1.930956	0	0	0	0	0	0	1	-360	360;
	4501	4268	0.02061	0.308595	0	0	0	0	0	0	1	-360	360;
	9088	4268	0.02549	0.367231	0	0	0	0	0	0	1	-360	360;
	8388	4268	0.00824	0.278492	0	0	0	0	0	0	1	-360	360;
	3583	8451	0.71662	3.572645	0	0	0	0	0	0	1	-360	360;
	3225	3261	0.045599	0.50595	0	0	0	0	0	0	1	-360	360;
	9154	683	0.00887	0.706717	0	0	0	0	0	0	1	-360	360;
	9154	7360	0.04089	0.90831	0	0	0	0	0	0	1	-360	360;
	9154	4557	0.04322	0.911288	0	0	0	0	0	0	1	-360	360;
	330	5530	0.032849	0.6625	0	0	0	0	0	0	1	-360	360;
	7781	5530	0.03726	0.637562	0	0	0	0	0	0	1	-360	360;
	873	5530	0.074341	0.632975	0	0	0	0	0	0	1	-360	360;
	6343	5530	0.07781	0.62157	0	0	0	0	0	0	1	-360	360;
	624	5958	0.534411	1.302908	0	0	0	0	0	0	1	-360	360;
	624	6603	0.02355	0.61356	0	0	0	0	0	0	1	-360	360;
	624	1690	0.420249	1.970499	0	0	0	0	0	0	1	-360	360;
	1485	5958	0.229855	0.963099	0	0	0	0	0	0	1	-360	360;
	4718	5958	0.074669	0.41343	0	0	0	0	0	0	1	-360	360;
	6977	822	0.06182	0.403213	0	0	0	0	0	0	1	-360	360;
	6864	8451	-0.029256	0.40562	0	0	0	0	0	0	1	-360	360;
	6083	6336	0.0033	0.455227	0	0	0	0	0	0	1	-360	360;
	3225	6336	0.00124	0.284731	0	0	0	0	0	0	1	-360	360;
	79	6886	0.04584	0.400152	0	0	0	0	0	0	1	-360	360;
	79	7161	0.03961	0.441877	0	0	0	0	0	0	1	-360	360;
	6083	2610	0.282149	1.82343	0	0	0	0	0	0	1	-360	360;
	3160	2610	-0.001467	0.288285	0	0	0	0	0	0	1	-360	360;
	8759	3479	0.000203	0.019723	0	1185	0	0	1.023144	-0.036838	1	-360	360;
	6083	3479	3.246901	6.102273	0	0	0	0	0	0	1	-360	360;
	5491	5673	0.000187	0.025917	0	1185	0	0	0.945816	-0.090914	1	-360	360;
	5491	5673	0.000198	0.025388	0	1185	0	0	0.945816	-0.090914	1	-360	360;
	5914	5673	0.046659	0.29938	0	0	0	0	0	0	1	-360	360;
	6524	8516	0.02508	0.489778	0	0	0	0	0	0	1	-360	360;
	6524	7655	0.0359	0.39268	0	0	0	0	0	0	1	-360	360;
	5969	8516	0.01931	0.550478	0	0	0	0	0	0	1	-360	360;
	5969	7655	0.0365	0.48437	0	0	0	0	0	0	1	-360	360;
	1485	6603	-0.015393	0.242665	0	0	0	0	0	0	1	-360	360;
	1485	6603	0.00024	0.016571	0	1185	0	0	0.988772	0	1	-360	360;
	4718	6603	0.778884	2.360744	0	0	0	0	0	0	1	-360	360;
	1485	1690	0.834349	2.163366	0	0	0	0	0	0	1	-360	360;
	2429	8516	0.02768	0.486653	0	0	0	0	0	0	1	-360	360;
	2429	8516	0.0005	0.0343	0	724	0	0	1.012424	0	1	-360	360;
	2429	7655	0.170748	0.871399	0	0	0	0	0	0	1	-360	360;
	6083	5232	3.440496	6.966735	0	0	0	0	0	0	1	-360	360;
	6083	8699	-0.025	0.350744	0	0	0	0	0	0	1	-360	360;
	6083	3777	3.294252	7.269391	0	0	0	0	0	0	1	-360	360;
	6083	5261	3.075277	7.063712	0	0	0	0	0	0	1	-360	360;
	6083	7620	2.188573	6.820568	0	0	0	0	0	0	1	-360	360;
	5914	2223	0.03254	0.295443	0	0	0	0	0	0	1	-360	360;
	3160	8552	0.000216	0.016549	0	1185	0	0	1.016627	0	1	-360	360;
	7116	4793	0.00914	0.394959	0	0	0	0	0	0	1	-360	360;
	7942	683	0.11943	0.839463	0	0	0	0	0	0	1	-360	360;
	7399	683	0.02832	0.560682	0	0	0	0	0	0	1	-360	360;
	7399	3854	0.025289	0.560351	0	0	0	0	0	0	1	-360	360;
	7116	8798	0.203901	1.623181	0	0	0	0	0	0	1	-360	360;
	6164	8033	0.000184	0.018745	0	1185	0	0	1.023144	-0.036838	1	-360	360;
	4500	6886	0.0086	0.46804	0	0	0	0	0	0	1	-360	360;
	4500	7161	0.00772	0.464148	0	0	0	0	0	0	1	-360	360;
	9160	8639	0.08261	0.429463	0	0	0	0	0	0	1	-360	360;
	9160	5580	0.09332	0.419277	0	0	0	0	0	0	1	-360	360;
	330	6981	0.06593	0.705471	0	0	0	0	0	0	1	-360	360;
	7781	6981	0.070298	0.687417	0	0	0	0	0	0	1	-360	360;
	8899	1508	0.000348	0.017003	0	790	0	0	0.967332	-0.0579	1	-360	360;
	9160	5324	0.000239	0.015483	0	1185	0	0	0.988494	-0.023724	1	-360	360;
	7942	7360	0.0275	0.64187	0	0	0	0	0	0	1	-360	360;
	7942	4557	0.02423	0.621711	0	0	0	0	0	0	1	-360	360;
	7116	1657	0.021711	0.411839	0	0	0	0	0	0	1	-360	360;
	7116	1657	0.000216	0.018022	0	1185	0	0	1.002506	0	1	-360	360;
	7116	8778	0.000173	0.017112	0	1185	0	0	1.016627	0	1	-360	360;
	7803	7620	0.000179	0.017308	0	1185	0	0	1.016325	-0.024392	1	-360	360;
	7803	7620	0.000209	0.016741	0	1185	0	0	1.016325	-0.024392	1	-360	360;
	4718	1690	-0.024793	0.19232	0	0	0	0	0	0	1	-360	360;
	4501	1497	0.000274	0.016144	0	1185	0	0	1.016325	-0.024392	1	-360	360;
	8388	1773	0.000476	0.030991	0	724	0	0	0.944892	0	1	-360	360;
	873	2223	0.11386	1.025723	0	0	0	0	0	0	1	-360	360;
	6343	2223	0.111979	1.046363	0	0	0	0	0	0	1	-360	360;
	6343	3155	0.00022	0.03264	0	1185	0	0	1.062286	-0.114968	1	-360	360;
	8779	8414	0.383283	2.185803	0	0	0	0	0	0	1	-360	360;
	8779	6794	0.497389	1.835595	0	0	0	0	0	0	1	-360	360;
	4670	7006	0.000615	0.029917	0	0	0	0	1.040949	0	1	-360	360;
	4670	7006	0.000634	0.029307	0	0	0	0	1.054426	0	1	-360	360;
	1900	8944	-0.02062	0.218203	0	0	0	0	0	0	1	-360	360;
	1900	2852	0.20759	1.109002	0	0	0	0	0	0	1	-360	360;
	1900	439	0.275429	1.015166	0	0	0	0	0	0	1	-360	360;
	1900	484	1.042036	2.256095	0	0	0	0	0	0	1	-360	360;
	511	8944	0.10976	0.74593	0	0	0	0	0	0	1	-360	360;
	3396	8944	0.287831	1.163678	0	0	0	0	0	0	1	-360	360;
	7799	8198	0.07	0.46124	0	0	0	0	0	0	1	-360	360;
	6956	8198	0.068479	0.427066	0	0	0	0	0	0	1	-360	360;
	7680	6621	0.103857	0.477763	0	0	0	0	0	0	1	-360	360;
	7680	4122	1.274515	2.49626	0	0	0	0	0	0	1	-360	360;
	6466	6621	0.107943	0.458262	0	0	0	0	0	0	1	-360	360;
	6466	4122	1.323546	2.392243	0	0	0	0	0	0	1	-360	360;
	8256	8414	0.209487	1.002909	0	0	0	0	0	0	1	-360	360;
	8256	6794	0.126607	0.574751	0	0	0	0	0	0	1	-360	360;
	2192	4218	0.00044	0.03688	0	757	0	0	1.002506	0	1	-360	360;
	2192	2418	0.01559	0.151898	0	0	0	0	0	0	1	-360	360;
	6900	2736	0.20449	0.679959	0	0	0	0	0	0	1	-360	360;
	6458	2852	0.105249	0.487299	0	0	0	0	0	0	1	-360	360;
	511	2852	0.000246	0.030366	0	954	0	0	1.015668	-0.130098	1	-360	360;
	511	2852	-0.01905	0.247025	0	0	0	0	0	0	1	-360	360;
	511	439	1.418837	2.394528	0	0	0	0	0	0	1	-360	360;
	511	484	0.205332	0.854986	0	0	0	0	0	0	1	-360	360;
	8415	2852	1.852727	2.747521	0	0	0	0	0	0	1	-360	360;
	3396	2852	1.359979	2.972315	0	0	0	0	0	0	1	-360	360;
	3396	439	0.02269	0.199771	0	0	0	0	0	0	1	-360	360;
	6900	3759	0.00688	0.14088	0	0	0	0	0	0	1	-360	360;
	7934	6621	0.00021	0.022916	0	954	0	0	0.976337	-0.045364	1	-360	360;
	7934	6621	0.00021	0.021219	0	954	0	0	0.976337	-0.045364	1	-360	360;
	1601	193	0.021781	0.193349	0	0	0	0	0	0	1	-360	360;
	3396	484	0.037729	0.222665	0	0	0	0	0	0	1	-360	360;
	5986	2418	0.05599	0.311752	0	0	0	0	0	0	1	-360	360;
	4732	5584	0.000687	0.037873	0	0	0	0	1.027812	0	1	-360	360;
	2932	5584	0.000697	0.03705	0	0	0	0	1.027812	0	1	-360	360;
	1601	5880	0.000258	0.015179	0	954	0	0	1.031474	0	1	-360	360;
	1601	5880	-0.008698	0.24938	0	0	0	0	0	0	1	-360	360;
	1601	5880	0.000218	0.015068	0	954	0	0	1.031407	0	1	-360	360;
	7799	5880	0.055651	0.515413	0	0	0	0	0	0	1	-360	360;
	6956	5880	0.05444	0.477231	0	0	0	0	0	0	1	-360	360;
	6900	3189	0.153248	0.626697	0	0	0	0	0	0	1	-360	360;
	3057	7741	-0.003823	0.150277	0	0	0	0	0	0	1	-360	360;
	3057	7741	0.00069	0.03981	0	0	0	0	1.125914	0	1	-360	360;
	197	7741	0.056219	0.23343	0	0	0	0	0	0	1	-360	360;
	2309	8414	0.000243	0.015553	0	954	0	0	0.946281	-0.010988	1	-360	360;
	2309	8414	-0.011591	0.366818	0	0	0	0	0	0	1	-360	360;
	2309	6794	0.145519	0.713643	0	0	0	0	0	0	1	-360	360;
	503	8414	0.000231	0.01967	0	954	0	0	0.993682	0.080858	1	-360	360;
	503	8414	-0.010888	0.388409	0	0	0	0	0	0	1	-360	360;
	503	6794	0.157909	0.659799	0	0	0	0	0	0	1	-360	360;
	6457	7514	0.00297	0.141711	0	0	0	0	0	0	1	-360	360;
	4498	3189	0.000245	0.011573	0	954	0	0	1.060747	0	1	-360	360;
	4498	3189	0.000255	0.011329	0	954	0	0	1.060747	0	1	-360	360;
	197	866	0.000697	0.027473	0	0	0	0	1.054426	0	1	-360	360;
	197	866	0.000627	0.031139	0	0	0	0	1.054426	0	1	-360	360;
	7692	4384	0.00203	0.101959	0	0	0	0	0.900834	0	1	-360	360;
	7692	4384	0.00195	0.103917	0	0	0	0	0.900834	0	1	-360	360;
	7843	4871	0.00341	0.149108	0	0	0	0	0.978276	0.039847	1	-360	360;
	6159	2530	0.001812	0.126988	0	0	0	0	0.89161	0.149487	1	-360	360;
	6159	2530	0.001982	0.124457	0	0	0	0	0.89161	0.149487	1	-360	360;
	4569	5702	0.00153	0.094124	0	0	0	0	0.939166	0	1	-360	360;
	4569	5702	0.00156	0.092157	0	0	0	0	0.939166	0	1	-360	360;
	2175	4515	0.00512	0.156934	0	0	0	0	1.075001	0	1	-360	360;
	2175	3358	0.00513	0.153636	0	0	0	0	1.03	0	1	-360	360;
	7876	5794	0.001473	0.055275	0	0	0	0	1.001399	0	1	-360	360;
	5600	7708	0.000628	0.026981	0	0	0	0	0.982004	0	1	-360	360;
	6653	8761	0.000768	0.027532	0	0	0	0	0.982004	0	1	-360	360;
	1636	4228	0.000778	0.05467	0	0	0	0	0.963346	0	1	-360	360;
	1636	4228	0.000767	0.055473	0	0	0	0	0.963346	0	1	-360	360;
	1953	8611	0.000905	0.055047	0	0	0	0	1.001399	0	1	-360	360;
	7416	2069	0.000902	0.025704	0	0	0	0	1.021576	0	1	-360	360;
	7416	2069	0.000902	0.024552	0	0	0	0	1.021576	0	1	-360	360;
	5046	2069	0.000437	0.013336	0	0	0	0	0.975505	0	1	-360	360;
	285	8032	0.001665	0.048767	0	0	0	0	1.001399	0	1	-360	360;
	4217	2285	0.00083	0.05142	0	0	0	0	1.001399	0	1	-360	360;
	4217	2285	0.000754	0.045701	0	0	0	0	1.021576	0	1	-360	360;
	8206	7559	0.000798	0.052223	0	0	0	0	0.982004	0	1	-360	360;
	6492	1673	0.000748	0.052717	0	0	0	0	0.963346	0	1	-360	360;
	3525	222	0.000435	0.02158	0	0	0	0	0.987916	0	1	-360	360;
	5101	8201	0.000871	0.0262	0	0	0	0	0.982004	0	1	-360	360;
	6084	4601	0.000881	0.047183	0	0	0	0	0.982004	0	1	-360	360;
	4241	2248	0.000214	0.018189	0	691	0	0	0.997518	0	1	-360	360;
	4717	5787	0.000545	0.028951	0	0	0	0	1.002304	0	1	-360	360;
	5425	7431	0.000433	0.016641	0	691	0	0	1.052632	0	1	-360	360;
	5425	7431	0.000271	0.019483	0	691	0	0	1.046093	0	1	-360	360;
	6632	3650	0.000292	0.041885	0	0	0	0	1.035986	0	1	-360	360;
	6632	3650	0.000283	0.025435	0	0	0	0	1.035986	0	1	-360	360;
	6917	4958	0.00019	0.020772	0	691	0	0	1.016388	0	1	-360	360;
	2962	8888	0.000567	0.0487	0	0	0	0	1.01961	0	1	-360	360;
	2962	8888	0.000292	0.024683	0	0	0	0	1.026095	0	1	-360	360;
	4542	7510	0.000238	0.020505	0	691	0	0	1.016388	0	1	-360	360;
	4542	7510	0.000302	0.024253	0	0	0	0	1.016399	0	1	-360	360;
	1549	3282	0.000394	0.025662	0	0	0	0	0.997518	0	1	-360	360;
	1549	3282	0.000403	0.026263	0	0	0	0	0.997518	0	1	-360	360;
	1549	3282	0.000358	0.02845	0	0	0	0	0.997518	0	1	-360	360;
	1200	3549	0.000184	0.020004	0	691	0	0	1.016388	0	1	-360	360;
	6771	4806	0.000287	0.030676	0	0	0	0	1.026104	0	1	-360	360;
	1086	6443	0.000183	0.021447	0	691	0	0	0.997518	0	1	-360	360;
	1086	6443	0.000183	0.019858	0	691	0	0	0.997518	0	1	-360	360;
	9159	6499	0.000196	0.019264	0	691	0	0	0.997518	0	1	-360	360;
	9159	6499	0.000205	0.018521	0	691	0	0	0.997518	0	1	-360	360;
	5685	3378	0.000205	0.018131	0	691	0	0	0.997518	0	1	-360	360;
	693	6994	0.000224	0.01875	0	691	0	0	1.035986	0	1	-360	360;
	693	6994	0.000685	0.016451	0	691	0	0	1.032682	0	1	-360	360;
	5332	3633	0.00021	0.024641	0	691	0	0	0.961805	0	1	-360	360;
	5332	3633	0.00021	0.025112	0	691	0	0	0.961805	0	1	-360	360;
	8034	4020	0	0.00554	0	1514	0	0	0	0	1	-360	360;
	1985	1023	0	0.006182	0	895	0	0	0.969385	0.248079	1	-360	360;
	477	6769	0.000205	0.016639	0	691	0	0	0.997518	0	1	-360	360;
	477	6769	0.000214	0.017035	0	691	0	0	0.997518	0	1	-360	360;
	477	6769	0.000214	0.016674	0	691	0	0	0.997518	0	1	-360	360;
	2188	5170	0.000224	0.018397	0	691	0	0	1.016388	0	1	-360	360;
	3061	699	0.000522	0.027299	0	0	0	0	1.006865	0	1	-360	360;
	3061	699	0.000646	0.037462	0	0	0	0	0.997538	0	1	-360	360;
	5823	7110	0.000244	0.033945	0	0	0	0	1.052632	0	1	-360	360;
	5823	7110	0.000196	0.023262	0	691	0	0	1.035986	0	1	-360	360;
	5280	7110	0.00091	0.001686	0	457	0	0	0.999945	-0.010479	1	-360	360;
	1947	4317	0.000179	0.016955	0	691	0	0	1.056356	0	1	-360	360;
	7303	429	0.000206	0.01883	0	691	0	0	1.016388	0	1	-360	360;
	6039	7973	0.000233	0.018474	0	691	0	0	1.006865	0	1	-360	360;
	6039	7973	0.000421	0.031827	0	0	0	0	1.014604	0	1	-360	360;
	1060	8981	0.000219	0.016973	0	691	0	0	1.026095	0	1	-360	360;
	8034	1752	0.000213	0.020004	0	691	0	0	0.997518	0	1	-360	360;
	739	1393	0.000556	0.033053	0	0	0	0	0.998185	0	1	-360	360;
	2235	8512	0.000508	0.028375	0	0	0	0	0.998185	0	1	-360	360;
	3428	1439	0.000179	0.022003	0	691	0	0	1.006865	0	1	-360	360;
	2730	6127	0.000209	0.02019	0	691	0	0	0.970491	0	1	-360	360;
	2730	6127	0.000214	0.020311	0	691	0	0	0.970491	0	1	-360	360;
	7395	5435	0.000616	0.030505	0	0	0	0	0.991996	0	1	-360	360;
	7395	5435	0.000626	0.029369	0	0	0	0	0.991996	0	1	-360	360;
	2437	3792	0.000214	0.019409	0	691	0	0	0.988343	0	1	-360	360;
	2437	3792	0.000192	0.022	0	691	0	0	0.979336	0	1	-360	360;
	5239	8438	0.000531	0.03466	0	0	0	0	0.991996	0	1	-360	360;
	5873	8147	0.000531	0.035228	0	0	0	0	0.99508	0	1	-360	360;
	5873	8147	0.00049	0.031535	0	0	0	0	0.995061	0	1	-360	360;
	5873	8147	0.000209	0.020964	0	691	0	0	0.988343	0	1	-360	360;
	7508	1240	0.000393	0.02663	0	0	0	0	1.035986	0	1	-360	360;
	7508	1240	0.000402	0.026088	0	0	0	0	1.035986	0	1	-360	360;
	3268	6397	0.000201	0.017787	0	691	0	0	1.006865	0	1	-360	360;
	4290	797	0.000289	0.018121	0	691	0	0	1.033257	0	1	-360	360;
	7929	4592	0.000643	0.024464	0	0	0	0	1.066843	0	1	-360	360;
	7929	4592	0.000568	0.027771	0	0	0	0	1.066843	0	1	-360	360;
	849	3840	0.000506	0.031833	0	0	0	0	0.998185	0	1	-360	360;
	849	3840	0.000196	0.024118	0	691	0	0	0.997518	0	1	-360	360;
	5919	8871	0.000205	0.022593	0	691	0	0	0.997518	0	1	-360	360;
	5919	8871	0.000186	0.01968	0	691	0	0	0.997518	0	1	-360	360;
	5919	8871	0.000205	0.01997	0	691	0	0	0.997518	0	1	-360	360;
	1168	2686	0.000215	0.018787	0	691	0	0	1.016388	0	1	-360	360;
	1168	2686	0.000215	0.018388	0	691	0	0	1.016388	0	1	-360	360;
	3275	6243	0.000224	0.018834	0	691	0	0	0.997518	0	1	-360	360;
	3798	1121	0.000433	0.043546	0	0	0	0	1.052632	0	1	-360	360;
	5884	8207	0.000234	0.018757	0	691	0	0	0.996896	0	1	-360	360;
	5884	8207	0.000197	0.019207	0	691	0	0	0.991986	0	1	-360	360;
	5884	8207	0.000197	0.019495	0	691	0	0	0.991986	0	1	-360	360;
	6681	7819	0.00018	0.021226	0	691	0	0	1.046093	0	1	-360	360;
	6681	7819	0.000812	0.03849	0	0	0	0	1.046093	0	1	-360	360;
	2261	4204	0.000205	0.02035	0	691	0	0	1.056356	0	1	-360	360;
	731	3728	0.000593	0.031521	0	0	0	0	1.001309	0	1	-360	360;
	731	3728	0.000612	0.031123	0	0	0	0	1.001309	0	1	-360	360;
	731	3728	0.000427	0.024994	0	0	0	0	1.006865	0	1	-360	360;
	2764	415	0.000667	0.027207	0	0	0	0	1.026095	0	1	-360	360;
	2764	415	0.000621	0.030836	0	0	0	0	1.026095	0	1	-360	360;
	299	947	0.000192	0.020791	0	691	0	0	1.066843	0	1	-360	360;
	299	947	0.000192	0.02119	0	691	0	0	1.066843	0	1	-360	360;
	8806	32	0.000686	0.030987	0	0	0	0	1.026958	0	1	-360	360;
	3008	5670	0.000194	0.020764	0	691	0	0	0.988343	0	1	-360	360;
	179	1850	0.000717	0.038087	0	0	0	0	1.035986	0	1	-360	360;
	7918	6660	0.000206	0.01883	0	691	0	0	1.016388	0	1	-360	360;
	7918	77	0.000206	0.018439	0	691	0	0	1.016388	0	1	-360	360;
	7029	33	0.000219	0.017019	0	691	0	0	1.035986	0	1	-360	360;
	7029	33	0.000224	0.016641	0	691	0	0	1.035986	0	1	-360	360;
	6607	9039	0.000676	0.047085	0	0	0	0	1.026095	0	1	-360	360;
	6607	9039	0.000667	0.048014	0	0	0	0	1.016388	0	1	-360	360;
	7343	8338	0.000196	0.019215	0	691	0	0	0.997518	0	1	-360	360;
	3966	5890	0.000253	0.025528	0	0	0	0	1.011609	0	1	-360	360;
	8210	6936	0.00021	0.017884	0	691	0	0	1.016388	0	1	-360	360;
	8210	6936	0.00021	0.017518	0	691	0	0	1.016388	0	1	-360	360;
	1716	6544	0.000219	0.017358	0	691	0	0	1.006865	0	1	-360	360;
	1716	6544	0.000219	0.016973	0	691	0	0	1.006865	0	1	-360	360;
	3544	869	0.000295	0.027995	0	0	0	0	1.021673	0	1	-360	360;
	3544	869	0.000277	0.027486	0	0	0	0	1.026115	0	1	-360	360;
	3662	8242	0.000489	0.032074	0	0	0	0	0.998185	0	1	-360	360;
	6092	3109	0.00018	0.02129	0	691	0	0	1.020734	0	1	-360	360;
	1074	1769	0.000366	0.029147	0	0	0	0	1.000611	0	1	-360	360;
	1074	1769	0.000366	0.028067	0	0	0	0	1.002235	0	1	-360	360;
	1074	1769	0.00042	0.026331	0	0	0	0	1.000611	0	1	-360	360;
	4957	1513	0.000441	0.01916	0	691	0	0	1.012394	0	1	-360	360;
	7428	6134	0.000391	0.024766	0	0	0	0	0.988343	0	1	-360	360;
	7428	6134	0.000401	0.024225	0	0	0	0	0.997518	0	1	-360	360;
	7428	4651	0.000408	0.028237	0	0	0	0	1.006865	0	1	-360	360;
	6970	4632	0.000199	0.018609	0	691	0	0	0.997518	0	1	-360	360;
	2700	8953	0.000196	0.021597	0	691	0	0	0.997518	0	1	-360	360;
	2700	8953	0.000186	0.022003	0	691	0	0	0.997518	0	1	-360	360;
	2849	3291	0.000186	0.019718	0	691	0	0	1.006865	0	1	-360	360;
	4446	1457	0.000205	0.022703	0	691	0	0	1.016388	0	1	-360	360;
	814	7820	0.000228	0.021908	0	691	0	0	0.988343	0	1	-360	360;
	7436	7629	0.000322	0.028734	0	0	0	0	0.996603	0	1	-360	360;
	7436	9106	0.000319	0.026574	0	0	0	0	1.042861	0	1	-360	360;
	3587	1241	0.000289	0.018298	0	691	0	0	1.046093	0	1	-360	360;
	3587	1241	0.000199	0.02111	0	691	0	0	1.046093	0	1	-360	360;
	7688	8301	0.000196	0.02119	0	691	0	0	1.016388	0	1	-360	360;
	7688	8301	0.000196	0.021597	0	691	0	0	1.016388	0	1	-360	360;
	4507	1503	0.001097	0.082091	0	0	0	0	1.006865	0	1	-360	360;
	4507	1503	0.000406	0.048834	0	0	0	0	1.033257	0	1	-360	360;
	4507	1503	7.2e-05	0.017798	0	691	0	0	0.988343	0	1	-360	360;
	4507	1503	0.000424	0.04684	0	0	0	0	1.033257	0	1	-360	360;
	8260	4081	0.000224	0.01875	0	691	0	0	1.006865	0	1	-360	360;
	8260	4081	0.000224	0.018334	0	691	0	0	1.006865	0	1	-360	360;
	1280	4279	0.000379	0.018493	0	691	0	0	1.046093	0	1	-360	360;
	1280	4279	0.000253	0.019711	0	691	0	0	1.046093	0	1	-360	360;
	385	9028	0.000196	0.016483	0	1053	0	0	0.997518	0	1	-360	360;
	385	9028	0.000186	0.022421	0	691	0	0	0.997518	0	1	-360	360;
	6768	4096	0.00046	0.030794	0	0	0	0	1.002506	0	1	-360	360;
	9203	2426	0.000695	0.049098	0	0	0	0	0.982143	0	1	-360	360;
	9203	2426	0.000355	0.024893	0	691	0	0	0.93617	0	1	-360	360;
	118	2426	0.000365	0.024372	0	691	0	0	0.979227	0	1	-360	360;
	3145	7770	0.000411	0.023652	0	691	0	0	0.93617	0	1	-360	360;
	7267	4918	0.000386	0.025673	0	691	0	0	0.93617	0	1	-360	360;
	7267	4918	0.000344	0.029336	0	691	0	0	0.93617	0	1	-360	360;
	839	4506	0.000653	0.052104	0	0	0	0	0.982143	0	1	-360	360;
	839	4506	0.000767	0.043531	0	0	0	0	0.982143	0	1	-360	360;
	3445	6114	0.000323	0.028314	0	691	0	0	0.93617	0	1	-360	360;
	3445	6114	0.000323	0.025999	0	691	0	0	0.93617	0	1	-360	360;
	3445	7264	0.000365	0.025478	0	691	0	0	0.93617	0	1	-360	360;
	891	6306	0.00086	0.039291	0	0	0	0	0.982143	0	1	-360	360;
	720	1711	0.000355	0.026586	0	691	0	0	0.93617	0	1	-360	360;
	720	1711	0.000375	0.026313	0	691	0	0	0.93617	0	1	-360	360;
	6206	3240	0.000344	0.023329	0	691	0	0	0.979227	0	1	-360	360;
	6206	3240	0.000302	0.026375	0	691	0	0	0.979227	0	1	-360	360;
	9174	2924	0.000788	0.04323	0	0	0	0	0.982143	0	1	-360	360;
	9174	2924	0.000767	0.04407	0	0	0	0	0.982143	0	1	-360	360;
	1355	5856	0.000767	0.043562	0	0	0	0	0.982143	0	1	-360	360;
	1355	5856	0.000788	0.041281	0	0	0	0	0.982143	0	1	-360	360;
	1355	5856	0.000892	0.040524	0	0	0	0	0.982143	0	1	-360	360;
	3205	4683	0.000975	0.043562	0	0	0	0	0.982143	0	1	-360	360;
	3205	4683	0.000757	0.042649	0	0	0	0	0.982143	0	1	-360	360;
	5350	2361	0.000261	0.030762	0	691	0	0	0.93617	0	1	-360	360;
	5350	2361	0.000261	0.030088	0	691	0	0	0.93617	0	1	-360	360;
	5016	6112	0.000323	0.026447	0	691	0	0	0.979227	0	1	-360	360;
	5016	8860	0.000323	0.026885	0	691	0	0	0.979227	0	1	-360	360;
	7471	549	0.000323	0.02748	0	691	0	0	0.93617	0	1	-360	360;
	7471	549	0.000292	0.028158	0	691	0	0	0.93617	0	1	-360	360;
	549	5002	0	0.010856	0	667	0	0	0	0.120551	1	-360	360;
	5490	3817	0.000736	0.044526	0	0	0	0	0.982143	0	1	-360	360;
	2854	9112	0.00086	0.039488	0	0	0	0	0.982143	0	1	-360	360;
	2918	6532	0.000365	0.024372	0	691	0	0	0.93617	0	1	-360	360;
	2918	6532	0.000375	0.023986	0	691	0	0	0.93617	0	1	-360	360;
	4494	7284	0.000386	0.025673	0	691	0	0	0.93617	0	1	-360	360;
	2656	1775	0.000809	0.042318	0	0	0	0	0.982143	0	1	-360	360;
	3519	5641	0.000705	0.042038	0	0	0	0	0.982143	0	1	-360	360;
	3543	367	0.000334	0.028022	0	691	0	0	0.93617	0	1	-360	360;
	3543	1833	0.000353	0.022403	0	691	0	0	0.982143	0	1	-360	360;
	2372	367	0.000313	0.026354	0	691	0	0	0.93617	0	1	-360	360;
	2372	367	0.000323	0.025478	0	691	0	0	0.93617	0	1	-360	360;
	1876	7752	0.000344	0.024622	0	691	0	0	0.979227	0	1	-360	360;
	113	7752	0.000355	0.024038	0	691	0	0	0.979227	0	1	-360	360;
	3499	7289	0.000881	0.038244	0	0	0	0	0.982143	0	1	-360	360;
	3499	7289	0.000757	0.040959	0	0	0	0	0.982143	0	1	-360	360;
	3082	2083	0.000354	0.028765	0	691	0	0	0.93617	0	1	-360	360;
	6224	2083	0.000354	0.027114	0	691	0	0	0.93617	0	1	-360	360;
	2174	2083	0.000695	0.04748	0	0	0	0	0.982143	0	1	-360	360;
	4623	1609	0.000323	0.028314	0	691	0	0	0.93617	0	1	-360	360;
	4623	1609	0.000323	0.026354	0	691	0	0	0.93617	0	1	-360	360;
	4623	1609	0.000323	0.025696	0	691	0	0	0.93617	0	1	-360	360;
	5589	1001	0.000355	0.027305	0	691	0	0	0.979227	0	1	-360	360;
	5589	1001	0.000355	0.026666	0	691	0	0	0.979227	0	1	-360	360;
	5589	1001	0.000923	0.037456	0	0	0	0	0.982143	0	1	-360	360;
	5589	1001	0.000788	0.035217	0	0	0	0	0.982143	0	1	-360	360;
	1198	7473	0.000323	0.026739	0	691	0	0	0.93617	0	1	-360	360;
	1198	7473	0.000323	0.02749	0	691	0	0	0.93617	0	1	-360	360;
	2479	5482	0.000313	0.027324	0	691	0	0	0.93617	0	1	-360	360;
	2479	5482	0.000313	0.028074	0	691	0	0	0.93617	0	1	-360	360;
	7831	5525	0.000601	0.024984	0	0	0	0	0.982143	0	1	-360	360;
	7831	5525	0.000715	0.022797	0	0	0	0	0.982143	0	1	-360	360;
	7831	5525	0.000344	0.024893	0	691	0	0	0.93617	0	1	-360	360;
	4970	7955	0.000365	0.024163	0	691	0	0	0.979227	0	1	-360	360;
	4970	7955	0.000344	0.02385	0	691	0	0	0.979227	0	1	-360	360;
	5007	7367	0.000757	0.040182	0	0	0	0	0.982143	0	1	-360	360;
	757	3758	0.000323	0.028937	0	691	0	0	0.979227	0	1	-360	360;
	3697	6952	0.000354	0.026958	0	691	0	0	0.93617	0	1	-360	360;
	3697	6952	0.000778	0.045656	0	0	0	0	0.982143	0	1	-360	360;
	3697	6952	0.000695	0.048932	0	0	0	0	0.982143	0	1	-360	360;
	3697	6952	0.000302	0.025926	0	691	0	0	0.93617	0	1	-360	360;
	6738	8180	0.000355	0.027328	0	691	0	0	0.93617	0	1	-360	360;
	6738	8180	0.000365	0.027156	0	691	0	0	0.93617	0	1	-360	360;
	9222	5935	0.000355	0.024299	0	691	0	0	0.979227	0	1	-360	360;
	7988	804	0.000375	0.026313	0	691	0	0	0.93617	0	1	-360	360;
	7988	804	0.000375	0.025605	0	691	0	0	0.93617	0	1	-360	360;
	3239	1539	0.000334	0.026291	0	691	0	0	0.979227	0	1	-360	360;
	3239	1539	0.000705	0.046735	0	0	0	0	0.982143	0	1	-360	360;
	2056	4084	0.000313	0.026875	0	691	0	0	0.979227	0	1	-360	360;
	2056	4084	0.000302	0.027376	0	691	0	0	0.979227	0	1	-360	360;
	4598	5110	0.000664	0.044868	0	0	0	0	0.982143	0	1	-360	360;
	2129	5110	0.000809	0.044526	0	0	0	0	0.982143	0	1	-360	360;
	5658	5550	0.00086	0.034253	0	0	0	0	0.982143	0	1	-360	360;
	5658	5550	0.000355	0.024372	0	691	0	0	0.93617	0	1	-360	360;
	2848	1526	0.000365	0.023788	0	691	0	0	0.93617	0	1	-360	360;
	4541	7464	0.000375	0.023204	0	691	0	0	0.979227	0	1	-360	360;
	4541	7464	0.000302	0.026593	0	691	0	0	0.979227	0	1	-360	360;
	1754	3857	0.000323	0.029256	0	691	0	0	0.93617	0	1	-360	360;
	1754	3857	0.000334	0.029986	0	691	0	0	0.93617	0	1	-360	360;
	1754	3857	0.000355	0.030979	0	691	0	0	0.93617	0	1	-360	360;
	2518	4544	0.000302	0.025707	0	691	0	0	0.93617	0	1	-360	360;
	2518	4544	0.000344	0.025405	0	691	0	0	0.93617	0	1	-360	360;
	3069	6115	0.00012	0.009612	0	591	0	0	0	-0.096484	1	-360	360;
	5067	7641	0.000365	0.023715	0	691	0	0	0.93617	0	1	-360	360;
	6036	8670	0.000365	0.023517	0	691	0	0	0.979227	0	1	-360	360;
	6036	5522	0.000334	0.026812	0	691	0	0	0.979227	0	1	-360	360;
	8486	8791	0.000695	0.047253	0	0	0	0	0.982143	0	1	-360	360;
	7530	8791	0.000778	0.043759	0	0	0	0	0.982143	0	1	-360	360;
	7522	7571	0.000313	0.030636	0	691	0	0	0.93617	0	1	-360	360;
	7522	7571	0.000313	0.028286	0	691	0	0	0.93617	0	1	-360	360;
	7522	7571	0.000344	0.028115	0	691	0	0	0.93617	0	1	-360	360;
	2526	1483	0.000902	0.039425	0	0	0	0	0.982143	0	1	-360	360;
	2365	8316	0.000355	0.024237	0	691	0	0	0.93617	0	1	-360	360;
	8672	1860	0.000375	0.023986	0	691	0	0	0.93617	0	1	-360	360;
	8672	1860	0.000375	0.023861	0	691	0	0	0.93617	0	1	-360	360;
	1494	583	0.000819	0.04153	0	0	0	0	0.982143	0	1	-360	360;
	1494	583	0.000809	0.042701	0	0	0	0	0.982143	0	1	-360	360;
	1502	2795	0.000334	0.028022	0	691	0	0	0.93617	0	1	-360	360;
	8195	933	0.000632	0.048237	0	0	0	0	0.982143	0	1	-360	360;
	8195	933	0.000632	0.045159	0	0	0	0	0.982143	0	1	-360	360;
	8195	933	0.000715	0.044039	0	0	0	0	0.982143	0	1	-360	360;
	3918	6802	0.00074	0.044479	0	0	0	0	0.986547	0	1	-360	360;
	3918	4025	0.00074	0.043811	0	0	0	0	0.986547	0	1	-360	360;
	4435	2079	0.000375	0.02385	0	691	0	0	0.93617	0	1	-360	360;
	4435	2079	0.000365	0.023267	0	691	0	0	0.93617	0	1	-360	360;
	4435	2079	0.000323	0.026739	0	691	0	0	0.93617	0	1	-360	360;
	1081	7361	0.000684	0.047025	0	0	0	0	0.982143	0	1	-360	360;
	9101	2177	0.000313	0.02748	0	691	0	0	0.93617	0	1	-360	360;
	6555	8475	0.000323	0.028554	0	691	0	0	0.93617	0	1	-360	360;
	6555	8475	0.000302	0.026437	0	691	0	0	0.93617	0	1	-360	360;
	7056	7577	0.000365	0.025551	0	691	0	0	0.93617	0	1	-360	360;
	7056	7577	0.000344	0.024893	0	691	0	0	0.93617	0	1	-360	360;
	7056	9051	0.000375	0.024508	0	691	0	0	0.979227	0	1	-360	360;
	6889	6785	0.000778	0.041157	0	0	0	0	0.982143	0	1	-360	360;
	6889	6785	0.000923	0.03702	0	0	0	0	0.982143	0	1	-360	360;
	6889	6785	0.000302	0.026593	0	691	0	0	0.979227	0	1	-360	360;
	1027	7148	0.000344	0.029667	0	691	0	0	0.93617	0	1	-360	360;
	1027	44	0.000334	0.02748	0	691	0	0	0.93617	0	1	-360	360;
	5308	7148	0.000323	0.030636	0	691	0	0	0.89674	0	1	-360	360;
	5308	44	0.000323	0.028526	0	691	0	0	0.93617	0	1	-360	360;
	9213	6231	0.000344	0.025332	0	691	0	0	0.979227	0	1	-360	360;
	9213	8291	0.000334	0.024674	0	691	0	0	0.979227	0	1	-360	360;
	9213	6842	0.000344	0.02457	0	691	0	0	0.979227	0	1	-360	360;
	3608	6664	0.000365	0.023788	0	691	0	0	0.93617	0	1	-360	360;
	4339	6570	0.000365	0.023601	0	691	0	0	0.93617	0	1	-360	360;
	4339	6570	0.000344	0.02652	0	691	0	0	0.93617	0	1	-360	360;
	7513	6570	0.000344	0.027272	0	691	0	0	0.93617	0	1	-360	360;
	7513	6570	0.000334	0.027563	0	691	0	0	0.93617	0	1	-360	360;
	8887	6639	0.000313	0.030636	0	691	0	0	0.979227	0	1	-360	360;
	8887	6639	0.000313	0.028286	0	691	0	0	0.979227	0	1	-360	360;
	5461	4816	0.000881	0.036803	0	0	0	0	0.982143	0	1	-360	360;
	5461	4816	0.000902	0.036668	0	0	0	0	0.982143	0	1	-360	360;
	5461	4816	0.000881	0.034014	0	0	0	0	0.982143	0	1	-360	360;
	5460	3786	0.000365	0.026027	0	691	0	0	0.93617	0	1	-360	360;
	5460	3786	0.000386	0.025673	0	691	0	0	0.93617	0	1	-360	360;
	7256	4491	0.00012	0.009966	0	686	0	0	0	0.036236	1	-360	360;
	6153	58	0.00017	0.007798	0	1316	0	0	0	-0.119536	1	-360	360;
	6153	6807	0.000664	0.047107	0	0	0	0	0.982143	0	1	-360	360;
	1465	1262	0.000313	0.028314	0	691	0	0	0.93617	0	1	-360	360;
	1465	1262	0.000334	0.025707	0	691	0	0	0.93617	0	1	-360	360;
	1465	1551	0.000344	0.025478	0	691	0	0	0.93617	0	1	-360	360;
	7537	1551	0.000344	0.027236	0	691	0	0	0.93617	0	1	-360	360;
	7537	1551	0.000375	0.026518	0	691	0	0	0.93617	0	1	-360	360;
	6901	4874	0.000386	0.02646	0	691	0	0	0.93617	0	1	-360	360;
	5241	4	0.000788	0.040317	0	0	0	0	0.982143	0	1	-360	360;
	7520	561	0.000323	0.025863	0	691	0	0	0.979227	0	1	-360	360;
	749	4324	1e-05	0.022863	0	1048	0	0	0	0.04898	1	-360	360;
	4141	7124	0.000334	0.030306	0	691	0	0	0.93617	0	1	-360	360;
	4141	7124	0.000302	0.030636	0	691	0	0	0.93617	0	1	-360	360;
	7466	3649	0.00011	0.012655	0	572	0	0	0	-0.110406	1	-360	360;
	8222	6556	0.000344	0.024841	0	691	0	0	0.979227	0	1	-360	360;
	8222	6556	0.000365	0.025029	0	691	0	0	0.979227	0	1	-360	360;
	8222	6556	0.000355	0.024101	0	691	0	0	0.979227	0	1	-360	360;
	4125	163	0.000344	0.02385	0	691	0	0	0.93617	0	1	-360	360;
	4125	163	0.000411	0.02313	0	691	0	0	0.93617	0	1	-360	360;
	4125	163	0.000829	0.042764	0	0	0	0	0.982143	0	1	-360	360;
	5837	9217	0.000705	0.04748	0	0	0	0	0.982143	0	1	-360	360;
	5837	1448	0.000323	0.029986	0	691	0	0	0.93617	0	1	-360	360;
	5837	1448	0.000302	0.029952	0	691	0	0	0.93617	0	1	-360	360;
	8267	5365	0.000746	0.041064	0	0	0	0	0.982143	0	1	-360	360;
	8267	5365	0.000881	0.042214	0	0	0	0	0.982143	0	1	-360	360;
	2359	8005	0.00086	0.041209	0	0	0	0	0.982143	0	1	-360	360;
	2359	8005	0.000912	0.038669	0	0	0	0	0.982143	0	1	-360	360;
	8109	171	0.000375	0.026244	0	691	0	0	0.979227	0	1	-360	360;
	8109	171	0.000375	0.025457	0	691	0	0	0.979227	0	1	-360	360;
	3485	1156	0.000302	0.029096	0	691	0	0	0.93617	0	1	-360	360;
	4734	2229	0.000302	0.029667	0	691	0	0	0.93617	0	1	-360	360;
	1709	7702	0.000334	0.027563	0	691	0	0	0.979227	0	1	-360	360;
	1709	7702	0.000323	0.028074	0	691	0	0	0.979227	0	1	-360	360;
	905	8189	0.000736	0.041064	0	0	0	0	0.982143	0	1	-360	360;
	8334	8189	0.001234	0.040109	0	0	0	0	0.982143	0	1	-360	360;
	8334	8189	0.000757	0.043209	0	0	0	0	0.982143	0	1	-360	360;
	7998	6734	0.000323	0.026586	0	954	0	0	0.93617	0	1	-360	360;
	4196	280	0.000788	0.040296	0	0	0	0	0.982143	0	1	-360	360;
	4196	280	0.000912	0.040514	0	0	0	0	0.982143	0	1	-360	360;
	6475	3503	0.000798	0.042183	0	0	0	0	0.982143	0	1	-360	360;
	6475	3503	0.000778	0.043147	0	0	0	0	0.982143	0	1	-360	360;
	7328	4235	0.000323	0.027793	0	691	0	0	0.93617	0	1	-360	360;
	7328	4235	0.000302	0.027918	0	691	0	0	0.93617	0	1	-360	360;
	7328	4235	0.000302	0.025926	0	691	0	0	0.93617	0	1	-360	360;
	6921	2949	0.000344	0.024956	0	691	0	0	0.979227	0	1	-360	360;
	6921	2949	0.000355	0.024435	0	691	0	0	0.979227	0	1	-360	360;
	3830	306	0.000386	0.026392	0	691	0	0	0.93617	0	1	-360	360;
	5648	306	0.000365	0.025525	0	691	0	0	0.93617	0	1	-360	360;
	1730	115	0.000344	0.029986	0	691	0	0	0.93617	0	1	-360	360;
	1730	115	0.000323	0.029838	0	691	0	0	0.93617	0	1	-360	360;
	2721	5720	0.000323	0.02748	0	691	0	0	0.93617	0	1	-360	360;
	2721	5720	0.000313	0.028001	0	691	0	0	0.93617	0	1	-360	360;
	432	3246	0.000179	0.012988	0	1251	0	0	0.990991	0	1	-360	360;
	432	8846	0.000198	0.012724	0	1251	0	0	0.990991	0	1	-360	360;
	432	2393	0.000207	0.012469	0	1251	0	0	0.990991	0	1	-360	360;
	432	1895	0.000207	0.012206	0	1251	0	0	0.990991	0	1	-360	360;
	432	3412	0.000217	0.012912	0	1251	0	0	0.990991	0	1	-360	360;
	432	5586	0.000217	0.012627	0	1251	0	0	0.990991	0	1	-360	360;
	432	3112	0.000198	0.014317	0	1251	0	0	0.990991	0	1	-360	360;
	432	3112	0.000189	0.014592	0	1251	0	0	0.990991	0	1	-360	360;
	432	7857	0.000354	0.02725	0	691	0	0	0.979227	0	1	-360	360;
	960	8497	0.000628	0.042428	0	0	0	0	1.009174	0	1	-360	360;
	960	8497	0.000726	0.044039	0	0	0	0	0.982143	0	1	-360	360;
	3577	5907	0.000365	0.024893	0	691	0	0	0.979227	0	1	-360	360;
	3577	5907	0.000438	0.02457	0	691	0	0	0.979227	0	1	-360	360;
	3577	5907	0.000365	0.023986	0	691	0	0	0.979227	0	1	-360	360;
	7164	5918	0.000375	0.023267	0	691	0	0	0.979227	0	1	-360	360;
	7164	5918	0.000323	0.026375	0	691	0	0	0.93617	0	1	-360	360;
	1798	2467	0.000334	0.029735	0	691	0	0	0.93617	0	1	-360	360;
	1798	2467	0.000313	0.030237	0	691	0	0	0.93617	0	1	-360	360;
	3346	2132	0.000767	0.044049	0	0	0	0	0.982143	0	1	-360	360;
	3346	2132	0.000302	0.026072	0	691	0	0	0.93617	0	1	-360	360;
	3346	2132	0.00085	0.039892	0	0	0	0	0.982143	0	1	-360	360;
	3483	8992	0.000344	0.024956	0	691	0	0	0.979227	0	1	-360	360;
	8886	8992	0.000334	0.026666	0	691	0	0	0.93617	0	1	-360	360;
	2093	805	0.000365	0.023788	0	691	0	0	0.93617	0	1	-360	360;
	2093	805	0.000365	0.023465	0	691	0	0	0.93617	0	1	-360	360;
	221	1541	0.000334	0.026447	0	691	0	0	0.93617	0	1	-360	360;
	8347	5383	0.000323	0.027042	0	691	0	0	0.93617	0	1	-360	360;
	8347	5383	0.000313	0.02748	0	691	0	0	0.93617	0	1	-360	360;
	8347	5383	0.000302	0.028314	0	691	0	0	0.93617	0	1	-360	360;
	8347	7507	0.000323	0.025853	0	691	0	0	0.93617	0	1	-360	360;
	8347	7507	0.000365	0.026197	0	691	0	0	0.93617	0	1	-360	360;
	8347	7507	0.000365	0.025165	0	691	0	0	0.93617	0	1	-360	360;
	4000	4852	0.000355	0.024237	0	691	0	0	0.979227	0	1	-360	360;
	4000	2012	0.000375	0.024247	0	691	0	0	0.93617	0	1	-360	360;
	2183	4852	0.000365	0.023267	0	691	0	0	0.979227	0	1	-360	360;
	2183	2012	0.000342	0.023139	0	691	0	0	0.982143	0	1	-360	360;
	1642	7098	0.000344	0.026666	0	691	0	0	0.979227	0	1	-360	360;
	1642	687	0.000334	0.027021	0	691	0	0	0.93617	0	1	-360	360;
	809	7098	0.000302	0.027615	0	691	0	0	0.979227	0	1	-360	360;
	809	687	0.000313	0.026135	0	691	0	0	0.93617	0	1	-360	360;
	4594	8804	0.000365	0.025905	0	691	0	0	0.93617	0	1	-360	360;
	4594	8804	0.000355	0.025238	0	691	0	0	0.93617	0	1	-360	360;
	2919	4215	0.000334	0.026814	0	691	0	0	0.93617	0	1	-360	360;
	1142	6577	0.001242	0.083148	0	0	0	0	0.945477	0	1	-360	360;
	1142	4137	0.0017	0.090508	0	0	0	0	0.945477	0	1	-360	360;
	1930	3448	0.001648	0.071935	0	0	0	0	1.018675	0	1	-360	360;
	1930	6445	0.001742	0.073131	0	0	0	0	0.999333	0	1	-360	360;
	165	7904	0.000914	0.088676	0	0	0	0	0.986842	0	1	-360	360;
	165	5140	0.000955	0.085404	0	0	0	0	0.986842	0	1	-360	360;
	1786	3879	0.001479	0.088626	0	0	0	0	0.986842	0	1	-360	360;
	1786	4210	0.00153	0.084928	0	0	0	0	0.986842	0	1	-360	360;
	7228	6121	0.00151	0.102577	0	0	0	0	0.945477	0	1	-360	360;
	8171	2528	0.000884	0.108054	0	0	0	0	0.945477	0	1	-360	360;
	8345	5553	0.001099	0.069265	0	0	0	0	1.018675	0	1	-360	360;
	14	2925	0.000904	0.088121	0	0	0	0	0.986842	0	1	-360	360;
	4346	2859	0.001386	0.094158	0	0	0	0	0.986842	0	1	-360	360;
	4346	2859	0.001355	0.096007	0	0	0	0	0.986842	0	1	-360	360;
	4985	1166	0.001456	0.098925	0	0	0	0	0.962772	0	1	-360	360;
	4985	1166	0.001424	0.100867	0	0	0	0	0.962772	0	1	-360	360;
	1972	1780	0.001442	0.086424	0	0	0	0	0.999333	0	1	-360	360;
	1132	8445	0.001317	0.070127	0	0	0	0	1.074114	0	1	-360	360;
	603	2617	0.000991	0.083602	0	0	0	0	0.999333	0	1	-360	360;
	603	2617	0.001272	0.097228	0	0	0	0	0.999333	0	1	-360	360;
	603	5798	0.000968	0.088836	0	0	0	0	0.974659	0	1	-360	360;
	603	5798	0.001537	0.08896	0	0	0	0	0.974659	0	1	-360	360;
	603	5798	0.0018	0.078579	0	0	0	0	0.974659	0	1	-360	360;
	436	2576	0.001071	0.079233	0	0	0	0	0.999333	0	1	-360	360;
	5883	3911	0.001301	0.088365	0	0	0	0	1.018675	0	1	-360	360;
	5883	4796	0.001436	0.082886	0	0	0	0	1.018675	0	1	-360	360;
	5883	3632	0.001701	0.102782	0	0	0	0	0.923361	0	1	-360	360;
	8845	8752	0.001352	0.091819	0	0	0	0	0.999333	0	1	-360	360;
	1533	1966	0.001737	0.085791	0	0	0	0	0.962772	0	1	-360	360;
	1533	3165	0.001331	0.087657	0	0	0	0	0.945477	0	1	-360	360;
	1533	1966	0.001877	0.07879	0	0	0	0	0.962772	0	1	-360	360;
	1533	4590	0.001672	0.089295	0	0	0	0	0.962772	0	1	-360	360;
	1533	3165	0.001421	0.10862	0	0	0	0	0.945477	0	1	-360	360;
	1533	4590	0.001338	0.09699	0	0	0	0	0.962772	0	1	-360	360;
	3056	5496	0.001224	0.093571	0	0	0	0	1.018675	0	1	-360	360;
	2000	2548	0.001667	0.092522	0	0	0	0	0.945477	0	1	-360	360;
	2000	1071	0.0017	0.090508	0	0	0	0	0.945477	0	1	-360	360;
	2869	3455	0.001177	0.089984	0	0	0	0	1.038781	0	1	-360	360;
	6072	9092	0.000976	0.083762	0	0	0	0	0.986842	0	1	-360	360;
	6072	2292	0.001221	0.074424	0	0	0	0	0.974659	0	1	-360	360;
	5226	8340	0.001205	0.088315	0	0	0	0	1.038781	0	1	-360	360;
	2522	8386	0.000951	0.081618	0	0	0	0	0.999333	0	1	-360	360;
	2522	5953	0.0011	0.085679	0	0	0	0	0.962772	0	1	-360	360;
	2522	5953	0.000895	0.099866	0	0	0	0	0.962772	0	1	-360	360;
	4490	2829	0	0.064	0	0	0	0	0.937501	0	1	-360	360;
	4490	2829	0	0.06267	0	0	0	0	0.937501	0	1	-360	360;
	4490	2829	0	0.06134	0	0	0	0	0.937501	0	1	-360	360;
	1575	4591	0	0.06295	0	0	0	0	0.975	0	1	-360	360;
	1575	4591	0	0.07135	0	0	0	0	0.975	0	1	-360	360;
	3208	2004	0	0.04729	0	0	0	0	1.019713	0	1	-360	360;
	3208	2004	0	0.0482	0	0	0	0	1.019713	0	1	-360	360;
	3208	2004	0	0.04911	0	0	0	0	1.019713	0	1	-360	360;
	3371	5721	0	0.06744	0	0	0	0	0.980435	0	1	-360	360;
	7656	738	0	0.145661	0	0	0	0	1.052174	0	1	-360	360;
	8974	825	0	0.03024	0	0	0	0	1.052632	0	1	-360	360;
	8974	825	0	0.029729	0	0	0	0	1.052632	0	1	-360	360;
	825	7111	0	0.07191	0	0	0	0	0.937501	0	1	-360	360;
	825	7111	0	0.0815	0	0	0	0	0.937501	0	1	-360	360;
	9073	2819	0	0.07843	0	0	0	0	0.944565	0	1	-360	360;
	2502	5229	0	0.033891	0	0	0	0	1.052632	0	1	-360	360;
	2502	5229	0	0.035419	0	0	0	0	1.052632	0	1	-360	360;
	5229	7916	0	0.07323	0	0	0	0	0.980435	0	1	-360	360;
	5229	7916	0	0.07177	0	0	0	0	0.980435	0	1	-360	360;
	5501	131	0	0.06789	0	0	0	0	0.956521	0	1	-360	360;
	5501	131	0	0.06524	0	0	0	0	0.956521	0	1	-360	360;
	5501	131	0	0.06385	0	0	0	0	0.956521	0	1	-360	360;
	2649	5585	0	0.04033	0	0	0	0	1.052174	0	1	-360	360;
	5161	3838	0	0.07024	0	0	0	0	0.980435	0	1	-360	360;
	5161	3838	0	0.07161	0	0	0	0	0.980435	0	1	-360	360;
	6182	861	0	0.06784	0	0	0	0	1.0119	0	1	-360	360;
	3198	8811	0	0.071916	0	0	0	0	0.948274	0	1	-360	360;
	7008	4583	0	0.07323	0	0	0	0	0.991844	0	1	-360	360;
	7008	4583	0	0.07177	0	0	0	0	0.991844	0	1	-360	360;
	7081	1928	0	0.0447	0	0	0	0	1.052174	0	1	-360	360;
	7081	1928	0	0.04377	0	0	0	0	1.052174	0	1	-360	360;
	7081	1928	0	0.04123	0	0	0	0	1.052174	0	1	-360	360;
	2389	8003	0	0.06295	0	0	0	0	0	0	1	-360	360;
	2389	8003	0	0.07135	0	0	0	0	0	0	1	-360	360;
	2480	181	0	0.07014	0	0	0	0	0.932609	0	1	-360	360;
	2480	181	0	0.07149	0	0	0	0	0.932609	0	1	-360	360;
	1429	4417	0	0.034531	0	0	0	0	1.035087	0	1	-360	360;
	1429	6856	0	0.04656	0	0	0	0	1.007866	0	1	-360	360;
	1429	6856	0	0.04563	0	0	0	0	1.007866	0	1	-360	360;
	1809	8416	0.000229	0.01234	0	1429	0	0	0.859545	0	1	-360	360;
	1809	8416	0.000229	0.0122	0	1494	0	0	0.859545	0	1	-360	360;
	8526	2840	0.001485	0.065805	0	0	0	0	0.956938	0	1	-360	360;
	8526	2840	0.001583	0.06586	0	0	0	0	0.956938	0	1	-360	360;
	5284	2840	0.00175	0.070701	0	0	0	0	0.952381	0	1	-360	360;
	5284	2840	0.00169	0.0715	0	0	0	0	0.952381	0	1	-360	360;
	6743	6609	0.001228	0.060916	0	0	0	0	0.992489	0	1	-360	360;
	6743	6609	0.00116	0.063798	0	0	0	0	1.010798	0	1	-360	360;
	6743	6609	0.00113	0.059295	0	0	0	0	1.010736	0	1	-360	360;
	4998	6609	0.00177	0.06853	0	0	0	0	0.973143	0	1	-360	360;
	7458	2173	0.001363	0.063835	0	0	0	0	0.980349	0	1	-360	360;
	7458	2173	0.001457	0.067155	0	0	0	0	0.980349	0	1	-360	360;
	7679	2173	0.00193	0.06377	0	0	0	0	0.952381	0	1	-360	360;
	7679	2173	0.00194	0.06188	0	0	0	0	0.952381	0	1	-360	360;
	8715	4912	0.00045	0.02807	0	0	0	0	1.002506	0	1	-360	360;
	8715	2144	0.00042	0.028021	0	0	0	0	1.002506	0	1	-360	360;
	8580	5624	0.001208	0.06631	0	0	0	0	1.030783	0	1	-360	360;
	8580	5624	0.001228	0.066264	0	0	0	0	1.030783	0	1	-360	360;
	5850	2858	0.00047	0.026469	0	0	0	0	1.002506	0	1	-360	360;
	8943	7152	0.00186	0.06826	0	0	0	0	0.931619	0	1	-360	360;
	569	146	0.001431	0.068667	0	0	0	0	0.956938	0	1	-360	360;
	569	146	0.001529	0.068787	0	0	0	0	0.956938	0	1	-360	360;
	3245	146	0.00193	0.06377	0	0	0	0	0.952381	0	1	-360	360;
	3245	146	0.00194	0.06188	0	0	0	0	0.952381	0	1	-360	360;
	3360	5138	0	0.00898	0	0	0	0	0	0	1	-360	360;
	2578	6334	0	0.00433	0	0	0	0	0	0	1	-360	360;
	6334	4820	0.007955	0.151636	0	0	0	0	0.929293	0	1	-360	360;
	7044	402	0.000271	0.030694	0	0	0	0	1.052632	0	1	-360	360;
	7044	402	0.000271	0.02842	0	0	0	0	1.052632	0	1	-360	360;
	2204	672	0.00064	0.0518	0	0	0	0	0	0	1	-360	360;
	9006	7212	0.00072	0.05075	0	0	0	0	0	0	1	-360	360;
	1759	8170	0	0.23657	0	0	0	0	0	0	1	-360	360;
	8532	6671	0	0.020113	0	0	0	0	1.006865	0	1	-360	360;
	8532	8723	0	0.019669	0	0	0	0	1.006865	0	1	-360	360;
	3563	6622	0.000325	0.028508	0	0	0	0	1.006865	0	1	-360	360;
	3563	6622	0.000316	0.031545	0	0	0	0	1.006865	0	1	-360	360;
	9139	3465	0.000355	0.023171	0	0	0	0	1.006865	0	1	-360	360;
	2756	1468	0.000296	0.032897	0	0	0	0	1.006865	0	1	-360	360;
	2756	1468	0.000296	0.030461	0	0	0	0	1.006865	0	1	-360	360;
	2316	1	5.9e-05	0.022767	0	0	0	0	1.006865	0	1	-360	360;
	1072	6210	0.000335	0.029237	0	0	0	0	1.006865	0	1	-360	360;
	5747	9068	0.000385	0.027846	0	0	0	0	1.006865	0	1	-360	360;
	3884	6863	0.000444	0.020064	0	0	0	0	1.006865	0	1	-360	360;
	8375	49	0.00033	0.02558	0	0	0	0	0	0	1	-360	360;
	7639	5714	0	0.026909	0	0	0	0	0	0	1	-360	360;
	7311	7538	0.000385	0.022737	0	0	0	0	1.006865	0	1	-360	360;
	3068	708	0.000375	0.022441	0	0	0	0	1.006865	0	1	-360	360;
	7922	6129	-0.002742	0.00908	0	0	0	0	0	0	1	-360	360;
	7922	2956	-0.00268	0.00837	0	0	0	0	0	0	1	-360	360;
	7922	6874	-0.00304	0.00824	0	0	0	0	0	0	1	-360	360;
	7197	5398	-0.003927	0.00795	0	0	0	0	0	0	1	-360	360;
	7197	6249	-0.00401	0.00794	0	0	0	0	0	0	1	-360	360;
	4610	9000	0	0.00765	0	0	0	0	0	0	1	-360	360;
	4610	1546	0	0.00748	0	0	0	0	0	0	1	-360	360;
	7359	9167	-0.000111	0.00565	0	0	0	0	0	0	1	-360	360;
	7359	4027	-0.000339	0.00645	0	0	0	0	0	0	1	-360	360;
	7359	5841	-0.000332	0.00657	0	0	0	0	0	0	1	-360	360;
	7359	4398	-0.000118	0.00525	0	0	0	0	0	0	1	-360	360;
	4836	348	0.00593	0.0602	0	0	0	0	0	0	1	-360	360;
	4836	3227	8e-05	0.00811	0	0	0	0	0	0	1	-360	360;
	4836	4994	-0.003927	0.00835	0	0	0	0	0	0	1	-360	360;
	2615	7753	0	0.028951	0	0	0	0	1.006865	0	1	-360	360;
	2615	3384	0	0.02833	0	0	0	0	1.006865	0	1	-360	360;
	99	3024	0.000424	0.020212	0	0	0	0	1.006865	0	1	-360	360;
	4564	1743	0.00128	0.05539	0	0	0	0	0	0	1	-360	360;
	4564	5572	0.00126	0.056469	0	0	0	0	0	0	1	-360	360;
	4564	4075	0.00123	0.05756	0	0	0	0	0	0	1	-360	360;
	1469	289	0	0.046831	0	0	0	0	0	0	1	-360	360;
	3123	759	0	0.03197	0	0	0	0	1.006865	0	1	-360	360;
	5109	7103	0	0.027151	0	0	0	0	1.044622	0	1	-360	360;
	3699	927	0	0.0197	0	0	0	0	0	0	1	-360	360;
	7248	4283	0.000335	0.028458	0	0	0	0	1.006865	0	1	-360	360;
	7248	4283	0.000375	0.027846	0	0	0	0	1.006865	0	1	-360	360;
	7564	1170	0.00073	0.01489	0	0	0	0	0	0	1	-360	360;
	8876	8157	0	0.00864	0	0	0	0	0	0	1	-360	360;
	8876	1365	0	0.00881	0	0	0	0	0	0	1	-360	360;
	7182	2040	0.000325	0.028961	0	0	0	0	1.006865	0	1	-360	360;
	3196	6067	0.000296	0.025084	0	0	0	0	1.006865	0	1	-360	360;
	1224	7494	0.000325	0.025765	0	0	0	0	1.006865	0	1	-360	360;
	1224	7494	0.000375	0.020301	0	0	0	0	1.006865	0	1	-360	360;
	1224	7494	0.000345	0.021109	0	0	0	0	1.006865	0	1	-360	360;
	3099	127	0.000385	0.02101	0	0	0	0	1.006865	0	1	-360	360;
	1610	7593	0.000424	0.019472	0	0	0	0	1.006865	0	1	-360	360;
	1610	7593	0.000434	0.019048	0	0	0	0	1.006865	0	1	-360	360;
	6316	4348	0.000325	0.025637	0	0	0	0	1.006865	0	1	-360	360;
	6316	4348	0.000355	0.026564	0	0	0	0	1.006865	0	1	-360	360;
	660	5532	0.000355	0.021553	0	0	0	0	1.006865	0	1	-360	360;
	660	5532	0.000128	0.022007	0	0	0	0	1.006865	0	1	-360	360;
	835	4680	6e-05	0.007574	0	1942	0	0	0	0	1	-360	360;
	3449	4680	7e-05	0.007392	0	1942	0	0	0	0	1	-360	360;
	4680	1034	0.000395	0.021454	0	0	0	0	1.006865	0	1	-360	360;
	4147	3284	-0.000388	0.0074	0	0	0	0	0	0	1	-360	360;
	2332	5059	-0.000402	0.00724	0	0	0	0	0	0	1	-360	360;
	5124	1620	-0.000589	0.01411	0	0	0	0	0	0	1	-360	360;
	3348	7947	-0.00036	0.00803	0	0	0	0	0	0	1	-360	360;
	8252	6362	0	0.040449	0	0	0	0	0.956938	0	1	-360	360;
	3963	5493	0.000345	0.026071	0	0	0	0	1.006865	0	1	-360	360;
	2634	8185	0	0.00046	0	0	0	0	0	0	1	-360	360;
	2634	6910	0	0.00043	0	0	0	0	0	0	1	-360	360;
	875	1867	0.000375	0.022007	0	0	0	0	1.006865	0	1	-360	360;
	946	7403	0.000385	0.020196	0	0	0	0	0.993901	0	1	-360	360;
	5355	7066	0.000375	0.026624	0	0	0	0	0.993901	0	1	-360	360;
	4414	1423	0.000425	0.021208	0	0	0	0	0.993901	0	1	-360	360;
	8117	2675	0.000466	0.020185	0	0	0	0	0.993901	0	1	-360	360;
	3874	1613	0.000324	0.022139	0	0	0	0	0.993901	0	1	-360	360;
	3874	1613	0.000324	0.022706	0	0	0	0	0.993901	0	1	-360	360;
	3874	1613	0.000314	0.023142	0	0	0	0	0.993901	0	1	-360	360;
	9009	7320	0.000344	0.024832	0	0	0	0	0.993901	0	1	-360	360;
	9009	8484	0.000334	0.023071	0	0	0	0	0.993901	0	1	-360	360;
	2710	368	0.000335	0.022767	0	0	0	0	1.006865	0	1	-360	360;
	2680	339	0.00071	0.030845	0	0	0	0	1.006865	0	1	-360	360;
	3972	6963	0.000385	0.019472	0	0	0	0	1.006865	0	1	-360	360;
	3972	4126	0.000395	0.019057	0	0	0	0	1.006865	0	1	-360	360;
	3972	1303	0.000404	0.018643	0	0	0	0	1.006865	0	1	-360	360;
	3972	5114	0	0.027866	0	0	0	0	1.006865	0	1	-360	360;
	1684	7574	0.000325	0.032245	0	0	0	0	1.006865	0	1	-360	360;
	4330	8509	0.000365	0.022599	0	0	0	0	1.006865	0	1	-360	360;
	4330	8509	0.000355	0.023023	0	0	0	0	1.006865	0	1	-360	360;
	8524	4017	0	0.032004	0	0	0	0	1.045454	0	1	-360	360;
	6195	4017	0.00141	0.103559	0	377	0	0	0.882842	-0.48891	1	-360	360;
	2684	2317	0.000434	0.019018	0	0	0	0	1.006865	0	1	-360	360;
	6460	8774	0.000404	0.020497	0	0	0	0	1.006865	0	1	-360	360;
	5623	7633	0.00074	0.029553	0	0	0	0	1.006865	0	1	-360	360;
	1031	1688	0.000454	0.019669	0	0	0	0	1.006865	0	1	-360	360;
	4810	2729	0.00206	0.09005	0	0	0	0	0	0	1	-360	360;
	3262	246	0.00039	0.03411	0	0	0	0	1.052632	0	1	-360	360;
	3262	246	0.0004	0.034469	0	0	0	0	1.052632	0	1	-360	360;
	8496	4270	0.00183	0.090471	0	0	0	0	1.026087	0	1	-360	360;
	8496	4270	0.00187	0.07429	0	0	0	0	1.026087	0	1	-360	360;
	7674	1874	0.00077	0.04355	0	0	0	0	0.999313	0	1	-360	360;
	7674	1874	0.00077	0.04266	0	0	0	0	0.999313	0	1	-360	360;
	7330	7033	0.00116	0.04106	0	0	0	0	1.052632	0	1	-360	360;
	7330	7033	0.00118	0.04119	0	0	0	0	1.052632	0	1	-360	360;
	8344	2950	0.0012	0.03931	0	0	0	0	1.052632	0	1	-360	360;
	8344	2950	0.00107	0.04567	0	0	0	0	1.052632	0	1	-360	360;
	6860	6573	0.002282	0.077224	0	0	0	0	0.942174	0	1	-360	360;
	4208	6573	0.00102	0.04862	0	0	0	0	1.052632	0	1	-360	360;
	4208	6573	0.001	0.04954	0	0	0	0	1.052632	0	1	-360	360;
	3259	1710	0.00098	0.04477	0	0	0	0	1.052632	0	1	-360	360;
	3259	1710	0.00111	0.04387	0	0	0	0	1.052632	0	1	-360	360;
	3146	5593	0.00113	0.04193	0	0	0	0	1.052632	0	1	-360	360;
	8859	5251	0.002522	0.074933	0	0	0	0	1.013913	0	1	-360	360;
	1392	5001	0.001252	0.073601	0	0	0	0	0.94	0	1	-360	360;
	3538	5001	0.001271	0.072001	0	0	0	0	0.94	0	1	-360	360;
	8249	6139	0.000442	0.040802	0	921	0	0	0.88	0	1	-360	360;
	806	91	0.00048	0.041601	0	0	0	0	0.97	0	1	-360	360;
	1747	6998	0.000543	0.04261	0	0	0	0	0	0	1	-360	360;
	593	6998	0.000534	0.04342	0	0	0	0	0	0	1	-360	360;
	8432	6295	0.003895	0.040564	0	0	0	0	0.97	0	1	-360	360;
	8432	6295	0.004413	0.039762	0	0	0	0	0.97	0	1	-360	360;
	8432	6295	0.004498	0.038954	0	0	0	0	0.97	0	1	-360	360;
	3190	7090	0.000621	0.0376	0	0	0	0	0	0	1	-360	360;
	3484	7090	0.00063	0.0368	0	0	0	0	0	0	1	-360	360;
	3484	7844	0.00064	0.036	0	0	0	0	0	0	1	-360	360;
	281	3960	0.000555	0.040801	0	0	0	0	0.984999	0	1	-360	360;
	8086	1477	0.000546	0.0416	0	0	0	0	0.984999	0	1	-360	360;
	8086	8037	0.000527	0.0424	0	0	0	0	0.984999	0	1	-360	360;
	5008	6354	0.000456	0.043201	0	0	0	0	0.97	0	1	-360	360;
	5008	6354	0.000447	0.040001	0	0	0	0	0.97	0	1	-360	360;
	919	6354	0.000505	0.039202	0	0	0	0	0.97	0	1	-360	360;
	919	6354	0.000513	0.038402	0	0	0	0	0.97	0	1	-360	360;
	4825	7884	0.000513	0.037602	0	0	0	0	0.954999	0	1	-360	360;
	3023	7884	0.00052	0.036802	0	0	0	0	0.954999	0	1	-360	360;
	6345	7884	0.000529	0.036002	0	0	0	0	0.954999	0	1	-360	360;
	8114	8738	0.000473	0.040802	0	0	0	0	0.984999	0	1	-360	360;
	8773	8738	0.000465	0.041601	0	0	0	0	0.984999	0	1	-360	360;
	4059	6282	7e-05	0.021201	0	790	0	0	0	0	1	-360	360;
	2002	4340	7e-05	0.021599	0	790	0	0	0	0	1	-360	360;
	5165	3363	0.00054	0.04	0	0	0	0	0.98421	0	1	-360	360;
	5165	3363	0.000472	0.039202	0	0	0	0	0.984999	0	1	-360	360;
	2722	3363	0.00048	0.038402	0	0	0	0	0.984999	0	1	-360	360;
	2722	3363	0.00064	0.0376	0	0	0	0	0.98421	0	1	-360	360;
	1591	1124	0.000333	0.029441	0	0	0	0	0.925	0	1	-360	360;
	5558	8510	0.000353	0.028801	0	0	0	0	0.925	0	1	-360	360;
	5558	8510	0.000312	0.032641	0	0	0	0	0.925	0	1	-360	360;
	1689	8814	0.000529	0.041601	0	0	0	0	0.94	0	1	-360	360;
	8230	8814	0.000511	0.042401	0	0	0	0	0.94	0	1	-360	360;
	8393	7565	0.000633	0.051521	0	0	0	0	0.94	0	1	-360	360;
	8393	7565	0.000616	0.047701	0	0	0	0	0.94	0	1	-360	360;
	139	7565	0.000701	0.046752	0	0	0	0	0.94	0	1	-360	360;
	3257	6058	0.000811	0.04579	0	0	0	0	0.98871	0	1	-360	360;
	369	6058	0.000831	0.04484	0	0	0	0	0.98871	0	1	-360	360;
	369	6058	0.000841	0.043881	0	0	0	0	0.98871	0	1	-360	360;
	6068	7450	0.000621	0.036001	0	0	0	0	1.015	0	1	-360	360;
	6068	2271	0.000555	0.040801	0	0	0	0	1.015	0	1	-360	360;
	6263	3846	0.000563	0.0416	0	0	0	0	1.015	0	1	-360	360;
	1788	1020	0.000543	0.0424	0	0	0	0	1.015	0	1	-360	360;
	2824	211	0.001026	0.086398	0	0	0	0	0.97	0	1	-360	360;
	140	211	0.001007	0.080001	0	0	0	0	0.955	0	1	-360	360;
	3210	1670	0.000522	0.039201	0	0	0	0	0.954999	0	1	-360	360;
	5674	1670	0.00053	0.038401	0	0	0	0	0.954999	0	1	-360	360;
	4337	1670	0.000548	0.037601	0	0	0	0	0.94	0	1	-360	360;
	4779	1956	3e-05	0.006499	0	0	0	0	0	0	1	-360	360;
	4320	4302	0.000907	0.042929	0	0	0	0	0.97	0	1	-360	360;
	4858	4248	7.7e-05	0.0204	0	724	0	0	1.105	0	1	-360	360;
	7235	4858	3e-05	0.007522	0	0	0	0	0.995481	-0.09511	1	-360	360;
	4248	4302	0.000773	0.05056	0	0	0	0	0.97	0	1	-360	360;
	5897	2720	0.000728	0.043197	0	0	0	0	1.045	0	1	-360	360;
	420	243	0.000447	0.040001	0	0	0	0	1.045	0	1	-360	360;
	8898	6052	0.00061	0.0392	0	0	0	0	1.03	0	1	-360	360;
	5311	8668	0.001203	0.076801	0	0	0	0	1.03	0	1	-360	360;
	3891	8668	0.001223	0.075201	0	0	0	0	1.03	0	1	-360	360;
	662	7073	0.004533	0.037356	0	0	0	0	0.984999	0	1	-360	360;
	6017	349	0.004615	0.036549	0	0	0	0	0.984999	0	1	-360	360;
	6766	7156	0.004113	0.041392	0	0	0	0	0.984999	0	1	-360	360;
	4077	7442	0.000627	0.041599	0	0	0	0	1.039999	0	1	-360	360;
	1052	8277	0.001048	0.084802	0	0	0	0	1.045	0	1	-360	360;
	3878	8277	0.001029	0.086397	0	0	0	0	1.045	0	1	-360	360;
	76	5346	0.000808	0.047709	0	0	0	0	0	0	1	-360	360;
	4485	3000	0.00061	0.0392	0	0	0	0	1.015	0	1	-360	360;
	4485	3000	0.00062	0.0384	0	0	0	0	1.015	0	1	-360	360;
	6283	3000	0.00064	0.0376	0	0	0	0	1.015	0	1	-360	360;
	395	4082	0.00063	0.0368	0	0	0	0	0.984999	0	1	-360	360;
	5843	5246	0.00064	0.036	0	0	0	0	0.984999	0	1	-360	360;
	5843	5246	0.000573	0.0408	0	0	0	0	0.984999	0	1	-360	360;
	6477	2478	0.001067	0.083199	0	0	0	0	0	0	1	-360	360;
	5295	2478	0.001048	0.084802	0	0	0	0	0	0	1	-360	360;
	1443	6268	6e-05	0.11487	0	1415	0	0	0	0	1	-360	360;
	4638	6268	0.20187	1.758347	0	1415	0	0	0	0	1	-360	360;
	1315	3848	0.03911	0.342562	0	1415	0	0	0	0	1	-360	360;
	2997	4595	0.01564	0.58593	0	1415	0	0	0	0	1	-360	360;
	2997	4581	-0.000248	0.227727	0	1415	0	0	0	0	1	-360	360;
	6190	977	0.01414	0.513223	0	1415	0	0	0	0	1	-360	360;
	3676	977	-0.007335	0.347624	0	1415	0	0	0	0	1	-360	360;
	5505	977	-0.006901	0.404112	0	1415	0	0	0	0	1	-360	360;
	7625	3438	0.051021	0.452583	0	1415	0	0	0	0	1	-360	360;
	3797	638	0.586797	1.843698	0	1415	0	0	0	0	1	-360	360;
	214	638	0.487562	1.572644	0	1415	0	0	0	0	1	-360	360;
	4847	638	0.12936	0.588864	0	1415	0	0	0	0	1	-360	360;
	6238	4163	0.000246	0.025121	0	0	0	0	0.963066	0	1	-360	360;
	4638	7293	0.10005	1.276798	0	1415	0	0	0	0	1	-360	360;
	3676	7293	0.21936	1.401591	0	1415	0	0	0	0	1	-360	360;
	5505	7293	0.22688	1.394834	0	1415	0	0	0	0	1	-360	360;
	1443	2509	0.04986	0.379174	0	1415	0	0	0	0	1	-360	360;
	3797	966	0.02303	0.62419	0	1415	0	0	0	0	1	-360	360;
	3797	2604	0.03435	0.639633	0	1415	0	0	0	0	1	-360	360;
	248	8449	0.01742	0.603616	0	1415	0	0	0	0	1	-360	360;
	4244	8449	0.14245	1.97967	0	1415	0	0	0	0	1	-360	360;
	1110	8449	-0.000145	0.32938	0	1415	0	0	0	0	1	-360	360;
	4533	3219	-0.000298	0.102701	0	1415	0	0	0	0	1	-360	360;
	3147	6634	0.000708	0.03271	0	0	0	0	1.065848	0	1	-360	360;
	3147	7854	0.06719	0.834339	0	1415	0	0	0	0	1	-360	360;
	6109	2274	0.050909	0.307686	0	1415	0	0	0	0	1	-360	360;
	993	2274	0.212521	0.943657	0	1415	0	0	0	0	1	-360	360;
	4257	8387	-0.002652	0.490201	0	1415	0	0	0	0	1	-360	360;
	4257	5640	0.02628	0.438469	0	1415	0	0	0	0	1	-360	360;
	5462	8387	-0.008241	0.388947	0	1415	0	0	0	0	1	-360	360;
	7553	8387	0.000226	0.024837	0	0	0	0	1.039739	0	1	-360	360;
	2758	8387	0.106109	0.699359	0	1415	0	0	0	0	1	-360	360;
	2758	5640	0.02855	0.418988	0	1415	0	0	0	0	1	-360	360;
	1110	328	0.292479	1.229545	0	1415	0	0	0	0	1	-360	360;
	214	3219	0.757066	1.631798	0	1415	0	0	0	0	1	-360	360;
	4847	3219	0.384752	1.021653	0	1415	0	0	0	0	1	-360	360;
	5209	4695	-0.014297	0.212851	0	1415	0	0	0	0	1	-360	360;
	5209	4695	0.000444	0.031171	0	0	0	0	0.974118	0	1	-360	360;
	5209	3515	-0.014091	0.217459	0	1415	0	0	0	0	1	-360	360;
	7625	6100	0.00894	0.300393	0	1415	0	0	0	0	1	-360	360;
	8822	8971	0.049771	0.546674	0	1415	0	0	0	0	1	-360	360;
	2216	8652	0.000737	0.026925	0	0	0	0	1.059199	0	1	-360	360;
	2216	8652	0.000845	0.026029	0	0	0	0	1.059199	0	1	-360	360;
	8822	3084	0.05157	0.475537	0	1415	0	0	0	0	1	-360	360;
	7625	2771	0.054781	0.446446	0	1415	0	0	0	0	1	-360	360;
	3454	2013	0.000452	0.023578	0	0	0	0	0.979739	0	1	-360	360;
	3368	2862	-0.010579	0.27814	0	1415	0	0	0	0	1	-360	360;
	3368	5075	-0.007004	0.324484	0	1415	0	0	0	0	1	-360	360;
	4244	8610	-0.034029	0.423698	0	1415	0	0	0	0	1	-360	360;
	2835	1492	0.000384	0.0242	0	0	0	0	1.027158	0	1	-360	360;
	1443	1492	0.299587	1.849794	0	1415	0	0	0	0	1	-360	360;
	1443	5305	0.301095	1.740868	0	1415	0	0	0	0	1	-360	360;
	1103	4844	0.000759	0.019191	0	921	0	0	1.046145	0	1	-360	360;
	248	7025	0.170173	0.568199	0	1415	0	0	0	0	1	-360	360;
	7625	6708	0.00708	0.453843	0	1415	0	0	0	0	1	-360	360;
	8822	6708	0.057719	0.644008	0	1415	0	0	0	0	1	-360	360;
	7089	7025	0.077609	0.938223	0	1415	0	0	0	0	1	-360	360;
	1110	7025	0.200409	0.618946	0	1415	0	0	0	0	1	-360	360;
	7625	3404	0.152771	0.863657	0	1415	0	0	0	0	1	-360	360;
	8822	3404	-0.001921	0.311818	0	1415	0	0	0	0	1	-360	360;
	1110	578	0.305992	1.948119	0	1415	0	0	0	0	1	-360	360;
	8822	5513	0.17005	0.993099	0	1415	0	0	0	0	1	-360	360;
	5479	1719	0.000257	0.029742	0	691	0	0	0.963066	0	1	-360	360;
	5479	3251	0.000268	0.029134	0	691	0	0	0.968561	0	1	-360	360;
	3762	4468	-0.014277	0.154039	0	1415	0	0	0	0	1	-360	360;
	3762	4468	0.000738	0.031616	0	0	0	0	0.967563	0	1	-360	360;
	499	4468	0.251178	0.958099	0	1415	0	0	0	0	1	-360	360;
	3762	378	0.08599	0.668788	0	1415	0	0	0	0	1	-360	360;
	3930	7854	0.000241	0.02217	0	0	0	0	0.996999	0	1	-360	360;
	499	2250	-0.003678	0.380909	0	1415	0	0	0	0	1	-360	360;
	8685	2250	-0.011302	0.627045	0	1415	0	0	0	0	1	-360	360;
	499	3411	0.00797	0.426157	0	1415	0	0	0	0	1	-360	360;
	8685	3411	0.058159	0.835806	0	1415	0	0	0	0	1	-360	360;
	499	2539	0.00907	0.406178	0	1415	0	0	0	0	1	-360	360;
	8685	2539	0.059789	0.790103	0	1415	0	0	0	0	1	-360	360;
	499	378	0.04738	0.355028	0	1415	0	0	0	0	1	-360	360;
	499	5386	0.00166	0.210118	0	1415	0	0	0	0	1	-360	360;
	993	770	0.000234	0.03334	0	0	0	0	0.974118	0	1	-360	360;
	993	7890	-0.007978	0.33365	0	1415	0	0	0	0	1	-360	360;
	993	1507	-0.007673	0.339439	0	1415	0	0	0	0	1	-360	360;
	8685	3015	0.000213	0.026084	0	0	0	0	0.952263	0	1	-360	360;
	8685	378	0.079162	0.676053	0	1415	0	0	0	0	1	-360	360;
	8685	3782	-0.012742	0.556163	0	1415	0	0	0	0	1	-360	360;
	8685	5386	0.05625	0.586697	0	1415	0	0	0	0	1	-360	360;
	1315	3953	0.11012	1.389876	0	1415	0	0	0	0	1	-360	360;
	4244	3326	0.12839	0.634174	0	1415	0	0	0	0	1	-360	360;
	4244	3326	0.000265	0.029415	0	0	0	0	0.963066	0	1	-360	360;
	2948	2573	0.01031	0.165651	0	1415	0	0	0	0	1	-360	360;
	791	6937	0.000906	0.032368	0	0	0	0	1.052632	0	1	-360	360;
	3368	6937	0.258574	1.91281	0	1415	0	0	0	0	1	-360	360;
	2997	8537	0.114169	0.69781	0	1415	0	0	0	0	1	-360	360;
	1110	7298	0.215475	0.921488	0	1415	0	0	0	0	1	-360	360;
	1110	262	0.000388	0.026559	0	0	0	0	1.020981	0	1	-360	360;
	1110	262	-0.012727	0.168081	0	1415	0	0	0	0	1	-360	360;
	1110	6169	0.11171	0.548317	0	1415	0	0	0	0	1	-360	360;
	2948	3782	0.04893	0.453587	0	1415	0	0	0	0	1	-360	360;
	2997	5158	0.000427	0.025869	0	0	0	0	1.059199	0	1	-360	360;
	2997	8758	0.000379	0.029319	0	0	0	0	1.065848	0	1	-360	360;
	7625	3003	0.086233	0.611399	0	1415	0	0	0	0	1	-360	360;
	7625	3991	0.03295	0.399238	0	1415	0	0	0	0	1	-360	360;
	8822	5656	0.034401	0.560661	0	1415	0	0	0	0	1	-360	360;
	5822	2553	0.04848	0.4525	0	1415	0	0	0	0	1	-360	360;
	5822	4051	0.05627	0.448553	0	1415	0	0	0	0	1	-360	360;
	27	173	0.000262	0.020993	0	0	0	0	1.03341	0	1	-360	360;
	4847	173	0.093521	0.726302	0	1415	0	0	0	0	1	-360	360;
	4847	6601	0.170831	0.824835	0	1415	0	0	0	0	1	-360	360;
	1315	3991	0.05836	0.35759	0	1415	0	0	0	0	1	-360	360;
	8401	3003	0.000563	0.031553	0	0	0	0	1.002888	0	1	-360	360;
	4847	3607	0.04447	0.619827	0	1415	0	0	0	0	1	-360	360;
	6504	3341	0.001519	0.050701	0	0	0	0	1.013513	0	1	-360	360;
	130	5452	0.000821	0.051524	0	0	0	0	1.052632	0	1	-360	360;
	130	5452	0.000803	0.047869	0	0	0	0	1.052632	0	1	-360	360;
	441	2233	0.000466	0.029318	0	0	0	0	1.025291	0	1	-360	360;
	8088	4721	0.000463	0.028006	0	0	0	0	1.038781	0	1	-360	360;
	8088	4721	0.000463	0.026569	0	0	0	0	1.038781	0	1	-360	360;
	3080	8789	0.000556	0.030088	0	0	0	0	1.08147	0	1	-360	360;
	3080	8789	0.000581	0.029019	0	0	0	0	1.08147	0	1	-360	360;
	5825	702	0.0017	0.077979	0	0	0	0	0	0	1	-360	360;
	8224	5517	0.000569	0.046371	0	0	0	0	1.052632	0	1	-360	360;
	8224	5517	0.00056	0.047255	0	0	0	0	1.052632	0	1	-360	360;
	536	4344	0.0008	0.050159	0	0	0	0	1.066857	0	1	-360	360;
	6080	5006	0.000379	0.039836	0	0	0	0	1.052632	0	1	-360	360;
	6080	5006	0.000433	0.039042	0	0	0	0	1.052632	0	1	-360	360;
	1899	7358	0.000448	0.032183	0	0	0	0	1.066857	0	1	-360	360;
	4712	1250	0.000498	0.034255	0	0	0	0	1.041667	0	1	-360	360;
	4712	1250	0.000507	0.033509	0	0	0	0	1.041667	0	1	-360	360;
	6187	2407	0.000415	0.040631	0	0	0	0	1.052632	0	1	-360	360;
	6187	2407	0.000406	0.041434	0	0	0	0	1.052632	0	1	-360	360;
	5562	1407	0.000397	0.037986	0	0	0	0	1.052632	0	1	-360	360;
	7827	1974	0.000322	0.035886	0	0	0	0	1.043149	0	1	-360	360;
	1025	8926	0.000408	0.039074	0	0	0	0	0.989653	0	1	-360	360;
	6298	6795	0.000417	0.026115	0	0	0	0	1.062288	0	1	-360	360;
	6298	6795	0.000443	0.02577	0	0	0	0	1.062288	0	1	-360	360;
	3007	6254	0.000396	0.033918	0	0	0	0	1.02924	0	1	-360	360;
	7969	177	0.000406	0.023438	0	0	0	0	1.052632	0	1	-360	360;
	7969	177	0.000352	0.026759	0	0	0	0	1.052632	0	1	-360	360;
	1703	7503	0.000334	0.036651	0	0	0	0	1.052632	0	1	-360	360;
	1703	7503	0.000334	0.037354	0	0	0	0	1.052632	0	1	-360	360;
	4704	4170	0.000305	0.035675	0	0	0	0	1.087225	0	1	-360	360;
	4704	4170	0.000296	0.033035	0	0	0	0	1.087225	0	1	-360	360;
	4704	4170	0.000338	0.032376	0	0	0	0	1.087225	0	1	-360	360;
	2031	8493	0.000285	0.02604	0	0	0	0	1.199891	0	1	-360	360;
	1025	1239	0.000419	0.02814	0	0	0	0	0.989653	0	1	-360	360;
	3007	9136	0.000453	0.025129	0	0	0	0	1.02924	0	1	-360	360;
	3007	4650	0.000453	0.02528	0	0	0	0	1.02924	0	1	-360	360;
	5050	1996	0.00174	0.07467	0	0	0	0	1.049587	0	1	-360	360;
	5050	1298	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	1360	5741	0.00172	0.099322	0	0	0	0	1.052632	0	1	-360	360;
	1360	5741	0.00168	0.101198	0	0	0	0	1.052632	0	1	-360	360;
	8820	6205	0.00363	0.146413	0	0	0	0	1.049587	0	1	-360	360;
	8820	6205	0.00182	0.07175	0	0	0	0	1.049587	0	1	-360	360;
	1732	456	0.0019	0.08995	0	0	0	0	1.052632	0	1	-360	360;
	4746	5447	0.00189	0.06882	0	0	0	0	1.049587	0	1	-360	360;
	4746	850	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	4187	7499	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	3731	760	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	5537	6063	0.00167	0.0776	0	0	0	0	1.049587	0	1	-360	360;
	7418	3120	0.000646	0.038909	0	0	0	0	1.002832	-0.000564	1	-360	360;
	7418	3120	0.000636	0.056994	0	0	0	0	1.003288	-0.001353	1	-360	360;
	3120	117	0.00182	0.07175	0	0	0	0	1.049587	0	1	-360	360;
	3120	3328	0.00185	0.07028	0	0	0	0	1.049587	0	1	-360	360;
	2427	6265	0.00194	0.088083	0	0	0	0	1.052632	0	1	-360	360;
	2427	6770	0.00198	0.086207	0	0	0	0	1.052632	0	1	-360	360;
	3570	6849	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	4396	6317	0.000697	0.032338	0	0	0	0	1.002506	0	1	-360	360;
	6317	3434	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	2707	2534	0.000667	0.033601	0	0	0	0	1.002506	0	1	-360	360;
	2707	2534	0.000647	0.034238	0	0	0	0	1.002506	0	1	-360	360;
	2534	3029	0.0016	0.07553	0	0	0	0	1.061984	0	1	-360	360;
	2534	1176	0.00182	0.07402	0	0	0	0	1.061984	0	1	-360	360;
	4335	5849	0.0019	0.08995	0	0	0	0	1.052632	0	1	-360	360;
	5619	3804	0.00189	0.06882	0	0	0	0	1.049587	0	1	-360	360;
	5619	617	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	6245	87	0.00201	0.084331	0	0	0	0	1.052632	0	1	-360	360;
	6245	87	0.00179	0.095578	0	0	0	0	1.052632	0	1	-360	360;
	9227	2061	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	6787	922	0.00167	0.0776	0	0	0	0	1.049587	0	1	-360	360;
	841	1988	0.00168	0.101198	0	0	0	0	1.052632	0	1	-360	360;
	6729	8754	0.00165	0.093702	0	0	0	0	1.052632	0	1	-360	360;
	6729	1412	0.00187	0.091826	0	0	0	0	1.052632	0	1	-360	360;
	6319	8922	0.0019	0.08995	0	0	0	0	1.052632	0	1	-360	360;
	3802	9056	0.00189	0.06882	0	0	0	0	1.049587	0	1	-360	360;
	3802	9056	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	4977	7344	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	4977	6465	0.00174	0.07467	0	0	0	0	1.049587	0	1	-360	360;
	7889	3153	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	2170	4881	0.00167	0.0776	0	0	0	0	1.049587	0	1	-360	360;
	2108	7853	0.00164	0.07907	0	0	0	0	1.049587	0	1	-360	360;
	9218	3129	0.00045	0.04299	0	0	0	0	1.113821	0	1	-360	360;
	9005	3207	0.00051	0.04213	0	0	0	0	1.113821	0	1	-360	360;
	1196	1197	0.00185	0.07028	0	0	0	0	1.049587	0	1	-360	360;
	1196	1197	0.00189	0.06882	0	0	0	0	1.049587	0	1	-360	360;
	204	675	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	204	2488	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	5009	3642	0.00179	0.095578	0	0	0	0	1.052632	0	1	-360	360;
	5009	3642	0.00176	0.097446	0	0	0	0	1.052632	0	1	-360	360;
	2027	1274	0.00167	0.0776	0	0	0	0	1.049587	0	1	-360	360;
	2027	1274	0.00164	0.07907	0	0	0	0	1.049587	0	1	-360	360;
	67	5537	0.000637	0.031701	0	0	0	0	1.002506	0	1	-360	360;
	67	7946	0.000716	0.031064	0	0	0	0	1.002506	0	1	-360	360;
	7946	8949	0.00185	0.07028	0	0	0	0	1.049587	0	1	-360	360;
	7946	215	0.00189	0.06882	0	0	0	0	1.049587	0	1	-360	360;
	2383	1921	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	2383	1921	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	6481	9035	0.00174	0.07467	0	0	0	0	1.049587	0	1	-360	360;
	1563	5322	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	6256	1756	0.00084	0.0322	0	0	0	0	1.052632	0	1	-360	360;
	1756	2984	0.00238	0.123248	0	0	0	0	1.049587	0	1	-360	360;
	5487	7109	0.00269	0.109562	0	0	0	0	1.049587	0	1	-360	360;
	6116	8455	0.00275	0.107281	0	0	0	0	1.049587	0	1	-360	360;
	4736	3050	0.000766	0.029164	0	0	0	0	1.002506	0	1	-360	360;
	4736	3050	0.000776	0.028526	0	0	0	0	1.002506	0	1	-360	360;
	3050	8351	0.00174	0.07467	0	0	0	0	1.049587	0	1	-360	360;
	3050	6550	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	3511	1783	0.00167	0.0776	0	0	0	0	1.049587	0	1	-360	360;
	3729	3393	0.00168	0.101198	0	0	0	0	1.052632	0	1	-360	360;
	3729	3732	0.00165	0.093702	0	0	0	0	1.052632	0	1	-360	360;
	971	7278	0.000716	0.031064	0	0	0	0	1.002506	0	1	-360	360;
	971	3441	0.000736	0.030437	0	0	0	0	1.002506	0	1	-360	360;
	3441	479	0.00189	0.06882	0	0	0	0	1.049587	0	1	-360	360;
	3441	479	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	6665	4364	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	7828	4537	0.00174	0.07467	0	0	0	0	1.049587	0	1	-360	360;
	1669	3810	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	1669	3810	0.00167	0.0776	0	0	0	0	1.049587	0	1	-360	360;
	4394	4667	0.00168	0.101198	0	0	0	0	1.052632	0	1	-360	360;
	4697	3853	0.00062	0.03068	0	0	0	0	1.015698	0	1	-360	360;
	4697	7671	0.000458	0.027443	0	0	0	0	1.002506	0	1	-360	360;
	4697	7671	0.000468	0.026875	0	0	0	0	1.002506	0	1	-360	360;
	5175	9105	0.001543	0.057853	0	0	0	0	0.892857	0	1	-360	360;
	2226	4759	0.00198	0.086207	0	0	0	0	1.052632	0	1	-360	360;
	7515	3595	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	3365	5964	0.00174	0.07467	0	0	0	0	1.049587	0	1	-360	360;
	1695	6733	0.00171	0.07614	0	0	0	0	1.049587	0	1	-360	360;
	3356	8692	0.000667	0.033601	0	0	0	0	1.002506	0	1	-360	360;
	3356	7549	0.00168	0.101198	0	0	0	0	1.052632	0	1	-360	360;
	8692	7549	0.0016	0.07321	0	0	0	0	1.049587	0	1	-360	360;
	6449	7578	0.0019	0.08995	0	0	0	0	1.052632	0	1	-360	360;
	8137	3640	0.000746	0.029801	0	0	0	0	1.002506	0	1	-360	360;
	7506	3815	0.00198	0.086207	0	0	0	0	1.052632	0	1	-360	360;
	5337	3030	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	8633	4992	0.00174	0.07467	0	0	0	0	1.049587	0	1	-360	360;
	3192	7120	0.000677	0.032965	0	0	0	0	1.002506	0	1	-360	360;
	3192	8583	0.000667	0.033601	0	0	0	0	1.002506	0	1	-360	360;
	3192	1321	0.00168	0.101198	0	0	0	0	1.052632	0	1	-360	360;
	3192	4988	0.00165	0.093702	0	0	0	0	1.052632	0	1	-360	360;
	7017	4716	0.000716	0.031064	0	0	0	0	1.002506	0	1	-360	360;
	7017	4716	0.000736	0.030437	0	0	0	0	1.002506	0	1	-360	360;
	7017	4716	0.000746	0.029801	0	0	0	0	1.002506	0	1	-360	360;
	4716	6825	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	4716	6825	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	1540	2092	0.00179	0.095578	0	0	0	0	1.052632	0	1	-360	360;
	1540	3921	0.00176	0.097446	0	0	0	0	1.052632	0	1	-360	360;
	8166	5231	0.00167	0.0776	0	0	0	0	1.049587	0	1	-360	360;
	4990	37	0.00164	0.07907	0	0	0	0	1.049587	0	1	-360	360;
	2152	508	0.0016	0.07321	0	0	0	0	1.049587	0	1	-360	360;
	2152	508	0.00182	0.07175	0	0	0	0	1.049587	0	1	-360	360;
	3017	7536	0.0019	0.08995	0	0	0	0	1.052632	0	1	-360	360;
	3017	7536	0.00194	0.088083	0	0	0	0	1.052632	0	1	-360	360;
	403	7536	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	7271	7585	0.00196	0.06589	0	0	0	0	1.049587	0	1	-360	360;
	5185	4741	0.00174	0.07467	0	0	0	0	1.049587	0	1	-360	360;
	4327	4351	0.00387	0.152273	0	0	0	0	1.049587	0	1	-360	360;
	887	7750	0.000667	0.033601	0	0	0	0	1.002506	0	1	-360	360;
	887	3161	0.00168	0.101198	0	0	0	0	1.052632	0	1	-360	360;
	887	2645	0.00165	0.093702	0	0	0	0	1.052632	0	1	-360	360;
	2251	1053	0.00182	0.07175	0	0	0	0	1.049587	0	1	-360	360;
	2251	1053	0.00185	0.07028	0	0	0	0	1.049587	0	1	-360	360;
	7225	941	0.00189	0.06882	0	0	0	0	1.049587	0	1	-360	360;
	7225	941	0.00192	0.06735	0	0	0	0	1.049587	0	1	-360	360;
	3915	3679	0.00196	0.07015	0	0	0	0	1.07438	0	1	-360	360;
	1127	3679	0.00174	0.0795	0	0	0	0	1.07438	0	1	-360	360;
	2904	2706	0.00136	0.05901	0	0	0	0	1.052632	0	1	-360	360;
	2904	576	0.00133	0.06014	0	0	0	0	1.052632	0	1	-360	360;
	3213	5632	0.00164	0.07907	0	0	0	0	1.049587	0	1	-360	360;
	3213	2333	0.0016	0.07321	0	0	0	0	1.049587	0	1	-360	360;
	1211	7660	0.000716	0.031064	0	0	0	0	1.002506	0	1	-360	360;
	7660	4378	0.00185	0.07028	0	0	0	0	1.049587	0	1	-360	360;
	7010	2198	0.00189	0.06882	0	0	0	0	1.049587	0	1	-360	360;
	4437	155	0.00158	0.05564	0	0	0	0	0	0	1	-360	360;
	4437	7730	0.00179	0.0695	0	0	0	0	0	0	1	-360	360;
	2939	341	0.00084	0.04445	0	0	0	0	0	0	1	-360	360;
	2939	341	0.0008	0.04546	0	0	0	0	0	0	1	-360	360;
	2939	341	0.00219	0.08229	0	0	0	0	0	0	1	-360	360;
	2939	341	0.00069	0.04566	0	0	0	0	1.047826	0	1	-360	360;
	9016	3924	0.00072	0.04835	0	0	0	0	0.999313	0	1	-360	360;
	9016	3924	0.00082	0.04738	0	0	0	0	0.999313	0	1	-360	360;
	5327	2345	0.00045	0.03151	0	0	0	0	1.052632	0	1	-360	360;
	5327	8470	0.00048	0.026419	0	0	0	0	1.052632	0	1	-360	360;
	3103	8471	0.0022	0.07023	0	0	0	0	0	0	1	-360	360;
	2855	223	0.00241	0.0727	0	0	0	0	1.052174	0	1	-360	360;
	2855	223	0.00081	0.04644	0	0	0	0	1.052174	0	1	-360	360;
	4361	3026	0.00158	0.08045	0	0	0	0	0	0	1	-360	360;
	7830	7739	0.00069	0.04707	0	0	0	0	1.052174	0	1	-360	360;
	1534	8306	0.00075	0.04866	0	0	0	0	1.052174	0	1	-360	360;
	1534	8306	0.00076	0.04586	0	0	0	0	1.052174	0	1	-360	360;
	8734	8570	0.00189	0.07781	0	0	0	0	0	0	1	-360	360;
	8734	8570	0.00205	0.07804	0	0	0	0	0	0	1	-360	360;
	663	7604	0.00207	0.091083	0	0	0	0	0	0	1	-360	360;
	663	7604	0.00211	0.06802	0	0	0	0	0	0	1	-360	360;
	1978	7477	0.0008	0.03997	0	0	0	0	1.052174	0	1	-360	360;
	3104	7477	0.00159	0.086876	0	0	0	0	0	0	1	-360	360;
	3104	7477	0.0026	0.07998	0	0	0	0	0	0	1	-360	360;
	4861	7528	0.00154	0.086488	0	0	0	0	0	0	1	-360	360;
	8562	4861	0.00052	0.033159	0	0	0	0	1.052632	0	1	-360	360;
	8562	7528	0.00072	0.0451	0	0	0	0	1.052174	0	1	-360	360;
	8562	7528	0.00068	0.04531	0	0	0	0	1.052174	0	1	-360	360;
	4801	3372	0.00042	0.030281	0	0	0	0	1.052632	0	1	-360	360;
	4801	8382	0.00079	0.04086	0	0	0	0	1.052174	0	1	-360	360;
	3372	5218	0.00168	0.07343	0	0	0	0	0	0	1	-360	360;
	7364	1964	0.00038	0.033591	0	0	0	0	1.052632	0	1	-360	360;
	7364	1964	0.00037	0.03405	0	0	0	0	1.052632	0	1	-360	360;
	1686	8618	0.00039	0.034849	0	0	0	0	1.052632	0	1	-360	360;
	1686	3748	0.00063	0.04764	0	0	0	0	1.052174	0	1	-360	360;
	1686	3748	0.00074	0.044626	0	0	0	0	1.006865	0	1	-360	360;
	7314	7848	0.00069	0.04232	0	0	0	0	1.052632	0	1	-360	360;
	7978	6673	0.00228	0.07286	0	0	0	0	0	0	1	-360	360;
	7978	6673	0.00189	0.07059	0	0	0	0	0	0	1	-360	360;
	2349	1714	0.00195	0.07374	0	0	0	0	0	0	1	-360	360;
	2349	1714	0.00199	0.07214	0	0	0	0	0	0	1	-360	360;
	3688	5629	0.00177	0.07931	0	0	0	0	0	0	1	-360	360;
	3688	5629	0.00174	0.08086	0	0	0	0	0	0	1	-360	360;
	465	4803	0.0015	0.08239	0	0	0	0	0	0	1	-360	360;
	465	4803	0.00149	0.091992	0	0	0	0	0	0	1	-360	360;
	8358	3470	0.00066	0.04441	0	0	0	0	1.052174	0	1	-360	360;
	1753	3470	0.00171	0.07418	0	0	0	0	0	0	1	-360	360;
	1753	3470	0.00175	0.07267	0	0	0	0	0	0	1	-360	360;
	2585	4899	0.00053	0.03038	0	0	0	0	1.052632	0	1	-360	360;
	4899	591	0.00165	0.07003	0	0	0	0	0	0	1	-360	360;
	4899	7705	0.00217	0.07323	0	0	0	0	0	0	1	-360	360;
	254	1427	0.00072	0.0453	0	0	0	0	1.052174	0	1	-360	360;
	7521	1087	0.0021	0.08257	0	0	0	0	0	0	1	-360	360;
	7521	1087	0.00186	0.087719	0	0	0	0	0	0	1	-360	360;
	895	8969	0.00076	0.04796	0	0	0	0	1.052174	0	1	-360	360;
	895	3232	0.00075	0.04466	0	0	0	0	1.052174	0	1	-360	360;
	3750	7687	0.00401	0.10005	0	0	0	0	0	0	1	-360	360;
	3750	7687	0.00272	0.07842	0	0	0	0	0	0	1	-360	360;
	3750	7687	0.00278	0.07679	0	0	0	0	0	0	1	-360	360;
	3750	7687	0.00283	0.07515	0	0	0	0	0	0	1	-360	360;
	2398	5000	0.00045	0.029271	0	0	0	0	1.052632	0	1	-360	360;
	2398	5000	0.0004	0.033169	0	0	0	0	1.052632	0	1	-360	360;
	2398	5000	0.00039	0.03382	0	0	0	0	1.052632	0	1	-360	360;
	2737	5190	0.00064	0.04577	0	0	0	0	1.052632	0	1	-360	360;
	694	1960	0.00147	0.08395	0	0	0	0	0	0	1	-360	360;
	694	180	0.00184	0.07633	0	0	0	0	0	0	1	-360	360;
	6098	9193	0.00074	0.04352	0	0	0	0	1.052174	0	1	-360	360;
	1473	4784	0.00176	0.08177	0	0	0	0	0	0	1	-360	360;
	1473	4784	0.00242	0.06613	0	0	0	0	0	0	1	-360	360;
	1920	3447	0.00042	0.030031	0	0	0	0	1.015038	0	1	-360	360;
	1920	3447	0.00043	0.02938	0	0	0	0	1.015038	0	1	-360	360;
	8556	9133	0.00014	0.019637	0	0	0	0	0	0.199334	1	-360	360;
	5824	8145	0.00037	0.03395	0	0	0	0	0.977444	0	1	-360	360;
	4111	3955	1.8e-05	0.052587	0	0	0	0	1.052632	0	1	-360	360;
	2846	3955	5.5e-05	0.082396	0	591	0	0	1.04762	0	1	-360	360;
	794	4924	0.001416	0.043162	0	0	0	0	0.92437	0	1	-360	360;
	794	4924	0.001603	0.042296	0	0	0	0	0.92437	0	1	-360	360;
	4190	3940	0.000806	0.025741	0	0	0	0	1.002506	0	1	-360	360;
	4190	3940	0.000826	0.025204	0	0	0	0	1.002506	0	1	-360	360;
	4190	3518	0.000107	0.01357	0	724	0	0	1.059051	0	1	-360	360;
	3940	956	0.00228	0.07939	0	0	0	0	0	0	1	-360	360;
	3518	3940	0.001087	0.076556	0	0	0	0	0.99449	0.048556	1	-360	360;
	3518	3940	0.001067	0.077593	0	0	0	0	1.00317	-0.04898	1	-360	360;
	6091	4063	0.001301	0.061931	0	0	0	0	0.948276	0	1	-360	360;
	6091	4063	0.001268	0.063098	0	0	0	0	0.948276	0	1	-360	360;
	7678	1075	0.002358	0.088053	0	0	0	0	0.948276	0	1	-360	360;
	7678	1075	0.00268	0.086295	0	0	0	0	0.948276	0	1	-360	360;
	7870	599	0.000779	0.028436	0	0	0	0	1.048339	-0.002864	1	-360	360;
	599	4242	0.001565	0.062128	0	0	0	0	0.932204	0	1	-360	360;
	599	4242	0.001588	0.060806	0	0	0	0	0.932204	0	1	-360	360;
	5389	8041	0.002713	0.092944	0	0	0	0	0.916667	0	1	-360	360;
	5389	8041	0.002464	0.122138	0	0	0	0	0.916667	0	1	-360	360;
	5052	446	0.002739	0.116189	0	0	0	0	0.932204	0	1	-360	360;
	5052	446	0.002681	0.118434	0	0	0	0	0.932204	0	1	-360	360;
	6398	607	0.000212	0.012061	0	0	0	0	0.956194	-0.071507	1	-360	360;
	4458	530	0	0.04332	0	0	0	0	1.041096	0	1	-360	360;
	4458	530	0	0.04332	0	0	0	0	1.041096	0	1	-360	360;
	4959	4654	0	0.0505	0	0	0	0	1.014795	0	1	-360	360;
	6370	6269	0	0.044	0	0	0	0	0.986875	0	1	-360	360;
	6370	6269	0	0.044	0	0	0	0	0.986875	0	1	-360	360;
	6588	2621	0	0.047975	0	0	0	0	0.987675	0	1	-360	360;
	6588	2621	0	0.047975	0	0	0	0	0.987675	0	1	-360	360;
	6588	5840	0	0.037831	0	0	0	0	0.987675	0	1	-360	360;
	6588	5840	0	0.037831	0	0	0	0	0.987675	0	1	-360	360;
	5757	3386	0	0.045294	0	0	0	0	1.001022	0	1	-360	360;
	5757	3386	0	0.045294	0	0	0	0	1.001022	0	1	-360	360;
	5757	2001	0	0.04563	0	0	0	0	0	0	1	-360	360;
	5757	2001	0	0.04563	0	0	0	0	0	0	1	-360	360;
	3628	3402	0	0.047975	0	0	0	0	0.987675	0	1	-360	360;
	3628	3402	0	0.047975	0	0	0	0	0.987675	0	1	-360	360;
	3628	6545	0	0.049246	0	0	0	0	0.987678	0	1	-360	360;
	3628	6545	0	0.049246	0	0	0	0	0.987678	0	1	-360	360;
	8390	7962	0	0.049246	0	0	0	0	0.987678	0	1	-360	360;
	8390	7962	0	0.049246	0	0	0	0	0.987678	0	1	-360	360;
	8793	6198	0	0.046721	0	0	0	0	0.974792	0	1	-360	360;
	8793	5345	0	0.047975	0	0	0	0	1.001022	0	1	-360	360;
	836	7084	0	0.047974	0	0	0	0	1.001022	0	1	-360	360;
	836	7084	0	0.047974	0	0	0	0	1.001022	0	1	-360	360;
	1941	8122	0	0.049246	0	0	0	0	0.987678	0	1	-360	360;
	1941	8122	0	0.049246	0	0	0	0	0.987678	0	1	-360	360;
	1941	7288	0	0.047975	0	0	0	0	0.987675	0	1	-360	360;
	1941	7288	0	0.047975	0	0	0	0	0.987675	0	1	-360	360;
	8089	4769	0	0.047975	0	0	0	0	0.987675	0	1	-360	360;
	8089	4769	0	0.047975	0	0	0	0	0.987675	0	1	-360	360;
	836	1995	0	0.045482	0	0	0	0	1.014736	0	1	-360	360;
	836	1995	0	0.045482	0	0	0	0	1.014736	0	1	-360	360;
	1594	1291	0	0.048	0	0	0	0	0.975	0	1	-360	360;
	1594	1291	0	0.048	0	0	0	0	0.975	0	1	-360	360;
	1594	3783	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	1594	3783	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	1420	1799	0	0.049041	0	0	0	0	0.95	0	1	-360	360;
	1420	1799	0	0.049041	0	0	0	0	0.95	0	1	-360	360;
	1420	4303	0	0.047975	0	0	0	0	0.95	0	1	-360	360;
	1420	4303	0	0.047975	0	0	0	0	0.95	0	1	-360	360;
	8680	8366	0	0.049246	0	0	0	0	0.962025	0	1	-360	360;
	8680	8366	0	0.049246	0	0	0	0	0.962025	0	1	-360	360;
	1827	4292	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	1827	4292	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	1827	5113	0	0.049246	0	0	0	0	0.962025	0	1	-360	360;
	1827	5113	0	0.049246	0	0	0	0	0.962025	0	1	-360	360;
	644	319	0	0.05533	0	0	0	0	0.975	0	1	-360	360;
	644	319	0	0.05533	0	0	0	0	0.975	0	1	-360	360;
	644	319	0	0.05533	0	0	0	0	0.975	0	1	-360	360;
	644	319	0	0.05533	0	0	0	0	0.975	0	1	-360	360;
	4888	7888	0	0.047975	0	0	0	0	0.987675	0	1	-360	360;
	4888	7888	0	0.047975	0	0	0	0	0.987675	0	1	-360	360;
	4917	777	0	0.0505	0	0	0	0	0.962289	0	1	-360	360;
	4917	777	0	0.0505	0	0	0	0	0.962289	0	1	-360	360;
	4917	777	0	0.0505	0	0	0	0	0.962289	0	1	-360	360;
	8657	4628	0	0.048	0	0	0	0	0.987294	0	1	-360	360;
	8657	4628	0	0.04563	0	0	0	0	0.987013	0	1	-360	360;
	8657	1444	0	0.049141	0	0	0	0	0.964134	0	1	-360	360;
	8657	1444	0	0.049141	0	0	0	0	0.964134	0	1	-360	360;
	1881	3459	0	0.067933	0	0	0	0	0	0	1	-360	360;
	1881	3459	0	0.067933	0	0	0	0	0	0	1	-360	360;
	1881	3459	0	0.047975	0	0	0	0	0.987675	0	1	-360	360;
	5293	4955	0	0.046721	0	0	0	0	0.962237	0	1	-360	360;
	7429	9220	0	0.070733	0	0	0	0	0.974792	0	1	-360	360;
	7429	9220	0	0.07187	0	0	0	0	0.97375	0	1	-360	360;
	7658	1080	0	0.049246	0	0	0	0	0.95	0	1	-360	360;
	7658	1080	0	0.049246	0	0	0	0	0.95	0	1	-360	360;
	7658	1080	0	0.049246	0	0	0	0	0.95	0	1	-360	360;
	6021	3002	0	0.0505	0	0	0	0	0.962289	0	1	-360	360;
	6021	3002	0	0.0505	0	0	0	0	0.962289	0	1	-360	360;
	6021	4193	0	0.046807	0	0	0	0	0.987013	0	1	-360	360;
	6021	4193	0	0.046807	0	0	0	0	0.987013	0	1	-360	360;
	6021	3058	0	0.049246	0	0	0	0	0.987678	0	1	-360	360;
	6021	3058	0	0.049246	0	0	0	0	0.987678	0	1	-360	360;
	4567	2735	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	4567	2735	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	4567	9085	0	0.0505	0	0	0	0	0.962289	0	1	-360	360;
	4567	9085	0	0.0505	0	0	0	0	0.962289	0	1	-360	360;
	2659	2999	0	0.047975	0	0	0	0	0.987675	0	1	-360	360;
	2659	2999	0	0.047975	0	0	0	0	0.987675	0	1	-360	360;
	2659	2999	0	0.047975	0	0	0	0	0.987675	0	1	-360	360;
	2659	2999	0	0.047975	0	0	0	0	0.987675	0	1	-360	360;
	2775	6890	0	0.05643	0	0	0	0	1.032609	0	1	-360	360;
	2775	6890	0	0.040627	0	0	0	0	1.032609	0	1	-360	360;
	2683	409	0	0.049246	0	0	0	0	0.95	0	1	-360	360;
	2683	409	0	0.049246	0	0	0	0	0.95	0	1	-360	360;
	926	2182	0	0.092641	0	0	0	0	0.987013	0	1	-360	360;
	926	2182	0	0.092641	0	0	0	0	0.987013	0	1	-360	360;
	926	2182	0	0.05415	0	0	0	0	0	0	1	-360	360;
	7728	1800	0	0.044468	0	0	0	0	0	0	1	-360	360;
	7728	1800	0	0.044468	0	0	0	0	0	0	1	-360	360;
	2455	2791	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	2455	2791	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	2455	2791	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	7498	5411	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	7498	5411	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	7498	5411	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	6181	5851	0	0.06556	0	0	0	0	0.97375	0	1	-360	360;
	6181	5851	0	0.0472	0	0	0	0	0.97375	0	1	-360	360;
	2436	6821	0	0.075233	0	0	0	0	0.95	0	1	-360	360;
	2436	6821	0	0.043977	0	0	0	0	0.962181	0	1	-360	360;
	2436	6821	0	0.043977	0	0	0	0	0.962181	0	1	-360	360;
	2436	6821	0	0.043977	0	0	0	0	0.962181	0	1	-360	360;
	7158	7481	0	0.047783	0	0	0	0	0.962025	0	1	-360	360;
	7158	7481	0	0.046807	0	0	0	0	0.962025	0	1	-360	360;
	7985	4109	0	0.071122	0	0	0	0	0.974792	0	1	-360	360;
	7985	4109	0	0.071122	0	0	0	0	0.974792	0	1	-360	360;
	7985	4109	0	0.07796	0	0	0	0	0.962181	0	1	-360	360;
	7985	4109	0	0.07796	0	0	0	0	0.962181	0	1	-360	360;
	5921	2881	0	0.046721	0	0	0	0	0.974792	0	1	-360	360;
	6518	2499	0	0.047975	0	0	0	0	0.962181	0	1	-360	360;
	134	2243	0	0.046721	0	0	0	0	0.974792	0	1	-360	360;
	1738	9	0	0.046721	0	0	0	0	0.98768	0	1	-360	360;
	2151	2483	0	0.08407	0	0	0	0	0.974587	0	1	-360	360;
	2151	2483	0	0.08407	0	0	0	0	0.974587	0	1	-360	360;
	2151	2483	0	0.08407	0	0	0	0	0.974587	0	1	-360	360;
	4920	6044	0	0.050122	0	0	0	0	0.95	0	1	-360	360;
	4920	6044	0	0.050122	0	0	0	0	0.95	0	1	-360	360;
	4920	3564	0	0.047176	0	0	0	0	0.962181	0	1	-360	360;
	4920	3564	0	0.047176	0	0	0	0	0.962181	0	1	-360	360;
	4305	5603	0	0.046721	0	0	0	0	0.974792	0	1	-360	360;
	4305	4749	0	0.046721	0	0	0	0	0.974792	0	1	-360	360;
	2325	2324	0	0.053184	0	0	0	0	0.95	0	1	-360	360;
	2325	2324	0	0.053184	0	0	0	0	0.95	0	1	-360	360;
	2325	2324	0	0.053184	0	0	0	0	0.95	0	1	-360	360;
	1152	8543	0	0.0505	0	0	0	0	0.914719	0	1	-360	360;
	1152	8543	0	0.0505	0	0	0	0	0.914719	0	1	-360	360;
	7928	8259	0	0.119999	0	0	0	0	0.95	0	1	-360	360;
	7928	8259	0	0.119999	0	0	0	0	0.95	0	1	-360	360;
	634	4358	0	0.074556	0	0	0	0	0.974683	0	1	-360	360;
	634	4358	0	0.074556	0	0	0	0	0.974683	0	1	-360	360;
	634	4358	0	0.074556	0	0	0	0	0.974683	0	1	-360	360;
	634	4358	0	0.078311	0	0	0	0	0.962025	0	1	-360	360;
	9022	5287	0	0.04602	0	0	0	0	0.974792	0	1	-360	360;
	9022	5287	0	0.08441	0	0	0	0	0.97375	0	1	-360	360;
	1925	8523	0	0.06561	0	0	0	0	0.97375	0	1	-360	360;
	1925	8523	0	0.07333	0	0	0	0	0.97375	0	1	-360	360;
	1515	50	0	0.049246	0	0	0	0	0.974683	0	1	-360	360;
	1515	50	0	0.049246	0	0	0	0	0.974683	0	1	-360	360;
	5934	5040	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	5934	5040	0	0.082076	0	0	0	0	0.962025	0	1	-360	360;
	5934	5040	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	8227	6796	0	0.076827	0	0	0	0	0.962025	0	1	-360	360;
	8227	6796	0	0.049734	0	0	0	0	0.962181	0	1	-360	360;
	6124	5810	0	0.0808	0	0	0	0	0.98363	0	1	-360	360;
	6124	5810	0	0.073292	0	0	0	0	0.962181	0	1	-360	360;
	6124	5810	0	0.08451	0	0	0	0	0.973754	0	1	-360	360;
	6124	5810	0	0.073292	0	0	0	0	0.962181	0	1	-360	360;
	7378	2007	0	0.047975	0	0	0	0	0.95	0	1	-360	360;
	7378	2007	0	0.047975	0	0	0	0	0.95	0	1	-360	360;
	3475	3799	0	0.070727	0	0	0	0	0.974359	0	1	-360	360;
	3475	3799	0	0.072561	0	0	0	0	0.974359	0	1	-360	360;
	6893	7217	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	6893	7217	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	7128	5667	0	0.07575	0	0	0	0	0.962289	0	1	-360	360;
	7128	5667	0	0.049246	0	0	0	0	0.974683	0	1	-360	360;
	7128	5667	0	0.044468	0	0	0	0	0	0	1	-360	360;
	4225	6951	0	0.082076	0	0	0	0	0.95	0	1	-360	360;
	4225	6951	0	0.046721	0	0	0	0	0.95	0	1	-360	360;
	7026	5889	0	0.073292	0	0	0	0	0.962181	0	1	-360	360;
	7026	5889	0	0.07618	0	0	0	0	0.974359	0	1	-360	360;
	320	20	0	0.034784	0	0	0	0	0.987013	0	1	-360	360;
	320	20	0	0.041737	0	0	0	0	0.987013	0	1	-360	360;
	715	1054	0	0.0842	0	0	0	0	0.97375	0	1	-360	360;
	715	1054	0	0.0842	0	0	0	0	0.97375	0	1	-360	360;
	2874	3660	0	0.048	0	0	0	0	0.9625	0	1	-360	360;
	2874	3660	0	0.048	0	0	0	0	0.9625	0	1	-360	360;
	2874	3660	0	0.048	0	0	0	0	0.9625	0	1	-360	360;
	2874	3660	0	0.048	0	0	0	0	0.9625	0	1	-360	360;
	3907	7894	0	0.080704	0	0	0	0	0.974687	0	1	-360	360;
	2241	2584	0	0.033909	0	0	0	0	0	0	1	-360	360;
	6008	6326	0	0.075925	0	0	0	0	0.949102	0	1	-360	360;
	6008	6326	0	0.075925	0	0	0	0	0.949102	0	1	-360	360;
	6008	6326	0	0.08451	0	0	0	0	0.987917	0	1	-360	360;
	9114	5538	0	0.082076	0	0	0	0	0.974683	0	1	-360	360;
	9114	5538	0	0.046721	0	0	0	0	0.98768	0	1	-360	360;
	9114	5538	0	0.046721	0	0	0	0	0.98768	0	1	-360	360;
	4469	5590	0	0.046417	0	0	0	0	0	0	1	-360	360;
	4469	5590	0	0.046417	0	0	0	0	0	0	1	-360	360;
	5098	3822	0	0.077314	0	0	0	0	0.974359	0	1	-360	360;
	5098	3822	0	0.04639	0	0	0	0	0.987013	0	1	-360	360;
	3194	1559	0	0.046721	0	0	0	0	1.001026	0	1	-360	360;
	407	723	0	0.071508	0	0	0	0	0.962025	0	1	-360	360;
	407	723	0	0.071508	0	0	0	0	0.962025	0	1	-360	360;
	407	723	0	0.042907	0	0	0	0	0.974359	0	1	-360	360;
	407	723	0	0.042907	0	0	0	0	0.974359	0	1	-360	360;
	6085	5744	0	0.079232	0	0	0	0	0.95	0	1	-360	360;
	6746	6410	0	0.045483	0	0	0	0	0.98776	0	1	-360	360;
	7642	7950	0	0.045483	0	0	0	0	0.974844	0	1	-360	360;
	7642	7950	0	0.045483	0	0	0	0	0.974844	0	1	-360	360;
	5581	5240	0	0.045483	0	0	0	0	0.98776	0	1	-360	360;
	8422	3260	0	0.047975	0	0	0	0	0.962181	0	1	-360	360;
	8422	3260	0	0.047975	0	0	0	0	0.962181	0	1	-360	360;
	4050	4370	0	0.044262	0	0	0	0	0.987813	0	1	-360	360;
	2028	2370	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	2028	2370	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	2028	2370	0	0.047975	0	0	0	0	0.974679	0	1	-360	360;
	294	9224	0	0.046721	0	0	0	0	0.98768	0	1	-360	360;
	294	9224	0	0.046721	0	0	0	0	0.98768	0	1	-360	360;
	3292	3612	0	0.047975	0	0	0	0	0.962181	0	1	-360	360;
	3292	3612	0	0.047975	0	0	0	0	0.962181	0	1	-360	360;
	1929	5090	0	0.0505	0	0	0	0	0.95	0	1	-360	360;
	1929	5090	0	0.0505	0	0	0	0	0.95	0	1	-360	360;
	1929	5090	0	0.0505	0	0	0	0	0.95	0	1	-360	360;
	8110	8441	0	0.070313	0	0	0	0	0.95	0	1	-360	360;
];

%%-----  OPF Data  -----%%
%% generator cost data
%	1	startup	shutdown	n	x1	y1	...	xn	yn
%	2	startup	shutdown	n	c(n-1)	...	c0
mpc.gencost = [
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
];
